<!DOCTYPE html>
<html lang="en">
<head>
  <!-- Compiled and minified CSS -->
    <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/materialize/1.0.0/css/materialize.min.css">

    <!-- Compiled and minified JavaScript -->
    <script src="https://cdnjs.cloudflare.com/ajax/libs/materialize/1.0.0/js/materialize.min.js"></script>
    <title>SBML Validation of {{ model }}</title>
</head>
<body>
    <h1>SBML Validation</h1>
    <div>
    {% for error in notifications.errors %}
        <div class="card-panel red lighten-4"> {{ error }} </div>
    {% endfor %}
    </div>
    <div>
    {% for warning in notifications.warnings %}
        <div class="card-panel yellow lighten-4"> {{ warning }} </div>
    {% endfor %}
    </div>
<a href="https://github.com/opencobra/memote">
</body>
</html>