<!doctype html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <title>MemoteReportApp</title>
  <base href="/">

  <meta name="viewport" content="width=device-width, initial-scale=1">
  <link rel="icon" type="image/x-icon" href="data:image/vnd.microsoft.icon;base64, ">
  <script type="text/javascript" async src="https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.5/latest.js?config=TeX-MML-AM_CHTML">
</script>
<script type="text/x-mathjax-config"> 
  MathJax.Hub.Config({ 
      "HTML-CSS": { scale: 50, linebreaks: { automatic: true } }, 
      SVG: { linebreaks: { automatic:true } }, 
      displayAlign: "left" });
</script>
</head>
<body style="background-color: rgba(177, 212, 237, 0.1)">
  <app-root></app-root>
  <script>
    window.data = $results;
    window.reportType = '$report_type';
  </script>
<script type="text/javascript">
!function(e){function r(r){for(var n,i,f=r[0],l=r[1],a=r[2],c=0,s=[];c<f.length;c++)i=f[c],o[i]&&s.push(o[i][0]),o[i]=0;for(n in l)Object.prototype.hasOwnProperty.call(l,n)&&(e[n]=l[n]);for(p&&p(r);s.length;)s.shift()();return u.push.apply(u,a||[]),t()}function t(){for(var e,r=0;r<u.length;r++){for(var t=u[r],n=!0,f=1;f<t.length;f++){var l=t[f];0!==o[l]&&(n=!1)}n&&(u.splice(r--,1),e=i(i.s=t[0]))}return e}var n={},o={runtime:0},u=[];function i(r){if(n[r])return n[r].exports;var t=n[r]={i:r,l:!1,exports:{}};return e[r].call(t.exports,t,t.exports,i),t.l=!0,t.exports}i.m=e,i.c=n,i.d=function(e,r,t){i.o(e,r)||Object.defineProperty(e,r,{enumerable:!0,get:t})},i.r=function(e){"undefined"!=typeof Symbol&&Symbol.toStringTag&&Object.defineProperty(e,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(e,"__esModule",{value:!0})},i.t=function(e,r){if(1&r&&(e=i(e)),8&r)return e;if(4&r&&"object"==typeof e&&e&&e.__esModule)return e;var t=Object.create(null);if(i.r(t),Object.defineProperty(t,"default",{enumerable:!0,value:e}),2&r&&"string"!=typeof e)for(var n in e)i.d(t,n,function(r){return e[r]}.bind(null,n));return t},i.n=function(e){var r=e&&e.__esModule?function(){return e.default}:function(){return e};return i.d(r,"a",r),r},i.o=function(e,r){return Object.prototype.hasOwnProperty.call(e,r)},i.p="";var f=window.webpackJsonp=window.webpackJsonp||[],l=f.push.bind(f);f.push=r,f=f.slice();for(var a=0;a<f.length;a++)r(f[a]);var p=l;t()}([]);
</script><script type="text/javascript">
(window.webpackJsonp=window.webpackJsonp||[]).push([["polyfills"],{"./node_modules/@angular-devkit/build-angular/src/angular-cli-files/models/jit-polyfills.js":function(e,o,t){"use strict";t.r(o);t("./node_modules/core-js/es7/reflect.js")},"./node_modules/core-js/es7/reflect.js":function(e,o,t){t("./node_modules/core-js/modules/es7.reflect.define-metadata.js"),t("./node_modules/core-js/modules/es7.reflect.delete-metadata.js"),t("./node_modules/core-js/modules/es7.reflect.get-metadata.js"),t("./node_modules/core-js/modules/es7.reflect.get-metadata-keys.js"),t("./node_modules/core-js/modules/es7.reflect.get-own-metadata.js"),t("./node_modules/core-js/modules/es7.reflect.get-own-metadata-keys.js"),t("./node_modules/core-js/modules/es7.reflect.has-metadata.js"),t("./node_modules/core-js/modules/es7.reflect.has-own-metadata.js"),t("./node_modules/core-js/modules/es7.reflect.metadata.js"),e.exports=t("./node_modules/core-js/modules/_core.js").Reflect},"./node_modules/core-js/modules/_a-function.js":function(e,o){e.exports=function(e){if("function"!=typeof e)throw TypeError(e+" is not a function!");return e}},"./node_modules/core-js/modules/_an-instance.js":function(e,o){e.exports=function(e,o,t,n){if(!(e instanceof o)||void 0!==n&&n in e)throw TypeError(t+": incorrect invocation!");return e}},"./node_modules/core-js/modules/_an-object.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_is-object.js");e.exports=function(e){if(!n(e))throw TypeError(e+" is not an object!");return e}},"./node_modules/core-js/modules/_array-from-iterable.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_for-of.js");e.exports=function(e,o){var t=[];return n(e,!1,t.push,t,o),t}},"./node_modules/core-js/modules/_array-includes.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_to-iobject.js"),r=t("./node_modules/core-js/modules/_to-length.js"),s=t("./node_modules/core-js/modules/_to-absolute-index.js");e.exports=function(e){return function(o,t,i){var a,u=n(o),c=r(u.length),l=s(i,c);if(e&&t!=t){for(;c>l;)if((a=u[l++])!=a)return!0}else for(;c>l;l++)if((e||l in u)&&u[l]===t)return e||l||0;return!e&&-1}}},"./node_modules/core-js/modules/_array-methods.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_ctx.js"),r=t("./node_modules/core-js/modules/_iobject.js"),s=t("./node_modules/core-js/modules/_to-object.js"),i=t("./node_modules/core-js/modules/_to-length.js"),a=t("./node_modules/core-js/modules/_array-species-create.js");e.exports=function(e,o){var t=1==e,u=2==e,c=3==e,l=4==e,d=6==e,f=5==e||d,m=o||a;return function(o,a,p){for(var _,h,j=s(o),v=r(j),g=n(a,p,3),y=i(v.length),b=0,k=t?m(o,y):u?m(o,0):void 0;y>b;b++)if((f||b in v)&&(h=g(_=v[b],b,j),e))if(t)k[b]=h;else if(h)switch(e){case 3:return!0;case 5:return _;case 6:return b;case 2:k.push(_)}else if(l)return!1;return d?-1:c||l?l:k}}},"./node_modules/core-js/modules/_array-species-constructor.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_is-object.js"),r=t("./node_modules/core-js/modules/_is-array.js"),s=t("./node_modules/core-js/modules/_wks.js")("species");e.exports=function(e){var o;return r(e)&&("function"!=typeof(o=e.constructor)||o!==Array&&!r(o.prototype)||(o=void 0),n(o)&&null===(o=o[s])&&(o=void 0)),void 0===o?Array:o}},"./node_modules/core-js/modules/_array-species-create.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_array-species-constructor.js");e.exports=function(e,o){return new(n(e))(o)}},"./node_modules/core-js/modules/_classof.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_cof.js"),r=t("./node_modules/core-js/modules/_wks.js")("toStringTag"),s="Arguments"==n(function(){return arguments}());e.exports=function(e){var o,t,i;return void 0===e?"Undefined":null===e?"Null":"string"==typeof(t=function(e,o){try{return e[o]}catch(e){}}(o=Object(e),r))?t:s?n(o):"Object"==(i=n(o))&&"function"==typeof o.callee?"Arguments":i}},"./node_modules/core-js/modules/_cof.js":function(e,o){var t={}.toString;e.exports=function(e){return t.call(e).slice(8,-1)}},"./node_modules/core-js/modules/_collection-strong.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_object-dp.js").f,r=t("./node_modules/core-js/modules/_object-create.js"),s=t("./node_modules/core-js/modules/_redefine-all.js"),i=t("./node_modules/core-js/modules/_ctx.js"),a=t("./node_modules/core-js/modules/_an-instance.js"),u=t("./node_modules/core-js/modules/_for-of.js"),c=t("./node_modules/core-js/modules/_iter-define.js"),l=t("./node_modules/core-js/modules/_iter-step.js"),d=t("./node_modules/core-js/modules/_set-species.js"),f=t("./node_modules/core-js/modules/_descriptors.js"),m=t("./node_modules/core-js/modules/_meta.js").fastKey,p=t("./node_modules/core-js/modules/_validate-collection.js"),_=f?"_s":"size",h=function(e,o){var t,n=m(o);if("F"!==n)return e._i[n];for(t=e._f;t;t=t.n)if(t.k==o)return t};e.exports={getConstructor:function(e,o,t,c){var l=e(function(e,n){a(e,l,o,"_i"),e._t=o,e._i=r(null),e._f=void 0,e._l=void 0,e[_]=0,void 0!=n&&u(n,t,e[c],e)});return s(l.prototype,{clear:function(){for(var e=p(this,o),t=e._i,n=e._f;n;n=n.n)n.r=!0,n.p&&(n.p=n.p.n=void 0),delete t[n.i];e._f=e._l=void 0,e[_]=0},delete:function(e){var t=p(this,o),n=h(t,e);if(n){var r=n.n,s=n.p;delete t._i[n.i],n.r=!0,s&&(s.n=r),r&&(r.p=s),t._f==n&&(t._f=r),t._l==n&&(t._l=s),t[_]--}return!!n},forEach:function(e){p(this,o);for(var t,n=i(e,arguments.length>1?arguments[1]:void 0,3);t=t?t.n:this._f;)for(n(t.v,t.k,this);t&&t.r;)t=t.p},has:function(e){return!!h(p(this,o),e)}}),f&&n(l.prototype,"size",{get:function(){return p(this,o)[_]}}),l},def:function(e,o,t){var n,r,s=h(e,o);return s?s.v=t:(e._l=s={i:r=m(o,!0),k:o,v:t,p:n=e._l,n:void 0,r:!1},e._f||(e._f=s),n&&(n.n=s),e[_]++,"F"!==r&&(e._i[r]=s)),e},getEntry:h,setStrong:function(e,o,t){c(e,o,function(e,t){this._t=p(e,o),this._k=t,this._l=void 0},function(){for(var e=this._k,o=this._l;o&&o.r;)o=o.p;return this._t&&(this._l=o=o?o.n:this._t._f)?l(0,"keys"==e?o.k:"values"==e?o.v:[o.k,o.v]):(this._t=void 0,l(1))},t?"entries":"values",!t,!0),d(o)}}},"./node_modules/core-js/modules/_collection-weak.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_redefine-all.js"),r=t("./node_modules/core-js/modules/_meta.js").getWeak,s=t("./node_modules/core-js/modules/_an-object.js"),i=t("./node_modules/core-js/modules/_is-object.js"),a=t("./node_modules/core-js/modules/_an-instance.js"),u=t("./node_modules/core-js/modules/_for-of.js"),c=t("./node_modules/core-js/modules/_array-methods.js"),l=t("./node_modules/core-js/modules/_has.js"),d=t("./node_modules/core-js/modules/_validate-collection.js"),f=c(5),m=c(6),p=0,_=function(e){return e._l||(e._l=new h)},h=function(){this.a=[]},j=function(e,o){return f(e.a,function(e){return e[0]===o})};h.prototype={get:function(e){var o=j(this,e);if(o)return o[1]},has:function(e){return!!j(this,e)},set:function(e,o){var t=j(this,e);t?t[1]=o:this.a.push([e,o])},delete:function(e){var o=m(this.a,function(o){return o[0]===e});return~o&&this.a.splice(o,1),!!~o}},e.exports={getConstructor:function(e,o,t,s){var c=e(function(e,n){a(e,c,o,"_i"),e._t=o,e._i=p++,e._l=void 0,void 0!=n&&u(n,t,e[s],e)});return n(c.prototype,{delete:function(e){if(!i(e))return!1;var t=r(e);return!0===t?_(d(this,o)).delete(e):t&&l(t,this._i)&&delete t[this._i]},has:function(e){if(!i(e))return!1;var t=r(e);return!0===t?_(d(this,o)).has(e):t&&l(t,this._i)}}),c},def:function(e,o,t){var n=r(s(o),!0);return!0===n?_(e).set(o,t):n[e._i]=t,e},ufstore:_}},"./node_modules/core-js/modules/_collection.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_global.js"),r=t("./node_modules/core-js/modules/_export.js"),s=t("./node_modules/core-js/modules/_redefine.js"),i=t("./node_modules/core-js/modules/_redefine-all.js"),a=t("./node_modules/core-js/modules/_meta.js"),u=t("./node_modules/core-js/modules/_for-of.js"),c=t("./node_modules/core-js/modules/_an-instance.js"),l=t("./node_modules/core-js/modules/_is-object.js"),d=t("./node_modules/core-js/modules/_fails.js"),f=t("./node_modules/core-js/modules/_iter-detect.js"),m=t("./node_modules/core-js/modules/_set-to-string-tag.js"),p=t("./node_modules/core-js/modules/_inherit-if-required.js");e.exports=function(e,o,t,_,h,j){var v=n[e],g=v,y=h?"set":"add",b=g&&g.prototype,k={},w=function(e){var o=b[e];s(b,e,"delete"==e?function(e){return!(j&&!l(e))&&o.call(this,0===e?0:e)}:"has"==e?function(e){return!(j&&!l(e))&&o.call(this,0===e?0:e)}:"get"==e?function(e){return j&&!l(e)?void 0:o.call(this,0===e?0:e)}:"add"==e?function(e){return o.call(this,0===e?0:e),this}:function(e,t){return o.call(this,0===e?0:e,t),this})};if("function"==typeof g&&(j||b.forEach&&!d(function(){(new g).entries().next()}))){var T=new g,E=T[y](j?{}:-0,1)!=T,x=d(function(){T.has(1)}),O=f(function(e){new g(e)}),S=!j&&d(function(){for(var e=new g,o=5;o--;)e[y](o,o);return!e.has(-0)});O||((g=o(function(o,t){c(o,g,e);var n=p(new v,o,g);return void 0!=t&&u(t,h,n[y],n),n})).prototype=b,b.constructor=g),(x||S)&&(w("delete"),w("has"),h&&w("get")),(S||E)&&w(y),j&&b.clear&&delete b.clear}else g=_.getConstructor(o,e,h,y),i(g.prototype,t),a.NEED=!0;return m(g,e),k[e]=g,r(r.G+r.W+r.F*(g!=v),k),j||_.setStrong(g,e,h),g}},"./node_modules/core-js/modules/_core.js":function(e,o){var t=e.exports={version:"2.5.1"};"number"==typeof __e&&(__e=t)},"./node_modules/core-js/modules/_ctx.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_a-function.js");e.exports=function(e,o,t){if(n(e),void 0===o)return e;switch(t){case 1:return function(t){return e.call(o,t)};case 2:return function(t,n){return e.call(o,t,n)};case 3:return function(t,n,r){return e.call(o,t,n,r)}}return function(){return e.apply(o,arguments)}}},"./node_modules/core-js/modules/_defined.js":function(e,o){e.exports=function(e){if(void 0==e)throw TypeError("Can't call method on  "+e);return e}},"./node_modules/core-js/modules/_descriptors.js":function(e,o,t){e.exports=!t("./node_modules/core-js/modules/_fails.js")(function(){return 7!=Object.defineProperty({},"a",{get:function(){return 7}}).a})},"./node_modules/core-js/modules/_dom-create.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_is-object.js"),r=t("./node_modules/core-js/modules/_global.js").document,s=n(r)&&n(r.createElement);e.exports=function(e){return s?r.createElement(e):{}}},"./node_modules/core-js/modules/_enum-bug-keys.js":function(e,o){e.exports="constructor,hasOwnProperty,isPrototypeOf,propertyIsEnumerable,toLocaleString,toString,valueOf".split(",")},"./node_modules/core-js/modules/_export.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_global.js"),r=t("./node_modules/core-js/modules/_core.js"),s=t("./node_modules/core-js/modules/_hide.js"),i=t("./node_modules/core-js/modules/_redefine.js"),a=t("./node_modules/core-js/modules/_ctx.js"),u="prototype",c=function(e,o,t){var l,d,f,m,p=e&c.F,_=e&c.G,h=e&c.S,j=e&c.P,v=e&c.B,g=_?n:h?n[o]||(n[o]={}):(n[o]||{})[u],y=_?r:r[o]||(r[o]={}),b=y[u]||(y[u]={});_&&(t=o);for(l in t)f=((d=!p&&g&&void 0!==g[l])?g:t)[l],m=v&&d?a(f,n):j&&"function"==typeof f?a(Function.call,f):f,g&&i(g,l,f,e&c.U),y[l]!=f&&s(y,l,m),j&&b[l]!=f&&(b[l]=f)};n.core=r,c.F=1,c.G=2,c.S=4,c.P=8,c.B=16,c.W=32,c.U=64,c.R=128,e.exports=c},"./node_modules/core-js/modules/_fails.js":function(e,o){e.exports=function(e){try{return!!e()}catch(e){return!0}}},"./node_modules/core-js/modules/_for-of.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_ctx.js"),r=t("./node_modules/core-js/modules/_iter-call.js"),s=t("./node_modules/core-js/modules/_is-array-iter.js"),i=t("./node_modules/core-js/modules/_an-object.js"),a=t("./node_modules/core-js/modules/_to-length.js"),u=t("./node_modules/core-js/modules/core.get-iterator-method.js"),c={},l={};(o=e.exports=function(e,o,t,d,f){var m,p,_,h,j=f?function(){return e}:u(e),v=n(t,d,o?2:1),g=0;if("function"!=typeof j)throw TypeError(e+" is not iterable!");if(s(j)){for(m=a(e.length);m>g;g++)if((h=o?v(i(p=e[g])[0],p[1]):v(e[g]))===c||h===l)return h}else for(_=j.call(e);!(p=_.next()).done;)if((h=r(_,v,p.value,o))===c||h===l)return h}).BREAK=c,o.RETURN=l},"./node_modules/core-js/modules/_global.js":function(e,o){var t=e.exports="undefined"!=typeof window&&window.Math==Math?window:"undefined"!=typeof self&&self.Math==Math?self:Function("return this")();"number"==typeof __g&&(__g=t)},"./node_modules/core-js/modules/_has.js":function(e,o){var t={}.hasOwnProperty;e.exports=function(e,o){return t.call(e,o)}},"./node_modules/core-js/modules/_hide.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_object-dp.js"),r=t("./node_modules/core-js/modules/_property-desc.js");e.exports=t("./node_modules/core-js/modules/_descriptors.js")?function(e,o,t){return n.f(e,o,r(1,t))}:function(e,o,t){return e[o]=t,e}},"./node_modules/core-js/modules/_html.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_global.js").document;e.exports=n&&n.documentElement},"./node_modules/core-js/modules/_ie8-dom-define.js":function(e,o,t){e.exports=!t("./node_modules/core-js/modules/_descriptors.js")&&!t("./node_modules/core-js/modules/_fails.js")(function(){return 7!=Object.defineProperty(t("./node_modules/core-js/modules/_dom-create.js")("div"),"a",{get:function(){return 7}}).a})},"./node_modules/core-js/modules/_inherit-if-required.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_is-object.js"),r=t("./node_modules/core-js/modules/_set-proto.js").set;e.exports=function(e,o,t){var s,i=o.constructor;return i!==t&&"function"==typeof i&&(s=i.prototype)!==t.prototype&&n(s)&&r&&r(e,s),e}},"./node_modules/core-js/modules/_iobject.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_cof.js");e.exports=Object("z").propertyIsEnumerable(0)?Object:function(e){return"String"==n(e)?e.split(""):Object(e)}},"./node_modules/core-js/modules/_is-array-iter.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_iterators.js"),r=t("./node_modules/core-js/modules/_wks.js")("iterator"),s=Array.prototype;e.exports=function(e){return void 0!==e&&(n.Array===e||s[r]===e)}},"./node_modules/core-js/modules/_is-array.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_cof.js");e.exports=Array.isArray||function(e){return"Array"==n(e)}},"./node_modules/core-js/modules/_is-object.js":function(e,o){e.exports=function(e){return"object"==typeof e?null!==e:"function"==typeof e}},"./node_modules/core-js/modules/_iter-call.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_an-object.js");e.exports=function(e,o,t,r){try{return r?o(n(t)[0],t[1]):o(t)}catch(o){var s=e.return;throw void 0!==s&&n(s.call(e)),o}}},"./node_modules/core-js/modules/_iter-create.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_object-create.js"),r=t("./node_modules/core-js/modules/_property-desc.js"),s=t("./node_modules/core-js/modules/_set-to-string-tag.js"),i={};t("./node_modules/core-js/modules/_hide.js")(i,t("./node_modules/core-js/modules/_wks.js")("iterator"),function(){return this}),e.exports=function(e,o,t){e.prototype=n(i,{next:r(1,t)}),s(e,o+" Iterator")}},"./node_modules/core-js/modules/_iter-define.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_library.js"),r=t("./node_modules/core-js/modules/_export.js"),s=t("./node_modules/core-js/modules/_redefine.js"),i=t("./node_modules/core-js/modules/_hide.js"),a=t("./node_modules/core-js/modules/_has.js"),u=t("./node_modules/core-js/modules/_iterators.js"),c=t("./node_modules/core-js/modules/_iter-create.js"),l=t("./node_modules/core-js/modules/_set-to-string-tag.js"),d=t("./node_modules/core-js/modules/_object-gpo.js"),f=t("./node_modules/core-js/modules/_wks.js")("iterator"),m=!([].keys&&"next"in[].keys()),p="values",_=function(){return this};e.exports=function(e,o,t,h,j,v,g){c(t,o,h);var y,b,k,w=function(e){if(!m&&e in O)return O[e];switch(e){case"keys":case p:return function(){return new t(this,e)}}return function(){return new t(this,e)}},T=o+" Iterator",E=j==p,x=!1,O=e.prototype,S=O[f]||O["@@iterator"]||j&&O[j],P=S||w(j),D=j?E?w("entries"):P:void 0,z="Array"==o?O.entries||S:S;if(z&&(k=d(z.call(new e)))!==Object.prototype&&k.next&&(l(k,T,!0),n||a(k,f)||i(k,f,_)),E&&S&&S.name!==p&&(x=!0,P=function(){return S.call(this)}),n&&!g||!m&&!x&&O[f]||i(O,f,P),u[o]=P,u[T]=_,j)if(y={values:E?P:w(p),keys:v?P:w("keys"),entries:D},g)for(b in y)b in O||s(O,b,y[b]);else r(r.P+r.F*(m||x),o,y);return y}},"./node_modules/core-js/modules/_iter-detect.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_wks.js")("iterator"),r=!1;try{var s=[7][n]();s.return=function(){r=!0},Array.from(s,function(){throw 2})}catch(e){}e.exports=function(e,o){if(!o&&!r)return!1;var t=!1;try{var s=[7],i=s[n]();i.next=function(){return{done:t=!0}},s[n]=function(){return i},e(s)}catch(e){}return t}},"./node_modules/core-js/modules/_iter-step.js":function(e,o){e.exports=function(e,o){return{value:o,done:!!e}}},"./node_modules/core-js/modules/_iterators.js":function(e,o){e.exports={}},"./node_modules/core-js/modules/_library.js":function(e,o){e.exports=!1},"./node_modules/core-js/modules/_meta.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_uid.js")("meta"),r=t("./node_modules/core-js/modules/_is-object.js"),s=t("./node_modules/core-js/modules/_has.js"),i=t("./node_modules/core-js/modules/_object-dp.js").f,a=0,u=Object.isExtensible||function(){return!0},c=!t("./node_modules/core-js/modules/_fails.js")(function(){return u(Object.preventExtensions({}))}),l=function(e){i(e,n,{value:{i:"O"+ ++a,w:{}}})},d=e.exports={KEY:n,NEED:!1,fastKey:function(e,o){if(!r(e))return"symbol"==typeof e?e:("string"==typeof e?"S":"P")+e;if(!s(e,n)){if(!u(e))return"F";if(!o)return"E";l(e)}return e[n].i},getWeak:function(e,o){if(!s(e,n)){if(!u(e))return!0;if(!o)return!1;l(e)}return e[n].w},onFreeze:function(e){return c&&d.NEED&&u(e)&&!s(e,n)&&l(e),e}}},"./node_modules/core-js/modules/_metadata.js":function(e,o,t){var n=t("./node_modules/core-js/modules/es6.map.js"),r=t("./node_modules/core-js/modules/_export.js"),s=t("./node_modules/core-js/modules/_shared.js")("metadata"),i=s.store||(s.store=new(t("./node_modules/core-js/modules/es6.weak-map.js"))),a=function(e,o,t){var r=i.get(e);if(!r){if(!t)return;i.set(e,r=new n)}var s=r.get(o);if(!s){if(!t)return;r.set(o,s=new n)}return s};e.exports={store:i,map:a,has:function(e,o,t){var n=a(o,t,!1);return void 0!==n&&n.has(e)},get:function(e,o,t){var n=a(o,t,!1);return void 0===n?void 0:n.get(e)},set:function(e,o,t,n){a(t,n,!0).set(e,o)},keys:function(e,o){var t=a(e,o,!1),n=[];return t&&t.forEach(function(e,o){n.push(o)}),n},key:function(e){return void 0===e||"symbol"==typeof e?e:String(e)},exp:function(e){r(r.S,"Reflect",e)}}},"./node_modules/core-js/modules/_object-assign.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_object-keys.js"),r=t("./node_modules/core-js/modules/_object-gops.js"),s=t("./node_modules/core-js/modules/_object-pie.js"),i=t("./node_modules/core-js/modules/_to-object.js"),a=t("./node_modules/core-js/modules/_iobject.js"),u=Object.assign;e.exports=!u||t("./node_modules/core-js/modules/_fails.js")(function(){var e={},o={},t=Symbol(),n="abcdefghijklmnopqrst";return e[t]=7,n.split("").forEach(function(e){o[e]=e}),7!=u({},e)[t]||Object.keys(u({},o)).join("")!=n})?function(e,o){for(var t=i(e),u=arguments.length,c=1,l=r.f,d=s.f;u>c;)for(var f,m=a(arguments[c++]),p=l?n(m).concat(l(m)):n(m),_=p.length,h=0;_>h;)d.call(m,f=p[h++])&&(t[f]=m[f]);return t}:u},"./node_modules/core-js/modules/_object-create.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_an-object.js"),r=t("./node_modules/core-js/modules/_object-dps.js"),s=t("./node_modules/core-js/modules/_enum-bug-keys.js"),i=t("./node_modules/core-js/modules/_shared-key.js")("IE_PROTO"),a=function(){},u="prototype",c=function(){var e,o=t("./node_modules/core-js/modules/_dom-create.js")("iframe"),n=s.length;for(o.style.display="none",t("./node_modules/core-js/modules/_html.js").appendChild(o),o.src="javascript:",(e=o.contentWindow.document).open(),e.write("<script>document.F=Object<\/script>"),e.close(),c=e.F;n--;)delete c[u][s[n]];return c()};e.exports=Object.create||function(e,o){var t;return null!==e?(a[u]=n(e),t=new a,a[u]=null,t[i]=e):t=c(),void 0===o?t:r(t,o)}},"./node_modules/core-js/modules/_object-dp.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_an-object.js"),r=t("./node_modules/core-js/modules/_ie8-dom-define.js"),s=t("./node_modules/core-js/modules/_to-primitive.js"),i=Object.defineProperty;o.f=t("./node_modules/core-js/modules/_descriptors.js")?Object.defineProperty:function(e,o,t){if(n(e),o=s(o,!0),n(t),r)try{return i(e,o,t)}catch(e){}if("get"in t||"set"in t)throw TypeError("Accessors not supported!");return"value"in t&&(e[o]=t.value),e}},"./node_modules/core-js/modules/_object-dps.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_object-dp.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=t("./node_modules/core-js/modules/_object-keys.js");e.exports=t("./node_modules/core-js/modules/_descriptors.js")?Object.defineProperties:function(e,o){r(e);for(var t,i=s(o),a=i.length,u=0;a>u;)n.f(e,t=i[u++],o[t]);return e}},"./node_modules/core-js/modules/_object-gopd.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_object-pie.js"),r=t("./node_modules/core-js/modules/_property-desc.js"),s=t("./node_modules/core-js/modules/_to-iobject.js"),i=t("./node_modules/core-js/modules/_to-primitive.js"),a=t("./node_modules/core-js/modules/_has.js"),u=t("./node_modules/core-js/modules/_ie8-dom-define.js"),c=Object.getOwnPropertyDescriptor;o.f=t("./node_modules/core-js/modules/_descriptors.js")?c:function(e,o){if(e=s(e),o=i(o,!0),u)try{return c(e,o)}catch(e){}if(a(e,o))return r(!n.f.call(e,o),e[o])}},"./node_modules/core-js/modules/_object-gops.js":function(e,o){o.f=Object.getOwnPropertySymbols},"./node_modules/core-js/modules/_object-gpo.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_has.js"),r=t("./node_modules/core-js/modules/_to-object.js"),s=t("./node_modules/core-js/modules/_shared-key.js")("IE_PROTO"),i=Object.prototype;e.exports=Object.getPrototypeOf||function(e){return e=r(e),n(e,s)?e[s]:"function"==typeof e.constructor&&e instanceof e.constructor?e.constructor.prototype:e instanceof Object?i:null}},"./node_modules/core-js/modules/_object-keys-internal.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_has.js"),r=t("./node_modules/core-js/modules/_to-iobject.js"),s=t("./node_modules/core-js/modules/_array-includes.js")(!1),i=t("./node_modules/core-js/modules/_shared-key.js")("IE_PROTO");e.exports=function(e,o){var t,a=r(e),u=0,c=[];for(t in a)t!=i&&n(a,t)&&c.push(t);for(;o.length>u;)n(a,t=o[u++])&&(~s(c,t)||c.push(t));return c}},"./node_modules/core-js/modules/_object-keys.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_object-keys-internal.js"),r=t("./node_modules/core-js/modules/_enum-bug-keys.js");e.exports=Object.keys||function(e){return n(e,r)}},"./node_modules/core-js/modules/_object-pie.js":function(e,o){o.f={}.propertyIsEnumerable},"./node_modules/core-js/modules/_property-desc.js":function(e,o){e.exports=function(e,o){return{enumerable:!(1&e),configurable:!(2&e),writable:!(4&e),value:o}}},"./node_modules/core-js/modules/_redefine-all.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_redefine.js");e.exports=function(e,o,t){for(var r in o)n(e,r,o[r],t);return e}},"./node_modules/core-js/modules/_redefine.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_global.js"),r=t("./node_modules/core-js/modules/_hide.js"),s=t("./node_modules/core-js/modules/_has.js"),i=t("./node_modules/core-js/modules/_uid.js")("src"),a=Function.toString,u=(""+a).split("toString");t("./node_modules/core-js/modules/_core.js").inspectSource=function(e){return a.call(e)},(e.exports=function(e,o,t,a){var c="function"==typeof t;c&&(s(t,"name")||r(t,"name",o)),e[o]!==t&&(c&&(s(t,i)||r(t,i,e[o]?""+e[o]:u.join(String(o)))),e===n?e[o]=t:a?e[o]?e[o]=t:r(e,o,t):(delete e[o],r(e,o,t)))})(Function.prototype,"toString",function(){return"function"==typeof this&&this[i]||a.call(this)})},"./node_modules/core-js/modules/_set-proto.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_is-object.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=function(e,o){if(r(e),!n(o)&&null!==o)throw TypeError(o+": can't set as prototype!")};e.exports={set:Object.setPrototypeOf||("__proto__"in{}?function(e,o,n){try{(n=t("./node_modules/core-js/modules/_ctx.js")(Function.call,t("./node_modules/core-js/modules/_object-gopd.js").f(Object.prototype,"__proto__").set,2))(e,[]),o=!(e instanceof Array)}catch(e){o=!0}return function(e,t){return s(e,t),o?e.__proto__=t:n(e,t),e}}({},!1):void 0),check:s}},"./node_modules/core-js/modules/_set-species.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_global.js"),r=t("./node_modules/core-js/modules/_object-dp.js"),s=t("./node_modules/core-js/modules/_descriptors.js"),i=t("./node_modules/core-js/modules/_wks.js")("species");e.exports=function(e){var o=n[e];s&&o&&!o[i]&&r.f(o,i,{configurable:!0,get:function(){return this}})}},"./node_modules/core-js/modules/_set-to-string-tag.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_object-dp.js").f,r=t("./node_modules/core-js/modules/_has.js"),s=t("./node_modules/core-js/modules/_wks.js")("toStringTag");e.exports=function(e,o,t){e&&!r(e=t?e:e.prototype,s)&&n(e,s,{configurable:!0,value:o})}},"./node_modules/core-js/modules/_shared-key.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_shared.js")("keys"),r=t("./node_modules/core-js/modules/_uid.js");e.exports=function(e){return n[e]||(n[e]=r(e))}},"./node_modules/core-js/modules/_shared.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_global.js"),r="__core-js_shared__",s=n[r]||(n[r]={});e.exports=function(e){return s[e]||(s[e]={})}},"./node_modules/core-js/modules/_to-absolute-index.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_to-integer.js"),r=Math.max,s=Math.min;e.exports=function(e,o){return(e=n(e))<0?r(e+o,0):s(e,o)}},"./node_modules/core-js/modules/_to-integer.js":function(e,o){var t=Math.ceil,n=Math.floor;e.exports=function(e){return isNaN(e=+e)?0:(e>0?n:t)(e)}},"./node_modules/core-js/modules/_to-iobject.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_iobject.js"),r=t("./node_modules/core-js/modules/_defined.js");e.exports=function(e){return n(r(e))}},"./node_modules/core-js/modules/_to-length.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_to-integer.js"),r=Math.min;e.exports=function(e){return e>0?r(n(e),9007199254740991):0}},"./node_modules/core-js/modules/_to-object.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_defined.js");e.exports=function(e){return Object(n(e))}},"./node_modules/core-js/modules/_to-primitive.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_is-object.js");e.exports=function(e,o){if(!n(e))return e;var t,r;if(o&&"function"==typeof(t=e.toString)&&!n(r=t.call(e)))return r;if("function"==typeof(t=e.valueOf)&&!n(r=t.call(e)))return r;if(!o&&"function"==typeof(t=e.toString)&&!n(r=t.call(e)))return r;throw TypeError("Can't convert object to primitive value")}},"./node_modules/core-js/modules/_uid.js":function(e,o){var t=0,n=Math.random();e.exports=function(e){return"Symbol(".concat(void 0===e?"":e,")_",(++t+n).toString(36))}},"./node_modules/core-js/modules/_validate-collection.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_is-object.js");e.exports=function(e,o){if(!n(e)||e._t!==o)throw TypeError("Incompatible receiver, "+o+" required!");return e}},"./node_modules/core-js/modules/_wks.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_shared.js")("wks"),r=t("./node_modules/core-js/modules/_uid.js"),s=t("./node_modules/core-js/modules/_global.js").Symbol,i="function"==typeof s;(e.exports=function(e){return n[e]||(n[e]=i&&s[e]||(i?s:r)("Symbol."+e))}).store=n},"./node_modules/core-js/modules/core.get-iterator-method.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_classof.js"),r=t("./node_modules/core-js/modules/_wks.js")("iterator"),s=t("./node_modules/core-js/modules/_iterators.js");e.exports=t("./node_modules/core-js/modules/_core.js").getIteratorMethod=function(e){if(void 0!=e)return e[r]||e["@@iterator"]||s[n(e)]}},"./node_modules/core-js/modules/es6.map.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_collection-strong.js"),r=t("./node_modules/core-js/modules/_validate-collection.js");e.exports=t("./node_modules/core-js/modules/_collection.js")("Map",function(e){return function(){return e(this,arguments.length>0?arguments[0]:void 0)}},{get:function(e){var o=n.getEntry(r(this,"Map"),e);return o&&o.v},set:function(e,o){return n.def(r(this,"Map"),0===e?0:e,o)}},n,!0)},"./node_modules/core-js/modules/es6.set.js":function(e,o,t){"use strict";var n=t("./node_modules/core-js/modules/_collection-strong.js"),r=t("./node_modules/core-js/modules/_validate-collection.js");e.exports=t("./node_modules/core-js/modules/_collection.js")("Set",function(e){return function(){return e(this,arguments.length>0?arguments[0]:void 0)}},{add:function(e){return n.def(r(this,"Set"),e=0===e?0:e,e)}},n)},"./node_modules/core-js/modules/es6.weak-map.js":function(e,o,t){"use strict";var n,r=t("./node_modules/core-js/modules/_array-methods.js")(0),s=t("./node_modules/core-js/modules/_redefine.js"),i=t("./node_modules/core-js/modules/_meta.js"),a=t("./node_modules/core-js/modules/_object-assign.js"),u=t("./node_modules/core-js/modules/_collection-weak.js"),c=t("./node_modules/core-js/modules/_is-object.js"),l=t("./node_modules/core-js/modules/_fails.js"),d=t("./node_modules/core-js/modules/_validate-collection.js"),f=i.getWeak,m=Object.isExtensible,p=u.ufstore,_={},h=function(e){return function(){return e(this,arguments.length>0?arguments[0]:void 0)}},j={get:function(e){if(c(e)){var o=f(e);return!0===o?p(d(this,"WeakMap")).get(e):o?o[this._i]:void 0}},set:function(e,o){return u.def(d(this,"WeakMap"),e,o)}},v=e.exports=t("./node_modules/core-js/modules/_collection.js")("WeakMap",h,j,u,!0,!0);l(function(){return 7!=(new v).set((Object.freeze||Object)(_),7).get(_)})&&(a((n=u.getConstructor(h,"WeakMap")).prototype,j),i.NEED=!0,r(["delete","has","get","set"],function(e){var o=v.prototype,t=o[e];s(o,e,function(o,r){if(c(o)&&!m(o)){this._f||(this._f=new n);var s=this._f[e](o,r);return"set"==e?this:s}return t.call(this,o,r)})}))},"./node_modules/core-js/modules/es7.reflect.define-metadata.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_metadata.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=n.key,i=n.set;n.exp({defineMetadata:function(e,o,t,n){i(e,o,r(t),s(n))}})},"./node_modules/core-js/modules/es7.reflect.delete-metadata.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_metadata.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=n.key,i=n.map,a=n.store;n.exp({deleteMetadata:function(e,o){var t=arguments.length<3?void 0:s(arguments[2]),n=i(r(o),t,!1);if(void 0===n||!n.delete(e))return!1;if(n.size)return!0;var u=a.get(o);return u.delete(t),!!u.size||a.delete(o)}})},"./node_modules/core-js/modules/es7.reflect.get-metadata-keys.js":function(e,o,t){var n=t("./node_modules/core-js/modules/es6.set.js"),r=t("./node_modules/core-js/modules/_array-from-iterable.js"),s=t("./node_modules/core-js/modules/_metadata.js"),i=t("./node_modules/core-js/modules/_an-object.js"),a=t("./node_modules/core-js/modules/_object-gpo.js"),u=s.keys,c=s.key,l=function(e,o){var t=u(e,o),s=a(e);if(null===s)return t;var i=l(s,o);return i.length?t.length?r(new n(t.concat(i))):i:t};s.exp({getMetadataKeys:function(e){return l(i(e),arguments.length<2?void 0:c(arguments[1]))}})},"./node_modules/core-js/modules/es7.reflect.get-metadata.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_metadata.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=t("./node_modules/core-js/modules/_object-gpo.js"),i=n.has,a=n.get,u=n.key,c=function(e,o,t){if(i(e,o,t))return a(e,o,t);var n=s(o);return null!==n?c(e,n,t):void 0};n.exp({getMetadata:function(e,o){return c(e,r(o),arguments.length<3?void 0:u(arguments[2]))}})},"./node_modules/core-js/modules/es7.reflect.get-own-metadata-keys.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_metadata.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=n.keys,i=n.key;n.exp({getOwnMetadataKeys:function(e){return s(r(e),arguments.length<2?void 0:i(arguments[1]))}})},"./node_modules/core-js/modules/es7.reflect.get-own-metadata.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_metadata.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=n.get,i=n.key;n.exp({getOwnMetadata:function(e,o){return s(e,r(o),arguments.length<3?void 0:i(arguments[2]))}})},"./node_modules/core-js/modules/es7.reflect.has-metadata.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_metadata.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=t("./node_modules/core-js/modules/_object-gpo.js"),i=n.has,a=n.key,u=function(e,o,t){if(i(e,o,t))return!0;var n=s(o);return null!==n&&u(e,n,t)};n.exp({hasMetadata:function(e,o){return u(e,r(o),arguments.length<3?void 0:a(arguments[2]))}})},"./node_modules/core-js/modules/es7.reflect.has-own-metadata.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_metadata.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=n.has,i=n.key;n.exp({hasOwnMetadata:function(e,o){return s(e,r(o),arguments.length<3?void 0:i(arguments[2]))}})},"./node_modules/core-js/modules/es7.reflect.metadata.js":function(e,o,t){var n=t("./node_modules/core-js/modules/_metadata.js"),r=t("./node_modules/core-js/modules/_an-object.js"),s=t("./node_modules/core-js/modules/_a-function.js"),i=n.key,a=n.set;n.exp({metadata:function(e,o){return function(t,n){a(e,o,(void 0!==n?r:s)(t),i(n))}}})},"./node_modules/zone.js/dist/zone.js":function(e,o,t){!function(){"use strict";!function(e){var o=e.performance;function t(e){o&&o.mark&&o.mark(e)}function n(e,t){o&&o.measure&&o.measure(e,t)}t("Zone");var r=!0===e.__zone_symbol__forceDuplicateZoneCheck;if(e.Zone){if(r||"function"!=typeof e.Zone.__symbol__)throw new Error("Zone already loaded.");return e.Zone}var s,i=function(){function o(e,o){this._parent=e,this._name=o?o.name||"unnamed":"<root>",this._properties=o&&o.properties||{},this._zoneDelegate=new u(this,this._parent&&this._parent._zoneDelegate,o)}return o.assertZonePatched=function(){if(e.Promise!==O.ZoneAwarePromise)throw new Error("Zone.js has detected that ZoneAwarePromise `(window|global).Promise` has been overwritten.\nMost likely cause is that a Promise polyfill has been loaded after Zone.js (Polyfilling Promise api is not necessary when zone.js is loaded. If you must load one, do so before loading zone.js.)")},Object.defineProperty(o,"root",{get:function(){for(var e=o.current;e.parent;)e=e.parent;return e},enumerable:!0,configurable:!0}),Object.defineProperty(o,"current",{get:function(){return P.zone},enumerable:!0,configurable:!0}),Object.defineProperty(o,"currentTask",{get:function(){return D},enumerable:!0,configurable:!0}),o.__load_patch=function(s,i){if(O.hasOwnProperty(s)){if(r)throw Error("Already loaded patch: "+s)}else if(!e["__Zone_disable_"+s]){var a="Zone:"+s;t(a),O[s]=i(e,o,S),n(a,a)}},Object.defineProperty(o.prototype,"parent",{get:function(){return this._parent},enumerable:!0,configurable:!0}),Object.defineProperty(o.prototype,"name",{get:function(){return this._name},enumerable:!0,configurable:!0}),o.prototype.get=function(e){var o=this.getZoneWith(e);if(o)return o._properties[e]},o.prototype.getZoneWith=function(e){for(var o=this;o;){if(o._properties.hasOwnProperty(e))return o;o=o._parent}return null},o.prototype.fork=function(e){if(!e)throw new Error("ZoneSpec required!");return this._zoneDelegate.fork(this,e)},o.prototype.wrap=function(e,o){if("function"!=typeof e)throw new Error("Expecting function got: "+e);var t=this._zoneDelegate.intercept(this,e,o),n=this;return function(){return n.runGuarded(t,this,arguments,o)}},o.prototype.run=function(e,o,t,n){P={parent:P,zone:this};try{return this._zoneDelegate.invoke(this,e,o,t,n)}finally{P=P.parent}},o.prototype.runGuarded=function(e,o,t,n){void 0===o&&(o=null),P={parent:P,zone:this};try{try{return this._zoneDelegate.invoke(this,e,o,t,n)}catch(e){if(this._zoneDelegate.handleError(this,e))throw e}}finally{P=P.parent}},o.prototype.runTask=function(e,o,t){if(e.zone!=this)throw new Error("A task can only be run in the zone of creation! (Creation: "+(e.zone||j).name+"; Execution: "+this.name+")");if(e.state!==v||e.type!==x&&e.type!==E){var n=e.state!=b;n&&e._transitionTo(b,y),e.runCount++;var r=D;D=e,P={parent:P,zone:this};try{e.type==E&&e.data&&!e.data.isPeriodic&&(e.cancelFn=void 0);try{return this._zoneDelegate.invokeTask(this,e,o,t)}catch(e){if(this._zoneDelegate.handleError(this,e))throw e}}finally{e.state!==v&&e.state!==w&&(e.type==x||e.data&&e.data.isPeriodic?n&&e._transitionTo(y,b):(e.runCount=0,this._updateTaskCount(e,-1),n&&e._transitionTo(v,b,v))),P=P.parent,D=r}}},o.prototype.scheduleTask=function(e){if(e.zone&&e.zone!==this)for(var o=this;o;){if(o===e.zone)throw Error("can not reschedule task to "+this.name+" which is descendants of the original zone "+e.zone.name);o=o.parent}e._transitionTo(g,v);var t=[];e._zoneDelegates=t,e._zone=this;try{e=this._zoneDelegate.scheduleTask(this,e)}catch(o){throw e._transitionTo(w,g,v),this._zoneDelegate.handleError(this,o),o}return e._zoneDelegates===t&&this._updateTaskCount(e,1),e.state==g&&e._transitionTo(y,g),e},o.prototype.scheduleMicroTask=function(e,o,t,n){return this.scheduleTask(new c(T,e,o,t,n,void 0))},o.prototype.scheduleMacroTask=function(e,o,t,n,r){return this.scheduleTask(new c(E,e,o,t,n,r))},o.prototype.scheduleEventTask=function(e,o,t,n,r){return this.scheduleTask(new c(x,e,o,t,n,r))},o.prototype.cancelTask=function(e){if(e.zone!=this)throw new Error("A task can only be cancelled in the zone of creation! (Creation: "+(e.zone||j).name+"; Execution: "+this.name+")");e._transitionTo(k,y,b);try{this._zoneDelegate.cancelTask(this,e)}catch(o){throw e._transitionTo(w,k),this._zoneDelegate.handleError(this,o),o}return this._updateTaskCount(e,-1),e._transitionTo(v,k),e.runCount=0,e},o.prototype._updateTaskCount=function(e,o){var t=e._zoneDelegates;-1==o&&(e._zoneDelegates=null);for(var n=0;n<t.length;n++)t[n]._updateTaskCount(e.type,o)},o.__symbol__=C,o}(),a={name:"",onHasTask:function(e,o,t,n){return e.hasTask(t,n)},onScheduleTask:function(e,o,t,n){return e.scheduleTask(t,n)},onInvokeTask:function(e,o,t,n,r,s){return e.invokeTask(t,n,r,s)},onCancelTask:function(e,o,t,n){return e.cancelTask(t,n)}},u=function(){function e(e,o,t){this._taskCounts={microTask:0,macroTask:0,eventTask:0},this.zone=e,this._parentDelegate=o,this._forkZS=t&&(t&&t.onFork?t:o._forkZS),this._forkDlgt=t&&(t.onFork?o:o._forkDlgt),this._forkCurrZone=t&&(t.onFork?this.zone:o.zone),this._interceptZS=t&&(t.onIntercept?t:o._interceptZS),this._interceptDlgt=t&&(t.onIntercept?o:o._interceptDlgt),this._interceptCurrZone=t&&(t.onIntercept?this.zone:o.zone),this._invokeZS=t&&(t.onInvoke?t:o._invokeZS),this._invokeDlgt=t&&(t.onInvoke?o:o._invokeDlgt),this._invokeCurrZone=t&&(t.onInvoke?this.zone:o.zone),this._handleErrorZS=t&&(t.onHandleError?t:o._handleErrorZS),this._handleErrorDlgt=t&&(t.onHandleError?o:o._handleErrorDlgt),this._handleErrorCurrZone=t&&(t.onHandleError?this.zone:o.zone),this._scheduleTaskZS=t&&(t.onScheduleTask?t:o._scheduleTaskZS),this._scheduleTaskDlgt=t&&(t.onScheduleTask?o:o._scheduleTaskDlgt),this._scheduleTaskCurrZone=t&&(t.onScheduleTask?this.zone:o.zone),this._invokeTaskZS=t&&(t.onInvokeTask?t:o._invokeTaskZS),this._invokeTaskDlgt=t&&(t.onInvokeTask?o:o._invokeTaskDlgt),this._invokeTaskCurrZone=t&&(t.onInvokeTask?this.zone:o.zone),this._cancelTaskZS=t&&(t.onCancelTask?t:o._cancelTaskZS),this._cancelTaskDlgt=t&&(t.onCancelTask?o:o._cancelTaskDlgt),this._cancelTaskCurrZone=t&&(t.onCancelTask?this.zone:o.zone),this._hasTaskZS=null,this._hasTaskDlgt=null,this._hasTaskDlgtOwner=null,this._hasTaskCurrZone=null;var n=t&&t.onHasTask,r=o&&o._hasTaskZS;(n||r)&&(this._hasTaskZS=n?t:a,this._hasTaskDlgt=o,this._hasTaskDlgtOwner=this,this._hasTaskCurrZone=e,t.onScheduleTask||(this._scheduleTaskZS=a,this._scheduleTaskDlgt=o,this._scheduleTaskCurrZone=this.zone),t.onInvokeTask||(this._invokeTaskZS=a,this._invokeTaskDlgt=o,this._invokeTaskCurrZone=this.zone),t.onCancelTask||(this._cancelTaskZS=a,this._cancelTaskDlgt=o,this._cancelTaskCurrZone=this.zone))}return e.prototype.fork=function(e,o){return this._forkZS?this._forkZS.onFork(this._forkDlgt,this.zone,e,o):new i(e,o)},e.prototype.intercept=function(e,o,t){return this._interceptZS?this._interceptZS.onIntercept(this._interceptDlgt,this._interceptCurrZone,e,o,t):o},e.prototype.invoke=function(e,o,t,n,r){return this._invokeZS?this._invokeZS.onInvoke(this._invokeDlgt,this._invokeCurrZone,e,o,t,n,r):o.apply(t,n)},e.prototype.handleError=function(e,o){return!this._handleErrorZS||this._handleErrorZS.onHandleError(this._handleErrorDlgt,this._handleErrorCurrZone,e,o)},e.prototype.scheduleTask=function(e,o){var t=o;if(this._scheduleTaskZS)this._hasTaskZS&&t._zoneDelegates.push(this._hasTaskDlgtOwner),(t=this._scheduleTaskZS.onScheduleTask(this._scheduleTaskDlgt,this._scheduleTaskCurrZone,e,o))||(t=o);else if(o.scheduleFn)o.scheduleFn(o);else{if(o.type!=T)throw new Error("Task is missing scheduleFn.");_(o)}return t},e.prototype.invokeTask=function(e,o,t,n){return this._invokeTaskZS?this._invokeTaskZS.onInvokeTask(this._invokeTaskDlgt,this._invokeTaskCurrZone,e,o,t,n):o.callback.apply(t,n)},e.prototype.cancelTask=function(e,o){var t;if(this._cancelTaskZS)t=this._cancelTaskZS.onCancelTask(this._cancelTaskDlgt,this._cancelTaskCurrZone,e,o);else{if(!o.cancelFn)throw Error("Task is not cancelable");t=o.cancelFn(o)}return t},e.prototype.hasTask=function(e,o){try{this._hasTaskZS&&this._hasTaskZS.onHasTask(this._hasTaskDlgt,this._hasTaskCurrZone,e,o)}catch(o){this.handleError(e,o)}},e.prototype._updateTaskCount=function(e,o){var t=this._taskCounts,n=t[e],r=t[e]=n+o;if(r<0)throw new Error("More tasks executed then were scheduled.");if(0==n||0==r){var s={microTask:t.microTask>0,macroTask:t.macroTask>0,eventTask:t.eventTask>0,change:e};this.hasTask(this.zone,s)}},e}(),c=function(){function o(t,n,r,s,i,a){this._zone=null,this.runCount=0,this._zoneDelegates=null,this._state="notScheduled",this.type=t,this.source=n,this.data=s,this.scheduleFn=i,this.cancelFn=a,this.callback=r;var u=this;t===x&&s&&s.useG?this.invoke=o.invokeTask:this.invoke=function(){return o.invokeTask.call(e,u,this,arguments)}}return o.invokeTask=function(e,o,t){e||(e=this),z++;try{return e.runCount++,e.zone.runTask(e,o,t)}finally{1==z&&h(),z--}},Object.defineProperty(o.prototype,"zone",{get:function(){return this._zone},enumerable:!0,configurable:!0}),Object.defineProperty(o.prototype,"state",{get:function(){return this._state},enumerable:!0,configurable:!0}),o.prototype.cancelScheduleRequest=function(){this._transitionTo(v,g)},o.prototype._transitionTo=function(e,o,t){if(this._state!==o&&this._state!==t)throw new Error(this.type+" '"+this.source+"': can not transition to '"+e+"', expecting state '"+o+"'"+(t?" or '"+t+"'":"")+", was '"+this._state+"'.");this._state=e,e==v&&(this._zoneDelegates=null)},o.prototype.toString=function(){return this.data&&void 0!==this.data.handleId?this.data.handleId.toString():Object.prototype.toString.call(this)},o.prototype.toJSON=function(){return{type:this.type,state:this.state,source:this.source,zone:this.zone.name,runCount:this.runCount}},o}(),l=C("setTimeout"),d=C("Promise"),f=C("then"),m=[],p=!1;function _(o){if(0===z&&0===m.length)if(s||e[d]&&(s=e[d].resolve(0)),s){var t=s[f];t||(t=s.then),t.call(s,h)}else e[l](h,0);o&&m.push(o)}function h(){if(!p){for(p=!0;m.length;){var e=m;m=[];for(var o=0;o<e.length;o++){var t=e[o];try{t.zone.runTask(t,null,null)}catch(e){S.onUnhandledError(e)}}}S.microtaskDrainDone(),p=!1}}var j={name:"NO ZONE"},v="notScheduled",g="scheduling",y="scheduled",b="running",k="canceling",w="unknown",T="microTask",E="macroTask",x="eventTask",O={},S={symbol:C,currentZoneFrame:function(){return P},onUnhandledError:Z,microtaskDrainDone:Z,scheduleMicroTask:_,showUncaughtError:function(){return!i[C("ignoreConsoleErrorUncaughtError")]},patchEventTarget:function(){return[]},patchOnProperties:Z,patchMethod:function(){return Z},bindArguments:function(){return[]},patchThen:function(){return Z},setNativePromise:function(e){e&&"function"==typeof e.resolve&&(s=e.resolve(0))}},P={parent:null,zone:new i(null,null)},D=null,z=0;function Z(){}function C(e){return"__zone_symbol__"+e}n("Zone","Zone"),e.Zone=i}("undefined"!=typeof window&&window||"undefined"!=typeof self&&self||global);var e=function(e){var o="function"==typeof Symbol&&e[Symbol.iterator],t=0;return o?o.call(e):{next:function(){return e&&t>=e.length&&(e=void 0),{value:e&&e[t++],done:!e}}}};Zone.__load_patch("ZoneAwarePromise",function(o,t,n){var r=Object.getOwnPropertyDescriptor,s=Object.defineProperty,i=n.symbol,a=[],u=i("Promise"),c=i("then"),l="__creationTrace__";n.onUnhandledError=function(e){if(n.showUncaughtError()){var o=e&&e.rejection;o?console.error("Unhandled Promise rejection:",o instanceof Error?o.message:o,"; Zone:",e.zone.name,"; Task:",e.task&&e.task.source,"; Value:",o,o instanceof Error?o.stack:void 0):console.error(e)}},n.microtaskDrainDone=function(){for(;a.length;)for(var e=function(){var e=a.shift();try{e.zone.runGuarded(function(){throw e})}catch(e){f(e)}};a.length;)e()};var d=i("unhandledPromiseRejectionHandler");function f(e){n.onUnhandledError(e);try{var o=t[d];o&&"function"==typeof o&&o.call(this,e)}catch(e){}}function m(e){return e&&e.then}function p(e){return e}function _(e){return M.reject(e)}var h=i("state"),j=i("value"),v=i("finally"),g=i("parentPromiseValue"),y=i("parentPromiseState"),b="Promise.then",k=null,w=!0,T=!1,E=0;function x(e,o){return function(t){try{D(e,o,t)}catch(o){D(e,!1,o)}}}var O=function(){var e=!1;return function(o){return function(){e||(e=!0,o.apply(null,arguments))}}},S="Promise resolved with itself",P=i("currentTaskTrace");function D(e,o,r){var i=O();if(e===r)throw new TypeError(S);if(e[h]===k){var u=null;try{"object"!=typeof r&&"function"!=typeof r||(u=r&&r.then)}catch(o){return i(function(){D(e,!1,o)})(),e}if(o!==T&&r instanceof M&&r.hasOwnProperty(h)&&r.hasOwnProperty(j)&&r[h]!==k)Z(r),D(e,r[h],r[j]);else if(o!==T&&"function"==typeof u)try{u.call(r,i(x(e,o)),i(x(e,!1)))}catch(o){i(function(){D(e,!1,o)})()}else{e[h]=o;var c=e[j];if(e[j]=r,e[v]===v&&o===w&&(e[h]=e[y],e[j]=e[g]),o===T&&r instanceof Error){var d=t.currentTask&&t.currentTask.data&&t.currentTask.data[l];d&&s(r,P,{configurable:!0,enumerable:!1,writable:!0,value:d})}for(var f=0;f<c.length;)C(e,c[f++],c[f++],c[f++],c[f++]);if(0==c.length&&o==T){e[h]=E;try{throw new Error("Uncaught (in promise): "+function(e){if(e&&e.toString===Object.prototype.toString){var o=e.constructor&&e.constructor.name;return(o||"")+": "+JSON.stringify(e)}return e?e.toString():Object.prototype.toString.call(e)}(r)+(r&&r.stack?"\n"+r.stack:""))}catch(o){var m=o;m.rejection=r,m.promise=e,m.zone=t.current,m.task=t.currentTask,a.push(m),n.scheduleMicroTask()}}}}return e}var z=i("rejectionHandledHandler");function Z(e){if(e[h]===E){try{var o=t[z];o&&"function"==typeof o&&o.call(this,{rejection:e[j],promise:e})}catch(e){}e[h]=T;for(var n=0;n<a.length;n++)e===a[n].promise&&a.splice(n,1)}}function C(e,o,t,n,r){Z(e);var s=e[h],i=s?"function"==typeof n?n:p:"function"==typeof r?r:_;o.scheduleMicroTask(b,function(){try{var n=e[j],r=t&&v===t[v];r&&(t[g]=n,t[y]=s);var a=o.run(i,void 0,r&&i!==_&&i!==p?[]:[n]);D(t,!0,a)}catch(e){D(t,!1,e)}},t)}var M=function(){function o(e){if(!(this instanceof o))throw new Error("Must be an instanceof Promise.");this[h]=k,this[j]=[];try{e&&e(x(this,w),x(this,T))}catch(e){D(this,!1,e)}}return o.toString=function(){return"function ZoneAwarePromise() { [native code] }"},o.resolve=function(e){return D(new this(null),w,e)},o.reject=function(e){return D(new this(null),T,e)},o.race=function(o){var t,n,r,s,i=new this(function(e,o){r=e,s=o});function a(e){i&&(i=r(e))}function u(e){i&&(i=s(e))}try{for(var c=e(o),l=c.next();!l.done;l=c.next()){var d=l.value;m(d)||(d=this.resolve(d)),d.then(a,u)}}catch(e){t={error:e}}finally{try{l&&!l.done&&(n=c.return)&&n.call(c)}finally{if(t)throw t.error}}return i},o.all=function(o){var t,n,r,s,i=new this(function(e,o){r=e,s=o}),a=2,u=0,c=[],l=function(e){m(e)||(e=d.resolve(e));var o=u;e.then(function(e){c[o]=e,0==--a&&r(c)},s),a++,u++},d=this;try{for(var f=e(o),p=f.next();!p.done;p=f.next()){var _=p.value;l(_)}}catch(e){t={error:e}}finally{try{p&&!p.done&&(n=f.return)&&n.call(f)}finally{if(t)throw t.error}}return 0==(a-=2)&&r(c),i},o.prototype.then=function(e,o){var n=new this.constructor(null),r=t.current;return this[h]==k?this[j].push(r,n,e,o):C(this,r,n,e,o),n},o.prototype.catch=function(e){return this.then(null,e)},o.prototype.finally=function(e){var o=new this.constructor(null);o[v]=v;var n=t.current;return this[h]==k?this[j].push(n,o,e,e):C(this,n,o,e,e),o},o}();M.resolve=M.resolve,M.reject=M.reject,M.race=M.race,M.all=M.all;var I=o[u]=o.Promise,L=t.__symbol__("ZoneAwarePromise"),R=r(o,"Promise");R&&!R.configurable||(R&&delete R.writable,R&&delete R.value,R||(R={configurable:!0,enumerable:!0}),R.get=function(){return o[L]?o[L]:o[u]},R.set=function(e){e===M?o[L]=e:(o[u]=e,e.prototype[c]||A(e),n.setNativePromise(e))},s(o,"Promise",R)),o.Promise=M;var F=i("thenPatched");function A(e){var o=e.prototype,t=r(o,"then");if(!t||!1!==t.writable&&t.configurable){var n=o.then;o[c]=n,e.prototype.then=function(e,o){var t=this,r=new M(function(e,o){n.call(t,e,o)});return r.then(e,o)},e[F]=!0}}return n.patchThen=A,I&&A(I),Promise[t.__symbol__("uncaughtPromiseErrors")]=a,M}),Zone.__load_patch("fetch",function(e,o,t){var n=e.fetch,r=e.Promise,s=t.symbol("thenPatched"),i=t.symbol("fetchTaskScheduling"),a=t.symbol("fetchTaskAborting");if("function"==typeof n){var u=e.AbortController,c="function"==typeof u,l=null;c&&(e.AbortController=function(){var e=new u,o=e.signal;return o.abortController=e,e},l=t.patchMethod(u.prototype,"abort",function(e){return function(o,t){return o.task?o.task.zone.cancelTask(o.task):e.apply(o,t)}}));var d=function(){};e.fetch=function(){var e=this,u=Array.prototype.slice.call(arguments),f=u.length>1?u[1]:null,m=f&&f.signal;return new Promise(function(f,p){var _=o.current.scheduleMacroTask("fetch",d,u,function(){var a,c=o.current;try{c[i]=!0,a=n.apply(e,u)}catch(e){return void p(e)}finally{c[i]=!1}if(!(a instanceof r)){var l=a.constructor;l[s]||t.patchThen(l)}a.then(function(e){"notScheduled"!==_.state&&_.invoke(),f(e)},function(e){"notScheduled"!==_.state&&_.invoke(),p(e)})},function(){if(c)if(m&&m.abortController&&!m.aborted&&"function"==typeof m.abortController.abort&&l)try{o.current[a]=!0,l.call(m.abortController)}finally{o.current[a]=!1}else p("cancel fetch need a AbortController.signal");else p("No AbortController supported, can not cancel fetch")});m&&m.abortController&&(m.abortController.task=_)})}}});var o=Object.getOwnPropertyDescriptor,t=Object.defineProperty,n=Object.getPrototypeOf,r=Object.create,s=Array.prototype.slice,i="addEventListener",a="removeEventListener",u=Zone.__symbol__(i),c=Zone.__symbol__(a),l="true",d="false",f="__zone_symbol__";function m(e,o){return Zone.current.wrap(e,o)}function p(e,o,t,n,r){return Zone.current.scheduleMacroTask(e,o,t,n,r)}var _=Zone.__symbol__,h="undefined"!=typeof window,j=h?window:void 0,v=h&&j||"object"==typeof self&&self||global,g="removeAttribute",y=[null];function b(e,o){for(var t=e.length-1;t>=0;t--)"function"==typeof e[t]&&(e[t]=m(e[t],o+"_"+t));return e}function k(e){return!e||!1!==e.writable&&!("function"==typeof e.get&&void 0===e.set)}var w="undefined"!=typeof WorkerGlobalScope&&self instanceof WorkerGlobalScope,T=!("nw"in v)&&void 0!==v.process&&"[object process]"==={}.toString.call(v.process),E=!T&&!w&&!(!h||!j.HTMLElement),x=void 0!==v.process&&"[object process]"==={}.toString.call(v.process)&&!w&&!(!h||!j.HTMLElement),O={},S=function(e){if(e=e||v.event){var o=O[e.type];o||(o=O[e.type]=_("ON_PROPERTY"+e.type));var t,n=this||e.target||v,r=n[o];if(E&&n===j&&"error"===e.type){var s=e;!0===(t=r&&r.call(this,s.message,s.filename,s.lineno,s.colno,s.error))&&e.preventDefault()}else void 0==(t=r&&r.apply(this,arguments))||t||e.preventDefault();return t}};function P(e,n,r){var s=o(e,n);if(!s&&r){var i=o(r,n);i&&(s={enumerable:!0,configurable:!0})}if(s&&s.configurable){var a=_("on"+n+"patched");if(!e.hasOwnProperty(a)||!e[a]){delete s.writable,delete s.value;var u=s.get,c=s.set,l=n.substr(2),d=O[l];d||(d=O[l]=_("ON_PROPERTY"+l)),s.set=function(o){var t=this;if(t||e!==v||(t=v),t){var n=t[d];n&&t.removeEventListener(l,S),c&&c.apply(t,y),"function"==typeof o?(t[d]=o,t.addEventListener(l,S,!1)):t[d]=null}},s.get=function(){var o=this;if(o||e!==v||(o=v),!o)return null;var t=o[d];if(t)return t;if(u){var r=u&&u.call(this);if(r)return s.set.call(this,r),"function"==typeof o[g]&&o.removeAttribute(n),r}return null},t(e,n,s),e[a]=!0}}}function D(e,o,t){if(o)for(var n=0;n<o.length;n++)P(e,"on"+o[n],t);else{var r=[];for(var s in e)"on"==s.substr(0,2)&&r.push(s);for(var i=0;i<r.length;i++)P(e,r[i],t)}}var z=_("originalInstance");function Z(e){var o=v[e];if(o){v[_(e)]=o,v[e]=function(){var t=b(arguments,e);switch(t.length){case 0:this[z]=new o;break;case 1:this[z]=new o(t[0]);break;case 2:this[z]=new o(t[0],t[1]);break;case 3:this[z]=new o(t[0],t[1],t[2]);break;case 4:this[z]=new o(t[0],t[1],t[2],t[3]);break;default:throw new Error("Arg list too long.")}},I(v[e],o);var n,r=new o(function(){});for(n in r)"XMLHttpRequest"===e&&"responseBlob"===n||function(o){"function"==typeof r[o]?v[e].prototype[o]=function(){return this[z][o].apply(this[z],arguments)}:t(v[e].prototype,o,{set:function(t){"function"==typeof t?(this[z][o]=m(t,e+"."+o),I(this[z][o],t)):this[z][o]=t},get:function(){return this[z][o]}})}(n);for(n in o)"prototype"!==n&&o.hasOwnProperty(n)&&(v[e][n]=o[n])}}var C=!1;function M(e,t,r){for(var s=e;s&&!s.hasOwnProperty(t);)s=n(s);!s&&e[t]&&(s=e);var i=_(t),a=null;if(s&&!(a=s[i])){a=s[i]=s[t];var u=s&&o(s,t);if(k(u)){var c=r(a,i,t);s[t]=function(){return c(this,arguments)},I(s[t],a),C&&function(e,o){"function"==typeof Object.getOwnPropertySymbols&&Object.getOwnPropertySymbols(e).forEach(function(t){var n=Object.getOwnPropertyDescriptor(e,t);Object.defineProperty(o,t,{get:function(){return e[t]},set:function(o){(!n||n.writable&&"function"==typeof n.set)&&(e[t]=o)},enumerable:!n||n.enumerable,configurable:!n||n.configurable})})}(a,s[t])}}return a}function I(e,o){e[_("OriginalDelegate")]=o}var L=!1,R=!1;Zone.__load_patch("toString",function(e){var o=Function.prototype.toString,t=_("OriginalDelegate"),n=_("Promise"),r=_("Error"),s=function(){if("function"==typeof this){var s=this[t];if(s)return"function"==typeof s?o.apply(this[t],arguments):Object.prototype.toString.call(s);if(this===Promise){var i=e[n];if(i)return o.apply(i,arguments)}if(this===Error){var a=e[r];if(a)return o.apply(a,arguments)}}return o.apply(this,arguments)};s[t]=o,Function.prototype.toString=s;var i=Object.prototype.toString;Object.prototype.toString=function(){return this instanceof Promise?"[object Promise]":i.apply(this,arguments)}});var F=!1;if("undefined"!=typeof window)try{var A=Object.defineProperty({},"passive",{get:function(){F=!0}});window.addEventListener("test",A,A),window.removeEventListener("test",A,A)}catch(e){F=!1}var H={useG:!0},N={},B={},q=/^__zone_symbol__(\w+)(true|false)$/,W="__zone_symbol__propagationStopped";function U(e,o,t){var r=t&&t.add||i,s=t&&t.rm||a,u=t&&t.listeners||"eventListeners",c=t&&t.rmAll||"removeAllListeners",m=_(r),p="."+r+":",h="prependListener",j="."+h+":",v=function(e,o,t){if(!e.isRemoved){var n=e.callback;"object"==typeof n&&n.handleEvent&&(e.callback=function(e){return n.handleEvent(e)},e.originalDelegate=n),e.invoke(e,o,[t]);var r=e.options;if(r&&"object"==typeof r&&r.once){var i=e.originalDelegate?e.originalDelegate:e.callback;o[s].call(o,t.type,i,r)}}},g=function(o){if(o=o||e.event){var t=this||o.target||e,n=t[N[o.type][d]];if(n)if(1===n.length)v(n[0],t,o);else for(var r=n.slice(),s=0;s<r.length&&(!o||!0!==o[W]);s++)v(r[s],t,o)}},y=function(o){if(o=o||e.event){var t=this||o.target||e,n=t[N[o.type][l]];if(n)if(1===n.length)v(n[0],t,o);else for(var r=n.slice(),s=0;s<r.length&&(!o||!0!==o[W]);s++)v(r[s],t,o)}};function b(o,t){if(!o)return!1;var i=!0;t&&void 0!==t.useG&&(i=t.useG);var a=t&&t.vh,v=!0;t&&void 0!==t.chkDup&&(v=t.chkDup);var b=!1;t&&void 0!==t.rt&&(b=t.rt);for(var k=o;k&&!k.hasOwnProperty(r);)k=n(k);if(!k&&o[r]&&(k=o),!k)return!1;if(k[m])return!1;var w,E=t&&t.eventNameToString,x={},O=k[m]=k[r],S=k[_(s)]=k[s],P=k[_(u)]=k[u],D=k[_(c)]=k[c];function z(e){F||"boolean"==typeof x.options||void 0===x.options||null===x.options||(e.options=!!x.options.capture,x.options=e.options)}t&&t.prepend&&(w=k[_(t.prepend)]=k[t.prepend]);var Z=i?function(e){if(!x.isExisting)return z(e),O.call(x.target,x.eventName,x.capture?y:g,x.options)}:function(e){return z(e),O.call(x.target,x.eventName,e.invoke,x.options)},C=i?function(e){if(!e.isRemoved){var o=N[e.eventName],t=void 0;o&&(t=o[e.capture?l:d]);var n=t&&e.target[t];if(n)for(var r=0;r<n.length;r++){var s=n[r];if(s===e){n.splice(r,1),e.isRemoved=!0,0===n.length&&(e.allRemoved=!0,e.target[t]=null);break}}}if(e.allRemoved)return S.call(e.target,e.eventName,e.capture?y:g,e.options)}:function(e){return S.call(e.target,e.eventName,e.invoke,e.options)},M=t&&t.diff?t.diff:function(e,o){var t=typeof o;return"function"===t&&e.callback===o||"object"===t&&e.originalDelegate===o},L=Zone[Zone.__symbol__("BLACK_LISTED_EVENTS")],R=function(o,t,n,r,s,u){return void 0===s&&(s=!1),void 0===u&&(u=!1),function(){var c=this||e,m=arguments[0],p=arguments[1];if(!p)return o.apply(this,arguments);if(T&&"uncaughtException"===m)return o.apply(this,arguments);var _=!1;if("function"!=typeof p){if(!p.handleEvent)return o.apply(this,arguments);_=!0}if(!a||a(o,p,c,arguments)){var h,j=arguments[2];if(L)for(var g=0;g<L.length;g++)if(m===L[g])return o.apply(this,arguments);var y=!1;void 0===j?h=!1:!0===j?h=!0:!1===j?h=!1:(h=!!j&&!!j.capture,y=!!j&&!!j.once);var b,k=Zone.current,w=N[m];if(w)b=w[h?l:d];else{var O=(E?E(m):m)+d,S=(E?E(m):m)+l,P=f+O,D=f+S;N[m]={},N[m][d]=P,N[m][l]=D,b=h?D:P}var z,Z=c[b],C=!1;if(Z){if(C=!0,v)for(var g=0;g<Z.length;g++)if(M(Z[g],p))return}else Z=c[b]=[];var I=c.constructor.name,R=B[I];R&&(z=R[m]),z||(z=I+t+(E?E(m):m)),x.options=j,y&&(x.options.once=!1),x.target=c,x.capture=h,x.eventName=m,x.isExisting=C;var A=i?H:void 0;A&&(A.taskData=x);var q=k.scheduleEventTask(z,p,A,n,r);return x.target=null,A&&(A.taskData=null),y&&(j.once=!0),(F||"boolean"!=typeof q.options)&&(q.options=j),q.target=c,q.capture=h,q.eventName=m,_&&(q.originalDelegate=p),u?Z.unshift(q):Z.push(q),s?c:void 0}}};return k[r]=R(O,p,Z,C,b),w&&(k[h]=R(w,j,function(e){return w.call(x.target,x.eventName,e.invoke,x.options)},C,b,!0)),k[s]=function(){var o,t=this||e,n=arguments[0],r=arguments[2];o=void 0!==r&&(!0===r||!1!==r&&!!r&&!!r.capture);var s=arguments[1];if(!s)return S.apply(this,arguments);if(!a||a(S,s,t,arguments)){var i,u=N[n];u&&(i=u[o?l:d]);var c=i&&t[i];if(c)for(var f=0;f<c.length;f++){var m=c[f];if(M(m,s))return c.splice(f,1),m.isRemoved=!0,0===c.length&&(m.allRemoved=!0,t[i]=null),m.zone.cancelTask(m),b?t:void 0}return S.apply(this,arguments)}},k[u]=function(){for(var o=this||e,t=arguments[0],n=[],r=G(o,E?E(t):t),s=0;s<r.length;s++){var i=r[s],a=i.originalDelegate?i.originalDelegate:i.callback;n.push(a)}return n},k[c]=function(){var o=this||e,t=arguments[0];if(t){var n=N[t];if(n){var r=n[d],i=n[l],a=o[r],u=o[i];if(a)for(var f=a.slice(),m=0;m<f.length;m++){var p=f[m],_=p.originalDelegate?p.originalDelegate:p.callback;this[s].call(this,t,_,p.options)}if(u)for(var f=u.slice(),m=0;m<f.length;m++){var p=f[m],_=p.originalDelegate?p.originalDelegate:p.callback;this[s].call(this,t,_,p.options)}}}else{for(var h=Object.keys(o),m=0;m<h.length;m++){var j=h[m],v=q.exec(j),g=v&&v[1];g&&"removeListener"!==g&&this[c].call(this,g)}this[c].call(this,"removeListener")}if(b)return this},I(k[r],O),I(k[s],S),D&&I(k[c],D),P&&I(k[u],P),!0}for(var k=[],w=0;w<o.length;w++)k[w]=b(o[w],t);return k}function G(e,o){var t=[];for(var n in e){var r=q.exec(n),s=r&&r[1];if(s&&(!o||s===o)){var i=e[n];if(i)for(var a=0;a<i.length;a++)t.push(i[a])}}return t}var X=_("zoneTask");function K(e,o,t,n){var r=null,s=null;t+=n;var i={};function a(o){var t=o.data;return t.args[0]=function(){try{o.invoke.apply(this,arguments)}finally{o.data&&o.data.isPeriodic||("number"==typeof t.handleId?delete i[t.handleId]:t.handleId&&(t.handleId[X]=null))}},t.handleId=r.apply(e,t.args),o}function u(e){return s(e.data.handleId)}r=M(e,o+=n,function(t){return function(r,s){if("function"==typeof s[0]){var c={isPeriodic:"Interval"===n,delay:"Timeout"===n||"Interval"===n?s[1]||0:void 0,args:s},l=p(o,s[0],c,a,u);if(!l)return l;var d=l.data.handleId;return"number"==typeof d?i[d]=l:d&&(d[X]=l),d&&d.ref&&d.unref&&"function"==typeof d.ref&&"function"==typeof d.unref&&(l.ref=d.ref.bind(d),l.unref=d.unref.bind(d)),"number"==typeof d||d?d:l}return t.apply(e,s)}}),s=M(e,t,function(o){return function(t,n){var r,s=n[0];"number"==typeof s?r=i[s]:(r=s&&s[X])||(r=s),r&&"string"==typeof r.type?"notScheduled"!==r.state&&(r.cancelFn&&r.data.isPeriodic||0===r.runCount)&&("number"==typeof s?delete i[s]:s&&(s[X]=null),r.zone.cancelTask(r)):o.apply(e,n)}})}var V=Object[_("defineProperty")]=Object.defineProperty,J=Object[_("getOwnPropertyDescriptor")]=Object.getOwnPropertyDescriptor,Y=Object.create,Q=_("unconfigurables");function $(e,o){return e&&e[Q]&&e[Q][o]}function ee(e,o,t){return Object.isFrozen(t)||(t.configurable=!0),t.configurable||(e[Q]||Object.isFrozen(e)||V(e,Q,{writable:!0,value:{}}),e[Q]&&(e[Q][o]=!0)),t}function oe(e,o,t,n){try{return V(e,o,t)}catch(s){if(!t.configurable)throw s;void 0===n?delete t.configurable:t.configurable=n;try{return V(e,o,t)}catch(n){var r=null;try{r=JSON.stringify(t)}catch(e){r=t.toString()}console.log("Attempting to configure '"+o+"' with descriptor '"+r+"' on object '"+e+"' and got error, giving up: "+n)}}}var te=["absolutedeviceorientation","afterinput","afterprint","appinstalled","beforeinstallprompt","beforeprint","beforeunload","devicelight","devicemotion","deviceorientation","deviceorientationabsolute","deviceproximity","hashchange","languagechange","message","mozbeforepaint","offline","online","paint","pageshow","pagehide","popstate","rejectionhandled","storage","unhandledrejection","unload","userproximity","vrdisplyconnected","vrdisplaydisconnected","vrdisplaypresentchange"],ne=["encrypted","waitingforkey","msneedkey","mozinterruptbegin","mozinterruptend"],re=["load"],se=["blur","error","focus","load","resize","scroll","messageerror"],ie=["bounce","finish","start"],ae=["loadstart","progress","abort","error","load","progress","timeout","loadend","readystatechange"],ue=["upgradeneeded","complete","abort","success","error","blocked","versionchange","close"],ce=["close","error","open","message"],le=["error","message"],de=["abort","animationcancel","animationend","animationiteration","auxclick","beforeinput","blur","cancel","canplay","canplaythrough","change","compositionstart","compositionupdate","compositionend","cuechange","click","close","contextmenu","curechange","dblclick","drag","dragend","dragenter","dragexit","dragleave","dragover","drop","durationchange","emptied","ended","error","focus","focusin","focusout","gotpointercapture","input","invalid","keydown","keypress","keyup","load","loadstart","loadeddata","loadedmetadata","lostpointercapture","mousedown","mouseenter","mouseleave","mousemove","mouseout","mouseover","mouseup","mousewheel","orientationchange","pause","play","playing","pointercancel","pointerdown","pointerenter","pointerleave","pointerlockchange","mozpointerlockchange","webkitpointerlockerchange","pointerlockerror","mozpointerlockerror","webkitpointerlockerror","pointermove","pointout","pointerover","pointerup","progress","ratechange","reset","resize","scroll","seeked","seeking","select","selectionchange","selectstart","show","sort","stalled","submit","suspend","timeupdate","volumechange","touchcancel","touchmove","touchstart","touchend","transitioncancel","transitionend","waiting","wheel"].concat(["webglcontextrestored","webglcontextlost","webglcontextcreationerror"],["autocomplete","autocompleteerror"],["toggle"],["afterscriptexecute","beforescriptexecute","DOMContentLoaded","freeze","fullscreenchange","mozfullscreenchange","webkitfullscreenchange","msfullscreenchange","fullscreenerror","mozfullscreenerror","webkitfullscreenerror","msfullscreenerror","readystatechange","visibilitychange","resume"],te,["beforecopy","beforecut","beforepaste","copy","cut","paste","dragstart","loadend","animationstart","search","transitionrun","transitionstart","webkitanimationend","webkitanimationiteration","webkitanimationstart","webkittransitionend"],["activate","afterupdate","ariarequest","beforeactivate","beforedeactivate","beforeeditfocus","beforeupdate","cellchange","controlselect","dataavailable","datasetchanged","datasetcomplete","errorupdate","filterchange","layoutcomplete","losecapture","move","moveend","movestart","propertychange","resizeend","resizestart","rowenter","rowexit","rowsdelete","rowsinserted","command","compassneedscalibration","deactivate","help","mscontentzoom","msmanipulationstatechanged","msgesturechange","msgesturedoubletap","msgestureend","msgesturehold","msgesturestart","msgesturetap","msgotpointercapture","msinertiastart","mslostpointercapture","mspointercancel","mspointerdown","mspointerenter","mspointerhover","mspointerleave","mspointermove","mspointerout","mspointerover","mspointerup","pointerout","mssitemodejumplistitemremoved","msthumbnailclick","stop","storagecommit"]);function fe(e,o,t,n){if(e){var r=function(e,o,t){if(!t||0===t.length)return o;var n=t.filter(function(o){return o.target===e});if(!n||0===n.length)return o;var r=n[0].ignoreProperties;return o.filter(function(e){return-1===r.indexOf(e)})}(e,o,t);D(e,r,n)}}function me(e,u){if(!T||x){var c="undefined"!=typeof WebSocket;if(function(){if((E||x)&&!o(HTMLElement.prototype,"onclick")&&"undefined"!=typeof Element){var e=o(Element.prototype,"onclick");if(e&&!e.configurable)return!1}var n="onreadystatechange",r=XMLHttpRequest.prototype,s=o(r,n);if(s){t(r,n,{enumerable:!0,configurable:!0,get:function(){return!0}});var i=new XMLHttpRequest,a=!!i.onreadystatechange;return t(r,n,s||{}),a}var u=_("fake");t(r,n,{enumerable:!0,configurable:!0,get:function(){return this[u]},set:function(e){this[u]=e}});var i=new XMLHttpRequest,c=function(){};i.onreadystatechange=c;var a=i[u]===c;return i.onreadystatechange=null,a}()){var l=u.__Zone_ignore_on_properties;if(E){var d=window,f=function(){try{var e=j.navigator.userAgent;if(-1!==e.indexOf("MSIE ")||-1!==e.indexOf("Trident/"))return!0}catch(e){}return!1}?[{target:d,ignoreProperties:["error"]}]:[];fe(d,de.concat(["messageerror"]),l?l.concat(f):l,n(d)),fe(Document.prototype,de,l),void 0!==d.SVGElement&&fe(d.SVGElement.prototype,de,l),fe(Element.prototype,de,l),fe(HTMLElement.prototype,de,l),fe(HTMLMediaElement.prototype,ne,l),fe(HTMLFrameSetElement.prototype,te.concat(se),l),fe(HTMLBodyElement.prototype,te.concat(se),l),fe(HTMLFrameElement.prototype,re,l),fe(HTMLIFrameElement.prototype,re,l);var p=d.HTMLMarqueeElement;p&&fe(p.prototype,ie,l);var h=d.Worker;h&&fe(h.prototype,le,l)}fe(XMLHttpRequest.prototype,ae,l);var v=u.XMLHttpRequestEventTarget;v&&fe(v&&v.prototype,ae,l),"undefined"!=typeof IDBIndex&&(fe(IDBIndex.prototype,ue,l),fe(IDBRequest.prototype,ue,l),fe(IDBOpenDBRequest.prototype,ue,l),fe(IDBDatabase.prototype,ue,l),fe(IDBTransaction.prototype,ue,l),fe(IDBCursor.prototype,ue,l)),c&&fe(WebSocket.prototype,ce,l)}else!function(){for(var e=function(e){var o=de[e],t="on"+o;self.addEventListener(o,function(e){var o,n,r=e.target;for(n=r?r.constructor.name+"."+t:"unknown."+t;r;)r[t]&&!r[t][pe]&&((o=m(r[t],n))[pe]=r[t],r[t]=o),r=r.parentElement},!0)},o=0;o<de.length;o++)e(o)}(),Z("XMLHttpRequest"),c&&function(e,t){var n=t.WebSocket;t.EventTarget||U(t,[n.prototype]),t.WebSocket=function(e,t){var u,c,l=arguments.length>1?new n(e,t):new n(e),d=o(l,"onmessage");return d&&!1===d.configurable?(u=r(l),c=l,[i,a,"send","close"].forEach(function(e){u[e]=function(){var o=s.call(arguments);if(e===i||e===a){var t=o.length>0?o[0]:void 0;if(t){var n=Zone.__symbol__("ON_PROPERTY"+t);l[n]=u[n]}}return l[e].apply(l,o)}})):u=l,D(u,["close","error","message","open"],c),u};var u=t.WebSocket;for(var c in n)u[c]=n[c]}(0,u)}}var pe=_("unbound");function _e(e,o){var t="Anchor,Area,Audio,BR,Base,BaseFont,Body,Button,Canvas,Content,DList,Directory,Div,Embed,FieldSet,Font,Form,Frame,FrameSet,HR,Head,Heading,Html,IFrame,Image,Input,Keygen,LI,Label,Legend,Link,Map,Marquee,Media,Menu,Meta,Meter,Mod,OList,Object,OptGroup,Option,Output,Paragraph,Pre,Progress,Quote,Script,Select,Source,Span,Style,TableCaption,TableCell,TableCol,Table,TableRow,TableSection,TextArea,Title,Track,UList,Unknown,Video",n="ApplicationCache,EventSource,FileReader,InputMethodContext,MediaController,MessagePort,Node,Performance,SVGElementInstance,SharedWorker,TextTrack,TextTrackCue,TextTrackList,WebKitNamedFlow,Window,Worker,WorkerGlobalScope,XMLHttpRequest,XMLHttpRequestEventTarget,XMLHttpRequestUpload,IDBRequest,IDBOpenDBRequest,IDBDatabase,IDBTransaction,IDBCursor,DBIndex,WebSocket".split(","),r=[],s=e.wtf,i=t.split(",");s?r=i.map(function(e){return"HTML"+e+"Element"}).concat(n):e.EventTarget?r.push("EventTarget"):r=n;for(var a=e.__Zone_disable_IE_check||!1,u=e.__Zone_enable_cross_context_check||!1,c=function(){if(L)return R;L=!0;try{var e=j.navigator.userAgent;return-1===e.indexOf("MSIE ")&&-1===e.indexOf("Trident/")&&-1===e.indexOf("Edge/")||(R=!0),R}catch(e){}}(),m="[object FunctionWrapper]",p="function __BROWSERTOOLS_CONSOLE_SAFEFUNC() { [native code] }",_=0;_<de.length;_++){var h=de[_],v=h+d,g=h+l,y=f+v,b=f+g;N[h]={},N[h][d]=y,N[h][l]=b}for(var _=0;_<t.length;_++)for(var k=i[_],w=B[k]={},T=0;T<de.length;T++){var h=de[T];w[h]=k+".addEventListener:"+h}for(var E=[],_=0;_<r.length;_++){var x=e[r[_]];E.push(x&&x.prototype)}return U(e,E,{vh:function(e,o,t,n){if(!a&&c)if(u)try{var r=o.toString();if(r===m||r==p)return e.apply(t,n),!1}catch(o){return e.apply(t,n),!1}else{var r=o.toString();if(r===m||r==p)return e.apply(t,n),!1}else if(u)try{o.toString()}catch(o){return e.apply(t,n),!1}return!0}}),o.patchEventTarget=U,!0}function he(e,o){var t,n;t=o,(n=e.Event)&&n.prototype&&t.patchMethod(n.prototype,"stopImmediatePropagation",function(e){return function(o,t){o[W]=!0,e&&e.apply(o,t)}})}function je(e,t,n,r){var s=Zone.__symbol__(n);if(!e[s]){var i=e[s]=e[n];e[n]=function(s,a,u){return a&&a.prototype&&r.forEach(function(e){var r,s,i,u,c=t+"."+n+"::"+e,l=a.prototype;if(l.hasOwnProperty(e)){var d=o(l,e);d&&d.value?(d.value=m(d.value,c),r=a.prototype,s=e,u=(i=d).configurable,i=ee(r,s,i),oe(r,s,i,u)):l[e]=m(l[e],c)}else l[e]&&(l[e]=m(l[e],c))}),i.call(e,s,a,u)},I(e[n],i)}}Zone.__load_patch("util",function(e,o,t){t.patchOnProperties=D,t.patchMethod=M,t.bindArguments=b}),Zone.__load_patch("timers",function(e){K(e,"set","clear","Timeout"),K(e,"set","clear","Interval"),K(e,"set","clear","Immediate")}),Zone.__load_patch("requestAnimationFrame",function(e){K(e,"request","cancel","AnimationFrame"),K(e,"mozRequest","mozCancel","AnimationFrame"),K(e,"webkitRequest","webkitCancel","AnimationFrame")}),Zone.__load_patch("blocking",function(e,o){for(var t=["alert","prompt","confirm"],n=0;n<t.length;n++){var r=t[n];M(e,r,function(t,n,r){return function(n,s){return o.current.run(t,e,s,r)}})}}),Zone.__load_patch("EventTarget",function(e,o,t){var n=o.__symbol__("BLACK_LISTED_EVENTS");e[n]&&(o[n]=e[n]),he(e,t),_e(e,t);var r=e.XMLHttpRequestEventTarget;r&&r.prototype&&t.patchEventTarget(e,[r.prototype]),Z("MutationObserver"),Z("WebKitMutationObserver"),Z("IntersectionObserver"),Z("FileReader")}),Zone.__load_patch("on_property",function(e,o,t){me(0,e),Object.defineProperty=function(e,o,t){if($(e,o))throw new TypeError("Cannot assign to read only property '"+o+"' of "+e);var n=t.configurable;return"prototype"!==o&&(t=ee(e,o,t)),oe(e,o,t,n)},Object.defineProperties=function(e,o){return Object.keys(o).forEach(function(t){Object.defineProperty(e,t,o[t])}),e},Object.create=function(e,o){return"object"!=typeof o||Object.isFrozen(o)||Object.keys(o).forEach(function(t){o[t]=ee(e,t,o[t])}),Y(e,o)},Object.getOwnPropertyDescriptor=function(e,o){var t=J(e,o);return t&&$(e,o)&&(t.configurable=!1),t}}),Zone.__load_patch("customElements",function(e,o,t){!function(e){(E||x)&&"registerElement"in e.document&&je(document,"Document","registerElement",["createdCallback","attachedCallback","detachedCallback","attributeChangedCallback"])}(e),function(e){(E||x)&&"customElements"in e&&je(e.customElements,"customElements","define",["connectedCallback","disconnectedCallback","adoptedCallback","attributeChangedCallback"])}(e)}),Zone.__load_patch("canvas",function(e){var o=e.HTMLCanvasElement;void 0!==o&&o.prototype&&o.prototype.toBlob&&function(e,o,t){var n=null;function r(e){var o=e.data;return o.args[o.cbIdx]=function(){e.invoke.apply(this,arguments)},n.apply(o.target,o.args),e}n=M(e,o,function(e){return function(o,n){var s=t(o,n);return s.cbIdx>=0&&"function"==typeof n[s.cbIdx]?p(s.name,n[s.cbIdx],s,r):e.apply(o,n)}})}(o.prototype,"toBlob",function(e,o){return{name:"HTMLCanvasElement.toBlob",target:e,cbIdx:0,args:o}})}),Zone.__load_patch("XHR",function(e,o){!function(e){var l=XMLHttpRequest.prototype,d=l[u],f=l[c];if(!d){var m=e.XMLHttpRequestEventTarget;if(m){var h=m.prototype;d=h[u],f=h[c]}}var j="readystatechange",v="scheduled";function g(e){var o=e.data,n=o.target;n[s]=!1,n[a]=!1;var i=n[r];d||(d=n[u],f=n[c]),i&&f.call(n,j,i);var l=n[r]=function(){if(n.readyState===n.DONE)if(!o.aborted&&n[s]&&e.state===v){var t=n.__zone_symbol__loadfalse;if(t&&t.length>0){var r=e.invoke;e.invoke=function(){for(var t=n.__zone_symbol__loadfalse,s=0;s<t.length;s++)t[s]===e&&t.splice(s,1);o.aborted||e.state!==v||r.call(e)},t.push(e)}else e.invoke()}else o.aborted||!1!==n[s]||(n[a]=!0)};d.call(n,j,l);var m=n[t];return m||(n[t]=e),E.apply(n,o.args),n[s]=!0,e}function y(){}function b(e){var o=e.data;return o.aborted=!0,x.apply(o.target,o.args)}var k=M(l,"open",function(){return function(e,o){return e[n]=0==o[2],e[i]=o[1],k.apply(e,o)}}),w=_("fetchTaskAborting"),T=_("fetchTaskScheduling"),E=M(l,"send",function(){return function(e,t){if(!0===o.current[T])return E.apply(e,t);if(e[n])return E.apply(e,t);var r={target:e,url:e[i],isPeriodic:!1,args:t,aborted:!1},s=p("XMLHttpRequest.send",y,r,g,b);e&&!0===e[a]&&!r.aborted&&s.state===v&&s.invoke()}}),x=M(l,"abort",function(){return function(e,n){var r=e[t];if(r&&"string"==typeof r.type){if(null==r.cancelFn||r.data&&r.data.aborted)return;r.zone.cancelTask(r)}else if(!0===o.current[w])return x.apply(e,n)}})}(e);var t=_("xhrTask"),n=_("xhrSync"),r=_("xhrListener"),s=_("xhrScheduled"),i=_("xhrURL"),a=_("xhrErrorBeforeScheduled")}),Zone.__load_patch("geolocation",function(e){e.navigator&&e.navigator.geolocation&&function(e,t){for(var n=e.constructor.name,r=function(r){var s=t[r],i=e[s];if(i){var a=o(e,s);if(!k(a))return"continue";e[s]=function(e){var o=function(){return e.apply(this,b(arguments,n+"."+s))};return I(o,e),o}(i)}},s=0;s<t.length;s++)r(s)}(e.navigator.geolocation,["getCurrentPosition","watchPosition"])}),Zone.__load_patch("PromiseRejectionEvent",function(e,o){function t(o){return function(t){var n=G(e,o);n.forEach(function(n){var r=e.PromiseRejectionEvent;if(r){var s=new r(o,{promise:t.promise,reason:t.rejection});n.invoke(s)}})}}e.PromiseRejectionEvent&&(o[_("unhandledPromiseRejectionHandler")]=t("unhandledrejection"),o[_("rejectionHandledHandler")]=t("rejectionhandled"))})}()},"./src/polyfills.ts":function(e,o,t){"use strict";t.r(o);t("./node_modules/zone.js/dist/zone.js")},4:function(e,o,t){t("./src/polyfills.ts"),e.exports=t("./node_modules/@angular-devkit/build-angular/src/angular-cli-files/models/jit-polyfills.js")}},[[4,"runtime"]]]);
</script><script type="text/javascript">
(window.webpackJsonp=window.webpackJsonp||[]).push([["styles"],{"./node_modules/@angular-devkit/build-angular/src/angular-cli-files/plugins/raw-css-loader.js!./node_modules/postcss-loader/src/index.js?!./node_modules/sass-loader/lib/loader.js?!./src/styles.scss":function(b,t){b.exports=[[b.i,'/**\n * Applies styles for users in high contrast mode. Note that this only applies\n * to Microsoft browsers. Chrome can be included by checking for the `html[hc]`\n * attribute, however Chrome handles high contrast differently.\n * @param target Which kind of high contrast setting to target. Defaults to `active`, can be\n *    `white-on-black` or `black-on-white`.\n */\n/* Theme for the ripple elements.*/\n/* stylelint-disable material/no-prefixes */\n/* stylelint-enable */\n.mat-badge-content {\n  font-weight: 600;\n  font-size: 12px;\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-badge-small .mat-badge-content {\n  font-size: 6px; }\n.mat-badge-large .mat-badge-content {\n  font-size: 24px; }\n.mat-h1, .mat-headline, .mat-typography h1 {\n  font: 400 24px/32px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 16px; }\n.mat-h2, .mat-title, .mat-typography h2 {\n  font: 500 20px/32px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 16px; }\n.mat-h3, .mat-subheading-2, .mat-typography h3 {\n  font: 400 16px/28px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 16px; }\n.mat-h4, .mat-subheading-1, .mat-typography h4 {\n  font: 400 15px/24px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 16px; }\n.mat-h5, .mat-typography h5 {\n  font: 400 11.62px/20px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 12px; }\n.mat-h6, .mat-typography h6 {\n  font: 400 9.38px/20px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 12px; }\n.mat-body-strong, .mat-body-2 {\n  font: 500 14px/24px Roboto, "Helvetica Neue", sans-serif; }\n.mat-body, .mat-body-1, .mat-typography {\n  font: 400 14px/20px Roboto, "Helvetica Neue", sans-serif; }\n.mat-body p, .mat-body-1 p, .mat-typography p {\n    margin: 0 0 12px; }\n.mat-small, .mat-caption {\n  font: 400 12px/20px Roboto, "Helvetica Neue", sans-serif; }\n.mat-display-4, .mat-typography .mat-display-4 {\n  font: 300 112px/112px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 56px;\n  letter-spacing: -0.05em; }\n.mat-display-3, .mat-typography .mat-display-3 {\n  font: 400 56px/56px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 64px;\n  letter-spacing: -0.02em; }\n.mat-display-2, .mat-typography .mat-display-2 {\n  font: 400 45px/48px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 64px;\n  letter-spacing: -0.005em; }\n.mat-display-1, .mat-typography .mat-display-1 {\n  font: 400 34px/40px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 64px; }\n.mat-bottom-sheet-container {\n  font: 400 14px/20px Roboto, "Helvetica Neue", sans-serif; }\n.mat-button, .mat-raised-button, .mat-icon-button, .mat-stroked-button,\n.mat-flat-button, .mat-fab, .mat-mini-fab {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 14px;\n  font-weight: 500; }\n.mat-button-toggle {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-card {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-card-title {\n  font-size: 24px;\n  font-weight: 500; }\n.mat-card-header .mat-card-title {\n  font-size: 20px; }\n.mat-card-subtitle,\n.mat-card-content {\n  font-size: 14px; }\n.mat-checkbox {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-checkbox-layout .mat-checkbox-label {\n  line-height: 24px; }\n.mat-chip {\n  font-size: 14px;\n  font-weight: 500; }\n.mat-chip .mat-chip-trailing-icon.mat-icon,\n  .mat-chip .mat-chip-remove.mat-icon {\n    font-size: 18px; }\n.mat-table {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-header-cell {\n  font-size: 12px;\n  font-weight: 500; }\n.mat-cell, .mat-footer-cell {\n  font-size: 14px; }\n.mat-calendar {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-calendar-body {\n  font-size: 13px; }\n.mat-calendar-body-label,\n.mat-calendar-period-button {\n  font-size: 14px;\n  font-weight: 500; }\n.mat-calendar-table-header th {\n  font-size: 11px;\n  font-weight: 400; }\n.mat-dialog-title {\n  font: 500 20px/32px Roboto, "Helvetica Neue", sans-serif; }\n.mat-expansion-panel-header {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 15px;\n  font-weight: 400; }\n.mat-expansion-panel-content {\n  font: 400 14px/20px Roboto, "Helvetica Neue", sans-serif; }\n.mat-form-field {\n  font-size: inherit;\n  font-weight: 400;\n  line-height: 1.125;\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-form-field-wrapper {\n  padding-bottom: 1.34375em; }\n.mat-form-field-prefix .mat-icon,\n.mat-form-field-suffix .mat-icon {\n  font-size: 150%;\n  line-height: 1.125; }\n.mat-form-field-prefix .mat-icon-button,\n.mat-form-field-suffix .mat-icon-button {\n  height: 1.5em;\n  width: 1.5em; }\n.mat-form-field-prefix .mat-icon-button .mat-icon,\n  .mat-form-field-suffix .mat-icon-button .mat-icon {\n    height: 1.125em;\n    line-height: 1.125; }\n.mat-form-field-infix {\n  padding: 0.5em 0;\n  border-top: 0.84375em solid transparent; }\n.mat-form-field-can-float.mat-form-field-should-float .mat-form-field-label,\n.mat-form-field-can-float .mat-input-server:focus + .mat-form-field-label-wrapper .mat-form-field-label {\n  -webkit-transform: translateY(-1.34375em) scale(0.75);\n          transform: translateY(-1.34375em) scale(0.75);\n  width: 133.33333333%; }\n.mat-form-field-can-float .mat-input-server[label]:not(:label-shown) + .mat-form-field-label-wrapper\n.mat-form-field-label {\n  -webkit-transform: translateY(-1.34374em) scale(0.75);\n          transform: translateY(-1.34374em) scale(0.75);\n  width: 133.33334333%; }\n.mat-form-field-label-wrapper {\n  top: -0.84375em;\n  padding-top: 0.84375em; }\n.mat-form-field-label {\n  top: 1.34375em; }\n.mat-form-field-underline {\n  bottom: 1.34375em; }\n.mat-form-field-subscript-wrapper {\n  font-size: 75%;\n  margin-top: 0.66666667em;\n  top: calc(100% - 1.79166667em); }\n.mat-form-field-appearance-legacy .mat-form-field-wrapper {\n  padding-bottom: 1.25em; }\n.mat-form-field-appearance-legacy .mat-form-field-infix {\n  padding: 0.4375em 0; }\n.mat-form-field-appearance-legacy.mat-form-field-can-float.mat-form-field-should-float .mat-form-field-label,\n.mat-form-field-appearance-legacy.mat-form-field-can-float .mat-input-server:focus + .mat-form-field-label-wrapper .mat-form-field-label {\n  -webkit-transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.001px);\n          transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.001px);\n  -ms-transform: translateY(-1.28125em) scale(0.75);\n  width: 133.33333333%; }\n.mat-form-field-appearance-legacy.mat-form-field-can-float .mat-form-field-autofill-control:-webkit-autofill + .mat-form-field-label-wrapper\n.mat-form-field-label {\n  -webkit-transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.00101px);\n          transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.00101px);\n  -ms-transform: translateY(-1.28124em) scale(0.75);\n  width: 133.33334333%; }\n.mat-form-field-appearance-legacy.mat-form-field-can-float .mat-input-server[label]:not(:label-shown) + .mat-form-field-label-wrapper\n.mat-form-field-label {\n  -webkit-transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.00102px);\n          transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.00102px);\n  -ms-transform: translateY(-1.28123em) scale(0.75);\n  width: 133.33335333%; }\n.mat-form-field-appearance-legacy .mat-form-field-label {\n  top: 1.28125em; }\n.mat-form-field-appearance-legacy .mat-form-field-underline {\n  bottom: 1.25em; }\n.mat-form-field-appearance-legacy .mat-form-field-subscript-wrapper {\n  margin-top: 0.54166667em;\n  top: calc(100% - 1.66666667em); }\n@media print {\n  .mat-form-field-appearance-legacy.mat-form-field-can-float.mat-form-field-should-float .mat-form-field-label,\n  .mat-form-field-appearance-legacy.mat-form-field-can-float .mat-input-server:focus + .mat-form-field-label-wrapper .mat-form-field-label {\n    -webkit-transform: translateY(-1.28122em) scale(0.75);\n            transform: translateY(-1.28122em) scale(0.75); }\n  .mat-form-field-appearance-legacy.mat-form-field-can-float .mat-form-field-autofill-control:-webkit-autofill + .mat-form-field-label-wrapper\n.mat-form-field-label {\n    -webkit-transform: translateY(-1.28121em) scale(0.75);\n            transform: translateY(-1.28121em) scale(0.75); }\n  .mat-form-field-appearance-legacy.mat-form-field-can-float .mat-input-server[label]:not(:label-shown) + .mat-form-field-label-wrapper\n.mat-form-field-label {\n    -webkit-transform: translateY(-1.2812em) scale(0.75);\n            transform: translateY(-1.2812em) scale(0.75); } }\n.mat-form-field-appearance-fill .mat-form-field-infix {\n  padding: 0.25em 0 0.75em 0; }\n.mat-form-field-appearance-fill .mat-form-field-label {\n  top: 1.09375em;\n  margin-top: -0.5em; }\n.mat-form-field-appearance-fill.mat-form-field-can-float.mat-form-field-should-float .mat-form-field-label,\n.mat-form-field-appearance-fill.mat-form-field-can-float .mat-input-server:focus + .mat-form-field-label-wrapper .mat-form-field-label {\n  -webkit-transform: translateY(-0.59375em) scale(0.75);\n          transform: translateY(-0.59375em) scale(0.75);\n  width: 133.33333333%; }\n.mat-form-field-appearance-fill.mat-form-field-can-float .mat-input-server[label]:not(:label-shown) + .mat-form-field-label-wrapper\n.mat-form-field-label {\n  -webkit-transform: translateY(-0.59374em) scale(0.75);\n          transform: translateY(-0.59374em) scale(0.75);\n  width: 133.33334333%; }\n.mat-form-field-appearance-outline .mat-form-field-infix {\n  padding: 1em 0 1em 0; }\n.mat-form-field-appearance-outline .mat-form-field-label {\n  top: 1.84375em;\n  margin-top: -0.25em; }\n.mat-form-field-appearance-outline.mat-form-field-can-float.mat-form-field-should-float .mat-form-field-label,\n.mat-form-field-appearance-outline.mat-form-field-can-float .mat-input-server:focus + .mat-form-field-label-wrapper .mat-form-field-label {\n  -webkit-transform: translateY(-1.59375em) scale(0.75);\n          transform: translateY(-1.59375em) scale(0.75);\n  width: 133.33333333%; }\n.mat-form-field-appearance-outline.mat-form-field-can-float .mat-input-server[label]:not(:label-shown) + .mat-form-field-label-wrapper\n.mat-form-field-label {\n  -webkit-transform: translateY(-1.59374em) scale(0.75);\n          transform: translateY(-1.59374em) scale(0.75);\n  width: 133.33334333%; }\n.mat-grid-tile-header,\n.mat-grid-tile-footer {\n  font-size: 14px; }\n.mat-grid-tile-header .mat-line,\n  .mat-grid-tile-footer .mat-line {\n    white-space: nowrap;\n    overflow: hidden;\n    text-overflow: ellipsis;\n    display: block;\n    box-sizing: border-box; }\n.mat-grid-tile-header .mat-line:nth-child(n+2),\n    .mat-grid-tile-footer .mat-line:nth-child(n+2) {\n      font-size: 12px; }\ninput.mat-input-element {\n  margin-top: -0.0625em; }\n.mat-menu-item {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 14px;\n  font-weight: 400; }\n.mat-paginator,\n.mat-paginator-page-size .mat-select-trigger {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 12px; }\n.mat-radio-button {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-select {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-select-trigger {\n  height: 1.125em; }\n.mat-slide-toggle-content {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-slider-thumb-label-text {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 12px;\n  font-weight: 500; }\n.mat-stepper-vertical, .mat-stepper-horizontal {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-step-label {\n  font-size: 14px;\n  font-weight: 400; }\n.mat-step-sub-label-error {\n  font-weight: normal; }\n.mat-step-label-error {\n  font-size: 14px; }\n.mat-step-label-selected {\n  font-size: 14px;\n  font-weight: 500; }\n.mat-tab-group {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-tab-label, .mat-tab-link {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 14px;\n  font-weight: 500; }\n.mat-toolbar,\n.mat-toolbar h1,\n.mat-toolbar h2,\n.mat-toolbar h3,\n.mat-toolbar h4,\n.mat-toolbar h5,\n.mat-toolbar h6 {\n  font: 500 20px/32px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0; }\n.mat-tooltip {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 10px;\n  padding-top: 6px;\n  padding-bottom: 6px; }\n.mat-tooltip-handset {\n  font-size: 14px;\n  padding-top: 8px;\n  padding-bottom: 8px; }\n.mat-list-item {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-list-option {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-list-base .mat-list-item {\n  font-size: 16px; }\n.mat-list-base .mat-list-item .mat-line {\n    white-space: nowrap;\n    overflow: hidden;\n    text-overflow: ellipsis;\n    display: block;\n    box-sizing: border-box; }\n.mat-list-base .mat-list-item .mat-line:nth-child(n+2) {\n      font-size: 14px; }\n.mat-list-base .mat-list-option {\n  font-size: 16px; }\n.mat-list-base .mat-list-option .mat-line {\n    white-space: nowrap;\n    overflow: hidden;\n    text-overflow: ellipsis;\n    display: block;\n    box-sizing: border-box; }\n.mat-list-base .mat-list-option .mat-line:nth-child(n+2) {\n      font-size: 14px; }\n.mat-list-base .mat-subheader {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 14px;\n  font-weight: 500; }\n.mat-list-base[dense] .mat-list-item {\n  font-size: 12px; }\n.mat-list-base[dense] .mat-list-item .mat-line {\n    white-space: nowrap;\n    overflow: hidden;\n    text-overflow: ellipsis;\n    display: block;\n    box-sizing: border-box; }\n.mat-list-base[dense] .mat-list-item .mat-line:nth-child(n+2) {\n      font-size: 12px; }\n.mat-list-base[dense] .mat-list-option {\n  font-size: 12px; }\n.mat-list-base[dense] .mat-list-option .mat-line {\n    white-space: nowrap;\n    overflow: hidden;\n    text-overflow: ellipsis;\n    display: block;\n    box-sizing: border-box; }\n.mat-list-base[dense] .mat-list-option .mat-line:nth-child(n+2) {\n      font-size: 12px; }\n.mat-list-base[dense] .mat-subheader {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 12px;\n  font-weight: 500; }\n.mat-option {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 16px; }\n.mat-optgroup-label {\n  font: 500 14px/24px Roboto, "Helvetica Neue", sans-serif; }\n.mat-simple-snackbar {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 14px; }\n.mat-simple-snackbar-action {\n  line-height: 1;\n  font-family: inherit;\n  font-size: inherit;\n  font-weight: 500; }\n.mat-tree {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n.mat-tree-node,\n.mat-nested-tree-node {\n  font-weight: 400;\n  font-size: 14px; }\n.mat-ripple {\n  overflow: hidden;\n  position: relative; }\n.mat-ripple.mat-ripple-unbounded {\n  overflow: visible; }\n.mat-ripple-element {\n  position: absolute;\n  border-radius: 50%;\n  pointer-events: none;\n  transition: opacity, -webkit-transform 0ms cubic-bezier(0, 0, 0.2, 1);\n  transition: opacity, transform 0ms cubic-bezier(0, 0, 0.2, 1);\n  transition: opacity, transform 0ms cubic-bezier(0, 0, 0.2, 1), -webkit-transform 0ms cubic-bezier(0, 0, 0.2, 1);\n  -webkit-transform: scale(0);\n          transform: scale(0); }\n@media (-ms-high-contrast: active) {\n    .mat-ripple-element {\n      display: none; } }\n.cdk-visually-hidden {\n  border: 0;\n  clip: rect(0 0 0 0);\n  height: 1px;\n  margin: -1px;\n  overflow: hidden;\n  padding: 0;\n  position: absolute;\n  width: 1px;\n  outline: 0;\n  -webkit-appearance: none;\n  -moz-appearance: none; }\n.cdk-overlay-container, .cdk-global-overlay-wrapper {\n  pointer-events: none;\n  top: 0;\n  left: 0;\n  height: 100%;\n  width: 100%; }\n.cdk-overlay-container {\n  position: fixed;\n  z-index: 1000; }\n.cdk-overlay-container:empty {\n    display: none; }\n.cdk-global-overlay-wrapper {\n  display: flex;\n  position: absolute;\n  z-index: 1000; }\n.cdk-overlay-pane {\n  position: absolute;\n  pointer-events: auto;\n  box-sizing: border-box;\n  z-index: 1000;\n  display: flex;\n  max-width: 100%;\n  max-height: 100%; }\n.cdk-overlay-backdrop {\n  position: absolute;\n  top: 0;\n  bottom: 0;\n  left: 0;\n  right: 0;\n  z-index: 1000;\n  pointer-events: auto;\n  -webkit-tap-highlight-color: transparent;\n  transition: opacity 400ms cubic-bezier(0.25, 0.8, 0.25, 1);\n  opacity: 0; }\n.cdk-overlay-backdrop.cdk-overlay-backdrop-showing {\n    opacity: 1; }\n@media screen and (-ms-high-contrast: active) {\n      .cdk-overlay-backdrop.cdk-overlay-backdrop-showing {\n        opacity: 0.6; } }\n.cdk-overlay-dark-backdrop {\n  background: rgba(0, 0, 0, 0.32); }\n.cdk-overlay-transparent-backdrop, .cdk-overlay-transparent-backdrop.cdk-overlay-backdrop-showing {\n  opacity: 0; }\n.cdk-overlay-connected-position-bounding-box {\n  position: absolute;\n  z-index: 1000;\n  display: flex;\n  flex-direction: column;\n  min-width: 1px;\n  min-height: 1px; }\n.cdk-global-scrollblock {\n  position: fixed;\n  width: 100%;\n  overflow-y: scroll; }\n@-webkit-keyframes cdk-text-field-autofill-start {\n  /*!*/ }\n@keyframes cdk-text-field-autofill-start {\n  /*!*/ }\n@-webkit-keyframes cdk-text-field-autofill-end {\n  /*!*/ }\n@keyframes cdk-text-field-autofill-end {\n  /*!*/ }\n.cdk-text-field-autofill-monitored:-webkit-autofill {\n  -webkit-animation-name: cdk-text-field-autofill-start;\n          animation-name: cdk-text-field-autofill-start; }\n.cdk-text-field-autofill-monitored:not(:-webkit-autofill) {\n  -webkit-animation-name: cdk-text-field-autofill-end;\n          animation-name: cdk-text-field-autofill-end; }\ntextarea.cdk-textarea-autosize {\n  resize: none; }\ntextarea.cdk-textarea-autosize-measuring {\n  height: auto !important;\n  overflow: hidden !important;\n  padding: 2px 0 !important;\n  box-sizing: content-box !important; }\n.mat-ripple-element {\n  background-color: rgba(0, 0, 0, 0.1); }\n.mat-option {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-option:hover:not(.mat-option-disabled), .mat-option:focus:not(.mat-option-disabled) {\n    background: rgba(0, 0, 0, 0.04); }\n.mat-option.mat-selected:not(.mat-option-multiple):not(.mat-option-disabled) {\n    background: rgba(0, 0, 0, 0.04); }\n.mat-option.mat-active {\n    background: rgba(0, 0, 0, 0.04);\n    color: rgba(0, 0, 0, 0.87); }\n.mat-option.mat-option-disabled {\n    color: rgba(0, 0, 0, 0.38); }\n.mat-primary .mat-option.mat-selected:not(.mat-option-disabled) {\n  color: #2a7bb8; }\n.mat-accent .mat-option.mat-selected:not(.mat-option-disabled) {\n  color: #81c6fa; }\n.mat-warn .mat-option.mat-selected:not(.mat-option-disabled) {\n  color: #f44336; }\n.mat-optgroup-label {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-optgroup-disabled .mat-optgroup-label {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-pseudo-checkbox {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-pseudo-checkbox::after {\n    color: #fafafa; }\n.mat-pseudo-checkbox-checked,\n.mat-pseudo-checkbox-indeterminate,\n.mat-accent .mat-pseudo-checkbox-checked,\n.mat-accent .mat-pseudo-checkbox-indeterminate {\n  background: #81c6fa; }\n.mat-primary .mat-pseudo-checkbox-checked,\n.mat-primary .mat-pseudo-checkbox-indeterminate {\n  background: #2a7bb8; }\n.mat-warn .mat-pseudo-checkbox-checked,\n.mat-warn .mat-pseudo-checkbox-indeterminate {\n  background: #f44336; }\n.mat-pseudo-checkbox-checked.mat-pseudo-checkbox-disabled,\n.mat-pseudo-checkbox-indeterminate.mat-pseudo-checkbox-disabled {\n  background: #b0b0b0; }\n.mat-elevation-z0 {\n  box-shadow: 0px 0px 0px 0px rgba(0, 0, 0, 0.2), 0px 0px 0px 0px rgba(0, 0, 0, 0.14), 0px 0px 0px 0px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z1 {\n  box-shadow: 0px 2px 1px -1px rgba(0, 0, 0, 0.2), 0px 1px 1px 0px rgba(0, 0, 0, 0.14), 0px 1px 3px 0px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z2 {\n  box-shadow: 0px 3px 1px -2px rgba(0, 0, 0, 0.2), 0px 2px 2px 0px rgba(0, 0, 0, 0.14), 0px 1px 5px 0px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z3 {\n  box-shadow: 0px 3px 3px -2px rgba(0, 0, 0, 0.2), 0px 3px 4px 0px rgba(0, 0, 0, 0.14), 0px 1px 8px 0px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z4 {\n  box-shadow: 0px 2px 4px -1px rgba(0, 0, 0, 0.2), 0px 4px 5px 0px rgba(0, 0, 0, 0.14), 0px 1px 10px 0px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z5 {\n  box-shadow: 0px 3px 5px -1px rgba(0, 0, 0, 0.2), 0px 5px 8px 0px rgba(0, 0, 0, 0.14), 0px 1px 14px 0px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z6 {\n  box-shadow: 0px 3px 5px -1px rgba(0, 0, 0, 0.2), 0px 6px 10px 0px rgba(0, 0, 0, 0.14), 0px 1px 18px 0px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z7 {\n  box-shadow: 0px 4px 5px -2px rgba(0, 0, 0, 0.2), 0px 7px 10px 1px rgba(0, 0, 0, 0.14), 0px 2px 16px 1px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z8 {\n  box-shadow: 0px 5px 5px -3px rgba(0, 0, 0, 0.2), 0px 8px 10px 1px rgba(0, 0, 0, 0.14), 0px 3px 14px 2px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z9 {\n  box-shadow: 0px 5px 6px -3px rgba(0, 0, 0, 0.2), 0px 9px 12px 1px rgba(0, 0, 0, 0.14), 0px 3px 16px 2px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z10 {\n  box-shadow: 0px 6px 6px -3px rgba(0, 0, 0, 0.2), 0px 10px 14px 1px rgba(0, 0, 0, 0.14), 0px 4px 18px 3px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z11 {\n  box-shadow: 0px 6px 7px -4px rgba(0, 0, 0, 0.2), 0px 11px 15px 1px rgba(0, 0, 0, 0.14), 0px 4px 20px 3px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z12 {\n  box-shadow: 0px 7px 8px -4px rgba(0, 0, 0, 0.2), 0px 12px 17px 2px rgba(0, 0, 0, 0.14), 0px 5px 22px 4px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z13 {\n  box-shadow: 0px 7px 8px -4px rgba(0, 0, 0, 0.2), 0px 13px 19px 2px rgba(0, 0, 0, 0.14), 0px 5px 24px 4px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z14 {\n  box-shadow: 0px 7px 9px -4px rgba(0, 0, 0, 0.2), 0px 14px 21px 2px rgba(0, 0, 0, 0.14), 0px 5px 26px 4px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z15 {\n  box-shadow: 0px 8px 9px -5px rgba(0, 0, 0, 0.2), 0px 15px 22px 2px rgba(0, 0, 0, 0.14), 0px 6px 28px 5px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z16 {\n  box-shadow: 0px 8px 10px -5px rgba(0, 0, 0, 0.2), 0px 16px 24px 2px rgba(0, 0, 0, 0.14), 0px 6px 30px 5px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z17 {\n  box-shadow: 0px 8px 11px -5px rgba(0, 0, 0, 0.2), 0px 17px 26px 2px rgba(0, 0, 0, 0.14), 0px 6px 32px 5px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z18 {\n  box-shadow: 0px 9px 11px -5px rgba(0, 0, 0, 0.2), 0px 18px 28px 2px rgba(0, 0, 0, 0.14), 0px 7px 34px 6px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z19 {\n  box-shadow: 0px 9px 12px -6px rgba(0, 0, 0, 0.2), 0px 19px 29px 2px rgba(0, 0, 0, 0.14), 0px 7px 36px 6px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z20 {\n  box-shadow: 0px 10px 13px -6px rgba(0, 0, 0, 0.2), 0px 20px 31px 3px rgba(0, 0, 0, 0.14), 0px 8px 38px 7px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z21 {\n  box-shadow: 0px 10px 13px -6px rgba(0, 0, 0, 0.2), 0px 21px 33px 3px rgba(0, 0, 0, 0.14), 0px 8px 40px 7px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z22 {\n  box-shadow: 0px 10px 14px -6px rgba(0, 0, 0, 0.2), 0px 22px 35px 3px rgba(0, 0, 0, 0.14), 0px 8px 42px 7px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z23 {\n  box-shadow: 0px 11px 14px -7px rgba(0, 0, 0, 0.2), 0px 23px 36px 3px rgba(0, 0, 0, 0.14), 0px 9px 44px 8px rgba(0, 0, 0, 0.12); }\n.mat-elevation-z24 {\n  box-shadow: 0px 11px 15px -7px rgba(0, 0, 0, 0.2), 0px 24px 38px 3px rgba(0, 0, 0, 0.14), 0px 9px 46px 8px rgba(0, 0, 0, 0.12); }\n.mat-app-background {\n  background-color: #fafafa;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-theme-loaded-marker {\n  display: none; }\n.mat-autocomplete-panel {\n  background: white;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-autocomplete-panel:not([class*=\'mat-elevation-z\']) {\n    box-shadow: 0px 2px 4px -1px rgba(0, 0, 0, 0.2), 0px 4px 5px 0px rgba(0, 0, 0, 0.14), 0px 1px 10px 0px rgba(0, 0, 0, 0.12); }\n.mat-autocomplete-panel .mat-option.mat-selected:not(.mat-active):not(:hover) {\n    background: white; }\n.mat-autocomplete-panel .mat-option.mat-selected:not(.mat-active):not(:hover):not(.mat-option-disabled) {\n      color: rgba(0, 0, 0, 0.87); }\n.mat-badge-content {\n  color: white;\n  background: #2a7bb8; }\n@media (-ms-high-contrast: active) {\n    .mat-badge-content {\n      outline: solid 1px;\n      border-radius: 0; } }\n.mat-badge-accent .mat-badge-content {\n  background: #81c6fa;\n  color: black; }\n.mat-badge-warn .mat-badge-content {\n  color: white;\n  background: #f44336; }\n.mat-badge {\n  position: relative; }\n.mat-badge-hidden .mat-badge-content {\n  display: none; }\n.mat-badge-disabled .mat-badge-content {\n  background: #b9b9b9;\n  color: rgba(0, 0, 0, 0.38); }\n.mat-badge-content {\n  position: absolute;\n  text-align: center;\n  display: inline-block;\n  border-radius: 50%;\n  transition: -webkit-transform 200ms ease-in-out;\n  transition: transform 200ms ease-in-out;\n  transition: transform 200ms ease-in-out, -webkit-transform 200ms ease-in-out;\n  -webkit-transform: scale(0.6);\n          transform: scale(0.6);\n  overflow: hidden;\n  white-space: nowrap;\n  text-overflow: ellipsis;\n  pointer-events: none; }\n.mat-badge-content.mat-badge-active {\n  -webkit-transform: none;\n          transform: none; }\n.mat-badge-small .mat-badge-content {\n  width: 16px;\n  height: 16px;\n  line-height: 16px; }\n.mat-badge-small.mat-badge-above .mat-badge-content {\n  top: -8px; }\n.mat-badge-small.mat-badge-below .mat-badge-content {\n  bottom: -8px; }\n.mat-badge-small.mat-badge-before .mat-badge-content {\n  left: -16px; }\n[dir=\'rtl\'] .mat-badge-small.mat-badge-before .mat-badge-content {\n  left: auto;\n  right: -16px; }\n.mat-badge-small.mat-badge-after .mat-badge-content {\n  right: -16px; }\n[dir=\'rtl\'] .mat-badge-small.mat-badge-after .mat-badge-content {\n  right: auto;\n  left: -16px; }\n.mat-badge-small.mat-badge-overlap.mat-badge-before .mat-badge-content {\n  left: -8px; }\n[dir=\'rtl\'] .mat-badge-small.mat-badge-overlap.mat-badge-before .mat-badge-content {\n  left: auto;\n  right: -8px; }\n.mat-badge-small.mat-badge-overlap.mat-badge-after .mat-badge-content {\n  right: -8px; }\n[dir=\'rtl\'] .mat-badge-small.mat-badge-overlap.mat-badge-after .mat-badge-content {\n  right: auto;\n  left: -8px; }\n.mat-badge-medium .mat-badge-content {\n  width: 22px;\n  height: 22px;\n  line-height: 22px; }\n.mat-badge-medium.mat-badge-above .mat-badge-content {\n  top: -11px; }\n.mat-badge-medium.mat-badge-below .mat-badge-content {\n  bottom: -11px; }\n.mat-badge-medium.mat-badge-before .mat-badge-content {\n  left: -22px; }\n[dir=\'rtl\'] .mat-badge-medium.mat-badge-before .mat-badge-content {\n  left: auto;\n  right: -22px; }\n.mat-badge-medium.mat-badge-after .mat-badge-content {\n  right: -22px; }\n[dir=\'rtl\'] .mat-badge-medium.mat-badge-after .mat-badge-content {\n  right: auto;\n  left: -22px; }\n.mat-badge-medium.mat-badge-overlap.mat-badge-before .mat-badge-content {\n  left: -11px; }\n[dir=\'rtl\'] .mat-badge-medium.mat-badge-overlap.mat-badge-before .mat-badge-content {\n  left: auto;\n  right: -11px; }\n.mat-badge-medium.mat-badge-overlap.mat-badge-after .mat-badge-content {\n  right: -11px; }\n[dir=\'rtl\'] .mat-badge-medium.mat-badge-overlap.mat-badge-after .mat-badge-content {\n  right: auto;\n  left: -11px; }\n.mat-badge-large .mat-badge-content {\n  width: 28px;\n  height: 28px;\n  line-height: 28px; }\n.mat-badge-large.mat-badge-above .mat-badge-content {\n  top: -14px; }\n.mat-badge-large.mat-badge-below .mat-badge-content {\n  bottom: -14px; }\n.mat-badge-large.mat-badge-before .mat-badge-content {\n  left: -28px; }\n[dir=\'rtl\'] .mat-badge-large.mat-badge-before .mat-badge-content {\n  left: auto;\n  right: -28px; }\n.mat-badge-large.mat-badge-after .mat-badge-content {\n  right: -28px; }\n[dir=\'rtl\'] .mat-badge-large.mat-badge-after .mat-badge-content {\n  right: auto;\n  left: -28px; }\n.mat-badge-large.mat-badge-overlap.mat-badge-before .mat-badge-content {\n  left: -14px; }\n[dir=\'rtl\'] .mat-badge-large.mat-badge-overlap.mat-badge-before .mat-badge-content {\n  left: auto;\n  right: -14px; }\n.mat-badge-large.mat-badge-overlap.mat-badge-after .mat-badge-content {\n  right: -14px; }\n[dir=\'rtl\'] .mat-badge-large.mat-badge-overlap.mat-badge-after .mat-badge-content {\n  right: auto;\n  left: -14px; }\n.mat-bottom-sheet-container {\n  box-shadow: 0px 8px 10px -5px rgba(0, 0, 0, 0.2), 0px 16px 24px 2px rgba(0, 0, 0, 0.14), 0px 6px 30px 5px rgba(0, 0, 0, 0.12);\n  background: white;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-button, .mat-icon-button, .mat-stroked-button {\n  color: inherit;\n  background: transparent; }\n.mat-button.mat-primary, .mat-icon-button.mat-primary, .mat-stroked-button.mat-primary {\n    color: #2a7bb8; }\n.mat-button.mat-accent, .mat-icon-button.mat-accent, .mat-stroked-button.mat-accent {\n    color: #81c6fa; }\n.mat-button.mat-warn, .mat-icon-button.mat-warn, .mat-stroked-button.mat-warn {\n    color: #f44336; }\n.mat-button.mat-primary[disabled], .mat-button.mat-accent[disabled], .mat-button.mat-warn[disabled], .mat-button[disabled][disabled], .mat-icon-button.mat-primary[disabled], .mat-icon-button.mat-accent[disabled], .mat-icon-button.mat-warn[disabled], .mat-icon-button[disabled][disabled], .mat-stroked-button.mat-primary[disabled], .mat-stroked-button.mat-accent[disabled], .mat-stroked-button.mat-warn[disabled], .mat-stroked-button[disabled][disabled] {\n    color: rgba(0, 0, 0, 0.26); }\n.mat-button.mat-primary .mat-button-focus-overlay, .mat-icon-button.mat-primary .mat-button-focus-overlay, .mat-stroked-button.mat-primary .mat-button-focus-overlay {\n    background-color: #2a7bb8; }\n.mat-button.mat-accent .mat-button-focus-overlay, .mat-icon-button.mat-accent .mat-button-focus-overlay, .mat-stroked-button.mat-accent .mat-button-focus-overlay {\n    background-color: #81c6fa; }\n.mat-button.mat-warn .mat-button-focus-overlay, .mat-icon-button.mat-warn .mat-button-focus-overlay, .mat-stroked-button.mat-warn .mat-button-focus-overlay {\n    background-color: #f44336; }\n.mat-button[disabled] .mat-button-focus-overlay, .mat-icon-button[disabled] .mat-button-focus-overlay, .mat-stroked-button[disabled] .mat-button-focus-overlay {\n    background-color: transparent; }\n.mat-button .mat-ripple-element, .mat-icon-button .mat-ripple-element, .mat-stroked-button .mat-ripple-element {\n    opacity: 0.1;\n    background-color: currentColor; }\n.mat-button-focus-overlay {\n  background: black; }\n.mat-stroked-button:not([disabled]) {\n  border-color: rgba(0, 0, 0, 0.12); }\n.mat-flat-button, .mat-raised-button, .mat-fab, .mat-mini-fab {\n  color: rgba(0, 0, 0, 0.87);\n  background-color: white; }\n.mat-flat-button.mat-primary, .mat-raised-button.mat-primary, .mat-fab.mat-primary, .mat-mini-fab.mat-primary {\n    color: white; }\n.mat-flat-button.mat-accent, .mat-raised-button.mat-accent, .mat-fab.mat-accent, .mat-mini-fab.mat-accent {\n    color: black; }\n.mat-flat-button.mat-warn, .mat-raised-button.mat-warn, .mat-fab.mat-warn, .mat-mini-fab.mat-warn {\n    color: white; }\n.mat-flat-button.mat-primary[disabled], .mat-flat-button.mat-accent[disabled], .mat-flat-button.mat-warn[disabled], .mat-flat-button[disabled][disabled], .mat-raised-button.mat-primary[disabled], .mat-raised-button.mat-accent[disabled], .mat-raised-button.mat-warn[disabled], .mat-raised-button[disabled][disabled], .mat-fab.mat-primary[disabled], .mat-fab.mat-accent[disabled], .mat-fab.mat-warn[disabled], .mat-fab[disabled][disabled], .mat-mini-fab.mat-primary[disabled], .mat-mini-fab.mat-accent[disabled], .mat-mini-fab.mat-warn[disabled], .mat-mini-fab[disabled][disabled] {\n    color: rgba(0, 0, 0, 0.26); }\n.mat-flat-button.mat-primary, .mat-raised-button.mat-primary, .mat-fab.mat-primary, .mat-mini-fab.mat-primary {\n    background-color: #2a7bb8; }\n.mat-flat-button.mat-accent, .mat-raised-button.mat-accent, .mat-fab.mat-accent, .mat-mini-fab.mat-accent {\n    background-color: #81c6fa; }\n.mat-flat-button.mat-warn, .mat-raised-button.mat-warn, .mat-fab.mat-warn, .mat-mini-fab.mat-warn {\n    background-color: #f44336; }\n.mat-flat-button.mat-primary[disabled], .mat-flat-button.mat-accent[disabled], .mat-flat-button.mat-warn[disabled], .mat-flat-button[disabled][disabled], .mat-raised-button.mat-primary[disabled], .mat-raised-button.mat-accent[disabled], .mat-raised-button.mat-warn[disabled], .mat-raised-button[disabled][disabled], .mat-fab.mat-primary[disabled], .mat-fab.mat-accent[disabled], .mat-fab.mat-warn[disabled], .mat-fab[disabled][disabled], .mat-mini-fab.mat-primary[disabled], .mat-mini-fab.mat-accent[disabled], .mat-mini-fab.mat-warn[disabled], .mat-mini-fab[disabled][disabled] {\n    background-color: rgba(0, 0, 0, 0.12); }\n.mat-flat-button.mat-primary .mat-ripple-element, .mat-raised-button.mat-primary .mat-ripple-element, .mat-fab.mat-primary .mat-ripple-element, .mat-mini-fab.mat-primary .mat-ripple-element {\n    background-color: rgba(255, 255, 255, 0.1); }\n.mat-flat-button.mat-accent .mat-ripple-element, .mat-raised-button.mat-accent .mat-ripple-element, .mat-fab.mat-accent .mat-ripple-element, .mat-mini-fab.mat-accent .mat-ripple-element {\n    background-color: rgba(0, 0, 0, 0.1); }\n.mat-flat-button.mat-warn .mat-ripple-element, .mat-raised-button.mat-warn .mat-ripple-element, .mat-fab.mat-warn .mat-ripple-element, .mat-mini-fab.mat-warn .mat-ripple-element {\n    background-color: rgba(255, 255, 255, 0.1); }\n.mat-stroked-button:not([class*=\'mat-elevation-z\']), .mat-flat-button:not([class*=\'mat-elevation-z\']) {\n  box-shadow: 0px 0px 0px 0px rgba(0, 0, 0, 0.2), 0px 0px 0px 0px rgba(0, 0, 0, 0.14), 0px 0px 0px 0px rgba(0, 0, 0, 0.12); }\n.mat-raised-button:not([class*=\'mat-elevation-z\']) {\n  box-shadow: 0px 3px 1px -2px rgba(0, 0, 0, 0.2), 0px 2px 2px 0px rgba(0, 0, 0, 0.14), 0px 1px 5px 0px rgba(0, 0, 0, 0.12); }\n.mat-raised-button:not([disabled]):active:not([class*=\'mat-elevation-z\']) {\n  box-shadow: 0px 5px 5px -3px rgba(0, 0, 0, 0.2), 0px 8px 10px 1px rgba(0, 0, 0, 0.14), 0px 3px 14px 2px rgba(0, 0, 0, 0.12); }\n.mat-raised-button[disabled]:not([class*=\'mat-elevation-z\']) {\n  box-shadow: 0px 0px 0px 0px rgba(0, 0, 0, 0.2), 0px 0px 0px 0px rgba(0, 0, 0, 0.14), 0px 0px 0px 0px rgba(0, 0, 0, 0.12); }\n.mat-fab:not([class*=\'mat-elevation-z\']), .mat-mini-fab:not([class*=\'mat-elevation-z\']) {\n  box-shadow: 0px 3px 5px -1px rgba(0, 0, 0, 0.2), 0px 6px 10px 0px rgba(0, 0, 0, 0.14), 0px 1px 18px 0px rgba(0, 0, 0, 0.12); }\n.mat-fab:not([disabled]):active:not([class*=\'mat-elevation-z\']), .mat-mini-fab:not([disabled]):active:not([class*=\'mat-elevation-z\']) {\n  box-shadow: 0px 7px 8px -4px rgba(0, 0, 0, 0.2), 0px 12px 17px 2px rgba(0, 0, 0, 0.14), 0px 5px 22px 4px rgba(0, 0, 0, 0.12); }\n.mat-fab[disabled]:not([class*=\'mat-elevation-z\']), .mat-mini-fab[disabled]:not([class*=\'mat-elevation-z\']) {\n  box-shadow: 0px 0px 0px 0px rgba(0, 0, 0, 0.2), 0px 0px 0px 0px rgba(0, 0, 0, 0.14), 0px 0px 0px 0px rgba(0, 0, 0, 0.12); }\n.mat-button-toggle-standalone,\n.mat-button-toggle-group {\n  box-shadow: 0px 3px 1px -2px rgba(0, 0, 0, 0.2), 0px 2px 2px 0px rgba(0, 0, 0, 0.14), 0px 1px 5px 0px rgba(0, 0, 0, 0.12); }\n.mat-button-toggle-standalone.mat-button-toggle-appearance-standard,\n.mat-button-toggle-group-appearance-standard {\n  box-shadow: none; }\n.mat-button-toggle {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-button-toggle .mat-button-toggle-focus-overlay {\n    background-color: rgba(0, 0, 0, 0.12); }\n.mat-button-toggle-appearance-standard {\n  color: rgba(0, 0, 0, 0.87);\n  background: white; }\n.mat-button-toggle-appearance-standard .mat-button-toggle-focus-overlay {\n    background-color: black; }\n.mat-button-toggle-group-appearance-standard .mat-button-toggle + .mat-button-toggle {\n  border-left: solid 1px rgba(0, 0, 0, 0.12); }\n[dir=\'rtl\'] .mat-button-toggle-group-appearance-standard .mat-button-toggle + .mat-button-toggle {\n  border-left: none;\n  border-right: solid 1px rgba(0, 0, 0, 0.12); }\n.mat-button-toggle-group-appearance-standard.mat-button-toggle-vertical .mat-button-toggle + .mat-button-toggle {\n  border-left: none;\n  border-right: none;\n  border-top: solid 1px rgba(0, 0, 0, 0.12); }\n.mat-button-toggle-checked {\n  background-color: #e0e0e0;\n  color: rgba(0, 0, 0, 0.54); }\n.mat-button-toggle-checked.mat-button-toggle-appearance-standard {\n    color: rgba(0, 0, 0, 0.87); }\n.mat-button-toggle-disabled {\n  color: rgba(0, 0, 0, 0.26);\n  background-color: #eeeeee; }\n.mat-button-toggle-disabled.mat-button-toggle-appearance-standard {\n    background: white; }\n.mat-button-toggle-disabled.mat-button-toggle-checked {\n    background-color: #bdbdbd; }\n.mat-button-toggle-standalone.mat-button-toggle-appearance-standard,\n.mat-button-toggle-group-appearance-standard {\n  border: solid 1px rgba(0, 0, 0, 0.12); }\n.mat-card {\n  background: white;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-card:not([class*=\'mat-elevation-z\']) {\n    box-shadow: 0px 2px 1px -1px rgba(0, 0, 0, 0.2), 0px 1px 1px 0px rgba(0, 0, 0, 0.14), 0px 1px 3px 0px rgba(0, 0, 0, 0.12); }\n.mat-card.mat-card-flat:not([class*=\'mat-elevation-z\']) {\n    box-shadow: 0px 0px 0px 0px rgba(0, 0, 0, 0.2), 0px 0px 0px 0px rgba(0, 0, 0, 0.14), 0px 0px 0px 0px rgba(0, 0, 0, 0.12); }\n.mat-card-subtitle {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-checkbox-frame {\n  border-color: rgba(0, 0, 0, 0.54); }\n.mat-checkbox-checkmark {\n  fill: #fafafa; }\n.mat-checkbox-checkmark-path {\n  stroke: #fafafa !important; }\n@media (-ms-high-contrast: black-on-white) {\n    .mat-checkbox-checkmark-path {\n      stroke: #000 !important; } }\n.mat-checkbox-mixedmark {\n  background-color: #fafafa; }\n.mat-checkbox-indeterminate.mat-primary .mat-checkbox-background, .mat-checkbox-checked.mat-primary .mat-checkbox-background {\n  background-color: #2a7bb8; }\n.mat-checkbox-indeterminate.mat-accent .mat-checkbox-background, .mat-checkbox-checked.mat-accent .mat-checkbox-background {\n  background-color: #81c6fa; }\n.mat-checkbox-indeterminate.mat-warn .mat-checkbox-background, .mat-checkbox-checked.mat-warn .mat-checkbox-background {\n  background-color: #f44336; }\n.mat-checkbox-disabled.mat-checkbox-checked .mat-checkbox-background, .mat-checkbox-disabled.mat-checkbox-indeterminate .mat-checkbox-background {\n  background-color: #b0b0b0; }\n.mat-checkbox-disabled:not(.mat-checkbox-checked) .mat-checkbox-frame {\n  border-color: #b0b0b0; }\n.mat-checkbox-disabled .mat-checkbox-label {\n  color: rgba(0, 0, 0, 0.54); }\n@media (-ms-high-contrast: active) {\n  .mat-checkbox-disabled {\n    opacity: 0.5; } }\n@media (-ms-high-contrast: active) {\n  .mat-checkbox-background {\n    background: none; } }\n.mat-checkbox:not(.mat-checkbox-disabled).mat-primary .mat-checkbox-ripple .mat-ripple-element {\n  background-color: #2a7bb8; }\n.mat-checkbox:not(.mat-checkbox-disabled).mat-accent .mat-checkbox-ripple .mat-ripple-element {\n  background-color: #81c6fa; }\n.mat-checkbox:not(.mat-checkbox-disabled).mat-warn .mat-checkbox-ripple .mat-ripple-element {\n  background-color: #f44336; }\n.mat-chip.mat-standard-chip {\n  background-color: #e0e0e0;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-chip.mat-standard-chip .mat-chip-remove {\n    color: rgba(0, 0, 0, 0.87);\n    opacity: 0.4; }\n.mat-chip.mat-standard-chip:not(.mat-chip-disabled):active {\n    box-shadow: 0px 3px 3px -2px rgba(0, 0, 0, 0.2), 0px 3px 4px 0px rgba(0, 0, 0, 0.14), 0px 1px 8px 0px rgba(0, 0, 0, 0.12); }\n.mat-chip.mat-standard-chip:not(.mat-chip-disabled) .mat-chip-remove:hover {\n    opacity: 0.54; }\n.mat-chip.mat-standard-chip.mat-chip-disabled {\n    opacity: 0.4; }\n.mat-chip.mat-standard-chip::after {\n    background: black; }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-primary {\n  background-color: #2a7bb8;\n  color: white; }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-primary .mat-chip-remove {\n    color: white;\n    opacity: 0.4; }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-primary .mat-ripple-element {\n    background: rgba(255, 255, 255, 0.1); }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-warn {\n  background-color: #f44336;\n  color: white; }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-warn .mat-chip-remove {\n    color: white;\n    opacity: 0.4; }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-warn .mat-ripple-element {\n    background: rgba(255, 255, 255, 0.1); }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-accent {\n  background-color: #81c6fa;\n  color: black; }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-accent .mat-chip-remove {\n    color: black;\n    opacity: 0.4; }\n.mat-chip.mat-standard-chip.mat-chip-selected.mat-accent .mat-ripple-element {\n    background: rgba(0, 0, 0, 0.1); }\n.mat-table {\n  background: white; }\n.mat-table thead, .mat-table tbody, .mat-table tfoot,\nmat-header-row, mat-row, mat-footer-row,\n[mat-header-row], [mat-row], [mat-footer-row],\n.mat-table-sticky {\n  background: inherit; }\nmat-row, mat-header-row, mat-footer-row,\nth.mat-header-cell, td.mat-cell, td.mat-footer-cell {\n  border-bottom-color: rgba(0, 0, 0, 0.12); }\n.mat-header-cell {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-cell, .mat-footer-cell {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-calendar-arrow {\n  border-top-color: rgba(0, 0, 0, 0.54); }\n.mat-datepicker-toggle,\n.mat-datepicker-content .mat-calendar-next-button,\n.mat-datepicker-content .mat-calendar-previous-button {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-calendar-table-header {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-calendar-table-header-divider::after {\n  background: rgba(0, 0, 0, 0.12); }\n.mat-calendar-body-label {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-calendar-body-cell-content {\n  color: rgba(0, 0, 0, 0.87);\n  border-color: transparent; }\n.mat-calendar-body-disabled > .mat-calendar-body-cell-content:not(.mat-calendar-body-selected) {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-calendar-body-cell:not(.mat-calendar-body-disabled):hover > .mat-calendar-body-cell-content:not(.mat-calendar-body-selected),\n.cdk-keyboard-focused .mat-calendar-body-active > .mat-calendar-body-cell-content:not(.mat-calendar-body-selected),\n.cdk-program-focused .mat-calendar-body-active > .mat-calendar-body-cell-content:not(.mat-calendar-body-selected) {\n  background-color: rgba(0, 0, 0, 0.04); }\n.mat-calendar-body-today:not(.mat-calendar-body-selected) {\n  border-color: rgba(0, 0, 0, 0.38); }\n.mat-calendar-body-disabled > .mat-calendar-body-today:not(.mat-calendar-body-selected) {\n  border-color: rgba(0, 0, 0, 0.18); }\n.mat-calendar-body-selected {\n  background-color: #2a7bb8;\n  color: white; }\n.mat-calendar-body-disabled > .mat-calendar-body-selected {\n  background-color: rgba(42, 123, 184, 0.4); }\n.mat-calendar-body-today.mat-calendar-body-selected {\n  box-shadow: inset 0 0 0 1px white; }\n.mat-datepicker-content {\n  box-shadow: 0px 2px 4px -1px rgba(0, 0, 0, 0.2), 0px 4px 5px 0px rgba(0, 0, 0, 0.14), 0px 1px 10px 0px rgba(0, 0, 0, 0.12);\n  background-color: white;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-datepicker-content.mat-accent .mat-calendar-body-selected {\n    background-color: #81c6fa;\n    color: black; }\n.mat-datepicker-content.mat-accent .mat-calendar-body-disabled > .mat-calendar-body-selected {\n    background-color: rgba(129, 198, 250, 0.4); }\n.mat-datepicker-content.mat-accent .mat-calendar-body-today.mat-calendar-body-selected {\n    box-shadow: inset 0 0 0 1px black; }\n.mat-datepicker-content.mat-warn .mat-calendar-body-selected {\n    background-color: #f44336;\n    color: white; }\n.mat-datepicker-content.mat-warn .mat-calendar-body-disabled > .mat-calendar-body-selected {\n    background-color: rgba(244, 67, 54, 0.4); }\n.mat-datepicker-content.mat-warn .mat-calendar-body-today.mat-calendar-body-selected {\n    box-shadow: inset 0 0 0 1px white; }\n.mat-datepicker-content-touch {\n  box-shadow: 0px 0px 0px 0px rgba(0, 0, 0, 0.2), 0px 0px 0px 0px rgba(0, 0, 0, 0.14), 0px 0px 0px 0px rgba(0, 0, 0, 0.12); }\n.mat-datepicker-toggle-active {\n  color: #2a7bb8; }\n.mat-datepicker-toggle-active.mat-accent {\n    color: #81c6fa; }\n.mat-datepicker-toggle-active.mat-warn {\n    color: #f44336; }\n.mat-dialog-container {\n  box-shadow: 0px 11px 15px -7px rgba(0, 0, 0, 0.2), 0px 24px 38px 3px rgba(0, 0, 0, 0.14), 0px 9px 46px 8px rgba(0, 0, 0, 0.12);\n  background: white;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-divider {\n  border-top-color: rgba(0, 0, 0, 0.12); }\n.mat-divider-vertical {\n  border-right-color: rgba(0, 0, 0, 0.12); }\n.mat-expansion-panel {\n  background: white;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-expansion-panel:not([class*=\'mat-elevation-z\']) {\n    box-shadow: 0px 3px 1px -2px rgba(0, 0, 0, 0.2), 0px 2px 2px 0px rgba(0, 0, 0, 0.14), 0px 1px 5px 0px rgba(0, 0, 0, 0.12); }\n.mat-action-row {\n  border-top-color: rgba(0, 0, 0, 0.12); }\n.mat-expansion-panel:not(.mat-expanded) .mat-expansion-panel-header:not([aria-disabled=\'true\']).cdk-keyboard-focused, .mat-expansion-panel:not(.mat-expanded) .mat-expansion-panel-header:not([aria-disabled=\'true\']).cdk-program-focused, .mat-expansion-panel:not(.mat-expanded) .mat-expansion-panel-header:not([aria-disabled=\'true\']):hover {\n  background: rgba(0, 0, 0, 0.04); }\n@media (hover: none) {\n  .mat-expansion-panel:not(.mat-expanded):not([aria-disabled=\'true\'])\n.mat-expansion-panel-header:hover {\n    background: white; } }\n.mat-expansion-panel-header-title {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-expansion-panel-header-description,\n.mat-expansion-indicator::after {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-expansion-panel-header[aria-disabled=\'true\'] {\n  color: rgba(0, 0, 0, 0.26); }\n.mat-expansion-panel-header[aria-disabled=\'true\'] .mat-expansion-panel-header-title,\n  .mat-expansion-panel-header[aria-disabled=\'true\'] .mat-expansion-panel-header-description {\n    color: inherit; }\n.mat-form-field-label {\n  color: rgba(0, 0, 0, 0.6); }\n.mat-hint {\n  color: rgba(0, 0, 0, 0.6); }\n.mat-form-field.mat-focused .mat-form-field-label {\n  color: #2a7bb8; }\n.mat-form-field.mat-focused .mat-form-field-label.mat-accent {\n    color: #81c6fa; }\n.mat-form-field.mat-focused .mat-form-field-label.mat-warn {\n    color: #f44336; }\n.mat-focused .mat-form-field-required-marker {\n  color: #81c6fa; }\n.mat-form-field-ripple {\n  background-color: rgba(0, 0, 0, 0.87); }\n.mat-form-field.mat-focused .mat-form-field-ripple {\n  background-color: #2a7bb8; }\n.mat-form-field.mat-focused .mat-form-field-ripple.mat-accent {\n    background-color: #81c6fa; }\n.mat-form-field.mat-focused .mat-form-field-ripple.mat-warn {\n    background-color: #f44336; }\n.mat-form-field-type-mat-native-select.mat-focused:not(.mat-form-field-invalid) .mat-form-field-infix::after {\n  color: #2a7bb8; }\n.mat-form-field-type-mat-native-select.mat-focused:not(.mat-form-field-invalid).mat-accent .mat-form-field-infix::after {\n  color: #81c6fa; }\n.mat-form-field-type-mat-native-select.mat-focused:not(.mat-form-field-invalid).mat-warn .mat-form-field-infix::after {\n  color: #f44336; }\n.mat-form-field.mat-form-field-invalid .mat-form-field-label {\n  color: #f44336; }\n.mat-form-field.mat-form-field-invalid .mat-form-field-label.mat-accent,\n  .mat-form-field.mat-form-field-invalid .mat-form-field-label .mat-form-field-required-marker {\n    color: #f44336; }\n.mat-form-field.mat-form-field-invalid .mat-form-field-ripple,\n.mat-form-field.mat-form-field-invalid .mat-form-field-ripple.mat-accent {\n  background-color: #f44336; }\n.mat-error {\n  color: #f44336; }\n.mat-form-field-appearance-legacy .mat-form-field-label {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-form-field-appearance-legacy .mat-hint {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-form-field-appearance-legacy .mat-form-field-underline {\n  background-color: rgba(0, 0, 0, 0.42); }\n.mat-form-field-appearance-legacy.mat-form-field-disabled .mat-form-field-underline {\n  background-image: linear-gradient(to right, rgba(0, 0, 0, 0.42) 0%, rgba(0, 0, 0, 0.42) 33%, transparent 0%);\n  background-size: 4px 100%;\n  background-repeat: repeat-x; }\n.mat-form-field-appearance-standard .mat-form-field-underline {\n  background-color: rgba(0, 0, 0, 0.42); }\n.mat-form-field-appearance-standard.mat-form-field-disabled .mat-form-field-underline {\n  background-image: linear-gradient(to right, rgba(0, 0, 0, 0.42) 0%, rgba(0, 0, 0, 0.42) 33%, transparent 0%);\n  background-size: 4px 100%;\n  background-repeat: repeat-x; }\n.mat-form-field-appearance-fill .mat-form-field-flex {\n  background-color: rgba(0, 0, 0, 0.04); }\n.mat-form-field-appearance-fill.mat-form-field-disabled .mat-form-field-flex {\n  background-color: rgba(0, 0, 0, 0.02); }\n.mat-form-field-appearance-fill .mat-form-field-underline::before {\n  background-color: rgba(0, 0, 0, 0.42); }\n.mat-form-field-appearance-fill.mat-form-field-disabled .mat-form-field-label {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-form-field-appearance-fill.mat-form-field-disabled .mat-form-field-underline::before {\n  background-color: transparent; }\n.mat-form-field-appearance-outline .mat-form-field-outline {\n  color: rgba(0, 0, 0, 0.12); }\n.mat-form-field-appearance-outline .mat-form-field-outline-thick {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-form-field-appearance-outline.mat-focused .mat-form-field-outline-thick {\n  color: #2a7bb8; }\n.mat-form-field-appearance-outline.mat-focused.mat-accent .mat-form-field-outline-thick {\n  color: #81c6fa; }\n.mat-form-field-appearance-outline.mat-focused.mat-warn .mat-form-field-outline-thick {\n  color: #f44336; }\n.mat-form-field-appearance-outline.mat-form-field-invalid.mat-form-field-invalid .mat-form-field-outline-thick {\n  color: #f44336; }\n.mat-form-field-appearance-outline.mat-form-field-disabled .mat-form-field-label {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-form-field-appearance-outline.mat-form-field-disabled .mat-form-field-outline {\n  color: rgba(0, 0, 0, 0.06); }\n.mat-icon.mat-primary {\n  color: #2a7bb8; }\n.mat-icon.mat-accent {\n  color: #81c6fa; }\n.mat-icon.mat-warn {\n  color: #f44336; }\n.mat-form-field-type-mat-native-select .mat-form-field-infix::after {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-input-element:disabled,\n.mat-form-field-type-mat-native-select.mat-form-field-disabled .mat-form-field-infix::after {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-input-element {\n  caret-color: #2a7bb8; }\n.mat-input-element::-webkit-input-placeholder {\n    color: rgba(0, 0, 0, 0.42); }\n.mat-input-element:-ms-input-placeholder {\n    color: rgba(0, 0, 0, 0.42); }\n.mat-input-element::-ms-input-placeholder {\n    color: rgba(0, 0, 0, 0.42); }\n.mat-input-element::placeholder {\n    color: rgba(0, 0, 0, 0.42); }\n.mat-input-element::-moz-placeholder {\n    color: rgba(0, 0, 0, 0.42); }\n.mat-input-element::-webkit-input-placeholder {\n    color: rgba(0, 0, 0, 0.42); }\n.mat-input-element:-ms-input-placeholder {\n    color: rgba(0, 0, 0, 0.42); }\n.mat-accent .mat-input-element {\n  caret-color: #81c6fa; }\n.mat-warn .mat-input-element,\n.mat-form-field-invalid .mat-input-element {\n  caret-color: #f44336; }\n.mat-form-field-type-mat-native-select.mat-form-field-invalid .mat-form-field-infix::after {\n  color: #f44336; }\n.mat-list-base .mat-list-item {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-list-base .mat-list-option {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-list-base .mat-subheader {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-list-item-disabled {\n  background-color: #eeeeee; }\n.mat-list-option:hover, .mat-list-option:focus,\n.mat-nav-list .mat-list-item:hover,\n.mat-nav-list .mat-list-item:focus {\n  background: rgba(0, 0, 0, 0.04); }\n.mat-menu-panel {\n  background: white; }\n.mat-menu-panel:not([class*=\'mat-elevation-z\']) {\n    box-shadow: 0px 2px 4px -1px rgba(0, 0, 0, 0.2), 0px 4px 5px 0px rgba(0, 0, 0, 0.14), 0px 1px 10px 0px rgba(0, 0, 0, 0.12); }\n.mat-menu-item {\n  background: transparent;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-menu-item[disabled], .mat-menu-item[disabled]::after {\n    color: rgba(0, 0, 0, 0.38); }\n.mat-menu-item .mat-icon-no-color,\n.mat-menu-item-submenu-trigger::after {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-menu-item:hover:not([disabled]),\n.mat-menu-item.cdk-program-focused:not([disabled]),\n.mat-menu-item.cdk-keyboard-focused:not([disabled]),\n.mat-menu-item-highlighted:not([disabled]) {\n  background: rgba(0, 0, 0, 0.04); }\n.mat-paginator {\n  background: white; }\n.mat-paginator,\n.mat-paginator-page-size .mat-select-trigger {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-paginator-decrement,\n.mat-paginator-increment {\n  border-top: 2px solid rgba(0, 0, 0, 0.54);\n  border-right: 2px solid rgba(0, 0, 0, 0.54); }\n.mat-paginator-first,\n.mat-paginator-last {\n  border-top: 2px solid rgba(0, 0, 0, 0.54); }\n.mat-icon-button[disabled] .mat-paginator-decrement,\n.mat-icon-button[disabled] .mat-paginator-increment,\n.mat-icon-button[disabled] .mat-paginator-first,\n.mat-icon-button[disabled] .mat-paginator-last {\n  border-color: rgba(0, 0, 0, 0.38); }\n.mat-progress-bar-background {\n  fill: #b1d4ed; }\n.mat-progress-bar-buffer {\n  background-color: #b1d4ed; }\n.mat-progress-bar-fill::after {\n  background-color: #2a7bb8; }\n.mat-progress-bar.mat-accent .mat-progress-bar-background {\n  fill: #b1d4ed; }\n.mat-progress-bar.mat-accent .mat-progress-bar-buffer {\n  background-color: #b1d4ed; }\n.mat-progress-bar.mat-accent .mat-progress-bar-fill::after {\n  background-color: #81c6fa; }\n.mat-progress-bar.mat-warn .mat-progress-bar-background {\n  fill: #ffcdd2; }\n.mat-progress-bar.mat-warn .mat-progress-bar-buffer {\n  background-color: #ffcdd2; }\n.mat-progress-bar.mat-warn .mat-progress-bar-fill::after {\n  background-color: #f44336; }\n.mat-progress-spinner circle, .mat-spinner circle {\n  stroke: #2a7bb8; }\n.mat-progress-spinner.mat-accent circle, .mat-spinner.mat-accent circle {\n  stroke: #81c6fa; }\n.mat-progress-spinner.mat-warn circle, .mat-spinner.mat-warn circle {\n  stroke: #f44336; }\n.mat-radio-outer-circle {\n  border-color: rgba(0, 0, 0, 0.54); }\n.mat-radio-button.mat-primary.mat-radio-checked .mat-radio-outer-circle {\n  border-color: #2a7bb8; }\n.mat-radio-button.mat-primary .mat-radio-inner-circle,\n.mat-radio-button.mat-primary .mat-radio-ripple .mat-ripple-element:not(.mat-radio-persistent-ripple),\n.mat-radio-button.mat-primary.mat-radio-checked .mat-radio-persistent-ripple,\n.mat-radio-button.mat-primary:active .mat-radio-persistent-ripple {\n  background-color: #2a7bb8; }\n.mat-radio-button.mat-accent.mat-radio-checked .mat-radio-outer-circle {\n  border-color: #81c6fa; }\n.mat-radio-button.mat-accent .mat-radio-inner-circle,\n.mat-radio-button.mat-accent .mat-radio-ripple .mat-ripple-element:not(.mat-radio-persistent-ripple),\n.mat-radio-button.mat-accent.mat-radio-checked .mat-radio-persistent-ripple,\n.mat-radio-button.mat-accent:active .mat-radio-persistent-ripple {\n  background-color: #81c6fa; }\n.mat-radio-button.mat-warn.mat-radio-checked .mat-radio-outer-circle {\n  border-color: #f44336; }\n.mat-radio-button.mat-warn .mat-radio-inner-circle,\n.mat-radio-button.mat-warn .mat-radio-ripple .mat-ripple-element:not(.mat-radio-persistent-ripple),\n.mat-radio-button.mat-warn.mat-radio-checked .mat-radio-persistent-ripple,\n.mat-radio-button.mat-warn:active .mat-radio-persistent-ripple {\n  background-color: #f44336; }\n.mat-radio-button.mat-radio-disabled.mat-radio-checked .mat-radio-outer-circle,\n.mat-radio-button.mat-radio-disabled .mat-radio-outer-circle {\n  border-color: rgba(0, 0, 0, 0.38); }\n.mat-radio-button.mat-radio-disabled .mat-radio-ripple .mat-ripple-element,\n.mat-radio-button.mat-radio-disabled .mat-radio-inner-circle {\n  background-color: rgba(0, 0, 0, 0.38); }\n.mat-radio-button.mat-radio-disabled .mat-radio-label-content {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-radio-button .mat-ripple-element {\n  background-color: black; }\n.mat-select-value {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-select-placeholder {\n  color: rgba(0, 0, 0, 0.42); }\n.mat-select-disabled .mat-select-value {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-select-arrow {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-select-panel {\n  background: white; }\n.mat-select-panel:not([class*=\'mat-elevation-z\']) {\n    box-shadow: 0px 2px 4px -1px rgba(0, 0, 0, 0.2), 0px 4px 5px 0px rgba(0, 0, 0, 0.14), 0px 1px 10px 0px rgba(0, 0, 0, 0.12); }\n.mat-select-panel .mat-option.mat-selected:not(.mat-option-multiple) {\n    background: rgba(0, 0, 0, 0.12); }\n.mat-form-field.mat-focused.mat-primary .mat-select-arrow {\n  color: #2a7bb8; }\n.mat-form-field.mat-focused.mat-accent .mat-select-arrow {\n  color: #81c6fa; }\n.mat-form-field.mat-focused.mat-warn .mat-select-arrow {\n  color: #f44336; }\n.mat-form-field .mat-select.mat-select-invalid .mat-select-arrow {\n  color: #f44336; }\n.mat-form-field .mat-select.mat-select-disabled .mat-select-arrow {\n  color: rgba(0, 0, 0, 0.38); }\n.mat-drawer-container {\n  background-color: #fafafa;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-drawer {\n  background-color: white;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-drawer.mat-drawer-push {\n    background-color: white; }\n.mat-drawer:not(.mat-drawer-side) {\n    box-shadow: 0px 8px 10px -5px rgba(0, 0, 0, 0.2), 0px 16px 24px 2px rgba(0, 0, 0, 0.14), 0px 6px 30px 5px rgba(0, 0, 0, 0.12); }\n.mat-drawer-side {\n  border-right: solid 1px rgba(0, 0, 0, 0.12); }\n.mat-drawer-side.mat-drawer-end {\n    border-left: solid 1px rgba(0, 0, 0, 0.12);\n    border-right: none; }\n[dir=\'rtl\'] .mat-drawer-side {\n  border-left: solid 1px rgba(0, 0, 0, 0.12);\n  border-right: none; }\n[dir=\'rtl\'] .mat-drawer-side.mat-drawer-end {\n    border-left: none;\n    border-right: solid 1px rgba(0, 0, 0, 0.12); }\n.mat-drawer-backdrop.mat-drawer-shown {\n  background-color: rgba(0, 0, 0, 0.6); }\n.mat-slide-toggle.mat-checked .mat-slide-toggle-thumb {\n  background-color: #81c6fa; }\n.mat-slide-toggle.mat-checked .mat-slide-toggle-bar {\n  background-color: rgba(129, 198, 250, 0.54); }\n.mat-slide-toggle.mat-checked .mat-ripple-element {\n  background-color: #81c6fa; }\n.mat-slide-toggle.mat-primary.mat-checked .mat-slide-toggle-thumb {\n  background-color: #2a7bb8; }\n.mat-slide-toggle.mat-primary.mat-checked .mat-slide-toggle-bar {\n  background-color: rgba(42, 123, 184, 0.54); }\n.mat-slide-toggle.mat-primary.mat-checked .mat-ripple-element {\n  background-color: #2a7bb8; }\n.mat-slide-toggle.mat-warn.mat-checked .mat-slide-toggle-thumb {\n  background-color: #f44336; }\n.mat-slide-toggle.mat-warn.mat-checked .mat-slide-toggle-bar {\n  background-color: rgba(244, 67, 54, 0.54); }\n.mat-slide-toggle.mat-warn.mat-checked .mat-ripple-element {\n  background-color: #f44336; }\n.mat-slide-toggle:not(.mat-checked) .mat-ripple-element {\n  background-color: black; }\n.mat-slide-toggle-thumb {\n  box-shadow: 0px 2px 1px -1px rgba(0, 0, 0, 0.2), 0px 1px 1px 0px rgba(0, 0, 0, 0.14), 0px 1px 3px 0px rgba(0, 0, 0, 0.12);\n  background-color: #fafafa; }\n.mat-slide-toggle-bar {\n  background-color: rgba(0, 0, 0, 0.38); }\n.mat-slider-track-background {\n  background-color: rgba(0, 0, 0, 0.26); }\n.mat-primary .mat-slider-track-fill,\n.mat-primary .mat-slider-thumb,\n.mat-primary .mat-slider-thumb-label {\n  background-color: #2a7bb8; }\n.mat-primary .mat-slider-thumb-label-text {\n  color: white; }\n.mat-accent .mat-slider-track-fill,\n.mat-accent .mat-slider-thumb,\n.mat-accent .mat-slider-thumb-label {\n  background-color: #81c6fa; }\n.mat-accent .mat-slider-thumb-label-text {\n  color: black; }\n.mat-warn .mat-slider-track-fill,\n.mat-warn .mat-slider-thumb,\n.mat-warn .mat-slider-thumb-label {\n  background-color: #f44336; }\n.mat-warn .mat-slider-thumb-label-text {\n  color: white; }\n.mat-slider-focus-ring {\n  background-color: rgba(129, 198, 250, 0.2); }\n.mat-slider:hover .mat-slider-track-background,\n.cdk-focused .mat-slider-track-background {\n  background-color: rgba(0, 0, 0, 0.38); }\n.mat-slider-disabled .mat-slider-track-background,\n.mat-slider-disabled .mat-slider-track-fill,\n.mat-slider-disabled .mat-slider-thumb {\n  background-color: rgba(0, 0, 0, 0.26); }\n.mat-slider-disabled:hover .mat-slider-track-background {\n  background-color: rgba(0, 0, 0, 0.26); }\n.mat-slider-min-value .mat-slider-focus-ring {\n  background-color: rgba(0, 0, 0, 0.12); }\n.mat-slider-min-value.mat-slider-thumb-label-showing .mat-slider-thumb,\n.mat-slider-min-value.mat-slider-thumb-label-showing .mat-slider-thumb-label {\n  background-color: rgba(0, 0, 0, 0.87); }\n.mat-slider-min-value.mat-slider-thumb-label-showing.cdk-focused .mat-slider-thumb,\n.mat-slider-min-value.mat-slider-thumb-label-showing.cdk-focused .mat-slider-thumb-label {\n  background-color: rgba(0, 0, 0, 0.26); }\n.mat-slider-min-value:not(.mat-slider-thumb-label-showing) .mat-slider-thumb {\n  border-color: rgba(0, 0, 0, 0.26);\n  background-color: transparent; }\n.mat-slider-min-value:not(.mat-slider-thumb-label-showing):hover .mat-slider-thumb, .mat-slider-min-value:not(.mat-slider-thumb-label-showing).cdk-focused .mat-slider-thumb {\n  border-color: rgba(0, 0, 0, 0.38); }\n.mat-slider-min-value:not(.mat-slider-thumb-label-showing):hover.mat-slider-disabled .mat-slider-thumb, .mat-slider-min-value:not(.mat-slider-thumb-label-showing).cdk-focused.mat-slider-disabled .mat-slider-thumb {\n  border-color: rgba(0, 0, 0, 0.26); }\n.mat-slider-has-ticks .mat-slider-wrapper::after {\n  border-color: rgba(0, 0, 0, 0.7); }\n.mat-slider-horizontal .mat-slider-ticks {\n  background-image: repeating-linear-gradient(to right, rgba(0, 0, 0, 0.7), rgba(0, 0, 0, 0.7) 2px, transparent 0, transparent);\n  background-image: -moz-repeating-linear-gradient(0.0001deg, rgba(0, 0, 0, 0.7), rgba(0, 0, 0, 0.7) 2px, transparent 0, transparent); }\n.mat-slider-vertical .mat-slider-ticks {\n  background-image: repeating-linear-gradient(to bottom, rgba(0, 0, 0, 0.7), rgba(0, 0, 0, 0.7) 2px, transparent 0, transparent); }\n.mat-step-header.cdk-keyboard-focused, .mat-step-header.cdk-program-focused, .mat-step-header:hover {\n  background-color: rgba(0, 0, 0, 0.04); }\n@media (hover: none) {\n  .mat-step-header:hover {\n    background: none; } }\n.mat-step-header .mat-step-label,\n.mat-step-header .mat-step-optional {\n  color: rgba(0, 0, 0, 0.54); }\n.mat-step-header .mat-step-icon {\n  background-color: rgba(0, 0, 0, 0.54);\n  color: white; }\n.mat-step-header .mat-step-icon-selected,\n.mat-step-header .mat-step-icon-state-done,\n.mat-step-header .mat-step-icon-state-edit {\n  background-color: #2a7bb8;\n  color: white; }\n.mat-step-header .mat-step-icon-state-error {\n  background-color: transparent;\n  color: #f44336; }\n.mat-step-header .mat-step-label.mat-step-label-active {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-step-header .mat-step-label.mat-step-label-error {\n  color: #f44336; }\n.mat-stepper-horizontal, .mat-stepper-vertical {\n  background-color: white; }\n.mat-stepper-vertical-line::before {\n  border-left-color: rgba(0, 0, 0, 0.12); }\n.mat-horizontal-stepper-header::before,\n.mat-horizontal-stepper-header::after,\n.mat-stepper-horizontal-line {\n  border-top-color: rgba(0, 0, 0, 0.12); }\n.mat-sort-header-arrow {\n  color: #757575; }\n.mat-tab-nav-bar,\n.mat-tab-header {\n  border-bottom: 1px solid rgba(0, 0, 0, 0.12); }\n.mat-tab-group-inverted-header .mat-tab-nav-bar,\n.mat-tab-group-inverted-header .mat-tab-header {\n  border-top: 1px solid rgba(0, 0, 0, 0.12);\n  border-bottom: none; }\n.mat-tab-label, .mat-tab-link {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-tab-label.mat-tab-disabled, .mat-tab-link.mat-tab-disabled {\n    color: rgba(0, 0, 0, 0.38); }\n.mat-tab-header-pagination-chevron {\n  border-color: rgba(0, 0, 0, 0.87); }\n.mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron {\n  border-color: rgba(0, 0, 0, 0.38); }\n.mat-tab-group[class*=\'mat-background-\'] .mat-tab-header,\n.mat-tab-nav-bar[class*=\'mat-background-\'] {\n  border-bottom: none;\n  border-top: none; }\n.mat-tab-group.mat-primary .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-group.mat-primary .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-primary .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-primary .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-primary .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-primary .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-primary .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-primary .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled) {\n  background-color: rgba(177, 212, 237, 0.3); }\n.mat-tab-group.mat-primary .mat-ink-bar, .mat-tab-nav-bar.mat-primary .mat-ink-bar {\n  background-color: #2a7bb8; }\n.mat-tab-group.mat-primary.mat-background-primary .mat-ink-bar, .mat-tab-nav-bar.mat-primary.mat-background-primary .mat-ink-bar {\n  background-color: white; }\n.mat-tab-group.mat-accent .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-group.mat-accent .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-accent .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-accent .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-accent .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-accent .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-accent .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-accent .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled) {\n  background-color: rgba(177, 212, 237, 0.3); }\n.mat-tab-group.mat-accent .mat-ink-bar, .mat-tab-nav-bar.mat-accent .mat-ink-bar {\n  background-color: #81c6fa; }\n.mat-tab-group.mat-accent.mat-background-accent .mat-ink-bar, .mat-tab-nav-bar.mat-accent.mat-background-accent .mat-ink-bar {\n  background-color: black; }\n.mat-tab-group.mat-warn .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-group.mat-warn .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-warn .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-warn .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-warn .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-warn .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-warn .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-warn .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled) {\n  background-color: rgba(255, 205, 210, 0.3); }\n.mat-tab-group.mat-warn .mat-ink-bar, .mat-tab-nav-bar.mat-warn .mat-ink-bar {\n  background-color: #f44336; }\n.mat-tab-group.mat-warn.mat-background-warn .mat-ink-bar, .mat-tab-nav-bar.mat-warn.mat-background-warn .mat-ink-bar {\n  background-color: white; }\n.mat-tab-group.mat-background-primary .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-group.mat-background-primary .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-background-primary .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-background-primary .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-background-primary .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-background-primary .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-background-primary .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-background-primary .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled) {\n  background-color: rgba(177, 212, 237, 0.3); }\n.mat-tab-group.mat-background-primary .mat-tab-header, .mat-tab-group.mat-background-primary .mat-tab-links, .mat-tab-nav-bar.mat-background-primary .mat-tab-header, .mat-tab-nav-bar.mat-background-primary .mat-tab-links {\n  background-color: #2a7bb8; }\n.mat-tab-group.mat-background-primary .mat-tab-label, .mat-tab-group.mat-background-primary .mat-tab-link, .mat-tab-nav-bar.mat-background-primary .mat-tab-label, .mat-tab-nav-bar.mat-background-primary .mat-tab-link {\n  color: white; }\n.mat-tab-group.mat-background-primary .mat-tab-label.mat-tab-disabled, .mat-tab-group.mat-background-primary .mat-tab-link.mat-tab-disabled, .mat-tab-nav-bar.mat-background-primary .mat-tab-label.mat-tab-disabled, .mat-tab-nav-bar.mat-background-primary .mat-tab-link.mat-tab-disabled {\n    color: rgba(255, 255, 255, 0.4); }\n.mat-tab-group.mat-background-primary .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-primary .mat-tab-header-pagination-chevron {\n  border-color: white; }\n.mat-tab-group.mat-background-primary .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-primary .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron {\n  border-color: rgba(255, 255, 255, 0.4); }\n.mat-tab-group.mat-background-primary .mat-ripple-element, .mat-tab-nav-bar.mat-background-primary .mat-ripple-element {\n  background-color: rgba(255, 255, 255, 0.12); }\n.mat-tab-group.mat-background-accent .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-group.mat-background-accent .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-background-accent .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-background-accent .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-background-accent .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-background-accent .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-background-accent .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-background-accent .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled) {\n  background-color: rgba(177, 212, 237, 0.3); }\n.mat-tab-group.mat-background-accent .mat-tab-header, .mat-tab-group.mat-background-accent .mat-tab-links, .mat-tab-nav-bar.mat-background-accent .mat-tab-header, .mat-tab-nav-bar.mat-background-accent .mat-tab-links {\n  background-color: #81c6fa; }\n.mat-tab-group.mat-background-accent .mat-tab-label, .mat-tab-group.mat-background-accent .mat-tab-link, .mat-tab-nav-bar.mat-background-accent .mat-tab-label, .mat-tab-nav-bar.mat-background-accent .mat-tab-link {\n  color: black; }\n.mat-tab-group.mat-background-accent .mat-tab-label.mat-tab-disabled, .mat-tab-group.mat-background-accent .mat-tab-link.mat-tab-disabled, .mat-tab-nav-bar.mat-background-accent .mat-tab-label.mat-tab-disabled, .mat-tab-nav-bar.mat-background-accent .mat-tab-link.mat-tab-disabled {\n    color: rgba(0, 0, 0, 0.4); }\n.mat-tab-group.mat-background-accent .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-accent .mat-tab-header-pagination-chevron {\n  border-color: black; }\n.mat-tab-group.mat-background-accent .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-accent .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron {\n  border-color: rgba(0, 0, 0, 0.4); }\n.mat-tab-group.mat-background-accent .mat-ripple-element, .mat-tab-nav-bar.mat-background-accent .mat-ripple-element {\n  background-color: rgba(0, 0, 0, 0.12); }\n.mat-tab-group.mat-background-warn .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-group.mat-background-warn .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-background-warn .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-group.mat-background-warn .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-background-warn .mat-tab-label.cdk-keyboard-focused:not(.mat-tab-disabled), .mat-tab-nav-bar.mat-background-warn .mat-tab-label.cdk-program-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-background-warn .mat-tab-link.cdk-keyboard-focused:not(.mat-tab-disabled),\n.mat-tab-nav-bar.mat-background-warn .mat-tab-link.cdk-program-focused:not(.mat-tab-disabled) {\n  background-color: rgba(255, 205, 210, 0.3); }\n.mat-tab-group.mat-background-warn .mat-tab-header, .mat-tab-group.mat-background-warn .mat-tab-links, .mat-tab-nav-bar.mat-background-warn .mat-tab-header, .mat-tab-nav-bar.mat-background-warn .mat-tab-links {\n  background-color: #f44336; }\n.mat-tab-group.mat-background-warn .mat-tab-label, .mat-tab-group.mat-background-warn .mat-tab-link, .mat-tab-nav-bar.mat-background-warn .mat-tab-label, .mat-tab-nav-bar.mat-background-warn .mat-tab-link {\n  color: white; }\n.mat-tab-group.mat-background-warn .mat-tab-label.mat-tab-disabled, .mat-tab-group.mat-background-warn .mat-tab-link.mat-tab-disabled, .mat-tab-nav-bar.mat-background-warn .mat-tab-label.mat-tab-disabled, .mat-tab-nav-bar.mat-background-warn .mat-tab-link.mat-tab-disabled {\n    color: rgba(255, 255, 255, 0.4); }\n.mat-tab-group.mat-background-warn .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-warn .mat-tab-header-pagination-chevron {\n  border-color: white; }\n.mat-tab-group.mat-background-warn .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-warn .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron {\n  border-color: rgba(255, 255, 255, 0.4); }\n.mat-tab-group.mat-background-warn .mat-ripple-element, .mat-tab-nav-bar.mat-background-warn .mat-ripple-element {\n  background-color: rgba(255, 255, 255, 0.12); }\n.mat-toolbar {\n  background: whitesmoke;\n  color: rgba(0, 0, 0, 0.87); }\n.mat-toolbar.mat-primary {\n    background: #2a7bb8;\n    color: white; }\n.mat-toolbar.mat-accent {\n    background: #81c6fa;\n    color: black; }\n.mat-toolbar.mat-warn {\n    background: #f44336;\n    color: white; }\n.mat-toolbar .mat-form-field-underline,\n  .mat-toolbar .mat-form-field-ripple,\n  .mat-toolbar .mat-focused .mat-form-field-ripple {\n    background-color: currentColor; }\n.mat-toolbar .mat-form-field-label,\n  .mat-toolbar .mat-focused .mat-form-field-label,\n  .mat-toolbar .mat-select-value,\n  .mat-toolbar .mat-select-arrow,\n  .mat-toolbar .mat-form-field.mat-focused .mat-select-arrow {\n    color: inherit; }\n.mat-toolbar .mat-input-element {\n    caret-color: currentColor; }\n.mat-tooltip {\n  background: rgba(97, 97, 97, 0.9); }\n.mat-tree {\n  background: white; }\n.mat-tree-node,\n.mat-nested-tree-node {\n  color: rgba(0, 0, 0, 0.87); }\n.mat-snack-bar-container {\n  color: rgba(255, 255, 255, 0.7);\n  background: #323232;\n  box-shadow: 0px 3px 5px -1px rgba(0, 0, 0, 0.2), 0px 6px 10px 0px rgba(0, 0, 0, 0.14), 0px 1px 18px 0px rgba(0, 0, 0, 0.12); }\n.mat-simple-snackbar-action {\n  color: #81c6fa; }\n.fixed-header {\n  position: fixed;\n  top: 0;\n  left: 0;\n  z-index: 2;\n  width: 100%  !important; }\nbody {\n  padding-top: 3%; }\ntextarea {\n  color: Gray;\n  border: none;\n  background-color: #f0f6fc;\n  overflow-wrap: break-word;\n  width: 100%; }\n.test-title {\n  color: black;\n  font-weight: 500; }\n.MatButtonEmulation {\n  font: caption;\n  display: inline-block;\n  margin: 0.05vh;\n  white-space: normal;\n  word-wrap: break-word;\n  border-radius: 2px;\n  text-align: center;\n  color: white;\n  background: #2a7bb8;\n  width: 70px;\n  padding: 0.5vw;\n  height: auto; }\n.badge {\n  position: relative;\n  text-align: center;\n  background: #ff4081;\n  width: 1vmax;\n  height: 1vmax;\n  line-height: 1vmax;\n  border-radius: 50%; }\n.vert-align {\n  display: inline-block;\n  vertical-align: middle;\n  line-height: normal;\n  color: white;\n  font-size: 0.6vmax; }\n\n/*# sourceMappingURL=data:application/json;base64,  */',"",""]]},"./node_modules/style-loader/lib/addStyles.js":function(b,t,g){var I,c,G,C={},m=(I=function(){return window&&document&&document.all&&!window.atob},function(){return void 0===c&&(c=I.apply(this,arguments)),c}),W=(G={},function(b,t){if("function"==typeof b)return b();if(void 0===G[b]){var g=function(b,t){return t?t.querySelector(b):document.querySelector(b)}.call(this,b,t);if(window.HTMLIFrameElement&&g instanceof window.HTMLIFrameElement)try{g=g.contentDocument.head}catch(b){g=null}G[b]=g}return G[b]}),Z=null,l=0,d=[],a=g("./node_modules/style-loader/lib/urls.js");function X(b,t){for(var g=0;g<b.length;g++){var I=b[g],c=C[I.id];if(c){c.refs++;for(var G=0;G<c.parts.length;G++)c.parts[G](I.parts[G]);for(;G<I.parts.length;G++)c.parts.push(n(I.parts[G],t))}else{var m=[];for(G=0;G<I.parts.length;G++)m.push(n(I.parts[G],t));C[I.id]={id:I.id,refs:1,parts:m}}}}function i(b,t){for(var g=[],I={},c=0;c<b.length;c++){var G=b[c],C=t.base?G[0]+t.base:G[0],m={css:G[1],media:G[2],sourceMap:G[3]};I[C]?I[C].parts.push(m):g.push(I[C]={id:C,parts:[m]})}return g}function Y(b,t){var g=W(b.insertInto);if(!g)throw new Error("Couldn't find a style target. This probably means that the value for the 'insertInto' parameter is invalid.");var I=d[d.length-1];if("top"===b.insertAt)I?I.nextSibling?g.insertBefore(t,I.nextSibling):g.appendChild(t):g.insertBefore(t,g.firstChild),d.push(t);else if("bottom"===b.insertAt)g.appendChild(t);else{if("object"!=typeof b.insertAt||!b.insertAt.before)throw new Error("[Style Loader]\n\n Invalid value for parameter 'insertAt' ('options.insertAt') found.\n Must be 'top', 'bottom', or Object.\n (https://github.com/webpack-contrib/style-loader#insertat)\n");var c=W(b.insertAt.before,g);g.insertBefore(t,c)}}function A(b){if(null===b.parentNode)return!1;b.parentNode.removeChild(b);var t=d.indexOf(b);t>=0&&d.splice(t,1)}function u(b){var t=document.createElement("style");if(void 0===b.attrs.type&&(b.attrs.type="text/css"),void 0===b.attrs.nonce){var I=function(){0;return g.nc}();I&&(b.attrs.nonce=I)}return L(t,b.attrs),Y(b,t),t}function L(b,t){Object.keys(t).forEach(function(g){b.setAttribute(g,t[g])})}function n(b,t){var g,I,c,G,C,m;if(t.transform&&b.css){if(!(G="function"==typeof t.transform?t.transform(b.css):t.transform.default(b.css)))return function(){};b.css=G}if(t.singleton){var W=l++;g=Z||(Z=u(t)),I=F.bind(null,g,W,!1),c=F.bind(null,g,W,!0)}else b.sourceMap&&"function"==typeof URL&&"function"==typeof URL.createObjectURL&&"function"==typeof URL.revokeObjectURL&&"function"==typeof Blob&&"function"==typeof btoa?(C=t,m=document.createElement("link"),void 0===C.attrs.type&&(C.attrs.type="text/css"),C.attrs.rel="stylesheet",L(m,C.attrs),Y(C,m),I=function(b,t,g){var I=g.css,c=g.sourceMap,G=void 0===t.convertToAbsoluteUrls&&c;(t.convertToAbsoluteUrls||G)&&(I=a(I));c&&(I+="\n/*# sourceMappingURL=data:application/json;base64,"+btoa(unescape(encodeURIComponent(JSON.stringify(c))))+" */");var C=new Blob([I],{type:"text/css"}),m=b.href;b.href=URL.createObjectURL(C),m&&URL.revokeObjectURL(m)}.bind(null,g=m,t),c=function(){A(g),g.href&&URL.revokeObjectURL(g.href)}):(g=u(t),I=function(b,t){var g=t.css,I=t.media;I&&b.setAttribute("media",I);if(b.styleSheet)b.styleSheet.cssText=g;else{for(;b.firstChild;)b.removeChild(b.firstChild);b.appendChild(document.createTextNode(g))}}.bind(null,g),c=function(){A(g)});return I(b),function(t){if(t){if(t.css===b.css&&t.media===b.media&&t.sourceMap===b.sourceMap)return;I(b=t)}else c()}}b.exports=function(b,t){if("undefined"!=typeof DEBUG&&DEBUG&&"object"!=typeof document)throw new Error("The style-loader cannot be used in a non-browser environment");(t=t||{}).attrs="object"==typeof t.attrs?t.attrs:{},t.singleton||"boolean"==typeof t.singleton||(t.singleton=m()),t.insertInto||(t.insertInto="head"),t.insertAt||(t.insertAt="bottom");var g=i(b,t);return X(g,t),function(b){for(var I=[],c=0;c<g.length;c++){var G=g[c];(m=C[G.id]).refs--,I.push(m)}b&&X(i(b,t),t);for(c=0;c<I.length;c++){var m;if(0===(m=I[c]).refs){for(var W=0;W<m.parts.length;W++)m.parts[W]();delete C[m.id]}}}};var B,h=(B=[],function(b,t){return B[b]=t,B.filter(Boolean).join("\n")});function F(b,t,g,I){var c=g?"":I.css;if(b.styleSheet)b.styleSheet.cssText=h(t,c);else{var G=document.createTextNode(c),C=b.childNodes;C[t]&&b.removeChild(C[t]),C.length?b.insertBefore(G,C[t]):b.appendChild(G)}}},"./node_modules/style-loader/lib/urls.js":function(b,t){b.exports=function(b){var t="undefined"!=typeof window&&window.location;if(!t)throw new Error("fixUrls requires window.location");if(!b||"string"!=typeof b)return b;var g=t.protocol+"//"+t.host,I=g+t.pathname.replace(/\/[^\/]*$/,"/");return b.replace(/url\s*\(((?:[^)(]|\((?:[^)(]+|\([^)(]*\))*\))*)\)/gi,function(b,t){var c,G=t.trim().replace(/^"(.*)"$/,function(b,t){return t}).replace(/^'(.*)'$/,function(b,t){return t});return/^(#|data:|http:\/\/|https:\/\/|file:\/\/\/|\s*$)/i.test(G)?b:(c=0===G.indexOf("//")?G:0===G.indexOf("/")?g+G:I+G.replace(/^\.\//,""),"url("+JSON.stringify(c)+")")})}},"./src/styles.scss":function(b,t,g){var I=g("./node_modules/@angular-devkit/build-angular/src/angular-cli-files/plugins/raw-css-loader.js!./node_modules/postcss-loader/src/index.js?!./node_modules/sass-loader/lib/loader.js?!./src/styles.scss");"string"==typeof I&&(I=[[b.i,I,""]]);var c={hmr:!0};c.transform=void 0,c.insertInto=void 0;g("./node_modules/style-loader/lib/addStyles.js")(I,c);I.locals&&(b.exports=I.locals)},5:function(b,t,g){b.exports=g("./src/styles.scss")}},[[5,"runtime"]]]);
</script><script type="text/javascript">
(window.webpackJsonp=window.webpackJsonp||[]).push([["vendor"],{"./node_modules/@angular/animations/fesm5/animations.js":function(e,t,n){"use strict";n.r(t),n.d(t,"AnimationBuilder",function(){return r}),n.d(t,"AnimationFactory",function(){return i}),n.d(t,"AUTO_STYLE",function(){return o}),n.d(t,"animate",function(){return s}),n.d(t,"animateChild",function(){return m}),n.d(t,"animation",function(){return h}),n.d(t,"group",function(){return u}),n.d(t,"keyframes",function(){return p}),n.d(t,"query",function(){return v}),n.d(t,"sequence",function(){return c}),n.d(t,"stagger",function(){return _}),n.d(t,"state",function(){return d}),n.d(t,"style",function(){return l}),n.d(t,"transition",function(){return f}),n.d(t,"trigger",function(){return a}),n.d(t,"useAnimation",function(){return g}),n.d(t,"NoopAnimationPlayer",function(){return b}),n.d(t,"ɵPRE_STYLE",function(){return x}),n.d(t,"ɵAnimationGroupPlayer",function(){return j});var r=function(){return function(){}}(),i=function(){return function(){}}(),o="*";function a(e,t){return{type:7,name:e,definitions:t,options:{}}}function s(e,t){return void 0===t&&(t=null),{type:4,styles:t,timings:e}}function u(e,t){return void 0===t&&(t=null),{type:3,steps:e,options:t}}function c(e,t){return void 0===t&&(t=null),{type:2,steps:e,options:t}}function l(e){return{type:6,styles:e,offset:null}}function d(e,t,n){return{type:0,name:e,styles:t,options:n}}function p(e){return{type:5,steps:e}}function f(e,t,n){return void 0===n&&(n=null),{type:1,expr:e,animation:t,options:n}}function h(e,t){return void 0===t&&(t=null),{type:8,animation:e,options:t}}function m(e){return void 0===e&&(e=null),{type:9,options:e}}function g(e,t){return void 0===t&&(t=null),{type:10,animation:e,options:t}}function v(e,t,n){return void 0===n&&(n=null),{type:11,selector:e,animation:t,options:n}}function _(e,t){return{type:12,timings:e,animation:t}}function y(e){Promise.resolve(null).then(e)}var b=function(){function e(e,t){void 0===e&&(e=0),void 0===t&&(t=0),this._onDoneFns=[],this._onStartFns=[],this._onDestroyFns=[],this._started=!1,this._destroyed=!1,this._finished=!1,this.parentPlayer=null,this.totalTime=e+t}return e.prototype._onFinish=function(){this._finished||(this._finished=!0,this._onDoneFns.forEach(function(e){return e()}),this._onDoneFns=[])},e.prototype.onStart=function(e){this._onStartFns.push(e)},e.prototype.onDone=function(e){this._onDoneFns.push(e)},e.prototype.onDestroy=function(e){this._onDestroyFns.push(e)},e.prototype.hasStarted=function(){return this._started},e.prototype.init=function(){},e.prototype.play=function(){this.hasStarted()||(this._onStart(),this.triggerMicrotask()),this._started=!0},e.prototype.triggerMicrotask=function(){var e=this;y(function(){return e._onFinish()})},e.prototype._onStart=function(){this._onStartFns.forEach(function(e){return e()}),this._onStartFns=[]},e.prototype.pause=function(){},e.prototype.restart=function(){},e.prototype.finish=function(){this._onFinish()},e.prototype.destroy=function(){this._destroyed||(this._destroyed=!0,this.hasStarted()||this._onStart(),this.finish(),this._onDestroyFns.forEach(function(e){return e()}),this._onDestroyFns=[])},e.prototype.reset=function(){},e.prototype.setPosition=function(e){},e.prototype.getPosition=function(){return 0},e.prototype.triggerCallback=function(e){var t="start"==e?this._onStartFns:this._onDoneFns;t.forEach(function(e){return e()}),t.length=0},e}(),j=function(){function e(e){var t=this;this._onDoneFns=[],this._onStartFns=[],this._finished=!1,this._started=!1,this._destroyed=!1,this._onDestroyFns=[],this.parentPlayer=null,this.totalTime=0,this.players=e;var n=0,r=0,i=0,o=this.players.length;0==o?y(function(){return t._onFinish()}):this.players.forEach(function(e){e.onDone(function(){++n==o&&t._onFinish()}),e.onDestroy(function(){++r==o&&t._onDestroy()}),e.onStart(function(){++i==o&&t._onStart()})}),this.totalTime=this.players.reduce(function(e,t){return Math.max(e,t.totalTime)},0)}return e.prototype._onFinish=function(){this._finished||(this._finished=!0,this._onDoneFns.forEach(function(e){return e()}),this._onDoneFns=[])},e.prototype.init=function(){this.players.forEach(function(e){return e.init()})},e.prototype.onStart=function(e){this._onStartFns.push(e)},e.prototype._onStart=function(){this.hasStarted()||(this._started=!0,this._onStartFns.forEach(function(e){return e()}),this._onStartFns=[])},e.prototype.onDone=function(e){this._onDoneFns.push(e)},e.prototype.onDestroy=function(e){this._onDestroyFns.push(e)},e.prototype.hasStarted=function(){return this._started},e.prototype.play=function(){this.parentPlayer||this.init(),this._onStart(),this.players.forEach(function(e){return e.play()})},e.prototype.pause=function(){this.players.forEach(function(e){return e.pause()})},e.prototype.restart=function(){this.players.forEach(function(e){return e.restart()})},e.prototype.finish=function(){this._onFinish(),this.players.forEach(function(e){return e.finish()})},e.prototype.destroy=function(){this._onDestroy()},e.prototype._onDestroy=function(){this._destroyed||(this._destroyed=!0,this._onFinish(),this.players.forEach(function(e){return e.destroy()}),this._onDestroyFns.forEach(function(e){return e()}),this._onDestroyFns=[])},e.prototype.reset=function(){this.players.forEach(function(e){return e.reset()}),this._destroyed=!1,this._finished=!1,this._started=!1},e.prototype.setPosition=function(e){var t=e*this.totalTime;this.players.forEach(function(e){var n=e.totalTime?Math.min(1,t/e.totalTime):1;e.setPosition(n)})},e.prototype.getPosition=function(){var e=0;return this.players.forEach(function(t){var n=t.getPosition();e=Math.min(n,e)}),e},e.prototype.beforeDestroy=function(){this.players.forEach(function(e){e.beforeDestroy&&e.beforeDestroy()})},e.prototype.triggerCallback=function(e){var t="start"==e?this._onStartFns:this._onDoneFns;t.forEach(function(e){return e()}),t.length=0},e}(),x="!"},"./node_modules/@angular/animations/fesm5/browser.js":function(e,t,n){"use strict";n.r(t),n.d(t,"AnimationDriver",function(){return k}),n.d(t,"ɵAnimation",function(){return Se}),n.d(t,"ɵAnimationStyleNormalizer",function(){return ke}),n.d(t,"ɵNoopAnimationStyleNormalizer",function(){return Me}),n.d(t,"ɵWebAnimationsStyleNormalizer",function(){return Ae}),n.d(t,"ɵAnimationDriver",function(){return k}),n.d(t,"ɵNoopAnimationDriver",function(){return S}),n.d(t,"ɵAnimationEngine",function(){return st}),n.d(t,"ɵCssKeyframesDriver",function(){return bt}),n.d(t,"ɵCssKeyframesPlayer",function(){return _t}),n.d(t,"ɵcontainsElement",function(){return w}),n.d(t,"ɵinvokeQuery",function(){return C}),n.d(t,"ɵmatchesElement",function(){return O}),n.d(t,"ɵvalidateStyleProperty",function(){return x}),n.d(t,"ɵWebAnimationsDriver",function(){return xt}),n.d(t,"ɵsupportsWebAnimations",function(){return Ot}),n.d(t,"ɵWebAnimationsPlayer",function(){return jt}),n.d(t,"ɵallowPreviousPlayerStylesMerge",function(){return Z});var r=n("./node_modules/@angular/animations/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/animations/fesm5/animations.js"),o=n("./node_modules/@angular/core/fesm5/core.js");function a(){return"undefined"!=typeof process}function s(e){switch(e.length){case 0:return new i.NoopAnimationPlayer;case 1:return e[0];default:return new i["ɵAnimationGroupPlayer"](e)}}function u(e,t,n,r,o,a){void 0===o&&(o={}),void 0===a&&(a={});var s=[],u=[],c=-1,l=null;if(r.forEach(function(e){var n=e.offset,r=n==c,d=r&&l||{};Object.keys(e).forEach(function(n){var r=n,u=e[n];if("offset"!==n)switch(r=t.normalizePropertyName(r,s),u){case i["ɵPRE_STYLE"]:u=o[n];break;case i.AUTO_STYLE:u=a[n];break;default:u=t.normalizeStyleValue(n,r,u,s)}d[r]=u}),r||u.push(d),l=d,c=n}),s.length){throw new Error("Unable to animate due to the following errors:\n - "+s.join("\n - "))}return u}function c(e,t,n,r){switch(t){case"start":e.onStart(function(){return r(n&&l(n,"start",e))});break;case"done":e.onDone(function(){return r(n&&l(n,"done",e))});break;case"destroy":e.onDestroy(function(){return r(n&&l(n,"destroy",e))})}}function l(e,t,n){var r=n.totalTime,i=!!n.disabled,o=d(e.element,e.triggerName,e.fromState,e.toState,t||e.phaseName,void 0==r?e.totalTime:r,i),a=e._data;return null!=a&&(o._data=a),o}function d(e,t,n,r,i,o,a){return void 0===i&&(i=""),void 0===o&&(o=0),{element:e,triggerName:t,fromState:n,toState:r,phaseName:i,totalTime:o,disabled:!!a}}function p(e,t,n){var r;return e instanceof Map?(r=e.get(t))||e.set(t,r=n):(r=e[t])||(r=e[t]=n),r}function f(e){var t=e.indexOf(":");return[e.substring(1,t),e.substr(t+1)]}var h=function(e,t){return!1},m=function(e,t){return!1},g=function(e,t,n){return[]},v=a();if(v||"undefined"!=typeof Element){if(h=function(e,t){return e.contains(t)},v||Element.prototype.matches)m=function(e,t){return e.matches(t)};else{var _=Element.prototype,y=_.matchesSelector||_.mozMatchesSelector||_.msMatchesSelector||_.oMatchesSelector||_.webkitMatchesSelector;y&&(m=function(e,t){return y.apply(e,[t])})}g=function(e,t,n){var i=[];if(n)i.push.apply(i,Object(r.__spread)(e.querySelectorAll(t)));else{var o=e.querySelector(t);o&&i.push(o)}return i}}var b=null,j=!1;function x(e){b||(b=function(){if("undefined"!=typeof document)return document.body;return null}()||{},j=!!b.style&&"WebkitAppearance"in b.style);var t=!0;b.style&&"ebkit"!=e.substring(1,6)&&(!(t=e in b.style)&&j&&(t="Webkit"+e.charAt(0).toUpperCase()+e.substr(1)in b.style));return t}var O=m,w=h,C=g;function E(e){var t={};return Object.keys(e).forEach(function(n){var r=n.replace(/([a-z])([A-Z])/g,"$1-$2");t[r]=e[n]}),t}var S=function(){function e(){}return e.prototype.validateStyleProperty=function(e){return x(e)},e.prototype.matchesElement=function(e,t){return O(e,t)},e.prototype.containsElement=function(e,t){return w(e,t)},e.prototype.query=function(e,t,n){return C(e,t,n)},e.prototype.computeStyle=function(e,t,n){return n||""},e.prototype.animate=function(e,t,n,r,o,a,s){return void 0===a&&(a=[]),new i.NoopAnimationPlayer(n,r)},e=Object(r.__decorate)([Object(o.Injectable)()],e)}(),k=function(){function e(){}return e.NOOP=new S,e}(),M=1e3,A="ng-leave",T="ng-trigger",D=".ng-trigger",I=".ng-animating";function P(e){if("number"==typeof e)return e;var t=e.match(/^(-?[\.\d]+)(m?s)/);return!t||t.length<2?0:R(parseFloat(t[1]),t[2])}function R(e,t){switch(t){case"s":return e*M;default:return e}}function N(e,t,n){return e.hasOwnProperty("duration")?e:function(e,t,n){var r,i=0,o="";if("string"==typeof e){var a=e.match(/^(-?[\.\d]+)(m?s)(?:\s+(-?[\.\d]+)(m?s))?(?:\s+([-a-z]+(?:\(.+?\))?))?$/i);if(null===a)return t.push('The provided timing value "'+e+'" is invalid.'),{duration:0,delay:0,easing:""};r=R(parseFloat(a[1]),a[2]);var s=a[3];null!=s&&(i=R(parseFloat(s),a[4]));var u=a[5];u&&(o=u)}else r=e;if(!n){var c=!1,l=t.length;r<0&&(t.push("Duration values below 0 are not allowed for this animation step."),c=!0),i<0&&(t.push("Delay values below 0 are not allowed for this animation step."),c=!0),c&&t.splice(l,0,'The provided timing value "'+e+'" is invalid.')}return{duration:r,delay:i,easing:o}}(e,t,n)}function F(e,t){return void 0===t&&(t={}),Object.keys(e).forEach(function(n){t[n]=e[n]}),t}function L(e){var t={};return Array.isArray(e)?e.forEach(function(e){return V(e,!1,t)}):V(e,!1,t),t}function V(e,t,n){if(void 0===n&&(n={}),t)for(var r in e)n[r]=e[r];else F(e,n);return n}function B(e,t,n){return n?t+":"+n+";":""}function z(e){for(var t="",n=0;n<e.style.length;n++){t+=B(0,r=e.style.item(n),e.style.getPropertyValue(r))}for(var r in e.style){if(e.style.hasOwnProperty(r)&&!r.startsWith("_"))t+=B(0,r.replace(/([a-z])([A-Z])/g,"$1-$2").toLowerCase(),e.style[r])}e.setAttribute("style",t)}function U(e,t){e.style&&(Object.keys(t).forEach(function(n){var r=X(n);e.style[r]=t[n]}),a()&&z(e))}function H(e,t){e.style&&(Object.keys(t).forEach(function(t){var n=X(t);e.style[n]=""}),a()&&z(e))}function q(e){return Array.isArray(e)?1==e.length?e[0]:Object(i.sequence)(e):e}var G=new RegExp("{{\\s*(.+?)\\s*}}","g");function W(e){var t=[];if("string"==typeof e){for(var n=e.toString(),r=void 0;r=G.exec(n);)t.push(r[1]);G.lastIndex=0}return t}function Y(e,t,n){var r=e.toString(),i=r.replace(G,function(e,r){var i=t[r];return t.hasOwnProperty(r)||(n.push("Please provide a value for the animation param "+r),i=""),i.toString()});return i==r?e:i}function K(e){for(var t=[],n=e.next();!n.done;)t.push(n.value),n=e.next();return t}var Q=/-+([a-z0-9])/g;function X(e){return e.replace(Q,function(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];return e[1].toUpperCase()})}function Z(e,t){return 0===e||0===t}function $(e,t,n){var r=Object.keys(n);if(r.length&&t.length){var i=t[0],o=[];if(r.forEach(function(e){i.hasOwnProperty(e)||o.push(e),i[e]=n[e]}),o.length)for(var a=function(){var n=t[s];o.forEach(function(t){n[t]=ee(e,t)})},s=1;s<t.length;s++)a()}return t}function J(e,t,n){switch(t.type){case 7:return e.visitTrigger(t,n);case 0:return e.visitState(t,n);case 1:return e.visitTransition(t,n);case 2:return e.visitSequence(t,n);case 3:return e.visitGroup(t,n);case 4:return e.visitAnimate(t,n);case 5:return e.visitKeyframes(t,n);case 6:return e.visitStyle(t,n);case 8:return e.visitReference(t,n);case 9:return e.visitAnimateChild(t,n);case 10:return e.visitAnimateRef(t,n);case 11:return e.visitQuery(t,n);case 12:return e.visitStagger(t,n);default:throw new Error("Unable to resolve animation metadata node #"+t.type)}}function ee(e,t){return window.getComputedStyle(e)[t]}var te="*";function ne(e,t){var n=[];return"string"==typeof e?e.split(/\s*,\s*/).forEach(function(e){return function(e,t,n){if(":"==e[0]){var r=function(e,t){switch(e){case":enter":return"void => *";case":leave":return"* => void";case":increment":return function(e,t){return parseFloat(t)>parseFloat(e)};case":decrement":return function(e,t){return parseFloat(t)<parseFloat(e)};default:return t.push('The transition alias value "'+e+'" is not supported'),"* => *"}}(e,n);if("function"==typeof r)return void t.push(r);e=r}var i=e.match(/^(\*|[-\w]+)\s*(<?[=-]>)\s*(\*|[-\w]+)$/);if(null==i||i.length<4)return n.push('The provided transition expression "'+e+'" is not supported'),t;var o=i[1],a=i[2],s=i[3];t.push(oe(o,s));var u=o==te&&s==te;"<"!=a[0]||u||t.push(oe(s,o))}(e,n,t)}):n.push(e),n}var re=new Set(["true","1"]),ie=new Set(["false","0"]);function oe(e,t){var n=re.has(e)||ie.has(e),r=re.has(t)||ie.has(t);return function(i,o){var a=e==te||e==i,s=t==te||t==o;return!a&&n&&"boolean"==typeof i&&(a=i?re.has(e):ie.has(e)),!s&&r&&"boolean"==typeof o&&(s=o?re.has(t):ie.has(t)),a&&s}}var ae=":self",se=new RegExp("s*"+ae+"s*,?","g");function ue(e,t,n){return new ce(e).build(t,n)}var ce=function(){function e(e){this._driver=e}return e.prototype.build=function(e,t){var n=new le(t);return this._resetContextStyleTimingState(n),J(this,q(e),n)},e.prototype._resetContextStyleTimingState=function(e){e.currentQuerySelector="",e.collectedStyles={},e.collectedStyles[""]={},e.currentTime=0},e.prototype.visitTrigger=function(e,t){var n=this,r=t.queryCount=0,i=t.depCount=0,o=[],a=[];return"@"==e.name.charAt(0)&&t.errors.push("animation triggers cannot be prefixed with an `@` sign (e.g. trigger('@foo', [...]))"),e.definitions.forEach(function(e){if(n._resetContextStyleTimingState(t),0==e.type){var s=e,u=s.name;u.toString().split(/\s*,\s*/).forEach(function(e){s.name=e,o.push(n.visitState(s,t))}),s.name=u}else if(1==e.type){var c=n.visitTransition(e,t);r+=c.queryCount,i+=c.depCount,a.push(c)}else t.errors.push("only state() and transition() definitions can sit inside of a trigger()")}),{type:7,name:e.name,states:o,transitions:a,queryCount:r,depCount:i,options:null}},e.prototype.visitState=function(e,t){var n=this.visitStyle(e.styles,t),r=e.options&&e.options.params||null;if(n.containsDynamicStyles){var i=new Set,o=r||{};if(n.styles.forEach(function(e){if(de(e)){var t=e;Object.keys(t).forEach(function(e){W(t[e]).forEach(function(e){o.hasOwnProperty(e)||i.add(e)})})}}),i.size){var a=K(i.values());t.errors.push('state("'+e.name+'", ...) must define default values for all the following style substitutions: '+a.join(", "))}}return{type:0,name:e.name,style:n,options:r?{params:r}:null}},e.prototype.visitTransition=function(e,t){t.queryCount=0,t.depCount=0;var n=J(this,q(e.animation),t);return{type:1,matchers:ne(e.expr,t.errors),animation:n,queryCount:t.queryCount,depCount:t.depCount,options:pe(e.options)}},e.prototype.visitSequence=function(e,t){var n=this;return{type:2,steps:e.steps.map(function(e){return J(n,e,t)}),options:pe(e.options)}},e.prototype.visitGroup=function(e,t){var n=this,r=t.currentTime,i=0,o=e.steps.map(function(e){t.currentTime=r;var o=J(n,e,t);return i=Math.max(i,t.currentTime),o});return t.currentTime=i,{type:3,steps:o,options:pe(e.options)}},e.prototype.visitAnimate=function(e,t){var n,r=function(e,t){var n=null;if(e.hasOwnProperty("duration"))n=e;else if("number"==typeof e){var r=N(e,t).duration;return fe(r,0,"")}var i=e;if(i.split(/\s+/).some(function(e){return"{"==e.charAt(0)&&"{"==e.charAt(1)})){var o=fe(0,0,"");return o.dynamic=!0,o.strValue=i,o}return fe((n=n||N(i,t)).duration,n.delay,n.easing)}(e.timings,t.errors);t.currentAnimateTimings=r;var o=e.styles?e.styles:Object(i.style)({});if(5==o.type)n=this.visitKeyframes(o,t);else{var a=e.styles,s=!1;if(!a){s=!0;var u={};r.easing&&(u.easing=r.easing),a=Object(i.style)(u)}t.currentTime+=r.duration+r.delay;var c=this.visitStyle(a,t);c.isEmptyStep=s,n=c}return t.currentAnimateTimings=null,{type:4,timings:r,style:n,options:null}},e.prototype.visitStyle=function(e,t){var n=this._makeStyleAst(e,t);return this._validateStyleAst(n,t),n},e.prototype._makeStyleAst=function(e,t){var n=[];Array.isArray(e.styles)?e.styles.forEach(function(e){"string"==typeof e?e==i.AUTO_STYLE?n.push(e):t.errors.push("The provided style string value "+e+" is not allowed."):n.push(e)}):n.push(e.styles);var r=!1,o=null;return n.forEach(function(e){if(de(e)){var t=e,n=t.easing;if(n&&(o=n,delete t.easing),!r)for(var i in t){if(t[i].toString().indexOf("{{")>=0){r=!0;break}}}}),{type:6,styles:n,easing:o,offset:e.offset,containsDynamicStyles:r,options:null}},e.prototype._validateStyleAst=function(e,t){var n=this,r=t.currentAnimateTimings,i=t.currentTime,o=t.currentTime;r&&o>0&&(o-=r.duration+r.delay),e.styles.forEach(function(e){"string"!=typeof e&&Object.keys(e).forEach(function(r){if(n._driver.validateStyleProperty(r)){var a,s,u,c,l,d=t.collectedStyles[t.currentQuerySelector],p=d[r],f=!0;p&&(o!=i&&o>=p.startTime&&i<=p.endTime&&(t.errors.push('The CSS property "'+r+'" that exists between the times of "'+p.startTime+'ms" and "'+p.endTime+'ms" is also being animated in a parallel animation between the times of "'+o+'ms" and "'+i+'ms"'),f=!1),o=p.startTime),f&&(d[r]={startTime:o,endTime:i}),t.options&&(a=e[r],s=t.options,u=t.errors,c=s.params||{},(l=W(a)).length&&l.forEach(function(e){c.hasOwnProperty(e)||u.push("Unable to resolve the local animation param "+e+" in the given list of values")}))}else t.errors.push('The provided animation property "'+r+'" is not a supported CSS property for animations')})})},e.prototype.visitKeyframes=function(e,t){var n=this,r={type:5,styles:[],options:null};if(!t.currentAnimateTimings)return t.errors.push("keyframes() must be placed inside of a call to animate()"),r;var i=0,o=[],a=!1,s=!1,u=0,c=e.steps.map(function(e){var r=n._makeStyleAst(e,t),c=null!=r.offset?r.offset:function(e){if("string"==typeof e)return null;var t=null;if(Array.isArray(e))e.forEach(function(e){if(de(e)&&e.hasOwnProperty("offset")){var n=e;t=parseFloat(n.offset),delete n.offset}});else if(de(e)&&e.hasOwnProperty("offset")){var n=e;t=parseFloat(n.offset),delete n.offset}return t}(r.styles),l=0;return null!=c&&(i++,l=r.offset=c),s=s||l<0||l>1,a=a||l<u,u=l,o.push(l),r});s&&t.errors.push("Please ensure that all keyframe offsets are between 0 and 1"),a&&t.errors.push("Please ensure that all keyframe offsets are in order");var l=e.steps.length,d=0;i>0&&i<l?t.errors.push("Not all style() steps within the declared keyframes() contain offsets"):0==i&&(d=1/(l-1));var p=l-1,f=t.currentTime,h=t.currentAnimateTimings,m=h.duration;return c.forEach(function(e,i){var a=d>0?i==p?1:d*i:o[i],s=a*m;t.currentTime=f+h.delay+s,h.duration=s,n._validateStyleAst(e,t),e.offset=a,r.styles.push(e)}),r},e.prototype.visitReference=function(e,t){return{type:8,animation:J(this,q(e.animation),t),options:pe(e.options)}},e.prototype.visitAnimateChild=function(e,t){return t.depCount++,{type:9,options:pe(e.options)}},e.prototype.visitAnimateRef=function(e,t){return{type:10,animation:this.visitReference(e.animation,t),options:pe(e.options)}},e.prototype.visitQuery=function(e,t){var n=t.currentQuerySelector,i=e.options||{};t.queryCount++,t.currentQuery=e;var o=Object(r.__read)(function(e){var t=!!e.split(/\s*,\s*/).find(function(e){return e==ae});t&&(e=e.replace(se,""));return[e=e.replace(/@\*/g,D).replace(/@\w+/g,function(e){return D+"-"+e.substr(1)}).replace(/:animating/g,I),t]}(e.selector),2),a=o[0],s=o[1];t.currentQuerySelector=n.length?n+" "+a:a,p(t.collectedStyles,t.currentQuerySelector,{});var u=J(this,q(e.animation),t);return t.currentQuery=null,t.currentQuerySelector=n,{type:11,selector:a,limit:i.limit||0,optional:!!i.optional,includeSelf:s,animation:u,originalSelector:e.selector,options:pe(e.options)}},e.prototype.visitStagger=function(e,t){t.currentQuery||t.errors.push("stagger() can only be used inside of query()");var n="full"===e.timings?{duration:0,delay:0,easing:"full"}:N(e.timings,t.errors,!0);return{type:12,animation:J(this,q(e.animation),t),timings:n,options:null}},e}();var le=function(){return function(e){this.errors=e,this.queryCount=0,this.depCount=0,this.currentTransition=null,this.currentQuery=null,this.currentQuerySelector=null,this.currentAnimateTimings=null,this.currentTime=0,this.collectedStyles={},this.options=null}}();function de(e){return!Array.isArray(e)&&"object"==typeof e}function pe(e){var t;return e?(e=F(e)).params&&(e.params=(t=e.params)?F(t):null):e={},e}function fe(e,t,n){return{duration:e,delay:t,easing:n}}function he(e,t,n,r,i,o,a,s){return void 0===a&&(a=null),void 0===s&&(s=!1),{type:1,element:e,keyframes:t,preStyleProps:n,postStyleProps:r,duration:i,delay:o,totalTime:i+o,easing:a,subTimeline:s}}var me=function(){function e(){this._map=new Map}return e.prototype.consume=function(e){var t=this._map.get(e);return t?this._map.delete(e):t=[],t},e.prototype.append=function(e,t){var n=this._map.get(e);n||this._map.set(e,n=[]),n.push.apply(n,Object(r.__spread)(t))},e.prototype.has=function(e){return this._map.has(e)},e.prototype.clear=function(){this._map.clear()},e}(),ge=new RegExp(":enter","g"),ve=new RegExp(":leave","g");function _e(e,t,n,r,i,o,a,s,u,c){return void 0===o&&(o={}),void 0===a&&(a={}),void 0===c&&(c=[]),(new ye).buildKeyframes(e,t,n,r,i,o,a,s,u,c)}var ye=function(){function e(){}return e.prototype.buildKeyframes=function(e,t,n,r,i,o,a,s,u,c){void 0===c&&(c=[]),u=u||new me;var l=new je(e,t,u,r,i,c,[]);l.options=s,l.currentTimeline.setStyles([o],null,l.errors,s),J(this,n,l);var d=l.timelines.filter(function(e){return e.containsAnimation()});if(d.length&&Object.keys(a).length){var p=d[d.length-1];p.allowOnlyTimelineStyles()||p.setStyles([a],null,l.errors,s)}return d.length?d.map(function(e){return e.buildKeyframes()}):[he(t,[],[],[],0,0,"",!1)]},e.prototype.visitTrigger=function(e,t){},e.prototype.visitState=function(e,t){},e.prototype.visitTransition=function(e,t){},e.prototype.visitAnimateChild=function(e,t){var n=t.subInstructions.consume(t.element);if(n){var r=t.createSubContext(e.options),i=t.currentTimeline.currentTime,o=this._visitSubInstructions(n,r,r.options);i!=o&&t.transformIntoNewTimeline(o)}t.previousNode=e},e.prototype.visitAnimateRef=function(e,t){var n=t.createSubContext(e.options);n.transformIntoNewTimeline(),this.visitReference(e.animation,n),t.transformIntoNewTimeline(n.currentTimeline.currentTime),t.previousNode=e},e.prototype._visitSubInstructions=function(e,t,n){var r=t.currentTimeline.currentTime,i=null!=n.duration?P(n.duration):null,o=null!=n.delay?P(n.delay):null;return 0!==i&&e.forEach(function(e){var n=t.appendInstructionToTimeline(e,i,o);r=Math.max(r,n.duration+n.delay)}),r},e.prototype.visitReference=function(e,t){t.updateOptions(e.options,!0),J(this,e.animation,t),t.previousNode=e},e.prototype.visitSequence=function(e,t){var n=this,r=t.subContextCount,i=t,o=e.options;if(o&&(o.params||o.delay)&&((i=t.createSubContext(o)).transformIntoNewTimeline(),null!=o.delay)){6==i.previousNode.type&&(i.currentTimeline.snapshotCurrentStyles(),i.previousNode=be);var a=P(o.delay);i.delayNextStep(a)}e.steps.length&&(e.steps.forEach(function(e){return J(n,e,i)}),i.currentTimeline.applyStylesToKeyframe(),i.subContextCount>r&&i.transformIntoNewTimeline()),t.previousNode=e},e.prototype.visitGroup=function(e,t){var n=this,r=[],i=t.currentTimeline.currentTime,o=e.options&&e.options.delay?P(e.options.delay):0;e.steps.forEach(function(a){var s=t.createSubContext(e.options);o&&s.delayNextStep(o),J(n,a,s),i=Math.max(i,s.currentTimeline.currentTime),r.push(s.currentTimeline)}),r.forEach(function(e){return t.currentTimeline.mergeTimelineCollectedStyles(e)}),t.transformIntoNewTimeline(i),t.previousNode=e},e.prototype._visitTiming=function(e,t){if(e.dynamic){var n=e.strValue;return N(t.params?Y(n,t.params,t.errors):n,t.errors)}return{duration:e.duration,delay:e.delay,easing:e.easing}},e.prototype.visitAnimate=function(e,t){var n=t.currentAnimateTimings=this._visitTiming(e.timings,t),r=t.currentTimeline;n.delay&&(t.incrementTime(n.delay),r.snapshotCurrentStyles());var i=e.style;5==i.type?this.visitKeyframes(i,t):(t.incrementTime(n.duration),this.visitStyle(i,t),r.applyStylesToKeyframe()),t.currentAnimateTimings=null,t.previousNode=e},e.prototype.visitStyle=function(e,t){var n=t.currentTimeline,r=t.currentAnimateTimings;!r&&n.getCurrentStyleProperties().length&&n.forwardFrame();var i=r&&r.easing||e.easing;e.isEmptyStep?n.applyEmptyStep(i):n.setStyles(e.styles,i,t.errors,t.options),t.previousNode=e},e.prototype.visitKeyframes=function(e,t){var n=t.currentAnimateTimings,r=t.currentTimeline.duration,i=n.duration,o=t.createSubContext().currentTimeline;o.easing=n.easing,e.styles.forEach(function(e){var n=e.offset||0;o.forwardTime(n*i),o.setStyles(e.styles,e.easing,t.errors,t.options),o.applyStylesToKeyframe()}),t.currentTimeline.mergeTimelineCollectedStyles(o),t.transformIntoNewTimeline(r+i),t.previousNode=e},e.prototype.visitQuery=function(e,t){var n=this,r=t.currentTimeline.currentTime,i=e.options||{},o=i.delay?P(i.delay):0;o&&(6===t.previousNode.type||0==r&&t.currentTimeline.getCurrentStyleProperties().length)&&(t.currentTimeline.snapshotCurrentStyles(),t.previousNode=be);var a=r,s=t.invokeQuery(e.selector,e.originalSelector,e.limit,e.includeSelf,!!i.optional,t.errors);t.currentQueryTotal=s.length;var u=null;s.forEach(function(r,i){t.currentQueryIndex=i;var s=t.createSubContext(e.options,r);o&&s.delayNextStep(o),r===t.element&&(u=s.currentTimeline),J(n,e.animation,s),s.currentTimeline.applyStylesToKeyframe();var c=s.currentTimeline.currentTime;a=Math.max(a,c)}),t.currentQueryIndex=0,t.currentQueryTotal=0,t.transformIntoNewTimeline(a),u&&(t.currentTimeline.mergeTimelineCollectedStyles(u),t.currentTimeline.snapshotCurrentStyles()),t.previousNode=e},e.prototype.visitStagger=function(e,t){var n=t.parentContext,r=t.currentTimeline,i=e.timings,o=Math.abs(i.duration),a=o*(t.currentQueryTotal-1),s=o*t.currentQueryIndex;switch(i.duration<0?"reverse":i.easing){case"reverse":s=a-s;break;case"full":s=n.currentStaggerTime}var u=t.currentTimeline;s&&u.delayNextStep(s);var c=u.currentTime;J(this,e.animation,t),t.previousNode=e,n.currentStaggerTime=r.currentTime-c+(r.startTime-n.currentTimeline.startTime)},e}(),be={},je=function(){function e(e,t,n,r,i,o,a,s){this._driver=e,this.element=t,this.subInstructions=n,this._enterClassName=r,this._leaveClassName=i,this.errors=o,this.timelines=a,this.parentContext=null,this.currentAnimateTimings=null,this.previousNode=be,this.subContextCount=0,this.options={},this.currentQueryIndex=0,this.currentQueryTotal=0,this.currentStaggerTime=0,this.currentTimeline=s||new xe(this._driver,t,0),a.push(this.currentTimeline)}return Object.defineProperty(e.prototype,"params",{get:function(){return this.options.params},enumerable:!0,configurable:!0}),e.prototype.updateOptions=function(e,t){var n=this;if(e){var r=e,i=this.options;null!=r.duration&&(i.duration=P(r.duration)),null!=r.delay&&(i.delay=P(r.delay));var o=r.params;if(o){var a=i.params;a||(a=this.options.params={}),Object.keys(o).forEach(function(e){t&&a.hasOwnProperty(e)||(a[e]=Y(o[e],a,n.errors))})}}},e.prototype._copyOptions=function(){var e={};if(this.options){var t=this.options.params;if(t){var n=e.params={};Object.keys(t).forEach(function(e){n[e]=t[e]})}}return e},e.prototype.createSubContext=function(t,n,r){void 0===t&&(t=null);var i=n||this.element,o=new e(this._driver,i,this.subInstructions,this._enterClassName,this._leaveClassName,this.errors,this.timelines,this.currentTimeline.fork(i,r||0));return o.previousNode=this.previousNode,o.currentAnimateTimings=this.currentAnimateTimings,o.options=this._copyOptions(),o.updateOptions(t),o.currentQueryIndex=this.currentQueryIndex,o.currentQueryTotal=this.currentQueryTotal,o.parentContext=this,this.subContextCount++,o},e.prototype.transformIntoNewTimeline=function(e){return this.previousNode=be,this.currentTimeline=this.currentTimeline.fork(this.element,e),this.timelines.push(this.currentTimeline),this.currentTimeline},e.prototype.appendInstructionToTimeline=function(e,t,n){var r={duration:null!=t?t:e.duration,delay:this.currentTimeline.currentTime+(null!=n?n:0)+e.delay,easing:""},i=new Oe(this._driver,e.element,e.keyframes,e.preStyleProps,e.postStyleProps,r,e.stretchStartingKeyframe);return this.timelines.push(i),r},e.prototype.incrementTime=function(e){this.currentTimeline.forwardTime(this.currentTimeline.duration+e)},e.prototype.delayNextStep=function(e){e>0&&this.currentTimeline.delayNextStep(e)},e.prototype.invokeQuery=function(e,t,n,i,o,a){var s=[];if(i&&s.push(this.element),e.length>0){e=(e=e.replace(ge,"."+this._enterClassName)).replace(ve,"."+this._leaveClassName);var u=1!=n,c=this._driver.query(this.element,e,u);0!==n&&(c=n<0?c.slice(c.length+n,c.length):c.slice(0,n)),s.push.apply(s,Object(r.__spread)(c))}return o||0!=s.length||a.push('`query("'+t+'")` returned zero elements. (Use `query("'+t+'", { optional: true })` if you wish to allow this.)'),s},e}(),xe=function(){function e(e,t,n,r){this._driver=e,this.element=t,this.startTime=n,this._elementTimelineStylesLookup=r,this.duration=0,this._previousKeyframe={},this._currentKeyframe={},this._keyframes=new Map,this._styleSummary={},this._pendingStyles={},this._backFill={},this._currentEmptyStepKeyframe=null,this._elementTimelineStylesLookup||(this._elementTimelineStylesLookup=new Map),this._localTimelineStyles=Object.create(this._backFill,{}),this._globalTimelineStyles=this._elementTimelineStylesLookup.get(t),this._globalTimelineStyles||(this._globalTimelineStyles=this._localTimelineStyles,this._elementTimelineStylesLookup.set(t,this._localTimelineStyles)),this._loadKeyframe()}return e.prototype.containsAnimation=function(){switch(this._keyframes.size){case 0:return!1;case 1:return this.getCurrentStyleProperties().length>0;default:return!0}},e.prototype.getCurrentStyleProperties=function(){return Object.keys(this._currentKeyframe)},Object.defineProperty(e.prototype,"currentTime",{get:function(){return this.startTime+this.duration},enumerable:!0,configurable:!0}),e.prototype.delayNextStep=function(e){var t=1==this._keyframes.size&&Object.keys(this._pendingStyles).length;this.duration||t?(this.forwardTime(this.currentTime+e),t&&this.snapshotCurrentStyles()):this.startTime+=e},e.prototype.fork=function(t,n){return this.applyStylesToKeyframe(),new e(this._driver,t,n||this.currentTime,this._elementTimelineStylesLookup)},e.prototype._loadKeyframe=function(){this._currentKeyframe&&(this._previousKeyframe=this._currentKeyframe),this._currentKeyframe=this._keyframes.get(this.duration),this._currentKeyframe||(this._currentKeyframe=Object.create(this._backFill,{}),this._keyframes.set(this.duration,this._currentKeyframe))},e.prototype.forwardFrame=function(){this.duration+=1,this._loadKeyframe()},e.prototype.forwardTime=function(e){this.applyStylesToKeyframe(),this.duration=e,this._loadKeyframe()},e.prototype._updateStyle=function(e,t){this._localTimelineStyles[e]=t,this._globalTimelineStyles[e]=t,this._styleSummary[e]={time:this.currentTime,value:t}},e.prototype.allowOnlyTimelineStyles=function(){return this._currentEmptyStepKeyframe!==this._currentKeyframe},e.prototype.applyEmptyStep=function(e){var t=this;e&&(this._previousKeyframe.easing=e),Object.keys(this._globalTimelineStyles).forEach(function(e){t._backFill[e]=t._globalTimelineStyles[e]||i.AUTO_STYLE,t._currentKeyframe[e]=i.AUTO_STYLE}),this._currentEmptyStepKeyframe=this._currentKeyframe},e.prototype.setStyles=function(e,t,n,r){var o=this;t&&(this._previousKeyframe.easing=t);var a,s,u,c,l=r&&r.params||{},d=(a=e,s=this._globalTimelineStyles,c={},a.forEach(function(e){"*"===e?(u=u||Object.keys(s)).forEach(function(e){c[e]=i.AUTO_STYLE}):V(e,!1,c)}),c);Object.keys(d).forEach(function(e){var t=Y(d[e],l,n);o._pendingStyles[e]=t,o._localTimelineStyles.hasOwnProperty(e)||(o._backFill[e]=o._globalTimelineStyles.hasOwnProperty(e)?o._globalTimelineStyles[e]:i.AUTO_STYLE),o._updateStyle(e,t)})},e.prototype.applyStylesToKeyframe=function(){var e=this,t=this._pendingStyles,n=Object.keys(t);0!=n.length&&(this._pendingStyles={},n.forEach(function(n){var r=t[n];e._currentKeyframe[n]=r}),Object.keys(this._localTimelineStyles).forEach(function(t){e._currentKeyframe.hasOwnProperty(t)||(e._currentKeyframe[t]=e._localTimelineStyles[t])}))},e.prototype.snapshotCurrentStyles=function(){var e=this;Object.keys(this._localTimelineStyles).forEach(function(t){var n=e._localTimelineStyles[t];e._pendingStyles[t]=n,e._updateStyle(t,n)})},e.prototype.getFinalKeyframe=function(){return this._keyframes.get(this.duration)},Object.defineProperty(e.prototype,"properties",{get:function(){var e=[];for(var t in this._currentKeyframe)e.push(t);return e},enumerable:!0,configurable:!0}),e.prototype.mergeTimelineCollectedStyles=function(e){var t=this;Object.keys(e._styleSummary).forEach(function(n){var r=t._styleSummary[n],i=e._styleSummary[n];(!r||i.time>r.time)&&t._updateStyle(n,i.value)})},e.prototype.buildKeyframes=function(){var e=this;this.applyStylesToKeyframe();var t=new Set,n=new Set,r=1===this._keyframes.size&&0===this.duration,o=[];this._keyframes.forEach(function(a,s){var u=V(a,!0);Object.keys(u).forEach(function(e){var r=u[e];r==i["ɵPRE_STYLE"]?t.add(e):r==i.AUTO_STYLE&&n.add(e)}),r||(u.offset=s/e.duration),o.push(u)});var a=t.size?K(t.values()):[],s=n.size?K(n.values()):[];if(r){var u=o[0],c=F(u);u.offset=0,c.offset=1,o=[u,c]}return he(this.element,o,a,s,this.duration,this.startTime,this.easing,!1)},e}(),Oe=function(e){function t(t,n,r,i,o,a,s){void 0===s&&(s=!1);var u=e.call(this,t,n,a.delay)||this;return u.element=n,u.keyframes=r,u.preStyleProps=i,u.postStyleProps=o,u._stretchStartingKeyframe=s,u.timings={duration:a.duration,delay:a.delay,easing:a.easing},u}return Object(r.__extends)(t,e),t.prototype.containsAnimation=function(){return this.keyframes.length>1},t.prototype.buildKeyframes=function(){var e=this.keyframes,t=this.timings,n=t.delay,r=t.duration,i=t.easing;if(this._stretchStartingKeyframe&&n){var o=[],a=r+n,s=n/a,u=V(e[0],!1);u.offset=0,o.push(u);var c=V(e[0],!1);c.offset=we(s),o.push(c);for(var l=e.length-1,d=1;d<=l;d++){var p=V(e[d],!1),f=n+p.offset*r;p.offset=we(f/a),o.push(p)}r=a,n=0,i="",e=o}return he(this.element,e,this.preStyleProps,this.postStyleProps,r,n,i,!0)},t}(xe);function we(e,t){void 0===t&&(t=3);var n=Math.pow(10,t-1);return Math.round(e*n)/n}var Ce,Ee,Se=function(){function e(e,t){this._driver=e;var n=[],r=ue(e,t,n);if(n.length){var i="animation validation failed:\n"+n.join("\n");throw new Error(i)}this._animationAst=r}return e.prototype.buildTimelines=function(e,t,n,r,i){var o=Array.isArray(t)?L(t):t,a=Array.isArray(n)?L(n):n,s=[];i=i||new me;var u=_e(this._driver,e,this._animationAst,"ng-enter",A,o,a,r,i,s);if(s.length){var c="animation building failed:\n"+s.join("\n");throw new Error(c)}return u},e}(),ke=function(){return function(){}}(),Me=function(){function e(){}return e.prototype.normalizePropertyName=function(e,t){return e},e.prototype.normalizeStyleValue=function(e,t,n,r){return n},e}(),Ae=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.normalizePropertyName=function(e,t){return X(e)},t.prototype.normalizeStyleValue=function(e,t,n,r){var i="",o=n.toString().trim();if(Te[t]&&0!==n&&"0"!==n)if("number"==typeof n)i="px";else{var a=n.match(/^[+-]?[\d\.]+([a-z]*)$/);a&&0==a[1].length&&r.push("Please provide a CSS unit value for "+e+":"+n)}return o+i},t}(ke),Te=(Ce="width,height,minWidth,minHeight,maxWidth,maxHeight,left,top,bottom,right,fontSize,outlineWidth,outlineOffset,paddingTop,paddingLeft,paddingBottom,paddingRight,marginTop,marginLeft,marginBottom,marginRight,borderRadius,borderWidth,borderTopWidth,borderLeftWidth,borderRightWidth,borderBottomWidth,textIndent,perspective".split(","),Ee={},Ce.forEach(function(e){return Ee[e]=!0}),Ee);function De(e,t,n,r,i,o,a,s,u,c,l,d,p){return{type:0,element:e,triggerName:t,isRemovalTransition:i,fromState:n,fromStyles:o,toState:r,toStyles:a,timelines:s,queriedElements:u,preStyleProps:c,postStyleProps:l,totalTime:d,errors:p}}var Ie={},Pe=function(){function e(e,t,n){this._triggerName=e,this.ast=t,this._stateStyles=n}return e.prototype.match=function(e,t,n,r){return i=this.ast.matchers,o=e,a=t,s=n,u=r,i.some(function(e){return e(o,a,s,u)});var i,o,a,s,u},e.prototype.buildStyles=function(e,t,n){var r=this._stateStyles["*"],i=this._stateStyles[e],o=r?r.buildStyles(t,n):{};return i?i.buildStyles(t,n):o},e.prototype.build=function(e,t,n,i,o,a,s,u,c,l){var d=[],f=this.ast.options&&this.ast.options.params||Ie,h=s&&s.params||Ie,m=this.buildStyles(n,h,d),g=u&&u.params||Ie,v=this.buildStyles(i,g,d),_=new Set,y=new Map,b=new Map,j="void"===i,x={params:Object(r.__assign)({},f,g)},O=l?[]:_e(e,t,this.ast.animation,o,a,m,v,x,c,d),w=0;if(O.forEach(function(e){w=Math.max(e.duration+e.delay,w)}),d.length)return De(t,this._triggerName,n,i,j,m,v,[],[],y,b,w,d);O.forEach(function(e){var n=e.element,r=p(y,n,{});e.preStyleProps.forEach(function(e){return r[e]=!0});var i=p(b,n,{});e.postStyleProps.forEach(function(e){return i[e]=!0}),n!==t&&_.add(n)});var C=K(_.values());return De(t,this._triggerName,n,i,j,m,v,O,C,y,b,w)},e}();var Re=function(){function e(e,t){this.styles=e,this.defaultParams=t}return e.prototype.buildStyles=function(e,t){var n={},r=F(this.defaultParams);return Object.keys(e).forEach(function(t){var n=e[t];null!=n&&(r[t]=n)}),this.styles.styles.forEach(function(e){if("string"!=typeof e){var i=e;Object.keys(i).forEach(function(e){var o=i[e];o.length>1&&(o=Y(o,r,t)),n[e]=o})}}),n},e}();var Ne=function(){function e(e,t){var n,r,i=this;this.name=e,this.ast=t,this.transitionFactories=[],this.states={},t.states.forEach(function(e){var t=e.options&&e.options.params||{};i.states[e.name]=new Re(e.style,t)}),Fe(this.states,"true","1"),Fe(this.states,"false","0"),t.transitions.forEach(function(t){i.transitionFactories.push(new Pe(e,t,i.states))}),this.fallbackTransition=(n=e,r=this.states,new Pe(n,{type:1,animation:{type:2,steps:[],options:null},matchers:[function(e,t){return!0}],options:null,queryCount:0,depCount:0},r))}return Object.defineProperty(e.prototype,"containsQueries",{get:function(){return this.ast.queryCount>0},enumerable:!0,configurable:!0}),e.prototype.matchTransition=function(e,t,n,r){return this.transitionFactories.find(function(i){return i.match(e,t,n,r)})||null},e.prototype.matchStyles=function(e,t,n){return this.fallbackTransition.buildStyles(e,t,n)},e}();function Fe(e,t,n){e.hasOwnProperty(t)?e.hasOwnProperty(n)||(e[n]=e[t]):e.hasOwnProperty(n)&&(e[t]=e[n])}var Le=new me,Ve=function(){function e(e,t,n){this.bodyNode=e,this._driver=t,this._normalizer=n,this._animations={},this._playersById={},this.players=[]}return e.prototype.register=function(e,t){var n=[],r=ue(this._driver,t,n);if(n.length)throw new Error("Unable to build the animation due to the following errors: "+n.join("\n"));this._animations[e]=r},e.prototype._buildPlayer=function(e,t,n){var r=e.element,i=u(this._driver,this._normalizer,0,e.keyframes,t,n);return this._driver.animate(r,i,e.duration,e.delay,e.easing,[],!0)},e.prototype.create=function(e,t,n){var r=this;void 0===n&&(n={});var o,a=[],u=this._animations[e],c=new Map;if(u?(o=_e(this._driver,t,u,"ng-enter",A,{},{},n,Le,a)).forEach(function(e){var t=p(c,e.element,{});e.postStyleProps.forEach(function(e){return t[e]=null})}):(a.push("The requested animation doesn't exist or has already been destroyed"),o=[]),a.length)throw new Error("Unable to create the animation due to the following errors: "+a.join("\n"));c.forEach(function(e,t){Object.keys(e).forEach(function(n){e[n]=r._driver.computeStyle(t,n,i.AUTO_STYLE)})});var l=s(o.map(function(e){var t=c.get(e.element);return r._buildPlayer(e,{},t)}));return this._playersById[e]=l,l.onDestroy(function(){return r.destroy(e)}),this.players.push(l),l},e.prototype.destroy=function(e){var t=this._getPlayer(e);t.destroy(),delete this._playersById[e];var n=this.players.indexOf(t);n>=0&&this.players.splice(n,1)},e.prototype._getPlayer=function(e){var t=this._playersById[e];if(!t)throw new Error("Unable to find the timeline player referenced by "+e);return t},e.prototype.listen=function(e,t,n,r){var i=d(t,"","","");return c(this._getPlayer(e),n,i,r),function(){}},e.prototype.command=function(e,t,n,r){if("register"!=n)if("create"!=n){var i=this._getPlayer(e);switch(n){case"play":i.play();break;case"pause":i.pause();break;case"reset":i.reset();break;case"restart":i.restart();break;case"finish":i.finish();break;case"init":i.init();break;case"setPosition":i.setPosition(parseFloat(r[0]));break;case"destroy":this.destroy(e)}}else{var o=r[0]||{};this.create(e,t,o)}else this.register(e,r[0])},e}(),Be="ng-animate-queued",ze="ng-animate-disabled",Ue=".ng-animate-disabled",He=[],qe={namespaceId:"",setForRemoval:!1,setForMove:!1,hasAnimation:!1,removedBeforeQueried:!1},Ge={namespaceId:"",setForMove:!1,setForRemoval:!1,hasAnimation:!1,removedBeforeQueried:!0},We="__ng_removed",Ye=function(){function e(e,t){void 0===t&&(t=""),this.namespaceId=t;var n,r=e&&e.hasOwnProperty("value"),i=r?e.value:e;if(this.value=null!=(n=i)?n:null,r){var o=F(e);delete o.value,this.options=o}else this.options={};this.options.params||(this.options.params={})}return Object.defineProperty(e.prototype,"params",{get:function(){return this.options.params},enumerable:!0,configurable:!0}),e.prototype.absorbOptions=function(e){var t=e.params;if(t){var n=this.options.params;Object.keys(t).forEach(function(e){null==n[e]&&(n[e]=t[e])})}},e}(),Ke=new Ye("void"),Qe=function(){function e(e,t,n){this.id=e,this.hostElement=t,this._engine=n,this.players=[],this._triggers={},this._queue=[],this._elementListeners=new Map,this._hostClassName="ng-tns-"+e,rt(t,this._hostClassName)}return e.prototype.listen=function(e,t,n,r){var i,o=this;if(!this._triggers.hasOwnProperty(t))throw new Error('Unable to listen on the animation trigger event "'+n+'" because the animation trigger "'+t+"\" doesn't exist!");if(null==n||0==n.length)throw new Error('Unable to listen on the animation trigger "'+t+'" because the provided event is undefined!');if("start"!=(i=n)&&"done"!=i)throw new Error('The provided animation trigger event "'+n+'" for the animation trigger "'+t+'" is not supported!');var a=p(this._elementListeners,e,[]),s={name:t,phase:n,callback:r};a.push(s);var u=p(this._engine.statesByElement,e,{});return u.hasOwnProperty(t)||(rt(e,T),rt(e,T+"-"+t),u[t]=Ke),function(){o._engine.afterFlush(function(){var e=a.indexOf(s);e>=0&&a.splice(e,1),o._triggers[t]||delete u[t]})}},e.prototype.register=function(e,t){return!this._triggers[e]&&(this._triggers[e]=t,!0)},e.prototype._getTrigger=function(e){var t=this._triggers[e];if(!t)throw new Error('The provided animation trigger "'+e+'" has not been registered!');return t},e.prototype.trigger=function(e,t,n,r){var i=this;void 0===r&&(r=!0);var o=this._getTrigger(t),a=new Ze(this.id,t,e),s=this._engine.statesByElement.get(e);s||(rt(e,T),rt(e,T+"-"+t),this._engine.statesByElement.set(e,s={}));var u=s[t],c=new Ye(n,this.id);if(!(n&&n.hasOwnProperty("value"))&&u&&c.absorbOptions(u.options),s[t]=c,u||(u=Ke),"void"===c.value||u.value!==c.value){var l=p(this._engine.playersByElement,e,[]);l.forEach(function(e){e.namespaceId==i.id&&e.triggerName==t&&e.queued&&e.destroy()});var d=o.matchTransition(u.value,c.value,e,c.params),f=!1;if(!d){if(!r)return;d=o.fallbackTransition,f=!0}return this._engine.totalQueuedPlayers++,this._queue.push({element:e,triggerName:t,transition:d,fromState:u,toState:c,player:a,isFallbackTransition:f}),f||(rt(e,Be),a.onStart(function(){it(e,Be)})),a.onDone(function(){var t=i.players.indexOf(a);t>=0&&i.players.splice(t,1);var n=i._engine.playersByElement.get(e);if(n){var r=n.indexOf(a);r>=0&&n.splice(r,1)}}),this.players.push(a),l.push(a),a}if(!function(e,t){var n=Object.keys(e),r=Object.keys(t);if(n.length!=r.length)return!1;for(var i=0;i<n.length;i++){var o=n[i];if(!t.hasOwnProperty(o)||e[o]!==t[o])return!1}return!0}(u.params,c.params)){var h=[],m=o.matchStyles(u.value,u.params,h),g=o.matchStyles(c.value,c.params,h);h.length?this._engine.reportError(h):this._engine.afterFlush(function(){H(e,m),U(e,g)})}},e.prototype.deregister=function(e){var t=this;delete this._triggers[e],this._engine.statesByElement.forEach(function(t,n){delete t[e]}),this._elementListeners.forEach(function(n,r){t._elementListeners.set(r,n.filter(function(t){return t.name!=e}))})},e.prototype.clearElementCache=function(e){this._engine.statesByElement.delete(e),this._elementListeners.delete(e);var t=this._engine.playersByElement.get(e);t&&(t.forEach(function(e){return e.destroy()}),this._engine.playersByElement.delete(e))},e.prototype._signalRemovalForInnerTriggers=function(e,t,n){var r=this;void 0===n&&(n=!1),this._engine.driver.query(e,D,!0).forEach(function(e){if(!e[We]){var n=r._engine.fetchNamespacesByElement(e);n.size?n.forEach(function(n){return n.triggerLeaveAnimation(e,t,!1,!0)}):r.clearElementCache(e)}})},e.prototype.triggerLeaveAnimation=function(e,t,n,r){var i=this,o=this._engine.statesByElement.get(e);if(o){var a=[];if(Object.keys(o).forEach(function(t){if(i._triggers[t]){var n=i.trigger(e,t,"void",r);n&&a.push(n)}}),a.length)return this._engine.markElementAsRemoved(this.id,e,!0,t),n&&s(a).onDone(function(){return i._engine.processLeaveNode(e)}),!0}return!1},e.prototype.prepareLeaveAnimationListeners=function(e){var t=this,n=this._elementListeners.get(e);if(n){var r=new Set;n.forEach(function(n){var i=n.name;if(!r.has(i)){r.add(i);var o=t._triggers[i].fallbackTransition,a=t._engine.statesByElement.get(e)[i]||Ke,s=new Ye("void"),u=new Ze(t.id,i,e);t._engine.totalQueuedPlayers++,t._queue.push({element:e,triggerName:i,transition:o,fromState:a,toState:s,player:u,isFallbackTransition:!0})}})}},e.prototype.removeNode=function(e,t){var n=this,r=this._engine;if(e.childElementCount&&this._signalRemovalForInnerTriggers(e,t,!0),!this.triggerLeaveAnimation(e,t,!0)){var i=!1;if(r.totalAnimations){var o=r.players.length?r.playersByQueriedElement.get(e):[];if(o&&o.length)i=!0;else for(var a=e;a=a.parentNode;){if(r.statesByElement.get(a)){i=!0;break}}}this.prepareLeaveAnimationListeners(e),i?r.markElementAsRemoved(this.id,e,!1,t):(r.afterFlush(function(){return n.clearElementCache(e)}),r.destroyInnerAnimations(e),r._onRemovalComplete(e,t))}},e.prototype.insertNode=function(e,t){rt(e,this._hostClassName)},e.prototype.drainQueuedTransitions=function(e){var t=this,n=[];return this._queue.forEach(function(r){var i=r.player;if(!i.destroyed){var o=r.element,a=t._elementListeners.get(o);a&&a.forEach(function(t){if(t.name==r.triggerName){var n=d(o,r.triggerName,r.fromState.value,r.toState.value);n._data=e,c(r.player,t.phase,n,t.callback)}}),i.markedForDestroy?t._engine.afterFlush(function(){i.destroy()}):n.push(r)}}),this._queue=[],n.sort(function(e,n){var r=e.transition.ast.depCount,i=n.transition.ast.depCount;return 0==r||0==i?r-i:t._engine.driver.containsElement(e.element,n.element)?1:-1})},e.prototype.destroy=function(e){this.players.forEach(function(e){return e.destroy()}),this._signalRemovalForInnerTriggers(this.hostElement,e)},e.prototype.elementContainsData=function(e){var t=!1;return this._elementListeners.has(e)&&(t=!0),t=!!this._queue.find(function(t){return t.element===e})||t},e}(),Xe=function(){function e(e,t,n){this.bodyNode=e,this.driver=t,this._normalizer=n,this.players=[],this.newHostElements=new Map,this.playersByElement=new Map,this.playersByQueriedElement=new Map,this.statesByElement=new Map,this.disabledNodes=new Set,this.totalAnimations=0,this.totalQueuedPlayers=0,this._namespaceLookup={},this._namespaceList=[],this._flushFns=[],this._whenQuietFns=[],this.namespacesByHostElement=new Map,this.collectedEnterElements=[],this.collectedLeaveElements=[],this.onRemovalComplete=function(e,t){}}return e.prototype._onRemovalComplete=function(e,t){this.onRemovalComplete(e,t)},Object.defineProperty(e.prototype,"queuedPlayers",{get:function(){var e=[];return this._namespaceList.forEach(function(t){t.players.forEach(function(t){t.queued&&e.push(t)})}),e},enumerable:!0,configurable:!0}),e.prototype.createNamespace=function(e,t){var n=new Qe(e,t,this);return t.parentNode?this._balanceNamespaceList(n,t):(this.newHostElements.set(t,n),this.collectEnterElement(t)),this._namespaceLookup[e]=n},e.prototype._balanceNamespaceList=function(e,t){var n=this._namespaceList.length-1;if(n>=0){for(var r=!1,i=n;i>=0;i--){var o=this._namespaceList[i];if(this.driver.containsElement(o.hostElement,t)){this._namespaceList.splice(i+1,0,e),r=!0;break}}r||this._namespaceList.splice(0,0,e)}else this._namespaceList.push(e);return this.namespacesByHostElement.set(t,e),e},e.prototype.register=function(e,t){var n=this._namespaceLookup[e];return n||(n=this.createNamespace(e,t)),n},e.prototype.registerTrigger=function(e,t,n){var r=this._namespaceLookup[e];r&&r.register(t,n)&&this.totalAnimations++},e.prototype.destroy=function(e,t){var n=this;if(e){var r=this._fetchNamespace(e);this.afterFlush(function(){n.namespacesByHostElement.delete(r.hostElement),delete n._namespaceLookup[e];var t=n._namespaceList.indexOf(r);t>=0&&n._namespaceList.splice(t,1)}),this.afterFlushAnimationsDone(function(){return r.destroy(t)})}},e.prototype._fetchNamespace=function(e){return this._namespaceLookup[e]},e.prototype.fetchNamespacesByElement=function(e){var t=new Set,n=this.statesByElement.get(e);if(n)for(var r=Object.keys(n),i=0;i<r.length;i++){var o=n[r[i]].namespaceId;if(o){var a=this._fetchNamespace(o);a&&t.add(a)}}return t},e.prototype.trigger=function(e,t,n,r){if($e(t)){var i=this._fetchNamespace(e);if(i)return i.trigger(t,n,r),!0}return!1},e.prototype.insertNode=function(e,t,n,r){if($e(t)){var i=t[We];if(i&&i.setForRemoval){i.setForRemoval=!1,i.setForMove=!0;var o=this.collectedLeaveElements.indexOf(t);o>=0&&this.collectedLeaveElements.splice(o,1)}if(e){var a=this._fetchNamespace(e);a&&a.insertNode(t,n)}r&&this.collectEnterElement(t)}},e.prototype.collectEnterElement=function(e){this.collectedEnterElements.push(e)},e.prototype.markElementAsDisabled=function(e,t){t?this.disabledNodes.has(e)||(this.disabledNodes.add(e),rt(e,ze)):this.disabledNodes.has(e)&&(this.disabledNodes.delete(e),it(e,ze))},e.prototype.removeNode=function(e,t,n){if($e(t)){var r=e?this._fetchNamespace(e):null;r?r.removeNode(t,n):this.markElementAsRemoved(e,t,!1,n)}else this._onRemovalComplete(t,n)},e.prototype.markElementAsRemoved=function(e,t,n,r){this.collectedLeaveElements.push(t),t[We]={namespaceId:e,setForRemoval:r,hasAnimation:n,removedBeforeQueried:!1}},e.prototype.listen=function(e,t,n,r,i){return $e(t)?this._fetchNamespace(e).listen(t,n,r,i):function(){}},e.prototype._buildInstruction=function(e,t,n,r,i){return e.transition.build(this.driver,e.element,e.fromState.value,e.toState.value,n,r,e.fromState.options,e.toState.options,t,i)},e.prototype.destroyInnerAnimations=function(e){var t=this,n=this.driver.query(e,D,!0);n.forEach(function(e){return t.destroyActiveAnimationsForElement(e)}),0!=this.playersByQueriedElement.size&&(n=this.driver.query(e,I,!0)).forEach(function(e){return t.finishActiveQueriedAnimationOnElement(e)})},e.prototype.destroyActiveAnimationsForElement=function(e){var t=this.playersByElement.get(e);t&&t.forEach(function(e){e.queued?e.markedForDestroy=!0:e.destroy()})},e.prototype.finishActiveQueriedAnimationOnElement=function(e){var t=this.playersByQueriedElement.get(e);t&&t.forEach(function(e){return e.finish()})},e.prototype.whenRenderingDone=function(){var e=this;return new Promise(function(t){if(e.players.length)return s(e.players).onDone(function(){return t()});t()})},e.prototype.processLeaveNode=function(e){var t=this,n=e[We];if(n&&n.setForRemoval){if(e[We]=qe,n.namespaceId){this.destroyInnerAnimations(e);var r=this._fetchNamespace(n.namespaceId);r&&r.clearElementCache(e)}this._onRemovalComplete(e,n.setForRemoval)}this.driver.matchesElement(e,Ue)&&this.markElementAsDisabled(e,!1),this.driver.query(e,Ue,!0).forEach(function(e){t.markElementAsDisabled(e,!1)})},e.prototype.flush=function(e){var t=this;void 0===e&&(e=-1);var n=[];if(this.newHostElements.size&&(this.newHostElements.forEach(function(e,n){return t._balanceNamespaceList(e,n)}),this.newHostElements.clear()),this.totalAnimations&&this.collectedEnterElements.length)for(var r=0;r<this.collectedEnterElements.length;r++){rt(this.collectedEnterElements[r],"ng-star-inserted")}if(this._namespaceList.length&&(this.totalQueuedPlayers||this.collectedLeaveElements.length)){var i=[];try{n=this._flushAnimations(i,e)}finally{for(r=0;r<i.length;r++)i[r]()}}else for(r=0;r<this.collectedLeaveElements.length;r++){var o=this.collectedLeaveElements[r];this.processLeaveNode(o)}if(this.totalQueuedPlayers=0,this.collectedEnterElements.length=0,this.collectedLeaveElements.length=0,this._flushFns.forEach(function(e){return e()}),this._flushFns=[],this._whenQuietFns.length){var a=this._whenQuietFns;this._whenQuietFns=[],n.length?s(n).onDone(function(){a.forEach(function(e){return e()})}):a.forEach(function(e){return e()})}},e.prototype.reportError=function(e){throw new Error("Unable to process animations due to the following failed trigger transitions\n "+e.join("\n"))},e.prototype._flushAnimations=function(e,t){var n=this,o=new me,a=[],u=new Map,c=[],l=new Map,d=new Map,f=new Map,h=new Set;this.disabledNodes.forEach(function(e){h.add(e);for(var t=n.driver.query(e,".ng-animate-queued",!0),r=0;r<t.length;r++)h.add(t[r])});var m=this.bodyNode,g=Array.from(this.statesByElement.keys()),v=tt(g,this.collectedEnterElements),_=new Map,y=0;v.forEach(function(e,t){var n="ng-enter"+y++;_.set(t,n),e.forEach(function(e){return rt(e,n)})});for(var b=[],j=new Set,x=new Set,O=0;O<this.collectedLeaveElements.length;O++){(q=(z=this.collectedLeaveElements[O])[We])&&q.setForRemoval&&(b.push(z),j.add(z),q.hasAnimation?this.driver.query(z,".ng-star-inserted",!0).forEach(function(e){return j.add(e)}):x.add(z))}var w=new Map,C=tt(g,Array.from(j));C.forEach(function(e,t){var n=A+y++;w.set(t,n),e.forEach(function(e){return rt(e,n)})}),e.push(function(){v.forEach(function(e,t){var n=_.get(t);e.forEach(function(e){return it(e,n)})}),C.forEach(function(e,t){var n=w.get(t);e.forEach(function(e){return it(e,n)})}),b.forEach(function(e){n.processLeaveNode(e)})});for(var E=[],S=[],k=this._namespaceList.length-1;k>=0;k--){this._namespaceList[k].drainQueuedTransitions(t).forEach(function(e){var t=e.player,r=e.element;if(E.push(t),n.collectedEnterElements.length){var i=r[We];if(i&&i.setForMove)return void t.destroy()}var s=!m||!n.driver.containsElement(m,r),u=w.get(r),h=_.get(r),g=n._buildInstruction(e,o,h,u,s);if(g.errors&&g.errors.length)S.push(g);else{if(s)return t.onStart(function(){return H(r,g.fromStyles)}),t.onDestroy(function(){return U(r,g.toStyles)}),void a.push(t);if(e.isFallbackTransition)return t.onStart(function(){return H(r,g.fromStyles)}),t.onDestroy(function(){return U(r,g.toStyles)}),void a.push(t);g.timelines.forEach(function(e){return e.stretchStartingKeyframe=!0}),o.append(r,g.timelines);var v={instruction:g,player:t,element:r};c.push(v),g.queriedElements.forEach(function(e){return p(l,e,[]).push(t)}),g.preStyleProps.forEach(function(e,t){var n=Object.keys(e);if(n.length){var r=d.get(t);r||d.set(t,r=new Set),n.forEach(function(e){return r.add(e)})}}),g.postStyleProps.forEach(function(e,t){var n=Object.keys(e),r=f.get(t);r||f.set(t,r=new Set),n.forEach(function(e){return r.add(e)})})}})}if(S.length){var M=[];S.forEach(function(e){M.push("@"+e.triggerName+" has failed due to:\n"),e.errors.forEach(function(e){return M.push("- "+e+"\n")})}),E.forEach(function(e){return e.destroy()}),this.reportError(M)}var T=new Map,D=new Map;c.forEach(function(e){var t=e.element;o.has(t)&&(D.set(t,t),n._beforeAnimationBuild(e.player.namespaceId,e.instruction,T))}),a.forEach(function(e){var t=e.element;n._getPreviousPlayers(t,!1,e.namespaceId,e.triggerName,null).forEach(function(e){p(T,t,[]).push(e),e.destroy()})});var P=b.filter(function(e){return at(e,d,f)}),R=new Map;et(R,this.driver,x,f,i.AUTO_STYLE).forEach(function(e){at(e,d,f)&&P.push(e)});var N=new Map;v.forEach(function(e,t){et(N,n.driver,new Set(e),d,i["ɵPRE_STYLE"])}),P.forEach(function(e){var t=R.get(e),n=N.get(e);R.set(e,Object(r.__assign)({},t,n))});var F=[],L=[],V={};c.forEach(function(e){var t=e.element,r=e.player,i=e.instruction;if(o.has(t)){if(h.has(t))return r.onDestroy(function(){return U(t,i.toStyles)}),r.disabled=!0,r.overrideTotalTime(i.totalTime),void a.push(r);var c=V;if(D.size>1){for(var l=t,d=[];l=l.parentNode;){var p=D.get(l);if(p){c=p;break}d.push(l)}d.forEach(function(e){return D.set(e,c)})}var f=n._buildAnimation(r.namespaceId,i,T,u,N,R);if(r.setRealPlayer(f),c===V)F.push(r);else{var m=n.playersByElement.get(c);m&&m.length&&(r.parentPlayer=s(m)),a.push(r)}}else H(t,i.fromStyles),r.onDestroy(function(){return U(t,i.toStyles)}),L.push(r),h.has(t)&&a.push(r)}),L.forEach(function(e){var t=u.get(e.element);if(t&&t.length){var n=s(t);e.setRealPlayer(n)}}),a.forEach(function(e){e.parentPlayer?e.syncPlayerEvents(e.parentPlayer):e.destroy()});for(var B=0;B<b.length;B++){var z,q=(z=b[B])[We];if(it(z,A),!q||!q.hasAnimation){var G=[];if(l.size){var W=l.get(z);W&&W.length&&G.push.apply(G,Object(r.__spread)(W));for(var Y=this.driver.query(z,I,!0),K=0;K<Y.length;K++){var Q=l.get(Y[K]);Q&&Q.length&&G.push.apply(G,Object(r.__spread)(Q))}}var X=G.filter(function(e){return!e.destroyed});X.length?ot(this,z,X):this.processLeaveNode(z)}}return b.length=0,F.forEach(function(e){n.players.push(e),e.onDone(function(){e.destroy();var t=n.players.indexOf(e);n.players.splice(t,1)}),e.play()}),F},e.prototype.elementContainsData=function(e,t){var n=!1,r=t[We];return r&&r.setForRemoval&&(n=!0),this.playersByElement.has(t)&&(n=!0),this.playersByQueriedElement.has(t)&&(n=!0),this.statesByElement.has(t)&&(n=!0),this._fetchNamespace(e).elementContainsData(t)||n},e.prototype.afterFlush=function(e){this._flushFns.push(e)},e.prototype.afterFlushAnimationsDone=function(e){this._whenQuietFns.push(e)},e.prototype._getPreviousPlayers=function(e,t,n,r,i){var o=[];if(t){var a=this.playersByQueriedElement.get(e);a&&(o=a)}else{var s=this.playersByElement.get(e);if(s){var u=!i||"void"==i;s.forEach(function(e){e.queued||(u||e.triggerName==r)&&o.push(e)})}}return(n||r)&&(o=o.filter(function(e){return(!n||n==e.namespaceId)&&(!r||r==e.triggerName)})),o},e.prototype._beforeAnimationBuild=function(e,t,n){var i,o,a=t.triggerName,s=t.element,u=t.isRemovalTransition?void 0:e,c=t.isRemovalTransition?void 0:a,l=function(e){var r=e.element,i=r!==s,o=p(n,r,[]);d._getPreviousPlayers(r,i,u,c,t.toState).forEach(function(e){var t=e.getRealPlayer();t.beforeDestroy&&t.beforeDestroy(),e.destroy(),o.push(e)})},d=this;try{for(var f=Object(r.__values)(t.timelines),h=f.next();!h.done;h=f.next()){l(h.value)}}catch(e){i={error:e}}finally{try{h&&!h.done&&(o=f.return)&&o.call(f)}finally{if(i)throw i.error}}H(s,t.fromStyles)},e.prototype._buildAnimation=function(e,t,n,r,o,a){var c=this,l=t.triggerName,d=t.element,f=[],h=new Set,m=new Set,g=t.timelines.map(function(t){var s=t.element;h.add(s);var p=s[We];if(p&&p.removedBeforeQueried)return new i.NoopAnimationPlayer(t.duration,t.delay);var g,v,_=s!==d,y=(g=(n.get(s)||He).map(function(e){return e.getRealPlayer()}),v=[],function e(t,n){for(var r=0;r<t.length;r++){var o=t[r];o instanceof i["ɵAnimationGroupPlayer"]?e(o.players,n):n.push(o)}}(g,v),v).filter(function(e){var t=e;return!!t.element&&t.element===s}),b=o.get(s),j=a.get(s),x=u(c.driver,c._normalizer,0,t.keyframes,b,j),O=c._buildPlayer(t,x,y);if(t.subTimeline&&r&&m.add(s),_){var w=new Ze(e,l,s);w.setRealPlayer(O),f.push(w)}return O});f.forEach(function(e){p(c.playersByQueriedElement,e.element,[]).push(e),e.onDone(function(){return function(e,t,n){var r;if(e instanceof Map){if(r=e.get(t)){if(r.length){var i=r.indexOf(n);r.splice(i,1)}0==r.length&&e.delete(t)}}else if(r=e[t]){if(r.length){var i=r.indexOf(n);r.splice(i,1)}0==r.length&&delete e[t]}return r}(c.playersByQueriedElement,e.element,e)})}),h.forEach(function(e){return rt(e,"ng-animating")});var v=s(g);return v.onDestroy(function(){h.forEach(function(e){return it(e,"ng-animating")}),U(d,t.toStyles)}),m.forEach(function(e){p(r,e,[]).push(v)}),v},e.prototype._buildPlayer=function(e,t,n){return t.length>0?this.driver.animate(e.element,t,e.duration,e.delay,e.easing,n):new i.NoopAnimationPlayer(e.duration,e.delay)},e}(),Ze=function(){function e(e,t,n){this.namespaceId=e,this.triggerName=t,this.element=n,this._player=new i.NoopAnimationPlayer,this._containsRealPlayer=!1,this._queuedCallbacks={},this.destroyed=!1,this.markedForDestroy=!1,this.disabled=!1,this.queued=!0,this.totalTime=0}return e.prototype.setRealPlayer=function(e){var t=this;this._containsRealPlayer||(this._player=e,Object.keys(this._queuedCallbacks).forEach(function(n){t._queuedCallbacks[n].forEach(function(t){return c(e,n,void 0,t)})}),this._queuedCallbacks={},this._containsRealPlayer=!0,this.overrideTotalTime(e.totalTime),this.queued=!1)},e.prototype.getRealPlayer=function(){return this._player},e.prototype.overrideTotalTime=function(e){this.totalTime=e},e.prototype.syncPlayerEvents=function(e){var t=this,n=this._player;n.triggerCallback&&e.onStart(function(){return n.triggerCallback("start")}),e.onDone(function(){return t.finish()}),e.onDestroy(function(){return t.destroy()})},e.prototype._queueEvent=function(e,t){p(this._queuedCallbacks,e,[]).push(t)},e.prototype.onDone=function(e){this.queued&&this._queueEvent("done",e),this._player.onDone(e)},e.prototype.onStart=function(e){this.queued&&this._queueEvent("start",e),this._player.onStart(e)},e.prototype.onDestroy=function(e){this.queued&&this._queueEvent("destroy",e),this._player.onDestroy(e)},e.prototype.init=function(){this._player.init()},e.prototype.hasStarted=function(){return!this.queued&&this._player.hasStarted()},e.prototype.play=function(){!this.queued&&this._player.play()},e.prototype.pause=function(){!this.queued&&this._player.pause()},e.prototype.restart=function(){!this.queued&&this._player.restart()},e.prototype.finish=function(){this._player.finish()},e.prototype.destroy=function(){this.destroyed=!0,this._player.destroy()},e.prototype.reset=function(){!this.queued&&this._player.reset()},e.prototype.setPosition=function(e){this.queued||this._player.setPosition(e)},e.prototype.getPosition=function(){return this.queued?0:this._player.getPosition()},e.prototype.triggerCallback=function(e){var t=this._player;t.triggerCallback&&t.triggerCallback(e)},e}();function $e(e){return e&&1===e.nodeType}function Je(e,t){var n=e.style.display;return e.style.display=null!=t?t:"none",n}function et(e,t,n,r,i){var o=[];n.forEach(function(e){return o.push(Je(e))});var a=[];r.forEach(function(n,r){var o={};n.forEach(function(e){var n=o[e]=t.computeStyle(r,e,i);n&&0!=n.length||(r[We]=Ge,a.push(r))}),e.set(r,o)});var s=0;return n.forEach(function(e){return Je(e,o[s++])}),a}function tt(e,t){var n=new Map;if(e.forEach(function(e){return n.set(e,[])}),0==t.length)return n;var r=1,i=new Set(t),o=new Map;return t.forEach(function(e){var t=function e(t){if(!t)return r;var a=o.get(t);if(a)return a;var s=t.parentNode;return a=n.has(s)?s:i.has(s)?r:e(s),o.set(t,a),a}(e);t!==r&&n.get(t).push(e)}),n}var nt="$$classes";function rt(e,t){if(e.classList)e.classList.add(t);else{var n=e[nt];n||(n=e[nt]={}),n[t]=!0}}function it(e,t){if(e.classList)e.classList.remove(t);else{var n=e[nt];n&&delete n[t]}}function ot(e,t,n){s(n).onDone(function(){return e.processLeaveNode(t)})}function at(e,t,n){var r=n.get(e);if(!r)return!1;var i=t.get(e);return i?r.forEach(function(e){return i.add(e)}):t.set(e,r),n.delete(e),!0}var st=function(){function e(e,t,n){var r=this;this.bodyNode=e,this._driver=t,this._triggerCache={},this.onRemovalComplete=function(e,t){},this._transitionEngine=new Xe(e,t,n),this._timelineEngine=new Ve(e,t,n),this._transitionEngine.onRemovalComplete=function(e,t){return r.onRemovalComplete(e,t)}}return e.prototype.registerTrigger=function(e,t,n,r,i){var o=e+"-"+r,a=this._triggerCache[o];if(!a){var s=[],u=ue(this._driver,i,s);if(s.length)throw new Error('The animation trigger "'+r+'" has failed to build due to the following errors:\n - '+s.join("\n - "));a=new Ne(r,u),this._triggerCache[o]=a}this._transitionEngine.registerTrigger(t,r,a)},e.prototype.register=function(e,t){this._transitionEngine.register(e,t)},e.prototype.destroy=function(e,t){this._transitionEngine.destroy(e,t)},e.prototype.onInsert=function(e,t,n,r){this._transitionEngine.insertNode(e,t,n,r)},e.prototype.onRemove=function(e,t,n){this._transitionEngine.removeNode(e,t,n)},e.prototype.disableAnimations=function(e,t){this._transitionEngine.markElementAsDisabled(e,t)},e.prototype.process=function(e,t,n,i){if("@"==n.charAt(0)){var o=Object(r.__read)(f(n),2),a=o[0],s=o[1],u=i;this._timelineEngine.command(a,t,s,u)}else this._transitionEngine.trigger(e,t,n,i)},e.prototype.listen=function(e,t,n,i,o){if("@"==n.charAt(0)){var a=Object(r.__read)(f(n),2),s=a[0],u=a[1];return this._timelineEngine.listen(s,t,u,o)}return this._transitionEngine.listen(e,t,n,i,o)},e.prototype.flush=function(e){void 0===e&&(e=-1),this._transitionEngine.flush(e)},Object.defineProperty(e.prototype,"players",{get:function(){return this._transitionEngine.players.concat(this._timelineEngine.players)},enumerable:!0,configurable:!0}),e.prototype.whenRenderingDone=function(){return this._transitionEngine.whenRenderingDone()},e}(),ut="animation",ct="animationend",lt=function(){function e(e,t,n,r,i,o,a){var s=this;this._element=e,this._name=t,this._duration=n,this._delay=r,this._easing=i,this._fillMode=o,this._onDoneFn=a,this._finished=!1,this._destroyed=!1,this._startTime=0,this._position=0,this._eventFn=function(e){return s._handleCallback(e)}}return e.prototype.apply=function(){!function(e,t){var n=gt(e,"").trim(),r=0;n.length&&(r=function(e,t){for(var n=0,r=0;r<e.length;r++){var i=e.charAt(r);i===t&&n++}return n}(n,",")+1,t=n+", "+t);mt(e,"",t)}(this._element,this._duration+"ms "+this._easing+" "+this._delay+"ms 1 normal "+this._fillMode+" "+this._name),ht(this._element,this._eventFn,!1),this._startTime=Date.now()},e.prototype.pause=function(){dt(this._element,this._name,"paused")},e.prototype.resume=function(){dt(this._element,this._name,"running")},e.prototype.setPosition=function(e){var t=pt(this._element,this._name);this._position=e*this._duration,mt(this._element,"Delay","-"+this._position+"ms",t)},e.prototype.getPosition=function(){return this._position},e.prototype._handleCallback=function(e){var t=e._ngTestManualTimestamp||Date.now(),n=1e3*parseFloat(e.elapsedTime.toFixed(3));e.animationName==this._name&&Math.max(t-this._startTime,0)>=this._delay&&n>=this._duration&&this.finish()},e.prototype.finish=function(){this._finished||(this._finished=!0,this._onDoneFn(),ht(this._element,this._eventFn,!0))},e.prototype.destroy=function(){this._destroyed||(this._destroyed=!0,this.finish(),function(e,t){var n=gt(e,"").split(","),r=ft(n,t);if(r>=0){n.splice(r,1);var i=n.join(",");mt(e,"",i)}}(this._element,this._name))},e}();function dt(e,t,n){mt(e,"PlayState",n,pt(e,t))}function pt(e,t){var n=gt(e,"");return n.indexOf(",")>0?ft(n.split(","),t):ft([n],t)}function ft(e,t){for(var n=0;n<e.length;n++)if(e[n].indexOf(t)>=0)return n;return-1}function ht(e,t,n){n?e.removeEventListener(ct,t):e.addEventListener(ct,t)}function mt(e,t,n,r){var i=ut+t;if(null!=r){var o=e.style[i];if(o.length){var a=o.split(",");a[r]=n,n=a.join(",")}}e.style[i]=n}function gt(e,t){return e.style[ut+t]}var vt="linear",_t=function(){function e(e,t,n,r,i,o,a){this.element=e,this.keyframes=t,this.animationName=n,this._duration=r,this._delay=i,this._finalStyles=a,this._onDoneFns=[],this._onStartFns=[],this._onDestroyFns=[],this._started=!1,this.currentSnapshot={},this._state=0,this.easing=o||vt,this.totalTime=r+i,this._buildStyler()}return e.prototype.onStart=function(e){this._onStartFns.push(e)},e.prototype.onDone=function(e){this._onDoneFns.push(e)},e.prototype.onDestroy=function(e){this._onDestroyFns.push(e)},e.prototype.destroy=function(){this.init(),this._state>=4||(this._state=4,this._styler.destroy(),this._flushStartFns(),this._flushDoneFns(),this._onDestroyFns.forEach(function(e){return e()}),this._onDestroyFns=[])},e.prototype._flushDoneFns=function(){this._onDoneFns.forEach(function(e){return e()}),this._onDoneFns=[]},e.prototype._flushStartFns=function(){this._onStartFns.forEach(function(e){return e()}),this._onStartFns=[]},e.prototype.finish=function(){this.init(),this._state>=3||(this._state=3,this._styler.finish(),this._flushStartFns(),this._flushDoneFns())},e.prototype.setPosition=function(e){this._styler.setPosition(e)},e.prototype.getPosition=function(){return this._styler.getPosition()},e.prototype.hasStarted=function(){return this._state>=2},e.prototype.init=function(){if(!(this._state>=1)){this._state=1;this.element;this._styler.apply(),this._delay&&this._styler.pause()}},e.prototype.play=function(){this.init(),this.hasStarted()||(this._flushStartFns(),this._state=2),this._styler.resume()},e.prototype.pause=function(){this.init(),this._styler.pause()},e.prototype.restart=function(){this.reset(),this.play()},e.prototype.reset=function(){this._styler.destroy(),this._buildStyler(),this._styler.apply()},e.prototype._buildStyler=function(){var e=this;this._styler=new lt(this.element,this.animationName,this._duration,this._delay,this.easing,"forwards",function(){return e.finish()})},e.prototype.triggerCallback=function(e){var t="start"==e?this._onStartFns:this._onDoneFns;t.forEach(function(e){return e()}),t.length=0},e.prototype.beforeDestroy=function(){var e=this;this.init();var t={};if(this.hasStarted()){var n=this._state>=3;Object.keys(this._finalStyles).forEach(function(r){"offset"!=r&&(t[r]=n?e._finalStyles[r]:ee(e.element,r))})}this.currentSnapshot=t},e}(),yt=function(e){function t(t,n){var r=e.call(this)||this;return r.element=t,r._startingStyles={},r.__initialized=!1,r._styles=E(n),r}return Object(r.__extends)(t,e),t.prototype.init=function(){var t=this;!this.__initialized&&this._startingStyles&&(this.__initialized=!0,Object.keys(this._styles).forEach(function(e){t._startingStyles[e]=t.element.style[e]}),e.prototype.init.call(this))},t.prototype.play=function(){var t=this;this._startingStyles&&(this.init(),Object.keys(this._styles).forEach(function(e){return t.element.style.setProperty(e,t._styles[e])}),e.prototype.play.call(this))},t.prototype.destroy=function(){var t=this;this._startingStyles&&(Object.keys(this._startingStyles).forEach(function(e){var n=t._startingStyles[e];n?t.element.style.setProperty(e,n):t.element.style.removeProperty(e)}),this._startingStyles=null,e.prototype.destroy.call(this))},t}(i.NoopAnimationPlayer),bt=function(){function e(){this._count=0,this._head=document.querySelector("head"),this._warningIssued=!1}return e.prototype.validateStyleProperty=function(e){return x(e)},e.prototype.matchesElement=function(e,t){return O(e,t)},e.prototype.containsElement=function(e,t){return w(e,t)},e.prototype.query=function(e,t,n){return C(e,t,n)},e.prototype.computeStyle=function(e,t,n){return window.getComputedStyle(e)[t]},e.prototype.buildKeyframeElement=function(e,t,n){var r="@keyframes "+t+" {\n",i="";(n=n.map(function(e){return E(e)})).forEach(function(e){i=" ";var t=parseFloat(e.offset);r+=""+i+100*t+"% {\n",i+=" ",Object.keys(e).forEach(function(t){var n=e[t];switch(t){case"offset":return;case"easing":return void(n&&(r+=i+"animation-timing-function: "+n+";\n"));default:return void(r+=""+i+t+": "+n+";\n")}}),r+=i+"}\n"}),r+="}\n";var o=document.createElement("style");return o.innerHTML=r,o},e.prototype.animate=function(e,t,n,r,i,o,a){void 0===o&&(o=[]),a&&this._notifyFaultyScrubber();var s=o.filter(function(e){return e instanceof _t}),u={};Z(n,r)&&s.forEach(function(e){var t=e.currentSnapshot;Object.keys(t).forEach(function(e){return u[e]=t[e]})});var c=function(e){var t={};if(e){var n=Array.isArray(e)?e:[e];n.forEach(function(e){Object.keys(e).forEach(function(n){"offset"!=n&&"easing"!=n&&(t[n]=e[n])})})}return t}(t=$(e,t,u));if(0==n)return new yt(e,c);var l="gen_css_kf_"+this._count++,d=this.buildKeyframeElement(e,l,t);document.querySelector("head").appendChild(d);var p=new _t(e,t,l,n,r,i,c);return p.onDestroy(function(){var e;(e=d).parentNode.removeChild(e)}),p},e.prototype._notifyFaultyScrubber=function(){this._warningIssued||(console.warn("@angular/animations: please load the web-animations.js polyfill to allow programmatic access...\n","  visit http://bit.ly/IWukam to learn more about using the web-animation-js polyfill."),this._warningIssued=!0)},e}();var jt=function(){function e(e,t,n){this.element=e,this.keyframes=t,this.options=n,this._onDoneFns=[],this._onStartFns=[],this._onDestroyFns=[],this._initialized=!1,this._finished=!1,this._started=!1,this._destroyed=!1,this.time=0,this.parentPlayer=null,this.currentSnapshot={},this._duration=n.duration,this._delay=n.delay||0,this.time=this._duration+this._delay}return e.prototype._onFinish=function(){this._finished||(this._finished=!0,this._onDoneFns.forEach(function(e){return e()}),this._onDoneFns=[])},e.prototype.init=function(){this._buildPlayer(),this._preparePlayerBeforeStart()},e.prototype._buildPlayer=function(){var e=this;if(!this._initialized){this._initialized=!0;var t=this.keyframes;this.domPlayer=this._triggerWebAnimation(this.element,t,this.options),this._finalKeyframe=t.length?t[t.length-1]:{},this.domPlayer.addEventListener("finish",function(){return e._onFinish()})}},e.prototype._preparePlayerBeforeStart=function(){this._delay?this._resetDomPlayerState():this.domPlayer.pause()},e.prototype._triggerWebAnimation=function(e,t,n){return e.animate(t,n)},e.prototype.onStart=function(e){this._onStartFns.push(e)},e.prototype.onDone=function(e){this._onDoneFns.push(e)},e.prototype.onDestroy=function(e){this._onDestroyFns.push(e)},e.prototype.play=function(){this._buildPlayer(),this.hasStarted()||(this._onStartFns.forEach(function(e){return e()}),this._onStartFns=[],this._started=!0),this.domPlayer.play()},e.prototype.pause=function(){this.init(),this.domPlayer.pause()},e.prototype.finish=function(){this.init(),this._onFinish(),this.domPlayer.finish()},e.prototype.reset=function(){this._resetDomPlayerState(),this._destroyed=!1,this._finished=!1,this._started=!1},e.prototype._resetDomPlayerState=function(){this.domPlayer&&this.domPlayer.cancel()},e.prototype.restart=function(){this.reset(),this.play()},e.prototype.hasStarted=function(){return this._started},e.prototype.destroy=function(){this._destroyed||(this._destroyed=!0,this._resetDomPlayerState(),this._onFinish(),this._onDestroyFns.forEach(function(e){return e()}),this._onDestroyFns=[])},e.prototype.setPosition=function(e){this.domPlayer.currentTime=e*this.time},e.prototype.getPosition=function(){return this.domPlayer.currentTime/this.time},Object.defineProperty(e.prototype,"totalTime",{get:function(){return this._delay+this._duration},enumerable:!0,configurable:!0}),e.prototype.beforeDestroy=function(){var e=this,t={};this.hasStarted()&&Object.keys(this._finalKeyframe).forEach(function(n){"offset"!=n&&(t[n]=e._finished?e._finalKeyframe[n]:ee(e.element,n))}),this.currentSnapshot=t},e.prototype.triggerCallback=function(e){var t="start"==e?this._onStartFns:this._onDoneFns;t.forEach(function(e){return e()}),t.length=0},e}(),xt=function(){function e(){this._isNativeImpl=/\{\s*\[native\s+code\]\s*\}/.test(wt().toString()),this._cssKeyframesDriver=new bt}return e.prototype.validateStyleProperty=function(e){return x(e)},e.prototype.matchesElement=function(e,t){return O(e,t)},e.prototype.containsElement=function(e,t){return w(e,t)},e.prototype.query=function(e,t,n){return C(e,t,n)},e.prototype.computeStyle=function(e,t,n){return window.getComputedStyle(e)[t]},e.prototype.overrideWebAnimationsSupport=function(e){this._isNativeImpl=e},e.prototype.animate=function(e,t,n,r,i,o,a){if(void 0===o&&(o=[]),!a&&!this._isNativeImpl)return this._cssKeyframesDriver.animate(e,t,n,r,i,o);var s={duration:n,delay:r,fill:0==r?"both":"forwards"};i&&(s.easing=i);var u={},c=o.filter(function(e){return e instanceof jt});return Z(n,r)&&c.forEach(function(e){var t=e.currentSnapshot;Object.keys(t).forEach(function(e){return u[e]=t[e]})}),t=$(e,t=t.map(function(e){return V(e,!1)}),u),new jt(e,t,s)},e}();function Ot(){return"function"==typeof wt()}function wt(){return"undefined"!=typeof window&&void 0!==window.document&&Element.prototype.animate||{}}},"./node_modules/@angular/animations/node_modules/tslib/tslib.es6.js":function(e,t,n){"use strict";n.r(t),n.d(t,"__extends",function(){return i}),n.d(t,"__assign",function(){return o}),n.d(t,"__rest",function(){return a}),n.d(t,"__decorate",function(){return s}),n.d(t,"__param",function(){return u}),n.d(t,"__metadata",function(){return c}),n.d(t,"__awaiter",function(){return l}),n.d(t,"__generator",function(){return d}),n.d(t,"__exportStar",function(){return p}),n.d(t,"__values",function(){return f}),n.d(t,"__read",function(){return h}),n.d(t,"__spread",function(){return m}),n.d(t,"__await",function(){return g}),n.d(t,"__asyncGenerator",function(){return v}),n.d(t,"__asyncDelegator",function(){return _}),n.d(t,"__asyncValues",function(){return y}),n.d(t,"__makeTemplateObject",function(){return b}),n.d(t,"__importStar",function(){return j}),n.d(t,"__importDefault",function(){return x});var r=function(e,t){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,t){e.__proto__=t}||function(e,t){for(var n in t)t.hasOwnProperty(n)&&(e[n]=t[n])})(e,t)};function i(e,t){function n(){this.constructor=e}r(e,t),e.prototype=null===t?Object.create(t):(n.prototype=t.prototype,new n)}var o=function(){return(o=Object.assign||function(e){for(var t,n=1,r=arguments.length;n<r;n++){t=arguments[n];for(var i in t)Object.prototype.hasOwnProperty.call(t,i)&&(e[i]=t[i])}return e}).apply(this,arguments)};function a(e,t){var n={};for(var r in e)Object.prototype.hasOwnProperty.call(e,r)&&t.indexOf(r)<0&&(n[r]=e[r]);if(null!=e&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(e);i<r.length;i++)t.indexOf(r[i])<0&&(n[r[i]]=e[r[i]])}return n}function s(e,t,n,r){var i,o=arguments.length,a=o<3?t:null===r?r=Object.getOwnPropertyDescriptor(t,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(e,t,n,r);else for(var s=e.length-1;s>=0;s--)(i=e[s])&&(a=(o<3?i(a):o>3?i(t,n,a):i(t,n))||a);return o>3&&a&&Object.defineProperty(t,n,a),a}function u(e,t){return function(n,r){t(n,r,e)}}function c(e,t){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(e,t)}function l(e,t,n,r){return new(n||(n=Promise))(function(i,o){function a(e){try{u(r.next(e))}catch(e){o(e)}}function s(e){try{u(r.throw(e))}catch(e){o(e)}}function u(e){e.done?i(e.value):new n(function(t){t(e.value)}).then(a,s)}u((r=r.apply(e,t||[])).next())})}function d(e,t){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=t.call(e,a)}catch(e){o=[6,e],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(e,t){for(var n in e)t.hasOwnProperty(n)||(t[n]=e[n])}function f(e){var t="function"==typeof Symbol&&e[Symbol.iterator],n=0;return t?t.call(e):{next:function(){return e&&n>=e.length&&(e=void 0),{value:e&&e[n++],done:!e}}}}function h(e,t){var n="function"==typeof Symbol&&e[Symbol.iterator];if(!n)return e;var r,i,o=n.call(e),a=[];try{for(;(void 0===t||t-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(e){i={error:e}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var e=[],t=0;t<arguments.length;t++)e=e.concat(h(arguments[t]));return e}function g(e){return this instanceof g?(this.v=e,this):new g(e)}function v(e,t,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(e,t||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(e){i[e]&&(r[e]=function(t){return new Promise(function(n,r){o.push([e,t,n,r])>1||s(e,t)})})}function s(e,t){try{(n=i[e](t)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(e){l(o[0][3],e)}var n}function u(e){s("next",e)}function c(e){s("throw",e)}function l(e,t){e(t),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(e){var t,n;return t={},r("next"),r("throw",function(e){throw e}),r("return"),t[Symbol.iterator]=function(){return this},t;function r(r,i){t[r]=e[r]?function(t){return(n=!n)?{value:g(e[r](t)),done:"return"===r}:i?i(t):t}:i}}function y(e){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var t,n=e[Symbol.asyncIterator];return n?n.call(e):(e=f(e),t={},r("next"),r("throw"),r("return"),t[Symbol.asyncIterator]=function(){return this},t);function r(n){t[n]=e[n]&&function(t){return new Promise(function(r,i){var o,a,s,u;t=e[n](t),o=r,a=i,s=t.done,u=t.value,Promise.resolve(u).then(function(e){o({value:e,done:s})},a)})}}}function b(e,t){return Object.defineProperty?Object.defineProperty(e,"raw",{value:t}):e.raw=t,e}function j(e){if(e&&e.__esModule)return e;var t={};if(null!=e)for(var n in e)Object.hasOwnProperty.call(e,n)&&(t[n]=e[n]);return t.default=e,t}function x(e){return e&&e.__esModule?e:{default:e}}},"./node_modules/@angular/cdk/esm5/a11y.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"ARIA_DESCRIBER_PROVIDER_FACTORY",function(){return j}),n.d(t,"MESSAGES_CONTAINER_ID",function(){return h}),n.d(t,"CDK_DESCRIBEDBY_ID_PREFIX",function(){return m}),n.d(t,"CDK_DESCRIBEDBY_HOST_ATTRIBUTE",function(){return g}),n.d(t,"AriaDescriber",function(){return b}),n.d(t,"ARIA_DESCRIBER_PROVIDER",function(){return x}),n.d(t,"ActiveDescendantKeyManager",function(){return w}),n.d(t,"FocusKeyManager",function(){return C}),n.d(t,"ListKeyManager",function(){return O}),n.d(t,"FocusTrap",function(){return M}),n.d(t,"FocusTrapFactory",function(){return A}),n.d(t,"CdkTrapFocus",function(){return T}),n.d(t,"InteractivityChecker",function(){return E}),n.d(t,"LIVE_ANNOUNCER_PROVIDER_FACTORY",function(){return N}),n.d(t,"LiveAnnouncer",function(){return P}),n.d(t,"CdkAriaLive",function(){return R}),n.d(t,"LIVE_ANNOUNCER_PROVIDER",function(){return F}),n.d(t,"LIVE_ANNOUNCER_ELEMENT_TOKEN_FACTORY",function(){return I}),n.d(t,"LIVE_ANNOUNCER_ELEMENT_TOKEN",function(){return D}),n.d(t,"FOCUS_MONITOR_PROVIDER_FACTORY",function(){return U}),n.d(t,"TOUCH_BUFFER_MS",function(){return L}),n.d(t,"FocusMonitor",function(){return B}),n.d(t,"CdkMonitorFocus",function(){return z}),n.d(t,"FOCUS_MONITOR_PROVIDER",function(){return H}),n.d(t,"isFakeMousedownFromScreenReader",function(){return q}),n.d(t,"A11yModule",function(){return G});var r=n("./node_modules/@angular/common/fesm5/common.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/rxjs/_esm5/index.js"),a=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),s=n("./node_modules/rxjs/_esm5/operators/index.js"),u=n("./node_modules/tslib/tslib.es6.js"),c=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),l=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),d=n("./node_modules/@angular/cdk/esm5/observers.es5.js"),p=" ";function f(e,t){return(e.getAttribute(t)||"").match(/\S+/g)||[]}var h="cdk-describedby-message-container",m="cdk-describedby-message",g="cdk-describedby-host",v=0,_=new Map,y=null,b=function(){function e(e){this._document=e}return e.prototype.describe=function(e,t){this._canBeDescribed(e,t)&&(_.has(t)||this._createMessageElement(t),this._isElementDescribedByMessage(e,t)||this._addMessageReference(e,t))},e.prototype.removeDescription=function(e,t){if(this._canBeDescribed(e,t)){this._isElementDescribedByMessage(e,t)&&this._removeMessageReference(e,t);var n=_.get(t);n&&0===n.referenceCount&&this._deleteMessageElement(t),y&&0===y.childNodes.length&&this._deleteMessagesContainer()}},e.prototype.ngOnDestroy=function(){for(var e=this._document.querySelectorAll("["+g+"]"),t=0;t<e.length;t++)this._removeCdkDescribedByReferenceIds(e[t]),e[t].removeAttribute(g);y&&this._deleteMessagesContainer(),_.clear()},e.prototype._createMessageElement=function(e){var t=this._document.createElement("div");t.setAttribute("id",m+"-"+v++),t.appendChild(this._document.createTextNode(e)),this._createMessagesContainer(),y.appendChild(t),_.set(e,{messageElement:t,referenceCount:0})},e.prototype._deleteMessageElement=function(e){var t=_.get(e),n=t&&t.messageElement;y&&n&&y.removeChild(n),_.delete(e)},e.prototype._createMessagesContainer=function(){if(!y){var e=this._document.getElementById(h);e&&e.parentNode.removeChild(e),(y=this._document.createElement("div")).id=h,y.setAttribute("aria-hidden","true"),y.style.display="none",this._document.body.appendChild(y)}},e.prototype._deleteMessagesContainer=function(){y&&y.parentNode&&(y.parentNode.removeChild(y),y=null)},e.prototype._removeCdkDescribedByReferenceIds=function(e){var t=f(e,"aria-describedby").filter(function(e){return 0!=e.indexOf(m)});e.setAttribute("aria-describedby",t.join(" "))},e.prototype._addMessageReference=function(e,t){var n,r,i,o,a=_.get(t);n=e,r="aria-describedby",i=a.messageElement.id,(o=f(n,r)).some(function(e){return e.trim()==i.trim()})||(o.push(i.trim()),n.setAttribute(r,o.join(p))),e.setAttribute(g,""),a.referenceCount++},e.prototype._removeMessageReference=function(e,t){var n,r,i,o,a=_.get(t);a.referenceCount--,n=e,r="aria-describedby",i=a.messageElement.id,o=f(n,r).filter(function(e){return e!=i.trim()}),n.setAttribute(r,o.join(p)),e.removeAttribute(g)},e.prototype._isElementDescribedByMessage=function(e,t){var n=f(e,"aria-describedby"),r=_.get(t),i=r&&r.messageElement.id;return!!i&&-1!=n.indexOf(i)},e.prototype._canBeDescribed=function(e,t){return e.nodeType===this._document.ELEMENT_NODE&&null!=t&&!!(""+t).trim()},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:i.Inject,args:[r.DOCUMENT]}]}]},e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e(Object(i.inject)(r.DOCUMENT))},token:e,providedIn:"root"}),e}();function j(e,t){return e||new b(t)}var x={provide:b,deps:[[new i.Optional,new i.SkipSelf,b],r.DOCUMENT],useFactory:j},O=function(){function e(e){var t=this;this._items=e,this._activeItemIndex=-1,this._activeItem=null,this._wrap=!1,this._letterKeyStream=new o.Subject,this._typeaheadSubscription=o.Subscription.EMPTY,this._vertical=!0,this._allowedModifierKeys=[],this._skipPredicateFn=function(e){return e.disabled},this._pressedLetters=[],this.tabOut=new o.Subject,this.change=new o.Subject,e instanceof i.QueryList&&e.changes.subscribe(function(e){if(t._activeItem){var n=e.toArray().indexOf(t._activeItem);n>-1&&n!==t._activeItemIndex&&(t._activeItemIndex=n)}})}return e.prototype.skipPredicate=function(e){return this._skipPredicateFn=e,this},e.prototype.withWrap=function(e){return void 0===e&&(e=!0),this._wrap=e,this},e.prototype.withVerticalOrientation=function(e){return void 0===e&&(e=!0),this._vertical=e,this},e.prototype.withHorizontalOrientation=function(e){return this._horizontal=e,this},e.prototype.withAllowedModifierKeys=function(e){return this._allowedModifierKeys=e,this},e.prototype.withTypeAhead=function(e){var t=this;if(void 0===e&&(e=200),this._items.length&&this._items.some(function(e){return"function"!=typeof e.getLabel}))throw Error("ListKeyManager items in typeahead mode must implement the `getLabel` method.");return this._typeaheadSubscription.unsubscribe(),this._typeaheadSubscription=this._letterKeyStream.pipe(Object(s.tap)(function(e){return t._pressedLetters.push(e)}),Object(s.debounceTime)(e),Object(s.filter)(function(){return t._pressedLetters.length>0}),Object(s.map)(function(){return t._pressedLetters.join("")})).subscribe(function(e){for(var n=t._getItemsArray(),r=1;r<n.length+1;r++){var i=(t._activeItemIndex+r)%n.length,o=n[i];if(!t._skipPredicateFn(o)&&0===o.getLabel().toUpperCase().trim().indexOf(e)){t.setActiveItem(i);break}}t._pressedLetters=[]}),this},e.prototype.setActiveItem=function(e){var t=this._activeItemIndex;this.updateActiveItem(e),this._activeItemIndex!==t&&this.change.next(this._activeItemIndex)},e.prototype.onKeydown=function(e){var t=this,n=e.keyCode,r=["altKey","ctrlKey","metaKey","shiftKey"].every(function(n){return!e[n]||t._allowedModifierKeys.indexOf(n)>-1});switch(n){case a.TAB:return void this.tabOut.next();case a.DOWN_ARROW:if(this._vertical&&r){this.setNextItemActive();break}return;case a.UP_ARROW:if(this._vertical&&r){this.setPreviousItemActive();break}return;case a.RIGHT_ARROW:if(this._horizontal&&r){"rtl"===this._horizontal?this.setPreviousItemActive():this.setNextItemActive();break}return;case a.LEFT_ARROW:if(this._horizontal&&r){"rtl"===this._horizontal?this.setNextItemActive():this.setPreviousItemActive();break}return;default:return void((r||Object(a.hasModifierKey)(e,"shiftKey"))&&(e.key&&1===e.key.length?this._letterKeyStream.next(e.key.toLocaleUpperCase()):(n>=a.A&&n<=a.Z||n>=a.ZERO&&n<=a.NINE)&&this._letterKeyStream.next(String.fromCharCode(n))))}this._pressedLetters=[],e.preventDefault()},Object.defineProperty(e.prototype,"activeItemIndex",{get:function(){return this._activeItemIndex},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"activeItem",{get:function(){return this._activeItem},enumerable:!0,configurable:!0}),e.prototype.setFirstItemActive=function(){this._setActiveItemByIndex(0,1)},e.prototype.setLastItemActive=function(){this._setActiveItemByIndex(this._items.length-1,-1)},e.prototype.setNextItemActive=function(){this._activeItemIndex<0?this.setFirstItemActive():this._setActiveItemByDelta(1)},e.prototype.setPreviousItemActive=function(){this._activeItemIndex<0&&this._wrap?this.setLastItemActive():this._setActiveItemByDelta(-1)},e.prototype.updateActiveItem=function(e){var t=this._getItemsArray(),n="number"==typeof e?e:t.indexOf(e),r=t[n];this._activeItem=null==r?null:r,this._activeItemIndex=n},e.prototype.updateActiveItemIndex=function(e){this.updateActiveItem(e)},e.prototype._setActiveItemByDelta=function(e){this._wrap?this._setActiveInWrapMode(e):this._setActiveInDefaultMode(e)},e.prototype._setActiveInWrapMode=function(e){for(var t=this._getItemsArray(),n=1;n<=t.length;n++){var r=(this._activeItemIndex+e*n+t.length)%t.length,i=t[r];if(!this._skipPredicateFn(i))return void this.setActiveItem(r)}},e.prototype._setActiveInDefaultMode=function(e){this._setActiveItemByIndex(this._activeItemIndex+e,e)},e.prototype._setActiveItemByIndex=function(e,t){var n=this._getItemsArray();if(n[e]){for(;this._skipPredicateFn(n[e]);)if(!n[e+=t])return;this.setActiveItem(e)}},e.prototype._getItemsArray=function(){return this._items instanceof i.QueryList?this._items.toArray():this._items},e}(),w=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(u.__extends)(t,e),t.prototype.setActiveItem=function(t){this.activeItem&&this.activeItem.setInactiveStyles(),e.prototype.setActiveItem.call(this,t),this.activeItem&&this.activeItem.setActiveStyles()},t}(O),C=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t._origin="program",t}return Object(u.__extends)(t,e),t.prototype.setFocusOrigin=function(e){return this._origin=e,this},t.prototype.setActiveItem=function(t){e.prototype.setActiveItem.call(this,t),this.activeItem&&this.activeItem.focus(this._origin)},t}(O),E=function(){function e(e){this._platform=e}return e.prototype.isDisabled=function(e){return e.hasAttribute("disabled")},e.prototype.isVisible=function(e){return!!((t=e).offsetWidth||t.offsetHeight||"function"==typeof t.getClientRects&&t.getClientRects().length)&&"visible"===getComputedStyle(e).visibility;var t},e.prototype.isTabbable=function(e){if(!this._platform.isBrowser)return!1;var t,n=function(e){try{return e.frameElement}catch(e){return null}}((t=e).ownerDocument&&t.ownerDocument.defaultView||window);if(n){var r=n&&n.nodeName.toLowerCase();if(-1===k(n))return!1;if((this._platform.BLINK||this._platform.WEBKIT)&&"object"===r)return!1;if((this._platform.BLINK||this._platform.WEBKIT)&&!this.isVisible(n))return!1}var i,o,a,s=e.nodeName.toLowerCase(),u=k(e);if(e.hasAttribute("contenteditable"))return-1!==u;if("iframe"===s)return!1;if("audio"===s){if(!e.hasAttribute("controls"))return!1;if(this._platform.BLINK)return!0}if("video"===s){if(!e.hasAttribute("controls")&&this._platform.TRIDENT)return!1;if(this._platform.BLINK||this._platform.FIREFOX)return!0}return("object"!==s||!this._platform.BLINK&&!this._platform.WEBKIT)&&((!this._platform.WEBKIT||!this._platform.IOS||(o=(i=e).nodeName.toLowerCase(),"text"===(a="input"===o&&i.type)||"password"===a||"select"===o||"textarea"===o))&&e.tabIndex>=0)},e.prototype.isFocusable=function(e){return function(e){if(t=e,n=t,"input"==n.nodeName.toLowerCase()&&"hidden"==t.type)return!1;var t,n;return o=e,a=o.nodeName.toLowerCase(),"input"===a||"select"===a||"button"===a||"textarea"===a||(r=e,i=r,"a"==i.nodeName.toLowerCase()&&r.hasAttribute("href"))||e.hasAttribute("contenteditable")||S(e);var r,i;var o,a}(e)&&!this.isDisabled(e)&&this.isVisible(e)},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:c.Platform}]},e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e(Object(i.inject)(c.Platform))},token:e,providedIn:"root"}),e}();function S(e){if(!e.hasAttribute("tabindex")||void 0===e.tabIndex)return!1;var t=e.getAttribute("tabindex");return"-32768"!=t&&!(!t||isNaN(parseInt(t,10)))}function k(e){if(!S(e))return null;var t=parseInt(e.getAttribute("tabindex")||"",10);return isNaN(t)?-1:t}var M=function(){function e(e,t,n,r,i){void 0===i&&(i=!1);var o=this;this._element=e,this._checker=t,this._ngZone=n,this._document=r,this._hasAttached=!1,this._startAnchorListener=function(){return o.focusLastTabbableElement()},this._endAnchorListener=function(){return o.focusFirstTabbableElement()},this._enabled=!0,i||this.attachAnchors()}return Object.defineProperty(e.prototype,"enabled",{get:function(){return this._enabled},set:function(e){this._enabled=e,this._startAnchor&&this._endAnchor&&(this._toggleAnchorTabIndex(e,this._startAnchor),this._toggleAnchorTabIndex(e,this._endAnchor))},enumerable:!0,configurable:!0}),e.prototype.destroy=function(){var e=this._startAnchor,t=this._endAnchor;e&&(e.removeEventListener("focus",this._startAnchorListener),e.parentNode&&e.parentNode.removeChild(e)),t&&(t.removeEventListener("focus",this._endAnchorListener),t.parentNode&&t.parentNode.removeChild(t)),this._startAnchor=this._endAnchor=null},e.prototype.attachAnchors=function(){var e=this;return!!this._hasAttached||(this._ngZone.runOutsideAngular(function(){e._startAnchor||(e._startAnchor=e._createAnchor(),e._startAnchor.addEventListener("focus",e._startAnchorListener)),e._endAnchor||(e._endAnchor=e._createAnchor(),e._endAnchor.addEventListener("focus",e._endAnchorListener))}),this._element.parentNode&&(this._element.parentNode.insertBefore(this._startAnchor,this._element),this._element.parentNode.insertBefore(this._endAnchor,this._element.nextSibling),this._hasAttached=!0),this._hasAttached)},e.prototype.focusInitialElementWhenReady=function(){var e=this;return new Promise(function(t){e._executeOnStable(function(){return t(e.focusInitialElement())})})},e.prototype.focusFirstTabbableElementWhenReady=function(){var e=this;return new Promise(function(t){e._executeOnStable(function(){return t(e.focusFirstTabbableElement())})})},e.prototype.focusLastTabbableElementWhenReady=function(){var e=this;return new Promise(function(t){e._executeOnStable(function(){return t(e.focusLastTabbableElement())})})},e.prototype._getRegionBoundary=function(e){for(var t=this._element.querySelectorAll("[cdk-focus-region-"+e+"], [cdkFocusRegion"+e+"], [cdk-focus-"+e+"]"),n=0;n<t.length;n++)t[n].hasAttribute("cdk-focus-"+e)?console.warn("Found use of deprecated attribute 'cdk-focus-"+e+"', use 'cdkFocusRegion"+e+"' instead. The deprecated attribute will be removed in 8.0.0.",t[n]):t[n].hasAttribute("cdk-focus-region-"+e)&&console.warn("Found use of deprecated attribute 'cdk-focus-region-"+e+"', use 'cdkFocusRegion"+e+"' instead. The deprecated attribute will be removed in 8.0.0.",t[n]);return"start"==e?t.length?t[0]:this._getFirstTabbableElement(this._element):t.length?t[t.length-1]:this._getLastTabbableElement(this._element)},e.prototype.focusInitialElement=function(){var e=this._element.querySelector("[cdk-focus-initial], [cdkFocusInitial]");return e?(e.hasAttribute("cdk-focus-initial")&&console.warn("Found use of deprecated attribute 'cdk-focus-initial', use 'cdkFocusInitial' instead. The deprecated attribute will be removed in 8.0.0",e),Object(i.isDevMode)()&&!this._checker.isFocusable(e)&&console.warn("Element matching '[cdkFocusInitial]' is not focusable.",e),e.focus(),!0):this.focusFirstTabbableElement()},e.prototype.focusFirstTabbableElement=function(){var e=this._getRegionBoundary("start");return e&&e.focus(),!!e},e.prototype.focusLastTabbableElement=function(){var e=this._getRegionBoundary("end");return e&&e.focus(),!!e},e.prototype.hasAttached=function(){return this._hasAttached},e.prototype._getFirstTabbableElement=function(e){if(this._checker.isFocusable(e)&&this._checker.isTabbable(e))return e;for(var t=e.children||e.childNodes,n=0;n<t.length;n++){var r=t[n].nodeType===this._document.ELEMENT_NODE?this._getFirstTabbableElement(t[n]):null;if(r)return r}return null},e.prototype._getLastTabbableElement=function(e){if(this._checker.isFocusable(e)&&this._checker.isTabbable(e))return e;for(var t=e.children||e.childNodes,n=t.length-1;n>=0;n--){var r=t[n].nodeType===this._document.ELEMENT_NODE?this._getLastTabbableElement(t[n]):null;if(r)return r}return null},e.prototype._createAnchor=function(){var e=this._document.createElement("div");return this._toggleAnchorTabIndex(this._enabled,e),e.classList.add("cdk-visually-hidden"),e.classList.add("cdk-focus-trap-anchor"),e},e.prototype._toggleAnchorTabIndex=function(e,t){e?t.setAttribute("tabindex","0"):t.removeAttribute("tabindex")},e.prototype._executeOnStable=function(e){this._ngZone.isStable?e():this._ngZone.onStable.asObservable().pipe(Object(s.take)(1)).subscribe(e)},e}(),A=function(){function e(e,t,n){this._checker=e,this._ngZone=t,this._document=n}return e.prototype.create=function(e,t){return void 0===t&&(t=!1),new M(e,this._checker,this._ngZone,this._document,t)},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:E},{type:i.NgZone},{type:void 0,decorators:[{type:i.Inject,args:[r.DOCUMENT]}]}]},e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e(Object(i.inject)(E),Object(i.inject)(i.NgZone),Object(i.inject)(r.DOCUMENT))},token:e,providedIn:"root"}),e}(),T=function(){function e(e,t,n){this._elementRef=e,this._focusTrapFactory=t,this._previouslyFocusedElement=null,this._document=n,this.focusTrap=this._focusTrapFactory.create(this._elementRef.nativeElement,!0)}return Object.defineProperty(e.prototype,"enabled",{get:function(){return this.focusTrap.enabled},set:function(e){this.focusTrap.enabled=Object(l.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"autoCapture",{get:function(){return this._autoCapture},set:function(e){this._autoCapture=Object(l.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),e.prototype.ngOnDestroy=function(){this.focusTrap.destroy(),this._previouslyFocusedElement&&(this._previouslyFocusedElement.focus(),this._previouslyFocusedElement=null)},e.prototype.ngAfterContentInit=function(){this.focusTrap.attachAnchors(),this.autoCapture&&(this._previouslyFocusedElement=this._document.activeElement,this.focusTrap.focusInitialElementWhenReady())},e.prototype.ngDoCheck=function(){this.focusTrap.hasAttached()||this.focusTrap.attachAnchors()},e.decorators=[{type:i.Directive,args:[{selector:"[cdkTrapFocus]",exportAs:"cdkTrapFocus"}]}],e.ctorParameters=function(){return[{type:i.ElementRef},{type:A},{type:void 0,decorators:[{type:i.Inject,args:[r.DOCUMENT]}]}]},e.propDecorators={enabled:[{type:i.Input,args:["cdkTrapFocus"]}],autoCapture:[{type:i.Input,args:["cdkTrapFocusAutoCapture"]}]},e}(),D=new i.InjectionToken("liveAnnouncerElement",{providedIn:"root",factory:I});function I(){return null}var P=function(){function e(e,t,n){this._ngZone=t,this._document=n,this._liveElement=e||this._createLiveElement()}return e.prototype.announce=function(e){for(var t,n,r=this,i=[],o=1;o<arguments.length;o++)i[o-1]=arguments[o];return 1===i.length&&"number"==typeof i[0]?n=i[0]:(t=i[0],n=i[1]),this.clear(),clearTimeout(this._previousTimeout),this._liveElement.setAttribute("aria-live",t||"polite"),this._ngZone.runOutsideAngular(function(){return new Promise(function(t){clearTimeout(r._previousTimeout),r._previousTimeout=setTimeout(function(){r._liveElement.textContent=e,t(),"number"==typeof n&&(r._previousTimeout=setTimeout(function(){return r.clear()},n))},100)})})},e.prototype.clear=function(){this._liveElement&&(this._liveElement.textContent="")},e.prototype.ngOnDestroy=function(){clearTimeout(this._previousTimeout),this._liveElement&&this._liveElement.parentNode&&(this._liveElement.parentNode.removeChild(this._liveElement),this._liveElement=null)},e.prototype._createLiveElement=function(){for(var e="cdk-live-announcer-element",t=this._document.getElementsByClassName(e),n=this._document.createElement("div"),r=0;r<t.length;r++)t[r].parentNode.removeChild(t[r]);return n.classList.add(e),n.classList.add("cdk-visually-hidden"),n.setAttribute("aria-atomic","true"),n.setAttribute("aria-live","polite"),this._document.body.appendChild(n),n},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:i.Optional},{type:i.Inject,args:[D]}]},{type:i.NgZone},{type:void 0,decorators:[{type:i.Inject,args:[r.DOCUMENT]}]}]},e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e(Object(i.inject)(D,8),Object(i.inject)(i.NgZone),Object(i.inject)(r.DOCUMENT))},token:e,providedIn:"root"}),e}(),R=function(){function e(e,t,n,r){this._elementRef=e,this._liveAnnouncer=t,this._contentObserver=n,this._ngZone=r,this._politeness="off"}return Object.defineProperty(e.prototype,"politeness",{get:function(){return this._politeness},set:function(e){var t=this;this._politeness="polite"===e||"assertive"===e?e:"off","off"===this._politeness?this._subscription&&(this._subscription.unsubscribe(),this._subscription=null):this._subscription||(this._subscription=this._ngZone.runOutsideAngular(function(){return t._contentObserver.observe(t._elementRef).subscribe(function(){var e=t._elementRef.nativeElement.textContent;e!==t._previousAnnouncedText&&(t._liveAnnouncer.announce(e,t._politeness),t._previousAnnouncedText=e)})}))},enumerable:!0,configurable:!0}),e.prototype.ngOnDestroy=function(){this._subscription&&this._subscription.unsubscribe()},e.decorators=[{type:i.Directive,args:[{selector:"[cdkAriaLive]",exportAs:"cdkAriaLive"}]}],e.ctorParameters=function(){return[{type:i.ElementRef},{type:P},{type:d.ContentObserver},{type:i.NgZone}]},e.propDecorators={politeness:[{type:i.Input,args:["cdkAriaLive"]}]},e}();function N(e,t,n,r){return e||new P(t,r,n)}var F={provide:P,deps:[[new i.Optional,new i.SkipSelf,P],[new i.Optional,new i.Inject(D)],r.DOCUMENT,i.NgZone],useFactory:N},L=650,V=Object(c.normalizePassiveListenerOptions)({passive:!0,capture:!0}),B=function(){function e(e,t){var n=this;this._ngZone=e,this._platform=t,this._origin=null,this._windowFocused=!1,this._elementInfo=new Map,this._monitoredElementCount=0,this._documentKeydownListener=function(){n._lastTouchTarget=null,n._setOriginForCurrentEventQueue("keyboard")},this._documentMousedownListener=function(){n._lastTouchTarget||n._setOriginForCurrentEventQueue("mouse")},this._documentTouchstartListener=function(e){null!=n._touchTimeoutId&&clearTimeout(n._touchTimeoutId),n._lastTouchTarget=e.target,n._touchTimeoutId=setTimeout(function(){return n._lastTouchTarget=null},L)},this._windowFocusListener=function(){n._windowFocused=!0,n._windowFocusTimeoutId=setTimeout(function(){return n._windowFocused=!1})}}return e.prototype.monitor=function(e,t){var n=this;if(void 0===t&&(t=!1),!this._platform.isBrowser)return Object(o.of)(null);var r=Object(l.coerceElement)(e);if(this._elementInfo.has(r)){var i=this._elementInfo.get(r);return i.checkChildren=t,i.subject.asObservable()}var a={unlisten:function(){},checkChildren:t,subject:new o.Subject};this._elementInfo.set(r,a),this._incrementMonitoredElementCount();var s=function(e){return n._onFocus(e,r)},u=function(e){return n._onBlur(e,r)};return this._ngZone.runOutsideAngular(function(){r.addEventListener("focus",s,!0),r.addEventListener("blur",u,!0)}),a.unlisten=function(){r.removeEventListener("focus",s,!0),r.removeEventListener("blur",u,!0)},a.subject.asObservable()},e.prototype.stopMonitoring=function(e){var t=Object(l.coerceElement)(e),n=this._elementInfo.get(t);n&&(n.unlisten(),n.subject.complete(),this._setClasses(t),this._elementInfo.delete(t),this._decrementMonitoredElementCount())},e.prototype.focusVia=function(e,t,n){var r=Object(l.coerceElement)(e);this._setOriginForCurrentEventQueue(t),"function"==typeof r.focus&&r.focus(n)},e.prototype.ngOnDestroy=function(){var e=this;this._elementInfo.forEach(function(t,n){return e.stopMonitoring(n)})},e.prototype._toggleClass=function(e,t,n){n?e.classList.add(t):e.classList.remove(t)},e.prototype._setClasses=function(e,t){this._elementInfo.get(e)&&(this._toggleClass(e,"cdk-focused",!!t),this._toggleClass(e,"cdk-touch-focused","touch"===t),this._toggleClass(e,"cdk-keyboard-focused","keyboard"===t),this._toggleClass(e,"cdk-mouse-focused","mouse"===t),this._toggleClass(e,"cdk-program-focused","program"===t))},e.prototype._setOriginForCurrentEventQueue=function(e){var t=this;this._ngZone.runOutsideAngular(function(){t._origin=e,t._originTimeoutId=setTimeout(function(){return t._origin=null},1)})},e.prototype._wasCausedByTouch=function(e){var t=e.target;return this._lastTouchTarget instanceof Node&&t instanceof Node&&(t===this._lastTouchTarget||t.contains(this._lastTouchTarget))},e.prototype._onFocus=function(e,t){var n=this._elementInfo.get(t);if(n&&(n.checkChildren||t===e.target)){var r=this._origin;r||(r=this._windowFocused&&this._lastFocusOrigin?this._lastFocusOrigin:this._wasCausedByTouch(e)?"touch":"program"),this._setClasses(t,r),this._emitOrigin(n.subject,r),this._lastFocusOrigin=r}},e.prototype._onBlur=function(e,t){var n=this._elementInfo.get(t);!n||n.checkChildren&&e.relatedTarget instanceof Node&&t.contains(e.relatedTarget)||(this._setClasses(t),this._emitOrigin(n.subject,null))},e.prototype._emitOrigin=function(e,t){this._ngZone.run(function(){return e.next(t)})},e.prototype._incrementMonitoredElementCount=function(){var e=this;1==++this._monitoredElementCount&&this._platform.isBrowser&&this._ngZone.runOutsideAngular(function(){document.addEventListener("keydown",e._documentKeydownListener,V),document.addEventListener("mousedown",e._documentMousedownListener,V),document.addEventListener("touchstart",e._documentTouchstartListener,V),window.addEventListener("focus",e._windowFocusListener)})},e.prototype._decrementMonitoredElementCount=function(){--this._monitoredElementCount||(document.removeEventListener("keydown",this._documentKeydownListener,V),document.removeEventListener("mousedown",this._documentMousedownListener,V),document.removeEventListener("touchstart",this._documentTouchstartListener,V),window.removeEventListener("focus",this._windowFocusListener),clearTimeout(this._windowFocusTimeoutId),clearTimeout(this._touchTimeoutId),clearTimeout(this._originTimeoutId))},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:i.NgZone},{type:c.Platform}]},e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e(Object(i.inject)(i.NgZone),Object(i.inject)(c.Platform))},token:e,providedIn:"root"}),e}(),z=function(){function e(e,t){var n=this;this._elementRef=e,this._focusMonitor=t,this.cdkFocusChange=new i.EventEmitter,this._monitorSubscription=this._focusMonitor.monitor(this._elementRef,this._elementRef.nativeElement.hasAttribute("cdkMonitorSubtreeFocus")).subscribe(function(e){return n.cdkFocusChange.emit(e)})}return e.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._elementRef),this._monitorSubscription.unsubscribe()},e.decorators=[{type:i.Directive,args:[{selector:"[cdkMonitorElementFocus], [cdkMonitorSubtreeFocus]"}]}],e.ctorParameters=function(){return[{type:i.ElementRef},{type:B}]},e.propDecorators={cdkFocusChange:[{type:i.Output}]},e}();function U(e,t,n){return e||new B(t,n)}var H={provide:B,deps:[[new i.Optional,new i.SkipSelf,B],i.NgZone,c.Platform],useFactory:U};function q(e){return 0===e.buttons}var G=function(){function e(){}return e.decorators=[{type:i.NgModule,args:[{imports:[r.CommonModule,c.PlatformModule,d.ObserversModule],declarations:[R,T,z],exports:[R,T,z]}]}],e}()},"./node_modules/@angular/cdk/esm5/accordion.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"CdkAccordionItem",function(){return l}),n.d(t,"CdkAccordion",function(){return u}),n.d(t,"CdkAccordionModule",function(){return d});var r=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/rxjs/_esm5/index.js"),a=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),s=0,u=function(){function e(){this._stateChanges=new o.Subject,this._openCloseAllActions=new o.Subject,this.id="cdk-accordion-"+s++,this._multi=!1}return Object.defineProperty(e.prototype,"multi",{get:function(){return this._multi},set:function(e){this._multi=Object(r.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),e.prototype.openAll=function(){this._openCloseAll(!0)},e.prototype.closeAll=function(){this._openCloseAll(!1)},e.prototype.ngOnChanges=function(e){this._stateChanges.next(e)},e.prototype.ngOnDestroy=function(){this._stateChanges.complete()},e.prototype._openCloseAll=function(e){this.multi&&this._openCloseAllActions.next(e)},e.decorators=[{type:i.Directive,args:[{selector:"cdk-accordion, [cdkAccordion]",exportAs:"cdkAccordion"}]}],e.propDecorators={multi:[{type:i.Input}]},e}(),c=0,l=function(){function e(e,t,n){var r=this;this.accordion=e,this._changeDetectorRef=t,this._expansionDispatcher=n,this._openCloseAllSubscription=o.Subscription.EMPTY,this.closed=new i.EventEmitter,this.opened=new i.EventEmitter,this.destroyed=new i.EventEmitter,this.expandedChange=new i.EventEmitter,this.id="cdk-accordion-child-"+c++,this._expanded=!1,this._disabled=!1,this._removeUniqueSelectionListener=function(){},this._removeUniqueSelectionListener=n.listen(function(e,t){r.accordion&&!r.accordion.multi&&r.accordion.id===t&&r.id!==e&&(r.expanded=!1)}),this.accordion&&(this._openCloseAllSubscription=this._subscribeToOpenCloseAllActions())}return Object.defineProperty(e.prototype,"expanded",{get:function(){return this._expanded},set:function(e){if(e=Object(r.coerceBooleanProperty)(e),this._expanded!==e){if(this._expanded=e,this.expandedChange.emit(e),e){this.opened.emit();var t=this.accordion?this.accordion.id:this.id;this._expansionDispatcher.notify(this.id,t)}else this.closed.emit();this._changeDetectorRef.markForCheck()}},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disabled",{get:function(){return this._disabled},set:function(e){this._disabled=Object(r.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),e.prototype.ngOnDestroy=function(){this.opened.complete(),this.closed.complete(),this.destroyed.emit(),this.destroyed.complete(),this._removeUniqueSelectionListener(),this._openCloseAllSubscription.unsubscribe()},e.prototype.toggle=function(){this.disabled||(this.expanded=!this.expanded)},e.prototype.close=function(){this.disabled||(this.expanded=!1)},e.prototype.open=function(){this.disabled||(this.expanded=!0)},e.prototype._subscribeToOpenCloseAllActions=function(){var e=this;return this.accordion._openCloseAllActions.subscribe(function(t){e.disabled||(e.expanded=t)})},e.decorators=[{type:i.Directive,args:[{selector:"cdk-accordion-item, [cdkAccordionItem]",exportAs:"cdkAccordionItem",providers:[{provide:u,useValue:void 0}]}]}],e.ctorParameters=function(){return[{type:u,decorators:[{type:i.Optional},{type:i.SkipSelf}]},{type:i.ChangeDetectorRef},{type:a.UniqueSelectionDispatcher}]},e.propDecorators={closed:[{type:i.Output}],opened:[{type:i.Output}],destroyed:[{type:i.Output}],expandedChange:[{type:i.Output}],expanded:[{type:i.Input}],disabled:[{type:i.Input}]},e}(),d=function(){function e(){}return e.decorators=[{type:i.NgModule,args:[{exports:[u,l],declarations:[u,l]}]}],e}()},"./node_modules/@angular/cdk/esm5/bidi.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Directionality",function(){return s}),n.d(t,"DIR_DOCUMENT",function(){return o}),n.d(t,"Dir",function(){return u}),n.d(t,"BidiModule",function(){return c}),n.d(t,"ɵa",function(){return a});var r=n("./node_modules/@angular/common/fesm5/common.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=new i.InjectionToken("cdk-dir-doc",{providedIn:"root",factory:a});function a(){return Object(i.inject)(r.DOCUMENT)}var s=function(){function e(e){if(this.value="ltr",this.change=new i.EventEmitter,e){var t=e.body?e.body.dir:null,n=e.documentElement?e.documentElement.dir:null,r=t||n;this.value="ltr"===r||"rtl"===r?r:"ltr"}}return e.prototype.ngOnDestroy=function(){this.change.complete()},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:i.Optional},{type:i.Inject,args:[o]}]}]},e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e(Object(i.inject)(o,8))},token:e,providedIn:"root"}),e}(),u=function(){function e(){this._dir="ltr",this._isInitialized=!1,this.change=new i.EventEmitter}return Object.defineProperty(e.prototype,"dir",{get:function(){return this._dir},set:function(e){var t=this._dir;this._rawDir=e,this._dir="ltr"===e||"rtl"===e?e:"ltr",t!==this._dir&&this._isInitialized&&this.change.emit(this._dir)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return this.dir},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){this._isInitialized=!0},e.prototype.ngOnDestroy=function(){this.change.complete()},e.decorators=[{type:i.Directive,args:[{selector:"[dir]",providers:[{provide:s,useExisting:e}],host:{"[attr.dir]":"_rawDir"},exportAs:"dir"}]}],e.propDecorators={change:[{type:i.Output,args:["dirChange"]}],dir:[{type:i.Input}]},e}(),c=function(){function e(){}return e.decorators=[{type:i.NgModule,args:[{exports:[u],declarations:[u]}]}],e}()},"./node_modules/@angular/cdk/esm5/coercion.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"coerceBooleanProperty",function(){return i}),n.d(t,"coerceNumberProperty",function(){return o}),n.d(t,"_isNumberValue",function(){return a}),n.d(t,"coerceArray",function(){return s}),n.d(t,"coerceCssPixelValue",function(){return u}),n.d(t,"coerceElement",function(){return c});var r=n("./node_modules/@angular/core/fesm5/core.js");function i(e){return null!=e&&""+e!="false"}function o(e,t){return void 0===t&&(t=0),a(e)?Number(e):t}function a(e){return!isNaN(parseFloat(e))&&!isNaN(Number(e))}function s(e){return Array.isArray(e)?e:[e]}function u(e){return null==e?"":"string"==typeof e?e:e+"px"}function c(e){return e instanceof r.ElementRef?e.nativeElement:e}},"./node_modules/@angular/cdk/esm5/collections.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"UniqueSelectionDispatcher",function(){return l}),n.d(t,"ArrayDataSource",function(){return s}),n.d(t,"DataSource",function(){return a}),n.d(t,"getMultipleValuesInSingleSelectionError",function(){return c}),n.d(t,"SelectionModel",function(){return u});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/index.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=function(){return function(){}}(),s=function(e){function t(t){var n=e.call(this)||this;return n._data=t,n}return Object(r.__extends)(t,e),t.prototype.connect=function(){return this._data instanceof i.Observable?this._data:Object(i.of)(this._data)},t.prototype.disconnect=function(){},t}(a),u=function(){function e(e,t,n){void 0===e&&(e=!1),void 0===n&&(n=!0);var r=this;this._multiple=e,this._emitChanges=n,this._selection=new Set,this._deselectedToEmit=[],this._selectedToEmit=[],this.changed=new i.Subject,this.onChange=this.changed,t&&t.length&&(e?t.forEach(function(e){return r._markSelected(e)}):this._markSelected(t[0]),this._selectedToEmit.length=0)}return Object.defineProperty(e.prototype,"selected",{get:function(){return this._selected||(this._selected=Array.from(this._selection.values())),this._selected},enumerable:!0,configurable:!0}),e.prototype.select=function(){for(var e=this,t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];this._verifyValueAssignment(t),t.forEach(function(t){return e._markSelected(t)}),this._emitChangeEvent()},e.prototype.deselect=function(){for(var e=this,t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];this._verifyValueAssignment(t),t.forEach(function(t){return e._unmarkSelected(t)}),this._emitChangeEvent()},e.prototype.toggle=function(e){this.isSelected(e)?this.deselect(e):this.select(e)},e.prototype.clear=function(){this._unmarkAll(),this._emitChangeEvent()},e.prototype.isSelected=function(e){return this._selection.has(e)},e.prototype.isEmpty=function(){return 0===this._selection.size},e.prototype.hasValue=function(){return!this.isEmpty()},e.prototype.sort=function(e){this._multiple&&this.selected&&this._selected.sort(e)},e.prototype.isMultipleSelection=function(){return this._multiple},e.prototype._emitChangeEvent=function(){this._selected=null,(this._selectedToEmit.length||this._deselectedToEmit.length)&&(this.changed.next({source:this,added:this._selectedToEmit,removed:this._deselectedToEmit}),this._deselectedToEmit=[],this._selectedToEmit=[])},e.prototype._markSelected=function(e){this.isSelected(e)||(this._multiple||this._unmarkAll(),this._selection.add(e),this._emitChanges&&this._selectedToEmit.push(e))},e.prototype._unmarkSelected=function(e){this.isSelected(e)&&(this._selection.delete(e),this._emitChanges&&this._deselectedToEmit.push(e))},e.prototype._unmarkAll=function(){var e=this;this.isEmpty()||this._selection.forEach(function(t){return e._unmarkSelected(t)})},e.prototype._verifyValueAssignment=function(e){if(e.length>1&&!this._multiple)throw c()},e}();function c(){return Error("Cannot pass multiple values into SelectionModel with single-value mode.")}var l=function(){function e(){this._listeners=[]}return e.prototype.notify=function(e,t){for(var n=0,r=this._listeners;n<r.length;n++){(0,r[n])(e,t)}},e.prototype.listen=function(e){var t=this;return this._listeners.push(e),function(){t._listeners=t._listeners.filter(function(t){return e!==t})}},e.prototype.ngOnDestroy=function(){this._listeners=[]},e.decorators=[{type:o.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(o.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}()},"./node_modules/@angular/cdk/esm5/keycodes.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MAC_ENTER",function(){return r}),n.d(t,"BACKSPACE",function(){return i}),n.d(t,"TAB",function(){return o}),n.d(t,"NUM_CENTER",function(){return a}),n.d(t,"ENTER",function(){return s}),n.d(t,"SHIFT",function(){return u}),n.d(t,"CONTROL",function(){return c}),n.d(t,"ALT",function(){return l}),n.d(t,"PAUSE",function(){return d}),n.d(t,"CAPS_LOCK",function(){return p}),n.d(t,"ESCAPE",function(){return f}),n.d(t,"SPACE",function(){return h}),n.d(t,"PAGE_UP",function(){return m}),n.d(t,"PAGE_DOWN",function(){return g}),n.d(t,"END",function(){return v}),n.d(t,"HOME",function(){return _}),n.d(t,"LEFT_ARROW",function(){return y}),n.d(t,"UP_ARROW",function(){return b}),n.d(t,"RIGHT_ARROW",function(){return j}),n.d(t,"DOWN_ARROW",function(){return x}),n.d(t,"PLUS_SIGN",function(){return O}),n.d(t,"PRINT_SCREEN",function(){return w}),n.d(t,"INSERT",function(){return C}),n.d(t,"DELETE",function(){return E}),n.d(t,"ZERO",function(){return S}),n.d(t,"ONE",function(){return k}),n.d(t,"TWO",function(){return M}),n.d(t,"THREE",function(){return A}),n.d(t,"FOUR",function(){return T}),n.d(t,"FIVE",function(){return D}),n.d(t,"SIX",function(){return I}),n.d(t,"SEVEN",function(){return P}),n.d(t,"EIGHT",function(){return R}),n.d(t,"NINE",function(){return N}),n.d(t,"FF_SEMICOLON",function(){return F}),n.d(t,"FF_EQUALS",function(){return L}),n.d(t,"QUESTION_MARK",function(){return V}),n.d(t,"AT_SIGN",function(){return B}),n.d(t,"A",function(){return z}),n.d(t,"B",function(){return U}),n.d(t,"C",function(){return H}),n.d(t,"D",function(){return q}),n.d(t,"E",function(){return G}),n.d(t,"F",function(){return W}),n.d(t,"G",function(){return Y}),n.d(t,"H",function(){return K}),n.d(t,"I",function(){return Q}),n.d(t,"J",function(){return X}),n.d(t,"K",function(){return Z}),n.d(t,"L",function(){return $}),n.d(t,"M",function(){return J}),n.d(t,"N",function(){return ee}),n.d(t,"O",function(){return te}),n.d(t,"P",function(){return ne}),n.d(t,"Q",function(){return re}),n.d(t,"R",function(){return ie}),n.d(t,"S",function(){return oe}),n.d(t,"T",function(){return ae}),n.d(t,"U",function(){return se}),n.d(t,"V",function(){return ue}),n.d(t,"W",function(){return ce}),n.d(t,"X",function(){return le}),n.d(t,"Y",function(){return de}),n.d(t,"Z",function(){return pe}),n.d(t,"META",function(){return fe}),n.d(t,"MAC_WK_CMD_LEFT",function(){return he}),n.d(t,"MAC_WK_CMD_RIGHT",function(){return me}),n.d(t,"CONTEXT_MENU",function(){return ge}),n.d(t,"NUMPAD_ZERO",function(){return ve}),n.d(t,"NUMPAD_ONE",function(){return _e}),n.d(t,"NUMPAD_TWO",function(){return ye}),n.d(t,"NUMPAD_THREE",function(){return be}),n.d(t,"NUMPAD_FOUR",function(){return je}),n.d(t,"NUMPAD_FIVE",function(){return xe}),n.d(t,"NUMPAD_SIX",function(){return Oe}),n.d(t,"NUMPAD_SEVEN",function(){return we}),n.d(t,"NUMPAD_EIGHT",function(){return Ce}),n.d(t,"NUMPAD_NINE",function(){return Ee}),n.d(t,"NUMPAD_MULTIPLY",function(){return Se}),n.d(t,"NUMPAD_PLUS",function(){return ke}),n.d(t,"NUMPAD_MINUS",function(){return Me}),n.d(t,"NUMPAD_PERIOD",function(){return Ae}),n.d(t,"NUMPAD_DIVIDE",function(){return Te}),n.d(t,"F1",function(){return De}),n.d(t,"F2",function(){return Ie}),n.d(t,"F3",function(){return Pe}),n.d(t,"F4",function(){return Re}),n.d(t,"F5",function(){return Ne}),n.d(t,"F6",function(){return Fe}),n.d(t,"F7",function(){return Le}),n.d(t,"F8",function(){return Ve}),n.d(t,"F9",function(){return Be}),n.d(t,"F10",function(){return ze}),n.d(t,"F11",function(){return Ue}),n.d(t,"F12",function(){return He}),n.d(t,"NUM_LOCK",function(){return qe}),n.d(t,"SCROLL_LOCK",function(){return Ge}),n.d(t,"FIRST_MEDIA",function(){return We}),n.d(t,"FF_MINUS",function(){return Ye}),n.d(t,"MUTE",function(){return Ke}),n.d(t,"VOLUME_DOWN",function(){return Qe}),n.d(t,"VOLUME_UP",function(){return Xe}),n.d(t,"FF_MUTE",function(){return Ze}),n.d(t,"FF_VOLUME_DOWN",function(){return $e}),n.d(t,"LAST_MEDIA",function(){return Je}),n.d(t,"FF_VOLUME_UP",function(){return et}),n.d(t,"SEMICOLON",function(){return tt}),n.d(t,"EQUALS",function(){return nt}),n.d(t,"COMMA",function(){return rt}),n.d(t,"DASH",function(){return it}),n.d(t,"SLASH",function(){return ot}),n.d(t,"APOSTROPHE",function(){return at}),n.d(t,"TILDE",function(){return st}),n.d(t,"OPEN_SQUARE_BRACKET",function(){return ut}),n.d(t,"BACKSLASH",function(){return ct}),n.d(t,"CLOSE_SQUARE_BRACKET",function(){return lt}),n.d(t,"SINGLE_QUOTE",function(){return dt}),n.d(t,"MAC_META",function(){return pt}),n.d(t,"hasModifierKey",function(){return ft});var r=3,i=8,o=9,a=12,s=13,u=16,c=17,l=18,d=19,p=20,f=27,h=32,m=33,g=34,v=35,_=36,y=37,b=38,j=39,x=40,O=43,w=44,C=45,E=46,S=48,k=49,M=50,A=51,T=52,D=53,I=54,P=55,R=56,N=57,F=59,L=61,V=63,B=64,z=65,U=66,H=67,q=68,G=69,W=70,Y=71,K=72,Q=73,X=74,Z=75,$=76,J=77,ee=78,te=79,ne=80,re=81,ie=82,oe=83,ae=84,se=85,ue=86,ce=87,le=88,de=89,pe=90,fe=91,he=91,me=93,ge=93,ve=96,_e=97,ye=98,be=99,je=100,xe=101,Oe=102,we=103,Ce=104,Ee=105,Se=106,ke=107,Me=109,Ae=110,Te=111,De=112,Ie=113,Pe=114,Re=115,Ne=116,Fe=117,Le=118,Ve=119,Be=120,ze=121,Ue=122,He=123,qe=144,Ge=145,We=166,Ye=173,Ke=173,Qe=174,Xe=175,Ze=181,$e=182,Je=183,et=183,tt=186,nt=187,rt=188,it=189,ot=191,at=192,st=192,ut=219,ct=220,lt=221,dt=222,pt=224;function ft(e){for(var t=[],n=1;n<arguments.length;n++)t[n-1]=arguments[n];return t.length?t.some(function(t){return e[t]}):e.altKey||e.shiftKey||e.ctrlKey||e.metaKey}},"./node_modules/@angular/cdk/esm5/layout.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"LayoutModule",function(){return c}),n.d(t,"BreakpointObserver",function(){return f}),n.d(t,"Breakpoints",function(){return m}),n.d(t,"MediaMatcher",function(){return d});var r,i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),a=n("./node_modules/rxjs/_esm5/index.js"),s=n("./node_modules/rxjs/_esm5/operators/index.js"),u=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),c=function(){function e(){}return e.decorators=[{type:i.NgModule,args:[{}]}],e}(),l=new Set,d=function(){function e(e){this.platform=e,this._matchMedia=this.platform.isBrowser&&window.matchMedia?window.matchMedia.bind(window):p}return e.prototype.matchMedia=function(e){return this.platform.WEBKIT&&function(e){if(l.has(e))return;try{r||((r=document.createElement("style")).setAttribute("type","text/css"),document.head.appendChild(r)),r.sheet&&(r.sheet.insertRule("@media "+e+" {.fx-query-test{ }}",0),l.add(e))}catch(e){console.error(e)}}(e),this._matchMedia(e)},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:o.Platform}]},e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e(Object(i.inject)(o.Platform))},token:e,providedIn:"root"}),e}();function p(e){return{matches:"all"===e||""===e,media:e,addListener:function(){},removeListener:function(){}}}var f=function(){function e(e,t){this.mediaMatcher=e,this.zone=t,this._queries=new Map,this._destroySubject=new a.Subject}return e.prototype.ngOnDestroy=function(){this._destroySubject.next(),this._destroySubject.complete()},e.prototype.isMatched=function(e){var t=this;return h(Object(u.coerceArray)(e)).some(function(e){return t._registerQuery(e).mql.matches})},e.prototype.observe=function(e){var t=this,n=h(Object(u.coerceArray)(e)).map(function(e){return t._registerQuery(e).observable});return Object(a.combineLatest)(n).pipe(Object(s.debounceTime)(0,a.asapScheduler),Object(s.map)(function(e){var t={matches:!1,breakpoints:{}};return e.forEach(function(e){t.matches=t.matches||e.matches,t.breakpoints[e.query]=e.matches}),t}))},e.prototype._registerQuery=function(e){var t=this;if(this._queries.has(e))return this._queries.get(e);var n,r=this.mediaMatcher.matchMedia(e),i={observable:Object(a.fromEventPattern)(function(e){n=function(n){return t.zone.run(function(){return e(n)})},r.addListener(n)},function(){return r.removeListener(n)}).pipe(Object(s.startWith)(r),Object(s.map)(function(t){return{query:e,matches:t.matches}}),Object(s.takeUntil)(this._destroySubject)),mql:r};return this._queries.set(e,i),i},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:d},{type:i.NgZone}]},e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e(Object(i.inject)(d),Object(i.inject)(i.NgZone))},token:e,providedIn:"root"}),e}();function h(e){return e.map(function(e){return e.split(",")}).reduce(function(e,t){return e.concat(t)}).map(function(e){return e.trim()})}var m={XSmall:"(max-width: 599.99px)",Small:"(min-width: 600px) and (max-width: 959.99px)",Medium:"(min-width: 960px) and (max-width: 1279.99px)",Large:"(min-width: 1280px) and (max-width: 1919.99px)",XLarge:"(min-width: 1920px)",Handset:"(max-width: 599.99px) and (orientation: portrait), (max-width: 959.99px) and (orientation: landscape)",Tablet:"(min-width: 600px) and (max-width: 839.99px) and (orientation: portrait), (min-width: 960px) and (max-width: 1279.99px) and (orientation: landscape)",Web:"(min-width: 840px) and (orientation: portrait), (min-width: 1280px) and (orientation: landscape)",HandsetPortrait:"(max-width: 599.99px) and (orientation: portrait)",TabletPortrait:"(min-width: 600px) and (max-width: 839.99px) and (orientation: portrait)",WebPortrait:"(min-width: 840px) and (orientation: portrait)",HandsetLandscape:"(max-width: 959.99px) and (orientation: landscape)",TabletLandscape:"(min-width: 960px) and (max-width: 1279.99px) and (orientation: landscape)",WebLandscape:"(min-width: 1280px) and (orientation: landscape)"}},"./node_modules/@angular/cdk/esm5/observers.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MutationObserverFactory",function(){return s}),n.d(t,"ContentObserver",function(){return u}),n.d(t,"CdkObserveContent",function(){return c}),n.d(t,"ObserversModule",function(){return l});var r=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/rxjs/_esm5/index.js"),a=n("./node_modules/rxjs/_esm5/operators/index.js"),s=function(){function e(){}return e.prototype.create=function(e){return"undefined"==typeof MutationObserver?null:new MutationObserver(e)},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(),u=function(){function e(e){this._mutationObserverFactory=e,this._observedElements=new Map}return e.prototype.ngOnDestroy=function(){var e=this;this._observedElements.forEach(function(t,n){return e._cleanupObserver(n)})},e.prototype.observe=function(e){var t=this,n=Object(r.coerceElement)(e);return o.Observable.create(function(e){var r=t._observeElement(n).subscribe(e);return function(){r.unsubscribe(),t._unobserveElement(n)}})},e.prototype._observeElement=function(e){if(this._observedElements.has(e))this._observedElements.get(e).count++;else{var t=new o.Subject,n=this._mutationObserverFactory.create(function(e){return t.next(e)});n&&n.observe(e,{characterData:!0,childList:!0,subtree:!0}),this._observedElements.set(e,{observer:n,stream:t,count:1})}return this._observedElements.get(e).stream},e.prototype._unobserveElement=function(e){this._observedElements.has(e)&&(this._observedElements.get(e).count--,this._observedElements.get(e).count||this._cleanupObserver(e))},e.prototype._cleanupObserver=function(e){if(this._observedElements.has(e)){var t=this._observedElements.get(e),n=t.observer,r=t.stream;n&&n.disconnect(),r.complete(),this._observedElements.delete(e)}},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:s}]},e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e(Object(i.inject)(s))},token:e,providedIn:"root"}),e}(),c=function(){function e(e,t,n){this._contentObserver=e,this._elementRef=t,this._ngZone=n,this.event=new i.EventEmitter,this._disabled=!1,this._currentSubscription=null}return Object.defineProperty(e.prototype,"disabled",{get:function(){return this._disabled},set:function(e){this._disabled=Object(r.coerceBooleanProperty)(e),this._disabled?this._unsubscribe():this._subscribe()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"debounce",{get:function(){return this._debounce},set:function(e){this._debounce=Object(r.coerceNumberProperty)(e),this._subscribe()},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){this._currentSubscription||this.disabled||this._subscribe()},e.prototype.ngOnDestroy=function(){this._unsubscribe()},e.prototype._subscribe=function(){var e=this;this._unsubscribe();var t=this._contentObserver.observe(this._elementRef);this._ngZone.runOutsideAngular(function(){e._currentSubscription=(e.debounce?t.pipe(Object(a.debounceTime)(e.debounce)):t).subscribe(e.event)})},e.prototype._unsubscribe=function(){this._currentSubscription&&this._currentSubscription.unsubscribe()},e.decorators=[{type:i.Directive,args:[{selector:"[cdkObserveContent]",exportAs:"cdkObserveContent"}]}],e.ctorParameters=function(){return[{type:u},{type:i.ElementRef},{type:i.NgZone}]},e.propDecorators={event:[{type:i.Output,args:["cdkObserveContent"]}],disabled:[{type:i.Input,args:["cdkObserveContentDisabled"]}],debounce:[{type:i.Input}]},e}(),l=function(){function e(){}return e.decorators=[{type:i.NgModule,args:[{exports:[c],declarations:[c],providers:[s]}]}],e}()},"./node_modules/@angular/cdk/esm5/overlay.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Overlay",function(){return H}),n.d(t,"OverlayContainer",function(){return T}),n.d(t,"CdkOverlayOrigin",function(){return W}),n.d(t,"CdkConnectedOverlay",function(){return Y}),n.d(t,"FullscreenOverlayContainer",function(){return $}),n.d(t,"OverlayRef",function(){return P}),n.d(t,"OverlayKeyboardDispatcher",function(){return k}),n.d(t,"OverlayPositionBuilder",function(){return z}),n.d(t,"GlobalPositionStrategy",function(){return B}),n.d(t,"ConnectedPositionStrategy",function(){return L}),n.d(t,"FlexibleConnectedPositionStrategy",function(){return N}),n.d(t,"OverlayConfig",function(){return x}),n.d(t,"validateVerticalPosition",function(){return E}),n.d(t,"validateHorizontalPosition",function(){return S}),n.d(t,"ConnectionPositionPair",function(){return O}),n.d(t,"ScrollingVisibility",function(){return w}),n.d(t,"ConnectedOverlayPositionChange",function(){return C}),n.d(t,"ScrollStrategyOptions",function(){return j}),n.d(t,"RepositionScrollStrategy",function(){return b}),n.d(t,"CloseScrollStrategy",function(){return g}),n.d(t,"NoopScrollStrategy",function(){return v}),n.d(t,"BlockScrollStrategy",function(){return h}),n.d(t,"OverlayModule",function(){return X}),n.d(t,"OVERLAY_PROVIDERS",function(){return Z}),n.d(t,"ɵg",function(){return A}),n.d(t,"ɵf",function(){return M}),n.d(t,"ɵb",function(){return I}),n.d(t,"ɵa",function(){return D}),n.d(t,"ɵc",function(){return G}),n.d(t,"ɵe",function(){return Q}),n.d(t,"ɵd",function(){return K});var r=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),i=n("./node_modules/@angular/cdk/esm5/scrolling.es5.js");n.d(t,"ViewportRuler",function(){return i.ViewportRuler}),n.d(t,"VIEWPORT_RULER_PROVIDER",function(){return i.VIEWPORT_RULER_PROVIDER}),n.d(t,"CdkScrollable",function(){return i.CdkScrollable}),n.d(t,"ScrollDispatcher",function(){return i.ScrollDispatcher});var o=n("./node_modules/@angular/common/fesm5/common.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),s=n("./node_modules/tslib/tslib.es6.js"),u=n("./node_modules/rxjs/_esm5/index.js"),c=n("./node_modules/rxjs/_esm5/operators/index.js"),l=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),d=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),p=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),f=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),h=function(){function e(e,t){this._viewportRuler=e,this._previousHTMLStyles={top:"",left:""},this._isEnabled=!1,this._document=t}return e.prototype.attach=function(){},e.prototype.enable=function(){if(this._canBeEnabled()){var e=this._document.documentElement;this._previousScrollPosition=this._viewportRuler.getViewportScrollPosition(),this._previousHTMLStyles.left=e.style.left||"",this._previousHTMLStyles.top=e.style.top||"",e.style.left=Object(r.coerceCssPixelValue)(-this._previousScrollPosition.left),e.style.top=Object(r.coerceCssPixelValue)(-this._previousScrollPosition.top),e.classList.add("cdk-global-scrollblock"),this._isEnabled=!0}},e.prototype.disable=function(){if(this._isEnabled){var e=this._document.documentElement,t=this._document.body,n=e.style,r=t.style,i=n.scrollBehavior||"",o=r.scrollBehavior||"";this._isEnabled=!1,n.left=this._previousHTMLStyles.left,n.top=this._previousHTMLStyles.top,e.classList.remove("cdk-global-scrollblock"),n.scrollBehavior=r.scrollBehavior="auto",window.scroll(this._previousScrollPosition.left,this._previousScrollPosition.top),n.scrollBehavior=i,r.scrollBehavior=o}},e.prototype._canBeEnabled=function(){if(this._document.documentElement.classList.contains("cdk-global-scrollblock")||this._isEnabled)return!1;var e=this._document.body,t=this._viewportRuler.getViewportSize();return e.scrollHeight>t.height||e.scrollWidth>t.width},e}();function m(){return Error("Scroll strategy has already been attached.")}var g=function(){function e(e,t,n,r){var i=this;this._scrollDispatcher=e,this._ngZone=t,this._viewportRuler=n,this._config=r,this._scrollSubscription=null,this._detach=function(){i.disable(),i._overlayRef.hasAttached()&&i._ngZone.run(function(){return i._overlayRef.detach()})}}return e.prototype.attach=function(e){if(this._overlayRef)throw m();this._overlayRef=e},e.prototype.enable=function(){var e=this;if(!this._scrollSubscription){var t=this._scrollDispatcher.scrolled(0);this._config&&this._config.threshold&&this._config.threshold>1?(this._initialScrollPosition=this._viewportRuler.getViewportScrollPosition().top,this._scrollSubscription=t.subscribe(function(){var t=e._viewportRuler.getViewportScrollPosition().top;Math.abs(t-e._initialScrollPosition)>e._config.threshold?e._detach():e._overlayRef.updatePosition()})):this._scrollSubscription=t.subscribe(this._detach)}},e.prototype.disable=function(){this._scrollSubscription&&(this._scrollSubscription.unsubscribe(),this._scrollSubscription=null)},e}(),v=function(){function e(){}return e.prototype.enable=function(){},e.prototype.disable=function(){},e.prototype.attach=function(){},e}();function _(e,t){return t.some(function(t){var n=e.bottom<t.top,r=e.top>t.bottom,i=e.right<t.left,o=e.left>t.right;return n||r||i||o})}function y(e,t){return t.some(function(t){var n=e.top<t.top,r=e.bottom>t.bottom,i=e.left<t.left,o=e.right>t.right;return n||r||i||o})}var b=function(){function e(e,t,n,r){this._scrollDispatcher=e,this._viewportRuler=t,this._ngZone=n,this._config=r,this._scrollSubscription=null}return e.prototype.attach=function(e){if(this._overlayRef)throw m();this._overlayRef=e},e.prototype.enable=function(){var e=this;if(!this._scrollSubscription){var t=this._config?this._config.scrollThrottle:0;this._scrollSubscription=this._scrollDispatcher.scrolled(t).subscribe(function(){if(e._overlayRef.updatePosition(),e._config&&e._config.autoClose){var t=e._overlayRef.overlayElement.getBoundingClientRect(),n=e._viewportRuler.getViewportSize(),r=n.width,i=n.height;_(t,[{width:r,height:i,bottom:i,right:r,top:0,left:0}])&&(e.disable(),e._ngZone.run(function(){return e._overlayRef.detach()}))}})}},e.prototype.disable=function(){this._scrollSubscription&&(this._scrollSubscription.unsubscribe(),this._scrollSubscription=null)},e}(),j=function(){function e(e,t,n,r){var i=this;this._scrollDispatcher=e,this._viewportRuler=t,this._ngZone=n,this.noop=function(){return new v},this.close=function(e){return new g(i._scrollDispatcher,i._ngZone,i._viewportRuler,e)},this.block=function(){return new h(i._viewportRuler,i._document)},this.reposition=function(e){return new b(i._scrollDispatcher,i._viewportRuler,i._ngZone,e)},this._document=r}return e.decorators=[{type:a.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:i.ScrollDispatcher},{type:i.ViewportRuler},{type:a.NgZone},{type:void 0,decorators:[{type:a.Inject,args:[o.DOCUMENT]}]}]},e.ngInjectableDef=Object(a.defineInjectable)({factory:function(){return new e(Object(a.inject)(i.ScrollDispatcher),Object(a.inject)(i.ViewportRuler),Object(a.inject)(a.NgZone),Object(a.inject)(o.DOCUMENT))},token:e,providedIn:"root"}),e}(),x=function(){return function(e){var t=this;this.scrollStrategy=new v,this.panelClass="",this.hasBackdrop=!1,this.backdropClass="cdk-overlay-dark-backdrop",this.disposeOnNavigation=!1,e&&Object.keys(e).forEach(function(n){var r=n;void 0!==e[r]&&(t[r]=e[r])})}}(),O=function(){return function(e,t,n,r,i){this.offsetX=n,this.offsetY=r,this.panelClass=i,this.originX=e.originX,this.originY=e.originY,this.overlayX=t.overlayX,this.overlayY=t.overlayY}}(),w=function(){return function(){}}(),C=function(){function e(e,t){this.connectionPair=e,this.scrollableViewProperties=t}return e.ctorParameters=function(){return[{type:O},{type:w,decorators:[{type:a.Optional}]}]},e}();function E(e,t){if("top"!==t&&"bottom"!==t&&"center"!==t)throw Error("ConnectedPosition: Invalid "+e+' "'+t+'". Expected "top", "bottom" or "center".')}function S(e,t){if("start"!==t&&"end"!==t&&"center"!==t)throw Error("ConnectedPosition: Invalid "+e+' "'+t+'". Expected "start", "end" or "center".')}var k=function(){function e(e){var t=this;this._attachedOverlays=[],this._keydownListener=function(e){for(var n=t._attachedOverlays,r=n.length-1;r>-1;r--)if(n[r]._keydownEventSubscriptions>0){n[r]._keydownEvents.next(e);break}},this._document=e}return e.prototype.ngOnDestroy=function(){this._detach()},e.prototype.add=function(e){this.remove(e),this._isAttached||(this._document.body.addEventListener("keydown",this._keydownListener,!0),this._isAttached=!0),this._attachedOverlays.push(e)},e.prototype.remove=function(e){var t=this._attachedOverlays.indexOf(e);t>-1&&this._attachedOverlays.splice(t,1),0===this._attachedOverlays.length&&this._detach()},e.prototype._detach=function(){this._isAttached&&(this._document.body.removeEventListener("keydown",this._keydownListener,!0),this._isAttached=!1)},e.decorators=[{type:a.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:a.Inject,args:[o.DOCUMENT]}]}]},e.ngInjectableDef=Object(a.defineInjectable)({factory:function(){return new e(Object(a.inject)(o.DOCUMENT))},token:e,providedIn:"root"}),e}();function M(e,t){return e||new k(t)}var A={provide:k,deps:[[new a.Optional,new a.SkipSelf,k],o.DOCUMENT],useFactory:M},T=function(){function e(e){this._document=e}return e.prototype.ngOnDestroy=function(){this._containerElement&&this._containerElement.parentNode&&this._containerElement.parentNode.removeChild(this._containerElement)},e.prototype.getContainerElement=function(){return this._containerElement||this._createContainer(),this._containerElement},e.prototype._createContainer=function(){var e=this._document.createElement("div");e.classList.add("cdk-overlay-container"),this._document.body.appendChild(e),this._containerElement=e},e.decorators=[{type:a.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:a.Inject,args:[o.DOCUMENT]}]}]},e.ngInjectableDef=Object(a.defineInjectable)({factory:function(){return new e(Object(a.inject)(o.DOCUMENT))},token:e,providedIn:"root"}),e}();function D(e,t){return e||new T(t)}var I={provide:T,deps:[[new a.Optional,new a.SkipSelf,T],o.DOCUMENT],useFactory:D},P=function(){function e(e,t,n,r,i,o,a,s){var c=this;this._portalOutlet=e,this._host=t,this._pane=n,this._config=r,this._ngZone=i,this._keyboardDispatcher=o,this._document=a,this._location=s,this._backdropElement=null,this._backdropClick=new u.Subject,this._attachments=new u.Subject,this._detachments=new u.Subject,this._locationChanges=u.Subscription.EMPTY,this._keydownEventsObservable=u.Observable.create(function(e){var t=c._keydownEvents.subscribe(e);return c._keydownEventSubscriptions++,function(){t.unsubscribe(),c._keydownEventSubscriptions--}}),this._keydownEvents=new u.Subject,this._keydownEventSubscriptions=0,r.scrollStrategy&&r.scrollStrategy.attach(this),this._positionStrategy=r.positionStrategy}return Object.defineProperty(e.prototype,"overlayElement",{get:function(){return this._pane},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"backdropElement",{get:function(){return this._backdropElement},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hostElement",{get:function(){return this._host},enumerable:!0,configurable:!0}),e.prototype.attach=function(e){var t=this,n=this._portalOutlet.attach(e);return this._positionStrategy&&this._positionStrategy.attach(this),!this._host.parentElement&&this._previousHostParent&&this._previousHostParent.appendChild(this._host),this._updateStackingOrder(),this._updateElementSize(),this._updateElementDirection(),this._config.scrollStrategy&&this._config.scrollStrategy.enable(),this._ngZone.onStable.asObservable().pipe(Object(c.take)(1)).subscribe(function(){t.hasAttached()&&t.updatePosition()}),this._togglePointerEvents(!0),this._config.hasBackdrop&&this._attachBackdrop(),this._config.panelClass&&this._toggleClasses(this._pane,this._config.panelClass,!0),this._attachments.next(),this._keyboardDispatcher.add(this),this._config.disposeOnNavigation&&this._location&&(this._locationChanges=this._location.subscribe(function(){return t.dispose()})),n},e.prototype.detach=function(){if(this.hasAttached()){this.detachBackdrop(),this._togglePointerEvents(!1),this._positionStrategy&&this._positionStrategy.detach&&this._positionStrategy.detach(),this._config.scrollStrategy&&this._config.scrollStrategy.disable();var e=this._portalOutlet.detach();return this._detachments.next(),this._keyboardDispatcher.remove(this),this._detachContentWhenStable(),this._locationChanges.unsubscribe(),e}},e.prototype.dispose=function(){var e=this.hasAttached();this._positionStrategy&&this._positionStrategy.dispose(),this._config.scrollStrategy&&this._config.scrollStrategy.disable(),this.detachBackdrop(),this._locationChanges.unsubscribe(),this._keyboardDispatcher.remove(this),this._portalOutlet.dispose(),this._attachments.complete(),this._backdropClick.complete(),this._keydownEvents.complete(),this._host&&this._host.parentNode&&(this._host.parentNode.removeChild(this._host),this._host=null),this._previousHostParent=this._pane=null,e&&this._detachments.next(),this._detachments.complete()},e.prototype.hasAttached=function(){return this._portalOutlet.hasAttached()},e.prototype.backdropClick=function(){return this._backdropClick.asObservable()},e.prototype.attachments=function(){return this._attachments.asObservable()},e.prototype.detachments=function(){return this._detachments.asObservable()},e.prototype.keydownEvents=function(){return this._keydownEventsObservable},e.prototype.getConfig=function(){return this._config},e.prototype.updatePosition=function(){this._positionStrategy&&this._positionStrategy.apply()},e.prototype.updatePositionStrategy=function(e){e!==this._positionStrategy&&(this._positionStrategy&&this._positionStrategy.dispose(),this._positionStrategy=e,this.hasAttached()&&(e.attach(this),this.updatePosition()))},e.prototype.updateSize=function(e){this._config=Object(s.__assign)({},this._config,e),this._updateElementSize()},e.prototype.setDirection=function(e){this._config=Object(s.__assign)({},this._config,{direction:e}),this._updateElementDirection()},e.prototype.getDirection=function(){var e=this._config.direction;return e?"string"==typeof e?e:e.value:"ltr"},e.prototype._updateElementDirection=function(){this._host.setAttribute("dir",this.getDirection())},e.prototype._updateElementSize=function(){var e=this._pane.style;e.width=Object(r.coerceCssPixelValue)(this._config.width),e.height=Object(r.coerceCssPixelValue)(this._config.height),e.minWidth=Object(r.coerceCssPixelValue)(this._config.minWidth),e.minHeight=Object(r.coerceCssPixelValue)(this._config.minHeight),e.maxWidth=Object(r.coerceCssPixelValue)(this._config.maxWidth),e.maxHeight=Object(r.coerceCssPixelValue)(this._config.maxHeight)},e.prototype._togglePointerEvents=function(e){this._pane.style.pointerEvents=e?"auto":"none"},e.prototype._attachBackdrop=function(){var e=this,t="cdk-overlay-backdrop-showing";this._backdropElement=this._document.createElement("div"),this._backdropElement.classList.add("cdk-overlay-backdrop"),this._config.backdropClass&&this._toggleClasses(this._backdropElement,this._config.backdropClass,!0),this._host.parentElement.insertBefore(this._backdropElement,this._host),this._backdropElement.addEventListener("click",function(t){return e._backdropClick.next(t)}),"undefined"!=typeof requestAnimationFrame?this._ngZone.runOutsideAngular(function(){requestAnimationFrame(function(){e._backdropElement&&e._backdropElement.classList.add(t)})}):this._backdropElement.classList.add(t)},e.prototype._updateStackingOrder=function(){this._host.nextSibling&&this._host.parentNode.appendChild(this._host)},e.prototype.detachBackdrop=function(){var e=this,t=this._backdropElement;if(t){var n,r=function(){t&&t.parentNode&&t.parentNode.removeChild(t),e._backdropElement==t&&(e._backdropElement=null),e._config.backdropClass&&e._toggleClasses(t,e._config.backdropClass,!1),clearTimeout(n)};t.classList.remove("cdk-overlay-backdrop-showing"),this._ngZone.runOutsideAngular(function(){t.addEventListener("transitionend",r)}),t.style.pointerEvents="none",n=this._ngZone.runOutsideAngular(function(){return setTimeout(r,500)})}},e.prototype._toggleClasses=function(e,t,n){var i=e.classList;Object(r.coerceArray)(t).forEach(function(e){n?i.add(e):i.remove(e)})},e.prototype._detachContentWhenStable=function(){var e=this;this._ngZone.runOutsideAngular(function(){var t=e._ngZone.onStable.asObservable().pipe(Object(c.takeUntil)(Object(u.merge)(e._attachments,e._detachments))).subscribe(function(){e._pane&&e._host&&0!==e._pane.children.length||(e._pane&&e._config.panelClass&&e._toggleClasses(e._pane,e._config.panelClass,!1),e._host&&e._host.parentElement&&(e._previousHostParent=e._host.parentElement,e._previousHostParent.removeChild(e._host)),t.unsubscribe())})})},e}(),R="cdk-overlay-connected-position-bounding-box",N=function(){function e(e,t,n,r,i){var o=this;this._viewportRuler=t,this._document=n,this._platform=r,this._overlayContainer=i,this._lastBoundingBoxSize={width:0,height:0},this._isPushed=!1,this._canPush=!0,this._growAfterOpen=!1,this._hasFlexibleDimensions=!0,this._positionLocked=!1,this._viewportMargin=0,this.scrollables=[],this._preferredPositions=[],this._positionChanges=new u.Subject,this._resizeSubscription=u.Subscription.EMPTY,this._offsetX=0,this._offsetY=0,this._positionChangeSubscriptions=0,this._appliedPanelClasses=[],this.positionChanges=u.Observable.create(function(e){var t=o._positionChanges.subscribe(e);return o._positionChangeSubscriptions++,function(){t.unsubscribe(),o._positionChangeSubscriptions--}}),this.setOrigin(e)}return Object.defineProperty(e.prototype,"positions",{get:function(){return this._preferredPositions},enumerable:!0,configurable:!0}),e.prototype.attach=function(e){var t=this;if(this._overlayRef&&e!==this._overlayRef)throw Error("This position strategy is already attached to an overlay");this._validatePositions(),e.hostElement.classList.add(R),this._overlayRef=e,this._boundingBox=e.hostElement,this._pane=e.overlayElement,this._isDisposed=!1,this._isInitialRender=!0,this._lastPosition=null,this._resizeSubscription.unsubscribe(),this._resizeSubscription=this._viewportRuler.change().subscribe(function(){t._isInitialRender=!0,t.apply()})},e.prototype.apply=function(){if(!(this._isDisposed||this._platform&&!this._platform.isBrowser))if(!this._isInitialRender&&this._positionLocked&&this._lastPosition)this.reapplyLastPosition();else{this._clearPanelClasses(),this._resetOverlayElementStyles(),this._resetBoundingBoxStyles(),this._viewportRect=this._getNarrowedViewportRect(),this._originRect=this._origin.getBoundingClientRect(),this._overlayRect=this._pane.getBoundingClientRect();for(var e,t=this._originRect,n=this._overlayRect,r=this._viewportRect,i=[],o=0,a=this._preferredPositions;o<a.length;o++){var s=a[o],u=this._getOriginPoint(t,s),c=this._getOverlayPoint(u,n,s),l=this._getOverlayFit(c,n,r,s);if(l.isCompletelyWithinViewport)return this._isPushed=!1,void this._applyPosition(s,u);this._canFitWithFlexibleDimensions(l,c,r)?i.push({position:s,origin:u,overlayRect:n,boundingBoxRect:this._calculateBoundingBoxRect(u,s)}):(!e||e.overlayFit.visibleArea<l.visibleArea)&&(e={overlayFit:l,overlayPoint:c,originPoint:u,position:s,overlayRect:n})}if(i.length){for(var d=null,p=-1,f=0,h=i;f<h.length;f++){var m=h[f],g=m.boundingBoxRect.width*m.boundingBoxRect.height*(m.position.weight||1);g>p&&(p=g,d=m)}return this._isPushed=!1,void this._applyPosition(d.position,d.origin)}if(this._canPush)return this._isPushed=!0,void this._applyPosition(e.position,e.originPoint);this._applyPosition(e.position,e.originPoint)}},e.prototype.detach=function(){this._clearPanelClasses(),this._lastPosition=null,this._previousPushAmount=null,this._resizeSubscription.unsubscribe()},e.prototype.dispose=function(){this._isDisposed||(this._boundingBox&&F(this._boundingBox.style,{top:"",left:"",right:"",bottom:"",height:"",width:"",alignItems:"",justifyContent:""}),this._pane&&this._resetOverlayElementStyles(),this._overlayRef&&this._overlayRef.hostElement.classList.remove(R),this.detach(),this._positionChanges.complete(),this._overlayRef=this._boundingBox=null,this._isDisposed=!0)},e.prototype.reapplyLastPosition=function(){if(!this._isDisposed&&(!this._platform||this._platform.isBrowser)){this._originRect=this._origin.getBoundingClientRect(),this._overlayRect=this._pane.getBoundingClientRect(),this._viewportRect=this._getNarrowedViewportRect();var e=this._lastPosition||this._preferredPositions[0],t=this._getOriginPoint(this._originRect,e);this._applyPosition(e,t)}},e.prototype.withScrollableContainers=function(e){this.scrollables=e},e.prototype.withPositions=function(e){return this._preferredPositions=e,-1===e.indexOf(this._lastPosition)&&(this._lastPosition=null),this._validatePositions(),this},e.prototype.withViewportMargin=function(e){return this._viewportMargin=e,this},e.prototype.withFlexibleDimensions=function(e){return void 0===e&&(e=!0),this._hasFlexibleDimensions=e,this},e.prototype.withGrowAfterOpen=function(e){return void 0===e&&(e=!0),this._growAfterOpen=e,this},e.prototype.withPush=function(e){return void 0===e&&(e=!0),this._canPush=e,this},e.prototype.withLockedPosition=function(e){return void 0===e&&(e=!0),this._positionLocked=e,this},e.prototype.setOrigin=function(e){return this._origin=Object(r.coerceElement)(e),this},e.prototype.withDefaultOffsetX=function(e){return this._offsetX=e,this},e.prototype.withDefaultOffsetY=function(e){return this._offsetY=e,this},e.prototype.withTransformOriginOn=function(e){return this._transformOriginSelector=e,this},e.prototype._getOriginPoint=function(e,t){var n;if("center"==t.originX)n=e.left+e.width/2;else{var r=this._isRtl()?e.right:e.left,i=this._isRtl()?e.left:e.right;n="start"==t.originX?r:i}return{x:n,y:"center"==t.originY?e.top+e.height/2:"top"==t.originY?e.top:e.bottom}},e.prototype._getOverlayPoint=function(e,t,n){var r,i;return r="center"==n.overlayX?-t.width/2:"start"===n.overlayX?this._isRtl()?-t.width:0:this._isRtl()?0:-t.width,i="center"==n.overlayY?-t.height/2:"top"==n.overlayY?0:-t.height,{x:e.x+r,y:e.y+i}},e.prototype._getOverlayFit=function(e,t,n,r){var i=e.x,o=e.y,a=this._getOffset(r,"x"),s=this._getOffset(r,"y");a&&(i+=a),s&&(o+=s);var u=0-i,c=i+t.width-n.width,l=0-o,d=o+t.height-n.height,p=this._subtractOverflows(t.width,u,c),f=this._subtractOverflows(t.height,l,d),h=p*f;return{visibleArea:h,isCompletelyWithinViewport:t.width*t.height===h,fitsInViewportVertically:f===t.height,fitsInViewportHorizontally:p==t.width}},e.prototype._canFitWithFlexibleDimensions=function(e,t,n){if(this._hasFlexibleDimensions){var r=n.bottom-t.y,i=n.right-t.x,o=this._overlayRef.getConfig().minHeight,a=this._overlayRef.getConfig().minWidth,s=e.fitsInViewportVertically||null!=o&&o<=r,u=e.fitsInViewportHorizontally||null!=a&&a<=i;return s&&u}},e.prototype._pushOverlayOnScreen=function(e,t,n){if(this._previousPushAmount&&this._positionLocked)return{x:e.x+this._previousPushAmount.x,y:e.y+this._previousPushAmount.y};var r=this._viewportRect,i=Math.max(e.x+t.width-r.right,0),o=Math.max(e.y+t.height-r.bottom,0),a=Math.max(r.top-n.top-e.y,0),s=Math.max(r.left-n.left-e.x,0),u=0,c=0;return u=t.width<r.width?s||-i:e.x<this._viewportMargin?r.left-n.left-e.x:0,c=t.height<r.height?a||-o:e.y<this._viewportMargin?r.top-n.top-e.y:0,this._previousPushAmount={x:u,y:c},{x:e.x+u,y:e.y+c}},e.prototype._applyPosition=function(e,t){if(this._setTransformOrigin(e),this._setOverlayElementStyles(t,e),this._setBoundingBoxStyles(t,e),e.panelClass&&this._addPanelClasses(e.panelClass),this._lastPosition=e,this._positionChangeSubscriptions>0){var n=this._getScrollVisibility(),r=new C(e,n);this._positionChanges.next(r)}this._isInitialRender=!1},e.prototype._setTransformOrigin=function(e){if(this._transformOriginSelector){var t,n=this._boundingBox.querySelectorAll(this._transformOriginSelector),r=e.overlayY;t="center"===e.overlayX?"center":this._isRtl()?"start"===e.overlayX?"right":"left":"start"===e.overlayX?"left":"right";for(var i=0;i<n.length;i++)n[i].style.transformOrigin=t+" "+r}},e.prototype._calculateBoundingBoxRect=function(e,t){var n,r,i,o=this._viewportRect,a=this._isRtl();if("top"===t.overlayY)r=e.y,n=o.bottom-e.y;else if("bottom"===t.overlayY)i=o.height-e.y+2*this._viewportMargin,n=o.height-i+this._viewportMargin;else{var s=Math.min(o.bottom-e.y+o.top,e.y),u=this._lastBoundingBoxSize.height;n=2*s,r=e.y-s,n>u&&!this._isInitialRender&&!this._growAfterOpen&&(r=e.y-u/2)}var c,l,d,p="start"===t.overlayX&&!a||"end"===t.overlayX&&a;if("end"===t.overlayX&&!a||"start"===t.overlayX&&a)d=o.right-e.x+this._viewportMargin,c=e.x-o.left;else if(p)l=e.x,c=o.right-e.x;else{s=Math.min(o.right-e.x+o.left,e.x);var f=this._lastBoundingBoxSize.width;c=2*s,l=e.x-s,c>f&&!this._isInitialRender&&!this._growAfterOpen&&(l=e.x-f/2)}return{top:r,left:l,bottom:i,right:d,width:c,height:n}},e.prototype._setBoundingBoxStyles=function(e,t){var n=this._calculateBoundingBoxRect(e,t);this._isInitialRender||this._growAfterOpen||(n.height=Math.min(n.height,this._lastBoundingBoxSize.height),n.width=Math.min(n.width,this._lastBoundingBoxSize.width));var i={};if(this._hasExactPosition())i.top=i.left="0",i.bottom=i.right="",i.width=i.height="100%";else{var o=this._overlayRef.getConfig().maxHeight,a=this._overlayRef.getConfig().maxWidth;i.height=Object(r.coerceCssPixelValue)(n.height),i.top=Object(r.coerceCssPixelValue)(n.top),i.bottom=Object(r.coerceCssPixelValue)(n.bottom),i.width=Object(r.coerceCssPixelValue)(n.width),i.left=Object(r.coerceCssPixelValue)(n.left),i.right=Object(r.coerceCssPixelValue)(n.right),"center"===t.overlayX?i.alignItems="center":i.alignItems="end"===t.overlayX?"flex-end":"flex-start","center"===t.overlayY?i.justifyContent="center":i.justifyContent="bottom"===t.overlayY?"flex-end":"flex-start",o&&(i.maxHeight=Object(r.coerceCssPixelValue)(o)),a&&(i.maxWidth=Object(r.coerceCssPixelValue)(a))}this._lastBoundingBoxSize=n,F(this._boundingBox.style,i)},e.prototype._resetBoundingBoxStyles=function(){F(this._boundingBox.style,{top:"0",left:"0",right:"0",bottom:"0",height:"",width:"",alignItems:"",justifyContent:""})},e.prototype._resetOverlayElementStyles=function(){F(this._pane.style,{top:"",left:"",bottom:"",right:"",position:""})},e.prototype._setOverlayElementStyles=function(e,t){var n={};if(this._hasExactPosition()){var r=this._viewportRuler.getViewportScrollPosition();F(n,this._getExactOverlayY(t,e,r)),F(n,this._getExactOverlayX(t,e,r))}else n.position="static";var i="",o=this._getOffset(t,"x"),a=this._getOffset(t,"y");o&&(i+="translateX("+o+"px) "),a&&(i+="translateY("+a+"px)"),n.transform=i.trim(),this._hasFlexibleDimensions&&this._overlayRef.getConfig().maxHeight&&(n.maxHeight=""),this._hasFlexibleDimensions&&this._overlayRef.getConfig().maxWidth&&(n.maxWidth=""),F(this._pane.style,n)},e.prototype._getExactOverlayY=function(e,t,n){var i={top:null,bottom:null},o=this._getOverlayPoint(t,this._overlayRect,e);this._isPushed&&(o=this._pushOverlayOnScreen(o,this._overlayRect,n));var a=this._overlayContainer?this._overlayContainer.getContainerElement().getBoundingClientRect().top:0;if(o.y-=a,"bottom"===e.overlayY){var s=this._document.documentElement.clientHeight;i.bottom=s-(o.y+this._overlayRect.height)+"px"}else i.top=Object(r.coerceCssPixelValue)(o.y);return i},e.prototype._getExactOverlayX=function(e,t,n){var i={left:null,right:null},o=this._getOverlayPoint(t,this._overlayRect,e);if(this._isPushed&&(o=this._pushOverlayOnScreen(o,this._overlayRect,n)),"right"===(this._isRtl()?"end"===e.overlayX?"left":"right":"end"===e.overlayX?"right":"left")){var a=this._document.documentElement.clientWidth;i.right=a-(o.x+this._overlayRect.width)+"px"}else i.left=Object(r.coerceCssPixelValue)(o.x);return i},e.prototype._getScrollVisibility=function(){var e=this._origin.getBoundingClientRect(),t=this._pane.getBoundingClientRect(),n=this.scrollables.map(function(e){return e.getElementRef().nativeElement.getBoundingClientRect()});return{isOriginClipped:y(e,n),isOriginOutsideView:_(e,n),isOverlayClipped:y(t,n),isOverlayOutsideView:_(t,n)}},e.prototype._subtractOverflows=function(e){for(var t=[],n=1;n<arguments.length;n++)t[n-1]=arguments[n];return t.reduce(function(e,t){return e-Math.max(t,0)},e)},e.prototype._getNarrowedViewportRect=function(){var e=this._document.documentElement.clientWidth,t=this._document.documentElement.clientHeight,n=this._viewportRuler.getViewportScrollPosition();return{top:n.top+this._viewportMargin,left:n.left+this._viewportMargin,right:n.left+e-this._viewportMargin,bottom:n.top+t-this._viewportMargin,width:e-2*this._viewportMargin,height:t-2*this._viewportMargin}},e.prototype._isRtl=function(){return"rtl"===this._overlayRef.getDirection()},e.prototype._hasExactPosition=function(){return!this._hasFlexibleDimensions||this._isPushed},e.prototype._getOffset=function(e,t){return"x"===t?null==e.offsetX?this._offsetX:e.offsetX:null==e.offsetY?this._offsetY:e.offsetY},e.prototype._validatePositions=function(){if(!this._preferredPositions.length)throw Error("FlexibleConnectedPositionStrategy: At least one position is required.");this._preferredPositions.forEach(function(e){S("originX",e.originX),E("originY",e.originY),S("overlayX",e.overlayX),E("overlayY",e.overlayY)})},e.prototype._addPanelClasses=function(e){var t=this;this._pane&&Object(r.coerceArray)(e).forEach(function(e){-1===t._appliedPanelClasses.indexOf(e)&&(t._appliedPanelClasses.push(e),t._pane.classList.add(e))})},e.prototype._clearPanelClasses=function(){var e=this;this._pane&&(this._appliedPanelClasses.forEach(function(t){return e._pane.classList.remove(t)}),this._appliedPanelClasses=[])},e}();function F(e,t){for(var n in t)t.hasOwnProperty(n)&&(e[n]=t[n]);return e}var L=function(){function e(e,t,n,r,i,o){this._preferredPositions=[],this._positionStrategy=new N(n,r,i,o).withFlexibleDimensions(!1).withPush(!1).withViewportMargin(0),this.withFallbackPosition(e,t)}return Object.defineProperty(e.prototype,"_isRtl",{get:function(){return"rtl"===this._overlayRef.getDirection()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"onPositionChange",{get:function(){return this._positionStrategy.positionChanges},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"positions",{get:function(){return this._preferredPositions},enumerable:!0,configurable:!0}),e.prototype.attach=function(e){this._overlayRef=e,this._positionStrategy.attach(e),this._direction&&(e.setDirection(this._direction),this._direction=null)},e.prototype.dispose=function(){this._positionStrategy.dispose()},e.prototype.detach=function(){this._positionStrategy.detach()},e.prototype.apply=function(){this._positionStrategy.apply()},e.prototype.recalculateLastPosition=function(){this._positionStrategy.reapplyLastPosition()},e.prototype.withScrollableContainers=function(e){this._positionStrategy.withScrollableContainers(e)},e.prototype.withFallbackPosition=function(e,t,n,r){var i=new O(e,t,n,r);return this._preferredPositions.push(i),this._positionStrategy.withPositions(this._preferredPositions),this},e.prototype.withDirection=function(e){return this._overlayRef?this._overlayRef.setDirection(e):this._direction=e,this},e.prototype.withOffsetX=function(e){return this._positionStrategy.withDefaultOffsetX(e),this},e.prototype.withOffsetY=function(e){return this._positionStrategy.withDefaultOffsetY(e),this},e.prototype.withLockedPosition=function(e){return this._positionStrategy.withLockedPosition(e),this},e.prototype.withPositions=function(e){return this._preferredPositions=e.slice(),this._positionStrategy.withPositions(this._preferredPositions),this},e.prototype.setOrigin=function(e){return this._positionStrategy.setOrigin(e),this},e}(),V="cdk-global-overlay-wrapper",B=function(){function e(){this._cssPosition="static",this._topOffset="",this._bottomOffset="",this._leftOffset="",this._rightOffset="",this._alignItems="",this._justifyContent="",this._width="",this._height=""}return e.prototype.attach=function(e){var t=e.getConfig();this._overlayRef=e,this._width&&!t.width&&e.updateSize({width:this._width}),this._height&&!t.height&&e.updateSize({height:this._height}),e.hostElement.classList.add(V),this._isDisposed=!1},e.prototype.top=function(e){return void 0===e&&(e=""),this._bottomOffset="",this._topOffset=e,this._alignItems="flex-start",this},e.prototype.left=function(e){return void 0===e&&(e=""),this._rightOffset="",this._leftOffset=e,this._justifyContent="flex-start",this},e.prototype.bottom=function(e){return void 0===e&&(e=""),this._topOffset="",this._bottomOffset=e,this._alignItems="flex-end",this},e.prototype.right=function(e){return void 0===e&&(e=""),this._leftOffset="",this._rightOffset=e,this._justifyContent="flex-end",this},e.prototype.width=function(e){return void 0===e&&(e=""),this._overlayRef?this._overlayRef.updateSize({width:e}):this._width=e,this},e.prototype.height=function(e){return void 0===e&&(e=""),this._overlayRef?this._overlayRef.updateSize({height:e}):this._height=e,this},e.prototype.centerHorizontally=function(e){return void 0===e&&(e=""),this.left(e),this._justifyContent="center",this},e.prototype.centerVertically=function(e){return void 0===e&&(e=""),this.top(e),this._alignItems="center",this},e.prototype.apply=function(){if(this._overlayRef&&this._overlayRef.hasAttached()){var e=this._overlayRef.overlayElement.style,t=this._overlayRef.hostElement.style,n=this._overlayRef.getConfig();e.position=this._cssPosition,e.marginLeft="100%"===n.width?"0":this._leftOffset,e.marginTop="100%"===n.height?"0":this._topOffset,e.marginBottom=this._bottomOffset,e.marginRight=this._rightOffset,"100%"===n.width?t.justifyContent="flex-start":"center"===this._justifyContent?t.justifyContent="center":"rtl"===this._overlayRef.getConfig().direction?"flex-start"===this._justifyContent?t.justifyContent="flex-end":"flex-end"===this._justifyContent&&(t.justifyContent="flex-start"):t.justifyContent=this._justifyContent,t.alignItems="100%"===n.height?"flex-start":this._alignItems}},e.prototype.dispose=function(){if(!this._isDisposed&&this._overlayRef){var e=this._overlayRef.overlayElement.style,t=this._overlayRef.hostElement,n=t.style;t.classList.remove(V),n.justifyContent=n.alignItems=e.marginTop=e.marginBottom=e.marginLeft=e.marginRight=e.position="",this._overlayRef=null,this._isDisposed=!0}},e}(),z=function(){function e(e,t,n,r){this._viewportRuler=e,this._document=t,this._platform=n,this._overlayContainer=r}return e.prototype.global=function(){return new B},e.prototype.connectedTo=function(e,t,n){return new L(t,n,e,this._viewportRuler,this._document)},e.prototype.flexibleConnectedTo=function(e){return new N(e,this._viewportRuler,this._document,this._platform,this._overlayContainer)},e.decorators=[{type:a.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:i.ViewportRuler},{type:void 0,decorators:[{type:a.Inject,args:[o.DOCUMENT]}]},{type:l.Platform,decorators:[{type:a.Optional}]},{type:T,decorators:[{type:a.Optional}]}]},e.ngInjectableDef=Object(a.defineInjectable)({factory:function(){return new e(Object(a.inject)(i.ViewportRuler),Object(a.inject)(o.DOCUMENT),Object(a.inject)(l.Platform,8),Object(a.inject)(T,8))},token:e,providedIn:"root"}),e}(),U=0,H=function(){function e(e,t,n,r,i,o,a,s,u,c){this.scrollStrategies=e,this._overlayContainer=t,this._componentFactoryResolver=n,this._positionBuilder=r,this._keyboardDispatcher=i,this._injector=o,this._ngZone=a,this._document=s,this._directionality=u,this._location=c}return e.prototype.create=function(e){var t=this._createHostElement(),n=this._createPaneElement(t),r=this._createPortalOutlet(n),i=new x(e);return i.direction=i.direction||this._directionality.value,new P(r,t,n,i,this._ngZone,this._keyboardDispatcher,this._document,this._location)},e.prototype.position=function(){return this._positionBuilder},e.prototype._createPaneElement=function(e){var t=this._document.createElement("div");return t.id="cdk-overlay-"+U++,t.classList.add("cdk-overlay-pane"),e.appendChild(t),t},e.prototype._createHostElement=function(){var e=this._document.createElement("div");return this._overlayContainer.getContainerElement().appendChild(e),e},e.prototype._createPortalOutlet=function(e){return this._appRef||(this._appRef=this._injector.get(a.ApplicationRef)),new p.DomPortalOutlet(e,this._componentFactoryResolver,this._appRef,this._injector)},e.decorators=[{type:a.Injectable}],e.ctorParameters=function(){return[{type:j},{type:T},{type:a.ComponentFactoryResolver},{type:z},{type:k},{type:a.Injector},{type:a.NgZone},{type:void 0,decorators:[{type:a.Inject,args:[o.DOCUMENT]}]},{type:d.Directionality},{type:o.Location,decorators:[{type:a.Optional}]}]},e}(),q=[{originX:"start",originY:"bottom",overlayX:"start",overlayY:"top"},{originX:"start",originY:"top",overlayX:"start",overlayY:"bottom"},{originX:"end",originY:"top",overlayX:"end",overlayY:"bottom"},{originX:"end",originY:"bottom",overlayX:"end",overlayY:"top"}],G=new a.InjectionToken("cdk-connected-overlay-scroll-strategy"),W=function(){function e(e){this.elementRef=e}return e.decorators=[{type:a.Directive,args:[{selector:"[cdk-overlay-origin], [overlay-origin], [cdkOverlayOrigin]",exportAs:"cdkOverlayOrigin"}]}],e.ctorParameters=function(){return[{type:a.ElementRef}]},e}(),Y=function(){function e(e,t,n,r,i){this._overlay=e,this._dir=i,this._hasBackdrop=!1,this._lockPosition=!1,this._growAfterOpen=!1,this._flexibleDimensions=!1,this._push=!1,this._backdropSubscription=u.Subscription.EMPTY,this.viewportMargin=0,this.open=!1,this.backdropClick=new a.EventEmitter,this.positionChange=new a.EventEmitter,this.attach=new a.EventEmitter,this.detach=new a.EventEmitter,this.overlayKeydown=new a.EventEmitter,this._templatePortal=new p.TemplatePortal(t,n),this._scrollStrategyFactory=r,this.scrollStrategy=this._scrollStrategyFactory()}return Object.defineProperty(e.prototype,"offsetX",{get:function(){return this._offsetX},set:function(e){this._offsetX=e,this._position&&this._updatePositionStrategy(this._position)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"offsetY",{get:function(){return this._offsetY},set:function(e){this._offsetY=e,this._position&&this._updatePositionStrategy(this._position)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hasBackdrop",{get:function(){return this._hasBackdrop},set:function(e){this._hasBackdrop=Object(r.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"lockPosition",{get:function(){return this._lockPosition},set:function(e){this._lockPosition=Object(r.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"flexibleDimensions",{get:function(){return this._flexibleDimensions},set:function(e){this._flexibleDimensions=Object(r.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"growAfterOpen",{get:function(){return this._growAfterOpen},set:function(e){this._growAfterOpen=Object(r.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"push",{get:function(){return this._push},set:function(e){this._push=Object(r.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"overlayRef",{get:function(){return this._overlayRef},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"dir",{get:function(){return this._dir?this._dir.value:"ltr"},enumerable:!0,configurable:!0}),e.prototype.ngOnDestroy=function(){this._overlayRef&&this._overlayRef.dispose(),this._backdropSubscription.unsubscribe()},e.prototype.ngOnChanges=function(e){this._position&&(this._updatePositionStrategy(this._position),e.origin&&this.open&&this._position.apply()),e.open&&(this.open?this._attachOverlay():this._detachOverlay())},e.prototype._createOverlay=function(){var e=this;this.positions&&this.positions.length||(this.positions=q),this._overlayRef=this._overlay.create(this._buildConfig()),this._overlayRef.keydownEvents().subscribe(function(t){e.overlayKeydown.next(t),t.keyCode===f.ESCAPE&&e._detachOverlay()})},e.prototype._buildConfig=function(){var e=this._position=this._createPositionStrategy(),t=new x({direction:this._dir,positionStrategy:e,scrollStrategy:this.scrollStrategy,hasBackdrop:this.hasBackdrop});return(this.width||0===this.width)&&(t.width=this.width),(this.height||0===this.height)&&(t.height=this.height),(this.minWidth||0===this.minWidth)&&(t.minWidth=this.minWidth),(this.minHeight||0===this.minHeight)&&(t.minHeight=this.minHeight),this.backdropClass&&(t.backdropClass=this.backdropClass),this.panelClass&&(t.panelClass=this.panelClass),t},e.prototype._updatePositionStrategy=function(e){var t=this,n=this.positions.map(function(e){return{originX:e.originX,originY:e.originY,overlayX:e.overlayX,overlayY:e.overlayY,offsetX:e.offsetX||t.offsetX,offsetY:e.offsetY||t.offsetY}});return e.setOrigin(this.origin.elementRef).withPositions(n).withFlexibleDimensions(this.flexibleDimensions).withPush(this.push).withGrowAfterOpen(this.growAfterOpen).withViewportMargin(this.viewportMargin).withLockedPosition(this.lockPosition)},e.prototype._createPositionStrategy=function(){var e=this,t=this._overlay.position().flexibleConnectedTo(this.origin.elementRef);return this._updatePositionStrategy(t),t.positionChanges.subscribe(function(t){return e.positionChange.emit(t)}),t},e.prototype._attachOverlay=function(){var e=this;this._overlayRef?(this._overlayRef.updateSize({width:this.width,minWidth:this.minWidth,height:this.height,minHeight:this.minHeight}),this._overlayRef.getConfig().hasBackdrop=this.hasBackdrop):this._createOverlay(),this._overlayRef.hasAttached()||(this._overlayRef.attach(this._templatePortal),this.attach.emit()),this.hasBackdrop?this._backdropSubscription=this._overlayRef.backdropClick().subscribe(function(t){e.backdropClick.emit(t)}):this._backdropSubscription.unsubscribe()},e.prototype._detachOverlay=function(){this._overlayRef&&(this._overlayRef.detach(),this.detach.emit()),this._backdropSubscription.unsubscribe()},e.decorators=[{type:a.Directive,args:[{selector:"[cdk-connected-overlay], [connected-overlay], [cdkConnectedOverlay]",exportAs:"cdkConnectedOverlay"}]}],e.ctorParameters=function(){return[{type:H},{type:a.TemplateRef},{type:a.ViewContainerRef},{type:void 0,decorators:[{type:a.Inject,args:[G]}]},{type:d.Directionality,decorators:[{type:a.Optional}]}]},e.propDecorators={origin:[{type:a.Input,args:["cdkConnectedOverlayOrigin"]}],positions:[{type:a.Input,args:["cdkConnectedOverlayPositions"]}],offsetX:[{type:a.Input,args:["cdkConnectedOverlayOffsetX"]}],offsetY:[{type:a.Input,args:["cdkConnectedOverlayOffsetY"]}],width:[{type:a.Input,args:["cdkConnectedOverlayWidth"]}],height:[{type:a.Input,args:["cdkConnectedOverlayHeight"]}],minWidth:[{type:a.Input,args:["cdkConnectedOverlayMinWidth"]}],minHeight:[{type:a.Input,args:["cdkConnectedOverlayMinHeight"]}],backdropClass:[{type:a.Input,args:["cdkConnectedOverlayBackdropClass"]}],panelClass:[{type:a.Input,args:["cdkConnectedOverlayPanelClass"]}],viewportMargin:[{type:a.Input,args:["cdkConnectedOverlayViewportMargin"]}],scrollStrategy:[{type:a.Input,args:["cdkConnectedOverlayScrollStrategy"]}],open:[{type:a.Input,args:["cdkConnectedOverlayOpen"]}],hasBackdrop:[{type:a.Input,args:["cdkConnectedOverlayHasBackdrop"]}],lockPosition:[{type:a.Input,args:["cdkConnectedOverlayLockPosition"]}],flexibleDimensions:[{type:a.Input,args:["cdkConnectedOverlayFlexibleDimensions"]}],growAfterOpen:[{type:a.Input,args:["cdkConnectedOverlayGrowAfterOpen"]}],push:[{type:a.Input,args:["cdkConnectedOverlayPush"]}],backdropClick:[{type:a.Output}],positionChange:[{type:a.Output}],attach:[{type:a.Output}],detach:[{type:a.Output}],overlayKeydown:[{type:a.Output}]},e}();function K(e){return function(){return e.scrollStrategies.reposition()}}var Q={provide:G,deps:[H],useFactory:K},X=function(){function e(){}return e.decorators=[{type:a.NgModule,args:[{imports:[d.BidiModule,p.PortalModule,i.ScrollingModule],exports:[Y,W,i.ScrollingModule],declarations:[Y,W],providers:[H,Q]}]}],e}(),Z=[H,z,A,i.VIEWPORT_RULER_PROVIDER,I,Q],$=function(e){function t(t){return e.call(this,t)||this}return Object(s.__extends)(t,e),t.prototype.ngOnDestroy=function(){e.prototype.ngOnDestroy.call(this),this._fullScreenEventName&&this._fullScreenListener&&this._document.removeEventListener(this._fullScreenEventName,this._fullScreenListener)},t.prototype._createContainer=function(){var t=this;e.prototype._createContainer.call(this),this._adjustParentForFullscreenChange(),this._addFullscreenChangeListener(function(){return t._adjustParentForFullscreenChange()})},t.prototype._adjustParentForFullscreenChange=function(){this._containerElement&&(this.getFullscreenElement()||this._document.body).appendChild(this._containerElement)},t.prototype._addFullscreenChangeListener=function(e){var t=this._getEventName();t&&(this._fullScreenListener&&this._document.removeEventListener(t,this._fullScreenListener),this._document.addEventListener(t,e),this._fullScreenListener=e)},t.prototype._getEventName=function(){return this._fullScreenEventName||(this._document.fullscreenEnabled?this._fullScreenEventName="fullscreenchange":this._document.webkitFullscreenEnabled?this._fullScreenEventName="webkitfullscreenchange":this._document.mozFullScreenEnabled?this._fullScreenEventName="mozfullscreenchange":this._document.msFullscreenEnabled&&(this._fullScreenEventName="MSFullscreenChange")),this._fullScreenEventName},t.prototype.getFullscreenElement=function(){return this._document.fullscreenElement||this._document.webkitFullscreenElement||this._document.mozFullScreenElement||this._document.msFullscreenElement||null},t.decorators=[{type:a.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:a.Inject,args:[o.DOCUMENT]}]}]},t.ngInjectableDef=Object(a.defineInjectable)({factory:function(){return new t(Object(a.inject)(o.DOCUMENT))},token:t,providedIn:"root"}),t}(T)},"./node_modules/@angular/cdk/esm5/platform.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Platform",function(){return u}),n.d(t,"PlatformModule",function(){return c}),n.d(t,"getSupportedInputTypes",function(){return d}),n.d(t,"supportsPassiveEventListeners",function(){return p}),n.d(t,"normalizePassiveListenerOptions",function(){return f}),n.d(t,"supportsScrollBehavior",function(){return g}),n.d(t,"getRtlScrollAxisType",function(){return v}),n.d(t,"RtlScrollAxisType",function(){return m});var r,i,o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/common/fesm5/common.js"),s="undefined"!=typeof Intl&&Intl.v8BreakIterator,u=function(){function e(e){this._platformId=e,this.isBrowser=this._platformId?Object(a.isPlatformBrowser)(this._platformId):"object"==typeof document&&!!document,this.EDGE=this.isBrowser&&/(edge)/i.test(navigator.userAgent),this.TRIDENT=this.isBrowser&&/(msie|trident)/i.test(navigator.userAgent),this.BLINK=this.isBrowser&&!(!window.chrome&&!s)&&"undefined"!=typeof CSS&&!this.EDGE&&!this.TRIDENT,this.WEBKIT=this.isBrowser&&/AppleWebKit/i.test(navigator.userAgent)&&!this.BLINK&&!this.EDGE&&!this.TRIDENT,this.IOS=this.isBrowser&&/iPad|iPhone|iPod/.test(navigator.userAgent)&&!("MSStream"in window),this.FIREFOX=this.isBrowser&&/(firefox|minefield)/i.test(navigator.userAgent),this.ANDROID=this.isBrowser&&/android/i.test(navigator.userAgent)&&!this.TRIDENT,this.SAFARI=this.isBrowser&&/safari/i.test(navigator.userAgent)&&this.WEBKIT}return e.decorators=[{type:o.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:Object,decorators:[{type:o.Optional},{type:o.Inject,args:[o.PLATFORM_ID]}]}]},e.ngInjectableDef=Object(o.defineInjectable)({factory:function(){return new e(Object(o.inject)(o.PLATFORM_ID,8))},token:e,providedIn:"root"}),e}(),c=function(){function e(){}return e.decorators=[{type:o.NgModule,args:[{}]}],e}(),l=["color","button","checkbox","date","datetime-local","email","file","hidden","image","month","number","password","radio","range","reset","search","submit","tel","text","time","url","week"];function d(){if(r)return r;if("object"!=typeof document||!document)return r=new Set(l);var e=document.createElement("input");return r=new Set(l.filter(function(t){return e.setAttribute("type",t),e.type===t}))}function p(){if(null==i&&"undefined"!=typeof window)try{window.addEventListener("test",null,Object.defineProperty({},"passive",{get:function(){return i=!0}}))}finally{i=i||!1}return i}function f(e){return p()?e:!!e.capture}var h,m={NORMAL:0,NEGATED:1,INVERTED:2};function g(){return!!("object"==typeof document&&"scrollBehavior"in document.documentElement.style)}function v(){if("object"!=typeof document||!document)return m.NORMAL;if(!h){var e=document.createElement("div"),t=e.style;e.dir="rtl",t.height="1px",t.width="1px",t.overflow="auto",t.visibility="hidden",t.pointerEvents="none",t.position="absolute";var n=document.createElement("div"),r=n.style;r.width="2px",r.height="1px",e.appendChild(n),document.body.appendChild(e),h=m.NORMAL,0===e.scrollLeft&&(e.scrollLeft=1,h=0===e.scrollLeft?m.NEGATED:m.INVERTED),e.parentNode.removeChild(e)}return h}m[m.NORMAL]="NORMAL",m[m.NEGATED]="NEGATED",m[m.INVERTED]="INVERTED"},"./node_modules/@angular/cdk/esm5/portal.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"DomPortalHost",function(){return l}),n.d(t,"PortalHostDirective",function(){return p}),n.d(t,"TemplatePortalDirective",function(){return d}),n.d(t,"BasePortalHost",function(){return c}),n.d(t,"Portal",function(){return a}),n.d(t,"ComponentPortal",function(){return s}),n.d(t,"TemplatePortal",function(){return u}),n.d(t,"BasePortalOutlet",function(){return c}),n.d(t,"DomPortalOutlet",function(){return l}),n.d(t,"CdkPortal",function(){return d}),n.d(t,"CdkPortalOutlet",function(){return p}),n.d(t,"PortalModule",function(){return f}),n.d(t,"PortalInjector",function(){return h});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js");function o(){throw Error("Host already has a portal attached")}var a=function(){function e(){}return e.prototype.attach=function(e){return null==e&&function(){throw Error("Attempting to attach a portal to a null PortalOutlet")}(),e.hasAttached()&&o(),this._attachedHost=e,e.attach(this)},e.prototype.detach=function(){var e=this._attachedHost;null==e?function(){throw Error("Attempting to detach a portal that is not attached to a host")}():(this._attachedHost=null,e.detach())},Object.defineProperty(e.prototype,"isAttached",{get:function(){return null!=this._attachedHost},enumerable:!0,configurable:!0}),e.prototype.setAttachedHost=function(e){this._attachedHost=e},e}(),s=function(e){function t(t,n,r,i){var o=e.call(this)||this;return o.component=t,o.viewContainerRef=n,o.injector=r,o.componentFactoryResolver=i,o}return Object(r.__extends)(t,e),t}(a),u=function(e){function t(t,n,r){var i=e.call(this)||this;return i.templateRef=t,i.viewContainerRef=n,i.context=r,i}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"origin",{get:function(){return this.templateRef.elementRef},enumerable:!0,configurable:!0}),t.prototype.attach=function(t,n){return void 0===n&&(n=this.context),this.context=n,e.prototype.attach.call(this,t)},t.prototype.detach=function(){return this.context=void 0,e.prototype.detach.call(this)},t}(a),c=function(){function e(){this._isDisposed=!1}return e.prototype.hasAttached=function(){return!!this._attachedPortal},e.prototype.attach=function(e){return e||function(){throw Error("Must provide a portal to attach")}(),this.hasAttached()&&o(),this._isDisposed&&function(){throw Error("This PortalOutlet has already been disposed")}(),e instanceof s?(this._attachedPortal=e,this.attachComponentPortal(e)):e instanceof u?(this._attachedPortal=e,this.attachTemplatePortal(e)):void function(){throw Error("Attempting to attach an unknown Portal type. BasePortalOutlet accepts either a ComponentPortal or a TemplatePortal.")}()},e.prototype.detach=function(){this._attachedPortal&&(this._attachedPortal.setAttachedHost(null),this._attachedPortal=null),this._invokeDisposeFn()},e.prototype.dispose=function(){this.hasAttached()&&this.detach(),this._invokeDisposeFn(),this._isDisposed=!0},e.prototype.setDisposeFn=function(e){this._disposeFn=e},e.prototype._invokeDisposeFn=function(){this._disposeFn&&(this._disposeFn(),this._disposeFn=null)},e}(),l=function(e){function t(t,n,r,i){var o=e.call(this)||this;return o.outletElement=t,o._componentFactoryResolver=n,o._appRef=r,o._defaultInjector=i,o}return Object(r.__extends)(t,e),t.prototype.attachComponentPortal=function(e){var t,n=this,r=(e.componentFactoryResolver||this._componentFactoryResolver).resolveComponentFactory(e.component);return e.viewContainerRef?(t=e.viewContainerRef.createComponent(r,e.viewContainerRef.length,e.injector||e.viewContainerRef.injector),this.setDisposeFn(function(){return t.destroy()})):(t=r.create(e.injector||this._defaultInjector),this._appRef.attachView(t.hostView),this.setDisposeFn(function(){n._appRef.detachView(t.hostView),t.destroy()})),this.outletElement.appendChild(this._getComponentRootNode(t)),t},t.prototype.attachTemplatePortal=function(e){var t=this,n=e.viewContainerRef,r=n.createEmbeddedView(e.templateRef,e.context);return r.detectChanges(),r.rootNodes.forEach(function(e){return t.outletElement.appendChild(e)}),this.setDisposeFn(function(){var e=n.indexOf(r);-1!==e&&n.remove(e)}),r},t.prototype.dispose=function(){e.prototype.dispose.call(this),null!=this.outletElement.parentNode&&this.outletElement.parentNode.removeChild(this.outletElement)},t.prototype._getComponentRootNode=function(e){return e.hostView.rootNodes[0]},t}(c),d=function(e){function t(t,n){return e.call(this,t,n)||this}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"[cdk-portal], [cdkPortal], [portal]",exportAs:"cdkPortal"}]}],t.ctorParameters=function(){return[{type:i.TemplateRef},{type:i.ViewContainerRef}]},t}(u),p=function(e){function t(t,n){var r=e.call(this)||this;return r._componentFactoryResolver=t,r._viewContainerRef=n,r._isInitialized=!1,r.attached=new i.EventEmitter,r}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"portal",{get:function(){return this._attachedPortal},set:function(t){(!this.hasAttached()||t||this._isInitialized)&&(this.hasAttached()&&e.prototype.detach.call(this),t&&e.prototype.attach.call(this,t),this._attachedPortal=t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"attachedRef",{get:function(){return this._attachedRef},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){this._isInitialized=!0},t.prototype.ngOnDestroy=function(){e.prototype.dispose.call(this),this._attachedPortal=null,this._attachedRef=null},t.prototype.attachComponentPortal=function(t){t.setAttachedHost(this);var n=null!=t.viewContainerRef?t.viewContainerRef:this._viewContainerRef,r=(t.componentFactoryResolver||this._componentFactoryResolver).resolveComponentFactory(t.component),i=n.createComponent(r,n.length,t.injector||n.injector);return e.prototype.setDisposeFn.call(this,function(){return i.destroy()}),this._attachedPortal=t,this._attachedRef=i,this.attached.emit(i),i},t.prototype.attachTemplatePortal=function(t){var n=this;t.setAttachedHost(this);var r=this._viewContainerRef.createEmbeddedView(t.templateRef,t.context);return e.prototype.setDisposeFn.call(this,function(){return n._viewContainerRef.clear()}),this._attachedPortal=t,this._attachedRef=r,this.attached.emit(r),r},t.decorators=[{type:i.Directive,args:[{selector:"[cdkPortalOutlet], [cdkPortalHost], [portalHost]",exportAs:"cdkPortalOutlet, cdkPortalHost",inputs:["portal: cdkPortalOutlet"]}]}],t.ctorParameters=function(){return[{type:i.ComponentFactoryResolver},{type:i.ViewContainerRef}]},t.propDecorators={attached:[{type:i.Output}]},t}(c),f=function(){function e(){}return e.decorators=[{type:i.NgModule,args:[{exports:[d,p],declarations:[d,p]}]}],e}(),h=function(){function e(e,t){this._parentInjector=e,this._customTokens=t}return e.prototype.get=function(e,t){var n=this._customTokens.get(e);return void 0!==n?n:this._parentInjector.get(e,t)},e}()},"./node_modules/@angular/cdk/esm5/scrolling.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"_fixedSizeVirtualScrollStrategyFactory",function(){return f}),n.d(t,"FixedSizeVirtualScrollStrategy",function(){return p}),n.d(t,"CdkFixedSizeVirtualScroll",function(){return h}),n.d(t,"SCROLL_DISPATCHER_PROVIDER_FACTORY",function(){return v}),n.d(t,"DEFAULT_SCROLL_TIME",function(){return m}),n.d(t,"ScrollDispatcher",function(){return g}),n.d(t,"SCROLL_DISPATCHER_PROVIDER",function(){return _}),n.d(t,"CdkScrollable",function(){return y}),n.d(t,"ScrollingModule",function(){return O}),n.d(t,"ScrollDispatchModule",function(){return w}),n.d(t,"VIEWPORT_RULER_PROVIDER_FACTORY",function(){return S}),n.d(t,"DEFAULT_RESIZE_TIME",function(){return C}),n.d(t,"ViewportRuler",function(){return E}),n.d(t,"VIEWPORT_RULER_PROVIDER",function(){return k}),n.d(t,"CdkVirtualForOf",function(){return x}),n.d(t,"VIRTUAL_SCROLL_STRATEGY",function(){return d}),n.d(t,"CdkVirtualScrollViewport",function(){return b});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),o=n("./node_modules/rxjs/_esm5/index.js"),a=n("./node_modules/rxjs/_esm5/operators/index.js"),s=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),u=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),c=n("./node_modules/tslib/tslib.es6.js"),l=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),d=new r.InjectionToken("VIRTUAL_SCROLL_STRATEGY"),p=function(){function e(e,t,n){this._scrolledIndexChange=new o.Subject,this.scrolledIndexChange=this._scrolledIndexChange.pipe(Object(a.distinctUntilChanged)()),this._viewport=null,this._itemSize=e,this._minBufferPx=t,this._maxBufferPx=n}return e.prototype.attach=function(e){this._viewport=e,this._updateTotalContentSize(),this._updateRenderedRange()},e.prototype.detach=function(){this._scrolledIndexChange.complete(),this._viewport=null},e.prototype.updateItemAndBufferSize=function(e,t,n){if(n<t)throw Error("CDK virtual scroll: maxBufferPx must be greater than or equal to minBufferPx");this._itemSize=e,this._minBufferPx=t,this._maxBufferPx=n,this._updateTotalContentSize(),this._updateRenderedRange()},e.prototype.onContentScrolled=function(){this._updateRenderedRange()},e.prototype.onDataLengthChanged=function(){this._updateTotalContentSize(),this._updateRenderedRange()},e.prototype.onContentRendered=function(){},e.prototype.onRenderedOffsetChanged=function(){},e.prototype.scrollToIndex=function(e,t){this._viewport&&this._viewport.scrollToOffset(e*this._itemSize,t)},e.prototype._updateTotalContentSize=function(){this._viewport&&this._viewport.setTotalContentSize(this._viewport.getDataLength()*this._itemSize)},e.prototype._updateRenderedRange=function(){if(this._viewport){var e=this._viewport.measureScrollOffset(),t=e/this._itemSize,n=this._viewport.getRenderedRange(),r={start:n.start,end:n.end},i=this._viewport.getViewportSize(),o=this._viewport.getDataLength(),a=e-r.start*this._itemSize;if(a<this._minBufferPx&&0!=r.start){var s=Math.ceil((this._maxBufferPx-a)/this._itemSize);r.start=Math.max(0,r.start-s),r.end=Math.min(o,Math.ceil(t+(i+this._minBufferPx)/this._itemSize))}else{var u=r.end*this._itemSize-(e+i);if(u<this._minBufferPx&&r.end!=o){var c=Math.ceil((this._maxBufferPx-u)/this._itemSize);c>0&&(r.end=Math.min(o,r.end+c),r.start=Math.max(0,Math.floor(t-this._minBufferPx/this._itemSize)))}}this._viewport.setRenderedRange(r),this._viewport.setRenderedContentOffset(this._itemSize*r.start),this._scrolledIndexChange.next(Math.floor(t))}},e}();function f(e){return e._scrollStrategy}var h=function(){function e(){this._itemSize=20,this._minBufferPx=100,this._maxBufferPx=200,this._scrollStrategy=new p(this.itemSize,this.minBufferPx,this.maxBufferPx)}return Object.defineProperty(e.prototype,"itemSize",{get:function(){return this._itemSize},set:function(e){this._itemSize=Object(i.coerceNumberProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"minBufferPx",{get:function(){return this._minBufferPx},set:function(e){this._minBufferPx=Object(i.coerceNumberProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"maxBufferPx",{get:function(){return this._maxBufferPx},set:function(e){this._maxBufferPx=Object(i.coerceNumberProperty)(e)},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(){this._scrollStrategy.updateItemAndBufferSize(this.itemSize,this.minBufferPx,this.maxBufferPx)},e.decorators=[{type:r.Directive,args:[{selector:"cdk-virtual-scroll-viewport[itemSize]",providers:[{provide:d,useFactory:f,deps:[Object(r.forwardRef)(function(){return e})]}]}]}],e.propDecorators={itemSize:[{type:r.Input}],minBufferPx:[{type:r.Input}],maxBufferPx:[{type:r.Input}]},e}(),m=20,g=function(){function e(e,t){this._ngZone=e,this._platform=t,this._scrolled=new o.Subject,this._globalSubscription=null,this._scrolledCount=0,this.scrollContainers=new Map}return e.prototype.register=function(e){var t=this;this.scrollContainers.has(e)||this.scrollContainers.set(e,e.elementScrolled().subscribe(function(){return t._scrolled.next(e)}))},e.prototype.deregister=function(e){var t=this.scrollContainers.get(e);t&&(t.unsubscribe(),this.scrollContainers.delete(e))},e.prototype.scrolled=function(e){var t=this;return void 0===e&&(e=m),this._platform.isBrowser?o.Observable.create(function(n){t._globalSubscription||t._addGlobalListener();var r=e>0?t._scrolled.pipe(Object(a.auditTime)(e)).subscribe(n):t._scrolled.subscribe(n);return t._scrolledCount++,function(){r.unsubscribe(),t._scrolledCount--,t._scrolledCount||t._removeGlobalListener()}}):Object(o.of)()},e.prototype.ngOnDestroy=function(){var e=this;this._removeGlobalListener(),this.scrollContainers.forEach(function(t,n){return e.deregister(n)}),this._scrolled.complete()},e.prototype.ancestorScrolled=function(e,t){var n=this.getAncestorScrollContainers(e);return this.scrolled(t).pipe(Object(a.filter)(function(e){return!e||n.indexOf(e)>-1}))},e.prototype.getAncestorScrollContainers=function(e){var t=this,n=[];return this.scrollContainers.forEach(function(r,i){t._scrollableContainsElement(i,e)&&n.push(i)}),n},e.prototype._scrollableContainsElement=function(e,t){var n=t.nativeElement,r=e.getElementRef().nativeElement;do{if(n==r)return!0}while(n=n.parentElement);return!1},e.prototype._addGlobalListener=function(){var e=this;this._globalSubscription=this._ngZone.runOutsideAngular(function(){return Object(o.fromEvent)(window.document,"scroll").subscribe(function(){return e._scrolled.next()})})},e.prototype._removeGlobalListener=function(){this._globalSubscription&&(this._globalSubscription.unsubscribe(),this._globalSubscription=null)},e.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:r.NgZone},{type:s.Platform}]},e.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new e(Object(r.inject)(r.NgZone),Object(r.inject)(s.Platform))},token:e,providedIn:"root"}),e}();function v(e,t,n){return e||new g(t,n)}var _={provide:g,deps:[[new r.Optional,new r.SkipSelf,g],r.NgZone,s.Platform],useFactory:v},y=function(){function e(e,t,n,r){var i=this;this.elementRef=e,this.scrollDispatcher=t,this.ngZone=n,this.dir=r,this._destroyed=new o.Subject,this._elementScrolled=o.Observable.create(function(e){return i.ngZone.runOutsideAngular(function(){return Object(o.fromEvent)(i.elementRef.nativeElement,"scroll").pipe(Object(a.takeUntil)(i._destroyed)).subscribe(e)})})}return e.prototype.ngOnInit=function(){this.scrollDispatcher.register(this)},e.prototype.ngOnDestroy=function(){this.scrollDispatcher.deregister(this),this._destroyed.next(),this._destroyed.complete()},e.prototype.elementScrolled=function(){return this._elementScrolled},e.prototype.getElementRef=function(){return this.elementRef},e.prototype.scrollTo=function(e){var t=this.elementRef.nativeElement,n=this.dir&&"rtl"==this.dir.value;e.left=null==e.left?n?e.end:e.start:e.left,e.right=null==e.right?n?e.start:e.end:e.right,null!=e.bottom&&(e.top=t.scrollHeight-t.clientHeight-e.bottom),n&&Object(s.getRtlScrollAxisType)()!=s.RtlScrollAxisType.NORMAL?(null!=e.left&&(e.right=t.scrollWidth-t.clientWidth-e.left),Object(s.getRtlScrollAxisType)()==s.RtlScrollAxisType.INVERTED?e.left=e.right:Object(s.getRtlScrollAxisType)()==s.RtlScrollAxisType.NEGATED&&(e.left=e.right?-e.right:e.right)):null!=e.right&&(e.left=t.scrollWidth-t.clientWidth-e.right),this._applyScrollToOptions(e)},e.prototype._applyScrollToOptions=function(e){var t=this.elementRef.nativeElement;Object(s.supportsScrollBehavior)()?t.scrollTo(e):(null!=e.top&&(t.scrollTop=e.top),null!=e.left&&(t.scrollLeft=e.left))},e.prototype.measureScrollOffset=function(e){var t=this.elementRef.nativeElement;if("top"==e)return t.scrollTop;if("bottom"==e)return t.scrollHeight-t.clientHeight-t.scrollTop;var n=this.dir&&"rtl"==this.dir.value;return"start"==e?e=n?"right":"left":"end"==e&&(e=n?"left":"right"),n&&Object(s.getRtlScrollAxisType)()==s.RtlScrollAxisType.INVERTED?"left"==e?t.scrollWidth-t.clientWidth-t.scrollLeft:t.scrollLeft:n&&Object(s.getRtlScrollAxisType)()==s.RtlScrollAxisType.NEGATED?"left"==e?t.scrollLeft+t.scrollWidth-t.clientWidth:-t.scrollLeft:"left"==e?t.scrollLeft:t.scrollWidth-t.clientWidth-t.scrollLeft},e.decorators=[{type:r.Directive,args:[{selector:"[cdk-scrollable], [cdkScrollable]"}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:g},{type:r.NgZone},{type:u.Directionality,decorators:[{type:r.Optional}]}]},e}();var b=function(e){function t(t,n,r,i,a,s){var u=e.call(this,t,s,r,a)||this;if(u.elementRef=t,u._changeDetectorRef=n,u._scrollStrategy=i,u._detachedSubject=new o.Subject,u._renderedRangeSubject=new o.Subject,u.orientation="vertical",u.scrolledIndexChange=o.Observable.create(function(e){return u._scrollStrategy.scrolledIndexChange.subscribe(function(t){return Promise.resolve().then(function(){return u.ngZone.run(function(){return e.next(t)})})})}),u.renderedRangeStream=u._renderedRangeSubject.asObservable(),u._totalContentSizeTransform="",u._totalContentSize=0,u._renderedRange={start:0,end:0},u._dataLength=0,u._viewportSize=0,u._renderedContentOffset=0,u._renderedContentOffsetNeedsRewrite=!1,u._isChangeDetectionPending=!1,u._runAfterChangeDetection=[],!i)throw Error('Error: cdk-virtual-scroll-viewport requires the "itemSize" property to be set.');return u}return Object(c.__extends)(t,e),t.prototype.ngOnInit=function(){var t=this;e.prototype.ngOnInit.call(this),this.ngZone.runOutsideAngular(function(){return Promise.resolve().then(function(){t._measureViewportSize(),t._scrollStrategy.attach(t),t.elementScrolled().pipe(Object(a.startWith)(null),Object(a.auditTime)(0,o.animationFrameScheduler)).subscribe(function(){return t._scrollStrategy.onContentScrolled()}),t._markChangeDetectionNeeded()})})},t.prototype.ngOnDestroy=function(){this.detach(),this._scrollStrategy.detach(),this._renderedRangeSubject.complete(),this._detachedSubject.complete(),e.prototype.ngOnDestroy.call(this)},t.prototype.attach=function(e){var t=this;if(this._forOf)throw Error("CdkVirtualScrollViewport is already attached.");this.ngZone.runOutsideAngular(function(){t._forOf=e,t._forOf.dataStream.pipe(Object(a.takeUntil)(t._detachedSubject)).subscribe(function(e){var n=e.length;n!==t._dataLength&&(t._dataLength=n,t._scrollStrategy.onDataLengthChanged()),t._doChangeDetection()})})},t.prototype.detach=function(){this._forOf=null,this._detachedSubject.next()},t.prototype.getDataLength=function(){return this._dataLength},t.prototype.getViewportSize=function(){return this._viewportSize},t.prototype.getRenderedRange=function(){return this._renderedRange},t.prototype.setTotalContentSize=function(e){if(this._totalContentSize!==e){this._totalContentSize=e;var t="horizontal"==this.orientation?"X":"Y";this._totalContentSizeTransform="scale"+t+"("+this._totalContentSize+")",this._markChangeDetectionNeeded()}},t.prototype.setRenderedRange=function(e){var t,n,r=this;t=this._renderedRange,n=e,(t.start!=n.start||t.end!=n.end)&&(this._renderedRangeSubject.next(this._renderedRange=e),this._markChangeDetectionNeeded(function(){return r._scrollStrategy.onContentRendered()}))},t.prototype.getOffsetToRenderedContentStart=function(){return this._renderedContentOffsetNeedsRewrite?null:this._renderedContentOffset},t.prototype.setRenderedContentOffset=function(e,t){var n=this;void 0===t&&(t="to-start");var r=this.dir&&"rtl"==this.dir.value,i="horizontal"==this.orientation,o=i?"X":"Y",a="translate"+o+"("+Number((i&&r?-1:1)*e)+"px)";this._renderedContentOffset=e,"to-end"===t&&(a+=" translate"+o+"(-100%)",this._renderedContentOffsetNeedsRewrite=!0),this._renderedContentTransform!=a&&(this._renderedContentTransform=a,this._markChangeDetectionNeeded(function(){n._renderedContentOffsetNeedsRewrite?(n._renderedContentOffset-=n.measureRenderedContentSize(),n._renderedContentOffsetNeedsRewrite=!1,n.setRenderedContentOffset(n._renderedContentOffset)):n._scrollStrategy.onRenderedOffsetChanged()}))},t.prototype.scrollToOffset=function(e,t){void 0===t&&(t="auto");var n={behavior:t};"horizontal"===this.orientation?n.start=e:n.top=e,this.scrollTo(n)},t.prototype.scrollToIndex=function(e,t){void 0===t&&(t="auto"),this._scrollStrategy.scrollToIndex(e,t)},t.prototype.measureScrollOffset=function(t){return e.prototype.measureScrollOffset.call(this,t||("horizontal"===this.orientation?"start":"top"))},t.prototype.measureRenderedContentSize=function(){var e=this._contentWrapper.nativeElement;return"horizontal"===this.orientation?e.offsetWidth:e.offsetHeight},t.prototype.measureRangeSize=function(e){return this._forOf?this._forOf.measureRangeSize(e,this.orientation):0},t.prototype.checkViewportSize=function(){this._measureViewportSize(),this._scrollStrategy.onDataLengthChanged()},t.prototype._measureViewportSize=function(){var e=this.elementRef.nativeElement;this._viewportSize="horizontal"===this.orientation?e.clientWidth:e.clientHeight},t.prototype._markChangeDetectionNeeded=function(e){var t=this;e&&this._runAfterChangeDetection.push(e),this._isChangeDetectionPending||(this._isChangeDetectionPending=!0,this.ngZone.runOutsideAngular(function(){return Promise.resolve().then(function(){t._doChangeDetection()})}))},t.prototype._doChangeDetection=function(){var e=this;this._isChangeDetectionPending=!1,this.ngZone.run(function(){return e._changeDetectorRef.markForCheck()}),this._contentWrapper.nativeElement.style.transform=this._renderedContentTransform;var t=this._runAfterChangeDetection;this._runAfterChangeDetection=[];for(var n=0,r=t;n<r.length;n++){(0,r[n])()}},t.decorators=[{type:r.Component,args:[{selector:"cdk-virtual-scroll-viewport",template:'<div #contentWrapper class="cdk-virtual-scroll-content-wrapper"><ng-content></ng-content></div><div class="cdk-virtual-scroll-spacer" [style.transform]="_totalContentSizeTransform"></div>',styles:["cdk-virtual-scroll-viewport{display:block;position:relative;overflow:auto;contain:strict;transform:translateZ(0);will-change:scroll-position;-webkit-overflow-scrolling:touch}.cdk-virtual-scroll-content-wrapper{position:absolute;top:0;left:0;contain:content}[dir=rtl] .cdk-virtual-scroll-content-wrapper{right:0;left:auto}.cdk-virtual-scroll-orientation-horizontal .cdk-virtual-scroll-content-wrapper{min-height:100%}.cdk-virtual-scroll-orientation-horizontal .cdk-virtual-scroll-content-wrapper>dl:not([cdkVirtualFor]),.cdk-virtual-scroll-orientation-horizontal .cdk-virtual-scroll-content-wrapper>ol:not([cdkVirtualFor]),.cdk-virtual-scroll-orientation-horizontal .cdk-virtual-scroll-content-wrapper>table:not([cdkVirtualFor]),.cdk-virtual-scroll-orientation-horizontal .cdk-virtual-scroll-content-wrapper>ul:not([cdkVirtualFor]){padding-left:0;padding-right:0;margin-left:0;margin-right:0;border-left-width:0;border-right-width:0;outline:0}.cdk-virtual-scroll-orientation-vertical .cdk-virtual-scroll-content-wrapper{min-width:100%}.cdk-virtual-scroll-orientation-vertical .cdk-virtual-scroll-content-wrapper>dl:not([cdkVirtualFor]),.cdk-virtual-scroll-orientation-vertical .cdk-virtual-scroll-content-wrapper>ol:not([cdkVirtualFor]),.cdk-virtual-scroll-orientation-vertical .cdk-virtual-scroll-content-wrapper>table:not([cdkVirtualFor]),.cdk-virtual-scroll-orientation-vertical .cdk-virtual-scroll-content-wrapper>ul:not([cdkVirtualFor]){padding-top:0;padding-bottom:0;margin-top:0;margin-bottom:0;border-top-width:0;border-bottom-width:0;outline:0}.cdk-virtual-scroll-spacer{position:absolute;top:0;left:0;height:1px;width:1px;transform-origin:0 0}[dir=rtl] .cdk-virtual-scroll-spacer{right:0;left:auto;transform-origin:100% 0}"],host:{class:"cdk-virtual-scroll-viewport","[class.cdk-virtual-scroll-orientation-horizontal]":'orientation === "horizontal"',"[class.cdk-virtual-scroll-orientation-vertical]":'orientation !== "horizontal"'},encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,providers:[{provide:y,useExisting:t}]}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:r.ChangeDetectorRef},{type:r.NgZone},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[d]}]},{type:u.Directionality,decorators:[{type:r.Optional}]},{type:g}]},t.propDecorators={orientation:[{type:r.Input}],scrolledIndexChange:[{type:r.Output}],_contentWrapper:[{type:r.ViewChild,args:["contentWrapper"]}]},t}(y);function j(e,t){var n=t;if(!n.getBoundingClientRect)return 0;var r=n.getBoundingClientRect();return"horizontal"==e?r.width:r.height}var x=function(){function e(e,t,n,r,i){var s=this;this._viewContainerRef=e,this._template=t,this._differs=n,this._viewport=r,this.viewChange=new o.Subject,this._dataSourceChanges=new o.Subject,this.cdkVirtualForTemplateCacheSize=20,this.dataStream=this._dataSourceChanges.pipe(Object(a.startWith)(null),Object(a.pairwise)(),Object(a.switchMap)(function(e){var t=e[0],n=e[1];return s._changeDataSource(t,n)}),Object(a.shareReplay)(1)),this._differ=null,this._templateCache=[],this._needsUpdate=!1,this._destroyed=new o.Subject,this.dataStream.subscribe(function(e){s._data=e,s._onRenderedDataChange()}),this._viewport.renderedRangeStream.pipe(Object(a.takeUntil)(this._destroyed)).subscribe(function(e){s._renderedRange=e,i.run(function(){return s.viewChange.next(s._renderedRange)}),s._onRenderedDataChange()}),this._viewport.attach(this)}return Object.defineProperty(e.prototype,"cdkVirtualForOf",{get:function(){return this._cdkVirtualForOf},set:function(e){this._cdkVirtualForOf=e;var t=e instanceof l.DataSource?e:new l.ArrayDataSource(e instanceof o.Observable?e:Array.prototype.slice.call(e||[]));this._dataSourceChanges.next(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"cdkVirtualForTrackBy",{get:function(){return this._cdkVirtualForTrackBy},set:function(e){var t=this;this._needsUpdate=!0,this._cdkVirtualForTrackBy=e?function(n,r){return e(n+(t._renderedRange?t._renderedRange.start:0),r)}:void 0},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"cdkVirtualForTemplate",{set:function(e){e&&(this._needsUpdate=!0,this._template=e)},enumerable:!0,configurable:!0}),e.prototype.measureRangeSize=function(e,t){if(e.start>=e.end)return 0;if(e.start<this._renderedRange.start||e.end>this._renderedRange.end)throw Error("Error: attempted to measure an item that isn't rendered.");for(var n=e.start-this._renderedRange.start,r=0,i=e.end-e.start;i--;)for(var o=this._viewContainerRef.get(i+n),a=o?o.rootNodes.length:0;a--;)r+=j(t,o.rootNodes[a]);return r},e.prototype.ngDoCheck=function(){if(this._differ&&this._needsUpdate){var e=this._differ.diff(this._renderedItems);e?this._applyChanges(e):this._updateContext(),this._needsUpdate=!1}},e.prototype.ngOnDestroy=function(){this._viewport.detach(),this._dataSourceChanges.complete(),this.viewChange.complete(),this._destroyed.next(),this._destroyed.complete();for(var e=0,t=this._templateCache;e<t.length;e++){t[e].destroy()}},e.prototype._onRenderedDataChange=function(){this._renderedRange&&(this._renderedItems=this._data.slice(this._renderedRange.start,this._renderedRange.end),this._differ||(this._differ=this._differs.find(this._renderedItems).create(this.cdkVirtualForTrackBy)),this._needsUpdate=!0)},e.prototype._changeDataSource=function(e,t){return e&&e.disconnect(this),this._needsUpdate=!0,t.connect(this)},e.prototype._updateContext=function(){for(var e=this._data.length,t=this._viewContainerRef.length;t--;){var n=this._viewContainerRef.get(t);n.context.index=this._renderedRange.start+t,n.context.count=e,this._updateComputedContextProperties(n.context),n.detectChanges()}},e.prototype._applyChanges=function(e){var t=this;e.forEachOperation(function(e,n,r){if(null==e.previousIndex){var i=t._getViewForNewItem();t._viewContainerRef.insert(i,r),i.context.$implicit=e.item}else if(null==r)t._cacheView(t._viewContainerRef.detach(n));else{i=t._viewContainerRef.get(n);t._viewContainerRef.move(i,r),i.context.$implicit=e.item}}),e.forEachIdentityChange(function(e){t._viewContainerRef.get(e.currentIndex).context.$implicit=e.item});for(var n=this._data.length,r=this._viewContainerRef.length;r--;){var i=this._viewContainerRef.get(r);i.context.index=this._renderedRange.start+r,i.context.count=n,this._updateComputedContextProperties(i.context)}},e.prototype._cacheView=function(e){if(this._templateCache.length<this.cdkVirtualForTemplateCacheSize)this._templateCache.push(e);else{var t=this._viewContainerRef.indexOf(e);-1===t?e.destroy():this._viewContainerRef.remove(t)}},e.prototype._getViewForNewItem=function(){return this._templateCache.pop()||this._viewContainerRef.createEmbeddedView(this._template,{$implicit:null,cdkVirtualForOf:this._cdkVirtualForOf,index:-1,count:-1,first:!1,last:!1,odd:!1,even:!1})},e.prototype._updateComputedContextProperties=function(e){e.first=0===e.index,e.last=e.index===e.count-1,e.even=e.index%2==0,e.odd=!e.even},e.decorators=[{type:r.Directive,args:[{selector:"[cdkVirtualFor][cdkVirtualForOf]"}]}],e.ctorParameters=function(){return[{type:r.ViewContainerRef},{type:r.TemplateRef},{type:r.IterableDiffers},{type:b,decorators:[{type:r.SkipSelf}]},{type:r.NgZone}]},e.propDecorators={cdkVirtualForOf:[{type:r.Input}],cdkVirtualForTrackBy:[{type:r.Input}],cdkVirtualForTemplate:[{type:r.Input}],cdkVirtualForTemplateCacheSize:[{type:r.Input}]},e}(),O=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[u.BidiModule,s.PlatformModule],exports:[u.BidiModule,h,y,x,b],declarations:[h,y,x,b]}]}],e}(),w=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[O],exports:[O]}]}],e}(),C=20,E=function(){function e(e,t){var n=this;this._platform=e,t.runOutsideAngular(function(){n._change=e.isBrowser?Object(o.merge)(Object(o.fromEvent)(window,"resize"),Object(o.fromEvent)(window,"orientationchange")):Object(o.of)(),n._invalidateCache=n.change().subscribe(function(){return n._updateViewportSize()})})}return e.prototype.ngOnDestroy=function(){this._invalidateCache.unsubscribe()},e.prototype.getViewportSize=function(){this._viewportSize||this._updateViewportSize();var e={width:this._viewportSize.width,height:this._viewportSize.height};return this._platform.isBrowser||(this._viewportSize=null),e},e.prototype.getViewportRect=function(){var e=this.getViewportScrollPosition(),t=this.getViewportSize(),n=t.width,r=t.height;return{top:e.top,left:e.left,bottom:e.top+r,right:e.left+n,height:r,width:n}},e.prototype.getViewportScrollPosition=function(){if(!this._platform.isBrowser)return{top:0,left:0};var e=document.documentElement,t=e.getBoundingClientRect();return{top:-t.top||document.body.scrollTop||window.scrollY||e.scrollTop||0,left:-t.left||document.body.scrollLeft||window.scrollX||e.scrollLeft||0}},e.prototype.change=function(e){return void 0===e&&(e=C),e>0?this._change.pipe(Object(a.auditTime)(e)):this._change},e.prototype._updateViewportSize=function(){this._viewportSize=this._platform.isBrowser?{width:window.innerWidth,height:window.innerHeight}:{width:0,height:0}},e.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:s.Platform},{type:r.NgZone}]},e.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new e(Object(r.inject)(s.Platform),Object(r.inject)(r.NgZone))},token:e,providedIn:"root"}),e}();function S(e,t,n){return e||new E(t,n)}var k={provide:E,deps:[[new r.Optional,new r.SkipSelf,E],s.Platform,r.NgZone],useFactory:S}},"./node_modules/@angular/cdk/esm5/stepper.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"StepperSelectionEvent",function(){return h}),n.d(t,"STEP_STATE",function(){return m}),n.d(t,"STEPPER_GLOBAL_OPTIONS",function(){return g}),n.d(t,"MAT_STEPPER_GLOBAL_OPTIONS",function(){return v}),n.d(t,"CdkStep",function(){return _}),n.d(t,"CdkStepper",function(){return y}),n.d(t,"CdkStepLabel",function(){return d}),n.d(t,"CdkStepperNext",function(){return b}),n.d(t,"CdkStepperPrevious",function(){return j}),n.d(t,"CdkStepperModule",function(){return x}),n.d(t,"CdkStepHeader",function(){return p});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),o=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),a=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),s=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),u=n("./node_modules/@angular/common/fesm5/common.js"),c=(n("./node_modules/@angular/forms/fesm5/forms.js"),n("./node_modules/rxjs/_esm5/index.js")),l=n("./node_modules/rxjs/_esm5/operators/index.js"),d=function(){function e(e){this.template=e}return e.decorators=[{type:r.Directive,args:[{selector:"[cdkStepLabel]"}]}],e.ctorParameters=function(){return[{type:r.TemplateRef}]},e}(),p=function(){function e(e){this._elementRef=e}return e.prototype.focus=function(){this._elementRef.nativeElement.focus()},e.decorators=[{type:r.Directive,args:[{selector:"[cdkStepHeader]",host:{role:"tab"}}]}],e.ctorParameters=function(){return[{type:r.ElementRef}]},e}(),f=0,h=function(){return function(){}}(),m={NUMBER:"number",EDIT:"edit",DONE:"done",ERROR:"error"},g=new r.InjectionToken("STEPPER_GLOBAL_OPTIONS"),v=g,_=function(){function e(e,t){this._stepper=e,this.interacted=!1,this._editable=!0,this._optional=!1,this._customCompleted=null,this._customError=null,this._stepperOptions=t||{},this._displayDefaultIndicatorType=!1!==this._stepperOptions.displayDefaultIndicatorType,this._showError=!!this._stepperOptions.showError}return Object.defineProperty(e.prototype,"editable",{get:function(){return this._editable},set:function(e){this._editable=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"optional",{get:function(){return this._optional},set:function(e){this._optional=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"completed",{get:function(){return null==this._customCompleted?this._getDefaultCompleted():this._customCompleted},set:function(e){this._customCompleted=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),e.prototype._getDefaultCompleted=function(){return this.stepControl?this.stepControl.valid&&this.interacted:this.interacted},Object.defineProperty(e.prototype,"hasError",{get:function(){return null==this._customError?this._getDefaultError():this._customError},set:function(e){this._customError=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),e.prototype._getDefaultError=function(){return this.stepControl&&this.stepControl.invalid&&this.interacted},e.prototype.select=function(){this._stepper.selected=this},e.prototype.reset=function(){this.interacted=!1,null!=this._customCompleted&&(this._customCompleted=!1),null!=this._customError&&(this._customError=!1),this.stepControl&&this.stepControl.reset()},e.prototype.ngOnChanges=function(){this._stepper._stateChanged()},e.decorators=[{type:r.Component,args:[{selector:"cdk-step",exportAs:"cdkStep",template:"<ng-template><ng-content></ng-content></ng-template>",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:y,decorators:[{type:r.Inject,args:[Object(r.forwardRef)(function(){return y})]}]},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[g]}]}]},e.propDecorators={stepLabel:[{type:r.ContentChild,args:[d]}],content:[{type:r.ViewChild,args:[r.TemplateRef]}],stepControl:[{type:r.Input}],label:[{type:r.Input}],errorMessage:[{type:r.Input}],ariaLabel:[{type:r.Input,args:["aria-label"]}],ariaLabelledby:[{type:r.Input,args:["aria-labelledby"]}],state:[{type:r.Input}],editable:[{type:r.Input}],optional:[{type:r.Input}],completed:[{type:r.Input}],hasError:[{type:r.Input}]},e}(),y=function(){function e(e,t,n,i){this._dir=e,this._changeDetectorRef=t,this._elementRef=n,this._destroyed=new c.Subject,this._linear=!1,this._selectedIndex=0,this.selectionChange=new r.EventEmitter,this._orientation="horizontal",this._groupId=f++,this._document=i}return Object.defineProperty(e.prototype,"linear",{get:function(){return this._linear},set:function(e){this._linear=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selectedIndex",{get:function(){return this._selectedIndex},set:function(e){if(this._steps){if(e<0||e>this._steps.length-1)throw Error("cdkStepper: Cannot assign out-of-bounds value to `selectedIndex`.");this._selectedIndex!=e&&!this._anyControlsInvalidOrPending(e)&&(e>=this._selectedIndex||this._steps.toArray()[e].editable)&&this._updateSelectedItemIndex(e)}else this._selectedIndex=e},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selected",{get:function(){return this._steps?this._steps.toArray()[this.selectedIndex]:void 0},set:function(e){this.selectedIndex=this._steps?this._steps.toArray().indexOf(e):-1},enumerable:!0,configurable:!0}),e.prototype.ngAfterViewInit=function(){var e=this;this._keyManager=new i.FocusKeyManager(this._stepHeader).withWrap().withVerticalOrientation("vertical"===this._orientation),(this._dir?this._dir.change:Object(c.of)()).pipe(Object(l.startWith)(this._layoutDirection()),Object(l.takeUntil)(this._destroyed)).subscribe(function(t){return e._keyManager.withHorizontalOrientation(t)}),this._keyManager.updateActiveItemIndex(this._selectedIndex),this._steps.changes.pipe(Object(l.takeUntil)(this._destroyed)).subscribe(function(){e.selected||(e._selectedIndex=Math.max(e._selectedIndex-1,0))})},e.prototype.ngOnDestroy=function(){this._destroyed.next(),this._destroyed.complete()},e.prototype.next=function(){this.selectedIndex=Math.min(this._selectedIndex+1,this._steps.length-1)},e.prototype.previous=function(){this.selectedIndex=Math.max(this._selectedIndex-1,0)},e.prototype.reset=function(){this._updateSelectedItemIndex(0),this._steps.forEach(function(e){return e.reset()}),this._stateChanged()},e.prototype._getStepLabelId=function(e){return"cdk-step-label-"+this._groupId+"-"+e},e.prototype._getStepContentId=function(e){return"cdk-step-content-"+this._groupId+"-"+e},e.prototype._stateChanged=function(){this._changeDetectorRef.markForCheck()},e.prototype._getAnimationDirection=function(e){var t=e-this._selectedIndex;return t<0?"rtl"===this._layoutDirection()?"next":"previous":t>0?"rtl"===this._layoutDirection()?"previous":"next":"current"},e.prototype._getIndicatorType=function(e,t){void 0===t&&(t=m.NUMBER);var n=this._steps.toArray()[e],r=this._isCurrentStep(e);return n._displayDefaultIndicatorType?this._getDefaultIndicatorLogic(n,r):this._getGuidelineLogic(n,r,t)},e.prototype._getDefaultIndicatorLogic=function(e,t){return e._showError&&e.hasError&&!t?m.ERROR:!e.completed||t?m.NUMBER:e.editable?m.EDIT:m.DONE},e.prototype._getGuidelineLogic=function(e,t,n){return void 0===n&&(n=m.NUMBER),e._showError&&e.hasError&&!t?m.ERROR:e.completed&&!t?m.DONE:e.completed&&t?n:e.editable&&t?m.EDIT:n},e.prototype._isCurrentStep=function(e){return this._selectedIndex===e},e.prototype._getFocusIndex=function(){return this._keyManager?this._keyManager.activeItemIndex:this._selectedIndex},e.prototype._updateSelectedItemIndex=function(e){var t=this._steps.toArray();this.selectionChange.emit({selectedIndex:e,previouslySelectedIndex:this._selectedIndex,selectedStep:t[e],previouslySelectedStep:t[this._selectedIndex]}),this._containsFocus()?this._keyManager.setActiveItem(e):this._keyManager.updateActiveItemIndex(e),this._selectedIndex=e,this._stateChanged()},e.prototype._onKeydown=function(e){var t=Object(s.hasModifierKey)(e),n=e.keyCode,r=this._keyManager;null==r.activeItemIndex||t||n!==s.SPACE&&n!==s.ENTER?n===s.HOME?(r.setFirstItemActive(),e.preventDefault()):n===s.END?(r.setLastItemActive(),e.preventDefault()):r.onKeydown(e):(this.selectedIndex=r.activeItemIndex,e.preventDefault())},e.prototype._anyControlsInvalidOrPending=function(e){var t=this._steps.toArray();return t[this._selectedIndex].interacted=!0,!!(this._linear&&e>=0)&&t.slice(0,e).some(function(e){var t=e.stepControl;return(t?t.invalid||t.pending||!e.interacted:!e.completed)&&!e.optional})},e.prototype._layoutDirection=function(){return this._dir&&"rtl"===this._dir.value?"rtl":"ltr"},e.prototype._containsFocus=function(){if(!this._document||!this._elementRef)return!1;var e=this._elementRef.nativeElement,t=this._document.activeElement;return e===t||e.contains(t)},e.decorators=[{type:r.Directive,args:[{selector:"[cdkStepper]",exportAs:"cdkStepper"}]}],e.ctorParameters=function(){return[{type:o.Directionality,decorators:[{type:r.Optional}]},{type:r.ChangeDetectorRef},{type:r.ElementRef},{type:void 0,decorators:[{type:r.Inject,args:[u.DOCUMENT]}]}]},e.propDecorators={_steps:[{type:r.ContentChildren,args:[_]}],_stepHeader:[{type:r.ContentChildren,args:[p]}],linear:[{type:r.Input}],selectedIndex:[{type:r.Input}],selected:[{type:r.Input}],selectionChange:[{type:r.Output}]},e}(),b=function(){function e(e){this._stepper=e,this.type="submit"}return e.decorators=[{type:r.Directive,args:[{selector:"button[cdkStepperNext]",host:{"(click)":"_stepper.next()","[type]":"type"}}]}],e.ctorParameters=function(){return[{type:y}]},e.propDecorators={type:[{type:r.Input}]},e}(),j=function(){function e(e){this._stepper=e,this.type="button"}return e.decorators=[{type:r.Directive,args:[{selector:"button[cdkStepperPrevious]",host:{"(click)":"_stepper.previous()","[type]":"type"}}]}],e.ctorParameters=function(){return[{type:y}]},e.propDecorators={type:[{type:r.Input}]},e}(),x=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[o.BidiModule,u.CommonModule],exports:[_,y,p,d,b,j],declarations:[_,y,p,d,b,j]}]}],e}()},"./node_modules/@angular/cdk/esm5/table.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"DataRowOutlet",function(){return L}),n.d(t,"HeaderRowOutlet",function(){return V}),n.d(t,"FooterRowOutlet",function(){return B}),n.d(t,"CDK_TABLE_TEMPLATE",function(){return z}),n.d(t,"CdkTable",function(){return U}),n.d(t,"CdkCellDef",function(){return f}),n.d(t,"CdkHeaderCellDef",function(){return h}),n.d(t,"CdkFooterCellDef",function(){return m}),n.d(t,"CdkColumnDefBase",function(){return g}),n.d(t,"_CdkColumnDefBase",function(){return v}),n.d(t,"CdkColumnDef",function(){return _}),n.d(t,"BaseCdkCell",function(){return y}),n.d(t,"CdkHeaderCell",function(){return b}),n.d(t,"CdkFooterCell",function(){return j}),n.d(t,"CdkCell",function(){return x}),n.d(t,"CDK_ROW_TEMPLATE",function(){return O}),n.d(t,"BaseRowDef",function(){return w}),n.d(t,"CdkHeaderRowDefBase",function(){return C}),n.d(t,"_CdkHeaderRowDefBase",function(){return E}),n.d(t,"CdkHeaderRowDef",function(){return S}),n.d(t,"CdkFooterRowDefBase",function(){return k}),n.d(t,"_CdkFooterRowDefBase",function(){return M}),n.d(t,"CdkFooterRowDef",function(){return A}),n.d(t,"CdkRowDef",function(){return T}),n.d(t,"CdkCellOutlet",function(){return D}),n.d(t,"CdkHeaderRow",function(){return I}),n.d(t,"CdkFooterRow",function(){return P}),n.d(t,"CdkRow",function(){return R}),n.d(t,"CdkTableModule",function(){return G}),n.d(t,"STICKY_DIRECTIONS",function(){return N}),n.d(t,"StickyStyler",function(){return F}),n.d(t,"mixinHasStickyInput",function(){return p});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/cdk/esm5/collections.es5.js");n.d(t,"DataSource",function(){return a.DataSource});var s=n("./node_modules/@angular/common/fesm5/common.js"),u=n("./node_modules/rxjs/_esm5/index.js"),c=n("./node_modules/rxjs/_esm5/operators/index.js"),l=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),d=n("./node_modules/@angular/cdk/esm5/platform.es5.js");function p(e){return function(e){function t(){for(var t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];var r=e.apply(this,t)||this;return r._sticky=!1,r._hasStickyChanged=!1,r}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"sticky",{get:function(){return this._sticky},set:function(e){var t=this._sticky;this._sticky=Object(i.coerceBooleanProperty)(e),this._hasStickyChanged=t!==this._sticky},enumerable:!0,configurable:!0}),t.prototype.hasStickyChanged=function(){var e=this._hasStickyChanged;return this._hasStickyChanged=!1,e},t.prototype.resetStickyChanged=function(){this._hasStickyChanged=!1},t}(e)}var f=function(){function e(e){this.template=e}return e.decorators=[{type:o.Directive,args:[{selector:"[cdkCellDef]"}]}],e.ctorParameters=function(){return[{type:o.TemplateRef}]},e}(),h=function(){function e(e){this.template=e}return e.decorators=[{type:o.Directive,args:[{selector:"[cdkHeaderCellDef]"}]}],e.ctorParameters=function(){return[{type:o.TemplateRef}]},e}(),m=function(){function e(e){this.template=e}return e.decorators=[{type:o.Directive,args:[{selector:"[cdkFooterCellDef]"}]}],e.ctorParameters=function(){return[{type:o.TemplateRef}]},e}(),g=function(){return function(){}}(),v=p(g),_=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t._stickyEnd=!1,t}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"name",{get:function(){return this._name},set:function(e){e&&(this._name=e,this.cssClassFriendlyName=e.replace(/[^a-z0-9_-]/gi,"-"))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"stickyEnd",{get:function(){return this._stickyEnd},set:function(e){var t=this._stickyEnd;this._stickyEnd=Object(i.coerceBooleanProperty)(e),this._hasStickyChanged=t!==this._stickyEnd},enumerable:!0,configurable:!0}),t.decorators=[{type:o.Directive,args:[{selector:"[cdkColumnDef]",inputs:["sticky"],providers:[{provide:"MAT_SORT_HEADER_COLUMN_DEF",useExisting:t}]}]}],t.propDecorators={name:[{type:o.Input,args:["cdkColumnDef"]}],stickyEnd:[{type:o.Input,args:["stickyEnd"]}],cell:[{type:o.ContentChild,args:[f]}],headerCell:[{type:o.ContentChild,args:[h]}],footerCell:[{type:o.ContentChild,args:[m]}]},t}(v),y=function(){return function(e,t){var n="cdk-column-"+e.cssClassFriendlyName;t.nativeElement.classList.add(n)}}(),b=function(e){function t(t,n){return e.call(this,t,n)||this}return Object(r.__extends)(t,e),t.decorators=[{type:o.Directive,args:[{selector:"cdk-header-cell, th[cdk-header-cell]",host:{class:"cdk-header-cell",role:"columnheader"}}]}],t.ctorParameters=function(){return[{type:_},{type:o.ElementRef}]},t}(y),j=function(e){function t(t,n){return e.call(this,t,n)||this}return Object(r.__extends)(t,e),t.decorators=[{type:o.Directive,args:[{selector:"cdk-footer-cell, td[cdk-footer-cell]",host:{class:"cdk-footer-cell",role:"gridcell"}}]}],t.ctorParameters=function(){return[{type:_},{type:o.ElementRef}]},t}(y),x=function(e){function t(t,n){return e.call(this,t,n)||this}return Object(r.__extends)(t,e),t.decorators=[{type:o.Directive,args:[{selector:"cdk-cell, td[cdk-cell]",host:{class:"cdk-cell",role:"gridcell"}}]}],t.ctorParameters=function(){return[{type:_},{type:o.ElementRef}]},t}(y),O="<ng-container cdkCellOutlet></ng-container>",w=function(){function e(e,t){this.template=e,this._differs=t}return e.prototype.ngOnChanges=function(e){if(!this._columnsDiffer){var t=e.columns&&e.columns.currentValue||[];this._columnsDiffer=this._differs.find(t).create(),this._columnsDiffer.diff(t)}},e.prototype.getColumnsDiff=function(){return this._columnsDiffer.diff(this.columns)},e.prototype.extractCellTemplate=function(e){return this instanceof S?e.headerCell.template:this instanceof A?e.footerCell.template:e.cell.template},e}(),C=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t}(w),E=p(C),S=function(e){function t(t,n){return e.call(this,t,n)||this}return Object(r.__extends)(t,e),t.prototype.ngOnChanges=function(t){e.prototype.ngOnChanges.call(this,t)},t.decorators=[{type:o.Directive,args:[{selector:"[cdkHeaderRowDef]",inputs:["columns: cdkHeaderRowDef","sticky: cdkHeaderRowDefSticky"]}]}],t.ctorParameters=function(){return[{type:o.TemplateRef},{type:o.IterableDiffers}]},t}(E),k=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t}(w),M=p(k),A=function(e){function t(t,n){return e.call(this,t,n)||this}return Object(r.__extends)(t,e),t.prototype.ngOnChanges=function(t){e.prototype.ngOnChanges.call(this,t)},t.decorators=[{type:o.Directive,args:[{selector:"[cdkFooterRowDef]",inputs:["columns: cdkFooterRowDef","sticky: cdkFooterRowDefSticky"]}]}],t.ctorParameters=function(){return[{type:o.TemplateRef},{type:o.IterableDiffers}]},t}(M),T=function(e){function t(t,n){return e.call(this,t,n)||this}return Object(r.__extends)(t,e),t.decorators=[{type:o.Directive,args:[{selector:"[cdkRowDef]",inputs:["columns: cdkRowDefColumns","when: cdkRowDefWhen"]}]}],t.ctorParameters=function(){return[{type:o.TemplateRef},{type:o.IterableDiffers}]},t}(w),D=function(){function e(t){this._viewContainer=t,e.mostRecentCellOutlet=this}return e.prototype.ngOnDestroy=function(){e.mostRecentCellOutlet===this&&(e.mostRecentCellOutlet=null)},e.mostRecentCellOutlet=null,e.decorators=[{type:o.Directive,args:[{selector:"[cdkCellOutlet]"}]}],e.ctorParameters=function(){return[{type:o.ViewContainerRef}]},e}(),I=function(){function e(){}return e.decorators=[{type:o.Component,args:[{selector:"cdk-header-row, tr[cdk-header-row]",template:O,host:{class:"cdk-header-row",role:"row"},changeDetection:o.ChangeDetectionStrategy.OnPush,encapsulation:o.ViewEncapsulation.None}]}],e}(),P=function(){function e(){}return e.decorators=[{type:o.Component,args:[{selector:"cdk-footer-row, tr[cdk-footer-row]",template:O,host:{class:"cdk-footer-row",role:"row"},changeDetection:o.ChangeDetectionStrategy.OnPush,encapsulation:o.ViewEncapsulation.None}]}],e}(),R=function(){function e(){}return e.decorators=[{type:o.Component,args:[{selector:"cdk-row, tr[cdk-row]",template:O,host:{class:"cdk-row",role:"row"},changeDetection:o.ChangeDetectionStrategy.OnPush,encapsulation:o.ViewEncapsulation.None}]}],e}();var N=["top","bottom","left","right"],F=function(){function e(e,t,n,r){void 0===r&&(r=!0),this.isNativeHtmlTable=e,this.stickCellCss=t,this.direction=n,this._isBrowser=r}return e.prototype.clearStickyPositioning=function(e,t){for(var n=0,r=e;n<r.length;n++){var i=r[n];if(i.nodeType===i.ELEMENT_NODE){this._removeStickyStyle(i,t);for(var o=0;o<i.children.length;o++){var a=i.children[o];this._removeStickyStyle(a,t)}}}},e.prototype.updateStickyColumns=function(e,t,n){var r=t.some(function(e){return e})||n.some(function(e){return e});if(e.length&&r&&this._isBrowser)for(var i=e[0],o=i.children.length,a=this._getCellWidths(i),s=this._getStickyStartColumnPositions(a,t),u=this._getStickyEndColumnPositions(a,n),c="rtl"===this.direction,l=0,d=e;l<d.length;l++)for(var p=d[l],f=0;f<o;f++){var h=p.children[f];t[f]&&this._addStickyStyle(h,c?"right":"left",s[f]),n[f]&&this._addStickyStyle(h,c?"left":"right",u[f])}},e.prototype.stickRows=function(e,t,n){if(this._isBrowser)for(var r="bottom"===n?e.reverse():e,i=0,o=0;o<r.length;o++)if(t[o]){var a=r[o];if(this.isNativeHtmlTable)for(var s=0;s<a.children.length;s++){var u=a.children[s];this._addStickyStyle(u,n,i)}else this._addStickyStyle(a,n,i);if(o===r.length-1)return;i+=a.getBoundingClientRect().height}},e.prototype.updateStickyFooterContainer=function(e,t){if(this.isNativeHtmlTable){var n=e.querySelector("tfoot");t.some(function(e){return!e})?this._removeStickyStyle(n,["bottom"]):this._addStickyStyle(n,"bottom",0)}},e.prototype._removeStickyStyle=function(e,t){for(var n=0,r=t;n<r.length;n++){var i=r[n];e.style[i]=""}e.style.zIndex=this._getCalculatedZIndex(e),N.some(function(t){return!!e.style[t]})||(e.style.position="",e.classList.remove(this.stickCellCss))},e.prototype._addStickyStyle=function(e,t,n){e.classList.add(this.stickCellCss),e.style[t]=n+"px",e.style.cssText+="position: -webkit-sticky; position: sticky; ",e.style.zIndex=this._getCalculatedZIndex(e)},e.prototype._getCalculatedZIndex=function(e){for(var t={top:100,bottom:10,left:1,right:1},n=0,r=0,i=N;r<i.length;r++){var o=i[r];e.style[o]&&(n+=t[o])}return n?""+n:""},e.prototype._getCellWidths=function(e){for(var t=[],n=e.children,r=0;r<n.length;r++){var i=n[r];t.push(i.getBoundingClientRect().width)}return t},e.prototype._getStickyStartColumnPositions=function(e,t){for(var n=[],r=0,i=0;i<e.length;i++)t[i]&&(n[i]=r,r+=e[i]);return n},e.prototype._getStickyEndColumnPositions=function(e,t){for(var n=[],r=0,i=e.length;i>0;i--)t[i]&&(n[i]=r,r+=e[i]);return n},e}(),L=function(){function e(e,t){this.viewContainer=e,this.elementRef=t}return e.decorators=[{type:o.Directive,args:[{selector:"[rowOutlet]"}]}],e.ctorParameters=function(){return[{type:o.ViewContainerRef},{type:o.ElementRef}]},e}(),V=function(){function e(e,t){this.viewContainer=e,this.elementRef=t}return e.decorators=[{type:o.Directive,args:[{selector:"[headerRowOutlet]"}]}],e.ctorParameters=function(){return[{type:o.ViewContainerRef},{type:o.ElementRef}]},e}(),B=function(){function e(e,t){this.viewContainer=e,this.elementRef=t}return e.decorators=[{type:o.Directive,args:[{selector:"[footerRowOutlet]"}]}],e.ctorParameters=function(){return[{type:o.ViewContainerRef},{type:o.ElementRef}]},e}(),z="\n  <ng-container headerRowOutlet></ng-container>\n  <ng-container rowOutlet></ng-container>\n  <ng-container footerRowOutlet></ng-container>",U=(function(e){function t(){return null!==e&&e.apply(this,arguments)||this}Object(r.__extends)(t,e)}(o.EmbeddedViewRef),function(){function e(e,t,n,r,i,o,a){this._differs=e,this._changeDetectorRef=t,this._elementRef=n,this._dir=i,this._platform=a,this._onDestroy=new u.Subject,this._columnDefsByName=new Map,this._customColumnDefs=new Set,this._customRowDefs=new Set,this._customHeaderRowDefs=new Set,this._customFooterRowDefs=new Set,this._headerRowDefChanged=!0,this._footerRowDefChanged=!0,this._cachedRenderRowsMap=new Map,this.stickyCssClass="cdk-table-sticky",this._multiTemplateDataRows=!1,this.viewChange=new u.BehaviorSubject({start:0,end:Number.MAX_VALUE}),r||this._elementRef.nativeElement.setAttribute("role","grid"),this._document=o,this._isNativeHtmlTable="TABLE"===this._elementRef.nativeElement.nodeName}return Object.defineProperty(e.prototype,"trackBy",{get:function(){return this._trackByFn},set:function(e){Object(o.isDevMode)()&&null!=e&&"function"!=typeof e&&console&&console.warn&&console.warn("trackBy must be a function, but received "+JSON.stringify(e)+"."),this._trackByFn=e},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"dataSource",{get:function(){return this._dataSource},set:function(e){this._dataSource!==e&&this._switchDataSource(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"multiTemplateDataRows",{get:function(){return this._multiTemplateDataRows},set:function(e){this._multiTemplateDataRows=Object(i.coerceBooleanProperty)(e),this._rowOutlet.viewContainer.length&&this._forceRenderDataRows()},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){var e=this;this._setupStickyStyler(),this._isNativeHtmlTable&&this._applyNativeTableSections(),this._dataDiffer=this._differs.find([]).create(function(t,n){return e.trackBy?e.trackBy(n.dataIndex,n.data):n})},e.prototype.ngAfterContentChecked=function(){if(this._cacheRowDefs(),this._cacheColumnDefs(),!this._headerRowDefs.length&&!this._footerRowDefs.length&&!this._rowDefs.length)throw Error("Missing definitions for header, footer, and row; cannot determine which columns should be rendered.");this._renderUpdatedColumns(),this._headerRowDefChanged&&(this._forceRenderHeaderRows(),this._headerRowDefChanged=!1),this._footerRowDefChanged&&(this._forceRenderFooterRows(),this._footerRowDefChanged=!1),this.dataSource&&this._rowDefs.length>0&&!this._renderChangeSubscription&&this._observeRenderChanges(),this._checkStickyStates()},e.prototype.ngOnDestroy=function(){this._rowOutlet.viewContainer.clear(),this._headerRowOutlet.viewContainer.clear(),this._footerRowOutlet.viewContainer.clear(),this._cachedRenderRowsMap.clear(),this._onDestroy.next(),this._onDestroy.complete(),this.dataSource instanceof a.DataSource&&this.dataSource.disconnect(this)},e.prototype.renderRows=function(){var e=this;this._renderRows=this._getAllRenderRows();var t=this._dataDiffer.diff(this._renderRows);if(t){var n=this._rowOutlet.viewContainer;t.forEachOperation(function(t,r,i){if(null==t.previousIndex)e._insertRow(t.item,i);else if(null==i)n.remove(r);else{var o=n.get(r);n.move(o,i)}}),this._updateRowIndexContext(),t.forEachIdentityChange(function(e){n.get(e.currentIndex).context.$implicit=e.item.data}),this.updateStickyColumnStyles()}},e.prototype.setHeaderRowDef=function(e){this._customHeaderRowDefs=new Set([e]),this._headerRowDefChanged=!0},e.prototype.setFooterRowDef=function(e){this._customFooterRowDefs=new Set([e]),this._footerRowDefChanged=!0},e.prototype.addColumnDef=function(e){this._customColumnDefs.add(e)},e.prototype.removeColumnDef=function(e){this._customColumnDefs.delete(e)},e.prototype.addRowDef=function(e){this._customRowDefs.add(e)},e.prototype.removeRowDef=function(e){this._customRowDefs.delete(e)},e.prototype.addHeaderRowDef=function(e){this._customHeaderRowDefs.add(e),this._headerRowDefChanged=!0},e.prototype.removeHeaderRowDef=function(e){this._customHeaderRowDefs.delete(e),this._headerRowDefChanged=!0},e.prototype.addFooterRowDef=function(e){this._customFooterRowDefs.add(e),this._footerRowDefChanged=!0},e.prototype.removeFooterRowDef=function(e){this._customFooterRowDefs.delete(e),this._footerRowDefChanged=!0},e.prototype.updateStickyHeaderRowStyles=function(){var e=this._getRenderedRows(this._headerRowOutlet);this._stickyStyler.clearStickyPositioning(e,["top"]);var t=this._headerRowDefs.map(function(e){return e.sticky});this._stickyStyler.stickRows(e,t,"top"),this._headerRowDefs.forEach(function(e){return e.resetStickyChanged()})},e.prototype.updateStickyFooterRowStyles=function(){var e=this._getRenderedRows(this._footerRowOutlet);this._stickyStyler.clearStickyPositioning(e,["bottom"]);var t=this._footerRowDefs.map(function(e){return e.sticky});this._stickyStyler.stickRows(e,t,"bottom"),this._stickyStyler.updateStickyFooterContainer(this._elementRef.nativeElement,t),this._footerRowDefs.forEach(function(e){return e.resetStickyChanged()})},e.prototype.updateStickyColumnStyles=function(){var e=this,t=this._getRenderedRows(this._headerRowOutlet),n=this._getRenderedRows(this._rowOutlet),r=this._getRenderedRows(this._footerRowOutlet);this._stickyStyler.clearStickyPositioning(t.concat(n,r),["left","right"]),t.forEach(function(t,n){e._addStickyColumnStyles([t],e._headerRowDefs[n])}),this._rowDefs.forEach(function(t){for(var r=[],i=0;i<n.length;i++)e._renderRows[i].rowDef===t&&r.push(n[i]);e._addStickyColumnStyles(r,t)}),r.forEach(function(t,n){e._addStickyColumnStyles([t],e._footerRowDefs[n])}),Array.from(this._columnDefsByName.values()).forEach(function(e){return e.resetStickyChanged()})},e.prototype._getAllRenderRows=function(){var e=[],t=this._cachedRenderRowsMap;this._cachedRenderRowsMap=new Map;for(var n=0;n<this._data.length;n++){var r=this._data[n],i=this._getRenderRowsForData(r,n,t.get(r));this._cachedRenderRowsMap.has(r)||this._cachedRenderRowsMap.set(r,new WeakMap);for(var o=0;o<i.length;o++){var a=i[o],s=this._cachedRenderRowsMap.get(a.data);s.has(a.rowDef)?s.get(a.rowDef).push(a):s.set(a.rowDef,[a]),e.push(a)}}return e},e.prototype._getRenderRowsForData=function(e,t,n){return this._getRowDefs(e,t).map(function(r){var i=n&&n.has(r)?n.get(r):[];if(i.length){var o=i.shift();return o.dataIndex=t,o}return{data:e,rowDef:r,dataIndex:t}})},e.prototype._cacheColumnDefs=function(){var e=this;this._columnDefsByName.clear(),H(this._contentColumnDefs,this._customColumnDefs).forEach(function(t){if(e._columnDefsByName.has(t.name))throw n=t.name,Error('Duplicate column definition name provided: "'+n+'".');var n;e._columnDefsByName.set(t.name,t)})},e.prototype._cacheRowDefs=function(){this._headerRowDefs=H(this._contentHeaderRowDefs,this._customHeaderRowDefs),this._footerRowDefs=H(this._contentFooterRowDefs,this._customFooterRowDefs),this._rowDefs=H(this._contentRowDefs,this._customRowDefs);var e=this._rowDefs.filter(function(e){return!e.when});if(!this.multiTemplateDataRows&&e.length>1)throw Error("There can only be one default row without a when predicate function.");this._defaultRowDef=e[0]},e.prototype._renderUpdatedColumns=function(){var e=function(e,t){return e||!!t.getColumnsDiff()};this._rowDefs.reduce(e,!1)&&this._forceRenderDataRows(),this._headerRowDefs.reduce(e,!1)&&this._forceRenderHeaderRows(),this._footerRowDefs.reduce(e,!1)&&this._forceRenderFooterRows()},e.prototype._switchDataSource=function(e){this._data=[],this.dataSource instanceof a.DataSource&&this.dataSource.disconnect(this),this._renderChangeSubscription&&(this._renderChangeSubscription.unsubscribe(),this._renderChangeSubscription=null),e||(this._dataDiffer&&this._dataDiffer.diff([]),this._rowOutlet.viewContainer.clear()),this._dataSource=e},e.prototype._observeRenderChanges=function(){var e=this;if(this.dataSource){var t;if(this.dataSource.connect instanceof Function?t=this.dataSource.connect(this):this.dataSource instanceof u.Observable?t=this.dataSource:Array.isArray(this.dataSource)&&(t=Object(u.of)(this.dataSource)),void 0===t)throw Error("Provided data source did not match an array, Observable, or DataSource");this._renderChangeSubscription=t.pipe(Object(c.takeUntil)(this._onDestroy)).subscribe(function(t){e._data=t||[],e.renderRows()})}},e.prototype._forceRenderHeaderRows=function(){var e=this;this._headerRowOutlet.viewContainer.length>0&&this._headerRowOutlet.viewContainer.clear(),this._headerRowDefs.forEach(function(t,n){return e._renderRow(e._headerRowOutlet,t,n)}),this.updateStickyHeaderRowStyles(),this.updateStickyColumnStyles()},e.prototype._forceRenderFooterRows=function(){var e=this;this._footerRowOutlet.viewContainer.length>0&&this._footerRowOutlet.viewContainer.clear(),this._footerRowDefs.forEach(function(t,n){return e._renderRow(e._footerRowOutlet,t,n)}),this.updateStickyFooterRowStyles(),this.updateStickyColumnStyles()},e.prototype._addStickyColumnStyles=function(e,t){var n=this,r=Array.from(t.columns||[]).map(function(e){return n._columnDefsByName.get(e)}),i=r.map(function(e){return e.sticky}),o=r.map(function(e){return e.stickyEnd});this._stickyStyler.updateStickyColumns(e,i,o)},e.prototype._getRenderedRows=function(e){for(var t=[],n=0;n<e.viewContainer.length;n++){var r=e.viewContainer.get(n);t.push(r.rootNodes[0])}return t},e.prototype._getRowDefs=function(e,t){if(1==this._rowDefs.length)return[this._rowDefs[0]];var n,r=[];if(this.multiTemplateDataRows)r=this._rowDefs.filter(function(n){return!n.when||n.when(t,e)});else{var i=this._rowDefs.find(function(n){return n.when&&n.when(t,e)})||this._defaultRowDef;i&&r.push(i)}if(!r.length)throw n=e,Error("Could not find a matching row definition for theprovided row data: "+JSON.stringify(n));return r},e.prototype._insertRow=function(e,t){var n=e.rowDef,r={$implicit:e.data};this._renderRow(this._rowOutlet,n,t,r)},e.prototype._renderRow=function(e,t,n,r){void 0===r&&(r={}),e.viewContainer.createEmbeddedView(t.template,r,n);for(var i=0,o=this._getCellTemplates(t);i<o.length;i++){var a=o[i];D.mostRecentCellOutlet&&D.mostRecentCellOutlet._viewContainer.createEmbeddedView(a,r)}this._changeDetectorRef.markForCheck()},e.prototype._updateRowIndexContext=function(){for(var e=this._rowOutlet.viewContainer,t=0,n=e.length;t<n;t++){var r=e.get(t).context;r.count=n,r.first=0===t,r.last=t===n-1,r.even=t%2==0,r.odd=!r.even,this.multiTemplateDataRows?(r.dataIndex=this._renderRows[t].dataIndex,r.renderIndex=t):r.index=this._renderRows[t].dataIndex}},e.prototype._getCellTemplates=function(e){var t=this;return e&&e.columns?Array.from(e.columns,function(n){var r=t._columnDefsByName.get(n);if(!r)throw Error('Could not find column with id "'+n+'".');return e.extractCellTemplate(r)}):[]},e.prototype._applyNativeTableSections=function(){for(var e=0,t=[{tag:"thead",outlet:this._headerRowOutlet},{tag:"tbody",outlet:this._rowOutlet},{tag:"tfoot",outlet:this._footerRowOutlet}];e<t.length;e++){var n=t[e],r=(this._document||document).createElement(n.tag);r.appendChild(n.outlet.elementRef.nativeElement),this._elementRef.nativeElement.appendChild(r)}},e.prototype._forceRenderDataRows=function(){this._dataDiffer.diff([]),this._rowOutlet.viewContainer.clear(),this.renderRows(),this.updateStickyColumnStyles()},e.prototype._checkStickyStates=function(){var e=function(e,t){return e||t.hasStickyChanged()};this._headerRowDefs.reduce(e,!1)&&this.updateStickyHeaderRowStyles(),this._footerRowDefs.reduce(e,!1)&&this.updateStickyFooterRowStyles(),Array.from(this._columnDefsByName.values()).reduce(e,!1)&&this.updateStickyColumnStyles()},e.prototype._setupStickyStyler=function(){var e=this,t=this._dir?this._dir.value:"ltr";this._stickyStyler=new F(this._isNativeHtmlTable,this.stickyCssClass,t,!this._platform||this._platform.isBrowser),(this._dir?this._dir.change:Object(u.of)()).pipe(Object(c.takeUntil)(this._onDestroy)).subscribe(function(t){e._stickyStyler.direction=t,e.updateStickyColumnStyles()})},e.decorators=[{type:o.Component,args:[{selector:"cdk-table, table[cdk-table]",exportAs:"cdkTable",template:z,host:{class:"cdk-table"},encapsulation:o.ViewEncapsulation.None,changeDetection:o.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:o.IterableDiffers},{type:o.ChangeDetectorRef},{type:o.ElementRef},{type:String,decorators:[{type:o.Attribute,args:["role"]}]},{type:l.Directionality,decorators:[{type:o.Optional}]},{type:void 0,decorators:[{type:o.Inject,args:[s.DOCUMENT]}]},{type:d.Platform}]},e.propDecorators={trackBy:[{type:o.Input}],dataSource:[{type:o.Input}],multiTemplateDataRows:[{type:o.Input}],_rowOutlet:[{type:o.ViewChild,args:[L]}],_headerRowOutlet:[{type:o.ViewChild,args:[V]}],_footerRowOutlet:[{type:o.ViewChild,args:[B]}],_contentColumnDefs:[{type:o.ContentChildren,args:[_]}],_contentRowDefs:[{type:o.ContentChildren,args:[T]}],_contentHeaderRowDefs:[{type:o.ContentChildren,args:[S]}],_contentFooterRowDefs:[{type:o.ContentChildren,args:[A]}]},e}());function H(e,t){return e.toArray().concat(Array.from(t))}var q=[U,T,f,D,h,m,_,x,R,b,j,I,S,P,A,L,V,B],G=function(){function e(){}return e.decorators=[{type:o.NgModule,args:[{imports:[s.CommonModule],exports:q,declarations:q}]}],e}()},"./node_modules/@angular/cdk/esm5/text-field.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"AutofillMonitor",function(){return c}),n.d(t,"CdkAutofill",function(){return l}),n.d(t,"CdkTextareaAutosize",function(){return d}),n.d(t,"TextFieldModule",function(){return p});var r=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),a=n("./node_modules/rxjs/_esm5/index.js"),s=n("./node_modules/rxjs/_esm5/operators/index.js"),u=Object(r.normalizePassiveListenerOptions)({passive:!0}),c=function(){function e(e,t){this._platform=e,this._ngZone=t,this._monitoredElements=new Map}return e.prototype.monitor=function(e){var t=this;if(!this._platform.isBrowser)return a.EMPTY;var n=Object(o.coerceElement)(e),r=this._monitoredElements.get(n);if(r)return r.subject.asObservable();var i=new a.Subject,s="cdk-text-field-autofilled",c=function(e){"cdk-text-field-autofill-start"!==e.animationName||n.classList.contains(s)?"cdk-text-field-autofill-end"===e.animationName&&n.classList.contains(s)&&(n.classList.remove(s),t._ngZone.run(function(){return i.next({target:e.target,isAutofilled:!1})})):(n.classList.add(s),t._ngZone.run(function(){return i.next({target:e.target,isAutofilled:!0})}))};return this._ngZone.runOutsideAngular(function(){n.addEventListener("animationstart",c,u),n.classList.add("cdk-text-field-autofill-monitored")}),this._monitoredElements.set(n,{subject:i,unlisten:function(){n.removeEventListener("animationstart",c,u)}}),i.asObservable()},e.prototype.stopMonitoring=function(e){var t=Object(o.coerceElement)(e),n=this._monitoredElements.get(t);n&&(n.unlisten(),n.subject.complete(),t.classList.remove("cdk-text-field-autofill-monitored"),t.classList.remove("cdk-text-field-autofilled"),this._monitoredElements.delete(t))},e.prototype.ngOnDestroy=function(){var e=this;this._monitoredElements.forEach(function(t,n){return e.stopMonitoring(n)})},e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:r.Platform},{type:i.NgZone}]},e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e(Object(i.inject)(r.Platform),Object(i.inject)(i.NgZone))},token:e,providedIn:"root"}),e}(),l=function(){function e(e,t){this._elementRef=e,this._autofillMonitor=t,this.cdkAutofill=new i.EventEmitter}return e.prototype.ngOnInit=function(){var e=this;this._autofillMonitor.monitor(this._elementRef).subscribe(function(t){return e.cdkAutofill.emit(t)})},e.prototype.ngOnDestroy=function(){this._autofillMonitor.stopMonitoring(this._elementRef)},e.decorators=[{type:i.Directive,args:[{selector:"[cdkAutofill]"}]}],e.ctorParameters=function(){return[{type:i.ElementRef},{type:c}]},e.propDecorators={cdkAutofill:[{type:i.Output}]},e}(),d=function(){function e(e,t,n){this._elementRef=e,this._platform=t,this._ngZone=n,this._destroyed=new a.Subject,this._enabled=!0,this._previousMinRows=-1,this._textareaElement=this._elementRef.nativeElement}return Object.defineProperty(e.prototype,"minRows",{get:function(){return this._minRows},set:function(e){this._minRows=e,this._setMinHeight()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"maxRows",{get:function(){return this._maxRows},set:function(e){this._maxRows=e,this._setMaxHeight()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"enabled",{get:function(){return this._enabled},set:function(e){e=Object(o.coerceBooleanProperty)(e),this._enabled!==e&&((this._enabled=e)?this.resizeToFitContent(!0):this.reset())},enumerable:!0,configurable:!0}),e.prototype._setMinHeight=function(){var e=this.minRows&&this._cachedLineHeight?this.minRows*this._cachedLineHeight+"px":null;e&&(this._textareaElement.style.minHeight=e)},e.prototype._setMaxHeight=function(){var e=this.maxRows&&this._cachedLineHeight?this.maxRows*this._cachedLineHeight+"px":null;e&&(this._textareaElement.style.maxHeight=e)},e.prototype.ngAfterViewInit=function(){var e=this;this._platform.isBrowser&&(this._initialHeight=this._textareaElement.style.height,this.resizeToFitContent(),this._ngZone.runOutsideAngular(function(){Object(a.fromEvent)(window,"resize").pipe(Object(s.auditTime)(16),Object(s.takeUntil)(e._destroyed)).subscribe(function(){return e.resizeToFitContent(!0)})}))},e.prototype.ngOnDestroy=function(){this._destroyed.next(),this._destroyed.complete()},e.prototype._cacheTextareaLineHeight=function(){if(!this._cachedLineHeight){var e=this._textareaElement.cloneNode(!1);e.rows=1,e.style.position="absolute",e.style.visibility="hidden",e.style.border="none",e.style.padding="0",e.style.height="",e.style.minHeight="",e.style.maxHeight="",e.style.overflow="hidden",this._textareaElement.parentNode.appendChild(e),this._cachedLineHeight=e.clientHeight,this._textareaElement.parentNode.removeChild(e),this._setMinHeight(),this._setMaxHeight()}},e.prototype.ngDoCheck=function(){this._platform.isBrowser&&this.resizeToFitContent()},e.prototype.resizeToFitContent=function(e){var t=this;if(void 0===e&&(e=!1),this._enabled&&(this._cacheTextareaLineHeight(),this._cachedLineHeight)){var n=this._elementRef.nativeElement,r=n.value;if(e||this._minRows!==this._previousMinRows||r!==this._previousValue){var i=n.placeholder;n.classList.add("cdk-textarea-autosize-measuring"),n.placeholder="";var o=n.scrollHeight-4;n.style.height=o+"px",n.classList.remove("cdk-textarea-autosize-measuring"),n.placeholder=i,this._ngZone.runOutsideAngular(function(){"undefined"!=typeof requestAnimationFrame?requestAnimationFrame(function(){return t._scrollToCaretPosition(n)}):setTimeout(function(){return t._scrollToCaretPosition(n)})}),this._previousValue=r,this._previousMinRows=this._minRows}}},e.prototype.reset=function(){void 0!==this._initialHeight&&(this._textareaElement.style.height=this._initialHeight)},e.prototype._noopInputHandler=function(){},e.prototype._scrollToCaretPosition=function(e){var t=e.selectionStart,n=e.selectionEnd;this._destroyed.isStopped||document.activeElement!==e||e.setSelectionRange(t,n)},e.decorators=[{type:i.Directive,args:[{selector:"textarea[cdkTextareaAutosize]",exportAs:"cdkTextareaAutosize",host:{class:"cdk-textarea-autosize",rows:"1","(input)":"_noopInputHandler()"}}]}],e.ctorParameters=function(){return[{type:i.ElementRef},{type:r.Platform},{type:i.NgZone}]},e.propDecorators={minRows:[{type:i.Input,args:["cdkAutosizeMinRows"]}],maxRows:[{type:i.Input,args:["cdkAutosizeMaxRows"]}],enabled:[{type:i.Input,args:["cdkTextareaAutosize"]}]},e}(),p=function(){function e(){}return e.decorators=[{type:i.NgModule,args:[{declarations:[l,d],imports:[r.PlatformModule],exports:[l,d]}]}],e}()},"./node_modules/@angular/cdk/esm5/tree.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"BaseTreeControl",function(){return p}),n.d(t,"FlatTreeControl",function(){return f}),n.d(t,"NestedTreeControl",function(){return h}),n.d(t,"CdkNestedTreeNode",function(){return C}),n.d(t,"CdkTreeNodeOutletContext",function(){return m}),n.d(t,"CdkTreeNodeDef",function(){return g}),n.d(t,"CdkTreeNodePadding",function(){return S}),n.d(t,"CdkTreeNodeOutlet",function(){return v}),n.d(t,"CdkTree",function(){return O}),n.d(t,"CdkTreeNode",function(){return w}),n.d(t,"getTreeNoValidDataSourceError",function(){return _}),n.d(t,"getTreeMultipleDefaultNodeDefsError",function(){return y}),n.d(t,"getTreeMissingMatchingNodeDefError",function(){return b}),n.d(t,"getTreeControlMissingError",function(){return j}),n.d(t,"getTreeControlFunctionsMissingError",function(){return x}),n.d(t,"CdkTreeModule",function(){return A}),n.d(t,"CdkTreeNodeToggle",function(){return k});var r=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/rxjs/_esm5/index.js"),a=n("./node_modules/rxjs/_esm5/operators/index.js"),s=n("./node_modules/@angular/core/fesm5/core.js"),u=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),c=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),l=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),d=n("./node_modules/@angular/common/fesm5/common.js"),p=function(){function e(){this.expansionModel=new r.SelectionModel(!0)}return e.prototype.toggle=function(e){this.expansionModel.toggle(e)},e.prototype.expand=function(e){this.expansionModel.select(e)},e.prototype.collapse=function(e){this.expansionModel.deselect(e)},e.prototype.isExpanded=function(e){return this.expansionModel.isSelected(e)},e.prototype.toggleDescendants=function(e){this.expansionModel.isSelected(e)?this.collapseDescendants(e):this.expandDescendants(e)},e.prototype.collapseAll=function(){this.expansionModel.clear()},e.prototype.expandDescendants=function(e){var t,n=[e];n.push.apply(n,this.getDescendants(e)),(t=this.expansionModel).select.apply(t,n)},e.prototype.collapseDescendants=function(e){var t,n=[e];n.push.apply(n,this.getDescendants(e)),(t=this.expansionModel).deselect.apply(t,n)},e}(),f=function(e){function t(t,n){var r=e.call(this)||this;return r.getLevel=t,r.isExpandable=n,r}return Object(i.__extends)(t,e),t.prototype.getDescendants=function(e){for(var t=[],n=this.dataNodes.indexOf(e)+1;n<this.dataNodes.length&&this.getLevel(e)<this.getLevel(this.dataNodes[n]);n++)t.push(this.dataNodes[n]);return t},t.prototype.expandAll=function(){var e;(e=this.expansionModel).select.apply(e,this.dataNodes)},t}(p),h=function(e){function t(t){var n=e.call(this)||this;return n.getChildren=t,n}return Object(i.__extends)(t,e),t.prototype.expandAll=function(){var e,t=this;this.expansionModel.clear();var n=this.dataNodes.reduce(function(e,n){return e.concat(t.getDescendants(n),[n])},[]);(e=this.expansionModel).select.apply(e,n)},t.prototype.getDescendants=function(e){var t=[];return this._getDescendants(t,e),t.splice(1)},t.prototype._getDescendants=function(e,t){var n=this;e.push(t);var r=this.getChildren(t);Array.isArray(r)?r.forEach(function(t){return n._getDescendants(e,t)}):r instanceof o.Observable&&r.pipe(Object(a.take)(1),Object(a.filter)(Boolean)).subscribe(function(t){t.forEach(function(t){return n._getDescendants(e,t)})})},t}(p),m=function(){return function(e){this.$implicit=e}}(),g=function(){function e(e){this.template=e}return e.decorators=[{type:s.Directive,args:[{selector:"[cdkTreeNodeDef]",inputs:["when: cdkTreeNodeDefWhen"]}]}],e.ctorParameters=function(){return[{type:s.TemplateRef}]},e}(),v=function(){function e(e){this.viewContainer=e}return e.decorators=[{type:s.Directive,args:[{selector:"[cdkTreeNodeOutlet]"}]}],e.ctorParameters=function(){return[{type:s.ViewContainerRef}]},e}();function _(){return Error("A valid data source must be provided.")}function y(){return Error("There can only be one default row without a when predicate function.")}function b(){return Error("Could not find a matching node definition for the provided node data.")}function j(){return Error("Could not find a tree control for the tree.")}function x(){return Error("Could not find functions for nested/flat tree in tree control.")}var O=function(){function e(e,t){this._differs=e,this._changeDetectorRef=t,this._onDestroy=new o.Subject,this._levels=new Map,this.viewChange=new o.BehaviorSubject({start:0,end:Number.MAX_VALUE})}return Object.defineProperty(e.prototype,"dataSource",{get:function(){return this._dataSource},set:function(e){this._dataSource!==e&&this._switchDataSource(e)},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){if(this._dataDiffer=this._differs.find([]).create(this.trackBy),!this.treeControl)throw j()},e.prototype.ngOnDestroy=function(){this._nodeOutlet.viewContainer.clear(),this._onDestroy.next(),this._onDestroy.complete(),this._dataSource&&"function"==typeof this._dataSource.disconnect&&this.dataSource.disconnect(this),this._dataSubscription&&(this._dataSubscription.unsubscribe(),this._dataSubscription=null)},e.prototype.ngAfterContentChecked=function(){var e=this._nodeDefs.filter(function(e){return!e.when});if(e.length>1)throw y();this._defaultNodeDef=e[0],this.dataSource&&this._nodeDefs&&!this._dataSubscription&&this._observeRenderChanges()},e.prototype._switchDataSource=function(e){this._dataSource&&"function"==typeof this._dataSource.disconnect&&this.dataSource.disconnect(this),this._dataSubscription&&(this._dataSubscription.unsubscribe(),this._dataSubscription=null),e||this._nodeOutlet.viewContainer.clear(),this._dataSource=e,this._nodeDefs&&this._observeRenderChanges()},e.prototype._observeRenderChanges=function(){var e,t=this;if("function"==typeof this._dataSource.connect?e=this._dataSource.connect(this):this._dataSource instanceof o.Observable?e=this._dataSource:Array.isArray(this._dataSource)&&(e=Object(o.of)(this._dataSource)),!e)throw _();this._dataSubscription=e.pipe(Object(a.takeUntil)(this._onDestroy)).subscribe(function(e){return t.renderNodeChanges(e)})},e.prototype.renderNodeChanges=function(e,t,n,r){var i=this;void 0===t&&(t=this._dataDiffer),void 0===n&&(n=this._nodeOutlet.viewContainer);var o=t.diff(e);o&&(o.forEachOperation(function(t,o,a){if(null==t.previousIndex)i.insertNode(e[a],a,n,r);else if(null==a)n.remove(o),i._levels.delete(t.item);else{var s=n.get(o);n.move(s,a)}}),this._changeDetectorRef.detectChanges())},e.prototype._getNodeDef=function(e,t){if(1===this._nodeDefs.length)return this._nodeDefs.first;var n=this._nodeDefs.find(function(n){return n.when&&n.when(t,e)})||this._defaultNodeDef;if(!n)throw b();return n},e.prototype.insertNode=function(e,t,n,r){var i=this._getNodeDef(e,t),o=new m(e);this.treeControl.getLevel?o.level=this.treeControl.getLevel(e):void 0!==r&&this._levels.has(r)?o.level=this._levels.get(r)+1:o.level=0,this._levels.set(e,o.level),(n||this._nodeOutlet.viewContainer).createEmbeddedView(i.template,o,t),w.mostRecentTreeNode&&(w.mostRecentTreeNode.data=e)},e.decorators=[{type:s.Component,args:[{selector:"cdk-tree",exportAs:"cdkTree",template:"<ng-container cdkTreeNodeOutlet></ng-container>",host:{class:"cdk-tree",role:"tree"},encapsulation:s.ViewEncapsulation.None,changeDetection:s.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:s.IterableDiffers},{type:s.ChangeDetectorRef}]},e.propDecorators={dataSource:[{type:s.Input}],treeControl:[{type:s.Input}],trackBy:[{type:s.Input}],_nodeOutlet:[{type:s.ViewChild,args:[v]}],_nodeDefs:[{type:s.ContentChildren,args:[g]}]},e}(),w=function(){function e(t,n){this._elementRef=t,this._tree=n,this._destroyed=new o.Subject,this.role="treeitem",e.mostRecentTreeNode=this}return Object.defineProperty(e.prototype,"data",{get:function(){return this._data},set:function(e){this._data=e,this._setRoleFromData()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"isExpanded",{get:function(){return this._tree.treeControl.isExpanded(this._data)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"level",{get:function(){return this._tree.treeControl.getLevel?this._tree.treeControl.getLevel(this._data):0},enumerable:!0,configurable:!0}),e.prototype.ngOnDestroy=function(){e.mostRecentTreeNode===this&&(e.mostRecentTreeNode=null),this._destroyed.next(),this._destroyed.complete()},e.prototype.focus=function(){this._elementRef.nativeElement.focus()},e.prototype._setRoleFromData=function(){var e=this;if(this._tree.treeControl.isExpandable)this.role=this._tree.treeControl.isExpandable(this._data)?"group":"treeitem";else{if(!this._tree.treeControl.getChildren)throw x();var t=this._tree.treeControl.getChildren(this._data);Array.isArray(t)?this._setRoleFromChildren(t):t instanceof o.Observable&&t.pipe(Object(a.takeUntil)(this._destroyed)).subscribe(function(t){return e._setRoleFromChildren(t)})}},e.prototype._setRoleFromChildren=function(e){this.role=e&&e.length?"group":"treeitem"},e.mostRecentTreeNode=null,e.decorators=[{type:s.Directive,args:[{selector:"cdk-tree-node",exportAs:"cdkTreeNode",host:{"[attr.aria-expanded]":"isExpanded","[attr.aria-level]":'role === "treeitem" ? level : null',"[attr.role]":"role",class:"cdk-tree-node"}}]}],e.ctorParameters=function(){return[{type:s.ElementRef},{type:O}]},e.propDecorators={role:[{type:s.Input}]},e}(),C=function(e){function t(t,n,r){var i=e.call(this,t,n)||this;return i._elementRef=t,i._tree=n,i._differs=r,i}return Object(i.__extends)(t,e),t.prototype.ngAfterContentInit=function(){var e=this;if(this._dataDiffer=this._differs.find([]).create(this._tree.trackBy),!this._tree.treeControl.getChildren)throw x();var t=this._tree.treeControl.getChildren(this.data);Array.isArray(t)?this.updateChildrenNodes(t):t instanceof o.Observable&&t.pipe(Object(a.takeUntil)(this._destroyed)).subscribe(function(t){return e.updateChildrenNodes(t)}),this.nodeOutlet.changes.pipe(Object(a.takeUntil)(this._destroyed)).subscribe(function(){return e.updateChildrenNodes()})},t.prototype.ngOnDestroy=function(){this._clear(),e.prototype.ngOnDestroy.call(this)},t.prototype.updateChildrenNodes=function(e){if(e&&(this._children=e),this.nodeOutlet.length&&this._children){var t=this.nodeOutlet.first.viewContainer;this._tree.renderNodeChanges(this._children,this._dataDiffer,t,this._data)}else this._dataDiffer.diff([])},t.prototype._clear=function(){this.nodeOutlet&&this.nodeOutlet.first&&(this.nodeOutlet.first.viewContainer.clear(),this._dataDiffer.diff([]))},t.decorators=[{type:s.Directive,args:[{selector:"cdk-nested-tree-node",exportAs:"cdkNestedTreeNode",host:{"[attr.aria-expanded]":"isExpanded","[attr.role]":"role",class:"cdk-tree-node cdk-nested-tree-node"},providers:[{provide:w,useExisting:t}]}]}],t.ctorParameters=function(){return[{type:s.ElementRef},{type:O},{type:s.IterableDiffers}]},t.propDecorators={nodeOutlet:[{type:s.ContentChildren,args:[v]}]},t}(w),E=/([A-Za-z%]+)$/,S=function(){function e(e,t,n,r,i){var s=this;this._treeNode=e,this._tree=t,this._renderer=n,this._element=r,this._dir=i,this._destroyed=new o.Subject,this.indentUnits="px",this._indent=40,this._setPadding(),this._dir&&this._dir.change.pipe(Object(a.takeUntil)(this._destroyed)).subscribe(function(){return s._setPadding()})}return Object.defineProperty(e.prototype,"level",{get:function(){return this._level},set:function(e){this._level=Object(c.coerceNumberProperty)(e),this._setPadding()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"indent",{get:function(){return this._indent},set:function(e){var t=e,n="px";if("string"==typeof e){var r=e.split(E);t=r[0],n=r[1]||n}this.indentUnits=n,this._indent=Object(c.coerceNumberProperty)(t),this._setPadding()},enumerable:!0,configurable:!0}),e.prototype.ngOnDestroy=function(){this._destroyed.next(),this._destroyed.complete()},e.prototype._paddingIndent=function(){var e=this._treeNode.data&&this._tree.treeControl.getLevel?this._tree.treeControl.getLevel(this._treeNode.data):null,t=this._level||e;return t?""+t*this._indent+this.indentUnits:null},e.prototype._setPadding=function(){var e=this._paddingIndent(),t=this._dir&&"rtl"===this._dir.value?"paddingRight":"paddingLeft";this._renderer.setStyle(this._element.nativeElement,t,e)},e.decorators=[{type:s.Directive,args:[{selector:"[cdkTreeNodePadding]"}]}],e.ctorParameters=function(){return[{type:w},{type:O},{type:s.Renderer2},{type:s.ElementRef},{type:u.Directionality,decorators:[{type:s.Optional}]}]},e.propDecorators={level:[{type:s.Input,args:["cdkTreeNodePadding"]}],indent:[{type:s.Input,args:["cdkTreeNodePaddingIndent"]}]},e}(),k=function(){function e(e,t){this._tree=e,this._treeNode=t,this._recursive=!1}return Object.defineProperty(e.prototype,"recursive",{get:function(){return this._recursive},set:function(e){this._recursive=Object(c.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),e.prototype._toggle=function(e){this.recursive?this._tree.treeControl.toggleDescendants(this._treeNode.data):this._tree.treeControl.toggle(this._treeNode.data),e.stopPropagation()},e.decorators=[{type:s.Directive,args:[{selector:"[cdkTreeNodeToggle]",host:{"(click)":"_toggle($event)"}}]}],e.ctorParameters=function(){return[{type:O},{type:w}]},e.propDecorators={recursive:[{type:s.Input,args:["cdkTreeNodeToggleRecursive"]}]},e}(),M=[C,g,S,k,O,w,v],A=function(){function e(){}return e.decorators=[{type:s.NgModule,args:[{imports:[d.CommonModule],exports:M,declarations:M,providers:[l.FocusMonitor,g]}]}],e}()},"./node_modules/@angular/common/fesm5/common.js":function(e,t,n){"use strict";n.r(t),n.d(t,"ɵangular_packages_common_common_e",function(){return at}),n.d(t,"ɵangular_packages_common_common_d",function(){return Z}),n.d(t,"ɵangular_packages_common_common_a",function(){return Le}),n.d(t,"ɵangular_packages_common_common_b",function(){return Ue}),n.d(t,"ɵangular_packages_common_common_g",function(){return At}),n.d(t,"ɵangular_packages_common_common_f",function(){return $t}),n.d(t,"ɵregisterLocaleData",function(){return m}),n.d(t,"formatDate",function(){return le}),n.d(t,"formatCurrency",function(){return Ie}),n.d(t,"formatNumber",function(){return Re}),n.d(t,"formatPercent",function(){return Pe}),n.d(t,"NgLocaleLocalization",function(){return ze}),n.d(t,"NgLocalization",function(){return Ve}),n.d(t,"registerLocaleData",function(){return m}),n.d(t,"Plural",function(){return _}),n.d(t,"NumberFormatStyle",function(){return g}),n.d(t,"FormStyle",function(){return b}),n.d(t,"TranslationWidth",function(){return x}),n.d(t,"FormatWidth",function(){return w}),n.d(t,"NumberSymbol",function(){return E}),n.d(t,"WeekDay",function(){return k}),n.d(t,"getNumberOfCurrencyDigits",function(){return ee}),n.d(t,"getCurrencySymbol",function(){return $}),n.d(t,"getLocaleDayPeriods",function(){return D}),n.d(t,"getLocaleDayNames",function(){return I}),n.d(t,"getLocaleMonthNames",function(){return P}),n.d(t,"getLocaleId",function(){return T}),n.d(t,"getLocaleEraNames",function(){return R}),n.d(t,"getLocaleWeekEndRange",function(){return F}),n.d(t,"getLocaleFirstDayOfWeek",function(){return N}),n.d(t,"getLocaleDateFormat",function(){return L}),n.d(t,"getLocaleDateTimeFormat",function(){return B}),n.d(t,"getLocaleExtraDayPeriodRules",function(){return Y}),n.d(t,"getLocaleExtraDayPeriods",function(){return K}),n.d(t,"getLocalePluralCase",function(){return G}),n.d(t,"getLocaleTimeFormat",function(){return V}),n.d(t,"getLocaleNumberSymbol",function(){return z}),n.d(t,"getLocaleNumberFormat",function(){return U}),n.d(t,"getLocaleCurrencyName",function(){return q}),n.d(t,"getLocaleCurrencySymbol",function(){return H}),n.d(t,"ɵparseCookieValue",function(){return He}),n.d(t,"CommonModule",function(){return Jt}),n.d(t,"DeprecatedI18NPipesModule",function(){return tn}),n.d(t,"NgClass",function(){return qe}),n.d(t,"NgForOf",function(){return Ye}),n.d(t,"NgForOfContext",function(){return We}),n.d(t,"NgIf",function(){return Qe}),n.d(t,"NgIfContext",function(){return Xe}),n.d(t,"NgPlural",function(){return nt}),n.d(t,"NgPluralCase",function(){return rt}),n.d(t,"NgStyle",function(){return it}),n.d(t,"NgSwitch",function(){return Je}),n.d(t,"NgSwitchCase",function(){return et}),n.d(t,"NgSwitchDefault",function(){return tt}),n.d(t,"NgTemplateOutlet",function(){return ot}),n.d(t,"NgComponentOutlet",function(){return Ge}),n.d(t,"DOCUMENT",function(){return nn}),n.d(t,"AsyncPipe",function(){return Pt}),n.d(t,"DatePipe",function(){return Vt}),n.d(t,"I18nPluralPipe",function(){return zt}),n.d(t,"I18nSelectPipe",function(){return Ut}),n.d(t,"JsonPipe",function(){return Ht}),n.d(t,"LowerCasePipe",function(){return Rt}),n.d(t,"CurrencyPipe",function(){return Kt}),n.d(t,"DecimalPipe",function(){return Wt}),n.d(t,"PercentPipe",function(){return Yt}),n.d(t,"SlicePipe",function(){return Zt}),n.d(t,"UpperCasePipe",function(){return Lt}),n.d(t,"TitleCasePipe",function(){return Ft}),n.d(t,"KeyValuePipe",function(){return qt}),n.d(t,"DeprecatedDatePipe",function(){return wt}),n.d(t,"DeprecatedCurrencyPipe",function(){return Mt}),n.d(t,"DeprecatedDecimalPipe",function(){return St}),n.d(t,"DeprecatedPercentPipe",function(){return kt}),n.d(t,"ɵPLATFORM_BROWSER_ID",function(){return rn}),n.d(t,"ɵPLATFORM_SERVER_ID",function(){return on}),n.d(t,"ɵPLATFORM_WORKER_APP_ID",function(){return an}),n.d(t,"ɵPLATFORM_WORKER_UI_ID",function(){return sn}),n.d(t,"isPlatformBrowser",function(){return un}),n.d(t,"isPlatformServer",function(){return cn}),n.d(t,"isPlatformWorkerApp",function(){return ln}),n.d(t,"isPlatformWorkerUi",function(){return dn}),n.d(t,"VERSION",function(){return pn}),n.d(t,"ViewportScroller",function(){return fn}),n.d(t,"ɵNullViewportScroller",function(){return mn}),n.d(t,"PlatformLocation",function(){return o}),n.d(t,"LOCATION_INITIALIZED",function(){return a}),n.d(t,"LocationStrategy",function(){return s}),n.d(t,"APP_BASE_HREF",function(){return u}),n.d(t,"HashLocationStrategy",function(){return d}),n.d(t,"PathLocationStrategy",function(){return p}),n.d(t,"Location",function(){return c});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/common/node_modules/tslib/tslib.es6.js"),o=function(){return function(){}}(),a=new r.InjectionToken("Location Initialized"),s=function(){return function(){}}(),u=new r.InjectionToken("appBaseHref"),c=function(){function e(e){var n=this;this._subject=new r.EventEmitter,this._platformStrategy=e;var i=this._platformStrategy.getBaseHref();this._baseHref=t.stripTrailingSlash(l(i)),this._platformStrategy.onPopState(function(e){n._subject.emit({url:n.path(!0),pop:!0,state:e.state,type:e.type})})}var t;return t=e,e.prototype.path=function(e){return void 0===e&&(e=!1),this.normalize(this._platformStrategy.path(e))},e.prototype.isCurrentPathEqualTo=function(e,n){return void 0===n&&(n=""),this.path()==this.normalize(e+t.normalizeQueryParams(n))},e.prototype.normalize=function(e){return t.stripTrailingSlash((n=this._baseHref,r=l(e),n&&r.startsWith(n)?r.substring(n.length):r));var n,r},e.prototype.prepareExternalUrl=function(e){return e&&"/"!==e[0]&&(e="/"+e),this._platformStrategy.prepareExternalUrl(e)},e.prototype.go=function(e,t,n){void 0===t&&(t=""),void 0===n&&(n=null),this._platformStrategy.pushState(n,"",e,t)},e.prototype.replaceState=function(e,t,n){void 0===t&&(t=""),void 0===n&&(n=null),this._platformStrategy.replaceState(n,"",e,t)},e.prototype.forward=function(){this._platformStrategy.forward()},e.prototype.back=function(){this._platformStrategy.back()},e.prototype.subscribe=function(e,t,n){return this._subject.subscribe({next:e,error:t,complete:n})},e.normalizeQueryParams=function(e){return e&&"?"!==e[0]?"?"+e:e},e.joinWithSlash=function(e,t){if(0==e.length)return t;if(0==t.length)return e;var n=0;return e.endsWith("/")&&n++,t.startsWith("/")&&n++,2==n?e+t.substring(1):1==n?e+t:e+"/"+t},e.stripTrailingSlash=function(e){var t=e.match(/#|\?|$/),n=t&&t.index||e.length,r=n-("/"===e[n-1]?1:0);return e.slice(0,r)+e.slice(n)},e=t=Object(i.__decorate)([Object(r.Injectable)(),Object(i.__metadata)("design:paramtypes",[s])],e)}();function l(e){return e.replace(/\/index.html$/,"")}var d=function(e){function t(t,n){var r=e.call(this)||this;return r._platformLocation=t,r._baseHref="",null!=n&&(r._baseHref=n),r}return Object(i.__extends)(t,e),t.prototype.onPopState=function(e){this._platformLocation.onPopState(e),this._platformLocation.onHashChange(e)},t.prototype.getBaseHref=function(){return this._baseHref},t.prototype.path=function(e){void 0===e&&(e=!1);var t=this._platformLocation.hash;return null==t&&(t="#"),t.length>0?t.substring(1):t},t.prototype.prepareExternalUrl=function(e){var t=c.joinWithSlash(this._baseHref,e);return t.length>0?"#"+t:t},t.prototype.pushState=function(e,t,n,r){var i=this.prepareExternalUrl(n+c.normalizeQueryParams(r));0==i.length&&(i=this._platformLocation.pathname),this._platformLocation.pushState(e,t,i)},t.prototype.replaceState=function(e,t,n,r){var i=this.prepareExternalUrl(n+c.normalizeQueryParams(r));0==i.length&&(i=this._platformLocation.pathname),this._platformLocation.replaceState(e,t,i)},t.prototype.forward=function(){this._platformLocation.forward()},t.prototype.back=function(){this._platformLocation.back()},t=Object(i.__decorate)([Object(r.Injectable)(),Object(i.__param)(1,Object(r.Optional)()),Object(i.__param)(1,Object(r.Inject)(u)),Object(i.__metadata)("design:paramtypes",[o,String])],t)}(s),p=function(e){function t(t,n){var r=e.call(this)||this;if(r._platformLocation=t,null==n&&(n=r._platformLocation.getBaseHrefFromDOM()),null==n)throw new Error("No base href set. Please provide a value for the APP_BASE_HREF token or add a base element to the document.");return r._baseHref=n,r}return Object(i.__extends)(t,e),t.prototype.onPopState=function(e){this._platformLocation.onPopState(e),this._platformLocation.onHashChange(e)},t.prototype.getBaseHref=function(){return this._baseHref},t.prototype.prepareExternalUrl=function(e){return c.joinWithSlash(this._baseHref,e)},t.prototype.path=function(e){void 0===e&&(e=!1);var t=this._platformLocation.pathname+c.normalizeQueryParams(this._platformLocation.search),n=this._platformLocation.hash;return n&&e?""+t+n:t},t.prototype.pushState=function(e,t,n,r){var i=this.prepareExternalUrl(n+c.normalizeQueryParams(r));this._platformLocation.pushState(e,t,i)},t.prototype.replaceState=function(e,t,n,r){var i=this.prepareExternalUrl(n+c.normalizeQueryParams(r));this._platformLocation.replaceState(e,t,i)},t.prototype.forward=function(){this._platformLocation.forward()},t.prototype.back=function(){this._platformLocation.back()},t=Object(i.__decorate)([Object(r.Injectable)(),Object(i.__param)(1,Object(r.Optional)()),Object(i.__param)(1,Object(r.Inject)(u)),Object(i.__metadata)("design:paramtypes",[o,String])],t)}(s);var f=["en",[["a","p"],["AM","PM"],void 0],[["AM","PM"],void 0,void 0],[["S","M","T","W","T","F","S"],["Sun","Mon","Tue","Wed","Thu","Fri","Sat"],["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"],["Su","Mo","Tu","We","Th","Fr","Sa"]],void 0,[["J","F","M","A","M","J","J","A","S","O","N","D"],["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"],["January","February","March","April","May","June","July","August","September","October","November","December"]],void 0,[["B","A"],["BC","AD"],["Before Christ","Anno Domini"]],0,[6,0],["M/d/yy","MMM d, y","MMMM d, y","EEEE, MMMM d, y"],["h:mm a","h:mm:ss a","h:mm:ss a z","h:mm:ss a zzzz"],["{1}, {0}",void 0,"{1} 'at' {0}",void 0],[".",",",";","%","+","-","E","×","‰","∞","NaN",":"],["#,##0.###","#,##0%","¤#,##0.00","#E0"],"$","US Dollar",{},function(e){var t=Math.floor(Math.abs(e)),n=e.toString().replace(/^[^.]*\.?/,"").length;return 1===t&&0===n?1:5}],h={};function m(e,t,n){"string"!=typeof t&&(n=t,t=e[0]),t=t.toLowerCase().replace(/_/g,"-"),h[t]=e,n&&(h[t][19]=n)}var g,v,_,y,b,j,x,O,w,C,E,S,k,M,A={ADP:[void 0,void 0,0],AFN:[void 0,void 0,0],ALL:[void 0,void 0,0],AMD:[void 0,void 0,0],AOA:[void 0,"Kz"],ARS:[void 0,"$"],AUD:["A$","$"],BAM:[void 0,"KM"],BBD:[void 0,"$"],BDT:[void 0,"৳"],BHD:[void 0,void 0,3],BIF:[void 0,void 0,0],BMD:[void 0,"$"],BND:[void 0,"$"],BOB:[void 0,"Bs"],BRL:["R$"],BSD:[void 0,"$"],BWP:[void 0,"P"],BYN:[void 0,"р.",2],BYR:[void 0,void 0,0],BZD:[void 0,"$"],CAD:["CA$","$",2],CHF:[void 0,void 0,2],CLF:[void 0,void 0,4],CLP:[void 0,"$",0],CNY:["CN¥","¥"],COP:[void 0,"$",0],CRC:[void 0,"₡",2],CUC:[void 0,"$"],CUP:[void 0,"$"],CZK:[void 0,"Kč",2],DJF:[void 0,void 0,0],DKK:[void 0,"kr",2],DOP:[void 0,"$"],EGP:[void 0,"E£"],ESP:[void 0,"₧",0],EUR:["€"],FJD:[void 0,"$"],FKP:[void 0,"£"],GBP:["£"],GEL:[void 0,"₾"],GIP:[void 0,"£"],GNF:[void 0,"FG",0],GTQ:[void 0,"Q"],GYD:[void 0,"$",0],HKD:["HK$","$"],HNL:[void 0,"L"],HRK:[void 0,"kn"],HUF:[void 0,"Ft",2],IDR:[void 0,"Rp",0],ILS:["₪"],INR:["₹"],IQD:[void 0,void 0,0],IRR:[void 0,void 0,0],ISK:[void 0,"kr",0],ITL:[void 0,void 0,0],JMD:[void 0,"$"],JOD:[void 0,void 0,3],JPY:["¥",void 0,0],KHR:[void 0,"៛"],KMF:[void 0,"CF",0],KPW:[void 0,"₩",0],KRW:["₩",void 0,0],KWD:[void 0,void 0,3],KYD:[void 0,"$"],KZT:[void 0,"₸"],LAK:[void 0,"₭",0],LBP:[void 0,"L£",0],LKR:[void 0,"Rs"],LRD:[void 0,"$"],LTL:[void 0,"Lt"],LUF:[void 0,void 0,0],LVL:[void 0,"Ls"],LYD:[void 0,void 0,3],MGA:[void 0,"Ar",0],MGF:[void 0,void 0,0],MMK:[void 0,"K",0],MNT:[void 0,"₮",0],MRO:[void 0,void 0,0],MUR:[void 0,"Rs",0],MXN:["MX$","$"],MYR:[void 0,"RM"],NAD:[void 0,"$"],NGN:[void 0,"₦"],NIO:[void 0,"C$"],NOK:[void 0,"kr",2],NPR:[void 0,"Rs"],NZD:["NZ$","$"],OMR:[void 0,void 0,3],PHP:[void 0,"₱"],PKR:[void 0,"Rs",0],PLN:[void 0,"zł"],PYG:[void 0,"₲",0],RON:[void 0,"lei"],RSD:[void 0,void 0,0],RUB:[void 0,"₽"],RUR:[void 0,"р."],RWF:[void 0,"RF",0],SBD:[void 0,"$"],SEK:[void 0,"kr",2],SGD:[void 0,"$"],SHP:[void 0,"£"],SLL:[void 0,void 0,0],SOS:[void 0,void 0,0],SRD:[void 0,"$"],SSP:[void 0,"£"],STD:[void 0,void 0,0],STN:[void 0,"Db"],SYP:[void 0,"£",0],THB:[void 0,"฿"],TMM:[void 0,void 0,0],TND:[void 0,void 0,3],TOP:[void 0,"T$"],TRL:[void 0,void 0,0],TRY:[void 0,"₺"],TTD:[void 0,"$"],TWD:["NT$","$",2],TZS:[void 0,void 0,0],UAH:[void 0,"₴"],UGX:[void 0,void 0,0],USD:["$"],UYI:[void 0,void 0,0],UYU:[void 0,"$"],UZS:[void 0,void 0,0],VEF:[void 0,"Bs"],VND:["₫",void 0,0],VUV:[void 0,void 0,0],XAF:["FCFA",void 0,0],XCD:["EC$","$"],XOF:["CFA",void 0,0],XPF:["CFPF",void 0,0],YER:[void 0,void 0,0],ZAR:[void 0,"R"],ZMK:[void 0,void 0,0],ZMW:[void 0,"ZK"],ZWD:[void 0,void 0,0]};function T(e){return Z(e)[0]}function D(e,t,n){var r=Z(e);return Q(Q([r[1],r[2]],t),n)}function I(e,t,n){var r=Z(e);return Q(Q([r[3],r[4]],t),n)}function P(e,t,n){var r=Z(e);return Q(Q([r[5],r[6]],t),n)}function R(e,t){return Q(Z(e)[7],t)}function N(e){return Z(e)[8]}function F(e){return Z(e)[9]}function L(e,t){return Q(Z(e)[10],t)}function V(e,t){return Q(Z(e)[11],t)}function B(e,t){return Q(Z(e)[12],t)}function z(e,t){var n=Z(e),r=n[13][t];if(void 0===r){if(t===E.CurrencyDecimal)return n[13][E.Decimal];if(t===E.CurrencyGroup)return n[13][E.Group]}return r}function U(e,t){return Z(e)[14][t]}function H(e){return Z(e)[15]||null}function q(e){return Z(e)[16]||null}function G(e){return Z(e)[18]}function W(e){if(!e[19])throw new Error('Missing extra locale data for the locale "'+e[0]+'". Use "registerLocaleData" to load new data. See the "I18n guide" on angular.io to know more.')}function Y(e){var t=Z(e);return W(t),(t[19][2]||[]).map(function(e){return"string"==typeof e?X(e):[X(e[0]),X(e[1])]})}function K(e,t,n){var r=Z(e);return W(r),Q(Q([r[19][0],r[19][1]],t)||[],n)||[]}function Q(e,t){for(var n=t;n>-1;n--)if(void 0!==e[n])return e[n];throw new Error("Locale data API: locale data undefined")}function X(e){var t=Object(i.__read)(e.split(":"),2);return{hours:+t[0],minutes:+t[1]}}function Z(e){var t=e.toLowerCase().replace(/_/g,"-"),n=h[t];if(n)return n;var r=t.split("-")[0];if(n=h[r])return n;if("en"===r)return f;throw new Error('Missing locale data for the locale "'+e+'".')}function $(e,t,n){void 0===n&&(n="en");var r,i=(r=n,Z(r)[17])[e]||A[e]||[],o=i[1];return"narrow"===t&&"string"==typeof o?o:i[0]||e}(v=g||(g={}))[v.Decimal=0]="Decimal",v[v.Percent=1]="Percent",v[v.Currency=2]="Currency",v[v.Scientific=3]="Scientific",(y=_||(_={}))[y.Zero=0]="Zero",y[y.One=1]="One",y[y.Two=2]="Two",y[y.Few=3]="Few",y[y.Many=4]="Many",y[y.Other=5]="Other",(j=b||(b={}))[j.Format=0]="Format",j[j.Standalone=1]="Standalone",(O=x||(x={}))[O.Narrow=0]="Narrow",O[O.Abbreviated=1]="Abbreviated",O[O.Wide=2]="Wide",O[O.Short=3]="Short",(C=w||(w={}))[C.Short=0]="Short",C[C.Medium=1]="Medium",C[C.Long=2]="Long",C[C.Full=3]="Full",(S=E||(E={}))[S.Decimal=0]="Decimal",S[S.Group=1]="Group",S[S.List=2]="List",S[S.PercentSign=3]="PercentSign",S[S.PlusSign=4]="PlusSign",S[S.MinusSign=5]="MinusSign",S[S.Exponential=6]="Exponential",S[S.SuperscriptingExponent=7]="SuperscriptingExponent",S[S.PerMille=8]="PerMille",S[S[1/0]=9]="Infinity",S[S.NaN=10]="NaN",S[S.TimeSeparator=11]="TimeSeparator",S[S.CurrencyDecimal=12]="CurrencyDecimal",S[S.CurrencyGroup=13]="CurrencyGroup",(M=k||(k={}))[M.Sunday=0]="Sunday",M[M.Monday=1]="Monday",M[M.Tuesday=2]="Tuesday",M[M.Wednesday=3]="Wednesday",M[M.Thursday=4]="Thursday",M[M.Friday=5]="Friday",M[M.Saturday=6]="Saturday";var J=2;function ee(e){var t,n=A[e];return n&&(t=n[2]),"number"==typeof t?t:J}var te,ne,re,ie,oe,ae,se=/^(\d{4})-?(\d\d)-?(\d\d)(?:T(\d\d)(?::?(\d\d)(?::?(\d\d)(?:\.(\d+))?)?)?(Z|([+-])(\d\d):?(\d\d))?)?$/,ue={},ce=/((?:[^GyMLwWdEabBhHmsSzZO']+)|(?:'(?:[^']|'')*')|(?:G{1,5}|y{1,4}|M{1,5}|L{1,5}|w{1,2}|W{1}|d{1,2}|E{1,6}|a{1,5}|b{1,5}|B{1,5}|h{1,2}|H{1,2}|m{1,2}|s{1,2}|S{1,3}|z{1,4}|Z{1,5}|O{1,4}))([\s\S]*)/;function le(e,t,n,r){var o=function(e){if(xe(e))return e;if("number"==typeof e&&!isNaN(e))return new Date(e);if("string"==typeof e){e=e.trim();var t=parseFloat(e);if(!isNaN(e-t))return new Date(t);if(/^(\d{4}-\d{1,2}-\d{1,2})$/.test(e)){var n=Object(i.__read)(e.split("-").map(function(e){return+e}),3),r=n[0],o=n[1],a=n[2];return new Date(r,o-1,a)}var s=void 0;if(s=e.match(se))return je(s)}var u=new Date(e);if(!xe(u))throw new Error('Unable to convert "'+e+'" into a date');return u}(e);t=function e(t,n){var r=T(t);ue[r]=ue[r]||{};if(ue[r][n])return ue[r][n];var i="";switch(n){case"shortDate":i=L(t,w.Short);break;case"mediumDate":i=L(t,w.Medium);break;case"longDate":i=L(t,w.Long);break;case"fullDate":i=L(t,w.Full);break;case"shortTime":i=V(t,w.Short);break;case"mediumTime":i=V(t,w.Medium);break;case"longTime":i=V(t,w.Long);break;case"fullTime":i=V(t,w.Full);break;case"short":var o=e(t,"shortTime"),a=e(t,"shortDate");i=de(B(t,w.Short),[o,a]);break;case"medium":var s=e(t,"mediumTime"),u=e(t,"mediumDate");i=de(B(t,w.Medium),[s,u]);break;case"long":var c=e(t,"longTime"),l=e(t,"longDate");i=de(B(t,w.Long),[c,l]);break;case"full":var d=e(t,"fullTime"),p=e(t,"fullDate");i=de(B(t,w.Full),[d,p])}i&&(ue[r][n]=i);return i}(n,t)||t;for(var a,s=[];t;){if(!(a=ce.exec(t))){s.push(t);break}var u=(s=s.concat(a.slice(1))).pop();if(!u)break;t=u}var c,l,d,p,f,h,m,g=o.getTimezoneOffset();r&&(g=be(r,g),c=o,l=r,d=!0?-1:1,p=c.getTimezoneOffset(),f=be(l,p),h=c,m=d*(f-p),(h=new Date(h.getTime())).setMinutes(h.getMinutes()+m),o=h);var v="";return s.forEach(function(e){var t=function(e){if(ye[e])return ye[e];var t;switch(e){case"G":case"GG":case"GGG":t=he(oe.Eras,x.Abbreviated);break;case"GGGG":t=he(oe.Eras,x.Wide);break;case"GGGGG":t=he(oe.Eras,x.Narrow);break;case"y":t=fe(re.FullYear,1,0,!1,!0);break;case"yy":t=fe(re.FullYear,2,0,!0,!0);break;case"yyy":t=fe(re.FullYear,3,0,!1,!0);break;case"yyyy":t=fe(re.FullYear,4,0,!1,!0);break;case"M":case"L":t=fe(re.Month,1,1);break;case"MM":case"LL":t=fe(re.Month,2,1);break;case"MMM":t=he(oe.Months,x.Abbreviated);break;case"MMMM":t=he(oe.Months,x.Wide);break;case"MMMMM":t=he(oe.Months,x.Narrow);break;case"LLL":t=he(oe.Months,x.Abbreviated,b.Standalone);break;case"LLLL":t=he(oe.Months,x.Wide,b.Standalone);break;case"LLLLL":t=he(oe.Months,x.Narrow,b.Standalone);break;case"w":t=_e(1);break;case"ww":t=_e(2);break;case"W":t=_e(1,!0);break;case"d":t=fe(re.Date,1);break;case"dd":t=fe(re.Date,2);break;case"E":case"EE":case"EEE":t=he(oe.Days,x.Abbreviated);break;case"EEEE":t=he(oe.Days,x.Wide);break;case"EEEEE":t=he(oe.Days,x.Narrow);break;case"EEEEEE":t=he(oe.Days,x.Short);break;case"a":case"aa":case"aaa":t=he(oe.DayPeriods,x.Abbreviated);break;case"aaaa":t=he(oe.DayPeriods,x.Wide);break;case"aaaaa":t=he(oe.DayPeriods,x.Narrow);break;case"b":case"bb":case"bbb":t=he(oe.DayPeriods,x.Abbreviated,b.Standalone,!0);break;case"bbbb":t=he(oe.DayPeriods,x.Wide,b.Standalone,!0);break;case"bbbbb":t=he(oe.DayPeriods,x.Narrow,b.Standalone,!0);break;case"B":case"BB":case"BBB":t=he(oe.DayPeriods,x.Abbreviated,b.Format,!0);break;case"BBBB":t=he(oe.DayPeriods,x.Wide,b.Format,!0);break;case"BBBBB":t=he(oe.DayPeriods,x.Narrow,b.Format,!0);break;case"h":t=fe(re.Hours,1,-12);break;case"hh":t=fe(re.Hours,2,-12);break;case"H":t=fe(re.Hours,1);break;case"HH":t=fe(re.Hours,2);break;case"m":t=fe(re.Minutes,1);break;case"mm":t=fe(re.Minutes,2);break;case"s":t=fe(re.Seconds,1);break;case"ss":t=fe(re.Seconds,2);break;case"S":t=fe(re.FractionalSeconds,1);break;case"SS":t=fe(re.FractionalSeconds,2);break;case"SSS":t=fe(re.FractionalSeconds,3);break;case"Z":case"ZZ":case"ZZZ":t=me(te.Short);break;case"ZZZZZ":t=me(te.Extended);break;case"O":case"OO":case"OOO":case"z":case"zz":case"zzz":t=me(te.ShortGMT);break;case"OOOO":case"ZZZZ":case"zzzz":t=me(te.Long);break;default:return null}return ye[e]=t,t}(e);v+=t?t(o,n,g):"''"===e?"'":e.replace(/(^'|'$)/g,"").replace(/''/g,"'")}),v}function de(e,t){return t&&(e=e.replace(/\{([^}]+)}/g,function(e,n){return null!=t&&n in t?t[n]:e})),e}function pe(e,t,n,r,i){void 0===n&&(n="-");var o="";(e<0||i&&e<=0)&&(i?e=1-e:(e=-e,o=n));for(var a=String(e);a.length<t;)a="0"+a;return r&&(a=a.substr(a.length-t)),o+a}function fe(e,t,n,r,i){return void 0===n&&(n=0),void 0===r&&(r=!1),void 0===i&&(i=!1),function(o,a){var s,u=function(e,t){switch(e){case re.FullYear:return t.getFullYear();case re.Month:return t.getMonth();case re.Date:return t.getDate();case re.Hours:return t.getHours();case re.Minutes:return t.getMinutes();case re.Seconds:return t.getSeconds();case re.FractionalSeconds:return t.getMilliseconds();case re.Day:return t.getDay();default:throw new Error('Unknown DateType value "'+e+'".')}}(e,o);if((n>0||u>-n)&&(u+=n),e===re.Hours)0===u&&-12===n&&(u=12);else if(e===re.FractionalSeconds)return s=t,pe(u,3).substr(0,s);var c=z(a,E.MinusSign);return pe(u,t,c,r,i)}}function he(e,t,n,r){return void 0===n&&(n=b.Format),void 0===r&&(r=!1),function(i,o){return function(e,t,n,r,i,o){switch(n){case oe.Months:return P(t,i,r)[e.getMonth()];case oe.Days:return I(t,i,r)[e.getDay()];case oe.DayPeriods:var a=e.getHours(),s=e.getMinutes();if(o){var u,c=Y(t),l=K(t,i,r);if(c.forEach(function(e,t){if(Array.isArray(e)){var n=e[0],r=n.hours,i=n.minutes,o=e[1],c=o.hours,d=o.minutes;a>=r&&s>=i&&(a<c||a===c&&s<d)&&(u=l[t])}else{var p=e.hours,f=e.minutes;p===a&&f===s&&(u=l[t])}}),u)return u}return D(t,i,r)[a<12?0:1];case oe.Eras:return R(t,r)[e.getFullYear()<=0?0:1];default:var d=n;throw new Error("unexpected translation type "+d)}}(i,o,e,t,n,r)}}function me(e){return function(t,n,r){var i=-1*r,o=z(n,E.MinusSign),a=i>0?Math.floor(i/60):Math.ceil(i/60);switch(e){case te.Short:return(i>=0?"+":"")+pe(a,2,o)+pe(Math.abs(i%60),2,o);case te.ShortGMT:return"GMT"+(i>=0?"+":"")+pe(a,1,o);case te.Long:return"GMT"+(i>=0?"+":"")+pe(a,2,o)+":"+pe(Math.abs(i%60),2,o);case te.Extended:return 0===r?"Z":(i>=0?"+":"")+pe(a,2,o)+":"+pe(Math.abs(i%60),2,o);default:throw new Error('Unknown zone width "'+e+'"')}}}(ne=te||(te={}))[ne.Short=0]="Short",ne[ne.ShortGMT=1]="ShortGMT",ne[ne.Long=2]="Long",ne[ne.Extended=3]="Extended",(ie=re||(re={}))[ie.FullYear=0]="FullYear",ie[ie.Month=1]="Month",ie[ie.Date=2]="Date",ie[ie.Hours=3]="Hours",ie[ie.Minutes=4]="Minutes",ie[ie.Seconds=5]="Seconds",ie[ie.FractionalSeconds=6]="FractionalSeconds",ie[ie.Day=7]="Day",(ae=oe||(oe={}))[ae.DayPeriods=0]="DayPeriods",ae[ae.Days=1]="Days",ae[ae.Months=2]="Months",ae[ae.Eras=3]="Eras";var ge=0,ve=4;function _e(e,t){return void 0===t&&(t=!1),function(n,r){var i,o,a,s;if(t){var u=new Date(n.getFullYear(),n.getMonth(),1).getDay()-1,c=n.getDate();i=1+Math.floor((c+u)/7)}else{var l=(a=n.getFullYear(),s=new Date(a,ge,1).getDay(),new Date(a,0,1+(s<=ve?ve:ve+7)-s)),d=(o=n,new Date(o.getFullYear(),o.getMonth(),o.getDate()+(ve-o.getDay()))).getTime()-l.getTime();i=1+Math.round(d/6048e5)}return pe(i,e,z(r,E.MinusSign))}}var ye={};function be(e,t){e=e.replace(/:/g,"");var n=Date.parse("Jan 01, 1970 00:00:00 "+e)/6e4;return isNaN(n)?t:n}function je(e){var t=new Date(0),n=0,r=0,i=e[8]?t.setUTCFullYear:t.setFullYear,o=e[8]?t.setUTCHours:t.setHours;e[9]&&(n=Number(e[9]+e[10]),r=Number(e[9]+e[11])),i.call(t,Number(e[1]),Number(e[2])-1,Number(e[3]));var a=Number(e[4]||0)-n,s=Number(e[5]||0)-r,u=Number(e[6]||0),c=Math.round(1e3*parseFloat("0."+(e[7]||0)));return o.call(t,a,s,u,c),t}function xe(e){return e instanceof Date&&!isNaN(e.valueOf())}var Oe=/^(\d+)?\.((\d+)(-(\d+))?)?$/,we=22,Ce=".",Ee="0",Se=";",ke=",",Me="#",Ae="¤",Te="%";function De(e,t,n,r,i,o,a){void 0===a&&(a=!1);var s="",u=!1;if(isFinite(e)){var c=function(e){var t,n,r,i,o,a=Math.abs(e)+"",s=0;(n=a.indexOf(Ce))>-1&&(a=a.replace(Ce,""));(r=a.search(/e/i))>0?(n<0&&(n=r),n+=+a.slice(r+1),a=a.substring(0,r)):n<0&&(n=a.length);for(r=0;a.charAt(r)===Ee;r++);if(r===(o=a.length))t=[0],n=1;else{for(o--;a.charAt(o)===Ee;)o--;for(n-=r,t=[],i=0;r<=o;r++,i++)t[i]=Number(a.charAt(r))}n>we&&(t=t.splice(0,we-1),s=n-1,n=1);return{digits:t,exponent:s,integerLen:n}}(e);a&&(c=function(e){if(0===e.digits[0])return e;var t=e.digits.length-e.integerLen;e.exponent?e.exponent+=2:(0===t?e.digits.push(0,0):1===t&&e.digits.push(0),e.integerLen+=2);return e}(c));var l=t.minInt,d=t.minFrac,p=t.maxFrac;if(o){var f=o.match(Oe);if(null===f)throw new Error(o+" is not a valid digit info");var h=f[1],m=f[3],g=f[5];null!=h&&(l=Fe(h)),null!=m&&(d=Fe(m)),null!=g?p=Fe(g):null!=m&&d>p&&(p=d)}!function(e,t,n){if(t>n)throw new Error("The minimum number of digits after fraction ("+t+") is higher than the maximum ("+n+").");var r=e.digits,i=r.length-e.integerLen,o=Math.min(Math.max(t,i),n),a=o+e.integerLen,s=r[a];if(a>0){r.splice(Math.max(e.integerLen,a));for(var u=a;u<r.length;u++)r[u]=0}else{i=Math.max(0,i),e.integerLen=1,r.length=Math.max(1,a=o+1),r[0]=0;for(var c=1;c<a;c++)r[c]=0}if(s>=5)if(a-1<0){for(var l=0;l>a;l--)r.unshift(0),e.integerLen++;r.unshift(1),e.integerLen++}else r[a-1]++;for(;i<Math.max(0,o);i++)r.push(0);var d=0!==o,p=t+e.integerLen,f=r.reduceRight(function(e,t,n,r){return t+=e,r[n]=t<10?t:t-10,d&&(0===r[n]&&n>=p?r.pop():d=!1),t>=10?1:0},0);f&&(r.unshift(f),e.integerLen++)}(c,d,p);var v=c.digits,_=c.integerLen,y=c.exponent,b=[];for(u=v.every(function(e){return!e});_<l;_++)v.unshift(0);for(;_<0;_++)v.unshift(0);_>0?b=v.splice(_,v.length):(b=v,v=[0]);var j=[];for(v.length>=t.lgSize&&j.unshift(v.splice(-t.lgSize,v.length).join(""));v.length>t.gSize;)j.unshift(v.splice(-t.gSize,v.length).join(""));v.length&&j.unshift(v.join("")),s=j.join(z(n,r)),b.length&&(s+=z(n,i)+b.join("")),y&&(s+=z(n,E.Exponential)+"+"+y)}else s=z(n,E.Infinity);return s=e<0&&!u?t.negPre+s+t.negSuf:t.posPre+s+t.posSuf}function Ie(e,t,n,r,i){var o=Ne(U(t,g.Currency),z(t,E.MinusSign));return o.minFrac=ee(r),o.maxFrac=o.minFrac,De(e,o,t,E.CurrencyGroup,E.CurrencyDecimal,i).replace(Ae,n).replace(Ae,"")}function Pe(e,t,n){return De(e,Ne(U(t,g.Percent),z(t,E.MinusSign)),t,E.Group,E.Decimal,n,!0).replace(new RegExp(Te,"g"),z(t,E.PercentSign))}function Re(e,t,n){return De(e,Ne(U(t,g.Decimal),z(t,E.MinusSign)),t,E.Group,E.Decimal,n)}function Ne(e,t){void 0===t&&(t="-");var n={minInt:1,minFrac:0,maxFrac:0,posPre:"",posSuf:"",negPre:"",negSuf:"",gSize:0,lgSize:0},r=e.split(Se),i=r[0],o=r[1],a=-1!==i.indexOf(Ce)?i.split(Ce):[i.substring(0,i.lastIndexOf(Ee)+1),i.substring(i.lastIndexOf(Ee)+1)],s=a[0],u=a[1]||"";n.posPre=s.substr(0,s.indexOf(Me));for(var c=0;c<u.length;c++){var l=u.charAt(c);l===Ee?n.minFrac=n.maxFrac=c+1:l===Me?n.maxFrac=c+1:n.posSuf+=l}var d=s.split(ke);if(n.gSize=d[1]?d[1].length:0,n.lgSize=d[2]||d[1]?(d[2]||d[1]).length:0,o){var p=i.length-n.posPre.length-n.posSuf.length,f=o.indexOf(Me);n.negPre=o.substr(0,f).replace(/'/g,""),n.negSuf=o.substr(f+p).replace(/'/g,"")}else n.negPre=t+n.posPre,n.negSuf=n.posSuf;return n}function Fe(e){var t=parseInt(e);if(isNaN(t))throw new Error("Invalid integer literal when parsing "+e);return t}var Le=new r.InjectionToken("UseV4Plurals"),Ve=function(){return function(){}}();function Be(e,t,n,r){var i="="+e;if(t.indexOf(i)>-1)return i;if(i=n.getPluralCategory(e,r),t.indexOf(i)>-1)return i;if(t.indexOf("other")>-1)return"other";throw new Error('No plural message found for value "'+e+'"')}var ze=function(e){function t(t,n){var r=e.call(this)||this;return r.locale=t,r.deprecatedPluralFn=n,r}return Object(i.__extends)(t,e),t.prototype.getPluralCategory=function(e,t){switch(this.deprecatedPluralFn?this.deprecatedPluralFn(t||this.locale,e):G(t||this.locale)(e)){case _.Zero:return"zero";case _.One:return"one";case _.Two:return"two";case _.Few:return"few";case _.Many:return"many";default:return"other"}},t=Object(i.__decorate)([Object(r.Injectable)(),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__param)(1,Object(r.Optional)()),Object(i.__param)(1,Object(r.Inject)(Le)),Object(i.__metadata)("design:paramtypes",[String,Object])],t)}(Ve);function Ue(e,t){"string"==typeof t&&(t=parseInt(t,10));var n=t,r=n.toString().replace(/^[^.]*\.?/,""),i=Math.floor(Math.abs(n)),o=r.length,a=parseInt(r,10),s=parseInt(n.toString().replace(/^[^.]*\.?|0+$/g,""),10)||0;switch(e.split("-")[0].toLowerCase()){case"af":case"asa":case"az":case"bem":case"bez":case"bg":case"brx":case"ce":case"cgg":case"chr":case"ckb":case"ee":case"el":case"eo":case"es":case"eu":case"fo":case"fur":case"gsw":case"ha":case"haw":case"hu":case"jgo":case"jmc":case"ka":case"kk":case"kkj":case"kl":case"ks":case"ksb":case"ky":case"lb":case"lg":case"mas":case"mgo":case"ml":case"mn":case"nb":case"nd":case"ne":case"nn":case"nnh":case"nyn":case"om":case"or":case"os":case"ps":case"rm":case"rof":case"rwk":case"saq":case"seh":case"sn":case"so":case"sq":case"ta":case"te":case"teo":case"tk":case"tr":case"ug":case"uz":case"vo":case"vun":case"wae":case"xog":return 1===n?_.One:_.Other;case"ak":case"ln":case"mg":case"pa":case"ti":return n===Math.floor(n)&&n>=0&&n<=1?_.One:_.Other;case"am":case"as":case"bn":case"fa":case"gu":case"hi":case"kn":case"mr":case"zu":return 0===i||1===n?_.One:_.Other;case"ar":return 0===n?_.Zero:1===n?_.One:2===n?_.Two:n%100===Math.floor(n%100)&&n%100>=3&&n%100<=10?_.Few:n%100===Math.floor(n%100)&&n%100>=11&&n%100<=99?_.Many:_.Other;case"ast":case"ca":case"de":case"en":case"et":case"fi":case"fy":case"gl":case"it":case"nl":case"sv":case"sw":case"ur":case"yi":return 1===i&&0===o?_.One:_.Other;case"be":return n%10==1&&n%100!=11?_.One:n%10===Math.floor(n%10)&&n%10>=2&&n%10<=4&&!(n%100>=12&&n%100<=14)?_.Few:n%10==0||n%10===Math.floor(n%10)&&n%10>=5&&n%10<=9||n%100===Math.floor(n%100)&&n%100>=11&&n%100<=14?_.Many:_.Other;case"br":return n%10==1&&n%100!=11&&n%100!=71&&n%100!=91?_.One:n%10==2&&n%100!=12&&n%100!=72&&n%100!=92?_.Two:n%10===Math.floor(n%10)&&(n%10>=3&&n%10<=4||n%10==9)&&!(n%100>=10&&n%100<=19||n%100>=70&&n%100<=79||n%100>=90&&n%100<=99)?_.Few:0!==n&&n%1e6==0?_.Many:_.Other;case"bs":case"hr":case"sr":return 0===o&&i%10==1&&i%100!=11||a%10==1&&a%100!=11?_.One:0===o&&i%10===Math.floor(i%10)&&i%10>=2&&i%10<=4&&!(i%100>=12&&i%100<=14)||a%10===Math.floor(a%10)&&a%10>=2&&a%10<=4&&!(a%100>=12&&a%100<=14)?_.Few:_.Other;case"cs":case"sk":return 1===i&&0===o?_.One:i===Math.floor(i)&&i>=2&&i<=4&&0===o?_.Few:0!==o?_.Many:_.Other;case"cy":return 0===n?_.Zero:1===n?_.One:2===n?_.Two:3===n?_.Few:6===n?_.Many:_.Other;case"da":return 1===n||0!==s&&(0===i||1===i)?_.One:_.Other;case"dsb":case"hsb":return 0===o&&i%100==1||a%100==1?_.One:0===o&&i%100==2||a%100==2?_.Two:0===o&&i%100===Math.floor(i%100)&&i%100>=3&&i%100<=4||a%100===Math.floor(a%100)&&a%100>=3&&a%100<=4?_.Few:_.Other;case"ff":case"fr":case"hy":case"kab":return 0===i||1===i?_.One:_.Other;case"fil":return 0===o&&(1===i||2===i||3===i)||0===o&&i%10!=4&&i%10!=6&&i%10!=9||0!==o&&a%10!=4&&a%10!=6&&a%10!=9?_.One:_.Other;case"ga":return 1===n?_.One:2===n?_.Two:n===Math.floor(n)&&n>=3&&n<=6?_.Few:n===Math.floor(n)&&n>=7&&n<=10?_.Many:_.Other;case"gd":return 1===n||11===n?_.One:2===n||12===n?_.Two:n===Math.floor(n)&&(n>=3&&n<=10||n>=13&&n<=19)?_.Few:_.Other;case"gv":return 0===o&&i%10==1?_.One:0===o&&i%10==2?_.Two:0!==o||i%100!=0&&i%100!=20&&i%100!=40&&i%100!=60&&i%100!=80?0!==o?_.Many:_.Other:_.Few;case"he":return 1===i&&0===o?_.One:2===i&&0===o?_.Two:0!==o||n>=0&&n<=10||n%10!=0?_.Other:_.Many;case"is":return 0===s&&i%10==1&&i%100!=11||0!==s?_.One:_.Other;case"ksh":return 0===n?_.Zero:1===n?_.One:_.Other;case"kw":case"naq":case"se":case"smn":return 1===n?_.One:2===n?_.Two:_.Other;case"lag":return 0===n?_.Zero:0!==i&&1!==i||0===n?_.Other:_.One;case"lt":return n%10!=1||n%100>=11&&n%100<=19?n%10===Math.floor(n%10)&&n%10>=2&&n%10<=9&&!(n%100>=11&&n%100<=19)?_.Few:0!==a?_.Many:_.Other:_.One;case"lv":case"prg":return n%10==0||n%100===Math.floor(n%100)&&n%100>=11&&n%100<=19||2===o&&a%100===Math.floor(a%100)&&a%100>=11&&a%100<=19?_.Zero:n%10==1&&n%100!=11||2===o&&a%10==1&&a%100!=11||2!==o&&a%10==1?_.One:_.Other;case"mk":return 0===o&&i%10==1||a%10==1?_.One:_.Other;case"mt":return 1===n?_.One:0===n||n%100===Math.floor(n%100)&&n%100>=2&&n%100<=10?_.Few:n%100===Math.floor(n%100)&&n%100>=11&&n%100<=19?_.Many:_.Other;case"pl":return 1===i&&0===o?_.One:0===o&&i%10===Math.floor(i%10)&&i%10>=2&&i%10<=4&&!(i%100>=12&&i%100<=14)?_.Few:0===o&&1!==i&&i%10===Math.floor(i%10)&&i%10>=0&&i%10<=1||0===o&&i%10===Math.floor(i%10)&&i%10>=5&&i%10<=9||0===o&&i%100===Math.floor(i%100)&&i%100>=12&&i%100<=14?_.Many:_.Other;case"pt":return n===Math.floor(n)&&n>=0&&n<=2&&2!==n?_.One:_.Other;case"ro":return 1===i&&0===o?_.One:0!==o||0===n||1!==n&&n%100===Math.floor(n%100)&&n%100>=1&&n%100<=19?_.Few:_.Other;case"ru":case"uk":return 0===o&&i%10==1&&i%100!=11?_.One:0===o&&i%10===Math.floor(i%10)&&i%10>=2&&i%10<=4&&!(i%100>=12&&i%100<=14)?_.Few:0===o&&i%10==0||0===o&&i%10===Math.floor(i%10)&&i%10>=5&&i%10<=9||0===o&&i%100===Math.floor(i%100)&&i%100>=11&&i%100<=14?_.Many:_.Other;case"shi":return 0===i||1===n?_.One:n===Math.floor(n)&&n>=2&&n<=10?_.Few:_.Other;case"si":return 0===n||1===n||0===i&&1===a?_.One:_.Other;case"sl":return 0===o&&i%100==1?_.One:0===o&&i%100==2?_.Two:0===o&&i%100===Math.floor(i%100)&&i%100>=3&&i%100<=4||0!==o?_.Few:_.Other;case"tzm":return n===Math.floor(n)&&n>=0&&n<=1||n===Math.floor(n)&&n>=11&&n<=99?_.One:_.Other;default:return _.Other}}function He(e,t){var n,r;t=encodeURIComponent(t);try{for(var o=Object(i.__values)(e.split(";")),a=o.next();!a.done;a=o.next()){var s=a.value,u=s.indexOf("="),c=Object(i.__read)(-1==u?[s,""]:[s.slice(0,u),s.slice(u+1)],2),l=c[0],d=c[1];if(l.trim()===t)return decodeURIComponent(d)}}catch(e){n={error:e}}finally{try{a&&!a.done&&(r=o.return)&&r.call(o)}finally{if(n)throw n.error}}return null}var qe=function(){function e(e,t,n,r){this._iterableDiffers=e,this._keyValueDiffers=t,this._ngEl=n,this._renderer=r,this._initialClasses=[]}return Object.defineProperty(e.prototype,"klass",{set:function(e){this._removeClasses(this._initialClasses),this._initialClasses="string"==typeof e?e.split(/\s+/):[],this._applyClasses(this._initialClasses),this._applyClasses(this._rawClass)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClass",{set:function(e){this._removeClasses(this._rawClass),this._applyClasses(this._initialClasses),this._iterableDiffer=null,this._keyValueDiffer=null,this._rawClass="string"==typeof e?e.split(/\s+/):e,this._rawClass&&(Object(r["ɵisListLikeIterable"])(this._rawClass)?this._iterableDiffer=this._iterableDiffers.find(this._rawClass).create():this._keyValueDiffer=this._keyValueDiffers.find(this._rawClass).create())},enumerable:!0,configurable:!0}),e.prototype.ngDoCheck=function(){if(this._iterableDiffer){var e=this._iterableDiffer.diff(this._rawClass);e&&this._applyIterableChanges(e)}else if(this._keyValueDiffer){var t=this._keyValueDiffer.diff(this._rawClass);t&&this._applyKeyValueChanges(t)}},e.prototype._applyKeyValueChanges=function(e){var t=this;e.forEachAddedItem(function(e){return t._toggleClass(e.key,e.currentValue)}),e.forEachChangedItem(function(e){return t._toggleClass(e.key,e.currentValue)}),e.forEachRemovedItem(function(e){e.previousValue&&t._toggleClass(e.key,!1)})},e.prototype._applyIterableChanges=function(e){var t=this;e.forEachAddedItem(function(e){if("string"!=typeof e.item)throw new Error("NgClass can only toggle CSS classes expressed as strings, got "+Object(r["ɵstringify"])(e.item));t._toggleClass(e.item,!0)}),e.forEachRemovedItem(function(e){return t._toggleClass(e.item,!1)})},e.prototype._applyClasses=function(e){var t=this;e&&(Array.isArray(e)||e instanceof Set?e.forEach(function(e){return t._toggleClass(e,!0)}):Object.keys(e).forEach(function(n){return t._toggleClass(n,!!e[n])}))},e.prototype._removeClasses=function(e){var t=this;e&&(Array.isArray(e)||e instanceof Set?e.forEach(function(e){return t._toggleClass(e,!1)}):Object.keys(e).forEach(function(e){return t._toggleClass(e,!1)}))},e.prototype._toggleClass=function(e,t){var n=this;(e=e.trim())&&e.split(/\s+/g).forEach(function(e){t?n._renderer.addClass(n._ngEl.nativeElement,e):n._renderer.removeClass(n._ngEl.nativeElement,e)})},Object(i.__decorate)([Object(r.Input)("class"),Object(i.__metadata)("design:type",String),Object(i.__metadata)("design:paramtypes",[String])],e.prototype,"klass",null),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object),Object(i.__metadata)("design:paramtypes",[Object])],e.prototype,"ngClass",null),e=Object(i.__decorate)([Object(r.Directive)({selector:"[ngClass]"}),Object(i.__metadata)("design:paramtypes",[r.IterableDiffers,r.KeyValueDiffers,r.ElementRef,r.Renderer2])],e)}(),Ge=function(){function e(e){this._viewContainerRef=e,this._componentRef=null,this._moduleRef=null}return e.prototype.ngOnChanges=function(e){if(this._viewContainerRef.clear(),this._componentRef=null,this.ngComponentOutlet){var t=this.ngComponentOutletInjector||this._viewContainerRef.parentInjector;if(e.ngComponentOutletNgModuleFactory)if(this._moduleRef&&this._moduleRef.destroy(),this.ngComponentOutletNgModuleFactory){var n=t.get(r.NgModuleRef);this._moduleRef=this.ngComponentOutletNgModuleFactory.create(n.injector)}else this._moduleRef=null;var i=(this._moduleRef?this._moduleRef.componentFactoryResolver:t.get(r.ComponentFactoryResolver)).resolveComponentFactory(this.ngComponentOutlet);this._componentRef=this._viewContainerRef.createComponent(i,this._viewContainerRef.length,t,this.ngComponentOutletContent)}},e.prototype.ngOnDestroy=function(){this._moduleRef&&this._moduleRef.destroy()},Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",r.Type)],e.prototype,"ngComponentOutlet",void 0),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",r.Injector)],e.prototype,"ngComponentOutletInjector",void 0),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Array)],e.prototype,"ngComponentOutletContent",void 0),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",r.NgModuleFactory)],e.prototype,"ngComponentOutletNgModuleFactory",void 0),e=Object(i.__decorate)([Object(r.Directive)({selector:"[ngComponentOutlet]"}),Object(i.__metadata)("design:paramtypes",[r.ViewContainerRef])],e)}(),We=function(){function e(e,t,n,r){this.$implicit=e,this.ngForOf=t,this.index=n,this.count=r}return Object.defineProperty(e.prototype,"first",{get:function(){return 0===this.index},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"last",{get:function(){return this.index===this.count-1},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"even",{get:function(){return this.index%2==0},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"odd",{get:function(){return!this.even},enumerable:!0,configurable:!0}),e}(),Ye=function(){function e(e,t,n){this._viewContainer=e,this._template=t,this._differs=n,this._ngForOfDirty=!0,this._differ=null}return Object.defineProperty(e.prototype,"ngForOf",{set:function(e){this._ngForOf=e,this._ngForOfDirty=!0},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngForTrackBy",{get:function(){return this._trackByFn},set:function(e){Object(r.isDevMode)()&&null!=e&&"function"!=typeof e&&console&&console.warn&&console.warn("trackBy must be a function, but received "+JSON.stringify(e)+". See https://angular.io/docs/ts/latest/api/common/index/NgFor-directive.html#!#change-propagation for more information."),this._trackByFn=e},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngForTemplate",{set:function(e){e&&(this._template=e)},enumerable:!0,configurable:!0}),e.prototype.ngDoCheck=function(){if(this._ngForOfDirty){this._ngForOfDirty=!1;var e=this._ngForOf;if(!this._differ&&e)try{this._differ=this._differs.find(e).create(this.ngForTrackBy)}catch(n){throw new Error("Cannot find a differ supporting object '"+e+"' of type '"+((t=e).name||typeof t)+"'. NgFor only supports binding to Iterables such as Arrays.")}}var t;if(this._differ){var n=this._differ.diff(this._ngForOf);n&&this._applyChanges(n)}},e.prototype._applyChanges=function(e){var t=this,n=[];e.forEachOperation(function(e,r,i){if(null==e.previousIndex){var o=t._viewContainer.createEmbeddedView(t._template,new We(null,t._ngForOf,-1,-1),i),a=new Ke(e,o);n.push(a)}else if(null==i)t._viewContainer.remove(r);else{o=t._viewContainer.get(r);t._viewContainer.move(o,i);a=new Ke(e,o);n.push(a)}});for(var r=0;r<n.length;r++)this._perViewChange(n[r].view,n[r].record);r=0;for(var i=this._viewContainer.length;r<i;r++){var o=this._viewContainer.get(r);o.context.index=r,o.context.count=i,o.context.ngForOf=this._ngForOf}e.forEachIdentityChange(function(e){t._viewContainer.get(e.currentIndex).context.$implicit=e.item})},e.prototype._perViewChange=function(e,t){e.context.$implicit=t.item},e.ngTemplateContextGuard=function(e,t){return!0},Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object),Object(i.__metadata)("design:paramtypes",[Object])],e.prototype,"ngForOf",null),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Function),Object(i.__metadata)("design:paramtypes",[Function])],e.prototype,"ngForTrackBy",null),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",r.TemplateRef),Object(i.__metadata)("design:paramtypes",[r.TemplateRef])],e.prototype,"ngForTemplate",null),e=Object(i.__decorate)([Object(r.Directive)({selector:"[ngFor][ngForOf]"}),Object(i.__metadata)("design:paramtypes",[r.ViewContainerRef,r.TemplateRef,r.IterableDiffers])],e)}(),Ke=function(){return function(e,t){this.record=e,this.view=t}}();var Qe=function(){function e(e,t){this._viewContainer=e,this._context=new Xe,this._thenTemplateRef=null,this._elseTemplateRef=null,this._thenViewRef=null,this._elseViewRef=null,this._thenTemplateRef=t}return Object.defineProperty(e.prototype,"ngIf",{set:function(e){this._context.$implicit=this._context.ngIf=e,this._updateView()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngIfThen",{set:function(e){Ze("ngIfThen",e),this._thenTemplateRef=e,this._thenViewRef=null,this._updateView()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngIfElse",{set:function(e){Ze("ngIfElse",e),this._elseTemplateRef=e,this._elseViewRef=null,this._updateView()},enumerable:!0,configurable:!0}),e.prototype._updateView=function(){this._context.$implicit?this._thenViewRef||(this._viewContainer.clear(),this._elseViewRef=null,this._thenTemplateRef&&(this._thenViewRef=this._viewContainer.createEmbeddedView(this._thenTemplateRef,this._context))):this._elseViewRef||(this._viewContainer.clear(),this._thenViewRef=null,this._elseTemplateRef&&(this._elseViewRef=this._viewContainer.createEmbeddedView(this._elseTemplateRef,this._context)))},e.ngTemplateGuard_ngIf=function(e,t){return!0},Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object),Object(i.__metadata)("design:paramtypes",[Object])],e.prototype,"ngIf",null),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object),Object(i.__metadata)("design:paramtypes",[Object])],e.prototype,"ngIfThen",null),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object),Object(i.__metadata)("design:paramtypes",[Object])],e.prototype,"ngIfElse",null),e=Object(i.__decorate)([Object(r.Directive)({selector:"[ngIf]"}),Object(i.__metadata)("design:paramtypes",[r.ViewContainerRef,r.TemplateRef])],e)}(),Xe=function(){return function(){this.$implicit=null,this.ngIf=null}}();function Ze(e,t){if(!!(t&&!t.createEmbeddedView))throw new Error(e+" must be a TemplateRef, but received '"+Object(r["ɵstringify"])(t)+"'.")}var $e=function(){function e(e,t){this._viewContainerRef=e,this._templateRef=t,this._created=!1}return e.prototype.create=function(){this._created=!0,this._viewContainerRef.createEmbeddedView(this._templateRef)},e.prototype.destroy=function(){this._created=!1,this._viewContainerRef.clear()},e.prototype.enforceState=function(e){e&&!this._created?this.create():!e&&this._created&&this.destroy()},e}(),Je=function(){function e(){this._defaultUsed=!1,this._caseCount=0,this._lastCaseCheckIndex=0,this._lastCasesMatched=!1}return Object.defineProperty(e.prototype,"ngSwitch",{set:function(e){this._ngSwitch=e,0===this._caseCount&&this._updateDefaultCases(!0)},enumerable:!0,configurable:!0}),e.prototype._addCase=function(){return this._caseCount++},e.prototype._addDefault=function(e){this._defaultViews||(this._defaultViews=[]),this._defaultViews.push(e)},e.prototype._matchCase=function(e){var t=e==this._ngSwitch;return this._lastCasesMatched=this._lastCasesMatched||t,this._lastCaseCheckIndex++,this._lastCaseCheckIndex===this._caseCount&&(this._updateDefaultCases(!this._lastCasesMatched),this._lastCaseCheckIndex=0,this._lastCasesMatched=!1),t},e.prototype._updateDefaultCases=function(e){if(this._defaultViews&&e!==this._defaultUsed){this._defaultUsed=e;for(var t=0;t<this._defaultViews.length;t++){this._defaultViews[t].enforceState(e)}}},Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object),Object(i.__metadata)("design:paramtypes",[Object])],e.prototype,"ngSwitch",null),e=Object(i.__decorate)([Object(r.Directive)({selector:"[ngSwitch]"})],e)}(),et=function(){function e(e,t,n){this.ngSwitch=n,n._addCase(),this._view=new $e(e,t)}return e.prototype.ngDoCheck=function(){this._view.enforceState(this.ngSwitch._matchCase(this.ngSwitchCase))},Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object)],e.prototype,"ngSwitchCase",void 0),e=Object(i.__decorate)([Object(r.Directive)({selector:"[ngSwitchCase]"}),Object(i.__param)(2,Object(r.Host)()),Object(i.__metadata)("design:paramtypes",[r.ViewContainerRef,r.TemplateRef,Je])],e)}(),tt=function(){function e(e,t,n){n._addDefault(new $e(e,t))}return e=Object(i.__decorate)([Object(r.Directive)({selector:"[ngSwitchDefault]"}),Object(i.__param)(2,Object(r.Host)()),Object(i.__metadata)("design:paramtypes",[r.ViewContainerRef,r.TemplateRef,Je])],e)}(),nt=function(){function e(e){this._localization=e,this._caseViews={}}return Object.defineProperty(e.prototype,"ngPlural",{set:function(e){this._switchValue=e,this._updateView()},enumerable:!0,configurable:!0}),e.prototype.addCase=function(e,t){this._caseViews[e]=t},e.prototype._updateView=function(){this._clearViews();var e=Object.keys(this._caseViews),t=Be(this._switchValue,e,this._localization);this._activateView(this._caseViews[t])},e.prototype._clearViews=function(){this._activeView&&this._activeView.destroy()},e.prototype._activateView=function(e){e&&(this._activeView=e,this._activeView.create())},Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Number),Object(i.__metadata)("design:paramtypes",[Number])],e.prototype,"ngPlural",null),e=Object(i.__decorate)([Object(r.Directive)({selector:"[ngPlural]"}),Object(i.__metadata)("design:paramtypes",[Ve])],e)}(),rt=function(){function e(e,t,n,r){this.value=e;var i=!isNaN(Number(e));r.addCase(i?"="+e:e,new $e(n,t))}return e=Object(i.__decorate)([Object(r.Directive)({selector:"[ngPluralCase]"}),Object(i.__param)(0,Object(r.Attribute)("ngPluralCase")),Object(i.__param)(3,Object(r.Host)()),Object(i.__metadata)("design:paramtypes",[String,r.TemplateRef,r.ViewContainerRef,nt])],e)}(),it=function(){function e(e,t,n){this._differs=e,this._ngEl=t,this._renderer=n}return Object.defineProperty(e.prototype,"ngStyle",{set:function(e){this._ngStyle=e,!this._differ&&e&&(this._differ=this._differs.find(e).create())},enumerable:!0,configurable:!0}),e.prototype.ngDoCheck=function(){if(this._differ){var e=this._differ.diff(this._ngStyle);e&&this._applyChanges(e)}},e.prototype._applyChanges=function(e){var t=this;e.forEachRemovedItem(function(e){return t._setStyle(e.key,null)}),e.forEachAddedItem(function(e){return t._setStyle(e.key,e.currentValue)}),e.forEachChangedItem(function(e){return t._setStyle(e.key,e.currentValue)})},e.prototype._setStyle=function(e,t){var n=Object(i.__read)(e.split("."),2),r=n[0],o=n[1];null!=(t=null!=t&&o?""+t+o:t)?this._renderer.setStyle(this._ngEl.nativeElement,r,t):this._renderer.removeStyle(this._ngEl.nativeElement,r)},Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object),Object(i.__metadata)("design:paramtypes",[Object])],e.prototype,"ngStyle",null),e=Object(i.__decorate)([Object(r.Directive)({selector:"[ngStyle]"}),Object(i.__metadata)("design:paramtypes",[r.KeyValueDiffers,r.ElementRef,r.Renderer2])],e)}(),ot=function(){function e(e){this._viewContainerRef=e}return e.prototype.ngOnChanges=function(e){this._shouldRecreateView(e)?(this._viewRef&&this._viewContainerRef.remove(this._viewContainerRef.indexOf(this._viewRef)),this.ngTemplateOutlet&&(this._viewRef=this._viewContainerRef.createEmbeddedView(this.ngTemplateOutlet,this.ngTemplateOutletContext))):this._viewRef&&this.ngTemplateOutletContext&&this._updateExistingContext(this.ngTemplateOutletContext)},e.prototype._shouldRecreateView=function(e){var t=e.ngTemplateOutletContext;return!!e.ngTemplateOutlet||t&&this._hasContextShapeChanged(t)},e.prototype._hasContextShapeChanged=function(e){var t,n,r=Object.keys(e.previousValue||{}),o=Object.keys(e.currentValue||{});if(r.length===o.length){try{for(var a=Object(i.__values)(o),s=a.next();!s.done;s=a.next()){var u=s.value;if(-1===r.indexOf(u))return!0}}catch(e){t={error:e}}finally{try{s&&!s.done&&(n=a.return)&&n.call(a)}finally{if(t)throw t.error}}return!1}return!0},e.prototype._updateExistingContext=function(e){var t,n;try{for(var r=Object(i.__values)(Object.keys(e)),o=r.next();!o.done;o=r.next()){var a=o.value;this._viewRef.context[a]=this.ngTemplateOutletContext[a]}}catch(e){t={error:e}}finally{try{o&&!o.done&&(n=r.return)&&n.call(r)}finally{if(t)throw t.error}}},Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",Object)],e.prototype,"ngTemplateOutletContext",void 0),Object(i.__decorate)([Object(r.Input)(),Object(i.__metadata)("design:type",r.TemplateRef)],e.prototype,"ngTemplateOutlet",void 0),e=Object(i.__decorate)([Object(r.Directive)({selector:"[ngTemplateOutlet]"}),Object(i.__metadata)("design:paramtypes",[r.ViewContainerRef])],e)}(),at=[qe,Ge,Ye,Qe,ot,it,Je,et,tt,nt,rt];function st(e,t){return Error("InvalidPipeArgument: '"+t+"' for pipe '"+Object(r["ɵstringify"])(e)+"'")}var ut,ct=function(){function e(){}return e.format=function(e,t,n,r){void 0===r&&(r={});var i=r.minimumIntegerDigits,o=r.minimumFractionDigits,a=r.maximumFractionDigits,s=r.currency,u=r.currencyAsSymbol,c=void 0!==u&&u,l={minimumIntegerDigits:i,minimumFractionDigits:o,maximumFractionDigits:a,style:g[n].toLowerCase()};return n==g.Currency&&(l.currency="string"==typeof s?s:void 0,l.currencyDisplay=c?"symbol":"code"),new Intl.NumberFormat(t,l).format(e)},e}(),lt=/((?:[^yMLdHhmsazZEwGjJ']+)|(?:'(?:[^']|'')*')|(?:E+|y+|M+|L+|d+|H+|h+|J+|j+|m+|s+|a|z|Z|G+|w+))(.*)/,dt={yMMMdjms:jt(bt([_t("year",1),yt("month",3),_t("day",1),_t("hour",1),_t("minute",1),_t("second",1)])),yMdjm:jt(bt([_t("year",1),_t("month",1),_t("day",1),_t("hour",1),_t("minute",1)])),yMMMMEEEEd:jt(bt([_t("year",1),yt("month",4),yt("weekday",4),_t("day",1)])),yMMMMd:jt(bt([_t("year",1),yt("month",4),_t("day",1)])),yMMMd:jt(bt([_t("year",1),yt("month",3),_t("day",1)])),yMd:jt(bt([_t("year",1),_t("month",1),_t("day",1)])),jms:jt(bt([_t("hour",1),_t("second",1),_t("minute",1)])),jm:jt(bt([_t("hour",1),_t("minute",1)]))},pt={yyyy:jt(_t("year",4)),yy:jt(_t("year",2)),y:jt(_t("year",1)),MMMM:jt(yt("month",4)),MMM:jt(yt("month",3)),MM:jt(_t("month",2)),M:jt(_t("month",1)),LLLL:jt(yt("month",4)),L:jt(yt("month",1)),dd:jt(_t("day",2)),d:jt(_t("day",1)),HH:ft(ht(jt(vt(_t("hour",2),!1)))),H:ht(jt(vt(_t("hour",1),!1))),hh:ft(ht(jt(vt(_t("hour",2),!0)))),h:ht(jt(vt(_t("hour",1),!0))),jj:jt(_t("hour",2)),j:jt(_t("hour",1)),mm:ft(jt(_t("minute",2))),m:jt(_t("minute",1)),ss:ft(jt(_t("second",2))),s:jt(_t("second",1)),sss:jt(_t("second",3)),EEEE:jt(yt("weekday",4)),EEE:jt(yt("weekday",3)),EE:jt(yt("weekday",2)),E:jt(yt("weekday",1)),a:(ut=jt(vt(_t("hour",1),!0)),function(e,t){return ut(e,t).split(" ")[1]}),Z:gt("short"),z:gt("long"),ww:jt({}),w:jt({}),G:jt(yt("era",1)),GG:jt(yt("era",2)),GGG:jt(yt("era",3)),GGGG:jt(yt("era",4))};function ft(e){return function(t,n){var r=e(t,n);return 1==r.length?"0"+r:r}}function ht(e){return function(t,n){return e(t,n).split(" ")[0]}}function mt(e,t,n){return new Intl.DateTimeFormat(t,n).format(e).replace(/[\u200e\u200f]/g,"")}function gt(e){var t={hour:"2-digit",hour12:!1,timeZoneName:e};return function(e,n){var r=mt(e,n,t);return r?r.substring(3):""}}function vt(e,t){return e.hour12=t,e}function _t(e,t){var n={};return n[e]=2===t?"2-digit":"numeric",n}function yt(e,t){var n={};return n[e]=t<4?t>1?"short":"narrow":"long",n}function bt(e){return e.reduce(function(e,t){return Object(i.__assign)({},e,t)},{})}function jt(e){return function(t,n){return mt(t,n,e)}}var xt=new Map;var Ot=function(){function e(){}return e.format=function(e,t,n){return function(e,t,n){var r=dt[e];if(r)return r(t,n);var i=e,o=xt.get(i);if(!o){o=[];var a=void 0;lt.exec(e);for(var s=e;s;)(a=lt.exec(s))?s=(o=o.concat(a.slice(1))).pop():(o.push(s),s=null);xt.set(i,o)}return o.reduce(function(e,r){var i,o=pt[r];return e+(o?o(t,n):"''"===(i=r)?"'":i.replace(/(^'|'$)/g,"").replace(/''/g,"'"))},"")}(n,e,t)},e}(),wt=function(){function e(e){this._locale=e}var t;return t=e,e.prototype.transform=function(e,n){if(void 0===n&&(n="mediumDate"),null==e||""===e||e!=e)return null;var r;if("string"==typeof e&&(e=e.trim()),Ct(e))r=e;else if(isNaN(e-parseFloat(e)))if("string"==typeof e&&/^(\d{4}-\d{1,2}-\d{1,2})$/.test(e)){var o=Object(i.__read)(e.split("-").map(function(e){return parseInt(e,10)}),3),a=o[0],s=o[1],u=o[2];r=new Date(a,s-1,u)}else r=new Date(e);else r=new Date(parseFloat(e));if(!Ct(r)){var c=void 0;if("string"!=typeof e||!(c=e.match(se)))throw st(t,e);r=je(c)}return Ot.format(r,this._locale,t._ALIASES[n]||n)},e._ALIASES={medium:"yMMMdjms",short:"yMdjm",fullDate:"yMMMMEEEEd",longDate:"yMMMMd",mediumDate:"yMMMd",shortDate:"yMd",mediumTime:"jms",shortTime:"jm"},e=t=Object(i.__decorate)([Object(r.Pipe)({name:"date",pure:!0}),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__metadata)("design:paramtypes",[String])],e)}();function Ct(e){return e instanceof Date&&!isNaN(e.valueOf())}function Et(e,t,n,r,i,o,a){if(void 0===o&&(o=null),void 0===a&&(a=!1),null==n)return null;if("number"!=typeof(n="string"!=typeof n||isNaN(+n-parseFloat(n))?n:+n))throw st(e,n);var s,u,c;if(r!==g.Currency&&(s=1,u=0,c=3),i){var l=i.match(Oe);if(null===l)throw new Error(i+" is not a valid digit info for number pipes");null!=l[1]&&(s=Fe(l[1])),null!=l[3]&&(u=Fe(l[3])),null!=l[5]&&(c=Fe(l[5]))}return ct.format(n,t,r,{minimumIntegerDigits:s,minimumFractionDigits:u,maximumFractionDigits:c,currency:o,currencyAsSymbol:a})}var St=function(){function e(e){this._locale=e}var t;return t=e,e.prototype.transform=function(e,n){return Et(t,this._locale,e,g.Decimal,n)},e=t=Object(i.__decorate)([Object(r.Pipe)({name:"number"}),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__metadata)("design:paramtypes",[String])],e)}(),kt=function(){function e(e){this._locale=e}var t;return t=e,e.prototype.transform=function(e,n){return Et(t,this._locale,e,g.Percent,n)},e=t=Object(i.__decorate)([Object(r.Pipe)({name:"percent"}),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__metadata)("design:paramtypes",[String])],e)}(),Mt=function(){function e(e){this._locale=e}var t;return t=e,e.prototype.transform=function(e,n,r,i){return void 0===n&&(n="USD"),void 0===r&&(r=!1),Et(t,this._locale,e,g.Currency,i,n,r)},e=t=Object(i.__decorate)([Object(r.Pipe)({name:"currency"}),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__metadata)("design:paramtypes",[String])],e)}(),At=[St,kt,Mt,wt],Tt=function(){function e(){}return e.prototype.createSubscription=function(e,t){return e.subscribe({next:t,error:function(e){throw e}})},e.prototype.dispose=function(e){e.unsubscribe()},e.prototype.onDestroy=function(e){e.unsubscribe()},e}(),Dt=new(function(){function e(){}return e.prototype.createSubscription=function(e,t){return e.then(t,function(e){throw e})},e.prototype.dispose=function(e){},e.prototype.onDestroy=function(e){},e}()),It=new Tt,Pt=function(){function e(e){this._ref=e,this._latestValue=null,this._latestReturnedValue=null,this._subscription=null,this._obj=null,this._strategy=null}var t;return t=e,e.prototype.ngOnDestroy=function(){this._subscription&&this._dispose()},e.prototype.transform=function(e){return this._obj?e!==this._obj?(this._dispose(),this.transform(e)):this._latestValue===this._latestReturnedValue?this._latestReturnedValue:(this._latestReturnedValue=this._latestValue,r.WrappedValue.wrap(this._latestValue)):(e&&this._subscribe(e),this._latestReturnedValue=this._latestValue,this._latestValue)},e.prototype._subscribe=function(e){var t=this;this._obj=e,this._strategy=this._selectStrategy(e),this._subscription=this._strategy.createSubscription(e,function(n){return t._updateLatestValue(e,n)})},e.prototype._selectStrategy=function(e){if(Object(r["ɵisPromise"])(e))return Dt;if(Object(r["ɵisObservable"])(e))return It;throw st(t,e)},e.prototype._dispose=function(){this._strategy.dispose(this._subscription),this._latestValue=null,this._latestReturnedValue=null,this._subscription=null,this._obj=null},e.prototype._updateLatestValue=function(e,t){e===this._obj&&(this._latestValue=t,this._ref.markForCheck())},e=t=Object(i.__decorate)([Object(r.Pipe)({name:"async",pure:!1}),Object(i.__metadata)("design:paramtypes",[r.ChangeDetectorRef])],e)}(),Rt=function(){function e(){}var t;return t=e,e.prototype.transform=function(e){if(!e)return e;if("string"!=typeof e)throw st(t,e);return e.toLowerCase()},e=t=Object(i.__decorate)([Object(r.Pipe)({name:"lowercase"})],e)}(),Nt=/(?:[A-Za-z\xAA\xB5\xBA\xC0-\xD6\xD8-\xF6\xF8-\u02C1\u02C6-\u02D1\u02E0-\u02E4\u02EC\u02EE\u0370-\u0374\u0376\u0377\u037A-\u037D\u037F\u0386\u0388-\u038A\u038C\u038E-\u03A1\u03A3-\u03F5\u03F7-\u0481\u048A-\u052F\u0531-\u0556\u0559\u0561-\u0587\u05D0-\u05EA\u05F0-\u05F2\u0620-\u064A\u066E\u066F\u0671-\u06D3\u06D5\u06E5\u06E6\u06EE\u06EF\u06FA-\u06FC\u06FF\u0710\u0712-\u072F\u074D-\u07A5\u07B1\u07CA-\u07EA\u07F4\u07F5\u07FA\u0800-\u0815\u081A\u0824\u0828\u0840-\u0858\u0860-\u086A\u08A0-\u08B4\u08B6-\u08BD\u0904-\u0939\u093D\u0950\u0958-\u0961\u0971-\u0980\u0985-\u098C\u098F\u0990\u0993-\u09A8\u09AA-\u09B0\u09B2\u09B6-\u09B9\u09BD\u09CE\u09DC\u09DD\u09DF-\u09E1\u09F0\u09F1\u09FC\u0A05-\u0A0A\u0A0F\u0A10\u0A13-\u0A28\u0A2A-\u0A30\u0A32\u0A33\u0A35\u0A36\u0A38\u0A39\u0A59-\u0A5C\u0A5E\u0A72-\u0A74\u0A85-\u0A8D\u0A8F-\u0A91\u0A93-\u0AA8\u0AAA-\u0AB0\u0AB2\u0AB3\u0AB5-\u0AB9\u0ABD\u0AD0\u0AE0\u0AE1\u0AF9\u0B05-\u0B0C\u0B0F\u0B10\u0B13-\u0B28\u0B2A-\u0B30\u0B32\u0B33\u0B35-\u0B39\u0B3D\u0B5C\u0B5D\u0B5F-\u0B61\u0B71\u0B83\u0B85-\u0B8A\u0B8E-\u0B90\u0B92-\u0B95\u0B99\u0B9A\u0B9C\u0B9E\u0B9F\u0BA3\u0BA4\u0BA8-\u0BAA\u0BAE-\u0BB9\u0BD0\u0C05-\u0C0C\u0C0E-\u0C10\u0C12-\u0C28\u0C2A-\u0C39\u0C3D\u0C58-\u0C5A\u0C60\u0C61\u0C80\u0C85-\u0C8C\u0C8E-\u0C90\u0C92-\u0CA8\u0CAA-\u0CB3\u0CB5-\u0CB9\u0CBD\u0CDE\u0CE0\u0CE1\u0CF1\u0CF2\u0D05-\u0D0C\u0D0E-\u0D10\u0D12-\u0D3A\u0D3D\u0D4E\u0D54-\u0D56\u0D5F-\u0D61\u0D7A-\u0D7F\u0D85-\u0D96\u0D9A-\u0DB1\u0DB3-\u0DBB\u0DBD\u0DC0-\u0DC6\u0E01-\u0E30\u0E32\u0E33\u0E40-\u0E46\u0E81\u0E82\u0E84\u0E87\u0E88\u0E8A\u0E8D\u0E94-\u0E97\u0E99-\u0E9F\u0EA1-\u0EA3\u0EA5\u0EA7\u0EAA\u0EAB\u0EAD-\u0EB0\u0EB2\u0EB3\u0EBD\u0EC0-\u0EC4\u0EC6\u0EDC-\u0EDF\u0F00\u0F40-\u0F47\u0F49-\u0F6C\u0F88-\u0F8C\u1000-\u102A\u103F\u1050-\u1055\u105A-\u105D\u1061\u1065\u1066\u106E-\u1070\u1075-\u1081\u108E\u10A0-\u10C5\u10C7\u10CD\u10D0-\u10FA\u10FC-\u1248\u124A-\u124D\u1250-\u1256\u1258\u125A-\u125D\u1260-\u1288\u128A-\u128D\u1290-\u12B0\u12B2-\u12B5\u12B8-\u12BE\u12C0\u12C2-\u12C5\u12C8-\u12D6\u12D8-\u1310\u1312-\u1315\u1318-\u135A\u1380-\u138F\u13A0-\u13F5\u13F8-\u13FD\u1401-\u166C\u166F-\u167F\u1681-\u169A\u16A0-\u16EA\u16F1-\u16F8\u1700-\u170C\u170E-\u1711\u1720-\u1731\u1740-\u1751\u1760-\u176C\u176E-\u1770\u1780-\u17B3\u17D7\u17DC\u1820-\u1877\u1880-\u1884\u1887-\u18A8\u18AA\u18B0-\u18F5\u1900-\u191E\u1950-\u196D\u1970-\u1974\u1980-\u19AB\u19B0-\u19C9\u1A00-\u1A16\u1A20-\u1A54\u1AA7\u1B05-\u1B33\u1B45-\u1B4B\u1B83-\u1BA0\u1BAE\u1BAF\u1BBA-\u1BE5\u1C00-\u1C23\u1C4D-\u1C4F\u1C5A-\u1C7D\u1C80-\u1C88\u1CE9-\u1CEC\u1CEE-\u1CF1\u1CF5\u1CF6\u1D00-\u1DBF\u1E00-\u1F15\u1F18-\u1F1D\u1F20-\u1F45\u1F48-\u1F4D\u1F50-\u1F57\u1F59\u1F5B\u1F5D\u1F5F-\u1F7D\u1F80-\u1FB4\u1FB6-\u1FBC\u1FBE\u1FC2-\u1FC4\u1FC6-\u1FCC\u1FD0-\u1FD3\u1FD6-\u1FDB\u1FE0-\u1FEC\u1FF2-\u1FF4\u1FF6-\u1FFC\u2071\u207F\u2090-\u209C\u2102\u2107\u210A-\u2113\u2115\u2119-\u211D\u2124\u2126\u2128\u212A-\u212D\u212F-\u2139\u213C-\u213F\u2145-\u2149\u214E\u2183\u2184\u2C00-\u2C2E\u2C30-\u2C5E\u2C60-\u2CE4\u2CEB-\u2CEE\u2CF2\u2CF3\u2D00-\u2D25\u2D27\u2D2D\u2D30-\u2D67\u2D6F\u2D80-\u2D96\u2DA0-\u2DA6\u2DA8-\u2DAE\u2DB0-\u2DB6\u2DB8-\u2DBE\u2DC0-\u2DC6\u2DC8-\u2DCE\u2DD0-\u2DD6\u2DD8-\u2DDE\u2E2F\u3005\u3006\u3031-\u3035\u303B\u303C\u3041-\u3096\u309D-\u309F\u30A1-\u30FA\u30FC-\u30FF\u3105-\u312E\u3131-\u318E\u31A0-\u31BA\u31F0-\u31FF\u3400-\u4DB5\u4E00-\u9FEA\uA000-\uA48C\uA4D0-\uA4FD\uA500-\uA60C\uA610-\uA61F\uA62A\uA62B\uA640-\uA66E\uA67F-\uA69D\uA6A0-\uA6E5\uA717-\uA71F\uA722-\uA788\uA78B-\uA7AE\uA7B0-\uA7B7\uA7F7-\uA801\uA803-\uA805\uA807-\uA80A\uA80C-\uA822\uA840-\uA873\uA882-\uA8B3\uA8F2-\uA8F7\uA8FB\uA8FD\uA90A-\uA925\uA930-\uA946\uA960-\uA97C\uA984-\uA9B2\uA9CF\uA9E0-\uA9E4\uA9E6-\uA9EF\uA9FA-\uA9FE\uAA00-\uAA28\uAA40-\uAA42\uAA44-\uAA4B\uAA60-\uAA76\uAA7A\uAA7E-\uAAAF\uAAB1\uAAB5\uAAB6\uAAB9-\uAABD\uAAC0\uAAC2\uAADB-\uAADD\uAAE0-\uAAEA\uAAF2-\uAAF4\uAB01-\uAB06\uAB09-\uAB0E\uAB11-\uAB16\uAB20-\uAB26\uAB28-\uAB2E\uAB30-\uAB5A\uAB5C-\uAB65\uAB70-\uABE2\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFA6D\uFA70-\uFAD9\uFB00-\uFB06\uFB13-\uFB17\uFB1D\uFB1F-\uFB28\uFB2A-\uFB36\uFB38-\uFB3C\uFB3E\uFB40\uFB41\uFB43\uFB44\uFB46-\uFBB1\uFBD3-\uFD3D\uFD50-\uFD8F\uFD92-\uFDC7\uFDF0-\uFDFB\uFE70-\uFE74\uFE76-\uFEFC\uFF21-\uFF3A\uFF41-\uFF5A\uFF66-\uFFBE\uFFC2-\uFFC7\uFFCA-\uFFCF\uFFD2-\uFFD7\uFFDA-\uFFDC]|\uD800[\uDC00-\uDC0B\uDC0D-\uDC26\uDC28-\uDC3A\uDC3C\uDC3D\uDC3F-\uDC4D\uDC50-\uDC5D\uDC80-\uDCFA\uDE80-\uDE9C\uDEA0-\uDED0\uDF00-\uDF1F\uDF2D-\uDF40\uDF42-\uDF49\uDF50-\uDF75\uDF80-\uDF9D\uDFA0-\uDFC3\uDFC8-\uDFCF]|\uD801[\uDC00-\uDC9D\uDCB0-\uDCD3\uDCD8-\uDCFB\uDD00-\uDD27\uDD30-\uDD63\uDE00-\uDF36\uDF40-\uDF55\uDF60-\uDF67]|\uD802[\uDC00-\uDC05\uDC08\uDC0A-\uDC35\uDC37\uDC38\uDC3C\uDC3F-\uDC55\uDC60-\uDC76\uDC80-\uDC9E\uDCE0-\uDCF2\uDCF4\uDCF5\uDD00-\uDD15\uDD20-\uDD39\uDD80-\uDDB7\uDDBE\uDDBF\uDE00\uDE10-\uDE13\uDE15-\uDE17\uDE19-\uDE33\uDE60-\uDE7C\uDE80-\uDE9C\uDEC0-\uDEC7\uDEC9-\uDEE4\uDF00-\uDF35\uDF40-\uDF55\uDF60-\uDF72\uDF80-\uDF91]|\uD803[\uDC00-\uDC48\uDC80-\uDCB2\uDCC0-\uDCF2]|\uD804[\uDC03-\uDC37\uDC83-\uDCAF\uDCD0-\uDCE8\uDD03-\uDD26\uDD50-\uDD72\uDD76\uDD83-\uDDB2\uDDC1-\uDDC4\uDDDA\uDDDC\uDE00-\uDE11\uDE13-\uDE2B\uDE80-\uDE86\uDE88\uDE8A-\uDE8D\uDE8F-\uDE9D\uDE9F-\uDEA8\uDEB0-\uDEDE\uDF05-\uDF0C\uDF0F\uDF10\uDF13-\uDF28\uDF2A-\uDF30\uDF32\uDF33\uDF35-\uDF39\uDF3D\uDF50\uDF5D-\uDF61]|\uD805[\uDC00-\uDC34\uDC47-\uDC4A\uDC80-\uDCAF\uDCC4\uDCC5\uDCC7\uDD80-\uDDAE\uDDD8-\uDDDB\uDE00-\uDE2F\uDE44\uDE80-\uDEAA\uDF00-\uDF19]|\uD806[\uDCA0-\uDCDF\uDCFF\uDE00\uDE0B-\uDE32\uDE3A\uDE50\uDE5C-\uDE83\uDE86-\uDE89\uDEC0-\uDEF8]|\uD807[\uDC00-\uDC08\uDC0A-\uDC2E\uDC40\uDC72-\uDC8F\uDD00-\uDD06\uDD08\uDD09\uDD0B-\uDD30\uDD46]|\uD808[\uDC00-\uDF99]|\uD809[\uDC80-\uDD43]|[\uD80C\uD81C-\uD820\uD840-\uD868\uD86A-\uD86C\uD86F-\uD872\uD874-\uD879][\uDC00-\uDFFF]|\uD80D[\uDC00-\uDC2E]|\uD811[\uDC00-\uDE46]|\uD81A[\uDC00-\uDE38\uDE40-\uDE5E\uDED0-\uDEED\uDF00-\uDF2F\uDF40-\uDF43\uDF63-\uDF77\uDF7D-\uDF8F]|\uD81B[\uDF00-\uDF44\uDF50\uDF93-\uDF9F\uDFE0\uDFE1]|\uD821[\uDC00-\uDFEC]|\uD822[\uDC00-\uDEF2]|\uD82C[\uDC00-\uDD1E\uDD70-\uDEFB]|\uD82F[\uDC00-\uDC6A\uDC70-\uDC7C\uDC80-\uDC88\uDC90-\uDC99]|\uD835[\uDC00-\uDC54\uDC56-\uDC9C\uDC9E\uDC9F\uDCA2\uDCA5\uDCA6\uDCA9-\uDCAC\uDCAE-\uDCB9\uDCBB\uDCBD-\uDCC3\uDCC5-\uDD05\uDD07-\uDD0A\uDD0D-\uDD14\uDD16-\uDD1C\uDD1E-\uDD39\uDD3B-\uDD3E\uDD40-\uDD44\uDD46\uDD4A-\uDD50\uDD52-\uDEA5\uDEA8-\uDEC0\uDEC2-\uDEDA\uDEDC-\uDEFA\uDEFC-\uDF14\uDF16-\uDF34\uDF36-\uDF4E\uDF50-\uDF6E\uDF70-\uDF88\uDF8A-\uDFA8\uDFAA-\uDFC2\uDFC4-\uDFCB]|\uD83A[\uDC00-\uDCC4\uDD00-\uDD43]|\uD83B[\uDE00-\uDE03\uDE05-\uDE1F\uDE21\uDE22\uDE24\uDE27\uDE29-\uDE32\uDE34-\uDE37\uDE39\uDE3B\uDE42\uDE47\uDE49\uDE4B\uDE4D-\uDE4F\uDE51\uDE52\uDE54\uDE57\uDE59\uDE5B\uDE5D\uDE5F\uDE61\uDE62\uDE64\uDE67-\uDE6A\uDE6C-\uDE72\uDE74-\uDE77\uDE79-\uDE7C\uDE7E\uDE80-\uDE89\uDE8B-\uDE9B\uDEA1-\uDEA3\uDEA5-\uDEA9\uDEAB-\uDEBB]|\uD869[\uDC00-\uDED6\uDF00-\uDFFF]|\uD86D[\uDC00-\uDF34\uDF40-\uDFFF]|\uD86E[\uDC00-\uDC1D\uDC20-\uDFFF]|\uD873[\uDC00-\uDEA1\uDEB0-\uDFFF]|\uD87A[\uDC00-\uDFE0]|\uD87E[\uDC00-\uDE1D])\S*/g,Ft=function(){function e(){}var t;return t=e,e.prototype.transform=function(e){if(!e)return e;if("string"!=typeof e)throw st(t,e);return e.replace(Nt,function(e){return e[0].toUpperCase()+e.substr(1).toLowerCase()})},e=t=Object(i.__decorate)([Object(r.Pipe)({name:"titlecase"})],e)}(),Lt=function(){function e(){}var t;return t=e,e.prototype.transform=function(e){if(!e)return e;if("string"!=typeof e)throw st(t,e);return e.toUpperCase()},e=t=Object(i.__decorate)([Object(r.Pipe)({name:"uppercase"})],e)}(),Vt=function(){function e(e){this.locale=e}var t;return t=e,e.prototype.transform=function(e,n,r,i){if(void 0===n&&(n="mediumDate"),null==e||""===e||e!=e)return null;try{return le(e,n,i||this.locale,r)}catch(e){throw st(t,e.message)}},e=t=Object(i.__decorate)([Object(r.Pipe)({name:"date",pure:!0}),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__metadata)("design:paramtypes",[String])],e)}(),Bt=/#/g,zt=function(){function e(e){this._localization=e}var t;return t=e,e.prototype.transform=function(e,n,r){if(null==e)return"";if("object"!=typeof n||null===n)throw st(t,n);return n[Be(e,Object.keys(n),this._localization,r)].replace(Bt,e.toString())},e=t=Object(i.__decorate)([Object(r.Pipe)({name:"i18nPlural",pure:!0}),Object(i.__metadata)("design:paramtypes",[Ve])],e)}(),Ut=function(){function e(){}var t;return t=e,e.prototype.transform=function(e,n){if(null==e)return"";if("object"!=typeof n||"string"!=typeof e)throw st(t,n);return n.hasOwnProperty(e)?n[e]:n.hasOwnProperty("other")?n.other:""},e=t=Object(i.__decorate)([Object(r.Pipe)({name:"i18nSelect",pure:!0})],e)}(),Ht=function(){function e(){}return e.prototype.transform=function(e){return JSON.stringify(e,null,2)},e=Object(i.__decorate)([Object(r.Pipe)({name:"json",pure:!1})],e)}();var qt=function(){function e(e){this.differs=e,this.keyValues=[]}return e.prototype.transform=function(e,t){var n=this;if(void 0===t&&(t=Gt),!e||!(e instanceof Map)&&"object"!=typeof e)return null;this.differ||(this.differ=this.differs.find(e).create());var r=this.differ.diff(e);return r&&(this.keyValues=[],r.forEachItem(function(e){var t,r;n.keyValues.push((t=e.key,r=e.currentValue,{key:t,value:r}))}),this.keyValues.sort(t)),this.keyValues},e=Object(i.__decorate)([Object(r.Pipe)({name:"keyvalue",pure:!1}),Object(i.__metadata)("design:paramtypes",[r.KeyValueDiffers])],e)}();function Gt(e,t){var n=e.key,r=t.key;if(n===r)return 0;if(void 0===n)return 1;if(void 0===r)return-1;if(null===n)return 1;if(null===r)return-1;if("string"==typeof n&&"string"==typeof r)return n<r?-1:1;if("number"==typeof n&&"number"==typeof r)return n-r;if("boolean"==typeof n&&"boolean"==typeof r)return n<r?-1:1;var i=String(n),o=String(r);return i==o?0:i<o?-1:1}var Wt=function(){function e(e){this._locale=e}var t;return t=e,e.prototype.transform=function(e,n,r){if(Qt(e))return null;r=r||this._locale;try{return Re(Xt(e),r,n)}catch(e){throw st(t,e.message)}},e=t=Object(i.__decorate)([Object(r.Pipe)({name:"number"}),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__metadata)("design:paramtypes",[String])],e)}(),Yt=function(){function e(e){this._locale=e}var t;return t=e,e.prototype.transform=function(e,n,r){if(Qt(e))return null;r=r||this._locale;try{return Pe(Xt(e),r,n)}catch(e){throw st(t,e.message)}},e=t=Object(i.__decorate)([Object(r.Pipe)({name:"percent"}),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__metadata)("design:paramtypes",[String])],e)}(),Kt=function(){function e(e){this._locale=e}var t;return t=e,e.prototype.transform=function(e,n,r,i,o){if(void 0===r&&(r="symbol"),Qt(e))return null;o=o||this._locale,"boolean"==typeof r&&(console&&console.warn&&console.warn('Warning: the currency pipe has been changed in Angular v5. The symbolDisplay option (third parameter) is now a string instead of a boolean. The accepted values are "code", "symbol" or "symbol-narrow".'),r=r?"symbol":"code");var a=n||"USD";"code"!==r&&(a="symbol"===r||"symbol-narrow"===r?$(a,"symbol"===r?"wide":"narrow",o):r);try{return Ie(Xt(e),o,a,n,i)}catch(e){throw st(t,e.message)}},e=t=Object(i.__decorate)([Object(r.Pipe)({name:"currency"}),Object(i.__param)(0,Object(r.Inject)(r.LOCALE_ID)),Object(i.__metadata)("design:paramtypes",[String])],e)}();function Qt(e){return null==e||""===e||e!=e}function Xt(e){if("string"==typeof e&&!isNaN(Number(e)-parseFloat(e)))return Number(e);if("number"!=typeof e)throw new Error(e+" is not a number");return e}var Zt=function(){function e(){}var t;return t=e,e.prototype.transform=function(e,n,r){if(null==e)return e;if(!this.supports(e))throw st(t,e);return e.slice(n,r)},e.prototype.supports=function(e){return"string"==typeof e||Array.isArray(e)},e=t=Object(i.__decorate)([Object(r.Pipe)({name:"slice",pure:!1})],e)}(),$t=[Pt,Lt,Rt,Ht,Zt,Wt,Yt,Ft,Kt,Vt,zt,Ut,qt],Jt=function(){function e(){}return e=Object(i.__decorate)([Object(r.NgModule)({declarations:[at,$t],exports:[at,$t],providers:[{provide:Ve,useClass:ze}]})],e)}(),en=Ue,tn=function(){function e(){}return e=Object(i.__decorate)([Object(r.NgModule)({declarations:[At],exports:[At],providers:[{provide:Le,useValue:en}]})],e)}(),nn=new r.InjectionToken("DocumentToken"),rn="browser",on="server",an="browserWorkerApp",sn="browserWorkerUi";function un(e){return e===rn}function cn(e){return e===on}function ln(e){return e===an}function dn(e){return e===sn}var pn=new r.Version("7.2.0"),fn=function(){function e(){}return e.ngInjectableDef=Object(r.defineInjectable)({providedIn:"root",factory:function(){return new hn(Object(r.inject)(nn),window)}}),e}(),hn=function(){function e(e,t){this.document=e,this.window=t,this.offset=function(){return[0,0]}}return e.prototype.setOffset=function(e){Array.isArray(e)?this.offset=function(){return e}:this.offset=e},e.prototype.getScrollPosition=function(){return this.supportScrollRestoration()?[this.window.scrollX,this.window.scrollY]:[0,0]},e.prototype.scrollToPosition=function(e){this.supportScrollRestoration()&&this.window.scrollTo(e[0],e[1])},e.prototype.scrollToAnchor=function(e){if(this.supportScrollRestoration()){var t=this.document.querySelector("#"+e);if(t)return void this.scrollToElement(t);var n=this.document.querySelector("[name='"+e+"']");if(n)return void this.scrollToElement(n)}},e.prototype.setHistoryScrollRestoration=function(e){if(this.supportScrollRestoration()){var t=this.window.history;t&&t.scrollRestoration&&(t.scrollRestoration=e)}},e.prototype.scrollToElement=function(e){var t=e.getBoundingClientRect(),n=t.left+this.window.pageXOffset,r=t.top+this.window.pageYOffset,i=this.offset();this.window.scrollTo(n-i[0],r-i[1])},e.prototype.supportScrollRestoration=function(){try{return!!this.window&&!!this.window.scrollTo}catch(e){return!1}},e}(),mn=function(){function e(){}return e.prototype.setOffset=function(e){},e.prototype.getScrollPosition=function(){return[0,0]},e.prototype.scrollToPosition=function(e){},e.prototype.scrollToAnchor=function(e){},e.prototype.setHistoryScrollRestoration=function(e){},e}()},"./node_modules/@angular/common/fesm5/http.js":function(e,t,n){"use strict";n.r(t),n.d(t,"ɵangular_packages_common_http_http_a",function(){return k}),n.d(t,"ɵangular_packages_common_http_http_b",function(){return A}),n.d(t,"ɵangular_packages_common_http_http_c",function(){return H}),n.d(t,"ɵangular_packages_common_http_http_d",function(){return R}),n.d(t,"ɵangular_packages_common_http_http_g",function(){return B}),n.d(t,"ɵangular_packages_common_http_http_h",function(){return z}),n.d(t,"ɵangular_packages_common_http_http_e",function(){return F}),n.d(t,"ɵangular_packages_common_http_http_f",function(){return L}),n.d(t,"HttpBackend",function(){return c}),n.d(t,"HttpHandler",function(){return u}),n.d(t,"HttpClient",function(){return C}),n.d(t,"HttpHeaders",function(){return l}),n.d(t,"HTTP_INTERCEPTORS",function(){return S}),n.d(t,"JsonpClientBackend",function(){return T}),n.d(t,"JsonpInterceptor",function(){return D}),n.d(t,"HttpClientJsonpModule",function(){return W}),n.d(t,"HttpClientModule",function(){return G}),n.d(t,"HttpClientXsrfModule",function(){return q}),n.d(t,"ɵHttpInterceptingHandler",function(){return U}),n.d(t,"HttpParams",function(){return f}),n.d(t,"HttpUrlEncodingCodec",function(){return d}),n.d(t,"HttpRequest",function(){return y}),n.d(t,"HttpErrorResponse",function(){return O}),n.d(t,"HttpEventType",function(){return v}),n.d(t,"HttpHeaderResponse",function(){return j}),n.d(t,"HttpResponse",function(){return x}),n.d(t,"HttpResponseBase",function(){return b}),n.d(t,"HttpXhrBackend",function(){return N}),n.d(t,"XhrFactory",function(){return P}),n.d(t,"HttpXsrfTokenExtractor",function(){return V});var r=n("./node_modules/@angular/common/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/rxjs/_esm5/index.js"),a=n("./node_modules/rxjs/_esm5/operators/index.js"),s=n("./node_modules/@angular/common/fesm5/common.js"),u=function(){return function(){}}(),c=function(){return function(){}}(),l=function(){function e(e){var t=this;this.normalizedNames=new Map,this.lazyUpdate=null,e?this.lazyInit="string"==typeof e?function(){t.headers=new Map,e.split("\n").forEach(function(e){var n=e.indexOf(":");if(n>0){var r=e.slice(0,n),i=r.toLowerCase(),o=e.slice(n+1).trim();t.maybeSetNormalizedName(r,i),t.headers.has(i)?t.headers.get(i).push(o):t.headers.set(i,[o])}})}:function(){t.headers=new Map,Object.keys(e).forEach(function(n){var r=e[n],i=n.toLowerCase();"string"==typeof r&&(r=[r]),r.length>0&&(t.headers.set(i,r),t.maybeSetNormalizedName(n,i))})}:this.headers=new Map}return e.prototype.has=function(e){return this.init(),this.headers.has(e.toLowerCase())},e.prototype.get=function(e){this.init();var t=this.headers.get(e.toLowerCase());return t&&t.length>0?t[0]:null},e.prototype.keys=function(){return this.init(),Array.from(this.normalizedNames.values())},e.prototype.getAll=function(e){return this.init(),this.headers.get(e.toLowerCase())||null},e.prototype.append=function(e,t){return this.clone({name:e,value:t,op:"a"})},e.prototype.set=function(e,t){return this.clone({name:e,value:t,op:"s"})},e.prototype.delete=function(e,t){return this.clone({name:e,value:t,op:"d"})},e.prototype.maybeSetNormalizedName=function(e,t){this.normalizedNames.has(t)||this.normalizedNames.set(t,e)},e.prototype.init=function(){var t=this;this.lazyInit&&(this.lazyInit instanceof e?this.copyFrom(this.lazyInit):this.lazyInit(),this.lazyInit=null,this.lazyUpdate&&(this.lazyUpdate.forEach(function(e){return t.applyUpdate(e)}),this.lazyUpdate=null))},e.prototype.copyFrom=function(e){var t=this;e.init(),Array.from(e.headers.keys()).forEach(function(n){t.headers.set(n,e.headers.get(n)),t.normalizedNames.set(n,e.normalizedNames.get(n))})},e.prototype.clone=function(t){var n=new e;return n.lazyInit=this.lazyInit&&this.lazyInit instanceof e?this.lazyInit:this,n.lazyUpdate=(this.lazyUpdate||[]).concat([t]),n},e.prototype.applyUpdate=function(e){var t=e.name.toLowerCase();switch(e.op){case"a":case"s":var n=e.value;if("string"==typeof n&&(n=[n]),0===n.length)return;this.maybeSetNormalizedName(e.name,t);var i=("a"===e.op?this.headers.get(t):void 0)||[];i.push.apply(i,Object(r.__spread)(n)),this.headers.set(t,i);break;case"d":var o=e.value;if(o){var a=this.headers.get(t);if(!a)return;0===(a=a.filter(function(e){return-1===o.indexOf(e)})).length?(this.headers.delete(t),this.normalizedNames.delete(t)):this.headers.set(t,a)}else this.headers.delete(t),this.normalizedNames.delete(t)}},e.prototype.forEach=function(e){var t=this;this.init(),Array.from(this.normalizedNames.keys()).forEach(function(n){return e(t.normalizedNames.get(n),t.headers.get(n))})},e}(),d=function(){function e(){}return e.prototype.encodeKey=function(e){return p(e)},e.prototype.encodeValue=function(e){return p(e)},e.prototype.decodeKey=function(e){return decodeURIComponent(e)},e.prototype.decodeValue=function(e){return decodeURIComponent(e)},e}();function p(e){return encodeURIComponent(e).replace(/%40/gi,"@").replace(/%3A/gi,":").replace(/%24/gi,"$").replace(/%2C/gi,",").replace(/%3B/gi,";").replace(/%2B/gi,"+").replace(/%3D/gi,"=").replace(/%3F/gi,"?").replace(/%2F/gi,"/")}var f=function(){function e(e){void 0===e&&(e={});var t,n,i,o=this;if(this.updates=null,this.cloneFrom=null,this.encoder=e.encoder||new d,e.fromString){if(e.fromObject)throw new Error("Cannot specify both fromString and fromObject.");this.map=(t=e.fromString,n=this.encoder,i=new Map,t.length>0&&t.split("&").forEach(function(e){var t=e.indexOf("="),o=Object(r.__read)(-1==t?[n.decodeKey(e),""]:[n.decodeKey(e.slice(0,t)),n.decodeValue(e.slice(t+1))],2),a=o[0],s=o[1],u=i.get(a)||[];u.push(s),i.set(a,u)}),i)}else e.fromObject?(this.map=new Map,Object.keys(e.fromObject).forEach(function(t){var n=e.fromObject[t];o.map.set(t,Array.isArray(n)?n:[n])})):this.map=null}return e.prototype.has=function(e){return this.init(),this.map.has(e)},e.prototype.get=function(e){this.init();var t=this.map.get(e);return t?t[0]:null},e.prototype.getAll=function(e){return this.init(),this.map.get(e)||null},e.prototype.keys=function(){return this.init(),Array.from(this.map.keys())},e.prototype.append=function(e,t){return this.clone({param:e,value:t,op:"a"})},e.prototype.set=function(e,t){return this.clone({param:e,value:t,op:"s"})},e.prototype.delete=function(e,t){return this.clone({param:e,value:t,op:"d"})},e.prototype.toString=function(){var e=this;return this.init(),this.keys().map(function(t){var n=e.encoder.encodeKey(t);return e.map.get(t).map(function(t){return n+"="+e.encoder.encodeValue(t)}).join("&")}).join("&")},e.prototype.clone=function(t){var n=new e({encoder:this.encoder});return n.cloneFrom=this.cloneFrom||this,n.updates=(this.updates||[]).concat([t]),n},e.prototype.init=function(){var e=this;null===this.map&&(this.map=new Map),null!==this.cloneFrom&&(this.cloneFrom.init(),this.cloneFrom.keys().forEach(function(t){return e.map.set(t,e.cloneFrom.map.get(t))}),this.updates.forEach(function(t){switch(t.op){case"a":case"s":var n=("a"===t.op?e.map.get(t.param):void 0)||[];n.push(t.value),e.map.set(t.param,n);break;case"d":if(void 0===t.value){e.map.delete(t.param);break}var r=e.map.get(t.param)||[],i=r.indexOf(t.value);-1!==i&&r.splice(i,1),r.length>0?e.map.set(t.param,r):e.map.delete(t.param)}}),this.cloneFrom=null)},e}();function h(e){return"undefined"!=typeof ArrayBuffer&&e instanceof ArrayBuffer}function m(e){return"undefined"!=typeof Blob&&e instanceof Blob}function g(e){return"undefined"!=typeof FormData&&e instanceof FormData}var v,_,y=function(){function e(e,t,n,r){var i;if(this.url=t,this.body=null,this.reportProgress=!1,this.withCredentials=!1,this.responseType="json",this.method=e.toUpperCase(),function(e){switch(e){case"DELETE":case"GET":case"HEAD":case"OPTIONS":case"JSONP":return!1;default:return!0}}(this.method)||r?(this.body=void 0!==n?n:null,i=r):i=n,i&&(this.reportProgress=!!i.reportProgress,this.withCredentials=!!i.withCredentials,i.responseType&&(this.responseType=i.responseType),i.headers&&(this.headers=i.headers),i.params&&(this.params=i.params)),this.headers||(this.headers=new l),this.params){var o=this.params.toString();if(0===o.length)this.urlWithParams=t;else{var a=t.indexOf("?"),s=-1===a?"?":a<t.length-1?"&":"";this.urlWithParams=t+s+o}}else this.params=new f,this.urlWithParams=t}return e.prototype.serializeBody=function(){return null===this.body?null:h(this.body)||m(this.body)||g(this.body)||"string"==typeof this.body?this.body:this.body instanceof f?this.body.toString():"object"==typeof this.body||"boolean"==typeof this.body||Array.isArray(this.body)?JSON.stringify(this.body):this.body.toString()},e.prototype.detectContentTypeHeader=function(){return null===this.body?null:g(this.body)?null:m(this.body)?this.body.type||null:h(this.body)?null:"string"==typeof this.body?"text/plain":this.body instanceof f?"application/x-www-form-urlencoded;charset=UTF-8":"object"==typeof this.body||"number"==typeof this.body||Array.isArray(this.body)?"application/json":null},e.prototype.clone=function(t){void 0===t&&(t={});var n=t.method||this.method,r=t.url||this.url,i=t.responseType||this.responseType,o=void 0!==t.body?t.body:this.body,a=void 0!==t.withCredentials?t.withCredentials:this.withCredentials,s=void 0!==t.reportProgress?t.reportProgress:this.reportProgress,u=t.headers||this.headers,c=t.params||this.params;return void 0!==t.setHeaders&&(u=Object.keys(t.setHeaders).reduce(function(e,n){return e.set(n,t.setHeaders[n])},u)),t.setParams&&(c=Object.keys(t.setParams).reduce(function(e,n){return e.set(n,t.setParams[n])},c)),new e(n,r,o,{params:c,headers:u,reportProgress:s,responseType:i,withCredentials:a})},e}();(_=v||(v={}))[_.Sent=0]="Sent",_[_.UploadProgress=1]="UploadProgress",_[_.ResponseHeader=2]="ResponseHeader",_[_.DownloadProgress=3]="DownloadProgress",_[_.Response=4]="Response",_[_.User=5]="User";var b=function(){return function(e,t,n){void 0===t&&(t=200),void 0===n&&(n="OK"),this.headers=e.headers||new l,this.status=void 0!==e.status?e.status:t,this.statusText=e.statusText||n,this.url=e.url||null,this.ok=this.status>=200&&this.status<300}}(),j=function(e){function t(t){void 0===t&&(t={});var n=e.call(this,t)||this;return n.type=v.ResponseHeader,n}return Object(r.__extends)(t,e),t.prototype.clone=function(e){return void 0===e&&(e={}),new t({headers:e.headers||this.headers,status:void 0!==e.status?e.status:this.status,statusText:e.statusText||this.statusText,url:e.url||this.url||void 0})},t}(b),x=function(e){function t(t){void 0===t&&(t={});var n=e.call(this,t)||this;return n.type=v.Response,n.body=void 0!==t.body?t.body:null,n}return Object(r.__extends)(t,e),t.prototype.clone=function(e){return void 0===e&&(e={}),new t({body:void 0!==e.body?e.body:this.body,headers:e.headers||this.headers,status:void 0!==e.status?e.status:this.status,statusText:e.statusText||this.statusText,url:e.url||this.url||void 0})},t}(b),O=function(e){function t(t){var n=e.call(this,t,0,"Unknown Error")||this;return n.name="HttpErrorResponse",n.ok=!1,n.status>=200&&n.status<300?n.message="Http failure during parsing for "+(t.url||"(unknown url)"):n.message="Http failure response for "+(t.url||"(unknown url)")+": "+t.status+" "+t.statusText,n.error=t.error||null,n}return Object(r.__extends)(t,e),t}(b);function w(e,t){return{body:t,headers:e.headers,observe:e.observe,params:e.params,reportProgress:e.reportProgress,responseType:e.responseType,withCredentials:e.withCredentials}}var C=function(){function e(e){this.handler=e}return e.prototype.request=function(e,t,n){var r,i=this;if(void 0===n&&(n={}),e instanceof y)r=e;else{var s=void 0;s=n.headers instanceof l?n.headers:new l(n.headers);var u=void 0;n.params&&(u=n.params instanceof f?n.params:new f({fromObject:n.params})),r=new y(e,t,void 0!==n.body?n.body:null,{headers:s,params:u,reportProgress:n.reportProgress,responseType:n.responseType||"json",withCredentials:n.withCredentials})}var c=Object(o.of)(r).pipe(Object(a.concatMap)(function(e){return i.handler.handle(e)}));if(e instanceof y||"events"===n.observe)return c;var d=c.pipe(Object(a.filter)(function(e){return e instanceof x}));switch(n.observe||"body"){case"body":switch(r.responseType){case"arraybuffer":return d.pipe(Object(a.map)(function(e){if(null!==e.body&&!(e.body instanceof ArrayBuffer))throw new Error("Response is not an ArrayBuffer.");return e.body}));case"blob":return d.pipe(Object(a.map)(function(e){if(null!==e.body&&!(e.body instanceof Blob))throw new Error("Response is not a Blob.");return e.body}));case"text":return d.pipe(Object(a.map)(function(e){if(null!==e.body&&"string"!=typeof e.body)throw new Error("Response is not a string.");return e.body}));case"json":default:return d.pipe(Object(a.map)(function(e){return e.body}))}case"response":return d;default:throw new Error("Unreachable: unhandled observe type "+n.observe+"}")}},e.prototype.delete=function(e,t){return void 0===t&&(t={}),this.request("DELETE",e,t)},e.prototype.get=function(e,t){return void 0===t&&(t={}),this.request("GET",e,t)},e.prototype.head=function(e,t){return void 0===t&&(t={}),this.request("HEAD",e,t)},e.prototype.jsonp=function(e,t){return this.request("JSONP",e,{params:(new f).append(t,"JSONP_CALLBACK"),observe:"body",responseType:"json"})},e.prototype.options=function(e,t){return void 0===t&&(t={}),this.request("OPTIONS",e,t)},e.prototype.patch=function(e,t,n){return void 0===n&&(n={}),this.request("PATCH",e,w(n,t))},e.prototype.post=function(e,t,n){return void 0===n&&(n={}),this.request("POST",e,w(n,t))},e.prototype.put=function(e,t,n){return void 0===n&&(n={}),this.request("PUT",e,w(n,t))},e=Object(r.__decorate)([Object(i.Injectable)(),Object(r.__metadata)("design:paramtypes",[u])],e)}(),E=function(){function e(e,t){this.next=e,this.interceptor=t}return e.prototype.handle=function(e){return this.interceptor.intercept(e,this.next)},e}(),S=new i.InjectionToken("HTTP_INTERCEPTORS"),k=function(){function e(){}return e.prototype.intercept=function(e,t){return t.handle(e)},e=Object(r.__decorate)([Object(i.Injectable)()],e)}(),M=0,A=function(){return function(){}}(),T=function(){function e(e,t){this.callbackMap=e,this.document=t}return e.prototype.nextCallback=function(){return"ng_jsonp_callback_"+M++},e.prototype.handle=function(e){var t=this;if("JSONP"!==e.method)throw new Error("JSONP requests must use JSONP request method.");if("json"!==e.responseType)throw new Error("JSONP requests must use Json response type.");return new o.Observable(function(n){var r=t.nextCallback(),i=e.urlWithParams.replace(/=JSONP_CALLBACK(&|$)/,"="+r+"$1"),o=t.document.createElement("script");o.src=i;var a=null,s=!1,u=!1;t.callbackMap[r]=function(e){delete t.callbackMap[r],u||(a=e,s=!0)};var c=function(){o.parentNode&&o.parentNode.removeChild(o),delete t.callbackMap[r]},l=function(e){u||(c(),s?(n.next(new x({body:a,status:200,statusText:"OK",url:i})),n.complete()):n.error(new O({url:i,status:0,statusText:"JSONP Error",error:new Error("JSONP injected script did not invoke callback.")})))},d=function(e){u||(c(),n.error(new O({error:e,status:0,statusText:"JSONP Error",url:i})))};return o.addEventListener("load",l),o.addEventListener("error",d),t.document.body.appendChild(o),n.next({type:v.Sent}),function(){u=!0,o.removeEventListener("load",l),o.removeEventListener("error",d),c()}})},e=Object(r.__decorate)([Object(i.Injectable)(),Object(r.__param)(1,Object(i.Inject)(s.DOCUMENT)),Object(r.__metadata)("design:paramtypes",[A,Object])],e)}(),D=function(){function e(e){this.jsonp=e}return e.prototype.intercept=function(e,t){return"JSONP"===e.method?this.jsonp.handle(e):t.handle(e)},e=Object(r.__decorate)([Object(i.Injectable)(),Object(r.__metadata)("design:paramtypes",[T])],e)}(),I=/^\)\]\}',?\n/;var P=function(){return function(){}}(),R=function(){function e(){}return e.prototype.build=function(){return new XMLHttpRequest},e=Object(r.__decorate)([Object(i.Injectable)(),Object(r.__metadata)("design:paramtypes",[])],e)}(),N=function(){function e(e){this.xhrFactory=e}return e.prototype.handle=function(e){var t=this;if("JSONP"===e.method)throw new Error("Attempted to construct Jsonp request without JsonpClientModule installed.");return new o.Observable(function(n){var r=t.xhrFactory.build();if(r.open(e.method,e.urlWithParams),e.withCredentials&&(r.withCredentials=!0),e.headers.forEach(function(e,t){return r.setRequestHeader(e,t.join(","))}),e.headers.has("Accept")||r.setRequestHeader("Accept","application/json, text/plain, */*"),!e.headers.has("Content-Type")){var i=e.detectContentTypeHeader();null!==i&&r.setRequestHeader("Content-Type",i)}if(e.responseType){var o=e.responseType.toLowerCase();r.responseType="json"!==o?o:"text"}var a=e.serializeBody(),s=null,u=function(){if(null!==s)return s;var t,n=1223===r.status?204:r.status,i=r.statusText||"OK",o=new l(r.getAllResponseHeaders()),a=("responseURL"in(t=r)&&t.responseURL?t.responseURL:/^X-Request-URL:/m.test(t.getAllResponseHeaders())?t.getResponseHeader("X-Request-URL"):null)||e.url;return s=new j({headers:o,status:n,statusText:i,url:a})},c=function(){var t=u(),i=t.headers,o=t.status,a=t.statusText,s=t.url,c=null;204!==o&&(c=void 0===r.response?r.responseText:r.response),0===o&&(o=c?200:0);var l=o>=200&&o<300;if("json"===e.responseType&&"string"==typeof c){var d=c;c=c.replace(I,"");try{c=""!==c?JSON.parse(c):null}catch(e){c=d,l&&(l=!1,c={error:e,text:c})}}l?(n.next(new x({body:c,headers:i,status:o,statusText:a,url:s||void 0})),n.complete()):n.error(new O({error:c,headers:i,status:o,statusText:a,url:s||void 0}))},d=function(e){var t=u().url,i=new O({error:e,status:r.status||0,statusText:r.statusText||"Unknown Error",url:t||void 0});n.error(i)},p=!1,f=function(t){p||(n.next(u()),p=!0);var i={type:v.DownloadProgress,loaded:t.loaded};t.lengthComputable&&(i.total=t.total),"text"===e.responseType&&r.responseText&&(i.partialText=r.responseText),n.next(i)},h=function(e){var t={type:v.UploadProgress,loaded:e.loaded};e.lengthComputable&&(t.total=e.total),n.next(t)};return r.addEventListener("load",c),r.addEventListener("error",d),e.reportProgress&&(r.addEventListener("progress",f),null!==a&&r.upload&&r.upload.addEventListener("progress",h)),r.send(a),n.next({type:v.Sent}),function(){r.removeEventListener("error",d),r.removeEventListener("load",c),e.reportProgress&&(r.removeEventListener("progress",f),null!==a&&r.upload&&r.upload.removeEventListener("progress",h)),r.abort()}})},e=Object(r.__decorate)([Object(i.Injectable)(),Object(r.__metadata)("design:paramtypes",[P])],e)}(),F=new i.InjectionToken("XSRF_COOKIE_NAME"),L=new i.InjectionToken("XSRF_HEADER_NAME"),V=function(){return function(){}}(),B=function(){function e(e,t,n){this.doc=e,this.platform=t,this.cookieName=n,this.lastCookieString="",this.lastToken=null,this.parseCount=0}return e.prototype.getToken=function(){if("server"===this.platform)return null;var e=this.doc.cookie||"";return e!==this.lastCookieString&&(this.parseCount++,this.lastToken=Object(s["ɵparseCookieValue"])(e,this.cookieName),this.lastCookieString=e),this.lastToken},e=Object(r.__decorate)([Object(i.Injectable)(),Object(r.__param)(0,Object(i.Inject)(s.DOCUMENT)),Object(r.__param)(1,Object(i.Inject)(i.PLATFORM_ID)),Object(r.__param)(2,Object(i.Inject)(F)),Object(r.__metadata)("design:paramtypes",[Object,String,String])],e)}(),z=function(){function e(e,t){this.tokenService=e,this.headerName=t}return e.prototype.intercept=function(e,t){var n=e.url.toLowerCase();if("GET"===e.method||"HEAD"===e.method||n.startsWith("http://")||n.startsWith("https://"))return t.handle(e);var r=this.tokenService.getToken();return null===r||e.headers.has(this.headerName)||(e=e.clone({headers:e.headers.set(this.headerName,r)})),t.handle(e)},e=Object(r.__decorate)([Object(i.Injectable)(),Object(r.__param)(1,Object(i.Inject)(L)),Object(r.__metadata)("design:paramtypes",[V,String])],e)}(),U=function(){function e(e,t){this.backend=e,this.injector=t,this.chain=null}return e.prototype.handle=function(e){if(null===this.chain){var t=this.injector.get(S,[]);this.chain=t.reduceRight(function(e,t){return new E(e,t)},this.backend)}return this.chain.handle(e)},e=Object(r.__decorate)([Object(i.Injectable)(),Object(r.__metadata)("design:paramtypes",[c,i.Injector])],e)}();function H(){return"object"==typeof window?window:{}}var q=function(){function e(){}var t;return t=e,e.disable=function(){return{ngModule:t,providers:[{provide:z,useClass:k}]}},e.withOptions=function(e){return void 0===e&&(e={}),{ngModule:t,providers:[e.cookieName?{provide:F,useValue:e.cookieName}:[],e.headerName?{provide:L,useValue:e.headerName}:[]]}},e=t=Object(r.__decorate)([Object(i.NgModule)({providers:[z,{provide:S,useExisting:z,multi:!0},{provide:V,useClass:B},{provide:F,useValue:"XSRF-TOKEN"},{provide:L,useValue:"X-XSRF-TOKEN"}]})],e)}(),G=function(){function e(){}return e=Object(r.__decorate)([Object(i.NgModule)({imports:[q.withOptions({cookieName:"XSRF-TOKEN",headerName:"X-XSRF-TOKEN"})],providers:[C,{provide:u,useClass:U},N,{provide:c,useExisting:N},R,{provide:P,useExisting:R}]})],e)}(),W=function(){function e(){}return e=Object(r.__decorate)([Object(i.NgModule)({providers:[T,{provide:A,useFactory:H},{provide:S,useClass:D,multi:!0}]})],e)}()},"./node_modules/@angular/common/node_modules/tslib/tslib.es6.js":function(e,t,n){"use strict";n.r(t),n.d(t,"__extends",function(){return i}),n.d(t,"__assign",function(){return o}),n.d(t,"__rest",function(){return a}),n.d(t,"__decorate",function(){return s}),n.d(t,"__param",function(){return u}),n.d(t,"__metadata",function(){return c}),n.d(t,"__awaiter",function(){return l}),n.d(t,"__generator",function(){return d}),n.d(t,"__exportStar",function(){return p}),n.d(t,"__values",function(){return f}),n.d(t,"__read",function(){return h}),n.d(t,"__spread",function(){return m}),n.d(t,"__await",function(){return g}),n.d(t,"__asyncGenerator",function(){return v}),n.d(t,"__asyncDelegator",function(){return _}),n.d(t,"__asyncValues",function(){return y}),n.d(t,"__makeTemplateObject",function(){return b}),n.d(t,"__importStar",function(){return j}),n.d(t,"__importDefault",function(){return x});var r=function(e,t){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,t){e.__proto__=t}||function(e,t){for(var n in t)t.hasOwnProperty(n)&&(e[n]=t[n])})(e,t)};function i(e,t){function n(){this.constructor=e}r(e,t),e.prototype=null===t?Object.create(t):(n.prototype=t.prototype,new n)}var o=function(){return(o=Object.assign||function(e){for(var t,n=1,r=arguments.length;n<r;n++){t=arguments[n];for(var i in t)Object.prototype.hasOwnProperty.call(t,i)&&(e[i]=t[i])}return e}).apply(this,arguments)};function a(e,t){var n={};for(var r in e)Object.prototype.hasOwnProperty.call(e,r)&&t.indexOf(r)<0&&(n[r]=e[r]);if(null!=e&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(e);i<r.length;i++)t.indexOf(r[i])<0&&(n[r[i]]=e[r[i]])}return n}function s(e,t,n,r){var i,o=arguments.length,a=o<3?t:null===r?r=Object.getOwnPropertyDescriptor(t,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(e,t,n,r);else for(var s=e.length-1;s>=0;s--)(i=e[s])&&(a=(o<3?i(a):o>3?i(t,n,a):i(t,n))||a);return o>3&&a&&Object.defineProperty(t,n,a),a}function u(e,t){return function(n,r){t(n,r,e)}}function c(e,t){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(e,t)}function l(e,t,n,r){return new(n||(n=Promise))(function(i,o){function a(e){try{u(r.next(e))}catch(e){o(e)}}function s(e){try{u(r.throw(e))}catch(e){o(e)}}function u(e){e.done?i(e.value):new n(function(t){t(e.value)}).then(a,s)}u((r=r.apply(e,t||[])).next())})}function d(e,t){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=t.call(e,a)}catch(e){o=[6,e],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(e,t){for(var n in e)t.hasOwnProperty(n)||(t[n]=e[n])}function f(e){var t="function"==typeof Symbol&&e[Symbol.iterator],n=0;return t?t.call(e):{next:function(){return e&&n>=e.length&&(e=void 0),{value:e&&e[n++],done:!e}}}}function h(e,t){var n="function"==typeof Symbol&&e[Symbol.iterator];if(!n)return e;var r,i,o=n.call(e),a=[];try{for(;(void 0===t||t-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(e){i={error:e}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var e=[],t=0;t<arguments.length;t++)e=e.concat(h(arguments[t]));return e}function g(e){return this instanceof g?(this.v=e,this):new g(e)}function v(e,t,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(e,t||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(e){i[e]&&(r[e]=function(t){return new Promise(function(n,r){o.push([e,t,n,r])>1||s(e,t)})})}function s(e,t){try{(n=i[e](t)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(e){l(o[0][3],e)}var n}function u(e){s("next",e)}function c(e){s("throw",e)}function l(e,t){e(t),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(e){var t,n;return t={},r("next"),r("throw",function(e){throw e}),r("return"),t[Symbol.iterator]=function(){return this},t;function r(r,i){t[r]=e[r]?function(t){return(n=!n)?{value:g(e[r](t)),done:"return"===r}:i?i(t):t}:i}}function y(e){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var t,n=e[Symbol.asyncIterator];return n?n.call(e):(e=f(e),t={},r("next"),r("throw"),r("return"),t[Symbol.asyncIterator]=function(){return this},t);function r(n){t[n]=e[n]&&function(t){return new Promise(function(r,i){var o,a,s,u;t=e[n](t),o=r,a=i,s=t.done,u=t.value,Promise.resolve(u).then(function(e){o({value:e,done:s})},a)})}}}function b(e,t){return Object.defineProperty?Object.defineProperty(e,"raw",{value:t}):e.raw=t,e}function j(e){if(e&&e.__esModule)return e;var t={};if(null!=e)for(var n in e)Object.hasOwnProperty.call(e,n)&&(t[n]=e[n]);return t.default=e,t}function x(e){return e&&e.__esModule?e:{default:e}}},"./node_modules/@angular/compiler/fesm5/compiler.js":function(e,t,n){"use strict";n.r(t),n.d(t,"core",function(){return oe}),n.d(t,"CompilerConfig",function(){return Sd}),n.d(t,"preserveWhitespacesDefault",function(){return kd}),n.d(t,"isLoweredSymbol",function(){return Cp}),n.d(t,"createLoweredSymbol",function(){return Ep}),n.d(t,"Identifiers",function(){return gn}),n.d(t,"JitCompiler",function(){return rm}),n.d(t,"ConstantPool",function(){return dn}),n.d(t,"DirectiveResolver",function(){return Dd}),n.d(t,"PipeResolver",function(){return ef}),n.d(t,"NgModuleResolver",function(){return Xp}),n.d(t,"DEFAULT_INTERPOLATION_CONFIG",function(){return Ui}),n.d(t,"InterpolationConfig",function(){return zi}),n.d(t,"NgModuleCompiler",function(){return Qp}),n.d(t,"ArrayType",function(){return fe}),n.d(t,"AssertNotNull",function(){return He}),n.d(t,"BinaryOperator",function(){return ce}),n.d(t,"BinaryOperatorExpr",function(){return Ye}),n.d(t,"BuiltinMethod",function(){return Ce}),n.d(t,"BuiltinType",function(){return de}),n.d(t,"BuiltinTypeName",function(){return ae}),n.d(t,"BuiltinVar",function(){return xe}),n.d(t,"CastExpr",function(){return qe}),n.d(t,"ClassField",function(){return ut}),n.d(t,"ClassMethod",function(){return ct}),n.d(t,"ClassStmt",function(){return dt}),n.d(t,"CommaExpr",function(){return Je}),n.d(t,"CommentStmt",function(){return ft}),n.d(t,"ConditionalExpr",function(){return ze}),n.d(t,"DeclareFunctionStmt",function(){return it}),n.d(t,"DeclareVarStmt",function(){return rt}),n.d(t,"Expression",function(){return we}),n.d(t,"ExpressionStatement",function(){return ot}),n.d(t,"ExpressionType",function(){return pe}),n.d(t,"ExternalExpr",function(){return Ve}),n.d(t,"ExternalReference",function(){return Be}),n.d(t,"FunctionExpr",function(){return We}),n.d(t,"IfStmt",function(){return pt}),n.d(t,"InstantiateExpr",function(){return Fe}),n.d(t,"InvokeFunctionExpr",function(){return Ne}),n.d(t,"InvokeMethodExpr",function(){return Re}),n.d(t,"JSDocCommentStmt",function(){return ht}),n.d(t,"LiteralArrayExpr",function(){return Xe}),n.d(t,"LiteralExpr",function(){return Le}),n.d(t,"LiteralMapExpr",function(){return $e}),n.d(t,"MapType",function(){return he}),n.d(t,"NotExpr",function(){return Ue}),n.d(t,"ReadKeyExpr",function(){return Qe}),n.d(t,"ReadPropExpr",function(){return Ke}),n.d(t,"ReadVarExpr",function(){return Se}),n.d(t,"ReturnStatement",function(){return at}),n.d(t,"ThrowStmt",function(){return gt}),n.d(t,"TryCatchStmt",function(){return mt}),n.d(t,"Type",function(){return ue}),n.d(t,"WrappedNodeExpr",function(){return Me}),n.d(t,"WriteKeyExpr",function(){return Te}),n.d(t,"WritePropExpr",function(){return De}),n.d(t,"WriteVarExpr",function(){return Ae}),n.d(t,"StmtModifier",function(){return Ie}),n.d(t,"Statement",function(){return nt}),n.d(t,"TypeofExpr",function(){return ke}),n.d(t,"collectExternalReferences",function(){return jt}),n.d(t,"EmitterVisitorContext",function(){return Ji}),n.d(t,"ViewCompiler",function(){return Tf}),n.d(t,"getParseErrors",function(){return Zt}),n.d(t,"isSyntaxError",function(){return Xt}),n.d(t,"syntaxError",function(){return Yt}),n.d(t,"Version",function(){return on}),n.d(t,"TmplAstBoundAttribute",function(){return mc}),n.d(t,"TmplAstBoundEvent",function(){return gc}),n.d(t,"TmplAstBoundText",function(){return fc}),n.d(t,"TmplAstContent",function(){return yc}),n.d(t,"TmplAstElement",function(){return vc}),n.d(t,"TmplAstReference",function(){return jc}),n.d(t,"TmplAstTemplate",function(){return _c}),n.d(t,"TmplAstText",function(){return pc}),n.d(t,"TmplAstTextAttribute",function(){return hc}),n.d(t,"TmplAstVariable",function(){return bc}),n.d(t,"jitExpression",function(){return ao}),n.d(t,"R3ResolvedDependencyType",function(){return xi}),n.d(t,"compileInjector",function(){return uo}),n.d(t,"compileNgModule",function(){return so}),n.d(t,"compilePipeFromMetadata",function(){return lo}),n.d(t,"makeBindingParser",function(){return wl}),n.d(t,"parseTemplate",function(){return Ol}),n.d(t,"compileBaseDefFromMetadata",function(){return Il}),n.d(t,"compileComponentFromMetadata",function(){return Pl}),n.d(t,"compileDirectiveFromMetadata",function(){return Dl}),n.d(t,"parseHostBindings",function(){return Gl}),n.d(t,"publishFacade",function(){return od}),n.d(t,"VERSION",function(){return cd}),n.d(t,"TextAst",function(){return ld}),n.d(t,"BoundTextAst",function(){return dd}),n.d(t,"AttrAst",function(){return pd}),n.d(t,"BoundElementPropertyAst",function(){return hd}),n.d(t,"BoundEventAst",function(){return md}),n.d(t,"ReferenceAst",function(){return gd}),n.d(t,"VariableAst",function(){return vd}),n.d(t,"ElementAst",function(){return _d}),n.d(t,"EmbeddedTemplateAst",function(){return yd}),n.d(t,"BoundDirectivePropertyAst",function(){return bd}),n.d(t,"DirectiveAst",function(){return jd}),n.d(t,"ProviderAst",function(){return xd}),n.d(t,"ProviderAstType",function(){return sd}),n.d(t,"NgContentAst",function(){return Od}),n.d(t,"NullTemplateVisitor",function(){return wd}),n.d(t,"RecursiveTemplateAstVisitor",function(){return Cd}),n.d(t,"templateVisitAll",function(){return Ed}),n.d(t,"sanitizeIdentifier",function(){return xn}),n.d(t,"identifierName",function(){return En}),n.d(t,"identifierModuleUrl",function(){return Sn}),n.d(t,"viewClassName",function(){return kn}),n.d(t,"rendererTypeName",function(){return Mn}),n.d(t,"hostViewClassName",function(){return An}),n.d(t,"componentFactoryName",function(){return Tn}),n.d(t,"CompileSummaryKind",function(){return On}),n.d(t,"tokenName",function(){return Dn}),n.d(t,"tokenReference",function(){return In}),n.d(t,"CompileStylesheetMetadata",function(){return Pn}),n.d(t,"CompileTemplateMetadata",function(){return Rn}),n.d(t,"CompileDirectiveMetadata",function(){return Nn}),n.d(t,"CompilePipeMetadata",function(){return Fn}),n.d(t,"CompileShallowModuleMetadata",function(){return Ln}),n.d(t,"CompileNgModuleMetadata",function(){return Vn}),n.d(t,"TransitiveCompileNgModuleMetadata",function(){return Bn}),n.d(t,"ProviderMeta",function(){return Un}),n.d(t,"flatten",function(){return Hn}),n.d(t,"templateSourceUrl",function(){return Gn}),n.d(t,"sharedStylesheetJitUrl",function(){return Wn}),n.d(t,"ngModuleJitUrl",function(){return Yn}),n.d(t,"templateJitUrl",function(){return Kn}),n.d(t,"createAotUrlResolver",function(){return Wh}),n.d(t,"createAotCompiler",function(){return Yh}),n.d(t,"AotCompiler",function(){return sh}),n.d(t,"analyzeNgModules",function(){return dh}),n.d(t,"analyzeAndValidateNgModules",function(){return ph}),n.d(t,"analyzeFile",function(){return hh}),n.d(t,"analyzeFileForInjectables",function(){return mh}),n.d(t,"mergeAnalyzedFiles",function(){return gh}),n.d(t,"GeneratedFile",function(){return Wf}),n.d(t,"toTypeScript",function(){return Yf}),n.d(t,"formattedError",function(){return bh}),n.d(t,"isFormattedError",function(){return jh}),n.d(t,"StaticReflector",function(){return Mh}),n.d(t,"StaticSymbol",function(){return yn}),n.d(t,"StaticSymbolCache",function(){return bn}),n.d(t,"ResolvedStaticSymbol",function(){return $f}),n.d(t,"StaticSymbolResolver",function(){return Jf}),n.d(t,"unescapeIdentifier",function(){return eh}),n.d(t,"unwrapResolvedMetadata",function(){return th}),n.d(t,"AotSummaryResolver",function(){return Gh}),n.d(t,"AstPath",function(){return du}),n.d(t,"SummaryResolver",function(){return Kh}),n.d(t,"JitSummaryResolver",function(){return Qh}),n.d(t,"CompileReflector",function(){return sm}),n.d(t,"createUrlResolverWithoutPackagePrefix",function(){return um}),n.d(t,"createOfflineCompileUrlResolver",function(){return cm}),n.d(t,"UrlResolver",function(){return lm}),n.d(t,"getUrlScheme",function(){return dm}),n.d(t,"ResourceLoader",function(){return vm}),n.d(t,"ElementSchemaRegistry",function(){return ec}),n.d(t,"Extractor",function(){return _m}),n.d(t,"I18NHtmlParser",function(){return up}),n.d(t,"MessageBundle",function(){return qf}),n.d(t,"Serializer",function(){return Er}),n.d(t,"Xliff",function(){return Qd}),n.d(t,"Xliff2",function(){return Jd}),n.d(t,"Xmb",function(){return Fr}),n.d(t,"Xtb",function(){return rp}),n.d(t,"DirectiveNormalizer",function(){return Md}),n.d(t,"ParserError",function(){return po}),n.d(t,"ParseSpan",function(){return fo}),n.d(t,"AST",function(){return ho}),n.d(t,"Quote",function(){return mo}),n.d(t,"EmptyExpr",function(){return go}),n.d(t,"ImplicitReceiver",function(){return vo}),n.d(t,"Chain",function(){return _o}),n.d(t,"Conditional",function(){return yo}),n.d(t,"PropertyRead",function(){return bo}),n.d(t,"PropertyWrite",function(){return jo}),n.d(t,"SafePropertyRead",function(){return xo}),n.d(t,"KeyedRead",function(){return Oo}),n.d(t,"KeyedWrite",function(){return wo}),n.d(t,"BindingPipe",function(){return Co}),n.d(t,"LiteralPrimitive",function(){return Eo}),n.d(t,"LiteralArray",function(){return So}),n.d(t,"LiteralMap",function(){return ko}),n.d(t,"Interpolation",function(){return Mo}),n.d(t,"Binary",function(){return Ao}),n.d(t,"PrefixNot",function(){return To}),n.d(t,"NonNullAssert",function(){return Do}),n.d(t,"MethodCall",function(){return Io}),n.d(t,"SafeMethodCall",function(){return Po}),n.d(t,"FunctionCall",function(){return Ro}),n.d(t,"ASTWithSource",function(){return No}),n.d(t,"TemplateBinding",function(){return Fo}),n.d(t,"NullAstVisitor",function(){return Lo}),n.d(t,"RecursiveAstVisitor",function(){return Vo}),n.d(t,"AstTransformer",function(){return Bo}),n.d(t,"AstMemoryEfficientTransformer",function(){return zo}),n.d(t,"visitAstChildren",function(){return Uo}),n.d(t,"ParsedProperty",function(){return Go}),n.d(t,"ParsedPropertyType",function(){return Ho}),n.d(t,"ParsedEvent",function(){return Wo}),n.d(t,"ParsedVariable",function(){return Yo}),n.d(t,"BoundElementProperty",function(){return Ko}),n.d(t,"TokenType",function(){return zs}),n.d(t,"Lexer",function(){return Ks}),n.d(t,"Token",function(){return Qs}),n.d(t,"EOF",function(){return $s}),n.d(t,"isIdentifier",function(){return tu}),n.d(t,"isQuote",function(){return ru}),n.d(t,"SplitInterpolation",function(){return ou}),n.d(t,"TemplateBindingParseResult",function(){return au}),n.d(t,"Parser",function(){return uu}),n.d(t,"_ParseAST",function(){return cu}),n.d(t,"ERROR_COMPONENT_TYPE",function(){return Sp}),n.d(t,"CompileMetadataResolver",function(){return kp}),n.d(t,"Text",function(){return pu}),n.d(t,"Expansion",function(){return fu}),n.d(t,"ExpansionCase",function(){return hu}),n.d(t,"Attribute",function(){return mu}),n.d(t,"Element",function(){return gu}),n.d(t,"Comment",function(){return vu}),n.d(t,"visitAll",function(){return _u}),n.d(t,"RecursiveVisitor",function(){return ju}),n.d(t,"findNode",function(){return xu}),n.d(t,"HtmlParser",function(){return Uu}),n.d(t,"ParseTreeResult",function(){return Lu}),n.d(t,"TreeError",function(){return Fu}),n.d(t,"HtmlTagDefinition",function(){return g}),n.d(t,"getHtmlTagDefinition",function(){return v}),n.d(t,"TagContentType",function(){return r}),n.d(t,"splitNsName",function(){return a}),n.d(t,"isNgContainer",function(){return s}),n.d(t,"isNgContent",function(){return u}),n.d(t,"isNgTemplate",function(){return c}),n.d(t,"getNsPrefix",function(){return l}),n.d(t,"mergeNsAndName",function(){return d}),n.d(t,"NAMED_ENTITIES",function(){return p}),n.d(t,"NGSP_UNICODE",function(){return f}),n.d(t,"debugOutputAstAsTypeScript",function(){return Zp}),n.d(t,"TypeScriptEmitter",function(){return $p}),n.d(t,"ParseLocation",function(){return ns}),n.d(t,"ParseSourceFile",function(){return rs}),n.d(t,"ParseSourceSpan",function(){return is}),n.d(t,"ParseErrorLevel",function(){return es}),n.d(t,"ParseError",function(){return os}),n.d(t,"typeSourceSpan",function(){return as}),n.d(t,"DomElementSchemaRegistry",function(){return sc}),n.d(t,"CssSelector",function(){return O}),n.d(t,"SelectorMatcher",function(){return w}),n.d(t,"SelectorListContext",function(){return C}),n.d(t,"SelectorContext",function(){return E}),n.d(t,"HOST_ATTR",function(){return Ps}),n.d(t,"CONTENT_ATTR",function(){return Rs}),n.d(t,"StylesCompileDependency",function(){return Ns}),n.d(t,"CompiledStylesheet",function(){return Fs}),n.d(t,"StyleCompiler",function(){return Ls}),n.d(t,"TemplateParseError",function(){return df}),n.d(t,"TemplateParseResult",function(){return pf}),n.d(t,"TemplateParser",function(){return ff}),n.d(t,"splitClasses",function(){return vf}),n.d(t,"createElementCssSelector",function(){return yf}),n.d(t,"removeSummaryDuplicates",function(){return Of}),n.d(t,"compileInjectable",function(){return Fi}),n.d(t,"R3TargetBinder",function(){return ym}),n.d(t,"R3BoundTarget",function(){return Om});var r,i,o=n("./node_modules/@angular/compiler/node_modules/tslib/tslib.es6.js");function a(e){if(":"!=e[0])return[null,e];var t=e.indexOf(":",1);if(-1==t)throw new Error('Unsupported format "'+e+'" expecting ":namespace:name"');return[e.slice(1,t),e.slice(t+1)]}function s(e){return"ng-container"===a(e)[1]}function u(e){return"ng-content"===a(e)[1]}function c(e){return"ng-template"===a(e)[1]}function l(e){return null===e?null:a(e)[0]}function d(e,t){return e?":"+e+":"+t:t}(i=r||(r={}))[i.RAW_TEXT=0]="RAW_TEXT",i[i.ESCAPABLE_RAW_TEXT=1]="ESCAPABLE_RAW_TEXT",i[i.PARSABLE_DATA=2]="PARSABLE_DATA";var p={Aacute:"Á",aacute:"á",Acirc:"Â",acirc:"â",acute:"´",AElig:"Æ",aelig:"æ",Agrave:"À",agrave:"à",alefsym:"ℵ",Alpha:"Α",alpha:"α",amp:"&",and:"∧",ang:"∠",apos:"'",Aring:"Å",aring:"å",asymp:"≈",Atilde:"Ã",atilde:"ã",Auml:"Ä",auml:"ä",bdquo:"„",Beta:"Β",beta:"β",brvbar:"¦",bull:"•",cap:"∩",Ccedil:"Ç",ccedil:"ç",cedil:"¸",cent:"¢",Chi:"Χ",chi:"χ",circ:"ˆ",clubs:"♣",cong:"≅",copy:"©",crarr:"↵",cup:"∪",curren:"¤",dagger:"†",Dagger:"‡",darr:"↓",dArr:"⇓",deg:"°",Delta:"Δ",delta:"δ",diams:"♦",divide:"÷",Eacute:"É",eacute:"é",Ecirc:"Ê",ecirc:"ê",Egrave:"È",egrave:"è",empty:"∅",emsp:" ",ensp:" ",Epsilon:"Ε",epsilon:"ε",equiv:"≡",Eta:"Η",eta:"η",ETH:"Ð",eth:"ð",Euml:"Ë",euml:"ë",euro:"€",exist:"∃",fnof:"ƒ",forall:"∀",frac12:"½",frac14:"¼",frac34:"¾",frasl:"⁄",Gamma:"Γ",gamma:"γ",ge:"≥",gt:">",harr:"↔",hArr:"⇔",hearts:"♥",hellip:"…",Iacute:"Í",iacute:"í",Icirc:"Î",icirc:"î",iexcl:"¡",Igrave:"Ì",igrave:"ì",image:"ℑ",infin:"∞",int:"∫",Iota:"Ι",iota:"ι",iquest:"¿",isin:"∈",Iuml:"Ï",iuml:"ï",Kappa:"Κ",kappa:"κ",Lambda:"Λ",lambda:"λ",lang:"⟨",laquo:"«",larr:"←",lArr:"⇐",lceil:"⌈",ldquo:"“",le:"≤",lfloor:"⌊",lowast:"∗",loz:"◊",lrm:"‎",lsaquo:"‹",lsquo:"‘",lt:"<",macr:"¯",mdash:"—",micro:"µ",middot:"·",minus:"−",Mu:"Μ",mu:"μ",nabla:"∇",nbsp:" ",ndash:"–",ne:"≠",ni:"∋",not:"¬",notin:"∉",nsub:"⊄",Ntilde:"Ñ",ntilde:"ñ",Nu:"Ν",nu:"ν",Oacute:"Ó",oacute:"ó",Ocirc:"Ô",ocirc:"ô",OElig:"Œ",oelig:"œ",Ograve:"Ò",ograve:"ò",oline:"‾",Omega:"Ω",omega:"ω",Omicron:"Ο",omicron:"ο",oplus:"⊕",or:"∨",ordf:"ª",ordm:"º",Oslash:"Ø",oslash:"ø",Otilde:"Õ",otilde:"õ",otimes:"⊗",Ouml:"Ö",ouml:"ö",para:"¶",permil:"‰",perp:"⊥",Phi:"Φ",phi:"φ",Pi:"Π",pi:"π",piv:"ϖ",plusmn:"±",pound:"£",prime:"′",Prime:"″",prod:"∏",prop:"∝",Psi:"Ψ",psi:"ψ",quot:'"',radic:"√",rang:"⟩",raquo:"»",rarr:"→",rArr:"⇒",rceil:"⌉",rdquo:"”",real:"ℜ",reg:"®",rfloor:"⌋",Rho:"Ρ",rho:"ρ",rlm:"‏",rsaquo:"›",rsquo:"’",sbquo:"‚",Scaron:"Š",scaron:"š",sdot:"⋅",sect:"§",shy:"­",Sigma:"Σ",sigma:"σ",sigmaf:"ς",sim:"∼",spades:"♠",sub:"⊂",sube:"⊆",sum:"∑",sup:"⊃",sup1:"¹",sup2:"²",sup3:"³",supe:"⊇",szlig:"ß",Tau:"Τ",tau:"τ",there4:"∴",Theta:"Θ",theta:"θ",thetasym:"ϑ",thinsp:" ",THORN:"Þ",thorn:"þ",tilde:"˜",times:"×",trade:"™",Uacute:"Ú",uacute:"ú",uarr:"↑",uArr:"⇑",Ucirc:"Û",ucirc:"û",Ugrave:"Ù",ugrave:"ù",uml:"¨",upsih:"ϒ",Upsilon:"Υ",upsilon:"υ",Uuml:"Ü",uuml:"ü",weierp:"℘",Xi:"Ξ",xi:"ξ",Yacute:"Ý",yacute:"ý",yen:"¥",yuml:"ÿ",Yuml:"Ÿ",Zeta:"Ζ",zeta:"ζ",zwj:"‍",zwnj:"‌"},f="";p.ngsp=f;var h,m,g=function(){function e(e){var t=void 0===e?{}:e,n=t.closedByChildren,i=t.requiredParents,o=t.implicitNamespacePrefix,a=t.contentType,s=void 0===a?r.PARSABLE_DATA:a,u=t.closedByParent,c=void 0!==u&&u,l=t.isVoid,d=void 0!==l&&l,p=t.ignoreFirstLf,f=void 0!==p&&p,h=this;this.closedByChildren={},this.closedByParent=!1,this.canSelfClose=!1,n&&n.length>0&&n.forEach(function(e){return h.closedByChildren[e]=!0}),this.isVoid=d,this.closedByParent=c||d,i&&i.length>0&&(this.requiredParents={},this.parentToAdd=i[0],i.forEach(function(e){return h.requiredParents[e]=!0})),this.implicitNamespacePrefix=o||null,this.contentType=s,this.ignoreFirstLf=f}return e.prototype.requireExtraParent=function(e){if(!this.requiredParents)return!1;if(!e)return!0;var t=e.toLowerCase();return!("template"===t||"ng-template"===e)&&1!=this.requiredParents[t]},e.prototype.isClosedByChild=function(e){return this.isVoid||e.toLowerCase()in this.closedByChildren},e}();function v(e){return m||(h=new g,m={base:new g({isVoid:!0}),meta:new g({isVoid:!0}),area:new g({isVoid:!0}),embed:new g({isVoid:!0}),link:new g({isVoid:!0}),img:new g({isVoid:!0}),input:new g({isVoid:!0}),param:new g({isVoid:!0}),hr:new g({isVoid:!0}),br:new g({isVoid:!0}),source:new g({isVoid:!0}),track:new g({isVoid:!0}),wbr:new g({isVoid:!0}),p:new g({closedByChildren:["address","article","aside","blockquote","div","dl","fieldset","footer","form","h1","h2","h3","h4","h5","h6","header","hgroup","hr","main","nav","ol","p","pre","section","table","ul"],closedByParent:!0}),thead:new g({closedByChildren:["tbody","tfoot"]}),tbody:new g({closedByChildren:["tbody","tfoot"],closedByParent:!0}),tfoot:new g({closedByChildren:["tbody"],closedByParent:!0}),tr:new g({closedByChildren:["tr"],requiredParents:["tbody","tfoot","thead"],closedByParent:!0}),td:new g({closedByChildren:["td","th"],closedByParent:!0}),th:new g({closedByChildren:["td","th"],closedByParent:!0}),col:new g({requiredParents:["colgroup"],isVoid:!0}),svg:new g({implicitNamespacePrefix:"svg"}),math:new g({implicitNamespacePrefix:"math"}),li:new g({closedByChildren:["li"],closedByParent:!0}),dt:new g({closedByChildren:["dt","dd"]}),dd:new g({closedByChildren:["dt","dd"],closedByParent:!0}),rb:new g({closedByChildren:["rb","rt","rtc","rp"],closedByParent:!0}),rt:new g({closedByChildren:["rb","rt","rtc","rp"],closedByParent:!0}),rtc:new g({closedByChildren:["rb","rtc","rp"],closedByParent:!0}),rp:new g({closedByChildren:["rb","rt","rtc","rp"],closedByParent:!0}),optgroup:new g({closedByChildren:["optgroup"],closedByParent:!0}),option:new g({closedByChildren:["option","optgroup"],closedByParent:!0}),pre:new g({ignoreFirstLf:!0}),listing:new g({ignoreFirstLf:!0}),style:new g({contentType:r.RAW_TEXT}),script:new g({contentType:r.RAW_TEXT}),title:new g({contentType:r.ESCAPABLE_RAW_TEXT}),textarea:new g({contentType:r.ESCAPABLE_RAW_TEXT,ignoreFirstLf:!0})}),m[e.toLowerCase()]||h}var _,y,b,j,x=new RegExp("(\\:not\\()|([-\\w]+)|(?:\\.([-\\w]+))|(?:\\[([-.\\w*]+)(?:=([\"']?)([^\\]\"']*)\\5)?\\])|(\\))|(\\s*,\\s*)","g"),O=function(){function e(){this.element=null,this.classNames=[],this.attrs=[],this.notSelectors=[]}return e.parse=function(t){var n,r=[],i=function(e,t){t.notSelectors.length>0&&!t.element&&0==t.classNames.length&&0==t.attrs.length&&(t.element="*"),e.push(t)},o=new e,a=o,s=!1;for(x.lastIndex=0;n=x.exec(t);){if(n[1]){if(s)throw new Error("Nesting :not is not allowed in a selector");s=!0,a=new e,o.notSelectors.push(a)}if(n[2]&&a.setElement(n[2]),n[3]&&a.addClassName(n[3]),n[4]&&a.addAttribute(n[4],n[6]),n[7]&&(s=!1,a=o),n[8]){if(s)throw new Error("Multiple selectors in :not are not supported");i(r,o),o=a=new e}}return i(r,o),r},e.prototype.isElementSelector=function(){return this.hasElementSelector()&&0==this.classNames.length&&0==this.attrs.length&&0===this.notSelectors.length},e.prototype.hasElementSelector=function(){return!!this.element},e.prototype.setElement=function(e){void 0===e&&(e=null),this.element=e},e.prototype.getMatchingElementTemplate=function(){for(var e=this.element||"div",t=this.classNames.length>0?' class="'+this.classNames.join(" ")+'"':"",n="",r=0;r<this.attrs.length;r+=2){n+=" "+this.attrs[r]+(""!==this.attrs[r+1]?'="'+this.attrs[r+1]+'"':"")}return v(e).isVoid?"<"+e+t+n+"/>":"<"+e+t+n+"></"+e+">"},e.prototype.getAttrs=function(){var e=[];return this.classNames.length>0&&e.push("class",this.classNames.join(" ")),e.concat(this.attrs)},e.prototype.addAttribute=function(e,t){void 0===t&&(t=""),this.attrs.push(e,t&&t.toLowerCase()||"")},e.prototype.addClassName=function(e){this.classNames.push(e.toLowerCase())},e.prototype.toString=function(){var e=this.element||"";if(this.classNames&&this.classNames.forEach(function(t){return e+="."+t}),this.attrs)for(var t=0;t<this.attrs.length;t+=2){var n=this.attrs[t],r=this.attrs[t+1];e+="["+n+(r?"="+r:"")+"]"}return this.notSelectors.forEach(function(t){return e+=":not("+t+")"}),e},e}(),w=function(){function e(){this._elementMap=new Map,this._elementPartialMap=new Map,this._classMap=new Map,this._classPartialMap=new Map,this._attrValueMap=new Map,this._attrValuePartialMap=new Map,this._listContexts=[]}return e.createNotMatcher=function(t){var n=new e;return n.addSelectables(t,null),n},e.prototype.addSelectables=function(e,t){var n=null;e.length>1&&(n=new C(e),this._listContexts.push(n));for(var r=0;r<e.length;r++)this._addSelectable(e[r],t,n)},e.prototype._addSelectable=function(e,t,n){var r=this,i=e.element,o=e.classNames,a=e.attrs,s=new E(e,t,n);i&&((c=0===a.length&&0===o.length)?this._addTerminal(r._elementMap,i,s):r=this._addPartial(r._elementPartialMap,i));if(o)for(var u=0;u<o.length;u++){var c=0===a.length&&u===o.length-1,l=o[u];c?this._addTerminal(r._classMap,l,s):r=this._addPartial(r._classPartialMap,l)}if(a)for(u=0;u<a.length;u+=2){c=u===a.length-2;var d=a[u],p=a[u+1];if(c){var f=r._attrValueMap,h=f.get(d);h||(h=new Map,f.set(d,h)),this._addTerminal(h,p,s)}else{var m=r._attrValuePartialMap,g=m.get(d);g||(g=new Map,m.set(d,g)),r=this._addPartial(g,p)}}},e.prototype._addTerminal=function(e,t,n){var r=e.get(t);r||(r=[],e.set(t,r)),r.push(n)},e.prototype._addPartial=function(t,n){var r=t.get(n);return r||(r=new e,t.set(n,r)),r},e.prototype.match=function(e,t){for(var n=!1,r=e.element,i=e.classNames,o=e.attrs,a=0;a<this._listContexts.length;a++)this._listContexts[a].alreadyMatched=!1;if(n=this._matchTerminal(this._elementMap,r,e,t)||n,n=this._matchPartial(this._elementPartialMap,r,e,t)||n,i)for(a=0;a<i.length;a++){var s=i[a];n=this._matchTerminal(this._classMap,s,e,t)||n,n=this._matchPartial(this._classPartialMap,s,e,t)||n}if(o)for(a=0;a<o.length;a+=2){var u=o[a],c=o[a+1],l=this._attrValueMap.get(u);c&&(n=this._matchTerminal(l,"",e,t)||n),n=this._matchTerminal(l,c,e,t)||n;var d=this._attrValuePartialMap.get(u);c&&(n=this._matchPartial(d,"",e,t)||n),n=this._matchPartial(d,c,e,t)||n}return n},e.prototype._matchTerminal=function(e,t,n,r){if(!e||"string"!=typeof t)return!1;var i=e.get(t)||[],o=e.get("*");if(o&&(i=i.concat(o)),0===i.length)return!1;for(var a=!1,s=0;s<i.length;s++)a=i[s].finalize(n,r)||a;return a},e.prototype._matchPartial=function(e,t,n,r){if(!e||"string"!=typeof t)return!1;var i=e.get(t);return!!i&&i.match(n,r)},e}(),C=function(){return function(e){this.selectors=e,this.alreadyMatched=!1}}(),E=function(){function e(e,t,n){this.selector=e,this.cbContext=t,this.listContext=n,this.notSelectors=e.notSelectors}return e.prototype.finalize=function(e,t){var n=!0;!(this.notSelectors.length>0)||this.listContext&&this.listContext.alreadyMatched||(n=!w.createNotMatcher(this.notSelectors).match(e,null));return!n||!t||this.listContext&&this.listContext.alreadyMatched||(this.listContext&&(this.listContext.alreadyMatched=!0),t(this.selector,this.cbContext)),n},e}(),S=ee("Inject",function(e){return{token:e}}),k=ee("InjectionToken",function(e){return{_desc:e,ngInjectableDef:void 0}}),M=ee("Attribute",function(e){return{attributeName:e}}),A=ee("ContentChildren",function(e,t){return void 0===t&&(t={}),Object(o.__assign)({selector:e,first:!1,isViewQuery:!1,descendants:!1},t)}),T=ee("ContentChild",function(e,t){return void 0===t&&(t={}),Object(o.__assign)({selector:e,first:!0,isViewQuery:!1,descendants:!0},t)}),D=ee("ViewChildren",function(e,t){return void 0===t&&(t={}),Object(o.__assign)({selector:e,first:!1,isViewQuery:!0,descendants:!0},t)}),I=ee("ViewChild",function(e,t){return Object(o.__assign)({selector:e,first:!0,isViewQuery:!0,descendants:!0},t)}),P=ee("Directive",function(e){return void 0===e&&(e={}),e});(y=_||(_={}))[y.Emulated=0]="Emulated",y[y.Native=1]="Native",y[y.None=2]="None",y[y.ShadowDom=3]="ShadowDom",(j=b||(b={}))[j.OnPush=0]="OnPush",j[j.Default=1]="Default";var R,N,F,L,V=ee("Component",function(e){return void 0===e&&(e={}),Object(o.__assign)({changeDetection:b.Default},e)}),B=ee("Pipe",function(e){return Object(o.__assign)({pure:!0},e)}),z=ee("Input",function(e){return{bindingPropertyName:e}}),U=ee("Output",function(e){return{bindingPropertyName:e}}),H=ee("HostBinding",function(e){return{hostPropertyName:e}}),q=ee("HostListener",function(e,t){return{eventName:e,args:t}}),G=ee("NgModule",function(e){return e}),W=ee("Injectable",function(e){return void 0===e&&(e={}),e}),Y={name:"custom-elements"},K={name:"no-errors-schema"},Q=ee("Optional"),X=ee("Self"),Z=ee("SkipSelf"),$=ee("Host"),J=Function;function ee(e,t){var n=function(){for(var n=[],r=0;r<arguments.length;r++)n[r]=arguments[r];var i=t?t.apply(void 0,Object(o.__spread)(n)):{};return Object(o.__assign)({ngMetadataName:e},i)};return n.isTypeOf=function(t){return t&&t.ngMetadataName===e},n.ngMetadataName=e,n}function te(e){var t,n,r,i=(n=(t=e).classNames&&t.classNames.length?Object(o.__spread)([8],t.classNames):[],r=t.element&&"*"!==t.element?t.element:"",Object(o.__spread)([r],t.attrs,n)),a=e.notSelectors&&e.notSelectors.length?e.notSelectors.map(function(e){return n=(t=e).classNames&&t.classNames.length?Object(o.__spread)([8],t.classNames):[],t.element?Object(o.__spread)([5,t.element],t.attrs,n):t.attrs.length?Object(o.__spread)([3],t.attrs,n):t.classNames&&t.classNames.length?Object(o.__spread)([9],t.classNames):[];var t,n}):[];return i.concat.apply(i,Object(o.__spread)(a))}function ne(e){return e?O.parse(e).map(te):[]}(N=R||(R={}))[N.NONE=0]="NONE",N[N.HTML=1]="HTML",N[N.STYLE=2]="STYLE",N[N.SCRIPT=3]="SCRIPT",N[N.URL=4]="URL",N[N.RESOURCE_URL=5]="RESOURCE_URL",(L=F||(F={}))[L.Error=0]="Error",L[L.Warning=1]="Warning",L[L.Ignore=2]="Ignore";var re,ie,oe=Object.freeze({createInject:S,createInjectionToken:k,createAttribute:M,createContentChildren:A,createContentChild:T,createViewChildren:D,createViewChild:I,createDirective:P,get ViewEncapsulation(){return _},get ChangeDetectionStrategy(){return b},createComponent:V,createPipe:B,createInput:z,createOutput:U,createHostBinding:H,createHostListener:q,createNgModule:G,createInjectable:W,CUSTOM_ELEMENTS_SCHEMA:Y,NO_ERRORS_SCHEMA:K,createOptional:Q,createSelf:X,createSkipSelf:Z,createHost:$,Type:J,get SecurityContext(){return R},get MissingTranslationStrategy(){return F},parseSelectorToR3Selector:ne});(ie=re||(re={}))[ie.Const=0]="Const";var ae,se,ue=function(){function e(e){void 0===e&&(e=null),this.modifiers=e,e||(this.modifiers=[])}return e.prototype.hasModifier=function(e){return-1!==this.modifiers.indexOf(e)},e}();(se=ae||(ae={}))[se.Dynamic=0]="Dynamic",se[se.Bool=1]="Bool",se[se.String=2]="String",se[se.Int=3]="Int",se[se.Number=4]="Number",se[se.Function=5]="Function",se[se.Inferred=6]="Inferred",se[se.None=7]="None";var ce,le,de=function(e){function t(t,n){void 0===n&&(n=null);var r=e.call(this,n)||this;return r.name=t,r}return Object(o.__extends)(t,e),t.prototype.visitType=function(e,t){return e.visitBuiltinType(this,t)},t}(ue),pe=function(e){function t(t,n,r){void 0===n&&(n=null),void 0===r&&(r=null);var i=e.call(this,n)||this;return i.value=t,i.typeParams=r,i}return Object(o.__extends)(t,e),t.prototype.visitType=function(e,t){return e.visitExpressionType(this,t)},t}(ue),fe=function(e){function t(t,n){void 0===n&&(n=null);var r=e.call(this,n)||this;return r.of=t,r}return Object(o.__extends)(t,e),t.prototype.visitType=function(e,t){return e.visitArrayType(this,t)},t}(ue),he=function(e){function t(t,n){void 0===n&&(n=null);var r=e.call(this,n)||this;return r.valueType=t||null,r}return Object(o.__extends)(t,e),t.prototype.visitType=function(e,t){return e.visitMapType(this,t)},t}(ue),me=new de(ae.Dynamic),ge=new de(ae.Inferred),ve=new de(ae.Bool),_e=(new de(ae.Int),new de(ae.Number)),ye=(new de(ae.String),new de(ae.Function),new de(ae.None));function be(e,t){return null==e||null==t?e==t:e.isEquivalent(t)}function je(e,t){var n=e.length;if(n!==t.length)return!1;for(var r=0;r<n;r++)if(!e[r].isEquivalent(t[r]))return!1;return!0}(le=ce||(ce={}))[le.Equals=0]="Equals",le[le.NotEquals=1]="NotEquals",le[le.Identical=2]="Identical",le[le.NotIdentical=3]="NotIdentical",le[le.Minus=4]="Minus",le[le.Plus=5]="Plus",le[le.Divide=6]="Divide",le[le.Multiply=7]="Multiply",le[le.Modulo=8]="Modulo",le[le.And=9]="And",le[le.Or=10]="Or",le[le.BitwiseAnd=11]="BitwiseAnd",le[le.Lower=12]="Lower",le[le.LowerEquals=13]="LowerEquals",le[le.Bigger=14]="Bigger",le[le.BiggerEquals=15]="BiggerEquals";var xe,Oe,we=function(){function e(e,t){this.type=e||null,this.sourceSpan=t||null}return e.prototype.prop=function(e,t){return new Ke(this,e,null,t)},e.prototype.key=function(e,t,n){return new Qe(this,e,t,n)},e.prototype.callMethod=function(e,t,n){return new Re(this,e,t,null,n)},e.prototype.callFn=function(e,t){return new Ne(this,e,null,t)},e.prototype.instantiate=function(e,t,n){return new Fe(this,e,t,n)},e.prototype.conditional=function(e,t,n){return void 0===t&&(t=null),new ze(this,e,t,null,n)},e.prototype.equals=function(e,t){return new Ye(ce.Equals,this,e,null,t)},e.prototype.notEquals=function(e,t){return new Ye(ce.NotEquals,this,e,null,t)},e.prototype.identical=function(e,t){return new Ye(ce.Identical,this,e,null,t)},e.prototype.notIdentical=function(e,t){return new Ye(ce.NotIdentical,this,e,null,t)},e.prototype.minus=function(e,t){return new Ye(ce.Minus,this,e,null,t)},e.prototype.plus=function(e,t){return new Ye(ce.Plus,this,e,null,t)},e.prototype.divide=function(e,t){return new Ye(ce.Divide,this,e,null,t)},e.prototype.multiply=function(e,t){return new Ye(ce.Multiply,this,e,null,t)},e.prototype.modulo=function(e,t){return new Ye(ce.Modulo,this,e,null,t)},e.prototype.and=function(e,t){return new Ye(ce.And,this,e,null,t)},e.prototype.bitwiseAnd=function(e,t,n){return void 0===n&&(n=!0),new Ye(ce.BitwiseAnd,this,e,null,t,n)},e.prototype.or=function(e,t){return new Ye(ce.Or,this,e,null,t)},e.prototype.lower=function(e,t){return new Ye(ce.Lower,this,e,null,t)},e.prototype.lowerEquals=function(e,t){return new Ye(ce.LowerEquals,this,e,null,t)},e.prototype.bigger=function(e,t){return new Ye(ce.Bigger,this,e,null,t)},e.prototype.biggerEquals=function(e,t){return new Ye(ce.BiggerEquals,this,e,null,t)},e.prototype.isBlank=function(e){return this.equals(tt,e)},e.prototype.cast=function(e,t){return new qe(this,e,t)},e.prototype.toStmt=function(){return new ot(this,null)},e}();(Oe=xe||(xe={}))[Oe.This=0]="This",Oe[Oe.Super=1]="Super",Oe[Oe.CatchError=2]="CatchError",Oe[Oe.CatchStack=3]="CatchStack";var Ce,Ee,Se=function(e){function t(t,n,r){var i=e.call(this,n,r)||this;return"string"==typeof t?(i.name=t,i.builtin=null):(i.name=null,i.builtin=t),i}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.name===e.name&&this.builtin===e.builtin},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitReadVarExpr(this,t)},t.prototype.set=function(e){if(!this.name)throw new Error("Built in variable "+this.builtin+" can not be assigned to.");return new Ae(this.name,e,null,this.sourceSpan)},t}(we),ke=function(e){function t(t,n,r){var i=e.call(this,n,r)||this;return i.expr=t,i}return Object(o.__extends)(t,e),t.prototype.visitExpression=function(e,t){return e.visitTypeofExpr(this,t)},t.prototype.isEquivalent=function(e){return e instanceof t&&e.expr.isEquivalent(this.expr)},t.prototype.isConstant=function(){return this.expr.isConstant()},t}(we),Me=function(e){function t(t,n,r){var i=e.call(this,n,r)||this;return i.node=t,i}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.node===e.node},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitWrappedNodeExpr(this,t)},t}(we),Ae=function(e){function t(t,n,r,i){var o=e.call(this,r||n.type,i)||this;return o.name=t,o.value=n,o}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.name===e.name&&this.value.isEquivalent(e.value)},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitWriteVarExpr(this,t)},t.prototype.toDeclStmt=function(e,t){return new rt(this.name,this.value,e,t,this.sourceSpan)},t.prototype.toConstDecl=function(){return this.toDeclStmt(ge,[Ie.Final])},t}(we),Te=function(e){function t(t,n,r,i,o){var a=e.call(this,i||r.type,o)||this;return a.receiver=t,a.index=n,a.value=r,a}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.receiver.isEquivalent(e.receiver)&&this.index.isEquivalent(e.index)&&this.value.isEquivalent(e.value)},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitWriteKeyExpr(this,t)},t}(we),De=function(e){function t(t,n,r,i,o){var a=e.call(this,i||r.type,o)||this;return a.receiver=t,a.name=n,a.value=r,a}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.receiver.isEquivalent(e.receiver)&&this.name===e.name&&this.value.isEquivalent(e.value)},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitWritePropExpr(this,t)},t}(we);(Ee=Ce||(Ce={}))[Ee.ConcatArray=0]="ConcatArray",Ee[Ee.SubscribeObservable=1]="SubscribeObservable",Ee[Ee.Bind=2]="Bind";var Ie,Pe,Re=function(e){function t(t,n,r,i,o){var a=e.call(this,i,o)||this;return a.receiver=t,a.args=r,"string"==typeof n?(a.name=n,a.builtin=null):(a.name=null,a.builtin=n),a}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.receiver.isEquivalent(e.receiver)&&this.name===e.name&&this.builtin===e.builtin&&je(this.args,e.args)},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitInvokeMethodExpr(this,t)},t}(we),Ne=function(e){function t(t,n,r,i,o){void 0===o&&(o=!1);var a=e.call(this,r,i)||this;return a.fn=t,a.args=n,a.pure=o,a}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.fn.isEquivalent(e.fn)&&je(this.args,e.args)&&this.pure===e.pure},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitInvokeFunctionExpr(this,t)},t}(we),Fe=function(e){function t(t,n,r,i){var o=e.call(this,r,i)||this;return o.classExpr=t,o.args=n,o}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.classExpr.isEquivalent(e.classExpr)&&je(this.args,e.args)},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitInstantiateExpr(this,t)},t}(we),Le=function(e){function t(t,n,r){var i=e.call(this,n,r)||this;return i.value=t,i}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.value===e.value},t.prototype.isConstant=function(){return!0},t.prototype.visitExpression=function(e,t){return e.visitLiteralExpr(this,t)},t}(we),Ve=function(e){function t(t,n,r,i){void 0===r&&(r=null);var o=e.call(this,n,i)||this;return o.value=t,o.typeParams=r,o}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.value.name===e.value.name&&this.value.moduleName===e.value.moduleName&&this.value.runtime===e.value.runtime},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitExternalExpr(this,t)},t}(we),Be=function(){return function(e,t,n){this.moduleName=e,this.name=t,this.runtime=n}}(),ze=function(e){function t(t,n,r,i,o){void 0===r&&(r=null);var a=e.call(this,i||n.type,o)||this;return a.condition=t,a.falseCase=r,a.trueCase=n,a}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.condition.isEquivalent(e.condition)&&this.trueCase.isEquivalent(e.trueCase)&&be(this.falseCase,e.falseCase)},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitConditionalExpr(this,t)},t}(we),Ue=function(e){function t(t,n){var r=e.call(this,ve,n)||this;return r.condition=t,r}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.condition.isEquivalent(e.condition)},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitNotExpr(this,t)},t}(we),He=function(e){function t(t,n){var r=e.call(this,t.type,n)||this;return r.condition=t,r}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.condition.isEquivalent(e.condition)},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitAssertNotNullExpr(this,t)},t}(we),qe=function(e){function t(t,n,r){var i=e.call(this,n,r)||this;return i.value=t,i}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.value.isEquivalent(e.value)},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitCastExpr(this,t)},t}(we),Ge=function(){function e(e,t){void 0===t&&(t=null),this.name=e,this.type=t}return e.prototype.isEquivalent=function(e){return this.name===e.name},e}(),We=function(e){function t(t,n,r,i,o){var a=e.call(this,r,i)||this;return a.params=t,a.statements=n,a.name=o,a}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&je(this.params,e.params)&&je(this.statements,e.statements)},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitFunctionExpr(this,t)},t.prototype.toDeclStmt=function(e,t){return void 0===t&&(t=null),new it(e,this.params,this.statements,this.type,t,this.sourceSpan)},t}(we),Ye=function(e){function t(t,n,r,i,o,a){void 0===a&&(a=!0);var s=e.call(this,i||n.type,o)||this;return s.operator=t,s.rhs=r,s.parens=a,s.lhs=n,s}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.operator===e.operator&&this.lhs.isEquivalent(e.lhs)&&this.rhs.isEquivalent(e.rhs)},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitBinaryOperatorExpr(this,t)},t}(we),Ke=function(e){function t(t,n,r,i){var o=e.call(this,r,i)||this;return o.receiver=t,o.name=n,o}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.receiver.isEquivalent(e.receiver)&&this.name===e.name},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitReadPropExpr(this,t)},t.prototype.set=function(e){return new De(this.receiver,this.name,e,null,this.sourceSpan)},t}(we),Qe=function(e){function t(t,n,r,i){var o=e.call(this,r,i)||this;return o.receiver=t,o.index=n,o}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.receiver.isEquivalent(e.receiver)&&this.index.isEquivalent(e.index)},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitReadKeyExpr(this,t)},t.prototype.set=function(e){return new Te(this.receiver,this.index,e,null,this.sourceSpan)},t}(we),Xe=function(e){function t(t,n,r){var i=e.call(this,n,r)||this;return i.entries=t,i}return Object(o.__extends)(t,e),t.prototype.isConstant=function(){return this.entries.every(function(e){return e.isConstant()})},t.prototype.isEquivalent=function(e){return e instanceof t&&je(this.entries,e.entries)},t.prototype.visitExpression=function(e,t){return e.visitLiteralArrayExpr(this,t)},t}(we),Ze=function(){function e(e,t,n){this.key=e,this.value=t,this.quoted=n}return e.prototype.isEquivalent=function(e){return this.key===e.key&&this.value.isEquivalent(e.value)},e}(),$e=function(e){function t(t,n,r){var i=e.call(this,n,r)||this;return i.entries=t,i.valueType=null,n&&(i.valueType=n.valueType),i}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&je(this.entries,e.entries)},t.prototype.isConstant=function(){return this.entries.every(function(e){return e.value.isConstant()})},t.prototype.visitExpression=function(e,t){return e.visitLiteralMapExpr(this,t)},t}(we),Je=function(e){function t(t,n){var r=e.call(this,t[t.length-1].type,n)||this;return r.parts=t,r}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&je(this.parts,e.parts)},t.prototype.isConstant=function(){return!1},t.prototype.visitExpression=function(e,t){return e.visitCommaExpr(this,t)},t}(we),et=(new Se(xe.This,null,null),new Se(xe.Super,null,null),new Se(xe.CatchError,null,null),new Se(xe.CatchStack,null,null),new Le(null,null,null)),tt=new Le(null,ge,null);(Pe=Ie||(Ie={}))[Pe.Final=0]="Final",Pe[Pe.Private=1]="Private",Pe[Pe.Exported=2]="Exported",Pe[Pe.Static=3]="Static";var nt=function(){function e(e,t){this.modifiers=e||[],this.sourceSpan=t||null}return e.prototype.hasModifier=function(e){return-1!==this.modifiers.indexOf(e)},e}(),rt=function(e){function t(t,n,r,i,o){void 0===i&&(i=null);var a=e.call(this,i,o)||this;return a.name=t,a.value=n,a.type=r||n&&n.type||null,a}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.name===e.name&&(this.value?!!e.value&&this.value.isEquivalent(e.value):!e.value)},t.prototype.visitStatement=function(e,t){return e.visitDeclareVarStmt(this,t)},t}(nt),it=function(e){function t(t,n,r,i,o,a){void 0===o&&(o=null);var s=e.call(this,o,a)||this;return s.name=t,s.params=n,s.statements=r,s.type=i||null,s}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&je(this.params,e.params)&&je(this.statements,e.statements)},t.prototype.visitStatement=function(e,t){return e.visitDeclareFunctionStmt(this,t)},t}(nt),ot=function(e){function t(t,n){var r=e.call(this,null,n)||this;return r.expr=t,r}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.expr.isEquivalent(e.expr)},t.prototype.visitStatement=function(e,t){return e.visitExpressionStmt(this,t)},t}(nt),at=function(e){function t(t,n){var r=e.call(this,null,n)||this;return r.value=t,r}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.value.isEquivalent(e.value)},t.prototype.visitStatement=function(e,t){return e.visitReturnStmt(this,t)},t}(nt),st=function(){function e(e,t){this.modifiers=t,t||(this.modifiers=[]),this.type=e||null}return e.prototype.hasModifier=function(e){return-1!==this.modifiers.indexOf(e)},e}(),ut=function(e){function t(t,n,r,i){void 0===r&&(r=null);var o=e.call(this,n,r)||this;return o.name=t,o.initializer=i,o}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return this.name===e.name},t}(st),ct=function(e){function t(t,n,r,i,o){void 0===o&&(o=null);var a=e.call(this,i,o)||this;return a.name=t,a.params=n,a.body=r,a}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return this.name===e.name&&je(this.body,e.body)},t}(st),lt=function(e){function t(t,n,r,i){void 0===i&&(i=null);var o=e.call(this,r,i)||this;return o.name=t,o.body=n,o}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return this.name===e.name&&je(this.body,e.body)},t}(st),dt=function(e){function t(t,n,r,i,o,a,s,u){void 0===s&&(s=null);var c=e.call(this,s,u)||this;return c.name=t,c.parent=n,c.fields=r,c.getters=i,c.constructorMethod=o,c.methods=a,c}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.name===e.name&&be(this.parent,e.parent)&&je(this.fields,e.fields)&&je(this.getters,e.getters)&&this.constructorMethod.isEquivalent(e.constructorMethod)&&je(this.methods,e.methods)},t.prototype.visitStatement=function(e,t){return e.visitDeclareClassStmt(this,t)},t}(nt),pt=function(e){function t(t,n,r,i){void 0===r&&(r=[]);var o=e.call(this,null,i)||this;return o.condition=t,o.trueCase=n,o.falseCase=r,o}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.condition.isEquivalent(e.condition)&&je(this.trueCase,e.trueCase)&&je(this.falseCase,e.falseCase)},t.prototype.visitStatement=function(e,t){return e.visitIfStmt(this,t)},t}(nt),ft=function(e){function t(t,n,r){void 0===n&&(n=!1);var i=e.call(this,null,r)||this;return i.comment=t,i.multiline=n,i}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t},t.prototype.visitStatement=function(e,t){return e.visitCommentStmt(this,t)},t}(nt),ht=function(e){function t(t,n){void 0===t&&(t=[]);var r=e.call(this,null,n)||this;return r.tags=t,r}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&this.toString()===e.toString()},t.prototype.visitStatement=function(e,t){return e.visitJSDocCommentStmt(this,t)},t.prototype.toString=function(){return function(e){var t,n;if(0===e.length)return"";var r="*\n";try{for(var i=Object(o.__values)(e),a=i.next();!a.done;a=i.next()){var s=a.value;r+=" *",r+=Rt(s).replace(/\n/g,"\n * "),r+="\n"}}catch(e){t={error:e}}finally{try{a&&!a.done&&(n=i.return)&&n.call(i)}finally{if(t)throw t.error}}return r+=" "}(this.tags)},t}(nt),mt=function(e){function t(t,n,r){var i=e.call(this,null,r)||this;return i.bodyStmts=t,i.catchStmts=n,i}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof t&&je(this.bodyStmts,e.bodyStmts)&&je(this.catchStmts,e.catchStmts)},t.prototype.visitStatement=function(e,t){return e.visitTryCatchStmt(this,t)},t}(nt),gt=function(e){function t(t,n){var r=e.call(this,null,n)||this;return r.error=t,r}return Object(o.__extends)(t,e),t.prototype.isEquivalent=function(e){return e instanceof mt&&this.error.isEquivalent(e.error)},t.prototype.visitStatement=function(e,t){return e.visitThrowStmt(this,t)},t}(nt),vt=function(){function e(){}return e.prototype.transformExpr=function(e,t){return e},e.prototype.transformStmt=function(e,t){return e},e.prototype.visitReadVarExpr=function(e,t){return this.transformExpr(e,t)},e.prototype.visitWrappedNodeExpr=function(e,t){return this.transformExpr(e,t)},e.prototype.visitTypeofExpr=function(e,t){return this.transformExpr(new ke(e.expr.visitExpression(this,t),e.type,e.sourceSpan),t)},e.prototype.visitWriteVarExpr=function(e,t){return this.transformExpr(new Ae(e.name,e.value.visitExpression(this,t),e.type,e.sourceSpan),t)},e.prototype.visitWriteKeyExpr=function(e,t){return this.transformExpr(new Te(e.receiver.visitExpression(this,t),e.index.visitExpression(this,t),e.value.visitExpression(this,t),e.type,e.sourceSpan),t)},e.prototype.visitWritePropExpr=function(e,t){return this.transformExpr(new De(e.receiver.visitExpression(this,t),e.name,e.value.visitExpression(this,t),e.type,e.sourceSpan),t)},e.prototype.visitInvokeMethodExpr=function(e,t){var n=e.builtin||e.name;return this.transformExpr(new Re(e.receiver.visitExpression(this,t),n,this.visitAllExpressions(e.args,t),e.type,e.sourceSpan),t)},e.prototype.visitInvokeFunctionExpr=function(e,t){return this.transformExpr(new Ne(e.fn.visitExpression(this,t),this.visitAllExpressions(e.args,t),e.type,e.sourceSpan),t)},e.prototype.visitInstantiateExpr=function(e,t){return this.transformExpr(new Fe(e.classExpr.visitExpression(this,t),this.visitAllExpressions(e.args,t),e.type,e.sourceSpan),t)},e.prototype.visitLiteralExpr=function(e,t){return this.transformExpr(e,t)},e.prototype.visitExternalExpr=function(e,t){return this.transformExpr(e,t)},e.prototype.visitConditionalExpr=function(e,t){return this.transformExpr(new ze(e.condition.visitExpression(this,t),e.trueCase.visitExpression(this,t),e.falseCase.visitExpression(this,t),e.type,e.sourceSpan),t)},e.prototype.visitNotExpr=function(e,t){return this.transformExpr(new Ue(e.condition.visitExpression(this,t),e.sourceSpan),t)},e.prototype.visitAssertNotNullExpr=function(e,t){return this.transformExpr(new He(e.condition.visitExpression(this,t),e.sourceSpan),t)},e.prototype.visitCastExpr=function(e,t){return this.transformExpr(new qe(e.value.visitExpression(this,t),e.type,e.sourceSpan),t)},e.prototype.visitFunctionExpr=function(e,t){return this.transformExpr(new We(e.params,this.visitAllStatements(e.statements,t),e.type,e.sourceSpan),t)},e.prototype.visitBinaryOperatorExpr=function(e,t){return this.transformExpr(new Ye(e.operator,e.lhs.visitExpression(this,t),e.rhs.visitExpression(this,t),e.type,e.sourceSpan),t)},e.prototype.visitReadPropExpr=function(e,t){return this.transformExpr(new Ke(e.receiver.visitExpression(this,t),e.name,e.type,e.sourceSpan),t)},e.prototype.visitReadKeyExpr=function(e,t){return this.transformExpr(new Qe(e.receiver.visitExpression(this,t),e.index.visitExpression(this,t),e.type,e.sourceSpan),t)},e.prototype.visitLiteralArrayExpr=function(e,t){return this.transformExpr(new Xe(this.visitAllExpressions(e.entries,t),e.type,e.sourceSpan),t)},e.prototype.visitLiteralMapExpr=function(e,t){var n=this,r=e.entries.map(function(e){return new Ze(e.key,e.value.visitExpression(n,t),e.quoted)}),i=new he(e.valueType,null);return this.transformExpr(new $e(r,i,e.sourceSpan),t)},e.prototype.visitCommaExpr=function(e,t){return this.transformExpr(new Je(this.visitAllExpressions(e.parts,t),e.sourceSpan),t)},e.prototype.visitAllExpressions=function(e,t){var n=this;return e.map(function(e){return e.visitExpression(n,t)})},e.prototype.visitDeclareVarStmt=function(e,t){var n=e.value&&e.value.visitExpression(this,t);return this.transformStmt(new rt(e.name,n,e.type,e.modifiers,e.sourceSpan),t)},e.prototype.visitDeclareFunctionStmt=function(e,t){return this.transformStmt(new it(e.name,e.params,this.visitAllStatements(e.statements,t),e.type,e.modifiers,e.sourceSpan),t)},e.prototype.visitExpressionStmt=function(e,t){return this.transformStmt(new ot(e.expr.visitExpression(this,t),e.sourceSpan),t)},e.prototype.visitReturnStmt=function(e,t){return this.transformStmt(new at(e.value.visitExpression(this,t),e.sourceSpan),t)},e.prototype.visitDeclareClassStmt=function(e,t){var n=this,r=e.parent.visitExpression(this,t),i=e.getters.map(function(e){return new lt(e.name,n.visitAllStatements(e.body,t),e.type,e.modifiers)}),o=e.constructorMethod&&new ct(e.constructorMethod.name,e.constructorMethod.params,this.visitAllStatements(e.constructorMethod.body,t),e.constructorMethod.type,e.constructorMethod.modifiers),a=e.methods.map(function(e){return new ct(e.name,e.params,n.visitAllStatements(e.body,t),e.type,e.modifiers)});return this.transformStmt(new dt(e.name,r,e.fields,i,o,a,e.modifiers,e.sourceSpan),t)},e.prototype.visitIfStmt=function(e,t){return this.transformStmt(new pt(e.condition.visitExpression(this,t),this.visitAllStatements(e.trueCase,t),this.visitAllStatements(e.falseCase,t),e.sourceSpan),t)},e.prototype.visitTryCatchStmt=function(e,t){return this.transformStmt(new mt(this.visitAllStatements(e.bodyStmts,t),this.visitAllStatements(e.catchStmts,t),e.sourceSpan),t)},e.prototype.visitThrowStmt=function(e,t){return this.transformStmt(new gt(e.error.visitExpression(this,t),e.sourceSpan),t)},e.prototype.visitCommentStmt=function(e,t){return this.transformStmt(e,t)},e.prototype.visitJSDocCommentStmt=function(e,t){return this.transformStmt(e,t)},e.prototype.visitAllStatements=function(e,t){var n=this;return e.map(function(e){return e.visitStatement(n,t)})},e}(),_t=function(){function e(){}return e.prototype.visitType=function(e,t){return e},e.prototype.visitExpression=function(e,t){return e.type&&e.type.visitType(this,t),e},e.prototype.visitBuiltinType=function(e,t){return this.visitType(e,t)},e.prototype.visitExpressionType=function(e,t){var n=this;return e.value.visitExpression(this,t),null!==e.typeParams&&e.typeParams.forEach(function(e){return n.visitType(e,t)}),this.visitType(e,t)},e.prototype.visitArrayType=function(e,t){return this.visitType(e,t)},e.prototype.visitMapType=function(e,t){return this.visitType(e,t)},e.prototype.visitWrappedNodeExpr=function(e,t){return e},e.prototype.visitTypeofExpr=function(e,t){return this.visitExpression(e,t)},e.prototype.visitReadVarExpr=function(e,t){return this.visitExpression(e,t)},e.prototype.visitWriteVarExpr=function(e,t){return e.value.visitExpression(this,t),this.visitExpression(e,t)},e.prototype.visitWriteKeyExpr=function(e,t){return e.receiver.visitExpression(this,t),e.index.visitExpression(this,t),e.value.visitExpression(this,t),this.visitExpression(e,t)},e.prototype.visitWritePropExpr=function(e,t){return e.receiver.visitExpression(this,t),e.value.visitExpression(this,t),this.visitExpression(e,t)},e.prototype.visitInvokeMethodExpr=function(e,t){return e.receiver.visitExpression(this,t),this.visitAllExpressions(e.args,t),this.visitExpression(e,t)},e.prototype.visitInvokeFunctionExpr=function(e,t){return e.fn.visitExpression(this,t),this.visitAllExpressions(e.args,t),this.visitExpression(e,t)},e.prototype.visitInstantiateExpr=function(e,t){return e.classExpr.visitExpression(this,t),this.visitAllExpressions(e.args,t),this.visitExpression(e,t)},e.prototype.visitLiteralExpr=function(e,t){return this.visitExpression(e,t)},e.prototype.visitExternalExpr=function(e,t){var n=this;return e.typeParams&&e.typeParams.forEach(function(e){return e.visitType(n,t)}),this.visitExpression(e,t)},e.prototype.visitConditionalExpr=function(e,t){return e.condition.visitExpression(this,t),e.trueCase.visitExpression(this,t),e.falseCase.visitExpression(this,t),this.visitExpression(e,t)},e.prototype.visitNotExpr=function(e,t){return e.condition.visitExpression(this,t),this.visitExpression(e,t)},e.prototype.visitAssertNotNullExpr=function(e,t){return e.condition.visitExpression(this,t),this.visitExpression(e,t)},e.prototype.visitCastExpr=function(e,t){return e.value.visitExpression(this,t),this.visitExpression(e,t)},e.prototype.visitFunctionExpr=function(e,t){return this.visitAllStatements(e.statements,t),this.visitExpression(e,t)},e.prototype.visitBinaryOperatorExpr=function(e,t){return e.lhs.visitExpression(this,t),e.rhs.visitExpression(this,t),this.visitExpression(e,t)},e.prototype.visitReadPropExpr=function(e,t){return e.receiver.visitExpression(this,t),this.visitExpression(e,t)},e.prototype.visitReadKeyExpr=function(e,t){return e.receiver.visitExpression(this,t),e.index.visitExpression(this,t),this.visitExpression(e,t)},e.prototype.visitLiteralArrayExpr=function(e,t){return this.visitAllExpressions(e.entries,t),this.visitExpression(e,t)},e.prototype.visitLiteralMapExpr=function(e,t){var n=this;return e.entries.forEach(function(e){return e.value.visitExpression(n,t)}),this.visitExpression(e,t)},e.prototype.visitCommaExpr=function(e,t){return this.visitAllExpressions(e.parts,t),this.visitExpression(e,t)},e.prototype.visitAllExpressions=function(e,t){var n=this;e.forEach(function(e){return e.visitExpression(n,t)})},e.prototype.visitDeclareVarStmt=function(e,t){return e.value&&e.value.visitExpression(this,t),e.type&&e.type.visitType(this,t),e},e.prototype.visitDeclareFunctionStmt=function(e,t){return this.visitAllStatements(e.statements,t),e.type&&e.type.visitType(this,t),e},e.prototype.visitExpressionStmt=function(e,t){return e.expr.visitExpression(this,t),e},e.prototype.visitReturnStmt=function(e,t){return e.value.visitExpression(this,t),e},e.prototype.visitDeclareClassStmt=function(e,t){var n=this;return e.parent.visitExpression(this,t),e.getters.forEach(function(e){return n.visitAllStatements(e.body,t)}),e.constructorMethod&&this.visitAllStatements(e.constructorMethod.body,t),e.methods.forEach(function(e){return n.visitAllStatements(e.body,t)}),e},e.prototype.visitIfStmt=function(e,t){return e.condition.visitExpression(this,t),this.visitAllStatements(e.trueCase,t),this.visitAllStatements(e.falseCase,t),e},e.prototype.visitTryCatchStmt=function(e,t){return this.visitAllStatements(e.bodyStmts,t),this.visitAllStatements(e.catchStmts,t),e},e.prototype.visitThrowStmt=function(e,t){return e.error.visitExpression(this,t),e},e.prototype.visitCommentStmt=function(e,t){return e},e.prototype.visitJSDocCommentStmt=function(e,t){return e},e.prototype.visitAllStatements=function(e,t){var n=this;e.forEach(function(e){return e.visitStatement(n,t)})},e}();function yt(e){var t=new bt;return t.visitAllStatements(e,null),t.varNames}var bt=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.varNames=new Set,t}return Object(o.__extends)(t,e),t.prototype.visitDeclareFunctionStmt=function(e,t){return e},t.prototype.visitDeclareClassStmt=function(e,t){return e},t.prototype.visitReadVarExpr=function(e,t){return e.name&&this.varNames.add(e.name),null},t}(_t);function jt(e){var t=new xt;return t.visitAllStatements(e,null),t.externalReferences}var xt=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.externalReferences=[],t}return Object(o.__extends)(t,e),t.prototype.visitExternalExpr=function(t,n){return this.externalReferences.push(t.value),e.prototype.visitExternalExpr.call(this,t,n)},t}(_t);function Ot(e,t){if(!t)return e;var n=new Ct(t);return e.visitStatement(n,null)}function wt(e,t){if(!t)return e;var n=new Ct(t);return e.visitExpression(n,null)}var Ct=function(e){function t(t){var n=e.call(this)||this;return n.sourceSpan=t,n}return Object(o.__extends)(t,e),t.prototype._clone=function(e){var t=Object.create(e.constructor.prototype);for(var n in e)t[n]=e[n];return t},t.prototype.transformExpr=function(e,t){return e.sourceSpan||((e=this._clone(e)).sourceSpan=this.sourceSpan),e},t.prototype.transformStmt=function(e,t){return e.sourceSpan||((e=this._clone(e)).sourceSpan=this.sourceSpan),e},t}(vt);function Et(e,t,n){return new Se(e,t,n)}function St(e,t,n){return void 0===t&&(t=null),new Ve(e,null,t,n)}function kt(e,t,n){return void 0===t&&(t=null),void 0===n&&(n=null),null!=e?Mt(St(e,t,null),n):null}function Mt(e,t,n){return void 0===t&&(t=null),void 0===n&&(n=null),new pe(e,t,n)}function At(e,t,n){return new Xe(e,t,n)}function Tt(e,t){return void 0===t&&(t=null),new $e(e.map(function(e){return new Ze(e.key,e.value,e.quoted)}),t,null)}function Dt(e,t,n,r,i){return new We(e,t,n,r,i)}function It(e,t,n){return new pt(e,t,n)}function Pt(e,t,n){return new Le(e,t,n)}function Rt(e){var t="";if(e.tagName&&(t+=" @"+e.tagName),e.text){if(e.text.match(/\/\*|\*\//))throw new Error('JSDoc text cannot contain "/*" and "*/"');t+=" "+e.text.replace(/@/g,"\\@")}return t}var Nt=/-+([a-z0-9])/g;function Ft(e,t){return Lt(e,":",t)}function Lt(e,t,n){var r=e.indexOf(t);return-1==r?n:[e.slice(0,r).trim(),e.slice(r+1).trim()]}function Vt(e,t,n){return Array.isArray(e)?t.visitArray(e,n):"object"==typeof(r=e)&&null!==r&&Object.getPrototypeOf(r)===Jt?t.visitStringMap(e,n):null==e||"string"==typeof e||"number"==typeof e||"boolean"==typeof e?t.visitPrimitive(e,n):t.visitOther(e,n);var r}function Bt(e){return null!==e&&void 0!==e}function zt(e){return void 0===e?null:e}var Ut=function(){function e(){}return e.prototype.visitArray=function(e,t){var n=this;return e.map(function(e){return Vt(e,n,t)})},e.prototype.visitStringMap=function(e,t){var n=this,r={};return Object.keys(e).forEach(function(i){r[i]=Vt(e[i],n,t)}),r},e.prototype.visitPrimitive=function(e,t){return e},e.prototype.visitOther=function(e,t){return e},e}(),Ht=function(e){if(rn(e))throw new Error("Illegal state: value cannot be a promise");return e},qt=function(e,t){return rn(e)?e.then(t):t(e)},Gt=function(e){return e.some(rn)?Promise.all(e):e};function Wt(e){throw new Error("Internal Error: "+e)}function Yt(e,t){var n=Error(e);return n[Kt]=!0,t&&(n[Qt]=t),n}var Kt="ngSyntaxError",Qt="ngParseErrors";function Xt(e){return e[Kt]}function Zt(e){return e[Qt]||[]}function $t(e){return e.replace(/([.*+?^=!:${}()|[\]\/\\])/g,"\\$1")}var Jt=Object.getPrototypeOf({});function en(e){for(var t="",n=0;n<e.length;n++){var r=e.charCodeAt(n);if(r>=55296&&r<=56319&&e.length>n+1){var i=e.charCodeAt(n+1);i>=56320&&i<=57343&&(n++,r=(r-55296<<10)+i-56320+65536)}r<=127?t+=String.fromCharCode(r):r<=2047?t+=String.fromCharCode(r>>6&31|192,63&r|128):r<=65535?t+=String.fromCharCode(r>>12|224,r>>6&63|128,63&r|128):r<=2097151&&(t+=String.fromCharCode(r>>18&7|240,r>>12&63|128,r>>6&63|128,63&r|128))}return t}function tn(e){if("string"==typeof e)return e;if(e instanceof Array)return"["+e.map(tn).join(", ")+"]";if(null==e)return""+e;if(e.overriddenName)return""+e.overriddenName;if(e.name)return""+e.name;var t=e.toString();if(null==t)return""+t;var n=t.indexOf("\n");return-1===n?t:t.substring(0,n)}function nn(e){return"function"==typeof e&&e.hasOwnProperty("__forward_ref__")?e():e}function rn(e){return!!e&&"function"==typeof e.then}var on=function(){return function(e){this.full=e;var t=e.split(".");this.major=t[0],this.minor=t[1],this.patch=t.slice(2).join(".")}}(),an="undefined"!=typeof window&&window,sn="undefined"!=typeof self&&"undefined"!=typeof WorkerGlobalScope&&self instanceof WorkerGlobalScope&&self,un="undefined"!=typeof global&&global||an||sn,cn={},ln=function(e){function t(t){var n=e.call(this,t.type)||this;return n.resolved=t,n.original=t,n}return Object(o.__extends)(t,e),t.prototype.visitExpression=function(e,t){return t===cn?this.original.visitExpression(e,t):this.resolved.visitExpression(e,t)},t.prototype.isEquivalent=function(e){return e instanceof t&&this.resolved.isEquivalent(e.resolved)},t.prototype.isConstant=function(){return!0},t.prototype.fixup=function(e){this.resolved=e,this.shared=!0},t}(we),dn=function(){function e(){this.statements=[],this.literals=new Map,this.literalFactories=new Map,this.injectorDefinitions=new Map,this.directiveDefinitions=new Map,this.componentDefinitions=new Map,this.pipeDefinitions=new Map,this.nextNameIndex=0}return e.prototype.getConstLiteral=function(e,t){if(e instanceof Le||e instanceof ln)return e;var n=this.keyOf(e),r=this.literals.get(n),i=!1;if(r||(r=new ln(e),this.literals.set(n,r),i=!0),!i&&!r.shared||i&&t){var o=this.freshName();this.statements.push(Et(o).set(e).toDeclStmt(ge,[Ie.Final])),r.fixup(Et(o))}return r},e.prototype.getDefinition=function(e,t,n,r){void 0===r&&(r=!1);var i=this.definitionsOf(t),o=i.get(e),a=!1;if(!o){var s=this.propertyNameOf(t);o=new ln(n.importExpr(e).prop(s)),i.set(e,o),a=!0}if(!a&&!o.shared||a&&r){var u=this.freshName();this.statements.push(Et(u).set(o.resolved).toDeclStmt(ge,[Ie.Final])),o.fixup(Et(u))}return o},e.prototype.getLiteralFactory=function(e){if(e instanceof Xe){var t=e.entries.map(function(e){return e.isConstant()?e:Pt(null)}),n=this.keyOf(At(t));return this._getLiteralFactory(n,e.entries,function(e){return At(e)})}var r=Tt(e.entries.map(function(e){return{key:e.key,value:e.value.isConstant()?e.value:Pt(null),quoted:e.quoted}}));n=this.keyOf(r);return this._getLiteralFactory(n,e.entries.map(function(e){return e.value}),function(t){return Tt(t.map(function(t,n){return{key:e.entries[n].key,value:t,quoted:e.entries[n].quoted}}))})},e.prototype._getLiteralFactory=function(e,t,n){var r=this,i=this.literalFactories.get(e),o=t.filter(function(e){return!e.isConstant()});if(!i){var a=t.map(function(e,t){return e.isConstant()?r.getConstLiteral(e,!0):Et("a"+t)}),s=Dt(a.filter(hn).map(function(e){return new Ge(e.name,me)}),[new at(n(a))],ge),u=this.freshName();this.statements.push(Et(u).set(s).toDeclStmt(ge,[Ie.Final])),i=Et(u),this.literalFactories.set(e,i)}return{literalFactory:i,literalFactoryArguments:o}},e.prototype.uniqueName=function(e){return""+e+this.nextNameIndex++},e.prototype.definitionsOf=function(e){switch(e){case 2:return this.componentDefinitions;case 1:return this.directiveDefinitions;case 0:return this.injectorDefinitions;case 3:return this.pipeDefinitions}return Wt("Unknown definition kind "+e),this.componentDefinitions},e.prototype.propertyNameOf=function(e){switch(e){case 2:return"ngComponentDef";case 1:return"ngDirectiveDef";case 0:return"ngInjectorDef";case 3:return"ngPipeDef"}return Wt("Unknown definition kind "+e),"<unknown>"},e.prototype.freshName=function(){return this.uniqueName("_c")},e.prototype.keyOf=function(e){return e.visitExpression(new pn,cn)},e}(),pn=function(){function e(){this.visitWrappedNodeExpr=fn,this.visitWriteVarExpr=fn,this.visitWriteKeyExpr=fn,this.visitWritePropExpr=fn,this.visitInvokeMethodExpr=fn,this.visitInvokeFunctionExpr=fn,this.visitInstantiateExpr=fn,this.visitConditionalExpr=fn,this.visitNotExpr=fn,this.visitAssertNotNullExpr=fn,this.visitCastExpr=fn,this.visitFunctionExpr=fn,this.visitBinaryOperatorExpr=fn,this.visitReadPropExpr=fn,this.visitReadKeyExpr=fn,this.visitCommaExpr=fn}return e.prototype.visitLiteralExpr=function(e){return""+("string"==typeof e.value?'"'+e.value+'"':e.value)},e.prototype.visitLiteralArrayExpr=function(e,t){var n=this;return"["+e.entries.map(function(e){return e.visitExpression(n,t)}).join(",")+"]"},e.prototype.visitLiteralMapExpr=function(e,t){var n=this;return"{"+e.entries.map(function(e){return(i=(r=e).quoted?'"':"")+r.key+i+":"+e.value.visitExpression(n,t);var r,i}).join(",")},e.prototype.visitExternalExpr=function(e){return e.value.moduleName?"EX:"+e.value.moduleName+":"+e.value.name:"EX:"+e.value.runtime.name},e.prototype.visitReadVarExpr=function(e){return"VAR:"+e.name},e.prototype.visitTypeofExpr=function(e,t){return"TYPEOF:"+e.expr.visitExpression(this,t)},e}();function fn(e){throw new Error("Invalid state: Visitor "+this.constructor.name+" doesn't handle "+e.constructor.name)}function hn(e){return e instanceof Se}var mn="@angular/core",gn=function(){function e(){}return e.ANALYZE_FOR_ENTRY_COMPONENTS={name:"ANALYZE_FOR_ENTRY_COMPONENTS",moduleName:mn},e.ElementRef={name:"ElementRef",moduleName:mn},e.NgModuleRef={name:"NgModuleRef",moduleName:mn},e.ViewContainerRef={name:"ViewContainerRef",moduleName:mn},e.ChangeDetectorRef={name:"ChangeDetectorRef",moduleName:mn},e.QueryList={name:"QueryList",moduleName:mn},e.TemplateRef={name:"TemplateRef",moduleName:mn},e.Renderer2={name:"Renderer2",moduleName:mn},e.CodegenComponentFactoryResolver={name:"ɵCodegenComponentFactoryResolver",moduleName:mn},e.ComponentFactoryResolver={name:"ComponentFactoryResolver",moduleName:mn},e.ComponentFactory={name:"ComponentFactory",moduleName:mn},e.ComponentRef={name:"ComponentRef",moduleName:mn},e.NgModuleFactory={name:"NgModuleFactory",moduleName:mn},e.createModuleFactory={name:"ɵcmf",moduleName:mn},e.moduleDef={name:"ɵmod",moduleName:mn},e.moduleProviderDef={name:"ɵmpd",moduleName:mn},e.RegisterModuleFactoryFn={name:"ɵregisterModuleFactory",moduleName:mn},e.inject={name:"inject",moduleName:mn},e.INJECTOR={name:"INJECTOR",moduleName:mn},e.Injector={name:"Injector",moduleName:mn},e.defineInjectable={name:"defineInjectable",moduleName:mn},e.InjectableDef={name:"ɵInjectableDef",moduleName:mn},e.ViewEncapsulation={name:"ViewEncapsulation",moduleName:mn},e.ChangeDetectionStrategy={name:"ChangeDetectionStrategy",moduleName:mn},e.SecurityContext={name:"SecurityContext",moduleName:mn},e.LOCALE_ID={name:"LOCALE_ID",moduleName:mn},e.TRANSLATIONS_FORMAT={name:"TRANSLATIONS_FORMAT",moduleName:mn},e.inlineInterpolate={name:"ɵinlineInterpolate",moduleName:mn},e.interpolate={name:"ɵinterpolate",moduleName:mn},e.EMPTY_ARRAY={name:"ɵEMPTY_ARRAY",moduleName:mn},e.EMPTY_MAP={name:"ɵEMPTY_MAP",moduleName:mn},e.Renderer={name:"Renderer",moduleName:mn},e.viewDef={name:"ɵvid",moduleName:mn},e.elementDef={name:"ɵeld",moduleName:mn},e.anchorDef={name:"ɵand",moduleName:mn},e.textDef={name:"ɵted",moduleName:mn},e.directiveDef={name:"ɵdid",moduleName:mn},e.providerDef={name:"ɵprd",moduleName:mn},e.queryDef={name:"ɵqud",moduleName:mn},e.pureArrayDef={name:"ɵpad",moduleName:mn},e.pureObjectDef={name:"ɵpod",moduleName:mn},e.purePipeDef={name:"ɵppd",moduleName:mn},e.pipeDef={name:"ɵpid",moduleName:mn},e.nodeValue={name:"ɵnov",moduleName:mn},e.ngContentDef={name:"ɵncd",moduleName:mn},e.unwrapValue={name:"ɵunv",moduleName:mn},e.createRendererType2={name:"ɵcrt",moduleName:mn},e.RendererType2={name:"RendererType2",moduleName:mn},e.ViewDefinition={name:"ɵViewDefinition",moduleName:mn},e.createComponentFactory={name:"ɵccf",moduleName:mn},e.setClassMetadata={name:"ɵsetClassMetadata",moduleName:mn},e}();function vn(e){return{identifier:{reference:e}}}function _n(e,t){return vn(e.resolveExternalReference(t))}var yn=function(){function e(e,t,n){this.filePath=e,this.name=t,this.members=n}return e.prototype.assertNoMembers=function(){if(this.members.length)throw new Error("Illegal state: symbol without members expected, but got "+JSON.stringify(this)+".")},e}(),bn=function(){function e(){this.cache=new Map}return e.prototype.get=function(e,t,n){var r='"'+e+'".'+t+((n=n||[]).length?"."+n.join("."):""),i=this.cache.get(r);return i||(i=new yn(e,t,n),this.cache.set(r,i)),i},e}(),jn=/^(?:(?:\[([^\]]+)\])|(?:\(([^\)]+)\)))|(\@[-\w]+)$/;function xn(e){return e.replace(/\W/g,"_")}var On,wn,Cn=0;function En(e){if(!e||!e.reference)return null;var t=e.reference;if(t instanceof yn)return t.name;if(t.__anonymousType)return t.__anonymousType;var n=tn(t);return n.indexOf("(")>=0?(n="anonymous_"+Cn++,t.__anonymousType=n):n=xn(n),n}function Sn(e){var t=e.reference;return t instanceof yn?t.filePath:"./"+tn(t)}function kn(e,t){return"View_"+En({reference:e})+"_"+t}function Mn(e){return"RenderType_"+En({reference:e})}function An(e){return"HostView_"+En({reference:e})}function Tn(e){return En({reference:e})+"NgFactory"}function Dn(e){return null!=e.value?xn(e.value):En(e.identifier)}function In(e){return null!=e.identifier?e.identifier.reference:e.value}(wn=On||(On={}))[wn.Pipe=0]="Pipe",wn[wn.Directive=1]="Directive",wn[wn.NgModule=2]="NgModule",wn[wn.Injectable=3]="Injectable";var Pn=function(){return function(e){var t=void 0===e?{}:e,n=t.moduleUrl,r=t.styles,i=t.styleUrls;this.moduleUrl=n||null,this.styles=zn(r),this.styleUrls=zn(i)}}(),Rn=function(){function e(e){var t=e.encapsulation,n=e.template,r=e.templateUrl,i=e.htmlAst,o=e.styles,a=e.styleUrls,s=e.externalStylesheets,u=e.animations,c=e.ngContentSelectors,l=e.interpolation,d=e.isInline,p=e.preserveWhitespaces;if(this.encapsulation=t,this.template=n,this.templateUrl=r,this.htmlAst=i,this.styles=zn(o),this.styleUrls=zn(a),this.externalStylesheets=zn(s),this.animations=u?Hn(u):[],this.ngContentSelectors=c||[],l&&2!=l.length)throw new Error("'interpolation' should have a start and an end symbol.");this.interpolation=l,this.isInline=d,this.preserveWhitespaces=p}return e.prototype.toSummary=function(){return{ngContentSelectors:this.ngContentSelectors,encapsulation:this.encapsulation,styles:this.styles,animations:this.animations}},e}(),Nn=function(){function e(e){var t=e.isHost,n=e.type,r=e.isComponent,i=e.selector,o=e.exportAs,a=e.changeDetection,s=e.inputs,u=e.outputs,c=e.hostListeners,l=e.hostProperties,d=e.hostAttributes,p=e.providers,f=e.viewProviders,h=e.queries,m=e.guards,g=e.viewQueries,v=e.entryComponents,_=e.template,y=e.componentViewType,b=e.rendererType,j=e.componentFactory;this.isHost=!!t,this.type=n,this.isComponent=r,this.selector=i,this.exportAs=o,this.changeDetection=a,this.inputs=s,this.outputs=u,this.hostListeners=c,this.hostProperties=l,this.hostAttributes=d,this.providers=zn(p),this.viewProviders=zn(f),this.queries=zn(h),this.guards=m,this.viewQueries=zn(g),this.entryComponents=zn(v),this.template=_,this.componentViewType=y,this.rendererType=b,this.componentFactory=j}return e.create=function(t){var n=t.isHost,r=t.type,i=t.isComponent,o=t.selector,a=t.exportAs,s=t.changeDetection,u=t.inputs,c=t.outputs,l=t.host,d=t.providers,p=t.viewProviders,f=t.queries,h=t.guards,m=t.viewQueries,g=t.entryComponents,v=t.template,_=t.componentViewType,y=t.rendererType,b=t.componentFactory,j={},x={},O={};null!=l&&Object.keys(l).forEach(function(e){var t=l[e],n=e.match(jn);null===n?O[e]=t:null!=n[1]?x[n[1]]=t:null!=n[2]&&(j[n[2]]=t)});var w={};null!=u&&u.forEach(function(e){var t=Ft(e,[e,e]);w[t[0]]=t[1]});var C={};return null!=c&&c.forEach(function(e){var t=Ft(e,[e,e]);C[t[0]]=t[1]}),new e({isHost:n,type:r,isComponent:!!i,selector:o,exportAs:a,changeDetection:s,inputs:w,outputs:C,hostListeners:j,hostProperties:x,hostAttributes:O,providers:d,viewProviders:p,queries:f,guards:h,viewQueries:m,entryComponents:g,template:v,componentViewType:_,rendererType:y,componentFactory:b})},e.prototype.toSummary=function(){return{summaryKind:On.Directive,type:this.type,isComponent:this.isComponent,selector:this.selector,exportAs:this.exportAs,inputs:this.inputs,outputs:this.outputs,hostListeners:this.hostListeners,hostProperties:this.hostProperties,hostAttributes:this.hostAttributes,providers:this.providers,viewProviders:this.viewProviders,queries:this.queries,guards:this.guards,viewQueries:this.viewQueries,entryComponents:this.entryComponents,changeDetection:this.changeDetection,template:this.template&&this.template.toSummary(),componentViewType:this.componentViewType,rendererType:this.rendererType,componentFactory:this.componentFactory}},e}(),Fn=function(){function e(e){var t=e.type,n=e.name,r=e.pure;this.type=t,this.name=n,this.pure=!!r}return e.prototype.toSummary=function(){return{summaryKind:On.Pipe,type:this.type,name:this.name,pure:this.pure}},e}(),Ln=function(){return function(){}}(),Vn=function(){function e(e){var t=e.type,n=e.providers,r=e.declaredDirectives,i=e.exportedDirectives,o=e.declaredPipes,a=e.exportedPipes,s=e.entryComponents,u=e.bootstrapComponents,c=e.importedModules,l=e.exportedModules,d=e.schemas,p=e.transitiveModule,f=e.id;this.type=t||null,this.declaredDirectives=zn(r),this.exportedDirectives=zn(i),this.declaredPipes=zn(o),this.exportedPipes=zn(a),this.providers=zn(n),this.entryComponents=zn(s),this.bootstrapComponents=zn(u),this.importedModules=zn(c),this.exportedModules=zn(l),this.schemas=zn(d),this.id=f||null,this.transitiveModule=p||null}return e.prototype.toSummary=function(){var e=this.transitiveModule;return{summaryKind:On.NgModule,type:this.type,entryComponents:e.entryComponents,providers:e.providers,modules:e.modules,exportedDirectives:e.exportedDirectives,exportedPipes:e.exportedPipes}},e}(),Bn=function(){function e(){this.directivesSet=new Set,this.directives=[],this.exportedDirectivesSet=new Set,this.exportedDirectives=[],this.pipesSet=new Set,this.pipes=[],this.exportedPipesSet=new Set,this.exportedPipes=[],this.modulesSet=new Set,this.modules=[],this.entryComponentsSet=new Set,this.entryComponents=[],this.providers=[]}return e.prototype.addProvider=function(e,t){this.providers.push({provider:e,module:t})},e.prototype.addDirective=function(e){this.directivesSet.has(e.reference)||(this.directivesSet.add(e.reference),this.directives.push(e))},e.prototype.addExportedDirective=function(e){this.exportedDirectivesSet.has(e.reference)||(this.exportedDirectivesSet.add(e.reference),this.exportedDirectives.push(e))},e.prototype.addPipe=function(e){this.pipesSet.has(e.reference)||(this.pipesSet.add(e.reference),this.pipes.push(e))},e.prototype.addExportedPipe=function(e){this.exportedPipesSet.has(e.reference)||(this.exportedPipesSet.add(e.reference),this.exportedPipes.push(e))},e.prototype.addModule=function(e){this.modulesSet.has(e.reference)||(this.modulesSet.add(e.reference),this.modules.push(e))},e.prototype.addEntryComponent=function(e){this.entryComponentsSet.has(e.componentType)||(this.entryComponentsSet.add(e.componentType),this.entryComponents.push(e))},e}();function zn(e){return e||[]}var Un=function(){return function(e,t){var n=t.useClass,r=t.useValue,i=t.useExisting,o=t.useFactory,a=t.deps,s=t.multi;this.token=e,this.useClass=n||null,this.useValue=r,this.useExisting=i,this.useFactory=o||null,this.dependencies=a||null,this.multi=!!s}}();function Hn(e){return e.reduce(function(e,t){var n=Array.isArray(t)?Hn(t):t;return e.concat(n)},[])}function qn(e){return e.replace(/(\w+:\/\/[\w:-]+)?(\/+)?/,"ng:///")}function Gn(e,t,n){var r;return r=n.isInline?t.type.reference instanceof yn?t.type.reference.filePath+"."+t.type.reference.name+".html":En(e)+"/"+En(t.type)+".html":n.templateUrl,t.type.reference instanceof yn?r:qn(r)}function Wn(e,t){var n=e.moduleUrl.split(/\/\\/g);return qn("css/"+t+n[n.length-1]+".ngstyle.js")}function Yn(e){return qn(En(e.type)+"/module.ngfactory.js")}function Kn(e,t){return qn(En(e)+"/"+En(t.type)+".ngfactory.js")}var Qn="@angular/core",Xn=function(){function e(){}return e.NEW_METHOD="factory",e.TRANSFORM_METHOD="transform",e.PATCH_DEPS="patchedDeps",e.namespaceHTML={name:"ɵnamespaceHTML",moduleName:Qn},e.namespaceMathML={name:"ɵnamespaceMathML",moduleName:Qn},e.namespaceSVG={name:"ɵnamespaceSVG",moduleName:Qn},e.element={name:"ɵelement",moduleName:Qn},e.elementStart={name:"ɵelementStart",moduleName:Qn},e.elementEnd={name:"ɵelementEnd",moduleName:Qn},e.elementProperty={name:"ɵelementProperty",moduleName:Qn},e.componentHostSyntheticProperty={name:"ɵcomponentHostSyntheticProperty",moduleName:Qn},e.elementAttribute={name:"ɵelementAttribute",moduleName:Qn},e.elementClassProp={name:"ɵelementClassProp",moduleName:Qn},e.elementContainerStart={name:"ɵelementContainerStart",moduleName:Qn},e.elementContainerEnd={name:"ɵelementContainerEnd",moduleName:Qn},e.elementStyling={name:"ɵelementStyling",moduleName:Qn},e.elementHostAttrs={name:"ɵelementHostAttrs",moduleName:Qn},e.elementStylingMap={name:"ɵelementStylingMap",moduleName:Qn},e.elementStyleProp={name:"ɵelementStyleProp",moduleName:Qn},e.elementStylingApply={name:"ɵelementStylingApply",moduleName:Qn},e.containerCreate={name:"ɵcontainer",moduleName:Qn},e.nextContext={name:"ɵnextContext",moduleName:Qn},e.templateCreate={name:"ɵtemplate",moduleName:Qn},e.text={name:"ɵtext",moduleName:Qn},e.textBinding={name:"ɵtextBinding",moduleName:Qn},e.bind={name:"ɵbind",moduleName:Qn},e.enableBindings={name:"ɵenableBindings",moduleName:Qn},e.disableBindings={name:"ɵdisableBindings",moduleName:Qn},e.allocHostVars={name:"ɵallocHostVars",moduleName:Qn},e.getCurrentView={name:"ɵgetCurrentView",moduleName:Qn},e.restoreView={name:"ɵrestoreView",moduleName:Qn},e.interpolation1={name:"ɵinterpolation1",moduleName:Qn},e.interpolation2={name:"ɵinterpolation2",moduleName:Qn},e.interpolation3={name:"ɵinterpolation3",moduleName:Qn},e.interpolation4={name:"ɵinterpolation4",moduleName:Qn},e.interpolation5={name:"ɵinterpolation5",moduleName:Qn},e.interpolation6={name:"ɵinterpolation6",moduleName:Qn},e.interpolation7={name:"ɵinterpolation7",moduleName:Qn},e.interpolation8={name:"ɵinterpolation8",moduleName:Qn},e.interpolationV={name:"ɵinterpolationV",moduleName:Qn},e.pureFunction0={name:"ɵpureFunction0",moduleName:Qn},e.pureFunction1={name:"ɵpureFunction1",moduleName:Qn},e.pureFunction2={name:"ɵpureFunction2",moduleName:Qn},e.pureFunction3={name:"ɵpureFunction3",moduleName:Qn},e.pureFunction4={name:"ɵpureFunction4",moduleName:Qn},e.pureFunction5={name:"ɵpureFunction5",moduleName:Qn},e.pureFunction6={name:"ɵpureFunction6",moduleName:Qn},e.pureFunction7={name:"ɵpureFunction7",moduleName:Qn},e.pureFunction8={name:"ɵpureFunction8",moduleName:Qn},e.pureFunctionV={name:"ɵpureFunctionV",moduleName:Qn},e.pipeBind1={name:"ɵpipeBind1",moduleName:Qn},e.pipeBind2={name:"ɵpipeBind2",moduleName:Qn},e.pipeBind3={name:"ɵpipeBind3",moduleName:Qn},e.pipeBind4={name:"ɵpipeBind4",moduleName:Qn},e.pipeBindV={name:"ɵpipeBindV",moduleName:Qn},e.i18n={name:"ɵi18n",moduleName:Qn},e.i18nAttributes={name:"ɵi18nAttributes",moduleName:Qn},e.i18nExp={name:"ɵi18nExp",moduleName:Qn},e.i18nStart={name:"ɵi18nStart",moduleName:Qn},e.i18nEnd={name:"ɵi18nEnd",moduleName:Qn},e.i18nApply={name:"ɵi18nApply",moduleName:Qn},e.i18nPostprocess={name:"ɵi18nPostprocess",moduleName:Qn},e.load={name:"ɵload",moduleName:Qn},e.loadQueryList={name:"ɵloadQueryList",moduleName:Qn},e.pipe={name:"ɵpipe",moduleName:Qn},e.projection={name:"ɵprojection",moduleName:Qn},e.projectionDef={name:"ɵprojectionDef",moduleName:Qn},e.reference={name:"ɵreference",moduleName:Qn},e.inject={name:"inject",moduleName:Qn},e.injectAttribute={name:"ɵinjectAttribute",moduleName:Qn},e.directiveInject={name:"ɵdirectiveInject",moduleName:Qn},e.templateRefExtractor={name:"ɵtemplateRefExtractor",moduleName:Qn},e.defineBase={name:"ɵdefineBase",moduleName:Qn},e.BaseDef={name:"ɵBaseDef",moduleName:Qn},e.defineComponent={name:"ɵdefineComponent",moduleName:Qn},e.ComponentDefWithMeta={name:"ɵComponentDefWithMeta",moduleName:Qn},e.defineDirective={name:"ɵdefineDirective",moduleName:Qn},e.DirectiveDefWithMeta={name:"ɵDirectiveDefWithMeta",moduleName:Qn},e.InjectorDef={name:"ɵInjectorDef",moduleName:Qn},e.defineInjector={name:"defineInjector",moduleName:Qn},e.NgModuleDefWithMeta={name:"ɵNgModuleDefWithMeta",moduleName:Qn},e.defineNgModule={name:"ɵdefineNgModule",moduleName:Qn},e.PipeDefWithMeta={name:"ɵPipeDefWithMeta",moduleName:Qn},e.definePipe={name:"ɵdefinePipe",moduleName:Qn},e.query={name:"ɵquery",moduleName:Qn},e.queryRefresh={name:"ɵqueryRefresh",moduleName:Qn},e.registerContentQuery={name:"ɵregisterContentQuery",moduleName:Qn},e.NgOnChangesFeature={name:"ɵNgOnChangesFeature",moduleName:Qn},e.InheritDefinitionFeature={name:"ɵInheritDefinitionFeature",moduleName:Qn},e.ProvidersFeature={name:"ɵProvidersFeature",moduleName:Qn},e.listener={name:"ɵlistener",moduleName:Qn},e.getFactoryOf={name:"ɵgetFactoryOf",moduleName:Qn},e.getInheritedFactory={name:"ɵgetInheritedFactory",moduleName:Qn},e.sanitizeHtml={name:"ɵsanitizeHtml",moduleName:Qn},e.sanitizeStyle={name:"ɵsanitizeStyle",moduleName:Qn},e.defaultStyleSanitizer={name:"ɵdefaultStyleSanitizer",moduleName:Qn},e.sanitizeResourceUrl={name:"ɵsanitizeResourceUrl",moduleName:Qn},e.sanitizeScript={name:"ɵsanitizeScript",moduleName:Qn},e.sanitizeUrl={name:"ɵsanitizeUrl",moduleName:Qn},e}(),Zn=function(){return function(e,t,n,r,i,o){this.nodes=e,this.placeholders=t,this.placeholderToMessage=n,this.meaning=r,this.description=i,this.id=o,e.length?this.sources=[{filePath:e[0].sourceSpan.start.file.url,startLine:e[0].sourceSpan.start.line+1,startCol:e[0].sourceSpan.start.col+1,endLine:e[e.length-1].sourceSpan.end.line+1,endCol:e[0].sourceSpan.start.col+1}]:this.sources=[]}}(),$n=function(){function e(e,t){this.value=e,this.sourceSpan=t}return e.prototype.visit=function(e,t){return e.visitText(this,t)},e}(),Jn=function(){function e(e,t){this.children=e,this.sourceSpan=t}return e.prototype.visit=function(e,t){return e.visitContainer(this,t)},e}(),er=function(){function e(e,t,n,r){this.expression=e,this.type=t,this.cases=n,this.sourceSpan=r}return e.prototype.visit=function(e,t){return e.visitIcu(this,t)},e}(),tr=function(){function e(e,t,n,r,i,o,a){this.tag=e,this.attrs=t,this.startName=n,this.closeName=r,this.children=i,this.isVoid=o,this.sourceSpan=a}return e.prototype.visit=function(e,t){return e.visitTagPlaceholder(this,t)},e}(),nr=function(){function e(e,t,n){this.value=e,this.name=t,this.sourceSpan=n}return e.prototype.visit=function(e,t){return e.visitPlaceholder(this,t)},e}(),rr=function(){function e(e,t,n){this.value=e,this.name=t,this.sourceSpan=n}return e.prototype.visit=function(e,t){return e.visitIcuPlaceholder(this,t)},e}(),ir=function(){function e(){}return e.prototype.visitText=function(e,t){return new $n(e.value,e.sourceSpan)},e.prototype.visitContainer=function(e,t){var n=this,r=e.children.map(function(e){return e.visit(n,t)});return new Jn(r,e.sourceSpan)},e.prototype.visitIcu=function(e,t){var n=this,r={};Object.keys(e.cases).forEach(function(i){return r[i]=e.cases[i].visit(n,t)});var i=new er(e.expression,e.type,r,e.sourceSpan);return i.expressionPlaceholder=e.expressionPlaceholder,i},e.prototype.visitTagPlaceholder=function(e,t){var n=this,r=e.children.map(function(e){return e.visit(n,t)});return new tr(e.tag,e.attrs,e.startName,e.closeName,r,e.isVoid,e.sourceSpan)},e.prototype.visitPlaceholder=function(e,t){return new nr(e.value,e.name,e.sourceSpan)},e.prototype.visitIcuPlaceholder=function(e,t){return new rr(e.value,e.name,e.sourceSpan)},e}(),or=function(){function e(){}return e.prototype.visitText=function(e,t){},e.prototype.visitContainer=function(e,t){var n=this;e.children.forEach(function(e){return e.visit(n)})},e.prototype.visitIcu=function(e,t){var n=this;Object.keys(e.cases).forEach(function(t){e.cases[t].visit(n)})},e.prototype.visitTagPlaceholder=function(e,t){var n=this;e.children.forEach(function(e){return e.visit(n)})},e.prototype.visitPlaceholder=function(e,t){},e.prototype.visitIcuPlaceholder=function(e,t){},e}();function ar(e){return e.id||function(e){var t,n,r=en(e),i=function(e,t){for(var n=Array(e.length+3>>>2),r=0;r<n.length;r++)n[r]=xr(e,4*r,t);return n}(r,lr.Big),a=8*r.length,s=new Array(80),u=Object(o.__read)([1732584193,4023233417,2562383102,271733878,3285377520],5),c=u[0],l=u[1],d=u[2],p=u[3],f=u[4];i[a>>5]|=128<<24-a%32,i[15+(a+64>>9<<4)]=a;for(var h=0;h<i.length;h+=16){for(var m=Object(o.__read)([c,l,d,p,f],5),g=m[0],v=m[1],_=m[2],y=m[3],b=m[4],j=0;j<80;j++){s[j]=j<16?i[h+j]:br(s[j-3]^s[j-8]^s[j-14]^s[j-16],1);var x=Object(o.__read)(fr(j,l,d,p),2),O=x[0],w=x[1],C=[br(c,5),O,f,w,s[j]].reduce(vr);t=Object(o.__read)([p,d,br(l,30),c,C],5),f=t[0],p=t[1],d=t[2],l=t[3],c=t[4]}n=Object(o.__read)([vr(c,g),vr(l,v),vr(d,_),vr(p,y),vr(f,b)],5),c=n[0],l=n[1],d=n[2],p=n[3],f=n[4]}return function(e){for(var t="",n=0;n<e.length;n++){var r=jr(e,n);t+=(r>>>4).toString(16)+(15&r).toString(16)}return t.toLowerCase()}(Or([c,l,d,p,f]))}((t=e.nodes,t.map(function(e){return e.visit(cr,null)})).join("")+"["+e.meaning+"]");var t}function sr(e){if(e.id)return e.id;var t=new pr;return function(e,t){var n,r=Object(o.__read)(hr(e),2),i=r[0],a=r[1];if(t){var s=Object(o.__read)(hr(t),2),u=s[0],c=s[1];n=Object(o.__read)((j=[i,a],x=1,O=Object(o.__read)(j,2),w=O[0],C=O[1],l=[w<<x|C>>>32-x,C<<x|w>>>32-x],d=[u,c],p=Object(o.__read)(l,2),f=p[0],h=p[1],m=Object(o.__read)(d,2),g=m[0],v=m[1],_=Object(o.__read)(_r(h,v),2),y=_[0],b=_[1],[vr(vr(f,g),y),b]),2),i=n[0],a=n[1]}var l,d,p,f,h,m,g,v,_,y,b;var j,x,O,w,C;return function(e){for(var t="",n="1",r=e.length-1;r>=0;r--)t=wr(t,Cr(jr(e,r),n)),n=Cr(256,n);return t.split("").reverse().join("")}(Or([2147483647&i,a]))}(e.nodes.map(function(e){return e.visit(t,null)}).join(""),e.meaning)}var ur=function(){function e(){}return e.prototype.visitText=function(e,t){return e.value},e.prototype.visitContainer=function(e,t){var n=this;return"["+e.children.map(function(e){return e.visit(n)}).join(", ")+"]"},e.prototype.visitIcu=function(e,t){var n=this,r=Object.keys(e.cases).map(function(t){return t+" {"+e.cases[t].visit(n)+"}"});return"{"+e.expression+", "+e.type+", "+r.join(", ")+"}"},e.prototype.visitTagPlaceholder=function(e,t){var n=this;return e.isVoid?'<ph tag name="'+e.startName+'"/>':'<ph tag name="'+e.startName+'">'+e.children.map(function(e){return e.visit(n)}).join(", ")+'</ph name="'+e.closeName+'">'},e.prototype.visitPlaceholder=function(e,t){return e.value?'<ph name="'+e.name+'">'+e.value+"</ph>":'<ph name="'+e.name+'"/>'},e.prototype.visitIcuPlaceholder=function(e,t){return'<ph icu name="'+e.name+'">'+e.value.visit(this)+"</ph>"},e}(),cr=new ur;var lr,dr,pr=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(o.__extends)(t,e),t.prototype.visitIcu=function(e,t){var n=this,r=Object.keys(e.cases).map(function(t){return t+" {"+e.cases[t].visit(n)+"}"});return"{"+e.type+", "+r.join(", ")+"}"},t}(ur);function fr(e,t,n,r){return e<20?[t&n|~t&r,1518500249]:e<40?[t^n^r,1859775393]:e<60?[t&n|t&r|n&r,2400959708]:[t^n^r,3395469782]}function hr(e){var t=en(e),n=Object(o.__read)([mr(t,0),mr(t,102072)],2),r=n[0],i=n[1];return 0!=r||0!=i&&1!=i||(r^=319790063,i^=-1801410264),[r,i]}function mr(e,t){var n,r,i=Object(o.__read)([2654435769,2654435769],2),a=i[0],s=i[1],u=e.length;for(r=0;r+12<=u;r+=12)a=vr(a,xr(e,r,lr.Little)),s=vr(s,xr(e,r+4,lr.Little)),t=vr(t,xr(e,r+8,lr.Little)),a=(n=Object(o.__read)(gr([a,s,t]),3))[0],s=n[1],t=n[2];return gr([a=vr(a,xr(e,r,lr.Little)),s=vr(s,xr(e,r+4,lr.Little)),t=vr(t=vr(t,u),xr(e,r+8,lr.Little)<<8)])[2]}function gr(e){var t=Object(o.__read)(e,3),n=t[0],r=t[1],i=t[2];return n=yr(n=yr(n,r),i),n^=i>>>13,r=yr(r=yr(r,i),n),r^=n<<8,i=yr(i=yr(i,n),r),i^=r>>>13,n=yr(n=yr(n,r),i),n^=i>>>12,r=yr(r=yr(r,i),n),r^=n<<16,i=yr(i=yr(i,n),r),i^=r>>>5,n=yr(n=yr(n,r),i),n^=i>>>3,r=yr(r=yr(r,i),n),r^=n<<10,i=yr(i=yr(i,n),r),[n,r,i^=r>>>15]}function vr(e,t){return _r(e,t)[1]}function _r(e,t){var n=(65535&e)+(65535&t),r=(e>>>16)+(t>>>16)+(n>>>16);return[r>>>16,r<<16|65535&n]}function yr(e,t){var n=(65535&e)-(65535&t);return(e>>16)-(t>>16)+(n>>16)<<16|65535&n}function br(e,t){return e<<t|e>>>32-t}function jr(e,t){return t>=e.length?0:255&e.charCodeAt(t)}function xr(e,t,n){var r=0;if(n===lr.Big)for(var i=0;i<4;i++)r+=jr(e,t+i)<<24-8*i;else for(i=0;i<4;i++)r+=jr(e,t+i)<<8*i;return r}function Or(e){return e.reduce(function(e,t){return e+function(e){for(var t="",n=0;n<4;n++)t+=String.fromCharCode(e>>>8*(3-n)&255);return t}(t)},"")}function wr(e,t){for(var n="",r=Math.max(e.length,t.length),i=0,o=0;i<r||o;i++){var a=o+ +(e[i]||0)+ +(t[i]||0);a>=10?(o=1,n+=a-10):(o=0,n+=a)}return n}function Cr(e,t){for(var n="",r=t;0!==e;e>>>=1)1&e&&(n=wr(n,r)),r=wr(r,r);return n}(dr=lr||(lr={}))[dr.Little=0]="Little",dr[dr.Big=1]="Big";var Er=function(){function e(){}return e.prototype.createNameMapper=function(e){return null},e}(),Sr=function(e){function t(t,n){var r=e.call(this)||this;return r.mapName=n,r.internalToPublic={},r.publicToNextId={},r.publicToInternal={},t.nodes.forEach(function(e){return e.visit(r)}),r}return Object(o.__extends)(t,e),t.prototype.toPublicName=function(e){return this.internalToPublic.hasOwnProperty(e)?this.internalToPublic[e]:null},t.prototype.toInternalName=function(e){return this.publicToInternal.hasOwnProperty(e)?this.publicToInternal[e]:null},t.prototype.visitText=function(e,t){return null},t.prototype.visitTagPlaceholder=function(t,n){this.visitPlaceholderName(t.startName),e.prototype.visitTagPlaceholder.call(this,t,n),this.visitPlaceholderName(t.closeName)},t.prototype.visitPlaceholder=function(e,t){this.visitPlaceholderName(e.name)},t.prototype.visitIcuPlaceholder=function(e,t){this.visitPlaceholderName(e.name)},t.prototype.visitPlaceholderName=function(e){if(e&&!this.internalToPublic.hasOwnProperty(e)){var t=this.mapName(e);if(this.publicToInternal.hasOwnProperty(t)){var n=this.publicToNextId[t];this.publicToNextId[t]=n+1,t=t+"_"+n}else this.publicToNextId[t]=1;this.internalToPublic[e]=t,this.publicToInternal[t]=e}},t}(or),kr=new(function(){function e(){}return e.prototype.visitTag=function(e){var t=this,n=this._serializeAttributes(e.attrs);if(0==e.children.length)return"<"+e.name+n+"/>";var r=e.children.map(function(e){return e.visit(t)});return"<"+e.name+n+">"+r.join("")+"</"+e.name+">"},e.prototype.visitText=function(e){return e.value},e.prototype.visitDeclaration=function(e){return"<?xml"+this._serializeAttributes(e.attrs)+" ?>"},e.prototype._serializeAttributes=function(e){var t=Object.keys(e).map(function(t){return t+'="'+e[t]+'"'}).join(" ");return t.length>0?" "+t:""},e.prototype.visitDoctype=function(e){return"<!DOCTYPE "+e.rootTag+" [\n"+e.dtd+"\n]>"},e}());function Mr(e){return e.map(function(e){return e.visit(kr)}).join("")}var Ar=function(){function e(e){var t=this;this.attrs={},Object.keys(e).forEach(function(n){t.attrs[n]=Nr(e[n])})}return e.prototype.visit=function(e){return e.visitDeclaration(this)},e}(),Tr=function(){function e(e,t){this.rootTag=e,this.dtd=t}return e.prototype.visit=function(e){return e.visitDoctype(this)},e}(),Dr=function(){function e(e,t,n){void 0===t&&(t={}),void 0===n&&(n=[]);var r=this;this.name=e,this.children=n,this.attrs={},Object.keys(t).forEach(function(e){r.attrs[e]=Nr(t[e])})}return e.prototype.visit=function(e){return e.visitTag(this)},e}(),Ir=function(){function e(e){this.value=Nr(e)}return e.prototype.visit=function(e){return e.visitText(this)},e}(),Pr=function(e){function t(t){return void 0===t&&(t=0),e.call(this,"\n"+new Array(t+1).join(" "))||this}return Object(o.__extends)(t,e),t}(Ir),Rr=[[/&/g,"&amp;"],[/"/g,"&quot;"],[/'/g,"&apos;"],[/</g,"&lt;"],[/>/g,"&gt;"]];function Nr(e){return Rr.reduce(function(e,t){return e.replace(t[0],t[1])},e)}var Fr=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(o.__extends)(t,e),t.prototype.write=function(e,t){var n=new Br,r=new Lr,i=new Dr("messagebundle");return e.forEach(function(e){var t={id:e.id};e.description&&(t.desc=e.description),e.meaning&&(t.meaning=e.meaning);var n=[];e.sources.forEach(function(e){n.push(new Dr("source",{},[new Ir(e.filePath+":"+e.startLine+(e.endLine!==e.startLine?","+e.endLine:""))]))}),i.children.push(new Pr(2),new Dr("msg",t,Object(o.__spread)(n,r.serialize(e.nodes))))}),i.children.push(new Pr),Mr([new Ar({version:"1.0",encoding:"UTF-8"}),new Pr,new Tr("messagebundle",'<!ELEMENT messagebundle (msg)*>\n<!ATTLIST messagebundle class CDATA #IMPLIED>\n\n<!ELEMENT msg (#PCDATA|ph|source)*>\n<!ATTLIST msg id CDATA #IMPLIED>\n<!ATTLIST msg seq CDATA #IMPLIED>\n<!ATTLIST msg name CDATA #IMPLIED>\n<!ATTLIST msg desc CDATA #IMPLIED>\n<!ATTLIST msg meaning CDATA #IMPLIED>\n<!ATTLIST msg obsolete (obsolete) #IMPLIED>\n<!ATTLIST msg xml:space (default|preserve) "default">\n<!ATTLIST msg is_hidden CDATA #IMPLIED>\n\n<!ELEMENT source (#PCDATA)>\n\n<!ELEMENT ph (#PCDATA|ex)*>\n<!ATTLIST ph name CDATA #REQUIRED>\n\n<!ELEMENT ex (#PCDATA)>'),new Pr,n.addDefaultExamples(i),new Pr])},t.prototype.load=function(e,t){throw new Error("Unsupported")},t.prototype.digest=function(e){return Vr(e)},t.prototype.createNameMapper=function(e){return new Sr(e,zr)},t}(Er),Lr=function(){function e(){}return e.prototype.visitText=function(e,t){return[new Ir(e.value)]},e.prototype.visitContainer=function(e,t){var n=this,r=[];return e.children.forEach(function(e){return r.push.apply(r,Object(o.__spread)(e.visit(n)))}),r},e.prototype.visitIcu=function(e,t){var n=this,r=[new Ir("{"+e.expressionPlaceholder+", "+e.type+", ")];return Object.keys(e.cases).forEach(function(t){r.push.apply(r,Object(o.__spread)([new Ir(t+" {")],e.cases[t].visit(n),[new Ir("} ")]))}),r.push(new Ir("}")),r},e.prototype.visitTagPlaceholder=function(e,t){var n=new Ir("<"+e.tag+">"),r=new Dr("ex",{},[n]),i=new Dr("ph",{name:e.startName},[r,n]);if(e.isVoid)return[i];var a=new Ir("</"+e.tag+">"),s=new Dr("ex",{},[a]),u=new Dr("ph",{name:e.closeName},[s,a]);return Object(o.__spread)([i],this.serialize(e.children),[u])},e.prototype.visitPlaceholder=function(e,t){var n=new Ir("{{"+e.value+"}}"),r=new Dr("ex",{},[n]);return[new Dr("ph",{name:e.name},[r,n])]},e.prototype.visitIcuPlaceholder=function(e,t){var n=e.value.expression,r=e.value.type,i=Object.keys(e.value.cases).map(function(e){return e+" {...}"}).join(" "),o=new Ir("{"+n+", "+r+", "+i+"}"),a=new Dr("ex",{},[o]);return[new Dr("ph",{name:e.name},[a,o])]},e.prototype.serialize=function(e){var t=this;return[].concat.apply([],Object(o.__spread)(e.map(function(e){return e.visit(t)})))},e}();function Vr(e){return sr(e)}var Br=function(){function e(){}return e.prototype.addDefaultExamples=function(e){return e.visit(this),e},e.prototype.visitTag=function(e){var t=this;if("ph"===e.name){if(!e.children||0==e.children.length){var n=new Ir(e.attrs.name||"...");e.children=[new Dr("ex",{},[n])]}}else e.children&&e.children.forEach(function(e){return e.visit(t)})},e.prototype.visitText=function(e){},e.prototype.visitDeclaration=function(e){},e.prototype.visitDoctype=function(e){},e}();function zr(e){return e.toUpperCase().replace(/[^A-Z0-9_]/g,"_")}function Ur(e,t){return void 0===t&&(t=!1),Tt(Object.keys(e).map(function(n){return{key:n,quoted:t,value:e[n]}}))}var Hr="MSG_",qr=new RegExp("^"+Hr),Gr="I18N_",Wr="goog.getMsg",Yr="|",Kr="@@",Qr="i18n",Xr="i18n-",Zr="�";function $r(e,t,n){var r=[Pt(t)];n&&Object.keys(n).length&&r.push(Ur(n,!0));var i=Et(Wr).callFn(r);return e.set(i).toDeclStmt(ge,[Ie.Final])}function Jr(e){return e===Qr||e.startsWith(Xr)}function ei(e){return e instanceof Zn}function ti(e){return ei(e)&&1===e.nodes.length&&e.nodes[0]instanceof er}function ni(e,t){return void 0===t&&(t=null),{id:"string"==typeof t?t:e.id||"",meaning:e.meaning||"",description:e.description||""}}function ri(e){return e.nodes[0]}function ii(e,t){return void 0===t&&(t=0),""+Zr+e+(t>0?":"+t:"")+Zr}function oi(e){var t={};return e.forEach(function(e,n){t[n]=Pt(e.length>1?"["+e.join("|")+"]":e[0])}),t}function ai(e,t){for(var n=[],r=2;r<arguments.length;r++)n[r-2]=arguments[r];var i=e.get(t)||[];i.push.apply(i,Object(o.__spread)(n)),e.set(t,i)}function si(e,t,n){void 0===t&&(t=0),void 0===n&&(n=0);var r=t,i=new Map,o=e instanceof Zn?e.nodes.find(function(e){return e instanceof Jn}):e;return o&&o.children.filter(function(e){return e instanceof nr}).forEach(function(e,t){var o=ii(r+t,n);ai(i,e.name,o)}),i}function ui(e){var t,n=zr(e).split("_");if(1===n.length)return e.toLowerCase();/^\d+$/.test(n[n.length-1])&&(t=n.pop());var r=n.shift().toLowerCase();return n.length&&(r+=n.map(function(e){return e.charAt(0).toUpperCase()+e.slice(1).toLowerCase()}).join("")),t?r+"_"+t:r}function ci(e){return(""+Hr+e).toUpperCase()}function li(e,t,n,r,i){void 0===r&&(r={});var o,a,s=[],u=(a=[],(o=n).description&&a.push({tagName:"desc",text:o.description}),o.meaning&&a.push({tagName:"meaning",text:o.meaning}),0==a.length?null:new ht(a));if(u&&s.push(u),i){s.push($r(e,t,r));var c=Et(e.name);e.name=e.name.replace(qr,Gr),s.push(e.set(i(c)).toDeclStmt(ge,[Ie.Final]))}else s.push($r(e,t,r));return s}var di="_t",pi="ctx",fi="rf";function hi(e,t){var n=null;return function(){return n||(e.push(new rt(di,void 0,me)),n=Et(t)),n}}function mi(e){if(this)throw new Error("Builder "+this.constructor.name+" doesn't support "+e+" yet");throw new Error("Feature "+e+" is not supported yet")}function gi(e){throw new Error("Invalid state: Visitor "+this.constructor.name+" doesn't handle "+void 0)}function vi(e){return Array.isArray(e)?At(e.map(vi)):Pt(e,ge)}function _i(e,t){return Object.getOwnPropertyNames(e).length>0?(n=e,r=t,Tt(Object.getOwnPropertyNames(n).map(function(e){var t,i,a,s,u=n[e];return Array.isArray(u)?(t=Object(o.__read)(u,2),s=t[0],a=t[1]):(i=Object(o.__read)(Ft(e,[e,u]),2),a=i[0],s=i[1]),{key:a,quoted:!1,value:r&&s!==a?At([vi(s),vi(a)]):vi(s)}}))):null;var n,r}function yi(e){for(;(t=e[e.length-1])instanceof Le&&null===t.value;)e.pop();var t;return e}var bi,ji,xi,Oi,wi=function(){function e(){this.values=[]}return e.prototype.set=function(e,t){t&&this.values.push({key:e,value:t,quoted:!1})},e.prototype.toLiteralMap=function(){return Tt(this.values)},e}();function Ci(e){var t={};return e.attributes.forEach(function(e){Jr(e.name)||(t[e.name]=e.value)}),e.inputs.forEach(function(e){t[e.name]=""}),e.outputs.forEach(function(e){t[e.name]=""}),t}function Ei(e){var t=Et("t"),n=[],r=Mi(e)?t:new Ye(ce.Or,t,e.type),i=null;if(null!==e.deps)i=new Fe(r,Si(e.deps,e.injectFn));else{var a=Et("ɵ"+e.name+"_BaseFactory"),s=St(Xn.getInheritedFactory),u=a.set(s.callFn([e.type])).toDeclStmt(ge,[Ie.Exported,Ie.Final]);n.push(u),i=a.callFn([r])}var c,l=i,d=[],p=null;function f(e){var n=Et("r");return d.push(n.set(et).toDeclStmt()),d.push(It(t,[n.set(l).toStmt()],[n.set(e).toStmt()])),n}if(Mi(e)&&e.delegateType===bi.Factory){var h=Et("ɵ"+e.name+"_BaseFactory"),m=St(Xn.getFactoryOf);if(e.delegate.isEquivalent(e.type))throw new Error("Illegal state: compiling factory that delegates to itself");var g=h.set(m.callFn([e.delegate])).toDeclStmt(ge,[Ie.Exported,Ie.Final]);n.push(g),p=f(h.callFn([]))}else if(Mi(e)){var v=Si(e.delegateDeps,e.injectFn);p=f(new(e.delegateType===bi.Class?Fe:Ne)(e.delegate,v))}else c=e,p=void 0!==c.expression?f(e.expression):i;return{factory:Dt([new Ge("t",me)],Object(o.__spread)(d,[new at(p)]),ge,void 0,e.name+"_Factory"),statements:n}}function Si(e,t){return e.map(function(e){return function(e,t){switch(e.resolved){case xi.Token:var n=0|(e.self?2:0)|(e.skipSelf?4:0)|(e.host?1:0)|(e.optional?8:0),r=[e.token];return(0!==n||e.optional)&&r.push(Pt(n)),St(t).callFn(r);case xi.Attribute:return St(Xn.injectAttribute).callFn([e.token]);default:return mi("Unknown R3ResolvedDependencyType: "+xi[e.resolved])}}(e,t)})}function ki(e,t,n){n.resolveExternalReference(gn.Injector);var r,i,a=[];try{for(var s=Object(o.__values)(e.diDeps),u=s.next();!u.done;u=s.next()){var c=u.value;if(c.token){var l=In(c.token),d=c.isAttribute?xi.Attribute:xi.Token,p=l instanceof yn?t.importExpr(l):Pt(l);a.push({token:p,resolved:d,host:!!c.isHost,optional:!!c.isOptional,self:!!c.isSelf,skipSelf:!!c.isSkipSelf})}else mi("dependency without a token")}}catch(e){r={error:e}}finally{try{u&&!u.done&&(i=s.return)&&i.call(s)}finally{if(r)throw r.error}}return a}function Mi(e){return void 0!==e.delegateType}function Ai(e){return Tt(Object.keys(e).map(function(t){return{key:t,value:e[t],quoted:!1}}))}function Ti(e,t){if(Array.isArray(e))return At(e.map(function(e){return Ti(e,t)}));if(e instanceof yn)return t.importExpr(e);if(null==e)return Pt(e);throw new Error("Internal error: Unsupported or unknown metadata: "+e)}function Di(e,t){var n=null;if(t>0){n=[];for(var r=0;r<t;r++)n.push(me)}return Mt(e,null,n)}(ji=bi||(bi={}))[ji.Class=0]="Class",ji[ji.Function=1]="Function",ji[ji.Factory=2]="Factory",(Oi=xi||(xi={}))[Oi.Token=0]="Token",Oi[Oi.Attribute=1]="Attribute";var Ii="@";function Pi(e){return""+Ii+e}function Ri(e,t){return""+Ii+e+"."+t}function Ni(e,t){return"animation_"+e+"_"+t}function Fi(e){var t=null,n={name:e.name,type:e.type,deps:e.ctorDeps,injectFn:gn.inject};if(void 0!==e.useClass){var r=e.useClass.isEquivalent(e.type),i=e.userDeps||r&&e.ctorDeps||void 0;t=Ei(void 0!==i?Object(o.__assign)({},n,{delegate:e.useClass,delegateDeps:i,delegateType:bi.Class}):Object(o.__assign)({},n,{delegate:e.useClass,delegateType:bi.Factory}))}else t=void 0!==e.useFactory?Ei(Object(o.__assign)({},n,{delegate:e.useFactory,delegateDeps:e.userDeps||[],delegateType:bi.Function})):void 0!==e.useValue?Ei(Object(o.__assign)({},n,{expression:e.useValue})):void 0!==e.useExisting?Ei(Object(o.__assign)({},n,{expression:St(gn.inject).callFn([e.useExisting])})):Ei(n);var a=e.type,s=e.providedIn;return{expression:St(gn.defineInjectable).callFn([Ai({token:a,factory:t.factory,providedIn:s})]),type:new pe(St(gn.InjectableDef,[Di(e.type,e.typeArgumentCount)])),statements:t.statements}}function Li(e,t){if(null!=t){if(!Array.isArray(t))throw new Error("Expected '"+e+"' to be an array of strings.");for(var n=0;n<t.length;n+=1)if("string"!=typeof t[n])throw new Error("Expected '"+e+"' to be an array of strings.")}}var Vi=[/^\s*$/,/[<>]/,/^[{}]$/,/&(#|[a-z])/i,/^\/\//];function Bi(e,t){if(!(null==t||Array.isArray(t)&&2==t.length))throw new Error("Expected '"+e+"' to be an array, [start, end].");if(null!=t){var n=t[0],r=t[1];Vi.forEach(function(e){if(e.test(n)||e.test(r))throw new Error("['"+n+"', '"+r+"'] contains unusable interpolation symbol.")})}}var zi=function(){function e(e,t){this.start=e,this.end=t}return e.fromArray=function(t){return t?(Bi("interpolation",t),new e(t[0],t[1])):Ui},e}(),Ui=new zi("{{","}}"),Hi=function(){function e(e){void 0===e&&(e=null),this.file=e,this.sourcesContent=new Map,this.lines=[],this.lastCol0=0,this.hasMappings=!1}return e.prototype.addSource=function(e,t){return void 0===t&&(t=null),this.sourcesContent.has(e)||this.sourcesContent.set(e,t),this},e.prototype.addLine=function(){return this.lines.push([]),this.lastCol0=0,this},e.prototype.addMapping=function(e,t,n,r){if(!this.currentLine)throw new Error("A line must be added before mappings can be added");if(null!=t&&!this.sourcesContent.has(t))throw new Error('Unknown source file "'+t+'"');if(null==e)throw new Error("The column in the generated code must be provided");if(e<this.lastCol0)throw new Error("Mapping should be added in output order");if(t&&(null==n||null==r))throw new Error("The source location must be provided when a source url is provided");return this.hasMappings=!0,this.lastCol0=e,this.currentLine.push({col0:e,sourceUrl:t,sourceLine0:n,sourceCol0:r}),this},Object.defineProperty(e.prototype,"currentLine",{get:function(){return this.lines.slice(-1)[0]},enumerable:!0,configurable:!0}),e.prototype.toJSON=function(){var e=this;if(!this.hasMappings)return null;var t=new Map,n=[],r=[];Array.from(this.sourcesContent.keys()).forEach(function(i,o){t.set(i,o),n.push(i),r.push(e.sourcesContent.get(i)||null)});var i="",o=0,a=0,s=0,u=0;return this.lines.forEach(function(e){o=0,i+=e.map(function(e){var n=qi(e.col0-o);return o=e.col0,null!=e.sourceUrl&&(n+=qi(t.get(e.sourceUrl)-a),a=t.get(e.sourceUrl),n+=qi(e.sourceLine0-s),s=e.sourceLine0,n+=qi(e.sourceCol0-u),u=e.sourceCol0),n}).join(","),i+=";"}),i=i.slice(0,-1),{file:this.file||"",version:3,sourceRoot:"",sources:n,sourcesContent:r,mappings:i}},e.prototype.toJsComment=function(){return this.hasMappings?"//# sourceMappingURL=data:application/json;base64,"+function(e){var t="";e=en(e);for(var n=0;n<e.length;){var r=e.charCodeAt(n++),i=e.charCodeAt(n++),o=e.charCodeAt(n++);t+=Wi(r>>2),t+=Wi((3&r)<<4|(isNaN(i)?0:i>>4)),t+=isNaN(i)?"=":Wi((15&i)<<2|o>>6),t+=isNaN(i)||isNaN(o)?"=":Wi(63&o)}return t}(JSON.stringify(this,null,0)):""},e}();function qi(e){e=e<0?1+(-e<<1):e<<1;var t="";do{var n=31&e;(e>>=5)>0&&(n|=32),t+=Wi(n)}while(e>0);return t}var Gi="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";function Wi(e){if(e<0||e>=64)throw new Error("Can only encode value in the range [0, 63]");return Gi[e]}var Yi=/'|\\|\n|\r|\$/g,Ki=/^[$A-Z_][0-9A-Z_$]*$/i,Qi="  ",Xi=Et("error",null,null),Zi=Et("stack",null,null),$i=function(){return function(e){this.indent=e,this.partsLength=0,this.parts=[],this.srcSpans=[]}}(),Ji=function(){function e(e){this._indent=e,this._classes=[],this._preambleLineCount=0,this._lines=[new $i(e)]}return e.createRoot=function(){return new e(0)},Object.defineProperty(e.prototype,"_currentLine",{get:function(){return this._lines[this._lines.length-1]},enumerable:!0,configurable:!0}),e.prototype.println=function(e,t){void 0===t&&(t=""),this.print(e||null,t,!0)},e.prototype.lineIsEmpty=function(){return 0===this._currentLine.parts.length},e.prototype.lineLength=function(){return this._currentLine.indent*Qi.length+this._currentLine.partsLength},e.prototype.print=function(e,t,n){void 0===n&&(n=!1),t.length>0&&(this._currentLine.parts.push(t),this._currentLine.partsLength+=t.length,this._currentLine.srcSpans.push(e&&e.sourceSpan||null)),n&&this._lines.push(new $i(this._indent))},e.prototype.removeEmptyLastLine=function(){this.lineIsEmpty()&&this._lines.pop()},e.prototype.incIndent=function(){this._indent++,this.lineIsEmpty()&&(this._currentLine.indent=this._indent)},e.prototype.decIndent=function(){this._indent--,this.lineIsEmpty()&&(this._currentLine.indent=this._indent)},e.prototype.pushClass=function(e){this._classes.push(e)},e.prototype.popClass=function(){return this._classes.pop()},Object.defineProperty(e.prototype,"currentClass",{get:function(){return this._classes.length>0?this._classes[this._classes.length-1]:null},enumerable:!0,configurable:!0}),e.prototype.toSource=function(){return this.sourceLines.map(function(e){return e.parts.length>0?no(e.indent)+e.parts.join(""):""}).join("\n")},e.prototype.toSourceMapGenerator=function(e,t){void 0===t&&(t=0);for(var n=new Hi(e),r=!1,i=function(){r||(n.addSource(e," ").addMapping(0,e,0,0),r=!0)},o=0;o<t;o++)n.addLine(),i();return this.sourceLines.forEach(function(e,t){n.addLine();for(var o=e.srcSpans,a=e.parts,s=e.indent*Qi.length,u=0;u<o.length&&!o[u];)s+=a[u].length,u++;for(u<o.length&&0===t&&0===s?r=!0:i();u<o.length;){var c=o[u],l=c.start.file,d=c.start.line,p=c.start.col;for(n.addSource(l.url,l.content).addMapping(s,l.url,d,p),s+=a[u].length,u++;u<o.length&&(c===o[u]||!o[u]);)s+=a[u].length,u++}}),n},e.prototype.setPreambleLineCount=function(e){return this._preambleLineCount=e},e.prototype.spanOf=function(e,t){var n=this._lines[e-this._preambleLineCount];if(n)for(var r=t-no(n.indent).length,i=0;i<n.parts.length;i++){var o=n.parts[i];if(o.length>r)return n.srcSpans[i];r-=o.length}return null},Object.defineProperty(e.prototype,"sourceLines",{get:function(){return this._lines.length&&0===this._lines[this._lines.length-1].parts.length?this._lines.slice(0,-1):this._lines},enumerable:!0,configurable:!0}),e}(),eo=function(){function e(e){this._escapeDollarInStrings=e}return e.prototype.visitExpressionStmt=function(e,t){return e.expr.visitExpression(this,t),t.println(e,";"),null},e.prototype.visitReturnStmt=function(e,t){return t.print(e,"return "),e.value.visitExpression(this,t),t.println(e,";"),null},e.prototype.visitIfStmt=function(e,t){t.print(e,"if ("),e.condition.visitExpression(this,t),t.print(e,") {");var n=null!=e.falseCase&&e.falseCase.length>0;return e.trueCase.length<=1&&!n?(t.print(e," "),this.visitAllStatements(e.trueCase,t),t.removeEmptyLastLine(),t.print(e," ")):(t.println(),t.incIndent(),this.visitAllStatements(e.trueCase,t),t.decIndent(),n&&(t.println(e,"} else {"),t.incIndent(),this.visitAllStatements(e.falseCase,t),t.decIndent())),t.println(e,"}"),null},e.prototype.visitThrowStmt=function(e,t){return t.print(e,"throw "),e.error.visitExpression(this,t),t.println(e,";"),null},e.prototype.visitCommentStmt=function(e,t){return e.multiline?t.println(e,"/* "+e.comment+" */"):e.comment.split("\n").forEach(function(n){t.println(e,"// "+n)}),null},e.prototype.visitJSDocCommentStmt=function(e,t){return t.println(e,"/*"+e.toString()+"*/"),null},e.prototype.visitWriteVarExpr=function(e,t){var n=t.lineIsEmpty();return n||t.print(e,"("),t.print(e,e.name+" = "),e.value.visitExpression(this,t),n||t.print(e,")"),null},e.prototype.visitWriteKeyExpr=function(e,t){var n=t.lineIsEmpty();return n||t.print(e,"("),e.receiver.visitExpression(this,t),t.print(e,"["),e.index.visitExpression(this,t),t.print(e,"] = "),e.value.visitExpression(this,t),n||t.print(e,")"),null},e.prototype.visitWritePropExpr=function(e,t){var n=t.lineIsEmpty();return n||t.print(e,"("),e.receiver.visitExpression(this,t),t.print(e,"."+e.name+" = "),e.value.visitExpression(this,t),n||t.print(e,")"),null},e.prototype.visitInvokeMethodExpr=function(e,t){e.receiver.visitExpression(this,t);var n=e.name;return null!=e.builtin&&null==(n=this.getBuiltinMethodName(e.builtin))?null:(t.print(e,"."+n+"("),this.visitAllExpressions(e.args,t,","),t.print(e,")"),null)},e.prototype.visitInvokeFunctionExpr=function(e,t){return e.fn.visitExpression(this,t),t.print(e,"("),this.visitAllExpressions(e.args,t,","),t.print(e,")"),null},e.prototype.visitWrappedNodeExpr=function(e,t){throw new Error("Abstract emitter cannot visit WrappedNodeExpr.")},e.prototype.visitTypeofExpr=function(e,t){t.print(e,"typeof "),e.expr.visitExpression(this,t)},e.prototype.visitReadVarExpr=function(e,t){var n=e.name;if(null!=e.builtin)switch(e.builtin){case xe.Super:n="super";break;case xe.This:n="this";break;case xe.CatchError:n=Xi.name;break;case xe.CatchStack:n=Zi.name;break;default:throw new Error("Unknown builtin variable "+e.builtin)}return t.print(e,n),null},e.prototype.visitInstantiateExpr=function(e,t){return t.print(e,"new "),e.classExpr.visitExpression(this,t),t.print(e,"("),this.visitAllExpressions(e.args,t,","),t.print(e,")"),null},e.prototype.visitLiteralExpr=function(e,t){var n=e.value;return"string"==typeof n?t.print(e,to(n,this._escapeDollarInStrings)):t.print(e,""+n),null},e.prototype.visitConditionalExpr=function(e,t){return t.print(e,"("),e.condition.visitExpression(this,t),t.print(e,"? "),e.trueCase.visitExpression(this,t),t.print(e,": "),e.falseCase.visitExpression(this,t),t.print(e,")"),null},e.prototype.visitNotExpr=function(e,t){return t.print(e,"!"),e.condition.visitExpression(this,t),null},e.prototype.visitAssertNotNullExpr=function(e,t){return e.condition.visitExpression(this,t),null},e.prototype.visitBinaryOperatorExpr=function(e,t){var n;switch(e.operator){case ce.Equals:n="==";break;case ce.Identical:n="===";break;case ce.NotEquals:n="!=";break;case ce.NotIdentical:n="!==";break;case ce.And:n="&&";break;case ce.BitwiseAnd:n="&";break;case ce.Or:n="||";break;case ce.Plus:n="+";break;case ce.Minus:n="-";break;case ce.Divide:n="/";break;case ce.Multiply:n="*";break;case ce.Modulo:n="%";break;case ce.Lower:n="<";break;case ce.LowerEquals:n="<=";break;case ce.Bigger:n=">";break;case ce.BiggerEquals:n=">=";break;default:throw new Error("Unknown operator "+e.operator)}return e.parens&&t.print(e,"("),e.lhs.visitExpression(this,t),t.print(e," "+n+" "),e.rhs.visitExpression(this,t),e.parens&&t.print(e,")"),null},e.prototype.visitReadPropExpr=function(e,t){return e.receiver.visitExpression(this,t),t.print(e,"."),t.print(e,e.name),null},e.prototype.visitReadKeyExpr=function(e,t){return e.receiver.visitExpression(this,t),t.print(e,"["),e.index.visitExpression(this,t),t.print(e,"]"),null},e.prototype.visitLiteralArrayExpr=function(e,t){return t.print(e,"["),this.visitAllExpressions(e.entries,t,","),t.print(e,"]"),null},e.prototype.visitLiteralMapExpr=function(e,t){var n=this;return t.print(e,"{"),this.visitAllObjects(function(r){t.print(e,to(r.key,n._escapeDollarInStrings,r.quoted)+":"),r.value.visitExpression(n,t)},e.entries,t,","),t.print(e,"}"),null},e.prototype.visitCommaExpr=function(e,t){return t.print(e,"("),this.visitAllExpressions(e.parts,t,","),t.print(e,")"),null},e.prototype.visitAllExpressions=function(e,t,n){var r=this;this.visitAllObjects(function(e){return e.visitExpression(r,t)},e,t,n)},e.prototype.visitAllObjects=function(e,t,n,r){for(var i=!1,o=0;o<t.length;o++)o>0&&(n.lineLength()>80?(n.print(null,r,!0),i||(n.incIndent(),n.incIndent(),i=!0)):n.print(null,r,!1)),e(t[o]);i&&(n.decIndent(),n.decIndent())},e.prototype.visitAllStatements=function(e,t){var n=this;e.forEach(function(e){return e.visitStatement(n,t)})},e}();function to(e,t,n){if(void 0===n&&(n=!0),null==e)return null;var r=e.replace(Yi,function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];return"$"==e[0]?t?"\\$":"$":"\n"==e[0]?"\\n":"\r"==e[0]?"\\r":"\\"+e[0]});return n||!Ki.test(r)?"'"+r+"'":r}function no(e){for(var t="",n=0;n<e;n++)t+=Qi;return t}function ro(e,t,n,r){var i=new io(n),a=Ji.createRoot();return i.visitAllStatements(t,a),i.createReturnStmt(a),function(e,t,n,r){var i=t.toSource()+"\n//# sourceURL="+e,a=[],s=[];for(var u in n)a.push(u),s.push(n[u]);if(r){var c=(new(Function.bind.apply(Function,Object(o.__spread)([void 0],a.concat("return null;"))))).toString(),l=c.slice(0,c.indexOf("return null;")).split("\n").length-1;i+="\n"+t.toSourceMapGenerator(e,l).toJsComment()}return(new(Function.bind.apply(Function,Object(o.__spread)([void 0],a.concat(i))))).apply(void 0,Object(o.__spread)(s))}(e,a,i.getArgs(),r)}var io=function(e){function t(t){var n=e.call(this)||this;return n.reflector=t,n._evalArgNames=[],n._evalArgValues=[],n._evalExportedVars=[],n}return Object(o.__extends)(t,e),t.prototype.createReturnStmt=function(e){new at(new $e(this._evalExportedVars.map(function(e){return new Ze(e,Et(e),!1)}))).visitStatement(this,e)},t.prototype.getArgs=function(){for(var e={},t=0;t<this._evalArgNames.length;t++)e[this._evalArgNames[t]]=this._evalArgValues[t];return e},t.prototype.visitExternalExpr=function(e,t){return this._emitReferenceToExternal(e,this.reflector.resolveExternalReference(e.value),t),null},t.prototype.visitWrappedNodeExpr=function(e,t){return this._emitReferenceToExternal(e,e.node,t),null},t.prototype.visitDeclareVarStmt=function(t,n){return t.hasModifier(Ie.Exported)&&this._evalExportedVars.push(t.name),e.prototype.visitDeclareVarStmt.call(this,t,n)},t.prototype.visitDeclareFunctionStmt=function(t,n){return t.hasModifier(Ie.Exported)&&this._evalExportedVars.push(t.name),e.prototype.visitDeclareFunctionStmt.call(this,t,n)},t.prototype.visitDeclareClassStmt=function(t,n){return t.hasModifier(Ie.Exported)&&this._evalExportedVars.push(t.name),e.prototype.visitDeclareClassStmt.call(this,t,n)},t.prototype._emitReferenceToExternal=function(e,t,n){var r=this._evalArgValues.indexOf(t);if(-1===r){r=this._evalArgValues.length,this._evalArgValues.push(t);var i=En({reference:t})||"val";this._evalArgNames.push("jit_"+i+"_"+r)}n.print(e,this._evalArgNames[r])},t}(function(e){function t(){return e.call(this,!1)||this}return Object(o.__extends)(t,e),t.prototype.visitDeclareClassStmt=function(e,t){var n=this;return t.pushClass(e),this._visitClassConstructor(e,t),null!=e.parent&&(t.print(e,e.name+".prototype = Object.create("),e.parent.visitExpression(this,t),t.println(e,".prototype);")),e.getters.forEach(function(r){return n._visitClassGetter(e,r,t)}),e.methods.forEach(function(r){return n._visitClassMethod(e,r,t)}),t.popClass(),null},t.prototype._visitClassConstructor=function(e,t){t.print(e,"function "+e.name+"("),null!=e.constructorMethod&&this._visitParams(e.constructorMethod.params,t),t.println(e,") {"),t.incIndent(),null!=e.constructorMethod&&e.constructorMethod.body.length>0&&(t.println(e,"var self = this;"),this.visitAllStatements(e.constructorMethod.body,t)),t.decIndent(),t.println(e,"}")},t.prototype._visitClassGetter=function(e,t,n){n.println(e,"Object.defineProperty("+e.name+".prototype, '"+t.name+"', { get: function() {"),n.incIndent(),t.body.length>0&&(n.println(e,"var self = this;"),this.visitAllStatements(t.body,n)),n.decIndent(),n.println(e,"}});")},t.prototype._visitClassMethod=function(e,t,n){n.print(e,e.name+".prototype."+t.name+" = function("),this._visitParams(t.params,n),n.println(e,") {"),n.incIndent(),t.body.length>0&&(n.println(e,"var self = this;"),this.visitAllStatements(t.body,n)),n.decIndent(),n.println(e,"};")},t.prototype.visitWrappedNodeExpr=function(e,t){throw new Error("Cannot emit a WrappedNodeExpr in Javascript.")},t.prototype.visitReadVarExpr=function(t,n){if(t.builtin===xe.This)n.print(t,"self");else{if(t.builtin===xe.Super)throw new Error("'super' needs to be handled at a parent ast node, not at the variable level!");e.prototype.visitReadVarExpr.call(this,t,n)}return null},t.prototype.visitDeclareVarStmt=function(e,t){return t.print(e,"var "+e.name),e.value&&(t.print(e," = "),e.value.visitExpression(this,t)),t.println(e,";"),null},t.prototype.visitCastExpr=function(e,t){return e.value.visitExpression(this,t),null},t.prototype.visitInvokeFunctionExpr=function(t,n){var r=t.fn;return r instanceof Se&&r.builtin===xe.Super?(n.currentClass.parent.visitExpression(this,n),n.print(t,".call(this"),t.args.length>0&&(n.print(t,", "),this.visitAllExpressions(t.args,n,",")),n.print(t,")")):e.prototype.visitInvokeFunctionExpr.call(this,t,n),null},t.prototype.visitFunctionExpr=function(e,t){return t.print(e,"function"+(e.name?" "+e.name:"")+"("),this._visitParams(e.params,t),t.println(e,") {"),t.incIndent(),this.visitAllStatements(e.statements,t),t.decIndent(),t.print(e,"}"),null},t.prototype.visitDeclareFunctionStmt=function(e,t){return t.print(e,"function "+e.name+"("),this._visitParams(e.params,t),t.println(e,") {"),t.incIndent(),this.visitAllStatements(e.statements,t),t.decIndent(),t.println(e,"}"),null},t.prototype.visitTryCatchStmt=function(e,t){t.println(e,"try {"),t.incIndent(),this.visitAllStatements(e.bodyStmts,t),t.decIndent(),t.println(e,"} catch ("+Xi.name+") {"),t.incIndent();var n=[Zi.set(Xi.prop("stack")).toDeclStmt(null,[Ie.Final])].concat(e.catchStmts);return this.visitAllStatements(n,t),t.decIndent(),t.println(e,"}"),null},t.prototype._visitParams=function(e,t){this.visitAllObjects(function(e){return t.print(null,e.name)},e,t,",")},t.prototype.getBuiltinMethodName=function(e){var t;switch(e){case Ce.ConcatArray:t="concat";break;case Ce.SubscribeObservable:t="subscribe";break;case Ce.Bind:t="bind";break;default:throw new Error("Unknown builtin method: "+e)}return t},t}(eo)),oo=function(){function e(e){this.context=e}return e.prototype.resolveExternalReference=function(e){if("@angular/core"!==e.moduleName)throw new Error("Cannot resolve external reference to "+e.moduleName+", only references to @angular/core are supported.");if(!this.context.hasOwnProperty(e.name))throw new Error("No value provided for @angular/core symbol '"+e.name+"'.");return this.context[e.name]},e.prototype.parameters=function(e){throw new Error("Not implemented.")},e.prototype.annotations=function(e){throw new Error("Not implemented.")},e.prototype.shallowAnnotations=function(e){throw new Error("Not implemented.")},e.prototype.tryAnnotations=function(e){throw new Error("Not implemented.")},e.prototype.propMetadata=function(e){throw new Error("Not implemented.")},e.prototype.hasLifecycleHook=function(e,t){throw new Error("Not implemented.")},e.prototype.guards=function(e){throw new Error("Not implemented.")},e.prototype.componentModuleUrl=function(e,t){throw new Error("Not implemented.")},e}();function ao(e,t,n,r){return ro(n,Object(o.__spread)(r,[new rt("$def",e,void 0,[Ie.Exported])]),new oo(t),!1).$def}function so(e){var t=e.type,n=e.bootstrap,r=e.declarations,i=e.imports,o=e.exports;return{expression:St(Xn.defineNgModule).callFn([Ai({type:t,bootstrap:At(n.map(function(e){return e.value})),declarations:At(r.map(function(e){return e.value})),imports:At(i.map(function(e){return e.value})),exports:At(o.map(function(e){return e.value}))})]),type:new pe(St(Xn.NgModuleDefWithMeta,[new pe(t),co(r),co(i),co(o)])),additionalStatements:[]}}function uo(e){var t=Ei({name:e.name,type:e.type,deps:e.deps,injectFn:Xn.inject});return{expression:St(Xn.defineInjector).callFn([Ai({factory:t.factory,providers:e.providers,imports:e.imports})]),type:new pe(St(Xn.InjectorDef,[new pe(e.type)])),statements:t.statements}}function co(e){var t=e.map(function(e){return t=e.type,new ke(t);var t});return e.length>0?Mt(At(t)):ye}function lo(e){var t=[];t.push({key:"name",value:Pt(e.pipeName),quoted:!1}),t.push({key:"type",value:e.type,quoted:!1});var n=Ei({name:e.name,type:e.type,deps:e.deps,injectFn:Xn.directiveInject});return t.push({key:"factory",value:n.factory,quoted:!1}),t.push({key:"pure",value:Pt(e.pure),quoted:!1}),{expression:St(Xn.definePipe).callFn([Tt(t)]),type:new pe(St(Xn.PipeDefWithMeta,[new pe(e.type),new pe(new Le(e.pipeName))])),statements:n.statements}}var po=function(){return function(e,t,n,r){this.input=t,this.errLocation=n,this.ctxLocation=r,this.message="Parser Error: "+e+" "+n+" ["+t+"] in "+r}}(),fo=function(){return function(e,t){this.start=e,this.end=t}}(),ho=function(){function e(e){this.span=e}return e.prototype.visit=function(e,t){return void 0===t&&(t=null),null},e.prototype.toString=function(){return"AST"},e}(),mo=function(e){function t(t,n,r,i){var o=e.call(this,t)||this;return o.prefix=n,o.uninterpretedExpression=r,o.location=i,o}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitQuote(this,t)},t.prototype.toString=function(){return"Quote"},t}(ho),go=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){void 0===t&&(t=null)},t}(ho),vo=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitImplicitReceiver(this,t)},t}(ho),_o=function(e){function t(t,n){var r=e.call(this,t)||this;return r.expressions=n,r}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitChain(this,t)},t}(ho),yo=function(e){function t(t,n,r,i){var o=e.call(this,t)||this;return o.condition=n,o.trueExp=r,o.falseExp=i,o}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitConditional(this,t)},t}(ho),bo=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.receiver=n,i.name=r,i}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitPropertyRead(this,t)},t}(ho),jo=function(e){function t(t,n,r,i){var o=e.call(this,t)||this;return o.receiver=n,o.name=r,o.value=i,o}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitPropertyWrite(this,t)},t}(ho),xo=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.receiver=n,i.name=r,i}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitSafePropertyRead(this,t)},t}(ho),Oo=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.obj=n,i.key=r,i}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitKeyedRead(this,t)},t}(ho),wo=function(e){function t(t,n,r,i){var o=e.call(this,t)||this;return o.obj=n,o.key=r,o.value=i,o}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitKeyedWrite(this,t)},t}(ho),Co=function(e){function t(t,n,r,i){var o=e.call(this,t)||this;return o.exp=n,o.name=r,o.args=i,o}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitPipe(this,t)},t}(ho),Eo=function(e){function t(t,n){var r=e.call(this,t)||this;return r.value=n,r}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitLiteralPrimitive(this,t)},t}(ho),So=function(e){function t(t,n){var r=e.call(this,t)||this;return r.expressions=n,r}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitLiteralArray(this,t)},t}(ho),ko=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.keys=n,i.values=r,i}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitLiteralMap(this,t)},t}(ho),Mo=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.strings=n,i.expressions=r,i}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitInterpolation(this,t)},t}(ho),Ao=function(e){function t(t,n,r,i){var o=e.call(this,t)||this;return o.operation=n,o.left=r,o.right=i,o}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitBinary(this,t)},t}(ho),To=function(e){function t(t,n){var r=e.call(this,t)||this;return r.expression=n,r}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitPrefixNot(this,t)},t}(ho),Do=function(e){function t(t,n){var r=e.call(this,t)||this;return r.expression=n,r}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitNonNullAssert(this,t)},t}(ho),Io=function(e){function t(t,n,r,i){var o=e.call(this,t)||this;return o.receiver=n,o.name=r,o.args=i,o}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitMethodCall(this,t)},t}(ho),Po=function(e){function t(t,n,r,i){var o=e.call(this,t)||this;return o.receiver=n,o.name=r,o.args=i,o}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitSafeMethodCall(this,t)},t}(ho),Ro=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.target=n,i.args=r,i}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),e.visitFunctionCall(this,t)},t}(ho),No=function(e){function t(t,n,r,i){var o=e.call(this,new fo(0,null==n?0:n.length))||this;return o.ast=t,o.source=n,o.location=r,o.errors=i,o}return Object(o.__extends)(t,e),t.prototype.visit=function(e,t){return void 0===t&&(t=null),this.ast.visit(e,t)},t.prototype.toString=function(){return this.source+" in "+this.location},t}(ho),Fo=function(){return function(e,t,n,r,i){this.span=e,this.key=t,this.keyIsVar=n,this.name=r,this.expression=i}}(),Lo=function(){function e(){}return e.prototype.visitBinary=function(e,t){},e.prototype.visitChain=function(e,t){},e.prototype.visitConditional=function(e,t){},e.prototype.visitFunctionCall=function(e,t){},e.prototype.visitImplicitReceiver=function(e,t){},e.prototype.visitInterpolation=function(e,t){},e.prototype.visitKeyedRead=function(e,t){},e.prototype.visitKeyedWrite=function(e,t){},e.prototype.visitLiteralArray=function(e,t){},e.prototype.visitLiteralMap=function(e,t){},e.prototype.visitLiteralPrimitive=function(e,t){},e.prototype.visitMethodCall=function(e,t){},e.prototype.visitPipe=function(e,t){},e.prototype.visitPrefixNot=function(e,t){},e.prototype.visitNonNullAssert=function(e,t){},e.prototype.visitPropertyRead=function(e,t){},e.prototype.visitPropertyWrite=function(e,t){},e.prototype.visitQuote=function(e,t){},e.prototype.visitSafeMethodCall=function(e,t){},e.prototype.visitSafePropertyRead=function(e,t){},e}(),Vo=function(){function e(){}return e.prototype.visitBinary=function(e,t){return e.left.visit(this),e.right.visit(this),null},e.prototype.visitChain=function(e,t){return this.visitAll(e.expressions,t)},e.prototype.visitConditional=function(e,t){return e.condition.visit(this),e.trueExp.visit(this),e.falseExp.visit(this),null},e.prototype.visitPipe=function(e,t){return e.exp.visit(this),this.visitAll(e.args,t),null},e.prototype.visitFunctionCall=function(e,t){return e.target.visit(this),this.visitAll(e.args,t),null},e.prototype.visitImplicitReceiver=function(e,t){return null},e.prototype.visitInterpolation=function(e,t){return this.visitAll(e.expressions,t)},e.prototype.visitKeyedRead=function(e,t){return e.obj.visit(this),e.key.visit(this),null},e.prototype.visitKeyedWrite=function(e,t){return e.obj.visit(this),e.key.visit(this),e.value.visit(this),null},e.prototype.visitLiteralArray=function(e,t){return this.visitAll(e.expressions,t)},e.prototype.visitLiteralMap=function(e,t){return this.visitAll(e.values,t)},e.prototype.visitLiteralPrimitive=function(e,t){return null},e.prototype.visitMethodCall=function(e,t){return e.receiver.visit(this),this.visitAll(e.args,t)},e.prototype.visitPrefixNot=function(e,t){return e.expression.visit(this),null},e.prototype.visitNonNullAssert=function(e,t){return e.expression.visit(this),null},e.prototype.visitPropertyRead=function(e,t){return e.receiver.visit(this),null},e.prototype.visitPropertyWrite=function(e,t){return e.receiver.visit(this),e.value.visit(this),null},e.prototype.visitSafePropertyRead=function(e,t){return e.receiver.visit(this),null},e.prototype.visitSafeMethodCall=function(e,t){return e.receiver.visit(this),this.visitAll(e.args,t)},e.prototype.visitAll=function(e,t){var n=this;return e.forEach(function(e){return e.visit(n,t)}),null},e.prototype.visitQuote=function(e,t){return null},e}(),Bo=function(){function e(){}return e.prototype.visitImplicitReceiver=function(e,t){return e},e.prototype.visitInterpolation=function(e,t){return new Mo(e.span,e.strings,this.visitAll(e.expressions))},e.prototype.visitLiteralPrimitive=function(e,t){return new Eo(e.span,e.value)},e.prototype.visitPropertyRead=function(e,t){return new bo(e.span,e.receiver.visit(this),e.name)},e.prototype.visitPropertyWrite=function(e,t){return new jo(e.span,e.receiver.visit(this),e.name,e.value.visit(this))},e.prototype.visitSafePropertyRead=function(e,t){return new xo(e.span,e.receiver.visit(this),e.name)},e.prototype.visitMethodCall=function(e,t){return new Io(e.span,e.receiver.visit(this),e.name,this.visitAll(e.args))},e.prototype.visitSafeMethodCall=function(e,t){return new Po(e.span,e.receiver.visit(this),e.name,this.visitAll(e.args))},e.prototype.visitFunctionCall=function(e,t){return new Ro(e.span,e.target.visit(this),this.visitAll(e.args))},e.prototype.visitLiteralArray=function(e,t){return new So(e.span,this.visitAll(e.expressions))},e.prototype.visitLiteralMap=function(e,t){return new ko(e.span,e.keys,this.visitAll(e.values))},e.prototype.visitBinary=function(e,t){return new Ao(e.span,e.operation,e.left.visit(this),e.right.visit(this))},e.prototype.visitPrefixNot=function(e,t){return new To(e.span,e.expression.visit(this))},e.prototype.visitNonNullAssert=function(e,t){return new Do(e.span,e.expression.visit(this))},e.prototype.visitConditional=function(e,t){return new yo(e.span,e.condition.visit(this),e.trueExp.visit(this),e.falseExp.visit(this))},e.prototype.visitPipe=function(e,t){return new Co(e.span,e.exp.visit(this),e.name,this.visitAll(e.args))},e.prototype.visitKeyedRead=function(e,t){return new Oo(e.span,e.obj.visit(this),e.key.visit(this))},e.prototype.visitKeyedWrite=function(e,t){return new wo(e.span,e.obj.visit(this),e.key.visit(this),e.value.visit(this))},e.prototype.visitAll=function(e){for(var t=new Array(e.length),n=0;n<e.length;++n)t[n]=e[n].visit(this);return t},e.prototype.visitChain=function(e,t){return new _o(e.span,this.visitAll(e.expressions))},e.prototype.visitQuote=function(e,t){return new mo(e.span,e.prefix,e.uninterpretedExpression,e.location)},e}(),zo=function(){function e(){}return e.prototype.visitImplicitReceiver=function(e,t){return e},e.prototype.visitInterpolation=function(e,t){var n=this.visitAll(e.expressions);return n!==e.expressions?new Mo(e.span,e.strings,n):e},e.prototype.visitLiteralPrimitive=function(e,t){return e},e.prototype.visitPropertyRead=function(e,t){var n=e.receiver.visit(this);return n!==e.receiver?new bo(e.span,n,e.name):e},e.prototype.visitPropertyWrite=function(e,t){var n=e.receiver.visit(this),r=e.value.visit(this);return n!==e.receiver||r!==e.value?new jo(e.span,n,e.name,r):e},e.prototype.visitSafePropertyRead=function(e,t){var n=e.receiver.visit(this);return n!==e.receiver?new xo(e.span,n,e.name):e},e.prototype.visitMethodCall=function(e,t){var n=e.receiver.visit(this);return n!==e.receiver?new Io(e.span,n,e.name,this.visitAll(e.args)):e},e.prototype.visitSafeMethodCall=function(e,t){var n=e.receiver.visit(this),r=this.visitAll(e.args);return n!==e.receiver||r!==e.args?new Po(e.span,n,e.name,r):e},e.prototype.visitFunctionCall=function(e,t){var n=e.target&&e.target.visit(this),r=this.visitAll(e.args);return n!==e.target||r!==e.args?new Ro(e.span,n,r):e},e.prototype.visitLiteralArray=function(e,t){var n=this.visitAll(e.expressions);return n!==e.expressions?new So(e.span,n):e},e.prototype.visitLiteralMap=function(e,t){var n=this.visitAll(e.values);return n!==e.values?new ko(e.span,e.keys,n):e},e.prototype.visitBinary=function(e,t){var n=e.left.visit(this),r=e.right.visit(this);return n!==e.left||r!==e.right?new Ao(e.span,e.operation,n,r):e},e.prototype.visitPrefixNot=function(e,t){var n=e.expression.visit(this);return n!==e.expression?new To(e.span,n):e},e.prototype.visitNonNullAssert=function(e,t){var n=e.expression.visit(this);return n!==e.expression?new Do(e.span,n):e},e.prototype.visitConditional=function(e,t){var n=e.condition.visit(this),r=e.trueExp.visit(this),i=e.falseExp.visit(this);return n!==e.condition||r!==e.trueExp||i!=i?new yo(e.span,n,r,i):e},e.prototype.visitPipe=function(e,t){var n=e.exp.visit(this),r=this.visitAll(e.args);return n!==e.exp||r!==e.args?new Co(e.span,n,e.name,r):e},e.prototype.visitKeyedRead=function(e,t){var n=e.obj.visit(this),r=e.key.visit(this);return n!==e.obj||r!==e.key?new Oo(e.span,n,r):e},e.prototype.visitKeyedWrite=function(e,t){var n=e.obj.visit(this),r=e.key.visit(this),i=e.value.visit(this);return n!==e.obj||r!==e.key||i!==e.value?new wo(e.span,n,r,i):e},e.prototype.visitAll=function(e){for(var t=new Array(e.length),n=!1,r=0;r<e.length;++r){var i=e[r],o=i.visit(this);t[r]=o,n=n||o!==i}return n?t:e},e.prototype.visitChain=function(e,t){var n=this.visitAll(e.expressions);return n!==e.expressions?new _o(e.span,n):e},e.prototype.visitQuote=function(e,t){return e},e}();function Uo(e,t,n){function r(e){t.visit&&t.visit(e,n)||e.visit(t,n)}function i(e){e.forEach(r)}e.visit({visitBinary:function(e){r(e.left),r(e.right)},visitChain:function(e){i(e.expressions)},visitConditional:function(e){r(e.condition),r(e.trueExp),r(e.falseExp)},visitFunctionCall:function(e){e.target&&r(e.target),i(e.args)},visitImplicitReceiver:function(e){},visitInterpolation:function(e){i(e.expressions)},visitKeyedRead:function(e){r(e.obj),r(e.key)},visitKeyedWrite:function(e){r(e.obj),r(e.key),r(e.obj)},visitLiteralArray:function(e){i(e.expressions)},visitLiteralMap:function(e){},visitLiteralPrimitive:function(e){},visitMethodCall:function(e){r(e.receiver),i(e.args)},visitPipe:function(e){r(e.exp),i(e.args)},visitPrefixNot:function(e){r(e.expression)},visitNonNullAssert:function(e){r(e.expression)},visitPropertyRead:function(e){r(e.receiver)},visitPropertyWrite:function(e){r(e.receiver),r(e.value)},visitQuote:function(e){},visitSafeMethodCall:function(e){r(e.receiver),i(e.args)},visitSafePropertyRead:function(e){r(e.receiver)}})}var Ho,qo,Go=function(){return function(e,t,n,r){this.name=e,this.expression=t,this.type=n,this.sourceSpan=r,this.isLiteral=this.type===Ho.LITERAL_ATTR,this.isAnimation=this.type===Ho.ANIMATION}}();(qo=Ho||(Ho={}))[qo.DEFAULT=0]="DEFAULT",qo[qo.LITERAL_ATTR=1]="LITERAL_ATTR",qo[qo.ANIMATION=2]="ANIMATION";var Wo=function(){return function(e,t,n,r,i){this.name=e,this.targetOrPhase=t,this.type=n,this.handler=r,this.sourceSpan=i}}(),Yo=function(){return function(e,t,n){this.name=e,this.value=t,this.sourceSpan=n}}(),Ko=function(){return function(e,t,n,r,i,o){this.name=e,this.type=t,this.securityContext=n,this.value=r,this.unit=i,this.sourceSpan=o}}(),Qo=function(){function e(){}return e.event=Et("$event"),e}(),Xo=function(){return function(e,t){this.stmts=e,this.allowDefault=t,this.render3Stmts=e.map(function(e){if(e instanceof rt&&e.name==t.name&&e.value instanceof Ye){var n=e.value.lhs;return new at(n.value)}return e})}}();function Zo(e,t,n,r,i){e||(e=new da);var o=$o({createLiteralArrayConverter:function(e){return function(e){return At(e)}},createLiteralMapConverter:function(e){return function(t){return Tt(e.map(function(e,n){return{key:e.key,value:t[n],quoted:e.quoted}}))}},createPipeConverter:function(e){throw new Error("Illegal State: Actions are not allowed to contain pipes. Pipe: "+e)}},n),a=new la(e,t,r,i),s=[];!function e(t,n){Array.isArray(t)?t.forEach(function(t){return e(t,n)}):n.push(t)}(o.visit(a,ta.Statement),s),function(e,t,n){for(var r=e-1;r>=0;r--)n.unshift(aa(t,r))}(a.temporaryCount,r,s);var u=s.length-1,c=null;if(u>=0){var l=function(e){if(e instanceof ot)return e.expr;if(e instanceof at)return e.value;return null}(s[u]);l&&(c=Et("pd_"+r),s[u]=c.set(l.cast(me).notIdentical(Pt(!1))).toDeclStmt(null,[Ie.Final]))}return new Xo(s,c)}function $o(e,t){return n=t,r=new ca(e),n.visit(r);var n,r}var Jo,ea,ta,na,ra=function(){return function(e,t){this.stmts=e,this.currValExpr=t}}();function ia(e,t,n,r,i,o){e||(e=new da);var a=Et("currVal_"+r),s=[],u=new la(e,t,r,o),c=n.visit(u,ta.Expression);if(u.temporaryCount)for(var l=0;l<u.temporaryCount;l++)s.push(aa(r,l));else if(i==Jo.TrySimple)return new ra([],c);return s.push(a.set(c).toDeclStmt(me,[Ie.Final])),new ra(s,a)}function oa(e,t){return"tmp_"+e+"_"+t}function aa(e,t){return new rt(oa(e,t),et)}function sa(e,t){if(e!==ta.Expression)throw new Error("Expected an expression, but saw "+t)}function ua(e,t){return e===ta.Statement?t.toStmt():t}(ea=Jo||(Jo={}))[ea.General=0]="General",ea[ea.TrySimple=1]="TrySimple",(na=ta||(ta={}))[na.Statement=0]="Statement",na[na.Expression=1]="Expression";var ca=function(e){function t(t){var n=e.call(this)||this;return n._converterFactory=t,n}return Object(o.__extends)(t,e),t.prototype.visitPipe=function(e,t){var n=this,r=Object(o.__spread)([e.exp],e.args).map(function(e){return e.visit(n,t)});return new ha(e.span,r,this._converterFactory.createPipeConverter(e.name,r.length))},t.prototype.visitLiteralArray=function(e,t){var n=this,r=e.expressions.map(function(e){return e.visit(n,t)});return new ha(e.span,r,this._converterFactory.createLiteralArrayConverter(e.expressions.length))},t.prototype.visitLiteralMap=function(e,t){var n=this,r=e.values.map(function(e){return e.visit(n,t)});return new ha(e.span,r,this._converterFactory.createLiteralMapConverter(e.keys))},t}(Bo),la=function(){function e(e,t,n,r){this._localResolver=e,this._implicitReceiver=t,this.bindingId=n,this.interpolationFunction=r,this._nodeMap=new Map,this._resultMap=new Map,this._currentTemporary=0,this.temporaryCount=0}return e.prototype.visitBinary=function(e,t){var n;switch(e.operation){case"+":n=ce.Plus;break;case"-":n=ce.Minus;break;case"*":n=ce.Multiply;break;case"/":n=ce.Divide;break;case"%":n=ce.Modulo;break;case"&&":n=ce.And;break;case"||":n=ce.Or;break;case"==":n=ce.Equals;break;case"!=":n=ce.NotEquals;break;case"===":n=ce.Identical;break;case"!==":n=ce.NotIdentical;break;case"<":n=ce.Lower;break;case">":n=ce.Bigger;break;case"<=":n=ce.LowerEquals;break;case">=":n=ce.BiggerEquals;break;default:throw new Error("Unsupported operation "+e.operation)}return ua(t,new Ye(n,this._visit(e.left,ta.Expression),this._visit(e.right,ta.Expression)))},e.prototype.visitChain=function(e,t){return function(e,t){if(e!==ta.Statement)throw new Error("Expected a statement, but saw "+t)}(t,e),this.visitAll(e.expressions,t)},e.prototype.visitConditional=function(e,t){return ua(t,this._visit(e.condition,ta.Expression).conditional(this._visit(e.trueExp,ta.Expression),this._visit(e.falseExp,ta.Expression)))},e.prototype.visitPipe=function(e,t){throw new Error("Illegal state: Pipes should have been converted into functions. Pipe: "+e.name)},e.prototype.visitFunctionCall=function(e,t){var n=this.visitAll(e.args,ta.Expression);return ua(t,e instanceof ha?e.converter(n):this._visit(e.target,ta.Expression).callFn(n))},e.prototype.visitImplicitReceiver=function(e,t){return sa(t,e),this._implicitReceiver},e.prototype.visitInterpolation=function(e,t){sa(t,e);for(var n=[Pt(e.expressions.length)],r=0;r<e.strings.length-1;r++)n.push(Pt(e.strings[r])),n.push(this._visit(e.expressions[r],ta.Expression));return n.push(Pt(e.strings[e.strings.length-1])),this.interpolationFunction?this.interpolationFunction(n):e.expressions.length<=9?St(gn.inlineInterpolate).callFn(n):St(gn.interpolate).callFn([n[0],At(n.slice(1))])},e.prototype.visitKeyedRead=function(e,t){var n=this.leftMostSafeNode(e);return n?this.convertSafeAccess(e,n,t):ua(t,this._visit(e.obj,ta.Expression).key(this._visit(e.key,ta.Expression)))},e.prototype.visitKeyedWrite=function(e,t){var n=this._visit(e.obj,ta.Expression),r=this._visit(e.key,ta.Expression),i=this._visit(e.value,ta.Expression);return ua(t,n.key(r).set(i))},e.prototype.visitLiteralArray=function(e,t){throw new Error("Illegal State: literal arrays should have been converted into functions")},e.prototype.visitLiteralMap=function(e,t){throw new Error("Illegal State: literal maps should have been converted into functions")},e.prototype.visitLiteralPrimitive=function(e,t){var n=null===e.value||void 0===e.value||!0===e.value||!0===e.value?ge:void 0;return ua(t,Pt(e.value,n))},e.prototype._getLocal=function(e){return this._localResolver.getLocal(e)},e.prototype.visitMethodCall=function(e,t){if(e.receiver instanceof vo&&"$any"==e.name){if(1!=(r=this.visitAll(e.args,ta.Expression)).length)throw new Error("Invalid call to $any, expected 1 argument but received "+(r.length||"none"));return r[0].cast(me)}var n=this.leftMostSafeNode(e);if(n)return this.convertSafeAccess(e,n,t);var r=this.visitAll(e.args,ta.Expression),i=null,o=this._visit(e.receiver,ta.Expression);if(o===this._implicitReceiver){var a=this._getLocal(e.name);a&&(i=a.callFn(r))}return null==i&&(i=o.callMethod(e.name,r)),ua(t,i)},e.prototype.visitPrefixNot=function(e,t){return ua(t,(n=this._visit(e.expression,ta.Expression),new Ue(n,r)));var n,r},e.prototype.visitNonNullAssert=function(e,t){return ua(t,(n=this._visit(e.expression,ta.Expression),new He(n,r)));var n,r},e.prototype.visitPropertyRead=function(e,t){var n=this.leftMostSafeNode(e);if(n)return this.convertSafeAccess(e,n,t);var r=null,i=this._visit(e.receiver,ta.Expression);return i===this._implicitReceiver&&(r=this._getLocal(e.name)),null==r&&(r=i.prop(e.name)),ua(t,r)},e.prototype.visitPropertyWrite=function(e,t){var n=this._visit(e.receiver,ta.Expression),r=null;if(n===this._implicitReceiver){var i=this._getLocal(e.name);if(i){if(!(i instanceof Ke))throw new Error("Cannot assign to a reference or variable!");r=i}}return null===r&&(r=n.prop(e.name)),ua(t,r.set(this._visit(e.value,ta.Expression)))},e.prototype.visitSafePropertyRead=function(e,t){return this.convertSafeAccess(e,this.leftMostSafeNode(e),t)},e.prototype.visitSafeMethodCall=function(e,t){return this.convertSafeAccess(e,this.leftMostSafeNode(e),t)},e.prototype.visitAll=function(e,t){var n=this;return e.map(function(e){return n._visit(e,t)})},e.prototype.visitQuote=function(e,t){throw new Error("Quotes are not supported for evaluation!\n        Statement: "+e.uninterpretedExpression+" located at "+e.location)},e.prototype._visit=function(e,t){var n=this._resultMap.get(e);return n||(this._nodeMap.get(e)||e).visit(this,t)},e.prototype.convertSafeAccess=function(e,t,n){var r=this._visit(t.receiver,ta.Expression),i=void 0;this.needsTemporary(t.receiver)&&(r=(i=this.allocateTemporary()).set(r),this._resultMap.set(t.receiver,i));var o=r.isBlank();t instanceof Po?this._nodeMap.set(t,new Io(t.span,t.receiver,t.name,t.args)):this._nodeMap.set(t,new bo(t.span,t.receiver,t.name));var a=this._visit(e,ta.Expression);return this._nodeMap.delete(t),i&&this.releaseTemporary(i),ua(n,o.conditional(Pt(null),a))},e.prototype.leftMostSafeNode=function(e){var t=this,n=function(e,n){return(t._nodeMap.get(n)||n).visit(e)};return e.visit({visitBinary:function(e){return null},visitChain:function(e){return null},visitConditional:function(e){return null},visitFunctionCall:function(e){return null},visitImplicitReceiver:function(e){return null},visitInterpolation:function(e){return null},visitKeyedRead:function(e){return n(this,e.obj)},visitKeyedWrite:function(e){return null},visitLiteralArray:function(e){return null},visitLiteralMap:function(e){return null},visitLiteralPrimitive:function(e){return null},visitMethodCall:function(e){return n(this,e.receiver)},visitPipe:function(e){return null},visitPrefixNot:function(e){return null},visitNonNullAssert:function(e){return null},visitPropertyRead:function(e){return n(this,e.receiver)},visitPropertyWrite:function(e){return null},visitQuote:function(e){return null},visitSafeMethodCall:function(e){return n(this,e.receiver)||e},visitSafePropertyRead:function(e){return n(this,e.receiver)||e}})},e.prototype.needsTemporary=function(e){var t=this,n=function(e,n){return n&&(t._nodeMap.get(n)||n).visit(e)};return e.visit({visitBinary:function(e){return n(this,e.left)||n(this,e.right)},visitChain:function(e){return!1},visitConditional:function(e){return n(this,e.condition)||n(this,e.trueExp)||n(this,e.falseExp)},visitFunctionCall:function(e){return!0},visitImplicitReceiver:function(e){return!1},visitInterpolation:function(e){return t=this,e.expressions.some(function(e){return n(t,e)});var t},visitKeyedRead:function(e){return!1},visitKeyedWrite:function(e){return!1},visitLiteralArray:function(e){return!0},visitLiteralMap:function(e){return!0},visitLiteralPrimitive:function(e){return!1},visitMethodCall:function(e){return!0},visitPipe:function(e){return!0},visitPrefixNot:function(e){return n(this,e.expression)},visitNonNullAssert:function(e){return n(this,e.expression)},visitPropertyRead:function(e){return!1},visitPropertyWrite:function(e){return!1},visitQuote:function(e){return!1},visitSafeMethodCall:function(e){return!0},visitSafePropertyRead:function(e){return!1}})},e.prototype.allocateTemporary=function(){var e=this._currentTemporary++;return this.temporaryCount=Math.max(this._currentTemporary,this.temporaryCount),new Se(oa(this.bindingId,e))},e.prototype.releaseTemporary=function(e){if(this._currentTemporary--,e.name!=oa(this.bindingId,this._currentTemporary))throw new Error("Temporary "+e.name+" released out of order")},e}();var da=function(){function e(){}return e.prototype.getLocal=function(e){return e===Qo.event.name?Qo.event:null},e}();var pa,fa,ha=function(e){function t(t,n,r){var i=e.call(this,t,null,n)||this;return i.args=n,i.converter=r,i}return Object(o.__extends)(t,e),t}(Ro);(fa=pa||(pa={}))[fa.OnInit=0]="OnInit",fa[fa.OnDestroy=1]="OnDestroy",fa[fa.DoCheck=2]="DoCheck",fa[fa.OnChanges=3]="OnChanges",fa[fa.AfterContentInit=4]="AfterContentInit",fa[fa.AfterContentChecked=5]="AfterContentChecked",fa[fa.AfterViewInit=6]="AfterViewInit",fa[fa.AfterViewChecked=7]="AfterViewChecked";var ma=[pa.OnInit,pa.OnDestroy,pa.DoCheck,pa.OnChanges,pa.AfterContentInit,pa.AfterContentChecked,pa.AfterViewInit,pa.AfterViewChecked];function ga(e,t,n){return e.hasLifecycleHook(n,function(e){switch(e){case pa.OnInit:return"ngOnInit";case pa.OnDestroy:return"ngOnDestroy";case pa.DoCheck:return"ngDoCheck";case pa.OnChanges:return"ngOnChanges";case pa.AfterContentInit:return"ngAfterContentInit";case pa.AfterContentChecked:return"ngAfterContentChecked";case pa.AfterViewInit:return"ngAfterViewInit";case pa.AfterViewChecked:return"ngAfterViewChecked";default:var t=e;throw new Error("unexpected "+t)}}(t))}var va=0,_a=9,ya=10,ba=11,ja=12,xa=13,Oa=32,wa=34,Ca=36,Ea=39,Sa=43,ka=45,Ma=47,Aa=59,Ta=61,Da=62,Ia=48,Pa=57,Ra=65,Na=69,Fa=70,La=90,Va=95,Ba=97,za=101,Ua=102,Ha=110,qa=114,Ga=116,Wa=118,Ya=122,Ka=123,Qa=160,Xa=96;function Za(e){return e>=_a&&e<=Oa||e==Qa}function $a(e){return Ia<=e&&e<=Pa}function Ja(e){return e>=Ba&&e<=Ya||e>=Ra&&e<=La}var es,ts,ns=function(){function e(e,t,n,r){this.file=e,this.offset=t,this.line=n,this.col=r}return e.prototype.toString=function(){return null!=this.offset?this.file.url+"@"+this.line+":"+this.col:this.file.url},e.prototype.moveBy=function(t){for(var n=this.file.content,r=n.length,i=this.offset,o=this.line,a=this.col;i>0&&t<0;){if(i--,t++,(u=n.charCodeAt(i))==ya){o--;var s=n.substr(0,i-1).lastIndexOf(String.fromCharCode(ya));a=s>0?i-s:i}else a--}for(;i<r&&t>0;){var u=n.charCodeAt(i);i++,t--,u==ya?(o++,a=0):a++}return new e(this.file,i,o,a)},e.prototype.getContext=function(e,t){var n=this.file.content,r=this.offset;if(null!=r){r>n.length-1&&(r=n.length-1);for(var i=r,o=0,a=0;o<e&&r>0&&(o++,"\n"!=n[--r]||++a!=t););for(o=0,a=0;o<e&&i<n.length-1&&(o++,"\n"!=n[++i]||++a!=t););return{before:n.substring(r,this.offset),after:n.substring(this.offset,i+1)}}return null},e}(),rs=function(){return function(e,t){this.content=e,this.url=t}}(),is=function(){function e(e,t,n){void 0===n&&(n=null),this.start=e,this.end=t,this.details=n}return e.prototype.toString=function(){return this.start.file.content.substring(this.start.offset,this.end.offset)},e}();(ts=es||(es={}))[ts.WARNING=0]="WARNING",ts[ts.ERROR=1]="ERROR";var os=function(){function e(e,t,n){void 0===n&&(n=es.ERROR),this.span=e,this.msg=t,this.level=n}return e.prototype.contextualMessage=function(){var e=this.span.start.getContext(100,3);return e?this.msg+' ("'+e.before+"["+es[this.level]+" ->]"+e.after+'")':this.msg},e.prototype.toString=function(){var e=this.span.details?", "+this.span.details:"";return this.contextualMessage()+": "+this.span.start+e},e}();function as(e,t){var n=Sn(t),r=null!=n?"in "+e+" "+En(t)+" in "+n:"in "+e+" "+En(t),i=new rs("",r);return new is(new ns(i,-1,-1,-1),new ns(i,-1,-1,-1))}var ss=function(){function e(){this.strictStyling=!0}return e.prototype.shimCssText=function(e,t,n){void 0===n&&(n="");var r=e.match(Es)||[];e=e.replace(Cs,""),e=this._insertDirectives(e);var i=this._scopeCssText(e,t,n);return Object(o.__spread)([i],r).join("\n")},e.prototype._insertDirectives=function(e){return e=this._insertPolyfillDirectivesInCssText(e),this._insertPolyfillRulesInCssText(e)},e.prototype._insertPolyfillDirectivesInCssText=function(e){return e.replace(cs,function(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];return e[2]+"{"})},e.prototype._insertPolyfillRulesInCssText=function(e){return e.replace(ls,function(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];var n=e[0].replace(e[1],"").replace(e[2],"");return e[4]+n})},e.prototype._scopeCssText=function(e,t,n){var r=this._extractUnscopedRulesFromCssText(e);return e=this._insertPolyfillHostInCssText(e),e=this._convertColonHost(e),e=this._convertColonHostContext(e),e=this._convertShadowDOMSelectors(e),t&&(e=this._scopeSelectors(e,t,n)),(e=e+"\n"+r).trim()},e.prototype._extractUnscopedRulesFromCssText=function(e){var t,n="";for(ds.lastIndex=0;null!==(t=ds.exec(e));){n+=t[0].replace(t[2],"").replace(t[1],t[4])+"\n\n"}return n},e.prototype._convertColonHost=function(e){return this._convertColonRule(e,ms,this._colonHostPartReplacer)},e.prototype._convertColonHostContext=function(e){return this._convertColonRule(e,gs,this._colonHostContextPartReplacer)},e.prototype._convertColonRule=function(e,t,n){return e.replace(t,function(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];if(e[2]){for(var r=e[2].split(","),i=[],o=0;o<r.length;o++){var a=r[o].trim();if(!a)break;i.push(n(vs,a,e[3]))}return i.join(",")}return vs+e[3]})},e.prototype._colonHostContextPartReplacer=function(e,t,n){return t.indexOf(ps)>-1?this._colonHostPartReplacer(e,t,n):e+t+n+", "+t+" "+e+n},e.prototype._colonHostPartReplacer=function(e,t,n){return e+t.replace(ps,"")+n},e.prototype._convertShadowDOMSelectors=function(e){return ys.reduce(function(e,t){return e.replace(t," ")},e)},e.prototype._scopeSelectors=function(e,t,n){var r,i,o,a=this;return r=function(e){var r=e.selector,i=e.content;return"@"!=e.selector[0]?r=a._scopeSelector(e.selector,t,n,a.strictStyling):(e.selector.startsWith("@media")||e.selector.startsWith("@supports")||e.selector.startsWith("@page")||e.selector.startsWith("@document"))&&(i=a._scopeSelectors(e.content,t,n)),new Ds(r,i)},i=function(e){for(var t=e.split(ks),n=[],r=[],i=0,o=[],a=0;a<t.length;a++){var s=t[a];s==As&&i--,i>0?o.push(s):(o.length>0&&(r.push(o.join("")),n.push(Ts),o=[]),n.push(s)),s==Ms&&i++}o.length>0&&(r.push(o.join("")),n.push(Ts));return new Is(n.join(""),r)}(e),o=0,i.escapedString.replace(Ss,function(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];var n=e[2],a="",s=e[4],u="";s&&s.startsWith("{"+Ts)&&(a=i.blocks[o++],s=s.substring(Ts.length+1),u="{");var c=r(new Ds(n,a));return""+e[1]+c.selector+e[3]+u+c.content+s})},e.prototype._scopeSelector=function(e,t,n,r){var i=this;return e.split(",").map(function(e){return e.trim().split(bs)}).map(function(e){var a,s=Object(o.__read)(e),u=s[0],c=s.slice(1);return Object(o.__spread)([(a=u,i._selectorNeedsScoping(a,t)?r?i._applyStrictSelectorScope(a,t,n):i._applySelectorScope(a,t,n):a)],c).join(" ")}).join(", ")},e.prototype._selectorNeedsScoping=function(e,t){return!this._makeScopeMatcher(t).test(e)},e.prototype._makeScopeMatcher=function(e){return e=e.replace(/\[/g,"\\[").replace(/\]/g,"\\]"),new RegExp("^("+e+")"+js,"m")},e.prototype._applySelectorScope=function(e,t,n){return this._applySimpleSelectorScope(e,t,n)},e.prototype._applySimpleSelectorScope=function(e,t,n){if(xs.lastIndex=0,xs.test(e)){var r=this.strictStyling?"["+n+"]":t;return e.replace(_s,function(e,t){return t.replace(/([^:]*)(:*)(.*)/,function(e,t,n,i){return t+r+n+i})}).replace(xs,r+" ")}return t+" "+e},e.prototype._applyStrictSelectorScope=function(e,t,n){for(var r,i=this,o="["+(t=t.replace(/\[is=([^\]]*)\]/g,function(e){for(var t=[],n=1;n<arguments.length;n++)t[n-1]=arguments[n];return t[0]}))+"]",a=function(e){var r=e.trim();if(!r)return"";if(e.indexOf(vs)>-1)r=i._applySimpleSelectorScope(e,t,n);else{var a=e.replace(xs,"");if(a.length>0){var s=a.match(/([^:]*)(:*)(.*)/);s&&(r=s[1]+o+s[2]+s[3])}}return r},s=new us(e),u="",c=0,l=/( |>|\+|~(?!=))\s*/g,d=!((e=s.content()).indexOf(vs)>-1);null!==(r=l.exec(e));){var p=r[1],f=e.slice(c,r.index).trim();u+=((d=d||f.indexOf(vs)>-1)?a(f):f)+" "+p+" ",c=l.lastIndex}var h=e.substring(c);return u+=(d=d||h.indexOf(vs)>-1)?a(h):h,s.restore(u)},e.prototype._insertPolyfillHostInCssText=function(e){return e.replace(ws,fs).replace(Os,ps)},e}(),us=function(){function e(e){var t=this;this.placeholders=[],this.index=0,e=e.replace(/(\[[^\]]*\])/g,function(e,n){var r="__ph-"+t.index+"__";return t.placeholders.push(n),t.index++,r}),this._content=e.replace(/(:nth-[-\w]+)(\([^)]+\))/g,function(e,n,r){var i="__ph-"+t.index+"__";return t.placeholders.push(r),t.index++,n+i})}return e.prototype.restore=function(e){var t=this;return e.replace(/__ph-(\d+)__/g,function(e,n){return t.placeholders[+n]})},e.prototype.content=function(){return this._content},e}(),cs=/polyfill-next-selector[^}]*content:[\s]*?(['"])(.*?)\1[;\s]*}([^{]*?){/gim,ls=/(polyfill-rule)[^}]*(content:[\s]*(['"])(.*?)\3)[;\s]*[^}]*}/gim,ds=/(polyfill-unscoped-rule)[^}]*(content:[\s]*(['"])(.*?)\3)[;\s]*[^}]*}/gim,ps="-shadowcsshost",fs="-shadowcsscontext",hs=")(?:\\(((?:\\([^)(]*\\)|[^)(]*)+?)\\))?([^,{]*)",ms=new RegExp("("+ps+hs,"gim"),gs=new RegExp("("+fs+hs,"gim"),vs=ps+"-no-combinator",_s=/-shadowcsshost-no-combinator([^\s]*)/,ys=[/::shadow/g,/::content/g,/\/shadow-deep\//g,/\/shadow\//g],bs=/(?:>>>)|(?:\/deep\/)|(?:::ng-deep)/g,js="([>\\s~+[.,{:][\\s\\S]*)?$",xs=/-shadowcsshost/gim,Os=/:host/gim,ws=/:host-context/gim,Cs=/\/\*\s*[\s\S]*?\*\//g;var Es=/\/\*\s*#\s*source(Mapping)?URL=[\s\S]+?\*\//g;var Ss=/(\s*)([^;\{\}]+?)(\s*)((?:{%BLOCK%}?\s*;?)|(?:\s*;))/g,ks=/([{}])/g,Ms="{",As="}",Ts="%BLOCK%",Ds=function(){return function(e,t){this.selector=e,this.content=t}}();var Is=function(){return function(e,t){this.escapedString=e,this.blocks=t}}();var Ps="_nghost-%COMP%",Rs="_ngcontent-%COMP%",Ns=function(){return function(e,t,n){this.name=e,this.moduleUrl=t,this.setValue=n}}(),Fs=function(){return function(e,t,n,r,i){this.outputCtx=e,this.stylesVar=t,this.dependencies=n,this.isShimmed=r,this.meta=i}}(),Ls=function(){function e(e){this._urlResolver=e,this._shadowCss=new ss}return e.prototype.compileComponent=function(e,t){var n=t.template;return this._compileStyles(e,t,new Pn({styles:n.styles,styleUrls:n.styleUrls,moduleUrl:Sn(t.type)}),this.needsStyleShim(t),!0)},e.prototype.compileStyles=function(e,t,n,r){return void 0===r&&(r=this.needsStyleShim(t)),this._compileStyles(e,t,n,r,!1)},e.prototype.needsStyleShim=function(e){return e.template.encapsulation===_.Emulated},e.prototype._compileStyles=function(e,t,n,r,i){var o=this,a=n.styles.map(function(e){return Pt(o._shimIfNeeded(e,r))}),s=[];n.styleUrls.forEach(function(t){var n=a.length;a.push(null),s.push(new Ns(Vs(null),t,function(t){return a[n]=e.importExpr(t)}))});var u=Vs(i?t:null),c=Et(u).set(At(a,new fe(me,[re.Const]))).toDeclStmt(null,i?[Ie.Final]:[Ie.Final,Ie.Exported]);return e.statements.push(c),new Fs(e,u,s,r,n)},e.prototype._shimIfNeeded=function(e,t){return t?this._shadowCss.shimCssText(e,Rs,Ps):e},e}();function Vs(e){var t="styles";return e&&(t+="_"+En(e.type)),t}function Bs(e){var t=e.charCodeAt(0);if(t==e.charCodeAt(e.length-1)&&(39==t||34==t)){var n=e.substring(1,e.length-1);-1==n.indexOf("'")&&-1==n.indexOf('"')&&(e=n)}return e}var zs,Us,Hs=function(){function e(e,t){this._elementIndexExpr=e,this._directiveExpr=t,this._hasInitialValues=!1,this._hasBindings=!1,this._classMapInput=null,this._styleMapInput=null,this._singleStyleInputs=null,this._singleClassInputs=null,this._lastStylingInput=null,this._stylesIndex=new Map,this._classesIndex=new Map,this._initialStyleValues=[],this._initialClassValues=[],this._useDefaultSanitizer=!1}return e.prototype.hasBindingsOrInitialValues=function(){return this._hasBindings||this._hasInitialValues},e.prototype.registerBoundInput=function(e){var t,n=e.name,r=null;switch(e.type){case 0:"style"==n?r=this.registerStyleInput(null,e.value,"",e.sourceSpan):("className"==(t=e.name)||"class"==t)&&(r=this.registerClassInput(null,e.value,e.sourceSpan));break;case 3:r=this.registerStyleInput(e.name,e.value,e.unit,e.sourceSpan);break;case 2:r=this.registerClassInput(e.name,e.value,e.sourceSpan)}return!!r},e.prototype.registerStyleInput=function(e,t,n,r){var i,o={name:e,unit:n,value:t,sourceSpan:r};return e?((this._singleStyleInputs=this._singleStyleInputs||[]).push(o),this._useDefaultSanitizer=this._useDefaultSanitizer||("background-image"===(i=e)||"background"===i||"border-image"===i||"filter"===i||"list-style"===i||"list-style-image"===i),qs(this._stylesIndex,e)):(this._useDefaultSanitizer=!0,this._styleMapInput=o),this._lastStylingInput=o,this._hasBindings=!0,o},e.prototype.registerClassInput=function(e,t,n){var r={name:e,value:t,sourceSpan:n};return e?((this._singleClassInputs=this._singleClassInputs||[]).push(r),qs(this._classesIndex,e)):this._classMapInput=r,this._lastStylingInput=r,this._hasBindings=!0,r},e.prototype.registerStyleAttr=function(e){this._initialStyleValues=function(e){for(var t=[],n=0,r=0,i=0,o=0,a=0,s=null,u=!1;n<e.length;)switch(e.charCodeAt(n++)){case 40:r++;break;case 41:r--;break;case 39:u=u||o>0,0===i?i=39:39===i&&92!==e.charCodeAt(n-1)&&(i=0);break;case 34:u=u||o>0,0===i?i=34:34===i&&92!==e.charCodeAt(n-1)&&(i=0);break;case 58:s||0!==r||0!==i||(s=e.substring(a,n-1).trim().replace(/[a-z][A-Z]/g,function(e){return e.charAt(0)+"-"+e.charAt(1)}).toLowerCase(),o=n);break;case 59:if(s&&o>0&&0===r&&0===i){var c=e.substring(o,n-1).trim();t.push(s,u?Bs(c):c),a=n,o=0,s=null,u=!1}}s&&o&&(c=e.substr(o).trim(),t.push(s,u?Bs(c):c));return t}(e),this._hasInitialValues=!0},e.prototype.registerClassAttr=function(e){this._initialClassValues=e.trim().split(/\s+/g),this._hasInitialValues=!0},e.prototype.populateInitialStylingAttrs=function(e){if(this._initialClassValues.length){e.push(Pt(1));for(var t=0;t<this._initialClassValues.length;t++)e.push(Pt(this._initialClassValues[t]))}if(this._initialStyleValues.length){e.push(Pt(2));for(t=0;t<this._initialStyleValues.length;t+=2)e.push(Pt(this._initialStyleValues[t]),Pt(this._initialStyleValues[t+1]))}},e.prototype.buildDirectiveHostAttrsInstruction=function(e,t){var n=this;return this._hasInitialValues&&this._directiveExpr?{sourceSpan:e,reference:Xn.elementHostAttrs,buildParams:function(){var e=[];return n.populateInitialStylingAttrs(e),[n._directiveExpr,Gs(t,e)]}}:null},e.prototype.buildElementStylingInstruction=function(e,t){var n=this;return this._hasBindings?{sourceSpan:e,reference:Xn.elementStyling,buildParams:function(){var e=n._singleStyleInputs?n._singleStyleInputs.map(function(e){return Pt(e.name)}):[],r=n._singleClassInputs?n._singleClassInputs.map(function(e){return Pt(e.name)}):[],i=0;n._directiveExpr?i=4:n._useDefaultSanitizer?i=3:e.length?i=2:r.length&&(i=1);var o=[];return Ws(o,r.length>0,Gs(t,r),1,i),Ws(o,e.length>0,Gs(t,e),2,i),Ws(o,n._useDefaultSanitizer,St(Xn.defaultStyleSanitizer),3,i),n._directiveExpr&&o.push(n._directiveExpr),o}}:null},e.prototype.buildElementStylingMapInstruction=function(e){var t=this;if(this._classMapInput||this._styleMapInput){var n=this._classMapInput||this._styleMapInput,r=this._classMapInput?this._classMapInput.value.visit(e):null,i=this._styleMapInput?this._styleMapInput.value.visit(e):null;return{sourceSpan:n.sourceSpan,reference:Xn.elementStylingMap,buildParams:function(e){var n=[t._elementIndexExpr];return r?n.push(e(r)):t._styleMapInput&&n.push(et),i?n.push(e(i)):t._directiveExpr&&n.push(et),t._directiveExpr&&n.push(t._directiveExpr),n}}}return null},e.prototype._buildSingleInputs=function(e,t,n,r,i){var o=this;return t.map(function(t){var a=n.get(t.name),s=t.value.visit(i);return{sourceSpan:t.sourceSpan,reference:e,buildParams:function(e){var n=[o._elementIndexExpr,Pt(a),e(s)];return r&&(t.unit?n.push(Pt(t.unit)):o._directiveExpr&&n.push(et)),o._directiveExpr&&n.push(o._directiveExpr),n}}})},e.prototype._buildClassInputs=function(e){return this._singleClassInputs?this._buildSingleInputs(Xn.elementClassProp,this._singleClassInputs,this._classesIndex,!1,e):[]},e.prototype._buildStyleInputs=function(e){return this._singleStyleInputs?this._buildSingleInputs(Xn.elementStyleProp,this._singleStyleInputs,this._stylesIndex,!0,e):[]},e.prototype._buildApplyFn=function(){var e=this;return{sourceSpan:this._lastStylingInput?this._lastStylingInput.sourceSpan:null,reference:Xn.elementStylingApply,buildParams:function(){var t=[e._elementIndexExpr];return e._directiveExpr&&t.push(e._directiveExpr),t}}},e.prototype.buildUpdateLevelInstructions=function(e){var t=[];if(this._hasBindings){var n=this.buildElementStylingMapInstruction(e);n&&t.push(n),t.push.apply(t,Object(o.__spread)(this._buildStyleInputs(e))),t.push.apply(t,Object(o.__spread)(this._buildClassInputs(e))),t.push(this._buildApplyFn())}return t},e}();function qs(e,t){e.has(t)||e.set(t,e.size)}function Gs(e,t){return t.length?e.getConstLiteral(At(t),!0):et}function Ws(e,t,n,r,i){t?e.push(n):r<i&&e.push(et)}(Us=zs||(zs={}))[Us.Character=0]="Character",Us[Us.Identifier=1]="Identifier",Us[Us.Keyword=2]="Keyword",Us[Us.String=3]="String",Us[Us.Operator=4]="Operator",Us[Us.Number=5]="Number",Us[Us.Error=6]="Error";var Ys=["var","let","as","null","undefined","true","false","if","else","this"],Ks=function(){function e(){}return e.prototype.tokenize=function(e){for(var t=new Js(e),n=[],r=t.scanToken();null!=r;)n.push(r),r=t.scanToken();return n},e}(),Qs=function(){function e(e,t,n,r){this.index=e,this.type=t,this.numValue=n,this.strValue=r}return e.prototype.isCharacter=function(e){return this.type==zs.Character&&this.numValue==e},e.prototype.isNumber=function(){return this.type==zs.Number},e.prototype.isString=function(){return this.type==zs.String},e.prototype.isOperator=function(e){return this.type==zs.Operator&&this.strValue==e},e.prototype.isIdentifier=function(){return this.type==zs.Identifier},e.prototype.isKeyword=function(){return this.type==zs.Keyword},e.prototype.isKeywordLet=function(){return this.type==zs.Keyword&&"let"==this.strValue},e.prototype.isKeywordAs=function(){return this.type==zs.Keyword&&"as"==this.strValue},e.prototype.isKeywordNull=function(){return this.type==zs.Keyword&&"null"==this.strValue},e.prototype.isKeywordUndefined=function(){return this.type==zs.Keyword&&"undefined"==this.strValue},e.prototype.isKeywordTrue=function(){return this.type==zs.Keyword&&"true"==this.strValue},e.prototype.isKeywordFalse=function(){return this.type==zs.Keyword&&"false"==this.strValue},e.prototype.isKeywordThis=function(){return this.type==zs.Keyword&&"this"==this.strValue},e.prototype.isError=function(){return this.type==zs.Error},e.prototype.toNumber=function(){return this.type==zs.Number?this.numValue:-1},e.prototype.toString=function(){switch(this.type){case zs.Character:case zs.Identifier:case zs.Keyword:case zs.Operator:case zs.String:case zs.Error:return this.strValue;case zs.Number:return this.numValue.toString();default:return null}},e}();function Xs(e,t){return new Qs(e,zs.Character,t,String.fromCharCode(t))}function Zs(e,t){return new Qs(e,zs.Operator,0,t)}var $s=new Qs(-1,zs.Character,0,""),Js=function(){function e(e){this.input=e,this.peek=0,this.index=-1,this.length=e.length,this.advance()}return e.prototype.advance=function(){this.peek=++this.index>=this.length?va:this.input.charCodeAt(this.index)},e.prototype.scanToken=function(){for(var e=this.input,t=this.length,n=this.peek,r=this.index;n<=Oa;){if(++r>=t){n=va;break}n=e.charCodeAt(r)}if(this.peek=n,this.index=r,r>=t)return null;if(eu(n))return this.scanIdentifier();if($a(n))return this.scanNumber(r);var i=r;switch(n){case 46:return this.advance(),$a(this.peek)?this.scanNumber(i):Xs(i,46);case 40:case 41:case Ka:case 125:case 91:case 93:case 44:case 58:case Aa:return this.scanCharacter(i,n);case Ea:case wa:return this.scanString();case 35:case Sa:case ka:case 42:case Ma:case 37:case 94:return this.scanOperator(i,String.fromCharCode(n));case 63:return this.scanComplexOperator(i,"?",46,".");case 60:case Da:return this.scanComplexOperator(i,String.fromCharCode(n),Ta,"=");case 33:case Ta:return this.scanComplexOperator(i,String.fromCharCode(n),Ta,"=",Ta,"=");case 38:return this.scanComplexOperator(i,"&",38,"&");case 124:return this.scanComplexOperator(i,"|",124,"|");case Qa:for(;Za(this.peek);)this.advance();return this.scanToken()}return this.advance(),this.error("Unexpected character ["+String.fromCharCode(n)+"]",0)},e.prototype.scanCharacter=function(e,t){return this.advance(),Xs(e,t)},e.prototype.scanOperator=function(e,t){return this.advance(),Zs(e,t)},e.prototype.scanComplexOperator=function(e,t,n,r,i,o){this.advance();var a=t;return this.peek==n&&(this.advance(),a+=r),null!=i&&this.peek==i&&(this.advance(),a+=o),Zs(e,a)},e.prototype.scanIdentifier=function(){var e=this.index;for(this.advance();nu(this.peek);)this.advance();var t,n,r=this.input.substring(e,this.index);return Ys.indexOf(r)>-1?(n=r,new Qs(e,zs.Keyword,0,n)):(t=r,new Qs(e,zs.Identifier,0,t))},e.prototype.scanNumber=function(e){var t,n,r=this.index===e;for(this.advance();;){if($a(this.peek));else if(46==this.peek)r=!1;else{if((n=this.peek)!=za&&n!=Na)break;if(this.advance(),((t=this.peek)==ka||t==Sa)&&this.advance(),!$a(this.peek))return this.error("Invalid exponent",-1);r=!1}this.advance()}var i,o=this.input.substring(e,this.index),a=r?function(e){var t=parseInt(e);if(isNaN(t))throw new Error("Invalid integer literal when parsing "+e);return t}(o):parseFloat(o);return i=a,new Qs(e,zs.Number,i,"")},e.prototype.scanString=function(){var e=this.index,t=this.peek;this.advance();for(var n="",r=this.index,i=this.input;this.peek!=t;)if(92==this.peek){n+=i.substring(r,this.index),this.advance();var o=void 0;if(this.peek=this.peek,117==this.peek){var a=i.substring(this.index+1,this.index+5);if(!/^[0-9a-f]+$/i.test(a))return this.error("Invalid unicode escape [\\u"+a+"]",0);o=parseInt(a,16);for(var s=0;s<5;s++)this.advance()}else o=iu(this.peek),this.advance();n+=String.fromCharCode(o),r=this.index}else{if(this.peek==va)return this.error("Unterminated quote",0);this.advance()}var u,c=i.substring(r,this.index);return this.advance(),u=n+c,new Qs(e,zs.String,0,u)},e.prototype.error=function(e,t){var n,r,i=this.index+t;return n=i,r="Lexer Error: "+e+" at column "+i+" in expression ["+this.input+"]",new Qs(n,zs.Error,0,r)},e}();function eu(e){return Ba<=e&&e<=Ya||Ra<=e&&e<=La||e==Va||e==Ca}function tu(e){if(0==e.length)return!1;var t=new Js(e);if(!eu(t.peek))return!1;for(t.advance();t.peek!==va;){if(!nu(t.peek))return!1;t.advance()}return!0}function nu(e){return Ja(e)||$a(e)||e==Va||e==Ca}function ru(e){return e===Ea||e===wa||e===Xa}function iu(e){switch(e){case Ha:return ya;case Ua:return ja;case qa:return xa;case Ga:return _a;case Wa:return ba;default:return e}}var ou=function(){return function(e,t,n){this.strings=e,this.expressions=t,this.offsets=n}}(),au=function(){return function(e,t,n){this.templateBindings=e,this.warnings=t,this.errors=n}}();function su(e){var t=$t(e.start)+"([\\s\\S]*?)"+$t(e.end);return new RegExp(t,"g")}var uu=function(){function e(e){this._lexer=e,this.errors=[]}return e.prototype.parseAction=function(e,t,n){void 0===n&&(n=Ui),this._checkNoInterpolation(e,t,n);var r=this._stripComments(e),i=this._lexer.tokenize(this._stripComments(e)),o=new cu(e,t,i,r.length,!0,this.errors,e.length-r.length).parseChain();return new No(o,e,t,this.errors)},e.prototype.parseBinding=function(e,t,n){void 0===n&&(n=Ui);var r=this._parseBindingAst(e,t,n);return new No(r,e,t,this.errors)},e.prototype.parseSimpleBinding=function(e,t,n){void 0===n&&(n=Ui);var r=this._parseBindingAst(e,t,n),i=lu.check(r);return i.length>0&&this._reportError("Host binding expression cannot contain "+i.join(" "),e,t),new No(r,e,t,this.errors)},e.prototype._reportError=function(e,t,n,r){this.errors.push(new po(e,t,n,r))},e.prototype._parseBindingAst=function(e,t,n){var r=this._parseQuote(e,t);if(null!=r)return r;this._checkNoInterpolation(e,t,n);var i=this._stripComments(e),o=this._lexer.tokenize(i);return new cu(e,t,o,i.length,!1,this.errors,e.length-i.length).parseChain()},e.prototype._parseQuote=function(e,t){if(null==e)return null;var n=e.indexOf(":");if(-1==n)return null;var r=e.substring(0,n).trim();if(!tu(r))return null;var i=e.substring(n+1);return new mo(new fo(0,e.length),r,i,t)},e.prototype.parseTemplateBindings=function(e,t,n){var r=this._lexer.tokenize(t);return new cu(t,n,r,t.length,!1,this.errors,0).parseTemplateBindings(e)},e.prototype.parseInterpolation=function(e,t,n){void 0===n&&(n=Ui);var r=this.splitInterpolation(e,t,n);if(null==r)return null;for(var i=[],o=0;o<r.expressions.length;++o){var a=r.expressions[o],s=this._stripComments(a),u=this._lexer.tokenize(s),c=new cu(e,t,u,s.length,!1,this.errors,r.offsets[o]+(a.length-s.length)).parseChain();i.push(c)}return new No(new Mo(new fo(0,null==e?0:e.length),r.strings,i),e,t,this.errors)},e.prototype.splitInterpolation=function(e,t,n){void 0===n&&(n=Ui);var r=su(n),i=e.split(r);if(i.length<=1)return null;for(var o=[],a=[],s=[],u=0,c=0;c<i.length;c++){var l=i[c];c%2==0?(o.push(l),u+=l.length):l.trim().length>0?(u+=n.start.length,a.push(l),s.push(u),u+=l.length+n.end.length):(this._reportError("Blank expressions are not allowed in interpolated strings",e,"at column "+this._findInterpolationErrorColumn(i,c,n)+" in",t),a.push("$implict"),s.push(u))}return new ou(o,a,s)},e.prototype.wrapLiteralPrimitive=function(e,t){return new No(new Eo(new fo(0,null==e?0:e.length),e),e,t,this.errors)},e.prototype._stripComments=function(e){var t=this._commentStart(e);return null!=t?e.substring(0,t).trim():e},e.prototype._commentStart=function(e){for(var t=null,n=0;n<e.length-1;n++){var r=e.charCodeAt(n),i=e.charCodeAt(n+1);if(r===Ma&&i==Ma&&null==t)return n;t===r?t=null:null==t&&ru(r)&&(t=r)}return null},e.prototype._checkNoInterpolation=function(e,t,n){var r=su(n),i=e.split(r);i.length>1&&this._reportError("Got interpolation ("+n.start+n.end+") where expression was expected",e,"at column "+this._findInterpolationErrorColumn(i,1,n)+" in",t)},e.prototype._findInterpolationErrorColumn=function(e,t,n){for(var r="",i=0;i<t;i++)r+=i%2==0?e[i]:""+n.start+e[i]+n.end;return r.length},e}(),cu=function(){function e(e,t,n,r,i,o,a){this.input=e,this.location=t,this.tokens=n,this.inputLength=r,this.parseAction=i,this.errors=o,this.offset=a,this.rparensExpected=0,this.rbracketsExpected=0,this.rbracesExpected=0,this.index=0}return e.prototype.peek=function(e){var t=this.index+e;return t<this.tokens.length?this.tokens[t]:$s},Object.defineProperty(e.prototype,"next",{get:function(){return this.peek(0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"inputIndex",{get:function(){return this.index<this.tokens.length?this.next.index+this.offset:this.inputLength+this.offset},enumerable:!0,configurable:!0}),e.prototype.span=function(e){return new fo(e,this.inputIndex)},e.prototype.advance=function(){this.index++},e.prototype.optionalCharacter=function(e){return!!this.next.isCharacter(e)&&(this.advance(),!0)},e.prototype.peekKeywordLet=function(){return this.next.isKeywordLet()},e.prototype.peekKeywordAs=function(){return this.next.isKeywordAs()},e.prototype.expectCharacter=function(e){this.optionalCharacter(e)||this.error("Missing expected "+String.fromCharCode(e))},e.prototype.optionalOperator=function(e){return!!this.next.isOperator(e)&&(this.advance(),!0)},e.prototype.expectOperator=function(e){this.optionalOperator(e)||this.error("Missing expected operator "+e)},e.prototype.expectIdentifierOrKeyword=function(){var e=this.next;return e.isIdentifier()||e.isKeyword()?(this.advance(),e.toString()):(this.error("Unexpected token "+e+", expected identifier or keyword"),"")},e.prototype.expectIdentifierOrKeywordOrString=function(){var e=this.next;return e.isIdentifier()||e.isKeyword()||e.isString()?(this.advance(),e.toString()):(this.error("Unexpected token "+e+", expected identifier, keyword, or string"),"")},e.prototype.parseChain=function(){for(var e=[],t=this.inputIndex;this.index<this.tokens.length;){var n=this.parsePipe();if(e.push(n),this.optionalCharacter(Aa))for(this.parseAction||this.error("Binding expression cannot contain chained expression");this.optionalCharacter(Aa););else this.index<this.tokens.length&&this.error("Unexpected token '"+this.next+"'")}return 0==e.length?new go(this.span(t)):1==e.length?e[0]:new _o(this.span(t),e)},e.prototype.parsePipe=function(){var e=this.parseExpression();if(this.optionalOperator("|")){this.parseAction&&this.error("Cannot have a pipe in an action expression");do{for(var t=this.expectIdentifierOrKeyword(),n=[];this.optionalCharacter(58);)n.push(this.parseExpression());e=new Co(this.span(e.span.start),e,t,n)}while(this.optionalOperator("|"))}return e},e.prototype.parseExpression=function(){return this.parseConditional()},e.prototype.parseConditional=function(){var e=this.inputIndex,t=this.parseLogicalOr();if(this.optionalOperator("?")){var n=this.parsePipe(),r=void 0;if(this.optionalCharacter(58))r=this.parsePipe();else{var i=this.inputIndex,o=this.input.substring(e,i);this.error("Conditional expression "+o+" requires all 3 expressions"),r=new go(this.span(e))}return new yo(this.span(e),t,n,r)}return t},e.prototype.parseLogicalOr=function(){for(var e=this.parseLogicalAnd();this.optionalOperator("||");){var t=this.parseLogicalAnd();e=new Ao(this.span(e.span.start),"||",e,t)}return e},e.prototype.parseLogicalAnd=function(){for(var e=this.parseEquality();this.optionalOperator("&&");){var t=this.parseEquality();e=new Ao(this.span(e.span.start),"&&",e,t)}return e},e.prototype.parseEquality=function(){for(var e=this.parseRelational();this.next.type==zs.Operator;){var t=this.next.strValue;switch(t){case"==":case"===":case"!=":case"!==":this.advance();var n=this.parseRelational();e=new Ao(this.span(e.span.start),t,e,n);continue}break}return e},e.prototype.parseRelational=function(){for(var e=this.parseAdditive();this.next.type==zs.Operator;){var t=this.next.strValue;switch(t){case"<":case">":case"<=":case">=":this.advance();var n=this.parseAdditive();e=new Ao(this.span(e.span.start),t,e,n);continue}break}return e},e.prototype.parseAdditive=function(){for(var e=this.parseMultiplicative();this.next.type==zs.Operator;){var t=this.next.strValue;switch(t){case"+":case"-":this.advance();var n=this.parseMultiplicative();e=new Ao(this.span(e.span.start),t,e,n);continue}break}return e},e.prototype.parseMultiplicative=function(){for(var e=this.parsePrefix();this.next.type==zs.Operator;){var t=this.next.strValue;switch(t){case"*":case"%":case"/":this.advance();var n=this.parsePrefix();e=new Ao(this.span(e.span.start),t,e,n);continue}break}return e},e.prototype.parsePrefix=function(){if(this.next.type==zs.Operator){var e=this.inputIndex,t=this.next.strValue,n=void 0;switch(t){case"+":return this.advance(),n=this.parsePrefix(),new Ao(this.span(e),"-",n,new Eo(new fo(e,e),0));case"-":return this.advance(),n=this.parsePrefix(),new Ao(this.span(e),t,new Eo(new fo(e,e),0),n);case"!":return this.advance(),n=this.parsePrefix(),new To(this.span(e),n)}}return this.parseCallChain()},e.prototype.parseCallChain=function(){for(var e=this.parsePrimary();;)if(this.optionalCharacter(46))e=this.parseAccessMemberOrMethodCall(e,!1);else if(this.optionalOperator("?."))e=this.parseAccessMemberOrMethodCall(e,!0);else if(this.optionalCharacter(91)){this.rbracketsExpected++;var t=this.parsePipe();if(this.rbracketsExpected--,this.expectCharacter(93),this.optionalOperator("=")){var n=this.parseConditional();e=new wo(this.span(e.span.start),e,t,n)}else e=new Oo(this.span(e.span.start),e,t)}else if(this.optionalCharacter(40)){this.rparensExpected++;var r=this.parseCallArguments();this.rparensExpected--,this.expectCharacter(41),e=new Ro(this.span(e.span.start),e,r)}else{if(!this.optionalOperator("!"))return e;e=new Do(this.span(e.span.start),e)}},e.prototype.parsePrimary=function(){var e=this.inputIndex;if(this.optionalCharacter(40)){this.rparensExpected++;var t=this.parsePipe();return this.rparensExpected--,this.expectCharacter(41),t}if(this.next.isKeywordNull())return this.advance(),new Eo(this.span(e),null);if(this.next.isKeywordUndefined())return this.advance(),new Eo(this.span(e),void 0);if(this.next.isKeywordTrue())return this.advance(),new Eo(this.span(e),!0);if(this.next.isKeywordFalse())return this.advance(),new Eo(this.span(e),!1);if(this.next.isKeywordThis())return this.advance(),new vo(this.span(e));if(this.optionalCharacter(91)){this.rbracketsExpected++;var n=this.parseExpressionList(93);return this.rbracketsExpected--,this.expectCharacter(93),new So(this.span(e),n)}if(this.next.isCharacter(Ka))return this.parseLiteralMap();if(this.next.isIdentifier())return this.parseAccessMemberOrMethodCall(new vo(this.span(e)),!1);if(this.next.isNumber()){var r=this.next.toNumber();return this.advance(),new Eo(this.span(e),r)}if(this.next.isString()){var i=this.next.toString();return this.advance(),new Eo(this.span(e),i)}return this.index>=this.tokens.length?(this.error("Unexpected end of expression: "+this.input),new go(this.span(e))):(this.error("Unexpected token "+this.next),new go(this.span(e)))},e.prototype.parseExpressionList=function(e){var t=[];if(!this.next.isCharacter(e))do{t.push(this.parsePipe())}while(this.optionalCharacter(44));return t},e.prototype.parseLiteralMap=function(){var e=[],t=[],n=this.inputIndex;if(this.expectCharacter(Ka),!this.optionalCharacter(125)){this.rbracesExpected++;do{var r=this.next.isString(),i=this.expectIdentifierOrKeywordOrString();e.push({key:i,quoted:r}),this.expectCharacter(58),t.push(this.parsePipe())}while(this.optionalCharacter(44));this.rbracesExpected--,this.expectCharacter(125)}return new ko(this.span(n),e,t)},e.prototype.parseAccessMemberOrMethodCall=function(e,t){void 0===t&&(t=!1);var n=e.span.start,r=this.expectIdentifierOrKeyword();if(this.optionalCharacter(40)){this.rparensExpected++;var i=this.parseCallArguments();this.expectCharacter(41),this.rparensExpected--;var o=this.span(n);return t?new Po(o,e,r,i):new Io(o,e,r,i)}if(t)return this.optionalOperator("=")?(this.error("The '?.' operator cannot be used in the assignment"),new go(this.span(n))):new xo(this.span(n),e,r);if(this.optionalOperator("=")){if(!this.parseAction)return this.error("Bindings cannot contain assignments"),new go(this.span(n));var a=this.parseConditional();return new jo(this.span(n),e,r,a)}return new bo(this.span(n),e,r)},e.prototype.parseCallArguments=function(){if(this.next.isCharacter(41))return[];var e=[];do{e.push(this.parsePipe())}while(this.optionalCharacter(44));return e},e.prototype.expectTemplateBindingKey=function(){var e="",t=!1;do{e+=this.expectIdentifierOrKeywordOrString(),(t=this.optionalOperator("-"))&&(e+="-")}while(t);return e.toString()},e.prototype.parseTemplateBindings=function(e){var t=!0,n=[];do{var r=this.inputIndex,i=void 0,o=void 0,a=!1;t?(i=o=e,t=!1):((a=this.peekKeywordLet())&&this.advance(),i=this.expectTemplateBindingKey(),o=a?i:e+i[0].toUpperCase()+i.substring(1),this.optionalCharacter(58));var s=null,u=null;if(a)s=this.optionalOperator("=")?this.expectTemplateBindingKey():"$implicit";else if(this.peekKeywordAs())this.advance(),s=i,o=this.expectTemplateBindingKey(),a=!0;else if(this.next!==$s&&!this.peekKeywordLet()){var c=this.inputIndex,l=this.parsePipe(),d=this.input.substring(c-this.offset,this.inputIndex-this.offset);u=new No(l,d,this.location,this.errors)}if(n.push(new Fo(this.span(r),o,a,s,u)),this.peekKeywordAs()&&!a){var p=this.inputIndex;this.advance();var f=this.expectTemplateBindingKey();n.push(new Fo(this.span(p),f,!0,o,null))}this.optionalCharacter(Aa)||this.optionalCharacter(44)}while(this.index<this.tokens.length);return new au(n,[],this.errors)},e.prototype.error=function(e,t){void 0===t&&(t=null),this.errors.push(new po(e,this.input,this.locationText(t),this.location)),this.skip()},e.prototype.locationText=function(e){return void 0===e&&(e=null),null==e&&(e=this.index),e<this.tokens.length?"at column "+(this.tokens[e].index+1)+" in":"at the end of the expression"},e.prototype.skip=function(){for(var e=this.next;this.index<this.tokens.length&&!e.isCharacter(Aa)&&(this.rparensExpected<=0||!e.isCharacter(41))&&(this.rbracesExpected<=0||!e.isCharacter(125))&&(this.rbracketsExpected<=0||!e.isCharacter(93));)this.next.isError()&&this.errors.push(new po(this.next.toString(),this.input,this.locationText(),this.location)),this.advance(),e=this.next},e}(),lu=function(){function e(){this.errors=[]}return e.check=function(t){var n=new e;return t.visit(n),n.errors},e.prototype.visitImplicitReceiver=function(e,t){},e.prototype.visitInterpolation=function(e,t){},e.prototype.visitLiteralPrimitive=function(e,t){},e.prototype.visitPropertyRead=function(e,t){},e.prototype.visitPropertyWrite=function(e,t){},e.prototype.visitSafePropertyRead=function(e,t){},e.prototype.visitMethodCall=function(e,t){},e.prototype.visitSafeMethodCall=function(e,t){},e.prototype.visitFunctionCall=function(e,t){},e.prototype.visitLiteralArray=function(e,t){this.visitAll(e.expressions)},e.prototype.visitLiteralMap=function(e,t){this.visitAll(e.values)},e.prototype.visitBinary=function(e,t){},e.prototype.visitPrefixNot=function(e,t){},e.prototype.visitNonNullAssert=function(e,t){},e.prototype.visitConditional=function(e,t){},e.prototype.visitPipe=function(e,t){this.errors.push("pipes")},e.prototype.visitKeyedRead=function(e,t){},e.prototype.visitKeyedWrite=function(e,t){},e.prototype.visitAll=function(e){var t=this;return e.map(function(e){return e.visit(t)})},e.prototype.visitChain=function(e,t){},e.prototype.visitQuote=function(e,t){},e}(),du=function(){function e(e,t){void 0===t&&(t=-1),this.path=e,this.position=t}return Object.defineProperty(e.prototype,"empty",{get:function(){return!this.path||!this.path.length},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"head",{get:function(){return this.path[0]},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"tail",{get:function(){return this.path[this.path.length-1]},enumerable:!0,configurable:!0}),e.prototype.parentOf=function(e){return e&&this.path[this.path.indexOf(e)-1]},e.prototype.childOf=function(e){return this.path[this.path.indexOf(e)+1]},e.prototype.first=function(e){for(var t=this.path.length-1;t>=0;t--){var n=this.path[t];if(n instanceof e)return n}},e.prototype.push=function(e){this.path.push(e)},e.prototype.pop=function(){return this.path.pop()},e}(),pu=function(){function e(e,t,n){this.value=e,this.sourceSpan=t,this.i18n=n}return e.prototype.visit=function(e,t){return e.visitText(this,t)},e}(),fu=function(){function e(e,t,n,r,i,o){this.switchValue=e,this.type=t,this.cases=n,this.sourceSpan=r,this.switchValueSourceSpan=i,this.i18n=o}return e.prototype.visit=function(e,t){return e.visitExpansion(this,t)},e}(),hu=function(){function e(e,t,n,r,i){this.value=e,this.expression=t,this.sourceSpan=n,this.valueSourceSpan=r,this.expSourceSpan=i}return e.prototype.visit=function(e,t){return e.visitExpansionCase(this,t)},e}(),mu=function(){function e(e,t,n,r,i){this.name=e,this.value=t,this.sourceSpan=n,this.valueSpan=r,this.i18n=i}return e.prototype.visit=function(e,t){return e.visitAttribute(this,t)},e}(),gu=function(){function e(e,t,n,r,i,o,a){void 0===i&&(i=null),void 0===o&&(o=null),this.name=e,this.attrs=t,this.children=n,this.sourceSpan=r,this.startSourceSpan=i,this.endSourceSpan=o,this.i18n=a}return e.prototype.visit=function(e,t){return e.visitElement(this,t)},e}(),vu=function(){function e(e,t){this.value=e,this.sourceSpan=t}return e.prototype.visit=function(e,t){return e.visitComment(this,t)},e}();function _u(e,t,n){void 0===n&&(n=null);var r=[],i=e.visit?function(t){return e.visit(t,n)||t.visit(e,n)}:function(t){return t.visit(e,n)};return t.forEach(function(e){var t=i(e);t&&r.push(t)}),r}var yu,bu,ju=function(){function e(){}return e.prototype.visitElement=function(e,t){this.visitChildren(t,function(t){t(e.attrs),t(e.children)})},e.prototype.visitAttribute=function(e,t){},e.prototype.visitText=function(e,t){},e.prototype.visitComment=function(e,t){},e.prototype.visitExpansion=function(e,t){return this.visitChildren(t,function(t){t(e.cases)})},e.prototype.visitExpansionCase=function(e,t){},e.prototype.visitChildren=function(e,t){var n=[],r=this;return t(function(t){t&&n.push(_u(r,t,e))}),[].concat.apply([],n)},e}();function xu(e,t){var n=[];return _u(new(function(e){function r(){return null!==e&&e.apply(this,arguments)||this}return Object(o.__extends)(r,e),r.prototype.visit=function(e,r){var i=function e(t){var n=t.sourceSpan.start.offset,r=t.sourceSpan.end.offset;return t instanceof gu&&(t.endSourceSpan?r=t.endSourceSpan.end.offset:t.children&&t.children.length&&(r=e(t.children[t.children.length-1]).end)),{start:n,end:r}}(e);if(!(i.start<=t&&t<i.end))return!0;n.push(e)},r}(ju)),e),new du(n,t)}(bu=yu||(yu={}))[bu.TAG_OPEN_START=0]="TAG_OPEN_START",bu[bu.TAG_OPEN_END=1]="TAG_OPEN_END",bu[bu.TAG_OPEN_END_VOID=2]="TAG_OPEN_END_VOID",bu[bu.TAG_CLOSE=3]="TAG_CLOSE",bu[bu.TEXT=4]="TEXT",bu[bu.ESCAPABLE_RAW_TEXT=5]="ESCAPABLE_RAW_TEXT",bu[bu.RAW_TEXT=6]="RAW_TEXT",bu[bu.COMMENT_START=7]="COMMENT_START",bu[bu.COMMENT_END=8]="COMMENT_END",bu[bu.CDATA_START=9]="CDATA_START",bu[bu.CDATA_END=10]="CDATA_END",bu[bu.ATTR_NAME=11]="ATTR_NAME",bu[bu.ATTR_VALUE=12]="ATTR_VALUE",bu[bu.DOC_TYPE=13]="DOC_TYPE",bu[bu.EXPANSION_FORM_START=14]="EXPANSION_FORM_START",bu[bu.EXPANSION_CASE_VALUE=15]="EXPANSION_CASE_VALUE",bu[bu.EXPANSION_CASE_EXP_START=16]="EXPANSION_CASE_EXP_START",bu[bu.EXPANSION_CASE_EXP_END=17]="EXPANSION_CASE_EXP_END",bu[bu.EXPANSION_FORM_END=18]="EXPANSION_FORM_END",bu[bu.EOF=19]="EOF";var Ou=function(){return function(e,t,n){this.type=e,this.parts=t,this.sourceSpan=n}}(),wu=function(e){function t(t,n,r){var i=e.call(this,r,t)||this;return i.tokenType=n,i}return Object(o.__extends)(t,e),t}(os),Cu=function(){return function(e,t){this.tokens=e,this.errors=t}}();var Eu=/\r\n?/g;function Su(e){return'Unexpected character "'+(e===va?"EOF":String.fromCharCode(e))+'"'}function ku(e){return'Unknown entity "'+e+'" - use the "&#<decimal>;" or  "&#x<hex>;" syntax'}var Mu=function(){return function(e){this.error=e}}(),Au=function(){function e(e,t,n,r){void 0===r&&(r=Ui),this._file=e,this._getTagDefinition=t,this._tokenizeIcu=n,this._interpolationConfig=r,this._peek=-1,this._nextPeek=-1,this._index=-1,this._line=0,this._column=-1,this._expansionCaseStack=[],this._inInterpolation=!1,this.tokens=[],this.errors=[],this._input=e.content,this._length=e.content.length,this._advance()}return e.prototype._processCarriageReturns=function(e){return e.replace(Eu,"\n")},e.prototype.tokenize=function(){for(;this._peek!==va;){var e=this._getLocation();try{this._attemptCharCode(60)?this._attemptCharCode(33)?this._attemptCharCode(91)?this._consumeCdata(e):this._attemptCharCode(ka)?this._consumeComment(e):this._consumeDocType(e):this._attemptCharCode(Ma)?this._consumeTagClose(e):this._consumeTagOpen(e):this._tokenizeIcu&&this._tokenizeExpansionForm()||this._consumeText()}catch(e){if(!(e instanceof Mu))throw e;this.errors.push(e.error)}}return this._beginToken(yu.EOF),this._endToken([]),new Cu(function(e){for(var t=[],n=void 0,r=0;r<e.length;r++){var i=e[r];n&&n.type==yu.TEXT&&i.type==yu.TEXT?(n.parts[0]+=i.parts[0],n.sourceSpan.end=i.sourceSpan.end):(n=i,t.push(n))}return t}(this.tokens),this.errors)},e.prototype._tokenizeExpansionForm=function(){if(Ru(this._input,this._index,this._interpolationConfig))return this._consumeExpansionFormStart(),!0;if(((e=this._peek)===Ta||Ja(e)||$a(e))&&this._isInExpansionForm())return this._consumeExpansionCaseStart(),!0;var e;if(125===this._peek){if(this._isInExpansionCase())return this._consumeExpansionCaseEnd(),!0;if(this._isInExpansionForm())return this._consumeExpansionFormEnd(),!0}return!1},e.prototype._getLocation=function(){return new ns(this._file,this._index,this._line,this._column)},e.prototype._getSpan=function(e,t){return void 0===e&&(e=this._getLocation()),void 0===t&&(t=this._getLocation()),new is(e,t)},e.prototype._beginToken=function(e,t){void 0===t&&(t=this._getLocation()),this._currentTokenStart=t,this._currentTokenType=e},e.prototype._endToken=function(e,t){void 0===t&&(t=this._getLocation());var n=new Ou(this._currentTokenType,e,new is(this._currentTokenStart,t));return this.tokens.push(n),this._currentTokenStart=null,this._currentTokenType=null,n},e.prototype._createError=function(e,t){this._isInExpansionForm()&&(e+=' (Do you have an unescaped "{" in your template? Use "{{ \'{\' }}") to escape it.)');var n=new wu(e,this._currentTokenType,t);return this._currentTokenStart=null,this._currentTokenType=null,new Mu(n)},e.prototype._advance=function(){if(this._index>=this._length)throw this._createError(Su(va),this._getSpan());this._peek===ya?(this._line++,this._column=0):this._peek!==ya&&this._peek!==xa&&this._column++,this._index++,this._peek=this._index>=this._length?va:this._input.charCodeAt(this._index),this._nextPeek=this._index+1>=this._length?va:this._input.charCodeAt(this._index+1)},e.prototype._attemptCharCode=function(e){return this._peek===e&&(this._advance(),!0)},e.prototype._attemptCharCodeCaseInsensitive=function(e){return t=this._peek,n=e,Nu(t)==Nu(n)&&(this._advance(),!0);var t,n},e.prototype._requireCharCode=function(e){var t=this._getLocation();if(!this._attemptCharCode(e))throw this._createError(Su(this._peek),this._getSpan(t,t))},e.prototype._attemptStr=function(e){var t=e.length;if(this._index+t>this._length)return!1;for(var n=this._savePosition(),r=0;r<t;r++)if(!this._attemptCharCode(e.charCodeAt(r)))return this._restorePosition(n),!1;return!0},e.prototype._attemptStrCaseInsensitive=function(e){for(var t=0;t<e.length;t++)if(!this._attemptCharCodeCaseInsensitive(e.charCodeAt(t)))return!1;return!0},e.prototype._requireStr=function(e){var t=this._getLocation();if(!this._attemptStr(e))throw this._createError(Su(this._peek),this._getSpan(t))},e.prototype._attemptCharCodeUntilFn=function(e){for(;!e(this._peek);)this._advance()},e.prototype._requireCharCodeUntilFn=function(e,t){var n=this._getLocation();if(this._attemptCharCodeUntilFn(e),this._index-n.offset<t)throw this._createError(Su(this._peek),this._getSpan(n,n))},e.prototype._attemptUntilChar=function(e){for(;this._peek!==e;)this._advance()},e.prototype._readChar=function(e){if(e&&38===this._peek)return this._decodeEntity();var t=this._index;return this._advance(),this._input[t]},e.prototype._decodeEntity=function(){var e=this._getLocation();if(this._advance(),!this._attemptCharCode(35)){var t=this._savePosition();if(this._attemptCharCodeUntilFn(Pu),this._peek!=Aa)return this._restorePosition(t),"&";this._advance();var n=this._input.substring(e.offset+1,this._index-1),r=p[n];if(!r)throw this._createError(ku(n),this._getSpan(e));return r}var i=this._attemptCharCode(120)||this._attemptCharCode(88),o=this._getLocation().offset;if(this._attemptCharCodeUntilFn(Iu),this._peek!=Aa)throw this._createError(Su(this._peek),this._getSpan());this._advance();var a=this._input.substring(o,this._index-1);try{var s=parseInt(a,i?16:10);return String.fromCharCode(s)}catch(t){var u=this._input.substring(e.offset+1,this._index-1);throw this._createError(ku(u),this._getSpan(e))}},e.prototype._consumeRawText=function(e,t,n){var r,i=this._getLocation();this._beginToken(e?yu.ESCAPABLE_RAW_TEXT:yu.RAW_TEXT,i);for(var o=[];r=this._getLocation(),!this._attemptCharCode(t)||!n();)for(this._index>r.offset&&o.push(this._input.substring(r.offset,this._index));this._peek!==t;)o.push(this._readChar(e));return this._endToken([this._processCarriageReturns(o.join(""))],r)},e.prototype._consumeComment=function(e){var t=this;this._beginToken(yu.COMMENT_START,e),this._requireCharCode(ka),this._endToken([]);var n=this._consumeRawText(!1,ka,function(){return t._attemptStr("->")});this._beginToken(yu.COMMENT_END,n.sourceSpan.end),this._endToken([])},e.prototype._consumeCdata=function(e){var t=this;this._beginToken(yu.CDATA_START,e),this._requireStr("CDATA["),this._endToken([]);var n=this._consumeRawText(!1,93,function(){return t._attemptStr("]>")});this._beginToken(yu.CDATA_END,n.sourceSpan.end),this._endToken([])},e.prototype._consumeDocType=function(e){this._beginToken(yu.DOC_TYPE,e),this._attemptUntilChar(Da),this._advance(),this._endToken([this._input.substring(e.offset+2,this._index-1)])},e.prototype._consumePrefixAndName=function(){for(var e,t,n=this._index,r=null;58!==this._peek&&!(((e=this._peek)<Ba||Ya<e)&&(e<Ra||La<e)&&(e<Ia||e>Pa));)this._advance();return 58===this._peek?(this._advance(),r=this._input.substring(n,this._index-1),t=this._index):t=n,this._requireCharCodeUntilFn(Du,this._index===t?1:0),[r,this._input.substring(t,this._index)]},e.prototype._consumeTagOpen=function(e){var t,n,i=this._savePosition();try{if(!Ja(this._peek))throw this._createError(Su(this._peek),this._getSpan());var o=this._index;for(this._consumeTagOpenStart(e),n=(t=this._input.substring(o,this._index)).toLowerCase(),this._attemptCharCodeUntilFn(Tu);this._peek!==Ma&&this._peek!==Da;)this._consumeAttributeName(),this._attemptCharCodeUntilFn(Tu),this._attemptCharCode(Ta)&&(this._attemptCharCodeUntilFn(Tu),this._consumeAttributeValue()),this._attemptCharCodeUntilFn(Tu);this._consumeTagOpenEnd()}catch(t){if(t instanceof Mu)return this._restorePosition(i),this._beginToken(yu.TEXT,e),void this._endToken(["<"]);throw t}var a=this._getTagDefinition(t).contentType;a===r.RAW_TEXT?this._consumeRawTextWithTagClose(n,!1):a===r.ESCAPABLE_RAW_TEXT&&this._consumeRawTextWithTagClose(n,!0)},e.prototype._consumeRawTextWithTagClose=function(e,t){var n=this,r=this._consumeRawText(t,60,function(){return!!n._attemptCharCode(Ma)&&(n._attemptCharCodeUntilFn(Tu),!!n._attemptStrCaseInsensitive(e)&&(n._attemptCharCodeUntilFn(Tu),n._attemptCharCode(Da)))});this._beginToken(yu.TAG_CLOSE,r.sourceSpan.end),this._endToken([null,e])},e.prototype._consumeTagOpenStart=function(e){this._beginToken(yu.TAG_OPEN_START,e);var t=this._consumePrefixAndName();this._endToken(t)},e.prototype._consumeAttributeName=function(){this._beginToken(yu.ATTR_NAME);var e=this._consumePrefixAndName();this._endToken(e)},e.prototype._consumeAttributeValue=function(){var e;if(this._beginToken(yu.ATTR_VALUE),this._peek===Ea||this._peek===wa){var t=this._peek;this._advance();for(var n=[];this._peek!==t;)n.push(this._readChar(!0));e=n.join(""),this._advance()}else{var r=this._index;this._requireCharCodeUntilFn(Du,1),e=this._input.substring(r,this._index)}this._endToken([this._processCarriageReturns(e)])},e.prototype._consumeTagOpenEnd=function(){var e=this._attemptCharCode(Ma)?yu.TAG_OPEN_END_VOID:yu.TAG_OPEN_END;this._beginToken(e),this._requireCharCode(Da),this._endToken([])},e.prototype._consumeTagClose=function(e){this._beginToken(yu.TAG_CLOSE,e),this._attemptCharCodeUntilFn(Tu);var t=this._consumePrefixAndName();this._attemptCharCodeUntilFn(Tu),this._requireCharCode(Da),this._endToken(t)},e.prototype._consumeExpansionFormStart=function(){this._beginToken(yu.EXPANSION_FORM_START,this._getLocation()),this._requireCharCode(Ka),this._endToken([]),this._expansionCaseStack.push(yu.EXPANSION_FORM_START),this._beginToken(yu.RAW_TEXT,this._getLocation());var e=this._readUntil(44);this._endToken([e],this._getLocation()),this._requireCharCode(44),this._attemptCharCodeUntilFn(Tu),this._beginToken(yu.RAW_TEXT,this._getLocation());var t=this._readUntil(44);this._endToken([t],this._getLocation()),this._requireCharCode(44),this._attemptCharCodeUntilFn(Tu)},e.prototype._consumeExpansionCaseStart=function(){this._beginToken(yu.EXPANSION_CASE_VALUE,this._getLocation());var e=this._readUntil(Ka).trim();this._endToken([e],this._getLocation()),this._attemptCharCodeUntilFn(Tu),this._beginToken(yu.EXPANSION_CASE_EXP_START,this._getLocation()),this._requireCharCode(Ka),this._endToken([],this._getLocation()),this._attemptCharCodeUntilFn(Tu),this._expansionCaseStack.push(yu.EXPANSION_CASE_EXP_START)},e.prototype._consumeExpansionCaseEnd=function(){this._beginToken(yu.EXPANSION_CASE_EXP_END,this._getLocation()),this._requireCharCode(125),this._endToken([],this._getLocation()),this._attemptCharCodeUntilFn(Tu),this._expansionCaseStack.pop()},e.prototype._consumeExpansionFormEnd=function(){this._beginToken(yu.EXPANSION_FORM_END,this._getLocation()),this._requireCharCode(125),this._endToken([]),this._expansionCaseStack.pop()},e.prototype._consumeText=function(){var e=this._getLocation();this._beginToken(yu.TEXT,e);var t=[];do{this._interpolationConfig&&this._attemptStr(this._interpolationConfig.start)?(t.push(this._interpolationConfig.start),this._inInterpolation=!0):this._interpolationConfig&&this._inInterpolation&&this._attemptStr(this._interpolationConfig.end)?(t.push(this._interpolationConfig.end),this._inInterpolation=!1):t.push(this._readChar(!0))}while(!this._isTextEnd());this._endToken([this._processCarriageReturns(t.join(""))])},e.prototype._isTextEnd=function(){if(60===this._peek||this._peek===va)return!0;if(this._tokenizeIcu&&!this._inInterpolation){if(Ru(this._input,this._index,this._interpolationConfig))return!0;if(125===this._peek&&this._isInExpansionCase())return!0}return!1},e.prototype._savePosition=function(){return[this._peek,this._index,this._column,this._line,this.tokens.length]},e.prototype._readUntil=function(e){var t=this._index;return this._attemptUntilChar(e),this._input.substring(t,this._index)},e.prototype._restorePosition=function(e){this._peek=e[0],this._index=e[1],this._column=e[2],this._line=e[3];var t=e[4];t<this.tokens.length&&(this.tokens=this.tokens.slice(0,t))},e.prototype._isInExpansionCase=function(){return this._expansionCaseStack.length>0&&this._expansionCaseStack[this._expansionCaseStack.length-1]===yu.EXPANSION_CASE_EXP_START},e.prototype._isInExpansionForm=function(){return this._expansionCaseStack.length>0&&this._expansionCaseStack[this._expansionCaseStack.length-1]===yu.EXPANSION_FORM_START},e}();function Tu(e){return!Za(e)||e===va}function Du(e){return Za(e)||e===Da||e===Ma||e===Ea||e===wa||e===Ta}function Iu(e){return e==Aa||e==va||!((t=e)>=Ba&&t<=Ua||t>=Ra&&t<=Fa||$a(t));var t}function Pu(e){return e==Aa||e==va||!Ja(e)}function Ru(e,t,n){var r=!!n&&e.indexOf(n.start,t)==t;return e.charCodeAt(t)==Ka&&!r}function Nu(e){return e>=Ba&&e<=Ya?e-Ba+Ra:e}var Fu=function(e){function t(t,n,r){var i=e.call(this,n,r)||this;return i.elementName=t,i}return Object(o.__extends)(t,e),t.create=function(e,n,r){return new t(e,n,r)},t}(os),Lu=function(){return function(e,t){this.rootNodes=e,this.errors=t}}(),Vu=function(){function e(e){this.getTagDefinition=e}return e.prototype.parse=function(e,t,n,r){void 0===n&&(n=!1),void 0===r&&(r=Ui);var i,o,a,s,u,c=(i=e,o=t,a=this.getTagDefinition,u=r,void 0===(s=n)&&(s=!1),void 0===u&&(u=Ui),new Au(new rs(i,o),a,s,u).tokenize()),l=new Bu(c.tokens,this.getTagDefinition).build();return new Lu(l.rootNodes,c.errors.concat(l.errors))},e}(),Bu=function(){function e(e,t){this.tokens=e,this.getTagDefinition=t,this._index=-1,this._rootNodes=[],this._errors=[],this._elementStack=[],this._advance()}return e.prototype.build=function(){for(;this._peek.type!==yu.EOF;)this._peek.type===yu.TAG_OPEN_START?this._consumeStartTag(this._advance()):this._peek.type===yu.TAG_CLOSE?this._consumeEndTag(this._advance()):this._peek.type===yu.CDATA_START?(this._closeVoidElement(),this._consumeCdata(this._advance())):this._peek.type===yu.COMMENT_START?(this._closeVoidElement(),this._consumeComment(this._advance())):this._peek.type===yu.TEXT||this._peek.type===yu.RAW_TEXT||this._peek.type===yu.ESCAPABLE_RAW_TEXT?(this._closeVoidElement(),this._consumeText(this._advance())):this._peek.type===yu.EXPANSION_FORM_START?this._consumeExpansion(this._advance()):this._advance();return new Lu(this._rootNodes,this._errors)},e.prototype._advance=function(){var e=this._peek;return this._index<this.tokens.length-1&&this._index++,this._peek=this.tokens[this._index],e},e.prototype._advanceIf=function(e){return this._peek.type===e?this._advance():null},e.prototype._consumeCdata=function(e){this._consumeText(this._advance()),this._advanceIf(yu.CDATA_END)},e.prototype._consumeComment=function(e){var t=this._advanceIf(yu.RAW_TEXT);this._advanceIf(yu.COMMENT_END);var n=null!=t?t.parts[0].trim():null;this._addToParent(new vu(n,e.sourceSpan))},e.prototype._consumeExpansion=function(e){for(var t=this._advance(),n=this._advance(),r=[];this._peek.type===yu.EXPANSION_CASE_VALUE;){var i=this._parseExpansionCase();if(!i)return;r.push(i)}if(this._peek.type===yu.EXPANSION_FORM_END){var o=new is(e.sourceSpan.start,this._peek.sourceSpan.end);this._addToParent(new fu(t.parts[0],n.parts[0],r,o,t.sourceSpan)),this._advance()}else this._errors.push(Fu.create(null,this._peek.sourceSpan,"Invalid ICU message. Missing '}'."))},e.prototype._parseExpansionCase=function(){var t=this._advance();if(this._peek.type!==yu.EXPANSION_CASE_EXP_START)return this._errors.push(Fu.create(null,this._peek.sourceSpan,"Invalid ICU message. Missing '{'.")),null;var n=this._advance(),r=this._collectExpansionExpTokens(n);if(!r)return null;var i=this._advance();r.push(new Ou(yu.EOF,[],i.sourceSpan));var o=new e(r,this.getTagDefinition).build();if(o.errors.length>0)return this._errors=this._errors.concat(o.errors),null;var a=new is(t.sourceSpan.start,i.sourceSpan.end),s=new is(n.sourceSpan.start,i.sourceSpan.end);return new hu(t.parts[0],o.rootNodes,a,t.sourceSpan,s)},e.prototype._collectExpansionExpTokens=function(e){for(var t=[],n=[yu.EXPANSION_CASE_EXP_START];;){if(this._peek.type!==yu.EXPANSION_FORM_START&&this._peek.type!==yu.EXPANSION_CASE_EXP_START||n.push(this._peek.type),this._peek.type===yu.EXPANSION_CASE_EXP_END){if(!zu(n,yu.EXPANSION_CASE_EXP_START))return this._errors.push(Fu.create(null,e.sourceSpan,"Invalid ICU message. Missing '}'.")),null;if(n.pop(),0==n.length)return t}if(this._peek.type===yu.EXPANSION_FORM_END){if(!zu(n,yu.EXPANSION_FORM_START))return this._errors.push(Fu.create(null,e.sourceSpan,"Invalid ICU message. Missing '}'.")),null;n.pop()}if(this._peek.type===yu.EOF)return this._errors.push(Fu.create(null,e.sourceSpan,"Invalid ICU message. Missing '}'.")),null;t.push(this._advance())}},e.prototype._consumeText=function(e){var t=e.parts[0];if(t.length>0&&"\n"==t[0]){var n=this._getParentElement();null!=n&&0==n.children.length&&this.getTagDefinition(n.name).ignoreFirstLf&&(t=t.substring(1))}t.length>0&&this._addToParent(new pu(t,e.sourceSpan))},e.prototype._closeVoidElement=function(){var e=this._getParentElement();e&&this.getTagDefinition(e.name).isVoid&&this._elementStack.pop()},e.prototype._consumeStartTag=function(e){for(var t=e.parts[0],n=e.parts[1],r=[];this._peek.type===yu.ATTR_NAME;)r.push(this._consumeAttr(this._advance()));var i=this._getElementFullName(t,n,this._getParentElement()),o=!1;if(this._peek.type===yu.TAG_OPEN_END_VOID){this._advance(),o=!0;var a=this.getTagDefinition(i);a.canSelfClose||null!==l(i)||a.isVoid||this._errors.push(Fu.create(i,e.sourceSpan,'Only void and foreign elements can be self closed "'+e.parts[1]+'"'))}else this._peek.type===yu.TAG_OPEN_END&&(this._advance(),o=!1);var s=this._peek.sourceSpan.start,u=new is(e.sourceSpan.start,s),c=new gu(i,r,[],u,u,void 0);this._pushElement(c),o&&(this._popElement(i),c.endSourceSpan=u)},e.prototype._pushElement=function(e){var t=this._getParentElement();t&&this.getTagDefinition(t.name).isClosedByChild(e.name)&&this._elementStack.pop();var n=this.getTagDefinition(e.name),r=this._getParentElementSkippingContainers(),i=r.parent,o=r.container;if(i&&n.requireExtraParent(i.name)){var a=new gu(n.parentToAdd,[],[],e.sourceSpan,e.startSourceSpan,e.endSourceSpan);this._insertBeforeContainer(i,o,a)}this._addToParent(e),this._elementStack.push(e)},e.prototype._consumeEndTag=function(e){var t=this._getElementFullName(e.parts[0],e.parts[1],this._getParentElement());if(this._getParentElement()&&(this._getParentElement().endSourceSpan=e.sourceSpan),this.getTagDefinition(t).isVoid)this._errors.push(Fu.create(t,e.sourceSpan,'Void elements do not have end tags "'+e.parts[1]+'"'));else if(!this._popElement(t)){var n='Unexpected closing tag "'+t+'". It may happen when the tag has already been closed by another tag. For more info see https://www.w3.org/TR/html5/syntax.html#closing-elements-that-have-implied-end-tags';this._errors.push(Fu.create(t,e.sourceSpan,n))}},e.prototype._popElement=function(e){for(var t=this._elementStack.length-1;t>=0;t--){var n=this._elementStack[t];if(n.name==e)return this._elementStack.splice(t,this._elementStack.length-t),!0;if(!this.getTagDefinition(n.name).closedByParent)return!1}return!1},e.prototype._consumeAttr=function(e){var t=d(e.parts[0],e.parts[1]),n=e.sourceSpan.end,r="",i=void 0;if(this._peek.type===yu.ATTR_VALUE){var o=this._advance();r=o.parts[0],n=o.sourceSpan.end,i=o.sourceSpan}return new mu(t,r,new is(e.sourceSpan.start,n),i)},e.prototype._getParentElement=function(){return this._elementStack.length>0?this._elementStack[this._elementStack.length-1]:null},e.prototype._getParentElementSkippingContainers=function(){for(var e=null,t=this._elementStack.length-1;t>=0;t--){if(!s(this._elementStack[t].name))return{parent:this._elementStack[t],container:e};e=this._elementStack[t]}return{parent:null,container:e}},e.prototype._addToParent=function(e){var t=this._getParentElement();null!=t?t.children.push(e):this._rootNodes.push(e)},e.prototype._insertBeforeContainer=function(e,t,n){if(t){if(e){var r=e.children.indexOf(t);e.children[r]=n}else this._rootNodes.push(n);n.children.push(t),this._elementStack.splice(this._elementStack.indexOf(t),0,n)}else this._addToParent(n),this._elementStack.push(n)},e.prototype._getElementFullName=function(e,t,n){return null==e&&null==(e=this.getTagDefinition(t).implicitNamespacePrefix)&&null!=n&&(e=l(n.name)),d(e,t)},e}();function zu(e,t){return e.length>0&&e[e.length-1]===t}var Uu=function(e){function t(){return e.call(this,v)||this}return Object(o.__extends)(t,e),t.prototype.parse=function(t,n,r,i){return void 0===r&&(r=!1),void 0===i&&(i=Ui),e.prototype.parse.call(this,t,n,r,i)},t}(Vu),Hu="ngPreserveWhitespaces",qu=new Set(["pre","template","textarea","script","style"]),Gu=" \f\n\r\t\v ᠎ - \u2028\u2029  　\ufeff",Wu=new RegExp("[^"+Gu+"]"),Yu=new RegExp("["+Gu+"]{2,}","g");function Ku(e){return e.replace(new RegExp(f,"g")," ")}var Qu,Xu=function(){function e(){}return e.prototype.visitElement=function(e,t){return qu.has(e.name)||e.attrs.some(function(e){return e.name===Hu})?new gu(e.name,_u(this,e.attrs),e.children,e.sourceSpan,e.startSourceSpan,e.endSourceSpan,e.i18n):new gu(e.name,e.attrs,_u(this,e.children),e.sourceSpan,e.startSourceSpan,e.endSourceSpan,e.i18n)},e.prototype.visitAttribute=function(e,t){return e.name!==Hu?e:null},e.prototype.visitText=function(e,t){return e.value.match(Wu)?new pu(Ku(e.value).replace(Yu," "),e.sourceSpan,e.i18n):null},e.prototype.visitComment=function(e,t){return e},e.prototype.visitExpansion=function(e,t){return e},e.prototype.visitExpansionCase=function(e,t){return e},e}();function Zu(e){return new Lu(_u(new Xu,e.rootNodes),e.errors)}function $u(){return Qu||(Qu={},Ju(R.HTML,["iframe|srcdoc","*|innerHTML","*|outerHTML"]),Ju(R.STYLE,["*|style"]),Ju(R.URL,["*|formAction","area|href","area|ping","audio|src","a|href","a|ping","blockquote|cite","body|background","del|cite","form|action","img|src","img|srcset","input|src","ins|cite","q|cite","source|src","source|srcset","track|src","video|poster","video|src"]),Ju(R.RESOURCE_URL,["applet|code","applet|codebase","base|href","embed|src","frame|src","head|profile","html|manifest","iframe|src","link|href","media|src","object|codebase","object|data","script|src"])),Qu}function Ju(e,t){var n,r;try{for(var i=Object(o.__values)(t),a=i.next();!a.done;a=i.next()){var s=a.value;Qu[s.toLowerCase()]=e}}catch(e){n={error:e}}finally{try{a&&!a.done&&(r=i.return)&&r.call(i)}finally{if(n)throw n.error}}}var ec=function(){return function(){}}(),tc="boolean",nc="number",rc="string",ic="object",oc=["[Element]|textContent,%classList,className,id,innerHTML,*beforecopy,*beforecut,*beforepaste,*copy,*cut,*paste,*search,*selectstart,*webkitfullscreenchange,*webkitfullscreenerror,*wheel,outerHTML,#scrollLeft,#scrollTop,slot,*message,*mozfullscreenchange,*mozfullscreenerror,*mozpointerlockchange,*mozpointerlockerror,*webglcontextcreationerror,*webglcontextlost,*webglcontextrestored","[HTMLElement]^[Element]|accessKey,contentEditable,dir,!draggable,!hidden,innerText,lang,*abort,*auxclick,*blur,*cancel,*canplay,*canplaythrough,*change,*click,*close,*contextmenu,*cuechange,*dblclick,*drag,*dragend,*dragenter,*dragleave,*dragover,*dragstart,*drop,*durationchange,*emptied,*ended,*error,*focus,*gotpointercapture,*input,*invalid,*keydown,*keypress,*keyup,*load,*loadeddata,*loadedmetadata,*loadstart,*lostpointercapture,*mousedown,*mouseenter,*mouseleave,*mousemove,*mouseout,*mouseover,*mouseup,*mousewheel,*pause,*play,*playing,*pointercancel,*pointerdown,*pointerenter,*pointerleave,*pointermove,*pointerout,*pointerover,*pointerup,*progress,*ratechange,*reset,*resize,*scroll,*seeked,*seeking,*select,*show,*stalled,*submit,*suspend,*timeupdate,*toggle,*volumechange,*waiting,outerText,!spellcheck,%style,#tabIndex,title,!translate","abbr,address,article,aside,b,bdi,bdo,cite,code,dd,dfn,dt,em,figcaption,figure,footer,header,i,kbd,main,mark,nav,noscript,rb,rp,rt,rtc,ruby,s,samp,section,small,strong,sub,sup,u,var,wbr^[HTMLElement]|accessKey,contentEditable,dir,!draggable,!hidden,innerText,lang,*abort,*auxclick,*blur,*cancel,*canplay,*canplaythrough,*change,*click,*close,*contextmenu,*cuechange,*dblclick,*drag,*dragend,*dragenter,*dragleave,*dragover,*dragstart,*drop,*durationchange,*emptied,*ended,*error,*focus,*gotpointercapture,*input,*invalid,*keydown,*keypress,*keyup,*load,*loadeddata,*loadedmetadata,*loadstart,*lostpointercapture,*mousedown,*mouseenter,*mouseleave,*mousemove,*mouseout,*mouseover,*mouseup,*mousewheel,*pause,*play,*playing,*pointercancel,*pointerdown,*pointerenter,*pointerleave,*pointermove,*pointerout,*pointerover,*pointerup,*progress,*ratechange,*reset,*resize,*scroll,*seeked,*seeking,*select,*show,*stalled,*submit,*suspend,*timeupdate,*toggle,*volumechange,*waiting,outerText,!spellcheck,%style,#tabIndex,title,!translate","media^[HTMLElement]|!autoplay,!controls,%controlsList,%crossOrigin,#currentTime,!defaultMuted,#defaultPlaybackRate,!disableRemotePlayback,!loop,!muted,*encrypted,*waitingforkey,#playbackRate,preload,src,%srcObject,#volume",":svg:^[HTMLElement]|*abort,*auxclick,*blur,*cancel,*canplay,*canplaythrough,*change,*click,*close,*contextmenu,*cuechange,*dblclick,*drag,*dragend,*dragenter,*dragleave,*dragover,*dragstart,*drop,*durationchange,*emptied,*ended,*error,*focus,*gotpointercapture,*input,*invalid,*keydown,*keypress,*keyup,*load,*loadeddata,*loadedmetadata,*loadstart,*lostpointercapture,*mousedown,*mouseenter,*mouseleave,*mousemove,*mouseout,*mouseover,*mouseup,*mousewheel,*pause,*play,*playing,*pointercancel,*pointerdown,*pointerenter,*pointerleave,*pointermove,*pointerout,*pointerover,*pointerup,*progress,*ratechange,*reset,*resize,*scroll,*seeked,*seeking,*select,*show,*stalled,*submit,*suspend,*timeupdate,*toggle,*volumechange,*waiting,%style,#tabIndex",":svg:graphics^:svg:|",":svg:animation^:svg:|*begin,*end,*repeat",":svg:geometry^:svg:|",":svg:componentTransferFunction^:svg:|",":svg:gradient^:svg:|",":svg:textContent^:svg:graphics|",":svg:textPositioning^:svg:textContent|","a^[HTMLElement]|charset,coords,download,hash,host,hostname,href,hreflang,name,password,pathname,ping,port,protocol,referrerPolicy,rel,rev,search,shape,target,text,type,username","area^[HTMLElement]|alt,coords,download,hash,host,hostname,href,!noHref,password,pathname,ping,port,protocol,referrerPolicy,rel,search,shape,target,username","audio^media|","br^[HTMLElement]|clear","base^[HTMLElement]|href,target","body^[HTMLElement]|aLink,background,bgColor,link,*beforeunload,*blur,*error,*focus,*hashchange,*languagechange,*load,*message,*offline,*online,*pagehide,*pageshow,*popstate,*rejectionhandled,*resize,*scroll,*storage,*unhandledrejection,*unload,text,vLink","button^[HTMLElement]|!autofocus,!disabled,formAction,formEnctype,formMethod,!formNoValidate,formTarget,name,type,value","canvas^[HTMLElement]|#height,#width","content^[HTMLElement]|select","dl^[HTMLElement]|!compact","datalist^[HTMLElement]|","details^[HTMLElement]|!open","dialog^[HTMLElement]|!open,returnValue","dir^[HTMLElement]|!compact","div^[HTMLElement]|align","embed^[HTMLElement]|align,height,name,src,type,width","fieldset^[HTMLElement]|!disabled,name","font^[HTMLElement]|color,face,size","form^[HTMLElement]|acceptCharset,action,autocomplete,encoding,enctype,method,name,!noValidate,target","frame^[HTMLElement]|frameBorder,longDesc,marginHeight,marginWidth,name,!noResize,scrolling,src","frameset^[HTMLElement]|cols,*beforeunload,*blur,*error,*focus,*hashchange,*languagechange,*load,*message,*offline,*online,*pagehide,*pageshow,*popstate,*rejectionhandled,*resize,*scroll,*storage,*unhandledrejection,*unload,rows","hr^[HTMLElement]|align,color,!noShade,size,width","head^[HTMLElement]|","h1,h2,h3,h4,h5,h6^[HTMLElement]|align","html^[HTMLElement]|version","iframe^[HTMLElement]|align,!allowFullscreen,frameBorder,height,longDesc,marginHeight,marginWidth,name,referrerPolicy,%sandbox,scrolling,src,srcdoc,width","img^[HTMLElement]|align,alt,border,%crossOrigin,#height,#hspace,!isMap,longDesc,lowsrc,name,referrerPolicy,sizes,src,srcset,useMap,#vspace,#width","input^[HTMLElement]|accept,align,alt,autocapitalize,autocomplete,!autofocus,!checked,!defaultChecked,defaultValue,dirName,!disabled,%files,formAction,formEnctype,formMethod,!formNoValidate,formTarget,#height,!incremental,!indeterminate,max,#maxLength,min,#minLength,!multiple,name,pattern,placeholder,!readOnly,!required,selectionDirection,#selectionEnd,#selectionStart,#size,src,step,type,useMap,value,%valueAsDate,#valueAsNumber,#width","li^[HTMLElement]|type,#value","label^[HTMLElement]|htmlFor","legend^[HTMLElement]|align","link^[HTMLElement]|as,charset,%crossOrigin,!disabled,href,hreflang,integrity,media,referrerPolicy,rel,%relList,rev,%sizes,target,type","map^[HTMLElement]|name","marquee^[HTMLElement]|behavior,bgColor,direction,height,#hspace,#loop,#scrollAmount,#scrollDelay,!trueSpeed,#vspace,width","menu^[HTMLElement]|!compact","meta^[HTMLElement]|content,httpEquiv,name,scheme","meter^[HTMLElement]|#high,#low,#max,#min,#optimum,#value","ins,del^[HTMLElement]|cite,dateTime","ol^[HTMLElement]|!compact,!reversed,#start,type","object^[HTMLElement]|align,archive,border,code,codeBase,codeType,data,!declare,height,#hspace,name,standby,type,useMap,#vspace,width","optgroup^[HTMLElement]|!disabled,label","option^[HTMLElement]|!defaultSelected,!disabled,label,!selected,text,value","output^[HTMLElement]|defaultValue,%htmlFor,name,value","p^[HTMLElement]|align","param^[HTMLElement]|name,type,value,valueType","picture^[HTMLElement]|","pre^[HTMLElement]|#width","progress^[HTMLElement]|#max,#value","q,blockquote,cite^[HTMLElement]|","script^[HTMLElement]|!async,charset,%crossOrigin,!defer,event,htmlFor,integrity,src,text,type","select^[HTMLElement]|!autofocus,!disabled,#length,!multiple,name,!required,#selectedIndex,#size,value","shadow^[HTMLElement]|","slot^[HTMLElement]|name","source^[HTMLElement]|media,sizes,src,srcset,type","span^[HTMLElement]|","style^[HTMLElement]|!disabled,media,type","caption^[HTMLElement]|align","th,td^[HTMLElement]|abbr,align,axis,bgColor,ch,chOff,#colSpan,headers,height,!noWrap,#rowSpan,scope,vAlign,width","col,colgroup^[HTMLElement]|align,ch,chOff,#span,vAlign,width","table^[HTMLElement]|align,bgColor,border,%caption,cellPadding,cellSpacing,frame,rules,summary,%tFoot,%tHead,width","tr^[HTMLElement]|align,bgColor,ch,chOff,vAlign","tfoot,thead,tbody^[HTMLElement]|align,ch,chOff,vAlign","template^[HTMLElement]|","textarea^[HTMLElement]|autocapitalize,!autofocus,#cols,defaultValue,dirName,!disabled,#maxLength,#minLength,name,placeholder,!readOnly,!required,#rows,selectionDirection,#selectionEnd,#selectionStart,value,wrap","title^[HTMLElement]|text","track^[HTMLElement]|!default,kind,label,src,srclang","ul^[HTMLElement]|!compact,type","unknown^[HTMLElement]|","video^media|#height,poster,#width",":svg:a^:svg:graphics|",":svg:animate^:svg:animation|",":svg:animateMotion^:svg:animation|",":svg:animateTransform^:svg:animation|",":svg:circle^:svg:geometry|",":svg:clipPath^:svg:graphics|",":svg:defs^:svg:graphics|",":svg:desc^:svg:|",":svg:discard^:svg:|",":svg:ellipse^:svg:geometry|",":svg:feBlend^:svg:|",":svg:feColorMatrix^:svg:|",":svg:feComponentTransfer^:svg:|",":svg:feComposite^:svg:|",":svg:feConvolveMatrix^:svg:|",":svg:feDiffuseLighting^:svg:|",":svg:feDisplacementMap^:svg:|",":svg:feDistantLight^:svg:|",":svg:feDropShadow^:svg:|",":svg:feFlood^:svg:|",":svg:feFuncA^:svg:componentTransferFunction|",":svg:feFuncB^:svg:componentTransferFunction|",":svg:feFuncG^:svg:componentTransferFunction|",":svg:feFuncR^:svg:componentTransferFunction|",":svg:feGaussianBlur^:svg:|",":svg:feImage^:svg:|",":svg:feMerge^:svg:|",":svg:feMergeNode^:svg:|",":svg:feMorphology^:svg:|",":svg:feOffset^:svg:|",":svg:fePointLight^:svg:|",":svg:feSpecularLighting^:svg:|",":svg:feSpotLight^:svg:|",":svg:feTile^:svg:|",":svg:feTurbulence^:svg:|",":svg:filter^:svg:|",":svg:foreignObject^:svg:graphics|",":svg:g^:svg:graphics|",":svg:image^:svg:graphics|",":svg:line^:svg:geometry|",":svg:linearGradient^:svg:gradient|",":svg:mpath^:svg:|",":svg:marker^:svg:|",":svg:mask^:svg:|",":svg:metadata^:svg:|",":svg:path^:svg:geometry|",":svg:pattern^:svg:|",":svg:polygon^:svg:geometry|",":svg:polyline^:svg:geometry|",":svg:radialGradient^:svg:gradient|",":svg:rect^:svg:geometry|",":svg:svg^:svg:graphics|#currentScale,#zoomAndPan",":svg:script^:svg:|type",":svg:set^:svg:animation|",":svg:stop^:svg:|",":svg:style^:svg:|!disabled,media,title,type",":svg:switch^:svg:graphics|",":svg:symbol^:svg:|",":svg:tspan^:svg:textPositioning|",":svg:text^:svg:textPositioning|",":svg:textPath^:svg:textContent|",":svg:title^:svg:|",":svg:use^:svg:graphics|",":svg:view^:svg:|#zoomAndPan","data^[HTMLElement]|value","keygen^[HTMLElement]|!autofocus,challenge,!disabled,form,keytype,name","menuitem^[HTMLElement]|type,label,icon,!disabled,!checked,radiogroup,!default","summary^[HTMLElement]|","time^[HTMLElement]|dateTime",":svg:cursor^:svg:|"],ac={class:"className",for:"htmlFor",formaction:"formAction",innerHtml:"innerHTML",readonly:"readOnly",tabindex:"tabIndex"},sc=function(e){function t(){var t=e.call(this)||this;return t._schema={},oc.forEach(function(e){var n={},r=Object(o.__read)(e.split("|"),2),i=r[0],a=r[1].split(","),s=Object(o.__read)(i.split("^"),2),u=s[0],c=s[1];u.split(",").forEach(function(e){return t._schema[e.toLowerCase()]=n});var l=c&&t._schema[c.toLowerCase()];l&&Object.keys(l).forEach(function(e){n[e]=l[e]}),a.forEach(function(e){if(e.length>0)switch(e[0]){case"*":break;case"!":n[e.substring(1)]=tc;break;case"#":n[e.substring(1)]=nc;break;case"%":n[e.substring(1)]=ic;break;default:n[e]=rc}})}),t}return Object(o.__extends)(t,e),t.prototype.hasProperty=function(e,t,n){if(n.some(function(e){return e.name===K.name}))return!0;if(e.indexOf("-")>-1){if(s(e)||u(e))return!1;if(n.some(function(e){return e.name===Y.name}))return!0}return!!(this._schema[e.toLowerCase()]||this._schema.unknown)[t]},t.prototype.hasElement=function(e,t){if(t.some(function(e){return e.name===K.name}))return!0;if(e.indexOf("-")>-1){if(s(e)||u(e))return!0;if(t.some(function(e){return e.name===Y.name}))return!0}return!!this._schema[e.toLowerCase()]},t.prototype.securityContext=function(e,t,n){n&&(t=this.getMappedPropName(t)),e=e.toLowerCase(),t=t.toLowerCase();var r=$u()[e+"|"+t];return r||((r=$u()["*|"+t])||R.NONE)},t.prototype.getMappedPropName=function(e){return ac[e]||e},t.prototype.getDefaultComponentElementName=function(){return"ng-component"},t.prototype.validateProperty=function(e){return e.toLowerCase().startsWith("on")?{error:!0,msg:"Binding to event property '"+e+"' is disallowed for security reasons, please use ("+e.slice(2)+")=...\nIf '"+e+"' is a directive input, make sure the directive is imported by the current module."}:{error:!1}},t.prototype.validateAttribute=function(e){return e.toLowerCase().startsWith("on")?{error:!0,msg:"Binding to event attribute '"+e+"' is disallowed for security reasons, please use ("+e.slice(2)+")=..."}:{error:!1}},t.prototype.allKnownElementNames=function(){return Object.keys(this._schema)},t.prototype.normalizeAnimationStyleProperty=function(e){return e.replace(Nt,function(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];return e[1].toUpperCase()})},t.prototype.normalizeAnimationStyleValue=function(e,t,n){var r="",i=n.toString().trim(),o=null;if(function(e){switch(e){case"width":case"height":case"minWidth":case"minHeight":case"maxWidth":case"maxHeight":case"left":case"top":case"bottom":case"right":case"fontSize":case"outlineWidth":case"outlineOffset":case"paddingTop":case"paddingLeft":case"paddingBottom":case"paddingRight":case"marginTop":case"marginLeft":case"marginBottom":case"marginRight":case"borderRadius":case"borderWidth":case"borderTopWidth":case"borderLeftWidth":case"borderRightWidth":case"borderBottomWidth":case"textIndent":return!0;default:return!1}}(e)&&0!==n&&"0"!==n)if("number"==typeof n)r="px";else{var a=n.match(/^[+-]?[\d\.]+([a-z]*)$/);a&&0==a[1].length&&(o="Please provide a CSS unit value for "+t+":"+n)}return{error:o,value:i+r}},t}(ec);var uc=function(){function e(e,t,n,r,i){if(this._exprParser=e,this._interpolationConfig=t,this._schemaRegistry=n,this.errors=i,this.pipesByName=null,this._usedPipes=new Map,r){var o=new Map;r.forEach(function(e){return o.set(e.name,e)}),this.pipesByName=o}}return Object.defineProperty(e.prototype,"interpolationConfig",{get:function(){return this._interpolationConfig},enumerable:!0,configurable:!0}),e.prototype.getUsedPipes=function(){return Array.from(this._usedPipes.values())},e.prototype.createBoundHostProperties=function(e,t){var n=this;if(e.hostProperties){var r=[];return Object.keys(e.hostProperties).forEach(function(i){var o=e.hostProperties[i];"string"==typeof o?n.parsePropertyBinding(i,o,!0,t,[],r):n._reportError('Value of the host property binding "'+i+'" needs to be a string representing an expression but got "'+o+'" ('+typeof o+")",t)}),r}return null},e.prototype.createDirectiveHostPropertyAsts=function(e,t,n){var r=this,i=this.createBoundHostProperties(e,n);return i&&i.map(function(e){return r.createBoundElementProperty(t,e)})},e.prototype.createDirectiveHostEventAsts=function(e,t){var n=this;if(e.hostListeners){var r=[];return Object.keys(e.hostListeners).forEach(function(i){var o=e.hostListeners[i];"string"==typeof o?n.parseEvent(i,o,t,[],r):n._reportError('Value of the host listener "'+i+'" needs to be a string representing an expression but got "'+o+'" ('+typeof o+")",t)}),r}return null},e.prototype.parseInterpolation=function(e,t){var n=t.start.toString();try{var r=this._exprParser.parseInterpolation(e,n,this._interpolationConfig);return r&&this._reportExpressionParserErrors(r.errors,t),this._checkPipes(r,t),r}catch(e){return this._reportError(""+e,t),this._exprParser.wrapLiteralPrimitive("ERROR",n)}},e.prototype.parseInlineTemplateBinding=function(e,t,n,r,i,o){for(var a=this._parseTemplateBindings(e,t,n),s=0;s<a.length;s++){var u=a[s];u.keyIsVar?o.push(new Yo(u.key,u.name,n)):u.expression?this._parsePropertyAst(u.key,u.expression,n,r,i):(r.push([u.key,""]),this.parseLiteralAttr(u.key,null,n,r,i))}},e.prototype._parseTemplateBindings=function(e,t,n){var r=this,i=n.start.toString();try{var o=this._exprParser.parseTemplateBindings(e,t,i);return this._reportExpressionParserErrors(o.errors,n),o.templateBindings.forEach(function(e){e.expression&&r._checkPipes(e.expression,n)}),o.warnings.forEach(function(e){r._reportError(e,n,es.WARNING)}),o.templateBindings}catch(e){return this._reportError(""+e,n),[]}},e.prototype.parseLiteralAttr=function(e,t,n,r,i){lc(e)?(e=e.substring(1),t&&this._reportError('Assigning animation triggers via @prop="exp" attributes with an expression is invalid. Use property bindings (e.g. [@prop]="exp") or use an attribute without a value (e.g. @prop) instead.',n,es.ERROR),this._parseAnimation(e,t,n,r,i)):i.push(new Go(e,this._exprParser.wrapLiteralPrimitive(t,""),Ho.LITERAL_ATTR,n))},e.prototype.parsePropertyBinding=function(e,t,n,r,i,o){var a=!1;e.startsWith("animate-")?(a=!0,e=e.substring("animate-".length)):lc(e)&&(a=!0,e=e.substring(1)),a?this._parseAnimation(e,t,r,i,o):this._parsePropertyAst(e,this._parseBinding(t,n,r),r,i,o)},e.prototype.parsePropertyInterpolation=function(e,t,n,r,i){var o=this.parseInterpolation(t,n);return!!o&&(this._parsePropertyAst(e,o,n,r,i),!0)},e.prototype._parsePropertyAst=function(e,t,n,r,i){r.push([e,t.source]),i.push(new Go(e,t,Ho.DEFAULT,n))},e.prototype._parseAnimation=function(e,t,n,r,i){var o=this._parseBinding(t||"undefined",!1,n);r.push([e,o.source]),i.push(new Go(e,o,Ho.ANIMATION,n))},e.prototype._parseBinding=function(e,t,n){var r=(n&&n.start||"(unknown)").toString();try{var i=t?this._exprParser.parseSimpleBinding(e,r,this._interpolationConfig):this._exprParser.parseBinding(e,r,this._interpolationConfig);return i&&this._reportExpressionParserErrors(i.errors,n),this._checkPipes(i,n),i}catch(e){return this._reportError(""+e,n),this._exprParser.wrapLiteralPrimitive("ERROR",r)}},e.prototype.createBoundElementProperty=function(e,t){if(t.isAnimation)return new Ko(t.name,4,R.NONE,t.expression,null,t.sourceSpan);var n=null,r=void 0,i=null,o=t.name.split("."),a=void 0;if(o.length>1)if("attr"==o[0]){i=o[1],this._validatePropertyOrAttributeName(i,t.sourceSpan,!0),a=dc(this._schemaRegistry,e,i,!0);var s=i.indexOf(":");if(s>-1)i=d(i.substring(0,s),i.substring(s+1));r=1}else"class"==o[0]?(i=o[1],r=2,a=[R.NONE]):"style"==o[0]&&(n=o.length>2?o[2]:null,i=o[1],r=3,a=[R.STYLE]);return null===i&&(i=this._schemaRegistry.getMappedPropName(t.name),a=dc(this._schemaRegistry,e,i,!1),r=0,this._validatePropertyOrAttributeName(i,t.sourceSpan,!1)),new Ko(i,r,a[0],t.expression,n,t.sourceSpan)},e.prototype.parseEvent=function(e,t,n,r,i){lc(e)?(e=e.substr(1),this._parseAnimationEvent(e,t,n,i)):this._parseRegularEvent(e,t,n,r,i)},e.prototype._parseAnimationEvent=function(e,t,n,r){var i=Lt(e,".",[e,""]),o=i[0],a=i[1].toLowerCase();if(a)switch(a){case"start":case"done":var s=this._parseAction(t,n);r.push(new Wo(o,a,1,s,n));break;default:this._reportError('The provided animation output phase value "'+a+'" for "@'+o+'" is not supported (use start or done)',n)}else this._reportError("The animation trigger output event (@"+o+") is missing its phase value name (start or done are currently supported)",n)},e.prototype._parseRegularEvent=function(e,t,n,r,i){var a=Object(o.__read)(Ft(e,[null,e]),2),s=a[0],u=a[1],c=this._parseAction(t,n);r.push([e,c.source]),i.push(new Wo(u,s,0,c,n))},e.prototype._parseAction=function(e,t){var n=(t&&t.start||"(unknown").toString();try{var r=this._exprParser.parseAction(e,n,this._interpolationConfig);return r&&this._reportExpressionParserErrors(r.errors,t),!r||r.ast instanceof go?(this._reportError("Empty expressions are not allowed",t),this._exprParser.wrapLiteralPrimitive("ERROR",n)):(this._checkPipes(r,t),r)}catch(e){return this._reportError(""+e,t),this._exprParser.wrapLiteralPrimitive("ERROR",n)}},e.prototype._reportError=function(e,t,n){void 0===n&&(n=es.ERROR),this.errors.push(new os(t,e,n))},e.prototype._reportExpressionParserErrors=function(e,t){var n,r;try{for(var i=Object(o.__values)(e),a=i.next();!a.done;a=i.next()){var s=a.value;this._reportError(s.message,t)}}catch(e){n={error:e}}finally{try{a&&!a.done&&(r=i.return)&&r.call(i)}finally{if(n)throw n.error}}},e.prototype._checkPipes=function(e,t){var n=this;if(e&&this.pipesByName){var r=new cc;e.visit(r),r.pipes.forEach(function(e,r){var i=n.pipesByName.get(r);i?n._usedPipes.set(r,i):n._reportError("The pipe '"+r+"' could not be found",new is(t.start.moveBy(e.span.start),t.start.moveBy(e.span.end)))})}},e.prototype._validatePropertyOrAttributeName=function(e,t,n){var r=n?this._schemaRegistry.validateAttribute(e):this._schemaRegistry.validateProperty(e);r.error&&this._reportError(r.msg,t,es.ERROR)},e}(),cc=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.pipes=new Map,t}return Object(o.__extends)(t,e),t.prototype.visitPipe=function(e,t){return this.pipes.set(e.name,e),e.exp.visit(this),this.visitAll(e.args,t),null},t}(Vo);function lc(e){return"@"==e[0]}function dc(e,t,n,r){var i=[];return O.parse(t).forEach(function(t){var a=t.element?[t.element]:e.allKnownElementNames(),s=new Set(t.notSelectors.filter(function(e){return e.isElementSelector()}).map(function(e){return e.element})),u=a.filter(function(e){return!s.has(e)});i.push.apply(i,Object(o.__spread)(u.map(function(t){return e.securityContext(t,n,r)})))}),0===i.length?[R.NONE]:Array.from(new Set(i)).sort()}var pc=function(){function e(e,t){this.value=e,this.sourceSpan=t}return e.prototype.visit=function(e){return e.visitText(this)},e}(),fc=function(){function e(e,t,n){this.value=e,this.sourceSpan=t,this.i18n=n}return e.prototype.visit=function(e){return e.visitBoundText(this)},e}(),hc=function(){function e(e,t,n,r,i){this.name=e,this.value=t,this.sourceSpan=n,this.valueSpan=r,this.i18n=i}return e.prototype.visit=function(e){return e.visitTextAttribute(this)},e}(),mc=function(){function e(e,t,n,r,i,o,a){this.name=e,this.type=t,this.securityContext=n,this.value=r,this.unit=i,this.sourceSpan=o,this.i18n=a}return e.fromBoundElementProperty=function(t,n){return new e(t.name,t.type,t.securityContext,t.value,t.unit,t.sourceSpan,n)},e.prototype.visit=function(e){return e.visitBoundAttribute(this)},e}(),gc=function(){function e(e,t,n,r,i,o){this.name=e,this.type=t,this.handler=n,this.target=r,this.phase=i,this.sourceSpan=o}return e.fromParsedEvent=function(t){var n=0===t.type?t.targetOrPhase:null,r=1===t.type?t.targetOrPhase:null;return new e(t.name,t.type,t.handler,n,r,t.sourceSpan)},e.prototype.visit=function(e){return e.visitBoundEvent(this)},e}(),vc=function(){function e(e,t,n,r,i,o,a,s,u,c){this.name=e,this.attributes=t,this.inputs=n,this.outputs=r,this.children=i,this.references=o,this.sourceSpan=a,this.startSourceSpan=s,this.endSourceSpan=u,this.i18n=c}return e.prototype.visit=function(e){return e.visitElement(this)},e}(),_c=function(){function e(e,t,n,r,i,o,a,s,u,c,l){this.tagName=e,this.attributes=t,this.inputs=n,this.outputs=r,this.children=i,this.references=o,this.variables=a,this.sourceSpan=s,this.startSourceSpan=u,this.endSourceSpan=c,this.i18n=l}return e.prototype.visit=function(e){return e.visitTemplate(this)},e}(),yc=function(){function e(e,t,n,r){this.selector=e,this.attributes=t,this.sourceSpan=n,this.i18n=r}return e.prototype.visit=function(e){return e.visitContent(this)},e}(),bc=function(){function e(e,t,n){this.name=e,this.value=t,this.sourceSpan=n}return e.prototype.visit=function(e){return e.visitVariable(this)},e}(),jc=function(){function e(e,t,n){this.name=e,this.value=t,this.sourceSpan=n}return e.prototype.visit=function(e){return e.visitReference(this)},e}(),xc=function(){function e(e,t,n,r){this.vars=e,this.placeholders=t,this.sourceSpan=n,this.i18n=r}return e.prototype.visit=function(e){return e.visitIcu(this)},e}();function Oc(e,t){var n,r,i,a,s=[];if(e.visit)try{for(var u=Object(o.__values)(t),c=u.next();!c.done;c=u.next())var l=c.value,d=e.visit(l)||l.visit(e)}catch(e){n={error:e}}finally{try{c&&!c.done&&(r=u.return)&&r.call(u)}finally{if(n)throw n.error}}else try{for(var p=Object(o.__values)(t),f=p.next();!f.done;f=p.next()){(d=(l=f.value).visit(e))&&s.push(d)}}catch(e){i={error:e}}finally{try{f&&!f.done&&(a=p.return)&&a.call(p)}finally{if(i)throw i.error}}return s}var wc=function(){return function(e,t){this.style=e,this.styleUrls=t}}();function Cc(e){if(null==e||0===e.length||"/"==e[0])return!1;var t=e.match(Ac);return null===t||"package"==t[1]||"asset"==t[1]}var Ec,Sc,kc=/@import\s+(?:url\()?\s*(?:(?:['"]([^'"]*))|([^;\)\s]*))[^;]*;?/g,Mc=/\/\*(?!#\s*(?:sourceURL|sourceMappingURL)=)[\s\S]+?\*\//g,Ac=/^([^:/?#]+):/,Tc="select",Dc="link",Ic="rel",Pc="href",Rc="stylesheet",Nc="style",Fc="script",Lc="ngNonBindable",Vc="ngProjectAs";function Bc(e){var t=null,n=null,r=null,i=!1,o="";e.attrs.forEach(function(e){var a=e.name.toLowerCase();a==Tc?t=e.value:a==Pc?n=e.value:a==Ic?r=e.value:e.name==Lc?i=!0:e.name==Vc&&e.value.length>0&&(o=e.value)}),t=function(e){if(null===e||0===e.length)return"*";return e}(t);var a=e.name.toLowerCase(),s=Ec.OTHER;return u(a)?s=Ec.NG_CONTENT:a==Nc?s=Ec.STYLE:a==Fc?s=Ec.SCRIPT:a==Dc&&r==Rc&&(s=Ec.STYLESHEET),new zc(s,t,n,i,o)}(Sc=Ec||(Ec={}))[Sc.NG_CONTENT=0]="NG_CONTENT",Sc[Sc.STYLE=1]="STYLE",Sc[Sc.STYLESHEET=2]="STYLESHEET",Sc[Sc.SCRIPT=3]="SCRIPT",Sc[Sc.OTHER=4]="OTHER";var zc=function(){return function(e,t,n,r,i){this.type=e,this.selectAttr=t,this.hrefAttr=n,this.nonBindable=r,this.projectAs=i}}();var Uc=/^(?:(?:(?:(bind-)|(let-)|(ref-|#)|(on-)|(bindon-)|(@))(.+))|\[\(([^\)]+)\)\]|\[([^\]]+)\]|\(([^\)]+)\))$/;function Hc(e,t){var n=new Wc(t),r=_u(n,e),i=t.errors.concat(n.errors),o=i.filter(function(e){return e.level===es.ERROR});if(o.length>0)throw Yt("Template parse errors:\n"+o.join("\n"),o);return{nodes:r,errors:i}}var qc,Gc,Wc=function(){function e(e){this.bindingParser=e,this.errors=[]}return e.prototype.visitElement=function(e){var t,n,r=this,i=Bc(e);if(i.type===Ec.SCRIPT||i.type===Ec.STYLE)return null;if(i.type===Ec.STYLESHEET&&Cc(i.hrefAttr))return null;var a=c(e.name),s=[],u=[],l=[],d=[],p=[],f={},h=[],m=[],g=!1;try{for(var v=Object(o.__values)(e.attrs),_=v.next();!_.done;_=v.next()){var y=_.value,b=!1,j=Kc(y.name),x=!1;if(y.i18n&&(f[y.name]=y.i18n),j.startsWith("*")){g&&this.reportError("Can't have multiple template bindings on one element. Use only one attribute prefixed with *",y.sourceSpan),x=!0,g=!0;var O=y.value,w=j.substring("*".length),C=[];this.bindingParser.parseInlineTemplateBinding(w,O,y.sourceSpan,[],h,C),m.push.apply(m,Object(o.__spread)(C.map(function(e){return new bc(e.name,e.value,e.sourceSpan)})))}else b=this.parseAttribute(a,y,[],s,u,l,d);b||x||p.push(this.visitAttribute(y))}}catch(e){t={error:e}}finally{try{_&&!_.done&&(n=v.return)&&n.call(v)}finally{if(t)throw t.error}}var E,S=_u(i.nonBindable?Yc:this,e.children);if(i.type===Ec.NG_CONTENT){e.children&&!e.children.every(Xc)&&this.reportError("<ng-content> element cannot have content.",e.sourceSpan);var k=i.selectAttr,M=e.attrs.map(function(e){return r.visitAttribute(e)});E=new yc(k,M,e.sourceSpan,e.i18n)}else if(a){M=this.extractAttributes(e.name,s,f);E=new _c(e.name,p,M.bound,u,S,d,l,e.sourceSpan,e.startSourceSpan,e.endSourceSpan,e.i18n)}else{M=this.extractAttributes(e.name,s,f);E=new vc(e.name,p,M.bound,u,S,d,e.sourceSpan,e.startSourceSpan,e.endSourceSpan,e.i18n)}if(g){M=this.extractAttributes("ng-template",h,f);E=new _c(E.name,M.literal,M.bound,[],[E],[],m,e.sourceSpan,e.startSourceSpan,e.endSourceSpan,e.i18n)}return E},e.prototype.visitAttribute=function(e){return new hc(e.name,e.value,e.sourceSpan,e.valueSpan,e.i18n)},e.prototype.visitText=function(e){return this._visitTextWithInterpolation(e.value,e.sourceSpan,e.i18n)},e.prototype.visitExpansion=function(e){var t=this,n=e.i18n;if(!n)return null;var r={},i={};return Object.keys(n.placeholders).forEach(function(o){var a=n.placeholders[o];if(o.startsWith("VAR_")){var s=t.bindingParser.interpolationConfig,u=""+s.start+a+s.end;r[o]=t._visitTextWithInterpolation(u,e.sourceSpan)}else i[o]=t._visitTextWithInterpolation(a,e.sourceSpan)}),new xc(r,i,e.sourceSpan,n)},e.prototype.visitExpansionCase=function(e){return null},e.prototype.visitComment=function(e){return null},e.prototype.extractAttributes=function(e,t,n){var r=this,i=[],o=[];return t.forEach(function(t){var a=n[t.name];if(t.isLiteral)o.push(new hc(t.name,t.expression.source||"",t.sourceSpan,void 0,a));else{var s=r.bindingParser.createBoundElementProperty(e,t);i.push(mc.fromBoundElementProperty(s,a))}}),{bound:i,literal:o}},e.prototype.parseAttribute=function(e,t,n,r,i,o,a){var s=Kc(t.name),u=t.value,c=t.sourceSpan,l=s.match(Uc),d=!1;if(l){if(d=!0,null!=l[1])this.bindingParser.parsePropertyBinding(l[7],u,!1,c,n,r);else if(l[2])if(e){var p=l[7];this.parseVariable(p,u,c,o)}else this.reportError('"let-" is only supported on ng-template elements.',c);else if(l[3]){p=l[7];this.parseReference(p,u,c,a)}else if(l[4]){var f=[];this.bindingParser.parseEvent(l[7],u,c,n,f),Qc(f,i)}else if(l[5])this.bindingParser.parsePropertyBinding(l[7],u,!1,c,n,r),this.parseAssignmentEvent(l[7],u,c,n,i);else if(l[6])this.bindingParser.parseLiteralAttr(s,u,c,n,r);else if(l[8])this.bindingParser.parsePropertyBinding(l[8],u,!1,c,n,r),this.parseAssignmentEvent(l[8],u,c,n,i);else if(l[9])this.bindingParser.parsePropertyBinding(l[9],u,!1,c,n,r);else if(l[10]){f=[];this.bindingParser.parseEvent(l[10],u,c,n,f),Qc(f,i)}}else d=this.bindingParser.parsePropertyInterpolation(s,u,c,n,r);return d},e.prototype._visitTextWithInterpolation=function(e,t,n){var r=Ku(e),i=this.bindingParser.parseInterpolation(r,t);return i?new fc(i,t,n):new pc(r,t)},e.prototype.parseVariable=function(e,t,n,r){e.indexOf("-")>-1&&this.reportError('"-" is not allowed in variable names',n),r.push(new bc(e,t,n))},e.prototype.parseReference=function(e,t,n,r){e.indexOf("-")>-1&&this.reportError('"-" is not allowed in reference names',n),r.push(new jc(e,t,n))},e.prototype.parseAssignmentEvent=function(e,t,n,r,i){var o=[];this.bindingParser.parseEvent(e+"Change",t+"=$event",n,r,o),Qc(o,i)},e.prototype.reportError=function(e,t,n){void 0===n&&(n=es.ERROR),this.errors.push(new os(t,e,n))},e}(),Yc=new(function(){function e(){}return e.prototype.visitElement=function(e){var t=Bc(e);if(t.type===Ec.SCRIPT||t.type===Ec.STYLE||t.type===Ec.STYLESHEET)return null;var n=_u(this,e.children,null);return new vc(e.name,_u(this,e.attrs),[],[],n,[],e.sourceSpan,e.startSourceSpan,e.endSourceSpan)},e.prototype.visitComment=function(e){return null},e.prototype.visitAttribute=function(e){return new hc(e.name,e.value,e.sourceSpan,void 0,e.i18n)},e.prototype.visitText=function(e){return new pc(e.value,e.sourceSpan)},e.prototype.visitExpansion=function(e){return null},e.prototype.visitExpansionCase=function(e){return null},e}());function Kc(e){return/^data-/i.test(e)?e.substring(5):e}function Qc(e,t){t.push.apply(t,Object(o.__spread)(e.map(function(e){return gc.fromParsedEvent(e)})))}function Xc(e){return e instanceof pu&&0==e.value.trim().length}(Gc=qc||(qc={}))[Gc.ELEMENT=0]="ELEMENT",Gc[Gc.TEMPLATE=1]="TEMPLATE";var Zc=function(){function e(e,t,n,r,i,o){void 0===n&&(n=0),void 0===r&&(r=null),this.index=e,this.ref=t,this.level=n,this.templateIndex=r,this.meta=i,this.registry=o,this.bindings=new Set,this.placeholders=new Map,this._unresolvedCtxCount=0,this._registry=o||{getUniqueId:function(e){void 0===e&&(e=0);var t=e;return function(){return t++}}(),icus:new Map},this.id=this._registry.getUniqueId()}return e.prototype.appendTag=function(e,t,n,r){if(!t.isVoid||!r){var i=t.isVoid||!r?t.startName:t.closeName,o={type:e,index:n,ctx:this.id,isVoid:t.isVoid,closed:r};ai(this.placeholders,i,o)}},Object.defineProperty(e.prototype,"icus",{get:function(){return this._registry.icus},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"isRoot",{get:function(){return 0===this.level},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"isResolved",{get:function(){return 0===this._unresolvedCtxCount},enumerable:!0,configurable:!0}),e.prototype.getSerializedPlaceholders=function(){var e=new Map;return this.placeholders.forEach(function(t,n){return e.set(n,t.map(tl))}),e},e.prototype.appendBinding=function(e){this.bindings.add(e)},e.prototype.appendIcu=function(e,t){ai(this._registry.icus,e,t)},e.prototype.appendBoundText=function(e){var t=this;si(e,this.bindings.size,this.id).forEach(function(e,n){return ai.apply(void 0,Object(o.__spread)([t.placeholders,n],e))})},e.prototype.appendTemplate=function(e,t){this.appendTag(qc.TEMPLATE,e,t,!1),this.appendTag(qc.TEMPLATE,e,t,!0),this._unresolvedCtxCount++},e.prototype.appendElement=function(e,t,n){this.appendTag(qc.ELEMENT,e,t,n)},e.prototype.forkChildContext=function(t,n,r){return new e(t,this.ref,this.level+1,n,r,this._registry)},e.prototype.reconcileChildContext=function(e){var t=this;["start","close"].forEach(function(n){var r=e.meta[n+"Name"],i=(t.placeholders.get(r)||[]).find(el(t.id,e.templateIndex));i&&(i.ctx=e.id)}),e.placeholders.forEach(function(n,r){var i=t.placeholders.get(r);if(i){var a=function(e,t){for(var n=0;n<e.length;n++)if(t(e[n]))return n;return-1}(i,el(e.id,e.templateIndex));if(a>=0){var s=r.startsWith("CLOSE");if(r.endsWith("NG-TEMPLATE"))i.splice.apply(i,Object(o.__spread)([a+(s?0:1),0],n));else n[s?n.length-1:0].tmpl=i[a],i.splice.apply(i,Object(o.__spread)([a,1],n))}else i.push.apply(i,Object(o.__spread)(n));t.placeholders.set(r,i)}else t.placeholders.set(r,n)}),this._unresolvedCtxCount--},e}();function $c(e,t,n,r){return ii(""+(r?"/":"")+e+t,n)}function Jc(e,t,n){var r=t.index,i=t.ctx;return t.isVoid?$c(e,r,i)+$c(e,r,i,!0):$c(e,r,i,n)}function el(e,t){return function(n){return"object"==typeof n&&n.type===qc.TEMPLATE&&n.index===t&&n.ctx===e}}function tl(e){var t=function(e,t){return Jc("#",e,t)},n=function(e,t){return Jc("*",e,t)};switch(e.type){case qc.ELEMENT:return e.closed?t(e,!0)+(e.tmpl?n(e.tmpl,!0):""):e.tmpl?n(e.tmpl)+t(e)+(e.isVoid?n(e.tmpl,!0):""):t(e);case qc.TEMPLATE:return n(e,e.closed);default:return e}}var nl={A:"LINK",B:"BOLD_TEXT",BR:"LINE_BREAK",EM:"EMPHASISED_TEXT",H1:"HEADING_LEVEL1",H2:"HEADING_LEVEL2",H3:"HEADING_LEVEL3",H4:"HEADING_LEVEL4",H5:"HEADING_LEVEL5",H6:"HEADING_LEVEL6",HR:"HORIZONTAL_RULE",I:"ITALIC_TEXT",LI:"LIST_ITEM",LINK:"MEDIA_LINK",OL:"ORDERED_LIST",P:"PARAGRAPH",Q:"QUOTATION",S:"STRIKETHROUGH_TEXT",SMALL:"SMALL_TEXT",SUB:"SUBSTRIPT",SUP:"SUPERSCRIPT",TBODY:"TABLE_BODY",TD:"TABLE_CELL",TFOOT:"TABLE_FOOTER",TH:"TABLE_HEADER_CELL",THEAD:"TABLE_HEADER",TR:"TABLE_ROW",TT:"MONOSPACED_TEXT",U:"UNDERLINED_TEXT",UL:"UNORDERED_LIST"},rl=function(){function e(){this._placeHolderNameCounts={},this._signatureToName={}}return e.prototype.getStartTagPlaceholderName=function(e,t,n){var r=this._hashTag(e,t,n);if(this._signatureToName[r])return this._signatureToName[r];var i=e.toUpperCase(),o=nl[i]||"TAG_"+i,a=this._generateUniqueName(n?o:"START_"+o);return this._signatureToName[r]=a,a},e.prototype.getCloseTagPlaceholderName=function(e){var t=this._hashClosingTag(e);if(this._signatureToName[t])return this._signatureToName[t];var n=e.toUpperCase(),r=nl[n]||"TAG_"+n,i=this._generateUniqueName("CLOSE_"+r);return this._signatureToName[t]=i,i},e.prototype.getPlaceholderName=function(e,t){var n=e.toUpperCase(),r="PH: "+n+"="+t;if(this._signatureToName[r])return this._signatureToName[r];var i=this._generateUniqueName(n);return this._signatureToName[r]=i,i},e.prototype.getUniquePlaceholder=function(e){return this._generateUniqueName(e.toUpperCase())},e.prototype._hashTag=function(e,t,n){return"<"+e+Object.keys(t).sort().map(function(e){return" "+e+"="+t[e]}).join("")+(n?"/>":"></"+e+">")},e.prototype._hashClosingTag=function(e){return this._hashTag("/"+e,{},!1)},e.prototype._generateUniqueName=function(e){if(!this._placeHolderNameCounts.hasOwnProperty(e))return this._placeHolderNameCounts[e]=1,e;var t=this._placeHolderNameCounts[e];return this._placeHolderNameCounts[e]=t+1,e+"_"+t},e}(),il=new uu(new Ks);function ol(e){var t=new al(il,e);return function(e,n,r,i,o){return t.toI18nMessage(e,n,r,i,o)}}var al=function(){function e(e,t){this._expressionParser=e,this._interpolationConfig=t}return e.prototype.toI18nMessage=function(e,t,n,r,i){this._isIcu=1==e.length&&e[0]instanceof fu,this._icuDepth=0,this._placeholderRegistry=new rl,this._placeholderToContent={},this._placeholderToMessage={},this._visitNodeFn=i;var o=_u(this,e,{});return new Zn(o,this._placeholderToContent,this._placeholderToMessage,t,n,r)},e.prototype._visitNode=function(e,t){return this._visitNodeFn&&this._visitNodeFn(e,t),t},e.prototype.visitElement=function(e,t){var n=_u(this,e.children),r={};e.attrs.forEach(function(e){r[e.name]=e.value});var i=v(e.name).isVoid,o=this._placeholderRegistry.getStartTagPlaceholderName(e.name,r,i);this._placeholderToContent[o]=e.sourceSpan.toString();var a="";i||(a=this._placeholderRegistry.getCloseTagPlaceholderName(e.name),this._placeholderToContent[a]="</"+e.name+">");var s=new tr(e.name,r,o,a,n,i,e.sourceSpan);return this._visitNode(e,s)},e.prototype.visitAttribute=function(e,t){var n=this._visitTextWithInterpolation(e.value,e.sourceSpan);return this._visitNode(e,n)},e.prototype.visitText=function(e,t){var n=this._visitTextWithInterpolation(e.value,e.sourceSpan);return this._visitNode(e,n)},e.prototype.visitComment=function(e,t){return null},e.prototype.visitExpansion=function(t,n){var r=this;this._icuDepth++;var i={},o=new er(t.switchValue,t.type,i,t.sourceSpan);if(t.cases.forEach(function(e){i[e.value]=new Jn(e.expression.map(function(e){return e.visit(r,{})}),e.expSourceSpan)}),this._icuDepth--,this._isIcu||this._icuDepth>0){var a=this._placeholderRegistry.getUniquePlaceholder("VAR_"+t.type);return o.expressionPlaceholder=a,this._placeholderToContent[a]=t.switchValue,this._visitNode(t,o)}var s=this._placeholderRegistry.getPlaceholderName("ICU",t.sourceSpan.toString()),u=new e(this._expressionParser,this._interpolationConfig);this._placeholderToMessage[s]=u.toI18nMessage([t],"","","");var c=new rr(o,s,t.sourceSpan);return this._visitNode(t,c)},e.prototype.visitExpansionCase=function(e,t){throw new Error("Unreachable code")},e.prototype._visitTextWithInterpolation=function(e,t){var n=this._expressionParser.splitInterpolation(e,t.start.toString(),this._interpolationConfig);if(!n)return new $n(e,t);for(var r=[],i=new Jn(r,t),o=this._interpolationConfig,a=o.start,s=o.end,u=0;u<n.strings.length-1;u++){var c=n.expressions[u],l=c.split(sl)[2]||"INTERPOLATION",d=this._placeholderRegistry.getPlaceholderName(l,c);n.strings[u].length&&r.push(new $n(n.strings[u],t)),r.push(new nr(c,d,t)),this._placeholderToContent[d]=a+c+s}var p=n.strings.length-1;return n.strings[p].length&&r.push(new $n(n.strings[p],t)),i},e}(),sl=/\/\/[\s\S]*i18n[\s\S]*\([\s\S]*ph[\s\S]*=[\s\S]*("|')([\s\S]*?)\1[\s\S]*\)/g;function ul(e,t){e.i18n=t}var cl=function(){function e(e,t){void 0===e&&(e=Ui),void 0===t&&(t=!1),this.interpolationConfig=e,this.keepI18nAttrs=t,this._createI18nMessage=ol(e)}return e.prototype._generateI18nMessage=function(e,t,n){void 0===t&&(t="");var r="string"==typeof t?function(e){var t,n,r,i,a;if(e){var s=e.indexOf(Kr),u=e.indexOf(Yr),c=void 0;c=(t=Object(o.__read)(s>-1?[e.slice(0,s),e.slice(s+2)]:[e,""],2))[0],r=t[1],i=(n=Object(o.__read)(u>-1?[c.slice(0,u),c.slice(u+1)]:["",c],2))[0],a=n[1]}return{id:r,meaning:i,description:a}}(t):ni(t),i=this._createI18nMessage(e,r.meaning||"",r.description||"",r.id||"",n);return i.id||(i.id="string"!=typeof t&&t.id||sr(i)),i},e.prototype.visitElement=function(e,t){var n,r,i,a;if(e.attrs.some(function(e){return Jr(e.name)})){var s=[],u={};try{for(var c=Object(o.__values)(e.attrs),l=c.next();!l.done;l=c.next()){if((m=l.value).name===Qr){var d=e.i18n||m.value,p=this._generateI18nMessage(e.children,d,ul);p.nodes.length&&(e.i18n=p)}else if(m.name.startsWith(Xr)){u[m.name.slice(Xr.length)]=m.value}else s.push(m)}}catch(e){n={error:e}}finally{try{l&&!l.done&&(r=c.return)&&r.call(c)}finally{if(n)throw n.error}}if(Object.keys(u).length)try{for(var f=Object(o.__values)(s),h=f.next();!h.done;h=f.next()){var m,g=u[(m=h.value).name];void 0!==g&&m.value&&(m.i18n=this._generateI18nMessage([m],m.i18n||g))}}catch(e){i={error:e}}finally{try{h&&!h.done&&(a=f.return)&&a.call(f)}finally{if(i)throw i.error}}this.keepI18nAttrs||(e.attrs=s)}return _u(this,e.children),e},e.prototype.visitExpansion=function(e,t){var n,r=e.i18n;if(r instanceof rr){var i=r.name;ri(n=this._generateI18nMessage([e],r)).name=i}else n=this._generateI18nMessage([e],r);return e.i18n=n,e},e.prototype.visitText=function(e,t){return e},e.prototype.visitAttribute=function(e,t){return e},e.prototype.visitComment=function(e,t){return e},e.prototype.visitExpansionCase=function(e,t){return e},e}(),ll=function(e){return"{$"+ui(e)+"}"},dl=new(function(){function e(){}return e.prototype.visitText=function(e,t){return e.value},e.prototype.visitContainer=function(e,t){var n=this;return e.children.map(function(e){return e.visit(n)}).join("")},e.prototype.visitIcu=function(e,t){var n=this,r=Object.keys(e.cases).map(function(t){return t+" {"+e.cases[t].visit(n)+"}"});return"{"+e.expressionPlaceholder+", "+e.type+", "+r.join(" ")+"}"},e.prototype.visitTagPlaceholder=function(e,t){var n=this;return e.isVoid?ll(e.startName):""+ll(e.startName)+e.children.map(function(e){return e.visit(n)}).join("")+ll(e.closeName)},e.prototype.visitPlaceholder=function(e,t){return ll(e.name)},e.prototype.visitIcuPlaceholder=function(e,t){return ll(e.name)},e}());function pl(e,t){return It(Et(fi).bitwiseAnd(Pt(e),null,!1),t)}var fl=function(){function e(e,t,n,r,i,o,a,s,u,c,l,d,p,f,h){void 0===n&&(n=0);var m=this;this.constantPool=e,this.level=n,this.contextName=r,this.i18nContext=i,this.templateIndex=o,this.templateName=a,this.viewQueries=s,this.directiveMatcher=u,this.directives=c,this.pipeTypeByName=l,this.pipes=d,this._namespace=p,this.relativeContextFilePath=f,this.i18nUseExternalIds=h,this._dataIndex=0,this._bindingContext=0,this._prefixCode=[],this._creationCodeFns=[],this._updateCodeFns=[],this._tempVariables=[],this._nestedTemplateFns=[],this._unsupported=mi,this.i18n=null,this._pureFunctionSlots=0,this._bindingSlots=0,this._hasNgContent=!1,this._ngContentSelectors=[],this._ngContentSelectorsOffset=0,this.visitReference=gi,this.visitVariable=gi,this.visitTextAttribute=gi,this.visitBoundAttribute=gi,this.visitBoundEvent=gi,this._dataIndex=s.length,this._bindingScope=t.nestedScope(n),this.fileBasedI18nSuffix=f.replace(/[^A-Za-z0-9]/g,"_")+"_",this._valueConverter=new hl(e,function(){return m.allocateDataSlot()},function(e){return m.allocatePureFunctionSlots(e)},function(e,t,n,r){var i=l.get(e);i&&m.pipes.add(i),m._bindingScope.set(m.level,t,r),m.creationInstruction(null,Xn.pipe,[Pt(n),Pt(e)])})}return e.prototype.registerContextVariables=function(e){var t=this._bindingScope.freshReferenceName(),n=this.level,r=Et(e.name+t);this._bindingScope.set(n,e.name,r,1,function(t,i){var o;if(t.bindingLevel===n)o=Et(pi);else{var a=t.getSharedContextName(n);o=a||_l(i)}return[r.set(o.prop(e.value||"$implicit")).toConstDecl()]})},e.prototype.buildTemplateFunction=function(e,t,n,r){var i=this;void 0===n&&(n=0),this._ngContentSelectorsOffset=n,this._namespace!==Xn.namespaceHTML&&this.creationInstruction(null,this._namespace),t.forEach(function(e){return i.registerContextVariables(e)});var a,s=this.i18nContext||ei(r)&&!ti(r)&&!(1===(a=e).length&&a[0]instanceof vc&&e[0].i18n===r),u=El(e);if(s&&this.i18nStart(null,r,u),Oc(this,e),this._pureFunctionSlots+=this._bindingSlots,this._valueConverter.updatePipeSlotOffsets(this._bindingSlots),this._nestedTemplateFns.forEach(function(e){return e()}),0===this.level&&this._hasNgContent){var c=[];if(this._ngContentSelectors.length){var l=this._ngContentSelectors.map(function(e){return ne(e)}),d=this.constantPool.getConstLiteral(vi(l),!0),p=this.constantPool.getConstLiteral(vi(this._ngContentSelectors),!0);c.push(d,p)}this.creationInstruction(null,Xn.projectionDef,c,!0)}s&&this.i18nEnd(null,u);var f=this._creationCodeFns.map(function(e){return e()}),h=this._updateCodeFns.map(function(e){return e()}),m=this._bindingScope.viewSnapshotStatements(),g=this._bindingScope.variableDeclarations().concat(this._tempVariables),v=f.length>0?[pl(1,m.concat(f))]:[],_=h.length>0?[pl(2,g.concat(h))]:[];return Dt([new Ge(fi,_e),new Ge(pi,null)],Object(o.__spread)(this._prefixCode,v,_),ge,null,this.templateName)},e.prototype.getLocal=function(e){return this._bindingScope.get(e)},e.prototype.i18nTranslate=function(e,t,n,r){var i;void 0===t&&(t={});var a=n||this.i18nAllocateRef(e.id),s={};t&&Object.keys(t).length&&Object.keys(t).forEach(function(e){return s[ui(e)]=t[e]});var u=ni(e),c=li(a,e.nodes.map(function(e){return e.visit(dl,null)}).join(""),u,s,r);return(i=this.constantPool.statements).push.apply(i,Object(o.__spread)(c)),a},e.prototype.i18nAppendBindings=function(e){var t=this;if(this.i18n&&e.length){var n=Et(pi);e.forEach(function(e){var r=t.convertExpressionBinding(n,e);t.i18n.appendBinding(r)})}},e.prototype.i18nBindProps=function(e){var t=this,n={};return Object.keys(e).forEach(function(r){var i=e[r];if(i instanceof pc)n[r]=Pt(i.value);else{var o=i.value.visit(t._valueConverter);if(t.allocateBindingSlots(o),o instanceof Mo){var a=o.strings,s=o.expressions,u=t.i18n,c=u.id,l=function(e,t,n){if(void 0===t&&(t=0),void 0===n&&(n=0),!e.length)return"";for(var r="",i=e.length-1,o=0;o<i;o++)r+=""+e[o]+ii(t+o,n);return r+=e[i]}(a,u.bindings.size,c);t.i18nAppendBindings(s),n[r]=Pt(l)}}}),n},e.prototype.i18nAllocateRef=function(e){var t,n=this.fileBasedI18nSuffix.toUpperCase();if(this.i18nUseExternalIds){t=""+(r=ci("EXTERNAL_"))+e+"$$"+this.constantPool.uniqueName(n)}else{var r=ci(n);t=this.constantPool.uniqueName(r)}return Et(t)},e.prototype.i18nUpdateRef=function(e){var t=e.icus,n=e.meta,r=e.isRoot,i=e.isResolved;if(r&&i&&!ti(n)){var o=e.getSerializedPlaceholders(),a={},s=o.size?oi(o):{};t.size&&t.forEach(function(e,t){if(1===e.length)s[t]=e[0];else{var n=ii("I18N_EXP_"+t);s[t]=Pt(n),a[t]=At(e)}});var u=void 0;(Array.from(o.values()).some(function(e){return e.length>1})||Object.keys(a).length)&&(u=function(e){var t=[e];return Object.keys(a).length&&t.push(Ur(a,!0)),vl(null,Xn.i18nPostprocess,t)}),this.i18nTranslate(n,s,e.ref,u)}},e.prototype.i18nStart=function(e,t,n){void 0===e&&(e=null);var r=this.allocateDataSlot();if(this.i18nContext)this.i18n=this.i18nContext.forkChildContext(r,this.templateIndex,t);else{var i=this.i18nAllocateRef(t.id);this.i18n=new Zc(r,i,0,this.templateIndex,t)}var o=this.i18n,a=o.id,s=o.ref,u=[Pt(r),s];a>0&&u.push(Pt(a)),this.creationInstruction(e,n?Xn.i18n:Xn.i18nStart,u)},e.prototype.i18nEnd=function(e,t){var n=this;if(void 0===e&&(e=null),!this.i18n)throw new Error("i18nEnd is executed with no i18n context present");this.i18nContext?(this.i18nContext.reconcileChildContext(this.i18n),this.i18nUpdateRef(this.i18nContext)):this.i18nUpdateRef(this.i18n);var r=this.i18n,i=r.index,o=r.bindings;o.size&&(o.forEach(function(t){return n.updateInstruction(e,Xn.i18nExp,[t])}),this.updateInstruction(e,Xn.i18nApply,[Pt(i)])),t||this.creationInstruction(e,Xn.i18nEnd),this.i18n=null},e.prototype.visitContent=function(e){this._hasNgContent=!0;var t=this.allocateDataSlot(),n="*"===e.selector?0:this._ngContentSelectors.push(e.selector)+this._ngContentSelectorsOffset,r=[Pt(t)],i=[];e.attributes.forEach(function(e){var t=e.name,n=e.value;"select"!==t.toLowerCase()&&i.push(t,n)}),i.length>0?r.push(Pt(n),vi(i)):0!==n&&r.push(Pt(n)),this.creationInstruction(e.sourceSpan,Xn.projection,r)},e.prototype.getNamespaceInstruction=function(e){switch(e){case"math":return Xn.namespaceMathML;case"svg":return Xn.namespaceSVG;default:return Xn.namespaceHTML}},e.prototype.addNamespaceInstruction=function(e,t){this._namespace=e,this.creationInstruction(t.sourceSpan,e)},e.prototype.visitElement=function(e){var t,n,r=this,i=this.allocateDataSlot(),u=new Hs(Pt(i),null),c=!1,l=ei(e.i18n)&&!ti(e.i18n);if(l&&this.i18n)throw new Error("Could not mark an element as translatable inside of a translatable section");var d=[],p=[],f=Object(o.__read)(a(e.name),2),h=f[0],m=f[1],g=s(e.name);try{for(var v=Object(o.__values)(e.attributes),_=v.next();!_.done;_=v.next()){var y=_.value,b=y.name,j=y.value;"ngNonBindable"===b?c=!0:"style"===b?u.registerStyleAttr(j):"class"===b?u.registerClassAttr(j):y.i18n?d.push(y):p.push(y)}}catch(e){t={error:e}}finally{try{_&&!_.done&&(n=v.return)&&n.call(v)}finally{if(t)throw t.error}}this.matchDirectives(e.name,e);var x=[Pt(i)];g||x.push(Pt(m));var O=[],w=[];e.inputs.forEach(function(e){u.registerBoundInput(e)||(0===e.type&&e.i18n?d.push(e):w.push(e))}),p.forEach(function(e){return O.push(Pt(e.name),Pt(e.value))}),O.push.apply(O,Object(o.__spread)(this.prepareSyntheticAndSelectOnlyAttrs(w,e.outputs,u))),x.push(this.toAttrsParam(O)),x.push(this.prepareRefsParameter(e.references));var C=this._namespace,E=this.getNamespaceInstruction(h);E!==C&&this.addNamespaceInstruction(E,e);var S=Et(pi);this.i18n&&this.i18n.appendElement(e.i18n,i);var k=!u.hasBindingsOrInitialValues()&&!g&&0===e.outputs.length&&0===d.length&&!(!l&&r.i18n?!El(e.children):e.children.length>0),M=!k&&!u.hasBindingsOrInitialValues()&&El(e.children);if(k)this.creationInstruction(e.sourceSpan,Xn.element,yi(x));else{if(this.creationInstruction(e.sourceSpan,g?Xn.elementContainerStart:Xn.elementStart,yi(x)),c&&this.creationInstruction(e.sourceSpan,Xn.disableBindings),l&&this.i18nStart(e.sourceSpan,e.i18n,M),d.length){var A=!1,T=[];if(d.forEach(function(t){var n=t.i18n;if(t instanceof hc)T.push(Pt(t.name),r.i18nTranslate(n));else{var i=t.value.visit(r._valueConverter);if(r.allocateBindingSlots(i),i instanceof Mo){var o=oi(si(n));T.push(Pt(t.name),r.i18nTranslate(n,o)),i.expressions.forEach(function(t){A=!0;var n=r.convertExpressionBinding(S,t);r.updateInstruction(e.sourceSpan,Xn.i18nExp,[n])})}}}),T.length){var D=Pt(this.allocateDataSlot()),I=this.constantPool.getConstLiteral(At(T),!0);this.creationInstruction(e.sourceSpan,Xn.i18nAttributes,[D,I]),A&&this.updateInstruction(e.sourceSpan,Xn.i18nApply,[D])}}this.processStylingInstruction(S,u.buildElementStylingInstruction(e.sourceSpan,this.constantPool),!0),e.outputs.forEach(function(t){r.creationInstruction(t.sourceSpan,Xn.listener,r.prepareListenerParameter(e.name,t,i))})}if(u.buildUpdateLevelInstructions(this._valueConverter).forEach(function(e){r.processStylingInstruction(S,e,!1)}),w.forEach(function(e){var t=function(e){switch(e){case 0:case 4:return Xn.elementProperty;case 2:return Xn.elementClassProp;case 1:return Xn.elementAttribute;default:return}}(e.type);if(4===e.type){var n=e.value.visit(r._valueConverter);if(n.name||n.value){var a=Pi(e.name);r.allocateBindingSlots(n),r.updateInstruction(e.sourceSpan,Xn.elementProperty,function(){return[Pt(i),Pt(a),r.convertPropertyBinding(S,n)]})}}else if(t){var s=[],u=function(e,t){switch(t){case R.HTML:return St(Xn.sanitizeHtml);case R.SCRIPT:return St(Xn.sanitizeScript);case R.STYLE:return 1===e.type?St(Xn.sanitizeStyle):null;case R.URL:return St(Xn.sanitizeUrl);case R.RESOURCE_URL:return St(Xn.sanitizeResourceUrl);default:return null}}(e,e.securityContext);u&&s.push(u);var c=e.value.visit(r._valueConverter);r.allocateBindingSlots(c),r.updateInstruction(e.sourceSpan,t,function(){return Object(o.__spread)([Pt(i),Pt(e.name),r.convertPropertyBinding(S,c)],s)})}else r._unsupported("binding type "+e.type)}),Oc(this,e.children),!l&&this.i18n&&this.i18n.appendElement(e.i18n,i,!0),!k){var P=e.endSourceSpan||e.sourceSpan;l&&this.i18nEnd(P,M),c&&this.creationInstruction(P,Xn.enableBindings),this.creationInstruction(P,g?Xn.elementContainerEnd:Xn.elementEnd)}},e.prototype.visitTemplate=function(t){var n=this,r=this.allocateDataSlot();this.i18n&&this.i18n.appendTemplate(t.i18n,r);var i=xn(t.tagName||""),a=(i?this.contextName+"_"+i:"")+"_"+r,s=a+"_Template",u=[Pt(r),Et(s),Pt(t.tagName)];this.matchDirectives("ng-template",t);var c=[];t.attributes.forEach(function(e){c.push(vi(e.name),vi(e.value))}),c.push.apply(c,Object(o.__spread)(this.prepareSyntheticAndSelectOnlyAttrs(t.inputs,t.outputs))),u.push(this.toAttrsParam(c)),t.references&&t.references.length&&(u.push(this.prepareRefsParameter(t.references)),u.push(St(Xn.templateRefExtractor)));var l=Et(pi);t.inputs.forEach(function(e){var i=e.value.visit(n._valueConverter);n.allocateBindingSlots(i),n.updateInstruction(t.sourceSpan,Xn.elementProperty,function(){return[Pt(r),Pt(e.name),n.convertPropertyBinding(l,i)]})});var d=new e(this.constantPool,this._bindingScope,this.level+1,a,this.i18n,r,s,[],this.directiveMatcher,this.directives,this.pipeTypeByName,this.pipes,this._namespace,this.fileBasedI18nSuffix,this.i18nUseExternalIds);this._nestedTemplateFns.push(function(){var e,r=d.buildTemplateFunction(t.children,t.variables,n._ngContentSelectors.length+n._ngContentSelectorsOffset,t.i18n);n.constantPool.statements.push(r.toDeclStmt(s,null)),d._hasNgContent&&(n._hasNgContent=!0,(e=n._ngContentSelectors).push.apply(e,Object(o.__spread)(d._ngContentSelectors)))}),this.creationInstruction(t.sourceSpan,Xn.templateCreate,function(){return u.splice(2,0,Pt(d.getConstCount()),Pt(d.getVarCount())),yi(u)}),t.outputs.forEach(function(e){n.creationInstruction(e.sourceSpan,Xn.listener,n.prepareListenerParameter("ng_template",e,r))})},e.prototype.visitBoundText=function(e){var t=this;if(this.i18n){var n=e.value.visit(this._valueConverter);return this.allocateBindingSlots(n),void(n instanceof Mo&&(this.i18n.appendBoundText(e.i18n),this.i18nAppendBindings(n.expressions)))}var r=this.allocateDataSlot();this.creationInstruction(e.sourceSpan,Xn.text,[Pt(r)]);var i=e.value.visit(this._valueConverter);this.allocateBindingSlots(i),this.updateInstruction(e.sourceSpan,Xn.textBinding,function(){return[Pt(r),t.convertPropertyBinding(Et(pi),i)]})},e.prototype.visitText=function(e){this.i18n||this.creationInstruction(e.sourceSpan,Xn.text,[Pt(this.allocateDataSlot()),Pt(e.value)])},e.prototype.visitIcu=function(e){var t=!1;this.i18n||(t=!0,this.i18nStart(null,e.i18n,!0));var n=this.i18n,r=this.i18nBindProps(e.vars),i=this.i18nBindProps(e.placeholders),o=e.i18n,a=function(e){return vl(null,Xn.i18nPostprocess,[e,Ur(r,!0)])};if(ti(n.meta))this.i18nTranslate(o,i,n.ref,a);else{var s=this.i18nTranslate(o,i,void 0,a);n.appendIcu(ri(o).name,s)}return t&&this.i18nEnd(null,!0),null},e.prototype.allocateDataSlot=function(){return this._dataIndex++},e.prototype.getConstCount=function(){return this._dataIndex},e.prototype.getVarCount=function(){return this._pureFunctionSlots},e.prototype.bindingContext=function(){return""+this._bindingContext++},e.prototype.instructionFn=function(e,t,n,r,i){void 0===i&&(i=!1),e[i?"unshift":"push"](function(){var e=Array.isArray(r)?r:r();return vl(t,n,e).toStmt()})},e.prototype.processStylingInstruction=function(e,t,n){var r=this;if(t){var i=function(){return t.buildParams(function(t){return r.convertPropertyBinding(e,t,!0)})};n?this.creationInstruction(t.sourceSpan,t.reference,i):this.updateInstruction(t.sourceSpan,t.reference,i)}},e.prototype.creationInstruction=function(e,t,n,r){this.instructionFn(this._creationCodeFns,e,t,n||[],r)},e.prototype.updateInstruction=function(e,t,n){this.instructionFn(this._updateCodeFns,e,t,n||[])},e.prototype.allocatePureFunctionSlots=function(e){var t=this._pureFunctionSlots;return this._pureFunctionSlots+=e,t},e.prototype.allocateBindingSlots=function(e){this._bindingSlots+=e instanceof Mo?e.expressions.length:1},e.prototype.convertExpressionBinding=function(e,t){var n=ia(this,e,t,this.bindingContext(),Jo.TrySimple).currValExpr;return St(Xn.bind).callFn([n])},e.prototype.convertPropertyBinding=function(e,t,n){var r,i=t instanceof Mo?xl:function(){return Wt("Unexpected interpolation")},a=ia(this,e,t,this.bindingContext(),Jo.TrySimple,i);(r=this._tempVariables).push.apply(r,Object(o.__spread)(a.stmts));var s=a.currValExpr;return t instanceof Mo||n?s:St(Xn.bind).callFn([s])},e.prototype.matchDirectives=function(e,t){var n,r,i,o=this;if(this.directiveMatcher){var a=(n=e,r=Ci(t),(i=new O).setElement(n),Object.getOwnPropertyNames(r).forEach(function(e){var t=r[e];i.addAttribute(e,t),"class"===e.toLowerCase()&&t.trim().split(/\s+/).forEach(function(e){return i.addClassName(e)})}),i);this.directiveMatcher.match(a,function(e,t){o.directives.add(t)})}},e.prototype.prepareSyntheticAndSelectOnlyAttrs=function(e,t,n){var r=[],i=[],o=new Set;function a(e,t){"string"==typeof e?o.has(e)||(r.push(Pt(e)),void 0!==t&&r.push(t),o.add(e)):r.push(Pt(e))}if(e.length){var s=vi("");e.forEach(function(e){if(4===e.type){var t=e.value;if(t instanceof No){var n=t.ast;n instanceof Eo&&void 0===n.value&&a(Pi(e.name),s)}}else i.push(e)})}return n&&n.populateInitialStylingAttrs(r),(i.length||t.length)&&(a(3),i.forEach(function(e){return a(e.name)}),t.forEach(function(e){var t,n;a(1===e.type?(t=e.name,n=t.indexOf("."),(t=n>0?t.substring(0,n):t).charAt(0)!==Ii&&(t=Ii+t),t):e.name)})),r},e.prototype.toAttrsParam=function(e){return e.length>0?this.constantPool.getConstLiteral(At(e),!0):tt},e.prototype.prepareRefsParameter=function(e){var t=this;if(!e||0===e.length)return tt;var n=Hn(e.map(function(e){var n=t.allocateDataSlot(),r=t._bindingScope.freshReferenceName(),i=t.level,o=Et(r);return t._bindingScope.set(i,e.name,o,0,function(e,t){var r=t>0?[_l(t).toStmt()]:[],i=o.set(St(Xn.reference).callFn([Pt(n)]));return r.concat(i.toConstDecl())},!0),[e.name,e.value]}));return this.constantPool.getConstLiteral(vi(n),!0)},e.prototype.prepareListenerParameter=function(e,t,n){var r,i=this,a=t.name;1===t.type?(r=Ni(a,t.phase),a=Ri(a,t.phase)):r=xn(a);var s=xn(e),u=this.templateName+"_"+s+"_"+r+"_"+n+"_listener";return function(){var e=i._bindingScope.nestedScope(i._bindingScope.bindingLevel),n=Zo(e,Et(pi),t.handler,"b",function(){return Wt("Unexpected interpolation")}),r=Object(o.__spread)(e.restoreViewStatement(),e.variableDeclarations(),n.render3Stmts),s=Dt([new Ge("$event",me)],r,ge,null,u);return[Pt(a),s]}},e}(),hl=function(e){function t(t,n,r,i){var o=e.call(this)||this;return o.constantPool=t,o.allocateSlot=n,o.allocatePureFunctionSlots=r,o.definePipe=i,o._pipeBindExprs=[],o}return Object(o.__extends)(t,e),t.prototype.visitPipe=function(e,t){var n,r,i=this.allocateSlot(),a="PIPE:"+i,s=this.allocatePureFunctionSlots(2+e.args.length),u=new bo(e.span,new vo(e.span),a),c=(n=e.args,{identifier:(r=ml[n.length])||Xn.pipeBindV,isVarLength:!r}),l=c.identifier,d=c.isVarLength;this.definePipe(e.name,a,i,St(l));var p=Object(o.__spread)([e.exp],e.args),f=d?this.visitAll([new So(e.span,p)]):this.visitAll(p),h=new Ro(e.span,u,Object(o.__spread)([new Eo(e.span,i),new Eo(e.span,s)],f));return this._pipeBindExprs.push(h),h},t.prototype.updatePipeSlotOffsets=function(e){this._pipeBindExprs.forEach(function(t){t.args[1].value+=e})},t.prototype.visitLiteralArray=function(e,t){var n=this;return new ha(e.span,this.visitAll(e.expressions),function(e){var t=At(e);return e.every(function(e){return e.isConstant()})?n.constantPool.getConstLiteral(t,!0):yl(n.constantPool,t,n.allocatePureFunctionSlots)})},t.prototype.visitLiteralMap=function(e,t){var n=this;return new ha(e.span,this.visitAll(e.values),function(t){var r=Tt(t.map(function(t,n){return{key:e.keys[n].key,value:t,quoted:e.keys[n].quoted}}));return t.every(function(e){return e.isConstant()})?n.constantPool.getConstLiteral(r,!0):yl(n.constantPool,r,n.allocatePureFunctionSlots)})},t}(zo),ml=[Xn.pipeBind1,Xn.pipeBind2,Xn.pipeBind3,Xn.pipeBind4];var gl=[Xn.pureFunction0,Xn.pureFunction1,Xn.pureFunction2,Xn.pureFunction3,Xn.pureFunction4,Xn.pureFunction5,Xn.pureFunction6,Xn.pureFunction7,Xn.pureFunction8];function vl(e,t,n){return St(t,null,e).callFn(n,e)}function _l(e){return St(Xn.nextContext).callFn(e>1?[Pt(e)]:[])}function yl(e,t,n){var r=e.getLiteralFactory(t),i=r.literalFactory,a=r.literalFactoryArguments,s=n(1+a.length);a.length>0||Wt("Expected arguments to a literal factory function");var u,c={identifier:(u=gl[a.length])||Xn.pureFunctionV,isVarLength:!u},l=c.identifier,d=c.isVarLength,p=[Pt(s),i];return d?p.push(At(a)):p.push.apply(p,Object(o.__spread)(a)),St(l).callFn(p)}var bl="$$shared_ctx$$",jl=function(){function e(e,t){void 0===e&&(e=0),void 0===t&&(t=null),this.bindingLevel=e,this.parent=t,this.map=new Map,this.referenceNameIndex=0,this.restoreViewVariable=null}return Object.defineProperty(e,"ROOT_SCOPE",{get:function(){return e._ROOT_SCOPE||(e._ROOT_SCOPE=(new e).set(0,"$event",Et("$event"))),e._ROOT_SCOPE},enumerable:!0,configurable:!0}),e.prototype.get=function(e){for(var t=this;t;){var n=t.map.get(e);if(null!=n)return t!==this&&(n={retrievalLevel:n.retrievalLevel,lhs:n.lhs,declareLocalCallback:n.declareLocalCallback,declare:!1,priority:n.priority,localRef:n.localRef},this.map.set(e,n),this.maybeGenerateSharedContextVar(n),this.maybeRestoreView(n.retrievalLevel,n.localRef)),n.declareLocalCallback&&!n.declare&&(n.declare=!0),n.lhs;t=t.parent}return 0===this.bindingLevel?null:this.getComponentProperty(e)},e.prototype.set=function(e,t,n,r,i,o){return void 0===r&&(r=0),!this.map.has(t)||Wt("The name "+t+" is already defined in scope to be "+this.map.get(t)),this.map.set(t,{retrievalLevel:e,lhs:n,declare:!1,declareLocalCallback:i,priority:r,localRef:o||!1}),this},e.prototype.getLocal=function(e){return this.get(e)},e.prototype.nestedScope=function(t){var n=new e(t,this);return t>0&&n.generateSharedContextVar(0),n},e.prototype.getSharedContextName=function(e){var t=this.map.get(bl+e);return t&&t.declare?t.lhs:null},e.prototype.maybeGenerateSharedContextVar=function(e){if(1===e.priority){var t=this.map.get(bl+e.retrievalLevel);t?t.declare=!0:this.generateSharedContextVar(e.retrievalLevel)}},e.prototype.generateSharedContextVar=function(e){var t=Et(pi+this.freshReferenceName());this.map.set(bl+e,{retrievalLevel:e,lhs:t,declareLocalCallback:function(e,n){return[t.set(_l(n)).toConstDecl()]},declare:!1,priority:2,localRef:!1})},e.prototype.getComponentProperty=function(e){var t=this.map.get(bl+0);return t.declare=!0,this.maybeRestoreView(0,!1),t.lhs.prop(e)},e.prototype.maybeRestoreView=function(e,t){this.isListenerScope()&&(e<this.bindingLevel||t)&&(this.parent.restoreViewVariable||(this.parent.restoreViewVariable=Et(this.parent.freshReferenceName())),this.restoreViewVariable=this.parent.restoreViewVariable)},e.prototype.restoreViewStatement=function(){return this.restoreViewVariable?[vl(null,Xn.restoreView,[this.restoreViewVariable]).toStmt()]:[]},e.prototype.viewSnapshotStatements=function(){var e=vl(null,Xn.getCurrentView,[]);return this.restoreViewVariable?[this.restoreViewVariable.set(e).toConstDecl()]:[]},e.prototype.isListenerScope=function(){return this.parent&&this.parent.bindingLevel===this.bindingLevel},e.prototype.variableDeclarations=function(){var e=this,t=0;return Array.from(this.map.values()).filter(function(e){return e.declare}).sort(function(e,t){return t.retrievalLevel-e.retrievalLevel||t.priority-e.priority}).reduce(function(n,r){var i=e.bindingLevel-r.retrievalLevel,o=r.declareLocalCallback(e,i-t);return t=i,n.concat(o)},[])},e.prototype.freshReferenceName=function(){for(var e=this;e.parent;)e=e.parent;return"_r"+e.referenceNameIndex++},e}();function xl(e){switch((e=e.slice(1)).length){case 3:return St(Xn.interpolation1).callFn(e);case 5:return St(Xn.interpolation2).callFn(e);case 7:return St(Xn.interpolation3).callFn(e);case 9:return St(Xn.interpolation4).callFn(e);case 11:return St(Xn.interpolation5).callFn(e);case 13:return St(Xn.interpolation6).callFn(e);case 15:return St(Xn.interpolation7).callFn(e);case 17:return St(Xn.interpolation8).callFn(e)}return e.length>=19&&e.length%2==1||Wt("Invalid interpolation argument length "+e.length),St(Xn.interpolationV).callFn([At(e)])}function Ol(e,t,n){void 0===n&&(n={});var r=n.interpolationConfig,i=n.preserveWhitespaces,o=wl(r),a=(new Uu).parse(e,t,!0,r);if(a.errors&&a.errors.length>0)return{errors:a.errors,nodes:[]};var s=a.rootNodes;s=_u(new cl(r,!i),s),i||(s=_u(new Xu,s),s=_u(new cl(r,!1),s));var u=Hc(s,o),c=u.nodes,l=u.errors;return l&&l.length>0?{errors:l,nodes:[]}:{nodes:c}}function wl(e){return void 0===e&&(e=Ui),new uc(new uu(new Ks),e,new sc,null,[])}function Cl(e){return e instanceof pc||e instanceof fc||e instanceof xc}function El(e){return e.every(Cl)}var Sl=[],kl=/attr\.([^\]]+)/;function Ml(e){return e.substring(0,5).toLowerCase()}function Al(e,t,n){var r=new wi;r.set("type",e.type),r.set("selectors",vi(ne(e.selector)));var i=Ei({name:e.name,type:e.type,deps:e.deps,injectFn:Xn.directiveInject});r.set("factory",i.factory),r.set("contentQueries",function(e,t){if(e.queries.length){var n=e.queries.map(function(e){var n=Ll(e,t,null);return St(Xn.registerContentQuery).callFn([n,Et("dirIndex")]).toStmt()}),r=e.name,i=[new Ge("dirIndex",_e)];return Dt(i,n,ge,null,r?r+"_ContentQueries":null)}return null}(e,t)),r.set("contentQueriesRefresh",function(e){if(e.queries.length>0){var t=[],n=e.name,r=[new Ge("dirIndex",_e),new Ge("queryStartIndex",_e)],i=Et("instance"),o=hi(t,di);return t.push(i.set(St(Xn.load).callFn([Et("dirIndex")])).toDeclStmt(ge,[Ie.Final])),e.queries.forEach(function(e,n){var r=Et("queryStartIndex"),a=St(Xn.loadQueryList).callFn([n>0?r.plus(Pt(n)):r]),s=o().set(a),u=St(Xn.queryRefresh).callFn([s]),c=i.prop(e.propertyName).set(e.first?o().prop("first"):o()),l=u.and(c);t.push(l.toStmt())}),Dt(r,t,ge,null,n?n+"_ContentQueriesRefresh":null)}return null}(e));for(var a=Object.keys(e.host.properties).filter(function(e){var t=Ml(e);return"style"!==t&&"class"!==t}).length,s=Et("elIndex"),u=Et(pi),c=new Hs(s,u),l={},d=Object.getOwnPropertyNames(e.host.attributes),p=0;p<d.length;p++){var f=d[p],h=e.host.attributes[f];switch(f){case"style":c.registerStyleAttr(h);break;case"class":c.registerClassAttr(h);break;default:l[f]=h}}return r.set("attributes",function(e){var t,n,r=[];try{for(var i=Object(o.__values)(Object.getOwnPropertyNames(e)),a=i.next();!a.done;a=i.next()){var s=a.value,u=e[s];r.push(Pt(s),Pt(u))}}catch(e){t={error:e}}finally{try{a&&!a.done&&(n=i.return)&&n.call(i)}finally{if(t)throw t.error}}if(r.length>0)return At(r);return null}(l)),r.set("hostBindings",function(e,t,n,r,i,a,s){var u,c,l=[],d=[],p=s,f=e.typeSourceSpan,h=(g=e,{hostAttributes:g.host.attributes,hostListeners:g.host.listeners,hostProperties:g.host.properties}),m=i.createDirectiveHostEventAsts(h,f);var g;if(m&&m.length){var v=(_=n,y=e,m.map(function(e){var t=Zo(null,_,e.handler,"b",function(){return Wt("Unexpected interpolation")}),n=e.name&&xn(e.name),r=n;1===e.type&&(r=Ni(n,e.targetOrPhase),n=Ri(n,e.targetOrPhase));var i=y.name,a=i&&n?i+"_"+r+"_HostBindingHandler":null,s=Dt([new Ge("$event",me)],Object(o.__spread)(t.render3Stmts),ge,null,a);return St(Xn.listener).callFn([Pt(n),s]).toStmt()}));l.push.apply(l,Object(o.__spread)(v))}var _,y;var b=i.createBoundHostProperties(h,f),j=function(e,t){return ia(null,e,t,"b",Jo.TrySimple,function(){return Wt("Unexpected interpolation")})};if(b){var x=new hl(a,function(){return Wt("Unexpected node")},function(e){var t=p;return p+=e,t},function(){return Wt("Unexpected pipe")});try{for(var O=Object(o.__values)(b),w=O.next();!w.done;w=O.next()){var C=w.value,E=C.name,S=Ml(E);if("style"===S){var k=Wl(E),M=k.propertyName,A=k.unit;r.registerStyleInput(M,C.expression,A,C.sourceSpan)}else if("class"===S)r.registerClassInput(Wl(E).propertyName,C.expression,C.sourceSpan);else{var T=C.expression.visit(x),D=j(n,T),I=Hl(C),P=I.bindingName,R=I.instruction,N=I.extraParams,F=[t,Pt(P),St(Xn.bind).callFn([D.currValExpr])];d.push.apply(d,Object(o.__spread)(D.stmts)),d.push(St(R).callFn(F.concat(N)).toStmt())}}}catch(e){u={error:e}}finally{try{w&&!w.done&&(c=O.return)&&c.call(O)}finally{if(u)throw u.error}}if(r.hasBindingsOrInitialValues()){var L=r.buildDirectiveHostAttrsInstruction(null,a);L&&l.push(Ul(L,n,j));var V=r.buildElementStylingInstruction(null,a);V&&l.push(Ul(V,n,j)),r.buildUpdateLevelInstructions(x).forEach(function(e){d.push(Ul(e,n,j))})}}p&&l.unshift(St(Xn.allocHostVars).callFn([Pt(p)]).toStmt());if(l.length>0||d.length>0){var B=e.name?e.name+"_HostBindings":null,z=[];return l.length>0&&z.push(pl(1,l)),d.length>0&&z.push(pl(2,d)),Dt([new Ge(fi,_e),new Ge(pi,null),new Ge(t.name,_e)],z,ge,null,B)}return null}(e,s,u,c,n,t,a)),r.set("inputs",_i(e.inputs,!0)),r.set("outputs",_i(e.outputs)),null!==e.exportAs&&r.set("exportAs",Pt(e.exportAs)),{definitionMap:r,statements:i.statements}}function Tl(e,t){var n=[],r=t.providers,i=t.viewProviders;if(r||i){var o=[r||new Xe([])];i&&o.push(i),n.push(St(Xn.ProvidersFeature).callFn(o))}t.usesInheritance&&n.push(St(Xn.InheritDefinitionFeature)),t.lifecycle.usesOnChanges&&n.push(St(Xn.NgOnChangesFeature)),n.length&&e.set("features",At(n))}function Dl(e,t,n){var r=Al(e,t,n),i=r.definitionMap,o=r.statements;Tl(i,e);var a=St(Xn.defineDirective).callFn([i.toLiteralMap()]);(e.selector||"").replace(/\n/g,"");return{expression:a,type:zl(e,Xn.DirectiveDefWithMeta),statements:o}}function Il(e){var t=new wi;if(e.inputs){var n=e.inputs,r=Object.keys(n).map(function(e){var t=n[e];return{key:e,value:Array.isArray(t)?At(t.map(function(e){return Pt(e)})):Pt(t),quoted:!1}});t.set("inputs",Tt(r))}if(e.outputs){var i=e.outputs,o=Object.keys(i).map(function(e){return{key:e,value:Pt(i[e]),quoted:!1}});t.set("outputs",Tt(o))}return{expression:St(Xn.defineBase).callFn([t.toLiteralMap()]),type:new pe(St(Xn.BaseDef))}}function Pl(e,t,n){var r,i,a=Al(e,t,n),s=a.definitionMap,u=a.statements;Tl(s,e);var c=e.selector&&O.parse(e.selector),l=c&&c[0];if(l){var d=l.getAttrs();d.length&&s.set("attrs",t.getConstLiteral(At(d.map(function(e){return Pt(null!=e?e:void 0)})),!0))}var p=null;if(e.directives.length>0){var f=new w;try{for(var h=Object(o.__values)(e.directives),m=h.next();!m.done;m=h.next()){var g=m.value,v=g.selector,y=g.expression;f.addSelectables(O.parse(v),y)}}catch(e){r={error:e}}finally{try{m&&!m.done&&(i=h.return)&&i.call(h)}finally{if(r)throw r.error}}p=f}e.viewQueries.length&&s.set("viewQuery",function(e,t){for(var n=[],r=[],i=hi(r,di),o=0;o<e.viewQueries.length;o++){var a=e.viewQueries[o],s=Ll(a,t,o);n.push(s.toStmt());var u=i(),c=St(Xn.load).callFn([Pt(o)]),l=St(Xn.queryRefresh).callFn([u.set(c)]),d=Et(pi).prop(a.propertyName).set(a.first?u.prop("first"):u);r.push(l.and(d).toStmt())}var p=e.name?e.name+"_Query":null;return Dt([new Ge(fi,_e),new Ge(pi,null)],[pl(1,n),pl(2,r)],ge,null,p)}(e,t));var j,x,C,E,S=e.name,k=S?S+"_Template":null,M=new Set,A=new Set,T=e.changeDetection,D=e.template,I=new fl(t,jl.ROOT_SCOPE,0,S,null,null,k,e.viewQueries,p,M,e.pipes,A,Xn.namespaceHTML,e.relativeContextFilePath,e.i18nUseExternalIds),P=I.buildTemplateFunction(D.nodes,[]);if(s.set("consts",Pt(I.getConstCount())),s.set("vars",Pt(I.getVarCount())),s.set("template",P),M.size){var R=At(Array.from(M));e.wrapDirectivesAndPipesInClosure&&(R=Dt([],[new at(R)])),s.set("directives",R)}if(A.size){var N=At(Array.from(A));e.wrapDirectivesAndPipesInClosure&&(N=Dt([],[new at(N)])),s.set("pipes",N)}if(null===e.encapsulation&&(e.encapsulation=_.Emulated),e.styles&&e.styles.length){var F=(e.encapsulation==_.Emulated?(j=e.styles,x=Rs,C=Ps,E=new ss,j.map(function(e){return E.shimCssText(e,x,C)})):e.styles).map(function(e){return Pt(e)});s.set("styles",At(F))}else e.encapsulation===_.Emulated&&(e.encapsulation=_.None);e.encapsulation!==_.Emulated&&s.set("encapsulation",Pt(e.encapsulation)),null!==e.animations&&s.set("data",Tt([{key:"animation",value:e.animations,quoted:!1}])),null!=T&&T!==b.Default&&s.set("changeDetection",Pt(T));(e.selector||"").replace(/\n/g,"");return{expression:St(Xn.defineComponent).callFn([s.toLiteralMap()]),type:zl(e,Xn.ComponentDefWithMeta),statements:u}}function Rl(e,t,n,r,i,a,s){var u=En(t.type);u||Wt("Cannot resolver the name of "+t.type);var c,l,d,p=e.constantPool.propertyNameOf(2),f=t.toSummary(),h=Pl(Object(o.__assign)({},Nl(t,e,r),{selector:t.selector,template:{nodes:n.nodes},directives:[],pipes:(c=s,l=e,d=Array.from(c).map(function(e){var t=Object(o.__read)(e,2),n=t[0],r=t[1];return[n,l.importExpr(r)]}),new Map(d)),viewQueries:Fl(t.viewQueries,e),wrapDirectivesAndPipesInClosure:!1,styles:f.template&&f.template.styles||Sl,encapsulation:f.template&&f.template.encapsulation||_.Emulated,interpolation:Ui,animations:null,viewProviders:t.viewProviders.length>0?new Me(t.viewProviders):null,relativeContextFilePath:"",i18nUseExternalIds:!0}),e.constantPool,i);e.statements.push(new dt(u,null,[new ut(p,ge,[Ie.Static],h.expression)],[],new ct(null,[],[]),[]))}function Nl(e,t,n){var r=e.toSummary(),i=En(e.type);return i||Wt("Cannot resolver the name of "+e.type),{name:i,type:t.importExpr(e.type.reference),typeArgumentCount:0,typeSourceSpan:as(e.isComponent?"Component":"Directive",e.type),selector:e.selector,deps:ki(e.type,t,n),queries:Fl(e.queries,t),lifecycle:{usesOnChanges:e.type.lifecycleHooks.some(function(e){return e==pa.OnChanges})},host:{attributes:e.hostAttributes,listeners:r.hostListeners,properties:r.hostProperties},inputs:e.inputs,outputs:e.outputs,usesInheritance:!1,exportAs:null,providers:e.providers.length>0?new Me(e.providers):null}}function Fl(e,t){return e.map(function(e){var n=null;return e.read&&e.read.identifier&&(n=t.importExpr(e.read.identifier.reference)),{propertyName:e.propertyName,first:e.first,predicate:function(e,t){if(e.length>1||1==e.length&&e[0].value){var n=e.map(function(e){return e.value});return n.some(function(e){return!e})&&Wt("Found a type among the string selectors expected"),t.constantPool.getConstLiteral(At(n.map(function(e){return Pt(e)})))}if(1==e.length){var r=e[0];if(r.identifier)return t.importExpr(r.identifier.reference)}return Wt("Unexpected query form"),et}(e.selectors,t),descendants:e.descendants,read:n}})}function Ll(e,t,n){var r=function(e,t){if(Array.isArray(e.predicate)){var n=[];return e.predicate.forEach(function(e){var t=e.split(",").map(function(e){return Pt(e.trim())});n.push.apply(n,Object(o.__spread)(t))}),t.getConstLiteral(At(n),!0)}return e.predicate}(e,t),i=[Pt(n,ge),r,Pt(e.descendants)];return e.read&&i.push(e.read),St(Xn.query).callFn(i)}function Vl(e){return Mt(Pt(e))}function Bl(e){return Mt(Tt(Object.keys(e).map(function(t){return{key:t,value:Pt(Array.isArray(e[t])?e[t][0]:e[t]),quoted:!0}})))}function zl(e,t){var n,r=(e.selector||"").replace(/\n/g,"");return Mt(St(t,[Di(e.type,e.typeArgumentCount),Vl(r),null!==e.exportAs?Vl(e.exportAs):ye,Bl(e.inputs),Bl(e.outputs),(n=e.queries.map(function(e){return e.propertyName}),n.length>0?Mt(At(n.map(function(e){return Pt(e)}))):ye)]))}function Ul(e,t,n){var r=e.buildParams(function(e){return n(t,e).currValExpr});return St(e.reference,null,e.sourceSpan).callFn(r,e.sourceSpan).toStmt()}function Hl(e){var t,n=e.name,r=[],i=n.match(kl);return i?(n=i[1],t=Xn.elementAttribute):(e.isAnimation?(n=Pi(n),t=Xn.componentHostSyntheticProperty):t=Xn.elementProperty,r.push(Pt(null),Pt(!0))),{bindingName:n,instruction:t,extraParams:r}}var ql=/^(?:\[([^\]]+)\])|(?:\(([^\)]+)\))$/;function Gl(e){var t={},n={},r={};return Object.keys(e).forEach(function(i){var o=e[i],a=i.match(ql);null===a?t[i]=o:null!=a[1]?r[a[1]]=o:null!=a[2]&&(n[a[2]]=o)}),{attributes:t,listeners:n,properties:r}}function Wl(e){var t="",n="",r=e.indexOf(".");if(r>0){var i=e.lastIndexOf(".");i!==r?(t=e.substring(i+1,e.length),n=e.substring(r+1,i)):n=e.substring(r+1,e.length)}return{propertyName:n,unit:t}}var Yl=function(){function e(){this.R3ResolvedDependencyType=xi,this.elementSchemaRegistry=new sc}return e.prototype.compilePipe=function(e,t,n){var r=lo({name:n.name,type:new Me(n.type),deps:rd(n.deps),pipeName:n.pipeName,pure:n.pure});return ao(r.expression,e,t,r.statements)},e.prototype.compileInjectable=function(e,t,n){var r,i=Fi({name:n.name,type:new Me(n.type),typeArgumentCount:n.typeArgumentCount,providedIn:(r=n.providedIn,null==r||"string"==typeof r?new Le(r):new Me(r)),useClass:td(n,Kl),useFactory:td(n,Ql),useValue:td(n,Xl),useExisting:td(n,Zl),ctorDeps:rd(n.ctorDeps),userDeps:rd(n.userDeps)||void 0});return ao(i.expression,e,t,i.statements)},e.prototype.compileInjector=function(e,t,n){var r=uo({name:n.name,type:new Me(n.type),deps:rd(n.deps),providers:new Me(n.providers),imports:new Me(n.imports)});return ao(r.expression,e,t,r.statements)},e.prototype.compileNgModule=function(e,t,n){return ao(so({type:new Me(n.type),bootstrap:n.bootstrap.map($l),declarations:n.declarations.map($l),imports:n.imports.map($l),exports:n.exports.map($l),emitInline:!0}).expression,e,t,[])},e.prototype.compileDirective=function(e,t,n){var r=new dn,i=wl(),a=Dl(ed(n),r,i),s=Object(o.__spread)(r.statements,a.statements);return ao(a.expression,e,t,s)},e.prototype.compileComponent=function(e,t,n){var r=new dn,i=n.interpolation?zi.fromArray(n.interpolation):Ui,a=Ol(n.template,t,{preserveWhitespaces:n.preserveWhitespaces||!1,interpolationConfig:i});if(void 0!==a.errors){var s=a.errors.map(function(e){return e.toString()}).join(", ");throw new Error("Errors during JIT compilation of template for "+n.name+": "+s)}var u=Pl(Object(o.__assign)({},n,ed(n),{selector:n.selector||this.elementSchemaRegistry.getDefaultComponentElementName(),template:a,viewQueries:n.viewQueries.map(Jl),wrapDirectivesAndPipesInClosure:!1,styles:n.styles||[],encapsulation:n.encapsulation,interpolation:i,changeDetection:n.changeDetection,animations:null!=n.animations?new Me(n.animations):null,viewProviders:null!=n.viewProviders?new Me(n.viewProviders):null,relativeContextFilePath:"",i18nUseExternalIds:!0}),r,wl(i)),c=Object(o.__spread)(r.statements,u.statements);return ao(u.expression,e,t,c)},e}(),Kl=Object.keys({useClass:null})[0],Ql=Object.keys({useFactory:null})[0],Xl=Object.keys({useValue:null})[0],Zl=Object.keys({useExisting:null})[0],$l=function(e){var t=new Me(e);return{value:t,type:t}};function Jl(e){return Object(o.__assign)({},e,{predicate:Array.isArray(e.predicate)?e.predicate:new Me(e.predicate),read:e.read?new Me(e.read):null})}function ed(e){var t=id(e.inputs||[]),n=id(e.outputs||[]),r=e.propMetadata,i={},a={},s=function(e){r.hasOwnProperty(e)&&r[e].forEach(function(t){"Input"===t.ngMetadataName?i[e]=t.bindingPropertyName?[t.bindingPropertyName,e]:e:"Output"===t.ngMetadataName&&(a[e]=t.bindingPropertyName||e)})};for(var u in r)s(u);return Object(o.__assign)({},e,{typeSourceSpan:null,type:new Me(e.type),deps:rd(e.deps),host:function(e,t){var n=Gl(e||{}),r=n.attributes,i=n.listeners,o=n.properties,a=function(e){t.hasOwnProperty(e)&&t[e].forEach(function(t){"HostBinding"===t.ngMetadataName?o[t.hostPropertyName||e]=e:"HostListener"===t.ngMetadataName&&(i[t.eventName||e]=e+"("+(t.args||[]).join(",")+")")})};for(var s in t)a(s);return{attributes:r,listeners:i,properties:o}}(e.host,e.propMetadata),inputs:Object(o.__assign)({},t,i),outputs:Object(o.__assign)({},n,a),queries:e.queries.map(Jl),providers:null!=e.providers?new Me(e.providers):null})}function td(e,t){return e.hasOwnProperty(t)?new Me(e[t]):void 0}function nd(e){return{token:null===e.token?new Le(null):e.resolved===xi.Attribute?new Le(e.token):new Me(e.token),resolved:e.resolved,host:e.host,optional:e.optional,self:e.self,skipSelf:e.skipSelf}}function rd(e){return null==e?null:e.map(nd)}function id(e){return e.reduce(function(e,t){var n=Object(o.__read)(t.split(",").map(function(e){return e.trim()}),2),r=n[0],i=n[1];return e[r]=i||r,e},{})}function od(e){(e.ng||(e.ng={})).ɵcompilerFacade=new Yl}var ad,sd,ud,cd=new on("7.2.0"),ld=function(){function e(e,t,n){this.value=e,this.ngContentIndex=t,this.sourceSpan=n}return e.prototype.visit=function(e,t){return e.visitText(this,t)},e}(),dd=function(){function e(e,t,n){this.value=e,this.ngContentIndex=t,this.sourceSpan=n}return e.prototype.visit=function(e,t){return e.visitBoundText(this,t)},e}(),pd=function(){function e(e,t,n){this.name=e,this.value=t,this.sourceSpan=n}return e.prototype.visit=function(e,t){return e.visitAttr(this,t)},e}(),fd=((ad={})[4]=4,ad[1]=1,ad[2]=2,ad[0]=0,ad[3]=3,ad),hd=function(){function e(e,t,n,r,i,o){this.name=e,this.type=t,this.securityContext=n,this.value=r,this.unit=i,this.sourceSpan=o,this.isAnimation=4===this.type}return e.fromBoundProperty=function(t){var n=fd[t.type];return new e(t.name,n,t.securityContext,t.value,t.unit,t.sourceSpan)},e.prototype.visit=function(e,t){return e.visitElementProperty(this,t)},e}(),md=function(){function e(t,n,r,i,o){this.name=t,this.target=n,this.phase=r,this.handler=i,this.sourceSpan=o,this.fullName=e.calcFullName(this.name,this.target,this.phase),this.isAnimation=!!this.phase}return e.calcFullName=function(e,t,n){return t?t+":"+e:n?"@"+e+"."+n:e},e.fromParsedEvent=function(t){var n=0===t.type?t.targetOrPhase:null,r=1===t.type?t.targetOrPhase:null;return new e(t.name,n,r,t.handler,t.sourceSpan)},e.prototype.visit=function(e,t){return e.visitEvent(this,t)},e}(),gd=function(){function e(e,t,n,r){this.name=e,this.value=t,this.originalValue=n,this.sourceSpan=r}return e.prototype.visit=function(e,t){return e.visitReference(this,t)},e}(),vd=function(){function e(e,t,n){this.name=e,this.value=t,this.sourceSpan=n}return e.fromParsedVariable=function(t){return new e(t.name,t.value,t.sourceSpan)},e.prototype.visit=function(e,t){return e.visitVariable(this,t)},e}(),_d=function(){function e(e,t,n,r,i,o,a,s,u,c,l,d,p){this.name=e,this.attrs=t,this.inputs=n,this.outputs=r,this.references=i,this.directives=o,this.providers=a,this.hasViewContainer=s,this.queryMatches=u,this.children=c,this.ngContentIndex=l,this.sourceSpan=d,this.endSourceSpan=p}return e.prototype.visit=function(e,t){return e.visitElement(this,t)},e}(),yd=function(){function e(e,t,n,r,i,o,a,s,u,c,l){this.attrs=e,this.outputs=t,this.references=n,this.variables=r,this.directives=i,this.providers=o,this.hasViewContainer=a,this.queryMatches=s,this.children=u,this.ngContentIndex=c,this.sourceSpan=l}return e.prototype.visit=function(e,t){return e.visitEmbeddedTemplate(this,t)},e}(),bd=function(){function e(e,t,n,r){this.directiveName=e,this.templateName=t,this.value=n,this.sourceSpan=r}return e.prototype.visit=function(e,t){return e.visitDirectiveProperty(this,t)},e}(),jd=function(){function e(e,t,n,r,i,o){this.directive=e,this.inputs=t,this.hostProperties=n,this.hostEvents=r,this.contentQueryStartId=i,this.sourceSpan=o}return e.prototype.visit=function(e,t){return e.visitDirective(this,t)},e}(),xd=function(){function e(e,t,n,r,i,o,a,s){this.token=e,this.multiProvider=t,this.eager=n,this.providers=r,this.providerType=i,this.lifecycleHooks=o,this.sourceSpan=a,this.isModule=s}return e.prototype.visit=function(e,t){return null},e}();(ud=sd||(sd={}))[ud.PublicService=0]="PublicService",ud[ud.PrivateService=1]="PrivateService",ud[ud.Component=2]="Component",ud[ud.Directive=3]="Directive",ud[ud.Builtin=4]="Builtin";var Od=function(){function e(e,t,n){this.index=e,this.ngContentIndex=t,this.sourceSpan=n}return e.prototype.visit=function(e,t){return e.visitNgContent(this,t)},e}(),wd=function(){function e(){}return e.prototype.visitNgContent=function(e,t){},e.prototype.visitEmbeddedTemplate=function(e,t){},e.prototype.visitElement=function(e,t){},e.prototype.visitReference=function(e,t){},e.prototype.visitVariable=function(e,t){},e.prototype.visitEvent=function(e,t){},e.prototype.visitElementProperty=function(e,t){},e.prototype.visitAttr=function(e,t){},e.prototype.visitBoundText=function(e,t){},e.prototype.visitText=function(e,t){},e.prototype.visitDirective=function(e,t){},e.prototype.visitDirectiveProperty=function(e,t){},e}(),Cd=function(e){function t(){return e.call(this)||this}return Object(o.__extends)(t,e),t.prototype.visitEmbeddedTemplate=function(e,t){return this.visitChildren(t,function(t){t(e.attrs),t(e.references),t(e.variables),t(e.directives),t(e.providers),t(e.children)})},t.prototype.visitElement=function(e,t){return this.visitChildren(t,function(t){t(e.attrs),t(e.inputs),t(e.outputs),t(e.references),t(e.directives),t(e.providers),t(e.children)})},t.prototype.visitDirective=function(e,t){return this.visitChildren(t,function(t){t(e.inputs),t(e.hostProperties),t(e.hostEvents)})},t.prototype.visitChildren=function(e,t){var n=[],r=this;return t(function(t){t&&t.length&&n.push(Ed(r,t,e))}),[].concat.apply([],n)},t}(wd);function Ed(e,t,n){void 0===n&&(n=null);var r=[],i=e.visit?function(t){return e.visit(t,n)||t.visit(e,n)}:function(t){return t.visit(e,n)};return t.forEach(function(e){var t=i(e);t&&r.push(t)}),r}var Sd=function(){return function(e){var t=void 0===e?{}:e,n=t.defaultEncapsulation,r=void 0===n?_.Emulated:n,i=t.useJit,o=void 0===i||i,a=t.jitDevMode,s=void 0!==a&&a,u=t.missingTranslation,c=void 0===u?null:u,l=t.preserveWhitespaces,d=t.strictInjectionParameters;this.defaultEncapsulation=r,this.useJit=!!o,this.jitDevMode=!!s,this.missingTranslation=c,this.preserveWhitespaces=kd(zt(l)),this.strictInjectionParameters=!0===d}}();function kd(e,t){return void 0===t&&(t=!1),null===e?t:e}var Md=function(){function e(e,t,n,r){this._resourceLoader=e,this._urlResolver=t,this._htmlParser=n,this._config=r,this._resourceLoaderCache=new Map}return e.prototype.clearCache=function(){this._resourceLoaderCache.clear()},e.prototype.clearCacheFor=function(e){var t=this;if(e.isComponent){var n=e.template;this._resourceLoaderCache.delete(n.templateUrl),n.externalStylesheets.forEach(function(e){t._resourceLoaderCache.delete(e.moduleUrl)})}},e.prototype._fetch=function(e){var t=this._resourceLoaderCache.get(e);return t||(t=this._resourceLoader.get(e),this._resourceLoaderCache.set(e,t)),t},e.prototype.normalizeTemplate=function(e){var t=this;if(Bt(e.template)){if(Bt(e.templateUrl))throw Yt("'"+tn(e.componentType)+"' component cannot define both template and templateUrl");if("string"!=typeof e.template)throw Yt("The template specified for component "+tn(e.componentType)+" is not a string")}else{if(!Bt(e.templateUrl))throw Yt("No template specified for component "+tn(e.componentType));if("string"!=typeof e.templateUrl)throw Yt("The templateUrl specified for component "+tn(e.componentType)+" is not a string")}if(Bt(e.preserveWhitespaces)&&"boolean"!=typeof e.preserveWhitespaces)throw Yt("The preserveWhitespaces option for component "+tn(e.componentType)+" must be a boolean");return qt(this._preParseTemplate(e),function(n){return t._normalizeTemplateMetadata(e,n)})},e.prototype._preParseTemplate=function(e){var t,n,r=this;return null!=e.template?(t=e.template,n=e.moduleUrl):(n=this._urlResolver.resolve(e.moduleUrl,e.templateUrl),t=this._fetch(n)),qt(t,function(t){return r._preparseLoadedTemplate(e,t,n)})},e.prototype._preparseLoadedTemplate=function(e,t,n){var r=!!e.template,i=zi.fromArray(e.interpolation),o=this._htmlParser.parse(t,Gn({reference:e.ngModuleType},{type:{reference:e.componentType}},{isInline:r,templateUrl:n}),!0,i);if(o.errors.length>0)throw Yt("Template parse errors:\n"+o.errors.join("\n"));var a=this._normalizeStylesheet(new Pn({styles:e.styles,moduleUrl:e.moduleUrl})),s=new Ad;_u(s,o.rootNodes);var u=this._normalizeStylesheet(new Pn({styles:s.styles,styleUrls:s.styleUrls,moduleUrl:n}));return{template:t,templateUrl:n,isInline:r,htmlAst:o,styles:a.styles.concat(u.styles),inlineStyleUrls:a.styleUrls.concat(u.styleUrls),styleUrls:this._normalizeStylesheet(new Pn({styleUrls:e.styleUrls,moduleUrl:e.moduleUrl})).styleUrls,ngContentSelectors:s.ngContentSelectors}},e.prototype._normalizeTemplateMetadata=function(e,t){var n=this;return qt(this._loadMissingExternalStylesheets(t.styleUrls.concat(t.inlineStyleUrls)),function(r){return n._normalizeLoadedTemplateMetadata(e,t,r)})},e.prototype._normalizeLoadedTemplateMetadata=function(e,t,n){var r=this,i=Object(o.__spread)(t.styles);this._inlineStyles(t.inlineStyleUrls,n,i);var a=t.styleUrls,s=a.map(function(e){var t=n.get(e),i=Object(o.__spread)(t.styles);return r._inlineStyles(t.styleUrls,n,i),new Pn({moduleUrl:e,styles:i})}),u=e.encapsulation;return null==u&&(u=this._config.defaultEncapsulation),u===_.Emulated&&0===i.length&&0===a.length&&(u=_.None),new Rn({encapsulation:u,template:t.template,templateUrl:t.templateUrl,htmlAst:t.htmlAst,styles:i,styleUrls:a,ngContentSelectors:t.ngContentSelectors,animations:e.animations,interpolation:e.interpolation,isInline:t.isInline,externalStylesheets:s,preserveWhitespaces:kd(e.preserveWhitespaces,this._config.preserveWhitespaces)})},e.prototype._inlineStyles=function(e,t,n){var r=this;e.forEach(function(e){var i=t.get(e);i.styles.forEach(function(e){return n.push(e)}),r._inlineStyles(i.styleUrls,t,n)})},e.prototype._loadMissingExternalStylesheets=function(e,t){var n=this;return void 0===t&&(t=new Map),qt(Gt(e.filter(function(e){return!t.has(e)}).map(function(e){return qt(n._fetch(e),function(r){var i=n._normalizeStylesheet(new Pn({styles:[r],moduleUrl:e}));return t.set(e,i),n._loadMissingExternalStylesheets(i.styleUrls,t)})})),function(e){return t})},e.prototype._normalizeStylesheet=function(e){var t=this,n=e.moduleUrl,r=e.styleUrls.filter(Cc).map(function(e){return t._urlResolver.resolve(n,e)}),i=e.styles.map(function(e){var i,a,s,u,c=(i=t._urlResolver,a=n,s=[],u=e.replace(Mc,"").replace(kc,function(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];var n=e[1]||e[2];return Cc(n)?(s.push(i.resolve(a,n)),""):e[0]}),new wc(u,s));return r.push.apply(r,Object(o.__spread)(c.styleUrls)),c.style});return new Pn({styles:i,styleUrls:r,moduleUrl:n})},e}(),Ad=function(){function e(){this.ngContentSelectors=[],this.styles=[],this.styleUrls=[],this.ngNonBindableStackCount=0}return e.prototype.visitElement=function(e,t){var n=Bc(e);switch(n.type){case Ec.NG_CONTENT:0===this.ngNonBindableStackCount&&this.ngContentSelectors.push(n.selectAttr);break;case Ec.STYLE:var r="";e.children.forEach(function(e){e instanceof pu&&(r+=e.value)}),this.styles.push(r);break;case Ec.STYLESHEET:this.styleUrls.push(n.hrefAttr)}return n.nonBindable&&this.ngNonBindableStackCount++,_u(this,e.children),n.nonBindable&&this.ngNonBindableStackCount--,null},e.prototype.visitExpansion=function(e,t){_u(this,e.cases)},e.prototype.visitExpansionCase=function(e,t){_u(this,e.expression)},e.prototype.visitComment=function(e,t){return null},e.prototype.visitAttribute=function(e,t){return null},e.prototype.visitText=function(e,t){return null},e}(),Td=[I,D,T,A],Dd=function(){function e(e){this._reflector=e}return e.prototype.isDirective=function(e){var t=this._reflector.annotations(nn(e));return t&&t.some(Id)},e.prototype.resolve=function(e,t){void 0===t&&(t=!0);var n=this._reflector.annotations(nn(e));if(n){var r=Pd(n,Id);if(r){var i=this._reflector.propMetadata(e),o=this._reflector.guards(e);return this._mergeWithPropertyMetadata(r,i,o,e)}}if(t)throw new Error("No Directive annotation found on "+tn(e));return null},e.prototype._mergeWithPropertyMetadata=function(e,t,n,r){var i=[],o=[],a={},s={};return Object.keys(t).forEach(function(e){var n=Pd(t[e],function(e){return z.isTypeOf(e)});n&&(n.bindingPropertyName?i.push(e+": "+n.bindingPropertyName):i.push(e));var r=Pd(t[e],function(e){return U.isTypeOf(e)});r&&(r.bindingPropertyName?o.push(e+": "+r.bindingPropertyName):o.push(e)),t[e].filter(function(e){return H.isTypeOf(e)}).forEach(function(t){if(t.hostPropertyName){var n=t.hostPropertyName[0];if("("===n)throw new Error("@HostBinding can not bind to events. Use @HostListener instead.");if("["===n)throw new Error("@HostBinding parameter should be a property name, 'class.<name>', or 'attr.<name>'.");a["["+t.hostPropertyName+"]"]=e}else a["["+e+"]"]=e}),t[e].filter(function(e){return q.isTypeOf(e)}).forEach(function(t){var n=t.args||[];a["("+t.eventName+")"]=e+"("+n.join(",")+")"});var u=Pd(t[e],function(e){return Td.some(function(t){return t.isTypeOf(e)})});u&&(s[e]=u)}),this._merge(e,i,o,a,s,n,r)},e.prototype._extractPublicName=function(e){return Ft(e,[null,e])[1].trim()},e.prototype._dedupeBindings=function(e){for(var t=new Set,n=new Set,r=[],i=e.length-1;i>=0;i--){var o=e[i],a=this._extractPublicName(o);n.add(a),t.has(a)||(t.add(a),r.push(o))}return r.reverse()},e.prototype._merge=function(e,t,n,r,i,a,s){var u=this._dedupeBindings(e.inputs?e.inputs.concat(t):t),c=this._dedupeBindings(e.outputs?e.outputs.concat(n):n),l=e.host?Object(o.__assign)({},e.host,r):r,d=e.queries?Object(o.__assign)({},e.queries,i):i;if(V.isTypeOf(e)){var p=e;return V({selector:p.selector,inputs:u,outputs:c,host:l,exportAs:p.exportAs,moduleId:p.moduleId,queries:d,changeDetection:p.changeDetection,providers:p.providers,viewProviders:p.viewProviders,entryComponents:p.entryComponents,template:p.template,templateUrl:p.templateUrl,styles:p.styles,styleUrls:p.styleUrls,encapsulation:p.encapsulation,animations:p.animations,interpolation:p.interpolation,preserveWhitespaces:e.preserveWhitespaces})}return P({selector:e.selector,inputs:u,outputs:c,host:l,exportAs:e.exportAs,queries:d,providers:e.providers,guards:a})},e}();function Id(e){return P.isTypeOf(e)||V.isTypeOf(e)}function Pd(e,t){for(var n=e.length-1;n>=0;n--)if(t(e[n]))return e[n];return null}var Rd=function(e){function t(t,n){return e.call(this,t,n)||this}return Object(o.__extends)(t,e),t}(os),Nd="i18n",Fd=/^i18n:?/,Ld="|",Vd="@@",Bd=!1;var zd,Ud,Hd=function(){return function(e,t){this.messages=e,this.errors=t}}();(Ud=zd||(zd={}))[Ud.Extract=0]="Extract",Ud[Ud.Merge=1]="Merge";var qd=function(){function e(e,t){this._implicitTags=e,this._implicitAttrs=t}return e.prototype.extract=function(e,t){var n=this;return this._init(zd.Extract,t),e.forEach(function(e){return e.visit(n,null)}),this._inI18nBlock&&this._reportError(e[e.length-1],"Unclosed block"),new Hd(this._messages,this._errors)},e.prototype.merge=function(e,t,n){this._init(zd.Merge,n),this._translations=t;var r=new gu("wrapper",[],e,void 0,void 0,void 0).visit(this,null);return this._inI18nBlock&&this._reportError(e[e.length-1],"Unclosed block"),new Lu(r.children,this._errors)},e.prototype.visitExpansionCase=function(e,t){var n=_u(this,e.expression,t);if(this._mode===zd.Merge)return new hu(e.value,n,e.sourceSpan,e.valueSourceSpan,e.expSourceSpan)},e.prototype.visitExpansion=function(e,t){this._mayBeAddBlockChildren(e);var n=this._inIcu;this._inIcu||(this._isInTranslatableSection&&this._addMessage([e]),this._inIcu=!0);var r=_u(this,e.cases,t);return this._mode===zd.Merge&&(e=new fu(e.switchValue,e.type,r,e.sourceSpan,e.switchValueSourceSpan)),this._inIcu=n,e},e.prototype.visitComment=function(e,t){var n,r=!!((n=e)instanceof vu&&n.value&&n.value.startsWith("i18n"));if(r&&this._isInTranslatableSection)this._reportError(e,"Could not start a block inside a translatable section");else{var i,o=!!((i=e)instanceof vu&&i.value&&"/i18n"===i.value);if(!o||this._inI18nBlock){if(!this._inI18nNode&&!this._inIcu)if(this._inI18nBlock){if(o){if(this._depth==this._blockStartDepth){this._closeTranslatableSection(e,this._blockChildren),this._inI18nBlock=!1;var a=this._addMessage(this._blockChildren,this._blockMeaningAndDesc);return _u(this,this._translateMessage(e,a))}return void this._reportError(e,"I18N blocks should not cross element boundaries")}}else if(r){if(!Bd&&console&&console.warn){Bd=!0;var s=e.sourceSpan.details?", "+e.sourceSpan.details:"";console.warn("I18n comments are deprecated, use an <ng-container> element instead ("+e.sourceSpan.start+s+")")}this._inI18nBlock=!0,this._blockStartDepth=this._depth,this._blockChildren=[],this._blockMeaningAndDesc=e.value.replace(Fd,"").trim(),this._openTranslatableSection(e)}}else this._reportError(e,"Trying to close an unopened block")}},e.prototype.visitText=function(e,t){return this._isInTranslatableSection&&this._mayBeAddBlockChildren(e),e},e.prototype.visitElement=function(e,t){var n=this;this._mayBeAddBlockChildren(e),this._depth++;var r=this._inI18nNode,i=this._inImplicitNode,o=[],a=void 0,s=e.attrs.find(function(e){return e.name===Nd})||null,u=s?s.value:"",c=this._implicitTags.some(function(t){return e.name===t})&&!this._inIcu&&!this._isInTranslatableSection,l=!i&&c;if(this._inImplicitNode=i||c,this._isInTranslatableSection||this._inIcu)(s||l)&&this._reportError(e,"Could not mark an element as translatable inside a translatable section"),this._mode==zd.Extract&&_u(this,e.children);else{if(s||l){this._inI18nNode=!0;var d=this._addMessage(e.children,u);a=this._translateMessage(e,d)}if(this._mode==zd.Extract){var p=s||l;p&&this._openTranslatableSection(e),_u(this,e.children),p&&this._closeTranslatableSection(e,e.children)}}this._mode===zd.Merge&&(a||e.children).forEach(function(e){var r=e.visit(n,t);r&&!n._isInTranslatableSection&&(o=o.concat(r))});if(this._visitAttributesOf(e),this._depth--,this._inI18nNode=r,this._inImplicitNode=i,this._mode===zd.Merge){var f=this._translateAttributes(e);return new gu(e.name,f,o,e.sourceSpan,e.startSourceSpan,e.endSourceSpan)}return null},e.prototype.visitAttribute=function(e,t){throw new Error("unreachable code")},e.prototype._init=function(e,t){this._mode=e,this._inI18nBlock=!1,this._inI18nNode=!1,this._depth=0,this._inIcu=!1,this._msgCountAtSectionStart=void 0,this._errors=[],this._messages=[],this._inImplicitNode=!1,this._createI18nMessage=ol(t)},e.prototype._visitAttributesOf=function(e){var t=this,n={},r=this._implicitAttrs[e.name]||[];e.attrs.filter(function(e){return e.name.startsWith("i18n-")}).forEach(function(e){return n[e.name.slice("i18n-".length)]=e.value}),e.attrs.forEach(function(e){e.name in n?t._addMessage([e],n[e.name]):r.some(function(t){return e.name===t})&&t._addMessage([e])})},e.prototype._addMessage=function(e,t){if(0==e.length||1==e.length&&e[0]instanceof mu&&!e[0].value)return null;var n=Gd(t),r=n.meaning,i=n.description,o=n.id,a=this._createI18nMessage(e,r,i,o);return this._messages.push(a),a},e.prototype._translateMessage=function(e,t){if(t&&this._mode===zd.Merge){var n=this._translations.get(t);if(n)return n;this._reportError(e,'Translation unavailable for message id="'+this._translations.digest(t)+'"')}return[]},e.prototype._translateAttributes=function(e){var t=this,n=e.attrs,r={};n.forEach(function(e){e.name.startsWith("i18n-")&&(r[e.name.slice("i18n-".length)]=Gd(e.value))});var i=[];return n.forEach(function(n){if(n.name!==Nd&&!n.name.startsWith("i18n-"))if(n.value&&""!=n.value&&r.hasOwnProperty(n.name)){var o=r[n.name],a=o.meaning,s=o.description,u=o.id,c=t._createI18nMessage([n],a,s,u),l=t._translations.get(c);if(l)if(0==l.length)i.push(new mu(n.name,"",n.sourceSpan));else if(l[0]instanceof pu){var d=l[0].value;i.push(new mu(n.name,d,n.sourceSpan))}else t._reportError(e,'Unexpected translation for attribute "'+n.name+'" (id="'+(u||t._translations.digest(c))+'")');else t._reportError(e,'Translation unavailable for attribute "'+n.name+'" (id="'+(u||t._translations.digest(c))+'")')}else i.push(n)}),i},e.prototype._mayBeAddBlockChildren=function(e){this._inI18nBlock&&!this._inIcu&&this._depth==this._blockStartDepth&&this._blockChildren.push(e)},e.prototype._openTranslatableSection=function(e){this._isInTranslatableSection?this._reportError(e,"Unexpected section start"):this._msgCountAtSectionStart=this._messages.length},Object.defineProperty(e.prototype,"_isInTranslatableSection",{get:function(){return void 0!==this._msgCountAtSectionStart},enumerable:!0,configurable:!0}),e.prototype._closeTranslatableSection=function(e,t){if(this._isInTranslatableSection){var n=this._msgCountAtSectionStart;if(1==t.reduce(function(e,t){return e+(t instanceof vu?0:1)},0))for(var r=this._messages.length-1;r>=n;r--){var i=this._messages[r].nodes;if(!(1==i.length&&i[0]instanceof $n)){this._messages.splice(r,1);break}}this._msgCountAtSectionStart=void 0}else this._reportError(e,"Unexpected section end")},e.prototype._reportError=function(e,t){this._errors.push(new Rd(e.sourceSpan,t))},e}();function Gd(e){if(!e)return{meaning:"",description:"",id:""};var t=e.indexOf(Vd),n=e.indexOf(Ld),r=Object(o.__read)(t>-1?[e.slice(0,t),e.slice(t+2)]:[e,""],2),i=r[0],a=r[1],s=Object(o.__read)(n>-1?[i.slice(0,n),i.slice(n+1)]:["",i],2);return{meaning:s[0],description:s[1],id:a}}var Wd=new(function(){function e(){this.closedByParent=!1,this.contentType=r.PARSABLE_DATA,this.isVoid=!1,this.ignoreFirstLf=!1,this.canSelfClose=!0}return e.prototype.requireExtraParent=function(e){return!1},e.prototype.isClosedByChild=function(e){return!1},e}());function Yd(e){return Wd}var Kd=function(e){function t(){return e.call(this,Yd)||this}return Object(o.__extends)(t,e),t.prototype.parse=function(t,n,r){return void 0===r&&(r=!1),e.prototype.parse.call(this,t,n,r)},t}(Vu),Qd=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(o.__extends)(t,e),t.prototype.write=function(e,t){var n=new Xd,r=[];e.forEach(function(e){var t,i=[];e.sources.forEach(function(e){var t=new Dr("context-group",{purpose:"location"});t.children.push(new Pr(10),new Dr("context",{"context-type":"sourcefile"},[new Ir(e.filePath)]),new Pr(10),new Dr("context",{"context-type":"linenumber"},[new Ir(""+e.startLine)]),new Pr(8)),i.push(new Pr(8),t)});var a=new Dr("trans-unit",{id:e.id,datatype:"html"});(t=a.children).push.apply(t,Object(o.__spread)([new Pr(8),new Dr("source",{},n.serialize(e.nodes))],i)),e.description&&a.children.push(new Pr(8),new Dr("note",{priority:"1",from:"description"},[new Ir(e.description)])),e.meaning&&a.children.push(new Pr(8),new Dr("note",{priority:"1",from:"meaning"},[new Ir(e.meaning)])),a.children.push(new Pr(6)),r.push(new Pr(6),a)});var i=new Dr("body",{},Object(o.__spread)(r,[new Pr(4)])),a=new Dr("file",{"source-language":t||"en",datatype:"plaintext",original:"ng2.template"},[new Pr(4),i,new Pr(2)]),s=new Dr("xliff",{version:"1.2",xmlns:"urn:oasis:names:tc:xliff:document:1.2"},[new Pr(2),a,new Pr]);return Mr([new Ar({version:"1.0",encoding:"UTF-8"}),new Pr,s,new Pr])},t.prototype.load=function(e,t){var n=(new Zd).parse(e,t),r=n.locale,i=n.msgIdToHtml,a=n.errors,s={},u=new $d;if(Object.keys(i).forEach(function(e){var n=u.convert(i[e],t),r=n.i18nNodes,c=n.errors;a.push.apply(a,Object(o.__spread)(c)),s[e]=r}),a.length)throw new Error("xliff parse errors:\n"+a.join("\n"));return{locale:r,i18nNodesByMsgId:s}},t.prototype.digest=function(e){return ar(e)},t}(Er),Xd=function(){function e(){}return e.prototype.visitText=function(e,t){return[new Ir(e.value)]},e.prototype.visitContainer=function(e,t){var n=this,r=[];return e.children.forEach(function(e){return r.push.apply(r,Object(o.__spread)(e.visit(n)))}),r},e.prototype.visitIcu=function(e,t){var n=this,r=[new Ir("{"+e.expressionPlaceholder+", "+e.type+", ")];return Object.keys(e.cases).forEach(function(t){r.push.apply(r,Object(o.__spread)([new Ir(t+" {")],e.cases[t].visit(n),[new Ir("} ")]))}),r.push(new Ir("}")),r},e.prototype.visitTagPlaceholder=function(e,t){var n=function(e){switch(e.toLowerCase()){case"br":return"lb";case"img":return"image";default:return"x-"+e}}(e.tag);if(e.isVoid)return[new Dr("x",{id:e.startName,ctype:n,"equiv-text":"<"+e.tag+"/>"})];var r=new Dr("x",{id:e.startName,ctype:n,"equiv-text":"<"+e.tag+">"}),i=new Dr("x",{id:e.closeName,ctype:n,"equiv-text":"</"+e.tag+">"});return Object(o.__spread)([r],this.serialize(e.children),[i])},e.prototype.visitPlaceholder=function(e,t){return[new Dr("x",{id:e.name,"equiv-text":"{{"+e.value+"}}"})]},e.prototype.visitIcuPlaceholder=function(e,t){var n="{"+e.value.expression+", "+e.value.type+", "+Object.keys(e.value.cases).map(function(e){return e+" {...}"}).join(" ")+"}";return[new Dr("x",{id:e.name,"equiv-text":n})]},e.prototype.serialize=function(e){var t=this;return[].concat.apply([],Object(o.__spread)(e.map(function(e){return e.visit(t)})))},e}(),Zd=function(){function e(){this._locale=null}return e.prototype.parse=function(e,t){this._unitMlString=null,this._msgIdToHtml={};var n=(new Kd).parse(e,t,!1);return this._errors=n.errors,_u(this,n.rootNodes,null),{msgIdToHtml:this._msgIdToHtml,errors:this._errors,locale:this._locale}},e.prototype.visitElement=function(e,t){switch(e.name){case"trans-unit":this._unitMlString=null;var n=e.attrs.find(function(e){return"id"===e.name});if(n){var r=n.value;this._msgIdToHtml.hasOwnProperty(r)?this._addError(e,"Duplicated translations for msg "+r):(_u(this,e.children,null),"string"==typeof this._unitMlString?this._msgIdToHtml[r]=this._unitMlString:this._addError(e,"Message "+r+" misses a translation"))}else this._addError(e,'<trans-unit> misses the "id" attribute');break;case"source":case"seg-source":break;case"target":var i=e.startSourceSpan.end.offset,o=e.endSourceSpan.start.offset,a=e.startSourceSpan.start.file.content.slice(i,o);this._unitMlString=a;break;case"file":var s=e.attrs.find(function(e){return"target-language"===e.name});s&&(this._locale=s.value),_u(this,e.children,null);break;default:_u(this,e.children,null)}},e.prototype.visitAttribute=function(e,t){},e.prototype.visitText=function(e,t){},e.prototype.visitComment=function(e,t){},e.prototype.visitExpansion=function(e,t){},e.prototype.visitExpansionCase=function(e,t){},e.prototype._addError=function(e,t){this._errors.push(new Rd(e.sourceSpan,t))},e}(),$d=function(){function e(){}return e.prototype.convert=function(e,t){var n=(new Kd).parse(e,t,!0);return this._errors=n.errors,{i18nNodes:this._errors.length>0||0==n.rootNodes.length?[]:[].concat.apply([],Object(o.__spread)(_u(this,n.rootNodes))),errors:this._errors}},e.prototype.visitText=function(e,t){return new $n(e.value,e.sourceSpan)},e.prototype.visitElement=function(e,t){if("x"===e.name){var n=e.attrs.find(function(e){return"id"===e.name});return n?new nr("",n.value,e.sourceSpan):(this._addError(e,'<x> misses the "id" attribute'),null)}return"mrk"===e.name?[].concat.apply([],Object(o.__spread)(_u(this,e.children))):(this._addError(e,"Unexpected tag"),null)},e.prototype.visitExpansion=function(e,t){var n={};return _u(this,e.cases).forEach(function(t){n[t.value]=new Jn(t.nodes,e.sourceSpan)}),new er(e.switchValue,e.type,n,e.sourceSpan)},e.prototype.visitExpansionCase=function(e,t){return{value:e.value,nodes:_u(this,e.expression)}},e.prototype.visitComment=function(e,t){},e.prototype.visitAttribute=function(e,t){},e.prototype._addError=function(e,t){this._errors.push(new Rd(e.sourceSpan,t))},e}();var Jd=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(o.__extends)(t,e),t.prototype.write=function(e,t){var n=new ep,r=[];e.forEach(function(e){var t=new Dr("unit",{id:e.id}),i=new Dr("notes");(e.description||e.meaning)&&(e.description&&i.children.push(new Pr(8),new Dr("note",{category:"description"},[new Ir(e.description)])),e.meaning&&i.children.push(new Pr(8),new Dr("note",{category:"meaning"},[new Ir(e.meaning)]))),e.sources.forEach(function(e){i.children.push(new Pr(8),new Dr("note",{category:"location"},[new Ir(e.filePath+":"+e.startLine+(e.endLine!==e.startLine?","+e.endLine:""))]))}),i.children.push(new Pr(6)),t.children.push(new Pr(6),i);var o=new Dr("segment");o.children.push(new Pr(8),new Dr("source",{},n.serialize(e.nodes)),new Pr(6)),t.children.push(new Pr(6),o,new Pr(4)),r.push(new Pr(4),t)});var i=new Dr("file",{original:"ng.template",id:"ngi18n"},Object(o.__spread)(r,[new Pr(2)])),a=new Dr("xliff",{version:"2.0",xmlns:"urn:oasis:names:tc:xliff:document:2.0",srcLang:t||"en"},[new Pr(2),i,new Pr]);return Mr([new Ar({version:"1.0",encoding:"UTF-8"}),new Pr,a,new Pr])},t.prototype.load=function(e,t){var n=(new tp).parse(e,t),r=n.locale,i=n.msgIdToHtml,a=n.errors,s={},u=new np;if(Object.keys(i).forEach(function(e){var n=u.convert(i[e],t),r=n.i18nNodes,c=n.errors;a.push.apply(a,Object(o.__spread)(c)),s[e]=r}),a.length)throw new Error("xliff2 parse errors:\n"+a.join("\n"));return{locale:r,i18nNodesByMsgId:s}},t.prototype.digest=function(e){return sr(e)},t}(Er),ep=function(){function e(){}return e.prototype.visitText=function(e,t){return[new Ir(e.value)]},e.prototype.visitContainer=function(e,t){var n=this,r=[];return e.children.forEach(function(e){return r.push.apply(r,Object(o.__spread)(e.visit(n)))}),r},e.prototype.visitIcu=function(e,t){var n=this,r=[new Ir("{"+e.expressionPlaceholder+", "+e.type+", ")];return Object.keys(e.cases).forEach(function(t){r.push.apply(r,Object(o.__spread)([new Ir(t+" {")],e.cases[t].visit(n),[new Ir("} ")]))}),r.push(new Ir("}")),r},e.prototype.visitTagPlaceholder=function(e,t){var n=this,r=function(e){switch(e.toLowerCase()){case"br":case"b":case"i":case"u":return"fmt";case"img":return"image";case"a":return"link";default:return"other"}}(e.tag);if(e.isVoid)return[new Dr("ph",{id:(this._nextPlaceholderId++).toString(),equiv:e.startName,type:r,disp:"<"+e.tag+"/>"})];var i=new Dr("pc",{id:(this._nextPlaceholderId++).toString(),equivStart:e.startName,equivEnd:e.closeName,type:r,dispStart:"<"+e.tag+">",dispEnd:"</"+e.tag+">"}),a=[].concat.apply([],Object(o.__spread)(e.children.map(function(e){return e.visit(n)})));return a.length?a.forEach(function(e){return i.children.push(e)}):i.children.push(new Ir("")),[i]},e.prototype.visitPlaceholder=function(e,t){var n=(this._nextPlaceholderId++).toString();return[new Dr("ph",{id:n,equiv:e.name,disp:"{{"+e.value+"}}"})]},e.prototype.visitIcuPlaceholder=function(e,t){var n=Object.keys(e.value.cases).map(function(e){return e+" {...}"}).join(" "),r=(this._nextPlaceholderId++).toString();return[new Dr("ph",{id:r,equiv:e.name,disp:"{"+e.value.expression+", "+e.value.type+", "+n+"}"})]},e.prototype.serialize=function(e){var t=this;return this._nextPlaceholderId=0,[].concat.apply([],Object(o.__spread)(e.map(function(e){return e.visit(t)})))},e}(),tp=function(){function e(){this._locale=null}return e.prototype.parse=function(e,t){this._unitMlString=null,this._msgIdToHtml={};var n=(new Kd).parse(e,t,!1);return this._errors=n.errors,_u(this,n.rootNodes,null),{msgIdToHtml:this._msgIdToHtml,errors:this._errors,locale:this._locale}},e.prototype.visitElement=function(e,t){switch(e.name){case"unit":this._unitMlString=null;var n=e.attrs.find(function(e){return"id"===e.name});if(n){var r=n.value;this._msgIdToHtml.hasOwnProperty(r)?this._addError(e,"Duplicated translations for msg "+r):(_u(this,e.children,null),"string"==typeof this._unitMlString?this._msgIdToHtml[r]=this._unitMlString:this._addError(e,"Message "+r+" misses a translation"))}else this._addError(e,'<unit> misses the "id" attribute');break;case"source":break;case"target":var i=e.startSourceSpan.end.offset,o=e.endSourceSpan.start.offset,a=e.startSourceSpan.start.file.content.slice(i,o);this._unitMlString=a;break;case"xliff":var s=e.attrs.find(function(e){return"trgLang"===e.name});s&&(this._locale=s.value);var u=e.attrs.find(function(e){return"version"===e.name});if(u){var c=u.value;"2.0"!==c?this._addError(e,"The XLIFF file version "+c+" is not compatible with XLIFF 2.0 serializer"):_u(this,e.children,null)}break;default:_u(this,e.children,null)}},e.prototype.visitAttribute=function(e,t){},e.prototype.visitText=function(e,t){},e.prototype.visitComment=function(e,t){},e.prototype.visitExpansion=function(e,t){},e.prototype.visitExpansionCase=function(e,t){},e.prototype._addError=function(e,t){this._errors.push(new Rd(e.sourceSpan,t))},e}(),np=function(){function e(){}return e.prototype.convert=function(e,t){var n=(new Kd).parse(e,t,!0);return this._errors=n.errors,{i18nNodes:this._errors.length>0||0==n.rootNodes.length?[]:[].concat.apply([],Object(o.__spread)(_u(this,n.rootNodes))),errors:this._errors}},e.prototype.visitText=function(e,t){return new $n(e.value,e.sourceSpan)},e.prototype.visitElement=function(e,t){var n=this;switch(e.name){case"ph":var r=e.attrs.find(function(e){return"equiv"===e.name});if(r)return[new nr("",r.value,e.sourceSpan)];this._addError(e,'<ph> misses the "equiv" attribute');break;case"pc":var i=e.attrs.find(function(e){return"equivStart"===e.name}),a=e.attrs.find(function(e){return"equivEnd"===e.name});if(i){if(a){var s=i.value,u=a.value,c=[];return c.concat.apply(c,Object(o.__spread)([new nr("",s,e.sourceSpan)],e.children.map(function(e){return e.visit(n,null)}),[new nr("",u,e.sourceSpan)]))}this._addError(e,'<ph> misses the "equivEnd" attribute')}else this._addError(e,'<ph> misses the "equivStart" attribute');break;case"mrk":return[].concat.apply([],Object(o.__spread)(_u(this,e.children)));default:this._addError(e,"Unexpected tag")}return null},e.prototype.visitExpansion=function(e,t){var n={};return _u(this,e.cases).forEach(function(t){n[t.value]=new Jn(t.nodes,e.sourceSpan)}),new er(e.switchValue,e.type,n,e.sourceSpan)},e.prototype.visitExpansionCase=function(e,t){return{value:e.value,nodes:[].concat.apply([],Object(o.__spread)(_u(this,e.expression)))}},e.prototype.visitComment=function(e,t){},e.prototype.visitAttribute=function(e,t){},e.prototype._addError=function(e,t){this._errors.push(new Rd(e.sourceSpan,t))},e}();var rp=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(o.__extends)(t,e),t.prototype.write=function(e,t){throw new Error("Unsupported")},t.prototype.load=function(e,t){var n=(new ip).parse(e,t),r=n.locale,i=n.msgIdToHtml,o=n.errors,a={},s=new op;if(Object.keys(i).forEach(function(e){var n,r,o;n=a,r=e,o=function(){var n=s.convert(i[e],t),r=n.i18nNodes,o=n.errors;if(o.length)throw new Error("xtb parse errors:\n"+o.join("\n"));return r},Object.defineProperty(n,r,{configurable:!0,enumerable:!0,get:function(){var e=o();return Object.defineProperty(n,r,{enumerable:!0,value:e}),e},set:function(e){throw new Error("Could not overwrite an XTB translation")}})}),o.length)throw new Error("xtb parse errors:\n"+o.join("\n"));return{locale:r,i18nNodesByMsgId:a}},t.prototype.digest=function(e){return Vr(e)},t.prototype.createNameMapper=function(e){return new Sr(e,zr)},t}(Er);var ip=function(){function e(){this._locale=null}return e.prototype.parse=function(e,t){this._bundleDepth=0,this._msgIdToHtml={};var n=(new Kd).parse(e,t,!1);return this._errors=n.errors,_u(this,n.rootNodes),{msgIdToHtml:this._msgIdToHtml,errors:this._errors,locale:this._locale}},e.prototype.visitElement=function(e,t){switch(e.name){case"translationbundle":this._bundleDepth++,this._bundleDepth>1&&this._addError(e,"<translationbundle> elements can not be nested");var n=e.attrs.find(function(e){return"lang"===e.name});n&&(this._locale=n.value),_u(this,e.children,null),this._bundleDepth--;break;case"translation":var r=e.attrs.find(function(e){return"id"===e.name});if(r){var i=r.value;if(this._msgIdToHtml.hasOwnProperty(i))this._addError(e,"Duplicated translations for msg "+i);else{var o=e.startSourceSpan.end.offset,a=e.endSourceSpan.start.offset,s=e.startSourceSpan.start.file.content.slice(o,a);this._msgIdToHtml[i]=s}}else this._addError(e,'<translation> misses the "id" attribute');break;default:this._addError(e,"Unexpected tag")}},e.prototype.visitAttribute=function(e,t){},e.prototype.visitText=function(e,t){},e.prototype.visitComment=function(e,t){},e.prototype.visitExpansion=function(e,t){},e.prototype.visitExpansionCase=function(e,t){},e.prototype._addError=function(e,t){this._errors.push(new Rd(e.sourceSpan,t))},e}(),op=function(){function e(){}return e.prototype.convert=function(e,t){var n=(new Kd).parse(e,t,!0);return this._errors=n.errors,{i18nNodes:this._errors.length>0||0==n.rootNodes.length?[]:_u(this,n.rootNodes),errors:this._errors}},e.prototype.visitText=function(e,t){return new $n(e.value,e.sourceSpan)},e.prototype.visitExpansion=function(e,t){var n={};return _u(this,e.cases).forEach(function(t){n[t.value]=new Jn(t.nodes,e.sourceSpan)}),new er(e.switchValue,e.type,n,e.sourceSpan)},e.prototype.visitExpansionCase=function(e,t){return{value:e.value,nodes:_u(this,e.expression)}},e.prototype.visitElement=function(e,t){if("ph"===e.name){var n=e.attrs.find(function(e){return"name"===e.name});if(n)return new nr("",n.value,e.sourceSpan);this._addError(e,'<ph> misses the "name" attribute')}else this._addError(e,"Unexpected tag");return null},e.prototype.visitComment=function(e,t){},e.prototype.visitAttribute=function(e,t){},e.prototype._addError=function(e,t){this._errors.push(new Rd(e.sourceSpan,t))},e}(),ap=function(){function e(e,t,n,r,i,o){void 0===e&&(e={}),void 0===i&&(i=F.Warning),this._i18nNodesByMsgId=e,this.digest=n,this.mapperFactory=r,this._i18nToHtml=new sp(e,t,n,r,i,o)}return e.load=function(t,n,r,i,o){var a=r.load(t,n),s=a.locale;return new e(a.i18nNodesByMsgId,s,function(e){return r.digest(e)},function(e){return r.createNameMapper(e)},i,o)},e.prototype.get=function(e){var t=this._i18nToHtml.convert(e);if(t.errors.length)throw new Error(t.errors.join("\n"));return t.nodes},e.prototype.has=function(e){return this.digest(e)in this._i18nNodesByMsgId},e}(),sp=function(){function e(e,t,n,r,i,o){void 0===e&&(e={}),this._i18nNodesByMsgId=e,this._locale=t,this._digest=n,this._mapperFactory=r,this._missingTranslationStrategy=i,this._console=o,this._contextStack=[],this._errors=[]}return e.prototype.convert=function(e){this._contextStack.length=0,this._errors.length=0;var t=this._convertToText(e),n=e.nodes[0].sourceSpan.start.file.url,r=(new Uu).parse(t,n,!0);return{nodes:r.rootNodes,errors:Object(o.__spread)(this._errors,r.errors)}},e.prototype.visitText=function(e,t){return Nr(e.value)},e.prototype.visitContainer=function(e,t){var n=this;return e.children.map(function(e){return e.visit(n)}).join("")},e.prototype.visitIcu=function(e,t){var n=this,r=Object.keys(e.cases).map(function(t){return t+" {"+e.cases[t].visit(n)+"}"});return"{"+(this._srcMsg.placeholders.hasOwnProperty(e.expression)?this._srcMsg.placeholders[e.expression]:e.expression)+", "+e.type+", "+r.join(" ")+"}"},e.prototype.visitPlaceholder=function(e,t){var n=this._mapper(e.name);return this._srcMsg.placeholders.hasOwnProperty(n)?this._srcMsg.placeholders[n]:this._srcMsg.placeholderToMessage.hasOwnProperty(n)?this._convertToText(this._srcMsg.placeholderToMessage[n]):(this._addError(e,'Unknown placeholder "'+e.name+'"'),"")},e.prototype.visitTagPlaceholder=function(e,t){var n=this,r=""+e.tag,i=Object.keys(e.attrs).map(function(t){return t+'="'+e.attrs[t]+'"'}).join(" ");return e.isVoid?"<"+r+" "+i+"/>":"<"+r+" "+i+">"+e.children.map(function(e){return e.visit(n)}).join("")+"</"+r+">"},e.prototype.visitIcuPlaceholder=function(e,t){return this._convertToText(this._srcMsg.placeholderToMessage[e.name])},e.prototype._convertToText=function(e){var t,n=this,r=this._digest(e),i=this._mapperFactory?this._mapperFactory(e):null;if(this._contextStack.push({msg:this._srcMsg,mapper:this._mapper}),this._srcMsg=e,this._i18nNodesByMsgId.hasOwnProperty(r))t=this._i18nNodesByMsgId[r],this._mapper=function(e){return i?i.toInternalName(e):e};else{if(this._missingTranslationStrategy===F.Error){var o=this._locale?' for locale "'+this._locale+'"':"";this._addError(e.nodes[0],'Missing translation for message "'+r+'"'+o)}else if(this._console&&this._missingTranslationStrategy===F.Warning){o=this._locale?' for locale "'+this._locale+'"':"";this._console.warn('Missing translation for message "'+r+'"'+o)}t=e.nodes,this._mapper=function(e){return e}}var a=t.map(function(e){return e.visit(n)}).join(""),s=this._contextStack.pop();return this._srcMsg=s.msg,this._mapper=s.mapper,a},e.prototype._addError=function(e,t){this._errors.push(new Rd(e.sourceSpan,t))},e}(),up=function(){function e(e,t,n,r,i){if(void 0===r&&(r=F.Warning),this._htmlParser=e,t){var o=function(e){switch(e=(e||"xlf").toLowerCase()){case"xmb":return new Fr;case"xtb":return new rp;case"xliff2":case"xlf2":return new Jd;case"xliff":case"xlf":default:return new Qd}}(n);this._translationBundle=ap.load(t,"i18n",o,r,i)}else this._translationBundle=new ap({},null,ar,void 0,r,i)}return e.prototype.parse=function(e,t,n,r){void 0===n&&(n=!1),void 0===r&&(r=Ui);var i,o,a,s=this._htmlParser.parse(e,t,n,r);return s.errors.length?new Lu(s.rootNodes,s.errors):(i=s.rootNodes,o=this._translationBundle,a=r,new qd([],{}).merge(i,o,a))},e}();function cp(e,t,n){return void 0===n&&(n=null),Vt(t,new lp(e),n)}var lp=function(){function e(e){this.ctx=e}return e.prototype.visitArray=function(e,t){var n=this;return At(e.map(function(e){return Vt(e,n,null)}),t)},e.prototype.visitStringMap=function(e,t){var n=this,r=[],i=new Set(e&&e.$quoted$);return Object.keys(e).forEach(function(t){r.push(new Ze(t,Vt(e[t],n,null),i.has(t)))}),new $e(r,t)},e.prototype.visitPrimitive=function(e,t){return Pt(e,t)},e.prototype.visitOther=function(e,t){return e instanceof we?e:this.ctx.importExpr(e)},e}();function dp(e,t){return{key:e,value:t,quoted:!1}}var pp=function(){function e(e,t){this.reflector=e,this.alwaysGenerateDef=t,this.tokenInjector=e.resolveExternalReference(gn.Injector)}return e.prototype.depsArray=function(e,t){var n=this;return e.map(function(e){var r,i=e,o=[i],a=0;if(Array.isArray(e))for(var s=0;s<e.length;s++){var u=e[s];u&&("Optional"===u.ngMetadataName?a|=8:"SkipSelf"===u.ngMetadataName?a|=4:"Self"===u.ngMetadataName?a|=2:i="Inject"===u.ngMetadataName?u.token:u)}return r="string"==typeof i?Pt(i):i===n.tokenInjector?St(gn.INJECTOR):t.importExpr(i),o=0!==a?[r,Pt(a)]:[r],St(gn.inject).callFn(o)})},e.prototype.factoryFor=function(e,t){var n;if(e.useExisting)n=St(gn.inject).callFn([t.importExpr(e.useExisting)]);else if(e.useFactory){var r=e.deps||[];if(!(r.length>0))return t.importExpr(e.useFactory);n=t.importExpr(e.useFactory).callFn(this.depsArray(r,t))}else if(e.useValue)n=cp(t,e.useValue);else{var i=e.useClass||e.symbol,o=this.depsArray(this.reflector.parameters(i),t);n=new Fe(t.importExpr(i),o)}return Dt([],[new at(n)],void 0,void 0,e.symbol.name+"_Factory")},e.prototype.injectableDef=function(e,t){var n=et;void 0!==e.providedIn&&(n=null===e.providedIn?et:"string"==typeof e.providedIn?Pt(e.providedIn):t.importExpr(e.providedIn));var r=[dp("factory",this.factoryFor(e,t)),dp("token",t.importExpr(e.type.reference)),dp("providedIn",n)];return St(gn.defineInjectable).callFn([Tt(r)])},e.prototype.compile=function(e,t){if(this.alwaysGenerateDef||void 0!==e.providedIn){var n=En(e.type),r=new dt(n,null,[new ut("ngInjectableDef",ge,[Ie.Static],this.injectableDef(e,t))],[],new ct(null,[],[]),[]);t.statements.push(r)}},e}(),fp=/(\.ts|\.d\.ts|\.js|\.jsx|\.tsx)$/,hp=/\.ngfactory\.|\.ngsummary\./,mp=/\.ngsummary\./,gp=/NgSummary$/;function vp(e,t){void 0===t&&(t=!1);var n=yp(e,t);return n[0]+".ngfactory"+bp(n[1])}function _p(e){return e.replace(hp,".")}function yp(e,t){if(void 0===t&&(t=!1),e.endsWith(".d.ts"))return[e.slice(0,-5),t?".ts":".d.ts"];var n=e.lastIndexOf(".");return-1!==n?[e.substring(0,n),e.substring(n)]:[e,""]}function bp(e){return".tsx"===e?".ts":e}function jp(e){return e.replace(fp,"")+".ngsummary.json"}function xp(e,t){void 0===t&&(t=!1);var n=yp(_p(e),t);return n[0]+".ngsummary"+n[1]}function Op(e){return e+"NgSummary"}var wp=/\u0275\d+/;function Cp(e){return wp.test(e)}function Ep(e){return"ɵ"+e}var Sp="ngComponentType",kp=function(){function e(e,t,n,r,i,o,a,s,u,c,l,d){this._config=e,this._htmlParser=t,this._ngModuleResolver=n,this._directiveResolver=r,this._pipeResolver=i,this._summaryResolver=o,this._schemaRegistry=a,this._directiveNormalizer=s,this._console=u,this._staticSymbolCache=c,this._reflector=l,this._errorCollector=d,this._nonNormalizedDirectiveCache=new Map,this._directiveCache=new Map,this._summaryCache=new Map,this._pipeCache=new Map,this._ngModuleCache=new Map,this._ngModuleOfTypes=new Map,this._shallowModuleCache=new Map}return e.prototype.getReflector=function(){return this._reflector},e.prototype.clearCacheFor=function(e){var t=this._directiveCache.get(e);this._directiveCache.delete(e),this._nonNormalizedDirectiveCache.delete(e),this._summaryCache.delete(e),this._pipeCache.delete(e),this._ngModuleOfTypes.delete(e),this._ngModuleCache.clear(),t&&this._directiveNormalizer.clearCacheFor(t)},e.prototype.clearCache=function(){this._directiveCache.clear(),this._nonNormalizedDirectiveCache.clear(),this._summaryCache.clear(),this._pipeCache.clear(),this._ngModuleCache.clear(),this._ngModuleOfTypes.clear(),this._directiveNormalizer.clearCache()},e.prototype._createProxyClass=function(e,t){var n=null,r=function(){if(!n)throw new Error("Illegal state: Class "+t+" for type "+tn(e)+" is not compiled yet!");return n.apply(this,arguments)};return r.setDelegate=function(e){n=e,r.prototype=e.prototype},r.overriddenName=t,r},e.prototype.getGeneratedClass=function(e,t){return e instanceof yn?this._staticSymbolCache.get(vp(e.filePath),t):this._createProxyClass(e,t)},e.prototype.getComponentViewClass=function(e){return this.getGeneratedClass(e,kn(e,0))},e.prototype.getHostComponentViewClass=function(e){return this.getGeneratedClass(e,An(e))},e.prototype.getHostComponentType=function(e){var t=En({reference:e})+"_Host";return e instanceof yn?this._staticSymbolCache.get(e.filePath,t):this._createProxyClass(e,t)},e.prototype.getRendererType=function(e){return e instanceof yn?this._staticSymbolCache.get(vp(e.filePath),Mn(e)):{}},e.prototype.getComponentFactory=function(e,t,n,r){if(t instanceof yn)return this._staticSymbolCache.get(vp(t.filePath),Tn(t));var i=this.getHostComponentViewClass(t);return this._reflector.resolveExternalReference(gn.createComponentFactory)(e,t,i,n,r,[])},e.prototype.initComponentFactory=function(e,t){var n;e instanceof yn||(n=e.ngContentSelectors).push.apply(n,Object(o.__spread)(t))},e.prototype._loadSummary=function(e,t){var n=this._summaryCache.get(e);if(!n){var r=this._summaryResolver.resolveSummary(e);n=r?r.type:null,this._summaryCache.set(e,n||null)}return n&&n.summaryKind===t?n:null},e.prototype.getHostComponentMetadata=function(e,t){var n=this.getHostComponentType(e.type.reference);t||(t=this.getHostComponentViewClass(n));var r=O.parse(e.selector)[0].getMatchingElementTemplate(),i=this._htmlParser.parse(r,"");return Nn.create({isHost:!0,type:{reference:n,diDeps:[],lifecycleHooks:[]},template:new Rn({encapsulation:_.None,template:r,templateUrl:"",htmlAst:i,styles:[],styleUrls:[],ngContentSelectors:[],animations:[],isInline:!0,externalStylesheets:[],interpolation:null,preserveWhitespaces:!1}),exportAs:null,changeDetection:b.Default,inputs:[],outputs:[],host:{},isComponent:!0,selector:"*",providers:[],viewProviders:[],queries:[],guards:{},viewQueries:[],componentViewType:t,rendererType:{id:"__Host__",encapsulation:_.None,styles:[],data:{}},entryComponents:[],componentFactory:null})},e.prototype.loadDirectiveMetadata=function(e,t,n){var r=this;if(this._directiveCache.has(t))return null;t=nn(t);var i,o,a=this.getNonNormalizedDirectiveMetadata(t),s=a.annotation,u=a.metadata,c=function(e){var n=new Nn({isHost:!1,type:u.type,isComponent:u.isComponent,selector:u.selector,exportAs:u.exportAs,changeDetection:u.changeDetection,inputs:u.inputs,outputs:u.outputs,hostListeners:u.hostListeners,hostProperties:u.hostProperties,hostAttributes:u.hostAttributes,providers:u.providers,viewProviders:u.viewProviders,queries:u.queries,guards:u.guards,viewQueries:u.viewQueries,entryComponents:u.entryComponents,componentViewType:u.componentViewType,rendererType:u.rendererType,componentFactory:u.componentFactory,template:e});return e&&r.initComponentFactory(u.componentFactory,e.ngContentSelectors),r._directiveCache.set(t,n),r._summaryCache.set(t,n.toSummary()),null};if(u.isComponent){var l=u.template,d=this._directiveNormalizer.normalizeTemplate({ngModuleType:e,componentType:t,moduleUrl:this._reflector.componentModuleUrl(t,s),encapsulation:l.encapsulation,template:l.template,templateUrl:l.templateUrl,styles:l.styles,styleUrls:l.styleUrls,animations:l.animations,interpolation:l.interpolation,preserveWhitespaces:l.preserveWhitespaces});return rn(d)&&n?(this._reportError((i=t,(o=Error("Can't compile synchronously as "+tn(i)+" is still being loaded!"))[Sp]=i,o),t),null):qt(d,c)}return c(null),null},e.prototype.getNonNormalizedDirectiveMetadata=function(e){var t=this;if(!(e=nn(e)))return null;var n=this._nonNormalizedDirectiveCache.get(e);if(n)return n;var r=this._directiveResolver.resolve(e,!1);if(!r)return null;var i=void 0;if(V.isTypeOf(r)){Li("styles",(a=r).styles),Li("styleUrls",a.styleUrls),Bi("interpolation",a.interpolation);var o=a.animations;i=new Rn({encapsulation:zt(a.encapsulation),template:zt(a.template),templateUrl:zt(a.templateUrl),htmlAst:null,styles:a.styles||[],styleUrls:a.styleUrls||[],animations:o||[],interpolation:zt(a.interpolation),isInline:!!a.template,externalStylesheets:[],ngContentSelectors:[],preserveWhitespaces:zt(r.preserveWhitespaces)})}var a,s=null,u=[],c=[],l=r.selector;V.isTypeOf(r)?(s=(a=r).changeDetection,a.viewProviders&&(u=this._getProvidersMetadata(a.viewProviders,c,'viewProviders for "'+Dp(e)+'"',[],e)),a.entryComponents&&(c=Mp(a.entryComponents).map(function(e){return t._getEntryComponentMetadata(e)}).concat(c)),l||(l=this._schemaRegistry.getDefaultComponentElementName())):l||(this._reportError(Yt("Directive "+Dp(e)+" has no selector, please add it!"),e),l="error");var d=[];null!=r.providers&&(d=this._getProvidersMetadata(r.providers,c,'providers for "'+Dp(e)+'"',[],e));var p=[],f=[];null!=r.queries&&(p=this._getQueriesMetadata(r.queries,!1,e),f=this._getQueriesMetadata(r.queries,!0,e));var h=Nn.create({isHost:!1,selector:l,exportAs:zt(r.exportAs),isComponent:!!i,type:this._getTypeMetadata(e),template:i,changeDetection:s,inputs:r.inputs||[],outputs:r.outputs||[],host:r.host||{},providers:d||[],viewProviders:u||[],queries:p||[],guards:r.guards||{},viewQueries:f||[],entryComponents:c,componentViewType:i?this.getComponentViewClass(e):null,rendererType:i?this.getRendererType(e):null,componentFactory:null});return i&&(h.componentFactory=this.getComponentFactory(l,e,h.inputs,h.outputs)),n={metadata:h,annotation:r},this._nonNormalizedDirectiveCache.set(e,n),n},e.prototype.getDirectiveMetadata=function(e){var t=this._directiveCache.get(e);return t||this._reportError(Yt("Illegal state: getDirectiveMetadata can only be called after loadNgModuleDirectiveAndPipeMetadata for a module that declares it. Directive "+Dp(e)+"."),e),t},e.prototype.getDirectiveSummary=function(e){var t=this._loadSummary(e,On.Directive);return t||this._reportError(Yt("Illegal state: Could not load the summary for directive "+Dp(e)+"."),e),t},e.prototype.isDirective=function(e){return!!this._loadSummary(e,On.Directive)||this._directiveResolver.isDirective(e)},e.prototype.isPipe=function(e){return!!this._loadSummary(e,On.Pipe)||this._pipeResolver.isPipe(e)},e.prototype.isNgModule=function(e){return!!this._loadSummary(e,On.NgModule)||this._ngModuleResolver.isNgModule(e)},e.prototype.getNgModuleSummary=function(e,t){void 0===t&&(t=null);var n=this._loadSummary(e,On.NgModule);if(!n){var r=this.getNgModuleMetadata(e,!1,t);(n=r?r.toSummary():null)&&this._summaryCache.set(e,n)}return n},e.prototype.loadNgModuleDirectiveAndPipeMetadata=function(e,t,n){var r=this;void 0===n&&(n=!0);var i=this.getNgModuleMetadata(e,n),o=[];return i&&(i.declaredDirectives.forEach(function(n){var i=r.loadDirectiveMetadata(e,n.reference,t);i&&o.push(i)}),i.declaredPipes.forEach(function(e){return r._loadPipeMetadata(e.reference)})),Promise.all(o)},e.prototype.getShallowModuleMetadata=function(e){var t=this._shallowModuleCache.get(e);if(t)return t;var n=Pd(this._reflector.shallowAnnotations(e),G.isTypeOf);return t={type:this._getTypeMetadata(e),rawExports:n.exports,rawImports:n.imports,rawProviders:n.providers},this._shallowModuleCache.set(e,t),t},e.prototype.getNgModuleMetadata=function(e,t,n){var r=this;void 0===t&&(t=!0),void 0===n&&(n=null),e=nn(e);var i=this._ngModuleCache.get(e);if(i)return i;var a=this._ngModuleResolver.resolve(e,t);if(!a)return null;var s=[],u=[],c=[],l=[],d=[],p=[],f=[],h=[],m=[];a.imports&&Mp(a.imports).forEach(function(t){var i=void 0;if(Ap(t))i=t;else if(t&&t.ngModule){var a=t;i=a.ngModule,a.providers&&p.push.apply(p,Object(o.__spread)(r._getProvidersMetadata(a.providers,f,"provider for the NgModule '"+Dp(i)+"'",[],t)))}if(i){if(!r._checkSelfImport(e,i))if(n||(n=new Set),n.has(i))r._reportError(Yt(r._getTypeDescriptor(i)+" '"+Dp(t)+"' is imported recursively by the module '"+Dp(e)+"'."),e);else{n.add(i);var s=r.getNgModuleSummary(i,n);n.delete(i),s?l.push(s):r._reportError(Yt("Unexpected "+r._getTypeDescriptor(t)+" '"+Dp(t)+"' imported by the module '"+Dp(e)+"'. Please add a @NgModule annotation."),e)}}else r._reportError(Yt("Unexpected value '"+Dp(t)+"' imported by the module '"+Dp(e)+"'"),e)}),a.exports&&Mp(a.exports).forEach(function(t){if(Ap(t))if(n||(n=new Set),n.has(t))r._reportError(Yt(r._getTypeDescriptor(t)+" '"+tn(t)+"' is exported recursively by the module '"+Dp(e)+"'"),e);else{n.add(t);var i=r.getNgModuleSummary(t,n);n.delete(t),i?d.push(i):u.push(r._getIdentifierMetadata(t))}else r._reportError(Yt("Unexpected value '"+Dp(t)+"' exported by the module '"+Dp(e)+"'"),e)});var g=this._getTransitiveNgModuleMetadata(l,d);a.declarations&&Mp(a.declarations).forEach(function(t){if(Ap(t)){var n=r._getIdentifierMetadata(t);if(r.isDirective(t))g.addDirective(n),s.push(n),r._addTypeToModule(t,e);else{if(!r.isPipe(t))return void r._reportError(Yt("Unexpected "+r._getTypeDescriptor(t)+" '"+Dp(t)+"' declared by the module '"+Dp(e)+"'. Please add a @Pipe/@Directive/@Component annotation."),e);g.addPipe(n),g.pipes.push(n),c.push(n),r._addTypeToModule(t,e)}}else r._reportError(Yt("Unexpected value '"+Dp(t)+"' declared by the module '"+Dp(e)+"'"),e)});var v=[],_=[];return u.forEach(function(t){if(g.directivesSet.has(t.reference))v.push(t),g.addExportedDirective(t);else{if(!g.pipesSet.has(t.reference))return void r._reportError(Yt("Can't export "+r._getTypeDescriptor(t.reference)+" "+Dp(t.reference)+" from "+Dp(e)+" as it was neither declared nor imported!"),e);_.push(t),g.addExportedPipe(t)}}),a.providers&&p.push.apply(p,Object(o.__spread)(this._getProvidersMetadata(a.providers,f,"provider for the NgModule '"+Dp(e)+"'",[],e))),a.entryComponents&&f.push.apply(f,Object(o.__spread)(Mp(a.entryComponents).map(function(e){return r._getEntryComponentMetadata(e)}))),a.bootstrap&&Mp(a.bootstrap).forEach(function(t){Ap(t)?h.push(r._getIdentifierMetadata(t)):r._reportError(Yt("Unexpected value '"+Dp(t)+"' used in the bootstrap property of module '"+Dp(e)+"'"),e)}),f.push.apply(f,Object(o.__spread)(h.map(function(e){return r._getEntryComponentMetadata(e.reference)}))),a.schemas&&m.push.apply(m,Object(o.__spread)(Mp(a.schemas))),i=new Vn({type:this._getTypeMetadata(e),providers:p,entryComponents:f,bootstrapComponents:h,schemas:m,declaredDirectives:s,exportedDirectives:v,declaredPipes:c,exportedPipes:_,importedModules:l,exportedModules:d,transitiveModule:g,id:a.id||null}),f.forEach(function(e){return g.addEntryComponent(e)}),p.forEach(function(e){return g.addProvider(e,i.type)}),g.addModule(i.type),this._ngModuleCache.set(e,i),i},e.prototype._checkSelfImport=function(e,t){return e===t&&(this._reportError(Yt("'"+Dp(e)+"' module can't import itself"),e),!0)},e.prototype._getTypeDescriptor=function(e){if(Ap(e)){if(this.isDirective(e))return"directive";if(this.isPipe(e))return"pipe";if(this.isNgModule(e))return"module"}return e.provide?"provider":"value"},e.prototype._addTypeToModule=function(e,t){var n=this._ngModuleOfTypes.get(e);n&&n!==t?this._reportError(Yt("Type "+Dp(e)+" is part of the declarations of 2 modules: "+Dp(n)+" and "+Dp(t)+"! Please consider moving "+Dp(e)+" to a higher module that imports "+Dp(n)+" and "+Dp(t)+". You can also create a new NgModule that exports and includes "+Dp(e)+" then import that NgModule in "+Dp(n)+" and "+Dp(t)+"."),t):this._ngModuleOfTypes.set(e,t)},e.prototype._getTransitiveNgModuleMetadata=function(e,t){var n=new Bn,r=new Map;return e.concat(t).forEach(function(e){e.modules.forEach(function(e){return n.addModule(e)}),e.entryComponents.forEach(function(e){return n.addEntryComponent(e)});var t=new Set;e.providers.forEach(function(e){var i=In(e.provider.token),o=r.get(i);o||(o=new Set,r.set(i,o));var a=e.module.reference;!t.has(i)&&o.has(a)||(o.add(a),t.add(i),n.addProvider(e.provider,e.module))})}),t.forEach(function(e){e.exportedDirectives.forEach(function(e){return n.addExportedDirective(e)}),e.exportedPipes.forEach(function(e){return n.addExportedPipe(e)})}),e.forEach(function(e){e.exportedDirectives.forEach(function(e){return n.addDirective(e)}),e.exportedPipes.forEach(function(e){return n.addPipe(e)})}),n},e.prototype._getIdentifierMetadata=function(e){return{reference:e=nn(e)}},e.prototype.isInjectable=function(e){return this._reflector.tryAnnotations(e).some(function(e){return W.isTypeOf(e)})},e.prototype.getInjectableSummary=function(e){return{summaryKind:On.Injectable,type:this._getTypeMetadata(e,null,!1)}},e.prototype.getInjectableMetadata=function(e,t,n){void 0===t&&(t=null),void 0===n&&(n=!0);var r=this._loadSummary(e,On.Injectable),i=r?r.type:this._getTypeMetadata(e,t,n),o=this._reflector.annotations(e).filter(function(e){return W.isTypeOf(e)});if(0===o.length)return null;var a=o[o.length-1];return{symbol:e,type:i,providedIn:a.providedIn,useValue:a.useValue,useClass:a.useClass,useExisting:a.useExisting,useFactory:a.useFactory,deps:a.deps}},e.prototype._getTypeMetadata=function(e,t,n){void 0===t&&(t=null),void 0===n&&(n=!0);var r,i,o=this._getIdentifierMetadata(e);return{reference:o.reference,diDeps:this._getDependenciesMetadata(o.reference,t,n),lifecycleHooks:(r=this._reflector,i=o.reference,ma.filter(function(e){return ga(r,e,i)}))}},e.prototype._getFactoryMetadata=function(e,t){return void 0===t&&(t=null),{reference:e=nn(e),diDeps:this._getDependenciesMetadata(e,t)}},e.prototype.getPipeMetadata=function(e){var t=this._pipeCache.get(e);return t||this._reportError(Yt("Illegal state: getPipeMetadata can only be called after loadNgModuleDirectiveAndPipeMetadata for a module that declares it. Pipe "+Dp(e)+"."),e),t||null},e.prototype.getPipeSummary=function(e){var t=this._loadSummary(e,On.Pipe);return t||this._reportError(Yt("Illegal state: Could not load the summary for pipe "+Dp(e)+"."),e),t},e.prototype.getOrLoadPipeMetadata=function(e){var t=this._pipeCache.get(e);return t||(t=this._loadPipeMetadata(e)),t},e.prototype._loadPipeMetadata=function(e){e=nn(e);var t=this._pipeResolver.resolve(e),n=new Fn({type:this._getTypeMetadata(e),name:t.name,pure:!!t.pure});return this._pipeCache.set(e,n),this._summaryCache.set(e,n.toSummary()),n},e.prototype._getDependenciesMetadata=function(e,t,n){var r=this;void 0===n&&(n=!0);var i=!1,o=(t||this._reflector.parameters(e)||[]).map(function(e){var t=!1,n=!1,o=!1,a=!1,s=!1,u=null;return Array.isArray(e)?e.forEach(function(e){$.isTypeOf(e)?n=!0:X.isTypeOf(e)?o=!0:Z.isTypeOf(e)?a=!0:Q.isTypeOf(e)?s=!0:M.isTypeOf(e)?(t=!0,u=e.attributeName):S.isTypeOf(e)?u=e.token:k.isTypeOf(e)||e instanceof yn?u=e:Ap(e)&&null==u&&(u=e)}):u=e,null==u?(i=!0,null):{isAttribute:t,isHost:n,isSelf:o,isSkipSelf:a,isOptional:s,token:r._getTokenMetadata(u)}});if(i){var a=o.map(function(e){return e?Dp(e.token):"?"}).join(", "),s="Can't resolve all parameters for "+Dp(e)+": ("+a+").";n||this._config.strictInjectionParameters?this._reportError(Yt(s),e):this._console.warn("Warning: "+s+" This will become an error in Angular v6.x")}return o},e.prototype._getTokenMetadata=function(e){return"string"==typeof(e=nn(e))?{value:e}:{identifier:{reference:e}}},e.prototype._getProvidersMetadata=function(e,t,n,r,i){var a=this;return void 0===r&&(r=[]),e.forEach(function(s,u){if(Array.isArray(s))a._getProvidersMetadata(s,t,n,r);else{var c=void 0;if((s=nn(s))&&"object"==typeof s&&s.hasOwnProperty("provide"))a._validateProvider(s),c=new Un(s.provide,s);else{if(!Ap(s)){if(void 0===s)return void a._reportError(Yt("Encountered undefined provider! Usually this means you have a circular dependencies. This might be caused by using 'barrel' index.ts files."));var l=e.reduce(function(e,t,n){return n<u?e.push(""+Dp(t)):n==u?e.push("?"+Dp(t)+"?"):n==u+1&&e.push("..."),e},[]).join(", ");return void a._reportError(Yt("Invalid "+(n||"provider")+" - only instances of Provider and Type are allowed, got: ["+l+"]"),i)}c=new Un(s,{useClass:s})}c.token===a._reflector.resolveExternalReference(gn.ANALYZE_FOR_ENTRY_COMPONENTS)?t.push.apply(t,Object(o.__spread)(a._getEntryComponentsFromProvider(c,i))):r.push(a.getProviderMetadata(c))}}),r},e.prototype._validateProvider=function(e){e.hasOwnProperty("useClass")&&null==e.useClass&&this._reportError(Yt("Invalid provider for "+Dp(e.provide)+". useClass cannot be "+e.useClass+".\n           Usually it happens when:\n           1. There's a circular dependency (might be caused by using index.ts (barrel) files).\n           2. Class was used before it was declared. Use forwardRef in this case."))},e.prototype._getEntryComponentsFromProvider=function(e,t){var n,r,i=this,o=[],a=[];return e.useFactory||e.useExisting||e.useClass?(this._reportError(Yt("The ANALYZE_FOR_ENTRY_COMPONENTS token only supports useValue!"),t),[]):e.multi?(n=e.useValue,r=a,Vt(n,new Tp,r),a.forEach(function(e){var t=i._getEntryComponentMetadata(e.reference,!1);t&&o.push(t)}),o):(this._reportError(Yt("The ANALYZE_FOR_ENTRY_COMPONENTS token only supports 'multi = true'!"),t),[])},e.prototype._getEntryComponentMetadata=function(e,t){void 0===t&&(t=!0);var n=this.getNonNormalizedDirectiveMetadata(e);if(n&&n.metadata.isComponent)return{componentType:e,componentFactory:n.metadata.componentFactory};var r=this._loadSummary(e,On.Directive);if(r&&r.isComponent)return{componentType:e,componentFactory:r.componentFactory};if(t)throw Yt(e.name+" cannot be used as an entry component.");return null},e.prototype._getInjectableTypeMetadata=function(e,t){void 0===t&&(t=null);var n=this._loadSummary(e,On.Injectable);return n?n.type:this._getTypeMetadata(e,t)},e.prototype.getProviderMetadata=function(e){var t=void 0,n=null,r=null,i=this._getTokenMetadata(e.token);return e.useClass?(t=(n=this._getInjectableTypeMetadata(e.useClass,e.dependencies)).diDeps,e.token===e.useClass&&(i={identifier:n})):e.useFactory&&(t=(r=this._getFactoryMetadata(e.useFactory,e.dependencies)).diDeps),{token:i,useClass:n,useValue:e.useValue,useFactory:r,useExisting:e.useExisting?this._getTokenMetadata(e.useExisting):void 0,deps:t,multi:e.multi}},e.prototype._getQueriesMetadata=function(e,t,n){var r=this,i=[];return Object.keys(e).forEach(function(o){var a=e[o];a.isViewQuery===t&&i.push(r._getQueryMetadata(a,o,n))}),i},e.prototype._queryVarBindings=function(e){return e.split(/\s*,\s*/)},e.prototype._getQueryMetadata=function(e,t,n){var r,i=this;return"string"==typeof e.selector?r=this._queryVarBindings(e.selector).map(function(e){return i._getTokenMetadata(e)}):e.selector?r=[this._getTokenMetadata(e.selector)]:(this._reportError(Yt("Can't construct a query for the property \""+t+'" of "'+Dp(n)+"\" since the query selector wasn't defined."),n),r=[]),{selectors:r,first:e.first,descendants:e.descendants,propertyName:t,read:e.read?this._getTokenMetadata(e.read):null}},e.prototype._reportError=function(e,t,n){if(!this._errorCollector)throw e;this._errorCollector(e,t),n&&this._errorCollector(e,n)},e}();function Mp(e){return(t=function e(t,n){if(void 0===n&&(n=[]),t)for(var r=0;r<t.length;r++){var i=nn(t[r]);Array.isArray(i)?e(i,n):n.push(i)}return n}(e))?Array.from(new Set(t)):[];var t}function Ap(e){return e instanceof yn||e instanceof J}var Tp=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(o.__extends)(t,e),t.prototype.visitOther=function(e,t){t.push({reference:e})},t}(Ut);function Dp(e){return e instanceof yn?e.name+" in "+e.filePath:tn(e)}var Ip=function(e){function t(t,n){return e.call(this,n,t)||this}return Object(o.__extends)(t,e),t}(os),Pp=function(){return function(e,t){var n=this;this.reflector=e,this.component=t,this.errors=[],this.viewQueries=(r=t,i=1,o=new Map,r.viewQueries&&r.viewQueries.forEach(function(e){return Bp(o,{meta:e,queryId:i++})}),o),this.viewProviders=new Map,t.viewProviders.forEach(function(e){null==n.viewProviders.get(In(e.token))&&n.viewProviders.set(In(e.token),!0)});var r,i,o}}(),Rp=function(){function e(e,t,n,r,i,o,a,s,u){var c=this;this.viewContext=e,this._parent=t,this._isViewRoot=n,this._directiveAsts=r,this._sourceSpan=u,this._transformedProviders=new Map,this._seenProviders=new Map,this._queriedTokens=new Map,this.transformedHasViewContainer=!1,this._attrs={},i.forEach(function(e){return c._attrs[e.name]=e.value});var l,d,p,f,h,m,g,v=r.map(function(e){return e.directive});if(this._allProviders=(l=v,d=u,p=e.errors,f=new Map,l.forEach(function(e){Vp([{token:{identifier:e.type},useClass:e.type}],e.isComponent?sd.Component:sd.Directive,!0,d,p,f,!1)}),l.filter(function(e){return e.isComponent}).concat(l.filter(function(e){return!e.isComponent})).forEach(function(e){Vp(e.providers,sd.PublicService,!1,d,p,f,!1),Vp(e.viewProviders,sd.PrivateService,!1,d,p,f,!1)}),f),this._contentQueries=(h=v,m=s,g=new Map,h.forEach(function(e,t){e.queries&&e.queries.forEach(function(e){return Bp(g,{meta:e,queryId:m++})})}),g),Array.from(this._allProviders.values()).forEach(function(e){c._addQueryReadsTo(e.token,e.token,c._queriedTokens)}),a){var _=_n(this.viewContext.reflector,gn.TemplateRef);this._addQueryReadsTo(_,_,this._queriedTokens)}o.forEach(function(e){var t=e.value||_n(c.viewContext.reflector,gn.ElementRef);c._addQueryReadsTo({value:e.name},t,c._queriedTokens)}),this._queriedTokens.get(this.viewContext.reflector.resolveExternalReference(gn.ViewContainerRef))&&(this.transformedHasViewContainer=!0),Array.from(this._allProviders.values()).forEach(function(e){(e.eager||c._queriedTokens.get(In(e.token)))&&c._getOrCreateLocalProvider(e.providerType,e.token,!0)})}return e.prototype.afterElement=function(){var e=this;Array.from(this._allProviders.values()).forEach(function(t){e._getOrCreateLocalProvider(t.providerType,t.token,!1)})},Object.defineProperty(e.prototype,"transformProviders",{get:function(){var e=[],t=[];return this._transformedProviders.forEach(function(n){n.eager?t.push(n):e.push(n)}),e.concat(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"transformedDirectiveAsts",{get:function(){var e=this.transformProviders.map(function(e){return e.token.identifier}),t=this._directiveAsts.slice();return t.sort(function(t,n){return e.indexOf(t.directive.type)-e.indexOf(n.directive.type)}),t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"queryMatches",{get:function(){var e=[];return this._queriedTokens.forEach(function(t){e.push.apply(e,Object(o.__spread)(t))}),e},enumerable:!0,configurable:!0}),e.prototype._addQueryReadsTo=function(e,t,n){this._getQueriesFor(e).forEach(function(e){var r=e.meta.read||t,i=In(r),o=n.get(i);o||(o=[],n.set(i,o)),o.push({queryId:e.queryId,value:r})})},e.prototype._getQueriesFor=function(e){for(var t,n=[],r=this,i=0;null!==r;)(t=r._contentQueries.get(In(e)))&&n.push.apply(n,Object(o.__spread)(t.filter(function(e){return e.meta.descendants||i<=1}))),r._directiveAsts.length>0&&i++,r=r._parent;return(t=this.viewContext.viewQueries.get(In(e)))&&n.push.apply(n,Object(o.__spread)(t)),n},e.prototype._getOrCreateLocalProvider=function(e,t,n){var r=this,i=this._allProviders.get(In(t));if(!i||(e===sd.Directive||e===sd.PublicService)&&i.providerType===sd.PrivateService||(e===sd.PrivateService||e===sd.PublicService)&&i.providerType===sd.Builtin)return null;var o=this._transformedProviders.get(In(t));if(o)return o;if(null!=this._seenProviders.get(In(t)))return this.viewContext.errors.push(new Ip("Cannot instantiate cyclic dependency! "+Dn(t),this._sourceSpan)),null;this._seenProviders.set(In(t),!0);var a=i.providers.map(function(e){var t=e.useValue,o=e.useExisting,a=void 0;if(null!=e.useExisting){var s=r._getDependency(i.providerType,{token:e.useExisting},n);null!=s.token?o=s.token:(o=null,t=s.value)}else if(e.useFactory){a=(e.deps||e.useFactory.diDeps).map(function(e){return r._getDependency(i.providerType,e,n)})}else if(e.useClass){a=(e.deps||e.useClass.diDeps).map(function(e){return r._getDependency(i.providerType,e,n)})}return Fp(e,{useExisting:o,useValue:t,deps:a})});return o=Lp(i,{eager:n,providers:a}),this._transformedProviders.set(In(t),o),o},e.prototype._getLocalDependency=function(e,t,n){if(void 0===n&&(n=!1),t.isAttribute){var r=this._attrs[t.token.value];return{isValue:!0,value:null==r?null:r}}if(null!=t.token){if(e===sd.Directive||e===sd.Component){if(In(t.token)===this.viewContext.reflector.resolveExternalReference(gn.Renderer)||In(t.token)===this.viewContext.reflector.resolveExternalReference(gn.ElementRef)||In(t.token)===this.viewContext.reflector.resolveExternalReference(gn.ChangeDetectorRef)||In(t.token)===this.viewContext.reflector.resolveExternalReference(gn.TemplateRef))return t;In(t.token)===this.viewContext.reflector.resolveExternalReference(gn.ViewContainerRef)&&(this.transformedHasViewContainer=!0)}if(In(t.token)===this.viewContext.reflector.resolveExternalReference(gn.Injector))return t;if(null!=this._getOrCreateLocalProvider(e,t.token,n))return t}return null},e.prototype._getDependency=function(e,t,n){void 0===n&&(n=!1);var r=this,i=n,o=null;if(t.isSkipSelf||(o=this._getLocalDependency(e,t,n)),t.isSelf)!o&&t.isOptional&&(o={isValue:!0,value:null});else{for(;!o&&r._parent;){var a=r;r=r._parent,a._isViewRoot&&(i=!1),o=r._getLocalDependency(sd.PublicService,t,i)}o||(o=!t.isHost||this.viewContext.component.isHost||this.viewContext.component.type.reference===In(t.token)||null!=this.viewContext.viewProviders.get(In(t.token))?t:t.isOptional?{isValue:!0,value:null}:null)}return o||this.viewContext.errors.push(new Ip("No provider for "+Dn(t.token),this._sourceSpan)),o},e}(),Np=function(){function e(e,t,n,r){var i=this;this.reflector=e,this._transformedProviders=new Map,this._seenProviders=new Map,this._errors=[],this._allProviders=new Map,t.transitiveModule.modules.forEach(function(e){Vp([{token:{identifier:e},useClass:e}],sd.PublicService,!0,r,i._errors,i._allProviders,!0)}),Vp(t.transitiveModule.providers.map(function(e){return e.provider}).concat(n),sd.PublicService,!1,r,this._errors,this._allProviders,!1)}return e.prototype.parse=function(){var e=this;if(Array.from(this._allProviders.values()).forEach(function(t){e._getOrCreateLocalProvider(t.token,t.eager)}),this._errors.length>0){var t=this._errors.join("\n");throw new Error("Provider parse errors:\n"+t)}var n=[],r=[];return this._transformedProviders.forEach(function(e){e.eager?r.push(e):n.push(e)}),n.concat(r)},e.prototype._getOrCreateLocalProvider=function(e,t){var n=this,r=this._allProviders.get(In(e));if(!r)return null;var i=this._transformedProviders.get(In(e));if(i)return i;if(null!=this._seenProviders.get(In(e)))return this._errors.push(new Ip("Cannot instantiate cyclic dependency! "+Dn(e),r.sourceSpan)),null;this._seenProviders.set(In(e),!0);var o=r.providers.map(function(e){var i=e.useValue,o=e.useExisting,a=void 0;if(null!=e.useExisting){var s=n._getDependency({token:e.useExisting},t,r.sourceSpan);null!=s.token?o=s.token:(o=null,i=s.value)}else if(e.useFactory){a=(e.deps||e.useFactory.diDeps).map(function(e){return n._getDependency(e,t,r.sourceSpan)})}else if(e.useClass){a=(e.deps||e.useClass.diDeps).map(function(e){return n._getDependency(e,t,r.sourceSpan)})}return Fp(e,{useExisting:o,useValue:i,deps:a})});return i=Lp(r,{eager:t,providers:o}),this._transformedProviders.set(In(e),i),i},e.prototype._getDependency=function(e,t,n){return void 0===t&&(t=!1),e.isSkipSelf||null==e.token||In(e.token)===this.reflector.resolveExternalReference(gn.Injector)||In(e.token)===this.reflector.resolveExternalReference(gn.ComponentFactoryResolver)||this._getOrCreateLocalProvider(e.token,t),e},e}();function Fp(e,t){var n=t.useExisting,r=t.useValue,i=t.deps;return{token:e.token,useClass:e.useClass,useExisting:n,useFactory:e.useFactory,useValue:r,deps:i,multi:e.multi}}function Lp(e,t){var n=t.eager,r=t.providers;return new xd(e.token,e.multiProvider,e.eager||n,r,e.providerType,e.lifecycleHooks,e.sourceSpan,e.isModule)}function Vp(e,t,n,r,i,o,a){e.forEach(function(e){var s=o.get(In(e.token));if(null!=s&&!!s.multiProvider!=!!e.multi&&i.push(new Ip("Mixing multi and non multi provider is not possible for token "+Dn(s.token),r)),s)e.multi||(s.providers.length=0),s.providers.push(e);else{var u=e.token.identifier&&e.token.identifier.lifecycleHooks?e.token.identifier.lifecycleHooks:[],c=!(e.useClass||e.useExisting||e.useFactory);s=new xd(e.token,!!e.multi,n||c,[e],t,u,r,a),o.set(In(e.token),s)}})}function Bp(e,t){t.meta.selectors.forEach(function(n){var r=e.get(In(n));r||(r=[],e.set(In(n),r)),r.push(t)})}function zp(e,t){var n=0;t.eager||(n|=4096),t.providerType===sd.PrivateService&&(n|=8192),t.isModule&&(n|=1073741824),t.lifecycleHooks.forEach(function(e){e!==pa.OnDestroy&&t.providerType!==sd.Directive&&t.providerType!==sd.Component||(n|=Gp(e))});var r=t.multiProvider?function(e,t,n){var r=[],i=[],o=n.map(function(t,n){var r;if(t.useClass){var i=a(n,t.deps||t.useClass.diDeps);r=e.importExpr(t.useClass.reference).instantiate(i)}else if(t.useFactory){var i=a(n,t.deps||t.useFactory.diDeps);r=e.importExpr(t.useFactory.reference).callFn(i)}else if(t.useExisting){var i=a(n,[{token:t.useExisting}]);r=i[0]}else r=cp(e,t.useValue);return r});return{providerExpr:Dt(i,[new at(At(o))],ge),flags:1024|t,depsExpr:At(r)};function a(t,n){return n.map(function(n,o){var a="p"+t+"_"+o;return i.push(new Ge(a,me)),r.push(qp(e,n)),Et(a)})}}(e,n,t.providers):Up(e,n,t.providerType,t.providers[0]);return{providerExpr:r.providerExpr,flags:r.flags,depsExpr:r.depsExpr,tokenExpr:Hp(e,t.token)}}function Up(e,t,n,r){var i,o;return n===sd.Directive||n===sd.Component?(i=e.importExpr(r.useClass.reference),t|=16384,o=r.deps||r.useClass.diDeps):r.useClass?(i=e.importExpr(r.useClass.reference),t|=512,o=r.deps||r.useClass.diDeps):r.useFactory?(i=e.importExpr(r.useFactory.reference),t|=1024,o=r.deps||r.useFactory.diDeps):r.useExisting?(i=et,t|=2048,o=[{token:r.useExisting}]):(i=cp(e,r.useValue),t|=256,o=[]),{providerExpr:i,flags:t,depsExpr:At(o.map(function(t){return qp(e,t)}))}}function Hp(e,t){return t.identifier?e.importExpr(t.identifier.reference):Pt(t.value)}function qp(e,t){var n=t.isValue?cp(e,t.value):Hp(e,t.token),r=0;return t.isSkipSelf&&(r|=1),t.isOptional&&(r|=2),t.isSelf&&(r|=4),t.isValue&&(r|=8),0===r?n:At([Pt(r),n])}function Gp(e){var t=0;switch(e){case pa.AfterContentChecked:t=2097152;break;case pa.AfterContentInit:t=1048576;break;case pa.AfterViewChecked:t=8388608;break;case pa.AfterViewInit:t=4194304;break;case pa.DoCheck:t=262144;break;case pa.OnChanges:t=524288;break;case pa.OnDestroy:t=131072;break;case pa.OnInit:t=65536}return t}function Wp(e,t,n,r){var i=r.map(function(e){return t.importExpr(e.componentFactory)}),o=_n(e,gn.ComponentFactoryResolver),a={diDeps:[{isValue:!0,value:At(i)},{token:o,isSkipSelf:!0,isOptional:!0},{token:_n(e,gn.NgModuleRef)}],lifecycleHooks:[],reference:e.resolveExternalReference(gn.CodegenComponentFactoryResolver)},s=Up(t,n,sd.PrivateService,{token:o,multi:!1,useClass:a});return{providerExpr:s.providerExpr,flags:s.flags,depsExpr:s.depsExpr,tokenExpr:Hp(t,o)}}var Yp=function(){return function(e){this.ngModuleFactoryVar=e}}(),Kp=Et("_l"),Qp=function(){function e(e){this.reflector=e}return e.prototype.compile=function(e,t,n){var r=as("NgModule",t.type),i=t.transitiveModule.entryComponents,o=t.bootstrapComponents,a=new Np(this.reflector,t,n,r),s=[Wp(this.reflector,e,0,i)].concat(a.parse().map(function(t){return zp(e,t)})).map(function(e){var t=e.providerExpr,n=e.depsExpr,r=e.flags,i=e.tokenExpr;return St(gn.moduleProviderDef).callFn([Pt(r),i,t,n])}),u=St(gn.moduleDef).callFn([At(s)]),c=Dt([new Ge(Kp.name)],[new at(u)],ge),l=En(t.type)+"NgFactory";if(this._createNgModuleFactory(e,t.type.reference,St(gn.createModuleFactory).callFn([e.importExpr(t.type.reference),At(o.map(function(t){return e.importExpr(t.reference)})),c])),t.id){var d="string"==typeof t.id?Pt(t.id):e.importExpr(t.id),p=St(gn.RegisterModuleFactoryFn).callFn([d,Et(l)]).toStmt();e.statements.push(p)}return new Yp(l)},e.prototype.createStub=function(e,t){this._createNgModuleFactory(e,t,et)},e.prototype._createNgModuleFactory=function(e,t,n){var r=Et(En({reference:t})+"NgFactory").set(n).toDeclStmt(kt(gn.NgModuleFactory,[Mt(e.importExpr(t))],[re.Const]),[Ie.Final,Ie.Exported]);e.statements.push(r)},e}(),Xp=function(){function e(e){this._reflector=e}return e.prototype.isNgModule=function(e){return this._reflector.annotations(e).some(G.isTypeOf)},e.prototype.resolve=function(e,t){void 0===t&&(t=!0);var n=Pd(this._reflector.annotations(e),G.isTypeOf);if(n)return n;if(t)throw new Error("No NgModule metadata found for '"+tn(e)+"'.");return null},e}();function Zp(e){var t=new Jp,n=Ji.createRoot();return(Array.isArray(e)?e:[e]).forEach(function(e){if(e instanceof nt)e.visitStatement(t,n);else if(e instanceof we)e.visitExpression(t,n);else{if(!(e instanceof ue))throw new Error("Don't know how to print debug info for "+e);e.visitType(t,n)}}),n.toSource()}var $p=function(){function e(){}return e.prototype.emitStatementsAndContext=function(e,t,n,r,i,a){void 0===n&&(n=""),void 0===r&&(r=!0);var s=new Jp(i,a),u=Ji.createRoot();s.visitAllStatements(t,u);var c=n?n.split("\n"):[];s.reexports.forEach(function(e,t){var n=e.map(function(e){return e.name+" as "+e.as}).join(",");c.push("export {"+n+"} from '"+t+"';")}),s.importsWithPrefixes.forEach(function(e,t){c.push("import * as "+e+" from '"+t+"';")});var l=r?u.toSourceMapGenerator(e,c.length).toJsComment():"",d=Object(o.__spread)(c,[u.toSource(),l]);return l&&d.push(""),u.setPreambleLineCount(c.length),{sourceText:d.join("\n"),context:u}},e.prototype.emitStatements=function(e,t,n){return void 0===n&&(n=""),this.emitStatementsAndContext(e,t,n).sourceText},e}(),Jp=function(e){function t(t,n){var r=e.call(this,!1)||this;return r.referenceFilter=t,r.importFilter=n,r.typeExpression=0,r.importsWithPrefixes=new Map,r.reexports=new Map,r}return Object(o.__extends)(t,e),t.prototype.visitType=function(e,t,n){void 0===n&&(n="any"),e?(this.typeExpression++,e.visitType(this,t),this.typeExpression--):t.print(null,n)},t.prototype.visitLiteralExpr=function(t,n){var r=t.value;return null==r&&t.type!=ge?(n.print(t,"("+r+" as any)"),null):e.prototype.visitLiteralExpr.call(this,t,n)},t.prototype.visitLiteralArrayExpr=function(t,n){0===t.entries.length&&n.print(t,"(");var r=e.prototype.visitLiteralArrayExpr.call(this,t,n);return 0===t.entries.length&&n.print(t," as any[])"),r},t.prototype.visitExternalExpr=function(e,t){return this._visitIdentifier(e.value,e.typeParams,t),null},t.prototype.visitAssertNotNullExpr=function(t,n){var r=e.prototype.visitAssertNotNullExpr.call(this,t,n);return n.print(t,"!"),r},t.prototype.visitDeclareVarStmt=function(e,t){if(e.hasModifier(Ie.Exported)&&e.value instanceof Ve&&!e.type){var n=e.value.value,r=n.name,i=n.moduleName;if(i){var o=this.reexports.get(i);return o||(o=[],this.reexports.set(i,o)),o.push({name:r,as:e.name}),null}}return e.hasModifier(Ie.Exported)&&t.print(e,"export "),e.hasModifier(Ie.Final)?t.print(e,"const"):t.print(e,"var"),t.print(e," "+e.name),this._printColonType(e.type,t),e.value&&(t.print(e," = "),e.value.visitExpression(this,t)),t.println(e,";"),null},t.prototype.visitWrappedNodeExpr=function(e,t){throw new Error("Cannot visit a WrappedNodeExpr when outputting Typescript.")},t.prototype.visitCastExpr=function(e,t){return t.print(e,"(<"),e.type.visitType(this,t),t.print(e,">"),e.value.visitExpression(this,t),t.print(e,")"),null},t.prototype.visitInstantiateExpr=function(e,t){return t.print(e,"new "),this.typeExpression++,e.classExpr.visitExpression(this,t),this.typeExpression--,t.print(e,"("),this.visitAllExpressions(e.args,t,","),t.print(e,")"),null},t.prototype.visitDeclareClassStmt=function(e,t){var n=this;return t.pushClass(e),e.hasModifier(Ie.Exported)&&t.print(e,"export "),t.print(e,"class "+e.name),null!=e.parent&&(t.print(e," extends "),this.typeExpression++,e.parent.visitExpression(this,t),this.typeExpression--),t.println(e," {"),t.incIndent(),e.fields.forEach(function(e){return n._visitClassField(e,t)}),null!=e.constructorMethod&&this._visitClassConstructor(e,t),e.getters.forEach(function(e){return n._visitClassGetter(e,t)}),e.methods.forEach(function(e){return n._visitClassMethod(e,t)}),t.decIndent(),t.println(e,"}"),t.popClass(),null},t.prototype._visitClassField=function(e,t){e.hasModifier(Ie.Private)&&t.print(null,"/*private*/ "),e.hasModifier(Ie.Static)&&t.print(null,"static "),t.print(null,e.name),this._printColonType(e.type,t),e.initializer&&(t.print(null," = "),e.initializer.visitExpression(this,t)),t.println(null,";")},t.prototype._visitClassGetter=function(e,t){e.hasModifier(Ie.Private)&&t.print(null,"private "),t.print(null,"get "+e.name+"()"),this._printColonType(e.type,t),t.println(null," {"),t.incIndent(),this.visitAllStatements(e.body,t),t.decIndent(),t.println(null,"}")},t.prototype._visitClassConstructor=function(e,t){t.print(e,"constructor("),this._visitParams(e.constructorMethod.params,t),t.println(e,") {"),t.incIndent(),this.visitAllStatements(e.constructorMethod.body,t),t.decIndent(),t.println(e,"}")},t.prototype._visitClassMethod=function(e,t){e.hasModifier(Ie.Private)&&t.print(null,"private "),t.print(null,e.name+"("),this._visitParams(e.params,t),t.print(null,")"),this._printColonType(e.type,t,"void"),t.println(null," {"),t.incIndent(),this.visitAllStatements(e.body,t),t.decIndent(),t.println(null,"}")},t.prototype.visitFunctionExpr=function(e,t){return e.name&&(t.print(e,"function "),t.print(e,e.name)),t.print(e,"("),this._visitParams(e.params,t),t.print(e,")"),this._printColonType(e.type,t,"void"),e.name||t.print(e," => "),t.println(e,"{"),t.incIndent(),this.visitAllStatements(e.statements,t),t.decIndent(),t.print(e,"}"),null},t.prototype.visitDeclareFunctionStmt=function(e,t){return e.hasModifier(Ie.Exported)&&t.print(e,"export "),t.print(e,"function "+e.name+"("),this._visitParams(e.params,t),t.print(e,")"),this._printColonType(e.type,t,"void"),t.println(e," {"),t.incIndent(),this.visitAllStatements(e.statements,t),t.decIndent(),t.println(e,"}"),null},t.prototype.visitTryCatchStmt=function(e,t){t.println(e,"try {"),t.incIndent(),this.visitAllStatements(e.bodyStmts,t),t.decIndent(),t.println(e,"} catch ("+Xi.name+") {"),t.incIndent();var n=[Zi.set(Xi.prop("stack",null)).toDeclStmt(null,[Ie.Final])].concat(e.catchStmts);return this.visitAllStatements(n,t),t.decIndent(),t.println(e,"}"),null},t.prototype.visitBuiltinType=function(e,t){var n;switch(e.name){case ae.Bool:n="boolean";break;case ae.Dynamic:n="any";break;case ae.Function:n="Function";break;case ae.Number:case ae.Int:n="number";break;case ae.String:n="string";break;case ae.None:n="never";break;default:throw new Error("Unsupported builtin type "+e.name)}return t.print(null,n),null},t.prototype.visitExpressionType=function(e,t){var n=this;return e.value.visitExpression(this,t),null!==e.typeParams&&(t.print(null,"<"),this.visitAllObjects(function(e){return n.visitType(e,t)},e.typeParams,t,","),t.print(null,">")),null},t.prototype.visitArrayType=function(e,t){return this.visitType(e.of,t),t.print(null,"[]"),null},t.prototype.visitMapType=function(e,t){return t.print(null,"{[key: string]:"),this.visitType(e.valueType,t),t.print(null,"}"),null},t.prototype.getBuiltinMethodName=function(e){var t;switch(e){case Ce.ConcatArray:t="concat";break;case Ce.SubscribeObservable:t="subscribe";break;case Ce.Bind:t="bind";break;default:throw new Error("Unknown builtin method: "+e)}return t},t.prototype._visitParams=function(e,t){var n=this;this.visitAllObjects(function(e){t.print(null,e.name),n._printColonType(e.type,t)},e,t,",")},t.prototype._visitIdentifier=function(e,t,n){var r=this,i=e.name,o=e.moduleName;if(this.referenceFilter&&this.referenceFilter(e))n.print(null,"(null as any)");else{if(o&&(!this.importFilter||!this.importFilter(e))){var a=this.importsWithPrefixes.get(o);null==a&&(a="i"+this.importsWithPrefixes.size,this.importsWithPrefixes.set(o,a)),n.print(null,a+".")}if(n.print(null,i),this.typeExpression>0)(t||[]).length>0&&(n.print(null,"<"),this.visitAllObjects(function(e){return e.visitType(r,n)},t,n,","),n.print(null,">"))}},t.prototype._printColonType=function(e,t,n){e!==ge&&(t.print(null,":"),this.visitType(e,t,n))},t}(eo),ef=function(){function e(e){this._reflector=e}return e.prototype.isPipe=function(e){var t=this._reflector.annotations(nn(e));return t&&t.some(B.isTypeOf)},e.prototype.resolve=function(e,t){void 0===t&&(t=!0);var n=this._reflector.annotations(nn(e));if(n){var r=Pd(n,B.isTypeOf);if(r)return r}if(t)throw new Error("No Pipe decorator found on "+tn(e));return null},e}(),tf=["zero","one","two","few","many","other"];function nf(e){var t=new af;return new rf(_u(t,e),t.isExpanded,t.errors)}var rf=function(){return function(e,t,n){this.nodes=e,this.expanded=t,this.errors=n}}(),of=function(e){function t(t,n){return e.call(this,t,n)||this}return Object(o.__extends)(t,e),t}(os),af=function(){function e(){this.isExpanded=!1,this.errors=[]}return e.prototype.visitElement=function(e,t){return new gu(e.name,e.attrs,_u(this,e.children),e.sourceSpan,e.startSourceSpan,e.endSourceSpan)},e.prototype.visitAttribute=function(e,t){return e},e.prototype.visitText=function(e,t){return e},e.prototype.visitComment=function(e,t){return e},e.prototype.visitExpansion=function(e,t){return this.isExpanded=!0,"plural"==e.type?(s=e,u=this.errors,c=s.cases.map(function(e){-1!=tf.indexOf(e.value)||e.value.match(/^=\d+$/)||u.push(new of(e.valueSourceSpan,'Plural cases should be "=<number>" or one of '+tf.join(", ")));var t=nf(e.expression);return u.push.apply(u,Object(o.__spread)(t.errors)),new gu("ng-template",[new mu("ngPluralCase",""+e.value,e.valueSourceSpan)],t.nodes,e.sourceSpan,e.sourceSpan,e.sourceSpan)}),l=new mu("[ngPlural]",s.switchValue,s.switchValueSourceSpan),new gu("ng-container",[l],c,s.sourceSpan,s.sourceSpan,s.sourceSpan)):(n=e,r=this.errors,i=n.cases.map(function(e){var t=nf(e.expression);return r.push.apply(r,Object(o.__spread)(t.errors)),"other"===e.value?new gu("ng-template",[new mu("ngSwitchDefault","",e.valueSourceSpan)],t.nodes,e.sourceSpan,e.sourceSpan,e.sourceSpan):new gu("ng-template",[new mu("ngSwitchCase",""+e.value,e.valueSourceSpan)],t.nodes,e.sourceSpan,e.sourceSpan,e.sourceSpan)}),a=new mu("[ngSwitch]",n.switchValue,n.switchValueSourceSpan),new gu("ng-container",[a],i,n.sourceSpan,n.sourceSpan,n.sourceSpan));var n,r,i,a,s,u,c,l},e.prototype.visitExpansionCase=function(e,t){throw new Error("Should not be reached")},e}();var sf,uf=/^(?:(?:(?:(bind-)|(let-)|(ref-|#)|(on-)|(bindon-)|(@))(.+))|\[\(([^\)]+)\)\]|\[([^\]]+)\]|\(([^\)]+)\))$/,cf="class";function lf(){return sf||(sf=O.parse("*")[0]),sf}var df=function(e){function t(t,n,r){return e.call(this,n,t,r)||this}return Object(o.__extends)(t,e),t}(os),pf=function(){return function(e,t,n){this.templateAst=e,this.usedPipes=t,this.errors=n}}(),ff=function(){function e(e,t,n,r,i,o,a){this._config=e,this._reflector=t,this._exprParser=n,this._schemaRegistry=r,this._htmlParser=i,this._console=o,this.transforms=a}return Object.defineProperty(e.prototype,"expressionParser",{get:function(){return this._exprParser},enumerable:!0,configurable:!0}),e.prototype.parse=function(e,t,n,r,i,o,a){var s=this.tryParse(e,t,n,r,i,o,a),u=s.errors.filter(function(e){return e.level===es.WARNING}),c=s.errors.filter(function(e){return e.level===es.ERROR});if(u.length>0&&this._console.warn("Template parse warnings:\n"+u.join("\n")),c.length>0)throw Yt("Template parse errors:\n"+c.join("\n"),c);return{template:s.templateAst,pipes:s.usedPipes}},e.prototype.tryParse=function(e,t,n,r,i,o,a){var s="string"==typeof t?this._htmlParser.parse(t,o,!0,this.getInterpolationConfig(e)):t;return a||(s=Zu(s)),this.tryParseHtml(this.expandHtml(s),e,n,r,i)},e.prototype.tryParseHtml=function(e,t,n,r,i){var a,s=e.errors,u=[];if(e.rootNodes.length>0){var c=Of(n),l=Of(r),d=new Pp(this._reflector,t),p=void 0;t.template&&t.template.interpolation&&(p={start:t.template.interpolation[0],end:t.template.interpolation[1]});var f=new uc(this._exprParser,p,this._schemaRegistry,l,s),h=new hf(this._reflector,this._config,d,c,f,this._schemaRegistry,i,s);a=_u(h,e.rootNodes,bf),s.push.apply(s,Object(o.__spread)(d.errors)),u.push.apply(u,Object(o.__spread)(f.getUsedPipes()))}else a=[];return this._assertNoReferenceDuplicationOnTemplate(a,s),s.length>0?new pf(a,u,s):(this.transforms&&this.transforms.forEach(function(e){a=Ed(e,a)}),new pf(a,u,s))},e.prototype.expandHtml=function(e,t){void 0===t&&(t=!1);var n=e.errors;if(0==n.length||t){var r=nf(e.rootNodes);n.push.apply(n,Object(o.__spread)(r.errors)),e=new Lu(r.nodes,n)}return e},e.prototype.getInterpolationConfig=function(e){if(e.template)return zi.fromArray(e.template.interpolation)},e.prototype._assertNoReferenceDuplicationOnTemplate=function(e,t){var n=[];e.filter(function(e){return!!e.references}).forEach(function(e){return e.references.forEach(function(e){var r=e.name;if(n.indexOf(r)<0)n.push(r);else{var i=new df('Reference "#'+r+'" is defined several times',e.sourceSpan,es.ERROR);t.push(i)}})})},e}(),hf=function(){function e(e,t,n,r,i,o,a,s){var u=this;this.reflector=e,this.config=t,this.providerViewContext=n,this._bindingParser=i,this._schemaRegistry=o,this._schemas=a,this._targetErrors=s,this.selectorMatcher=new w,this.directivesIndex=new Map,this.ngContentCount=0,this.contentQueryStartId=n.component.viewQueries.length+1,r.forEach(function(e,t){var n=O.parse(e.selector);u.selectorMatcher.addSelectables(n,e),u.directivesIndex.set(e,t)})}return e.prototype.visitExpansion=function(e,t){return null},e.prototype.visitExpansionCase=function(e,t){return null},e.prototype.visitText=function(e,t){var n=t.findNgContentIndex(lf()),r=Ku(e.value),i=this._bindingParser.parseInterpolation(r,e.sourceSpan);return i?new dd(i,n,e.sourceSpan):new ld(r,n,e.sourceSpan)},e.prototype.visitAttribute=function(e,t){return new pd(e.name,e.value,e.sourceSpan)},e.prototype.visitComment=function(e,t){return null},e.prototype.visitElement=function(e,t){var n=this,r=this.contentQueryStartId,i=e.name,a=Bc(e);if(a.type===Ec.SCRIPT||a.type===Ec.STYLE)return null;if(a.type===Ec.STYLESHEET&&Cc(a.hrefAttr))return null;var s=[],u=[],l=[],d=[],p=[],f=[],h=[],m=[],g=!1,v=[],_=c(e.name);e.attrs.forEach(function(e){var t,r,i=n._parseAttr(_,e,s,u,p,l,d);d.push.apply(d,Object(o.__spread)([].map(function(e){return vd.fromParsedVariable(e)})));var a=n._normalizeAttributeName(e.name);a.startsWith("*")&&(t=e.value,r=a.substring("*".length));var c=null!=t;if(c){g&&n._reportError("Can't have multiple template bindings on one element. Use only one attribute prefixed with *",e.sourceSpan),g=!0;var y=[];n._bindingParser.parseInlineTemplateBinding(r,t,e.sourceSpan,h,f,y),m.push.apply(m,Object(o.__spread)(y.map(function(e){return vd.fromParsedVariable(e)})))}i||c||(v.push(n.visitAttribute(e,null)),s.push([e.name,e.value]))});var y=yf(i,s),b=this._parseDirectives(this.selectorMatcher,y),j=b.directives,x=b.matchElement,w=[],C=new Set,E=this._createDirectiveAsts(_,e.name,j,u,l,e.sourceSpan,w,C),S=this._createElementPropertyAsts(e.name,u,C),k=t.isTemplateElement||g,M=new Rp(this.providerViewContext,t.providerContext,k,E,v,w,_,r,e.sourceSpan),A=_u(a.nonBindable?jf:this,e.children,_f.create(_,E,_?t.providerContext:M));M.afterElement();var T,D=""!=a.projectAs?O.parse(a.projectAs)[0]:y,I=t.findNgContentIndex(D);if(a.type===Ec.NG_CONTENT)e.children&&!e.children.every(xf)&&this._reportError("<ng-content> element cannot have content.",e.sourceSpan),T=new Od(this.ngContentCount++,g?null:I,e.sourceSpan);else if(_)this._assertAllEventsPublishedByDirectives(E,p),this._assertNoComponentsNorElementBindingsOnTemplate(E,S,e.sourceSpan),T=new yd(v,p,w,d,M.transformedDirectiveAsts,M.transformProviders,M.transformedHasViewContainer,M.queryMatches,A,g?null:I,e.sourceSpan);else{this._assertElementExists(x,e),this._assertOnlyOneComponent(E,e.sourceSpan);var P=g?null:t.findNgContentIndex(D);T=new _d(i,v,S,p,w,M.transformedDirectiveAsts,M.transformProviders,M.transformedHasViewContainer,M.queryMatches,A,g?null:P,e.sourceSpan,e.endSourceSpan||null)}if(g){var R=this.contentQueryStartId,N=yf("ng-template",h),F=this._parseDirectives(this.selectorMatcher,N).directives,L=new Set,V=this._createDirectiveAsts(!0,i,F,f,[],e.sourceSpan,[],L),B=this._createElementPropertyAsts(i,f,L);this._assertNoComponentsNorElementBindingsOnTemplate(V,B,e.sourceSpan);var z=new Rp(this.providerViewContext,t.providerContext,t.isTemplateElement,V,[],[],!0,R,e.sourceSpan);z.afterElement(),T=new yd([],[],[],m,z.transformedDirectiveAsts,z.transformProviders,z.transformedHasViewContainer,z.queryMatches,[T],I,e.sourceSpan)}return T},e.prototype._parseAttr=function(e,t,n,r,i,a,s){var u=this._normalizeAttributeName(t.name),c=t.value,l=t.sourceSpan,d=[],p=u.match(uf),f=!1;if(null!==p)if(f=!0,null!=p[1])this._bindingParser.parsePropertyBinding(p[7],c,!1,l,n,r);else if(p[2])if(e){var h=p[7];this._parseVariable(h,c,l,s)}else this._reportError('"let-" is only supported on ng-template elements.',l);else if(p[3]){h=p[7];this._parseReference(h,c,l,a)}else p[4]?this._bindingParser.parseEvent(p[7],c,l,n,d):p[5]?(this._bindingParser.parsePropertyBinding(p[7],c,!1,l,n,r),this._parseAssignmentEvent(p[7],c,l,n,d)):p[6]?this._bindingParser.parseLiteralAttr(u,c,l,n,r):p[8]?(this._bindingParser.parsePropertyBinding(p[8],c,!1,l,n,r),this._parseAssignmentEvent(p[8],c,l,n,d)):p[9]?this._bindingParser.parsePropertyBinding(p[9],c,!1,l,n,r):p[10]&&this._bindingParser.parseEvent(p[10],c,l,n,d);else f=this._bindingParser.parsePropertyInterpolation(u,c,l,n,r);return f||this._bindingParser.parseLiteralAttr(u,c,l,n,r),i.push.apply(i,Object(o.__spread)(d.map(function(e){return md.fromParsedEvent(e)}))),f},e.prototype._normalizeAttributeName=function(e){return/^data-/i.test(e)?e.substring(5):e},e.prototype._parseVariable=function(e,t,n,r){e.indexOf("-")>-1&&this._reportError('"-" is not allowed in variable names',n),r.push(new vd(e,t,n))},e.prototype._parseReference=function(e,t,n,r){e.indexOf("-")>-1&&this._reportError('"-" is not allowed in reference names',n),r.push(new gf(e,t,n))},e.prototype._parseAssignmentEvent=function(e,t,n,r,i){this._bindingParser.parseEvent(e+"Change",t+"=$event",n,r,i)},e.prototype._parseDirectives=function(e,t){var n=this,r=new Array(this.directivesIndex.size),i=!1;return e.match(t,function(e,t){r[n.directivesIndex.get(t)]=t,i=i||e.hasElementSelector()}),{directives:r.filter(function(e){return!!e}),matchElement:i}},e.prototype._createDirectiveAsts=function(e,t,n,r,i,o,a,s){var u=this,c=new Set,l=null,d=n.map(function(e){var n=new is(o.start,o.end,"Directive "+En(e.type));e.isComponent&&(l=e);var d=[],p=u._bindingParser.createDirectiveHostPropertyAsts(e,t,n).map(function(e){return hd.fromBoundProperty(e)});p=u._checkPropertiesInSchema(t,p);var f=u._bindingParser.createDirectiveHostEventAsts(e,n);u._createDirectivePropertyAsts(e.inputs,r,d,s),i.forEach(function(t){(0===t.value.length&&e.isComponent||t.isReferenceToDirective(e))&&(a.push(new gd(t.name,vn(e.type.reference),t.value,t.sourceSpan)),c.add(t.name))});var h=f.map(function(e){return md.fromParsedEvent(e)}),m=u.contentQueryStartId;return u.contentQueryStartId+=e.queries.length,new jd(e,d,p,h,m,n)});return i.forEach(function(t){if(t.value.length>0)c.has(t.name)||u._reportError('There is no directive with "exportAs" set to "'+t.value+'"',t.sourceSpan);else if(!l){var n=null;e&&(n=_n(u.reflector,gn.TemplateRef)),a.push(new gd(t.name,n,t.value,t.sourceSpan))}}),d},e.prototype._createDirectivePropertyAsts=function(e,t,n,r){if(e){var i=new Map;t.forEach(function(e){var t=i.get(e.name);t&&!t.isLiteral||i.set(e.name,e)}),Object.keys(e).forEach(function(t){var o=e[t],a=i.get(o);a&&(r.add(a.name),wf(a.expression)||n.push(new bd(t,a.name,a.expression,a.sourceSpan)))})}},e.prototype._createElementPropertyAsts=function(e,t,n){var r=this,i=[];return t.forEach(function(t){if(!t.isLiteral&&!n.has(t.name)){var o=r._bindingParser.createBoundElementProperty(e,t);i.push(hd.fromBoundProperty(o))}}),this._checkPropertiesInSchema(e,i)},e.prototype._findComponentDirectives=function(e){return e.filter(function(e){return e.directive.isComponent})},e.prototype._findComponentDirectiveNames=function(e){return this._findComponentDirectives(e).map(function(e){return En(e.directive.type)})},e.prototype._assertOnlyOneComponent=function(e,t){var n=this._findComponentDirectiveNames(e);n.length>1&&this._reportError("More than one component matched on this element.\nMake sure that only one component's selector can match a given element.\nConflicting components: "+n.join(","),t)},e.prototype._assertElementExists=function(e,t){var n=t.name.replace(/^:xhtml:/,"");if(!e&&!this._schemaRegistry.hasElement(n,this._schemas)){var r="'"+n+"' is not a known element:\n";r+="1. If '"+n+"' is an Angular component, then verify that it is part of this module.\n",n.indexOf("-")>-1?r+="2. If '"+n+"' is a Web Component then add 'CUSTOM_ELEMENTS_SCHEMA' to the '@NgModule.schemas' of this component to suppress this message.":r+="2. To allow any element add 'NO_ERRORS_SCHEMA' to the '@NgModule.schemas' of this component.",this._reportError(r,t.sourceSpan)}},e.prototype._assertNoComponentsNorElementBindingsOnTemplate=function(e,t,n){var r=this,i=this._findComponentDirectiveNames(e);i.length>0&&this._reportError("Components on an embedded template: "+i.join(","),n),t.forEach(function(e){r._reportError("Property binding "+e.name+' not used by any directive on an embedded template. Make sure that the property name is spelled correctly and all directives are listed in the "@NgModule.declarations".',n)})},e.prototype._assertAllEventsPublishedByDirectives=function(e,t){var n=this,r=new Set;e.forEach(function(e){Object.keys(e.directive.outputs).forEach(function(t){var n=e.directive.outputs[t];r.add(n)})}),t.forEach(function(e){null==e.target&&r.has(e.name)||n._reportError("Event binding "+e.fullName+' not emitted by any directive on an embedded template. Make sure that the event name is spelled correctly and all directives are listed in the "@NgModule.declarations".',e.sourceSpan)})},e.prototype._checkPropertiesInSchema=function(e,t){var n=this;return t.filter(function(t){if(0===t.type&&!n._schemaRegistry.hasProperty(e,t.name,n._schemas)){var r="Can't bind to '"+t.name+"' since it isn't a known property of '"+e+"'.";e.startsWith("ng-")?r+="\n1. If '"+t.name+"' is an Angular directive, then add 'CommonModule' to the '@NgModule.imports' of this component.\n2. To allow any property add 'NO_ERRORS_SCHEMA' to the '@NgModule.schemas' of this component.":e.indexOf("-")>-1&&(r+="\n1. If '"+e+"' is an Angular component and it has '"+t.name+"' input, then verify that it is part of this module.\n2. If '"+e+"' is a Web Component then add 'CUSTOM_ELEMENTS_SCHEMA' to the '@NgModule.schemas' of this component to suppress this message.\n3. To allow any property add 'NO_ERRORS_SCHEMA' to the '@NgModule.schemas' of this component."),n._reportError(r,t.sourceSpan)}return!wf(t.value)})},e.prototype._reportError=function(e,t,n){void 0===n&&(n=es.ERROR),this._targetErrors.push(new os(t,e,n))},e}(),mf=function(){function e(){}return e.prototype.visitElement=function(e,t){var n=Bc(e);if(n.type===Ec.SCRIPT||n.type===Ec.STYLE||n.type===Ec.STYLESHEET)return null;var r=e.attrs.map(function(e){return[e.name,e.value]}),i=yf(e.name,r),o=t.findNgContentIndex(i),a=_u(this,e.children,bf);return new _d(e.name,_u(this,e.attrs),[],[],[],[],[],!1,[],a,o,e.sourceSpan,e.endSourceSpan)},e.prototype.visitComment=function(e,t){return null},e.prototype.visitAttribute=function(e,t){return new pd(e.name,e.value,e.sourceSpan)},e.prototype.visitText=function(e,t){var n=t.findNgContentIndex(lf());return new ld(e.value,n,e.sourceSpan)},e.prototype.visitExpansion=function(e,t){return e},e.prototype.visitExpansionCase=function(e,t){return e},e}(),gf=function(){function e(e,t,n){this.name=e,this.value=t,this.sourceSpan=n}return e.prototype.isReferenceToDirective=function(e){return-1!==(t=e.exportAs,t?t.split(",").map(function(e){return e.trim()}):[]).indexOf(this.value);var t},e}();function vf(e){return e.trim().split(/\s+/g)}var _f=function(){function e(e,t,n,r){this.isTemplateElement=e,this._ngContentIndexMatcher=t,this._wildcardNgContentIndex=n,this.providerContext=r}return e.create=function(t,n,r){var i=new w,o=null,a=n.find(function(e){return e.directive.isComponent});if(a)for(var s=a.directive.template.ngContentSelectors,u=0;u<s.length;u++){"*"===s[u]?o=u:i.addSelectables(O.parse(s[u]),u)}return new e(t,i,o,r)},e.prototype.findNgContentIndex=function(e){var t=[];return this._ngContentIndexMatcher.match(e,function(e,n){t.push(n)}),t.sort(),null!=this._wildcardNgContentIndex&&t.push(this._wildcardNgContentIndex),t.length>0?t[0]:null},e}();function yf(e,t){var n=new O,r=a(e)[1];n.setElement(r);for(var i=0;i<t.length;i++){var o=t[i][0],s=a(o)[1],u=t[i][1];if(n.addAttribute(s,u),o.toLowerCase()==cf)vf(u).forEach(function(e){return n.addClassName(e)})}return n}var bf=new _f(!0,new w,null,null),jf=new mf;function xf(e){return e instanceof pu&&0==e.value.trim().length}function Of(e){var t=new Map;return e.forEach(function(e){t.get(e.type.reference)||t.set(e.type.reference,e)}),Array.from(t.values())}function wf(e){return e instanceof No&&(e=e.ast),e instanceof go}var Cf=function(){function e(e,t){this.options=e,this.reflector=t}return e.prototype.compileComponent=function(e,t,n,r,i,o){var a=this,s=new Map;r.forEach(function(e){return s.set(e.name,e.type.reference)});var u=0,c=function(e,n){var r=u++;return new Sf(a.options,a.reflector,i,e,t.type.reference,t.isHost,r,s,n,o,c)},l=c(null,[]);return l.visitAll([],n),l.build(e)},e}(),Ef=new(function(){function e(){}return e.prototype.getLocal=function(e){return e===Qo.event.name?Et("_any"):null},e}()),Sf=function(){function e(e,t,n,r,i,o,a,s,u,c,l){this.options=e,this.reflector=t,this.externalReferenceVars=n,this.parent=r,this.component=i,this.isHostComponent=o,this.embeddedViewIndex=a,this.pipes=s,this.guards=u,this.ctx=c,this.viewBuilderFactory=l,this.refOutputVars=new Map,this.variables=[],this.children=[],this.updates=[],this.actions=[]}return e.prototype.getOutputVar=function(e){var t;if(!(t=e===this.component&&this.isHostComponent?"_any":e instanceof yn?this.externalReferenceVars.get(e):"_any"))throw new Error("Illegal State: referring to a type without a variable "+JSON.stringify(e));return t},e.prototype.getTypeGuardExpressions=function(e){var t,n,r,i,a=Object(o.__spread)(this.guards);try{for(var s=Object(o.__values)(e.directives),u=s.next();!u.done;u=s.next()){var c=u.value;try{for(var l=Object(o.__values)(c.inputs),d=l.next();!d.done;d=l.next()){var p=d.value,f=c.directive.guards[p.directiveName];if(f){var h="UseIf"===f;a.push({guard:f,useIf:h,expression:{context:this.component,value:p.value}})}}}catch(e){r={error:e}}finally{try{d&&!d.done&&(i=l.return)&&i.call(l)}finally{if(r)throw r.error}}}}catch(e){t={error:e}}finally{try{u&&!u.done&&(n=s.return)&&n.call(s)}finally{if(t)throw t.error}}return a},e.prototype.visitAll=function(e,t){this.variables=e,Ed(this,t)},e.prototype.build=function(e,t){var n,r,i=this;void 0===t&&(t=[]),this.children.forEach(function(n){return n.build(e,t)});var a=[Et("_any").set(et).toDeclStmt(me)],s=0;if(this.updates.forEach(function(e){var t=i.preprocessUpdateExpression(e),n=t.sourceSpan,r=t.context,u=t.value,c=""+s++,l=ia(r===i.component?i:Ef,Et(i.getOutputVar(r)),u,c,Jo.General),d=l.stmts,p=l.currValExpr;d.push(new ot(p)),a.push.apply(a,Object(o.__spread)(d.map(function(e){return Ot(e,n)})))}),this.actions.forEach(function(e){var t=e.sourceSpan,n=e.context,r=e.value,u=""+s++,c=Zo(n===i.component?i:Ef,Et(i.getOutputVar(n)),r,u).stmts;a.push.apply(a,Object(o.__spread)(c.map(function(e){return Ot(e,t)})))}),this.guards.length){var u=void 0;try{for(var c=Object(o.__values)(this.guards),l=c.next();!l.done;l=c.next()){var d=l.value,p=this.preprocessUpdateExpression(d.expression),f=p.context,h=p.value,m=""+s++,g=ia(f===this.component?this:Ef,Et(this.getOutputVar(f)),h,m,Jo.TrySimple),v=g.stmts,_=g.currValExpr;if(0==v.length){var y=d.useIf?_:this.ctx.importExpr(d.guard).callFn([_]);u=u?u.and(y):y}}}catch(e){n={error:e}}finally{try{l&&!l.done&&(r=c.return)&&r.call(c)}finally{if(n)throw n.error}}u&&(a=[new pt(u,a)])}var b="_View_"+e+"_"+this.embeddedViewIndex,j=new it(b,[],a);return t.push(j),t},e.prototype.visitBoundText=function(e,t){var n=this;e.value.ast.expressions.forEach(function(t){return n.updates.push({context:n.component,value:t,sourceSpan:e.sourceSpan})})},e.prototype.visitEmbeddedTemplate=function(e,t){if(this.visitElementOrTemplate(e),this.options.fullTemplateTypeCheck){var n=this.getTypeGuardExpressions(e),r=this.viewBuilderFactory(this,n);this.children.push(r),r.visitAll(e.variables,e.children)}},e.prototype.visitElement=function(e,t){var n=this;this.visitElementOrTemplate(e),e.inputs.forEach(function(e){n.updates.push({context:n.component,value:e.value,sourceSpan:e.sourceSpan})}),Ed(this,e.children)},e.prototype.visitElementOrTemplate=function(e){var t=this;e.directives.forEach(function(e){t.visitDirective(e)}),e.references.forEach(function(e){var n=null;n=e.value&&e.value.identifier&&t.options.fullTemplateTypeCheck?e.value.identifier.reference:ae.Dynamic,t.refOutputVars.set(e.name,n)}),e.outputs.forEach(function(e){t.actions.push({context:t.component,value:e.handler,sourceSpan:e.sourceSpan})})},e.prototype.visitDirective=function(e){var t=this,n=e.directive.type.reference;e.inputs.forEach(function(e){return t.updates.push({context:t.component,value:e.value,sourceSpan:e.sourceSpan})}),this.options.fullTemplateTypeCheck&&(e.hostProperties.forEach(function(e){return t.updates.push({context:n,value:e.value,sourceSpan:e.sourceSpan})}),e.hostEvents.forEach(function(e){return t.actions.push({context:n,value:e.handler,sourceSpan:e.sourceSpan})}))},e.prototype.getLocal=function(e){if(e==Qo.event.name)return Et(this.getOutputVar(ae.Dynamic));for(var t=this;t;t=t.parent){var n=void 0;if(null==(n=t.refOutputVars.get(e)))t.variables.find(function(t){return t.name===e})&&(n=ae.Dynamic);if(null!=n)return Et(this.getOutputVar(n))}return null},e.prototype.pipeOutputVar=function(e){var t=this.pipes.get(e);if(!t)throw new Error("Illegal State: Could not find pipe "+e+" in template of "+this.component);return this.getOutputVar(t)},e.prototype.preprocessUpdateExpression=function(e){var t=this;return{sourceSpan:e.sourceSpan,context:e.context,value:$o({createLiteralArrayConverter:function(e){return function(e){var n=At(e);return t.options.fullTemplateTypeCheck?n:n.cast(me)}},createLiteralMapConverter:function(e){return function(n){var r=Tt(e.map(function(e,t){return{key:e.key,value:n[t],quoted:e.quoted}}));return t.options.fullTemplateTypeCheck?r:r.cast(me)}},createPipeConverter:function(e,n){return function(n){return(t.options.fullTemplateTypeCheck?Et(t.pipeOutputVar(e)):Et(t.getOutputVar(ae.Dynamic))).callMethod("transform",n)}}},e.value)}},e.prototype.visitNgContent=function(e,t){},e.prototype.visitText=function(e,t){},e.prototype.visitDirectiveProperty=function(e,t){},e.prototype.visitReference=function(e,t){},e.prototype.visitVariable=function(e,t){},e.prototype.visitEvent=function(e,t){},e.prototype.visitElementProperty=function(e,t){},e.prototype.visitAttr=function(e,t){},e}(),kf="class",Mf="style",Af=function(){return function(e,t){this.viewClassVar=e,this.rendererTypeVar=t}}(),Tf=function(){function e(e){this._reflector=e}return e.prototype.compileComponent=function(e,t,n,r,i){var a,s=this,u=0,c=function e(t,n){void 0===n&&(n=new Map);t.forEach(function(t){var r=new Set,i=new Set,o=void 0;t instanceof _d?(e(t.children,n),t.children.forEach(function(e){var t=n.get(e);t.staticQueryIds.forEach(function(e){return r.add(e)}),t.dynamicQueryIds.forEach(function(e){return i.add(e)})}),o=t.queryMatches):t instanceof yd&&(e(t.children,n),t.children.forEach(function(e){var t=n.get(e);t.staticQueryIds.forEach(function(e){return i.add(e)}),t.dynamicQueryIds.forEach(function(e){return i.add(e)})}),o=t.queryMatches),o&&o.forEach(function(e){return r.add(e.queryId)}),i.forEach(function(e){return r.delete(e)}),n.set(t,{staticQueryIds:r,dynamicQueryIds:i})});return n}(n),l=void 0;if(!t.isHost){var d=t.template,p=[];d.animations&&d.animations.length&&p.push(new Ze("animation",cp(e,d.animations),!0));var f=Et(Mn(t.type.reference));l=f.name,e.statements.push(f.set(St(gn.createRendererType2).callFn([new $e([new Ze("encapsulation",Pt(d.encapsulation),!1),new Ze("styles",r,!1),new Ze("data",new $e(p),!1)])])).toDeclStmt(kt(gn.RendererType2),[Ie.Final,Ie.Exported]))}var h=function(n){var r=u++;return new Lf(s._reflector,e,n,t,r,i,c,h)},m=h(null);return m.visitAll([],n),(a=e.statements).push.apply(a,Object(o.__spread)(m.build())),new Af(m.viewName,l)},e}(),Df=Et("_l"),If=Et("_v"),Pf=Et("_ck"),Rf=Et("_co"),Nf=Et("en"),Ff=Et("ad"),Lf=function(){function e(e,t,n,r,i,o,a,s){this.reflector=e,this.outputCtx=t,this.parent=n,this.component=r,this.embeddedViewIndex=i,this.usedPipes=o,this.staticQueryIds=a,this.viewBuilderFactory=s,this.nodes=[],this.purePipeNodeIndices=Object.create(null),this.refNodeIndices=Object.create(null),this.variables=[],this.children=[],this.compType=this.embeddedViewIndex>0?me:Mt(t.importExpr(this.component.type.reference)),this.viewName=kn(this.component.type.reference,this.embeddedViewIndex)}return e.prototype.visitAll=function(e,t){var n,r,i,o=this;if(this.variables=e,this.parent||this.usedPipes.forEach(function(e){e.pure&&(o.purePipeNodeIndices[e.name]=o._createPipe(null,e))}),!this.parent){var a=(n=this.staticQueryIds,r=new Set,i=new Set,Array.from(n.values()).forEach(function(e){e.staticQueryIds.forEach(function(e){return r.add(e)}),e.dynamicQueryIds.forEach(function(e){return i.add(e)})}),i.forEach(function(e){return r.delete(e)}),{staticQueryIds:r,dynamicQueryIds:i});this.component.viewQueries.forEach(function(e,t){var n=t+1,r=e.first?0:1,i=134217728|Uf(a,n,e.first);o.nodes.push(function(){return{sourceSpan:null,nodeFlags:i,nodeDef:St(gn.queryDef).callFn([Pt(i),Pt(n),new $e([new Ze(e.propertyName,Pt(r),!1)])])}})})}Ed(this,t),this.parent&&(0===t.length||function e(t){var n=t[t.length-1];if(n instanceof yd)return n.hasViewContainer;if(n instanceof _d)return s(n.name)&&n.children.length?e(n.children):n.hasViewContainer;return n instanceof Od}(t))&&this.nodes.push(function(){return{sourceSpan:null,nodeFlags:1,nodeDef:St(gn.anchorDef).callFn([Pt(0),et,et,Pt(0)])}})},e.prototype.build=function(e){void 0===e&&(e=[]),this.children.forEach(function(t){return t.build(e)});var t=this._createNodeExpressions(),n=t.updateRendererStmts,r=t.updateDirectivesStmts,i=t.nodeDefExprs,o=this._createUpdateFn(n),a=this._createUpdateFn(r),s=0;this.parent||this.component.changeDetection!==b.OnPush||(s|=2);var u=new it(this.viewName,[new Ge(Df.name)],[new at(St(gn.viewDef).callFn([Pt(s),At(i),a,o]))],kt(gn.ViewDefinition),0===this.embeddedViewIndex?[Ie.Exported]:[]);return e.push(u),e},e.prototype._createUpdateFn=function(e){var t;if(e.length>0){var n=[];!this.component.isHost&&yt(e).has(Rf.name)&&n.push(Rf.set(If.prop("component")).toDeclStmt(this.compType)),t=Dt([new Ge(Pf.name,ge),new Ge(If.name,ge)],Object(o.__spread)(n,e),ge)}else t=et;return t},e.prototype.visitNgContent=function(e,t){this.nodes.push(function(){return{sourceSpan:e.sourceSpan,nodeFlags:8,nodeDef:St(gn.ngContentDef).callFn([Pt(e.ngContentIndex),Pt(e.index)])}})},e.prototype.visitText=function(e,t){this.nodes.push(function(){return{sourceSpan:e.sourceSpan,nodeFlags:2,nodeDef:St(gn.textDef).callFn([Pt(-1),Pt(e.ngContentIndex),At([Pt(e.value)])])}})},e.prototype.visitBoundText=function(e,t){var n=this,r=this.nodes.length;this.nodes.push(null);var i=e.value.ast,o=i.expressions.map(function(t,i){return n._preprocessUpdateExpression({nodeIndex:r,bindingIndex:i,sourceSpan:e.sourceSpan,context:Rf,value:t})}),a=r;this.nodes[r]=function(){return{sourceSpan:e.sourceSpan,nodeFlags:2,nodeDef:St(gn.textDef).callFn([Pt(a),Pt(e.ngContentIndex),At(i.strings.map(function(e){return Pt(e)}))]),updateRenderer:o}}},e.prototype.visitEmbeddedTemplate=function(e,t){var n=this,r=this.nodes.length;this.nodes.push(null);var i=this._visitElementOrTemplate(r,e),o=i.flags,a=i.queryMatchesExpr,s=i.hostEvents,u=this.viewBuilderFactory(this);this.children.push(u),u.visitAll(e.variables,e.children);var c=this.nodes.length-r-1;this.nodes[r]=function(){return{sourceSpan:e.sourceSpan,nodeFlags:1|o,nodeDef:St(gn.anchorDef).callFn([Pt(o),a,Pt(e.ngContentIndex),Pt(c),n._createElementHandleEventFn(r,s),Et(u.viewName)])}}},e.prototype.visitElement=function(e,t){var n=this,r=this.nodes.length;this.nodes.push(null);var i=s(e.name)?null:e.name,a=this._visitElementOrTemplate(r,e),u=a.flags,c=a.usedEvents,l=a.queryMatchesExpr,d=a.hostBindings,p=a.hostEvents,f=[],h=[],m=[];if(i){var g=e.inputs.map(function(e){return{context:Rf,inputAst:e,dirAst:null}}).concat(d);g.length&&(h=g.map(function(e,t){return n._preprocessUpdateExpression({context:e.context,nodeIndex:r,bindingIndex:t,sourceSpan:e.inputAst.sourceSpan,value:e.inputAst.value})}),f=g.map(function(e){return function(e,t){var n=e.type;switch(n){case 1:return At([Pt(1),Pt(e.name),Pt(e.securityContext)]);case 0:return At([Pt(8),Pt(e.name),Pt(e.securityContext)]);case 4:var r=8|(t&&t.directive.isComponent?32:16);return At([Pt(r),Pt("@"+e.name),Pt(e.securityContext)]);case 2:return At([Pt(2),Pt(e.name),et]);case 3:return At([Pt(4),Pt(e.name),Pt(e.unit)]);default:var i=n;throw new Error("unexpected "+i)}}(e.inputAst,e.dirAst)})),m=c.map(function(e){var t=Object(o.__read)(e,2),n=t[0],r=t[1];return At([Pt(n),Pt(r)])})}Ed(this,e.children);var v=this.nodes.length-r-1,_=e.directives.find(function(e){return e.directive.isComponent}),y=et,b=et;_&&(b=this.outputCtx.importExpr(_.directive.componentViewType),y=this.outputCtx.importExpr(_.directive.rendererType));var j=r;this.nodes[r]=function(){return{sourceSpan:e.sourceSpan,nodeFlags:1|u,nodeDef:St(gn.elementDef).callFn([Pt(j),Pt(u),l,Pt(e.ngContentIndex),Pt(v),Pt(i),i?(t=e,o=Object.create(null),t.attrs.forEach(function(e){o[e.name]=e.value}),t.directives.forEach(function(e){Object.keys(e.directive.hostAttributes).forEach(function(t){var n,r,i=e.directive.hostAttributes[t],a=o[t];o[t]=null!=a?(r=i,(n=t)==kf||n==Mf?a+" "+r:r):i})}),At(Object.keys(o).sort().map(function(e){return At([Pt(e),Pt(o[e])])}))):et,f.length?At(f):et,m.length?At(m):et,n._createElementHandleEventFn(r,p),b,y]),updateRenderer:h};var t,o}},e.prototype._visitElementOrTemplate=function(e,t){var n=this,r=0;t.hasViewContainer&&(r|=16777216);var i=new Map;t.outputs.forEach(function(e){var t=zf(e,null),n=t.name,r=t.target;i.set(Hf(r,n),[r,n])}),t.directives.forEach(function(e){e.hostEvents.forEach(function(t){var n=zf(t,e),r=n.name,o=n.target;i.set(Hf(o,r),[o,r])})});var a=[],s=[];this._visitComponentFactoryResolverProvider(t.directives),t.providers.forEach(function(r,u){var c=void 0,l=void 0;if(t.directives.forEach(function(e,t){e.directive.type.reference===In(r.token)&&(c=e,l=t)}),c){var d=n._visitDirective(r,c,l,e,t.references,t.queryMatches,i,n.staticQueryIds.get(t)),p=d.hostBindings,f=d.hostEvents;a.push.apply(a,Object(o.__spread)(p)),s.push.apply(s,Object(o.__spread)(f))}else n._visitProvider(r,t.queryMatches)});var u=[];return t.queryMatches.forEach(function(e){var t=void 0;In(e.value)===n.reflector.resolveExternalReference(gn.ElementRef)?t=0:In(e.value)===n.reflector.resolveExternalReference(gn.ViewContainerRef)?t=3:In(e.value)===n.reflector.resolveExternalReference(gn.TemplateRef)&&(t=2),null!=t&&u.push(At([Pt(e.queryId),Pt(t)]))}),t.references.forEach(function(t){var r=void 0;t.value?In(t.value)===n.reflector.resolveExternalReference(gn.TemplateRef)&&(r=2):r=1,null!=r&&(n.refNodeIndices[t.name]=e,u.push(At([Pt(t.name),Pt(r)])))}),t.outputs.forEach(function(e){s.push({context:Rf,eventAst:e,dirAst:null})}),{flags:r,usedEvents:Array.from(i.values()),queryMatchesExpr:u.length?At(u):et,hostBindings:a,hostEvents:s}},e.prototype._visitDirective=function(e,t,n,r,i,o,a,s){var u=this,c=this.nodes.length;this.nodes.push(null),t.directive.queries.forEach(function(e,n){var r=t.contentQueryStartId+n,i=67108864|Uf(s,r,e.first),o=e.first?0:1;u.nodes.push(function(){return{sourceSpan:t.sourceSpan,nodeFlags:i,nodeDef:St(gn.queryDef).callFn([Pt(i),Pt(r),new $e([new Ze(e.propertyName,Pt(o),!1)])])}})});var l=this.nodes.length-c-1,d=this._visitProviderOrDirective(e,o),p=d.flags,f=d.queryMatchExprs,h=d.providerExpr,m=d.depsExpr;i.forEach(function(t){t.value&&In(t.value)===In(e.token)&&(u.refNodeIndices[t.name]=c,f.push(At([Pt(t.name),Pt(4)])))}),t.directive.isComponent&&(p|=32768);var g=t.inputs.map(function(e,t){var n=At([Pt(t),Pt(e.directiveName)]);return new Ze(e.directiveName,n,!1)}),v=[],_=t.directive;Object.keys(_.outputs).forEach(function(e){var t=_.outputs[e];a.has(t)&&v.push(new Ze(e,Pt(t),!1))});var y=[];(t.inputs.length||(327680&p)>0)&&(y=t.inputs.map(function(e,t){return u._preprocessUpdateExpression({nodeIndex:c,bindingIndex:t,sourceSpan:e.sourceSpan,context:Rf,value:e.value})}));var b=St(gn.nodeValue).callFn([If,Pt(c)]),j=t.hostProperties.map(function(e){return{context:b,dirAst:t,inputAst:e}}),x=t.hostEvents.map(function(e){return{context:b,eventAst:e,dirAst:t}}),O=c;return this.nodes[c]=function(){return{sourceSpan:t.sourceSpan,nodeFlags:16384|p,nodeDef:St(gn.directiveDef).callFn([Pt(O),Pt(p),f.length?At(f):et,Pt(l),h,m,g.length?new $e(g):et,v.length?new $e(v):et]),updateDirectives:y,directive:t.directive.type}},{hostBindings:j,hostEvents:x}},e.prototype._visitProvider=function(e,t){this._addProviderNode(this._visitProviderOrDirective(e,t))},e.prototype._visitComponentFactoryResolverProvider=function(e){var t=e.find(function(e){return e.directive.isComponent});if(t&&t.directive.entryComponents.length){var n=Wp(this.reflector,this.outputCtx,8192,t.directive.entryComponents),r=n.providerExpr,i=n.depsExpr,o=n.flags,a=n.tokenExpr;this._addProviderNode({providerExpr:r,depsExpr:i,flags:o,tokenExpr:a,queryMatchExprs:[],sourceSpan:t.sourceSpan})}},e.prototype._addProviderNode=function(e){this.nodes.length;this.nodes.push(function(){return{sourceSpan:e.sourceSpan,nodeFlags:e.flags,nodeDef:St(gn.providerDef).callFn([Pt(e.flags),e.queryMatchExprs.length?At(e.queryMatchExprs):et,e.tokenExpr,e.providerExpr,e.depsExpr])}})},e.prototype._visitProviderOrDirective=function(e,t){var n=[];t.forEach(function(t){In(t.value)===In(e.token)&&n.push(At([Pt(t.queryId),Pt(4)]))});var r=zp(this.outputCtx,e),i=r.providerExpr,o=r.depsExpr,a=r.flags,s=r.tokenExpr;return{flags:0|a,queryMatchExprs:n,providerExpr:i,depsExpr:o,tokenExpr:s,sourceSpan:e.sourceSpan}},e.prototype.getLocal=function(e){if(e==Qo.event.name)return Qo.event;for(var t=If,n=this;n;n=n.parent,t=t.prop("parent").cast(me)){var r=n.refNodeIndices[e];if(null!=r)return St(gn.nodeValue).callFn([t,Pt(r)]);var i=n.variables.find(function(t){return t.name===e});if(i){var o=i.value||"$implicit";return t.prop("context").prop(o)}}return null},e.prototype._createLiteralArrayConverter=function(e,t){if(0===t){var n=St(gn.EMPTY_ARRAY);return function(){return n}}var r=this.nodes.length;return this.nodes.push(function(){return{sourceSpan:e,nodeFlags:32,nodeDef:St(gn.pureArrayDef).callFn([Pt(r),Pt(t)])}}),function(e){return Vf(r,e)}},e.prototype._createLiteralMapConverter=function(e,t){if(0===t.length){var n=St(gn.EMPTY_MAP);return function(){return n}}var r=Tt(t.map(function(e,t){return Object(o.__assign)({},e,{value:Pt(t)})})),i=this.nodes.length;return this.nodes.push(function(){return{sourceSpan:e,nodeFlags:64,nodeDef:St(gn.pureObjectDef).callFn([Pt(i),r])}}),function(e){return Vf(i,e)}},e.prototype._createPipeConverter=function(e,t,n){var r=this.usedPipes.find(function(e){return e.name===t});if(r.pure){var i=this.nodes.length;this.nodes.push(function(){return{sourceSpan:e.sourceSpan,nodeFlags:128,nodeDef:St(gn.purePipeDef).callFn([Pt(i),Pt(n)])}});for(var o=If,a=this;a.parent;)a=a.parent,o=o.prop("parent").cast(me);var s=a.purePipeNodeIndices[t],u=St(gn.nodeValue).callFn([o,Pt(s)]);return function(t){return Bf(e.nodeIndex,e.bindingIndex,Vf(i,[u].concat(t)))}}var c=this._createPipe(e.sourceSpan,r),l=St(gn.nodeValue).callFn([If,Pt(c)]);return function(t){return Bf(e.nodeIndex,e.bindingIndex,l.callMethod("transform",t))}},e.prototype._createPipe=function(e,t){var n=this,r=this.nodes.length,i=0;t.type.lifecycleHooks.forEach(function(e){e===pa.OnDestroy&&(i|=Gp(e))});var o=t.type.diDeps.map(function(e){return qp(n.outputCtx,e)});return this.nodes.push(function(){return{sourceSpan:e,nodeFlags:16,nodeDef:St(gn.pipeDef).callFn([Pt(i),n.outputCtx.importExpr(t.type.reference),At(o)])}}),r},e.prototype._preprocessUpdateExpression=function(e){var t=this;return{nodeIndex:e.nodeIndex,bindingIndex:e.bindingIndex,sourceSpan:e.sourceSpan,context:e.context,value:$o({createLiteralArrayConverter:function(n){return t._createLiteralArrayConverter(e.sourceSpan,n)},createLiteralMapConverter:function(n){return t._createLiteralMapConverter(e.sourceSpan,n)},createPipeConverter:function(n,r){return t._createPipeConverter(e,n,r)}},e.value)}},e.prototype._createNodeExpressions=function(){var e=this,t=0,n=[],r=[],i=this.nodes.map(function(e,t){var i=e(),s=i.nodeDef,u=i.nodeFlags,c=i.updateDirectives,l=i.updateRenderer,d=i.sourceSpan;return l&&n.push.apply(n,Object(o.__spread)(a(t,d,l,!1))),c&&r.push.apply(r,Object(o.__spread)(a(t,d,c,(327680&u)>0))),wt(3&u?new Je([Df.callFn([]).callFn([]),s]):s,d)});return{updateRendererStmts:n,updateDirectivesStmts:r,nodeDefExprs:i};function a(n,r,i,a){var s=[],u=i.map(function(n){var r=n.sourceSpan,i=n.context,a=n.value,u=""+t++,c=ia(i===Rf?e:null,i,a,u,Jo.General),l=c.stmts,d=c.currValExpr;return s.push.apply(s,Object(o.__spread)(l.map(function(e){return Ot(e,r)}))),wt(d,r)});return(i.length||a)&&s.push(Ot(Vf(n,u).toStmt(),r)),s}},e.prototype._createElementHandleEventFn=function(e,t){var n,r=this,i=[],a=0;if(t.forEach(function(e){var t=e.context,n=e.eventAst,o=e.dirAst,s=""+a++,u=Zo(t===Rf?r:null,t,n.handler,s),c=u.stmts,l=u.allowDefault,d=c;l&&d.push(Ff.set(l.and(Ff)).toStmt());var p=zf(n,o),f=Hf(p.target,p.name);i.push(Ot(new pt(Pt(f).identical(Nf),d),n.sourceSpan))}),i.length>0){var s=[Ff.set(Pt(!0)).toDeclStmt(ve)];!this.component.isHost&&yt(i).has(Rf.name)&&s.push(Rf.set(If.prop("component")).toDeclStmt(this.compType)),n=Dt([new Ge(If.name,ge),new Ge(Nf.name,ge),new Ge(Qo.event.name,ge)],Object(o.__spread)(s,i,[new at(Ff)]),ge)}else n=et;return n},e.prototype.visitDirective=function(e,t){},e.prototype.visitDirectiveProperty=function(e,t){},e.prototype.visitReference=function(e,t){},e.prototype.visitVariable=function(e,t){},e.prototype.visitEvent=function(e,t){},e.prototype.visitElementProperty=function(e,t){},e.prototype.visitAttr=function(e,t){},e}();function Vf(e,t){return t.length>10?Pf.callFn([If,Pt(e),Pt(1),At(t)]):Pf.callFn(Object(o.__spread)([If,Pt(e),Pt(0)],t))}function Bf(e,t,n){return St(gn.unwrapValue).callFn([If,Pt(e),Pt(t),n])}function zf(e,t){return e.isAnimation?{name:"@"+e.name+"."+e.phase,target:t&&t.directive.isComponent?"component":null}:e}function Uf(e,t,n){var r=0;return!n||!e.staticQueryIds.has(t)&&e.dynamicQueryIds.has(t)?r|=536870912:r|=268435456,r}function Hf(e,t){return e?e+":"+t:t}var qf=function(){function e(e,t,n,r){void 0===r&&(r=null),this._htmlParser=e,this._implicitTags=t,this._implicitAttrs=n,this._locale=r,this._messages=[]}return e.prototype.updateFromTemplate=function(e,t,n){var r,i=this._htmlParser.parse(e,t,!0,n);if(i.errors.length)return i.errors;var a,s,u,c,l=(a=i.rootNodes,s=n,u=this._implicitTags,c=this._implicitAttrs,new qd(u,c).extract(a,s));return l.errors.length?l.errors:((r=this._messages).push.apply(r,Object(o.__spread)(l.messages)),[])},e.prototype.getMessages=function(){return this._messages},e.prototype.write=function(e,t){var n={},r=new Gf;this._messages.forEach(function(t){var r,i=e.digest(t);n.hasOwnProperty(i)?(r=n[i].sources).push.apply(r,Object(o.__spread)(t.sources)):n[i]=t});var i=Object.keys(n).map(function(i){var o=e.createNameMapper(n[i]),a=n[i],s=o?r.convert(a.nodes,o):a.nodes,u=new Zn(s,{},{},a.meaning,a.description,i);return u.sources=a.sources,t&&u.sources.forEach(function(e){return e.filePath=t(e.filePath)}),u});return e.write(i,this._locale)},e}(),Gf=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(o.__extends)(t,e),t.prototype.convert=function(e,t){var n=this;return t?e.map(function(e){return e.visit(n,t)}):e},t.prototype.visitTagPlaceholder=function(e,t){var n=this,r=t.toPublicName(e.startName),i=e.closeName?t.toPublicName(e.closeName):e.closeName,o=e.children.map(function(e){return e.visit(n,t)});return new tr(e.tag,e.attrs,r,i,o,e.isVoid,e.sourceSpan)},t.prototype.visitPlaceholder=function(e,t){return new nr(e.value,t.toPublicName(e.name),e.sourceSpan)},t.prototype.visitIcuPlaceholder=function(e,t){return new rr(e.value,t.toPublicName(e.name),e.sourceSpan)},t}(ir),Wf=function(){function e(e,t,n){this.srcFileUrl=e,this.genFileUrl=t,"string"==typeof n?(this.source=n,this.stmts=null):(this.source=null,this.stmts=n)}return e.prototype.isEquivalent=function(e){return this.genFileUrl===e.genFileUrl&&(this.source?this.source===e.source:null!=e.stmts&&je(this.stmts,e.stmts))},e}();function Yf(e,t){if(void 0===t&&(t=""),!e.stmts)throw new Error("Illegal state: No stmts present on GeneratedFile "+e.genFileUrl);return(new $p).emitStatements(e.genFileUrl,e.stmts,t)}function Kf(e,t){var n,r,i,a,s=[];try{for(var u=Object(o.__values)(e.transitiveModule.providers),c=u.next();!c.done;c=u.next()){var l=c.value,d=l.provider,p=l.module;if(In(d.token)===t.ROUTES){var f=Qf(d.useValue);try{for(var h=Object(o.__values)(f),m=h.next();!m.done;m=h.next()){var g=m.value;s.push(Xf(g,t,p.reference))}}catch(e){i={error:e}}finally{try{m&&!m.done&&(a=h.return)&&a.call(h)}finally{if(i)throw i.error}}}}}catch(e){n={error:e}}finally{try{c&&!c.done&&(r=u.return)&&r.call(u)}finally{if(n)throw n.error}}return s}function Qf(e,t){var n,r;if(void 0===t&&(t=[]),"string"==typeof e)t.push(e);else if(Array.isArray(e))try{for(var i=Object(o.__values)(e),a=i.next();!a.done;a=i.next()){Qf(a.value,t)}}catch(e){n={error:e}}finally{try{a&&!a.done&&(r=i.return)&&r.call(i)}finally{if(n)throw n.error}}else e.loadChildren?Qf(e.loadChildren,t):e.children&&Qf(e.children,t);return t}function Xf(e,t,n){var r=Object(o.__read)(e.split("#"),2),i=r[0],a=r[1],s=t.resolveExternalReference({moduleName:i,name:a},n?n.filePath:void 0);return{route:e,module:n||s,referencedModule:s}}var Zf=/^(?!.*\.d\.ts$).*\.ts$/,$f=function(){return function(e,t){this.symbol=e,this.metadata=t}}(),Jf=function(){function e(e,t,n,r){this.host=e,this.staticSymbolCache=t,this.summaryResolver=n,this.errorRecorder=r,this.metadataCache=new Map,this.resolvedSymbols=new Map,this.resolvedFilePaths=new Set,this.importAs=new Map,this.symbolResourcePaths=new Map,this.symbolFromFile=new Map,this.knownFileNameToModuleNames=new Map}return e.prototype.resolveSymbol=function(e){if(e.members.length>0)return this._resolveSymbolMembers(e);var t=this._resolveSymbolFromSummary(e);if(t)return t;var n=this.resolvedSymbols.get(e);return n||(this._createSymbolsOf(e.filePath),this.resolvedSymbols.get(e))},e.prototype.getImportAs=function(e,t){if(void 0===t&&(t=!0),e.members.length){var n=this.getStaticSymbol(e.filePath,e.name);return(i=this.getImportAs(n,t))?this.getStaticSymbol(i.filePath,i.name,e.members):null}var r=e.filePath.replace(mp,".");if(r!==e.filePath){var i,o=e.name.replace(gp,"");n=this.getStaticSymbol(r,o,e.members);return(i=this.getImportAs(n,t))?this.getStaticSymbol(xp(i.filePath),Op(i.name),n.members):null}var a=t&&this.summaryResolver.getImportAs(e)||null;return a||(a=this.importAs.get(e)),a},e.prototype.getResourcePath=function(e){return this.symbolResourcePaths.get(e)||e.filePath},e.prototype.getTypeArity=function(e){if(t=e.filePath,hp.test(t))return null;for(var t,n=th(this.resolveSymbol(e));n&&n.metadata instanceof yn;)n=th(this.resolveSymbol(n.metadata));return n&&n.metadata&&n.metadata.arity||null},e.prototype.getKnownModuleName=function(e){return this.knownFileNameToModuleNames.get(e)||null},e.prototype.recordImportAs=function(e,t){e.assertNoMembers(),t.assertNoMembers(),this.importAs.set(e,t)},e.prototype.recordModuleNameForFileName=function(e,t){this.knownFileNameToModuleNames.set(e,t)},e.prototype.invalidateFile=function(e){var t,n;this.metadataCache.delete(e),this.resolvedFilePaths.delete(e);var r=this.symbolFromFile.get(e);if(r){this.symbolFromFile.delete(e);try{for(var i=Object(o.__values)(r),a=i.next();!a.done;a=i.next()){var s=a.value;this.resolvedSymbols.delete(s),this.importAs.delete(s),this.symbolResourcePaths.delete(s)}}catch(e){t={error:e}}finally{try{a&&!a.done&&(n=i.return)&&n.call(i)}finally{if(t)throw t.error}}}},e.prototype.ignoreErrorsFor=function(e){var t=this.errorRecorder;this.errorRecorder=function(){};try{return e()}finally{this.errorRecorder=t}},e.prototype._resolveSymbolMembers=function(e){var t=e.members,n=this.resolveSymbol(this.getStaticSymbol(e.filePath,e.name));if(!n)return null;var r=th(n.metadata);if(r instanceof yn)return new $f(e,this.getStaticSymbol(r.filePath,r.name,t));if(!r||"class"!==r.__symbolic){for(var i=r,o=0;o<t.length&&i;o++)i=i[t[o]];return new $f(e,i)}return r.statics&&1===t.length?new $f(e,r.statics[t[0]]):null},e.prototype._resolveSymbolFromSummary=function(e){var t=this.summaryResolver.resolveSummary(e);return t?new $f(e,t.metadata):null},e.prototype.getStaticSymbol=function(e,t,n){return this.staticSymbolCache.get(e,t,n)},e.prototype.hasDecorators=function(e){var t=this.getModuleMetadata(e);return!!t.metadata&&Object.keys(t.metadata).some(function(e){var n=t.metadata[e];return n&&"class"===n.__symbolic&&n.decorators})},e.prototype.getSymbolsOf=function(e){var t=this.summaryResolver.getSymbolsOf(e);if(t)return t;this._createSymbolsOf(e);var n=[];return this.resolvedSymbols.forEach(function(t){t.symbol.filePath===e&&n.push(t.symbol)}),n},e.prototype._createSymbolsOf=function(e){var t,n,r=this;if(!this.resolvedFilePaths.has(e)){this.resolvedFilePaths.add(e);var i=[],a=this.getModuleMetadata(e);if(a.importAs&&this.knownFileNameToModuleNames.set(e,a.importAs),a.exports){var s=function(t){if(t.export)t.export.forEach(function(n){var o,a=o=eh(o="string"==typeof n?n:n.as);"string"!=typeof n&&(a=eh(n.name));var s=r.resolveModule(t.from,e);if(s){var u=r.getStaticSymbol(s,a),c=r.getStaticSymbol(e,o);i.push(r.createExport(c,u))}});else{var n=u.resolveModule(t.from,e);if(n)u.getSymbolsOf(n).forEach(function(t){var n=r.getStaticSymbol(e,t.name);i.push(r.createExport(n,t))})}},u=this;try{for(var c=Object(o.__values)(a.exports),l=c.next();!l.done;l=c.next()){s(l.value)}}catch(e){t={error:e}}finally{try{l&&!l.done&&(n=c.return)&&n.call(c)}finally{if(t)throw t.error}}}if(a.metadata){var d=new Set(Object.keys(a.metadata).map(eh)),p=a.origins||{};Object.keys(a.metadata).forEach(function(t){var n=a.metadata[t],o=eh(t),s=r.getStaticSymbol(e,o),u=p.hasOwnProperty(t)&&p[t];if(u){var c=r.resolveModule(u,e);c?r.symbolResourcePaths.set(s,c):r.reportError(new Error("Couldn't resolve original symbol for "+u+" from "+r.host.getOutputName(e)))}i.push(r.createResolvedSymbol(s,e,d,n))})}i.forEach(function(e){return r.resolvedSymbols.set(e.symbol,e)}),this.symbolFromFile.set(e,i.map(function(e){return e.symbol}))}},e.prototype.createResolvedSymbol=function(e,t,n,r){var i,a=this,s=Zf.test(e.filePath);if(this.summaryResolver.isLibraryFile(e.filePath)&&!s&&r&&"class"===r.__symbolic){var u={__symbolic:"class",arity:r.arity};return new $f(e,u)}var c=function(){return i||(i=a.host.getOutputName(t.replace(/((\.ts)|(\.d\.ts)|)$/,".ts").replace(/^.*node_modules[/\\]/,""))),i},l=this,d=Vt(r,new(function(r){function i(){return null!==r&&r.apply(this,arguments)||this}return Object(o.__extends)(i,r),i.prototype.visitStringMap=function(i,a){var s=i.__symbolic;if("function"===s){var u=a.length;a.push.apply(a,Object(o.__spread)(i.parameters||[]));var d=r.prototype.visitStringMap.call(this,i,a);return a.length=u,d}if("reference"!==s)return"error"===s?Object(o.__assign)({},i,{fileName:c()}):r.prototype.visitStringMap.call(this,i,a);var p=i.module,f=i.name?eh(i.name):i.name;if(!f)return null;var h=void 0;return p?(h=l.resolveModule(p,e.filePath))?{__symbolic:"resolved",symbol:l.getStaticSymbol(h,f),line:i.line,character:i.character,fileName:c()}:{__symbolic:"error",message:"Could not resolve "+p+" relative to "+l.host.getMetadataFor(e.filePath)+".",line:i.line,character:i.character,fileName:c()}:a.indexOf(f)>=0?{__symbolic:"reference",name:f}:n.has(f)?l.getStaticSymbol(t,f):void 0},i}(Ut)),[]),p=th(d);return p instanceof yn?this.createExport(e,p):new $f(e,d)},e.prototype.createExport=function(e,t){return e.assertNoMembers(),t.assertNoMembers(),this.summaryResolver.isLibraryFile(e.filePath)&&this.summaryResolver.isLibraryFile(t.filePath)&&this.importAs.set(t,this.getImportAs(e)||e),new $f(e,t)},e.prototype.reportError=function(e,t,n){if(!this.errorRecorder)throw e;this.errorRecorder(e,t&&t.filePath||n)},e.prototype.getModuleMetadata=function(e){var t=this.metadataCache.get(e);if(!t){var n=this.host.getMetadataFor(e);if(n){var r=-1;n.forEach(function(e){e&&e.version>r&&(r=e.version,t=e)})}if(t||(t={__symbolic:"module",version:4,module:e,metadata:{}}),4!=t.version){var i=2==t.version?"Unsupported metadata version "+t.version+" for module "+e+". This module should be compiled with a newer version of ngc":"Metadata version mismatch for module "+this.host.getOutputName(e)+", found version "+t.version+", expected 4";this.reportError(new Error(i))}this.metadataCache.set(e,t)}return t},e.prototype.getSymbolByModule=function(e,t,n){var r=this.resolveModule(e,n);return r?this.getStaticSymbol(r,t):(this.reportError(new Error("Could not resolve module "+e+(n?" relative to "+this.host.getOutputName(n):""))),this.getStaticSymbol("ERROR:"+e,t))},e.prototype.resolveModule=function(e,t){try{return this.host.moduleNameToFileName(e,t)}catch(n){console.error("Could not resolve module '"+e+"' relative to file "+t),this.reportError(n,void 0,t)}return null},e}();function eh(e){return e.startsWith("___")?e.substr(1):e}function th(e){return e&&"resolved"===e.__symbolic?e.symbol:e}function nh(e,t,n){var r=Op(t.name);e.statements.push(Dt([],[new at(n)],new fe(me)).toDeclStmt(r,[Ie.Final,Ie.Exported]))}var rh=function(e){function t(t,n,r){var i=e.call(this)||this;return i.symbolResolver=t,i.summaryResolver=n,i.srcFileName=r,i.symbols=[],i.indexBySymbol=new Map,i.reexportedBy=new Map,i.processedSummaryBySymbol=new Map,i.processedSummaries=[],i.unprocessedSymbolSummariesBySymbol=new Map,i.moduleName=t.getKnownModuleName(r),i}return Object(o.__extends)(t,e),t.prototype.addSummary=function(e){var t,n,r=this,i=this.unprocessedSymbolSummariesBySymbol.get(e.symbol),o=this.processedSummaryBySymbol.get(e.symbol);if(i||(i={symbol:e.symbol,metadata:void 0},this.unprocessedSymbolSummariesBySymbol.set(e.symbol,i),o={symbol:this.processValue(e.symbol,0)},this.processedSummaries.push(o),this.processedSummaryBySymbol.set(e.symbol,o)),!i.metadata&&e.metadata){var a=e.metadata||{};if("class"===a.__symbolic){var s={};Object.keys(a).forEach(function(e){"decorators"!==e&&(s[e]=a[e])}),a=s}else ah(a)&&(ah(n=a)&&th(n.expression)instanceof yn||ah(t=a)&&t.expression&&"select"===t.expression.__symbolic&&th(t.expression.expression)instanceof yn||(a={__symbolic:"error",message:"Complex function calls are not supported."}));if(i.metadata=a,o.metadata=this.processValue(a,1),a instanceof yn&&this.summaryResolver.isLibraryFile(a.filePath)){var u=this.symbols[this.indexBySymbol.get(a)];Cp(u.name)||this.reexportedBy.set(u,e.symbol)}}if(!i.type&&e.type&&(i.type=e.type,o.type=this.processValue(e.type,0),e.type.summaryKind===On.NgModule)){var c=e.type;c.exportedDirectives.concat(c.exportedPipes).forEach(function(e){var t=e.reference;if(r.summaryResolver.isLibraryFile(t.filePath)&&!r.unprocessedSymbolSummariesBySymbol.has(t)){var n=r.summaryResolver.resolveSummary(t);n&&r.addSummary(n)}})}},t.prototype.serialize=function(){var e=this,t=[];return{json:JSON.stringify({moduleName:this.moduleName,summaries:this.processedSummaries,symbols:this.symbols.map(function(n,r){n.assertNoMembers();var i=void 0;if(e.summaryResolver.isLibraryFile(n.filePath)){var o=e.reexportedBy.get(n);if(o)i=e.indexBySymbol.get(o);else{var a=e.unprocessedSymbolSummariesBySymbol.get(n);a&&a.metadata&&"interface"===a.metadata.__symbolic||(i=n.name+"_"+r,t.push({symbol:n,exportAs:i}))}}return{__symbol:r,name:n.name,filePath:e.summaryResolver.toSummaryFileName(n.filePath,e.srcFileName),importAs:i}})}),exportAs:t}},t.prototype.processValue=function(e,t){return Vt(e,this,t)},t.prototype.visitOther=function(e,t){if(e instanceof yn){var n=this.symbolResolver.getStaticSymbol(e.filePath,e.name);return{__symbol:this.visitStaticSymbol(n,t),members:e.members}}},t.prototype.visitStringMap=function(t,n){return"resolved"===t.__symbolic?Vt(t.symbol,this,n):("error"===t.__symbolic&&(delete t.line,delete t.character),e.prototype.visitStringMap.call(this,t,n))},t.prototype.visitStaticSymbol=function(e,t){var n=this.indexBySymbol.get(e),r=null;if(1&t&&this.summaryResolver.isLibraryFile(e.filePath)){if(this.unprocessedSymbolSummariesBySymbol.has(e))return n;(r=this.loadSummary(e))&&r.metadata instanceof yn&&(n=this.visitStaticSymbol(r.metadata,t),r=null)}else if(null!=n)return n;return null==n&&(n=this.symbols.length,this.symbols.push(e)),this.indexBySymbol.set(e,n),r&&this.addSummary(r),n},t.prototype.loadSummary=function(e){var t=this.summaryResolver.resolveSummary(e);if(!t){var n=this.symbolResolver.resolveSymbol(e);n&&(t={symbol:n.symbol,metadata:n.metadata})}return t},t}(Ut),ih=function(){function e(e,t,n){this.outputCtx=e,this.symbolResolver=t,this.summaryResolver=n,this.data=[]}return e.prototype.addSourceType=function(e,t){this.data.push({summary:e,metadata:t,isLibrary:!1})},e.prototype.addLibType=function(e){this.data.push({summary:e,metadata:null,isLibrary:!0})},e.prototype.serialize=function(e){var t,n,r,i,a,s,u=this,c=new Map;try{for(var l=Object(o.__values)(e),d=l.next();!d.done;d=l.next()){var p=d.value,f=p.symbol,h=p.exportAs;c.set(f,h)}}catch(e){t={error:e}}finally{try{d&&!d.done&&(n=l.return)&&n.call(l)}finally{if(t)throw t.error}}var m=new Set;try{for(var g=Object(o.__values)(this.data),v=g.next();!v.done;v=g.next()){var _=v.value,y=_.summary,b=_.metadata,j=_.isLibrary;if(y.summaryKind===On.NgModule){m.add(y.type.reference);var x=y;try{for(var O=Object(o.__values)(x.modules),w=O.next();!w.done;w=O.next()){var C=w.value;m.add(C.reference)}}catch(e){a={error:e}}finally{try{w&&!w.done&&(s=O.return)&&s.call(O)}finally{if(a)throw a.error}}}if(!j){Op(y.type.reference.name);nh(this.outputCtx,y.type.reference,this.serializeSummaryWithDeps(y,b))}}}catch(e){r={error:e}}finally{try{v&&!v.done&&(i=g.return)&&i.call(g)}finally{if(r)throw r.error}}m.forEach(function(e){if(u.summaryResolver.isLibraryFile(e.filePath)){var t=Op(c.get(e)||e.name);u.outputCtx.statements.push(Et(t).set(u.serializeSummaryRef(e)).toDeclStmt(null,[Ie.Exported]))}})},e.prototype.serializeSummaryWithDeps=function(e,t){var n=this,r=[this.serializeSummary(e)],i=[];if(t instanceof Vn)r.push.apply(r,Object(o.__spread)(t.declaredDirectives.concat(t.declaredPipes).map(function(e){return e.reference}).concat(t.transitiveModule.modules.map(function(e){return e.reference}).filter(function(e){return e!==t.type.reference})).map(function(e){return n.serializeSummaryRef(e)}))),i=t.providers;else if(e.summaryKind===On.Directive){var a=e;i=a.providers.concat(a.viewProviders)}return r.push.apply(r,Object(o.__spread)(i.filter(function(e){return!!e.useClass}).map(function(e){return n.serializeSummary({summaryKind:On.Injectable,type:e.useClass})}))),At(r)},e.prototype.serializeSummaryRef=function(e){var t=this.symbolResolver.getStaticSymbol(xp(e.filePath),Op(e.name));return this.outputCtx.importExpr(t)},e.prototype.serializeSummary=function(e){var t=this.outputCtx;return Vt(e,new(function(){function e(){}return e.prototype.visitArray=function(e,t){var n=this;return At(e.map(function(e){return Vt(e,n,t)}))},e.prototype.visitStringMap=function(e,t){var n=this;return new $e(Object.keys(e).map(function(r){return new Ze(r,Vt(e[r],n,t),!1)}))},e.prototype.visitPrimitive=function(e,t){return Pt(e)},e.prototype.visitOther=function(e,n){if(e instanceof yn)return t.importExpr(e);throw new Error("Illegal State: Encountered value "+e)},e}()),null)},e}(),oh=function(e){function t(t,n){var r=e.call(this)||this;return r.symbolCache=t,r.summaryResolver=n,r}return Object(o.__extends)(t,e),t.prototype.deserialize=function(e,t){var n=this,r=JSON.parse(t),i=[];this.symbols=r.symbols.map(function(t){return n.symbolCache.get(n.summaryResolver.fromSummaryFileName(t.filePath,e),t.name)}),r.symbols.forEach(function(t,r){var o=n.symbols[r],a=t.importAs;"number"==typeof a?i.push({symbol:o,importAs:n.symbols[a]}):"string"==typeof a&&i.push({symbol:o,importAs:n.symbolCache.get(vp(e),a)})});var o=Vt(r.summaries,this,null);return{moduleName:r.moduleName,summaries:o,importAs:i}},t.prototype.visitStringMap=function(t,n){if("__symbol"in t){var r=this.symbols[t.__symbol],i=t.members;return i.length?this.symbolCache.get(r.filePath,r.name,i):r}return e.prototype.visitStringMap.call(this,t,n)},t}(Ut);function ah(e){return e&&"call"===e.__symbolic}var sh=function(){function e(e,t,n,r,i,o,a,s,u,c,l,d,p,f){this._config=e,this._options=t,this._host=n,this.reflector=r,this._metadataResolver=i,this._templateParser=o,this._styleCompiler=a,this._viewCompiler=s,this._typeCheckCompiler=u,this._ngModuleCompiler=c,this._injectableCompiler=l,this._outputEmitter=d,this._summaryResolver=p,this._symbolResolver=f,this._templateAstCache=new Map,this._analyzedFiles=new Map,this._analyzedFilesForInjectables=new Map}return e.prototype.clearCache=function(){this._metadataResolver.clearCache()},e.prototype.analyzeModulesSync=function(e){var t=this,n=ph(e,this._host,this._symbolResolver,this._metadataResolver);return n.ngModules.forEach(function(e){return t._metadataResolver.loadNgModuleDirectiveAndPipeMetadata(e.type.reference,!0)}),n},e.prototype.analyzeModulesAsync=function(e){var t=this,n=ph(e,this._host,this._symbolResolver,this._metadataResolver);return Promise.all(n.ngModules.map(function(e){return t._metadataResolver.loadNgModuleDirectiveAndPipeMetadata(e.type.reference,!1)})).then(function(){return n})},e.prototype._analyzeFile=function(e){var t=this._analyzedFiles.get(e);return t||(t=hh(this._host,this._symbolResolver,this._metadataResolver,e),this._analyzedFiles.set(e,t)),t},e.prototype._analyzeFileForInjectables=function(e){var t=this._analyzedFilesForInjectables.get(e);return t||(t=mh(this._host,this._symbolResolver,this._metadataResolver,e),this._analyzedFilesForInjectables.set(e,t)),t},e.prototype.findGeneratedFileNames=function(e){var t=this,n=[],r=this._analyzeFile(e);(this._options.allowEmptyCodegenFiles||r.directives.length||r.pipes.length||r.injectables.length||r.ngModules.length||r.exportsNonSourceFiles)&&(n.push(vp(r.fileName,!0)),this._options.enableSummariesForJit&&n.push(xp(r.fileName,!0)));var i=bp(yp(r.fileName,!0)[1]);return r.directives.forEach(function(e){var o=t._metadataResolver.getNonNormalizedDirectiveMetadata(e).metadata;o.isComponent&&o.template.styleUrls.forEach(function(e){var a=t._host.resourceNameToFileName(e,r.fileName);if(!a)throw Yt("Couldn't resolve resource "+e+" relative to "+r.fileName);var s=(o.template.encapsulation||t._config.defaultEncapsulation)===_.Emulated;n.push(lh(a,s,i)),t._options.allowEmptyCodegenFiles&&n.push(lh(a,!s,i))})}),n},e.prototype.emitBasicStub=function(e,t){var n=this._createOutputContext(e);if(e.endsWith(".ngfactory.ts")){if(!t)throw new Error("Assertion error: require the original file for .ngfactory.ts stubs. File: "+e);var r=this._analyzeFile(t);this._createNgFactoryStub(n,r,1)}else if(e.endsWith(".ngsummary.ts")){if(this._options.enableSummariesForJit){if(!t)throw new Error("Assertion error: require the original file for .ngsummary.ts stubs. File: "+e);r=this._analyzeFile(t);uh(n),r.ngModules.forEach(function(e){var t,r;t=n,r=e.type.reference,nh(t,r,et)})}}else e.endsWith(".ngstyle.ts")&&uh(n);return this._codegenSourceModule("unknown",n)},e.prototype.emitTypeCheckStub=function(e,t){var n=this._analyzeFile(t),r=this._createOutputContext(e);return e.endsWith(".ngfactory.ts")&&this._createNgFactoryStub(r,n,2),r.statements.length>0?this._codegenSourceModule(n.fileName,r):null},e.prototype.loadFilesAsync=function(e,t){var n=this,r=e.map(function(e){return n._analyzeFile(e)}),i=[];r.forEach(function(e){return e.ngModules.forEach(function(e){return i.push(n._metadataResolver.loadNgModuleDirectiveAndPipeMetadata(e.type.reference,!1))})});var o=t.map(function(e){return n._analyzeFileForInjectables(e)});return Promise.all(i).then(function(e){return{analyzedModules:vh(r),analyzedInjectables:o}})},e.prototype.loadFilesSync=function(e,t){var n=this,r=e.map(function(e){return n._analyzeFile(e)});r.forEach(function(e){return e.ngModules.forEach(function(e){return n._metadataResolver.loadNgModuleDirectiveAndPipeMetadata(e.type.reference,!0)})});var i=t.map(function(e){return n._analyzeFileForInjectables(e)});return{analyzedModules:vh(r),analyzedInjectables:i}},e.prototype._createNgFactoryStub=function(e,t,n){var r=this,i=0;t.ngModules.forEach(function(t,a){r._ngModuleCompiler.createStub(e,t.type.reference);var s=Object(o.__spread)(t.transitiveModule.directives.map(function(e){return e.reference}),t.transitiveModule.pipes.map(function(e){return e.reference}),t.importedModules.map(function(e){return e.type.reference}),t.exportedModules.map(function(e){return e.type.reference}),r._externalIdentifierReferences([gn.TemplateRef,gn.ElementRef])),u=new Map;s.forEach(function(e,t){u.set(e,"_decl"+a+"_"+t)}),u.forEach(function(t,n){e.statements.push(Et(t).set(et.cast(me)).toDeclStmt(Mt(e.importExpr(n,null,!1))))}),2&n&&t.declaredDirectives.forEach(function(n){var o=r._metadataResolver.getDirectiveMetadata(n.reference);o.isComponent&&(i++,r._createTypeCheckBlock(e,o.type.reference.name+"_Host_"+i,t,r._metadataResolver.getHostComponentMetadata(o),[o.type],u),r._createTypeCheckBlock(e,o.type.reference.name+"_"+i,t,o,t.transitiveModule.directives,u))})}),0===e.statements.length&&uh(e)},e.prototype._externalIdentifierReferences=function(e){var t,n,r=[];try{for(var i=Object(o.__values)(e),a=i.next();!a.done;a=i.next()){var s=a.value,u=_n(this.reflector,s);u.identifier&&r.push(u.identifier.reference)}}catch(e){t={error:e}}finally{try{a&&!a.done&&(n=i.return)&&n.call(i)}finally{if(t)throw t.error}}return r},e.prototype._createTypeCheckBlock=function(e,t,n,r,i,a){var s,u=this._parseTemplate(r,n,i),c=u.template,l=u.pipes;(s=e.statements).push.apply(s,Object(o.__spread)(this._typeCheckCompiler.compileComponent(t,r,c,l,a,e)))},e.prototype.emitMessageBundle=function(e,t){var n=this,r=[],i=new Uu,a=new qf(i,[],{},t);if(e.files.forEach(function(e){var t=[];e.directives.forEach(function(e){var r=n._metadataResolver.getDirectiveMetadata(e);r&&r.isComponent&&t.push(r)}),t.forEach(function(e){var t=e.template.template,n=e.template.templateUrl,i=zi.fromArray(e.template.interpolation);r.push.apply(r,Object(o.__spread)(a.updateFromTemplate(t,n,i)))})}),r.length)throw new Error(r.map(function(e){return e.toString()}).join("\n"));return a},e.prototype.emitAllPartialModules=function(e,t){var n=this,r=e.ngModuleByPipeOrDirective,i=e.files,a=new Map,s=function(e){return a.has(e)||a.set(e,n._createOutputContext(e)),a.get(e)};return i.forEach(function(e){return n._compilePartialModule(e.fileName,r,e.directives,e.pipes,e.ngModules,e.injectables,s(e.fileName))}),t.forEach(function(e){return n._compileShallowModules(e.fileName,e.shallowModules,s(e.fileName))}),Array.from(a.values()).map(function(e){return{fileName:e.genFilePath,statements:Object(o.__spread)(e.constantPool.statements,e.statements)}})},e.prototype._compileShallowModules=function(e,t,n){var r=this;t.forEach(function(e){return t=n,i=e,a=r._injectableCompiler,s=En(i.type),u=i.rawImports?[i.rawImports]:[],c=i.rawExports?[i.rawExports]:[],l=Ur({factory:a.factoryFor({type:i.type,symbol:i.type.reference},t),providers:Ti(i.rawProviders,t),imports:Ti(Object(o.__spread)(u,c),t)}),d=St(Xn.defineInjector).callFn([l]),void t.statements.push(new dt(s,null,[new ut("ngInjectorDef",ge,[Ie.Static],d)],[],new ct(null,[],[]),[]));var t,i,a,s,u,c,l,d})},e.prototype._compilePartialModule=function(e,t,n,r,i,o,a){var s=this,u=new sc,c=new uc(this._templateParser.expressionParser,Ui,u,[],[]);n.forEach(function(e){var n=s._metadataResolver.getDirectiveMetadata(e);if(n.isComponent){var r=t.get(e);r||Wt("Cannot determine the module for component '"+En(n.type)+"'");var i=n.template.htmlAst;n.template.preserveWhitespaces||(i=Zu(i));var o=Hc(i.rootNodes,c),u=new Map;r.transitiveModule.directives.map(function(e){return s._metadataResolver.getDirectiveSummary(e.reference)}).forEach(function(e){e.selector&&u.set(e.selector,e.type.reference)});var l=new Map;r.transitiveModule.pipes.map(function(e){return s._metadataResolver.getPipeSummary(e.reference)}).forEach(function(e){l.set(e.name,e.type.reference)}),Rl(a,n,o,s.reflector,c,0,l)}else!function(e,t,n,r){var i=En(t.type);i||Wt("Cannot resolver the name of "+t.type);var o=e.constantPool.propertyNameOf(1),a=Dl(Nl(t,e,n),e.constantPool,r);e.statements.push(new dt(i,null,[new ut(o,ge,[Ie.Static],a.expression)],[],new ct(null,[],[]),[]))}(a,n,s.reflector,c)}),r.forEach(function(e){var t=s._metadataResolver.getPipeMetadata(e);t&&function(e,t,n){var r=En(t.type);if(!r)return Wt("Cannot resolve the name of "+t.type);var i=lo({name:r,pipeName:t.name,type:e.importExpr(t.type.reference),deps:ki(t.type,e,n),pure:t.pure}),o=e.constantPool.propertyNameOf(3);e.statements.push(new dt(r,null,[new ut(o,ge,[Ie.Static],i.expression)],[],new ct(null,[],[]),[]))}(a,t,s.reflector)}),o.forEach(function(e){return s._injectableCompiler.compile(e,a)})},e.prototype.emitAllPartialModules2=function(e){var t=this;return e.reduce(function(e,n){return e.push.apply(e,Object(o.__spread)(t._emitPartialModule2(n.fileName,n.injectables))),e},[])},e.prototype._emitPartialModule2=function(e,t){var n=this,r=this._createOutputContext(e);return t.forEach(function(e){return n._injectableCompiler.compile(e,r)}),r.statements&&r.statements.length>0?[{fileName:e,statements:Object(o.__spread)(r.constantPool.statements,r.statements)}]:[]},e.prototype.emitAllImpls=function(e){var t=this,n=e.ngModuleByPipeOrDirective;return Hn(e.files.map(function(e){return t._compileImplFile(e.fileName,n,e.directives,e.pipes,e.ngModules,e.injectables)}))},e.prototype._compileImplFile=function(e,t,n,r,i,a){var s=this,u=bp(yp(e,!0)[1]),c=[],l=this._createOutputContext(vp(e,!0));if(c.push.apply(c,Object(o.__spread)(this._createSummary(e,n,r,i,a,l))),i.forEach(function(e){return s._compileModule(l,e)}),n.forEach(function(n){var r=s._metadataResolver.getDirectiveMetadata(n);if(r.isComponent){var i=t.get(n);if(!i)throw new Error("Internal Error: cannot determine the module for component "+En(r.type)+"!");var o=s._styleCompiler.compileComponent(l,r);r.template.externalStylesheets.forEach(function(t){var n=s._styleCompiler.needsStyleShim(r);c.push(s._codegenStyles(e,r,t,n,u)),s._options.allowEmptyCodegenFiles&&c.push(s._codegenStyles(e,r,t,!n,u))});s._compileComponent(l,r,i,i.transitiveModule.directives,o,u);s._compileComponentFactory(l,r,i,u)}}),l.statements.length>0||this._options.allowEmptyCodegenFiles){var d=this._codegenSourceModule(e,l);c.unshift(d)}return c},e.prototype._createSummary=function(e,t,n,r,i,a){var s=this,u=this._symbolResolver.getSymbolsOf(e).map(function(e){return s._symbolResolver.resolveSymbol(e)}),c=Object(o.__spread)(r.map(function(e){return{summary:s._metadataResolver.getNgModuleSummary(e.type.reference),metadata:s._metadataResolver.getNgModuleMetadata(e.type.reference)}}),t.map(function(e){return{summary:s._metadataResolver.getDirectiveSummary(e),metadata:s._metadataResolver.getDirectiveMetadata(e)}}),n.map(function(e){return{summary:s._metadataResolver.getPipeSummary(e),metadata:s._metadataResolver.getPipeMetadata(e)}}),i.map(function(e){return{summary:s._metadataResolver.getInjectableSummary(e.symbol),metadata:s._metadataResolver.getInjectableSummary(e.symbol).type}})),l=this._options.enableSummariesForJit?this._createOutputContext(xp(e,!0)):null,d=function(e,t,n,r,i,o){var a=new rh(r,n,e);i.forEach(function(e){return a.addSummary({symbol:e.symbol,metadata:e.metadata})}),o.forEach(function(e){var t=e.summary;e.metadata,a.addSummary({symbol:t.type.reference,metadata:void 0,type:t})});var s=a.serialize(),u=s.json,c=s.exportAs;if(t){var l=new ih(t,r,n);o.forEach(function(e){var t=e.summary,n=e.metadata;l.addSourceType(t,n)}),a.unprocessedSymbolSummariesBySymbol.forEach(function(e){n.isLibraryFile(e.symbol.filePath)&&e.type&&l.addLibType(e.type)}),l.serialize(c)}return{json:u,exportAs:c}}(e,l,this._summaryResolver,this._symbolResolver,u,c),p=d.json;d.exportAs.forEach(function(e){a.statements.push(Et(e.exportAs).set(a.importExpr(e.symbol)).toDeclStmt(null,[Ie.Exported]))});var f=[new Wf(e,jp(e),p)];return l&&f.push(this._codegenSourceModule(e,l)),f},e.prototype._compileModule=function(e,t){var n=[];if(this._options.locale){var r=this._options.locale.replace(/_/g,"-");n.push({token:_n(this.reflector,gn.LOCALE_ID),useValue:r})}this._options.i18nFormat&&n.push({token:_n(this.reflector,gn.TRANSLATIONS_FORMAT),useValue:this._options.i18nFormat}),this._ngModuleCompiler.compile(e,t,n)},e.prototype._compileComponentFactory=function(e,t,n,r){var i=this._metadataResolver.getHostComponentMetadata(t),o=this._compileComponent(e,i,n,[t.type],null,r).viewClassVar,a=Tn(t.type.reference),s=[];for(var u in t.inputs){var c=t.inputs[u];s.push(new Ze(u,Pt(c),!1))}var l=[];for(var u in t.outputs){c=t.outputs[u];l.push(new Ze(u,Pt(c),!1))}e.statements.push(Et(a).set(St(gn.createComponentFactory).callFn([Pt(t.selector),e.importExpr(t.type.reference),Et(o),new $e(s),new $e(l),At(t.template.ngContentSelectors.map(function(e){return Pt(e)}))])).toDeclStmt(kt(gn.ComponentFactory,[Mt(e.importExpr(t.type.reference))],[re.Const]),[Ie.Final,Ie.Exported]))},e.prototype._compileComponent=function(e,t,n,r,i,o){var a=this._parseTemplate(t,n,r),s=a.template,u=a.pipes,c=i?Et(i.stylesVar):At([]),l=this._viewCompiler.compileComponent(e,t,s,c,u);return i&&ch(this._symbolResolver,i,this._styleCompiler.needsStyleShim(t),o),l},e.prototype._parseTemplate=function(e,t,n){var r=this;if(this._templateAstCache.has(e.type.reference))return this._templateAstCache.get(e.type.reference);var i=e.template.preserveWhitespaces,o=n.map(function(e){return r._metadataResolver.getDirectiveSummary(e.reference)}),a=t.transitiveModule.pipes.map(function(e){return r._metadataResolver.getPipeSummary(e.reference)}),s=this._templateParser.parse(e,e.template.htmlAst,o,a,t.schemas,Gn(t.type,e,e.template),i);return this._templateAstCache.set(e.type.reference,s),s},e.prototype._createOutputContext=function(e){var t=this;return{statements:[],genFilePath:e,importExpr:function(n,r,i){if(void 0===r&&(r=null),void 0===i&&(i=!0),!(n instanceof yn))throw new Error("Internal error: unknown identifier "+JSON.stringify(n));var o=t._symbolResolver.getTypeArity(n)||0,a=t._symbolResolver.getImportAs(n,i)||n,s=a.filePath,u=a.name,c=a.members,l=t._fileNameToModuleName(s,e),d=l===t._fileNameToModuleName(e,e)?null:l,p=r||[],f=o-p.length,h=p.concat(new Array(f).fill(me));return c.reduce(function(e,t){return e.prop(t)},St(new Be(d,u,null),h))},constantPool:new dn}},e.prototype._fileNameToModuleName=function(e,t){return this._summaryResolver.getKnownModuleName(e)||this._symbolResolver.getKnownModuleName(e)||this._host.fileNameToModuleName(e,t)},e.prototype._codegenStyles=function(e,t,n,r,i){var o=this._createOutputContext(lh(n.moduleUrl,r,i)),a=this._styleCompiler.compileStyles(o,t,n,r);return ch(this._symbolResolver,a,r,i),this._codegenSourceModule(e,o)},e.prototype._codegenSourceModule=function(e,t){return new Wf(e,t.genFilePath,t.statements)},e.prototype.listLazyRoutes=function(e,t){var n,r,i,a,s=this;if(e)return function e(t,n,r){void 0===n&&(n=new Set);void 0===r&&(r=[]);var i,a;if(n.has(t)||!t.name)return r;n.add(t);var u=Kf(s._metadataResolver.getNgModuleMetadata(t,!0),s.reflector);try{for(var c=Object(o.__values)(u),l=c.next();!l.done;l=c.next()){var d=l.value;r.push(d),e(d.referencedModule,n,r)}}catch(e){i={error:e}}finally{try{l&&!l.done&&(a=c.return)&&a.call(c)}finally{if(i)throw i.error}}return r}(Xf(e,this.reflector).referencedModule);if(t){var u=[];try{for(var c=Object(o.__values)(t.ngModules),l=c.next();!l.done;l=c.next()){var d=Kf(l.value,this.reflector);try{for(var p=Object(o.__values)(d),f=p.next();!f.done;f=p.next()){var h=f.value;u.push(h)}}catch(e){i={error:e}}finally{try{f&&!f.done&&(a=p.return)&&a.call(p)}finally{if(i)throw i.error}}}}catch(e){n={error:e}}finally{try{l&&!l.done&&(r=c.return)&&r.call(c)}finally{if(n)throw n.error}}return u}throw new Error("Either route or analyzedModules has to be specified!")},e}();function uh(e){e.statements.push(St(gn.ComponentFactory).toStmt())}function ch(e,t,n,r){t.dependencies.forEach(function(t){t.setValue(e.getStaticSymbol(lh(t.moduleUrl,n,r),t.name))})}function lh(e,t,n){return e+(t?".shim":"")+".ngstyle"+n}function dh(e,t,n,r){var i,o,a,s,u,c,l;return gh((i=e,o=t,a=n,s=r,u=new Set,c=[],l=function(e){if(u.has(e)||!o.isSourceFile(e))return!1;u.add(e);var t=hh(o,a,s,e);c.push(t),t.ngModules.forEach(function(e){e.transitiveModule.modules.forEach(function(e){return l(e.reference.filePath)})})},i.forEach(function(e){return l(e)}),c))}function ph(e,t,n,r){return fh(dh(e,t,n,r))}function fh(e){if(e.symbolsMissingModule&&e.symbolsMissingModule.length)throw Yt(e.symbolsMissingModule.map(function(e){return"Cannot determine the module for class "+e.name+" in "+e.filePath+"! Add "+e.name+" to the NgModule to fix it."}).join("\n"));return e}function hh(e,t,n,r){var i=[],o=[],a=[],s=[],u=t.hasDecorators(r),c=!1;return r.endsWith(".d.ts")&&!u||t.getSymbolsOf(r).forEach(function(r){var u=t.resolveSymbol(r).metadata;if(u&&"error"!==u.__symbolic){var l,d,p,f,h=!1;if("class"===u.__symbolic)if(n.isDirective(r))h=!0,i.push(r);else if(n.isPipe(r))h=!0,o.push(r);else if(n.isNgModule(r)){var m=n.getNgModuleMetadata(r,!1);m&&(h=!0,s.push(m))}else if(n.isInjectable(r)){h=!0;var g=n.getInjectableMetadata(r,null,!1);g&&a.push(g)}h||(c=c||(l=e,d=u,p=!1,f=function(){function e(){}return e.prototype.visitArray=function(e,t){var n=this;e.forEach(function(e){return Vt(e,n,t)})},e.prototype.visitStringMap=function(e,t){var n=this;Object.keys(e).forEach(function(r){return Vt(e[r],n,t)})},e.prototype.visitPrimitive=function(e,t){},e.prototype.visitOther=function(e,t){e instanceof yn&&!l.isSourceFile(e.filePath)&&(p=!0)},e}(),Vt(d,new f,null),p))}}),{fileName:r,directives:i,pipes:o,ngModules:s,injectables:a,exportsNonSourceFiles:c}}function mh(e,t,n,r){var i=[],o=[];return t.hasDecorators(r)&&t.getSymbolsOf(r).forEach(function(e){var r=t.resolveSymbol(e).metadata;if(r&&"error"!==r.__symbolic&&"class"===r.__symbolic)if(n.isInjectable(e)){var a=n.getInjectableMetadata(e,null,!1);a&&i.push(a)}else if(n.isNgModule(e)){var s=n.getShallowModuleMetadata(e);s&&o.push(s)}}),{fileName:r,injectables:i,shallowModules:o}}function gh(e){var t=[],n=new Map,r=new Set;e.forEach(function(e){e.ngModules.forEach(function(e){t.push(e),e.declaredDirectives.forEach(function(t){return n.set(t.reference,e)}),e.declaredPipes.forEach(function(t){return n.set(t.reference,e)})}),e.directives.forEach(function(e){return r.add(e)}),e.pipes.forEach(function(e){return r.add(e)})});var i=[];return r.forEach(function(e){n.has(e)||i.push(e)}),{ngModules:t,ngModuleByPipeOrDirective:n,symbolsMissingModule:i,files:e}}function vh(e){return fh(gh(e))}var _h="ngFormattedMessage";function yh(e,t){if(void 0===t&&(t=0),!e)return"";var n=e.position?e.position.fileName+"("+(e.position.line+1)+","+(e.position.column+1)+")":"",r=n&&0!==t?" at "+n:"",i=""+(n&&0===t?n+": ":"")+e.message+r;return""+function e(t){if(t<=0)return"";if(t<6)return[""," ","  ","   ","    ","     "][t];var n=e(Math.floor(t/2));return n+n+(t%2==1?" ":"")}(t)+i+(e.next&&"\n"+yh(e.next,t+2)||"")}function bh(e){var t=Yt(yh(e)+".");return t[_h]=!0,t.chain=e,t.position=e.position,t}function jh(e){return!!e[_h]}var xh="@angular/core",Oh=/^\$.*\$$/,wh={__symbolic:"ignore"},Ch="useValue",Eh="provide",Sh=new Set([Ch,"useFactory","data","id","loadChildren"]);function kh(e){return e&&"ignore"==e.__symbolic}var Mh=function(){function e(e,t,n,r,i){void 0===n&&(n=[]),void 0===r&&(r=[]);var o=this;this.summaryResolver=e,this.symbolResolver=t,this.errorRecorder=i,this.annotationCache=new Map,this.shallowAnnotationCache=new Map,this.propertyCache=new Map,this.parameterCache=new Map,this.methodCache=new Map,this.staticCache=new Map,this.conversionMap=new Map,this.resolvedExternalReferences=new Map,this.annotationForParentClassWithSummaryKind=new Map,this.initializeConversionMap(),n.forEach(function(e){return o._registerDecoratorOrConstructor(o.getStaticSymbol(e.filePath,e.name),e.ctor)}),r.forEach(function(e){return o._registerFunction(o.getStaticSymbol(e.filePath,e.name),e.fn)}),this.annotationForParentClassWithSummaryKind.set(On.Directive,[P,V]),this.annotationForParentClassWithSummaryKind.set(On.Pipe,[B]),this.annotationForParentClassWithSummaryKind.set(On.NgModule,[G]),this.annotationForParentClassWithSummaryKind.set(On.Injectable,[W,B,P,V,G])}return e.prototype.componentModuleUrl=function(e){var t=this.findSymbolDeclaration(e);return this.symbolResolver.getResourcePath(t)},e.prototype.resolveExternalReference=function(e,t){var n=void 0;if(!t){n=e.moduleName+":"+e.name;var r=this.resolvedExternalReferences.get(n);if(r)return r}var i=this.symbolResolver.getSymbolByModule(e.moduleName,e.name,t),o=this.findSymbolDeclaration(i);return t||(this.symbolResolver.recordModuleNameForFileName(i.filePath,e.moduleName),this.symbolResolver.recordImportAs(o,i)),n&&this.resolvedExternalReferences.set(n,o),o},e.prototype.findDeclaration=function(e,t,n){return this.findSymbolDeclaration(this.symbolResolver.getSymbolByModule(e,t,n))},e.prototype.tryFindDeclaration=function(e,t,n){var r=this;return this.symbolResolver.ignoreErrorsFor(function(){return r.findDeclaration(e,t,n)})},e.prototype.findSymbolDeclaration=function(e){var t=this.symbolResolver.resolveSymbol(e);if(t){var n=t.metadata;if(n&&"resolved"===n.__symbolic&&(n=n.symbol),n instanceof yn)return this.findSymbolDeclaration(t.metadata)}return e},e.prototype.tryAnnotations=function(e){var t=this.errorRecorder;this.errorRecorder=function(e,t){};try{return this.annotations(e)}finally{this.errorRecorder=t}},e.prototype.annotations=function(e){var t=this;return this._annotations(e,function(e,n){return t.simplify(e,n)},this.annotationCache)},e.prototype.shallowAnnotations=function(e){var t=this;return this._annotations(e,function(e,n){return t.simplify(e,n,!0)},this.shallowAnnotationCache)},e.prototype._annotations=function(e,t,n){var r=n.get(e);if(!r){r=[];var i=this.getTypeMetadata(e),a=this.findParentType(e,i);if(a){var s=this.annotations(a);r.push.apply(r,Object(o.__spread)(s))}var u=[];if(i.decorators&&(u=t(e,i.decorators))&&r.push.apply(r,Object(o.__spread)(u)),a&&!this.summaryResolver.isLibraryFile(e.filePath)&&this.summaryResolver.isLibraryFile(a.filePath)){var c=this.summaryResolver.resolveSummary(a);if(c&&c.type){var l=this.annotationForParentClassWithSummaryKind.get(c.type.summaryKind);l.some(function(e){return u.some(function(t){return e.isTypeOf(t)})})||this.reportError(qh(Th("Class "+e.name+" in "+e.filePath+" extends from a "+On[c.type.summaryKind]+" in another compilation unit without duplicating the decorator",void 0,"Please add a "+l.map(function(e){return e.ngMetadataName}).join(" or ")+" decorator to the class"),e),e)}}n.set(e,r.filter(function(e){return!!e}))}return r},e.prototype.propMetadata=function(e){var t=this,n=this.propertyCache.get(e);if(!n){var r=this.getTypeMetadata(e);n={};var i=this.findParentType(e,r);if(i){var a=this.propMetadata(i);Object.keys(a).forEach(function(e){n[e]=a[e]})}var s=r.members||{};Object.keys(s).forEach(function(r){var i=s[r].find(function(e){return"property"==e.__symbolic||"method"==e.__symbolic}),a=[];n[r]&&a.push.apply(a,Object(o.__spread)(n[r])),n[r]=a,i&&i.decorators&&a.push.apply(a,Object(o.__spread)(t.simplify(e,i.decorators)))}),this.propertyCache.set(e,n)}return n},e.prototype.parameters=function(e){var t=this;if(!(e instanceof yn))return this.reportError(new Error("parameters received "+JSON.stringify(e)+" which is not a StaticSymbol"),e),[];try{var n=this.parameterCache.get(e);if(!n){var r=this.getTypeMetadata(e),i=this.findParentType(e,r),a=r?r.members:null,s=a?a.__ctor__:null;if(s){var u=s.find(function(e){return"constructor"==e.__symbolic}),c=u.parameters||[],l=this.simplify(e,u.parameterDecorators||[]);n=[],c.forEach(function(r,i){var a=[],s=t.trySimplify(e,r);s&&a.push(s);var u=l?l[i]:null;u&&a.push.apply(a,Object(o.__spread)(u)),n.push(a)})}else i&&(n=this.parameters(i));n||(n=[]),this.parameterCache.set(e,n)}return n}catch(t){throw console.error("Failed on type "+JSON.stringify(e)+" with error "+t),t}},e.prototype._methodNames=function(e){var t=this.methodCache.get(e);if(!t){var n=this.getTypeMetadata(e);t={};var r=this.findParentType(e,n);if(r){var i=this._methodNames(r);Object.keys(i).forEach(function(e){t[e]=i[e]})}var o=n.members||{};Object.keys(o).forEach(function(e){var n=o[e].some(function(e){return"method"==e.__symbolic});t[e]=t[e]||n}),this.methodCache.set(e,t)}return t},e.prototype._staticMembers=function(e){var t=this.staticCache.get(e);if(!t){var n=this.getTypeMetadata(e).statics||{};t=Object.keys(n),this.staticCache.set(e,t)}return t},e.prototype.findParentType=function(e,t){var n=this.trySimplify(e,t.extends);if(n instanceof yn)return n},e.prototype.hasLifecycleHook=function(e,t){e instanceof yn||this.reportError(new Error("hasLifecycleHook received "+JSON.stringify(e)+" which is not a StaticSymbol"),e);try{return!!this._methodNames(e)[t]}catch(t){throw console.error("Failed on type "+JSON.stringify(e)+" with error "+t),t}},e.prototype.guards=function(e){var t,n;if(!(e instanceof yn))return this.reportError(new Error("guards received "+JSON.stringify(e)+" which is not a StaticSymbol"),e),{};var r=this._staticMembers(e),i={};try{for(var a=Object(o.__values)(r),s=a.next();!s.done;s=a.next()){var u=s.value;if(u.endsWith("TypeGuard")){var c=u.substr(0,u.length-"TypeGuard".length),l=void 0;c.endsWith("UseIf")?(c=u.substr(0,c.length-"UseIf".length),l="UseIf"):l=this.getStaticSymbol(e.filePath,e.name,[u]),i[c]=l}}}catch(e){t={error:e}}finally{try{s&&!s.done&&(n=a.return)&&n.call(a)}finally{if(t)throw t.error}}return i},e.prototype._registerDecoratorOrConstructor=function(e,t){this.conversionMap.set(e,function(e,n){return new(t.bind.apply(t,Object(o.__spread)([void 0],n)))})},e.prototype._registerFunction=function(e,t){this.conversionMap.set(e,function(e,n){return t.apply(void 0,n)})},e.prototype.initializeConversionMap=function(){this._registerDecoratorOrConstructor(this.findDeclaration(xh,"Injectable"),W),this.injectionToken=this.findDeclaration(xh,"InjectionToken"),this.opaqueToken=this.findDeclaration(xh,"OpaqueToken"),this.ROUTES=this.tryFindDeclaration("@angular/router","ROUTES"),this.ANALYZE_FOR_ENTRY_COMPONENTS=this.findDeclaration(xh,"ANALYZE_FOR_ENTRY_COMPONENTS"),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"Host"),$),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"Self"),X),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"SkipSelf"),Z),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"Inject"),S),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"Optional"),Q),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"Attribute"),M),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"ContentChild"),T),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"ContentChildren"),A),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"ViewChild"),I),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"ViewChildren"),D),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"Input"),z),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"Output"),U),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"Pipe"),B),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"HostBinding"),H),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"HostListener"),q),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"Directive"),P),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"Component"),V),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"NgModule"),G),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"Host"),$),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"Self"),X),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"SkipSelf"),Z),this._registerDecoratorOrConstructor(this.findDeclaration(xh,"Optional"),Q)},e.prototype.getStaticSymbol=function(e,t,n){return this.symbolResolver.getStaticSymbol(e,t,n)},e.prototype.trySimplify=function(e,t){var n=this.errorRecorder;this.errorRecorder=function(e,t){};var r=this.simplify(e,t);return this.errorRecorder=n,r},e.prototype.simplify=function(e,t,n){void 0===n&&(n=!1);var r,i=this,a=zh.empty,s=new Map;try{r=function e(t,n,r,u){function c(e){var t=i.symbolResolver.resolveSymbol(e);return t?t.metadata:null}function l(n){return e(t,n,r,0)}function d(n,o){if(n===t)return e(n,o,r+1,u);try{return e(n,o,r+1,u)}catch(e){if(!Dh(e))throw e;var a=e.chain?"references '"+e.symbol.name+"'":function(e){if(e.summary)return e.summary;switch(e.message){case Ih:if(e.context&&e.context.className)return"references non-exported class "+e.context.className;break;case Ph:return"is not initialized";case Rh:return"is a destructured variable";case Nh:return"could not be resolved";case Fh:return e.context&&e.context.name?"calls '"+e.context.name+"'":"calls a function";case Lh:return e.context&&e.context.name?"references local variable "+e.context.name:"references a local variable"}return"contains the error"}(e),s={message:"'"+n.name+"' "+a,position:e.position,next:e.chain};i.error({message:e.message,advise:e.advise,context:e.context,chain:s,symbol:n},t)}}function p(n){var f,h,m,g;if(Bh(n))return n;if(n instanceof Array){var v=[];try{for(var _=Object(o.__values)(n),y=_.next();!y.done;y=_.next()){var b=y.value;if(b&&"spread"===b.__symbolic){var j=l(b.expression);if(Array.isArray(j)){try{for(var x=Object(o.__values)(j),O=x.next();!O.done;O=x.next()){var w=O.value;v.push(w)}}catch(e){m={error:e}}finally{try{O&&!O.done&&(g=x.return)&&g.call(x)}finally{if(m)throw m.error}}continue}}var C=p(b);kh(C)||v.push(C)}}catch(e){f={error:e}}finally{try{y&&!y.done&&(h=_.return)&&h.call(_)}finally{if(f)throw f.error}}return v}if(n instanceof yn)return n===i.injectionToken||i.conversionMap.has(n)||u>0&&!n.members.length?n:null!=(R=c(E=n))?d(E,R):E;if(n){if(n.__symbolic){var E=void 0;switch(n.__symbolic){case"binop":var S=p(n.left);if(kh(S))return S;var k=p(n.right);if(kh(k))return k;switch(n.operator){case"&&":return S&&k;case"||":return S||k;case"|":return S|k;case"^":return S^k;case"&":return S&k;case"==":return S==k;case"!=":return S!=k;case"===":return S===k;case"!==":return S!==k;case"<":return S<k;case">":return S>k;case"<=":return S<=k;case">=":return S>=k;case"<<":return S<<k;case">>":return S>>k;case"+":return S+k;case"-":return S-k;case"*":return S*k;case"/":return S/k;case"%":return S%k}return null;case"if":return p(p(n.condition)?n.thenExpression:n.elseExpression);case"pre":var M=p(n.operand);if(kh(M))return M;switch(n.operator){case"+":return M;case"-":return-M;case"!":return!M;case"~":return~M}return null;case"index":var A=l(n.expression),T=l(n.index);return A&&Bh(T)?A[T]:null;case"select":var D=n.member,I=t,P=p(n.expression);if(P instanceof yn){var R,N=P.members.concat(D);return null!=(R=c(I=i.getStaticSymbol(P.filePath,P.name,N)))?d(I,R):I}return P&&Bh(D)?d(I,P[D]):null;case"reference":var F=n.name,L=a.resolve(F);if(L!=zh.missing)return L;break;case"resolved":try{return p(n.symbol)}catch(e){throw Dh(e)&&null!=n.fileName&&null!=n.line&&null!=n.character&&(e.position={fileName:n.fileName,line:n.line,column:n.character}),e}case"class":case"function":return t;case"new":case"call":if((E=e(t,n.expression,r+1,0))instanceof yn){if(E===i.injectionToken||E===i.opaqueToken)return t;var V=n.arguments||[],B=i.conversionMap.get(E);if(B){var z=V.map(function(e){return d(t,e)}).map(function(e){return kh(e)?void 0:e});return B(t,z)}return function(e,n,u,c){if(n&&"function"==n.__symbolic){s.get(e)&&i.error({message:"Recursion is not supported",summary:"called '"+e.name+"' recursively",value:n},e);try{var l=n.value;if(l&&(0!=r||"error"!=l.__symbolic)){var f=n.parameters,h=n.defaults;u=u.map(function(e){return d(t,e)}).map(function(e){return kh(e)?void 0:e}),h&&h.length>u.length&&u.push.apply(u,Object(o.__spread)(h.slice(u.length).map(function(e){return p(e)}))),s.set(e,!0);for(var m=zh.build(),g=0;g<f.length;g++)m.define(f[g],u[g]);var v,_=a;try{a=m.done(),v=d(e,l)}finally{a=_}return v}}finally{s.delete(e)}}if(0===r)return wh;var y=void 0;if(c&&"resolved"==c.__symbolic){var b=c.line,j=c.character,x=c.fileName;null!=x&&null!=b&&null!=j&&(y={fileName:x,line:b,column:j})}i.error({message:Fh,context:e,value:n,position:y},t)}(E,c(E),V,n.expression)}return wh;case"error":var U=n.message;return null!=n.line?i.error({message:U,context:n.context,value:n,position:{fileName:n.fileName,line:n.line,column:n.character}},t):i.error({message:U,context:n.context},t),wh;case"ignore":return n}return null}return function(e,t){if(!e)return{};var n={};return Object.keys(e).forEach(function(r){var i=t(e[r],r);kh(i)||(Oh.test(r)?Object.defineProperty(n,r,{enumerable:!1,configurable:!0,value:i}):n[r]=i)}),n}(n,function(o,a){if(Sh.has(a)){if(a===Ch&&Eh in n){var s=p(n.provide);if(s===i.ROUTES||s==i.ANALYZE_FOR_ENTRY_COMPONENTS)return p(o)}return e(t,o,r,u+1)}return p(o)})}return wh}return p(n)}(e,t,0,n?1:0)}catch(t){if(!this.errorRecorder)throw qh(t,e);this.reportError(t,e)}if(!kh(r))return r},e.prototype.getTypeMetadata=function(e){var t=this.symbolResolver.resolveSymbol(e);return t&&t.metadata?t.metadata:{__symbolic:"class"}},e.prototype.reportError=function(e,t,n){if(!this.errorRecorder)throw e;this.errorRecorder(qh(e,t),t&&t.filePath||n)},e.prototype.error=function(e,t){var n=e.message,r=e.summary,i=e.advise,o=e.position,a=e.context,s=(e.value,e.symbol),u=e.chain;this.reportError(Th(n,r,i,o,s,a,u),t)},e}(),Ah="ngMetadataError";function Th(e,t,n,r,i,o,a){var s=Yt(e);return s[Ah]=!0,n&&(s.advise=n),r&&(s.position=r),t&&(s.summary=t),o&&(s.context=o),a&&(s.chain=a),i&&(s.symbol=i),s}function Dh(e){return!!e[Ah]}var Ih="Reference to non-exported class",Ph="Variable not initialized",Rh="Destructuring not supported",Nh="Could not resolve type",Fh="Function call not supported",Lh="Reference to a local symbol",Vh="Lambda not supported";function Bh(e){return null===e||"function"!=typeof e&&"object"!=typeof e}var zh=function(){function e(){}return e.build=function(){var t=new Map;return{define:function(e,n){return t.set(e,n),this},done:function(){return t.size>0?new Uh(t):e.empty}}},e.missing={},e.empty={resolve:function(t){return e.missing}},e}(),Uh=function(e){function t(t){var n=e.call(this)||this;return n.bindings=t,n}return Object(o.__extends)(t,e),t.prototype.resolve=function(e){return this.bindings.has(e)?this.bindings.get(e):zh.missing},t}(zh);function Hh(e,t){return{message:""+function(e,t){switch(e){case Ih:if(t&&t.className)return"References to a non-exported class are not supported in decorators but "+t.className+" was referenced.";break;case Ph:return"Only initialized variables and constants can be referenced in decorators because the value of this variable is needed by the template compiler";case Rh:return"Referencing an exported destructured variable or constant is not supported in decorators and this value is needed by the template compiler";case Nh:if(t&&t.typeName)return"Could not resolve type "+t.typeName;break;case Fh:return t&&t.name?"Function calls are not supported in decorators but '"+t.name+"' was called":"Function calls are not supported in decorators";case Lh:if(t&&t.name)return"Reference to a local (non-exported) symbols are not supported in decorators but '"+t.name+"' was referenced";break;case Vh:return"Function expressions are not supported in decorators"}return e}(e.message,e.context)+(e.symbol?" in '"+e.symbol.name+"'":""),position:e.position,next:e.next?Hh(e.next,t):t?{message:t}:void 0}}function qh(e,t){if(Dh(e)){var n=e.position;return bh(Hh({message:"Error during template compile of '"+t.name+"'",position:n,next:{message:e.message,next:e.chain,context:e.context,symbol:e.symbol}},e.advise||function(e,t){switch(e){case Ih:if(t&&t.className)return"Consider exporting '"+t.className+"'";break;case Rh:return"Consider simplifying to avoid destructuring";case Lh:if(t&&t.name)return"Consider exporting '"+t.name+"'";break;case Vh:return"Consider changing the function expression into an exported function"}}(e.message,e.context)))}return e}var Gh=function(){function e(e,t){this.host=e,this.staticSymbolCache=t,this.summaryCache=new Map,this.loadedFilePaths=new Map,this.importAs=new Map,this.knownFileNameToModuleNames=new Map}return e.prototype.isLibraryFile=function(e){return!this.host.isSourceFile(_p(e))},e.prototype.toSummaryFileName=function(e,t){return this.host.toSummaryFileName(e,t)},e.prototype.fromSummaryFileName=function(e,t){return this.host.fromSummaryFileName(e,t)},e.prototype.resolveSummary=function(e){var t=e.members.length?this.staticSymbolCache.get(e.filePath,e.name):e,n=this.summaryCache.get(t);return n||(this._loadSummaryFile(e.filePath),n=this.summaryCache.get(e)),t===e&&n||null},e.prototype.getSymbolsOf=function(e){return this._loadSummaryFile(e)?Array.from(this.summaryCache.keys()).filter(function(t){return t.filePath===e}):null},e.prototype.getImportAs=function(e){return e.assertNoMembers(),this.importAs.get(e)},e.prototype.getKnownModuleName=function(e){return this.knownFileNameToModuleNames.get(e)||null},e.prototype.addSummary=function(e){this.summaryCache.set(e.symbol,e)},e.prototype._loadSummaryFile=function(e){var t=this,n=this.loadedFilePaths.get(e);if(null!=n)return n;var r,i,o,a=null;if(this.isLibraryFile(e)){var s=jp(e);try{a=this.host.loadSummary(s)}catch(e){throw console.error("Error loading summary file "+s),e}}if(n=null!=a,this.loadedFilePaths.set(e,n),a){var u=(r=this.staticSymbolCache,i=e,o=a,new oh(r,this).deserialize(i,o)),c=u.moduleName,l=u.summaries,d=u.importAs;l.forEach(function(e){return t.summaryCache.set(e.symbol,e)}),c&&this.knownFileNameToModuleNames.set(e,c),d.forEach(function(e){t.importAs.set(e.symbol,e.importAs)})}return n},e}();function Wh(e){return{resolve:function(t,n){var r=e.resourceNameToFileName(n,t);if(!r)throw Yt("Couldn't resolve resource "+n+" from "+t);return r}}}function Yh(e,t,n){var r,i=t.translations||"",o=Wh(e),a=new bn,s=new Gh(e,a),u=new Jf(e,a,s),c=new Mh(s,u,[],[],n);r=t.enableIvy?new Uu:new up(new Uu,i,t.i18nFormat,t.missingTranslation,console);var l=new Sd({defaultEncapsulation:_.Emulated,useJit:!1,missingTranslation:t.missingTranslation,preserveWhitespaces:t.preserveWhitespaces,strictInjectionParameters:t.strictInjectionParameters}),d=new Md({get:function(t){return e.loadResource(t)}},o,r,l),p=new uu(new Ks),f=new sc,h=new ff(l,c,p,f,r,console,[]),m=new kp(l,r,new Xp(c),new Dd(c),new ef(c),s,f,d,console,a,c,n),g=new Tf(c),v=new Cf(t,c);return{compiler:new sh(l,t,e,c,m,h,new Ls(o),g,v,new Qp(c),new pp(c,!!t.enableIvy),new $p,s,u),reflector:c}}var Kh=function(){return function(){}}(),Qh=function(){function e(){this._summaries=new Map}return e.prototype.isLibraryFile=function(){return!1},e.prototype.toSummaryFileName=function(e){return e},e.prototype.fromSummaryFileName=function(e){return e},e.prototype.resolveSummary=function(e){return this._summaries.get(e)||null},e.prototype.getSymbolsOf=function(){return[]},e.prototype.getImportAs=function(e){return e},e.prototype.getKnownModuleName=function(e){return null},e.prototype.addSummary=function(e){this._summaries.set(e.symbol,e)},e}();function Xh(e,t,n,r,i){for(var o=r.createChildWihtLocalVars(),a=0;a<e.length;a++)o.vars.set(e[a],t[a]);var s=i.visitAllStatements(n,o);return s?s.value:null}var Zh=function(){function e(e,t,n,r){this.parent=e,this.instance=t,this.className=n,this.vars=r,this.exports=[]}return e.prototype.createChildWihtLocalVars=function(){return new e(this,this.instance,this.className,new Map)},e}(),$h=function(){return function(e){this.value=e}}();var Jh=function(){function e(e){this.reflector=e}return e.prototype.debugAst=function(e){return Zp(e)},e.prototype.visitDeclareVarStmt=function(e,t){var n=e.value?e.value.visitExpression(this,t):void 0;return t.vars.set(e.name,n),e.hasModifier(Ie.Exported)&&t.exports.push(e.name),null},e.prototype.visitWriteVarExpr=function(e,t){for(var n=e.value.visitExpression(this,t),r=t;null!=r;){if(r.vars.has(e.name))return r.vars.set(e.name,n),n;r=r.parent}throw new Error("Not declared variable "+e.name)},e.prototype.visitWrappedNodeExpr=function(e,t){throw new Error("Cannot interpret a WrappedNodeExpr.")},e.prototype.visitTypeofExpr=function(e,t){throw new Error("Cannot interpret a TypeofExpr")},e.prototype.visitReadVarExpr=function(e,t){var n=e.name;if(null!=e.builtin)switch(e.builtin){case xe.Super:return t.instance.__proto__;case xe.This:return t.instance;case xe.CatchError:n=tm;break;case xe.CatchStack:n=nm;break;default:throw new Error("Unknown builtin variable "+e.builtin)}for(var r=t;null!=r;){if(r.vars.has(n))return r.vars.get(n);r=r.parent}throw new Error("Not declared variable "+n)},e.prototype.visitWriteKeyExpr=function(e,t){var n=e.receiver.visitExpression(this,t),r=e.index.visitExpression(this,t),i=e.value.visitExpression(this,t);return n[r]=i,i},e.prototype.visitWritePropExpr=function(e,t){var n=e.receiver.visitExpression(this,t),r=e.value.visitExpression(this,t);return n[e.name]=r,r},e.prototype.visitInvokeMethodExpr=function(e,t){var n,r=e.receiver.visitExpression(this,t),i=this.visitAllExpressions(e.args,t);if(null!=e.builtin)switch(e.builtin){case Ce.ConcatArray:n=r.concat.apply(r,Object(o.__spread)(i));break;case Ce.SubscribeObservable:n=r.subscribe({next:i[0]});break;case Ce.Bind:n=r.bind.apply(r,Object(o.__spread)(i));break;default:throw new Error("Unknown builtin method "+e.builtin)}else n=r[e.name].apply(r,i);return n},e.prototype.visitInvokeFunctionExpr=function(e,t){var n=this.visitAllExpressions(e.args,t),r=e.fn;return r instanceof Se&&r.builtin===xe.Super?(t.instance.constructor.prototype.constructor.apply(t.instance,n),null):e.fn.visitExpression(this,t).apply(null,n)},e.prototype.visitReturnStmt=function(e,t){return new $h(e.value.visitExpression(this,t))},e.prototype.visitDeclareClassStmt=function(e,t){var n=function(e,t,n){var r={};e.getters.forEach(function(i){r[i.name]={configurable:!1,get:function(){var r=new Zh(t,this,e.name,t.vars);return Xh([],[],i.body,r,n)}}}),e.methods.forEach(function(i){var o=i.params.map(function(e){return e.name});r[i.name]={writable:!1,configurable:!1,value:function(){for(var r=[],a=0;a<arguments.length;a++)r[a]=arguments[a];var s=new Zh(t,this,e.name,t.vars);return Xh(o,r,i.body,s,n)}}});var i=e.constructorMethod.params.map(function(e){return e.name}),o=function(){for(var r=this,o=[],a=0;a<arguments.length;a++)o[a]=arguments[a];var s=new Zh(t,this,e.name,t.vars);e.fields.forEach(function(e){r[e.name]=void 0}),Xh(i,o,e.constructorMethod.body,s,n)},a=e.parent?e.parent.visitExpression(n,t):Object;return o.prototype=Object.create(a.prototype,r),o}(e,t,this);return t.vars.set(e.name,n),e.hasModifier(Ie.Exported)&&t.exports.push(e.name),null},e.prototype.visitExpressionStmt=function(e,t){return e.expr.visitExpression(this,t)},e.prototype.visitIfStmt=function(e,t){return e.condition.visitExpression(this,t)?this.visitAllStatements(e.trueCase,t):null!=e.falseCase?this.visitAllStatements(e.falseCase,t):null},e.prototype.visitTryCatchStmt=function(e,t){try{return this.visitAllStatements(e.bodyStmts,t)}catch(r){var n=t.createChildWihtLocalVars();return n.vars.set(tm,r),n.vars.set(nm,r.stack),this.visitAllStatements(e.catchStmts,n)}},e.prototype.visitThrowStmt=function(e,t){throw e.error.visitExpression(this,t)},e.prototype.visitCommentStmt=function(e,t){return null},e.prototype.visitJSDocCommentStmt=function(e,t){return null},e.prototype.visitInstantiateExpr=function(e,t){var n=this.visitAllExpressions(e.args,t),r=e.classExpr.visitExpression(this,t);return new(r.bind.apply(r,Object(o.__spread)([void 0],n)))},e.prototype.visitLiteralExpr=function(e,t){return e.value},e.prototype.visitExternalExpr=function(e,t){return this.reflector.resolveExternalReference(e.value)},e.prototype.visitConditionalExpr=function(e,t){return e.condition.visitExpression(this,t)?e.trueCase.visitExpression(this,t):null!=e.falseCase?e.falseCase.visitExpression(this,t):null},e.prototype.visitNotExpr=function(e,t){return!e.condition.visitExpression(this,t)},e.prototype.visitAssertNotNullExpr=function(e,t){return e.condition.visitExpression(this,t)},e.prototype.visitCastExpr=function(e,t){return e.value.visitExpression(this,t)},e.prototype.visitFunctionExpr=function(e,t){return em(e.params.map(function(e){return e.name}),e.statements,t,this)},e.prototype.visitDeclareFunctionStmt=function(e,t){var n=e.params.map(function(e){return e.name});return t.vars.set(e.name,em(n,e.statements,t,this)),e.hasModifier(Ie.Exported)&&t.exports.push(e.name),null},e.prototype.visitBinaryOperatorExpr=function(e,t){var n=this,r=function(){return e.lhs.visitExpression(n,t)},i=function(){return e.rhs.visitExpression(n,t)};switch(e.operator){case ce.Equals:return r()==i();case ce.Identical:return r()===i();case ce.NotEquals:return r()!=i();case ce.NotIdentical:return r()!==i();case ce.And:return r()&&i();case ce.Or:return r()||i();case ce.Plus:return r()+i();case ce.Minus:return r()-i();case ce.Divide:return r()/i();case ce.Multiply:return r()*i();case ce.Modulo:return r()%i();case ce.Lower:return r()<i();case ce.LowerEquals:return r()<=i();case ce.Bigger:return r()>i();case ce.BiggerEquals:return r()>=i();default:throw new Error("Unknown operator "+e.operator)}},e.prototype.visitReadPropExpr=function(e,t){return e.receiver.visitExpression(this,t)[e.name]},e.prototype.visitReadKeyExpr=function(e,t){return e.receiver.visitExpression(this,t)[e.index.visitExpression(this,t)]},e.prototype.visitLiteralArrayExpr=function(e,t){return this.visitAllExpressions(e.entries,t)},e.prototype.visitLiteralMapExpr=function(e,t){var n=this,r={};return e.entries.forEach(function(e){return r[e.key]=e.value.visitExpression(n,t)}),r},e.prototype.visitCommaExpr=function(e,t){var n=this.visitAllExpressions(e.parts,t);return n[n.length-1]},e.prototype.visitAllExpressions=function(e,t){var n=this;return e.map(function(e){return e.visitExpression(n,t)})},e.prototype.visitAllStatements=function(e,t){for(var n=0;n<e.length;n++){var r=e[n].visitStatement(this,t);if(r instanceof $h)return r}return null},e}();function em(e,t,n,r){return function(){for(var i=[],o=0;o<arguments.length;o++)i[o]=arguments[o];return Xh(e,i,t,n,r)}}var tm="error",nm="stack",rm=function(){function e(e,t,n,r,i,o,a,s,u,c){this._metadataResolver=e,this._templateParser=t,this._styleCompiler=n,this._viewCompiler=r,this._ngModuleCompiler=i,this._summaryResolver=o,this._reflector=a,this._compilerConfig=s,this._console=u,this.getExtraNgModuleProviders=c,this._compiledTemplateCache=new Map,this._compiledHostTemplateCache=new Map,this._compiledDirectiveWrapperCache=new Map,this._compiledNgModuleCache=new Map,this._sharedStylesheetCount=0,this._addedAotSummaries=new Set}return e.prototype.compileModuleSync=function(e){return Ht(this._compileModuleAndComponents(e,!0))},e.prototype.compileModuleAsync=function(e){return Promise.resolve(this._compileModuleAndComponents(e,!1))},e.prototype.compileModuleAndAllComponentsSync=function(e){return Ht(this._compileModuleAndAllComponents(e,!0))},e.prototype.compileModuleAndAllComponentsAsync=function(e){return Promise.resolve(this._compileModuleAndAllComponents(e,!1))},e.prototype.getComponentFactory=function(e){return this._metadataResolver.getDirectiveSummary(e).componentFactory},e.prototype.loadAotSummaries=function(e){this.clearCache(),this._addAotSummaries(e)},e.prototype._addAotSummaries=function(e){if(!this._addedAotSummaries.has(e)){this._addedAotSummaries.add(e);for(var t=e(),n=0;n<t.length;n++){var r=t[n];if("function"==typeof r)this._addAotSummaries(r);else{var i=r;this._summaryResolver.addSummary({symbol:i.type.reference,metadata:null,type:i})}}}},e.prototype.hasAotSummary=function(e){return!!this._summaryResolver.resolveSummary(e)},e.prototype._filterJitIdentifiers=function(e){var t=this;return e.map(function(e){return e.reference}).filter(function(e){return!t.hasAotSummary(e)})},e.prototype._compileModuleAndComponents=function(e,t){var n=this;return qt(this._loadModules(e,t),function(){return n._compileComponents(e,null),n._compileModule(e)})},e.prototype._compileModuleAndAllComponents=function(e,t){var n=this;return qt(this._loadModules(e,t),function(){var t=[];return n._compileComponents(e,t),{ngModuleFactory:n._compileModule(e),componentFactories:t}})},e.prototype._loadModules=function(e,t){var n=this,r=[],i=this._metadataResolver.getNgModuleMetadata(e);return this._filterJitIdentifiers(i.transitiveModule.modules).forEach(function(e){var i=n._metadataResolver.getNgModuleMetadata(e);n._filterJitIdentifiers(i.declaredDirectives).forEach(function(e){var o=n._metadataResolver.loadDirectiveMetadata(i.type.reference,e,t);o&&r.push(o)}),n._filterJitIdentifiers(i.declaredPipes).forEach(function(e){return n._metadataResolver.getOrLoadPipeMetadata(e)})}),Gt(r)},e.prototype._compileModule=function(e){var t=this._compiledNgModuleCache.get(e);if(!t){var n=this._metadataResolver.getNgModuleMetadata(e),r=this.getExtraNgModuleProviders(n.type.reference),i=am(),o=this._ngModuleCompiler.compile(i,n,r);t=this._interpretOrJit(Yn(n),i.statements)[o.ngModuleFactoryVar],this._compiledNgModuleCache.set(n.type.reference,t)}return t},e.prototype._compileComponents=function(e,t){var n=this,r=this._metadataResolver.getNgModuleMetadata(e),i=new Map,o=new Set,a=this._filterJitIdentifiers(r.transitiveModule.modules);a.forEach(function(e){var r=n._metadataResolver.getNgModuleMetadata(e);n._filterJitIdentifiers(r.declaredDirectives).forEach(function(e){i.set(e,r);var a=n._metadataResolver.getDirectiveMetadata(e);if(a.isComponent&&(o.add(n._createCompiledTemplate(a,r)),t)){var s=n._createCompiledHostTemplate(a.type.reference,r);o.add(s),t.push(a.componentFactory)}})}),a.forEach(function(e){var t=n._metadataResolver.getNgModuleMetadata(e);n._filterJitIdentifiers(t.declaredDirectives).forEach(function(e){var t=n._metadataResolver.getDirectiveMetadata(e);t.isComponent&&t.entryComponents.forEach(function(e){var t=i.get(e.componentType);o.add(n._createCompiledHostTemplate(e.componentType,t))})}),t.entryComponents.forEach(function(e){if(!n.hasAotSummary(e.componentType)){var t=i.get(e.componentType);o.add(n._createCompiledHostTemplate(e.componentType,t))}})}),o.forEach(function(e){return n._compileTemplate(e)})},e.prototype.clearCacheFor=function(e){this._compiledNgModuleCache.delete(e),this._metadataResolver.clearCacheFor(e),this._compiledHostTemplateCache.delete(e),this._compiledTemplateCache.get(e)&&this._compiledTemplateCache.delete(e)},e.prototype.clearCache=function(){this._metadataResolver.clearCache(),this._compiledTemplateCache.clear(),this._compiledHostTemplateCache.clear(),this._compiledNgModuleCache.clear()},e.prototype._createCompiledHostTemplate=function(e,t){if(!t)throw new Error("Component "+tn(e)+" is not part of any NgModule or the module has not been imported into your module.");var n=this._compiledHostTemplateCache.get(e);if(!n){var r=this._metadataResolver.getDirectiveMetadata(e);om(r);var i=this._metadataResolver.getHostComponentMetadata(r,r.componentFactory.viewDefFactory);n=new im(!0,r.type,i,t,[r.type]),this._compiledHostTemplateCache.set(e,n)}return n},e.prototype._createCompiledTemplate=function(e,t){var n=this._compiledTemplateCache.get(e.type.reference);return n||(om(e),n=new im(!1,e.type,e,t,t.transitiveModule.directives),this._compiledTemplateCache.set(e.type.reference,n)),n},e.prototype._compileTemplate=function(e){var t=this;if(!e.isCompiled){var n=e.compMeta,r=new Map,i=am(),o=this._styleCompiler.compileComponent(i,n);n.template.externalStylesheets.forEach(function(e){var i=t._styleCompiler.compileStyles(am(),n,e);r.set(e.moduleUrl,i)}),this._resolveStylesCompileResult(o,r);e.ngModule.transitiveModule.pipes.map(function(e){return t._metadataResolver.getPipeSummary(e.reference)});var a=this._parseTemplate(n,e.ngModule,e.directives),s=a.template,u=a.pipes,c=this._viewCompiler.compileComponent(i,n,s,Et(o.stylesVar),u),l=this._interpretOrJit(Kn(e.ngModule.type,e.compMeta),i.statements),d=l[c.viewClassVar],p=l[c.rendererTypeVar];e.compiled(d,p)}},e.prototype._parseTemplate=function(e,t,n){var r=this,i=e.template.preserveWhitespaces,o=n.map(function(e){return r._metadataResolver.getDirectiveSummary(e.reference)}),a=t.transitiveModule.pipes.map(function(e){return r._metadataResolver.getPipeSummary(e.reference)});return this._templateParser.parse(e,e.template.htmlAst,o,a,t.schemas,Gn(t.type,e,e.template),i)},e.prototype._resolveStylesCompileResult=function(e,t){var n=this;e.dependencies.forEach(function(e,r){var i=t.get(e.moduleUrl),o=n._resolveAndEvalStylesCompileResult(i,t);e.setValue(o)})},e.prototype._resolveAndEvalStylesCompileResult=function(e,t){return this._resolveStylesCompileResult(e,t),this._interpretOrJit(Wn(e.meta,this._sharedStylesheetCount++),e.outputCtx.statements)[e.stylesVar]},e.prototype._interpretOrJit=function(e,t){return this._compilerConfig.useJit?ro(e,t,this._reflector,this._compilerConfig.jitDevMode):function(e,t){var n=new Zh(null,null,null,new Map);new Jh(t).visitAllStatements(e,n);var r={};return n.exports.forEach(function(e){r[e]=n.vars.get(e)}),r}(t,this._reflector)},e}(),im=function(){function e(e,t,n,r,i){this.isHost=e,this.compType=t,this.compMeta=n,this.ngModule=r,this.directives=i,this._viewClass=null,this.isCompiled=!1}return e.prototype.compiled=function(e,t){this._viewClass=e,this.compMeta.componentViewType.setDelegate(e);for(var n in t)this.compMeta.rendererType[n]=t[n];this.isCompiled=!0},e}();function om(e){if(!e.isComponent)throw new Error("Could not compile '"+En(e.type)+"' because it is not a component.")}function am(){return{statements:[],genFilePath:"",importExpr:function(e){return St({name:En(e),moduleName:null,runtime:e})},constantPool:new dn}}var sm=function(){return function(){}}();function um(){return new lm}function cm(){return new lm(".")}var lm=function(){function e(e){void 0===e&&(e=null),this._packagePrefix=e}return e.prototype.resolve=function(e,t){var n=t;null!=e&&e.length>0&&(n=function(e,t){var n=mm(encodeURI(t)),r=mm(e);if(null!=n[pm.Scheme])return gm(n);n[pm.Scheme]=r[pm.Scheme];for(var i=pm.Scheme;i<=pm.Port;i++)null==n[i]&&(n[i]=r[i]);if("/"==n[pm.Path][0])return gm(n);var o=r[pm.Path];null==o&&(o="/");var a=o.lastIndexOf("/");return o=o.substring(0,a+1)+n[pm.Path],n[pm.Path]=o,gm(n)}(e,n));var r=mm(n),i=this._packagePrefix;if(null!=i&&null!=r&&"package"==r[pm.Scheme]){var o=r[pm.Path];return(i=i.replace(/\/+$/,""))+"/"+(o=o.replace(/^\/+/,""))}return n},e}();function dm(e){var t=mm(e);return t&&t[pm.Scheme]||""}var pm,fm,hm=new RegExp("^(?:([^:/?#.]+):)?(?://(?:([^/?#]*)@)?([\\w\\d\\-\\u0100-\\uffff.%]*)(?::([0-9]+))?)?([^?#]+)?(?:\\?([^#]*))?(?:#(.*))?$");function mm(e){return e.match(hm)}function gm(e){var t,n,r,i,o,a,s,u,c=e[pm.Path];return c=null==c?"":function(e){if("/"==e)return"/";for(var t="/"==e[0]?"/":"",n="/"===e[e.length-1]?"/":"",r=e.split("/"),i=[],o=0,a=0;a<r.length;a++){var s=r[a];switch(s){case"":case".":break;case"..":i.length>0?i.pop():o++;break;default:i.push(s)}}if(""==t){for(;o-- >0;)i.unshift("..");0===i.length&&i.push(".")}return t+i.join("/")+n}(c),e[pm.Path]=c,t=e[pm.Scheme],n=e[pm.UserInfo],r=e[pm.Domain],i=e[pm.Port],o=c,a=e[pm.QueryData],s=e[pm.Fragment],u=[],null!=t&&u.push(t+":"),null!=r&&(u.push("//"),null!=n&&u.push(n+"@"),u.push(r),null!=i&&u.push(":"+i)),null!=o&&u.push(o),null!=a&&u.push("?"+a),null!=s&&u.push("#"+s),u.join("")}(fm=pm||(pm={}))[fm.Scheme=1]="Scheme",fm[fm.UserInfo=2]="UserInfo",fm[fm.Domain=3]="Domain",fm[fm.Port=4]="Port",fm[fm.Path=5]="Path",fm[fm.QueryData=6]="QueryData",fm[fm.Fragment=7]="Fragment";var vm=function(){function e(){}return e.prototype.get=function(e){return""},e}(),_m=function(){function e(e,t,n,r){this.host=e,this.staticSymbolResolver=t,this.messageBundle=n,this.metadataResolver=r}return e.prototype.extract=function(e){var t=this,n=ph(e,this.host,this.staticSymbolResolver,this.metadataResolver),r=n.files,i=n.ngModules;return Promise.all(i.map(function(e){return t.metadataResolver.loadNgModuleDirectiveAndPipeMetadata(e.type.reference,!1)})).then(function(){var e=[];if(r.forEach(function(n){var r=[];n.directives.forEach(function(e){var n=t.metadataResolver.getDirectiveMetadata(e);n&&n.isComponent&&r.push(n)}),r.forEach(function(n){var r=n.template.template,i=n.template.templateUrl,a=zi.fromArray(n.template.interpolation);e.push.apply(e,Object(o.__spread)(t.messageBundle.updateFromTemplate(r,i,a)))})}),e.length)throw new Error(e.map(function(e){return e.toString()}).join("\n"));return t.messageBundle})},e.create=function(t,n){var r=new Uu,i=Wh(t),o=new bn,a=new Gh(t,o),s=new Jf(t,o,a),u=new Mh(a,s),c=new Sd({defaultEncapsulation:_.Emulated,useJit:!1}),l=new Md({get:function(e){return t.loadResource(e)}},i,r,c),d=new sc,p=new kp(c,r,new Xp(u),new Dd(u),new ef(u),a,d,l,console,o,u),f=new qf(r,[],{},n);return{extractor:new e(t,s,f,p),staticReflector:u}},e}(),ym=function(){function e(e){this.directiveMatcher=e}return e.prototype.bind=function(e){if(!e.template)throw new Error("Binding without a template not yet supported");var t=bm.apply(e.template),n=jm.apply(e.template,this.directiveMatcher),r=n.directives,i=n.bindings,o=n.references,a=xm.apply(e.template,t),s=a.expressions,u=a.symbols,c=a.nestingLevel;return new Om(e,r,i,o,s,u,c)},e}(),bm=function(){function e(e){this.parentScope=e,this.namedEntities=new Map,this.childScopes=new Map}return e.apply=function(t){var n=new e;return n.ingest(t),n},e.prototype.ingest=function(e){var t=this;e instanceof _c?(e.variables.forEach(function(e){return t.visitVariable(e)}),e.children.forEach(function(e){return e.visit(t)})):e.forEach(function(e){return e.visit(t)})},e.prototype.visitElement=function(e){var t=this;e.references.forEach(function(e){return t.visitReference(e)}),e.children.forEach(function(e){return e.visit(t)})},e.prototype.visitTemplate=function(t){var n=this;t.references.forEach(function(e){return n.visitReference(e)});var r=new e(this);r.ingest(t),this.childScopes.set(t,r)},e.prototype.visitVariable=function(e){this.maybeDeclare(e)},e.prototype.visitReference=function(e){this.maybeDeclare(e)},e.prototype.visitContent=function(e){},e.prototype.visitBoundAttribute=function(e){},e.prototype.visitBoundEvent=function(e){},e.prototype.visitBoundText=function(e){},e.prototype.visitText=function(e){},e.prototype.visitTextAttribute=function(e){},e.prototype.visitIcu=function(e){},e.prototype.maybeDeclare=function(e){this.namedEntities.has(e.name)||this.namedEntities.set(e.name,e)},e.prototype.lookup=function(e){return this.namedEntities.has(e)?this.namedEntities.get(e):void 0!==this.parentScope?this.parentScope.lookup(e):null},e.prototype.getChildScope=function(e){var t=this.childScopes.get(e);if(void 0===t)throw new Error("Assertion error: child scope for "+e+" not found");return t},e}(),jm=function(){function e(e,t,n,r){this.matcher=e,this.directives=t,this.bindings=n,this.references=r}return e.apply=function(t,n){var r=new Map,i=new Map,o=new Map;return new e(n,r,i,o).ingest(t),{directives:r,bindings:i,references:o}},e.prototype.ingest=function(e){var t=this;e.forEach(function(e){return e.visit(t)})},e.prototype.visitElement=function(e){this.visitElementOrTemplate(e.name,e)},e.prototype.visitTemplate=function(e){this.visitElementOrTemplate("ng-template",e)},e.prototype.visitElementOrTemplate=function(e,t){var n=this,r=new O;r.setElement(e);var i=Ci(t);Object.getOwnPropertyNames(i).forEach(function(e){var t=i[e];(r.addAttribute(e,t),"class"===e.toLowerCase())&&t.trim().split(/\s+/g).forEach(function(e){return r.addClassName(e)})});var a=[];this.matcher.match(r,function(e,t){return a.push(t)}),a.length>0&&this.directives.set(t,a),t.references.forEach(function(e){var r=null;if(""===e.value.trim())r=a.find(function(e){return e.isComponent})||null;else if(null===(r=a.find(function(t){return t.exportAs===e.value})||null))throw new Error("Assertion error: failed to find directive with exportAs: "+e.value);null!==r?n.references.set(e,{directive:r,node:t}):n.references.set(e,t)}),Object(o.__spread)(t.attributes,t.inputs).forEach(function(e){var r=a.find(function(t){return t.inputs.hasOwnProperty(e.name)});void 0!==r?n.bindings.set(e,r):n.bindings.set(e,t)}),t.outputs.forEach(function(e){var r=a.find(function(t){return t.outputs.hasOwnProperty(e.name)});void 0!==r?n.bindings.set(e,r):n.bindings.set(e,t)}),t.children.forEach(function(e){return e.visit(n)})},e.prototype.visitContent=function(e){},e.prototype.visitVariable=function(e){},e.prototype.visitReference=function(e){},e.prototype.visitTextAttribute=function(e){},e.prototype.visitBoundAttribute=function(e){},e.prototype.visitBoundEvent=function(e){},e.prototype.visitBoundAttributeOrEvent=function(e){},e.prototype.visitText=function(e){},e.prototype.visitBoundText=function(e){},e.prototype.visitIcu=function(e){},e}(),xm=function(e){function t(t,n,r,i,o,a){var s=e.call(this)||this;return s.bindings=t,s.symbols=n,s.nestingLevel=r,s.scope=i,s.template=o,s.level=a,s.visitNode=function(e){return e.visit(s)},s}return Object(o.__extends)(t,e),t.apply=function(e,n){var r=new Map,i=new Map,o=new Map;return new t(r,i,o,n,e instanceof _c?e:null,0).ingest(e),{expressions:r,symbols:i,nestingLevel:o}},t.prototype.ingest=function(e){e instanceof _c?(e.inputs.forEach(this.visitNode),e.outputs.forEach(this.visitNode),e.variables.forEach(this.visitNode),e.children.forEach(this.visitNode),this.nestingLevel.set(e,this.level)):e.forEach(this.visitNode)},t.prototype.visitElement=function(e){e.inputs.forEach(this.visitNode),e.outputs.forEach(this.visitNode),e.children.forEach(this.visitNode)},t.prototype.visitTemplate=function(e){e.inputs.forEach(this.visitNode),e.outputs.forEach(this.visitNode),e.references.forEach(this.visitNode);var n=this.scope.getChildScope(e);new t(this.bindings,this.symbols,this.nestingLevel,n,e,this.level+1).ingest(e)},t.prototype.visitVariable=function(e){null!==this.template&&this.symbols.set(e,this.template)},t.prototype.visitReference=function(e){null!==this.template&&this.symbols.set(e,this.template)},t.prototype.visitText=function(e){},t.prototype.visitContent=function(e){},t.prototype.visitTextAttribute=function(e){},t.prototype.visitIcu=function(e){},t.prototype.visitBoundAttribute=function(e){e.value.visit(this)},t.prototype.visitBoundEvent=function(e){e.handler.visit(this)},t.prototype.visitBoundText=function(e){e.value.visit(this)},t.prototype.visitPropertyRead=function(t,n){return this.maybeMap(n,t,t.name),e.prototype.visitPropertyRead.call(this,t,n)},t.prototype.visitSafePropertyRead=function(t,n){return this.maybeMap(n,t,t.name),e.prototype.visitSafePropertyRead.call(this,t,n)},t.prototype.visitPropertyWrite=function(t,n){return this.maybeMap(n,t,t.name),e.prototype.visitPropertyWrite.call(this,t,n)},t.prototype.visitMethodCall=function(t,n){return this.maybeMap(n,t,t.name),e.prototype.visitMethodCall.call(this,t,n)},t.prototype.visitSafeMethodCall=function(t,n){return this.maybeMap(n,t,t.name),e.prototype.visitSafeMethodCall.call(this,t,n)},t.prototype.maybeMap=function(e,t,n){if(t.receiver instanceof vo){var r=this.scope.lookup(n);null!==r&&this.bindings.set(t,r)}},t}(Vo),Om=function(){function e(e,t,n,r,i,o,a){this.target=e,this.directives=t,this.bindings=n,this.references=r,this.exprTargets=i,this.symbols=o,this.nestingLevel=a}return e.prototype.getDirectivesOfNode=function(e){return this.directives.get(e)||null},e.prototype.getReferenceTarget=function(e){return this.references.get(e)||null},e.prototype.getConsumerOfBinding=function(e){return this.bindings.get(e)||null},e.prototype.getExpressionTarget=function(e){return this.exprTargets.get(e)||null},e.prototype.getTemplateOfSymbol=function(e){return this.symbols.get(e)||null},e.prototype.getNestingLevel=function(e){return this.nestingLevel.get(e)||0},e.prototype.getUsedDirectives=function(){var e=new Set;return this.directives.forEach(function(t){return t.forEach(function(t){return e.add(t)})}),Array.from(e.values())},e}();od(un)},"./node_modules/@angular/compiler/node_modules/tslib/tslib.es6.js":function(e,t,n){"use strict";n.r(t),n.d(t,"__extends",function(){return i}),n.d(t,"__assign",function(){return o}),n.d(t,"__rest",function(){return a}),n.d(t,"__decorate",function(){return s}),n.d(t,"__param",function(){return u}),n.d(t,"__metadata",function(){return c}),n.d(t,"__awaiter",function(){return l}),n.d(t,"__generator",function(){return d}),n.d(t,"__exportStar",function(){return p}),n.d(t,"__values",function(){return f}),n.d(t,"__read",function(){return h}),n.d(t,"__spread",function(){return m}),n.d(t,"__await",function(){return g}),n.d(t,"__asyncGenerator",function(){return v}),n.d(t,"__asyncDelegator",function(){return _}),n.d(t,"__asyncValues",function(){return y}),n.d(t,"__makeTemplateObject",function(){return b}),n.d(t,"__importStar",function(){return j}),n.d(t,"__importDefault",function(){return x});var r=function(e,t){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,t){e.__proto__=t}||function(e,t){for(var n in t)t.hasOwnProperty(n)&&(e[n]=t[n])})(e,t)};function i(e,t){function n(){this.constructor=e}r(e,t),e.prototype=null===t?Object.create(t):(n.prototype=t.prototype,new n)}var o=function(){return(o=Object.assign||function(e){for(var t,n=1,r=arguments.length;n<r;n++){t=arguments[n];for(var i in t)Object.prototype.hasOwnProperty.call(t,i)&&(e[i]=t[i])}return e}).apply(this,arguments)};function a(e,t){var n={};for(var r in e)Object.prototype.hasOwnProperty.call(e,r)&&t.indexOf(r)<0&&(n[r]=e[r]);if(null!=e&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(e);i<r.length;i++)t.indexOf(r[i])<0&&(n[r[i]]=e[r[i]])}return n}function s(e,t,n,r){var i,o=arguments.length,a=o<3?t:null===r?r=Object.getOwnPropertyDescriptor(t,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(e,t,n,r);else for(var s=e.length-1;s>=0;s--)(i=e[s])&&(a=(o<3?i(a):o>3?i(t,n,a):i(t,n))||a);return o>3&&a&&Object.defineProperty(t,n,a),a}function u(e,t){return function(n,r){t(n,r,e)}}function c(e,t){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(e,t)}function l(e,t,n,r){return new(n||(n=Promise))(function(i,o){function a(e){try{u(r.next(e))}catch(e){o(e)}}function s(e){try{u(r.throw(e))}catch(e){o(e)}}function u(e){e.done?i(e.value):new n(function(t){t(e.value)}).then(a,s)}u((r=r.apply(e,t||[])).next())})}function d(e,t){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=t.call(e,a)}catch(e){o=[6,e],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(e,t){for(var n in e)t.hasOwnProperty(n)||(t[n]=e[n])}function f(e){var t="function"==typeof Symbol&&e[Symbol.iterator],n=0;return t?t.call(e):{next:function(){return e&&n>=e.length&&(e=void 0),{value:e&&e[n++],done:!e}}}}function h(e,t){var n="function"==typeof Symbol&&e[Symbol.iterator];if(!n)return e;var r,i,o=n.call(e),a=[];try{for(;(void 0===t||t-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(e){i={error:e}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var e=[],t=0;t<arguments.length;t++)e=e.concat(h(arguments[t]));return e}function g(e){return this instanceof g?(this.v=e,this):new g(e)}function v(e,t,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(e,t||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(e){i[e]&&(r[e]=function(t){return new Promise(function(n,r){o.push([e,t,n,r])>1||s(e,t)})})}function s(e,t){try{(n=i[e](t)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(e){l(o[0][3],e)}var n}function u(e){s("next",e)}function c(e){s("throw",e)}function l(e,t){e(t),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(e){var t,n;return t={},r("next"),r("throw",function(e){throw e}),r("return"),t[Symbol.iterator]=function(){return this},t;function r(r,i){t[r]=e[r]?function(t){return(n=!n)?{value:g(e[r](t)),done:"return"===r}:i?i(t):t}:i}}function y(e){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var t,n=e[Symbol.asyncIterator];return n?n.call(e):(e=f(e),t={},r("next"),r("throw"),r("return"),t[Symbol.asyncIterator]=function(){return this},t);function r(n){t[n]=e[n]&&function(t){return new Promise(function(r,i){var o,a,s,u;t=e[n](t),o=r,a=i,s=t.done,u=t.value,Promise.resolve(u).then(function(e){o({value:e,done:s})},a)})}}}function b(e,t){return Object.defineProperty?Object.defineProperty(e,"raw",{value:t}):e.raw=t,e}function j(e){if(e&&e.__esModule)return e;var t={};if(null!=e)for(var n in e)Object.hasOwnProperty.call(e,n)&&(t[n]=e[n]);return t.default=e,t}function x(e){return e&&e.__esModule?e:{default:e}}},"./node_modules/@angular/core/fesm5/core.js":function(e,t,n){"use strict";n.r(t),n.d(t,"ɵangular_packages_core_core_s",function(){return Im}),n.d(t,"ɵangular_packages_core_core_p",function(){return Am}),n.d(t,"ɵangular_packages_core_core_q",function(){return Tm}),n.d(t,"ɵangular_packages_core_core_r",function(){return Dm}),n.d(t,"ɵangular_packages_core_core_g",function(){return Tf}),n.d(t,"ɵangular_packages_core_core_n",function(){return um}),n.d(t,"ɵangular_packages_core_core_o",function(){return mm}),n.d(t,"ɵangular_packages_core_core_m",function(){return Zh}),n.d(t,"ɵangular_packages_core_core_l",function(){return Xh}),n.d(t,"ɵangular_packages_core_core_c",function(){return Wt}),n.d(t,"ɵangular_packages_core_core_d",function(){return Cf}),n.d(t,"ɵangular_packages_core_core_e",function(){return ff}),n.d(t,"ɵangular_packages_core_core_f",function(){return yf}),n.d(t,"ɵangular_packages_core_core_k",function(){return Id}),n.d(t,"ɵangular_packages_core_core_t",function(){return oh}),n.d(t,"ɵangular_packages_core_core_v",function(){return th}),n.d(t,"ɵangular_packages_core_core_u",function(){return eh}),n.d(t,"ɵangular_packages_core_core_y",function(){return ih}),n.d(t,"ɵangular_packages_core_core_w",function(){return nh}),n.d(t,"ɵangular_packages_core_core_x",function(){return rh}),n.d(t,"ɵangular_packages_core_core_bb",function(){return Wn}),n.d(t,"ɵangular_packages_core_core_bi",function(){return l}),n.d(t,"ɵangular_packages_core_core_bc",function(){return gn}),n.d(t,"ɵangular_packages_core_core_bd",function(){return _n}),n.d(t,"ɵangular_packages_core_core_be",function(){return Rn}),n.d(t,"ɵangular_packages_core_core_bh",function(){return wi}),n.d(t,"ɵangular_packages_core_core_bl",function(){return ct}),n.d(t,"ɵangular_packages_core_core_h",function(){return Fu}),n.d(t,"ɵangular_packages_core_core_i",function(){return Lu}),n.d(t,"ɵangular_packages_core_core_j",function(){return Vu}),n.d(t,"ɵangular_packages_core_core_a",function(){return E}),n.d(t,"ɵangular_packages_core_core_b",function(){return S}),n.d(t,"ɵangular_packages_core_core_bj",function(){return a}),n.d(t,"ɵangular_packages_core_core_z",function(){return wv}),n.d(t,"ɵangular_packages_core_core_ba",function(){return qm}),n.d(t,"createPlatform",function(){return Sh}),n.d(t,"assertPlatform",function(){return Mh}),n.d(t,"destroyPlatform",function(){return Ah}),n.d(t,"getPlatform",function(){return Th}),n.d(t,"PlatformRef",function(){return Dh}),n.d(t,"ApplicationRef",function(){return Ph}),n.d(t,"createPlatformFactory",function(){return kh}),n.d(t,"NgProbeToken",function(){return Eh}),n.d(t,"enableProdMode",function(){return hc}),n.d(t,"isDevMode",function(){return fc}),n.d(t,"APP_ID",function(){return Af}),n.d(t,"PACKAGE_ROOT_URL",function(){return Ff}),n.d(t,"PLATFORM_INITIALIZER",function(){return Pf}),n.d(t,"PLATFORM_ID",function(){return Rf}),n.d(t,"APP_BOOTSTRAP_LISTENER",function(){return Nf}),n.d(t,"APP_INITIALIZER",function(){return kf}),n.d(t,"ApplicationInitStatus",function(){return Mf}),n.d(t,"DebugElement",function(){return sm}),n.d(t,"DebugNode",function(){return am}),n.d(t,"asNativeElements",function(){return $h}),n.d(t,"getDebugNode",function(){return im}),n.d(t,"Testability",function(){return _h}),n.d(t,"TestabilityRegistry",function(){return yh}),n.d(t,"setTestabilityGetter",function(){return bh}),n.d(t,"TRANSLATIONS",function(){return km}),n.d(t,"TRANSLATIONS_FORMAT",function(){return Mm}),n.d(t,"LOCALE_ID",function(){return Sm}),n.d(t,"MissingTranslationStrategy",function(){return ym}),n.d(t,"ApplicationModule",function(){return Pm}),n.d(t,"wtfCreateScope",function(){return sh}),n.d(t,"wtfLeave",function(){return uh}),n.d(t,"wtfStartTimeRange",function(){return ch}),n.d(t,"wtfEndTimeRange",function(){return lh}),n.d(t,"Type",function(){return Fd}),n.d(t,"EventEmitter",function(){return Yl}),n.d(t,"ErrorHandler",function(){return of}),n.d(t,"Sanitizer",function(){return tc}),n.d(t,"SecurityContext",function(){return Xu}),n.d(t,"ANALYZE_FOR_ENTRY_COMPONENTS",function(){return D}),n.d(t,"Attribute",function(){return I}),n.d(t,"ContentChild",function(){return N}),n.d(t,"ContentChildren",function(){return R}),n.d(t,"Query",function(){return P}),n.d(t,"ViewChild",function(){return L}),n.d(t,"ViewChildren",function(){return F}),n.d(t,"Component",function(){return Op}),n.d(t,"Directive",function(){return xp}),n.d(t,"HostBinding",function(){return kp}),n.d(t,"HostListener",function(){return Mp}),n.d(t,"Input",function(){return Ep}),n.d(t,"Output",function(){return Sp}),n.d(t,"Pipe",function(){return wp}),n.d(t,"CUSTOM_ELEMENTS_SCHEMA",function(){return Vp}),n.d(t,"NO_ERRORS_SCHEMA",function(){return Bp}),n.d(t,"NgModule",function(){return zp}),n.d(t,"ViewEncapsulation",function(){return J}),n.d(t,"Version",function(){return nc}),n.d(t,"VERSION",function(){return rc}),n.d(t,"defineInjectable",function(){return g}),n.d(t,"defineInjector",function(){return v}),n.d(t,"forwardRef",function(){return X}),n.d(t,"resolveForwardRef",function(){return Z}),n.d(t,"Injectable",function(){return Qp}),n.d(t,"INJECTOR",function(){return Ns}),n.d(t,"Injector",function(){return Ls}),n.d(t,"inject",function(){return Yt}),n.d(t,"InjectFlags",function(){return Pt}),n.d(t,"ReflectiveInjector",function(){return wf}),n.d(t,"createInjector",function(){return su}),n.d(t,"ResolvedReflectiveFactory",function(){return gf}),n.d(t,"ReflectiveKey",function(){return lf}),n.d(t,"InjectionToken",function(){return b}),n.d(t,"Inject",function(){return Nt}),n.d(t,"Optional",function(){return Ft}),n.d(t,"Self",function(){return Lt}),n.d(t,"SkipSelf",function(){return Vt}),n.d(t,"Host",function(){return Bt}),n.d(t,"NgZone",function(){return dh}),n.d(t,"ɵNoopNgZone",function(){return vh}),n.d(t,"RenderComponentType",function(){return Gu}),n.d(t,"Renderer",function(){return Yu}),n.d(t,"Renderer2",function(){return $u}),n.d(t,"RendererFactory2",function(){return Qu}),n.d(t,"RendererStyleFlags2",function(){return Bu}),n.d(t,"RootRenderer",function(){return Ku}),n.d(t,"COMPILER_OPTIONS",function(){return $f}),n.d(t,"Compiler",function(){return Zf}),n.d(t,"CompilerFactory",function(){return Jf}),n.d(t,"ModuleWithComponentFactories",function(){return Vf}),n.d(t,"ComponentFactory",function(){return Ou}),n.d(t,"ComponentRef",function(){return xu}),n.d(t,"ComponentFactoryResolver",function(){return Au}),n.d(t,"ElementRef",function(){return Uu}),n.d(t,"NgModuleFactory",function(){return Pu}),n.d(t,"NgModuleRef",function(){return Iu}),n.d(t,"NgModuleFactoryLoader",function(){return Md}),n.d(t,"getModuleFactory",function(){return Rd}),n.d(t,"QueryList",function(){return Nh}),n.d(t,"SystemJsNgModuleLoader",function(){return Vh}),n.d(t,"SystemJsNgModuleLoaderConfig",function(){return Fh}),n.d(t,"TemplateRef",function(){return Kl}),n.d(t,"ViewContainerRef",function(){return zh}),n.d(t,"EmbeddedViewRef",function(){return Kh}),n.d(t,"ViewRef",function(){return Yh}),n.d(t,"ChangeDetectionStrategy",function(){return k}),n.d(t,"ChangeDetectorRef",function(){return qh}),n.d(t,"DefaultIterableDiffer",function(){return lm}),n.d(t,"IterableDiffers",function(){return _m}),n.d(t,"KeyValueDiffers",function(){return jm}),n.d(t,"SimpleChange",function(){return Ar}),n.d(t,"WrappedValue",function(){return Mr}),n.d(t,"platformCore",function(){return Em}),n.d(t,"ɵALLOW_MULTIPLE_PLATFORMS",function(){return Ch}),n.d(t,"ɵAPP_ID_RANDOM_PROVIDER",function(){return Df}),n.d(t,"ɵdefaultIterableDiffers",function(){return wm}),n.d(t,"ɵdefaultKeyValueDiffers",function(){return Cm}),n.d(t,"ɵdevModeEqual",function(){return kr}),n.d(t,"ɵisListLikeIterable",function(){return Tr}),n.d(t,"ɵChangeDetectorStatus",function(){return A}),n.d(t,"ɵisDefaultChangeDetectionStrategy",function(){return V}),n.d(t,"ɵConsole",function(){return Lf}),n.d(t,"ɵgetInjectableDef",function(){return _}),n.d(t,"ɵinject",function(){return Yt}),n.d(t,"ɵsetCurrentInjector",function(){return qt}),n.d(t,"ɵAPP_ROOT",function(){return tu}),n.d(t,"ɵivyEnabled",function(){return Nm}),n.d(t,"ɵComponentFactory",function(){return Ou}),n.d(t,"ɵCodegenComponentFactoryResolver",function(){return Tu}),n.d(t,"ɵresolveComponentResources",function(){return $}),n.d(t,"ɵReflectionCapabilities",function(){return Ud}),n.d(t,"ɵRenderDebugInfo",function(){return Wu}),n.d(t,"ɵ_sanitizeHtml",function(){return Lc}),n.d(t,"ɵ_sanitizeStyle",function(){return xd}),n.d(t,"ɵ_sanitizeUrl",function(){return _c}),n.d(t,"ɵglobal",function(){return U}),n.d(t,"ɵlooseIdentical",function(){return Y}),n.d(t,"ɵstringify",function(){return K}),n.d(t,"ɵmakeDecorator",function(){return w}),n.d(t,"ɵisObservable",function(){return Sf}),n.d(t,"ɵisPromise",function(){return Ef}),n.d(t,"ɵclearOverrides",function(){return ay}),n.d(t,"ɵinitServicesIfNeeded",function(){return b_}),n.d(t,"ɵoverrideComponentView",function(){return oy}),n.d(t,"ɵoverrideProvider",function(){return iy}),n.d(t,"ɵNOT_FOUND_CHECK_ONLY_ELEMENT_INJECTOR",function(){return Dv}),n.d(t,"ɵdefineBase",function(){return fe}),n.d(t,"ɵdefineComponent",function(){return ue}),n.d(t,"ɵdefineDirective",function(){return he}),n.d(t,"ɵdefinePipe",function(){return me}),n.d(t,"ɵdefineNgModule",function(){return de}),n.d(t,"ɵdetectChanges",function(){return Ba}),n.d(t,"ɵrenderComponent",function(){return bs}),n.d(t,"ɵRender3ComponentFactory",function(){return cc}),n.d(t,"ɵRender3ComponentRef",function(){return lc}),n.d(t,"ɵdirectiveInject",function(){return as}),n.d(t,"ɵinjectAttribute",function(){return ss}),n.d(t,"ɵgetFactoryOf",function(){return tr}),n.d(t,"ɵgetInheritedFactory",function(){return nr}),n.d(t,"ɵtemplateRefExtractor",function(){return dd}),n.d(t,"ɵProvidersFeature",function(){return ju}),n.d(t,"ɵInheritDefinitionFeature",function(){return Ss}),n.d(t,"ɵNgOnChangesFeature",function(){return As}),n.d(t,"ɵLifecycleHooksFeature",function(){return ws}),n.d(t,"ɵRender3NgModuleRef",function(){return El}),n.d(t,"ɵmarkDirty",function(){return qa}),n.d(t,"ɵNgModuleFactory",function(){return Sl}),n.d(t,"ɵNO_CHANGE",function(){return Pr}),n.d(t,"ɵcontainer",function(){return Ea}),n.d(t,"ɵnextContext",function(){return Po}),n.d(t,"ɵelementStart",function(){return qo}),n.d(t,"ɵnamespaceHTML",function(){return Bo}),n.d(t,"ɵnamespaceMathML",function(){return Vo}),n.d(t,"ɵnamespaceSVG",function(){return Lo}),n.d(t,"ɵelement",function(){return zo}),n.d(t,"ɵlistener",function(){return $o}),n.d(t,"ɵtext",function(){return ha}),n.d(t,"ɵembeddedViewStart",function(){return Aa}),n.d(t,"ɵquery",function(){return cd}),n.d(t,"ɵregisterContentQuery",function(){return us}),n.d(t,"ɵprojection",function(){return Ra}),n.d(t,"ɵbind",function(){return Ga}),n.d(t,"ɵinterpolation1",function(){return Ka}),n.d(t,"ɵinterpolation2",function(){return Qa}),n.d(t,"ɵinterpolation3",function(){return Xa}),n.d(t,"ɵinterpolation4",function(){return Za}),n.d(t,"ɵinterpolation5",function(){return $a}),n.d(t,"ɵinterpolation6",function(){return Ja}),n.d(t,"ɵinterpolation7",function(){return es}),n.d(t,"ɵinterpolation8",function(){return ts}),n.d(t,"ɵinterpolationV",function(){return Ya}),n.d(t,"ɵpipeBind1",function(){return Bl}),n.d(t,"ɵpipeBind2",function(){return zl}),n.d(t,"ɵpipeBind3",function(){return Ul}),n.d(t,"ɵpipeBind4",function(){return Hl}),n.d(t,"ɵpipeBindV",function(){return ql}),n.d(t,"ɵpureFunction0",function(){return Ml}),n.d(t,"ɵpureFunction1",function(){return Al}),n.d(t,"ɵpureFunction2",function(){return Tl}),n.d(t,"ɵpureFunction3",function(){return Dl}),n.d(t,"ɵpureFunction4",function(){return Il}),n.d(t,"ɵpureFunction5",function(){return Pl}),n.d(t,"ɵpureFunction6",function(){return Rl}),n.d(t,"ɵpureFunction7",function(){return Nl}),n.d(t,"ɵpureFunction8",function(){return Fl}),n.d(t,"ɵpureFunctionV",function(){return Ll}),n.d(t,"ɵgetCurrentView",function(){return ds}),n.d(t,"ɵgetHostElement",function(){return br}),n.d(t,"ɵrestoreView",function(){return vn}),n.d(t,"ɵcontainerRefreshStart",function(){return ka}),n.d(t,"ɵcontainerRefreshEnd",function(){return Ma}),n.d(t,"ɵqueryRefresh",function(){return ld}),n.d(t,"ɵloadQueryList",function(){return is}),n.d(t,"ɵelementEnd",function(){return Jo}),n.d(t,"ɵelementProperty",function(){return ta}),n.d(t,"ɵcomponentHostSyntheticProperty",function(){return na}),n.d(t,"ɵprojectionDef",function(){return Ia}),n.d(t,"ɵreference",function(){return rs}),n.d(t,"ɵenableBindings",function(){return hn}),n.d(t,"ɵdisableBindings",function(){return mn}),n.d(t,"ɵallocHostVars",function(){return Wa}),n.d(t,"ɵelementAttribute",function(){return ea}),n.d(t,"ɵelementContainerStart",function(){return Uo}),n.d(t,"ɵelementContainerEnd",function(){return Ho}),n.d(t,"ɵelementStyling",function(){return ua}),n.d(t,"ɵelementHostAttrs",function(){return ca}),n.d(t,"ɵelementStylingMap",function(){return fa}),n.d(t,"ɵelementStyleProp",function(){return da}),n.d(t,"ɵelementStylingApply",function(){return la}),n.d(t,"ɵelementClassProp",function(){return pa}),n.d(t,"ɵtextBinding",function(){return ma}),n.d(t,"ɵtemplate",function(){return Ca}),n.d(t,"ɵembeddedViewEnd",function(){return Ta}),n.d(t,"ɵstore",function(){return ns}),n.d(t,"ɵload",function(){return os}),n.d(t,"ɵpipe",function(){return Vl}),n.d(t,"ɵwhenRendered",function(){return Cs}),n.d(t,"ɵi18n",function(){return pl}),n.d(t,"ɵi18nAttributes",function(){return fl}),n.d(t,"ɵi18nExp",function(){return _l}),n.d(t,"ɵi18nStart",function(){return al}),n.d(t,"ɵi18nEnd",function(){return cl}),n.d(t,"ɵi18nApply",function(){return yl}),n.d(t,"ɵi18nPostprocess",function(){return ul}),n.d(t,"ɵsetClassMetadata",function(){return kl}),n.d(t,"ɵcompileComponent",function(){return hp}),n.d(t,"ɵcompileDirective",function(){return mp}),n.d(t,"ɵcompileNgModule",function(){return Jd}),n.d(t,"ɵcompileNgModuleDefs",function(){return ep}),n.d(t,"ɵpatchComponentDefWithScope",function(){return cp}),n.d(t,"ɵresetCompiledComponents",function(){return ap}),n.d(t,"ɵcompilePipe",function(){return jp}),n.d(t,"ɵsanitizeHtml",function(){return Od}),n.d(t,"ɵsanitizeStyle",function(){return wd}),n.d(t,"ɵsanitizeUrl",function(){return Cd}),n.d(t,"ɵsanitizeResourceUrl",function(){return Ed}),n.d(t,"ɵbypassSanitizationTrustHtml",function(){return hd}),n.d(t,"ɵbypassSanitizationTrustStyle",function(){return md}),n.d(t,"ɵbypassSanitizationTrustScript",function(){return gd}),n.d(t,"ɵbypassSanitizationTrustUrl",function(){return vd}),n.d(t,"ɵbypassSanitizationTrustResourceUrl",function(){return _d}),n.d(t,"ɵgetLContext",function(){return rr}),n.d(t,"ɵbindPlayerFactory",function(){return Oi}),n.d(t,"ɵaddPlayer",function(){return hs}),n.d(t,"ɵgetPlayers",function(){return ms}),n.d(t,"ɵcompileNgModuleFactory__POST_R3__",function(){return wh}),n.d(t,"ɵSWITCH_COMPILE_COMPONENT__POST_R3__",function(){return Ap}),n.d(t,"ɵSWITCH_COMPILE_DIRECTIVE__POST_R3__",function(){return Tp}),n.d(t,"ɵSWITCH_COMPILE_PIPE__POST_R3__",function(){return Dp}),n.d(t,"ɵSWITCH_COMPILE_NGMODULE__POST_R3__",function(){return Up}),n.d(t,"ɵgetDebugNode__POST_R3__",function(){return rm}),n.d(t,"ɵSWITCH_COMPILE_INJECTABLE__POST_R3__",function(){return Xp}),n.d(t,"ɵSWITCH_IVY_ENABLED__POST_R3__",function(){return Rm}),n.d(t,"ɵSWITCH_CHANGE_DETECTOR_REF_FACTORY__POST_R3__",function(){return Gh}),n.d(t,"ɵCompiler_compileModuleSync__POST_R3__",function(){return Hf}),n.d(t,"ɵCompiler_compileModuleAsync__POST_R3__",function(){return Gf}),n.d(t,"ɵCompiler_compileModuleAndAllComponentsSync__POST_R3__",function(){return Yf}),n.d(t,"ɵCompiler_compileModuleAndAllComponentsAsync__POST_R3__",function(){return Qf}),n.d(t,"ɵSWITCH_ELEMENT_REF_FACTORY__POST_R3__",function(){return Hu}),n.d(t,"ɵSWITCH_TEMPLATE_REF_FACTORY__POST_R3__",function(){return Ql}),n.d(t,"ɵSWITCH_VIEW_CONTAINER_REF_FACTORY__POST_R3__",function(){return Uh}),n.d(t,"ɵSWITCH_RENDERER2_FACTORY__POST_R3__",function(){return Ju}),n.d(t,"ɵgetModuleFactory__POST_R3__",function(){return Pd}),n.d(t,"ɵpublishGlobalUtil",function(){return ys}),n.d(t,"ɵpublishDefaultGlobalUtils",function(){return _s}),n.d(t,"ɵSWITCH_INJECTOR_FACTORY__POST_R3__",function(){return Vs}),n.d(t,"ɵregisterModuleFactory",function(){return Td}),n.d(t,"ɵEMPTY_ARRAY",function(){return Tg}),n.d(t,"ɵEMPTY_MAP",function(){return Dg}),n.d(t,"ɵand",function(){return Ig}),n.d(t,"ɵccf",function(){return ev}),n.d(t,"ɵcmf",function(){return sy}),n.d(t,"ɵcrt",function(){return tg}),n.d(t,"ɵdid",function(){return jv}),n.d(t,"ɵeld",function(){return Pg}),n.d(t,"ɵelementEventFullName",function(){return pg}),n.d(t,"ɵgetComponentViewDefinitionFactory",function(){return tv}),n.d(t,"ɵinlineInterpolate",function(){return Mg}),n.d(t,"ɵinterpolate",function(){return kg}),n.d(t,"ɵmod",function(){return qg}),n.d(t,"ɵmpd",function(){return Hg}),n.d(t,"ɵncd",function(){return qv}),n.d(t,"ɵnov",function(){return cv}),n.d(t,"ɵpid",function(){return xv}),n.d(t,"ɵprd",function(){return Ov}),n.d(t,"ɵpad",function(){return Wv}),n.d(t,"ɵpod",function(){return Yv}),n.d(t,"ɵppd",function(){return Gv}),n.d(t,"ɵqud",function(){return Vv}),n.d(t,"ɵted",function(){return Qv}),n.d(t,"ɵunv",function(){return $m}),n.d(t,"ɵvid",function(){return $v});var r=n("./node_modules/@angular/core/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/index.js"),o=n("./node_modules/rxjs/_esm5/operators/index.js");function a(e){for(var t in e)if(e[t]===a)return t;throw Error("Could not find renamed property on target object.")}function s(e,t){for(var n in t)t.hasOwnProperty(n)&&!e.hasOwnProperty(n)&&(e[n]=t[n])}var u=a({ngComponentDef:a}),c=a({ngDirectiveDef:a}),l=a({ngInjectableDef:a}),d=a({ngInjectorDef:a}),p=a({ngPipeDef:a}),f=a({ngModuleDef:a}),h=a({ngBaseDef:a}),m=a({__NG_ELEMENT_ID__:a});function g(e){return{providedIn:e.providedIn||null,factory:e.factory,value:void 0}}function v(e){return{factory:e.factory,providers:e.providers||[],imports:e.imports||[]}}function _(e){return e&&e.hasOwnProperty(l)?e[l]:null}function y(e){return e&&e.hasOwnProperty(d)?e[d]:null}var b=function(){function e(e,t){this._desc=e,this.ngMetadataName="InjectionToken",this.ngInjectableDef=void 0!==t?g({providedIn:t.providedIn||"root",factory:t.factory}):void 0}return e.prototype.toString=function(){return"InjectionToken "+this._desc},e}(),j="__annotations__",x="__parameters__",O="__prop__metadata__";function w(e,t,n,i,o){var a=C(t);function s(){for(var e,t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];if(this instanceof s)return a.call.apply(a,Object(r.__spread)([this],t)),this;var u=new((e=s).bind.apply(e,Object(r.__spread)([void 0],t)));return function(e){return o&&o.apply(void 0,Object(r.__spread)([e],t)),(e.hasOwnProperty(j)?e[j]:Object.defineProperty(e,j,{value:[]})[j]).push(u),i&&i(e),e}}return n&&(s.prototype=Object.create(n.prototype)),s.prototype.ngMetadataName=e,s.annotationCls=s,s}function C(e){return function(){for(var t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];if(e){var i=e.apply(void 0,Object(r.__spread)(t));for(var o in i)this[o]=i[o]}}}function E(e,t,n){var i=C(t);function o(){for(var e,t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];if(this instanceof o)return i.apply(this,t),this;var a=new((e=o).bind.apply(e,Object(r.__spread)([void 0],t)));return s.annotation=a,s;function s(e,t,n){for(var r=e.hasOwnProperty(x)?e[x]:Object.defineProperty(e,x,{value:[]})[x];r.length<=n;)r.push(null);return(r[n]=r[n]||[]).push(a),e}}return n&&(o.prototype=Object.create(n.prototype)),o.prototype.ngMetadataName=e,o.annotationCls=o,o}function S(e,t,n,i){var o=C(t);function a(){for(var e,t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];if(this instanceof a)return o.apply(this,t),this;var s=new((e=a).bind.apply(e,Object(r.__spread)([void 0],t)));return function(e,n){var o=e.constructor,a=o.hasOwnProperty(O)?o[O]:Object.defineProperty(o,O,{value:{}})[O];a[n]=a.hasOwnProperty(n)&&a[n]||[],a[n].unshift(s),i&&i.apply(void 0,Object(r.__spread)([e,n],t))}}return n&&(a.prototype=Object.create(n.prototype)),a.prototype.ngMetadataName=e,a.annotationCls=a,a}var k,M,A,T,D=new b("AnalyzeForEntryComponents"),I=E("Attribute",function(e){return{attributeName:e}}),P=function(){return function(){}}(),R=S("ContentChildren",function(e,t){return void 0===t&&(t={}),Object(r.__assign)({selector:e,first:!1,isViewQuery:!1,descendants:!1},t)},P),N=S("ContentChild",function(e,t){return void 0===t&&(t={}),Object(r.__assign)({selector:e,first:!0,isViewQuery:!1,descendants:!0},t)},P),F=S("ViewChildren",function(e,t){return void 0===t&&(t={}),Object(r.__assign)({selector:e,first:!1,isViewQuery:!0,descendants:!0},t)},P),L=S("ViewChild",function(e,t){return Object(r.__assign)({selector:e,first:!0,isViewQuery:!0,descendants:!0},t)},P);function V(e){return null==e||e===k.Default}(M=k||(k={}))[M.OnPush=0]="OnPush",M[M.Default=1]="Default",(T=A||(A={}))[T.CheckOnce=0]="CheckOnce",T[T.Checked=1]="Checked",T[T.CheckAlways=2]="CheckAlways",T[T.Detached=3]="Detached",T[T.Errored=4]="Errored",T[T.Destroyed=5]="Destroyed";var B="undefined"!=typeof window&&window,z="undefined"!=typeof self&&"undefined"!=typeof WorkerGlobalScope&&self instanceof WorkerGlobalScope&&self,U="undefined"!=typeof global&&global||B||z,H=Promise.resolve(0),q=null;function G(){if(!q){var e=U.Symbol;if(e&&e.iterator)q=e.iterator;else for(var t=Object.getOwnPropertyNames(Map.prototype),n=0;n<t.length;++n){var r=t[n];"entries"!==r&&"size"!==r&&Map.prototype[r]===Map.prototype.entries&&(q=r)}}return q}function W(e){"undefined"==typeof Zone?H.then(function(){e&&e.apply(null,null)}):Zone.current.scheduleMicroTask("scheduleMicrotask",e)}function Y(e,t){return e===t||"number"==typeof e&&"number"==typeof t&&isNaN(e)&&isNaN(t)}function K(e){if("string"==typeof e)return e;if(e instanceof Array)return"["+e.map(K).join(", ")+"]";if(null==e)return""+e;if(e.overriddenName)return""+e.overriddenName;if(e.name)return""+e.name;var t=e.toString();if(null==t)return""+t;var n=t.indexOf("\n");return-1===n?t:t.substring(0,n)}var Q=a({__forward_ref__:a});function X(e){return e.__forward_ref__=X,e.toString=function(){return K(this())},e}function Z(e){var t=e;return"function"==typeof t&&t.hasOwnProperty(Q)&&t.__forward_ref__===X?t():e}function $(e){var t=[],n=new Map;function r(r){var i=n.get(r);if(!i){var o=e(r);n.set(r,i=o.then(ie)),t.push(i)}return i}return ne.forEach(function(e){e.templateUrl&&r(e.templateUrl).then(function(t){e.template=t,e.templateUrl=void 0});var t=e.styleUrls,n=e.styles||(e.styles=[]),i=e.styles.length;t&&t.forEach(function(o,a){n.push(""),r(o).then(function(r){n[i+a]=r,t.splice(t.indexOf(o),1),0==t.length&&(e.styleUrls=void 0)})})}),ne.clear(),Promise.all(t).then(function(){return null})}var J,ee,te,ne=new Set;function re(e){return e.templateUrl||e.styleUrls&&e.styleUrls.length}function ie(e){return"string"==typeof e?e:e.text()}(ee=J||(J={}))[ee.Emulated=0]="Emulated",ee[ee.Native=1]="Native",ee[ee.None=2]="None",ee[ee.ShadowDom=3]="ShadowDom",("undefined"==typeof ngDevMode||ngDevMode)&&(te={firstTemplatePass:0,tNode:0,tView:0,rendererCreateTextNode:0,rendererSetText:0,rendererCreateElement:0,rendererAddEventListener:0,rendererSetAttribute:0,rendererRemoveAttribute:0,rendererSetProperty:0,rendererSetClassName:0,rendererAddClass:0,rendererRemoveClass:0,rendererSetStyle:0,rendererRemoveStyle:0,rendererDestroy:0,rendererDestroyNode:0,rendererMoveNode:0,rendererRemoveNode:0,rendererCreateComment:0},"undefined"!=typeof window&&(window.ngDevMode=te),"undefined"!=typeof global&&(global.ngDevMode=te),"undefined"!=typeof self&&(self.ngDevMode=te));var oe={},ae=[];"undefined"!=typeof ngDevMode&&ngDevMode&&(Object.freeze(oe),Object.freeze(ae));var se=0;function ue(e){var t=e.type,n=t.prototype,r={},i={type:t,providersResolver:null,consts:e.consts,vars:e.vars,factory:e.factory,template:e.template||null,hostBindings:e.hostBindings||null,contentQueries:e.contentQueries||null,contentQueriesRefresh:e.contentQueriesRefresh||null,attributes:e.attributes||null,declaredInputs:r,inputs:null,outputs:null,exportAs:e.exportAs||null,onInit:n.ngOnInit||null,doCheck:n.ngDoCheck||null,afterContentInit:n.ngAfterContentInit||null,afterContentChecked:n.ngAfterContentChecked||null,afterViewInit:n.ngAfterViewInit||null,afterViewChecked:n.ngAfterViewChecked||null,onDestroy:n.ngOnDestroy||null,onPush:e.changeDetection===k.OnPush,directiveDefs:null,pipeDefs:null,selectors:e.selectors,viewQuery:e.viewQuery||null,features:e.features||null,data:e.data||{},encapsulation:e.encapsulation||J.Emulated,id:"c",styles:e.styles||ae,_:null};return i._=""+{toString:function(){var t=e.directives,n=e.features,o=e.pipes;i.id+=se++,i.inputs=pe(e.inputs,r),i.outputs=pe(e.outputs),n&&n.forEach(function(e){return e(i)}),i.directiveDefs=t?function(){return("function"==typeof t?t():t).map(ce)}:null,i.pipeDefs=o?function(){return("function"==typeof o?o():o).map(le)}:null}},i}function ce(e){var t=ge(e)||ve(e);if(ngDevMode&&!t)throw new Error("'"+e.name+"' is neither 'ComponentType' or 'DirectiveType'.");return t}function le(e){var t=_e(e);if(ngDevMode&&!t)throw new Error("'"+e.name+"' is not a 'PipeType'.");return t}function de(e){return{type:e.type,bootstrap:e.bootstrap||ae,declarations:e.declarations||ae,imports:e.imports||ae,exports:e.exports||ae,transitiveCompileScopes:null}}function pe(e,t){if(null==e)return oe;var n={};for(var r in e)if(e.hasOwnProperty(r)){var i=e[r],o=i;Array.isArray(i)&&(o=i[1],i=i[0]),n[i]=r,t&&(t[i]=o)}return n}function fe(e){var t={};return{inputs:pe(e.inputs,t),declaredInputs:t,outputs:pe(e.outputs)}}var he=ue;function me(e){return{name:e.name,factory:e.factory,pure:!1!==e.pure,onDestroy:e.type.prototype.ngOnDestroy||null}}function ge(e){return e[u]||null}function ve(e){return e[c]||null}function _e(e){return e[p]||null}function ye(e,t){var n=e[f]||null;if(!n&&!0===t)throw new Error("Type "+K(e)+" does not have 'ngModuleDef' property.");return n}function be(e,t,n){e!=t&&Ee(n)}function je(e,t,n){e==t&&Ee(n)}function xe(e,t,n){e>=t&&Ee(n)}function Oe(e,t,n){e<=t&&Ee(n)}function we(e,t){null==e&&Ee(t)}function Ce(e,t){void 0===t&&(t="Type passed in is not ComponentType, it does not have 'ngComponentDef' property."),ge(e)||Ee(t)}function Ee(e){throw new Error("ASSERTION ERROR: "+e)}function Se(e){be(e instanceof Node,!0,"The provided value must be an instance of a DOM Node")}function ke(e){be(e,!0,"previousOrParentTNode should be a parent")}function Me(e){we(e.parent,"previousOrParentTNode should have a parent")}function Ae(e,t){xe(t,e?e.length:0,"index expected to be a valid data index")}var Te=0,De=1,Ie=2,Pe=3,Re=4,Ne=5,Fe=6,Le=7,Ve=8,Be=9,ze=10,Ue=11,He=12,qe=13,Ge=14,We=15,Ye=16,Ke=17,Qe=18,Xe=0,Ze=1,$e=6,Je=7,et=8,tt="__ngContext__",nt=8,rt=8,it=9,ot=-1,at=function(){return function(e,t,n){this.factory=e,this.resolving=!1,this.canSeeViewProviders=t,this.injectImpl=n}}(),st=at.prototype;function ut(e){return"function"==typeof e?e.name||e:"string"==typeof e?e:null==e?"":"object"==typeof e&&"function"==typeof e.type?e.type.name||e.type:""+e}function ct(e,t){return ngDevMode&&Ae(e,t+Qe),e[t+Qe]}function lt(e){for(;Array.isArray(e);)e=e[Ne];return e}function dt(e,t){return lt(t[e+Qe])}function pt(e,t){return lt(t[e.index])}function ft(e,t){return ngDevMode&&Oe(e,-1,"wrong index for TNode"),ngDevMode&&xe(e,t[Te].data.length,"wrong index for TNode"),t[Te].data[e+Qe]}function ht(e,t){var n=t[e];return n.length>=Qe?n:n[Ne]}function mt(e){return 0!=(4&e.flags)}function gt(e){return 1==(1&e.flags)}function vt(e){return null!==e.template}function _t(e){return Array.isArray(e)&&e.length===et}function yt(e){return 0!=(128&e[De])}function bt(e){ngDevMode&&we(e,"component");for(var t=Array.isArray(e)?e:Ot(e);t&&!(128&t[De]);)t=t[Ie];return t}function jt(e){var t=bt(e);return ngDevMode&&we(t[Be],"RootView has no context. Perhaps it is disconnected?"),t[Be]}function xt(e){return ngDevMode&&we(e,"Target expected"),e[tt]}function Ot(e){var t=xt(e);return t?Array.isArray(t)?t:t.lView:null}function wt(e){return e!==ot}function Ct(e){return 32767&e}function Et(e){return e>>16}function St(e,t){for(var n=Et(e),r=t;n>0;)r=r[Ke],n--;return r}var kt,Mt,At=("undefined"!=typeof requestAnimationFrame&&requestAnimationFrame||setTimeout).bind(U);function Tt(e,t){for(var n=0;n<e.length;n++)t.push(e[n])}function Dt(e){for(var t=e[Fe];t&&2===t.type;)ngDevMode&&we(e[Ke],"lView[DECLARATION_VIEW]"),t=(e=e[Ke])[Fe];return e}function It(){var e=U.ng;if(!e||!e.ɵcompilerFacade)throw new Error("Angular JIT compilation failed: '@angular/compiler' not loaded!\n  - JIT compilation is discouraged for production use-cases! Consider AOT mode instead.\n  - Did you bootstrap using '@angular/platform-browser-dynamic' or '@angular/platform-server'?\n  - Alternatively provide the compiler with 'import \"@angular/compiler\";' before bootstrapping.");return e.ɵcompilerFacade}(Mt=kt||(kt={}))[Mt.Token=0]="Token",Mt[Mt.Attribute=1]="Attribute";var Pt,Rt,Nt=E("Inject",function(e){return{token:e}}),Ft=E("Optional"),Lt=E("Self"),Vt=E("SkipSelf"),Bt=E("Host");(Rt=Pt||(Pt={}))[Rt.Default=0]="Default",Rt[Rt.Host=1]="Host",Rt[Rt.Self=2]="Self",Rt[Rt.SkipSelf=4]="SkipSelf",Rt[Rt.Optional=8]="Optional";var zt,Ut,Ht=void 0;function qt(e){var t=Ht;return Ht=e,t}function Gt(e){var t=zt;return zt=e,t}function Wt(e,t){if(void 0===t&&(t=Pt.Default),void 0===Ht)throw new Error("inject() must be called from an injection context");return null===Ht?Kt(e,void 0,t):Ht.get(e,t&Pt.Optional?null:void 0,t)}function Yt(e,t){return void 0===t&&(t=Pt.Default),(zt||Wt)(e,t)}function Kt(e,t,n){var r=_(e);if(r&&"root"==r.providedIn)return void 0===r.value?r.value=r.factory():r.value;if(n&Pt.Optional)return null;if(void 0!==t)return t;throw new Error("Injector: NOT_FOUND ["+K(e)+"]")}function Qt(e){for(var t=[],n=0;n<e.length;n++){var r=e[n];if(Array.isArray(r)){if(0===r.length)throw new Error("Arguments array must have arguments.");for(var i=void 0,o=Pt.Default,a=0;a<r.length;a++){var s=r[a];s instanceof Ft||"Optional"===s.ngMetadataName?o|=Pt.Optional:s instanceof Vt||"SkipSelf"===s.ngMetadataName?o|=Pt.SkipSelf:s instanceof Lt||"Self"===s.ngMetadataName?o|=Pt.Self:i=s instanceof Nt?s.token:s}t.push(Yt(i,o))}else t.push(Yt(r))}return t}function Xt(e,t){we(e,"should be called with a TNode"),be(e.type,t,"should be a "+$t(t))}function Zt(e){for(var t=[],n=1;n<arguments.length;n++)t[n-1]=arguments[n];we(e,"should be called with a TNode"),be(t.some(function(t){return e.type===t}),!0,"Should be one of "+t.map($t).join(", ")+" but got "+$t(e.type))}function $t(e){return 1==e?"Projection":0==e?"Container":2==e?"View":3==e?"Element":4==e?"ElementContainer":"<unknown>"}function Jt(e,t,n,r){ngDevMode&&be(r.firstTemplatePass,!0,"Should only be called on first template pass"),t&&(r.initHooks||(r.initHooks=[])).push(e,t),n&&((r.initHooks||(r.initHooks=[])).push(e,n),(r.checkHooks||(r.checkHooks=[])).push(e,n))}function en(e,t){if(e.firstTemplatePass)for(var n=t.directiveStart,r=t.directiveEnd;n<r;n++){var i=e.data[n];tn(i,e,n),nn(i,e,n),rn(i,e,n)}}function tn(e,t,n){e.afterContentInit&&(t.contentHooks||(t.contentHooks=[])).push(n,e.afterContentInit),e.afterContentChecked&&((t.contentHooks||(t.contentHooks=[])).push(n,e.afterContentChecked),(t.contentCheckHooks||(t.contentCheckHooks=[])).push(n,e.afterContentChecked))}function nn(e,t,n){e.afterViewInit&&(t.viewHooks||(t.viewHooks=[])).push(n,e.afterViewInit),e.afterViewChecked&&((t.viewHooks||(t.viewHooks=[])).push(n,e.afterViewChecked),(t.viewCheckHooks||(t.viewCheckHooks=[])).push(n,e.afterViewChecked))}function rn(e,t,n){null!=e.onDestroy&&(t.destroyHooks||(t.destroyHooks=[])).push(n,e.onDestroy)}function on(e,t,n){!n&&32&e[De]&&(an(e,t.initHooks,t.checkHooks,n),e[De]&=-33)}function an(e,t,n,r){if(!r){var i=2&e[De]?t:n;i&&sn(e,i)}}function sn(e,t){for(var n=0;n<t.length;n+=2)t[n+1].call(e[t[n]])}var un,cn,ln,dn,pn=null;function fn(e){pn=e}function hn(){un=!0}function mn(){un=!1}function gn(){return dn}function vn(e){wn=e}function _n(){return cn}function yn(e){cn=e}function bn(e,t){cn=e,dn=t}function jn(){return ln}function xn(e){ln=e}function On(e){return void 0===e&&(e=dn),1==(1&e[De])}var wn=null;var Cn=!1;function En(){return Cn}function Sn(e){Cn=e}var kn=!0;function Mn(){return kn}function An(e){kn=e}var Tn=-1;function Dn(){return Tn}function In(e){Tn=e}function Pn(e,t){var n=dn;if(e){var r=e[Te];kn=r.firstTemplatePass,Tn=r.bindingStartIndex}return cn=t,ln=!0,dn=wn=e,n}function Rn(e){return void 0===e&&(e=1),(wn=function(e,t){for(;e>0;)ngDevMode&&we(t[Ke],"Declaration view should be defined if nesting level is greater than 0."),t=t[Ke],e--;return t}(e,wn))[Be]}function Nn(e){var t=dn[Te];On(dn)?dn[De]&=-2:(an(dn,t.viewHooks,t.viewCheckHooks,Cn),dn[De]&=-11,dn[De]|=32,dn[Le]=t.bindingStartIndex),Pn(e,null)}var Fn=!0;function Ln(e){var t=Fn;return Fn=e,t}var Vn=255,Bn=0;function zn(e,t){var n=Hn(e,t);if(-1!==n)return n;var r=t[Te];r.firstTemplatePass&&(e.injectorIndex=t.length,Un(r.data,e),Un(t,null),Un(r.blueprint,null),ngDevMode&&be(0===e.flags||1===e.flags,!0,"expected tNode.flags to not be initialized"));var i=qn(e,t),o=Ct(i),a=St(i,t),s=e.injectorIndex;if(wt(i))for(var u=a[Te].data,c=0;c<8;c++)t[s+c]=a[o+c]|u[o+c];return t[s+rt]=i,s}function Un(e,t){e.push(0,0,0,0,0,0,0,0,t)}function Hn(e,t){return-1===e.injectorIndex||e.parent&&e.parent.injectorIndex===e.injectorIndex||null==t[e.injectorIndex+rt]?-1:e.injectorIndex}function qn(e,t){if(e.parent&&-1!==e.parent.injectorIndex)return e.parent.injectorIndex;for(var n=t[Fe],r=1;n&&-1===n.injectorIndex;)n=(t=t[Ke])?t[Fe]:null,r++;return n?n.injectorIndex|r<<16:-1}function Gn(e,t,n){!function(e,t,n){ngDevMode&&be(t.firstTemplatePass,!0,"expected firstTemplatePass to be true");var r="string"!=typeof n?n[m]:n.charCodeAt(0)||0;null==r&&(r=n[m]=Bn++);var i=r&Vn,o=1<<i,a=128&i,s=64&i,u=32&i,c=t.data;a?s?u?c[e+7]|=o:c[e+6]|=o:u?c[e+5]|=o:c[e+4]|=o:s?u?c[e+3]|=o:c[e+2]|=o:u?c[e+1]|=o:c[e]|=o}(e,t[Te],n)}function Wn(e,t){ngDevMode&&Zt(e,0,3,4),ngDevMode&&we(e,"expecting tNode");var n=e.attrs;if(n)for(var r=0;r<n.length;r+=2){var i=n[r];if(3===i)break;if(i==t)return n[r+1]}return null}function Yn(e,t,n,r,i){if(void 0===r&&(r=Pt.Default),e){var o=function(e){if(ngDevMode&&we(e,"token must be defined"),"string"==typeof e)return e.charCodeAt(0)||0;var t=e[m];return"number"==typeof t?t&Vn:t}(n);if("function"==typeof o){var a=_n(),s=gn();bn(e,t);try{var u=o();if(null!=u||r&Pt.Optional)return u;throw new Error("No provider for "+ut(n)+"!")}finally{bn(a,s)}}else if("number"==typeof o){var c=null,l=Hn(e,t),d=ot,p=r&Pt.Host?Dt(t)[Fe]:null;for((-1===l||r&Pt.SkipSelf)&&(d=-1===l?qn(e,t):t[l+rt],Jn(r,!1)?(c=t[Te],l=Ct(d),t=St(d,t)):l=-1);-1!==l;){d=t[l+rt];var f=t[Te];if($n(o,l,f.data)){var h=Qn(l,t,n,c,r,p);if(h!==Kn)return h}Jn(r,t[Te].data[l+nt]===p)&&$n(o,l,t)?(c=f,l=Ct(d),t=St(d,t)):l=-1}}}if(r&Pt.Optional&&void 0===i&&(i=null),0==(r&(Pt.Self|Pt.Host))){var g=t[ze];return g?g.get(n,i,r&Pt.Optional):Kt(n,i,r&Pt.Optional)}if(r&Pt.Optional)return i;throw new Error("NodeInjector: NOT_FOUND ["+ut(n)+"]")}var Kn={};function Qn(e,t,n,r,i,o){var a=t[Te],s=a.data[e+nt],u=Xn(s,t,n,null==r?gt(s)&&Fn:r!=a&&3===s.type,i&Pt.Host&&o===s);return null!==u?Zn(a.data,t,u,s):Kn}function Xn(e,t,n,r,i){for(var o=t[Te],a=e.providerIndexes,s=o.data,u=65535&a,c=e.directiveStart,l=e.directiveEnd,d=a>>16,p=i?u+d:l,f=r?u:u+d;f<p;f++){var h=s[f];if(f<c&&n===h||f>=c&&h.type===n)return f}if(i){var m=s[c];if(m&&vt(m)&&m.type===n)return c}return null}function Zn(e,t,n,r){var i,o=t[n];if(null!=(i=o)&&"object"==typeof i&&Object.getPrototypeOf(i)==st){var a=o;if(a.resolving)throw new Error("Circular dep for "+ut(e[n]));var s=Ln(a.canSeeViewProviders);a.resolving=!0;var u=void 0;a.injectImpl&&(u=Gt(a.injectImpl));var c=_n(),l=gn();bn(r,t);try{o=t[n]=a.factory(null,e,t,r)}finally{a.injectImpl&&Gt(u),Ln(s),a.resolving=!1,bn(c,l)}}return o}function $n(e,t,n){var r=1<<e,i=64&e,o=32&e;return!!((128&e?i?o?n[t+7]:n[t+6]:o?n[t+5]:n[t+4]:i?o?n[t+3]:n[t+2]:o?n[t+1]:n[t])&r)}function Jn(e,t){return!(e&Pt.Self||e&Pt.Host&&t)}var er=function(){function e(e,t){this._tNode=e,this._lView=t}return e.prototype.get=function(e,t){return Yn(this._tNode,this._lView,e,void 0,t)},e}();function tr(e){var t=e,n=ge(t)||ve(t)||_e(t)||_(t)||y(t);return n&&void 0!==n.factory?n.factory:null}function nr(e){var t=tr(Object.getPrototypeOf(e.prototype).constructor);return null!==t?t:function(e){return new e}}function rr(e){var t,n,r=xt(e);if(r){if(Array.isArray(r)){var i=r,o=void 0,a=void 0,s=void 0;if((n=e)&&n.constructor&&n.constructor.ngComponentDef){if(-1==(o=cr(i,e)))throw new Error("The provided component was not found in the application");a=e}else if(t=e,t&&t.constructor&&t.constructor.ngDirectiveDef){if(-1==(o=function(e,t){var n=e[Te].firstChild;for(;n;){for(var r=n.directiveStart,i=n.directiveEnd,o=r;o<i;o++)if(e[o]===t)return n.index;n=ur(n)}return-1}(i,e)))throw new Error("The provided directive was not found in the application");s=lr(o,i,!1)}else if(-1==(o=sr(i,e)))return null;var u=xt(m=lt(i[o])),c=u&&!Array.isArray(u)?u:ir(i,o,m);if(a&&void 0===c.component&&(c.component=a,ar(c.component,c)),s&&void 0===c.directives){c.directives=s;for(var l=0;l<s.length;l++)ar(s[l],c)}ar(c.native,c),r=c}}else{var d=e;ngDevMode&&Se(d);for(var p=d;p=p.parentNode;){var f=xt(p);if(f){i=void 0;if(!(i=Array.isArray(f)?f:f.lView))return null;var h=sr(i,d);if(h>=0){var m;ar(m=lt(i[h]),c=ir(i,h,m)),r=c;break}}}}return r||null}function ir(e,t,n){return{lView:e,nodeIndex:t,native:n,component:void 0,directives:void 0,localRefs:void 0}}function or(e){var t,n=xt(e);if(Array.isArray(n)){var r=cr(n,e);(i=ir(n,r,(t=ht(r,n))[Ne])).component=e,ar(e,i),ar(i.native,i)}else{var i;t=ht((i=n).nodeIndex,i.lView)}return t}function ar(e,t){e[tt]=t}function sr(e,t){for(var n=e[Te].firstChild;n;){if(pt(n,e)===t)return n.index;n=ur(n)}return-1}function ur(e){if(e.child)return e.child;if(e.next)return e.next;for(;e.parent&&!e.parent.next;)e=e.parent;return e.parent&&e.parent.next}function cr(e,t){var n=e[Te].components;if(n)for(var r=0;r<n.length;r++){var i=n[r];if(ht(i,e)[Be]===t)return i}else if(ht(Qe,e)[Be]===t)return Qe;return-1}function lr(e,t,n){var r=t[Te].data[e],i=r.directiveStart;if(0==i)return ae;var o=r.directiveEnd;return!n&&1&r.flags&&i++,t.slice(i,o)}function dr(e){var t,n,r,i,o=jr(e);return void 0===o.component&&(o.component=(t=o.nodeIndex,n=o.lView,r=n[Te].data[t],i=r.directiveStart,1&r.flags?n[i]:null)),o.component}function pr(e){return jr(e).lView[Be]}function fr(e){for(var t=_r(e).lView;t[Ie]&&null===t[Ne];)t=t[Ie];return 128&t[De]?null:t[Be]}function hr(e){return function(e){var t;Array.isArray(e)?(ngDevMode&&we(e,"lView"),t=e):(ngDevMode&&we(e,"component"),t=Ot(e));for(;t&&!(128&t[De]);)t=t[Ie];return t}(Array.isArray(e)?e:_r(e).lView)[Be]}function mr(e){return Object(r.__spread)(hr(e).components)}function gr(e){var t=_r(e),n=t.lView[Te].data[t.nodeIndex];return new er(n,t.lView)}function vr(e){var t=_r(e);return void 0===t.directives&&(t.directives=lr(t.nodeIndex,t.lView,!1)),t.directives||[]}function _r(e,t){void 0===t&&(t=!0);var n=rr(e);if(!n&&t)throw new Error(ngDevMode?"Unable to find context associated with "+ut(e):"Invalid ng target");return n}function yr(e){var t=_r(e);return void 0===t.localRefs&&(t.localRefs=function(e,t){var n=e[Te].data[t];if(n&&n.localNames){for(var r={},i=0;i<n.localNames.length;i+=2){var o=n.localNames[i],a=n.localNames[i+1];r[o]=-1===a?pt(n,e):e[a]}return r}return null}(t.lView,t.nodeIndex)),t.localRefs||{}}function br(e){return rr(e).native}function jr(e){if(!(e instanceof Node))throw new Error("Expecting instance of DOM Node");return _r(e)}function xr(e){return"boolean"==typeof e.useCapture}function Or(e){var t=jr(e).lView,n=t[Te],r=t[Ve],i=n.cleanup,o=[];if(i&&r)for(var a=0;a<i.length;){var s=i[a++],u=i[a++];if("string"==typeof s){var c=s,l=lt(t[u]),d=r[i[a++]],p=i[a++],f="boolean"==typeof p?p:!(p>=0)&&null;e==l&&o.push({element:e,name:c,callback:d,useCapture:f})}}return o.sort(wr),o}function wr(e,t){return e.name==t.name?0:e.name<t.name?-1:1}function Cr(e){return"ng-reflect-"+(e=e.replace(/[$@]/g,"_").replace(Er,function(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];return"-"+e[1].toLowerCase()}))}var Er=/([A-Z])/g;function Sr(e){try{return null!=e?e.toString().slice(0,30):e}catch(e){return"[ERROR] Exception while trying to serialize the value"}}function kr(e,t){var n=Tr(e),r=Tr(t);return n&&r?function(e,t,n){var r=e[G()](),i=t[G()]();for(;;){var o=r.next(),a=i.next();if(o.done&&a.done)return!0;if(o.done||a.done)return!1;if(!n(o.value,a.value))return!1}}(e,t,kr):!(n||!(e&&("object"==typeof e||"function"==typeof e))||r||!(t&&("object"==typeof t||"function"==typeof t)))||Y(e,t)}var Mr=function(){function e(e){this.wrapped=e}return e.wrap=function(t){return new e(t)},e.unwrap=function(t){return e.isWrapped(t)?t.wrapped:t},e.isWrapped=function(t){return t instanceof e},e}(),Ar=function(){function e(e,t,n){this.previousValue=e,this.currentValue=t,this.firstChange=n}return e.prototype.isFirstChange=function(){return this.firstChange},e}();function Tr(e){return!!Dr(e)&&(Array.isArray(e)||!(e instanceof Map)&&G()in e)}function Dr(e){return null!==e&&("function"==typeof e||"object"==typeof e)}function Ir(e){throw new Error("Multiple components match node with tagname "+e.tagName)}var Pr={};function Rr(e,t,n){return e[t]=n}function Nr(e,t){return ngDevMode&&Ae(e,e[t]),ngDevMode&&je(e[t],Pr,"Stored value should never be NO_CHANGE."),e[t]}function Fr(e,t,n){if(ngDevMode&&je(n,Pr,"Incoming value should never be NO_CHANGE."),ngDevMode&&xe(t,e.length,"Slot should have been initialized to NO_CHANGE"),e[t]===Pr)e[t]=n;else{if(r=e[t],i=n,r!=r&&i!=i||r===i)return!1;ngDevMode&&En()&&(kr(e[t],n)||function(e,t,n){var r="ExpressionChangedAfterItHasBeenCheckedError: Expression has changed after it was checked. Previous value: '"+t+"'. Current value: '"+n+"'.";throw e&&(r+=" It seems like the view has been created after its parent and its children have been dirty checked. Has it been created in a change detection hook ?"),new Error(r)}(On(e),e[t],n)),e[t]=n}var r,i;return!0}function Lr(e,t,n,r){var i=Fr(e,t,n);return Fr(e,t+1,r)||i}function Vr(e,t,n,r,i){var o=Lr(e,t,n,r);return Fr(e,t+2,i)||o}function Br(e,t,n,r,i,o){var a=Lr(e,t,n,r);return Lr(e,t+2,i,o)||a}var zr,Ur,Hr="ngProjectAs";function qr(e){return!!e.listen}(Ur=zr||(zr={}))[Ur.Important=1]="Important",Ur[Ur.DashCase=2]="DashCase";var Gr={createRenderer:function(e,t){return document}};function Wr(e,t){return null==e.parent?(r=(n=t)[Fe])&&2!==r.type?pt(r,n[Ie]):null:pt(Yr(e),t);var n,r}function Yr(e){for(var t=e.parent;t&&5===t.type;)t=t.parent;return t}function Kr(e,t){if(-1===e.index){var n=t[We];return n>-1?t[Ie][n]:null}return t[Ie][e.parent.index]}function Qr(e,t){var n=Kr(e,t);return n?n[Je]:null}var Xr=[];function Zr(e,t,n,r,i){for(var o=e[Te].node,a=-1,s=e,u=o.child;u;){var c=null;if(3===u.type){$r(t,n,r,pt(u,s),i);var l=s[u.index];_t(l)&&$r(t,n,r,l[$e],i)}else if(0===u.type){var d=s[u.index];$r(t,n,r,d[$e],i),r&&(d[Je]=r),d[Ze].length&&(c=(s=d[Ze][0])[Te].node,i=d[$e])}else if(1===u.type){var p=Dt(s),f=p[Fe].projection[u.projection];Xr[++a]=u,Xr[++a]=s,f&&(c=(s=p[Ie])[Te].data[f.index])}else c=u.child;if(null===c)for(null===u.next&&2&u.flags&&(s=Xr[a--],u=Xr[a--]),c=u.next;!c;){if(null===(u=u.parent||s[Te].node)||u===o)return null;0===u.type&&(i=(s=s[Ie])[u.index][$e]),c=2===u.type&&s[Pe]?(s=s[Pe])[Te].node:u.next}u=c}}function $r(e,t,n,r,i){0===e?qr(t)?t.insertBefore(n,r,i):n.insertBefore(r,i,!0):1===e?qr(t)?t.removeChild(n,r):n.removeChild(r):2===e&&(ngDevMode&&ngDevMode.rendererDestroyNode++,t.destroyNode(r))}function Jr(e,t){return qr(t)?t.createText(ut(e)):t.createTextNode(ut(e))}function ei(e,t,n){var r=Qr(e[Te].node,e);if(ngDevMode&&Xt(e[Te].node,2),r){var i=e[He];Zr(e,t?0:1,i,r,n)}}function ti(e,t,n,r,i){var o=t[Ze];r>0&&(o[r-1][Pe]=e),r<o.length?(e[Pe]=o[r],o.splice(r,0,e)):(o.push(e),e[Pe]=null),i>-1&&(e[We]=i,e[Ie]=n),e[Re]&&e[Re].insertView(r),e[De]|=16}function ni(e,t,n){var r=e[Ze],i=r[t];return t>0&&(r[t-1][Pe]=i[Pe]),r.splice(t,1),n||ei(i,!1),i[Re]&&i[Re].removeView(),i[We]=-1,i[Ie]=null,i[De]&=-17,i}function ri(e,t,n){var r=e[Ze][n];ni(e,n,!!t.detached),oi(r)}function ii(e){var t=e[Te].childIndex;return-1===t?null:e[t]}function oi(e){var t=e[He];qr(t)&&t.destroyNode&&Zr(e,2,t,null),function(e){if(-1===e[Te].childIndex)return si(e);for(var t=ii(e);t;){var n=null;if(t.length>=Qe){var r=t;r[Te].childIndex>-1&&(n=ii(r))}else{var i=t;i[Ze].length&&(n=i[Ze][0])}if(null==n){for(;t&&!t[Pe]&&t!==e;)si(t),t=ai(t,e);si(t||e),n=t&&t[Pe]}t=n}}(e),e[De]|=64}function ai(e,t){var n;return e.length>=Qe&&(n=e[Fe])&&2===n.type?Kr(n,e):e[Ie]===t?null:e[Ie]}function si(e){if(e.length>=Qe){var t=e;null!=(s=(o=t)[Te])&&null!=(a=s.destroyHooks)&&sn(o,a),(i=(r=t)[Te]&&r[Te].pipeDestroyHooks)&&sn(r,i),function(e){var t=e[Te].cleanup;if(null!=t){for(var n=e[Ve],r=0;r<t.length-1;r+=2)if("string"==typeof t[r]){var i=t[r+1],o=n[t[r+2]],a=lt(e[i]),s=t[r+3];"boolean"==typeof s?a.removeEventListener(t[r],o,s):s>=0?n[s]():n[-s].unsubscribe(),r+=2}else if("number"==typeof t[r]){var u=n[t[r]];u()}else{var c=n[t[r+1]];t[r].call(c)}e[Ve]=null}}(t);var n=t[Fe];n&&3===n.type&&qr(t[He])&&(ngDevMode&&ngDevMode.rendererDestroy++,t[He].destroy())}var r,i,o,a,s}function ui(e,t){if(ci(e,t)){if(yt(t))return di(t[He],pt(e,t));var n=t[Fe],r=e.parent;return null!=r&&4===r.type&&(e=fi(r)),null==e.parent&&2===n.type?Qr(n,t):Wr(e,t)}return null}function ci(e,t){var n,r,i=e,o=e.parent;return e.parent&&(4===e.parent.type?o=(i=fi(e)).parent:5===e.parent.type&&(o=(i=Yr(i)).parent)),null===o&&(o=t[Fe]),o&&2===o.type?null!=(r=Kr(o,t))&&null!=r[Je]:null==(n=i).parent||!(3!==n.parent.type||1&n.parent.flags)}function li(e,t,n,r){qr(e)?e.insertBefore(t,n,r):t.insertBefore(n,r,!0)}function di(e,t){return qr(e)?e.parentNode(t):t.parentNode}function pi(e,t,n){if(void 0===e&&(e=null),null!==e&&ci(t,n)){var r=n[He],i=Wr(t,n),o=t.parent||n[Fe];if(2===o.type){var a=Kr(o,n),s=a[Ze],u=s.indexOf(n);li(r,a[Je],e,hi(u,s,a[$e]))}else if(4===o.type){li(r,ui(t,n),e,i)}else if(5===o.type){li(r,i,e,pt(t.parent,n))}else qr(r)?r.appendChild(i,e):i.appendChild(e);return!0}return!1}function fi(e){for(;null!=e.parent&&4===e.parent.type;)e=e.parent;return e}function hi(e,t,n){if(e+1<t.length){var r=t[e+1],i=r[Fe];return i.child?pt(i.child,r):n}return n}function mi(e,t,n){if(null!==t&&ci(e,n)){var r=Wr(e,n),i=n[He];return qr(i)?i.removeChild(r,t):r.removeChild(t),!0}return!1}function gi(e,t,n,r){var i=pt(e,r);pi(i,t,n),ar(i,r);var o=ui(t,n),a=r[e.index];if(0===e.type){a[Je]=o;for(var s=a[Ze],u=0;u<s.length;u++)ei(s[u],!0,a[$e])}else{if(4===e.type)for(var c=e.child;c;)gi(c,t,n,r),c=c.next;_t(a)&&(a[Je]=o,pi(a[$e],t,n))}}var vi="ng-template";function _i(e,t,n){ngDevMode&&we(t[0],"Selector should have a tag name");for(var r,i,o,a,s,u,c,l,d=4,p=e.attrs,f=p?p.indexOf(3):-1,h=!1,m=0;m<t.length;m++){var g=t[m];if("number"!=typeof g){if(!h)if(4&d){if(d=2|1&d,""!==g&&(l=n,(c=g)!==(u=e).tagName||!(l||3===u.type||4===u.type||0===u.type&&c===vi))||""===g&&1===t.length){if(yi(d))return!1;h=!0}}else{var v=bi(8&d?"class":g,p);if(-1===v){if(yi(d))return!1;h=!0;continue}var _=8&d?g:t[++m];if(""!==_){var y=void 0,b=p[v];if(f>-1&&v>f?y="":(ngDevMode&&je(b,0,"We do not match directives on namespaced attributes"),y=p[v+1]),8&d&&(i=_,void 0,void 0,void 0,o=(r=y).length,a=r.indexOf(i),s=a+i.length,-1===a||a>0&&" "!==r[a-1]||s<o&&" "!==r[s])||2&d&&_!==y){if(yi(d))return!1;h=!0}}}}else{if(!h&&!yi(d)&&!yi(g))return!1;if(h&&yi(g))continue;h=!1,d=g|1&d}}return yi(d)||h}function yi(e){return 0==(1&e)}function bi(e,t){if(null===t)return-1;for(var n=!1,r=0;r<t.length;){var i=t[r];if(i===e)return r;0===i?r+=4:(3===i&&(n=!0),r+=n?1:2)}return-1}function ji(e,t,n){void 0===n&&(n=!1);for(var r=0;r<t.length;r++)if(_i(e,t[r],n))return!0;return!1}function xi(e,t,n){for(var r=function(e){var t=e.attrs;if(null!=t){var n=t.indexOf(Hr);if(0==(1&n))return t[n+1]}return null}(e),i=0;i<t.length;i++)if(r===n[i]||null===r&&ji(e,t[i],!0))return i+1;return 0}function Oi(e,t){return new wi(e,t)}var wi=function(){return function(e,t){this.fn=e,this.value=t}}(),Ci=function(){function e(){this._players=[]}return e.prototype.flushPlayers=function(){for(var e=0;e<this._players.length;e++){var t=this._players[e];t.parent||0!==t.state||t.play()}this._players.length=0},e.prototype.queuePlayer=function(e){this._players.push(e)},e}(),Ei="@";function Si(e,t,n,r){return[0,[null,-1,!1,t||null],n||[null],r||[null],[0,0],e||null,null,null,null]}function ki(e,t){for(var n=e,r=t[n],i=t;Array.isArray(r);)i=r,r=r[Ne];if(s=i,Array.isArray(s)&&"number"==typeof s[0]&&Array.isArray(s[2]))return i;var o,a,s,u=ft(e-Qe,t).stylingTemplate;return i!==t&&(n=Ne),i[n]=u?(o=r,(a=u.slice())[5]=o,a[0]|=32,a):Si(r)}function Mi(e){return e[0]===Ei}function Ai(e,t,n,r,i,o){return o=o||n,i?e[i]=r:e.push(r),!!r&&(r.addEventListener(200,function(){var t=e.indexOf(r),n=e[0];t&&(t<n?e[t]=null:e.splice(t,1)),r.destroy()}),(t.playerHandler||(t.playerHandler=new Ci)).queuePlayer(r,o),!0)}function Ti(e){return e[8]}function Di(e){return e[8]=[5,null,null,null,null]}function Ii(){throw new Error("Only elements that exist in an Angular application can be used for animations")}function Pi(e){return!!(8&e.flags)}function Ri(e){for(var t=Si(),n=t[3]=[null],r=t[2]=[null],i=-1,o=0;o<e.length;o++){var a=e[o];if("number"==typeof a)i=a;else if(2===i)r.push(a,e[++o]);else if(1===i)n.push(a,!0);else if(3===i)break}return t}function Ni(e,t,n){for(var r=1;r<e.length;){if(e[r]===t){var i=e[r+1];return void(null!=i&&0!=i||(e[r+1]=n))}r+=2}e.push(t,n)}function Fi(e,t,n,r){for(var i=1;i<n.length;i+=2){var o=n[i+1];o&&(r?Hi(e,n[i+0],!0,t,null):Ui(e,n[i+0],o,t,null))}}function Li(e,t,n,r,i,o){if(!(32&e[0])){var a=function(e,t,n){var r,i=e[1],o=e[4].length,a=xo(i,t);if(-1===a)r=i.length/4,i.push(t,o,!1,n||null);else{var s=a+1;if(i[s]>=0)return-1;r=a/4;var u=a+1;i[u]=o;var c=a+3;i[c]=n||null}return r}(e,t,i);if(-1!==a){var s=e[4],u=s[1],c=s[0],l=4*c,d=9+l,p=d+4*u,f=p+l,h=s.length;s.push(r?r.length:0,n?n.length:0);var m=0,g=[];if(r&&r.length)for(var v=0;v<r.length;v++){var _=r[v];-1==(j=Vi(e,_,9,d))&&(j=d+m,m+=4,g.push(_)),s.push(j)}var y=[];if(n&&n.length)for(var b=0;b<n.length;b++){var j,x=n[b];-1==(j=Vi(e,x,d,p))?(j=p+m,m+=4,y.push(x)):j+=4*g.length,s.push(j)}var O=2;if(g.length)for(;O<h;){var w=s[O+0],C=s[O+1];if(C)for(var E=O+2+w,S=E;S<E+C;S++)s[S]+=4*g.length;O+=2+(w+C)}for(var k=y.length+g.length,M=9;M<e.length;M+=4){var A=M>=p,T=M>=(A?f:d),D=ao(e,M),I=Qi(D),P=Xi(D);oo(e,M,Yi(D,I,P+=A?T?4*g.length:0:4*k+4*(T?g.length:0)))}for(var R=0;R<4*g.length;R++)e.splice(f,0,null),e.splice(d,0,null),d++,p++,f+=2;for(var N=0;N<4*y.length;N++)e.splice(p,0,null),e.push(null),p++,f++;for(var F=e[3],L=e[2],V=0;V<k;V++){var B=V>=g.length,z=B?V-g.length:V,U=B?y[z]:g[z],H=void 0,q=void 0;B?(H=f+4*(u+z),q=d+4*(u+z)):(H=p+4*(c+z),q=9+4*(c+z));var G=B?F:L,W=Oo(G,U);-1===W?(W=G.length+1,G.push(U,!B&&null)):W+=1;var Y=vo(e,U,B,i||null);oo(e,q,Yi(Y,W,H)),$i(e,q,U),Ji(e,q,null),no(e,q,0,a),oo(e,H,Yi(Y,W,q)),$i(e,H,U),Ji(e,H,null),no(e,H,0,a)}s[1]=u+y.length,s[0]=c+g.length,oo(e,0,Yi(0,0,p)|(o?16:0))}}}function Vi(e,t,n,r){for(var i=n;i<r;i+=4)if(uo(e,i)===t)return i;return-1}function Bi(e,t,n,r,i){var o,a,s,u,c,l,d,p,f,h=jo(e,i||null),m=(a=h,s=t,u=r,c=(o=e)[1][4*a+1],l=o[4],d=c+2+(u?l[c+0]:0)+s,l[d]),g=so(e,m),v=ao(e,m),_=bo(e,m),y=n instanceof wi?n.value:n;if(_o(v,g,y)&&function(e,t,n,r){if(e)return t?r<=n:n===r;return!0}(g,y,_,h)){var b=2==(2&v),j=e[5],x=n instanceof wi?new yo(n,j,b?1:2):null,O=x?n.value:n,w=ro(e,m),C=!1,E=x?w:0;if(eo(e,x,w)){var S=to(e,x,w);E=x?S:0,C=!0}if((C||_!==h)&&no(e,m,E,h),_!==h){var k=uo(e,m),M=wo(e,h);p=e,f=m,!(!M||!M(k))?p[f]|=4:p[f]&=-5}Ji(e,m,O);var A=Xi(v),T=so(e,A);if(!T||_o(v,T,O)){var D=!1,I=!0;!go(O,b)&&go(T,b)&&(D=!0,I=!1),qi(e,A,D),qi(e,m,I),Co(e,h,!0),lo(e,!0)}C&&po(e,!0)}}function zi(e,t,n,r,i,o,a){var s,u=0,c=jo(e,a||null);if(Gi(e,0)&&(s=c,e[1][4*s+2])){for(var l=8&e[0],d=e[5],p=Zi(e),f=co(e),h=!1,m=9;m<e.length;m+=4)if(Gi(e,m)){var g=ao(e,m),v=bo(e,m);if(c!==v){h=!0;continue}var _=uo(e,m),y=so(e,m),b=4&g?wo(e,v):null,j=io(e,m),x=!!(2&g),O=!x||!f,w=y;if(m<p&&!go(w,x))w=so(e,Xi(g));x||go(w,x)||!O||(w=Ki(e,g)),(!r||w)&&(x?Hi(d,_,!!w,t,i,j):Ui(d,_,w,t,b,o,j)),qi(e,m,!1)}if(l){var C=Array.isArray(n)?jt(n):n,E=Ti(e),S=E[0];for(m=1;m<S;m+=2){var k=E[m],M=m+1,A=E[M];if(k){var T=k.buildPlayer(A,r);if(void 0!==T){if(null!=T)Ai(E,C,d,T,M)&&u++;A&&A.destroy()}}else A&&A.destroy()}po(e,!1)}Co(e,c,!1),lo(e,h)}return u}function Ui(e,t,n,r,i,o,a){n=i&&n?i(t,n):n,o||a?(o&&o.setValue(t,n),a&&a.setValue(t,n)):n?(n=n.toString(),ngDevMode&&ngDevMode.rendererSetStyle++,qr(r)?r.setStyle(e,t,n,zr.DashCase):e.style.setProperty(t,n)):(ngDevMode&&ngDevMode.rendererRemoveStyle++,qr(r)?r.removeStyle(e,t,zr.DashCase):e.style.removeProperty(t))}function Hi(e,t,n,r,i,o){i||o?(i&&i.setValue(t,n),o&&o.setValue(t,n)):n?(ngDevMode&&ngDevMode.rendererAddClass++,qr(r)?r.addClass(e,t):e.classList.add(t)):(ngDevMode&&ngDevMode.rendererRemoveClass++,qr(r)?r.removeClass(e,t):e.classList.remove(t))}function qi(e,t,n){var r=t>=9?t+0:t;n?e[r]|=1:e[r]&=-2}function Gi(e,t){return 1==(1&e[t>=9?t+0:t])}function Wi(e,t){return 2==(2&e[t>=9?t+0:t])}function Yi(e,t,n){return 63&e|t<<6|n<<20}function Ki(e,t){var n=Qi(t);return(2&t?e[3]:e[2])[n]}function Qi(e){return e>>6&16383}function Xi(e){var t=e>>20&16383;return t>=9?t:-1}function Zi(e){return Xi(e[0])}function $i(e,t,n){e[t+1]=n}function Ji(e,t,n){e[t+2]=n}function eo(e,t,n){var r=e[8];if(t){if(!r||0===n)return!0}else if(!r)return!1;return r[n]!==t}function to(e,t,n){var r=e[8]||Di(e);return n>0?r[n]=t:(n=r[0],r.splice(n,0,t,null),r[0]+=2),n}function no(e,t,n,r){var i=n<<16|r;e[t+3]=i}function ro(e,t){return e[t+3]>>16&65535}function io(e,t){var n=ro(e,t);if(n){var r=e[8];if(r)return r[n]}return null}function oo(e,t,n){e[0===t?t:t+0]=n}function ao(e,t){return e[0===t?t:t+0]}function so(e,t){return e[t+2]}function uo(e,t){return e[t+1]}function co(e){return 16&e[0]}function lo(e,t){qi(e,0,t)}function po(e,t){t?e[0]|=8:e[0]&=-9}function fo(e,t,n){for(var r=(n||0)+1;r<e.length;r+=4){if(e[r]==t)return r-1}return-1}function ho(e,t,n){var r=so(e,t),i=uo(e,t),o=ao(e,t),a=ro(e,t),s=o,u=ao(e,n),c=Xi(s);c>=0&&oo(e,c,Yi(l=ao(e,c),Qi(l),n));var l,d=Xi(u);d>=0&&oo(e,d,Yi(l=ao(e,d),Qi(l),t));Ji(e,t,so(e,n)),$i(e,t,uo(e,n)),oo(e,t,ao(e,n));no(e,t,ro(e,n),0),Ji(e,n,r),$i(e,n,i),oo(e,n,o),no(e,n,a,0)}function mo(e,t,n,r,i,o,a,s){var u=t<e.length;e.splice(t,0,1|i|(n?2:0),r,o,0),no(e,t,s,a),u&&function(e,t){for(var n=t;n<e.length;n+=4){var r=Xi(ao(e,n));if(r>0){var i=Qi(ao(e,r));oo(e,r,Yi((Gi(e,r)?1:0)|(Wi(e,r)?2:0)|(4==(4&e[(o=r)>=9?o+0:o])?4:0),i,n))}}var o}(e,t+4)}function go(e,t){return t?!!e:null!==e}function vo(e,t,n,r){var i,o=r&&r(t)?4:0;return n?(o|=2,i=Oo(e[3],t)):i=Oo(e[2],t),Yi(o,i=i>0?i+1:0,0)}function _o(e,t,n){return!(2&e)&&(t&&n)&&4&e?t.toString()!==n.toString():t!==n}var yo=function(){function e(e,t,n){this._element=t,this._type=n,this._values={},this._dirty=!1,this._factory=e}return e.prototype.setValue=function(e,t){this._values[e]!==t&&(this._values[e]=t,this._dirty=!0)},e.prototype.buildPlayer=function(e,t){if(this._dirty){var n=this._factory.fn(this._element,this._type,this._values,t,e||null);return this._values={},this._dirty=!1,n}},e}();function bo(e,t){return 65535&e[t+3]}function jo(e,t){var n=xo(e[1],t);return ngDevMode&&je(n,-1,"The provided directive "+t+" has not been allocated to the element's style/class bindings"),n>0?n/4:0}function xo(e,t){for(var n=0;n<e.length;n+=4)if(e[n]===t)return n;return-1}function Oo(e,t){for(var n=1;n<e.length;n+=2)if(e[n]===t)return n;return-1}function wo(e,t){var n=e[1];return n[4*t+3]||n[3]||null}function Co(e,t,n){e[1][4*t+2]=n}function Eo(e){var t=e[6];if(null==t){t="";for(var n=e[3],r=1;r<n.length;r+=2){n[r+1]&&(t+=(t.length?" ":"")+n[r])}e[6]=t}return t}var So=Promise.resolve(null);function ko(e){var t=e[Te];if(t.firstTemplatePass=!1,An(!1),!On(e)){var n=En();on(e,t,n),function(e){for(var t=ii(e);null!==t;t=t[Pe])if(t.length<Qe&&-1===t[Xe])for(var n=t,r=0;r<n[Ze].length;r++){var i=n[Ze][r];ngDevMode&&we(i[Te],"TView must be allocated"),Io(i,i[Te],i[Be])}}(e),function(e){if(null!=e.contentQueries)for(var t=0;t<e.contentQueries.length;t+=2){var n=e.contentQueries[t],r=e.data[n];r.contentQueriesRefresh(n-Qe,e.contentQueries[t+1])}}(t),an(e,t.contentHooks,t.contentCheckHooks,n),function(e,t){if(e.expandoInstructions){var n=t[Le]=e.expandoStartIndex;In(n);for(var r=-1,i=-1,o=0;o<e.expandoInstructions.length;o++){var a=e.expandoInstructions[o];if("number"==typeof a){if(a<=0){i=-a;var s=e.expandoInstructions[++o];r=n+=it+s}else n+=a;In(n)}else null!==a&&(t[Le]=n,a(2,lt(t[r]),i)),r++}}}(t,e)}!function(e){if(null!=e)for(var t=0;t<e.length;t++)Da(e[t])}(t.components)}function Mo(e,t,n,r,i,o,a,s){var u=t.blueprint.slice();return u[De]=51|r,u[Ie]=u[Ke]=e,u[Be]=n,u[Ue]=i||e&&e[Ue],ngDevMode&&we(u[Ue],"RendererFactory is required"),u[He]=o||e&&e[He],ngDevMode&&we(u[He],"Renderer is required"),u[qe]=a||e&&e[qe]||null,u[ze]=s||e&&e[ze]||null,u}function Ao(e,t,n,r,i){var o=gn(),a=o[Te],s=e+Qe;ngDevMode&&xe(s,o.length,"Slot should have been initialized with null"),o[s]=n;var u=a.data[s];null==u&&(u=a.data[s]=oa(o,t,s,r,i,null));var c=_n(),l=jn();return c&&(!l||null!=c.child||null===u.parent&&2!==c.type?l||(c.next=u):c.child=u),null==a.firstChild&&(a.firstChild=u),yn(u),xn(!0),u}function To(e,t){return null==t[Te].node&&(t[Te].node=oa(t,2,e,null,null,null)),t[Fe]=t[Te].node}function Do(e){var t=e[Te];t.firstTemplatePass&&(t.expandoStartIndex++,t.blueprint.push(null),t.data.push(null),e.push(null))}function Io(e,t,n){var r,i=jn(),o=_n();if(xn(!0),yn(null),128&e[De])Va(jt(e));else try{xn(!0),yn(null),r=Pn(e,e[Fe]),Bo(),t.template(No(e),n),e[Te].firstTemplatePass=!1,An(!1),ko(e)}finally{Nn(r),xn(i),yn(o)}}function Po(e){return void 0===e&&(e=1),Rn(e)}function Ro(e,t,n){var r=e[Ue],i=Pn(e,e[Fe]),o=!En();try{o&&r.begin&&r.begin(),On(e)&&(n&&(Bo(),n(1,t)),ko(e),e[De]&=-2),n&&n(2,t),ko(e)}finally{o&&r.end&&r.end(),Nn(i)}}function No(e){return On(e)?1:2}var Fo=null;function Lo(){Fo="http://www.w3.org/2000/svg"}function Vo(){Fo="http://www.w3.org/1998/MathML/"}function Bo(){Fo=null}function zo(e,t,n,r){qo(e,t,n,r),Jo()}function Uo(e,t,n){var r=gn(),i=r[Te],o=r[He];ngDevMode&&be(r[Le],i.bindingStartIndex,"element containers should be created before any bindings"),ngDevMode&&ngDevMode.rendererCreateComment++;var a=o.createComment(ngDevMode?"ng-container":"");ngDevMode&&Ae(r,e-1),pi(a,Ao(e,4,a,"ng-container",t||null),r),Wo(i,r,n),ar(a,r)}function Ho(){var e=_n(),t=gn(),n=t[Te];jn()?xn(!1):(ngDevMode&&Me(_n()),yn(e=e.parent)),ngDevMode&&Xt(e,4);var r=t[Re];r&&(t[Re]=r.addNode(e)),en(n,e)}function qo(e,t,n,r){var i=gn(),o=i[Te];ngDevMode&&be(i[Le],o.bindingStartIndex,"elements should be created before any bindings "),ngDevMode&&ngDevMode.rendererCreateElement++;var a=Go(t);ngDevMode&&Ae(i,e-1);var s,u,c,l=Ao(e,3,a,t,n||null);if(n&&(o.firstTemplatePass&&!l.stylingTemplate&&function(e){for(var t=0;t<e.length;t++){var n=e[t];if(1==n||2==n)return!0}return!1}(n)&&(l.stylingTemplate=Ri(n)),Qo(a,n)),pi(a,l,i),Wo(o,i,r),0===Ut&&ar(a,i),Ut++,o.firstTemplatePass){var d=ls(l);d&&d.hasOwnProperty("class")&&(l.flags|=8)}l.stylingTemplate&&0==(8&l.flags)&&(s=a,u=l.stylingTemplate,c=i[He],Fi(s,c,u[3],!0),Fi(s,c,u[2],!1))}function Go(e,t){var n=t||gn()[He];return qr(n)?n.createElement(e,Fo):null===Fo?n.createElement(e):n.createElementNS(Fo,e)}function Wo(e,t,n,r){if(void 0===r&&(r=pt),un){var i=_n();Mn()&&(ngDevMode&&ngDevMode.firstTemplatePass++,function(e,t,n,r,i){ngDevMode&&be(Mn(),!0,"should run on first template pass only");var o=i?{"":-1}:null;if(n){ja(r,e.data.length,n.length);for(var a=0;a<n.length;a++){var s=n[a];s.providersResolver&&s.providersResolver(s)}ga(e,r,n.length);for(var a=0;a<n.length;a++){var s=n[a],u=e.data.length;xa(e,t,s,s.factory),ba(e.data.length-1,s,o),Jt(u,s.onInit,s.doCheck,e)}}o&&function(e,t,n){if(t)for(var r=e.localNames=[],i=0;i<t.length;i+=2){var o=n[t[i+1]];if(null==o)throw new Error("Export of name '"+t[i+1]+"' not found!");r.push(t[i],o)}}(r,i,o)}(e,t,function(e,t,n){ngDevMode&&be(Mn(),!0,"should run on first template pass only");var r=e.directiveRegistry,i=null;if(r)for(var o=0;o<r.length;o++){var a=r[o];ji(n,a.selectors,!1)&&(i||(i=[]),Gn(zn(_n(),t),t,a.type),vt(a)?(1&n.flags&&Ir(n),n.flags=1,i.unshift(a)):i.push(a))}return i}(e,t,i),i,n||null)),function(e,t,n){var r=n.directiveStart,i=n.directiveEnd;!Mn()&&r<i&&zn(n,t);for(var o=r;o<i;o++){var a=e.data[o];vt(a)&&Oa(t,n,a);var s=Zn(e.data,t,o,n);va(t,s,a,o)}}(e,t,i),function(e,t,n){for(var r=n.directiveStart,i=n.directiveEnd,o=e.expandoInstructions,a=Mn(),s=r;s<i;s++){var u=e.data[s],c=t[s];if(u.hostBindings){var l=o.length;fn(u),u.hostBindings(1,c,n.index-Qe),fn(null),l===o.length&&a&&o.push(u.hostBindings)}else a&&o.push(null)}}(e,t,i),function(e,t,n){var r=t.localNames;if(r)for(var i=t.index+1,o=0;o<r.length;o+=2){var a=r[o+1],s=-1===a?n(t,e):e[a];e[i++]=s}}(t,i,r)}}function Yo(e,t,n,r,i,o){return e.ngPrivateData||(e.ngPrivateData=Ko(-1,e,t,n,r,i,o))}function Ko(e,t,n,r,i,o,a){ngDevMode&&ngDevMode.tView++;var s,u,c=Qe+n,l=c+r,d=(s=c,(u=new Array(l).fill(null,0,s).fill(Pr,s))[We]=-1,u[Le]=s,u);return d[Te]={id:e,blueprint:d,template:t,viewQuery:a,node:null,data:d.slice(),childIndex:-1,bindingStartIndex:c,expandoStartIndex:l,expandoInstructions:null,firstTemplatePass:!0,initHooks:null,checkHooks:null,contentHooks:null,contentCheckHooks:null,viewHooks:null,viewCheckHooks:null,destroyHooks:null,pipeDestroyHooks:null,cleanup:null,contentQueries:null,components:null,directiveRegistry:"function"==typeof i?i():i,pipeRegistry:"function"==typeof o?o():o,firstChild:null}}function Qo(e,t){for(var n=gn()[He],r=qr(n),i=0;i<t.length;){var o=t[i++];if("number"==typeof o){if(0!==o)break;var a=t[i++],s=t[i++],u=t[i++];ngDevMode&&ngDevMode.rendererSetAttribute++,r?n.setAttribute(e,s,u,a):e.setAttributeNS(a,s,u)}else{u=t[i++];o!==Hr&&(ngDevMode&&ngDevMode.rendererSetAttribute++,Mi(o)?r&&n.setProperty(e,o,u):r?n.setAttribute(e,o,u):e.setAttribute(o,u))}}}function Xo(e,t){return new Error("Renderer: "+e+" ["+ut(t)+"]")}function Zo(e,t){var n=e.createRenderer(null,null),r="string"==typeof t?qr(n)?n.selectRootElement(t):n.querySelector(t):t;if(ngDevMode&&!r)throw Xo("string"==typeof t?"Host node with selector not found:":"Host node is required:",t);return r}function $o(e,t,n){void 0===n&&(n=!1);var r,i=gn(),o=_n(),a=i[Te],s=a.firstTemplatePass&&(a.cleanup||(a.cleanup=[]));if(ngDevMode&&Zt(o,3,0,4),3===o.type){var u=pt(o,i);ngDevMode&&ngDevMode.rendererAddEventListener++;var c=i[He],l=(v=ps(i)).length,d=n;if(qr(c)){var p=c.listen(u,e,t);v.push(t,p),d=l+1}else{var f=(r=t,function(e){!1===r(e)&&(e.preventDefault(),e.returnValue=!1)});u.addEventListener(e,f,n),v.push(f)}s&&s.push(e,o.index,l,d)}void 0===o.outputs&&(o.outputs=sa(o,1));var h,m=o.outputs;if(m&&(h=m[e])){var g=h.length;if(g)for(var v=ps(i),_=0;_<g;_+=2){ngDevMode&&Ae(i,h[_]);var y=i[h[_]][h[_+1]].subscribe(t),b=v.length;v.push(t,y),s&&s.push(e,o.index,b,-(b+1))}}}function Jo(){var e=_n();jn()?xn(!1):(ngDevMode&&Me(_n()),yn(e=e.parent)),ngDevMode&&Xt(e,3);var t=gn(),n=t[Re];if(n&&(t[Re]=n.addNode(e)),en(gn()[Te],e),Ut--,Pi(e)){var r=ki(e.index,t);aa(t,e.inputs.class,Eo(r))}}function ea(e,t,n,r){if(n!==Pr){var i=gn(),o=i[He],a=dt(e,i);if(null==n)ngDevMode&&ngDevMode.rendererRemoveAttribute++,qr(o)?o.removeAttribute(a,t):a.removeAttribute(t);else{ngDevMode&&ngDevMode.rendererSetAttribute++;var s=null==r?ut(n):r(n);qr(o)?o.setAttribute(a,t,s):a.setAttribute(t,s)}}}function ta(e,t,n,r,i){ia(e,t,n,r,i)}function na(e,t,n,r,i){ia(e,t,n,r,i,ra)}function ra(e,t){return t[e.index][He]}function ia(e,t,n,r,i,o){if(n!==Pr){var a,s,u,c=gn(),l=dt(e,c),d=ft(e,c);if(!i&&(a=ls(d))&&(s=a[t]))aa(c,s,n),gt(d)&&(4&(u=ht(e+Qe,c))[De]||(u[De]|=8)),ngDevMode&&(3!==d.type&&0!==d.type||function(e,t,n,r,i){for(var o,a=0;a<r.length;a+=2){var s=e[He],u=Cr(r[a+1]),c=Sr(i);if(3===n)qr(s)?s.setAttribute(t,u,c):t.setAttribute(u,c);else if(void 0!==i){var l="bindings="+JSON.stringify(((o={})[u]=c,o),null,2);qr(s)?s.setValue(t,l):t.textContent=l}}}(c,l,d.type,s,n));else if(3===d.type){var p=o?o(d,c):c[He];n=null!=r?r(n):n,ngDevMode&&ngDevMode.rendererSetProperty++,qr(p)?p.setProperty(l,t,n):Mi(t)||(l.setProperty?l.setProperty(t,n):l[t]=n)}}}function oa(e,t,n,r,i,o){var a=_n();ngDevMode&&ngDevMode.tNode++;var s=jn()?a:a&&a.parent,u=s&&e&&s!==e[Fe]?s:null;return{type:t,index:n,injectorIndex:u?u.injectorIndex:-1,directiveStart:-1,directiveEnd:-1,flags:0,providerIndexes:0,tagName:r,attrs:i,localNames:null,initialInputs:void 0,inputs:void 0,outputs:void 0,tViews:o,next:null,child:null,parent:u,detached:null,stylingTemplate:null,projection:null}}function aa(e,t,n){for(var r=0;r<t.length;r+=2)ngDevMode&&Ae(e,t[r]),e[t[r]][t[r+1]]=n}function sa(e,t){var n=gn()[Te],r=null,i=e.directiveStart,o=e.directiveEnd;if(o>i)for(var a=0===t,s=n.data,u=i;u<o;u++){var c=s[u],l=a?c.inputs:c.outputs;for(var d in l)if(l.hasOwnProperty(d)){r=r||{};var p=l[d];r.hasOwnProperty(d)?r[d].push(u,p):r[d]=[u,p]}}return r}function ua(e,t,n,r){var i=_n();i.stylingTemplate||(i.stylingTemplate=Si()),Li(i.stylingTemplate,r||null,e,t,n,Pi(i))}function ca(e,t){var n=_n();n.stylingTemplate||(n.stylingTemplate=Ri(t)),function(e,t,n){var r=e[1];if(-1==xo(r,n)){r.push(n,-1,!1,null);for(var i=null,o=null,a=-1,s=0;s<t.length;s++){var u=t[s];"number"==typeof u?a=u:1==a?Ni(i=i||e[3],u,!0):2==a&&Ni(o=o||e[2],u,t[++s])}}}(n.stylingTemplate,t,e)}function la(e,t){var n=gn(),r=0!=(2&n[De]);zi(ki(e+Qe,n),n[He],n,r,null,null,t)>0&&La(jt(n),2)}function da(e,t,n,r,i){var o=null;null!==n&&(o=r?ut(n)+r:n),Bi(ki(e+Qe,gn()),t,o,!1,i)}function pa(e,t,n,r){var i=n instanceof wi?n:!!n;Bi(ki(e+Qe,gn()),t,i,!0,r)}function fa(e,t,n,r){if(void 0!=r)return function(e,t,n,r){throw new Error("unimplemented. Should not be needed by ViewEngine compatibility")}();var i=gn(),o=ft(e,i),a=ki(e+Qe,i);if(Pi(o)&&t!==Pr){var s=Eo(a),u=(s.length?s+" ":"")+t;aa(i,o.inputs.class,u)}else!function(e,t,n,r){n=n||null;var i=jo(e,r||null),o=e[5],a=t instanceof wi?new yo(t,o,1):null,s=n instanceof wi?new yo(n,o,2):null,u=a?t.value:t,c=s?n.value:n,l=co(e)||u===Pr||u===e[6],d=c===Pr||c===e[7];if(!l||!d){e[6]=u,e[7]=c;var p=ae,f=!1,h=!1,m=a?1:0;eo(e,a,1)&&(to(e,a,1),h=!0);var g=s?3:0;eo(e,s,3)&&(to(e,s,3),h=!0),l||("string"==typeof u?(p=u.split(/\s+/),f=!0):p=u?Object.keys(u):ae);for(var v=u||oe,_=c?Object.keys(c):ae,y=c||oe,b=_.length,j=!1,x=Zi(e),O=0,w=_.length+p.length;x<e.length&&O<w;){if(!(F=O>=b)&&!d||F&&!l){var C=F?O-b:O,E=F?p[C]:_[C],S=F?!!f||v[E]:y[E],k=F?m:g;if((L=uo(e,x))===E){var M=so(e,x),A=ao(e,x);if(no(e,x,k,i),_o(A,M,S)){Ji(e,x,S),h=h||!!k;var T=Ki(e,A);(null!=M||_o(A,T,S))&&(qi(e,x,!0),j=!0)}}else{var D=fo(e,E,x);if(D>0){var I=so(e,D),P=ao(e,D);ho(e,x,D),_o(P,I,S)&&(T=Ki(e,P),Ji(e,x,S),(null!=I||_o(P,T,S))&&(qi(e,x,!0),h=h||!!k,j=!0))}else{var R=vo(e,E,F,wo(e,i));h=h||!!k,mo(e,x,F,E,R,S,i,k),j=!0}}}x+=4,O++}for(;x<e.length;)(!(F=2==(2&(A=ao(e,x))))&&!d||F&&!l)&&go(M=so(e,x),F)&&(qi(e,x,!0),Ji(e,x,null),no(e,x,k=F?m:g,i),j=!0),x+=4;for(var N=wo(e,i);O<w;){var F;if(!(F=O>=b)&&!d||F&&!l){C=F?O-b:O;var L=F?p[C]:_[C],V=(M=F?!!f||v[L]:y[L],A=1|vo(e,L,F,N),k=F?m:g,e.length);e.push(A,L,M,0),no(e,V,k,i),j=!0}O++}j&&(lo(e,!0),Co(e,i,!0)),h&&po(e,!0)}}(a,t,n)}function ha(e,t){var n=gn();ngDevMode&&be(n[Le],n[Te].bindingStartIndex,"text nodes should be created before any bindings"),ngDevMode&&ngDevMode.rendererCreateTextNode++;var r=Jr(t,n[He]),i=Ao(e,3,r,null,null);xn(!1),pi(r,i,n)}function ma(e,t){if(t!==Pr){var n=gn();ngDevMode&&Ae(n,e+Qe);var r=dt(e,n);ngDevMode&&we(r,"native element should exist"),ngDevMode&&ngDevMode.rendererSetText++;var i=n[He];qr(i)?i.setValue(r,ut(t)):r.textContent=ut(t)}}function ga(e,t,n){ngDevMode&&be(e.firstTemplatePass,!0,"Expando block should only be generated on first template pass.");var r=-(t.index-Qe),i=65535&t.providerIndexes,o=e.data.length-i;(e.expandoInstructions||(e.expandoInstructions=[])).push(r,o,n)}function va(e,t,n,r){var i=_n();(_a(e,i,t,n),ngDevMode&&we(i,"previousOrParentTNode"),i&&i.attrs&&function(e,t,n,r){var i=r.initialInputs;(void 0===i||e>=i.length)&&(i=function(e,t,n){var r=n.initialInputs||(n.initialInputs=[]);r[e]=null;var i=n.attrs,o=0;for(;o<i.length;){var a=i[o];if(3===a)break;if(0!==a){var s=t[a],u=i[o+1];if(void 0!==s){var c=r[e]||(r[e]=[]);c.push(s,u)}o+=2}else o+=4}return r}(e,n,r));var o=i[e];if(o)for(var a=0;a<o.length;a+=2)t[o[a]]=o[a+1]}(r,t,n.inputs,i),n.contentQueries&&n.contentQueries(r),vt(n))&&(ht(i.index,e)[Be]=t)}function _a(e,t,n,r){var i=pt(t,e);ngDevMode&&be(e[Le],e[Te].bindingStartIndex,"directives should be created before any bindings"),ngDevMode&&ke(jn()),ar(n,e),i&&ar(i,e),null!=r.attributes&&3==t.type&&Qo(i,r.attributes)}function ya(e){ngDevMode&&be(Mn(),!0,"Should only be called in first template pass.");var t=gn()[Te];(t.components||(t.components=[])).push(e.index)}function ba(e,t,n){n&&(t.exportAs&&(n[t.exportAs]=e),t.template&&(n[""]=e))}function ja(e,t,n){ngDevMode&&be(Mn(),!0,"expected firstTemplatePass to be true");var r=e.flags;ngDevMode&&be(0===r||1===r,!0,"expected node flags to not be initialized"),ngDevMode&&je(n,e.directiveEnd-e.directiveStart,"Reached the max number of directives"),e.flags=1&r,e.directiveStart=t,e.directiveEnd=t+n,e.providerIndexes=t}function xa(e,t,n,r){e.data.push(n);var i=new at(r,vt(n),null);e.blueprint.push(i),t.push(i)}function Oa(e,t,n){var r=pt(t,e),i=Yo(n.template,n.consts,n.vars,n.directiveDefs,n.pipeDefs,n.viewQuery),o=e[Ue],a=Na(e,t.index,Mo(e,i,null,n.onPush?8:4,o,e[Ue].createRenderer(r,n)));a[Fe]=t,a[Ne]=e[t.index],e[t.index]=a,Mn()&&ya(t)}function wa(e,t,n,r,i){return[i?-1:0,[],n,null,null,e,r,ui(t,n)]}function Ca(e,t,n,r,i,o,a,s){var u=gn(),c=u[Te],l=Sa(e,i||null,o||null);Mn()&&(l.tViews=Ko(-1,t,n,r,c.directiveRegistry,c.pipeRegistry,null)),Wo(c,u,a,s);var d=u[Re],p=_n();ar(pt(p,u),u),d&&(u[Re]=d.addNode(p)),en(c,l),xn(!1)}function Ea(e){var t=Sa(e,null,null);Mn()&&(t.tViews=[]),xn(!1)}function Sa(e,t,n){var r=gn();ngDevMode&&be(r[Le],r[Te].bindingStartIndex,"container nodes should be created before any bindings");var i=e+Qe,o=r[He].createComment(ngDevMode?"container":"");ngDevMode&&ngDevMode.rendererCreateComment++;var a=Ao(e,0,o,t,n),s=r[i]=wa(r[i],a,r,o);pi(o,a,r),Na(r,e+Qe,s);var u=r[Re];return u&&(s[Re]=u.container()),ngDevMode&&Xt(_n(),0),a}function ka(e){var t=gn(),n=t[Te],r=ct(n.data,e);yn(r),ngDevMode&&Xt(r,0),xn(!0),t[e+Qe][Xe]=0,on(t,n,En())}function Ma(){var e=_n();jn()?xn(!1):(ngDevMode&&Xt(e,2),ngDevMode&&Me(e),yn(e=e.parent)),ngDevMode&&Xt(e,0);for(var t=gn()[e.index],n=t[Xe];n<t[Ze].length;)ri(t,e,n)}function Aa(e,t,n){var r=gn(),i=_n(),o=2===i.type?i.parent:i,a=r[o.index];ngDevMode&&Xt(o,0);var s=function(e,t,n,r){for(var i=e[Ze],o=n;o<i.length;o++){var a=i[o][Te].id;if(a===r)return i[o];if(!(a<r))break;ri(e,t,o)}return null}(a,o,a[Xe],e);return s?(xn(!0),Pn(s,s[Te].node)):(s=Mo(r,function(e,t,n,r){var i=gn()[Te];ngDevMode&&Xt(r,0);var o=r.tViews;ngDevMode&&we(o,"TView expected"),ngDevMode&&be(Array.isArray(o),!0,"TViews should be in an array"),(e>=o.length||null==o[e])&&(o[e]=Ko(e,null,t,n,i.directiveRegistry,i.pipeRegistry,null));return o[e]}(e,t,n,o),null,4),a[Re]&&(s[Re]=a[Re].createView()),To(e,s),Pn(s,s[Te].node)),a&&(On(s)&&ti(s,a,r,a[Xe],-1),a[Xe]++),On(s)?3:2}function Ta(){var e=gn(),t=e[Fe];On(e)&&(ko(e),e[De]&=-2),ko(e),Nn(e[Ie]),yn(t),xn(!1)}function Da(e){var t=gn();ngDevMode&&Ae(t,e);var n=ht(e,t);ngDevMode&&Xt(t[Te].data[e],3),16==(16&n[De])&&12&n[De]&&(!function(e){for(var t=e[Te],n=e.length;n<t.blueprint.length;n++)e[n]=t.blueprint[n]}(n),Ha(n,n[Be]))}function Ia(e,t){var n=Dt(gn())[Fe];if(!n.projection)for(var r=e?e.length+1:1,i=n.projection=new Array(r).fill(null),o=i.slice(),a=n.child;null!==a;){var s=e?xi(a,e,t):0,u=a.next;o[s]?o[s].next=a:(i[s]=a,a.next=null),o[s]=a,a=u}}var Pa=[];function Ra(e,t,n){void 0===t&&(t=0);var r=gn(),i=Ao(e,1,null,null,n||null);null===i.projection&&(i.projection=t),xn(!1);for(var o=Dt(r),a=o[Fe].projection[t],s=o[Ie],u=-1;a;){if(1===a.type){var c=Dt(s),l=c[Fe].projection[a.projection];if(l){Pa[++u]=a,Pa[++u]=s,a=l,s=c[Ie];continue}}else a.flags|=2,gi(a,i,r,s);null===a.next&&s!==o[Ie]&&(s=Pa[u--],a=Pa[u--]),a=a.next}}function Na(e,t,n){var r=e[Te],i=Mn();return e[Ge]?e[Ge][Pe]=n:i&&(r.childIndex=t),e[Ge]=n,n}function Fa(e){for(;e&&!(128&e[De]);)e[De]|=8,e=e[Ie];e[De]|=8,ngDevMode&&we(e[Be],"rootContext should be defined"),La(e[Be],1)}function La(e,t){var n,r=0===e.flags;(e.flags|=t,r&&e.clean==So)&&(e.clean=new Promise(function(e){return n=e}),e.scheduler(function(){if(1&e.flags&&(e.flags&=-2,Va(e)),2&e.flags){e.flags&=-3;var t=e.playerHandler;t&&t.flushPlayers()}e.clean=So,n(null)}))}function Va(e){for(var t=0;t<e.components.length;t++){var n=e.components[t];Ro(Ot(n),n)}}function Ba(e){za(or(e),e)}function za(e,t){var n=e[Ue];n.begin&&n.begin(),On(e)&&Ha(e,t),Ha(e,t),n.end&&n.end()}function Ua(e){Va(e[Be])}function Ha(e,t){var n,r,i,o,a=e[Te],s=Pn(e,e[Fe]),u=a.template,c=a.viewQuery;try{Bo(),o=t,(i=c)&&On(e)&&i(1,o),u(No(e),t),ko(e),r=t,(n=c)&&!On(e)&&n(2,r)}finally{Nn(s)}}function qa(e){ngDevMode&&we(e,"component"),Fa(or(e))}function Ga(e){var t=gn();return Fr(t,t[Le]++,e)?e:Pr}function Wa(e){if(Mn()){var t=gn(),n=t[Te];!function(e,t,n){ngDevMode&&be(Mn(),!0,"Should only be called in first template pass.");var r=e.expandoInstructions,i=r.length;i>=2&&r[i-2]===t.hostBindings?r[i-1]=r[i-1]+n:r.push(t.hostBindings,n)}(n,pn,e),function(e,t,n){ngDevMode&&be(Mn(),!0,"Should only be called in first template pass.");for(var r=0;r<n;r++)t.push(Pr),e.blueprint.push(Pr),e.data.push(null)}(n,t,e)}}function Ya(e){ngDevMode&&xe(2,e.length,"should have at least 3 values"),ngDevMode&&be(e.length%2,1,"should have an odd number of values");for(var t=!1,n=gn(),r=n[Le],i=1;i<e.length;i+=2)Fr(n,r++,e[i])&&(t=!0);if(n[Le]=r,!t)return Pr;var o=e[0];for(i=1;i<e.length;i+=2)o+=ut(e[i])+e[i+1];return o}function Ka(e,t,n){var r=gn(),i=Fr(r,r[Le],t);return r[Le]+=1,i?e+ut(t)+n:Pr}function Qa(e,t,n,r,i){var o=gn(),a=Lr(o,o[Le],t,r);return o[Le]+=2,a?e+ut(t)+n+ut(r)+i:Pr}function Xa(e,t,n,r,i,o,a){var s=gn(),u=Vr(s,s[Le],t,r,o);return s[Le]+=3,u?e+ut(t)+n+ut(r)+i+ut(o)+a:Pr}function Za(e,t,n,r,i,o,a,s,u){var c=gn(),l=Br(c,c[Le],t,r,o,s);return c[Le]+=4,l?e+ut(t)+n+ut(r)+i+ut(o)+a+ut(s)+u:Pr}function $a(e,t,n,r,i,o,a,s,u,c,l){var d=gn(),p=d[Le],f=Br(d,p,t,r,o,s);return f=Fr(d,p+4,c)||f,d[Le]+=5,f?e+ut(t)+n+ut(r)+i+ut(o)+a+ut(s)+u+ut(c)+l:Pr}function Ja(e,t,n,r,i,o,a,s,u,c,l,d,p){var f=gn(),h=f[Le],m=Br(f,h,t,r,o,s);return m=Lr(f,h+4,c,d)||m,f[Le]+=6,m?e+ut(t)+n+ut(r)+i+ut(o)+a+ut(s)+u+ut(c)+l+ut(d)+p:Pr}function es(e,t,n,r,i,o,a,s,u,c,l,d,p,f,h){var m=gn(),g=m[Le],v=Br(m,g,t,r,o,s);return v=Vr(m,g+4,c,d,f)||v,m[Le]+=7,v?e+ut(t)+n+ut(r)+i+ut(o)+a+ut(s)+u+ut(c)+l+ut(d)+p+ut(f)+h:Pr}function ts(e,t,n,r,i,o,a,s,u,c,l,d,p,f,h,m,g){var v=gn(),_=v[Le],y=Br(v,_,t,r,o,s);return y=Br(v,_+4,c,d,f,m)||y,v[Le]+=8,y?e+ut(t)+n+ut(r)+i+ut(o)+a+ut(s)+u+ut(c)+l+ut(d)+p+ut(f)+h+ut(m)+g:Pr}function ns(e,t){var n=gn(),r=n[Te],i=e+Qe;i>=r.data.length&&(r.data[i]=null),n[i]=t}function rs(e){return ct(wn,e)}function is(e){var t=gn();return ngDevMode&&we(t[Ye],"Content QueryList array should be defined if reading a query."),ngDevMode&&Ae(t[Ye],e),t[Ye][e]}function os(e){return ct(gn(),e)}function as(e,t){return void 0===t&&(t=Pt.Default),e=Z(e),Yn(_n(),gn(),e,t)}function ss(e){return Wn(_n(),e)}function us(e,t){var n=gn(),r=n[Te],i=(n[Ye]||(n[Ye]=[])).push(e);if(Mn()){var o=r.contentQueries||(r.contentQueries=[]);t!==(r.contentQueries.length?r.contentQueries[r.contentQueries.length-2]:-1)&&o.push(t,i-1)}}var cs=So;function ls(e){return e?(void 0===e.inputs&&(e.inputs=sa(e,0)),e.inputs):null}function ds(){return gn()}function ps(e){return e[Ve]||(e[Ve]=[])}function fs(e){return e[Te].cleanup||(e[Te].cleanup=[])}function hs(e,t){var n=rr(e);if(n){var r=n.native,i=n.lView,o=function(e,t){if(!(t=t||rr(e)))return ngDevMode&&Ii(),null;var n=t.lView,r=ki(t.nodeIndex,n);return Ti(r)||Di(r)}(r,n),a=hr(i);Ai(o,a,r,t,0,e),La(a,2)}else ngDevMode&&Ii()}function ms(e){var t=rr(e);if(!t)return ngDevMode&&Ii(),[];var n=ki(t.nodeIndex,t.lView),r=n?Ti(n):null;return r?function(e){for(var t=[],n=e[0],r=2;r<n;r+=2){var i=e[r];i&&t.push(i)}for(r=n;r<e.length;r++)t.push(e[r]);return t}(r):[]}var gs="ng",vs=!1;function _s(){vs||(vs=!0,ys("getComponent",dr),ys("getContext",pr),ys("getListeners",Or),ys("getViewComponent",fr),ys("getHostElement",br),ys("getInjector",gr),ys("getRootComponents",mr),ys("getDirectives",vr),ys("getPlayers",ms),ys("markDirty",qa))}function ys(e,t){var n=U;if(ngDevMode&&we(t,"function not defined"),n){var r=n[gs];r||(r=n[gs]={}),r[e]=t}}function bs(e,t){void 0===t&&(t={}),ngDevMode&&_s(),ngDevMode&&Ce(e);var n=t.rendererFactory||Gr,r=t.sanitizer||null,i=ge(e);i.type!=e&&(i.type=e);var o,a=i.selectors[0][0],s=Zo(n,t.host||a),u=i.onPush?136:132,c=Os(t.scheduler,t.playerHandler),l=n.createRenderer(s,i),d=Mo(null,Ko(-1,null,1,0,null,null,null),c,u,n,l,void 0,t.injector||null),p=Pn(d,null);try{n.begin&&n.begin(),o=xs(js(s,i,d,n,l,r),i,d,c,t.hostFeatures||null),ko(d),d[De]&=-2,ko(d)}finally{Nn(p),n.end&&n.end()}return o}function js(e,t,n,r,i,o){ln=!1,cn=null,Ut=0,un=!0;var a=n[Te],s=Mo(n,Yo(t.template,t.consts,t.vars,t.directiveDefs,t.pipeDefs,t.viewQuery),null,t.onPush?8:4,r,i,o),u=Ao(0,3,e,null,null);return a.firstTemplatePass&&(Gn(zn(u,n),n,t.type),u.flags=1,ja(u,n.length,1),ya(u)),s[Ne]=n[Qe],s[Fe]=u,n[Qe]=s}function xs(e,t,n,r,i){var o=n[Te],a=function(e,t,n){var r=_n();e.firstTemplatePass&&(n.providersResolver&&n.providersResolver(n),ga(e,r,1),xa(e,t,n,n.factory));var i=Zn(e.data,t,t.length-1,r);return _a(t,r,i,n),i}(o,n,t);if(r.components.push(a),e[Be]=a,i&&i.forEach(function(e){return e(a,t)}),o.firstTemplatePass&&t.hostBindings){var s=_n();fn(t),t.hostBindings(1,a,s.index-Qe),fn(null)}return a}function Os(e,t){return{components:[],scheduler:e||At,clean:cs,playerHandler:t||null,flags:0}}function ws(e,t){var n=Ot(e)[Te],r=n.data.length-1;Jt(r,t.onInit,t.doCheck,n),en(n,{directiveStart:r,directiveEnd:r+1})}function Cs(e){return(t=e,n=bt(t)[Be],ngDevMode&&we(n,"rootContext"),n).clean;var t,n}function Es(e){return"function"==typeof e.template}function Ss(e){for(var t,n=(t=e.type,Object.getPrototypeOf(t.prototype).constructor),i=function(){var t,i,o=void 0;if(Es(e))o=n.ngComponentDef||n.ngDirectiveDef;else{if(n.ngComponentDef)throw new Error("Directives cannot inherit Components");o=n.ngDirectiveDef}var a=n.ngBaseDef;if(a||o){var u=e;u.inputs=ks(e.inputs),u.declaredInputs=ks(e.declaredInputs),u.outputs=ks(e.outputs)}if(a&&(s(e.inputs,a.inputs),s(e.declaredInputs,a.declaredInputs),s(e.outputs,a.outputs)),o){var c=e.hostBindings,l=o.hostBindings;if(l&&(e.hostBindings=c?function(e,t,n){l(e,t,n),c(e,t,n)}:l),Es(e)&&Es(o)){var d=e.viewQuery,p=o.viewQuery;p&&(e.viewQuery=d?function(e,t){p(e,t),d(e,t)}:p)}var f=e.contentQueries,h=o.contentQueries;h&&(e.contentQueries=f?function(e){h(e),f(e)}:h);var m=e.contentQueriesRefresh,g=o.contentQueriesRefresh;g&&(e.contentQueriesRefresh=m?function(e,t){g(e,t),m(e,t)}:g),s(e.inputs,o.inputs),s(e.declaredInputs,o.declaredInputs),s(e.outputs,o.outputs),e.afterContentChecked=e.afterContentChecked||o.afterContentChecked,e.afterContentInit=e.afterContentInit||o.afterContentInit,e.afterViewChecked=e.afterViewChecked||o.afterViewChecked,e.afterViewInit=e.afterViewInit||o.afterViewInit,e.doCheck=e.doCheck||o.doCheck,e.onDestroy=e.onDestroy||o.onDestroy,e.onInit=e.onInit||o.onInit;var v=o.features;if(v)try{for(var _=Object(r.__values)(v),y=_.next();!y.done;y=_.next()){var b=y.value;b&&b.ngInherit&&b(e)}}catch(e){t={error:e}}finally{try{y&&!y.done&&(i=_.return)&&i.call(_)}finally{if(t)throw t.error}}return"break"}var j=n.prototype;j&&(e.afterContentChecked=e.afterContentChecked||j.afterContentChecked,e.afterContentInit=e.afterContentInit||j.afterContentInit,e.afterViewChecked=e.afterViewChecked||j.afterViewChecked,e.afterViewInit=e.afterViewInit||j.afterViewInit,e.doCheck=e.doCheck||j.doCheck,e.onDestroy=e.onDestroy||j.onDestroy,e.onInit=e.onInit||j.onInit),n=Object.getPrototypeOf(n)};n;){if("break"===i())break}}function ks(e){return e===oe?{}:e===ae?[]:e}var Ms="__ngOnChanges_";function As(e){var t=e.declaredInputs,n=e.inputs,r=e.type.prototype,i=function(e){if(t.hasOwnProperty(e)){for(var i=n[e],o=t[e],a=Ms+i,s=void 0,u=r;!s&&u&&Object.getPrototypeOf(u)!==Object.getPrototypeOf(Object.prototype);)s=Object.getOwnPropertyDescriptor(u,i),u=Object.getPrototypeOf(u);var c=s&&s.get,l=s&&s.set;Object.defineProperty(r,i,{get:c||(l?void 0:function(){return this[a]}),set:function(e){var t=this[Ms];t||(t={},Object.defineProperty(this,Ms,{value:t,writable:!0}));var n=!this.hasOwnProperty(a),r=t[o];r?r.currentValue=e:t[o]=new Ar(this[a],e,n),n?Object.defineProperty(this,a,{value:e,writable:!0}):this[a]=e,l&&l.call(this,e)},configurable:!!ngDevMode})}};for(var o in t)i(o);null!=e.onInit&&(e.onInit=Ts(e.onInit)),e.doCheck=Ts(e.doCheck)}function Ts(e){return function(){var t=this[Ms];null!=t&&(this.ngOnChanges(t),this[Ms]=null),e&&e.apply(this)}}function Ds(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t]}As.ngInherit=!0;var Is="__source",Ps=new Object,Rs=Ps,Ns=new b("INJECTOR"),Fs=function(){function e(){}return e.prototype.get=function(e,t){if(void 0===t&&(t=Ps),t===Ps)throw new Error("NullInjectorError: No provider for "+K(e)+"!");return t},e}(),Ls=function(){function e(){}return e.create=function(e,t){return Array.isArray(e)?new Xs(e,t):new Xs(e.providers,e.parent,e.name||null)},e.THROW_IF_NOT_FOUND=Ps,e.NULL=new Fs,e.ngInjectableDef=g({providedIn:"any",factory:function(){return Yt(Ns)}}),e.__NG_ELEMENT_ID__=function(){return Bs()},e}(),Vs=function(){return e=_n(),new er(e,gn());var e},Bs=Ds,zs=function(e){return e},Us=[],Hs=zs,qs=function(){return Array.prototype.slice.call(arguments)},Gs=a({provide:String,useValue:a}),Ws="ngTempTokenPath",Ys=Ls.NULL,Ks=/\n/gm,Qs="ɵ",Xs=function(){function e(e,t,n){void 0===t&&(t=Ys),void 0===n&&(n=null),this.parent=t,this.source=n;var r=this._records=new Map;r.set(Ls,{token:Ls,fn:zs,deps:Us,value:this,useNew:!1}),r.set(Ns,{token:Ns,fn:zs,deps:Us,value:this,useNew:!1}),function e(t,n){if(n)if((n=Z(n))instanceof Array)for(var r=0;r<n.length;r++)e(t,n[r]);else{if("function"==typeof n)throw eu("Function/Class not supported",n);if(!n||"object"!=typeof n||!n.provide)throw eu("Unexpected provider",n);var i=Z(n.provide),o=function(e){var t=function(e){var t=Us,n=e.deps;if(n&&n.length){t=[];for(var r=0;r<n.length;r++){var i=6,o=Z(n[r]);if(o instanceof Array)for(var a=0,s=o;a<s.length;a++){var u=s[a];u instanceof Ft||u==Ft?i|=1:u instanceof Vt||u==Vt?i&=-3:u instanceof Lt||u==Lt?i&=-5:o=u instanceof Nt?u.token:Z(u)}t.push({token:o,options:i})}}else if(e.useExisting){var o=Z(e.useExisting);t=[{token:o,options:6}]}else if(!(n||Gs in e))throw eu("'deps' required",e);return t}(e),n=zs,r=Us,i=!1,o=Z(e.provide);if(Gs in e)r=e.useValue;else if(e.useFactory)n=e.useFactory;else if(e.useExisting);else if(e.useClass)i=!0,n=Z(e.useClass);else{if("function"!=typeof o)throw eu("StaticProvider does not have [useValue|useFactory|useExisting|useClass] or [provide] is not newable",e);i=!0,n=o}return{deps:t,fn:n,useNew:i,value:r}}(n);if(!0===n.multi){var a=t.get(i);if(a){if(a.fn!==qs)throw Zs(i)}else t.set(i,a={token:n.provide,deps:[],useNew:!1,fn:qs,value:Us});i=n,a.deps.push({token:i,options:6})}var s=t.get(i);if(s&&s.fn==qs)throw Zs(i);t.set(i,o)}}(r,e)}return e.prototype.get=function(e,t,n){void 0===n&&(n=Pt.Default);var r=this._records.get(e);try{return $s(e,r,this._records,this.parent,t,n)}catch(t){var i=t[Ws];throw e[Is]&&i.unshift(e[Is]),t.message=Js("\n"+t.message,i,this.source),t.ngTokenPath=i,t[Ws]=null,t}},e.prototype.toString=function(){var e=[];return this._records.forEach(function(t,n){return e.push(K(n))}),"StaticInjector["+e.join(", ")+"]"},e}();function Zs(e){return eu("Cannot mix multi providers and regular providers",e)}function $s(e,t,n,i,o,a){try{return function(e,t,n,i,o,a){var s,u;if(!t||a&Pt.SkipSelf)a&Pt.Self||(u=i.get(e,o,Pt.Default));else{if((u=t.value)==Hs)throw Error(Qs+"Circular dependency");if(u===Us){t.value=Hs;var c=void 0,l=t.useNew,d=t.fn,p=t.deps,f=Us;if(p.length){f=[];for(var h=0;h<p.length;h++){var m=p[h],g=m.options,v=2&g?n.get(m.token):void 0;f.push($s(m.token,v,n,v||4&g?i:Ys,1&g?null:Ls.THROW_IF_NOT_FOUND,Pt.Default))}}t.value=u=l?new((s=d).bind.apply(s,Object(r.__spread)([void 0],f))):d.apply(c,f)}}return u}(e,t,n,i,o,a)}catch(n){throw n instanceof Error||(n=new Error(n)),(n[Ws]=n[Ws]||[]).unshift(e),t&&t.value==Hs&&(t.value=Us),n}}function Js(e,t,n){void 0===n&&(n=null),e=e&&"\n"===e.charAt(0)&&e.charAt(1)==Qs?e.substr(2):e;var r=K(t);if(t instanceof Array)r=t.map(K).join(" -> ");else if("object"==typeof t){var i=[];for(var o in t)if(t.hasOwnProperty(o)){var a=t[o];i.push(o+":"+("string"==typeof a?JSON.stringify(a):K(a)))}r="{"+i.join(", ")+"}"}return"StaticInjectorError"+(n?"("+n+")":"")+"["+r+"]: "+e.replace(Ks,"\n  ")}function eu(e,t){return new Error(Js(e,t))}var tu=new b("The presence of this token marks an injector as being the root injector."),nu={},ru={},iu=[],ou=void 0;function au(){return void 0===ou&&(ou=new Fs),ou}function su(e,t,n){return void 0===t&&(t=null),void 0===n&&(n=null),t=t||au(),new uu(e,n,t)}var uu=function(){function e(e,t,n){var r=this;this.parent=n,this.records=new Map,this.injectorDefTypes=new Set,this.onDestroy=new Set,this.destroyed=!1;var i=[];pu([e],function(e){return r.processInjectorType(e,[],i)}),t&&pu(t,function(n){return r.processProvider(n,e,t)}),this.records.set(Ns,du(void 0,this)),this.isRootInjector=this.records.has(tu),this.injectorDefTypes.forEach(function(e){return r.get(e)})}return e.prototype.destroy=function(){this.assertNotDestroyed(),this.destroyed=!0;try{this.onDestroy.forEach(function(e){return e.ngOnDestroy()})}finally{this.records.clear(),this.onDestroy.clear(),this.injectorDefTypes.clear()}},e.prototype.get=function(e,t,n){void 0===t&&(t=Rs),void 0===n&&(n=Pt.Default),this.assertNotDestroyed();var r,i=qt(this);try{if(!(n&Pt.SkipSelf)){var o=this.records.get(e);if(void 0===o){var a=("function"==typeof(r=e)||"object"==typeof r&&r instanceof b)&&_(e);a&&this.injectableDefInScope(a)&&(o=du(cu(e),nu),this.records.set(e,o))}if(void 0!==o)return this.hydrate(e,o)}return(n&Pt.Self?au():this.parent).get(e,t)}finally{qt(i)}},e.prototype.assertNotDestroyed=function(){if(this.destroyed)throw new Error("Injector has already been destroyed.")},e.prototype.processInjectorType=function(e,t,n){var r=this;if(e=Z(e)){var i=y(e),o=null==i&&e.ngModule||void 0,a=void 0===o?e:o;if(ngDevMode&&-1!==t.indexOf(a)){var s=K(a);throw new Error("Circular dependency in DI detected for type "+s+". Dependency path: "+t.map(function(e){return K(e)}).join(" > ")+" > "+s+".")}var u=-1!==n.indexOf(a),c=void 0!==o&&e.providers||iu;if(void 0!==o&&(i=y(o)),null!=i){if(this.injectorDefTypes.add(a),this.records.set(a,du(i.factory,nu)),null!=i.imports&&!u){ngDevMode&&t.push(a),n.push(a);try{pu(i.imports,function(e){return r.processInjectorType(e,t,n)})}finally{ngDevMode&&t.pop()}}var l=i.providers;if(null!=l&&!u){var d=e;pu(l,function(e){return r.processProvider(e,d,l)})}var p=e.ngModule;pu(c,function(e){return r.processProvider(e,p,c)})}}},e.prototype.processProvider=function(e,t,n){var r,i,o=hu(e=Z(e))?e:Z(e&&e.provide),a=(i=lu(r=e,t,n),fu(r)?du(void 0,r.useValue):du(i,nu));if(hu(e)||!0!==e.multi){var s=this.records.get(o);if(s&&void 0!==s.multi)throw new Error("Mixed multi-provider for "+K(o))}else{var u=this.records.get(o);if(u){if(void 0===u.multi)throw new Error("Mixed multi-provider for "+o+".")}else(u=du(void 0,nu,!0)).factory=function(){return Qt(u.multi)},this.records.set(o,u);o=e,u.multi.push(e)}this.records.set(o,a)},e.prototype.hydrate=function(e,t){if(t.value===ru)throw new Error("Cannot instantiate cyclic dependency! "+K(e));var n;return t.value===nu&&(t.value=ru,t.value=t.factory()),"object"==typeof t.value&&t.value&&("object"==typeof(n=t.value)&&null!=n&&n.ngOnDestroy&&"function"==typeof n.ngOnDestroy)&&this.onDestroy.add(t.value),t.value},e.prototype.injectableDefInScope=function(e){return!!e.providedIn&&("string"==typeof e.providedIn?"any"===e.providedIn||"root"===e.providedIn&&this.isRootInjector:this.injectorDefTypes.has(e.providedIn))},e}();function cu(e){var t=_(e);if(null===t){var n=y(e);if(null!==n)return n.factory;if(e instanceof b)throw new Error("Token "+K(e)+" is missing an ngInjectableDef definition.");if(e instanceof Function){var r=e.length;if(r>0){var i=new Array(r).fill("?");throw new Error("Can't resolve all parameters for "+K(e)+": ("+i.join(", ")+").")}return function(){return new e}}throw new Error("unreachable")}return t.factory}function lu(e,t,n){var i,o,a=void 0;if(hu(e))return cu(Z(e));if(fu(e))a=function(){return Z(e.useValue)};else if(o=e,o&&o.useExisting)a=function(){return Yt(Z(e.useExisting))};else if(i=e,i&&i.useFactory)a=function(){return e.useFactory.apply(e,Object(r.__spread)(Qt(e.deps||[])))};else{var s=Z(e&&(e.useClass||e.provide));if(!s){var u="";if(t&&n)u=" - only instances of Provider and Type are allowed, got: ["+n.map(function(t){return t==e?"?"+e+"?":"..."}).join(", ")+"]";throw new Error("Invalid provider for the NgModule '"+K(t)+"'"+u)}if(!e.deps)return cu(s);a=function(){return new(s.bind.apply(s,Object(r.__spread)([void 0],Qt(e.deps))))}}return a}function du(e,t,n){return void 0===n&&(n=!1),{factory:e,value:t,multi:n?[]:void 0}}function pu(e,t){e.forEach(function(e){return Array.isArray(e)?pu(e,t):t(e)})}function fu(e){return e&&"object"==typeof e&&Gs in e}function hu(e){return"function"==typeof e}function mu(e,t,n,r,i){if(e=Z(e),Array.isArray(e))for(var o=0;o<e.length;o++)mu(e[o],t,n,r,i);else{var a=gn(),s=hu(e)?e:Z(e.provide),u=lu(e),c=_n(),l=65535&c.providerIndexes,d=c.directiveStart,p=c.providerIndexes>>16;if(hu(e)||!e.multi){var f=new at(u,i,as),h=vu(s,t,i?l:l+p,d);-1==h?(Gn(zn(c,a),a,s),t.push(s),c.directiveStart++,c.directiveEnd++,i&&(c.providerIndexes+=65536),n.push(f),a.push(f)):(n[h]=f,a[h]=f)}else{var m=vu(s,t,l+p,d),g=vu(s,t,l,l+p),v=m>=0&&n[m],_=g>=0&&n[g];if(i&&!_||!i&&!v){Gn(zn(c,a),a,s);f=(y=i?yu:_u,b=n.length,x=r,O=u,(w=new at(y,j=i,as)).multi=[],w.index=b,w.componentProviders=0,gu(w,O,x&&!j),w);!i&&_&&(n[g].providerFactory=f),t.push(s),c.directiveStart++,c.directiveEnd++,i&&(c.providerIndexes+=65536),n.push(f),a.push(f)}else gu(n[i?g:m],u,!i&&r);!i&&r&&_&&n[g].componentProviders++}}var y,b,j,x,O,w}function gu(e,t,n){e.multi.push(t),n&&e.componentProviders++}function vu(e,t,n,r){for(var i=n;i<r;i++)if(t[i]===e)return i;return-1}function _u(e,t,n,r){return bu(this.multi,[])}function yu(e,t,n,r){var i,o=this.multi;if(this.providerFactory){var a=this.providerFactory.componentProviders,s=Zn(t,n,this.providerFactory.index,r);bu(o,i=s.slice(0,a));for(var u=a;u<s.length;u++)i.push(s[u])}else bu(o,i=[]);return i}function bu(e,t){for(var n=0;n<e.length;n++){var r=e[n];t.push(r())}return t}function ju(e,t){return void 0===t&&(t=[]),function(n){n.providersResolver=function(n){return function(e,t,n){var r=gn()[Te];if(r.firstTemplatePass){var i=vt(e);mu(n,r.data,r.blueprint,i,!0),mu(t,r.data,r.blueprint,i,!1)}}(n,e,t)}}}var xu=function(){return function(){}}(),Ou=function(){return function(){}}();function wu(e){var t=Error("No component factory found for "+K(e)+". Did you add it to @NgModule.entryComponents?");return t[ku]=e,t}var Cu,Eu,Su,ku="ngComponent",Mu=function(){function e(){}return e.prototype.resolveComponentFactory=function(e){throw wu(e)},e}(),Au=function(){function e(){}return e.NULL=new Mu,e}(),Tu=function(){function e(e,t,n){this._parent=t,this._ngModule=n,this._factories=new Map;for(var r=0;r<e.length;r++){var i=e[r];this._factories.set(i.componentType,i)}}return e.prototype.resolveComponentFactory=function(e){var t=this._factories.get(e);if(!t&&this._parent&&(t=this._parent.resolveComponentFactory(e)),!t)throw wu(e);return new Du(t,this._ngModule)},e}(),Du=function(e){function t(t,n){var r=e.call(this)||this;return r.factory=t,r.ngModule=n,r.selector=t.selector,r.componentType=t.componentType,r.ngContentSelectors=t.ngContentSelectors,r.inputs=t.inputs,r.outputs=t.outputs,r}return Object(r.__extends)(t,e),t.prototype.create=function(e,t,n,r){return this.factory.create(e,t,n,r||this.ngModule)},t}(Ou),Iu=function(){return function(){}}(),Pu=function(){return function(){}}(),Ru=function(){function e(e,t,n){this._context=t,this._componentIndex=n,this._appRef=null,this._viewContainerRef=null,this._tViewNode=null,this._lView=e}return Object.defineProperty(e.prototype,"rootNodes",{get:function(){if(null==this._lView[Ne]){var e=this._lView[Fe];return function e(t,n,r){var i=n.child;for(;i;)r.push(pt(i,t)),4===i.type&&e(t,i,r),i=i.next;return r}(this._lView,e,[])}return[]},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"context",{get:function(){return this._context?this._context:this._lookUpContext()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"destroyed",{get:function(){return 64==(64&this._lView[De])},enumerable:!0,configurable:!0}),e.prototype.destroy=function(){if(this._appRef)this._appRef.detachView(this);else if(this._viewContainerRef){var e=this._viewContainerRef.indexOf(this);e>-1&&this._viewContainerRef.detach(e),this._viewContainerRef=null}oi(this._lView)},e.prototype.onDestroy=function(e){var t,n;t=this._lView,n=e,ps(t).push(n),t[Te].firstTemplatePass&&fs(t).push(t[Ve].length-1,null)},e.prototype.markForCheck=function(){Fa(this._lView)},e.prototype.detach=function(){this._lView[De]&=-17},e.prototype.reattach=function(){this._lView[De]|=16},e.prototype.detectChanges=function(){za(this._lView,this.context)},e.prototype.checkNoChanges=function(){!function(e){Sn(!0);try{Ba(e)}finally{Sn(!1)}}(this.context)},e.prototype.attachToViewContainerRef=function(e){if(this._appRef)throw new Error("This view is already attached directly to the ApplicationRef!");this._viewContainerRef=e},e.prototype.detachFromAppRef=function(){this._appRef=null},e.prototype.attachToAppRef=function(e){if(this._viewContainerRef)throw new Error("This view is already attached to a ViewContainer!");this._appRef=e},e.prototype._lookUpContext=function(){return this._context=this._lView[Ie][this._componentIndex]},e}(),Nu=function(e){function t(t){var n=e.call(this,t,null,-1)||this;return n._view=t,n}return Object(r.__extends)(t,e),t.prototype.detectChanges=function(){Ua(this._view)},t.prototype.checkNoChanges=function(){!function(e){Sn(!0);try{Ua(e)}finally{Sn(!1)}}(this._view)},Object.defineProperty(t.prototype,"context",{get:function(){return null},enumerable:!0,configurable:!0}),t}(Ru);function Fu(e,t,n){return Cu||(Cu=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t}(e)),new Cu(pt(t,n))}function Lu(e,t,n,i){if(Eu||(Eu=function(e){function t(t,n,r,i,o,a){var s=e.call(this)||this;return s._declarationParentView=t,s.elementRef=n,s._tView=r,s._renderer=i,s._queries=o,s._injectorIndex=a,s}return Object(r.__extends)(t,e),t.prototype.createEmbeddedView=function(e,t,n,r,i){var o=function(e,t,n,r,i,o){var a=jn(),s=_n();xn(!0),yn(null);var u=Mo(n,e,t,4);return u[Ke]=n,i&&(u[Re]=i.createView()),To(-1,u),e.firstTemplatePass&&(e.node.injectorIndex=o),xn(a),yn(s),u}(this._tView,e,this._declarationParentView,this._renderer,this._queries,this._injectorIndex);t&&ti(o,t,r,i,n.index),Io(o,this._tView,e);var a=new Ru(o,e,-1);return a._tViewNode=o[Fe],a},t}(e)),0===n.type){var o=i[n.index];return ngDevMode&&we(n.tViews,"TView must be allocated"),new Eu(i,Fu(t,n,i),n.tViews,gn()[He],o[Re],n.injectorIndex)}return null}function Vu(e,t,n){if(gt(e)){var r=e.directiveStart,i=ht(e.index,t);return new Ru(i,n,r)}if(3===e.type){var o=Dt(t);return new Ru(o,o[Be],-1)}return null}var Bu,zu,Uu=function(){function e(e){this.nativeElement=e}return e.__NG_ELEMENT_ID__=function(){return qu(e)},e}(),Hu=function(e){return Fu(e,_n(),gn())},qu=Ds,Gu=function(){return function(e,t,n,r,i,o){this.id=e,this.templateUrl=t,this.slotCount=n,this.encapsulation=r,this.styles=i,this.animations=o}}(),Wu=function(){return function(){}}(),Yu=function(){return function(){}}(),Ku=(new b("Renderer2Interceptor"),function(){return function(){}}()),Qu=function(){return function(){}}();(zu=Bu||(Bu={}))[zu.Important=1]="Important",zu[zu.DashCase=2]="DashCase";var Xu,Zu,$u=function(){function e(){}return e.__NG_ELEMENT_ID__=function(){return ec()},e}(),Ju=function(){return function(e){var t=e[He];if(qr(t))return t;throw new Error("Cannot inject Renderer2 when the application uses Renderer3!")}(gn())},ec=Ds;(Zu=Xu||(Xu={}))[Zu.NONE=0]="NONE",Zu[Zu.HTML=1]="HTML",Zu[Zu.STYLE=2]="STYLE",Zu[Zu.SCRIPT=3]="SCRIPT",Zu[Zu.URL=4]="URL",Zu[Zu.RESOURCE_URL=5]="RESOURCE_URL";var tc=function(){return function(){}}(),nc=function(){return function(e){this.full=e,this.major=e.split(".")[0],this.minor=e.split(".")[1],this.patch=e.split(".").slice(2).join(".")}}(),rc=new nc("7.2.0"),ic=function(e){function t(t){var n=e.call(this)||this;return n.ngModule=t,n}return Object(r.__extends)(t,e),t.prototype.resolveComponentFactory=function(e){ngDevMode&&Ce(e);var t=ge(e);return new cc(t,this.ngModule)},t}(Au);function oc(e){var t=[];for(var n in e)if(e.hasOwnProperty(n)){var r=e[n];t.push({propName:r,templateName:n})}return t}var ac=new b("ROOT_CONTEXT_TOKEN",{providedIn:"root",factory:function(){return Os(Yt(sc))}}),sc=new b("SCHEDULER_TOKEN",{providedIn:"root",factory:function(){return At}}),uc={};var cc=function(e){function t(t,n){var r=e.call(this)||this;return r.componentDef=t,r.ngModule=n,r.componentType=t.type,r.selector=t.selectors[0][0],r.ngContentSelectors=[],r}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"inputs",{get:function(){return oc(this.componentDef.inputs)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"outputs",{get:function(){return oc(this.componentDef.outputs)},enumerable:!0,configurable:!0}),t.prototype.create=function(e,t,n,r){var i,o,a=void 0===n,s=(r=r||this.ngModule)?(i=e,o=r.injector,{get:function(e,t){var n=i.get(e,uc);return n!==uc||t===uc?n:o.get(e,t)}}):e,u=s.get(Qu,Gr),c=s.get(tc,null),l=a?Go(this.selector,u.createRenderer(null,this.componentDef)):Zo(u,n),d=this.componentDef.onPush?136:132,p=a?Os():s.get(ac),f=u.createRenderer(l,this.componentDef);n&&l&&(ngDevMode&&ngDevMode.rendererSetAttribute++,qr(f)?f.setAttribute(l,"ng-version",rc.full):l.setAttribute("ng-version",rc.full));var h,m,g=Mo(null,Ko(-1,null,1,0,null,null,null),p,d,u,f,c,s),v=Pn(g,null);try{u.begin&&u.begin();var _=js(l,this.componentDef,g,u,f);if(m=ft(0,g),t)for(var y=0,b=g[Te],j=m.projection=[],x=0;x<t.length;x++){for(var O=t[x],w=null,C=null,E=0;E<O.length;E++){b.firstTemplatePass&&(b.expandoStartIndex++,b.blueprint.splice(++y+Qe,0,null),b.data.splice(y+Qe,0,null),g.splice(y+Qe,0,null));var S=Ao(y,3,O[E],null,null);C?C.next=S:w=S,C=S}j.push(w)}h=xs(_,this.componentDef,g,p,[ws]),Na(g,Qe,_),ko(g)}finally{Nn(v),u.end&&u.end()}var k=new lc(this.componentType,h,Fu(Uu,m,g),g,m);return a&&(k.hostView._tViewNode.child=m),k},t}(Ou),lc=(new ic,function(e){function t(t,n,r,i,o){var a=e.call(this)||this;return a.location=r,a._rootLView=i,a._tNode=o,a.destroyCbs=[],a.instance=n,a.hostView=a.changeDetectorRef=new Nu(i),a.hostView._tViewNode=To(-1,i),a.componentType=t,a}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"injector",{get:function(){return new er(this._tNode,this._rootLView)},enumerable:!0,configurable:!0}),t.prototype.destroy=function(){ngDevMode&&we(this.destroyCbs,"NgModule already destroyed"),this.destroyCbs.forEach(function(e){return e()}),this.destroyCbs=null,this.hostView.destroy()},t.prototype.onDestroy=function(e){ngDevMode&&we(this.destroyCbs,"NgModule already destroyed"),this.destroyCbs.push(e)},t}(xu)),dc=!0,pc=!1;function fc(){return pc=!0,dc}function hc(){if(pc)throw new Error("Cannot enable prod mode after platform setup.");dc=!1}var mc=function(){function e(e){if(this.defaultDoc=e,this.inertDocument=this.defaultDoc.implementation.createHTMLDocument("sanitization-inert"),this.inertBodyElement=this.inertDocument.body,null==this.inertBodyElement){var t=this.inertDocument.createElement("html");this.inertDocument.appendChild(t),this.inertBodyElement=this.inertDocument.createElement("body"),t.appendChild(this.inertBodyElement)}this.inertBodyElement.innerHTML='<svg><g onload="this.parentNode.remove()"></g></svg>',!this.inertBodyElement.querySelector||this.inertBodyElement.querySelector("svg")?(this.inertBodyElement.innerHTML='<svg><p><style><img src="</style><img src=x onerror=alert(1)//">',this.inertBodyElement.querySelector&&this.inertBodyElement.querySelector("svg img")&&function(){try{return!!window.DOMParser}catch(e){return!1}}()?this.getInertBodyElement=this.getInertBodyElement_DOMParser:this.getInertBodyElement=this.getInertBodyElement_InertDocument):this.getInertBodyElement=this.getInertBodyElement_XHR}return e.prototype.getInertBodyElement_XHR=function(e){e="<body><remove></remove>"+e+"</body>";try{e=encodeURI(e)}catch(e){return null}var t=new XMLHttpRequest;t.responseType="document",t.open("GET","data:text/html;charset=utf-8,"+e,!1),t.send(void 0);var n=t.response.body;return n.removeChild(n.firstChild),n},e.prototype.getInertBodyElement_DOMParser=function(e){e="<body><remove></remove>"+e+"</body>";try{var t=(new window.DOMParser).parseFromString(e,"text/html").body;return t.removeChild(t.firstChild),t}catch(e){return null}},e.prototype.getInertBodyElement_InertDocument=function(e){var t=this.inertDocument.createElement("template");return"content"in t?(t.innerHTML=e,t):(this.inertBodyElement.innerHTML=e,this.defaultDoc.documentMode&&this.stripCustomNsAttrs(this.inertBodyElement),this.inertBodyElement)},e.prototype.stripCustomNsAttrs=function(e){for(var t=e.attributes,n=t.length-1;0<n;n--){var r=t.item(n).name;"xmlns:ns1"!==r&&0!==r.indexOf("ns1:")||e.removeAttribute(r)}for(var i=e.firstChild;i;)i.nodeType===Node.ELEMENT_NODE&&this.stripCustomNsAttrs(i),i=i.nextSibling},e}();var gc=/^(?:(?:https?|mailto|ftp|tel|file):|[^&:/?#]*(?:[/?#]|$))/gi,vc=/^data:(?:image\/(?:bmp|gif|jpeg|jpg|png|tiff|webp)|video\/(?:mpeg|mp4|ogg|webm)|audio\/(?:mp3|oga|ogg|opus));base64,[a-z0-9+\/]+=*$/i;function _c(e){return(e=String(e)).match(gc)||e.match(vc)?e:(fc()&&console.warn("WARNING: sanitizing unsafe URL value "+e+" (see http://g.co/ng/security#xss)"),"unsafe:"+e)}function yc(e){return(e=String(e)).split(",").map(function(e){return _c(e.trim())}).join(", ")}function bc(e){var t,n,i={};try{for(var o=Object(r.__values)(e.split(",")),a=o.next();!a.done;a=o.next()){i[a.value]=!0}}catch(e){t={error:e}}finally{try{a&&!a.done&&(n=o.return)&&n.call(o)}finally{if(t)throw t.error}}return i}function jc(){for(var e,t,n=[],i=0;i<arguments.length;i++)n[i]=arguments[i];var o={};try{for(var a=Object(r.__values)(n),s=a.next();!s.done;s=a.next()){var u=s.value;for(var c in u)u.hasOwnProperty(c)&&(o[c]=!0)}}catch(t){e={error:t}}finally{try{s&&!s.done&&(t=a.return)&&t.call(a)}finally{if(e)throw e.error}}return o}var xc,Oc=bc("area,br,col,hr,img,wbr"),wc=bc("colgroup,dd,dt,li,p,tbody,td,tfoot,th,thead,tr"),Cc=bc("rp,rt"),Ec=jc(Cc,wc),Sc=jc(wc,bc("address,article,aside,blockquote,caption,center,del,details,dialog,dir,div,dl,figure,figcaption,footer,h1,h2,h3,h4,h5,h6,header,hgroup,hr,ins,main,map,menu,nav,ol,pre,section,summary,table,ul")),kc=jc(Cc,bc("a,abbr,acronym,audio,b,bdi,bdo,big,br,cite,code,del,dfn,em,font,i,img,ins,kbd,label,map,mark,picture,q,ruby,rp,rt,s,samp,small,source,span,strike,strong,sub,sup,time,track,tt,u,var,video")),Mc=jc(Oc,Sc,kc,Ec),Ac=bc("background,cite,href,itemtype,longdesc,poster,src,xlink:href"),Tc=bc("srcset"),Dc=bc("abbr,accesskey,align,alt,autoplay,axis,bgcolor,border,cellpadding,cellspacing,class,clear,color,cols,colspan,compact,controls,coords,datetime,default,dir,download,face,headers,height,hidden,hreflang,hspace,ismap,itemscope,itemprop,kind,label,lang,language,loop,media,muted,nohref,nowrap,open,preload,rel,rev,role,rows,rowspan,rules,scope,scrolling,shape,size,sizes,span,srclang,start,summary,tabindex,target,title,translate,type,usemap,valign,value,vspace,width"),Ic=jc(Ac,Tc,Dc),Pc=function(){function e(){this.sanitizedSomething=!1,this.buf=[]}return e.prototype.sanitizeChildren=function(e){for(var t=e.firstChild,n=!0;t;)if(t.nodeType===Node.ELEMENT_NODE?n=this.startElement(t):t.nodeType===Node.TEXT_NODE?this.chars(t.nodeValue):this.sanitizedSomething=!0,n&&t.firstChild)t=t.firstChild;else for(;t;){t.nodeType===Node.ELEMENT_NODE&&this.endElement(t);var r=this.checkClobberedElement(t,t.nextSibling);if(r){t=r;break}t=this.checkClobberedElement(t,t.parentNode)}return this.buf.join("")},e.prototype.startElement=function(e){var t=e.nodeName.toLowerCase();if(!Mc.hasOwnProperty(t))return this.sanitizedSomething=!0,!1;this.buf.push("<"),this.buf.push(t);for(var n=e.attributes,r=0;r<n.length;r++){var i=n.item(r),o=i.name,a=o.toLowerCase();if(Ic.hasOwnProperty(a)){var s=i.value;Ac[a]&&(s=_c(s)),Tc[a]&&(s=yc(s)),this.buf.push(" ",o,'="',Fc(s),'"')}else this.sanitizedSomething=!0}return this.buf.push(">"),!0},e.prototype.endElement=function(e){var t=e.nodeName.toLowerCase();Mc.hasOwnProperty(t)&&!Oc.hasOwnProperty(t)&&(this.buf.push("</"),this.buf.push(t),this.buf.push(">"))},e.prototype.chars=function(e){this.buf.push(Fc(e))},e.prototype.checkClobberedElement=function(e,t){if(t&&(e.compareDocumentPosition(t)&Node.DOCUMENT_POSITION_CONTAINED_BY)===Node.DOCUMENT_POSITION_CONTAINED_BY)throw new Error("Failed to sanitize html because the element is clobbered: "+e.outerHTML);return t},e}(),Rc=/[\uD800-\uDBFF][\uDC00-\uDFFF]/g,Nc=/([^\#-~ |!])/g;function Fc(e){return e.replace(/&/g,"&amp;").replace(Rc,function(e){return"&#"+(1024*(e.charCodeAt(0)-55296)+(e.charCodeAt(1)-56320)+65536)+";"}).replace(Nc,function(e){return"&#"+e.charCodeAt(0)+";"}).replace(/</g,"&lt;").replace(/>/g,"&gt;")}function Lc(e,t){var n=null;try{xc=xc||new mc(e);var r=t?String(t):"";n=xc.getInertBodyElement(r);var i=5,o=r;do{if(0===i)throw new Error("Failed to sanitize html because the input is unstable");i--,r=o,o=n.innerHTML,n=xc.getInertBodyElement(r)}while(r!==o);var a=new Pc,s=a.sanitizeChildren(Vc(n)||n);return fc()&&a.sanitizedSomething&&console.warn("WARNING: sanitizing HTML stripped some content (see http://g.co/ng/security#xss)."),s}finally{if(n)for(var u=Vc(n)||n;u.firstChild;)u.removeChild(u.firstChild)}}function Vc(e){return"content"in e&&((t=e).nodeType===Node.ELEMENT_NODE&&"TEMPLATE"===t.nodeName)?e.content:null;var t}var Bc={marker:"element"},zc={marker:"comment"},Uc="�",Hc=/^\s*(�\d+:?\d*�)\s*,\s*(select|plural)\s*,/,qc=/�\/?\*(\d+:\d+)�/gi,Gc=/�(\/?[#*]\d+):?\d*�/gi,Wc=/�(\d+):?\d*�/gi,Yc=/({\s*�\d+:?\d*�\s*,\s*\S{6}\s*,[\s\S]*})/gi,Kc=/\[(�.+?�?)\]/g,Qc=/({\s*)(VAR_(PLURAL|SELECT)(_\d+)?)(\s*,)/g,Xc=/�I18N_EXP_(ICU(_\d+)?)�/g;function Zc(e){if(!e)return[];var t,n=0,r=[],i=[],o=/[{}]/g;for(o.lastIndex=0;t=o.exec(e);){var a=t.index;if("}"==t[0]){if(r.pop(),0==r.length){var s=e.substring(n,a);Hc.test(s)?i.push($c(s)):s&&i.push(s),n=a+1}}else{if(0==r.length){var u=e.substring(n,a);i.push(u),n=a+1}r.push("{")}}var c=e.substring(n);return""!=c&&i.push(c),i}function $c(e){for(var t=[],n=[],r=1,i=0,o=Zc(e=e.replace(Hc,function(e,t,n){return r="select"===n?0:1,i=parseInt(t.substr(1),10),""})),a=0;a<o.length;){var s=o[a++].trim();1===r&&(s=s.replace(/\s*(?:=)?(\w+)\s*/,"$1")),s.length&&t.push(s);var u=Zc(o[a++]);u.length&&n.push(u)}return Oe(t.indexOf("other"),-1,'Missing key "other" in ICU statement.'),{type:r,mainBinding:i,cases:t,values:n}}function Jc(e){for(var t,n,r="",i=0,o=!1;null!==(t=qc.exec(e));)o?t[0]===Uc+"/*"+n+Uc&&(i=t.index,o=!1):(r+=e.substring(i,t.index+t[0].length),n=t[1],o=!0);return ngDevMode&&be(o,!1,'Tag mismatch: unable to find the end of the sub-template in the translation "'+e+'"'),r+=e.substr(i)}function el(e,t,n,r){void 0===r&&(r=null);for(var i=[null,null],o=e.split(Wc),a=0,s=0;s<o.length;s++){var u=o[s];if(1&s){var c=parseInt(u,10);i.push(-1-c),a|=il(c)}else""!==u&&i.push(u)}return i.push(t<<2|(n?1:0)),n&&i.push(n,r),i[0]=a,i[1]=i.length-2,i}function tl(e,t){var n;void 0===t&&(t=0),t|=il(e.mainBinding);for(var r=0;r<e.values.length;r++)for(var i=e.values[r],o=0;o<i.length;o++){var a=i[o];if("string"==typeof a)for(;n=Wc.exec(a);)t|=il(parseInt(n[1],10));else t=tl(a,t)}return t}var nl=[],rl=-1;function il(e){return 1<<Math.min(e,31)}var ol=[];function al(e,t,n){var r=gn()[Te];ngDevMode&&we(r,"tView should be defined"),nl[++rl]=e,r.firstTemplatePass&&null===r.data[e+Qe]&&function(e,t,n,r){var i=gn(),o=e.blueprint.length-Qe,a=_n(),s=jn()?_n():a&&a.parent,u=s&&s!==i[Fe]?s.index-Qe:t,c=0;ol[c]=u;var l=[];t>0&&a!==s&&l.push(a.index<<3|0);for(var d=[],p=[],f=function(e,t){if("number"!=typeof t)return Jc(e);var n=e.indexOf(":"+t+Uc)+2+t.toString().length,r=e.search(new RegExp(Uc+"\\/\\*\\d+:"+t+Uc));return Jc(e.substring(n,r))}(n,r).split(Gc),h=0;h<f.length;h++){var m=f[h];if(1&h)if("/"===m.charAt(0)){if("#"===m.charAt(1)){var g=parseInt(m.substr(2),10);u=ol[--c],l.push(g<<3|5)}}else{var g=parseInt(m.substr(1),10);l.push(g<<3|0,u<<17|1),"#"===m.charAt(0)&&(ol[++c]=u=g)}else for(var v=m.split(Yc),_=0;_<v.length;_++)if(m=v[_],1&_){Do(i);var y=e.blueprint.length-1-Qe;l.push(zc,ngDevMode?"ICU "+y:"",u<<17|1);var b=$c(m.substr(1,m.length-2)),j=tl(b);xl(p,b,y,y);var x=p.length-1;d.push(il(b.mainBinding),3,-1-b.mainBinding,y<<2|2,x,j,2,y<<2|3,x)}else if(""!==m){var O=m.match(Wc);Do(i),l.push(O?"":m,u<<17|1),O&&Tt(el(m,e.blueprint.length-1-Qe),d)}}var w={vars:e.blueprint.length-Qe-o,expandoStartIndex:o,create:l,update:d,icus:p.length?p:null};e.data[t+Qe]=w}(r,e,t,n)}function sl(e,t,n){ngDevMode&&ngDevMode.rendererMoveNode++;var r=gn();n||(n=t),n===t&&e!==t.child?(e.next=t.child,t.child=e):n!==t&&e!==n.next?(e.next=n.next,n.next=e):e.next=null,t!==r[Fe]&&(e.parent=t),pi(pt(e,r),e,r);var i=r[e.index];return 0!==e.type&&_t(i)&&pi(i[$e],e,r),e}function ul(e,t){var n={},r=e.replace(Kc,function(e,t){if(n[t]||(n[t]=t.split("|")),!n[t].length)throw new Error("i18n postprocess: unmatched placeholder - "+t);return n[t].shift()});if(Object.keys(n).some(function(e){return!!n[e].length}))throw new Error("i18n postprocess: unmatched values - "+JSON.stringify(n));return Object.keys(t).length?r=(r=r.replace(Qc,function(e,n,r,i,o,a){return t.hasOwnProperty(r)?""+n+t[r]+a:e})).replace(Xc,function(e,n){if(t.hasOwnProperty(n)){var r=t[n];if(!r.length)throw new Error("i18n postprocess: unmatched ICU - "+e+" with key: "+n);return r.shift()}return e}):r}function cl(){var e=gn()[Te];ngDevMode&&we(e,"tView should be defined"),function(e){var t=gn();ngDevMode&&be(t[Le],t[Te].bindingStartIndex,"i18nEnd should be called before any binding");var n=nl[rl--],r=e.data[n+Qe];ngDevMode&&we(r,"You should call i18nStart before i18nEnd");for(var i=_n(),o=ll(n,r.create,r.expandoStartIndex,t),a=n+1;a<=i.index-Qe;a++)-1===o.indexOf(a)&&dl(a,t)}(e)}function ll(e,t,n,r){for(var i=gn()[He],o=null,a=null,s=[],u=0;u<t.length;u++){var c=t[u];if("string"==typeof c){var l=Jr(c,i);ngDevMode&&ngDevMode.rendererCreateTextNode++,a=o,o=Ao(n++,3,l,null,null),xn(!1)}else if("number"==typeof c)switch(7&c){case 1:var d=c>>>17,p=void 0;p=d===e?r[Fe]:ft(d,r),ngDevMode&&we(o,"You need to create or select a node before you can insert it into the DOM"),a=sl(o,p,a),p.next=null;break;case 0:var f=c>>>3;s.push(f),a=o,(o=ft(f,r))&&(yn(o),3===o.type&&xn(!0));break;case 5:a=o=ft(c>>>3,r),yn(o),xn(!1);break;case 4:ea(c>>>3,t[++u],t[++u]);break;default:throw new Error('Unable to determine the type of mutate operation for "'+c+'"')}else switch(c){case zc:var h=t[++u];ngDevMode&&be(typeof h,"string",'Expected "'+h+'" to be a comment node value');var m=i.createComment(h);ngDevMode&&ngDevMode.rendererCreateComment++,a=o,o=Ao(n++,5,m,null,null),ar(m,r),o.activeCaseIndex=null,xn(!1);break;case Bc:var g=t[++u];ngDevMode&&be(typeof g,"string",'Expected "'+g+'" to be an element node tag name');var v=i.createElement(g);ngDevMode&&ngDevMode.rendererCreateElement++,a=o,o=Ao(n++,3,v,g,null);break;default:throw new Error('Unable to determine the type of mutate operation for "'+c+'"')}}return xn(!1),s}function dl(e,t){var n=ft(e,t);mi(n,dt(e,t)||null,t),n.detached=!0,ngDevMode&&ngDevMode.rendererRemoveNode++;var r=os(e);if(_t(r)){var i=r;0!==n.type&&mi(n,i[$e]||null,t),i[Je]=null}}function pl(e,t,n){al(e,t,n),cl()}function fl(e,t){var n=gn()[Te];ngDevMode&&we(n,"tView should be defined"),ngDevMode&&be(n.firstTemplatePass,!0,"You should only call i18nEnd on first template pass"),n.firstTemplatePass&&null===n.data[e+Qe]&&function(e,t,n){for(var r=_n().index-Qe,i=[],o=0;o<n.length;o+=2)for(var a=n[o],s=n[o+1],u=s.split(Yc),c=0;c<u.length;c++){var l=u[c];if(1&c);else if(""!==l){var d=!!l.match(Wc);d?Tt(el(l,r,a),i):ea(r,a,l)}}e.data[t+Qe]=i}(n,e,t)}var hl,ml,gl=0,vl=0;function _l(e){e!==Pr&&(gl|=1<<vl),vl++}function yl(e){if(vl){var t=gn(),n=t[Te];ngDevMode&&we(n,"tView should be defined");var r=n.data[e+Qe],i=void 0,o=null;Array.isArray(r)?i=r:(i=r.update,o=r.icus),!function e(t,n,r,i,o,a){void 0===a&&(a=!1);for(var s=!1,u=0;u<t.length;u++){var c=t[u],l=t[++u];if(a||c&i)for(var d="",p=u+1;p<=u+l;p++){var f=t[p];if("string"==typeof f)d+=f;else if("number"==typeof f)if(f<0)d+=ut(o[r-f]);else{var h=f>>>2;switch(3&f){case 1:ea(h,t[++p],d,t[++p]);break;case 0:ma(h,d);break;case 2:var m=t[++p],g=n[m],v=ft(h,o);if(null!==v.activeCaseIndex)for(var _=g.remove[v.activeCaseIndex],y=0;y<_.length;y++){var b=_[y];switch(7&b){case 3:dl(b>>>3,o);break;case 6:var j=ft(_[y+1]>>>3,o).activeCaseIndex;null!==j&&Tt(n[b>>>3].remove[j],_)}}var x=jl(g,d);v.activeCaseIndex=-1!==x?x:null,ll(-1,g.create[x],g.expandoStartIndex,o),s=!0;break;case 3:g=n[m=t[++p]],v=ft(h,o),e(g.update[v.activeCaseIndex],n,r,i,o,s)}}}u+=l}}(i,o,t[Le]-vl-1,gl,t),gl=0,vl=0}}function bl(e,t){switch(function(e,t){"string"==typeof t&&(t=parseInt(t,10));var n=t,r=n.toString().replace(/^[^.]*\.?/,""),i=Math.floor(Math.abs(n)),o=r.length,a=parseInt(r,10),s=parseInt(n.toString().replace(/^[^.]*\.?|0+$/g,""),10)||0;switch(e.split("-")[0].toLowerCase()){case"af":case"asa":case"az":case"bem":case"bez":case"bg":case"brx":case"ce":case"cgg":case"chr":case"ckb":case"ee":case"el":case"eo":case"es":case"eu":case"fo":case"fur":case"gsw":case"ha":case"haw":case"hu":case"jgo":case"jmc":case"ka":case"kk":case"kkj":case"kl":case"ks":case"ksb":case"ky":case"lb":case"lg":case"mas":case"mgo":case"ml":case"mn":case"nb":case"nd":case"ne":case"nn":case"nnh":case"nyn":case"om":case"or":case"os":case"ps":case"rm":case"rof":case"rwk":case"saq":case"seh":case"sn":case"so":case"sq":case"ta":case"te":case"teo":case"tk":case"tr":case"ug":case"uz":case"vo":case"vun":case"wae":case"xog":return 1===n?hl.One:hl.Other;case"ak":case"ln":case"mg":case"pa":case"ti":return n===Math.floor(n)&&n>=0&&n<=1?hl.One:hl.Other;case"am":case"as":case"bn":case"fa":case"gu":case"hi":case"kn":case"mr":case"zu":return 0===i||1===n?hl.One:hl.Other;case"ar":return 0===n?hl.Zero:1===n?hl.One:2===n?hl.Two:n%100===Math.floor(n%100)&&n%100>=3&&n%100<=10?hl.Few:n%100===Math.floor(n%100)&&n%100>=11&&n%100<=99?hl.Many:hl.Other;case"ast":case"ca":case"de":case"en":case"et":case"fi":case"fy":case"gl":case"it":case"nl":case"sv":case"sw":case"ur":case"yi":return 1===i&&0===o?hl.One:hl.Other;case"be":return n%10==1&&n%100!=11?hl.One:n%10===Math.floor(n%10)&&n%10>=2&&n%10<=4&&!(n%100>=12&&n%100<=14)?hl.Few:n%10==0||n%10===Math.floor(n%10)&&n%10>=5&&n%10<=9||n%100===Math.floor(n%100)&&n%100>=11&&n%100<=14?hl.Many:hl.Other;case"br":return n%10==1&&n%100!=11&&n%100!=71&&n%100!=91?hl.One:n%10==2&&n%100!=12&&n%100!=72&&n%100!=92?hl.Two:n%10===Math.floor(n%10)&&(n%10>=3&&n%10<=4||n%10==9)&&!(n%100>=10&&n%100<=19||n%100>=70&&n%100<=79||n%100>=90&&n%100<=99)?hl.Few:0!==n&&n%1e6==0?hl.Many:hl.Other;case"bs":case"hr":case"sr":return 0===o&&i%10==1&&i%100!=11||a%10==1&&a%100!=11?hl.One:0===o&&i%10===Math.floor(i%10)&&i%10>=2&&i%10<=4&&!(i%100>=12&&i%100<=14)||a%10===Math.floor(a%10)&&a%10>=2&&a%10<=4&&!(a%100>=12&&a%100<=14)?hl.Few:hl.Other;case"cs":case"sk":return 1===i&&0===o?hl.One:i===Math.floor(i)&&i>=2&&i<=4&&0===o?hl.Few:0!==o?hl.Many:hl.Other;case"cy":return 0===n?hl.Zero:1===n?hl.One:2===n?hl.Two:3===n?hl.Few:6===n?hl.Many:hl.Other;case"da":return 1===n||0!==s&&(0===i||1===i)?hl.One:hl.Other;case"dsb":case"hsb":return 0===o&&i%100==1||a%100==1?hl.One:0===o&&i%100==2||a%100==2?hl.Two:0===o&&i%100===Math.floor(i%100)&&i%100>=3&&i%100<=4||a%100===Math.floor(a%100)&&a%100>=3&&a%100<=4?hl.Few:hl.Other;case"ff":case"fr":case"hy":case"kab":return 0===i||1===i?hl.One:hl.Other;case"fil":return 0===o&&(1===i||2===i||3===i)||0===o&&i%10!=4&&i%10!=6&&i%10!=9||0!==o&&a%10!=4&&a%10!=6&&a%10!=9?hl.One:hl.Other;case"ga":return 1===n?hl.One:2===n?hl.Two:n===Math.floor(n)&&n>=3&&n<=6?hl.Few:n===Math.floor(n)&&n>=7&&n<=10?hl.Many:hl.Other;case"gd":return 1===n||11===n?hl.One:2===n||12===n?hl.Two:n===Math.floor(n)&&(n>=3&&n<=10||n>=13&&n<=19)?hl.Few:hl.Other;case"gv":return 0===o&&i%10==1?hl.One:0===o&&i%10==2?hl.Two:0!==o||i%100!=0&&i%100!=20&&i%100!=40&&i%100!=60&&i%100!=80?0!==o?hl.Many:hl.Other:hl.Few;case"he":return 1===i&&0===o?hl.One:2===i&&0===o?hl.Two:0!==o||n>=0&&n<=10||n%10!=0?hl.Other:hl.Many;case"is":return 0===s&&i%10==1&&i%100!=11||0!==s?hl.One:hl.Other;case"ksh":return 0===n?hl.Zero:1===n?hl.One:hl.Other;case"kw":case"naq":case"se":case"smn":return 1===n?hl.One:2===n?hl.Two:hl.Other;case"lag":return 0===n?hl.Zero:0!==i&&1!==i||0===n?hl.Other:hl.One;case"lt":return n%10!=1||n%100>=11&&n%100<=19?n%10===Math.floor(n%10)&&n%10>=2&&n%10<=9&&!(n%100>=11&&n%100<=19)?hl.Few:0!==a?hl.Many:hl.Other:hl.One;case"lv":case"prg":return n%10==0||n%100===Math.floor(n%100)&&n%100>=11&&n%100<=19||2===o&&a%100===Math.floor(a%100)&&a%100>=11&&a%100<=19?hl.Zero:n%10==1&&n%100!=11||2===o&&a%10==1&&a%100!=11||2!==o&&a%10==1?hl.One:hl.Other;case"mk":return 0===o&&i%10==1||a%10==1?hl.One:hl.Other;case"mt":return 1===n?hl.One:0===n||n%100===Math.floor(n%100)&&n%100>=2&&n%100<=10?hl.Few:n%100===Math.floor(n%100)&&n%100>=11&&n%100<=19?hl.Many:hl.Other;case"pl":return 1===i&&0===o?hl.One:0===o&&i%10===Math.floor(i%10)&&i%10>=2&&i%10<=4&&!(i%100>=12&&i%100<=14)?hl.Few:0===o&&1!==i&&i%10===Math.floor(i%10)&&i%10>=0&&i%10<=1||0===o&&i%10===Math.floor(i%10)&&i%10>=5&&i%10<=9||0===o&&i%100===Math.floor(i%100)&&i%100>=12&&i%100<=14?hl.Many:hl.Other;case"pt":return n===Math.floor(n)&&n>=0&&n<=2&&2!==n?hl.One:hl.Other;case"ro":return 1===i&&0===o?hl.One:0!==o||0===n||1!==n&&n%100===Math.floor(n%100)&&n%100>=1&&n%100<=19?hl.Few:hl.Other;case"ru":case"uk":return 0===o&&i%10==1&&i%100!=11?hl.One:0===o&&i%10===Math.floor(i%10)&&i%10>=2&&i%10<=4&&!(i%100>=12&&i%100<=14)?hl.Few:0===o&&i%10==0||0===o&&i%10===Math.floor(i%10)&&i%10>=5&&i%10<=9||0===o&&i%100===Math.floor(i%100)&&i%100>=11&&i%100<=14?hl.Many:hl.Other;case"shi":return 0===i||1===n?hl.One:n===Math.floor(n)&&n>=2&&n<=10?hl.Few:hl.Other;case"si":return 0===n||1===n||0===i&&1===a?hl.One:hl.Other;case"sl":return 0===o&&i%100==1?hl.One:0===o&&i%100==2?hl.Two:0===o&&i%100===Math.floor(i%100)&&i%100>=3&&i%100<=4||0!==o?hl.Few:hl.Other;case"tzm":return n===Math.floor(n)&&n>=0&&n<=1||n===Math.floor(n)&&n>=11&&n<=99?hl.One:hl.Other;default:return hl.Other}}(t,e)){case hl.Zero:return"zero";case hl.One:return"one";case hl.Two:return"two";case hl.Few:return"few";case hl.Many:return"many";default:return"other"}}function jl(e,t){var n=e.cases.indexOf(t);if(-1===n)switch(e.type){case 1:var r=bl(t,"en-US");-1===(n=e.cases.indexOf(r))&&"other"!==r&&(n=e.cases.indexOf("other"));break;case 0:n=e.cases.indexOf("other")}return n}function xl(e,t,n,i){for(var o=[],a=[],s=[],u=[],c=[],l=0;l<t.values.length;l++){for(var d=t.values[l],p=[],f=0;f<d.length;f++){var h=d[f];if("string"!=typeof h){var m=p.push(h)-1;d[f]="\x3c!--�"+m+"�--\x3e"}}var g=Ol(d.join(""),n,p,e,i);o.push(g.create),a.push(g.remove),s.push(g.update),u.push(g.vars),c.push(g.childIcus)}var v={type:t.type,vars:u,expandoStartIndex:i+1,childIcus:c,cases:t.cases,create:o,remove:a,update:s};e.push(v);var _=gn(),y=Math.max.apply(Math,Object(r.__spread)(u));for(l=0;l<y;l++)Do(_)}function Ol(e,t,n,i,o){var a=new mc(document).getInertBodyElement(e);if(!a)throw new Error("Unable to generate inert body element");var s={vars:0,childIcus:[],create:[],remove:[],update:[]};return function e(t,n,i,o,a,s){if(t){for(var u=[];t;){var c=t.nextSibling,l=s+ ++n.vars;switch(t.nodeType){case Node.ELEMENT_NODE:var d=t,p=d.tagName.toLowerCase();if(Mc.hasOwnProperty(p)){n.create.push(Bc,p,i<<17|1);for(var f=d.attributes,h=0;h<f.length;h++){var m=f.item(h),g=m.name.toLowerCase(),v=!!m.value.match(Wc);v?Ic.hasOwnProperty(g)?Ac[g]?Tt(el(m.value,l,m.name,_c),n.update):Tc[g]?Tt(el(m.value,l,m.name,yc),n.update):Tt(el(m.value,l,m.name),n.update):ngDevMode&&console.warn("WARNING: ignoring unsafe attribute value "+g+" on element "+p+" (see http://g.co/ng/security#xss)"):n.create.push(l<<3|4,m.name,m.value)}e(t.firstChild,n,l,o,a,s),n.remove.push(l<<3|3)}else n.vars--;break;case Node.TEXT_NODE:var _=t.textContent||"",y=_.match(Wc);n.create.push(y?"":_,i<<17|1),n.remove.push(l<<3|3),y&&Tt(el(_,l),n.update);break;case Node.COMMENT_NODE:var b=wl.exec(t.textContent||"");if(b){var j=parseInt(b[1],10),x=ngDevMode?"nested ICU "+j:"";n.create.push(zc,x,i<<17|1);var O=o[j];u.push([O,l])}else n.vars--;break;default:n.vars--}t=c}for(var h=0;h<u.length;h++){var O=u[h][0],w=u[h][1];xl(a,O,w,s+n.vars);var C=a.length-1;n.vars+=Math.max.apply(Math,Object(r.__spread)(a[C].vars)),n.childIcus.push(C);var E=tl(O);n.update.push(il(O.mainBinding),3,-1-O.mainBinding,w<<2|2,C,E,2,w<<2|3,C),n.remove.push(C<<3|6,w<<3|3)}}}((Vc(a)||a).firstChild,s,t,n,i,o),s}(ml=hl||(hl={}))[ml.Zero=0]="Zero",ml[ml.One=1]="One",ml[ml.Two=2]="Two",ml[ml.Few=3]="Few",ml[ml.Many=4]="Many",ml[ml.Other=5]="Other";var wl=/�(\d+)�/;var Cl={provide:Au,useClass:ic,deps:[Iu]},El=function(e){function t(t,n){var r=e.call(this)||this;r._parent=n,r._bootstrapComponents=[],r.injector=r,r.destroyCbs=[];var i=ye(t);ngDevMode&&we(i,"NgModule '"+K(t)+"' is not a subtype of 'NgModuleType'."),r._bootstrapComponents=i.bootstrap;var o=[{provide:Iu,useValue:r},Cl];return r._r3Injector=su(t,n,o),r.instance=r.get(t),r}return Object(r.__extends)(t,e),t.prototype.get=function(e,t,n){return void 0===t&&(t=Ls.THROW_IF_NOT_FOUND),void 0===n&&(n=Pt.Default),e===Ls||e===Iu||e===Ns?this:this._r3Injector.get(e,t,n)},Object.defineProperty(t.prototype,"componentFactoryResolver",{get:function(){return this.get(Au)},enumerable:!0,configurable:!0}),t.prototype.destroy=function(){ngDevMode&&we(this.destroyCbs,"NgModule already destroyed"),this.destroyCbs.forEach(function(e){return e()}),this.destroyCbs=null},t.prototype.onDestroy=function(e){ngDevMode&&we(this.destroyCbs,"NgModule already destroyed"),this.destroyCbs.push(e)},t}(Iu),Sl=function(e){function t(t){var n=e.call(this)||this;return n.moduleType=t,n}return Object(r.__extends)(t,e),t.prototype.create=function(e){return new El(this.moduleType,e)},t}(Pu);function kl(e,t,n,i){var o,a=e;null!==t&&(void 0!==a.decorators?(o=a.decorators).push.apply(o,Object(r.__spread)(t)):a.decorators=t),null!==n&&(a.ctorParameters=n),null!==i&&(void 0!==a.propDecorators?a.propDecorators=Object(r.__assign)({},a.propDecorators,i):a.propDecorators=i)}function Ml(e,t,n){var r=Dn()+e,i=gn();return On()?Rr(i,r,n?t.call(n):t()):Nr(i,r)}function Al(e,t,n,r){var i=gn(),o=Dn()+e;return Fr(i,o,n)?Rr(i,o+1,r?t.call(r,n):t(n)):Nr(i,o+1)}function Tl(e,t,n,r,i){var o=Dn()+e,a=gn();return Lr(a,o,n,r)?Rr(a,o+2,i?t.call(i,n,r):t(n,r)):Nr(a,o+2)}function Dl(e,t,n,r,i,o){var a=Dn()+e,s=gn();return Vr(s,a,n,r,i)?Rr(s,a+3,o?t.call(o,n,r,i):t(n,r,i)):Nr(s,a+3)}function Il(e,t,n,r,i,o,a){var s=Dn()+e,u=gn();return Br(u,s,n,r,i,o)?Rr(u,s+4,a?t.call(a,n,r,i,o):t(n,r,i,o)):Nr(u,s+4)}function Pl(e,t,n,r,i,o,a,s){var u=Dn()+e,c=gn(),l=Br(c,u,n,r,i,o);return Fr(c,u+4,a)||l?Rr(c,u+5,s?t.call(s,n,r,i,o,a):t(n,r,i,o,a)):Nr(c,u+5)}function Rl(e,t,n,r,i,o,a,s,u){var c=Dn()+e,l=gn(),d=Br(l,c,n,r,i,o);return Lr(l,c+4,a,s)||d?Rr(l,c+6,u?t.call(u,n,r,i,o,a,s):t(n,r,i,o,a,s)):Nr(l,c+6)}function Nl(e,t,n,r,i,o,a,s,u,c){var l=Dn()+e,d=gn(),p=Br(d,l,n,r,i,o);return Vr(d,l+4,a,s,u)||p?Rr(d,l+7,c?t.call(c,n,r,i,o,a,s,u):t(n,r,i,o,a,s,u)):Nr(d,l+7)}function Fl(e,t,n,r,i,o,a,s,u,c,l){var d=Dn()+e,p=gn(),f=Br(p,d,n,r,i,o);return Br(p,d+4,a,s,u,c)||f?Rr(p,d+8,l?t.call(l,n,r,i,o,a,s,u,c):t(n,r,i,o,a,s,u,c)):Nr(p,d+8)}function Ll(e,t,n,r){for(var i=Dn()+e,o=!1,a=gn(),s=0;s<n.length;s++)Fr(a,i++,n[s])&&(o=!0);return o?Rr(a,i,t.apply(r,n)):Nr(a,i)}function Vl(e,t){var n,r=gn()[Te],i=e+Qe;r.firstTemplatePass?(n=function(e,t){if(t)for(var n=t.length-1;n>=0;n--){var r=t[n];if(e===r.name)return r}throw new Error("The pipe '"+e+"' could not be found!")}(t,r.pipeRegistry),r.data[i]=n,n.onDestroy&&(r.pipeDestroyHooks||(r.pipeDestroyHooks=[])).push(i,n.onDestroy)):n=r.data[i];var o=n.factory(null);return ns(e,o),o}function Bl(e,t,n){var r=os(e);return Wl(Gl(e)?Al(t,r.transform,n,r):r.transform(n))}function zl(e,t,n,r){var i=os(e);return Wl(Gl(e)?Tl(t,i.transform,n,r,i):i.transform(n,r))}function Ul(e,t,n,r,i){var o=os(e);return Wl(Gl(e)?Dl(t,o.transform,n,r,i,o):o.transform(n,r,i))}function Hl(e,t,n,r,i,o){var a=os(e);return Wl(Gl(e)?Il(t,a.transform,n,r,i,o,a):a.transform(n,r,i,o))}function ql(e,t,n){var r=os(e);return Wl(Gl(e)?Ll(t,r.transform,n,r):r.transform.apply(r,n))}function Gl(e){return gn()[Te].data[e+Qe].pure}function Wl(e){return Mr.isWrapped(e)&&(e=Mr.unwrap(e),gn()[Dn()]=Pr),e}var Yl=function(e){function t(t){void 0===t&&(t=!1);var n=e.call(this)||this;return n.__isAsync=t,n}return Object(r.__extends)(t,e),t.prototype.emit=function(t){e.prototype.next.call(this,t)},t.prototype.subscribe=function(t,n,r){var o,a=function(e){return null},s=function(){return null};t&&"object"==typeof t?(o=this.__isAsync?function(e){setTimeout(function(){return t.next(e)})}:function(e){t.next(e)},t.error&&(a=this.__isAsync?function(e){setTimeout(function(){return t.error(e)})}:function(e){t.error(e)}),t.complete&&(s=this.__isAsync?function(){setTimeout(function(){return t.complete()})}:function(){t.complete()})):(o=this.__isAsync?function(e){setTimeout(function(){return t(e)})}:function(e){t(e)},n&&(a=this.__isAsync?function(e){setTimeout(function(){return n(e)})}:function(e){n(e)}),r&&(s=this.__isAsync?function(){setTimeout(function(){return r()})}:function(){r()}));var u=e.prototype.subscribe.call(this,o,a,s);return t instanceof i.Subscription&&t.add(u),u},t}(i.Subject),Kl=function(){function e(){}return e.__NG_ELEMENT_ID__=function(){return Xl(e,Uu)},e}(),Ql=function(e,t){return Lu(e,t,_n(),gn())},Xl=Ds,Zl=function(){function e(e,t,n){this.parent=e,this.shallow=t,this.deep=n}return e.prototype.track=function(e,t,n,r){n?this.deep=sd(this.deep,e,t,null!=r?r:null):this.shallow=sd(this.shallow,e,t,null!=r?r:null)},e.prototype.clone=function(){return new e(this,null,this.deep)},e.prototype.container=function(){var t=$l(this.shallow),n=$l(this.deep);return t||n?new e(this,t,n):null},e.prototype.createView=function(){var t=Jl(this.shallow),n=Jl(this.deep);return t||n?new e(this,t,n):null},e.prototype.insertView=function(e){ed(e,this.shallow),ed(e,this.deep)},e.prototype.addNode=function(e){return od(this.deep,e),mt(e)?(od(this.shallow,e),e.parent&&mt(e.parent)&&od(this.parent.shallow,e),this.parent):((null===(t=e).parent||mt(t.parent))&&od(this.shallow,e),this);var t},e.prototype.removeView=function(){td(this.shallow),td(this.deep)},e}();function $l(e){for(var t=null;e;){var n=[];e.values.push(n),t={next:t,list:e.list,predicate:e.predicate,values:n,containerValues:null},e=e.next}return t}function Jl(e){for(var t=null;e;){t={next:t,list:e.list,predicate:e.predicate,values:[],containerValues:e.values},e=e.next}return t}function ed(e,t){for(;t;)ngDevMode&&we(t.containerValues,"View queries need to have a pointer to container values."),t.containerValues.splice(e,0,t.values),t=t.next}function td(e){for(;e;){ngDevMode&&we(e.containerValues,"View queries need to have a pointer to container values.");var t=e.containerValues,n=t.indexOf(e.values),r=t.splice(n,1);ngDevMode&&be(r.length,1,"removed.length"),r[0].length&&e.list.setDirty(),e=e.next}}function nd(e,t){var n=e.localNames;if(n)for(var r=0;r<n.length;r+=2)if(n[r]===t)return n[r+1];return null}function rd(e,t,n){var r=e[m];if("function"==typeof r)return r();var i=Xn(t,n,e,!1,!1);return null!==i?Zn(n[Te].data,n,i,t):null}function id(e,t,n,r){return n?rd(n,e,t):r>-1?Zn(t[Te].data,t,r,e):(o=t,3===(i=e).type||4===i.type?Fu(Uu,i,o):0===i.type?Lu(Kl,Uu,i,o):null);var i,o}function od(e,t){for(var n,r,i,o,a,s=gn();e;){var u=e.predicate,c=u.type;if(c){var l=null;if(c===Kl)n=c,r=t,i=s,o=u.read,void 0,a=n[m](),l=o?a?rd(o,r,i):null:a;else null!==(f=Xn(t,s,c,!1,!1))&&(l=id(t,s,u.read,f));null!==l&&ad(e,l)}else for(var d=u.selector,p=0;p<d.length;p++){var f;if(null!==(f=nd(t,d[p])))null!==(l=id(t,s,u.read,f))&&ad(e,l)}e=e.next}}function ad(e,t){e.values.push(t),e.list.setDirty()}function sd(e,t,n,r){return{next:e,list:t,predicate:(i=n,o=r,a=Array.isArray(i),{type:a?null:i,selector:a?i:null,read:o}),values:t._valuesTree,containerValues:null};var i,o,a}var ud=function(){function e(){this.dirty=!0,this.changes=new Yl,this._values=[],this._valuesTree=[]}return Object.defineProperty(e.prototype,"length",{get:function(){return this._values.length},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"first",{get:function(){var e=this._values;return e.length?e[0]:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"last",{get:function(){var e=this._values;return e.length?e[e.length-1]:null},enumerable:!0,configurable:!0}),e.prototype.map=function(e){return this._values.map(e)},e.prototype.filter=function(e){return this._values.filter(e)},e.prototype.find=function(e){return this._values.find(e)},e.prototype.reduce=function(e,t){return this._values.reduce(e,t)},e.prototype.forEach=function(e){this._values.forEach(e)},e.prototype.some=function(e){return this._values.some(e)},e.prototype.toArray=function(){return this._values.slice(0)},e.prototype[G()]=function(){return this._values[G()]()},e.prototype.toString=function(){return this._values.toString()},e.prototype.reset=function(e){this._values=function(e){for(var t=[],n=0;n<e.length;){var r=e[n];Array.isArray(r)?r.length>0?(e=r.concat(e.slice(n+1)),n=0):n++:(t.push(r),n++)}return t}(e),this.dirty=!1},e.prototype.notifyOnChanges=function(){this.changes.emit(this)},e.prototype.setDirty=function(){this.dirty=!0},e.prototype.destroy=function(){this.changes.complete(),this.changes.unsubscribe()},e}();function cd(e,t,n,r){ngDevMode&&ke(jn());var i,o,a,s,u,c,l,d=new ud;return(i=Zl,o=gn(),a=o[Re],cn&&cn!==o[Fe]&&!mt(cn)&&(a&&(a=o[Re]=a.clone()),cn.flags|=4),a||(o[Re]=new i(null,null,null))).track(d,t,n,r),s=gn(),u=d,c=d.destroy,(l=ps(s)).push(u),s[Te].firstTemplatePass&&fs(s).push(c,l.length-1),null!=e&&ns(e,d),d}function ld(e){var t=e;return!!e.dirty&&(e.reset(t._valuesTree),e.notifyOnChanges(),!0)}function dd(e,t){return Lu(Kl,Uu,e,t)}var pd="__SANITIZER_TRUSTED_BRAND__";function fd(e,t){return e instanceof String&&e[pd]===t}function hd(e){return yd(e,"Html")}function md(e){return yd(e,"Style")}function gd(e){return yd(e,"Script")}function vd(e){return yd(e,"Url")}function _d(e){return yd(e,"ResourceUrl")}function yd(e,t){var n=new String(e);return n[pd]=t,n}var bd=new RegExp("^([-,.\"'%_!# a-zA-Z0-9]+|(?:(?:matrix|translate|scale|rotate|skew|perspective)(?:X|Y|3d)?|(?:rgb|hsl)a?|(?:repeating-)?(?:linear|radial)-gradient|(?:calc|attr))\\([-0-9.%, #a-zA-Z]+\\))$","g"),jd=/^url\(([^)]+)\)$/;function xd(e){if(!(e=String(e).trim()))return"";var t=e.match(jd);return t&&_c(t[1])===t[1]||e.match(bd)&&function(e){for(var t=!0,n=!0,r=0;r<e.length;r++){var i=e.charAt(r);"'"===i&&n?t=!t:'"'===i&&t&&(n=!n)}return t&&n}(e)?e:(fc()&&console.warn("WARNING: sanitizing unsafe style value "+e+" (see http://g.co/ng/security#xss)."),"unsafe")}function Od(e){var t=Sd();return t?t.sanitize(Xu.HTML,e)||"":fd(e,"Html")?e.toString():Lc(document,ut(e))}function wd(e){var t=Sd();return t?t.sanitize(Xu.STYLE,e)||"":fd(e,"Style")?e.toString():xd(ut(e))}function Cd(e){var t=Sd();return t?t.sanitize(Xu.URL,e)||"":fd(e,"Url")?e.toString():_c(ut(e))}function Ed(e){var t=Sd();if(t)return t.sanitize(Xu.RESOURCE_URL,e)||"";if(fd(e,"ResourceUrl"))return e.toString();throw new Error("unsafe value used in a resource URL context (see http://g.co/ng/security#xss)")}function Sd(){var e=gn();return e&&e[qe]}var kd={"ɵdefineBase":fe,"ɵdefineComponent":ue,"ɵdefineDirective":he,defineInjectable:g,defineInjector:v,"ɵdefineNgModule":de,"ɵdefinePipe":me,"ɵdirectiveInject":as,"ɵgetFactoryOf":tr,"ɵgetInheritedFactory":nr,inject:Yt,"ɵinjectAttribute":ss,"ɵtemplateRefExtractor":dd,"ɵNgOnChangesFeature":As,"ɵProvidersFeature":ju,"ɵInheritDefinitionFeature":Ss,"ɵelementAttribute":ea,"ɵbind":Ga,"ɵcontainer":Ea,"ɵnextContext":Po,"ɵcontainerRefreshStart":ka,"ɵcontainerRefreshEnd":Ma,"ɵloadQueryList":is,"ɵnamespaceHTML":Bo,"ɵnamespaceMathML":Vo,"ɵnamespaceSVG":Lo,"ɵenableBindings":hn,"ɵdisableBindings":mn,"ɵallocHostVars":Wa,"ɵelementStart":qo,"ɵelementEnd":Jo,"ɵelement":zo,"ɵelementContainerStart":Uo,"ɵelementContainerEnd":Ho,"ɵpureFunction0":Ml,"ɵpureFunction1":Al,"ɵpureFunction2":Tl,"ɵpureFunction3":Dl,"ɵpureFunction4":Il,"ɵpureFunction5":Pl,"ɵpureFunction6":Rl,"ɵpureFunction7":Nl,"ɵpureFunction8":Fl,"ɵpureFunctionV":Ll,"ɵgetCurrentView":ds,"ɵrestoreView":vn,"ɵinterpolation1":Ka,"ɵinterpolation2":Qa,"ɵinterpolation3":Xa,"ɵinterpolation4":Za,"ɵinterpolation5":$a,"ɵinterpolation6":Ja,"ɵinterpolation7":es,"ɵinterpolation8":ts,"ɵinterpolationV":Ya,"ɵelementClassProp":pa,"ɵlistener":$o,"ɵload":os,"ɵprojection":Ra,"ɵelementProperty":ta,"ɵcomponentHostSyntheticProperty":na,"ɵpipeBind1":Bl,"ɵpipeBind2":zl,"ɵpipeBind3":Ul,"ɵpipeBind4":Hl,"ɵpipeBindV":ql,"ɵprojectionDef":Ia,"ɵpipe":Vl,"ɵquery":cd,"ɵqueryRefresh":ld,"ɵregisterContentQuery":us,"ɵreference":rs,"ɵelementStyling":ua,"ɵelementHostAttrs":ca,"ɵelementStylingMap":fa,"ɵelementStyleProp":da,"ɵelementStylingApply":la,"ɵtemplate":Ca,"ɵtext":ha,"ɵtextBinding":ma,"ɵembeddedViewStart":Aa,"ɵembeddedViewEnd":Ta,"ɵi18n":pl,"ɵi18nAttributes":fl,"ɵi18nExp":_l,"ɵi18nStart":al,"ɵi18nEnd":cl,"ɵi18nApply":yl,"ɵi18nPostprocess":ul,"ɵsanitizeHtml":Od,"ɵsanitizeStyle":wd,"ɵdefaultStyleSanitizer":function(e,t){return void 0===t?"background-image"===e||"background"===e||"border-image"===e||"filter"===e||"filter"===e||"list-style"===e||"list-style-image"===e:wd(t)},"ɵsanitizeResourceUrl":Ed,"ɵsanitizeScript":function(e){var t=Sd();if(t)return t.sanitize(Xu.SCRIPT,e)||"";if(fd(e,"Script"))return e.toString();throw new Error("unsafe value used in a script context")},"ɵsanitizeUrl":Cd},Md=function(){return function(){}}(),Ad=new Map;function Td(e,t){var n=Ad.get(e);Dd(e,n&&n.moduleType),Ad.set(e,t)}function Dd(e,t){if(t)throw new Error("Duplicate module registered for "+e+" - "+K(t)+" vs "+K(t.name))}function Id(e){var t=Ad.get(e);if(!t)throw Nd(e);return t}function Pd(e){var t=Ad.get(e);if(!t)throw Nd(e);return new Sl(t)}var Rd=Id;function Nd(e){return new Error("No module with ID "+e+" loaded")}var Fd=Function;function Ld(e){return"function"==typeof e}var Vd=/^function\s+\S+\(\)\s*{[\s\S]+\.apply\(this,\s*arguments\)/,Bd=/^class\s+[A-Za-z\d$_]*\s*extends\s+[^{]+{/,zd=/^class\s+[A-Za-z\d$_]*\s*extends\s+[^{]+{[\s\S]*constructor\s*\(/,Ud=function(){function e(e){this._reflect=e||U.Reflect}return e.prototype.isReflectionEnabled=function(){return!0},e.prototype.factory=function(e){return function(){for(var t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];return new(e.bind.apply(e,Object(r.__spread)([void 0],t)))}},e.prototype._zipTypesAndAnnotations=function(e,t){var n;n=void 0===e?new Array(t.length):new Array(e.length);for(var r=0;r<n.length;r++)void 0===e?n[r]=[]:e[r]!=Object?n[r]=[e[r]]:n[r]=[],t&&null!=t[r]&&(n[r]=n[r].concat(t[r]));return n},e.prototype._ownParameters=function(e,t){var n=e.toString();if(Vd.exec(n)||Bd.exec(n)&&!zd.exec(n))return null;if(e.parameters&&e.parameters!==t.parameters)return e.parameters;var r=e.ctorParameters;if(r&&r!==t.ctorParameters){var i="function"==typeof r?r():r,o=i.map(function(e){return e&&e.type}),a=i.map(function(e){return e&&Hd(e.decorators)});return this._zipTypesAndAnnotations(o,a)}var s=e.hasOwnProperty(x)&&e[x],u=this._reflect&&this._reflect.getOwnMetadata&&this._reflect.getOwnMetadata("design:paramtypes",e);return u||s?this._zipTypesAndAnnotations(u,s):new Array(e.length).fill(void 0)},e.prototype.parameters=function(e){if(!Ld(e))return[];var t=qd(e),n=this._ownParameters(e,t);return n||t===Object||(n=this.parameters(t)),n||[]},e.prototype._ownAnnotations=function(e,t){if(e.annotations&&e.annotations!==t.annotations){var n=e.annotations;return"function"==typeof n&&n.annotations&&(n=n.annotations),n}return e.decorators&&e.decorators!==t.decorators?Hd(e.decorators):e.hasOwnProperty(j)?e[j]:null},e.prototype.annotations=function(e){if(!Ld(e))return[];var t=qd(e),n=this._ownAnnotations(e,t)||[];return(t!==Object?this.annotations(t):[]).concat(n)},e.prototype._ownPropMetadata=function(e,t){if(e.propMetadata&&e.propMetadata!==t.propMetadata){var n=e.propMetadata;return"function"==typeof n&&n.propMetadata&&(n=n.propMetadata),n}if(e.propDecorators&&e.propDecorators!==t.propDecorators){var r=e.propDecorators,i={};return Object.keys(r).forEach(function(e){i[e]=Hd(r[e])}),i}return e.hasOwnProperty(O)?e[O]:null},e.prototype.propMetadata=function(e){if(!Ld(e))return{};var t=qd(e),n={};if(t!==Object){var i=this.propMetadata(t);Object.keys(i).forEach(function(e){n[e]=i[e]})}var o=this._ownPropMetadata(e,t);return o&&Object.keys(o).forEach(function(e){var t=[];n.hasOwnProperty(e)&&t.push.apply(t,Object(r.__spread)(n[e])),t.push.apply(t,Object(r.__spread)(o[e])),n[e]=t}),n},e.prototype.hasLifecycleHook=function(e,t){return e instanceof Fd&&t in e.prototype},e.prototype.guards=function(e){return{}},e.prototype.getter=function(e){return new Function("o","return o."+e+";")},e.prototype.setter=function(e){return new Function("o","v","return o."+e+" = v;")},e.prototype.method=function(e){return new Function("o","args","if (!o."+e+") throw new Error('\""+e+"\" is undefined');\n        return o."+e+".apply(o, args);")},e.prototype.importUri=function(e){return"object"==typeof e&&e.filePath?e.filePath:"./"+K(e)},e.prototype.resourceUri=function(e){return"./"+K(e)},e.prototype.resolveIdentifier=function(e,t,n,r){return r},e.prototype.resolveEnum=function(e,t){return e[t]},e}();function Hd(e){return e?e.map(function(e){var t=e.type.annotationCls,n=e.args?e.args:[];return new(t.bind.apply(t,Object(r.__spread)([void 0],n)))}):[]}function qd(e){var t=e.prototype?Object.getPrototypeOf(e.prototype):null;return(t?t.constructor:null)||Object}var Gd=null;function Wd(){return Gd=Gd||new Ud}function Yd(e){return Kd(Wd().parameters(e))}function Kd(e){var t=It();return e.map(function(e){return function(e,t){var n={token:null,host:!1,optional:!1,resolved:e.R3ResolvedDependencyType.Token,self:!1,skipSelf:!1};function r(t){n.resolved=e.R3ResolvedDependencyType.Token,n.token=t}if(Array.isArray(t)){if(0===t.length)throw new Error("Dependency array must have arguments.");for(var i=0;i<t.length;i++){var o=t[i];if(void 0!==o)if(o instanceof Ft||"Optional"===o.__proto__.ngMetadataName)n.optional=!0;else if(o instanceof Vt||"SkipSelf"===o.__proto__.ngMetadataName)n.skipSelf=!0;else if(o instanceof Lt||"Self"===o.__proto__.ngMetadataName)n.self=!0;else if(o instanceof Bt||"Host"===o.__proto__.ngMetadataName)n.host=!0;else if(o instanceof Nt)n.token=o.token;else if(o instanceof I){if(void 0===o.attributeName)throw new Error("Attribute name must be defined.");n.token=o.attributeName,n.resolved=e.R3ResolvedDependencyType.Attribute}else r(o)}}else r(t);return n}(t,e)})}var Qd=[],Xd=[];var Zd=!1;function $d(e){return Array.isArray(e)?e.every($d):!!Z(e)}function Jd(e,t){var n,r;void 0===t&&(t={}),ep(e,t),n=e,r=t,Xd.push({moduleType:n,ngModule:r})}function ep(e,t){ngDevMode&&we(e,"Required value moduleType"),ngDevMode&&we(t,"Required value ngModule");var n,r,i=dp(t.declarations||Qd),o=null;Object.defineProperty(e,f,{configurable:!0,get:function(){return null===o&&(o=It().compileNgModule(kd,"ng://"+e.name+"/ngModuleDef.js",{type:e,bootstrap:dp(t.bootstrap||Qd,Z),declarations:i.map(Z),imports:dp(t.imports||Qd,Z).map(pp),exports:dp(t.exports||Qd,Z).map(pp),emitInline:!0})),o}}),t.id&&(n=t.id,r=e,Dd(n,Ad.get(n)),Ad.set(n,r));var a=null;Object.defineProperty(e,d,{get:function(){if(null===a){ngDevMode&&tp(e);var n={name:e.name,type:e,deps:Yd(e),providers:t.providers||Qd,imports:[(t.imports||Qd).map(Z),(t.exports||Qd).map(Z)]};a=It().compileInjector(kd,"ng://"+e.name+"/ngInjectorDef.js",n)}return a},configurable:!!ngDevMode})}function tp(e){if(!op.get(e)){op.set(e,!0);var t=ye(e=Z(e),!0),n=[];t.declarations.forEach(function(t){ge(t=Z(t))||ve(t)||_e(t)||n.push("Unexpected value '"+ut(t)+"' declared by the module '"+ut(e)+"'. Please add a @Pipe/@Directive/@Component annotation.")});var i=Object(r.__spread)(t.declarations.map(Z),dp(t.imports.map(sp),Z));t.exports.forEach(function(t){var r=(ge(t=Z(t))?"component":ve(t)&&"directive")||_e(t)&&"pipe";r&&-1===i.lastIndexOf(t)&&n.push("Can't export "+r+" "+ut(t)+" from "+ut(e)+" as it was neither declared nor imported!")}),t.declarations.forEach(function(t){t=Z(t);var r=ip.get(t);if(r&&r!==e){var i=[r,e].map(ut).sort();n.push("Type "+ut(t)+" is part of the declarations of 2 modules: "+i[0]+" and "+i[1]+"! Please consider moving "+ut(t)+" to a higher module that imports "+i[0]+" and "+i[1]+". You can also create a new NgModule that exports and includes "+ut(t)+" then import that NgModule in "+i[0]+" and "+i[1]+".")}else ip.set(t,e)}),t.declarations.forEach(function(e){if(ge(e=Z(e))){var t=rp(e,"Component");t&&t.entryComponents&&t.entryComponents.forEach(a)}});var o=rp(e,"NgModule");if(o&&(o.imports&&dp(o.imports,np).forEach(tp),o.bootstrap&&o.bootstrap.forEach(a),o.entryComponents&&o.entryComponents.forEach(a)),n.length)throw new Error(n.join("\n"))}function a(e){e=Z(e),ip.get(e)||n.push("Component "+ut(e)+" is not part of any NgModule or the module has not been imported into your module.")}}function np(e){return(e=Z(e)).ngModule||e}function rp(e,t){var n=null;return r(e.__annotations__),r(e.decorators),n;function r(e){e&&e.forEach(i)}function i(e){if(!n)if(Object.getPrototypeOf(e).ngMetadataName==t)n=e;else if(e.type){Object.getPrototypeOf(e.type).ngMetadataName==t&&(n=e.args[0])}}}var ip=new Map,op=new Map;function ap(){ip=new Map,op=new Map,Xd.length=0}function sp(e){var t=ye(e=Z(e),!0);return Object(r.__spread)(dp(t.exports.map(function(e){return ye(e)?(tp(e),sp(e)):e})))}function up(e,t){var n=dp(t.declarations||Qd),r=lp(e);n.forEach(function(t){t.hasOwnProperty(u)?cp(ge(t),r):t.hasOwnProperty(c)||t.hasOwnProperty(p)||(t.ngSelectorScope=e)})}function cp(e,t){e.directiveDefs=function(){return Array.from(t.compilation.directives).map(function(e){return ve(e)||ge(e)}).filter(function(e){return!!e})},e.pipeDefs=function(){return Array.from(t.compilation.pipes).map(function(e){return _e(e)})}}function lp(e){if(!fp(e))throw new Error(e.name+" does not have an ngModuleDef");var t=ye(e);if(null!==t.transitiveCompileScopes)return t.transitiveCompileScopes;var n={compilation:{directives:new Set,pipes:new Set},exported:{directives:new Set,pipes:new Set}};return t.declarations.forEach(function(e){_e(e)?n.compilation.pipes.add(e):n.compilation.directives.add(e)}),t.imports.forEach(function(e){var t=e;if(!fp(t))throw new Error("Importing "+t.name+" which does not have an ngModuleDef");var r=lp(t);r.exported.directives.forEach(function(e){return n.compilation.directives.add(e)}),r.exported.pipes.forEach(function(e){return n.compilation.pipes.add(e)})}),t.exports.forEach(function(e){var t=e;if(fp(t)){var r=lp(t);r.exported.directives.forEach(function(e){n.compilation.directives.add(e),n.exported.directives.add(e)}),r.exported.pipes.forEach(function(e){n.compilation.pipes.add(e),n.exported.pipes.add(e)})}else _e(t)?n.exported.pipes.add(t):n.exported.directives.add(t)}),t.transitiveCompileScopes=n,n}function dp(e,t){var n=[];return e.forEach(function(e){Array.isArray(e)?n.push.apply(n,Object(r.__spread)(dp(e,t))):n.push(t?t(e):e)}),n}function pp(e){return void 0!==e.ngModule?e.ngModule:e}function fp(e){return!!ye(e)}function hp(e,t){var n,i=null;re(n=t)&&ne.add(n),Object.defineProperty(e,u,{get:function(){var n=It();if(null===i){if(re(t)){var o=["Component '"+ut(e)+"' is not resolved:"];throw t.templateUrl&&o.push(" - templateUrl: "+ut(t.templateUrl)),t.styleUrls&&t.styleUrls.length&&o.push(" - styleUrls: "+JSON.stringify(t.styleUrls)),o.push("Did you run and wait for 'resolveComponentResources()'?"),new Error(o.join("\n"))}var a=Object(r.__assign)({},gp(e,t),{template:t.template||"",preserveWhitespaces:t.preserveWhitespaces||!1,styles:t.styles||ae,animations:t.animations,viewQueries:_p(e,Wd().propMetadata(e),bp),directives:[],changeDetection:t.changeDetection,pipes:new Map,encapsulation:t.encapsulation||J.Emulated,interpolation:t.interpolation,viewProviders:t.viewProviders||null});if(i=n.compileComponent(kd,"ng://"+ut(e)+"/template.html",a),function(){if(!Zd){Zd=!0;try{for(var e=Xd.length-1;e>=0;e--){var t=Xd[e],n=t.moduleType,r=t.ngModule;r.declarations&&r.declarations.every($d)&&(Xd.splice(e,1),up(n,r))}}finally{Zd=!1}}}(),void 0!==e.ngSelectorScope){var s=lp(e.ngSelectorScope);cp(i,s)}}return i},configurable:!!ngDevMode})}function mp(e,t){var n=null;Object.defineProperty(e,c,{get:function(){if(null===n){var r=gp(e,t);n=It().compileDirective(kd,"ng://"+(e&&e.name)+"/ngDirectiveDef.js",r)}return n},configurable:!!ngDevMode})}function gp(e,t){var n,r=Wd().propMetadata(e);return{name:e.name,type:e,typeArgumentCount:0,selector:t.selector,deps:Yd(e),host:t.host||oe,propMetadata:r,inputs:t.inputs||ae,outputs:t.outputs||ae,queries:_p(e,r,yp),lifecycle:{usesOnChanges:void 0!==e.prototype.ngOnChanges},typeSourceSpan:null,usesInheritance:(n=e,!(Object.getPrototypeOf(n.prototype)===Object.prototype)),exportAs:t.exportAs||null,providers:t.providers||null}}function vp(e,t){return{propertyName:e,predicate:(n=t.selector,"string"==typeof n?n.split(",").map(function(e){return e.trim()}):Z(n)),descendants:t.descendants,first:t.first,read:t.read?t.read:null};var n}function _p(e,t,n){var r=[],i=function(i){t.hasOwnProperty(i)&&t[i].forEach(function(t){if(n(t)){if(!t.selector)throw new Error("Can't construct a query for the property \""+i+'" of "'+ut(e)+"\" since the query selector wasn't defined.");r.push(vp(i,t))}})};for(var o in t)i(o);return r}function yp(e){var t=e.ngMetadataName;return"ContentChild"===t||"ContentChildren"===t}function bp(e){var t=e.ngMetadataName;return"ViewChild"===t||"ViewChildren"===t}function jp(e,t){var n=null;Object.defineProperty(e,p,{get:function(){return null===n&&(n=It().compilePipe(kd,"ng://"+ut(e)+"/ngPipeDef.js",{type:e,name:e.name,deps:Yd(e),pipeName:t.name,pure:void 0===t.pure||t.pure})),n},configurable:!!ngDevMode})}var xp=w("Directive",function(e){return void 0===e&&(e={}),e},void 0,void 0,function(e,t){return Pp(e,t)}),Op=w("Component",function(e){return void 0===e&&(e={}),Object(r.__assign)({changeDetection:k.Default},e)},xp,void 0,function(e,t){return Ip(e,t)}),wp=w("Pipe",function(e){return Object(r.__assign)({pure:!0},e)},void 0,void 0,function(e,t){return Rp(e,t)}),Cp=function(e){return function(t,n){for(var r=[],i=2;i<arguments.length;i++)r[i-2]=arguments[i];var o,a,u,c=t.constructor;c.hasOwnProperty(h)||(o=t.constructor,a=o.ngBaseDef,u=o.ngBaseDef={inputs:{},outputs:{},declaredInputs:{}},a&&(s(u.inputs,a.inputs),s(u.outputs,a.outputs),s(u.declaredInputs,a.declaredInputs)));var l=c.ngBaseDef;e(l)[n]=r[0]}},Ep=S("Input",function(e){return{bindingPropertyName:e}},void 0,Cp(function(e){return e.inputs||{}})),Sp=S("Output",function(e){return{bindingPropertyName:e}},void 0,Cp(function(e){return e.outputs||{}})),kp=S("HostBinding",function(e){return{hostPropertyName:e}}),Mp=S("HostListener",function(e,t){return{eventName:e,args:t}}),Ap=hp,Tp=mp,Dp=jp,Ip=Ds,Pp=Ds,Rp=Ds,Np=a({provide:String,useValue:a}),Fp=[];function Lp(e,t){if(!t){var n=(c=new Ud).parameters(e);return function(){return new(e.bind.apply(e,Object(r.__spread)([void 0],Qt(n))))}}if(Np in t){var i=t;return function(){return i.useValue}}if(t.useExisting){var o=t;return function(){return Yt(o.useExisting)}}if(t.useFactory){var a=t;return function(){return a.useFactory.apply(a,Object(r.__spread)(Qt(a.deps||Fp)))}}if(t.useClass){var s=t,u=t.deps;if(!u){var c=new Ud;u=c.parameters(e)}return function(){var e;return new((e=s.useClass).bind.apply(e,Object(r.__spread)([void 0],Qt(u))))}}var l=t.deps;if(!l){c=new Ud;l=c.parameters(e)}return function(){return new(e.bind.apply(e,Object(r.__spread)([void 0],Qt(l))))}}var Vp={name:"custom-elements"},Bp={name:"no-errors-schema"},zp=w("NgModule",function(e){return e},void 0,void 0,function(e,t){return Hp(e,t)});var Up=Jd,Hp=function(e,t){var n=t&&t.imports||[];t&&t.exports&&(n=Object(r.__spread)(n,[t.exports])),e.ngInjectorDef=v({factory:Lp(e,{useClass:e}),providers:t&&t.providers,imports:n})};var qp=a({provide:String,useValue:a});function Gp(e){return void 0!==e.useClass}function Wp(e){return qp in e}function Yp(e){return void 0!==e.useFactory}function Kp(e){return void 0!==e.useExisting}var Qp=w("Injectable",void 0,void 0,void 0,function(e,t){return Zp(e,t)});var Xp=function(e,t){var n=null;e.hasOwnProperty(l)||Object.defineProperty(e,l,{get:function(){if(null===n){var r=t||{providedIn:null},i=Gp(r)||Yp(r)||Wp(r)||Kp(r),o={name:e.name,type:e,typeArgumentCount:0,providedIn:r.providedIn,ctorDeps:Yd(e),userDeps:void 0};if((Gp(r)||Yp(r))&&void 0!==r.deps&&(o.userDeps=Kd(r.deps)),i)if(Gp(r))o.useClass=r.useClass;else if(Wp(r))o.useValue=r.useValue;else if(Yp(r))o.useFactory=r.useFactory;else{if(!Kp(r))throw new Error("Unreachable state.");o.useExisting=r.useExisting}else o.useClass=e;n=It().compileInjectable(kd,"ng://"+e.name+"/ngInjectableDef.js",o)}return n}})},Zp=function(e,t){t&&void 0!==t.providedIn&&!_(e)&&(e.ngInjectableDef=g({providedIn:t.providedIn,factory:Lp(e,t)}))},$p="ngDebugContext",Jp="ngOriginalError",ef="ngErrorLogger";function tf(e){return e[$p]}function nf(e){return e[Jp]}function rf(e){for(var t=[],n=1;n<arguments.length;n++)t[n-1]=arguments[n];e.error.apply(e,Object(r.__spread)(t))}var of=function(){function e(){this._console=console}return e.prototype.handleError=function(e){var t=this._findOriginalError(e),n=this._findContext(e),r=e[ef]||rf;r(this._console,"ERROR",e),t&&r(this._console,"ORIGINAL ERROR",t),n&&r(this._console,"ERROR CONTEXT",n)},e.prototype._findContext=function(e){return e?tf(e)?tf(e):this._findContext(nf(e)):null},e.prototype._findOriginalError=function(e){for(var t=nf(e);t&&nf(t);)t=nf(t);return t},e}();function af(e){return e.length>1?" ("+function(e){for(var t=[],n=0;n<e.length;++n){if(t.indexOf(e[n])>-1)return t.push(e[n]),t;t.push(e[n])}return t}(e.slice().reverse()).map(function(e){return K(e.token)}).join(" -> ")+")":""}function sf(e,t,n,r){var i,o,a,s=[t],u=n(s),c=r?(o=u+" caused by: "+((i=r)instanceof Error?i.message:i),(a=Error(o))[Jp]=i,a):Error(u);return c.addKey=uf,c.keys=s,c.injectors=[e],c.constructResolvingMessage=n,c[Jp]=r,c}function uf(e,t){this.injectors.push(e),this.keys.push(t),this.message=this.constructResolvingMessage(this.keys)}function cf(e,t){for(var n=[],r=0,i=t.length;r<i;r++){var o=t[r];o&&0!=o.length?n.push(o.map(K).join(" ")):n.push("?")}return Error("Cannot resolve all parameters for '"+K(e)+"'("+n.join(", ")+"). Make sure that all the parameters are decorated with Inject or have valid type annotations and that '"+K(e)+"' is decorated with Injectable.")}var lf=function(){function e(e,t){if(this.token=e,this.id=t,!e)throw new Error("Token must be defined!");this.displayName=K(this.token)}return e.get=function(e){return df.get(Z(e))},Object.defineProperty(e,"numberOfKeys",{get:function(){return df.numberOfKeys},enumerable:!0,configurable:!0}),e}(),df=new(function(){function e(){this._allKeys=new Map}return e.prototype.get=function(e){if(e instanceof lf)return e;if(this._allKeys.has(e))return this._allKeys.get(e);var t=new lf(e,lf.numberOfKeys);return this._allKeys.set(e,t),t},Object.defineProperty(e.prototype,"numberOfKeys",{get:function(){return this._allKeys.size},enumerable:!0,configurable:!0}),e}()),pf=new(function(){function e(e){this.reflectionCapabilities=e}return e.prototype.updateCapabilities=function(e){this.reflectionCapabilities=e},e.prototype.factory=function(e){return this.reflectionCapabilities.factory(e)},e.prototype.parameters=function(e){return this.reflectionCapabilities.parameters(e)},e.prototype.annotations=function(e){return this.reflectionCapabilities.annotations(e)},e.prototype.propMetadata=function(e){return this.reflectionCapabilities.propMetadata(e)},e.prototype.hasLifecycleHook=function(e,t){return this.reflectionCapabilities.hasLifecycleHook(e,t)},e.prototype.getter=function(e){return this.reflectionCapabilities.getter(e)},e.prototype.setter=function(e){return this.reflectionCapabilities.setter(e)},e.prototype.method=function(e){return this.reflectionCapabilities.method(e)},e.prototype.importUri=function(e){return this.reflectionCapabilities.importUri(e)},e.prototype.resourceUri=function(e){return this.reflectionCapabilities.resourceUri(e)},e.prototype.resolveIdentifier=function(e,t,n,r){return this.reflectionCapabilities.resolveIdentifier(e,t,n,r)},e.prototype.resolveEnum=function(e,t){return this.reflectionCapabilities.resolveEnum(e,t)},e}())(new Ud),ff=function(){function e(e,t,n){this.key=e,this.optional=t,this.visibility=n}return e.fromKey=function(t){return new e(t,!1,null)},e}(),hf=[],mf=function(){return function(e,t,n){this.key=e,this.resolvedFactories=t,this.multiProvider=n,this.resolvedFactory=this.resolvedFactories[0]}}(),gf=function(){return function(e,t){this.factory=e,this.dependencies=t}}();function vf(e){var t,n;if(e.useClass){var r=Z(e.useClass);t=pf.factory(r),n=bf(r)}else e.useExisting?(t=function(e){return e},n=[ff.fromKey(lf.get(e.useExisting))]):e.useFactory?(t=e.useFactory,n=function(e,t){if(t){var n=t.map(function(e){return[e]});return t.map(function(t){return jf(e,t,n)})}return bf(e)}(e.useFactory,e.deps)):(t=function(){return e.useValue},n=hf);return new gf(t,n)}function _f(e){return new mf(lf.get(e.provide),[vf(e)],e.multi||!1)}function yf(e){var t=function(e,t){for(var n=0;n<e.length;n++){var r=e[n],i=t.get(r.key.id);if(i){if(r.multiProvider!==i.multiProvider)throw Error("Cannot mix multi providers and regular providers, got: "+i+" "+r);if(r.multiProvider)for(var o=0;o<r.resolvedFactories.length;o++)i.resolvedFactories.push(r.resolvedFactories[o]);else t.set(r.key.id,r)}else{var a=void 0;a=r.multiProvider?new mf(r.key,r.resolvedFactories.slice(),r.multiProvider):r,t.set(r.key.id,a)}}return t}(function e(t,n){t.forEach(function(t){if(t instanceof Fd)n.push({provide:t,useClass:t});else if(t&&"object"==typeof t&&void 0!==t.provide)n.push(t);else{if(!(t instanceof Array))throw Error("Invalid provider - only instances of Provider and Type are allowed, got: "+t);e(t,n)}});return n}(e,[]).map(_f),new Map);return Array.from(t.values())}function bf(e){var t=pf.parameters(e);if(!t)return[];if(t.some(function(e){return null==e}))throw cf(e,t);return t.map(function(n){return jf(e,n,t)})}function jf(e,t,n){var r=null,i=!1;if(!Array.isArray(t))return xf(t instanceof Nt?t.token:t,i,null);for(var o=null,a=0;a<t.length;++a){var s=t[a];s instanceof Fd?r=s:s instanceof Nt?r=s.token:s instanceof Ft?i=!0:s instanceof Lt||s instanceof Vt?o=s:s instanceof b&&(r=s)}if(null!=(r=Z(r)))return xf(r,i,o);throw cf(e,n)}function xf(e,t,n){return new ff(lf.get(e),t,n)}var Of=new Object,wf=function(){function e(){}return e.resolve=function(e){return yf(e)},e.resolveAndCreate=function(t,n){var r=e.resolve(t);return e.fromResolvedProviders(r,n)},e.fromResolvedProviders=function(e,t){return new Cf(e,t)},e}(),Cf=function(){function e(e,t){this._constructionCounter=0,this._providers=e,this.parent=t||null;var n=e.length;this.keyIds=new Array(n),this.objs=new Array(n);for(var r=0;r<n;r++)this.keyIds[r]=e[r].key.id,this.objs[r]=Of}return e.prototype.get=function(e,t){return void 0===t&&(t=Rs),this._getByKey(lf.get(e),null,t)},e.prototype.resolveAndCreateChild=function(e){var t=wf.resolve(e);return this.createChildFromResolved(t)},e.prototype.createChildFromResolved=function(t){var n=new e(t);return n.parent=this,n},e.prototype.resolveAndInstantiate=function(e){return this.instantiateResolved(wf.resolve([e])[0])},e.prototype.instantiateResolved=function(e){return this._instantiateProvider(e)},e.prototype.getProviderAtIndex=function(e){if(e<0||e>=this._providers.length)throw Error("Index "+e+" is out-of-bounds.");return this._providers[e]},e.prototype._new=function(e){if(this._constructionCounter++>this._getMaxNumberOfObjects())throw t=this,n=e.key,sf(t,n,function(e){return"Cannot instantiate cyclic dependency!"+af(e)});var t,n;return this._instantiateProvider(e)},e.prototype._getMaxNumberOfObjects=function(){return this.objs.length},e.prototype._instantiateProvider=function(e){if(e.multiProvider){for(var t=new Array(e.resolvedFactories.length),n=0;n<e.resolvedFactories.length;++n)t[n]=this._instantiate(e,e.resolvedFactories[n]);return t}return this._instantiate(e,e.resolvedFactories[0])},e.prototype._instantiate=function(e,t){var n,i,o,a,s,u=this,c=t.factory;try{n=t.dependencies.map(function(e){return u._getByReflectiveDependency(e)})}catch(t){throw t.addKey&&t.addKey(this,e.key),t}try{i=c.apply(void 0,Object(r.__spread)(n))}catch(t){throw o=this,a=t,t.stack,s=e.key,sf(o,s,function(e){var t=K(e[0].token);return a.message+": Error during instantiation of "+t+"!"+af(e)+"."},a)}return i},e.prototype._getByReflectiveDependency=function(e){return this._getByKey(e.key,e.visibility,e.optional?null:Rs)},e.prototype._getByKey=function(t,n,r){return t===e.INJECTOR_KEY?this:n instanceof Lt?this._getByKeySelf(t,r):this._getByKeyDefault(t,r,n)},e.prototype._getObjByKeyId=function(e){for(var t=0;t<this.keyIds.length;t++)if(this.keyIds[t]===e)return this.objs[t]===Of&&(this.objs[t]=this._new(this._providers[t])),this.objs[t];return Of},e.prototype._throwOrNull=function(e,t){if(t!==Rs)return t;throw sf(this,e,function(e){return"No provider for "+K(e[0].token)+"!"+af(e)})},e.prototype._getByKeySelf=function(e,t){var n=this._getObjByKeyId(e.id);return n!==Of?n:this._throwOrNull(e,t)},e.prototype._getByKeyDefault=function(t,n,r){var i;for(i=r instanceof Vt?this.parent:this;i instanceof e;){var o=i,a=o._getObjByKeyId(t.id);if(a!==Of)return a;i=o.parent}return null!==i?i.get(t.token,n):this._throwOrNull(t,n)},Object.defineProperty(e.prototype,"displayName",{get:function(){return"ReflectiveInjector(providers: ["+function(e,t){for(var n=new Array(e._providers.length),r=0;r<e._providers.length;++r)n[r]=t(e.getProviderAtIndex(r));return n}(this,function(e){return' "'+e.key.displayName+'" '}).join(", ")+"])"},enumerable:!0,configurable:!0}),e.prototype.toString=function(){return this.displayName},e.INJECTOR_KEY=lf.get(Ls),e}();function Ef(e){return!!e&&"function"==typeof e.then}function Sf(e){return!!e&&"function"==typeof e.subscribe}var kf=new b("Application Initializer"),Mf=function(){function e(e){var t=this;this.appInits=e,this.initialized=!1,this.done=!1,this.donePromise=new Promise(function(e,n){t.resolve=e,t.reject=n})}return e.prototype.runInitializers=function(){var e=this;if(!this.initialized){var t=[],n=function(){e.done=!0,e.resolve()};if(this.appInits)for(var r=0;r<this.appInits.length;r++){var i=this.appInits[r]();Ef(i)&&t.push(i)}Promise.all(t).then(function(){n()}).catch(function(t){e.reject(t)}),0===t.length&&n(),this.initialized=!0}},e=Object(r.__decorate)([Qp(),Object(r.__param)(0,Nt(kf)),Object(r.__param)(0,Ft()),Object(r.__metadata)("design:paramtypes",[Array])],e)}(),Af=new b("AppId");function Tf(){return""+If()+If()+If()}var Df={provide:Af,useFactory:Tf,deps:[]};function If(){return String.fromCharCode(97+Math.floor(25*Math.random()))}var Pf=new b("Platform Initializer"),Rf=new b("Platform ID"),Nf=new b("appBootstrapListener"),Ff=new b("Application Packages Root URL"),Lf=function(){function e(){}return e.prototype.log=function(e){console.log(e)},e.prototype.warn=function(e){console.warn(e)},e=Object(r.__decorate)([Qp()],e)}(),Vf=function(){return function(e,t){this.ngModuleFactory=e,this.componentFactories=t}}();function Bf(){throw new Error("Runtime compiler is not loaded")}var zf,Uf,Hf=function(e){return new Sl(e)},qf=Bf,Gf=function(e){return Promise.resolve(Hf(e))},Wf=Bf,Yf=function(e){return new Vf(Hf(e),[])},Kf=Bf,Qf=function(e){return Promise.resolve(Yf(e))},Xf=Bf,Zf=function(){function e(){this.compileModuleSync=qf,this.compileModuleAsync=Wf,this.compileModuleAndAllComponentsSync=Kf,this.compileModuleAndAllComponentsAsync=Xf}return e.prototype.clearCache=function(){},e.prototype.clearCacheFor=function(e){},e.prototype.getModuleId=function(e){},e=Object(r.__decorate)([Qp()],e)}(),$f=new b("compilerOptions"),Jf=function(){return function(){}}();function eh(){var e=U.wtf;return!(!e||!(zf=e.trace))&&(Uf=zf.events,!0)}function th(e,t){return void 0===t&&(t=null),Uf.createScope(e,t)}function nh(e,t){return zf.leaveScope(e,t),t}function rh(e,t){return zf.beginTimeRange(e,t)}function ih(e){zf.endTimeRange(e)}var oh=eh();function ah(e,t){return null}var sh=oh?th:function(e,t){return ah},uh=oh?nh:function(e,t){return t},ch=oh?rh:function(e,t){return null},lh=oh?ih:function(e){return null},dh=function(){function e(e){var t=e.enableLongStackTrace,n=void 0!==t&&t;if(this.hasPendingMicrotasks=!1,this.hasPendingMacrotasks=!1,this.isStable=!0,this.onUnstable=new Yl(!1),this.onMicrotaskEmpty=new Yl(!1),this.onStable=new Yl(!1),this.onError=new Yl(!1),"undefined"==typeof Zone)throw new Error("In this configuration Angular requires Zone.js");Zone.assertZonePatched();var r;this._nesting=0,this._outer=this._inner=Zone.current,Zone.wtfZoneSpec&&(this._inner=this._inner.fork(Zone.wtfZoneSpec)),Zone.TaskTrackingZoneSpec&&(this._inner=this._inner.fork(new Zone.TaskTrackingZoneSpec)),n&&Zone.longStackTraceZoneSpec&&(this._inner=this._inner.fork(Zone.longStackTraceZoneSpec)),(r=this)._inner=r._inner.fork({name:"angular",properties:{isAngularZone:!0},onInvokeTask:function(e,t,n,i,o,a){try{return mh(r),e.invokeTask(n,i,o,a)}finally{gh(r)}},onInvoke:function(e,t,n,i,o,a,s){try{return mh(r),e.invoke(n,i,o,a,s)}finally{gh(r)}},onHasTask:function(e,t,n,i){e.hasTask(n,i),t===n&&("microTask"==i.change?(r.hasPendingMicrotasks=i.microTask,hh(r)):"macroTask"==i.change&&(r.hasPendingMacrotasks=i.macroTask))},onHandleError:function(e,t,n,i){return e.handleError(n,i),r.runOutsideAngular(function(){return r.onError.emit(i)}),!1}})}return e.isInAngularZone=function(){return!0===Zone.current.get("isAngularZone")},e.assertInAngularZone=function(){if(!e.isInAngularZone())throw new Error("Expected to be in Angular Zone, but it is not!")},e.assertNotInAngularZone=function(){if(e.isInAngularZone())throw new Error("Expected to not be in Angular Zone, but it is!")},e.prototype.run=function(e,t,n){return this._inner.run(e,t,n)},e.prototype.runTask=function(e,t,n,r){var i=this._inner,o=i.scheduleEventTask("NgZoneEvent: "+r,e,fh,ph,ph);try{return i.runTask(o,t,n)}finally{i.cancelTask(o)}},e.prototype.runGuarded=function(e,t,n){return this._inner.runGuarded(e,t,n)},e.prototype.runOutsideAngular=function(e){return this._outer.run(e)},e}();function ph(){}var fh={};function hh(e){if(0==e._nesting&&!e.hasPendingMicrotasks&&!e.isStable)try{e._nesting++,e.onMicrotaskEmpty.emit(null)}finally{if(e._nesting--,!e.hasPendingMicrotasks)try{e.runOutsideAngular(function(){return e.onStable.emit(null)})}finally{e.isStable=!0}}}function mh(e){e._nesting++,e.isStable&&(e.isStable=!1,e.onUnstable.emit(null))}function gh(e){e._nesting--,hh(e)}var vh=function(){function e(){this.hasPendingMicrotasks=!1,this.hasPendingMacrotasks=!1,this.isStable=!0,this.onUnstable=new Yl,this.onMicrotaskEmpty=new Yl,this.onStable=new Yl,this.onError=new Yl}return e.prototype.run=function(e){return e()},e.prototype.runGuarded=function(e){return e()},e.prototype.runOutsideAngular=function(e){return e()},e.prototype.runTask=function(e){return e()},e}(),_h=function(){function e(e){var t=this;this._ngZone=e,this._pendingCount=0,this._isZoneStable=!0,this._didWork=!1,this._callbacks=[],this.taskTrackingZone=null,this._watchAngularEvents(),e.run(function(){t.taskTrackingZone="undefined"==typeof Zone?null:Zone.current.get("TaskTrackingZone")})}return e.prototype._watchAngularEvents=function(){var e=this;this._ngZone.onUnstable.subscribe({next:function(){e._didWork=!0,e._isZoneStable=!1}}),this._ngZone.runOutsideAngular(function(){e._ngZone.onStable.subscribe({next:function(){dh.assertNotInAngularZone(),W(function(){e._isZoneStable=!0,e._runCallbacksIfReady()})}})})},e.prototype.increasePendingRequestCount=function(){return this._pendingCount+=1,this._didWork=!0,this._pendingCount},e.prototype.decreasePendingRequestCount=function(){if(this._pendingCount-=1,this._pendingCount<0)throw new Error("pending async requests below zero");return this._runCallbacksIfReady(),this._pendingCount},e.prototype.isStable=function(){return this._isZoneStable&&0===this._pendingCount&&!this._ngZone.hasPendingMacrotasks},e.prototype._runCallbacksIfReady=function(){var e=this;if(this.isStable())W(function(){for(;0!==e._callbacks.length;){var t=e._callbacks.pop();clearTimeout(t.timeoutId),t.doneCb(e._didWork)}e._didWork=!1});else{var t=this.getPendingTasks();this._callbacks=this._callbacks.filter(function(e){return!e.updateCb||!e.updateCb(t)||(clearTimeout(e.timeoutId),!1)}),this._didWork=!0}},e.prototype.getPendingTasks=function(){return this.taskTrackingZone?this.taskTrackingZone.macroTasks.map(function(e){return{source:e.source,creationLocation:e.creationLocation,data:e.data}}):[]},e.prototype.addCallback=function(e,t,n){var r=this,i=-1;t&&t>0&&(i=setTimeout(function(){r._callbacks=r._callbacks.filter(function(e){return e.timeoutId!==i}),e(r._didWork,r.getPendingTasks())},t)),this._callbacks.push({doneCb:e,timeoutId:i,updateCb:n})},e.prototype.whenStable=function(e,t,n){if(n&&!this.taskTrackingZone)throw new Error('Task tracking zone is required when passing an update callback to whenStable(). Is "zone.js/dist/task-tracking.js" loaded?');this.addCallback(e,t,n),this._runCallbacksIfReady()},e.prototype.getPendingRequestCount=function(){return this._pendingCount},e.prototype.findProviders=function(e,t,n){return[]},e=Object(r.__decorate)([Qp(),Object(r.__metadata)("design:paramtypes",[dh])],e)}(),yh=function(){function e(){this._applications=new Map,xh.addToWindow(this)}return e.prototype.registerApplication=function(e,t){this._applications.set(e,t)},e.prototype.unregisterApplication=function(e){this._applications.delete(e)},e.prototype.unregisterAllApplications=function(){this._applications.clear()},e.prototype.getTestability=function(e){return this._applications.get(e)||null},e.prototype.getAllTestabilities=function(){return Array.from(this._applications.values())},e.prototype.getAllRootElements=function(){return Array.from(this._applications.keys())},e.prototype.findTestabilityInTree=function(e,t){return void 0===t&&(t=!0),xh.findTestabilityInTree(this,e,t)},e=Object(r.__decorate)([Qp(),Object(r.__metadata)("design:paramtypes",[])],e)}();function bh(e){xh=e}var jh,xh=new(function(){function e(){}return e.prototype.addToWindow=function(e){},e.prototype.findTestabilityInTree=function(e,t,n){return null},e}()),Oh=function(e,t,n){return e.get(Jf).createCompiler([t]).compileModuleAsync(n)};function wh(e,t,n){var r;return ngDevMode&&(void 0===r&&(r="Type passed in is not NgModuleType, it does not have 'ngModuleDef' property."),ye(n)||Ee(r)),Promise.resolve(new Sl(n))}var Ch=new b("AllowMultipleToken"),Eh=function(){return function(e,t){this.name=e,this.token=t}}();function Sh(e){if(jh&&!jh.destroyed&&!jh.injector.get(Ch,!1))throw new Error("There can be only one platform. Destroy the previous one to create a new one.");jh=e.get(Dh);var t=e.get(Pf,null);return t&&t.forEach(function(e){return e()}),jh}function kh(e,t,n){void 0===n&&(n=[]);var r="Platform: "+t,i=new b(r);return function(t){void 0===t&&(t=[]);var o=Th();if(!o||o.injector.get(Ch,!1))if(e)e(n.concat(t).concat({provide:i,useValue:!0}));else{var a=n.concat(t).concat({provide:i,useValue:!0});Sh(Ls.create({providers:a,name:r}))}return Mh(i)}}function Mh(e){var t=Th();if(!t)throw new Error("No platform exists!");if(!t.injector.get(e,null))throw new Error("A platform with a different configuration has been created. Please destroy it first.");return t}function Ah(){jh&&!jh.destroyed&&jh.destroy()}function Th(){return jh&&!jh.destroyed?jh:null}var Dh=function(){function e(e){this._injector=e,this._modules=[],this._destroyListeners=[],this._destroyed=!1}return e.prototype.bootstrapModuleFactory=function(e,t){var n=this,r=function(e){var t;t="noop"===e?new vh:("zone.js"===e?void 0:e)||new dh({enableLongStackTrace:fc()});return t}(t?t.ngZone:void 0),i=[{provide:dh,useValue:r}];return r.run(function(){var t=Ls.create({providers:i,parent:n.injector,name:e.moduleType.name}),o=e.create(t),a=o.injector.get(of,null);if(!a)throw new Error("No ErrorHandler. Is platform module (BrowserModule) included?");return o.onDestroy(function(){return Rh(n._modules,o)}),r.runOutsideAngular(function(){return r.onError.subscribe({next:function(e){a.handleError(e)}})}),function(e,t,n){try{var r=n();return Ef(r)?r.catch(function(n){throw t.runOutsideAngular(function(){return e.handleError(n)}),n}):r}catch(n){throw t.runOutsideAngular(function(){return e.handleError(n)}),n}}(a,r,function(){var e=o.injector.get(Mf);return e.runInitializers(),e.donePromise.then(function(){return n._moduleDoBootstrap(o),o})})})},e.prototype.bootstrapModule=function(e,t){var n=this;void 0===t&&(t=[]);var r=Ih({},t);return Oh(this.injector,r,e).then(function(e){return n.bootstrapModuleFactory(e,r)})},e.prototype._moduleDoBootstrap=function(e){var t=e.injector.get(Ph);if(e._bootstrapComponents.length>0)e._bootstrapComponents.forEach(function(e){return t.bootstrap(e)});else{if(!e.instance.ngDoBootstrap)throw new Error("The module "+K(e.instance.constructor)+' was bootstrapped, but it does not declare "@NgModule.bootstrap" components nor a "ngDoBootstrap" method. Please define one of these.');e.instance.ngDoBootstrap(t)}this._modules.push(e)},e.prototype.onDestroy=function(e){this._destroyListeners.push(e)},Object.defineProperty(e.prototype,"injector",{get:function(){return this._injector},enumerable:!0,configurable:!0}),e.prototype.destroy=function(){if(this._destroyed)throw new Error("The platform has already been destroyed!");this._modules.slice().forEach(function(e){return e.destroy()}),this._destroyListeners.forEach(function(e){return e()}),this._destroyed=!0},Object.defineProperty(e.prototype,"destroyed",{get:function(){return this._destroyed},enumerable:!0,configurable:!0}),e=Object(r.__decorate)([Qp(),Object(r.__metadata)("design:paramtypes",[Ls])],e)}();function Ih(e,t){return e=Array.isArray(t)?t.reduce(Ih,e):Object(r.__assign)({},e,t)}var Ph=function(){function e(e,t,n,r,a,s){var u=this;this._zone=e,this._console=t,this._injector=n,this._exceptionHandler=r,this._componentFactoryResolver=a,this._initStatus=s,this._bootstrapListeners=[],this._views=[],this._runningTick=!1,this._enforceNoNewChanges=!1,this._stable=!0,this.componentTypes=[],this.components=[],this._enforceNoNewChanges=fc(),this._zone.onMicrotaskEmpty.subscribe({next:function(){u._zone.run(function(){u.tick()})}});var c=new i.Observable(function(e){u._stable=u._zone.isStable&&!u._zone.hasPendingMacrotasks&&!u._zone.hasPendingMicrotasks,u._zone.runOutsideAngular(function(){e.next(u._stable),e.complete()})}),l=new i.Observable(function(e){var t;u._zone.runOutsideAngular(function(){t=u._zone.onStable.subscribe(function(){dh.assertNotInAngularZone(),W(function(){u._stable||u._zone.hasPendingMacrotasks||u._zone.hasPendingMicrotasks||(u._stable=!0,e.next(!0))})})});var n=u._zone.onUnstable.subscribe(function(){dh.assertInAngularZone(),u._stable&&(u._stable=!1,u._zone.runOutsideAngular(function(){e.next(!1)}))});return function(){t.unsubscribe(),n.unsubscribe()}});this.isStable=Object(i.merge)(c,l.pipe(Object(o.share)()))}var t;return t=e,e.prototype.bootstrap=function(e,t){var n,r=this;if(!this._initStatus.done)throw new Error("Cannot bootstrap as there are still asynchronous initializers running. Bootstrap components in the `ngDoBootstrap` method of the root module.");n=e instanceof Ou?e:this._componentFactoryResolver.resolveComponentFactory(e),this.componentTypes.push(n.componentType);var i=n instanceof Du?null:this._injector.get(Iu),o=t||n.selector,a=n.create(Ls.NULL,[],o,i);a.onDestroy(function(){r._unloadComponent(a)});var s=a.injector.get(_h,null);return s&&a.injector.get(yh).registerApplication(a.location.nativeElement,s),this._loadComponent(a),fc()&&this._console.log("Angular is running in the development mode. Call enableProdMode() to enable the production mode."),a},e.prototype.tick=function(){var e=this;if(this._runningTick)throw new Error("ApplicationRef.tick is called recursively");var n=t._tickScope();try{this._runningTick=!0,this._views.forEach(function(e){return e.detectChanges()}),this._enforceNoNewChanges&&this._views.forEach(function(e){return e.checkNoChanges()})}catch(t){this._zone.runOutsideAngular(function(){return e._exceptionHandler.handleError(t)})}finally{this._runningTick=!1,uh(n)}},e.prototype.attachView=function(e){var t=e;this._views.push(t),t.attachToAppRef(this)},e.prototype.detachView=function(e){var t=e;Rh(this._views,t),t.detachFromAppRef()},e.prototype._loadComponent=function(e){this.attachView(e.hostView),this.tick(),this.components.push(e),this._injector.get(Nf,[]).concat(this._bootstrapListeners).forEach(function(t){return t(e)})},e.prototype._unloadComponent=function(e){this.detachView(e.hostView),Rh(this.components,e)},e.prototype.ngOnDestroy=function(){this._views.slice().forEach(function(e){return e.destroy()})},Object.defineProperty(e.prototype,"viewCount",{get:function(){return this._views.length},enumerable:!0,configurable:!0}),e._tickScope=sh("ApplicationRef#tick()"),e=t=Object(r.__decorate)([Qp(),Object(r.__metadata)("design:paramtypes",[dh,Lf,Ls,of,Au,Mf])],e)}();function Rh(e,t){var n=e.indexOf(t);n>-1&&e.splice(n,1)}var Nh=function(){function e(){this.dirty=!0,this._results=[],this.changes=new Yl,this.length=0}return e.prototype.map=function(e){return this._results.map(e)},e.prototype.filter=function(e){return this._results.filter(e)},e.prototype.find=function(e){return this._results.find(e)},e.prototype.reduce=function(e,t){return this._results.reduce(e,t)},e.prototype.forEach=function(e){this._results.forEach(e)},e.prototype.some=function(e){return this._results.some(e)},e.prototype.toArray=function(){return this._results.slice()},e.prototype[G()]=function(){return this._results[G()]()},e.prototype.toString=function(){return this._results.toString()},e.prototype.reset=function(e){this._results=function e(t){return t.reduce(function(t,n){var r=Array.isArray(n)?e(n):n;return t.concat(r)},[])}(e),this.dirty=!1,this.length=this._results.length,this.last=this._results[this.length-1],this.first=this._results[0]},e.prototype.notifyOnChanges=function(){this.changes.emit(this)},e.prototype.setDirty=function(){this.dirty=!0},e.prototype.destroy=function(){this.changes.complete(),this.changes.unsubscribe()},e}();var Fh=function(){return function(){}}(),Lh={factoryPathPrefix:"",factoryPathSuffix:".ngfactory"},Vh=function(){function e(e,t){this._compiler=e,this._config=t||Lh}return e.prototype.load=function(e){return this._compiler instanceof Zf?this.loadFactory(e):this.loadAndCompile(e)},e.prototype.loadAndCompile=function(e){var t=this,i=Object(r.__read)(e.split("#"),2),o=i[0],a=i[1];return void 0===a&&(a="default"),n("./src/$$_lazy_route_resource lazy recursive")(o).then(function(e){return e[a]}).then(function(e){return Bh(e,o,a)}).then(function(e){return t._compiler.compileModuleAsync(e)})},e.prototype.loadFactory=function(e){var t=Object(r.__read)(e.split("#"),2),i=t[0],o=t[1],a="NgFactory";return void 0===o&&(o="default",a=""),n("./src/$$_lazy_route_resource lazy recursive")(this._config.factoryPathPrefix+i+this._config.factoryPathSuffix).then(function(e){return e[o+a]}).then(function(e){return Bh(e,i,o)})},e=Object(r.__decorate)([Qp(),Object(r.__param)(1,Ft()),Object(r.__metadata)("design:paramtypes",[Zf,Fh])],e)}();function Bh(e,t,n){if(!e)throw new Error("Cannot find '"+n+"' in '"+t+"'");return e}var zh=function(){function e(){}return e.__NG_ELEMENT_ID__=function(){return Hh(e,Uu)},e}(),Uh=function(e,t){return function(e,t,n,i){var o;Su||(Su=function(e){function n(t,n,r){var i=e.call(this)||this;return i._lContainer=t,i._hostTNode=n,i._hostView=r,i._viewRefs=[],i}return Object(r.__extends)(n,e),Object.defineProperty(n.prototype,"element",{get:function(){return Fu(t,this._hostTNode,this._hostView)},enumerable:!0,configurable:!0}),Object.defineProperty(n.prototype,"injector",{get:function(){return new er(this._hostTNode,this._hostView)},enumerable:!0,configurable:!0}),Object.defineProperty(n.prototype,"parentInjector",{get:function(){var e=qn(this._hostTNode,this._hostView),t=St(e,this._hostView),n=function(e,t,n){if(n.parent&&-1!==n.parent.injectorIndex){for(var r=n.parent.injectorIndex,i=n.parent;null!=i.parent&&r==i.injectorIndex;)i=i.parent;return i}for(var o=Et(e),a=t,s=t[Fe];o>1;)s=(a=a[Ke])[Fe],o--;return s}(e,this._hostView,this._hostTNode);return wt(e)&&null!=n?new er(n,t):new er(null,this._hostView)},enumerable:!0,configurable:!0}),n.prototype.clear=function(){for(;this._lContainer[Ze].length;)this.remove(0)},n.prototype.get=function(e){return this._viewRefs[e]||null},Object.defineProperty(n.prototype,"length",{get:function(){return this._lContainer[Ze].length},enumerable:!0,configurable:!0}),n.prototype.createEmbeddedView=function(e,t,n){var r=this._adjustIndex(n),i=e.createEmbeddedView(t||{},this._lContainer,this._hostTNode,this._hostView,r);return i.attachToViewContainerRef(this),this._viewRefs.splice(r,0,i),i},n.prototype.createComponent=function(e,t,n,r,i){var o=n||this.parentInjector;!i&&null==e.ngModule&&o&&(i=o.get(Iu,null));var a=e.create(o,r,void 0,i);return this.insert(a.hostView,t),a},n.prototype.insert=function(e,t){if(e.destroyed)throw new Error("Cannot insert a destroyed View in a ViewContainer!");var n=e._lView,r=this._adjustIndex(t);ti(n,this._lContainer,this._hostView,r,this._hostTNode.index);var i=hi(r,this._lContainer[Ze],this._lContainer[$e]);return ei(n,!0,i),e.attachToViewContainerRef(this),this._viewRefs.splice(r,0,e),e},n.prototype.move=function(e,t){if(e.destroyed)throw new Error("Cannot move a destroyed View in a ViewContainer!");var n=this.indexOf(e);return this.detach(n),this.insert(e,this._adjustIndex(t)),e},n.prototype.indexOf=function(e){return this._viewRefs.indexOf(e)},n.prototype.remove=function(e){var t=this._adjustIndex(e,-1);ri(this._lContainer,this._hostTNode,t),this._viewRefs.splice(t,1)},n.prototype.detach=function(e){var t=this._adjustIndex(e,-1),n=ni(this._lContainer,t,!!this._hostTNode.detached),r=null!=this._viewRefs.splice(t,1)[0];return r?new Ru(n,n[Be],n[We]):null},n.prototype._adjustIndex=function(e,t){return void 0===t&&(t=0),null==e?this._lContainer[Ze].length+t:(ngDevMode&&(Oe(e,-1,"index must be positive"),xe(e,this._lContainer[Ze].length+1+t,"index")),e)},n}(e)),ngDevMode&&Zt(n,0,3,4);var a,s,u=i[n.index];if(_t(u))(o=u)[Xe]=-1;else{var c=i[He].createComment(ngDevMode?"container":"");if(ngDevMode&&ngDevMode.rendererCreateComment++,yt(i)){var l=i[He],d=pt(n,i),p=di(l,d);li(l,p,c,(s=d,qr(a=l)?a.nextSibling(s):s.nextSibling))}else pi(c,n,i);i[n.index]=o=wa(u,n,i,c,!0),Na(i,n.index,o)}return new Su(o,n,i)}(e,t,_n(),gn())},Hh=Ds,qh=function(){function e(){}return e.__NG_ELEMENT_ID__=function(){return Wh()},e}(),Gh=function(){return Vu(_n(),gn(),null)},Wh=function(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t]},Yh=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t}(qh),Kh=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t}(Yh),Qh=function(){return function(e,t){this.name=e,this.callback=t}}(),Xh=function(){function e(e,t,n){this.listeners=[],this.parent=null,this._debugContext=n,this.nativeNode=e,t&&t instanceof Zh&&t.addChild(this)}return Object.defineProperty(e.prototype,"injector",{get:function(){return this._debugContext.injector},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"componentInstance",{get:function(){return this._debugContext.component},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"context",{get:function(){return this._debugContext.context},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"references",{get:function(){return this._debugContext.references},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"providerTokens",{get:function(){return this._debugContext.providerTokens},enumerable:!0,configurable:!0}),e}(),Zh=function(e){function t(t,n,r){var i=e.call(this,t,n,r)||this;return i.properties={},i.attributes={},i.classes={},i.styles={},i.childNodes=[],i.nativeElement=t,i}return Object(r.__extends)(t,e),t.prototype.addChild=function(e){e&&(this.childNodes.push(e),e.parent=this)},t.prototype.removeChild=function(e){var t=this.childNodes.indexOf(e);-1!==t&&(e.parent=null,this.childNodes.splice(t,1))},t.prototype.insertChildrenAfter=function(e,t){var n,i=this,o=this.childNodes.indexOf(e);-1!==o&&((n=this.childNodes).splice.apply(n,Object(r.__spread)([o+1,0],t)),t.forEach(function(t){t.parent&&t.parent.removeChild(t),e.parent=i}))},t.prototype.insertBefore=function(e,t){var n=this.childNodes.indexOf(e);-1===n?this.addChild(t):(t.parent&&t.parent.removeChild(t),t.parent=this,this.childNodes.splice(n,0,t))},t.prototype.query=function(e){return this.queryAll(e)[0]||null},t.prototype.queryAll=function(e){var t=[];return function e(t,n,r){t.childNodes.forEach(function(t){t instanceof Zh&&(n(t)&&r.push(t),e(t,n,r))})}(this,e,t),t},t.prototype.queryAllNodes=function(e){var t=[];return function e(t,n,r){t instanceof Zh&&t.childNodes.forEach(function(t){n(t)&&r.push(t),t instanceof Zh&&e(t,n,r)})}(this,e,t),t},Object.defineProperty(t.prototype,"children",{get:function(){return this.childNodes.filter(function(e){return e instanceof t})},enumerable:!0,configurable:!0}),t.prototype.triggerEventHandler=function(e,t){this.listeners.forEach(function(n){n.name==e&&n.callback(t)})},t}(Xh);function $h(e){return e.map(function(e){return e.nativeElement})}var Jh=function(){function e(e){this.nativeNode=e}return Object.defineProperty(e.prototype,"parent",{get:function(){var e=this.nativeNode.parentNode;return e?new em(e):null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"injector",{get:function(){return gr(this.nativeNode)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"componentInstance",{get:function(){var e=this.nativeNode;return e&&dr(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"context",{get:function(){return pr(this.nativeNode)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"listeners",{get:function(){return Or(this.nativeNode).filter(xr)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"references",{get:function(){return yr(this.nativeNode)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"providerTokens",{get:function(){return function(e){var t=_r(e,!1);if(!t)return[];for(var n,r=t.lView[Te],i=r.data[t.nodeIndex],o=[],a=65535&i.providerIndexes,s=i.directiveEnd,u=a;u<s;u++){var c=r.data[u];void 0!==(n=c).type&&void 0!==n.template&&void 0!==n.declaredInputs&&(c=c.type),o.push(c)}return o}(this.nativeNode)},enumerable:!0,configurable:!0}),e}(),em=function(e){function t(t){return ngDevMode&&Se(t),e.call(this,t)||this}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"nativeElement",{get:function(){return this.nativeNode.nodeType==Node.ELEMENT_NODE?this.nativeNode:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"name",{get:function(){return this.nativeElement.nodeName},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"properties",{get:function(){var e=_r(this.nativeNode);e.lView[Te].data[e.nodeIndex];return{}},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"attributes",{get:function(){var e={},t=this.nativeElement;if(t)for(var n=t.attributes,r=0;r<n.length;r++){var i=n[r];e[i.name]=i.value}return e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"classes",{get:function(){var e={},t=this.nativeElement;if(t){var n=jr(t),r=n.lView[n.nodeIndex];if(ki(n.nodeIndex,n.lView)){for(var i=9;i<r.length;i+=4)if(Wi(r,i)){var o=uo(r,i),a=so(r,i);"boolean"==typeof a&&(e[o]=a)}}else{var s=t.classList;for(i=0;i<s.length;i++)e[s[i]]=!0}}return e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"styles",{get:function(){var e={},t=this.nativeElement;if(t){var n=jr(t),r=n.lView[n.nodeIndex];if(ki(n.nodeIndex,n.lView)){for(var i=9;i<r.length;i+=4)if(!Wi(r,i)){var o=uo(r,i),a=so(r,i);null!==a&&(e[o]=a)}}else{var s=t.style;for(i=0;i<s.length;i++){var u=s.item(i);e[u]=s.getPropertyValue(u)}}}return e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"childNodes",{get:function(){for(var e=this.nativeNode.childNodes,t=[],n=0;n<e.length;n++){var r=e[n];t.push(rm(r))}return t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"children",{get:function(){var e=this.nativeElement;if(!e)return[];for(var t=e.children,n=[],r=0;r<t.length;r++){var i=t[r];n.push(rm(i))}return n},enumerable:!0,configurable:!0}),t.prototype.query=function(e){return this.queryAll(e)[0]||null},t.prototype.queryAll=function(e){var t=[];return tm(this,e,t,!0),t},t.prototype.queryAllNodes=function(e){var t=[];return tm(this,e,t,!1),t},t.prototype.triggerEventHandler=function(e,t){this.listeners.forEach(function(n){n.name===e&&n.callback(t)})},t}(Jh);function tm(e,t,n,r){e instanceof em&&e.childNodes.forEach(function(e){t(e)&&n.push(e),e instanceof em&&(r&&!e.nativeElement||tm(e,t,n,r))})}var nm=new Map;function rm(e){return e instanceof Node?e.nodeType==Node.ELEMENT_NODE?new em(e):new Jh(e):null}var im=function(e){return nm.get(e)||null};function om(e){nm.set(e.nativeNode,e)}var am=Xh,sm=Zh,um=function(){function e(){}return e.prototype.supports=function(e){return Tr(e)},e.prototype.create=function(e){return new lm(e)},e}(),cm=function(e,t){return t},lm=function(){function e(e){this.length=0,this._linkedRecords=null,this._unlinkedRecords=null,this._previousItHead=null,this._itHead=null,this._itTail=null,this._additionsHead=null,this._additionsTail=null,this._movesHead=null,this._movesTail=null,this._removalsHead=null,this._removalsTail=null,this._identityChangesHead=null,this._identityChangesTail=null,this._trackByFn=e||cm}return e.prototype.forEachItem=function(e){var t;for(t=this._itHead;null!==t;t=t._next)e(t)},e.prototype.forEachOperation=function(e){for(var t=this._itHead,n=this._removalsHead,r=0,i=null;t||n;){var o=!n||t&&t.currentIndex<hm(n,r,i)?t:n,a=hm(o,r,i),s=o.currentIndex;if(o===n)r--,n=n._nextRemoved;else if(t=t._next,null==o.previousIndex)r++;else{i||(i=[]);var u=a-r,c=s-r;if(u!=c){for(var l=0;l<u;l++){var d=l<i.length?i[l]:i[l]=0,p=d+l;c<=p&&p<u&&(i[l]=d+1)}i[o.previousIndex]=c-u}}a!==s&&e(o,a,s)}},e.prototype.forEachPreviousItem=function(e){var t;for(t=this._previousItHead;null!==t;t=t._nextPrevious)e(t)},e.prototype.forEachAddedItem=function(e){var t;for(t=this._additionsHead;null!==t;t=t._nextAdded)e(t)},e.prototype.forEachMovedItem=function(e){var t;for(t=this._movesHead;null!==t;t=t._nextMoved)e(t)},e.prototype.forEachRemovedItem=function(e){var t;for(t=this._removalsHead;null!==t;t=t._nextRemoved)e(t)},e.prototype.forEachIdentityChange=function(e){var t;for(t=this._identityChangesHead;null!==t;t=t._nextIdentityChange)e(t)},e.prototype.diff=function(e){if(null==e&&(e=[]),!Tr(e))throw new Error("Error trying to diff '"+K(e)+"'. Only arrays and iterables are allowed");return this.check(e)?this:null},e.prototype.onDestroy=function(){},e.prototype.check=function(e){var t=this;this._reset();var n,r,i,o=this._itHead,a=!1;if(Array.isArray(e)){this.length=e.length;for(var s=0;s<this.length;s++)r=e[s],i=this._trackByFn(s,r),null!==o&&Y(o.trackById,i)?(a&&(o=this._verifyReinsertion(o,r,i,s)),Y(o.item,r)||this._addIdentityChange(o,r)):(o=this._mismatch(o,r,i,s),a=!0),o=o._next}else n=0,function(e,t){if(Array.isArray(e))for(var n=0;n<e.length;n++)t(e[n]);else for(var r=e[G()](),i=void 0;!(i=r.next()).done;)t(i.value)}(e,function(e){i=t._trackByFn(n,e),null!==o&&Y(o.trackById,i)?(a&&(o=t._verifyReinsertion(o,e,i,n)),Y(o.item,e)||t._addIdentityChange(o,e)):(o=t._mismatch(o,e,i,n),a=!0),o=o._next,n++}),this.length=n;return this._truncate(o),this.collection=e,this.isDirty},Object.defineProperty(e.prototype,"isDirty",{get:function(){return null!==this._additionsHead||null!==this._movesHead||null!==this._removalsHead||null!==this._identityChangesHead},enumerable:!0,configurable:!0}),e.prototype._reset=function(){if(this.isDirty){var e=void 0,t=void 0;for(e=this._previousItHead=this._itHead;null!==e;e=e._next)e._nextPrevious=e._next;for(e=this._additionsHead;null!==e;e=e._nextAdded)e.previousIndex=e.currentIndex;for(this._additionsHead=this._additionsTail=null,e=this._movesHead;null!==e;e=t)e.previousIndex=e.currentIndex,t=e._nextMoved;this._movesHead=this._movesTail=null,this._removalsHead=this._removalsTail=null,this._identityChangesHead=this._identityChangesTail=null}},e.prototype._mismatch=function(e,t,n,r){var i;return null===e?i=this._itTail:(i=e._prev,this._remove(e)),null!==(e=null===this._linkedRecords?null:this._linkedRecords.get(n,r))?(Y(e.item,t)||this._addIdentityChange(e,t),this._moveAfter(e,i,r)):null!==(e=null===this._unlinkedRecords?null:this._unlinkedRecords.get(n,null))?(Y(e.item,t)||this._addIdentityChange(e,t),this._reinsertAfter(e,i,r)):e=this._addAfter(new dm(t,n),i,r),e},e.prototype._verifyReinsertion=function(e,t,n,r){var i=null===this._unlinkedRecords?null:this._unlinkedRecords.get(n,null);return null!==i?e=this._reinsertAfter(i,e._prev,r):e.currentIndex!=r&&(e.currentIndex=r,this._addToMoves(e,r)),e},e.prototype._truncate=function(e){for(;null!==e;){var t=e._next;this._addToRemovals(this._unlink(e)),e=t}null!==this._unlinkedRecords&&this._unlinkedRecords.clear(),null!==this._additionsTail&&(this._additionsTail._nextAdded=null),null!==this._movesTail&&(this._movesTail._nextMoved=null),null!==this._itTail&&(this._itTail._next=null),null!==this._removalsTail&&(this._removalsTail._nextRemoved=null),null!==this._identityChangesTail&&(this._identityChangesTail._nextIdentityChange=null)},e.prototype._reinsertAfter=function(e,t,n){null!==this._unlinkedRecords&&this._unlinkedRecords.remove(e);var r=e._prevRemoved,i=e._nextRemoved;return null===r?this._removalsHead=i:r._nextRemoved=i,null===i?this._removalsTail=r:i._prevRemoved=r,this._insertAfter(e,t,n),this._addToMoves(e,n),e},e.prototype._moveAfter=function(e,t,n){return this._unlink(e),this._insertAfter(e,t,n),this._addToMoves(e,n),e},e.prototype._addAfter=function(e,t,n){return this._insertAfter(e,t,n),null===this._additionsTail?this._additionsTail=this._additionsHead=e:this._additionsTail=this._additionsTail._nextAdded=e,e},e.prototype._insertAfter=function(e,t,n){var r=null===t?this._itHead:t._next;return e._next=r,e._prev=t,null===r?this._itTail=e:r._prev=e,null===t?this._itHead=e:t._next=e,null===this._linkedRecords&&(this._linkedRecords=new fm),this._linkedRecords.put(e),e.currentIndex=n,e},e.prototype._remove=function(e){return this._addToRemovals(this._unlink(e))},e.prototype._unlink=function(e){null!==this._linkedRecords&&this._linkedRecords.remove(e);var t=e._prev,n=e._next;return null===t?this._itHead=n:t._next=n,null===n?this._itTail=t:n._prev=t,e},e.prototype._addToMoves=function(e,t){return e.previousIndex===t?e:(null===this._movesTail?this._movesTail=this._movesHead=e:this._movesTail=this._movesTail._nextMoved=e,e)},e.prototype._addToRemovals=function(e){return null===this._unlinkedRecords&&(this._unlinkedRecords=new fm),this._unlinkedRecords.put(e),e.currentIndex=null,e._nextRemoved=null,null===this._removalsTail?(this._removalsTail=this._removalsHead=e,e._prevRemoved=null):(e._prevRemoved=this._removalsTail,this._removalsTail=this._removalsTail._nextRemoved=e),e},e.prototype._addIdentityChange=function(e,t){return e.item=t,null===this._identityChangesTail?this._identityChangesTail=this._identityChangesHead=e:this._identityChangesTail=this._identityChangesTail._nextIdentityChange=e,e},e}(),dm=function(){return function(e,t){this.item=e,this.trackById=t,this.currentIndex=null,this.previousIndex=null,this._nextPrevious=null,this._prev=null,this._next=null,this._prevDup=null,this._nextDup=null,this._prevRemoved=null,this._nextRemoved=null,this._nextAdded=null,this._nextMoved=null,this._nextIdentityChange=null}}(),pm=function(){function e(){this._head=null,this._tail=null}return e.prototype.add=function(e){null===this._head?(this._head=this._tail=e,e._nextDup=null,e._prevDup=null):(this._tail._nextDup=e,e._prevDup=this._tail,e._nextDup=null,this._tail=e)},e.prototype.get=function(e,t){var n;for(n=this._head;null!==n;n=n._nextDup)if((null===t||t<=n.currentIndex)&&Y(n.trackById,e))return n;return null},e.prototype.remove=function(e){var t=e._prevDup,n=e._nextDup;return null===t?this._head=n:t._nextDup=n,null===n?this._tail=t:n._prevDup=t,null===this._head},e}(),fm=function(){function e(){this.map=new Map}return e.prototype.put=function(e){var t=e.trackById,n=this.map.get(t);n||(n=new pm,this.map.set(t,n)),n.add(e)},e.prototype.get=function(e,t){var n=e,r=this.map.get(n);return r?r.get(e,t):null},e.prototype.remove=function(e){var t=e.trackById;return this.map.get(t).remove(e)&&this.map.delete(t),e},Object.defineProperty(e.prototype,"isEmpty",{get:function(){return 0===this.map.size},enumerable:!0,configurable:!0}),e.prototype.clear=function(){this.map.clear()},e}();function hm(e,t,n){var r=e.previousIndex;if(null===r)return r;var i=0;return n&&r<n.length&&(i=n[r]),r+t+i}var mm=function(){function e(){}return e.prototype.supports=function(e){return e instanceof Map||Dr(e)},e.prototype.create=function(){return new gm},e}(),gm=function(){function e(){this._records=new Map,this._mapHead=null,this._appendAfter=null,this._previousMapHead=null,this._changesHead=null,this._changesTail=null,this._additionsHead=null,this._additionsTail=null,this._removalsHead=null,this._removalsTail=null}return Object.defineProperty(e.prototype,"isDirty",{get:function(){return null!==this._additionsHead||null!==this._changesHead||null!==this._removalsHead},enumerable:!0,configurable:!0}),e.prototype.forEachItem=function(e){var t;for(t=this._mapHead;null!==t;t=t._next)e(t)},e.prototype.forEachPreviousItem=function(e){var t;for(t=this._previousMapHead;null!==t;t=t._nextPrevious)e(t)},e.prototype.forEachChangedItem=function(e){var t;for(t=this._changesHead;null!==t;t=t._nextChanged)e(t)},e.prototype.forEachAddedItem=function(e){var t;for(t=this._additionsHead;null!==t;t=t._nextAdded)e(t)},e.prototype.forEachRemovedItem=function(e){var t;for(t=this._removalsHead;null!==t;t=t._nextRemoved)e(t)},e.prototype.diff=function(e){if(e){if(!(e instanceof Map||Dr(e)))throw new Error("Error trying to diff '"+K(e)+"'. Only maps and objects are allowed")}else e=new Map;return this.check(e)?this:null},e.prototype.onDestroy=function(){},e.prototype.check=function(e){var t=this;this._reset();var n=this._mapHead;if(this._appendAfter=null,this._forEach(e,function(e,r){if(n&&n.key===r)t._maybeAddToChanges(n,e),t._appendAfter=n,n=n._next;else{var i=t._getOrCreateRecordForKey(r,e);n=t._insertBeforeOrAppend(n,i)}}),n){n._prev&&(n._prev._next=null),this._removalsHead=n;for(var r=n;null!==r;r=r._nextRemoved)r===this._mapHead&&(this._mapHead=null),this._records.delete(r.key),r._nextRemoved=r._next,r.previousValue=r.currentValue,r.currentValue=null,r._prev=null,r._next=null}return this._changesTail&&(this._changesTail._nextChanged=null),this._additionsTail&&(this._additionsTail._nextAdded=null),this.isDirty},e.prototype._insertBeforeOrAppend=function(e,t){if(e){var n=e._prev;return t._next=e,t._prev=n,e._prev=t,n&&(n._next=t),e===this._mapHead&&(this._mapHead=t),this._appendAfter=e,e}return this._appendAfter?(this._appendAfter._next=t,t._prev=this._appendAfter):this._mapHead=t,this._appendAfter=t,null},e.prototype._getOrCreateRecordForKey=function(e,t){if(this._records.has(e)){var n=this._records.get(e);this._maybeAddToChanges(n,t);var r=n._prev,i=n._next;return r&&(r._next=i),i&&(i._prev=r),n._next=null,n._prev=null,n}var o=new vm(e);return this._records.set(e,o),o.currentValue=t,this._addToAdditions(o),o},e.prototype._reset=function(){if(this.isDirty){var e=void 0;for(this._previousMapHead=this._mapHead,e=this._previousMapHead;null!==e;e=e._next)e._nextPrevious=e._next;for(e=this._changesHead;null!==e;e=e._nextChanged)e.previousValue=e.currentValue;for(e=this._additionsHead;null!=e;e=e._nextAdded)e.previousValue=e.currentValue;this._changesHead=this._changesTail=null,this._additionsHead=this._additionsTail=null,this._removalsHead=null}},e.prototype._maybeAddToChanges=function(e,t){Y(t,e.currentValue)||(e.previousValue=e.currentValue,e.currentValue=t,this._addToChanges(e))},e.prototype._addToAdditions=function(e){null===this._additionsHead?this._additionsHead=this._additionsTail=e:(this._additionsTail._nextAdded=e,this._additionsTail=e)},e.prototype._addToChanges=function(e){null===this._changesHead?this._changesHead=this._changesTail=e:(this._changesTail._nextChanged=e,this._changesTail=e)},e.prototype._forEach=function(e,t){e instanceof Map?e.forEach(t):Object.keys(e).forEach(function(n){return t(e[n],n)})},e}(),vm=function(){return function(e){this.key=e,this.previousValue=null,this.currentValue=null,this._nextPrevious=null,this._next=null,this._prev=null,this._nextAdded=null,this._nextRemoved=null,this._nextChanged=null}}(),_m=function(){function e(e){this.factories=e}return e.create=function(t,n){if(null!=n){var r=n.factories.slice();t=t.concat(r)}return new e(t)},e.extend=function(t){return{provide:e,useFactory:function(n){if(!n)throw new Error("Cannot extend IterableDiffers without a parent injector");return e.create(t,n)},deps:[[e,new Vt,new Ft]]}},e.prototype.find=function(e){var t,n=this.factories.find(function(t){return t.supports(e)});if(null!=n)return n;throw new Error("Cannot find a differ supporting object '"+e+"' of type '"+((t=e).name||typeof t)+"'")},e.ngInjectableDef=g({providedIn:"root",factory:function(){return new e([new um])}}),e}();var ym,bm,jm=function(){function e(e){this.factories=e}return e.create=function(t,n){if(n){var r=n.factories.slice();t=t.concat(r)}return new e(t)},e.extend=function(t){return{provide:e,useFactory:function(n){if(!n)throw new Error("Cannot extend KeyValueDiffers without a parent injector");return e.create(t,n)},deps:[[e,new Vt,new Ft]]}},e.prototype.find=function(e){var t=this.factories.find(function(t){return t.supports(e)});if(t)return t;throw new Error("Cannot find a differ supporting object '"+e+"'")},e.ngInjectableDef=g({providedIn:"root",factory:function(){return new e([new mm])}}),e}(),xm=[new mm],Om=[new um],wm=new _m(Om),Cm=new jm(xm),Em=kh(null,"core",[{provide:Rf,useValue:"unknown"},{provide:Dh,deps:[Ls]},{provide:yh,deps:[]},{provide:Lf,deps:[]}]),Sm=new b("LocaleId"),km=new b("Translations"),Mm=new b("TranslationsFormat");function Am(){return wm}function Tm(){return Cm}function Dm(e){return e||"en-US"}(bm=ym||(ym={}))[bm.Error=0]="Error",bm[bm.Warning=1]="Warning",bm[bm.Ignore=2]="Ignore";var Im=[{provide:Ph,useClass:Ph,deps:[dh,Lf,Ls,of,Au,Mf]},{provide:Mf,useClass:Mf,deps:[[new Ft,kf]]},{provide:Zf,useClass:Zf,deps:[]},Df,{provide:_m,useFactory:Am,deps:[]},{provide:jm,useFactory:Tm,deps:[]},{provide:Sm,useFactory:Dm,deps:[[new Nt(Sm),new Ft,new Vt]]}],Pm=function(){function e(e){}return e=Object(r.__decorate)([zp({providers:Im}),Object(r.__metadata)("design:paramtypes",[Ph])],e)}(),Rm=!0,Nm=!1;function Fm(e,t,n){var r=e.state,i=1792&r;return i===t?(e.state=-1793&r|n,e.initIndex=-1,!0):i===n}function Lm(e,t,n){return(1792&e.state)===t&&e.initIndex<=n&&(e.initIndex=n+1,!0)}function Vm(e,t){return e.nodes[t]}function Bm(e,t){return e.nodes[t]}function zm(e,t){return e.nodes[t]}function Um(e,t){return e.nodes[t]}function Hm(e,t){return e.nodes[t]}var qm=function(){return function(){}}(),Gm={setCurrentNode:void 0,createRootView:void 0,createEmbeddedView:void 0,createComponentView:void 0,createNgModuleRef:void 0,overrideProvider:void 0,overrideComponentView:void 0,clearOverrides:void 0,checkAndUpdateView:void 0,checkNoChangesView:void 0,destroyView:void 0,resolveDep:void 0,createDebugContext:void 0,handleEvent:void 0,updateDirectives:void 0,updateRenderer:void 0,dirtyParentQueries:void 0};function Wm(e,t,n,r){var i,o,a="ExpressionChangedAfterItHasBeenCheckedError: Expression has changed after it was checked. Previous value: '"+t+"'. Current value: '"+n+"'.";return r&&(a+=" It seems like the view has been created after its parent and its children have been dirty checked. Has it been created in a change detection hook ?"),i=e,Ym(o=new Error(a),i),o}function Ym(e,t){e[$p]=t,e[ef]=t.logError.bind(t)}function Km(e){return new Error("ViewDestroyedError: Attempt to use a destroyed view: "+e)}var Qm=function(){},Xm=new Map;function Zm(e){var t=Xm.get(e);return t||(t=K(e)+"_"+Xm.size,Xm.set(e,t)),t}function $m(e,t,n,r){if(Mr.isWrapped(r)){r=Mr.unwrap(r);var i=e.def.nodes[t].bindingIndex+n,o=Mr.unwrap(e.oldValues[i]);e.oldValues[i]=new Mr(o)}return r}var Jm="$$undefined",eg="$$empty";function tg(e){return{id:Jm,styles:e.styles,encapsulation:e.encapsulation,data:e.data}}var ng=0;function rg(e,t,n,r){var i=e.oldValues;return!(!(2&e.state)&&Y(i[t.bindingIndex+n],r))}function ig(e,t,n,r){return!!rg(e,t,n,r)&&(e.oldValues[t.bindingIndex+n]=r,!0)}function og(e,t,n,r){var i=e.oldValues[t.bindingIndex+n];if(1&e.state||!kr(i,r)){var o=t.bindings[n].name;throw Wm(Gm.createDebugContext(e,t.nodeIndex),o+": "+i,o+": "+r,0!=(1&e.state))}}function ag(e){for(var t=e;t;)2&t.def.flags&&(t.state|=8),t=t.viewContainerParent||t.parent}function sg(e,t){for(var n=e;n&&n!==t;)n.state|=64,n=n.viewContainerParent||n.parent}function ug(e,t,n,r){try{return ag(33554432&e.def.nodes[t].flags?Bm(e,t).componentView:e),Gm.handleEvent(e,t,n,r)}catch(t){e.root.errorHandler.handleError(t)}}function cg(e){return e.parent?Bm(e.parent,e.parentNodeDef.nodeIndex):null}function lg(e){return e.parent?e.parentNodeDef.parent:null}function dg(e,t){switch(201347067&t.flags){case 1:return Bm(e,t.nodeIndex).renderElement;case 2:return Vm(e,t.nodeIndex).renderText}}function pg(e,t){return e?e+":"+t:t}function fg(e){return!!e.parent&&!!(32768&e.parentNodeDef.flags)}function hg(e){return 1<<e%32}function mg(e){var t={},n=0,i={};return e&&e.forEach(function(e){var o=Object(r.__read)(e,2),a=o[0],s=o[1];"number"==typeof a?(t[a]=s,n|=hg(a)):i[a]=s}),{matchedQueries:t,references:i,matchedQueryIds:n}}function gg(e,t){return e.map(function(e){var n,i,o;return Array.isArray(e)?(o=(n=Object(r.__read)(e,2))[0],i=n[1]):(o=0,i=e),i&&("function"==typeof i||"object"==typeof i)&&t&&Object.defineProperty(i,Is,{value:t,configurable:!0}),{flags:o,token:i,tokenKey:Zm(i)}})}function vg(e,t,n){var r=n.renderParent;return r?0==(1&r.flags)||0==(33554432&r.flags)||r.element.componentRendererType&&r.element.componentRendererType.encapsulation===J.Native?Bm(e,n.renderParent.nodeIndex).renderElement:void 0:t}var _g=new WeakMap;function yg(e){var t=_g.get(e);return t||((t=e(function(){return Qm})).factory=e,_g.set(e,t)),t}function bg(e,t,n,r,i){3===t&&(n=e.renderer.parentNode(dg(e,e.def.lastRenderRootNode))),jg(e,t,0,e.def.nodes.length-1,n,r,i)}function jg(e,t,n,r,i,o,a){for(var s=n;s<=r;s++){var u=e.def.nodes[s];11&u.flags&&Og(e,u,t,i,o,a),s+=u.childCount}}function xg(e,t,n,r,i,o){for(var a=e;a&&!fg(a);)a=a.parent;for(var s=a.parent,u=lg(a),c=u.nodeIndex+1,l=u.nodeIndex+u.childCount,d=c;d<=l;d++){var p=s.def.nodes[d];p.ngContentIndex===t&&Og(s,p,n,r,i,o),d+=p.childCount}if(!s.parent){var f=e.root.projectableNodes[t];if(f)for(d=0;d<f.length;d++)wg(e,f[d],n,r,i,o)}}function Og(e,t,n,r,i,o){if(8&t.flags)xg(e,t.ngContent.index,n,r,i,o);else{var a=dg(e,t);if(3===n&&33554432&t.flags&&48&t.bindingFlags){if(16&t.bindingFlags&&wg(e,a,n,r,i,o),32&t.bindingFlags)wg(Bm(e,t.nodeIndex).componentView,a,n,r,i,o)}else wg(e,a,n,r,i,o);if(16777216&t.flags)for(var s=Bm(e,t.nodeIndex).viewContainer._embeddedViews,u=0;u<s.length;u++)bg(s[u],n,r,i,o);1&t.flags&&!t.element.name&&jg(e,n,t.nodeIndex+1,t.nodeIndex+t.childCount,r,i,o)}}function wg(e,t,n,r,i,o){var a=e.renderer;switch(n){case 1:a.appendChild(r,t);break;case 2:a.insertBefore(r,t,i);break;case 3:a.removeChild(r,t);break;case 0:o.push(t)}}var Cg=/^:([^:]+):(.+)$/;function Eg(e){if(":"===e[0]){var t=e.match(Cg);return[t[1],t[2]]}return["",e]}function Sg(e){for(var t=0,n=0;n<e.length;n++)t|=e[n].flags;return t}function kg(e,t){for(var n="",r=0;r<2*e;r+=2)n=n+t[r]+Ag(t[r+1]);return n+t[2*e]}function Mg(e,t,n,r,i,o,a,s,u,c,l,d,p,f,h,m,g,v,_,y){switch(e){case 1:return t+Ag(n)+r;case 2:return t+Ag(n)+r+Ag(i)+o;case 3:return t+Ag(n)+r+Ag(i)+o+Ag(a)+s;case 4:return t+Ag(n)+r+Ag(i)+o+Ag(a)+s+Ag(u)+c;case 5:return t+Ag(n)+r+Ag(i)+o+Ag(a)+s+Ag(u)+c+Ag(l)+d;case 6:return t+Ag(n)+r+Ag(i)+o+Ag(a)+s+Ag(u)+c+Ag(l)+d+Ag(p)+f;case 7:return t+Ag(n)+r+Ag(i)+o+Ag(a)+s+Ag(u)+c+Ag(l)+d+Ag(p)+f+Ag(h)+m;case 8:return t+Ag(n)+r+Ag(i)+o+Ag(a)+s+Ag(u)+c+Ag(l)+d+Ag(p)+f+Ag(h)+m+Ag(g)+v;case 9:return t+Ag(n)+r+Ag(i)+o+Ag(a)+s+Ag(u)+c+Ag(l)+d+Ag(p)+f+Ag(h)+m+Ag(g)+v+Ag(_)+y;default:throw new Error("Does not support more than 9 expressions")}}function Ag(e){return null!=e?e.toString():""}var Tg=[],Dg={};function Ig(e,t,n,r,i,o){e|=1;var a=mg(t),s=a.matchedQueries,u=a.references;return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,flags:e,checkIndex:-1,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:s,matchedQueryIds:a.matchedQueryIds,references:u,ngContentIndex:n,childCount:r,bindings:[],bindingFlags:0,outputs:[],element:{ns:null,name:null,attrs:null,template:o?yg(o):null,componentProvider:null,componentView:null,componentRendererType:null,publicProviders:null,allProviders:null,handleEvent:i||Qm},provider:null,text:null,query:null,ngContent:null}}function Pg(e,t,n,i,o,a,s,u,c,l,d,p){var f;void 0===s&&(s=[]),l||(l=Qm);var h=mg(n),m=h.matchedQueries,g=h.references,v=h.matchedQueryIds,_=null,y=null;a&&(_=(f=Object(r.__read)(Eg(a),2))[0],y=f[1]),u=u||[];for(var b=new Array(u.length),j=0;j<u.length;j++){var x=Object(r.__read)(u[j],3),O=x[0],w=x[1],C=x[2],E=Object(r.__read)(Eg(w),2),S=E[0],k=E[1],M=void 0,A=void 0;switch(15&O){case 4:A=C;break;case 1:case 8:M=C}b[j]={flags:O,ns:S,name:k,nonMinifiedName:k,securityContext:M,suffix:A}}c=c||[];var T=new Array(c.length);for(j=0;j<c.length;j++){var D=Object(r.__read)(c[j],2),I=D[0],P=D[1];T[j]={type:0,target:I,eventName:P,propName:null}}var R=(s=s||[]).map(function(e){var t=Object(r.__read)(e,2),n=t[0],i=t[1],o=Object(r.__read)(Eg(n),2);return[o[0],o[1],i]});return p=function(e){if(e&&e.id===Jm){var t=null!=e.encapsulation&&e.encapsulation!==J.None||e.styles.length||Object.keys(e.data).length;e.id=t?"c"+ng++:eg}return e&&e.id===eg&&(e=null),e||null}(p),d&&(t|=33554432),{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:e,flags:t|=1,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:m,matchedQueryIds:v,references:g,ngContentIndex:i,childCount:o,bindings:b,bindingFlags:Sg(b),outputs:T,element:{ns:_,name:y,attrs:R,template:null,componentProvider:null,componentView:d||null,componentRendererType:p,publicProviders:null,allProviders:null,handleEvent:l||Qm},provider:null,text:null,query:null,ngContent:null}}function Rg(e,t,n){var i,o=n.element,a=e.root.selectorOrNode,s=e.renderer;if(e.parent||!a){i=o.name?s.createElement(o.name,o.ns):s.createComment("");var u=vg(e,t,n);u&&s.appendChild(u,i)}else{var c=!!o.componentRendererType&&o.componentRendererType.encapsulation===J.ShadowDom;i=s.selectRootElement(a,c)}if(o.attrs)for(var l=0;l<o.attrs.length;l++){var d=Object(r.__read)(o.attrs[l],3),p=d[0],f=d[1],h=d[2];s.setAttribute(i,f,h,p)}return i}function Ng(e,t,n,r){for(var i=0;i<n.outputs.length;i++){var o=n.outputs[i],a=Fg(e,n.nodeIndex,pg(o.target,o.eventName)),s=o.target,u=e;"component"===o.target&&(s=null,u=t);var c=u.renderer.listen(s||r,o.eventName,a);e.disposables[n.outputIndex+i]=c}}function Fg(e,t,n){return function(r){return ug(e,t,n,r)}}function Lg(e,t,n,r){if(!ig(e,t,n,r))return!1;var i,o,a,s,u,c,l,d,p,f,h=t.bindings[n],m=Bm(e,t.nodeIndex),g=m.renderElement,v=h.name;switch(15&h.flags){case 1:!function(e,t,n,r,i,o){var a=t.securityContext,s=a?e.root.sanitizer.sanitize(a,o):o;s=null!=s?s.toString():null;var u=e.renderer;null!=o?u.setAttribute(n,i,s,r):u.removeAttribute(n,i,r)}(e,h,g,h.ns,v,r);break;case 2:l=g,d=v,p=r,f=e.renderer,p?f.addClass(l,d):f.removeClass(l,d);break;case 4:!function(e,t,n,r,i){var o=e.root.sanitizer.sanitize(Xu.STYLE,i);if(null!=o){o=o.toString();var a=t.suffix;null!=a&&(o+=a)}else o=null;var s=e.renderer;null!=o?s.setStyle(n,r,o):s.removeStyle(n,r)}(e,h,g,v,r);break;case 8:var _=33554432&t.flags&&32&h.flags?m.componentView:e;i=_,o=g,a=v,s=r,u=h.securityContext,c=u?i.root.sanitizer.sanitize(u,s):s,i.renderer.setProperty(o,a,c)}return!0}var Vg=new Object,Bg=Zm(Ls),zg=Zm(Ns),Ug=Zm(Iu);function Hg(e,t,n,r){return n=Z(n),{index:-1,deps:gg(r,K(t)),flags:e,token:t,value:n}}function qg(e){for(var t={},n=[],r=!1,i=0;i<e.length;i++){var o=e[i];o.token===tu&&!0===o.value&&(r=!0),1073741824&o.flags&&n.push(o.token),o.index=i,t[Zm(o.token)]=o}return{factory:null,providersByKey:t,providers:e,modules:n,isRoot:r}}function Gg(e,t,n){void 0===n&&(n=Ls.THROW_IF_NOT_FOUND);var r,i,o,a,s=qt(e);try{if(8&t.flags)return t.token;if(2&t.flags&&(n=null),1&t.flags)return e._parent.get(t.token,n);var u=t.tokenKey;switch(u){case Bg:case zg:case Ug:return e}var c=e._def.providersByKey[u],l=void 0;if(c){var d=e._providers[c.index];return void 0===d&&(d=e._providers[c.index]=Wg(e,c)),d===Vg?void 0:d}if((l=_(t.token))&&(r=e,null!=(i=l).providedIn&&(o=r,a=i.providedIn,o._def.modules.indexOf(a)>-1||"root"===i.providedIn&&r._def.isRoot))){var p=e._providers.length;return e._def.providersByKey[t.tokenKey]={flags:5120,value:l.factory,deps:[],index:p,token:t.token},e._providers[p]=Vg,e._providers[p]=Wg(e,e._def.providersByKey[t.tokenKey])}return 4&t.flags?n:e._parent.get(t.token,n)}finally{qt(s)}}function Wg(e,t){var n;switch(201347067&t.flags){case 512:n=function(e,t,n){var i=n.length;switch(i){case 0:return new t;case 1:return new t(Gg(e,n[0]));case 2:return new t(Gg(e,n[0]),Gg(e,n[1]));case 3:return new t(Gg(e,n[0]),Gg(e,n[1]),Gg(e,n[2]));default:for(var o=new Array(i),a=0;a<i;a++)o[a]=Gg(e,n[a]);return new(t.bind.apply(t,Object(r.__spread)([void 0],o)))}}(e,t.value,t.deps);break;case 1024:n=function(e,t,n){var i=n.length;switch(i){case 0:return t();case 1:return t(Gg(e,n[0]));case 2:return t(Gg(e,n[0]),Gg(e,n[1]));case 3:return t(Gg(e,n[0]),Gg(e,n[1]),Gg(e,n[2]));default:for(var o=Array(i),a=0;a<i;a++)o[a]=Gg(e,n[a]);return t.apply(void 0,Object(r.__spread)(o))}}(e,t.value,t.deps);break;case 2048:n=Gg(e,t.deps[0]);break;case 256:n=t.value}return n===Vg||null==n||"object"!=typeof n||131072&t.flags||"function"!=typeof n.ngOnDestroy||(t.flags|=131072),void 0===n?Vg:n}function Yg(e,t,n,r){var i=t.viewContainer._embeddedViews;null!==n&&void 0!==n||(n=i.length),r.viewContainerParent=e,Zg(i,n,r),function(e,t){var n=cg(t);if(!n||n===e||16&t.state)return;t.state|=16;var r=n.template._projectedViews;r||(r=n.template._projectedViews=[]);r.push(t),function(e,t){if(4&t.flags)return;e.nodeFlags|=4,t.flags|=4;var n=t.parent;for(;n;)n.childFlags|=4,n=n.parent}(t.parent.def,t.parentNodeDef)}(t,r),Gm.dirtyParentQueries(r),Qg(t,n>0?i[n-1]:null,r)}function Kg(e,t){var n=e.viewContainer._embeddedViews;if((null==t||t>=n.length)&&(t=n.length-1),t<0)return null;var r=n[t];return r.viewContainerParent=null,$g(n,t),Gm.dirtyParentQueries(r),Xg(r),r}function Qg(e,t,n){var r=t?dg(t,t.def.lastRenderRootNode):e.renderElement;bg(n,2,n.renderer.parentNode(r),n.renderer.nextSibling(r),void 0)}function Xg(e){bg(e,3,null,null,void 0)}function Zg(e,t,n){t>=e.length?e.push(n):e.splice(t,0,n)}function $g(e,t){t>=e.length-1?e.pop():e.splice(t,1)}var Jg=new Object;function ev(e,t,n,r,i,o){return new nv(e,t,n,r,i,o)}function tv(e){return e.viewDefFactory}var nv=function(e){function t(t,n,r,i,o,a){var s=e.call(this)||this;return s.selector=t,s.componentType=n,s._inputs=i,s._outputs=o,s.ngContentSelectors=a,s.viewDefFactory=r,s}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"inputs",{get:function(){var e=[],t=this._inputs;for(var n in t){var r=t[n];e.push({propName:n,templateName:r})}return e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"outputs",{get:function(){var e=[];for(var t in this._outputs){var n=this._outputs[t];e.push({propName:t,templateName:n})}return e},enumerable:!0,configurable:!0}),t.prototype.create=function(e,t,n,r){if(!r)throw new Error("ngModule should be provided");var i=yg(this.viewDefFactory),o=i.nodes[0].element.componentProvider.nodeIndex,a=Gm.createRootView(e,t||[],n,i,r,Jg),s=zm(a,o).instance;return n&&a.renderer.setAttribute(Bm(a,0).renderElement,"ng-version",rc.full),new rv(a,new ov(a),s)},t}(Ou),rv=function(e){function t(t,n,r){var i=e.call(this)||this;return i._view=t,i._viewRef=n,i._component=r,i._elDef=i._view.def.nodes[0],i.hostView=n,i.changeDetectorRef=n,i.instance=r,i}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"location",{get:function(){return new Uu(Bm(this._view,this._elDef.nodeIndex).renderElement)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"injector",{get:function(){return new uv(this._view,this._elDef)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"componentType",{get:function(){return this._component.constructor},enumerable:!0,configurable:!0}),t.prototype.destroy=function(){this._viewRef.destroy()},t.prototype.onDestroy=function(e){this._viewRef.onDestroy(e)},t}(xu);var iv=function(){function e(e,t,n){this._view=e,this._elDef=t,this._data=n,this._embeddedViews=[]}return Object.defineProperty(e.prototype,"element",{get:function(){return new Uu(this._data.renderElement)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"injector",{get:function(){return new uv(this._view,this._elDef)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"parentInjector",{get:function(){for(var e=this._view,t=this._elDef.parent;!t&&e;)t=lg(e),e=e.parent;return e?new uv(e,t):new uv(this._view,null)},enumerable:!0,configurable:!0}),e.prototype.clear=function(){for(var e=this._embeddedViews.length-1;e>=0;e--){var t=Kg(this._data,e);Gm.destroyView(t)}},e.prototype.get=function(e){var t=this._embeddedViews[e];if(t){var n=new ov(t);return n.attachToViewContainerRef(this),n}return null},Object.defineProperty(e.prototype,"length",{get:function(){return this._embeddedViews.length},enumerable:!0,configurable:!0}),e.prototype.createEmbeddedView=function(e,t,n){var r=e.createEmbeddedView(t||{});return this.insert(r,n),r},e.prototype.createComponent=function(e,t,n,r,i){var o=n||this.parentInjector;i||e instanceof Du||(i=o.get(Iu));var a=e.create(o,r,void 0,i);return this.insert(a.hostView,t),a},e.prototype.insert=function(e,t){if(e.destroyed)throw new Error("Cannot insert a destroyed View in a ViewContainer!");var n=e,r=n._view;return Yg(this._view,this._data,t,r),n.attachToViewContainerRef(this),e},e.prototype.move=function(e,t){if(e.destroyed)throw new Error("Cannot move a destroyed View in a ViewContainer!");var n,r,i,o,a,s=this._embeddedViews.indexOf(e._view);return n=this._data,r=s,i=t,o=n.viewContainer._embeddedViews,a=o[r],$g(o,r),null==i&&(i=o.length),Zg(o,i,a),Gm.dirtyParentQueries(a),Xg(a),Qg(n,i>0?o[i-1]:null,a),e},e.prototype.indexOf=function(e){return this._embeddedViews.indexOf(e._view)},e.prototype.remove=function(e){var t=Kg(this._data,e);t&&Gm.destroyView(t)},e.prototype.detach=function(e){var t=Kg(this._data,e);return t?new ov(t):null},e}();var ov=function(){function e(e){this._view=e,this._viewContainerRef=null,this._appRef=null}return Object.defineProperty(e.prototype,"rootNodes",{get:function(){return bg(this._view,0,void 0,void 0,e=[]),e;var e},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"context",{get:function(){return this._view.context},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"destroyed",{get:function(){return 0!=(128&this._view.state)},enumerable:!0,configurable:!0}),e.prototype.markForCheck=function(){ag(this._view)},e.prototype.detach=function(){this._view.state&=-5},e.prototype.detectChanges=function(){var e=this._view.root.rendererFactory;e.begin&&e.begin();try{Gm.checkAndUpdateView(this._view)}finally{e.end&&e.end()}},e.prototype.checkNoChanges=function(){Gm.checkNoChangesView(this._view)},e.prototype.reattach=function(){this._view.state|=4},e.prototype.onDestroy=function(e){this._view.disposables||(this._view.disposables=[]),this._view.disposables.push(e)},e.prototype.destroy=function(){this._appRef?this._appRef.detachView(this):this._viewContainerRef&&this._viewContainerRef.detach(this._viewContainerRef.indexOf(this)),Gm.destroyView(this._view)},e.prototype.detachFromAppRef=function(){this._appRef=null,Xg(this._view),Gm.dirtyParentQueries(this._view)},e.prototype.attachToAppRef=function(e){if(this._viewContainerRef)throw new Error("This view is already attached to a ViewContainer!");this._appRef=e},e.prototype.attachToViewContainerRef=function(e){if(this._appRef)throw new Error("This view is already attached directly to the ApplicationRef!");this._viewContainerRef=e},e}();var av=function(e){function t(t,n){var r=e.call(this)||this;return r._parentView=t,r._def=n,r}return Object(r.__extends)(t,e),t.prototype.createEmbeddedView=function(e){return new ov(Gm.createEmbeddedView(this._parentView,this._def,this._def.element.template,e))},Object.defineProperty(t.prototype,"elementRef",{get:function(){return new Uu(Bm(this._parentView,this._def.nodeIndex).renderElement)},enumerable:!0,configurable:!0}),t}(Kl);function sv(e,t){return new uv(e,t)}var uv=function(){function e(e,t){this.view=e,this.elDef=t}return e.prototype.get=function(e,t){void 0===t&&(t=Ls.THROW_IF_NOT_FOUND);var n=!!this.elDef&&0!=(33554432&this.elDef.flags);return Gm.resolveDep(this.view,this.elDef,n,{flags:0,token:e,tokenKey:Zm(e)},t)},e}();function cv(e,t){var n=e.def.nodes[t];if(1&n.flags){var r=Bm(e,n.nodeIndex);return n.element.template?r.template:r.renderElement}if(2&n.flags)return Vm(e,n.nodeIndex).renderText;if(20240&n.flags)return zm(e,n.nodeIndex).instance;throw new Error("Illegal state: read nodeValue for node index "+t)}var lv=function(){function e(e){this.delegate=e}return e.prototype.selectRootElement=function(e){return this.delegate.selectRootElement(e)},e.prototype.createElement=function(e,t){var n=Object(r.__read)(Eg(t),2),i=n[0],o=n[1],a=this.delegate.createElement(o,i);return e&&this.delegate.appendChild(e,a),a},e.prototype.createViewRoot=function(e){return e},e.prototype.createTemplateAnchor=function(e){var t=this.delegate.createComment("");return e&&this.delegate.appendChild(e,t),t},e.prototype.createText=function(e,t){var n=this.delegate.createText(t);return e&&this.delegate.appendChild(e,n),n},e.prototype.projectNodes=function(e,t){for(var n=0;n<t.length;n++)this.delegate.appendChild(e,t[n])},e.prototype.attachViewAfter=function(e,t){for(var n=this.delegate.parentNode(e),r=this.delegate.nextSibling(e),i=0;i<t.length;i++)this.delegate.insertBefore(n,t[i],r)},e.prototype.detachView=function(e){for(var t=0;t<e.length;t++){var n=e[t],r=this.delegate.parentNode(n);this.delegate.removeChild(r,n)}},e.prototype.destroyView=function(e,t){for(var n=0;n<t.length;n++)this.delegate.destroyNode(t[n])},e.prototype.listen=function(e,t,n){return this.delegate.listen(e,t,n)},e.prototype.listenGlobal=function(e,t,n){return this.delegate.listen(e,t,n)},e.prototype.setElementProperty=function(e,t,n){this.delegate.setProperty(e,t,n)},e.prototype.setElementAttribute=function(e,t,n){var i=Object(r.__read)(Eg(t),2),o=i[0],a=i[1];null!=n?this.delegate.setAttribute(e,a,n,o):this.delegate.removeAttribute(e,a,o)},e.prototype.setBindingDebugInfo=function(e,t,n){},e.prototype.setElementClass=function(e,t,n){n?this.delegate.addClass(e,t):this.delegate.removeClass(e,t)},e.prototype.setElementStyle=function(e,t,n){null!=n?this.delegate.setStyle(e,t,n):this.delegate.removeStyle(e,t)},e.prototype.invokeElementMethod=function(e,t,n){e[t].apply(e,n)},e.prototype.setText=function(e,t){this.delegate.setValue(e,t)},e.prototype.animate=function(){throw new Error("Renderer.animate is no longer supported!")},e}();function dv(e,t,n,r){return new pv(e,t,n,r)}var pv=function(){function e(e,t,n,r){this._moduleType=e,this._parent=t,this._bootstrapComponents=n,this._def=r,this._destroyListeners=[],this._destroyed=!1,this.injector=this,function(e){for(var t=e._def,n=e._providers=new Array(t.providers.length),r=0;r<t.providers.length;r++){var i=t.providers[r];4096&i.flags||void 0===n[r]&&(n[r]=Wg(e,i))}}(this)}return e.prototype.get=function(e,t,n){void 0===t&&(t=Ls.THROW_IF_NOT_FOUND),void 0===n&&(n=Pt.Default);var r=0;return n&Pt.SkipSelf?r|=1:n&Pt.Self&&(r|=4),Gg(this,{token:e,tokenKey:Zm(e),flags:r},t)},Object.defineProperty(e.prototype,"instance",{get:function(){return this.get(this._moduleType)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"componentFactoryResolver",{get:function(){return this.get(Au)},enumerable:!0,configurable:!0}),e.prototype.destroy=function(){if(this._destroyed)throw new Error("The ng module "+K(this.instance.constructor)+" has already been destroyed.");this._destroyed=!0,function(e,t){for(var n=e._def,r=new Set,i=0;i<n.providers.length;i++)if(131072&n.providers[i].flags){var o=e._providers[i];if(o&&o!==Vg){var a=o.ngOnDestroy;"function"!=typeof a||r.has(o)||(a.apply(o),r.add(o))}}}(this),this._destroyListeners.forEach(function(e){return e()})},e.prototype.onDestroy=function(e){this._destroyListeners.push(e)},e}(),fv=Zm(Yu),hv=Zm($u),mv=Zm(Uu),gv=Zm(zh),vv=Zm(Kl),_v=Zm(qh),yv=Zm(Ls),bv=Zm(Ns);function jv(e,t,n,i,o,a,s,u){var c=[];if(s)for(var l in s){var d=Object(r.__read)(s[l],2),p=d[0],f=d[1];c[p]={flags:8,name:l,nonMinifiedName:f,ns:null,securityContext:null,suffix:null}}var h=[];if(u)for(var m in u)h.push({type:1,propName:m,target:null,eventName:u[m]});return wv(e,t|=16384,n,i,o,o,a,c,h)}function xv(e,t,n){return wv(-1,e|=16,null,0,t,t,n)}function Ov(e,t,n,r,i){return wv(-1,e,t,0,n,r,i)}function wv(e,t,n,r,i,o,a,s,u){var c=mg(n),l=c.matchedQueries,d=c.references,p=c.matchedQueryIds;u||(u=[]),s||(s=[]),o=Z(o);var f=gg(a,K(i));return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:e,flags:t,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:l,matchedQueryIds:p,references:d,ngContentIndex:-1,childCount:r,bindings:s,bindingFlags:Sg(s),outputs:u,element:null,provider:{token:i,value:o,deps:f},text:null,query:null,ngContent:null}}function Cv(e,t){for(var n=e;n.parent&&!fg(n);)n=n.parent;return Mv(n.parent,lg(n),!0,t.provider.value,t.provider.deps)}function Ev(e,t){var n=(32768&t.flags)>0,r=Mv(e,t.parent,n,t.provider.value,t.provider.deps);if(t.outputs.length)for(var i=0;i<t.outputs.length;i++){var o=t.outputs[i],a=r[o.propName];if(!Sf(a))throw new Error("@Output "+o.propName+" not initialized in '"+r.constructor.name+"'.");var s=a.subscribe(Sv(e,t.parent.nodeIndex,o.eventName));e.disposables[t.outputIndex+i]=s.unsubscribe.bind(s)}return r}function Sv(e,t,n){return function(r){return ug(e,t,n,r)}}function kv(e,t){var n=(8192&t.flags)>0,i=t.provider;switch(201347067&t.flags){case 512:return Mv(e,t.parent,n,i.value,i.deps);case 1024:return function(e,t,n,i,o){var a=o.length;switch(a){case 0:return i();case 1:return i(Iv(e,t,n,o[0]));case 2:return i(Iv(e,t,n,o[0]),Iv(e,t,n,o[1]));case 3:return i(Iv(e,t,n,o[0]),Iv(e,t,n,o[1]),Iv(e,t,n,o[2]));default:for(var s=Array(a),u=0;u<a;u++)s[u]=Iv(e,t,n,o[u]);return i.apply(void 0,Object(r.__spread)(s))}}(e,t.parent,n,i.value,i.deps);case 2048:return Iv(e,t.parent,n,i.deps[0]);case 256:return i.value}}function Mv(e,t,n,i,o){var a=o.length;switch(a){case 0:return new i;case 1:return new i(Iv(e,t,n,o[0]));case 2:return new i(Iv(e,t,n,o[0]),Iv(e,t,n,o[1]));case 3:return new i(Iv(e,t,n,o[0]),Iv(e,t,n,o[1]),Iv(e,t,n,o[2]));default:for(var s=new Array(a),u=0;u<a;u++)s[u]=Iv(e,t,n,o[u]);return new(i.bind.apply(i,Object(r.__spread)([void 0],s)))}}var Av,Tv,Dv={};function Iv(e,t,n,r,i){if(void 0===i&&(i=Ls.THROW_IF_NOT_FOUND),8&r.flags)return r.token;var o=e;2&r.flags&&(i=null);var a=r.tokenKey;a===_v&&(n=!(!t||!t.element.componentView)),t&&1&r.flags&&(n=!1,t=t.parent);for(var s=e;s;){if(t)switch(a){case fv:var u=Pv(s,t,n);return new lv(u.renderer);case hv:return(u=Pv(s,t,n)).renderer;case mv:return new Uu(Bm(s,t.nodeIndex).renderElement);case gv:return Bm(s,t.nodeIndex).viewContainer;case vv:if(t.element.template)return Bm(s,t.nodeIndex).template;break;case _v:var c=Pv(s,t,n);return new ov(c);case yv:case bv:return sv(s,t);default:var l=(n?t.element.allProviders:t.element.publicProviders)[a];if(l){var d=zm(s,l.nodeIndex);return d||(d={instance:kv(s,l)},s.nodes[l.nodeIndex]=d),d.instance}}n=fg(s),t=lg(s),s=s.parent,4&r.flags&&(s=null)}var p=o.root.injector.get(r.token,Dv);return p!==Dv||i===Dv?p:o.root.ngModule.injector.get(r.token,i)}function Pv(e,t,n){var r;if(n)r=Bm(e,t.nodeIndex).componentView;else for(r=e;r.parent&&!fg(r);)r=r.parent;return r}function Rv(e,t,n,r,i,o){if(32768&n.flags){var a=Bm(e,n.parent.nodeIndex).componentView;2&a.def.flags&&(a.state|=8)}var s=n.bindings[r].name;if(t.instance[s]=i,524288&n.flags){o=o||{};var u=Mr.unwrap(e.oldValues[n.bindingIndex+r]);o[n.bindings[r].nonMinifiedName]=new Ar(u,i,0!=(2&e.state))}return e.oldValues[n.bindingIndex+r]=i,o}function Nv(e,t){if(e.def.nodeFlags&t)for(var n=e.def.nodes,r=0,i=0;i<n.length;i++){var o=n[i],a=o.parent;for(!a&&o.flags&t&&Lv(e,i,o.flags&t,r++),0==(o.childFlags&t)&&(i+=o.childCount);a&&1&a.flags&&i===a.nodeIndex+a.childCount;)a.directChildFlags&t&&(r=Fv(e,a,t,r)),a=a.parent}}function Fv(e,t,n,r){for(var i=t.nodeIndex+1;i<=t.nodeIndex+t.childCount;i++){var o=e.def.nodes[i];o.flags&n&&Lv(e,i,o.flags&n,r++),i+=o.childCount}return r}function Lv(e,t,n,r){var i=zm(e,t);if(i){var o=i.instance;o&&(Gm.setCurrentNode(e,t),1048576&n&&Lm(e,512,r)&&o.ngAfterContentInit(),2097152&n&&o.ngAfterContentChecked(),4194304&n&&Lm(e,768,r)&&o.ngAfterViewInit(),8388608&n&&o.ngAfterViewChecked(),131072&n&&o.ngOnDestroy())}}function Vv(e,t,n){var r=[];for(var i in n){var o=n[i];r.push({propName:i,bindingType:o})}return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:-1,flags:e,childFlags:0,directChildFlags:0,childMatchedQueries:0,ngContentIndex:-1,matchedQueries:{},matchedQueryIds:0,references:{},childCount:0,bindings:[],bindingFlags:0,outputs:[],element:null,provider:null,text:null,query:{id:t,filterId:hg(t),bindings:r},ngContent:null}}function Bv(e){for(var t,n=e.def.nodeMatchedQueries;e.parent&&((t=e).parent&&!(32768&t.parentNodeDef.flags));){var r=e.parentNodeDef;e=e.parent;for(var i=r.nodeIndex+r.childCount,o=0;o<=i;o++){67108864&(a=e.def.nodes[o]).flags&&536870912&a.flags&&(a.query.filterId&n)===a.query.filterId&&Hm(e,o).setDirty(),!(1&a.flags&&o+a.childCount<r.nodeIndex)&&67108864&a.childFlags&&536870912&a.childFlags||(o+=a.childCount)}}if(134217728&e.def.nodeFlags)for(o=0;o<e.def.nodes.length;o++){var a;134217728&(a=e.def.nodes[o]).flags&&536870912&a.flags&&Hm(e,o).setDirty(),o+=a.childCount}}function zv(e,t){var n=Hm(e,t.nodeIndex);if(n.dirty){var r,i=void 0;if(67108864&t.flags){var o=t.parent.parent;i=Uv(e,o.nodeIndex,o.nodeIndex+o.childCount,t.query,[]),r=zm(e,t.parent.nodeIndex).instance}else 134217728&t.flags&&(i=Uv(e,0,e.def.nodes.length-1,t.query,[]),r=e.component);n.reset(i);for(var a=t.query.bindings,s=!1,u=0;u<a.length;u++){var c=a[u],l=void 0;switch(c.bindingType){case 0:l=n.first;break;case 1:l=n,s=!0}r[c.propName]=l}s&&n.notifyOnChanges()}}function Uv(e,t,n,r,i){for(var o=t;o<=n;o++){var a=e.def.nodes[o],s=a.matchedQueries[r.id];if(null!=s&&i.push(Hv(e,a,s)),1&a.flags&&a.element.template&&(a.element.template.nodeMatchedQueries&r.filterId)===r.filterId){var u=Bm(e,o);if((a.childMatchedQueries&r.filterId)===r.filterId&&(Uv(e,o+1,o+a.childCount,r,i),o+=a.childCount),16777216&a.flags)for(var c=u.viewContainer._embeddedViews,l=0;l<c.length;l++){var d=c[l],p=cg(d);p&&p===u&&Uv(d,0,d.def.nodes.length-1,r,i)}var f=u.template._projectedViews;if(f)for(l=0;l<f.length;l++){var h=f[l];Uv(h,0,h.def.nodes.length-1,r,i)}}(a.childMatchedQueries&r.filterId)!==r.filterId&&(o+=a.childCount)}return i}function Hv(e,t,n){if(null!=n)switch(n){case 1:return Bm(e,t.nodeIndex).renderElement;case 0:return new Uu(Bm(e,t.nodeIndex).renderElement);case 2:return Bm(e,t.nodeIndex).template;case 3:return Bm(e,t.nodeIndex).viewContainer;case 4:return zm(e,t.nodeIndex).instance}}function qv(e,t){return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:-1,flags:8,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:{},matchedQueryIds:0,references:{},ngContentIndex:e,childCount:0,bindings:[],bindingFlags:0,outputs:[],element:null,provider:null,text:null,query:null,ngContent:{index:t}}}function Gv(e,t){return Kv(128,e,new Array(t+1))}function Wv(e,t){return Kv(32,e,new Array(t))}function Yv(e,t){for(var n=Object.keys(t),r=n.length,i=new Array(r),o=0;o<r;o++){var a=n[o];i[t[a]]=a}return Kv(64,e,i)}function Kv(e,t,n){for(var r=new Array(n.length),i=0;i<n.length;i++){var o=n[i];r[i]={flags:8,name:o,ns:null,nonMinifiedName:o,securityContext:null,suffix:null}}return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:t,flags:e,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:{},matchedQueryIds:0,references:{},ngContentIndex:-1,childCount:0,bindings:r,bindingFlags:Sg(r),outputs:[],element:null,provider:null,text:null,query:null,ngContent:null}}function Qv(e,t,n){for(var r=new Array(n.length-1),i=1;i<n.length;i++)r[i-1]={flags:8,name:null,ns:null,nonMinifiedName:null,securityContext:null,suffix:n[i]};return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:e,flags:2,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:{},matchedQueryIds:0,references:{},ngContentIndex:t,childCount:0,bindings:r,bindingFlags:8,outputs:[],element:null,provider:null,text:{prefix:n[0]},query:null,ngContent:null}}function Xv(e,t,n){var r,i=e.renderer;r=i.createText(n.text.prefix);var o=vg(e,t,n);return o&&i.appendChild(o,r),{renderText:r}}function Zv(e,t){return(null!=e?e.toString():"")+t.suffix}function $v(e,t,n,r){for(var i=0,o=0,a=0,s=0,u=0,c=null,l=null,d=!1,p=!1,f=null,h=0;h<t.length;h++){var m=t[h];if(m.nodeIndex=h,m.parent=c,m.bindingIndex=i,m.outputIndex=o,m.renderParent=l,a|=m.flags,u|=m.matchedQueryIds,m.element){var g=m.element;g.publicProviders=c?c.element.publicProviders:Object.create(null),g.allProviders=g.publicProviders,d=!1,p=!1,m.element.template&&(u|=m.element.template.nodeMatchedQueries)}if(e_(c,m,t.length),i+=m.bindings.length,o+=m.outputs.length,!l&&3&m.flags&&(f=m),20224&m.flags){d||(d=!0,c.element.publicProviders=Object.create(c.element.publicProviders),c.element.allProviders=c.element.publicProviders);var v=0!=(8192&m.flags),_=0!=(32768&m.flags);!v||_?c.element.publicProviders[Zm(m.provider.token)]=m:(p||(p=!0,c.element.allProviders=Object.create(c.element.publicProviders)),c.element.allProviders[Zm(m.provider.token)]=m),_&&(c.element.componentProvider=m)}if(c?(c.childFlags|=m.flags,c.directChildFlags|=m.flags,c.childMatchedQueries|=m.matchedQueryIds,m.element&&m.element.template&&(c.childMatchedQueries|=m.element.template.nodeMatchedQueries)):s|=m.flags,m.childCount>0)c=m,Jv(m)||(l=m);else for(;c&&h===c.nodeIndex+c.childCount;){var y=c.parent;y&&(y.childFlags|=c.childFlags,y.childMatchedQueries|=c.childMatchedQueries),l=(c=y)&&Jv(c)?c.renderParent:c}}return{factory:null,nodeFlags:a,rootNodeFlags:s,nodeMatchedQueries:u,flags:e,nodes:t,updateDirectives:n||Qm,updateRenderer:r||Qm,handleEvent:function(e,n,r,i){return t[n].element.handleEvent(e,r,i)},bindingCount:i,outputCount:o,lastRenderRootNode:f}}function Jv(e){return 0!=(1&e.flags)&&null===e.element.name}function e_(e,t,n){var r=t.element&&t.element.template;if(r){if(!r.lastRenderRootNode)throw new Error("Illegal State: Embedded templates without nodes are not allowed!");if(r.lastRenderRootNode&&16777216&r.lastRenderRootNode.flags)throw new Error("Illegal State: Last root node of a template can't have embedded views, at index "+t.nodeIndex+"!")}if(20224&t.flags&&0==(1&(e?e.flags:0)))throw new Error("Illegal State: StaticProvider/Directive nodes need to be children of elements or anchors, at index "+t.nodeIndex+"!");if(t.query){if(67108864&t.flags&&(!e||0==(16384&e.flags)))throw new Error("Illegal State: Content Query nodes need to be children of directives, at index "+t.nodeIndex+"!");if(134217728&t.flags&&e)throw new Error("Illegal State: View Query nodes have to be top level nodes, at index "+t.nodeIndex+"!")}if(t.childCount){var i=e?e.nodeIndex+e.childCount:n-1;if(t.nodeIndex<=i&&t.nodeIndex+t.childCount>i)throw new Error("Illegal State: childCount of node leads outside of parent, at index "+t.nodeIndex+"!")}}function t_(e,t,n,r){var i=i_(e.root,e.renderer,e,t,n);return o_(i,e.component,r),a_(i),i}function n_(e,t,n){var r=i_(e,e.renderer,null,null,t);return o_(r,n,n),a_(r),r}function r_(e,t,n,r){var i,o=t.element.componentRendererType;return i=o?e.root.rendererFactory.createRenderer(r,o):e.root.renderer,i_(e.root,i,e,t.element.componentProvider,n)}function i_(e,t,n,r,i){var o=new Array(i.nodes.length),a=i.outputCount?new Array(i.outputCount):null;return{def:i,parent:n,viewContainerParent:null,parentNodeDef:r,context:null,component:null,nodes:o,state:13,root:e,renderer:t,oldValues:new Array(i.bindingCount),disposables:a,initIndex:-1}}function o_(e,t,n){e.component=t,e.context=n}function a_(e){var t;if(fg(e)){var n=e.parentNodeDef;t=Bm(e.parent,n.parent.nodeIndex).renderElement}for(var r,i,o,a,s,u=e.def,c=e.nodes,l=0;l<u.nodes.length;l++){var d=u.nodes[l];Gm.setCurrentNode(e,l);var p=void 0;switch(201347067&d.flags){case 1:var f=Rg(e,t,d),h=void 0;if(33554432&d.flags){var m=yg(d.element.componentView);h=Gm.createComponentView(e,d,m,f)}Ng(e,h,d,f),p={renderElement:f,componentView:h,viewContainer:null,template:d.element.template?(a=e,s=d,new av(a,s)):void 0},16777216&d.flags&&(p.viewContainer=new iv(e,d,p));break;case 2:p=Xv(e,t,d);break;case 512:case 1024:case 2048:case 256:if(!((p=c[l])||4096&d.flags))p={instance:kv(e,d)};break;case 16:p={instance:Cv(e,d)};break;case 16384:if(!(p=c[l]))p={instance:Ev(e,d)};if(32768&d.flags)o_(Bm(e,d.parent.nodeIndex).componentView,p.instance,p.instance);break;case 32:case 64:case 128:p={value:void 0};break;case 67108864:case 134217728:p=new Nh;break;case 8:void 0,(o=vg(r=e,t,i=d))&&xg(r,i.ngContent.index,1,o,null,void 0),p=void 0}c[l]=p}h_(e,Av.CreateViewNodes),__(e,201326592,268435456,0)}function s_(e){l_(e),Gm.updateDirectives(e,1),m_(e,Av.CheckNoChanges),Gm.updateRenderer(e,1),h_(e,Av.CheckNoChanges),e.state&=-97}function u_(e){1&e.state?(e.state&=-2,e.state|=2):e.state&=-3,Fm(e,0,256),l_(e),Gm.updateDirectives(e,0),m_(e,Av.CheckAndUpdate),__(e,67108864,536870912,0);var t=Fm(e,256,512);Nv(e,2097152|(t?1048576:0)),Gm.updateRenderer(e,0),h_(e,Av.CheckAndUpdate),__(e,134217728,536870912,0),Nv(e,8388608|((t=Fm(e,512,768))?4194304:0)),2&e.def.flags&&(e.state&=-9),e.state&=-97,Fm(e,768,1024)}function c_(e,t,n,i,o,a,s,u,c,l,d,p,f){return 0===n?function(e,t,n,r,i,o,a,s,u,c,l,d){switch(201347067&t.flags){case 1:return M=e,T=n,D=r,I=i,P=o,R=a,N=s,F=u,L=c,V=l,B=d,z=(A=t).bindings.length,U=!1,z>0&&Lg(M,A,0,T)&&(U=!0),z>1&&Lg(M,A,1,D)&&(U=!0),z>2&&Lg(M,A,2,I)&&(U=!0),z>3&&Lg(M,A,3,P)&&(U=!0),z>4&&Lg(M,A,4,R)&&(U=!0),z>5&&Lg(M,A,5,N)&&(U=!0),z>6&&Lg(M,A,6,F)&&(U=!0),z>7&&Lg(M,A,7,L)&&(U=!0),z>8&&Lg(M,A,8,V)&&(U=!0),z>9&&Lg(M,A,9,B)&&(U=!0),U;case 2:return function(e,t,n,r,i,o,a,s,u,c,l,d){var p=!1,f=t.bindings,h=f.length;if(h>0&&ig(e,t,0,n)&&(p=!0),h>1&&ig(e,t,1,r)&&(p=!0),h>2&&ig(e,t,2,i)&&(p=!0),h>3&&ig(e,t,3,o)&&(p=!0),h>4&&ig(e,t,4,a)&&(p=!0),h>5&&ig(e,t,5,s)&&(p=!0),h>6&&ig(e,t,6,u)&&(p=!0),h>7&&ig(e,t,7,c)&&(p=!0),h>8&&ig(e,t,8,l)&&(p=!0),h>9&&ig(e,t,9,d)&&(p=!0),p){var m=t.text.prefix;h>0&&(m+=Zv(n,f[0])),h>1&&(m+=Zv(r,f[1])),h>2&&(m+=Zv(i,f[2])),h>3&&(m+=Zv(o,f[3])),h>4&&(m+=Zv(a,f[4])),h>5&&(m+=Zv(s,f[5])),h>6&&(m+=Zv(u,f[6])),h>7&&(m+=Zv(c,f[7])),h>8&&(m+=Zv(l,f[8])),h>9&&(m+=Zv(d,f[9]));var g=Vm(e,t.nodeIndex).renderText;e.renderer.setValue(g,m)}return p}(e,t,n,r,i,o,a,s,u,c,l,d);case 16384:return h=n,m=r,g=i,v=o,_=a,y=s,b=u,j=c,x=l,O=d,w=zm(p=e,(f=t).nodeIndex),C=w.instance,E=!1,S=void 0,(k=f.bindings.length)>0&&rg(p,f,0,h)&&(E=!0,S=Rv(p,w,f,0,h,S)),k>1&&rg(p,f,1,m)&&(E=!0,S=Rv(p,w,f,1,m,S)),k>2&&rg(p,f,2,g)&&(E=!0,S=Rv(p,w,f,2,g,S)),k>3&&rg(p,f,3,v)&&(E=!0,S=Rv(p,w,f,3,v,S)),k>4&&rg(p,f,4,_)&&(E=!0,S=Rv(p,w,f,4,_,S)),k>5&&rg(p,f,5,y)&&(E=!0,S=Rv(p,w,f,5,y,S)),k>6&&rg(p,f,6,b)&&(E=!0,S=Rv(p,w,f,6,b,S)),k>7&&rg(p,f,7,j)&&(E=!0,S=Rv(p,w,f,7,j,S)),k>8&&rg(p,f,8,x)&&(E=!0,S=Rv(p,w,f,8,x,S)),k>9&&rg(p,f,9,O)&&(E=!0,S=Rv(p,w,f,9,O,S)),S&&C.ngOnChanges(S),65536&f.flags&&Lm(p,256,f.nodeIndex)&&C.ngOnInit(),262144&f.flags&&C.ngDoCheck(),E;case 32:case 64:case 128:return function(e,t,n,r,i,o,a,s,u,c,l,d){var p=t.bindings,f=!1,h=p.length;if(h>0&&ig(e,t,0,n)&&(f=!0),h>1&&ig(e,t,1,r)&&(f=!0),h>2&&ig(e,t,2,i)&&(f=!0),h>3&&ig(e,t,3,o)&&(f=!0),h>4&&ig(e,t,4,a)&&(f=!0),h>5&&ig(e,t,5,s)&&(f=!0),h>6&&ig(e,t,6,u)&&(f=!0),h>7&&ig(e,t,7,c)&&(f=!0),h>8&&ig(e,t,8,l)&&(f=!0),h>9&&ig(e,t,9,d)&&(f=!0),f){var m=Um(e,t.nodeIndex),g=void 0;switch(201347067&t.flags){case 32:g=new Array(p.length),h>0&&(g[0]=n),h>1&&(g[1]=r),h>2&&(g[2]=i),h>3&&(g[3]=o),h>4&&(g[4]=a),h>5&&(g[5]=s),h>6&&(g[6]=u),h>7&&(g[7]=c),h>8&&(g[8]=l),h>9&&(g[9]=d);break;case 64:g={},h>0&&(g[p[0].name]=n),h>1&&(g[p[1].name]=r),h>2&&(g[p[2].name]=i),h>3&&(g[p[3].name]=o),h>4&&(g[p[4].name]=a),h>5&&(g[p[5].name]=s),h>6&&(g[p[6].name]=u),h>7&&(g[p[7].name]=c),h>8&&(g[p[8].name]=l),h>9&&(g[p[9].name]=d);break;case 128:var v=n;switch(h){case 1:g=v.transform(n);break;case 2:g=v.transform(r);break;case 3:g=v.transform(r,i);break;case 4:g=v.transform(r,i,o);break;case 5:g=v.transform(r,i,o,a);break;case 6:g=v.transform(r,i,o,a,s);break;case 7:g=v.transform(r,i,o,a,s,u);break;case 8:g=v.transform(r,i,o,a,s,u,c);break;case 9:g=v.transform(r,i,o,a,s,u,c,l);break;case 10:g=v.transform(r,i,o,a,s,u,c,l,d)}}m.value=g}return f}(e,t,n,r,i,o,a,s,u,c,l,d);default:throw"unreachable"}var p,f,h,m,g,v,_,y,b,j,x,O,w,C,E,S,k;var M,A,T,D,I,P,R,N,F,L,V,B,z,U}(e,t,i,o,a,s,u,c,l,d,p,f):function(e,t,n){switch(201347067&t.flags){case 1:return function(e,t,n){for(var r=!1,i=0;i<n.length;i++)Lg(e,t,i,n[i])&&(r=!0);return r}(e,t,n);case 2:return function(e,t,n){for(var r=t.bindings,i=!1,o=0;o<n.length;o++)ig(e,t,o,n[o])&&(i=!0);if(i){var a="";for(o=0;o<n.length;o++)a+=Zv(n[o],r[o]);a=t.text.prefix+a;var s=Vm(e,t.nodeIndex).renderText;e.renderer.setValue(s,a)}return i}(e,t,n);case 16384:return function(e,t,n){for(var r=zm(e,t.nodeIndex),i=r.instance,o=!1,a=void 0,s=0;s<n.length;s++)rg(e,t,s,n[s])&&(o=!0,a=Rv(e,r,t,s,n[s],a));return a&&i.ngOnChanges(a),65536&t.flags&&Lm(e,256,t.nodeIndex)&&i.ngOnInit(),262144&t.flags&&i.ngDoCheck(),o}(e,t,n);case 32:case 64:case 128:return function(e,t,n){for(var i=t.bindings,o=!1,a=0;a<n.length;a++)ig(e,t,a,n[a])&&(o=!0);if(o){var s=Um(e,t.nodeIndex),u=void 0;switch(201347067&t.flags){case 32:u=n;break;case 64:for(u={},a=0;a<n.length;a++)u[i[a].name]=n[a];break;case 128:var c=n[0],l=n.slice(1);u=c.transform.apply(c,Object(r.__spread)(l))}s.value=u}return o}(e,t,n);default:throw"unreachable"}}(e,t,i)}function l_(e){var t=e.def;if(4&t.nodeFlags)for(var n=0;n<t.nodes.length;n++){var r=t.nodes[n];if(4&r.flags){var i=Bm(e,n).template._projectedViews;if(i)for(var o=0;o<i.length;o++){var a=i[o];a.state|=32,sg(a,e)}}else 0==(4&r.childFlags)&&(n+=r.childCount)}}function d_(e,t,n,r,i,o,a,s,u,c,l,d,p){return 0===n?function(e,t,n,r,i,o,a,s,u,c,l,d){var p=t.bindings.length;p>0&&og(e,t,0,n);p>1&&og(e,t,1,r);p>2&&og(e,t,2,i);p>3&&og(e,t,3,o);p>4&&og(e,t,4,a);p>5&&og(e,t,5,s);p>6&&og(e,t,6,u);p>7&&og(e,t,7,c);p>8&&og(e,t,8,l);p>9&&og(e,t,9,d)}(e,t,r,i,o,a,s,u,c,l,d,p):function(e,t,n){for(var r=0;r<n.length;r++)og(e,t,r,n[r])}(e,t,r),!1}function p_(e,t){if(Hm(e,t.nodeIndex).dirty)throw Wm(Gm.createDebugContext(e,t.nodeIndex),"Query "+t.query.id+" not dirty","Query "+t.query.id+" dirty",0!=(1&e.state))}function f_(e){if(!(128&e.state)){if(m_(e,Av.Destroy),h_(e,Av.Destroy),Nv(e,131072),e.disposables)for(var t=0;t<e.disposables.length;t++)e.disposables[t]();!function(e){if(16&e.state){var t=cg(e);if(t){var n=t.template._projectedViews;n&&($g(n,n.indexOf(e)),Gm.dirtyParentQueries(e))}}}(e),e.renderer.destroyNode&&function(e){for(var t=e.def.nodes.length,n=0;n<t;n++){var r=e.def.nodes[n];1&r.flags?e.renderer.destroyNode(Bm(e,n).renderElement):2&r.flags?e.renderer.destroyNode(Vm(e,n).renderText):(67108864&r.flags||134217728&r.flags)&&Hm(e,n).destroy()}}(e),fg(e)&&e.renderer.destroy(),e.state|=128}}function h_(e,t){var n=e.def;if(33554432&n.nodeFlags)for(var r=0;r<n.nodes.length;r++){var i=n.nodes[r];33554432&i.flags?g_(Bm(e,r).componentView,t):0==(33554432&i.childFlags)&&(r+=i.childCount)}}function m_(e,t){var n=e.def;if(16777216&n.nodeFlags)for(var r=0;r<n.nodes.length;r++){var i=n.nodes[r];if(16777216&i.flags)for(var o=Bm(e,r).viewContainer._embeddedViews,a=0;a<o.length;a++)g_(o[a],t);else 0==(16777216&i.childFlags)&&(r+=i.childCount)}}function g_(e,t){var n=e.state;switch(t){case Av.CheckNoChanges:0==(128&n)&&(12==(12&n)?s_(e):64&n&&v_(e,Av.CheckNoChangesProjectedViews));break;case Av.CheckNoChangesProjectedViews:0==(128&n)&&(32&n?s_(e):64&n&&v_(e,t));break;case Av.CheckAndUpdate:0==(128&n)&&(12==(12&n)?u_(e):64&n&&v_(e,Av.CheckAndUpdateProjectedViews));break;case Av.CheckAndUpdateProjectedViews:0==(128&n)&&(32&n?u_(e):64&n&&v_(e,t));break;case Av.Destroy:f_(e);break;case Av.CreateViewNodes:a_(e)}}function v_(e,t){m_(e,t),h_(e,t)}function __(e,t,n,r){if(e.def.nodeFlags&t&&e.def.nodeFlags&n)for(var i=e.def.nodes.length,o=0;o<i;o++){var a=e.def.nodes[o];if(a.flags&t&&a.flags&n)switch(Gm.setCurrentNode(e,a.nodeIndex),r){case 0:zv(e,a);break;case 1:p_(e,a)}a.childFlags&t&&a.childFlags&n||(o+=a.childCount)}}(Tv=Av||(Av={}))[Tv.CreateViewNodes=0]="CreateViewNodes",Tv[Tv.CheckNoChanges=1]="CheckNoChanges",Tv[Tv.CheckNoChangesProjectedViews=2]="CheckNoChangesProjectedViews",Tv[Tv.CheckAndUpdate=3]="CheckAndUpdate",Tv[Tv.CheckAndUpdateProjectedViews=4]="CheckAndUpdateProjectedViews",Tv[Tv.Destroy=5]="Destroy";var y_=!1;function b_(){if(!y_){y_=!0;var e=fc()?{setCurrentNode:q_,createRootView:x_,createEmbeddedView:w_,createComponentView:C_,createNgModuleRef:E_,overrideProvider:R_,overrideComponentView:N_,clearOverrides:F_,checkAndUpdateView:z_,checkNoChangesView:U_,destroyView:H_,createDebugContext:function(e,t){return new $_(e,t)},handleEvent:G_,updateDirectives:W_,updateRenderer:Y_}:{setCurrentNode:function(){},createRootView:j_,createEmbeddedView:t_,createComponentView:r_,createNgModuleRef:dv,overrideProvider:Qm,overrideComponentView:Qm,clearOverrides:Qm,checkAndUpdateView:u_,checkNoChangesView:s_,destroyView:f_,createDebugContext:function(e,t){return new $_(e,t)},handleEvent:function(e,t,n,r){return e.def.handleEvent(e,t,n,r)},updateDirectives:function(e,t){return e.def.updateDirectives(0===t?V_:B_,e)},updateRenderer:function(e,t){return e.def.updateRenderer(0===t?V_:B_,e)}};Gm.setCurrentNode=e.setCurrentNode,Gm.createRootView=e.createRootView,Gm.createEmbeddedView=e.createEmbeddedView,Gm.createComponentView=e.createComponentView,Gm.createNgModuleRef=e.createNgModuleRef,Gm.overrideProvider=e.overrideProvider,Gm.overrideComponentView=e.overrideComponentView,Gm.clearOverrides=e.clearOverrides,Gm.checkAndUpdateView=e.checkAndUpdateView,Gm.checkNoChangesView=e.checkNoChangesView,Gm.destroyView=e.destroyView,Gm.resolveDep=Iv,Gm.createDebugContext=e.createDebugContext,Gm.handleEvent=e.handleEvent,Gm.updateDirectives=e.updateDirectives,Gm.updateRenderer=e.updateRenderer,Gm.dirtyParentQueries=Bv}}function j_(e,t,n,r,i,o){return n_(O_(e,i,i.injector.get(Qu),t,n),r,o)}function x_(e,t,n,r,i,o){var a=i.injector.get(Qu),s=O_(e,i,new ny(a),t,n),u=L_(r);return ey(S_.create,n_,null,[s,u,o])}function O_(e,t,n,r,i){var o=t.injector.get(tc),a=t.injector.get(of);return{ngModule:t,injector:e,projectableNodes:r,selectorOrNode:i,sanitizer:o,rendererFactory:n,renderer:n.createRenderer(null,null),errorHandler:a}}function w_(e,t,n,r){var i=L_(n);return ey(S_.create,t_,null,[e,t,i,r])}function C_(e,t,n,r){var i=P_.get(t.element.componentProvider.provider.token);return n=i||L_(n),ey(S_.create,r_,null,[e,t,n,r])}function E_(e,t,n,r){return dv(e,t,n,function(e){var t=function(e){var t=!1,n=!1;if(0===D_.size)return{hasOverrides:t,hasDeprecatedOverrides:n};return e.providers.forEach(function(e){var r=D_.get(e.token);3840&e.flags&&r&&(t=!0,n=n||r.deprecatedBehavior)}),e.modules.forEach(function(e){I_.forEach(function(r,i){_(i).providedIn===e&&(t=!0,n=n||r.deprecatedBehavior)})}),{hasOverrides:t,hasDeprecatedOverrides:n}}(e),n=t.hasOverrides,r=t.hasDeprecatedOverrides;if(!n)return e;return function(e){for(var t=0;t<e.providers.length;t++){var n=e.providers[t];r&&(n.flags|=4096);var i=D_.get(n.token);i&&(n.flags=-3841&n.flags|i.flags,n.deps=gg(i.deps),n.value=i.value)}if(I_.size>0){var o=new Set(e.modules);I_.forEach(function(t,n){if(o.has(_(n).providedIn)){var i={token:n,flags:t.flags|(r?4096:0),deps:gg(t.deps),value:t.value,index:e.providers.length};e.providers.push(i),e.providersByKey[Zm(n)]=i}})}}(e=e.factory(function(){return Qm})),e}(r))}var S_,k_,M_,A_,T_,D_=new Map,I_=new Map,P_=new Map;function R_(e){var t;D_.set(e.token,e),"function"==typeof e.token&&(t=_(e.token))&&"function"==typeof t.providedIn&&I_.set(e.token,e)}function N_(e,t){var n=yg(yg(tv(t)).nodes[0].element.componentView);P_.set(e,n)}function F_(){D_.clear(),I_.clear(),P_.clear()}function L_(e){if(0===D_.size)return e;var t=function(e){for(var t=[],n=null,r=0;r<e.nodes.length;r++){var i=e.nodes[r];1&i.flags&&(n=i),n&&3840&i.flags&&D_.has(i.provider.token)&&(t.push(n.nodeIndex),n=null)}return t}(e);if(0===t.length)return e;e=e.factory(function(){return Qm});for(var n=0;n<t.length;n++)r(e,t[n]);return e;function r(e,t){for(var n=t+1;n<e.nodes.length;n++){var r=e.nodes[n];if(1&r.flags)return;if(3840&r.flags){var i=r.provider,o=D_.get(i.token);o&&(r.flags=-3841&r.flags|o.flags,i.deps=gg(o.deps),i.value=o.value)}}}}function V_(e,t,n,r,i,o,a,s,u,c,l,d,p){var f=e.def.nodes[t];return c_(e,f,n,r,i,o,a,s,u,c,l,d,p),224&f.flags?Um(e,t).value:void 0}function B_(e,t,n,r,i,o,a,s,u,c,l,d,p){var f=e.def.nodes[t];return d_(e,f,n,r,i,o,a,s,u,c,l,d,p),224&f.flags?Um(e,t).value:void 0}function z_(e){return ey(S_.detectChanges,u_,null,[e])}function U_(e){return ey(S_.checkNoChanges,s_,null,[e])}function H_(e){return ey(S_.destroy,f_,null,[e])}function q_(e,t){A_=e,T_=t}function G_(e,t,n,r){return q_(e,t),ey(S_.handleEvent,e.def.handleEvent,null,[e,t,n,r])}function W_(e,t){if(128&e.state)throw Km(S_[M_]);return q_(e,X_(e,0)),e.def.updateDirectives(function(e,n,r){for(var i=[],o=3;o<arguments.length;o++)i[o-3]=arguments[o];var a=e.def.nodes[n];0===t?K_(e,a,r,i):Q_(e,a,r,i);16384&a.flags&&q_(e,X_(e,n));return 224&a.flags?Um(e,a.nodeIndex).value:void 0},e)}function Y_(e,t){if(128&e.state)throw Km(S_[M_]);return q_(e,Z_(e,0)),e.def.updateRenderer(function(e,n,r){for(var i=[],o=3;o<arguments.length;o++)i[o-3]=arguments[o];var a=e.def.nodes[n];0===t?K_(e,a,r,i):Q_(e,a,r,i);3&a.flags&&q_(e,Z_(e,n));return 224&a.flags?Um(e,a.nodeIndex).value:void 0},e)}function K_(e,t,n,i){if(c_.apply(void 0,Object(r.__spread)([e,t,n],i))){var o=1===n?i[0]:i;if(16384&t.flags){for(var a={},s=0;s<t.bindings.length;s++){var u=t.bindings[s],c=o[s];8&u.flags&&(a[Cr(u.nonMinifiedName)]=Sr(c))}var l=t.parent,d=Bm(e,l.nodeIndex).renderElement;if(l.element.name)for(var p in a){null!=(c=a[p])?e.renderer.setAttribute(d,p,c):e.renderer.removeAttribute(d,p)}else e.renderer.setValue(d,"bindings="+JSON.stringify(a,null,2))}}}function Q_(e,t,n,i){d_.apply(void 0,Object(r.__spread)([e,t,n],i))}function X_(e,t){for(var n=t;n<e.def.nodes.length;n++){var r=e.def.nodes[n];if(16384&r.flags&&r.bindings&&r.bindings.length)return n}return null}function Z_(e,t){for(var n=t;n<e.def.nodes.length;n++){var r=e.def.nodes[n];if(3&r.flags&&r.bindings&&r.bindings.length)return n}return null}(k_=S_||(S_={}))[k_.create=0]="create",k_[k_.detectChanges=1]="detectChanges",k_[k_.checkNoChanges=2]="checkNoChanges",k_[k_.destroy=3]="destroy",k_[k_.handleEvent=4]="handleEvent";var $_=function(){function e(e,t){this.view=e,this.nodeIndex=t,null==t&&(this.nodeIndex=t=0),this.nodeDef=e.def.nodes[t];for(var n=this.nodeDef,r=e;n&&0==(1&n.flags);)n=n.parent;if(!n)for(;!n&&r;)n=lg(r),r=r.parent;this.elDef=n,this.elView=r}return Object.defineProperty(e.prototype,"elOrCompView",{get:function(){return Bm(this.elView,this.elDef.nodeIndex).componentView||this.view},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"injector",{get:function(){return sv(this.elView,this.elDef)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"component",{get:function(){return this.elOrCompView.component},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"context",{get:function(){return this.elOrCompView.context},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"providerTokens",{get:function(){var e=[];if(this.elDef)for(var t=this.elDef.nodeIndex+1;t<=this.elDef.nodeIndex+this.elDef.childCount;t++){var n=this.elView.def.nodes[t];20224&n.flags&&e.push(n.provider.token),t+=n.childCount}return e},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"references",{get:function(){var e={};if(this.elDef){J_(this.elView,this.elDef,e);for(var t=this.elDef.nodeIndex+1;t<=this.elDef.nodeIndex+this.elDef.childCount;t++){var n=this.elView.def.nodes[t];20224&n.flags&&J_(this.elView,n,e),t+=n.childCount}}return e},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"componentRenderElement",{get:function(){var e=function(e){for(;e&&!fg(e);)e=e.parent;if(e.parent)return Bm(e.parent,lg(e).nodeIndex);return null}(this.elOrCompView);return e?e.renderElement:void 0},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"renderNode",{get:function(){return 2&this.nodeDef.flags?dg(this.view,this.nodeDef):dg(this.elView,this.elDef)},enumerable:!0,configurable:!0}),e.prototype.logError=function(e){for(var t,n,i=[],o=1;o<arguments.length;o++)i[o-1]=arguments[o];2&this.nodeDef.flags?(t=this.view.def,n=this.nodeDef.nodeIndex):(t=this.elView.def,n=this.elDef.nodeIndex);var a=function(e,t){for(var n=-1,r=0;r<=t;r++){var i=e.nodes[r];3&i.flags&&n++}return n}(t,n),s=-1;t.factory(function(){var t;return++s===a?(t=e.error).bind.apply(t,Object(r.__spread)([e],i)):Qm}),s<a&&(e.error("Illegal state: the ViewDefinitionFactory did not call the logger!"),e.error.apply(e,Object(r.__spread)(i)))},e}();function J_(e,t,n){for(var r in t.references)n[r]=Hv(e,t,t.references[r])}function ey(e,t,n,r){var i,o,a=M_,s=A_,u=T_;try{M_=e;var c=t.apply(n,r);return A_=s,T_=u,M_=a,c}catch(e){if(tf(e)||!A_)throw e;throw i=e,o=ty(),i instanceof Error||(i=new Error(i.toString())),Ym(i,o),i}}function ty(){return A_?new $_(A_,T_):null}var ny=function(){function e(e){this.delegate=e}return e.prototype.createRenderer=function(e,t){return new ry(this.delegate.createRenderer(e,t))},e.prototype.begin=function(){this.delegate.begin&&this.delegate.begin()},e.prototype.end=function(){this.delegate.end&&this.delegate.end()},e.prototype.whenRenderingDone=function(){return this.delegate.whenRenderingDone?this.delegate.whenRenderingDone():Promise.resolve(null)},e}(),ry=function(){function e(e){this.delegate=e,this.debugContextFactory=ty,this.data=this.delegate.data}return e.prototype.createDebugContext=function(e){return this.debugContextFactory(e)},e.prototype.destroyNode=function(e){var t;t=im(e),nm.delete(t.nativeNode),this.delegate.destroyNode&&this.delegate.destroyNode(e)},e.prototype.destroy=function(){this.delegate.destroy()},e.prototype.createElement=function(e,t){var n=this.delegate.createElement(e,t),r=this.createDebugContext(n);if(r){var i=new Zh(n,null,r);i.name=e,om(i)}return n},e.prototype.createComment=function(e){var t=this.delegate.createComment(e),n=this.createDebugContext(t);return n&&om(new Xh(t,null,n)),t},e.prototype.createText=function(e){var t=this.delegate.createText(e),n=this.createDebugContext(t);return n&&om(new Xh(t,null,n)),t},e.prototype.appendChild=function(e,t){var n=im(e),r=im(t);n&&r&&n instanceof Zh&&n.addChild(r),this.delegate.appendChild(e,t)},e.prototype.insertBefore=function(e,t,n){var r=im(e),i=im(t),o=im(n);r&&i&&r instanceof Zh&&r.insertBefore(o,i),this.delegate.insertBefore(e,t,n)},e.prototype.removeChild=function(e,t){var n=im(e),r=im(t);n&&r&&n instanceof Zh&&n.removeChild(r),this.delegate.removeChild(e,t)},e.prototype.selectRootElement=function(e,t){var n=this.delegate.selectRootElement(e,t),r=ty();return r&&om(new Zh(n,null,r)),n},e.prototype.setAttribute=function(e,t,n,r){var i=im(e);if(i&&i instanceof Zh){var o=r?r+":"+t:t;i.attributes[o]=n}this.delegate.setAttribute(e,t,n,r)},e.prototype.removeAttribute=function(e,t,n){var r=im(e);if(r&&r instanceof Zh){var i=n?n+":"+t:t;r.attributes[i]=null}this.delegate.removeAttribute(e,t,n)},e.prototype.addClass=function(e,t){var n=im(e);n&&n instanceof Zh&&(n.classes[t]=!0),this.delegate.addClass(e,t)},e.prototype.removeClass=function(e,t){var n=im(e);n&&n instanceof Zh&&(n.classes[t]=!1),this.delegate.removeClass(e,t)},e.prototype.setStyle=function(e,t,n,r){var i=im(e);i&&i instanceof Zh&&(i.styles[t]=n),this.delegate.setStyle(e,t,n,r)},e.prototype.removeStyle=function(e,t,n){var r=im(e);r&&r instanceof Zh&&(r.styles[t]=null),this.delegate.removeStyle(e,t,n)},e.prototype.setProperty=function(e,t,n){var r=im(e);r&&r instanceof Zh&&(r.properties[t]=n),this.delegate.setProperty(e,t,n)},e.prototype.listen=function(e,t,n){if("string"!=typeof e){var r=im(e);r&&r.listeners.push(new Qh(t,n))}return this.delegate.listen(e,t,n)},e.prototype.parentNode=function(e){return this.delegate.parentNode(e)},e.prototype.nextSibling=function(e){return this.delegate.nextSibling(e)},e.prototype.setValue=function(e,t){return this.delegate.setValue(e,t)},e}();function iy(e){return b_(),Gm.overrideProvider(e)}function oy(e,t){return b_(),Gm.overrideComponentView(e,t)}function ay(){return b_(),Gm.clearOverrides()}function sy(e,t,n){return new uy(e,t,n)}var uy=function(e){function t(t,n,r){var i=e.call(this)||this;return i.moduleType=t,i._bootstrapComponents=n,i._ngModuleDefFactory=r,i}return Object(r.__extends)(t,e),t.prototype.create=function(e){b_();var t=function(e){var t=Array.from(e.providers),n=Array.from(e.modules),r={};for(var i in e.providersByKey)r[i]=e.providersByKey[i];return{factory:e.factory,isRoot:e.isRoot,providers:t,modules:n,providersByKey:r}}(yg(this._ngModuleDefFactory));return Gm.createNgModuleRef(this.moduleType,e||Ls.NULL,this._bootstrapComponents,t)},t}(Pu)},"./node_modules/@angular/core/node_modules/tslib/tslib.es6.js":function(e,t,n){"use strict";n.r(t),n.d(t,"__extends",function(){return i}),n.d(t,"__assign",function(){return o}),n.d(t,"__rest",function(){return a}),n.d(t,"__decorate",function(){return s}),n.d(t,"__param",function(){return u}),n.d(t,"__metadata",function(){return c}),n.d(t,"__awaiter",function(){return l}),n.d(t,"__generator",function(){return d}),n.d(t,"__exportStar",function(){return p}),n.d(t,"__values",function(){return f}),n.d(t,"__read",function(){return h}),n.d(t,"__spread",function(){return m}),n.d(t,"__await",function(){return g}),n.d(t,"__asyncGenerator",function(){return v}),n.d(t,"__asyncDelegator",function(){return _}),n.d(t,"__asyncValues",function(){return y}),n.d(t,"__makeTemplateObject",function(){return b}),n.d(t,"__importStar",function(){return j}),n.d(t,"__importDefault",function(){return x});var r=function(e,t){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,t){e.__proto__=t}||function(e,t){for(var n in t)t.hasOwnProperty(n)&&(e[n]=t[n])})(e,t)};function i(e,t){function n(){this.constructor=e}r(e,t),e.prototype=null===t?Object.create(t):(n.prototype=t.prototype,new n)}var o=function(){return(o=Object.assign||function(e){for(var t,n=1,r=arguments.length;n<r;n++){t=arguments[n];for(var i in t)Object.prototype.hasOwnProperty.call(t,i)&&(e[i]=t[i])}return e}).apply(this,arguments)};function a(e,t){var n={};for(var r in e)Object.prototype.hasOwnProperty.call(e,r)&&t.indexOf(r)<0&&(n[r]=e[r]);if(null!=e&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(e);i<r.length;i++)t.indexOf(r[i])<0&&(n[r[i]]=e[r[i]])}return n}function s(e,t,n,r){var i,o=arguments.length,a=o<3?t:null===r?r=Object.getOwnPropertyDescriptor(t,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(e,t,n,r);else for(var s=e.length-1;s>=0;s--)(i=e[s])&&(a=(o<3?i(a):o>3?i(t,n,a):i(t,n))||a);return o>3&&a&&Object.defineProperty(t,n,a),a}function u(e,t){return function(n,r){t(n,r,e)}}function c(e,t){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(e,t)}function l(e,t,n,r){return new(n||(n=Promise))(function(i,o){function a(e){try{u(r.next(e))}catch(e){o(e)}}function s(e){try{u(r.throw(e))}catch(e){o(e)}}function u(e){e.done?i(e.value):new n(function(t){t(e.value)}).then(a,s)}u((r=r.apply(e,t||[])).next())})}function d(e,t){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=t.call(e,a)}catch(e){o=[6,e],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(e,t){for(var n in e)t.hasOwnProperty(n)||(t[n]=e[n])}function f(e){var t="function"==typeof Symbol&&e[Symbol.iterator],n=0;return t?t.call(e):{next:function(){return e&&n>=e.length&&(e=void 0),{value:e&&e[n++],done:!e}}}}function h(e,t){var n="function"==typeof Symbol&&e[Symbol.iterator];if(!n)return e;var r,i,o=n.call(e),a=[];try{for(;(void 0===t||t-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(e){i={error:e}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var e=[],t=0;t<arguments.length;t++)e=e.concat(h(arguments[t]));return e}function g(e){return this instanceof g?(this.v=e,this):new g(e)}function v(e,t,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(e,t||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(e){i[e]&&(r[e]=function(t){return new Promise(function(n,r){o.push([e,t,n,r])>1||s(e,t)})})}function s(e,t){try{(n=i[e](t)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(e){l(o[0][3],e)}var n}function u(e){s("next",e)}function c(e){s("throw",e)}function l(e,t){e(t),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(e){var t,n;return t={},r("next"),r("throw",function(e){throw e}),r("return"),t[Symbol.iterator]=function(){return this},t;function r(r,i){t[r]=e[r]?function(t){return(n=!n)?{value:g(e[r](t)),done:"return"===r}:i?i(t):t}:i}}function y(e){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var t,n=e[Symbol.asyncIterator];return n?n.call(e):(e=f(e),t={},r("next"),r("throw"),r("return"),t[Symbol.asyncIterator]=function(){return this},t);function r(n){t[n]=e[n]&&function(t){return new Promise(function(r,i){var o,a,s,u;t=e[n](t),o=r,a=i,s=t.done,u=t.value,Promise.resolve(u).then(function(e){o({value:e,done:s})},a)})}}}function b(e,t){return Object.defineProperty?Object.defineProperty(e,"raw",{value:t}):e.raw=t,e}function j(e){if(e&&e.__esModule)return e;var t={};if(null!=e)for(var n in e)Object.hasOwnProperty.call(e,n)&&(t[n]=e[n]);return t.default=e,t}function x(e){return e&&e.__esModule?e:{default:e}}},"./node_modules/@angular/flex-layout/esm5/core.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"removeStyles",function(){return u}),n.d(t,"BROWSER_PROVIDER",function(){return c}),n.d(t,"CLASS_NAME",function(){return l}),n.d(t,"CoreModule",function(){return d}),n.d(t,"MediaChange",function(){return p}),n.d(t,"StylesheetMap",function(){return f}),n.d(t,"DEFAULT_CONFIG",function(){return h}),n.d(t,"LAYOUT_CONFIG",function(){return m}),n.d(t,"SERVER_TOKEN",function(){return g}),n.d(t,"BREAKPOINT",function(){return v}),n.d(t,"BaseDirective2",function(){return j}),n.d(t,"sortDescendingPriority",function(){return R}),n.d(t,"sortAscendingPriority",function(){return N}),n.d(t,"DEFAULT_BREAKPOINTS",function(){return x}),n.d(t,"ScreenTypes",function(){return M}),n.d(t,"ORIENTATION_BREAKPOINTS",function(){return A}),n.d(t,"BreakPointRegistry",function(){return L}),n.d(t,"BREAKPOINTS",function(){return F}),n.d(t,"MatchMedia",function(){return V}),n.d(t,"MockMatchMedia",function(){return z}),n.d(t,"MockMediaQueryList",function(){return U}),n.d(t,"MockMatchMediaProvider",function(){return H}),n.d(t,"ServerMediaQueryList",function(){return q}),n.d(t,"ServerMatchMedia",function(){return G}),n.d(t,"MediaObserver",function(){return Z}),n.d(t,"StyleUtils",function(){return J}),n.d(t,"StyleBuilder",function(){return ee}),n.d(t,"validateBasis",function(){return te}),n.d(t,"MediaMarshaller",function(){return re}),n.d(t,"BREAKPOINT_PRINT",function(){return K}),n.d(t,"PrintHook",function(){return Q});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/common/fesm5/common.js"),o=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/rxjs/_esm5/index.js"),s=n("./node_modules/rxjs/_esm5/operators/index.js");function u(e,t){return function(){if(Object(i.isPlatformBrowser)(t)){var n=/\bflex-layout-.+?\b/g;Array.from(e.querySelectorAll("[class*="+l+"]")).forEach(function(e){e.classList.contains(l+"ssr")&&e.parentNode?e.parentNode.removeChild(e):e.className.replace(n,"")})}}}var c={provide:r.APP_BOOTSTRAP_LISTENER,useFactory:u,deps:[i.DOCUMENT,r.PLATFORM_ID],multi:!0},l="flex-layout-",d=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{providers:[c]}]}],e}(),p=function(){function e(e,t,n,r){void 0===e&&(e=!1),void 0===t&&(t="all"),void 0===n&&(n=""),void 0===r&&(r=""),this.matches=e,this.mediaQuery=t,this.mqAlias=n,this.suffix=r,this.property=""}return e.prototype.clone=function(){return new e(this.matches,this.mediaQuery,this.mqAlias,this.suffix)},e}(),f=function(){function e(){this.stylesheet=new Map}return e.prototype.addStyleToElement=function(e,t,n){var r=this.stylesheet.get(e);r?r.set(t,n):this.stylesheet.set(e,new Map([[t,n]]))},e.prototype.clearStyles=function(){this.stylesheet.clear()},e.prototype.getStyleForElement=function(e,t){var n=this.stylesheet.get(e),r="";if(n){var i=n.get(t);"number"!=typeof i&&"string"!=typeof i||(r=i+"")}return r},e.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(),h={addFlexToParent:!0,addOrientationBps:!1,disableDefaultBps:!1,disableVendorPrefixes:!1,serverLoaded:!1,useColumnBasisZero:!0,printWithBreakpoints:[]},m=new r.InjectionToken("Flex Layout token, config options for the library",{providedIn:"root",factory:function(){return h}}),g=new r.InjectionToken("FlexLayoutServerLoaded",{providedIn:"root",factory:function(){return!1}}),v=new r.InjectionToken("Flex Layout token, collect all breakpoints into one provider",{providedIn:"root",factory:function(){return null}}),_="inline",y=["row","column","row-reverse","column-reverse"];function b(e){var t=function(e){var t=(e=e?e.toLowerCase():"").split(" "),n=t[0],r=t[1],i=t[2];y.find(function(e){return e===n})||(n=y[0]);r===_&&(r=i!==_?i:"",i=_);return[n,function(e){if(e)switch(e.toLowerCase()){case"reverse":case"wrap-reverse":case"reverse-wrap":e="wrap-reverse";break;case"no":case"none":case"nowrap":e="nowrap";break;default:e="wrap"}return e}(r),!!i]}(e);return function(e,t,n){void 0===t&&(t=null);void 0===n&&(n=!1);return{display:n?"inline-flex":"flex","box-sizing":"border-box","flex-direction":e,"flex-wrap":t||null}}(t[0],t[1],t[2])}var j=function(){function e(e,t,n,r){this.elementRef=e,this.styleBuilder=t,this.styler=n,this.marshal=r,this.DIRECTIVE_KEY="",this.inputs=[],this.mru={},this.destroySubject=new a.Subject,this.styleCache=new Map}return Object.defineProperty(e.prototype,"parentElement",{get:function(){return this.elementRef.nativeElement.parentElement},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"nativeElement",{get:function(){return this.elementRef.nativeElement},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"activatedValue",{get:function(){return this.marshal.getValue(this.nativeElement,this.DIRECTIVE_KEY)},set:function(e){this.marshal.setValue(this.nativeElement,this.DIRECTIVE_KEY,e,this.marshal.activatedAlias)},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(e){var t=this;Object.keys(e).forEach(function(n){if(-1!==t.inputs.indexOf(n)){var r=n.split(".").slice(1).join("."),i=e[n].currentValue;t.setValue(i,r)}})},e.prototype.ngOnDestroy=function(){this.destroySubject.next(),this.destroySubject.complete(),this.marshal.releaseElement(this.nativeElement)},e.prototype.init=function(e){void 0===e&&(e=[]),this.marshal.init(this.elementRef.nativeElement,this.DIRECTIVE_KEY,this.updateWithValue.bind(this),this.clearStyles.bind(this),e)},e.prototype.addStyles=function(e,t){var n=this.styleBuilder,r=n.shouldCache,i=this.styleCache.get(e);i&&r||(i=n.buildStyles(e,t),r&&this.styleCache.set(e,i)),this.mru=Object(o.__assign)({},i),this.applyStyleToElement(i),n.sideEffect(e,i,t)},e.prototype.clearStyles=function(){var e=this;Object.keys(this.mru).forEach(function(t){e.mru[t]=""}),this.applyStyleToElement(this.mru),this.mru={}},e.prototype.triggerUpdate=function(){var e=this.marshal.getValue(this.nativeElement,this.DIRECTIVE_KEY);void 0!==e&&this.marshal.updateElement(this.nativeElement,this.DIRECTIVE_KEY,e)},e.prototype.getFlexFlowDirection=function(e,t){if(void 0===t&&(t=!1),e){var n=this.styler.getFlowDirection(e),r=n[0];if(!n[1]&&t){var i=b(r),o=[e];this.styler.applyStyleToElements(i,o)}return r.trim()}return"row"},e.prototype.applyStyleToElement=function(e,t,n){void 0===n&&(n=this.nativeElement),this.styler.applyStyleToElement(n,e,t)},e.prototype.setValue=function(e,t){this.marshal.setValue(this.nativeElement,this.DIRECTIVE_KEY,e,t)},e.prototype.updateWithValue=function(e){this.addStyles(e)},e}(),x=[{alias:"xs",mediaQuery:"screen and (min-width: 0px) and (max-width: 599px)",priority:1e3},{alias:"sm",mediaQuery:"screen and (min-width: 600px) and (max-width: 959px)",priority:900},{alias:"md",mediaQuery:"screen and (min-width: 960px) and (max-width: 1279px)",priority:800},{alias:"lg",mediaQuery:"screen and (min-width: 1280px) and (max-width: 1919px)",priority:700},{alias:"xl",mediaQuery:"screen and (min-width: 1920px) and (max-width: 5000px)",priority:600},{alias:"lt-sm",overlapping:!0,mediaQuery:"screen and (max-width: 599px)",priority:950},{alias:"lt-md",overlapping:!0,mediaQuery:"screen and (max-width: 959px)",priority:850},{alias:"lt-lg",overlapping:!0,mediaQuery:"screen and (max-width: 1279px)",priority:750},{alias:"lt-xl",overlapping:!0,priority:650,mediaQuery:"screen and (max-width: 1919px)"},{alias:"gt-xs",overlapping:!0,mediaQuery:"screen and (min-width: 600px)",priority:-950},{alias:"gt-sm",overlapping:!0,mediaQuery:"screen and (min-width: 960px)",priority:-850},{alias:"gt-md",overlapping:!0,mediaQuery:"screen and (min-width: 1280px)",priority:-750},{alias:"gt-lg",overlapping:!0,mediaQuery:"screen and (min-width: 1920px)",priority:-650}],O="(orientation: portrait) and (max-width: 599px)",w="(orientation: landscape) and (max-width: 959px)",C="(orientation: landscape) and (min-width: 960px) and (max-width: 1279px)",E="(orientation: portrait) and (min-width: 600px) and (max-width: 839px)",S="(orientation: portrait) and (min-width: 840px)",k="(orientation: landscape) and (min-width: 1280px)",M={HANDSET:O+", "+w,TABLET:E+" , "+C,WEB:S+", "+k+" ",HANDSET_PORTRAIT:""+O,TABLET_PORTRAIT:E+" ",WEB_PORTRAIT:""+S,HANDSET_LANDSCAPE:w+"]",TABLET_LANDSCAPE:""+C,WEB_LANDSCAPE:""+k},A=[{alias:"handset",priority:2e3,mediaQuery:M.HANDSET},{alias:"handset.landscape",priority:2e3,mediaQuery:M.HANDSET_LANDSCAPE},{alias:"handset.portrait",priority:2e3,mediaQuery:M.HANDSET_PORTRAIT},{alias:"tablet",priority:2100,mediaQuery:M.TABLET},{alias:"tablet.landscape",priority:2100,mediaQuery:M.TABLET},{alias:"tablet.portrait",priority:2100,mediaQuery:M.TABLET_PORTRAIT},{alias:"web",priority:2200,mediaQuery:M.WEB,overlapping:!0},{alias:"web.landscape",priority:2200,mediaQuery:M.WEB_LANDSCAPE,overlapping:!0},{alias:"web.portrait",priority:2200,mediaQuery:M.WEB_PORTRAIT,overlapping:!0}];function T(e){for(var t=[],n=1;n<arguments.length;n++)t[n-1]=arguments[n];if(null==e)throw TypeError("Cannot convert undefined or null to object");for(var r=0,i=t;r<i.length;r++){var o=i[r];if(null!=o)for(var a in o)o.hasOwnProperty(a)&&(e[a]=o[a])}return e}var D=/(\.|-|_)/g;function I(e){var t=e.length>0?e.charAt(0):"",n=e.length>1?e.slice(1):"";return t.toUpperCase()+n}function P(e){return e.forEach(function(e){e.suffix||(e.suffix=e.alias.replace(D,"|").split("|").map(I).join(""),e.overlapping=!!e.overlapping)}),e}function R(e,t){var n=e?e.priority||0:0;return(t?t.priority||0:0)-n}function N(e,t){return(e.priority||0)-(t.priority||0)}var F=new r.InjectionToken("Token (@angular/flex-layout) Breakpoints",{providedIn:"root",factory:function(){var e=Object(r.inject)(v),t=Object(r.inject)(m),n=[].concat.apply([],(e||[]).map(function(e){return Array.isArray(e)?e:[e]}));return function(e,t){void 0===t&&(t=[]);var n={};return e.forEach(function(e){n[e.alias]=e}),t.forEach(function(e){n[e.alias]?T(n[e.alias],e):n[e.alias]=e}),P(Object.keys(n).map(function(e){return n[e]}))}((t.disableDefaultBps?[]:x).concat(t.addOrientationBps?A:[]),n)}}),L=function(){function e(e){this.findByMap=new Map,this.items=e.slice().sort(N)}return e.prototype.findByAlias=function(e){return e?this.findWithPredicate(e,function(t){return t.alias==e}):null},e.prototype.findByQuery=function(e){return this.findWithPredicate(e,function(t){return t.mediaQuery==e})},Object.defineProperty(e.prototype,"overlappings",{get:function(){return this.items.filter(function(e){return 1==e.overlapping})},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"aliases",{get:function(){return this.items.map(function(e){return e.alias})},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"suffixes",{get:function(){return this.items.map(function(e){return e.suffix?e.suffix:""})},enumerable:!0,configurable:!0}),e.prototype.findWithPredicate=function(e,t){var n=this.findByMap.get(e);return n||(n=this.items.find(t)||null,this.findByMap.set(e,n)),n||null},e.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:Array,decorators:[{type:r.Inject,args:[F]}]}]},e.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new e(Object(r.inject)(F))},token:e,providedIn:"root"}),e}(),V=function(){function e(e,t,n){this._zone=e,this._platformId=t,this._document=n,this._source=new a.BehaviorSubject(new p(!0)),this._registry=new Map,this._observable$=this._source.asObservable()}return e.prototype.isActive=function(e){var t=this._registry.get(e);return!!t&&t.matches},e.prototype.observe=function(e,t){var n=this;if(void 0===t&&(t=!1),e){var r=this._observable$.pipe(Object(s.filter)(function(n){return!t||e.indexOf(n.mediaQuery)>-1})),i=new a.Observable(function(t){var r=n.registerQuery(e);if(r.length){var i=r.pop();r.forEach(function(e){t.next(e)}),n._source.next(i)}t.complete()});return Object(a.merge)(i,r)}return this._observable$},e.prototype.registerQuery=function(e){var t=this,n=Array.isArray(e)?e:[e],r=[];return function(e,t){var n=e.filter(function(e){return!B[e]});if(n.length>0){var r=n.join(", ");try{var i=t.createElement("style");if(i.setAttribute("type","text/css"),!i.styleSheet){var o="\n/*\n  @angular/flex-layout - workaround for possible browser quirk with mediaQuery listeners\n  see http://bit.ly/2sd4HMP\n*/\n@media "+r+" {.fx-query-test{ }}\n";i.appendChild(t.createTextNode(o))}t.head.appendChild(i),n.forEach(function(e){return B[e]=i})}catch(e){console.error(e)}}}(n,this._document),n.forEach(function(e){var n=t._registry.get(e);n||((n=t.buildMQL(e)).addListener(function(n){t._zone.run(function(){return t._source.next(new p(n.matches,e))})}),t._registry.set(e,n)),n.matches&&r.push(new p(!0,e))}),r},e.prototype.buildMQL=function(e){return t=e,Object(i.isPlatformBrowser)(this._platformId)&&window.matchMedia("all").addListener?window.matchMedia(t):{matches:"all"===t||""===t,media:t,addListener:function(){},removeListener:function(){}};var t},e.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:r.NgZone},{type:Object,decorators:[{type:r.Inject,args:[r.PLATFORM_ID]}]},{type:void 0,decorators:[{type:r.Inject,args:[i.DOCUMENT]}]}]},e.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new e(Object(r.inject)(r.NgZone),Object(r.inject)(r.PLATFORM_ID),Object(r.inject)(i.DOCUMENT))},token:e,providedIn:"root"}),e}(),B={};var z=function(e){function t(t,n,r,i){var o=e.call(this,t,n,r)||this;return o._breakpoints=i,o.autoRegisterQueries=!0,o.useOverlaps=!1,o._registry=new Map,o._actives=[],o}return Object(o.__extends)(t,e),t.prototype.clearAll=function(){this._registry.forEach(function(e){e.destroy()}),this._registry.clear(),this.useOverlaps=!1},t.prototype.activate=function(e,t){return void 0===t&&(t=!1),t=t||this.useOverlaps,e=this._validateQuery(e),!t&&this.isActive(e)||(this._deactivateAll(),this._registerMediaQuery(e),this._activateWithOverlaps(e,t)),this.hasActivated},t.prototype._validateQuery=function(e){var t=this._breakpoints.findByAlias(e);return t&&t.mediaQuery||e},t.prototype._activateWithOverlaps=function(e,t){if(t){var n=this._breakpoints.findByQuery(e),r=n?n.alias:"unknown";switch(r){case"lg":this._activateByAlias("lt-xl");break;case"md":this._activateByAlias("lt-xl, lt-lg");break;case"sm":this._activateByAlias("lt-xl, lt-lg, lt-md");break;case"xs":this._activateByAlias("lt-xl, lt-lg, lt-md, lt-sm")}switch(r){case"xl":this._activateByAlias("gt-lg, gt-md, gt-sm, gt-xs");break;case"lg":this._activateByAlias("gt-md, gt-sm, gt-xs");break;case"md":this._activateByAlias("gt-sm, gt-xs");break;case"sm":this._activateByAlias("gt-xs")}}return this._activateByQuery(e)},t.prototype._activateByAlias=function(e){var t=this;e.split(",").forEach(function(e){return n=e.trim(),r=t._breakpoints.findByAlias(n),void t._activateByQuery(r?r.mediaQuery:n);var n,r})},t.prototype._activateByQuery=function(e){var t=this._registry.get(e),n=this._actives.reduce(function(e,n){return e||!!t&&n.media===t.media},!1);return t&&!n&&this._actives.push(t.activate()),this.hasActivated},t.prototype._deactivateAll=function(){return this._actives.forEach(function(e){return e.deactivate()}),this._actives=[],this},t.prototype._registerMediaQuery=function(e){!this._registry.has(e)&&this.autoRegisterQueries&&this.registerQuery(e)},t.prototype.buildMQL=function(e){return new U(e)},Object.defineProperty(t.prototype,"hasActivated",{get:function(){return this._actives.length>0},enumerable:!0,configurable:!0}),t.decorators=[{type:r.Injectable}],t.ctorParameters=function(){return[{type:r.NgZone},{type:Object,decorators:[{type:r.Inject,args:[r.PLATFORM_ID]}]},{type:void 0,decorators:[{type:r.Inject,args:[i.DOCUMENT]}]},{type:L}]},t}(V),U=function(){function e(e){this._mediaQuery=e,this._isActive=!1,this._listeners=[],this.onchange=null}return Object.defineProperty(e.prototype,"matches",{get:function(){return this._isActive},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"media",{get:function(){return this._mediaQuery},enumerable:!0,configurable:!0}),e.prototype.destroy=function(){this.deactivate(),this._listeners=[]},e.prototype.activate=function(){var e=this;return this._isActive||(this._isActive=!0,this._listeners.forEach(function(t){t.call(null,e)})),this},e.prototype.deactivate=function(){var e=this;return this._isActive&&(this._isActive=!1,this._listeners.forEach(function(t){t.call(null,e)})),this},e.prototype.addListener=function(e){(-1===this._listeners.indexOf(e)&&this._listeners.push(e),this._isActive)&&e.call(null,this)},e.prototype.removeListener=function(e){},e.prototype.addEventListener=function(e,t,n){},e.prototype.removeEventListener=function(e,t,n){},e.prototype.dispatchEvent=function(e){return!1},e}(),H={provide:V,useClass:z},q=function(){function e(e){this._mediaQuery=e,this._isActive=!1,this._listeners=[],this.onchange=null}return Object.defineProperty(e.prototype,"matches",{get:function(){return this._isActive},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"media",{get:function(){return this._mediaQuery},enumerable:!0,configurable:!0}),e.prototype.destroy=function(){this.deactivate(),this._listeners=[]},e.prototype.activate=function(){var e=this;return this._isActive||(this._isActive=!0,this._listeners.forEach(function(t){t.call(null,e)})),this},e.prototype.deactivate=function(){var e=this;return this._isActive&&(this._isActive=!1,this._listeners.forEach(function(t){t.call(null,e)})),this},e.prototype.addListener=function(e){(-1===this._listeners.indexOf(e)&&this._listeners.push(e),this._isActive)&&e.call(null,this)},e.prototype.removeListener=function(e){},e.prototype.addEventListener=function(e,t,n){},e.prototype.removeEventListener=function(e,t,n){},e.prototype.dispatchEvent=function(e){return!1},e}(),G=function(e){function t(t,n,r){var i=e.call(this,t,n,r)||this;return i._zone=t,i._platformId=n,i._document=r,i._registry=new Map,i}return Object(o.__extends)(t,e),t.prototype.activateBreakpoint=function(e){var t=this._registry.get(e.mediaQuery);t&&t.activate()},t.prototype.deactivateBreakpoint=function(e){var t=this._registry.get(e.mediaQuery);t&&t.deactivate()},t.prototype.buildMQL=function(e){return new q(e)},t.decorators=[{type:r.Injectable}],t.ctorParameters=function(){return[{type:r.NgZone},{type:Object,decorators:[{type:r.Inject,args:[r.PLATFORM_ID]}]},{type:void 0,decorators:[{type:r.Inject,args:[i.DOCUMENT]}]}]},t}(V);function W(e,t){return T(e||{},t?{mqAlias:t.alias,suffix:t.suffix}:{})}var Y="print",K={alias:Y,mediaQuery:Y,priority:1e3},Q=function(){function e(e,t){this.breakpoints=e,this.layoutConfig=t,this.isPrinting=!1,this.queue=new X,this.deactivations=[]}return e.prototype.withPrintQuery=function(e){return e.concat([Y])},e.prototype.isPrintEvent=function(e){return e.mediaQuery.startsWith(Y)},Object.defineProperty(e.prototype,"printAlias",{get:function(){return this.layoutConfig.printWithBreakpoints||[]},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"printBreakPoints",{get:function(){var e=this;return this.printAlias.map(function(t){return e.breakpoints.findByAlias(t)}).filter(function(e){return null!==e})},enumerable:!0,configurable:!0}),e.prototype.getEventBreakpoints=function(e){var t=e.mediaQuery,n=this.breakpoints.findByQuery(t);return(n?this.printBreakPoints.concat([n]):this.printBreakPoints).sort(R)},e.prototype.updateEvent=function(e){var t=this.breakpoints.findByQuery(e.mediaQuery);return this.isPrintEvent(e)&&(t=this.getEventBreakpoints(e)[0],e.mediaQuery=t?t.mediaQuery:""),W(e,t)},e.prototype.interceptEvents=function(e){var t=this;return function(n){t.isPrintEvent(n)?n.matches&&!t.isPrinting?(t.startPrinting(e,t.getEventBreakpoints(n)),e.updateStyles()):!n.matches&&t.isPrinting&&(t.stopPrinting(e),e.updateStyles()):t.collectActivations(n)}},e.prototype.blockPropagation=function(){var e=this;return function(t){return!(e.isPrinting||e.isPrintEvent(t))}},e.prototype.startPrinting=function(e,t){this.isPrinting=!0,e.activatedBreakpoints=this.queue.addPrintBreakpoints(t)},e.prototype.stopPrinting=function(e){e.activatedBreakpoints=this.deactivations,this.deactivations=[],this.queue.clear(),this.isPrinting=!1},e.prototype.collectActivations=function(e){if(!this.isPrinting)if(e.matches)this.deactivations=[];else{var t=this.breakpoints.findByQuery(e.mediaQuery);t&&(this.deactivations.push(t),this.deactivations.sort(R))}},e.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:L},{type:void 0,decorators:[{type:r.Inject,args:[m]}]}]},e.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new e(Object(r.inject)(L),Object(r.inject)(m))},token:e,providedIn:"root"}),e}(),X=function(){function e(){this.printBreakpoints=[]}return e.prototype.addPrintBreakpoints=function(e){var t=this;return e.push(K),e.sort(R),e.forEach(function(e){return t.addBreakpoint(e)}),this.printBreakpoints},e.prototype.addBreakpoint=function(e){var t;e&&(void 0===this.printBreakpoints.find(function(t){return t.mediaQuery===e.mediaQuery})&&(this.printBreakpoints=(t=e)&&t.mediaQuery.startsWith(Y)?[e].concat(this.printBreakpoints):this.printBreakpoints.concat([e])))},e.prototype.clear=function(){this.printBreakpoints=[]},e}();var Z=function(){function e(e,t,n){this.breakpoints=e,this.mediaWatcher=t,this.hook=n,this.filterOverlaps=!0,this.media$=this.watchActivations()}return e.prototype.isActive=function(e){return this.mediaWatcher.isActive(this.toMediaQuery(e))},e.prototype.watchActivations=function(){var e=this.breakpoints.items.map(function(e){return e.mediaQuery});return this.buildObservable(e)},e.prototype.buildObservable=function(e){var t=this,n=this.breakpoints;return this.mediaWatcher.observe(this.hook.withPrintQuery(e)).pipe(Object(s.filter)(function(e){return e.matches}),Object(s.filter)(function(e){var r=n.findByQuery(e.mediaQuery);return!r||!(t.filterOverlaps&&r.overlapping)}),Object(s.map)(function(e){return t.hook.isPrintEvent(e)?t.hook.updateEvent(e):W(e,n.findByQuery(e.mediaQuery))}),Object(s.filter)(function(e){return!e.mediaQuery.startsWith("print")}),Object(s.filter)(function(e){return""!==e.mediaQuery}))},e.prototype.toMediaQuery=function(e){var t=this.breakpoints,n=t.findByAlias(e)||t.findByQuery(e);return n?n.mediaQuery:e},e.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:L},{type:V},{type:Q}]},e.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new e(Object(r.inject)(L),Object(r.inject)(V),Object(r.inject)(Q))},token:e,providedIn:"root"}),e}();function $(e){for(var t in e){var n=e[t]||"";switch(t){case"display":e.display="flex"===n?["-webkit-flex","flex"]:"inline-flex"===n?["-webkit-inline-flex","inline-flex"]:n;break;case"align-items":case"align-self":case"align-content":case"flex":case"flex-basis":case"flex-flow":case"flex-grow":case"flex-shrink":case"flex-wrap":case"justify-content":e["-webkit-"+t]=n;break;case"flex-direction":n=n||"row",e["-webkit-flex-direction"]=n,e["flex-direction"]=n;break;case"order":e.order=e["-webkit-"+t]=isNaN(+n)?"0":n}}return e}var J=function(){function e(e,t,n,r){this._serverStylesheet=e,this._serverModuleLoaded=t,this._platformId=n,this.layoutConfig=r}return e.prototype.applyStyleToElement=function(e,t,n){void 0===n&&(n=null);var r={};"string"==typeof t&&(r[t]=n,t=r),r=this.layoutConfig.disableVendorPrefixes?t:$(t),this._applyMultiValueStyleToElement(r,e)},e.prototype.applyStyleToElements=function(e,t){var n=this;void 0===t&&(t=[]);var r=this.layoutConfig.disableVendorPrefixes?e:$(e);t.forEach(function(e){n._applyMultiValueStyleToElement(r,e)})},e.prototype.getFlowDirection=function(e){var t=this.lookupStyle(e,"flex-direction"),n=this.lookupInlineStyle(e,"flex-direction")||Object(i.isPlatformServer)(this._platformId)&&this._serverModuleLoaded?t:"";return[t||"row",n]},e.prototype.lookupAttributeValue=function(e,t){return e.getAttribute(t)||""},e.prototype.lookupInlineStyle=function(e,t){return Object(i.isPlatformBrowser)(this._platformId)?e.style.getPropertyValue(t):this._getServerStyle(e,t)},e.prototype.lookupStyle=function(e,t,n){void 0===n&&(n=!1);var r="";e&&((r=this.lookupInlineStyle(e,t))||(Object(i.isPlatformBrowser)(this._platformId)?n||(r=getComputedStyle(e).getPropertyValue(t)):this._serverModuleLoaded&&(r=this._serverStylesheet.getStyleForElement(e,t))));return r.trim()},e.prototype._applyMultiValueStyleToElement=function(e,t){var n=this;Object.keys(e).sort().forEach(function(r){var o=e[r],a=Array.isArray(o)?o:[o];a.sort();for(var s=0,u=a;s<u.length;s++){var c=u[s];c=c?c+"":"",Object(i.isPlatformBrowser)(n._platformId)||!n._serverModuleLoaded?Object(i.isPlatformBrowser)(n._platformId)?t.style.setProperty(r,c):n._setServerStyle(t,r,c):n._serverStylesheet.addStyleToElement(t,r,c)}})},e.prototype._setServerStyle=function(e,t,n){t=t.replace(/([a-z])([A-Z])/g,"$1-$2").toLowerCase();var r=this._readStyleAttribute(e);r[t]=n||"",this._writeStyleAttribute(e,r)},e.prototype._getServerStyle=function(e,t){return this._readStyleAttribute(e)[t]||""},e.prototype._readStyleAttribute=function(e){var t={},n=e.getAttribute("style");if(n)for(var r=n.split(/;+/g),i=0;i<r.length;i++){var o=r[i].trim();if(o.length>0){var a=o.indexOf(":");if(-1===a)throw new Error("Invalid CSS style: "+o);t[o.substr(0,a).trim()]=o.substr(a+1).trim()}}return t},e.prototype._writeStyleAttribute=function(e,t){var n="";for(var r in t){t[r]&&(n+=r+":"+t[r]+";")}e.setAttribute("style",n)},e.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:f,decorators:[{type:r.Optional}]},{type:Boolean,decorators:[{type:r.Optional},{type:r.Inject,args:[g]}]},{type:Object,decorators:[{type:r.Inject,args:[r.PLATFORM_ID]}]},{type:void 0,decorators:[{type:r.Inject,args:[m]}]}]},e.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new e(Object(r.inject)(f,8),Object(r.inject)(g,8),Object(r.inject)(r.PLATFORM_ID),Object(r.inject)(m))},token:e,providedIn:"root"}),e}(),ee=function(){function e(){this.shouldCache=!0}return e.prototype.sideEffect=function(e,t,n){},e}();function te(e,t,n){void 0===t&&(t="1"),void 0===n&&(n="1");var r=[t,n,e],i=e.indexOf("calc");if(i>0)r[2]=ne(e.substring(i).trim()),2==(o=e.substr(0,i).trim().split(" ")).length&&(r[0]=o[0],r[1]=o[1]);else if(0==i)r[2]=ne(e.trim());else{var o;r=3===(o=e.split(" ")).length?o:[t,n,e]}return r}function ne(e){return e.replace(/[\s]/g,"").replace(/[\/\*\+\-]/g," $& ")}var re=function(){function e(e,t,n){this.matchMedia=e,this.breakpoints=t,this.hook=n,this.activatedBreakpoints=[],this.elementMap=new Map,this.elementKeyMap=new WeakMap,this.watcherMap=new WeakMap,this.updateMap=new WeakMap,this.clearMap=new WeakMap,this.subject=new a.Subject,this.observeActivations()}return Object.defineProperty(e.prototype,"activatedAlias",{get:function(){return this.activatedBreakpoints[0]?this.activatedBreakpoints[0].alias:""},enumerable:!0,configurable:!0}),e.prototype.onMediaChange=function(e){var t=this.findByQuery(e.mediaQuery);t&&((e=W(e,t)).matches&&-1===this.activatedBreakpoints.indexOf(t)?(this.activatedBreakpoints.push(t),this.activatedBreakpoints.sort(R),this.updateStyles()):e.matches||-1===this.activatedBreakpoints.indexOf(t)||(this.activatedBreakpoints.splice(this.activatedBreakpoints.indexOf(t),1),this.activatedBreakpoints.sort(R),this.updateStyles()))},e.prototype.init=function(e,t,n,r,i){void 0===i&&(i=[]),ie(this.updateMap,e,t,n),ie(this.clearMap,e,t,r),this.buildElementKeyMap(e,t),this.watchExtraTriggers(e,t,i)},e.prototype.getValue=function(e,t,n){var r=this.elementMap.get(e);if(r){var i=void 0!==n?r.get(n):this.getActivatedValues(r,t);if(i)return i.get(t)}},e.prototype.hasValue=function(e,t){var n=this.elementMap.get(e);if(n){var r=this.getActivatedValues(n,t);if(r)return void 0!==r.get(t)||!1}return!1},e.prototype.setValue=function(e,t,n,r){var i=this.elementMap.get(e);if(i){var o=(i.get(r)||new Map).set(t,n);i.set(r,o),this.elementMap.set(e,i)}else i=(new Map).set(r,(new Map).set(t,n)),this.elementMap.set(e,i);var a=this.getValue(e,t);void 0!==a&&this.updateElement(e,t,a)},e.prototype.trackValue=function(e,t){return this.subject.asObservable().pipe(Object(s.filter)(function(n){return n.element===e&&n.key===t}))},e.prototype.updateStyles=function(){var e=this;this.elementMap.forEach(function(t,n){var r=new Set(e.elementKeyMap.get(n)),i=e.getActivatedValues(t);i&&i.forEach(function(t,i){e.updateElement(n,i,t),r.delete(i)}),r.forEach(function(r){if(i=e.getActivatedValues(t,r)){var o=i.get(r);e.updateElement(n,r,o)}else e.clearElement(n,r)})})},e.prototype.clearElement=function(e,t){var n=this.clearMap.get(e);if(n){var r=n.get(t);r&&(r(),this.subject.next({element:e,key:t,value:""}))}},e.prototype.updateElement=function(e,t,n){var r=this.updateMap.get(e);if(r){var i=r.get(t);i&&(i(n),this.subject.next({element:e,key:t,value:n}))}},e.prototype.releaseElement=function(e){var t=this.watcherMap.get(e);t&&(t.forEach(function(e){return e.unsubscribe()}),this.watcherMap.delete(e));var n=this.elementMap.get(e);n&&(n.forEach(function(e,t){return n.delete(t)}),this.elementMap.delete(e))},e.prototype.buildElementKeyMap=function(e,t){var n=this.elementKeyMap.get(e);n||(n=new Set,this.elementKeyMap.set(e,n)),n.add(t)},e.prototype.watchExtraTriggers=function(e,t,n){var r=this;if(n&&n.length){var i=this.watcherMap.get(e);if(i||(i=new Map,this.watcherMap.set(e,i)),!i.get(t)){var o=a.merge.apply(void 0,n).subscribe(function(){var n=r.getValue(e,t);r.updateElement(e,t,n)});i.set(t,o)}}},e.prototype.findByQuery=function(e){return this.breakpoints.findByQuery(e)},e.prototype.getActivatedValues=function(e,t){for(var n=0;n<this.activatedBreakpoints.length;n++){var r=this.activatedBreakpoints[n],i=e.get(r.alias);if(i&&(void 0===t||i.has(t)))return i}var o=e.get("");return void 0===t||o&&o.has(t)?o:void 0},e.prototype.observeActivations=function(){var e=this.breakpoints.items.map(function(e){return e.mediaQuery});this.matchMedia.observe(this.hook.withPrintQuery(e)).pipe(Object(s.tap)(this.hook.interceptEvents(this)),Object(s.filter)(this.hook.blockPropagation())).subscribe(this.onMediaChange.bind(this))},e.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:V},{type:L},{type:Q}]},e.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new e(Object(r.inject)(V),Object(r.inject)(L),Object(r.inject)(Q))},token:e,providedIn:"root"}),e}();function ie(e,t,n,r){if(void 0!==r){var i=e.get(t);i||(i=new Map,e.set(t,i)),i.set(n,r)}}},"./node_modules/@angular/flex-layout/esm5/extended.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"ExtendedModule",function(){return I}),n.d(t,"ClassDirective",function(){return m}),n.d(t,"DefaultClassDirective",function(){return v}),n.d(t,"ImgSrcStyleBuilder",function(){return l}),n.d(t,"ImgSrcDirective",function(){return d}),n.d(t,"DefaultImgSrcDirective",function(){return h}),n.d(t,"ShowHideStyleBuilder",function(){return _}),n.d(t,"ShowHideDirective",function(){return y}),n.d(t,"DefaultShowHideDirective",function(){return x}),n.d(t,"StyleDirective",function(){return k}),n.d(t,"DefaultStyleDirective",function(){return A});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/common/fesm5/common.js"),a=n("./node_modules/@angular/flex-layout/esm5/core.es5.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/rxjs/_esm5/operators/index.js"),c=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),l=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e){return{content:e?"url("+e+")":""}},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(a.StyleBuilder),d=function(e){function t(t,n,r,i,a,s){var u=e.call(this,t,n,r,i)||this;return u.elementRef=t,u.styleBuilder=n,u.styler=r,u.marshal=i,u.platformId=a,u.serverModuleLoaded=s,u.DIRECTIVE_KEY="img-src",u.defaultSrc="",u.styleCache=p,u.init(),u.setValue("",u.nativeElement.getAttribute("src")||""),Object(o.isPlatformServer)(u.platformId)&&u.serverModuleLoaded&&u.nativeElement.setAttribute("src",""),u}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"src",{set:function(e){this.defaultSrc=e,this.setValue("",this.defaultSrc)},enumerable:!0,configurable:!0}),t.prototype.updateWithValue=function(){var e=this.activatedValue||this.defaultSrc;Object(o.isPlatformServer)(this.platformId)&&this.serverModuleLoaded?this.addStyles(e):this.nativeElement.setAttribute("src",String(e))},t.ctorParameters=function(){return[{type:i.ElementRef},{type:l},{type:a.StyleUtils},{type:a.MediaMarshaller},{type:Object,decorators:[{type:i.Inject,args:[i.PLATFORM_ID]}]},{type:Boolean,decorators:[{type:i.Inject,args:[a.SERVER_TOKEN]}]}]},t.propDecorators={src:[{type:i.Input,args:["src"]}]},t}(a.BaseDirective2),p=new Map,f=["src.xs","src.sm","src.md","src.lg","src.xl","src.lt-sm","src.lt-md","src.lt-lg","src.lt-xl","src.gt-xs","src.gt-sm","src.gt-md","src.gt-lg"],h=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=f,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  img[src.xs],    img[src.sm],    img[src.md],    img[src.lg],   img[src.xl],\n  img[src.lt-sm], img[src.lt-md], img[src.lt-lg], img[src.lt-xl],\n  img[src.gt-xs], img[src.gt-sm], img[src.gt-md], img[src.gt-lg]\n",inputs:f}]}],t}(d),m=function(e){function t(t,n,r,i,a,s,u){var c=e.call(this,t,null,n,r)||this;return c.elementRef=t,c.styler=n,c.marshal=r,c.iterableDiffers=i,c.keyValueDiffers=a,c.renderer=s,c.ngClassInstance=u,c.DIRECTIVE_KEY="ngClass",c.ngClassInstance||(c.ngClassInstance=new o.NgClass(c.iterableDiffers,c.keyValueDiffers,c.elementRef,c.renderer)),c.init(),c}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"klass",{set:function(e){this.ngClassInstance.klass=e,this.setValue(e,"")},enumerable:!0,configurable:!0}),t.prototype.updateWithValue=function(e){this.ngClassInstance.ngClass=e,this.ngClassInstance.ngDoCheck()},t.prototype.ngDoCheck=function(){this.ngClassInstance.ngDoCheck()},t.ctorParameters=function(){return[{type:i.ElementRef},{type:a.StyleUtils},{type:a.MediaMarshaller},{type:i.IterableDiffers},{type:i.KeyValueDiffers},{type:i.Renderer2},{type:o.NgClass,decorators:[{type:i.Optional},{type:i.Self}]}]},t.propDecorators={klass:[{type:i.Input,args:["class"]}]},t}(a.BaseDirective2),g=["ngClass","ngClass.xs","ngClass.sm","ngClass.md","ngClass.lg","ngClass.xl","ngClass.lt-sm","ngClass.lt-md","ngClass.lt-lg","ngClass.lt-xl","ngClass.gt-xs","ngClass.gt-sm","ngClass.gt-md","ngClass.gt-lg"],v=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=g,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [ngClass], [ngClass.xs], [ngClass.sm], [ngClass.md], [ngClass.lg], [ngClass.xl],\n  [ngClass.lt-sm], [ngClass.lt-md], [ngClass.lt-lg], [ngClass.lt-xl],\n  [ngClass.gt-xs], [ngClass.gt-sm], [ngClass.gt-md], [ngClass.gt-lg]\n",inputs:g}]}],t}(m),_=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e,t){return{display:"true"===e?t.display:"none"}},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(a.StyleBuilder),y=function(e){function t(t,n,r,i,o,a,s){var u=e.call(this,t,n,r,i)||this;return u.elementRef=t,u.styleBuilder=n,u.styler=r,u.marshal=i,u.layoutConfig=o,u.platformId=a,u.serverModuleLoaded=s,u.DIRECTIVE_KEY="show-hide",u.display="",u.hasLayout=!1,u.hasFlexChild=!1,u}return Object(r.__extends)(t,e),t.prototype.ngAfterViewInit=function(){this.trackExtraTriggers();for(var e=Array.from(this.nativeElement.children),t=0;t<e.length;t++)if(this.marshal.hasValue(e[t],"flex")){this.hasFlexChild=!0;break}b.has(this.nativeElement)?this.display=b.get(this.nativeElement):(this.display=this.getDisplayStyle(),b.set(this.nativeElement,this.display)),this.init();var n=this.marshal.getValue(this.nativeElement,this.DIRECTIVE_KEY,"");void 0===n||""===n?this.setValue(!0,""):this.triggerUpdate()},t.prototype.ngOnChanges=function(e){var t=this;Object.keys(e).forEach(function(n){if(-1!==t.inputs.indexOf(n)){var r=n.split("."),i=r.slice(1).join("."),o=e[n].currentValue,a=""===o||0!==o&&Object(s.coerceBooleanProperty)(o);"fxHide"===r[0]&&(a=!a),t.setValue(a,i)}})},t.prototype.trackExtraTriggers=function(){var e=this;this.hasLayout=this.marshal.hasValue(this.nativeElement,"layout"),["layout","layout-align"].forEach(function(t){e.marshal.trackValue(e.nativeElement,t).pipe(Object(u.takeUntil)(e.destroySubject)).subscribe(e.triggerUpdate.bind(e))})},t.prototype.getDisplayStyle=function(){return this.hasLayout||this.hasFlexChild&&this.layoutConfig.addFlexToParent?"flex":this.styler.lookupStyle(this.nativeElement,"display",!0)},t.prototype.updateWithValue=function(e){void 0===e&&(e=!0),""!==e&&(this.addStyles(e?"true":"false",{display:this.display}),Object(o.isPlatformServer)(this.platformId)&&this.serverModuleLoaded&&this.nativeElement.style.setProperty("display",""))},t.ctorParameters=function(){return[{type:i.ElementRef},{type:_},{type:a.StyleUtils},{type:a.MediaMarshaller},{type:void 0,decorators:[{type:i.Inject,args:[a.LAYOUT_CONFIG]}]},{type:Object,decorators:[{type:i.Inject,args:[i.PLATFORM_ID]}]},{type:Boolean,decorators:[{type:i.Optional},{type:i.Inject,args:[a.SERVER_TOKEN]}]}]},t}(a.BaseDirective2),b=new WeakMap,j=["fxShow","fxShow.print","fxShow.xs","fxShow.sm","fxShow.md","fxShow.lg","fxShow.xl","fxShow.lt-sm","fxShow.lt-md","fxShow.lt-lg","fxShow.lt-xl","fxShow.gt-xs","fxShow.gt-sm","fxShow.gt-md","fxShow.gt-lg","fxHide","fxHide.print","fxHide.xs","fxHide.sm","fxHide.md","fxHide.lg","fxHide.xl","fxHide.lt-sm","fxHide.lt-md","fxHide.lt-lg","fxHide.lt-xl","fxHide.gt-xs","fxHide.gt-sm","fxHide.gt-md","fxHide.gt-lg"],x=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=j,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [fxShow], [fxShow.print],\n  [fxShow.xs], [fxShow.sm], [fxShow.md], [fxShow.lg], [fxShow.xl],\n  [fxShow.lt-sm], [fxShow.lt-md], [fxShow.lt-lg], [fxShow.lt-xl],\n  [fxShow.gt-xs], [fxShow.gt-sm], [fxShow.gt-md], [fxShow.gt-lg],\n  [fxHide], [fxHide.print],\n  [fxHide.xs], [fxHide.sm], [fxHide.md], [fxHide.lg], [fxHide.xl],\n  [fxHide.lt-sm], [fxHide.lt-md], [fxHide.lt-lg], [fxHide.lt-xl],\n  [fxHide.gt-xs], [fxHide.gt-sm], [fxHide.gt-md], [fxHide.gt-lg]\n",inputs:j}]}],t}(y),O=function(){return function(e,t,n){void 0===n&&(n=!0),this.key=e,this.value=t,this.key=n?e.replace(/['"]/g,"").trim():e.trim(),this.value=n?t.replace(/['"]/g,"").trim():t.trim(),this.value=this.value.replace(/;/,"")}}();function w(e){var t=typeof e;return"object"===t?e.constructor===Array?"array":e.constructor===Set?"set":"object":t}function C(e,t){var n,r=[];return"set"===w(e)?e.forEach(function(e){return r.push(e)}):Object.keys(e).forEach(function(t){r.push(t+":"+e[t])}),n=t,r.map(E).filter(function(e){return!!e}).map(function(e){return n&&(e.value=n(e.value)),e}).reduce(S,{})}function E(e){var t=e.split(":"),n=t[0],r=t.slice(1);return new O(n,r.join(":"))}function S(e,t){return t.key&&(e[t.key]=t.value),e}var k=function(e){function t(t,n,r,i,a,s,u,c,l){var d=e.call(this,t,null,n,r)||this;d.elementRef=t,d.styler=n,d.marshal=r,d.keyValueDiffers=i,d.renderer=a,d.sanitizer=s,d.ngStyleInstance=u,d.DIRECTIVE_KEY="ngStyle",d.ngStyleInstance||(d.ngStyleInstance=new o.NgStyle(d.keyValueDiffers,d.elementRef,d.renderer)),d.init();var p=d.nativeElement.getAttribute("style")||"";return d.fallbackStyles=d.buildStyleMap(p),d.isServer=c&&Object(o.isPlatformServer)(l),d}return Object(r.__extends)(t,e),t.prototype.updateWithValue=function(e){var t=this.buildStyleMap(e);this.ngStyleInstance.ngStyle=Object(r.__assign)({},this.fallbackStyles,t),this.isServer&&this.applyStyleToElement(t),this.ngStyleInstance.ngDoCheck()},t.prototype.clearStyles=function(){this.ngStyleInstance.ngStyle=this.fallbackStyles,this.ngStyleInstance.ngDoCheck()},t.prototype.buildStyleMap=function(e){var t,n=this,r=function(e){return n.sanitizer.sanitize(i.SecurityContext.STYLE,e)||""};if(e)switch(w(e)){case"string":return T((void 0===t&&(t=";"),String(e).trim().split(t).map(function(e){return e.trim()}).filter(function(e){return""!==e})),r);case"array":return T(e,r);case"set":default:return C(e,r)}return{}},t.prototype.ngDoCheck=function(){this.ngStyleInstance.ngDoCheck()},t.ctorParameters=function(){return[{type:i.ElementRef},{type:a.StyleUtils},{type:a.MediaMarshaller},{type:i.KeyValueDiffers},{type:i.Renderer2},{type:c.DomSanitizer},{type:o.NgStyle,decorators:[{type:i.Optional},{type:i.Self}]},{type:Boolean,decorators:[{type:i.Optional},{type:i.Inject,args:[a.SERVER_TOKEN]}]},{type:Object,decorators:[{type:i.Inject,args:[i.PLATFORM_ID]}]}]},t}(a.BaseDirective2),M=["ngStyle","ngStyle.xs","ngStyle.sm","ngStyle.md","ngStyle.lg","ngStyle.xl","ngStyle.lt-sm","ngStyle.lt-md","ngStyle.lt-lg","ngStyle.lt-xl","ngStyle.gt-xs","ngStyle.gt-sm","ngStyle.gt-md","ngStyle.gt-lg"],A=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=M,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [ngStyle],\n  [ngStyle.xs], [ngStyle.sm], [ngStyle.md], [ngStyle.lg], [ngStyle.xl],\n  [ngStyle.lt-sm], [ngStyle.lt-md], [ngStyle.lt-lg], [ngStyle.lt-xl],\n  [ngStyle.gt-xs], [ngStyle.gt-sm], [ngStyle.gt-md], [ngStyle.gt-lg]\n",inputs:M}]}],t}(k);function T(e,t){return e.map(E).filter(function(e){return!!e}).map(function(e){return t&&(e.value=t(e.value)),e}).reduce(S,{})}var D=[x,v,A,h],I=function(){function e(){}return e.decorators=[{type:i.NgModule,args:[{imports:[a.CoreModule],declarations:D.slice(),exports:D.slice()}]}],e}()},"./node_modules/@angular/flex-layout/esm5/flex-layout.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"VERSION",function(){return l}),n.d(t,"FlexLayoutModule",function(){return d});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/common/fesm5/common.js"),a=n("./node_modules/@angular/flex-layout/esm5/core.es5.js");n.d(t,"removeStyles",function(){return a.removeStyles}),n.d(t,"BROWSER_PROVIDER",function(){return a.BROWSER_PROVIDER}),n.d(t,"CLASS_NAME",function(){return a.CLASS_NAME}),n.d(t,"CoreModule",function(){return a.CoreModule}),n.d(t,"MediaChange",function(){return a.MediaChange}),n.d(t,"StylesheetMap",function(){return a.StylesheetMap}),n.d(t,"DEFAULT_CONFIG",function(){return a.DEFAULT_CONFIG}),n.d(t,"LAYOUT_CONFIG",function(){return a.LAYOUT_CONFIG}),n.d(t,"SERVER_TOKEN",function(){return a.SERVER_TOKEN}),n.d(t,"BREAKPOINT",function(){return a.BREAKPOINT}),n.d(t,"BaseDirective2",function(){return a.BaseDirective2}),n.d(t,"sortDescendingPriority",function(){return a.sortDescendingPriority}),n.d(t,"sortAscendingPriority",function(){return a.sortAscendingPriority}),n.d(t,"DEFAULT_BREAKPOINTS",function(){return a.DEFAULT_BREAKPOINTS}),n.d(t,"ScreenTypes",function(){return a.ScreenTypes}),n.d(t,"ORIENTATION_BREAKPOINTS",function(){return a.ORIENTATION_BREAKPOINTS}),n.d(t,"BreakPointRegistry",function(){return a.BreakPointRegistry}),n.d(t,"BREAKPOINTS",function(){return a.BREAKPOINTS}),n.d(t,"MatchMedia",function(){return a.MatchMedia}),n.d(t,"MockMatchMedia",function(){return a.MockMatchMedia}),n.d(t,"MockMediaQueryList",function(){return a.MockMediaQueryList}),n.d(t,"MockMatchMediaProvider",function(){return a.MockMatchMediaProvider}),n.d(t,"ServerMediaQueryList",function(){return a.ServerMediaQueryList}),n.d(t,"ServerMatchMedia",function(){return a.ServerMatchMedia}),n.d(t,"MediaObserver",function(){return a.MediaObserver}),n.d(t,"StyleUtils",function(){return a.StyleUtils}),n.d(t,"StyleBuilder",function(){return a.StyleBuilder}),n.d(t,"validateBasis",function(){return a.validateBasis}),n.d(t,"MediaMarshaller",function(){return a.MediaMarshaller}),n.d(t,"BREAKPOINT_PRINT",function(){return a.BREAKPOINT_PRINT}),n.d(t,"PrintHook",function(){return a.PrintHook});var s=n("./node_modules/@angular/flex-layout/esm5/extended.es5.js");n.d(t,"ExtendedModule",function(){return s.ExtendedModule}),n.d(t,"ClassDirective",function(){return s.ClassDirective}),n.d(t,"DefaultClassDirective",function(){return s.DefaultClassDirective}),n.d(t,"ImgSrcStyleBuilder",function(){return s.ImgSrcStyleBuilder}),n.d(t,"ImgSrcDirective",function(){return s.ImgSrcDirective}),n.d(t,"DefaultImgSrcDirective",function(){return s.DefaultImgSrcDirective}),n.d(t,"ShowHideStyleBuilder",function(){return s.ShowHideStyleBuilder}),n.d(t,"ShowHideDirective",function(){return s.ShowHideDirective}),n.d(t,"DefaultShowHideDirective",function(){return s.DefaultShowHideDirective}),n.d(t,"StyleDirective",function(){return s.StyleDirective}),n.d(t,"DefaultStyleDirective",function(){return s.DefaultStyleDirective});var u=n("./node_modules/@angular/flex-layout/esm5/flex.es5.js");n.d(t,"FlexModule",function(){return u.FlexModule}),n.d(t,"FlexStyleBuilder",function(){return u.FlexStyleBuilder}),n.d(t,"FlexDirective",function(){return u.FlexDirective}),n.d(t,"DefaultFlexDirective",function(){return u.DefaultFlexDirective}),n.d(t,"FlexAlignStyleBuilder",function(){return u.FlexAlignStyleBuilder}),n.d(t,"FlexAlignDirective",function(){return u.FlexAlignDirective}),n.d(t,"DefaultFlexAlignDirective",function(){return u.DefaultFlexAlignDirective}),n.d(t,"FlexFillStyleBuilder",function(){return u.FlexFillStyleBuilder}),n.d(t,"FlexFillDirective",function(){return u.FlexFillDirective}),n.d(t,"FlexOffsetStyleBuilder",function(){return u.FlexOffsetStyleBuilder}),n.d(t,"FlexOffsetDirective",function(){return u.FlexOffsetDirective}),n.d(t,"DefaultFlexOffsetDirective",function(){return u.DefaultFlexOffsetDirective}),n.d(t,"FlexOrderStyleBuilder",function(){return u.FlexOrderStyleBuilder}),n.d(t,"FlexOrderDirective",function(){return u.FlexOrderDirective}),n.d(t,"DefaultFlexOrderDirective",function(){return u.DefaultFlexOrderDirective}),n.d(t,"LayoutStyleBuilder",function(){return u.LayoutStyleBuilder}),n.d(t,"LayoutDirective",function(){return u.LayoutDirective}),n.d(t,"DefaultLayoutDirective",function(){return u.DefaultLayoutDirective}),n.d(t,"LayoutAlignStyleBuilder",function(){return u.LayoutAlignStyleBuilder}),n.d(t,"LayoutAlignDirective",function(){return u.LayoutAlignDirective}),n.d(t,"DefaultLayoutAlignDirective",function(){return u.DefaultLayoutAlignDirective}),n.d(t,"LayoutGapStyleBuilder",function(){return u.LayoutGapStyleBuilder}),n.d(t,"LayoutGapDirective",function(){return u.LayoutGapDirective}),n.d(t,"DefaultLayoutGapDirective",function(){return u.DefaultLayoutGapDirective});var c=n("./node_modules/@angular/flex-layout/esm5/grid.es5.js");n.d(t,"ɵf2",function(){return c["ɵf2"]}),n.d(t,"ɵe2",function(){return c["ɵe2"]}),n.d(t,"ɵd2",function(){return c["ɵd2"]}),n.d(t,"ɵi2",function(){return c["ɵi2"]}),n.d(t,"ɵh2",function(){return c["ɵh2"]}),n.d(t,"ɵg2",function(){return c["ɵg2"]}),n.d(t,"ɵl2",function(){return c["ɵl2"]}),n.d(t,"ɵk2",function(){return c["ɵk2"]}),n.d(t,"ɵj2",function(){return c["ɵj2"]}),n.d(t,"ɵo2",function(){return c["ɵo2"]}),n.d(t,"ɵn2",function(){return c["ɵn2"]}),n.d(t,"ɵm2",function(){return c["ɵm2"]}),n.d(t,"ɵr2",function(){return c["ɵr2"]}),n.d(t,"ɵq2",function(){return c["ɵq2"]}),n.d(t,"ɵp2",function(){return c["ɵp2"]}),n.d(t,"ɵu2",function(){return c["ɵu2"]}),n.d(t,"ɵt2",function(){return c["ɵt2"]}),n.d(t,"ɵs2",function(){return c["ɵs2"]}),n.d(t,"ɵx2",function(){return c["ɵx2"]}),n.d(t,"ɵw2",function(){return c["ɵw2"]}),n.d(t,"ɵv2",function(){return c["ɵv2"]}),n.d(t,"ɵba2",function(){return c["ɵba2"]}),n.d(t,"ɵz2",function(){return c["ɵz2"]}),n.d(t,"ɵy2",function(){return c["ɵy2"]}),n.d(t,"ɵc2",function(){return c["ɵc2"]}),n.d(t,"ɵb2",function(){return c["ɵb2"]}),n.d(t,"ɵa2",function(){return c["ɵa2"]}),n.d(t,"ɵbd2",function(){return c["ɵbd2"]}),n.d(t,"ɵbc2",function(){return c["ɵbc2"]}),n.d(t,"ɵbb2",function(){return c["ɵbb2"]}),n.d(t,"ɵbg2",function(){return c["ɵbg2"]}),n.d(t,"ɵbf2",function(){return c["ɵbf2"]}),n.d(t,"ɵbe2",function(){return c["ɵbe2"]}),n.d(t,"GridModule",function(){return c.GridModule});var l=new r.Version("7.0.0-beta.23"),d=function(){function e(e,t){Object(o.isPlatformServer)(t)&&!e&&console.warn("Warning: Flex Layout loaded on the server without FlexLayoutServerModule")}return e.withConfig=function(t,n){return void 0===n&&(n=[]),{ngModule:e,providers:t.serverLoaded?[{provide:a.LAYOUT_CONFIG,useValue:Object(i.__assign)({},a.DEFAULT_CONFIG,t)},{provide:a.BREAKPOINT,useValue:n,multi:!0},{provide:a.SERVER_TOKEN,useValue:!0}]:[{provide:a.LAYOUT_CONFIG,useValue:Object(i.__assign)({},a.DEFAULT_CONFIG,t)},{provide:a.BREAKPOINT,useValue:n,multi:!0}]}},e.decorators=[{type:r.NgModule,args:[{imports:[u.FlexModule,s.ExtendedModule,c.GridModule],exports:[u.FlexModule,s.ExtendedModule,c.GridModule]}]}],e.ctorParameters=function(){return[{type:Boolean,decorators:[{type:r.Optional},{type:r.Inject,args:[a.SERVER_TOKEN]}]},{type:Object,decorators:[{type:r.Inject,args:[r.PLATFORM_ID]}]}]},e}()},"./node_modules/@angular/flex-layout/esm5/flex.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"FlexModule",function(){return he}),n.d(t,"FlexStyleBuilder",function(){return M}),n.d(t,"FlexDirective",function(){return T}),n.d(t,"DefaultFlexDirective",function(){return D}),n.d(t,"FlexAlignStyleBuilder",function(){return X}),n.d(t,"FlexAlignDirective",function(){return $}),n.d(t,"DefaultFlexAlignDirective",function(){return ee}),n.d(t,"FlexFillStyleBuilder",function(){return ne}),n.d(t,"FlexFillDirective",function(){return re}),n.d(t,"FlexOffsetStyleBuilder",function(){return U}),n.d(t,"FlexOffsetDirective",function(){return q}),n.d(t,"DefaultFlexOffsetDirective",function(){return G}),n.d(t,"FlexOrderStyleBuilder",function(){return F}),n.d(t,"FlexOrderDirective",function(){return V}),n.d(t,"DefaultFlexOrderDirective",function(){return z}),n.d(t,"LayoutStyleBuilder",function(){return f}),n.d(t,"LayoutDirective",function(){return m}),n.d(t,"DefaultLayoutDirective",function(){return g}),n.d(t,"LayoutAlignStyleBuilder",function(){return oe}),n.d(t,"LayoutAlignDirective",function(){return se}),n.d(t,"DefaultLayoutAlignDirective",function(){return ue}),n.d(t,"LayoutGapStyleBuilder",function(){return y}),n.d(t,"LayoutGapDirective",function(){return j}),n.d(t,"DefaultLayoutGapDirective",function(){return x});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/flex-layout/esm5/core.es5.js"),a=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),s=n("./node_modules/rxjs/_esm5/index.js"),u=n("./node_modules/rxjs/_esm5/operators/index.js"),c="inline",l=["row","column","row-reverse","column-reverse"];function d(e){var t=(e=e?e.toLowerCase():"").split(" "),n=t[0],r=t[1],i=t[2];return l.find(function(e){return e===n})||(n=l[0]),r===c&&(r=i!==c?i:"",i=c),[n,function(e){if(e)switch(e.toLowerCase()){case"reverse":case"wrap-reverse":case"reverse-wrap":e="wrap-reverse";break;case"no":case"none":case"nowrap":e="nowrap";break;default:e="wrap"}return e}(r),!!i]}function p(e){return d(e)[0].indexOf("row")>-1}var f=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e){return t=d(e),n=t[0],r=t[1],i=t[2],void 0===r&&(r=null),void 0===i&&(i=!1),{display:i?"inline-flex":"flex","box-sizing":"border-box","flex-direction":n,"flex-wrap":r||null};var t,n,r,i},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(o.StyleBuilder),h=["fxLayout","fxLayout.xs","fxLayout.sm","fxLayout.md","fxLayout.lg","fxLayout.xl","fxLayout.lt-sm","fxLayout.lt-md","fxLayout.lt-lg","fxLayout.lt-xl","fxLayout.gt-xs","fxLayout.gt-sm","fxLayout.gt-md","fxLayout.gt-lg"],m=function(e){function t(t,n,r,i){var o=e.call(this,t,r,n,i)||this;return o.elRef=t,o.styleUtils=n,o.styleBuilder=r,o.marshal=i,o.DIRECTIVE_KEY="layout",o.styleCache=v,o.init(),o}return Object(r.__extends)(t,e),t.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:f,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller}]},t}(o.BaseDirective2),g=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=h,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [fxLayout], [fxLayout.xs], [fxLayout.sm], [fxLayout.md],\n  [fxLayout.lg], [fxLayout.xl], [fxLayout.lt-sm], [fxLayout.lt-md],\n  [fxLayout.lt-lg], [fxLayout.lt-xl], [fxLayout.gt-xs], [fxLayout.gt-sm],\n  [fxLayout.gt-md], [fxLayout.gt-lg]\n",inputs:h}]}],t}(m),v=new Map,_={"margin-left":null,"margin-right":null,"margin-top":null,"margin-bottom":null},y=function(e){function t(t){var n=e.call(this)||this;return n._styler=t,n}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e,t){return e.endsWith(S)?function(e,t){var n="0px",r="0px";"rtl"===t?r="-"+e:n="-"+e;return{margin:"0px "+n+" -"+e+" "+r}}(e=e.slice(0,e.indexOf(S)),t.directionality):{}},t.prototype.sideEffect=function(e,t,n){var i=n.items;if(e.endsWith(S)){var o=function(e,t){var n="0px",r="0px";"rtl"===t?r=e:n=e;return{padding:"0px "+n+" "+e+" "+r}}(e=e.slice(0,e.indexOf(S)),n.directionality);this._styler.applyStyleToElements(o,n.items)}else{var a=i.pop(),s=function(e,t){var n,i=Object(r.__assign)({},_);switch(t.layout){case"column":n="margin-bottom";break;case"column-reverse":n="margin-top";break;case"row":n="rtl"===t.directionality?"margin-left":"margin-right";break;case"row-reverse":n="rtl"===t.directionality?"margin-right":"margin-left";break;default:n="rtl"===t.directionality?"margin-left":"margin-right"}return i[n]=e,i}(e,n);this._styler.applyStyleToElements(s,i),this._styler.applyStyleToElements(_,[a])}},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:o.StyleUtils}]},t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t(Object(i.inject)(o.StyleUtils))},token:t,providedIn:"root"}),t}(o.StyleBuilder),b=["fxLayoutGap","fxLayoutGap.xs","fxLayoutGap.sm","fxLayoutGap.md","fxLayoutGap.lg","fxLayoutGap.xl","fxLayoutGap.lt-sm","fxLayoutGap.lt-md","fxLayoutGap.lt-lg","fxLayoutGap.lt-xl","fxLayoutGap.gt-xs","fxLayoutGap.gt-sm","fxLayoutGap.gt-md","fxLayoutGap.gt-lg"],j=function(e){function t(t,n,r,i,o,a){var c=e.call(this,t,o,i,a)||this;c.elRef=t,c.zone=n,c.directionality=r,c.styleUtils=i,c.styleBuilder=o,c.marshal=a,c.layout="row",c.DIRECTIVE_KEY="layout-gap",c.observerSubject=new s.Subject;var l=[c.directionality.change,c.observerSubject.asObservable()];return c.init(l),c.marshal.trackValue(c.nativeElement,"layout").pipe(Object(u.takeUntil)(c.destroySubject)).subscribe(c.onLayoutChange.bind(c)),c}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"childrenNodes",{get:function(){for(var e=this.nativeElement.children,t=[],n=e.length;n--;)t[n]=e[n];return t},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){this.buildChildObservable(),this.triggerUpdate()},t.prototype.ngOnDestroy=function(){e.prototype.ngOnDestroy.call(this),this.observer&&this.observer.disconnect()},t.prototype.onLayoutChange=function(e){var t=this,n=e.value.split(" ");this.layout=n[0],l.find(function(e){return e===t.layout})||(this.layout="row"),this.triggerUpdate()},t.prototype.updateWithValue=function(e){var t=this,n=this.childrenNodes.filter(function(e){return 1===e.nodeType&&t.willDisplay(e)}).sort(function(e,n){var r=+t.styler.lookupStyle(e,"order"),i=+t.styler.lookupStyle(n,"order");return isNaN(r)||isNaN(i)||r===i?0:r>i?1:-1});if(n.length>0){var r=this.directionality.value,i=this.layout;"row"===i&&"rtl"===r?this.styleCache=O:"row"===i&&"rtl"!==r?this.styleCache=C:"column"===i&&"rtl"===r?this.styleCache=w:"column"===i&&"rtl"!==r&&(this.styleCache=E),this.addStyles(e,{directionality:r,items:n,layout:i})}},t.prototype.willDisplay=function(e){var t=this.marshal.getValue(e,"show-hide");return!0===t||void 0===t&&"none"!==this.styleUtils.lookupStyle(e,"display")},t.prototype.buildChildObservable=function(){var e=this;this.zone.runOutsideAngular(function(){"undefined"!=typeof MutationObserver&&(e.observer=new MutationObserver(function(t){t.some(function(e){return e.addedNodes&&e.addedNodes.length>0||e.removedNodes&&e.removedNodes.length>0})&&e.observerSubject.next()}),e.observer.observe(e.nativeElement,{childList:!0}))})},t.ctorParameters=function(){return[{type:i.ElementRef},{type:i.NgZone},{type:a.Directionality},{type:o.StyleUtils},{type:y,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller}]},t}(o.BaseDirective2),x=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=b,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [fxLayoutGap], [fxLayoutGap.xs], [fxLayoutGap.sm], [fxLayoutGap.md],\n  [fxLayoutGap.lg], [fxLayoutGap.xl], [fxLayoutGap.lt-sm], [fxLayoutGap.lt-md],\n  [fxLayoutGap.lt-lg], [fxLayoutGap.lt-xl], [fxLayoutGap.gt-xs], [fxLayoutGap.gt-sm],\n  [fxLayoutGap.gt-md], [fxLayoutGap.gt-lg]\n",inputs:b}]}],t}(j),O=new Map,w=new Map,C=new Map,E=new Map,S=" grid";function k(e){for(var t=[],n=1;n<arguments.length;n++)t[n-1]=arguments[n];if(null==e)throw TypeError("Cannot convert undefined or null to object");for(var r=0,i=t;r<i.length;r++){var o=i[r];if(null!=o)for(var a in o)o.hasOwnProperty(a)&&(e[a]=o[a])}return e}var M=function(e){function t(t){var n=e.call(this)||this;return n.layoutConfig=t,n}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e,t){var n=e.split(" "),r=n[0],i=n[1],o=n.slice(2).join(" "),a=t.direction.indexOf("column")>-1?"column":"row",s=p(a)?"max-width":"max-height",u=p(a)?"min-width":"min-height",c=String(o).indexOf("calc")>-1,l=c||"auto"===o,d=String(o).indexOf("%")>-1&&!c,f=String(o).indexOf("px")>-1||String(o).indexOf("rem")>-1||String(o).indexOf("em")>-1||String(o).indexOf("vw")>-1||String(o).indexOf("vh")>-1,h=c||f;r="0"==r?0:r,i="0"==i?0:i;var m=!r&&!i,g={},v={"max-width":null,"max-height":null,"min-width":null,"min-height":null};switch(o||""){case"":var _=!1!==this.layoutConfig.useColumnBasisZero;o="row"===a?"0%":_?"0.000000001px":"auto";break;case"initial":case"nogrow":r=0,o="auto";break;case"grow":o="100%";break;case"noshrink":i=0,o="auto";break;case"auto":break;case"none":r=0,i=0,o="auto";break;default:h||d||isNaN(o)||(o+="%"),"0%"===o&&(h=!0),"0px"===o&&(o="0%"),g=k(v,c?{"flex-grow":r,"flex-shrink":i,"flex-basis":h?o:"100%"}:{flex:r+" "+i+" "+(h?o:"100%")})}return g.flex||g["flex-grow"]||(g=k(v,c?{"flex-grow":r,"flex-shrink":i,"flex-basis":o}:{flex:r+" "+i+" "+o})),"0%"!==o&&"0px"!==o&&"0.000000001px"!==o&&"auto"!==o&&(g[u]=m||h&&r?o:null,g[s]=m||!l&&i?o:null),g[u]||g[s]?t.hasWrap&&(g[c?"flex-basis":"flex"]=g[s]?c?g[s]:r+" "+i+" "+g[s]:c?g[u]:r+" "+i+" "+g[u]):g=k(v,c?{"flex-grow":r,"flex-shrink":i,"flex-basis":o}:{flex:r+" "+i+" "+o}),k(g,{"box-sizing":"border-box"})},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:i.Inject,args:[o.LAYOUT_CONFIG]}]}]},t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t(Object(i.inject)(o.LAYOUT_CONFIG))},token:t,providedIn:"root"}),t}(o.StyleBuilder),A=["fxFlex","fxFlex.xs","fxFlex.sm","fxFlex.md","fxFlex.lg","fxFlex.xl","fxFlex.lt-sm","fxFlex.lt-md","fxFlex.lt-lg","fxFlex.lt-xl","fxFlex.gt-xs","fxFlex.gt-sm","fxFlex.gt-md","fxFlex.gt-lg"],T=function(e){function t(t,n,r,i,o){var a=e.call(this,t,i,n,o)||this;return a.elRef=t,a.styleUtils=n,a.layoutConfig=r,a.styleBuilder=i,a.marshal=o,a.DIRECTIVE_KEY="flex",a.direction="",a.wrap=!1,a.flexGrow="1",a.flexShrink="1",a.init(),a.parentElement&&a.marshal.trackValue(a.parentElement,"layout").pipe(Object(u.takeUntil)(a.destroySubject)).subscribe(a.onLayoutChange.bind(a)),a}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"shrink",{get:function(){return this.flexShrink},set:function(e){this.flexShrink=e||"1",this.triggerReflow()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"grow",{get:function(){return this.flexGrow},set:function(e){this.flexGrow=e||"1",this.triggerReflow()},enumerable:!0,configurable:!0}),t.prototype.onLayoutChange=function(e){var t=e.value.split(" ");this.direction=t[0],this.wrap=void 0!==t[1]&&"wrap"===t[1],this.triggerUpdate()},t.prototype.updateWithValue=function(e){var t=!1!==this.layoutConfig.addFlexToParent;this.direction||(this.direction=this.getFlexFlowDirection(this.parentElement,t));var n=this.direction,r=n.startsWith("row"),i=this.wrap;r&&i?this.styleCache=R:r&&!i?this.styleCache=I:!r&&i?this.styleCache=N:r||i||(this.styleCache=P);var a=String(e).replace(";",""),s=Object(o.validateBasis)(a,this.flexGrow,this.flexShrink);this.addStyles(s.join(" "),{direction:n,hasWrap:i})},t.prototype.triggerReflow=function(){var e=this.activatedValue;if(void 0!==e){var t=Object(o.validateBasis)(e,this.flexGrow,this.flexShrink);this.marshal.updateElement(this.nativeElement,this.DIRECTIVE_KEY,t.join(" "))}},t.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:void 0,decorators:[{type:i.Inject,args:[o.LAYOUT_CONFIG]}]},{type:M},{type:o.MediaMarshaller}]},t.propDecorators={shrink:[{type:i.Input,args:["fxShrink"]}],grow:[{type:i.Input,args:["fxGrow"]}]},t}(o.BaseDirective2),D=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=A,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{inputs:A,selector:"\n  [fxFlex], [fxFlex.xs], [fxFlex.sm], [fxFlex.md],\n  [fxFlex.lg], [fxFlex.xl], [fxFlex.lt-sm], [fxFlex.lt-md],\n  [fxFlex.lt-lg], [fxFlex.lt-xl], [fxFlex.gt-xs], [fxFlex.gt-sm],\n  [fxFlex.gt-md], [fxFlex.gt-lg]\n"}]}],t}(T),I=new Map,P=new Map,R=new Map,N=new Map,F=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e){return{order:e&&parseInt(e,10)||""}},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(o.StyleBuilder),L=["fxFlexOrder","fxFlexOrder.xs","fxFlexOrder.sm","fxFlexOrder.md","fxFlexOrder.lg","fxFlexOrder.xl","fxFlexOrder.lt-sm","fxFlexOrder.lt-md","fxFlexOrder.lt-lg","fxFlexOrder.lt-xl","fxFlexOrder.gt-xs","fxFlexOrder.gt-sm","fxFlexOrder.gt-md","fxFlexOrder.gt-lg"],V=function(e){function t(t,n,r,i){var o=e.call(this,t,r,n,i)||this;return o.elRef=t,o.styleUtils=n,o.styleBuilder=r,o.marshal=i,o.DIRECTIVE_KEY="flex-order",o.styleCache=B,o.init(),o}return Object(r.__extends)(t,e),t.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:F,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller}]},t}(o.BaseDirective2),B=new Map,z=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=L,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [fxFlexOrder], [fxFlexOrder.xs], [fxFlexOrder.sm], [fxFlexOrder.md],\n  [fxFlexOrder.lg], [fxFlexOrder.xl], [fxFlexOrder.lt-sm], [fxFlexOrder.lt-md],\n  [fxFlexOrder.lt-lg], [fxFlexOrder.lt-xl], [fxFlexOrder.gt-xs], [fxFlexOrder.gt-sm],\n  [fxFlexOrder.gt-md], [fxFlexOrder.gt-lg]\n",inputs:L}]}],t}(V),U=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e,t){var n;""===e&&(e="0");var r=String(e).indexOf("%")>-1;String(e).indexOf("px")>-1||r||isNaN(+e)||(e+="%");var i=t.isRtl?"margin-right":"margin-left";return p(t.layout)?((n={})[i]=""+e,n):{"margin-top":""+e}},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(o.StyleBuilder),H=["fxFlexOffset","fxFlexOffset.xs","fxFlexOffset.sm","fxFlexOffset.md","fxFlexOffset.lg","fxFlexOffset.xl","fxFlexOffset.lt-sm","fxFlexOffset.lt-md","fxFlexOffset.lt-lg","fxFlexOffset.lt-xl","fxFlexOffset.gt-xs","fxFlexOffset.gt-sm","fxFlexOffset.gt-md","fxFlexOffset.gt-lg"],q=function(e){function t(t,n,r,i,o){var a=e.call(this,t,r,o,i)||this;return a.elRef=t,a.directionality=n,a.styleBuilder=r,a.marshal=i,a.styler=o,a.DIRECTIVE_KEY="flex-offset",a.init([a.directionality.change]),a.parentElement&&a.marshal.trackValue(a.parentElement,"layout-gap").pipe(Object(u.takeUntil)(a.destroySubject)).subscribe(a.triggerUpdate.bind(a)),a}return Object(r.__extends)(t,e),t.prototype.updateWithValue=function(e){void 0===e&&(e="");var t=this.getFlexFlowDirection(this.parentElement,!0),n="rtl"===this.directionality.value;"row"===t&&n?this.styleCache=W:"row"!==t||n?"column"===t&&n?this.styleCache=Y:"column"!==t||n||(this.styleCache=Q):this.styleCache=K,this.addStyles(e+"",{layout:t,isRtl:n})},t.ctorParameters=function(){return[{type:i.ElementRef},{type:a.Directionality},{type:U,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller},{type:o.StyleUtils}]},t}(o.BaseDirective2),G=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=H,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [fxFlexOffset], [fxFlexOffset.xs], [fxFlexOffset.sm], [fxFlexOffset.md],\n  [fxFlexOffset.lg], [fxFlexOffset.xl], [fxFlexOffset.lt-sm], [fxFlexOffset.lt-md],\n  [fxFlexOffset.lt-lg], [fxFlexOffset.lt-xl], [fxFlexOffset.gt-xs], [fxFlexOffset.gt-sm],\n  [fxFlexOffset.gt-md], [fxFlexOffset.gt-lg]\n",inputs:H}]}],t}(q),W=new Map,Y=new Map,K=new Map,Q=new Map,X=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e){var t={};switch(e=e||"stretch"){case"start":t["align-self"]="flex-start";break;case"end":t["align-self"]="flex-end";break;default:t["align-self"]=e}return t},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(o.StyleBuilder),Z=["fxFlexAlign","fxFlexAlign.xs","fxFlexAlign.sm","fxFlexAlign.md","fxFlexAlign.lg","fxFlexAlign.xl","fxFlexAlign.lt-sm","fxFlexAlign.lt-md","fxFlexAlign.lt-lg","fxFlexAlign.lt-xl","fxFlexAlign.gt-xs","fxFlexAlign.gt-sm","fxFlexAlign.gt-md","fxFlexAlign.gt-lg"],$=function(e){function t(t,n,r,i){var o=e.call(this,t,r,n,i)||this;return o.elRef=t,o.styleUtils=n,o.styleBuilder=r,o.marshal=i,o.DIRECTIVE_KEY="flex-align",o.styleCache=J,o.init(),o}return Object(r.__extends)(t,e),t.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:X,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller}]},t}(o.BaseDirective2),J=new Map,ee=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=Z,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [fxFlexAlign], [fxFlexAlign.xs], [fxFlexAlign.sm], [fxFlexAlign.md],\n  [fxFlexAlign.lg], [fxFlexAlign.xl], [fxFlexAlign.lt-sm], [fxFlexAlign.lt-md],\n  [fxFlexAlign.lt-lg], [fxFlexAlign.lt-xl], [fxFlexAlign.gt-xs], [fxFlexAlign.gt-sm],\n  [fxFlexAlign.gt-md], [fxFlexAlign.gt-lg]\n",inputs:Z}]}],t}($),te={margin:0,width:"100%",height:"100%","min-width":"100%","min-height":"100%"},ne=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e){return te},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(o.StyleBuilder),re=function(e){function t(t,n,r,i){var o=e.call(this,t,r,n,i)||this;return o.elRef=t,o.styleUtils=n,o.styleBuilder=r,o.marshal=i,o.styleCache=ie,o.addStyles(""),o}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"[fxFill], [fxFlexFill]"}]}],t.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:ne},{type:o.MediaMarshaller}]},t}(o.BaseDirective2),ie=new Map,oe=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e,t){var n={},r=e.split(" "),i=r[0],o=r[1];switch(i){case"center":n["justify-content"]="center";break;case"space-around":n["justify-content"]="space-around";break;case"space-between":n["justify-content"]="space-between";break;case"space-evenly":n["justify-content"]="space-evenly";break;case"end":case"flex-end":n["justify-content"]="flex-end";break;case"start":case"flex-start":default:n["justify-content"]="flex-start"}switch(o){case"start":case"flex-start":n["align-items"]=n["align-content"]="flex-start";break;case"center":n["align-items"]=n["align-content"]="center";break;case"end":case"flex-end":n["align-items"]=n["align-content"]="flex-end";break;case"space-between":n["align-content"]="space-between",n["align-items"]="stretch";break;case"space-around":n["align-content"]="space-around",n["align-items"]="stretch";break;case"baseline":n["align-content"]="stretch",n["align-items"]="baseline";break;case"stretch":default:n["align-items"]=n["align-content"]="stretch"}return k(n,{display:"flex","flex-direction":t.layout,"box-sizing":"border-box","max-width":"stretch"===o?p(t.layout)?null:"100%":null,"max-height":"stretch"===o&&p(t.layout)?"100%":null})},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(o.StyleBuilder),ae=["fxLayoutAlign","fxLayoutAlign.xs","fxLayoutAlign.sm","fxLayoutAlign.md","fxLayoutAlign.lg","fxLayoutAlign.xl","fxLayoutAlign.lt-sm","fxLayoutAlign.lt-md","fxLayoutAlign.lt-lg","fxLayoutAlign.lt-xl","fxLayoutAlign.gt-xs","fxLayoutAlign.gt-sm","fxLayoutAlign.gt-md","fxLayoutAlign.gt-lg"],se=function(e){function t(t,n,r,i){var o=e.call(this,t,r,n,i)||this;return o.elRef=t,o.styleUtils=n,o.styleBuilder=r,o.marshal=i,o.DIRECTIVE_KEY="layout-align",o.layout="row",o.init(),o.marshal.trackValue(o.nativeElement,"layout").pipe(Object(u.takeUntil)(o.destroySubject)).subscribe(o.onLayoutChange.bind(o)),o}return Object(r.__extends)(t,e),t.prototype.updateWithValue=function(e){var t=this.layout||"row";"row"===t?this.styleCache=ce:"row-reverse"===t?this.styleCache=de:"column"===t?this.styleCache=le:"column-reverse"===t&&(this.styleCache=pe),this.addStyles(e,{layout:t})},t.prototype.onLayoutChange=function(e){var t=this,n=e.value;this.layout=n.split(" ")[0],l.find(function(e){return e===t.layout})||(this.layout="row"),this.triggerUpdate()},t.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:oe,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller}]},t}(o.BaseDirective2),ue=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=ae,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [fxLayoutAlign], [fxLayoutAlign.xs], [fxLayoutAlign.sm], [fxLayoutAlign.md],\n  [fxLayoutAlign.lg], [fxLayoutAlign.xl], [fxLayoutAlign.lt-sm], [fxLayoutAlign.lt-md],\n  [fxLayoutAlign.lt-lg], [fxLayoutAlign.lt-xl], [fxLayoutAlign.gt-xs], [fxLayoutAlign.gt-sm],\n  [fxLayoutAlign.gt-md], [fxLayoutAlign.gt-lg]\n",inputs:ae}]}],t}(se),ce=new Map,le=new Map,de=new Map,pe=new Map,fe=[g,x,ue,z,G,re,ee,D],he=function(){function e(){}return e.decorators=[{type:i.NgModule,args:[{imports:[o.CoreModule,a.BidiModule],declarations:fe.slice(),exports:fe.slice()}]}],e}()},"./node_modules/@angular/flex-layout/esm5/grid.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"GridModule",function(){return xe}),n.d(t,"ɵf2",function(){return j}),n.d(t,"ɵe2",function(){return v}),n.d(t,"ɵd2",function(){return g}),n.d(t,"ɵi2",function(){return M}),n.d(t,"ɵh2",function(){return C}),n.d(t,"ɵg2",function(){return w}),n.d(t,"ɵl2",function(){return P}),n.d(t,"ɵk2",function(){return T}),n.d(t,"ɵj2",function(){return A}),n.d(t,"ɵo2",function(){return B}),n.d(t,"ɵn2",function(){return N}),n.d(t,"ɵm2",function(){return R}),n.d(t,"ɵr2",function(){return W}),n.d(t,"ɵq2",function(){return U}),n.d(t,"ɵp2",function(){return z}),n.d(t,"ɵu2",function(){return Z}),n.d(t,"ɵt2",function(){return K}),n.d(t,"ɵs2",function(){return Y}),n.d(t,"ɵx2",function(){return re}),n.d(t,"ɵw2",function(){return J}),n.d(t,"ɵv2",function(){return $}),n.d(t,"ɵba2",function(){return ce}),n.d(t,"ɵz2",function(){return oe}),n.d(t,"ɵy2",function(){return ie}),n.d(t,"ɵc2",function(){return f}),n.d(t,"ɵb2",function(){return l}),n.d(t,"ɵa2",function(){return c}),n.d(t,"ɵbd2",function(){return he}),n.d(t,"ɵbc2",function(){return de}),n.d(t,"ɵbb2",function(){return le}),n.d(t,"ɵbg2",function(){return be}),n.d(t,"ɵbf2",function(){return ge}),n.d(t,"ɵbe2",function(){return me});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/flex-layout/esm5/core.es5.js"),a=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),s="stretch",u="stretch",c=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e){return function(e){void 0===e&&(e="");var t={},n=e.split(" "),r=n[0],i=n[1];switch(r){case"end":t["justify-self"]="end";break;case"center":t["justify-self"]="center";break;case"stretch":t["justify-self"]="stretch";break;case"start":t["justify-self"]="start";break;default:t["justify-self"]=s}switch(i){case"end":t["align-self"]="end";break;case"center":t["align-self"]="center";break;case"stretch":t["align-self"]="stretch";break;case"start":t["align-self"]="start";break;default:t["align-self"]=u}return t}(e||s)},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(o.StyleBuilder),l=function(e){function t(t,n,r,i){var o=e.call(this,t,n,r,i)||this;return o.elementRef=t,o.styleBuilder=n,o.styler=r,o.marshal=i,o.DIRECTIVE_KEY="grid-align",o.styleCache=d,o.init(),o}return Object(r.__extends)(t,e),t.ctorParameters=function(){return[{type:i.ElementRef},{type:c,decorators:[{type:i.Optional}]},{type:o.StyleUtils},{type:o.MediaMarshaller}]},t}(o.BaseDirective2),d=new Map,p=["gdGridAlign","gdGridAlign.xs","gdGridAlign.sm","gdGridAlign.md","gdGridAlign.lg","gdGridAlign.xl","gdGridAlign.lt-sm","gdGridAlign.lt-md","gdGridAlign.lt-lg","gdGridAlign.lt-xl","gdGridAlign.gt-xs","gdGridAlign.gt-sm","gdGridAlign.gt-md","gdGridAlign.gt-lg"],f=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=p,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [gdGridAlign],\n  [gdGridAlign.xs], [gdGridAlign.sm], [gdGridAlign.md], [gdGridAlign.lg],[gdGridAlign.xl],\n  [gdGridAlign.lt-sm], [gdGridAlign.lt-md], [gdGridAlign.lt-lg], [gdGridAlign.lt-xl],\n  [gdGridAlign.gt-xs], [gdGridAlign.gt-sm], [gdGridAlign.gt-md], [gdGridAlign.gt-lg]\n",inputs:p}]}],t}(l);var h="start",m="stretch",g=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e,t){return function(e,t){var n={},r=e.split(" "),i=r[0],o=r[1];switch(i){case"center":n["align-content"]="center";break;case"space-around":n["align-content"]="space-around";break;case"space-between":n["align-content"]="space-between";break;case"space-evenly":n["align-content"]="space-evenly";break;case"end":n["align-content"]="end";break;case"start":n["align-content"]="start";break;case"stretch":n["align-content"]="stretch";break;default:n["align-content"]=h}switch(o){case"start":n["align-items"]="start";break;case"center":n["align-items"]="center";break;case"end":n["align-items"]="end";break;case"stretch":n["align-items"]="stretch";break;default:n["align-items"]=m}return n.display=t?"inline-grid":"grid",n}(e||h+" "+m,t.inline)},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(o.StyleBuilder),v=function(e){function t(t,n,r,i){var o=e.call(this,t,n,r,i)||this;return o.elementRef=t,o.styleBuilder=n,o.styler=r,o.marshal=i,o.DIRECTIVE_KEY="grid-align-columns",o._inline=!1,o.init(),o}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"inline",{get:function(){return this._inline},set:function(e){this._inline=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t.prototype.updateWithValue=function(e){this.styleCache=this.inline?y:_,this.addStyles(e,{inline:this.inline})},t.ctorParameters=function(){return[{type:i.ElementRef},{type:g,decorators:[{type:i.Optional}]},{type:o.StyleUtils},{type:o.MediaMarshaller}]},t.propDecorators={inline:[{type:i.Input,args:["gdInline"]}]},t}(o.BaseDirective2),_=new Map,y=new Map,b=["gdAlignColumns","gdAlignColumns.xs","gdAlignColumns.sm","gdAlignColumns.md","gdAlignColumns.lg","gdAlignColumns.xl","gdAlignColumns.lt-sm","gdAlignColumns.lt-md","gdAlignColumns.lt-lg","gdAlignColumns.lt-xl","gdAlignColumns.gt-xs","gdAlignColumns.gt-sm","gdAlignColumns.gt-md","gdAlignColumns.gt-lg"],j=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=b,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [gdAlignColumns],\n  [gdAlignColumns.xs], [gdAlignColumns.sm], [gdAlignColumns.md],\n  [gdAlignColumns.lg], [gdAlignColumns.xl], [gdAlignColumns.lt-sm],\n  [gdAlignColumns.lt-md], [gdAlignColumns.lt-lg], [gdAlignColumns.lt-xl],\n  [gdAlignColumns.gt-xs], [gdAlignColumns.gt-sm], [gdAlignColumns.gt-md],\n  [gdAlignColumns.gt-lg]\n",inputs:b}]}],t}(v);var x="start",O="stretch",w=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e,t){return function(e,t){var n={},r=e.split(" "),i=r[0],o=r[1];switch(i){case"center":case"space-around":case"space-between":case"space-evenly":case"end":case"start":case"stretch":n["justify-content"]=i;break;default:n["justify-content"]=x}switch(o){case"start":case"center":case"end":case"stretch":n["justify-items"]=o;break;default:n["justify-items"]=O}return n.display=t?"inline-grid":"grid",n}(e||x+" "+O,t.inline)},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(o.StyleBuilder),C=function(e){function t(t,n,r,i){var o=e.call(this,t,n,r,i)||this;return o.elementRef=t,o.styleBuilder=n,o.styler=r,o.marshal=i,o.DIRECTIVE_KEY="grid-align-rows",o._inline=!1,o.init(),o}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"inline",{get:function(){return this._inline},set:function(e){this._inline=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t.prototype.updateWithValue=function(e){this.styleCache=this.inline?S:E,this.addStyles(e,{inline:this.inline})},t.ctorParameters=function(){return[{type:i.ElementRef},{type:w,decorators:[{type:i.Optional}]},{type:o.StyleUtils},{type:o.MediaMarshaller}]},t.propDecorators={inline:[{type:i.Input,args:["gdInline"]}]},t}(o.BaseDirective2),E=new Map,S=new Map,k=["gdAlignRows","gdAlignRows.xs","gdAlignRows.sm","gdAlignRows.md","gdAlignRows.lg","gdAlignRows.xl","gdAlignRows.lt-sm","gdAlignRows.lt-md","gdAlignRows.lt-lg","gdAlignRows.lt-xl","gdAlignRows.gt-xs","gdAlignRows.gt-sm","gdAlignRows.gt-md","gdAlignRows.gt-lg"],M=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=k,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [gdAlignRows],\n  [gdAlignRows.xs], [gdAlignRows.sm], [gdAlignRows.md],\n  [gdAlignRows.lg], [gdAlignRows.xl], [gdAlignRows.lt-sm],\n  [gdAlignRows.lt-md], [gdAlignRows.lt-lg], [gdAlignRows.lt-xl],\n  [gdAlignRows.gt-xs], [gdAlignRows.gt-sm], [gdAlignRows.gt-md],\n  [gdAlignRows.gt-lg]\n",inputs:k}]}],t}(C);var A=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e){return{"grid-area":e||"auto"}},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(o.StyleBuilder),T=function(e){function t(t,n,r,i){var o=e.call(this,t,r,n,i)||this;return o.elRef=t,o.styleUtils=n,o.styleBuilder=r,o.marshal=i,o.DIRECTIVE_KEY="grid-area",o.styleCache=D,o.init(),o}return Object(r.__extends)(t,e),t.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:A,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller}]},t}(o.BaseDirective2),D=new Map,I=["gdArea","gdArea.xs","gdArea.sm","gdArea.md","gdArea.lg","gdArea.xl","gdArea.lt-sm","gdArea.lt-md","gdArea.lt-lg","gdArea.lt-xl","gdArea.gt-xs","gdArea.gt-sm","gdArea.gt-md","gdArea.gt-lg"],P=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=I,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [gdArea],\n  [gdArea.xs], [gdArea.sm], [gdArea.md], [gdArea.lg], [gdArea.xl],\n  [gdArea.lt-sm], [gdArea.lt-md], [gdArea.lt-lg], [gdArea.lt-xl],\n  [gdArea.gt-xs], [gdArea.gt-sm], [gdArea.gt-md], [gdArea.gt-lg]\n",inputs:I}]}],t}(T),R=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e,t){var n=(e||"none").split("|").map(function(e){return'"'+e.trim()+'"'});return{display:t.inline?"inline-grid":"grid","grid-template-areas":n.join(" ")}},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(o.StyleBuilder),N=function(e){function t(t,n,r,i){var o=e.call(this,t,r,n,i)||this;return o.elRef=t,o.styleUtils=n,o.styleBuilder=r,o.marshal=i,o.DIRECTIVE_KEY="grid-areas",o._inline=!1,o.init(),o}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"inline",{get:function(){return this._inline},set:function(e){this._inline=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t.prototype.updateWithValue=function(e){this.styleCache=this.inline?L:F,this.addStyles(e,{inline:this.inline})},t.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:R,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller}]},t.propDecorators={inline:[{type:i.Input,args:["gdInline"]}]},t}(o.BaseDirective2),F=new Map,L=new Map,V=["gdAreas","gdAreas.xs","gdAreas.sm","gdAreas.md","gdAreas.lg","gdAreas.xl","gdAreas.lt-sm","gdAreas.lt-md","gdAreas.lt-lg","gdAreas.lt-xl","gdAreas.gt-xs","gdAreas.gt-sm","gdAreas.gt-md","gdAreas.gt-lg"],B=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=V,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [gdAreas],\n  [gdAreas.xs], [gdAreas.sm], [gdAreas.md], [gdAreas.lg], [gdAreas.xl],\n  [gdAreas.lt-sm], [gdAreas.lt-md], [gdAreas.lt-lg], [gdAreas.lt-xl],\n  [gdAreas.gt-xs], [gdAreas.gt-sm], [gdAreas.gt-md], [gdAreas.gt-lg]\n",inputs:V}]}],t}(N),z=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e,t){var n=(e||"initial").split(" "),r=n[0],i=n[1];return"column"!==r&&"row"!==r&&"dense"!==r&&(r="row"),i="dense"===i&&"dense"!==r?" dense":"",{display:t.inline?"inline-grid":"grid","grid-auto-flow":r+i}},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(o.StyleBuilder),U=function(e){function t(t,n,r,i){var o=e.call(this,t,n,r,i)||this;return o.elementRef=t,o.styleBuilder=n,o.styler=r,o.marshal=i,o._inline=!1,o.DIRECTIVE_KEY="grid-auto",o.init(),o}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"inline",{get:function(){return this._inline},set:function(e){this._inline=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t.prototype.updateWithValue=function(e){this.styleCache=this.inline?q:H,this.addStyles(e,{inline:this.inline})},t.ctorParameters=function(){return[{type:i.ElementRef},{type:z,decorators:[{type:i.Optional}]},{type:o.StyleUtils},{type:o.MediaMarshaller}]},t.propDecorators={inline:[{type:i.Input,args:["gdInline"]}]},t}(o.BaseDirective2),H=new Map,q=new Map,G=["gdAuto","gdAuto.xs","gdAuto.sm","gdAuto.md","gdAuto.lg","gdAuto.xl","gdAuto.lt-sm","gdAuto.lt-md","gdAuto.lt-lg","gdAuto.lt-xl","gdAuto.gt-xs","gdAuto.gt-sm","gdAuto.gt-md","gdAuto.gt-lg"],W=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=G,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [gdAuto],\n  [gdAuto.xs], [gdAuto.sm], [gdAuto.md], [gdAuto.lg], [gdAuto.xl],\n  [gdAuto.lt-sm], [gdAuto.lt-md], [gdAuto.lt-lg], [gdAuto.lt-xl],\n  [gdAuto.gt-xs], [gdAuto.gt-sm], [gdAuto.gt-md], [gdAuto.gt-lg]\n",inputs:G}]}],t}(U),Y=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e){return{"grid-column":e||"auto"}},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(o.StyleBuilder),K=function(e){function t(t,n,r,i){var o=e.call(this,t,n,r,i)||this;return o.elementRef=t,o.styleBuilder=n,o.styler=r,o.marshal=i,o.DIRECTIVE_KEY="grid-column",o.styleCache=Q,o.init(),o}return Object(r.__extends)(t,e),t.ctorParameters=function(){return[{type:i.ElementRef},{type:Y,decorators:[{type:i.Optional}]},{type:o.StyleUtils},{type:o.MediaMarshaller}]},t}(o.BaseDirective2),Q=new Map,X=["gdColumn","gdColumn.xs","gdColumn.sm","gdColumn.md","gdColumn.lg","gdColumn.xl","gdColumn.lt-sm","gdColumn.lt-md","gdColumn.lt-lg","gdColumn.lt-xl","gdColumn.gt-xs","gdColumn.gt-sm","gdColumn.gt-md","gdColumn.gt-lg"],Z=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=X,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [gdColumn],\n  [gdColumn.xs], [gdColumn.sm], [gdColumn.md], [gdColumn.lg], [gdColumn.xl],\n  [gdColumn.lt-sm], [gdColumn.lt-md], [gdColumn.lt-lg], [gdColumn.lt-xl],\n  [gdColumn.gt-xs], [gdColumn.gt-sm], [gdColumn.gt-md], [gdColumn.gt-lg]\n",inputs:X}]}],t}(K),$=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e,t){var n=!1;(e=e||"none").endsWith("!")&&(e=e.substring(0,e.indexOf("!")),n=!0);var r={display:t.inline?"inline-grid":"grid","grid-auto-columns":"","grid-template-columns":""};return r[n?"grid-auto-columns":"grid-template-columns"]=e,r},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(o.StyleBuilder),J=function(e){function t(t,n,r,i){var o=e.call(this,t,n,r,i)||this;return o.elementRef=t,o.styleBuilder=n,o.styler=r,o.marshal=i,o.DIRECTIVE_KEY="grid-columns",o._inline=!1,o.init(),o}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"inline",{get:function(){return this._inline},set:function(e){this._inline=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t.prototype.updateWithValue=function(e){this.styleCache=this.inline?te:ee,this.addStyles(e,{inline:this.inline})},t.ctorParameters=function(){return[{type:i.ElementRef},{type:$,decorators:[{type:i.Optional}]},{type:o.StyleUtils},{type:o.MediaMarshaller}]},t.propDecorators={inline:[{type:i.Input,args:["gdInline"]}]},t}(o.BaseDirective2),ee=new Map,te=new Map,ne=["gdColumns","gdColumns.xs","gdColumns.sm","gdColumns.md","gdColumns.lg","gdColumns.xl","gdColumns.lt-sm","gdColumns.lt-md","gdColumns.lt-lg","gdColumns.lt-xl","gdColumns.gt-xs","gdColumns.gt-sm","gdColumns.gt-md","gdColumns.gt-lg"],re=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=ne,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [gdColumns],\n  [gdColumns.xs], [gdColumns.sm], [gdColumns.md], [gdColumns.lg], [gdColumns.xl],\n  [gdColumns.lt-sm], [gdColumns.lt-md], [gdColumns.lt-lg], [gdColumns.lt-xl],\n  [gdColumns.gt-xs], [gdColumns.gt-sm], [gdColumns.gt-md], [gdColumns.gt-lg]\n",inputs:ne}]}],t}(J),ie=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e,t){return{display:t.inline?"inline-grid":"grid","grid-gap":e||"0"}},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(o.StyleBuilder),oe=function(e){function t(t,n,r,i){var o=e.call(this,t,r,n,i)||this;return o.elRef=t,o.styleUtils=n,o.styleBuilder=r,o.marshal=i,o.DIRECTIVE_KEY="grid-gap",o._inline=!1,o.init(),o}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"inline",{get:function(){return this._inline},set:function(e){this._inline=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t.prototype.updateWithValue=function(e){this.styleCache=this.inline?se:ae,this.addStyles(e,{inline:this.inline})},t.ctorParameters=function(){return[{type:i.ElementRef},{type:o.StyleUtils},{type:ie,decorators:[{type:i.Optional}]},{type:o.MediaMarshaller}]},t.propDecorators={inline:[{type:i.Input,args:["gdInline"]}]},t}(o.BaseDirective2),ae=new Map,se=new Map,ue=["gdGap","gdGap.xs","gdGap.sm","gdGap.md","gdGap.lg","gdGap.xl","gdGap.lt-sm","gdGap.lt-md","gdGap.lt-lg","gdGap.lt-xl","gdGap.gt-xs","gdGap.gt-sm","gdGap.gt-md","gdGap.gt-lg"],ce=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=ue,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [gdGap],\n  [gdGap.xs], [gdGap.sm], [gdGap.md], [gdGap.lg], [gdGap.xl],\n  [gdGap.lt-sm], [gdGap.lt-md], [gdGap.lt-lg], [gdGap.lt-xl],\n  [gdGap.gt-xs], [gdGap.gt-sm], [gdGap.gt-md], [gdGap.gt-lg]\n",inputs:ue}]}],t}(oe),le=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e){return{"grid-row":e||"auto"}},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(o.StyleBuilder),de=function(e){function t(t,n,r,i){var o=e.call(this,t,n,r,i)||this;return o.elementRef=t,o.styleBuilder=n,o.styler=r,o.marshal=i,o.DIRECTIVE_KEY="grid-row",o.styleCache=pe,o.init(),o}return Object(r.__extends)(t,e),t.ctorParameters=function(){return[{type:i.ElementRef},{type:le,decorators:[{type:i.Optional}]},{type:o.StyleUtils},{type:o.MediaMarshaller}]},t}(o.BaseDirective2),pe=new Map,fe=["gdRow","gdRow.xs","gdRow.sm","gdRow.md","gdRow.lg","gdRow.xl","gdRow.lt-sm","gdRow.lt-md","gdRow.lt-lg","gdRow.lt-xl","gdRow.gt-xs","gdRow.gt-sm","gdRow.gt-md","gdRow.gt-lg"],he=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=fe,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [gdRow],\n  [gdRow.xs], [gdRow.sm], [gdRow.md], [gdRow.lg], [gdRow.xl],\n  [gdRow.lt-sm], [gdRow.lt-md], [gdRow.lt-lg], [gdRow.lt-xl],\n  [gdRow.gt-xs], [gdRow.gt-sm], [gdRow.gt-md], [gdRow.gt-lg]\n",inputs:fe}]}],t}(de),me=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.buildStyles=function(e,t){var n=!1;(e=e||"none").endsWith("!")&&(e=e.substring(0,e.indexOf("!")),n=!0);var r={display:t.inline?"inline-grid":"grid","grid-auto-rows":"","grid-template-rows":""};return r[n?"grid-auto-rows":"grid-template-rows"]=e,r},t.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],t.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new t},token:t,providedIn:"root"}),t}(o.StyleBuilder),ge=function(e){function t(t,n,r,i){var o=e.call(this,t,n,r,i)||this;return o.elementRef=t,o.styleBuilder=n,o.styler=r,o.marshal=i,o.DIRECTIVE_KEY="grid-rows",o._inline=!1,o.init(),o}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"inline",{get:function(){return this._inline},set:function(e){this._inline=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t.prototype.updateWithValue=function(e){this.styleCache=this.inline?_e:ve,this.addStyles(e,{inline:this.inline})},t.ctorParameters=function(){return[{type:i.ElementRef},{type:me,decorators:[{type:i.Optional}]},{type:o.StyleUtils},{type:o.MediaMarshaller}]},t.propDecorators={inline:[{type:i.Input,args:["gdInline"]}]},t}(o.BaseDirective2),ve=new Map,_e=new Map,ye=["gdRows","gdRows.xs","gdRows.sm","gdRows.md","gdRows.lg","gdRows.xl","gdRows.lt-sm","gdRows.lt-md","gdRows.lt-lg","gdRows.lt-xl","gdRows.gt-xs","gdRows.gt-sm","gdRows.gt-md","gdRows.gt-lg"],be=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.inputs=ye,t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"\n  [gdRows],\n  [gdRows.xs], [gdRows.sm], [gdRows.md], [gdRows.lg], [gdRows.xl],\n  [gdRows.lt-sm], [gdRows.lt-md], [gdRows.lt-lg], [gdRows.lt-xl],\n  [gdRows.gt-xs], [gdRows.gt-sm], [gdRows.gt-md], [gdRows.gt-lg]\n",inputs:ye}]}],t}(ge),je=[f,j,M,P,B,W,Z,re,ce,he,be],xe=function(){function e(){}return e.decorators=[{type:i.NgModule,args:[{imports:[o.CoreModule],declarations:je.slice(),exports:je.slice()}]}],e}()},"./node_modules/@angular/forms/fesm5/forms.js":function(e,t,n){"use strict";n.r(t),n.d(t,"ɵangular_packages_forms_forms_bc",function(){return lt}),n.d(t,"ɵangular_packages_forms_forms_bb",function(){return ct}),n.d(t,"ɵangular_packages_forms_forms_z",function(){return st}),n.d(t,"ɵangular_packages_forms_forms_ba",function(){return ut}),n.d(t,"ɵangular_packages_forms_forms_a",function(){return y}),n.d(t,"ɵangular_packages_forms_forms_b",function(){return j}),n.d(t,"ɵangular_packages_forms_forms_c",function(){return le}),n.d(t,"ɵangular_packages_forms_forms_d",function(){return de}),n.d(t,"ɵangular_packages_forms_forms_e",function(){return Oe}),n.d(t,"ɵangular_packages_forms_forms_f",function(){return Se}),n.d(t,"ɵangular_packages_forms_forms_g",function(){return Te}),n.d(t,"ɵangular_packages_forms_forms_h",function(){return Me}),n.d(t,"ɵangular_packages_forms_forms_bh",function(){return at}),n.d(t,"ɵangular_packages_forms_forms_bd",function(){return E}),n.d(t,"ɵangular_packages_forms_forms_be",function(){return S}),n.d(t,"ɵangular_packages_forms_forms_i",function(){return A}),n.d(t,"ɵangular_packages_forms_forms_j",function(){return T}),n.d(t,"ɵangular_packages_forms_forms_bf",function(){return I}),n.d(t,"ɵangular_packages_forms_forms_bg",function(){return P}),n.d(t,"ɵangular_packages_forms_forms_k",function(){return Pe}),n.d(t,"ɵangular_packages_forms_forms_l",function(){return Re}),n.d(t,"ɵangular_packages_forms_forms_m",function(){return qe}),n.d(t,"ɵangular_packages_forms_forms_n",function(){return Fe}),n.d(t,"ɵangular_packages_forms_forms_p",function(){return ze}),n.d(t,"ɵangular_packages_forms_forms_o",function(){return Ve}),n.d(t,"ɵangular_packages_forms_forms_q",function(){return z}),n.d(t,"ɵangular_packages_forms_forms_s",function(){return K}),n.d(t,"ɵangular_packages_forms_forms_r",function(){return G}),n.d(t,"ɵangular_packages_forms_forms_u",function(){return Ye}),n.d(t,"ɵangular_packages_forms_forms_v",function(){return Xe}),n.d(t,"ɵangular_packages_forms_forms_x",function(){return et}),n.d(t,"ɵangular_packages_forms_forms_w",function(){return $e}),n.d(t,"ɵangular_packages_forms_forms_y",function(){return nt}),n.d(t,"ɵangular_packages_forms_forms_t",function(){return We}),n.d(t,"AbstractControlDirective",function(){return u}),n.d(t,"AbstractFormGroupDirective",function(){return ce}),n.d(t,"CheckboxControlValueAccessor",function(){return b}),n.d(t,"ControlContainer",function(){return c}),n.d(t,"NG_VALUE_ACCESSOR",function(){return _}),n.d(t,"COMPOSITION_BUFFER_MODE",function(){return x}),n.d(t,"DefaultValueAccessor",function(){return O}),n.d(t,"NgControl",function(){return M}),n.d(t,"NgControlStatus",function(){return pe}),n.d(t,"NgControlStatusGroup",function(){return fe}),n.d(t,"NgForm",function(){return Ce}),n.d(t,"NgFormSelectorWarning",function(){return ke}),n.d(t,"NgModel",function(){return Ie}),n.d(t,"NgModelGroup",function(){return Ae}),n.d(t,"RadioControlValueAccessor",function(){return D}),n.d(t,"FormControlDirective",function(){return Ne}),n.d(t,"FormControlName",function(){return Ge}),n.d(t,"FormGroupDirective",function(){return Le}),n.d(t,"FormArrayName",function(){return Ue}),n.d(t,"FormGroupName",function(){return Be}),n.d(t,"NgSelectOption",function(){return q}),n.d(t,"SelectControlValueAccessor",function(){return H}),n.d(t,"SelectMultipleControlValueAccessor",function(){return Y}),n.d(t,"CheckboxRequiredValidator",function(){return Qe}),n.d(t,"EmailValidator",function(){return Ze}),n.d(t,"MaxLengthValidator",function(){return tt}),n.d(t,"MinLengthValidator",function(){return Je}),n.d(t,"PatternValidator",function(){return rt}),n.d(t,"RequiredValidator",function(){return Ke}),n.d(t,"FormBuilder",function(){return it}),n.d(t,"AbstractControl",function(){return ye}),n.d(t,"FormArray",function(){return xe}),n.d(t,"FormControl",function(){return be}),n.d(t,"FormGroup",function(){return je}),n.d(t,"NG_ASYNC_VALIDATORS",function(){return p}),n.d(t,"NG_VALIDATORS",function(){return d}),n.d(t,"Validators",function(){return h}),n.d(t,"VERSION",function(){return ot}),n.d(t,"FormsModule",function(){return dt}),n.d(t,"ReactiveFormsModule",function(){return pt});var r=n("./node_modules/@angular/forms/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/rxjs/_esm5/index.js"),a=n("./node_modules/rxjs/_esm5/operators/index.js"),s=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),u=function(){function e(){}return Object.defineProperty(e.prototype,"value",{get:function(){return this.control?this.control.value:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"valid",{get:function(){return this.control?this.control.valid:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"invalid",{get:function(){return this.control?this.control.invalid:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"pending",{get:function(){return this.control?this.control.pending:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disabled",{get:function(){return this.control?this.control.disabled:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"enabled",{get:function(){return this.control?this.control.enabled:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"errors",{get:function(){return this.control?this.control.errors:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"pristine",{get:function(){return this.control?this.control.pristine:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"dirty",{get:function(){return this.control?this.control.dirty:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"touched",{get:function(){return this.control?this.control.touched:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"status",{get:function(){return this.control?this.control.status:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"untouched",{get:function(){return this.control?this.control.untouched:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"statusChanges",{get:function(){return this.control?this.control.statusChanges:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"valueChanges",{get:function(){return this.control?this.control.valueChanges:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"path",{get:function(){return null},enumerable:!0,configurable:!0}),e.prototype.reset=function(e){void 0===e&&(e=void 0),this.control&&this.control.reset(e)},e.prototype.hasError=function(e,t){return!!this.control&&this.control.hasError(e,t)},e.prototype.getError=function(e,t){return this.control?this.control.getError(e,t):null},e}(),c=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"formDirective",{get:function(){return null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"path",{get:function(){return null},enumerable:!0,configurable:!0}),t}(u);function l(e){return null==e||0===e.length}var d=new i.InjectionToken("NgValidators"),p=new i.InjectionToken("NgAsyncValidators"),f=/^(?=.{1,254}$)(?=.{1,64}@)[-!#$%&'*+/0-9=?A-Z^_`a-z{|}~]+(\.[-!#$%&'*+/0-9=?A-Z^_`a-z{|}~]+)*@[A-Za-z0-9]([A-Za-z0-9-]{0,61}[A-Za-z0-9])?(\.[A-Za-z0-9]([A-Za-z0-9-]{0,61}[A-Za-z0-9])?)*$/,h=function(){function e(){}return e.min=function(e){return function(t){if(l(t.value)||l(e))return null;var n=parseFloat(t.value);return!isNaN(n)&&n<e?{min:{min:e,actual:t.value}}:null}},e.max=function(e){return function(t){if(l(t.value)||l(e))return null;var n=parseFloat(t.value);return!isNaN(n)&&n>e?{max:{max:e,actual:t.value}}:null}},e.required=function(e){return l(e.value)?{required:!0}:null},e.requiredTrue=function(e){return!0===e.value?null:{required:!0}},e.email=function(e){return l(e.value)?null:f.test(e.value)?null:{email:!0}},e.minLength=function(e){return function(t){if(l(t.value))return null;var n=t.value?t.value.length:0;return n<e?{minlength:{requiredLength:e,actualLength:n}}:null}},e.maxLength=function(e){return function(t){var n=t.value?t.value.length:0;return n>e?{maxlength:{requiredLength:e,actualLength:n}}:null}},e.pattern=function(t){return t?("string"==typeof t?(r="","^"!==t.charAt(0)&&(r+="^"),r+=t,"$"!==t.charAt(t.length-1)&&(r+="$"),n=new RegExp(r)):(r=t.toString(),n=t),function(e){if(l(e.value))return null;var t=e.value;return n.test(t)?null:{pattern:{requiredPattern:r,actualValue:t}}}):e.nullValidator;var n,r},e.nullValidator=function(e){return null},e.compose=function(e){if(!e)return null;var t=e.filter(m);return 0==t.length?null:function(e){return v((n=e,t.map(function(e){return e(n)})));var n}},e.composeAsync=function(e){if(!e)return null;var t=e.filter(m);return 0==t.length?null:function(e){var n,r,i=(n=e,r=t,r.map(function(e){return e(n)})).map(g);return Object(o.forkJoin)(i).pipe(Object(a.map)(v))}},e}();function m(e){return null!=e}function g(e){var t=Object(i["ɵisPromise"])(e)?Object(o.from)(e):e;if(!Object(i["ɵisObservable"])(t))throw new Error("Expected validator to return Promise or Observable.");return t}function v(e){var t=e.reduce(function(e,t){return null!=t?Object(r.__assign)({},e,t):e},{});return 0===Object.keys(t).length?null:t}var _=new i.InjectionToken("NgValueAccessor"),y={provide:_,useExisting:Object(i.forwardRef)(function(){return b}),multi:!0},b=function(){function e(e,t){this._renderer=e,this._elementRef=t,this.onChange=function(e){},this.onTouched=function(){}}return e.prototype.writeValue=function(e){this._renderer.setProperty(this._elementRef.nativeElement,"checked",e)},e.prototype.registerOnChange=function(e){this.onChange=e},e.prototype.registerOnTouched=function(e){this.onTouched=e},e.prototype.setDisabledState=function(e){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",e)},e=Object(r.__decorate)([Object(i.Directive)({selector:"input[type=checkbox][formControlName],input[type=checkbox][formControl],input[type=checkbox][ngModel]",host:{"(change)":"onChange($event.target.checked)","(blur)":"onTouched()"},providers:[y]}),Object(r.__metadata)("design:paramtypes",[i.Renderer2,i.ElementRef])],e)}(),j={provide:_,useExisting:Object(i.forwardRef)(function(){return O}),multi:!0};var x=new i.InjectionToken("CompositionEventMode"),O=function(){function e(e,t,n){var r;this._renderer=e,this._elementRef=t,this._compositionMode=n,this.onChange=function(e){},this.onTouched=function(){},this._composing=!1,null==this._compositionMode&&(this._compositionMode=(r=Object(s["ɵgetDOM"])()?Object(s["ɵgetDOM"])().getUserAgent():"",!/android (\d+)/.test(r.toLowerCase())))}return e.prototype.writeValue=function(e){var t=null==e?"":e;this._renderer.setProperty(this._elementRef.nativeElement,"value",t)},e.prototype.registerOnChange=function(e){this.onChange=e},e.prototype.registerOnTouched=function(e){this.onTouched=e},e.prototype.setDisabledState=function(e){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",e)},e.prototype._handleInput=function(e){(!this._compositionMode||this._compositionMode&&!this._composing)&&this.onChange(e)},e.prototype._compositionStart=function(){this._composing=!0},e.prototype._compositionEnd=function(e){this._composing=!1,this._compositionMode&&this.onChange(e)},e=Object(r.__decorate)([Object(i.Directive)({selector:"input:not([type=checkbox])[formControlName],textarea[formControlName],input:not([type=checkbox])[formControl],textarea[formControl],input:not([type=checkbox])[ngModel],textarea[ngModel],[ngDefaultControl]",host:{"(input)":"$any(this)._handleInput($event.target.value)","(blur)":"onTouched()","(compositionstart)":"$any(this)._compositionStart()","(compositionend)":"$any(this)._compositionEnd($event.target.value)"},providers:[j]}),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Inject)(x)),Object(r.__metadata)("design:paramtypes",[i.Renderer2,i.ElementRef,Boolean])],e)}();function w(e){return e.validate?function(t){return e.validate(t)}:e}function C(e){return e.validate?function(t){return e.validate(t)}:e}var E={provide:_,useExisting:Object(i.forwardRef)(function(){return S}),multi:!0},S=function(){function e(e,t){this._renderer=e,this._elementRef=t,this.onChange=function(e){},this.onTouched=function(){}}return e.prototype.writeValue=function(e){var t=null==e?"":e;this._renderer.setProperty(this._elementRef.nativeElement,"value",t)},e.prototype.registerOnChange=function(e){this.onChange=function(t){e(""==t?null:parseFloat(t))}},e.prototype.registerOnTouched=function(e){this.onTouched=e},e.prototype.setDisabledState=function(e){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",e)},e=Object(r.__decorate)([Object(i.Directive)({selector:"input[type=number][formControlName],input[type=number][formControl],input[type=number][ngModel]",host:{"(change)":"onChange($event.target.value)","(input)":"onChange($event.target.value)","(blur)":"onTouched()"},providers:[E]}),Object(r.__metadata)("design:paramtypes",[i.Renderer2,i.ElementRef])],e)}();function k(){throw new Error("unimplemented")}var M=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t._parent=null,t.name=null,t.valueAccessor=null,t._rawValidators=[],t._rawAsyncValidators=[],t}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"validator",{get:function(){return k()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"asyncValidator",{get:function(){return k()},enumerable:!0,configurable:!0}),t}(u),A={provide:_,useExisting:Object(i.forwardRef)(function(){return D}),multi:!0},T=function(){function e(){this._accessors=[]}return e.prototype.add=function(e,t){this._accessors.push([e,t])},e.prototype.remove=function(e){for(var t=this._accessors.length-1;t>=0;--t)if(this._accessors[t][1]===e)return void this._accessors.splice(t,1)},e.prototype.select=function(e){var t=this;this._accessors.forEach(function(n){t._isSameGroup(n,e)&&n[1]!==e&&n[1].fireUncheck(e.value)})},e.prototype._isSameGroup=function(e,t){return!!e[0].control&&(e[0]._parent===t._control._parent&&e[1].name===t.name)},e=Object(r.__decorate)([Object(i.Injectable)()],e)}(),D=function(){function e(e,t,n,r){this._renderer=e,this._elementRef=t,this._registry=n,this._injector=r,this.onChange=function(){},this.onTouched=function(){}}return e.prototype.ngOnInit=function(){this._control=this._injector.get(M),this._checkName(),this._registry.add(this._control,this)},e.prototype.ngOnDestroy=function(){this._registry.remove(this)},e.prototype.writeValue=function(e){this._state=e===this.value,this._renderer.setProperty(this._elementRef.nativeElement,"checked",this._state)},e.prototype.registerOnChange=function(e){var t=this;this._fn=e,this.onChange=function(){e(t.value),t._registry.select(t)}},e.prototype.fireUncheck=function(e){this.writeValue(e)},e.prototype.registerOnTouched=function(e){this.onTouched=e},e.prototype.setDisabledState=function(e){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",e)},e.prototype._checkName=function(){this.name&&this.formControlName&&this.name!==this.formControlName&&this._throwNameError(),!this.name&&this.formControlName&&(this.name=this.formControlName)},e.prototype._throwNameError=function(){throw new Error('\n      If you define both a name and a formControlName attribute on your radio button, their values\n      must match. Ex: <input type="radio" formControlName="food" name="food">\n    ')},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",String)],e.prototype,"name",void 0),Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",String)],e.prototype,"formControlName",void 0),Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",Object)],e.prototype,"value",void 0),e=Object(r.__decorate)([Object(i.Directive)({selector:"input[type=radio][formControlName],input[type=radio][formControl],input[type=radio][ngModel]",host:{"(change)":"onChange()","(blur)":"onTouched()"},providers:[A]}),Object(r.__metadata)("design:paramtypes",[i.Renderer2,i.ElementRef,T,i.Injector])],e)}(),I={provide:_,useExisting:Object(i.forwardRef)(function(){return P}),multi:!0},P=function(){function e(e,t){this._renderer=e,this._elementRef=t,this.onChange=function(e){},this.onTouched=function(){}}return e.prototype.writeValue=function(e){this._renderer.setProperty(this._elementRef.nativeElement,"value",parseFloat(e))},e.prototype.registerOnChange=function(e){this.onChange=function(t){e(""==t?null:parseFloat(t))}},e.prototype.registerOnTouched=function(e){this.onTouched=e},e.prototype.setDisabledState=function(e){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",e)},e=Object(r.__decorate)([Object(i.Directive)({selector:"input[type=range][formControlName],input[type=range][formControl],input[type=range][ngModel]",host:{"(change)":"onChange($event.target.value)","(input)":"onChange($event.target.value)","(blur)":"onTouched()"},providers:[I]}),Object(r.__metadata)("design:paramtypes",[i.Renderer2,i.ElementRef])],e)}(),R='\n    <div [formGroup]="myGroup">\n      <input formControlName="firstName">\n    </div>\n\n    In your class:\n\n    this.myGroup = new FormGroup({\n       firstName: new FormControl()\n    });',N='\n    <div [formGroup]="myGroup">\n       <div formGroupName="person">\n          <input formControlName="firstName">\n       </div>\n    </div>\n\n    In your class:\n\n    this.myGroup = new FormGroup({\n       person: new FormGroup({ firstName: new FormControl() })\n    });',F='\n    <div [formGroup]="myGroup">\n      <div formArrayName="cities">\n        <div *ngFor="let city of cityArray.controls; index as i">\n          <input [formControlName]="i">\n        </div>\n      </div>\n    </div>\n\n    In your class:\n\n    this.cityArray = new FormArray([new FormControl(\'SF\')]);\n    this.myGroup = new FormGroup({\n      cities: this.cityArray\n    });',L='\n    <form>\n       <div ngModelGroup="person">\n          <input [(ngModel)]="person.name" name="firstName">\n       </div>\n    </form>',V='\n    <div [formGroup]="myGroup">\n       <input formControlName="firstName">\n       <input [(ngModel)]="showMoreControls" [ngModelOptions]="{standalone: true}">\n    </div>\n  ',B=function(){function e(){}return e.controlParentException=function(){throw new Error("formControlName must be used with a parent formGroup directive.  You'll want to add a formGroup\n       directive and pass it an existing FormGroup instance (you can create one in your class).\n\n      Example:\n\n      "+R)},e.ngModelGroupException=function(){throw new Error('formControlName cannot be used with an ngModelGroup parent. It is only compatible with parents\n       that also have a "form" prefix: formGroupName, formArrayName, or formGroup.\n\n       Option 1:  Update the parent to be formGroupName (reactive form strategy)\n\n        '+N+"\n\n        Option 2: Use ngModel instead of formControlName (template-driven strategy)\n\n        "+L)},e.missingFormException=function(){throw new Error("formGroup expects a FormGroup instance. Please pass one in.\n\n       Example:\n\n       "+R)},e.groupParentException=function(){throw new Error("formGroupName must be used with a parent formGroup directive.  You'll want to add a formGroup\n      directive and pass it an existing FormGroup instance (you can create one in your class).\n\n      Example:\n\n      "+N)},e.arrayParentException=function(){throw new Error("formArrayName must be used with a parent formGroup directive.  You'll want to add a formGroup\n       directive and pass it an existing FormGroup instance (you can create one in your class).\n\n        Example:\n\n        "+F)},e.disabledAttrWarning=function(){console.warn("\n      It looks like you're using the disabled attribute with a reactive form directive. If you set disabled to true\n      when you set up this control in your component class, the disabled attribute will actually be set in the DOM for\n      you. We recommend using this approach to avoid 'changed after checked' errors.\n       \n      Example: \n      form = new FormGroup({\n        first: new FormControl({value: 'Nancy', disabled: true}, Validators.required),\n        last: new FormControl('Drew', Validators.required)\n      });\n    ")},e.ngModelWarning=function(e){console.warn("\n    It looks like you're using ngModel on the same form field as "+e+". \n    Support for using the ngModel input property and ngModelChange event with \n    reactive form directives has been deprecated in Angular v6 and will be removed \n    in Angular v7.\n    \n    For more information on this, see our API docs here:\n    https://angular.io/api/forms/"+("formControl"===e?"FormControlDirective":"FormControlName")+"#use-with-ngmodel\n    ")},e}(),z={provide:_,useExisting:Object(i.forwardRef)(function(){return H}),multi:!0};function U(e,t){return null==e?""+t:(t&&"object"==typeof t&&(t="Object"),(e+": "+t).slice(0,50))}var H=function(){function e(e,t){this._renderer=e,this._elementRef=t,this._optionMap=new Map,this._idCounter=0,this.onChange=function(e){},this.onTouched=function(){},this._compareWith=i["ɵlooseIdentical"]}return Object.defineProperty(e.prototype,"compareWith",{set:function(e){if("function"!=typeof e)throw new Error("compareWith must be a function, but received "+JSON.stringify(e));this._compareWith=e},enumerable:!0,configurable:!0}),e.prototype.writeValue=function(e){this.value=e;var t=this._getOptionId(e);null==t&&this._renderer.setProperty(this._elementRef.nativeElement,"selectedIndex",-1);var n=U(t,e);this._renderer.setProperty(this._elementRef.nativeElement,"value",n)},e.prototype.registerOnChange=function(e){var t=this;this.onChange=function(n){t.value=t._getOptionValue(n),e(t.value)}},e.prototype.registerOnTouched=function(e){this.onTouched=e},e.prototype.setDisabledState=function(e){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",e)},e.prototype._registerOption=function(){return(this._idCounter++).toString()},e.prototype._getOptionId=function(e){var t,n;try{for(var i=Object(r.__values)(Array.from(this._optionMap.keys())),o=i.next();!o.done;o=i.next()){var a=o.value;if(this._compareWith(this._optionMap.get(a),e))return a}}catch(e){t={error:e}}finally{try{o&&!o.done&&(n=i.return)&&n.call(i)}finally{if(t)throw t.error}}return null},e.prototype._getOptionValue=function(e){var t=e.split(":")[0];return this._optionMap.has(t)?this._optionMap.get(t):e},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",Function),Object(r.__metadata)("design:paramtypes",[Function])],e.prototype,"compareWith",null),e=Object(r.__decorate)([Object(i.Directive)({selector:"select:not([multiple])[formControlName],select:not([multiple])[formControl],select:not([multiple])[ngModel]",host:{"(change)":"onChange($event.target.value)","(blur)":"onTouched()"},providers:[z]}),Object(r.__metadata)("design:paramtypes",[i.Renderer2,i.ElementRef])],e)}(),q=function(){function e(e,t,n){this._element=e,this._renderer=t,this._select=n,this._select&&(this.id=this._select._registerOption())}return Object.defineProperty(e.prototype,"ngValue",{set:function(e){null!=this._select&&(this._select._optionMap.set(this.id,e),this._setElementValue(U(this.id,e)),this._select.writeValue(this._select.value))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{set:function(e){this._setElementValue(e),this._select&&this._select.writeValue(this._select.value)},enumerable:!0,configurable:!0}),e.prototype._setElementValue=function(e){this._renderer.setProperty(this._element.nativeElement,"value",e)},e.prototype.ngOnDestroy=function(){this._select&&(this._select._optionMap.delete(this.id),this._select.writeValue(this._select.value))},Object(r.__decorate)([Object(i.Input)("ngValue"),Object(r.__metadata)("design:type",Object),Object(r.__metadata)("design:paramtypes",[Object])],e.prototype,"ngValue",null),Object(r.__decorate)([Object(i.Input)("value"),Object(r.__metadata)("design:type",Object),Object(r.__metadata)("design:paramtypes",[Object])],e.prototype,"value",null),e=Object(r.__decorate)([Object(i.Directive)({selector:"option"}),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Host)()),Object(r.__metadata)("design:paramtypes",[i.ElementRef,i.Renderer2,H])],e)}(),G={provide:_,useExisting:Object(i.forwardRef)(function(){return Y}),multi:!0};function W(e,t){return null==e?""+t:("string"==typeof t&&(t="'"+t+"'"),t&&"object"==typeof t&&(t="Object"),(e+": "+t).slice(0,50))}var Y=function(){function e(e,t){this._renderer=e,this._elementRef=t,this._optionMap=new Map,this._idCounter=0,this.onChange=function(e){},this.onTouched=function(){},this._compareWith=i["ɵlooseIdentical"]}return Object.defineProperty(e.prototype,"compareWith",{set:function(e){if("function"!=typeof e)throw new Error("compareWith must be a function, but received "+JSON.stringify(e));this._compareWith=e},enumerable:!0,configurable:!0}),e.prototype.writeValue=function(e){var t,n=this;if(this.value=e,Array.isArray(e)){var r=e.map(function(e){return n._getOptionId(e)});t=function(e,t){e._setSelected(r.indexOf(t.toString())>-1)}}else t=function(e,t){e._setSelected(!1)};this._optionMap.forEach(t)},e.prototype.registerOnChange=function(e){var t=this;this.onChange=function(n){var r=[];if(n.hasOwnProperty("selectedOptions"))for(var i=n.selectedOptions,o=0;o<i.length;o++){var a=i.item(o),s=t._getOptionValue(a.value);r.push(s)}else for(i=n.options,o=0;o<i.length;o++){if((a=i.item(o)).selected){s=t._getOptionValue(a.value);r.push(s)}}t.value=r,e(r)}},e.prototype.registerOnTouched=function(e){this.onTouched=e},e.prototype.setDisabledState=function(e){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",e)},e.prototype._registerOption=function(e){var t=(this._idCounter++).toString();return this._optionMap.set(t,e),t},e.prototype._getOptionId=function(e){var t,n;try{for(var i=Object(r.__values)(Array.from(this._optionMap.keys())),o=i.next();!o.done;o=i.next()){var a=o.value;if(this._compareWith(this._optionMap.get(a)._value,e))return a}}catch(e){t={error:e}}finally{try{o&&!o.done&&(n=i.return)&&n.call(i)}finally{if(t)throw t.error}}return null},e.prototype._getOptionValue=function(e){var t=e.split(":")[0];return this._optionMap.has(t)?this._optionMap.get(t)._value:e},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",Function),Object(r.__metadata)("design:paramtypes",[Function])],e.prototype,"compareWith",null),e=Object(r.__decorate)([Object(i.Directive)({selector:"select[multiple][formControlName],select[multiple][formControl],select[multiple][ngModel]",host:{"(change)":"onChange($event.target)","(blur)":"onTouched()"},providers:[G]}),Object(r.__metadata)("design:paramtypes",[i.Renderer2,i.ElementRef])],e)}(),K=function(){function e(e,t,n){this._element=e,this._renderer=t,this._select=n,this._select&&(this.id=this._select._registerOption(this))}return Object.defineProperty(e.prototype,"ngValue",{set:function(e){null!=this._select&&(this._value=e,this._setElementValue(W(this.id,e)),this._select.writeValue(this._select.value))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{set:function(e){this._select?(this._value=e,this._setElementValue(W(this.id,e)),this._select.writeValue(this._select.value)):this._setElementValue(e)},enumerable:!0,configurable:!0}),e.prototype._setElementValue=function(e){this._renderer.setProperty(this._element.nativeElement,"value",e)},e.prototype._setSelected=function(e){this._renderer.setProperty(this._element.nativeElement,"selected",e)},e.prototype.ngOnDestroy=function(){this._select&&(this._select._optionMap.delete(this.id),this._select.writeValue(this._select.value))},Object(r.__decorate)([Object(i.Input)("ngValue"),Object(r.__metadata)("design:type",Object),Object(r.__metadata)("design:paramtypes",[Object])],e.prototype,"ngValue",null),Object(r.__decorate)([Object(i.Input)("value"),Object(r.__metadata)("design:type",Object),Object(r.__metadata)("design:paramtypes",[Object])],e.prototype,"value",null),e=Object(r.__decorate)([Object(i.Directive)({selector:"option"}),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Host)()),Object(r.__metadata)("design:paramtypes",[i.ElementRef,i.Renderer2,Y])],e)}();function Q(e,t){return Object(r.__spread)(t.path,[e])}function X(e,t){var n,r,i,o,a;e||ee(t,"Cannot find control with"),t.valueAccessor||ee(t,"No value accessor for form control with"),e.validator=h.compose([e.validator,t.validator]),e.asyncValidator=h.composeAsync([e.asyncValidator,t.asyncValidator]),t.valueAccessor.writeValue(e.value),n=e,(r=t).valueAccessor.registerOnChange(function(e){n._pendingValue=e,n._pendingChange=!0,n._pendingDirty=!0,"change"===n.updateOn&&Z(n,r)}),i=t,e.registerOnChange(function(e,t){i.valueAccessor.writeValue(e),t&&i.viewToModelUpdate(e)}),o=e,(a=t).valueAccessor.registerOnTouched(function(){o._pendingTouched=!0,"blur"===o.updateOn&&o._pendingChange&&Z(o,a),"submit"!==o.updateOn&&o.markAsTouched()}),t.valueAccessor.setDisabledState&&e.registerOnDisabledChange(function(e){t.valueAccessor.setDisabledState(e)}),t._rawValidators.forEach(function(t){t.registerOnValidatorChange&&t.registerOnValidatorChange(function(){return e.updateValueAndValidity()})}),t._rawAsyncValidators.forEach(function(t){t.registerOnValidatorChange&&t.registerOnValidatorChange(function(){return e.updateValueAndValidity()})})}function Z(e,t){e._pendingDirty&&e.markAsDirty(),e.setValue(e._pendingValue,{emitModelToViewChange:!1}),t.viewToModelUpdate(e._pendingValue),e._pendingChange=!1}function $(e,t){null==e&&ee(t,"Cannot find control with"),e.validator=h.compose([e.validator,t.validator]),e.asyncValidator=h.composeAsync([e.asyncValidator,t.asyncValidator])}function J(e){return ee(e,"There is no FormControl instance attached to form control element with")}function ee(e,t){var n;throw n=e.path.length>1?"path: '"+e.path.join(" -> ")+"'":e.path[0]?"name: '"+e.path+"'":"unspecified name attribute",new Error(t+" "+n)}function te(e){return null!=e?h.compose(e.map(w)):null}function ne(e){return null!=e?h.composeAsync(e.map(C)):null}function re(e,t){if(!e.hasOwnProperty("model"))return!1;var n=e.model;return!!n.isFirstChange()||!Object(i["ɵlooseIdentical"])(t,n.currentValue)}var ie=[b,P,S,H,Y,D];function oe(e,t){e._syncPendingControls(),t.forEach(function(e){var t=e.control;"submit"===t.updateOn&&t._pendingChange&&(e.viewToModelUpdate(t._pendingValue),t._pendingChange=!1)})}function ae(e,t){if(!t)return null;Array.isArray(t)||ee(e,"Value accessor was not provided as an array for form control with");var n=void 0,r=void 0,i=void 0;return t.forEach(function(t){var o;t.constructor===O?n=t:(o=t,ie.some(function(e){return o.constructor===e})?(r&&ee(e,"More than one built-in value accessor matches form control with"),r=t):(i&&ee(e,"More than one custom value accessor matches form control with"),i=t))}),i||(r||(n||(ee(e,"No valid value accessor for form control with"),null)))}function se(e,t){var n=e.indexOf(t);n>-1&&e.splice(n,1)}function ue(e,t,n,r){Object(i.isDevMode)()&&"never"!==r&&((null!==r&&"once"!==r||t._ngModelWarningSentOnce)&&("always"!==r||n._ngModelWarningSent)||(B.ngModelWarning(e),t._ngModelWarningSentOnce=!0,n._ngModelWarningSent=!0))}var ce=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.ngOnInit=function(){this._checkParentType(),this.formDirective.addFormGroup(this)},t.prototype.ngOnDestroy=function(){this.formDirective&&this.formDirective.removeFormGroup(this)},Object.defineProperty(t.prototype,"control",{get:function(){return this.formDirective.getFormGroup(this)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"path",{get:function(){return Q(this.name,this._parent)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"formDirective",{get:function(){return this._parent?this._parent.formDirective:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"validator",{get:function(){return te(this._validators)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"asyncValidator",{get:function(){return ne(this._asyncValidators)},enumerable:!0,configurable:!0}),t.prototype._checkParentType=function(){},t}(c),le=function(){function e(e){this._cd=e}return Object.defineProperty(e.prototype,"ngClassUntouched",{get:function(){return!!this._cd.control&&this._cd.control.untouched},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassTouched",{get:function(){return!!this._cd.control&&this._cd.control.touched},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassPristine",{get:function(){return!!this._cd.control&&this._cd.control.pristine},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassDirty",{get:function(){return!!this._cd.control&&this._cd.control.dirty},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassValid",{get:function(){return!!this._cd.control&&this._cd.control.valid},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassInvalid",{get:function(){return!!this._cd.control&&this._cd.control.invalid},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassPending",{get:function(){return!!this._cd.control&&this._cd.control.pending},enumerable:!0,configurable:!0}),e}(),de={"[class.ng-untouched]":"ngClassUntouched","[class.ng-touched]":"ngClassTouched","[class.ng-pristine]":"ngClassPristine","[class.ng-dirty]":"ngClassDirty","[class.ng-valid]":"ngClassValid","[class.ng-invalid]":"ngClassInvalid","[class.ng-pending]":"ngClassPending"},pe=function(e){function t(t){return e.call(this,t)||this}return Object(r.__extends)(t,e),t=Object(r.__decorate)([Object(i.Directive)({selector:"[formControlName],[ngModel],[formControl]",host:de}),Object(r.__param)(0,Object(i.Self)()),Object(r.__metadata)("design:paramtypes",[M])],t)}(le),fe=function(e){function t(t){return e.call(this,t)||this}return Object(r.__extends)(t,e),t=Object(r.__decorate)([Object(i.Directive)({selector:"[formGroupName],[formArrayName],[ngModelGroup],[formGroup],form:not([ngNoForm]),[ngForm]",host:de}),Object(r.__param)(0,Object(i.Self)()),Object(r.__metadata)("design:paramtypes",[c])],t)}(le),he="PENDING",me="DISABLED";function ge(e){var t=_e(e)?e.validators:e;return Array.isArray(t)?te(t):t||null}function ve(e,t){var n=_e(t)?t.asyncValidators:e;return Array.isArray(n)?ne(n):n||null}function _e(e){return null!=e&&!Array.isArray(e)&&"object"==typeof e}var ye=function(){function e(e,t){this.validator=e,this.asyncValidator=t,this._onCollectionChange=function(){},this.pristine=!0,this.touched=!1,this._onDisabledChange=[]}return Object.defineProperty(e.prototype,"parent",{get:function(){return this._parent},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"valid",{get:function(){return"VALID"===this.status},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"invalid",{get:function(){return"INVALID"===this.status},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"pending",{get:function(){return this.status==he},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disabled",{get:function(){return this.status===me},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"enabled",{get:function(){return this.status!==me},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"dirty",{get:function(){return!this.pristine},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"untouched",{get:function(){return!this.touched},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"updateOn",{get:function(){return this._updateOn?this._updateOn:this.parent?this.parent.updateOn:"change"},enumerable:!0,configurable:!0}),e.prototype.setValidators=function(e){this.validator=ge(e)},e.prototype.setAsyncValidators=function(e){this.asyncValidator=ve(e)},e.prototype.clearValidators=function(){this.validator=null},e.prototype.clearAsyncValidators=function(){this.asyncValidator=null},e.prototype.markAsTouched=function(e){void 0===e&&(e={}),this.touched=!0,this._parent&&!e.onlySelf&&this._parent.markAsTouched(e)},e.prototype.markAsUntouched=function(e){void 0===e&&(e={}),this.touched=!1,this._pendingTouched=!1,this._forEachChild(function(e){e.markAsUntouched({onlySelf:!0})}),this._parent&&!e.onlySelf&&this._parent._updateTouched(e)},e.prototype.markAsDirty=function(e){void 0===e&&(e={}),this.pristine=!1,this._parent&&!e.onlySelf&&this._parent.markAsDirty(e)},e.prototype.markAsPristine=function(e){void 0===e&&(e={}),this.pristine=!0,this._pendingDirty=!1,this._forEachChild(function(e){e.markAsPristine({onlySelf:!0})}),this._parent&&!e.onlySelf&&this._parent._updatePristine(e)},e.prototype.markAsPending=function(e){void 0===e&&(e={}),this.status=he,!1!==e.emitEvent&&this.statusChanges.emit(this.status),this._parent&&!e.onlySelf&&this._parent.markAsPending(e)},e.prototype.disable=function(e){void 0===e&&(e={}),this.status=me,this.errors=null,this._forEachChild(function(t){t.disable(Object(r.__assign)({},e,{onlySelf:!0}))}),this._updateValue(),!1!==e.emitEvent&&(this.valueChanges.emit(this.value),this.statusChanges.emit(this.status)),this._updateAncestors(e),this._onDisabledChange.forEach(function(e){return e(!0)})},e.prototype.enable=function(e){void 0===e&&(e={}),this.status="VALID",this._forEachChild(function(t){t.enable(Object(r.__assign)({},e,{onlySelf:!0}))}),this.updateValueAndValidity({onlySelf:!0,emitEvent:e.emitEvent}),this._updateAncestors(e),this._onDisabledChange.forEach(function(e){return e(!1)})},e.prototype._updateAncestors=function(e){this._parent&&!e.onlySelf&&(this._parent.updateValueAndValidity(e),this._parent._updatePristine(),this._parent._updateTouched())},e.prototype.setParent=function(e){this._parent=e},e.prototype.updateValueAndValidity=function(e){void 0===e&&(e={}),this._setInitialStatus(),this._updateValue(),this.enabled&&(this._cancelExistingSubscription(),this.errors=this._runValidator(),this.status=this._calculateStatus(),"VALID"!==this.status&&this.status!==he||this._runAsyncValidator(e.emitEvent)),!1!==e.emitEvent&&(this.valueChanges.emit(this.value),this.statusChanges.emit(this.status)),this._parent&&!e.onlySelf&&this._parent.updateValueAndValidity(e)},e.prototype._updateTreeValidity=function(e){void 0===e&&(e={emitEvent:!0}),this._forEachChild(function(t){return t._updateTreeValidity(e)}),this.updateValueAndValidity({onlySelf:!0,emitEvent:e.emitEvent})},e.prototype._setInitialStatus=function(){this.status=this._allControlsDisabled()?me:"VALID"},e.prototype._runValidator=function(){return this.validator?this.validator(this):null},e.prototype._runAsyncValidator=function(e){var t=this;if(this.asyncValidator){this.status=he;var n=g(this.asyncValidator(this));this._asyncValidationSubscription=n.subscribe(function(n){return t.setErrors(n,{emitEvent:e})})}},e.prototype._cancelExistingSubscription=function(){this._asyncValidationSubscription&&this._asyncValidationSubscription.unsubscribe()},e.prototype.setErrors=function(e,t){void 0===t&&(t={}),this.errors=e,this._updateControlsErrors(!1!==t.emitEvent)},e.prototype.get=function(e){return t=this,r=".",null==(n=e)?null:(n instanceof Array||(n=n.split(r)),n instanceof Array&&0===n.length?null:n.reduce(function(e,t){return e instanceof je?e.controls.hasOwnProperty(t)?e.controls[t]:null:e instanceof xe?e.at(t)||null:null},t));var t,n,r},e.prototype.getError=function(e,t){var n=t?this.get(t):this;return n&&n.errors?n.errors[e]:null},e.prototype.hasError=function(e,t){return!!this.getError(e,t)},Object.defineProperty(e.prototype,"root",{get:function(){for(var e=this;e._parent;)e=e._parent;return e},enumerable:!0,configurable:!0}),e.prototype._updateControlsErrors=function(e){this.status=this._calculateStatus(),e&&this.statusChanges.emit(this.status),this._parent&&this._parent._updateControlsErrors(e)},e.prototype._initObservables=function(){this.valueChanges=new i.EventEmitter,this.statusChanges=new i.EventEmitter},e.prototype._calculateStatus=function(){return this._allControlsDisabled()?me:this.errors?"INVALID":this._anyControlsHaveStatus(he)?he:this._anyControlsHaveStatus("INVALID")?"INVALID":"VALID"},e.prototype._anyControlsHaveStatus=function(e){return this._anyControls(function(t){return t.status===e})},e.prototype._anyControlsDirty=function(){return this._anyControls(function(e){return e.dirty})},e.prototype._anyControlsTouched=function(){return this._anyControls(function(e){return e.touched})},e.prototype._updatePristine=function(e){void 0===e&&(e={}),this.pristine=!this._anyControlsDirty(),this._parent&&!e.onlySelf&&this._parent._updatePristine(e)},e.prototype._updateTouched=function(e){void 0===e&&(e={}),this.touched=this._anyControlsTouched(),this._parent&&!e.onlySelf&&this._parent._updateTouched(e)},e.prototype._isBoxedValue=function(e){return"object"==typeof e&&null!==e&&2===Object.keys(e).length&&"value"in e&&"disabled"in e},e.prototype._registerOnCollectionChange=function(e){this._onCollectionChange=e},e.prototype._setUpdateStrategy=function(e){_e(e)&&null!=e.updateOn&&(this._updateOn=e.updateOn)},e}(),be=function(e){function t(t,n,r){void 0===t&&(t=null);var i=e.call(this,ge(n),ve(r,n))||this;return i._onChange=[],i._applyFormState(t),i._setUpdateStrategy(n),i.updateValueAndValidity({onlySelf:!0,emitEvent:!1}),i._initObservables(),i}return Object(r.__extends)(t,e),t.prototype.setValue=function(e,t){var n=this;void 0===t&&(t={}),this.value=this._pendingValue=e,this._onChange.length&&!1!==t.emitModelToViewChange&&this._onChange.forEach(function(e){return e(n.value,!1!==t.emitViewToModelChange)}),this.updateValueAndValidity(t)},t.prototype.patchValue=function(e,t){void 0===t&&(t={}),this.setValue(e,t)},t.prototype.reset=function(e,t){void 0===e&&(e=null),void 0===t&&(t={}),this._applyFormState(e),this.markAsPristine(t),this.markAsUntouched(t),this.setValue(this.value,t),this._pendingChange=!1},t.prototype._updateValue=function(){},t.prototype._anyControls=function(e){return!1},t.prototype._allControlsDisabled=function(){return this.disabled},t.prototype.registerOnChange=function(e){this._onChange.push(e)},t.prototype._clearChangeFns=function(){this._onChange=[],this._onDisabledChange=[],this._onCollectionChange=function(){}},t.prototype.registerOnDisabledChange=function(e){this._onDisabledChange.push(e)},t.prototype._forEachChild=function(e){},t.prototype._syncPendingControls=function(){return!("submit"!==this.updateOn||(this._pendingDirty&&this.markAsDirty(),this._pendingTouched&&this.markAsTouched(),!this._pendingChange))&&(this.setValue(this._pendingValue,{onlySelf:!0,emitModelToViewChange:!1}),!0)},t.prototype._applyFormState=function(e){this._isBoxedValue(e)?(this.value=this._pendingValue=e.value,e.disabled?this.disable({onlySelf:!0,emitEvent:!1}):this.enable({onlySelf:!0,emitEvent:!1})):this.value=this._pendingValue=e},t}(ye),je=function(e){function t(t,n,r){var i=e.call(this,ge(n),ve(r,n))||this;return i.controls=t,i._initObservables(),i._setUpdateStrategy(n),i._setUpControls(),i.updateValueAndValidity({onlySelf:!0,emitEvent:!1}),i}return Object(r.__extends)(t,e),t.prototype.registerControl=function(e,t){return this.controls[e]?this.controls[e]:(this.controls[e]=t,t.setParent(this),t._registerOnCollectionChange(this._onCollectionChange),t)},t.prototype.addControl=function(e,t){this.registerControl(e,t),this.updateValueAndValidity(),this._onCollectionChange()},t.prototype.removeControl=function(e){this.controls[e]&&this.controls[e]._registerOnCollectionChange(function(){}),delete this.controls[e],this.updateValueAndValidity(),this._onCollectionChange()},t.prototype.setControl=function(e,t){this.controls[e]&&this.controls[e]._registerOnCollectionChange(function(){}),delete this.controls[e],t&&this.registerControl(e,t),this.updateValueAndValidity(),this._onCollectionChange()},t.prototype.contains=function(e){return this.controls.hasOwnProperty(e)&&this.controls[e].enabled},t.prototype.setValue=function(e,t){var n=this;void 0===t&&(t={}),this._checkAllValuesPresent(e),Object.keys(e).forEach(function(r){n._throwIfControlMissing(r),n.controls[r].setValue(e[r],{onlySelf:!0,emitEvent:t.emitEvent})}),this.updateValueAndValidity(t)},t.prototype.patchValue=function(e,t){var n=this;void 0===t&&(t={}),Object.keys(e).forEach(function(r){n.controls[r]&&n.controls[r].patchValue(e[r],{onlySelf:!0,emitEvent:t.emitEvent})}),this.updateValueAndValidity(t)},t.prototype.reset=function(e,t){void 0===e&&(e={}),void 0===t&&(t={}),this._forEachChild(function(n,r){n.reset(e[r],{onlySelf:!0,emitEvent:t.emitEvent})}),this.updateValueAndValidity(t),this._updatePristine(t),this._updateTouched(t)},t.prototype.getRawValue=function(){return this._reduceChildren({},function(e,t,n){return e[n]=t instanceof be?t.value:t.getRawValue(),e})},t.prototype._syncPendingControls=function(){var e=this._reduceChildren(!1,function(e,t){return!!t._syncPendingControls()||e});return e&&this.updateValueAndValidity({onlySelf:!0}),e},t.prototype._throwIfControlMissing=function(e){if(!Object.keys(this.controls).length)throw new Error("\n        There are no form controls registered with this group yet.  If you're using ngModel,\n        you may want to check next tick (e.g. use setTimeout).\n      ");if(!this.controls[e])throw new Error("Cannot find form control with name: "+e+".")},t.prototype._forEachChild=function(e){var t=this;Object.keys(this.controls).forEach(function(n){return e(t.controls[n],n)})},t.prototype._setUpControls=function(){var e=this;this._forEachChild(function(t){t.setParent(e),t._registerOnCollectionChange(e._onCollectionChange)})},t.prototype._updateValue=function(){this.value=this._reduceValue()},t.prototype._anyControls=function(e){var t=this,n=!1;return this._forEachChild(function(r,i){n=n||t.contains(i)&&e(r)}),n},t.prototype._reduceValue=function(){var e=this;return this._reduceChildren({},function(t,n,r){return(n.enabled||e.disabled)&&(t[r]=n.value),t})},t.prototype._reduceChildren=function(e,t){var n=e;return this._forEachChild(function(e,r){n=t(n,e,r)}),n},t.prototype._allControlsDisabled=function(){var e,t;try{for(var n=Object(r.__values)(Object.keys(this.controls)),i=n.next();!i.done;i=n.next()){var o=i.value;if(this.controls[o].enabled)return!1}}catch(t){e={error:t}}finally{try{i&&!i.done&&(t=n.return)&&t.call(n)}finally{if(e)throw e.error}}return Object.keys(this.controls).length>0||this.disabled},t.prototype._checkAllValuesPresent=function(e){this._forEachChild(function(t,n){if(void 0===e[n])throw new Error("Must supply a value for form control with name: '"+n+"'.")})},t}(ye),xe=function(e){function t(t,n,r){var i=e.call(this,ge(n),ve(r,n))||this;return i.controls=t,i._initObservables(),i._setUpdateStrategy(n),i._setUpControls(),i.updateValueAndValidity({onlySelf:!0,emitEvent:!1}),i}return Object(r.__extends)(t,e),t.prototype.at=function(e){return this.controls[e]},t.prototype.push=function(e){this.controls.push(e),this._registerControl(e),this.updateValueAndValidity(),this._onCollectionChange()},t.prototype.insert=function(e,t){this.controls.splice(e,0,t),this._registerControl(t),this.updateValueAndValidity()},t.prototype.removeAt=function(e){this.controls[e]&&this.controls[e]._registerOnCollectionChange(function(){}),this.controls.splice(e,1),this.updateValueAndValidity()},t.prototype.setControl=function(e,t){this.controls[e]&&this.controls[e]._registerOnCollectionChange(function(){}),this.controls.splice(e,1),t&&(this.controls.splice(e,0,t),this._registerControl(t)),this.updateValueAndValidity(),this._onCollectionChange()},Object.defineProperty(t.prototype,"length",{get:function(){return this.controls.length},enumerable:!0,configurable:!0}),t.prototype.setValue=function(e,t){var n=this;void 0===t&&(t={}),this._checkAllValuesPresent(e),e.forEach(function(e,r){n._throwIfControlMissing(r),n.at(r).setValue(e,{onlySelf:!0,emitEvent:t.emitEvent})}),this.updateValueAndValidity(t)},t.prototype.patchValue=function(e,t){var n=this;void 0===t&&(t={}),e.forEach(function(e,r){n.at(r)&&n.at(r).patchValue(e,{onlySelf:!0,emitEvent:t.emitEvent})}),this.updateValueAndValidity(t)},t.prototype.reset=function(e,t){void 0===e&&(e=[]),void 0===t&&(t={}),this._forEachChild(function(n,r){n.reset(e[r],{onlySelf:!0,emitEvent:t.emitEvent})}),this.updateValueAndValidity(t),this._updatePristine(t),this._updateTouched(t)},t.prototype.getRawValue=function(){return this.controls.map(function(e){return e instanceof be?e.value:e.getRawValue()})},t.prototype._syncPendingControls=function(){var e=this.controls.reduce(function(e,t){return!!t._syncPendingControls()||e},!1);return e&&this.updateValueAndValidity({onlySelf:!0}),e},t.prototype._throwIfControlMissing=function(e){if(!this.controls.length)throw new Error("\n        There are no form controls registered with this array yet.  If you're using ngModel,\n        you may want to check next tick (e.g. use setTimeout).\n      ");if(!this.at(e))throw new Error("Cannot find form control at index "+e)},t.prototype._forEachChild=function(e){this.controls.forEach(function(t,n){e(t,n)})},t.prototype._updateValue=function(){var e=this;this.value=this.controls.filter(function(t){return t.enabled||e.disabled}).map(function(e){return e.value})},t.prototype._anyControls=function(e){return this.controls.some(function(t){return t.enabled&&e(t)})},t.prototype._setUpControls=function(){var e=this;this._forEachChild(function(t){return e._registerControl(t)})},t.prototype._checkAllValuesPresent=function(e){this._forEachChild(function(t,n){if(void 0===e[n])throw new Error("Must supply a value for form control at index: "+n+".")})},t.prototype._allControlsDisabled=function(){var e,t;try{for(var n=Object(r.__values)(this.controls),i=n.next();!i.done;i=n.next()){if(i.value.enabled)return!1}}catch(t){e={error:t}}finally{try{i&&!i.done&&(t=n.return)&&t.call(n)}finally{if(e)throw e.error}}return this.controls.length>0||this.disabled},t.prototype._registerControl=function(e){e.setParent(this),e._registerOnCollectionChange(this._onCollectionChange)},t}(ye),Oe={provide:c,useExisting:Object(i.forwardRef)(function(){return Ce})},we=Promise.resolve(null),Ce=function(e){function t(t,n){var r=e.call(this)||this;return r.submitted=!1,r._directives=[],r.ngSubmit=new i.EventEmitter,r.form=new je({},te(t),ne(n)),r}return Object(r.__extends)(t,e),t.prototype.ngAfterViewInit=function(){this._setUpdateStrategy()},Object.defineProperty(t.prototype,"formDirective",{get:function(){return this},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"control",{get:function(){return this.form},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"path",{get:function(){return[]},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"controls",{get:function(){return this.form.controls},enumerable:!0,configurable:!0}),t.prototype.addControl=function(e){var t=this;we.then(function(){var n=t._findContainer(e.path);e.control=n.registerControl(e.name,e.control),X(e.control,e),e.control.updateValueAndValidity({emitEvent:!1}),t._directives.push(e)})},t.prototype.getControl=function(e){return this.form.get(e.path)},t.prototype.removeControl=function(e){var t=this;we.then(function(){var n=t._findContainer(e.path);n&&n.removeControl(e.name),se(t._directives,e)})},t.prototype.addFormGroup=function(e){var t=this;we.then(function(){var n=t._findContainer(e.path),r=new je({});$(r,e),n.registerControl(e.name,r),r.updateValueAndValidity({emitEvent:!1})})},t.prototype.removeFormGroup=function(e){var t=this;we.then(function(){var n=t._findContainer(e.path);n&&n.removeControl(e.name)})},t.prototype.getFormGroup=function(e){return this.form.get(e.path)},t.prototype.updateModel=function(e,t){var n=this;we.then(function(){n.form.get(e.path).setValue(t)})},t.prototype.setValue=function(e){this.control.setValue(e)},t.prototype.onSubmit=function(e){return this.submitted=!0,oe(this.form,this._directives),this.ngSubmit.emit(e),!1},t.prototype.onReset=function(){this.resetForm()},t.prototype.resetForm=function(e){void 0===e&&(e=void 0),this.form.reset(e),this.submitted=!1},t.prototype._setUpdateStrategy=function(){this.options&&null!=this.options.updateOn&&(this.form._updateOn=this.options.updateOn)},t.prototype._findContainer=function(e){return e.pop(),e.length?this.form.get(e):this.form},Object(r.__decorate)([Object(i.Input)("ngFormOptions"),Object(r.__metadata)("design:type",Object)],t.prototype,"options",void 0),t=Object(r.__decorate)([Object(i.Directive)({selector:"form:not([ngNoForm]):not([formGroup]),ngForm,ng-form,[ngForm]",providers:[Oe],host:{"(submit)":"onSubmit($event)","(reset)":"onReset()"},outputs:["ngSubmit"],exportAs:"ngForm"}),Object(r.__param)(0,Object(i.Optional)()),Object(r.__param)(0,Object(i.Self)()),Object(r.__param)(0,Object(i.Inject)(d)),Object(r.__param)(1,Object(i.Optional)()),Object(r.__param)(1,Object(i.Self)()),Object(r.__param)(1,Object(i.Inject)(p)),Object(r.__metadata)("design:paramtypes",[Array,Array])],t)}(c),Ee=function(){function e(){}return e.modelParentException=function(){throw new Error('\n      ngModel cannot be used to register form controls with a parent formGroup directive.  Try using\n      formGroup\'s partner directive "formControlName" instead.  Example:\n\n      '+R+"\n\n      Or, if you'd like to avoid registering this form control, indicate that it's standalone in ngModelOptions:\n\n      Example:\n\n      "+V)},e.formGroupNameException=function(){throw new Error("\n      ngModel cannot be used to register form controls with a parent formGroupName or formArrayName directive.\n\n      Option 1: Use formControlName instead of ngModel (reactive strategy):\n\n      "+N+"\n\n      Option 2:  Update ngModel's parent be ngModelGroup (template-driven strategy):\n\n      "+L)},e.missingNameException=function(){throw new Error('If ngModel is used within a form tag, either the name attribute must be set or the form\n      control must be defined as \'standalone\' in ngModelOptions.\n\n      Example 1: <input [(ngModel)]="person.firstName" name="first">\n      Example 2: <input [(ngModel)]="person.firstName" [ngModelOptions]="{standalone: true}">')},e.modelGroupParentException=function(){throw new Error("\n      ngModelGroup cannot be used with a parent formGroup directive.\n\n      Option 1: Use formGroupName instead of ngModelGroup (reactive strategy):\n\n      "+N+"\n\n      Option 2:  Use a regular form tag instead of the formGroup directive (template-driven strategy):\n\n      "+L)},e.ngFormWarning=function(){console.warn("\n    It looks like you're using 'ngForm'.\n\n    Support for using the 'ngForm' element selector has been deprecated in Angular v6 and will be removed\n    in Angular v9.\n\n    Use 'ng-form' instead.\n\n    Before:\n    <ngForm #myForm=\"ngForm\">\n\n    After:\n    <ng-form #myForm=\"ngForm\">\n    ")},e}(),Se=new i.InjectionToken("NgFormSelectorWarning"),ke=function(){function e(e){(e&&"once"!==e||t._ngFormWarning)&&"always"!==e||(Ee.ngFormWarning(),t._ngFormWarning=!0)}var t;return t=e,e._ngFormWarning=!1,e=t=Object(r.__decorate)([Object(i.Directive)({selector:"ngForm"}),Object(r.__param)(0,Object(i.Optional)()),Object(r.__param)(0,Object(i.Inject)(Se)),Object(r.__metadata)("design:paramtypes",[Object])],e)}(),Me={provide:c,useExisting:Object(i.forwardRef)(function(){return Ae})},Ae=function(e){function t(t,n,r){var i=e.call(this)||this;return i._parent=t,i._validators=n,i._asyncValidators=r,i}var n;return Object(r.__extends)(t,e),n=t,t.prototype._checkParentType=function(){this._parent instanceof n||this._parent instanceof Ce||Ee.modelGroupParentException()},Object(r.__decorate)([Object(i.Input)("ngModelGroup"),Object(r.__metadata)("design:type",String)],t.prototype,"name",void 0),t=n=Object(r.__decorate)([Object(i.Directive)({selector:"[ngModelGroup]",providers:[Me],exportAs:"ngModelGroup"}),Object(r.__param)(0,Object(i.Host)()),Object(r.__param)(0,Object(i.SkipSelf)()),Object(r.__param)(1,Object(i.Optional)()),Object(r.__param)(1,Object(i.Self)()),Object(r.__param)(1,Object(i.Inject)(d)),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Self)()),Object(r.__param)(2,Object(i.Inject)(p)),Object(r.__metadata)("design:paramtypes",[c,Array,Array])],t)}(ce),Te={provide:M,useExisting:Object(i.forwardRef)(function(){return Ie})},De=Promise.resolve(null),Ie=function(e){function t(t,n,r,o){var a=e.call(this)||this;return a.control=new be,a._registered=!1,a.update=new i.EventEmitter,a._parent=t,a._rawValidators=n||[],a._rawAsyncValidators=r||[],a.valueAccessor=ae(a,o),a}return Object(r.__extends)(t,e),t.prototype.ngOnChanges=function(e){this._checkForErrors(),this._registered||this._setUpControl(),"isDisabled"in e&&this._updateDisabled(e),re(e,this.viewModel)&&(this._updateValue(this.model),this.viewModel=this.model)},t.prototype.ngOnDestroy=function(){this.formDirective&&this.formDirective.removeControl(this)},Object.defineProperty(t.prototype,"path",{get:function(){return this._parent?Q(this.name,this._parent):[this.name]},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"formDirective",{get:function(){return this._parent?this._parent.formDirective:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"validator",{get:function(){return te(this._rawValidators)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"asyncValidator",{get:function(){return ne(this._rawAsyncValidators)},enumerable:!0,configurable:!0}),t.prototype.viewToModelUpdate=function(e){this.viewModel=e,this.update.emit(e)},t.prototype._setUpControl=function(){this._setUpdateStrategy(),this._isStandalone()?this._setUpStandalone():this.formDirective.addControl(this),this._registered=!0},t.prototype._setUpdateStrategy=function(){this.options&&null!=this.options.updateOn&&(this.control._updateOn=this.options.updateOn)},t.prototype._isStandalone=function(){return!this._parent||!(!this.options||!this.options.standalone)},t.prototype._setUpStandalone=function(){X(this.control,this),this.control.updateValueAndValidity({emitEvent:!1})},t.prototype._checkForErrors=function(){this._isStandalone()||this._checkParentType(),this._checkName()},t.prototype._checkParentType=function(){!(this._parent instanceof Ae)&&this._parent instanceof ce?Ee.formGroupNameException():this._parent instanceof Ae||this._parent instanceof Ce||Ee.modelParentException()},t.prototype._checkName=function(){this.options&&this.options.name&&(this.name=this.options.name),this._isStandalone()||this.name||Ee.missingNameException()},t.prototype._updateValue=function(e){var t=this;De.then(function(){t.control.setValue(e,{emitViewToModelChange:!1})})},t.prototype._updateDisabled=function(e){var t=this,n=e.isDisabled.currentValue,r=""===n||n&&"false"!==n;De.then(function(){r&&!t.control.disabled?t.control.disable():!r&&t.control.disabled&&t.control.enable()})},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",String)],t.prototype,"name",void 0),Object(r.__decorate)([Object(i.Input)("disabled"),Object(r.__metadata)("design:type",Boolean)],t.prototype,"isDisabled",void 0),Object(r.__decorate)([Object(i.Input)("ngModel"),Object(r.__metadata)("design:type",Object)],t.prototype,"model",void 0),Object(r.__decorate)([Object(i.Input)("ngModelOptions"),Object(r.__metadata)("design:type",Object)],t.prototype,"options",void 0),Object(r.__decorate)([Object(i.Output)("ngModelChange"),Object(r.__metadata)("design:type",Object)],t.prototype,"update",void 0),t=Object(r.__decorate)([Object(i.Directive)({selector:"[ngModel]:not([formControlName]):not([formControl])",providers:[Te],exportAs:"ngModel"}),Object(r.__param)(0,Object(i.Optional)()),Object(r.__param)(0,Object(i.Host)()),Object(r.__param)(1,Object(i.Optional)()),Object(r.__param)(1,Object(i.Self)()),Object(r.__param)(1,Object(i.Inject)(d)),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Self)()),Object(r.__param)(2,Object(i.Inject)(p)),Object(r.__param)(3,Object(i.Optional)()),Object(r.__param)(3,Object(i.Self)()),Object(r.__param)(3,Object(i.Inject)(_)),Object(r.__metadata)("design:paramtypes",[c,Array,Array,Array])],t)}(M),Pe=new i.InjectionToken("NgModelWithFormControlWarning"),Re={provide:M,useExisting:Object(i.forwardRef)(function(){return Ne})},Ne=function(e){function t(t,n,r,o){var a=e.call(this)||this;return a._ngModelWarningConfig=o,a.update=new i.EventEmitter,a._ngModelWarningSent=!1,a._rawValidators=t||[],a._rawAsyncValidators=n||[],a.valueAccessor=ae(a,r),a}var n;return Object(r.__extends)(t,e),n=t,Object.defineProperty(t.prototype,"isDisabled",{set:function(e){B.disabledAttrWarning()},enumerable:!0,configurable:!0}),t.prototype.ngOnChanges=function(e){this._isControlChanged(e)&&(X(this.form,this),this.control.disabled&&this.valueAccessor.setDisabledState&&this.valueAccessor.setDisabledState(!0),this.form.updateValueAndValidity({emitEvent:!1})),re(e,this.viewModel)&&(ue("formControl",n,this,this._ngModelWarningConfig),this.form.setValue(this.model),this.viewModel=this.model)},Object.defineProperty(t.prototype,"path",{get:function(){return[]},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"validator",{get:function(){return te(this._rawValidators)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"asyncValidator",{get:function(){return ne(this._rawAsyncValidators)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"control",{get:function(){return this.form},enumerable:!0,configurable:!0}),t.prototype.viewToModelUpdate=function(e){this.viewModel=e,this.update.emit(e)},t.prototype._isControlChanged=function(e){return e.hasOwnProperty("form")},t._ngModelWarningSentOnce=!1,Object(r.__decorate)([Object(i.Input)("formControl"),Object(r.__metadata)("design:type",be)],t.prototype,"form",void 0),Object(r.__decorate)([Object(i.Input)("disabled"),Object(r.__metadata)("design:type",Boolean),Object(r.__metadata)("design:paramtypes",[Boolean])],t.prototype,"isDisabled",null),Object(r.__decorate)([Object(i.Input)("ngModel"),Object(r.__metadata)("design:type",Object)],t.prototype,"model",void 0),Object(r.__decorate)([Object(i.Output)("ngModelChange"),Object(r.__metadata)("design:type",Object)],t.prototype,"update",void 0),t=n=Object(r.__decorate)([Object(i.Directive)({selector:"[formControl]",providers:[Re],exportAs:"ngForm"}),Object(r.__param)(0,Object(i.Optional)()),Object(r.__param)(0,Object(i.Self)()),Object(r.__param)(0,Object(i.Inject)(d)),Object(r.__param)(1,Object(i.Optional)()),Object(r.__param)(1,Object(i.Self)()),Object(r.__param)(1,Object(i.Inject)(p)),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Self)()),Object(r.__param)(2,Object(i.Inject)(_)),Object(r.__param)(3,Object(i.Optional)()),Object(r.__param)(3,Object(i.Inject)(Pe)),Object(r.__metadata)("design:paramtypes",[Array,Array,Array,Object])],t)}(M),Fe={provide:c,useExisting:Object(i.forwardRef)(function(){return Le})},Le=function(e){function t(t,n){var r=e.call(this)||this;return r._validators=t,r._asyncValidators=n,r.submitted=!1,r.directives=[],r.form=null,r.ngSubmit=new i.EventEmitter,r}return Object(r.__extends)(t,e),t.prototype.ngOnChanges=function(e){this._checkFormPresent(),e.hasOwnProperty("form")&&(this._updateValidators(),this._updateDomValue(),this._updateRegistrations())},Object.defineProperty(t.prototype,"formDirective",{get:function(){return this},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"control",{get:function(){return this.form},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"path",{get:function(){return[]},enumerable:!0,configurable:!0}),t.prototype.addControl=function(e){var t=this.form.get(e.path);return X(t,e),t.updateValueAndValidity({emitEvent:!1}),this.directives.push(e),t},t.prototype.getControl=function(e){return this.form.get(e.path)},t.prototype.removeControl=function(e){se(this.directives,e)},t.prototype.addFormGroup=function(e){var t=this.form.get(e.path);$(t,e),t.updateValueAndValidity({emitEvent:!1})},t.prototype.removeFormGroup=function(e){},t.prototype.getFormGroup=function(e){return this.form.get(e.path)},t.prototype.addFormArray=function(e){var t=this.form.get(e.path);$(t,e),t.updateValueAndValidity({emitEvent:!1})},t.prototype.removeFormArray=function(e){},t.prototype.getFormArray=function(e){return this.form.get(e.path)},t.prototype.updateModel=function(e,t){this.form.get(e.path).setValue(t)},t.prototype.onSubmit=function(e){return this.submitted=!0,oe(this.form,this.directives),this.ngSubmit.emit(e),!1},t.prototype.onReset=function(){this.resetForm()},t.prototype.resetForm=function(e){void 0===e&&(e=void 0),this.form.reset(e),this.submitted=!1},t.prototype._updateDomValue=function(){var e=this;this.directives.forEach(function(t){var n,r,i=e.form.get(t.path);t.control!==i&&(n=t.control,(r=t).valueAccessor.registerOnChange(function(){return J(r)}),r.valueAccessor.registerOnTouched(function(){return J(r)}),r._rawValidators.forEach(function(e){e.registerOnValidatorChange&&e.registerOnValidatorChange(null)}),r._rawAsyncValidators.forEach(function(e){e.registerOnValidatorChange&&e.registerOnValidatorChange(null)}),n&&n._clearChangeFns(),i&&X(i,t),t.control=i)}),this.form._updateTreeValidity({emitEvent:!1})},t.prototype._updateRegistrations=function(){var e=this;this.form._registerOnCollectionChange(function(){return e._updateDomValue()}),this._oldForm&&this._oldForm._registerOnCollectionChange(function(){}),this._oldForm=this.form},t.prototype._updateValidators=function(){var e=te(this._validators);this.form.validator=h.compose([this.form.validator,e]);var t=ne(this._asyncValidators);this.form.asyncValidator=h.composeAsync([this.form.asyncValidator,t])},t.prototype._checkFormPresent=function(){this.form||B.missingFormException()},Object(r.__decorate)([Object(i.Input)("formGroup"),Object(r.__metadata)("design:type",je)],t.prototype,"form",void 0),Object(r.__decorate)([Object(i.Output)(),Object(r.__metadata)("design:type",Object)],t.prototype,"ngSubmit",void 0),t=Object(r.__decorate)([Object(i.Directive)({selector:"[formGroup]",providers:[Fe],host:{"(submit)":"onSubmit($event)","(reset)":"onReset()"},exportAs:"ngForm"}),Object(r.__param)(0,Object(i.Optional)()),Object(r.__param)(0,Object(i.Self)()),Object(r.__param)(0,Object(i.Inject)(d)),Object(r.__param)(1,Object(i.Optional)()),Object(r.__param)(1,Object(i.Self)()),Object(r.__param)(1,Object(i.Inject)(p)),Object(r.__metadata)("design:paramtypes",[Array,Array])],t)}(c),Ve={provide:c,useExisting:Object(i.forwardRef)(function(){return Be})},Be=function(e){function t(t,n,r){var i=e.call(this)||this;return i._parent=t,i._validators=n,i._asyncValidators=r,i}return Object(r.__extends)(t,e),t.prototype._checkParentType=function(){He(this._parent)&&B.groupParentException()},Object(r.__decorate)([Object(i.Input)("formGroupName"),Object(r.__metadata)("design:type",String)],t.prototype,"name",void 0),t=Object(r.__decorate)([Object(i.Directive)({selector:"[formGroupName]",providers:[Ve]}),Object(r.__param)(0,Object(i.Optional)()),Object(r.__param)(0,Object(i.Host)()),Object(r.__param)(0,Object(i.SkipSelf)()),Object(r.__param)(1,Object(i.Optional)()),Object(r.__param)(1,Object(i.Self)()),Object(r.__param)(1,Object(i.Inject)(d)),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Self)()),Object(r.__param)(2,Object(i.Inject)(p)),Object(r.__metadata)("design:paramtypes",[c,Array,Array])],t)}(ce),ze={provide:c,useExisting:Object(i.forwardRef)(function(){return Ue})},Ue=function(e){function t(t,n,r){var i=e.call(this)||this;return i._parent=t,i._validators=n,i._asyncValidators=r,i}return Object(r.__extends)(t,e),t.prototype.ngOnInit=function(){this._checkParentType(),this.formDirective.addFormArray(this)},t.prototype.ngOnDestroy=function(){this.formDirective&&this.formDirective.removeFormArray(this)},Object.defineProperty(t.prototype,"control",{get:function(){return this.formDirective.getFormArray(this)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"formDirective",{get:function(){return this._parent?this._parent.formDirective:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"path",{get:function(){return Q(this.name,this._parent)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"validator",{get:function(){return te(this._validators)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"asyncValidator",{get:function(){return ne(this._asyncValidators)},enumerable:!0,configurable:!0}),t.prototype._checkParentType=function(){He(this._parent)&&B.arrayParentException()},Object(r.__decorate)([Object(i.Input)("formArrayName"),Object(r.__metadata)("design:type",String)],t.prototype,"name",void 0),t=Object(r.__decorate)([Object(i.Directive)({selector:"[formArrayName]",providers:[ze]}),Object(r.__param)(0,Object(i.Optional)()),Object(r.__param)(0,Object(i.Host)()),Object(r.__param)(0,Object(i.SkipSelf)()),Object(r.__param)(1,Object(i.Optional)()),Object(r.__param)(1,Object(i.Self)()),Object(r.__param)(1,Object(i.Inject)(d)),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Self)()),Object(r.__param)(2,Object(i.Inject)(p)),Object(r.__metadata)("design:paramtypes",[c,Array,Array])],t)}(c);function He(e){return!(e instanceof Be||e instanceof Le||e instanceof Ue)}var qe={provide:M,useExisting:Object(i.forwardRef)(function(){return Ge})},Ge=function(e){function t(t,n,r,o,a){var s=e.call(this)||this;return s._ngModelWarningConfig=a,s._added=!1,s.update=new i.EventEmitter,s._ngModelWarningSent=!1,s._parent=t,s._rawValidators=n||[],s._rawAsyncValidators=r||[],s.valueAccessor=ae(s,o),s}var n;return Object(r.__extends)(t,e),n=t,Object.defineProperty(t.prototype,"isDisabled",{set:function(e){B.disabledAttrWarning()},enumerable:!0,configurable:!0}),t.prototype.ngOnChanges=function(e){this._added||this._setUpControl(),re(e,this.viewModel)&&(ue("formControlName",n,this,this._ngModelWarningConfig),this.viewModel=this.model,this.formDirective.updateModel(this,this.model))},t.prototype.ngOnDestroy=function(){this.formDirective&&this.formDirective.removeControl(this)},t.prototype.viewToModelUpdate=function(e){this.viewModel=e,this.update.emit(e)},Object.defineProperty(t.prototype,"path",{get:function(){return Q(this.name,this._parent)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"formDirective",{get:function(){return this._parent?this._parent.formDirective:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"validator",{get:function(){return te(this._rawValidators)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"asyncValidator",{get:function(){return ne(this._rawAsyncValidators)},enumerable:!0,configurable:!0}),t.prototype._checkParentType=function(){!(this._parent instanceof Be)&&this._parent instanceof ce?B.ngModelGroupException():this._parent instanceof Be||this._parent instanceof Le||this._parent instanceof Ue||B.controlParentException()},t.prototype._setUpControl=function(){this._checkParentType(),this.control=this.formDirective.addControl(this),this.control.disabled&&this.valueAccessor.setDisabledState&&this.valueAccessor.setDisabledState(!0),this._added=!0},t._ngModelWarningSentOnce=!1,Object(r.__decorate)([Object(i.Input)("formControlName"),Object(r.__metadata)("design:type",String)],t.prototype,"name",void 0),Object(r.__decorate)([Object(i.Input)("disabled"),Object(r.__metadata)("design:type",Boolean),Object(r.__metadata)("design:paramtypes",[Boolean])],t.prototype,"isDisabled",null),Object(r.__decorate)([Object(i.Input)("ngModel"),Object(r.__metadata)("design:type",Object)],t.prototype,"model",void 0),Object(r.__decorate)([Object(i.Output)("ngModelChange"),Object(r.__metadata)("design:type",Object)],t.prototype,"update",void 0),t=n=Object(r.__decorate)([Object(i.Directive)({selector:"[formControlName]",providers:[qe]}),Object(r.__param)(0,Object(i.Optional)()),Object(r.__param)(0,Object(i.Host)()),Object(r.__param)(0,Object(i.SkipSelf)()),Object(r.__param)(1,Object(i.Optional)()),Object(r.__param)(1,Object(i.Self)()),Object(r.__param)(1,Object(i.Inject)(d)),Object(r.__param)(2,Object(i.Optional)()),Object(r.__param)(2,Object(i.Self)()),Object(r.__param)(2,Object(i.Inject)(p)),Object(r.__param)(3,Object(i.Optional)()),Object(r.__param)(3,Object(i.Self)()),Object(r.__param)(3,Object(i.Inject)(_)),Object(r.__param)(4,Object(i.Optional)()),Object(r.__param)(4,Object(i.Inject)(Pe)),Object(r.__metadata)("design:paramtypes",[c,Array,Array,Array,Object])],t)}(M),We={provide:d,useExisting:Object(i.forwardRef)(function(){return Ke}),multi:!0},Ye={provide:d,useExisting:Object(i.forwardRef)(function(){return Qe}),multi:!0},Ke=function(){function e(){}return Object.defineProperty(e.prototype,"required",{get:function(){return this._required},set:function(e){this._required=null!=e&&!1!==e&&""+e!="false",this._onChange&&this._onChange()},enumerable:!0,configurable:!0}),e.prototype.validate=function(e){return this.required?h.required(e):null},e.prototype.registerOnValidatorChange=function(e){this._onChange=e},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",Object),Object(r.__metadata)("design:paramtypes",[Object])],e.prototype,"required",null),e=Object(r.__decorate)([Object(i.Directive)({selector:":not([type=checkbox])[required][formControlName],:not([type=checkbox])[required][formControl],:not([type=checkbox])[required][ngModel]",providers:[We],host:{"[attr.required]":'required ? "" : null'}})],e)}(),Qe=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.validate=function(e){return this.required?h.requiredTrue(e):null},t=Object(r.__decorate)([Object(i.Directive)({selector:"input[type=checkbox][required][formControlName],input[type=checkbox][required][formControl],input[type=checkbox][required][ngModel]",providers:[Ye],host:{"[attr.required]":'required ? "" : null'}})],t)}(Ke),Xe={provide:d,useExisting:Object(i.forwardRef)(function(){return Ze}),multi:!0},Ze=function(){function e(){}return Object.defineProperty(e.prototype,"email",{set:function(e){this._enabled=""===e||!0===e||"true"===e,this._onChange&&this._onChange()},enumerable:!0,configurable:!0}),e.prototype.validate=function(e){return this._enabled?h.email(e):null},e.prototype.registerOnValidatorChange=function(e){this._onChange=e},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",Object),Object(r.__metadata)("design:paramtypes",[Object])],e.prototype,"email",null),e=Object(r.__decorate)([Object(i.Directive)({selector:"[email][formControlName],[email][formControl],[email][ngModel]",providers:[Xe]})],e)}(),$e={provide:d,useExisting:Object(i.forwardRef)(function(){return Je}),multi:!0},Je=function(){function e(){}return e.prototype.ngOnChanges=function(e){"minlength"in e&&(this._createValidator(),this._onChange&&this._onChange())},e.prototype.validate=function(e){return null==this.minlength?null:this._validator(e)},e.prototype.registerOnValidatorChange=function(e){this._onChange=e},e.prototype._createValidator=function(){this._validator=h.minLength(parseInt(this.minlength,10))},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",String)],e.prototype,"minlength",void 0),e=Object(r.__decorate)([Object(i.Directive)({selector:"[minlength][formControlName],[minlength][formControl],[minlength][ngModel]",providers:[$e],host:{"[attr.minlength]":"minlength ? minlength : null"}})],e)}(),et={provide:d,useExisting:Object(i.forwardRef)(function(){return tt}),multi:!0},tt=function(){function e(){}return e.prototype.ngOnChanges=function(e){"maxlength"in e&&(this._createValidator(),this._onChange&&this._onChange())},e.prototype.validate=function(e){return null!=this.maxlength?this._validator(e):null},e.prototype.registerOnValidatorChange=function(e){this._onChange=e},e.prototype._createValidator=function(){this._validator=h.maxLength(parseInt(this.maxlength,10))},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",String)],e.prototype,"maxlength",void 0),e=Object(r.__decorate)([Object(i.Directive)({selector:"[maxlength][formControlName],[maxlength][formControl],[maxlength][ngModel]",providers:[et],host:{"[attr.maxlength]":"maxlength ? maxlength : null"}})],e)}(),nt={provide:d,useExisting:Object(i.forwardRef)(function(){return rt}),multi:!0},rt=function(){function e(){}return e.prototype.ngOnChanges=function(e){"pattern"in e&&(this._createValidator(),this._onChange&&this._onChange())},e.prototype.validate=function(e){return this._validator(e)},e.prototype.registerOnValidatorChange=function(e){this._onChange=e},e.prototype._createValidator=function(){this._validator=h.pattern(this.pattern)},Object(r.__decorate)([Object(i.Input)(),Object(r.__metadata)("design:type",Object)],e.prototype,"pattern",void 0),e=Object(r.__decorate)([Object(i.Directive)({selector:"[pattern][formControlName],[pattern][formControl],[pattern][ngModel]",providers:[nt],host:{"[attr.pattern]":"pattern ? pattern : null"}})],e)}();var it=function(){function e(){}return e.prototype.group=function(e,t){void 0===t&&(t=null);var n,r=this._reduceControls(e),i=null,o=null,a=void 0;return null!=t&&(void 0!==(n=t).asyncValidators||void 0!==n.validators||void 0!==n.updateOn?(i=null!=t.validators?t.validators:null,o=null!=t.asyncValidators?t.asyncValidators:null,a=null!=t.updateOn?t.updateOn:void 0):(i=null!=t.validator?t.validator:null,o=null!=t.asyncValidator?t.asyncValidator:null)),new je(r,{asyncValidators:o,updateOn:a,validators:i})},e.prototype.control=function(e,t,n){return new be(e,t,n)},e.prototype.array=function(e,t,n){var r=this,i=e.map(function(e){return r._createControl(e)});return new xe(i,t,n)},e.prototype._reduceControls=function(e){var t=this,n={};return Object.keys(e).forEach(function(r){n[r]=t._createControl(e[r])}),n},e.prototype._createControl=function(e){if(e instanceof be||e instanceof je||e instanceof xe)return e;if(Array.isArray(e)){var t=e[0],n=e.length>1?e[1]:null,r=e.length>2?e[2]:null;return this.control(t,n,r)}return this.control(e)},e=Object(r.__decorate)([Object(i.Injectable)()],e)}(),ot=new i.Version("7.2.0"),at=function(){function e(){}return e=Object(r.__decorate)([Object(i.Directive)({selector:"form:not([ngNoForm]):not([ngNativeValidate])",host:{novalidate:""}})],e)}(),st=[at,q,K,O,S,P,b,H,Y,D,pe,fe,Ke,Je,tt,rt,Qe,Ze],ut=[Ie,Ae,Ce,ke],ct=[Ne,Le,Ge,Be,Ue],lt=function(){function e(){}return e=Object(r.__decorate)([Object(i.NgModule)({declarations:st,exports:st})],e)}(),dt=function(){function e(){}var t;return t=e,e.withConfig=function(e){return{ngModule:t,providers:[{provide:Se,useValue:e.warnOnDeprecatedNgFormSelector}]}},e=t=Object(r.__decorate)([Object(i.NgModule)({declarations:ut,providers:[T],exports:[lt,ut]})],e)}(),pt=function(){function e(){}var t;return t=e,e.withConfig=function(e){return{ngModule:t,providers:[{provide:Pe,useValue:e.warnOnNgModelWithFormControl}]}},e=t=Object(r.__decorate)([Object(i.NgModule)({declarations:[ct],providers:[it,T],exports:[lt,ct]})],e)}()},"./node_modules/@angular/forms/node_modules/tslib/tslib.es6.js":function(e,t,n){"use strict";n.r(t),n.d(t,"__extends",function(){return i}),n.d(t,"__assign",function(){return o}),n.d(t,"__rest",function(){return a}),n.d(t,"__decorate",function(){return s}),n.d(t,"__param",function(){return u}),n.d(t,"__metadata",function(){return c}),n.d(t,"__awaiter",function(){return l}),n.d(t,"__generator",function(){return d}),n.d(t,"__exportStar",function(){return p}),n.d(t,"__values",function(){return f}),n.d(t,"__read",function(){return h}),n.d(t,"__spread",function(){return m}),n.d(t,"__await",function(){return g}),n.d(t,"__asyncGenerator",function(){return v}),n.d(t,"__asyncDelegator",function(){return _}),n.d(t,"__asyncValues",function(){return y}),n.d(t,"__makeTemplateObject",function(){return b}),n.d(t,"__importStar",function(){return j}),n.d(t,"__importDefault",function(){return x});var r=function(e,t){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,t){e.__proto__=t}||function(e,t){for(var n in t)t.hasOwnProperty(n)&&(e[n]=t[n])})(e,t)};function i(e,t){function n(){this.constructor=e}r(e,t),e.prototype=null===t?Object.create(t):(n.prototype=t.prototype,new n)}var o=function(){return(o=Object.assign||function(e){for(var t,n=1,r=arguments.length;n<r;n++){t=arguments[n];for(var i in t)Object.prototype.hasOwnProperty.call(t,i)&&(e[i]=t[i])}return e}).apply(this,arguments)};function a(e,t){var n={};for(var r in e)Object.prototype.hasOwnProperty.call(e,r)&&t.indexOf(r)<0&&(n[r]=e[r]);if(null!=e&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(e);i<r.length;i++)t.indexOf(r[i])<0&&(n[r[i]]=e[r[i]])}return n}function s(e,t,n,r){var i,o=arguments.length,a=o<3?t:null===r?r=Object.getOwnPropertyDescriptor(t,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(e,t,n,r);else for(var s=e.length-1;s>=0;s--)(i=e[s])&&(a=(o<3?i(a):o>3?i(t,n,a):i(t,n))||a);return o>3&&a&&Object.defineProperty(t,n,a),a}function u(e,t){return function(n,r){t(n,r,e)}}function c(e,t){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(e,t)}function l(e,t,n,r){return new(n||(n=Promise))(function(i,o){function a(e){try{u(r.next(e))}catch(e){o(e)}}function s(e){try{u(r.throw(e))}catch(e){o(e)}}function u(e){e.done?i(e.value):new n(function(t){t(e.value)}).then(a,s)}u((r=r.apply(e,t||[])).next())})}function d(e,t){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=t.call(e,a)}catch(e){o=[6,e],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(e,t){for(var n in e)t.hasOwnProperty(n)||(t[n]=e[n])}function f(e){var t="function"==typeof Symbol&&e[Symbol.iterator],n=0;return t?t.call(e):{next:function(){return e&&n>=e.length&&(e=void 0),{value:e&&e[n++],done:!e}}}}function h(e,t){var n="function"==typeof Symbol&&e[Symbol.iterator];if(!n)return e;var r,i,o=n.call(e),a=[];try{for(;(void 0===t||t-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(e){i={error:e}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var e=[],t=0;t<arguments.length;t++)e=e.concat(h(arguments[t]));return e}function g(e){return this instanceof g?(this.v=e,this):new g(e)}function v(e,t,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(e,t||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(e){i[e]&&(r[e]=function(t){return new Promise(function(n,r){o.push([e,t,n,r])>1||s(e,t)})})}function s(e,t){try{(n=i[e](t)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(e){l(o[0][3],e)}var n}function u(e){s("next",e)}function c(e){s("throw",e)}function l(e,t){e(t),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(e){var t,n;return t={},r("next"),r("throw",function(e){throw e}),r("return"),t[Symbol.iterator]=function(){return this},t;function r(r,i){t[r]=e[r]?function(t){return(n=!n)?{value:g(e[r](t)),done:"return"===r}:i?i(t):t}:i}}function y(e){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var t,n=e[Symbol.asyncIterator];return n?n.call(e):(e=f(e),t={},r("next"),r("throw"),r("return"),t[Symbol.asyncIterator]=function(){return this},t);function r(n){t[n]=e[n]&&function(t){return new Promise(function(r,i){var o,a,s,u;t=e[n](t),o=r,a=i,s=t.done,u=t.value,Promise.resolve(u).then(function(e){o({value:e,done:s})},a)})}}}function b(e,t){return Object.defineProperty?Object.defineProperty(e,"raw",{value:t}):e.raw=t,e}function j(e){if(e&&e.__esModule)return e;var t={};if(null!=e)for(var n in e)Object.hasOwnProperty.call(e,n)&&(t[n]=e[n]);return t.default=e,t}function x(e){return e&&e.__esModule?e:{default:e}}},"./node_modules/@angular/material/esm5/autocomplete.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MAT_AUTOCOMPLETE_DEFAULT_OPTIONS_FACTORY",function(){return O}),n.d(t,"MatAutocompleteSelectedEvent",function(){return y}),n.d(t,"MatAutocompleteBase",function(){return b}),n.d(t,"_MatAutocompleteMixinBase",function(){return j}),n.d(t,"MAT_AUTOCOMPLETE_DEFAULT_OPTIONS",function(){return x}),n.d(t,"MatAutocomplete",function(){return w}),n.d(t,"MatAutocompleteModule",function(){return P}),n.d(t,"MAT_AUTOCOMPLETE_SCROLL_STRATEGY_FACTORY",function(){return M}),n.d(t,"getMatAutocompleteMissingPanelError",function(){return D}),n.d(t,"AUTOCOMPLETE_OPTION_HEIGHT",function(){return E}),n.d(t,"AUTOCOMPLETE_PANEL_HEIGHT",function(){return S}),n.d(t,"MAT_AUTOCOMPLETE_SCROLL_STRATEGY",function(){return k}),n.d(t,"MAT_AUTOCOMPLETE_SCROLL_STRATEGY_FACTORY_PROVIDER",function(){return A}),n.d(t,"MAT_AUTOCOMPLETE_VALUE_ACCESSOR",function(){return T}),n.d(t,"MatAutocompleteTrigger",function(){return I}),n.d(t,"ɵa28",function(){return C});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),o=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),s=n("./node_modules/@angular/material/esm5/core.es5.js"),u=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),c=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),l=n("./node_modules/@angular/cdk/esm5/overlay.es5.js"),d=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),p=n("./node_modules/@angular/common/fesm5/common.js"),f=n("./node_modules/rxjs/_esm5/operators/index.js"),h=n("./node_modules/@angular/cdk/esm5/scrolling.es5.js"),m=n("./node_modules/@angular/forms/fesm5/forms.js"),g=n("./node_modules/@angular/material/esm5/form-field.es5.js"),v=n("./node_modules/rxjs/_esm5/index.js"),_=0,y=function(){return function(e,t){this.source=e,this.option=t}}(),b=function(){return function(){}}(),j=Object(s.mixinDisableRipple)(b),x=new a.InjectionToken("mat-autocomplete-default-options",{providedIn:"root",factory:O});function O(){return{autoActiveFirstOption:!1}}var w=function(e){function t(t,n,r){var i=e.call(this)||this;return i._changeDetectorRef=t,i._elementRef=n,i.showPanel=!1,i._isOpen=!1,i.displayWith=null,i.optionSelected=new a.EventEmitter,i.opened=new a.EventEmitter,i.closed=new a.EventEmitter,i._classList={},i.id="mat-autocomplete-"+_++,i._autoActiveFirstOption=!!r.autoActiveFirstOption,i}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"isOpen",{get:function(){return this._isOpen&&this.showPanel},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"autoActiveFirstOption",{get:function(){return this._autoActiveFirstOption},set:function(e){this._autoActiveFirstOption=Object(o.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"classList",{set:function(e){var t=this;e&&e.length&&(e.split(" ").forEach(function(e){return t._classList[e.trim()]=!0}),this._elementRef.nativeElement.className="")},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){this._keyManager=new i.ActiveDescendantKeyManager(this.options).withWrap(),this._setVisibility()},t.prototype._setScrollTop=function(e){this.panel&&(this.panel.nativeElement.scrollTop=e)},t.prototype._getScrollTop=function(){return this.panel?this.panel.nativeElement.scrollTop:0},t.prototype._setVisibility=function(){this.showPanel=!!this.options.length,this._classList["mat-autocomplete-visible"]=this.showPanel,this._classList["mat-autocomplete-hidden"]=!this.showPanel,this._changeDetectorRef.markForCheck()},t.prototype._emitSelectEvent=function(e){var t=new y(this,e);this.optionSelected.emit(t)},t.decorators=[{type:a.Component,args:[{selector:"mat-autocomplete",template:'<ng-template><div class="mat-autocomplete-panel" role="listbox" [id]="id" [ngClass]="_classList" #panel><ng-content></ng-content></div></ng-template>',styles:[".mat-autocomplete-panel{min-width:112px;max-width:280px;overflow:auto;-webkit-overflow-scrolling:touch;visibility:hidden;max-width:none;max-height:256px;position:relative;width:100%;border-bottom-left-radius:4px;border-bottom-right-radius:4px}.mat-autocomplete-panel.mat-autocomplete-visible{visibility:visible}.mat-autocomplete-panel.mat-autocomplete-hidden{visibility:hidden}.mat-autocomplete-panel-above .mat-autocomplete-panel{border-radius:0;border-top-left-radius:4px;border-top-right-radius:4px}.mat-autocomplete-panel .mat-divider-horizontal{margin-top:-1px}@media (-ms-high-contrast:active){.mat-autocomplete-panel{outline:solid 1px}}"],encapsulation:a.ViewEncapsulation.None,changeDetection:a.ChangeDetectionStrategy.OnPush,exportAs:"matAutocomplete",inputs:["disableRipple"],host:{class:"mat-autocomplete"},providers:[{provide:s.MAT_OPTION_PARENT_COMPONENT,useExisting:t}]}]}],t.ctorParameters=function(){return[{type:a.ChangeDetectorRef},{type:a.ElementRef},{type:void 0,decorators:[{type:a.Inject,args:[x]}]}]},t.propDecorators={template:[{type:a.ViewChild,args:[a.TemplateRef]}],panel:[{type:a.ViewChild,args:["panel"]}],options:[{type:a.ContentChildren,args:[s.MatOption,{descendants:!0}]}],optionGroups:[{type:a.ContentChildren,args:[s.MatOptgroup]}],displayWith:[{type:a.Input}],autoActiveFirstOption:[{type:a.Input}],panelWidth:[{type:a.Input}],optionSelected:[{type:a.Output}],opened:[{type:a.Output}],closed:[{type:a.Output}],classList:[{type:a.Input,args:["class"]}]},t}(j),C=function(){function e(e){this.elementRef=e}return e.decorators=[{type:a.Directive,args:[{selector:"[matAutocompleteOrigin]",exportAs:"matAutocompleteOrigin"}]}],e.ctorParameters=function(){return[{type:a.ElementRef}]},e}(),E=48,S=256,k=new a.InjectionToken("mat-autocomplete-scroll-strategy");function M(e){return function(){return e.scrollStrategies.reposition()}}var A={provide:k,deps:[l.Overlay],useFactory:M},T={provide:m.NG_VALUE_ACCESSOR,useExisting:Object(a.forwardRef)(function(){return I}),multi:!0};function D(){return Error("Attempting to open an undefined instance of `mat-autocomplete`. Make sure that the id passed to the `matAutocomplete` is correct and that you're attempting to open it after the ngAfterContentInit hook.")}var I=function(){function e(e,t,n,r,i,o,a,s,u,c){var l=this;this._element=e,this._overlay=t,this._viewContainerRef=n,this._zone=r,this._changeDetectorRef=i,this._dir=a,this._formField=s,this._document=u,this._viewportRuler=c,this._componentDestroyed=!1,this._autocompleteDisabled=!1,this._manuallyFloatingLabel=!1,this._viewportSubscription=v.Subscription.EMPTY,this._canOpenOnNextFocus=!0,this._closeKeyEventStream=new v.Subject,this._windowBlurHandler=function(){l._canOpenOnNextFocus=document.activeElement!==l._element.nativeElement||l.panelOpen},this._onChange=function(){},this._onTouched=function(){},this.autocompleteAttribute="off",this._overlayAttached=!1,this.optionSelections=Object(v.defer)(function(){return l.autocomplete&&l.autocomplete.options?v.merge.apply(void 0,l.autocomplete.options.map(function(e){return e.onSelectionChange})):l._zone.onStable.asObservable().pipe(Object(f.take)(1),Object(f.switchMap)(function(){return l.optionSelections}))}),"undefined"!=typeof window&&r.runOutsideAngular(function(){window.addEventListener("blur",l._windowBlurHandler)}),this._scrollStrategy=o}return Object.defineProperty(e.prototype,"autocompleteDisabled",{get:function(){return this._autocompleteDisabled},set:function(e){this._autocompleteDisabled=Object(o.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),e.prototype.ngOnDestroy=function(){"undefined"!=typeof window&&window.removeEventListener("blur",this._windowBlurHandler),this._viewportSubscription.unsubscribe(),this._componentDestroyed=!0,this._destroyPanel(),this._closeKeyEventStream.complete()},Object.defineProperty(e.prototype,"panelOpen",{get:function(){return this._overlayAttached&&this.autocomplete.showPanel},enumerable:!0,configurable:!0}),e.prototype.openPanel=function(){this._attachOverlay(),this._floatLabel()},e.prototype.closePanel=function(){this._resetLabel(),this._overlayAttached&&(this.panelOpen&&this.autocomplete.closed.emit(),this.autocomplete._isOpen=this._overlayAttached=!1,this._overlayRef&&this._overlayRef.hasAttached()&&(this._overlayRef.detach(),this._closingActionsSubscription.unsubscribe()),this._componentDestroyed||this._changeDetectorRef.detectChanges())},e.prototype.updatePosition=function(){this._overlayAttached&&this._overlayRef.updatePosition()},Object.defineProperty(e.prototype,"panelClosingActions",{get:function(){var e=this;return Object(v.merge)(this.optionSelections,this.autocomplete._keyManager.tabOut.pipe(Object(f.filter)(function(){return e._overlayAttached})),this._closeKeyEventStream,this._getOutsideClickStream(),this._overlayRef?this._overlayRef.detachments().pipe(Object(f.filter)(function(){return e._overlayAttached})):Object(v.of)()).pipe(Object(f.map)(function(e){return e instanceof s.MatOptionSelectionChange?e:null}))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"activeOption",{get:function(){return this.autocomplete&&this.autocomplete._keyManager?this.autocomplete._keyManager.activeItem:null},enumerable:!0,configurable:!0}),e.prototype._getOutsideClickStream=function(){var e=this;return this._document?Object(v.merge)(Object(v.fromEvent)(this._document,"click"),Object(v.fromEvent)(this._document,"touchend")).pipe(Object(f.filter)(function(t){var n=t.target,r=e._formField?e._formField._elementRef.nativeElement:null;return e._overlayAttached&&n!==e._element.nativeElement&&(!r||!r.contains(n))&&!!e._overlayRef&&!e._overlayRef.overlayElement.contains(n)})):Object(v.of)(null)},e.prototype.writeValue=function(e){var t=this;Promise.resolve(null).then(function(){return t._setTriggerValue(e)})},e.prototype.registerOnChange=function(e){this._onChange=e},e.prototype.registerOnTouched=function(e){this._onTouched=e},e.prototype.setDisabledState=function(e){this._element.nativeElement.disabled=e},e.prototype._handleKeydown=function(e){var t=e.keyCode;if(t===c.ESCAPE&&e.preventDefault(),this.activeOption&&t===c.ENTER&&this.panelOpen)this.activeOption._selectViaInteraction(),this._resetActiveItem(),e.preventDefault();else if(this.autocomplete){var n=this.autocomplete._keyManager.activeItem,r=t===c.UP_ARROW||t===c.DOWN_ARROW;this.panelOpen||t===c.TAB?this.autocomplete._keyManager.onKeydown(e):r&&this._canOpen()&&this.openPanel(),(r||this.autocomplete._keyManager.activeItem!==n)&&this._scrollToOption()}},e.prototype._handleInput=function(e){var t=e.target,n=t.value;"number"===t.type&&(n=""==n?null:parseFloat(n)),this._previousValue!==n&&(this._previousValue=n,this._onChange(n),this._canOpen()&&document.activeElement===e.target&&this.openPanel())},e.prototype._handleFocus=function(){this._canOpenOnNextFocus?this._canOpen()&&(this._previousValue=this._element.nativeElement.value,this._attachOverlay(),this._floatLabel(!0)):this._canOpenOnNextFocus=!0},e.prototype._floatLabel=function(e){void 0===e&&(e=!1),this._formField&&"auto"===this._formField.floatLabel&&(e?this._formField._animateAndLockLabel():this._formField.floatLabel="always",this._manuallyFloatingLabel=!0)},e.prototype._resetLabel=function(){this._manuallyFloatingLabel&&(this._formField.floatLabel="auto",this._manuallyFloatingLabel=!1)},e.prototype._scrollToOption=function(){var e=this.autocomplete._keyManager.activeItemIndex||0,t=Object(s._countGroupLabelsBeforeOption)(e,this.autocomplete.options,this.autocomplete.optionGroups),n=Object(s._getOptionScrollPosition)(e+t,E,this.autocomplete._getScrollTop(),S);this.autocomplete._setScrollTop(n)},e.prototype._subscribeToClosingActions=function(){var e=this,t=this._zone.onStable.asObservable().pipe(Object(f.take)(1)),n=this.autocomplete.options.changes.pipe(Object(f.tap)(function(){return e._positionStrategy.reapplyLastPosition()}),Object(f.delay)(0));return Object(v.merge)(t,n).pipe(Object(f.switchMap)(function(){return e._resetActiveItem(),e.autocomplete._setVisibility(),e.panelOpen&&e._overlayRef.updatePosition(),e.panelClosingActions}),Object(f.take)(1)).subscribe(function(t){return e._setValueAndClose(t)})},e.prototype._destroyPanel=function(){this._overlayRef&&(this.closePanel(),this._overlayRef.dispose(),this._overlayRef=null)},e.prototype._setTriggerValue=function(e){var t=this.autocomplete&&this.autocomplete.displayWith?this.autocomplete.displayWith(e):e,n=null!=t?t:"";this._formField?this._formField._control.value=n:this._element.nativeElement.value=n,this._previousValue=n},e.prototype._setValueAndClose=function(e){e&&e.source&&(this._clearPreviousSelectedOption(e.source),this._setTriggerValue(e.source.value),this._onChange(e.source.value),this._element.nativeElement.focus(),this.autocomplete._emitSelectEvent(e.source)),this.closePanel()},e.prototype._clearPreviousSelectedOption=function(e){this.autocomplete.options.forEach(function(t){t!=e&&t.selected&&t.deselect()})},e.prototype._attachOverlay=function(){var e=this;if(!this.autocomplete)throw D();this._overlayRef?this._overlayRef.updateSize({width:this._getPanelWidth()}):(this._portal=new d.TemplatePortal(this.autocomplete.template,this._viewContainerRef),this._overlayRef=this._overlay.create(this._getOverlayConfig()),this._overlayRef.keydownEvents().subscribe(function(t){(t.keyCode===c.ESCAPE||t.keyCode===c.UP_ARROW&&t.altKey)&&(e._resetActiveItem(),e._closeKeyEventStream.next())}),this._viewportRuler&&(this._viewportSubscription=this._viewportRuler.change().subscribe(function(){e.panelOpen&&e._overlayRef&&e._overlayRef.updateSize({width:e._getPanelWidth()})}))),this._overlayRef&&!this._overlayRef.hasAttached()&&(this._overlayRef.attach(this._portal),this._closingActionsSubscription=this._subscribeToClosingActions());var t=this.panelOpen;this.autocomplete._setVisibility(),this.autocomplete._isOpen=this._overlayAttached=!0,this.panelOpen&&t!==this.panelOpen&&this.autocomplete.opened.emit()},e.prototype._getOverlayConfig=function(){return new l.OverlayConfig({positionStrategy:this._getOverlayPosition(),scrollStrategy:this._scrollStrategy(),width:this._getPanelWidth(),direction:this._dir})},e.prototype._getOverlayPosition=function(){return this._positionStrategy=this._overlay.position().flexibleConnectedTo(this._getConnectedElement()).withFlexibleDimensions(!1).withPush(!1).withPositions([{originX:"start",originY:"bottom",overlayX:"start",overlayY:"top"},{originX:"start",originY:"top",overlayX:"start",overlayY:"bottom",panelClass:"mat-autocomplete-panel-above"}]),this._positionStrategy},e.prototype._getConnectedElement=function(){return this.connectedTo?this.connectedTo.elementRef:this._formField?this._formField.getConnectedOverlayOrigin():this._element},e.prototype._getPanelWidth=function(){return this.autocomplete.panelWidth||this._getHostWidth()},e.prototype._getHostWidth=function(){return this._getConnectedElement().nativeElement.getBoundingClientRect().width},e.prototype._resetActiveItem=function(){this.autocomplete._keyManager.setActiveItem(this.autocomplete.autoActiveFirstOption?0:-1)},e.prototype._canOpen=function(){var e=this._element.nativeElement;return!e.readOnly&&!e.disabled&&!this._autocompleteDisabled},e.decorators=[{type:a.Directive,args:[{selector:"input[matAutocomplete], textarea[matAutocomplete]",host:{"[attr.autocomplete]":"autocompleteAttribute","[attr.role]":'autocompleteDisabled ? null : "combobox"',"[attr.aria-autocomplete]":'autocompleteDisabled ? null : "list"',"[attr.aria-activedescendant]":"(panelOpen && activeOption) ? activeOption.id : null","[attr.aria-expanded]":"autocompleteDisabled ? null : panelOpen.toString()","[attr.aria-owns]":"(autocompleteDisabled || !panelOpen) ? null : autocomplete?.id","(focusin)":"_handleFocus()","(blur)":"_onTouched()","(input)":"_handleInput($event)","(keydown)":"_handleKeydown($event)"},exportAs:"matAutocompleteTrigger",providers:[T]}]}],e.ctorParameters=function(){return[{type:a.ElementRef},{type:l.Overlay},{type:a.ViewContainerRef},{type:a.NgZone},{type:a.ChangeDetectorRef},{type:void 0,decorators:[{type:a.Inject,args:[k]}]},{type:u.Directionality,decorators:[{type:a.Optional}]},{type:g.MatFormField,decorators:[{type:a.Optional},{type:a.Host}]},{type:void 0,decorators:[{type:a.Optional},{type:a.Inject,args:[p.DOCUMENT]}]},{type:h.ViewportRuler}]},e.propDecorators={autocomplete:[{type:a.Input,args:["matAutocomplete"]}],connectedTo:[{type:a.Input,args:["matAutocompleteConnectedTo"]}],autocompleteAttribute:[{type:a.Input,args:["autocomplete"]}],autocompleteDisabled:[{type:a.Input,args:["matAutocompleteDisabled"]}]},e}(),P=function(){function e(){}return e.decorators=[{type:a.NgModule,args:[{imports:[s.MatOptionModule,l.OverlayModule,s.MatCommonModule,p.CommonModule],exports:[w,s.MatOptionModule,I,C,s.MatCommonModule],declarations:[w,I,C],providers:[A]}]}],e}()},"./node_modules/@angular/material/esm5/badge.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatBadgeModule",function(){return f}),n.d(t,"MatBadgeBase",function(){return l}),n.d(t,"_MatBadgeMixinBase",function(){return d}),n.d(t,"MatBadge",function(){return p});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),o=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),a=n("./node_modules/@angular/common/fesm5/common.js"),s=n("./node_modules/@angular/core/fesm5/core.js"),u=n("./node_modules/@angular/material/esm5/core.es5.js"),c=0,l=function(){return function(){}}(),d=Object(u.mixinDisabled)(l),p=function(e){function t(t,n,r,i,o){var a=e.call(this)||this;return a._document=t,a._ngZone=n,a._elementRef=r,a._ariaDescriber=i,a._renderer=o,a._hasContent=!1,a._color="primary",a._overlap=!0,a.position="above after",a.size="medium",a._id=c++,a}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"color",{get:function(){return this._color},set:function(e){this._setColor(e),this._color=e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"overlap",{get:function(){return this._overlap},set:function(e){this._overlap=Object(o.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"content",{get:function(){return this._content},set:function(e){this._content=e,this._hasContent=null!=e&&(""+e).trim().length>0,this._updateTextContent()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"description",{get:function(){return this._description},set:function(e){if(e!==this._description){var t=this._badgeElement;this._updateHostAriaDescription(e,this._description),this._description=e,t&&(e?t.setAttribute("aria-label",e):t.removeAttribute("aria-label"))}},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"hidden",{get:function(){return this._hidden},set:function(e){this._hidden=Object(o.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t.prototype.isAbove=function(){return-1===this.position.indexOf("below")},t.prototype.isAfter=function(){return-1===this.position.indexOf("before")},t.prototype.ngOnDestroy=function(){var e=this._badgeElement;e&&(this.description&&this._ariaDescriber.removeDescription(e,this.description),this._renderer&&this._renderer.destroyNode&&this._renderer.destroyNode(e))},t.prototype._updateTextContent=function(){return this._badgeElement?this._badgeElement.textContent=this.content:this._badgeElement=this._createBadgeElement(),this._badgeElement},t.prototype._createBadgeElement=function(){var e=(this._renderer||this._document).createElement("span");return e.setAttribute("id","mat-badge-content-"+this._id),e.classList.add("mat-badge-content"),e.textContent=this.content,this.description&&e.setAttribute("aria-label",this.description),this._elementRef.nativeElement.appendChild(e),"function"==typeof requestAnimationFrame?this._ngZone.runOutsideAngular(function(){requestAnimationFrame(function(){e.classList.add("mat-badge-active")})}):e.classList.add("mat-badge-active"),e},t.prototype._updateHostAriaDescription=function(e,t){var n=this._updateTextContent();t&&this._ariaDescriber.removeDescription(n,t),e&&this._ariaDescriber.describe(n,e)},t.prototype._setColor=function(e){e!==this._color&&(this._color&&this._elementRef.nativeElement.classList.remove("mat-badge-"+this._color),e&&this._elementRef.nativeElement.classList.add("mat-badge-"+e))},t.decorators=[{type:s.Directive,args:[{selector:"[matBadge]",inputs:["disabled: matBadgeDisabled"],host:{class:"mat-badge","[class.mat-badge-overlap]":"overlap","[class.mat-badge-above]":"isAbove()","[class.mat-badge-below]":"!isAbove()","[class.mat-badge-before]":"!isAfter()","[class.mat-badge-after]":"isAfter()","[class.mat-badge-small]":'size === "small"',"[class.mat-badge-medium]":'size === "medium"',"[class.mat-badge-large]":'size === "large"',"[class.mat-badge-hidden]":"hidden || !_hasContent","[class.mat-badge-disabled]":"disabled"}}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:s.Optional},{type:s.Inject,args:[a.DOCUMENT]}]},{type:s.NgZone},{type:s.ElementRef},{type:i.AriaDescriber},{type:s.Renderer2}]},t.propDecorators={color:[{type:s.Input,args:["matBadgeColor"]}],overlap:[{type:s.Input,args:["matBadgeOverlap"]}],position:[{type:s.Input,args:["matBadgePosition"]}],content:[{type:s.Input,args:["matBadge"]}],description:[{type:s.Input,args:["matBadgeDescription"]}],size:[{type:s.Input,args:["matBadgeSize"]}],hidden:[{type:s.Input,args:["matBadgeHidden"]}]},t}(d),f=function(){function e(){}return e.decorators=[{type:s.NgModule,args:[{imports:[i.A11yModule,u.MatCommonModule],exports:[p],declarations:[p]}]}],e}()},"./node_modules/@angular/material/esm5/bottom-sheet.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatBottomSheetModule",function(){return b}),n.d(t,"MAT_BOTTOM_SHEET_DEFAULT_OPTIONS",function(){return x}),n.d(t,"MatBottomSheet",function(){return O}),n.d(t,"MAT_BOTTOM_SHEET_DATA",function(){return g}),n.d(t,"MatBottomSheetConfig",function(){return v}),n.d(t,"MatBottomSheetContainer",function(){return y}),n.d(t,"matBottomSheetAnimations",function(){return _}),n.d(t,"MatBottomSheetRef",function(){return j});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/animations/fesm5/animations.js"),o=n("./node_modules/@angular/material/esm5/core.es5.js"),a=n("./node_modules/tslib/tslib.es6.js"),s=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),u=n("./node_modules/@angular/cdk/esm5/layout.es5.js"),c=n("./node_modules/@angular/common/fesm5/common.js"),l=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),d=n("./node_modules/@angular/cdk/esm5/overlay.es5.js"),p=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),f=n("./node_modules/rxjs/_esm5/index.js"),h=n("./node_modules/rxjs/_esm5/operators/index.js"),m=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),g=new r.InjectionToken("MatBottomSheetData"),v=function(){return function(){this.data=null,this.hasBackdrop=!0,this.disableClose=!1,this.ariaLabel=null,this.closeOnNavigation=!0,this.autoFocus=!1,this.restoreFocus=!0}}(),_={bottomSheetState:Object(i.trigger)("state",[Object(i.state)("void, hidden",Object(i.style)({transform:"translateY(100%)"})),Object(i.state)("visible",Object(i.style)({transform:"translateY(0%)"})),Object(i.transition)("visible => void, visible => hidden",Object(i.animate)(o.AnimationDurations.COMPLEX+" "+o.AnimationCurves.ACCELERATION_CURVE)),Object(i.transition)("void => visible",Object(i.animate)(o.AnimationDurations.EXITING+" "+o.AnimationCurves.DECELERATION_CURVE))])},y=function(e){function t(t,n,i,o,a,s){var c=e.call(this)||this;return c._elementRef=t,c._changeDetectorRef=n,c._focusTrapFactory=i,c.bottomSheetConfig=s,c._animationState="void",c._animationStateChanged=new r.EventEmitter,c._elementFocusedBeforeOpened=null,c._document=a,c._breakpointSubscription=o.observe([u.Breakpoints.Medium,u.Breakpoints.Large,u.Breakpoints.XLarge]).subscribe(function(){c._toggleClass("mat-bottom-sheet-container-medium",o.isMatched(u.Breakpoints.Medium)),c._toggleClass("mat-bottom-sheet-container-large",o.isMatched(u.Breakpoints.Large)),c._toggleClass("mat-bottom-sheet-container-xlarge",o.isMatched(u.Breakpoints.XLarge))}),c}return Object(a.__extends)(t,e),t.prototype.attachComponentPortal=function(e){return this._validatePortalAttached(),this._setPanelClass(),this._savePreviouslyFocusedElement(),this._portalOutlet.attachComponentPortal(e)},t.prototype.attachTemplatePortal=function(e){return this._validatePortalAttached(),this._setPanelClass(),this._savePreviouslyFocusedElement(),this._portalOutlet.attachTemplatePortal(e)},t.prototype.enter=function(){this._destroyed||(this._animationState="visible",this._changeDetectorRef.detectChanges())},t.prototype.exit=function(){this._destroyed||(this._animationState="hidden",this._changeDetectorRef.markForCheck())},t.prototype.ngOnDestroy=function(){this._breakpointSubscription.unsubscribe(),this._destroyed=!0},t.prototype._onAnimationDone=function(e){"hidden"===e.toState?this._restoreFocus():"visible"===e.toState&&this.bottomSheetConfig.autoFocus&&this._trapFocus(),this._animationStateChanged.emit(e)},t.prototype._onAnimationStart=function(e){this._animationStateChanged.emit(e)},t.prototype._toggleClass=function(e,t){var n=this._elementRef.nativeElement.classList;t?n.add(e):n.remove(e)},t.prototype._validatePortalAttached=function(){if(this._portalOutlet.hasAttached())throw Error("Attempting to attach bottom sheet content after content is already attached")},t.prototype._setPanelClass=function(){var e=this._elementRef.nativeElement,t=this.bottomSheetConfig.panelClass;Array.isArray(t)?t.forEach(function(t){return e.classList.add(t)}):t&&e.classList.add(t)},t.prototype._trapFocus=function(){this._focusTrap||(this._focusTrap=this._focusTrapFactory.create(this._elementRef.nativeElement)),this._focusTrap.focusInitialElementWhenReady()},t.prototype._restoreFocus=function(){var e=this._elementFocusedBeforeOpened;this.bottomSheetConfig.restoreFocus&&e&&"function"==typeof e.focus&&e.focus(),this._focusTrap&&this._focusTrap.destroy()},t.prototype._savePreviouslyFocusedElement=function(){var e=this;this._elementFocusedBeforeOpened=this._document.activeElement,this._elementRef.nativeElement.focus&&Promise.resolve().then(function(){return e._elementRef.nativeElement.focus()})},t.decorators=[{type:r.Component,args:[{selector:"mat-bottom-sheet-container",template:"<ng-template cdkPortalOutlet></ng-template>",styles:[".mat-bottom-sheet-container{padding:8px 16px;min-width:100vw;box-sizing:border-box;display:block;outline:0;max-height:80vh;overflow:auto}@media (-ms-high-contrast:active){.mat-bottom-sheet-container{outline:1px solid}}.mat-bottom-sheet-container-large,.mat-bottom-sheet-container-medium,.mat-bottom-sheet-container-xlarge{border-top-left-radius:4px;border-top-right-radius:4px}.mat-bottom-sheet-container-medium{min-width:384px;max-width:calc(100vw - 128px)}.mat-bottom-sheet-container-large{min-width:512px;max-width:calc(100vw - 256px)}.mat-bottom-sheet-container-xlarge{min-width:576px;max-width:calc(100vw - 384px)}"],changeDetection:r.ChangeDetectionStrategy.OnPush,encapsulation:r.ViewEncapsulation.None,animations:[_.bottomSheetState],host:{class:"mat-bottom-sheet-container",tabindex:"-1",role:"dialog","aria-modal":"true","[attr.aria-label]":"bottomSheetConfig?.ariaLabel","[@state]":"_animationState","(@state.start)":"_onAnimationStart($event)","(@state.done)":"_onAnimationDone($event)"}}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:r.ChangeDetectorRef},{type:l.FocusTrapFactory},{type:u.BreakpointObserver},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[c.DOCUMENT]}]},{type:v}]},t.propDecorators={_portalOutlet:[{type:r.ViewChild,args:[s.CdkPortalOutlet]}]},t}(s.BasePortalOutlet),b=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[c.CommonModule,d.OverlayModule,o.MatCommonModule,s.PortalModule],exports:[y,o.MatCommonModule],declarations:[y],entryComponents:[y]}]}],e}(),j=function(){function e(e,t,n){var r=this;this._overlayRef=t,this._afterDismissed=new f.Subject,this._afterOpened=new f.Subject,this.containerInstance=e,e._animationStateChanged.pipe(Object(h.filter)(function(e){return"done"===e.phaseName&&"visible"===e.toState}),Object(h.take)(1)).subscribe(function(){r._afterOpened.next(),r._afterOpened.complete()}),e._animationStateChanged.pipe(Object(h.filter)(function(e){return"done"===e.phaseName&&"hidden"===e.toState}),Object(h.take)(1)).subscribe(function(){r._overlayRef.dispose(),r._afterDismissed.next(r._result),r._afterDismissed.complete()}),e.bottomSheetConfig.disableClose||Object(f.merge)(t.backdropClick(),t.keydownEvents().pipe(Object(h.filter)(function(e){return e.keyCode===p.ESCAPE}))).subscribe(function(){return r.dismiss()})}return e.prototype.dismiss=function(e){var t=this;this._afterDismissed.closed||(this.containerInstance._animationStateChanged.pipe(Object(h.filter)(function(e){return"start"===e.phaseName}),Object(h.take)(1)).subscribe(function(){return t._overlayRef.detachBackdrop()}),this._result=e,this.containerInstance.exit())},e.prototype.afterDismissed=function(){return this._afterDismissed.asObservable()},e.prototype.afterOpened=function(){return this._afterOpened.asObservable()},e.prototype.backdropClick=function(){return this._overlayRef.backdropClick()},e.prototype.keydownEvents=function(){return this._overlayRef.keydownEvents()},e}(),x=new r.InjectionToken("mat-bottom-sheet-default-options"),O=function(){function e(e,t,n,r,i){this._overlay=e,this._injector=t,this._parentBottomSheet=n,this._location=r,this._defaultOptions=i,this._bottomSheetRefAtThisLevel=null}return Object.defineProperty(e.prototype,"_openedBottomSheetRef",{get:function(){var e=this._parentBottomSheet;return e?e._openedBottomSheetRef:this._bottomSheetRefAtThisLevel},set:function(e){this._parentBottomSheet?this._parentBottomSheet._openedBottomSheetRef=e:this._bottomSheetRefAtThisLevel=e},enumerable:!0,configurable:!0}),e.prototype.open=function(e,t){var n,i,o=this,u=(n=this._defaultOptions||new v,i=t,Object(a.__assign)({},n,i)),c=this._createOverlay(u),l=this._attachContainer(c,u),d=new j(l,c,this._location);if(e instanceof r.TemplateRef)l.attachTemplatePortal(new s.TemplatePortal(e,null,{$implicit:u.data,bottomSheetRef:d}));else{var p=new s.ComponentPortal(e,void 0,this._createInjector(u,d)),f=l.attachComponentPortal(p);d.instance=f.instance}return d.afterDismissed().subscribe(function(){o._openedBottomSheetRef==d&&(o._openedBottomSheetRef=null)}),this._openedBottomSheetRef?(this._openedBottomSheetRef.afterDismissed().subscribe(function(){return d.containerInstance.enter()}),this._openedBottomSheetRef.dismiss()):d.containerInstance.enter(),this._openedBottomSheetRef=d,d},e.prototype.dismiss=function(){this._openedBottomSheetRef&&this._openedBottomSheetRef.dismiss()},e.prototype.ngOnDestroy=function(){this._bottomSheetRefAtThisLevel&&this._bottomSheetRefAtThisLevel.dismiss()},e.prototype._attachContainer=function(e,t){var n=t&&t.viewContainerRef&&t.viewContainerRef.injector,r=new s.PortalInjector(n||this._injector,new WeakMap([[v,t]])),i=new s.ComponentPortal(y,t.viewContainerRef,r);return e.attach(i).instance},e.prototype._createOverlay=function(e){var t=new d.OverlayConfig({direction:e.direction,hasBackdrop:e.hasBackdrop,disposeOnNavigation:e.closeOnNavigation,maxWidth:"100%",scrollStrategy:this._overlay.scrollStrategies.block(),positionStrategy:this._overlay.position().global().centerHorizontally().bottom("0")});return e.backdropClass&&(t.backdropClass=e.backdropClass),this._overlay.create(t)},e.prototype._createInjector=function(e,t){var n=e&&e.viewContainerRef&&e.viewContainerRef.injector,r=new WeakMap([[j,t],[g,e.data]]);return!e.direction||n&&n.get(m.Directionality,null)||r.set(m.Directionality,{value:e.direction,change:Object(f.of)()}),new s.PortalInjector(n||this._injector,r)},e.decorators=[{type:r.Injectable,args:[{providedIn:b}]}],e.ctorParameters=function(){return[{type:d.Overlay},{type:r.Injector},{type:e,decorators:[{type:r.Optional},{type:r.SkipSelf}]},{type:c.Location,decorators:[{type:r.Optional}]},{type:v,decorators:[{type:r.Optional},{type:r.Inject,args:[x]}]}]},e.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new e(Object(r.inject)(d.Overlay),Object(r.inject)(r.INJECTOR),Object(r.inject)(e,12),Object(r.inject)(c.Location,8),Object(r.inject)(x,8))},token:e,providedIn:b}),e}()},"./node_modules/@angular/material/esm5/button-toggle.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MAT_BUTTON_TOGGLE_DEFAULT_OPTIONS",function(){return l}),n.d(t,"MAT_BUTTON_TOGGLE_GROUP_VALUE_ACCESSOR",function(){return d}),n.d(t,"MatButtonToggleGroupMultiple",function(){return p}),n.d(t,"MatButtonToggleChange",function(){return h}),n.d(t,"MatButtonToggleGroup",function(){return m}),n.d(t,"MatButtonToggleBase",function(){return g}),n.d(t,"_MatButtonToggleMixinBase",function(){return v}),n.d(t,"MatButtonToggle",function(){return _}),n.d(t,"MatButtonToggleModule",function(){return y});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),o=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),a=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),s=n("./node_modules/@angular/core/fesm5/core.js"),u=n("./node_modules/@angular/forms/fesm5/forms.js"),c=n("./node_modules/@angular/material/esm5/core.es5.js"),l=new s.InjectionToken("MAT_BUTTON_TOGGLE_DEFAULT_OPTIONS"),d={provide:u.NG_VALUE_ACCESSOR,useExisting:Object(s.forwardRef)(function(){return m}),multi:!0},p=function(){return function(){}}(),f=0,h=function(){return function(e,t){this.source=e,this.value=t}}(),m=function(){function e(e,t){this._changeDetector=e,this._vertical=!1,this._multiple=!1,this._disabled=!1,this._controlValueAccessorChangeFn=function(){},this._onTouched=function(){},this._name="mat-button-toggle-group-"+f++,this.valueChange=new s.EventEmitter,this.change=new s.EventEmitter,this.appearance=t&&t.appearance?t.appearance:"standard"}return Object.defineProperty(e.prototype,"name",{get:function(){return this._name},set:function(e){var t=this;this._name=e,this._buttonToggles&&this._buttonToggles.forEach(function(e){return e.name=t._name})},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"vertical",{get:function(){return this._vertical},set:function(e){this._vertical=Object(o.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){var e=this._selectionModel?this._selectionModel.selected:[];return this.multiple?e.map(function(e){return e.value}):e[0]?e[0].value:void 0},set:function(e){this._setSelectionByValue(e),this.valueChange.emit(this.value)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selected",{get:function(){var e=this._selectionModel.selected;return this.multiple?e:e[0]||null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"multiple",{get:function(){return this._multiple},set:function(e){this._multiple=Object(o.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disabled",{get:function(){return this._disabled},set:function(e){this._disabled=Object(o.coerceBooleanProperty)(e),this._buttonToggles&&this._buttonToggles.forEach(function(e){return e._markForCheck()})},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){this._selectionModel=new a.SelectionModel(this.multiple,void 0,!1)},e.prototype.ngAfterContentInit=function(){var e;(e=this._selectionModel).select.apply(e,this._buttonToggles.filter(function(e){return e.checked}))},e.prototype.writeValue=function(e){this.value=e,this._changeDetector.markForCheck()},e.prototype.registerOnChange=function(e){this._controlValueAccessorChangeFn=e},e.prototype.registerOnTouched=function(e){this._onTouched=e},e.prototype.setDisabledState=function(e){this.disabled=e},e.prototype._emitChangeEvent=function(){var e=this.selected,t=Array.isArray(e)?e[e.length-1]:e,n=new h(t,this.value);this._controlValueAccessorChangeFn(n.value),this.change.emit(n)},e.prototype._syncButtonToggle=function(e,t,n){void 0===n&&(n=!1),this.multiple||!this.selected||e.checked||(this.selected.checked=!1),t?this._selectionModel.select(e):this._selectionModel.deselect(e),n&&this._emitChangeEvent(),this.valueChange.emit(this.value)},e.prototype._isSelected=function(e){return this._selectionModel.isSelected(e)},e.prototype._isPrechecked=function(e){return void 0!==this._rawValue&&(this.multiple&&Array.isArray(this._rawValue)?this._rawValue.some(function(t){return null!=e.value&&t===e.value}):e.value===this._rawValue)},e.prototype._setSelectionByValue=function(e){var t=this;if(this._rawValue=e,this._buttonToggles)if(this.multiple&&e){if(!Array.isArray(e))throw Error("Value must be an array in multiple-selection mode.");this._clearSelection(),e.forEach(function(e){return t._selectValue(e)})}else this._clearSelection(),this._selectValue(e)},e.prototype._clearSelection=function(){this._selectionModel.clear(),this._buttonToggles.forEach(function(e){return e.checked=!1})},e.prototype._selectValue=function(e){var t=this._buttonToggles.find(function(t){return null!=t.value&&t.value===e});t&&(t.checked=!0,this._selectionModel.select(t))},e.decorators=[{type:s.Directive,args:[{selector:"mat-button-toggle-group",providers:[d,{provide:p,useExisting:e}],host:{role:"group",class:"mat-button-toggle-group","[attr.aria-disabled]":"disabled","[class.mat-button-toggle-vertical]":"vertical","[class.mat-button-toggle-group-appearance-standard]":'appearance === "standard"'},exportAs:"matButtonToggleGroup"}]}],e.ctorParameters=function(){return[{type:s.ChangeDetectorRef},{type:void 0,decorators:[{type:s.Optional},{type:s.Inject,args:[l]}]}]},e.propDecorators={_buttonToggles:[{type:s.ContentChildren,args:[Object(s.forwardRef)(function(){return _})]}],appearance:[{type:s.Input}],name:[{type:s.Input}],vertical:[{type:s.Input}],value:[{type:s.Input}],valueChange:[{type:s.Output}],multiple:[{type:s.Input}],disabled:[{type:s.Input}],change:[{type:s.Output}]},e}(),g=function(){return function(){}}(),v=Object(c.mixinDisableRipple)(g),_=function(e){function t(t,n,r,i,o,a){var u=e.call(this)||this;u._changeDetectorRef=n,u._elementRef=r,u._focusMonitor=i,u._isSingleSelector=!1,u._checked=!1,u.ariaLabelledby=null,u._disabled=!1,u.change=new s.EventEmitter;var c=Number(o);return u.tabIndex=c||0===c?c:null,u.buttonToggleGroup=t,u.appearance=a&&a.appearance?a.appearance:"standard",u}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"buttonId",{get:function(){return this.id+"-button"},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"appearance",{get:function(){return this.buttonToggleGroup?this.buttonToggleGroup.appearance:this._appearance},set:function(e){this._appearance=e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"checked",{get:function(){return this.buttonToggleGroup?this.buttonToggleGroup._isSelected(this):this._checked},set:function(e){var t=Object(o.coerceBooleanProperty)(e);t!==this._checked&&(this._checked=t,this.buttonToggleGroup&&this.buttonToggleGroup._syncButtonToggle(this,this._checked),this._changeDetectorRef.markForCheck())},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return this._disabled||this.buttonToggleGroup&&this.buttonToggleGroup.disabled},set:function(e){this._disabled=Object(o.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){this._isSingleSelector=this.buttonToggleGroup&&!this.buttonToggleGroup.multiple,this._type=this._isSingleSelector?"radio":"checkbox",this.id=this.id||"mat-button-toggle-"+f++,this._isSingleSelector&&(this.name=this.buttonToggleGroup.name),this.buttonToggleGroup&&this.buttonToggleGroup._isPrechecked(this)&&(this.checked=!0),this._focusMonitor.monitor(this._elementRef,!0)},t.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._elementRef)},t.prototype.focus=function(){this._buttonElement.nativeElement.focus()},t.prototype._onButtonClick=function(){var e=!!this._isSingleSelector||!this._checked;e!==this._checked&&(this._checked=e,this.buttonToggleGroup&&(this.buttonToggleGroup._syncButtonToggle(this,this._checked,!0),this.buttonToggleGroup._onTouched())),this.change.emit(new h(this,this.value))},t.prototype._markForCheck=function(){this._changeDetectorRef.markForCheck()},t.decorators=[{type:s.Component,args:[{selector:"mat-button-toggle",template:'<button #button class="mat-button-toggle-button" type="button" [id]="buttonId" [attr.tabindex]="disabled ? -1 : tabIndex" [attr.aria-pressed]="checked" [disabled]="disabled || null" [attr.name]="name || null" [attr.aria-label]="ariaLabel" [attr.aria-labelledby]="ariaLabelledby" (click)="_onButtonClick()"><div class="mat-button-toggle-label-content"><ng-content></ng-content></div></button><div class="mat-button-toggle-focus-overlay"></div><div class="mat-button-toggle-ripple" matRipple [matRippleTrigger]="button" [matRippleDisabled]="this.disableRipple || this.disabled"></div>',styles:[".mat-button-toggle-group,.mat-button-toggle-standalone{position:relative;display:inline-flex;flex-direction:row;white-space:nowrap;overflow:hidden;border-radius:2px;-webkit-tap-highlight-color:transparent}@media (-ms-high-contrast:active){.mat-button-toggle-group,.mat-button-toggle-standalone{outline:solid 1px}}.mat-button-toggle-group-appearance-standard,.mat-button-toggle-standalone.mat-button-toggle-appearance-standard{border-radius:4px}@media (-ms-high-contrast:active){.mat-button-toggle-group-appearance-standard,.mat-button-toggle-standalone.mat-button-toggle-appearance-standard{outline:0}}.mat-button-toggle-vertical{flex-direction:column}.mat-button-toggle-vertical .mat-button-toggle-label-content{display:block}.mat-button-toggle{white-space:nowrap;position:relative}.mat-button-toggle .mat-icon svg{vertical-align:top}.mat-button-toggle.cdk-keyboard-focused .mat-button-toggle-focus-overlay{opacity:1}@media (-ms-high-contrast:active){.mat-button-toggle.cdk-keyboard-focused .mat-button-toggle-focus-overlay{opacity:.5}}.mat-button-toggle-appearance-standard:not(.mat-button-toggle-disabled):hover .mat-button-toggle-focus-overlay{opacity:.04}.mat-button-toggle-appearance-standard.cdk-keyboard-focused:not(.mat-button-toggle-disabled) .mat-button-toggle-focus-overlay{opacity:.12}@media (-ms-high-contrast:active){.mat-button-toggle-appearance-standard.cdk-keyboard-focused:not(.mat-button-toggle-disabled) .mat-button-toggle-focus-overlay{opacity:.5}}@media (hover:none){.mat-button-toggle-appearance-standard:not(.mat-button-toggle-disabled):hover .mat-button-toggle-focus-overlay{display:none}}.mat-button-toggle-label-content{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;display:inline-block;line-height:36px;padding:0 16px;position:relative}.mat-button-toggle-appearance-standard .mat-button-toggle-label-content{line-height:48px;padding:0 12px}.mat-button-toggle-label-content>*{vertical-align:middle}.mat-button-toggle-focus-overlay{border-radius:inherit;pointer-events:none;opacity:0;top:0;left:0;right:0;bottom:0;position:absolute}.mat-button-toggle-checked .mat-button-toggle-focus-overlay{border-bottom:solid 36px}@media (-ms-high-contrast:active){.mat-button-toggle-checked .mat-button-toggle-focus-overlay{opacity:.5;height:0}}@media (-ms-high-contrast:active){.mat-button-toggle-checked.mat-button-toggle-appearance-standard .mat-button-toggle-focus-overlay{border-bottom:solid 48px}}.mat-button-toggle .mat-button-toggle-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-button-toggle-button{border:0;background:0 0;color:inherit;padding:0;margin:0;font:inherit;outline:0;width:100%;cursor:pointer}.mat-button-toggle-disabled .mat-button-toggle-button{cursor:default}.mat-button-toggle-button::-moz-focus-inner{border:0}"],encapsulation:s.ViewEncapsulation.None,exportAs:"matButtonToggle",changeDetection:s.ChangeDetectionStrategy.OnPush,inputs:["disableRipple"],host:{"[class.mat-button-toggle-standalone]":"!buttonToggleGroup","[class.mat-button-toggle-checked]":"checked","[class.mat-button-toggle-disabled]":"disabled","[class.mat-button-toggle-appearance-standard]":'appearance === "standard"',class:"mat-button-toggle","[attr.tabindex]":"-1","[attr.id]":"id","(focus)":"focus()"}}]}],t.ctorParameters=function(){return[{type:m,decorators:[{type:s.Optional}]},{type:s.ChangeDetectorRef},{type:s.ElementRef},{type:i.FocusMonitor},{type:String,decorators:[{type:s.Attribute,args:["tabindex"]}]},{type:void 0,decorators:[{type:s.Optional},{type:s.Inject,args:[l]}]}]},t.propDecorators={ariaLabel:[{type:s.Input,args:["aria-label"]}],ariaLabelledby:[{type:s.Input,args:["aria-labelledby"]}],_buttonElement:[{type:s.ViewChild,args:["button"]}],id:[{type:s.Input}],name:[{type:s.Input}],value:[{type:s.Input}],tabIndex:[{type:s.Input}],appearance:[{type:s.Input}],checked:[{type:s.Input}],disabled:[{type:s.Input}],change:[{type:s.Output}]},t}(v),y=function(){function e(){}return e.decorators=[{type:s.NgModule,args:[{imports:[c.MatCommonModule,c.MatRippleModule],exports:[c.MatCommonModule,m,_],declarations:[m,_]}]}],e}()},"./node_modules/@angular/material/esm5/button.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatButtonModule",function(){return g}),n.d(t,"MatButtonBase",function(){return p}),n.d(t,"_MatButtonMixinBase",function(){return f}),n.d(t,"MatButton",function(){return h}),n.d(t,"MatAnchor",function(){return m});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),o=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),s=n("./node_modules/@angular/material/esm5/core.es5.js"),u=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),c=n("./node_modules/@angular/common/fesm5/common.js"),l="accent",d=["mat-button","mat-flat-button","mat-icon-button","mat-raised-button","mat-stroked-button","mat-mini-fab","mat-fab"],p=function(){return function(e){this._elementRef=e}}(),f=Object(s.mixinColor)(Object(s.mixinDisabled)(Object(s.mixinDisableRipple)(p))),h=function(e){function t(t,n,r,i){var o=e.call(this,t)||this;o._focusMonitor=r,o._animationMode=i,o.isRoundButton=o._hasHostAttributes("mat-fab","mat-mini-fab"),o.isIconButton=o._hasHostAttributes("mat-icon-button");for(var a=0,s=d;a<s.length;a++){var u=s[a];o._hasHostAttributes(u)&&t.nativeElement.classList.add(u)}return o._focusMonitor.monitor(o._elementRef,!0),o.isRoundButton&&(o.color=l),o}return Object(r.__extends)(t,e),t.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._elementRef)},t.prototype.focus=function(){this._getHostElement().focus()},t.prototype._getHostElement=function(){return this._elementRef.nativeElement},t.prototype._isRippleDisabled=function(){return this.disableRipple||this.disabled},t.prototype._hasHostAttributes=function(){for(var e=this,t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];return t.some(function(t){return e._getHostElement().hasAttribute(t)})},t.decorators=[{type:a.Component,args:[{selector:"button[mat-button], button[mat-raised-button], button[mat-icon-button],\n             button[mat-fab], button[mat-mini-fab], button[mat-stroked-button],\n             button[mat-flat-button]",exportAs:"matButton",host:{"[disabled]":"disabled || null","[class._mat-animation-noopable]":'_animationMode === "NoopAnimations"'},template:'<span class="mat-button-wrapper"><ng-content></ng-content></span><div matRipple class="mat-button-ripple" [class.mat-button-ripple-round]="isRoundButton || isIconButton" [matRippleDisabled]="_isRippleDisabled()" [matRippleCentered]="isIconButton" [matRippleTrigger]="_getHostElement()"></div><div class="mat-button-focus-overlay"></div>',styles:[".mat-button .mat-button-focus-overlay,.mat-icon-button .mat-button-focus-overlay{opacity:0}.mat-button:hover .mat-button-focus-overlay,.mat-stroked-button:hover .mat-button-focus-overlay{opacity:.04}@media (hover:none){.mat-button:hover .mat-button-focus-overlay,.mat-stroked-button:hover .mat-button-focus-overlay{opacity:0}}.mat-button,.mat-flat-button,.mat-icon-button,.mat-stroked-button{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:64px;line-height:36px;padding:0 16px;border-radius:4px;overflow:visible}.mat-button::-moz-focus-inner,.mat-flat-button::-moz-focus-inner,.mat-icon-button::-moz-focus-inner,.mat-stroked-button::-moz-focus-inner{border:0}.mat-button[disabled],.mat-flat-button[disabled],.mat-icon-button[disabled],.mat-stroked-button[disabled]{cursor:default}.mat-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-button.cdk-program-focused .mat-button-focus-overlay,.mat-flat-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-flat-button.cdk-program-focused .mat-button-focus-overlay,.mat-icon-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-icon-button.cdk-program-focused .mat-button-focus-overlay,.mat-stroked-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-stroked-button.cdk-program-focused .mat-button-focus-overlay{opacity:.12}.mat-button::-moz-focus-inner,.mat-flat-button::-moz-focus-inner,.mat-icon-button::-moz-focus-inner,.mat-stroked-button::-moz-focus-inner{border:0}.mat-raised-button{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:64px;line-height:36px;padding:0 16px;border-radius:4px;overflow:visible;transform:translate3d(0,0,0);transition:background .4s cubic-bezier(.25,.8,.25,1),box-shadow 280ms cubic-bezier(.4,0,.2,1)}.mat-raised-button::-moz-focus-inner{border:0}.mat-raised-button[disabled]{cursor:default}.mat-raised-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-raised-button.cdk-program-focused .mat-button-focus-overlay{opacity:.12}.mat-raised-button::-moz-focus-inner{border:0}._mat-animation-noopable.mat-raised-button{transition:none;animation:none}.mat-stroked-button{border:1px solid currentColor;padding:0 15px;line-height:34px}.mat-stroked-button .mat-button-focus-overlay,.mat-stroked-button .mat-button-ripple.mat-ripple{top:-1px;left:-1px;right:-1px;bottom:-1px}.mat-fab{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:64px;line-height:36px;padding:0 16px;border-radius:4px;overflow:visible;transform:translate3d(0,0,0);transition:background .4s cubic-bezier(.25,.8,.25,1),box-shadow 280ms cubic-bezier(.4,0,.2,1);min-width:0;border-radius:50%;width:56px;height:56px;padding:0;flex-shrink:0}.mat-fab::-moz-focus-inner{border:0}.mat-fab[disabled]{cursor:default}.mat-fab.cdk-keyboard-focused .mat-button-focus-overlay,.mat-fab.cdk-program-focused .mat-button-focus-overlay{opacity:.12}.mat-fab::-moz-focus-inner{border:0}._mat-animation-noopable.mat-fab{transition:none;animation:none}.mat-fab .mat-button-wrapper{padding:16px 0;display:inline-block;line-height:24px}.mat-mini-fab{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:64px;line-height:36px;padding:0 16px;border-radius:4px;overflow:visible;transform:translate3d(0,0,0);transition:background .4s cubic-bezier(.25,.8,.25,1),box-shadow 280ms cubic-bezier(.4,0,.2,1);min-width:0;border-radius:50%;width:40px;height:40px;padding:0;flex-shrink:0}.mat-mini-fab::-moz-focus-inner{border:0}.mat-mini-fab[disabled]{cursor:default}.mat-mini-fab.cdk-keyboard-focused .mat-button-focus-overlay,.mat-mini-fab.cdk-program-focused .mat-button-focus-overlay{opacity:.12}.mat-mini-fab::-moz-focus-inner{border:0}._mat-animation-noopable.mat-mini-fab{transition:none;animation:none}.mat-mini-fab .mat-button-wrapper{padding:8px 0;display:inline-block;line-height:24px}.mat-icon-button{padding:0;min-width:0;width:40px;height:40px;flex-shrink:0;line-height:40px;border-radius:50%}.mat-icon-button .mat-icon,.mat-icon-button i{line-height:24px}.mat-button-focus-overlay,.mat-button-ripple.mat-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none;border-radius:inherit}.mat-button-focus-overlay{opacity:0;transition:opacity .2s cubic-bezier(.35,0,.25,1),background-color .2s cubic-bezier(.35,0,.25,1)}._mat-animation-noopable .mat-button-focus-overlay{transition:none}@media (-ms-high-contrast:active){.mat-button-focus-overlay{background-color:rgba(255,255,255,.5)}}.mat-button-ripple-round{border-radius:50%;z-index:1}.mat-button .mat-button-wrapper>*,.mat-fab .mat-button-wrapper>*,.mat-flat-button .mat-button-wrapper>*,.mat-icon-button .mat-button-wrapper>*,.mat-mini-fab .mat-button-wrapper>*,.mat-raised-button .mat-button-wrapper>*,.mat-stroked-button .mat-button-wrapper>*{vertical-align:middle}.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-prefix .mat-icon-button,.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-suffix .mat-icon-button{display:block;font-size:inherit;width:2.5em;height:2.5em}@media (-ms-high-contrast:active){.mat-button,.mat-fab,.mat-flat-button,.mat-icon-button,.mat-mini-fab,.mat-raised-button{outline:solid 1px}}"],inputs:["disabled","disableRipple","color"],encapsulation:a.ViewEncapsulation.None,changeDetection:a.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:a.ElementRef},{type:o.Platform},{type:i.FocusMonitor},{type:String,decorators:[{type:a.Optional},{type:a.Inject,args:[u.ANIMATION_MODULE_TYPE]}]}]},t.propDecorators={ripple:[{type:a.ViewChild,args:[s.MatRipple]}]},t}(f),m=function(e){function t(t,n,r,i){return e.call(this,r,t,n,i)||this}return Object(r.__extends)(t,e),t.prototype._haltDisabledEvents=function(e){this.disabled&&(e.preventDefault(),e.stopImmediatePropagation())},t.decorators=[{type:a.Component,args:[{selector:"a[mat-button], a[mat-raised-button], a[mat-icon-button], a[mat-fab],\n             a[mat-mini-fab], a[mat-stroked-button], a[mat-flat-button]",exportAs:"matButton, matAnchor",host:{"[attr.tabindex]":"disabled ? -1 : (tabIndex || 0)","[attr.disabled]":"disabled || null","[attr.aria-disabled]":"disabled.toString()","(click)":"_haltDisabledEvents($event)","[class._mat-animation-noopable]":'_animationMode === "NoopAnimations"'},inputs:["disabled","disableRipple","color"],template:'<span class="mat-button-wrapper"><ng-content></ng-content></span><div matRipple class="mat-button-ripple" [class.mat-button-ripple-round]="isRoundButton || isIconButton" [matRippleDisabled]="_isRippleDisabled()" [matRippleCentered]="isIconButton" [matRippleTrigger]="_getHostElement()"></div><div class="mat-button-focus-overlay"></div>',styles:[".mat-button .mat-button-focus-overlay,.mat-icon-button .mat-button-focus-overlay{opacity:0}.mat-button:hover .mat-button-focus-overlay,.mat-stroked-button:hover .mat-button-focus-overlay{opacity:.04}@media (hover:none){.mat-button:hover .mat-button-focus-overlay,.mat-stroked-button:hover .mat-button-focus-overlay{opacity:0}}.mat-button,.mat-flat-button,.mat-icon-button,.mat-stroked-button{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:64px;line-height:36px;padding:0 16px;border-radius:4px;overflow:visible}.mat-button::-moz-focus-inner,.mat-flat-button::-moz-focus-inner,.mat-icon-button::-moz-focus-inner,.mat-stroked-button::-moz-focus-inner{border:0}.mat-button[disabled],.mat-flat-button[disabled],.mat-icon-button[disabled],.mat-stroked-button[disabled]{cursor:default}.mat-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-button.cdk-program-focused .mat-button-focus-overlay,.mat-flat-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-flat-button.cdk-program-focused .mat-button-focus-overlay,.mat-icon-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-icon-button.cdk-program-focused .mat-button-focus-overlay,.mat-stroked-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-stroked-button.cdk-program-focused .mat-button-focus-overlay{opacity:.12}.mat-button::-moz-focus-inner,.mat-flat-button::-moz-focus-inner,.mat-icon-button::-moz-focus-inner,.mat-stroked-button::-moz-focus-inner{border:0}.mat-raised-button{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:64px;line-height:36px;padding:0 16px;border-radius:4px;overflow:visible;transform:translate3d(0,0,0);transition:background .4s cubic-bezier(.25,.8,.25,1),box-shadow 280ms cubic-bezier(.4,0,.2,1)}.mat-raised-button::-moz-focus-inner{border:0}.mat-raised-button[disabled]{cursor:default}.mat-raised-button.cdk-keyboard-focused .mat-button-focus-overlay,.mat-raised-button.cdk-program-focused .mat-button-focus-overlay{opacity:.12}.mat-raised-button::-moz-focus-inner{border:0}._mat-animation-noopable.mat-raised-button{transition:none;animation:none}.mat-stroked-button{border:1px solid currentColor;padding:0 15px;line-height:34px}.mat-stroked-button .mat-button-focus-overlay,.mat-stroked-button .mat-button-ripple.mat-ripple{top:-1px;left:-1px;right:-1px;bottom:-1px}.mat-fab{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:64px;line-height:36px;padding:0 16px;border-radius:4px;overflow:visible;transform:translate3d(0,0,0);transition:background .4s cubic-bezier(.25,.8,.25,1),box-shadow 280ms cubic-bezier(.4,0,.2,1);min-width:0;border-radius:50%;width:56px;height:56px;padding:0;flex-shrink:0}.mat-fab::-moz-focus-inner{border:0}.mat-fab[disabled]{cursor:default}.mat-fab.cdk-keyboard-focused .mat-button-focus-overlay,.mat-fab.cdk-program-focused .mat-button-focus-overlay{opacity:.12}.mat-fab::-moz-focus-inner{border:0}._mat-animation-noopable.mat-fab{transition:none;animation:none}.mat-fab .mat-button-wrapper{padding:16px 0;display:inline-block;line-height:24px}.mat-mini-fab{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:64px;line-height:36px;padding:0 16px;border-radius:4px;overflow:visible;transform:translate3d(0,0,0);transition:background .4s cubic-bezier(.25,.8,.25,1),box-shadow 280ms cubic-bezier(.4,0,.2,1);min-width:0;border-radius:50%;width:40px;height:40px;padding:0;flex-shrink:0}.mat-mini-fab::-moz-focus-inner{border:0}.mat-mini-fab[disabled]{cursor:default}.mat-mini-fab.cdk-keyboard-focused .mat-button-focus-overlay,.mat-mini-fab.cdk-program-focused .mat-button-focus-overlay{opacity:.12}.mat-mini-fab::-moz-focus-inner{border:0}._mat-animation-noopable.mat-mini-fab{transition:none;animation:none}.mat-mini-fab .mat-button-wrapper{padding:8px 0;display:inline-block;line-height:24px}.mat-icon-button{padding:0;min-width:0;width:40px;height:40px;flex-shrink:0;line-height:40px;border-radius:50%}.mat-icon-button .mat-icon,.mat-icon-button i{line-height:24px}.mat-button-focus-overlay,.mat-button-ripple.mat-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none;border-radius:inherit}.mat-button-focus-overlay{opacity:0;transition:opacity .2s cubic-bezier(.35,0,.25,1),background-color .2s cubic-bezier(.35,0,.25,1)}._mat-animation-noopable .mat-button-focus-overlay{transition:none}@media (-ms-high-contrast:active){.mat-button-focus-overlay{background-color:rgba(255,255,255,.5)}}.mat-button-ripple-round{border-radius:50%;z-index:1}.mat-button .mat-button-wrapper>*,.mat-fab .mat-button-wrapper>*,.mat-flat-button .mat-button-wrapper>*,.mat-icon-button .mat-button-wrapper>*,.mat-mini-fab .mat-button-wrapper>*,.mat-raised-button .mat-button-wrapper>*,.mat-stroked-button .mat-button-wrapper>*{vertical-align:middle}.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-prefix .mat-icon-button,.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-suffix .mat-icon-button{display:block;font-size:inherit;width:2.5em;height:2.5em}@media (-ms-high-contrast:active){.mat-button,.mat-fab,.mat-flat-button,.mat-icon-button,.mat-mini-fab,.mat-raised-button{outline:solid 1px}}"],encapsulation:a.ViewEncapsulation.None,changeDetection:a.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:o.Platform},{type:i.FocusMonitor},{type:a.ElementRef},{type:String,decorators:[{type:a.Optional},{type:a.Inject,args:[u.ANIMATION_MODULE_TYPE]}]}]},t.propDecorators={tabIndex:[{type:a.Input}]},t}(h),g=function(){function e(){}return e.decorators=[{type:a.NgModule,args:[{imports:[c.CommonModule,s.MatRippleModule,s.MatCommonModule],exports:[h,m,s.MatCommonModule],declarations:[h,m]}]}],e}()},"./node_modules/@angular/material/esm5/card.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatCardContent",function(){return o}),n.d(t,"MatCardTitle",function(){return a}),n.d(t,"MatCardSubtitle",function(){return s}),n.d(t,"MatCardActions",function(){return u}),n.d(t,"MatCardFooter",function(){return c}),n.d(t,"MatCardImage",function(){return l}),n.d(t,"MatCardSmImage",function(){return d}),n.d(t,"MatCardMdImage",function(){return p}),n.d(t,"MatCardLgImage",function(){return f}),n.d(t,"MatCardXlImage",function(){return h}),n.d(t,"MatCardAvatar",function(){return m}),n.d(t,"MatCard",function(){return g}),n.d(t,"MatCardHeader",function(){return v}),n.d(t,"MatCardTitleGroup",function(){return _}),n.d(t,"MatCardModule",function(){return y});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/material/esm5/core.es5.js"),o=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"mat-card-content",host:{class:"mat-card-content"}}]}],e}(),a=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"mat-card-title, [mat-card-title], [matCardTitle]",host:{class:"mat-card-title"}}]}],e}(),s=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"mat-card-subtitle, [mat-card-subtitle], [matCardSubtitle]",host:{class:"mat-card-subtitle"}}]}],e}(),u=function(){function e(){this.align="start"}return e.decorators=[{type:r.Directive,args:[{selector:"mat-card-actions",exportAs:"matCardActions",host:{class:"mat-card-actions","[class.mat-card-actions-align-end]":'align === "end"'}}]}],e.propDecorators={align:[{type:r.Input}]},e}(),c=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"mat-card-footer",host:{class:"mat-card-footer"}}]}],e}(),l=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"[mat-card-image], [matCardImage]",host:{class:"mat-card-image"}}]}],e}(),d=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"[mat-card-sm-image], [matCardImageSmall]",host:{class:"mat-card-sm-image"}}]}],e}(),p=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"[mat-card-md-image], [matCardImageMedium]",host:{class:"mat-card-md-image"}}]}],e}(),f=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"[mat-card-lg-image], [matCardImageLarge]",host:{class:"mat-card-lg-image"}}]}],e}(),h=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"[mat-card-xl-image], [matCardImageXLarge]",host:{class:"mat-card-xl-image"}}]}],e}(),m=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"[mat-card-avatar], [matCardAvatar]",host:{class:"mat-card-avatar"}}]}],e}(),g=function(){function e(){}return e.decorators=[{type:r.Component,args:[{selector:"mat-card",exportAs:"matCard",template:'<ng-content></ng-content><ng-content select="mat-card-footer"></ng-content>',styles:[".mat-card{transition:box-shadow 280ms cubic-bezier(.4,0,.2,1);display:block;position:relative;padding:16px;border-radius:4px}.mat-card .mat-divider-horizontal{position:absolute;left:0;width:100%}[dir=rtl] .mat-card .mat-divider-horizontal{left:auto;right:0}.mat-card .mat-divider-horizontal.mat-divider-inset{position:static;margin:0}[dir=rtl] .mat-card .mat-divider-horizontal.mat-divider-inset{margin-right:0}@media (-ms-high-contrast:active){.mat-card{outline:solid 1px}}.mat-card-actions,.mat-card-content,.mat-card-subtitle{display:block;margin-bottom:16px}.mat-card-title{display:block;margin-bottom:8px}.mat-card-actions{margin-left:-8px;margin-right:-8px;padding:8px 0}.mat-card-actions-align-end{display:flex;justify-content:flex-end}.mat-card-image{width:calc(100% + 32px);margin:0 -16px 16px -16px}.mat-card-footer{display:block;margin:0 -16px -16px -16px}.mat-card-actions .mat-button,.mat-card-actions .mat-raised-button{margin:0 8px}.mat-card-header{display:flex;flex-direction:row}.mat-card-header .mat-card-title{margin-bottom:12px}.mat-card-header-text{margin:0 16px}.mat-card-avatar{height:40px;width:40px;border-radius:50%;flex-shrink:0;object-fit:cover}.mat-card-title-group{display:flex;justify-content:space-between}.mat-card-sm-image{width:80px;height:80px}.mat-card-md-image{width:112px;height:112px}.mat-card-lg-image{width:152px;height:152px}.mat-card-xl-image{width:240px;height:240px;margin:-8px}.mat-card-title-group>.mat-card-xl-image{margin:-8px 0 8px 0}@media (max-width:599px){.mat-card-title-group{margin:0}.mat-card-xl-image{margin-left:0;margin-right:0}}.mat-card-content>:first-child,.mat-card>:first-child{margin-top:0}.mat-card-content>:last-child:not(.mat-card-footer),.mat-card>:last-child:not(.mat-card-footer){margin-bottom:0}.mat-card-image:first-child{margin-top:-16px;border-top-left-radius:inherit;border-top-right-radius:inherit}.mat-card>.mat-card-actions:last-child{margin-bottom:-8px;padding-bottom:0}.mat-card-actions .mat-button:first-child,.mat-card-actions .mat-raised-button:first-child{margin-left:0;margin-right:0}.mat-card-subtitle:not(:first-child),.mat-card-title:not(:first-child){margin-top:-4px}.mat-card-header .mat-card-subtitle:not(:first-child){margin-top:-8px}.mat-card>.mat-card-xl-image:first-child{margin-top:-8px}.mat-card>.mat-card-xl-image:last-child{margin-bottom:-8px}"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,host:{class:"mat-card"}}]}],e}(),v=function(){function e(){}return e.decorators=[{type:r.Component,args:[{selector:"mat-card-header",template:'<ng-content select="[mat-card-avatar], [matCardAvatar]"></ng-content><div class="mat-card-header-text"><ng-content select="mat-card-title, mat-card-subtitle, [mat-card-title], [mat-card-subtitle], [matCardTitle], [matCardSubtitle]"></ng-content></div><ng-content></ng-content>',encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,host:{class:"mat-card-header"}}]}],e}(),_=function(){function e(){}return e.decorators=[{type:r.Component,args:[{selector:"mat-card-title-group",template:'<div><ng-content select="mat-card-title, mat-card-subtitle, [mat-card-title], [mat-card-subtitle], [matCardTitle], [matCardSubtitle]"></ng-content></div><ng-content select="img"></ng-content><ng-content></ng-content>',encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,host:{class:"mat-card-title-group"}}]}],e}(),y=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[i.MatCommonModule],exports:[g,v,_,o,a,s,u,c,d,p,f,l,h,m,i.MatCommonModule],declarations:[g,v,_,o,a,s,u,c,d,p,f,l,h,m]}]}],e}()},"./node_modules/@angular/material/esm5/checkbox.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MAT_CHECKBOX_CONTROL_VALUE_ACCESSOR",function(){return m}),n.d(t,"TransitionCheckState",function(){return g}),n.d(t,"MatCheckboxChange",function(){return v}),n.d(t,"MatCheckboxBase",function(){return _}),n.d(t,"_MatCheckboxMixinBase",function(){return y}),n.d(t,"MatCheckbox",function(){return b}),n.d(t,"MAT_CHECKBOX_CLICK_ACTION",function(){return p}),n.d(t,"MatCheckboxModule",function(){return O}),n.d(t,"MAT_CHECKBOX_REQUIRED_VALIDATOR",function(){return j}),n.d(t,"MatCheckboxRequiredValidator",function(){return x});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),a=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),s=n("./node_modules/@angular/forms/fesm5/forms.js"),u=n("./node_modules/@angular/material/esm5/core.es5.js"),c=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),l=n("./node_modules/@angular/cdk/esm5/observers.es5.js"),d=n("./node_modules/@angular/common/fesm5/common.js"),p=new r.InjectionToken("mat-checkbox-click-action"),f=0,h=0,m={provide:s.NG_VALUE_ACCESSOR,useExisting:Object(r.forwardRef)(function(){return b}),multi:!0},g={Init:0,Checked:1,Unchecked:2,Indeterminate:3};g[g.Init]="Init",g[g.Checked]="Checked",g[g.Unchecked]="Unchecked",g[g.Indeterminate]="Indeterminate";var v=function(){return function(){}}(),_=function(){return function(e){this._elementRef=e}}(),y=Object(u.mixinTabIndex)(Object(u.mixinColor)(Object(u.mixinDisableRipple)(Object(u.mixinDisabled)(_)),"accent")),b=function(e){function t(t,n,i,o,a,s,u){var c=e.call(this,t)||this;return c._changeDetectorRef=n,c._focusMonitor=i,c._ngZone=o,c._clickAction=s,c._animationMode=u,c.ariaLabel="",c.ariaLabelledby=null,c._uniqueId="mat-checkbox-"+ ++f,c.id=c._uniqueId,c.labelPosition="after",c.name=null,c.change=new r.EventEmitter,c.indeterminateChange=new r.EventEmitter,c._onTouched=function(){},c._currentAnimationClass="",c._currentCheckState=g.Init,c._controlValueAccessorChangeFn=function(){},c._checked=!1,c._disabled=!1,c._indeterminate=!1,c.tabIndex=parseInt(a)||0,c._focusMonitor.monitor(t,!0).subscribe(function(e){e||Promise.resolve().then(function(){return c._onTouched()})}),c}return Object(i.__extends)(t,e),Object.defineProperty(t.prototype,"inputId",{get:function(){return(this.id||this._uniqueId)+"-input"},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"required",{get:function(){return this._required},set:function(e){this._required=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t.prototype.ngAfterViewChecked=function(){this._calculateRippleRadius()},t.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._elementRef)},Object.defineProperty(t.prototype,"checked",{get:function(){return this._checked},set:function(e){e!=this.checked&&(this._checked=e,this._changeDetectorRef.markForCheck())},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return this._disabled},set:function(e){var t=Object(a.coerceBooleanProperty)(e);t!==this.disabled&&(this._disabled=t,this._changeDetectorRef.markForCheck())},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"indeterminate",{get:function(){return this._indeterminate},set:function(e){var t=e!=this._indeterminate;this._indeterminate=e,t&&(this._indeterminate?this._transitionCheckState(g.Indeterminate):this._transitionCheckState(this.checked?g.Checked:g.Unchecked),this.indeterminateChange.emit(this._indeterminate))},enumerable:!0,configurable:!0}),t.prototype._isRippleDisabled=function(){return this.disableRipple||this.disabled},t.prototype._onLabelTextChange=function(){this._changeDetectorRef.detectChanges()},t.prototype.writeValue=function(e){this.checked=!!e},t.prototype.registerOnChange=function(e){this._controlValueAccessorChangeFn=e},t.prototype.registerOnTouched=function(e){this._onTouched=e},t.prototype.setDisabledState=function(e){this.disabled=e},t.prototype._getAriaChecked=function(){return this.checked?"true":this.indeterminate?"mixed":"false"},t.prototype._transitionCheckState=function(e){var t=this._currentCheckState,n=this._elementRef.nativeElement;if(t!==e&&(this._currentAnimationClass.length>0&&n.classList.remove(this._currentAnimationClass),this._currentAnimationClass=this._getAnimationClassForCheckStateTransition(t,e),this._currentCheckState=e,this._currentAnimationClass.length>0)){n.classList.add(this._currentAnimationClass);var r=this._currentAnimationClass;this._ngZone.runOutsideAngular(function(){setTimeout(function(){n.classList.remove(r)},1e3)})}},t.prototype._emitChangeEvent=function(){var e=new v;e.source=this,e.checked=this.checked,this._controlValueAccessorChangeFn(this.checked),this.change.emit(e)},t.prototype.toggle=function(){this.checked=!this.checked},t.prototype._onInputClick=function(e){var t=this;e.stopPropagation(),this.disabled||"noop"===this._clickAction?this.disabled||"noop"!==this._clickAction||(this._inputElement.nativeElement.checked=this.checked,this._inputElement.nativeElement.indeterminate=this.indeterminate):(this.indeterminate&&"check"!==this._clickAction&&Promise.resolve().then(function(){t._indeterminate=!1,t.indeterminateChange.emit(t._indeterminate)}),this.toggle(),this._transitionCheckState(this._checked?g.Checked:g.Unchecked),this._emitChangeEvent())},t.prototype.focus=function(){this._focusMonitor.focusVia(this._inputElement,"keyboard")},t.prototype._onInteractionEvent=function(e){e.stopPropagation()},t.prototype._getAnimationClassForCheckStateTransition=function(e,t){if("NoopAnimations"===this._animationMode)return"";var n="";switch(e){case g.Init:if(t===g.Checked)n="unchecked-checked";else{if(t!=g.Indeterminate)return"";n="unchecked-indeterminate"}break;case g.Unchecked:n=t===g.Checked?"unchecked-checked":"unchecked-indeterminate";break;case g.Checked:n=t===g.Unchecked?"checked-unchecked":"checked-indeterminate";break;case g.Indeterminate:n=t===g.Checked?"indeterminate-checked":"indeterminate-unchecked"}return"mat-checkbox-anim-"+n},t.prototype._calculateRippleRadius=function(){if(!h){var e=this._elementRef.nativeElement.querySelector(".mat-checkbox-ripple").clientWidth||0;h=e/2}this.ripple.radius=h},t.decorators=[{type:r.Component,args:[{selector:"mat-checkbox",template:'<label [attr.for]="inputId" class="mat-checkbox-layout" #label><div class="mat-checkbox-inner-container" [class.mat-checkbox-inner-container-no-side-margin]="!checkboxLabel.textContent || !checkboxLabel.textContent.trim()"><input #input class="mat-checkbox-input cdk-visually-hidden" type="checkbox" [id]="inputId" [required]="required" [checked]="checked" [attr.value]="value" [disabled]="disabled" [attr.name]="name" [tabIndex]="tabIndex" [indeterminate]="indeterminate" [attr.aria-label]="ariaLabel || null" [attr.aria-labelledby]="ariaLabelledby" [attr.aria-checked]="_getAriaChecked()" (change)="_onInteractionEvent($event)" (click)="_onInputClick($event)"><div matRipple class="mat-checkbox-ripple" [matRippleTrigger]="label" [matRippleDisabled]="_isRippleDisabled()" [matRippleRadius]="20" [matRippleCentered]="true" [matRippleAnimation]="{enterDuration: 150}"><div class="mat-ripple-element mat-checkbox-persistent-ripple"></div></div><div class="mat-checkbox-frame"></div><div class="mat-checkbox-background"><svg version="1.1" focusable="false" class="mat-checkbox-checkmark" viewBox="0 0 24 24" xml:space="preserve"><path class="mat-checkbox-checkmark-path" fill="none" stroke="white" d="M4.1,12.7 9,17.6 20.3,6.3"/></svg><div class="mat-checkbox-mixedmark"></div></div></div><span class="mat-checkbox-label" #checkboxLabel (cdkObserveContent)="_onLabelTextChange()"><span style="display:none">&nbsp;</span><ng-content></ng-content></span></label>',styles:["@keyframes mat-checkbox-fade-in-background{0%{opacity:0}50%{opacity:1}}@keyframes mat-checkbox-fade-out-background{0%,50%{opacity:1}100%{opacity:0}}@keyframes mat-checkbox-unchecked-checked-checkmark-path{0%,50%{stroke-dashoffset:22.91026}50%{animation-timing-function:cubic-bezier(0,0,.2,.1)}100%{stroke-dashoffset:0}}@keyframes mat-checkbox-unchecked-indeterminate-mixedmark{0%,68.2%{transform:scaleX(0)}68.2%{animation-timing-function:cubic-bezier(0,0,0,1)}100%{transform:scaleX(1)}}@keyframes mat-checkbox-checked-unchecked-checkmark-path{from{animation-timing-function:cubic-bezier(.4,0,1,1);stroke-dashoffset:0}to{stroke-dashoffset:-22.91026}}@keyframes mat-checkbox-checked-indeterminate-checkmark{from{animation-timing-function:cubic-bezier(0,0,.2,.1);opacity:1;transform:rotate(0)}to{opacity:0;transform:rotate(45deg)}}@keyframes mat-checkbox-indeterminate-checked-checkmark{from{animation-timing-function:cubic-bezier(.14,0,0,1);opacity:0;transform:rotate(45deg)}to{opacity:1;transform:rotate(360deg)}}@keyframes mat-checkbox-checked-indeterminate-mixedmark{from{animation-timing-function:cubic-bezier(0,0,.2,.1);opacity:0;transform:rotate(-45deg)}to{opacity:1;transform:rotate(0)}}@keyframes mat-checkbox-indeterminate-checked-mixedmark{from{animation-timing-function:cubic-bezier(.14,0,0,1);opacity:1;transform:rotate(0)}to{opacity:0;transform:rotate(315deg)}}@keyframes mat-checkbox-indeterminate-unchecked-mixedmark{0%{animation-timing-function:linear;opacity:1;transform:scaleX(1)}100%,32.8%{opacity:0;transform:scaleX(0)}}.mat-checkbox-background,.mat-checkbox-frame{top:0;left:0;right:0;bottom:0;position:absolute;border-radius:2px;box-sizing:border-box;pointer-events:none}.mat-checkbox{transition:background .4s cubic-bezier(.25,.8,.25,1),box-shadow 280ms cubic-bezier(.4,0,.2,1);cursor:pointer;-webkit-tap-highlight-color:transparent}._mat-animation-noopable.mat-checkbox{transition:none;animation:none}.mat-checkbox .mat-ripple-element:not(.mat-checkbox-persistent-ripple){opacity:.16}.mat-checkbox-layout{cursor:inherit;align-items:baseline;vertical-align:middle;display:inline-flex;white-space:nowrap}.mat-checkbox-inner-container{display:inline-block;height:16px;line-height:0;margin:auto;margin-right:8px;order:0;position:relative;vertical-align:middle;white-space:nowrap;width:16px;flex-shrink:0}[dir=rtl] .mat-checkbox-inner-container{margin-left:8px;margin-right:auto}.mat-checkbox-inner-container-no-side-margin{margin-left:0;margin-right:0}.mat-checkbox-frame{background-color:transparent;transition:border-color 90ms cubic-bezier(0,0,.2,.1);border-width:2px;border-style:solid}._mat-animation-noopable .mat-checkbox-frame{transition:none}@media (-ms-high-contrast:active){.mat-checkbox.cdk-keyboard-focused .mat-checkbox-frame{border-style:dotted}}.mat-checkbox-background{align-items:center;display:inline-flex;justify-content:center;transition:background-color 90ms cubic-bezier(0,0,.2,.1),opacity 90ms cubic-bezier(0,0,.2,.1)}._mat-animation-noopable .mat-checkbox-background{transition:none}.mat-checkbox-persistent-ripple{width:100%;height:100%;transform:none}.mat-checkbox-inner-container:hover .mat-checkbox-persistent-ripple{opacity:.04}.mat-checkbox.cdk-keyboard-focused .mat-checkbox-persistent-ripple{opacity:.12}.mat-checkbox-persistent-ripple,.mat-checkbox.mat-disabled .mat-checkbox-inner-container:hover .mat-checkbox-persistent-ripple{opacity:0}.mat-checkbox-checkmark{top:0;left:0;right:0;bottom:0;position:absolute;width:100%}.mat-checkbox-checkmark-path{stroke-dashoffset:22.91026;stroke-dasharray:22.91026;stroke-width:2.13333px}.mat-checkbox-mixedmark{width:calc(100% - 6px);height:2px;opacity:0;transform:scaleX(0) rotate(0);border-radius:2px}@media (-ms-high-contrast:active){.mat-checkbox-mixedmark{height:0;border-top:solid 2px;margin-top:2px}}.mat-checkbox-label-before .mat-checkbox-inner-container{order:1;margin-left:8px;margin-right:auto}[dir=rtl] .mat-checkbox-label-before .mat-checkbox-inner-container{margin-left:auto;margin-right:8px}.mat-checkbox-checked .mat-checkbox-checkmark{opacity:1}.mat-checkbox-checked .mat-checkbox-checkmark-path{stroke-dashoffset:0}.mat-checkbox-checked .mat-checkbox-mixedmark{transform:scaleX(1) rotate(-45deg)}.mat-checkbox-indeterminate .mat-checkbox-checkmark{opacity:0;transform:rotate(45deg)}.mat-checkbox-indeterminate .mat-checkbox-checkmark-path{stroke-dashoffset:0}.mat-checkbox-indeterminate .mat-checkbox-mixedmark{opacity:1;transform:scaleX(1) rotate(0)}.mat-checkbox-unchecked .mat-checkbox-background{background-color:transparent}.mat-checkbox-disabled{cursor:default}.mat-checkbox-anim-unchecked-checked .mat-checkbox-background{animation:180ms linear 0s mat-checkbox-fade-in-background}.mat-checkbox-anim-unchecked-checked .mat-checkbox-checkmark-path{animation:180ms linear 0s mat-checkbox-unchecked-checked-checkmark-path}.mat-checkbox-anim-unchecked-indeterminate .mat-checkbox-background{animation:180ms linear 0s mat-checkbox-fade-in-background}.mat-checkbox-anim-unchecked-indeterminate .mat-checkbox-mixedmark{animation:90ms linear 0s mat-checkbox-unchecked-indeterminate-mixedmark}.mat-checkbox-anim-checked-unchecked .mat-checkbox-background{animation:180ms linear 0s mat-checkbox-fade-out-background}.mat-checkbox-anim-checked-unchecked .mat-checkbox-checkmark-path{animation:90ms linear 0s mat-checkbox-checked-unchecked-checkmark-path}.mat-checkbox-anim-checked-indeterminate .mat-checkbox-checkmark{animation:90ms linear 0s mat-checkbox-checked-indeterminate-checkmark}.mat-checkbox-anim-checked-indeterminate .mat-checkbox-mixedmark{animation:90ms linear 0s mat-checkbox-checked-indeterminate-mixedmark}.mat-checkbox-anim-indeterminate-checked .mat-checkbox-checkmark{animation:.5s linear 0s mat-checkbox-indeterminate-checked-checkmark}.mat-checkbox-anim-indeterminate-checked .mat-checkbox-mixedmark{animation:.5s linear 0s mat-checkbox-indeterminate-checked-mixedmark}.mat-checkbox-anim-indeterminate-unchecked .mat-checkbox-background{animation:180ms linear 0s mat-checkbox-fade-out-background}.mat-checkbox-anim-indeterminate-unchecked .mat-checkbox-mixedmark{animation:.3s linear 0s mat-checkbox-indeterminate-unchecked-mixedmark}.mat-checkbox-input{bottom:0;left:50%}.mat-checkbox .mat-checkbox-ripple{position:absolute;left:calc(50% - 20px);top:calc(50% - 20px);height:40px;width:40px;z-index:1;pointer-events:none}"],exportAs:"matCheckbox",host:{class:"mat-checkbox","[id]":"id","[attr.tabindex]":"null","[class.mat-checkbox-indeterminate]":"indeterminate","[class.mat-checkbox-checked]":"checked","[class.mat-checkbox-disabled]":"disabled","[class.mat-checkbox-label-before]":'labelPosition == "before"',"[class._mat-animation-noopable]":"_animationMode === 'NoopAnimations'"},providers:[m],inputs:["disableRipple","color","tabIndex"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:r.ChangeDetectorRef},{type:o.FocusMonitor},{type:r.NgZone},{type:String,decorators:[{type:r.Attribute,args:["tabindex"]}]},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[p]}]},{type:String,decorators:[{type:r.Optional},{type:r.Inject,args:[c.ANIMATION_MODULE_TYPE]}]}]},t.propDecorators={ariaLabel:[{type:r.Input,args:["aria-label"]}],ariaLabelledby:[{type:r.Input,args:["aria-labelledby"]}],id:[{type:r.Input}],required:[{type:r.Input}],labelPosition:[{type:r.Input}],name:[{type:r.Input}],change:[{type:r.Output}],indeterminateChange:[{type:r.Output}],value:[{type:r.Input}],_inputElement:[{type:r.ViewChild,args:["input"]}],ripple:[{type:r.ViewChild,args:[u.MatRipple]}],checked:[{type:r.Input}],disabled:[{type:r.Input}],indeterminate:[{type:r.Input}]},t}(y),j={provide:s.NG_VALIDATORS,useExisting:Object(r.forwardRef)(function(){return x}),multi:!0},x=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(i.__extends)(t,e),t.decorators=[{type:r.Directive,args:[{selector:"mat-checkbox[required][formControlName],\n             mat-checkbox[required][formControl], mat-checkbox[required][ngModel]",providers:[j],host:{"[attr.required]":'required ? "" : null'}}]}],t}(s.CheckboxRequiredValidator),O=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[d.CommonModule,u.MatRippleModule,u.MatCommonModule,l.ObserversModule],exports:[b,x,u.MatCommonModule],declarations:[b,x]}]}],e}()},"./node_modules/@angular/material/esm5/chips.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatChipsModule",function(){return P}),n.d(t,"MatChipListBase",function(){return C}),n.d(t,"_MatChipListMixinBase",function(){return E}),n.d(t,"MatChipListChange",function(){return k}),n.d(t,"MatChipList",function(){return M}),n.d(t,"MatChipSelectionChange",function(){return g}),n.d(t,"MatChipBase",function(){return v}),n.d(t,"_MatChipMixinBase",function(){return _}),n.d(t,"MatChipAvatar",function(){return b}),n.d(t,"MatChipTrailingIcon",function(){return j}),n.d(t,"MatChip",function(){return x}),n.d(t,"MatChipRemove",function(){return O}),n.d(t,"MatChipInput",function(){return T}),n.d(t,"MAT_CHIPS_DEFAULT_OPTIONS",function(){return w});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),o=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),a=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),s=n("./node_modules/@angular/core/fesm5/core.js"),u=n("./node_modules/@angular/material/esm5/core.es5.js"),c=n("./node_modules/rxjs/_esm5/index.js"),l=n("./node_modules/rxjs/_esm5/operators/index.js"),d=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),p=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),f=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),h=n("./node_modules/@angular/forms/fesm5/forms.js"),m=n("./node_modules/@angular/material/esm5/form-field.es5.js"),g=function(){return function(e,t,n){void 0===n&&(n=!1),this.source=e,this.selected=t,this.isUserInput=n}}(),v=function(){return function(e){this._elementRef=e}}(),_=Object(u.mixinColor)(Object(u.mixinDisableRipple)(Object(u.mixinDisabled)(v)),"primary"),y=["mat-basic-chip"],b=function(){function e(){}return e.decorators=[{type:s.Directive,args:[{selector:"mat-chip-avatar, [matChipAvatar]",host:{class:"mat-chip-avatar"}}]}],e}(),j=function(){function e(){}return e.decorators=[{type:s.Directive,args:[{selector:"mat-chip-trailing-icon, [matChipTrailingIcon]",host:{class:"mat-chip-trailing-icon"}}]}],e}(),x=function(e){function t(t,n,r,i){var o=e.call(this,t)||this;return o._elementRef=t,o._ngZone=n,o._ripplesGloballyDisabled=!1,o.rippleConfig={},o._hasFocus=!1,o.chipListSelectable=!0,o._selected=!1,o._selectable=!0,o._removable=!0,o._onFocus=new c.Subject,o._onBlur=new c.Subject,o.selectionChange=new s.EventEmitter,o.destroyed=new s.EventEmitter,o.removed=new s.EventEmitter,o._addHostClassName(),o._chipRipple=new u.RippleRenderer(o,n,t,r),o._chipRipple.setupTriggerEvents(t.nativeElement),i&&(o._ripplesGloballyDisabled=!!i.disabled,o.rippleConfig={animation:i.animation,terminateOnPointerUp:i.terminateOnPointerUp}),o}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"rippleDisabled",{get:function(){return this.disabled||this.disableRipple||this._ripplesGloballyDisabled},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selected",{get:function(){return this._selected},set:function(e){var t=Object(i.coerceBooleanProperty)(e);t!==this._selected&&(this._selected=t,this._dispatchSelectionChange())},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{get:function(){return void 0!=this._value?this._value:this._elementRef.nativeElement.textContent},set:function(e){this._value=e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selectable",{get:function(){return this._selectable&&this.chipListSelectable},set:function(e){this._selectable=Object(i.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"removable",{get:function(){return this._removable},set:function(e){this._removable=Object(i.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ariaSelected",{get:function(){return this.selectable?this.selected.toString():null},enumerable:!0,configurable:!0}),t.prototype._addHostClassName=function(){for(var e=0,t=y;e<t.length;e++){var n=t[e];if(this._elementRef.nativeElement.hasAttribute(n)||this._elementRef.nativeElement.tagName.toLowerCase()===n)return void this._elementRef.nativeElement.classList.add(n)}this._elementRef.nativeElement.classList.add("mat-standard-chip")},t.prototype.ngOnDestroy=function(){this.destroyed.emit({chip:this}),this._chipRipple._removeTriggerEvents()},t.prototype.select=function(){this._selected||(this._selected=!0,this._dispatchSelectionChange())},t.prototype.deselect=function(){this._selected&&(this._selected=!1,this._dispatchSelectionChange())},t.prototype.selectViaInteraction=function(){this._selected||(this._selected=!0,this._dispatchSelectionChange(!0))},t.prototype.toggleSelected=function(e){return void 0===e&&(e=!1),this._selected=!this.selected,this._dispatchSelectionChange(e),this.selected},t.prototype.focus=function(){this._hasFocus||(this._elementRef.nativeElement.focus(),this._onFocus.next({chip:this})),this._hasFocus=!0},t.prototype.remove=function(){this.removable&&this.removed.emit({chip:this})},t.prototype._handleClick=function(e){this.disabled?e.preventDefault():e.stopPropagation()},t.prototype._handleKeydown=function(e){if(!this.disabled)switch(e.keyCode){case o.DELETE:case o.BACKSPACE:this.remove(),e.preventDefault();break;case o.SPACE:this.selectable&&this.toggleSelected(!0),e.preventDefault()}},t.prototype._blur=function(){var e=this;this._ngZone.onStable.asObservable().pipe(Object(l.take)(1)).subscribe(function(){e._ngZone.run(function(){e._hasFocus=!1,e._onBlur.next({chip:e})})})},t.prototype._dispatchSelectionChange=function(e){void 0===e&&(e=!1),this.selectionChange.emit({source:this,isUserInput:e,selected:this._selected})},t.decorators=[{type:s.Directive,args:[{selector:"mat-basic-chip, [mat-basic-chip], mat-chip, [mat-chip]",inputs:["color","disabled","disableRipple"],exportAs:"matChip",host:{class:"mat-chip","[attr.tabindex]":"disabled ? null : -1",role:"option","[class.mat-chip-selected]":"selected","[class.mat-chip-with-avatar]":"avatar","[class.mat-chip-with-trailing-icon]":"trailingIcon || removeIcon","[class.mat-chip-disabled]":"disabled","[attr.disabled]":"disabled || null","[attr.aria-disabled]":"disabled.toString()","[attr.aria-selected]":"ariaSelected","(click)":"_handleClick($event)","(keydown)":"_handleKeydown($event)","(focus)":"focus()","(blur)":"_blur()"}}]}],t.ctorParameters=function(){return[{type:s.ElementRef},{type:s.NgZone},{type:a.Platform},{type:void 0,decorators:[{type:s.Optional},{type:s.Inject,args:[u.MAT_RIPPLE_GLOBAL_OPTIONS]}]}]},t.propDecorators={avatar:[{type:s.ContentChild,args:[b]}],trailingIcon:[{type:s.ContentChild,args:[j]}],removeIcon:[{type:s.ContentChild,args:[Object(s.forwardRef)(function(){return O})]}],selected:[{type:s.Input}],value:[{type:s.Input}],selectable:[{type:s.Input}],removable:[{type:s.Input}],selectionChange:[{type:s.Output}],destroyed:[{type:s.Output}],removed:[{type:s.Output}]},t}(_),O=function(){function e(e){this._parentChip=e}return e.prototype._handleClick=function(e){this._parentChip.removable&&this._parentChip.remove(),e.stopPropagation()},e.decorators=[{type:s.Directive,args:[{selector:"[matChipRemove]",host:{class:"mat-chip-remove mat-chip-trailing-icon","(click)":"_handleClick($event)"}}]}],e.ctorParameters=function(){return[{type:x}]},e}(),w=new s.InjectionToken("mat-chips-default-options"),C=function(){return function(e,t,n,r){this._defaultErrorStateMatcher=e,this._parentForm=t,this._parentFormGroup=n,this.ngControl=r}}(),E=Object(u.mixinErrorState)(C),S=0,k=function(){return function(e,t){this.source=e,this.value=t}}(),M=function(e){function t(t,n,r,i,o,a,u){var l=e.call(this,a,i,o,u)||this;return l._elementRef=t,l._changeDetectorRef=n,l._dir=r,l.ngControl=u,l.controlType="mat-chip-list",l._lastDestroyedChipIndex=null,l._destroyed=new c.Subject,l._uid="mat-chip-list-"+S++,l._tabIndex=0,l._userTabIndex=null,l._onTouched=function(){},l._onChange=function(){},l._multiple=!1,l._compareWith=function(e,t){return e===t},l._required=!1,l._disabled=!1,l.ariaOrientation="horizontal",l._selectable=!0,l.change=new s.EventEmitter,l.valueChange=new s.EventEmitter,l.ngControl&&(l.ngControl.valueAccessor=l),l}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"selected",{get:function(){return this.multiple?this._selectionModel.selected:this._selectionModel.selected[0]},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"role",{get:function(){return this.empty?null:"listbox"},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"multiple",{get:function(){return this._multiple},set:function(e){this._multiple=Object(i.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"compareWith",{get:function(){return this._compareWith},set:function(e){this._compareWith=e,this._selectionModel&&this._initializeSelection()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{get:function(){return this._value},set:function(e){this.writeValue(e),this._value=e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"id",{get:function(){return this._chipInput?this._chipInput.id:this._uid},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"required",{get:function(){return this._required},set:function(e){this._required=Object(i.coerceBooleanProperty)(e),this.stateChanges.next()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"placeholder",{get:function(){return this._chipInput?this._chipInput.placeholder:this._placeholder},set:function(e){this._placeholder=e,this.stateChanges.next()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"focused",{get:function(){return this._chipInput&&this._chipInput.focused||this._hasFocusedChip()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"empty",{get:function(){return(!this._chipInput||this._chipInput.empty)&&0===this.chips.length},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"shouldLabelFloat",{get:function(){return!this.empty||this.focused},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return this.ngControl?!!this.ngControl.disabled:this._disabled},set:function(e){var t=this;this._disabled=Object(i.coerceBooleanProperty)(e),this.chips&&this.chips.forEach(function(e){return e.disabled=t._disabled})},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selectable",{get:function(){return this._selectable},set:function(e){var t=this;this._selectable=Object(i.coerceBooleanProperty)(e),this.chips&&this.chips.forEach(function(e){return e.chipListSelectable=t._selectable})},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"tabIndex",{set:function(e){this._userTabIndex=e,this._tabIndex=e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"chipSelectionChanges",{get:function(){return c.merge.apply(void 0,this.chips.map(function(e){return e.selectionChange}))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"chipFocusChanges",{get:function(){return c.merge.apply(void 0,this.chips.map(function(e){return e._onFocus}))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"chipBlurChanges",{get:function(){return c.merge.apply(void 0,this.chips.map(function(e){return e._onBlur}))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"chipRemoveChanges",{get:function(){return c.merge.apply(void 0,this.chips.map(function(e){return e.destroyed}))},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){var e=this;this._keyManager=new d.FocusKeyManager(this.chips).withWrap().withVerticalOrientation().withHorizontalOrientation(this._dir?this._dir.value:"ltr"),this._dir&&this._dir.change.pipe(Object(l.takeUntil)(this._destroyed)).subscribe(function(t){return e._keyManager.withHorizontalOrientation(t)}),this._keyManager.tabOut.pipe(Object(l.takeUntil)(this._destroyed)).subscribe(function(){e._tabIndex=-1,setTimeout(function(){return e._tabIndex=e._userTabIndex||0})}),this.chips.changes.pipe(Object(l.startWith)(null),Object(l.takeUntil)(this._destroyed)).subscribe(function(){e._resetChips(),e._initializeSelection(),e._updateTabIndex(),e._updateFocusForDestroyedChips(),e.stateChanges.next()})},t.prototype.ngOnInit=function(){this._selectionModel=new f.SelectionModel(this.multiple,void 0,!1),this.stateChanges.next()},t.prototype.ngDoCheck=function(){this.ngControl&&this.updateErrorState()},t.prototype.ngOnDestroy=function(){this._destroyed.next(),this._destroyed.complete(),this.stateChanges.complete(),this._dropSubscriptions()},t.prototype.registerInput=function(e){this._chipInput=e},t.prototype.setDescribedByIds=function(e){this._ariaDescribedby=e.join(" ")},t.prototype.writeValue=function(e){this.chips&&this._setSelectionByValue(e,!1)},t.prototype.registerOnChange=function(e){this._onChange=e},t.prototype.registerOnTouched=function(e){this._onTouched=e},t.prototype.setDisabledState=function(e){this.disabled=e,this.stateChanges.next()},t.prototype.onContainerClick=function(e){this._originatesFromChip(e)||this.focus()},t.prototype.focus=function(){this.disabled||this._chipInput&&this._chipInput.focused||(this.chips.length>0?(this._keyManager.setFirstItemActive(),this.stateChanges.next()):(this._focusInput(),this.stateChanges.next()))},t.prototype._focusInput=function(){this._chipInput&&this._chipInput.focus()},t.prototype._keydown=function(e){var t=e.target;e.keyCode===o.BACKSPACE&&this._isInputEmpty(t)?(this._keyManager.setLastItemActive(),e.preventDefault()):t&&t.classList.contains("mat-chip")&&(e.keyCode===o.HOME?(this._keyManager.setFirstItemActive(),e.preventDefault()):e.keyCode===o.END?(this._keyManager.setLastItemActive(),e.preventDefault()):this._keyManager.onKeydown(e),this.stateChanges.next())},t.prototype._updateTabIndex=function(){this._tabIndex=this._userTabIndex||(0===this.chips.length?-1:0)},t.prototype._updateFocusForDestroyedChips=function(){if(null!=this._lastDestroyedChipIndex&&this.chips.length){var e=Math.min(this._lastDestroyedChipIndex,this.chips.length-1);this._keyManager.setActiveItem(e)}this._lastDestroyedChipIndex=null},t.prototype._isValidIndex=function(e){return e>=0&&e<this.chips.length},t.prototype._isInputEmpty=function(e){return!(!e||"input"!==e.nodeName.toLowerCase())&&!e.value},t.prototype._setSelectionByValue=function(e,t){var n=this;if(void 0===t&&(t=!0),this._clearSelection(),this.chips.forEach(function(e){return e.deselect()}),Array.isArray(e))e.forEach(function(e){return n._selectValue(e,t)}),this._sortValues();else{var r=this._selectValue(e,t);r&&t&&this._keyManager.setActiveItem(r)}},t.prototype._selectValue=function(e,t){var n=this;void 0===t&&(t=!0);var r=this.chips.find(function(t){return null!=t.value&&n._compareWith(t.value,e)});return r&&(t?r.selectViaInteraction():r.select(),this._selectionModel.select(r)),r},t.prototype._initializeSelection=function(){var e=this;Promise.resolve().then(function(){(e.ngControl||e._value)&&(e._setSelectionByValue(e.ngControl?e.ngControl.value:e._value,!1),e.stateChanges.next())})},t.prototype._clearSelection=function(e){this._selectionModel.clear(),this.chips.forEach(function(t){t!==e&&t.deselect()}),this.stateChanges.next()},t.prototype._sortValues=function(){var e=this;this._multiple&&(this._selectionModel.clear(),this.chips.forEach(function(t){t.selected&&e._selectionModel.select(t)}),this.stateChanges.next())},t.prototype._propagateChanges=function(e){var t=null;t=Array.isArray(this.selected)?this.selected.map(function(e){return e.value}):this.selected?this.selected.value:e,this._value=t,this.change.emit(new k(this,t)),this.valueChange.emit(t),this._onChange(t),this._changeDetectorRef.markForCheck()},t.prototype._blur=function(){var e=this;this._hasFocusedChip()||this._keyManager.setActiveItem(-1),this.disabled||(this._chipInput?setTimeout(function(){e.focused||e._markAsTouched()}):this._markAsTouched())},t.prototype._markAsTouched=function(){this._onTouched(),this._changeDetectorRef.markForCheck(),this.stateChanges.next()},t.prototype._resetChips=function(){this._dropSubscriptions(),this._listenToChipsFocus(),this._listenToChipsSelection(),this._listenToChipsRemoved()},t.prototype._dropSubscriptions=function(){this._chipFocusSubscription&&(this._chipFocusSubscription.unsubscribe(),this._chipFocusSubscription=null),this._chipBlurSubscription&&(this._chipBlurSubscription.unsubscribe(),this._chipBlurSubscription=null),this._chipSelectionSubscription&&(this._chipSelectionSubscription.unsubscribe(),this._chipSelectionSubscription=null),this._chipRemoveSubscription&&(this._chipRemoveSubscription.unsubscribe(),this._chipRemoveSubscription=null)},t.prototype._listenToChipsSelection=function(){var e=this;this._chipSelectionSubscription=this.chipSelectionChanges.subscribe(function(t){t.source.selected?e._selectionModel.select(t.source):e._selectionModel.deselect(t.source),e.multiple||e.chips.forEach(function(t){!e._selectionModel.isSelected(t)&&t.selected&&t.deselect()}),t.isUserInput&&e._propagateChanges()})},t.prototype._listenToChipsFocus=function(){var e=this;this._chipFocusSubscription=this.chipFocusChanges.subscribe(function(t){var n=e.chips.toArray().indexOf(t.chip);e._isValidIndex(n)&&e._keyManager.updateActiveItemIndex(n),e.stateChanges.next()}),this._chipBlurSubscription=this.chipBlurChanges.subscribe(function(){e._blur(),e.stateChanges.next()})},t.prototype._listenToChipsRemoved=function(){var e=this;this._chipRemoveSubscription=this.chipRemoveChanges.subscribe(function(t){var n=t.chip,r=e.chips.toArray().indexOf(t.chip);e._isValidIndex(r)&&n._hasFocus&&(e._lastDestroyedChipIndex=r)})},t.prototype._originatesFromChip=function(e){for(var t=e.target;t&&t!==this._elementRef.nativeElement;){if(t.classList.contains("mat-chip"))return!0;t=t.parentElement}return!1},t.prototype._hasFocusedChip=function(){return this.chips.some(function(e){return e._hasFocus})},t.decorators=[{type:s.Component,args:[{selector:"mat-chip-list",template:'<div class="mat-chip-list-wrapper"><ng-content></ng-content></div>',exportAs:"matChipList",host:{"[attr.tabindex]":"disabled ? null : _tabIndex","[attr.aria-describedby]":"_ariaDescribedby || null","[attr.aria-required]":"required.toString()","[attr.aria-disabled]":"disabled.toString()","[attr.aria-invalid]":"errorState","[attr.aria-multiselectable]":"multiple","[attr.role]":"role","[class.mat-chip-list-disabled]":"disabled","[class.mat-chip-list-invalid]":"errorState","[class.mat-chip-list-required]":"required","[attr.aria-orientation]":"ariaOrientation",class:"mat-chip-list","(focus)":"focus()","(blur)":"_blur()","(keydown)":"_keydown($event)","[id]":"_uid"},providers:[{provide:m.MatFormFieldControl,useExisting:t}],styles:[".mat-chip{position:relative;overflow:hidden;box-sizing:border-box;-webkit-tap-highlight-color:transparent;transform:translateZ(0)}.mat-standard-chip{transition:box-shadow 280ms cubic-bezier(.4,0,.2,1);display:inline-flex;padding:7px 12px;border-radius:16px;align-items:center;cursor:default;min-height:32px;height:1px}.mat-standard-chip .mat-chip-remove.mat-icon{width:18px;height:18px}.mat-standard-chip::after{top:0;left:0;right:0;bottom:0;position:absolute;border-radius:inherit;opacity:0;content:'';pointer-events:none;transition:opacity .2s cubic-bezier(.35,0,.25,1)}.mat-standard-chip:hover::after{opacity:.12}.mat-standard-chip:focus{outline:0}.mat-standard-chip:focus::after{opacity:.16}@media (-ms-high-contrast:active){.mat-standard-chip{outline:solid 1px}.mat-standard-chip:focus{outline:dotted 2px}}.mat-standard-chip.mat-chip-disabled::after{opacity:0}.mat-standard-chip.mat-chip-disabled .mat-chip-remove,.mat-standard-chip.mat-chip-disabled .mat-chip-trailing-icon{cursor:default}.mat-standard-chip.mat-chip-with-avatar,.mat-standard-chip.mat-chip-with-trailing-icon.mat-chip-with-avatar{padding-top:0;padding-bottom:0}.mat-standard-chip.mat-chip-with-trailing-icon.mat-chip-with-avatar{padding-right:8px;padding-left:0}[dir=rtl] .mat-standard-chip.mat-chip-with-trailing-icon.mat-chip-with-avatar{padding-left:8px;padding-right:0}.mat-standard-chip.mat-chip-with-trailing-icon{padding-top:7px;padding-bottom:7px;padding-right:8px;padding-left:12px}[dir=rtl] .mat-standard-chip.mat-chip-with-trailing-icon{padding-left:8px;padding-right:12px}.mat-standard-chip.mat-chip-with-avatar{padding-left:0;padding-right:12px}[dir=rtl] .mat-standard-chip.mat-chip-with-avatar{padding-right:0;padding-left:12px}.mat-standard-chip .mat-chip-avatar{width:24px;height:24px;margin-right:8px;margin-left:4px}[dir=rtl] .mat-standard-chip .mat-chip-avatar{margin-left:8px;margin-right:4px}.mat-standard-chip .mat-chip-remove,.mat-standard-chip .mat-chip-trailing-icon{width:18px;height:18px;cursor:pointer}.mat-standard-chip .mat-chip-remove,.mat-standard-chip .mat-chip-trailing-icon{margin-left:8px;margin-right:0}[dir=rtl] .mat-standard-chip .mat-chip-remove,[dir=rtl] .mat-standard-chip .mat-chip-trailing-icon{margin-right:8px;margin-left:0}.mat-chip-list-wrapper{display:flex;flex-direction:row;flex-wrap:wrap;align-items:center;margin:-4px}.mat-chip-list-wrapper .mat-standard-chip,.mat-chip-list-wrapper input.mat-input-element{margin:4px}.mat-chip-list-stacked .mat-chip-list-wrapper{flex-direction:column;align-items:flex-start}.mat-chip-list-stacked .mat-chip-list-wrapper .mat-standard-chip{width:100%}.mat-chip-avatar{border-radius:50%;justify-content:center;align-items:center;display:flex;overflow:hidden;object-fit:cover}input.mat-chip-input{width:150px;margin:4px;flex:1 0 150px}"],encapsulation:s.ViewEncapsulation.None,changeDetection:s.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:s.ElementRef},{type:s.ChangeDetectorRef},{type:p.Directionality,decorators:[{type:s.Optional}]},{type:h.NgForm,decorators:[{type:s.Optional}]},{type:h.FormGroupDirective,decorators:[{type:s.Optional}]},{type:u.ErrorStateMatcher},{type:h.NgControl,decorators:[{type:s.Optional},{type:s.Self}]}]},t.propDecorators={errorStateMatcher:[{type:s.Input}],multiple:[{type:s.Input}],compareWith:[{type:s.Input}],value:[{type:s.Input}],required:[{type:s.Input}],placeholder:[{type:s.Input}],disabled:[{type:s.Input}],ariaOrientation:[{type:s.Input,args:["aria-orientation"]}],selectable:[{type:s.Input}],tabIndex:[{type:s.Input}],change:[{type:s.Output}],valueChange:[{type:s.Output}],chips:[{type:s.ContentChildren,args:[x]}]},t}(E),A=0,T=function(){function e(e,t){this._elementRef=e,this._defaultOptions=t,this.focused=!1,this._addOnBlur=!1,this.separatorKeyCodes=this._defaultOptions.separatorKeyCodes,this.chipEnd=new s.EventEmitter,this.placeholder="",this.id="mat-chip-list-input-"+A++,this._disabled=!1,this._inputElement=this._elementRef.nativeElement}return Object.defineProperty(e.prototype,"chipList",{set:function(e){e&&(this._chipList=e,this._chipList.registerInput(this))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"addOnBlur",{get:function(){return this._addOnBlur},set:function(e){this._addOnBlur=Object(i.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disabled",{get:function(){return this._disabled||this._chipList&&this._chipList.disabled},set:function(e){this._disabled=Object(i.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"empty",{get:function(){return!this._inputElement.value},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(){this._chipList.stateChanges.next()},e.prototype._keydown=function(e){this._emitChipEnd(e)},e.prototype._blur=function(){this.addOnBlur&&this._emitChipEnd(),this.focused=!1,this._chipList.focused||this._chipList._blur(),this._chipList.stateChanges.next()},e.prototype._focus=function(){this.focused=!0,this._chipList.stateChanges.next()},e.prototype._emitChipEnd=function(e){!this._inputElement.value&&e&&this._chipList._keydown(e),e&&!this._isSeparatorKey(e)||(this.chipEnd.emit({input:this._inputElement,value:this._inputElement.value}),e&&e.preventDefault())},e.prototype._onInput=function(){this._chipList.stateChanges.next()},e.prototype.focus=function(){this._inputElement.focus()},e.prototype._isSeparatorKey=function(e){if(Object(o.hasModifierKey)(e))return!1;var t=this.separatorKeyCodes,n=e.keyCode;return Array.isArray(t)?t.indexOf(n)>-1:t.has(n)},e.decorators=[{type:s.Directive,args:[{selector:"input[matChipInputFor]",exportAs:"matChipInput, matChipInputFor",host:{class:"mat-chip-input mat-input-element","(keydown)":"_keydown($event)","(blur)":"_blur()","(focus)":"_focus()","(input)":"_onInput()","[id]":"id","[attr.disabled]":"disabled || null","[attr.placeholder]":"placeholder || null","[attr.aria-invalid]":"_chipList && _chipList.ngControl ? _chipList.ngControl.invalid : null"}}]}],e.ctorParameters=function(){return[{type:s.ElementRef},{type:void 0,decorators:[{type:s.Inject,args:[w]}]}]},e.propDecorators={chipList:[{type:s.Input,args:["matChipInputFor"]}],addOnBlur:[{type:s.Input,args:["matChipInputAddOnBlur"]}],separatorKeyCodes:[{type:s.Input,args:["matChipInputSeparatorKeyCodes"]}],chipEnd:[{type:s.Output,args:["matChipInputTokenEnd"]}],placeholder:[{type:s.Input}],id:[{type:s.Input}],disabled:[{type:s.Input}]},e}(),D=[M,x,T,O,b,j],I={separatorKeyCodes:[o.ENTER]},P=function(){function e(){}return e.decorators=[{type:s.NgModule,args:[{exports:D,declarations:D,providers:[u.ErrorStateMatcher,{provide:w,useValue:I}]}]}],e}()},"./node_modules/@angular/material/esm5/core.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"AnimationCurves",function(){return m}),n.d(t,"AnimationDurations",function(){return g}),n.d(t,"MatCommonModule",function(){return y}),n.d(t,"MATERIAL_SANITY_CHECKS",function(){return v}),n.d(t,"mixinDisabled",function(){return b}),n.d(t,"mixinColor",function(){return j}),n.d(t,"mixinDisableRipple",function(){return x}),n.d(t,"mixinTabIndex",function(){return O}),n.d(t,"mixinErrorState",function(){return w}),n.d(t,"mixinInitialized",function(){return C}),n.d(t,"NativeDateModule",function(){return V}),n.d(t,"MatNativeDateModule",function(){return z}),n.d(t,"MAT_DATE_LOCALE_FACTORY",function(){return S}),n.d(t,"MAT_DATE_LOCALE",function(){return E}),n.d(t,"MAT_DATE_LOCALE_PROVIDER",function(){return k}),n.d(t,"DateAdapter",function(){return M}),n.d(t,"MAT_DATE_FORMATS",function(){return A}),n.d(t,"NativeDateAdapter",function(){return F}),n.d(t,"MAT_NATIVE_DATE_FORMATS",function(){return L}),n.d(t,"ShowOnDirtyErrorStateMatcher",function(){return U}),n.d(t,"ErrorStateMatcher",function(){return H}),n.d(t,"MAT_HAMMER_OPTIONS",function(){return q}),n.d(t,"GestureConfig",function(){return Y}),n.d(t,"setLines",function(){return Q}),n.d(t,"MatLine",function(){return K}),n.d(t,"MatLineSetter",function(){return Z}),n.d(t,"MatLineModule",function(){return $}),n.d(t,"MatOptionModule",function(){return be}),n.d(t,"_countGroupLabelsBeforeOption",function(){return _e}),n.d(t,"_getOptionScrollPosition",function(){return ye}),n.d(t,"MatOptionSelectionChange",function(){return me}),n.d(t,"MAT_OPTION_PARENT_COMPONENT",function(){return ge}),n.d(t,"MatOption",function(){return ve}),n.d(t,"MatOptgroupBase",function(){return le}),n.d(t,"_MatOptgroupMixinBase",function(){return de}),n.d(t,"MatOptgroup",function(){return fe}),n.d(t,"MAT_LABEL_GLOBAL_OPTIONS",function(){return je}),n.d(t,"MatRippleModule",function(){return se}),n.d(t,"MAT_RIPPLE_GLOBAL_OPTIONS",function(){return oe}),n.d(t,"MatRipple",function(){return ae}),n.d(t,"RippleState",function(){return J}),n.d(t,"RippleRef",function(){return ee}),n.d(t,"defaultRippleAnimationConfig",function(){return te}),n.d(t,"RippleRenderer",function(){return ie}),n.d(t,"MatPseudoCheckboxModule",function(){return ce}),n.d(t,"MatPseudoCheckbox",function(){return ue}),n.d(t,"JAN",function(){return xe}),n.d(t,"FEB",function(){return Oe}),n.d(t,"MAR",function(){return we}),n.d(t,"APR",function(){return Ce}),n.d(t,"MAY",function(){return Ee}),n.d(t,"JUN",function(){return Se}),n.d(t,"JUL",function(){return ke}),n.d(t,"AUG",function(){return Me}),n.d(t,"SEP",function(){return Ae}),n.d(t,"OCT",function(){return Te}),n.d(t,"NOV",function(){return De}),n.d(t,"DEC",function(){return Ie}),n.d(t,"ɵa1",function(){return _});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),o=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),a=n("./node_modules/tslib/tslib.es6.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/rxjs/_esm5/index.js"),c=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),l=n("./node_modules/rxjs/_esm5/operators/index.js"),d=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),p=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),f=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),h=n("./node_modules/@angular/common/fesm5/common.js"),m=function(){function e(){}return e.STANDARD_CURVE="cubic-bezier(0.4,0.0,0.2,1)",e.DECELERATION_CURVE="cubic-bezier(0.0,0.0,0.2,1)",e.ACCELERATION_CURVE="cubic-bezier(0.4,0.0,1,1)",e.SHARP_CURVE="cubic-bezier(0.4,0.0,0.6,1)",e}(),g=function(){function e(){}return e.COMPLEX="375ms",e.ENTERING="225ms",e.EXITING="195ms",e}(),v=new r.InjectionToken("mat-sanity-checks",{providedIn:"root",factory:_});function _(){return!0}var y=function(){function e(e,t){this._sanityChecksEnabled=e,this._hammerLoader=t,this._hasDoneGlobalChecks=!1,this._hasCheckedHammer=!1,this._document="object"==typeof document&&document?document:null,this._window="object"==typeof window&&window?window:null,this._areChecksEnabled()&&!this._hasDoneGlobalChecks&&(this._checkDoctypeIsDefined(),this._checkThemeIsPresent(),this._hasDoneGlobalChecks=!0)}return e.prototype._areChecksEnabled=function(){return this._sanityChecksEnabled&&Object(r.isDevMode)()&&!this._isTestEnv()},e.prototype._isTestEnv=function(){var e=this._window;return e&&(e.__karma__||e.jasmine)},e.prototype._checkDoctypeIsDefined=function(){this._document&&!this._document.doctype&&console.warn("Current document does not have a doctype. This may cause some Angular Material components not to behave as expected.")},e.prototype._checkThemeIsPresent=function(){if(this._document&&this._document.body&&"function"==typeof getComputedStyle){var e=this._document.createElement("div");e.classList.add("mat-theme-loaded-marker"),this._document.body.appendChild(e);var t=getComputedStyle(e);t&&"none"!==t.display&&console.warn("Could not find Angular Material core theme. Most Material components may not work as expected. For more info refer to the theming guide: https://material.angular.io/guide/theming"),this._document.body.removeChild(e)}},e.prototype._checkHammerIsAvailable=function(){!this._hasCheckedHammer&&this._window&&(!this._areChecksEnabled()||this._window.Hammer||this._hammerLoader||console.warn("Could not find HammerJS. Certain Angular Material components may not work correctly."),this._hasCheckedHammer=!0)},e.decorators=[{type:r.NgModule,args:[{imports:[o.BidiModule],exports:[o.BidiModule]}]}],e.ctorParameters=function(){return[{type:Boolean,decorators:[{type:r.Optional},{type:r.Inject,args:[v]}]},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[i.HAMMER_LOADER]}]}]},e}();function b(e){return function(e){function t(){for(var t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];var r=e.apply(this,t)||this;return r._disabled=!1,r}return Object(a.__extends)(t,e),Object.defineProperty(t.prototype,"disabled",{get:function(){return this._disabled},set:function(e){this._disabled=Object(s.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t}(e)}function j(e,t){return function(e){function n(){for(var n=[],r=0;r<arguments.length;r++)n[r]=arguments[r];var i=e.apply(this,n)||this;return i.color=t,i}return Object(a.__extends)(n,e),Object.defineProperty(n.prototype,"color",{get:function(){return this._color},set:function(e){var n=e||t;n!==this._color&&(this._color&&this._elementRef.nativeElement.classList.remove("mat-"+this._color),n&&this._elementRef.nativeElement.classList.add("mat-"+n),this._color=n)},enumerable:!0,configurable:!0}),n}(e)}function x(e){return function(e){function t(){for(var t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];var r=e.apply(this,t)||this;return r._disableRipple=!1,r}return Object(a.__extends)(t,e),Object.defineProperty(t.prototype,"disableRipple",{get:function(){return this._disableRipple},set:function(e){this._disableRipple=Object(s.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t}(e)}function O(e,t){return void 0===t&&(t=0),function(e){function n(){for(var n=[],r=0;r<arguments.length;r++)n[r]=arguments[r];var i=e.apply(this,n)||this;return i._tabIndex=t,i}return Object(a.__extends)(n,e),Object.defineProperty(n.prototype,"tabIndex",{get:function(){return this.disabled?-1:this._tabIndex},set:function(e){this._tabIndex=null!=e?e:t},enumerable:!0,configurable:!0}),n}(e)}function w(e){return function(e){function t(){for(var t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];var r=e.apply(this,t)||this;return r.errorState=!1,r.stateChanges=new u.Subject,r}return Object(a.__extends)(t,e),t.prototype.updateErrorState=function(){var e=this.errorState,t=this._parentFormGroup||this._parentForm,n=this.errorStateMatcher||this._defaultErrorStateMatcher,r=this.ngControl?this.ngControl.control:null,i=n.isErrorState(r,t);i!==e&&(this.errorState=i,this.stateChanges.next())},t}(e)}function C(e){return function(e){function t(){for(var t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];var r=e.apply(this,t)||this;return r._isInitialized=!1,r._pendingSubscribers=[],r.initialized=new u.Observable(function(e){r._isInitialized?r._notifySubscriber(e):r._pendingSubscribers.push(e)}),r}return Object(a.__extends)(t,e),t.prototype._markInitialized=function(){if(this._isInitialized)throw Error("This directive has already been marked as initialized and should not be called twice.");this._isInitialized=!0,this._pendingSubscribers.forEach(this._notifySubscriber),this._pendingSubscribers=null},t.prototype._notifySubscriber=function(e){e.next(),e.complete()},t}(e)}var E=new r.InjectionToken("MAT_DATE_LOCALE",{providedIn:"root",factory:S});function S(){return Object(r.inject)(r.LOCALE_ID)}var k={provide:E,useExisting:r.LOCALE_ID},M=function(){function e(){this._localeChanges=new u.Subject}return Object.defineProperty(e.prototype,"localeChanges",{get:function(){return this._localeChanges},enumerable:!0,configurable:!0}),e.prototype.deserialize=function(e){return null==e||this.isDateInstance(e)&&this.isValid(e)?e:this.invalid()},e.prototype.setLocale=function(e){this.locale=e,this._localeChanges.next()},e.prototype.compareDate=function(e,t){return this.getYear(e)-this.getYear(t)||this.getMonth(e)-this.getMonth(t)||this.getDate(e)-this.getDate(t)},e.prototype.sameDate=function(e,t){if(e&&t){var n=this.isValid(e),r=this.isValid(t);return n&&r?!this.compareDate(e,t):n==r}return e==t},e.prototype.clampDate=function(e,t,n){return t&&this.compareDate(e,t)<0?t:n&&this.compareDate(e,n)>0?n:e},e}(),A=new r.InjectionToken("mat-date-formats"),T="undefined"!=typeof Intl,D={long:["January","February","March","April","May","June","July","August","September","October","November","December"],short:["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"],narrow:["J","F","M","A","M","J","J","A","S","O","N","D"]},I=N(31,function(e){return String(e+1)}),P={long:["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"],short:["Sun","Mon","Tue","Wed","Thu","Fri","Sat"],narrow:["S","M","T","W","T","F","S"]},R=/^\d{4}-\d{2}-\d{2}(?:T\d{2}:\d{2}:\d{2}(?:\.\d+)?(?:Z|(?:(?:\+|-)\d{2}:\d{2}))?)?$/;function N(e,t){for(var n=Array(e),r=0;r<e;r++)n[r]=t(r);return n}var F=function(e){function t(t,n){var r=e.call(this)||this;return r.useUtcForDisplay=!0,e.prototype.setLocale.call(r,t),r.useUtcForDisplay=!n.TRIDENT,r._clampDate=n.TRIDENT||n.EDGE,r}return Object(a.__extends)(t,e),t.prototype.getYear=function(e){return e.getFullYear()},t.prototype.getMonth=function(e){return e.getMonth()},t.prototype.getDate=function(e){return e.getDate()},t.prototype.getDayOfWeek=function(e){return e.getDay()},t.prototype.getMonthNames=function(e){var t=this;if(T){var n=new Intl.DateTimeFormat(this.locale,{month:e,timeZone:"utc"});return N(12,function(e){return t._stripDirectionalityCharacters(t._format(n,new Date(2017,e,1)))})}return D[e]},t.prototype.getDateNames=function(){var e=this;if(T){var t=new Intl.DateTimeFormat(this.locale,{day:"numeric",timeZone:"utc"});return N(31,function(n){return e._stripDirectionalityCharacters(e._format(t,new Date(2017,0,n+1)))})}return I},t.prototype.getDayOfWeekNames=function(e){var t=this;if(T){var n=new Intl.DateTimeFormat(this.locale,{weekday:e,timeZone:"utc"});return N(7,function(e){return t._stripDirectionalityCharacters(t._format(n,new Date(2017,0,e+1)))})}return P[e]},t.prototype.getYearName=function(e){if(T){var t=new Intl.DateTimeFormat(this.locale,{year:"numeric",timeZone:"utc"});return this._stripDirectionalityCharacters(this._format(t,e))}return String(this.getYear(e))},t.prototype.getFirstDayOfWeek=function(){return 0},t.prototype.getNumDaysInMonth=function(e){return this.getDate(this._createDateWithOverflow(this.getYear(e),this.getMonth(e)+1,0))},t.prototype.clone=function(e){return new Date(e.getTime())},t.prototype.createDate=function(e,t,n){if(t<0||t>11)throw Error('Invalid month index "'+t+'". Month index has to be between 0 and 11.');if(n<1)throw Error('Invalid date "'+n+'". Date has to be greater than 0.');var r=this._createDateWithOverflow(e,t,n);if(r.getMonth()!=t)throw Error('Invalid date "'+n+'" for month with index "'+t+'".');return r},t.prototype.today=function(){return new Date},t.prototype.parse=function(e){return"number"==typeof e?new Date(e):e?new Date(Date.parse(e)):null},t.prototype.format=function(e,t){if(!this.isValid(e))throw Error("NativeDateAdapter: Cannot format invalid date.");if(T){this._clampDate&&(e.getFullYear()<1||e.getFullYear()>9999)&&(e=this.clone(e)).setFullYear(Math.max(1,Math.min(9999,e.getFullYear()))),t=Object(a.__assign)({},t,{timeZone:"utc"});var n=new Intl.DateTimeFormat(this.locale,t);return this._stripDirectionalityCharacters(this._format(n,e))}return this._stripDirectionalityCharacters(e.toDateString())},t.prototype.addCalendarYears=function(e,t){return this.addCalendarMonths(e,12*t)},t.prototype.addCalendarMonths=function(e,t){var n=this._createDateWithOverflow(this.getYear(e),this.getMonth(e)+t,this.getDate(e));return this.getMonth(n)!=((this.getMonth(e)+t)%12+12)%12&&(n=this._createDateWithOverflow(this.getYear(n),this.getMonth(n),0)),n},t.prototype.addCalendarDays=function(e,t){return this._createDateWithOverflow(this.getYear(e),this.getMonth(e),this.getDate(e)+t)},t.prototype.toIso8601=function(e){return[e.getUTCFullYear(),this._2digit(e.getUTCMonth()+1),this._2digit(e.getUTCDate())].join("-")},t.prototype.deserialize=function(t){if("string"==typeof t){if(!t)return null;if(R.test(t)){var n=new Date(t);if(this.isValid(n))return n}}return e.prototype.deserialize.call(this,t)},t.prototype.isDateInstance=function(e){return e instanceof Date},t.prototype.isValid=function(e){return!isNaN(e.getTime())},t.prototype.invalid=function(){return new Date(NaN)},t.prototype._createDateWithOverflow=function(e,t,n){var r=new Date(e,t,n);return e>=0&&e<100&&r.setFullYear(this.getYear(r)-1900),r},t.prototype._2digit=function(e){return("00"+e).slice(-2)},t.prototype._stripDirectionalityCharacters=function(e){return e.replace(/[\u200e\u200f]/g,"")},t.prototype._format=function(e,t){var n=new Date(Date.UTC(t.getFullYear(),t.getMonth(),t.getDate(),t.getHours(),t.getMinutes(),t.getSeconds(),t.getMilliseconds()));return e.format(n)},t.decorators=[{type:r.Injectable}],t.ctorParameters=function(){return[{type:String,decorators:[{type:r.Optional},{type:r.Inject,args:[E]}]},{type:c.Platform}]},t}(M),L={parse:{dateInput:null},display:{dateInput:{year:"numeric",month:"numeric",day:"numeric"},monthYearLabel:{year:"numeric",month:"short"},dateA11yLabel:{year:"numeric",month:"long",day:"numeric"},monthYearA11yLabel:{year:"numeric",month:"long"}}},V=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[c.PlatformModule],providers:[{provide:M,useClass:F}]}]}],e}(),B=L,z=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[V],providers:[{provide:A,useValue:B}]}]}],e}(),U=function(){function e(){}return e.prototype.isErrorState=function(e,t){return!!(e&&e.invalid&&(e.dirty||t&&t.submitted))},e.decorators=[{type:r.Injectable}],e}(),H=function(){function e(){}return e.prototype.isErrorState=function(e,t){return!!(e&&e.invalid&&(e.touched||t&&t.submitted))},e.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(),q=new r.InjectionToken("MAT_HAMMER_OPTIONS"),G=["longpress","slide","slidestart","slideend","slideright","slideleft"],W={on:function(){},off:function(){}},Y=function(e){function t(t,n){var r=e.call(this)||this;return r._hammerOptions=t,r.events=G,n&&n._checkHammerIsAvailable(),r}return Object(a.__extends)(t,e),t.prototype.buildHammer=function(e){var t="undefined"!=typeof window?window.Hammer:null;if(!t)return W;var n=new t(e,this._hammerOptions||void 0),r=new t.Pan,i=new t.Swipe,o=new t.Press,a=this._createRecognizer(r,{event:"slide",threshold:0},i),s=this._createRecognizer(o,{event:"longpress",time:500});return r.recognizeWith(i),s.recognizeWith(a),n.add([i,o,r,a,s]),n},t.prototype._createRecognizer=function(e,t){for(var n=[],r=2;r<arguments.length;r++)n[r-2]=arguments[r];var i=new e.constructor(t);return n.push(e),n.forEach(function(e){return i.recognizeWith(e)}),i},t.decorators=[{type:r.Injectable}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[q]}]},{type:y,decorators:[{type:r.Optional}]}]},t}(i.HammerGestureConfig),K=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"[mat-line], [matLine]",host:{class:"mat-line"}}]}],e}();function Q(e,t){e.changes.pipe(Object(l.startWith)(e)).subscribe(function(e){var n=e.length;X(t,"mat-2-line",!1),X(t,"mat-3-line",!1),X(t,"mat-multi-line",!1),2===n||3===n?X(t,"mat-"+n+"-line",!0):n>3&&X(t,"mat-multi-line",!0)})}function X(e,t,n){var r=e.nativeElement.classList;n?r.add(t):r.remove(t)}var Z=function(){return function(e,t){Q(e,t)}}(),$=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[y],exports:[K,y],declarations:[K]}]}],e}(),J={FADING_IN:0,VISIBLE:1,FADING_OUT:2,HIDDEN:3};J[J.FADING_IN]="FADING_IN",J[J.VISIBLE]="VISIBLE",J[J.FADING_OUT]="FADING_OUT",J[J.HIDDEN]="HIDDEN";var ee=function(){function e(e,t,n){this._renderer=e,this.element=t,this.config=n,this.state=J.HIDDEN}return e.prototype.fadeOut=function(){this._renderer.fadeOutRipple(this)},e}(),te={enterDuration:450,exitDuration:400},ne=800,re=Object(c.normalizePassiveListenerOptions)({passive:!0}),ie=function(){function e(e,t,n,r){var i=this;this._target=e,this._ngZone=t,this._isPointerDown=!1,this._triggerEvents=new Map,this._activeRipples=new Set,this.onMousedown=function(e){var t=Object(d.isFakeMousedownFromScreenReader)(e),n=i._lastTouchStartEvent&&Date.now()<i._lastTouchStartEvent+ne;i._target.rippleDisabled||t||n||(i._isPointerDown=!0,i.fadeInRipple(e.clientX,e.clientY,i._target.rippleConfig))},this.onTouchStart=function(e){if(!i._target.rippleDisabled){i._lastTouchStartEvent=Date.now(),i._isPointerDown=!0;for(var t=e.changedTouches,n=0;n<t.length;n++)i.fadeInRipple(t[n].clientX,t[n].clientY,i._target.rippleConfig)}},this.onPointerUp=function(){i._isPointerDown&&(i._isPointerDown=!1,i._activeRipples.forEach(function(e){var t=e.state===J.VISIBLE||e.config.terminateOnPointerUp&&e.state===J.FADING_IN;!e.config.persistent&&t&&e.fadeOut()}))},r.isBrowser&&(this._containerElement=n.nativeElement,this._triggerEvents.set("mousedown",this.onMousedown).set("mouseup",this.onPointerUp).set("mouseleave",this.onPointerUp).set("touchstart",this.onTouchStart).set("touchend",this.onPointerUp).set("touchcancel",this.onPointerUp))}return e.prototype.fadeInRipple=function(e,t,n){var r=this;void 0===n&&(n={});var i=this._containerRect=this._containerRect||this._containerElement.getBoundingClientRect(),o=Object(a.__assign)({},te,n.animation);n.centered&&(e=i.left+i.width/2,t=i.top+i.height/2);var s,u,c,l,d,p,f=n.radius||(s=e,u=t,c=i,l=Math.max(Math.abs(s-c.left),Math.abs(s-c.right)),d=Math.max(Math.abs(u-c.top),Math.abs(u-c.bottom)),Math.sqrt(l*l+d*d)),h=e-i.left,m=t-i.top,g=o.enterDuration,v=document.createElement("div");v.classList.add("mat-ripple-element"),v.style.left=h-f+"px",v.style.top=m-f+"px",v.style.height=2*f+"px",v.style.width=2*f+"px",v.style.backgroundColor=n.color||null,v.style.transitionDuration=g+"ms",this._containerElement.appendChild(v),p=v,window.getComputedStyle(p).getPropertyValue("opacity"),v.style.transform="scale(1)";var _=new ee(this,v,n);return _.state=J.FADING_IN,this._activeRipples.add(_),n.persistent||(this._mostRecentTransientRipple=_),this.runTimeoutOutsideZone(function(){var e=_===r._mostRecentTransientRipple;_.state=J.VISIBLE,n.persistent||e&&r._isPointerDown||_.fadeOut()},g),_},e.prototype.fadeOutRipple=function(e){var t=this._activeRipples.delete(e);if(e===this._mostRecentTransientRipple&&(this._mostRecentTransientRipple=null),this._activeRipples.size||(this._containerRect=null),t){var n=e.element,r=Object(a.__assign)({},te,e.config.animation);n.style.transitionDuration=r.exitDuration+"ms",n.style.opacity="0",e.state=J.FADING_OUT,this.runTimeoutOutsideZone(function(){e.state=J.HIDDEN,n.parentNode.removeChild(n)},r.exitDuration)}},e.prototype.fadeOutAll=function(){this._activeRipples.forEach(function(e){return e.fadeOut()})},e.prototype.setupTriggerEvents=function(e){var t=this;e&&e!==this._triggerElement&&(this._removeTriggerEvents(),this._ngZone.runOutsideAngular(function(){t._triggerEvents.forEach(function(t,n){e.addEventListener(n,t,re)})}),this._triggerElement=e)},e.prototype.runTimeoutOutsideZone=function(e,t){void 0===t&&(t=0),this._ngZone.runOutsideAngular(function(){return setTimeout(e,t)})},e.prototype._removeTriggerEvents=function(){var e=this;this._triggerElement&&this._triggerEvents.forEach(function(t,n){e._triggerElement.removeEventListener(n,t,re)})},e}();var oe=new r.InjectionToken("mat-ripple-global-options"),ae=function(){function e(e,t,n,r,i){this._elementRef=e,this.radius=0,this._disabled=!1,this._isInitialized=!1,this._globalOptions=r||{},this._rippleRenderer=new ie(this,t,e,n),"NoopAnimations"===i&&(this._globalOptions.animation={enterDuration:0,exitDuration:0})}return Object.defineProperty(e.prototype,"disabled",{get:function(){return this._disabled},set:function(e){this._disabled=e,this._setupTriggerEventsIfEnabled()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"trigger",{get:function(){return this._trigger||this._elementRef.nativeElement},set:function(e){this._trigger=e,this._setupTriggerEventsIfEnabled()},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){this._isInitialized=!0,this._setupTriggerEventsIfEnabled()},e.prototype.ngOnDestroy=function(){this._rippleRenderer._removeTriggerEvents()},e.prototype.fadeOutAll=function(){this._rippleRenderer.fadeOutAll()},Object.defineProperty(e.prototype,"rippleConfig",{get:function(){return{centered:this.centered,radius:this.radius,color:this.color,animation:Object(a.__assign)({},this._globalOptions.animation,this.animation),terminateOnPointerUp:this._globalOptions.terminateOnPointerUp}},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"rippleDisabled",{get:function(){return this.disabled||!!this._globalOptions.disabled},enumerable:!0,configurable:!0}),e.prototype._setupTriggerEventsIfEnabled=function(){!this.disabled&&this._isInitialized&&this._rippleRenderer.setupTriggerEvents(this.trigger)},e.prototype.launch=function(e,t,n){return void 0===t&&(t=0),"number"==typeof e?this._rippleRenderer.fadeInRipple(e,t,Object(a.__assign)({},this.rippleConfig,n)):this._rippleRenderer.fadeInRipple(0,0,Object(a.__assign)({},this.rippleConfig,e))},e.decorators=[{type:r.Directive,args:[{selector:"[mat-ripple], [matRipple]",exportAs:"matRipple",host:{class:"mat-ripple","[class.mat-ripple-unbounded]":"unbounded"}}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:r.NgZone},{type:c.Platform},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[oe]}]},{type:String,decorators:[{type:r.Optional},{type:r.Inject,args:[p.ANIMATION_MODULE_TYPE]}]}]},e.propDecorators={color:[{type:r.Input,args:["matRippleColor"]}],unbounded:[{type:r.Input,args:["matRippleUnbounded"]}],centered:[{type:r.Input,args:["matRippleCentered"]}],radius:[{type:r.Input,args:["matRippleRadius"]}],animation:[{type:r.Input,args:["matRippleAnimation"]}],disabled:[{type:r.Input,args:["matRippleDisabled"]}],trigger:[{type:r.Input,args:["matRippleTrigger"]}]},e}(),se=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[y,c.PlatformModule],exports:[ae,y],declarations:[ae]}]}],e}(),ue=function(){function e(e){this._animationMode=e,this.state="unchecked",this.disabled=!1}return e.decorators=[{type:r.Component,args:[{encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,selector:"mat-pseudo-checkbox",styles:[".mat-pseudo-checkbox{width:16px;height:16px;border:2px solid;border-radius:2px;cursor:pointer;display:inline-block;vertical-align:middle;box-sizing:border-box;position:relative;flex-shrink:0;transition:border-color 90ms cubic-bezier(0,0,.2,.1),background-color 90ms cubic-bezier(0,0,.2,.1)}.mat-pseudo-checkbox::after{position:absolute;opacity:0;content:'';border-bottom:2px solid currentColor;transition:opacity 90ms cubic-bezier(0,0,.2,.1)}.mat-pseudo-checkbox.mat-pseudo-checkbox-checked,.mat-pseudo-checkbox.mat-pseudo-checkbox-indeterminate{border-color:transparent}._mat-animation-noopable.mat-pseudo-checkbox{transition:none;animation:none}._mat-animation-noopable.mat-pseudo-checkbox::after{transition:none}.mat-pseudo-checkbox-disabled{cursor:default}.mat-pseudo-checkbox-indeterminate::after{top:5px;left:1px;width:10px;opacity:1;border-radius:2px}.mat-pseudo-checkbox-checked::after{top:2.4px;left:1px;width:8px;height:3px;border-left:2px solid currentColor;transform:rotate(-45deg);opacity:1}"],template:"",host:{class:"mat-pseudo-checkbox","[class.mat-pseudo-checkbox-indeterminate]":'state === "indeterminate"',"[class.mat-pseudo-checkbox-checked]":'state === "checked"',"[class.mat-pseudo-checkbox-disabled]":"disabled","[class._mat-animation-noopable]":'_animationMode === "NoopAnimations"'}}]}],e.ctorParameters=function(){return[{type:String,decorators:[{type:r.Optional},{type:r.Inject,args:[p.ANIMATION_MODULE_TYPE]}]}]},e.propDecorators={state:[{type:r.Input}],disabled:[{type:r.Input}]},e}(),ce=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{exports:[ue],declarations:[ue]}]}],e}(),le=function(){return function(){}}(),de=b(le),pe=0,fe=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t._labelId="mat-optgroup-label-"+pe++,t}return Object(a.__extends)(t,e),t.decorators=[{type:r.Component,args:[{selector:"mat-optgroup",exportAs:"matOptgroup",template:'<label class="mat-optgroup-label" [id]="_labelId">{{ label }}<ng-content></ng-content></label><ng-content select="mat-option, ng-container"></ng-content>',encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,inputs:["disabled"],styles:[".mat-optgroup-label{white-space:nowrap;overflow:hidden;text-overflow:ellipsis;display:block;line-height:48px;height:48px;padding:0 16px;text-align:left;text-decoration:none;max-width:100%;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:default}.mat-optgroup-label[disabled]{cursor:default}[dir=rtl] .mat-optgroup-label{text-align:right}.mat-optgroup-label .mat-icon{margin-right:16px;vertical-align:middle}.mat-optgroup-label .mat-icon svg{vertical-align:top}[dir=rtl] .mat-optgroup-label .mat-icon{margin-left:16px;margin-right:0}"],host:{class:"mat-optgroup",role:"group","[class.mat-optgroup-disabled]":"disabled","[attr.aria-disabled]":"disabled.toString()","[attr.aria-labelledby]":"_labelId"}}]}],t.propDecorators={label:[{type:r.Input}]},t}(de),he=0,me=function(){return function(e,t){void 0===t&&(t=!1),this.source=e,this.isUserInput=t}}(),ge=new r.InjectionToken("MAT_OPTION_PARENT_COMPONENT"),ve=function(){function e(e,t,n,i){this._element=e,this._changeDetectorRef=t,this._parent=n,this.group=i,this._selected=!1,this._active=!1,this._disabled=!1,this._mostRecentViewValue="",this.id="mat-option-"+he++,this.onSelectionChange=new r.EventEmitter,this._stateChanges=new u.Subject}return Object.defineProperty(e.prototype,"multiple",{get:function(){return this._parent&&this._parent.multiple},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selected",{get:function(){return this._selected},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disabled",{get:function(){return this.group&&this.group.disabled||this._disabled},set:function(e){this._disabled=Object(s.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disableRipple",{get:function(){return this._parent&&this._parent.disableRipple},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"active",{get:function(){return this._active},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"viewValue",{get:function(){return(this._getHostElement().textContent||"").trim()},enumerable:!0,configurable:!0}),e.prototype.select=function(){this._selected||(this._selected=!0,this._changeDetectorRef.markForCheck(),this._emitSelectionChangeEvent())},e.prototype.deselect=function(){this._selected&&(this._selected=!1,this._changeDetectorRef.markForCheck(),this._emitSelectionChangeEvent())},e.prototype.focus=function(){var e=this._getHostElement();"function"==typeof e.focus&&e.focus()},e.prototype.setActiveStyles=function(){this._active||(this._active=!0,this._changeDetectorRef.markForCheck())},e.prototype.setInactiveStyles=function(){this._active&&(this._active=!1,this._changeDetectorRef.markForCheck())},e.prototype.getLabel=function(){return this.viewValue},e.prototype._handleKeydown=function(e){e.keyCode!==f.ENTER&&e.keyCode!==f.SPACE||Object(f.hasModifierKey)(e)||(this._selectViaInteraction(),e.preventDefault())},e.prototype._selectViaInteraction=function(){this.disabled||(this._selected=!this.multiple||!this._selected,this._changeDetectorRef.markForCheck(),this._emitSelectionChangeEvent(!0))},e.prototype._getTabIndex=function(){return this.disabled?"-1":"0"},e.prototype._getHostElement=function(){return this._element.nativeElement},e.prototype.ngAfterViewChecked=function(){if(this._selected){var e=this.viewValue;e!==this._mostRecentViewValue&&(this._mostRecentViewValue=e,this._stateChanges.next())}},e.prototype.ngOnDestroy=function(){this._stateChanges.complete()},e.prototype._emitSelectionChangeEvent=function(e){void 0===e&&(e=!1),this.onSelectionChange.emit(new me(this,e))},e.decorators=[{type:r.Component,args:[{selector:"mat-option",exportAs:"matOption",host:{role:"option","[attr.tabindex]":"_getTabIndex()","[class.mat-selected]":"selected","[class.mat-option-multiple]":"multiple","[class.mat-active]":"active","[id]":"id","[attr.aria-selected]":"selected.toString()","[attr.aria-disabled]":"disabled.toString()","[class.mat-option-disabled]":"disabled","(click)":"_selectViaInteraction()","(keydown)":"_handleKeydown($event)",class:"mat-option"},styles:[".mat-option{white-space:nowrap;overflow:hidden;text-overflow:ellipsis;display:block;line-height:48px;height:48px;padding:0 16px;text-align:left;text-decoration:none;max-width:100%;position:relative;cursor:pointer;outline:0;display:flex;flex-direction:row;max-width:100%;box-sizing:border-box;align-items:center;-webkit-tap-highlight-color:transparent}.mat-option[disabled]{cursor:default}[dir=rtl] .mat-option{text-align:right}.mat-option .mat-icon{margin-right:16px;vertical-align:middle}.mat-option .mat-icon svg{vertical-align:top}[dir=rtl] .mat-option .mat-icon{margin-left:16px;margin-right:0}.mat-option[aria-disabled=true]{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:default}.mat-optgroup .mat-option:not(.mat-option-multiple){padding-left:32px}[dir=rtl] .mat-optgroup .mat-option:not(.mat-option-multiple){padding-left:16px;padding-right:32px}@media (-ms-high-contrast:active){.mat-option{margin:0 1px}.mat-option.mat-active{border:solid 1px currentColor;margin:0}}.mat-option-text{display:inline-block;flex-grow:1;overflow:hidden;text-overflow:ellipsis}.mat-option .mat-option-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}@media (-ms-high-contrast:active){.mat-option .mat-option-ripple{opacity:.5}}.mat-option-pseudo-checkbox{margin-right:8px}[dir=rtl] .mat-option-pseudo-checkbox{margin-left:8px;margin-right:0}"],template:'<mat-pseudo-checkbox *ngIf="multiple" class="mat-option-pseudo-checkbox" [state]="selected ? \'checked\' : \'\'" [disabled]="disabled"></mat-pseudo-checkbox><span class="mat-option-text"><ng-content></ng-content></span><div class="mat-option-ripple" mat-ripple [matRippleTrigger]="_getHostElement()" [matRippleDisabled]="disabled || disableRipple"></div>',encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:r.ChangeDetectorRef},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[ge]}]},{type:fe,decorators:[{type:r.Optional}]}]},e.propDecorators={value:[{type:r.Input}],id:[{type:r.Input}],disabled:[{type:r.Input}],onSelectionChange:[{type:r.Output}]},e}();function _e(e,t,n){if(n.length){for(var r=t.toArray(),i=n.toArray(),o=0,a=0;a<e+1;a++)r[a].group&&r[a].group===i[o]&&o++;return o}return 0}function ye(e,t,n,r){var i=e*t;return i<n?i:i+t>n+r?Math.max(0,i-r+t):n}var be=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[se,h.CommonModule,ce],exports:[ve,fe],declarations:[ve,fe]}]}],e}(),je=new r.InjectionToken("mat-label-global-options"),xe=0,Oe=1,we=2,Ce=3,Ee=4,Se=5,ke=6,Me=7,Ae=8,Te=9,De=10,Ie=11},"./node_modules/@angular/material/esm5/datepicker.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatDatepickerModule",function(){return G}),n.d(t,"MatCalendarHeader",function(){return k}),n.d(t,"MatCalendar",function(){return M}),n.d(t,"MatCalendarCell",function(){return O}),n.d(t,"MatCalendarBody",function(){return w}),n.d(t,"MAT_DATEPICKER_SCROLL_STRATEGY_FACTORY",function(){return I}),n.d(t,"MAT_DATEPICKER_SCROLL_STRATEGY",function(){return D}),n.d(t,"MAT_DATEPICKER_SCROLL_STRATEGY_FACTORY_PROVIDER",function(){return P}),n.d(t,"MatDatepickerContentBase",function(){return R}),n.d(t,"_MatDatepickerContentMixinBase",function(){return N}),n.d(t,"MatDatepickerContent",function(){return F}),n.d(t,"MatDatepicker",function(){return L}),n.d(t,"matDatepickerAnimations",function(){return A}),n.d(t,"MAT_DATEPICKER_VALUE_ACCESSOR",function(){return V}),n.d(t,"MAT_DATEPICKER_VALIDATORS",function(){return B}),n.d(t,"MatDatepickerInputEvent",function(){return z}),n.d(t,"MatDatepickerInput",function(){return U}),n.d(t,"MatDatepickerIntl",function(){return x}),n.d(t,"MatDatepickerToggleIcon",function(){return H}),n.d(t,"MatDatepickerToggle",function(){return q}),n.d(t,"MatMonthView",function(){return C}),n.d(t,"MatYearView",function(){return S}),n.d(t,"ɵa34",function(){return E});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/rxjs/_esm5/index.js"),o=n("./node_modules/rxjs/_esm5/operators/index.js"),a=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),s=n("./node_modules/@angular/material/esm5/core.es5.js"),u=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),c=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),l=n("./node_modules/@angular/animations/fesm5/animations.js"),d=n("./node_modules/tslib/tslib.es6.js"),p=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),f=n("./node_modules/@angular/cdk/esm5/overlay.es5.js"),h=n("./node_modules/@angular/common/fesm5/common.js"),m=n("./node_modules/@angular/material/esm5/dialog.es5.js"),g=n("./node_modules/@angular/forms/fesm5/forms.js"),v=n("./node_modules/@angular/material/esm5/form-field.es5.js"),_=n("./node_modules/@angular/material/esm5/input.es5.js"),y=n("./node_modules/@angular/material/esm5/button.es5.js"),b=n("./node_modules/@angular/cdk/esm5/a11y.es5.js");function j(e){return Error("MatDatepicker: No provider found for "+e+". You must import one of the following modules at your application root: MatNativeDateModule, MatMomentDateModule, or provide a custom implementation.")}var x=function(){function e(){this.changes=new i.Subject,this.calendarLabel="Calendar",this.openCalendarLabel="Open calendar",this.prevMonthLabel="Previous month",this.nextMonthLabel="Next month",this.prevYearLabel="Previous year",this.nextYearLabel="Next year",this.prevMultiYearLabel="Previous 20 years",this.nextMultiYearLabel="Next 20 years",this.switchToMonthViewLabel="Choose date",this.switchToMultiYearViewLabel="Choose month and year"}return e.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}(),O=function(){return function(e,t,n,r,i){this.value=e,this.displayValue=t,this.ariaLabel=n,this.enabled=r,this.cssClasses=i}}(),w=function(){function e(e,t){this._elementRef=e,this._ngZone=t,this.numCols=7,this.activeCell=0,this.cellAspectRatio=1,this.selectedValueChange=new r.EventEmitter}return e.prototype._cellClicked=function(e){e.enabled&&this.selectedValueChange.emit(e.value)},e.prototype.ngOnChanges=function(e){var t=e.numCols,n=this.rows,r=this.numCols;(e.rows||t)&&(this._firstRowOffset=n&&n.length&&n[0].length?r-n[0].length:0),(e.cellAspectRatio||t||!this._cellPadding)&&(this._cellPadding=50*this.cellAspectRatio/r+"%"),!t&&this._cellWidth||(this._cellWidth=100/r+"%")},e.prototype._isActiveCell=function(e,t){var n=e*this.numCols+t;return e&&(n-=this._firstRowOffset),n==this.activeCell},e.prototype._focusActiveCell=function(){var e=this;this._ngZone.runOutsideAngular(function(){e._ngZone.onStable.asObservable().pipe(Object(o.take)(1)).subscribe(function(){var t=e._elementRef.nativeElement.querySelector(".mat-calendar-body-active");t&&t.focus()})})},e.decorators=[{type:r.Component,args:[{selector:"[mat-calendar-body]",template:'<tr *ngIf="_firstRowOffset < labelMinRequiredCells" aria-hidden="true"><td class="mat-calendar-body-label" [attr.colspan]="numCols" [style.paddingTop]="_cellPadding" [style.paddingBottom]="_cellPadding">{{label}}</td></tr><tr *ngFor="let row of rows; let rowIndex = index" role="row"><td *ngIf="rowIndex === 0 && _firstRowOffset" aria-hidden="true" class="mat-calendar-body-label" [attr.colspan]="_firstRowOffset" [style.paddingTop]="_cellPadding" [style.paddingBottom]="_cellPadding">{{_firstRowOffset >= labelMinRequiredCells ? label : \'\'}}</td><td *ngFor="let item of row; let colIndex = index" role="gridcell" class="mat-calendar-body-cell" [ngClass]="item.cssClasses" [tabindex]="_isActiveCell(rowIndex, colIndex) ? 0 : -1" [class.mat-calendar-body-disabled]="!item.enabled" [class.mat-calendar-body-active]="_isActiveCell(rowIndex, colIndex)" [attr.aria-label]="item.ariaLabel" [attr.aria-disabled]="!item.enabled || null" [attr.aria-selected]="selectedValue === item.value" (click)="_cellClicked(item)" [style.width]="_cellWidth" [style.paddingTop]="_cellPadding" [style.paddingBottom]="_cellPadding"><div class="mat-calendar-body-cell-content" [class.mat-calendar-body-selected]="selectedValue === item.value" [class.mat-calendar-body-today]="todayValue === item.value">{{item.displayValue}}</div></td></tr>',styles:[".mat-calendar-body{min-width:224px}.mat-calendar-body-label{height:0;line-height:0;text-align:left;padding-left:4.71429%;padding-right:4.71429%}.mat-calendar-body-cell{position:relative;height:0;line-height:0;text-align:center;outline:0;cursor:pointer}.mat-calendar-body-disabled{cursor:default}.mat-calendar-body-cell-content{position:absolute;top:5%;left:5%;display:flex;align-items:center;justify-content:center;box-sizing:border-box;width:90%;height:90%;line-height:1;border-width:1px;border-style:solid;border-radius:999px}@media (-ms-high-contrast:active){.mat-calendar-body-cell-content{border:none}}@media (-ms-high-contrast:active){.mat-calendar-body-selected,.mat-datepicker-popup:not(:empty){outline:solid 1px}.mat-calendar-body-today{outline:dotted 1px}}[dir=rtl] .mat-calendar-body-label{text-align:right}"],host:{class:"mat-calendar-body",role:"grid","aria-readonly":"true"},exportAs:"matCalendarBody",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:r.NgZone}]},e.propDecorators={label:[{type:r.Input}],rows:[{type:r.Input}],todayValue:[{type:r.Input}],selectedValue:[{type:r.Input}],labelMinRequiredCells:[{type:r.Input}],numCols:[{type:r.Input}],activeCell:[{type:r.Input}],cellAspectRatio:[{type:r.Input}],selectedValueChange:[{type:r.Output}]},e}(),C=function(){function e(e,t,n,i){if(this._changeDetectorRef=e,this._dateFormats=t,this._dateAdapter=n,this._dir=i,this.selectedChange=new r.EventEmitter,this._userSelection=new r.EventEmitter,this.activeDateChange=new r.EventEmitter,!this._dateAdapter)throw j("DateAdapter");if(!this._dateFormats)throw j("MAT_DATE_FORMATS");var o=this._dateAdapter.getFirstDayOfWeek(),a=this._dateAdapter.getDayOfWeekNames("narrow"),s=this._dateAdapter.getDayOfWeekNames("long").map(function(e,t){return{long:e,narrow:a[t]}});this._weekdays=s.slice(o).concat(s.slice(0,o)),this._activeDate=this._dateAdapter.today()}return Object.defineProperty(e.prototype,"activeDate",{get:function(){return this._activeDate},set:function(e){var t=this._activeDate,n=this._getValidDateOrNull(this._dateAdapter.deserialize(e))||this._dateAdapter.today();this._activeDate=this._dateAdapter.clampDate(n,this.minDate,this.maxDate),this._hasSameMonthAndYear(t,this._activeDate)||this._init()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selected",{get:function(){return this._selected},set:function(e){this._selected=this._getValidDateOrNull(this._dateAdapter.deserialize(e)),this._selectedDate=this._getDateInCurrentMonth(this._selected)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"minDate",{get:function(){return this._minDate},set:function(e){this._minDate=this._getValidDateOrNull(this._dateAdapter.deserialize(e))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"maxDate",{get:function(){return this._maxDate},set:function(e){this._maxDate=this._getValidDateOrNull(this._dateAdapter.deserialize(e))},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){this._init()},e.prototype._dateSelected=function(e){if(this._selectedDate!=e){var t=this._dateAdapter.getYear(this.activeDate),n=this._dateAdapter.getMonth(this.activeDate),r=this._dateAdapter.createDate(t,n,e);this.selectedChange.emit(r)}this._userSelection.emit()},e.prototype._handleCalendarBodyKeydown=function(e){var t=this._activeDate,n=this._isRtl();switch(e.keyCode){case a.LEFT_ARROW:this.activeDate=this._dateAdapter.addCalendarDays(this._activeDate,n?1:-1);break;case a.RIGHT_ARROW:this.activeDate=this._dateAdapter.addCalendarDays(this._activeDate,n?-1:1);break;case a.UP_ARROW:this.activeDate=this._dateAdapter.addCalendarDays(this._activeDate,-7);break;case a.DOWN_ARROW:this.activeDate=this._dateAdapter.addCalendarDays(this._activeDate,7);break;case a.HOME:this.activeDate=this._dateAdapter.addCalendarDays(this._activeDate,1-this._dateAdapter.getDate(this._activeDate));break;case a.END:this.activeDate=this._dateAdapter.addCalendarDays(this._activeDate,this._dateAdapter.getNumDaysInMonth(this._activeDate)-this._dateAdapter.getDate(this._activeDate));break;case a.PAGE_UP:this.activeDate=e.altKey?this._dateAdapter.addCalendarYears(this._activeDate,-1):this._dateAdapter.addCalendarMonths(this._activeDate,-1);break;case a.PAGE_DOWN:this.activeDate=e.altKey?this._dateAdapter.addCalendarYears(this._activeDate,1):this._dateAdapter.addCalendarMonths(this._activeDate,1);break;case a.ENTER:case a.SPACE:return void(this.dateFilter&&!this.dateFilter(this._activeDate)||(this._dateSelected(this._dateAdapter.getDate(this._activeDate)),this._userSelection.emit(),e.preventDefault()));default:return}this._dateAdapter.compareDate(t,this.activeDate)&&this.activeDateChange.emit(this.activeDate),this._focusActiveCell(),e.preventDefault()},e.prototype._init=function(){this._selectedDate=this._getDateInCurrentMonth(this.selected),this._todayDate=this._getDateInCurrentMonth(this._dateAdapter.today()),this._monthLabel=this._dateAdapter.getMonthNames("short")[this._dateAdapter.getMonth(this.activeDate)].toLocaleUpperCase();var e=this._dateAdapter.createDate(this._dateAdapter.getYear(this.activeDate),this._dateAdapter.getMonth(this.activeDate),1);this._firstWeekOffset=(7+this._dateAdapter.getDayOfWeek(e)-this._dateAdapter.getFirstDayOfWeek())%7,this._createWeekCells(),this._changeDetectorRef.markForCheck()},e.prototype._focusActiveCell=function(){this._matCalendarBody._focusActiveCell()},e.prototype._createWeekCells=function(){var e=this._dateAdapter.getNumDaysInMonth(this.activeDate),t=this._dateAdapter.getDateNames();this._weeks=[[]];for(var n=0,r=this._firstWeekOffset;n<e;n++,r++){7==r&&(this._weeks.push([]),r=0);var i=this._dateAdapter.createDate(this._dateAdapter.getYear(this.activeDate),this._dateAdapter.getMonth(this.activeDate),n+1),o=this._shouldEnableDate(i),a=this._dateAdapter.format(i,this._dateFormats.display.dateA11yLabel),s=this.dateClass?this.dateClass(i):void 0;this._weeks[this._weeks.length-1].push(new O(n+1,t[n],a,o,s))}},e.prototype._shouldEnableDate=function(e){return!!e&&(!this.dateFilter||this.dateFilter(e))&&(!this.minDate||this._dateAdapter.compareDate(e,this.minDate)>=0)&&(!this.maxDate||this._dateAdapter.compareDate(e,this.maxDate)<=0)},e.prototype._getDateInCurrentMonth=function(e){return e&&this._hasSameMonthAndYear(e,this.activeDate)?this._dateAdapter.getDate(e):null},e.prototype._hasSameMonthAndYear=function(e,t){return!(!e||!t||this._dateAdapter.getMonth(e)!=this._dateAdapter.getMonth(t)||this._dateAdapter.getYear(e)!=this._dateAdapter.getYear(t))},e.prototype._getValidDateOrNull=function(e){return this._dateAdapter.isDateInstance(e)&&this._dateAdapter.isValid(e)?e:null},e.prototype._isRtl=function(){return this._dir&&"rtl"===this._dir.value},e.decorators=[{type:r.Component,args:[{selector:"mat-month-view",template:'<table class="mat-calendar-table"><thead class="mat-calendar-table-header"><tr><th *ngFor="let day of _weekdays" [attr.aria-label]="day.long">{{day.narrow}}</th></tr><tr><th class="mat-calendar-table-header-divider" colspan="7" aria-hidden="true"></th></tr></thead><tbody mat-calendar-body [label]="_monthLabel" [rows]="_weeks" [todayValue]="_todayDate" [selectedValue]="_selectedDate" [labelMinRequiredCells]="3" [activeCell]="_dateAdapter.getDate(activeDate) - 1" (selectedValueChange)="_dateSelected($event)" (keydown)="_handleCalendarBodyKeydown($event)"></tbody></table>',exportAs:"matMonthView",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:r.ChangeDetectorRef},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[s.MAT_DATE_FORMATS]}]},{type:s.DateAdapter,decorators:[{type:r.Optional}]},{type:u.Directionality,decorators:[{type:r.Optional}]}]},e.propDecorators={activeDate:[{type:r.Input}],selected:[{type:r.Input}],minDate:[{type:r.Input}],maxDate:[{type:r.Input}],dateFilter:[{type:r.Input}],dateClass:[{type:r.Input}],selectedChange:[{type:r.Output}],_userSelection:[{type:r.Output}],activeDateChange:[{type:r.Output}],_matCalendarBody:[{type:r.ViewChild,args:[w]}]},e}(),E=function(){function e(e,t,n){if(this._changeDetectorRef=e,this._dateAdapter=t,this._dir=n,this.selectedChange=new r.EventEmitter,this.yearSelected=new r.EventEmitter,this.activeDateChange=new r.EventEmitter,!this._dateAdapter)throw j("DateAdapter");this._activeDate=this._dateAdapter.today()}return Object.defineProperty(e.prototype,"activeDate",{get:function(){return this._activeDate},set:function(e){var t=this._activeDate,n=this._getValidDateOrNull(this._dateAdapter.deserialize(e))||this._dateAdapter.today();this._activeDate=this._dateAdapter.clampDate(n,this.minDate,this.maxDate),Math.floor(this._dateAdapter.getYear(t)/24)!=Math.floor(this._dateAdapter.getYear(this._activeDate)/24)&&this._init()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selected",{get:function(){return this._selected},set:function(e){this._selected=this._getValidDateOrNull(this._dateAdapter.deserialize(e)),this._selectedYear=this._selected&&this._dateAdapter.getYear(this._selected)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"minDate",{get:function(){return this._minDate},set:function(e){this._minDate=this._getValidDateOrNull(this._dateAdapter.deserialize(e))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"maxDate",{get:function(){return this._maxDate},set:function(e){this._maxDate=this._getValidDateOrNull(this._dateAdapter.deserialize(e))},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){this._init()},e.prototype._init=function(){var e=this;this._todayYear=this._dateAdapter.getYear(this._dateAdapter.today());var t=this._dateAdapter.getYear(this._activeDate),n=t%24;this._years=[];for(var r=0,i=[];r<24;r++)i.push(t-n+r),4==i.length&&(this._years.push(i.map(function(t){return e._createCellForYear(t)})),i=[]);this._changeDetectorRef.markForCheck()},e.prototype._yearSelected=function(e){this.yearSelected.emit(this._dateAdapter.createDate(e,0,1));var t=this._dateAdapter.getMonth(this.activeDate),n=this._dateAdapter.getNumDaysInMonth(this._dateAdapter.createDate(e,t,1));this.selectedChange.emit(this._dateAdapter.createDate(e,t,Math.min(this._dateAdapter.getDate(this.activeDate),n)))},e.prototype._handleCalendarBodyKeydown=function(e){var t=this._activeDate,n=this._isRtl();switch(e.keyCode){case a.LEFT_ARROW:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,n?1:-1);break;case a.RIGHT_ARROW:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,n?-1:1);break;case a.UP_ARROW:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,-4);break;case a.DOWN_ARROW:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,4);break;case a.HOME:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,-this._dateAdapter.getYear(this._activeDate)%24);break;case a.END:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,24-this._dateAdapter.getYear(this._activeDate)%24-1);break;case a.PAGE_UP:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,e.altKey?-240:-24);break;case a.PAGE_DOWN:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,e.altKey?240:24);break;case a.ENTER:case a.SPACE:this._yearSelected(this._dateAdapter.getYear(this._activeDate));break;default:return}this._dateAdapter.compareDate(t,this.activeDate)&&this.activeDateChange.emit(this.activeDate),this._focusActiveCell(),e.preventDefault()},e.prototype._getActiveCell=function(){return this._dateAdapter.getYear(this.activeDate)%24},e.prototype._focusActiveCell=function(){this._matCalendarBody._focusActiveCell()},e.prototype._createCellForYear=function(e){var t=this._dateAdapter.getYearName(this._dateAdapter.createDate(e,0,1));return new O(e,t,t,this._shouldEnableYear(e))},e.prototype._shouldEnableYear=function(e){if(void 0===e||null===e||this.maxDate&&e>this._dateAdapter.getYear(this.maxDate)||this.minDate&&e<this._dateAdapter.getYear(this.minDate))return!1;if(!this.dateFilter)return!0;for(var t=this._dateAdapter.createDate(e,0,1);this._dateAdapter.getYear(t)==e;t=this._dateAdapter.addCalendarDays(t,1))if(this.dateFilter(t))return!0;return!1},e.prototype._getValidDateOrNull=function(e){return this._dateAdapter.isDateInstance(e)&&this._dateAdapter.isValid(e)?e:null},e.prototype._isRtl=function(){return this._dir&&"rtl"===this._dir.value},e.decorators=[{type:r.Component,args:[{selector:"mat-multi-year-view",template:'<table class="mat-calendar-table"><thead class="mat-calendar-table-header"><tr><th class="mat-calendar-table-header-divider" colspan="4"></th></tr></thead><tbody mat-calendar-body [rows]="_years" [todayValue]="_todayYear" [selectedValue]="_selectedYear" [numCols]="4" [cellAspectRatio]="4 / 7" [activeCell]="_getActiveCell()" (selectedValueChange)="_yearSelected($event)" (keydown)="_handleCalendarBodyKeydown($event)"></tbody></table>',exportAs:"matMultiYearView",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:r.ChangeDetectorRef},{type:s.DateAdapter,decorators:[{type:r.Optional}]},{type:u.Directionality,decorators:[{type:r.Optional}]}]},e.propDecorators={activeDate:[{type:r.Input}],selected:[{type:r.Input}],minDate:[{type:r.Input}],maxDate:[{type:r.Input}],dateFilter:[{type:r.Input}],selectedChange:[{type:r.Output}],yearSelected:[{type:r.Output}],activeDateChange:[{type:r.Output}],_matCalendarBody:[{type:r.ViewChild,args:[w]}]},e}(),S=function(){function e(e,t,n,i){if(this._changeDetectorRef=e,this._dateFormats=t,this._dateAdapter=n,this._dir=i,this.selectedChange=new r.EventEmitter,this.monthSelected=new r.EventEmitter,this.activeDateChange=new r.EventEmitter,!this._dateAdapter)throw j("DateAdapter");if(!this._dateFormats)throw j("MAT_DATE_FORMATS");this._activeDate=this._dateAdapter.today()}return Object.defineProperty(e.prototype,"activeDate",{get:function(){return this._activeDate},set:function(e){var t=this._activeDate,n=this._getValidDateOrNull(this._dateAdapter.deserialize(e))||this._dateAdapter.today();this._activeDate=this._dateAdapter.clampDate(n,this.minDate,this.maxDate),this._dateAdapter.getYear(t)!==this._dateAdapter.getYear(this._activeDate)&&this._init()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selected",{get:function(){return this._selected},set:function(e){this._selected=this._getValidDateOrNull(this._dateAdapter.deserialize(e)),this._selectedMonth=this._getMonthInCurrentYear(this._selected)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"minDate",{get:function(){return this._minDate},set:function(e){this._minDate=this._getValidDateOrNull(this._dateAdapter.deserialize(e))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"maxDate",{get:function(){return this._maxDate},set:function(e){this._maxDate=this._getValidDateOrNull(this._dateAdapter.deserialize(e))},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){this._init()},e.prototype._monthSelected=function(e){var t=this._dateAdapter.createDate(this._dateAdapter.getYear(this.activeDate),e,1);this.monthSelected.emit(t);var n=this._dateAdapter.getNumDaysInMonth(t);this.selectedChange.emit(this._dateAdapter.createDate(this._dateAdapter.getYear(this.activeDate),e,Math.min(this._dateAdapter.getDate(this.activeDate),n)))},e.prototype._handleCalendarBodyKeydown=function(e){var t=this._activeDate,n=this._isRtl();switch(e.keyCode){case a.LEFT_ARROW:this.activeDate=this._dateAdapter.addCalendarMonths(this._activeDate,n?1:-1);break;case a.RIGHT_ARROW:this.activeDate=this._dateAdapter.addCalendarMonths(this._activeDate,n?-1:1);break;case a.UP_ARROW:this.activeDate=this._dateAdapter.addCalendarMonths(this._activeDate,-4);break;case a.DOWN_ARROW:this.activeDate=this._dateAdapter.addCalendarMonths(this._activeDate,4);break;case a.HOME:this.activeDate=this._dateAdapter.addCalendarMonths(this._activeDate,-this._dateAdapter.getMonth(this._activeDate));break;case a.END:this.activeDate=this._dateAdapter.addCalendarMonths(this._activeDate,11-this._dateAdapter.getMonth(this._activeDate));break;case a.PAGE_UP:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,e.altKey?-10:-1);break;case a.PAGE_DOWN:this.activeDate=this._dateAdapter.addCalendarYears(this._activeDate,e.altKey?10:1);break;case a.ENTER:case a.SPACE:this._monthSelected(this._dateAdapter.getMonth(this._activeDate));break;default:return}this._dateAdapter.compareDate(t,this.activeDate)&&this.activeDateChange.emit(this.activeDate),this._focusActiveCell(),e.preventDefault()},e.prototype._init=function(){var e=this;this._selectedMonth=this._getMonthInCurrentYear(this.selected),this._todayMonth=this._getMonthInCurrentYear(this._dateAdapter.today()),this._yearLabel=this._dateAdapter.getYearName(this.activeDate);var t=this._dateAdapter.getMonthNames("short");this._months=[[0,1,2,3],[4,5,6,7],[8,9,10,11]].map(function(n){return n.map(function(n){return e._createCellForMonth(n,t[n])})}),this._changeDetectorRef.markForCheck()},e.prototype._focusActiveCell=function(){this._matCalendarBody._focusActiveCell()},e.prototype._getMonthInCurrentYear=function(e){return e&&this._dateAdapter.getYear(e)==this._dateAdapter.getYear(this.activeDate)?this._dateAdapter.getMonth(e):null},e.prototype._createCellForMonth=function(e,t){var n=this._dateAdapter.format(this._dateAdapter.createDate(this._dateAdapter.getYear(this.activeDate),e,1),this._dateFormats.display.monthYearA11yLabel);return new O(e,t.toLocaleUpperCase(),n,this._shouldEnableMonth(e))},e.prototype._shouldEnableMonth=function(e){var t=this._dateAdapter.getYear(this.activeDate);if(void 0===e||null===e||this._isYearAndMonthAfterMaxDate(t,e)||this._isYearAndMonthBeforeMinDate(t,e))return!1;if(!this.dateFilter)return!0;for(var n=this._dateAdapter.createDate(t,e,1);this._dateAdapter.getMonth(n)==e;n=this._dateAdapter.addCalendarDays(n,1))if(this.dateFilter(n))return!0;return!1},e.prototype._isYearAndMonthAfterMaxDate=function(e,t){if(this.maxDate){var n=this._dateAdapter.getYear(this.maxDate),r=this._dateAdapter.getMonth(this.maxDate);return e>n||e===n&&t>r}return!1},e.prototype._isYearAndMonthBeforeMinDate=function(e,t){if(this.minDate){var n=this._dateAdapter.getYear(this.minDate),r=this._dateAdapter.getMonth(this.minDate);return e<n||e===n&&t<r}return!1},e.prototype._getValidDateOrNull=function(e){return this._dateAdapter.isDateInstance(e)&&this._dateAdapter.isValid(e)?e:null},e.prototype._isRtl=function(){return this._dir&&"rtl"===this._dir.value},e.decorators=[{type:r.Component,args:[{selector:"mat-year-view",template:'<table class="mat-calendar-table"><thead class="mat-calendar-table-header"><tr><th class="mat-calendar-table-header-divider" colspan="4"></th></tr></thead><tbody mat-calendar-body [label]="_yearLabel" [rows]="_months" [todayValue]="_todayMonth" [selectedValue]="_selectedMonth" [labelMinRequiredCells]="2" [numCols]="4" [cellAspectRatio]="4 / 7" [activeCell]="_dateAdapter.getMonth(activeDate)" (selectedValueChange)="_monthSelected($event)" (keydown)="_handleCalendarBodyKeydown($event)"></tbody></table>',exportAs:"matYearView",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:r.ChangeDetectorRef},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[s.MAT_DATE_FORMATS]}]},{type:s.DateAdapter,decorators:[{type:r.Optional}]},{type:u.Directionality,decorators:[{type:r.Optional}]}]},e.propDecorators={activeDate:[{type:r.Input}],selected:[{type:r.Input}],minDate:[{type:r.Input}],maxDate:[{type:r.Input}],dateFilter:[{type:r.Input}],selectedChange:[{type:r.Output}],monthSelected:[{type:r.Output}],activeDateChange:[{type:r.Output}],_matCalendarBody:[{type:r.ViewChild,args:[w]}]},e}(),k=function(){function e(e,t,n,r,i){this._intl=e,this.calendar=t,this._dateAdapter=n,this._dateFormats=r,this.calendar.stateChanges.subscribe(function(){return i.markForCheck()})}return Object.defineProperty(e.prototype,"periodButtonText",{get:function(){if("month"==this.calendar.currentView)return this._dateAdapter.format(this.calendar.activeDate,this._dateFormats.display.monthYearLabel).toLocaleUpperCase();if("year"==this.calendar.currentView)return this._dateAdapter.getYearName(this.calendar.activeDate);var e=this._dateAdapter.getYear(this.calendar.activeDate);return this._dateAdapter.getYearName(this._dateAdapter.createDate(e-e%24,0,1))+" – "+this._dateAdapter.getYearName(this._dateAdapter.createDate(e+24-1-e%24,0,1))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"periodButtonLabel",{get:function(){return"month"==this.calendar.currentView?this._intl.switchToMultiYearViewLabel:this._intl.switchToMonthViewLabel},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"prevButtonLabel",{get:function(){return{month:this._intl.prevMonthLabel,year:this._intl.prevYearLabel,"multi-year":this._intl.prevMultiYearLabel}[this.calendar.currentView]},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"nextButtonLabel",{get:function(){return{month:this._intl.nextMonthLabel,year:this._intl.nextYearLabel,"multi-year":this._intl.nextMultiYearLabel}[this.calendar.currentView]},enumerable:!0,configurable:!0}),e.prototype.currentPeriodClicked=function(){this.calendar.currentView="month"==this.calendar.currentView?"multi-year":"month"},e.prototype.previousClicked=function(){this.calendar.activeDate="month"==this.calendar.currentView?this._dateAdapter.addCalendarMonths(this.calendar.activeDate,-1):this._dateAdapter.addCalendarYears(this.calendar.activeDate,"year"==this.calendar.currentView?-1:-24)},e.prototype.nextClicked=function(){this.calendar.activeDate="month"==this.calendar.currentView?this._dateAdapter.addCalendarMonths(this.calendar.activeDate,1):this._dateAdapter.addCalendarYears(this.calendar.activeDate,"year"==this.calendar.currentView?1:24)},e.prototype.previousEnabled=function(){return!this.calendar.minDate||(!this.calendar.minDate||!this._isSameView(this.calendar.activeDate,this.calendar.minDate))},e.prototype.nextEnabled=function(){return!this.calendar.maxDate||!this._isSameView(this.calendar.activeDate,this.calendar.maxDate)},e.prototype._isSameView=function(e,t){return"month"==this.calendar.currentView?this._dateAdapter.getYear(e)==this._dateAdapter.getYear(t)&&this._dateAdapter.getMonth(e)==this._dateAdapter.getMonth(t):"year"==this.calendar.currentView?this._dateAdapter.getYear(e)==this._dateAdapter.getYear(t):Math.floor(this._dateAdapter.getYear(e)/24)==Math.floor(this._dateAdapter.getYear(t)/24)},e.decorators=[{type:r.Component,args:[{selector:"mat-calendar-header",template:'<div class="mat-calendar-header"><div class="mat-calendar-controls"><button mat-button type="button" class="mat-calendar-period-button" (click)="currentPeriodClicked()" [attr.aria-label]="periodButtonLabel" cdkAriaLive="polite">{{periodButtonText}}<div class="mat-calendar-arrow" [class.mat-calendar-invert]="calendar.currentView != \'month\'"></div></button><div class="mat-calendar-spacer"></div><ng-content></ng-content><button mat-icon-button type="button" class="mat-calendar-previous-button" [disabled]="!previousEnabled()" (click)="previousClicked()" [attr.aria-label]="prevButtonLabel"></button> <button mat-icon-button type="button" class="mat-calendar-next-button" [disabled]="!nextEnabled()" (click)="nextClicked()" [attr.aria-label]="nextButtonLabel"></button></div></div>',exportAs:"matCalendarHeader",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:x},{type:M,decorators:[{type:r.Inject,args:[Object(r.forwardRef)(function(){return M})]}]},{type:s.DateAdapter,decorators:[{type:r.Optional}]},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[s.MAT_DATE_FORMATS]}]},{type:r.ChangeDetectorRef}]},e}(),M=function(){function e(e,t,n,o){var a=this;if(this._dateAdapter=t,this._dateFormats=n,this._changeDetectorRef=o,this._moveFocusOnNextTick=!1,this.startView="month",this.selectedChange=new r.EventEmitter,this.yearSelected=new r.EventEmitter,this.monthSelected=new r.EventEmitter,this._userSelection=new r.EventEmitter,this.stateChanges=new i.Subject,!this._dateAdapter)throw j("DateAdapter");if(!this._dateFormats)throw j("MAT_DATE_FORMATS");this._intlChanges=e.changes.subscribe(function(){o.markForCheck(),a.stateChanges.next()})}return Object.defineProperty(e.prototype,"startAt",{get:function(){return this._startAt},set:function(e){this._startAt=this._getValidDateOrNull(this._dateAdapter.deserialize(e))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selected",{get:function(){return this._selected},set:function(e){this._selected=this._getValidDateOrNull(this._dateAdapter.deserialize(e))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"minDate",{get:function(){return this._minDate},set:function(e){this._minDate=this._getValidDateOrNull(this._dateAdapter.deserialize(e))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"maxDate",{get:function(){return this._maxDate},set:function(e){this._maxDate=this._getValidDateOrNull(this._dateAdapter.deserialize(e))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"activeDate",{get:function(){return this._clampedActiveDate},set:function(e){this._clampedActiveDate=this._dateAdapter.clampDate(e,this.minDate,this.maxDate),this.stateChanges.next()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"currentView",{get:function(){return this._currentView},set:function(e){this._currentView=e,this._moveFocusOnNextTick=!0},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){this._calendarHeaderPortal=new c.ComponentPortal(this.headerComponent||k),this.activeDate=this.startAt||this._dateAdapter.today(),this._currentView=this.startView},e.prototype.ngAfterViewChecked=function(){this._moveFocusOnNextTick&&(this._moveFocusOnNextTick=!1,this.focusActiveCell())},e.prototype.ngOnDestroy=function(){this._intlChanges.unsubscribe(),this.stateChanges.complete()},e.prototype.ngOnChanges=function(e){var t=e.minDate||e.maxDate||e.dateFilter;if(t&&!t.firstChange){var n=this._getCurrentViewComponent();n&&(this._changeDetectorRef.detectChanges(),n._init())}this.stateChanges.next()},e.prototype.focusActiveCell=function(){this._getCurrentViewComponent()._focusActiveCell()},e.prototype.updateTodaysDate=function(){("month"==this.currentView?this.monthView:"year"==this.currentView?this.yearView:this.multiYearView).ngAfterContentInit()},e.prototype._dateSelected=function(e){this._dateAdapter.sameDate(e,this.selected)||this.selectedChange.emit(e)},e.prototype._yearSelectedInMultiYearView=function(e){this.yearSelected.emit(e)},e.prototype._monthSelectedInYearView=function(e){this.monthSelected.emit(e)},e.prototype._userSelected=function(){this._userSelection.emit()},e.prototype._goToDateInView=function(e,t){this.activeDate=e,this.currentView=t},e.prototype._getValidDateOrNull=function(e){return this._dateAdapter.isDateInstance(e)&&this._dateAdapter.isValid(e)?e:null},e.prototype._getCurrentViewComponent=function(){return this.monthView||this.yearView||this.multiYearView},e.decorators=[{type:r.Component,args:[{selector:"mat-calendar",template:'<ng-template [cdkPortalOutlet]="_calendarHeaderPortal"></ng-template><div class="mat-calendar-content" [ngSwitch]="currentView" cdkMonitorSubtreeFocus tabindex="-1"><mat-month-view *ngSwitchCase="\'month\'" [(activeDate)]="activeDate" [selected]="selected" [dateFilter]="dateFilter" [maxDate]="maxDate" [minDate]="minDate" [dateClass]="dateClass" (selectedChange)="_dateSelected($event)" (_userSelection)="_userSelected()"></mat-month-view><mat-year-view *ngSwitchCase="\'year\'" [(activeDate)]="activeDate" [selected]="selected" [dateFilter]="dateFilter" [maxDate]="maxDate" [minDate]="minDate" (monthSelected)="_monthSelectedInYearView($event)" (selectedChange)="_goToDateInView($event, \'month\')"></mat-year-view><mat-multi-year-view *ngSwitchCase="\'multi-year\'" [(activeDate)]="activeDate" [selected]="selected" [dateFilter]="dateFilter" [maxDate]="maxDate" [minDate]="minDate" (yearSelected)="_yearSelectedInMultiYearView($event)" (selectedChange)="_goToDateInView($event, \'year\')"></mat-multi-year-view></div>',styles:[".mat-calendar{display:block}.mat-calendar-header{padding:8px 8px 0 8px}.mat-calendar-content{padding:0 8px 8px 8px;outline:0}.mat-calendar-controls{display:flex;margin:5% calc(33% / 7 - 16px)}.mat-calendar-spacer{flex:1 1 auto}.mat-calendar-period-button{min-width:0}.mat-calendar-arrow{display:inline-block;width:0;height:0;border-left:5px solid transparent;border-right:5px solid transparent;border-top-width:5px;border-top-style:solid;margin:0 0 0 5px;vertical-align:middle}.mat-calendar-arrow.mat-calendar-invert{transform:rotate(180deg)}[dir=rtl] .mat-calendar-arrow{margin:0 5px 0 0}.mat-calendar-next-button,.mat-calendar-previous-button{position:relative}.mat-calendar-next-button::after,.mat-calendar-previous-button::after{top:0;left:0;right:0;bottom:0;position:absolute;content:'';margin:15.5px;border:0 solid currentColor;border-top-width:2px}[dir=rtl] .mat-calendar-next-button,[dir=rtl] .mat-calendar-previous-button{transform:rotate(180deg)}.mat-calendar-previous-button::after{border-left-width:2px;transform:translateX(2px) rotate(-45deg)}.mat-calendar-next-button::after{border-right-width:2px;transform:translateX(-2px) rotate(45deg)}.mat-calendar-table{border-spacing:0;border-collapse:collapse;width:100%}.mat-calendar-table-header th{text-align:center;padding:0 0 8px 0}.mat-calendar-table-header-divider{position:relative;height:1px}.mat-calendar-table-header-divider::after{content:'';position:absolute;top:0;left:-8px;right:-8px;height:1px}"],host:{class:"mat-calendar"},exportAs:"matCalendar",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:x},{type:s.DateAdapter,decorators:[{type:r.Optional}]},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[s.MAT_DATE_FORMATS]}]},{type:r.ChangeDetectorRef}]},e.propDecorators={headerComponent:[{type:r.Input}],startAt:[{type:r.Input}],startView:[{type:r.Input}],selected:[{type:r.Input}],minDate:[{type:r.Input}],maxDate:[{type:r.Input}],dateFilter:[{type:r.Input}],dateClass:[{type:r.Input}],selectedChange:[{type:r.Output}],yearSelected:[{type:r.Output}],monthSelected:[{type:r.Output}],_userSelection:[{type:r.Output}],monthView:[{type:r.ViewChild,args:[C]}],yearView:[{type:r.ViewChild,args:[S]}],multiYearView:[{type:r.ViewChild,args:[E]}]},e}(),A={transformPanel:Object(l.trigger)("transformPanel",[Object(l.state)("void",Object(l.style)({opacity:0,transform:"scale(1, 0.8)"})),Object(l.transition)("void => enter",Object(l.animate)("120ms cubic-bezier(0, 0, 0.2, 1)",Object(l.style)({opacity:1,transform:"scale(1, 1)"}))),Object(l.transition)("* => void",Object(l.animate)("100ms linear",Object(l.style)({opacity:0})))]),fadeInCalendar:Object(l.trigger)("fadeInCalendar",[Object(l.state)("void",Object(l.style)({opacity:0})),Object(l.state)("enter",Object(l.style)({opacity:1})),Object(l.transition)("void => *",Object(l.animate)("120ms 100ms cubic-bezier(0.55, 0, 0.55, 0.2)"))])},T=0,D=new r.InjectionToken("mat-datepicker-scroll-strategy");function I(e){return function(){return e.scrollStrategies.reposition()}}var P={provide:D,deps:[f.Overlay],useFactory:I},R=function(){return function(e){this._elementRef=e}}(),N=Object(s.mixinColor)(R),F=function(e){function t(t){return e.call(this,t)||this}return Object(d.__extends)(t,e),t.prototype.ngAfterViewInit=function(){this._calendar.focusActiveCell()},t.decorators=[{type:r.Component,args:[{selector:"mat-datepicker-content",template:'<mat-calendar cdkTrapFocus [id]="datepicker.id" [ngClass]="datepicker.panelClass" [startAt]="datepicker.startAt" [startView]="datepicker.startView" [minDate]="datepicker._minDate" [maxDate]="datepicker._maxDate" [dateFilter]="datepicker._dateFilter" [headerComponent]="datepicker.calendarHeaderComponent" [selected]="datepicker._selected" [dateClass]="datepicker.dateClass" [@fadeInCalendar]="\'enter\'" (selectedChange)="datepicker.select($event)" (yearSelected)="datepicker._selectYear($event)" (monthSelected)="datepicker._selectMonth($event)" (_userSelection)="datepicker.close()"></mat-calendar>',styles:[".mat-datepicker-content{display:block;border-radius:4px}.mat-datepicker-content .mat-calendar{width:296px;height:354px}.mat-datepicker-content-touch{display:block;max-height:80vh;overflow:auto;margin:-24px}.mat-datepicker-content-touch .mat-calendar{min-width:250px;min-height:312px;max-width:750px;max-height:788px}@media all and (orientation:landscape){.mat-datepicker-content-touch .mat-calendar{width:64vh;height:80vh}}@media all and (orientation:portrait){.mat-datepicker-content-touch .mat-calendar{width:80vw;height:100vw}}"],host:{class:"mat-datepicker-content","[@transformPanel]":'"enter"',"[class.mat-datepicker-content-touch]":"datepicker.touchUi"},animations:[A.transformPanel,A.fadeInCalendar],exportAs:"matDatepickerContent",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,inputs:["color"]}]}],t.ctorParameters=function(){return[{type:r.ElementRef}]},t.propDecorators={_calendar:[{type:r.ViewChild,args:[M]}]},t}(N),L=function(){function e(e,t,n,o,a,s,u,c){if(this._dialog=e,this._overlay=t,this._ngZone=n,this._viewContainerRef=o,this._dateAdapter=s,this._dir=u,this._document=c,this.startView="month",this._touchUi=!1,this.yearSelected=new r.EventEmitter,this.monthSelected=new r.EventEmitter,this.openedStream=new r.EventEmitter,this.closedStream=new r.EventEmitter,this._opened=!1,this.id="mat-datepicker-"+T++,this._validSelected=null,this._focusedElementBeforeOpen=null,this._inputSubscription=i.Subscription.EMPTY,this._disabledChange=new i.Subject,this._selectedChanged=new i.Subject,!this._dateAdapter)throw j("DateAdapter");this._scrollStrategy=a}return Object.defineProperty(e.prototype,"startAt",{get:function(){return this._startAt||(this._datepickerInput?this._datepickerInput.value:null)},set:function(e){this._startAt=this._getValidDateOrNull(this._dateAdapter.deserialize(e))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"color",{get:function(){return this._color||(this._datepickerInput?this._datepickerInput._getThemePalette():void 0)},set:function(e){this._color=e},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"touchUi",{get:function(){return this._touchUi},set:function(e){this._touchUi=Object(p.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disabled",{get:function(){return void 0===this._disabled&&this._datepickerInput?this._datepickerInput.disabled:!!this._disabled},set:function(e){var t=Object(p.coerceBooleanProperty)(e);t!==this._disabled&&(this._disabled=t,this._disabledChange.next(t))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"opened",{get:function(){return this._opened},set:function(e){e?this.open():this.close()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_selected",{get:function(){return this._validSelected},set:function(e){this._validSelected=e},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_minDate",{get:function(){return this._datepickerInput&&this._datepickerInput.min},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_maxDate",{get:function(){return this._datepickerInput&&this._datepickerInput.max},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_dateFilter",{get:function(){return this._datepickerInput&&this._datepickerInput._dateFilter},enumerable:!0,configurable:!0}),e.prototype.ngOnDestroy=function(){this.close(),this._inputSubscription.unsubscribe(),this._disabledChange.complete(),this._popupRef&&(this._popupRef.dispose(),this._popupComponentRef=null)},e.prototype.select=function(e){var t=this._selected;this._selected=e,this._dateAdapter.sameDate(t,this._selected)||this._selectedChanged.next(e)},e.prototype._selectYear=function(e){this.yearSelected.emit(e)},e.prototype._selectMonth=function(e){this.monthSelected.emit(e)},e.prototype._registerInput=function(e){var t=this;if(this._datepickerInput)throw Error("A MatDatepicker can only be associated with a single input.");this._datepickerInput=e,this._inputSubscription=this._datepickerInput._valueChange.subscribe(function(e){return t._selected=e})},e.prototype.open=function(){if(!this._opened&&!this.disabled){if(!this._datepickerInput)throw Error("Attempted to open an MatDatepicker with no associated input.");this._document&&(this._focusedElementBeforeOpen=this._document.activeElement),this.touchUi?this._openAsDialog():this._openAsPopup(),this._opened=!0,this.openedStream.emit()}},e.prototype.close=function(){var e=this;if(this._opened){this._popupRef&&this._popupRef.hasAttached()&&this._popupRef.detach(),this._dialogRef&&(this._dialogRef.close(),this._dialogRef=null),this._calendarPortal&&this._calendarPortal.isAttached&&this._calendarPortal.detach();var t=function(){e._opened&&(e._opened=!1,e.closedStream.emit(),e._focusedElementBeforeOpen=null)};this._focusedElementBeforeOpen&&"function"==typeof this._focusedElementBeforeOpen.focus?(this._focusedElementBeforeOpen.focus(),setTimeout(t)):t()}},e.prototype._openAsDialog=function(){var e=this;this._dialogRef&&this._dialogRef.close(),this._dialogRef=this._dialog.open(F,{direction:this._dir?this._dir.value:"ltr",viewContainerRef:this._viewContainerRef,panelClass:"mat-datepicker-dialog"}),this._dialogRef.afterClosed().subscribe(function(){return e.close()}),this._dialogRef.componentInstance.datepicker=this,this._setColor()},e.prototype._openAsPopup=function(){var e=this;this._calendarPortal||(this._calendarPortal=new c.ComponentPortal(F,this._viewContainerRef)),this._popupRef||this._createPopup(),this._popupRef.hasAttached()||(this._popupComponentRef=this._popupRef.attach(this._calendarPortal),this._popupComponentRef.instance.datepicker=this,this._setColor(),this._ngZone.onStable.asObservable().pipe(Object(o.take)(1)).subscribe(function(){e._popupRef.updatePosition()}))},e.prototype._createPopup=function(){var e=this,t=new f.OverlayConfig({positionStrategy:this._createPopupPositionStrategy(),hasBackdrop:!0,backdropClass:"mat-overlay-transparent-backdrop",direction:this._dir,scrollStrategy:this._scrollStrategy(),panelClass:"mat-datepicker-popup"});this._popupRef=this._overlay.create(t),this._popupRef.overlayElement.setAttribute("role","dialog"),Object(i.merge)(this._popupRef.backdropClick(),this._popupRef.detachments(),this._popupRef.keydownEvents().pipe(Object(o.filter)(function(t){return t.keyCode===a.ESCAPE||e._datepickerInput&&t.altKey&&t.keyCode===a.UP_ARROW}))).subscribe(function(){return e.close()})},e.prototype._createPopupPositionStrategy=function(){return this._overlay.position().flexibleConnectedTo(this._datepickerInput.getConnectedOverlayOrigin()).withTransformOriginOn(".mat-datepicker-content").withFlexibleDimensions(!1).withViewportMargin(8).withLockedPosition().withPositions([{originX:"start",originY:"bottom",overlayX:"start",overlayY:"top"},{originX:"start",originY:"top",overlayX:"start",overlayY:"bottom"},{originX:"end",originY:"bottom",overlayX:"end",overlayY:"top"},{originX:"end",originY:"top",overlayX:"end",overlayY:"bottom"}])},e.prototype._getValidDateOrNull=function(e){return this._dateAdapter.isDateInstance(e)&&this._dateAdapter.isValid(e)?e:null},e.prototype._setColor=function(){var e=this.color;this._popupComponentRef&&(this._popupComponentRef.instance.color=e),this._dialogRef&&(this._dialogRef.componentInstance.color=e)},e.decorators=[{type:r.Component,args:[{selector:"mat-datepicker",template:"",exportAs:"matDatepicker",changeDetection:r.ChangeDetectionStrategy.OnPush,encapsulation:r.ViewEncapsulation.None}]}],e.ctorParameters=function(){return[{type:m.MatDialog},{type:f.Overlay},{type:r.NgZone},{type:r.ViewContainerRef},{type:void 0,decorators:[{type:r.Inject,args:[D]}]},{type:s.DateAdapter,decorators:[{type:r.Optional}]},{type:u.Directionality,decorators:[{type:r.Optional}]},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[h.DOCUMENT]}]}]},e.propDecorators={calendarHeaderComponent:[{type:r.Input}],startAt:[{type:r.Input}],startView:[{type:r.Input}],color:[{type:r.Input}],touchUi:[{type:r.Input}],disabled:[{type:r.Input}],yearSelected:[{type:r.Output}],monthSelected:[{type:r.Output}],panelClass:[{type:r.Input}],dateClass:[{type:r.Input}],openedStream:[{type:r.Output,args:["opened"]}],closedStream:[{type:r.Output,args:["closed"]}],opened:[{type:r.Input}]},e}(),V={provide:g.NG_VALUE_ACCESSOR,useExisting:Object(r.forwardRef)(function(){return U}),multi:!0},B={provide:g.NG_VALIDATORS,useExisting:Object(r.forwardRef)(function(){return U}),multi:!0},z=function(){return function(e,t){this.target=e,this.targetElement=t,this.value=this.target.value}}(),U=function(){function e(e,t,n,o){var a=this;if(this._elementRef=e,this._dateAdapter=t,this._dateFormats=n,this._formField=o,this.dateChange=new r.EventEmitter,this.dateInput=new r.EventEmitter,this._valueChange=new r.EventEmitter,this._disabledChange=new r.EventEmitter,this._onTouched=function(){},this._cvaOnChange=function(){},this._validatorOnChange=function(){},this._datepickerSubscription=i.Subscription.EMPTY,this._localeSubscription=i.Subscription.EMPTY,this._parseValidator=function(){return a._lastValueValid?null:{matDatepickerParse:{text:a._elementRef.nativeElement.value}}},this._minValidator=function(e){var t=a._getValidDateOrNull(a._dateAdapter.deserialize(e.value));return!a.min||!t||a._dateAdapter.compareDate(a.min,t)<=0?null:{matDatepickerMin:{min:a.min,actual:t}}},this._maxValidator=function(e){var t=a._getValidDateOrNull(a._dateAdapter.deserialize(e.value));return!a.max||!t||a._dateAdapter.compareDate(a.max,t)>=0?null:{matDatepickerMax:{max:a.max,actual:t}}},this._filterValidator=function(e){var t=a._getValidDateOrNull(a._dateAdapter.deserialize(e.value));return a._dateFilter&&t&&!a._dateFilter(t)?{matDatepickerFilter:!0}:null},this._validator=g.Validators.compose([this._parseValidator,this._minValidator,this._maxValidator,this._filterValidator]),this._lastValueValid=!1,!this._dateAdapter)throw j("DateAdapter");if(!this._dateFormats)throw j("MAT_DATE_FORMATS");this._localeSubscription=t.localeChanges.subscribe(function(){a.value=a.value})}return Object.defineProperty(e.prototype,"matDatepicker",{set:function(e){var t=this;e&&(this._datepicker=e,this._datepicker._registerInput(this),this._datepickerSubscription.unsubscribe(),this._datepickerSubscription=this._datepicker._selectedChanged.subscribe(function(e){t.value=e,t._cvaOnChange(e),t._onTouched(),t.dateInput.emit(new z(t,t._elementRef.nativeElement)),t.dateChange.emit(new z(t,t._elementRef.nativeElement))}))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"matDatepickerFilter",{set:function(e){this._dateFilter=e,this._validatorOnChange()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return this._value},set:function(e){e=this._dateAdapter.deserialize(e),this._lastValueValid=!e||this._dateAdapter.isValid(e),e=this._getValidDateOrNull(e);var t=this.value;this._value=e,this._formatValue(e),this._dateAdapter.sameDate(t,e)||this._valueChange.emit(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"min",{get:function(){return this._min},set:function(e){this._min=this._getValidDateOrNull(this._dateAdapter.deserialize(e)),this._validatorOnChange()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"max",{get:function(){return this._max},set:function(e){this._max=this._getValidDateOrNull(this._dateAdapter.deserialize(e)),this._validatorOnChange()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disabled",{get:function(){return!!this._disabled},set:function(e){var t=Object(p.coerceBooleanProperty)(e),n=this._elementRef.nativeElement;this._disabled!==t&&(this._disabled=t,this._disabledChange.emit(t)),t&&n.blur&&n.blur()},enumerable:!0,configurable:!0}),e.prototype.ngOnDestroy=function(){this._datepickerSubscription.unsubscribe(),this._localeSubscription.unsubscribe(),this._valueChange.complete(),this._disabledChange.complete()},e.prototype.registerOnValidatorChange=function(e){this._validatorOnChange=e},e.prototype.validate=function(e){return this._validator?this._validator(e):null},e.prototype.getPopupConnectionElementRef=function(){return this.getConnectedOverlayOrigin()},e.prototype.getConnectedOverlayOrigin=function(){return this._formField?this._formField.getConnectedOverlayOrigin():this._elementRef},e.prototype.writeValue=function(e){this.value=e},e.prototype.registerOnChange=function(e){this._cvaOnChange=e},e.prototype.registerOnTouched=function(e){this._onTouched=e},e.prototype.setDisabledState=function(e){this.disabled=e},e.prototype._onKeydown=function(e){var t=e.altKey&&e.keyCode===a.DOWN_ARROW;this._datepicker&&t&&!this._elementRef.nativeElement.readOnly&&(this._datepicker.open(),e.preventDefault())},e.prototype._onInput=function(e){var t=this._dateAdapter.parse(e,this._dateFormats.parse.dateInput);this._lastValueValid=!t||this._dateAdapter.isValid(t),t=this._getValidDateOrNull(t),this._dateAdapter.sameDate(t,this._value)||(this._value=t,this._cvaOnChange(t),this._valueChange.emit(t),this.dateInput.emit(new z(this,this._elementRef.nativeElement)))},e.prototype._onChange=function(){this.dateChange.emit(new z(this,this._elementRef.nativeElement))},e.prototype._getThemePalette=function(){return this._formField?this._formField.color:void 0},e.prototype._onBlur=function(){this.value&&this._formatValue(this.value),this._onTouched()},e.prototype._formatValue=function(e){this._elementRef.nativeElement.value=e?this._dateAdapter.format(e,this._dateFormats.display.dateInput):""},e.prototype._getValidDateOrNull=function(e){return this._dateAdapter.isDateInstance(e)&&this._dateAdapter.isValid(e)?e:null},e.decorators=[{type:r.Directive,args:[{selector:"input[matDatepicker]",providers:[V,B,{provide:_.MAT_INPUT_VALUE_ACCESSOR,useExisting:e}],host:{"[attr.aria-haspopup]":"true","[attr.aria-owns]":"(_datepicker?.opened && _datepicker.id) || null","[attr.min]":"min ? _dateAdapter.toIso8601(min) : null","[attr.max]":"max ? _dateAdapter.toIso8601(max) : null","[disabled]":"disabled","(input)":"_onInput($event.target.value)","(change)":"_onChange()","(blur)":"_onBlur()","(keydown)":"_onKeydown($event)"},exportAs:"matDatepickerInput"}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:s.DateAdapter,decorators:[{type:r.Optional}]},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[s.MAT_DATE_FORMATS]}]},{type:v.MatFormField,decorators:[{type:r.Optional}]}]},e.propDecorators={matDatepicker:[{type:r.Input}],matDatepickerFilter:[{type:r.Input}],value:[{type:r.Input}],min:[{type:r.Input}],max:[{type:r.Input}],disabled:[{type:r.Input}],dateChange:[{type:r.Output}],dateInput:[{type:r.Output}]},e}(),H=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"[matDatepickerToggleIcon]"}]}],e}(),q=function(){function e(e,t,n){this._intl=e,this._changeDetectorRef=t,this._stateChanges=i.Subscription.EMPTY;var r=Number(n);this.tabIndex=r||0===r?r:null}return Object.defineProperty(e.prototype,"disabled",{get:function(){return void 0===this._disabled?this.datepicker.disabled:!!this._disabled},set:function(e){this._disabled=Object(p.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(e){e.datepicker&&this._watchStateChanges()},e.prototype.ngOnDestroy=function(){this._stateChanges.unsubscribe()},e.prototype.ngAfterContentInit=function(){this._watchStateChanges()},e.prototype._open=function(e){this.datepicker&&!this.disabled&&(this.datepicker.open(),e.stopPropagation())},e.prototype._watchStateChanges=function(){var e=this,t=this.datepicker?this.datepicker._disabledChange:Object(i.of)(),n=this.datepicker&&this.datepicker._datepickerInput?this.datepicker._datepickerInput._disabledChange:Object(i.of)(),r=this.datepicker?Object(i.merge)(this.datepicker.openedStream,this.datepicker.closedStream):Object(i.of)();this._stateChanges.unsubscribe(),this._stateChanges=Object(i.merge)(this._intl.changes,t,n,r).subscribe(function(){return e._changeDetectorRef.markForCheck()})},e.decorators=[{type:r.Component,args:[{selector:"mat-datepicker-toggle",template:'<button #button mat-icon-button type="button" aria-haspopup="true" [attr.aria-label]="_intl.openCalendarLabel" [attr.tabindex]="disabled ? -1 : tabIndex" [disabled]="disabled" [disableRipple]="disableRipple" (click)="_open($event)"><svg *ngIf="!_customIcon" class="mat-datepicker-toggle-default-icon" viewBox="0 0 24 24" width="24px" height="24px" fill="currentColor" focusable="false"><path d="M19 3h-1V1h-2v2H8V1H6v2H5c-1.11 0-1.99.9-1.99 2L3 19c0 1.1.89 2 2 2h14c1.1 0 2-.9 2-2V5c0-1.1-.9-2-2-2zm0 16H5V8h14v11zM7 10h5v5H7z"/></svg><ng-content select="[matDatepickerToggleIcon]"></ng-content></button>',styles:[".mat-form-field-appearance-legacy .mat-form-field-prefix .mat-datepicker-toggle-default-icon,.mat-form-field-appearance-legacy .mat-form-field-suffix .mat-datepicker-toggle-default-icon{width:1em}.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-prefix .mat-datepicker-toggle-default-icon,.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-suffix .mat-datepicker-toggle-default-icon{display:block;width:1.5em;height:1.5em}.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-prefix .mat-icon-button .mat-datepicker-toggle-default-icon,.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-suffix .mat-icon-button .mat-datepicker-toggle-default-icon{margin:auto}"],host:{class:"mat-datepicker-toggle","[attr.tabindex]":"-1","[class.mat-datepicker-toggle-active]":"datepicker && datepicker.opened","[class.mat-accent]":'datepicker && datepicker.color === "accent"',"[class.mat-warn]":'datepicker && datepicker.color === "warn"',"(focus)":"_button.focus()"},exportAs:"matDatepickerToggle",encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:x},{type:r.ChangeDetectorRef},{type:String,decorators:[{type:r.Attribute,args:["tabindex"]}]}]},e.propDecorators={datepicker:[{type:r.Input,args:["for"]}],tabIndex:[{type:r.Input}],disabled:[{type:r.Input}],disableRipple:[{type:r.Input}],_customIcon:[{type:r.ContentChild,args:[H]}],_button:[{type:r.ViewChild,args:["button"]}]},e}(),G=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[h.CommonModule,y.MatButtonModule,m.MatDialogModule,f.OverlayModule,b.A11yModule,c.PortalModule],exports:[M,w,L,F,U,q,H,C,S,E,k],declarations:[M,w,L,F,U,q,H,C,S,E,k],providers:[x,P],entryComponents:[F,k]}]}],e}()},"./node_modules/@angular/material/esm5/dialog.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatDialogModule",function(){return R}),n.d(t,"MAT_DIALOG_SCROLL_STRATEGY_FACTORY",function(){return C}),n.d(t,"MAT_DIALOG_SCROLL_STRATEGY_PROVIDER_FACTORY",function(){return E}),n.d(t,"MAT_DIALOG_DATA",function(){return x}),n.d(t,"MAT_DIALOG_DEFAULT_OPTIONS",function(){return O}),n.d(t,"MAT_DIALOG_SCROLL_STRATEGY",function(){return w}),n.d(t,"MAT_DIALOG_SCROLL_STRATEGY_PROVIDER",function(){return S}),n.d(t,"MatDialog",function(){return k}),n.d(t,"throwMatDialogContentAlreadyAttachedError",function(){return _}),n.d(t,"MatDialogContainer",function(){return y}),n.d(t,"MatDialogClose",function(){return A}),n.d(t,"MatDialogTitle",function(){return T}),n.d(t,"MatDialogContent",function(){return D}),n.d(t,"MatDialogActions",function(){return I}),n.d(t,"MatDialogConfig",function(){return m}),n.d(t,"MatDialogRef",function(){return j}),n.d(t,"matDialogAnimations",function(){return v});var r=n("./node_modules/@angular/animations/fesm5/animations.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/common/fesm5/common.js"),s=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),u=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),c=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),l=n("./node_modules/rxjs/_esm5/index.js"),d=n("./node_modules/rxjs/_esm5/operators/index.js"),p=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),f=n("./node_modules/@angular/cdk/esm5/overlay.es5.js"),h=n("./node_modules/@angular/material/esm5/core.es5.js"),m=function(){return function(){this.role="dialog",this.panelClass="",this.hasBackdrop=!0,this.backdropClass="",this.disableClose=!1,this.width="",this.height="",this.maxWidth="80vw",this.data=null,this.ariaDescribedBy=null,this.ariaLabel=null,this.autoFocus=!0,this.restoreFocus=!0,this.closeOnNavigation=!0}}(),g=[Object(r.state)("void, exit",Object(r.style)({opacity:0,transform:"scale(0.7)"})),Object(r.state)("enter",Object(r.style)({transform:"none"})),Object(r.transition)("* => enter",Object(r.animate)("150ms cubic-bezier(0, 0, 0.2, 1)",Object(r.style)({transform:"none",opacity:1}))),Object(r.transition)("* => void, * => exit",Object(r.animate)("75ms cubic-bezier(0.4, 0.0, 0.2, 1)",Object(r.style)({opacity:0})))],v={dialogContainer:Object(r.trigger)("dialogContainer",g),slideDialog:Object(r.trigger)("slideDialog",g)};function _(){throw Error("Attempting to attach dialog content after content is already attached")}var y=function(e){function t(t,n,r,i,a){var s=e.call(this)||this;return s._elementRef=t,s._focusTrapFactory=n,s._changeDetectorRef=r,s._document=i,s._config=a,s._elementFocusedBeforeDialogWasOpened=null,s._state="enter",s._animationStateChanged=new o.EventEmitter,s._ariaLabelledBy=null,s}return Object(i.__extends)(t,e),t.prototype.attachComponentPortal=function(e){return this._portalOutlet.hasAttached()&&_(),this._savePreviouslyFocusedElement(),this._portalOutlet.attachComponentPortal(e)},t.prototype.attachTemplatePortal=function(e){return this._portalOutlet.hasAttached()&&_(),this._savePreviouslyFocusedElement(),this._portalOutlet.attachTemplatePortal(e)},t.prototype._trapFocus=function(){this._focusTrap||(this._focusTrap=this._focusTrapFactory.create(this._elementRef.nativeElement)),this._config.autoFocus&&this._focusTrap.focusInitialElementWhenReady()},t.prototype._restoreFocus=function(){var e=this._elementFocusedBeforeDialogWasOpened;this._config.restoreFocus&&e&&"function"==typeof e.focus&&e.focus(),this._focusTrap&&this._focusTrap.destroy()},t.prototype._savePreviouslyFocusedElement=function(){var e=this;this._document&&(this._elementFocusedBeforeDialogWasOpened=this._document.activeElement,this._elementRef.nativeElement.focus&&Promise.resolve().then(function(){return e._elementRef.nativeElement.focus()}))},t.prototype._onAnimationDone=function(e){"enter"===e.toState?this._trapFocus():"exit"===e.toState&&this._restoreFocus(),this._animationStateChanged.emit(e)},t.prototype._onAnimationStart=function(e){this._animationStateChanged.emit(e)},t.prototype._startExitAnimation=function(){this._state="exit",this._changeDetectorRef.markForCheck()},t.decorators=[{type:o.Component,args:[{selector:"mat-dialog-container",template:"<ng-template cdkPortalOutlet></ng-template>",styles:[".mat-dialog-container{display:block;padding:24px;border-radius:4px;box-sizing:border-box;overflow:auto;outline:0;width:100%;height:100%;min-height:inherit;max-height:inherit}@media (-ms-high-contrast:active){.mat-dialog-container{outline:solid 1px}}.mat-dialog-content{display:block;margin:0 -24px;padding:0 24px;max-height:65vh;overflow:auto;-webkit-overflow-scrolling:touch}.mat-dialog-title{margin:0 0 20px;display:block}.mat-dialog-actions{padding:8px 0;display:flex;flex-wrap:wrap;min-height:52px;align-items:center;margin-bottom:-24px}.mat-dialog-actions[align=end]{justify-content:flex-end}.mat-dialog-actions[align=center]{justify-content:center}.mat-dialog-actions .mat-button+.mat-button,.mat-dialog-actions .mat-button+.mat-raised-button,.mat-dialog-actions .mat-raised-button+.mat-button,.mat-dialog-actions .mat-raised-button+.mat-raised-button{margin-left:8px}[dir=rtl] .mat-dialog-actions .mat-button+.mat-button,[dir=rtl] .mat-dialog-actions .mat-button+.mat-raised-button,[dir=rtl] .mat-dialog-actions .mat-raised-button+.mat-button,[dir=rtl] .mat-dialog-actions .mat-raised-button+.mat-raised-button{margin-left:0;margin-right:8px}"],encapsulation:o.ViewEncapsulation.None,changeDetection:o.ChangeDetectionStrategy.Default,animations:[v.dialogContainer],host:{class:"mat-dialog-container",tabindex:"-1","aria-modal":"true","[attr.id]":"_id","[attr.role]":"_config.role","[attr.aria-labelledby]":"_config.ariaLabel ? null : _ariaLabelledBy","[attr.aria-label]":"_config.ariaLabel","[attr.aria-describedby]":"_config.ariaDescribedBy || null","[@dialogContainer]":"_state","(@dialogContainer.start)":"_onAnimationStart($event)","(@dialogContainer.done)":"_onAnimationDone($event)"}}]}],t.ctorParameters=function(){return[{type:o.ElementRef},{type:u.FocusTrapFactory},{type:o.ChangeDetectorRef},{type:void 0,decorators:[{type:o.Optional},{type:o.Inject,args:[a.DOCUMENT]}]},{type:m}]},t.propDecorators={_portalOutlet:[{type:o.ViewChild,args:[s.CdkPortalOutlet]}]},t}(s.BasePortalOutlet),b=0,j=function(){function e(e,t,n,r){void 0===r&&(r="mat-dialog-"+b++);var i=this;this._overlayRef=e,this._containerInstance=t,this.id=r,this.disableClose=this._containerInstance._config.disableClose,this._afterOpened=new l.Subject,this._afterClosed=new l.Subject,this._beforeClosed=new l.Subject,t._id=r,t._animationStateChanged.pipe(Object(d.filter)(function(e){return"done"===e.phaseName&&"enter"===e.toState}),Object(d.take)(1)).subscribe(function(){i._afterOpened.next(),i._afterOpened.complete()}),t._animationStateChanged.pipe(Object(d.filter)(function(e){return"done"===e.phaseName&&"exit"===e.toState}),Object(d.take)(1)).subscribe(function(){return i._overlayRef.dispose()}),e.detachments().subscribe(function(){i._beforeClosed.next(i._result),i._beforeClosed.complete(),i._afterClosed.next(i._result),i._afterClosed.complete(),i.componentInstance=null,i._overlayRef.dispose()}),e.keydownEvents().pipe(Object(d.filter)(function(e){return e.keyCode===c.ESCAPE&&!i.disableClose})).subscribe(function(){return i.close()})}return e.prototype.close=function(e){var t=this;this._result=e,this._containerInstance._animationStateChanged.pipe(Object(d.filter)(function(e){return"start"===e.phaseName}),Object(d.take)(1)).subscribe(function(){t._beforeClosed.next(e),t._beforeClosed.complete(),t._overlayRef.detachBackdrop()}),this._containerInstance._startExitAnimation()},e.prototype.afterOpened=function(){return this._afterOpened.asObservable()},e.prototype.afterClosed=function(){return this._afterClosed.asObservable()},e.prototype.beforeClosed=function(){return this._beforeClosed.asObservable()},e.prototype.backdropClick=function(){return this._overlayRef.backdropClick()},e.prototype.keydownEvents=function(){return this._overlayRef.keydownEvents()},e.prototype.updatePosition=function(e){var t=this._getPositionStrategy();return e&&(e.left||e.right)?e.left?t.left(e.left):t.right(e.right):t.centerHorizontally(),e&&(e.top||e.bottom)?e.top?t.top(e.top):t.bottom(e.bottom):t.centerVertically(),this._overlayRef.updatePosition(),this},e.prototype.updateSize=function(e,t){return void 0===e&&(e=""),void 0===t&&(t=""),this._getPositionStrategy().width(e).height(t),this._overlayRef.updatePosition(),this},e.prototype.afterOpen=function(){return this.afterOpened()},e.prototype.beforeClose=function(){return this.beforeClosed()},e.prototype._getPositionStrategy=function(){return this._overlayRef.getConfig().positionStrategy},e}(),x=new o.InjectionToken("MatDialogData"),O=new o.InjectionToken("mat-dialog-default-options"),w=new o.InjectionToken("mat-dialog-scroll-strategy");function C(e){return function(){return e.scrollStrategies.block()}}function E(e){return function(){return e.scrollStrategies.block()}}var S={provide:w,deps:[f.Overlay],useFactory:E},k=function(){function e(e,t,n,r,i,o,a){var s=this;this._overlay=e,this._injector=t,this._location=n,this._defaultOptions=r,this._parentDialog=o,this._overlayContainer=a,this._openDialogsAtThisLevel=[],this._afterAllClosedAtThisLevel=new l.Subject,this._afterOpenedAtThisLevel=new l.Subject,this._ariaHiddenElements=new Map,this.afterAllClosed=Object(l.defer)(function(){return s.openDialogs.length?s._afterAllClosed:s._afterAllClosed.pipe(Object(d.startWith)(void 0))}),this._scrollStrategy=i}return Object.defineProperty(e.prototype,"openDialogs",{get:function(){return this._parentDialog?this._parentDialog.openDialogs:this._openDialogsAtThisLevel},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"afterOpened",{get:function(){return this._parentDialog?this._parentDialog.afterOpened:this._afterOpenedAtThisLevel},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"afterOpen",{get:function(){return this.afterOpened},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_afterAllClosed",{get:function(){var e=this._parentDialog;return e?e._afterAllClosed:this._afterAllClosedAtThisLevel},enumerable:!0,configurable:!0}),e.prototype.open=function(e,t){var n,r,o=this;if(n=t,r=this._defaultOptions||new m,(t=Object(i.__assign)({},r,n)).id&&this.getDialogById(t.id))throw Error('Dialog with id "'+t.id+'" exists already. The dialog id must be unique.');var a=this._createOverlay(t),s=this._attachDialogContainer(a,t),u=this._attachDialogContent(e,s,a,t);return this.openDialogs.length||this._hideNonDialogContentFromAssistiveTechnology(),this.openDialogs.push(u),u.afterClosed().subscribe(function(){return o._removeOpenDialog(u)}),this.afterOpened.next(u),u},e.prototype.closeAll=function(){this._closeDialogs(this.openDialogs)},e.prototype.getDialogById=function(e){return this.openDialogs.find(function(t){return t.id===e})},e.prototype.ngOnDestroy=function(){this._closeDialogs(this._openDialogsAtThisLevel),this._afterAllClosedAtThisLevel.complete(),this._afterOpenedAtThisLevel.complete()},e.prototype._createOverlay=function(e){var t=this._getOverlayConfig(e);return this._overlay.create(t)},e.prototype._getOverlayConfig=function(e){var t=new f.OverlayConfig({positionStrategy:this._overlay.position().global(),scrollStrategy:e.scrollStrategy||this._scrollStrategy(),panelClass:e.panelClass,hasBackdrop:e.hasBackdrop,direction:e.direction,minWidth:e.minWidth,minHeight:e.minHeight,maxWidth:e.maxWidth,maxHeight:e.maxHeight,disposeOnNavigation:e.closeOnNavigation});return e.backdropClass&&(t.backdropClass=e.backdropClass),t},e.prototype._attachDialogContainer=function(e,t){var n=t&&t.viewContainerRef&&t.viewContainerRef.injector,r=new s.PortalInjector(n||this._injector,new WeakMap([[m,t]])),i=new s.ComponentPortal(y,t.viewContainerRef,r);return e.attach(i).instance},e.prototype._attachDialogContent=function(e,t,n,r){var i=new j(n,t,this._location,r.id);if(r.hasBackdrop&&n.backdropClick().subscribe(function(){i.disableClose||i.close()}),e instanceof o.TemplateRef)t.attachTemplatePortal(new s.TemplatePortal(e,null,{$implicit:r.data,dialogRef:i}));else{var a=this._createInjector(r,i,t),u=t.attachComponentPortal(new s.ComponentPortal(e,void 0,a));i.componentInstance=u.instance}return i.updateSize(r.width,r.height).updatePosition(r.position),i},e.prototype._createInjector=function(e,t,n){var r=e&&e.viewContainerRef&&e.viewContainerRef.injector,i=new WeakMap([[y,n],[x,e.data],[j,t]]);return!e.direction||r&&r.get(p.Directionality,null)||i.set(p.Directionality,{value:e.direction,change:Object(l.of)()}),new s.PortalInjector(r||this._injector,i)},e.prototype._removeOpenDialog=function(e){var t=this.openDialogs.indexOf(e);t>-1&&(this.openDialogs.splice(t,1),this.openDialogs.length||(this._ariaHiddenElements.forEach(function(e,t){e?t.setAttribute("aria-hidden",e):t.removeAttribute("aria-hidden")}),this._ariaHiddenElements.clear(),this._afterAllClosed.next()))},e.prototype._hideNonDialogContentFromAssistiveTechnology=function(){var e=this._overlayContainer.getContainerElement();if(e.parentElement)for(var t=e.parentElement.children,n=t.length-1;n>-1;n--){var r=t[n];r===e||"SCRIPT"===r.nodeName||"STYLE"===r.nodeName||r.hasAttribute("aria-live")||(this._ariaHiddenElements.set(r,r.getAttribute("aria-hidden")),r.setAttribute("aria-hidden","true"))}},e.prototype._closeDialogs=function(e){for(var t=e.length;t--;)e[t].close()},e.decorators=[{type:o.Injectable}],e.ctorParameters=function(){return[{type:f.Overlay},{type:o.Injector},{type:a.Location,decorators:[{type:o.Optional}]},{type:m,decorators:[{type:o.Optional},{type:o.Inject,args:[O]}]},{type:void 0,decorators:[{type:o.Inject,args:[w]}]},{type:e,decorators:[{type:o.Optional},{type:o.SkipSelf}]},{type:f.OverlayContainer}]},e}();var M=0,A=function(){function e(e,t,n){this.dialogRef=e,this._elementRef=t,this._dialog=n,this.ariaLabel="Close dialog"}return e.prototype.ngOnInit=function(){this.dialogRef||(this.dialogRef=P(this._elementRef,this._dialog.openDialogs))},e.prototype.ngOnChanges=function(e){var t=e._matDialogClose||e._matDialogCloseResult;t&&(this.dialogResult=t.currentValue)},e.decorators=[{type:o.Directive,args:[{selector:"button[mat-dialog-close], button[matDialogClose]",exportAs:"matDialogClose",host:{"(click)":"dialogRef.close(dialogResult)","[attr.aria-label]":"ariaLabel",type:"button"}}]}],e.ctorParameters=function(){return[{type:j,decorators:[{type:o.Optional}]},{type:o.ElementRef},{type:k}]},e.propDecorators={ariaLabel:[{type:o.Input,args:["aria-label"]}],dialogResult:[{type:o.Input,args:["mat-dialog-close"]}],_matDialogClose:[{type:o.Input,args:["matDialogClose"]}]},e}(),T=function(){function e(e,t,n){this._dialogRef=e,this._elementRef=t,this._dialog=n,this.id="mat-dialog-title-"+M++}return e.prototype.ngOnInit=function(){var e=this;this._dialogRef||(this._dialogRef=P(this._elementRef,this._dialog.openDialogs)),this._dialogRef&&Promise.resolve().then(function(){var t=e._dialogRef._containerInstance;t&&!t._ariaLabelledBy&&(t._ariaLabelledBy=e.id)})},e.decorators=[{type:o.Directive,args:[{selector:"[mat-dialog-title], [matDialogTitle]",exportAs:"matDialogTitle",host:{class:"mat-dialog-title","[id]":"id"}}]}],e.ctorParameters=function(){return[{type:j,decorators:[{type:o.Optional}]},{type:o.ElementRef},{type:k}]},e.propDecorators={id:[{type:o.Input}]},e}(),D=function(){function e(){}return e.decorators=[{type:o.Directive,args:[{selector:"[mat-dialog-content], mat-dialog-content, [matDialogContent]",host:{class:"mat-dialog-content"}}]}],e}(),I=function(){function e(){}return e.decorators=[{type:o.Directive,args:[{selector:"[mat-dialog-actions], mat-dialog-actions, [matDialogActions]",host:{class:"mat-dialog-actions"}}]}],e}();function P(e,t){for(var n=e.nativeElement.parentElement;n&&!n.classList.contains("mat-dialog-container");)n=n.parentElement;return n?t.find(function(e){return e.id===n.id}):null}var R=function(){function e(){}return e.decorators=[{type:o.NgModule,args:[{imports:[a.CommonModule,f.OverlayModule,s.PortalModule,h.MatCommonModule],exports:[y,A,T,D,I,h.MatCommonModule],declarations:[y,A,T,I,D],providers:[k,S],entryComponents:[y]}]}],e}()},"./node_modules/@angular/material/esm5/divider.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatDivider",function(){return s}),n.d(t,"MatDividerModule",function(){return u});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),o=n("./node_modules/@angular/common/fesm5/common.js"),a=n("./node_modules/@angular/material/esm5/core.es5.js"),s=function(){function e(){this._vertical=!1,this._inset=!1}return Object.defineProperty(e.prototype,"vertical",{get:function(){return this._vertical},set:function(e){this._vertical=Object(i.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"inset",{get:function(){return this._inset},set:function(e){this._inset=Object(i.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),e.decorators=[{type:r.Component,args:[{selector:"mat-divider",host:{role:"separator","[attr.aria-orientation]":'vertical ? "vertical" : "horizontal"',"[class.mat-divider-vertical]":"vertical","[class.mat-divider-horizontal]":"!vertical","[class.mat-divider-inset]":"inset",class:"mat-divider"},template:"",styles:[".mat-divider{display:block;margin:0;border-top-width:1px;border-top-style:solid}.mat-divider.mat-divider-vertical{border-top:0;border-right-width:1px;border-right-style:solid}.mat-divider.mat-divider-inset{margin-left:80px}[dir=rtl] .mat-divider.mat-divider-inset{margin-left:auto;margin-right:80px}"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],e.propDecorators={vertical:[{type:r.Input}],inset:[{type:r.Input}]},e}(),u=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[a.MatCommonModule,o.CommonModule],exports:[s,a.MatCommonModule],declarations:[s]}]}],e}()},"./node_modules/@angular/material/esm5/expansion.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatExpansionModule",function(){return k}),n.d(t,"MatAccordion",function(){return S}),n.d(t,"MAT_ACCORDION",function(){return g}),n.d(t,"MAT_EXPANSION_PANEL_DEFAULT_OPTIONS",function(){return j}),n.d(t,"MatExpansionPanel",function(){return x}),n.d(t,"MatExpansionPanelActionRow",function(){return O}),n.d(t,"MatExpansionPanelHeader",function(){return w}),n.d(t,"MatExpansionPanelDescription",function(){return C}),n.d(t,"MatExpansionPanelTitle",function(){return E}),n.d(t,"MatExpansionPanelContent",function(){return y}),n.d(t,"EXPANSION_PANEL_ANIMATION_TIMING",function(){return v}),n.d(t,"matExpansionAnimations",function(){return _});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/animations/fesm5/animations.js"),o=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/cdk/esm5/accordion.es5.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),c=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),l=n("./node_modules/@angular/common/fesm5/common.js"),d=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),p=n("./node_modules/rxjs/_esm5/index.js"),f=n("./node_modules/rxjs/_esm5/operators/index.js"),h=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),m=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),g=new r.InjectionToken("MAT_ACCORDION"),v="225ms cubic-bezier(0.4,0.0,0.2,1)",_={indicatorRotate:Object(i.trigger)("indicatorRotate",[Object(i.state)("collapsed, void",Object(i.style)({transform:"rotate(0deg)"})),Object(i.state)("expanded",Object(i.style)({transform:"rotate(180deg)"})),Object(i.transition)("expanded <=> collapsed, void => collapsed",Object(i.animate)(v))]),expansionHeaderHeight:Object(i.trigger)("expansionHeight",[Object(i.state)("collapsed, void",Object(i.style)({height:"{{collapsedHeight}}"}),{params:{collapsedHeight:"48px"}}),Object(i.state)("expanded",Object(i.style)({height:"{{expandedHeight}}"}),{params:{expandedHeight:"64px"}}),Object(i.transition)("expanded <=> collapsed, void => collapsed",Object(i.group)([Object(i.query)("@indicatorRotate",Object(i.animateChild)(),{optional:!0}),Object(i.animate)(v)]))]),bodyExpansion:Object(i.trigger)("bodyExpansion",[Object(i.state)("collapsed, void",Object(i.style)({height:"0px",visibility:"hidden"})),Object(i.state)("expanded",Object(i.style)({height:"*",visibility:"visible"})),Object(i.transition)("expanded <=> collapsed, void => collapsed",Object(i.animate)(v))])},y=function(){function e(e){this._template=e}return e.decorators=[{type:r.Directive,args:[{selector:"ng-template[matExpansionPanelContent]"}]}],e.ctorParameters=function(){return[{type:r.TemplateRef}]},e}(),b=0,j=new r.InjectionToken("MAT_EXPANSION_PANEL_DEFAULT_OPTIONS"),x=function(e){function t(t,n,i,o,a,s,u){var c=e.call(this,t,n,i)||this;return c._viewContainerRef=o,c._animationMode=s,c._hideToggle=!1,c.afterExpand=new r.EventEmitter,c.afterCollapse=new r.EventEmitter,c._inputChanges=new p.Subject,c._headerId="mat-expansion-panel-header-"+b++,c._bodyAnimationDone=new p.Subject,c.accordion=t,c._document=a,c._bodyAnimationDone.pipe(Object(f.distinctUntilChanged)(function(e,t){return e.fromState===t.fromState&&e.toState===t.toState})).subscribe(function(e){"void"!==e.fromState&&("expanded"===e.toState?c.afterExpand.emit():"collapsed"===e.toState&&c.afterCollapse.emit())}),u&&(c.hideToggle=u.hideToggle),c}return Object(o.__extends)(t,e),Object.defineProperty(t.prototype,"hideToggle",{get:function(){return this._hideToggle||this.accordion&&this.accordion.hideToggle},set:function(e){this._hideToggle=Object(s.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t.prototype._hasSpacing=function(){return!!this.accordion&&"default"===(this.expanded?this.accordion.displayMode:this._getExpandedState())},t.prototype._getExpandedState=function(){return this.expanded?"expanded":"collapsed"},t.prototype.ngAfterContentInit=function(){var e=this;this._lazyContent&&this.opened.pipe(Object(f.startWith)(null),Object(f.filter)(function(){return e.expanded&&!e._portal}),Object(f.take)(1)).subscribe(function(){e._portal=new c.TemplatePortal(e._lazyContent._template,e._viewContainerRef)})},t.prototype.ngOnChanges=function(e){this._inputChanges.next(e)},t.prototype.ngOnDestroy=function(){e.prototype.ngOnDestroy.call(this),this._bodyAnimationDone.complete(),this._inputChanges.complete()},t.prototype._containsFocus=function(){if(this._body&&this._document){var e=this._document.activeElement,t=this._body.nativeElement;return e===t||t.contains(e)}return!1},t.decorators=[{type:r.Component,args:[{styles:[".mat-expansion-panel{box-sizing:content-box;display:block;margin:0;border-radius:4px;overflow:hidden;transition:margin 225ms cubic-bezier(.4,0,.2,1),box-shadow 280ms cubic-bezier(.4,0,.2,1)}.mat-accordion .mat-expansion-panel:not(.mat-expanded),.mat-accordion .mat-expansion-panel:not(.mat-expansion-panel-spacing){border-radius:0}.mat-accordion .mat-expansion-panel:first-of-type{border-top-right-radius:4px;border-top-left-radius:4px}.mat-accordion .mat-expansion-panel:last-of-type{border-bottom-right-radius:4px;border-bottom-left-radius:4px}@media (-ms-high-contrast:active){.mat-expansion-panel{outline:solid 1px}}.mat-expansion-panel._mat-animation-noopable,.mat-expansion-panel.ng-animate-disabled,.ng-animate-disabled .mat-expansion-panel{transition:none}.mat-expansion-panel-content{display:flex;flex-direction:column;overflow:visible}.mat-expansion-panel-body{padding:0 24px 16px}.mat-expansion-panel-spacing{margin:16px 0}.mat-accordion>.mat-expansion-panel-spacing:first-child,.mat-accordion>:first-child:not(.mat-expansion-panel) .mat-expansion-panel-spacing{margin-top:0}.mat-accordion>.mat-expansion-panel-spacing:last-child,.mat-accordion>:last-child:not(.mat-expansion-panel) .mat-expansion-panel-spacing{margin-bottom:0}.mat-action-row{border-top-style:solid;border-top-width:1px;display:flex;flex-direction:row;justify-content:flex-end;padding:16px 8px 16px 24px}.mat-action-row button.mat-button{margin-left:8px}[dir=rtl] .mat-action-row button.mat-button{margin-left:0;margin-right:8px}"],selector:"mat-expansion-panel",exportAs:"matExpansionPanel",template:'<ng-content select="mat-expansion-panel-header"></ng-content><div class="mat-expansion-panel-content" role="region" [@bodyExpansion]="_getExpandedState()" (@bodyExpansion.done)="_bodyAnimationDone.next($event)" [attr.aria-labelledby]="_headerId" [id]="id" #body><div class="mat-expansion-panel-body"><ng-content></ng-content><ng-template [cdkPortalOutlet]="_portal"></ng-template></div><ng-content select="mat-action-row"></ng-content></div>',encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,inputs:["disabled","expanded"],outputs:["opened","closed","expandedChange"],animations:[_.bodyExpansion],providers:[{provide:g,useValue:void 0}],host:{class:"mat-expansion-panel","[class.mat-expanded]":"expanded","[class._mat-animation-noopable]":'_animationMode === "NoopAnimations"',"[class.mat-expansion-panel-spacing]":"_hasSpacing()"}}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.Optional},{type:r.SkipSelf},{type:r.Inject,args:[g]}]},{type:r.ChangeDetectorRef},{type:u.UniqueSelectionDispatcher},{type:r.ViewContainerRef},{type:void 0,decorators:[{type:r.Inject,args:[l.DOCUMENT]}]},{type:String,decorators:[{type:r.Optional},{type:r.Inject,args:[d.ANIMATION_MODULE_TYPE]}]},{type:void 0,decorators:[{type:r.Inject,args:[j]},{type:r.Optional}]}]},t.propDecorators={hideToggle:[{type:r.Input}],afterExpand:[{type:r.Output}],afterCollapse:[{type:r.Output}],_lazyContent:[{type:r.ContentChild,args:[y]}],_body:[{type:r.ViewChild,args:["body"]}]},t}(a.CdkAccordionItem),O=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"mat-action-row",host:{class:"mat-action-row"}}]}],e}(),w=function(){function e(e,t,n,r,i){var o=this;this.panel=e,this._element=t,this._focusMonitor=n,this._changeDetectorRef=r,this._parentChangeSubscription=p.Subscription.EMPTY;var a=e.accordion?e.accordion._stateChanges.pipe(Object(f.filter)(function(e){return!!e.hideToggle})):p.EMPTY;this._parentChangeSubscription=Object(p.merge)(e.opened,e.closed,a,e._inputChanges.pipe(Object(f.filter)(function(e){return!(!e.hideToggle&&!e.disabled)}))).subscribe(function(){return o._changeDetectorRef.markForCheck()}),e.closed.pipe(Object(f.filter)(function(){return e._containsFocus()})).subscribe(function(){return n.focusVia(t,"program")}),n.monitor(t).subscribe(function(t){t&&e.accordion&&e.accordion._handleHeaderFocus(o)}),i&&(this.expandedHeight=i.expandedHeight,this.collapsedHeight=i.collapsedHeight)}return Object.defineProperty(e.prototype,"disabled",{get:function(){return this.panel.disabled},enumerable:!0,configurable:!0}),e.prototype._toggle=function(){this.panel.toggle()},e.prototype._isExpanded=function(){return this.panel.expanded},e.prototype._getExpandedState=function(){return this.panel._getExpandedState()},e.prototype._getPanelId=function(){return this.panel.id},e.prototype._showToggle=function(){return!this.panel.hideToggle&&!this.panel.disabled},e.prototype._keydown=function(e){switch(e.keyCode){case m.SPACE:case m.ENTER:Object(m.hasModifierKey)(e)||(e.preventDefault(),this._toggle());break;default:return void(this.panel.accordion&&this.panel.accordion._handleHeaderKeydown(e))}},e.prototype.focus=function(e){void 0===e&&(e="program"),this._focusMonitor.focusVia(this._element,e)},e.prototype.ngOnDestroy=function(){this._parentChangeSubscription.unsubscribe(),this._focusMonitor.stopMonitoring(this._element)},e.decorators=[{type:r.Component,args:[{selector:"mat-expansion-panel-header",styles:[".mat-expansion-panel-header{display:flex;flex-direction:row;align-items:center;padding:0 24px;border-radius:inherit}.mat-expansion-panel-header:focus,.mat-expansion-panel-header:hover{outline:0}.mat-expansion-panel-header.mat-expanded:focus,.mat-expansion-panel-header.mat-expanded:hover{background:inherit}.mat-expansion-panel-header:not([aria-disabled=true]){cursor:pointer}.mat-content{display:flex;flex:1;flex-direction:row;overflow:hidden}.mat-expansion-panel-header-description,.mat-expansion-panel-header-title{display:flex;flex-grow:1;margin-right:16px}[dir=rtl] .mat-expansion-panel-header-description,[dir=rtl] .mat-expansion-panel-header-title{margin-right:0;margin-left:16px}.mat-expansion-panel-header-description{flex-grow:2}.mat-expansion-indicator::after{border-style:solid;border-width:0 2px 2px 0;content:'';display:inline-block;padding:3px;transform:rotate(45deg);vertical-align:middle}"],template:'<span class="mat-content"><ng-content select="mat-panel-title"></ng-content><ng-content select="mat-panel-description"></ng-content><ng-content></ng-content></span><span [@indicatorRotate]="_getExpandedState()" *ngIf="_showToggle()" class="mat-expansion-indicator"></span>',encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,animations:[_.indicatorRotate,_.expansionHeaderHeight],host:{class:"mat-expansion-panel-header",role:"button","[attr.id]":"panel._headerId","[attr.tabindex]":"disabled ? -1 : 0","[attr.aria-controls]":"_getPanelId()","[attr.aria-expanded]":"_isExpanded()","[attr.aria-disabled]":"panel.disabled","[class.mat-expanded]":"_isExpanded()","(click)":"_toggle()","(keydown)":"_keydown($event)","[@expansionHeight]":"{\n        value: _getExpandedState(),\n        params: {\n          collapsedHeight: collapsedHeight,\n          expandedHeight: expandedHeight\n        }\n    }"}}]}],e.ctorParameters=function(){return[{type:x,decorators:[{type:r.Host}]},{type:r.ElementRef},{type:h.FocusMonitor},{type:r.ChangeDetectorRef},{type:void 0,decorators:[{type:r.Inject,args:[j]},{type:r.Optional}]}]},e.propDecorators={expandedHeight:[{type:r.Input}],collapsedHeight:[{type:r.Input}]},e}(),C=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"mat-panel-description",host:{class:"mat-expansion-panel-header-description"}}]}],e}(),E=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"mat-panel-title",host:{class:"mat-expansion-panel-header-title"}}]}],e}(),S=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t._hideToggle=!1,t.displayMode="default",t}return Object(o.__extends)(t,e),Object.defineProperty(t.prototype,"hideToggle",{get:function(){return this._hideToggle},set:function(e){this._hideToggle=Object(s.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){this._keyManager=new h.FocusKeyManager(this._headers).withWrap()},t.prototype._handleHeaderKeydown=function(e){var t=e.keyCode,n=this._keyManager;t===m.HOME?(n.setFirstItemActive(),e.preventDefault()):t===m.END?(n.setLastItemActive(),e.preventDefault()):this._keyManager.onKeydown(e)},t.prototype._handleHeaderFocus=function(e){this._keyManager.updateActiveItem(e)},t.decorators=[{type:r.Directive,args:[{selector:"mat-accordion",exportAs:"matAccordion",inputs:["multi"],providers:[{provide:g,useExisting:t}],host:{class:"mat-accordion"}}]}],t.propDecorators={_headers:[{type:r.ContentChildren,args:[w,{descendants:!0}]}],hideToggle:[{type:r.Input}],displayMode:[{type:r.Input}]},t}(a.CdkAccordion),k=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[l.CommonModule,a.CdkAccordionModule,c.PortalModule],exports:[S,x,O,w,E,C,y],declarations:[S,x,O,w,E,C,y]}]}],e}()},"./node_modules/@angular/material/esm5/form-field.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatFormFieldModule",function(){return I}),n.d(t,"MatError",function(){return g}),n.d(t,"MatFormFieldBase",function(){return M}),n.d(t,"_MatFormFieldMixinBase",function(){return A}),n.d(t,"MAT_FORM_FIELD_DEFAULT_OPTIONS",function(){return T}),n.d(t,"MatFormField",function(){return D}),n.d(t,"MatFormFieldControl",function(){return _}),n.d(t,"getMatFormFieldPlaceholderConflictError",function(){return y}),n.d(t,"getMatFormFieldDuplicatedHintError",function(){return b}),n.d(t,"getMatFormFieldMissingControlError",function(){return j}),n.d(t,"MatHint",function(){return O}),n.d(t,"MatPlaceholder",function(){return C}),n.d(t,"MatPrefix",function(){return E}),n.d(t,"MatSuffix",function(){return S}),n.d(t,"MatLabel",function(){return w}),n.d(t,"matFormFieldAnimations",function(){return v});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/animations/fesm5/animations.js"),o=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/@angular/material/esm5/core.es5.js"),c=n("./node_modules/rxjs/_esm5/index.js"),l=n("./node_modules/rxjs/_esm5/operators/index.js"),d=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),p=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),f=n("./node_modules/@angular/common/fesm5/common.js"),h=n("./node_modules/@angular/cdk/esm5/observers.es5.js"),m=0,g=function(){function e(){this.id="mat-error-"+m++}return e.decorators=[{type:r.Directive,args:[{selector:"mat-error",host:{class:"mat-error",role:"alert","[attr.id]":"id"}}]}],e.propDecorators={id:[{type:r.Input}]},e}(),v={transitionMessages:Object(i.trigger)("transitionMessages",[Object(i.state)("enter",Object(i.style)({opacity:1,transform:"translateY(0%)"})),Object(i.transition)("void => enter",[Object(i.style)({opacity:0,transform:"translateY(-100%)"}),Object(i.animate)("300ms cubic-bezier(0.55, 0, 0.55, 0.2)")])])},_=function(){return function(){}}();function y(){return Error("Placeholder attribute and child element were both specified.")}function b(e){return Error("A hint was already declared for 'align=\""+e+"\"'.")}function j(){return Error("mat-form-field must contain a MatFormFieldControl.")}var x=0,O=function(){function e(){this.align="start",this.id="mat-hint-"+x++}return e.decorators=[{type:r.Directive,args:[{selector:"mat-hint",host:{class:"mat-hint","[class.mat-right]":'align == "end"',"[attr.id]":"id","[attr.align]":"null"}}]}],e.propDecorators={align:[{type:r.Input}],id:[{type:r.Input}]},e}(),w=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"mat-label"}]}],e}(),C=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"mat-placeholder"}]}],e}(),E=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"[matPrefix]"}]}],e}(),S=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"[matSuffix]"}]}],e}(),k=0,M=function(){return function(e){this._elementRef=e}}(),A=Object(u.mixinColor)(M,"primary"),T=new r.InjectionToken("MAT_FORM_FIELD_DEFAULT_OPTIONS"),D=function(e){function t(t,n,r,i,o,a,s,u){var l=e.call(this,t)||this;return l._elementRef=t,l._changeDetectorRef=n,l._dir=i,l._defaults=o,l._platform=a,l._ngZone=s,l._outlineGapCalculationNeededImmediately=!1,l._outlineGapCalculationNeededOnStable=!1,l._destroyed=new c.Subject,l._showAlwaysAnimate=!1,l._subscriptAnimationState="",l._hintLabel="",l._hintLabelId="mat-hint-"+k++,l._labelId="mat-form-field-label-"+k++,l._labelOptions=r||{},l.floatLabel=l._labelOptions.float||"auto",l._animationsEnabled="NoopAnimations"!==u,l.appearance=o&&o.appearance?o.appearance:"legacy",l}return Object(o.__extends)(t,e),Object.defineProperty(t.prototype,"appearance",{get:function(){return this._appearance},set:function(e){var t=this._appearance;this._appearance=e||this._defaults&&this._defaults.appearance||"legacy","outline"===this._appearance&&t!==e&&this._updateOutlineGapOnStable()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"hideRequiredMarker",{get:function(){return this._hideRequiredMarker},set:function(e){this._hideRequiredMarker=Object(s.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_shouldAlwaysFloat",{get:function(){return"always"===this.floatLabel&&!this._showAlwaysAnimate},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_canLabelFloat",{get:function(){return"never"!==this.floatLabel},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"hintLabel",{get:function(){return this._hintLabel},set:function(e){this._hintLabel=e,this._processHints()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"floatLabel",{get:function(){return"legacy"!==this.appearance&&"never"===this._floatLabel?"auto":this._floatLabel},set:function(e){e!==this._floatLabel&&(this._floatLabel=e||this._labelOptions.float||"auto",this._changeDetectorRef.markForCheck())},enumerable:!0,configurable:!0}),t.prototype.getConnectedOverlayOrigin=function(){return this._connectionContainerRef||this._elementRef},t.prototype.ngAfterContentInit=function(){var e=this;this._validateControlChild();var t=this._control;t.controlType&&this._elementRef.nativeElement.classList.add("mat-form-field-type-"+t.controlType),t.stateChanges.pipe(Object(l.startWith)(null)).subscribe(function(){e._validatePlaceholders(),e._syncDescribedByIds(),e._changeDetectorRef.markForCheck()}),t.ngControl&&t.ngControl.valueChanges&&t.ngControl.valueChanges.pipe(Object(l.takeUntil)(this._destroyed)).subscribe(function(){return e._changeDetectorRef.markForCheck()}),this._ngZone&&this._ngZone.onStable.asObservable().pipe(Object(l.takeUntil)(this._destroyed)).subscribe(function(){e._outlineGapCalculationNeededOnStable&&e.updateOutlineGap()}),Object(c.merge)(this._prefixChildren.changes,this._suffixChildren.changes).subscribe(function(){e._updateOutlineGapOnStable(),e._changeDetectorRef.markForCheck()}),this._hintChildren.changes.pipe(Object(l.startWith)(null)).subscribe(function(){e._processHints(),e._changeDetectorRef.markForCheck()}),this._errorChildren.changes.pipe(Object(l.startWith)(null)).subscribe(function(){e._syncDescribedByIds(),e._changeDetectorRef.markForCheck()}),this._dir&&this._dir.change.pipe(Object(l.takeUntil)(this._destroyed)).subscribe(function(){return e.updateOutlineGap()})},t.prototype.ngAfterContentChecked=function(){this._validateControlChild(),this._outlineGapCalculationNeededImmediately&&this.updateOutlineGap()},t.prototype.ngAfterViewInit=function(){this._subscriptAnimationState="enter",this._changeDetectorRef.detectChanges()},t.prototype.ngOnDestroy=function(){this._destroyed.next(),this._destroyed.complete()},t.prototype._shouldForward=function(e){var t=this._control?this._control.ngControl:null;return t&&t[e]},t.prototype._hasPlaceholder=function(){return!!(this._control&&this._control.placeholder||this._placeholderChild)},t.prototype._hasLabel=function(){return!!this._labelChild},t.prototype._shouldLabelFloat=function(){return this._canLabelFloat&&(this._control.shouldLabelFloat||this._shouldAlwaysFloat)},t.prototype._hideControlPlaceholder=function(){return"legacy"===this.appearance&&!this._hasLabel()||this._hasLabel()&&!this._shouldLabelFloat()},t.prototype._hasFloatingLabel=function(){return this._hasLabel()||"legacy"===this.appearance&&this._hasPlaceholder()},t.prototype._getDisplayedMessages=function(){return this._errorChildren&&this._errorChildren.length>0&&this._control.errorState?"error":"hint"},t.prototype._animateAndLockLabel=function(){var e=this;this._hasFloatingLabel()&&this._canLabelFloat&&(this._animationsEnabled&&(this._showAlwaysAnimate=!0,Object(c.fromEvent)(this._label.nativeElement,"transitionend").pipe(Object(l.take)(1)).subscribe(function(){e._showAlwaysAnimate=!1})),this.floatLabel="always",this._changeDetectorRef.markForCheck())},t.prototype._validatePlaceholders=function(){if(this._control.placeholder&&this._placeholderChild)throw y()},t.prototype._processHints=function(){this._validateHints(),this._syncDescribedByIds()},t.prototype._validateHints=function(){var e,t,n=this;this._hintChildren&&this._hintChildren.forEach(function(r){if("start"===r.align){if(e||n.hintLabel)throw b("start");e=r}else if("end"===r.align){if(t)throw b("end");t=r}})},t.prototype._syncDescribedByIds=function(){if(this._control){var e=[];if("hint"===this._getDisplayedMessages()){var t=this._hintChildren?this._hintChildren.find(function(e){return"start"===e.align}):null,n=this._hintChildren?this._hintChildren.find(function(e){return"end"===e.align}):null;t?e.push(t.id):this._hintLabel&&e.push(this._hintLabelId),n&&e.push(n.id)}else this._errorChildren&&(e=this._errorChildren.map(function(e){return e.id}));this._control.setDescribedByIds(e)}},t.prototype._validateControlChild=function(){if(!this._control)throw j()},t.prototype.updateOutlineGap=function(){var e=this._label?this._label.nativeElement:null;if("outline"===this.appearance&&e&&e.children.length&&e.textContent.trim()&&(!this._platform||this._platform.isBrowser))if(document.documentElement.contains(this._elementRef.nativeElement)){var t=0,n=0,r=this._connectionContainerRef.nativeElement,i=r.querySelectorAll(".mat-form-field-outline-start"),o=r.querySelectorAll(".mat-form-field-outline-gap");if(this._label&&this._label.nativeElement.children.length){var a=r.getBoundingClientRect();if(0===a.width&&0===a.height)return this._outlineGapCalculationNeededOnStable=!0,void(this._outlineGapCalculationNeededImmediately=!1);for(var s=this._getStartEnd(a),u=this._getStartEnd(e.children[0].getBoundingClientRect()),c=0,l=0,d=e.children;l<d.length;l++){c+=d[l].offsetWidth}t=u-s-5,n=c>0?.75*c+10:0}for(var p=0;p<i.length;p++)i.item(p).style.width=t+"px";for(p=0;p<o.length;p++)o.item(p).style.width=n+"px";this._outlineGapCalculationNeededOnStable=this._outlineGapCalculationNeededImmediately=!1}else this._outlineGapCalculationNeededImmediately=!0},t.prototype._getStartEnd=function(e){return this._dir&&"rtl"===this._dir.value?e.right:e.left},t.prototype._updateOutlineGapOnStable=function(){var e=this;this._ngZone?this._outlineGapCalculationNeededOnStable=!0:Promise.resolve().then(function(){return e.updateOutlineGap()})},t.decorators=[{type:r.Component,args:[{selector:"mat-form-field",exportAs:"matFormField",template:'<div class="mat-form-field-wrapper"><div class="mat-form-field-flex" #connectionContainer (click)="_control.onContainerClick && _control.onContainerClick($event)"><ng-container *ngIf="appearance == \'outline\'"><div class="mat-form-field-outline"><div class="mat-form-field-outline-start"></div><div class="mat-form-field-outline-gap"></div><div class="mat-form-field-outline-end"></div></div><div class="mat-form-field-outline mat-form-field-outline-thick"><div class="mat-form-field-outline-start"></div><div class="mat-form-field-outline-gap"></div><div class="mat-form-field-outline-end"></div></div></ng-container><div class="mat-form-field-prefix" *ngIf="_prefixChildren.length"><ng-content select="[matPrefix]"></ng-content></div><div class="mat-form-field-infix" #inputContainer><ng-content></ng-content><span class="mat-form-field-label-wrapper"><label class="mat-form-field-label" (cdkObserveContent)="updateOutlineGap()" [cdkObserveContentDisabled]="appearance != \'outline\'" [id]="_labelId" [attr.for]="_control.id" [attr.aria-owns]="_control.id" [class.mat-empty]="_control.empty && !_shouldAlwaysFloat" [class.mat-form-field-empty]="_control.empty && !_shouldAlwaysFloat" [class.mat-accent]="color == \'accent\'" [class.mat-warn]="color == \'warn\'" #label *ngIf="_hasFloatingLabel()" [ngSwitch]="_hasLabel()"><ng-container *ngSwitchCase="false"><ng-content select="mat-placeholder"></ng-content>{{_control.placeholder}}</ng-container><ng-content select="mat-label" *ngSwitchCase="true"></ng-content><span class="mat-placeholder-required mat-form-field-required-marker" aria-hidden="true" *ngIf="!hideRequiredMarker && _control.required && !_control.disabled">&nbsp;*</span></label></span></div><div class="mat-form-field-suffix" *ngIf="_suffixChildren.length"><ng-content select="[matSuffix]"></ng-content></div></div><div class="mat-form-field-underline" #underline *ngIf="appearance != \'outline\'"><span class="mat-form-field-ripple" [class.mat-accent]="color == \'accent\'" [class.mat-warn]="color == \'warn\'"></span></div><div class="mat-form-field-subscript-wrapper" [ngSwitch]="_getDisplayedMessages()"><div *ngSwitchCase="\'error\'" [@transitionMessages]="_subscriptAnimationState"><ng-content select="mat-error"></ng-content></div><div class="mat-form-field-hint-wrapper" *ngSwitchCase="\'hint\'" [@transitionMessages]="_subscriptAnimationState"><div *ngIf="hintLabel" [id]="_hintLabelId" class="mat-hint">{{hintLabel}}</div><ng-content select="mat-hint:not([align=\'end\'])"></ng-content><div class="mat-form-field-hint-spacer"></div><ng-content select="mat-hint[align=\'end\']"></ng-content></div></div></div>',styles:[".mat-form-field{display:inline-block;position:relative;text-align:left}[dir=rtl] .mat-form-field{text-align:right}.mat-form-field-wrapper{position:relative}.mat-form-field-flex{display:inline-flex;align-items:baseline;box-sizing:border-box;width:100%}.mat-form-field-prefix,.mat-form-field-suffix{white-space:nowrap;flex:none;position:relative}.mat-form-field-infix{display:block;position:relative;flex:auto;min-width:0;width:180px}@media (-ms-high-contrast:active){.mat-form-field-infix{border-image:linear-gradient(transparent,transparent)}}.mat-form-field-label-wrapper{position:absolute;left:0;box-sizing:content-box;width:100%;height:100%;overflow:hidden;pointer-events:none}[dir=rtl] .mat-form-field-label-wrapper{left:auto;right:0}.mat-form-field-label{position:absolute;left:0;font:inherit;pointer-events:none;width:100%;white-space:nowrap;text-overflow:ellipsis;overflow:hidden;transform-origin:0 0;transition:transform .4s cubic-bezier(.25,.8,.25,1),color .4s cubic-bezier(.25,.8,.25,1),width .4s cubic-bezier(.25,.8,.25,1);display:none}[dir=rtl] .mat-form-field-label{transform-origin:100% 0;left:auto;right:0}.mat-form-field-can-float.mat-form-field-should-float .mat-form-field-label,.mat-form-field-empty.mat-form-field-label{display:block}.mat-form-field-autofill-control:-webkit-autofill+.mat-form-field-label-wrapper .mat-form-field-label{display:none}.mat-form-field-can-float .mat-form-field-autofill-control:-webkit-autofill+.mat-form-field-label-wrapper .mat-form-field-label{display:block;transition:none}.mat-input-server:focus+.mat-form-field-label-wrapper .mat-form-field-label,.mat-input-server[placeholder]:not(:placeholder-shown)+.mat-form-field-label-wrapper .mat-form-field-label{display:none}.mat-form-field-can-float .mat-input-server:focus+.mat-form-field-label-wrapper .mat-form-field-label,.mat-form-field-can-float .mat-input-server[placeholder]:not(:placeholder-shown)+.mat-form-field-label-wrapper .mat-form-field-label{display:block}.mat-form-field-label:not(.mat-form-field-empty){transition:none}.mat-form-field-underline{position:absolute;width:100%;pointer-events:none;transform:scaleY(1.0001)}.mat-form-field-ripple{position:absolute;left:0;width:100%;transform-origin:50%;transform:scaleX(.5);opacity:0;transition:background-color .3s cubic-bezier(.55,0,.55,.2)}.mat-form-field.mat-focused .mat-form-field-ripple,.mat-form-field.mat-form-field-invalid .mat-form-field-ripple{opacity:1;transform:scaleX(1);transition:transform .3s cubic-bezier(.25,.8,.25,1),opacity .1s cubic-bezier(.25,.8,.25,1),background-color .3s cubic-bezier(.25,.8,.25,1)}.mat-form-field-subscript-wrapper{position:absolute;box-sizing:border-box;width:100%;overflow:hidden}.mat-form-field-label-wrapper .mat-icon,.mat-form-field-subscript-wrapper .mat-icon{width:1em;height:1em;font-size:inherit;vertical-align:baseline}.mat-form-field-hint-wrapper{display:flex}.mat-form-field-hint-spacer{flex:1 0 1em}.mat-error{display:block}.mat-form-field-control-wrapper{position:relative}.mat-form-field._mat-animation-noopable .mat-form-field-label,.mat-form-field._mat-animation-noopable .mat-form-field-ripple{transition:none} .mat-form-field-appearance-fill .mat-form-field-flex{border-radius:4px 4px 0 0;padding:.75em .75em 0 .75em}@media (-ms-high-contrast:active){.mat-form-field-appearance-fill .mat-form-field-flex{outline:solid 1px}}.mat-form-field-appearance-fill .mat-form-field-underline::before{content:'';display:block;position:absolute;bottom:0;height:1px;width:100%}.mat-form-field-appearance-fill .mat-form-field-ripple{bottom:0;height:2px}@media (-ms-high-contrast:active){.mat-form-field-appearance-fill .mat-form-field-ripple{height:0;border-top:solid 2px}}.mat-form-field-appearance-fill:not(.mat-form-field-disabled) .mat-form-field-flex:hover~.mat-form-field-underline .mat-form-field-ripple{opacity:1;transform:none;transition:opacity .6s cubic-bezier(.25,.8,.25,1)}.mat-form-field-appearance-fill._mat-animation-noopable:not(.mat-form-field-disabled) .mat-form-field-flex:hover~.mat-form-field-underline .mat-form-field-ripple{transition:none}.mat-form-field-appearance-fill .mat-form-field-subscript-wrapper{padding:0 1em} .mat-input-element{font:inherit;background:0 0;color:currentColor;border:none;outline:0;padding:0;margin:0;width:100%;max-width:100%;vertical-align:bottom;text-align:inherit}.mat-input-element:-moz-ui-invalid{box-shadow:none}.mat-input-element::-ms-clear,.mat-input-element::-ms-reveal{display:none}.mat-input-element,.mat-input-element::-webkit-search-cancel-button,.mat-input-element::-webkit-search-decoration,.mat-input-element::-webkit-search-results-button,.mat-input-element::-webkit-search-results-decoration{-webkit-appearance:none}.mat-input-element::-webkit-caps-lock-indicator,.mat-input-element::-webkit-contacts-auto-fill-button,.mat-input-element::-webkit-credentials-auto-fill-button{visibility:hidden}.mat-input-element[type=date]::after,.mat-input-element[type=datetime-local]::after,.mat-input-element[type=datetime]::after,.mat-input-element[type=month]::after,.mat-input-element[type=time]::after,.mat-input-element[type=week]::after{content:' ';white-space:pre;width:1px}.mat-input-element::-webkit-calendar-picker-indicator,.mat-input-element::-webkit-clear-button,.mat-input-element::-webkit-inner-spin-button{font-size:.75em}.mat-input-element::placeholder{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;transition:color .4s .133s cubic-bezier(.25,.8,.25,1)}.mat-input-element::-moz-placeholder{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;transition:color .4s .133s cubic-bezier(.25,.8,.25,1)}.mat-input-element::-webkit-input-placeholder{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;transition:color .4s .133s cubic-bezier(.25,.8,.25,1)}.mat-input-element:-ms-input-placeholder{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;transition:color .4s .133s cubic-bezier(.25,.8,.25,1)}.mat-form-field-hide-placeholder .mat-input-element::placeholder{color:transparent!important;-webkit-text-fill-color:transparent;transition:none}.mat-form-field-hide-placeholder .mat-input-element::-moz-placeholder{color:transparent!important;-webkit-text-fill-color:transparent;transition:none}.mat-form-field-hide-placeholder .mat-input-element::-webkit-input-placeholder{color:transparent!important;-webkit-text-fill-color:transparent;transition:none}.mat-form-field-hide-placeholder .mat-input-element:-ms-input-placeholder{color:transparent!important;-webkit-text-fill-color:transparent;transition:none}textarea.mat-input-element{resize:vertical;overflow:auto}textarea.mat-input-element.cdk-textarea-autosize{resize:none}textarea.mat-input-element{padding:2px 0;margin:-2px 0}select.mat-input-element{-moz-appearance:none;-webkit-appearance:none;position:relative;background-color:transparent;display:inline-flex;box-sizing:border-box;padding-top:1em;top:-1em;margin-bottom:-1em}select.mat-input-element::-ms-expand{display:none}select.mat-input-element::-moz-focus-inner{border:0}select.mat-input-element:not(:disabled){cursor:pointer}select.mat-input-element::-ms-value{color:inherit;background:0 0}.mat-form-field-type-mat-native-select .mat-form-field-infix::after{content:'';width:0;height:0;border-left:5px solid transparent;border-right:5px solid transparent;border-top:5px solid;position:absolute;top:50%;right:0;margin-top:-2.5px}[dir=rtl] .mat-form-field-type-mat-native-select .mat-form-field-infix::after{right:auto;left:0}.mat-form-field-type-mat-native-select .mat-input-element{padding-right:15px}[dir=rtl] .mat-form-field-type-mat-native-select .mat-input-element{padding-right:0;padding-left:15px}.mat-form-field-type-mat-native-select .mat-form-field-label-wrapper{max-width:calc(100% - 10px)}.mat-form-field-type-mat-native-select.mat-form-field-appearance-outline .mat-form-field-infix::after{margin-top:-5px}.mat-form-field-type-mat-native-select.mat-form-field-appearance-fill .mat-form-field-infix::after{margin-top:-10px} .mat-form-field-appearance-legacy .mat-form-field-label{transform:perspective(100px);-ms-transform:none}.mat-form-field-appearance-legacy .mat-form-field-prefix .mat-icon,.mat-form-field-appearance-legacy .mat-form-field-suffix .mat-icon{width:1em}.mat-form-field-appearance-legacy .mat-form-field-prefix .mat-icon-button,.mat-form-field-appearance-legacy .mat-form-field-suffix .mat-icon-button{font:inherit;vertical-align:baseline}.mat-form-field-appearance-legacy .mat-form-field-prefix .mat-icon-button .mat-icon,.mat-form-field-appearance-legacy .mat-form-field-suffix .mat-icon-button .mat-icon{font-size:inherit}.mat-form-field-appearance-legacy .mat-form-field-underline{height:1px}@media (-ms-high-contrast:active){.mat-form-field-appearance-legacy .mat-form-field-underline{height:0;border-top:solid 1px}}.mat-form-field-appearance-legacy .mat-form-field-ripple{top:0;height:2px;overflow:hidden}@media (-ms-high-contrast:active){.mat-form-field-appearance-legacy .mat-form-field-ripple{height:0;border-top:solid 2px}}.mat-form-field-appearance-legacy.mat-form-field-disabled .mat-form-field-underline{background-position:0;background-color:transparent}@media (-ms-high-contrast:active){.mat-form-field-appearance-legacy.mat-form-field-disabled .mat-form-field-underline{border-top-style:dotted;border-top-width:2px}}.mat-form-field-appearance-legacy.mat-form-field-invalid:not(.mat-focused) .mat-form-field-ripple{height:1px} .mat-form-field-appearance-outline .mat-form-field-wrapper{margin:.25em 0}.mat-form-field-appearance-outline .mat-form-field-flex{padding:0 .75em 0 .75em;margin-top:-.25em;position:relative}.mat-form-field-appearance-outline .mat-form-field-prefix,.mat-form-field-appearance-outline .mat-form-field-suffix{top:.25em}.mat-form-field-appearance-outline .mat-form-field-outline{display:flex;position:absolute;top:.25em;left:0;right:0;bottom:0;pointer-events:none}.mat-form-field-appearance-outline .mat-form-field-outline-end,.mat-form-field-appearance-outline .mat-form-field-outline-start{border:1px solid currentColor;min-width:5px}.mat-form-field-appearance-outline .mat-form-field-outline-start{border-radius:5px 0 0 5px;border-right-style:none}[dir=rtl] .mat-form-field-appearance-outline .mat-form-field-outline-start{border-right-style:solid;border-left-style:none;border-radius:0 5px 5px 0}.mat-form-field-appearance-outline .mat-form-field-outline-end{border-radius:0 5px 5px 0;border-left-style:none;flex-grow:1}[dir=rtl] .mat-form-field-appearance-outline .mat-form-field-outline-end{border-left-style:solid;border-right-style:none;border-radius:5px 0 0 5px}.mat-form-field-appearance-outline .mat-form-field-outline-gap{border-radius:.000001px;border:1px solid currentColor;border-left-style:none;border-right-style:none}.mat-form-field-appearance-outline.mat-form-field-can-float.mat-form-field-should-float .mat-form-field-outline-gap{border-top-color:transparent}.mat-form-field-appearance-outline .mat-form-field-outline-thick{opacity:0}.mat-form-field-appearance-outline .mat-form-field-outline-thick .mat-form-field-outline-end,.mat-form-field-appearance-outline .mat-form-field-outline-thick .mat-form-field-outline-gap,.mat-form-field-appearance-outline .mat-form-field-outline-thick .mat-form-field-outline-start{border-width:2px;transition:border-color .3s cubic-bezier(.25,.8,.25,1)}.mat-form-field-appearance-outline.mat-focused .mat-form-field-outline,.mat-form-field-appearance-outline.mat-form-field-invalid .mat-form-field-outline{opacity:0;transition:opacity .1s cubic-bezier(.25,.8,.25,1)}.mat-form-field-appearance-outline.mat-focused .mat-form-field-outline-thick,.mat-form-field-appearance-outline.mat-form-field-invalid .mat-form-field-outline-thick{opacity:1}.mat-form-field-appearance-outline:not(.mat-form-field-disabled) .mat-form-field-flex:hover .mat-form-field-outline{opacity:0;transition:opacity .6s cubic-bezier(.25,.8,.25,1)}.mat-form-field-appearance-outline:not(.mat-form-field-disabled) .mat-form-field-flex:hover .mat-form-field-outline-thick{opacity:1}.mat-form-field-appearance-outline .mat-form-field-subscript-wrapper{padding:0 1em}.mat-form-field-appearance-outline._mat-animation-noopable .mat-form-field-outline,.mat-form-field-appearance-outline._mat-animation-noopable .mat-form-field-outline-end,.mat-form-field-appearance-outline._mat-animation-noopable .mat-form-field-outline-gap,.mat-form-field-appearance-outline._mat-animation-noopable .mat-form-field-outline-start,.mat-form-field-appearance-outline._mat-animation-noopable:not(.mat-form-field-disabled) .mat-form-field-flex:hover~.mat-form-field-outline{transition:none} .mat-form-field-appearance-standard .mat-form-field-flex{padding-top:.75em}.mat-form-field-appearance-standard .mat-form-field-underline{height:1px}@media (-ms-high-contrast:active){.mat-form-field-appearance-standard .mat-form-field-underline{height:0;border-top:solid 1px}}.mat-form-field-appearance-standard .mat-form-field-ripple{bottom:0;height:2px}@media (-ms-high-contrast:active){.mat-form-field-appearance-standard .mat-form-field-ripple{height:0;border-top:2px}}.mat-form-field-appearance-standard.mat-form-field-disabled .mat-form-field-underline{background-position:0;background-color:transparent}@media (-ms-high-contrast:active){.mat-form-field-appearance-standard.mat-form-field-disabled .mat-form-field-underline{border-top-style:dotted;border-top-width:2px}}.mat-form-field-appearance-standard:not(.mat-form-field-disabled) .mat-form-field-flex:hover~.mat-form-field-underline .mat-form-field-ripple{opacity:1;transform:none;transition:opacity .6s cubic-bezier(.25,.8,.25,1)}.mat-form-field-appearance-standard._mat-animation-noopable:not(.mat-form-field-disabled) .mat-form-field-flex:hover~.mat-form-field-underline .mat-form-field-ripple{transition:none}"],animations:[v.transitionMessages],host:{class:"mat-form-field","[class.mat-form-field-appearance-standard]":'appearance == "standard"',"[class.mat-form-field-appearance-fill]":'appearance == "fill"',"[class.mat-form-field-appearance-outline]":'appearance == "outline"',"[class.mat-form-field-appearance-legacy]":'appearance == "legacy"',"[class.mat-form-field-invalid]":"_control.errorState","[class.mat-form-field-can-float]":"_canLabelFloat","[class.mat-form-field-should-float]":"_shouldLabelFloat()","[class.mat-form-field-has-label]":"_hasFloatingLabel()","[class.mat-form-field-hide-placeholder]":"_hideControlPlaceholder()","[class.mat-form-field-disabled]":"_control.disabled","[class.mat-form-field-autofilled]":"_control.autofilled","[class.mat-focused]":"_control.focused","[class.mat-accent]":'color == "accent"',"[class.mat-warn]":'color == "warn"',"[class.ng-untouched]":'_shouldForward("untouched")',"[class.ng-touched]":'_shouldForward("touched")',"[class.ng-pristine]":'_shouldForward("pristine")',"[class.ng-dirty]":'_shouldForward("dirty")',"[class.ng-valid]":'_shouldForward("valid")',"[class.ng-invalid]":'_shouldForward("invalid")',"[class.ng-pending]":'_shouldForward("pending")',"[class._mat-animation-noopable]":"!_animationsEnabled"},inputs:["color"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:r.ChangeDetectorRef},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[u.MAT_LABEL_GLOBAL_OPTIONS]}]},{type:a.Directionality,decorators:[{type:r.Optional}]},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[T]}]},{type:d.Platform},{type:r.NgZone},{type:String,decorators:[{type:r.Optional},{type:r.Inject,args:[p.ANIMATION_MODULE_TYPE]}]}]},t.propDecorators={appearance:[{type:r.Input}],hideRequiredMarker:[{type:r.Input}],hintLabel:[{type:r.Input}],floatLabel:[{type:r.Input}],underlineRef:[{type:r.ViewChild,args:["underline"]}],_connectionContainerRef:[{type:r.ViewChild,args:["connectionContainer"]}],_inputContainerRef:[{type:r.ViewChild,args:["inputContainer"]}],_label:[{type:r.ViewChild,args:["label"]}],_control:[{type:r.ContentChild,args:[_]}],_placeholderChild:[{type:r.ContentChild,args:[C]}],_labelChild:[{type:r.ContentChild,args:[w]}],_errorChildren:[{type:r.ContentChildren,args:[g]}],_hintChildren:[{type:r.ContentChildren,args:[O]}],_prefixChildren:[{type:r.ContentChildren,args:[E]}],_suffixChildren:[{type:r.ContentChildren,args:[S]}]},t}(A),I=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{declarations:[g,D,O,w,C,E,S],imports:[f.CommonModule,h.ObserversModule],exports:[g,D,O,w,C,E,S]}]}],e}()},"./node_modules/@angular/material/esm5/grid-list.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatGridListModule",function(){return w}),n.d(t,"MatGridList",function(){return O}),n.d(t,"MatGridTile",function(){return c}),n.d(t,"MatGridTileText",function(){return l}),n.d(t,"MatGridAvatarCssMatStyler",function(){return d}),n.d(t,"MatGridTileHeaderCssMatStyler",function(){return p}),n.d(t,"MatGridTileFooterCssMatStyler",function(){return f}),n.d(t,"ɵa15",function(){return u});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/material/esm5/core.es5.js"),o=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),a=n("./node_modules/tslib/tslib.es6.js"),s=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),u=new r.InjectionToken("MAT_GRID_LIST"),c=function(){function e(e,t){this._element=e,this._gridList=t,this._rowspan=1,this._colspan=1}return Object.defineProperty(e.prototype,"rowspan",{get:function(){return this._rowspan},set:function(e){this._rowspan=Math.round(Object(o.coerceNumberProperty)(e))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"colspan",{get:function(){return this._colspan},set:function(e){this._colspan=Math.round(Object(o.coerceNumberProperty)(e))},enumerable:!0,configurable:!0}),e.prototype._setStyle=function(e,t){this._element.nativeElement.style[e]=t},e.decorators=[{type:r.Component,args:[{selector:"mat-grid-tile",exportAs:"matGridTile",host:{class:"mat-grid-tile"},template:'<figure class="mat-figure"><ng-content></ng-content></figure>',styles:[".mat-grid-list{display:block;position:relative}.mat-grid-tile{display:block;position:absolute;overflow:hidden}.mat-grid-tile .mat-figure{top:0;left:0;right:0;bottom:0;position:absolute;display:flex;align-items:center;justify-content:center;height:100%;padding:0;margin:0}.mat-grid-tile .mat-grid-tile-footer,.mat-grid-tile .mat-grid-tile-header{display:flex;align-items:center;height:48px;color:#fff;background:rgba(0,0,0,.38);overflow:hidden;padding:0 16px;position:absolute;left:0;right:0}.mat-grid-tile .mat-grid-tile-footer>*,.mat-grid-tile .mat-grid-tile-header>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-grid-tile .mat-grid-tile-footer.mat-2-line,.mat-grid-tile .mat-grid-tile-header.mat-2-line{height:68px}.mat-grid-tile .mat-grid-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden}.mat-grid-tile .mat-grid-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-grid-tile .mat-grid-list-text:empty{display:none}.mat-grid-tile .mat-grid-tile-header{top:0}.mat-grid-tile .mat-grid-tile-footer{bottom:0}.mat-grid-tile .mat-grid-avatar{padding-right:16px}[dir=rtl] .mat-grid-tile .mat-grid-avatar{padding-right:0;padding-left:16px}.mat-grid-tile .mat-grid-avatar:empty{display:none}"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[u]}]}]},e.propDecorators={rowspan:[{type:r.Input}],colspan:[{type:r.Input}]},e}(),l=function(){function e(e){this._element=e}return e.prototype.ngAfterContentInit=function(){Object(i.setLines)(this._lines,this._element)},e.decorators=[{type:r.Component,args:[{selector:"mat-grid-tile-header, mat-grid-tile-footer",template:'<ng-content select="[mat-grid-avatar], [matGridAvatar]"></ng-content><div class="mat-grid-list-text"><ng-content select="[mat-line], [matLine]"></ng-content></div><ng-content></ng-content>',changeDetection:r.ChangeDetectionStrategy.OnPush,encapsulation:r.ViewEncapsulation.None}]}],e.ctorParameters=function(){return[{type:r.ElementRef}]},e.propDecorators={_lines:[{type:r.ContentChildren,args:[i.MatLine]}]},e}(),d=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"[mat-grid-avatar], [matGridAvatar]",host:{class:"mat-grid-avatar"}}]}],e}(),p=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"mat-grid-tile-header",host:{class:"mat-grid-tile-header"}}]}],e}(),f=function(){function e(){}return e.decorators=[{type:r.Directive,args:[{selector:"mat-grid-tile-footer",host:{class:"mat-grid-tile-footer"}}]}],e}(),h=function(){function e(){this.columnIndex=0,this.rowIndex=0}return Object.defineProperty(e.prototype,"rowCount",{get:function(){return this.rowIndex+1},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"rowspan",{get:function(){var e=Math.max.apply(Math,this.tracker);return e>1?this.rowCount+e-1:this.rowCount},enumerable:!0,configurable:!0}),e.prototype.update=function(e,t){var n=this;this.columnIndex=0,this.rowIndex=0,this.tracker=new Array(e),this.tracker.fill(0,0,this.tracker.length),this.positions=t.map(function(e){return n._trackTile(e)})},e.prototype._trackTile=function(e){var t=this._findMatchingGap(e.colspan);return this._markTilePosition(t,e),this.columnIndex=t+e.colspan,new m(this.rowIndex,t)},e.prototype._findMatchingGap=function(e){if(e>this.tracker.length)throw Error("mat-grid-list: tile with colspan "+e+' is wider than grid with cols="'+this.tracker.length+'".');var t=-1,n=-1;do{this.columnIndex+e>this.tracker.length?(this._nextRow(),t=this.tracker.indexOf(0,this.columnIndex),n=this._findGapEndIndex(t)):-1!=(t=this.tracker.indexOf(0,this.columnIndex))?(n=this._findGapEndIndex(t),this.columnIndex=t+1):(this._nextRow(),t=this.tracker.indexOf(0,this.columnIndex),n=this._findGapEndIndex(t))}while(n-t<e||0==n);return Math.max(t,0)},e.prototype._nextRow=function(){this.columnIndex=0,this.rowIndex++;for(var e=0;e<this.tracker.length;e++)this.tracker[e]=Math.max(0,this.tracker[e]-1)},e.prototype._findGapEndIndex=function(e){for(var t=e+1;t<this.tracker.length;t++)if(0!=this.tracker[t])return t;return this.tracker.length},e.prototype._markTilePosition=function(e,t){for(var n=0;n<t.colspan;n++)this.tracker[e+n]=t.rowspan},e}(),m=function(){return function(e,t){this.row=e,this.col=t}}(),g=/^-?\d+((\.\d+)?[A-Za-z%$]?)+$/,v=function(){function e(){this._rows=0,this._rowspan=0}return e.prototype.init=function(e,t,n,r){this._gutterSize=x(e),this._rows=t.rowCount,this._rowspan=t.rowspan,this._cols=n,this._direction=r},e.prototype.getBaseTileSize=function(e,t){return"("+e+"% - ("+this._gutterSize+" * "+t+"))"},e.prototype.getTilePosition=function(e,t){return 0===t?"0":j("("+e+" + "+this._gutterSize+") * "+t)},e.prototype.getTileSize=function(e,t){return"("+e+" * "+t+") + ("+(t-1)+" * "+this._gutterSize+")"},e.prototype.setStyle=function(e,t,n){var r=100/this._cols,i=(this._cols-1)/this._cols;this.setColStyles(e,n,r,i),this.setRowStyles(e,t,r,i)},e.prototype.setColStyles=function(e,t,n,r){var i=this.getBaseTileSize(n,r),o="rtl"===this._direction?"right":"left";e._setStyle(o,this.getTilePosition(i,t)),e._setStyle("width",j(this.getTileSize(i,e.colspan)))},e.prototype.getGutterSpan=function(){return this._gutterSize+" * ("+this._rowspan+" - 1)"},e.prototype.getTileSpan=function(e){return this._rowspan+" * "+this.getTileSize(e,1)},e.prototype.getComputedHeight=function(){return null},e}(),_=function(e){function t(t){var n=e.call(this)||this;return n.fixedRowHeight=t,n}return Object(a.__extends)(t,e),t.prototype.init=function(t,n,r,i){if(e.prototype.init.call(this,t,n,r,i),this.fixedRowHeight=x(this.fixedRowHeight),!g.test(this.fixedRowHeight))throw Error('Invalid value "'+this.fixedRowHeight+'" set as rowHeight.')},t.prototype.setRowStyles=function(e,t){e._setStyle("top",this.getTilePosition(this.fixedRowHeight,t)),e._setStyle("height",j(this.getTileSize(this.fixedRowHeight,e.rowspan)))},t.prototype.getComputedHeight=function(){return["height",j(this.getTileSpan(this.fixedRowHeight)+" + "+this.getGutterSpan())]},t.prototype.reset=function(e){e._setListStyle(["height",null]),e._tiles&&e._tiles.forEach(function(e){e._setStyle("top",null),e._setStyle("height",null)})},t}(v),y=function(e){function t(t){var n=e.call(this)||this;return n._parseRatio(t),n}return Object(a.__extends)(t,e),t.prototype.setRowStyles=function(e,t,n,r){var i=n/this.rowHeightRatio;this.baseTileHeight=this.getBaseTileSize(i,r),e._setStyle("marginTop",this.getTilePosition(this.baseTileHeight,t)),e._setStyle("paddingTop",j(this.getTileSize(this.baseTileHeight,e.rowspan)))},t.prototype.getComputedHeight=function(){return["paddingBottom",j(this.getTileSpan(this.baseTileHeight)+" + "+this.getGutterSpan())]},t.prototype.reset=function(e){e._setListStyle(["paddingBottom",null]),e._tiles.forEach(function(e){e._setStyle("marginTop",null),e._setStyle("paddingTop",null)})},t.prototype._parseRatio=function(e){var t=e.split(":");if(2!==t.length)throw Error('mat-grid-list: invalid ratio given for row-height: "'+e+'"');this.rowHeightRatio=parseFloat(t[0])/parseFloat(t[1])},t}(v),b=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(a.__extends)(t,e),t.prototype.setRowStyles=function(e,t){var n=100/this._rowspan,r=(this._rows-1)/this._rows,i=this.getBaseTileSize(n,r);e._setStyle("top",this.getTilePosition(i,t)),e._setStyle("height",j(this.getTileSize(i,e.rowspan)))},t.prototype.reset=function(e){e._tiles&&e._tiles.forEach(function(e){e._setStyle("top",null),e._setStyle("height",null)})},t}(v);function j(e){return"calc("+e+")"}function x(e){return e.match(/([A-Za-z%]+)$/)?e:e+"px"}var O=function(){function e(e,t){this._element=e,this._dir=t,this._gutter="1px"}return Object.defineProperty(e.prototype,"cols",{get:function(){return this._cols},set:function(e){this._cols=Math.max(1,Math.round(Object(o.coerceNumberProperty)(e)))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"gutterSize",{get:function(){return this._gutter},set:function(e){this._gutter=""+(null==e?"":e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"rowHeight",{get:function(){return this._rowHeight},set:function(e){var t=""+(null==e?"":e);t!==this._rowHeight&&(this._rowHeight=t,this._setTileStyler(this._rowHeight))},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){this._checkCols(),this._checkRowHeight()},e.prototype.ngAfterContentChecked=function(){this._layoutTiles()},e.prototype._checkCols=function(){if(!this.cols)throw Error('mat-grid-list: must pass in number of columns. Example: <mat-grid-list cols="3">')},e.prototype._checkRowHeight=function(){this._rowHeight||this._setTileStyler("1:1")},e.prototype._setTileStyler=function(e){this._tileStyler&&this._tileStyler.reset(this),"fit"===e?this._tileStyler=new b:e&&e.indexOf(":")>-1?this._tileStyler=new y(e):this._tileStyler=new _(e)},e.prototype._layoutTiles=function(){var e=this;this._tileCoordinator||(this._tileCoordinator=new h);var t=this._tileCoordinator,n=this._tiles.filter(function(t){return!t._gridList||t._gridList===e}),r=this._dir?this._dir.value:"ltr";this._tileCoordinator.update(this.cols,n),this._tileStyler.init(this.gutterSize,t,this.cols,r),n.forEach(function(n,r){var i=t.positions[r];e._tileStyler.setStyle(n,i.row,i.col)}),this._setListStyle(this._tileStyler.getComputedHeight())},e.prototype._setListStyle=function(e){e&&(this._element.nativeElement.style[e[0]]=e[1])},e.decorators=[{type:r.Component,args:[{selector:"mat-grid-list",exportAs:"matGridList",template:"<div><ng-content></ng-content></div>",styles:[".mat-grid-list{display:block;position:relative}.mat-grid-tile{display:block;position:absolute;overflow:hidden}.mat-grid-tile .mat-figure{top:0;left:0;right:0;bottom:0;position:absolute;display:flex;align-items:center;justify-content:center;height:100%;padding:0;margin:0}.mat-grid-tile .mat-grid-tile-footer,.mat-grid-tile .mat-grid-tile-header{display:flex;align-items:center;height:48px;color:#fff;background:rgba(0,0,0,.38);overflow:hidden;padding:0 16px;position:absolute;left:0;right:0}.mat-grid-tile .mat-grid-tile-footer>*,.mat-grid-tile .mat-grid-tile-header>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-grid-tile .mat-grid-tile-footer.mat-2-line,.mat-grid-tile .mat-grid-tile-header.mat-2-line{height:68px}.mat-grid-tile .mat-grid-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden}.mat-grid-tile .mat-grid-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-grid-tile .mat-grid-list-text:empty{display:none}.mat-grid-tile .mat-grid-tile-header{top:0}.mat-grid-tile .mat-grid-tile-footer{bottom:0}.mat-grid-tile .mat-grid-avatar{padding-right:16px}[dir=rtl] .mat-grid-tile .mat-grid-avatar{padding-right:0;padding-left:16px}.mat-grid-tile .mat-grid-avatar:empty{display:none}"],host:{class:"mat-grid-list"},providers:[{provide:u,useExisting:e}],changeDetection:r.ChangeDetectionStrategy.OnPush,encapsulation:r.ViewEncapsulation.None}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:s.Directionality,decorators:[{type:r.Optional}]}]},e.propDecorators={_tiles:[{type:r.ContentChildren,args:[c,{descendants:!0}]}],cols:[{type:r.Input}],gutterSize:[{type:r.Input}],rowHeight:[{type:r.Input}]},e}(),w=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[i.MatLineModule,i.MatCommonModule],exports:[O,c,l,i.MatLineModule,i.MatCommonModule,p,f,d],declarations:[O,c,l,p,f,d]}]}],e}()},"./node_modules/@angular/material/esm5/icon.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatIconModule",function(){return A}),n.d(t,"MAT_ICON_LOCATION_FACTORY",function(){return C}),n.d(t,"MatIconBase",function(){return x}),n.d(t,"_MatIconMixinBase",function(){return O}),n.d(t,"MAT_ICON_LOCATION",function(){return w}),n.d(t,"MatIcon",function(){return M}),n.d(t,"getMatIconNameNotFoundError",function(){return p}),n.d(t,"getMatIconNoHttpProviderError",function(){return f}),n.d(t,"getMatIconFailedToSanitizeUrlError",function(){return h}),n.d(t,"getMatIconFailedToSanitizeLiteralError",function(){return m}),n.d(t,"ICON_REGISTRY_PROVIDER_FACTORY",function(){return _}),n.d(t,"MatIconRegistry",function(){return v}),n.d(t,"ICON_REGISTRY_PROVIDER",function(){return y});var r=n("./node_modules/@angular/common/fesm5/common.js"),i=n("./node_modules/@angular/common/fesm5/http.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),s=n("./node_modules/rxjs/_esm5/index.js"),u=n("./node_modules/rxjs/_esm5/operators/index.js"),c=n("./node_modules/tslib/tslib.es6.js"),l=n("./node_modules/@angular/material/esm5/core.es5.js"),d=n("./node_modules/@angular/cdk/esm5/coercion.es5.js");function p(e){return Error('Unable to find icon with the name "'+e+'"')}function f(){return Error("Could not find HttpClient provider for use with Angular Material icons. Please include the HttpClientModule from @angular/common/http in your app imports.")}function h(e){return Error("The URL provided to MatIconRegistry was not trusted as a resource URL via Angular's DomSanitizer. Attempted URL was \""+e+'".')}function m(e){return Error("The literal provided to MatIconRegistry was not trusted as safe HTML by Angular's DomSanitizer. Attempted literal was \""+e+'".')}var g=function(){return function(e){e.nodeName?this.svgElement=e:this.url=e}}(),v=function(){function e(e,t,n){this._httpClient=e,this._sanitizer=t,this._svgIconConfigs=new Map,this._iconSetConfigs=new Map,this._cachedIconsByUrl=new Map,this._inProgressUrlFetches=new Map,this._fontCssClassesByAlias=new Map,this._defaultFontSetClass="material-icons",this._document=n}return e.prototype.addSvgIcon=function(e,t){return this.addSvgIconInNamespace("",e,t)},e.prototype.addSvgIconLiteral=function(e,t){return this.addSvgIconLiteralInNamespace("",e,t)},e.prototype.addSvgIconInNamespace=function(e,t,n){return this._addSvgIconConfig(e,t,new g(n))},e.prototype.addSvgIconLiteralInNamespace=function(e,t,n){var r=this._sanitizer.sanitize(o.SecurityContext.HTML,n);if(!r)throw m(n);var i=this._createSvgElementForSingleIcon(r);return this._addSvgIconConfig(e,t,new g(i))},e.prototype.addSvgIconSet=function(e){return this.addSvgIconSetInNamespace("",e)},e.prototype.addSvgIconSetLiteral=function(e){return this.addSvgIconSetLiteralInNamespace("",e)},e.prototype.addSvgIconSetInNamespace=function(e,t){return this._addSvgIconSetConfig(e,new g(t))},e.prototype.addSvgIconSetLiteralInNamespace=function(e,t){var n=this._sanitizer.sanitize(o.SecurityContext.HTML,t);if(!n)throw m(t);var r=this._svgElementFromString(n);return this._addSvgIconSetConfig(e,new g(r))},e.prototype.registerFontClassAlias=function(e,t){return void 0===t&&(t=e),this._fontCssClassesByAlias.set(e,t),this},e.prototype.classNameForFontAlias=function(e){return this._fontCssClassesByAlias.get(e)||e},e.prototype.setDefaultFontSetClass=function(e){return this._defaultFontSetClass=e,this},e.prototype.getDefaultFontSetClass=function(){return this._defaultFontSetClass},e.prototype.getSvgIconFromUrl=function(e){var t=this,n=this._sanitizer.sanitize(o.SecurityContext.RESOURCE_URL,e);if(!n)throw h(e);var r=this._cachedIconsByUrl.get(n);return r?Object(s.of)(b(r)):this._loadSvgIconFromConfig(new g(e)).pipe(Object(u.tap)(function(e){return t._cachedIconsByUrl.set(n,e)}),Object(u.map)(function(e){return b(e)}))},e.prototype.getNamedSvgIcon=function(e,t){void 0===t&&(t="");var n=j(t,e),r=this._svgIconConfigs.get(n);if(r)return this._getSvgFromConfig(r);var i=this._iconSetConfigs.get(t);return i?this._getSvgFromIconSetConfigs(e,i):Object(s.throwError)(p(n))},e.prototype._getSvgFromConfig=function(e){return e.svgElement?Object(s.of)(b(e.svgElement)):this._loadSvgIconFromConfig(e).pipe(Object(u.tap)(function(t){return e.svgElement=t}),Object(u.map)(function(e){return b(e)}))},e.prototype._getSvgFromIconSetConfigs=function(e,t){var n=this,r=this._extractIconWithNameFromAnySet(e,t);if(r)return Object(s.of)(r);var i=t.filter(function(e){return!e.svgElement}).map(function(e){return n._loadSvgIconSetFromConfig(e).pipe(Object(u.catchError)(function(t){var r=n._sanitizer.sanitize(o.SecurityContext.RESOURCE_URL,e.url);return console.error("Loading icon set URL: "+r+" failed: "+t.message),Object(s.of)(null)}))});return Object(s.forkJoin)(i).pipe(Object(u.map)(function(){var r=n._extractIconWithNameFromAnySet(e,t);if(!r)throw p(e);return r}))},e.prototype._extractIconWithNameFromAnySet=function(e,t){for(var n=t.length-1;n>=0;n--){var r=t[n];if(r.svgElement){var i=this._extractSvgIconFromSet(r.svgElement,e);if(i)return i}}return null},e.prototype._loadSvgIconFromConfig=function(e){var t=this;return this._fetchUrl(e.url).pipe(Object(u.map)(function(e){return t._createSvgElementForSingleIcon(e)}))},e.prototype._loadSvgIconSetFromConfig=function(e){var t=this;return e.svgElement?Object(s.of)(e.svgElement):this._fetchUrl(e.url).pipe(Object(u.map)(function(n){return e.svgElement||(e.svgElement=t._svgElementFromString(n)),e.svgElement}))},e.prototype._createSvgElementForSingleIcon=function(e){var t=this._svgElementFromString(e);return this._setSvgAttributes(t),t},e.prototype._extractSvgIconFromSet=function(e,t){var n=e.querySelector("#"+t);if(!n)return null;var r=n.cloneNode(!0);if(r.removeAttribute("id"),"svg"===r.nodeName.toLowerCase())return this._setSvgAttributes(r);if("symbol"===r.nodeName.toLowerCase())return this._setSvgAttributes(this._toSvgElement(r));var i=this._svgElementFromString("<svg></svg>");return i.appendChild(r),this._setSvgAttributes(i)},e.prototype._svgElementFromString=function(e){var t=this._document.createElement("DIV");t.innerHTML=e;var n=t.querySelector("svg");if(!n)throw Error("<svg> tag not found");return n},e.prototype._toSvgElement=function(e){for(var t=this._svgElementFromString("<svg></svg>"),n=0;n<e.childNodes.length;n++)e.childNodes[n].nodeType===this._document.ELEMENT_NODE&&t.appendChild(e.childNodes[n].cloneNode(!0));return t},e.prototype._setSvgAttributes=function(e){return e.setAttribute("fit",""),e.setAttribute("height","100%"),e.setAttribute("width","100%"),e.setAttribute("preserveAspectRatio","xMidYMid meet"),e.setAttribute("focusable","false"),e},e.prototype._fetchUrl=function(e){var t=this;if(!this._httpClient)throw f();if(null==e)throw Error('Cannot fetch icon from URL "'+e+'".');var n=this._sanitizer.sanitize(o.SecurityContext.RESOURCE_URL,e);if(!n)throw h(e);var r=this._inProgressUrlFetches.get(n);if(r)return r;var i=this._httpClient.get(n,{responseType:"text"}).pipe(Object(u.finalize)(function(){return t._inProgressUrlFetches.delete(n)}),Object(u.share)());return this._inProgressUrlFetches.set(n,i),i},e.prototype._addSvgIconConfig=function(e,t,n){return this._svgIconConfigs.set(j(e,t),n),this},e.prototype._addSvgIconSetConfig=function(e,t){var n=this._iconSetConfigs.get(e);return n?n.push(t):this._iconSetConfigs.set(e,[t]),this},e.decorators=[{type:o.Injectable,args:[{providedIn:"root"}]}],e.ctorParameters=function(){return[{type:i.HttpClient,decorators:[{type:o.Optional}]},{type:a.DomSanitizer},{type:void 0,decorators:[{type:o.Optional},{type:o.Inject,args:[r.DOCUMENT]}]}]},e.ngInjectableDef=Object(o.defineInjectable)({factory:function(){return new e(Object(o.inject)(i.HttpClient,8),Object(o.inject)(a.DomSanitizer),Object(o.inject)(r.DOCUMENT,8))},token:e,providedIn:"root"}),e}();function _(e,t,n,r){return e||new v(t,n,r)}var y={provide:v,deps:[[new o.Optional,new o.SkipSelf,v],[new o.Optional,i.HttpClient],a.DomSanitizer,[new o.Optional,r.DOCUMENT]],useFactory:_};function b(e){return e.cloneNode(!0)}function j(e,t){return e+":"+t}var x=function(){return function(e){this._elementRef=e}}(),O=Object(l.mixinColor)(x),w=new o.InjectionToken("mat-icon-location",{providedIn:"root",factory:C});function C(){var e=Object(o.inject)(r.DOCUMENT),t=e?e.location:null;return{getPathname:function(){return t?t.pathname+t.search:""}}}var E=["clip-path","color-profile","src","cursor","fill","filter","marker","marker-start","marker-mid","marker-end","mask","stroke"],S=E.map(function(e){return"["+e+"]"}).join(", "),k=/^url\(['"]?#(.*?)['"]?\)$/,M=function(e){function t(t,n,r,i){var o=e.call(this,t)||this;return o._iconRegistry=n,o._location=i,o._inline=!1,r||t.nativeElement.setAttribute("aria-hidden","true"),o}return Object(c.__extends)(t,e),Object.defineProperty(t.prototype,"inline",{get:function(){return this._inline},set:function(e){this._inline=Object(d.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"fontSet",{get:function(){return this._fontSet},set:function(e){this._fontSet=this._cleanupFontValue(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"fontIcon",{get:function(){return this._fontIcon},set:function(e){this._fontIcon=this._cleanupFontValue(e)},enumerable:!0,configurable:!0}),t.prototype._splitIconName=function(e){if(!e)return["",""];var t=e.split(":");switch(t.length){case 1:return["",t[0]];case 2:return t;default:throw Error('Invalid icon name: "'+e+'"')}},t.prototype.ngOnChanges=function(e){var t=this;if(e.svgIcon)if(this.svgIcon){var n=this._splitIconName(this.svgIcon),r=n[0],i=n[1];this._iconRegistry.getNamedSvgIcon(i,r).pipe(Object(u.take)(1)).subscribe(function(e){return t._setSvgElement(e)},function(e){return console.log("Error retrieving icon: "+e.message)})}else this._clearSvgElement();this._usingFontIcon()&&this._updateFontIconClasses()},t.prototype.ngOnInit=function(){this._usingFontIcon()&&this._updateFontIconClasses()},t.prototype.ngAfterViewChecked=function(){var e=this._elementsWithExternalReferences;if(e&&this._location&&e.size){var t=this._location.getPathname();t!==this._previousPath&&(this._previousPath=t,this._prependPathToReferences(t))}},t.prototype.ngOnDestroy=function(){this._elementsWithExternalReferences&&this._elementsWithExternalReferences.clear()},t.prototype._usingFontIcon=function(){return!this.svgIcon},t.prototype._setSvgElement=function(e){this._clearSvgElement();for(var t=e.querySelectorAll("style"),n=0;n<t.length;n++)t[n].textContent+=" ";if(this._location){var r=this._location.getPathname();this._previousPath=r,this._cacheChildrenWithExternalReferences(e),this._prependPathToReferences(r)}this._elementRef.nativeElement.appendChild(e)},t.prototype._clearSvgElement=function(){var e=this._elementRef.nativeElement,t=e.childNodes.length;for(this._elementsWithExternalReferences&&this._elementsWithExternalReferences.clear();t--;){var n=e.childNodes[t];1===n.nodeType&&"svg"!==n.nodeName.toLowerCase()||e.removeChild(n)}},t.prototype._updateFontIconClasses=function(){if(this._usingFontIcon()){var e=this._elementRef.nativeElement,t=this.fontSet?this._iconRegistry.classNameForFontAlias(this.fontSet):this._iconRegistry.getDefaultFontSetClass();t!=this._previousFontSetClass&&(this._previousFontSetClass&&e.classList.remove(this._previousFontSetClass),t&&e.classList.add(t),this._previousFontSetClass=t),this.fontIcon!=this._previousFontIconClass&&(this._previousFontIconClass&&e.classList.remove(this._previousFontIconClass),this.fontIcon&&e.classList.add(this.fontIcon),this._previousFontIconClass=this.fontIcon)}},t.prototype._cleanupFontValue=function(e){return"string"==typeof e?e.trim().split(" ")[0]:e},t.prototype._prependPathToReferences=function(e){var t=this._elementsWithExternalReferences;t&&t.forEach(function(t,n){t.forEach(function(t){n.setAttribute(t.name,"url('"+e+"#"+t.value+"')")})})},t.prototype._cacheChildrenWithExternalReferences=function(e){for(var t=e.querySelectorAll(S),n=this._elementsWithExternalReferences=this._elementsWithExternalReferences||new Map,r=function(e){E.forEach(function(r){var i=t[e],o=i.getAttribute(r),a=o?o.match(k):null;if(a){var s=n.get(i);s||(s=[],n.set(i,s)),s.push({name:r,value:a[1]})}})},i=0;i<t.length;i++)r(i)},t.decorators=[{type:o.Component,args:[{template:"<ng-content></ng-content>",selector:"mat-icon",exportAs:"matIcon",styles:[".mat-icon{background-repeat:no-repeat;display:inline-block;fill:currentColor;height:24px;width:24px}.mat-icon.mat-icon-inline{font-size:inherit;height:inherit;line-height:inherit;width:inherit}[dir=rtl] .mat-icon-rtl-mirror{transform:scale(-1,1)}.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-prefix .mat-icon,.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-suffix .mat-icon{display:block}.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-prefix .mat-icon-button .mat-icon,.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-suffix .mat-icon-button .mat-icon{margin:auto}"],inputs:["color"],host:{role:"img",class:"mat-icon","[class.mat-icon-inline]":"inline","[class.mat-icon-no-color]":'color !== "primary" && color !== "accent" && color !== "warn"'},encapsulation:o.ViewEncapsulation.None,changeDetection:o.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:o.ElementRef},{type:v},{type:String,decorators:[{type:o.Attribute,args:["aria-hidden"]}]},{type:void 0,decorators:[{type:o.Optional},{type:o.Inject,args:[w]}]}]},t.propDecorators={inline:[{type:o.Input}],svgIcon:[{type:o.Input}],fontSet:[{type:o.Input}],fontIcon:[{type:o.Input}]},t}(O),A=function(){function e(){}return e.decorators=[{type:o.NgModule,args:[{imports:[l.MatCommonModule],exports:[M,l.MatCommonModule],declarations:[M]}]}],e}()},"./node_modules/@angular/material/esm5/input.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatTextareaAutosize",function(){return f}),n.d(t,"MatInputBase",function(){return _}),n.d(t,"_MatInputMixinBase",function(){return y}),n.d(t,"MatInput",function(){return b}),n.d(t,"getMatInputUnsupportedTypeError",function(){return h}),n.d(t,"MatInputModule",function(){return j}),n.d(t,"MAT_INPUT_VALUE_ACCESSOR",function(){return m});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/text-field.es5.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),s=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),u=n("./node_modules/@angular/forms/fesm5/forms.js"),c=n("./node_modules/@angular/material/esm5/core.es5.js"),l=n("./node_modules/@angular/material/esm5/form-field.es5.js"),d=n("./node_modules/rxjs/_esm5/index.js"),p=n("./node_modules/@angular/common/fesm5/common.js"),f=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"matAutosizeMinRows",{get:function(){return this.minRows},set:function(e){this.minRows=e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"matAutosizeMaxRows",{get:function(){return this.maxRows},set:function(e){this.maxRows=e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"matAutosize",{get:function(){return this.enabled},set:function(e){this.enabled=e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"matTextareaAutosize",{get:function(){return this.enabled},set:function(e){this.enabled=e},enumerable:!0,configurable:!0}),t.decorators=[{type:o.Directive,args:[{selector:"textarea[mat-autosize], textarea[matTextareaAutosize]",exportAs:"matTextareaAutosize",inputs:["cdkAutosizeMinRows","cdkAutosizeMaxRows"],host:{class:"cdk-textarea-autosize mat-autosize",rows:"1","(input)":"_noopInputHandler()"}}]}],t.propDecorators={matAutosizeMinRows:[{type:o.Input}],matAutosizeMaxRows:[{type:o.Input}],matAutosize:[{type:o.Input,args:["mat-autosize"]}],matTextareaAutosize:[{type:o.Input}]},t}(i.CdkTextareaAutosize);function h(e){return Error('Input type "'+e+"\" isn't supported by matInput.")}var m=new o.InjectionToken("MAT_INPUT_VALUE_ACCESSOR"),g=["button","checkbox","file","hidden","image","radio","range","reset","submit"],v=0,_=function(){return function(e,t,n,r){this._defaultErrorStateMatcher=e,this._parentForm=t,this._parentFormGroup=n,this.ngControl=r}}(),y=Object(c.mixinErrorState)(_),b=function(e){function t(t,n,r,i,o,a,u,c,l){var p=e.call(this,a,i,o,r)||this;p._elementRef=t,p._platform=n,p.ngControl=r,p._autofillMonitor=c,p._uid="mat-input-"+v++,p._isServer=!1,p._isNativeSelect=!1,p.focused=!1,p.stateChanges=new d.Subject,p.controlType="mat-input",p.autofilled=!1,p._disabled=!1,p._required=!1,p._type="text",p._readonly=!1,p._neverEmptyInputTypes=["date","datetime","datetime-local","month","time","week"].filter(function(e){return Object(s.getSupportedInputTypes)().has(e)});var f=p._elementRef.nativeElement;return p._inputValueAccessor=u||f,p._previousNativeValue=p.value,p.id=p.id,n.IOS&&l.runOutsideAngular(function(){t.nativeElement.addEventListener("keyup",function(e){var t=e.target;t.value||t.selectionStart||t.selectionEnd||(t.setSelectionRange(1,1),t.setSelectionRange(0,0))})}),p._isServer=!p._platform.isBrowser,p._isNativeSelect="select"===f.nodeName.toLowerCase(),p._isNativeSelect&&(p.controlType=f.multiple?"mat-native-select-multiple":"mat-native-select"),p}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"disabled",{get:function(){return this.ngControl&&null!==this.ngControl.disabled?this.ngControl.disabled:this._disabled},set:function(e){this._disabled=Object(a.coerceBooleanProperty)(e),this.focused&&(this.focused=!1,this.stateChanges.next())},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"id",{get:function(){return this._id},set:function(e){this._id=e||this._uid},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"required",{get:function(){return this._required},set:function(e){this._required=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"type",{get:function(){return this._type},set:function(e){this._type=e||"text",this._validateType(),!this._isTextarea()&&Object(s.getSupportedInputTypes)().has(this._type)&&(this._elementRef.nativeElement.type=this._type)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{get:function(){return this._inputValueAccessor.value},set:function(e){e!==this.value&&(this._inputValueAccessor.value=e,this.stateChanges.next())},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"readonly",{get:function(){return this._readonly},set:function(e){this._readonly=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){var e=this;this._platform.isBrowser&&this._autofillMonitor.monitor(this._elementRef.nativeElement).subscribe(function(t){e.autofilled=t.isAutofilled,e.stateChanges.next()})},t.prototype.ngOnChanges=function(){this.stateChanges.next()},t.prototype.ngOnDestroy=function(){this.stateChanges.complete(),this._platform.isBrowser&&this._autofillMonitor.stopMonitoring(this._elementRef.nativeElement)},t.prototype.ngDoCheck=function(){this.ngControl&&this.updateErrorState(),this._dirtyCheckNativeValue()},t.prototype.focus=function(){this._elementRef.nativeElement.focus()},t.prototype._focusChanged=function(e){e===this.focused||this.readonly&&e||(this.focused=e,this.stateChanges.next())},t.prototype._onInput=function(){},t.prototype._dirtyCheckNativeValue=function(){var e=this._elementRef.nativeElement.value;this._previousNativeValue!==e&&(this._previousNativeValue=e,this.stateChanges.next())},t.prototype._validateType=function(){if(g.indexOf(this._type)>-1)throw h(this._type)},t.prototype._isNeverEmpty=function(){return this._neverEmptyInputTypes.indexOf(this._type)>-1},t.prototype._isBadInput=function(){var e=this._elementRef.nativeElement.validity;return e&&e.badInput},t.prototype._isTextarea=function(){return"textarea"===this._elementRef.nativeElement.nodeName.toLowerCase()},Object.defineProperty(t.prototype,"empty",{get:function(){return!(this._isNeverEmpty()||this._elementRef.nativeElement.value||this._isBadInput()||this.autofilled)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"shouldLabelFloat",{get:function(){if(this._isNativeSelect){var e=this._elementRef.nativeElement,t=e.options[0];return this.focused||e.multiple||!this.empty||!!(e.selectedIndex>-1&&t&&t.label)}return this.focused||!this.empty},enumerable:!0,configurable:!0}),t.prototype.setDescribedByIds=function(e){this._ariaDescribedby=e.join(" ")},t.prototype.onContainerClick=function(){this.focused||this.focus()},t.decorators=[{type:o.Directive,args:[{selector:"input[matInput], textarea[matInput], select[matNativeControl],\n      input[matNativeControl], textarea[matNativeControl]",exportAs:"matInput",host:{class:"mat-input-element mat-form-field-autofill-control","[class.mat-input-server]":"_isServer","[attr.id]":"id","[attr.placeholder]":"placeholder","[disabled]":"disabled","[required]":"required","[attr.readonly]":"readonly && !_isNativeSelect || null","[attr.aria-describedby]":"_ariaDescribedby || null","[attr.aria-invalid]":"errorState","[attr.aria-required]":"required.toString()","(blur)":"_focusChanged(false)","(focus)":"_focusChanged(true)","(input)":"_onInput()"},providers:[{provide:l.MatFormFieldControl,useExisting:t}]}]}],t.ctorParameters=function(){return[{type:o.ElementRef},{type:s.Platform},{type:u.NgControl,decorators:[{type:o.Optional},{type:o.Self}]},{type:u.NgForm,decorators:[{type:o.Optional}]},{type:u.FormGroupDirective,decorators:[{type:o.Optional}]},{type:c.ErrorStateMatcher},{type:void 0,decorators:[{type:o.Optional},{type:o.Self},{type:o.Inject,args:[m]}]},{type:i.AutofillMonitor},{type:o.NgZone}]},t.propDecorators={disabled:[{type:o.Input}],id:[{type:o.Input}],placeholder:[{type:o.Input}],required:[{type:o.Input}],type:[{type:o.Input}],errorStateMatcher:[{type:o.Input}],value:[{type:o.Input}],readonly:[{type:o.Input}]},t}(y),j=function(){function e(){}return e.decorators=[{type:o.NgModule,args:[{declarations:[b,f],imports:[p.CommonModule,i.TextFieldModule,l.MatFormFieldModule],exports:[i.TextFieldModule,l.MatFormFieldModule,b,f],providers:[c.ErrorStateMatcher]}]}],e}()},"./node_modules/@angular/material/esm5/list.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatListModule",function(){return D}),n.d(t,"MatListBase",function(){return h}),n.d(t,"_MatListMixinBase",function(){return m}),n.d(t,"MatListItemBase",function(){return g}),n.d(t,"_MatListItemMixinBase",function(){return v}),n.d(t,"MatNavList",function(){return _}),n.d(t,"MatList",function(){return y}),n.d(t,"MatListAvatarCssMatStyler",function(){return b}),n.d(t,"MatListIconCssMatStyler",function(){return j}),n.d(t,"MatListSubheaderCssMatStyler",function(){return x}),n.d(t,"MatListItem",function(){return O}),n.d(t,"MatSelectionListBase",function(){return w}),n.d(t,"_MatSelectionListMixinBase",function(){return C}),n.d(t,"MatListOptionBase",function(){return E}),n.d(t,"_MatListOptionMixinBase",function(){return S}),n.d(t,"MAT_SELECTION_LIST_VALUE_ACCESSOR",function(){return k}),n.d(t,"MatSelectionListChange",function(){return M}),n.d(t,"MatListOption",function(){return A}),n.d(t,"MatSelectionList",function(){return T});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/material/esm5/core.es5.js"),a=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),c=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),l=n("./node_modules/@angular/forms/fesm5/forms.js"),d=n("./node_modules/rxjs/_esm5/index.js"),p=n("./node_modules/@angular/common/fesm5/common.js"),f=n("./node_modules/@angular/material/esm5/divider.es5.js"),h=function(){return function(){}}(),m=Object(o.mixinDisableRipple)(h),g=function(){return function(){}}(),v=Object(o.mixinDisableRipple)(g),_=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.decorators=[{type:i.Component,args:[{selector:"mat-nav-list",exportAs:"matNavList",host:{role:"navigation",class:"mat-nav-list mat-list-base"},template:"<ng-content></ng-content>",styles:[".mat-subheader{display:flex;box-sizing:border-box;padding:16px;align-items:center}.mat-list-base .mat-subheader{margin:0}.mat-list-base{padding-top:8px;display:block;-webkit-tap-highlight-color:transparent}.mat-list-base .mat-subheader{height:48px;line-height:16px}.mat-list-base .mat-subheader:first-child{margin-top:-8px}.mat-list-base .mat-list-item,.mat-list-base .mat-list-option{display:block;height:48px;-webkit-tap-highlight-color:transparent;width:100%;padding:0}.mat-list-base .mat-list-item .mat-list-item-content,.mat-list-base .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list-base .mat-list-item .mat-list-item-content-reverse,.mat-list-base .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list-base .mat-list-item .mat-list-item-ripple,.mat-list-base .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list-base .mat-list-item.mat-list-item-with-avatar,.mat-list-base .mat-list-option.mat-list-item-with-avatar{height:56px}.mat-list-base .mat-list-item.mat-2-line,.mat-list-base .mat-list-option.mat-2-line{height:72px}.mat-list-base .mat-list-item.mat-3-line,.mat-list-base .mat-list-option.mat-3-line{height:88px}.mat-list-base .mat-list-item.mat-multi-line,.mat-list-base .mat-list-option.mat-multi-line{height:auto}.mat-list-base .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list-base .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list-base .mat-list-item .mat-list-text,.mat-list-base .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0}.mat-list-base .mat-list-item .mat-list-text>*,.mat-list-base .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list-base .mat-list-item .mat-list-text:empty,.mat-list-base .mat-list-option .mat-list-text:empty{display:none}.mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:0;padding-left:16px}[dir=rtl] .mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:16px;padding-left:0}.mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-left:0;padding-right:16px}[dir=rtl] .mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:0;padding-left:16px}.mat-list-base .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:16px;padding-left:16px}.mat-list-base .mat-list-item .mat-list-avatar,.mat-list-base .mat-list-option .mat-list-avatar{flex-shrink:0;width:40px;height:40px;border-radius:50%;object-fit:cover}.mat-list-base .mat-list-item .mat-list-avatar~.mat-divider-inset,.mat-list-base .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:72px;width:calc(100% - 72px)}[dir=rtl] .mat-list-base .mat-list-item .mat-list-avatar~.mat-divider-inset,[dir=rtl] .mat-list-base .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:auto;margin-right:72px}.mat-list-base .mat-list-item .mat-list-icon,.mat-list-base .mat-list-option .mat-list-icon{flex-shrink:0;width:24px;height:24px;font-size:24px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-list-base .mat-list-item .mat-list-icon~.mat-divider-inset,.mat-list-base .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:64px;width:calc(100% - 64px)}[dir=rtl] .mat-list-base .mat-list-item .mat-list-icon~.mat-divider-inset,[dir=rtl] .mat-list-base .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:auto;margin-right:64px}.mat-list-base .mat-list-item .mat-divider,.mat-list-base .mat-list-option .mat-divider{position:absolute;bottom:0;left:0;width:100%;margin:0}[dir=rtl] .mat-list-base .mat-list-item .mat-divider,[dir=rtl] .mat-list-base .mat-list-option .mat-divider{margin-left:auto;margin-right:0}.mat-list-base .mat-list-item .mat-divider.mat-divider-inset,.mat-list-base .mat-list-option .mat-divider.mat-divider-inset{position:absolute}.mat-list-base[dense]{padding-top:4px;display:block}.mat-list-base[dense] .mat-subheader{height:40px;line-height:8px}.mat-list-base[dense] .mat-subheader:first-child{margin-top:-4px}.mat-list-base[dense] .mat-list-item,.mat-list-base[dense] .mat-list-option{display:block;height:40px;-webkit-tap-highlight-color:transparent;width:100%;padding:0}.mat-list-base[dense] .mat-list-item .mat-list-item-content,.mat-list-base[dense] .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list-base[dense] .mat-list-item .mat-list-item-content-reverse,.mat-list-base[dense] .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list-base[dense] .mat-list-item .mat-list-item-ripple,.mat-list-base[dense] .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar{height:48px}.mat-list-base[dense] .mat-list-item.mat-2-line,.mat-list-base[dense] .mat-list-option.mat-2-line{height:60px}.mat-list-base[dense] .mat-list-item.mat-3-line,.mat-list-base[dense] .mat-list-option.mat-3-line{height:76px}.mat-list-base[dense] .mat-list-item.mat-multi-line,.mat-list-base[dense] .mat-list-option.mat-multi-line{height:auto}.mat-list-base[dense] .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list-base[dense] .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list-base[dense] .mat-list-item .mat-list-text,.mat-list-base[dense] .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0}.mat-list-base[dense] .mat-list-item .mat-list-text>*,.mat-list-base[dense] .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list-base[dense] .mat-list-item .mat-list-text:empty,.mat-list-base[dense] .mat-list-option .mat-list-text:empty{display:none}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:0;padding-left:16px}[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:16px;padding-left:0}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-left:0;padding-right:16px}[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:0;padding-left:16px}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:16px;padding-left:16px}.mat-list-base[dense] .mat-list-item .mat-list-avatar,.mat-list-base[dense] .mat-list-option .mat-list-avatar{flex-shrink:0;width:36px;height:36px;border-radius:50%;object-fit:cover}.mat-list-base[dense] .mat-list-item .mat-list-avatar~.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:68px;width:calc(100% - 68px)}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-list-avatar~.mat-divider-inset,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:auto;margin-right:68px}.mat-list-base[dense] .mat-list-item .mat-list-icon,.mat-list-base[dense] .mat-list-option .mat-list-icon{flex-shrink:0;width:20px;height:20px;font-size:20px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-list-base[dense] .mat-list-item .mat-list-icon~.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:60px;width:calc(100% - 60px)}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-list-icon~.mat-divider-inset,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:auto;margin-right:60px}.mat-list-base[dense] .mat-list-item .mat-divider,.mat-list-base[dense] .mat-list-option .mat-divider{position:absolute;bottom:0;left:0;width:100%;margin:0}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-divider,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-divider{margin-left:auto;margin-right:0}.mat-list-base[dense] .mat-list-item .mat-divider.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-divider.mat-divider-inset{position:absolute}.mat-nav-list a{text-decoration:none;color:inherit}.mat-nav-list .mat-list-item{cursor:pointer;outline:0}mat-action-list button{background:0 0;color:inherit;border:none;font:inherit;outline:inherit}mat-action-list .mat-list-item{cursor:pointer;outline:inherit}.mat-list-option:not(.mat-list-item-disabled){cursor:pointer;outline:0}@media (-ms-high-contrast:active){.mat-selection-list:focus{outline-style:dotted}.mat-list-option:focus,.mat-list-option:hover,.mat-nav-list .mat-list-item:focus,.mat-nav-list .mat-list-item:hover,mat-action-list .mat-list-item:focus,mat-action-list .mat-list-item:hover{outline:dotted 1px}}@media (hover:none){.mat-list-option:not(.mat-list-item-disabled):hover,.mat-nav-list .mat-list-item:not(.mat-list-item-disabled):hover{background:0 0}}"],inputs:["disableRipple"],encapsulation:i.ViewEncapsulation.None,changeDetection:i.ChangeDetectionStrategy.OnPush}]}],t}(m),y=function(e){function t(t){var n=e.call(this)||this;return n._elementRef=t,n}return Object(r.__extends)(t,e),t.prototype._getListType=function(){var e=this._elementRef;if(e){var t=e.nativeElement.nodeName.toLowerCase();if("mat-list"===t)return"list";if("mat-action-list"===t)return"action-list"}return null},t.decorators=[{type:i.Component,args:[{selector:"mat-list, mat-action-list",exportAs:"matList",template:"<ng-content></ng-content>",host:{class:"mat-list mat-list-base"},styles:[".mat-subheader{display:flex;box-sizing:border-box;padding:16px;align-items:center}.mat-list-base .mat-subheader{margin:0}.mat-list-base{padding-top:8px;display:block;-webkit-tap-highlight-color:transparent}.mat-list-base .mat-subheader{height:48px;line-height:16px}.mat-list-base .mat-subheader:first-child{margin-top:-8px}.mat-list-base .mat-list-item,.mat-list-base .mat-list-option{display:block;height:48px;-webkit-tap-highlight-color:transparent;width:100%;padding:0}.mat-list-base .mat-list-item .mat-list-item-content,.mat-list-base .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list-base .mat-list-item .mat-list-item-content-reverse,.mat-list-base .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list-base .mat-list-item .mat-list-item-ripple,.mat-list-base .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list-base .mat-list-item.mat-list-item-with-avatar,.mat-list-base .mat-list-option.mat-list-item-with-avatar{height:56px}.mat-list-base .mat-list-item.mat-2-line,.mat-list-base .mat-list-option.mat-2-line{height:72px}.mat-list-base .mat-list-item.mat-3-line,.mat-list-base .mat-list-option.mat-3-line{height:88px}.mat-list-base .mat-list-item.mat-multi-line,.mat-list-base .mat-list-option.mat-multi-line{height:auto}.mat-list-base .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list-base .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list-base .mat-list-item .mat-list-text,.mat-list-base .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0}.mat-list-base .mat-list-item .mat-list-text>*,.mat-list-base .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list-base .mat-list-item .mat-list-text:empty,.mat-list-base .mat-list-option .mat-list-text:empty{display:none}.mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:0;padding-left:16px}[dir=rtl] .mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:16px;padding-left:0}.mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-left:0;padding-right:16px}[dir=rtl] .mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:0;padding-left:16px}.mat-list-base .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:16px;padding-left:16px}.mat-list-base .mat-list-item .mat-list-avatar,.mat-list-base .mat-list-option .mat-list-avatar{flex-shrink:0;width:40px;height:40px;border-radius:50%;object-fit:cover}.mat-list-base .mat-list-item .mat-list-avatar~.mat-divider-inset,.mat-list-base .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:72px;width:calc(100% - 72px)}[dir=rtl] .mat-list-base .mat-list-item .mat-list-avatar~.mat-divider-inset,[dir=rtl] .mat-list-base .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:auto;margin-right:72px}.mat-list-base .mat-list-item .mat-list-icon,.mat-list-base .mat-list-option .mat-list-icon{flex-shrink:0;width:24px;height:24px;font-size:24px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-list-base .mat-list-item .mat-list-icon~.mat-divider-inset,.mat-list-base .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:64px;width:calc(100% - 64px)}[dir=rtl] .mat-list-base .mat-list-item .mat-list-icon~.mat-divider-inset,[dir=rtl] .mat-list-base .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:auto;margin-right:64px}.mat-list-base .mat-list-item .mat-divider,.mat-list-base .mat-list-option .mat-divider{position:absolute;bottom:0;left:0;width:100%;margin:0}[dir=rtl] .mat-list-base .mat-list-item .mat-divider,[dir=rtl] .mat-list-base .mat-list-option .mat-divider{margin-left:auto;margin-right:0}.mat-list-base .mat-list-item .mat-divider.mat-divider-inset,.mat-list-base .mat-list-option .mat-divider.mat-divider-inset{position:absolute}.mat-list-base[dense]{padding-top:4px;display:block}.mat-list-base[dense] .mat-subheader{height:40px;line-height:8px}.mat-list-base[dense] .mat-subheader:first-child{margin-top:-4px}.mat-list-base[dense] .mat-list-item,.mat-list-base[dense] .mat-list-option{display:block;height:40px;-webkit-tap-highlight-color:transparent;width:100%;padding:0}.mat-list-base[dense] .mat-list-item .mat-list-item-content,.mat-list-base[dense] .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list-base[dense] .mat-list-item .mat-list-item-content-reverse,.mat-list-base[dense] .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list-base[dense] .mat-list-item .mat-list-item-ripple,.mat-list-base[dense] .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar{height:48px}.mat-list-base[dense] .mat-list-item.mat-2-line,.mat-list-base[dense] .mat-list-option.mat-2-line{height:60px}.mat-list-base[dense] .mat-list-item.mat-3-line,.mat-list-base[dense] .mat-list-option.mat-3-line{height:76px}.mat-list-base[dense] .mat-list-item.mat-multi-line,.mat-list-base[dense] .mat-list-option.mat-multi-line{height:auto}.mat-list-base[dense] .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list-base[dense] .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list-base[dense] .mat-list-item .mat-list-text,.mat-list-base[dense] .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0}.mat-list-base[dense] .mat-list-item .mat-list-text>*,.mat-list-base[dense] .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list-base[dense] .mat-list-item .mat-list-text:empty,.mat-list-base[dense] .mat-list-option .mat-list-text:empty{display:none}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:0;padding-left:16px}[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:16px;padding-left:0}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-left:0;padding-right:16px}[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:0;padding-left:16px}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:16px;padding-left:16px}.mat-list-base[dense] .mat-list-item .mat-list-avatar,.mat-list-base[dense] .mat-list-option .mat-list-avatar{flex-shrink:0;width:36px;height:36px;border-radius:50%;object-fit:cover}.mat-list-base[dense] .mat-list-item .mat-list-avatar~.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:68px;width:calc(100% - 68px)}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-list-avatar~.mat-divider-inset,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:auto;margin-right:68px}.mat-list-base[dense] .mat-list-item .mat-list-icon,.mat-list-base[dense] .mat-list-option .mat-list-icon{flex-shrink:0;width:20px;height:20px;font-size:20px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-list-base[dense] .mat-list-item .mat-list-icon~.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:60px;width:calc(100% - 60px)}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-list-icon~.mat-divider-inset,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:auto;margin-right:60px}.mat-list-base[dense] .mat-list-item .mat-divider,.mat-list-base[dense] .mat-list-option .mat-divider{position:absolute;bottom:0;left:0;width:100%;margin:0}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-divider,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-divider{margin-left:auto;margin-right:0}.mat-list-base[dense] .mat-list-item .mat-divider.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-divider.mat-divider-inset{position:absolute}.mat-nav-list a{text-decoration:none;color:inherit}.mat-nav-list .mat-list-item{cursor:pointer;outline:0}mat-action-list button{background:0 0;color:inherit;border:none;font:inherit;outline:inherit}mat-action-list .mat-list-item{cursor:pointer;outline:inherit}.mat-list-option:not(.mat-list-item-disabled){cursor:pointer;outline:0}@media (-ms-high-contrast:active){.mat-selection-list:focus{outline-style:dotted}.mat-list-option:focus,.mat-list-option:hover,.mat-nav-list .mat-list-item:focus,.mat-nav-list .mat-list-item:hover,mat-action-list .mat-list-item:focus,mat-action-list .mat-list-item:hover{outline:dotted 1px}}@media (hover:none){.mat-list-option:not(.mat-list-item-disabled):hover,.mat-nav-list .mat-list-item:not(.mat-list-item-disabled):hover{background:0 0}}"],inputs:["disableRipple"],encapsulation:i.ViewEncapsulation.None,changeDetection:i.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:i.ElementRef}]},t}(m),b=function(){function e(){}return e.decorators=[{type:i.Directive,args:[{selector:"[mat-list-avatar], [matListAvatar]",host:{class:"mat-list-avatar"}}]}],e}(),j=function(){function e(){}return e.decorators=[{type:i.Directive,args:[{selector:"[mat-list-icon], [matListIcon]",host:{class:"mat-list-icon"}}]}],e}(),x=function(){function e(){}return e.decorators=[{type:i.Directive,args:[{selector:"[mat-subheader], [matSubheader]",host:{class:"mat-subheader"}}]}],e}(),O=function(e){function t(t,n,r){var i=e.call(this)||this;i._element=t,i._isInteractiveList=!1,i._isInteractiveList=!!(n||r&&"action-list"===r._getListType()),i._list=n||r;var o=i._getHostElement();return"button"!==o.nodeName.toLowerCase()||o.hasAttribute("type")||o.setAttribute("type","button"),i}return Object(r.__extends)(t,e),t.prototype.ngAfterContentInit=function(){Object(o.setLines)(this._lines,this._element)},t.prototype._isRippleDisabled=function(){return!this._isInteractiveList||this.disableRipple||!(!this._list||!this._list.disableRipple)},t.prototype._getHostElement=function(){return this._element.nativeElement},t.decorators=[{type:i.Component,args:[{selector:"mat-list-item, a[mat-list-item], button[mat-list-item]",exportAs:"matListItem",host:{class:"mat-list-item","[class.mat-list-item-avatar]":"_avatar || _icon","[class.mat-list-item-with-avatar]":"_avatar || _icon"},inputs:["disableRipple"],template:'<div class="mat-list-item-content"><div class="mat-list-item-ripple" mat-ripple [matRippleTrigger]="_getHostElement()" [matRippleDisabled]="_isRippleDisabled()"></div><ng-content select="[mat-list-avatar], [mat-list-icon], [matListAvatar], [matListIcon]"></ng-content><div class="mat-list-text"><ng-content select="[mat-line], [matLine]"></ng-content></div><ng-content></ng-content></div>',encapsulation:i.ViewEncapsulation.None,changeDetection:i.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:i.ElementRef},{type:_,decorators:[{type:i.Optional}]},{type:y,decorators:[{type:i.Optional}]}]},t.propDecorators={_lines:[{type:i.ContentChildren,args:[o.MatLine]}],_avatar:[{type:i.ContentChild,args:[b]}],_icon:[{type:i.ContentChild,args:[j]}]},t}(v),w=function(){return function(){}}(),C=Object(o.mixinDisableRipple)(w),E=function(){return function(){}}(),S=Object(o.mixinDisableRipple)(E),k={provide:l.NG_VALUE_ACCESSOR,useExisting:Object(i.forwardRef)(function(){return T}),multi:!0},M=function(){return function(e,t){this.source=e,this.option=t}}(),A=function(e){function t(t,n,r){var i=e.call(this)||this;return i._element=t,i._changeDetector=n,i.selectionList=r,i._selected=!1,i._disabled=!1,i._hasFocus=!1,i.checkboxPosition="after",i}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"disabled",{get:function(){return this._disabled||this.selectionList&&this.selectionList.disabled},set:function(e){var t=Object(s.coerceBooleanProperty)(e);t!==this._disabled&&(this._disabled=t,this._changeDetector.markForCheck())},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selected",{get:function(){return this.selectionList.selectedOptions.isSelected(this)},set:function(e){var t=Object(s.coerceBooleanProperty)(e);t!==this._selected&&(this._setSelected(t),this.selectionList._reportValueChange())},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){var e=this,t=this._selected;Promise.resolve().then(function(){(e._selected||t)&&(e.selected=!0,e._changeDetector.markForCheck())})},t.prototype.ngAfterContentInit=function(){Object(o.setLines)(this._lines,this._element)},t.prototype.ngOnDestroy=function(){var e=this;this.selected&&Promise.resolve().then(function(){return e.selected=!1});var t=this._hasFocus,n=this.selectionList._removeOptionFromList(this);t&&n&&n.focus()},t.prototype.toggle=function(){this.selected=!this.selected},t.prototype.focus=function(){this._element.nativeElement.focus()},t.prototype.getLabel=function(){return this._text?this._text.nativeElement.textContent||"":""},t.prototype._isRippleDisabled=function(){return this.disabled||this.disableRipple||this.selectionList.disableRipple},t.prototype._handleClick=function(){this.disabled||(this.toggle(),this.selectionList._emitChangeEvent(this))},t.prototype._handleFocus=function(){this.selectionList._setFocusedOption(this),this._hasFocus=!0},t.prototype._handleBlur=function(){this.selectionList._onTouched(),this._hasFocus=!1},t.prototype._getHostElement=function(){return this._element.nativeElement},t.prototype._setSelected=function(e){return e!==this._selected&&(this._selected=e,e?this.selectionList.selectedOptions.select(this):this.selectionList.selectedOptions.deselect(this),this._changeDetector.markForCheck(),!0)},t.prototype._markForCheck=function(){this._changeDetector.markForCheck()},t.decorators=[{type:i.Component,args:[{selector:"mat-list-option",exportAs:"matListOption",inputs:["disableRipple"],host:{role:"option",class:"mat-list-item mat-list-option","(focus)":"_handleFocus()","(blur)":"_handleBlur()","(click)":"_handleClick()",tabindex:"-1","[class.mat-list-item-disabled]":"disabled","[class.mat-list-item-with-avatar]":"_avatar || _icon","[attr.aria-selected]":"selected.toString()","[attr.aria-disabled]":"disabled.toString()"},template:'<div class="mat-list-item-content" [class.mat-list-item-content-reverse]="checkboxPosition == \'after\'"><div mat-ripple class="mat-list-item-ripple" [matRippleTrigger]="_getHostElement()" [matRippleDisabled]="_isRippleDisabled()"></div><mat-pseudo-checkbox [state]="selected ? \'checked\' : \'unchecked\'" [disabled]="disabled"></mat-pseudo-checkbox><div class="mat-list-text" #text><ng-content></ng-content></div><ng-content select="[mat-list-avatar], [mat-list-icon], [matListAvatar], [matListIcon]"></ng-content></div>',encapsulation:i.ViewEncapsulation.None,changeDetection:i.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:i.ElementRef},{type:i.ChangeDetectorRef},{type:T,decorators:[{type:i.Inject,args:[Object(i.forwardRef)(function(){return T})]}]}]},t.propDecorators={_avatar:[{type:i.ContentChild,args:[b]}],_icon:[{type:i.ContentChild,args:[j]}],_lines:[{type:i.ContentChildren,args:[o.MatLine]}],_text:[{type:i.ViewChild,args:["text"]}],checkboxPosition:[{type:i.Input}],value:[{type:i.Input}],disabled:[{type:i.Input}],selected:[{type:i.Input}]},t}(S),T=function(e){function t(t,n){var r=e.call(this)||this;return r._element=t,r.selectionChange=new i.EventEmitter,r.tabIndex=0,r._disabled=!1,r.selectedOptions=new u.SelectionModel(!0),r._onChange=function(e){},r._modelChanges=d.Subscription.EMPTY,r._onTouched=function(){},r.tabIndex=parseInt(n)||0,r}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"disabled",{get:function(){return this._disabled},set:function(e){this._disabled=Object(s.coerceBooleanProperty)(e),this.options&&this.options.forEach(function(e){return e._markForCheck()})},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){this._keyManager=new a.FocusKeyManager(this.options).withWrap().withTypeAhead().skipPredicate(function(){return!1}).withAllowedModifierKeys(["shiftKey"]),this._tempValues&&(this._setOptionsFromValues(this._tempValues),this._tempValues=null),this._modelChanges=this.selectedOptions.onChange.subscribe(function(e){if(e.added)for(var t=0,n=e.added;t<n.length;t++){n[t].selected=!0}if(e.removed)for(var r=0,i=e.removed;r<i.length;r++){i[r].selected=!1}})},t.prototype.ngOnDestroy=function(){this._modelChanges.unsubscribe()},t.prototype.focus=function(){this._element.nativeElement.focus()},t.prototype.selectAll=function(){this._setAllOptionsSelected(!0)},t.prototype.deselectAll=function(){this._setAllOptionsSelected(!1)},t.prototype._setFocusedOption=function(e){this._keyManager.updateActiveItemIndex(this._getOptionIndex(e))},t.prototype._removeOptionFromList=function(e){var t=this._getOptionIndex(e);return t>-1&&this._keyManager.activeItemIndex===t&&(t>0?this._keyManager.updateActiveItemIndex(t-1):0===t&&this.options.length>1&&this._keyManager.updateActiveItemIndex(Math.min(t+1,this.options.length-1))),this._keyManager.activeItem},t.prototype._keydown=function(e){var t=e.keyCode,n=this._keyManager,r=n.activeItemIndex,i=Object(c.hasModifierKey)(e);switch(t){case c.SPACE:case c.ENTER:i||(this._toggleFocusedOption(),e.preventDefault());break;case c.HOME:case c.END:i||(t===c.HOME?n.setFirstItemActive():n.setLastItemActive(),e.preventDefault());break;case c.A:Object(c.hasModifierKey)(e,"ctrlKey")&&(this.options.find(function(e){return!e.selected})?this.selectAll():this.deselectAll(),e.preventDefault());break;default:n.onKeydown(e)}t!==c.UP_ARROW&&t!==c.DOWN_ARROW||!e.shiftKey||n.activeItemIndex===r||this._toggleFocusedOption()},t.prototype._reportValueChange=function(){this.options&&this._onChange(this._getSelectedOptionValues())},t.prototype._emitChangeEvent=function(e){this.selectionChange.emit(new M(this,e))},t.prototype.writeValue=function(e){this.options?this._setOptionsFromValues(e||[]):this._tempValues=e},t.prototype.setDisabledState=function(e){this.disabled=e},t.prototype.registerOnChange=function(e){this._onChange=e},t.prototype.registerOnTouched=function(e){this._onTouched=e},t.prototype._setOptionsFromValues=function(e){var t=this;this.options.forEach(function(e){return e._setSelected(!1)}),e.forEach(function(e){var n=t.options.find(function(n){return!n.selected&&(t.compareWith?t.compareWith(n.value,e):n.value===e)});n&&n._setSelected(!0)})},t.prototype._getSelectedOptionValues=function(){return this.options.filter(function(e){return e.selected}).map(function(e){return e.value})},t.prototype._toggleFocusedOption=function(){var e=this._keyManager.activeItemIndex;if(null!=e&&this._isValidIndex(e)){var t=this.options.toArray()[e];t&&!t.disabled&&(t.toggle(),this._emitChangeEvent(t))}},t.prototype._setAllOptionsSelected=function(e){var t=!1;this.options.forEach(function(n){n._setSelected(e)&&(t=!0)}),t&&this._reportValueChange()},t.prototype._isValidIndex=function(e){return e>=0&&e<this.options.length},t.prototype._getOptionIndex=function(e){return this.options.toArray().indexOf(e)},t.decorators=[{type:i.Component,args:[{selector:"mat-selection-list",exportAs:"matSelectionList",inputs:["disabled","disableRipple","tabIndex"],host:{role:"listbox","[tabIndex]":"tabIndex",class:"mat-selection-list mat-list-base","(focus)":"focus()","(blur)":"_onTouched()","(keydown)":"_keydown($event)","aria-multiselectable":"true","[attr.aria-disabled]":"disabled.toString()"},template:"<ng-content></ng-content>",styles:[".mat-subheader{display:flex;box-sizing:border-box;padding:16px;align-items:center}.mat-list-base .mat-subheader{margin:0}.mat-list-base{padding-top:8px;display:block;-webkit-tap-highlight-color:transparent}.mat-list-base .mat-subheader{height:48px;line-height:16px}.mat-list-base .mat-subheader:first-child{margin-top:-8px}.mat-list-base .mat-list-item,.mat-list-base .mat-list-option{display:block;height:48px;-webkit-tap-highlight-color:transparent;width:100%;padding:0}.mat-list-base .mat-list-item .mat-list-item-content,.mat-list-base .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list-base .mat-list-item .mat-list-item-content-reverse,.mat-list-base .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list-base .mat-list-item .mat-list-item-ripple,.mat-list-base .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list-base .mat-list-item.mat-list-item-with-avatar,.mat-list-base .mat-list-option.mat-list-item-with-avatar{height:56px}.mat-list-base .mat-list-item.mat-2-line,.mat-list-base .mat-list-option.mat-2-line{height:72px}.mat-list-base .mat-list-item.mat-3-line,.mat-list-base .mat-list-option.mat-3-line{height:88px}.mat-list-base .mat-list-item.mat-multi-line,.mat-list-base .mat-list-option.mat-multi-line{height:auto}.mat-list-base .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list-base .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list-base .mat-list-item .mat-list-text,.mat-list-base .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0}.mat-list-base .mat-list-item .mat-list-text>*,.mat-list-base .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list-base .mat-list-item .mat-list-text:empty,.mat-list-base .mat-list-option .mat-list-text:empty{display:none}.mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:0;padding-left:16px}[dir=rtl] .mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:16px;padding-left:0}.mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-left:0;padding-right:16px}[dir=rtl] .mat-list-base .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:0;padding-left:16px}.mat-list-base .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:16px;padding-left:16px}.mat-list-base .mat-list-item .mat-list-avatar,.mat-list-base .mat-list-option .mat-list-avatar{flex-shrink:0;width:40px;height:40px;border-radius:50%;object-fit:cover}.mat-list-base .mat-list-item .mat-list-avatar~.mat-divider-inset,.mat-list-base .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:72px;width:calc(100% - 72px)}[dir=rtl] .mat-list-base .mat-list-item .mat-list-avatar~.mat-divider-inset,[dir=rtl] .mat-list-base .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:auto;margin-right:72px}.mat-list-base .mat-list-item .mat-list-icon,.mat-list-base .mat-list-option .mat-list-icon{flex-shrink:0;width:24px;height:24px;font-size:24px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-list-base .mat-list-item .mat-list-icon~.mat-divider-inset,.mat-list-base .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:64px;width:calc(100% - 64px)}[dir=rtl] .mat-list-base .mat-list-item .mat-list-icon~.mat-divider-inset,[dir=rtl] .mat-list-base .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:auto;margin-right:64px}.mat-list-base .mat-list-item .mat-divider,.mat-list-base .mat-list-option .mat-divider{position:absolute;bottom:0;left:0;width:100%;margin:0}[dir=rtl] .mat-list-base .mat-list-item .mat-divider,[dir=rtl] .mat-list-base .mat-list-option .mat-divider{margin-left:auto;margin-right:0}.mat-list-base .mat-list-item .mat-divider.mat-divider-inset,.mat-list-base .mat-list-option .mat-divider.mat-divider-inset{position:absolute}.mat-list-base[dense]{padding-top:4px;display:block}.mat-list-base[dense] .mat-subheader{height:40px;line-height:8px}.mat-list-base[dense] .mat-subheader:first-child{margin-top:-4px}.mat-list-base[dense] .mat-list-item,.mat-list-base[dense] .mat-list-option{display:block;height:40px;-webkit-tap-highlight-color:transparent;width:100%;padding:0}.mat-list-base[dense] .mat-list-item .mat-list-item-content,.mat-list-base[dense] .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list-base[dense] .mat-list-item .mat-list-item-content-reverse,.mat-list-base[dense] .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list-base[dense] .mat-list-item .mat-list-item-ripple,.mat-list-base[dense] .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar{height:48px}.mat-list-base[dense] .mat-list-item.mat-2-line,.mat-list-base[dense] .mat-list-option.mat-2-line{height:60px}.mat-list-base[dense] .mat-list-item.mat-3-line,.mat-list-base[dense] .mat-list-option.mat-3-line{height:76px}.mat-list-base[dense] .mat-list-item.mat-multi-line,.mat-list-base[dense] .mat-list-option.mat-multi-line{height:auto}.mat-list-base[dense] .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list-base[dense] .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list-base[dense] .mat-list-item .mat-list-text,.mat-list-base[dense] .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0}.mat-list-base[dense] .mat-list-item .mat-list-text>*,.mat-list-base[dense] .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list-base[dense] .mat-list-item .mat-list-text:empty,.mat-list-base[dense] .mat-list-option .mat-list-text:empty{display:none}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:0;padding-left:16px}[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content .mat-list-text{padding-right:16px;padding-left:0}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-left:0;padding-right:16px}[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-item.mat-list-option .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar .mat-list-item-content-reverse .mat-list-text,[dir=rtl] .mat-list-base[dense] .mat-list-option.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:0;padding-left:16px}.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-item.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content .mat-list-text,.mat-list-base[dense] .mat-list-option.mat-list-item-with-avatar.mat-list-option .mat-list-item-content-reverse .mat-list-text{padding-right:16px;padding-left:16px}.mat-list-base[dense] .mat-list-item .mat-list-avatar,.mat-list-base[dense] .mat-list-option .mat-list-avatar{flex-shrink:0;width:36px;height:36px;border-radius:50%;object-fit:cover}.mat-list-base[dense] .mat-list-item .mat-list-avatar~.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:68px;width:calc(100% - 68px)}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-list-avatar~.mat-divider-inset,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-list-avatar~.mat-divider-inset{margin-left:auto;margin-right:68px}.mat-list-base[dense] .mat-list-item .mat-list-icon,.mat-list-base[dense] .mat-list-option .mat-list-icon{flex-shrink:0;width:20px;height:20px;font-size:20px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-list-base[dense] .mat-list-item .mat-list-icon~.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:60px;width:calc(100% - 60px)}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-list-icon~.mat-divider-inset,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-list-icon~.mat-divider-inset{margin-left:auto;margin-right:60px}.mat-list-base[dense] .mat-list-item .mat-divider,.mat-list-base[dense] .mat-list-option .mat-divider{position:absolute;bottom:0;left:0;width:100%;margin:0}[dir=rtl] .mat-list-base[dense] .mat-list-item .mat-divider,[dir=rtl] .mat-list-base[dense] .mat-list-option .mat-divider{margin-left:auto;margin-right:0}.mat-list-base[dense] .mat-list-item .mat-divider.mat-divider-inset,.mat-list-base[dense] .mat-list-option .mat-divider.mat-divider-inset{position:absolute}.mat-nav-list a{text-decoration:none;color:inherit}.mat-nav-list .mat-list-item{cursor:pointer;outline:0}mat-action-list button{background:0 0;color:inherit;border:none;font:inherit;outline:inherit}mat-action-list .mat-list-item{cursor:pointer;outline:inherit}.mat-list-option:not(.mat-list-item-disabled){cursor:pointer;outline:0}@media (-ms-high-contrast:active){.mat-selection-list:focus{outline-style:dotted}.mat-list-option:focus,.mat-list-option:hover,.mat-nav-list .mat-list-item:focus,.mat-nav-list .mat-list-item:hover,mat-action-list .mat-list-item:focus,mat-action-list .mat-list-item:hover{outline:dotted 1px}}@media (hover:none){.mat-list-option:not(.mat-list-item-disabled):hover,.mat-nav-list .mat-list-item:not(.mat-list-item-disabled):hover{background:0 0}}"],encapsulation:i.ViewEncapsulation.None,providers:[k],changeDetection:i.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:i.ElementRef},{type:String,decorators:[{type:i.Attribute,args:["tabindex"]}]}]},t.propDecorators={options:[{type:i.ContentChildren,args:[A]}],selectionChange:[{type:i.Output}],tabIndex:[{type:i.Input}],compareWith:[{type:i.Input}],disabled:[{type:i.Input}]},t}(C),D=function(){function e(){}return e.decorators=[{type:i.NgModule,args:[{imports:[o.MatLineModule,o.MatRippleModule,o.MatCommonModule,o.MatPseudoCheckboxModule,p.CommonModule],exports:[y,_,O,b,o.MatLineModule,o.MatCommonModule,j,x,o.MatPseudoCheckboxModule,T,A,f.MatDividerModule],declarations:[y,_,O,b,j,x,T,A]}]}],e}()},"./node_modules/@angular/material/esm5/material.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"VERSION",function(){return V});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/material/esm5/autocomplete.es5.js");n.d(t,"ɵa28",function(){return i["ɵa28"]}),n.d(t,"MAT_AUTOCOMPLETE_DEFAULT_OPTIONS_FACTORY",function(){return i.MAT_AUTOCOMPLETE_DEFAULT_OPTIONS_FACTORY}),n.d(t,"MatAutocompleteSelectedEvent",function(){return i.MatAutocompleteSelectedEvent}),n.d(t,"MatAutocompleteBase",function(){return i.MatAutocompleteBase}),n.d(t,"_MatAutocompleteMixinBase",function(){return i._MatAutocompleteMixinBase}),n.d(t,"MAT_AUTOCOMPLETE_DEFAULT_OPTIONS",function(){return i.MAT_AUTOCOMPLETE_DEFAULT_OPTIONS}),n.d(t,"MatAutocomplete",function(){return i.MatAutocomplete}),n.d(t,"MatAutocompleteModule",function(){return i.MatAutocompleteModule}),n.d(t,"MAT_AUTOCOMPLETE_SCROLL_STRATEGY_FACTORY",function(){return i.MAT_AUTOCOMPLETE_SCROLL_STRATEGY_FACTORY}),n.d(t,"getMatAutocompleteMissingPanelError",function(){return i.getMatAutocompleteMissingPanelError}),n.d(t,"AUTOCOMPLETE_OPTION_HEIGHT",function(){return i.AUTOCOMPLETE_OPTION_HEIGHT}),n.d(t,"AUTOCOMPLETE_PANEL_HEIGHT",function(){return i.AUTOCOMPLETE_PANEL_HEIGHT}),n.d(t,"MAT_AUTOCOMPLETE_SCROLL_STRATEGY",function(){return i.MAT_AUTOCOMPLETE_SCROLL_STRATEGY}),n.d(t,"MAT_AUTOCOMPLETE_SCROLL_STRATEGY_FACTORY_PROVIDER",function(){return i.MAT_AUTOCOMPLETE_SCROLL_STRATEGY_FACTORY_PROVIDER}),n.d(t,"MAT_AUTOCOMPLETE_VALUE_ACCESSOR",function(){return i.MAT_AUTOCOMPLETE_VALUE_ACCESSOR}),n.d(t,"MatAutocompleteTrigger",function(){return i.MatAutocompleteTrigger});var o=n("./node_modules/@angular/material/esm5/badge.es5.js");n.d(t,"MatBadgeModule",function(){return o.MatBadgeModule}),n.d(t,"MatBadgeBase",function(){return o.MatBadgeBase}),n.d(t,"_MatBadgeMixinBase",function(){return o._MatBadgeMixinBase}),n.d(t,"MatBadge",function(){return o.MatBadge});var a=n("./node_modules/@angular/material/esm5/bottom-sheet.es5.js");n.d(t,"MatBottomSheetModule",function(){return a.MatBottomSheetModule}),n.d(t,"MAT_BOTTOM_SHEET_DEFAULT_OPTIONS",function(){return a.MAT_BOTTOM_SHEET_DEFAULT_OPTIONS}),n.d(t,"MatBottomSheet",function(){return a.MatBottomSheet}),n.d(t,"MAT_BOTTOM_SHEET_DATA",function(){return a.MAT_BOTTOM_SHEET_DATA}),n.d(t,"MatBottomSheetConfig",function(){return a.MatBottomSheetConfig}),n.d(t,"MatBottomSheetContainer",function(){return a.MatBottomSheetContainer}),n.d(t,"matBottomSheetAnimations",function(){return a.matBottomSheetAnimations}),n.d(t,"MatBottomSheetRef",function(){return a.MatBottomSheetRef});var s=n("./node_modules/@angular/material/esm5/button.es5.js");n.d(t,"MatButtonModule",function(){return s.MatButtonModule}),n.d(t,"MatButtonBase",function(){return s.MatButtonBase}),n.d(t,"_MatButtonMixinBase",function(){return s._MatButtonMixinBase}),n.d(t,"MatButton",function(){return s.MatButton}),n.d(t,"MatAnchor",function(){return s.MatAnchor});var u=n("./node_modules/@angular/material/esm5/button-toggle.es5.js");n.d(t,"MAT_BUTTON_TOGGLE_DEFAULT_OPTIONS",function(){return u.MAT_BUTTON_TOGGLE_DEFAULT_OPTIONS}),n.d(t,"MAT_BUTTON_TOGGLE_GROUP_VALUE_ACCESSOR",function(){return u.MAT_BUTTON_TOGGLE_GROUP_VALUE_ACCESSOR}),n.d(t,"MatButtonToggleGroupMultiple",function(){return u.MatButtonToggleGroupMultiple}),n.d(t,"MatButtonToggleChange",function(){return u.MatButtonToggleChange}),n.d(t,"MatButtonToggleGroup",function(){return u.MatButtonToggleGroup}),n.d(t,"MatButtonToggleBase",function(){return u.MatButtonToggleBase}),n.d(t,"_MatButtonToggleMixinBase",function(){return u._MatButtonToggleMixinBase}),n.d(t,"MatButtonToggle",function(){return u.MatButtonToggle}),n.d(t,"MatButtonToggleModule",function(){return u.MatButtonToggleModule});var c=n("./node_modules/@angular/material/esm5/card.es5.js");n.d(t,"MatCardContent",function(){return c.MatCardContent}),n.d(t,"MatCardTitle",function(){return c.MatCardTitle}),n.d(t,"MatCardSubtitle",function(){return c.MatCardSubtitle}),n.d(t,"MatCardActions",function(){return c.MatCardActions}),n.d(t,"MatCardFooter",function(){return c.MatCardFooter}),n.d(t,"MatCardImage",function(){return c.MatCardImage}),n.d(t,"MatCardSmImage",function(){return c.MatCardSmImage}),n.d(t,"MatCardMdImage",function(){return c.MatCardMdImage}),n.d(t,"MatCardLgImage",function(){return c.MatCardLgImage}),n.d(t,"MatCardXlImage",function(){return c.MatCardXlImage}),n.d(t,"MatCardAvatar",function(){return c.MatCardAvatar}),n.d(t,"MatCard",function(){return c.MatCard}),n.d(t,"MatCardHeader",function(){return c.MatCardHeader}),n.d(t,"MatCardTitleGroup",function(){return c.MatCardTitleGroup}),n.d(t,"MatCardModule",function(){return c.MatCardModule});var l=n("./node_modules/@angular/material/esm5/checkbox.es5.js");n.d(t,"MAT_CHECKBOX_CONTROL_VALUE_ACCESSOR",function(){return l.MAT_CHECKBOX_CONTROL_VALUE_ACCESSOR}),n.d(t,"TransitionCheckState",function(){return l.TransitionCheckState}),n.d(t,"MatCheckboxChange",function(){return l.MatCheckboxChange}),n.d(t,"MatCheckboxBase",function(){return l.MatCheckboxBase}),n.d(t,"_MatCheckboxMixinBase",function(){return l._MatCheckboxMixinBase}),n.d(t,"MatCheckbox",function(){return l.MatCheckbox}),n.d(t,"MAT_CHECKBOX_CLICK_ACTION",function(){return l.MAT_CHECKBOX_CLICK_ACTION}),n.d(t,"MatCheckboxModule",function(){return l.MatCheckboxModule}),n.d(t,"MAT_CHECKBOX_REQUIRED_VALIDATOR",function(){return l.MAT_CHECKBOX_REQUIRED_VALIDATOR}),n.d(t,"MatCheckboxRequiredValidator",function(){return l.MatCheckboxRequiredValidator});var d=n("./node_modules/@angular/material/esm5/chips.es5.js");n.d(t,"MatChipsModule",function(){return d.MatChipsModule}),n.d(t,"MatChipListBase",function(){return d.MatChipListBase}),n.d(t,"_MatChipListMixinBase",function(){return d._MatChipListMixinBase}),n.d(t,"MatChipListChange",function(){return d.MatChipListChange}),n.d(t,"MatChipList",function(){return d.MatChipList}),n.d(t,"MatChipSelectionChange",function(){return d.MatChipSelectionChange}),n.d(t,"MatChipBase",function(){return d.MatChipBase}),n.d(t,"_MatChipMixinBase",function(){return d._MatChipMixinBase}),n.d(t,"MatChipAvatar",function(){return d.MatChipAvatar}),n.d(t,"MatChipTrailingIcon",function(){return d.MatChipTrailingIcon}),n.d(t,"MatChip",function(){return d.MatChip}),n.d(t,"MatChipRemove",function(){return d.MatChipRemove}),n.d(t,"MatChipInput",function(){return d.MatChipInput}),n.d(t,"MAT_CHIPS_DEFAULT_OPTIONS",function(){return d.MAT_CHIPS_DEFAULT_OPTIONS});var p=n("./node_modules/@angular/material/esm5/core.es5.js");n.d(t,"ɵa1",function(){return p["ɵa1"]}),n.d(t,"AnimationCurves",function(){return p.AnimationCurves}),n.d(t,"AnimationDurations",function(){return p.AnimationDurations}),n.d(t,"MatCommonModule",function(){return p.MatCommonModule}),n.d(t,"MATERIAL_SANITY_CHECKS",function(){return p.MATERIAL_SANITY_CHECKS}),n.d(t,"mixinDisabled",function(){return p.mixinDisabled}),n.d(t,"mixinColor",function(){return p.mixinColor}),n.d(t,"mixinDisableRipple",function(){return p.mixinDisableRipple}),n.d(t,"mixinTabIndex",function(){return p.mixinTabIndex}),n.d(t,"mixinErrorState",function(){return p.mixinErrorState}),n.d(t,"mixinInitialized",function(){return p.mixinInitialized}),n.d(t,"NativeDateModule",function(){return p.NativeDateModule}),n.d(t,"MatNativeDateModule",function(){return p.MatNativeDateModule}),n.d(t,"MAT_DATE_LOCALE_FACTORY",function(){return p.MAT_DATE_LOCALE_FACTORY}),n.d(t,"MAT_DATE_LOCALE",function(){return p.MAT_DATE_LOCALE}),n.d(t,"MAT_DATE_LOCALE_PROVIDER",function(){return p.MAT_DATE_LOCALE_PROVIDER}),n.d(t,"DateAdapter",function(){return p.DateAdapter}),n.d(t,"MAT_DATE_FORMATS",function(){return p.MAT_DATE_FORMATS}),n.d(t,"NativeDateAdapter",function(){return p.NativeDateAdapter}),n.d(t,"MAT_NATIVE_DATE_FORMATS",function(){return p.MAT_NATIVE_DATE_FORMATS}),n.d(t,"ShowOnDirtyErrorStateMatcher",function(){return p.ShowOnDirtyErrorStateMatcher}),n.d(t,"ErrorStateMatcher",function(){return p.ErrorStateMatcher}),n.d(t,"MAT_HAMMER_OPTIONS",function(){return p.MAT_HAMMER_OPTIONS}),n.d(t,"GestureConfig",function(){return p.GestureConfig}),n.d(t,"setLines",function(){return p.setLines}),n.d(t,"MatLine",function(){return p.MatLine}),n.d(t,"MatLineSetter",function(){return p.MatLineSetter}),n.d(t,"MatLineModule",function(){return p.MatLineModule}),n.d(t,"MatOptionModule",function(){return p.MatOptionModule}),n.d(t,"_countGroupLabelsBeforeOption",function(){return p._countGroupLabelsBeforeOption}),n.d(t,"_getOptionScrollPosition",function(){return p._getOptionScrollPosition}),n.d(t,"MatOptionSelectionChange",function(){return p.MatOptionSelectionChange}),n.d(t,"MAT_OPTION_PARENT_COMPONENT",function(){return p.MAT_OPTION_PARENT_COMPONENT}),n.d(t,"MatOption",function(){return p.MatOption}),n.d(t,"MatOptgroupBase",function(){return p.MatOptgroupBase}),n.d(t,"_MatOptgroupMixinBase",function(){return p._MatOptgroupMixinBase}),n.d(t,"MatOptgroup",function(){return p.MatOptgroup}),n.d(t,"MAT_LABEL_GLOBAL_OPTIONS",function(){return p.MAT_LABEL_GLOBAL_OPTIONS}),n.d(t,"MatRippleModule",function(){return p.MatRippleModule}),n.d(t,"MAT_RIPPLE_GLOBAL_OPTIONS",function(){return p.MAT_RIPPLE_GLOBAL_OPTIONS}),n.d(t,"MatRipple",function(){return p.MatRipple}),n.d(t,"RippleState",function(){return p.RippleState}),n.d(t,"RippleRef",function(){return p.RippleRef}),n.d(t,"defaultRippleAnimationConfig",function(){return p.defaultRippleAnimationConfig}),n.d(t,"RippleRenderer",function(){return p.RippleRenderer}),n.d(t,"MatPseudoCheckboxModule",function(){return p.MatPseudoCheckboxModule}),n.d(t,"MatPseudoCheckbox",function(){return p.MatPseudoCheckbox}),n.d(t,"JAN",function(){return p.JAN}),n.d(t,"FEB",function(){return p.FEB}),n.d(t,"MAR",function(){return p.MAR}),n.d(t,"APR",function(){return p.APR}),n.d(t,"MAY",function(){return p.MAY}),n.d(t,"JUN",function(){return p.JUN}),n.d(t,"JUL",function(){return p.JUL}),n.d(t,"AUG",function(){return p.AUG}),n.d(t,"SEP",function(){return p.SEP}),n.d(t,"OCT",function(){return p.OCT}),n.d(t,"NOV",function(){return p.NOV}),n.d(t,"DEC",function(){return p.DEC});var f=n("./node_modules/@angular/material/esm5/datepicker.es5.js");n.d(t,"ɵa34",function(){return f["ɵa34"]}),n.d(t,"MatDatepickerModule",function(){return f.MatDatepickerModule}),n.d(t,"MatCalendarHeader",function(){return f.MatCalendarHeader}),n.d(t,"MatCalendar",function(){return f.MatCalendar}),n.d(t,"MatCalendarCell",function(){return f.MatCalendarCell}),n.d(t,"MatCalendarBody",function(){return f.MatCalendarBody}),n.d(t,"MAT_DATEPICKER_SCROLL_STRATEGY_FACTORY",function(){return f.MAT_DATEPICKER_SCROLL_STRATEGY_FACTORY}),n.d(t,"MAT_DATEPICKER_SCROLL_STRATEGY",function(){return f.MAT_DATEPICKER_SCROLL_STRATEGY}),n.d(t,"MAT_DATEPICKER_SCROLL_STRATEGY_FACTORY_PROVIDER",function(){return f.MAT_DATEPICKER_SCROLL_STRATEGY_FACTORY_PROVIDER}),n.d(t,"MatDatepickerContentBase",function(){return f.MatDatepickerContentBase}),n.d(t,"_MatDatepickerContentMixinBase",function(){return f._MatDatepickerContentMixinBase}),n.d(t,"MatDatepickerContent",function(){return f.MatDatepickerContent}),n.d(t,"MatDatepicker",function(){return f.MatDatepicker}),n.d(t,"matDatepickerAnimations",function(){return f.matDatepickerAnimations}),n.d(t,"MAT_DATEPICKER_VALUE_ACCESSOR",function(){return f.MAT_DATEPICKER_VALUE_ACCESSOR}),n.d(t,"MAT_DATEPICKER_VALIDATORS",function(){return f.MAT_DATEPICKER_VALIDATORS}),n.d(t,"MatDatepickerInputEvent",function(){return f.MatDatepickerInputEvent}),n.d(t,"MatDatepickerInput",function(){return f.MatDatepickerInput}),n.d(t,"MatDatepickerIntl",function(){return f.MatDatepickerIntl}),n.d(t,"MatDatepickerToggleIcon",function(){return f.MatDatepickerToggleIcon}),n.d(t,"MatDatepickerToggle",function(){return f.MatDatepickerToggle}),n.d(t,"MatMonthView",function(){return f.MatMonthView}),n.d(t,"MatYearView",function(){return f.MatYearView});var h=n("./node_modules/@angular/material/esm5/dialog.es5.js");n.d(t,"MatDialogModule",function(){return h.MatDialogModule}),n.d(t,"MAT_DIALOG_SCROLL_STRATEGY_FACTORY",function(){return h.MAT_DIALOG_SCROLL_STRATEGY_FACTORY}),n.d(t,"MAT_DIALOG_SCROLL_STRATEGY_PROVIDER_FACTORY",function(){return h.MAT_DIALOG_SCROLL_STRATEGY_PROVIDER_FACTORY}),n.d(t,"MAT_DIALOG_DATA",function(){return h.MAT_DIALOG_DATA}),n.d(t,"MAT_DIALOG_DEFAULT_OPTIONS",function(){return h.MAT_DIALOG_DEFAULT_OPTIONS}),n.d(t,"MAT_DIALOG_SCROLL_STRATEGY",function(){return h.MAT_DIALOG_SCROLL_STRATEGY}),n.d(t,"MAT_DIALOG_SCROLL_STRATEGY_PROVIDER",function(){return h.MAT_DIALOG_SCROLL_STRATEGY_PROVIDER}),n.d(t,"MatDialog",function(){return h.MatDialog}),n.d(t,"throwMatDialogContentAlreadyAttachedError",function(){return h.throwMatDialogContentAlreadyAttachedError}),n.d(t,"MatDialogContainer",function(){return h.MatDialogContainer}),n.d(t,"MatDialogClose",function(){return h.MatDialogClose}),n.d(t,"MatDialogTitle",function(){return h.MatDialogTitle}),n.d(t,"MatDialogContent",function(){return h.MatDialogContent}),n.d(t,"MatDialogActions",function(){return h.MatDialogActions}),n.d(t,"MatDialogConfig",function(){return h.MatDialogConfig}),n.d(t,"MatDialogRef",function(){return h.MatDialogRef}),n.d(t,"matDialogAnimations",function(){return h.matDialogAnimations});var m=n("./node_modules/@angular/material/esm5/divider.es5.js");n.d(t,"MatDivider",function(){return m.MatDivider}),n.d(t,"MatDividerModule",function(){return m.MatDividerModule});var g=n("./node_modules/@angular/material/esm5/expansion.es5.js");n.d(t,"MatExpansionModule",function(){return g.MatExpansionModule}),n.d(t,"MatAccordion",function(){return g.MatAccordion}),n.d(t,"MAT_ACCORDION",function(){return g.MAT_ACCORDION}),n.d(t,"MAT_EXPANSION_PANEL_DEFAULT_OPTIONS",function(){return g.MAT_EXPANSION_PANEL_DEFAULT_OPTIONS}),n.d(t,"MatExpansionPanel",function(){return g.MatExpansionPanel}),n.d(t,"MatExpansionPanelActionRow",function(){return g.MatExpansionPanelActionRow}),n.d(t,"MatExpansionPanelHeader",function(){return g.MatExpansionPanelHeader}),n.d(t,"MatExpansionPanelDescription",function(){return g.MatExpansionPanelDescription}),n.d(t,"MatExpansionPanelTitle",function(){return g.MatExpansionPanelTitle}),n.d(t,"MatExpansionPanelContent",function(){return g.MatExpansionPanelContent}),n.d(t,"EXPANSION_PANEL_ANIMATION_TIMING",function(){return g.EXPANSION_PANEL_ANIMATION_TIMING}),n.d(t,"matExpansionAnimations",function(){return g.matExpansionAnimations});var v=n("./node_modules/@angular/material/esm5/form-field.es5.js");n.d(t,"MatFormFieldModule",function(){return v.MatFormFieldModule}),n.d(t,"MatError",function(){return v.MatError}),n.d(t,"MatFormFieldBase",function(){return v.MatFormFieldBase}),n.d(t,"_MatFormFieldMixinBase",function(){return v._MatFormFieldMixinBase}),n.d(t,"MAT_FORM_FIELD_DEFAULT_OPTIONS",function(){return v.MAT_FORM_FIELD_DEFAULT_OPTIONS}),n.d(t,"MatFormField",function(){return v.MatFormField}),n.d(t,"MatFormFieldControl",function(){return v.MatFormFieldControl}),n.d(t,"getMatFormFieldPlaceholderConflictError",function(){return v.getMatFormFieldPlaceholderConflictError}),n.d(t,"getMatFormFieldDuplicatedHintError",function(){return v.getMatFormFieldDuplicatedHintError}),n.d(t,"getMatFormFieldMissingControlError",function(){return v.getMatFormFieldMissingControlError}),n.d(t,"MatHint",function(){return v.MatHint}),n.d(t,"MatPlaceholder",function(){return v.MatPlaceholder}),n.d(t,"MatPrefix",function(){return v.MatPrefix}),n.d(t,"MatSuffix",function(){return v.MatSuffix}),n.d(t,"MatLabel",function(){return v.MatLabel}),n.d(t,"matFormFieldAnimations",function(){return v.matFormFieldAnimations});var _=n("./node_modules/@angular/material/esm5/grid-list.es5.js");n.d(t,"ɵa15",function(){return _["ɵa15"]}),n.d(t,"MatGridListModule",function(){return _.MatGridListModule}),n.d(t,"MatGridList",function(){return _.MatGridList}),n.d(t,"MatGridTile",function(){return _.MatGridTile}),n.d(t,"MatGridTileText",function(){return _.MatGridTileText}),n.d(t,"MatGridAvatarCssMatStyler",function(){return _.MatGridAvatarCssMatStyler}),n.d(t,"MatGridTileHeaderCssMatStyler",function(){return _.MatGridTileHeaderCssMatStyler}),n.d(t,"MatGridTileFooterCssMatStyler",function(){return _.MatGridTileFooterCssMatStyler});var y=n("./node_modules/@angular/material/esm5/icon.es5.js");n.d(t,"MatIconModule",function(){return y.MatIconModule}),n.d(t,"MAT_ICON_LOCATION_FACTORY",function(){return y.MAT_ICON_LOCATION_FACTORY}),n.d(t,"MatIconBase",function(){return y.MatIconBase}),n.d(t,"_MatIconMixinBase",function(){return y._MatIconMixinBase}),n.d(t,"MAT_ICON_LOCATION",function(){return y.MAT_ICON_LOCATION}),n.d(t,"MatIcon",function(){return y.MatIcon}),n.d(t,"getMatIconNameNotFoundError",function(){return y.getMatIconNameNotFoundError}),n.d(t,"getMatIconNoHttpProviderError",function(){return y.getMatIconNoHttpProviderError}),n.d(t,"getMatIconFailedToSanitizeUrlError",function(){return y.getMatIconFailedToSanitizeUrlError}),n.d(t,"getMatIconFailedToSanitizeLiteralError",function(){return y.getMatIconFailedToSanitizeLiteralError}),n.d(t,"ICON_REGISTRY_PROVIDER_FACTORY",function(){return y.ICON_REGISTRY_PROVIDER_FACTORY}),n.d(t,"MatIconRegistry",function(){return y.MatIconRegistry}),n.d(t,"ICON_REGISTRY_PROVIDER",function(){return y.ICON_REGISTRY_PROVIDER});var b=n("./node_modules/@angular/material/esm5/input.es5.js");n.d(t,"MatTextareaAutosize",function(){return b.MatTextareaAutosize}),n.d(t,"MatInputBase",function(){return b.MatInputBase}),n.d(t,"_MatInputMixinBase",function(){return b._MatInputMixinBase}),n.d(t,"MatInput",function(){return b.MatInput}),n.d(t,"getMatInputUnsupportedTypeError",function(){return b.getMatInputUnsupportedTypeError}),n.d(t,"MatInputModule",function(){return b.MatInputModule}),n.d(t,"MAT_INPUT_VALUE_ACCESSOR",function(){return b.MAT_INPUT_VALUE_ACCESSOR});var j=n("./node_modules/@angular/material/esm5/list.es5.js");n.d(t,"MatListModule",function(){return j.MatListModule}),n.d(t,"MatListBase",function(){return j.MatListBase}),n.d(t,"_MatListMixinBase",function(){return j._MatListMixinBase}),n.d(t,"MatListItemBase",function(){return j.MatListItemBase}),n.d(t,"_MatListItemMixinBase",function(){return j._MatListItemMixinBase}),n.d(t,"MatNavList",function(){return j.MatNavList}),n.d(t,"MatList",function(){return j.MatList}),n.d(t,"MatListAvatarCssMatStyler",function(){return j.MatListAvatarCssMatStyler}),n.d(t,"MatListIconCssMatStyler",function(){return j.MatListIconCssMatStyler}),n.d(t,"MatListSubheaderCssMatStyler",function(){return j.MatListSubheaderCssMatStyler}),n.d(t,"MatListItem",function(){return j.MatListItem}),n.d(t,"MatSelectionListBase",function(){return j.MatSelectionListBase}),n.d(t,"_MatSelectionListMixinBase",function(){return j._MatSelectionListMixinBase}),n.d(t,"MatListOptionBase",function(){return j.MatListOptionBase}),n.d(t,"_MatListOptionMixinBase",function(){return j._MatListOptionMixinBase}),n.d(t,"MAT_SELECTION_LIST_VALUE_ACCESSOR",function(){return j.MAT_SELECTION_LIST_VALUE_ACCESSOR}),n.d(t,"MatSelectionListChange",function(){return j.MatSelectionListChange}),n.d(t,"MatListOption",function(){return j.MatListOption}),n.d(t,"MatSelectionList",function(){return j.MatSelectionList});var x=n("./node_modules/@angular/material/esm5/menu.es5.js");n.d(t,"ɵa23",function(){return x["ɵa23"]}),n.d(t,"ɵb23",function(){return x["ɵb23"]}),n.d(t,"ɵc23",function(){return x["ɵc23"]}),n.d(t,"ɵf23",function(){return x["ɵf23"]}),n.d(t,"ɵd23",function(){return x["ɵd23"]}),n.d(t,"ɵe23",function(){return x["ɵe23"]}),n.d(t,"MAT_MENU_SCROLL_STRATEGY",function(){return x.MAT_MENU_SCROLL_STRATEGY}),n.d(t,"MatMenuModule",function(){return x.MatMenuModule}),n.d(t,"MatMenu",function(){return x.MatMenu}),n.d(t,"MAT_MENU_DEFAULT_OPTIONS",function(){return x.MAT_MENU_DEFAULT_OPTIONS}),n.d(t,"MatMenuItem",function(){return x.MatMenuItem}),n.d(t,"MatMenuTrigger",function(){return x.MatMenuTrigger}),n.d(t,"matMenuAnimations",function(){return x.matMenuAnimations}),n.d(t,"fadeInItems",function(){return x.fadeInItems}),n.d(t,"transformMenu",function(){return x.transformMenu}),n.d(t,"MatMenuContent",function(){return x.MatMenuContent});var O=n("./node_modules/@angular/material/esm5/paginator.es5.js");n.d(t,"MatPaginatorModule",function(){return O.MatPaginatorModule}),n.d(t,"PageEvent",function(){return O.PageEvent}),n.d(t,"MatPaginatorBase",function(){return O.MatPaginatorBase}),n.d(t,"_MatPaginatorBase",function(){return O._MatPaginatorBase}),n.d(t,"MatPaginator",function(){return O.MatPaginator}),n.d(t,"MAT_PAGINATOR_INTL_PROVIDER_FACTORY",function(){return O.MAT_PAGINATOR_INTL_PROVIDER_FACTORY}),n.d(t,"MatPaginatorIntl",function(){return O.MatPaginatorIntl}),n.d(t,"MAT_PAGINATOR_INTL_PROVIDER",function(){return O.MAT_PAGINATOR_INTL_PROVIDER});var w=n("./node_modules/@angular/material/esm5/progress-bar.es5.js");n.d(t,"MatProgressBarModule",function(){return w.MatProgressBarModule}),n.d(t,"MAT_PROGRESS_BAR_LOCATION_FACTORY",function(){return w.MAT_PROGRESS_BAR_LOCATION_FACTORY}),n.d(t,"MatProgressBarBase",function(){return w.MatProgressBarBase}),n.d(t,"_MatProgressBarMixinBase",function(){return w._MatProgressBarMixinBase}),n.d(t,"MAT_PROGRESS_BAR_LOCATION",function(){return w.MAT_PROGRESS_BAR_LOCATION}),n.d(t,"MatProgressBar",function(){return w.MatProgressBar});var C=n("./node_modules/@angular/material/esm5/progress-spinner.es5.js");n.d(t,"MatProgressSpinnerModule",function(){return C.MatProgressSpinnerModule}),n.d(t,"MAT_PROGRESS_SPINNER_DEFAULT_OPTIONS_FACTORY",function(){return C.MAT_PROGRESS_SPINNER_DEFAULT_OPTIONS_FACTORY}),n.d(t,"MatProgressSpinnerBase",function(){return C.MatProgressSpinnerBase}),n.d(t,"_MatProgressSpinnerMixinBase",function(){return C._MatProgressSpinnerMixinBase}),n.d(t,"MAT_PROGRESS_SPINNER_DEFAULT_OPTIONS",function(){return C.MAT_PROGRESS_SPINNER_DEFAULT_OPTIONS}),n.d(t,"MatProgressSpinner",function(){return C.MatProgressSpinner}),n.d(t,"MatSpinner",function(){return C.MatSpinner});var E=n("./node_modules/@angular/material/esm5/radio.es5.js");n.d(t,"MatRadioModule",function(){return E.MatRadioModule}),n.d(t,"MAT_RADIO_GROUP_CONTROL_VALUE_ACCESSOR",function(){return E.MAT_RADIO_GROUP_CONTROL_VALUE_ACCESSOR}),n.d(t,"MatRadioChange",function(){return E.MatRadioChange}),n.d(t,"MatRadioGroup",function(){return E.MatRadioGroup}),n.d(t,"MatRadioButtonBase",function(){return E.MatRadioButtonBase}),n.d(t,"_MatRadioButtonMixinBase",function(){return E._MatRadioButtonMixinBase}),n.d(t,"MatRadioButton",function(){return E.MatRadioButton});var S=n("./node_modules/@angular/material/esm5/select.es5.js");n.d(t,"MatSelectModule",function(){return S.MatSelectModule}),n.d(t,"MAT_SELECT_SCROLL_STRATEGY_PROVIDER_FACTORY",function(){return S.MAT_SELECT_SCROLL_STRATEGY_PROVIDER_FACTORY}),n.d(t,"SELECT_PANEL_MAX_HEIGHT",function(){return S.SELECT_PANEL_MAX_HEIGHT}),n.d(t,"SELECT_PANEL_PADDING_X",function(){return S.SELECT_PANEL_PADDING_X}),n.d(t,"SELECT_PANEL_INDENT_PADDING_X",function(){return S.SELECT_PANEL_INDENT_PADDING_X}),n.d(t,"SELECT_ITEM_HEIGHT_EM",function(){return S.SELECT_ITEM_HEIGHT_EM}),n.d(t,"SELECT_MULTIPLE_PANEL_PADDING_X",function(){return S.SELECT_MULTIPLE_PANEL_PADDING_X}),n.d(t,"SELECT_PANEL_VIEWPORT_PADDING",function(){return S.SELECT_PANEL_VIEWPORT_PADDING}),n.d(t,"MAT_SELECT_SCROLL_STRATEGY",function(){return S.MAT_SELECT_SCROLL_STRATEGY}),n.d(t,"MAT_SELECT_SCROLL_STRATEGY_PROVIDER",function(){return S.MAT_SELECT_SCROLL_STRATEGY_PROVIDER}),n.d(t,"MatSelectChange",function(){return S.MatSelectChange}),n.d(t,"MatSelectBase",function(){return S.MatSelectBase}),n.d(t,"_MatSelectMixinBase",function(){return S._MatSelectMixinBase}),n.d(t,"MatSelectTrigger",function(){return S.MatSelectTrigger}),n.d(t,"MatSelect",function(){return S.MatSelect}),n.d(t,"matSelectAnimations",function(){return S.matSelectAnimations}),n.d(t,"transformPanel",function(){return S.transformPanel}),n.d(t,"fadeInContent",function(){return S.fadeInContent});var k=n("./node_modules/@angular/material/esm5/sidenav.es5.js");n.d(t,"MatSidenavModule",function(){return k.MatSidenavModule}),n.d(t,"throwMatDuplicatedDrawerError",function(){return k.throwMatDuplicatedDrawerError}),n.d(t,"MAT_DRAWER_DEFAULT_AUTOSIZE_FACTORY",function(){return k.MAT_DRAWER_DEFAULT_AUTOSIZE_FACTORY}),n.d(t,"MAT_DRAWER_DEFAULT_AUTOSIZE",function(){return k.MAT_DRAWER_DEFAULT_AUTOSIZE}),n.d(t,"MatDrawerContent",function(){return k.MatDrawerContent}),n.d(t,"MatDrawer",function(){return k.MatDrawer}),n.d(t,"MatDrawerContainer",function(){return k.MatDrawerContainer}),n.d(t,"MatSidenavContent",function(){return k.MatSidenavContent}),n.d(t,"MatSidenav",function(){return k.MatSidenav}),n.d(t,"MatSidenavContainer",function(){return k.MatSidenavContainer}),n.d(t,"matDrawerAnimations",function(){return k.matDrawerAnimations});var M=n("./node_modules/@angular/material/esm5/slide-toggle.es5.js");n.d(t,"MatSlideToggleModule",function(){return M.MatSlideToggleModule}),n.d(t,"MAT_SLIDE_TOGGLE_VALUE_ACCESSOR",function(){return M.MAT_SLIDE_TOGGLE_VALUE_ACCESSOR}),n.d(t,"MatSlideToggleChange",function(){return M.MatSlideToggleChange}),n.d(t,"MatSlideToggleBase",function(){return M.MatSlideToggleBase}),n.d(t,"_MatSlideToggleMixinBase",function(){return M._MatSlideToggleMixinBase}),n.d(t,"MatSlideToggle",function(){return M.MatSlideToggle}),n.d(t,"MAT_SLIDE_TOGGLE_DEFAULT_OPTIONS",function(){return M.MAT_SLIDE_TOGGLE_DEFAULT_OPTIONS});var A=n("./node_modules/@angular/material/esm5/slider.es5.js");n.d(t,"MatSliderModule",function(){return A.MatSliderModule}),n.d(t,"MAT_SLIDER_VALUE_ACCESSOR",function(){return A.MAT_SLIDER_VALUE_ACCESSOR}),n.d(t,"MatSliderChange",function(){return A.MatSliderChange}),n.d(t,"MatSliderBase",function(){return A.MatSliderBase}),n.d(t,"_MatSliderMixinBase",function(){return A._MatSliderMixinBase}),n.d(t,"MatSlider",function(){return A.MatSlider});var T=n("./node_modules/@angular/material/esm5/snack-bar.es5.js");n.d(t,"MatSnackBarModule",function(){return T.MatSnackBarModule}),n.d(t,"MAT_SNACK_BAR_DEFAULT_OPTIONS_FACTORY",function(){return T.MAT_SNACK_BAR_DEFAULT_OPTIONS_FACTORY}),n.d(t,"MAT_SNACK_BAR_DEFAULT_OPTIONS",function(){return T.MAT_SNACK_BAR_DEFAULT_OPTIONS}),n.d(t,"MatSnackBar",function(){return T.MatSnackBar}),n.d(t,"MatSnackBarContainer",function(){return T.MatSnackBarContainer}),n.d(t,"MAT_SNACK_BAR_DATA",function(){return T.MAT_SNACK_BAR_DATA}),n.d(t,"MatSnackBarConfig",function(){return T.MatSnackBarConfig}),n.d(t,"MatSnackBarRef",function(){return T.MatSnackBarRef}),n.d(t,"SimpleSnackBar",function(){return T.SimpleSnackBar}),n.d(t,"matSnackBarAnimations",function(){return T.matSnackBarAnimations});var D=n("./node_modules/@angular/material/esm5/sort.es5.js");n.d(t,"MatSortModule",function(){return D.MatSortModule}),n.d(t,"MatSortHeaderBase",function(){return D.MatSortHeaderBase}),n.d(t,"_MatSortHeaderMixinBase",function(){return D._MatSortHeaderMixinBase}),n.d(t,"MatSortHeader",function(){return D.MatSortHeader}),n.d(t,"MAT_SORT_HEADER_INTL_PROVIDER_FACTORY",function(){return D.MAT_SORT_HEADER_INTL_PROVIDER_FACTORY}),n.d(t,"MatSortHeaderIntl",function(){return D.MatSortHeaderIntl}),n.d(t,"MAT_SORT_HEADER_INTL_PROVIDER",function(){return D.MAT_SORT_HEADER_INTL_PROVIDER}),n.d(t,"MatSortBase",function(){return D.MatSortBase}),n.d(t,"_MatSortMixinBase",function(){return D._MatSortMixinBase}),n.d(t,"MatSort",function(){return D.MatSort}),n.d(t,"matSortAnimations",function(){return D.matSortAnimations});var I=n("./node_modules/@angular/material/esm5/stepper.es5.js");n.d(t,"MatStepperModule",function(){return I.MatStepperModule}),n.d(t,"MatStepLabel",function(){return I.MatStepLabel}),n.d(t,"MatStep",function(){return I.MatStep}),n.d(t,"MatStepper",function(){return I.MatStepper}),n.d(t,"MatHorizontalStepper",function(){return I.MatHorizontalStepper}),n.d(t,"MatVerticalStepper",function(){return I.MatVerticalStepper}),n.d(t,"MatStepperNext",function(){return I.MatStepperNext}),n.d(t,"MatStepperPrevious",function(){return I.MatStepperPrevious}),n.d(t,"MatStepHeader",function(){return I.MatStepHeader}),n.d(t,"MAT_STEPPER_INTL_PROVIDER_FACTORY",function(){return I.MAT_STEPPER_INTL_PROVIDER_FACTORY}),n.d(t,"MatStepperIntl",function(){return I.MatStepperIntl}),n.d(t,"MAT_STEPPER_INTL_PROVIDER",function(){return I.MAT_STEPPER_INTL_PROVIDER}),n.d(t,"matStepperAnimations",function(){return I.matStepperAnimations}),n.d(t,"MatStepperIcon",function(){return I.MatStepperIcon});var P=n("./node_modules/@angular/material/esm5/table.es5.js");n.d(t,"MatTableModule",function(){return P.MatTableModule}),n.d(t,"MatCellDef",function(){return P.MatCellDef}),n.d(t,"MatHeaderCellDef",function(){return P.MatHeaderCellDef}),n.d(t,"MatFooterCellDef",function(){return P.MatFooterCellDef}),n.d(t,"MatColumnDef",function(){return P.MatColumnDef}),n.d(t,"MatHeaderCell",function(){return P.MatHeaderCell}),n.d(t,"MatFooterCell",function(){return P.MatFooterCell}),n.d(t,"MatCell",function(){return P.MatCell}),n.d(t,"MatTable",function(){return P.MatTable}),n.d(t,"MatHeaderRowDef",function(){return P.MatHeaderRowDef}),n.d(t,"MatFooterRowDef",function(){return P.MatFooterRowDef}),n.d(t,"MatRowDef",function(){return P.MatRowDef}),n.d(t,"MatHeaderRow",function(){return P.MatHeaderRow}),n.d(t,"MatFooterRow",function(){return P.MatFooterRow}),n.d(t,"MatRow",function(){return P.MatRow}),n.d(t,"MatTableDataSource",function(){return P.MatTableDataSource});var R=n("./node_modules/@angular/material/esm5/tabs.es5.js");n.d(t,"ɵa17",function(){return R["ɵa17"]}),n.d(t,"ɵf17",function(){return R["ɵf17"]}),n.d(t,"ɵg17",function(){return R["ɵg17"]}),n.d(t,"ɵb17",function(){return R["ɵb17"]}),n.d(t,"ɵc17",function(){return R["ɵc17"]}),n.d(t,"ɵd17",function(){return R["ɵd17"]}),n.d(t,"ɵe17",function(){return R["ɵe17"]}),n.d(t,"ɵj17",function(){return R["ɵj17"]}),n.d(t,"ɵh17",function(){return R["ɵh17"]}),n.d(t,"ɵk17",function(){return R["ɵk17"]}),n.d(t,"ɵi17",function(){return R["ɵi17"]}),n.d(t,"MatInkBar",function(){return R.MatInkBar}),n.d(t,"_MAT_INK_BAR_POSITIONER",function(){return R._MAT_INK_BAR_POSITIONER}),n.d(t,"MatTabBody",function(){return R.MatTabBody}),n.d(t,"MatTabBodyPortal",function(){return R.MatTabBodyPortal}),n.d(t,"MatTabHeader",function(){return R.MatTabHeader}),n.d(t,"MatTabLabelWrapper",function(){return R.MatTabLabelWrapper}),n.d(t,"MatTab",function(){return R.MatTab}),n.d(t,"MatTabLabel",function(){return R.MatTabLabel}),n.d(t,"MatTabNav",function(){return R.MatTabNav}),n.d(t,"MatTabLink",function(){return R.MatTabLink}),n.d(t,"MatTabContent",function(){return R.MatTabContent}),n.d(t,"MatTabsModule",function(){return R.MatTabsModule}),n.d(t,"MatTabChangeEvent",function(){return R.MatTabChangeEvent}),n.d(t,"MAT_TABS_CONFIG",function(){return R.MAT_TABS_CONFIG}),n.d(t,"MatTabGroupBase",function(){return R.MatTabGroupBase}),n.d(t,"_MatTabGroupMixinBase",function(){return R._MatTabGroupMixinBase}),n.d(t,"MatTabGroup",function(){return R.MatTabGroup}),n.d(t,"matTabsAnimations",function(){return R.matTabsAnimations});var N=n("./node_modules/@angular/material/esm5/toolbar.es5.js");n.d(t,"MatToolbarModule",function(){return N.MatToolbarModule}),n.d(t,"throwToolbarMixedModesError",function(){return N.throwToolbarMixedModesError}),n.d(t,"MatToolbarBase",function(){return N.MatToolbarBase}),n.d(t,"_MatToolbarMixinBase",function(){return N._MatToolbarMixinBase}),n.d(t,"MatToolbarRow",function(){return N.MatToolbarRow}),n.d(t,"MatToolbar",function(){return N.MatToolbar});var F=n("./node_modules/@angular/material/esm5/tooltip.es5.js");n.d(t,"MatTooltipModule",function(){return F.MatTooltipModule}),n.d(t,"getMatTooltipInvalidPositionError",function(){return F.getMatTooltipInvalidPositionError}),n.d(t,"MAT_TOOLTIP_SCROLL_STRATEGY_FACTORY",function(){return F.MAT_TOOLTIP_SCROLL_STRATEGY_FACTORY}),n.d(t,"MAT_TOOLTIP_DEFAULT_OPTIONS_FACTORY",function(){return F.MAT_TOOLTIP_DEFAULT_OPTIONS_FACTORY}),n.d(t,"SCROLL_THROTTLE_MS",function(){return F.SCROLL_THROTTLE_MS}),n.d(t,"TOOLTIP_PANEL_CLASS",function(){return F.TOOLTIP_PANEL_CLASS}),n.d(t,"MAT_TOOLTIP_SCROLL_STRATEGY",function(){return F.MAT_TOOLTIP_SCROLL_STRATEGY}),n.d(t,"MAT_TOOLTIP_SCROLL_STRATEGY_FACTORY_PROVIDER",function(){return F.MAT_TOOLTIP_SCROLL_STRATEGY_FACTORY_PROVIDER}),n.d(t,"MAT_TOOLTIP_DEFAULT_OPTIONS",function(){return F.MAT_TOOLTIP_DEFAULT_OPTIONS}),n.d(t,"MatTooltip",function(){return F.MatTooltip}),n.d(t,"TooltipComponent",function(){return F.TooltipComponent}),n.d(t,"matTooltipAnimations",function(){return F.matTooltipAnimations});var L=n("./node_modules/@angular/material/esm5/tree.es5.js");n.d(t,"_MatTreeNodeMixinBase",function(){return L._MatTreeNodeMixinBase}),n.d(t,"_MatNestedTreeNodeMixinBase",function(){return L._MatNestedTreeNodeMixinBase}),n.d(t,"MatTreeNode",function(){return L.MatTreeNode}),n.d(t,"MatTreeNodeDef",function(){return L.MatTreeNodeDef}),n.d(t,"MatNestedTreeNode",function(){return L.MatNestedTreeNode}),n.d(t,"MatTreeNodePadding",function(){return L.MatTreeNodePadding}),n.d(t,"MatTree",function(){return L.MatTree}),n.d(t,"MatTreeModule",function(){return L.MatTreeModule}),n.d(t,"MatTreeNodeToggle",function(){return L.MatTreeNodeToggle}),n.d(t,"MatTreeNodeOutlet",function(){return L.MatTreeNodeOutlet}),n.d(t,"MatTreeFlattener",function(){return L.MatTreeFlattener}),n.d(t,"MatTreeFlatDataSource",function(){return L.MatTreeFlatDataSource}),n.d(t,"MatTreeNestedDataSource",function(){return L.MatTreeNestedDataSource});var V=new r.Version("7.2.1")},"./node_modules/@angular/material/esm5/menu.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MAT_MENU_SCROLL_STRATEGY",function(){return k}),n.d(t,"MatMenuModule",function(){return I}),n.d(t,"MatMenu",function(){return S}),n.d(t,"MAT_MENU_DEFAULT_OPTIONS",function(){return C}),n.d(t,"MatMenuItem",function(){return w}),n.d(t,"MatMenuTrigger",function(){return D}),n.d(t,"matMenuAnimations",function(){return _}),n.d(t,"fadeInItems",function(){return y}),n.d(t,"transformMenu",function(){return b}),n.d(t,"MatMenuContent",function(){return v}),n.d(t,"ɵa23",function(){return E}),n.d(t,"ɵb23",function(){return x}),n.d(t,"ɵc23",function(){return O}),n.d(t,"ɵf23",function(){return j}),n.d(t,"ɵd23",function(){return M}),n.d(t,"ɵe23",function(){return A});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),o=n("./node_modules/@angular/common/fesm5/common.js"),a=n("./node_modules/rxjs/_esm5/index.js"),s=n("./node_modules/@angular/animations/fesm5/animations.js"),u=n("./node_modules/tslib/tslib.es6.js"),c=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),l=n("./node_modules/@angular/material/esm5/core.es5.js"),d=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),p=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),f=n("./node_modules/rxjs/_esm5/operators/index.js"),h=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),m=n("./node_modules/@angular/cdk/esm5/overlay.es5.js"),g=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),v=function(){function e(e,t,n,r,i,o){this._template=e,this._componentFactoryResolver=t,this._appRef=n,this._injector=r,this._viewContainerRef=i,this._document=o,this._attached=new a.Subject}return e.prototype.attach=function(e){void 0===e&&(e={}),this._portal||(this._portal=new i.TemplatePortal(this._template,this._viewContainerRef)),this.detach(),this._outlet||(this._outlet=new i.DomPortalOutlet(this._document.createElement("div"),this._componentFactoryResolver,this._appRef,this._injector));var t=this._template.elementRef.nativeElement;t.parentNode.insertBefore(this._outlet.outletElement,t),this._portal.attach(this._outlet,e),this._attached.next()},e.prototype.detach=function(){this._portal.isAttached&&this._portal.detach()},e.prototype.ngOnDestroy=function(){this._outlet&&this._outlet.dispose()},e.decorators=[{type:r.Directive,args:[{selector:"ng-template[matMenuContent]"}]}],e.ctorParameters=function(){return[{type:r.TemplateRef},{type:r.ComponentFactoryResolver},{type:r.ApplicationRef},{type:r.Injector},{type:r.ViewContainerRef},{type:void 0,decorators:[{type:r.Inject,args:[o.DOCUMENT]}]}]},e}(),_={transformMenu:Object(s.trigger)("transformMenu",[Object(s.state)("void",Object(s.style)({opacity:0,transform:"scale(0.8)"})),Object(s.transition)("void => enter",Object(s.group)([Object(s.query)(".mat-menu-content",Object(s.animate)("100ms linear",Object(s.style)({opacity:1}))),Object(s.animate)("120ms cubic-bezier(0, 0, 0.2, 1)",Object(s.style)({transform:"scale(1)"}))])),Object(s.transition)("* => void",Object(s.animate)("100ms 25ms linear",Object(s.style)({opacity:0})))]),fadeInItems:Object(s.trigger)("fadeInItems",[Object(s.state)("showing",Object(s.style)({opacity:1})),Object(s.transition)("void => *",[Object(s.style)({opacity:0}),Object(s.animate)("400ms 100ms cubic-bezier(0.55, 0, 0.55, 0.2)")])])},y=_.fadeInItems,b=_.transformMenu;var j=new r.InjectionToken("MAT_MENU_PANEL"),x=function(){return function(){}}(),O=Object(l.mixinDisableRipple)(Object(l.mixinDisabled)(x)),w=function(e){function t(t,n,r,i){var o=e.call(this)||this;return o._elementRef=t,o._focusMonitor=r,o._parentMenu=i,o.role="menuitem",o._hovered=new a.Subject,o._highlighted=!1,o._triggersSubmenu=!1,r&&r.monitor(o._elementRef,!1),i&&i.addItem&&i.addItem(o),o._document=n,o}return Object(u.__extends)(t,e),t.prototype.focus=function(e){void 0===e&&(e="program"),this._focusMonitor?this._focusMonitor.focusVia(this._getHostElement(),e):this._getHostElement().focus()},t.prototype.ngOnDestroy=function(){this._focusMonitor&&this._focusMonitor.stopMonitoring(this._elementRef),this._parentMenu&&this._parentMenu.removeItem&&this._parentMenu.removeItem(this),this._hovered.complete()},t.prototype._getTabIndex=function(){return this.disabled?"-1":"0"},t.prototype._getHostElement=function(){return this._elementRef.nativeElement},t.prototype._checkDisabled=function(e){this.disabled&&(e.preventDefault(),e.stopPropagation())},t.prototype._handleMouseEnter=function(){this._hovered.next(this)},t.prototype.getLabel=function(){var e=this._elementRef.nativeElement,t=this._document?this._document.TEXT_NODE:3,n="";if(e.childNodes)for(var r=e.childNodes.length,i=0;i<r;i++)e.childNodes[i].nodeType===t&&(n+=e.childNodes[i].textContent);return n.trim()},t.decorators=[{type:r.Component,args:[{selector:"[mat-menu-item]",exportAs:"matMenuItem",inputs:["disabled","disableRipple"],host:{"[attr.role]":"role",class:"mat-menu-item","[class.mat-menu-item-highlighted]":"_highlighted","[class.mat-menu-item-submenu-trigger]":"_triggersSubmenu","[attr.tabindex]":"_getTabIndex()","[attr.aria-disabled]":"disabled.toString()","[attr.disabled]":"disabled || null","(click)":"_checkDisabled($event)","(mouseenter)":"_handleMouseEnter()"},changeDetection:r.ChangeDetectionStrategy.OnPush,encapsulation:r.ViewEncapsulation.None,template:'<ng-content></ng-content><div class="mat-menu-ripple" matRipple [matRippleDisabled]="disableRipple || disabled" [matRippleTrigger]="_getHostElement()"></div>'}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:void 0,decorators:[{type:r.Inject,args:[o.DOCUMENT]}]},{type:c.FocusMonitor},{type:void 0,decorators:[{type:r.Inject,args:[j]},{type:r.Optional}]}]},t.propDecorators={role:[{type:r.Input}]},t}(O),C=new r.InjectionToken("mat-menu-default-options",{providedIn:"root",factory:E});function E(){return{overlapTrigger:!1,xPosition:"after",yPosition:"below",backdropClass:"cdk-overlay-transparent-backdrop"}}var S=function(){function e(e,t,n){this._elementRef=e,this._ngZone=t,this._defaultOptions=n,this._xPosition=this._defaultOptions.xPosition,this._yPosition=this._defaultOptions.yPosition,this._items=[],this._itemChanges=new a.Subject,this._tabSubscription=a.Subscription.EMPTY,this._classList={},this._panelAnimationState="void",this._animationDone=new a.Subject,this.backdropClass=this._defaultOptions.backdropClass,this._overlapTrigger=this._defaultOptions.overlapTrigger,this._hasBackdrop=this._defaultOptions.hasBackdrop,this.closed=new r.EventEmitter,this.close=this.closed}return Object.defineProperty(e.prototype,"xPosition",{get:function(){return this._xPosition},set:function(e){"before"!==e&&"after"!==e&&function(){throw Error('xPosition value must be either \'before\' or after\'.\n      Example: <mat-menu xPosition="before" #menu="matMenu"></mat-menu>')}(),this._xPosition=e,this.setPositionClasses()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"yPosition",{get:function(){return this._yPosition},set:function(e){"above"!==e&&"below"!==e&&function(){throw Error('yPosition value must be either \'above\' or below\'.\n      Example: <mat-menu yPosition="above" #menu="matMenu"></mat-menu>')}(),this._yPosition=e,this.setPositionClasses()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"overlapTrigger",{get:function(){return this._overlapTrigger},set:function(e){this._overlapTrigger=Object(d.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hasBackdrop",{get:function(){return this._hasBackdrop},set:function(e){this._hasBackdrop=Object(d.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"panelClass",{set:function(e){e&&e.length&&(this._classList=e.split(" ").reduce(function(e,t){return e[t]=!0,e},{}),this._elementRef.nativeElement.className="")},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"classList",{get:function(){return this.panelClass},set:function(e){this.panelClass=e},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){this.setPositionClasses()},e.prototype.ngAfterContentInit=function(){var e=this;this._keyManager=new c.FocusKeyManager(this._items).withWrap().withTypeAhead(),this._tabSubscription=this._keyManager.tabOut.subscribe(function(){return e.closed.emit("tab")})},e.prototype.ngOnDestroy=function(){this._tabSubscription.unsubscribe(),this.closed.complete()},e.prototype._hovered=function(){return this._itemChanges.pipe(Object(f.startWith)(this._items),Object(f.switchMap)(function(e){return a.merge.apply(void 0,e.map(function(e){return e._hovered}))}))},e.prototype._handleKeydown=function(e){var t=e.keyCode;switch(t){case p.ESCAPE:this.closed.emit("keydown");break;case p.LEFT_ARROW:this.parentMenu&&"ltr"===this.direction&&this.closed.emit("keydown");break;case p.RIGHT_ARROW:this.parentMenu&&"rtl"===this.direction&&this.closed.emit("keydown");break;default:t!==p.UP_ARROW&&t!==p.DOWN_ARROW||this._keyManager.setFocusOrigin("keyboard"),this._keyManager.onKeydown(e)}},e.prototype.focusFirstItem=function(e){var t=this;void 0===e&&(e="program"),this.lazyContent?this._ngZone.onStable.asObservable().pipe(Object(f.take)(1)).subscribe(function(){return t._keyManager.setFocusOrigin(e).setFirstItemActive()}):this._keyManager.setFocusOrigin(e).setFirstItemActive()},e.prototype.resetActiveItem=function(){this._keyManager.setActiveItem(-1)},e.prototype.setElevation=function(e){var t="mat-elevation-z"+(4+e),n=Object.keys(this._classList).find(function(e){return e.startsWith("mat-elevation-z")});n&&n!==this._previousElevation||(this._previousElevation&&(this._classList[this._previousElevation]=!1),this._classList[t]=!0,this._previousElevation=t)},e.prototype.addItem=function(e){-1===this._items.indexOf(e)&&(this._items.push(e),this._itemChanges.next(this._items))},e.prototype.removeItem=function(e){var t=this._items.indexOf(e);this._items.indexOf(e)>-1&&(this._items.splice(t,1),this._itemChanges.next(this._items))},e.prototype.setPositionClasses=function(e,t){void 0===e&&(e=this.xPosition),void 0===t&&(t=this.yPosition);var n=this._classList;n["mat-menu-before"]="before"===e,n["mat-menu-after"]="after"===e,n["mat-menu-above"]="above"===t,n["mat-menu-below"]="below"===t},e.prototype._startAnimation=function(){this._panelAnimationState="enter"},e.prototype._resetAnimation=function(){this._panelAnimationState="void"},e.prototype._onAnimationDone=function(e){this._animationDone.next(e),this._isAnimating=!1},e.prototype._onAnimationStart=function(e){this._isAnimating=!0,"enter"===e.toState&&0===this._keyManager.activeItemIndex&&(e.element.scrollTop=0)},e.decorators=[{type:r.Component,args:[{selector:"mat-menu",template:'<ng-template><div class="mat-menu-panel" [ngClass]="_classList" (keydown)="_handleKeydown($event)" (click)="closed.emit(\'click\')" [@transformMenu]="_panelAnimationState" (@transformMenu.start)="_onAnimationStart($event)" (@transformMenu.done)="_onAnimationDone($event)" tabindex="-1" role="menu"><div class="mat-menu-content"><ng-content></ng-content></div></div></ng-template>',styles:[".mat-menu-panel{min-width:112px;max-width:280px;overflow:auto;-webkit-overflow-scrolling:touch;max-height:calc(100vh - 48px);border-radius:4px;outline:0}.mat-menu-panel.ng-animating{pointer-events:none}@media (-ms-high-contrast:active){.mat-menu-panel{outline:solid 1px}}.mat-menu-content:not(:empty){padding-top:8px;padding-bottom:8px}.mat-menu-item{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;white-space:nowrap;overflow:hidden;text-overflow:ellipsis;display:block;line-height:48px;height:48px;padding:0 16px;text-align:left;text-decoration:none;max-width:100%;position:relative}.mat-menu-item::-moz-focus-inner{border:0}.mat-menu-item[disabled]{cursor:default}[dir=rtl] .mat-menu-item{text-align:right}.mat-menu-item .mat-icon{margin-right:16px;vertical-align:middle}.mat-menu-item .mat-icon svg{vertical-align:top}[dir=rtl] .mat-menu-item .mat-icon{margin-left:16px;margin-right:0}@media (-ms-high-contrast:active){.mat-menu-item-highlighted,.mat-menu-item.cdk-keyboard-focused,.mat-menu-item.cdk-program-focused{outline:dotted 1px}}.mat-menu-item-submenu-trigger{padding-right:32px}.mat-menu-item-submenu-trigger::after{width:0;height:0;border-style:solid;border-width:5px 0 5px 5px;border-color:transparent transparent transparent currentColor;content:'';display:inline-block;position:absolute;top:50%;right:16px;transform:translateY(-50%)}[dir=rtl] .mat-menu-item-submenu-trigger{padding-right:16px;padding-left:32px}[dir=rtl] .mat-menu-item-submenu-trigger::after{right:auto;left:16px;transform:rotateY(180deg) translateY(-50%)}button.mat-menu-item{width:100%}.mat-menu-item .mat-menu-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}"],changeDetection:r.ChangeDetectionStrategy.OnPush,encapsulation:r.ViewEncapsulation.None,exportAs:"matMenu",animations:[_.transformMenu,_.fadeInItems],providers:[{provide:j,useExisting:e}]}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:r.NgZone},{type:void 0,decorators:[{type:r.Inject,args:[C]}]}]},e.propDecorators={backdropClass:[{type:r.Input}],xPosition:[{type:r.Input}],yPosition:[{type:r.Input}],templateRef:[{type:r.ViewChild,args:[r.TemplateRef]}],items:[{type:r.ContentChildren,args:[w]}],lazyContent:[{type:r.ContentChild,args:[v]}],overlapTrigger:[{type:r.Input}],hasBackdrop:[{type:r.Input}],panelClass:[{type:r.Input,args:["class"]}],classList:[{type:r.Input}],closed:[{type:r.Output}],close:[{type:r.Output}]},e}(),k=new r.InjectionToken("mat-menu-scroll-strategy");function M(e){return function(){return e.scrollStrategies.reposition()}}var A={provide:k,deps:[m.Overlay],useFactory:M},T=Object(g.normalizePassiveListenerOptions)({passive:!0}),D=function(){function e(e,t,n,i,o,s,u,c){var l=this;this._overlay=e,this._element=t,this._viewContainerRef=n,this._parentMenu=o,this._menuItemInstance=s,this._dir=u,this._focusMonitor=c,this._overlayRef=null,this._menuOpen=!1,this._closeSubscription=a.Subscription.EMPTY,this._hoverSubscription=a.Subscription.EMPTY,this._menuCloseSubscription=a.Subscription.EMPTY,this._handleTouchStart=function(){return l._openedBy="touch"},this._openedBy=null,this.menuOpened=new r.EventEmitter,this.onMenuOpen=this.menuOpened,this.menuClosed=new r.EventEmitter,this.onMenuClose=this.menuClosed,t.nativeElement.addEventListener("touchstart",this._handleTouchStart,T),s&&(s._triggersSubmenu=this.triggersSubmenu()),this._scrollStrategy=i}return Object.defineProperty(e.prototype,"_deprecatedMatMenuTriggerFor",{get:function(){return this.menu},set:function(e){this.menu=e},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"menu",{get:function(){return this._menu},set:function(e){var t=this;e!==this._menu&&(this._menu=e,this._menuCloseSubscription.unsubscribe(),e&&(this._menuCloseSubscription=e.close.asObservable().subscribe(function(e){t._destroyMenu(),"click"!==e&&"tab"!==e||!t._parentMenu||t._parentMenu.closed.emit(e)})))},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){this._checkMenu(),this._handleHover()},e.prototype.ngOnDestroy=function(){this._overlayRef&&(this._overlayRef.dispose(),this._overlayRef=null),this._element.nativeElement.removeEventListener("touchstart",this._handleTouchStart,T),this._cleanUpSubscriptions()},Object.defineProperty(e.prototype,"menuOpen",{get:function(){return this._menuOpen},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"dir",{get:function(){return this._dir&&"rtl"===this._dir.value?"rtl":"ltr"},enumerable:!0,configurable:!0}),e.prototype.triggersSubmenu=function(){return!(!this._menuItemInstance||!this._parentMenu)},e.prototype.toggleMenu=function(){return this._menuOpen?this.closeMenu():this.openMenu()},e.prototype.openMenu=function(){var e=this;if(!this._menuOpen){this._checkMenu();var t=this._createOverlay(),n=t.getConfig();this._setPosition(n.positionStrategy),n.hasBackdrop=null==this.menu.hasBackdrop?!this.triggersSubmenu():this.menu.hasBackdrop,t.attach(this._getPortal()),this.menu.lazyContent&&this.menu.lazyContent.attach(this.menuData),this._closeSubscription=this._menuClosingActions().subscribe(function(){return e.closeMenu()}),this._initMenu(),this.menu instanceof S&&this.menu._startAnimation()}},e.prototype.closeMenu=function(){this.menu.close.emit()},e.prototype.focus=function(e){void 0===e&&(e="program"),this._focusMonitor?this._focusMonitor.focusVia(this._element,e):this._element.nativeElement.focus()},e.prototype._destroyMenu=function(){var e=this;if(this._overlayRef&&this.menuOpen){var t=this.menu;this._closeSubscription.unsubscribe(),this._overlayRef.detach(),t instanceof S?(t._resetAnimation(),t.lazyContent?t._animationDone.pipe(Object(f.filter)(function(e){return"void"===e.toState}),Object(f.take)(1),Object(f.takeUntil)(t.lazyContent._attached)).subscribe(function(){return t.lazyContent.detach()},void 0,function(){e._resetMenu()}):this._resetMenu()):(this._resetMenu(),t.lazyContent&&t.lazyContent.detach())}},e.prototype._initMenu=function(){this.menu.parentMenu=this.triggersSubmenu()?this._parentMenu:void 0,this.menu.direction=this.dir,this._setMenuElevation(),this._setIsMenuOpen(!0),this.menu.focusFirstItem(this._openedBy||"program")},e.prototype._setMenuElevation=function(){if(this.menu.setElevation){for(var e=0,t=this.menu.parentMenu;t;)e++,t=t.parentMenu;this.menu.setElevation(e)}},e.prototype._resetMenu=function(){this._setIsMenuOpen(!1),this._openedBy?this.triggersSubmenu()||this.focus(this._openedBy):this.focus(),this._openedBy=null},e.prototype._setIsMenuOpen=function(e){this._menuOpen=e,this._menuOpen?this.menuOpened.emit():this.menuClosed.emit(),this.triggersSubmenu()&&(this._menuItemInstance._highlighted=e)},e.prototype._checkMenu=function(){this.menu||function(){throw Error('matMenuTriggerFor: must pass in an mat-menu instance.\n\n    Example:\n      <mat-menu #menu="matMenu"></mat-menu>\n      <button [matMenuTriggerFor]="menu"></button>')}()},e.prototype._createOverlay=function(){if(!this._overlayRef){var e=this._getOverlayConfig();this._subscribeToPositions(e.positionStrategy),this._overlayRef=this._overlay.create(e),this._overlayRef.keydownEvents().subscribe()}return this._overlayRef},e.prototype._getOverlayConfig=function(){return new m.OverlayConfig({positionStrategy:this._overlay.position().flexibleConnectedTo(this._element).withLockedPosition().withTransformOriginOn(".mat-menu-panel"),backdropClass:this.menu.backdropClass||"cdk-overlay-transparent-backdrop",scrollStrategy:this._scrollStrategy(),direction:this._dir})},e.prototype._subscribeToPositions=function(e){var t=this;this.menu.setPositionClasses&&e.positionChanges.subscribe(function(e){var n="start"===e.connectionPair.overlayX?"after":"before",r="top"===e.connectionPair.overlayY?"below":"above";t.menu.setPositionClasses(n,r)})},e.prototype._setPosition=function(e){var t="before"===this.menu.xPosition?["end","start"]:["start","end"],n=t[0],r=t[1],i="above"===this.menu.yPosition?["bottom","top"]:["top","bottom"],o=i[0],a=i[1],s=[o,a],u=s[0],c=s[1],l=[n,r],d=l[0],p=l[1],f=0;this.triggersSubmenu()?(p=n="before"===this.menu.xPosition?"start":"end",r=d="end"===n?"start":"end",f="bottom"===o?8:-8):this.menu.overlapTrigger||(u="top"===o?"bottom":"top",c="top"===a?"bottom":"top"),e.withPositions([{originX:n,originY:u,overlayX:d,overlayY:o,offsetY:f},{originX:r,originY:u,overlayX:p,overlayY:o,offsetY:f},{originX:n,originY:c,overlayX:d,overlayY:a,offsetY:-f},{originX:r,originY:c,overlayX:p,overlayY:a,offsetY:-f}])},e.prototype._cleanUpSubscriptions=function(){this._closeSubscription.unsubscribe(),this._hoverSubscription.unsubscribe()},e.prototype._menuClosingActions=function(){var e=this,t=this._overlayRef.backdropClick(),n=this._overlayRef.detachments(),r=this._parentMenu?this._parentMenu.closed:Object(a.of)(),i=this._parentMenu?this._parentMenu._hovered().pipe(Object(f.filter)(function(t){return t!==e._menuItemInstance}),Object(f.filter)(function(){return e._menuOpen})):Object(a.of)();return Object(a.merge)(t,r,i,n)},e.prototype._handleMousedown=function(e){Object(c.isFakeMousedownFromScreenReader)(e)||(this._openedBy=0===e.button?"mouse":null,this.triggersSubmenu()&&e.preventDefault())},e.prototype._handleKeydown=function(e){var t=e.keyCode;this.triggersSubmenu()&&(t===p.RIGHT_ARROW&&"ltr"===this.dir||t===p.LEFT_ARROW&&"rtl"===this.dir)&&this.openMenu()},e.prototype._handleClick=function(e){this.triggersSubmenu()?(e.stopPropagation(),this.openMenu()):this.toggleMenu()},e.prototype._handleHover=function(){var e=this;this.triggersSubmenu()&&(this._hoverSubscription=this._parentMenu._hovered().pipe(Object(f.filter)(function(t){return t===e._menuItemInstance&&!t.disabled}),Object(f.delay)(0,a.asapScheduler)).subscribe(function(){e._openedBy="mouse",e.menu instanceof S&&e.menu._isAnimating?e.menu._animationDone.pipe(Object(f.take)(1),Object(f.delay)(0,a.asapScheduler),Object(f.takeUntil)(e._parentMenu._hovered())).subscribe(function(){return e.openMenu()}):e.openMenu()}))},e.prototype._getPortal=function(){return this._portal&&this._portal.templateRef===this.menu.templateRef||(this._portal=new i.TemplatePortal(this.menu.templateRef,this._viewContainerRef)),this._portal},e.decorators=[{type:r.Directive,args:[{selector:"[mat-menu-trigger-for], [matMenuTriggerFor]",host:{"aria-haspopup":"true","[attr.aria-expanded]":"menuOpen || null","(mousedown)":"_handleMousedown($event)","(keydown)":"_handleKeydown($event)","(click)":"_handleClick($event)"},exportAs:"matMenuTrigger"}]}],e.ctorParameters=function(){return[{type:m.Overlay},{type:r.ElementRef},{type:r.ViewContainerRef},{type:void 0,decorators:[{type:r.Inject,args:[k]}]},{type:S,decorators:[{type:r.Optional}]},{type:w,decorators:[{type:r.Optional},{type:r.Self}]},{type:h.Directionality,decorators:[{type:r.Optional}]},{type:c.FocusMonitor}]},e.propDecorators={_deprecatedMatMenuTriggerFor:[{type:r.Input,args:["mat-menu-trigger-for"]}],menu:[{type:r.Input,args:["matMenuTriggerFor"]}],menuData:[{type:r.Input,args:["matMenuTriggerData"]}],menuOpened:[{type:r.Output}],onMenuOpen:[{type:r.Output}],menuClosed:[{type:r.Output}],onMenuClose:[{type:r.Output}]},e}(),I=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[o.CommonModule,l.MatCommonModule,l.MatRippleModule,m.OverlayModule],exports:[S,w,D,v,l.MatCommonModule],declarations:[S,w,D,v],providers:[A]}]}],e}()},"./node_modules/@angular/material/esm5/paginator.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatPaginatorModule",function(){return y}),n.d(t,"PageEvent",function(){return m}),n.d(t,"MatPaginatorBase",function(){return g}),n.d(t,"_MatPaginatorBase",function(){return v}),n.d(t,"MatPaginator",function(){return _}),n.d(t,"MAT_PAGINATOR_INTL_PROVIDER_FACTORY",function(){return f}),n.d(t,"MatPaginatorIntl",function(){return p}),n.d(t,"MAT_PAGINATOR_INTL_PROVIDER",function(){return h});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/rxjs/_esm5/index.js"),o=n("./node_modules/tslib/tslib.es6.js"),a=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),s=n("./node_modules/@angular/material/esm5/core.es5.js"),u=n("./node_modules/@angular/common/fesm5/common.js"),c=n("./node_modules/@angular/material/esm5/button.es5.js"),l=n("./node_modules/@angular/material/esm5/select.es5.js"),d=n("./node_modules/@angular/material/esm5/tooltip.es5.js"),p=function(){function e(){this.changes=new i.Subject,this.itemsPerPageLabel="Items per page:",this.nextPageLabel="Next page",this.previousPageLabel="Previous page",this.firstPageLabel="First page",this.lastPageLabel="Last page",this.getRangeLabel=function(e,t,n){if(0==n||0==t)return"0 of "+n;var r=e*t;return r+1+" - "+(r<(n=Math.max(n,0))?Math.min(r+t,n):r+t)+" of "+n}}return e.decorators=[{type:r.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(r.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}();function f(e){return e||new p}var h={provide:p,deps:[[new r.Optional,new r.SkipSelf,p]],useFactory:f},m=function(){return function(){}}(),g=function(){return function(){}}(),v=Object(s.mixinDisabled)(Object(s.mixinInitialized)(g)),_=function(e){function t(t,n){var i=e.call(this)||this;return i._intl=t,i._changeDetectorRef=n,i._pageIndex=0,i._length=0,i._pageSizeOptions=[],i._hidePageSize=!1,i._showFirstLastButtons=!1,i.page=new r.EventEmitter,i._intlChanges=t.changes.subscribe(function(){return i._changeDetectorRef.markForCheck()}),i}return Object(o.__extends)(t,e),Object.defineProperty(t.prototype,"pageIndex",{get:function(){return this._pageIndex},set:function(e){this._pageIndex=Math.max(Object(a.coerceNumberProperty)(e),0),this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"length",{get:function(){return this._length},set:function(e){this._length=Object(a.coerceNumberProperty)(e),this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"pageSize",{get:function(){return this._pageSize},set:function(e){this._pageSize=Math.max(Object(a.coerceNumberProperty)(e),0),this._updateDisplayedPageSizeOptions()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"pageSizeOptions",{get:function(){return this._pageSizeOptions},set:function(e){this._pageSizeOptions=(e||[]).map(function(e){return Object(a.coerceNumberProperty)(e)}),this._updateDisplayedPageSizeOptions()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"hidePageSize",{get:function(){return this._hidePageSize},set:function(e){this._hidePageSize=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"showFirstLastButtons",{get:function(){return this._showFirstLastButtons},set:function(e){this._showFirstLastButtons=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){this._initialized=!0,this._updateDisplayedPageSizeOptions(),this._markInitialized()},t.prototype.ngOnDestroy=function(){this._intlChanges.unsubscribe()},t.prototype.nextPage=function(){if(this.hasNextPage()){var e=this.pageIndex;this.pageIndex++,this._emitPageEvent(e)}},t.prototype.previousPage=function(){if(this.hasPreviousPage()){var e=this.pageIndex;this.pageIndex--,this._emitPageEvent(e)}},t.prototype.firstPage=function(){if(this.hasPreviousPage()){var e=this.pageIndex;this.pageIndex=0,this._emitPageEvent(e)}},t.prototype.lastPage=function(){if(this.hasNextPage()){var e=this.pageIndex;this.pageIndex=this.getNumberOfPages()-1,this._emitPageEvent(e)}},t.prototype.hasPreviousPage=function(){return this.pageIndex>=1&&0!=this.pageSize},t.prototype.hasNextPage=function(){var e=this.getNumberOfPages()-1;return this.pageIndex<e&&0!=this.pageSize},t.prototype.getNumberOfPages=function(){return this.pageSize?Math.ceil(this.length/this.pageSize):0},t.prototype._changePageSize=function(e){var t=this.pageIndex*this.pageSize,n=this.pageIndex;this.pageIndex=Math.floor(t/e)||0,this.pageSize=e,this._emitPageEvent(n)},t.prototype._nextButtonsDisabled=function(){return this.disabled||!this.hasNextPage()},t.prototype._previousButtonsDisabled=function(){return this.disabled||!this.hasPreviousPage()},t.prototype._updateDisplayedPageSizeOptions=function(){this._initialized&&(this.pageSize||(this._pageSize=0!=this.pageSizeOptions.length?this.pageSizeOptions[0]:50),this._displayedPageSizeOptions=this.pageSizeOptions.slice(),-1===this._displayedPageSizeOptions.indexOf(this.pageSize)&&this._displayedPageSizeOptions.push(this.pageSize),this._displayedPageSizeOptions.sort(function(e,t){return e-t}),this._changeDetectorRef.markForCheck())},t.prototype._emitPageEvent=function(e){this.page.emit({previousPageIndex:e,pageIndex:this.pageIndex,pageSize:this.pageSize,length:this.length})},t.decorators=[{type:r.Component,args:[{selector:"mat-paginator",exportAs:"matPaginator",template:'<div class="mat-paginator-outer-container"><div class="mat-paginator-container"><div class="mat-paginator-page-size" *ngIf="!hidePageSize"><div class="mat-paginator-page-size-label">{{_intl.itemsPerPageLabel}}</div><mat-form-field *ngIf="_displayedPageSizeOptions.length > 1" [color]="color" class="mat-paginator-page-size-select"><mat-select [value]="pageSize" [disabled]="disabled" [aria-label]="_intl.itemsPerPageLabel" (selectionChange)="_changePageSize($event.value)"><mat-option *ngFor="let pageSizeOption of _displayedPageSizeOptions" [value]="pageSizeOption">{{pageSizeOption}}</mat-option></mat-select></mat-form-field><div *ngIf="_displayedPageSizeOptions.length <= 1">{{pageSize}}</div></div><div class="mat-paginator-range-actions"><div class="mat-paginator-range-label">{{_intl.getRangeLabel(pageIndex, pageSize, length)}}</div><button mat-icon-button type="button" class="mat-paginator-navigation-first" (click)="firstPage()" [attr.aria-label]="_intl.firstPageLabel" [matTooltip]="_intl.firstPageLabel" [matTooltipDisabled]="_previousButtonsDisabled()" [matTooltipPosition]="\'above\'" [disabled]="_previousButtonsDisabled()" *ngIf="showFirstLastButtons"><svg class="mat-paginator-icon" viewBox="0 0 24 24" focusable="false"><path d="M18.41 16.59L13.82 12l4.59-4.59L17 6l-6 6 6 6zM6 6h2v12H6z"/></svg></button> <button mat-icon-button type="button" class="mat-paginator-navigation-previous" (click)="previousPage()" [attr.aria-label]="_intl.previousPageLabel" [matTooltip]="_intl.previousPageLabel" [matTooltipDisabled]="_previousButtonsDisabled()" [matTooltipPosition]="\'above\'" [disabled]="_previousButtonsDisabled()"><svg class="mat-paginator-icon" viewBox="0 0 24 24" focusable="false"><path d="M15.41 7.41L14 6l-6 6 6 6 1.41-1.41L10.83 12z"/></svg></button> <button mat-icon-button type="button" class="mat-paginator-navigation-next" (click)="nextPage()" [attr.aria-label]="_intl.nextPageLabel" [matTooltip]="_intl.nextPageLabel" [matTooltipDisabled]="_nextButtonsDisabled()" [matTooltipPosition]="\'above\'" [disabled]="_nextButtonsDisabled()"><svg class="mat-paginator-icon" viewBox="0 0 24 24" focusable="false"><path d="M10 6L8.59 7.41 13.17 12l-4.58 4.59L10 18l6-6z"/></svg></button> <button mat-icon-button type="button" class="mat-paginator-navigation-last" (click)="lastPage()" [attr.aria-label]="_intl.lastPageLabel" [matTooltip]="_intl.lastPageLabel" [matTooltipDisabled]="_nextButtonsDisabled()" [matTooltipPosition]="\'above\'" [disabled]="_nextButtonsDisabled()" *ngIf="showFirstLastButtons"><svg class="mat-paginator-icon" viewBox="0 0 24 24" focusable="false"><path d="M5.59 7.41L10.18 12l-4.59 4.59L7 18l6-6-6-6zM16 6h2v12h-2z"/></svg></button></div></div></div>',styles:[".mat-paginator{display:block}.mat-paginator-outer-container{display:flex}.mat-paginator-container{display:flex;align-items:center;justify-content:flex-end;min-height:56px;padding:0 8px;flex-wrap:wrap-reverse;width:100%}.mat-paginator-page-size{display:flex;align-items:baseline;margin-right:8px}[dir=rtl] .mat-paginator-page-size{margin-right:0;margin-left:8px}.mat-paginator-page-size-label{margin:0 4px}.mat-paginator-page-size-select{margin:6px 4px 0 4px;width:56px}.mat-paginator-page-size-select.mat-form-field-appearance-outline{width:64px}.mat-paginator-page-size-select.mat-form-field-appearance-fill{width:64px}.mat-paginator-range-label{margin:0 32px 0 24px}.mat-paginator-range-actions{display:flex;align-items:center}.mat-paginator-icon{width:28px;fill:currentColor}[dir=rtl] .mat-paginator-icon{transform:rotate(180deg)}"],inputs:["disabled"],host:{class:"mat-paginator"},changeDetection:r.ChangeDetectionStrategy.OnPush,encapsulation:r.ViewEncapsulation.None}]}],t.ctorParameters=function(){return[{type:p},{type:r.ChangeDetectorRef}]},t.propDecorators={color:[{type:r.Input}],pageIndex:[{type:r.Input}],length:[{type:r.Input}],pageSize:[{type:r.Input}],pageSizeOptions:[{type:r.Input}],hidePageSize:[{type:r.Input}],showFirstLastButtons:[{type:r.Input}],page:[{type:r.Output}]},t}(v),y=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[u.CommonModule,c.MatButtonModule,l.MatSelectModule,d.MatTooltipModule],exports:[_],declarations:[_],providers:[h]}]}],e}()},"./node_modules/@angular/material/esm5/progress-bar.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatProgressBarModule",function(){return v}),n.d(t,"MAT_PROGRESS_BAR_LOCATION_FACTORY",function(){return f}),n.d(t,"MatProgressBarBase",function(){return l}),n.d(t,"_MatProgressBarMixinBase",function(){return d}),n.d(t,"MAT_PROGRESS_BAR_LOCATION",function(){return p}),n.d(t,"MatProgressBar",function(){return m});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/rxjs/_esm5/index.js"),a=n("./node_modules/rxjs/_esm5/operators/index.js"),s=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),u=n("./node_modules/@angular/material/esm5/core.es5.js"),c=n("./node_modules/@angular/common/fesm5/common.js"),l=function(){return function(e){this._elementRef=e}}(),d=Object(u.mixinColor)(l,"primary"),p=new i.InjectionToken("mat-progress-bar-location",{providedIn:"root",factory:f});function f(){var e=Object(i.inject)(c.DOCUMENT),t=e?e.location:null;return{getPathname:function(){return t?t.pathname+t.search:""}}}var h=0,m=function(e){function t(t,n,r,a){var s=e.call(this,t)||this;s._elementRef=t,s._ngZone=n,s._animationMode=r,s._isNoopAnimation=!1,s._value=0,s._bufferValue=0,s.animationEnd=new i.EventEmitter,s._animationEndSubscription=o.Subscription.EMPTY,s.mode="determinate",s.progressbarId="mat-progress-bar-"+h++;var u=a?a.getPathname().split("#")[0]:"";return s._rectangleFillValue="url('"+u+"#"+s.progressbarId+"')",s._isNoopAnimation="NoopAnimations"===r,s}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"value",{get:function(){return this._value},set:function(e){this._value=g(e||0),this._isNoopAnimation&&this.emitAnimationEnd()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"bufferValue",{get:function(){return this._bufferValue},set:function(e){this._bufferValue=g(e||0)},enumerable:!0,configurable:!0}),t.prototype._primaryTransform=function(){return{transform:"scaleX("+this.value/100+")"}},t.prototype._bufferTransform=function(){if("buffer"===this.mode)return{transform:"scaleX("+this.bufferValue/100+")"}},t.prototype.ngAfterViewInit=function(){var e=this;this._isNoopAnimation||this._ngZone.runOutsideAngular(function(){e._animationEndSubscription=Object(o.fromEvent)(e._primaryValueBar.nativeElement,"transitionend").pipe(Object(a.filter)(function(t){return t.target===e._primaryValueBar.nativeElement})).subscribe(function(t){return e._ngZone.run(function(){return e.emitAnimationEnd()})})})},t.prototype.ngOnDestroy=function(){this._animationEndSubscription.unsubscribe()},t.prototype.emitAnimationEnd=function(){"determinate"!==this.mode&&"buffer"!==this.mode||this.animationEnd.next({value:this.value})},t.decorators=[{type:i.Component,args:[{selector:"mat-progress-bar",exportAs:"matProgressBar",host:{role:"progressbar","aria-valuemin":"0","aria-valuemax":"100","[attr.aria-valuenow]":"value","[attr.mode]":"mode",class:"mat-progress-bar","[class._mat-animation-noopable]":"_isNoopAnimation"},inputs:["color"],template:'<svg width="100%" height="4" focusable="false" class="mat-progress-bar-background mat-progress-bar-element"><defs><pattern [id]="progressbarId" x="4" y="0" width="8" height="4" patternUnits="userSpaceOnUse"><circle cx="2" cy="2" r="2"/></pattern></defs><rect [attr.fill]="_rectangleFillValue" width="100%" height="100%"/></svg><div class="mat-progress-bar-buffer mat-progress-bar-element" [ngStyle]="_bufferTransform()"></div><div class="mat-progress-bar-primary mat-progress-bar-fill mat-progress-bar-element" [ngStyle]="_primaryTransform()" #primaryValueBar></div><div class="mat-progress-bar-secondary mat-progress-bar-fill mat-progress-bar-element"></div>',styles:[".mat-progress-bar{display:block;height:4px;overflow:hidden;position:relative;transition:opacity 250ms linear;width:100%}._mat-animation-noopable.mat-progress-bar{transition:none;animation:none}.mat-progress-bar .mat-progress-bar-element,.mat-progress-bar .mat-progress-bar-fill::after{height:100%;position:absolute;width:100%}.mat-progress-bar .mat-progress-bar-background{width:calc(100% + 10px)}@media (-ms-high-contrast:active){.mat-progress-bar .mat-progress-bar-background{display:none}}.mat-progress-bar .mat-progress-bar-buffer{transform-origin:top left;transition:transform 250ms ease}@media (-ms-high-contrast:active){.mat-progress-bar .mat-progress-bar-buffer{border-top:solid 5px;opacity:.5}}.mat-progress-bar .mat-progress-bar-secondary{display:none}.mat-progress-bar .mat-progress-bar-fill{animation:none;transform-origin:top left;transition:transform 250ms ease}@media (-ms-high-contrast:active){.mat-progress-bar .mat-progress-bar-fill{border-top:solid 4px}}.mat-progress-bar .mat-progress-bar-fill::after{animation:none;content:'';display:inline-block;left:0}.mat-progress-bar[dir=rtl],[dir=rtl] .mat-progress-bar{transform:rotateY(180deg)}.mat-progress-bar[mode=query]{transform:rotateZ(180deg)}.mat-progress-bar[mode=query][dir=rtl],[dir=rtl] .mat-progress-bar[mode=query]{transform:rotateZ(180deg) rotateY(180deg)}.mat-progress-bar[mode=indeterminate] .mat-progress-bar-fill,.mat-progress-bar[mode=query] .mat-progress-bar-fill{transition:none}.mat-progress-bar[mode=indeterminate] .mat-progress-bar-primary,.mat-progress-bar[mode=query] .mat-progress-bar-primary{-webkit-backface-visibility:hidden;backface-visibility:hidden;animation:mat-progress-bar-primary-indeterminate-translate 2s infinite linear;left:-145.166611%}.mat-progress-bar[mode=indeterminate] .mat-progress-bar-primary.mat-progress-bar-fill::after,.mat-progress-bar[mode=query] .mat-progress-bar-primary.mat-progress-bar-fill::after{-webkit-backface-visibility:hidden;backface-visibility:hidden;animation:mat-progress-bar-primary-indeterminate-scale 2s infinite linear}.mat-progress-bar[mode=indeterminate] .mat-progress-bar-secondary,.mat-progress-bar[mode=query] .mat-progress-bar-secondary{-webkit-backface-visibility:hidden;backface-visibility:hidden;animation:mat-progress-bar-secondary-indeterminate-translate 2s infinite linear;left:-54.888891%;display:block}.mat-progress-bar[mode=indeterminate] .mat-progress-bar-secondary.mat-progress-bar-fill::after,.mat-progress-bar[mode=query] .mat-progress-bar-secondary.mat-progress-bar-fill::after{-webkit-backface-visibility:hidden;backface-visibility:hidden;animation:mat-progress-bar-secondary-indeterminate-scale 2s infinite linear}.mat-progress-bar[mode=buffer] .mat-progress-bar-background{-webkit-backface-visibility:hidden;backface-visibility:hidden;animation:mat-progress-bar-background-scroll 250ms infinite linear;display:block}.mat-progress-bar._mat-animation-noopable .mat-progress-bar-background,.mat-progress-bar._mat-animation-noopable .mat-progress-bar-buffer,.mat-progress-bar._mat-animation-noopable .mat-progress-bar-fill,.mat-progress-bar._mat-animation-noopable .mat-progress-bar-fill::after,.mat-progress-bar._mat-animation-noopable .mat-progress-bar-primary,.mat-progress-bar._mat-animation-noopable .mat-progress-bar-primary.mat-progress-bar-fill::after,.mat-progress-bar._mat-animation-noopable .mat-progress-bar-secondary,.mat-progress-bar._mat-animation-noopable .mat-progress-bar-secondary.mat-progress-bar-fill::after{animation:none;transition:none}@keyframes mat-progress-bar-primary-indeterminate-translate{0%{transform:translateX(0)}20%{animation-timing-function:cubic-bezier(.5,0,.70173,.49582);transform:translateX(0)}59.15%{animation-timing-function:cubic-bezier(.30244,.38135,.55,.95635);transform:translateX(83.67142%)}100%{transform:translateX(200.61106%)}}@keyframes mat-progress-bar-primary-indeterminate-scale{0%{transform:scaleX(.08)}36.65%{animation-timing-function:cubic-bezier(.33473,.12482,.78584,1);transform:scaleX(.08)}69.15%{animation-timing-function:cubic-bezier(.06,.11,.6,1);transform:scaleX(.66148)}100%{transform:scaleX(.08)}}@keyframes mat-progress-bar-secondary-indeterminate-translate{0%{animation-timing-function:cubic-bezier(.15,0,.51506,.40969);transform:translateX(0)}25%{animation-timing-function:cubic-bezier(.31033,.28406,.8,.73371);transform:translateX(37.65191%)}48.35%{animation-timing-function:cubic-bezier(.4,.62704,.6,.90203);transform:translateX(84.38617%)}100%{transform:translateX(160.27778%)}}@keyframes mat-progress-bar-secondary-indeterminate-scale{0%{animation-timing-function:cubic-bezier(.15,0,.51506,.40969);transform:scaleX(.08)}19.15%{animation-timing-function:cubic-bezier(.31033,.28406,.8,.73371);transform:scaleX(.4571)}44.15%{animation-timing-function:cubic-bezier(.4,.62704,.6,.90203);transform:scaleX(.72796)}100%{transform:scaleX(.08)}}@keyframes mat-progress-bar-background-scroll{to{transform:translateX(-8px)}}"],changeDetection:i.ChangeDetectionStrategy.OnPush,encapsulation:i.ViewEncapsulation.None}]}],t.ctorParameters=function(){return[{type:i.ElementRef},{type:i.NgZone},{type:String,decorators:[{type:i.Optional},{type:i.Inject,args:[s.ANIMATION_MODULE_TYPE]}]},{type:void 0,decorators:[{type:i.Optional},{type:i.Inject,args:[p]}]}]},t.propDecorators={value:[{type:i.Input}],bufferValue:[{type:i.Input}],_primaryValueBar:[{type:i.ViewChild,args:["primaryValueBar"]}],animationEnd:[{type:i.Output}],mode:[{type:i.Input}]},t}(d);function g(e,t,n){return void 0===t&&(t=0),void 0===n&&(n=100),Math.max(t,Math.min(n,e))}var v=function(){function e(){}return e.decorators=[{type:i.NgModule,args:[{imports:[c.CommonModule,u.MatCommonModule],exports:[m,u.MatCommonModule],declarations:[m]}]}],e}()},"./node_modules/@angular/material/esm5/progress-spinner.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatProgressSpinnerModule",function(){return v}),n.d(t,"MAT_PROGRESS_SPINNER_DEFAULT_OPTIONS_FACTORY",function(){return h}),n.d(t,"MatProgressSpinnerBase",function(){return d}),n.d(t,"_MatProgressSpinnerMixinBase",function(){return p}),n.d(t,"MAT_PROGRESS_SPINNER_DEFAULT_OPTIONS",function(){return f}),n.d(t,"MatProgressSpinner",function(){return m}),n.d(t,"MatSpinner",function(){return g});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),o=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),a=n("./node_modules/@angular/common/fesm5/common.js"),s=n("./node_modules/@angular/core/fesm5/core.js"),u=n("./node_modules/@angular/material/esm5/core.es5.js"),c=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),l=100,d=function(){return function(e){this._elementRef=e}}(),p=Object(u.mixinColor)(d,"primary"),f=new s.InjectionToken("mat-progress-spinner-default-options",{providedIn:"root",factory:h});function h(){return{diameter:l}}var m=function(e){function t(t,n,r,i,o){var a=e.call(this,t)||this;a._elementRef=t,a._document=r,a.animationMode=i,a.defaults=o,a._value=0,a._fallbackAnimation=!1,a._noopAnimations="NoopAnimations"===a.animationMode&&!!a.defaults&&!a.defaults._forceAnimations,a._diameter=l,a.mode="determinate",a._fallbackAnimation=n.EDGE||n.TRIDENT,o&&(o.diameter&&(a.diameter=o.diameter),o.strokeWidth&&(a.strokeWidth=o.strokeWidth));var s="mat-progress-spinner-indeterminate"+(a._fallbackAnimation?"-fallback":"")+"-animation";return t.nativeElement.classList.add(s),a}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"diameter",{get:function(){return this._diameter},set:function(e){this._diameter=Object(i.coerceNumberProperty)(e),this._fallbackAnimation||t.diameters.has(this._diameter)||this._attachStyleNode()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"strokeWidth",{get:function(){return this._strokeWidth||this.diameter/10},set:function(e){this._strokeWidth=Object(i.coerceNumberProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{get:function(){return"determinate"===this.mode?this._value:0},set:function(e){this._value=Math.max(0,Math.min(100,Object(i.coerceNumberProperty)(e)))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_circleRadius",{get:function(){return(this.diameter-10)/2},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_viewBox",{get:function(){var e=2*this._circleRadius+this.strokeWidth;return"0 0 "+e+" "+e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_strokeCircumference",{get:function(){return 2*Math.PI*this._circleRadius},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_strokeDashOffset",{get:function(){return"determinate"===this.mode?this._strokeCircumference*(100-this._value)/100:this._fallbackAnimation&&"indeterminate"===this.mode?.2*this._strokeCircumference:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_circleStrokeWidth",{get:function(){return this.strokeWidth/this.diameter*100},enumerable:!0,configurable:!0}),t.prototype._attachStyleNode=function(){var e=t.styleTag;e||(e=this._document.createElement("style"),this._document.head.appendChild(e),t.styleTag=e),e&&e.sheet&&e.sheet.insertRule(this._getAnimationText(),0),t.diameters.add(this.diameter)},t.prototype._getAnimationText=function(){return"\n @keyframes mat-progress-spinner-stroke-rotate-DIAMETER {\n    0%      { stroke-dashoffset: START_VALUE;  transform: rotate(0); }\n    12.5%   { stroke-dashoffset: END_VALUE;    transform: rotate(0); }\n    12.5001%  { stroke-dashoffset: END_VALUE;    transform: rotateX(180deg) rotate(72.5deg); }\n    25%     { stroke-dashoffset: START_VALUE;  transform: rotateX(180deg) rotate(72.5deg); }\n\n    25.0001%   { stroke-dashoffset: START_VALUE;  transform: rotate(270deg); }\n    37.5%   { stroke-dashoffset: END_VALUE;    transform: rotate(270deg); }\n    37.5001%  { stroke-dashoffset: END_VALUE;    transform: rotateX(180deg) rotate(161.5deg); }\n    50%     { stroke-dashoffset: START_VALUE;  transform: rotateX(180deg) rotate(161.5deg); }\n\n    50.0001%  { stroke-dashoffset: START_VALUE;  transform: rotate(180deg); }\n    62.5%   { stroke-dashoffset: END_VALUE;    transform: rotate(180deg); }\n    62.5001%  { stroke-dashoffset: END_VALUE;    transform: rotateX(180deg) rotate(251.5deg); }\n    75%     { stroke-dashoffset: START_VALUE;  transform: rotateX(180deg) rotate(251.5deg); }\n\n    75.0001%  { stroke-dashoffset: START_VALUE;  transform: rotate(90deg); }\n    87.5%   { stroke-dashoffset: END_VALUE;    transform: rotate(90deg); }\n    87.5001%  { stroke-dashoffset: END_VALUE;    transform: rotateX(180deg) rotate(341.5deg); }\n    100%    { stroke-dashoffset: START_VALUE;  transform: rotateX(180deg) rotate(341.5deg); }\n  }\n".replace(/START_VALUE/g,""+.95*this._strokeCircumference).replace(/END_VALUE/g,""+.2*this._strokeCircumference).replace(/DIAMETER/g,""+this.diameter)},t.diameters=new Set([l]),t.styleTag=null,t.decorators=[{type:s.Component,args:[{selector:"mat-progress-spinner",exportAs:"matProgressSpinner",host:{role:"progressbar",class:"mat-progress-spinner","[class._mat-animation-noopable]":"_noopAnimations","[style.width.px]":"diameter","[style.height.px]":"diameter","[attr.aria-valuemin]":'mode === "determinate" ? 0 : null',"[attr.aria-valuemax]":'mode === "determinate" ? 100 : null',"[attr.aria-valuenow]":"value","[attr.mode]":"mode"},inputs:["color"],template:'<svg [style.width.px]="diameter" [style.height.px]="diameter" [attr.viewBox]="_viewBox" preserveAspectRatio="xMidYMid meet" focusable="false" [ngSwitch]="mode === \'indeterminate\'"><circle *ngSwitchCase="true" cx="50%" cy="50%" [attr.r]="_circleRadius" [style.animation-name]="\'mat-progress-spinner-stroke-rotate-\' + diameter" [style.stroke-dashoffset.px]="_strokeDashOffset" [style.stroke-dasharray.px]="_strokeCircumference" [style.stroke-width.%]="_circleStrokeWidth"></circle><circle *ngSwitchCase="false" cx="50%" cy="50%" [attr.r]="_circleRadius" [style.stroke-dashoffset.px]="_strokeDashOffset" [style.stroke-dasharray.px]="_strokeCircumference" [style.stroke-width.%]="_circleStrokeWidth"></circle></svg>',styles:[".mat-progress-spinner{display:block;position:relative}.mat-progress-spinner svg{position:absolute;transform:rotate(-90deg);top:0;left:0;transform-origin:center;overflow:visible}.mat-progress-spinner circle{fill:transparent;transform-origin:center;transition:stroke-dashoffset 225ms linear}._mat-animation-noopable.mat-progress-spinner circle{transition:none;animation:none}.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate]{animation:mat-progress-spinner-linear-rotate 2s linear infinite}._mat-animation-noopable.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate]{transition:none;animation:none}.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate] circle{transition-property:stroke;animation-duration:4s;animation-timing-function:cubic-bezier(.35,0,.25,1);animation-iteration-count:infinite}._mat-animation-noopable.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate] circle{transition:none;animation:none}.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate]{animation:mat-progress-spinner-stroke-rotate-fallback 10s cubic-bezier(.87,.03,.33,1) infinite}._mat-animation-noopable.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate]{transition:none;animation:none}.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate] circle{transition-property:stroke}._mat-animation-noopable.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate] circle{transition:none;animation:none}@keyframes mat-progress-spinner-linear-rotate{0%{transform:rotate(0)}100%{transform:rotate(360deg)}}@keyframes mat-progress-spinner-stroke-rotate-100{0%{stroke-dashoffset:268.60617px;transform:rotate(0)}12.5%{stroke-dashoffset:56.54867px;transform:rotate(0)}12.5001%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(72.5deg)}25%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(72.5deg)}25.0001%{stroke-dashoffset:268.60617px;transform:rotate(270deg)}37.5%{stroke-dashoffset:56.54867px;transform:rotate(270deg)}37.5001%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(161.5deg)}50%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(161.5deg)}50.0001%{stroke-dashoffset:268.60617px;transform:rotate(180deg)}62.5%{stroke-dashoffset:56.54867px;transform:rotate(180deg)}62.5001%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(251.5deg)}75%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(251.5deg)}75.0001%{stroke-dashoffset:268.60617px;transform:rotate(90deg)}87.5%{stroke-dashoffset:56.54867px;transform:rotate(90deg)}87.5001%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(341.5deg)}100%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(341.5deg)}}@keyframes mat-progress-spinner-stroke-rotate-fallback{0%{transform:rotate(0)}25%{transform:rotate(1170deg)}50%{transform:rotate(2340deg)}75%{transform:rotate(3510deg)}100%{transform:rotate(4680deg)}}"],changeDetection:s.ChangeDetectionStrategy.OnPush,encapsulation:s.ViewEncapsulation.None}]}],t.ctorParameters=function(){return[{type:s.ElementRef},{type:o.Platform},{type:void 0,decorators:[{type:s.Optional},{type:s.Inject,args:[a.DOCUMENT]}]},{type:String,decorators:[{type:s.Optional},{type:s.Inject,args:[c.ANIMATION_MODULE_TYPE]}]},{type:void 0,decorators:[{type:s.Inject,args:[f]}]}]},t.propDecorators={diameter:[{type:s.Input}],strokeWidth:[{type:s.Input}],mode:[{type:s.Input}],value:[{type:s.Input}]},t}(p),g=function(e){function t(t,n,r,i,o){var a=e.call(this,t,n,r,i,o)||this;return a.mode="indeterminate",a}return Object(r.__extends)(t,e),t.decorators=[{type:s.Component,args:[{selector:"mat-spinner",host:{role:"progressbar",mode:"indeterminate",class:"mat-spinner mat-progress-spinner","[class._mat-animation-noopable]":"_noopAnimations","[style.width.px]":"diameter","[style.height.px]":"diameter"},inputs:["color"],template:'<svg [style.width.px]="diameter" [style.height.px]="diameter" [attr.viewBox]="_viewBox" preserveAspectRatio="xMidYMid meet" focusable="false" [ngSwitch]="mode === \'indeterminate\'"><circle *ngSwitchCase="true" cx="50%" cy="50%" [attr.r]="_circleRadius" [style.animation-name]="\'mat-progress-spinner-stroke-rotate-\' + diameter" [style.stroke-dashoffset.px]="_strokeDashOffset" [style.stroke-dasharray.px]="_strokeCircumference" [style.stroke-width.%]="_circleStrokeWidth"></circle><circle *ngSwitchCase="false" cx="50%" cy="50%" [attr.r]="_circleRadius" [style.stroke-dashoffset.px]="_strokeDashOffset" [style.stroke-dasharray.px]="_strokeCircumference" [style.stroke-width.%]="_circleStrokeWidth"></circle></svg>',styles:[".mat-progress-spinner{display:block;position:relative}.mat-progress-spinner svg{position:absolute;transform:rotate(-90deg);top:0;left:0;transform-origin:center;overflow:visible}.mat-progress-spinner circle{fill:transparent;transform-origin:center;transition:stroke-dashoffset 225ms linear}._mat-animation-noopable.mat-progress-spinner circle{transition:none;animation:none}.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate]{animation:mat-progress-spinner-linear-rotate 2s linear infinite}._mat-animation-noopable.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate]{transition:none;animation:none}.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate] circle{transition-property:stroke;animation-duration:4s;animation-timing-function:cubic-bezier(.35,0,.25,1);animation-iteration-count:infinite}._mat-animation-noopable.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate] circle{transition:none;animation:none}.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate]{animation:mat-progress-spinner-stroke-rotate-fallback 10s cubic-bezier(.87,.03,.33,1) infinite}._mat-animation-noopable.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate]{transition:none;animation:none}.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate] circle{transition-property:stroke}._mat-animation-noopable.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate] circle{transition:none;animation:none}@keyframes mat-progress-spinner-linear-rotate{0%{transform:rotate(0)}100%{transform:rotate(360deg)}}@keyframes mat-progress-spinner-stroke-rotate-100{0%{stroke-dashoffset:268.60617px;transform:rotate(0)}12.5%{stroke-dashoffset:56.54867px;transform:rotate(0)}12.5001%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(72.5deg)}25%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(72.5deg)}25.0001%{stroke-dashoffset:268.60617px;transform:rotate(270deg)}37.5%{stroke-dashoffset:56.54867px;transform:rotate(270deg)}37.5001%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(161.5deg)}50%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(161.5deg)}50.0001%{stroke-dashoffset:268.60617px;transform:rotate(180deg)}62.5%{stroke-dashoffset:56.54867px;transform:rotate(180deg)}62.5001%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(251.5deg)}75%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(251.5deg)}75.0001%{stroke-dashoffset:268.60617px;transform:rotate(90deg)}87.5%{stroke-dashoffset:56.54867px;transform:rotate(90deg)}87.5001%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(341.5deg)}100%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(341.5deg)}}@keyframes mat-progress-spinner-stroke-rotate-fallback{0%{transform:rotate(0)}25%{transform:rotate(1170deg)}50%{transform:rotate(2340deg)}75%{transform:rotate(3510deg)}100%{transform:rotate(4680deg)}}"],changeDetection:s.ChangeDetectionStrategy.OnPush,encapsulation:s.ViewEncapsulation.None}]}],t.ctorParameters=function(){return[{type:s.ElementRef},{type:o.Platform},{type:void 0,decorators:[{type:s.Optional},{type:s.Inject,args:[a.DOCUMENT]}]},{type:String,decorators:[{type:s.Optional},{type:s.Inject,args:[c.ANIMATION_MODULE_TYPE]}]},{type:void 0,decorators:[{type:s.Inject,args:[f]}]}]},t}(m),v=function(){function e(){}return e.decorators=[{type:s.NgModule,args:[{imports:[u.MatCommonModule,a.CommonModule],exports:[m,g,u.MatCommonModule],declarations:[m,g]}]}],e}()},"./node_modules/@angular/material/esm5/radio.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatRadioModule",function(){return y}),n.d(t,"MAT_RADIO_GROUP_CONTROL_VALUE_ACCESSOR",function(){return f}),n.d(t,"MatRadioChange",function(){return h}),n.d(t,"MatRadioGroup",function(){return m}),n.d(t,"MatRadioButtonBase",function(){return g}),n.d(t,"_MatRadioButtonMixinBase",function(){return v}),n.d(t,"MatRadioButton",function(){return _});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),o=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),a=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),s=n("./node_modules/@angular/core/fesm5/core.js"),u=n("./node_modules/@angular/forms/fesm5/forms.js"),c=n("./node_modules/@angular/material/esm5/core.es5.js"),l=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),d=n("./node_modules/@angular/common/fesm5/common.js"),p=0,f={provide:u.NG_VALUE_ACCESSOR,useExisting:Object(s.forwardRef)(function(){return m}),multi:!0},h=function(){return function(e,t){this.source=e,this.value=t}}(),m=function(){function e(e){this._changeDetector=e,this._value=null,this._name="mat-radio-group-"+p++,this._selected=null,this._isInitialized=!1,this._labelPosition="after",this._disabled=!1,this._required=!1,this._controlValueAccessorChangeFn=function(){},this.onTouched=function(){},this.change=new s.EventEmitter}return Object.defineProperty(e.prototype,"name",{get:function(){return this._name},set:function(e){this._name=e,this._updateRadioButtonNames()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"labelPosition",{get:function(){return this._labelPosition},set:function(e){this._labelPosition="before"===e?"before":"after",this._markRadiosForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return this._value},set:function(e){this._value!==e&&(this._value=e,this._updateSelectedRadioFromValue(),this._checkSelectedRadioButton())},enumerable:!0,configurable:!0}),e.prototype._checkSelectedRadioButton=function(){this._selected&&!this._selected.checked&&(this._selected.checked=!0)},Object.defineProperty(e.prototype,"selected",{get:function(){return this._selected},set:function(e){this._selected=e,this.value=e?e.value:null,this._checkSelectedRadioButton()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disabled",{get:function(){return this._disabled},set:function(e){this._disabled=Object(o.coerceBooleanProperty)(e),this._markRadiosForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"required",{get:function(){return this._required},set:function(e){this._required=Object(o.coerceBooleanProperty)(e),this._markRadiosForCheck()},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){this._isInitialized=!0},e.prototype._touch=function(){this.onTouched&&this.onTouched()},e.prototype._updateRadioButtonNames=function(){var e=this;this._radios&&this._radios.forEach(function(t){t.name=e.name})},e.prototype._updateSelectedRadioFromValue=function(){var e=this,t=null!==this._selected&&this._selected.value===this._value;this._radios&&!t&&(this._selected=null,this._radios.forEach(function(t){t.checked=e.value===t.value,t.checked&&(e._selected=t)}))},e.prototype._emitChangeEvent=function(){this._isInitialized&&this.change.emit(new h(this._selected,this._value))},e.prototype._markRadiosForCheck=function(){this._radios&&this._radios.forEach(function(e){return e._markForCheck()})},e.prototype.writeValue=function(e){this.value=e,this._changeDetector.markForCheck()},e.prototype.registerOnChange=function(e){this._controlValueAccessorChangeFn=e},e.prototype.registerOnTouched=function(e){this.onTouched=e},e.prototype.setDisabledState=function(e){this.disabled=e,this._changeDetector.markForCheck()},e.decorators=[{type:s.Directive,args:[{selector:"mat-radio-group",exportAs:"matRadioGroup",providers:[f],host:{role:"radiogroup",class:"mat-radio-group"}}]}],e.ctorParameters=function(){return[{type:s.ChangeDetectorRef}]},e.propDecorators={change:[{type:s.Output}],_radios:[{type:s.ContentChildren,args:[Object(s.forwardRef)(function(){return _}),{descendants:!0}]}],name:[{type:s.Input}],labelPosition:[{type:s.Input}],value:[{type:s.Input}],selected:[{type:s.Input}],disabled:[{type:s.Input}],required:[{type:s.Input}]},e}(),g=function(){return function(e){this._elementRef=e}}(),v=Object(c.mixinColor)(Object(c.mixinDisableRipple)(Object(c.mixinTabIndex)(g)),"accent"),_=function(e){function t(t,n,r,i,o,a){var u=e.call(this,n)||this;return u._changeDetector=r,u._focusMonitor=i,u._radioDispatcher=o,u._animationMode=a,u._uniqueId="mat-radio-"+ ++p,u.id=u._uniqueId,u.change=new s.EventEmitter,u._checked=!1,u._value=null,u._removeUniqueSelectionListener=function(){},u.radioGroup=t,u._removeUniqueSelectionListener=o.listen(function(e,t){e!==u.id&&t===u.name&&(u.checked=!1)}),u}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"checked",{get:function(){return this._checked},set:function(e){var t=Object(o.coerceBooleanProperty)(e);this._checked!==t&&(this._checked=t,t&&this.radioGroup&&this.radioGroup.value!==this.value?this.radioGroup.selected=this:!t&&this.radioGroup&&this.radioGroup.value===this.value&&(this.radioGroup.selected=null),t&&this._radioDispatcher.notify(this.id,this.name),this._changeDetector.markForCheck())},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{get:function(){return this._value},set:function(e){this._value!==e&&(this._value=e,null!==this.radioGroup&&(this.checked||(this.checked=this.radioGroup.value===e),this.checked&&(this.radioGroup.selected=this)))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"labelPosition",{get:function(){return this._labelPosition||this.radioGroup&&this.radioGroup.labelPosition||"after"},set:function(e){this._labelPosition=e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return this._disabled||null!==this.radioGroup&&this.radioGroup.disabled},set:function(e){var t=Object(o.coerceBooleanProperty)(e);this._disabled!==t&&(this._disabled=t,this._changeDetector.markForCheck())},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"required",{get:function(){return this._required||this.radioGroup&&this.radioGroup.required},set:function(e){this._required=Object(o.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"inputId",{get:function(){return(this.id||this._uniqueId)+"-input"},enumerable:!0,configurable:!0}),t.prototype.focus=function(){this._focusMonitor.focusVia(this._inputElement,"keyboard")},t.prototype._markForCheck=function(){this._changeDetector.markForCheck()},t.prototype.ngOnInit=function(){this.radioGroup&&(this.checked=this.radioGroup.value===this._value,this.name=this.radioGroup.name)},t.prototype.ngAfterViewInit=function(){var e=this;this._focusMonitor.monitor(this._elementRef,!0).subscribe(function(t){!t&&e.radioGroup&&e.radioGroup._touch()})},t.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._elementRef),this._removeUniqueSelectionListener()},t.prototype._emitChangeEvent=function(){this.change.emit(new h(this,this._value))},t.prototype._isRippleDisabled=function(){return this.disableRipple||this.disabled},t.prototype._onInputClick=function(e){e.stopPropagation()},t.prototype._onInputChange=function(e){e.stopPropagation();var t=this.radioGroup&&this.value!==this.radioGroup.value;this.checked=!0,this._emitChangeEvent(),this.radioGroup&&(this.radioGroup._controlValueAccessorChangeFn(this.value),this.radioGroup._touch(),t&&this.radioGroup._emitChangeEvent())},t.decorators=[{type:s.Component,args:[{selector:"mat-radio-button",template:'<label [attr.for]="inputId" class="mat-radio-label" #label><div class="mat-radio-container"><div class="mat-radio-outer-circle"></div><div class="mat-radio-inner-circle"></div><div mat-ripple class="mat-radio-ripple" [matRippleTrigger]="label" [matRippleDisabled]="_isRippleDisabled()" [matRippleCentered]="true" [matRippleRadius]="20" [matRippleAnimation]="{enterDuration: 150}"><div class="mat-ripple-element mat-radio-persistent-ripple"></div></div><input #input class="mat-radio-input cdk-visually-hidden" type="radio" [id]="inputId" [checked]="checked" [disabled]="disabled" [tabIndex]="tabIndex" [attr.name]="name" [required]="required" [attr.aria-label]="ariaLabel" [attr.aria-labelledby]="ariaLabelledby" [attr.aria-describedby]="ariaDescribedby" (change)="_onInputChange($event)" (click)="_onInputClick($event)"></div><div class="mat-radio-label-content" [class.mat-radio-label-before]="labelPosition == \'before\'"><span style="display:none">&nbsp;</span><ng-content></ng-content></div></label>',styles:[".mat-radio-button{display:inline-block;-webkit-tap-highlight-color:transparent;outline:0}.mat-radio-label{cursor:pointer;display:inline-flex;align-items:center;white-space:nowrap;vertical-align:middle}.mat-radio-container{box-sizing:border-box;display:inline-block;position:relative;width:20px;height:20px;flex-shrink:0}.mat-radio-outer-circle{box-sizing:border-box;height:20px;left:0;position:absolute;top:0;transition:border-color ease 280ms;width:20px;border-width:2px;border-style:solid;border-radius:50%}._mat-animation-noopable .mat-radio-outer-circle{transition:none}.mat-radio-inner-circle{border-radius:50%;box-sizing:border-box;height:20px;left:0;position:absolute;top:0;transition:transform ease 280ms,background-color ease 280ms;width:20px;transform:scale(.001)}._mat-animation-noopable .mat-radio-inner-circle{transition:none}.mat-radio-checked .mat-radio-inner-circle{transform:scale(.5)}@media (-ms-high-contrast:active){.mat-radio-checked .mat-radio-inner-circle{border:solid 10px}}.mat-radio-label-content{display:inline-block;order:0;line-height:inherit;padding-left:8px;padding-right:0}[dir=rtl] .mat-radio-label-content{padding-right:8px;padding-left:0}.mat-radio-label-content.mat-radio-label-before{order:-1;padding-left:0;padding-right:8px}[dir=rtl] .mat-radio-label-content.mat-radio-label-before{padding-right:0;padding-left:8px}.mat-radio-disabled,.mat-radio-disabled .mat-radio-label{cursor:default}.mat-radio-button .mat-radio-ripple{position:absolute;left:calc(50% - 20px);top:calc(50% - 20px);height:40px;width:40px;z-index:1;pointer-events:none}.mat-radio-button .mat-radio-ripple .mat-ripple-element:not(.mat-radio-persistent-ripple){opacity:.16}.mat-radio-persistent-ripple{width:100%;height:100%;transform:none}.mat-radio-container:hover .mat-radio-persistent-ripple{opacity:.04}.mat-radio-button:not(.mat-radio-disabled).cdk-focused .mat-radio-persistent-ripple{opacity:.12}.mat-radio-disabled .mat-radio-container:hover .mat-radio-persistent-ripple,.mat-radio-persistent-ripple{opacity:0}@media (hover:none){.mat-radio-container:hover .mat-radio-persistent-ripple{display:none}}.mat-radio-input{bottom:0;left:50%}"],inputs:["color","disableRipple","tabIndex"],encapsulation:s.ViewEncapsulation.None,exportAs:"matRadioButton",host:{class:"mat-radio-button","[class.mat-radio-checked]":"checked","[class.mat-radio-disabled]":"disabled","[class._mat-animation-noopable]":'_animationMode === "NoopAnimations"',"[attr.tabindex]":"-1","[attr.id]":"id","(focus)":"_inputElement.nativeElement.focus()"},changeDetection:s.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:m,decorators:[{type:s.Optional}]},{type:s.ElementRef},{type:s.ChangeDetectorRef},{type:i.FocusMonitor},{type:a.UniqueSelectionDispatcher},{type:String,decorators:[{type:s.Optional},{type:s.Inject,args:[l.ANIMATION_MODULE_TYPE]}]}]},t.propDecorators={id:[{type:s.Input}],name:[{type:s.Input}],ariaLabel:[{type:s.Input,args:["aria-label"]}],ariaLabelledby:[{type:s.Input,args:["aria-labelledby"]}],ariaDescribedby:[{type:s.Input,args:["aria-describedby"]}],checked:[{type:s.Input}],value:[{type:s.Input}],labelPosition:[{type:s.Input}],disabled:[{type:s.Input}],required:[{type:s.Input}],change:[{type:s.Output}],_inputElement:[{type:s.ViewChild,args:["input"]}]},t}(v),y=function(){function e(){}return e.decorators=[{type:s.NgModule,args:[{imports:[d.CommonModule,c.MatRippleModule,c.MatCommonModule],exports:[m,_,c.MatCommonModule],declarations:[m,_]}]}],e}()},"./node_modules/@angular/material/esm5/select.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatSelectModule",function(){return F}),n.d(t,"MAT_SELECT_SCROLL_STRATEGY_PROVIDER_FACTORY",function(){return A}),n.d(t,"SELECT_PANEL_MAX_HEIGHT",function(){return O}),n.d(t,"SELECT_PANEL_PADDING_X",function(){return w}),n.d(t,"SELECT_PANEL_INDENT_PADDING_X",function(){return C}),n.d(t,"SELECT_ITEM_HEIGHT_EM",function(){return E}),n.d(t,"SELECT_MULTIPLE_PANEL_PADDING_X",function(){return S}),n.d(t,"SELECT_PANEL_VIEWPORT_PADDING",function(){return k}),n.d(t,"MAT_SELECT_SCROLL_STRATEGY",function(){return M}),n.d(t,"MAT_SELECT_SCROLL_STRATEGY_PROVIDER",function(){return T}),n.d(t,"MatSelectChange",function(){return D}),n.d(t,"MatSelectBase",function(){return I}),n.d(t,"_MatSelectMixinBase",function(){return P}),n.d(t,"MatSelectTrigger",function(){return R}),n.d(t,"MatSelect",function(){return N}),n.d(t,"matSelectAnimations",function(){return y}),n.d(t,"transformPanel",function(){return b}),n.d(t,"fadeInContent",function(){return j});var r=n("./node_modules/@angular/animations/fesm5/animations.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),a=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),c=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),l=n("./node_modules/@angular/cdk/esm5/overlay.es5.js"),d=n("./node_modules/@angular/cdk/esm5/scrolling.es5.js"),p=n("./node_modules/@angular/core/fesm5/core.js"),f=n("./node_modules/@angular/forms/fesm5/forms.js"),h=n("./node_modules/@angular/material/esm5/core.es5.js"),m=n("./node_modules/@angular/material/esm5/form-field.es5.js"),g=n("./node_modules/rxjs/_esm5/index.js"),v=n("./node_modules/rxjs/_esm5/operators/index.js"),_=n("./node_modules/@angular/common/fesm5/common.js"),y={transformPanel:Object(r.trigger)("transformPanel",[Object(r.state)("void",Object(r.style)({transform:"scaleY(0.8)",minWidth:"100%",opacity:0})),Object(r.state)("showing",Object(r.style)({opacity:1,minWidth:"calc(100% + 32px)",transform:"scaleY(1)"})),Object(r.state)("showing-multiple",Object(r.style)({opacity:1,minWidth:"calc(100% + 64px)",transform:"scaleY(1)"})),Object(r.transition)("void => *",Object(r.animate)("120ms cubic-bezier(0, 0, 0.2, 1)")),Object(r.transition)("* => void",Object(r.animate)("100ms 25ms linear",Object(r.style)({opacity:0})))]),fadeInContent:Object(r.trigger)("fadeInContent",[Object(r.state)("showing",Object(r.style)({opacity:1})),Object(r.transition)("void => showing",[Object(r.style)({opacity:0}),Object(r.animate)("150ms 100ms cubic-bezier(0.55, 0, 0.55, 0.2)")])])},b=y.transformPanel,j=y.fadeInContent;var x=0,O=256,w=16,C=2*w,E=3,S=0,k=8,M=new p.InjectionToken("mat-select-scroll-strategy");function A(e){return function(){return e.scrollStrategies.reposition()}}var T={provide:M,deps:[l.Overlay],useFactory:A},D=function(){return function(e,t){this.source=e,this.value=t}}(),I=function(){return function(e,t,n,r,i){this._elementRef=e,this._defaultErrorStateMatcher=t,this._parentForm=n,this._parentFormGroup=r,this.ngControl=i}}(),P=Object(h.mixinDisableRipple)(Object(h.mixinTabIndex)(Object(h.mixinDisabled)(Object(h.mixinErrorState)(I)))),R=function(){function e(){}return e.decorators=[{type:p.Directive,args:[{selector:"mat-select-trigger"}]}],e}(),N=function(e){function t(t,n,r,i,o,a,s,u,c,l,d,f,h){var m=e.call(this,o,i,s,u,l)||this;return m._viewportRuler=t,m._changeDetectorRef=n,m._ngZone=r,m._dir=a,m._parentFormField=c,m.ngControl=l,m._liveAnnouncer=h,m._panelOpen=!1,m._required=!1,m._scrollTop=0,m._multiple=!1,m._compareWith=function(e,t){return e===t},m._uid="mat-select-"+x++,m._destroy=new g.Subject,m._triggerFontSize=0,m._onChange=function(){},m._onTouched=function(){},m._optionIds="",m._transformOrigin="top",m._panelDoneAnimatingStream=new g.Subject,m._offsetY=0,m._positions=[{originX:"start",originY:"top",overlayX:"start",overlayY:"top"},{originX:"start",originY:"bottom",overlayX:"start",overlayY:"bottom"}],m._disableOptionCentering=!1,m._focused=!1,m.controlType="mat-select",m.ariaLabel="",m.optionSelectionChanges=Object(g.defer)(function(){return m.options?g.merge.apply(void 0,m.options.map(function(e){return e.onSelectionChange})):m._ngZone.onStable.asObservable().pipe(Object(v.take)(1),Object(v.switchMap)(function(){return m.optionSelectionChanges}))}),m.openedChange=new p.EventEmitter,m._openedStream=m.openedChange.pipe(Object(v.filter)(function(e){return e}),Object(v.map)(function(){})),m._closedStream=m.openedChange.pipe(Object(v.filter)(function(e){return!e}),Object(v.map)(function(){})),m.selectionChange=new p.EventEmitter,m.valueChange=new p.EventEmitter,m.ngControl&&(m.ngControl.valueAccessor=m),m._scrollStrategyFactory=f,m._scrollStrategy=m._scrollStrategyFactory(),m.tabIndex=parseInt(d)||0,m.id=m.id,m}return Object(i.__extends)(t,e),Object.defineProperty(t.prototype,"focused",{get:function(){return this._focused||this._panelOpen},set:function(e){this._focused=e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"placeholder",{get:function(){return this._placeholder},set:function(e){this._placeholder=e,this.stateChanges.next()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"required",{get:function(){return this._required},set:function(e){this._required=Object(s.coerceBooleanProperty)(e),this.stateChanges.next()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"multiple",{get:function(){return this._multiple},set:function(e){if(this._selectionModel)throw Error("Cannot change `multiple` mode of select after initialization.");this._multiple=Object(s.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disableOptionCentering",{get:function(){return this._disableOptionCentering},set:function(e){this._disableOptionCentering=Object(s.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"compareWith",{get:function(){return this._compareWith},set:function(e){if("function"!=typeof e)throw Error("`compareWith` must be a function.");this._compareWith=e,this._selectionModel&&this._initializeSelection()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{get:function(){return this._value},set:function(e){e!==this._value&&(this.writeValue(e),this._value=e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"id",{get:function(){return this._id},set:function(e){this._id=e||this._uid,this.stateChanges.next()},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){var e=this;this._selectionModel=new u.SelectionModel(this.multiple),this.stateChanges.next(),this._panelDoneAnimatingStream.pipe(Object(v.distinctUntilChanged)(),Object(v.takeUntil)(this._destroy)).subscribe(function(){e.panelOpen?(e._scrollTop=0,e.openedChange.emit(!0)):(e.openedChange.emit(!1),e.overlayDir.offsetX=0,e._changeDetectorRef.markForCheck())})},t.prototype.ngAfterContentInit=function(){var e=this;this._initKeyManager(),this._selectionModel.onChange.pipe(Object(v.takeUntil)(this._destroy)).subscribe(function(e){e.added.forEach(function(e){return e.select()}),e.removed.forEach(function(e){return e.deselect()})}),this.options.changes.pipe(Object(v.startWith)(null),Object(v.takeUntil)(this._destroy)).subscribe(function(){e._resetOptions(),e._initializeSelection()})},t.prototype.ngDoCheck=function(){this.ngControl&&this.updateErrorState()},t.prototype.ngOnChanges=function(e){e.disabled&&this.stateChanges.next()},t.prototype.ngOnDestroy=function(){this._destroy.next(),this._destroy.complete(),this.stateChanges.complete()},t.prototype.toggle=function(){this.panelOpen?this.close():this.open()},t.prototype.open=function(){var e=this;!this.disabled&&this.options&&this.options.length&&!this._panelOpen&&(this._triggerRect=this.trigger.nativeElement.getBoundingClientRect(),this._triggerFontSize=parseInt(getComputedStyle(this.trigger.nativeElement).fontSize||"0"),this._panelOpen=!0,this._keyManager.withHorizontalOrientation(null),this._calculateOverlayPosition(),this._highlightCorrectOption(),this._changeDetectorRef.markForCheck(),this._ngZone.onStable.asObservable().pipe(Object(v.take)(1)).subscribe(function(){e._triggerFontSize&&e.overlayDir.overlayRef&&e.overlayDir.overlayRef.overlayElement&&(e.overlayDir.overlayRef.overlayElement.style.fontSize=e._triggerFontSize+"px")}))},t.prototype.close=function(){this._panelOpen&&(this._panelOpen=!1,this._keyManager.withHorizontalOrientation(this._isRtl()?"rtl":"ltr"),this._changeDetectorRef.markForCheck(),this._onTouched())},t.prototype.writeValue=function(e){this.options&&this._setSelectionByValue(e)},t.prototype.registerOnChange=function(e){this._onChange=e},t.prototype.registerOnTouched=function(e){this._onTouched=e},t.prototype.setDisabledState=function(e){this.disabled=e,this._changeDetectorRef.markForCheck(),this.stateChanges.next()},Object.defineProperty(t.prototype,"panelOpen",{get:function(){return this._panelOpen},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selected",{get:function(){return this.multiple?this._selectionModel.selected:this._selectionModel.selected[0]},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"triggerValue",{get:function(){if(this.empty)return"";if(this._multiple){var e=this._selectionModel.selected.map(function(e){return e.viewValue});return this._isRtl()&&e.reverse(),e.join(", ")}return this._selectionModel.selected[0].viewValue},enumerable:!0,configurable:!0}),t.prototype._isRtl=function(){return!!this._dir&&"rtl"===this._dir.value},t.prototype._handleKeydown=function(e){this.disabled||(this.panelOpen?this._handleOpenKeydown(e):this._handleClosedKeydown(e))},t.prototype._handleClosedKeydown=function(e){var t=e.keyCode,n=t===c.DOWN_ARROW||t===c.UP_ARROW||t===c.LEFT_ARROW||t===c.RIGHT_ARROW,r=t===c.ENTER||t===c.SPACE,i=this._keyManager;if(r&&!Object(c.hasModifierKey)(e)||(this.multiple||e.altKey)&&n)e.preventDefault(),this.open();else if(!this.multiple){var o=this.selected;t===c.HOME||t===c.END?(t===c.HOME?i.setFirstItemActive():i.setLastItemActive(),e.preventDefault()):i.onKeydown(e),this._liveAnnouncer&&o!==this.selected&&this._liveAnnouncer.announce(this.selected.viewValue)}},t.prototype._handleOpenKeydown=function(e){var t=e.keyCode,n=t===c.DOWN_ARROW||t===c.UP_ARROW,r=this._keyManager;if(t===c.HOME||t===c.END)e.preventDefault(),t===c.HOME?r.setFirstItemActive():r.setLastItemActive();else if(n&&e.altKey)e.preventDefault(),this.close();else if(t!==c.ENTER&&t!==c.SPACE||!r.activeItem||Object(c.hasModifierKey)(e))if(this._multiple&&t===c.A&&e.ctrlKey){e.preventDefault();var i=this.options.some(function(e){return!e.disabled&&!e.selected});this.options.forEach(function(e){e.disabled||(i?e.select():e.deselect())})}else{var o=r.activeItemIndex;r.onKeydown(e),this._multiple&&n&&e.shiftKey&&r.activeItem&&r.activeItemIndex!==o&&r.activeItem._selectViaInteraction()}else e.preventDefault(),r.activeItem._selectViaInteraction()},t.prototype._onFocus=function(){this.disabled||(this._focused=!0,this.stateChanges.next())},t.prototype._onBlur=function(){this._focused=!1,this.disabled||this.panelOpen||(this._onTouched(),this._changeDetectorRef.markForCheck(),this.stateChanges.next())},t.prototype._onAttached=function(){var e=this;this.overlayDir.positionChange.pipe(Object(v.take)(1)).subscribe(function(){e._setPseudoCheckboxPaddingSize(),e._changeDetectorRef.detectChanges(),e._calculateOverlayOffsetX(),e.panel.nativeElement.scrollTop=e._scrollTop})},t.prototype._getPanelTheme=function(){return this._parentFormField?"mat-"+this._parentFormField.color:""},t.prototype._setPseudoCheckboxPaddingSize=function(){if(!S&&this.multiple){var e=this.panel.nativeElement.querySelector(".mat-pseudo-checkbox");e&&(S=1.5*w+e.offsetWidth)}},Object.defineProperty(t.prototype,"empty",{get:function(){return!this._selectionModel||this._selectionModel.isEmpty()},enumerable:!0,configurable:!0}),t.prototype._initializeSelection=function(){var e=this;Promise.resolve().then(function(){e._setSelectionByValue(e.ngControl?e.ngControl.value:e._value),e.stateChanges.next()})},t.prototype._setSelectionByValue=function(e){var t=this;if(this.multiple&&e){if(!Array.isArray(e))throw Error("Value must be an array in multiple-selection mode.");this._selectionModel.clear(),e.forEach(function(e){return t._selectValue(e)}),this._sortValues()}else{this._selectionModel.clear();var n=this._selectValue(e);n&&this._keyManager.setActiveItem(n)}this._changeDetectorRef.markForCheck()},t.prototype._selectValue=function(e){var t=this,n=this.options.find(function(n){try{return null!=n.value&&t._compareWith(n.value,e)}catch(e){return Object(p.isDevMode)()&&console.warn(e),!1}});return n&&this._selectionModel.select(n),n},t.prototype._initKeyManager=function(){var e=this;this._keyManager=new o.ActiveDescendantKeyManager(this.options).withTypeAhead().withVerticalOrientation().withHorizontalOrientation(this._isRtl()?"rtl":"ltr").withAllowedModifierKeys(["shiftKey"]),this._keyManager.tabOut.pipe(Object(v.takeUntil)(this._destroy)).subscribe(function(){e.focus(),e.close()}),this._keyManager.change.pipe(Object(v.takeUntil)(this._destroy)).subscribe(function(){e._panelOpen&&e.panel?e._scrollActiveOptionIntoView():e._panelOpen||e.multiple||!e._keyManager.activeItem||e._keyManager.activeItem._selectViaInteraction()})},t.prototype._resetOptions=function(){var e=this,t=Object(g.merge)(this.options.changes,this._destroy);this.optionSelectionChanges.pipe(Object(v.takeUntil)(t)).subscribe(function(t){e._onSelect(t.source,t.isUserInput),t.isUserInput&&!e.multiple&&e._panelOpen&&(e.close(),e.focus())}),g.merge.apply(void 0,this.options.map(function(e){return e._stateChanges})).pipe(Object(v.takeUntil)(t)).subscribe(function(){e._changeDetectorRef.markForCheck(),e.stateChanges.next()}),this._setOptionIds()},t.prototype._onSelect=function(e,t){var n=this._selectionModel.isSelected(e);null!=e.value||this._multiple?(e.selected?this._selectionModel.select(e):this._selectionModel.deselect(e),t&&this._keyManager.setActiveItem(e),this.multiple&&(this._sortValues(),t&&this.focus())):(e.deselect(),this._selectionModel.clear(),this._propagateChanges(e.value)),n!==this._selectionModel.isSelected(e)&&this._propagateChanges(),this.stateChanges.next()},t.prototype._sortValues=function(){var e=this;if(this.multiple){var t=this.options.toArray();this._selectionModel.sort(function(n,r){return e.sortComparator?e.sortComparator(n,r,t):t.indexOf(n)-t.indexOf(r)}),this.stateChanges.next()}},t.prototype._propagateChanges=function(e){var t=null;t=this.multiple?this.selected.map(function(e){return e.value}):this.selected?this.selected.value:e,this._value=t,this.valueChange.emit(t),this._onChange(t),this.selectionChange.emit(new D(this,t)),this._changeDetectorRef.markForCheck()},t.prototype._setOptionIds=function(){this._optionIds=this.options.map(function(e){return e.id}).join(" ")},t.prototype._highlightCorrectOption=function(){this._keyManager&&(this.empty?this._keyManager.setFirstItemActive():this._keyManager.setActiveItem(this._selectionModel.selected[0]))},t.prototype._scrollActiveOptionIntoView=function(){var e=this._keyManager.activeItemIndex||0,t=Object(h._countGroupLabelsBeforeOption)(e,this.options,this.optionGroups);this.panel.nativeElement.scrollTop=Object(h._getOptionScrollPosition)(e+t,this._getItemHeight(),this.panel.nativeElement.scrollTop,O)},t.prototype.focus=function(){this._elementRef.nativeElement.focus()},t.prototype._getOptionIndex=function(e){return this.options.reduce(function(t,n,r){return void 0===t?e===n?r:void 0:t},void 0)},t.prototype._calculateOverlayPosition=function(){var e=this._getItemHeight(),t=this._getItemCount(),n=Math.min(t*e,O),r=t*e-n,i=this.empty?0:this._getOptionIndex(this._selectionModel.selected[0]);i+=Object(h._countGroupLabelsBeforeOption)(i,this.options,this.optionGroups);var o=n/2;this._scrollTop=this._calculateOverlayScroll(i,o,r),this._offsetY=this._calculateOverlayOffsetY(i,o,r),this._checkOverlayWithinViewport(r)},t.prototype._calculateOverlayScroll=function(e,t,n){var r=this._getItemHeight(),i=r*e-t+r/2;return Math.min(Math.max(0,i),n)},t.prototype._getAriaLabel=function(){return this.ariaLabelledby?null:this.ariaLabel||this.placeholder},t.prototype._getAriaLabelledby=function(){return this.ariaLabelledby?this.ariaLabelledby:this._parentFormField&&this._parentFormField._hasFloatingLabel()&&!this._getAriaLabel()?this._parentFormField._labelId||null:null},t.prototype._getAriaActiveDescendant=function(){return this.panelOpen&&this._keyManager&&this._keyManager.activeItem?this._keyManager.activeItem.id:null},t.prototype._calculateOverlayOffsetX=function(){var e,t=this.overlayDir.overlayRef.overlayElement.getBoundingClientRect(),n=this._viewportRuler.getViewportSize(),r=this._isRtl(),i=this.multiple?S+w:2*w;if(this.multiple)e=S;else{var o=this._selectionModel.selected[0]||this.options.first;e=o&&o.group?C:w}r||(e*=-1);var a=0-(t.left+e-(r?i:0)),s=t.right+e-n.width+(r?0:i);a>0?e+=a+k:s>0&&(e-=s+k),this.overlayDir.offsetX=Math.round(e),this.overlayDir.overlayRef.updatePosition()},t.prototype._calculateOverlayOffsetY=function(e,t,n){var r,i=this._getItemHeight(),o=(i-this._triggerRect.height)/2,a=Math.floor(O/i);if(this._disableOptionCentering)return 0;if(0===this._scrollTop)r=e*i;else if(this._scrollTop===n){r=(e-(this._getItemCount()-a))*i+(i-(this._getItemCount()*i-O)%i)}else r=t-i/2;return Math.round(-1*r-o)},t.prototype._checkOverlayWithinViewport=function(e){var t=this._getItemHeight(),n=this._viewportRuler.getViewportSize(),r=this._triggerRect.top-k,i=n.height-this._triggerRect.bottom-k,o=Math.abs(this._offsetY),a=Math.min(this._getItemCount()*t,O)-o-this._triggerRect.height;a>i?this._adjustPanelUp(a,i):o>r?this._adjustPanelDown(o,r,e):this._transformOrigin=this._getOriginBasedOnOption()},t.prototype._adjustPanelUp=function(e,t){var n=Math.round(e-t);this._scrollTop-=n,this._offsetY-=n,this._transformOrigin=this._getOriginBasedOnOption(),this._scrollTop<=0&&(this._scrollTop=0,this._offsetY=0,this._transformOrigin="50% bottom 0px")},t.prototype._adjustPanelDown=function(e,t,n){var r=Math.round(e-t);if(this._scrollTop+=r,this._offsetY+=r,this._transformOrigin=this._getOriginBasedOnOption(),this._scrollTop>=n)return this._scrollTop=n,this._offsetY=0,void(this._transformOrigin="50% top 0px")},t.prototype._getOriginBasedOnOption=function(){var e=this._getItemHeight(),t=(e-this._triggerRect.height)/2;return"50% "+(Math.abs(this._offsetY)-t+e/2)+"px 0px"},t.prototype._getItemCount=function(){return this.options.length+this.optionGroups.length},t.prototype._getItemHeight=function(){return this._triggerFontSize*E},t.prototype.setDescribedByIds=function(e){this._ariaDescribedby=e.join(" ")},t.prototype.onContainerClick=function(){this.focus(),this.open()},Object.defineProperty(t.prototype,"shouldLabelFloat",{get:function(){return this._panelOpen||!this.empty},enumerable:!0,configurable:!0}),t.decorators=[{type:p.Component,args:[{selector:"mat-select",exportAs:"matSelect",template:'<div cdk-overlay-origin class="mat-select-trigger" aria-hidden="true" (click)="toggle()" #origin="cdkOverlayOrigin" #trigger><div class="mat-select-value" [ngSwitch]="empty"><span class="mat-select-placeholder" *ngSwitchCase="true">{{placeholder || \' \'}}</span> <span class="mat-select-value-text" *ngSwitchCase="false" [ngSwitch]="!!customTrigger"><span *ngSwitchDefault>{{triggerValue || \' \'}}</span><ng-content select="mat-select-trigger" *ngSwitchCase="true"></ng-content></span></div><div class="mat-select-arrow-wrapper"><div class="mat-select-arrow"></div></div></div><ng-template cdk-connected-overlay cdkConnectedOverlayLockPosition cdkConnectedOverlayHasBackdrop cdkConnectedOverlayBackdropClass="cdk-overlay-transparent-backdrop" [cdkConnectedOverlayScrollStrategy]="_scrollStrategy" [cdkConnectedOverlayOrigin]="origin" [cdkConnectedOverlayOpen]="panelOpen" [cdkConnectedOverlayPositions]="_positions" [cdkConnectedOverlayMinWidth]="_triggerRect?.width" [cdkConnectedOverlayOffsetY]="_offsetY" (backdropClick)="close()" (attach)="_onAttached()" (detach)="close()"><div #panel class="mat-select-panel {{ _getPanelTheme() }}" [ngClass]="panelClass" [@transformPanel]="multiple ? \'showing-multiple\' : \'showing\'" (@transformPanel.done)="_panelDoneAnimatingStream.next($event.toState)" [style.transformOrigin]="_transformOrigin" [style.font-size.px]="_triggerFontSize" (keydown)="_handleKeydown($event)"><ng-content></ng-content></div></ng-template>',styles:[".mat-select{display:inline-block;width:100%;outline:0}.mat-select-trigger{display:inline-table;cursor:pointer;position:relative;box-sizing:border-box}.mat-select-disabled .mat-select-trigger{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:default}.mat-select-value{display:table-cell;max-width:0;width:100%;overflow:hidden;text-overflow:ellipsis;white-space:nowrap}.mat-select-value-text{white-space:nowrap;overflow:hidden;text-overflow:ellipsis}.mat-select-arrow-wrapper{display:table-cell;vertical-align:middle}.mat-form-field-appearance-fill .mat-select-arrow-wrapper{transform:translateY(-50%)}.mat-form-field-appearance-outline .mat-select-arrow-wrapper{transform:translateY(-25%)}.mat-form-field-appearance-standard.mat-form-field-has-label .mat-select:not(.mat-select-empty) .mat-select-arrow-wrapper{transform:translateY(-50%)}.mat-form-field-appearance-standard .mat-select.mat-select-empty .mat-select-arrow-wrapper{transition:transform .4s cubic-bezier(.25,.8,.25,1)}._mat-animation-noopable.mat-form-field-appearance-standard .mat-select.mat-select-empty .mat-select-arrow-wrapper{transition:none}.mat-select-arrow{width:0;height:0;border-left:5px solid transparent;border-right:5px solid transparent;border-top:5px solid;margin:0 4px}.mat-select-panel{min-width:112px;max-width:280px;overflow:auto;-webkit-overflow-scrolling:touch;padding-top:0;padding-bottom:0;max-height:256px;min-width:100%;border-radius:4px}@media (-ms-high-contrast:active){.mat-select-panel{outline:solid 1px}}.mat-select-panel .mat-optgroup-label,.mat-select-panel .mat-option{font-size:inherit;line-height:3em;height:3em}.mat-form-field-type-mat-select:not(.mat-form-field-disabled) .mat-form-field-flex{cursor:pointer}.mat-form-field-type-mat-select .mat-form-field-label{width:calc(100% - 18px)}.mat-select-placeholder{transition:color .4s .133s cubic-bezier(.25,.8,.25,1)}._mat-animation-noopable .mat-select-placeholder{transition:none}.mat-form-field-hide-placeholder .mat-select-placeholder{color:transparent;-webkit-text-fill-color:transparent;transition:none;display:block}"],inputs:["disabled","disableRipple","tabIndex"],encapsulation:p.ViewEncapsulation.None,changeDetection:p.ChangeDetectionStrategy.OnPush,host:{role:"listbox","[attr.id]":"id","[attr.tabindex]":"tabIndex","[attr.aria-label]":"_getAriaLabel()","[attr.aria-labelledby]":"_getAriaLabelledby()","[attr.aria-required]":"required.toString()","[attr.aria-disabled]":"disabled.toString()","[attr.aria-invalid]":"errorState","[attr.aria-owns]":"panelOpen ? _optionIds : null","[attr.aria-multiselectable]":"multiple","[attr.aria-describedby]":"_ariaDescribedby || null","[attr.aria-activedescendant]":"_getAriaActiveDescendant()","[class.mat-select-disabled]":"disabled","[class.mat-select-invalid]":"errorState","[class.mat-select-required]":"required","[class.mat-select-empty]":"empty",class:"mat-select","(keydown)":"_handleKeydown($event)","(focus)":"_onFocus()","(blur)":"_onBlur()"},animations:[y.transformPanel],providers:[{provide:m.MatFormFieldControl,useExisting:t},{provide:h.MAT_OPTION_PARENT_COMPONENT,useExisting:t}]}]}],t.ctorParameters=function(){return[{type:d.ViewportRuler},{type:p.ChangeDetectorRef},{type:p.NgZone},{type:h.ErrorStateMatcher},{type:p.ElementRef},{type:a.Directionality,decorators:[{type:p.Optional}]},{type:f.NgForm,decorators:[{type:p.Optional}]},{type:f.FormGroupDirective,decorators:[{type:p.Optional}]},{type:m.MatFormField,decorators:[{type:p.Optional}]},{type:f.NgControl,decorators:[{type:p.Self},{type:p.Optional}]},{type:String,decorators:[{type:p.Attribute,args:["tabindex"]}]},{type:void 0,decorators:[{type:p.Inject,args:[M]}]},{type:o.LiveAnnouncer}]},t.propDecorators={trigger:[{type:p.ViewChild,args:["trigger"]}],panel:[{type:p.ViewChild,args:["panel"]}],overlayDir:[{type:p.ViewChild,args:[l.CdkConnectedOverlay]}],options:[{type:p.ContentChildren,args:[h.MatOption,{descendants:!0}]}],optionGroups:[{type:p.ContentChildren,args:[h.MatOptgroup]}],panelClass:[{type:p.Input}],customTrigger:[{type:p.ContentChild,args:[R]}],placeholder:[{type:p.Input}],required:[{type:p.Input}],multiple:[{type:p.Input}],disableOptionCentering:[{type:p.Input}],compareWith:[{type:p.Input}],value:[{type:p.Input}],ariaLabel:[{type:p.Input,args:["aria-label"]}],ariaLabelledby:[{type:p.Input,args:["aria-labelledby"]}],errorStateMatcher:[{type:p.Input}],sortComparator:[{type:p.Input}],id:[{type:p.Input}],openedChange:[{type:p.Output}],_openedStream:[{type:p.Output,args:["opened"]}],_closedStream:[{type:p.Output,args:["closed"]}],selectionChange:[{type:p.Output}],valueChange:[{type:p.Output}]},t}(P),F=function(){function e(){}return e.decorators=[{type:p.NgModule,args:[{imports:[_.CommonModule,l.OverlayModule,h.MatOptionModule,h.MatCommonModule],exports:[m.MatFormFieldModule,N,R,h.MatOptionModule,h.MatCommonModule],declarations:[N,R],providers:[T]}]}],e}()},"./node_modules/@angular/material/esm5/sidenav.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatSidenavModule",function(){return S}),n.d(t,"throwMatDuplicatedDrawerError",function(){return _}),n.d(t,"MAT_DRAWER_DEFAULT_AUTOSIZE_FACTORY",function(){return b}),n.d(t,"MAT_DRAWER_DEFAULT_AUTOSIZE",function(){return y}),n.d(t,"MatDrawerContent",function(){return j}),n.d(t,"MatDrawer",function(){return x}),n.d(t,"MatDrawerContainer",function(){return O}),n.d(t,"MatSidenavContent",function(){return w}),n.d(t,"MatSidenav",function(){return C}),n.d(t,"MatSidenavContainer",function(){return E}),n.d(t,"matDrawerAnimations",function(){return v});var r=n("./node_modules/@angular/animations/fesm5/animations.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),a=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),c=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),l=n("./node_modules/@angular/cdk/esm5/scrolling.es5.js"),d=n("./node_modules/@angular/common/fesm5/common.js"),p=n("./node_modules/@angular/core/fesm5/core.js"),f=n("./node_modules/rxjs/_esm5/index.js"),h=n("./node_modules/rxjs/_esm5/operators/index.js"),m=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),g=n("./node_modules/@angular/material/esm5/core.es5.js"),v={transformDrawer:Object(r.trigger)("transform",[Object(r.state)("open, open-instant",Object(r.style)({transform:"none",visibility:"visible"})),Object(r.state)("void",Object(r.style)({"box-shadow":"none",visibility:"hidden"})),Object(r.transition)("void => open-instant",Object(r.animate)("0ms")),Object(r.transition)("void <=> open, open-instant => void",Object(r.animate)("400ms cubic-bezier(0.25, 0.8, 0.25, 1)"))])};function _(e){throw Error("A drawer was already declared for 'position=\""+e+"\"'")}var y=new p.InjectionToken("MAT_DRAWER_DEFAULT_AUTOSIZE",{providedIn:"root",factory:b});function b(){return!1}var j=function(e){function t(t,n,r,i,o){var a=e.call(this,r,i,o)||this;return a._changeDetectorRef=t,a._container=n,a}return Object(i.__extends)(t,e),t.prototype.ngAfterContentInit=function(){var e=this;this._container._contentMarginChanges.subscribe(function(){e._changeDetectorRef.markForCheck()})},t.decorators=[{type:p.Component,args:[{selector:"mat-drawer-content",template:"<ng-content></ng-content>",host:{class:"mat-drawer-content","[style.margin-left.px]":"_container._contentMargins.left","[style.margin-right.px]":"_container._contentMargins.right"},changeDetection:p.ChangeDetectionStrategy.OnPush,encapsulation:p.ViewEncapsulation.None}]}],t.ctorParameters=function(){return[{type:p.ChangeDetectorRef},{type:O,decorators:[{type:p.Inject,args:[Object(p.forwardRef)(function(){return O})]}]},{type:p.ElementRef},{type:l.ScrollDispatcher},{type:p.NgZone}]},t}(l.CdkScrollable),x=function(){function e(e,t,n,r,i,o){var a=this;this._elementRef=e,this._focusTrapFactory=t,this._focusMonitor=n,this._platform=r,this._ngZone=i,this._doc=o,this._elementFocusedBeforeDrawerWasOpened=null,this._enableAnimations=!1,this._position="start",this._mode="over",this._disableClose=!1,this._autoFocus=!0,this._animationStarted=new f.Subject,this._animationEnd=new f.Subject,this._animationState="void",this.openedChange=new p.EventEmitter(!0),this._destroyed=new f.Subject,this.onPositionChanged=new p.EventEmitter,this._modeChanged=new f.Subject,this._opened=!1,this.openedChange.subscribe(function(e){e?(a._doc&&(a._elementFocusedBeforeDrawerWasOpened=a._doc.activeElement),a._isFocusTrapEnabled&&a._focusTrap&&a._trapFocus()):a._restoreFocus()}),this._ngZone.runOutsideAngular(function(){Object(f.fromEvent)(a._elementRef.nativeElement,"keydown").pipe(Object(h.filter)(function(e){return e.keyCode===u.ESCAPE&&!a.disableClose}),Object(h.takeUntil)(a._destroyed)).subscribe(function(e){return a._ngZone.run(function(){a.close(),e.stopPropagation()})})}),this._animationEnd.pipe(Object(h.distinctUntilChanged)(function(e,t){return e.fromState===t.fromState&&e.toState===t.toState})).subscribe(function(e){var t=e.fromState,n=e.toState;(0===n.indexOf("open")&&"void"===t||"void"===n&&0===t.indexOf("open"))&&a.openedChange.emit(a._opened)})}return Object.defineProperty(e.prototype,"position",{get:function(){return this._position},set:function(e){(e="end"===e?"end":"start")!=this._position&&(this._position=e,this.onPositionChanged.emit())},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"mode",{get:function(){return this._mode},set:function(e){this._mode=e,this._modeChanged.next()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disableClose",{get:function(){return this._disableClose},set:function(e){this._disableClose=Object(s.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"autoFocus",{get:function(){return this._autoFocus},set:function(e){this._autoFocus=Object(s.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_openedStream",{get:function(){return this.openedChange.pipe(Object(h.filter)(function(e){return e}),Object(h.map)(function(){}))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"openedStart",{get:function(){return this._animationStarted.pipe(Object(h.filter)(function(e){return e.fromState!==e.toState&&0===e.toState.indexOf("open")}),Object(h.map)(function(){}))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_closedStream",{get:function(){return this.openedChange.pipe(Object(h.filter)(function(e){return!e}),Object(h.map)(function(){}))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"closedStart",{get:function(){return this._animationStarted.pipe(Object(h.filter)(function(e){return e.fromState!==e.toState&&"void"===e.toState}),Object(h.map)(function(){}))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_isFocusTrapEnabled",{get:function(){return this.opened&&"side"!==this.mode},enumerable:!0,configurable:!0}),e.prototype._trapFocus=function(){var e=this;this.autoFocus&&this._focusTrap.focusInitialElementWhenReady().then(function(t){t||"function"!=typeof e._elementRef.nativeElement.focus||e._elementRef.nativeElement.focus()})},e.prototype._restoreFocus=function(){if(this.autoFocus){var e=this._doc&&this._doc.activeElement;e&&this._elementRef.nativeElement.contains(e)&&(this._elementFocusedBeforeDrawerWasOpened instanceof HTMLElement?this._focusMonitor.focusVia(this._elementFocusedBeforeDrawerWasOpened,this._openedVia):this._elementRef.nativeElement.blur()),this._elementFocusedBeforeDrawerWasOpened=null,this._openedVia=null}},e.prototype.ngAfterContentInit=function(){this._focusTrap=this._focusTrapFactory.create(this._elementRef.nativeElement),this._focusTrap.enabled=this._isFocusTrapEnabled},e.prototype.ngAfterContentChecked=function(){this._platform.isBrowser&&(this._enableAnimations=!0)},e.prototype.ngOnDestroy=function(){this._focusTrap&&this._focusTrap.destroy(),this._animationStarted.complete(),this._animationEnd.complete(),this._destroyed.next(),this._destroyed.complete()},Object.defineProperty(e.prototype,"opened",{get:function(){return this._opened},set:function(e){this.toggle(Object(s.coerceBooleanProperty)(e))},enumerable:!0,configurable:!0}),e.prototype.open=function(e){return this.toggle(!0,e)},e.prototype.close=function(){return this.toggle(!1)},e.prototype.toggle=function(e,t){var n=this;return void 0===e&&(e=!this.opened),void 0===t&&(t="program"),this._opened=e,e?(this._animationState=this._enableAnimations?"open":"open-instant",this._openedVia=t):(this._animationState="void",this._restoreFocus()),this._focusTrap&&(this._focusTrap.enabled=this._isFocusTrapEnabled),new Promise(function(e){n.openedChange.pipe(Object(h.take)(1)).subscribe(function(t){return e(t?"open":"close")})})},Object.defineProperty(e.prototype,"_width",{get:function(){return this._elementRef.nativeElement?this._elementRef.nativeElement.offsetWidth||0:0},enumerable:!0,configurable:!0}),e.decorators=[{type:p.Component,args:[{selector:"mat-drawer",exportAs:"matDrawer",template:'<div class="mat-drawer-inner-container"><ng-content></ng-content></div>',animations:[v.transformDrawer],host:{class:"mat-drawer","[@transform]":"_animationState","(@transform.start)":"_animationStarted.next($event)","(@transform.done)":"_animationEnd.next($event)","[attr.align]":"null","[class.mat-drawer-end]":'position === "end"',"[class.mat-drawer-over]":'mode === "over"',"[class.mat-drawer-push]":'mode === "push"',"[class.mat-drawer-side]":'mode === "side"',tabIndex:"-1"},changeDetection:p.ChangeDetectionStrategy.OnPush,encapsulation:p.ViewEncapsulation.None}]}],e.ctorParameters=function(){return[{type:p.ElementRef},{type:o.FocusTrapFactory},{type:o.FocusMonitor},{type:c.Platform},{type:p.NgZone},{type:void 0,decorators:[{type:p.Optional},{type:p.Inject,args:[d.DOCUMENT]}]}]},e.propDecorators={position:[{type:p.Input}],mode:[{type:p.Input}],disableClose:[{type:p.Input}],autoFocus:[{type:p.Input}],openedChange:[{type:p.Output}],_openedStream:[{type:p.Output,args:["opened"]}],openedStart:[{type:p.Output}],_closedStream:[{type:p.Output,args:["closed"]}],closedStart:[{type:p.Output}],onPositionChanged:[{type:p.Output,args:["positionChanged"]}],opened:[{type:p.Input}]},e}(),O=function(){function e(e,t,n,r,i,o,a){void 0===i&&(i=!1);var s=this;this._dir=e,this._element=t,this._ngZone=n,this._changeDetectorRef=r,this._animationMode=o,this.backdropClick=new p.EventEmitter,this._destroyed=new f.Subject,this._doCheckSubject=new f.Subject,this._contentMargins={left:null,right:null},this._contentMarginChanges=new f.Subject,e&&e.change.pipe(Object(h.takeUntil)(this._destroyed)).subscribe(function(){s._validateDrawers(),s._updateContentMargins()}),a&&a.change().pipe(Object(h.takeUntil)(this._destroyed)).subscribe(function(){return s._updateContentMargins()}),this._autosize=i}return Object.defineProperty(e.prototype,"start",{get:function(){return this._start},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"end",{get:function(){return this._end},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"autosize",{get:function(){return this._autosize},set:function(e){this._autosize=Object(s.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hasBackdrop",{get:function(){return null==this._backdropOverride?!this._start||"side"!==this._start.mode||!this._end||"side"!==this._end.mode:this._backdropOverride},set:function(e){this._backdropOverride=null==e?null:Object(s.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"scrollable",{get:function(){return this._userContent||this._content},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){var e=this;this._drawers.changes.pipe(Object(h.startWith)(null)).subscribe(function(){e._validateDrawers(),e._drawers.forEach(function(t){e._watchDrawerToggle(t),e._watchDrawerPosition(t),e._watchDrawerMode(t)}),(!e._drawers.length||e._isDrawerOpen(e._start)||e._isDrawerOpen(e._end))&&e._updateContentMargins(),e._changeDetectorRef.markForCheck()}),this._doCheckSubject.pipe(Object(h.debounceTime)(10),Object(h.takeUntil)(this._destroyed)).subscribe(function(){return e._updateContentMargins()})},e.prototype.ngOnDestroy=function(){this._doCheckSubject.complete(),this._destroyed.next(),this._destroyed.complete()},e.prototype.open=function(){this._drawers.forEach(function(e){return e.open()})},e.prototype.close=function(){this._drawers.forEach(function(e){return e.close()})},e.prototype.ngDoCheck=function(){var e=this;this._autosize&&this._isPushed()&&this._ngZone.runOutsideAngular(function(){return e._doCheckSubject.next()})},e.prototype._watchDrawerToggle=function(e){var t=this;e._animationStarted.pipe(Object(h.filter)(function(e){return e.fromState!==e.toState}),Object(h.takeUntil)(this._drawers.changes)).subscribe(function(e){"open-instant"!==e.toState&&"NoopAnimations"!==t._animationMode&&t._element.nativeElement.classList.add("mat-drawer-transition"),t._updateContentMargins(),t._changeDetectorRef.markForCheck()}),"side"!==e.mode&&e.openedChange.pipe(Object(h.takeUntil)(this._drawers.changes)).subscribe(function(){return t._setContainerClass(e.opened)})},e.prototype._watchDrawerPosition=function(e){var t=this;e&&e.onPositionChanged.pipe(Object(h.takeUntil)(this._drawers.changes)).subscribe(function(){t._ngZone.onMicrotaskEmpty.asObservable().pipe(Object(h.take)(1)).subscribe(function(){t._validateDrawers()})})},e.prototype._watchDrawerMode=function(e){var t=this;e&&e._modeChanged.pipe(Object(h.takeUntil)(Object(f.merge)(this._drawers.changes,this._destroyed))).subscribe(function(){t._updateContentMargins(),t._changeDetectorRef.markForCheck()})},e.prototype._setContainerClass=function(e){e?this._element.nativeElement.classList.add("mat-drawer-opened"):this._element.nativeElement.classList.remove("mat-drawer-opened")},e.prototype._validateDrawers=function(){var e=this;this._start=this._end=null,this._drawers.forEach(function(t){"end"==t.position?(null!=e._end&&_("end"),e._end=t):(null!=e._start&&_("start"),e._start=t)}),this._right=this._left=null,this._dir&&"rtl"===this._dir.value?(this._left=this._end,this._right=this._start):(this._left=this._start,this._right=this._end)},e.prototype._isPushed=function(){return this._isDrawerOpen(this._start)&&"over"!=this._start.mode||this._isDrawerOpen(this._end)&&"over"!=this._end.mode},e.prototype._onBackdropClicked=function(){this.backdropClick.emit(),this._closeModalDrawer()},e.prototype._closeModalDrawer=function(){var e=this;[this._start,this._end].filter(function(t){return t&&!t.disableClose&&e._canHaveBackdrop(t)}).forEach(function(e){return e.close()})},e.prototype._isShowingBackdrop=function(){return this._isDrawerOpen(this._start)&&this._canHaveBackdrop(this._start)||this._isDrawerOpen(this._end)&&this._canHaveBackdrop(this._end)},e.prototype._canHaveBackdrop=function(e){return"side"!==e.mode||!!this._backdropOverride},e.prototype._isDrawerOpen=function(e){return null!=e&&e.opened},e.prototype._updateContentMargins=function(){var e=this,t=0,n=0;if(this._left&&this._left.opened)if("side"==this._left.mode)t+=this._left._width;else if("push"==this._left.mode){t+=r=this._left._width,n-=r}if(this._right&&this._right.opened)if("side"==this._right.mode)n+=this._right._width;else if("push"==this._right.mode){var r;n+=r=this._right._width,t-=r}t=t||null,n=n||null,t===this._contentMargins.left&&n===this._contentMargins.right||(this._contentMargins={left:t,right:n},this._ngZone.run(function(){return e._contentMarginChanges.next(e._contentMargins)}))},e.decorators=[{type:p.Component,args:[{selector:"mat-drawer-container",exportAs:"matDrawerContainer",template:'<div class="mat-drawer-backdrop" (click)="_onBackdropClicked()" *ngIf="hasBackdrop" [class.mat-drawer-shown]="_isShowingBackdrop()"></div><ng-content select="mat-drawer"></ng-content><ng-content select="mat-drawer-content"></ng-content><mat-drawer-content *ngIf="!_content"><ng-content></ng-content></mat-drawer-content>',styles:[".mat-drawer-container{position:relative;z-index:1;box-sizing:border-box;-webkit-overflow-scrolling:touch;display:block;overflow:hidden}.mat-drawer-container[fullscreen]{top:0;left:0;right:0;bottom:0;position:absolute}.mat-drawer-container[fullscreen].mat-drawer-opened{overflow:hidden}.mat-drawer-container.mat-drawer-container-explicit-backdrop .mat-drawer-side{z-index:3}.mat-drawer-container.ng-animate-disabled .mat-drawer-backdrop,.mat-drawer-container.ng-animate-disabled .mat-drawer-content,.ng-animate-disabled .mat-drawer-container .mat-drawer-backdrop,.ng-animate-disabled .mat-drawer-container .mat-drawer-content{transition:none}.mat-drawer-backdrop{top:0;left:0;right:0;bottom:0;position:absolute;display:block;z-index:3;visibility:hidden}.mat-drawer-backdrop.mat-drawer-shown{visibility:visible}.mat-drawer-transition .mat-drawer-backdrop{transition-duration:.4s;transition-timing-function:cubic-bezier(.25,.8,.25,1);transition-property:background-color,visibility}@media (-ms-high-contrast:active){.mat-drawer-backdrop{opacity:.5}}.mat-drawer-content{position:relative;z-index:1;display:block;height:100%;overflow:auto}.mat-drawer-transition .mat-drawer-content{transition-duration:.4s;transition-timing-function:cubic-bezier(.25,.8,.25,1);transition-property:transform,margin-left,margin-right}.mat-drawer{position:relative;z-index:4;display:block;position:absolute;top:0;bottom:0;z-index:3;outline:0;box-sizing:border-box;overflow-y:auto;transform:translate3d(-100%,0,0)}@media (-ms-high-contrast:active){.mat-drawer,[dir=rtl] .mat-drawer.mat-drawer-end{border-right:solid 1px currentColor}}@media (-ms-high-contrast:active){.mat-drawer.mat-drawer-end,[dir=rtl] .mat-drawer{border-left:solid 1px currentColor;border-right:none}}.mat-drawer.mat-drawer-side{z-index:2}.mat-drawer.mat-drawer-end{right:0;transform:translate3d(100%,0,0)}[dir=rtl] .mat-drawer{transform:translate3d(100%,0,0)}[dir=rtl] .mat-drawer.mat-drawer-end{left:0;right:auto;transform:translate3d(-100%,0,0)}.mat-drawer-inner-container{width:100%;height:100%;overflow:auto;-webkit-overflow-scrolling:touch}.mat-sidenav-fixed{position:fixed}"],host:{class:"mat-drawer-container","[class.mat-drawer-container-explicit-backdrop]":"_backdropOverride"},changeDetection:p.ChangeDetectionStrategy.OnPush,encapsulation:p.ViewEncapsulation.None}]}],e.ctorParameters=function(){return[{type:a.Directionality,decorators:[{type:p.Optional}]},{type:p.ElementRef},{type:p.NgZone},{type:p.ChangeDetectorRef},{type:void 0,decorators:[{type:p.Inject,args:[y]}]},{type:String,decorators:[{type:p.Optional},{type:p.Inject,args:[m.ANIMATION_MODULE_TYPE]}]},{type:l.ViewportRuler,decorators:[{type:p.Optional}]}]},e.propDecorators={_drawers:[{type:p.ContentChildren,args:[x]}],_content:[{type:p.ContentChild,args:[j]}],_userContent:[{type:p.ViewChild,args:[j]}],autosize:[{type:p.Input}],hasBackdrop:[{type:p.Input}],backdropClick:[{type:p.Output}]},e}(),w=function(e){function t(t,n,r,i,o){return e.call(this,t,n,r,i,o)||this}return Object(i.__extends)(t,e),t.decorators=[{type:p.Component,args:[{selector:"mat-sidenav-content",template:"<ng-content></ng-content>",host:{class:"mat-drawer-content mat-sidenav-content","[style.margin-left.px]":"_container._contentMargins.left","[style.margin-right.px]":"_container._contentMargins.right"},changeDetection:p.ChangeDetectionStrategy.OnPush,encapsulation:p.ViewEncapsulation.None}]}],t.ctorParameters=function(){return[{type:p.ChangeDetectorRef},{type:E,decorators:[{type:p.Inject,args:[Object(p.forwardRef)(function(){return E})]}]},{type:p.ElementRef},{type:l.ScrollDispatcher},{type:p.NgZone}]},t}(j),C=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t._fixedInViewport=!1,t._fixedTopGap=0,t._fixedBottomGap=0,t}return Object(i.__extends)(t,e),Object.defineProperty(t.prototype,"fixedInViewport",{get:function(){return this._fixedInViewport},set:function(e){this._fixedInViewport=Object(s.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"fixedTopGap",{get:function(){return this._fixedTopGap},set:function(e){this._fixedTopGap=Object(s.coerceNumberProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"fixedBottomGap",{get:function(){return this._fixedBottomGap},set:function(e){this._fixedBottomGap=Object(s.coerceNumberProperty)(e)},enumerable:!0,configurable:!0}),t.decorators=[{type:p.Component,args:[{selector:"mat-sidenav",exportAs:"matSidenav",template:'<div class="mat-drawer-inner-container"><ng-content></ng-content></div>',animations:[v.transformDrawer],host:{class:"mat-drawer mat-sidenav",tabIndex:"-1","[@transform]":"_animationState","(@transform.start)":"_animationStarted.next($event)","(@transform.done)":"_animationEnd.next($event)","[attr.align]":"null","[class.mat-drawer-end]":'position === "end"',"[class.mat-drawer-over]":'mode === "over"',"[class.mat-drawer-push]":'mode === "push"',"[class.mat-drawer-side]":'mode === "side"',"[class.mat-sidenav-fixed]":"fixedInViewport","[style.top.px]":"fixedInViewport ? fixedTopGap : null","[style.bottom.px]":"fixedInViewport ? fixedBottomGap : null"},changeDetection:p.ChangeDetectionStrategy.OnPush,encapsulation:p.ViewEncapsulation.None}]}],t.propDecorators={fixedInViewport:[{type:p.Input}],fixedTopGap:[{type:p.Input}],fixedBottomGap:[{type:p.Input}]},t}(x),E=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(i.__extends)(t,e),t.decorators=[{type:p.Component,args:[{selector:"mat-sidenav-container",exportAs:"matSidenavContainer",template:'<div class="mat-drawer-backdrop" (click)="_onBackdropClicked()" *ngIf="hasBackdrop" [class.mat-drawer-shown]="_isShowingBackdrop()"></div><ng-content select="mat-sidenav"></ng-content><ng-content select="mat-sidenav-content"></ng-content><mat-sidenav-content *ngIf="!_content" cdkScrollable><ng-content></ng-content></mat-sidenav-content>',styles:[".mat-drawer-container{position:relative;z-index:1;box-sizing:border-box;-webkit-overflow-scrolling:touch;display:block;overflow:hidden}.mat-drawer-container[fullscreen]{top:0;left:0;right:0;bottom:0;position:absolute}.mat-drawer-container[fullscreen].mat-drawer-opened{overflow:hidden}.mat-drawer-container.mat-drawer-container-explicit-backdrop .mat-drawer-side{z-index:3}.mat-drawer-container.ng-animate-disabled .mat-drawer-backdrop,.mat-drawer-container.ng-animate-disabled .mat-drawer-content,.ng-animate-disabled .mat-drawer-container .mat-drawer-backdrop,.ng-animate-disabled .mat-drawer-container .mat-drawer-content{transition:none}.mat-drawer-backdrop{top:0;left:0;right:0;bottom:0;position:absolute;display:block;z-index:3;visibility:hidden}.mat-drawer-backdrop.mat-drawer-shown{visibility:visible}.mat-drawer-transition .mat-drawer-backdrop{transition-duration:.4s;transition-timing-function:cubic-bezier(.25,.8,.25,1);transition-property:background-color,visibility}@media (-ms-high-contrast:active){.mat-drawer-backdrop{opacity:.5}}.mat-drawer-content{position:relative;z-index:1;display:block;height:100%;overflow:auto}.mat-drawer-transition .mat-drawer-content{transition-duration:.4s;transition-timing-function:cubic-bezier(.25,.8,.25,1);transition-property:transform,margin-left,margin-right}.mat-drawer{position:relative;z-index:4;display:block;position:absolute;top:0;bottom:0;z-index:3;outline:0;box-sizing:border-box;overflow-y:auto;transform:translate3d(-100%,0,0)}@media (-ms-high-contrast:active){.mat-drawer,[dir=rtl] .mat-drawer.mat-drawer-end{border-right:solid 1px currentColor}}@media (-ms-high-contrast:active){.mat-drawer.mat-drawer-end,[dir=rtl] .mat-drawer{border-left:solid 1px currentColor;border-right:none}}.mat-drawer.mat-drawer-side{z-index:2}.mat-drawer.mat-drawer-end{right:0;transform:translate3d(100%,0,0)}[dir=rtl] .mat-drawer{transform:translate3d(100%,0,0)}[dir=rtl] .mat-drawer.mat-drawer-end{left:0;right:auto;transform:translate3d(-100%,0,0)}.mat-drawer-inner-container{width:100%;height:100%;overflow:auto;-webkit-overflow-scrolling:touch}.mat-sidenav-fixed{position:fixed}"],host:{class:"mat-drawer-container mat-sidenav-container","[class.mat-drawer-container-explicit-backdrop]":"_backdropOverride"},changeDetection:p.ChangeDetectionStrategy.OnPush,encapsulation:p.ViewEncapsulation.None}]}],t.propDecorators={_drawers:[{type:p.ContentChildren,args:[C]}],_content:[{type:p.ContentChild,args:[w]}]},t}(O),S=function(){function e(){}return e.decorators=[{type:p.NgModule,args:[{imports:[d.CommonModule,g.MatCommonModule,l.ScrollingModule,c.PlatformModule],exports:[g.MatCommonModule,x,O,j,C,E,w],declarations:[x,O,j,C,E,w]}]}],e}()},"./node_modules/@angular/material/esm5/slide-toggle.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatSlideToggleModule",function(){return j}),n.d(t,"MAT_SLIDE_TOGGLE_VALUE_ACCESSOR",function(){return g}),n.d(t,"MatSlideToggleChange",function(){return v}),n.d(t,"MatSlideToggleBase",function(){return _}),n.d(t,"_MatSlideToggleMixinBase",function(){return y}),n.d(t,"MatSlideToggle",function(){return b}),n.d(t,"MAT_SLIDE_TOGGLE_DEFAULT_OPTIONS",function(){return h});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),a=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),c=n("./node_modules/@angular/forms/fesm5/forms.js"),l=n("./node_modules/@angular/material/esm5/core.es5.js"),d=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),p=n("./node_modules/@angular/cdk/esm5/observers.es5.js"),f=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),h=new r.InjectionToken("mat-slide-toggle-default-options",{providedIn:"root",factory:function(){return{disableToggleValue:!1,disableDragValue:!1}}}),m=0,g={provide:c.NG_VALUE_ACCESSOR,useExisting:Object(r.forwardRef)(function(){return b}),multi:!0},v=function(){return function(e,t){this.source=e,this.checked=t}}(),_=function(){return function(e){this._elementRef=e}}(),y=Object(l.mixinTabIndex)(Object(l.mixinColor)(Object(l.mixinDisableRipple)(Object(l.mixinDisabled)(_)),"accent")),b=function(e){function t(t,n,i,o,a,s,u,c,l){var d=e.call(this,t)||this;return d._focusMonitor=i,d._changeDetectorRef=o,d._ngZone=s,d.defaults=u,d._animationMode=c,d._dir=l,d.onChange=function(e){},d.onTouched=function(){},d._uniqueId="mat-slide-toggle-"+ ++m,d._required=!1,d._checked=!1,d._dragging=!1,d.name=null,d.id=d._uniqueId,d.labelPosition="after",d.ariaLabel=null,d.ariaLabelledby=null,d.change=new r.EventEmitter,d.toggleChange=new r.EventEmitter,d.dragChange=new r.EventEmitter,d.tabIndex=parseInt(a)||0,d}return Object(i.__extends)(t,e),Object.defineProperty(t.prototype,"required",{get:function(){return this._required},set:function(e){this._required=Object(s.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"checked",{get:function(){return this._checked},set:function(e){this._checked=Object(s.coerceBooleanProperty)(e),this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"inputId",{get:function(){return(this.id||this._uniqueId)+"-input"},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){var e=this;this._focusMonitor.monitor(this._elementRef,!0).subscribe(function(t){t||Promise.resolve().then(function(){return e.onTouched()})})},t.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._elementRef)},t.prototype._onChangeEvent=function(e){e.stopPropagation(),this._dragging||this.toggleChange.emit(),this._dragging||this.defaults.disableToggleValue?this._inputElement.nativeElement.checked=this.checked:(this.checked=this._inputElement.nativeElement.checked,this._emitChangeEvent())},t.prototype._onInputClick=function(e){e.stopPropagation()},t.prototype.writeValue=function(e){this.checked=!!e},t.prototype.registerOnChange=function(e){this.onChange=e},t.prototype.registerOnTouched=function(e){this.onTouched=e},t.prototype.setDisabledState=function(e){this.disabled=e,this._changeDetectorRef.markForCheck()},t.prototype.focus=function(){this._focusMonitor.focusVia(this._inputElement,"keyboard")},t.prototype.toggle=function(){this.checked=!this.checked,this.onChange(this.checked)},t.prototype._emitChangeEvent=function(){this.onChange(this.checked),this.change.emit(new v(this,this.checked))},t.prototype._getDragPercentage=function(e){var t=e/this._thumbBarWidth*100;return this._previousChecked&&(t+=100),Math.max(0,Math.min(t,100))},t.prototype._onDragStart=function(){if(!this.disabled&&!this._dragging){var e=this._thumbEl.nativeElement;this._thumbBarWidth=this._thumbBarEl.nativeElement.clientWidth-e.clientWidth,e.classList.add("mat-dragging"),this._previousChecked=this.checked,this._dragging=!0}},t.prototype._onDrag=function(e){if(this._dragging){var t=this._dir&&"rtl"===this._dir.value?-1:1;this._dragPercentage=this._getDragPercentage(e.deltaX*t);var n=this._dragPercentage/100*this._thumbBarWidth*t;this._thumbEl.nativeElement.style.transform="translate3d("+n+"px, 0, 0)"}},t.prototype._onDragEnd=function(){var e=this;if(this._dragging){var t=this._dragPercentage>50;t!==this.checked&&(this.dragChange.emit(),this.defaults.disableDragValue||(this.checked=t,this._emitChangeEvent())),this._ngZone.runOutsideAngular(function(){return setTimeout(function(){e._dragging&&(e._dragging=!1,e._thumbEl.nativeElement.classList.remove("mat-dragging"),e._thumbEl.nativeElement.style.transform="")})})}},t.prototype._onLabelTextChange=function(){this._changeDetectorRef.detectChanges()},t.decorators=[{type:r.Component,args:[{selector:"mat-slide-toggle",exportAs:"matSlideToggle",host:{class:"mat-slide-toggle","[id]":"id","[attr.tabindex]":"-1","[class.mat-checked]":"checked","[class.mat-disabled]":"disabled","[class.mat-slide-toggle-label-before]":'labelPosition == "before"',"[class._mat-animation-noopable]":'_animationMode === "NoopAnimations"',"(focus)":"_inputElement.nativeElement.focus()"},template:'<label [attr.for]="inputId" class="mat-slide-toggle-label" #label><div #toggleBar class="mat-slide-toggle-bar" [class.mat-slide-toggle-bar-no-side-margin]="!labelContent.textContent || !labelContent.textContent.trim()"><input #input class="mat-slide-toggle-input cdk-visually-hidden" type="checkbox" [id]="inputId" [required]="required" [tabIndex]="tabIndex" [checked]="checked" [disabled]="disabled" [attr.name]="name" [attr.aria-label]="ariaLabel" [attr.aria-labelledby]="ariaLabelledby" (change)="_onChangeEvent($event)" (click)="_onInputClick($event)"><div class="mat-slide-toggle-thumb-container" #thumbContainer (slidestart)="_onDragStart()" (slide)="_onDrag($event)" (slideend)="_onDragEnd()"><div class="mat-slide-toggle-thumb"></div><div class="mat-slide-toggle-ripple" mat-ripple [matRippleTrigger]="label" [matRippleDisabled]="disableRipple || disabled" [matRippleCentered]="true" [matRippleRadius]="20" [matRippleAnimation]="{enterDuration: 150}"><div class="mat-ripple-element mat-slide-toggle-persistent-ripple"></div></div></div></div><span class="mat-slide-toggle-content" #labelContent (cdkObserveContent)="_onLabelTextChange()"><span style="display:none">&nbsp;</span><ng-content></ng-content></span></label>',styles:[".mat-slide-toggle{display:inline-block;height:24px;max-width:100%;line-height:24px;white-space:nowrap;outline:0;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;-webkit-tap-highlight-color:transparent}.mat-slide-toggle.mat-checked .mat-slide-toggle-thumb-container{transform:translate3d(16px,0,0)}[dir=rtl] .mat-slide-toggle.mat-checked .mat-slide-toggle-thumb-container{transform:translate3d(-16px,0,0)}.mat-slide-toggle.mat-disabled{opacity:.38}.mat-slide-toggle.mat-disabled .mat-slide-toggle-label,.mat-slide-toggle.mat-disabled .mat-slide-toggle-thumb-container{cursor:default}.mat-slide-toggle-label{display:flex;flex:1;flex-direction:row;align-items:center;height:inherit;cursor:pointer}.mat-slide-toggle-content{white-space:nowrap;overflow:hidden;text-overflow:ellipsis}.mat-slide-toggle-label-before .mat-slide-toggle-label{order:1}.mat-slide-toggle-label-before .mat-slide-toggle-bar{order:2}.mat-slide-toggle-bar,[dir=rtl] .mat-slide-toggle-label-before .mat-slide-toggle-bar{margin-right:8px;margin-left:0}.mat-slide-toggle-label-before .mat-slide-toggle-bar,[dir=rtl] .mat-slide-toggle-bar{margin-left:8px;margin-right:0}.mat-slide-toggle-bar-no-side-margin{margin-left:0;margin-right:0}.mat-slide-toggle-thumb-container{position:absolute;z-index:1;width:20px;height:20px;top:-3px;left:0;transform:translate3d(0,0,0);transition:all 80ms linear;transition-property:transform;cursor:-webkit-grab;cursor:grab}.mat-slide-toggle-thumb-container.mat-dragging,.mat-slide-toggle-thumb-container:active{cursor:-webkit-grabbing;cursor:grabbing;transition-duration:0s}._mat-animation-noopable .mat-slide-toggle-thumb-container{transition:none}[dir=rtl] .mat-slide-toggle-thumb-container{left:auto;right:0}.mat-slide-toggle-thumb{height:20px;width:20px;border-radius:50%}.mat-slide-toggle-bar{position:relative;width:36px;height:14px;flex-shrink:0;border-radius:8px}.mat-slide-toggle-input{bottom:0;left:10px}[dir=rtl] .mat-slide-toggle-input{left:auto;right:10px}.mat-slide-toggle-bar,.mat-slide-toggle-thumb{transition:all 80ms linear;transition-property:background-color;transition-delay:50ms}._mat-animation-noopable .mat-slide-toggle-bar,._mat-animation-noopable .mat-slide-toggle-thumb{transition:none}.mat-slide-toggle .mat-slide-toggle-ripple{position:absolute;top:calc(50% - 20px);left:calc(50% - 20px);height:40px;width:40px;z-index:1;pointer-events:none}.mat-slide-toggle .mat-slide-toggle-ripple .mat-ripple-element:not(.mat-slide-toggle-persistent-ripple){opacity:.12}.mat-slide-toggle-persistent-ripple{width:100%;height:100%;transform:none}.mat-slide-toggle-bar:hover .mat-slide-toggle-persistent-ripple{opacity:.04}.mat-slide-toggle:not(.mat-disabled).cdk-focused .mat-slide-toggle-persistent-ripple{opacity:.12}.mat-slide-toggle-persistent-ripple,.mat-slide-toggle.mat-disabled .mat-slide-toggle-bar:hover .mat-slide-toggle-persistent-ripple{opacity:0}@media (hover:none){.mat-slide-toggle-bar:hover .mat-slide-toggle-persistent-ripple{display:none}}@media (-ms-high-contrast:active){.mat-slide-toggle-thumb{background:#fff;border:1px solid #000}.mat-slide-toggle.mat-checked .mat-slide-toggle-thumb{background:#000;border:1px solid #fff}.mat-slide-toggle-bar{background:#fff}.mat-slide-toggle.cdk-keyboard-focused .mat-slide-toggle-bar{outline:1px dotted;outline-offset:5px}}@media (-ms-high-contrast:black-on-white){.mat-slide-toggle-bar{border:1px solid #000}}"],providers:[g],inputs:["disabled","disableRipple","color","tabIndex"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:u.Platform},{type:o.FocusMonitor},{type:r.ChangeDetectorRef},{type:String,decorators:[{type:r.Attribute,args:["tabindex"]}]},{type:r.NgZone},{type:void 0,decorators:[{type:r.Inject,args:[h]}]},{type:String,decorators:[{type:r.Optional},{type:r.Inject,args:[d.ANIMATION_MODULE_TYPE]}]},{type:a.Directionality,decorators:[{type:r.Optional}]}]},t.propDecorators={_thumbEl:[{type:r.ViewChild,args:["thumbContainer"]}],_thumbBarEl:[{type:r.ViewChild,args:["toggleBar"]}],name:[{type:r.Input}],id:[{type:r.Input}],labelPosition:[{type:r.Input}],ariaLabel:[{type:r.Input,args:["aria-label"]}],ariaLabelledby:[{type:r.Input,args:["aria-labelledby"]}],required:[{type:r.Input}],checked:[{type:r.Input}],change:[{type:r.Output}],toggleChange:[{type:r.Output}],dragChange:[{type:r.Output}],_inputElement:[{type:r.ViewChild,args:["input"]}]},t}(y),j=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[l.MatRippleModule,l.MatCommonModule,p.ObserversModule],exports:[b,l.MatCommonModule],declarations:[b],providers:[{provide:f.HAMMER_GESTURE_CONFIG,useClass:l.GestureConfig}]}]}],e}()},"./node_modules/@angular/material/esm5/slider.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatSliderModule",function(){return b}),n.d(t,"MAT_SLIDER_VALUE_ACCESSOR",function(){return m}),n.d(t,"MatSliderChange",function(){return g}),n.d(t,"MatSliderBase",function(){return v}),n.d(t,"_MatSliderMixinBase",function(){return _}),n.d(t,"MatSlider",function(){return y});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),o=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),a=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),s=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),u=n("./node_modules/@angular/core/fesm5/core.js"),c=n("./node_modules/@angular/forms/fesm5/forms.js"),l=n("./node_modules/@angular/material/esm5/core.es5.js"),d=n("./node_modules/@angular/platform-browser/fesm5/animations.js"),p=n("./node_modules/rxjs/_esm5/index.js"),f=n("./node_modules/@angular/common/fesm5/common.js"),h=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),m={provide:c.NG_VALUE_ACCESSOR,useExisting:Object(u.forwardRef)(function(){return y}),multi:!0},g=function(){return function(){}}(),v=function(){return function(e){this._elementRef=e}}(),_=Object(l.mixinTabIndex)(Object(l.mixinColor)(Object(l.mixinDisabled)(v),"accent")),y=function(e){function t(t,n,r,i,o,a){var s=e.call(this,t)||this;return s._focusMonitor=n,s._changeDetectorRef=r,s._dir=i,s._animationMode=a,s._invert=!1,s._max=100,s._min=0,s._step=1,s._thumbLabel=!1,s._tickInterval=0,s._value=null,s._vertical=!1,s.change=new u.EventEmitter,s.input=new u.EventEmitter,s.valueChange=new u.EventEmitter,s.onTouched=function(){},s._percent=0,s._isSliding=!1,s._isActive=!1,s._tickIntervalPercent=0,s._sliderDimensions=null,s._controlValueAccessorChangeFn=function(){},s._dirChangeSubscription=p.Subscription.EMPTY,s.tabIndex=parseInt(o)||0,s}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"invert",{get:function(){return this._invert},set:function(e){this._invert=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"max",{get:function(){return this._max},set:function(e){this._max=Object(a.coerceNumberProperty)(e,this._max),this._percent=this._calculatePercentage(this._value),this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"min",{get:function(){return this._min},set:function(e){this._min=Object(a.coerceNumberProperty)(e,this._min),null===this._value&&(this.value=this._min),this._percent=this._calculatePercentage(this._value),this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"step",{get:function(){return this._step},set:function(e){this._step=Object(a.coerceNumberProperty)(e,this._step),this._step%1!=0&&(this._roundToDecimal=this._step.toString().split(".").pop().length),this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"thumbLabel",{get:function(){return this._thumbLabel},set:function(e){this._thumbLabel=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"tickInterval",{get:function(){return this._tickInterval},set:function(e){this._tickInterval="auto"===e?"auto":"number"==typeof e||"string"==typeof e?Object(a.coerceNumberProperty)(e,this._tickInterval):0},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{get:function(){return null===this._value&&(this.value=this._min),this._value},set:function(e){if(e!==this._value){var t=Object(a.coerceNumberProperty)(e);this._roundToDecimal&&(t=parseFloat(t.toFixed(this._roundToDecimal))),this._value=t,this._percent=this._calculatePercentage(this._value),this._changeDetectorRef.markForCheck()}},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"vertical",{get:function(){return this._vertical},set:function(e){this._vertical=Object(a.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"displayValue",{get:function(){return this.displayWith?this.displayWith(this.value):this._roundToDecimal&&this.value&&this.value%1!=0?this.value.toFixed(this._roundToDecimal):this.value||0},enumerable:!0,configurable:!0}),t.prototype.focus=function(){this._focusHostElement()},t.prototype.blur=function(){this._blurHostElement()},Object.defineProperty(t.prototype,"percent",{get:function(){return this._clamp(this._percent)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_invertAxis",{get:function(){return this.vertical?!this.invert:this.invert},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_isMinValue",{get:function(){return 0===this.percent},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_thumbGap",{get:function(){return this.disabled?7:this._isMinValue&&!this.thumbLabel?this._isActive?10:7:0},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_trackBackgroundStyles",{get:function(){var e=this.vertical?"Y":"X",t=this.vertical?"1, "+(1-this.percent)+", 1":1-this.percent+", 1, 1";return{transform:"translate"+e+"("+(this._shouldInvertMouseCoords()?"-":"")+this._thumbGap+"px) scale3d("+t+")"}},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_trackFillStyles",{get:function(){var e=this.vertical?"Y":"X",t=this.vertical?"1, "+this.percent+", 1":this.percent+", 1, 1";return{transform:"translate"+e+"("+(this._shouldInvertMouseCoords()?"":"-")+this._thumbGap+"px) scale3d("+t+")"}},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_ticksContainerStyles",{get:function(){return{transform:"translate"+(this.vertical?"Y":"X")+"("+(this.vertical||"rtl"!=this._getDirection()?"-":"")+this._tickIntervalPercent/2*100+"%)"}},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_ticksStyles",{get:function(){var e=100*this._tickIntervalPercent,t={backgroundSize:this.vertical?"2px "+e+"%":e+"% 2px",transform:"translateZ(0) translate"+(this.vertical?"Y":"X")+"("+(this.vertical||"rtl"!=this._getDirection()?"":"-")+e/2+"%)"+(this.vertical||"rtl"!=this._getDirection()?"":" rotate(180deg)")};this._isMinValue&&this._thumbGap&&(t["padding"+(this.vertical?this._invertAxis?"Bottom":"Top":this._invertAxis?"Right":"Left")]=this._thumbGap+"px");return t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_thumbContainerStyles",{get:function(){return{transform:"translate"+(this.vertical?"Y":"X")+"(-"+100*(("rtl"!=this._getDirection()||this.vertical?this._invertAxis:!this._invertAxis)?this.percent:1-this.percent)+"%)"}},enumerable:!0,configurable:!0}),t.prototype._shouldInvertMouseCoords=function(){return"rtl"!=this._getDirection()||this.vertical?this._invertAxis:!this._invertAxis},t.prototype._getDirection=function(){return this._dir&&"rtl"==this._dir.value?"rtl":"ltr"},t.prototype.ngOnInit=function(){var e=this;this._focusMonitor.monitor(this._elementRef,!0).subscribe(function(t){e._isActive=!!t&&"keyboard"!==t,e._changeDetectorRef.detectChanges()}),this._dir&&(this._dirChangeSubscription=this._dir.change.subscribe(function(){e._changeDetectorRef.markForCheck()}))},t.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._elementRef),this._dirChangeSubscription.unsubscribe()},t.prototype._onMouseenter=function(){this.disabled||(this._sliderDimensions=this._getSliderDimensions(),this._updateTickIntervalPercent())},t.prototype._onMousedown=function(e){if(!this.disabled&&0===e.button){var t=this.value;this._isSliding=!1,this._focusHostElement(),this._updateValueFromPosition({x:e.clientX,y:e.clientY}),t!=this.value&&(this._emitInputEvent(),this._emitChangeEvent())}},t.prototype._onSlide=function(e){if(!this.disabled){this._isSliding||this._onSlideStart(null),e.preventDefault();var t=this.value;this._updateValueFromPosition({x:e.center.x,y:e.center.y}),t!=this.value&&this._emitInputEvent()}},t.prototype._onSlideStart=function(e){this.disabled||this._isSliding||(this._onMouseenter(),this._isSliding=!0,this._focusHostElement(),this._valueOnSlideStart=this.value,e&&(this._updateValueFromPosition({x:e.center.x,y:e.center.y}),e.preventDefault()))},t.prototype._onSlideEnd=function(){this._isSliding=!1,this._valueOnSlideStart==this.value||this.disabled||this._emitChangeEvent(),this._valueOnSlideStart=null},t.prototype._onFocus=function(){this._sliderDimensions=this._getSliderDimensions(),this._updateTickIntervalPercent()},t.prototype._onBlur=function(){this.onTouched()},t.prototype._onKeydown=function(e){if(!this.disabled&&!Object(s.hasModifierKey)(e)){var t=this.value;switch(e.keyCode){case s.PAGE_UP:this._increment(10);break;case s.PAGE_DOWN:this._increment(-10);break;case s.END:this.value=this.max;break;case s.HOME:this.value=this.min;break;case s.LEFT_ARROW:this._increment("rtl"==this._getDirection()?1:-1);break;case s.UP_ARROW:this._increment(1);break;case s.RIGHT_ARROW:this._increment("rtl"==this._getDirection()?-1:1);break;case s.DOWN_ARROW:this._increment(-1);break;default:return}t!=this.value&&(this._emitInputEvent(),this._emitChangeEvent()),this._isSliding=!0,e.preventDefault()}},t.prototype._onKeyup=function(){this._isSliding=!1},t.prototype._increment=function(e){this.value=this._clamp((this.value||0)+this.step*e,this.min,this.max)},t.prototype._updateValueFromPosition=function(e){if(this._sliderDimensions){var t=this.vertical?this._sliderDimensions.top:this._sliderDimensions.left,n=this.vertical?this._sliderDimensions.height:this._sliderDimensions.width,r=this.vertical?e.y:e.x,i=this._clamp((r-t)/n);if(this._shouldInvertMouseCoords()&&(i=1-i),0===i)this.value=this.min;else if(1===i)this.value=this.max;else{var o=this._calculateValue(i),a=Math.round((o-this.min)/this.step)*this.step+this.min;this.value=this._clamp(a,this.min,this.max)}}},t.prototype._emitChangeEvent=function(){this._controlValueAccessorChangeFn(this.value),this.valueChange.emit(this.value),this.change.emit(this._createChangeEvent())},t.prototype._emitInputEvent=function(){this.input.emit(this._createChangeEvent())},t.prototype._updateTickIntervalPercent=function(){if(this.tickInterval&&this._sliderDimensions)if("auto"==this.tickInterval){var e=this.vertical?this._sliderDimensions.height:this._sliderDimensions.width,t=e*this.step/(this.max-this.min),n=Math.ceil(30/t)*this.step;this._tickIntervalPercent=n/e}else this._tickIntervalPercent=this.tickInterval*this.step/(this.max-this.min)},t.prototype._createChangeEvent=function(e){void 0===e&&(e=this.value);var t=new g;return t.source=this,t.value=e,t},t.prototype._calculatePercentage=function(e){return((e||0)-this.min)/(this.max-this.min)},t.prototype._calculateValue=function(e){return this.min+e*(this.max-this.min)},t.prototype._clamp=function(e,t,n){return void 0===t&&(t=0),void 0===n&&(n=1),Math.max(t,Math.min(e,n))},t.prototype._getSliderDimensions=function(){return this._sliderWrapper?this._sliderWrapper.nativeElement.getBoundingClientRect():null},t.prototype._focusHostElement=function(){this._elementRef.nativeElement.focus()},t.prototype._blurHostElement=function(){this._elementRef.nativeElement.blur()},t.prototype.writeValue=function(e){this.value=e},t.prototype.registerOnChange=function(e){this._controlValueAccessorChangeFn=e},t.prototype.registerOnTouched=function(e){this.onTouched=e},t.prototype.setDisabledState=function(e){this.disabled=e},t.decorators=[{type:u.Component,args:[{selector:"mat-slider",exportAs:"matSlider",providers:[m],host:{"(focus)":"_onFocus()","(blur)":"_onBlur()","(mousedown)":"_onMousedown($event)","(keydown)":"_onKeydown($event)","(keyup)":"_onKeyup()","(mouseenter)":"_onMouseenter()","(slide)":"_onSlide($event)","(slideend)":"_onSlideEnd()","(slidestart)":"_onSlideStart($event)",class:"mat-slider",role:"slider","[tabIndex]":"tabIndex","[attr.aria-disabled]":"disabled","[attr.aria-valuemax]":"max","[attr.aria-valuemin]":"min","[attr.aria-valuenow]":"value","[attr.aria-orientation]":'vertical ? "vertical" : "horizontal"',"[class.mat-slider-disabled]":"disabled","[class.mat-slider-has-ticks]":"tickInterval","[class.mat-slider-horizontal]":"!vertical","[class.mat-slider-axis-inverted]":"_invertAxis","[class.mat-slider-sliding]":"_isSliding","[class.mat-slider-thumb-label-showing]":"thumbLabel","[class.mat-slider-vertical]":"vertical","[class.mat-slider-min-value]":"_isMinValue","[class.mat-slider-hide-last-tick]":"disabled || _isMinValue && _thumbGap && _invertAxis","[class._mat-animation-noopable]":'_animationMode === "NoopAnimations"'},template:'<div class="mat-slider-wrapper" #sliderWrapper><div class="mat-slider-track-wrapper"><div class="mat-slider-track-background" [ngStyle]="_trackBackgroundStyles"></div><div class="mat-slider-track-fill" [ngStyle]="_trackFillStyles"></div></div><div class="mat-slider-ticks-container" [ngStyle]="_ticksContainerStyles"><div class="mat-slider-ticks" [ngStyle]="_ticksStyles"></div></div><div class="mat-slider-thumb-container" [ngStyle]="_thumbContainerStyles"><div class="mat-slider-focus-ring"></div><div class="mat-slider-thumb"></div><div class="mat-slider-thumb-label"><span class="mat-slider-thumb-label-text">{{displayValue}}</span></div></div></div>',styles:[".mat-slider{display:inline-block;position:relative;box-sizing:border-box;padding:8px;outline:0;vertical-align:middle}.mat-slider-wrapper{position:absolute}.mat-slider-track-wrapper{position:absolute;top:0;left:0;overflow:hidden}.mat-slider-track-fill{position:absolute;transform-origin:0 0;transition:transform .4s cubic-bezier(.25,.8,.25,1),background-color .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-track-background{position:absolute;transform-origin:100% 100%;transition:transform .4s cubic-bezier(.25,.8,.25,1),background-color .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-ticks-container{position:absolute;left:0;top:0;overflow:hidden}.mat-slider-ticks{background-repeat:repeat;background-clip:content-box;box-sizing:border-box;opacity:0;transition:opacity .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-thumb-container{position:absolute;z-index:1;transition:transform .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-focus-ring{position:absolute;width:30px;height:30px;border-radius:50%;transform:scale(0);opacity:0;transition:transform .4s cubic-bezier(.25,.8,.25,1),background-color .4s cubic-bezier(.25,.8,.25,1),opacity .4s cubic-bezier(.25,.8,.25,1)}.cdk-keyboard-focused .mat-slider-focus-ring,.cdk-program-focused .mat-slider-focus-ring{transform:scale(1);opacity:1}.mat-slider:not(.mat-slider-disabled) .mat-slider-thumb,.mat-slider:not(.mat-slider-disabled) .mat-slider-thumb-label{cursor:-webkit-grab;cursor:grab}.mat-slider-sliding:not(.mat-slider-disabled) .mat-slider-thumb,.mat-slider-sliding:not(.mat-slider-disabled) .mat-slider-thumb-label,.mat-slider:not(.mat-slider-disabled) .mat-slider-thumb-label:active,.mat-slider:not(.mat-slider-disabled) .mat-slider-thumb:active{cursor:-webkit-grabbing;cursor:grabbing}.mat-slider-thumb{position:absolute;right:-10px;bottom:-10px;box-sizing:border-box;width:20px;height:20px;border:3px solid transparent;border-radius:50%;transform:scale(.7);transition:transform .4s cubic-bezier(.25,.8,.25,1),background-color .4s cubic-bezier(.25,.8,.25,1),border-color .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-thumb-label{display:none;align-items:center;justify-content:center;position:absolute;width:28px;height:28px;border-radius:50%;transition:transform .4s cubic-bezier(.25,.8,.25,1),border-radius .4s cubic-bezier(.25,.8,.25,1),background-color .4s cubic-bezier(.25,.8,.25,1)}@media (-ms-high-contrast:active){.mat-slider-thumb-label{outline:solid 1px}}.mat-slider-thumb-label-text{z-index:1;opacity:0;transition:opacity .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-sliding .mat-slider-thumb-container,.mat-slider-sliding .mat-slider-track-background,.mat-slider-sliding .mat-slider-track-fill{transition-duration:0s}.mat-slider-has-ticks .mat-slider-wrapper::after{content:'';position:absolute;border-width:0;border-style:solid;opacity:0;transition:opacity .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-has-ticks.cdk-focused:not(.mat-slider-hide-last-tick) .mat-slider-wrapper::after,.mat-slider-has-ticks:hover:not(.mat-slider-hide-last-tick) .mat-slider-wrapper::after{opacity:1}.mat-slider-has-ticks.cdk-focused:not(.mat-slider-disabled) .mat-slider-ticks,.mat-slider-has-ticks:hover:not(.mat-slider-disabled) .mat-slider-ticks{opacity:1}.mat-slider-thumb-label-showing .mat-slider-focus-ring{transform:scale(0);opacity:0}.mat-slider-thumb-label-showing .mat-slider-thumb-label{display:flex}.mat-slider-axis-inverted .mat-slider-track-fill{transform-origin:100% 100%}.mat-slider-axis-inverted .mat-slider-track-background{transform-origin:0 0}.mat-slider:not(.mat-slider-disabled).cdk-focused.mat-slider-thumb-label-showing .mat-slider-thumb{transform:scale(0)}.mat-slider:not(.mat-slider-disabled).cdk-focused .mat-slider-thumb-label{border-radius:50% 50% 0}.mat-slider:not(.mat-slider-disabled).cdk-focused .mat-slider-thumb-label-text{opacity:1}.mat-slider:not(.mat-slider-disabled).cdk-mouse-focused .mat-slider-thumb,.mat-slider:not(.mat-slider-disabled).cdk-program-focused .mat-slider-thumb,.mat-slider:not(.mat-slider-disabled).cdk-touch-focused .mat-slider-thumb{border-width:2px;transform:scale(1)}.mat-slider-disabled .mat-slider-focus-ring{transform:scale(0);opacity:0}.mat-slider-disabled .mat-slider-thumb{border-width:4px;transform:scale(.5)}.mat-slider-disabled .mat-slider-thumb-label{display:none}.mat-slider-horizontal{height:48px;min-width:128px}.mat-slider-horizontal .mat-slider-wrapper{height:2px;top:23px;left:8px;right:8px}.mat-slider-horizontal .mat-slider-wrapper::after{height:2px;border-left-width:2px;right:0;top:0}.mat-slider-horizontal .mat-slider-track-wrapper{height:2px;width:100%}.mat-slider-horizontal .mat-slider-track-fill{height:2px;width:100%;transform:scaleX(0)}.mat-slider-horizontal .mat-slider-track-background{height:2px;width:100%;transform:scaleX(1)}.mat-slider-horizontal .mat-slider-ticks-container{height:2px;width:100%}@media (-ms-high-contrast:active){.mat-slider-horizontal .mat-slider-ticks-container{height:0;outline:solid 2px;top:1px}}.mat-slider-horizontal .mat-slider-ticks{height:2px;width:100%}.mat-slider-horizontal .mat-slider-thumb-container{width:100%;height:0;top:50%}.mat-slider-horizontal .mat-slider-focus-ring{top:-15px;right:-15px}.mat-slider-horizontal .mat-slider-thumb-label{right:-14px;top:-40px;transform:translateY(26px) scale(.01) rotate(45deg)}.mat-slider-horizontal .mat-slider-thumb-label-text{transform:rotate(-45deg)}.mat-slider-horizontal.cdk-focused .mat-slider-thumb-label{transform:rotate(45deg)}@media (-ms-high-contrast:active){.mat-slider-horizontal.cdk-focused .mat-slider-thumb-label,.mat-slider-horizontal.cdk-focused .mat-slider-thumb-label-text{transform:none}}.mat-slider-vertical{width:48px;min-height:128px}.mat-slider-vertical .mat-slider-wrapper{width:2px;top:8px;bottom:8px;left:23px}.mat-slider-vertical .mat-slider-wrapper::after{width:2px;border-top-width:2px;bottom:0;left:0}.mat-slider-vertical .mat-slider-track-wrapper{height:100%;width:2px}.mat-slider-vertical .mat-slider-track-fill{height:100%;width:2px;transform:scaleY(0)}.mat-slider-vertical .mat-slider-track-background{height:100%;width:2px;transform:scaleY(1)}.mat-slider-vertical .mat-slider-ticks-container{width:2px;height:100%}@media (-ms-high-contrast:active){.mat-slider-vertical .mat-slider-ticks-container{width:0;outline:solid 2px;left:1px}}.mat-slider-vertical .mat-slider-focus-ring{bottom:-15px;left:-15px}.mat-slider-vertical .mat-slider-ticks{width:2px;height:100%}.mat-slider-vertical .mat-slider-thumb-container{height:100%;width:0;left:50%}.mat-slider-vertical .mat-slider-thumb{-webkit-backface-visibility:hidden;backface-visibility:hidden}.mat-slider-vertical .mat-slider-thumb-label{bottom:-14px;left:-40px;transform:translateX(26px) scale(.01) rotate(-45deg)}.mat-slider-vertical .mat-slider-thumb-label-text{transform:rotate(45deg)}.mat-slider-vertical.cdk-focused .mat-slider-thumb-label{transform:rotate(-45deg)}[dir=rtl] .mat-slider-wrapper::after{left:0;right:auto}[dir=rtl] .mat-slider-horizontal .mat-slider-track-fill{transform-origin:100% 100%}[dir=rtl] .mat-slider-horizontal .mat-slider-track-background{transform-origin:0 0}[dir=rtl] .mat-slider-horizontal.mat-slider-axis-inverted .mat-slider-track-fill{transform-origin:0 0}[dir=rtl] .mat-slider-horizontal.mat-slider-axis-inverted .mat-slider-track-background{transform-origin:100% 100%}.mat-slider._mat-animation-noopable .mat-slider-focus-ring,.mat-slider._mat-animation-noopable .mat-slider-has-ticks .mat-slider-wrapper::after,.mat-slider._mat-animation-noopable .mat-slider-thumb,.mat-slider._mat-animation-noopable .mat-slider-thumb-container,.mat-slider._mat-animation-noopable .mat-slider-thumb-label,.mat-slider._mat-animation-noopable .mat-slider-thumb-label-text,.mat-slider._mat-animation-noopable .mat-slider-ticks,.mat-slider._mat-animation-noopable .mat-slider-track-background,.mat-slider._mat-animation-noopable .mat-slider-track-fill{transition:none}"],inputs:["disabled","color","tabIndex"],encapsulation:u.ViewEncapsulation.None,changeDetection:u.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:u.ElementRef},{type:i.FocusMonitor},{type:u.ChangeDetectorRef},{type:o.Directionality,decorators:[{type:u.Optional}]},{type:String,decorators:[{type:u.Attribute,args:["tabindex"]}]},{type:String,decorators:[{type:u.Optional},{type:u.Inject,args:[d.ANIMATION_MODULE_TYPE]}]}]},t.propDecorators={invert:[{type:u.Input}],max:[{type:u.Input}],min:[{type:u.Input}],step:[{type:u.Input}],thumbLabel:[{type:u.Input}],tickInterval:[{type:u.Input}],value:[{type:u.Input}],displayWith:[{type:u.Input}],vertical:[{type:u.Input}],change:[{type:u.Output}],input:[{type:u.Output}],valueChange:[{type:u.Output}],_sliderWrapper:[{type:u.ViewChild,args:["sliderWrapper"]}]},t}(_),b=function(){function e(){}return e.decorators=[{type:u.NgModule,args:[{imports:[f.CommonModule,l.MatCommonModule],exports:[y,l.MatCommonModule],declarations:[y],providers:[{provide:h.HAMMER_GESTURE_CONFIG,useClass:l.GestureConfig}]}]}],e}()},"./node_modules/@angular/material/esm5/snack-bar.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatSnackBarModule",function(){return j}),n.d(t,"MAT_SNACK_BAR_DEFAULT_OPTIONS_FACTORY",function(){return O}),n.d(t,"MAT_SNACK_BAR_DEFAULT_OPTIONS",function(){return x}),n.d(t,"MatSnackBar",function(){return w}),n.d(t,"MatSnackBarContainer",function(){return b}),n.d(t,"MAT_SNACK_BAR_DATA",function(){return g}),n.d(t,"MatSnackBarConfig",function(){return v}),n.d(t,"MatSnackBarRef",function(){return m}),n.d(t,"SimpleSnackBar",function(){return _}),n.d(t,"matSnackBarAnimations",function(){return y});var r=n("./node_modules/rxjs/_esm5/index.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/animations/fesm5/animations.js"),a=n("./node_modules/tslib/tslib.es6.js"),s=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),u=n("./node_modules/rxjs/_esm5/operators/index.js"),c=n("./node_modules/@angular/cdk/esm5/overlay.es5.js"),l=n("./node_modules/@angular/common/fesm5/common.js"),d=n("./node_modules/@angular/material/esm5/core.es5.js"),p=n("./node_modules/@angular/material/esm5/button.es5.js"),f=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),h=n("./node_modules/@angular/cdk/esm5/layout.es5.js"),m=function(){function e(e,t){var n=this;this._overlayRef=t,this._afterDismissed=new r.Subject,this._afterOpened=new r.Subject,this._onAction=new r.Subject,this._dismissedByAction=!1,this.containerInstance=e,this.onAction().subscribe(function(){return n.dismiss()}),e._onExit.subscribe(function(){return n._finishDismiss()})}return e.prototype.dismiss=function(){this._afterDismissed.closed||this.containerInstance.exit(),clearTimeout(this._durationTimeoutId)},e.prototype.dismissWithAction=function(){this._onAction.closed||(this._dismissedByAction=!0,this._onAction.next(),this._onAction.complete())},e.prototype.closeWithAction=function(){this.dismissWithAction()},e.prototype._dismissAfter=function(e){var t=this;this._durationTimeoutId=setTimeout(function(){return t.dismiss()},e)},e.prototype._open=function(){this._afterOpened.closed||(this._afterOpened.next(),this._afterOpened.complete())},e.prototype._finishDismiss=function(){this._overlayRef.dispose(),this._onAction.closed||this._onAction.complete(),this._afterDismissed.next({dismissedByAction:this._dismissedByAction}),this._afterDismissed.complete(),this._dismissedByAction=!1},e.prototype.afterDismissed=function(){return this._afterDismissed.asObservable()},e.prototype.afterOpened=function(){return this.containerInstance._onEnter},e.prototype.onAction=function(){return this._onAction.asObservable()},e}(),g=new i.InjectionToken("MatSnackBarData"),v=function(){return function(){this.politeness="assertive",this.announcementMessage="",this.duration=0,this.data=null,this.horizontalPosition="center",this.verticalPosition="bottom"}}(),_=function(){function e(e,t){this.snackBarRef=e,this.data=t}return e.prototype.action=function(){this.snackBarRef.dismissWithAction()},Object.defineProperty(e.prototype,"hasAction",{get:function(){return!!this.data.action},enumerable:!0,configurable:!0}),e.decorators=[{type:i.Component,args:[{selector:"simple-snack-bar",template:'<span>{{data.message}}</span><div class="mat-simple-snackbar-action" *ngIf="hasAction"><button mat-button (click)="action()">{{data.action}}</button></div>',styles:[".mat-simple-snackbar{display:flex;justify-content:space-between;align-items:center;height:100%;line-height:20px;opacity:1}.mat-simple-snackbar-action{flex-shrink:0;margin:-8px -8px -8px 8px}.mat-simple-snackbar-action button{max-height:36px;min-width:0}[dir=rtl] .mat-simple-snackbar-action{margin-left:-8px;margin-right:8px}"],encapsulation:i.ViewEncapsulation.None,changeDetection:i.ChangeDetectionStrategy.OnPush,host:{class:"mat-simple-snackbar"}}]}],e.ctorParameters=function(){return[{type:m},{type:void 0,decorators:[{type:i.Inject,args:[g]}]}]},e}(),y={snackBarState:Object(o.trigger)("state",[Object(o.state)("void, hidden",Object(o.style)({transform:"scale(0.8)",opacity:0})),Object(o.state)("visible",Object(o.style)({transform:"scale(1)",opacity:1})),Object(o.transition)("* => visible",Object(o.animate)("150ms cubic-bezier(0, 0, 0.2, 1)")),Object(o.transition)("* => void, * => hidden",Object(o.animate)("75ms cubic-bezier(0.4, 0.0, 1, 1)",Object(o.style)({opacity:0})))])},b=function(e){function t(t,n,i,o){var a=e.call(this)||this;return a._ngZone=t,a._elementRef=n,a._changeDetectorRef=i,a.snackBarConfig=o,a._destroyed=!1,a._onExit=new r.Subject,a._onEnter=new r.Subject,a._animationState="void","assertive"!==o.politeness||o.announcementMessage?"off"===o.politeness?a._role=null:a._role="status":a._role="alert",a}return Object(a.__extends)(t,e),t.prototype.attachComponentPortal=function(e){return this._assertNotAttached(),this._applySnackBarClasses(),this._portalOutlet.attachComponentPortal(e)},t.prototype.attachTemplatePortal=function(e){return this._assertNotAttached(),this._applySnackBarClasses(),this._portalOutlet.attachTemplatePortal(e)},t.prototype.onAnimationEnd=function(e){var t=e.fromState,n=e.toState;if(("void"===n&&"void"!==t||"hidden"===n)&&this._completeExit(),"visible"===n){var r=this._onEnter;this._ngZone.run(function(){r.next(),r.complete()})}},t.prototype.enter=function(){this._destroyed||(this._animationState="visible",this._changeDetectorRef.detectChanges())},t.prototype.exit=function(){return this._animationState="hidden",this._onExit},t.prototype.ngOnDestroy=function(){this._destroyed=!0,this._completeExit()},t.prototype._completeExit=function(){var e=this;this._ngZone.onMicrotaskEmpty.asObservable().pipe(Object(u.take)(1)).subscribe(function(){e._onExit.next(),e._onExit.complete()})},t.prototype._applySnackBarClasses=function(){var e=this._elementRef.nativeElement,t=this.snackBarConfig.panelClass;t&&(Array.isArray(t)?t.forEach(function(t){return e.classList.add(t)}):e.classList.add(t)),"center"===this.snackBarConfig.horizontalPosition&&e.classList.add("mat-snack-bar-center"),"top"===this.snackBarConfig.verticalPosition&&e.classList.add("mat-snack-bar-top")},t.prototype._assertNotAttached=function(){if(this._portalOutlet.hasAttached())throw Error("Attempting to attach snack bar content after content is already attached")},t.decorators=[{type:i.Component,args:[{selector:"snack-bar-container",template:"<ng-template cdkPortalOutlet></ng-template>",styles:[".mat-snack-bar-container{border-radius:4px;box-sizing:border-box;display:block;margin:24px;max-width:33vw;min-width:344px;padding:14px 16px;min-height:48px;transform-origin:center}@media (-ms-high-contrast:active){.mat-snack-bar-container{border:solid 1px}}.mat-snack-bar-handset{width:100%}.mat-snack-bar-handset .mat-snack-bar-container{margin:8px;max-width:100%;min-width:0;width:100%}"],changeDetection:i.ChangeDetectionStrategy.OnPush,encapsulation:i.ViewEncapsulation.None,animations:[y.snackBarState],host:{"[attr.role]":"_role",class:"mat-snack-bar-container","[@state]":"_animationState","(@state.done)":"onAnimationEnd($event)"}}]}],t.ctorParameters=function(){return[{type:i.NgZone},{type:i.ElementRef},{type:i.ChangeDetectorRef},{type:v}]},t.propDecorators={_portalOutlet:[{type:i.ViewChild,args:[s.CdkPortalOutlet]}]},t}(s.BasePortalOutlet),j=function(){function e(){}return e.decorators=[{type:i.NgModule,args:[{imports:[c.OverlayModule,s.PortalModule,l.CommonModule,p.MatButtonModule,d.MatCommonModule],exports:[b,d.MatCommonModule],declarations:[b,_],entryComponents:[b,_]}]}],e}(),x=new i.InjectionToken("mat-snack-bar-default-options",{providedIn:"root",factory:O});function O(){return new v}var w=function(){function e(e,t,n,r,i,o){this._overlay=e,this._live=t,this._injector=n,this._breakpointObserver=r,this._parentSnackBar=i,this._defaultConfig=o,this._snackBarRefAtThisLevel=null}return Object.defineProperty(e.prototype,"_openedSnackBarRef",{get:function(){var e=this._parentSnackBar;return e?e._openedSnackBarRef:this._snackBarRefAtThisLevel},set:function(e){this._parentSnackBar?this._parentSnackBar._openedSnackBarRef=e:this._snackBarRefAtThisLevel=e},enumerable:!0,configurable:!0}),e.prototype.openFromComponent=function(e,t){return this._attach(e,t)},e.prototype.openFromTemplate=function(e,t){return this._attach(e,t)},e.prototype.open=function(e,t,n){void 0===t&&(t="");var r=Object(a.__assign)({},this._defaultConfig,n);return r.data={message:e,action:t},r.announcementMessage||(r.announcementMessage=e),this.openFromComponent(_,r)},e.prototype.dismiss=function(){this._openedSnackBarRef&&this._openedSnackBarRef.dismiss()},e.prototype.ngOnDestroy=function(){this._snackBarRefAtThisLevel&&this._snackBarRefAtThisLevel.dismiss()},e.prototype._attachSnackBarContainer=function(e,t){var n=t&&t.viewContainerRef&&t.viewContainerRef.injector,r=new s.PortalInjector(n||this._injector,new WeakMap([[v,t]])),i=new s.ComponentPortal(b,t.viewContainerRef,r),o=e.attach(i);return o.instance.snackBarConfig=t,o.instance},e.prototype._attach=function(e,t){var n=Object(a.__assign)({},new v,this._defaultConfig,t),r=this._createOverlay(n),o=this._attachSnackBarContainer(r,n),c=new m(o,r);if(e instanceof i.TemplateRef){var l=new s.TemplatePortal(e,null,{$implicit:n.data,snackBarRef:c});c.instance=o.attachTemplatePortal(l)}else{var d=this._createInjector(n,c),p=(l=new s.ComponentPortal(e,void 0,d),o.attachComponentPortal(l));c.instance=p.instance}return this._breakpointObserver.observe(h.Breakpoints.Handset).pipe(Object(u.takeUntil)(r.detachments().pipe(Object(u.take)(1)))).subscribe(function(e){e.matches?r.overlayElement.classList.add("mat-snack-bar-handset"):r.overlayElement.classList.remove("mat-snack-bar-handset")}),this._animateSnackBar(c,n),this._openedSnackBarRef=c,this._openedSnackBarRef},e.prototype._animateSnackBar=function(e,t){var n=this;e.afterDismissed().subscribe(function(){n._openedSnackBarRef==e&&(n._openedSnackBarRef=null),t.announcementMessage&&n._live.clear()}),this._openedSnackBarRef?(this._openedSnackBarRef.afterDismissed().subscribe(function(){e.containerInstance.enter()}),this._openedSnackBarRef.dismiss()):e.containerInstance.enter(),t.duration&&t.duration>0&&e.afterOpened().subscribe(function(){return e._dismissAfter(t.duration)}),t.announcementMessage&&this._live.announce(t.announcementMessage,t.politeness)},e.prototype._createOverlay=function(e){var t=new c.OverlayConfig;t.direction=e.direction;var n=this._overlay.position().global(),r="rtl"===e.direction,i="left"===e.horizontalPosition||"start"===e.horizontalPosition&&!r||"end"===e.horizontalPosition&&r,o=!i&&"center"!==e.horizontalPosition;return i?n.left("0"):o?n.right("0"):n.centerHorizontally(),"top"===e.verticalPosition?n.top("0"):n.bottom("0"),t.positionStrategy=n,this._overlay.create(t)},e.prototype._createInjector=function(e,t){var n=e&&e.viewContainerRef&&e.viewContainerRef.injector;return new s.PortalInjector(n||this._injector,new WeakMap([[m,t],[g,e.data]]))},e.decorators=[{type:i.Injectable,args:[{providedIn:j}]}],e.ctorParameters=function(){return[{type:c.Overlay},{type:f.LiveAnnouncer},{type:i.Injector},{type:h.BreakpointObserver},{type:e,decorators:[{type:i.Optional},{type:i.SkipSelf}]},{type:v,decorators:[{type:i.Inject,args:[x]}]}]},e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e(Object(i.inject)(c.Overlay),Object(i.inject)(f.LiveAnnouncer),Object(i.inject)(i.INJECTOR),Object(i.inject)(h.BreakpointObserver),Object(i.inject)(e,12),Object(i.inject)(x))},token:e,providedIn:j}),e}()},"./node_modules/@angular/material/esm5/sort.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatSortModule",function(){return j}),n.d(t,"MatSortHeaderBase",function(){return _}),n.d(t,"_MatSortHeaderMixinBase",function(){return y}),n.d(t,"MatSortHeader",function(){return b}),n.d(t,"MAT_SORT_HEADER_INTL_PROVIDER_FACTORY",function(){return g}),n.d(t,"MatSortHeaderIntl",function(){return m}),n.d(t,"MAT_SORT_HEADER_INTL_PROVIDER",function(){return v}),n.d(t,"MatSortBase",function(){return l}),n.d(t,"_MatSortMixinBase",function(){return d}),n.d(t,"MatSort",function(){return p}),n.d(t,"matSortAnimations",function(){return h});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/material/esm5/core.es5.js"),s=n("./node_modules/rxjs/_esm5/index.js"),u=n("./node_modules/@angular/animations/fesm5/animations.js"),c=n("./node_modules/@angular/common/fesm5/common.js");var l=function(){return function(){}}(),d=Object(a.mixinInitialized)(Object(a.mixinDisabled)(l)),p=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.sortables=new Map,t._stateChanges=new s.Subject,t.start="asc",t._direction="",t.sortChange=new o.EventEmitter,t}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"direction",{get:function(){return this._direction},set:function(e){if(Object(o.isDevMode)()&&e&&"asc"!==e&&"desc"!==e)throw Error(e+" is not a valid sort direction ('asc' or 'desc').");this._direction=e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disableClear",{get:function(){return this._disableClear},set:function(e){this._disableClear=Object(i.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t.prototype.register=function(e){if(!e.id)throw Error("MatSortHeader must be provided with a unique id.");if(this.sortables.has(e.id))throw t=e.id,Error("Cannot have two MatSortables with the same id ("+t+").");var t;this.sortables.set(e.id,e)},t.prototype.deregister=function(e){this.sortables.delete(e.id)},t.prototype.sort=function(e){this.active!=e.id?(this.active=e.id,this.direction=e.start?e.start:this.start):this.direction=this.getNextSortDirection(e),this.sortChange.emit({active:this.active,direction:this.direction})},t.prototype.getNextSortDirection=function(e){if(!e)return"";var t=null!=e.disableClear?e.disableClear:this.disableClear,n=function(e,t){var n=["asc","desc"];"desc"==e&&n.reverse();t||n.push("");return n}(e.start||this.start,t),r=n.indexOf(this.direction)+1;return r>=n.length&&(r=0),n[r]},t.prototype.ngOnInit=function(){this._markInitialized()},t.prototype.ngOnChanges=function(){this._stateChanges.next()},t.prototype.ngOnDestroy=function(){this._stateChanges.complete()},t.decorators=[{type:o.Directive,args:[{selector:"[matSort]",exportAs:"matSort",inputs:["disabled: matSortDisabled"]}]}],t.propDecorators={active:[{type:o.Input,args:["matSortActive"]}],start:[{type:o.Input,args:["matSortStart"]}],direction:[{type:o.Input,args:["matSortDirection"]}],disableClear:[{type:o.Input,args:["matSortDisableClear"]}],sortChange:[{type:o.Output,args:["matSortChange"]}]},t}(d);var f=a.AnimationDurations.ENTERING+" "+a.AnimationCurves.STANDARD_CURVE,h={indicator:Object(u.trigger)("indicator",[Object(u.state)("active-asc, asc",Object(u.style)({transform:"translateY(0px)"})),Object(u.state)("active-desc, desc",Object(u.style)({transform:"translateY(10px)"})),Object(u.transition)("active-asc <=> active-desc",Object(u.animate)(f))]),leftPointer:Object(u.trigger)("leftPointer",[Object(u.state)("active-asc, asc",Object(u.style)({transform:"rotate(-45deg)"})),Object(u.state)("active-desc, desc",Object(u.style)({transform:"rotate(45deg)"})),Object(u.transition)("active-asc <=> active-desc",Object(u.animate)(f))]),rightPointer:Object(u.trigger)("rightPointer",[Object(u.state)("active-asc, asc",Object(u.style)({transform:"rotate(45deg)"})),Object(u.state)("active-desc, desc",Object(u.style)({transform:"rotate(-45deg)"})),Object(u.transition)("active-asc <=> active-desc",Object(u.animate)(f))]),arrowOpacity:Object(u.trigger)("arrowOpacity",[Object(u.state)("desc-to-active, asc-to-active, active",Object(u.style)({opacity:1})),Object(u.state)("desc-to-hint, asc-to-hint, hint",Object(u.style)({opacity:.54})),Object(u.state)("hint-to-desc, active-to-desc, desc, hint-to-asc, active-to-asc, asc, void",Object(u.style)({opacity:0})),Object(u.transition)("* => asc, * => desc, * => active, * => hint, * => void",Object(u.animate)("0ms")),Object(u.transition)("* <=> *",Object(u.animate)(f))]),arrowPosition:Object(u.trigger)("arrowPosition",[Object(u.transition)("* => desc-to-hint, * => desc-to-active",Object(u.animate)(f,Object(u.keyframes)([Object(u.style)({transform:"translateY(-25%)"}),Object(u.style)({transform:"translateY(0)"})]))),Object(u.transition)("* => hint-to-desc, * => active-to-desc",Object(u.animate)(f,Object(u.keyframes)([Object(u.style)({transform:"translateY(0)"}),Object(u.style)({transform:"translateY(25%)"})]))),Object(u.transition)("* => asc-to-hint, * => asc-to-active",Object(u.animate)(f,Object(u.keyframes)([Object(u.style)({transform:"translateY(25%)"}),Object(u.style)({transform:"translateY(0)"})]))),Object(u.transition)("* => hint-to-asc, * => active-to-asc",Object(u.animate)(f,Object(u.keyframes)([Object(u.style)({transform:"translateY(0)"}),Object(u.style)({transform:"translateY(-25%)"})]))),Object(u.state)("desc-to-hint, asc-to-hint, hint, desc-to-active, asc-to-active, active",Object(u.style)({transform:"translateY(0)"})),Object(u.state)("hint-to-desc, active-to-desc, desc",Object(u.style)({transform:"translateY(-25%)"})),Object(u.state)("hint-to-asc, active-to-asc, asc",Object(u.style)({transform:"translateY(25%)"}))]),allowChildren:Object(u.trigger)("allowChildren",[Object(u.transition)("* <=> *",[Object(u.query)("@*",Object(u.animateChild)(),{optional:!0})])])},m=function(){function e(){this.changes=new s.Subject,this.sortButtonLabel=function(e){return"Change sorting for "+e}}return e.decorators=[{type:o.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(o.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}();function g(e){return e||new m}var v={provide:m,deps:[[new o.Optional,new o.SkipSelf,m]],useFactory:g},_=function(){return function(){}}(),y=Object(a.mixinDisabled)(_),b=function(e){function t(t,n,r,i){var o=e.call(this)||this;if(o._intl=t,o._sort=r,o._columnDef=i,o._showIndicatorHint=!1,o._arrowDirection="",o._disableViewStateAnimation=!1,o.arrowPosition="after",!r)throw Error("MatSortHeader must be placed within a parent element with the MatSort directive.");return o._rerenderSubscription=Object(s.merge)(r.sortChange,r._stateChanges,t.changes).subscribe(function(){o._isSorted()&&o._updateArrowDirection(),!o._isSorted()&&o._viewState&&"active"===o._viewState.toState&&(o._disableViewStateAnimation=!1,o._setAnimationTransitionState({fromState:"active",toState:o._arrowDirection})),n.markForCheck()}),o}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"disableClear",{get:function(){return this._disableClear},set:function(e){this._disableClear=Object(i.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){!this.id&&this._columnDef&&(this.id=this._columnDef.name),this._updateArrowDirection(),this._setAnimationTransitionState({toState:this._isSorted()?"active":this._arrowDirection}),this._sort.register(this)},t.prototype.ngOnDestroy=function(){this._sort.deregister(this),this._rerenderSubscription.unsubscribe()},t.prototype._setIndicatorHintVisible=function(e){this._isDisabled()&&e||(this._showIndicatorHint=e,this._isSorted()||(this._updateArrowDirection(),this._showIndicatorHint?this._setAnimationTransitionState({fromState:this._arrowDirection,toState:"hint"}):this._setAnimationTransitionState({fromState:"hint",toState:this._arrowDirection})))},t.prototype._setAnimationTransitionState=function(e){this._viewState=e,this._disableViewStateAnimation&&(this._viewState={toState:e.toState})},t.prototype._handleClick=function(){if(!this._isDisabled()){this._sort.sort(this),"hint"!==this._viewState.toState&&"active"!==this._viewState.toState||(this._disableViewStateAnimation=!0);var e=this._isSorted()?{fromState:this._arrowDirection,toState:"active"}:{fromState:"active",toState:this._arrowDirection};this._setAnimationTransitionState(e),this._showIndicatorHint=!1}},t.prototype._isSorted=function(){return this._sort.active==this.id&&("asc"===this._sort.direction||"desc"===this._sort.direction)},t.prototype._getArrowDirectionState=function(){return(this._isSorted()?"active-":"")+this._arrowDirection},t.prototype._getArrowViewState=function(){var e=this._viewState.fromState;return(e?e+"-to-":"")+this._viewState.toState},t.prototype._updateArrowDirection=function(){this._arrowDirection=this._isSorted()?this._sort.direction:this.start||this._sort.start},t.prototype._isDisabled=function(){return this._sort.disabled||this.disabled},t.prototype._getAriaSortAttribute=function(){return this._isSorted()?"asc"==this._sort.direction?"ascending":"descending":null},t.decorators=[{type:o.Component,args:[{selector:"[mat-sort-header]",exportAs:"matSortHeader",template:'<div class="mat-sort-header-container" [class.mat-sort-header-sorted]="_isSorted()" [class.mat-sort-header-position-before]="arrowPosition == \'before\'"><button class="mat-sort-header-button" type="button" [attr.disabled]="_isDisabled() || null" [attr.aria-label]="_intl.sortButtonLabel(id)" (focus)="_setIndicatorHintVisible(true)" (blur)="_setIndicatorHintVisible(false)"><ng-content></ng-content></button><div class="mat-sort-header-arrow" [@arrowOpacity]="_getArrowViewState()" [@arrowPosition]="_getArrowViewState()" [@allowChildren]="_getArrowDirectionState()" (@arrowPosition.start)="_disableViewStateAnimation = true" (@arrowPosition.done)="_disableViewStateAnimation = false"><div class="mat-sort-header-stem"></div><div class="mat-sort-header-indicator" [@indicator]="_getArrowDirectionState()"><div class="mat-sort-header-pointer-left" [@leftPointer]="_getArrowDirectionState()"></div><div class="mat-sort-header-pointer-right" [@rightPointer]="_getArrowDirectionState()"></div><div class="mat-sort-header-pointer-middle"></div></div></div></div>',styles:[".mat-sort-header-container{display:flex;cursor:pointer;align-items:center}.mat-sort-header-disabled .mat-sort-header-container{cursor:default}.mat-sort-header-position-before{flex-direction:row-reverse}.mat-sort-header-button{border:none;background:0 0;display:flex;align-items:center;padding:0;cursor:inherit;outline:0;font:inherit;color:currentColor}.mat-sort-header-arrow{height:12px;width:12px;min-width:12px;position:relative;display:flex;opacity:0}.mat-sort-header-arrow,[dir=rtl] .mat-sort-header-position-before .mat-sort-header-arrow{margin:0 0 0 6px}.mat-sort-header-position-before .mat-sort-header-arrow,[dir=rtl] .mat-sort-header-arrow{margin:0 6px 0 0}.mat-sort-header-stem{background:currentColor;height:10px;width:2px;margin:auto;display:flex;align-items:center}@media (-ms-high-contrast:active){.mat-sort-header-stem{width:0;border-left:solid 2px}}.mat-sort-header-indicator{width:100%;height:2px;display:flex;align-items:center;position:absolute;top:0;left:0}.mat-sort-header-pointer-middle{margin:auto;height:2px;width:2px;background:currentColor;transform:rotate(45deg)}@media (-ms-high-contrast:active){.mat-sort-header-pointer-middle{width:0;height:0;border-top:solid 2px;border-left:solid 2px}}.mat-sort-header-pointer-left,.mat-sort-header-pointer-right{background:currentColor;width:6px;height:2px;position:absolute;top:0}@media (-ms-high-contrast:active){.mat-sort-header-pointer-left,.mat-sort-header-pointer-right{width:0;height:0;border-left:solid 6px;border-top:solid 2px}}.mat-sort-header-pointer-left{transform-origin:right;left:0}.mat-sort-header-pointer-right{transform-origin:left;right:0}"],host:{"(click)":"_handleClick()","(mouseenter)":"_setIndicatorHintVisible(true)","(longpress)":"_setIndicatorHintVisible(true)","(mouseleave)":"_setIndicatorHintVisible(false)","[attr.aria-sort]":"_getAriaSortAttribute()","[class.mat-sort-header-disabled]":"_isDisabled()"},encapsulation:o.ViewEncapsulation.None,changeDetection:o.ChangeDetectionStrategy.OnPush,inputs:["disabled"],animations:[h.indicator,h.leftPointer,h.rightPointer,h.arrowOpacity,h.arrowPosition,h.allowChildren]}]}],t.ctorParameters=function(){return[{type:m},{type:o.ChangeDetectorRef},{type:p,decorators:[{type:o.Optional}]},{type:void 0,decorators:[{type:o.Inject,args:["MAT_SORT_HEADER_COLUMN_DEF"]},{type:o.Optional}]}]},t.propDecorators={id:[{type:o.Input,args:["mat-sort-header"]}],arrowPosition:[{type:o.Input}],start:[{type:o.Input}],disableClear:[{type:o.Input}]},t}(y),j=function(){function e(){}return e.decorators=[{type:o.NgModule,args:[{imports:[c.CommonModule],exports:[p,b],declarations:[p,b],providers:[v]}]}],e}()},"./node_modules/@angular/material/esm5/stepper.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatStepperModule",function(){return M}),n.d(t,"MatStepLabel",function(){return g}),n.d(t,"MatStep",function(){return O}),n.d(t,"MatStepper",function(){return w}),n.d(t,"MatHorizontalStepper",function(){return C}),n.d(t,"MatVerticalStepper",function(){return E}),n.d(t,"MatStepperNext",function(){return S}),n.d(t,"MatStepperPrevious",function(){return k}),n.d(t,"MatStepHeader",function(){return b}),n.d(t,"MAT_STEPPER_INTL_PROVIDER_FACTORY",function(){return _}),n.d(t,"MatStepperIntl",function(){return v}),n.d(t,"MAT_STEPPER_INTL_PROVIDER",function(){return y}),n.d(t,"matStepperAnimations",function(){return j}),n.d(t,"MatStepperIcon",function(){return x});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/core/fesm5/core.js"),o=n("./node_modules/@angular/cdk/esm5/stepper.es5.js"),a=n("./node_modules/rxjs/_esm5/index.js"),s=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),u=n("./node_modules/@angular/animations/fesm5/animations.js"),c=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),l=n("./node_modules/@angular/common/fesm5/common.js"),d=n("./node_modules/@angular/material/esm5/core.es5.js"),p=n("./node_modules/rxjs/_esm5/operators/index.js"),f=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),h=n("./node_modules/@angular/material/esm5/button.es5.js"),m=n("./node_modules/@angular/material/esm5/icon.es5.js"),g=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"[matStepLabel]"}]}],t}(o.CdkStepLabel),v=function(){function e(){this.changes=new a.Subject,this.optionalLabel="Optional"}return e.decorators=[{type:i.Injectable,args:[{providedIn:"root"}]}],e.ngInjectableDef=Object(i.defineInjectable)({factory:function(){return new e},token:e,providedIn:"root"}),e}();function _(e){return e||new v}var y={provide:v,deps:[[new i.Optional,new i.SkipSelf,v]],useFactory:_},b=function(e){function t(t,n,r,i){var o=e.call(this,r)||this;return o._intl=t,o._focusMonitor=n,n.monitor(r,!0),o._intlSubscription=t.changes.subscribe(function(){return i.markForCheck()}),o}return Object(r.__extends)(t,e),t.prototype.ngOnDestroy=function(){this._intlSubscription.unsubscribe(),this._focusMonitor.stopMonitoring(this._elementRef)},t.prototype._stringLabel=function(){return this.label instanceof g?null:this.label},t.prototype._templateLabel=function(){return this.label instanceof g?this.label:null},t.prototype._getHostElement=function(){return this._elementRef.nativeElement},t.prototype._getIconContext=function(){return{index:this.index,active:this.active,optional:this.optional}},t.decorators=[{type:i.Component,args:[{selector:"mat-step-header",template:'<div class="mat-step-header-ripple" mat-ripple [matRippleTrigger]="_getHostElement()"></div><div class="mat-step-icon-state-{{state}} mat-step-icon" [class.mat-step-icon-selected]="selected" [ngSwitch]="state"><div class="mat-step-icon-content"><ng-container *ngSwitchCase="\'number\'" [ngSwitch]="!!(iconOverrides && iconOverrides.number)"><ng-container *ngSwitchCase="true" [ngTemplateOutlet]="iconOverrides.number" [ngTemplateOutletContext]="_getIconContext()"></ng-container><span *ngSwitchDefault>{{index + 1}}</span></ng-container><ng-container *ngSwitchCase="\'edit\'" [ngSwitch]="!!(iconOverrides && iconOverrides.edit)"><ng-container *ngSwitchCase="true" [ngTemplateOutlet]="iconOverrides.edit" [ngTemplateOutletContext]="_getIconContext()"></ng-container><mat-icon *ngSwitchDefault>create</mat-icon></ng-container><ng-container *ngSwitchCase="\'done\'" [ngSwitch]="!!(iconOverrides && iconOverrides.done)"><ng-container *ngSwitchCase="true" [ngTemplateOutlet]="iconOverrides.done" [ngTemplateOutletContext]="_getIconContext()"></ng-container><mat-icon *ngSwitchDefault>done</mat-icon></ng-container><ng-container *ngSwitchCase="\'error\'" [ngSwitch]="!!(iconOverrides && iconOverrides.error)"><ng-container *ngSwitchCase="true" [ngTemplateOutlet]="iconOverrides.error" [ngTemplateOutletContext]="_getIconContext()"></ng-container><mat-icon *ngSwitchDefault>warning</mat-icon></ng-container><ng-container *ngSwitchDefault [ngSwitch]="!!(iconOverrides && iconOverrides[state])"><ng-container *ngSwitchCase="true" [ngTemplateOutlet]="iconOverrides[state]" [ngTemplateOutletContext]="_getIconContext()"></ng-container><mat-icon *ngSwitchDefault>{{state}}</mat-icon></ng-container></div></div><div class="mat-step-label" [class.mat-step-label-active]="active" [class.mat-step-label-selected]="selected" [class.mat-step-label-error]="state == \'error\'"><ng-container *ngIf="_templateLabel()" [ngTemplateOutlet]="_templateLabel()!.template"></ng-container><div class="mat-step-text-label" *ngIf="_stringLabel()">{{label}}</div><div class="mat-step-optional" *ngIf="optional && state != \'error\'">{{_intl.optionalLabel}}</div><div class="mat-step-sub-label-error" *ngIf="state == \'error\'">{{errorMessage}}</div></div>',styles:[".mat-step-header{overflow:hidden;outline:0;cursor:pointer;position:relative;box-sizing:content-box;-webkit-tap-highlight-color:transparent}.mat-step-optional,.mat-step-sub-label-error{font-size:12px}.mat-step-icon{border-radius:50%;height:24px;width:24px;flex-shrink:0;position:relative}.mat-step-icon .mat-icon,.mat-step-icon-content{position:absolute;top:50%;left:50%;transform:translate(-50%,-50%)}.mat-step-icon .mat-icon{font-size:16px;height:16px;width:16px}.mat-step-icon-state-error .mat-icon{font-size:24px;height:24px;width:24px}.mat-step-label{display:inline-block;white-space:nowrap;overflow:hidden;text-overflow:ellipsis;min-width:50px;vertical-align:middle}.mat-step-text-label{text-overflow:ellipsis;overflow:hidden}.mat-step-header .mat-step-header-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}"],host:{class:"mat-step-header",role:"tab"},encapsulation:i.ViewEncapsulation.None,changeDetection:i.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:v},{type:s.FocusMonitor},{type:i.ElementRef},{type:i.ChangeDetectorRef}]},t.propDecorators={state:[{type:i.Input}],label:[{type:i.Input}],errorMessage:[{type:i.Input}],iconOverrides:[{type:i.Input}],index:[{type:i.Input}],selected:[{type:i.Input}],active:[{type:i.Input}],optional:[{type:i.Input}]},t}(o.CdkStepHeader),j={horizontalStepTransition:Object(u.trigger)("stepTransition",[Object(u.state)("previous",Object(u.style)({transform:"translate3d(-100%, 0, 0)",visibility:"hidden"})),Object(u.state)("current",Object(u.style)({transform:"none",visibility:"visible"})),Object(u.state)("next",Object(u.style)({transform:"translate3d(100%, 0, 0)",visibility:"hidden"})),Object(u.transition)("* => *",Object(u.animate)("500ms cubic-bezier(0.35, 0, 0.25, 1)"))]),verticalStepTransition:Object(u.trigger)("stepTransition",[Object(u.state)("previous",Object(u.style)({height:"0px",visibility:"hidden"})),Object(u.state)("next",Object(u.style)({height:"0px",visibility:"hidden"})),Object(u.state)("current",Object(u.style)({height:"*",visibility:"visible"})),Object(u.transition)("* <=> current",Object(u.animate)("225ms cubic-bezier(0.4, 0.0, 0.2, 1)"))])},x=function(){function e(e){this.templateRef=e}return e.decorators=[{type:i.Directive,args:[{selector:"ng-template[matStepperIcon]"}]}],e.ctorParameters=function(){return[{type:i.TemplateRef}]},e.propDecorators={name:[{type:i.Input,args:["matStepperIcon"]}]},e}(),O=function(e){function t(t,n,r){var i=e.call(this,t,r)||this;return i._errorStateMatcher=n,i}return Object(r.__extends)(t,e),t.prototype.isErrorState=function(e,t){var n=this._errorStateMatcher.isErrorState(e,t),r=!!(e&&e.invalid&&this.interacted);return n||r},t.decorators=[{type:i.Component,args:[{selector:"mat-step",template:"<ng-template><ng-content></ng-content></ng-template>",providers:[{provide:d.ErrorStateMatcher,useExisting:t}],encapsulation:i.ViewEncapsulation.None,exportAs:"matStep",changeDetection:i.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:w,decorators:[{type:i.Inject,args:[Object(i.forwardRef)(function(){return w})]}]},{type:d.ErrorStateMatcher,decorators:[{type:i.SkipSelf}]},{type:void 0,decorators:[{type:i.Optional},{type:i.Inject,args:[o.STEPPER_GLOBAL_OPTIONS]}]}]},t.propDecorators={stepLabel:[{type:i.ContentChild,args:[g]}]},t}(o.CdkStep),w=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.animationDone=new i.EventEmitter,t._iconOverrides={},t._animationDone=new a.Subject,t}return Object(r.__extends)(t,e),t.prototype.ngAfterContentInit=function(){var e=this;this._icons.forEach(function(t){var n=t.name,r=t.templateRef;return e._iconOverrides[n]=r}),this._steps.changes.pipe(Object(p.takeUntil)(this._destroyed)).subscribe(function(){return e._stateChanged()}),this._animationDone.pipe(Object(p.distinctUntilChanged)(function(e,t){return e.fromState===t.fromState&&e.toState===t.toState}),Object(p.takeUntil)(this._destroyed)).subscribe(function(t){"current"===t.toState&&e.animationDone.emit()})},t.decorators=[{type:i.Directive,args:[{selector:"[matStepper]"}]}],t.propDecorators={_stepHeader:[{type:i.ViewChildren,args:[b]}],_steps:[{type:i.ContentChildren,args:[O]}],_icons:[{type:i.ContentChildren,args:[x]}],animationDone:[{type:i.Output}]},t}(o.CdkStepper),C=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.labelPosition="end",t}return Object(r.__extends)(t,e),t.decorators=[{type:i.Component,args:[{selector:"mat-horizontal-stepper",exportAs:"matHorizontalStepper",template:'<div class="mat-horizontal-stepper-header-container"><ng-container *ngFor="let step of _steps; let i = index; let isLast = last"><mat-step-header class="mat-horizontal-stepper-header" (click)="step.select()" (keydown)="_onKeydown($event)" [tabIndex]="_getFocusIndex() === i ? 0 : -1" [id]="_getStepLabelId(i)" [attr.aria-posinset]="i + 1" [attr.aria-setsize]="_steps.length" [attr.aria-controls]="_getStepContentId(i)" [attr.aria-selected]="selectedIndex == i" [attr.aria-label]="step.ariaLabel || null" [attr.aria-labelledby]="(!step.ariaLabel && step.ariaLabelledby) ? step.ariaLabelledby : null" [index]="i" [state]="_getIndicatorType(i, step.state)" [label]="step.stepLabel || step.label" [selected]="selectedIndex === i" [active]="step.completed || selectedIndex === i || !linear" [optional]="step.optional" [errorMessage]="step.errorMessage" [iconOverrides]="_iconOverrides"></mat-step-header><div *ngIf="!isLast" class="mat-stepper-horizontal-line"></div></ng-container></div><div class="mat-horizontal-content-container"><div *ngFor="let step of _steps; let i = index" class="mat-horizontal-stepper-content" role="tabpanel" [@stepTransition]="_getAnimationDirection(i)" (@stepTransition.done)="_animationDone.next($event)" [id]="_getStepContentId(i)" [attr.aria-labelledby]="_getStepLabelId(i)" [attr.aria-expanded]="selectedIndex === i"><ng-container [ngTemplateOutlet]="step.content"></ng-container></div></div>',styles:[".mat-stepper-horizontal,.mat-stepper-vertical{display:block}.mat-horizontal-stepper-header-container{white-space:nowrap;display:flex;align-items:center}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header-container{align-items:flex-start}.mat-stepper-horizontal-line{border-top-width:1px;border-top-style:solid;flex:auto;height:0;margin:0 -16px;min-width:32px}.mat-stepper-label-position-bottom .mat-stepper-horizontal-line{margin:0;min-width:0;position:relative;top:36px}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:first-child)::before,.mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:last-child)::after,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:first-child)::after,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:last-child)::before{border-top-width:1px;border-top-style:solid;content:'';display:inline-block;height:0;position:absolute;top:36px;width:calc(50% - 20px)}.mat-horizontal-stepper-header{display:flex;height:72px;overflow:hidden;align-items:center;padding:0 24px}.mat-horizontal-stepper-header .mat-step-icon{margin-right:8px;flex:none}[dir=rtl] .mat-horizontal-stepper-header .mat-step-icon{margin-right:0;margin-left:8px}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header{box-sizing:border-box;flex-direction:column;height:auto;padding:24px}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:last-child)::after,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:first-child)::after{right:0}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:first-child)::before,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:last-child)::before{left:0}[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:first-child::after,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:last-child::before{display:none}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header .mat-step-icon,.mat-stepper-label-position-bottom .mat-horizontal-stepper-header .mat-step-icon-not-touched{margin-right:0;margin-left:0}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header .mat-step-label{padding:16px 0 0 0;text-align:center;width:100%}.mat-vertical-stepper-header{display:flex;align-items:center;padding:24px;height:24px}.mat-vertical-stepper-header .mat-step-icon{margin-right:12px}[dir=rtl] .mat-vertical-stepper-header .mat-step-icon{margin-right:0;margin-left:12px}.mat-horizontal-stepper-content[aria-expanded=false]{height:0;overflow:hidden}.mat-horizontal-content-container{overflow:hidden;padding:0 24px 24px 24px}.mat-vertical-content-container{margin-left:36px;border:0;position:relative}[dir=rtl] .mat-vertical-content-container{margin-left:0;margin-right:36px}.mat-stepper-vertical-line::before{content:'';position:absolute;top:-16px;bottom:-16px;left:0;border-left-width:1px;border-left-style:solid}[dir=rtl] .mat-stepper-vertical-line::before{left:auto;right:0}.mat-vertical-stepper-content{overflow:hidden}.mat-vertical-content{padding:0 24px 24px 24px}.mat-step:last-child .mat-vertical-content-container{border:none}"],inputs:["selectedIndex"],host:{class:"mat-stepper-horizontal","[class.mat-stepper-label-position-end]":'labelPosition == "end"',"[class.mat-stepper-label-position-bottom]":'labelPosition == "bottom"',"aria-orientation":"horizontal",role:"tablist"},animations:[j.horizontalStepTransition],providers:[{provide:w,useExisting:t}],encapsulation:i.ViewEncapsulation.None,changeDetection:i.ChangeDetectionStrategy.OnPush}]}],t.propDecorators={labelPosition:[{type:i.Input}]},t}(w),E=function(e){function t(t,n,r,i){var o=e.call(this,t,n,r,i)||this;return o._orientation="vertical",o}return Object(r.__extends)(t,e),t.decorators=[{type:i.Component,args:[{selector:"mat-vertical-stepper",exportAs:"matVerticalStepper",template:'<div class="mat-step" *ngFor="let step of _steps; let i = index; let isLast = last"><mat-step-header class="mat-vertical-stepper-header" (click)="step.select()" (keydown)="_onKeydown($event)" [tabIndex]="_getFocusIndex() == i ? 0 : -1" [id]="_getStepLabelId(i)" [attr.aria-posinset]="i + 1" [attr.aria-setsize]="_steps.length" [attr.aria-controls]="_getStepContentId(i)" [attr.aria-selected]="selectedIndex === i" [attr.aria-label]="step.ariaLabel || null" [attr.aria-labelledby]="(!step.ariaLabel && step.ariaLabelledby) ? step.ariaLabelledby : null" [index]="i" [state]="_getIndicatorType(i, step.state)" [label]="step.stepLabel || step.label" [selected]="selectedIndex === i" [active]="step.completed || selectedIndex === i || !linear" [optional]="step.optional" [errorMessage]="step.errorMessage" [iconOverrides]="_iconOverrides"></mat-step-header><div class="mat-vertical-content-container" [class.mat-stepper-vertical-line]="!isLast"><div class="mat-vertical-stepper-content" role="tabpanel" [@stepTransition]="_getAnimationDirection(i)" (@stepTransition.done)="_animationDone.next($event)" [id]="_getStepContentId(i)" [attr.aria-labelledby]="_getStepLabelId(i)" [attr.aria-expanded]="selectedIndex === i"><div class="mat-vertical-content"><ng-container [ngTemplateOutlet]="step.content"></ng-container></div></div></div></div>',styles:[".mat-stepper-horizontal,.mat-stepper-vertical{display:block}.mat-horizontal-stepper-header-container{white-space:nowrap;display:flex;align-items:center}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header-container{align-items:flex-start}.mat-stepper-horizontal-line{border-top-width:1px;border-top-style:solid;flex:auto;height:0;margin:0 -16px;min-width:32px}.mat-stepper-label-position-bottom .mat-stepper-horizontal-line{margin:0;min-width:0;position:relative;top:36px}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:first-child)::before,.mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:last-child)::after,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:first-child)::after,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:last-child)::before{border-top-width:1px;border-top-style:solid;content:'';display:inline-block;height:0;position:absolute;top:36px;width:calc(50% - 20px)}.mat-horizontal-stepper-header{display:flex;height:72px;overflow:hidden;align-items:center;padding:0 24px}.mat-horizontal-stepper-header .mat-step-icon{margin-right:8px;flex:none}[dir=rtl] .mat-horizontal-stepper-header .mat-step-icon{margin-right:0;margin-left:8px}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header{box-sizing:border-box;flex-direction:column;height:auto;padding:24px}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:last-child)::after,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:first-child)::after{right:0}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:first-child)::before,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:not(:last-child)::before{left:0}[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:first-child::after,[dir=rtl] .mat-stepper-label-position-bottom .mat-horizontal-stepper-header:last-child::before{display:none}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header .mat-step-icon,.mat-stepper-label-position-bottom .mat-horizontal-stepper-header .mat-step-icon-not-touched{margin-right:0;margin-left:0}.mat-stepper-label-position-bottom .mat-horizontal-stepper-header .mat-step-label{padding:16px 0 0 0;text-align:center;width:100%}.mat-vertical-stepper-header{display:flex;align-items:center;padding:24px;height:24px}.mat-vertical-stepper-header .mat-step-icon{margin-right:12px}[dir=rtl] .mat-vertical-stepper-header .mat-step-icon{margin-right:0;margin-left:12px}.mat-horizontal-stepper-content[aria-expanded=false]{height:0;overflow:hidden}.mat-horizontal-content-container{overflow:hidden;padding:0 24px 24px 24px}.mat-vertical-content-container{margin-left:36px;border:0;position:relative}[dir=rtl] .mat-vertical-content-container{margin-left:0;margin-right:36px}.mat-stepper-vertical-line::before{content:'';position:absolute;top:-16px;bottom:-16px;left:0;border-left-width:1px;border-left-style:solid}[dir=rtl] .mat-stepper-vertical-line::before{left:auto;right:0}.mat-vertical-stepper-content{overflow:hidden}.mat-vertical-content{padding:0 24px 24px 24px}.mat-step:last-child .mat-vertical-content-container{border:none}"],inputs:["selectedIndex"],host:{class:"mat-stepper-vertical","aria-orientation":"vertical",role:"tablist"},animations:[j.verticalStepTransition],providers:[{provide:w,useExisting:t}],encapsulation:i.ViewEncapsulation.None,changeDetection:i.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:c.Directionality,decorators:[{type:i.Optional}]},{type:i.ChangeDetectorRef},{type:i.ElementRef},{type:void 0,decorators:[{type:i.Inject,args:[l.DOCUMENT]}]}]},t}(w),S=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"button[matStepperNext]",host:{"(click)":"_stepper.next()","[type]":"type"},inputs:["type"],providers:[{provide:o.CdkStepper,useExisting:w}]}]}],t}(o.CdkStepperNext),k=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.decorators=[{type:i.Directive,args:[{selector:"button[matStepperPrevious]",host:{"(click)":"_stepper.previous()","[type]":"type"},inputs:["type"],providers:[{provide:o.CdkStepper,useExisting:w}]}]}],t}(o.CdkStepperPrevious),M=function(){function e(){}return e.decorators=[{type:i.NgModule,args:[{imports:[d.MatCommonModule,l.CommonModule,f.PortalModule,h.MatButtonModule,o.CdkStepperModule,m.MatIconModule,d.MatRippleModule],exports:[d.MatCommonModule,C,E,O,g,w,S,k,b,x],declarations:[C,E,O,g,w,S,k,b,x],providers:[y,d.ErrorStateMatcher]}]}],e}()},"./node_modules/@angular/material/esm5/table.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatTableModule",function(){return E}),n.d(t,"MatCellDef",function(){return p}),n.d(t,"MatHeaderCellDef",function(){return f}),n.d(t,"MatFooterCellDef",function(){return h}),n.d(t,"MatColumnDef",function(){return m}),n.d(t,"MatHeaderCell",function(){return g}),n.d(t,"MatFooterCell",function(){return v}),n.d(t,"MatCell",function(){return _}),n.d(t,"MatTable",function(){return d}),n.d(t,"MatHeaderRowDef",function(){return y}),n.d(t,"MatFooterRowDef",function(){return b}),n.d(t,"MatRowDef",function(){return j}),n.d(t,"MatHeaderRow",function(){return x}),n.d(t,"MatFooterRow",function(){return O}),n.d(t,"MatRow",function(){return w}),n.d(t,"MatTableDataSource",function(){return k});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/table.es5.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/common/fesm5/common.js"),s=n("./node_modules/@angular/material/esm5/core.es5.js"),u=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),c=n("./node_modules/rxjs/_esm5/index.js"),l=n("./node_modules/rxjs/_esm5/operators/index.js"),d=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.stickyCssClass="mat-table-sticky",t}return Object(r.__extends)(t,e),t.decorators=[{type:o.Component,args:[{selector:"mat-table, table[mat-table]",exportAs:"matTable",template:i.CDK_TABLE_TEMPLATE,styles:["mat-table{display:block}mat-header-row{min-height:56px}mat-footer-row,mat-row{min-height:48px}mat-footer-row,mat-header-row,mat-row{display:flex;border-width:0;border-bottom-width:1px;border-style:solid;align-items:center;box-sizing:border-box}mat-footer-row::after,mat-header-row::after,mat-row::after{display:inline-block;min-height:inherit;content:''}mat-cell:first-of-type,mat-footer-cell:first-of-type,mat-header-cell:first-of-type{padding-left:24px}[dir=rtl] mat-cell:first-of-type,[dir=rtl] mat-footer-cell:first-of-type,[dir=rtl] mat-header-cell:first-of-type{padding-left:0;padding-right:24px}mat-cell:last-of-type,mat-footer-cell:last-of-type,mat-header-cell:last-of-type{padding-right:24px}[dir=rtl] mat-cell:last-of-type,[dir=rtl] mat-footer-cell:last-of-type,[dir=rtl] mat-header-cell:last-of-type{padding-right:0;padding-left:24px}mat-cell,mat-footer-cell,mat-header-cell{flex:1;display:flex;align-items:center;overflow:hidden;word-wrap:break-word;min-height:inherit}table.mat-table{border-spacing:0}tr.mat-header-row{height:56px}tr.mat-footer-row,tr.mat-row{height:48px}th.mat-header-cell{text-align:left}[dir=rtl] th.mat-header-cell{text-align:right}td.mat-cell,td.mat-footer-cell,th.mat-header-cell{padding:0;border-bottom-width:1px;border-bottom-style:solid}td.mat-cell:first-of-type,td.mat-footer-cell:first-of-type,th.mat-header-cell:first-of-type{padding-left:24px}[dir=rtl] td.mat-cell:first-of-type,[dir=rtl] td.mat-footer-cell:first-of-type,[dir=rtl] th.mat-header-cell:first-of-type{padding-left:0;padding-right:24px}td.mat-cell:last-of-type,td.mat-footer-cell:last-of-type,th.mat-header-cell:last-of-type{padding-right:24px}[dir=rtl] td.mat-cell:last-of-type,[dir=rtl] td.mat-footer-cell:last-of-type,[dir=rtl] th.mat-header-cell:last-of-type{padding-right:0;padding-left:24px}"],host:{class:"mat-table"},encapsulation:o.ViewEncapsulation.None,changeDetection:o.ChangeDetectionStrategy.OnPush}]}],t}(i.CdkTable),p=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.decorators=[{type:o.Directive,args:[{selector:"[matCellDef]",providers:[{provide:i.CdkCellDef,useExisting:t}]}]}],t}(i.CdkCellDef),f=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.decorators=[{type:o.Directive,args:[{selector:"[matHeaderCellDef]",providers:[{provide:i.CdkHeaderCellDef,useExisting:t}]}]}],t}(i.CdkHeaderCellDef),h=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.decorators=[{type:o.Directive,args:[{selector:"[matFooterCellDef]",providers:[{provide:i.CdkFooterCellDef,useExisting:t}]}]}],t}(i.CdkFooterCellDef),m=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.decorators=[{type:o.Directive,args:[{selector:"[matColumnDef]",providers:[{provide:i.CdkColumnDef,useExisting:t},{provide:"MAT_SORT_HEADER_COLUMN_DEF",useExisting:t}]}]}],t.propDecorators={name:[{type:o.Input,args:["matColumnDef"]}],sticky:[{type:o.Input}],stickyEnd:[{type:o.Input}]},t}(i.CdkColumnDef),g=function(e){function t(t,n){var r=e.call(this,t,n)||this;return n.nativeElement.classList.add("mat-column-"+t.cssClassFriendlyName),r}return Object(r.__extends)(t,e),t.decorators=[{type:o.Directive,args:[{selector:"mat-header-cell, th[mat-header-cell]",host:{class:"mat-header-cell",role:"columnheader"}}]}],t.ctorParameters=function(){return[{type:i.CdkColumnDef},{type:o.ElementRef}]},t}(i.CdkHeaderCell),v=function(e){function t(t,n){var r=e.call(this,t,n)||this;return n.nativeElement.classList.add("mat-column-"+t.cssClassFriendlyName),r}return Object(r.__extends)(t,e),t.decorators=[{type:o.Directive,args:[{selector:"mat-footer-cell, td[mat-footer-cell]",host:{class:"mat-footer-cell",role:"gridcell"}}]}],t.ctorParameters=function(){return[{type:i.CdkColumnDef},{type:o.ElementRef}]},t}(i.CdkFooterCell),_=function(e){function t(t,n){var r=e.call(this,t,n)||this;return n.nativeElement.classList.add("mat-column-"+t.cssClassFriendlyName),r}return Object(r.__extends)(t,e),t.decorators=[{type:o.Directive,args:[{selector:"mat-cell, td[mat-cell]",host:{class:"mat-cell",role:"gridcell"}}]}],t.ctorParameters=function(){return[{type:i.CdkColumnDef},{type:o.ElementRef}]},t}(i.CdkCell),y=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.decorators=[{type:o.Directive,args:[{selector:"[matHeaderRowDef]",providers:[{provide:i.CdkHeaderRowDef,useExisting:t}],inputs:["columns: matHeaderRowDef","sticky: matHeaderRowDefSticky"]}]}],t}(i.CdkHeaderRowDef),b=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.decorators=[{type:o.Directive,args:[{selector:"[matFooterRowDef]",providers:[{provide:i.CdkFooterRowDef,useExisting:t}],inputs:["columns: matFooterRowDef","sticky: matFooterRowDefSticky"]}]}],t}(i.CdkFooterRowDef),j=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.decorators=[{type:o.Directive,args:[{selector:"[matRowDef]",providers:[{provide:i.CdkRowDef,useExisting:t}],inputs:["columns: matRowDefColumns","when: matRowDefWhen"]}]}],t}(i.CdkRowDef),x=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.decorators=[{type:o.Component,args:[{selector:"mat-header-row, tr[mat-header-row]",template:i.CDK_ROW_TEMPLATE,host:{class:"mat-header-row",role:"row"},changeDetection:o.ChangeDetectionStrategy.OnPush,encapsulation:o.ViewEncapsulation.None,exportAs:"matHeaderRow",providers:[{provide:i.CdkHeaderRow,useExisting:t}]}]}],t}(i.CdkHeaderRow),O=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.decorators=[{type:o.Component,args:[{selector:"mat-footer-row, tr[mat-footer-row]",template:i.CDK_ROW_TEMPLATE,host:{class:"mat-footer-row",role:"row"},changeDetection:o.ChangeDetectionStrategy.OnPush,encapsulation:o.ViewEncapsulation.None,exportAs:"matFooterRow",providers:[{provide:i.CdkFooterRow,useExisting:t}]}]}],t}(i.CdkFooterRow),w=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.decorators=[{type:o.Component,args:[{selector:"mat-row, tr[mat-row]",template:i.CDK_ROW_TEMPLATE,host:{class:"mat-row",role:"row"},changeDetection:o.ChangeDetectionStrategy.OnPush,encapsulation:o.ViewEncapsulation.None,exportAs:"matRow",providers:[{provide:i.CdkRow,useExisting:t}]}]}],t}(i.CdkRow),C=[d,f,y,m,p,j,h,b,g,_,v,x,w,O],E=function(){function e(){}return e.decorators=[{type:o.NgModule,args:[{imports:[i.CdkTableModule,a.CommonModule,s.MatCommonModule],exports:C,declarations:C}]}],e}(),S=9007199254740991,k=function(e){function t(t){void 0===t&&(t=[]);var n=e.call(this)||this;return n._renderData=new c.BehaviorSubject([]),n._filter=new c.BehaviorSubject(""),n._renderChangesSubscription=c.Subscription.EMPTY,n.sortingDataAccessor=function(e,t){var n=e[t];if(Object(u._isNumberValue)(n)){var r=Number(n);return r<S?r:n}return n},n.sortData=function(e,t){var r=t.active,i=t.direction;return r&&""!=i?e.sort(function(e,t){var o=n.sortingDataAccessor(e,r),a=n.sortingDataAccessor(t,r),s=0;return null!=o&&null!=a?o>a?s=1:o<a&&(s=-1):null!=o?s=1:null!=a&&(s=-1),s*("asc"==i?1:-1)}):e},n.filterPredicate=function(e,t){var n=Object.keys(e).reduce(function(t,n){return t+e[n]+"◬"},"").toLowerCase(),r=t.trim().toLowerCase();return-1!=n.indexOf(r)},n._data=new c.BehaviorSubject(t),n._updateChangeSubscription(),n}return Object(r.__extends)(t,e),Object.defineProperty(t.prototype,"data",{get:function(){return this._data.value},set:function(e){this._data.next(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"filter",{get:function(){return this._filter.value},set:function(e){this._filter.next(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"sort",{get:function(){return this._sort},set:function(e){this._sort=e,this._updateChangeSubscription()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"paginator",{get:function(){return this._paginator},set:function(e){this._paginator=e,this._updateChangeSubscription()},enumerable:!0,configurable:!0}),t.prototype._updateChangeSubscription=function(){var e=this,t=this._sort?Object(c.merge)(this._sort.sortChange,this._sort.initialized):Object(c.of)(null),n=this._paginator?Object(c.merge)(this._paginator.page,this._paginator.initialized):Object(c.of)(null),r=this._data,i=Object(c.combineLatest)(r,this._filter).pipe(Object(l.map)(function(t){var n=t[0];return e._filterData(n)})),o=Object(c.combineLatest)(i,t).pipe(Object(l.map)(function(t){var n=t[0];return e._orderData(n)})),a=Object(c.combineLatest)(o,n).pipe(Object(l.map)(function(t){var n=t[0];return e._pageData(n)}));this._renderChangesSubscription.unsubscribe(),this._renderChangesSubscription=a.subscribe(function(t){return e._renderData.next(t)})},t.prototype._filterData=function(e){var t=this;return this.filteredData=this.filter?e.filter(function(e){return t.filterPredicate(e,t.filter)}):e,this.paginator&&this._updatePaginator(this.filteredData.length),this.filteredData},t.prototype._orderData=function(e){return this.sort?this.sortData(e.slice(),this.sort):e},t.prototype._pageData=function(e){if(!this.paginator)return e;var t=this.paginator.pageIndex*this.paginator.pageSize;return e.slice().splice(t,this.paginator.pageSize)},t.prototype._updatePaginator=function(e){var t=this;Promise.resolve().then(function(){if(t.paginator&&(t.paginator.length=e,t.paginator.pageIndex>0)){var n=Math.ceil(t.paginator.length/t.paginator.pageSize)-1||0;t.paginator.pageIndex=Math.min(t.paginator.pageIndex,n)}})},t.prototype.connect=function(){return this._renderData},t.prototype.disconnect=function(){},t}(i.DataSource)},"./node_modules/@angular/material/esm5/tabs.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatInkBar",function(){return b}),n.d(t,"_MAT_INK_BAR_POSITIONER",function(){return _}),n.d(t,"MatTabBody",function(){return k}),n.d(t,"MatTabBodyPortal",function(){return S}),n.d(t,"MatTabHeader",function(){return P}),n.d(t,"MatTabLabelWrapper",function(){return T}),n.d(t,"MatTab",function(){return C}),n.d(t,"MatTabLabel",function(){return x}),n.d(t,"MatTabNav",function(){return H}),n.d(t,"MatTabLink",function(){return W}),n.d(t,"MatTabContent",function(){return j}),n.d(t,"MatTabsModule",function(){return Y}),n.d(t,"MatTabChangeEvent",function(){return N}),n.d(t,"MAT_TABS_CONFIG",function(){return F}),n.d(t,"MatTabGroupBase",function(){return L}),n.d(t,"_MatTabGroupMixinBase",function(){return V}),n.d(t,"MatTabGroup",function(){return B}),n.d(t,"matTabsAnimations",function(){return E}),n.d(t,"ɵa17",function(){return y}),n.d(t,"ɵf17",function(){return O}),n.d(t,"ɵg17",function(){return w}),n.d(t,"ɵb17",function(){return D}),n.d(t,"ɵc17",function(){return I}),n.d(t,"ɵd17",function(){return M}),n.d(t,"ɵe17",function(){return A}),n.d(t,"ɵj17",function(){return q}),n.d(t,"ɵh17",function(){return z}),n.d(t,"ɵk17",function(){return G}),n.d(t,"ɵi17",function(){return U});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),a=n("./node_modules/@angular/material/esm5/core.es5.js"),s=n("./node_modules/rxjs/_esm5/index.js"),u=n("./node_modules/@angular/animations/fesm5/animations.js"),c=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),l=n("./node_modules/rxjs/_esm5/operators/index.js"),d=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),p=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),f=n("./node_modules/@angular/cdk/esm5/scrolling.es5.js"),h=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),m=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),g=n("./node_modules/@angular/cdk/esm5/observers.es5.js"),v=n("./node_modules/@angular/common/fesm5/common.js"),_=new r.InjectionToken("MatInkBarPositioner",{providedIn:"root",factory:y});function y(){return function(e){return{left:e?(e.offsetLeft||0)+"px":"0",width:e?(e.offsetWidth||0)+"px":"0"}}}var b=function(){function e(e,t,n){this._elementRef=e,this._ngZone=t,this._inkBarPositioner=n}return e.prototype.alignToElement=function(e){var t=this;this.show(),"undefined"!=typeof requestAnimationFrame?this._ngZone.runOutsideAngular(function(){requestAnimationFrame(function(){return t._setStyles(e)})}):this._setStyles(e)},e.prototype.show=function(){this._elementRef.nativeElement.style.visibility="visible"},e.prototype.hide=function(){this._elementRef.nativeElement.style.visibility="hidden"},e.prototype._setStyles=function(e){var t=this._inkBarPositioner(e),n=this._elementRef.nativeElement;n.style.left=t.left,n.style.width=t.width},e.decorators=[{type:r.Directive,args:[{selector:"mat-ink-bar",host:{class:"mat-ink-bar"}}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:r.NgZone},{type:void 0,decorators:[{type:r.Inject,args:[_]}]}]},e}(),j=function(){function e(e){this.template=e}return e.decorators=[{type:r.Directive,args:[{selector:"[matTabContent]"}]}],e.ctorParameters=function(){return[{type:r.TemplateRef}]},e}(),x=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(i.__extends)(t,e),t.decorators=[{type:r.Directive,args:[{selector:"[mat-tab-label], [matTabLabel]"}]}],t}(o.CdkPortal),O=function(){return function(){}}(),w=Object(a.mixinDisabled)(O),C=function(e){function t(t){var n=e.call(this)||this;return n._viewContainerRef=t,n.textLabel="",n._contentPortal=null,n._stateChanges=new s.Subject,n.position=null,n.origin=null,n.isActive=!1,n}return Object(i.__extends)(t,e),Object.defineProperty(t.prototype,"content",{get:function(){return this._contentPortal},enumerable:!0,configurable:!0}),t.prototype.ngOnChanges=function(e){(e.hasOwnProperty("textLabel")||e.hasOwnProperty("disabled"))&&this._stateChanges.next()},t.prototype.ngOnDestroy=function(){this._stateChanges.complete()},t.prototype.ngOnInit=function(){this._contentPortal=new o.TemplatePortal(this._explicitContent||this._implicitContent,this._viewContainerRef)},t.decorators=[{type:r.Component,args:[{selector:"mat-tab",template:"<ng-template><ng-content></ng-content></ng-template>",inputs:["disabled"],changeDetection:r.ChangeDetectionStrategy.OnPush,encapsulation:r.ViewEncapsulation.None,exportAs:"matTab"}]}],t.ctorParameters=function(){return[{type:r.ViewContainerRef}]},t.propDecorators={templateLabel:[{type:r.ContentChild,args:[x]}],_explicitContent:[{type:r.ContentChild,args:[j,{read:r.TemplateRef}]}],_implicitContent:[{type:r.ViewChild,args:[r.TemplateRef]}],textLabel:[{type:r.Input,args:["label"]}],ariaLabel:[{type:r.Input,args:["aria-label"]}],ariaLabelledby:[{type:r.Input,args:["aria-labelledby"]}]},t}(w),E={translateTab:Object(u.trigger)("translateTab",[Object(u.state)("center, void, left-origin-center, right-origin-center",Object(u.style)({transform:"none"})),Object(u.state)("left",Object(u.style)({transform:"translate3d(-100%, 0, 0)",minHeight:"1px"})),Object(u.state)("right",Object(u.style)({transform:"translate3d(100%, 0, 0)",minHeight:"1px"})),Object(u.transition)("* => left, * => right, left => center, right => center",Object(u.animate)("{{animationDuration}} cubic-bezier(0.35, 0, 0.25, 1)")),Object(u.transition)("void => left-origin-center",[Object(u.style)({transform:"translate3d(-100%, 0, 0)"}),Object(u.animate)("{{animationDuration}} cubic-bezier(0.35, 0, 0.25, 1)")]),Object(u.transition)("void => right-origin-center",[Object(u.style)({transform:"translate3d(100%, 0, 0)"}),Object(u.animate)("{{animationDuration}} cubic-bezier(0.35, 0, 0.25, 1)")])])},S=function(e){function t(t,n,r){var i=e.call(this,t,n)||this;return i._host=r,i._centeringSub=s.Subscription.EMPTY,i._leavingSub=s.Subscription.EMPTY,i}return Object(i.__extends)(t,e),t.prototype.ngOnInit=function(){var t=this;e.prototype.ngOnInit.call(this),this._centeringSub=this._host._beforeCentering.pipe(Object(l.startWith)(this._host._isCenterPosition(this._host._position))).subscribe(function(e){e&&!t.hasAttached()&&t.attach(t._host._content)}),this._leavingSub=this._host._afterLeavingCenter.subscribe(function(){t.detach()})},t.prototype.ngOnDestroy=function(){e.prototype.ngOnDestroy.call(this),this._centeringSub.unsubscribe(),this._leavingSub.unsubscribe()},t.decorators=[{type:r.Directive,args:[{selector:"[matTabBodyHost]"}]}],t.ctorParameters=function(){return[{type:r.ComponentFactoryResolver},{type:r.ViewContainerRef},{type:k,decorators:[{type:r.Inject,args:[Object(r.forwardRef)(function(){return k})]}]}]},t}(o.CdkPortalOutlet),k=function(){function e(e,t,n){var i=this;this._elementRef=e,this._dir=t,this._dirChangeSubscription=s.Subscription.EMPTY,this._translateTabComplete=new s.Subject,this._onCentering=new r.EventEmitter,this._beforeCentering=new r.EventEmitter,this._afterLeavingCenter=new r.EventEmitter,this._onCentered=new r.EventEmitter(!0),this.animationDuration="500ms",this._dir&&n&&(this._dirChangeSubscription=this._dir.change.subscribe(function(e){i._computePositionAnimationState(e),n.markForCheck()})),this._translateTabComplete.pipe(Object(l.distinctUntilChanged)(function(e,t){return e.fromState===t.fromState&&e.toState===t.toState})).subscribe(function(e){i._isCenterPosition(e.toState)&&i._isCenterPosition(i._position)&&i._onCentered.emit(),i._isCenterPosition(e.fromState)&&!i._isCenterPosition(i._position)&&i._afterLeavingCenter.emit()})}return Object.defineProperty(e.prototype,"position",{set:function(e){this._positionIndex=e,this._computePositionAnimationState()},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){"center"==this._position&&null!=this.origin&&(this._position=this._computePositionFromOrigin())},e.prototype.ngOnDestroy=function(){this._dirChangeSubscription.unsubscribe(),this._translateTabComplete.complete()},e.prototype._onTranslateTabStarted=function(e){var t=this._isCenterPosition(e.toState);this._beforeCentering.emit(t),t&&this._onCentering.emit(this._elementRef.nativeElement.clientHeight)},e.prototype._getLayoutDirection=function(){return this._dir&&"rtl"===this._dir.value?"rtl":"ltr"},e.prototype._isCenterPosition=function(e){return"center"==e||"left-origin-center"==e||"right-origin-center"==e},e.prototype._computePositionAnimationState=function(e){void 0===e&&(e=this._getLayoutDirection()),this._positionIndex<0?this._position="ltr"==e?"left":"right":this._positionIndex>0?this._position="ltr"==e?"right":"left":this._position="center"},e.prototype._computePositionFromOrigin=function(){var e=this._getLayoutDirection();return"ltr"==e&&this.origin<=0||"rtl"==e&&this.origin>0?"left-origin-center":"right-origin-center"},e.decorators=[{type:r.Component,args:[{selector:"mat-tab-body",template:'<div class="mat-tab-body-content" #content [@translateTab]="{ value: _position, params: {animationDuration: animationDuration} }" (@translateTab.start)="_onTranslateTabStarted($event)" (@translateTab.done)="_translateTabComplete.next($event)"><ng-template matTabBodyHost></ng-template></div>',styles:[".mat-tab-body-content{height:100%;overflow:auto}.mat-tab-group-dynamic-height .mat-tab-body-content{overflow:hidden}"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,animations:[E.translateTab],host:{class:"mat-tab-body"}}]}],e.ctorParameters=function(){return[{type:r.ElementRef},{type:c.Directionality,decorators:[{type:r.Optional}]},{type:r.ChangeDetectorRef}]},e.propDecorators={_onCentering:[{type:r.Output}],_beforeCentering:[{type:r.Output}],_afterLeavingCenter:[{type:r.Output}],_onCentered:[{type:r.Output}],_portalHost:[{type:r.ViewChild,args:[o.PortalHostDirective]}],_content:[{type:r.Input,args:["content"]}],origin:[{type:r.Input}],animationDuration:[{type:r.Input}],position:[{type:r.Input}]},e}(),M=function(){return function(){}}(),A=Object(a.mixinDisabled)(M),T=function(e){function t(t){var n=e.call(this)||this;return n.elementRef=t,n}return Object(i.__extends)(t,e),t.prototype.focus=function(){this.elementRef.nativeElement.focus()},t.prototype.getOffsetLeft=function(){return this.elementRef.nativeElement.offsetLeft},t.prototype.getOffsetWidth=function(){return this.elementRef.nativeElement.offsetWidth},t.decorators=[{type:r.Directive,args:[{selector:"[matTabLabelWrapper]",inputs:["disabled"],host:{"[class.mat-tab-disabled]":"disabled","[attr.aria-disabled]":"!!disabled"}}]}],t.ctorParameters=function(){return[{type:r.ElementRef}]},t}(A),D=function(){return function(){}}(),I=Object(a.mixinDisableRipple)(D),P=function(e){function t(t,n,i,o,a){var u=e.call(this)||this;return u._elementRef=t,u._changeDetectorRef=n,u._viewportRuler=i,u._dir=o,u._ngZone=a,u._scrollDistance=0,u._selectedIndexChanged=!1,u._destroyed=new s.Subject,u._showPaginationControls=!1,u._disableScrollAfter=!0,u._disableScrollBefore=!0,u._selectedIndex=0,u.selectFocusedIndex=new r.EventEmitter,u.indexFocused=new r.EventEmitter,u}return Object(i.__extends)(t,e),Object.defineProperty(t.prototype,"selectedIndex",{get:function(){return this._selectedIndex},set:function(e){e=Object(d.coerceNumberProperty)(e),this._selectedIndexChanged=this._selectedIndex!=e,this._selectedIndex=e,this._keyManager&&this._keyManager.updateActiveItemIndex(e)},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentChecked=function(){this._tabLabelCount!=this._labelWrappers.length&&(this.updatePagination(),this._tabLabelCount=this._labelWrappers.length,this._changeDetectorRef.markForCheck()),this._selectedIndexChanged&&(this._scrollToLabel(this._selectedIndex),this._checkScrollingControls(),this._alignInkBarToSelectedTab(),this._selectedIndexChanged=!1,this._changeDetectorRef.markForCheck()),this._scrollDistanceChanged&&(this._updateTabScrollPosition(),this._scrollDistanceChanged=!1,this._changeDetectorRef.markForCheck())},t.prototype._handleKeydown=function(e){switch(e.keyCode){case p.HOME:this._keyManager.setFirstItemActive(),e.preventDefault();break;case p.END:this._keyManager.setLastItemActive(),e.preventDefault();break;case p.ENTER:case p.SPACE:this.selectFocusedIndex.emit(this.focusIndex),e.preventDefault();break;default:this._keyManager.onKeydown(e)}},t.prototype.ngAfterContentInit=function(){var e=this,t=this._dir?this._dir.change:Object(s.of)(null),n=this._viewportRuler.change(150),r=function(){e.updatePagination(),e._alignInkBarToSelectedTab()};this._keyManager=new h.FocusKeyManager(this._labelWrappers).withHorizontalOrientation(this._getLayoutDirection()).withWrap(),this._keyManager.updateActiveItem(0),"undefined"!=typeof requestAnimationFrame?requestAnimationFrame(r):r(),Object(s.merge)(t,n).pipe(Object(l.takeUntil)(this._destroyed)).subscribe(function(){r(),e._keyManager.withHorizontalOrientation(e._getLayoutDirection())}),this._keyManager.change.pipe(Object(l.takeUntil)(this._destroyed)).subscribe(function(t){e.indexFocused.emit(t),e._setTabFocus(t)})},t.prototype.ngOnDestroy=function(){this._destroyed.next(),this._destroyed.complete()},t.prototype._onContentChanges=function(){var e=this,t=this._elementRef.nativeElement.textContent;if(t!==this._currentTextContent){this._currentTextContent=t;var n=function(){e.updatePagination(),e._alignInkBarToSelectedTab(),e._changeDetectorRef.markForCheck()};this._ngZone?this._ngZone.run(n):n()}},t.prototype.updatePagination=function(){this._checkPaginationEnabled(),this._checkScrollingControls(),this._updateTabScrollPosition()},Object.defineProperty(t.prototype,"focusIndex",{get:function(){return this._keyManager?this._keyManager.activeItemIndex:0},set:function(e){this._isValidIndex(e)&&this.focusIndex!==e&&this._keyManager&&this._keyManager.setActiveItem(e)},enumerable:!0,configurable:!0}),t.prototype._isValidIndex=function(e){if(!this._labelWrappers)return!0;var t=this._labelWrappers?this._labelWrappers.toArray()[e]:null;return!!t&&!t.disabled},t.prototype._setTabFocus=function(e){if(this._showPaginationControls&&this._scrollToLabel(e),this._labelWrappers&&this._labelWrappers.length){this._labelWrappers.toArray()[e].focus();var t=this._tabListContainer.nativeElement,n=this._getLayoutDirection();t.scrollLeft="ltr"==n?0:t.scrollWidth-t.offsetWidth}},t.prototype._getLayoutDirection=function(){return this._dir&&"rtl"===this._dir.value?"rtl":"ltr"},t.prototype._updateTabScrollPosition=function(){var e=this.scrollDistance,t="ltr"===this._getLayoutDirection()?-e:e;this._tabList.nativeElement.style.transform="translateX("+Math.round(t)+"px)",this._tabListContainer.nativeElement.scrollLeft=0},Object.defineProperty(t.prototype,"scrollDistance",{get:function(){return this._scrollDistance},set:function(e){this._scrollDistance=Math.max(0,Math.min(this._getMaxScrollDistance(),e)),this._scrollDistanceChanged=!0,this._checkScrollingControls()},enumerable:!0,configurable:!0}),t.prototype._scrollHeader=function(e){var t=this._tabListContainer.nativeElement.offsetWidth;this.scrollDistance+=("before"==e?-1:1)*t/3},t.prototype._scrollToLabel=function(e){var t=this._labelWrappers?this._labelWrappers.toArray()[e]:null;if(t){var n,r,i=this._tabListContainer.nativeElement.offsetWidth;"ltr"==this._getLayoutDirection()?r=(n=t.getOffsetLeft())+t.getOffsetWidth():n=(r=this._tabList.nativeElement.offsetWidth-t.getOffsetLeft())-t.getOffsetWidth();var o=this.scrollDistance,a=this.scrollDistance+i;n<o?this.scrollDistance-=o-n+60:r>a&&(this.scrollDistance+=r-a+60)}},t.prototype._checkPaginationEnabled=function(){var e=this._tabList.nativeElement.scrollWidth>this._elementRef.nativeElement.offsetWidth;e||(this.scrollDistance=0),e!==this._showPaginationControls&&this._changeDetectorRef.markForCheck(),this._showPaginationControls=e},t.prototype._checkScrollingControls=function(){this._disableScrollBefore=0==this.scrollDistance,this._disableScrollAfter=this.scrollDistance==this._getMaxScrollDistance(),this._changeDetectorRef.markForCheck()},t.prototype._getMaxScrollDistance=function(){return this._tabList.nativeElement.scrollWidth-this._tabListContainer.nativeElement.offsetWidth||0},t.prototype._alignInkBarToSelectedTab=function(){var e=this._labelWrappers&&this._labelWrappers.length?this._labelWrappers.toArray()[this.selectedIndex].elementRef.nativeElement:null;this._inkBar.alignToElement(e)},t.decorators=[{type:r.Component,args:[{selector:"mat-tab-header",template:'<div class="mat-tab-header-pagination mat-tab-header-pagination-before mat-elevation-z4" aria-hidden="true" mat-ripple [matRippleDisabled]="_disableScrollBefore || disableRipple" [class.mat-tab-header-pagination-disabled]="_disableScrollBefore" (click)="_scrollHeader(\'before\')"><div class="mat-tab-header-pagination-chevron"></div></div><div class="mat-tab-label-container" #tabListContainer (keydown)="_handleKeydown($event)"><div class="mat-tab-list" #tabList role="tablist" (cdkObserveContent)="_onContentChanges()"><div class="mat-tab-labels"><ng-content></ng-content></div><mat-ink-bar></mat-ink-bar></div></div><div class="mat-tab-header-pagination mat-tab-header-pagination-after mat-elevation-z4" aria-hidden="true" mat-ripple [matRippleDisabled]="_disableScrollAfter || disableRipple" [class.mat-tab-header-pagination-disabled]="_disableScrollAfter" (click)="_scrollHeader(\'after\')"><div class="mat-tab-header-pagination-chevron"></div></div>',styles:[".mat-tab-header{display:flex;overflow:hidden;position:relative;flex-shrink:0}.mat-tab-label{height:48px;padding:0 24px;cursor:pointer;box-sizing:border-box;opacity:.6;min-width:160px;text-align:center;display:inline-flex;justify-content:center;align-items:center;white-space:nowrap;position:relative}.mat-tab-label:focus{outline:0}.mat-tab-label:focus:not(.mat-tab-disabled){opacity:1}@media (-ms-high-contrast:active){.mat-tab-label:focus{outline:dotted 2px}}.mat-tab-label.mat-tab-disabled{cursor:default}@media (-ms-high-contrast:active){.mat-tab-label.mat-tab-disabled{opacity:.5}}.mat-tab-label .mat-tab-label-content{display:inline-flex;justify-content:center;align-items:center;white-space:nowrap}@media (-ms-high-contrast:active){.mat-tab-label{opacity:1}}@media (max-width:599px){.mat-tab-label{min-width:72px}}.mat-ink-bar{position:absolute;bottom:0;height:2px;transition:.5s cubic-bezier(.35,0,.25,1)}.mat-tab-group-inverted-header .mat-ink-bar{bottom:auto;top:0}@media (-ms-high-contrast:active){.mat-ink-bar{outline:solid 2px;height:0}}.mat-tab-header-pagination{position:relative;display:none;justify-content:center;align-items:center;min-width:32px;cursor:pointer;z-index:2}.mat-tab-header-pagination-controls-enabled .mat-tab-header-pagination{display:flex}.mat-tab-header-pagination-before,.mat-tab-header-rtl .mat-tab-header-pagination-after{padding-left:4px}.mat-tab-header-pagination-before .mat-tab-header-pagination-chevron,.mat-tab-header-rtl .mat-tab-header-pagination-after .mat-tab-header-pagination-chevron{transform:rotate(-135deg)}.mat-tab-header-pagination-after,.mat-tab-header-rtl .mat-tab-header-pagination-before{padding-right:4px}.mat-tab-header-pagination-after .mat-tab-header-pagination-chevron,.mat-tab-header-rtl .mat-tab-header-pagination-before .mat-tab-header-pagination-chevron{transform:rotate(45deg)}.mat-tab-header-pagination-chevron{border-style:solid;border-width:2px 2px 0 0;content:'';height:8px;width:8px}.mat-tab-header-pagination-disabled{box-shadow:none;cursor:default}.mat-tab-label-container{display:flex;flex-grow:1;overflow:hidden;z-index:1}.mat-tab-list{flex-grow:1;position:relative;transition:transform .5s cubic-bezier(.35,0,.25,1)}.mat-tab-labels{display:flex}[mat-align-tabs=center] .mat-tab-labels{justify-content:center}[mat-align-tabs=end] .mat-tab-labels{justify-content:flex-end}"],inputs:["disableRipple"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,host:{class:"mat-tab-header","[class.mat-tab-header-pagination-controls-enabled]":"_showPaginationControls","[class.mat-tab-header-rtl]":"_getLayoutDirection() == 'rtl'"}}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:r.ChangeDetectorRef},{type:f.ViewportRuler},{type:c.Directionality,decorators:[{type:r.Optional}]},{type:r.NgZone}]},t.propDecorators={_labelWrappers:[{type:r.ContentChildren,args:[T]}],_inkBar:[{type:r.ViewChild,args:[b]}],_tabListContainer:[{type:r.ViewChild,args:["tabListContainer"]}],_tabList:[{type:r.ViewChild,args:["tabList"]}],selectedIndex:[{type:r.Input}],selectFocusedIndex:[{type:r.Output}],indexFocused:[{type:r.Output}]},t}(I),R=0,N=function(){return function(){}}(),F=new r.InjectionToken("MAT_TABS_CONFIG"),L=function(){return function(e){this._elementRef=e}}(),V=Object(a.mixinColor)(Object(a.mixinDisableRipple)(L),"primary"),B=function(e){function t(t,n,i){var o=e.call(this,t)||this;return o._changeDetectorRef=n,o._indexToSelect=0,o._tabBodyWrapperHeight=0,o._tabsSubscription=s.Subscription.EMPTY,o._tabLabelSubscription=s.Subscription.EMPTY,o._dynamicHeight=!1,o._selectedIndex=null,o.headerPosition="above",o.selectedIndexChange=new r.EventEmitter,o.focusChange=new r.EventEmitter,o.animationDone=new r.EventEmitter,o.selectedTabChange=new r.EventEmitter(!0),o._groupId=R++,o.animationDuration=i&&i.animationDuration?i.animationDuration:"500ms",o}return Object(i.__extends)(t,e),Object.defineProperty(t.prototype,"dynamicHeight",{get:function(){return this._dynamicHeight},set:function(e){this._dynamicHeight=Object(d.coerceBooleanProperty)(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selectedIndex",{get:function(){return this._selectedIndex},set:function(e){this._indexToSelect=Object(d.coerceNumberProperty)(e,null)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"backgroundColor",{get:function(){return this._backgroundColor},set:function(e){var t=this._elementRef.nativeElement;t.classList.remove("mat-background-"+this.backgroundColor),e&&t.classList.add("mat-background-"+e),this._backgroundColor=e},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentChecked=function(){var e=this,t=this._indexToSelect=this._clampTabIndex(this._indexToSelect);if(this._selectedIndex!=t){var n=null==this._selectedIndex;n||this.selectedTabChange.emit(this._createChangeEvent(t)),Promise.resolve().then(function(){e._tabs.forEach(function(e,n){return e.isActive=n===t}),n||e.selectedIndexChange.emit(t)})}this._tabs.forEach(function(n,r){n.position=r-t,null==e._selectedIndex||0!=n.position||n.origin||(n.origin=t-e._selectedIndex)}),this._selectedIndex!==t&&(this._selectedIndex=t,this._changeDetectorRef.markForCheck())},t.prototype.ngAfterContentInit=function(){var e=this;this._subscribeToTabLabels(),this._tabsSubscription=this._tabs.changes.subscribe(function(){if(e._clampTabIndex(e._indexToSelect)===e._selectedIndex)for(var t=e._tabs.toArray(),n=0;n<t.length;n++)if(t[n].isActive){e._indexToSelect=e._selectedIndex=n;break}e._subscribeToTabLabels(),e._changeDetectorRef.markForCheck()})},t.prototype.ngOnDestroy=function(){this._tabsSubscription.unsubscribe(),this._tabLabelSubscription.unsubscribe()},t.prototype.realignInkBar=function(){this._tabHeader&&this._tabHeader._alignInkBarToSelectedTab()},t.prototype._focusChanged=function(e){this.focusChange.emit(this._createChangeEvent(e))},t.prototype._createChangeEvent=function(e){var t=new N;return t.index=e,this._tabs&&this._tabs.length&&(t.tab=this._tabs.toArray()[e]),t},t.prototype._subscribeToTabLabels=function(){var e=this;this._tabLabelSubscription&&this._tabLabelSubscription.unsubscribe(),this._tabLabelSubscription=s.merge.apply(void 0,this._tabs.map(function(e){return e._stateChanges})).subscribe(function(){return e._changeDetectorRef.markForCheck()})},t.prototype._clampTabIndex=function(e){return Math.min(this._tabs.length-1,Math.max(e||0,0))},t.prototype._getTabLabelId=function(e){return"mat-tab-label-"+this._groupId+"-"+e},t.prototype._getTabContentId=function(e){return"mat-tab-content-"+this._groupId+"-"+e},t.prototype._setTabBodyWrapperHeight=function(e){if(this._dynamicHeight&&this._tabBodyWrapperHeight){var t=this._tabBodyWrapper.nativeElement;t.style.height=this._tabBodyWrapperHeight+"px",this._tabBodyWrapper.nativeElement.offsetHeight&&(t.style.height=e+"px")}},t.prototype._removeTabBodyWrapperHeight=function(){var e=this._tabBodyWrapper.nativeElement;this._tabBodyWrapperHeight=e.clientHeight,e.style.height="",this.animationDone.emit()},t.prototype._handleClick=function(e,t,n){e.disabled||(this.selectedIndex=t.focusIndex=n)},t.prototype._getTabIndex=function(e,t){return e.disabled?null:this.selectedIndex===t?0:-1},t.decorators=[{type:r.Component,args:[{selector:"mat-tab-group",exportAs:"matTabGroup",template:'<mat-tab-header #tabHeader [selectedIndex]="selectedIndex" [disableRipple]="disableRipple" (indexFocused)="_focusChanged($event)" (selectFocusedIndex)="selectedIndex = $event"><div class="mat-tab-label" role="tab" matTabLabelWrapper mat-ripple cdkMonitorElementFocus *ngFor="let tab of _tabs; let i = index" [id]="_getTabLabelId(i)" [attr.tabIndex]="_getTabIndex(tab, i)" [attr.aria-posinset]="i + 1" [attr.aria-setsize]="_tabs.length" [attr.aria-controls]="_getTabContentId(i)" [attr.aria-selected]="selectedIndex == i" [attr.aria-label]="tab.ariaLabel || null" [attr.aria-labelledby]="(!tab.ariaLabel && tab.ariaLabelledby) ? tab.ariaLabelledby : null" [class.mat-tab-label-active]="selectedIndex == i" [disabled]="tab.disabled" [matRippleDisabled]="tab.disabled || disableRipple" (click)="_handleClick(tab, tabHeader, i)"><div class="mat-tab-label-content"><ng-template [ngIf]="tab.templateLabel"><ng-template [cdkPortalOutlet]="tab.templateLabel"></ng-template></ng-template><ng-template [ngIf]="!tab.templateLabel">{{tab.textLabel}}</ng-template></div></div></mat-tab-header><div class="mat-tab-body-wrapper" #tabBodyWrapper><mat-tab-body role="tabpanel" *ngFor="let tab of _tabs; let i = index" [id]="_getTabContentId(i)" [attr.aria-labelledby]="_getTabLabelId(i)" [class.mat-tab-body-active]="selectedIndex == i" [content]="tab.content" [position]="tab.position" [origin]="tab.origin" [animationDuration]="animationDuration" (_onCentered)="_removeTabBodyWrapperHeight()" (_onCentering)="_setTabBodyWrapperHeight($event)"></mat-tab-body></div>',styles:[".mat-tab-group{display:flex;flex-direction:column}.mat-tab-group.mat-tab-group-inverted-header{flex-direction:column-reverse}.mat-tab-label{height:48px;padding:0 24px;cursor:pointer;box-sizing:border-box;opacity:.6;min-width:160px;text-align:center;display:inline-flex;justify-content:center;align-items:center;white-space:nowrap;position:relative}.mat-tab-label:focus{outline:0}.mat-tab-label:focus:not(.mat-tab-disabled){opacity:1}@media (-ms-high-contrast:active){.mat-tab-label:focus{outline:dotted 2px}}.mat-tab-label.mat-tab-disabled{cursor:default}@media (-ms-high-contrast:active){.mat-tab-label.mat-tab-disabled{opacity:.5}}.mat-tab-label .mat-tab-label-content{display:inline-flex;justify-content:center;align-items:center;white-space:nowrap}@media (-ms-high-contrast:active){.mat-tab-label{opacity:1}}@media (max-width:599px){.mat-tab-label{padding:0 12px}}@media (max-width:959px){.mat-tab-label{padding:0 12px}}.mat-tab-group[mat-stretch-tabs]>.mat-tab-header .mat-tab-label{flex-basis:0;flex-grow:1}.mat-tab-body-wrapper{position:relative;overflow:hidden;display:flex;transition:height .5s cubic-bezier(.35,0,.25,1)}.mat-tab-body{top:0;left:0;right:0;bottom:0;position:absolute;display:block;overflow:hidden;flex-basis:100%}.mat-tab-body.mat-tab-body-active{position:relative;overflow-x:hidden;overflow-y:auto;z-index:1;flex-grow:1}.mat-tab-group.mat-tab-group-dynamic-height .mat-tab-body.mat-tab-body-active{overflow-y:hidden}"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,inputs:["color","disableRipple"],host:{class:"mat-tab-group","[class.mat-tab-group-dynamic-height]":"dynamicHeight","[class.mat-tab-group-inverted-header]":'headerPosition === "below"'}}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:r.ChangeDetectorRef},{type:void 0,decorators:[{type:r.Inject,args:[F]},{type:r.Optional}]}]},t.propDecorators={_tabs:[{type:r.ContentChildren,args:[C]}],_tabBodyWrapper:[{type:r.ViewChild,args:["tabBodyWrapper"]}],_tabHeader:[{type:r.ViewChild,args:["tabHeader"]}],dynamicHeight:[{type:r.Input}],selectedIndex:[{type:r.Input}],headerPosition:[{type:r.Input}],animationDuration:[{type:r.Input}],backgroundColor:[{type:r.Input}],selectedIndexChange:[{type:r.Output}],focusChange:[{type:r.Output}],animationDone:[{type:r.Output}],selectedTabChange:[{type:r.Output}]},t}(V),z=function(){return function(e){this._elementRef=e}}(),U=Object(a.mixinDisableRipple)(Object(a.mixinColor)(z,"primary")),H=function(e){function t(t,n,r,i,o){var a=e.call(this,t)||this;return a._dir=n,a._ngZone=r,a._changeDetectorRef=i,a._viewportRuler=o,a._onDestroy=new s.Subject,a}return Object(i.__extends)(t,e),Object.defineProperty(t.prototype,"backgroundColor",{get:function(){return this._backgroundColor},set:function(e){var t=this._elementRef.nativeElement;t.classList.remove("mat-background-"+this.backgroundColor),e&&t.classList.add("mat-background-"+e),this._backgroundColor=e},enumerable:!0,configurable:!0}),t.prototype.updateActiveLink=function(e){this._activeLinkChanged=!!e,this._changeDetectorRef.markForCheck()},t.prototype.ngAfterContentInit=function(){var e=this;this._ngZone.runOutsideAngular(function(){var t=e._dir?e._dir.change:Object(s.of)(null);return Object(s.merge)(t,e._viewportRuler.change(10)).pipe(Object(l.takeUntil)(e._onDestroy)).subscribe(function(){return e._alignInkBar()})})},t.prototype.ngAfterContentChecked=function(){if(this._activeLinkChanged){var e=this._tabLinks.find(function(e){return e.active});this._activeLinkElement=e?e._elementRef:null,this._alignInkBar(),this._activeLinkChanged=!1}},t.prototype.ngOnDestroy=function(){this._onDestroy.next(),this._onDestroy.complete()},t.prototype._alignInkBar=function(){this._activeLinkElement?(this._inkBar.show(),this._inkBar.alignToElement(this._activeLinkElement.nativeElement)):this._inkBar.hide()},t.decorators=[{type:r.Component,args:[{selector:"[mat-tab-nav-bar]",exportAs:"matTabNavBar, matTabNav",inputs:["color","disableRipple"],template:'<div class="mat-tab-links" (cdkObserveContent)="_alignInkBar()"><ng-content></ng-content><mat-ink-bar></mat-ink-bar></div>',styles:[".mat-tab-nav-bar{overflow:hidden;position:relative;flex-shrink:0}.mat-tab-links{position:relative;display:flex}[mat-align-tabs=center] .mat-tab-links{justify-content:center}[mat-align-tabs=end] .mat-tab-links{justify-content:flex-end}.mat-tab-link{height:48px;padding:0 24px;cursor:pointer;box-sizing:border-box;opacity:.6;min-width:160px;text-align:center;display:inline-flex;justify-content:center;align-items:center;white-space:nowrap;vertical-align:top;text-decoration:none;position:relative;overflow:hidden;-webkit-tap-highlight-color:transparent}.mat-tab-link:focus{outline:0}.mat-tab-link:focus:not(.mat-tab-disabled){opacity:1}@media (-ms-high-contrast:active){.mat-tab-link:focus{outline:dotted 2px}}.mat-tab-link.mat-tab-disabled{cursor:default}@media (-ms-high-contrast:active){.mat-tab-link.mat-tab-disabled{opacity:.5}}.mat-tab-link .mat-tab-label-content{display:inline-flex;justify-content:center;align-items:center;white-space:nowrap}@media (-ms-high-contrast:active){.mat-tab-link{opacity:1}}[mat-stretch-tabs] .mat-tab-link{flex-basis:0;flex-grow:1}.mat-tab-link.mat-tab-disabled{pointer-events:none}@media (max-width:599px){.mat-tab-link{min-width:72px}}.mat-ink-bar{position:absolute;bottom:0;height:2px;transition:.5s cubic-bezier(.35,0,.25,1)}.mat-tab-group-inverted-header .mat-ink-bar{bottom:auto;top:0}@media (-ms-high-contrast:active){.mat-ink-bar{outline:solid 2px;height:0}}"],host:{class:"mat-tab-nav-bar"},encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:c.Directionality,decorators:[{type:r.Optional}]},{type:r.NgZone},{type:r.ChangeDetectorRef},{type:f.ViewportRuler}]},t.propDecorators={_inkBar:[{type:r.ViewChild,args:[b]}],_tabLinks:[{type:r.ContentChildren,args:[Object(r.forwardRef)(function(){return W}),{descendants:!0}]}],backgroundColor:[{type:r.Input}]},t}(U),q=function(){return function(){}}(),G=Object(a.mixinTabIndex)(Object(a.mixinDisableRipple)(Object(a.mixinDisabled)(q))),W=function(e){function t(t,n,r,i,o,s,u){var c=e.call(this)||this;return c._tabNavBar=t,c._elementRef=n,c._focusMonitor=u,c._isActive=!1,c._ripplesGloballyDisabled=!1,c.rippleConfig={},c._tabLinkRipple=new a.RippleRenderer(c,r,n,i),c._tabLinkRipple.setupTriggerEvents(n.nativeElement),c.tabIndex=parseInt(s)||0,o&&(c._ripplesGloballyDisabled=!!o.disabled,c.rippleConfig={terminateOnPointerUp:o.terminateOnPointerUp,animation:o.animation}),u&&u.monitor(n),c}return Object(i.__extends)(t,e),Object.defineProperty(t.prototype,"active",{get:function(){return this._isActive},set:function(e){e!==this._isActive&&(this._isActive=e,this._tabNavBar.updateActiveLink(this._elementRef))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"rippleDisabled",{get:function(){return this.disabled||this.disableRipple||this._tabNavBar.disableRipple||this._ripplesGloballyDisabled},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){this._tabLinkRipple._removeTriggerEvents(),this._focusMonitor&&this._focusMonitor.stopMonitoring(this._elementRef)},t.decorators=[{type:r.Directive,args:[{selector:"[mat-tab-link], [matTabLink]",exportAs:"matTabLink",inputs:["disabled","disableRipple","tabIndex"],host:{class:"mat-tab-link","[attr.aria-current]":"active","[attr.aria-disabled]":"disabled.toString()","[attr.tabIndex]":"tabIndex","[class.mat-tab-disabled]":"disabled","[class.mat-tab-label-active]":"active"}}]}],t.ctorParameters=function(){return[{type:H},{type:r.ElementRef},{type:r.NgZone},{type:m.Platform},{type:void 0,decorators:[{type:r.Optional},{type:r.Inject,args:[a.MAT_RIPPLE_GLOBAL_OPTIONS]}]},{type:String,decorators:[{type:r.Attribute,args:["tabindex"]}]},{type:h.FocusMonitor}]},t.propDecorators={active:[{type:r.Input}]},t}(G),Y=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[v.CommonModule,a.MatCommonModule,o.PortalModule,a.MatRippleModule,g.ObserversModule,h.A11yModule],exports:[a.MatCommonModule,B,x,C,H,W,j],declarations:[B,x,C,b,T,H,W,k,S,P,j]}]}],e}()},"./node_modules/@angular/material/esm5/toolbar.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatToolbarModule",function(){return f}),n.d(t,"throwToolbarMixedModesError",function(){return p}),n.d(t,"MatToolbarBase",function(){return u}),n.d(t,"_MatToolbarMixinBase",function(){return c}),n.d(t,"MatToolbarRow",function(){return l}),n.d(t,"MatToolbar",function(){return d});var r=n("./node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),o=n("./node_modules/@angular/common/fesm5/common.js"),a=n("./node_modules/@angular/core/fesm5/core.js"),s=n("./node_modules/@angular/material/esm5/core.es5.js"),u=function(){return function(e){this._elementRef=e}}(),c=Object(s.mixinColor)(u),l=function(){function e(){}return e.decorators=[{type:a.Directive,args:[{selector:"mat-toolbar-row",exportAs:"matToolbarRow",host:{class:"mat-toolbar-row"}}]}],e}(),d=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i._platform=n,i._document=r,i}return Object(r.__extends)(t,e),t.prototype.ngAfterViewInit=function(){var e=this;Object(a.isDevMode)()&&this._platform.isBrowser&&(this._checkToolbarMixedModes(),this._toolbarRows.changes.subscribe(function(){return e._checkToolbarMixedModes()}))},t.prototype._checkToolbarMixedModes=function(){var e=this;this._toolbarRows.length&&(Array.from(this._elementRef.nativeElement.childNodes).filter(function(e){return!(e.classList&&e.classList.contains("mat-toolbar-row"))}).filter(function(t){return t.nodeType!==(e._document?e._document.COMMENT_NODE:8)}).some(function(e){return!(!e.textContent||!e.textContent.trim())})&&p())},t.decorators=[{type:a.Component,args:[{selector:"mat-toolbar",exportAs:"matToolbar",template:'<ng-content></ng-content><ng-content select="mat-toolbar-row"></ng-content>',styles:["@media (-ms-high-contrast:active){.mat-toolbar{outline:solid 1px}}.mat-toolbar-row,.mat-toolbar-single-row{display:flex;box-sizing:border-box;padding:0 16px;width:100%;flex-direction:row;align-items:center;white-space:nowrap}.mat-toolbar-multiple-rows{display:flex;box-sizing:border-box;flex-direction:column;width:100%}.mat-toolbar-multiple-rows{min-height:64px}.mat-toolbar-row,.mat-toolbar-single-row{height:64px}@media (max-width:599px){.mat-toolbar-multiple-rows{min-height:56px}.mat-toolbar-row,.mat-toolbar-single-row{height:56px}}"],inputs:["color"],host:{class:"mat-toolbar","[class.mat-toolbar-multiple-rows]":"_toolbarRows.length > 0","[class.mat-toolbar-single-row]":"_toolbarRows.length === 0"},changeDetection:a.ChangeDetectionStrategy.OnPush,encapsulation:a.ViewEncapsulation.None}]}],t.ctorParameters=function(){return[{type:a.ElementRef},{type:i.Platform},{type:void 0,decorators:[{type:a.Inject,args:[o.DOCUMENT]}]}]},t.propDecorators={_toolbarRows:[{type:a.ContentChildren,args:[l]}]},t}(c);function p(){throw Error("MatToolbar: Attempting to combine different toolbar modes. Either specify multiple `<mat-toolbar-row>` elements explicitly or just place content inside of a `<mat-toolbar>` for a single row.")}var f=function(){function e(){}return e.decorators=[{type:a.NgModule,args:[{imports:[s.MatCommonModule],exports:[d,l,s.MatCommonModule],declarations:[d,l]}]}],e}()},"./node_modules/@angular/material/esm5/tooltip.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MatTooltipModule",function(){return T}),n.d(t,"getMatTooltipInvalidPositionError",function(){return O}),n.d(t,"MAT_TOOLTIP_SCROLL_STRATEGY_FACTORY",function(){return C}),n.d(t,"MAT_TOOLTIP_DEFAULT_OPTIONS_FACTORY",function(){return k}),n.d(t,"SCROLL_THROTTLE_MS",function(){return j}),n.d(t,"TOOLTIP_PANEL_CLASS",function(){return x}),n.d(t,"MAT_TOOLTIP_SCROLL_STRATEGY",function(){return w}),n.d(t,"MAT_TOOLTIP_SCROLL_STRATEGY_FACTORY_PROVIDER",function(){return E}),n.d(t,"MAT_TOOLTIP_DEFAULT_OPTIONS",function(){return S}),n.d(t,"MatTooltip",function(){return M}),n.d(t,"TooltipComponent",function(){return A}),n.d(t,"matTooltipAnimations",function(){return b});var r=n("./node_modules/@angular/animations/fesm5/animations.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/cdk/esm5/a11y.es5.js"),a=n("./node_modules/@angular/cdk/esm5/bidi.es5.js"),s=n("./node_modules/@angular/cdk/esm5/coercion.es5.js"),u=n("./node_modules/@angular/cdk/esm5/keycodes.es5.js"),c=n("./node_modules/@angular/cdk/esm5/layout.es5.js"),l=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),d=n("./node_modules/@angular/cdk/esm5/overlay.es5.js"),p=n("./node_modules/@angular/cdk/esm5/scrolling.es5.js"),f=n("./node_modules/@angular/cdk/esm5/platform.es5.js"),h=n("./node_modules/@angular/cdk/esm5/portal.es5.js"),m=n("./node_modules/rxjs/_esm5/operators/index.js"),g=n("./node_modules/@angular/core/fesm5/core.js"),v=n("./node_modules/rxjs/_esm5/index.js"),_=n("./node_modules/@angular/common/fesm5/common.js"),y=n("./node_modules/@angular/material/esm5/core.es5.js"),b={tooltipState:Object(r.trigger)("state",[Object(r.state)("initial, void, hidden",Object(r.style)({opacity:0,transform:"scale(0)"})),Object(r.state)("visible",Object(r.style)({transform:"scale(1)"})),Object(r.transition)("* => visible",Object(r.animate)("200ms cubic-bezier(0, 0, 0.2, 1)",Object(r.keyframes)([Object(r.style)({opacity:0,transform:"scale(0)",offset:0}),Object(r.style)({opacity:.5,transform:"scale(0.99)",offset:.5}),Object(r.style)({opacity:1,transform:"scale(1)",offset:1})]))),Object(r.transition)("* => hidden",Object(r.animate)("100ms cubic-bezier(0, 0, 0.2, 1)",Object(r.style)({opacity:0})))])},j=20,x="mat-tooltip-panel";function O(e){return Error('Tooltip position "'+e+'" is invalid.')}var w=new g.InjectionToken("mat-tooltip-scroll-strategy");function C(e){return function(){return e.scrollStrategies.reposition({scrollThrottle:j})}}var E={provide:w,deps:[d.Overlay],useFactory:C},S=new g.InjectionToken("mat-tooltip-default-options",{providedIn:"root",factory:k});function k(){return{showDelay:0,hideDelay:0,touchendHideDelay:1500}}var M=function(){function e(e,t,n,r,i,o,a,s,u,c,l,d){var p=this;this._overlay=e,this._elementRef=t,this._scrollDispatcher=n,this._viewContainerRef=r,this._ngZone=i,this._ariaDescriber=a,this._focusMonitor=s,this._dir=c,this._defaultOptions=l,this._position="below",this._disabled=!1,this.showDelay=this._defaultOptions.showDelay,this.hideDelay=this._defaultOptions.hideDelay,this._message="",this._manualListeners=new Map,this._destroyed=new v.Subject,this._scrollStrategy=u;var f=t.nativeElement,h=f.style,g="undefined"==typeof window||window.Hammer||d;o.IOS||o.ANDROID?g||this._manualListeners.set("touchstart",function(){return p.show()}):this._manualListeners.set("mouseenter",function(){return p.show()}).set("mouseleave",function(){return p.hide()}),this._manualListeners.forEach(function(e,t){return f.addEventListener(t,e)}),"INPUT"!==f.nodeName&&"TEXTAREA"!==f.nodeName||(h.webkitUserSelect=h.userSelect=h.msUserSelect=""),f.draggable&&"none"===h.webkitUserDrag&&(h.webkitUserDrag=""),s.monitor(t).pipe(Object(m.takeUntil)(this._destroyed)).subscribe(function(e){e?"keyboard"===e&&i.run(function(){return p.show()}):i.run(function(){return p.hide(0)})})}return Object.defineProperty(e.prototype,"position",{get:function(){return this._position},set:function(e){e!==this._position&&(this._position=e,this._overlayRef&&(this._updatePosition(),this._tooltipInstance&&this._tooltipInstance.show(0),this._overlayRef.updatePosition()))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disabled",{get:function(){return this._disabled},set:function(e){this._disabled=Object(s.coerceBooleanProperty)(e),this._disabled&&this.hide(0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"message",{get:function(){return this._message},set:function(e){this._ariaDescriber.removeDescription(this._elementRef.nativeElement,this._message),this._message=null!=e?(""+e).trim():"",!this._message&&this._isTooltipVisible()?this.hide(0):(this._updateTooltipMessage(),this._ariaDescriber.describe(this._elementRef.nativeElement,this.message))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"tooltipClass",{get:function(){return this._tooltipClass},set:function(e){this._tooltipClass=e,this._tooltipInstance&&this._setTooltipClass(this._tooltipClass)},enumerable:!0,configurable:!0}),e.prototype.ngOnDestroy=function(){var e=this;this._overlayRef&&(this._overlayRef.dispose(),this._tooltipInstance=null),this._manualListeners.forEach(function(t,n){e._elementRef.nativeElement.removeEventListener(n,t)}),this._manualListeners.clear(),this._destroyed.next(),this._destroyed.complete(),this._ariaDescriber.removeDescription(this._elementRef.nativeElement,this.message),this._focusMonitor.stopMonitoring(this._elementRef)},e.prototype.show=function(e){var t=this;if(void 0===e&&(e=this.showDelay),!this.disabled&&this.message&&(!this._isTooltipVisible()||this._tooltipInstance._showTimeoutId||this._tooltipInstance._hideTimeoutId)){var n=this._createOverlay();this._detach(),this._portal=this._portal||new h.ComponentPortal(A,this._viewContainerRef),this._tooltipInstance=n.attach(this._portal).instance,this._tooltipInstance.afterHidden().pipe(Object(m.takeUntil)(this._destroyed)).subscribe(function(){return t._detach()}),this._setTooltipClass(this._tooltipClass),this._updateTooltipMessage(),this._tooltipInstance.show(e)}},e.prototype.hide=function(e){void 0===e&&(e=this.hideDelay),this._tooltipInstance&&this._tooltipInstance.hide(e)},e.prototype.toggle=function(){this._isTooltipVisible()?this.hide():this.show()},e.prototype._isTooltipVisible=function(){return!!this._tooltipInstance&&this._tooltipInstance.isVisible()},e.prototype._handleKeydown=function(e){this._isTooltipVisible()&&e.keyCode===u.ESCAPE&&(e.stopPropagation(),this.hide(0))},e.prototype._handleTouchend=function(){this.hide(this._defaultOptions.touchendHideDelay)},e.prototype._createOverlay=function(){var e=this;if(this._overlayRef)return this._overlayRef;var t=this._overlay.position().flexibleConnectedTo(this._elementRef).withTransformOriginOn(".mat-tooltip").withFlexibleDimensions(!1).withViewportMargin(8),n=this._scrollDispatcher.getAncestorScrollContainers(this._elementRef);return t.withScrollableContainers(n),t.positionChanges.pipe(Object(m.takeUntil)(this._destroyed)).subscribe(function(t){e._tooltipInstance&&t.scrollableViewProperties.isOverlayClipped&&e._tooltipInstance.isVisible()&&e._ngZone.run(function(){return e.hide(0)})}),this._overlayRef=this._overlay.create({direction:this._dir,positionStrategy:t,panelClass:x,scrollStrategy:this._scrollStrategy()}),this._updatePosition(),this._overlayRef.detachments().pipe(Object(m.takeUntil)(this._destroyed)).subscribe(function(){return e._detach()}),this._overlayRef},e.prototype._detach=function(){this._overlayRef&&this._overlayRef.hasAttached()&&this._overlayRef.detach(),this._tooltipInstance=null},e.prototype._updatePosition=function(){var e=this._overlayRef.getConfig().positionStrategy,t=this._getOrigin(),n=this._getOverlayPosition();e.withPositions([Object(i.__assign)({},t.main,n.main),Object(i.__assign)({},t.fallback,n.fallback)])},e.prototype._getOrigin=function(){var e,t=!this._dir||"ltr"==this._dir.value,n=this.position;if("above"==n||"below"==n)e={originX:"center",originY:"above"==n?"top":"bottom"};else if("before"==n||"left"==n&&t||"right"==n&&!t)e={originX:"start",originY:"center"};else{if(!("after"==n||"right"==n&&t||"left"==n&&!t))throw O(n);e={originX:"end",originY:"center"}}var r=this._invertPosition(e.originX,e.originY);return{main:e,fallback:{originX:r.x,originY:r.y}}},e.prototype._getOverlayPosition=function(){var e,t=!this._dir||"ltr"==this._dir.value,n=this.position;if("above"==n)e={overlayX:"center",overlayY:"bottom"};else if("below"==n)e={overlayX:"center",overlayY:"top"};else if("before"==n||"left"==n&&t||"right"==n&&!t)e={overlayX:"end",overlayY:"center"};else{if(!("after"==n||"right"==n&&t||"left"==n&&!t))throw O(n);e={overlayX:"start",overlayY:"center"}}var r=this._invertPosition(e.overlayX,e.overlayY);return{main:e,fallback:{overlayX:r.x,overlayY:r.y}}},e.prototype._updateTooltipMessage=function(){var e=this;this._tooltipInstance&&(this._tooltipInstance.message=this.message,this._tooltipInstance._markForCheck(),this._ngZone.onMicrotaskEmpty.asObservable().pipe(Object(m.take)(1),Object(m.takeUntil)(this._destroyed)).subscribe(function(){e._tooltipInstance&&e._overlayRef.updatePosition()}))},e.prototype._setTooltipClass=function(e){this._tooltipInstance&&(this._tooltipInstance.tooltipClass=e,this._tooltipInstance._markForCheck())},e.prototype._invertPosition=function(e,t){return"above"===this.position||"below"===this.position?"top"===t?t="bottom":"bottom"===t&&(t="top"):"end"===e?e="start":"start"===e&&(e="end"),{x:e,y:t}},e.decorators=[{type:g.Directive,args:[{selector:"[matTooltip]",exportAs:"matTooltip",host:{"(longpress)":"show()","(keydown)":"_handleKeydown($event)","(touchend)":"_handleTouchend()"}}]}],e.ctorParameters=function(){return[{type:d.Overlay},{type:g.ElementRef},{type:p.ScrollDispatcher},{type:g.ViewContainerRef},{type:g.NgZone},{type:f.Platform},{type:o.AriaDescriber},{type:o.FocusMonitor},{type:void 0,decorators:[{type:g.Inject,args:[w]}]},{type:a.Directionality,decorators:[{type:g.Optional}]},{type:void 0,decorators:[{type:g.Optional},{type:g.Inject,args:[S]}]},{type:void 0,decorators:[{type:g.Optional},{type:g.Inject,args:[l.HAMMER_LOADER]}]}]},e.propDecorators={position:[{type:g.Input,args:["matTooltipPosition"]}],disabled:[{type:g.Input,args:["matTooltipDisabled"]}],showDelay:[{type:g.Input,args:["matTooltipShowDelay"]}],hideDelay:[{type:g.Input,args:["matTooltipHideDelay"]}],message:[{type:g.Input,args:["matTooltip"]}],tooltipClass:[{type:g.Input,args:["matTooltipClass"]}]},e}(),A=function(){function e(e,t){this._changeDetectorRef=e,this._breakpointObserver=t,this._visibility="initial",this._closeOnInteraction=!1,this._onHide=new v.Subject,this._isHandset=this._breakpointObserver.observe(c.Breakpoints.Handset)}return e.prototype.show=function(e){var t=this;this._hideTimeoutId&&(clearTimeout(this._hideTimeoutId),this._hideTimeoutId=null),this._closeOnInteraction=!0,this._showTimeoutId=setTimeout(function(){t._visibility="visible",t._showTimeoutId=null,t._markForCheck()},e)},e.prototype.hide=function(e){var t=this;this._showTimeoutId&&(clearTimeout(this._showTimeoutId),this._showTimeoutId=null),this._hideTimeoutId=setTimeout(function(){t._visibility="hidden",t._hideTimeoutId=null,t._markForCheck()},e)},e.prototype.afterHidden=function(){return this._onHide.asObservable()},e.prototype.isVisible=function(){return"visible"===this._visibility},e.prototype._animationStart=function(){this._closeOnInteraction=!1},e.prototype._animationDone=function(e){var t=e.toState;"hidden"!==t||this.isVisible()||this._onHide.next(),"visible"!==t&&"hidden"!==t||(this._closeOnInteraction=!0)},e.prototype._handleBodyInteraction=function(){this._closeOnInteraction&&this.hide(0)},e.prototype._markForCheck=function(){this._changeDetectorRef.markForCheck()},e.decorators=[{type:g.Component,args:[{selector:"mat-tooltip-component",template:'<div class="mat-tooltip" [ngClass]="tooltipClass" [class.mat-tooltip-handset]="(_isHandset | async)?.matches" [@state]="_visibility" (@state.start)="_animationStart()" (@state.done)="_animationDone($event)">{{message}}</div>',styles:[".mat-tooltip-panel{pointer-events:none!important}.mat-tooltip{color:#fff;border-radius:4px;margin:14px;max-width:250px;padding-left:8px;padding-right:8px;overflow:hidden;text-overflow:ellipsis}@media (-ms-high-contrast:active){.mat-tooltip{outline:solid 1px}}.mat-tooltip-handset{margin:24px;padding-left:16px;padding-right:16px}"],encapsulation:g.ViewEncapsulation.None,changeDetection:g.ChangeDetectionStrategy.OnPush,animations:[b.tooltipState],host:{"[style.zoom]":'_visibility === "visible" ? 1 : null',"(body:click)":"this._handleBodyInteraction()","aria-hidden":"true"}}]}],e.ctorParameters=function(){return[{type:g.ChangeDetectorRef},{type:c.BreakpointObserver}]},e}(),T=function(){function e(){}return e.decorators=[{type:g.NgModule,args:[{imports:[o.A11yModule,_.CommonModule,d.OverlayModule,y.MatCommonModule],exports:[M,A,y.MatCommonModule],declarations:[M,A],entryComponents:[A],providers:[E,{provide:l.HAMMER_GESTURE_CONFIG,useClass:y.GestureConfig}]}]}],e}()},"./node_modules/@angular/material/esm5/tree.es5.js":function(e,t,n){"use strict";n.r(t),n.d(t,"_MatTreeNodeMixinBase",function(){return p}),n.d(t,"_MatNestedTreeNodeMixinBase",function(){return f}),n.d(t,"MatTreeNode",function(){return h}),n.d(t,"MatTreeNodeDef",function(){return m}),n.d(t,"MatNestedTreeNode",function(){return g}),n.d(t,"MatTreeNodePadding",function(){return v}),n.d(t,"MatTree",function(){return _}),n.d(t,"MatTreeModule",function(){return j}),n.d(t,"MatTreeNodeToggle",function(){return y}),n.d(t,"MatTreeNodeOutlet",function(){return d}),n.d(t,"MatTreeFlattener",function(){return x}),n.d(t,"MatTreeFlatDataSource",function(){return O}),n.d(t,"MatTreeNestedDataSource",function(){return w});var r=n("./node_modules/@angular/core/fesm5/core.js"),i=n("./node_modules/tslib/tslib.es6.js"),o=n("./node_modules/@angular/cdk/esm5/tree.es5.js"),a=n("./node_modules/@angular/material/esm5/core.es5.js"),s=n("./node_modules/@angular/common/fesm5/common.js"),u=n("./node_modules/@angular/cdk/esm5/collections.es5.js"),c=n("./node_modules/rxjs/_esm5/index.js"),l=n("./node_modules/rxjs/_esm5/operators/index.js"),d=function(){function e(e){this.viewContainer=e}return e.decorators=[{type:r.Directive,args:[{selector:"[matTreeNodeOutlet]"}]}],e.ctorParameters=function(){return[{type:r.ViewContainerRef}]},e}(),p=Object(a.mixinTabIndex)(Object(a.mixinDisabled)(o.CdkTreeNode)),f=Object(a.mixinTabIndex)(Object(a.mixinDisabled)(o.CdkNestedTreeNode)),h=function(e){function t(t,n,r){var i=e.call(this,t,n)||this;return i._elementRef=t,i._tree=n,i.role="treeitem",i.tabIndex=Number(r)||0,i}return Object(i.__extends)(t,e),t.decorators=[{type:r.Directive,args:[{selector:"mat-tree-node",exportAs:"matTreeNode",inputs:["disabled","tabIndex"],host:{"[attr.aria-expanded]":"isExpanded","[attr.aria-level]":'role === "treeitem" ? level : null',"[attr.role]":"role",class:"mat-tree-node"},providers:[{provide:o.CdkTreeNode,useExisting:t}]}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:o.CdkTree},{type:String,decorators:[{type:r.Attribute,args:["tabindex"]}]}]},t.propDecorators={role:[{type:r.Input}]},t}(p),m=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(i.__extends)(t,e),t.decorators=[{type:r.Directive,args:[{selector:"[matTreeNodeDef]",inputs:["when: matTreeNodeDefWhen"],providers:[{provide:o.CdkTreeNodeDef,useExisting:t}]}]}],t.propDecorators={data:[{type:r.Input,args:["matTreeNode"]}]},t}(o.CdkTreeNodeDef),g=function(e){function t(t,n,r,i){var o=e.call(this,t,n,r)||this;return o._elementRef=t,o._tree=n,o._differs=r,o.tabIndex=Number(i)||0,o}return Object(i.__extends)(t,e),t.prototype.ngAfterContentInit=function(){e.prototype.ngAfterContentInit.call(this)},t.prototype.ngOnDestroy=function(){e.prototype.ngOnDestroy.call(this)},t.decorators=[{type:r.Directive,args:[{selector:"mat-nested-tree-node",exportAs:"matNestedTreeNode",host:{"[attr.aria-expanded]":"isExpanded","[attr.role]":"role",class:"mat-nested-tree-node"},inputs:["disabled","tabIndex"],providers:[{provide:o.CdkNestedTreeNode,useExisting:t},{provide:o.CdkTreeNode,useExisting:t}]}]}],t.ctorParameters=function(){return[{type:r.ElementRef},{type:o.CdkTree},{type:r.IterableDiffers},{type:String,decorators:[{type:r.Attribute,args:["tabindex"]}]}]},t.propDecorators={node:[{type:r.Input,args:["matNestedTreeNode"]}],nodeOutlet:[{type:r.ContentChildren,args:[d]}]},t}(f),v=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(i.__extends)(t,e),t.decorators=[{type:r.Directive,args:[{selector:"[matTreeNodePadding]",providers:[{provide:o.CdkTreeNodePadding,useExisting:t}]}]}],t.propDecorators={level:[{type:r.Input,args:["matTreeNodePadding"]}],indent:[{type:r.Input,args:["matTreeNodePaddingIndent"]}]},t}(o.CdkTreeNodePadding),_=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(i.__extends)(t,e),t.decorators=[{type:r.Component,args:[{selector:"mat-tree",exportAs:"matTree",template:"<ng-container matTreeNodeOutlet></ng-container>",host:{class:"mat-tree",role:"tree"},styles:[".mat-tree{display:block}.mat-tree-node{display:flex;align-items:center;min-height:48px;flex:1;overflow:hidden;word-wrap:break-word}.mat-nested-tree-ndoe{border-bottom-width:0}"],encapsulation:r.ViewEncapsulation.None,changeDetection:r.ChangeDetectionStrategy.OnPush,providers:[{provide:o.CdkTree,useExisting:t}]}]}],t.propDecorators={_nodeOutlet:[{type:r.ViewChild,args:[d]}]},t}(o.CdkTree),y=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.recursive=!1,t}return Object(i.__extends)(t,e),t.decorators=[{type:r.Directive,args:[{selector:"[matTreeNodeToggle]",host:{"(click)":"_toggle($event)"},providers:[{provide:o.CdkTreeNodeToggle,useExisting:t}]}]}],t.propDecorators={recursive:[{type:r.Input,args:["matTreeNodeToggleRecursive"]}]},t}(o.CdkTreeNodeToggle),b=[g,m,v,y,_,h,d],j=function(){function e(){}return e.decorators=[{type:r.NgModule,args:[{imports:[o.CdkTreeModule,s.CommonModule,a.MatCommonModule],exports:b,declarations:b}]}],e}(),x=function(){function e(e,t,n,r){this.transformFunction=e,this.getLevel=t,this.isExpandable=n,this.getChildren=r}return e.prototype._flattenNode=function(e,t,n,r){var i=this,o=this.transformFunction(e,t);if(n.push(o),this.isExpandable(o)){var a=this.getChildren(e);a&&(Array.isArray(a)?this._flattenChildren(a,t,n,r):a.pipe(Object(l.take)(1)).subscribe(function(e){i._flattenChildren(e,t,n,r)}))}return n},e.prototype._flattenChildren=function(e,t,n,r){var i=this;e.forEach(function(o,a){var s=r.slice();s.push(a!=e.length-1),i._flattenNode(o,t+1,n,s)})},e.prototype.flattenNodes=function(e){var t=this,n=[];return e.forEach(function(e){return t._flattenNode(e,0,n,[])}),n},e.prototype.expandFlattenedNodes=function(e,t){var n=this,r=[],i=[];return i[0]=!0,e.forEach(function(e){for(var o=!0,a=0;a<=n.getLevel(e);a++)o=o&&i[a];o&&r.push(e),n.isExpandable(e)&&(i[n.getLevel(e)+1]=t.isExpanded(e))}),r},e}(),O=function(e){function t(t,n,r){void 0===r&&(r=[]);var i=e.call(this)||this;return i.treeControl=t,i.treeFlattener=n,i._flattenedData=new c.BehaviorSubject([]),i._expandedData=new c.BehaviorSubject([]),i._data=new c.BehaviorSubject(r),i}return Object(i.__extends)(t,e),Object.defineProperty(t.prototype,"data",{get:function(){return this._data.value},set:function(e){this._data.next(e),this._flattenedData.next(this.treeFlattener.flattenNodes(this.data)),this.treeControl.dataNodes=this._flattenedData.value},enumerable:!0,configurable:!0}),t.prototype.connect=function(e){var t=this,n=[e.viewChange,this.treeControl.expansionModel.onChange,this._flattenedData];return c.merge.apply(void 0,n).pipe(Object(l.map)(function(){return t._expandedData.next(t.treeFlattener.expandFlattenedNodes(t._flattenedData.value,t.treeControl)),t._expandedData.value}))},t.prototype.disconnect=function(){},t}(u.DataSource),w=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t._data=new c.BehaviorSubject([]),t}return Object(i.__extends)(t,e),Object.defineProperty(t.prototype,"data",{get:function(){return this._data.value},set:function(e){this._data.next(e)},enumerable:!0,configurable:!0}),t.prototype.connect=function(e){var t=this;return c.merge.apply(void 0,[e.viewChange,this._data]).pipe(Object(l.map)(function(){return t.data}))},t.prototype.disconnect=function(){},t}(u.DataSource)},"./node_modules/@angular/platform-browser-dynamic/fesm5/platform-browser-dynamic.js":function(e,t,n){"use strict";n.r(t),n.d(t,"ɵangular_packages_platform_browser_dynamic_platform_browser_dynamic_a",function(){return x}),n.d(t,"RESOURCE_CACHE_PROVIDER",function(){return w}),n.d(t,"platformBrowserDynamic",function(){return C}),n.d(t,"VERSION",function(){return O}),n.d(t,"JitCompilerFactory",function(){return v}),n.d(t,"ɵCompilerImpl",function(){return m}),n.d(t,"ɵplatformCoreDynamic",function(){return y}),n.d(t,"ɵINTERNAL_BROWSER_DYNAMIC_PLATFORM_PROVIDERS",function(){return j}),n.d(t,"ɵResourceLoaderImpl",function(){return b});var r,i=n("./node_modules/@angular/compiler/fesm5/compiler.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/platform-browser-dynamic/node_modules/tslib/tslib.es6.js"),s=n("./node_modules/@angular/common/fesm5/common.js"),u=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),c=((r=new Map).set(i.Identifiers.ANALYZE_FOR_ENTRY_COMPONENTS,o.ANALYZE_FOR_ENTRY_COMPONENTS),r.set(i.Identifiers.ElementRef,o.ElementRef),r.set(i.Identifiers.NgModuleRef,o.NgModuleRef),r.set(i.Identifiers.ViewContainerRef,o.ViewContainerRef),r.set(i.Identifiers.ChangeDetectorRef,o.ChangeDetectorRef),r.set(i.Identifiers.Renderer2,o.Renderer2),r.set(i.Identifiers.QueryList,o.QueryList),r.set(i.Identifiers.TemplateRef,o.TemplateRef),r.set(i.Identifiers.CodegenComponentFactoryResolver,o["ɵCodegenComponentFactoryResolver"]),r.set(i.Identifiers.ComponentFactoryResolver,o.ComponentFactoryResolver),r.set(i.Identifiers.ComponentFactory,o.ComponentFactory),r.set(i.Identifiers.ComponentRef,o.ComponentRef),r.set(i.Identifiers.NgModuleFactory,o.NgModuleFactory),r.set(i.Identifiers.createModuleFactory,o["ɵcmf"]),r.set(i.Identifiers.moduleDef,o["ɵmod"]),r.set(i.Identifiers.moduleProviderDef,o["ɵmpd"]),r.set(i.Identifiers.RegisterModuleFactoryFn,o["ɵregisterModuleFactory"]),r.set(i.Identifiers.Injector,o.Injector),r.set(i.Identifiers.ViewEncapsulation,o.ViewEncapsulation),r.set(i.Identifiers.ChangeDetectionStrategy,o.ChangeDetectionStrategy),r.set(i.Identifiers.SecurityContext,o.SecurityContext),r.set(i.Identifiers.LOCALE_ID,o.LOCALE_ID),r.set(i.Identifiers.TRANSLATIONS_FORMAT,o.TRANSLATIONS_FORMAT),r.set(i.Identifiers.inlineInterpolate,o["ɵinlineInterpolate"]),r.set(i.Identifiers.interpolate,o["ɵinterpolate"]),r.set(i.Identifiers.EMPTY_ARRAY,o["ɵEMPTY_ARRAY"]),r.set(i.Identifiers.EMPTY_MAP,o["ɵEMPTY_MAP"]),r.set(i.Identifiers.Renderer,o.Renderer),r.set(i.Identifiers.viewDef,o["ɵvid"]),r.set(i.Identifiers.elementDef,o["ɵeld"]),r.set(i.Identifiers.anchorDef,o["ɵand"]),r.set(i.Identifiers.textDef,o["ɵted"]),r.set(i.Identifiers.directiveDef,o["ɵdid"]),r.set(i.Identifiers.providerDef,o["ɵprd"]),r.set(i.Identifiers.queryDef,o["ɵqud"]),r.set(i.Identifiers.pureArrayDef,o["ɵpad"]),r.set(i.Identifiers.pureObjectDef,o["ɵpod"]),r.set(i.Identifiers.purePipeDef,o["ɵppd"]),r.set(i.Identifiers.pipeDef,o["ɵpid"]),r.set(i.Identifiers.nodeValue,o["ɵnov"]),r.set(i.Identifiers.ngContentDef,o["ɵncd"]),r.set(i.Identifiers.unwrapValue,o["ɵunv"]),r.set(i.Identifiers.createRendererType2,o["ɵcrt"]),r.set(i.Identifiers.createComponentFactory,o["ɵccf"]),r),l=function(){function e(){this.reflectionCapabilities=new o["ɵReflectionCapabilities"]}return e.prototype.componentModuleUrl=function(e,t){var n=t.moduleId;if("string"==typeof n)return Object(i.getUrlScheme)(n)?n:"package:"+n;if(null!==n&&void 0!==n)throw Object(i.syntaxError)('moduleId should be a string in "'+Object(o["ɵstringify"])(e)+"\". See https://goo.gl/wIDDiL for more information.\nIf you're using Webpack you should inline the template and the styles, see https://goo.gl/X2J8zc.");return"./"+Object(o["ɵstringify"])(e)},e.prototype.parameters=function(e){return this.reflectionCapabilities.parameters(e)},e.prototype.tryAnnotations=function(e){return this.annotations(e)},e.prototype.annotations=function(e){return this.reflectionCapabilities.annotations(e)},e.prototype.shallowAnnotations=function(e){throw new Error("Not supported in JIT mode")},e.prototype.propMetadata=function(e){return this.reflectionCapabilities.propMetadata(e)},e.prototype.hasLifecycleHook=function(e,t){return this.reflectionCapabilities.hasLifecycleHook(e,t)},e.prototype.guards=function(e){return this.reflectionCapabilities.guards(e)},e.prototype.resolveExternalReference=function(e){return c.get(e)||e.runtime},e}();var d=new o.InjectionToken("ErrorCollector"),p={provide:o.PACKAGE_ROOT_URL,useValue:"/"},f={get:function(e){throw new Error("No ResourceLoader implementation has been provided. Can't read the url \""+e+'"')}},h=new o.InjectionToken("HtmlParser"),m=function(){function e(e,t,n,r,o,a,s,u,c,l){this._metadataResolver=t,this._delegate=new i.JitCompiler(t,n,r,o,a,s,u,c,l,this.getExtraNgModuleProviders.bind(this)),this.injector=e}return e.prototype.getExtraNgModuleProviders=function(){return[this._metadataResolver.getProviderMetadata(new i.ProviderMeta(o.Compiler,{useValue:this}))]},e.prototype.compileModuleSync=function(e){return this._delegate.compileModuleSync(e)},e.prototype.compileModuleAsync=function(e){return this._delegate.compileModuleAsync(e)},e.prototype.compileModuleAndAllComponentsSync=function(e){var t=this._delegate.compileModuleAndAllComponentsSync(e);return{ngModuleFactory:t.ngModuleFactory,componentFactories:t.componentFactories}},e.prototype.compileModuleAndAllComponentsAsync=function(e){return this._delegate.compileModuleAndAllComponentsAsync(e).then(function(e){return{ngModuleFactory:e.ngModuleFactory,componentFactories:e.componentFactories}})},e.prototype.loadAotSummaries=function(e){this._delegate.loadAotSummaries(e)},e.prototype.hasAotSummary=function(e){return this._delegate.hasAotSummary(e)},e.prototype.getComponentFactory=function(e){return this._delegate.getComponentFactory(e)},e.prototype.clearCache=function(){this._delegate.clearCache()},e.prototype.clearCacheFor=function(e){this._delegate.clearCacheFor(e)},e.prototype.getModuleId=function(e){var t=this._metadataResolver.getNgModuleMetadata(e);return t&&t.id||void 0},e}(),g=[{provide:i.CompileReflector,useValue:new l},{provide:i.ResourceLoader,useValue:f},{provide:i.JitSummaryResolver,deps:[]},{provide:i.SummaryResolver,useExisting:i.JitSummaryResolver},{provide:o["ɵConsole"],deps:[]},{provide:i.Lexer,deps:[]},{provide:i.Parser,deps:[i.Lexer]},{provide:h,useClass:i.HtmlParser,deps:[]},{provide:i.I18NHtmlParser,useFactory:function(e,t,n,r,a){var s=(t=t||"")?r.missingTranslation:o.MissingTranslationStrategy.Ignore;return new i.I18NHtmlParser(e,t,n,s,a)},deps:[h,[new o.Optional,new o.Inject(o.TRANSLATIONS)],[new o.Optional,new o.Inject(o.TRANSLATIONS_FORMAT)],[i.CompilerConfig],[o["ɵConsole"]]]},{provide:i.HtmlParser,useExisting:i.I18NHtmlParser},{provide:i.TemplateParser,deps:[i.CompilerConfig,i.CompileReflector,i.Parser,i.ElementSchemaRegistry,i.I18NHtmlParser,o["ɵConsole"]]},{provide:i.DirectiveNormalizer,deps:[i.ResourceLoader,i.UrlResolver,i.HtmlParser,i.CompilerConfig]},{provide:i.CompileMetadataResolver,deps:[i.CompilerConfig,i.HtmlParser,i.NgModuleResolver,i.DirectiveResolver,i.PipeResolver,i.SummaryResolver,i.ElementSchemaRegistry,i.DirectiveNormalizer,o["ɵConsole"],[o.Optional,i.StaticSymbolCache],i.CompileReflector,[o.Optional,d]]},p,{provide:i.StyleCompiler,deps:[i.UrlResolver]},{provide:i.ViewCompiler,deps:[i.CompileReflector]},{provide:i.NgModuleCompiler,deps:[i.CompileReflector]},{provide:i.CompilerConfig,useValue:new i.CompilerConfig},{provide:o.Compiler,useClass:m,deps:[o.Injector,i.CompileMetadataResolver,i.TemplateParser,i.StyleCompiler,i.ViewCompiler,i.NgModuleCompiler,i.SummaryResolver,i.CompileReflector,i.CompilerConfig,o["ɵConsole"]]},{provide:i.DomElementSchemaRegistry,deps:[]},{provide:i.ElementSchemaRegistry,useExisting:i.DomElementSchemaRegistry},{provide:i.UrlResolver,deps:[o.PACKAGE_ROOT_URL]},{provide:i.DirectiveResolver,deps:[i.CompileReflector]},{provide:i.PipeResolver,deps:[i.CompileReflector]},{provide:i.NgModuleResolver,deps:[i.CompileReflector]}],v=function(){function e(e){var t={useJit:!0,defaultEncapsulation:o.ViewEncapsulation.Emulated,missingTranslation:o.MissingTranslationStrategy.Warning};this._defaultOptions=Object(a.__spread)([t],e)}return e.prototype.createCompiler=function(e){void 0===e&&(e=[]);var t,n,r,s={useJit:_((t=this._defaultOptions.concat(e)).map(function(e){return e.useJit})),defaultEncapsulation:_(t.map(function(e){return e.defaultEncapsulation})),providers:(n=t.map(function(e){return e.providers}),r=[],n.forEach(function(e){return e&&r.push.apply(r,Object(a.__spread)(e))}),r),missingTranslation:_(t.map(function(e){return e.missingTranslation})),preserveWhitespaces:_(t.map(function(e){return e.preserveWhitespaces}))};return o.Injector.create([g,{provide:i.CompilerConfig,useFactory:function(){return new i.CompilerConfig({useJit:s.useJit,jitDevMode:Object(o.isDevMode)(),defaultEncapsulation:s.defaultEncapsulation,missingTranslation:s.missingTranslation,preserveWhitespaces:s.preserveWhitespaces})},deps:[]},s.providers]).get(o.Compiler)},e}();function _(e){for(var t=e.length-1;t>=0;t--)if(void 0!==e[t])return e[t]}var y=Object(o.createPlatformFactory)(o.platformCore,"coreDynamic",[{provide:o.COMPILER_OPTIONS,useValue:{},multi:!0},{provide:o.CompilerFactory,useClass:v,deps:[o.COMPILER_OPTIONS]}]),b=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(a.__extends)(t,e),t.prototype.get=function(e){var t,n,r=new Promise(function(e,r){t=e,n=r}),i=new XMLHttpRequest;return i.open("GET",e,!0),i.responseType="text",i.onload=function(){var r=i.response||i.responseText,o=1223===i.status?204:i.status;0===o&&(o=r?200:0),200<=o&&o<=300?t(r):n("Failed to load "+e)},i.onerror=function(){n("Failed to load "+e)},i.send(),r},t=Object(a.__decorate)([Object(o.Injectable)()],t)}(i.ResourceLoader),j=[u["ɵINTERNAL_BROWSER_PLATFORM_PROVIDERS"],{provide:o.COMPILER_OPTIONS,useValue:{providers:[{provide:i.ResourceLoader,useClass:b,deps:[]}]},multi:!0},{provide:o.PLATFORM_ID,useValue:s["ɵPLATFORM_BROWSER_ID"]}],x=function(e){function t(){var t=e.call(this)||this;if(t._cache=o["ɵglobal"].$templateCache,null==t._cache)throw new Error("CachedResourceLoader: Template cache was not found in $templateCache.");return t}return Object(a.__extends)(t,e),t.prototype.get=function(e){return this._cache.hasOwnProperty(e)?Promise.resolve(this._cache[e]):Promise.reject("CachedResourceLoader: Did not find cached template for "+e)},t}(i.ResourceLoader),O=new o.Version("7.2.0"),w=[{provide:i.ResourceLoader,useClass:x,deps:[]}],C=Object(o.createPlatformFactory)(y,"browserDynamic",j)},"./node_modules/@angular/platform-browser-dynamic/node_modules/tslib/tslib.es6.js":function(e,t,n){"use strict";n.r(t),n.d(t,"__extends",function(){return i}),n.d(t,"__assign",function(){return o}),n.d(t,"__rest",function(){return a}),n.d(t,"__decorate",function(){return s}),n.d(t,"__param",function(){return u}),n.d(t,"__metadata",function(){return c}),n.d(t,"__awaiter",function(){return l}),n.d(t,"__generator",function(){return d}),n.d(t,"__exportStar",function(){return p}),n.d(t,"__values",function(){return f}),n.d(t,"__read",function(){return h}),n.d(t,"__spread",function(){return m}),n.d(t,"__await",function(){return g}),n.d(t,"__asyncGenerator",function(){return v}),n.d(t,"__asyncDelegator",function(){return _}),n.d(t,"__asyncValues",function(){return y}),n.d(t,"__makeTemplateObject",function(){return b}),n.d(t,"__importStar",function(){return j}),n.d(t,"__importDefault",function(){return x});var r=function(e,t){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,t){e.__proto__=t}||function(e,t){for(var n in t)t.hasOwnProperty(n)&&(e[n]=t[n])})(e,t)};function i(e,t){function n(){this.constructor=e}r(e,t),e.prototype=null===t?Object.create(t):(n.prototype=t.prototype,new n)}var o=function(){return(o=Object.assign||function(e){for(var t,n=1,r=arguments.length;n<r;n++){t=arguments[n];for(var i in t)Object.prototype.hasOwnProperty.call(t,i)&&(e[i]=t[i])}return e}).apply(this,arguments)};function a(e,t){var n={};for(var r in e)Object.prototype.hasOwnProperty.call(e,r)&&t.indexOf(r)<0&&(n[r]=e[r]);if(null!=e&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(e);i<r.length;i++)t.indexOf(r[i])<0&&(n[r[i]]=e[r[i]])}return n}function s(e,t,n,r){var i,o=arguments.length,a=o<3?t:null===r?r=Object.getOwnPropertyDescriptor(t,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(e,t,n,r);else for(var s=e.length-1;s>=0;s--)(i=e[s])&&(a=(o<3?i(a):o>3?i(t,n,a):i(t,n))||a);return o>3&&a&&Object.defineProperty(t,n,a),a}function u(e,t){return function(n,r){t(n,r,e)}}function c(e,t){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(e,t)}function l(e,t,n,r){return new(n||(n=Promise))(function(i,o){function a(e){try{u(r.next(e))}catch(e){o(e)}}function s(e){try{u(r.throw(e))}catch(e){o(e)}}function u(e){e.done?i(e.value):new n(function(t){t(e.value)}).then(a,s)}u((r=r.apply(e,t||[])).next())})}function d(e,t){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=t.call(e,a)}catch(e){o=[6,e],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(e,t){for(var n in e)t.hasOwnProperty(n)||(t[n]=e[n])}function f(e){var t="function"==typeof Symbol&&e[Symbol.iterator],n=0;return t?t.call(e):{next:function(){return e&&n>=e.length&&(e=void 0),{value:e&&e[n++],done:!e}}}}function h(e,t){var n="function"==typeof Symbol&&e[Symbol.iterator];if(!n)return e;var r,i,o=n.call(e),a=[];try{for(;(void 0===t||t-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(e){i={error:e}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var e=[],t=0;t<arguments.length;t++)e=e.concat(h(arguments[t]));return e}function g(e){return this instanceof g?(this.v=e,this):new g(e)}function v(e,t,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(e,t||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(e){i[e]&&(r[e]=function(t){return new Promise(function(n,r){o.push([e,t,n,r])>1||s(e,t)})})}function s(e,t){try{(n=i[e](t)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(e){l(o[0][3],e)}var n}function u(e){s("next",e)}function c(e){s("throw",e)}function l(e,t){e(t),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(e){var t,n;return t={},r("next"),r("throw",function(e){throw e}),r("return"),t[Symbol.iterator]=function(){return this},t;function r(r,i){t[r]=e[r]?function(t){return(n=!n)?{value:g(e[r](t)),done:"return"===r}:i?i(t):t}:i}}function y(e){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var t,n=e[Symbol.asyncIterator];return n?n.call(e):(e=f(e),t={},r("next"),r("throw"),r("return"),t[Symbol.asyncIterator]=function(){return this},t);function r(n){t[n]=e[n]&&function(t){return new Promise(function(r,i){var o,a,s,u;t=e[n](t),o=r,a=i,s=t.done,u=t.value,Promise.resolve(u).then(function(e){o({value:e,done:s})},a)})}}}function b(e,t){return Object.defineProperty?Object.defineProperty(e,"raw",{value:t}):e.raw=t,e}function j(e){if(e&&e.__esModule)return e;var t={};if(null!=e)for(var n in e)Object.hasOwnProperty.call(e,n)&&(t[n]=e[n]);return t.default=e,t}function x(e){return e&&e.__esModule?e:{default:e}}},"./node_modules/@angular/platform-browser/fesm5/animations.js":function(e,t,n){"use strict";n.r(t),n.d(t,"ɵangular_packages_platform_browser_animations_animations_f",function(){return h}),n.d(t,"ɵangular_packages_platform_browser_animations_animations_d",function(){return x}),n.d(t,"ɵangular_packages_platform_browser_animations_animations_e",function(){return O}),n.d(t,"ɵangular_packages_platform_browser_animations_animations_b",function(){return _}),n.d(t,"ɵangular_packages_platform_browser_animations_animations_c",function(){return y}),n.d(t,"ɵangular_packages_platform_browser_animations_animations_a",function(){return v}),n.d(t,"BrowserAnimationsModule",function(){return w}),n.d(t,"NoopAnimationsModule",function(){return C}),n.d(t,"ANIMATION_MODULE_TYPE",function(){return b}),n.d(t,"ɵBrowserAnimationBuilder",function(){return c}),n.d(t,"ɵBrowserAnimationFactory",function(){return l}),n.d(t,"ɵAnimationRenderer",function(){return m}),n.d(t,"ɵAnimationRendererFactory",function(){return f}),n.d(t,"ɵInjectableAnimationEngine",function(){return g});var r=n("./node_modules/@angular/platform-browser/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/animations/fesm5/animations.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=n("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),s=n("./node_modules/@angular/animations/fesm5/browser.js"),u=n("./node_modules/@angular/common/fesm5/common.js"),c=function(e){function t(t,n){var r=e.call(this)||this;r._nextAnimationId=0;var i={id:"0",encapsulation:o.ViewEncapsulation.None,styles:[],data:{animation:[]}};return r._renderer=t.createRenderer(n.body,i),r}return Object(r.__extends)(t,e),t.prototype.build=function(e){var t=this._nextAnimationId.toString();this._nextAnimationId++;var n=Array.isArray(e)?Object(i.sequence)(e):e;return p(this._renderer,null,t,"register",[n]),new l(t,this._renderer)},t=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(1,Object(o.Inject)(a.DOCUMENT)),Object(r.__metadata)("design:paramtypes",[o.RendererFactory2,Object])],t)}(i.AnimationBuilder),l=function(e){function t(t,n){var r=e.call(this)||this;return r._id=t,r._renderer=n,r}return Object(r.__extends)(t,e),t.prototype.create=function(e,t){return new d(this._id,e,t||{},this._renderer)},t}(i.AnimationFactory),d=function(){function e(e,t,n,r){this.id=e,this.element=t,this._renderer=r,this.parentPlayer=null,this._started=!1,this.totalTime=0,this._command("create",n)}return e.prototype._listen=function(e,t){return this._renderer.listen(this.element,"@@"+this.id+":"+e,t)},e.prototype._command=function(e){for(var t=[],n=1;n<arguments.length;n++)t[n-1]=arguments[n];return p(this._renderer,this.element,this.id,e,t)},e.prototype.onDone=function(e){this._listen("done",e)},e.prototype.onStart=function(e){this._listen("start",e)},e.prototype.onDestroy=function(e){this._listen("destroy",e)},e.prototype.init=function(){this._command("init")},e.prototype.hasStarted=function(){return this._started},e.prototype.play=function(){this._command("play"),this._started=!0},e.prototype.pause=function(){this._command("pause")},e.prototype.restart=function(){this._command("restart")},e.prototype.finish=function(){this._command("finish")},e.prototype.destroy=function(){this._command("destroy")},e.prototype.reset=function(){this._command("reset")},e.prototype.setPosition=function(e){this._command("setPosition",e)},e.prototype.getPosition=function(){return 0},e}();function p(e,t,n,r,i){return e.setProperty(t,"@@"+n+":"+r,i)}var f=function(){function e(e,t,n){this.delegate=e,this.engine=t,this._zone=n,this._currentId=0,this._microtaskId=1,this._animationCallbacksBuffer=[],this._rendererCache=new Map,this._cdRecurDepth=0,this.promise=Promise.resolve(0),t.onRemovalComplete=function(e,t){t&&t.parentNode(e)&&t.removeChild(e.parentNode,e)}}return e.prototype.createRenderer=function(e,t){var n=this,r=this.delegate.createRenderer(e,t);if(!(e&&t&&t.data&&t.data.animation)){var i=this._rendererCache.get(r);return i||(i=new h("",r,this.engine),this._rendererCache.set(r,i)),i}var o=t.id,a=t.id+"-"+this._currentId;return this._currentId++,this.engine.register(a,e),t.data.animation.forEach(function(t){return n.engine.registerTrigger(o,a,e,t.name,t)}),new m(this,a,r,this.engine)},e.prototype.begin=function(){this._cdRecurDepth++,this.delegate.begin&&this.delegate.begin()},e.prototype._scheduleCountTask=function(){var e=this;this.promise.then(function(){e._microtaskId++})},e.prototype.scheduleListenerCallback=function(e,t,n){var i=this;e>=0&&e<this._microtaskId?this._zone.run(function(){return t(n)}):(0==this._animationCallbacksBuffer.length&&Promise.resolve(null).then(function(){i._zone.run(function(){i._animationCallbacksBuffer.forEach(function(e){var t=Object(r.__read)(e,2);(0,t[0])(t[1])}),i._animationCallbacksBuffer=[]})}),this._animationCallbacksBuffer.push([t,n]))},e.prototype.end=function(){var e=this;this._cdRecurDepth--,0==this._cdRecurDepth&&this._zone.runOutsideAngular(function(){e._scheduleCountTask(),e.engine.flush(e._microtaskId)}),this.delegate.end&&this.delegate.end()},e.prototype.whenRenderingDone=function(){return this.engine.whenRenderingDone()},e=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__metadata)("design:paramtypes",[o.RendererFactory2,s["ɵAnimationEngine"],o.NgZone])],e)}(),h=function(){function e(e,t,n){this.namespaceId=e,this.delegate=t,this.engine=n,this.destroyNode=this.delegate.destroyNode?function(e){return t.destroyNode(e)}:null}return Object.defineProperty(e.prototype,"data",{get:function(){return this.delegate.data},enumerable:!0,configurable:!0}),e.prototype.destroy=function(){this.engine.destroy(this.namespaceId,this.delegate),this.delegate.destroy()},e.prototype.createElement=function(e,t){return this.delegate.createElement(e,t)},e.prototype.createComment=function(e){return this.delegate.createComment(e)},e.prototype.createText=function(e){return this.delegate.createText(e)},e.prototype.appendChild=function(e,t){this.delegate.appendChild(e,t),this.engine.onInsert(this.namespaceId,t,e,!1)},e.prototype.insertBefore=function(e,t,n){this.delegate.insertBefore(e,t,n),this.engine.onInsert(this.namespaceId,t,e,!0)},e.prototype.removeChild=function(e,t){this.engine.onRemove(this.namespaceId,t,this.delegate)},e.prototype.selectRootElement=function(e,t){return this.delegate.selectRootElement(e,t)},e.prototype.parentNode=function(e){return this.delegate.parentNode(e)},e.prototype.nextSibling=function(e){return this.delegate.nextSibling(e)},e.prototype.setAttribute=function(e,t,n,r){this.delegate.setAttribute(e,t,n,r)},e.prototype.removeAttribute=function(e,t,n){this.delegate.removeAttribute(e,t,n)},e.prototype.addClass=function(e,t){this.delegate.addClass(e,t)},e.prototype.removeClass=function(e,t){this.delegate.removeClass(e,t)},e.prototype.setStyle=function(e,t,n,r){this.delegate.setStyle(e,t,n,r)},e.prototype.removeStyle=function(e,t,n){this.delegate.removeStyle(e,t,n)},e.prototype.setProperty=function(e,t,n){"@"==t.charAt(0)&&"@.disabled"==t?this.disableAnimations(e,!!n):this.delegate.setProperty(e,t,n)},e.prototype.setValue=function(e,t){this.delegate.setValue(e,t)},e.prototype.listen=function(e,t,n){return this.delegate.listen(e,t,n)},e.prototype.disableAnimations=function(e,t){this.engine.disableAnimations(e,t)},e}(),m=function(e){function t(t,n,r,i){var o=e.call(this,n,r,i)||this;return o.factory=t,o.namespaceId=n,o}return Object(r.__extends)(t,e),t.prototype.setProperty=function(e,t,n){"@"==t.charAt(0)?"."==t.charAt(1)&&"@.disabled"==t?(n=void 0===n||!!n,this.disableAnimations(e,n)):this.engine.process(this.namespaceId,e,t.substr(1),n):this.delegate.setProperty(e,t,n)},t.prototype.listen=function(e,t,n){var i,o,a,s,u,c=this;if("@"==t.charAt(0)){var l=function(e){switch(e){case"body":return document.body;case"document":return document;case"window":return window;default:return e}}(e),d=t.substr(1),p="";return"@"!=d.charAt(0)&&(d=(i=Object(r.__read)((a=(o=d).indexOf("."),s=o.substring(0,a),u=o.substr(a+1),[s,u]),2))[0],p=i[1]),this.engine.listen(this.namespaceId,l,d,p,function(e){var t=e._data||-1;c.factory.scheduleListenerCallback(t,n,e)})}return this.delegate.listen(e,t,n)},t}(h);var g=function(e){function t(t,n,r){return e.call(this,t.body,n,r)||this}return Object(r.__extends)(t,e),t=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(0,Object(o.Inject)(u.DOCUMENT)),Object(r.__metadata)("design:paramtypes",[Object,s.AnimationDriver,s["ɵAnimationStyleNormalizer"]])],t)}(s["ɵAnimationEngine"]);function v(){return Object(s["ɵsupportsWebAnimations"])()?new s["ɵWebAnimationsDriver"]:new s["ɵCssKeyframesDriver"]}function _(){return new s["ɵWebAnimationsStyleNormalizer"]}function y(e,t,n){return new f(e,t,n)}var b=new o.InjectionToken("AnimationModuleType"),j=[{provide:i.AnimationBuilder,useClass:c},{provide:s["ɵAnimationStyleNormalizer"],useFactory:_},{provide:s["ɵAnimationEngine"],useClass:g},{provide:o.RendererFactory2,useFactory:y,deps:[a["ɵDomRendererFactory2"],s["ɵAnimationEngine"],o.NgZone]}],x=Object(r.__spread)([{provide:s.AnimationDriver,useFactory:v},{provide:b,useValue:"BrowserAnimations"}],j),O=Object(r.__spread)([{provide:s.AnimationDriver,useClass:s["ɵNoopAnimationDriver"]},{provide:b,useValue:"NoopAnimations"}],j),w=function(){function e(){}return e=Object(r.__decorate)([Object(o.NgModule)({exports:[a.BrowserModule],providers:x})],e)}(),C=function(){function e(){}return e=Object(r.__decorate)([Object(o.NgModule)({exports:[a.BrowserModule],providers:O})],e)}()},"./node_modules/@angular/platform-browser/fesm5/platform-browser.js":function(e,t,n){"use strict";n.r(t),n.d(t,"ɵangular_packages_platform_browser_platform_browser_c",function(){return Te}),n.d(t,"ɵangular_packages_platform_browser_platform_browser_b",function(){return Ae}),n.d(t,"ɵangular_packages_platform_browser_platform_browser_a",function(){return Me}),n.d(t,"ɵangular_packages_platform_browser_platform_browser_k",function(){return d}),n.d(t,"ɵangular_packages_platform_browser_platform_browser_d",function(){return Ie}),n.d(t,"ɵangular_packages_platform_browser_platform_browser_i",function(){return O}),n.d(t,"ɵangular_packages_platform_browser_platform_browser_h",function(){return x}),n.d(t,"ɵangular_packages_platform_browser_platform_browser_e",function(){return Re}),n.d(t,"ɵangular_packages_platform_browser_platform_browser_f",function(){return We}),n.d(t,"ɵangular_packages_platform_browser_platform_browser_j",function(){return A}),n.d(t,"ɵangular_packages_platform_browser_platform_browser_g",function(){return P}),n.d(t,"BrowserModule",function(){return De}),n.d(t,"platformBrowser",function(){return Se}),n.d(t,"Meta",function(){return Pe}),n.d(t,"Title",function(){return Ne}),n.d(t,"disableDebugTools",function(){return Ue}),n.d(t,"enableDebugTools",function(){return ze}),n.d(t,"BrowserTransferStateModule",function(){return Ye}),n.d(t,"TransferState",function(){return Ge}),n.d(t,"makeStateKey",function(){return qe}),n.d(t,"By",function(){return Ke}),n.d(t,"DOCUMENT",function(){return _}),n.d(t,"EVENT_MANAGER_PLUGINS",function(){return D}),n.d(t,"EventManager",function(){return I}),n.d(t,"HAMMER_GESTURE_CONFIG",function(){return le}),n.d(t,"HAMMER_LOADER",function(){return de}),n.d(t,"HammerGestureConfig",function(){return pe}),n.d(t,"DomSanitizer",function(){return ve}),n.d(t,"VERSION",function(){return Qe}),n.d(t,"ɵBROWSER_SANITIZATION_PROVIDERS",function(){return Ee}),n.d(t,"ɵINTERNAL_BROWSER_PLATFORM_PROVIDERS",function(){return Ce}),n.d(t,"ɵinitDomAdapter",function(){return ke}),n.d(t,"ɵBrowserDomAdapter",function(){return g}),n.d(t,"ɵBrowserPlatformLocation",function(){return b}),n.d(t,"ɵTRANSITION_ID",function(){return j}),n.d(t,"ɵBrowserGetTestability",function(){return w}),n.d(t,"ɵescapeHtml",function(){return He}),n.d(t,"ɵELEMENT_PROBE_PROVIDERS",function(){return T}),n.d(t,"ɵDomAdapter",function(){return l}),n.d(t,"ɵgetDOM",function(){return s}),n.d(t,"ɵsetRootDomAdapter",function(){return u}),n.d(t,"ɵDomRendererFactory2",function(){return G}),n.d(t,"ɵNAMESPACE_URIS",function(){return F}),n.d(t,"ɵflattenStyles",function(){return H}),n.d(t,"ɵshimContentAttribute",function(){return z}),n.d(t,"ɵshimHostAttribute",function(){return U}),n.d(t,"ɵDomEventsPlugin",function(){return ue}),n.d(t,"ɵHammerGesturesPlugin",function(){return fe}),n.d(t,"ɵKeyEventsPlugin",function(){return ge}),n.d(t,"ɵDomSharedStylesHost",function(){return N}),n.d(t,"ɵSharedStylesHost",function(){return R}),n.d(t,"ɵDomSanitizerImpl",function(){return _e});var r=n("./node_modules/@angular/platform-browser/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/@angular/common/fesm5/common.js"),o=n("./node_modules/@angular/core/fesm5/core.js"),a=null;function s(){return a}function u(e){a||(a=e)}var c,l=function(){function e(){this.resourceLoaderType=null}return Object.defineProperty(e.prototype,"attrToPropMap",{get:function(){return this._attrToPropMap},set:function(e){this._attrToPropMap=e},enumerable:!0,configurable:!0}),e}(),d=function(e){function t(){var t=e.call(this)||this;t._animationPrefix=null,t._transitionEnd=null;try{var n=t.createElement("div",document);if(null!=t.getStyle(n,"animationName"))t._animationPrefix="";else for(var r=["Webkit","Moz","O","ms"],i=0;i<r.length;i++)if(null!=t.getStyle(n,r[i]+"AnimationName")){t._animationPrefix="-"+r[i].toLowerCase()+"-";break}var o={WebkitTransition:"webkitTransitionEnd",MozTransition:"transitionend",OTransition:"oTransitionEnd otransitionend",transition:"transitionend"};Object.keys(o).forEach(function(e){null!=t.getStyle(n,e)&&(t._transitionEnd=o[e])})}catch(e){t._animationPrefix=null,t._transitionEnd=null}return t}return Object(r.__extends)(t,e),t.prototype.getDistributedNodes=function(e){return e.getDistributedNodes()},t.prototype.resolveAndSetHref=function(e,t,n){e.href=null==n?t:t+"/../"+n},t.prototype.supportsDOMEvents=function(){return!0},t.prototype.supportsNativeShadowDOM=function(){return"function"==typeof document.body.createShadowRoot},t.prototype.getAnimationPrefix=function(){return this._animationPrefix?this._animationPrefix:""},t.prototype.getTransitionEnd=function(){return this._transitionEnd?this._transitionEnd:""},t.prototype.supportsAnimation=function(){return null!=this._animationPrefix&&null!=this._transitionEnd},t}(l),p={class:"className",innerHtml:"innerHTML",readonly:"readOnly",tabindex:"tabIndex"},f={"\b":"Backspace","\t":"Tab","":"Delete","":"Escape",Del:"Delete",Esc:"Escape",Left:"ArrowLeft",Right:"ArrowRight",Up:"ArrowUp",Down:"ArrowDown",Menu:"ContextMenu",Scroll:"ScrollLock",Win:"OS"},h={A:"1",B:"2",C:"3",D:"4",E:"5",F:"6",G:"7",H:"8",I:"9",J:"*",K:"+",M:"-",N:".",O:"/","`":"0","":"NumLock"};o["ɵglobal"].Node&&(c=o["ɵglobal"].Node.prototype.contains||function(e){return!!(16&this.compareDocumentPosition(e))});var m,g=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.parse=function(e){throw new Error("parse not implemented")},t.makeCurrent=function(){u(new t)},t.prototype.hasProperty=function(e,t){return t in e},t.prototype.setProperty=function(e,t,n){e[t]=n},t.prototype.getProperty=function(e,t){return e[t]},t.prototype.invoke=function(e,t,n){var i;(i=e)[t].apply(i,Object(r.__spread)(n))},t.prototype.logError=function(e){window.console&&(console.error?console.error(e):console.log(e))},t.prototype.log=function(e){window.console&&window.console.log&&window.console.log(e)},t.prototype.logGroup=function(e){window.console&&window.console.group&&window.console.group(e)},t.prototype.logGroupEnd=function(){window.console&&window.console.groupEnd&&window.console.groupEnd()},Object.defineProperty(t.prototype,"attrToPropMap",{get:function(){return p},enumerable:!0,configurable:!0}),t.prototype.contains=function(e,t){return c.call(e,t)},t.prototype.querySelector=function(e,t){return e.querySelector(t)},t.prototype.querySelectorAll=function(e,t){return e.querySelectorAll(t)},t.prototype.on=function(e,t,n){e.addEventListener(t,n,!1)},t.prototype.onAndCancel=function(e,t,n){return e.addEventListener(t,n,!1),function(){e.removeEventListener(t,n,!1)}},t.prototype.dispatchEvent=function(e,t){e.dispatchEvent(t)},t.prototype.createMouseEvent=function(e){var t=this.getDefaultDocument().createEvent("MouseEvent");return t.initEvent(e,!0,!0),t},t.prototype.createEvent=function(e){var t=this.getDefaultDocument().createEvent("Event");return t.initEvent(e,!0,!0),t},t.prototype.preventDefault=function(e){e.preventDefault(),e.returnValue=!1},t.prototype.isPrevented=function(e){return e.defaultPrevented||null!=e.returnValue&&!e.returnValue},t.prototype.getInnerHTML=function(e){return e.innerHTML},t.prototype.getTemplateContent=function(e){return"content"in e&&this.isTemplateElement(e)?e.content:null},t.prototype.getOuterHTML=function(e){return e.outerHTML},t.prototype.nodeName=function(e){return e.nodeName},t.prototype.nodeValue=function(e){return e.nodeValue},t.prototype.type=function(e){return e.type},t.prototype.content=function(e){return this.hasProperty(e,"content")?e.content:e},t.prototype.firstChild=function(e){return e.firstChild},t.prototype.nextSibling=function(e){return e.nextSibling},t.prototype.parentElement=function(e){return e.parentNode},t.prototype.childNodes=function(e){return e.childNodes},t.prototype.childNodesAsList=function(e){for(var t=e.childNodes,n=new Array(t.length),r=0;r<t.length;r++)n[r]=t[r];return n},t.prototype.clearNodes=function(e){for(;e.firstChild;)e.removeChild(e.firstChild)},t.prototype.appendChild=function(e,t){e.appendChild(t)},t.prototype.removeChild=function(e,t){e.removeChild(t)},t.prototype.replaceChild=function(e,t,n){e.replaceChild(t,n)},t.prototype.remove=function(e){return e.parentNode&&e.parentNode.removeChild(e),e},t.prototype.insertBefore=function(e,t,n){e.insertBefore(n,t)},t.prototype.insertAllBefore=function(e,t,n){n.forEach(function(n){return e.insertBefore(n,t)})},t.prototype.insertAfter=function(e,t,n){e.insertBefore(n,t.nextSibling)},t.prototype.setInnerHTML=function(e,t){e.innerHTML=t},t.prototype.getText=function(e){return e.textContent},t.prototype.setText=function(e,t){e.textContent=t},t.prototype.getValue=function(e){return e.value},t.prototype.setValue=function(e,t){e.value=t},t.prototype.getChecked=function(e){return e.checked},t.prototype.setChecked=function(e,t){e.checked=t},t.prototype.createComment=function(e){return this.getDefaultDocument().createComment(e)},t.prototype.createTemplate=function(e){var t=this.getDefaultDocument().createElement("template");return t.innerHTML=e,t},t.prototype.createElement=function(e,t){return(t=t||this.getDefaultDocument()).createElement(e)},t.prototype.createElementNS=function(e,t,n){return(n=n||this.getDefaultDocument()).createElementNS(e,t)},t.prototype.createTextNode=function(e,t){return(t=t||this.getDefaultDocument()).createTextNode(e)},t.prototype.createScriptTag=function(e,t,n){var r=(n=n||this.getDefaultDocument()).createElement("SCRIPT");return r.setAttribute(e,t),r},t.prototype.createStyleElement=function(e,t){var n=(t=t||this.getDefaultDocument()).createElement("style");return this.appendChild(n,this.createTextNode(e,t)),n},t.prototype.createShadowRoot=function(e){return e.createShadowRoot()},t.prototype.getShadowRoot=function(e){return e.shadowRoot},t.prototype.getHost=function(e){return e.host},t.prototype.clone=function(e){return e.cloneNode(!0)},t.prototype.getElementsByClassName=function(e,t){return e.getElementsByClassName(t)},t.prototype.getElementsByTagName=function(e,t){return e.getElementsByTagName(t)},t.prototype.classList=function(e){return Array.prototype.slice.call(e.classList,0)},t.prototype.addClass=function(e,t){e.classList.add(t)},t.prototype.removeClass=function(e,t){e.classList.remove(t)},t.prototype.hasClass=function(e,t){return e.classList.contains(t)},t.prototype.setStyle=function(e,t,n){e.style[t]=n},t.prototype.removeStyle=function(e,t){e.style[t]=""},t.prototype.getStyle=function(e,t){return e.style[t]},t.prototype.hasStyle=function(e,t,n){var r=this.getStyle(e,t)||"";return n?r==n:r.length>0},t.prototype.tagName=function(e){return e.tagName},t.prototype.attributeMap=function(e){for(var t=new Map,n=e.attributes,r=0;r<n.length;r++){var i=n.item(r);t.set(i.name,i.value)}return t},t.prototype.hasAttribute=function(e,t){return e.hasAttribute(t)},t.prototype.hasAttributeNS=function(e,t,n){return e.hasAttributeNS(t,n)},t.prototype.getAttribute=function(e,t){return e.getAttribute(t)},t.prototype.getAttributeNS=function(e,t,n){return e.getAttributeNS(t,n)},t.prototype.setAttribute=function(e,t,n){e.setAttribute(t,n)},t.prototype.setAttributeNS=function(e,t,n,r){e.setAttributeNS(t,n,r)},t.prototype.removeAttribute=function(e,t){e.removeAttribute(t)},t.prototype.removeAttributeNS=function(e,t,n){e.removeAttributeNS(t,n)},t.prototype.templateAwareRoot=function(e){return this.isTemplateElement(e)?this.content(e):e},t.prototype.createHtmlDocument=function(){return document.implementation.createHTMLDocument("fakeTitle")},t.prototype.getDefaultDocument=function(){return document},t.prototype.getBoundingClientRect=function(e){try{return e.getBoundingClientRect()}catch(e){return{top:0,bottom:0,left:0,right:0,width:0,height:0}}},t.prototype.getTitle=function(e){return e.title},t.prototype.setTitle=function(e,t){e.title=t||""},t.prototype.elementMatches=function(e,t){return!!this.isElementNode(e)&&(e.matches&&e.matches(t)||e.msMatchesSelector&&e.msMatchesSelector(t)||e.webkitMatchesSelector&&e.webkitMatchesSelector(t))},t.prototype.isTemplateElement=function(e){return this.isElementNode(e)&&"TEMPLATE"===e.nodeName},t.prototype.isTextNode=function(e){return e.nodeType===Node.TEXT_NODE},t.prototype.isCommentNode=function(e){return e.nodeType===Node.COMMENT_NODE},t.prototype.isElementNode=function(e){return e.nodeType===Node.ELEMENT_NODE},t.prototype.hasShadowRoot=function(e){return null!=e.shadowRoot&&e instanceof HTMLElement},t.prototype.isShadowRoot=function(e){return e instanceof DocumentFragment},t.prototype.importIntoDoc=function(e){return document.importNode(this.templateAwareRoot(e),!0)},t.prototype.adoptNode=function(e){return document.adoptNode(e)},t.prototype.getHref=function(e){return e.getAttribute("href")},t.prototype.getEventKey=function(e){var t=e.key;if(null==t){if(null==(t=e.keyIdentifier))return"Unidentified";t.startsWith("U+")&&(t=String.fromCharCode(parseInt(t.substring(2),16)),3===e.location&&h.hasOwnProperty(t)&&(t=h[t]))}return f[t]||t},t.prototype.getGlobalEventTarget=function(e,t){return"window"===t?window:"document"===t?e:"body"===t?e.body:null},t.prototype.getHistory=function(){return window.history},t.prototype.getLocation=function(){return window.location},t.prototype.getBaseHref=function(e){var t=function(){if(!v&&!(v=document.querySelector("base")))return null;return v.getAttribute("href")}();return null==t?null:function(e){m||(m=document.createElement("a"));return m.setAttribute("href",e),"/"===m.pathname.charAt(0)?m.pathname:"/"+m.pathname}(t)},t.prototype.resetBaseElement=function(){v=null},t.prototype.getUserAgent=function(){return window.navigator.userAgent},t.prototype.setData=function(e,t,n){this.setAttribute(e,"data-"+t,n)},t.prototype.getData=function(e,t){return this.getAttribute(e,"data-"+t)},t.prototype.getComputedStyle=function(e){return getComputedStyle(e)},t.prototype.supportsWebAnimation=function(){return"function"==typeof Element.prototype.animate},t.prototype.performanceNow=function(){return window.performance&&window.performance.now?window.performance.now():(new Date).getTime()},t.prototype.supportsCookies=function(){return!0},t.prototype.getCookie=function(e){return Object(i["ɵparseCookieValue"])(document.cookie,e)},t.prototype.setCookie=function(e,t){document.cookie=encodeURIComponent(e)+"="+encodeURIComponent(t)},t}(d),v=null;var _=i.DOCUMENT;function y(){return!!window.history.pushState}var b=function(e){function t(t){var n=e.call(this)||this;return n._doc=t,n._init(),n}return Object(r.__extends)(t,e),t.prototype._init=function(){this.location=s().getLocation(),this._history=s().getHistory()},t.prototype.getBaseHrefFromDOM=function(){return s().getBaseHref(this._doc)},t.prototype.onPopState=function(e){s().getGlobalEventTarget(this._doc,"window").addEventListener("popstate",e,!1)},t.prototype.onHashChange=function(e){s().getGlobalEventTarget(this._doc,"window").addEventListener("hashchange",e,!1)},Object.defineProperty(t.prototype,"pathname",{get:function(){return this.location.pathname},set:function(e){this.location.pathname=e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"search",{get:function(){return this.location.search},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"hash",{get:function(){return this.location.hash},enumerable:!0,configurable:!0}),t.prototype.pushState=function(e,t,n){y()?this._history.pushState(e,t,n):this.location.hash=n},t.prototype.replaceState=function(e,t,n){y()?this._history.replaceState(e,t,n):this.location.hash=n},t.prototype.forward=function(){this._history.forward()},t.prototype.back=function(){this._history.back()},t=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(0,Object(o.Inject)(_)),Object(r.__metadata)("design:paramtypes",[Object])],t)}(i.PlatformLocation),j=new o.InjectionToken("TRANSITION_ID");function x(e,t,n){return function(){n.get(o.ApplicationInitStatus).donePromise.then(function(){var n=s();Array.prototype.slice.apply(n.querySelectorAll(t,"style[ng-transition]")).filter(function(t){return n.getAttribute(t,"ng-transition")===e}).forEach(function(e){return n.remove(e)})})}}var O=[{provide:o.APP_INITIALIZER,useFactory:x,deps:[j,_,o.Injector],multi:!0}],w=function(){function e(){}return e.init=function(){Object(o.setTestabilityGetter)(new e)},e.prototype.addToWindow=function(e){o["ɵglobal"].getAngularTestability=function(t,n){void 0===n&&(n=!0);var r=e.findTestabilityInTree(t,n);if(null==r)throw new Error("Could not find testability for element.");return r},o["ɵglobal"].getAllAngularTestabilities=function(){return e.getAllTestabilities()},o["ɵglobal"].getAllAngularRootElements=function(){return e.getAllRootElements()};o["ɵglobal"].frameworkStabilizers||(o["ɵglobal"].frameworkStabilizers=[]),o["ɵglobal"].frameworkStabilizers.push(function(e){var t=o["ɵglobal"].getAllAngularTestabilities(),n=t.length,r=!1,i=function(t){r=r||t,0==--n&&e(r)};t.forEach(function(e){e.whenStable(i)})})},e.prototype.findTestabilityInTree=function(e,t,n){if(null==t)return null;var r=e.getTestability(t);return null!=r?r:n?s().isShadowRoot(t)?this.findTestabilityInTree(e,s().getHost(t),!0):this.findTestabilityInTree(e,s().parentElement(t),!0):null},e}();function C(e,t){"undefined"!=typeof COMPILED&&COMPILED||((o["ɵglobal"].ng=o["ɵglobal"].ng||{})[e]=t)}var E={ApplicationRef:o.ApplicationRef,NgZone:o.NgZone},S="probe",k="coreTokens";function M(e){return Object(o.getDebugNode)(e)}function A(e){return C(S,M),C(k,Object(r.__assign)({},E,(e||[]).reduce(function(e,t){return e[t.name]=t.token,e},{}))),function(){return M}}var T=[{provide:o.APP_INITIALIZER,useFactory:A,deps:[[o.NgProbeToken,new o.Optional]],multi:!0}],D=new o.InjectionToken("EventManagerPlugins"),I=function(){function e(e,t){var n=this;this._zone=t,this._eventNameToPlugin=new Map,e.forEach(function(e){return e.manager=n}),this._plugins=e.slice().reverse()}return e.prototype.addEventListener=function(e,t,n){return this._findPluginFor(t).addEventListener(e,t,n)},e.prototype.addGlobalEventListener=function(e,t,n){return this._findPluginFor(t).addGlobalEventListener(e,t,n)},e.prototype.getZone=function(){return this._zone},e.prototype._findPluginFor=function(e){var t=this._eventNameToPlugin.get(e);if(t)return t;for(var n=this._plugins,r=0;r<n.length;r++){var i=n[r];if(i.supports(e))return this._eventNameToPlugin.set(e,i),i}throw new Error("No event manager plugin found for event "+e)},e=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(0,Object(o.Inject)(D)),Object(r.__metadata)("design:paramtypes",[Array,o.NgZone])],e)}(),P=function(){function e(e){this._doc=e}return e.prototype.addGlobalEventListener=function(e,t,n){var r=s().getGlobalEventTarget(this._doc,e);if(!r)throw new Error("Unsupported event target "+r+" for event "+t);return this.addEventListener(r,t,n)},e}(),R=function(){function e(){this._stylesSet=new Set}return e.prototype.addStyles=function(e){var t=this,n=new Set;e.forEach(function(e){t._stylesSet.has(e)||(t._stylesSet.add(e),n.add(e))}),this.onStylesAdded(n)},e.prototype.onStylesAdded=function(e){},e.prototype.getAllStyles=function(){return Array.from(this._stylesSet)},e=Object(r.__decorate)([Object(o.Injectable)()],e)}(),N=function(e){function t(t){var n=e.call(this)||this;return n._doc=t,n._hostNodes=new Set,n._styleNodes=new Set,n._hostNodes.add(t.head),n}return Object(r.__extends)(t,e),t.prototype._addStylesToHost=function(e,t){var n=this;e.forEach(function(e){var r=n._doc.createElement("style");r.textContent=e,n._styleNodes.add(t.appendChild(r))})},t.prototype.addHost=function(e){this._addStylesToHost(this._stylesSet,e),this._hostNodes.add(e)},t.prototype.removeHost=function(e){this._hostNodes.delete(e)},t.prototype.onStylesAdded=function(e){var t=this;this._hostNodes.forEach(function(n){return t._addStylesToHost(e,n)})},t.prototype.ngOnDestroy=function(){this._styleNodes.forEach(function(e){return s().remove(e)})},t=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(0,Object(o.Inject)(_)),Object(r.__metadata)("design:paramtypes",[Object])],t)}(R),F={svg:"http://www.w3.org/2000/svg",xhtml:"http://www.w3.org/1999/xhtml",xlink:"http://www.w3.org/1999/xlink",xml:"http://www.w3.org/XML/1998/namespace",xmlns:"http://www.w3.org/2000/xmlns/"},L=/%COMP%/g,V="_nghost-%COMP%",B="_ngcontent-%COMP%";function z(e){return B.replace(L,e)}function U(e){return V.replace(L,e)}function H(e,t,n){for(var r=0;r<t.length;r++){var i=t[r];Array.isArray(i)?H(e,i,n):(i=i.replace(L,e),n.push(i))}return n}function q(e){return function(t){!1===e(t)&&(t.preventDefault(),t.returnValue=!1)}}var G=function(){function e(e,t){this.eventManager=e,this.sharedStylesHost=t,this.rendererByCompId=new Map,this.defaultRenderer=new W(e)}return e.prototype.createRenderer=function(e,t){if(!e||!t)return this.defaultRenderer;switch(t.encapsulation){case o.ViewEncapsulation.Emulated:var n=this.rendererByCompId.get(t.id);return n||(n=new X(this.eventManager,this.sharedStylesHost,t),this.rendererByCompId.set(t.id,n)),n.applyToHost(e),n;case o.ViewEncapsulation.Native:case o.ViewEncapsulation.ShadowDom:return new Z(this.eventManager,this.sharedStylesHost,e,t);default:if(!this.rendererByCompId.has(t.id)){var r=H(t.id,t.styles,[]);this.sharedStylesHost.addStyles(r),this.rendererByCompId.set(t.id,this.defaultRenderer)}return this.defaultRenderer}},e.prototype.begin=function(){},e.prototype.end=function(){},e=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__metadata)("design:paramtypes",[I,N])],e)}(),W=function(){function e(e){this.eventManager=e,this.data=Object.create(null)}return e.prototype.destroy=function(){},e.prototype.createElement=function(e,t){return t?document.createElementNS(F[t],e):document.createElement(e)},e.prototype.createComment=function(e){return document.createComment(e)},e.prototype.createText=function(e){return document.createTextNode(e)},e.prototype.appendChild=function(e,t){e.appendChild(t)},e.prototype.insertBefore=function(e,t,n){e&&e.insertBefore(t,n)},e.prototype.removeChild=function(e,t){e&&e.removeChild(t)},e.prototype.selectRootElement=function(e,t){var n="string"==typeof e?document.querySelector(e):e;if(!n)throw new Error('The selector "'+e+'" did not match any elements');return t||(n.textContent=""),n},e.prototype.parentNode=function(e){return e.parentNode},e.prototype.nextSibling=function(e){return e.nextSibling},e.prototype.setAttribute=function(e,t,n,r){if(r){t=r+":"+t;var i=F[r];i?e.setAttributeNS(i,t,n):e.setAttribute(t,n)}else e.setAttribute(t,n)},e.prototype.removeAttribute=function(e,t,n){if(n){var r=F[n];r?e.removeAttributeNS(r,t):e.removeAttribute(n+":"+t)}else e.removeAttribute(t)},e.prototype.addClass=function(e,t){e.classList.add(t)},e.prototype.removeClass=function(e,t){e.classList.remove(t)},e.prototype.setStyle=function(e,t,n,r){r&o.RendererStyleFlags2.DashCase?e.style.setProperty(t,n,r&o.RendererStyleFlags2.Important?"important":""):e.style[t]=n},e.prototype.removeStyle=function(e,t,n){n&o.RendererStyleFlags2.DashCase?e.style.removeProperty(t):e.style[t]=""},e.prototype.setProperty=function(e,t,n){K(t,"property"),e[t]=n},e.prototype.setValue=function(e,t){e.nodeValue=t},e.prototype.listen=function(e,t,n){return K(t,"listener"),"string"==typeof e?this.eventManager.addGlobalEventListener(e,t,q(n)):this.eventManager.addEventListener(e,t,q(n))},e}(),Y="@".charCodeAt(0);function K(e,t){if(e.charCodeAt(0)===Y)throw new Error("Found the synthetic "+t+" "+e+'. Please include either "BrowserAnimationsModule" or "NoopAnimationsModule" in your application.')}var Q,X=function(e){function t(t,n,r){var i=e.call(this,t)||this;i.component=r;var o=H(r.id,r.styles,[]);return n.addStyles(o),i.contentAttr=z(r.id),i.hostAttr=U(r.id),i}return Object(r.__extends)(t,e),t.prototype.applyToHost=function(t){e.prototype.setAttribute.call(this,t,this.hostAttr,"")},t.prototype.createElement=function(t,n){var r=e.prototype.createElement.call(this,t,n);return e.prototype.setAttribute.call(this,r,this.contentAttr,""),r},t}(W),Z=function(e){function t(t,n,r,i){var a=e.call(this,t)||this;a.sharedStylesHost=n,a.hostEl=r,a.component=i,i.encapsulation===o.ViewEncapsulation.ShadowDom?a.shadowRoot=r.attachShadow({mode:"open"}):a.shadowRoot=r.createShadowRoot(),a.sharedStylesHost.addHost(a.shadowRoot);for(var s=H(i.id,i.styles,[]),u=0;u<s.length;u++){var c=document.createElement("style");c.textContent=s[u],a.shadowRoot.appendChild(c)}return a}return Object(r.__extends)(t,e),t.prototype.nodeOrShadowRoot=function(e){return e===this.hostEl?this.shadowRoot:e},t.prototype.destroy=function(){this.sharedStylesHost.removeHost(this.shadowRoot)},t.prototype.appendChild=function(t,n){return e.prototype.appendChild.call(this,this.nodeOrShadowRoot(t),n)},t.prototype.insertBefore=function(t,n,r){return e.prototype.insertBefore.call(this,this.nodeOrShadowRoot(t),n,r)},t.prototype.removeChild=function(t,n){return e.prototype.removeChild.call(this,this.nodeOrShadowRoot(t),n)},t.prototype.parentNode=function(t){return this.nodeOrShadowRoot(e.prototype.parentNode.call(this,this.nodeOrShadowRoot(t)))},t}(W),$="undefined"!=typeof Zone&&Zone.__symbol__||function(e){return"__zone_symbol__"+e},J=$("addEventListener"),ee=$("removeEventListener"),te={},ne="removeEventListener",re="__zone_symbol__propagationStopped",ie="__zone_symbol__stopImmediatePropagation",oe="undefined"!=typeof Zone&&Zone[$("BLACK_LISTED_EVENTS")];oe&&(Q={},oe.forEach(function(e){Q[e]=e}));var ae=function(e){return!!Q&&Q.hasOwnProperty(e)},se=function(e){var t=te[e.type];if(t){var n=this[t];if(n){var r=[e];if(1===n.length)return(a=n[0]).zone!==Zone.current?a.zone.run(a.handler,this,r):a.handler.apply(this,r);for(var i=n.slice(),o=0;o<i.length&&!0!==e[re];o++){var a;(a=i[o]).zone!==Zone.current?a.zone.run(a.handler,this,r):a.handler.apply(this,r)}}}},ue=function(e){function t(t,n,r){var o=e.call(this,t)||this;return o.ngZone=n,r&&Object(i.isPlatformServer)(r)||o.patchEvent(),o}return Object(r.__extends)(t,e),t.prototype.patchEvent=function(){if("undefined"!=typeof Event&&Event&&Event.prototype&&!Event.prototype[ie]){var e=Event.prototype[ie]=Event.prototype.stopImmediatePropagation;Event.prototype.stopImmediatePropagation=function(){this&&(this[re]=!0),e&&e.apply(this,arguments)}}},t.prototype.supports=function(e){return!0},t.prototype.addEventListener=function(e,t,n){var r=this,i=n;if(!e[J]||o.NgZone.isInAngularZone()&&!ae(t))e.addEventListener(t,i,!1);else{var a=te[t];a||(a=te[t]=$("ANGULAR"+t+"FALSE"));var s=e[a],u=s&&s.length>0;s||(s=e[a]=[]);var c=ae(t)?Zone.root:Zone.current;if(0===s.length)s.push({zone:c,handler:i});else{for(var l=!1,d=0;d<s.length;d++)if(s[d].handler===i){l=!0;break}l||s.push({zone:c,handler:i})}u||e[J](t,se,!1)}return function(){return r.removeEventListener(e,t,i)}},t.prototype.removeEventListener=function(e,t,n){var r=e[ee];if(!r)return e[ne].apply(e,[t,n,!1]);var i=te[t],o=i&&e[i];if(!o)return e[ne].apply(e,[t,n,!1]);for(var a=!1,s=0;s<o.length;s++)if(o[s].handler===n){a=!0,o.splice(s,1);break}a?0===o.length&&r.apply(e,[t,se,!1]):e[ne].apply(e,[t,n,!1])},t=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(0,Object(o.Inject)(_)),Object(r.__param)(2,Object(o.Optional)()),Object(r.__param)(2,Object(o.Inject)(o.PLATFORM_ID)),Object(r.__metadata)("design:paramtypes",[Object,o.NgZone,Object])],t)}(P),ce={pan:!0,panstart:!0,panmove:!0,panend:!0,pancancel:!0,panleft:!0,panright:!0,panup:!0,pandown:!0,pinch:!0,pinchstart:!0,pinchmove:!0,pinchend:!0,pinchcancel:!0,pinchin:!0,pinchout:!0,press:!0,pressup:!0,rotate:!0,rotatestart:!0,rotatemove:!0,rotateend:!0,rotatecancel:!0,swipe:!0,swipeleft:!0,swiperight:!0,swipeup:!0,swipedown:!0,tap:!0},le=new o.InjectionToken("HammerGestureConfig"),de=new o.InjectionToken("HammerLoader"),pe=function(){function e(){this.events=[],this.overrides={}}return e.prototype.buildHammer=function(e){var t=new Hammer(e,this.options);t.get("pinch").set({enable:!0}),t.get("rotate").set({enable:!0});for(var n in this.overrides)t.get(n).set(this.overrides[n]);return t},e=Object(r.__decorate)([Object(o.Injectable)()],e)}(),fe=function(e){function t(t,n,r,i){var o=e.call(this,t)||this;return o._config=n,o.console=r,o.loader=i,o}return Object(r.__extends)(t,e),t.prototype.supports=function(e){return!(!ce.hasOwnProperty(e.toLowerCase())&&!this.isCustomEvent(e))&&(!(!window.Hammer&&!this.loader)||(this.console.warn('The "'+e+'" event cannot be bound because Hammer.JS is not loaded and no custom loader has been specified.'),!1))},t.prototype.addEventListener=function(e,t,n){var r=this,i=this.manager.getZone();if(t=t.toLowerCase(),!window.Hammer&&this.loader){var o=!1,a=function(){o=!0};return this.loader().then(function(){if(!window.Hammer)return r.console.warn("The custom HAMMER_LOADER completed, but Hammer.JS is not present."),void(a=function(){});o||(a=r.addEventListener(e,t,n))}).catch(function(){r.console.warn('The "'+t+'" event cannot be bound because the custom Hammer.JS loader failed.'),a=function(){}}),function(){a()}}return i.runOutsideAngular(function(){var o=r._config.buildHammer(e),a=function(e){i.runGuarded(function(){n(e)})};return o.on(t,a),function(){o.off(t,a),"function"==typeof o.destroy&&o.destroy()}})},t.prototype.isCustomEvent=function(e){return this._config.events.indexOf(e)>-1},t=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(0,Object(o.Inject)(_)),Object(r.__param)(1,Object(o.Inject)(le)),Object(r.__param)(3,Object(o.Optional)()),Object(r.__param)(3,Object(o.Inject)(de)),Object(r.__metadata)("design:paramtypes",[Object,pe,o["ɵConsole"],Object])],t)}(P),he=["alt","control","meta","shift"],me={alt:function(e){return e.altKey},control:function(e){return e.ctrlKey},meta:function(e){return e.metaKey},shift:function(e){return e.shiftKey}},ge=function(e){function t(t){return e.call(this,t)||this}var n;return Object(r.__extends)(t,e),n=t,t.prototype.supports=function(e){return null!=n.parseEventName(e)},t.prototype.addEventListener=function(e,t,r){var i=n.parseEventName(t),o=n.eventCallback(i.fullKey,r,this.manager.getZone());return this.manager.getZone().runOutsideAngular(function(){return s().onAndCancel(e,i.domEventName,o)})},t.parseEventName=function(e){var t=e.toLowerCase().split("."),r=t.shift();if(0===t.length||"keydown"!==r&&"keyup"!==r)return null;var i=n._normalizeKey(t.pop()),o="";if(he.forEach(function(e){var n=t.indexOf(e);n>-1&&(t.splice(n,1),o+=e+".")}),o+=i,0!=t.length||0===i.length)return null;var a={};return a.domEventName=r,a.fullKey=o,a},t.getEventFullKey=function(e){var t="",n=s().getEventKey(e);return" "===(n=n.toLowerCase())?n="space":"."===n&&(n="dot"),he.forEach(function(r){r!=n&&((0,me[r])(e)&&(t+=r+"."))}),t+=n},t.eventCallback=function(e,t,r){return function(i){n.getEventFullKey(i)===e&&r.runGuarded(function(){return t(i)})}},t._normalizeKey=function(e){switch(e){case"esc":return"escape";default:return e}},t=n=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(0,Object(o.Inject)(_)),Object(r.__metadata)("design:paramtypes",[Object])],t)}(P),ve=function(){return function(){}}(),_e=function(e){function t(t){var n=e.call(this)||this;return n._doc=t,n}return Object(r.__extends)(t,e),t.prototype.sanitize=function(e,t){if(null==t)return null;switch(e){case o.SecurityContext.NONE:return t;case o.SecurityContext.HTML:return t instanceof be?t.changingThisBreaksApplicationSecurity:(this.checkNotSafeValue(t,"HTML"),Object(o["ɵ_sanitizeHtml"])(this._doc,String(t)));case o.SecurityContext.STYLE:return t instanceof je?t.changingThisBreaksApplicationSecurity:(this.checkNotSafeValue(t,"Style"),Object(o["ɵ_sanitizeStyle"])(t));case o.SecurityContext.SCRIPT:if(t instanceof xe)return t.changingThisBreaksApplicationSecurity;throw this.checkNotSafeValue(t,"Script"),new Error("unsafe value used in a script context");case o.SecurityContext.URL:return t instanceof we||t instanceof Oe?t.changingThisBreaksApplicationSecurity:(this.checkNotSafeValue(t,"URL"),Object(o["ɵ_sanitizeUrl"])(String(t)));case o.SecurityContext.RESOURCE_URL:if(t instanceof we)return t.changingThisBreaksApplicationSecurity;throw this.checkNotSafeValue(t,"ResourceURL"),new Error("unsafe value used in a resource URL context (see http://g.co/ng/security#xss)");default:throw new Error("Unexpected SecurityContext "+e+" (see http://g.co/ng/security#xss)")}},t.prototype.checkNotSafeValue=function(e,t){if(e instanceof ye)throw new Error("Required a safe "+t+", got a "+e.getTypeName()+" (see http://g.co/ng/security#xss)")},t.prototype.bypassSecurityTrustHtml=function(e){return new be(e)},t.prototype.bypassSecurityTrustStyle=function(e){return new je(e)},t.prototype.bypassSecurityTrustScript=function(e){return new xe(e)},t.prototype.bypassSecurityTrustUrl=function(e){return new Oe(e)},t.prototype.bypassSecurityTrustResourceUrl=function(e){return new we(e)},t=Object(r.__decorate)([Object(o.Injectable)(),Object(r.__param)(0,Object(o.Inject)(_)),Object(r.__metadata)("design:paramtypes",[Object])],t)}(ve),ye=function(){function e(e){this.changingThisBreaksApplicationSecurity=e}return e.prototype.toString=function(){return"SafeValue must use [property]=binding: "+this.changingThisBreaksApplicationSecurity+" (see http://g.co/ng/security#xss)"},e}(),be=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.getTypeName=function(){return"HTML"},t}(ye),je=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.getTypeName=function(){return"Style"},t}(ye),xe=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.getTypeName=function(){return"Script"},t}(ye),Oe=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.getTypeName=function(){return"URL"},t}(ye),we=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return Object(r.__extends)(t,e),t.prototype.getTypeName=function(){return"ResourceURL"},t}(ye),Ce=[{provide:o.PLATFORM_ID,useValue:i["ɵPLATFORM_BROWSER_ID"]},{provide:o.PLATFORM_INITIALIZER,useValue:ke,multi:!0},{provide:i.PlatformLocation,useClass:b,deps:[_]},{provide:_,useFactory:Ae,deps:[]}],Ee=[{provide:o.Sanitizer,useExisting:ve},{provide:ve,useClass:_e,deps:[_]}],Se=Object(o.createPlatformFactory)(o.platformCore,"browser",Ce);function ke(){g.makeCurrent(),w.init()}function Me(){return new o.ErrorHandler}function Ae(){return document}var Te=[Ee,{provide:o["ɵAPP_ROOT"],useValue:!0},{provide:o.ErrorHandler,useFactory:Me,deps:[]},{provide:D,useClass:ue,multi:!0,deps:[_,o.NgZone,o.PLATFORM_ID]},{provide:D,useClass:ge,multi:!0,deps:[_]},{provide:D,useClass:fe,multi:!0,deps:[_,le,o["ɵConsole"],[new o.Optional,de]]},{provide:le,useClass:pe,deps:[]},{provide:G,useClass:G,deps:[I,N]},{provide:o.RendererFactory2,useExisting:G},{provide:R,useExisting:N},{provide:N,useClass:N,deps:[_]},{provide:o.Testability,useClass:o.Testability,deps:[o.NgZone]},{provide:I,useClass:I,deps:[D,o.NgZone]},T],De=function(){function e(e){if(e)throw new Error("BrowserModule has already been loaded. If you need access to common directives such as NgIf and NgFor from a lazy loaded module, import CommonModule instead.")}var t;return t=e,e.withServerTransition=function(e){return{ngModule:t,providers:[{provide:o.APP_ID,useValue:e.appId},{provide:j,useExisting:o.APP_ID},O]}},e=t=Object(r.__decorate)([Object(o.NgModule)({providers:Te,exports:[i.CommonModule,o.ApplicationModule]}),Object(r.__param)(0,Object(o.Optional)()),Object(r.__param)(0,Object(o.SkipSelf)()),Object(r.__param)(0,Object(o.Inject)(t)),Object(r.__metadata)("design:paramtypes",[Object])],e)}();function Ie(){return new Pe(Object(o.inject)(_))}var Pe=function(){function e(e){this._doc=e,this._dom=s()}return e.prototype.addTag=function(e,t){return void 0===t&&(t=!1),e?this._getOrCreateElement(e,t):null},e.prototype.addTags=function(e,t){var n=this;return void 0===t&&(t=!1),e?e.reduce(function(e,r){return r&&e.push(n._getOrCreateElement(r,t)),e},[]):[]},e.prototype.getTag=function(e){return e?this._dom.querySelector(this._doc,"meta["+e+"]")||null:null},e.prototype.getTags=function(e){if(!e)return[];var t=this._dom.querySelectorAll(this._doc,"meta["+e+"]");return t?[].slice.call(t):[]},e.prototype.updateTag=function(e,t){if(!e)return null;t=t||this._parseSelector(e);var n=this.getTag(t);return n?this._setMetaElementAttributes(e,n):this._getOrCreateElement(e,!0)},e.prototype.removeTag=function(e){this.removeTagElement(this.getTag(e))},e.prototype.removeTagElement=function(e){e&&this._dom.remove(e)},e.prototype._getOrCreateElement=function(e,t){if(void 0===t&&(t=!1),!t){var n=this._parseSelector(e),r=this.getTag(n);if(r&&this._containsAttributes(e,r))return r}var i=this._dom.createElement("meta");this._setMetaElementAttributes(e,i);var o=this._dom.getElementsByTagName(this._doc,"head")[0];return this._dom.appendChild(o,i),i},e.prototype._setMetaElementAttributes=function(e,t){var n=this;return Object.keys(e).forEach(function(r){return n._dom.setAttribute(t,r,e[r])}),t},e.prototype._parseSelector=function(e){var t=e.name?"name":"property";return t+'="'+e[t]+'"'},e.prototype._containsAttributes=function(e,t){var n=this;return Object.keys(e).every(function(r){return n._dom.getAttribute(t,r)===e[r]})},e.ngInjectableDef=Object(o.defineInjectable)({factory:Ie,token:e,providedIn:"root"}),e=Object(r.__decorate)([Object(o.Injectable)({providedIn:"root",useFactory:Ie,deps:[]}),Object(r.__param)(0,Object(o.Inject)(_)),Object(r.__metadata)("design:paramtypes",[Object])],e)}();function Re(){return new Ne(Object(o.inject)(_))}var Ne=function(){function e(e){this._doc=e}return e.prototype.getTitle=function(){return s().getTitle(this._doc)},e.prototype.setTitle=function(e){s().setTitle(this._doc,e)},e.ngInjectableDef=Object(o.defineInjectable)({factory:Re,token:e,providedIn:"root"}),e=Object(r.__decorate)([Object(o.Injectable)({providedIn:"root",useFactory:Re,deps:[]}),Object(r.__param)(0,Object(o.Inject)(_)),Object(r.__metadata)("design:paramtypes",[Object])],e)}(),Fe="undefined"!=typeof window&&window||{},Le=function(){return function(e,t){this.msPerTick=e,this.numTicks=t}}(),Ve=function(){function e(e){this.appRef=e.injector.get(o.ApplicationRef)}return e.prototype.timeChangeDetection=function(e){var t=e&&e.record,n=null!=Fe.console.profile;t&&n&&Fe.console.profile("Change Detection");for(var r=s().performanceNow(),i=0;i<5||s().performanceNow()-r<500;)this.appRef.tick(),i++;var o=s().performanceNow();t&&n&&Fe.console.profileEnd("Change Detection");var a=(o-r)/i;return Fe.console.log("ran "+i+" change detection cycles"),Fe.console.log(a.toFixed(2)+" ms per check"),new Le(a,i)},e}(),Be="profiler";function ze(e){return C(Be,new Ve(e)),e}function Ue(){C(Be,null)}function He(e){var t={"&":"&a;",'"':"&q;","'":"&s;","<":"&l;",">":"&g;"};return e.replace(/[&"'<>]/g,function(e){return t[e]})}function qe(e){return e}var Ge=function(){function e(){this.store={},this.onSerializeCallbacks={}}var t;return t=e,e.init=function(e){var n=new t;return n.store=e,n},e.prototype.get=function(e,t){return void 0!==this.store[e]?this.store[e]:t},e.prototype.set=function(e,t){this.store[e]=t},e.prototype.remove=function(e){delete this.store[e]},e.prototype.hasKey=function(e){return this.store.hasOwnProperty(e)},e.prototype.onSerialize=function(e,t){this.onSerializeCallbacks[e]=t},e.prototype.toJson=function(){for(var e in this.onSerializeCallbacks)if(this.onSerializeCallbacks.hasOwnProperty(e))try{this.store[e]=this.onSerializeCallbacks[e]()}catch(e){console.warn("Exception in onSerialize callback: ",e)}return JSON.stringify(this.store)},e=t=Object(r.__decorate)([Object(o.Injectable)()],e)}();function We(e,t){var n,r,i=e.getElementById(t+"-state"),o={};if(i&&i.textContent)try{o=JSON.parse((n=i.textContent,r={"&a;":"&","&q;":'"',"&s;":"'","&l;":"<","&g;":">"},n.replace(/&[^;]+;/g,function(e){return r[e]})))}catch(e){console.warn("Exception while restoring TransferState for app "+t,e)}return Ge.init(o)}var Ye=function(){function e(){}return e=Object(r.__decorate)([Object(o.NgModule)({providers:[{provide:Ge,useFactory:We,deps:[_,o.APP_ID]}]})],e)}(),Ke=function(){function e(){}return e.all=function(){return function(e){return!0}},e.css=function(e){return function(t){return null!=t.nativeElement&&s().elementMatches(t.nativeElement,e)}},e.directive=function(e){return function(t){return-1!==t.providerTokens.indexOf(e)}},e}(),Qe=new o.Version("7.2.0")},"./node_modules/@angular/platform-browser/node_modules/tslib/tslib.es6.js":function(e,t,n){"use strict";n.r(t),n.d(t,"__extends",function(){return i}),n.d(t,"__assign",function(){return o}),n.d(t,"__rest",function(){return a}),n.d(t,"__decorate",function(){return s}),n.d(t,"__param",function(){return u}),n.d(t,"__metadata",function(){return c}),n.d(t,"__awaiter",function(){return l}),n.d(t,"__generator",function(){return d}),n.d(t,"__exportStar",function(){return p}),n.d(t,"__values",function(){return f}),n.d(t,"__read",function(){return h}),n.d(t,"__spread",function(){return m}),n.d(t,"__await",function(){return g}),n.d(t,"__asyncGenerator",function(){return v}),n.d(t,"__asyncDelegator",function(){return _}),n.d(t,"__asyncValues",function(){return y}),n.d(t,"__makeTemplateObject",function(){return b}),n.d(t,"__importStar",function(){return j}),n.d(t,"__importDefault",function(){return x});var r=function(e,t){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,t){e.__proto__=t}||function(e,t){for(var n in t)t.hasOwnProperty(n)&&(e[n]=t[n])})(e,t)};function i(e,t){function n(){this.constructor=e}r(e,t),e.prototype=null===t?Object.create(t):(n.prototype=t.prototype,new n)}var o=function(){return(o=Object.assign||function(e){for(var t,n=1,r=arguments.length;n<r;n++){t=arguments[n];for(var i in t)Object.prototype.hasOwnProperty.call(t,i)&&(e[i]=t[i])}return e}).apply(this,arguments)};function a(e,t){var n={};for(var r in e)Object.prototype.hasOwnProperty.call(e,r)&&t.indexOf(r)<0&&(n[r]=e[r]);if(null!=e&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(e);i<r.length;i++)t.indexOf(r[i])<0&&(n[r[i]]=e[r[i]])}return n}function s(e,t,n,r){var i,o=arguments.length,a=o<3?t:null===r?r=Object.getOwnPropertyDescriptor(t,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(e,t,n,r);else for(var s=e.length-1;s>=0;s--)(i=e[s])&&(a=(o<3?i(a):o>3?i(t,n,a):i(t,n))||a);return o>3&&a&&Object.defineProperty(t,n,a),a}function u(e,t){return function(n,r){t(n,r,e)}}function c(e,t){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(e,t)}function l(e,t,n,r){return new(n||(n=Promise))(function(i,o){function a(e){try{u(r.next(e))}catch(e){o(e)}}function s(e){try{u(r.throw(e))}catch(e){o(e)}}function u(e){e.done?i(e.value):new n(function(t){t(e.value)}).then(a,s)}u((r=r.apply(e,t||[])).next())})}function d(e,t){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=t.call(e,a)}catch(e){o=[6,e],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(e,t){for(var n in e)t.hasOwnProperty(n)||(t[n]=e[n])}function f(e){var t="function"==typeof Symbol&&e[Symbol.iterator],n=0;return t?t.call(e):{next:function(){return e&&n>=e.length&&(e=void 0),{value:e&&e[n++],done:!e}}}}function h(e,t){var n="function"==typeof Symbol&&e[Symbol.iterator];if(!n)return e;var r,i,o=n.call(e),a=[];try{for(;(void 0===t||t-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(e){i={error:e}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var e=[],t=0;t<arguments.length;t++)e=e.concat(h(arguments[t]));return e}function g(e){return this instanceof g?(this.v=e,this):new g(e)}function v(e,t,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(e,t||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(e){i[e]&&(r[e]=function(t){return new Promise(function(n,r){o.push([e,t,n,r])>1||s(e,t)})})}function s(e,t){try{(n=i[e](t)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(e){l(o[0][3],e)}var n}function u(e){s("next",e)}function c(e){s("throw",e)}function l(e,t){e(t),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(e){var t,n;return t={},r("next"),r("throw",function(e){throw e}),r("return"),t[Symbol.iterator]=function(){return this},t;function r(r,i){t[r]=e[r]?function(t){return(n=!n)?{value:g(e[r](t)),done:"return"===r}:i?i(t):t}:i}}function y(e){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var t,n=e[Symbol.asyncIterator];return n?n.call(e):(e=f(e),t={},r("next"),r("throw"),r("return"),t[Symbol.asyncIterator]=function(){return this},t);function r(n){t[n]=e[n]&&function(t){return new Promise(function(r,i){var o,a,s,u;t=e[n](t),o=r,a=i,s=t.done,u=t.value,Promise.resolve(u).then(function(e){o({value:e,done:s})},a)})}}}function b(e,t){return Object.defineProperty?Object.defineProperty(e,"raw",{value:t}):e.raw=t,e}function j(e){if(e&&e.__esModule)return e;var t={};if(null!=e)for(var n in e)Object.hasOwnProperty.call(e,n)&&(t[n]=e[n]);return t.default=e,t}function x(e){return e&&e.__esModule?e:{default:e}}},"./node_modules/d3-array/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-array/src/bisect.js");n.d(t,"bisect",function(){return r.default}),n.d(t,"bisectRight",function(){return r.bisectRight}),n.d(t,"bisectLeft",function(){return r.bisectLeft});var i=n("./node_modules/d3-array/src/ascending.js");n.d(t,"ascending",function(){return i.default});var o=n("./node_modules/d3-array/src/bisector.js");n.d(t,"bisector",function(){return o.default});var a=n("./node_modules/d3-array/src/cross.js");n.d(t,"cross",function(){return a.default});var s=n("./node_modules/d3-array/src/descending.js");n.d(t,"descending",function(){return s.default});var u=n("./node_modules/d3-array/src/deviation.js");n.d(t,"deviation",function(){return u.default});var c=n("./node_modules/d3-array/src/extent.js");n.d(t,"extent",function(){return c.default});var l=n("./node_modules/d3-array/src/histogram.js");n.d(t,"histogram",function(){return l.default});var d=n("./node_modules/d3-array/src/threshold/freedmanDiaconis.js");n.d(t,"thresholdFreedmanDiaconis",function(){return d.default});var p=n("./node_modules/d3-array/src/threshold/scott.js");n.d(t,"thresholdScott",function(){return p.default});var f=n("./node_modules/d3-array/src/threshold/sturges.js");n.d(t,"thresholdSturges",function(){return f.default});var h=n("./node_modules/d3-array/src/max.js");n.d(t,"max",function(){return h.default});var m=n("./node_modules/d3-array/src/mean.js");n.d(t,"mean",function(){return m.default});var g=n("./node_modules/d3-array/src/median.js");n.d(t,"median",function(){return g.default});var v=n("./node_modules/d3-array/src/merge.js");n.d(t,"merge",function(){return v.default});var _=n("./node_modules/d3-array/src/min.js");n.d(t,"min",function(){return _.default});var y=n("./node_modules/d3-array/src/pairs.js");n.d(t,"pairs",function(){return y.default});var b=n("./node_modules/d3-array/src/permute.js");n.d(t,"permute",function(){return b.default});var j=n("./node_modules/d3-array/src/quantile.js");n.d(t,"quantile",function(){return j.default});var x=n("./node_modules/d3-array/src/range.js");n.d(t,"range",function(){return x.default});var O=n("./node_modules/d3-array/src/scan.js");n.d(t,"scan",function(){return O.default});var w=n("./node_modules/d3-array/src/shuffle.js");n.d(t,"shuffle",function(){return w.default});var C=n("./node_modules/d3-array/src/sum.js");n.d(t,"sum",function(){return C.default});var E=n("./node_modules/d3-array/src/ticks.js");n.d(t,"ticks",function(){return E.default}),n.d(t,"tickIncrement",function(){return E.tickIncrement}),n.d(t,"tickStep",function(){return E.tickStep});var S=n("./node_modules/d3-array/src/transpose.js");n.d(t,"transpose",function(){return S.default});var k=n("./node_modules/d3-array/src/variance.js");n.d(t,"variance",function(){return k.default});var M=n("./node_modules/d3-array/src/zip.js");n.d(t,"zip",function(){return M.default})},"./node_modules/d3-array/src/array.js":function(e,t,n){"use strict";n.r(t),n.d(t,"slice",function(){return i}),n.d(t,"map",function(){return o});var r=Array.prototype,i=r.slice,o=r.map},"./node_modules/d3-array/src/ascending.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){return e<t?-1:e>t?1:e>=t?0:NaN}},"./node_modules/d3-array/src/bisect.js":function(e,t,n){"use strict";n.r(t),n.d(t,"bisectRight",function(){return a}),n.d(t,"bisectLeft",function(){return s});var r=n("./node_modules/d3-array/src/ascending.js"),i=n("./node_modules/d3-array/src/bisector.js"),o=Object(i.default)(r.default),a=o.right,s=o.left;t.default=a},"./node_modules/d3-array/src/bisector.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-array/src/ascending.js");t.default=function(e){var t;return 1===e.length&&(t=e,e=function(e,n){return Object(r.default)(t(e),n)}),{left:function(t,n,r,i){for(null==r&&(r=0),null==i&&(i=t.length);r<i;){var o=r+i>>>1;e(t[o],n)<0?r=o+1:i=o}return r},right:function(t,n,r,i){for(null==r&&(r=0),null==i&&(i=t.length);r<i;){var o=r+i>>>1;e(t[o],n)>0?i=o:r=o+1}return r}}}},"./node_modules/d3-array/src/constant.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return function(){return e}}},"./node_modules/d3-array/src/cross.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-array/src/pairs.js");t.default=function(e,t,n){var i,o,a,s,u=e.length,c=t.length,l=new Array(u*c);for(null==n&&(n=r.pair),i=a=0;i<u;++i)for(s=e[i],o=0;o<c;++o,++a)l[a]=n(s,t[o]);return l}},"./node_modules/d3-array/src/descending.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){return t<e?-1:t>e?1:t>=e?0:NaN}},"./node_modules/d3-array/src/deviation.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-array/src/variance.js");t.default=function(e,t){var n=Object(r.default)(e,t);return n?Math.sqrt(n):n}},"./node_modules/d3-array/src/extent.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){var n,r,i,o=e.length,a=-1;if(null==t){for(;++a<o;)if(null!=(n=e[a])&&n>=n)for(r=i=n;++a<o;)null!=(n=e[a])&&(r>n&&(r=n),i<n&&(i=n))}else for(;++a<o;)if(null!=(n=t(e[a],a,e))&&n>=n)for(r=i=n;++a<o;)null!=(n=t(e[a],a,e))&&(r>n&&(r=n),i<n&&(i=n));return[r,i]}},"./node_modules/d3-array/src/histogram.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-array/src/array.js"),i=n("./node_modules/d3-array/src/bisect.js"),o=n("./node_modules/d3-array/src/constant.js"),a=n("./node_modules/d3-array/src/extent.js"),s=n("./node_modules/d3-array/src/identity.js"),u=n("./node_modules/d3-array/src/range.js"),c=n("./node_modules/d3-array/src/ticks.js"),l=n("./node_modules/d3-array/src/threshold/sturges.js");t.default=function(){var e=s.default,t=a.default,n=l.default;function d(r){var o,a,s=r.length,l=new Array(s);for(o=0;o<s;++o)l[o]=e(r[o],o,r);var d=t(l),p=d[0],f=d[1],h=n(l,p,f);Array.isArray(h)||(h=Object(c.tickStep)(p,f,h),h=Object(u.default)(Math.ceil(p/h)*h,Math.floor(f/h)*h,h));for(var m=h.length;h[0]<=p;)h.shift(),--m;for(;h[m-1]>f;)h.pop(),--m;var g,v=new Array(m+1);for(o=0;o<=m;++o)(g=v[o]=[]).x0=o>0?h[o-1]:p,g.x1=o<m?h[o]:f;for(o=0;o<s;++o)p<=(a=l[o])&&a<=f&&v[Object(i.default)(h,a,0,m)].push(r[o]);return v}return d.value=function(t){return arguments.length?(e="function"==typeof t?t:Object(o.default)(t),d):e},d.domain=function(e){return arguments.length?(t="function"==typeof e?e:Object(o.default)([e[0],e[1]]),d):t},d.thresholds=function(e){return arguments.length?(n="function"==typeof e?e:Array.isArray(e)?Object(o.default)(r.slice.call(e)):Object(o.default)(e),d):n},d}},"./node_modules/d3-array/src/identity.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return e}},"./node_modules/d3-array/src/max.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){var n,r,i=e.length,o=-1;if(null==t){for(;++o<i;)if(null!=(n=e[o])&&n>=n)for(r=n;++o<i;)null!=(n=e[o])&&n>r&&(r=n)}else for(;++o<i;)if(null!=(n=t(e[o],o,e))&&n>=n)for(r=n;++o<i;)null!=(n=t(e[o],o,e))&&n>r&&(r=n);return r}},"./node_modules/d3-array/src/mean.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-array/src/number.js");t.default=function(e,t){var n,i=e.length,o=i,a=-1,s=0;if(null==t)for(;++a<i;)isNaN(n=Object(r.default)(e[a]))?--o:s+=n;else for(;++a<i;)isNaN(n=Object(r.default)(t(e[a],a,e)))?--o:s+=n;if(o)return s/o}},"./node_modules/d3-array/src/median.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-array/src/ascending.js"),i=n("./node_modules/d3-array/src/number.js"),o=n("./node_modules/d3-array/src/quantile.js");t.default=function(e,t){var n,a=e.length,s=-1,u=[];if(null==t)for(;++s<a;)isNaN(n=Object(i.default)(e[s]))||u.push(n);else for(;++s<a;)isNaN(n=Object(i.default)(t(e[s],s,e)))||u.push(n);return Object(o.default)(u.sort(r.default),.5)}},"./node_modules/d3-array/src/merge.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){for(var t,n,r,i=e.length,o=-1,a=0;++o<i;)a+=e[o].length;for(n=new Array(a);--i>=0;)for(t=(r=e[i]).length;--t>=0;)n[--a]=r[t];return n}},"./node_modules/d3-array/src/min.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){var n,r,i=e.length,o=-1;if(null==t){for(;++o<i;)if(null!=(n=e[o])&&n>=n)for(r=n;++o<i;)null!=(n=e[o])&&r>n&&(r=n)}else for(;++o<i;)if(null!=(n=t(e[o],o,e))&&n>=n)for(r=n;++o<i;)null!=(n=t(e[o],o,e))&&r>n&&(r=n);return r}},"./node_modules/d3-array/src/number.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return null===e?NaN:+e}},"./node_modules/d3-array/src/pairs.js":function(e,t,n){"use strict";function r(e,t){return[e,t]}n.r(t),n.d(t,"pair",function(){return r}),t.default=function(e,t){null==t&&(t=r);for(var n=0,i=e.length-1,o=e[0],a=new Array(i<0?0:i);n<i;)a[n]=t(o,o=e[++n]);return a}},"./node_modules/d3-array/src/permute.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){for(var n=t.length,r=new Array(n);n--;)r[n]=e[t[n]];return r}},"./node_modules/d3-array/src/quantile.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-array/src/number.js");t.default=function(e,t,n){if(null==n&&(n=r.default),i=e.length){if((t=+t)<=0||i<2)return+n(e[0],0,e);if(t>=1)return+n(e[i-1],i-1,e);var i,o=(i-1)*t,a=Math.floor(o),s=+n(e[a],a,e);return s+(+n(e[a+1],a+1,e)-s)*(o-a)}}},"./node_modules/d3-array/src/range.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n){e=+e,t=+t,n=(i=arguments.length)<2?(t=e,e=0,1):i<3?1:+n;for(var r=-1,i=0|Math.max(0,Math.ceil((t-e)/n)),o=new Array(i);++r<i;)o[r]=e+r*n;return o}},"./node_modules/d3-array/src/scan.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-array/src/ascending.js");t.default=function(e,t){if(n=e.length){var n,i,o=0,a=0,s=e[a];for(null==t&&(t=r.default);++o<n;)(t(i=e[o],s)<0||0!==t(s,s))&&(s=i,a=o);return 0===t(s,s)?a:void 0}}},"./node_modules/d3-array/src/shuffle.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n){for(var r,i,o=(null==n?e.length:n)-(t=null==t?0:+t);o;)i=Math.random()*o--|0,r=e[o+t],e[o+t]=e[i+t],e[i+t]=r;return e}},"./node_modules/d3-array/src/sum.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){var n,r=e.length,i=-1,o=0;if(null==t)for(;++i<r;)(n=+e[i])&&(o+=n);else for(;++i<r;)(n=+t(e[i],i,e))&&(o+=n);return o}},"./node_modules/d3-array/src/threshold/freedmanDiaconis.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-array/src/array.js"),i=n("./node_modules/d3-array/src/ascending.js"),o=n("./node_modules/d3-array/src/number.js"),a=n("./node_modules/d3-array/src/quantile.js");t.default=function(e,t,n){return e=r.map.call(e,o.default).sort(i.default),Math.ceil((n-t)/(2*(Object(a.default)(e,.75)-Object(a.default)(e,.25))*Math.pow(e.length,-1/3)))}},"./node_modules/d3-array/src/threshold/scott.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-array/src/deviation.js");t.default=function(e,t,n){return Math.ceil((n-t)/(3.5*Object(r.default)(e)*Math.pow(e.length,-1/3)))}},"./node_modules/d3-array/src/threshold/sturges.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return Math.ceil(Math.log(e.length)/Math.LN2)+1}},"./node_modules/d3-array/src/ticks.js":function(e,t,n){"use strict";n.r(t),n.d(t,"tickIncrement",function(){return a}),n.d(t,"tickStep",function(){return s});var r=Math.sqrt(50),i=Math.sqrt(10),o=Math.sqrt(2);function a(e,t,n){var a=(t-e)/Math.max(0,n),s=Math.floor(Math.log(a)/Math.LN10),u=a/Math.pow(10,s);return s>=0?(u>=r?10:u>=i?5:u>=o?2:1)*Math.pow(10,s):-Math.pow(10,-s)/(u>=r?10:u>=i?5:u>=o?2:1)}function s(e,t,n){var a=Math.abs(t-e)/Math.max(0,n),s=Math.pow(10,Math.floor(Math.log(a)/Math.LN10)),u=a/s;return u>=r?s*=10:u>=i?s*=5:u>=o&&(s*=2),t<e?-s:s}t.default=function(e,t,n){var r,i,o,s,u=-1;if(n=+n,(e=+e)===(t=+t)&&n>0)return[e];if((r=t<e)&&(i=e,e=t,t=i),0===(s=a(e,t,n))||!isFinite(s))return[];if(s>0)for(e=Math.ceil(e/s),t=Math.floor(t/s),o=new Array(i=Math.ceil(t-e+1));++u<i;)o[u]=(e+u)*s;else for(e=Math.floor(e*s),t=Math.ceil(t*s),o=new Array(i=Math.ceil(e-t+1));++u<i;)o[u]=(e-u)/s;return r&&o.reverse(),o}},"./node_modules/d3-array/src/transpose.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-array/src/min.js");function i(e){return e.length}t.default=function(e){if(!(a=e.length))return[];for(var t=-1,n=Object(r.default)(e,i),o=new Array(n);++t<n;)for(var a,s=-1,u=o[t]=new Array(a);++s<a;)u[s]=e[s][t];return o}},"./node_modules/d3-array/src/variance.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-array/src/number.js");t.default=function(e,t){var n,i,o=e.length,a=0,s=-1,u=0,c=0;if(null==t)for(;++s<o;)isNaN(n=Object(r.default)(e[s]))||(c+=(i=n-u)*(n-(u+=i/++a)));else for(;++s<o;)isNaN(n=Object(r.default)(t(e[s],s,e)))||(c+=(i=n-u)*(n-(u+=i/++a)));if(a>1)return c/(a-1)}},"./node_modules/d3-array/src/zip.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-array/src/transpose.js");t.default=function(){return Object(r.default)(arguments)}},"./node_modules/d3-collection/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-collection/src/nest.js");n.d(t,"nest",function(){return r.default});var i=n("./node_modules/d3-collection/src/set.js");n.d(t,"set",function(){return i.default});var o=n("./node_modules/d3-collection/src/map.js");n.d(t,"map",function(){return o.default});var a=n("./node_modules/d3-collection/src/keys.js");n.d(t,"keys",function(){return a.default});var s=n("./node_modules/d3-collection/src/values.js");n.d(t,"values",function(){return s.default});var u=n("./node_modules/d3-collection/src/entries.js");n.d(t,"entries",function(){return u.default})},"./node_modules/d3-collection/src/entries.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){var t=[];for(var n in e)t.push({key:n,value:e[n]});return t}},"./node_modules/d3-collection/src/keys.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){var t=[];for(var n in e)t.push(n);return t}},"./node_modules/d3-collection/src/map.js":function(e,t,n){"use strict";n.r(t),n.d(t,"prefix",function(){return r});var r="$";function i(){}function o(e,t){var n=new i;if(e instanceof i)e.each(function(e,t){n.set(t,e)});else if(Array.isArray(e)){var r,o=-1,a=e.length;if(null==t)for(;++o<a;)n.set(o,e[o]);else for(;++o<a;)n.set(t(r=e[o],o,e),r)}else if(e)for(var s in e)n.set(s,e[s]);return n}i.prototype=o.prototype={constructor:i,has:function(e){return r+e in this},get:function(e){return this[r+e]},set:function(e,t){return this[r+e]=t,this},remove:function(e){var t=r+e;return t in this&&delete this[t]},clear:function(){for(var e in this)e[0]===r&&delete this[e]},keys:function(){var e=[];for(var t in this)t[0]===r&&e.push(t.slice(1));return e},values:function(){var e=[];for(var t in this)t[0]===r&&e.push(this[t]);return e},entries:function(){var e=[];for(var t in this)t[0]===r&&e.push({key:t.slice(1),value:this[t]});return e},size:function(){var e=0;for(var t in this)t[0]===r&&++e;return e},empty:function(){for(var e in this)if(e[0]===r)return!1;return!0},each:function(e){for(var t in this)t[0]===r&&e(this[t],t.slice(1),this)}},t.default=o},"./node_modules/d3-collection/src/nest.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-collection/src/map.js");function i(){return{}}function o(e,t,n){e[t]=n}function a(){return Object(r.default)()}function s(e,t,n){e.set(t,n)}t.default=function(){var e,t,n,u=[],c=[];function l(n,i,o,a){if(i>=u.length)return null!=e&&n.sort(e),null!=t?t(n):n;for(var s,c,d,p=-1,f=n.length,h=u[i++],m=Object(r.default)(),g=o();++p<f;)(d=m.get(s=h(c=n[p])+""))?d.push(c):m.set(s,[c]);return m.each(function(e,t){a(g,t,l(e,i,o,a))}),g}return n={object:function(e){return l(e,0,i,o)},map:function(e){return l(e,0,a,s)},entries:function(e){return function e(n,r){if(++r>u.length)return n;var i,o=c[r-1];return null!=t&&r>=u.length?i=n.entries():(i=[],n.each(function(t,n){i.push({key:n,values:e(t,r)})})),null!=o?i.sort(function(e,t){return o(e.key,t.key)}):i}(l(e,0,a,s),0)},key:function(e){return u.push(e),n},sortKeys:function(e){return c[u.length-1]=e,n},sortValues:function(t){return e=t,n},rollup:function(e){return t=e,n}}}},"./node_modules/d3-collection/src/set.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-collection/src/map.js");function i(){}var o=r.default.prototype;function a(e,t){var n=new i;if(e instanceof i)e.each(function(e){n.add(e)});else if(e){var r=-1,o=e.length;if(null==t)for(;++r<o;)n.add(e[r]);else for(;++r<o;)n.add(t(e[r],r,e))}return n}i.prototype=a.prototype={constructor:i,has:o.has,add:function(e){return e+="",this[r.prefix+e]=e,this},remove:o.remove,clear:o.clear,values:o.keys,size:o.size,empty:o.empty,each:o.each},t.default=a},"./node_modules/d3-collection/src/values.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){var t=[];for(var n in e)t.push(e[n]);return t}},"./node_modules/d3-color/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-color/src/color.js");n.d(t,"color",function(){return r.default}),n.d(t,"rgb",function(){return r.rgb}),n.d(t,"hsl",function(){return r.hsl});var i=n("./node_modules/d3-color/src/lab.js");n.d(t,"lab",function(){return i.default}),n.d(t,"hcl",function(){return i.hcl});var o=n("./node_modules/d3-color/src/cubehelix.js");n.d(t,"cubehelix",function(){return o.default})},"./node_modules/d3-color/src/color.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Color",function(){return i}),n.d(t,"darker",function(){return o}),n.d(t,"brighter",function(){return a}),n.d(t,"default",function(){return y}),n.d(t,"rgbConvert",function(){return x}),n.d(t,"rgb",function(){return O}),n.d(t,"Rgb",function(){return w}),n.d(t,"hslConvert",function(){return E}),n.d(t,"hsl",function(){return S});var r=n("./node_modules/d3-color/src/define.js");function i(){}var o=.7,a=1/o,s="\\s*([+-]?\\d+)\\s*",u="\\s*([+-]?\\d*\\.?\\d+(?:[eE][+-]?\\d+)?)\\s*",c="\\s*([+-]?\\d*\\.?\\d+(?:[eE][+-]?\\d+)?)%\\s*",l=/^#([0-9a-f]{3})$/,d=/^#([0-9a-f]{6})$/,p=new RegExp("^rgb\\("+[s,s,s]+"\\)$"),f=new RegExp("^rgb\\("+[c,c,c]+"\\)$"),h=new RegExp("^rgba\\("+[s,s,s,u]+"\\)$"),m=new RegExp("^rgba\\("+[c,c,c,u]+"\\)$"),g=new RegExp("^hsl\\("+[u,c,c]+"\\)$"),v=new RegExp("^hsla\\("+[u,c,c,u]+"\\)$"),_={aliceblue:15792383,antiquewhite:16444375,aqua:65535,aquamarine:8388564,azure:15794175,beige:16119260,bisque:16770244,black:0,blanchedalmond:16772045,blue:255,blueviolet:9055202,brown:10824234,burlywood:14596231,cadetblue:6266528,chartreuse:8388352,chocolate:13789470,coral:16744272,cornflowerblue:6591981,cornsilk:16775388,crimson:14423100,cyan:65535,darkblue:139,darkcyan:35723,darkgoldenrod:12092939,darkgray:11119017,darkgreen:25600,darkgrey:11119017,darkkhaki:12433259,darkmagenta:9109643,darkolivegreen:5597999,darkorange:16747520,darkorchid:10040012,darkred:9109504,darksalmon:15308410,darkseagreen:9419919,darkslateblue:4734347,darkslategray:3100495,darkslategrey:3100495,darkturquoise:52945,darkviolet:9699539,deeppink:16716947,deepskyblue:49151,dimgray:6908265,dimgrey:6908265,dodgerblue:2003199,firebrick:11674146,floralwhite:16775920,forestgreen:2263842,fuchsia:16711935,gainsboro:14474460,ghostwhite:16316671,gold:16766720,goldenrod:14329120,gray:8421504,green:32768,greenyellow:11403055,grey:8421504,honeydew:15794160,hotpink:16738740,indianred:13458524,indigo:4915330,ivory:16777200,khaki:15787660,lavender:15132410,lavenderblush:16773365,lawngreen:8190976,lemonchiffon:16775885,lightblue:11393254,lightcoral:15761536,lightcyan:14745599,lightgoldenrodyellow:16448210,lightgray:13882323,lightgreen:9498256,lightgrey:13882323,lightpink:16758465,lightsalmon:16752762,lightseagreen:2142890,lightskyblue:8900346,lightslategray:7833753,lightslategrey:7833753,lightsteelblue:11584734,lightyellow:16777184,lime:65280,limegreen:3329330,linen:16445670,magenta:16711935,maroon:8388608,mediumaquamarine:6737322,mediumblue:205,mediumorchid:12211667,mediumpurple:9662683,mediumseagreen:3978097,mediumslateblue:8087790,mediumspringgreen:64154,mediumturquoise:4772300,mediumvioletred:13047173,midnightblue:1644912,mintcream:16121850,mistyrose:16770273,moccasin:16770229,navajowhite:16768685,navy:128,oldlace:16643558,olive:8421376,olivedrab:7048739,orange:16753920,orangered:16729344,orchid:14315734,palegoldenrod:15657130,palegreen:10025880,paleturquoise:11529966,palevioletred:14381203,papayawhip:16773077,peachpuff:16767673,peru:13468991,pink:16761035,plum:14524637,powderblue:11591910,purple:8388736,rebeccapurple:6697881,red:16711680,rosybrown:12357519,royalblue:4286945,saddlebrown:9127187,salmon:16416882,sandybrown:16032864,seagreen:3050327,seashell:16774638,sienna:10506797,silver:12632256,skyblue:8900331,slateblue:6970061,slategray:7372944,slategrey:7372944,snow:16775930,springgreen:65407,steelblue:4620980,tan:13808780,teal:32896,thistle:14204888,tomato:16737095,turquoise:4251856,violet:15631086,wheat:16113331,white:16777215,whitesmoke:16119285,yellow:16776960,yellowgreen:10145074};function y(e){var t;return e=(e+"").trim().toLowerCase(),(t=l.exec(e))?new w((t=parseInt(t[1],16))>>8&15|t>>4&240,t>>4&15|240&t,(15&t)<<4|15&t,1):(t=d.exec(e))?b(parseInt(t[1],16)):(t=p.exec(e))?new w(t[1],t[2],t[3],1):(t=f.exec(e))?new w(255*t[1]/100,255*t[2]/100,255*t[3]/100,1):(t=h.exec(e))?j(t[1],t[2],t[3],t[4]):(t=m.exec(e))?j(255*t[1]/100,255*t[2]/100,255*t[3]/100,t[4]):(t=g.exec(e))?C(t[1],t[2]/100,t[3]/100,1):(t=v.exec(e))?C(t[1],t[2]/100,t[3]/100,t[4]):_.hasOwnProperty(e)?b(_[e]):"transparent"===e?new w(NaN,NaN,NaN,0):null}function b(e){return new w(e>>16&255,e>>8&255,255&e,1)}function j(e,t,n,r){return r<=0&&(e=t=n=NaN),new w(e,t,n,r)}function x(e){return e instanceof i||(e=y(e)),e?new w((e=e.rgb()).r,e.g,e.b,e.opacity):new w}function O(e,t,n,r){return 1===arguments.length?x(e):new w(e,t,n,null==r?1:r)}function w(e,t,n,r){this.r=+e,this.g=+t,this.b=+n,this.opacity=+r}function C(e,t,n,r){return r<=0?e=t=n=NaN:n<=0||n>=1?e=t=NaN:t<=0&&(e=NaN),new k(e,t,n,r)}function E(e){if(e instanceof k)return new k(e.h,e.s,e.l,e.opacity);if(e instanceof i||(e=y(e)),!e)return new k;if(e instanceof k)return e;var t=(e=e.rgb()).r/255,n=e.g/255,r=e.b/255,o=Math.min(t,n,r),a=Math.max(t,n,r),s=NaN,u=a-o,c=(a+o)/2;return u?(s=t===a?(n-r)/u+6*(n<r):n===a?(r-t)/u+2:(t-n)/u+4,u/=c<.5?a+o:2-a-o,s*=60):u=c>0&&c<1?0:s,new k(s,u,c,e.opacity)}function S(e,t,n,r){return 1===arguments.length?E(e):new k(e,t,n,null==r?1:r)}function k(e,t,n,r){this.h=+e,this.s=+t,this.l=+n,this.opacity=+r}function M(e,t,n){return 255*(e<60?t+(n-t)*e/60:e<180?n:e<240?t+(n-t)*(240-e)/60:t)}Object(r.default)(i,y,{displayable:function(){return this.rgb().displayable()},toString:function(){return this.rgb()+""}}),Object(r.default)(w,O,Object(r.extend)(i,{brighter:function(e){return e=null==e?a:Math.pow(a,e),new w(this.r*e,this.g*e,this.b*e,this.opacity)},darker:function(e){return e=null==e?o:Math.pow(o,e),new w(this.r*e,this.g*e,this.b*e,this.opacity)},rgb:function(){return this},displayable:function(){return 0<=this.r&&this.r<=255&&0<=this.g&&this.g<=255&&0<=this.b&&this.b<=255&&0<=this.opacity&&this.opacity<=1},toString:function(){var e=this.opacity;return(1===(e=isNaN(e)?1:Math.max(0,Math.min(1,e)))?"rgb(":"rgba(")+Math.max(0,Math.min(255,Math.round(this.r)||0))+", "+Math.max(0,Math.min(255,Math.round(this.g)||0))+", "+Math.max(0,Math.min(255,Math.round(this.b)||0))+(1===e?")":", "+e+")")}})),Object(r.default)(k,S,Object(r.extend)(i,{brighter:function(e){return e=null==e?a:Math.pow(a,e),new k(this.h,this.s,this.l*e,this.opacity)},darker:function(e){return e=null==e?o:Math.pow(o,e),new k(this.h,this.s,this.l*e,this.opacity)},rgb:function(){var e=this.h%360+360*(this.h<0),t=isNaN(e)||isNaN(this.s)?0:this.s,n=this.l,r=n+(n<.5?n:1-n)*t,i=2*n-r;return new w(M(e>=240?e-240:e+120,i,r),M(e,i,r),M(e<120?e+240:e-120,i,r),this.opacity)},displayable:function(){return(0<=this.s&&this.s<=1||isNaN(this.s))&&0<=this.l&&this.l<=1&&0<=this.opacity&&this.opacity<=1}}))},"./node_modules/d3-color/src/cubehelix.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return p}),n.d(t,"Cubehelix",function(){return f});var r=n("./node_modules/d3-color/src/define.js"),i=n("./node_modules/d3-color/src/color.js"),o=n("./node_modules/d3-color/src/math.js"),a=-.29227,s=-.90649,u=1.97294,c=u*s,l=1.78277*u,d=1.78277*a- -.14861*s;function p(e,t,n,r){return 1===arguments.length?function(e){if(e instanceof f)return new f(e.h,e.s,e.l,e.opacity);e instanceof i.Rgb||(e=Object(i.rgbConvert)(e));var t=e.r/255,n=e.g/255,r=e.b/255,p=(d*r+c*t-l*n)/(d+c-l),h=r-p,m=(u*(n-p)-a*h)/s,g=Math.sqrt(m*m+h*h)/(u*p*(1-p)),v=g?Math.atan2(m,h)*o.rad2deg-120:NaN;return new f(v<0?v+360:v,g,p,e.opacity)}(e):new f(e,t,n,null==r?1:r)}function f(e,t,n,r){this.h=+e,this.s=+t,this.l=+n,this.opacity=+r}Object(r.default)(f,p,Object(r.extend)(i.Color,{brighter:function(e){return e=null==e?i.brighter:Math.pow(i.brighter,e),new f(this.h,this.s,this.l*e,this.opacity)},darker:function(e){return e=null==e?i.darker:Math.pow(i.darker,e),new f(this.h,this.s,this.l*e,this.opacity)},rgb:function(){var e=isNaN(this.h)?0:(this.h+120)*o.deg2rad,t=+this.l,n=isNaN(this.s)?0:this.s*t*(1-t),r=Math.cos(e),c=Math.sin(e);return new i.Rgb(255*(t+n*(-.14861*r+1.78277*c)),255*(t+n*(a*r+s*c)),255*(t+n*(u*r)),this.opacity)}}))},"./node_modules/d3-color/src/define.js":function(e,t,n){"use strict";function r(e,t){var n=Object.create(e.prototype);for(var r in t)n[r]=t[r];return n}n.r(t),n.d(t,"extend",function(){return r}),t.default=function(e,t,n){e.prototype=t.prototype=n,n.constructor=e}},"./node_modules/d3-color/src/lab.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return h}),n.d(t,"Lab",function(){return m}),n.d(t,"hcl",function(){return b}),n.d(t,"Hcl",function(){return j});var r=n("./node_modules/d3-color/src/define.js"),i=n("./node_modules/d3-color/src/color.js"),o=n("./node_modules/d3-color/src/math.js"),a=.95047,s=1,u=1.08883,c=4/29,l=6/29,d=3*l*l,p=l*l*l;function f(e){if(e instanceof m)return new m(e.l,e.a,e.b,e.opacity);if(e instanceof j){var t=e.h*o.deg2rad;return new m(e.l,Math.cos(t)*e.c,Math.sin(t)*e.c,e.opacity)}e instanceof i.Rgb||(e=Object(i.rgbConvert)(e));var n=y(e.r),r=y(e.g),c=y(e.b),l=g((.4124564*n+.3575761*r+.1804375*c)/a),d=g((.2126729*n+.7151522*r+.072175*c)/s);return new m(116*d-16,500*(l-d),200*(d-g((.0193339*n+.119192*r+.9503041*c)/u)),e.opacity)}function h(e,t,n,r){return 1===arguments.length?f(e):new m(e,t,n,null==r?1:r)}function m(e,t,n,r){this.l=+e,this.a=+t,this.b=+n,this.opacity=+r}function g(e){return e>p?Math.pow(e,1/3):e/d+c}function v(e){return e>l?e*e*e:d*(e-c)}function _(e){return 255*(e<=.0031308?12.92*e:1.055*Math.pow(e,1/2.4)-.055)}function y(e){return(e/=255)<=.04045?e/12.92:Math.pow((e+.055)/1.055,2.4)}function b(e,t,n,r){return 1===arguments.length?function(e){if(e instanceof j)return new j(e.h,e.c,e.l,e.opacity);e instanceof m||(e=f(e));var t=Math.atan2(e.b,e.a)*o.rad2deg;return new j(t<0?t+360:t,Math.sqrt(e.a*e.a+e.b*e.b),e.l,e.opacity)}(e):new j(e,t,n,null==r?1:r)}function j(e,t,n,r){this.h=+e,this.c=+t,this.l=+n,this.opacity=+r}Object(r.default)(m,h,Object(r.extend)(i.Color,{brighter:function(e){return new m(this.l+18*(null==e?1:e),this.a,this.b,this.opacity)},darker:function(e){return new m(this.l-18*(null==e?1:e),this.a,this.b,this.opacity)},rgb:function(){var e=(this.l+16)/116,t=isNaN(this.a)?e:e+this.a/500,n=isNaN(this.b)?e:e-this.b/200;return e=s*v(e),t=a*v(t),n=u*v(n),new i.Rgb(_(3.2404542*t-1.5371385*e-.4985314*n),_(-.969266*t+1.8760108*e+.041556*n),_(.0556434*t-.2040259*e+1.0572252*n),this.opacity)}})),Object(r.default)(j,b,Object(r.extend)(i.Color,{brighter:function(e){return new j(this.h,this.c,this.l+18*(null==e?1:e),this.opacity)},darker:function(e){return new j(this.h,this.c,this.l-18*(null==e?1:e),this.opacity)},rgb:function(){return f(this).rgb()}}))},"./node_modules/d3-color/src/math.js":function(e,t,n){"use strict";n.r(t),n.d(t,"deg2rad",function(){return r}),n.d(t,"rad2deg",function(){return i});var r=Math.PI/180,i=180/Math.PI},"./node_modules/d3-contour/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-contour/src/contours.js");n.d(t,"contours",function(){return r.default});var i=n("./node_modules/d3-contour/src/density.js");n.d(t,"contourDensity",function(){return i.default})},"./node_modules/d3-contour/src/area.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){for(var t=0,n=e.length,r=e[n-1][1]*e[0][0]-e[n-1][0]*e[0][1];++t<n;)r+=e[t-1][1]*e[t][0]-e[t-1][0]*e[t][1];return r}},"./node_modules/d3-contour/src/array.js":function(e,t,n){"use strict";n.r(t),n.d(t,"slice",function(){return r});var r=Array.prototype.slice},"./node_modules/d3-contour/src/ascending.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){return e-t}},"./node_modules/d3-contour/src/blur.js":function(e,t,n){"use strict";function r(e,t,n){for(var r=e.width,i=e.height,o=1+(n<<1),a=0;a<i;++a)for(var s=0,u=0;s<r+n;++s)s<r&&(u+=e.data[s+a*r]),s>=n&&(s>=o&&(u-=e.data[s-o+a*r]),t.data[s-n+a*r]=u/Math.min(s+1,r-1+o-s,o))}function i(e,t,n){for(var r=e.width,i=e.height,o=1+(n<<1),a=0;a<r;++a)for(var s=0,u=0;s<i+n;++s)s<i&&(u+=e.data[a+s*r]),s>=n&&(s>=o&&(u-=e.data[a+(s-o)*r]),t.data[a+(s-n)*r]=u/Math.min(s+1,i-1+o-s,o))}n.r(t),n.d(t,"blurX",function(){return r}),n.d(t,"blurY",function(){return i})},"./node_modules/d3-contour/src/constant.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return function(){return e}}},"./node_modules/d3-contour/src/contains.js":function(e,t,n){"use strict";function r(e,t){for(var n,r,i,o,a,s,u,c,l,d,p=t[0],f=t[1],h=-1,m=0,g=e.length,v=g-1;m<g;v=m++){var _=e[m],y=_[0],b=_[1],j=e[v],x=j[0],O=j[1];if(d=void 0,o=u=_,s=l=t,((a=c=j)[0]-o[0])*(s[1]-o[1])==(s[0]-o[0])*(a[1]-o[1])&&(n=u[d=+(u[0]===c[0])],r=l[d],i=c[d],n<=r&&r<=i||i<=r&&r<=n))return 0;b>f!=O>f&&p<(x-y)*(f-b)/(O-b)+y&&(h=-h)}return h}n.r(t),t.default=function(e,t){for(var n,i=-1,o=t.length;++i<o;)if(n=r(e,t[i]))return n;return 0}},"./node_modules/d3-contour/src/contours.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/d3-contour/src/array.js"),o=n("./node_modules/d3-contour/src/ascending.js"),a=n("./node_modules/d3-contour/src/area.js"),s=n("./node_modules/d3-contour/src/constant.js"),u=n("./node_modules/d3-contour/src/contains.js"),c=n("./node_modules/d3-contour/src/noop.js"),l=[[],[[[1,1.5],[.5,1]]],[[[1.5,1],[1,1.5]]],[[[1.5,1],[.5,1]]],[[[1,.5],[1.5,1]]],[[[1,1.5],[.5,1]],[[1,.5],[1.5,1]]],[[[1,.5],[1,1.5]]],[[[1,.5],[.5,1]]],[[[.5,1],[1,.5]]],[[[1,1.5],[1,.5]]],[[[.5,1],[1,.5]],[[1.5,1],[1,1.5]]],[[[1.5,1],[1,.5]]],[[[.5,1],[1.5,1]]],[[[1,1.5],[1.5,1]]],[[[.5,1],[1,1.5]]],[]];t.default=function(){var e=1,t=1,n=r.thresholdSturges,d=h;function p(i){var s=n(i);if(Array.isArray(s))s=s.slice().sort(o.default);else{var c=Object(r.extent)(i),p=c[0],h=c[1];s=Object(r.tickStep)(p,h,s),s=Object(r.range)(Math.floor(p/s)*s,Math.floor(h/s)*s,s)}return s.map(function(n){var r=[],o=[];return function(n,r,i){var o,a,s,u,c,d,p=new Array,h=new Array;o=a=-1,u=n[0]>=r,l[u<<1].forEach(m);for(;++o<e-1;)s=u,u=n[o+1]>=r,l[s|u<<1].forEach(m);l[u<<0].forEach(m);for(;++a<t-1;){for(o=-1,u=n[a*e+e]>=r,c=n[a*e]>=r,l[u<<1|c<<2].forEach(m);++o<e-1;)s=u,u=n[a*e+e+o+1]>=r,d=c,c=n[a*e+o+1]>=r,l[s|u<<1|c<<2|d<<3].forEach(m);l[u|c<<3].forEach(m)}o=-1,c=n[a*e]>=r,l[c<<2].forEach(m);for(;++o<e-1;)d=c,c=n[a*e+o+1]>=r,l[c<<2|d<<3].forEach(m);function m(e){var t,n,r=[e[0][0]+o,e[0][1]+a],s=[e[1][0]+o,e[1][1]+a],u=f(r),c=f(s);(t=h[u])?(n=p[c])?(delete h[t.end],delete p[n.start],t===n?(t.ring.push(s),i(t.ring)):p[t.start]=h[n.end]={start:t.start,end:n.end,ring:t.ring.concat(n.ring)}):(delete h[t.end],t.ring.push(s),h[t.end=c]=t):(t=p[c])?(n=h[u])?(delete p[t.start],delete h[n.end],t===n?(t.ring.push(s),i(t.ring)):p[n.start]=h[t.end]={start:n.start,end:t.end,ring:n.ring.concat(t.ring)}):(delete p[t.start],t.ring.unshift(r),p[t.start=u]=t):p[u]=h[c]={start:u,end:c,ring:[r,s]}}l[c<<3].forEach(m)}(i,n,function(e){d(e,i,n),Object(a.default)(e)>0?r.push([e]):o.push(e)}),o.forEach(function(e){for(var t,n=0,i=r.length;n<i;++n)if(-1!==Object(u.default)((t=r[n])[0],e))return void t.push(e)}),r}).map(function(e,t){return{type:"MultiPolygon",value:s[t],coordinates:e}})}function f(t){return 2*t[0]+t[1]*(e+1)*4}function h(n,r,i){n.forEach(function(n){var o,a=n[0],s=n[1],u=0|a,c=0|s,l=r[c*e+u];a>0&&a<e&&u===a&&(o=r[c*e+u-1],n[0]=a+(i-o)/(l-o)-.5),s>0&&s<t&&c===s&&(o=r[(c-1)*e+u],n[1]=s+(i-o)/(l-o)-.5)})}return p.size=function(n){if(!arguments.length)return[e,t];var r=Math.ceil(n[0]),i=Math.ceil(n[1]);if(!(r>0&&i>0))throw new Error("invalid size");return e=r,t=i,p},p.thresholds=function(e){return arguments.length?(n="function"==typeof e?e:Array.isArray(e)?Object(s.default)(i.slice.call(e)):Object(s.default)(e),p):n},p.smooth=function(e){return arguments.length?(d=e?h:c.default,p):d===h},p}},"./node_modules/d3-contour/src/density.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/d3-contour/src/array.js"),o=n("./node_modules/d3-contour/src/blur.js"),a=n("./node_modules/d3-contour/src/constant.js"),s=n("./node_modules/d3-contour/src/contours.js");function u(e){return e[0]}function c(e){return e[1]}t.default=function(){var e=u,t=c,n=960,l=500,d=20,p=2,f=3*d,h=n+2*f>>p,m=l+2*f>>p,g=Object(a.default)(20);function v(n){var i=new Float32Array(h*m),a=new Float32Array(h*m);n.forEach(function(n,r,o){var a=e(n,r,o)+f>>p,s=t(n,r,o)+f>>p;a>=0&&a<h&&s>=0&&s<m&&++i[a+s*h]}),Object(o.blurX)({width:h,height:m,data:i},{width:h,height:m,data:a},d>>p),Object(o.blurY)({width:h,height:m,data:a},{width:h,height:m,data:i},d>>p),Object(o.blurX)({width:h,height:m,data:i},{width:h,height:m,data:a},d>>p),Object(o.blurY)({width:h,height:m,data:a},{width:h,height:m,data:i},d>>p),Object(o.blurX)({width:h,height:m,data:i},{width:h,height:m,data:a},d>>p),Object(o.blurY)({width:h,height:m,data:a},{width:h,height:m,data:i},d>>p);var u=g(i);if(!Array.isArray(u)){var c=Object(r.max)(i);u=Object(r.tickStep)(0,c,u),(u=Object(r.range)(0,Math.floor(c/u)*u,u)).shift()}return Object(s.default)().thresholds(u).size([h,m])(i).map(_)}function _(e){return e.value*=Math.pow(2,-2*p),e.coordinates.forEach(y),e}function y(e){e.forEach(b)}function b(e){e.forEach(j)}function j(e){e[0]=e[0]*Math.pow(2,p)-f,e[1]=e[1]*Math.pow(2,p)-f}function x(){return h=n+2*(f=3*d)>>p,m=l+2*f>>p,v}return v.x=function(t){return arguments.length?(e="function"==typeof t?t:Object(a.default)(+t),v):e},v.y=function(e){return arguments.length?(t="function"==typeof e?e:Object(a.default)(+e),v):t},v.size=function(e){if(!arguments.length)return[n,l];var t=Math.ceil(e[0]),r=Math.ceil(e[1]);if(!(t>=0||t>=0))throw new Error("invalid size");return n=t,l=r,x()},v.cellSize=function(e){if(!arguments.length)return 1<<p;if(!((e=+e)>=1))throw new Error("invalid cell size");return p=Math.floor(Math.log(e)/Math.LN2),x()},v.thresholds=function(e){return arguments.length?(g="function"==typeof e?e:Array.isArray(e)?Object(a.default)(i.slice.call(e)):Object(a.default)(e),v):g},v.bandwidth=function(e){if(!arguments.length)return Math.sqrt(d*(d+1));if(!((e=+e)>=0))throw new Error("invalid bandwidth");return d=Math.round((Math.sqrt(4*e*e+1)-1)/2),x()},v}},"./node_modules/d3-contour/src/noop.js":function(e,t,n){"use strict";n.r(t),t.default=function(){}},"./node_modules/d3-dispatch/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-dispatch/src/dispatch.js");n.d(t,"dispatch",function(){return r.default})},"./node_modules/d3-dispatch/src/dispatch.js":function(e,t,n){"use strict";n.r(t);var r={value:function(){}};function i(){for(var e,t=0,n=arguments.length,r={};t<n;++t){if(!(e=arguments[t]+"")||e in r)throw new Error("illegal type: "+e);r[e]=[]}return new o(r)}function o(e){this._=e}function a(e,t){for(var n,r=0,i=e.length;r<i;++r)if((n=e[r]).name===t)return n.value}function s(e,t,n){for(var i=0,o=e.length;i<o;++i)if(e[i].name===t){e[i]=r,e=e.slice(0,i).concat(e.slice(i+1));break}return null!=n&&e.push({name:t,value:n}),e}o.prototype=i.prototype={constructor:o,on:function(e,t){var n,r,i=this._,o=(r=i,(e+"").trim().split(/^|\s+/).map(function(e){var t="",n=e.indexOf(".");if(n>=0&&(t=e.slice(n+1),e=e.slice(0,n)),e&&!r.hasOwnProperty(e))throw new Error("unknown type: "+e);return{type:e,name:t}})),u=-1,c=o.length;if(!(arguments.length<2)){if(null!=t&&"function"!=typeof t)throw new Error("invalid callback: "+t);for(;++u<c;)if(n=(e=o[u]).type)i[n]=s(i[n],e.name,t);else if(null==t)for(n in i)i[n]=s(i[n],e.name,null);return this}for(;++u<c;)if((n=(e=o[u]).type)&&(n=a(i[n],e.name)))return n},copy:function(){var e={},t=this._;for(var n in t)e[n]=t[n].slice();return new o(e)},call:function(e,t){if((n=arguments.length-2)>0)for(var n,r,i=new Array(n),o=0;o<n;++o)i[o]=arguments[o+2];if(!this._.hasOwnProperty(e))throw new Error("unknown type: "+e);for(o=0,n=(r=this._[e]).length;o<n;++o)r[o].value.apply(t,i)},apply:function(e,t,n){if(!this._.hasOwnProperty(e))throw new Error("unknown type: "+e);for(var r=this._[e],i=0,o=r.length;i<o;++i)r[i].value.apply(t,n)}},t.default=i},"./node_modules/d3-dsv/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-dsv/src/dsv.js");n.d(t,"dsvFormat",function(){return r.default});var i=n("./node_modules/d3-dsv/src/csv.js");n.d(t,"csvParse",function(){return i.csvParse}),n.d(t,"csvParseRows",function(){return i.csvParseRows}),n.d(t,"csvFormat",function(){return i.csvFormat}),n.d(t,"csvFormatRows",function(){return i.csvFormatRows});var o=n("./node_modules/d3-dsv/src/tsv.js");n.d(t,"tsvParse",function(){return o.tsvParse}),n.d(t,"tsvParseRows",function(){return o.tsvParseRows}),n.d(t,"tsvFormat",function(){return o.tsvFormat}),n.d(t,"tsvFormatRows",function(){return o.tsvFormatRows})},"./node_modules/d3-dsv/src/csv.js":function(e,t,n){"use strict";n.r(t),n.d(t,"csvParse",function(){return o}),n.d(t,"csvParseRows",function(){return a}),n.d(t,"csvFormat",function(){return s}),n.d(t,"csvFormatRows",function(){return u});var r=n("./node_modules/d3-dsv/src/dsv.js"),i=Object(r.default)(","),o=i.parse,a=i.parseRows,s=i.format,u=i.formatRows},"./node_modules/d3-dsv/src/dsv.js":function(e,t,n){"use strict";n.r(t);var r={},i={},o=34,a=10,s=13;function u(e){return new Function("d","return {"+e.map(function(e,t){return JSON.stringify(e)+": d["+t+"]"}).join(",")+"}")}t.default=function(e){var t=new RegExp('["'+e+"\n\r]"),n=e.charCodeAt(0);function c(e,t){var u,c=[],l=e.length,d=0,p=0,f=l<=0,h=!1;function m(){if(f)return i;if(h)return h=!1,r;var t,u,c=d;if(e.charCodeAt(c)===o){for(;d++<l&&e.charCodeAt(d)!==o||e.charCodeAt(++d)===o;);return(t=d)>=l?f=!0:(u=e.charCodeAt(d++))===a?h=!0:u===s&&(h=!0,e.charCodeAt(d)===a&&++d),e.slice(c+1,t-1).replace(/""/g,'"')}for(;d<l;){if((u=e.charCodeAt(t=d++))===a)h=!0;else if(u===s)h=!0,e.charCodeAt(d)===a&&++d;else if(u!==n)continue;return e.slice(c,t)}return f=!0,e.slice(c,l)}for(e.charCodeAt(l-1)===a&&--l,e.charCodeAt(l-1)===s&&--l;(u=m())!==i;){for(var g=[];u!==r&&u!==i;)g.push(u),u=m();t&&null==(g=t(g,p++))||c.push(g)}return c}function l(t){return t.map(d).join(e)}function d(e){return null==e?"":t.test(e+="")?'"'+e.replace(/"/g,'""')+'"':e}return{parse:function(e,t){var n,r,i=c(e,function(e,i){if(n)return n(e,i-1);var o,a,s;r=e,n=t?(a=t,s=u(o=e),function(e,t){return a(s(e),t,o)}):u(e)});return i.columns=r||[],i},parseRows:c,format:function(t,n){var r,i,o;return null==n&&(r=t,i=Object.create(null),o=[],r.forEach(function(e){for(var t in e)t in i||o.push(i[t]=t)}),n=o),[n.map(d).join(e)].concat(t.map(function(t){return n.map(function(e){return d(t[e])}).join(e)})).join("\n")},formatRows:function(e){return e.map(l).join("\n")}}}},"./node_modules/d3-dsv/src/tsv.js":function(e,t,n){"use strict";n.r(t),n.d(t,"tsvParse",function(){return o}),n.d(t,"tsvParseRows",function(){return a}),n.d(t,"tsvFormat",function(){return s}),n.d(t,"tsvFormatRows",function(){return u});var r=n("./node_modules/d3-dsv/src/dsv.js"),i=Object(r.default)("\t"),o=i.parse,a=i.parseRows,s=i.format,u=i.formatRows},"./node_modules/d3-force/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-force/src/center.js");n.d(t,"forceCenter",function(){return r.default});var i=n("./node_modules/d3-force/src/collide.js");n.d(t,"forceCollide",function(){return i.default});var o=n("./node_modules/d3-force/src/link.js");n.d(t,"forceLink",function(){return o.default});var a=n("./node_modules/d3-force/src/manyBody.js");n.d(t,"forceManyBody",function(){return a.default});var s=n("./node_modules/d3-force/src/radial.js");n.d(t,"forceRadial",function(){return s.default});var u=n("./node_modules/d3-force/src/simulation.js");n.d(t,"forceSimulation",function(){return u.default});var c=n("./node_modules/d3-force/src/x.js");n.d(t,"forceX",function(){return c.default});var l=n("./node_modules/d3-force/src/y.js");n.d(t,"forceY",function(){return l.default})},"./node_modules/d3-force/src/center.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){var n;function r(){var r,i,o=n.length,a=0,s=0;for(r=0;r<o;++r)a+=(i=n[r]).x,s+=i.y;for(a=a/o-e,s=s/o-t,r=0;r<o;++r)(i=n[r]).x-=a,i.y-=s}return null==e&&(e=0),null==t&&(t=0),r.initialize=function(e){n=e},r.x=function(t){return arguments.length?(e=+t,r):e},r.y=function(e){return arguments.length?(t=+e,r):t},r}},"./node_modules/d3-force/src/collide.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-force/src/constant.js"),i=n("./node_modules/d3-force/src/jiggle.js"),o=n("./node_modules/d3-quadtree/index.js");function a(e){return e.x+e.vx}function s(e){return e.y+e.vy}t.default=function(e){var t,n,u=1,c=1;function l(){for(var e,r,l,p,f,h,m,g=t.length,v=0;v<c;++v)for(r=Object(o.quadtree)(t,a,s).visitAfter(d),e=0;e<g;++e)l=t[e],h=n[l.index],m=h*h,p=l.x+l.vx,f=l.y+l.vy,r.visit(_);function _(e,t,n,r,o){var a=e.data,s=e.r,c=h+s;if(!a)return t>p+c||r<p-c||n>f+c||o<f-c;if(a.index>l.index){var d=p-a.x-a.vx,g=f-a.y-a.vy,v=d*d+g*g;v<c*c&&(0===d&&(v+=(d=Object(i.default)())*d),0===g&&(v+=(g=Object(i.default)())*g),v=(c-(v=Math.sqrt(v)))/v*u,l.vx+=(d*=v)*(c=(s*=s)/(m+s)),l.vy+=(g*=v)*c,a.vx-=d*(c=1-c),a.vy-=g*c)}}}function d(e){if(e.data)return e.r=n[e.data.index];for(var t=e.r=0;t<4;++t)e[t]&&e[t].r>e.r&&(e.r=e[t].r)}function p(){if(t){var r,i,o=t.length;for(n=new Array(o),r=0;r<o;++r)i=t[r],n[i.index]=+e(i,r,t)}}return"function"!=typeof e&&(e=Object(r.default)(null==e?1:+e)),l.initialize=function(e){t=e,p()},l.iterations=function(e){return arguments.length?(c=+e,l):c},l.strength=function(e){return arguments.length?(u=+e,l):u},l.radius=function(t){return arguments.length?(e="function"==typeof t?t:Object(r.default)(+t),p(),l):e},l}},"./node_modules/d3-force/src/constant.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return function(){return e}}},"./node_modules/d3-force/src/jiggle.js":function(e,t,n){"use strict";n.r(t),t.default=function(){return 1e-6*(Math.random()-.5)}},"./node_modules/d3-force/src/link.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-force/src/constant.js"),i=n("./node_modules/d3-force/src/jiggle.js"),o=n("./node_modules/d3-collection/index.js");function a(e){return e.index}function s(e,t){var n=e.get(t);if(!n)throw new Error("missing: "+t);return n}t.default=function(e){var t,n,u,c,l,d=a,p=function(e){return 1/Math.min(c[e.source.index],c[e.target.index])},f=Object(r.default)(30),h=1;function m(r){for(var o=0,a=e.length;o<h;++o)for(var s,u,c,d,p,f,m,g=0;g<a;++g)u=(s=e[g]).source,d=(c=s.target).x+c.vx-u.x-u.vx||Object(i.default)(),p=c.y+c.vy-u.y-u.vy||Object(i.default)(),d*=f=((f=Math.sqrt(d*d+p*p))-n[g])/f*r*t[g],p*=f,c.vx-=d*(m=l[g]),c.vy-=p*m,u.vx+=d*(m=1-m),u.vy+=p*m}function g(){if(u){var r,i,a=u.length,p=e.length,f=Object(o.map)(u,d);for(r=0,c=new Array(a);r<p;++r)(i=e[r]).index=r,"object"!=typeof i.source&&(i.source=s(f,i.source)),"object"!=typeof i.target&&(i.target=s(f,i.target)),c[i.source.index]=(c[i.source.index]||0)+1,c[i.target.index]=(c[i.target.index]||0)+1;for(r=0,l=new Array(p);r<p;++r)i=e[r],l[r]=c[i.source.index]/(c[i.source.index]+c[i.target.index]);t=new Array(p),v(),n=new Array(p),_()}}function v(){if(u)for(var n=0,r=e.length;n<r;++n)t[n]=+p(e[n],n,e)}function _(){if(u)for(var t=0,r=e.length;t<r;++t)n[t]=+f(e[t],t,e)}return null==e&&(e=[]),m.initialize=function(e){u=e,g()},m.links=function(t){return arguments.length?(e=t,g(),m):e},m.id=function(e){return arguments.length?(d=e,m):d},m.iterations=function(e){return arguments.length?(h=+e,m):h},m.strength=function(e){return arguments.length?(p="function"==typeof e?e:Object(r.default)(+e),v(),m):p},m.distance=function(e){return arguments.length?(f="function"==typeof e?e:Object(r.default)(+e),_(),m):f},m}},"./node_modules/d3-force/src/manyBody.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-force/src/constant.js"),i=n("./node_modules/d3-force/src/jiggle.js"),o=n("./node_modules/d3-quadtree/index.js"),a=n("./node_modules/d3-force/src/simulation.js");t.default=function(){var e,t,n,s,u=Object(r.default)(-30),c=1,l=1/0,d=.81;function p(r){var i,s=e.length,u=Object(o.quadtree)(e,a.x,a.y).visitAfter(h);for(n=r,i=0;i<s;++i)t=e[i],u.visit(m)}function f(){if(e){var t,n,r=e.length;for(s=new Array(r),t=0;t<r;++t)n=e[t],s[n.index]=+u(n,t,e)}}function h(e){var t,n,r,i,o,a=0,u=0;if(e.length){for(r=i=o=0;o<4;++o)(t=e[o])&&(n=Math.abs(t.value))&&(a+=t.value,u+=n,r+=n*t.x,i+=n*t.y);e.x=r/u,e.y=i/u}else{(t=e).x=t.data.x,t.y=t.data.y;do{a+=s[t.data.index]}while(t=t.next)}e.value=a}function m(e,r,o,a){if(!e.value)return!0;var u=e.x-t.x,p=e.y-t.y,f=a-r,h=u*u+p*p;if(f*f/d<h)return h<l&&(0===u&&(h+=(u=Object(i.default)())*u),0===p&&(h+=(p=Object(i.default)())*p),h<c&&(h=Math.sqrt(c*h)),t.vx+=u*e.value*n/h,t.vy+=p*e.value*n/h),!0;if(!(e.length||h>=l)){(e.data!==t||e.next)&&(0===u&&(h+=(u=Object(i.default)())*u),0===p&&(h+=(p=Object(i.default)())*p),h<c&&(h=Math.sqrt(c*h)));do{e.data!==t&&(f=s[e.data.index]*n/h,t.vx+=u*f,t.vy+=p*f)}while(e=e.next)}}return p.initialize=function(t){e=t,f()},p.strength=function(e){return arguments.length?(u="function"==typeof e?e:Object(r.default)(+e),f(),p):u},p.distanceMin=function(e){return arguments.length?(c=e*e,p):Math.sqrt(c)},p.distanceMax=function(e){return arguments.length?(l=e*e,p):Math.sqrt(l)},p.theta=function(e){return arguments.length?(d=e*e,p):Math.sqrt(d)},p}},"./node_modules/d3-force/src/radial.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-force/src/constant.js");t.default=function(e,t,n){var i,o,a,s=Object(r.default)(.1);function u(e){for(var r=0,s=i.length;r<s;++r){var u=i[r],c=u.x-t||1e-6,l=u.y-n||1e-6,d=Math.sqrt(c*c+l*l),p=(a[r]-d)*o[r]*e/d;u.vx+=c*p,u.vy+=l*p}}function c(){if(i){var t,n=i.length;for(o=new Array(n),a=new Array(n),t=0;t<n;++t)a[t]=+e(i[t],t,i),o[t]=isNaN(a[t])?0:+s(i[t],t,i)}}return"function"!=typeof e&&(e=Object(r.default)(+e)),null==t&&(t=0),null==n&&(n=0),u.initialize=function(e){i=e,c()},u.strength=function(e){return arguments.length?(s="function"==typeof e?e:Object(r.default)(+e),c(),u):s},u.radius=function(t){return arguments.length?(e="function"==typeof t?t:Object(r.default)(+t),c(),u):e},u.x=function(e){return arguments.length?(t=+e,u):t},u.y=function(e){return arguments.length?(n=+e,u):n},u}},"./node_modules/d3-force/src/simulation.js":function(e,t,n){"use strict";n.r(t),n.d(t,"x",function(){return a}),n.d(t,"y",function(){return s});var r=n("./node_modules/d3-dispatch/index.js"),i=n("./node_modules/d3-collection/index.js"),o=n("./node_modules/d3-timer/index.js");function a(e){return e.x}function s(e){return e.y}var u=10,c=Math.PI*(3-Math.sqrt(5));t.default=function(e){var t,n=1,a=.001,s=1-Math.pow(a,1/300),l=0,d=.6,p=Object(i.map)(),f=Object(o.timer)(m),h=Object(r.dispatch)("tick","end");function m(){g(),h.call("tick",t),n<a&&(f.stop(),h.call("end",t))}function g(){var t,r,i=e.length;for(n+=(l-n)*s,p.each(function(e){e(n)}),t=0;t<i;++t)null==(r=e[t]).fx?r.x+=r.vx*=d:(r.x=r.fx,r.vx=0),null==r.fy?r.y+=r.vy*=d:(r.y=r.fy,r.vy=0)}function v(){for(var t,n=0,r=e.length;n<r;++n){if((t=e[n]).index=n,isNaN(t.x)||isNaN(t.y)){var i=u*Math.sqrt(n),o=n*c;t.x=i*Math.cos(o),t.y=i*Math.sin(o)}(isNaN(t.vx)||isNaN(t.vy))&&(t.vx=t.vy=0)}}function _(t){return t.initialize&&t.initialize(e),t}return null==e&&(e=[]),v(),t={tick:g,restart:function(){return f.restart(m),t},stop:function(){return f.stop(),t},nodes:function(n){return arguments.length?(e=n,v(),p.each(_),t):e},alpha:function(e){return arguments.length?(n=+e,t):n},alphaMin:function(e){return arguments.length?(a=+e,t):a},alphaDecay:function(e){return arguments.length?(s=+e,t):+s},alphaTarget:function(e){return arguments.length?(l=+e,t):l},velocityDecay:function(e){return arguments.length?(d=1-e,t):1-d},force:function(e,n){return arguments.length>1?(null==n?p.remove(e):p.set(e,_(n)),t):p.get(e)},find:function(t,n,r){var i,o,a,s,u,c=0,l=e.length;for(null==r?r=1/0:r*=r,c=0;c<l;++c)(a=(i=t-(s=e[c]).x)*i+(o=n-s.y)*o)<r&&(u=s,r=a);return u},on:function(e,n){return arguments.length>1?(h.on(e,n),t):h.on(e)}}}},"./node_modules/d3-force/src/x.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-force/src/constant.js");t.default=function(e){var t,n,i,o=Object(r.default)(.1);function a(e){for(var r,o=0,a=t.length;o<a;++o)(r=t[o]).vx+=(i[o]-r.x)*n[o]*e}function s(){if(t){var r,a=t.length;for(n=new Array(a),i=new Array(a),r=0;r<a;++r)n[r]=isNaN(i[r]=+e(t[r],r,t))?0:+o(t[r],r,t)}}return"function"!=typeof e&&(e=Object(r.default)(null==e?0:+e)),a.initialize=function(e){t=e,s()},a.strength=function(e){return arguments.length?(o="function"==typeof e?e:Object(r.default)(+e),s(),a):o},a.x=function(t){return arguments.length?(e="function"==typeof t?t:Object(r.default)(+t),s(),a):e},a}},"./node_modules/d3-force/src/y.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-force/src/constant.js");t.default=function(e){var t,n,i,o=Object(r.default)(.1);function a(e){for(var r,o=0,a=t.length;o<a;++o)(r=t[o]).vy+=(i[o]-r.y)*n[o]*e}function s(){if(t){var r,a=t.length;for(n=new Array(a),i=new Array(a),r=0;r<a;++r)n[r]=isNaN(i[r]=+e(t[r],r,t))?0:+o(t[r],r,t)}}return"function"!=typeof e&&(e=Object(r.default)(null==e?0:+e)),a.initialize=function(e){t=e,s()},a.strength=function(e){return arguments.length?(o="function"==typeof e?e:Object(r.default)(+e),s(),a):o},a.y=function(t){return arguments.length?(e="function"==typeof t?t:Object(r.default)(+t),s(),a):e},a}},"./node_modules/d3-format/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-format/src/defaultLocale.js");n.d(t,"formatDefaultLocale",function(){return r.default}),n.d(t,"format",function(){return r.format}),n.d(t,"formatPrefix",function(){return r.formatPrefix});var i=n("./node_modules/d3-format/src/locale.js");n.d(t,"formatLocale",function(){return i.default});var o=n("./node_modules/d3-format/src/formatSpecifier.js");n.d(t,"formatSpecifier",function(){return o.default});var a=n("./node_modules/d3-format/src/precisionFixed.js");n.d(t,"precisionFixed",function(){return a.default});var s=n("./node_modules/d3-format/src/precisionPrefix.js");n.d(t,"precisionPrefix",function(){return s.default});var u=n("./node_modules/d3-format/src/precisionRound.js");n.d(t,"precisionRound",function(){return u.default})},"./node_modules/d3-format/src/defaultLocale.js":function(e,t,n){"use strict";n.r(t),n.d(t,"format",function(){return i}),n.d(t,"formatPrefix",function(){return o}),n.d(t,"default",function(){return s});var r,i,o,a=n("./node_modules/d3-format/src/locale.js");function s(e){return r=Object(a.default)(e),i=r.format,o=r.formatPrefix,r}s({decimal:".",thousands:",",grouping:[3],currency:["$",""]})},"./node_modules/d3-format/src/exponent.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-format/src/formatDecimal.js");t.default=function(e){return(e=Object(r.default)(Math.abs(e)))?e[1]:NaN}},"./node_modules/d3-format/src/formatDecimal.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){if((n=(e=t?e.toExponential(t-1):e.toExponential()).indexOf("e"))<0)return null;var n,r=e.slice(0,n);return[r.length>1?r[0]+r.slice(2):r,+e.slice(n+1)]}},"./node_modules/d3-format/src/formatDefault.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){e=e.toPrecision(t);e:for(var n,r=e.length,i=1,o=-1;i<r;++i)switch(e[i]){case".":o=n=i;break;case"0":0===o&&(o=i),n=i;break;case"e":break e;default:o>0&&(o=0)}return o>0?e.slice(0,o)+e.slice(n+1):e}},"./node_modules/d3-format/src/formatGroup.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){return function(n,r){for(var i=n.length,o=[],a=0,s=e[0],u=0;i>0&&s>0&&(u+s+1>r&&(s=Math.max(1,r-u)),o.push(n.substring(i-=s,i+s)),!((u+=s+1)>r));)s=e[a=(a+1)%e.length];return o.reverse().join(t)}}},"./node_modules/d3-format/src/formatNumerals.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return function(t){return t.replace(/[0-9]/g,function(t){return e[+t]})}}},"./node_modules/d3-format/src/formatPrefixAuto.js":function(e,t,n){"use strict";n.r(t),n.d(t,"prefixExponent",function(){return r});var r,i=n("./node_modules/d3-format/src/formatDecimal.js");t.default=function(e,t){var n=Object(i.default)(e,t);if(!n)return e+"";var o=n[0],a=n[1],s=a-(r=3*Math.max(-8,Math.min(8,Math.floor(a/3))))+1,u=o.length;return s===u?o:s>u?o+new Array(s-u+1).join("0"):s>0?o.slice(0,s)+"."+o.slice(s):"0."+new Array(1-s).join("0")+Object(i.default)(e,Math.max(0,t+s-1))[0]}},"./node_modules/d3-format/src/formatRounded.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-format/src/formatDecimal.js");t.default=function(e,t){var n=Object(r.default)(e,t);if(!n)return e+"";var i=n[0],o=n[1];return o<0?"0."+new Array(-o).join("0")+i:i.length>o+1?i.slice(0,o+1)+"."+i.slice(o+1):i+new Array(o-i.length+2).join("0")}},"./node_modules/d3-format/src/formatSpecifier.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/d3-format/src/formatTypes.js"),i=/^(?:(.)?([<>=^]))?([+\-\( ])?([$#])?(0)?(\d+)?(,)?(\.\d+)?([a-z%])?$/i;function o(e){return new a(e)}function a(e){if(!(t=i.exec(e)))throw new Error("invalid format: "+e);var t,n=t[1]||" ",o=t[2]||">",a=t[3]||"-",s=t[4]||"",u=!!t[5],c=t[6]&&+t[6],l=!!t[7],d=t[8]&&+t[8].slice(1),p=t[9]||"";"n"===p?(l=!0,p="g"):r.default[p]||(p=""),(u||"0"===n&&"="===o)&&(u=!0,n="0",o="="),this.fill=n,this.align=o,this.sign=a,this.symbol=s,this.zero=u,this.width=c,this.comma=l,this.precision=d,this.type=p}o.prototype=a.prototype,a.prototype.toString=function(){return this.fill+this.align+this.sign+this.symbol+(this.zero?"0":"")+(null==this.width?"":Math.max(1,0|this.width))+(this.comma?",":"")+(null==this.precision?"":"."+Math.max(0,0|this.precision))+this.type}},"./node_modules/d3-format/src/formatTypes.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-format/src/formatDefault.js"),i=n("./node_modules/d3-format/src/formatPrefixAuto.js"),o=n("./node_modules/d3-format/src/formatRounded.js");t.default={"":r.default,"%":function(e,t){return(100*e).toFixed(t)},b:function(e){return Math.round(e).toString(2)},c:function(e){return e+""},d:function(e){return Math.round(e).toString(10)},e:function(e,t){return e.toExponential(t)},f:function(e,t){return e.toFixed(t)},g:function(e,t){return e.toPrecision(t)},o:function(e){return Math.round(e).toString(8)},p:function(e,t){return Object(o.default)(100*e,t)},r:o.default,s:i.default,X:function(e){return Math.round(e).toString(16).toUpperCase()},x:function(e){return Math.round(e).toString(16)}}},"./node_modules/d3-format/src/identity.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return e}},"./node_modules/d3-format/src/locale.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-format/src/exponent.js"),i=n("./node_modules/d3-format/src/formatGroup.js"),o=n("./node_modules/d3-format/src/formatNumerals.js"),a=n("./node_modules/d3-format/src/formatSpecifier.js"),s=n("./node_modules/d3-format/src/formatTypes.js"),u=n("./node_modules/d3-format/src/formatPrefixAuto.js"),c=n("./node_modules/d3-format/src/identity.js"),l=["y","z","a","f","p","n","µ","m","","k","M","G","T","P","E","Z","Y"];t.default=function(e){var t=e.grouping&&e.thousands?Object(i.default)(e.grouping,e.thousands):c.default,n=e.currency,d=e.decimal,p=e.numerals?Object(o.default)(e.numerals):c.default,f=e.percent||"%";function h(e){var r=(e=Object(a.default)(e)).fill,i=e.align,o=e.sign,c=e.symbol,h=e.zero,m=e.width,g=e.comma,v=e.precision,_=e.type,y="$"===c?n[0]:"#"===c&&/[boxX]/.test(_)?"0"+_.toLowerCase():"",b="$"===c?n[1]:/[%p]/.test(_)?f:"",j=s.default[_],x=!_||/[defgprs%]/.test(_);function O(e){var n,a,s,c=y,f=b;if("c"===_)f=j(e)+f,e="";else{var O=(e=+e)<0;if(e=j(Math.abs(e),v),O&&0==+e&&(O=!1),c=(O?"("===o?o:"-":"-"===o||"("===o?"":o)+c,f=("s"===_?l[8+u.prefixExponent/3]:"")+f+(O&&"("===o?")":""),x)for(n=-1,a=e.length;++n<a;)if(48>(s=e.charCodeAt(n))||s>57){f=(46===s?d+e.slice(n+1):e.slice(n))+f,e=e.slice(0,n);break}}g&&!h&&(e=t(e,1/0));var w=c.length+e.length+f.length,C=w<m?new Array(m-w+1).join(r):"";switch(g&&h&&(e=t(C+e,C.length?m-f.length:1/0),C=""),i){case"<":e=c+e+f+C;break;case"=":e=c+C+e+f;break;case"^":e=C.slice(0,w=C.length>>1)+c+e+f+C.slice(w);break;default:e=C+c+e+f}return p(e)}return v=null==v?_?6:12:/[gprs]/.test(_)?Math.max(1,Math.min(21,v)):Math.max(0,Math.min(20,v)),O.toString=function(){return e+""},O}return{format:h,formatPrefix:function(e,t){var n=h(((e=Object(a.default)(e)).type="f",e)),i=3*Math.max(-8,Math.min(8,Math.floor(Object(r.default)(t)/3))),o=Math.pow(10,-i),s=l[8+i/3];return function(e){return n(o*e)+s}}}}},"./node_modules/d3-format/src/precisionFixed.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-format/src/exponent.js");t.default=function(e){return Math.max(0,-Object(r.default)(Math.abs(e)))}},"./node_modules/d3-format/src/precisionPrefix.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-format/src/exponent.js");t.default=function(e,t){return Math.max(0,3*Math.max(-8,Math.min(8,Math.floor(Object(r.default)(t)/3)))-Object(r.default)(Math.abs(e)))}},"./node_modules/d3-format/src/precisionRound.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-format/src/exponent.js");t.default=function(e,t){return e=Math.abs(e),t=Math.abs(t)-e,Math.max(0,Object(r.default)(t)-Object(r.default)(e))+1}},"./node_modules/d3-geo/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/area.js");n.d(t,"geoArea",function(){return r.default});var i=n("./node_modules/d3-geo/src/bounds.js");n.d(t,"geoBounds",function(){return i.default});var o=n("./node_modules/d3-geo/src/centroid.js");n.d(t,"geoCentroid",function(){return o.default});var a=n("./node_modules/d3-geo/src/circle.js");n.d(t,"geoCircle",function(){return a.default});var s=n("./node_modules/d3-geo/src/clip/antimeridian.js");n.d(t,"geoClipAntimeridian",function(){return s.default});var u=n("./node_modules/d3-geo/src/clip/circle.js");n.d(t,"geoClipCircle",function(){return u.default});var c=n("./node_modules/d3-geo/src/clip/extent.js");n.d(t,"geoClipExtent",function(){return c.default});var l=n("./node_modules/d3-geo/src/clip/rectangle.js");n.d(t,"geoClipRectangle",function(){return l.default});var d=n("./node_modules/d3-geo/src/contains.js");n.d(t,"geoContains",function(){return d.default});var p=n("./node_modules/d3-geo/src/distance.js");n.d(t,"geoDistance",function(){return p.default});var f=n("./node_modules/d3-geo/src/graticule.js");n.d(t,"geoGraticule",function(){return f.default}),n.d(t,"geoGraticule10",function(){return f.graticule10});var h=n("./node_modules/d3-geo/src/interpolate.js");n.d(t,"geoInterpolate",function(){return h.default});var m=n("./node_modules/d3-geo/src/length.js");n.d(t,"geoLength",function(){return m.default});var g=n("./node_modules/d3-geo/src/path/index.js");n.d(t,"geoPath",function(){return g.default});var v=n("./node_modules/d3-geo/src/projection/albers.js");n.d(t,"geoAlbers",function(){return v.default});var _=n("./node_modules/d3-geo/src/projection/albersUsa.js");n.d(t,"geoAlbersUsa",function(){return _.default});var y=n("./node_modules/d3-geo/src/projection/azimuthalEqualArea.js");n.d(t,"geoAzimuthalEqualArea",function(){return y.default}),n.d(t,"geoAzimuthalEqualAreaRaw",function(){return y.azimuthalEqualAreaRaw});var b=n("./node_modules/d3-geo/src/projection/azimuthalEquidistant.js");n.d(t,"geoAzimuthalEquidistant",function(){return b.default}),n.d(t,"geoAzimuthalEquidistantRaw",function(){return b.azimuthalEquidistantRaw});var j=n("./node_modules/d3-geo/src/projection/conicConformal.js");n.d(t,"geoConicConformal",function(){return j.default}),n.d(t,"geoConicConformalRaw",function(){return j.conicConformalRaw});var x=n("./node_modules/d3-geo/src/projection/conicEqualArea.js");n.d(t,"geoConicEqualArea",function(){return x.default}),n.d(t,"geoConicEqualAreaRaw",function(){return x.conicEqualAreaRaw});var O=n("./node_modules/d3-geo/src/projection/conicEquidistant.js");n.d(t,"geoConicEquidistant",function(){return O.default}),n.d(t,"geoConicEquidistantRaw",function(){return O.conicEquidistantRaw});var w=n("./node_modules/d3-geo/src/projection/equirectangular.js");n.d(t,"geoEquirectangular",function(){return w.default}),n.d(t,"geoEquirectangularRaw",function(){return w.equirectangularRaw});var C=n("./node_modules/d3-geo/src/projection/gnomonic.js");n.d(t,"geoGnomonic",function(){return C.default}),n.d(t,"geoGnomonicRaw",function(){return C.gnomonicRaw});var E=n("./node_modules/d3-geo/src/projection/identity.js");n.d(t,"geoIdentity",function(){return E.default});var S=n("./node_modules/d3-geo/src/projection/index.js");n.d(t,"geoProjection",function(){return S.default}),n.d(t,"geoProjectionMutator",function(){return S.projectionMutator});var k=n("./node_modules/d3-geo/src/projection/mercator.js");n.d(t,"geoMercator",function(){return k.default}),n.d(t,"geoMercatorRaw",function(){return k.mercatorRaw});var M=n("./node_modules/d3-geo/src/projection/naturalEarth1.js");n.d(t,"geoNaturalEarth1",function(){return M.default}),n.d(t,"geoNaturalEarth1Raw",function(){return M.naturalEarth1Raw});var A=n("./node_modules/d3-geo/src/projection/orthographic.js");n.d(t,"geoOrthographic",function(){return A.default}),n.d(t,"geoOrthographicRaw",function(){return A.orthographicRaw});var T=n("./node_modules/d3-geo/src/projection/stereographic.js");n.d(t,"geoStereographic",function(){return T.default}),n.d(t,"geoStereographicRaw",function(){return T.stereographicRaw});var D=n("./node_modules/d3-geo/src/projection/transverseMercator.js");n.d(t,"geoTransverseMercator",function(){return D.default}),n.d(t,"geoTransverseMercatorRaw",function(){return D.transverseMercatorRaw});var I=n("./node_modules/d3-geo/src/rotation.js");n.d(t,"geoRotation",function(){return I.default});var P=n("./node_modules/d3-geo/src/stream.js");n.d(t,"geoStream",function(){return P.default});var R=n("./node_modules/d3-geo/src/transform.js");n.d(t,"geoTransform",function(){return R.default})},"./node_modules/d3-geo/src/adder.js":function(e,t,n){"use strict";function r(){this.reset()}n.r(t),t.default=function(){return new r},r.prototype={constructor:r,reset:function(){this.s=this.t=0},add:function(e){o(i,e,this.t),o(this,i.s,this.s),this.s?this.t+=i.t:this.s=i.t},valueOf:function(){return this.s}};var i=new r;function o(e,t,n){var r=e.s=t+n,i=r-t,o=r-i;e.t=t-o+(n-i)}},"./node_modules/d3-geo/src/area.js":function(e,t,n){"use strict";n.r(t),n.d(t,"areaRingSum",function(){return p}),n.d(t,"areaStream",function(){return h});var r,i,o,a,s,u=n("./node_modules/d3-geo/src/adder.js"),c=n("./node_modules/d3-geo/src/math.js"),l=n("./node_modules/d3-geo/src/noop.js"),d=n("./node_modules/d3-geo/src/stream.js"),p=Object(u.default)(),f=Object(u.default)(),h={point:l.default,lineStart:l.default,lineEnd:l.default,polygonStart:function(){p.reset(),h.lineStart=m,h.lineEnd=g},polygonEnd:function(){var e=+p;f.add(e<0?c.tau+e:e),this.lineStart=this.lineEnd=this.point=l.default},sphere:function(){f.add(c.tau)}};function m(){h.point=v}function g(){_(r,i)}function v(e,t){h.point=_,r=e,i=t,e*=c.radians,t*=c.radians,o=e,a=Object(c.cos)(t=t/2+c.quarterPi),s=Object(c.sin)(t)}function _(e,t){e*=c.radians,t=(t*=c.radians)/2+c.quarterPi;var n=e-o,r=n>=0?1:-1,i=r*n,u=Object(c.cos)(t),l=Object(c.sin)(t),d=s*l,f=a*u+d*Object(c.cos)(i),h=d*r*Object(c.sin)(i);p.add(Object(c.atan2)(h,f)),o=e,a=u,s=l}t.default=function(e){return f.reset(),Object(d.default)(e,h),2*f}},"./node_modules/d3-geo/src/bounds.js":function(e,t,n){"use strict";n.r(t);var r,i,o,a,s,u,c,l,d,p,f=n("./node_modules/d3-geo/src/adder.js"),h=n("./node_modules/d3-geo/src/area.js"),m=n("./node_modules/d3-geo/src/cartesian.js"),g=n("./node_modules/d3-geo/src/math.js"),v=n("./node_modules/d3-geo/src/stream.js"),_=Object(f.default)(),y={point:b,lineStart:x,lineEnd:O,polygonStart:function(){y.point=w,y.lineStart=C,y.lineEnd=E,_.reset(),h.areaStream.polygonStart()},polygonEnd:function(){h.areaStream.polygonEnd(),y.point=b,y.lineStart=x,y.lineEnd=O,h.areaRingSum<0?(r=-(o=180),i=-(a=90)):_>g.epsilon?a=90:_<-g.epsilon&&(i=-90),p[0]=r,p[1]=o}};function b(e,t){d.push(p=[r=e,o=e]),t<i&&(i=t),t>a&&(a=t)}function j(e,t){var n=Object(m.cartesian)([e*g.radians,t*g.radians]);if(l){var u=Object(m.cartesianCross)(l,n),c=[u[1],-u[0],0],f=Object(m.cartesianCross)(c,u);Object(m.cartesianNormalizeInPlace)(f),f=Object(m.spherical)(f);var h,v=e-s,_=v>0?1:-1,y=f[0]*g.degrees*_,b=Object(g.abs)(v)>180;b^(_*s<y&&y<_*e)?(h=f[1]*g.degrees)>a&&(a=h):b^(_*s<(y=(y+360)%360-180)&&y<_*e)?(h=-f[1]*g.degrees)<i&&(i=h):(t<i&&(i=t),t>a&&(a=t)),b?e<s?S(r,e)>S(r,o)&&(o=e):S(e,o)>S(r,o)&&(r=e):o>=r?(e<r&&(r=e),e>o&&(o=e)):e>s?S(r,e)>S(r,o)&&(o=e):S(e,o)>S(r,o)&&(r=e)}else d.push(p=[r=e,o=e]);t<i&&(i=t),t>a&&(a=t),l=n,s=e}function x(){y.point=j}function O(){p[0]=r,p[1]=o,y.point=b,l=null}function w(e,t){if(l){var n=e-s;_.add(Object(g.abs)(n)>180?n+(n>0?360:-360):n)}else u=e,c=t;h.areaStream.point(e,t),j(e,t)}function C(){h.areaStream.lineStart()}function E(){w(u,c),h.areaStream.lineEnd(),Object(g.abs)(_)>g.epsilon&&(r=-(o=180)),p[0]=r,p[1]=o,l=null}function S(e,t){return(t-=e)<0?t+360:t}function k(e,t){return e[0]-t[0]}function M(e,t){return e[0]<=e[1]?e[0]<=t&&t<=e[1]:t<e[0]||e[1]<t}t.default=function(e){var t,n,s,u,c,l,f;if(a=o=-(r=i=1/0),d=[],Object(v.default)(e,y),n=d.length){for(d.sort(k),t=1,c=[s=d[0]];t<n;++t)M(s,(u=d[t])[0])||M(s,u[1])?(S(s[0],u[1])>S(s[0],s[1])&&(s[1]=u[1]),S(u[0],s[1])>S(s[0],s[1])&&(s[0]=u[0])):c.push(s=u);for(l=-1/0,t=0,s=c[n=c.length-1];t<=n;s=u,++t)u=c[t],(f=S(s[1],u[0]))>l&&(l=f,r=u[0],o=s[1])}return d=p=null,r===1/0||i===1/0?[[NaN,NaN],[NaN,NaN]]:[[r,i],[o,a]]}},"./node_modules/d3-geo/src/cartesian.js":function(e,t,n){"use strict";n.r(t),n.d(t,"spherical",function(){return i}),n.d(t,"cartesian",function(){return o}),n.d(t,"cartesianDot",function(){return a}),n.d(t,"cartesianCross",function(){return s}),n.d(t,"cartesianAddInPlace",function(){return u}),n.d(t,"cartesianScale",function(){return c}),n.d(t,"cartesianNormalizeInPlace",function(){return l});var r=n("./node_modules/d3-geo/src/math.js");function i(e){return[Object(r.atan2)(e[1],e[0]),Object(r.asin)(e[2])]}function o(e){var t=e[0],n=e[1],i=Object(r.cos)(n);return[i*Object(r.cos)(t),i*Object(r.sin)(t),Object(r.sin)(n)]}function a(e,t){return e[0]*t[0]+e[1]*t[1]+e[2]*t[2]}function s(e,t){return[e[1]*t[2]-e[2]*t[1],e[2]*t[0]-e[0]*t[2],e[0]*t[1]-e[1]*t[0]]}function u(e,t){e[0]+=t[0],e[1]+=t[1],e[2]+=t[2]}function c(e,t){return[e[0]*t,e[1]*t,e[2]*t]}function l(e){var t=Object(r.sqrt)(e[0]*e[0]+e[1]*e[1]+e[2]*e[2]);e[0]/=t,e[1]/=t,e[2]/=t}},"./node_modules/d3-geo/src/centroid.js":function(e,t,n){"use strict";n.r(t);var r,i,o,a,s,u,c,l,d,p,f,h,m,g,v,_,y=n("./node_modules/d3-geo/src/math.js"),b=n("./node_modules/d3-geo/src/noop.js"),j=n("./node_modules/d3-geo/src/stream.js"),x={sphere:b.default,point:O,lineStart:C,lineEnd:k,polygonStart:function(){x.lineStart=M,x.lineEnd=A},polygonEnd:function(){x.lineStart=C,x.lineEnd=k}};function O(e,t){e*=y.radians,t*=y.radians;var n=Object(y.cos)(t);w(n*Object(y.cos)(e),n*Object(y.sin)(e),Object(y.sin)(t))}function w(e,t,n){o+=(e-o)/++r,a+=(t-a)/r,s+=(n-s)/r}function C(){x.point=E}function E(e,t){e*=y.radians,t*=y.radians;var n=Object(y.cos)(t);g=n*Object(y.cos)(e),v=n*Object(y.sin)(e),_=Object(y.sin)(t),x.point=S,w(g,v,_)}function S(e,t){e*=y.radians,t*=y.radians;var n=Object(y.cos)(t),r=n*Object(y.cos)(e),o=n*Object(y.sin)(e),a=Object(y.sin)(t),s=Object(y.atan2)(Object(y.sqrt)((s=v*a-_*o)*s+(s=_*r-g*a)*s+(s=g*o-v*r)*s),g*r+v*o+_*a);i+=s,u+=s*(g+(g=r)),c+=s*(v+(v=o)),l+=s*(_+(_=a)),w(g,v,_)}function k(){x.point=O}function M(){x.point=T}function A(){D(h,m),x.point=O}function T(e,t){h=e,m=t,e*=y.radians,t*=y.radians,x.point=D;var n=Object(y.cos)(t);g=n*Object(y.cos)(e),v=n*Object(y.sin)(e),_=Object(y.sin)(t),w(g,v,_)}function D(e,t){e*=y.radians,t*=y.radians;var n=Object(y.cos)(t),r=n*Object(y.cos)(e),o=n*Object(y.sin)(e),a=Object(y.sin)(t),s=v*a-_*o,h=_*r-g*a,m=g*o-v*r,b=Object(y.sqrt)(s*s+h*h+m*m),j=Object(y.asin)(b),x=b&&-j/b;d+=x*s,p+=x*h,f+=x*m,i+=j,u+=j*(g+(g=r)),c+=j*(v+(v=o)),l+=j*(_+(_=a)),w(g,v,_)}t.default=function(e){r=i=o=a=s=u=c=l=d=p=f=0,Object(j.default)(e,x);var t=d,n=p,h=f,m=t*t+n*n+h*h;return m<y.epsilon2&&(t=u,n=c,h=l,i<y.epsilon&&(t=o,n=a,h=s),(m=t*t+n*n+h*h)<y.epsilon2)?[NaN,NaN]:[Object(y.atan2)(n,t)*y.degrees,Object(y.asin)(h/Object(y.sqrt)(m))*y.degrees]}},"./node_modules/d3-geo/src/circle.js":function(e,t,n){"use strict";n.r(t),n.d(t,"circleStream",function(){return s});var r=n("./node_modules/d3-geo/src/cartesian.js"),i=n("./node_modules/d3-geo/src/constant.js"),o=n("./node_modules/d3-geo/src/math.js"),a=n("./node_modules/d3-geo/src/rotation.js");function s(e,t,n,i,a,s){if(n){var c=Object(o.cos)(t),l=Object(o.sin)(t),d=i*n;null==a?(a=t+i*o.tau,s=t-d/2):(a=u(c,a),s=u(c,s),(i>0?a<s:a>s)&&(a+=i*o.tau));for(var p,f=a;i>0?f>s:f<s;f-=d)p=Object(r.spherical)([c,-l*Object(o.cos)(f),-l*Object(o.sin)(f)]),e.point(p[0],p[1])}}function u(e,t){(t=Object(r.cartesian)(t))[0]-=e,Object(r.cartesianNormalizeInPlace)(t);var n=Object(o.acos)(-t[1]);return((-t[2]<0?-n:n)+o.tau-o.epsilon)%o.tau}t.default=function(){var e,t,n=Object(i.default)([0,0]),r=Object(i.default)(90),u=Object(i.default)(6),c={point:function(n,r){e.push(n=t(n,r)),n[0]*=o.degrees,n[1]*=o.degrees}};function l(){var i=n.apply(this,arguments),l=r.apply(this,arguments)*o.radians,d=u.apply(this,arguments)*o.radians;return e=[],t=Object(a.rotateRadians)(-i[0]*o.radians,-i[1]*o.radians,0).invert,s(c,l,d,1),i={type:"Polygon",coordinates:[e]},e=t=null,i}return l.center=function(e){return arguments.length?(n="function"==typeof e?e:Object(i.default)([+e[0],+e[1]]),l):n},l.radius=function(e){return arguments.length?(r="function"==typeof e?e:Object(i.default)(+e),l):r},l.precision=function(e){return arguments.length?(u="function"==typeof e?e:Object(i.default)(+e),l):u},l}},"./node_modules/d3-geo/src/clip/antimeridian.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/clip/index.js"),i=n("./node_modules/d3-geo/src/math.js");t.default=Object(r.default)(function(){return!0},function(e){var t,n=NaN,r=NaN,o=NaN;return{lineStart:function(){e.lineStart(),t=1},point:function(a,s){var u,c,l,d,p,f,h,m=a>0?i.pi:-i.pi,g=Object(i.abs)(a-n);Object(i.abs)(g-i.pi)<i.epsilon?(e.point(n,r=(r+s)/2>0?i.halfPi:-i.halfPi),e.point(o,r),e.lineEnd(),e.lineStart(),e.point(m,r),e.point(a,r),t=0):o!==m&&g>=i.pi&&(Object(i.abs)(n-o)<i.epsilon&&(n-=o*i.epsilon),Object(i.abs)(a-m)<i.epsilon&&(a-=m*i.epsilon),u=n,c=r,l=a,d=s,h=Object(i.sin)(u-l),r=Object(i.abs)(h)>i.epsilon?Object(i.atan)((Object(i.sin)(c)*(f=Object(i.cos)(d))*Object(i.sin)(l)-Object(i.sin)(d)*(p=Object(i.cos)(c))*Object(i.sin)(u))/(p*f*h)):(c+d)/2,e.point(o,r),e.lineEnd(),e.lineStart(),e.point(m,r),t=0),e.point(n=a,r=s),o=m},lineEnd:function(){e.lineEnd(),n=r=NaN},clean:function(){return 2-t}}},function(e,t,n,r){var o;if(null==e)o=n*i.halfPi,r.point(-i.pi,o),r.point(0,o),r.point(i.pi,o),r.point(i.pi,0),r.point(i.pi,-o),r.point(0,-o),r.point(-i.pi,-o),r.point(-i.pi,0),r.point(-i.pi,o);else if(Object(i.abs)(e[0]-t[0])>i.epsilon){var a=e[0]<t[0]?i.pi:-i.pi;o=n*a/2,r.point(-a,o),r.point(0,o),r.point(a,o)}else r.point(t[0],t[1])},[-i.pi,-i.halfPi])},"./node_modules/d3-geo/src/clip/buffer.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/noop.js");t.default=function(){var e,t=[];return{point:function(t,n){e.push([t,n])},lineStart:function(){t.push(e=[])},lineEnd:r.default,rejoin:function(){t.length>1&&t.push(t.pop().concat(t.shift()))},result:function(){var n=t;return t=[],e=null,n}}}},"./node_modules/d3-geo/src/clip/circle.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/cartesian.js"),i=n("./node_modules/d3-geo/src/circle.js"),o=n("./node_modules/d3-geo/src/math.js"),a=n("./node_modules/d3-geo/src/pointEqual.js"),s=n("./node_modules/d3-geo/src/clip/index.js");t.default=function(e){var t=Object(o.cos)(e),n=6*o.radians,u=t>0,c=Object(o.abs)(t)>o.epsilon;function l(e,n){return Object(o.cos)(e)*Object(o.cos)(n)>t}function d(e,n,i){var a=Object(r.cartesian)(e),s=Object(r.cartesian)(n),u=[1,0,0],c=Object(r.cartesianCross)(a,s),l=Object(r.cartesianDot)(c,c),d=c[0],p=l-d*d;if(!p)return!i&&e;var f=t*l/p,h=-t*d/p,m=Object(r.cartesianCross)(u,c),g=Object(r.cartesianScale)(u,f),v=Object(r.cartesianScale)(c,h);Object(r.cartesianAddInPlace)(g,v);var _=m,y=Object(r.cartesianDot)(g,_),b=Object(r.cartesianDot)(_,_),j=y*y-b*(Object(r.cartesianDot)(g,g)-1);if(!(j<0)){var x=Object(o.sqrt)(j),O=Object(r.cartesianScale)(_,(-y-x)/b);if(Object(r.cartesianAddInPlace)(O,g),O=Object(r.spherical)(O),!i)return O;var w,C=e[0],E=n[0],S=e[1],k=n[1];E<C&&(w=C,C=E,E=w);var M=E-C,A=Object(o.abs)(M-o.pi)<o.epsilon,T=A||M<o.epsilon;if(!A&&k<S&&(w=S,S=k,k=w),T?A?S+k>0^O[1]<(Object(o.abs)(O[0]-C)<o.epsilon?S:k):S<=O[1]&&O[1]<=k:M>o.pi^(C<=O[0]&&O[0]<=E)){var D=Object(r.cartesianScale)(_,(-y+x)/b);return Object(r.cartesianAddInPlace)(D,g),[O,Object(r.spherical)(D)]}}}function p(t,n){var r=u?e:o.pi-e,i=0;return t<-r?i|=1:t>r&&(i|=2),n<-r?i|=4:n>r&&(i|=8),i}return Object(s.default)(l,function(e){var t,n,r,i,s;return{lineStart:function(){i=r=!1,s=1},point:function(f,h){var m,g=[f,h],v=l(f,h),_=u?v?0:p(f,h):v?p(f+(f<0?o.pi:-o.pi),h):0;if(!t&&(i=r=v)&&e.lineStart(),v!==r&&(!(m=d(t,g))||Object(a.default)(t,m)||Object(a.default)(g,m))&&(g[0]+=o.epsilon,g[1]+=o.epsilon,v=l(g[0],g[1])),v!==r)s=0,v?(e.lineStart(),m=d(g,t),e.point(m[0],m[1])):(m=d(t,g),e.point(m[0],m[1]),e.lineEnd()),t=m;else if(c&&t&&u^v){var y;_&n||!(y=d(g,t,!0))||(s=0,u?(e.lineStart(),e.point(y[0][0],y[0][1]),e.point(y[1][0],y[1][1]),e.lineEnd()):(e.point(y[1][0],y[1][1]),e.lineEnd(),e.lineStart(),e.point(y[0][0],y[0][1])))}!v||t&&Object(a.default)(t,g)||e.point(g[0],g[1]),t=g,r=v,n=_},lineEnd:function(){r&&e.lineEnd(),t=null},clean:function(){return s|(i&&r)<<1}}},function(t,r,o,a){Object(i.circleStream)(a,e,n,o,t,r)},u?[0,-e]:[-o.pi,e-o.pi])}},"./node_modules/d3-geo/src/clip/extent.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/clip/rectangle.js");t.default=function(){var e,t,n,i=0,o=0,a=960,s=500;return n={stream:function(n){return e&&t===n?e:e=Object(r.default)(i,o,a,s)(t=n)},extent:function(r){return arguments.length?(i=+r[0][0],o=+r[0][1],a=+r[1][0],s=+r[1][1],e=t=null,n):[[i,o],[a,s]]}}}},"./node_modules/d3-geo/src/clip/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/clip/buffer.js"),i=n("./node_modules/d3-geo/src/clip/rejoin.js"),o=n("./node_modules/d3-geo/src/math.js"),a=n("./node_modules/d3-geo/src/polygonContains.js"),s=n("./node_modules/d3-array/index.js");function u(e){return e.length>1}function c(e,t){return((e=e.x)[0]<0?e[1]-o.halfPi-o.epsilon:o.halfPi-e[1])-((t=t.x)[0]<0?t[1]-o.halfPi-o.epsilon:o.halfPi-t[1])}t.default=function(e,t,n,o){return function(l){var d,p,f,h=t(l),m=Object(r.default)(),g=t(m),v=!1,_={point:y,lineStart:j,lineEnd:x,polygonStart:function(){_.point=O,_.lineStart=w,_.lineEnd=C,p=[],d=[]},polygonEnd:function(){_.point=y,_.lineStart=j,_.lineEnd=x,p=Object(s.merge)(p);var e=Object(a.default)(d,o);p.length?(v||(l.polygonStart(),v=!0),Object(i.default)(p,c,e,n,l)):e&&(v||(l.polygonStart(),v=!0),l.lineStart(),n(null,null,1,l),l.lineEnd()),v&&(l.polygonEnd(),v=!1),p=d=null},sphere:function(){l.polygonStart(),l.lineStart(),n(null,null,1,l),l.lineEnd(),l.polygonEnd()}};function y(t,n){e(t,n)&&l.point(t,n)}function b(e,t){h.point(e,t)}function j(){_.point=b,h.lineStart()}function x(){_.point=y,h.lineEnd()}function O(e,t){f.push([e,t]),g.point(e,t)}function w(){g.lineStart(),f=[]}function C(){O(f[0][0],f[0][1]),g.lineEnd();var e,t,n,r,i=g.clean(),o=m.result(),a=o.length;if(f.pop(),d.push(f),f=null,a)if(1&i){if((t=(n=o[0]).length-1)>0){for(v||(l.polygonStart(),v=!0),l.lineStart(),e=0;e<t;++e)l.point((r=n[e])[0],r[1]);l.lineEnd()}}else a>1&&2&i&&o.push(o.pop().concat(o.shift())),p.push(o.filter(u))}return _}}},"./node_modules/d3-geo/src/clip/line.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n,r,i,o){var a,s=e[0],u=e[1],c=0,l=1,d=t[0]-s,p=t[1]-u;if(a=n-s,d||!(a>0)){if(a/=d,d<0){if(a<c)return;a<l&&(l=a)}else if(d>0){if(a>l)return;a>c&&(c=a)}if(a=i-s,d||!(a<0)){if(a/=d,d<0){if(a>l)return;a>c&&(c=a)}else if(d>0){if(a<c)return;a<l&&(l=a)}if(a=r-u,p||!(a>0)){if(a/=p,p<0){if(a<c)return;a<l&&(l=a)}else if(p>0){if(a>l)return;a>c&&(c=a)}if(a=o-u,p||!(a<0)){if(a/=p,p<0){if(a>l)return;a>c&&(c=a)}else if(p>0){if(a<c)return;a<l&&(l=a)}return c>0&&(e[0]=s+c*d,e[1]=u+c*p),l<1&&(t[0]=s+l*d,t[1]=u+l*p),!0}}}}}},"./node_modules/d3-geo/src/clip/rectangle.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return l});var r=n("./node_modules/d3-geo/src/math.js"),i=n("./node_modules/d3-geo/src/clip/buffer.js"),o=n("./node_modules/d3-geo/src/clip/line.js"),a=n("./node_modules/d3-geo/src/clip/rejoin.js"),s=n("./node_modules/d3-array/index.js"),u=1e9,c=-u;function l(e,t,n,l){function d(r,i){return e<=r&&r<=n&&t<=i&&i<=l}function p(r,i,o,a){var s=0,u=0;if(null==r||(s=f(r,o))!==(u=f(i,o))||m(r,i)<0^o>0)do{a.point(0===s||3===s?e:n,s>1?l:t)}while((s=(s+o+4)%4)!==u);else a.point(i[0],i[1])}function f(i,o){return Object(r.abs)(i[0]-e)<r.epsilon?o>0?0:3:Object(r.abs)(i[0]-n)<r.epsilon?o>0?2:1:Object(r.abs)(i[1]-t)<r.epsilon?o>0?1:0:o>0?3:2}function h(e,t){return m(e.x,t.x)}function m(e,t){var n=f(e,1),r=f(t,1);return n!==r?n-r:0===n?t[1]-e[1]:1===n?e[0]-t[0]:2===n?e[1]-t[1]:t[0]-e[0]}return function(r){var f,m,g,v,_,y,b,j,x,O,w,C=r,E=Object(i.default)(),S={point:k,lineStart:function(){S.point=M,m&&m.push(g=[]);O=!0,x=!1,b=j=NaN},lineEnd:function(){f&&(M(v,_),y&&x&&E.rejoin(),f.push(E.result()));S.point=k,x&&C.lineEnd()},polygonStart:function(){C=E,f=[],m=[],w=!0},polygonEnd:function(){var t=function(){for(var t=0,n=0,r=m.length;n<r;++n)for(var i,o,a=m[n],s=1,u=a.length,c=a[0],d=c[0],p=c[1];s<u;++s)i=d,o=p,c=a[s],d=c[0],p=c[1],o<=l?p>l&&(d-i)*(l-o)>(p-o)*(e-i)&&++t:p<=l&&(d-i)*(l-o)<(p-o)*(e-i)&&--t;return t}(),n=w&&t,i=(f=Object(s.merge)(f)).length;(n||i)&&(r.polygonStart(),n&&(r.lineStart(),p(null,null,1,r),r.lineEnd()),i&&Object(a.default)(f,h,t,p,r),r.polygonEnd());C=r,f=m=g=null}};function k(e,t){d(e,t)&&C.point(e,t)}function M(r,i){var a=d(r,i);if(m&&g.push([r,i]),O)v=r,_=i,y=a,O=!1,a&&(C.lineStart(),C.point(r,i));else if(a&&x)C.point(r,i);else{var s=[b=Math.max(c,Math.min(u,b)),j=Math.max(c,Math.min(u,j))],p=[r=Math.max(c,Math.min(u,r)),i=Math.max(c,Math.min(u,i))];Object(o.default)(s,p,e,t,n,l)?(x||(C.lineStart(),C.point(s[0],s[1])),C.point(p[0],p[1]),a||C.lineEnd(),w=!1):a&&(C.lineStart(),C.point(r,i),w=!1)}b=r,j=i,x=a}return S}}},"./node_modules/d3-geo/src/clip/rejoin.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/pointEqual.js");function i(e,t,n,r){this.x=e,this.z=t,this.o=n,this.e=r,this.v=!1,this.n=this.p=null}function o(e){if(t=e.length){for(var t,n,r=0,i=e[0];++r<t;)i.n=n=e[r],n.p=i,i=n;i.n=n=e[0],n.p=i}}t.default=function(e,t,n,a,s){var u,c,l=[],d=[];if(e.forEach(function(e){if(!((t=e.length-1)<=0)){var t,n,o=e[0],a=e[t];if(Object(r.default)(o,a)){for(s.lineStart(),u=0;u<t;++u)s.point((o=e[u])[0],o[1]);s.lineEnd()}else l.push(n=new i(o,e,null,!0)),d.push(n.o=new i(o,null,n,!1)),l.push(n=new i(a,e,null,!1)),d.push(n.o=new i(a,null,n,!0))}}),l.length){for(d.sort(t),o(l),o(d),u=0,c=d.length;u<c;++u)d[u].e=n=!n;for(var p,f,h=l[0];;){for(var m=h,g=!0;m.v;)if((m=m.n)===h)return;p=m.z,s.lineStart();do{if(m.v=m.o.v=!0,m.e){if(g)for(u=0,c=p.length;u<c;++u)s.point((f=p[u])[0],f[1]);else a(m.x,m.n.x,1,s);m=m.n}else{if(g)for(p=m.p.z,u=p.length-1;u>=0;--u)s.point((f=p[u])[0],f[1]);else a(m.x,m.p.x,-1,s);m=m.p}p=(m=m.o).z,g=!g}while(!m.v);s.lineEnd()}}}},"./node_modules/d3-geo/src/compose.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){function n(n,r){return n=e(n,r),t(n[0],n[1])}return e.invert&&t.invert&&(n.invert=function(n,r){return(n=t.invert(n,r))&&e.invert(n[0],n[1])}),n}},"./node_modules/d3-geo/src/constant.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return function(){return e}}},"./node_modules/d3-geo/src/contains.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/polygonContains.js"),i=n("./node_modules/d3-geo/src/distance.js"),o=n("./node_modules/d3-geo/src/math.js"),a={Feature:function(e,t){return u(e.geometry,t)},FeatureCollection:function(e,t){for(var n=e.features,r=-1,i=n.length;++r<i;)if(u(n[r].geometry,t))return!0;return!1}},s={Sphere:function(){return!0},Point:function(e,t){return c(e.coordinates,t)},MultiPoint:function(e,t){for(var n=e.coordinates,r=-1,i=n.length;++r<i;)if(c(n[r],t))return!0;return!1},LineString:function(e,t){return l(e.coordinates,t)},MultiLineString:function(e,t){for(var n=e.coordinates,r=-1,i=n.length;++r<i;)if(l(n[r],t))return!0;return!1},Polygon:function(e,t){return d(e.coordinates,t)},MultiPolygon:function(e,t){for(var n=e.coordinates,r=-1,i=n.length;++r<i;)if(d(n[r],t))return!0;return!1},GeometryCollection:function(e,t){for(var n=e.geometries,r=-1,i=n.length;++r<i;)if(u(n[r],t))return!0;return!1}};function u(e,t){return!(!e||!s.hasOwnProperty(e.type))&&s[e.type](e,t)}function c(e,t){return 0===Object(i.default)(e,t)}function l(e,t){var n=Object(i.default)(e[0],e[1]);return Object(i.default)(e[0],t)+Object(i.default)(t,e[1])<=n+o.epsilon}function d(e,t){return!!Object(r.default)(e.map(p),f(t))}function p(e){return(e=e.map(f)).pop(),e}function f(e){return[e[0]*o.radians,e[1]*o.radians]}t.default=function(e,t){return(e&&a.hasOwnProperty(e.type)?a[e.type]:u)(e,t)}},"./node_modules/d3-geo/src/distance.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/length.js"),i=[null,null],o={type:"LineString",coordinates:i};t.default=function(e,t){return i[0]=e,i[1]=t,Object(r.default)(o)}},"./node_modules/d3-geo/src/graticule.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return s}),n.d(t,"graticule10",function(){return u});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/d3-geo/src/math.js");function o(e,t,n){var o=Object(r.range)(e,t-i.epsilon,n).concat(t);return function(e){return o.map(function(t){return[e,t]})}}function a(e,t,n){var o=Object(r.range)(e,t-i.epsilon,n).concat(t);return function(e){return o.map(function(t){return[t,e]})}}function s(){var e,t,n,s,u,c,l,d,p,f,h,m,g=10,v=g,_=90,y=360,b=2.5;function j(){return{type:"MultiLineString",coordinates:x()}}function x(){return Object(r.range)(Object(i.ceil)(s/_)*_,n,_).map(h).concat(Object(r.range)(Object(i.ceil)(d/y)*y,l,y).map(m)).concat(Object(r.range)(Object(i.ceil)(t/g)*g,e,g).filter(function(e){return Object(i.abs)(e%_)>i.epsilon}).map(p)).concat(Object(r.range)(Object(i.ceil)(c/v)*v,u,v).filter(function(e){return Object(i.abs)(e%y)>i.epsilon}).map(f))}return j.lines=function(){return x().map(function(e){return{type:"LineString",coordinates:e}})},j.outline=function(){return{type:"Polygon",coordinates:[h(s).concat(m(l).slice(1),h(n).reverse().slice(1),m(d).reverse().slice(1))]}},j.extent=function(e){return arguments.length?j.extentMajor(e).extentMinor(e):j.extentMinor()},j.extentMajor=function(e){return arguments.length?(s=+e[0][0],n=+e[1][0],d=+e[0][1],l=+e[1][1],s>n&&(e=s,s=n,n=e),d>l&&(e=d,d=l,l=e),j.precision(b)):[[s,d],[n,l]]},j.extentMinor=function(n){return arguments.length?(t=+n[0][0],e=+n[1][0],c=+n[0][1],u=+n[1][1],t>e&&(n=t,t=e,e=n),c>u&&(n=c,c=u,u=n),j.precision(b)):[[t,c],[e,u]]},j.step=function(e){return arguments.length?j.stepMajor(e).stepMinor(e):j.stepMinor()},j.stepMajor=function(e){return arguments.length?(_=+e[0],y=+e[1],j):[_,y]},j.stepMinor=function(e){return arguments.length?(g=+e[0],v=+e[1],j):[g,v]},j.precision=function(r){return arguments.length?(b=+r,p=o(c,u,90),f=a(t,e,b),h=o(d,l,90),m=a(s,n,b),j):b},j.extentMajor([[-180,-90+i.epsilon],[180,90-i.epsilon]]).extentMinor([[-180,-80-i.epsilon],[180,80+i.epsilon]])}function u(){return s()()}},"./node_modules/d3-geo/src/identity.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return e}},"./node_modules/d3-geo/src/interpolate.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/math.js");t.default=function(e,t){var n=e[0]*r.radians,i=e[1]*r.radians,o=t[0]*r.radians,a=t[1]*r.radians,s=Object(r.cos)(i),u=Object(r.sin)(i),c=Object(r.cos)(a),l=Object(r.sin)(a),d=s*Object(r.cos)(n),p=s*Object(r.sin)(n),f=c*Object(r.cos)(o),h=c*Object(r.sin)(o),m=2*Object(r.asin)(Object(r.sqrt)(Object(r.haversin)(a-i)+s*c*Object(r.haversin)(o-n))),g=Object(r.sin)(m),v=m?function(e){var t=Object(r.sin)(e*=m)/g,n=Object(r.sin)(m-e)/g,i=n*d+t*f,o=n*p+t*h,a=n*u+t*l;return[Object(r.atan2)(o,i)*r.degrees,Object(r.atan2)(a,Object(r.sqrt)(i*i+o*o))*r.degrees]}:function(){return[n*r.degrees,i*r.degrees]};return v.distance=m,v}},"./node_modules/d3-geo/src/length.js":function(e,t,n){"use strict";n.r(t);var r,i,o,a=n("./node_modules/d3-geo/src/adder.js"),s=n("./node_modules/d3-geo/src/math.js"),u=n("./node_modules/d3-geo/src/noop.js"),c=n("./node_modules/d3-geo/src/stream.js"),l=Object(a.default)(),d={sphere:u.default,point:u.default,lineStart:function(){d.point=f,d.lineEnd=p},lineEnd:u.default,polygonStart:u.default,polygonEnd:u.default};function p(){d.point=d.lineEnd=u.default}function f(e,t){e*=s.radians,t*=s.radians,r=e,i=Object(s.sin)(t),o=Object(s.cos)(t),d.point=h}function h(e,t){e*=s.radians,t*=s.radians;var n=Object(s.sin)(t),a=Object(s.cos)(t),u=Object(s.abs)(e-r),c=Object(s.cos)(u),d=a*Object(s.sin)(u),p=o*n-i*a*c,f=i*n+o*a*c;l.add(Object(s.atan2)(Object(s.sqrt)(d*d+p*p),f)),r=e,i=n,o=a}t.default=function(e){return l.reset(),Object(c.default)(e,d),+l}},"./node_modules/d3-geo/src/math.js":function(e,t,n){"use strict";n.r(t),n.d(t,"epsilon",function(){return r}),n.d(t,"epsilon2",function(){return i}),n.d(t,"pi",function(){return o}),n.d(t,"halfPi",function(){return a}),n.d(t,"quarterPi",function(){return s}),n.d(t,"tau",function(){return u}),n.d(t,"degrees",function(){return c}),n.d(t,"radians",function(){return l}),n.d(t,"abs",function(){return d}),n.d(t,"atan",function(){return p}),n.d(t,"atan2",function(){return f}),n.d(t,"cos",function(){return h}),n.d(t,"ceil",function(){return m}),n.d(t,"exp",function(){return g}),n.d(t,"floor",function(){return v}),n.d(t,"log",function(){return _}),n.d(t,"pow",function(){return y}),n.d(t,"sin",function(){return b}),n.d(t,"sign",function(){return j}),n.d(t,"sqrt",function(){return x}),n.d(t,"tan",function(){return O}),n.d(t,"acos",function(){return w}),n.d(t,"asin",function(){return C}),n.d(t,"haversin",function(){return E});var r=1e-6,i=1e-12,o=Math.PI,a=o/2,s=o/4,u=2*o,c=180/o,l=o/180,d=Math.abs,p=Math.atan,f=Math.atan2,h=Math.cos,m=Math.ceil,g=Math.exp,v=Math.floor,_=Math.log,y=Math.pow,b=Math.sin,j=Math.sign||function(e){return e>0?1:e<0?-1:0},x=Math.sqrt,O=Math.tan;function w(e){return e>1?0:e<-1?o:Math.acos(e)}function C(e){return e>1?a:e<-1?-a:Math.asin(e)}function E(e){return(e=b(e/2))*e}},"./node_modules/d3-geo/src/noop.js":function(e,t,n){"use strict";function r(){}n.r(t),n.d(t,"default",function(){return r})},"./node_modules/d3-geo/src/path/area.js":function(e,t,n){"use strict";n.r(t);var r,i,o,a,s=n("./node_modules/d3-geo/src/adder.js"),u=n("./node_modules/d3-geo/src/math.js"),c=n("./node_modules/d3-geo/src/noop.js"),l=Object(s.default)(),d=Object(s.default)(),p={point:c.default,lineStart:c.default,lineEnd:c.default,polygonStart:function(){p.lineStart=f,p.lineEnd=g},polygonEnd:function(){p.lineStart=p.lineEnd=p.point=c.default,l.add(Object(u.abs)(d)),d.reset()},result:function(){var e=l/2;return l.reset(),e}};function f(){p.point=h}function h(e,t){p.point=m,r=o=e,i=a=t}function m(e,t){d.add(a*e-o*t),o=e,a=t}function g(){m(r,i)}t.default=p},"./node_modules/d3-geo/src/path/bounds.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/noop.js"),i=1/0,o=i,a=-i,s=a,u={point:function(e,t){e<i&&(i=e);e>a&&(a=e);t<o&&(o=t);t>s&&(s=t)},lineStart:r.default,lineEnd:r.default,polygonStart:r.default,polygonEnd:r.default,result:function(){var e=[[i,o],[a,s]];return a=s=-(o=i=1/0),e}};t.default=u},"./node_modules/d3-geo/src/path/centroid.js":function(e,t,n){"use strict";n.r(t);var r,i,o,a,s=n("./node_modules/d3-geo/src/math.js"),u=0,c=0,l=0,d=0,p=0,f=0,h=0,m=0,g=0,v={point:_,lineStart:y,lineEnd:x,polygonStart:function(){v.lineStart=O,v.lineEnd=w},polygonEnd:function(){v.point=_,v.lineStart=y,v.lineEnd=x},result:function(){var e=g?[h/g,m/g]:f?[d/f,p/f]:l?[u/l,c/l]:[NaN,NaN];return u=c=l=d=p=f=h=m=g=0,e}};function _(e,t){u+=e,c+=t,++l}function y(){v.point=b}function b(e,t){v.point=j,_(o=e,a=t)}function j(e,t){var n=e-o,r=t-a,i=Object(s.sqrt)(n*n+r*r);d+=i*(o+e)/2,p+=i*(a+t)/2,f+=i,_(o=e,a=t)}function x(){v.point=_}function O(){v.point=C}function w(){E(r,i)}function C(e,t){v.point=E,_(r=o=e,i=a=t)}function E(e,t){var n=e-o,r=t-a,i=Object(s.sqrt)(n*n+r*r);d+=i*(o+e)/2,p+=i*(a+t)/2,f+=i,h+=(i=a*e-o*t)*(o+e),m+=i*(a+t),g+=3*i,_(o=e,a=t)}t.default=v},"./node_modules/d3-geo/src/path/context.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/d3-geo/src/math.js"),i=n("./node_modules/d3-geo/src/noop.js");function o(e){this._context=e}o.prototype={_radius:4.5,pointRadius:function(e){return this._radius=e,this},polygonStart:function(){this._line=0},polygonEnd:function(){this._line=NaN},lineStart:function(){this._point=0},lineEnd:function(){0===this._line&&this._context.closePath(),this._point=NaN},point:function(e,t){switch(this._point){case 0:this._context.moveTo(e,t),this._point=1;break;case 1:this._context.lineTo(e,t);break;default:this._context.moveTo(e+this._radius,t),this._context.arc(e,t,this._radius,0,r.tau)}},result:i.default}},"./node_modules/d3-geo/src/path/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/identity.js"),i=n("./node_modules/d3-geo/src/stream.js"),o=n("./node_modules/d3-geo/src/path/area.js"),a=n("./node_modules/d3-geo/src/path/bounds.js"),s=n("./node_modules/d3-geo/src/path/centroid.js"),u=n("./node_modules/d3-geo/src/path/context.js"),c=n("./node_modules/d3-geo/src/path/measure.js"),l=n("./node_modules/d3-geo/src/path/string.js");t.default=function(e,t){var n,d,p=4.5;function f(e){return e&&("function"==typeof p&&d.pointRadius(+p.apply(this,arguments)),Object(i.default)(e,n(d))),d.result()}return f.area=function(e){return Object(i.default)(e,n(o.default)),o.default.result()},f.measure=function(e){return Object(i.default)(e,n(c.default)),c.default.result()},f.bounds=function(e){return Object(i.default)(e,n(a.default)),a.default.result()},f.centroid=function(e){return Object(i.default)(e,n(s.default)),s.default.result()},f.projection=function(t){return arguments.length?(n=null==t?(e=null,r.default):(e=t).stream,f):e},f.context=function(e){return arguments.length?(d=null==e?(t=null,new l.default):new u.default(t=e),"function"!=typeof p&&d.pointRadius(p),f):t},f.pointRadius=function(e){return arguments.length?(p="function"==typeof e?e:(d.pointRadius(+e),+e),f):p},f.projection(e).context(t)}},"./node_modules/d3-geo/src/path/measure.js":function(e,t,n){"use strict";n.r(t);var r,i,o,a,s,u=n("./node_modules/d3-geo/src/adder.js"),c=n("./node_modules/d3-geo/src/math.js"),l=n("./node_modules/d3-geo/src/noop.js"),d=Object(u.default)(),p={point:l.default,lineStart:function(){p.point=f},lineEnd:function(){r&&h(i,o),p.point=l.default},polygonStart:function(){r=!0},polygonEnd:function(){r=null},result:function(){var e=+d;return d.reset(),e}};function f(e,t){p.point=h,i=a=e,o=s=t}function h(e,t){a-=e,s-=t,d.add(Object(c.sqrt)(a*a+s*s)),a=e,s=t}t.default=p},"./node_modules/d3-geo/src/path/string.js":function(e,t,n){"use strict";function r(){this._string=[]}function i(e){return"m0,"+e+"a"+e+","+e+" 0 1,1 0,"+-2*e+"a"+e+","+e+" 0 1,1 0,"+2*e+"z"}n.r(t),n.d(t,"default",function(){return r}),r.prototype={_radius:4.5,_circle:i(4.5),pointRadius:function(e){return(e=+e)!==this._radius&&(this._radius=e,this._circle=null),this},polygonStart:function(){this._line=0},polygonEnd:function(){this._line=NaN},lineStart:function(){this._point=0},lineEnd:function(){0===this._line&&this._string.push("Z"),this._point=NaN},point:function(e,t){switch(this._point){case 0:this._string.push("M",e,",",t),this._point=1;break;case 1:this._string.push("L",e,",",t);break;default:null==this._circle&&(this._circle=i(this._radius)),this._string.push("M",e,",",t,this._circle)}},result:function(){if(this._string.length){var e=this._string.join("");return this._string=[],e}return null}}},"./node_modules/d3-geo/src/pointEqual.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/math.js");t.default=function(e,t){return Object(r.abs)(e[0]-t[0])<r.epsilon&&Object(r.abs)(e[1]-t[1])<r.epsilon}},"./node_modules/d3-geo/src/polygonContains.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/adder.js"),i=n("./node_modules/d3-geo/src/cartesian.js"),o=n("./node_modules/d3-geo/src/math.js"),a=Object(r.default)();t.default=function(e,t){var n=t[0],r=t[1],s=[Object(o.sin)(n),-Object(o.cos)(n),0],u=0,c=0;a.reset();for(var l=0,d=e.length;l<d;++l)if(f=(p=e[l]).length)for(var p,f,h=p[f-1],m=h[0],g=h[1]/2+o.quarterPi,v=Object(o.sin)(g),_=Object(o.cos)(g),y=0;y<f;++y,m=j,v=O,_=w,h=b){var b=p[y],j=b[0],x=b[1]/2+o.quarterPi,O=Object(o.sin)(x),w=Object(o.cos)(x),C=j-m,E=C>=0?1:-1,S=E*C,k=S>o.pi,M=v*O;if(a.add(Object(o.atan2)(M*E*Object(o.sin)(S),_*w+M*Object(o.cos)(S))),u+=k?C+E*o.tau:C,k^m>=n^j>=n){var A=Object(i.cartesianCross)(Object(i.cartesian)(h),Object(i.cartesian)(b));Object(i.cartesianNormalizeInPlace)(A);var T=Object(i.cartesianCross)(s,A);Object(i.cartesianNormalizeInPlace)(T);var D=(k^C>=0?-1:1)*Object(o.asin)(T[2]);(r>D||r===D&&(A[0]||A[1]))&&(c+=k^C>=0?1:-1)}}return(u<-o.epsilon||u<o.epsilon&&a<-o.epsilon)^1&c}},"./node_modules/d3-geo/src/projection/albers.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/projection/conicEqualArea.js");t.default=function(){return Object(r.default)().parallels([29.5,45.5]).scale(1070).translate([480,250]).rotate([96,0]).center([-.6,38.7])}},"./node_modules/d3-geo/src/projection/albersUsa.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/math.js"),i=n("./node_modules/d3-geo/src/projection/albers.js"),o=n("./node_modules/d3-geo/src/projection/conicEqualArea.js"),a=n("./node_modules/d3-geo/src/projection/fit.js");t.default=function(){var e,t,n,s,u,c,l=Object(i.default)(),d=Object(o.default)().rotate([154,0]).center([-2,58.5]).parallels([55,65]),p=Object(o.default)().rotate([157,0]).center([-3,19.9]).parallels([8,18]),f={point:function(e,t){c=[e,t]}};function h(e){var t=e[0],r=e[1];return c=null,n.point(t,r),c||(s.point(t,r),c)||(u.point(t,r),c)}function m(){return e=t=null,h}return h.invert=function(e){var t=l.scale(),n=l.translate(),r=(e[0]-n[0])/t,i=(e[1]-n[1])/t;return(i>=.12&&i<.234&&r>=-.425&&r<-.214?d:i>=.166&&i<.234&&r>=-.214&&r<-.115?p:l).invert(e)},h.stream=function(n){return e&&t===n?e:(r=[l.stream(t=n),d.stream(n),p.stream(n)],i=r.length,e={point:function(e,t){for(var n=-1;++n<i;)r[n].point(e,t)},sphere:function(){for(var e=-1;++e<i;)r[e].sphere()},lineStart:function(){for(var e=-1;++e<i;)r[e].lineStart()},lineEnd:function(){for(var e=-1;++e<i;)r[e].lineEnd()},polygonStart:function(){for(var e=-1;++e<i;)r[e].polygonStart()},polygonEnd:function(){for(var e=-1;++e<i;)r[e].polygonEnd()}});var r,i},h.precision=function(e){return arguments.length?(l.precision(e),d.precision(e),p.precision(e),m()):l.precision()},h.scale=function(e){return arguments.length?(l.scale(e),d.scale(.35*e),p.scale(e),h.translate(l.translate())):l.scale()},h.translate=function(e){if(!arguments.length)return l.translate();var t=l.scale(),i=+e[0],o=+e[1];return n=l.translate(e).clipExtent([[i-.455*t,o-.238*t],[i+.455*t,o+.238*t]]).stream(f),s=d.translate([i-.307*t,o+.201*t]).clipExtent([[i-.425*t+r.epsilon,o+.12*t+r.epsilon],[i-.214*t-r.epsilon,o+.234*t-r.epsilon]]).stream(f),u=p.translate([i-.205*t,o+.212*t]).clipExtent([[i-.214*t+r.epsilon,o+.166*t+r.epsilon],[i-.115*t-r.epsilon,o+.234*t-r.epsilon]]).stream(f),m()},h.fitExtent=function(e,t){return Object(a.fitExtent)(h,e,t)},h.fitSize=function(e,t){return Object(a.fitSize)(h,e,t)},h.fitWidth=function(e,t){return Object(a.fitWidth)(h,e,t)},h.fitHeight=function(e,t){return Object(a.fitHeight)(h,e,t)},h.scale(1070)}},"./node_modules/d3-geo/src/projection/azimuthal.js":function(e,t,n){"use strict";n.r(t),n.d(t,"azimuthalRaw",function(){return i}),n.d(t,"azimuthalInvert",function(){return o});var r=n("./node_modules/d3-geo/src/math.js");function i(e){return function(t,n){var i=Object(r.cos)(t),o=Object(r.cos)(n),a=e(i*o);return[a*o*Object(r.sin)(t),a*Object(r.sin)(n)]}}function o(e){return function(t,n){var i=Object(r.sqrt)(t*t+n*n),o=e(i),a=Object(r.sin)(o),s=Object(r.cos)(o);return[Object(r.atan2)(t*a,i*s),Object(r.asin)(i&&n*a/i)]}}},"./node_modules/d3-geo/src/projection/azimuthalEqualArea.js":function(e,t,n){"use strict";n.r(t),n.d(t,"azimuthalEqualAreaRaw",function(){return a});var r=n("./node_modules/d3-geo/src/math.js"),i=n("./node_modules/d3-geo/src/projection/azimuthal.js"),o=n("./node_modules/d3-geo/src/projection/index.js"),a=Object(i.azimuthalRaw)(function(e){return Object(r.sqrt)(2/(1+e))});a.invert=Object(i.azimuthalInvert)(function(e){return 2*Object(r.asin)(e/2)}),t.default=function(){return Object(o.default)(a).scale(124.75).clipAngle(179.999)}},"./node_modules/d3-geo/src/projection/azimuthalEquidistant.js":function(e,t,n){"use strict";n.r(t),n.d(t,"azimuthalEquidistantRaw",function(){return a});var r=n("./node_modules/d3-geo/src/math.js"),i=n("./node_modules/d3-geo/src/projection/azimuthal.js"),o=n("./node_modules/d3-geo/src/projection/index.js"),a=Object(i.azimuthalRaw)(function(e){return(e=Object(r.acos)(e))&&e/Object(r.sin)(e)});a.invert=Object(i.azimuthalInvert)(function(e){return e}),t.default=function(){return Object(o.default)(a).scale(79.4188).clipAngle(179.999)}},"./node_modules/d3-geo/src/projection/conic.js":function(e,t,n){"use strict";n.r(t),n.d(t,"conicProjection",function(){return o});var r=n("./node_modules/d3-geo/src/math.js"),i=n("./node_modules/d3-geo/src/projection/index.js");function o(e){var t=0,n=r.pi/3,o=Object(i.projectionMutator)(e),a=o(t,n);return a.parallels=function(e){return arguments.length?o(t=e[0]*r.radians,n=e[1]*r.radians):[t*r.degrees,n*r.degrees]},a}},"./node_modules/d3-geo/src/projection/conicConformal.js":function(e,t,n){"use strict";n.r(t),n.d(t,"conicConformalRaw",function(){return s});var r=n("./node_modules/d3-geo/src/math.js"),i=n("./node_modules/d3-geo/src/projection/conic.js"),o=n("./node_modules/d3-geo/src/projection/mercator.js");function a(e){return Object(r.tan)((r.halfPi+e)/2)}function s(e,t){var n=Object(r.cos)(e),i=e===t?Object(r.sin)(e):Object(r.log)(n/Object(r.cos)(t))/Object(r.log)(a(t)/a(e)),s=n*Object(r.pow)(a(e),i)/i;if(!i)return o.mercatorRaw;function u(e,t){s>0?t<-r.halfPi+r.epsilon&&(t=-r.halfPi+r.epsilon):t>r.halfPi-r.epsilon&&(t=r.halfPi-r.epsilon);var n=s/Object(r.pow)(a(t),i);return[n*Object(r.sin)(i*e),s-n*Object(r.cos)(i*e)]}return u.invert=function(e,t){var n=s-t,o=Object(r.sign)(i)*Object(r.sqrt)(e*e+n*n);return[Object(r.atan2)(e,Object(r.abs)(n))/i*Object(r.sign)(n),2*Object(r.atan)(Object(r.pow)(s/o,1/i))-r.halfPi]},u}t.default=function(){return Object(i.conicProjection)(s).scale(109.5).parallels([30,30])}},"./node_modules/d3-geo/src/projection/conicEqualArea.js":function(e,t,n){"use strict";n.r(t),n.d(t,"conicEqualAreaRaw",function(){return a});var r=n("./node_modules/d3-geo/src/math.js"),i=n("./node_modules/d3-geo/src/projection/conic.js"),o=n("./node_modules/d3-geo/src/projection/cylindricalEqualArea.js");function a(e,t){var n=Object(r.sin)(e),i=(n+Object(r.sin)(t))/2;if(Object(r.abs)(i)<r.epsilon)return Object(o.cylindricalEqualAreaRaw)(e);var a=1+n*(2*i-n),s=Object(r.sqrt)(a)/i;function u(e,t){var n=Object(r.sqrt)(a-2*i*Object(r.sin)(t))/i;return[n*Object(r.sin)(e*=i),s-n*Object(r.cos)(e)]}return u.invert=function(e,t){var n=s-t;return[Object(r.atan2)(e,Object(r.abs)(n))/i*Object(r.sign)(n),Object(r.asin)((a-(e*e+n*n)*i*i)/(2*i))]},u}t.default=function(){return Object(i.conicProjection)(a).scale(155.424).center([0,33.6442])}},"./node_modules/d3-geo/src/projection/conicEquidistant.js":function(e,t,n){"use strict";n.r(t),n.d(t,"conicEquidistantRaw",function(){return a});var r=n("./node_modules/d3-geo/src/math.js"),i=n("./node_modules/d3-geo/src/projection/conic.js"),o=n("./node_modules/d3-geo/src/projection/equirectangular.js");function a(e,t){var n=Object(r.cos)(e),i=e===t?Object(r.sin)(e):(n-Object(r.cos)(t))/(t-e),a=n/i+e;if(Object(r.abs)(i)<r.epsilon)return o.equirectangularRaw;function s(e,t){var n=a-t,o=i*e;return[n*Object(r.sin)(o),a-n*Object(r.cos)(o)]}return s.invert=function(e,t){var n=a-t;return[Object(r.atan2)(e,Object(r.abs)(n))/i*Object(r.sign)(n),a-Object(r.sign)(i)*Object(r.sqrt)(e*e+n*n)]},s}t.default=function(){return Object(i.conicProjection)(a).scale(131.154).center([0,13.9389])}},"./node_modules/d3-geo/src/projection/cylindricalEqualArea.js":function(e,t,n){"use strict";n.r(t),n.d(t,"cylindricalEqualAreaRaw",function(){return i});var r=n("./node_modules/d3-geo/src/math.js");function i(e){var t=Object(r.cos)(e);function n(e,n){return[e*t,Object(r.sin)(n)/t]}return n.invert=function(e,n){return[e/t,Object(r.asin)(n*t)]},n}},"./node_modules/d3-geo/src/projection/equirectangular.js":function(e,t,n){"use strict";n.r(t),n.d(t,"equirectangularRaw",function(){return i});var r=n("./node_modules/d3-geo/src/projection/index.js");function i(e,t){return[e,t]}i.invert=i,t.default=function(){return Object(r.default)(i).scale(152.63)}},"./node_modules/d3-geo/src/projection/fit.js":function(e,t,n){"use strict";n.r(t),n.d(t,"fitExtent",function(){return a}),n.d(t,"fitSize",function(){return s}),n.d(t,"fitWidth",function(){return u}),n.d(t,"fitHeight",function(){return c});var r=n("./node_modules/d3-geo/src/stream.js"),i=n("./node_modules/d3-geo/src/path/bounds.js");function o(e,t,n){var o=e.clipExtent&&e.clipExtent();return e.scale(150).translate([0,0]),null!=o&&e.clipExtent(null),Object(r.default)(n,e.stream(i.default)),t(i.default.result()),null!=o&&e.clipExtent(o),e}function a(e,t,n){return o(e,function(n){var r=t[1][0]-t[0][0],i=t[1][1]-t[0][1],o=Math.min(r/(n[1][0]-n[0][0]),i/(n[1][1]-n[0][1])),a=+t[0][0]+(r-o*(n[1][0]+n[0][0]))/2,s=+t[0][1]+(i-o*(n[1][1]+n[0][1]))/2;e.scale(150*o).translate([a,s])},n)}function s(e,t,n){return a(e,[[0,0],t],n)}function u(e,t,n){return o(e,function(n){var r=+t,i=r/(n[1][0]-n[0][0]),o=(r-i*(n[1][0]+n[0][0]))/2,a=-i*n[0][1];e.scale(150*i).translate([o,a])},n)}function c(e,t,n){return o(e,function(n){var r=+t,i=r/(n[1][1]-n[0][1]),o=-i*n[0][0],a=(r-i*(n[1][1]+n[0][1]))/2;e.scale(150*i).translate([o,a])},n)}},"./node_modules/d3-geo/src/projection/gnomonic.js":function(e,t,n){"use strict";n.r(t),n.d(t,"gnomonicRaw",function(){return a});var r=n("./node_modules/d3-geo/src/math.js"),i=n("./node_modules/d3-geo/src/projection/azimuthal.js"),o=n("./node_modules/d3-geo/src/projection/index.js");function a(e,t){var n=Object(r.cos)(t),i=Object(r.cos)(e)*n;return[n*Object(r.sin)(e)/i,Object(r.sin)(t)/i]}a.invert=Object(i.azimuthalInvert)(r.atan),t.default=function(){return Object(o.default)(a).scale(144.049).clipAngle(60)}},"./node_modules/d3-geo/src/projection/identity.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/clip/rectangle.js"),i=n("./node_modules/d3-geo/src/identity.js"),o=n("./node_modules/d3-geo/src/transform.js"),a=n("./node_modules/d3-geo/src/projection/fit.js");function s(e,t,n,r){return 1===e&&1===t&&0===n&&0===r?i.default:Object(o.transformer)({point:function(i,o){this.stream.point(i*e+n,o*t+r)}})}t.default=function(){var e,t,n,o,u,c,l=1,d=0,p=0,f=1,h=1,m=i.default,g=null,v=i.default;function _(){return o=u=null,c}return c={stream:function(e){return o&&u===e?o:o=m(v(u=e))},postclip:function(r){return arguments.length?(v=r,g=e=t=n=null,_()):v},clipExtent:function(o){return arguments.length?(v=null==o?(g=e=t=n=null,i.default):Object(r.default)(g=+o[0][0],e=+o[0][1],t=+o[1][0],n=+o[1][1]),_()):null==g?null:[[g,e],[t,n]]},scale:function(e){return arguments.length?(m=s((l=+e)*f,l*h,d,p),_()):l},translate:function(e){return arguments.length?(m=s(l*f,l*h,d=+e[0],p=+e[1]),_()):[d,p]},reflectX:function(e){return arguments.length?(m=s(l*(f=e?-1:1),l*h,d,p),_()):f<0},reflectY:function(e){return arguments.length?(m=s(l*f,l*(h=e?-1:1),d,p),_()):h<0},fitExtent:function(e,t){return Object(a.fitExtent)(c,e,t)},fitSize:function(e,t){return Object(a.fitSize)(c,e,t)},fitWidth:function(e,t){return Object(a.fitWidth)(c,e,t)},fitHeight:function(e,t){return Object(a.fitHeight)(c,e,t)}}}},"./node_modules/d3-geo/src/projection/index.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return h}),n.d(t,"projectionMutator",function(){return m});var r=n("./node_modules/d3-geo/src/clip/antimeridian.js"),i=n("./node_modules/d3-geo/src/clip/circle.js"),o=n("./node_modules/d3-geo/src/clip/rectangle.js"),a=n("./node_modules/d3-geo/src/compose.js"),s=n("./node_modules/d3-geo/src/identity.js"),u=n("./node_modules/d3-geo/src/math.js"),c=n("./node_modules/d3-geo/src/rotation.js"),l=n("./node_modules/d3-geo/src/transform.js"),d=n("./node_modules/d3-geo/src/projection/fit.js"),p=n("./node_modules/d3-geo/src/projection/resample.js"),f=Object(l.transformer)({point:function(e,t){this.stream.point(e*u.radians,t*u.radians)}});function h(e){return m(function(){return e})()}function m(e){var t,n,h,m,g,v,_,y,b,j,x=150,O=480,w=250,C=0,E=0,S=0,k=0,M=0,A=null,T=r.default,D=null,I=s.default,P=.5,R=Object(p.default)(L,P);function N(e){return[(e=g(e[0]*u.radians,e[1]*u.radians))[0]*x+n,h-e[1]*x]}function F(e){return(e=g.invert((e[0]-n)/x,(h-e[1])/x))&&[e[0]*u.degrees,e[1]*u.degrees]}function L(e,r){return[(e=t(e,r))[0]*x+n,h-e[1]*x]}function V(){g=Object(a.default)(m=Object(c.rotateRadians)(S,k,M),t);var e=t(C,E);return n=O-e[0]*x,h=w+e[1]*x,B()}function B(){return b=j=null,N}return N.stream=function(e){return b&&j===e?b:b=f((t=m,Object(l.transformer)({point:function(e,n){var r=t(e,n);return this.stream.point(r[0],r[1])}}))(T(R(I(j=e)))));var t},N.preclip=function(e){return arguments.length?(T=e,A=void 0,B()):T},N.postclip=function(e){return arguments.length?(I=e,D=v=_=y=null,B()):I},N.clipAngle=function(e){return arguments.length?(T=+e?Object(i.default)(A=e*u.radians):(A=null,r.default),B()):A*u.degrees},N.clipExtent=function(e){return arguments.length?(I=null==e?(D=v=_=y=null,s.default):Object(o.default)(D=+e[0][0],v=+e[0][1],_=+e[1][0],y=+e[1][1]),B()):null==D?null:[[D,v],[_,y]]},N.scale=function(e){return arguments.length?(x=+e,V()):x},N.translate=function(e){return arguments.length?(O=+e[0],w=+e[1],V()):[O,w]},N.center=function(e){return arguments.length?(C=e[0]%360*u.radians,E=e[1]%360*u.radians,V()):[C*u.degrees,E*u.degrees]},N.rotate=function(e){return arguments.length?(S=e[0]%360*u.radians,k=e[1]%360*u.radians,M=e.length>2?e[2]%360*u.radians:0,V()):[S*u.degrees,k*u.degrees,M*u.degrees]},N.precision=function(e){return arguments.length?(R=Object(p.default)(L,P=e*e),B()):Object(u.sqrt)(P)},N.fitExtent=function(e,t){return Object(d.fitExtent)(N,e,t)},N.fitSize=function(e,t){return Object(d.fitSize)(N,e,t)},N.fitWidth=function(e,t){return Object(d.fitWidth)(N,e,t)},N.fitHeight=function(e,t){return Object(d.fitHeight)(N,e,t)},function(){return t=e.apply(this,arguments),N.invert=t.invert&&F,V()}}},"./node_modules/d3-geo/src/projection/mercator.js":function(e,t,n){"use strict";n.r(t),n.d(t,"mercatorRaw",function(){return a}),n.d(t,"mercatorProjection",function(){return s});var r=n("./node_modules/d3-geo/src/math.js"),i=n("./node_modules/d3-geo/src/rotation.js"),o=n("./node_modules/d3-geo/src/projection/index.js");function a(e,t){return[e,Object(r.log)(Object(r.tan)((r.halfPi+t)/2))]}function s(e){var t,n,s,u=Object(o.default)(e),c=u.center,l=u.scale,d=u.translate,p=u.clipExtent,f=null;function h(){var o=r.pi*l(),c=u(Object(i.default)(u.rotate()).invert([0,0]));return p(null==f?[[c[0]-o,c[1]-o],[c[0]+o,c[1]+o]]:e===a?[[Math.max(c[0]-o,f),t],[Math.min(c[0]+o,n),s]]:[[f,Math.max(c[1]-o,t)],[n,Math.min(c[1]+o,s)]])}return u.scale=function(e){return arguments.length?(l(e),h()):l()},u.translate=function(e){return arguments.length?(d(e),h()):d()},u.center=function(e){return arguments.length?(c(e),h()):c()},u.clipExtent=function(e){return arguments.length?(null==e?f=t=n=s=null:(f=+e[0][0],t=+e[0][1],n=+e[1][0],s=+e[1][1]),h()):null==f?null:[[f,t],[n,s]]},h()}a.invert=function(e,t){return[e,2*Object(r.atan)(Object(r.exp)(t))-r.halfPi]},t.default=function(){return s(a).scale(961/r.tau)}},"./node_modules/d3-geo/src/projection/naturalEarth1.js":function(e,t,n){"use strict";n.r(t),n.d(t,"naturalEarth1Raw",function(){return o});var r=n("./node_modules/d3-geo/src/projection/index.js"),i=n("./node_modules/d3-geo/src/math.js");function o(e,t){var n=t*t,r=n*n;return[e*(.8707-.131979*n+r*(r*(.003971*n-.001529*r)-.013791)),t*(1.007226+n*(.015085+r*(.028874*n-.044475-.005916*r)))]}o.invert=function(e,t){var n,r=t,o=25;do{var a=r*r,s=a*a;r-=n=(r*(1.007226+a*(.015085+s*(.028874*a-.044475-.005916*s)))-t)/(1.007226+a*(.045255+s*(.259866*a-.311325-.005916*11*s)))}while(Object(i.abs)(n)>i.epsilon&&--o>0);return[e/(.8707+(a=r*r)*(a*(a*a*a*(.003971-.001529*a)-.013791)-.131979)),r]},t.default=function(){return Object(r.default)(o).scale(175.295)}},"./node_modules/d3-geo/src/projection/orthographic.js":function(e,t,n){"use strict";n.r(t),n.d(t,"orthographicRaw",function(){return a});var r=n("./node_modules/d3-geo/src/math.js"),i=n("./node_modules/d3-geo/src/projection/azimuthal.js"),o=n("./node_modules/d3-geo/src/projection/index.js");function a(e,t){return[Object(r.cos)(t)*Object(r.sin)(e),Object(r.sin)(t)]}a.invert=Object(i.azimuthalInvert)(r.asin),t.default=function(){return Object(o.default)(a).scale(249.5).clipAngle(90+r.epsilon)}},"./node_modules/d3-geo/src/projection/resample.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-geo/src/cartesian.js"),i=n("./node_modules/d3-geo/src/math.js"),o=n("./node_modules/d3-geo/src/transform.js"),a=16,s=Object(i.cos)(30*i.radians);t.default=function(e,t){return+t?function(e,t){function n(r,o,a,u,c,l,d,p,f,h,m,g,v,_){var y=d-r,b=p-o,j=y*y+b*b;if(j>4*t&&v--){var x=u+h,O=c+m,w=l+g,C=Object(i.sqrt)(x*x+O*O+w*w),E=Object(i.asin)(w/=C),S=Object(i.abs)(Object(i.abs)(w)-1)<i.epsilon||Object(i.abs)(a-f)<i.epsilon?(a+f)/2:Object(i.atan2)(O,x),k=e(S,E),M=k[0],A=k[1],T=M-r,D=A-o,I=b*T-y*D;(I*I/j>t||Object(i.abs)((y*T+b*D)/j-.5)>.3||u*h+c*m+l*g<s)&&(n(r,o,a,u,c,l,M,A,S,x/=C,O/=C,w,v,_),_.point(M,A),n(M,A,S,x,O,w,d,p,f,h,m,g,v,_))}}return function(t){var i,o,s,u,c,l,d,p,f,h,m,g,v={point:_,lineStart:y,lineEnd:j,polygonStart:function(){t.polygonStart(),v.lineStart=x},polygonEnd:function(){t.polygonEnd(),v.lineStart=y}};function _(n,r){n=e(n,r),t.point(n[0],n[1])}function y(){p=NaN,v.point=b,t.lineStart()}function b(i,o){var s=Object(r.cartesian)([i,o]),u=e(i,o);n(p,f,d,h,m,g,p=u[0],f=u[1],d=i,h=s[0],m=s[1],g=s[2],a,t),t.point(p,f)}function j(){v.point=_,t.lineEnd()}function x(){y(),v.point=O,v.lineEnd=w}function O(e,t){b(i=e,t),o=p,s=f,u=h,c=m,l=g,v.point=b}function w(){n(p,f,d,h,m,g,o,s,i,u,c,l,a,t),v.lineEnd=j,j()}return v}}(e,t):(n=e,Object(o.transformer)({point:function(e,t){e=n(e,t),this.stream.point(e[0],e[1])}}));var n}},"./node_modules/d3-geo/src/projection/stereographic.js":function(e,t,n){"use strict";n.r(t),n.d(t,"stereographicRaw",function(){return a});var r=n("./node_modules/d3-geo/src/math.js"),i=n("./node_modules/d3-geo/src/projection/azimuthal.js"),o=n("./node_modules/d3-geo/src/projection/index.js");function a(e,t){var n=Object(r.cos)(t),i=1+Object(r.cos)(e)*n;return[n*Object(r.sin)(e)/i,Object(r.sin)(t)/i]}a.invert=Object(i.azimuthalInvert)(function(e){return 2*Object(r.atan)(e)}),t.default=function(){return Object(o.default)(a).scale(250).clipAngle(142)}},"./node_modules/d3-geo/src/projection/transverseMercator.js":function(e,t,n){"use strict";n.r(t),n.d(t,"transverseMercatorRaw",function(){return o});var r=n("./node_modules/d3-geo/src/math.js"),i=n("./node_modules/d3-geo/src/projection/mercator.js");function o(e,t){return[Object(r.log)(Object(r.tan)((r.halfPi+t)/2)),-e]}o.invert=function(e,t){return[-t,2*Object(r.atan)(Object(r.exp)(e))-r.halfPi]},t.default=function(){var e=Object(i.mercatorProjection)(o),t=e.center,n=e.rotate;return e.center=function(e){return arguments.length?t([-e[1],e[0]]):[(e=t())[1],-e[0]]},e.rotate=function(e){return arguments.length?n([e[0],e[1],e.length>2?e[2]+90:90]):[(e=n())[0],e[1],e[2]-90]},n([0,0,90]).scale(159.155)}},"./node_modules/d3-geo/src/rotation.js":function(e,t,n){"use strict";n.r(t),n.d(t,"rotateRadians",function(){return a});var r=n("./node_modules/d3-geo/src/compose.js"),i=n("./node_modules/d3-geo/src/math.js");function o(e,t){return[e>i.pi?e-i.tau:e<-i.pi?e+i.tau:e,t]}function a(e,t,n){return(e%=i.tau)?t||n?Object(r.default)(u(e),c(t,n)):u(e):t||n?c(t,n):o}function s(e){return function(t,n){return[(t+=e)>i.pi?t-i.tau:t<-i.pi?t+i.tau:t,n]}}function u(e){var t=s(e);return t.invert=s(-e),t}function c(e,t){var n=Object(i.cos)(e),r=Object(i.sin)(e),o=Object(i.cos)(t),a=Object(i.sin)(t);function s(e,t){var s=Object(i.cos)(t),u=Object(i.cos)(e)*s,c=Object(i.sin)(e)*s,l=Object(i.sin)(t),d=l*n+u*r;return[Object(i.atan2)(c*o-d*a,u*n-l*r),Object(i.asin)(d*o+c*a)]}return s.invert=function(e,t){var s=Object(i.cos)(t),u=Object(i.cos)(e)*s,c=Object(i.sin)(e)*s,l=Object(i.sin)(t),d=l*o-c*a;return[Object(i.atan2)(c*o+l*a,u*n+d*r),Object(i.asin)(d*n-u*r)]},s}o.invert=o,t.default=function(e){function t(t){return(t=e(t[0]*i.radians,t[1]*i.radians))[0]*=i.degrees,t[1]*=i.degrees,t}return e=a(e[0]*i.radians,e[1]*i.radians,e.length>2?e[2]*i.radians:0),t.invert=function(t){return(t=e.invert(t[0]*i.radians,t[1]*i.radians))[0]*=i.degrees,t[1]*=i.degrees,t},t}},"./node_modules/d3-geo/src/stream.js":function(e,t,n){"use strict";function r(e,t){e&&o.hasOwnProperty(e.type)&&o[e.type](e,t)}n.r(t);var i={Feature:function(e,t){r(e.geometry,t)},FeatureCollection:function(e,t){for(var n=e.features,i=-1,o=n.length;++i<o;)r(n[i].geometry,t)}},o={Sphere:function(e,t){t.sphere()},Point:function(e,t){e=e.coordinates,t.point(e[0],e[1],e[2])},MultiPoint:function(e,t){for(var n=e.coordinates,r=-1,i=n.length;++r<i;)e=n[r],t.point(e[0],e[1],e[2])},LineString:function(e,t){a(e.coordinates,t,0)},MultiLineString:function(e,t){for(var n=e.coordinates,r=-1,i=n.length;++r<i;)a(n[r],t,0)},Polygon:function(e,t){s(e.coordinates,t)},MultiPolygon:function(e,t){for(var n=e.coordinates,r=-1,i=n.length;++r<i;)s(n[r],t)},GeometryCollection:function(e,t){for(var n=e.geometries,i=-1,o=n.length;++i<o;)r(n[i],t)}};function a(e,t,n){var r,i=-1,o=e.length-n;for(t.lineStart();++i<o;)r=e[i],t.point(r[0],r[1],r[2]);t.lineEnd()}function s(e,t){var n=-1,r=e.length;for(t.polygonStart();++n<r;)a(e[n],t,1);t.polygonEnd()}t.default=function(e,t){e&&i.hasOwnProperty(e.type)?i[e.type](e,t):r(e,t)}},"./node_modules/d3-geo/src/transform.js":function(e,t,n){"use strict";function r(e){return function(t){var n=new i;for(var r in e)n[r]=e[r];return n.stream=t,n}}function i(){}n.r(t),n.d(t,"transformer",function(){return r}),t.default=function(e){return{stream:r(e)}},i.prototype={constructor:i,point:function(e,t){this.stream.point(e,t)},sphere:function(){this.stream.sphere()},lineStart:function(){this.stream.lineStart()},lineEnd:function(){this.stream.lineEnd()},polygonStart:function(){this.stream.polygonStart()},polygonEnd:function(){this.stream.polygonEnd()}}},"./node_modules/d3-hierarchy/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-hierarchy/src/cluster.js");n.d(t,"cluster",function(){return r.default});var i=n("./node_modules/d3-hierarchy/src/hierarchy/index.js");n.d(t,"hierarchy",function(){return i.default});var o=n("./node_modules/d3-hierarchy/src/pack/index.js");n.d(t,"pack",function(){return o.default});var a=n("./node_modules/d3-hierarchy/src/pack/siblings.js");n.d(t,"packSiblings",function(){return a.default});var s=n("./node_modules/d3-hierarchy/src/pack/enclose.js");n.d(t,"packEnclose",function(){return s.default});var u=n("./node_modules/d3-hierarchy/src/partition.js");n.d(t,"partition",function(){return u.default});var c=n("./node_modules/d3-hierarchy/src/stratify.js");n.d(t,"stratify",function(){return c.default});var l=n("./node_modules/d3-hierarchy/src/tree.js");n.d(t,"tree",function(){return l.default});var d=n("./node_modules/d3-hierarchy/src/treemap/index.js");n.d(t,"treemap",function(){return d.default});var p=n("./node_modules/d3-hierarchy/src/treemap/binary.js");n.d(t,"treemapBinary",function(){return p.default});var f=n("./node_modules/d3-hierarchy/src/treemap/dice.js");n.d(t,"treemapDice",function(){return f.default});var h=n("./node_modules/d3-hierarchy/src/treemap/slice.js");n.d(t,"treemapSlice",function(){return h.default});var m=n("./node_modules/d3-hierarchy/src/treemap/sliceDice.js");n.d(t,"treemapSliceDice",function(){return m.default});var g=n("./node_modules/d3-hierarchy/src/treemap/squarify.js");n.d(t,"treemapSquarify",function(){return g.default});var v=n("./node_modules/d3-hierarchy/src/treemap/resquarify.js");n.d(t,"treemapResquarify",function(){return v.default})},"./node_modules/d3-hierarchy/src/accessors.js":function(e,t,n){"use strict";function r(e){return null==e?null:i(e)}function i(e){if("function"!=typeof e)throw new Error;return e}n.r(t),n.d(t,"optional",function(){return r}),n.d(t,"required",function(){return i})},"./node_modules/d3-hierarchy/src/array.js":function(e,t,n){"use strict";n.r(t),n.d(t,"slice",function(){return r}),n.d(t,"shuffle",function(){return i});var r=Array.prototype.slice;function i(e){for(var t,n,r=e.length;r;)n=Math.random()*r--|0,t=e[r],e[r]=e[n],e[n]=t;return e}},"./node_modules/d3-hierarchy/src/cluster.js":function(e,t,n){"use strict";function r(e,t){return e.parent===t.parent?1:2}function i(e,t){return e+t.x}function o(e,t){return Math.max(e,t.y)}n.r(t),t.default=function(){var e=r,t=1,n=1,a=!1;function s(r){var s,u=0;r.eachAfter(function(t){var n,r=t.children;r?(t.x=(n=r).reduce(i,0)/n.length,t.y=1+r.reduce(o,0)):(t.x=s?u+=e(t,s):0,t.y=0,s=t)});var c=function(e){for(var t;t=e.children;)e=t[0];return e}(r),l=function(e){for(var t;t=e.children;)e=t[t.length-1];return e}(r),d=c.x-e(c,l)/2,p=l.x+e(l,c)/2;return r.eachAfter(a?function(e){e.x=(e.x-r.x)*t,e.y=(r.y-e.y)*n}:function(e){e.x=(e.x-d)/(p-d)*t,e.y=(1-(r.y?e.y/r.y:1))*n})}return s.separation=function(t){return arguments.length?(e=t,s):e},s.size=function(e){return arguments.length?(a=!1,t=+e[0],n=+e[1],s):a?null:[t,n]},s.nodeSize=function(e){return arguments.length?(a=!0,t=+e[0],n=+e[1],s):a?[t,n]:null},s}},"./node_modules/d3-hierarchy/src/constant.js":function(e,t,n){"use strict";function r(){return 0}n.r(t),n.d(t,"constantZero",function(){return r}),t.default=function(e){return function(){return e}}},"./node_modules/d3-hierarchy/src/hierarchy/ancestors.js":function(e,t,n){"use strict";n.r(t),t.default=function(){for(var e=this,t=[e];e=e.parent;)t.push(e);return t}},"./node_modules/d3-hierarchy/src/hierarchy/count.js":function(e,t,n){"use strict";function r(e){var t=0,n=e.children,r=n&&n.length;if(r)for(;--r>=0;)t+=n[r].value;else t=1;e.value=t}n.r(t),t.default=function(){return this.eachAfter(r)}},"./node_modules/d3-hierarchy/src/hierarchy/descendants.js":function(e,t,n){"use strict";n.r(t),t.default=function(){var e=[];return this.each(function(t){e.push(t)}),e}},"./node_modules/d3-hierarchy/src/hierarchy/each.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){var t,n,r,i,o=this,a=[o];do{for(t=a.reverse(),a=[];o=t.pop();)if(e(o),n=o.children)for(r=0,i=n.length;r<i;++r)a.push(n[r])}while(a.length);return this}},"./node_modules/d3-hierarchy/src/hierarchy/eachAfter.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){for(var t,n,r,i=this,o=[i],a=[];i=o.pop();)if(a.push(i),t=i.children)for(n=0,r=t.length;n<r;++n)o.push(t[n]);for(;i=a.pop();)e(i);return this}},"./node_modules/d3-hierarchy/src/hierarchy/eachBefore.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){for(var t,n,r=this,i=[r];r=i.pop();)if(e(r),t=r.children)for(n=t.length-1;n>=0;--n)i.push(t[n]);return this}},"./node_modules/d3-hierarchy/src/hierarchy/index.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return h}),n.d(t,"computeHeight",function(){return v}),n.d(t,"Node",function(){return _});var r=n("./node_modules/d3-hierarchy/src/hierarchy/count.js"),i=n("./node_modules/d3-hierarchy/src/hierarchy/each.js"),o=n("./node_modules/d3-hierarchy/src/hierarchy/eachBefore.js"),a=n("./node_modules/d3-hierarchy/src/hierarchy/eachAfter.js"),s=n("./node_modules/d3-hierarchy/src/hierarchy/sum.js"),u=n("./node_modules/d3-hierarchy/src/hierarchy/sort.js"),c=n("./node_modules/d3-hierarchy/src/hierarchy/path.js"),l=n("./node_modules/d3-hierarchy/src/hierarchy/ancestors.js"),d=n("./node_modules/d3-hierarchy/src/hierarchy/descendants.js"),p=n("./node_modules/d3-hierarchy/src/hierarchy/leaves.js"),f=n("./node_modules/d3-hierarchy/src/hierarchy/links.js");function h(e,t){var n,r,i,o,a,s=new _(e),u=+e.value&&(s.value=e.value),c=[s];for(null==t&&(t=m);n=c.pop();)if(u&&(n.value=+n.data.value),(i=t(n.data))&&(a=i.length))for(n.children=new Array(a),o=a-1;o>=0;--o)c.push(r=n.children[o]=new _(i[o])),r.parent=n,r.depth=n.depth+1;return s.eachBefore(v)}function m(e){return e.children}function g(e){e.data=e.data.data}function v(e){var t=0;do{e.height=t}while((e=e.parent)&&e.height<++t)}function _(e){this.data=e,this.depth=this.height=0,this.parent=null}_.prototype=h.prototype={constructor:_,count:r.default,each:i.default,eachAfter:a.default,eachBefore:o.default,sum:s.default,sort:u.default,path:c.default,ancestors:l.default,descendants:d.default,leaves:p.default,links:f.default,copy:function(){return h(this).eachBefore(g)}}},"./node_modules/d3-hierarchy/src/hierarchy/leaves.js":function(e,t,n){"use strict";n.r(t),t.default=function(){var e=[];return this.eachBefore(function(t){t.children||e.push(t)}),e}},"./node_modules/d3-hierarchy/src/hierarchy/links.js":function(e,t,n){"use strict";n.r(t),t.default=function(){var e=this,t=[];return e.each(function(n){n!==e&&t.push({source:n.parent,target:n})}),t}},"./node_modules/d3-hierarchy/src/hierarchy/path.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){for(var t=this,n=function(e,t){if(e===t)return e;var n=e.ancestors(),r=t.ancestors(),i=null;e=n.pop(),t=r.pop();for(;e===t;)i=e,e=n.pop(),t=r.pop();return i}(t,e),r=[t];t!==n;)t=t.parent,r.push(t);for(var i=r.length;e!==n;)r.splice(i,0,e),e=e.parent;return r}},"./node_modules/d3-hierarchy/src/hierarchy/sort.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return this.eachBefore(function(t){t.children&&t.children.sort(e)})}},"./node_modules/d3-hierarchy/src/hierarchy/sum.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return this.eachAfter(function(t){for(var n=+e(t.data)||0,r=t.children,i=r&&r.length;--i>=0;)n+=r[i].value;t.value=n})}},"./node_modules/d3-hierarchy/src/pack/enclose.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-hierarchy/src/array.js");function i(e,t){var n,r;if(s(t,e))return[t];for(n=0;n<e.length;++n)if(o(t,e[n])&&s(c(e[n],t),e))return[e[n],t];for(n=0;n<e.length-1;++n)for(r=n+1;r<e.length;++r)if(o(c(e[n],e[r]),t)&&o(c(e[n],t),e[r])&&o(c(e[r],t),e[n])&&s(l(e[n],e[r],t),e))return[e[n],e[r],t];throw new Error}function o(e,t){var n=e.r-t.r,r=t.x-e.x,i=t.y-e.y;return n<0||n*n<r*r+i*i}function a(e,t){var n=e.r-t.r+1e-6,r=t.x-e.x,i=t.y-e.y;return n>0&&n*n>r*r+i*i}function s(e,t){for(var n=0;n<t.length;++n)if(!a(e,t[n]))return!1;return!0}function u(e){switch(e.length){case 1:return{x:(t=e[0]).x,y:t.y,r:t.r};case 2:return c(e[0],e[1]);case 3:return l(e[0],e[1],e[2])}var t}function c(e,t){var n=e.x,r=e.y,i=e.r,o=t.x,a=t.y,s=t.r,u=o-n,c=a-r,l=s-i,d=Math.sqrt(u*u+c*c);return{x:(n+o+u/d*l)/2,y:(r+a+c/d*l)/2,r:(d+i+s)/2}}function l(e,t,n){var r=e.x,i=e.y,o=e.r,a=t.x,s=t.y,u=t.r,c=n.x,l=n.y,d=n.r,p=r-a,f=r-c,h=i-s,m=i-l,g=u-o,v=d-o,_=r*r+i*i-o*o,y=_-a*a-s*s+u*u,b=_-c*c-l*l+d*d,j=f*h-p*m,x=(h*b-m*y)/(2*j)-r,O=(m*g-h*v)/j,w=(f*y-p*b)/(2*j)-i,C=(p*v-f*g)/j,E=O*O+C*C-1,S=2*(o+x*O+w*C),k=x*x+w*w-o*o,M=-(E?(S+Math.sqrt(S*S-4*E*k))/(2*E):k/S);return{x:r+x+O*M,y:i+w+C*M,r:M}}t.default=function(e){for(var t,n,o=0,s=(e=Object(r.shuffle)(r.slice.call(e))).length,c=[];o<s;)t=e[o],n&&a(n,t)?++o:(n=u(c=i(c,t)),o=0);return n}},"./node_modules/d3-hierarchy/src/pack/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-hierarchy/src/pack/siblings.js"),i=n("./node_modules/d3-hierarchy/src/accessors.js"),o=n("./node_modules/d3-hierarchy/src/constant.js");function a(e){return Math.sqrt(e.value)}function s(e){return function(t){t.children||(t.r=Math.max(0,+e(t)||0))}}function u(e,t){return function(n){if(i=n.children){var i,o,a,s=i.length,u=e(n)*t||0;if(u)for(o=0;o<s;++o)i[o].r+=u;if(a=Object(r.packEnclose)(i),u)for(o=0;o<s;++o)i[o].r-=u;n.r=a+u}}}function c(e){return function(t){var n=t.parent;t.r*=e,n&&(t.x=n.x+e*t.x,t.y=n.y+e*t.y)}}t.default=function(){var e=null,t=1,n=1,r=o.constantZero;function l(i){return i.x=t/2,i.y=n/2,e?i.eachBefore(s(e)).eachAfter(u(r,.5)).eachBefore(c(1)):i.eachBefore(s(a)).eachAfter(u(o.constantZero,1)).eachAfter(u(r,i.r/Math.min(t,n))).eachBefore(c(Math.min(t,n)/(2*i.r))),i}return l.radius=function(t){return arguments.length?(e=Object(i.optional)(t),l):e},l.size=function(e){return arguments.length?(t=+e[0],n=+e[1],l):[t,n]},l.padding=function(e){return arguments.length?(r="function"==typeof e?e:Object(o.default)(+e),l):r},l}},"./node_modules/d3-hierarchy/src/pack/siblings.js":function(e,t,n){"use strict";n.r(t),n.d(t,"packEnclose",function(){return u});var r=n("./node_modules/d3-hierarchy/src/pack/enclose.js");function i(e,t,n){var r=e.x,i=e.y,o=t.r+n.r,a=e.r+n.r,s=t.x-r,u=t.y-i,c=s*s+u*u;if(c){var l=.5+((a*=a)-(o*=o))/(2*c),d=Math.sqrt(Math.max(0,2*o*(a+c)-(a-=c)*a-o*o))/(2*c);n.x=r+l*s+d*u,n.y=i+l*u-d*s}else n.x=r+a,n.y=i}function o(e,t){var n=t.x-e.x,r=t.y-e.y,i=e.r+t.r;return i*i-1e-6>n*n+r*r}function a(e){var t=e._,n=e.next._,r=t.r+n.r,i=(t.x*n.r+n.x*t.r)/r,o=(t.y*n.r+n.y*t.r)/r;return i*i+o*o}function s(e){this._=e,this.next=null,this.previous=null}function u(e){if(!(c=e.length))return 0;var t,n,u,c,l,d,p,f,h,m,g;if((t=e[0]).x=0,t.y=0,!(c>1))return t.r;if(n=e[1],t.x=-n.r,n.x=t.r,n.y=0,!(c>2))return t.r+n.r;i(n,t,u=e[2]),t=new s(t),n=new s(n),u=new s(u),t.next=u.previous=n,n.next=t.previous=u,u.next=n.previous=t;e:for(p=3;p<c;++p){i(t._,n._,u=e[p]),u=new s(u),f=n.next,h=t.previous,m=n._.r,g=t._.r;do{if(m<=g){if(o(f._,u._)){n=f,t.next=n,n.previous=t,--p;continue e}m+=f._.r,f=f.next}else{if(o(h._,u._)){(t=h).next=n,n.previous=t,--p;continue e}g+=h._.r,h=h.previous}}while(f!==h.next);for(u.previous=t,u.next=n,t.next=n.previous=n=u,l=a(t);(u=u.next)!==n;)(d=a(u))<l&&(t=u,l=d);n=t.next}for(t=[n._],u=n;(u=u.next)!==n;)t.push(u._);for(u=Object(r.default)(t),p=0;p<c;++p)t=e[p],t.x-=u.x,t.y-=u.y;return u.r}t.default=function(e){return u(e),e}},"./node_modules/d3-hierarchy/src/partition.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-hierarchy/src/treemap/round.js"),i=n("./node_modules/d3-hierarchy/src/treemap/dice.js");t.default=function(){var e=1,t=1,n=0,o=!1;function a(a){var s,u,c=a.height+1;return a.x0=a.y0=n,a.x1=e,a.y1=t/c,a.eachBefore((s=t,u=c,function(e){e.children&&Object(i.default)(e,e.x0,s*(e.depth+1)/u,e.x1,s*(e.depth+2)/u);var t=e.x0,r=e.y0,o=e.x1-n,a=e.y1-n;o<t&&(t=o=(t+o)/2),a<r&&(r=a=(r+a)/2),e.x0=t,e.y0=r,e.x1=o,e.y1=a})),o&&a.eachBefore(r.default),a}return a.round=function(e){return arguments.length?(o=!!e,a):o},a.size=function(n){return arguments.length?(e=+n[0],t=+n[1],a):[e,t]},a.padding=function(e){return arguments.length?(n=+e,a):n},a}},"./node_modules/d3-hierarchy/src/stratify.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-hierarchy/src/accessors.js"),i=n("./node_modules/d3-hierarchy/src/hierarchy/index.js"),o="$",a={depth:-1},s={};function u(e){return e.id}function c(e){return e.parentId}t.default=function(){var e=u,t=c;function n(n){var r,u,c,l,d,p,f,h=n.length,m=new Array(h),g={};for(u=0;u<h;++u)r=n[u],d=m[u]=new i.Node(r),null!=(p=e(r,u,n))&&(p+="")&&(g[f=o+(d.id=p)]=f in g?s:d);for(u=0;u<h;++u)if(d=m[u],null!=(p=t(n[u],u,n))&&(p+="")){if(!(l=g[o+p]))throw new Error("missing: "+p);if(l===s)throw new Error("ambiguous: "+p);l.children?l.children.push(d):l.children=[d],d.parent=l}else{if(c)throw new Error("multiple roots");c=d}if(!c)throw new Error("no root");if(c.parent=a,c.eachBefore(function(e){e.depth=e.parent.depth+1,--h}).eachBefore(i.computeHeight),c.parent=null,h>0)throw new Error("cycle");return c}return n.id=function(t){return arguments.length?(e=Object(r.required)(t),n):e},n.parentId=function(e){return arguments.length?(t=Object(r.required)(e),n):t},n}},"./node_modules/d3-hierarchy/src/tree.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-hierarchy/src/hierarchy/index.js");function i(e,t){return e.parent===t.parent?1:2}function o(e){var t=e.children;return t?t[0]:e.t}function a(e){var t=e.children;return t?t[t.length-1]:e.t}function s(e,t){this._=e,this.parent=null,this.children=null,this.A=null,this.a=this,this.z=0,this.m=0,this.c=0,this.s=0,this.t=null,this.i=t}s.prototype=Object.create(r.Node.prototype),t.default=function(){var e=i,t=1,n=1,r=null;function u(i){var o=function(e){for(var t,n,r,i,o,a=new s(e,0),u=[a];t=u.pop();)if(r=t._.children)for(t.children=new Array(o=r.length),i=o-1;i>=0;--i)u.push(n=t.children[i]=new s(r[i],i)),n.parent=t;return(a.parent=new s(null,0)).children=[a],a}(i);if(o.eachAfter(c),o.parent.m=-o.z,o.eachBefore(l),r)i.eachBefore(d);else{var a=i,u=i,p=i;i.eachBefore(function(e){e.x<a.x&&(a=e),e.x>u.x&&(u=e),e.depth>p.depth&&(p=e)});var f=a===u?1:e(a,u)/2,h=f-a.x,m=t/(u.x+f+h),g=n/(p.depth||1);i.eachBefore(function(e){e.x=(e.x+h)*m,e.y=e.depth*g})}return i}function c(t){var n=t.children,r=t.parent.children,i=t.i?r[t.i-1]:null;if(n){!function(e){for(var t,n=0,r=0,i=e.children,o=i.length;--o>=0;)(t=i[o]).z+=n,t.m+=n,n+=t.s+(r+=t.c)}(t);var s=(n[0].z+n[n.length-1].z)/2;i?(t.z=i.z+e(t._,i._),t.m=t.z-s):t.z=s}else i&&(t.z=i.z+e(t._,i._));t.parent.A=function(t,n,r){if(n){for(var i,s=t,u=t,c=n,l=s.parent.children[0],d=s.m,p=u.m,f=c.m,h=l.m;c=a(c),s=o(s),c&&s;)l=o(l),(u=a(u)).a=t,(i=c.z+f-s.z-d+e(c._,s._))>0&&(b=t,j=r,m=(y=c).a.parent===b.parent?y.a:j,void 0,_=(v=i)/((g=t).i-m.i),g.c-=_,g.s+=v,m.c+=_,g.z+=v,g.m+=v,d+=i,p+=i),f+=c.m,d+=s.m,h+=l.m,p+=u.m;c&&!a(u)&&(u.t=c,u.m+=f-p),s&&!o(l)&&(l.t=s,l.m+=d-h,r=t)}var m,g,v,_;var y,b,j;return r}(t,i,t.parent.A||r[0])}function l(e){e._.x=e.z+e.parent.m,e.m+=e.parent.m}function d(e){e.x*=t,e.y=e.depth*n}return u.separation=function(t){return arguments.length?(e=t,u):e},u.size=function(e){return arguments.length?(r=!1,t=+e[0],n=+e[1],u):r?null:[t,n]},u.nodeSize=function(e){return arguments.length?(r=!0,t=+e[0],n=+e[1],u):r?[t,n]:null},u}},"./node_modules/d3-hierarchy/src/treemap/binary.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n,r,i){var o,a,s=e.children,u=s.length,c=new Array(u+1);for(c[0]=a=o=0;o<u;++o)c[o+1]=a+=s[o].value;!function e(t,n,r,i,o,a,u){if(t>=n-1){var l=s[t];return l.x0=i,l.y0=o,l.x1=a,void(l.y1=u)}var d=c[t],p=r/2+d,f=t+1,h=n-1;for(;f<h;){var m=f+h>>>1;c[m]<p?f=m+1:h=m}p-c[f-1]<c[f]-p&&t+1<f&&--f;var g=c[f]-d,v=r-g;if(a-i>u-o){var _=(i*v+a*g)/r;e(t,f,g,i,o,_,u),e(f,n,v,_,o,a,u)}else{var y=(o*v+u*g)/r;e(t,f,g,i,o,a,y),e(f,n,v,i,y,a,u)}}(0,u,e.value,t,n,r,i)}},"./node_modules/d3-hierarchy/src/treemap/dice.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n,r,i){for(var o,a=e.children,s=-1,u=a.length,c=e.value&&(r-t)/e.value;++s<u;)(o=a[s]).y0=n,o.y1=i,o.x0=t,o.x1=t+=o.value*c}},"./node_modules/d3-hierarchy/src/treemap/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-hierarchy/src/treemap/round.js"),i=n("./node_modules/d3-hierarchy/src/treemap/squarify.js"),o=n("./node_modules/d3-hierarchy/src/accessors.js"),a=n("./node_modules/d3-hierarchy/src/constant.js");t.default=function(){var e=i.default,t=!1,n=1,s=1,u=[0],c=a.constantZero,l=a.constantZero,d=a.constantZero,p=a.constantZero,f=a.constantZero;function h(e){return e.x0=e.y0=0,e.x1=n,e.y1=s,e.eachBefore(m),u=[0],t&&e.eachBefore(r.default),e}function m(t){var n=u[t.depth],r=t.x0+n,i=t.y0+n,o=t.x1-n,a=t.y1-n;o<r&&(r=o=(r+o)/2),a<i&&(i=a=(i+a)/2),t.x0=r,t.y0=i,t.x1=o,t.y1=a,t.children&&(n=u[t.depth+1]=c(t)/2,r+=f(t)-n,i+=l(t)-n,o-=d(t)-n,a-=p(t)-n,o<r&&(r=o=(r+o)/2),a<i&&(i=a=(i+a)/2),e(t,r,i,o,a))}return h.round=function(e){return arguments.length?(t=!!e,h):t},h.size=function(e){return arguments.length?(n=+e[0],s=+e[1],h):[n,s]},h.tile=function(t){return arguments.length?(e=Object(o.required)(t),h):e},h.padding=function(e){return arguments.length?h.paddingInner(e).paddingOuter(e):h.paddingInner()},h.paddingInner=function(e){return arguments.length?(c="function"==typeof e?e:Object(a.default)(+e),h):c},h.paddingOuter=function(e){return arguments.length?h.paddingTop(e).paddingRight(e).paddingBottom(e).paddingLeft(e):h.paddingTop()},h.paddingTop=function(e){return arguments.length?(l="function"==typeof e?e:Object(a.default)(+e),h):l},h.paddingRight=function(e){return arguments.length?(d="function"==typeof e?e:Object(a.default)(+e),h):d},h.paddingBottom=function(e){return arguments.length?(p="function"==typeof e?e:Object(a.default)(+e),h):p},h.paddingLeft=function(e){return arguments.length?(f="function"==typeof e?e:Object(a.default)(+e),h):f},h}},"./node_modules/d3-hierarchy/src/treemap/resquarify.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-hierarchy/src/treemap/dice.js"),i=n("./node_modules/d3-hierarchy/src/treemap/slice.js"),o=n("./node_modules/d3-hierarchy/src/treemap/squarify.js");t.default=function e(t){function n(e,n,a,s,u){if((c=e._squarify)&&c.ratio===t)for(var c,l,d,p,f,h=-1,m=c.length,g=e.value;++h<m;){for(d=(l=c[h]).children,p=l.value=0,f=d.length;p<f;++p)l.value+=d[p].value;l.dice?Object(r.default)(l,n,a,s,a+=(u-a)*l.value/g):Object(i.default)(l,n,a,n+=(s-n)*l.value/g,u),g-=l.value}else e._squarify=c=Object(o.squarifyRatio)(t,e,n,a,s,u),c.ratio=t}return n.ratio=function(t){return e((t=+t)>1?t:1)},n}(o.phi)},"./node_modules/d3-hierarchy/src/treemap/round.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){e.x0=Math.round(e.x0),e.y0=Math.round(e.y0),e.x1=Math.round(e.x1),e.y1=Math.round(e.y1)}},"./node_modules/d3-hierarchy/src/treemap/slice.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n,r,i){for(var o,a=e.children,s=-1,u=a.length,c=e.value&&(i-n)/e.value;++s<u;)(o=a[s]).x0=t,o.x1=r,o.y0=n,o.y1=n+=o.value*c}},"./node_modules/d3-hierarchy/src/treemap/sliceDice.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-hierarchy/src/treemap/dice.js"),i=n("./node_modules/d3-hierarchy/src/treemap/slice.js");t.default=function(e,t,n,o,a){(1&e.depth?i.default:r.default)(e,t,n,o,a)}},"./node_modules/d3-hierarchy/src/treemap/squarify.js":function(e,t,n){"use strict";n.r(t),n.d(t,"phi",function(){return o}),n.d(t,"squarifyRatio",function(){return a});var r=n("./node_modules/d3-hierarchy/src/treemap/dice.js"),i=n("./node_modules/d3-hierarchy/src/treemap/slice.js"),o=(1+Math.sqrt(5))/2;function a(e,t,n,o,a,s){for(var u,c,l,d,p,f,h,m,g,v,_,y=[],b=t.children,j=0,x=0,O=b.length,w=t.value;j<O;){l=a-n,d=s-o;do{p=b[x++].value}while(!p&&x<O);for(f=h=p,_=p*p*(v=Math.max(d/l,l/d)/(w*e)),g=Math.max(h/_,_/f);x<O;++x){if(p+=c=b[x].value,c<f&&(f=c),c>h&&(h=c),_=p*p*v,(m=Math.max(h/_,_/f))>g){p-=c;break}g=m}y.push(u={value:p,dice:l<d,children:b.slice(j,x)}),u.dice?Object(r.default)(u,n,o,a,w?o+=d*p/w:s):Object(i.default)(u,n,o,w?n+=l*p/w:a,s),w-=p,j=x}return y}t.default=function e(t){function n(e,n,r,i,o){a(t,e,n,r,i,o)}return n.ratio=function(t){return e((t=+t)>1?t:1)},n}(o)},"./node_modules/d3-interpolate/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-interpolate/src/value.js");n.d(t,"interpolate",function(){return r.default});var i=n("./node_modules/d3-interpolate/src/array.js");n.d(t,"interpolateArray",function(){return i.default});var o=n("./node_modules/d3-interpolate/src/basis.js");n.d(t,"interpolateBasis",function(){return o.default});var a=n("./node_modules/d3-interpolate/src/basisClosed.js");n.d(t,"interpolateBasisClosed",function(){return a.default});var s=n("./node_modules/d3-interpolate/src/date.js");n.d(t,"interpolateDate",function(){return s.default});var u=n("./node_modules/d3-interpolate/src/number.js");n.d(t,"interpolateNumber",function(){return u.default});var c=n("./node_modules/d3-interpolate/src/object.js");n.d(t,"interpolateObject",function(){return c.default});var l=n("./node_modules/d3-interpolate/src/round.js");n.d(t,"interpolateRound",function(){return l.default});var d=n("./node_modules/d3-interpolate/src/string.js");n.d(t,"interpolateString",function(){return d.default});var p=n("./node_modules/d3-interpolate/src/transform/index.js");n.d(t,"interpolateTransformCss",function(){return p.interpolateTransformCss}),n.d(t,"interpolateTransformSvg",function(){return p.interpolateTransformSvg});var f=n("./node_modules/d3-interpolate/src/zoom.js");n.d(t,"interpolateZoom",function(){return f.default});var h=n("./node_modules/d3-interpolate/src/rgb.js");n.d(t,"interpolateRgb",function(){return h.default}),n.d(t,"interpolateRgbBasis",function(){return h.rgbBasis}),n.d(t,"interpolateRgbBasisClosed",function(){return h.rgbBasisClosed});var m=n("./node_modules/d3-interpolate/src/hsl.js");n.d(t,"interpolateHsl",function(){return m.default}),n.d(t,"interpolateHslLong",function(){return m.hslLong});var g=n("./node_modules/d3-interpolate/src/lab.js");n.d(t,"interpolateLab",function(){return g.default});var v=n("./node_modules/d3-interpolate/src/hcl.js");n.d(t,"interpolateHcl",function(){return v.default}),n.d(t,"interpolateHclLong",function(){return v.hclLong});var _=n("./node_modules/d3-interpolate/src/cubehelix.js");n.d(t,"interpolateCubehelix",function(){return _.default}),n.d(t,"interpolateCubehelixLong",function(){return _.cubehelixLong});var y=n("./node_modules/d3-interpolate/src/quantize.js");n.d(t,"quantize",function(){return y.default})},"./node_modules/d3-interpolate/src/array.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-interpolate/src/value.js");t.default=function(e,t){var n,i=t?t.length:0,o=e?Math.min(i,e.length):0,a=new Array(o),s=new Array(i);for(n=0;n<o;++n)a[n]=Object(r.default)(e[n],t[n]);for(;n<i;++n)s[n]=t[n];return function(e){for(n=0;n<o;++n)s[n]=a[n](e);return s}}},"./node_modules/d3-interpolate/src/basis.js":function(e,t,n){"use strict";function r(e,t,n,r,i){var o=e*e,a=o*e;return((1-3*e+3*o-a)*t+(4-6*o+3*a)*n+(1+3*e+3*o-3*a)*r+a*i)/6}n.r(t),n.d(t,"basis",function(){return r}),t.default=function(e){var t=e.length-1;return function(n){var i=n<=0?n=0:n>=1?(n=1,t-1):Math.floor(n*t),o=e[i],a=e[i+1],s=i>0?e[i-1]:2*o-a,u=i<t-1?e[i+2]:2*a-o;return r((n-i/t)*t,s,o,a,u)}}},"./node_modules/d3-interpolate/src/basisClosed.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-interpolate/src/basis.js");t.default=function(e){var t=e.length;return function(n){var i=Math.floor(((n%=1)<0?++n:n)*t),o=e[(i+t-1)%t],a=e[i%t],s=e[(i+1)%t],u=e[(i+2)%t];return Object(r.basis)((n-i/t)*t,o,a,s,u)}}},"./node_modules/d3-interpolate/src/color.js":function(e,t,n){"use strict";n.r(t),n.d(t,"hue",function(){return o}),n.d(t,"gamma",function(){return a}),n.d(t,"default",function(){return s});var r=n("./node_modules/d3-interpolate/src/constant.js");function i(e,t){return function(n){return e+n*t}}function o(e,t){var n=t-e;return n?i(e,n>180||n<-180?n-360*Math.round(n/360):n):Object(r.default)(isNaN(e)?t:e)}function a(e){return 1==(e=+e)?s:function(t,n){return n-t?(i=t,o=n,a=e,i=Math.pow(i,a),o=Math.pow(o,a)-i,a=1/a,function(e){return Math.pow(i+e*o,a)}):Object(r.default)(isNaN(t)?n:t);var i,o,a}}function s(e,t){var n=t-e;return n?i(e,n):Object(r.default)(isNaN(e)?t:e)}},"./node_modules/d3-interpolate/src/constant.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return function(){return e}}},"./node_modules/d3-interpolate/src/cubehelix.js":function(e,t,n){"use strict";n.r(t),n.d(t,"cubehelixLong",function(){return a});var r=n("./node_modules/d3-color/index.js"),i=n("./node_modules/d3-interpolate/src/color.js");function o(e){return function t(n){function o(t,o){var a=e((t=Object(r.cubehelix)(t)).h,(o=Object(r.cubehelix)(o)).h),s=Object(i.default)(t.s,o.s),u=Object(i.default)(t.l,o.l),c=Object(i.default)(t.opacity,o.opacity);return function(e){return t.h=a(e),t.s=s(e),t.l=u(Math.pow(e,n)),t.opacity=c(e),t+""}}return n=+n,o.gamma=t,o}(1)}t.default=o(i.hue);var a=o(i.default)},"./node_modules/d3-interpolate/src/date.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){var n=new Date;return t-=e=+e,function(r){return n.setTime(e+t*r),n}}},"./node_modules/d3-interpolate/src/hcl.js":function(e,t,n){"use strict";n.r(t),n.d(t,"hclLong",function(){return a});var r=n("./node_modules/d3-color/index.js"),i=n("./node_modules/d3-interpolate/src/color.js");function o(e){return function(t,n){var o=e((t=Object(r.hcl)(t)).h,(n=Object(r.hcl)(n)).h),a=Object(i.default)(t.c,n.c),s=Object(i.default)(t.l,n.l),u=Object(i.default)(t.opacity,n.opacity);return function(e){return t.h=o(e),t.c=a(e),t.l=s(e),t.opacity=u(e),t+""}}}t.default=o(i.hue);var a=o(i.default)},"./node_modules/d3-interpolate/src/hsl.js":function(e,t,n){"use strict";n.r(t),n.d(t,"hslLong",function(){return a});var r=n("./node_modules/d3-color/index.js"),i=n("./node_modules/d3-interpolate/src/color.js");function o(e){return function(t,n){var o=e((t=Object(r.hsl)(t)).h,(n=Object(r.hsl)(n)).h),a=Object(i.default)(t.s,n.s),s=Object(i.default)(t.l,n.l),u=Object(i.default)(t.opacity,n.opacity);return function(e){return t.h=o(e),t.s=a(e),t.l=s(e),t.opacity=u(e),t+""}}}t.default=o(i.hue);var a=o(i.default)},"./node_modules/d3-interpolate/src/lab.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/d3-color/index.js"),i=n("./node_modules/d3-interpolate/src/color.js");function o(e,t){var n=Object(i.default)((e=Object(r.lab)(e)).l,(t=Object(r.lab)(t)).l),o=Object(i.default)(e.a,t.a),a=Object(i.default)(e.b,t.b),s=Object(i.default)(e.opacity,t.opacity);return function(t){return e.l=n(t),e.a=o(t),e.b=a(t),e.opacity=s(t),e+""}}},"./node_modules/d3-interpolate/src/number.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){return t-=e=+e,function(n){return e+t*n}}},"./node_modules/d3-interpolate/src/object.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-interpolate/src/value.js");t.default=function(e,t){var n,i={},o={};null!==e&&"object"==typeof e||(e={}),null!==t&&"object"==typeof t||(t={});for(n in t)n in e?i[n]=Object(r.default)(e[n],t[n]):o[n]=t[n];return function(e){for(n in i)o[n]=i[n](e);return o}}},"./node_modules/d3-interpolate/src/quantize.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){for(var n=new Array(t),r=0;r<t;++r)n[r]=e(r/(t-1));return n}},"./node_modules/d3-interpolate/src/rgb.js":function(e,t,n){"use strict";n.r(t),n.d(t,"rgbBasis",function(){return u}),n.d(t,"rgbBasisClosed",function(){return c});var r=n("./node_modules/d3-color/index.js"),i=n("./node_modules/d3-interpolate/src/basis.js"),o=n("./node_modules/d3-interpolate/src/basisClosed.js"),a=n("./node_modules/d3-interpolate/src/color.js");function s(e){return function(t){var n,i,o=t.length,a=new Array(o),s=new Array(o),u=new Array(o);for(n=0;n<o;++n)i=Object(r.rgb)(t[n]),a[n]=i.r||0,s[n]=i.g||0,u[n]=i.b||0;return a=e(a),s=e(s),u=e(u),i.opacity=1,function(e){return i.r=a(e),i.g=s(e),i.b=u(e),i+""}}}t.default=function e(t){var n=Object(a.gamma)(t);function i(e,t){var i=n((e=Object(r.rgb)(e)).r,(t=Object(r.rgb)(t)).r),o=n(e.g,t.g),s=n(e.b,t.b),u=Object(a.default)(e.opacity,t.opacity);return function(t){return e.r=i(t),e.g=o(t),e.b=s(t),e.opacity=u(t),e+""}}return i.gamma=e,i}(1);var u=s(i.default),c=s(o.default)},"./node_modules/d3-interpolate/src/round.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){return t-=e=+e,function(n){return Math.round(e+t*n)}}},"./node_modules/d3-interpolate/src/string.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-interpolate/src/number.js"),i=/[-+]?(?:\d+\.?\d*|\.?\d+)(?:[eE][-+]?\d+)?/g,o=new RegExp(i.source,"g");t.default=function(e,t){var n,a,s,u,c,l=i.lastIndex=o.lastIndex=0,d=-1,p=[],f=[];for(e+="",t+="";(n=i.exec(e))&&(a=o.exec(t));)(s=a.index)>l&&(s=t.slice(l,s),p[d]?p[d]+=s:p[++d]=s),(n=n[0])===(a=a[0])?p[d]?p[d]+=a:p[++d]=a:(p[++d]=null,f.push({i:d,x:Object(r.default)(n,a)})),l=o.lastIndex;return l<t.length&&(s=t.slice(l),p[d]?p[d]+=s:p[++d]=s),p.length<2?f[0]?(c=f[0].x,function(e){return c(e)+""}):(u=t,function(){return u}):(t=f.length,function(e){for(var n,r=0;r<t;++r)p[(n=f[r]).i]=n.x(e);return p.join("")})}},"./node_modules/d3-interpolate/src/transform/decompose.js":function(e,t,n){"use strict";n.r(t),n.d(t,"identity",function(){return i});var r=180/Math.PI,i={translateX:0,translateY:0,rotate:0,skewX:0,scaleX:1,scaleY:1};t.default=function(e,t,n,i,o,a){var s,u,c;return(s=Math.sqrt(e*e+t*t))&&(e/=s,t/=s),(c=e*n+t*i)&&(n-=e*c,i-=t*c),(u=Math.sqrt(n*n+i*i))&&(n/=u,i/=u,c/=u),e*i<t*n&&(e=-e,t=-t,c=-c,s=-s),{translateX:o,translateY:a,rotate:Math.atan2(t,e)*r,skewX:Math.atan(c)*r,scaleX:s,scaleY:u}}},"./node_modules/d3-interpolate/src/transform/index.js":function(e,t,n){"use strict";n.r(t),n.d(t,"interpolateTransformCss",function(){return a}),n.d(t,"interpolateTransformSvg",function(){return s});var r=n("./node_modules/d3-interpolate/src/number.js"),i=n("./node_modules/d3-interpolate/src/transform/parse.js");function o(e,t,n,i){function o(e){return e.length?e.pop()+" ":""}return function(a,s){var u,c,l,d,p,f,h=[],m=[];return a=e(a),s=e(s),function(e,i,o,a,s,u){if(e!==o||i!==a){var c=s.push("translate(",null,t,null,n);u.push({i:c-4,x:Object(r.default)(e,o)},{i:c-2,x:Object(r.default)(i,a)})}else(o||a)&&s.push("translate("+o+t+a+n)}(a.translateX,a.translateY,s.translateX,s.translateY,h,m),u=a.rotate,c=s.rotate,l=h,u!==c?(u-c>180?c+=360:c-u>180&&(u+=360),m.push({i:l.push(o(l)+"rotate(",null,i)-2,x:Object(r.default)(u,c)})):c&&l.push(o(l)+"rotate("+c+i),d=a.skewX,p=s.skewX,f=h,d!==p?m.push({i:f.push(o(f)+"skewX(",null,i)-2,x:Object(r.default)(d,p)}):p&&f.push(o(f)+"skewX("+p+i),function(e,t,n,i,a,s){if(e!==n||t!==i){var u=a.push(o(a)+"scale(",null,",",null,")");s.push({i:u-4,x:Object(r.default)(e,n)},{i:u-2,x:Object(r.default)(t,i)})}else 1===n&&1===i||a.push(o(a)+"scale("+n+","+i+")")}(a.scaleX,a.scaleY,s.scaleX,s.scaleY,h,m),a=s=null,function(e){for(var t,n=-1,r=m.length;++n<r;)h[(t=m[n]).i]=t.x(e);return h.join("")}}}var a=o(i.parseCss,"px, ","px)","deg)"),s=o(i.parseSvg,", ",")",")")},"./node_modules/d3-interpolate/src/transform/parse.js":function(e,t,n){"use strict";n.r(t),n.d(t,"parseCss",function(){return u}),n.d(t,"parseSvg",function(){return c});var r,i,o,a,s=n("./node_modules/d3-interpolate/src/transform/decompose.js");function u(e){return"none"===e?s.identity:(r||(r=document.createElement("DIV"),i=document.documentElement,o=document.defaultView),r.style.transform=e,e=o.getComputedStyle(i.appendChild(r),null).getPropertyValue("transform"),i.removeChild(r),e=e.slice(7,-1).split(","),Object(s.default)(+e[0],+e[1],+e[2],+e[3],+e[4],+e[5]))}function c(e){return null==e?s.identity:(a||(a=document.createElementNS("http://www.w3.org/2000/svg","g")),a.setAttribute("transform",e),(e=a.transform.baseVal.consolidate())?(e=e.matrix,Object(s.default)(e.a,e.b,e.c,e.d,e.e,e.f)):s.identity)}},"./node_modules/d3-interpolate/src/value.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-color/index.js"),i=n("./node_modules/d3-interpolate/src/rgb.js"),o=n("./node_modules/d3-interpolate/src/array.js"),a=n("./node_modules/d3-interpolate/src/date.js"),s=n("./node_modules/d3-interpolate/src/number.js"),u=n("./node_modules/d3-interpolate/src/object.js"),c=n("./node_modules/d3-interpolate/src/string.js"),l=n("./node_modules/d3-interpolate/src/constant.js");t.default=function(e,t){var n,d=typeof t;return null==t||"boolean"===d?Object(l.default)(t):("number"===d?s.default:"string"===d?(n=Object(r.color)(t))?(t=n,i.default):c.default:t instanceof r.color?i.default:t instanceof Date?a.default:Array.isArray(t)?o.default:"function"!=typeof t.valueOf&&"function"!=typeof t.toString||isNaN(t)?u.default:s.default)(e,t)}},"./node_modules/d3-interpolate/src/zoom.js":function(e,t,n){"use strict";n.r(t);var r=Math.SQRT2;function i(e){return((e=Math.exp(e))+1/e)/2}t.default=function(e,t){var n,o,a=e[0],s=e[1],u=e[2],c=t[0],l=t[1],d=t[2],p=c-a,f=l-s,h=p*p+f*f;if(h<1e-12)o=Math.log(d/u)/r,n=function(e){return[a+e*p,s+e*f,u*Math.exp(r*e*o)]};else{var m=Math.sqrt(h),g=(d*d-u*u+4*h)/(2*u*2*m),v=(d*d-u*u-4*h)/(2*d*2*m),_=Math.log(Math.sqrt(g*g+1)-g),y=Math.log(Math.sqrt(v*v+1)-v);o=(y-_)/r,n=function(e){var t,n,c=e*o,l=i(_),d=u/(2*m)*(l*(n=r*c+_,((n=Math.exp(2*n))-1)/(n+1))-(t=_,((t=Math.exp(t))-1/t)/2));return[a+d*p,s+d*f,u*l/i(r*c+_)]}}return n.duration=1e3*o,n}},"./node_modules/d3-path/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-path/src/path.js");n.d(t,"path",function(){return r.default})},"./node_modules/d3-path/src/path.js":function(e,t,n){"use strict";n.r(t);var r=Math.PI,i=2*r,o=i-1e-6;function a(){this._x0=this._y0=this._x1=this._y1=null,this._=""}function s(){return new a}a.prototype=s.prototype={constructor:a,moveTo:function(e,t){this._+="M"+(this._x0=this._x1=+e)+","+(this._y0=this._y1=+t)},closePath:function(){null!==this._x1&&(this._x1=this._x0,this._y1=this._y0,this._+="Z")},lineTo:function(e,t){this._+="L"+(this._x1=+e)+","+(this._y1=+t)},quadraticCurveTo:function(e,t,n,r){this._+="Q"+ +e+","+ +t+","+(this._x1=+n)+","+(this._y1=+r)},bezierCurveTo:function(e,t,n,r,i,o){this._+="C"+ +e+","+ +t+","+ +n+","+ +r+","+(this._x1=+i)+","+(this._y1=+o)},arcTo:function(e,t,n,i,o){e=+e,t=+t,n=+n,i=+i,o=+o;var a=this._x1,s=this._y1,u=n-e,c=i-t,l=a-e,d=s-t,p=l*l+d*d;if(o<0)throw new Error("negative radius: "+o);if(null===this._x1)this._+="M"+(this._x1=e)+","+(this._y1=t);else if(p>1e-6)if(Math.abs(d*u-c*l)>1e-6&&o){var f=n-a,h=i-s,m=u*u+c*c,g=f*f+h*h,v=Math.sqrt(m),_=Math.sqrt(p),y=o*Math.tan((r-Math.acos((m+p-g)/(2*v*_)))/2),b=y/_,j=y/v;Math.abs(b-1)>1e-6&&(this._+="L"+(e+b*l)+","+(t+b*d)),this._+="A"+o+","+o+",0,0,"+ +(d*f>l*h)+","+(this._x1=e+j*u)+","+(this._y1=t+j*c)}else this._+="L"+(this._x1=e)+","+(this._y1=t);else;},arc:function(e,t,n,a,s,u){e=+e,t=+t;var c=(n=+n)*Math.cos(a),l=n*Math.sin(a),d=e+c,p=t+l,f=1^u,h=u?a-s:s-a;if(n<0)throw new Error("negative radius: "+n);null===this._x1?this._+="M"+d+","+p:(Math.abs(this._x1-d)>1e-6||Math.abs(this._y1-p)>1e-6)&&(this._+="L"+d+","+p),n&&(h<0&&(h=h%i+i),h>o?this._+="A"+n+","+n+",0,1,"+f+","+(e-c)+","+(t-l)+"A"+n+","+n+",0,1,"+f+","+(this._x1=d)+","+(this._y1=p):h>1e-6&&(this._+="A"+n+","+n+",0,"+ +(h>=r)+","+f+","+(this._x1=e+n*Math.cos(s))+","+(this._y1=t+n*Math.sin(s))))},rect:function(e,t,n,r){this._+="M"+(this._x0=this._x1=+e)+","+(this._y0=this._y1=+t)+"h"+ +n+"v"+ +r+"h"+-n+"Z"},toString:function(){return this._}},t.default=s},"./node_modules/d3-quadtree/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-quadtree/src/quadtree.js");n.d(t,"quadtree",function(){return r.default})},"./node_modules/d3-quadtree/src/add.js":function(e,t,n){"use strict";function r(e,t,n,r){if(isNaN(t)||isNaN(n))return e;var i,o,a,s,u,c,l,d,p,f=e._root,h={data:r},m=e._x0,g=e._y0,v=e._x1,_=e._y1;if(!f)return e._root=h,e;for(;f.length;)if((c=t>=(o=(m+v)/2))?m=o:v=o,(l=n>=(a=(g+_)/2))?g=a:_=a,i=f,!(f=f[d=l<<1|c]))return i[d]=h,e;if(s=+e._x.call(null,f.data),u=+e._y.call(null,f.data),t===s&&n===u)return h.next=f,i?i[d]=h:e._root=h,e;do{i=i?i[d]=new Array(4):e._root=new Array(4),(c=t>=(o=(m+v)/2))?m=o:v=o,(l=n>=(a=(g+_)/2))?g=a:_=a}while((d=l<<1|c)==(p=(u>=a)<<1|s>=o));return i[p]=f,i[d]=h,e}function i(e){var t,n,i,o,a=e.length,s=new Array(a),u=new Array(a),c=1/0,l=1/0,d=-1/0,p=-1/0;for(n=0;n<a;++n)isNaN(i=+this._x.call(null,t=e[n]))||isNaN(o=+this._y.call(null,t))||(s[n]=i,u[n]=o,i<c&&(c=i),i>d&&(d=i),o<l&&(l=o),o>p&&(p=o));for(d<c&&(c=this._x0,d=this._x1),p<l&&(l=this._y0,p=this._y1),this.cover(c,l).cover(d,p),n=0;n<a;++n)r(this,s[n],u[n],e[n]);return this}n.r(t),n.d(t,"addAll",function(){return i}),t.default=function(e){var t=+this._x.call(null,e),n=+this._y.call(null,e);return r(this.cover(t,n),t,n,e)}},"./node_modules/d3-quadtree/src/cover.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){if(isNaN(e=+e)||isNaN(t=+t))return this;var n=this._x0,r=this._y0,i=this._x1,o=this._y1;if(isNaN(n))i=(n=Math.floor(e))+1,o=(r=Math.floor(t))+1;else{if(!(n>e||e>i||r>t||t>o))return this;var a,s,u=i-n,c=this._root;switch(s=(t<(r+o)/2)<<1|e<(n+i)/2){case 0:do{a=new Array(4),a[s]=c,c=a}while(o=r+(u*=2),e>(i=n+u)||t>o);break;case 1:do{a=new Array(4),a[s]=c,c=a}while(o=r+(u*=2),(n=i-u)>e||t>o);break;case 2:do{a=new Array(4),a[s]=c,c=a}while(r=o-(u*=2),e>(i=n+u)||r>t);break;case 3:do{a=new Array(4),a[s]=c,c=a}while(r=o-(u*=2),(n=i-u)>e||r>t)}this._root&&this._root.length&&(this._root=c)}return this._x0=n,this._y0=r,this._x1=i,this._y1=o,this}},"./node_modules/d3-quadtree/src/data.js":function(e,t,n){"use strict";n.r(t),t.default=function(){var e=[];return this.visit(function(t){if(!t.length)do{e.push(t.data)}while(t=t.next)}),e}},"./node_modules/d3-quadtree/src/extent.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return arguments.length?this.cover(+e[0][0],+e[0][1]).cover(+e[1][0],+e[1][1]):isNaN(this._x0)?void 0:[[this._x0,this._y0],[this._x1,this._y1]]}},"./node_modules/d3-quadtree/src/find.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-quadtree/src/quad.js");t.default=function(e,t,n){var i,o,a,s,u,c,l,d=this._x0,p=this._y0,f=this._x1,h=this._y1,m=[],g=this._root;for(g&&m.push(new r.default(g,d,p,f,h)),null==n?n=1/0:(d=e-n,p=t-n,f=e+n,h=t+n,n*=n);c=m.pop();)if(!(!(g=c.node)||(o=c.x0)>f||(a=c.y0)>h||(s=c.x1)<d||(u=c.y1)<p))if(g.length){var v=(o+s)/2,_=(a+u)/2;m.push(new r.default(g[3],v,_,s,u),new r.default(g[2],o,_,v,u),new r.default(g[1],v,a,s,_),new r.default(g[0],o,a,v,_)),(l=(t>=_)<<1|e>=v)&&(c=m[m.length-1],m[m.length-1]=m[m.length-1-l],m[m.length-1-l]=c)}else{var y=e-+this._x.call(null,g.data),b=t-+this._y.call(null,g.data),j=y*y+b*b;if(j<n){var x=Math.sqrt(n=j);d=e-x,p=t-x,f=e+x,h=t+x,i=g.data}}return i}},"./node_modules/d3-quadtree/src/quad.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n,r,i){this.node=e,this.x0=t,this.y0=n,this.x1=r,this.y1=i}},"./node_modules/d3-quadtree/src/quadtree.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return m});var r=n("./node_modules/d3-quadtree/src/add.js"),i=n("./node_modules/d3-quadtree/src/cover.js"),o=n("./node_modules/d3-quadtree/src/data.js"),a=n("./node_modules/d3-quadtree/src/extent.js"),s=n("./node_modules/d3-quadtree/src/find.js"),u=n("./node_modules/d3-quadtree/src/remove.js"),c=n("./node_modules/d3-quadtree/src/root.js"),l=n("./node_modules/d3-quadtree/src/size.js"),d=n("./node_modules/d3-quadtree/src/visit.js"),p=n("./node_modules/d3-quadtree/src/visitAfter.js"),f=n("./node_modules/d3-quadtree/src/x.js"),h=n("./node_modules/d3-quadtree/src/y.js");function m(e,t,n){var r=new g(null==t?f.defaultX:t,null==n?h.defaultY:n,NaN,NaN,NaN,NaN);return null==e?r:r.addAll(e)}function g(e,t,n,r,i,o){this._x=e,this._y=t,this._x0=n,this._y0=r,this._x1=i,this._y1=o,this._root=void 0}function v(e){for(var t={data:e.data},n=t;e=e.next;)n=n.next={data:e.data};return t}var _=m.prototype=g.prototype;_.copy=function(){var e,t,n=new g(this._x,this._y,this._x0,this._y0,this._x1,this._y1),r=this._root;if(!r)return n;if(!r.length)return n._root=v(r),n;for(e=[{source:r,target:n._root=new Array(4)}];r=e.pop();)for(var i=0;i<4;++i)(t=r.source[i])&&(t.length?e.push({source:t,target:r.target[i]=new Array(4)}):r.target[i]=v(t));return n},_.add=r.default,_.addAll=r.addAll,_.cover=i.default,_.data=o.default,_.extent=a.default,_.find=s.default,_.remove=u.default,_.removeAll=u.removeAll,_.root=c.default,_.size=l.default,_.visit=d.default,_.visitAfter=p.default,_.x=f.default,_.y=h.default},"./node_modules/d3-quadtree/src/remove.js":function(e,t,n){"use strict";function r(e){for(var t=0,n=e.length;t<n;++t)this.remove(e[t]);return this}n.r(t),n.d(t,"removeAll",function(){return r}),t.default=function(e){if(isNaN(o=+this._x.call(null,e))||isNaN(a=+this._y.call(null,e)))return this;var t,n,r,i,o,a,s,u,c,l,d,p,f=this._root,h=this._x0,m=this._y0,g=this._x1,v=this._y1;if(!f)return this;if(f.length)for(;;){if((c=o>=(s=(h+g)/2))?h=s:g=s,(l=a>=(u=(m+v)/2))?m=u:v=u,t=f,!(f=f[d=l<<1|c]))return this;if(!f.length)break;(t[d+1&3]||t[d+2&3]||t[d+3&3])&&(n=t,p=d)}for(;f.data!==e;)if(r=f,!(f=f.next))return this;return(i=f.next)&&delete f.next,r?(i?r.next=i:delete r.next,this):t?(i?t[d]=i:delete t[d],(f=t[0]||t[1]||t[2]||t[3])&&f===(t[3]||t[2]||t[1]||t[0])&&!f.length&&(n?n[p]=f:this._root=f),this):(this._root=i,this)}},"./node_modules/d3-quadtree/src/root.js":function(e,t,n){"use strict";n.r(t),t.default=function(){return this._root}},"./node_modules/d3-quadtree/src/size.js":function(e,t,n){"use strict";n.r(t),t.default=function(){var e=0;return this.visit(function(t){if(!t.length)do{++e}while(t=t.next)}),e}},"./node_modules/d3-quadtree/src/visit.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-quadtree/src/quad.js");t.default=function(e){var t,n,i,o,a,s,u=[],c=this._root;for(c&&u.push(new r.default(c,this._x0,this._y0,this._x1,this._y1));t=u.pop();)if(!e(c=t.node,i=t.x0,o=t.y0,a=t.x1,s=t.y1)&&c.length){var l=(i+a)/2,d=(o+s)/2;(n=c[3])&&u.push(new r.default(n,l,d,a,s)),(n=c[2])&&u.push(new r.default(n,i,d,l,s)),(n=c[1])&&u.push(new r.default(n,l,o,a,d)),(n=c[0])&&u.push(new r.default(n,i,o,l,d))}return this}},"./node_modules/d3-quadtree/src/visitAfter.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-quadtree/src/quad.js");t.default=function(e){var t,n=[],i=[];for(this._root&&n.push(new r.default(this._root,this._x0,this._y0,this._x1,this._y1));t=n.pop();){var o=t.node;if(o.length){var a,s=t.x0,u=t.y0,c=t.x1,l=t.y1,d=(s+c)/2,p=(u+l)/2;(a=o[0])&&n.push(new r.default(a,s,u,d,p)),(a=o[1])&&n.push(new r.default(a,d,u,c,p)),(a=o[2])&&n.push(new r.default(a,s,p,d,l)),(a=o[3])&&n.push(new r.default(a,d,p,c,l))}i.push(t)}for(;t=i.pop();)e(t.node,t.x0,t.y0,t.x1,t.y1);return this}},"./node_modules/d3-quadtree/src/x.js":function(e,t,n){"use strict";function r(e){return e[0]}n.r(t),n.d(t,"defaultX",function(){return r}),t.default=function(e){return arguments.length?(this._x=e,this):this._x}},"./node_modules/d3-quadtree/src/y.js":function(e,t,n){"use strict";function r(e){return e[1]}n.r(t),n.d(t,"defaultY",function(){return r}),t.default=function(e){return arguments.length?(this._y=e,this):this._y}},"./node_modules/d3-request/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-request/src/request.js");n.d(t,"request",function(){return r.default});var i=n("./node_modules/d3-request/src/html.js");n.d(t,"html",function(){return i.default});var o=n("./node_modules/d3-request/src/json.js");n.d(t,"json",function(){return o.default});var a=n("./node_modules/d3-request/src/text.js");n.d(t,"text",function(){return a.default});var s=n("./node_modules/d3-request/src/xml.js");n.d(t,"xml",function(){return s.default});var u=n("./node_modules/d3-request/src/csv.js");n.d(t,"csv",function(){return u.default});var c=n("./node_modules/d3-request/src/tsv.js");n.d(t,"tsv",function(){return c.default})},"./node_modules/d3-request/src/csv.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-dsv/index.js"),i=n("./node_modules/d3-request/src/dsv.js");t.default=Object(i.default)("text/csv",r.csvParse)},"./node_modules/d3-request/src/dsv.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-request/src/request.js");t.default=function(e,t){return function(n,i,o){arguments.length<3&&(o=i,i=null);var a=Object(r.default)(n).mimeType(e);return a.row=function(e){return arguments.length?a.response((n=t,r=i=e,function(e){return n(e.responseText,r)})):i;var n,r},a.row(i),o?a.get(o):a}}},"./node_modules/d3-request/src/html.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-request/src/type.js");t.default=Object(r.default)("text/html",function(e){return document.createRange().createContextualFragment(e.responseText)})},"./node_modules/d3-request/src/json.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-request/src/type.js");t.default=Object(r.default)("application/json",function(e){return JSON.parse(e.responseText)})},"./node_modules/d3-request/src/request.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-collection/index.js"),i=n("./node_modules/d3-dispatch/index.js");t.default=function(e,t){var n,o,a,s,u=Object(i.dispatch)("beforesend","progress","load","error"),c=Object(r.map)(),l=new XMLHttpRequest,d=null,p=null,f=0;function h(e){var t,r,i,o=l.status;if(!o&&((i=(r=l).responseType)&&"text"!==i?r.response:r.responseText)||o>=200&&o<300||304===o){if(a)try{t=a.call(n,l)}catch(e){return void u.call("error",n,e)}else t=l;u.call("load",n,t)}else u.call("error",n,e)}if("undefined"==typeof XDomainRequest||"withCredentials"in l||!/^(http(s)?:)?\/\//.test(e)||(l=new XDomainRequest),"onload"in l?l.onload=l.onerror=l.ontimeout=h:l.onreadystatechange=function(e){l.readyState>3&&h(e)},l.onprogress=function(e){u.call("progress",n,e)},n={header:function(e,t){return e=(e+"").toLowerCase(),arguments.length<2?c.get(e):(null==t?c.remove(e):c.set(e,t+""),n)},mimeType:function(e){return arguments.length?(o=null==e?null:e+"",n):o},responseType:function(e){return arguments.length?(s=e,n):s},timeout:function(e){return arguments.length?(f=+e,n):f},user:function(e){return arguments.length<1?d:(d=null==e?null:e+"",n)},password:function(e){return arguments.length<1?p:(p=null==e?null:e+"",n)},response:function(e){return a=e,n},get:function(e,t){return n.send("GET",e,t)},post:function(e,t){return n.send("POST",e,t)},send:function(t,r,i){var a;return l.open(t,e,!0,d,p),null==o||c.has("accept")||c.set("accept",o+",*/*"),l.setRequestHeader&&c.each(function(e,t){l.setRequestHeader(t,e)}),null!=o&&l.overrideMimeType&&l.overrideMimeType(o),null!=s&&(l.responseType=s),f>0&&(l.timeout=f),null==i&&"function"==typeof r&&(i=r,r=null),null!=i&&1===i.length&&(a=i,i=function(e,t){a(null==e?t:null)}),null!=i&&n.on("error",i).on("load",function(e){i(null,e)}),u.call("beforesend",n,l),l.send(null==r?null:r),n},abort:function(){return l.abort(),n},on:function(){var e=u.on.apply(u,arguments);return e===u?n:e}},null!=t){if("function"!=typeof t)throw new Error("invalid callback: "+t);return n.get(t)}return n}},"./node_modules/d3-request/src/text.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-request/src/type.js");t.default=Object(r.default)("text/plain",function(e){return e.responseText})},"./node_modules/d3-request/src/tsv.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-dsv/index.js"),i=n("./node_modules/d3-request/src/dsv.js");t.default=Object(i.default)("text/tab-separated-values",r.tsvParse)},"./node_modules/d3-request/src/type.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-request/src/request.js");t.default=function(e,t){return function(n,i){var o=Object(r.default)(n).mimeType(e).response(t);if(null!=i){if("function"!=typeof i)throw new Error("invalid callback: "+i);return o.get(i)}return o}}},"./node_modules/d3-request/src/xml.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-request/src/type.js");t.default=Object(r.default)("application/xml",function(e){var t=e.responseXML;if(!t)throw new Error("parse error");return t})},"./node_modules/d3-scale-chromatic/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-scale-chromatic/src/categorical/category10.js");n.d(t,"schemeCategory10",function(){return r.default});var i=n("./node_modules/d3-scale-chromatic/src/categorical/Accent.js");n.d(t,"schemeAccent",function(){return i.default});var o=n("./node_modules/d3-scale-chromatic/src/categorical/Dark2.js");n.d(t,"schemeDark2",function(){return o.default});var a=n("./node_modules/d3-scale-chromatic/src/categorical/Paired.js");n.d(t,"schemePaired",function(){return a.default});var s=n("./node_modules/d3-scale-chromatic/src/categorical/Pastel1.js");n.d(t,"schemePastel1",function(){return s.default});var u=n("./node_modules/d3-scale-chromatic/src/categorical/Pastel2.js");n.d(t,"schemePastel2",function(){return u.default});var c=n("./node_modules/d3-scale-chromatic/src/categorical/Set1.js");n.d(t,"schemeSet1",function(){return c.default});var l=n("./node_modules/d3-scale-chromatic/src/categorical/Set2.js");n.d(t,"schemeSet2",function(){return l.default});var d=n("./node_modules/d3-scale-chromatic/src/categorical/Set3.js");n.d(t,"schemeSet3",function(){return d.default});var p=n("./node_modules/d3-scale-chromatic/src/diverging/BrBG.js");n.d(t,"interpolateBrBG",function(){return p.default}),n.d(t,"schemeBrBG",function(){return p.scheme});var f=n("./node_modules/d3-scale-chromatic/src/diverging/PRGn.js");n.d(t,"interpolatePRGn",function(){return f.default}),n.d(t,"schemePRGn",function(){return f.scheme});var h=n("./node_modules/d3-scale-chromatic/src/diverging/PiYG.js");n.d(t,"interpolatePiYG",function(){return h.default}),n.d(t,"schemePiYG",function(){return h.scheme});var m=n("./node_modules/d3-scale-chromatic/src/diverging/PuOr.js");n.d(t,"interpolatePuOr",function(){return m.default}),n.d(t,"schemePuOr",function(){return m.scheme});var g=n("./node_modules/d3-scale-chromatic/src/diverging/RdBu.js");n.d(t,"interpolateRdBu",function(){return g.default}),n.d(t,"schemeRdBu",function(){return g.scheme});var v=n("./node_modules/d3-scale-chromatic/src/diverging/RdGy.js");n.d(t,"interpolateRdGy",function(){return v.default}),n.d(t,"schemeRdGy",function(){return v.scheme});var _=n("./node_modules/d3-scale-chromatic/src/diverging/RdYlBu.js");n.d(t,"interpolateRdYlBu",function(){return _.default}),n.d(t,"schemeRdYlBu",function(){return _.scheme});var y=n("./node_modules/d3-scale-chromatic/src/diverging/RdYlGn.js");n.d(t,"interpolateRdYlGn",function(){return y.default}),n.d(t,"schemeRdYlGn",function(){return y.scheme});var b=n("./node_modules/d3-scale-chromatic/src/diverging/Spectral.js");n.d(t,"interpolateSpectral",function(){return b.default}),n.d(t,"schemeSpectral",function(){return b.scheme});var j=n("./node_modules/d3-scale-chromatic/src/sequential-multi/BuGn.js");n.d(t,"interpolateBuGn",function(){return j.default}),n.d(t,"schemeBuGn",function(){return j.scheme});var x=n("./node_modules/d3-scale-chromatic/src/sequential-multi/BuPu.js");n.d(t,"interpolateBuPu",function(){return x.default}),n.d(t,"schemeBuPu",function(){return x.scheme});var O=n("./node_modules/d3-scale-chromatic/src/sequential-multi/GnBu.js");n.d(t,"interpolateGnBu",function(){return O.default}),n.d(t,"schemeGnBu",function(){return O.scheme});var w=n("./node_modules/d3-scale-chromatic/src/sequential-multi/OrRd.js");n.d(t,"interpolateOrRd",function(){return w.default}),n.d(t,"schemeOrRd",function(){return w.scheme});var C=n("./node_modules/d3-scale-chromatic/src/sequential-multi/PuBuGn.js");n.d(t,"interpolatePuBuGn",function(){return C.default}),n.d(t,"schemePuBuGn",function(){return C.scheme});var E=n("./node_modules/d3-scale-chromatic/src/sequential-multi/PuBu.js");n.d(t,"interpolatePuBu",function(){return E.default}),n.d(t,"schemePuBu",function(){return E.scheme});var S=n("./node_modules/d3-scale-chromatic/src/sequential-multi/PuRd.js");n.d(t,"interpolatePuRd",function(){return S.default}),n.d(t,"schemePuRd",function(){return S.scheme});var k=n("./node_modules/d3-scale-chromatic/src/sequential-multi/RdPu.js");n.d(t,"interpolateRdPu",function(){return k.default}),n.d(t,"schemeRdPu",function(){return k.scheme});var M=n("./node_modules/d3-scale-chromatic/src/sequential-multi/YlGnBu.js");n.d(t,"interpolateYlGnBu",function(){return M.default}),n.d(t,"schemeYlGnBu",function(){return M.scheme});var A=n("./node_modules/d3-scale-chromatic/src/sequential-multi/YlGn.js");n.d(t,"interpolateYlGn",function(){return A.default}),n.d(t,"schemeYlGn",function(){return A.scheme});var T=n("./node_modules/d3-scale-chromatic/src/sequential-multi/YlOrBr.js");n.d(t,"interpolateYlOrBr",function(){return T.default}),n.d(t,"schemeYlOrBr",function(){return T.scheme});var D=n("./node_modules/d3-scale-chromatic/src/sequential-multi/YlOrRd.js");n.d(t,"interpolateYlOrRd",function(){return D.default}),n.d(t,"schemeYlOrRd",function(){return D.scheme});var I=n("./node_modules/d3-scale-chromatic/src/sequential-single/Blues.js");n.d(t,"interpolateBlues",function(){return I.default}),n.d(t,"schemeBlues",function(){return I.scheme});var P=n("./node_modules/d3-scale-chromatic/src/sequential-single/Greens.js");n.d(t,"interpolateGreens",function(){return P.default}),n.d(t,"schemeGreens",function(){return P.scheme});var R=n("./node_modules/d3-scale-chromatic/src/sequential-single/Greys.js");n.d(t,"interpolateGreys",function(){return R.default}),n.d(t,"schemeGreys",function(){return R.scheme});var N=n("./node_modules/d3-scale-chromatic/src/sequential-single/Purples.js");n.d(t,"interpolatePurples",function(){return N.default}),n.d(t,"schemePurples",function(){return N.scheme});var F=n("./node_modules/d3-scale-chromatic/src/sequential-single/Reds.js");n.d(t,"interpolateReds",function(){return F.default}),n.d(t,"schemeReds",function(){return F.scheme});var L=n("./node_modules/d3-scale-chromatic/src/sequential-single/Oranges.js");n.d(t,"interpolateOranges",function(){return L.default}),n.d(t,"schemeOranges",function(){return L.scheme});var V=n("./node_modules/d3-scale-chromatic/src/sequential-multi/cubehelix.js");n.d(t,"interpolateCubehelixDefault",function(){return V.default});var B=n("./node_modules/d3-scale-chromatic/src/sequential-multi/rainbow.js");n.d(t,"interpolateRainbow",function(){return B.default}),n.d(t,"interpolateWarm",function(){return B.warm}),n.d(t,"interpolateCool",function(){return B.cool});var z=n("./node_modules/d3-scale-chromatic/src/sequential-multi/viridis.js");n.d(t,"interpolateViridis",function(){return z.default}),n.d(t,"interpolateMagma",function(){return z.magma}),n.d(t,"interpolateInferno",function(){return z.inferno}),n.d(t,"interpolatePlasma",function(){return z.plasma})},"./node_modules/d3-scale-chromatic/src/categorical/Accent.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-scale-chromatic/src/colors.js");t.default=Object(r.default)("7fc97fbeaed4fdc086ffff99386cb0f0027fbf5b17666666")},"./node_modules/d3-scale-chromatic/src/categorical/Dark2.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-scale-chromatic/src/colors.js");t.default=Object(r.default)("1b9e77d95f027570b3e7298a66a61ee6ab02a6761d666666")},"./node_modules/d3-scale-chromatic/src/categorical/Paired.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-scale-chromatic/src/colors.js");t.default=Object(r.default)("a6cee31f78b4b2df8a33a02cfb9a99e31a1cfdbf6fff7f00cab2d66a3d9affff99b15928")},"./node_modules/d3-scale-chromatic/src/categorical/Pastel1.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-scale-chromatic/src/colors.js");t.default=Object(r.default)("fbb4aeb3cde3ccebc5decbe4fed9a6ffffcce5d8bdfddaecf2f2f2")},"./node_modules/d3-scale-chromatic/src/categorical/Pastel2.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-scale-chromatic/src/colors.js");t.default=Object(r.default)("b3e2cdfdcdaccbd5e8f4cae4e6f5c9fff2aef1e2cccccccc")},"./node_modules/d3-scale-chromatic/src/categorical/Set1.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-scale-chromatic/src/colors.js");t.default=Object(r.default)("e41a1c377eb84daf4a984ea3ff7f00ffff33a65628f781bf999999")},"./node_modules/d3-scale-chromatic/src/categorical/Set2.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-scale-chromatic/src/colors.js");t.default=Object(r.default)("66c2a5fc8d628da0cbe78ac3a6d854ffd92fe5c494b3b3b3")},"./node_modules/d3-scale-chromatic/src/categorical/Set3.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-scale-chromatic/src/colors.js");t.default=Object(r.default)("8dd3c7ffffb3bebadafb807280b1d3fdb462b3de69fccde5d9d9d9bc80bdccebc5ffed6f")},"./node_modules/d3-scale-chromatic/src/categorical/category10.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-scale-chromatic/src/colors.js");t.default=Object(r.default)("1f77b4ff7f0e2ca02cd627289467bd8c564be377c27f7f7fbcbd2217becf")},"./node_modules/d3-scale-chromatic/src/colors.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){for(var t=e.length/6|0,n=new Array(t),r=0;r<t;)n[r]="#"+e.slice(6*r,6*++r);return n}},"./node_modules/d3-scale-chromatic/src/diverging/BrBG.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("d8b365f5f5f55ab4ac","a6611adfc27d80cdc1018571","a6611adfc27df5f5f580cdc1018571","8c510ad8b365f6e8c3c7eae55ab4ac01665e","8c510ad8b365f6e8c3f5f5f5c7eae55ab4ac01665e","8c510abf812ddfc27df6e8c3c7eae580cdc135978f01665e","8c510abf812ddfc27df6e8c3f5f5f5c7eae580cdc135978f01665e","5430058c510abf812ddfc27df6e8c3c7eae580cdc135978f01665e003c30","5430058c510abf812ddfc27df6e8c3f5f5f5c7eae580cdc135978f01665e003c30").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/diverging/PRGn.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("af8dc3f7f7f77fbf7b","7b3294c2a5cfa6dba0008837","7b3294c2a5cff7f7f7a6dba0008837","762a83af8dc3e7d4e8d9f0d37fbf7b1b7837","762a83af8dc3e7d4e8f7f7f7d9f0d37fbf7b1b7837","762a839970abc2a5cfe7d4e8d9f0d3a6dba05aae611b7837","762a839970abc2a5cfe7d4e8f7f7f7d9f0d3a6dba05aae611b7837","40004b762a839970abc2a5cfe7d4e8d9f0d3a6dba05aae611b783700441b","40004b762a839970abc2a5cfe7d4e8f7f7f7d9f0d3a6dba05aae611b783700441b").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/diverging/PiYG.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("e9a3c9f7f7f7a1d76a","d01c8bf1b6dab8e1864dac26","d01c8bf1b6daf7f7f7b8e1864dac26","c51b7de9a3c9fde0efe6f5d0a1d76a4d9221","c51b7de9a3c9fde0eff7f7f7e6f5d0a1d76a4d9221","c51b7dde77aef1b6dafde0efe6f5d0b8e1867fbc414d9221","c51b7dde77aef1b6dafde0eff7f7f7e6f5d0b8e1867fbc414d9221","8e0152c51b7dde77aef1b6dafde0efe6f5d0b8e1867fbc414d9221276419","8e0152c51b7dde77aef1b6dafde0eff7f7f7e6f5d0b8e1867fbc414d9221276419").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/diverging/PuOr.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("998ec3f7f7f7f1a340","5e3c99b2abd2fdb863e66101","5e3c99b2abd2f7f7f7fdb863e66101","542788998ec3d8daebfee0b6f1a340b35806","542788998ec3d8daebf7f7f7fee0b6f1a340b35806","5427888073acb2abd2d8daebfee0b6fdb863e08214b35806","5427888073acb2abd2d8daebf7f7f7fee0b6fdb863e08214b35806","2d004b5427888073acb2abd2d8daebfee0b6fdb863e08214b358067f3b08","2d004b5427888073acb2abd2d8daebf7f7f7fee0b6fdb863e08214b358067f3b08").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/diverging/RdBu.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("ef8a62f7f7f767a9cf","ca0020f4a58292c5de0571b0","ca0020f4a582f7f7f792c5de0571b0","b2182bef8a62fddbc7d1e5f067a9cf2166ac","b2182bef8a62fddbc7f7f7f7d1e5f067a9cf2166ac","b2182bd6604df4a582fddbc7d1e5f092c5de4393c32166ac","b2182bd6604df4a582fddbc7f7f7f7d1e5f092c5de4393c32166ac","67001fb2182bd6604df4a582fddbc7d1e5f092c5de4393c32166ac053061","67001fb2182bd6604df4a582fddbc7f7f7f7d1e5f092c5de4393c32166ac053061").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/diverging/RdGy.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("ef8a62ffffff999999","ca0020f4a582bababa404040","ca0020f4a582ffffffbababa404040","b2182bef8a62fddbc7e0e0e09999994d4d4d","b2182bef8a62fddbc7ffffffe0e0e09999994d4d4d","b2182bd6604df4a582fddbc7e0e0e0bababa8787874d4d4d","b2182bd6604df4a582fddbc7ffffffe0e0e0bababa8787874d4d4d","67001fb2182bd6604df4a582fddbc7e0e0e0bababa8787874d4d4d1a1a1a","67001fb2182bd6604df4a582fddbc7ffffffe0e0e0bababa8787874d4d4d1a1a1a").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/diverging/RdYlBu.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("fc8d59ffffbf91bfdb","d7191cfdae61abd9e92c7bb6","d7191cfdae61ffffbfabd9e92c7bb6","d73027fc8d59fee090e0f3f891bfdb4575b4","d73027fc8d59fee090ffffbfe0f3f891bfdb4575b4","d73027f46d43fdae61fee090e0f3f8abd9e974add14575b4","d73027f46d43fdae61fee090ffffbfe0f3f8abd9e974add14575b4","a50026d73027f46d43fdae61fee090e0f3f8abd9e974add14575b4313695","a50026d73027f46d43fdae61fee090ffffbfe0f3f8abd9e974add14575b4313695").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/diverging/RdYlGn.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("fc8d59ffffbf91cf60","d7191cfdae61a6d96a1a9641","d7191cfdae61ffffbfa6d96a1a9641","d73027fc8d59fee08bd9ef8b91cf601a9850","d73027fc8d59fee08bffffbfd9ef8b91cf601a9850","d73027f46d43fdae61fee08bd9ef8ba6d96a66bd631a9850","d73027f46d43fdae61fee08bffffbfd9ef8ba6d96a66bd631a9850","a50026d73027f46d43fdae61fee08bd9ef8ba6d96a66bd631a9850006837","a50026d73027f46d43fdae61fee08bffffbfd9ef8ba6d96a66bd631a9850006837").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/diverging/Spectral.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("fc8d59ffffbf99d594","d7191cfdae61abdda42b83ba","d7191cfdae61ffffbfabdda42b83ba","d53e4ffc8d59fee08be6f59899d5943288bd","d53e4ffc8d59fee08bffffbfe6f59899d5943288bd","d53e4ff46d43fdae61fee08be6f598abdda466c2a53288bd","d53e4ff46d43fdae61fee08bffffbfe6f598abdda466c2a53288bd","9e0142d53e4ff46d43fdae61fee08be6f598abdda466c2a53288bd5e4fa2","9e0142d53e4ff46d43fdae61fee08bffffbfe6f598abdda466c2a53288bd5e4fa2").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/ramp.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-interpolate/index.js");t.default=function(e){return Object(r.interpolateRgbBasis)(e[e.length-1])}},"./node_modules/d3-scale-chromatic/src/sequential-multi/BuGn.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("e5f5f999d8c92ca25f","edf8fbb2e2e266c2a4238b45","edf8fbb2e2e266c2a42ca25f006d2c","edf8fbccece699d8c966c2a42ca25f006d2c","edf8fbccece699d8c966c2a441ae76238b45005824","f7fcfde5f5f9ccece699d8c966c2a441ae76238b45005824","f7fcfde5f5f9ccece699d8c966c2a441ae76238b45006d2c00441b").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/sequential-multi/BuPu.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("e0ecf49ebcda8856a7","edf8fbb3cde38c96c688419d","edf8fbb3cde38c96c68856a7810f7c","edf8fbbfd3e69ebcda8c96c68856a7810f7c","edf8fbbfd3e69ebcda8c96c68c6bb188419d6e016b","f7fcfde0ecf4bfd3e69ebcda8c96c68c6bb188419d6e016b","f7fcfde0ecf4bfd3e69ebcda8c96c68c6bb188419d810f7c4d004b").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/sequential-multi/GnBu.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("e0f3dba8ddb543a2ca","f0f9e8bae4bc7bccc42b8cbe","f0f9e8bae4bc7bccc443a2ca0868ac","f0f9e8ccebc5a8ddb57bccc443a2ca0868ac","f0f9e8ccebc5a8ddb57bccc44eb3d32b8cbe08589e","f7fcf0e0f3dbccebc5a8ddb57bccc44eb3d32b8cbe08589e","f7fcf0e0f3dbccebc5a8ddb57bccc44eb3d32b8cbe0868ac084081").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/sequential-multi/OrRd.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("fee8c8fdbb84e34a33","fef0d9fdcc8afc8d59d7301f","fef0d9fdcc8afc8d59e34a33b30000","fef0d9fdd49efdbb84fc8d59e34a33b30000","fef0d9fdd49efdbb84fc8d59ef6548d7301f990000","fff7ecfee8c8fdd49efdbb84fc8d59ef6548d7301f990000","fff7ecfee8c8fdd49efdbb84fc8d59ef6548d7301fb300007f0000").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/sequential-multi/PuBu.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("ece7f2a6bddb2b8cbe","f1eef6bdc9e174a9cf0570b0","f1eef6bdc9e174a9cf2b8cbe045a8d","f1eef6d0d1e6a6bddb74a9cf2b8cbe045a8d","f1eef6d0d1e6a6bddb74a9cf3690c00570b0034e7b","fff7fbece7f2d0d1e6a6bddb74a9cf3690c00570b0034e7b","fff7fbece7f2d0d1e6a6bddb74a9cf3690c00570b0045a8d023858").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/sequential-multi/PuBuGn.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("ece2f0a6bddb1c9099","f6eff7bdc9e167a9cf02818a","f6eff7bdc9e167a9cf1c9099016c59","f6eff7d0d1e6a6bddb67a9cf1c9099016c59","f6eff7d0d1e6a6bddb67a9cf3690c002818a016450","fff7fbece2f0d0d1e6a6bddb67a9cf3690c002818a016450","fff7fbece2f0d0d1e6a6bddb67a9cf3690c002818a016c59014636").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/sequential-multi/PuRd.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("e7e1efc994c7dd1c77","f1eef6d7b5d8df65b0ce1256","f1eef6d7b5d8df65b0dd1c77980043","f1eef6d4b9dac994c7df65b0dd1c77980043","f1eef6d4b9dac994c7df65b0e7298ace125691003f","f7f4f9e7e1efd4b9dac994c7df65b0e7298ace125691003f","f7f4f9e7e1efd4b9dac994c7df65b0e7298ace125698004367001f").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/sequential-multi/RdPu.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("fde0ddfa9fb5c51b8a","feebe2fbb4b9f768a1ae017e","feebe2fbb4b9f768a1c51b8a7a0177","feebe2fcc5c0fa9fb5f768a1c51b8a7a0177","feebe2fcc5c0fa9fb5f768a1dd3497ae017e7a0177","fff7f3fde0ddfcc5c0fa9fb5f768a1dd3497ae017e7a0177","fff7f3fde0ddfcc5c0fa9fb5f768a1dd3497ae017e7a017749006a").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/sequential-multi/YlGn.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("f7fcb9addd8e31a354","ffffccc2e69978c679238443","ffffccc2e69978c67931a354006837","ffffccd9f0a3addd8e78c67931a354006837","ffffccd9f0a3addd8e78c67941ab5d238443005a32","ffffe5f7fcb9d9f0a3addd8e78c67941ab5d238443005a32","ffffe5f7fcb9d9f0a3addd8e78c67941ab5d238443006837004529").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/sequential-multi/YlGnBu.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("edf8b17fcdbb2c7fb8","ffffcca1dab441b6c4225ea8","ffffcca1dab441b6c42c7fb8253494","ffffccc7e9b47fcdbb41b6c42c7fb8253494","ffffccc7e9b47fcdbb41b6c41d91c0225ea80c2c84","ffffd9edf8b1c7e9b47fcdbb41b6c41d91c0225ea80c2c84","ffffd9edf8b1c7e9b47fcdbb41b6c41d91c0225ea8253494081d58").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/sequential-multi/YlOrBr.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("fff7bcfec44fd95f0e","ffffd4fed98efe9929cc4c02","ffffd4fed98efe9929d95f0e993404","ffffd4fee391fec44ffe9929d95f0e993404","ffffd4fee391fec44ffe9929ec7014cc4c028c2d04","ffffe5fff7bcfee391fec44ffe9929ec7014cc4c028c2d04","ffffe5fff7bcfee391fec44ffe9929ec7014cc4c02993404662506").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/sequential-multi/YlOrRd.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("ffeda0feb24cf03b20","ffffb2fecc5cfd8d3ce31a1c","ffffb2fecc5cfd8d3cf03b20bd0026","ffffb2fed976feb24cfd8d3cf03b20bd0026","ffffb2fed976feb24cfd8d3cfc4e2ae31a1cb10026","ffffccffeda0fed976feb24cfd8d3cfc4e2ae31a1cb10026","ffffccffeda0fed976feb24cfd8d3cfc4e2ae31a1cbd0026800026").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/sequential-multi/cubehelix.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-color/index.js"),i=n("./node_modules/d3-interpolate/index.js");t.default=Object(i.interpolateCubehelixLong)(Object(r.cubehelix)(300,.5,0),Object(r.cubehelix)(-240,.5,1))},"./node_modules/d3-scale-chromatic/src/sequential-multi/rainbow.js":function(e,t,n){"use strict";n.r(t),n.d(t,"warm",function(){return o}),n.d(t,"cool",function(){return a});var r=n("./node_modules/d3-color/index.js"),i=n("./node_modules/d3-interpolate/index.js"),o=Object(i.interpolateCubehelixLong)(Object(r.cubehelix)(-100,.75,.35),Object(r.cubehelix)(80,1.5,.8)),a=Object(i.interpolateCubehelixLong)(Object(r.cubehelix)(260,.75,.35),Object(r.cubehelix)(80,1.5,.8)),s=Object(r.cubehelix)();t.default=function(e){(e<0||e>1)&&(e-=Math.floor(e));var t=Math.abs(e-.5);return s.h=360*e-100,s.s=1.5-1.5*t,s.l=.8-.9*t,s+""}},"./node_modules/d3-scale-chromatic/src/sequential-multi/viridis.js":function(e,t,n){"use strict";n.r(t),n.d(t,"magma",function(){return o}),n.d(t,"inferno",function(){return a}),n.d(t,"plasma",function(){return s});var r=n("./node_modules/d3-scale-chromatic/src/colors.js");function i(e){var t=e.length;return function(n){return e[Math.max(0,Math.min(t-1,Math.floor(n*t)))]}}t.default=i(Object(r.default)("44015444025645045745055946075a46085c460a5d460b5e470d60470e6147106347116447136548146748166848176948186a481a6c481b6d481c6e481d6f481f70482071482173482374482475482576482677482878482979472a7a472c7a472d7b472e7c472f7d46307e46327e46337f463480453581453781453882443983443a83443b84433d84433e85423f854240864241864142874144874045884046883f47883f48893e49893e4a893e4c8a3d4d8a3d4e8a3c4f8a3c508b3b518b3b528b3a538b3a548c39558c39568c38588c38598c375a8c375b8d365c8d365d8d355e8d355f8d34608d34618d33628d33638d32648e32658e31668e31678e31688e30698e306a8e2f6b8e2f6c8e2e6d8e2e6e8e2e6f8e2d708e2d718e2c718e2c728e2c738e2b748e2b758e2a768e2a778e2a788e29798e297a8e297b8e287c8e287d8e277e8e277f8e27808e26818e26828e26828e25838e25848e25858e24868e24878e23888e23898e238a8d228b8d228c8d228d8d218e8d218f8d21908d21918c20928c20928c20938c1f948c1f958b1f968b1f978b1f988b1f998a1f9a8a1e9b8a1e9c891e9d891f9e891f9f881fa0881fa1881fa1871fa28720a38620a48621a58521a68522a78522a88423a98324aa8325ab8225ac8226ad8127ad8128ae8029af7f2ab07f2cb17e2db27d2eb37c2fb47c31b57b32b67a34b67935b77937b87838b9773aba763bbb753dbc743fbc7340bd7242be7144bf7046c06f48c16e4ac16d4cc26c4ec36b50c46a52c56954c56856c66758c7655ac8645cc8635ec96260ca6063cb5f65cb5e67cc5c69cd5b6ccd5a6ece5870cf5773d05675d05477d1537ad1517cd2507fd34e81d34d84d44b86d54989d5488bd6468ed64590d74393d74195d84098d83e9bd93c9dd93ba0da39a2da37a5db36a8db34aadc32addc30b0dd2fb2dd2db5de2bb8de29bade28bddf26c0df25c2df23c5e021c8e020cae11fcde11dd0e11cd2e21bd5e21ad8e219dae319dde318dfe318e2e418e5e419e7e419eae51aece51befe51cf1e51df4e61ef6e620f8e621fbe723fde725"));var o=i(Object(r.default)("00000401000501010601010802010902020b02020d03030f03031204041405041606051806051a07061c08071e0907200a08220b09240c09260d0a290e0b2b100b2d110c2f120d31130d34140e36150e38160f3b180f3d19103f1a10421c10441d11471e114920114b21114e22115024125325125527125829115a2a115c2c115f2d11612f116331116533106734106936106b38106c390f6e3b0f703d0f713f0f72400f74420f75440f764510774710784910784a10794c117a4e117b4f127b51127c52137c54137d56147d57157e59157e5a167e5c167f5d177f5f187f601880621980641a80651a80671b80681c816a1c816b1d816d1d816e1e81701f81721f817320817521817621817822817922827b23827c23827e24828025828125818326818426818627818827818928818b29818c29818e2a81902a81912b81932b80942c80962c80982d80992d809b2e7f9c2e7f9e2f7fa02f7fa1307ea3307ea5317ea6317da8327daa337dab337cad347cae347bb0357bb2357bb3367ab5367ab73779b83779ba3878bc3978bd3977bf3a77c03a76c23b75c43c75c53c74c73d73c83e73ca3e72cc3f71cd4071cf4070d0416fd2426fd3436ed5446dd6456cd8456cd9466bdb476adc4869de4968df4a68e04c67e24d66e34e65e44f64e55064e75263e85362e95462ea5661eb5760ec5860ed5a5fee5b5eef5d5ef05f5ef1605df2625df2645cf3655cf4675cf4695cf56b5cf66c5cf66e5cf7705cf7725cf8745cf8765cf9785df9795df97b5dfa7d5efa7f5efa815ffb835ffb8560fb8761fc8961fc8a62fc8c63fc8e64fc9065fd9266fd9467fd9668fd9869fd9a6afd9b6bfe9d6cfe9f6dfea16efea36ffea571fea772fea973feaa74feac76feae77feb078feb27afeb47bfeb67cfeb77efeb97ffebb81febd82febf84fec185fec287fec488fec68afec88cfeca8dfecc8ffecd90fecf92fed194fed395fed597fed799fed89afdda9cfddc9efddea0fde0a1fde2a3fde3a5fde5a7fde7a9fde9aafdebacfcecaefceeb0fcf0b2fcf2b4fcf4b6fcf6b8fcf7b9fcf9bbfcfbbdfcfdbf")),a=i(Object(r.default)("00000401000501010601010802010a02020c02020e03021004031204031405041706041907051b08051d09061f0a07220b07240c08260d08290e092b10092d110a30120a32140b34150b37160b39180c3c190c3e1b0c411c0c431e0c451f0c48210c4a230c4c240c4f260c51280b53290b552b0b572d0b592f0a5b310a5c320a5e340a5f3609613809623909633b09643d09653e0966400a67420a68440a68450a69470b6a490b6a4a0c6b4c0c6b4d0d6c4f0d6c510e6c520e6d540f6d550f6d57106e59106e5a116e5c126e5d126e5f136e61136e62146e64156e65156e67166e69166e6a176e6c186e6d186e6f196e71196e721a6e741a6e751b6e771c6d781c6d7a1d6d7c1d6d7d1e6d7f1e6c801f6c82206c84206b85216b87216b88226a8a226a8c23698d23698f24699025689225689326679526679727669827669a28659b29649d29649f2a63a02a63a22b62a32c61a52c60a62d60a82e5fa92e5eab2f5ead305dae305cb0315bb1325ab3325ab43359b63458b73557b93556ba3655bc3754bd3853bf3952c03a51c13a50c33b4fc43c4ec63d4dc73e4cc83f4bca404acb4149cc4248ce4347cf4446d04545d24644d34743d44842d54a41d74b3fd84c3ed94d3dda4e3cdb503bdd513ade5238df5337e05536e15635e25734e35933e45a31e55c30e65d2fe75e2ee8602de9612bea632aeb6429eb6628ec6726ed6925ee6a24ef6c23ef6e21f06f20f1711ff1731df2741cf3761bf37819f47918f57b17f57d15f67e14f68013f78212f78410f8850ff8870ef8890cf98b0bf98c0af98e09fa9008fa9207fa9407fb9606fb9706fb9906fb9b06fb9d07fc9f07fca108fca309fca50afca60cfca80dfcaa0ffcac11fcae12fcb014fcb216fcb418fbb61afbb81dfbba1ffbbc21fbbe23fac026fac228fac42afac62df9c72ff9c932f9cb35f8cd37f8cf3af7d13df7d340f6d543f6d746f5d949f5db4cf4dd4ff4df53f4e156f3e35af3e55df2e661f2e865f2ea69f1ec6df1ed71f1ef75f1f179f2f27df2f482f3f586f3f68af4f88ef5f992f6fa96f8fb9af9fc9dfafda1fcffa4")),s=i(Object(r.default)("0d088710078813078916078a19068c1b068d1d068e20068f2206902406912605912805922a05932c05942e05952f059631059733059735049837049938049a3a049a3c049b3e049c3f049c41049d43039e44039e46039f48039f4903a04b03a14c02a14e02a25002a25102a35302a35502a45601a45801a45901a55b01a55c01a65e01a66001a66100a76300a76400a76600a76700a86900a86a00a86c00a86e00a86f00a87100a87201a87401a87501a87701a87801a87a02a87b02a87d03a87e03a88004a88104a78305a78405a78606a68707a68808a68a09a58b0aa58d0ba58e0ca48f0da4910ea3920fa39410a29511a19613a19814a099159f9a169f9c179e9d189d9e199da01a9ca11b9ba21d9aa31e9aa51f99a62098a72197a82296aa2395ab2494ac2694ad2793ae2892b02991b12a90b22b8fb32c8eb42e8db52f8cb6308bb7318ab83289ba3388bb3488bc3587bd3786be3885bf3984c03a83c13b82c23c81c33d80c43e7fc5407ec6417dc7427cc8437bc9447aca457acb4679cc4778cc4977cd4a76ce4b75cf4c74d04d73d14e72d24f71d35171d45270d5536fd5546ed6556dd7566cd8576bd9586ada5a6ada5b69db5c68dc5d67dd5e66de5f65de6164df6263e06363e16462e26561e26660e3685fe4695ee56a5de56b5de66c5ce76e5be76f5ae87059e97158e97257ea7457eb7556eb7655ec7754ed7953ed7a52ee7b51ef7c51ef7e50f07f4ff0804ef1814df1834cf2844bf3854bf3874af48849f48948f58b47f58c46f68d45f68f44f79044f79143f79342f89441f89540f9973ff9983ef99a3efa9b3dfa9c3cfa9e3bfb9f3afba139fba238fca338fca537fca636fca835fca934fdab33fdac33fdae32fdaf31fdb130fdb22ffdb42ffdb52efeb72dfeb82cfeba2cfebb2bfebd2afebe2afec029fdc229fdc328fdc527fdc627fdc827fdca26fdcb26fccd25fcce25fcd025fcd225fbd324fbd524fbd724fad824fada24f9dc24f9dd25f8df25f8e125f7e225f7e425f6e626f6e826f5e926f5eb27f4ed27f3ee27f3f027f2f227f1f426f1f525f0f724f0f921"))},"./node_modules/d3-scale-chromatic/src/sequential-single/Blues.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("deebf79ecae13182bd","eff3ffbdd7e76baed62171b5","eff3ffbdd7e76baed63182bd08519c","eff3ffc6dbef9ecae16baed63182bd08519c","eff3ffc6dbef9ecae16baed64292c62171b5084594","f7fbffdeebf7c6dbef9ecae16baed64292c62171b5084594","f7fbffdeebf7c6dbef9ecae16baed64292c62171b508519c08306b").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/sequential-single/Greens.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("e5f5e0a1d99b31a354","edf8e9bae4b374c476238b45","edf8e9bae4b374c47631a354006d2c","edf8e9c7e9c0a1d99b74c47631a354006d2c","edf8e9c7e9c0a1d99b74c47641ab5d238b45005a32","f7fcf5e5f5e0c7e9c0a1d99b74c47641ab5d238b45005a32","f7fcf5e5f5e0c7e9c0a1d99b74c47641ab5d238b45006d2c00441b").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/sequential-single/Greys.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("f0f0f0bdbdbd636363","f7f7f7cccccc969696525252","f7f7f7cccccc969696636363252525","f7f7f7d9d9d9bdbdbd969696636363252525","f7f7f7d9d9d9bdbdbd969696737373525252252525","fffffff0f0f0d9d9d9bdbdbd969696737373525252252525","fffffff0f0f0d9d9d9bdbdbd969696737373525252252525000000").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/sequential-single/Oranges.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("fee6cefdae6be6550d","feeddefdbe85fd8d3cd94701","feeddefdbe85fd8d3ce6550da63603","feeddefdd0a2fdae6bfd8d3ce6550da63603","feeddefdd0a2fdae6bfd8d3cf16913d948018c2d04","fff5ebfee6cefdd0a2fdae6bfd8d3cf16913d948018c2d04","fff5ebfee6cefdd0a2fdae6bfd8d3cf16913d94801a636037f2704").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/sequential-single/Purples.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("efedf5bcbddc756bb1","f2f0f7cbc9e29e9ac86a51a3","f2f0f7cbc9e29e9ac8756bb154278f","f2f0f7dadaebbcbddc9e9ac8756bb154278f","f2f0f7dadaebbcbddc9e9ac8807dba6a51a34a1486","fcfbfdefedf5dadaebbcbddc9e9ac8807dba6a51a34a1486","fcfbfdefedf5dadaebbcbddc9e9ac8807dba6a51a354278f3f007d").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale-chromatic/src/sequential-single/Reds.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scheme",function(){return o});var r=n("./node_modules/d3-scale-chromatic/src/colors.js"),i=n("./node_modules/d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("fee0d2fc9272de2d26","fee5d9fcae91fb6a4acb181d","fee5d9fcae91fb6a4ade2d26a50f15","fee5d9fcbba1fc9272fb6a4ade2d26a50f15","fee5d9fcbba1fc9272fb6a4aef3b2ccb181d99000d","fff5f0fee0d2fcbba1fc9272fb6a4aef3b2ccb181d99000d","fff5f0fee0d2fcbba1fc9272fb6a4aef3b2ccb181da50f1567000d").map(r.default);t.default=Object(i.default)(o)},"./node_modules/d3-scale/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-scale/src/band.js");n.d(t,"scaleBand",function(){return r.default}),n.d(t,"scalePoint",function(){return r.point});var i=n("./node_modules/d3-scale/src/identity.js");n.d(t,"scaleIdentity",function(){return i.default});var o=n("./node_modules/d3-scale/src/linear.js");n.d(t,"scaleLinear",function(){return o.default});var a=n("./node_modules/d3-scale/src/log.js");n.d(t,"scaleLog",function(){return a.default});var s=n("./node_modules/d3-scale/src/ordinal.js");n.d(t,"scaleOrdinal",function(){return s.default}),n.d(t,"scaleImplicit",function(){return s.implicit});var u=n("./node_modules/d3-scale/src/pow.js");n.d(t,"scalePow",function(){return u.default}),n.d(t,"scaleSqrt",function(){return u.sqrt});var c=n("./node_modules/d3-scale/src/quantile.js");n.d(t,"scaleQuantile",function(){return c.default});var l=n("./node_modules/d3-scale/src/quantize.js");n.d(t,"scaleQuantize",function(){return l.default});var d=n("./node_modules/d3-scale/src/threshold.js");n.d(t,"scaleThreshold",function(){return d.default});var p=n("./node_modules/d3-scale/src/time.js");n.d(t,"scaleTime",function(){return p.default});var f=n("./node_modules/d3-scale/src/utcTime.js");n.d(t,"scaleUtc",function(){return f.default});var h=n("./node_modules/d3-scale/src/sequential.js");n.d(t,"scaleSequential",function(){return h.default})},"./node_modules/d3-scale/src/array.js":function(e,t,n){"use strict";n.r(t),n.d(t,"map",function(){return i}),n.d(t,"slice",function(){return o});var r=Array.prototype,i=r.map,o=r.slice},"./node_modules/d3-scale/src/band.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o}),n.d(t,"point",function(){return a});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/d3-scale/src/ordinal.js");function o(){var e,t,n=Object(i.default)().unknown(void 0),a=n.domain,s=n.range,u=[0,1],c=!1,l=0,d=0,p=.5;function f(){var n=a().length,i=u[1]<u[0],o=u[i-0],f=u[1-i];e=(f-o)/Math.max(1,n-l+2*d),c&&(e=Math.floor(e)),o+=(f-o-e*(n-l))*p,t=e*(1-l),c&&(o=Math.round(o),t=Math.round(t));var h=Object(r.range)(n).map(function(t){return o+e*t});return s(i?h.reverse():h)}return delete n.unknown,n.domain=function(e){return arguments.length?(a(e),f()):a()},n.range=function(e){return arguments.length?(u=[+e[0],+e[1]],f()):u.slice()},n.rangeRound=function(e){return u=[+e[0],+e[1]],c=!0,f()},n.bandwidth=function(){return t},n.step=function(){return e},n.round=function(e){return arguments.length?(c=!!e,f()):c},n.padding=function(e){return arguments.length?(l=d=Math.max(0,Math.min(1,e)),f()):l},n.paddingInner=function(e){return arguments.length?(l=Math.max(0,Math.min(1,e)),f()):l},n.paddingOuter=function(e){return arguments.length?(d=Math.max(0,Math.min(1,e)),f()):d},n.align=function(e){return arguments.length?(p=Math.max(0,Math.min(1,e)),f()):p},n.copy=function(){return o().domain(a()).range(u).round(c).paddingInner(l).paddingOuter(d).align(p)},f()}function a(){return function e(t){var n=t.copy;return t.padding=t.paddingOuter,delete t.paddingInner,delete t.paddingOuter,t.copy=function(){return e(n())},t}(o().paddingInner(1))}},"./node_modules/d3-scale/src/constant.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return function(){return e}}},"./node_modules/d3-scale/src/continuous.js":function(e,t,n){"use strict";n.r(t),n.d(t,"deinterpolateLinear",function(){return c}),n.d(t,"copy",function(){return p}),n.d(t,"default",function(){return f});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/d3-interpolate/index.js"),o=n("./node_modules/d3-scale/src/array.js"),a=n("./node_modules/d3-scale/src/constant.js"),s=n("./node_modules/d3-scale/src/number.js"),u=[0,1];function c(e,t){return(t-=e=+e)?function(n){return(n-e)/t}:Object(a.default)(t)}function l(e,t,n,r){var i=e[0],o=e[1],a=t[0],s=t[1];return o<i?(i=n(o,i),a=r(s,a)):(i=n(i,o),a=r(a,s)),function(e){return a(i(e))}}function d(e,t,n,i){var o=Math.min(e.length,t.length)-1,a=new Array(o),s=new Array(o),u=-1;for(e[o]<e[0]&&(e=e.slice().reverse(),t=t.slice().reverse());++u<o;)a[u]=n(e[u],e[u+1]),s[u]=i(t[u],t[u+1]);return function(t){var n=Object(r.bisect)(e,t,1,o)-1;return s[n](a[n](t))}}function p(e,t){return t.domain(e.domain()).range(e.range()).interpolate(e.interpolate()).clamp(e.clamp())}function f(e,t){var n,r,a,p=u,f=u,h=i.interpolate,m=!1;function g(){return n=Math.min(p.length,f.length)>2?d:l,r=a=null,v}function v(t){return(r||(r=n(p,f,m?(i=e,function(e,t){var n=i(e=+e,t=+t);return function(r){return r<=e?0:r>=t?1:n(r)}}):e,h)))(+t);var i}return v.invert=function(e){return(a||(a=n(f,p,c,m?(r=t,function(e,t){var n=r(e=+e,t=+t);return function(r){return r<=0?e:r>=1?t:n(r)}}):t)))(+e);var r},v.domain=function(e){return arguments.length?(p=o.map.call(e,s.default),g()):p.slice()},v.range=function(e){return arguments.length?(f=o.slice.call(e),g()):f.slice()},v.rangeRound=function(e){return f=o.slice.call(e),h=i.interpolateRound,g()},v.clamp=function(e){return arguments.length?(m=!!e,g()):m},v.interpolate=function(e){return arguments.length?(h=e,g()):h},g()}},"./node_modules/d3-scale/src/identity.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/d3-scale/src/array.js"),i=n("./node_modules/d3-scale/src/linear.js"),o=n("./node_modules/d3-scale/src/number.js");function a(){var e=[0,1];function t(e){return+e}return t.invert=t,t.domain=t.range=function(n){return arguments.length?(e=r.map.call(n,o.default),t):e.slice()},t.copy=function(){return a().domain(e)},Object(i.linearish)(t)}},"./node_modules/d3-scale/src/linear.js":function(e,t,n){"use strict";n.r(t),n.d(t,"linearish",function(){return s}),n.d(t,"default",function(){return u});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/d3-interpolate/index.js"),o=n("./node_modules/d3-scale/src/continuous.js"),a=n("./node_modules/d3-scale/src/tickFormat.js");function s(e){var t=e.domain;return e.ticks=function(e){var n=t();return Object(r.ticks)(n[0],n[n.length-1],null==e?10:e)},e.tickFormat=function(e,n){return Object(a.default)(t(),e,n)},e.nice=function(n){null==n&&(n=10);var i,o=t(),a=0,s=o.length-1,u=o[a],c=o[s];return c<u&&(i=u,u=c,c=i,i=a,a=s,s=i),(i=Object(r.tickIncrement)(u,c,n))>0?(u=Math.floor(u/i)*i,c=Math.ceil(c/i)*i,i=Object(r.tickIncrement)(u,c,n)):i<0&&(u=Math.ceil(u*i)/i,c=Math.floor(c*i)/i,i=Object(r.tickIncrement)(u,c,n)),i>0?(o[a]=Math.floor(u/i)*i,o[s]=Math.ceil(c/i)*i,t(o)):i<0&&(o[a]=Math.ceil(u*i)/i,o[s]=Math.floor(c*i)/i,t(o)),e},e}function u(){var e=Object(o.default)(o.deinterpolateLinear,i.interpolateNumber);return e.copy=function(){return Object(o.copy)(e,u())},s(e)}},"./node_modules/d3-scale/src/log.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return h});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/d3-format/index.js"),o=n("./node_modules/d3-scale/src/constant.js"),a=n("./node_modules/d3-scale/src/nice.js"),s=n("./node_modules/d3-scale/src/continuous.js");function u(e,t){return(t=Math.log(t/e))?function(n){return Math.log(n/e)/t}:Object(o.default)(t)}function c(e,t){return e<0?function(n){return-Math.pow(-t,n)*Math.pow(-e,1-n)}:function(n){return Math.pow(t,n)*Math.pow(e,1-n)}}function l(e){return isFinite(e)?+("1e"+e):e<0?0:e}function d(e){return 10===e?l:e===Math.E?Math.exp:function(t){return Math.pow(e,t)}}function p(e){return e===Math.E?Math.log:10===e&&Math.log10||2===e&&Math.log2||(e=Math.log(e),function(t){return Math.log(t)/e})}function f(e){return function(t){return-e(-t)}}function h(){var e=Object(s.default)(u,c).domain([1,10]),t=e.domain,n=10,o=p(10),l=d(10);function m(){return o=p(n),l=d(n),t()[0]<0&&(o=f(o),l=f(l)),e}return e.base=function(e){return arguments.length?(n=+e,m()):n},e.domain=function(e){return arguments.length?(t(e),m()):t()},e.ticks=function(e){var i,a=t(),s=a[0],u=a[a.length-1];(i=u<s)&&(f=s,s=u,u=f);var c,d,p,f=o(s),h=o(u),m=null==e?10:+e,g=[];if(!(n%1)&&h-f<m){if(f=Math.round(f)-1,h=Math.round(h)+1,s>0){for(;f<h;++f)for(d=1,c=l(f);d<n;++d)if(!((p=c*d)<s)){if(p>u)break;g.push(p)}}else for(;f<h;++f)for(d=n-1,c=l(f);d>=1;--d)if(!((p=c*d)<s)){if(p>u)break;g.push(p)}}else g=Object(r.ticks)(f,h,Math.min(h-f,m)).map(l);return i?g.reverse():g},e.tickFormat=function(t,r){if(null==r&&(r=10===n?".0e":","),"function"!=typeof r&&(r=Object(i.format)(r)),t===1/0)return r;null==t&&(t=10);var a=Math.max(1,n*t/e.ticks().length);return function(e){var t=e/l(Math.round(o(e)));return t*n<n-.5&&(t*=n),t<=a?r(e):""}},e.nice=function(){return t(Object(a.default)(t(),{floor:function(e){return l(Math.floor(o(e)))},ceil:function(e){return l(Math.ceil(o(e)))}}))},e.copy=function(){return Object(s.copy)(e,h().base(n))},e}},"./node_modules/d3-scale/src/nice.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){var n,r=0,i=(e=e.slice()).length-1,o=e[r],a=e[i];return a<o&&(n=r,r=i,i=n,n=o,o=a,a=n),e[r]=t.floor(o),e[i]=t.ceil(a),e}},"./node_modules/d3-scale/src/number.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return+e}},"./node_modules/d3-scale/src/ordinal.js":function(e,t,n){"use strict";n.r(t),n.d(t,"implicit",function(){return o}),n.d(t,"default",function(){return a});var r=n("./node_modules/d3-collection/index.js"),i=n("./node_modules/d3-scale/src/array.js"),o={name:"implicit"};function a(e){var t=Object(r.map)(),n=[],s=o;function u(r){var i=r+"",a=t.get(i);if(!a){if(s!==o)return s;t.set(i,a=n.push(r))}return e[(a-1)%e.length]}return e=null==e?[]:i.slice.call(e),u.domain=function(e){if(!arguments.length)return n.slice();n=[],t=Object(r.map)();for(var i,o,a=-1,s=e.length;++a<s;)t.has(o=(i=e[a])+"")||t.set(o,n.push(i));return u},u.range=function(t){return arguments.length?(e=i.slice.call(t),u):e.slice()},u.unknown=function(e){return arguments.length?(s=e,u):s},u.copy=function(){return a().domain(n).range(e).unknown(s)},u}},"./node_modules/d3-scale/src/pow.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return s}),n.d(t,"sqrt",function(){return u});var r=n("./node_modules/d3-scale/src/constant.js"),i=n("./node_modules/d3-scale/src/linear.js"),o=n("./node_modules/d3-scale/src/continuous.js");function a(e,t){return e<0?-Math.pow(-e,t):Math.pow(e,t)}function s(){var e=1,t=Object(o.default)(function(t,n){return(n=a(n,e)-(t=a(t,e)))?function(r){return(a(r,e)-t)/n}:Object(r.default)(n)},function(t,n){return n=a(n,e)-(t=a(t,e)),function(r){return a(t+n*r,1/e)}}),n=t.domain;return t.exponent=function(t){return arguments.length?(e=+t,n(n())):e},t.copy=function(){return Object(o.copy)(t,s().exponent(e))},Object(i.linearish)(t)}function u(){return s().exponent(.5)}},"./node_modules/d3-scale/src/quantile.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/d3-scale/src/array.js");function o(){var e=[],t=[],n=[];function a(){var i=0,o=Math.max(1,t.length);for(n=new Array(o-1);++i<o;)n[i-1]=Object(r.quantile)(e,i/o);return s}function s(e){if(!isNaN(e=+e))return t[Object(r.bisect)(n,e)]}return s.invertExtent=function(r){var i=t.indexOf(r);return i<0?[NaN,NaN]:[i>0?n[i-1]:e[0],i<n.length?n[i]:e[e.length-1]]},s.domain=function(t){if(!arguments.length)return e.slice();e=[];for(var n,i=0,o=t.length;i<o;++i)n=t[i],null==n||isNaN(n=+n)||e.push(n);return e.sort(r.ascending),a()},s.range=function(e){return arguments.length?(t=i.slice.call(e),a()):t.slice()},s.quantiles=function(){return n.slice()},s.copy=function(){return o().domain(e).range(t)},s}},"./node_modules/d3-scale/src/quantize.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/d3-scale/src/array.js"),o=n("./node_modules/d3-scale/src/linear.js");function a(){var e=0,t=1,n=1,s=[.5],u=[0,1];function c(e){if(e<=e)return u[Object(r.bisect)(s,e,0,n)]}function l(){var r=-1;for(s=new Array(n);++r<n;)s[r]=((r+1)*t-(r-n)*e)/(n+1);return c}return c.domain=function(n){return arguments.length?(e=+n[0],t=+n[1],l()):[e,t]},c.range=function(e){return arguments.length?(n=(u=i.slice.call(e)).length-1,l()):u.slice()},c.invertExtent=function(r){var i=u.indexOf(r);return i<0?[NaN,NaN]:i<1?[e,s[0]]:i>=n?[s[n-1],t]:[s[i-1],s[i]]},c.copy=function(){return a().domain([e,t]).range(u)},Object(o.linearish)(c)}},"./node_modules/d3-scale/src/sequential.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return i});var r=n("./node_modules/d3-scale/src/linear.js");function i(e){var t=0,n=1,o=!1;function a(r){var i=(r-t)/(n-t);return e(o?Math.max(0,Math.min(1,i)):i)}return a.domain=function(e){return arguments.length?(t=+e[0],n=+e[1],a):[t,n]},a.clamp=function(e){return arguments.length?(o=!!e,a):o},a.interpolator=function(t){return arguments.length?(e=t,a):e},a.copy=function(){return i(e).domain([t,n]).clamp(o)},Object(r.linearish)(a)}},"./node_modules/d3-scale/src/threshold.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/d3-scale/src/array.js");function o(){var e=[.5],t=[0,1],n=1;function a(i){if(i<=i)return t[Object(r.bisect)(e,i,0,n)]}return a.domain=function(r){return arguments.length?(e=i.slice.call(r),n=Math.min(e.length,t.length-1),a):e.slice()},a.range=function(r){return arguments.length?(t=i.slice.call(r),n=Math.min(e.length,t.length-1),a):t.slice()},a.invertExtent=function(n){var r=t.indexOf(n);return[e[r-1],e[r]]},a.copy=function(){return o().domain(e).range(t)},a}},"./node_modules/d3-scale/src/tickFormat.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/d3-format/index.js");t.default=function(e,t,n){var o,a=e[0],s=e[e.length-1],u=Object(r.tickStep)(a,s,null==t?10:t);switch((n=Object(i.formatSpecifier)(null==n?",f":n)).type){case"s":var c=Math.max(Math.abs(a),Math.abs(s));return null!=n.precision||isNaN(o=Object(i.precisionPrefix)(u,c))||(n.precision=o),Object(i.formatPrefix)(n,c);case"":case"e":case"g":case"p":case"r":null!=n.precision||isNaN(o=Object(i.precisionRound)(u,Math.max(Math.abs(a),Math.abs(s))))||(n.precision=o-("e"===n.type));break;case"f":case"%":null!=n.precision||isNaN(o=Object(i.precisionFixed)(u))||(n.precision=o-2*("%"===n.type))}return Object(i.format)(n)}},"./node_modules/d3-scale/src/time.js":function(e,t,n){"use strict";n.r(t),n.d(t,"calendar",function(){return y});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/d3-interpolate/index.js"),o=n("./node_modules/d3-time/index.js"),a=n("./node_modules/d3-time-format/index.js"),s=n("./node_modules/d3-scale/src/array.js"),u=n("./node_modules/d3-scale/src/continuous.js"),c=n("./node_modules/d3-scale/src/nice.js"),l=1e3,d=60*l,p=60*d,f=24*p,h=7*f,m=30*f,g=365*f;function v(e){return new Date(e)}function _(e){return e instanceof Date?+e:+new Date(+e)}function y(e,t,n,o,a,b,j,x,O){var w=Object(u.default)(u.deinterpolateLinear,i.interpolateNumber),C=w.invert,E=w.domain,S=O(".%L"),k=O(":%S"),M=O("%I:%M"),A=O("%I %p"),T=O("%a %d"),D=O("%b %d"),I=O("%B"),P=O("%Y"),R=[[j,1,l],[j,5,5*l],[j,15,15*l],[j,30,30*l],[b,1,d],[b,5,5*d],[b,15,15*d],[b,30,30*d],[a,1,p],[a,3,3*p],[a,6,6*p],[a,12,12*p],[o,1,f],[o,2,2*f],[n,1,h],[t,1,m],[t,3,3*m],[e,1,g]];function N(r){return(j(r)<r?S:b(r)<r?k:a(r)<r?M:o(r)<r?A:t(r)<r?n(r)<r?T:D:e(r)<r?I:P)(r)}function F(t,n,i,o){if(null==t&&(t=10),"number"==typeof t){var a=Math.abs(i-n)/t,s=Object(r.bisector)(function(e){return e[2]}).right(R,a);s===R.length?(o=Object(r.tickStep)(n/g,i/g,t),t=e):s?(o=(s=R[a/R[s-1][2]<R[s][2]/a?s-1:s])[1],t=s[0]):(o=Math.max(Object(r.tickStep)(n,i,t),1),t=x)}return null==o?t:t.every(o)}return w.invert=function(e){return new Date(C(e))},w.domain=function(e){return arguments.length?E(s.map.call(e,_)):E().map(v)},w.ticks=function(e,t){var n,r=E(),i=r[0],o=r[r.length-1],a=o<i;return a&&(n=i,i=o,o=n),n=(n=F(e,i,o,t))?n.range(i,o+1):[],a?n.reverse():n},w.tickFormat=function(e,t){return null==t?N:O(t)},w.nice=function(e,t){var n=E();return(e=F(e,n[0],n[n.length-1],t))?E(Object(c.default)(n,e)):w},w.copy=function(){return Object(u.copy)(w,y(e,t,n,o,a,b,j,x,O))},w}t.default=function(){return y(o.timeYear,o.timeMonth,o.timeWeek,o.timeDay,o.timeHour,o.timeMinute,o.timeSecond,o.timeMillisecond,a.timeFormat).domain([new Date(2e3,0,1),new Date(2e3,0,2)])}},"./node_modules/d3-scale/src/utcTime.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-scale/src/time.js"),i=n("./node_modules/d3-time-format/index.js"),o=n("./node_modules/d3-time/index.js");t.default=function(){return Object(r.calendar)(o.utcYear,o.utcMonth,o.utcWeek,o.utcDay,o.utcHour,o.utcMinute,o.utcSecond,o.utcMillisecond,i.utcFormat).domain([Date.UTC(2e3,0,1),Date.UTC(2e3,0,2)])}},"./node_modules/d3-shape/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/arc.js");n.d(t,"arc",function(){return r.default});var i=n("./node_modules/d3-shape/src/area.js");n.d(t,"area",function(){return i.default});var o=n("./node_modules/d3-shape/src/line.js");n.d(t,"line",function(){return o.default});var a=n("./node_modules/d3-shape/src/pie.js");n.d(t,"pie",function(){return a.default});var s=n("./node_modules/d3-shape/src/areaRadial.js");n.d(t,"areaRadial",function(){return s.default}),n.d(t,"radialArea",function(){return s.default});var u=n("./node_modules/d3-shape/src/lineRadial.js");n.d(t,"lineRadial",function(){return u.default}),n.d(t,"radialLine",function(){return u.default});var c=n("./node_modules/d3-shape/src/pointRadial.js");n.d(t,"pointRadial",function(){return c.default});var l=n("./node_modules/d3-shape/src/link/index.js");n.d(t,"linkHorizontal",function(){return l.linkHorizontal}),n.d(t,"linkVertical",function(){return l.linkVertical}),n.d(t,"linkRadial",function(){return l.linkRadial});var d=n("./node_modules/d3-shape/src/symbol.js");n.d(t,"symbol",function(){return d.default}),n.d(t,"symbols",function(){return d.symbols});var p=n("./node_modules/d3-shape/src/symbol/circle.js");n.d(t,"symbolCircle",function(){return p.default});var f=n("./node_modules/d3-shape/src/symbol/cross.js");n.d(t,"symbolCross",function(){return f.default});var h=n("./node_modules/d3-shape/src/symbol/diamond.js");n.d(t,"symbolDiamond",function(){return h.default});var m=n("./node_modules/d3-shape/src/symbol/square.js");n.d(t,"symbolSquare",function(){return m.default});var g=n("./node_modules/d3-shape/src/symbol/star.js");n.d(t,"symbolStar",function(){return g.default});var v=n("./node_modules/d3-shape/src/symbol/triangle.js");n.d(t,"symbolTriangle",function(){return v.default});var _=n("./node_modules/d3-shape/src/symbol/wye.js");n.d(t,"symbolWye",function(){return _.default});var y=n("./node_modules/d3-shape/src/curve/basisClosed.js");n.d(t,"curveBasisClosed",function(){return y.default});var b=n("./node_modules/d3-shape/src/curve/basisOpen.js");n.d(t,"curveBasisOpen",function(){return b.default});var j=n("./node_modules/d3-shape/src/curve/basis.js");n.d(t,"curveBasis",function(){return j.default});var x=n("./node_modules/d3-shape/src/curve/bundle.js");n.d(t,"curveBundle",function(){return x.default});var O=n("./node_modules/d3-shape/src/curve/cardinalClosed.js");n.d(t,"curveCardinalClosed",function(){return O.default});var w=n("./node_modules/d3-shape/src/curve/cardinalOpen.js");n.d(t,"curveCardinalOpen",function(){return w.default});var C=n("./node_modules/d3-shape/src/curve/cardinal.js");n.d(t,"curveCardinal",function(){return C.default});var E=n("./node_modules/d3-shape/src/curve/catmullRomClosed.js");n.d(t,"curveCatmullRomClosed",function(){return E.default});var S=n("./node_modules/d3-shape/src/curve/catmullRomOpen.js");n.d(t,"curveCatmullRomOpen",function(){return S.default});var k=n("./node_modules/d3-shape/src/curve/catmullRom.js");n.d(t,"curveCatmullRom",function(){return k.default});var M=n("./node_modules/d3-shape/src/curve/linearClosed.js");n.d(t,"curveLinearClosed",function(){return M.default});var A=n("./node_modules/d3-shape/src/curve/linear.js");n.d(t,"curveLinear",function(){return A.default});var T=n("./node_modules/d3-shape/src/curve/monotone.js");n.d(t,"curveMonotoneX",function(){return T.monotoneX}),n.d(t,"curveMonotoneY",function(){return T.monotoneY});var D=n("./node_modules/d3-shape/src/curve/natural.js");n.d(t,"curveNatural",function(){return D.default});var I=n("./node_modules/d3-shape/src/curve/step.js");n.d(t,"curveStep",function(){return I.default}),n.d(t,"curveStepAfter",function(){return I.stepAfter}),n.d(t,"curveStepBefore",function(){return I.stepBefore});var P=n("./node_modules/d3-shape/src/stack.js");n.d(t,"stack",function(){return P.default});var R=n("./node_modules/d3-shape/src/offset/expand.js");n.d(t,"stackOffsetExpand",function(){return R.default});var N=n("./node_modules/d3-shape/src/offset/diverging.js");n.d(t,"stackOffsetDiverging",function(){return N.default});var F=n("./node_modules/d3-shape/src/offset/none.js");n.d(t,"stackOffsetNone",function(){return F.default});var L=n("./node_modules/d3-shape/src/offset/silhouette.js");n.d(t,"stackOffsetSilhouette",function(){return L.default});var V=n("./node_modules/d3-shape/src/offset/wiggle.js");n.d(t,"stackOffsetWiggle",function(){return V.default});var B=n("./node_modules/d3-shape/src/order/ascending.js");n.d(t,"stackOrderAscending",function(){return B.default});var z=n("./node_modules/d3-shape/src/order/descending.js");n.d(t,"stackOrderDescending",function(){return z.default});var U=n("./node_modules/d3-shape/src/order/insideOut.js");n.d(t,"stackOrderInsideOut",function(){return U.default});var H=n("./node_modules/d3-shape/src/order/none.js");n.d(t,"stackOrderNone",function(){return H.default});var q=n("./node_modules/d3-shape/src/order/reverse.js");n.d(t,"stackOrderReverse",function(){return q.default})},"./node_modules/d3-shape/src/arc.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-path/index.js"),i=n("./node_modules/d3-shape/src/constant.js"),o=n("./node_modules/d3-shape/src/math.js");function a(e){return e.innerRadius}function s(e){return e.outerRadius}function u(e){return e.startAngle}function c(e){return e.endAngle}function l(e){return e&&e.padAngle}function d(e,t,n,r,i,a,s){var u=e-n,c=t-r,l=(s?a:-a)/Object(o.sqrt)(u*u+c*c),d=l*c,p=-l*u,f=e+d,h=t+p,m=n+d,g=r+p,v=(f+m)/2,_=(h+g)/2,y=m-f,b=g-h,j=y*y+b*b,x=i-a,O=f*g-m*h,w=(b<0?-1:1)*Object(o.sqrt)(Object(o.max)(0,x*x*j-O*O)),C=(O*b-y*w)/j,E=(-O*y-b*w)/j,S=(O*b+y*w)/j,k=(-O*y+b*w)/j,M=C-v,A=E-_,T=S-v,D=k-_;return M*M+A*A>T*T+D*D&&(C=S,E=k),{cx:C,cy:E,x01:-d,y01:-p,x11:C*(i/x-1),y11:E*(i/x-1)}}t.default=function(){var e=a,t=s,n=Object(i.default)(0),p=null,f=u,h=c,m=l,g=null;function v(){var i,a,s,u,c,l,v,_,y,b,j,x=+e.apply(this,arguments),O=+t.apply(this,arguments),w=f.apply(this,arguments)-o.halfPi,C=h.apply(this,arguments)-o.halfPi,E=Object(o.abs)(C-w),S=C>w;if(g||(g=i=Object(r.path)()),O<x&&(a=O,O=x,x=a),O>o.epsilon)if(E>o.tau-o.epsilon)g.moveTo(O*Object(o.cos)(w),O*Object(o.sin)(w)),g.arc(0,0,O,w,C,!S),x>o.epsilon&&(g.moveTo(x*Object(o.cos)(C),x*Object(o.sin)(C)),g.arc(0,0,x,C,w,S));else{var k,M,A=w,T=C,D=w,I=C,P=E,R=E,N=m.apply(this,arguments)/2,F=N>o.epsilon&&(p?+p.apply(this,arguments):Object(o.sqrt)(x*x+O*O)),L=Object(o.min)(Object(o.abs)(O-x)/2,+n.apply(this,arguments)),V=L,B=L;if(F>o.epsilon){var z=Object(o.asin)(F/x*Object(o.sin)(N)),U=Object(o.asin)(F/O*Object(o.sin)(N));(P-=2*z)>o.epsilon?(D+=z*=S?1:-1,I-=z):(P=0,D=I=(w+C)/2),(R-=2*U)>o.epsilon?(A+=U*=S?1:-1,T-=U):(R=0,A=T=(w+C)/2)}var H=O*Object(o.cos)(A),q=O*Object(o.sin)(A),G=x*Object(o.cos)(I),W=x*Object(o.sin)(I);if(L>o.epsilon){var Y=O*Object(o.cos)(T),K=O*Object(o.sin)(T),Q=x*Object(o.cos)(D),X=x*Object(o.sin)(D);if(E<o.pi){var Z=P>o.epsilon?[(s=H)+(j=((y=G-(c=Y))*((u=q)-(l=K))-(b=W-l)*(s-c))/(b*(v=Q-s)-y*(_=X-u)))*v,u+j*_]:[G,W],$=H-Z[0],J=q-Z[1],ee=Y-Z[0],te=K-Z[1],ne=1/Object(o.sin)(Object(o.acos)(($*ee+J*te)/(Object(o.sqrt)($*$+J*J)*Object(o.sqrt)(ee*ee+te*te)))/2),re=Object(o.sqrt)(Z[0]*Z[0]+Z[1]*Z[1]);V=Object(o.min)(L,(x-re)/(ne-1)),B=Object(o.min)(L,(O-re)/(ne+1))}}R>o.epsilon?B>o.epsilon?(k=d(Q,X,H,q,O,B,S),M=d(Y,K,G,W,O,B,S),g.moveTo(k.cx+k.x01,k.cy+k.y01),B<L?g.arc(k.cx,k.cy,B,Object(o.atan2)(k.y01,k.x01),Object(o.atan2)(M.y01,M.x01),!S):(g.arc(k.cx,k.cy,B,Object(o.atan2)(k.y01,k.x01),Object(o.atan2)(k.y11,k.x11),!S),g.arc(0,0,O,Object(o.atan2)(k.cy+k.y11,k.cx+k.x11),Object(o.atan2)(M.cy+M.y11,M.cx+M.x11),!S),g.arc(M.cx,M.cy,B,Object(o.atan2)(M.y11,M.x11),Object(o.atan2)(M.y01,M.x01),!S))):(g.moveTo(H,q),g.arc(0,0,O,A,T,!S)):g.moveTo(H,q),x>o.epsilon&&P>o.epsilon?V>o.epsilon?(k=d(G,W,Y,K,x,-V,S),M=d(H,q,Q,X,x,-V,S),g.lineTo(k.cx+k.x01,k.cy+k.y01),V<L?g.arc(k.cx,k.cy,V,Object(o.atan2)(k.y01,k.x01),Object(o.atan2)(M.y01,M.x01),!S):(g.arc(k.cx,k.cy,V,Object(o.atan2)(k.y01,k.x01),Object(o.atan2)(k.y11,k.x11),!S),g.arc(0,0,x,Object(o.atan2)(k.cy+k.y11,k.cx+k.x11),Object(o.atan2)(M.cy+M.y11,M.cx+M.x11),S),g.arc(M.cx,M.cy,V,Object(o.atan2)(M.y11,M.x11),Object(o.atan2)(M.y01,M.x01),!S))):g.arc(0,0,x,I,D,S):g.lineTo(G,W)}else g.moveTo(0,0);if(g.closePath(),i)return g=null,i+""||null}return v.centroid=function(){var n=(+e.apply(this,arguments)+ +t.apply(this,arguments))/2,r=(+f.apply(this,arguments)+ +h.apply(this,arguments))/2-o.pi/2;return[Object(o.cos)(r)*n,Object(o.sin)(r)*n]},v.innerRadius=function(t){return arguments.length?(e="function"==typeof t?t:Object(i.default)(+t),v):e},v.outerRadius=function(e){return arguments.length?(t="function"==typeof e?e:Object(i.default)(+e),v):t},v.cornerRadius=function(e){return arguments.length?(n="function"==typeof e?e:Object(i.default)(+e),v):n},v.padRadius=function(e){return arguments.length?(p=null==e?null:"function"==typeof e?e:Object(i.default)(+e),v):p},v.startAngle=function(e){return arguments.length?(f="function"==typeof e?e:Object(i.default)(+e),v):f},v.endAngle=function(e){return arguments.length?(h="function"==typeof e?e:Object(i.default)(+e),v):h},v.padAngle=function(e){return arguments.length?(m="function"==typeof e?e:Object(i.default)(+e),v):m},v.context=function(e){return arguments.length?(g=null==e?null:e,v):g},v}},"./node_modules/d3-shape/src/area.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-path/index.js"),i=n("./node_modules/d3-shape/src/constant.js"),o=n("./node_modules/d3-shape/src/curve/linear.js"),a=n("./node_modules/d3-shape/src/line.js"),s=n("./node_modules/d3-shape/src/point.js");t.default=function(){var e=s.x,t=null,n=Object(i.default)(0),u=s.y,c=Object(i.default)(!0),l=null,d=o.default,p=null;function f(i){var o,a,s,f,h,m=i.length,g=!1,v=new Array(m),_=new Array(m);for(null==l&&(p=d(h=Object(r.path)())),o=0;o<=m;++o){if(!(o<m&&c(f=i[o],o,i))===g)if(g=!g)a=o,p.areaStart(),p.lineStart();else{for(p.lineEnd(),p.lineStart(),s=o-1;s>=a;--s)p.point(v[s],_[s]);p.lineEnd(),p.areaEnd()}g&&(v[o]=+e(f,o,i),_[o]=+n(f,o,i),p.point(t?+t(f,o,i):v[o],u?+u(f,o,i):_[o]))}if(h)return p=null,h+""||null}function h(){return Object(a.default)().defined(c).curve(d).context(l)}return f.x=function(n){return arguments.length?(e="function"==typeof n?n:Object(i.default)(+n),t=null,f):e},f.x0=function(t){return arguments.length?(e="function"==typeof t?t:Object(i.default)(+t),f):e},f.x1=function(e){return arguments.length?(t=null==e?null:"function"==typeof e?e:Object(i.default)(+e),f):t},f.y=function(e){return arguments.length?(n="function"==typeof e?e:Object(i.default)(+e),u=null,f):n},f.y0=function(e){return arguments.length?(n="function"==typeof e?e:Object(i.default)(+e),f):n},f.y1=function(e){return arguments.length?(u=null==e?null:"function"==typeof e?e:Object(i.default)(+e),f):u},f.lineX0=f.lineY0=function(){return h().x(e).y(n)},f.lineY1=function(){return h().x(e).y(u)},f.lineX1=function(){return h().x(t).y(n)},f.defined=function(e){return arguments.length?(c="function"==typeof e?e:Object(i.default)(!!e),f):c},f.curve=function(e){return arguments.length?(d=e,null!=l&&(p=d(l)),f):d},f.context=function(e){return arguments.length?(null==e?l=p=null:p=d(l=e),f):l},f}},"./node_modules/d3-shape/src/areaRadial.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/curve/radial.js"),i=n("./node_modules/d3-shape/src/area.js"),o=n("./node_modules/d3-shape/src/lineRadial.js");t.default=function(){var e=Object(i.default)().curve(r.curveRadialLinear),t=e.curve,n=e.lineX0,a=e.lineX1,s=e.lineY0,u=e.lineY1;return e.angle=e.x,delete e.x,e.startAngle=e.x0,delete e.x0,e.endAngle=e.x1,delete e.x1,e.radius=e.y,delete e.y,e.innerRadius=e.y0,delete e.y0,e.outerRadius=e.y1,delete e.y1,e.lineStartAngle=function(){return Object(o.lineRadial)(n())},delete e.lineX0,e.lineEndAngle=function(){return Object(o.lineRadial)(a())},delete e.lineX1,e.lineInnerRadius=function(){return Object(o.lineRadial)(s())},delete e.lineY0,e.lineOuterRadius=function(){return Object(o.lineRadial)(u())},delete e.lineY1,e.curve=function(e){return arguments.length?t(Object(r.default)(e)):t()._curve},e}},"./node_modules/d3-shape/src/array.js":function(e,t,n){"use strict";n.r(t),n.d(t,"slice",function(){return r});var r=Array.prototype.slice},"./node_modules/d3-shape/src/constant.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return function(){return e}}},"./node_modules/d3-shape/src/curve/basis.js":function(e,t,n){"use strict";function r(e,t,n){e._context.bezierCurveTo((2*e._x0+e._x1)/3,(2*e._y0+e._y1)/3,(e._x0+2*e._x1)/3,(e._y0+2*e._y1)/3,(e._x0+4*e._x1+t)/6,(e._y0+4*e._y1+n)/6)}function i(e){this._context=e}n.r(t),n.d(t,"point",function(){return r}),n.d(t,"Basis",function(){return i}),i.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=NaN,this._point=0},lineEnd:function(){switch(this._point){case 3:r(this,this._x1,this._y1);case 2:this._context.lineTo(this._x1,this._y1)}(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2;break;case 2:this._point=3,this._context.lineTo((5*this._x0+this._x1)/6,(5*this._y0+this._y1)/6);default:r(this,e,t)}this._x0=this._x1,this._x1=e,this._y0=this._y1,this._y1=t}},t.default=function(e){return new i(e)}},"./node_modules/d3-shape/src/curve/basisClosed.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/noop.js"),i=n("./node_modules/d3-shape/src/curve/basis.js");function o(e){this._context=e}o.prototype={areaStart:r.default,areaEnd:r.default,lineStart:function(){this._x0=this._x1=this._x2=this._x3=this._x4=this._y0=this._y1=this._y2=this._y3=this._y4=NaN,this._point=0},lineEnd:function(){switch(this._point){case 1:this._context.moveTo(this._x2,this._y2),this._context.closePath();break;case 2:this._context.moveTo((this._x2+2*this._x3)/3,(this._y2+2*this._y3)/3),this._context.lineTo((this._x3+2*this._x2)/3,(this._y3+2*this._y2)/3),this._context.closePath();break;case 3:this.point(this._x2,this._y2),this.point(this._x3,this._y3),this.point(this._x4,this._y4)}},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._x2=e,this._y2=t;break;case 1:this._point=2,this._x3=e,this._y3=t;break;case 2:this._point=3,this._x4=e,this._y4=t,this._context.moveTo((this._x0+4*this._x1+e)/6,(this._y0+4*this._y1+t)/6);break;default:Object(i.point)(this,e,t)}this._x0=this._x1,this._x1=e,this._y0=this._y1,this._y1=t}},t.default=function(e){return new o(e)}},"./node_modules/d3-shape/src/curve/basisOpen.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/curve/basis.js");function i(e){this._context=e}i.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=NaN,this._point=0},lineEnd:function(){(this._line||0!==this._line&&3===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1;break;case 1:this._point=2;break;case 2:this._point=3;var n=(this._x0+4*this._x1+e)/6,i=(this._y0+4*this._y1+t)/6;this._line?this._context.lineTo(n,i):this._context.moveTo(n,i);break;case 3:this._point=4;default:Object(r.point)(this,e,t)}this._x0=this._x1,this._x1=e,this._y0=this._y1,this._y1=t}},t.default=function(e){return new i(e)}},"./node_modules/d3-shape/src/curve/bundle.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/curve/basis.js");function i(e,t){this._basis=new r.Basis(e),this._beta=t}i.prototype={lineStart:function(){this._x=[],this._y=[],this._basis.lineStart()},lineEnd:function(){var e=this._x,t=this._y,n=e.length-1;if(n>0)for(var r,i=e[0],o=t[0],a=e[n]-i,s=t[n]-o,u=-1;++u<=n;)r=u/n,this._basis.point(this._beta*e[u]+(1-this._beta)*(i+r*a),this._beta*t[u]+(1-this._beta)*(o+r*s));this._x=this._y=null,this._basis.lineEnd()},point:function(e,t){this._x.push(+e),this._y.push(+t)}},t.default=function e(t){function n(e){return 1===t?new r.Basis(e):new i(e,t)}return n.beta=function(t){return e(+t)},n}(.85)},"./node_modules/d3-shape/src/curve/cardinal.js":function(e,t,n){"use strict";function r(e,t,n){e._context.bezierCurveTo(e._x1+e._k*(e._x2-e._x0),e._y1+e._k*(e._y2-e._y0),e._x2+e._k*(e._x1-t),e._y2+e._k*(e._y1-n),e._x2,e._y2)}function i(e,t){this._context=e,this._k=(1-t)/6}n.r(t),n.d(t,"point",function(){return r}),n.d(t,"Cardinal",function(){return i}),i.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._x2=this._y0=this._y1=this._y2=NaN,this._point=0},lineEnd:function(){switch(this._point){case 2:this._context.lineTo(this._x2,this._y2);break;case 3:r(this,this._x1,this._y1)}(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2,this._x1=e,this._y1=t;break;case 2:this._point=3;default:r(this,e,t)}this._x0=this._x1,this._x1=this._x2,this._x2=e,this._y0=this._y1,this._y1=this._y2,this._y2=t}},t.default=function e(t){function n(e){return new i(e,t)}return n.tension=function(t){return e(+t)},n}(0)},"./node_modules/d3-shape/src/curve/cardinalClosed.js":function(e,t,n){"use strict";n.r(t),n.d(t,"CardinalClosed",function(){return o});var r=n("./node_modules/d3-shape/src/noop.js"),i=n("./node_modules/d3-shape/src/curve/cardinal.js");function o(e,t){this._context=e,this._k=(1-t)/6}o.prototype={areaStart:r.default,areaEnd:r.default,lineStart:function(){this._x0=this._x1=this._x2=this._x3=this._x4=this._x5=this._y0=this._y1=this._y2=this._y3=this._y4=this._y5=NaN,this._point=0},lineEnd:function(){switch(this._point){case 1:this._context.moveTo(this._x3,this._y3),this._context.closePath();break;case 2:this._context.lineTo(this._x3,this._y3),this._context.closePath();break;case 3:this.point(this._x3,this._y3),this.point(this._x4,this._y4),this.point(this._x5,this._y5)}},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._x3=e,this._y3=t;break;case 1:this._point=2,this._context.moveTo(this._x4=e,this._y4=t);break;case 2:this._point=3,this._x5=e,this._y5=t;break;default:Object(i.point)(this,e,t)}this._x0=this._x1,this._x1=this._x2,this._x2=e,this._y0=this._y1,this._y1=this._y2,this._y2=t}},t.default=function e(t){function n(e){return new o(e,t)}return n.tension=function(t){return e(+t)},n}(0)},"./node_modules/d3-shape/src/curve/cardinalOpen.js":function(e,t,n){"use strict";n.r(t),n.d(t,"CardinalOpen",function(){return i});var r=n("./node_modules/d3-shape/src/curve/cardinal.js");function i(e,t){this._context=e,this._k=(1-t)/6}i.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._x2=this._y0=this._y1=this._y2=NaN,this._point=0},lineEnd:function(){(this._line||0!==this._line&&3===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1;break;case 1:this._point=2;break;case 2:this._point=3,this._line?this._context.lineTo(this._x2,this._y2):this._context.moveTo(this._x2,this._y2);break;case 3:this._point=4;default:Object(r.point)(this,e,t)}this._x0=this._x1,this._x1=this._x2,this._x2=e,this._y0=this._y1,this._y1=this._y2,this._y2=t}},t.default=function e(t){function n(e){return new i(e,t)}return n.tension=function(t){return e(+t)},n}(0)},"./node_modules/d3-shape/src/curve/catmullRom.js":function(e,t,n){"use strict";n.r(t),n.d(t,"point",function(){return o});var r=n("./node_modules/d3-shape/src/math.js"),i=n("./node_modules/d3-shape/src/curve/cardinal.js");function o(e,t,n){var i=e._x1,o=e._y1,a=e._x2,s=e._y2;if(e._l01_a>r.epsilon){var u=2*e._l01_2a+3*e._l01_a*e._l12_a+e._l12_2a,c=3*e._l01_a*(e._l01_a+e._l12_a);i=(i*u-e._x0*e._l12_2a+e._x2*e._l01_2a)/c,o=(o*u-e._y0*e._l12_2a+e._y2*e._l01_2a)/c}if(e._l23_a>r.epsilon){var l=2*e._l23_2a+3*e._l23_a*e._l12_a+e._l12_2a,d=3*e._l23_a*(e._l23_a+e._l12_a);a=(a*l+e._x1*e._l23_2a-t*e._l12_2a)/d,s=(s*l+e._y1*e._l23_2a-n*e._l12_2a)/d}e._context.bezierCurveTo(i,o,a,s,e._x2,e._y2)}function a(e,t){this._context=e,this._alpha=t}a.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._x2=this._y0=this._y1=this._y2=NaN,this._l01_a=this._l12_a=this._l23_a=this._l01_2a=this._l12_2a=this._l23_2a=this._point=0},lineEnd:function(){switch(this._point){case 2:this._context.lineTo(this._x2,this._y2);break;case 3:this.point(this._x2,this._y2)}(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){if(e=+e,t=+t,this._point){var n=this._x2-e,r=this._y2-t;this._l23_a=Math.sqrt(this._l23_2a=Math.pow(n*n+r*r,this._alpha))}switch(this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2;break;case 2:this._point=3;default:o(this,e,t)}this._l01_a=this._l12_a,this._l12_a=this._l23_a,this._l01_2a=this._l12_2a,this._l12_2a=this._l23_2a,this._x0=this._x1,this._x1=this._x2,this._x2=e,this._y0=this._y1,this._y1=this._y2,this._y2=t}},t.default=function e(t){function n(e){return t?new a(e,t):new i.Cardinal(e,0)}return n.alpha=function(t){return e(+t)},n}(.5)},"./node_modules/d3-shape/src/curve/catmullRomClosed.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/curve/cardinalClosed.js"),i=n("./node_modules/d3-shape/src/noop.js"),o=n("./node_modules/d3-shape/src/curve/catmullRom.js");function a(e,t){this._context=e,this._alpha=t}a.prototype={areaStart:i.default,areaEnd:i.default,lineStart:function(){this._x0=this._x1=this._x2=this._x3=this._x4=this._x5=this._y0=this._y1=this._y2=this._y3=this._y4=this._y5=NaN,this._l01_a=this._l12_a=this._l23_a=this._l01_2a=this._l12_2a=this._l23_2a=this._point=0},lineEnd:function(){switch(this._point){case 1:this._context.moveTo(this._x3,this._y3),this._context.closePath();break;case 2:this._context.lineTo(this._x3,this._y3),this._context.closePath();break;case 3:this.point(this._x3,this._y3),this.point(this._x4,this._y4),this.point(this._x5,this._y5)}},point:function(e,t){if(e=+e,t=+t,this._point){var n=this._x2-e,r=this._y2-t;this._l23_a=Math.sqrt(this._l23_2a=Math.pow(n*n+r*r,this._alpha))}switch(this._point){case 0:this._point=1,this._x3=e,this._y3=t;break;case 1:this._point=2,this._context.moveTo(this._x4=e,this._y4=t);break;case 2:this._point=3,this._x5=e,this._y5=t;break;default:Object(o.point)(this,e,t)}this._l01_a=this._l12_a,this._l12_a=this._l23_a,this._l01_2a=this._l12_2a,this._l12_2a=this._l23_2a,this._x0=this._x1,this._x1=this._x2,this._x2=e,this._y0=this._y1,this._y1=this._y2,this._y2=t}},t.default=function e(t){function n(e){return t?new a(e,t):new r.CardinalClosed(e,0)}return n.alpha=function(t){return e(+t)},n}(.5)},"./node_modules/d3-shape/src/curve/catmullRomOpen.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/curve/cardinalOpen.js"),i=n("./node_modules/d3-shape/src/curve/catmullRom.js");function o(e,t){this._context=e,this._alpha=t}o.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._x2=this._y0=this._y1=this._y2=NaN,this._l01_a=this._l12_a=this._l23_a=this._l01_2a=this._l12_2a=this._l23_2a=this._point=0},lineEnd:function(){(this._line||0!==this._line&&3===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){if(e=+e,t=+t,this._point){var n=this._x2-e,r=this._y2-t;this._l23_a=Math.sqrt(this._l23_2a=Math.pow(n*n+r*r,this._alpha))}switch(this._point){case 0:this._point=1;break;case 1:this._point=2;break;case 2:this._point=3,this._line?this._context.lineTo(this._x2,this._y2):this._context.moveTo(this._x2,this._y2);break;case 3:this._point=4;default:Object(i.point)(this,e,t)}this._l01_a=this._l12_a,this._l12_a=this._l23_a,this._l01_2a=this._l12_2a,this._l12_2a=this._l23_2a,this._x0=this._x1,this._x1=this._x2,this._x2=e,this._y0=this._y1,this._y1=this._y2,this._y2=t}},t.default=function e(t){function n(e){return t?new o(e,t):new r.CardinalOpen(e,0)}return n.alpha=function(t){return e(+t)},n}(.5)},"./node_modules/d3-shape/src/curve/linear.js":function(e,t,n){"use strict";function r(e){this._context=e}n.r(t),r.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._point=0},lineEnd:function(){(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2;default:this._context.lineTo(e,t)}}},t.default=function(e){return new r(e)}},"./node_modules/d3-shape/src/curve/linearClosed.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/noop.js");function i(e){this._context=e}i.prototype={areaStart:r.default,areaEnd:r.default,lineStart:function(){this._point=0},lineEnd:function(){this._point&&this._context.closePath()},point:function(e,t){e=+e,t=+t,this._point?this._context.lineTo(e,t):(this._point=1,this._context.moveTo(e,t))}},t.default=function(e){return new i(e)}},"./node_modules/d3-shape/src/curve/monotone.js":function(e,t,n){"use strict";function r(e){return e<0?-1:1}function i(e,t,n){var i=e._x1-e._x0,o=t-e._x1,a=(e._y1-e._y0)/(i||o<0&&-0),s=(n-e._y1)/(o||i<0&&-0),u=(a*o+s*i)/(i+o);return(r(a)+r(s))*Math.min(Math.abs(a),Math.abs(s),.5*Math.abs(u))||0}function o(e,t){var n=e._x1-e._x0;return n?(3*(e._y1-e._y0)/n-t)/2:t}function a(e,t,n){var r=e._x0,i=e._y0,o=e._x1,a=e._y1,s=(o-r)/3;e._context.bezierCurveTo(r+s,i+s*t,o-s,a-s*n,o,a)}function s(e){this._context=e}function u(e){this._context=new c(e)}function c(e){this._context=e}function l(e){return new s(e)}function d(e){return new u(e)}n.r(t),n.d(t,"monotoneX",function(){return l}),n.d(t,"monotoneY",function(){return d}),s.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=this._t0=NaN,this._point=0},lineEnd:function(){switch(this._point){case 2:this._context.lineTo(this._x1,this._y1);break;case 3:a(this,this._t0,o(this,this._t0))}(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){var n=NaN;if(t=+t,(e=+e)!==this._x1||t!==this._y1){switch(this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2;break;case 2:this._point=3,a(this,o(this,n=i(this,e,t)),n);break;default:a(this,this._t0,n=i(this,e,t))}this._x0=this._x1,this._x1=e,this._y0=this._y1,this._y1=t,this._t0=n}}},(u.prototype=Object.create(s.prototype)).point=function(e,t){s.prototype.point.call(this,t,e)},c.prototype={moveTo:function(e,t){this._context.moveTo(t,e)},closePath:function(){this._context.closePath()},lineTo:function(e,t){this._context.lineTo(t,e)},bezierCurveTo:function(e,t,n,r,i,o){this._context.bezierCurveTo(t,e,r,n,o,i)}}},"./node_modules/d3-shape/src/curve/natural.js":function(e,t,n){"use strict";function r(e){this._context=e}function i(e){var t,n,r=e.length-1,i=new Array(r),o=new Array(r),a=new Array(r);for(i[0]=0,o[0]=2,a[0]=e[0]+2*e[1],t=1;t<r-1;++t)i[t]=1,o[t]=4,a[t]=4*e[t]+2*e[t+1];for(i[r-1]=2,o[r-1]=7,a[r-1]=8*e[r-1]+e[r],t=1;t<r;++t)n=i[t]/o[t-1],o[t]-=n,a[t]-=n*a[t-1];for(i[r-1]=a[r-1]/o[r-1],t=r-2;t>=0;--t)i[t]=(a[t]-i[t+1])/o[t];for(o[r-1]=(e[r]+i[r-1])/2,t=0;t<r-1;++t)o[t]=2*e[t+1]-i[t+1];return[i,o]}n.r(t),r.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x=[],this._y=[]},lineEnd:function(){var e=this._x,t=this._y,n=e.length;if(n)if(this._line?this._context.lineTo(e[0],t[0]):this._context.moveTo(e[0],t[0]),2===n)this._context.lineTo(e[1],t[1]);else for(var r=i(e),o=i(t),a=0,s=1;s<n;++a,++s)this._context.bezierCurveTo(r[0][a],o[0][a],r[1][a],o[1][a],e[s],t[s]);(this._line||0!==this._line&&1===n)&&this._context.closePath(),this._line=1-this._line,this._x=this._y=null},point:function(e,t){this._x.push(+e),this._y.push(+t)}},t.default=function(e){return new r(e)}},"./node_modules/d3-shape/src/curve/radial.js":function(e,t,n){"use strict";n.r(t),n.d(t,"curveRadialLinear",function(){return r}),n.d(t,"default",function(){return o});var r=o(n("./node_modules/d3-shape/src/curve/linear.js").default);function i(e){this._curve=e}function o(e){function t(t){return new i(e(t))}return t._curve=e,t}i.prototype={areaStart:function(){this._curve.areaStart()},areaEnd:function(){this._curve.areaEnd()},lineStart:function(){this._curve.lineStart()},lineEnd:function(){this._curve.lineEnd()},point:function(e,t){this._curve.point(t*Math.sin(e),t*-Math.cos(e))}}},"./node_modules/d3-shape/src/curve/step.js":function(e,t,n){"use strict";function r(e,t){this._context=e,this._t=t}function i(e){return new r(e,0)}function o(e){return new r(e,1)}n.r(t),n.d(t,"stepBefore",function(){return i}),n.d(t,"stepAfter",function(){return o}),r.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x=this._y=NaN,this._point=0},lineEnd:function(){0<this._t&&this._t<1&&2===this._point&&this._context.lineTo(this._x,this._y),(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line>=0&&(this._t=1-this._t,this._line=1-this._line)},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2;default:if(this._t<=0)this._context.lineTo(this._x,t),this._context.lineTo(e,t);else{var n=this._x*(1-this._t)+e*this._t;this._context.lineTo(n,this._y),this._context.lineTo(n,t)}}this._x=e,this._y=t}},t.default=function(e){return new r(e,.5)}},"./node_modules/d3-shape/src/descending.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){return t<e?-1:t>e?1:t>=e?0:NaN}},"./node_modules/d3-shape/src/identity.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return e}},"./node_modules/d3-shape/src/line.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-path/index.js"),i=n("./node_modules/d3-shape/src/constant.js"),o=n("./node_modules/d3-shape/src/curve/linear.js"),a=n("./node_modules/d3-shape/src/point.js");t.default=function(){var e=a.x,t=a.y,n=Object(i.default)(!0),s=null,u=o.default,c=null;function l(i){var o,a,l,d=i.length,p=!1;for(null==s&&(c=u(l=Object(r.path)())),o=0;o<=d;++o)!(o<d&&n(a=i[o],o,i))===p&&((p=!p)?c.lineStart():c.lineEnd()),p&&c.point(+e(a,o,i),+t(a,o,i));if(l)return c=null,l+""||null}return l.x=function(t){return arguments.length?(e="function"==typeof t?t:Object(i.default)(+t),l):e},l.y=function(e){return arguments.length?(t="function"==typeof e?e:Object(i.default)(+e),l):t},l.defined=function(e){return arguments.length?(n="function"==typeof e?e:Object(i.default)(!!e),l):n},l.curve=function(e){return arguments.length?(u=e,null!=s&&(c=u(s)),l):u},l.context=function(e){return arguments.length?(null==e?s=c=null:c=u(s=e),l):s},l}},"./node_modules/d3-shape/src/lineRadial.js":function(e,t,n){"use strict";n.r(t),n.d(t,"lineRadial",function(){return o});var r=n("./node_modules/d3-shape/src/curve/radial.js"),i=n("./node_modules/d3-shape/src/line.js");function o(e){var t=e.curve;return e.angle=e.x,delete e.x,e.radius=e.y,delete e.y,e.curve=function(e){return arguments.length?t(Object(r.default)(e)):t()._curve},e}t.default=function(){return o(Object(i.default)().curve(r.curveRadialLinear))}},"./node_modules/d3-shape/src/link/index.js":function(e,t,n){"use strict";n.r(t),n.d(t,"linkHorizontal",function(){return h}),n.d(t,"linkVertical",function(){return m}),n.d(t,"linkRadial",function(){return g});var r=n("./node_modules/d3-path/index.js"),i=n("./node_modules/d3-shape/src/array.js"),o=n("./node_modules/d3-shape/src/constant.js"),a=n("./node_modules/d3-shape/src/point.js"),s=n("./node_modules/d3-shape/src/pointRadial.js");function u(e){return e.source}function c(e){return e.target}function l(e){var t=u,n=c,s=a.x,l=a.y,d=null;function p(){var o,a=i.slice.call(arguments),u=t.apply(this,a),c=n.apply(this,a);if(d||(d=o=Object(r.path)()),e(d,+s.apply(this,(a[0]=u,a)),+l.apply(this,a),+s.apply(this,(a[0]=c,a)),+l.apply(this,a)),o)return d=null,o+""||null}return p.source=function(e){return arguments.length?(t=e,p):t},p.target=function(e){return arguments.length?(n=e,p):n},p.x=function(e){return arguments.length?(s="function"==typeof e?e:Object(o.default)(+e),p):s},p.y=function(e){return arguments.length?(l="function"==typeof e?e:Object(o.default)(+e),p):l},p.context=function(e){return arguments.length?(d=null==e?null:e,p):d},p}function d(e,t,n,r,i){e.moveTo(t,n),e.bezierCurveTo(t=(t+r)/2,n,t,i,r,i)}function p(e,t,n,r,i){e.moveTo(t,n),e.bezierCurveTo(t,n=(n+i)/2,r,n,r,i)}function f(e,t,n,r,i){var o=Object(s.default)(t,n),a=Object(s.default)(t,n=(n+i)/2),u=Object(s.default)(r,n),c=Object(s.default)(r,i);e.moveTo(o[0],o[1]),e.bezierCurveTo(a[0],a[1],u[0],u[1],c[0],c[1])}function h(){return l(d)}function m(){return l(p)}function g(){var e=l(f);return e.angle=e.x,delete e.x,e.radius=e.y,delete e.y,e}},"./node_modules/d3-shape/src/math.js":function(e,t,n){"use strict";n.r(t),n.d(t,"abs",function(){return r}),n.d(t,"atan2",function(){return i}),n.d(t,"cos",function(){return o}),n.d(t,"max",function(){return a}),n.d(t,"min",function(){return s}),n.d(t,"sin",function(){return u}),n.d(t,"sqrt",function(){return c}),n.d(t,"epsilon",function(){return l}),n.d(t,"pi",function(){return d}),n.d(t,"halfPi",function(){return p}),n.d(t,"tau",function(){return f}),n.d(t,"acos",function(){return h}),n.d(t,"asin",function(){return m});var r=Math.abs,i=Math.atan2,o=Math.cos,a=Math.max,s=Math.min,u=Math.sin,c=Math.sqrt,l=1e-12,d=Math.PI,p=d/2,f=2*d;function h(e){return e>1?0:e<-1?d:Math.acos(e)}function m(e){return e>=1?p:e<=-1?-p:Math.asin(e)}},"./node_modules/d3-shape/src/noop.js":function(e,t,n){"use strict";n.r(t),t.default=function(){}},"./node_modules/d3-shape/src/offset/diverging.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){if((s=e.length)>1)for(var n,r,i,o,a,s,u=0,c=e[t[0]].length;u<c;++u)for(o=a=0,n=0;n<s;++n)(i=(r=e[t[n]][u])[1]-r[0])>=0?(r[0]=o,r[1]=o+=i):i<0?(r[1]=a,r[0]=a+=i):r[0]=o}},"./node_modules/d3-shape/src/offset/expand.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/offset/none.js");t.default=function(e,t){if((i=e.length)>0){for(var n,i,o,a=0,s=e[0].length;a<s;++a){for(o=n=0;n<i;++n)o+=e[n][a][1]||0;if(o)for(n=0;n<i;++n)e[n][a][1]/=o}Object(r.default)(e,t)}}},"./node_modules/d3-shape/src/offset/none.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){if((i=e.length)>1)for(var n,r,i,o=1,a=e[t[0]],s=a.length;o<i;++o)for(r=a,a=e[t[o]],n=0;n<s;++n)a[n][1]+=a[n][0]=isNaN(r[n][1])?r[n][0]:r[n][1]}},"./node_modules/d3-shape/src/offset/silhouette.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/offset/none.js");t.default=function(e,t){if((n=e.length)>0){for(var n,i=0,o=e[t[0]],a=o.length;i<a;++i){for(var s=0,u=0;s<n;++s)u+=e[s][i][1]||0;o[i][1]+=o[i][0]=-u/2}Object(r.default)(e,t)}}},"./node_modules/d3-shape/src/offset/wiggle.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/offset/none.js");t.default=function(e,t){if((o=e.length)>0&&(i=(n=e[t[0]]).length)>0){for(var n,i,o,a=0,s=1;s<i;++s){for(var u=0,c=0,l=0;u<o;++u){for(var d=e[t[u]],p=d[s][1]||0,f=(p-(d[s-1][1]||0))/2,h=0;h<u;++h){var m=e[t[h]];f+=(m[s][1]||0)-(m[s-1][1]||0)}c+=p,l+=f*p}n[s-1][1]+=n[s-1][0]=a,c&&(a-=l/c)}n[s-1][1]+=n[s-1][0]=a,Object(r.default)(e,t)}}},"./node_modules/d3-shape/src/order/ascending.js":function(e,t,n){"use strict";n.r(t),n.d(t,"sum",function(){return i});var r=n("./node_modules/d3-shape/src/order/none.js");function i(e){for(var t,n=0,r=-1,i=e.length;++r<i;)(t=+e[r][1])&&(n+=t);return n}t.default=function(e){var t=e.map(i);return Object(r.default)(e).sort(function(e,n){return t[e]-t[n]})}},"./node_modules/d3-shape/src/order/descending.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/order/ascending.js");t.default=function(e){return Object(r.default)(e).reverse()}},"./node_modules/d3-shape/src/order/insideOut.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/order/none.js"),i=n("./node_modules/d3-shape/src/order/ascending.js");t.default=function(e){var t,n,o=e.length,a=e.map(i.sum),s=Object(r.default)(e).sort(function(e,t){return a[t]-a[e]}),u=0,c=0,l=[],d=[];for(t=0;t<o;++t)n=s[t],u<c?(u+=a[n],l.push(n)):(c+=a[n],d.push(n));return d.reverse().concat(l)}},"./node_modules/d3-shape/src/order/none.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){for(var t=e.length,n=new Array(t);--t>=0;)n[t]=t;return n}},"./node_modules/d3-shape/src/order/reverse.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/order/none.js");t.default=function(e){return Object(r.default)(e).reverse()}},"./node_modules/d3-shape/src/pie.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/constant.js"),i=n("./node_modules/d3-shape/src/descending.js"),o=n("./node_modules/d3-shape/src/identity.js"),a=n("./node_modules/d3-shape/src/math.js");t.default=function(){var e=o.default,t=i.default,n=null,s=Object(r.default)(0),u=Object(r.default)(a.tau),c=Object(r.default)(0);function l(r){var i,o,l,d,p,f=r.length,h=0,m=new Array(f),g=new Array(f),v=+s.apply(this,arguments),_=Math.min(a.tau,Math.max(-a.tau,u.apply(this,arguments)-v)),y=Math.min(Math.abs(_)/f,c.apply(this,arguments)),b=y*(_<0?-1:1);for(i=0;i<f;++i)(p=g[m[i]=i]=+e(r[i],i,r))>0&&(h+=p);for(null!=t?m.sort(function(e,n){return t(g[e],g[n])}):null!=n&&m.sort(function(e,t){return n(r[e],r[t])}),i=0,l=h?(_-f*b)/h:0;i<f;++i,v=d)o=m[i],d=v+((p=g[o])>0?p*l:0)+b,g[o]={data:r[o],index:i,value:p,startAngle:v,endAngle:d,padAngle:y};return g}return l.value=function(t){return arguments.length?(e="function"==typeof t?t:Object(r.default)(+t),l):e},l.sortValues=function(e){return arguments.length?(t=e,n=null,l):t},l.sort=function(e){return arguments.length?(n=e,t=null,l):n},l.startAngle=function(e){return arguments.length?(s="function"==typeof e?e:Object(r.default)(+e),l):s},l.endAngle=function(e){return arguments.length?(u="function"==typeof e?e:Object(r.default)(+e),l):u},l.padAngle=function(e){return arguments.length?(c="function"==typeof e?e:Object(r.default)(+e),l):c},l}},"./node_modules/d3-shape/src/point.js":function(e,t,n){"use strict";function r(e){return e[0]}function i(e){return e[1]}n.r(t),n.d(t,"x",function(){return r}),n.d(t,"y",function(){return i})},"./node_modules/d3-shape/src/pointRadial.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){return[(t=+t)*Math.cos(e-=Math.PI/2),t*Math.sin(e)]}},"./node_modules/d3-shape/src/stack.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/array.js"),i=n("./node_modules/d3-shape/src/constant.js"),o=n("./node_modules/d3-shape/src/offset/none.js"),a=n("./node_modules/d3-shape/src/order/none.js");function s(e,t){return e[t]}t.default=function(){var e=Object(i.default)([]),t=a.default,n=o.default,u=s;function c(r){var i,o,a=e.apply(this,arguments),s=r.length,c=a.length,l=new Array(c);for(i=0;i<c;++i){for(var d,p=a[i],f=l[i]=new Array(s),h=0;h<s;++h)f[h]=d=[0,+u(r[h],p,h,r)],d.data=r[h];f.key=p}for(i=0,o=t(l);i<c;++i)l[o[i]].index=i;return n(l,o),l}return c.keys=function(t){return arguments.length?(e="function"==typeof t?t:Object(i.default)(r.slice.call(t)),c):e},c.value=function(e){return arguments.length?(u="function"==typeof e?e:Object(i.default)(+e),c):u},c.order=function(e){return arguments.length?(t=null==e?a.default:"function"==typeof e?e:Object(i.default)(r.slice.call(e)),c):t},c.offset=function(e){return arguments.length?(n=null==e?o.default:e,c):n},c}},"./node_modules/d3-shape/src/symbol.js":function(e,t,n){"use strict";n.r(t),n.d(t,"symbols",function(){return p});var r=n("./node_modules/d3-path/index.js"),i=n("./node_modules/d3-shape/src/symbol/circle.js"),o=n("./node_modules/d3-shape/src/symbol/cross.js"),a=n("./node_modules/d3-shape/src/symbol/diamond.js"),s=n("./node_modules/d3-shape/src/symbol/star.js"),u=n("./node_modules/d3-shape/src/symbol/square.js"),c=n("./node_modules/d3-shape/src/symbol/triangle.js"),l=n("./node_modules/d3-shape/src/symbol/wye.js"),d=n("./node_modules/d3-shape/src/constant.js"),p=[i.default,o.default,a.default,u.default,s.default,c.default,l.default];t.default=function(){var e=Object(d.default)(i.default),t=Object(d.default)(64),n=null;function o(){var i;if(n||(n=i=Object(r.path)()),e.apply(this,arguments).draw(n,+t.apply(this,arguments)),i)return n=null,i+""||null}return o.type=function(t){return arguments.length?(e="function"==typeof t?t:Object(d.default)(t),o):e},o.size=function(e){return arguments.length?(t="function"==typeof e?e:Object(d.default)(+e),o):t},o.context=function(e){return arguments.length?(n=null==e?null:e,o):n},o}},"./node_modules/d3-shape/src/symbol/circle.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/math.js");t.default={draw:function(e,t){var n=Math.sqrt(t/r.pi);e.moveTo(n,0),e.arc(0,0,n,0,r.tau)}}},"./node_modules/d3-shape/src/symbol/cross.js":function(e,t,n){"use strict";n.r(t),t.default={draw:function(e,t){var n=Math.sqrt(t/5)/2;e.moveTo(-3*n,-n),e.lineTo(-n,-n),e.lineTo(-n,-3*n),e.lineTo(n,-3*n),e.lineTo(n,-n),e.lineTo(3*n,-n),e.lineTo(3*n,n),e.lineTo(n,n),e.lineTo(n,3*n),e.lineTo(-n,3*n),e.lineTo(-n,n),e.lineTo(-3*n,n),e.closePath()}}},"./node_modules/d3-shape/src/symbol/diamond.js":function(e,t,n){"use strict";n.r(t);var r=Math.sqrt(1/3),i=2*r;t.default={draw:function(e,t){var n=Math.sqrt(t/i),o=n*r;e.moveTo(0,-n),e.lineTo(o,0),e.lineTo(0,n),e.lineTo(-o,0),e.closePath()}}},"./node_modules/d3-shape/src/symbol/square.js":function(e,t,n){"use strict";n.r(t),t.default={draw:function(e,t){var n=Math.sqrt(t),r=-n/2;e.rect(r,r,n,n)}}},"./node_modules/d3-shape/src/symbol/star.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-shape/src/math.js"),i=Math.sin(r.pi/10)/Math.sin(7*r.pi/10),o=Math.sin(r.tau/10)*i,a=-Math.cos(r.tau/10)*i;t.default={draw:function(e,t){var n=Math.sqrt(.8908130915292852*t),i=o*n,s=a*n;e.moveTo(0,-n),e.lineTo(i,s);for(var u=1;u<5;++u){var c=r.tau*u/5,l=Math.cos(c),d=Math.sin(c);e.lineTo(d*n,-l*n),e.lineTo(l*i-d*s,d*i+l*s)}e.closePath()}}},"./node_modules/d3-shape/src/symbol/triangle.js":function(e,t,n){"use strict";n.r(t);var r=Math.sqrt(3);t.default={draw:function(e,t){var n=-Math.sqrt(t/(3*r));e.moveTo(0,2*n),e.lineTo(-r*n,-n),e.lineTo(r*n,-n),e.closePath()}}},"./node_modules/d3-shape/src/symbol/wye.js":function(e,t,n){"use strict";n.r(t);var r=Math.sqrt(3)/2,i=1/Math.sqrt(12),o=3*(i/2+1);t.default={draw:function(e,t){var n=Math.sqrt(t/o),a=n/2,s=n*i,u=a,c=n*i+n,l=-u,d=c;e.moveTo(a,s),e.lineTo(u,c),e.lineTo(l,d),e.lineTo(-.5*a-r*s,r*a+-.5*s),e.lineTo(-.5*u-r*c,r*u+-.5*c),e.lineTo(-.5*l-r*d,r*l+-.5*d),e.lineTo(-.5*a+r*s,-.5*s-r*a),e.lineTo(-.5*u+r*c,-.5*c-r*u),e.lineTo(-.5*l+r*d,-.5*d-r*l),e.closePath()}}},"./node_modules/d3-time-format/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-time-format/src/defaultLocale.js");n.d(t,"timeFormatDefaultLocale",function(){return r.default}),n.d(t,"timeFormat",function(){return r.timeFormat}),n.d(t,"timeParse",function(){return r.timeParse}),n.d(t,"utcFormat",function(){return r.utcFormat}),n.d(t,"utcParse",function(){return r.utcParse});var i=n("./node_modules/d3-time-format/src/locale.js");n.d(t,"timeFormatLocale",function(){return i.default});var o=n("./node_modules/d3-time-format/src/isoFormat.js");n.d(t,"isoFormat",function(){return o.default});var a=n("./node_modules/d3-time-format/src/isoParse.js");n.d(t,"isoParse",function(){return a.default})},"./node_modules/d3-time-format/src/defaultLocale.js":function(e,t,n){"use strict";n.r(t),n.d(t,"timeFormat",function(){return i}),n.d(t,"timeParse",function(){return o}),n.d(t,"utcFormat",function(){return a}),n.d(t,"utcParse",function(){return s}),n.d(t,"default",function(){return c});var r,i,o,a,s,u=n("./node_modules/d3-time-format/src/locale.js");function c(e){return r=Object(u.default)(e),i=r.format,o=r.parse,a=r.utcFormat,s=r.utcParse,r}c({dateTime:"%x, %X",date:"%-m/%-d/%Y",time:"%-I:%M:%S %p",periods:["AM","PM"],days:["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"],shortDays:["Sun","Mon","Tue","Wed","Thu","Fri","Sat"],months:["January","February","March","April","May","June","July","August","September","October","November","December"],shortMonths:["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"]})},"./node_modules/d3-time-format/src/isoFormat.js":function(e,t,n){"use strict";n.r(t),n.d(t,"isoSpecifier",function(){return i});var r=n("./node_modules/d3-time-format/src/defaultLocale.js"),i="%Y-%m-%dT%H:%M:%S.%LZ";var o=Date.prototype.toISOString?function(e){return e.toISOString()}:Object(r.utcFormat)(i);t.default=o},"./node_modules/d3-time-format/src/isoParse.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-time-format/src/isoFormat.js"),i=n("./node_modules/d3-time-format/src/defaultLocale.js");var o=+new Date("2000-01-01T00:00:00.000Z")?function(e){var t=new Date(e);return isNaN(t)?null:t}:Object(i.utcParse)(r.isoSpecifier);t.default=o},"./node_modules/d3-time-format/src/locale.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return s});var r=n("./node_modules/d3-time/index.js");function i(e){if(0<=e.y&&e.y<100){var t=new Date(-1,e.m,e.d,e.H,e.M,e.S,e.L);return t.setFullYear(e.y),t}return new Date(e.y,e.m,e.d,e.H,e.M,e.S,e.L)}function o(e){if(0<=e.y&&e.y<100){var t=new Date(Date.UTC(-1,e.m,e.d,e.H,e.M,e.S,e.L));return t.setUTCFullYear(e.y),t}return new Date(Date.UTC(e.y,e.m,e.d,e.H,e.M,e.S,e.L))}function a(e){return{y:e,m:0,d:1,H:0,M:0,S:0,L:0}}function s(e){var t=e.dateTime,n=e.date,s=e.time,c=e.periods,l=e.days,d=e.shortDays,p=e.months,f=e.shortMonths,_e=h(c),ye=m(c),be=h(l),je=m(l),xe=h(d),Oe=m(d),we=h(p),Ce=m(p),Ee=h(f),Se=m(f),ke={a:function(e){return d[e.getDay()]},A:function(e){return l[e.getDay()]},b:function(e){return f[e.getMonth()]},B:function(e){return p[e.getMonth()]},c:null,d:R,e:R,f:B,H:N,I:F,j:L,L:V,m:z,M:U,p:function(e){return c[+(e.getHours()>=12)]},Q:ge,s:ve,S:H,u:q,U:G,V:W,w:Y,W:K,x:null,X:null,y:Q,Y:X,Z:Z,"%":me},Me={a:function(e){return d[e.getUTCDay()]},A:function(e){return l[e.getUTCDay()]},b:function(e){return f[e.getUTCMonth()]},B:function(e){return p[e.getUTCMonth()]},c:null,d:$,e:$,f:re,H:J,I:ee,j:te,L:ne,m:ie,M:oe,p:function(e){return c[+(e.getUTCHours()>=12)]},Q:ge,s:ve,S:ae,u:se,U:ue,V:ce,w:le,W:de,x:null,X:null,y:pe,Y:fe,Z:he,"%":me},Ae={a:function(e,t,n){var r=xe.exec(t.slice(n));return r?(e.w=Oe[r[0].toLowerCase()],n+r[0].length):-1},A:function(e,t,n){var r=be.exec(t.slice(n));return r?(e.w=je[r[0].toLowerCase()],n+r[0].length):-1},b:function(e,t,n){var r=Ee.exec(t.slice(n));return r?(e.m=Se[r[0].toLowerCase()],n+r[0].length):-1},B:function(e,t,n){var r=we.exec(t.slice(n));return r?(e.m=Ce[r[0].toLowerCase()],n+r[0].length):-1},c:function(e,n,r){return Ie(e,t,n,r)},d:C,e:C,f:T,H:S,I:S,j:E,L:A,m:w,M:k,p:function(e,t,n){var r=_e.exec(t.slice(n));return r?(e.p=ye[r[0].toLowerCase()],n+r[0].length):-1},Q:I,s:P,S:M,u:v,U:_,V:y,w:g,W:b,x:function(e,t,r){return Ie(e,n,t,r)},X:function(e,t,n){return Ie(e,s,t,n)},y:x,Y:j,Z:O,"%":D};function Te(e,t){return function(n){var r,i,o,a=[],s=-1,c=0,l=e.length;for(n instanceof Date||(n=new Date(+n));++s<l;)37===e.charCodeAt(s)&&(a.push(e.slice(c,s)),null!=(i=u[r=e.charAt(++s)])?r=e.charAt(++s):i="e"===r?" ":"0",(o=t[r])&&(r=o(n,i)),a.push(r),c=s+1);return a.push(e.slice(c,s)),a.join("")}}function De(e,t){return function(n){var i,s,u=a(1900);if(Ie(u,e,n+="",0)!=n.length)return null;if("Q"in u)return new Date(u.Q);if("p"in u&&(u.H=u.H%12+12*u.p),"V"in u){if(u.V<1||u.V>53)return null;"w"in u||(u.w=1),"Z"in u?(i=(s=(i=o(a(u.y))).getUTCDay())>4||0===s?r.utcMonday.ceil(i):Object(r.utcMonday)(i),i=r.utcDay.offset(i,7*(u.V-1)),u.y=i.getUTCFullYear(),u.m=i.getUTCMonth(),u.d=i.getUTCDate()+(u.w+6)%7):(i=(s=(i=t(a(u.y))).getDay())>4||0===s?r.timeMonday.ceil(i):Object(r.timeMonday)(i),i=r.timeDay.offset(i,7*(u.V-1)),u.y=i.getFullYear(),u.m=i.getMonth(),u.d=i.getDate()+(u.w+6)%7)}else("W"in u||"U"in u)&&("w"in u||(u.w="u"in u?u.u%7:"W"in u?1:0),s="Z"in u?o(a(u.y)).getUTCDay():t(a(u.y)).getDay(),u.m=0,u.d="W"in u?(u.w+6)%7+7*u.W-(s+5)%7:u.w+7*u.U-(s+6)%7);return"Z"in u?(u.H+=u.Z/100|0,u.M+=u.Z%100,o(u)):t(u)}}function Ie(e,t,n,r){for(var i,o,a=0,s=t.length,c=n.length;a<s;){if(r>=c)return-1;if(37===(i=t.charCodeAt(a++))){if(i=t.charAt(a++),!(o=Ae[i in u?t.charAt(a++):i])||(r=o(e,n,r))<0)return-1}else if(i!=n.charCodeAt(r++))return-1}return r}return ke.x=Te(n,ke),ke.X=Te(s,ke),ke.c=Te(t,ke),Me.x=Te(n,Me),Me.X=Te(s,Me),Me.c=Te(t,Me),{format:function(e){var t=Te(e+="",ke);return t.toString=function(){return e},t},parse:function(e){var t=De(e+="",i);return t.toString=function(){return e},t},utcFormat:function(e){var t=Te(e+="",Me);return t.toString=function(){return e},t},utcParse:function(e){var t=De(e,o);return t.toString=function(){return e},t}}}var u={"-":"",_:" ",0:"0"},c=/^\s*\d+/,l=/^%/,d=/[\\^$*+?|[\]().{}]/g;function p(e,t,n){var r=e<0?"-":"",i=(r?-e:e)+"",o=i.length;return r+(o<n?new Array(n-o+1).join(t)+i:i)}function f(e){return e.replace(d,"\\$&")}function h(e){return new RegExp("^(?:"+e.map(f).join("|")+")","i")}function m(e){for(var t={},n=-1,r=e.length;++n<r;)t[e[n].toLowerCase()]=n;return t}function g(e,t,n){var r=c.exec(t.slice(n,n+1));return r?(e.w=+r[0],n+r[0].length):-1}function v(e,t,n){var r=c.exec(t.slice(n,n+1));return r?(e.u=+r[0],n+r[0].length):-1}function _(e,t,n){var r=c.exec(t.slice(n,n+2));return r?(e.U=+r[0],n+r[0].length):-1}function y(e,t,n){var r=c.exec(t.slice(n,n+2));return r?(e.V=+r[0],n+r[0].length):-1}function b(e,t,n){var r=c.exec(t.slice(n,n+2));return r?(e.W=+r[0],n+r[0].length):-1}function j(e,t,n){var r=c.exec(t.slice(n,n+4));return r?(e.y=+r[0],n+r[0].length):-1}function x(e,t,n){var r=c.exec(t.slice(n,n+2));return r?(e.y=+r[0]+(+r[0]>68?1900:2e3),n+r[0].length):-1}function O(e,t,n){var r=/^(Z)|([+-]\d\d)(?::?(\d\d))?/.exec(t.slice(n,n+6));return r?(e.Z=r[1]?0:-(r[2]+(r[3]||"00")),n+r[0].length):-1}function w(e,t,n){var r=c.exec(t.slice(n,n+2));return r?(e.m=r[0]-1,n+r[0].length):-1}function C(e,t,n){var r=c.exec(t.slice(n,n+2));return r?(e.d=+r[0],n+r[0].length):-1}function E(e,t,n){var r=c.exec(t.slice(n,n+3));return r?(e.m=0,e.d=+r[0],n+r[0].length):-1}function S(e,t,n){var r=c.exec(t.slice(n,n+2));return r?(e.H=+r[0],n+r[0].length):-1}function k(e,t,n){var r=c.exec(t.slice(n,n+2));return r?(e.M=+r[0],n+r[0].length):-1}function M(e,t,n){var r=c.exec(t.slice(n,n+2));return r?(e.S=+r[0],n+r[0].length):-1}function A(e,t,n){var r=c.exec(t.slice(n,n+3));return r?(e.L=+r[0],n+r[0].length):-1}function T(e,t,n){var r=c.exec(t.slice(n,n+6));return r?(e.L=Math.floor(r[0]/1e3),n+r[0].length):-1}function D(e,t,n){var r=l.exec(t.slice(n,n+1));return r?n+r[0].length:-1}function I(e,t,n){var r=c.exec(t.slice(n));return r?(e.Q=+r[0],n+r[0].length):-1}function P(e,t,n){var r=c.exec(t.slice(n));return r?(e.Q=1e3*+r[0],n+r[0].length):-1}function R(e,t){return p(e.getDate(),t,2)}function N(e,t){return p(e.getHours(),t,2)}function F(e,t){return p(e.getHours()%12||12,t,2)}function L(e,t){return p(1+r.timeDay.count(Object(r.timeYear)(e),e),t,3)}function V(e,t){return p(e.getMilliseconds(),t,3)}function B(e,t){return V(e,t)+"000"}function z(e,t){return p(e.getMonth()+1,t,2)}function U(e,t){return p(e.getMinutes(),t,2)}function H(e,t){return p(e.getSeconds(),t,2)}function q(e){var t=e.getDay();return 0===t?7:t}function G(e,t){return p(r.timeSunday.count(Object(r.timeYear)(e),e),t,2)}function W(e,t){var n=e.getDay();return e=n>=4||0===n?Object(r.timeThursday)(e):r.timeThursday.ceil(e),p(r.timeThursday.count(Object(r.timeYear)(e),e)+(4===Object(r.timeYear)(e).getDay()),t,2)}function Y(e){return e.getDay()}function K(e,t){return p(r.timeMonday.count(Object(r.timeYear)(e),e),t,2)}function Q(e,t){return p(e.getFullYear()%100,t,2)}function X(e,t){return p(e.getFullYear()%1e4,t,4)}function Z(e){var t=e.getTimezoneOffset();return(t>0?"-":(t*=-1,"+"))+p(t/60|0,"0",2)+p(t%60,"0",2)}function $(e,t){return p(e.getUTCDate(),t,2)}function J(e,t){return p(e.getUTCHours(),t,2)}function ee(e,t){return p(e.getUTCHours()%12||12,t,2)}function te(e,t){return p(1+r.utcDay.count(Object(r.utcYear)(e),e),t,3)}function ne(e,t){return p(e.getUTCMilliseconds(),t,3)}function re(e,t){return ne(e,t)+"000"}function ie(e,t){return p(e.getUTCMonth()+1,t,2)}function oe(e,t){return p(e.getUTCMinutes(),t,2)}function ae(e,t){return p(e.getUTCSeconds(),t,2)}function se(e){var t=e.getUTCDay();return 0===t?7:t}function ue(e,t){return p(r.utcSunday.count(Object(r.utcYear)(e),e),t,2)}function ce(e,t){var n=e.getUTCDay();return e=n>=4||0===n?Object(r.utcThursday)(e):r.utcThursday.ceil(e),p(r.utcThursday.count(Object(r.utcYear)(e),e)+(4===Object(r.utcYear)(e).getUTCDay()),t,2)}function le(e){return e.getUTCDay()}function de(e,t){return p(r.utcMonday.count(Object(r.utcYear)(e),e),t,2)}function pe(e,t){return p(e.getUTCFullYear()%100,t,2)}function fe(e,t){return p(e.getUTCFullYear()%1e4,t,4)}function he(){return"+0000"}function me(){return"%"}function ge(e){return+e}function ve(e){return Math.floor(+e/1e3)}},"./node_modules/d3-time/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-time/src/interval.js");n.d(t,"timeInterval",function(){return r.default});var i=n("./node_modules/d3-time/src/millisecond.js");n.d(t,"timeMillisecond",function(){return i.default}),n.d(t,"timeMilliseconds",function(){return i.milliseconds}),n.d(t,"utcMillisecond",function(){return i.default}),n.d(t,"utcMilliseconds",function(){return i.milliseconds});var o=n("./node_modules/d3-time/src/second.js");n.d(t,"timeSecond",function(){return o.default}),n.d(t,"timeSeconds",function(){return o.seconds}),n.d(t,"utcSecond",function(){return o.default}),n.d(t,"utcSeconds",function(){return o.seconds});var a=n("./node_modules/d3-time/src/minute.js");n.d(t,"timeMinute",function(){return a.default}),n.d(t,"timeMinutes",function(){return a.minutes});var s=n("./node_modules/d3-time/src/hour.js");n.d(t,"timeHour",function(){return s.default}),n.d(t,"timeHours",function(){return s.hours});var u=n("./node_modules/d3-time/src/day.js");n.d(t,"timeDay",function(){return u.default}),n.d(t,"timeDays",function(){return u.days});var c=n("./node_modules/d3-time/src/week.js");n.d(t,"timeWeek",function(){return c.sunday}),n.d(t,"timeWeeks",function(){return c.sundays}),n.d(t,"timeSunday",function(){return c.sunday}),n.d(t,"timeSundays",function(){return c.sundays}),n.d(t,"timeMonday",function(){return c.monday}),n.d(t,"timeMondays",function(){return c.mondays}),n.d(t,"timeTuesday",function(){return c.tuesday}),n.d(t,"timeTuesdays",function(){return c.tuesdays}),n.d(t,"timeWednesday",function(){return c.wednesday}),n.d(t,"timeWednesdays",function(){return c.wednesdays}),n.d(t,"timeThursday",function(){return c.thursday}),n.d(t,"timeThursdays",function(){return c.thursdays}),n.d(t,"timeFriday",function(){return c.friday}),n.d(t,"timeFridays",function(){return c.fridays}),n.d(t,"timeSaturday",function(){return c.saturday}),n.d(t,"timeSaturdays",function(){return c.saturdays});var l=n("./node_modules/d3-time/src/month.js");n.d(t,"timeMonth",function(){return l.default}),n.d(t,"timeMonths",function(){return l.months});var d=n("./node_modules/d3-time/src/year.js");n.d(t,"timeYear",function(){return d.default}),n.d(t,"timeYears",function(){return d.years});var p=n("./node_modules/d3-time/src/utcMinute.js");n.d(t,"utcMinute",function(){return p.default}),n.d(t,"utcMinutes",function(){return p.utcMinutes});var f=n("./node_modules/d3-time/src/utcHour.js");n.d(t,"utcHour",function(){return f.default}),n.d(t,"utcHours",function(){return f.utcHours});var h=n("./node_modules/d3-time/src/utcDay.js");n.d(t,"utcDay",function(){return h.default}),n.d(t,"utcDays",function(){return h.utcDays});var m=n("./node_modules/d3-time/src/utcWeek.js");n.d(t,"utcWeek",function(){return m.utcSunday}),n.d(t,"utcWeeks",function(){return m.utcSundays}),n.d(t,"utcSunday",function(){return m.utcSunday}),n.d(t,"utcSundays",function(){return m.utcSundays}),n.d(t,"utcMonday",function(){return m.utcMonday}),n.d(t,"utcMondays",function(){return m.utcMondays}),n.d(t,"utcTuesday",function(){return m.utcTuesday}),n.d(t,"utcTuesdays",function(){return m.utcTuesdays}),n.d(t,"utcWednesday",function(){return m.utcWednesday}),n.d(t,"utcWednesdays",function(){return m.utcWednesdays}),n.d(t,"utcThursday",function(){return m.utcThursday}),n.d(t,"utcThursdays",function(){return m.utcThursdays}),n.d(t,"utcFriday",function(){return m.utcFriday}),n.d(t,"utcFridays",function(){return m.utcFridays}),n.d(t,"utcSaturday",function(){return m.utcSaturday}),n.d(t,"utcSaturdays",function(){return m.utcSaturdays});var g=n("./node_modules/d3-time/src/utcMonth.js");n.d(t,"utcMonth",function(){return g.default}),n.d(t,"utcMonths",function(){return g.utcMonths});var v=n("./node_modules/d3-time/src/utcYear.js");n.d(t,"utcYear",function(){return v.default}),n.d(t,"utcYears",function(){return v.utcYears})},"./node_modules/d3-time/src/day.js":function(e,t,n){"use strict";n.r(t),n.d(t,"days",function(){return a});var r=n("./node_modules/d3-time/src/interval.js"),i=n("./node_modules/d3-time/src/duration.js"),o=Object(r.default)(function(e){e.setHours(0,0,0,0)},function(e,t){e.setDate(e.getDate()+t)},function(e,t){return(t-e-(t.getTimezoneOffset()-e.getTimezoneOffset())*i.durationMinute)/i.durationDay},function(e){return e.getDate()-1});t.default=o;var a=o.range},"./node_modules/d3-time/src/duration.js":function(e,t,n){"use strict";n.r(t),n.d(t,"durationSecond",function(){return r}),n.d(t,"durationMinute",function(){return i}),n.d(t,"durationHour",function(){return o}),n.d(t,"durationDay",function(){return a}),n.d(t,"durationWeek",function(){return s});var r=1e3,i=6e4,o=36e5,a=864e5,s=6048e5},"./node_modules/d3-time/src/hour.js":function(e,t,n){"use strict";n.r(t),n.d(t,"hours",function(){return a});var r=n("./node_modules/d3-time/src/interval.js"),i=n("./node_modules/d3-time/src/duration.js"),o=Object(r.default)(function(e){var t=e.getTimezoneOffset()*i.durationMinute%i.durationHour;t<0&&(t+=i.durationHour),e.setTime(Math.floor((+e-t)/i.durationHour)*i.durationHour+t)},function(e,t){e.setTime(+e+t*i.durationHour)},function(e,t){return(t-e)/i.durationHour},function(e){return e.getHours()});t.default=o;var a=o.range},"./node_modules/d3-time/src/interval.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=new Date,i=new Date;function o(e,t,n,a){function s(t){return e(t=new Date(+t)),t}return s.floor=s,s.ceil=function(n){return e(n=new Date(n-1)),t(n,1),e(n),n},s.round=function(e){var t=s(e),n=s.ceil(e);return e-t<n-e?t:n},s.offset=function(e,n){return t(e=new Date(+e),null==n?1:Math.floor(n)),e},s.range=function(n,r,i){var o,a=[];if(n=s.ceil(n),i=null==i?1:Math.floor(i),!(n<r&&i>0))return a;do{a.push(o=new Date(+n)),t(n,i),e(n)}while(o<n&&n<r);return a},s.filter=function(n){return o(function(t){if(t>=t)for(;e(t),!n(t);)t.setTime(t-1)},function(e,r){if(e>=e)if(r<0)for(;++r<=0;)for(;t(e,-1),!n(e););else for(;--r>=0;)for(;t(e,1),!n(e););})},n&&(s.count=function(t,o){return r.setTime(+t),i.setTime(+o),e(r),e(i),Math.floor(n(r,i))},s.every=function(e){return e=Math.floor(e),isFinite(e)&&e>0?e>1?s.filter(a?function(t){return a(t)%e==0}:function(t){return s.count(0,t)%e==0}):s:null}),s}},"./node_modules/d3-time/src/millisecond.js":function(e,t,n){"use strict";n.r(t),n.d(t,"milliseconds",function(){return o});var r=n("./node_modules/d3-time/src/interval.js"),i=Object(r.default)(function(){},function(e,t){e.setTime(+e+t)},function(e,t){return t-e});i.every=function(e){return e=Math.floor(e),isFinite(e)&&e>0?e>1?Object(r.default)(function(t){t.setTime(Math.floor(t/e)*e)},function(t,n){t.setTime(+t+n*e)},function(t,n){return(n-t)/e}):i:null},t.default=i;var o=i.range},"./node_modules/d3-time/src/minute.js":function(e,t,n){"use strict";n.r(t),n.d(t,"minutes",function(){return a});var r=n("./node_modules/d3-time/src/interval.js"),i=n("./node_modules/d3-time/src/duration.js"),o=Object(r.default)(function(e){e.setTime(Math.floor(e/i.durationMinute)*i.durationMinute)},function(e,t){e.setTime(+e+t*i.durationMinute)},function(e,t){return(t-e)/i.durationMinute},function(e){return e.getMinutes()});t.default=o;var a=o.range},"./node_modules/d3-time/src/month.js":function(e,t,n){"use strict";n.r(t),n.d(t,"months",function(){return o});var r=n("./node_modules/d3-time/src/interval.js"),i=Object(r.default)(function(e){e.setDate(1),e.setHours(0,0,0,0)},function(e,t){e.setMonth(e.getMonth()+t)},function(e,t){return t.getMonth()-e.getMonth()+12*(t.getFullYear()-e.getFullYear())},function(e){return e.getMonth()});t.default=i;var o=i.range},"./node_modules/d3-time/src/second.js":function(e,t,n){"use strict";n.r(t),n.d(t,"seconds",function(){return a});var r=n("./node_modules/d3-time/src/interval.js"),i=n("./node_modules/d3-time/src/duration.js"),o=Object(r.default)(function(e){e.setTime(Math.floor(e/i.durationSecond)*i.durationSecond)},function(e,t){e.setTime(+e+t*i.durationSecond)},function(e,t){return(t-e)/i.durationSecond},function(e){return e.getUTCSeconds()});t.default=o;var a=o.range},"./node_modules/d3-time/src/utcDay.js":function(e,t,n){"use strict";n.r(t),n.d(t,"utcDays",function(){return a});var r=n("./node_modules/d3-time/src/interval.js"),i=n("./node_modules/d3-time/src/duration.js"),o=Object(r.default)(function(e){e.setUTCHours(0,0,0,0)},function(e,t){e.setUTCDate(e.getUTCDate()+t)},function(e,t){return(t-e)/i.durationDay},function(e){return e.getUTCDate()-1});t.default=o;var a=o.range},"./node_modules/d3-time/src/utcHour.js":function(e,t,n){"use strict";n.r(t),n.d(t,"utcHours",function(){return a});var r=n("./node_modules/d3-time/src/interval.js"),i=n("./node_modules/d3-time/src/duration.js"),o=Object(r.default)(function(e){e.setUTCMinutes(0,0,0)},function(e,t){e.setTime(+e+t*i.durationHour)},function(e,t){return(t-e)/i.durationHour},function(e){return e.getUTCHours()});t.default=o;var a=o.range},"./node_modules/d3-time/src/utcMinute.js":function(e,t,n){"use strict";n.r(t),n.d(t,"utcMinutes",function(){return a});var r=n("./node_modules/d3-time/src/interval.js"),i=n("./node_modules/d3-time/src/duration.js"),o=Object(r.default)(function(e){e.setUTCSeconds(0,0)},function(e,t){e.setTime(+e+t*i.durationMinute)},function(e,t){return(t-e)/i.durationMinute},function(e){return e.getUTCMinutes()});t.default=o;var a=o.range},"./node_modules/d3-time/src/utcMonth.js":function(e,t,n){"use strict";n.r(t),n.d(t,"utcMonths",function(){return o});var r=n("./node_modules/d3-time/src/interval.js"),i=Object(r.default)(function(e){e.setUTCDate(1),e.setUTCHours(0,0,0,0)},function(e,t){e.setUTCMonth(e.getUTCMonth()+t)},function(e,t){return t.getUTCMonth()-e.getUTCMonth()+12*(t.getUTCFullYear()-e.getUTCFullYear())},function(e){return e.getUTCMonth()});t.default=i;var o=i.range},"./node_modules/d3-time/src/utcWeek.js":function(e,t,n){"use strict";n.r(t),n.d(t,"utcSunday",function(){return a}),n.d(t,"utcMonday",function(){return s}),n.d(t,"utcTuesday",function(){return u}),n.d(t,"utcWednesday",function(){return c}),n.d(t,"utcThursday",function(){return l}),n.d(t,"utcFriday",function(){return d}),n.d(t,"utcSaturday",function(){return p}),n.d(t,"utcSundays",function(){return f}),n.d(t,"utcMondays",function(){return h}),n.d(t,"utcTuesdays",function(){return m}),n.d(t,"utcWednesdays",function(){return g}),n.d(t,"utcThursdays",function(){return v}),n.d(t,"utcFridays",function(){return _}),n.d(t,"utcSaturdays",function(){return y});var r=n("./node_modules/d3-time/src/interval.js"),i=n("./node_modules/d3-time/src/duration.js");function o(e){return Object(r.default)(function(t){t.setUTCDate(t.getUTCDate()-(t.getUTCDay()+7-e)%7),t.setUTCHours(0,0,0,0)},function(e,t){e.setUTCDate(e.getUTCDate()+7*t)},function(e,t){return(t-e)/i.durationWeek})}var a=o(0),s=o(1),u=o(2),c=o(3),l=o(4),d=o(5),p=o(6),f=a.range,h=s.range,m=u.range,g=c.range,v=l.range,_=d.range,y=p.range},"./node_modules/d3-time/src/utcYear.js":function(e,t,n){"use strict";n.r(t),n.d(t,"utcYears",function(){return o});var r=n("./node_modules/d3-time/src/interval.js"),i=Object(r.default)(function(e){e.setUTCMonth(0,1),e.setUTCHours(0,0,0,0)},function(e,t){e.setUTCFullYear(e.getUTCFullYear()+t)},function(e,t){return t.getUTCFullYear()-e.getUTCFullYear()},function(e){return e.getUTCFullYear()});i.every=function(e){return isFinite(e=Math.floor(e))&&e>0?Object(r.default)(function(t){t.setUTCFullYear(Math.floor(t.getUTCFullYear()/e)*e),t.setUTCMonth(0,1),t.setUTCHours(0,0,0,0)},function(t,n){t.setUTCFullYear(t.getUTCFullYear()+n*e)}):null},t.default=i;var o=i.range},"./node_modules/d3-time/src/week.js":function(e,t,n){"use strict";n.r(t),n.d(t,"sunday",function(){return a}),n.d(t,"monday",function(){return s}),n.d(t,"tuesday",function(){return u}),n.d(t,"wednesday",function(){return c}),n.d(t,"thursday",function(){return l}),n.d(t,"friday",function(){return d}),n.d(t,"saturday",function(){return p}),n.d(t,"sundays",function(){return f}),n.d(t,"mondays",function(){return h}),n.d(t,"tuesdays",function(){return m}),n.d(t,"wednesdays",function(){return g}),n.d(t,"thursdays",function(){return v}),n.d(t,"fridays",function(){return _}),n.d(t,"saturdays",function(){return y});var r=n("./node_modules/d3-time/src/interval.js"),i=n("./node_modules/d3-time/src/duration.js");function o(e){return Object(r.default)(function(t){t.setDate(t.getDate()-(t.getDay()+7-e)%7),t.setHours(0,0,0,0)},function(e,t){e.setDate(e.getDate()+7*t)},function(e,t){return(t-e-(t.getTimezoneOffset()-e.getTimezoneOffset())*i.durationMinute)/i.durationWeek})}var a=o(0),s=o(1),u=o(2),c=o(3),l=o(4),d=o(5),p=o(6),f=a.range,h=s.range,m=u.range,g=c.range,v=l.range,_=d.range,y=p.range},"./node_modules/d3-time/src/year.js":function(e,t,n){"use strict";n.r(t),n.d(t,"years",function(){return o});var r=n("./node_modules/d3-time/src/interval.js"),i=Object(r.default)(function(e){e.setMonth(0,1),e.setHours(0,0,0,0)},function(e,t){e.setFullYear(e.getFullYear()+t)},function(e,t){return t.getFullYear()-e.getFullYear()},function(e){return e.getFullYear()});i.every=function(e){return isFinite(e=Math.floor(e))&&e>0?Object(r.default)(function(t){t.setFullYear(Math.floor(t.getFullYear()/e)*e),t.setMonth(0,1),t.setHours(0,0,0,0)},function(t,n){t.setFullYear(t.getFullYear()+n*e)}):null},t.default=i;var o=i.range},"./node_modules/d3-timer/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-timer/src/timer.js");n.d(t,"now",function(){return r.now}),n.d(t,"timer",function(){return r.timer}),n.d(t,"timerFlush",function(){return r.timerFlush});var i=n("./node_modules/d3-timer/src/timeout.js");n.d(t,"timeout",function(){return i.default});var o=n("./node_modules/d3-timer/src/interval.js");n.d(t,"interval",function(){return o.default})},"./node_modules/d3-timer/src/interval.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-timer/src/timer.js");t.default=function(e,t,n){var i=new r.Timer,o=t;return null==t?(i.restart(e,t,n),i):(t=+t,n=null==n?Object(r.now)():+n,i.restart(function r(a){a+=o,i.restart(r,o+=t,n),e(a)},t,n),i)}},"./node_modules/d3-timer/src/timeout.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-timer/src/timer.js");t.default=function(e,t,n){var i=new r.Timer;return t=null==t?0:+t,i.restart(function(n){i.stop(),e(n+t)},t,n),i}},"./node_modules/d3-timer/src/timer.js":function(e,t,n){"use strict";n.r(t),n.d(t,"now",function(){return h}),n.d(t,"Timer",function(){return g}),n.d(t,"timer",function(){return v}),n.d(t,"timerFlush",function(){return _});var r,i,o=0,a=0,s=0,u=1e3,c=0,l=0,d=0,p="object"==typeof performance&&performance.now?performance:Date,f="object"==typeof window&&window.requestAnimationFrame?window.requestAnimationFrame.bind(window):function(e){setTimeout(e,17)};function h(){return l||(f(m),l=p.now()+d)}function m(){l=0}function g(){this._call=this._time=this._next=null}function v(e,t,n){var r=new g;return r.restart(e,t,n),r}function _(){h(),++o;for(var e,t=r;t;)(e=l-t._time)>=0&&t._call.call(null,e),t=t._next;--o}function y(){l=(c=p.now())+d,o=a=0;try{_()}finally{o=0,function(){var e,t,n=r,o=1/0;for(;n;)n._call?(o>n._time&&(o=n._time),e=n,n=n._next):(t=n._next,n._next=null,n=e?e._next=t:r=t);i=e,j(o)}(),l=0}}function b(){var e=p.now(),t=e-c;t>u&&(d-=t,c=e)}function j(e){o||(a&&(a=clearTimeout(a)),e-l>24?(e<1/0&&(a=setTimeout(y,e-p.now()-d)),s&&(s=clearInterval(s))):(s||(c=p.now(),s=setInterval(b,u)),o=1,f(y)))}g.prototype=v.prototype={constructor:g,restart:function(e,t,n){if("function"!=typeof e)throw new TypeError("callback is not a function");n=(null==n?h():+n)+(null==t?0:+t),this._next||i===this||(i?i._next=this:r=this,i=this),this._call=e,this._time=n,j()},stop:function(){this._call&&(this._call=null,this._time=1/0,j())}}},"./node_modules/d3-voronoi/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-voronoi/src/voronoi.js");n.d(t,"voronoi",function(){return r.default})},"./node_modules/d3-voronoi/src/Beach.js":function(e,t,n){"use strict";n.r(t),n.d(t,"removeBeach",function(){return d}),n.d(t,"addBeach",function(){return p});var r=n("./node_modules/d3-voronoi/src/RedBlackTree.js"),i=n("./node_modules/d3-voronoi/src/Cell.js"),o=n("./node_modules/d3-voronoi/src/Circle.js"),a=n("./node_modules/d3-voronoi/src/Edge.js"),s=n("./node_modules/d3-voronoi/src/Diagram.js"),u=[];function c(e){var t=u.pop()||new function(){Object(r.RedBlackNode)(this),this.edge=this.site=this.circle=null};return t.site=e,t}function l(e){Object(o.detachCircle)(e),s.beaches.remove(e),u.push(e),Object(r.RedBlackNode)(e)}function d(e){var t=e.circle,n=t.x,r=t.cy,i=[n,r],u=e.P,c=e.N,d=[e];l(e);for(var p=u;p.circle&&Math.abs(n-p.circle.x)<s.epsilon&&Math.abs(r-p.circle.cy)<s.epsilon;)u=p.P,d.unshift(p),l(p),p=u;d.unshift(p),Object(o.detachCircle)(p);for(var f=c;f.circle&&Math.abs(n-f.circle.x)<s.epsilon&&Math.abs(r-f.circle.cy)<s.epsilon;)c=f.N,d.push(f),l(f),f=c;d.push(f),Object(o.detachCircle)(f);var h,m=d.length;for(h=1;h<m;++h)f=d[h],p=d[h-1],Object(a.setEdgeEnd)(f.edge,p.site,f.site,i);p=d[0],(f=d[m-1]).edge=Object(a.createEdge)(p.site,f.site,null,i),Object(o.attachCircle)(p),Object(o.attachCircle)(f)}function p(e){for(var t,n,r,u,l=e[0],d=e[1],p=s.beaches._;p;)if((r=f(p,d)-l)>s.epsilon)p=p.L;else{if(!((u=l-h(p,d))>s.epsilon)){r>-s.epsilon?(t=p.P,n=p):u>-s.epsilon?(t=p,n=p.N):t=n=p;break}if(!p.R){t=p;break}p=p.R}Object(i.createCell)(e);var m=c(e);if(s.beaches.insert(t,m),t||n){if(t===n)return Object(o.detachCircle)(t),n=c(t.site),s.beaches.insert(m,n),m.edge=n.edge=Object(a.createEdge)(t.site,m.site),Object(o.attachCircle)(t),void Object(o.attachCircle)(n);if(n){Object(o.detachCircle)(t),Object(o.detachCircle)(n);var g=t.site,v=g[0],_=g[1],y=e[0]-v,b=e[1]-_,j=n.site,x=j[0]-v,O=j[1]-_,w=2*(y*O-b*x),C=y*y+b*b,E=x*x+O*O,S=[(O*C-b*E)/w+v,(y*E-x*C)/w+_];Object(a.setEdgeEnd)(n.edge,g,j,S),m.edge=Object(a.createEdge)(g,e,null,S),n.edge=Object(a.createEdge)(e,j,null,S),Object(o.attachCircle)(t),Object(o.attachCircle)(n)}else m.edge=Object(a.createEdge)(t.site,m.site)}}function f(e,t){var n=e.site,r=n[0],i=n[1],o=i-t;if(!o)return r;var a=e.P;if(!a)return-1/0;var s=(n=a.site)[0],u=n[1],c=u-t;if(!c)return s;var l=s-r,d=1/o-1/c,p=l/c;return d?(-p+Math.sqrt(p*p-2*d*(l*l/(-2*c)-u+c/2+i-o/2)))/d+r:(r+s)/2}function h(e,t){var n=e.N;if(n)return f(n,t);var r=e.site;return r[1]===t?r[0]:1/0}},"./node_modules/d3-voronoi/src/Cell.js":function(e,t,n){"use strict";n.r(t),n.d(t,"createCell",function(){return o}),n.d(t,"cellHalfedgeStart",function(){return a}),n.d(t,"cellHalfedgeEnd",function(){return s}),n.d(t,"sortCellHalfedges",function(){return u}),n.d(t,"clipCells",function(){return c});var r=n("./node_modules/d3-voronoi/src/Edge.js"),i=n("./node_modules/d3-voronoi/src/Diagram.js");function o(e){return i.cells[e.index]={site:e,halfedges:[]}}function a(e,t){return t[+(t.left!==e.site)]}function s(e,t){return t[+(t.left===e.site)]}function u(){for(var e,t,n,r,o=0,a=i.cells.length;o<a;++o)if((e=i.cells[o])&&(r=(t=e.halfedges).length)){var s=new Array(r),u=new Array(r);for(n=0;n<r;++n)s[n]=n,u[n]=(c=e,l=i.edges[t[n]],void 0,p=void 0,f=void 0,d=c.site,p=l.left,f=l.right,d===f&&(f=p,p=d),f?Math.atan2(f[1]-p[1],f[0]-p[0]):(d===p?(p=l[1],f=l[0]):(p=l[0],f=l[1]),Math.atan2(p[0]-f[0],f[1]-p[1])));for(s.sort(function(e,t){return u[t]-u[e]}),n=0;n<r;++n)u[n]=t[s[n]];for(n=0;n<r;++n)t[n]=u[n]}var c,l,d,p,f}function c(e,t,n,o){var u,c,l,d,p,f,h,m,g,v,_,y,b=i.cells.length,j=!0;for(u=0;u<b;++u)if(c=i.cells[u]){for(l=c.site,d=(p=c.halfedges).length;d--;)i.edges[p[d]]||p.splice(d,1);for(d=0,f=p.length;d<f;)_=(v=s(c,i.edges[p[d]]))[0],y=v[1],m=(h=a(c,i.edges[p[++d%f]]))[0],g=h[1],(Math.abs(_-m)>i.epsilon||Math.abs(y-g)>i.epsilon)&&(p.splice(d,0,i.edges.push(Object(r.createBorderEdge)(l,v,Math.abs(_-e)<i.epsilon&&o-y>i.epsilon?[e,Math.abs(m-e)<i.epsilon?g:o]:Math.abs(y-o)<i.epsilon&&n-_>i.epsilon?[Math.abs(g-o)<i.epsilon?m:n,o]:Math.abs(_-n)<i.epsilon&&y-t>i.epsilon?[n,Math.abs(m-n)<i.epsilon?g:t]:Math.abs(y-t)<i.epsilon&&_-e>i.epsilon?[Math.abs(g-t)<i.epsilon?m:e,t]:null))-1),++f);f&&(j=!1)}if(j){var x,O,w,C=1/0;for(u=0,j=null;u<b;++u)(c=i.cells[u])&&(w=(x=(l=c.site)[0]-e)*x+(O=l[1]-t)*O)<C&&(C=w,j=c);if(j){var E=[e,t],S=[e,o],k=[n,o],M=[n,t];j.halfedges.push(i.edges.push(Object(r.createBorderEdge)(l=j.site,E,S))-1,i.edges.push(Object(r.createBorderEdge)(l,S,k))-1,i.edges.push(Object(r.createBorderEdge)(l,k,M))-1,i.edges.push(Object(r.createBorderEdge)(l,M,E))-1)}}for(u=0;u<b;++u)(c=i.cells[u])&&(c.halfedges.length||delete i.cells[u])}},"./node_modules/d3-voronoi/src/Circle.js":function(e,t,n){"use strict";n.r(t),n.d(t,"firstCircle",function(){return r}),n.d(t,"attachCircle",function(){return s}),n.d(t,"detachCircle",function(){return u});var r,i=n("./node_modules/d3-voronoi/src/RedBlackTree.js"),o=n("./node_modules/d3-voronoi/src/Diagram.js"),a=[];function s(e){var t=e.P,n=e.N;if(t&&n){var s=t.site,u=e.site,c=n.site;if(s!==c){var l=u[0],d=u[1],p=s[0]-l,f=s[1]-d,h=c[0]-l,m=c[1]-d,g=2*(p*m-f*h);if(!(g>=-o.epsilon2)){var v=p*p+f*f,_=h*h+m*m,y=(m*v-f*_)/g,b=(p*_-h*v)/g,j=a.pop()||new function(){Object(i.RedBlackNode)(this),this.x=this.y=this.arc=this.site=this.cy=null};j.arc=e,j.site=u,j.x=y+l,j.y=(j.cy=b+d)+Math.sqrt(y*y+b*b),e.circle=j;for(var x=null,O=o.circles._;O;)if(j.y<O.y||j.y===O.y&&j.x<=O.x){if(!O.L){x=O.P;break}O=O.L}else{if(!O.R){x=O;break}O=O.R}o.circles.insert(x,j),x||(r=j)}}}}function u(e){var t=e.circle;t&&(t.P||(r=t.N),o.circles.remove(t),a.push(t),Object(i.RedBlackNode)(t),e.circle=null)}},"./node_modules/d3-voronoi/src/Diagram.js":function(e,t,n){"use strict";n.r(t),n.d(t,"epsilon",function(){return p}),n.d(t,"epsilon2",function(){return f}),n.d(t,"beaches",function(){return r}),n.d(t,"cells",function(){return i}),n.d(t,"circles",function(){return o}),n.d(t,"edges",function(){return a}),n.d(t,"default",function(){return m});var r,i,o,a,s=n("./node_modules/d3-voronoi/src/Beach.js"),u=n("./node_modules/d3-voronoi/src/Cell.js"),c=n("./node_modules/d3-voronoi/src/Circle.js"),l=n("./node_modules/d3-voronoi/src/Edge.js"),d=n("./node_modules/d3-voronoi/src/RedBlackTree.js"),p=1e-6,f=1e-12;function h(e,t){return t[1]-e[1]||t[0]-e[0]}function m(e,t){var n,p,f,m=e.sort(h).pop();for(a=[],i=new Array(e.length),r=new d.default,o=new d.default;;)if(f=c.firstCircle,m&&(!f||m[1]<f.y||m[1]===f.y&&m[0]<f.x))m[0]===n&&m[1]===p||(Object(s.addBeach)(m),n=m[0],p=m[1]),m=e.pop();else{if(!f)break;Object(s.removeBeach)(f.arc)}if(Object(u.sortCellHalfedges)(),t){var g=+t[0][0],v=+t[0][1],_=+t[1][0],y=+t[1][1];Object(l.clipEdges)(g,v,_,y),Object(u.clipCells)(g,v,_,y)}this.edges=a,this.cells=i,r=o=a=i=null}m.prototype={constructor:m,polygons:function(){var e=this.edges;return this.cells.map(function(t){var n=t.halfedges.map(function(n){return Object(u.cellHalfedgeStart)(t,e[n])});return n.data=t.site.data,n})},triangles:function(){var e=[],t=this.edges;return this.cells.forEach(function(n,r){if(o=(i=n.halfedges).length)for(var i,o,a,s,u,c,l=n.site,d=-1,p=t[i[o-1]],f=p.left===l?p.right:p.left;++d<o;)a=f,f=(p=t[i[d]]).left===l?p.right:p.left,a&&f&&r<a.index&&r<f.index&&(u=a,c=f,((s=l)[0]-c[0])*(u[1]-s[1])-(s[0]-u[0])*(c[1]-s[1])<0)&&e.push([l.data,a.data,f.data])}),e},links:function(){return this.edges.filter(function(e){return e.right}).map(function(e){return{source:e.left.data,target:e.right.data}})},find:function(e,t,n){for(var r,i,o=this,a=o._found||0,s=o.cells.length;!(i=o.cells[a]);)if(++a>=s)return null;var u=e-i.site[0],c=t-i.site[1],l=u*u+c*c;do{i=o.cells[r=a],a=null,i.halfedges.forEach(function(n){var r=o.edges[n],s=r.left;if(s!==i.site&&s||(s=r.right)){var u=e-s[0],c=t-s[1],d=u*u+c*c;d<l&&(l=d,a=s.index)}})}while(null!==a);return o._found=r,null==n||l<=n*n?i.site:null}}},"./node_modules/d3-voronoi/src/Edge.js":function(e,t,n){"use strict";n.r(t),n.d(t,"createEdge",function(){return i}),n.d(t,"createBorderEdge",function(){return o}),n.d(t,"setEdgeEnd",function(){return a}),n.d(t,"clipEdges",function(){return c});var r=n("./node_modules/d3-voronoi/src/Diagram.js");function i(e,t,n,i){var o=[null,null],s=r.edges.push(o)-1;return o.left=e,o.right=t,n&&a(o,e,t,n),i&&a(o,t,e,i),r.cells[e.index].halfedges.push(s),r.cells[t.index].halfedges.push(s),o}function o(e,t,n){var r=[t,n];return r.left=e,r}function a(e,t,n,r){e[0]||e[1]?e.left===n?e[1]=r:e[0]=r:(e[0]=r,e.left=t,e.right=n)}function s(e,t,n,r,i){var o,a=e[0],s=e[1],u=a[0],c=a[1],l=0,d=1,p=s[0]-u,f=s[1]-c;if(o=t-u,p||!(o>0)){if(o/=p,p<0){if(o<l)return;o<d&&(d=o)}else if(p>0){if(o>d)return;o>l&&(l=o)}if(o=r-u,p||!(o<0)){if(o/=p,p<0){if(o>d)return;o>l&&(l=o)}else if(p>0){if(o<l)return;o<d&&(d=o)}if(o=n-c,f||!(o>0)){if(o/=f,f<0){if(o<l)return;o<d&&(d=o)}else if(f>0){if(o>d)return;o>l&&(l=o)}if(o=i-c,f||!(o<0)){if(o/=f,f<0){if(o>d)return;o>l&&(l=o)}else if(f>0){if(o<l)return;o<d&&(d=o)}return!(l>0||d<1)||(l>0&&(e[0]=[u+l*p,c+l*f]),d<1&&(e[1]=[u+d*p,c+d*f]),!0)}}}}}function u(e,t,n,r,i){var o=e[1];if(o)return!0;var a,s,u=e[0],c=e.left,l=e.right,d=c[0],p=c[1],f=l[0],h=l[1],m=(d+f)/2,g=(p+h)/2;if(h===p){if(m<t||m>=r)return;if(d>f){if(u){if(u[1]>=i)return}else u=[m,n];o=[m,i]}else{if(u){if(u[1]<n)return}else u=[m,i];o=[m,n]}}else if(s=g-(a=(d-f)/(h-p))*m,a<-1||a>1)if(d>f){if(u){if(u[1]>=i)return}else u=[(n-s)/a,n];o=[(i-s)/a,i]}else{if(u){if(u[1]<n)return}else u=[(i-s)/a,i];o=[(n-s)/a,n]}else if(p<h){if(u){if(u[0]>=r)return}else u=[t,a*t+s];o=[r,a*r+s]}else{if(u){if(u[0]<t)return}else u=[r,a*r+s];o=[t,a*t+s]}return e[0]=u,e[1]=o,!0}function c(e,t,n,i){for(var o,a=r.edges.length;a--;)u(o=r.edges[a],e,t,n,i)&&s(o,e,t,n,i)&&(Math.abs(o[0][0]-o[1][0])>r.epsilon||Math.abs(o[0][1]-o[1][1])>r.epsilon)||delete r.edges[a]}},"./node_modules/d3-voronoi/src/RedBlackTree.js":function(e,t,n){"use strict";function r(){this._=null}function i(e){e.U=e.C=e.L=e.R=e.P=e.N=null}function o(e,t){var n=t,r=t.R,i=n.U;i?i.L===n?i.L=r:i.R=r:e._=r,r.U=i,n.U=r,n.R=r.L,n.R&&(n.R.U=n),r.L=n}function a(e,t){var n=t,r=t.L,i=n.U;i?i.L===n?i.L=r:i.R=r:e._=r,r.U=i,n.U=r,n.L=r.R,n.L&&(n.L.U=n),r.R=n}function s(e){for(;e.L;)e=e.L;return e}n.r(t),n.d(t,"RedBlackNode",function(){return i}),r.prototype={constructor:r,insert:function(e,t){var n,r,i;if(e){if(t.P=e,t.N=e.N,e.N&&(e.N.P=t),e.N=t,e.R){for(e=e.R;e.L;)e=e.L;e.L=t}else e.R=t;n=e}else this._?(e=s(this._),t.P=null,t.N=e,e.P=e.L=t,n=e):(t.P=t.N=null,this._=t,n=null);for(t.L=t.R=null,t.U=n,t.C=!0,e=t;n&&n.C;)n===(r=n.U).L?(i=r.R)&&i.C?(n.C=i.C=!1,r.C=!0,e=r):(e===n.R&&(o(this,n),n=(e=n).U),n.C=!1,r.C=!0,a(this,r)):(i=r.L)&&i.C?(n.C=i.C=!1,r.C=!0,e=r):(e===n.L&&(a(this,n),n=(e=n).U),n.C=!1,r.C=!0,o(this,r)),n=e.U;this._.C=!1},remove:function(e){e.N&&(e.N.P=e.P),e.P&&(e.P.N=e.N),e.N=e.P=null;var t,n,r,i=e.U,u=e.L,c=e.R;if(n=u?c?s(c):u:c,i?i.L===e?i.L=n:i.R=n:this._=n,u&&c?(r=n.C,n.C=e.C,n.L=u,u.U=n,n!==c?(i=n.U,n.U=e.U,e=n.R,i.L=e,n.R=c,c.U=n):(n.U=i,i=n,e=n.R)):(r=e.C,e=n),e&&(e.U=i),!r)if(e&&e.C)e.C=!1;else{do{if(e===this._)break;if(e===i.L){if((t=i.R).C&&(t.C=!1,i.C=!0,o(this,i),t=i.R),t.L&&t.L.C||t.R&&t.R.C){t.R&&t.R.C||(t.L.C=!1,t.C=!0,a(this,t),t=i.R),t.C=i.C,i.C=t.R.C=!1,o(this,i),e=this._;break}}else if((t=i.L).C&&(t.C=!1,i.C=!0,a(this,i),t=i.L),t.L&&t.L.C||t.R&&t.R.C){t.L&&t.L.C||(t.R.C=!1,t.C=!0,o(this,t),t=i.L),t.C=i.C,i.C=t.L.C=!1,a(this,i),e=this._;break}t.C=!0,e=i,i=i.U}while(!e.C);e&&(e.C=!1)}}},t.default=r},"./node_modules/d3-voronoi/src/constant.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return function(){return e}}},"./node_modules/d3-voronoi/src/point.js":function(e,t,n){"use strict";function r(e){return e[0]}function i(e){return e[1]}n.r(t),n.d(t,"x",function(){return r}),n.d(t,"y",function(){return i})},"./node_modules/d3-voronoi/src/voronoi.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-voronoi/src/constant.js"),i=n("./node_modules/d3-voronoi/src/point.js"),o=n("./node_modules/d3-voronoi/src/Diagram.js");t.default=function(){var e=i.x,t=i.y,n=null;function a(r){return new o.default(r.map(function(n,i){var a=[Math.round(e(n,i,r)/o.epsilon)*o.epsilon,Math.round(t(n,i,r)/o.epsilon)*o.epsilon];return a.index=i,a.data=n,a}),n)}return a.polygons=function(e){return a(e).polygons()},a.links=function(e){return a(e).links()},a.triangles=function(e){return a(e).triangles()},a.x=function(t){return arguments.length?(e="function"==typeof t?t:Object(r.default)(+t),a):e},a.y=function(e){return arguments.length?(t="function"==typeof e?e:Object(r.default)(+e),a):t},a.extent=function(e){return arguments.length?(n=null==e?null:[[+e[0][0],+e[0][1]],[+e[1][0],+e[1][1]]],a):n&&[[n[0][0],n[0][1]],[n[1][0],n[1][1]]]},a.size=function(e){return arguments.length?(n=null==e?null:[[0,0],[+e[0],+e[1]]],a):n&&[n[1][0]-n[0][0],n[1][1]-n[0][1]]},a}},"./node_modules/rxjs/_esm5/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/rxjs/_esm5/internal/Observable.js");n.d(t,"Observable",function(){return r.Observable});var i=n("./node_modules/rxjs/_esm5/internal/observable/ConnectableObservable.js");n.d(t,"ConnectableObservable",function(){return i.ConnectableObservable});var o=n("./node_modules/rxjs/_esm5/internal/operators/groupBy.js");n.d(t,"GroupedObservable",function(){return o.GroupedObservable});var a=n("./node_modules/rxjs/_esm5/internal/symbol/observable.js");n.d(t,"observable",function(){return a.observable});var s=n("./node_modules/rxjs/_esm5/internal/Subject.js");n.d(t,"Subject",function(){return s.Subject});var u=n("./node_modules/rxjs/_esm5/internal/BehaviorSubject.js");n.d(t,"BehaviorSubject",function(){return u.BehaviorSubject});var c=n("./node_modules/rxjs/_esm5/internal/ReplaySubject.js");n.d(t,"ReplaySubject",function(){return c.ReplaySubject});var l=n("./node_modules/rxjs/_esm5/internal/AsyncSubject.js");n.d(t,"AsyncSubject",function(){return l.AsyncSubject});var d=n("./node_modules/rxjs/_esm5/internal/scheduler/asap.js");n.d(t,"asapScheduler",function(){return d.asap});var p=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js");n.d(t,"asyncScheduler",function(){return p.async});var f=n("./node_modules/rxjs/_esm5/internal/scheduler/queue.js");n.d(t,"queueScheduler",function(){return f.queue});var h=n("./node_modules/rxjs/_esm5/internal/scheduler/animationFrame.js");n.d(t,"animationFrameScheduler",function(){return h.animationFrame});var m=n("./node_modules/rxjs/_esm5/internal/scheduler/VirtualTimeScheduler.js");n.d(t,"VirtualTimeScheduler",function(){return m.VirtualTimeScheduler}),n.d(t,"VirtualAction",function(){return m.VirtualAction});var g=n("./node_modules/rxjs/_esm5/internal/Scheduler.js");n.d(t,"Scheduler",function(){return g.Scheduler});var v=n("./node_modules/rxjs/_esm5/internal/Subscription.js");n.d(t,"Subscription",function(){return v.Subscription});var _=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");n.d(t,"Subscriber",function(){return _.Subscriber});var y=n("./node_modules/rxjs/_esm5/internal/Notification.js");n.d(t,"Notification",function(){return y.Notification});var b=n("./node_modules/rxjs/_esm5/internal/util/pipe.js");n.d(t,"pipe",function(){return b.pipe});var j=n("./node_modules/rxjs/_esm5/internal/util/noop.js");n.d(t,"noop",function(){return j.noop});var x=n("./node_modules/rxjs/_esm5/internal/util/identity.js");n.d(t,"identity",function(){return x.identity});var O=n("./node_modules/rxjs/_esm5/internal/util/isObservable.js");n.d(t,"isObservable",function(){return O.isObservable});var w=n("./node_modules/rxjs/_esm5/internal/util/ArgumentOutOfRangeError.js");n.d(t,"ArgumentOutOfRangeError",function(){return w.ArgumentOutOfRangeError});var C=n("./node_modules/rxjs/_esm5/internal/util/EmptyError.js");n.d(t,"EmptyError",function(){return C.EmptyError});var E=n("./node_modules/rxjs/_esm5/internal/util/ObjectUnsubscribedError.js");n.d(t,"ObjectUnsubscribedError",function(){return E.ObjectUnsubscribedError});var S=n("./node_modules/rxjs/_esm5/internal/util/UnsubscriptionError.js");n.d(t,"UnsubscriptionError",function(){return S.UnsubscriptionError});var k=n("./node_modules/rxjs/_esm5/internal/util/TimeoutError.js");n.d(t,"TimeoutError",function(){return k.TimeoutError});var M=n("./node_modules/rxjs/_esm5/internal/observable/bindCallback.js");n.d(t,"bindCallback",function(){return M.bindCallback});var A=n("./node_modules/rxjs/_esm5/internal/observable/bindNodeCallback.js");n.d(t,"bindNodeCallback",function(){return A.bindNodeCallback});var T=n("./node_modules/rxjs/_esm5/internal/observable/combineLatest.js");n.d(t,"combineLatest",function(){return T.combineLatest});var D=n("./node_modules/rxjs/_esm5/internal/observable/concat.js");n.d(t,"concat",function(){return D.concat});var I=n("./node_modules/rxjs/_esm5/internal/observable/defer.js");n.d(t,"defer",function(){return I.defer});var P=n("./node_modules/rxjs/_esm5/internal/observable/empty.js");n.d(t,"empty",function(){return P.empty});var R=n("./node_modules/rxjs/_esm5/internal/observable/forkJoin.js");n.d(t,"forkJoin",function(){return R.forkJoin});var N=n("./node_modules/rxjs/_esm5/internal/observable/from.js");n.d(t,"from",function(){return N.from});var F=n("./node_modules/rxjs/_esm5/internal/observable/fromEvent.js");n.d(t,"fromEvent",function(){return F.fromEvent});var L=n("./node_modules/rxjs/_esm5/internal/observable/fromEventPattern.js");n.d(t,"fromEventPattern",function(){return L.fromEventPattern});var V=n("./node_modules/rxjs/_esm5/internal/observable/generate.js");n.d(t,"generate",function(){return V.generate});var B=n("./node_modules/rxjs/_esm5/internal/observable/iif.js");n.d(t,"iif",function(){return B.iif});var z=n("./node_modules/rxjs/_esm5/internal/observable/interval.js");n.d(t,"interval",function(){return z.interval});var U=n("./node_modules/rxjs/_esm5/internal/observable/merge.js");n.d(t,"merge",function(){return U.merge});var H=n("./node_modules/rxjs/_esm5/internal/observable/never.js");n.d(t,"never",function(){return H.never});var q=n("./node_modules/rxjs/_esm5/internal/observable/of.js");n.d(t,"of",function(){return q.of});var G=n("./node_modules/rxjs/_esm5/internal/observable/onErrorResumeNext.js");n.d(t,"onErrorResumeNext",function(){return G.onErrorResumeNext});var W=n("./node_modules/rxjs/_esm5/internal/observable/pairs.js");n.d(t,"pairs",function(){return W.pairs});var Y=n("./node_modules/rxjs/_esm5/internal/observable/race.js");n.d(t,"race",function(){return Y.race});var K=n("./node_modules/rxjs/_esm5/internal/observable/range.js");n.d(t,"range",function(){return K.range});var Q=n("./node_modules/rxjs/_esm5/internal/observable/throwError.js");n.d(t,"throwError",function(){return Q.throwError});var X=n("./node_modules/rxjs/_esm5/internal/observable/timer.js");n.d(t,"timer",function(){return X.timer});var Z=n("./node_modules/rxjs/_esm5/internal/observable/using.js");n.d(t,"using",function(){return Z.using});var $=n("./node_modules/rxjs/_esm5/internal/observable/zip.js");n.d(t,"zip",function(){return $.zip}),n.d(t,"EMPTY",function(){return P.EMPTY}),n.d(t,"NEVER",function(){return H.NEVER});var J=n("./node_modules/rxjs/_esm5/internal/config.js");n.d(t,"config",function(){return J.config})},"./node_modules/rxjs/_esm5/internal/AsyncSubject.js":function(e,t,n){"use strict";n.r(t),n.d(t,"AsyncSubject",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),a=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.value=null,t.hasNext=!1,t.hasCompleted=!1,t}return r.__extends(t,e),t.prototype._subscribe=function(t){return this.hasError?(t.error(this.thrownError),o.Subscription.EMPTY):this.hasCompleted&&this.hasNext?(t.next(this.value),t.complete(),o.Subscription.EMPTY):e.prototype._subscribe.call(this,t)},t.prototype.next=function(e){this.hasCompleted||(this.value=e,this.hasNext=!0)},t.prototype.error=function(t){this.hasCompleted||e.prototype.error.call(this,t)},t.prototype.complete=function(){this.hasCompleted=!0,this.hasNext&&e.prototype.next.call(this,this.value),e.prototype.complete.call(this)},t}(i.Subject)},"./node_modules/rxjs/_esm5/internal/BehaviorSubject.js":function(e,t,n){"use strict";n.r(t),n.d(t,"BehaviorSubject",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/util/ObjectUnsubscribedError.js"),a=function(e){function t(t){var n=e.call(this)||this;return n._value=t,n}return r.__extends(t,e),Object.defineProperty(t.prototype,"value",{get:function(){return this.getValue()},enumerable:!0,configurable:!0}),t.prototype._subscribe=function(t){var n=e.prototype._subscribe.call(this,t);return n&&!n.closed&&t.next(this._value),n},t.prototype.getValue=function(){if(this.hasError)throw this.thrownError;if(this.closed)throw new o.ObjectUnsubscribedError;return this._value},t.prototype.next=function(t){e.prototype.next.call(this,this._value=t)},t}(i.Subject)},"./node_modules/rxjs/_esm5/internal/InnerSubscriber.js":function(e,t,n){"use strict";n.r(t),n.d(t,"InnerSubscriber",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(e){function t(t,n,r){var i=e.call(this)||this;return i.parent=t,i.outerValue=n,i.outerIndex=r,i.index=0,i}return r.__extends(t,e),t.prototype._next=function(e){this.parent.notifyNext(this.outerValue,e,this.outerIndex,this.index++,this)},t.prototype._error=function(e){this.parent.notifyError(e,this),this.unsubscribe()},t.prototype._complete=function(){this.parent.notifyComplete(this),this.unsubscribe()},t}(n("./node_modules/rxjs/_esm5/internal/Subscriber.js").Subscriber)},"./node_modules/rxjs/_esm5/internal/Notification.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Notification",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/observable/empty.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/of.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/throwError.js"),a=function(){function e(e,t,n){this.kind=e,this.value=t,this.error=n,this.hasValue="N"===e}return e.prototype.observe=function(e){switch(this.kind){case"N":return e.next&&e.next(this.value);case"E":return e.error&&e.error(this.error);case"C":return e.complete&&e.complete()}},e.prototype.do=function(e,t,n){switch(this.kind){case"N":return e&&e(this.value);case"E":return t&&t(this.error);case"C":return n&&n()}},e.prototype.accept=function(e,t,n){return e&&"function"==typeof e.next?this.observe(e):this.do(e,t,n)},e.prototype.toObservable=function(){switch(this.kind){case"N":return Object(i.of)(this.value);case"E":return Object(o.throwError)(this.error);case"C":return Object(r.empty)()}throw new Error("unexpected notification kind value")},e.createNext=function(t){return void 0!==t?new e("N",t):e.undefinedValueNotification},e.createError=function(t){return new e("E",void 0,t)},e.createComplete=function(){return e.completeNotification},e.completeNotification=new e("C"),e.undefinedValueNotification=new e("N",void 0),e}()},"./node_modules/rxjs/_esm5/internal/Observable.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Observable",function(){return u});var r=n("./node_modules/rxjs/_esm5/internal/util/canReportError.js"),i=n("./node_modules/rxjs/_esm5/internal/util/toSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/symbol/observable.js"),a=n("./node_modules/rxjs/_esm5/internal/util/pipe.js"),s=n("./node_modules/rxjs/_esm5/internal/config.js"),u=function(){function e(e){this._isScalar=!1,e&&(this._subscribe=e)}return e.prototype.lift=function(t){var n=new e;return n.source=this,n.operator=t,n},e.prototype.subscribe=function(e,t,n){var r=this.operator,o=Object(i.toSubscriber)(e,t,n);if(r?r.call(o,this.source):o.add(this.source||s.config.useDeprecatedSynchronousErrorHandling&&!o.syncErrorThrowable?this._subscribe(o):this._trySubscribe(o)),s.config.useDeprecatedSynchronousErrorHandling&&o.syncErrorThrowable&&(o.syncErrorThrowable=!1,o.syncErrorThrown))throw o.syncErrorValue;return o},e.prototype._trySubscribe=function(e){try{return this._subscribe(e)}catch(t){s.config.useDeprecatedSynchronousErrorHandling&&(e.syncErrorThrown=!0,e.syncErrorValue=t),Object(r.canReportError)(e)?e.error(t):console.warn(t)}},e.prototype.forEach=function(e,t){var n=this;return new(t=c(t))(function(t,r){var i;i=n.subscribe(function(t){try{e(t)}catch(e){r(e),i&&i.unsubscribe()}},r,t)})},e.prototype._subscribe=function(e){var t=this.source;return t&&t.subscribe(e)},e.prototype[o.observable]=function(){return this},e.prototype.pipe=function(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];return 0===e.length?this:Object(a.pipeFromArray)(e)(this)},e.prototype.toPromise=function(e){var t=this;return new(e=c(e))(function(e,n){var r;t.subscribe(function(e){return r=e},function(e){return n(e)},function(){return e(r)})})},e.create=function(t){return new e(t)},e}();function c(e){if(e||(e=s.config.Promise||Promise),!e)throw new Error("no Promise impl found");return e}},"./node_modules/rxjs/_esm5/internal/Observer.js":function(e,t,n){"use strict";n.r(t),n.d(t,"empty",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/config.js"),i=n("./node_modules/rxjs/_esm5/internal/util/hostReportError.js"),o={closed:!0,next:function(e){},error:function(e){if(r.config.useDeprecatedSynchronousErrorHandling)throw e;Object(i.hostReportError)(e)},complete:function(){}}},"./node_modules/rxjs/_esm5/internal/OuterSubscriber.js":function(e,t,n){"use strict";n.r(t),n.d(t,"OuterSubscriber",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return r.__extends(t,e),t.prototype.notifyNext=function(e,t,n,r,i){this.destination.next(t)},t.prototype.notifyError=function(e,t){this.destination.error(e)},t.prototype.notifyComplete=function(e){this.destination.complete()},t}(n("./node_modules/rxjs/_esm5/internal/Subscriber.js").Subscriber)},"./node_modules/rxjs/_esm5/internal/ReplaySubject.js":function(e,t,n){"use strict";n.r(t),n.d(t,"ReplaySubject",function(){return l});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/scheduler/queue.js"),a=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),s=n("./node_modules/rxjs/_esm5/internal/operators/observeOn.js"),u=n("./node_modules/rxjs/_esm5/internal/util/ObjectUnsubscribedError.js"),c=n("./node_modules/rxjs/_esm5/internal/SubjectSubscription.js"),l=function(e){function t(t,n,r){void 0===t&&(t=Number.POSITIVE_INFINITY),void 0===n&&(n=Number.POSITIVE_INFINITY);var i=e.call(this)||this;return i.scheduler=r,i._events=[],i._infiniteTimeWindow=!1,i._bufferSize=t<1?1:t,i._windowTime=n<1?1:n,n===Number.POSITIVE_INFINITY?(i._infiniteTimeWindow=!0,i.next=i.nextInfiniteTimeWindow):i.next=i.nextTimeWindow,i}return r.__extends(t,e),t.prototype.nextInfiniteTimeWindow=function(t){var n=this._events;n.push(t),n.length>this._bufferSize&&n.shift(),e.prototype.next.call(this,t)},t.prototype.nextTimeWindow=function(t){this._events.push(new d(this._getNow(),t)),this._trimBufferThenGetEvents(),e.prototype.next.call(this,t)},t.prototype._subscribe=function(e){var t,n=this._infiniteTimeWindow,r=n?this._events:this._trimBufferThenGetEvents(),i=this.scheduler,o=r.length;if(this.closed)throw new u.ObjectUnsubscribedError;if(this.isStopped||this.hasError?t=a.Subscription.EMPTY:(this.observers.push(e),t=new c.SubjectSubscription(this,e)),i&&e.add(e=new s.ObserveOnSubscriber(e,i)),n)for(var l=0;l<o&&!e.closed;l++)e.next(r[l]);else for(l=0;l<o&&!e.closed;l++)e.next(r[l].value);return this.hasError?e.error(this.thrownError):this.isStopped&&e.complete(),t},t.prototype._getNow=function(){return(this.scheduler||o.queue).now()},t.prototype._trimBufferThenGetEvents=function(){for(var e=this._getNow(),t=this._bufferSize,n=this._windowTime,r=this._events,i=r.length,o=0;o<i&&!(e-r[o].time<n);)o++;return i>t&&(o=Math.max(o,i-t)),o>0&&r.splice(0,o),r},t}(i.Subject),d=function(){return function(e,t){this.time=e,this.value=t}}()},"./node_modules/rxjs/_esm5/internal/Scheduler.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Scheduler",function(){return r});var r=function(){function e(t,n){void 0===n&&(n=e.now),this.SchedulerAction=t,this.now=n}return e.prototype.schedule=function(e,t,n){return void 0===t&&(t=0),new this.SchedulerAction(this,e).schedule(n,t)},e.now=function(){return Date.now()},e}()},"./node_modules/rxjs/_esm5/internal/Subject.js":function(e,t,n){"use strict";n.r(t),n.d(t,"SubjectSubscriber",function(){return l}),n.d(t,"Subject",function(){return d}),n.d(t,"AnonymousSubject",function(){return p});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Observable.js"),o=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),a=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),s=n("./node_modules/rxjs/_esm5/internal/util/ObjectUnsubscribedError.js"),u=n("./node_modules/rxjs/_esm5/internal/SubjectSubscription.js"),c=n("./node_modules/rxjs/_esm5/internal/symbol/rxSubscriber.js"),l=function(e){function t(t){var n=e.call(this,t)||this;return n.destination=t,n}return r.__extends(t,e),t}(o.Subscriber),d=function(e){function t(){var t=e.call(this)||this;return t.observers=[],t.closed=!1,t.isStopped=!1,t.hasError=!1,t.thrownError=null,t}return r.__extends(t,e),t.prototype[c.rxSubscriber]=function(){return new l(this)},t.prototype.lift=function(e){var t=new p(this,this);return t.operator=e,t},t.prototype.next=function(e){if(this.closed)throw new s.ObjectUnsubscribedError;if(!this.isStopped)for(var t=this.observers,n=t.length,r=t.slice(),i=0;i<n;i++)r[i].next(e)},t.prototype.error=function(e){if(this.closed)throw new s.ObjectUnsubscribedError;this.hasError=!0,this.thrownError=e,this.isStopped=!0;for(var t=this.observers,n=t.length,r=t.slice(),i=0;i<n;i++)r[i].error(e);this.observers.length=0},t.prototype.complete=function(){if(this.closed)throw new s.ObjectUnsubscribedError;this.isStopped=!0;for(var e=this.observers,t=e.length,n=e.slice(),r=0;r<t;r++)n[r].complete();this.observers.length=0},t.prototype.unsubscribe=function(){this.isStopped=!0,this.closed=!0,this.observers=null},t.prototype._trySubscribe=function(t){if(this.closed)throw new s.ObjectUnsubscribedError;return e.prototype._trySubscribe.call(this,t)},t.prototype._subscribe=function(e){if(this.closed)throw new s.ObjectUnsubscribedError;return this.hasError?(e.error(this.thrownError),a.Subscription.EMPTY):this.isStopped?(e.complete(),a.Subscription.EMPTY):(this.observers.push(e),new u.SubjectSubscription(this,e))},t.prototype.asObservable=function(){var e=new i.Observable;return e.source=this,e},t.create=function(e,t){return new p(e,t)},t}(i.Observable),p=function(e){function t(t,n){var r=e.call(this)||this;return r.destination=t,r.source=n,r}return r.__extends(t,e),t.prototype.next=function(e){var t=this.destination;t&&t.next&&t.next(e)},t.prototype.error=function(e){var t=this.destination;t&&t.error&&this.destination.error(e)},t.prototype.complete=function(){var e=this.destination;e&&e.complete&&this.destination.complete()},t.prototype._subscribe=function(e){return this.source?this.source.subscribe(e):a.Subscription.EMPTY},t}(d)},"./node_modules/rxjs/_esm5/internal/SubjectSubscription.js":function(e,t,n){"use strict";n.r(t),n.d(t,"SubjectSubscription",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(e){function t(t,n){var r=e.call(this)||this;return r.subject=t,r.subscriber=n,r.closed=!1,r}return r.__extends(t,e),t.prototype.unsubscribe=function(){if(!this.closed){this.closed=!0;var e=this.subject,t=e.observers;if(this.subject=null,t&&0!==t.length&&!e.isStopped&&!e.closed){var n=t.indexOf(this.subscriber);-1!==n&&t.splice(n,1)}}},t}(n("./node_modules/rxjs/_esm5/internal/Subscription.js").Subscription)},"./node_modules/rxjs/_esm5/internal/Subscriber.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Subscriber",function(){return l}),n.d(t,"SafeSubscriber",function(){return d});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/util/isFunction.js"),o=n("./node_modules/rxjs/_esm5/internal/Observer.js"),a=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),s=n("./node_modules/rxjs/_esm5/internal/symbol/rxSubscriber.js"),u=n("./node_modules/rxjs/_esm5/internal/config.js"),c=n("./node_modules/rxjs/_esm5/internal/util/hostReportError.js"),l=function(e){function t(n,r,i){var a=e.call(this)||this;switch(a.syncErrorValue=null,a.syncErrorThrown=!1,a.syncErrorThrowable=!1,a.isStopped=!1,a._parentSubscription=null,arguments.length){case 0:a.destination=o.empty;break;case 1:if(!n){a.destination=o.empty;break}if("object"==typeof n){n instanceof t?(a.syncErrorThrowable=n.syncErrorThrowable,a.destination=n,n.add(a)):(a.syncErrorThrowable=!0,a.destination=new d(a,n));break}default:a.syncErrorThrowable=!0,a.destination=new d(a,n,r,i)}return a}return r.__extends(t,e),t.prototype[s.rxSubscriber]=function(){return this},t.create=function(e,n,r){var i=new t(e,n,r);return i.syncErrorThrowable=!1,i},t.prototype.next=function(e){this.isStopped||this._next(e)},t.prototype.error=function(e){this.isStopped||(this.isStopped=!0,this._error(e))},t.prototype.complete=function(){this.isStopped||(this.isStopped=!0,this._complete())},t.prototype.unsubscribe=function(){this.closed||(this.isStopped=!0,e.prototype.unsubscribe.call(this))},t.prototype._next=function(e){this.destination.next(e)},t.prototype._error=function(e){this.destination.error(e),this.unsubscribe()},t.prototype._complete=function(){this.destination.complete(),this.unsubscribe()},t.prototype._unsubscribeAndRecycle=function(){var e=this._parent,t=this._parents;return this._parent=null,this._parents=null,this.unsubscribe(),this.closed=!1,this.isStopped=!1,this._parent=e,this._parents=t,this._parentSubscription=null,this},t}(a.Subscription),d=function(e){function t(t,n,r,a){var s,u=e.call(this)||this;u._parentSubscriber=t;var c=u;return Object(i.isFunction)(n)?s=n:n&&(s=n.next,r=n.error,a=n.complete,n!==o.empty&&(c=Object.create(n),Object(i.isFunction)(c.unsubscribe)&&u.add(c.unsubscribe.bind(c)),c.unsubscribe=u.unsubscribe.bind(u))),u._context=c,u._next=s,u._error=r,u._complete=a,u}return r.__extends(t,e),t.prototype.next=function(e){if(!this.isStopped&&this._next){var t=this._parentSubscriber;u.config.useDeprecatedSynchronousErrorHandling&&t.syncErrorThrowable?this.__tryOrSetError(t,this._next,e)&&this.unsubscribe():this.__tryOrUnsub(this._next,e)}},t.prototype.error=function(e){if(!this.isStopped){var t=this._parentSubscriber,n=u.config.useDeprecatedSynchronousErrorHandling;if(this._error)n&&t.syncErrorThrowable?(this.__tryOrSetError(t,this._error,e),this.unsubscribe()):(this.__tryOrUnsub(this._error,e),this.unsubscribe());else if(t.syncErrorThrowable)n?(t.syncErrorValue=e,t.syncErrorThrown=!0):Object(c.hostReportError)(e),this.unsubscribe();else{if(this.unsubscribe(),n)throw e;Object(c.hostReportError)(e)}}},t.prototype.complete=function(){var e=this;if(!this.isStopped){var t=this._parentSubscriber;if(this._complete){var n=function(){return e._complete.call(e._context)};u.config.useDeprecatedSynchronousErrorHandling&&t.syncErrorThrowable?(this.__tryOrSetError(t,n),this.unsubscribe()):(this.__tryOrUnsub(n),this.unsubscribe())}else this.unsubscribe()}},t.prototype.__tryOrUnsub=function(e,t){try{e.call(this._context,t)}catch(e){if(this.unsubscribe(),u.config.useDeprecatedSynchronousErrorHandling)throw e;Object(c.hostReportError)(e)}},t.prototype.__tryOrSetError=function(e,t,n){if(!u.config.useDeprecatedSynchronousErrorHandling)throw new Error("bad call");try{t.call(this._context,n)}catch(t){return u.config.useDeprecatedSynchronousErrorHandling?(e.syncErrorValue=t,e.syncErrorThrown=!0,!0):(Object(c.hostReportError)(t),!0)}return!1},t.prototype._unsubscribe=function(){var e=this._parentSubscriber;this._context=null,this._parentSubscriber=null,e.unsubscribe()},t}(l)},"./node_modules/rxjs/_esm5/internal/Subscription.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Subscription",function(){return c});var r=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),i=n("./node_modules/rxjs/_esm5/internal/util/isObject.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isFunction.js"),a=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),s=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),u=n("./node_modules/rxjs/_esm5/internal/util/UnsubscriptionError.js"),c=function(){function e(e){this.closed=!1,this._parent=null,this._parents=null,this._subscriptions=null,e&&(this._unsubscribe=e)}var t;return e.prototype.unsubscribe=function(){var e,t=!1;if(!this.closed){var n=this._parent,c=this._parents,d=this._unsubscribe,p=this._subscriptions;this.closed=!0,this._parent=null,this._parents=null,this._subscriptions=null;for(var f=-1,h=c?c.length:0;n;)n.remove(this),n=++f<h&&c[f]||null;if(Object(o.isFunction)(d))Object(a.tryCatch)(d).call(this)===s.errorObject&&(t=!0,e=e||(s.errorObject.e instanceof u.UnsubscriptionError?l(s.errorObject.e.errors):[s.errorObject.e]));if(Object(r.isArray)(p))for(f=-1,h=p.length;++f<h;){var m=p[f];if(Object(i.isObject)(m))if(Object(a.tryCatch)(m.unsubscribe).call(m)===s.errorObject){t=!0,e=e||[];var g=s.errorObject.e;g instanceof u.UnsubscriptionError?e=e.concat(l(g.errors)):e.push(g)}}if(t)throw new u.UnsubscriptionError(e)}},e.prototype.add=function(t){if(!t||t===e.EMPTY)return e.EMPTY;if(t===this)return this;var n=t;switch(typeof t){case"function":n=new e(t);case"object":if(n.closed||"function"!=typeof n.unsubscribe)return n;if(this.closed)return n.unsubscribe(),n;if("function"!=typeof n._addParent){var r=n;(n=new e)._subscriptions=[r]}break;default:throw new Error("unrecognized teardown "+t+" added to Subscription.")}return(this._subscriptions||(this._subscriptions=[])).push(n),n._addParent(this),n},e.prototype.remove=function(e){var t=this._subscriptions;if(t){var n=t.indexOf(e);-1!==n&&t.splice(n,1)}},e.prototype._addParent=function(e){var t=this._parent,n=this._parents;t&&t!==e?n?-1===n.indexOf(e)&&n.push(e):this._parents=[e]:this._parent=e},e.EMPTY=((t=new e).closed=!0,t),e}();function l(e){return e.reduce(function(e,t){return e.concat(t instanceof u.UnsubscriptionError?t.errors:t)},[])}},"./node_modules/rxjs/_esm5/internal/config.js":function(e,t,n){"use strict";n.r(t),n.d(t,"config",function(){return i});var r=!1,i={Promise:void 0,set useDeprecatedSynchronousErrorHandling(e){if(e){var t=new Error;t.stack}r=e},get useDeprecatedSynchronousErrorHandling(){return r}}},"./node_modules/rxjs/_esm5/internal/observable/ConnectableObservable.js":function(e,t,n){"use strict";n.r(t),n.d(t,"ConnectableObservable",function(){return c}),n.d(t,"connectableObservableDescriptor",function(){return d});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/Observable.js"),a=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),u=n("./node_modules/rxjs/_esm5/internal/operators/refCount.js"),c=function(e){function t(t,n){var r=e.call(this)||this;return r.source=t,r.subjectFactory=n,r._refCount=0,r._isComplete=!1,r}return r.__extends(t,e),t.prototype._subscribe=function(e){return this.getSubject().subscribe(e)},t.prototype.getSubject=function(){var e=this._subject;return e&&!e.isStopped||(this._subject=this.subjectFactory()),this._subject},t.prototype.connect=function(){var e=this._connection;return e||(this._isComplete=!1,(e=this._connection=new s.Subscription).add(this.source.subscribe(new p(this.getSubject(),this))),e.closed?(this._connection=null,e=s.Subscription.EMPTY):this._connection=e),e},t.prototype.refCount=function(){return Object(u.refCount)()(this)},t}(o.Observable),l=c.prototype,d={operator:{value:null},_refCount:{value:0,writable:!0},_subject:{value:null,writable:!0},_connection:{value:null,writable:!0},_subscribe:{value:l._subscribe},_isComplete:{value:l._isComplete,writable:!0},getSubject:{value:l.getSubject},connect:{value:l.connect},refCount:{value:l.refCount}},p=function(e){function t(t,n){var r=e.call(this,t)||this;return r.connectable=n,r}return r.__extends(t,e),t.prototype._error=function(t){this._unsubscribe(),e.prototype._error.call(this,t)},t.prototype._complete=function(){this.connectable._isComplete=!0,this._unsubscribe(),e.prototype._complete.call(this)},t.prototype._unsubscribe=function(){var e=this.connectable;if(e){this.connectable=null;var t=e._connection;e._refCount=0,e._subject=null,e._connection=null,t&&t.unsubscribe()}},t}(i.SubjectSubscriber);a.Subscriber},"./node_modules/rxjs/_esm5/internal/observable/SubscribeOnObservable.js":function(e,t,n){"use strict";n.r(t),n.d(t,"SubscribeOnObservable",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Observable.js"),o=n("./node_modules/rxjs/_esm5/internal/scheduler/asap.js"),a=n("./node_modules/rxjs/_esm5/internal/util/isNumeric.js"),s=function(e){function t(t,n,r){void 0===n&&(n=0),void 0===r&&(r=o.asap);var i=e.call(this)||this;return i.source=t,i.delayTime=n,i.scheduler=r,(!Object(a.isNumeric)(n)||n<0)&&(i.delayTime=0),r&&"function"==typeof r.schedule||(i.scheduler=o.asap),i}return r.__extends(t,e),t.create=function(e,n,r){return void 0===n&&(n=0),void 0===r&&(r=o.asap),new t(e,n,r)},t.dispatch=function(e){var t=e.source,n=e.subscriber;return this.add(t.subscribe(n))},t.prototype._subscribe=function(e){var n=this.delayTime,r=this.source;return this.scheduler.schedule(t.dispatch,n,{source:r,subscriber:e})},t}(i.Observable)},"./node_modules/rxjs/_esm5/internal/observable/bindCallback.js":function(e,t,n){"use strict";n.r(t),n.d(t,"bindCallback",function(){return c});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/AsyncSubject.js"),o=n("./node_modules/rxjs/_esm5/internal/operators/map.js"),a=n("./node_modules/rxjs/_esm5/internal/util/canReportError.js"),s=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),u=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js");function c(e,t,n){if(t){if(!Object(u.isScheduler)(t))return function(){for(var r=[],i=0;i<arguments.length;i++)r[i]=arguments[i];return c(e,n).apply(void 0,r).pipe(Object(o.map)(function(e){return Object(s.isArray)(e)?t.apply(void 0,e):t(e)}))};n=t}return function(){for(var t=[],o=0;o<arguments.length;o++)t[o]=arguments[o];var s,u=this,c={context:u,subject:s,callbackFunc:e,scheduler:n};return new r.Observable(function(r){if(n){var o={args:t,subscriber:r,params:c};return n.schedule(l,0,o)}if(!s){s=new i.AsyncSubject;try{e.apply(u,t.concat([function(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];s.next(e.length<=1?e[0]:e),s.complete()}]))}catch(e){Object(a.canReportError)(s)?s.error(e):console.warn(e)}}return s.subscribe(r)})}}function l(e){var t=this,n=e.args,r=e.subscriber,o=e.params,a=o.callbackFunc,s=o.context,u=o.scheduler,c=o.subject;if(!c){c=o.subject=new i.AsyncSubject;try{a.apply(s,n.concat([function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];var r=e.length<=1?e[0]:e;t.add(u.schedule(d,0,{value:r,subject:c}))}]))}catch(e){c.error(e)}}this.add(c.subscribe(r))}function d(e){var t=e.value,n=e.subject;n.next(t),n.complete()}},"./node_modules/rxjs/_esm5/internal/observable/bindNodeCallback.js":function(e,t,n){"use strict";n.r(t),n.d(t,"bindNodeCallback",function(){return c});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/AsyncSubject.js"),o=n("./node_modules/rxjs/_esm5/internal/operators/map.js"),a=n("./node_modules/rxjs/_esm5/internal/util/canReportError.js"),s=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js"),u=n("./node_modules/rxjs/_esm5/internal/util/isArray.js");function c(e,t,n){if(t){if(!Object(s.isScheduler)(t))return function(){for(var r=[],i=0;i<arguments.length;i++)r[i]=arguments[i];return c(e,n).apply(void 0,r).pipe(Object(o.map)(function(e){return Object(u.isArray)(e)?t.apply(void 0,e):t(e)}))};n=t}return function(){for(var t=[],o=0;o<arguments.length;o++)t[o]=arguments[o];var s={subject:void 0,args:t,callbackFunc:e,scheduler:n,context:this};return new r.Observable(function(r){var o=s.context,u=s.subject;if(n)return n.schedule(l,0,{params:s,subscriber:r,context:o});if(!u){u=s.subject=new i.AsyncSubject;try{e.apply(o,t.concat([function(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];var n=e.shift();n?u.error(n):(u.next(e.length<=1?e[0]:e),u.complete())}]))}catch(e){Object(a.canReportError)(u)?u.error(e):console.warn(e)}}return u.subscribe(r)})}}function l(e){var t=this,n=e.params,r=e.subscriber,o=e.context,a=n.callbackFunc,s=n.args,u=n.scheduler,c=n.subject;if(!c){c=n.subject=new i.AsyncSubject;try{a.apply(o,s.concat([function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];var r=e.shift();if(r)t.add(u.schedule(p,0,{err:r,subject:c}));else{var i=e.length<=1?e[0]:e;t.add(u.schedule(d,0,{value:i,subject:c}))}}]))}catch(e){this.add(u.schedule(p,0,{err:e,subject:c}))}}this.add(c.subscribe(r))}function d(e){var t=e.value,n=e.subject;n.next(t),n.complete()}function p(e){var t=e.err;e.subject.error(t)}},"./node_modules/rxjs/_esm5/internal/observable/combineLatest.js":function(e,t,n){"use strict";n.r(t),n.d(t,"combineLatest",function(){return l}),n.d(t,"CombineLatestOperator",function(){return d}),n.d(t,"CombineLatestSubscriber",function(){return p});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),a=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),u=n("./node_modules/rxjs/_esm5/internal/observable/fromArray.js"),c={};function l(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];var n=null,r=null;return Object(i.isScheduler)(e[e.length-1])&&(r=e.pop()),"function"==typeof e[e.length-1]&&(n=e.pop()),1===e.length&&Object(o.isArray)(e[0])&&(e=e[0]),Object(u.fromArray)(e,r).lift(new d(n))}var d=function(){function e(e){this.resultSelector=e}return e.prototype.call=function(e,t){return t.subscribe(new p(e,this.resultSelector))},e}(),p=function(e){function t(t,n){var r=e.call(this,t)||this;return r.resultSelector=n,r.active=0,r.values=[],r.observables=[],r}return r.__extends(t,e),t.prototype._next=function(e){this.values.push(c),this.observables.push(e)},t.prototype._complete=function(){var e=this.observables,t=e.length;if(0===t)this.destination.complete();else{this.active=t,this.toRespond=t;for(var n=0;n<t;n++){var r=e[n];this.add(Object(s.subscribeToResult)(this,r,r,n))}}},t.prototype.notifyComplete=function(e){0==(this.active-=1)&&this.destination.complete()},t.prototype.notifyNext=function(e,t,n,r,i){var o=this.values,a=o[n],s=this.toRespond?a===c?--this.toRespond:this.toRespond:0;o[n]=t,0===s&&(this.resultSelector?this._tryResultSelector(o):this.destination.next(o.slice()))},t.prototype._tryResultSelector=function(e){var t;try{t=this.resultSelector.apply(this,e)}catch(e){return void this.destination.error(e)}this.destination.next(t)},t}(a.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/observable/concat.js":function(e,t,n){"use strict";n.r(t),n.d(t,"concat",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/of.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/from.js"),a=n("./node_modules/rxjs/_esm5/internal/operators/concatAll.js");function s(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];return 1===e.length||2===e.length&&Object(r.isScheduler)(e[1])?Object(o.from)(e[0]):Object(a.concatAll)()(i.of.apply(void 0,e))}},"./node_modules/rxjs/_esm5/internal/observable/defer.js":function(e,t,n){"use strict";n.r(t),n.d(t,"defer",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/from.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/empty.js");function a(e){return new r.Observable(function(t){var n;try{n=e()}catch(e){return void t.error(e)}return(n?Object(i.from)(n):Object(o.empty)()).subscribe(t)})}},"./node_modules/rxjs/_esm5/internal/observable/empty.js":function(e,t,n){"use strict";n.r(t),n.d(t,"EMPTY",function(){return i}),n.d(t,"empty",function(){return o}),n.d(t,"emptyScheduled",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=new r.Observable(function(e){return e.complete()});function o(e){return e?a(e):i}function a(e){return new r.Observable(function(t){return e.schedule(function(){return t.complete()})})}},"./node_modules/rxjs/_esm5/internal/observable/forkJoin.js":function(e,t,n){"use strict";n.r(t),n.d(t,"forkJoin",function(){return l});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Observable.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/empty.js"),s=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),u=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),c=n("./node_modules/rxjs/_esm5/internal/operators/map.js");function l(){for(var e,t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];return"function"==typeof t[t.length-1]&&(e=t.pop()),1===t.length&&Object(o.isArray)(t[0])&&(t=t[0]),0===t.length?a.EMPTY:e?l(t).pipe(Object(c.map)(function(t){return e.apply(void 0,t)})):new i.Observable(function(e){return new d(e,t)})}var d=function(e){function t(t,n){var r=e.call(this,t)||this;r.sources=n,r.completed=0,r.haveValues=0;var i=n.length;r.values=new Array(i);for(var o=0;o<i;o++){var a=n[o],u=Object(s.subscribeToResult)(r,a,null,o);u&&r.add(u)}return r}return r.__extends(t,e),t.prototype.notifyNext=function(e,t,n,r,i){this.values[n]=t,i._hasValue||(i._hasValue=!0,this.haveValues++)},t.prototype.notifyComplete=function(e){var t=this.destination,n=this.haveValues,r=this.values,i=r.length;e._hasValue?(this.completed++,this.completed===i&&(n===i&&t.next(r),t.complete())):t.complete()},t}(u.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/observable/from.js":function(e,t,n){"use strict";n.r(t),n.d(t,"from",function(){return f});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/util/isPromise.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isArrayLike.js"),a=n("./node_modules/rxjs/_esm5/internal/util/isInteropObservable.js"),s=n("./node_modules/rxjs/_esm5/internal/util/isIterable.js"),u=n("./node_modules/rxjs/_esm5/internal/observable/fromArray.js"),c=n("./node_modules/rxjs/_esm5/internal/observable/fromPromise.js"),l=n("./node_modules/rxjs/_esm5/internal/observable/fromIterable.js"),d=n("./node_modules/rxjs/_esm5/internal/observable/fromObservable.js"),p=n("./node_modules/rxjs/_esm5/internal/util/subscribeTo.js");function f(e,t){if(!t)return e instanceof r.Observable?e:new r.Observable(Object(p.subscribeTo)(e));if(null!=e){if(Object(a.isInteropObservable)(e))return Object(d.fromObservable)(e,t);if(Object(i.isPromise)(e))return Object(c.fromPromise)(e,t);if(Object(o.isArrayLike)(e))return Object(u.fromArray)(e,t);if(Object(s.isIterable)(e)||"string"==typeof e)return Object(l.fromIterable)(e,t)}throw new TypeError((null!==e&&typeof e||e)+" is not observable")}},"./node_modules/rxjs/_esm5/internal/observable/fromArray.js":function(e,t,n){"use strict";n.r(t),n.d(t,"fromArray",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToArray.js");function a(e,t){return t?new r.Observable(function(n){var r=new i.Subscription,o=0;return r.add(t.schedule(function(){o!==e.length?(n.next(e[o++]),n.closed||r.add(this.schedule())):n.complete()})),r}):new r.Observable(Object(o.subscribeToArray)(e))}},"./node_modules/rxjs/_esm5/internal/observable/fromEvent.js":function(e,t,n){"use strict";n.r(t),n.d(t,"fromEvent",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isFunction.js"),a=n("./node_modules/rxjs/_esm5/internal/operators/map.js");Object.prototype.toString;function s(e,t,n,u){return Object(o.isFunction)(n)&&(u=n,n=void 0),u?s(e,t,n).pipe(Object(a.map)(function(e){return Object(i.isArray)(e)?u.apply(void 0,e):u(e)})):new r.Observable(function(r){!function e(t,n,r,i,o){var a;if(h=t,h&&"function"==typeof h.addEventListener&&"function"==typeof h.removeEventListener){var s=t;t.addEventListener(n,r,o),a=function(){return s.removeEventListener(n,r,o)}}else if(f=t,f&&"function"==typeof f.on&&"function"==typeof f.off){var u=t;t.on(n,r),a=function(){return u.off(n,r)}}else if(p=t,p&&"function"==typeof p.addListener&&"function"==typeof p.removeListener){var c=t;t.addListener(n,r),a=function(){return c.removeListener(n,r)}}else{if(!t||!t.length)throw new TypeError("Invalid event target");for(var l=0,d=t.length;l<d;l++)e(t[l],n,r,i,o)}var p;var f;var h;i.add(a)}(e,t,function(e){arguments.length>1?r.next(Array.prototype.slice.call(arguments)):r.next(e)},r,n)})}},"./node_modules/rxjs/_esm5/internal/observable/fromEventPattern.js":function(e,t,n){"use strict";n.r(t),n.d(t,"fromEventPattern",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isFunction.js"),a=n("./node_modules/rxjs/_esm5/internal/operators/map.js");function s(e,t,n){return n?s(e,t).pipe(Object(a.map)(function(e){return Object(i.isArray)(e)?n.apply(void 0,e):n(e)})):new r.Observable(function(n){var r,i=function(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];return n.next(1===e.length?e[0]:e)};try{r=e(i)}catch(e){return void n.error(e)}if(Object(o.isFunction)(t))return function(){return t(i,r)}})}},"./node_modules/rxjs/_esm5/internal/observable/fromIterable.js":function(e,t,n){"use strict";n.r(t),n.d(t,"fromIterable",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),o=n("./node_modules/rxjs/_esm5/internal/symbol/iterator.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToIterable.js");function s(e,t){if(!e)throw new Error("Iterable cannot be null");return t?new r.Observable(function(n){var r,a=new i.Subscription;return a.add(function(){r&&"function"==typeof r.return&&r.return()}),a.add(t.schedule(function(){r=e[o.iterator](),a.add(t.schedule(function(){if(!n.closed){var e,t;try{var i=r.next();e=i.value,t=i.done}catch(e){return void n.error(e)}t?n.complete():(n.next(e),this.schedule())}}))})),a}):new r.Observable(Object(a.subscribeToIterable)(e))}},"./node_modules/rxjs/_esm5/internal/observable/fromObservable.js":function(e,t,n){"use strict";n.r(t),n.d(t,"fromObservable",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),o=n("./node_modules/rxjs/_esm5/internal/symbol/observable.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToObservable.js");function s(e,t){return t?new r.Observable(function(n){var r=new i.Subscription;return r.add(t.schedule(function(){var i=e[o.observable]();r.add(i.subscribe({next:function(e){r.add(t.schedule(function(){return n.next(e)}))},error:function(e){r.add(t.schedule(function(){return n.error(e)}))},complete:function(){r.add(t.schedule(function(){return n.complete()}))}}))})),r}):new r.Observable(Object(a.subscribeToObservable)(e))}},"./node_modules/rxjs/_esm5/internal/observable/fromPromise.js":function(e,t,n){"use strict";n.r(t),n.d(t,"fromPromise",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToPromise.js");function a(e,t){return t?new r.Observable(function(n){var r=new i.Subscription;return r.add(t.schedule(function(){return e.then(function(e){r.add(t.schedule(function(){n.next(e),r.add(t.schedule(function(){return n.complete()}))}))},function(e){r.add(t.schedule(function(){return n.error(e)}))})})),r}):new r.Observable(Object(o.subscribeToPromise)(e))}},"./node_modules/rxjs/_esm5/internal/observable/generate.js":function(e,t,n){"use strict";n.r(t),n.d(t,"generate",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/util/identity.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js");function a(e,t,n,a,u){var c,l;if(1==arguments.length){var d=e;l=d.initialState,t=d.condition,n=d.iterate,c=d.resultSelector||i.identity,u=d.scheduler}else void 0===a||Object(o.isScheduler)(a)?(l=e,c=i.identity,u=a):(l=e,c=a);return new r.Observable(function(e){var r=l;if(u)return u.schedule(s,0,{subscriber:e,iterate:n,condition:t,resultSelector:c,state:r});for(;;){if(t){var i=void 0;try{i=t(r)}catch(t){return void e.error(t)}if(!i){e.complete();break}}var o=void 0;try{o=c(r)}catch(t){return void e.error(t)}if(e.next(o),e.closed)break;try{r=n(r)}catch(t){return void e.error(t)}}})}function s(e){var t=e.subscriber,n=e.condition;if(!t.closed){if(e.needIterate)try{e.state=e.iterate(e.state)}catch(e){return void t.error(e)}else e.needIterate=!0;if(n){var r=void 0;try{r=n(e.state)}catch(e){return void t.error(e)}if(!r)return void t.complete();if(t.closed)return}var i;try{i=e.resultSelector(e.state)}catch(e){return void t.error(e)}if(!t.closed&&(t.next(i),!t.closed))return this.schedule(e)}}},"./node_modules/rxjs/_esm5/internal/observable/iif.js":function(e,t,n){"use strict";n.r(t),n.d(t,"iif",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/observable/defer.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/empty.js");function o(e,t,n){return void 0===t&&(t=i.EMPTY),void 0===n&&(n=i.EMPTY),Object(r.defer)(function(){return e()?t:n})}},"./node_modules/rxjs/_esm5/internal/observable/interval.js":function(e,t,n){"use strict";n.r(t),n.d(t,"interval",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isNumeric.js");function a(e,t){return void 0===e&&(e=0),void 0===t&&(t=i.async),(!Object(o.isNumeric)(e)||e<0)&&(e=0),t&&"function"==typeof t.schedule||(t=i.async),new r.Observable(function(n){return n.add(t.schedule(s,e,{subscriber:n,counter:0,period:e})),n})}function s(e){var t=e.subscriber,n=e.counter,r=e.period;t.next(n),this.schedule({subscriber:t,counter:n+1,period:r},r)}},"./node_modules/rxjs/_esm5/internal/observable/merge.js":function(e,t,n){"use strict";n.r(t),n.d(t,"merge",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js"),o=n("./node_modules/rxjs/_esm5/internal/operators/mergeAll.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/fromArray.js");function s(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];var n=Number.POSITIVE_INFINITY,s=null,u=e[e.length-1];return Object(i.isScheduler)(u)?(s=e.pop(),e.length>1&&"number"==typeof e[e.length-1]&&(n=e.pop())):"number"==typeof u&&(n=e.pop()),null===s&&1===e.length&&e[0]instanceof r.Observable?e[0]:Object(o.mergeAll)(n)(Object(a.fromArray)(e,s))}},"./node_modules/rxjs/_esm5/internal/observable/never.js":function(e,t,n){"use strict";n.r(t),n.d(t,"NEVER",function(){return o}),n.d(t,"never",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/util/noop.js"),o=new r.Observable(i.noop);function a(){return o}},"./node_modules/rxjs/_esm5/internal/observable/of.js":function(e,t,n){"use strict";n.r(t),n.d(t,"of",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/fromArray.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/empty.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/scalar.js");function s(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];var n=e[e.length-1];switch(Object(r.isScheduler)(n)?e.pop():n=void 0,e.length){case 0:return Object(o.empty)(n);case 1:return n?Object(i.fromArray)(e,n):Object(a.scalar)(e[0]);default:return Object(i.fromArray)(e,n)}}},"./node_modules/rxjs/_esm5/internal/observable/onErrorResumeNext.js":function(e,t,n){"use strict";n.r(t),n.d(t,"onErrorResumeNext",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/from.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/empty.js");function s(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];if(0===e.length)return a.EMPTY;var n=e[0],u=e.slice(1);return 1===e.length&&Object(o.isArray)(n)?s.apply(void 0,n):new r.Observable(function(e){var t=function(){return e.add(s.apply(void 0,u).subscribe(e))};return Object(i.from)(n).subscribe({next:function(t){e.next(t)},error:t,complete:t})})}},"./node_modules/rxjs/_esm5/internal/observable/pairs.js":function(e,t,n){"use strict";n.r(t),n.d(t,"pairs",function(){return o}),n.d(t,"dispatch",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscription.js");function o(e,t){return t?new r.Observable(function(n){var r=Object.keys(e),o=new i.Subscription;return o.add(t.schedule(a,0,{keys:r,index:0,subscriber:n,subscription:o,obj:e})),o}):new r.Observable(function(t){for(var n=Object.keys(e),r=0;r<n.length&&!t.closed;r++){var i=n[r];e.hasOwnProperty(i)&&t.next([i,e[i]])}t.complete()})}function a(e){var t=e.keys,n=e.index,r=e.subscriber,i=e.subscription,o=e.obj;if(!r.closed)if(n<t.length){var a=t[n];r.next([a,o[a]]),i.add(this.schedule({keys:t,index:n+1,subscriber:r,subscription:i,obj:o}))}else r.complete()}},"./node_modules/rxjs/_esm5/internal/observable/race.js":function(e,t,n){"use strict";n.r(t),n.d(t,"race",function(){return u}),n.d(t,"RaceOperator",function(){return c}),n.d(t,"RaceSubscriber",function(){return l});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/fromArray.js"),a=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function u(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];if(1===e.length){if(!Object(i.isArray)(e[0]))return e[0];e=e[0]}return Object(o.fromArray)(e,void 0).lift(new c)}var c=function(){function e(){}return e.prototype.call=function(e,t){return t.subscribe(new l(e))},e}(),l=function(e){function t(t){var n=e.call(this,t)||this;return n.hasFirst=!1,n.observables=[],n.subscriptions=[],n}return r.__extends(t,e),t.prototype._next=function(e){this.observables.push(e)},t.prototype._complete=function(){var e=this.observables,t=e.length;if(0===t)this.destination.complete();else{for(var n=0;n<t&&!this.hasFirst;n++){var r=e[n],i=Object(s.subscribeToResult)(this,r,r,n);this.subscriptions&&this.subscriptions.push(i),this.add(i)}this.observables=null}},t.prototype.notifyNext=function(e,t,n,r,i){if(!this.hasFirst){this.hasFirst=!0;for(var o=0;o<this.subscriptions.length;o++)if(o!==n){var a=this.subscriptions[o];a.unsubscribe(),this.remove(a)}this.subscriptions=null}this.destination.next(t)},t}(a.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/observable/range.js":function(e,t,n){"use strict";n.r(t),n.d(t,"range",function(){return i}),n.d(t,"dispatch",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js");function i(e,t,n){return void 0===e&&(e=0),void 0===t&&(t=0),new r.Observable(function(r){var i=0,a=e;if(n)return n.schedule(o,0,{index:i,count:t,start:e,subscriber:r});for(;;){if(i++>=t){r.complete();break}if(r.next(a++),r.closed)break}})}function o(e){var t=e.start,n=e.index,r=e.count,i=e.subscriber;n>=r?i.complete():(i.next(t),i.closed||(e.index=n+1,e.start=t+1,this.schedule(e)))}},"./node_modules/rxjs/_esm5/internal/observable/scalar.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scalar",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js");function i(e){var t=new r.Observable(function(t){t.next(e),t.complete()});return t._isScalar=!0,t.value=e,t}},"./node_modules/rxjs/_esm5/internal/observable/throwError.js":function(e,t,n){"use strict";n.r(t),n.d(t,"throwError",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js");function i(e,t){return t?new r.Observable(function(n){return t.schedule(o,0,{error:e,subscriber:n})}):new r.Observable(function(t){return t.error(e)})}function o(e){var t=e.error;e.subscriber.error(t)}},"./node_modules/rxjs/_esm5/internal/observable/timer.js":function(e,t,n){"use strict";n.r(t),n.d(t,"timer",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isNumeric.js"),a=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js");function s(e,t,n){void 0===e&&(e=0);var s=-1;return Object(o.isNumeric)(t)?s=Number(t)<1?1:Number(t):Object(a.isScheduler)(t)&&(n=t),Object(a.isScheduler)(n)||(n=i.async),new r.Observable(function(t){var r=Object(o.isNumeric)(e)?e:+e-n.now();return n.schedule(u,r,{index:0,period:s,subscriber:t})})}function u(e){var t=e.index,n=e.period,r=e.subscriber;if(r.next(t),!r.closed){if(-1===n)return r.complete();e.index=t+1,this.schedule(e,n)}}},"./node_modules/rxjs/_esm5/internal/observable/using.js":function(e,t,n){"use strict";n.r(t),n.d(t,"using",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/from.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/empty.js");function a(e,t){return new r.Observable(function(n){var r,a;try{r=e()}catch(e){return void n.error(e)}try{a=t(r)}catch(e){return void n.error(e)}var s=(a?Object(i.from)(a):o.EMPTY).subscribe(n);return function(){s.unsubscribe(),r&&r.unsubscribe()}})}},"./node_modules/rxjs/_esm5/internal/observable/zip.js":function(e,t,n){"use strict";n.r(t),n.d(t,"zip",function(){return l}),n.d(t,"ZipOperator",function(){return d}),n.d(t,"ZipSubscriber",function(){return p});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/fromArray.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),a=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),u=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),c=n("./node_modules/rxjs/_esm5/internal/symbol/iterator.js");function l(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];var n=e[e.length-1];return"function"==typeof n&&e.pop(),Object(i.fromArray)(e,void 0).lift(new d(n))}var d=function(){function e(e){this.resultSelector=e}return e.prototype.call=function(e,t){return t.subscribe(new p(e,this.resultSelector))},e}(),p=function(e){function t(t,n,r){void 0===r&&(r=Object.create(null));var i=e.call(this,t)||this;return i.iterators=[],i.active=0,i.resultSelector="function"==typeof n?n:null,i.values=r,i}return r.__extends(t,e),t.prototype._next=function(e){var t=this.iterators;Object(o.isArray)(e)?t.push(new h(e)):"function"==typeof e[c.iterator]?t.push(new f(e[c.iterator]())):t.push(new m(this.destination,this,e))},t.prototype._complete=function(){var e=this.iterators,t=e.length;if(this.unsubscribe(),0!==t){this.active=t;for(var n=0;n<t;n++){var r=e[n];if(r.stillUnsubscribed)this.destination.add(r.subscribe(r,n));else this.active--}}else this.destination.complete()},t.prototype.notifyInactive=function(){this.active--,0===this.active&&this.destination.complete()},t.prototype.checkIterators=function(){for(var e=this.iterators,t=e.length,n=this.destination,r=0;r<t;r++){if("function"==typeof(a=e[r]).hasValue&&!a.hasValue())return}var i=!1,o=[];for(r=0;r<t;r++){var a,s=(a=e[r]).next();if(a.hasCompleted()&&(i=!0),s.done)return void n.complete();o.push(s.value)}this.resultSelector?this._tryresultSelector(o):n.next(o),i&&n.complete()},t.prototype._tryresultSelector=function(e){var t;try{t=this.resultSelector.apply(this,e)}catch(e){return void this.destination.error(e)}this.destination.next(t)},t}(a.Subscriber),f=function(){function e(e){this.iterator=e,this.nextResult=e.next()}return e.prototype.hasValue=function(){return!0},e.prototype.next=function(){var e=this.nextResult;return this.nextResult=this.iterator.next(),e},e.prototype.hasCompleted=function(){var e=this.nextResult;return e&&e.done},e}(),h=function(){function e(e){this.array=e,this.index=0,this.length=0,this.length=e.length}return e.prototype[c.iterator]=function(){return this},e.prototype.next=function(e){var t=this.index++,n=this.array;return t<this.length?{value:n[t],done:!1}:{value:null,done:!0}},e.prototype.hasValue=function(){return this.array.length>this.index},e.prototype.hasCompleted=function(){return this.array.length===this.index},e}(),m=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.parent=n,i.observable=r,i.stillUnsubscribed=!0,i.buffer=[],i.isComplete=!1,i}return r.__extends(t,e),t.prototype[c.iterator]=function(){return this},t.prototype.next=function(){var e=this.buffer;return 0===e.length&&this.isComplete?{value:null,done:!0}:{value:e.shift(),done:!1}},t.prototype.hasValue=function(){return this.buffer.length>0},t.prototype.hasCompleted=function(){return 0===this.buffer.length&&this.isComplete},t.prototype.notifyComplete=function(){this.buffer.length>0?(this.isComplete=!0,this.parent.notifyInactive()):this.destination.complete()},t.prototype.notifyNext=function(e,t,n,r,i){this.buffer.push(t),this.parent.checkIterators()},t.prototype.subscribe=function(e,t){return Object(u.subscribeToResult)(this,this.observable,this,t)},t}(s.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/audit.js":function(e,t,n){"use strict";n.r(t),n.d(t,"audit",function(){return u});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),o=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),a=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function u(e){return function(t){return t.lift(new c(e))}}var c=function(){function e(e){this.durationSelector=e}return e.prototype.call=function(e,t){return t.subscribe(new l(e,this.durationSelector))},e}(),l=function(e){function t(t,n){var r=e.call(this,t)||this;return r.durationSelector=n,r.hasValue=!1,r}return r.__extends(t,e),t.prototype._next=function(e){if(this.value=e,this.hasValue=!0,!this.throttled){var t=Object(i.tryCatch)(this.durationSelector)(e);if(t===o.errorObject)this.destination.error(o.errorObject.e);else{var n=Object(s.subscribeToResult)(this,t);!n||n.closed?this.clearThrottle():this.add(this.throttled=n)}}},t.prototype.clearThrottle=function(){var e=this.value,t=this.hasValue,n=this.throttled;n&&(this.remove(n),this.throttled=null,n.unsubscribe()),t&&(this.value=null,this.hasValue=!1,this.destination.next(e))},t.prototype.notifyNext=function(e,t,n,r){this.clearThrottle()},t.prototype.notifyComplete=function(){this.clearThrottle()},t}(a.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/auditTime.js":function(e,t,n){"use strict";n.r(t),n.d(t,"auditTime",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/audit.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/timer.js");function a(e,t){return void 0===t&&(t=r.async),Object(i.audit)(function(){return Object(o.timer)(e,t)})}},"./node_modules/rxjs/_esm5/internal/operators/buffer.js":function(e,t,n){"use strict";n.r(t),n.d(t,"buffer",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function a(e){return function(t){return t.lift(new s(e))}}var s=function(){function e(e){this.closingNotifier=e}return e.prototype.call=function(e,t){return t.subscribe(new u(e,this.closingNotifier))},e}(),u=function(e){function t(t,n){var r=e.call(this,t)||this;return r.buffer=[],r.add(Object(o.subscribeToResult)(r,n)),r}return r.__extends(t,e),t.prototype._next=function(e){this.buffer.push(e)},t.prototype.notifyNext=function(e,t,n,r,i){var o=this.buffer;this.buffer=[],this.destination.next(o)},t}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/bufferCount.js":function(e,t,n){"use strict";n.r(t),n.d(t,"bufferCount",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(e,t){return void 0===t&&(t=null),function(n){return n.lift(new a(e,t))}}var a=function(){function e(e,t){this.bufferSize=e,this.startBufferEvery=t,this.subscriberClass=t&&e!==t?u:s}return e.prototype.call=function(e,t){return t.subscribe(new this.subscriberClass(e,this.bufferSize,this.startBufferEvery))},e}(),s=function(e){function t(t,n){var r=e.call(this,t)||this;return r.bufferSize=n,r.buffer=[],r}return r.__extends(t,e),t.prototype._next=function(e){var t=this.buffer;t.push(e),t.length==this.bufferSize&&(this.destination.next(t),this.buffer=[])},t.prototype._complete=function(){var t=this.buffer;t.length>0&&this.destination.next(t),e.prototype._complete.call(this)},t}(i.Subscriber),u=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.bufferSize=n,i.startBufferEvery=r,i.buffers=[],i.count=0,i}return r.__extends(t,e),t.prototype._next=function(e){var t=this.bufferSize,n=this.startBufferEvery,r=this.buffers,i=this.count;this.count++,i%n==0&&r.push([]);for(var o=r.length;o--;){var a=r[o];a.push(e),a.length===t&&(r.splice(o,1),this.destination.next(a))}},t.prototype._complete=function(){for(var t=this.buffers,n=this.destination;t.length>0;){var r=t.shift();r.length>0&&n.next(r)}e.prototype._complete.call(this)},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/bufferTime.js":function(e,t,n){"use strict";n.r(t),n.d(t,"bufferTime",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),o=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),a=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js");function s(e){var t=arguments.length,n=i.async;Object(a.isScheduler)(arguments[arguments.length-1])&&(n=arguments[arguments.length-1],t--);var r=null;t>=2&&(r=arguments[1]);var o=Number.POSITIVE_INFINITY;return t>=3&&(o=arguments[2]),function(t){return t.lift(new u(e,r,o,n))}}var u=function(){function e(e,t,n,r){this.bufferTimeSpan=e,this.bufferCreationInterval=t,this.maxBufferSize=n,this.scheduler=r}return e.prototype.call=function(e,t){return t.subscribe(new l(e,this.bufferTimeSpan,this.bufferCreationInterval,this.maxBufferSize,this.scheduler))},e}(),c=function(){return function(){this.buffer=[]}}(),l=function(e){function t(t,n,r,i,o){var a=e.call(this,t)||this;a.bufferTimeSpan=n,a.bufferCreationInterval=r,a.maxBufferSize=i,a.scheduler=o,a.contexts=[];var s=a.openContext();if(a.timespanOnly=null==r||r<0,a.timespanOnly){var u={subscriber:a,context:s,bufferTimeSpan:n};a.add(s.closeAction=o.schedule(d,n,u))}else{var c={subscriber:a,context:s},l={bufferTimeSpan:n,bufferCreationInterval:r,subscriber:a,scheduler:o};a.add(s.closeAction=o.schedule(f,n,c)),a.add(o.schedule(p,r,l))}return a}return r.__extends(t,e),t.prototype._next=function(e){for(var t,n=this.contexts,r=n.length,i=0;i<r;i++){var o=n[i],a=o.buffer;a.push(e),a.length==this.maxBufferSize&&(t=o)}t&&this.onBufferFull(t)},t.prototype._error=function(t){this.contexts.length=0,e.prototype._error.call(this,t)},t.prototype._complete=function(){for(var t=this.contexts,n=this.destination;t.length>0;){var r=t.shift();n.next(r.buffer)}e.prototype._complete.call(this)},t.prototype._unsubscribe=function(){this.contexts=null},t.prototype.onBufferFull=function(e){this.closeContext(e);var t=e.closeAction;if(t.unsubscribe(),this.remove(t),!this.closed&&this.timespanOnly){e=this.openContext();var n=this.bufferTimeSpan,r={subscriber:this,context:e,bufferTimeSpan:n};this.add(e.closeAction=this.scheduler.schedule(d,n,r))}},t.prototype.openContext=function(){var e=new c;return this.contexts.push(e),e},t.prototype.closeContext=function(e){this.destination.next(e.buffer);var t=this.contexts;(t?t.indexOf(e):-1)>=0&&t.splice(t.indexOf(e),1)},t}(o.Subscriber);function d(e){var t=e.subscriber,n=e.context;n&&t.closeContext(n),t.closed||(e.context=t.openContext(),e.context.closeAction=this.schedule(e,e.bufferTimeSpan))}function p(e){var t=e.bufferCreationInterval,n=e.bufferTimeSpan,r=e.subscriber,i=e.scheduler,o=r.openContext();r.closed||(r.add(o.closeAction=i.schedule(f,n,{subscriber:r,context:o})),this.schedule(e,t))}function f(e){var t=e.subscriber,n=e.context;t.closeContext(n)}},"./node_modules/rxjs/_esm5/internal/operators/bufferToggle.js":function(e,t,n){"use strict";n.r(t),n.d(t,"bufferToggle",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),a=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js");function s(e,t){return function(n){return n.lift(new u(e,t))}}var u=function(){function e(e,t){this.openings=e,this.closingSelector=t}return e.prototype.call=function(e,t){return t.subscribe(new c(e,this.openings,this.closingSelector))},e}(),c=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.openings=n,i.closingSelector=r,i.contexts=[],i.add(Object(o.subscribeToResult)(i,n)),i}return r.__extends(t,e),t.prototype._next=function(e){for(var t=this.contexts,n=t.length,r=0;r<n;r++)t[r].buffer.push(e)},t.prototype._error=function(t){for(var n=this.contexts;n.length>0;){var r=n.shift();r.subscription.unsubscribe(),r.buffer=null,r.subscription=null}this.contexts=null,e.prototype._error.call(this,t)},t.prototype._complete=function(){for(var t=this.contexts;t.length>0;){var n=t.shift();this.destination.next(n.buffer),n.subscription.unsubscribe(),n.buffer=null,n.subscription=null}this.contexts=null,e.prototype._complete.call(this)},t.prototype.notifyNext=function(e,t,n,r,i){e?this.closeBuffer(e):this.openBuffer(t)},t.prototype.notifyComplete=function(e){this.closeBuffer(e.context)},t.prototype.openBuffer=function(e){try{var t=this.closingSelector.call(this,e);t&&this.trySubscribe(t)}catch(e){this._error(e)}},t.prototype.closeBuffer=function(e){var t=this.contexts;if(t&&e){var n=e.buffer,r=e.subscription;this.destination.next(n),t.splice(t.indexOf(e),1),this.remove(r),r.unsubscribe()}},t.prototype.trySubscribe=function(e){var t=this.contexts,n=new i.Subscription,r={buffer:[],subscription:n};t.push(r);var a=Object(o.subscribeToResult)(this,e,r);!a||a.closed?this.closeBuffer(r):(a.context=r,this.add(a),n.add(a))},t}(a.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/bufferWhen.js":function(e,t,n){"use strict";n.r(t),n.d(t,"bufferWhen",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),o=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),a=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),s=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),u=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function c(e){return function(t){return t.lift(new l(e))}}var l=function(){function e(e){this.closingSelector=e}return e.prototype.call=function(e,t){return t.subscribe(new d(e,this.closingSelector))},e}(),d=function(e){function t(t,n){var r=e.call(this,t)||this;return r.closingSelector=n,r.subscribing=!1,r.openBuffer(),r}return r.__extends(t,e),t.prototype._next=function(e){this.buffer.push(e)},t.prototype._complete=function(){var t=this.buffer;t&&this.destination.next(t),e.prototype._complete.call(this)},t.prototype._unsubscribe=function(){this.buffer=null,this.subscribing=!1},t.prototype.notifyNext=function(e,t,n,r,i){this.openBuffer()},t.prototype.notifyComplete=function(){this.subscribing?this.complete():this.openBuffer()},t.prototype.openBuffer=function(){var e=this.closingSubscription;e&&(this.remove(e),e.unsubscribe());var t=this.buffer;this.buffer&&this.destination.next(t),this.buffer=[];var n=Object(o.tryCatch)(this.closingSelector)();n===a.errorObject?this.error(a.errorObject.e):(e=new i.Subscription,this.closingSubscription=e,this.add(e),this.subscribing=!0,e.add(Object(u.subscribeToResult)(this,n)),this.subscribing=!1)},t}(s.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/catchError.js":function(e,t,n){"use strict";n.r(t),n.d(t,"catchError",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/InnerSubscriber.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function s(e){return function(t){var n=new u(e),r=t.lift(n);return n.caught=r}}var u=function(){function e(e){this.selector=e}return e.prototype.call=function(e,t){return t.subscribe(new c(e,this.selector,this.caught))},e}(),c=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.selector=n,i.caught=r,i}return r.__extends(t,e),t.prototype.error=function(t){if(!this.isStopped){var n=void 0;try{n=this.selector(t,this.caught)}catch(t){return void e.prototype.error.call(this,t)}this._unsubscribeAndRecycle();var r=new o.InnerSubscriber(this,void 0,void 0);this.add(r),Object(a.subscribeToResult)(this,n,void 0,void 0,r)}},t}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/combineAll.js":function(e,t,n){"use strict";n.r(t),n.d(t,"combineAll",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/observable/combineLatest.js");function i(e){return function(t){return t.lift(new r.CombineLatestOperator(e))}}},"./node_modules/rxjs/_esm5/internal/operators/combineLatest.js":function(e,t,n){"use strict";n.r(t),n.d(t,"combineLatest",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/combineLatest.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/from.js");function a(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];var n=null;return"function"==typeof e[e.length-1]&&(n=e.pop()),1===e.length&&Object(r.isArray)(e[0])&&(e=e[0].slice()),function(t){return t.lift.call(Object(o.from)([t].concat(e)),new i.CombineLatestOperator(n))}}},"./node_modules/rxjs/_esm5/internal/operators/concat.js":function(e,t,n){"use strict";n.r(t),n.d(t,"concat",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/observable/concat.js");function i(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];return function(t){return t.lift.call(r.concat.apply(void 0,[t].concat(e)))}}},"./node_modules/rxjs/_esm5/internal/operators/concatAll.js":function(e,t,n){"use strict";n.r(t),n.d(t,"concatAll",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/mergeAll.js");function i(){return Object(r.mergeAll)(1)}},"./node_modules/rxjs/_esm5/internal/operators/concatMap.js":function(e,t,n){"use strict";n.r(t),n.d(t,"concatMap",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/mergeMap.js");function i(e,t){return Object(r.mergeMap)(e,t,1)}},"./node_modules/rxjs/_esm5/internal/operators/concatMapTo.js":function(e,t,n){"use strict";n.r(t),n.d(t,"concatMapTo",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/concatMap.js");function i(e,t){return Object(r.concatMap)(function(){return e},t)}},"./node_modules/rxjs/_esm5/internal/operators/count.js":function(e,t,n){"use strict";n.r(t),n.d(t,"count",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(e){return function(t){return t.lift(new a(e,t))}}var a=function(){function e(e,t){this.predicate=e,this.source=t}return e.prototype.call=function(e,t){return t.subscribe(new s(e,this.predicate,this.source))},e}(),s=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.predicate=n,i.source=r,i.count=0,i.index=0,i}return r.__extends(t,e),t.prototype._next=function(e){this.predicate?this._tryPredicate(e):this.count++},t.prototype._tryPredicate=function(e){var t;try{t=this.predicate(e,this.index++,this.source)}catch(e){return void this.destination.error(e)}t&&this.count++},t.prototype._complete=function(){this.destination.next(this.count),this.destination.complete()},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/debounce.js":function(e,t,n){"use strict";n.r(t),n.d(t,"debounce",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function a(e){return function(t){return t.lift(new s(e))}}var s=function(){function e(e){this.durationSelector=e}return e.prototype.call=function(e,t){return t.subscribe(new u(e,this.durationSelector))},e}(),u=function(e){function t(t,n){var r=e.call(this,t)||this;return r.durationSelector=n,r.hasValue=!1,r.durationSubscription=null,r}return r.__extends(t,e),t.prototype._next=function(e){try{var t=this.durationSelector.call(this,e);t&&this._tryNext(e,t)}catch(e){this.destination.error(e)}},t.prototype._complete=function(){this.emitValue(),this.destination.complete()},t.prototype._tryNext=function(e,t){var n=this.durationSubscription;this.value=e,this.hasValue=!0,n&&(n.unsubscribe(),this.remove(n)),(n=Object(o.subscribeToResult)(this,t))&&!n.closed&&this.add(this.durationSubscription=n)},t.prototype.notifyNext=function(e,t,n,r,i){this.emitValue()},t.prototype.notifyComplete=function(){this.emitValue()},t.prototype.emitValue=function(){if(this.hasValue){var t=this.value,n=this.durationSubscription;n&&(this.durationSubscription=null,n.unsubscribe(),this.remove(n)),this.value=null,this.hasValue=!1,e.prototype._next.call(this,t)}},t}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/debounceTime.js":function(e,t,n){"use strict";n.r(t),n.d(t,"debounceTime",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js");function a(e,t){return void 0===t&&(t=o.async),function(n){return n.lift(new s(e,t))}}var s=function(){function e(e,t){this.dueTime=e,this.scheduler=t}return e.prototype.call=function(e,t){return t.subscribe(new u(e,this.dueTime,this.scheduler))},e}(),u=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.dueTime=n,i.scheduler=r,i.debouncedSubscription=null,i.lastValue=null,i.hasValue=!1,i}return r.__extends(t,e),t.prototype._next=function(e){this.clearDebounce(),this.lastValue=e,this.hasValue=!0,this.add(this.debouncedSubscription=this.scheduler.schedule(c,this.dueTime,this))},t.prototype._complete=function(){this.debouncedNext(),this.destination.complete()},t.prototype.debouncedNext=function(){if(this.clearDebounce(),this.hasValue){var e=this.lastValue;this.lastValue=null,this.hasValue=!1,this.destination.next(e)}},t.prototype.clearDebounce=function(){var e=this.debouncedSubscription;null!==e&&(this.remove(e),e.unsubscribe(),this.debouncedSubscription=null)},t}(i.Subscriber);function c(e){e.debouncedNext()}},"./node_modules/rxjs/_esm5/internal/operators/defaultIfEmpty.js":function(e,t,n){"use strict";n.r(t),n.d(t,"defaultIfEmpty",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(e){return void 0===e&&(e=null),function(t){return t.lift(new a(e))}}var a=function(){function e(e){this.defaultValue=e}return e.prototype.call=function(e,t){return t.subscribe(new s(e,this.defaultValue))},e}(),s=function(e){function t(t,n){var r=e.call(this,t)||this;return r.defaultValue=n,r.isEmpty=!0,r}return r.__extends(t,e),t.prototype._next=function(e){this.isEmpty=!1,this.destination.next(e)},t.prototype._complete=function(){this.isEmpty&&this.destination.next(this.defaultValue),this.destination.complete()},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/delay.js":function(e,t,n){"use strict";n.r(t),n.d(t,"delay",function(){return u});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isDate.js"),a=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/Notification.js");function u(e,t){void 0===t&&(t=i.async);var n=Object(o.isDate)(e)?+e-t.now():Math.abs(e);return function(e){return e.lift(new c(n,t))}}var c=function(){function e(e,t){this.delay=e,this.scheduler=t}return e.prototype.call=function(e,t){return t.subscribe(new l(e,this.delay,this.scheduler))},e}(),l=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.delay=n,i.scheduler=r,i.queue=[],i.active=!1,i.errored=!1,i}return r.__extends(t,e),t.dispatch=function(e){for(var t=e.source,n=t.queue,r=e.scheduler,i=e.destination;n.length>0&&n[0].time-r.now()<=0;)n.shift().notification.observe(i);if(n.length>0){var o=Math.max(0,n[0].time-r.now());this.schedule(e,o)}else this.unsubscribe(),t.active=!1},t.prototype._schedule=function(e){this.active=!0,this.destination.add(e.schedule(t.dispatch,this.delay,{source:this,destination:this.destination,scheduler:e}))},t.prototype.scheduleNotification=function(e){if(!0!==this.errored){var t=this.scheduler,n=new d(t.now()+this.delay,e);this.queue.push(n),!1===this.active&&this._schedule(t)}},t.prototype._next=function(e){this.scheduleNotification(s.Notification.createNext(e))},t.prototype._error=function(e){this.errored=!0,this.queue=[],this.destination.error(e),this.unsubscribe()},t.prototype._complete=function(){this.scheduleNotification(s.Notification.createComplete()),this.unsubscribe()},t}(a.Subscriber),d=function(){return function(e,t){this.time=e,this.notification=t}}()},"./node_modules/rxjs/_esm5/internal/operators/delayWhen.js":function(e,t,n){"use strict";n.r(t),n.d(t,"delayWhen",function(){return u});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/Observable.js"),a=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function u(e,t){return t?function(n){return new d(n,t).lift(new c(e))}:function(t){return t.lift(new c(e))}}var c=function(){function e(e){this.delayDurationSelector=e}return e.prototype.call=function(e,t){return t.subscribe(new l(e,this.delayDurationSelector))},e}(),l=function(e){function t(t,n){var r=e.call(this,t)||this;return r.delayDurationSelector=n,r.completed=!1,r.delayNotifierSubscriptions=[],r.index=0,r}return r.__extends(t,e),t.prototype.notifyNext=function(e,t,n,r,i){this.destination.next(e),this.removeSubscription(i),this.tryComplete()},t.prototype.notifyError=function(e,t){this._error(e)},t.prototype.notifyComplete=function(e){var t=this.removeSubscription(e);t&&this.destination.next(t),this.tryComplete()},t.prototype._next=function(e){var t=this.index++;try{var n=this.delayDurationSelector(e,t);n&&this.tryDelay(n,e)}catch(e){this.destination.error(e)}},t.prototype._complete=function(){this.completed=!0,this.tryComplete(),this.unsubscribe()},t.prototype.removeSubscription=function(e){e.unsubscribe();var t=this.delayNotifierSubscriptions.indexOf(e);return-1!==t&&this.delayNotifierSubscriptions.splice(t,1),e.outerValue},t.prototype.tryDelay=function(e,t){var n=Object(s.subscribeToResult)(this,e,t);n&&!n.closed&&(this.destination.add(n),this.delayNotifierSubscriptions.push(n))},t.prototype.tryComplete=function(){this.completed&&0===this.delayNotifierSubscriptions.length&&this.destination.complete()},t}(a.OuterSubscriber),d=function(e){function t(t,n){var r=e.call(this)||this;return r.source=t,r.subscriptionDelay=n,r}return r.__extends(t,e),t.prototype._subscribe=function(e){this.subscriptionDelay.subscribe(new p(e,this.source))},t}(o.Observable),p=function(e){function t(t,n){var r=e.call(this)||this;return r.parent=t,r.source=n,r.sourceSubscribed=!1,r}return r.__extends(t,e),t.prototype._next=function(e){this.subscribeToSource()},t.prototype._error=function(e){this.unsubscribe(),this.parent.error(e)},t.prototype._complete=function(){this.unsubscribe(),this.subscribeToSource()},t.prototype.subscribeToSource=function(){this.sourceSubscribed||(this.sourceSubscribed=!0,this.unsubscribe(),this.source.subscribe(this.parent))},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/dematerialize.js":function(e,t,n){"use strict";n.r(t),n.d(t,"dematerialize",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(){return function(e){return e.lift(new a)}}var a=function(){function e(){}return e.prototype.call=function(e,t){return t.subscribe(new s(e))},e}(),s=function(e){function t(t){return e.call(this,t)||this}return r.__extends(t,e),t.prototype._next=function(e){e.observe(this.destination)},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/distinct.js":function(e,t,n){"use strict";n.r(t),n.d(t,"distinct",function(){return a}),n.d(t,"DistinctSubscriber",function(){return u});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function a(e,t){return function(n){return n.lift(new s(e,t))}}var s=function(){function e(e,t){this.keySelector=e,this.flushes=t}return e.prototype.call=function(e,t){return t.subscribe(new u(e,this.keySelector,this.flushes))},e}(),u=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.keySelector=n,i.values=new Set,r&&i.add(Object(o.subscribeToResult)(i,r)),i}return r.__extends(t,e),t.prototype.notifyNext=function(e,t,n,r,i){this.values.clear()},t.prototype.notifyError=function(e,t){this._error(e)},t.prototype._next=function(e){this.keySelector?this._useKeySelector(e):this._finalizeNext(e,e)},t.prototype._useKeySelector=function(e){var t,n=this.destination;try{t=this.keySelector(e)}catch(e){return void n.error(e)}this._finalizeNext(t,e)},t.prototype._finalizeNext=function(e,t){var n=this.values;n.has(e)||(n.add(e),this.destination.next(t))},t}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/distinctUntilChanged.js":function(e,t,n){"use strict";n.r(t),n.d(t,"distinctUntilChanged",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),a=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js");function s(e,t){return function(n){return n.lift(new u(e,t))}}var u=function(){function e(e,t){this.compare=e,this.keySelector=t}return e.prototype.call=function(e,t){return t.subscribe(new c(e,this.compare,this.keySelector))},e}(),c=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.keySelector=r,i.hasKey=!1,"function"==typeof n&&(i.compare=n),i}return r.__extends(t,e),t.prototype.compare=function(e,t){return e===t},t.prototype._next=function(e){var t=e;if(this.keySelector&&(t=Object(o.tryCatch)(this.keySelector)(e))===a.errorObject)return this.destination.error(a.errorObject.e);var n=!1;if(this.hasKey){if((n=Object(o.tryCatch)(this.compare)(this.key,t))===a.errorObject)return this.destination.error(a.errorObject.e)}else this.hasKey=!0;!1===Boolean(n)&&(this.key=t,this.destination.next(e))},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/distinctUntilKeyChanged.js":function(e,t,n){"use strict";n.r(t),n.d(t,"distinctUntilKeyChanged",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/distinctUntilChanged.js");function i(e,t){return Object(r.distinctUntilChanged)(function(n,r){return t?t(n[e],r[e]):n[e]===r[e]})}},"./node_modules/rxjs/_esm5/internal/operators/elementAt.js":function(e,t,n){"use strict";n.r(t),n.d(t,"elementAt",function(){return u});var r=n("./node_modules/rxjs/_esm5/internal/util/ArgumentOutOfRangeError.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/filter.js"),o=n("./node_modules/rxjs/_esm5/internal/operators/throwIfEmpty.js"),a=n("./node_modules/rxjs/_esm5/internal/operators/defaultIfEmpty.js"),s=n("./node_modules/rxjs/_esm5/internal/operators/take.js");function u(e,t){if(e<0)throw new r.ArgumentOutOfRangeError;var n=arguments.length>=2;return function(u){return u.pipe(Object(i.filter)(function(t,n){return n===e}),Object(s.take)(1),n?Object(a.defaultIfEmpty)(t):Object(o.throwIfEmpty)(function(){return new r.ArgumentOutOfRangeError}))}}},"./node_modules/rxjs/_esm5/internal/operators/endWith.js":function(e,t,n){"use strict";n.r(t),n.d(t,"endWith",function(){return u});var r=n("./node_modules/rxjs/_esm5/internal/observable/fromArray.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/scalar.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/empty.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/concat.js"),s=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js");function u(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];return function(t){var n=e[e.length-1];Object(s.isScheduler)(n)?e.pop():n=null;var u=e.length;return 1!==u||n?u>0?Object(a.concat)(t,Object(r.fromArray)(e,n)):Object(a.concat)(t,Object(o.empty)(n)):Object(a.concat)(t,Object(i.scalar)(e[0]))}}},"./node_modules/rxjs/_esm5/internal/operators/every.js":function(e,t,n){"use strict";n.r(t),n.d(t,"every",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(e,t){return function(n){return n.lift(new a(e,t,n))}}var a=function(){function e(e,t,n){this.predicate=e,this.thisArg=t,this.source=n}return e.prototype.call=function(e,t){return t.subscribe(new s(e,this.predicate,this.thisArg,this.source))},e}(),s=function(e){function t(t,n,r,i){var o=e.call(this,t)||this;return o.predicate=n,o.thisArg=r,o.source=i,o.index=0,o.thisArg=r||o,o}return r.__extends(t,e),t.prototype.notifyComplete=function(e){this.destination.next(e),this.destination.complete()},t.prototype._next=function(e){var t=!1;try{t=this.predicate.call(this.thisArg,e,this.index++,this.source)}catch(e){return void this.destination.error(e)}t||this.notifyComplete(!1)},t.prototype._complete=function(){this.notifyComplete(!0)},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/exhaust.js":function(e,t,n){"use strict";n.r(t),n.d(t,"exhaust",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function a(){return function(e){return e.lift(new s)}}var s=function(){function e(){}return e.prototype.call=function(e,t){return t.subscribe(new u(e))},e}(),u=function(e){function t(t){var n=e.call(this,t)||this;return n.hasCompleted=!1,n.hasSubscription=!1,n}return r.__extends(t,e),t.prototype._next=function(e){this.hasSubscription||(this.hasSubscription=!0,this.add(Object(o.subscribeToResult)(this,e)))},t.prototype._complete=function(){this.hasCompleted=!0,this.hasSubscription||this.destination.complete()},t.prototype.notifyComplete=function(e){this.remove(e),this.hasSubscription=!1,this.hasCompleted&&this.destination.complete()},t}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/exhaustMap.js":function(e,t,n){"use strict";n.r(t),n.d(t,"exhaustMap",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/InnerSubscriber.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),s=n("./node_modules/rxjs/_esm5/internal/operators/map.js"),u=n("./node_modules/rxjs/_esm5/internal/observable/from.js");function c(e,t){return t?function(n){return n.pipe(c(function(n,r){return Object(u.from)(e(n,r)).pipe(Object(s.map)(function(e,i){return t(n,e,r,i)}))}))}:function(t){return t.lift(new l(e))}}var l=function(){function e(e){this.project=e}return e.prototype.call=function(e,t){return t.subscribe(new d(e,this.project))},e}(),d=function(e){function t(t,n){var r=e.call(this,t)||this;return r.project=n,r.hasSubscription=!1,r.hasCompleted=!1,r.index=0,r}return r.__extends(t,e),t.prototype._next=function(e){this.hasSubscription||this.tryNext(e)},t.prototype.tryNext=function(e){var t,n=this.index++;try{t=this.project(e,n)}catch(e){return void this.destination.error(e)}this.hasSubscription=!0,this._innerSub(t,e,n)},t.prototype._innerSub=function(e,t,n){var r=new o.InnerSubscriber(this,void 0,void 0);this.destination.add(r),Object(a.subscribeToResult)(this,e,t,n,r)},t.prototype._complete=function(){this.hasCompleted=!0,this.hasSubscription||this.destination.complete(),this.unsubscribe()},t.prototype.notifyNext=function(e,t,n,r,i){this.destination.next(t)},t.prototype.notifyError=function(e){this.destination.error(e)},t.prototype.notifyComplete=function(e){this.destination.remove(e),this.hasSubscription=!1,this.hasCompleted&&this.destination.complete()},t}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/expand.js":function(e,t,n){"use strict";n.r(t),n.d(t,"expand",function(){return u}),n.d(t,"ExpandOperator",function(){return c}),n.d(t,"ExpandSubscriber",function(){return l});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),o=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),a=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function u(e,t,n){return void 0===t&&(t=Number.POSITIVE_INFINITY),void 0===n&&(n=void 0),t=(t||0)<1?Number.POSITIVE_INFINITY:t,function(r){return r.lift(new c(e,t,n))}}var c=function(){function e(e,t,n){this.project=e,this.concurrent=t,this.scheduler=n}return e.prototype.call=function(e,t){return t.subscribe(new l(e,this.project,this.concurrent,this.scheduler))},e}(),l=function(e){function t(t,n,r,i){var o=e.call(this,t)||this;return o.project=n,o.concurrent=r,o.scheduler=i,o.index=0,o.active=0,o.hasCompleted=!1,r<Number.POSITIVE_INFINITY&&(o.buffer=[]),o}return r.__extends(t,e),t.dispatch=function(e){var t=e.subscriber,n=e.result,r=e.value,i=e.index;t.subscribeToProjection(n,r,i)},t.prototype._next=function(e){var n=this.destination;if(n.closed)this._complete();else{var r=this.index++;if(this.active<this.concurrent){n.next(e);var a=Object(i.tryCatch)(this.project)(e,r);if(a===o.errorObject)n.error(o.errorObject.e);else if(this.scheduler){var s={subscriber:this,result:a,value:e,index:r};this.destination.add(this.scheduler.schedule(t.dispatch,0,s))}else this.subscribeToProjection(a,e,r)}else this.buffer.push(e)}},t.prototype.subscribeToProjection=function(e,t,n){this.active++,this.destination.add(Object(s.subscribeToResult)(this,e,t,n))},t.prototype._complete=function(){this.hasCompleted=!0,this.hasCompleted&&0===this.active&&this.destination.complete(),this.unsubscribe()},t.prototype.notifyNext=function(e,t,n,r,i){this._next(t)},t.prototype.notifyComplete=function(e){var t=this.buffer;this.destination.remove(e),this.active--,t&&t.length>0&&this._next(t.shift()),this.hasCompleted&&0===this.active&&this.destination.complete()},t}(a.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/filter.js":function(e,t,n){"use strict";n.r(t),n.d(t,"filter",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(e,t){return function(n){return n.lift(new a(e,t))}}var a=function(){function e(e,t){this.predicate=e,this.thisArg=t}return e.prototype.call=function(e,t){return t.subscribe(new s(e,this.predicate,this.thisArg))},e}(),s=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.predicate=n,i.thisArg=r,i.count=0,i}return r.__extends(t,e),t.prototype._next=function(e){var t;try{t=this.predicate.call(this.thisArg,e,this.count++)}catch(e){return void this.destination.error(e)}t&&this.destination.next(e)},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/finalize.js":function(e,t,n){"use strict";n.r(t),n.d(t,"finalize",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/Subscription.js");function a(e){return function(t){return t.lift(new s(e))}}var s=function(){function e(e){this.callback=e}return e.prototype.call=function(e,t){return t.subscribe(new u(e,this.callback))},e}(),u=function(e){function t(t,n){var r=e.call(this,t)||this;return r.add(new o.Subscription(n)),r}return r.__extends(t,e),t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/find.js":function(e,t,n){"use strict";n.r(t),n.d(t,"find",function(){return o}),n.d(t,"FindValueOperator",function(){return a}),n.d(t,"FindValueSubscriber",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(e,t){if("function"!=typeof e)throw new TypeError("predicate is not a function");return function(n){return n.lift(new a(e,n,!1,t))}}var a=function(){function e(e,t,n,r){this.predicate=e,this.source=t,this.yieldIndex=n,this.thisArg=r}return e.prototype.call=function(e,t){return t.subscribe(new s(e,this.predicate,this.source,this.yieldIndex,this.thisArg))},e}(),s=function(e){function t(t,n,r,i,o){var a=e.call(this,t)||this;return a.predicate=n,a.source=r,a.yieldIndex=i,a.thisArg=o,a.index=0,a}return r.__extends(t,e),t.prototype.notifyComplete=function(e){var t=this.destination;t.next(e),t.complete(),this.unsubscribe()},t.prototype._next=function(e){var t=this.predicate,n=this.thisArg,r=this.index++;try{t.call(n||this,e,r,this.source)&&this.notifyComplete(this.yieldIndex?r:e)}catch(e){this.destination.error(e)}},t.prototype._complete=function(){this.notifyComplete(this.yieldIndex?-1:void 0)},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/findIndex.js":function(e,t,n){"use strict";n.r(t),n.d(t,"findIndex",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/find.js");function i(e,t){return function(n){return n.lift(new r.FindValueOperator(e,n,!0,t))}}},"./node_modules/rxjs/_esm5/internal/operators/first.js":function(e,t,n){"use strict";n.r(t),n.d(t,"first",function(){return c});var r=n("./node_modules/rxjs/_esm5/internal/util/EmptyError.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/filter.js"),o=n("./node_modules/rxjs/_esm5/internal/operators/take.js"),a=n("./node_modules/rxjs/_esm5/internal/operators/defaultIfEmpty.js"),s=n("./node_modules/rxjs/_esm5/internal/operators/throwIfEmpty.js"),u=n("./node_modules/rxjs/_esm5/internal/util/identity.js");function c(e,t){var n=arguments.length>=2;return function(c){return c.pipe(e?Object(i.filter)(function(t,n){return e(t,n,c)}):u.identity,Object(o.take)(1),n?Object(a.defaultIfEmpty)(t):Object(s.throwIfEmpty)(function(){return new r.EmptyError}))}}},"./node_modules/rxjs/_esm5/internal/operators/groupBy.js":function(e,t,n){"use strict";n.r(t),n.d(t,"groupBy",function(){return u}),n.d(t,"GroupedObservable",function(){return p});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),a=n("./node_modules/rxjs/_esm5/internal/Observable.js"),s=n("./node_modules/rxjs/_esm5/internal/Subject.js");function u(e,t,n,r){return function(i){return i.lift(new c(e,t,n,r))}}var c=function(){function e(e,t,n,r){this.keySelector=e,this.elementSelector=t,this.durationSelector=n,this.subjectSelector=r}return e.prototype.call=function(e,t){return t.subscribe(new l(e,this.keySelector,this.elementSelector,this.durationSelector,this.subjectSelector))},e}(),l=function(e){function t(t,n,r,i,o){var a=e.call(this,t)||this;return a.keySelector=n,a.elementSelector=r,a.durationSelector=i,a.subjectSelector=o,a.groups=null,a.attemptedToUnsubscribe=!1,a.count=0,a}return r.__extends(t,e),t.prototype._next=function(e){var t;try{t=this.keySelector(e)}catch(e){return void this.error(e)}this._group(e,t)},t.prototype._group=function(e,t){var n=this.groups;n||(n=this.groups=new Map);var r,i=n.get(t);if(this.elementSelector)try{r=this.elementSelector(e)}catch(e){this.error(e)}else r=e;if(!i){i=this.subjectSelector?this.subjectSelector():new s.Subject,n.set(t,i);var o=new p(t,i,this);if(this.destination.next(o),this.durationSelector){var a=void 0;try{a=this.durationSelector(new p(t,i))}catch(e){return void this.error(e)}this.add(a.subscribe(new d(t,i,this)))}}i.closed||i.next(r)},t.prototype._error=function(e){var t=this.groups;t&&(t.forEach(function(t,n){t.error(e)}),t.clear()),this.destination.error(e)},t.prototype._complete=function(){var e=this.groups;e&&(e.forEach(function(e,t){e.complete()}),e.clear()),this.destination.complete()},t.prototype.removeGroup=function(e){this.groups.delete(e)},t.prototype.unsubscribe=function(){this.closed||(this.attemptedToUnsubscribe=!0,0===this.count&&e.prototype.unsubscribe.call(this))},t}(i.Subscriber),d=function(e){function t(t,n,r){var i=e.call(this,n)||this;return i.key=t,i.group=n,i.parent=r,i}return r.__extends(t,e),t.prototype._next=function(e){this.complete()},t.prototype._unsubscribe=function(){var e=this.parent,t=this.key;this.key=this.parent=null,e&&e.removeGroup(t)},t}(i.Subscriber),p=function(e){function t(t,n,r){var i=e.call(this)||this;return i.key=t,i.groupSubject=n,i.refCountSubscription=r,i}return r.__extends(t,e),t.prototype._subscribe=function(e){var t=new o.Subscription,n=this.refCountSubscription,r=this.groupSubject;return n&&!n.closed&&t.add(new f(n)),t.add(r.subscribe(e)),t},t}(a.Observable),f=function(e){function t(t){var n=e.call(this)||this;return n.parent=t,t.count++,n}return r.__extends(t,e),t.prototype.unsubscribe=function(){var t=this.parent;t.closed||this.closed||(e.prototype.unsubscribe.call(this),t.count-=1,0===t.count&&t.attemptedToUnsubscribe&&t.unsubscribe())},t}(o.Subscription)},"./node_modules/rxjs/_esm5/internal/operators/ignoreElements.js":function(e,t,n){"use strict";n.r(t),n.d(t,"ignoreElements",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(){return function(e){return e.lift(new a)}}var a=function(){function e(){}return e.prototype.call=function(e,t){return t.subscribe(new s(e))},e}(),s=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return r.__extends(t,e),t.prototype._next=function(e){},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/isEmpty.js":function(e,t,n){"use strict";n.r(t),n.d(t,"isEmpty",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(){return function(e){return e.lift(new a)}}var a=function(){function e(){}return e.prototype.call=function(e,t){return t.subscribe(new s(e))},e}(),s=function(e){function t(t){return e.call(this,t)||this}return r.__extends(t,e),t.prototype.notifyComplete=function(e){var t=this.destination;t.next(e),t.complete()},t.prototype._next=function(e){this.notifyComplete(!1)},t.prototype._complete=function(){this.notifyComplete(!0)},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/last.js":function(e,t,n){"use strict";n.r(t),n.d(t,"last",function(){return c});var r=n("./node_modules/rxjs/_esm5/internal/util/EmptyError.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/filter.js"),o=n("./node_modules/rxjs/_esm5/internal/operators/takeLast.js"),a=n("./node_modules/rxjs/_esm5/internal/operators/throwIfEmpty.js"),s=n("./node_modules/rxjs/_esm5/internal/operators/defaultIfEmpty.js"),u=n("./node_modules/rxjs/_esm5/internal/util/identity.js");function c(e,t){var n=arguments.length>=2;return function(c){return c.pipe(e?Object(i.filter)(function(t,n){return e(t,n,c)}):u.identity,Object(o.takeLast)(1),n?Object(s.defaultIfEmpty)(t):Object(a.throwIfEmpty)(function(){return new r.EmptyError}))}}},"./node_modules/rxjs/_esm5/internal/operators/map.js":function(e,t,n){"use strict";n.r(t),n.d(t,"map",function(){return o}),n.d(t,"MapOperator",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(e,t){return function(n){if("function"!=typeof e)throw new TypeError("argument is not a function. Are you looking for `mapTo()`?");return n.lift(new a(e,t))}}var a=function(){function e(e,t){this.project=e,this.thisArg=t}return e.prototype.call=function(e,t){return t.subscribe(new s(e,this.project,this.thisArg))},e}(),s=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.project=n,i.count=0,i.thisArg=r||i,i}return r.__extends(t,e),t.prototype._next=function(e){var t;try{t=this.project.call(this.thisArg,e,this.count++)}catch(e){return void this.destination.error(e)}this.destination.next(t)},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/mapTo.js":function(e,t,n){"use strict";n.r(t),n.d(t,"mapTo",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(e){return function(t){return t.lift(new a(e))}}var a=function(){function e(e){this.value=e}return e.prototype.call=function(e,t){return t.subscribe(new s(e,this.value))},e}(),s=function(e){function t(t,n){var r=e.call(this,t)||this;return r.value=n,r}return r.__extends(t,e),t.prototype._next=function(e){this.destination.next(this.value)},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/materialize.js":function(e,t,n){"use strict";n.r(t),n.d(t,"materialize",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/Notification.js");function a(){return function(e){return e.lift(new s)}}var s=function(){function e(){}return e.prototype.call=function(e,t){return t.subscribe(new u(e))},e}(),u=function(e){function t(t){return e.call(this,t)||this}return r.__extends(t,e),t.prototype._next=function(e){this.destination.next(o.Notification.createNext(e))},t.prototype._error=function(e){var t=this.destination;t.next(o.Notification.createError(e)),t.complete()},t.prototype._complete=function(){var e=this.destination;e.next(o.Notification.createComplete()),e.complete()},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/max.js":function(e,t,n){"use strict";n.r(t),n.d(t,"max",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/reduce.js");function i(e){var t="function"==typeof e?function(t,n){return e(t,n)>0?t:n}:function(e,t){return e>t?e:t};return Object(r.reduce)(t)}},"./node_modules/rxjs/_esm5/internal/operators/merge.js":function(e,t,n){"use strict";n.r(t),n.d(t,"merge",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/observable/merge.js");function i(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];return function(t){return t.lift.call(r.merge.apply(void 0,[t].concat(e)))}}},"./node_modules/rxjs/_esm5/internal/operators/mergeAll.js":function(e,t,n){"use strict";n.r(t),n.d(t,"mergeAll",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/operators/mergeMap.js"),i=n("./node_modules/rxjs/_esm5/internal/util/identity.js");function o(e){return void 0===e&&(e=Number.POSITIVE_INFINITY),Object(r.mergeMap)(i.identity,e)}},"./node_modules/rxjs/_esm5/internal/operators/mergeMap.js":function(e,t,n){"use strict";n.r(t),n.d(t,"mergeMap",function(){return c}),n.d(t,"MergeMapOperator",function(){return l}),n.d(t,"MergeMapSubscriber",function(){return d});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),o=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),a=n("./node_modules/rxjs/_esm5/internal/InnerSubscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/operators/map.js"),u=n("./node_modules/rxjs/_esm5/internal/observable/from.js");function c(e,t,n){return void 0===n&&(n=Number.POSITIVE_INFINITY),"function"==typeof t?function(r){return r.pipe(c(function(n,r){return Object(u.from)(e(n,r)).pipe(Object(s.map)(function(e,i){return t(n,e,r,i)}))},n))}:("number"==typeof t&&(n=t),function(t){return t.lift(new l(e,n))})}var l=function(){function e(e,t){void 0===t&&(t=Number.POSITIVE_INFINITY),this.project=e,this.concurrent=t}return e.prototype.call=function(e,t){return t.subscribe(new d(e,this.project,this.concurrent))},e}(),d=function(e){function t(t,n,r){void 0===r&&(r=Number.POSITIVE_INFINITY);var i=e.call(this,t)||this;return i.project=n,i.concurrent=r,i.hasCompleted=!1,i.buffer=[],i.active=0,i.index=0,i}return r.__extends(t,e),t.prototype._next=function(e){this.active<this.concurrent?this._tryNext(e):this.buffer.push(e)},t.prototype._tryNext=function(e){var t,n=this.index++;try{t=this.project(e,n)}catch(e){return void this.destination.error(e)}this.active++,this._innerSub(t,e,n)},t.prototype._innerSub=function(e,t,n){var r=new a.InnerSubscriber(this,void 0,void 0);this.destination.add(r),Object(i.subscribeToResult)(this,e,t,n,r)},t.prototype._complete=function(){this.hasCompleted=!0,0===this.active&&0===this.buffer.length&&this.destination.complete(),this.unsubscribe()},t.prototype.notifyNext=function(e,t,n,r,i){this.destination.next(t)},t.prototype.notifyComplete=function(e){var t=this.buffer;this.remove(e),this.active--,t.length>0?this._next(t.shift()):0===this.active&&this.hasCompleted&&this.destination.complete()},t}(o.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/mergeMapTo.js":function(e,t,n){"use strict";n.r(t),n.d(t,"mergeMapTo",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/mergeMap.js");function i(e,t,n){return void 0===n&&(n=Number.POSITIVE_INFINITY),"function"==typeof t?Object(r.mergeMap)(function(){return e},t,n):("number"==typeof t&&(n=t),Object(r.mergeMap)(function(){return e},n))}},"./node_modules/rxjs/_esm5/internal/operators/mergeScan.js":function(e,t,n){"use strict";n.r(t),n.d(t,"mergeScan",function(){return c}),n.d(t,"MergeScanOperator",function(){return l}),n.d(t,"MergeScanSubscriber",function(){return d});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),o=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),s=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),u=n("./node_modules/rxjs/_esm5/internal/InnerSubscriber.js");function c(e,t,n){return void 0===n&&(n=Number.POSITIVE_INFINITY),function(r){return r.lift(new l(e,t,n))}}var l=function(){function e(e,t,n){this.accumulator=e,this.seed=t,this.concurrent=n}return e.prototype.call=function(e,t){return t.subscribe(new d(e,this.accumulator,this.seed,this.concurrent))},e}(),d=function(e){function t(t,n,r,i){var o=e.call(this,t)||this;return o.accumulator=n,o.acc=r,o.concurrent=i,o.hasValue=!1,o.hasCompleted=!1,o.buffer=[],o.active=0,o.index=0,o}return r.__extends(t,e),t.prototype._next=function(e){if(this.active<this.concurrent){var t=this.index++,n=Object(i.tryCatch)(this.accumulator)(this.acc,e),r=this.destination;n===o.errorObject?r.error(o.errorObject.e):(this.active++,this._innerSub(n,e,t))}else this.buffer.push(e)},t.prototype._innerSub=function(e,t,n){var r=new u.InnerSubscriber(this,void 0,void 0);this.destination.add(r),Object(a.subscribeToResult)(this,e,t,n,r)},t.prototype._complete=function(){this.hasCompleted=!0,0===this.active&&0===this.buffer.length&&(!1===this.hasValue&&this.destination.next(this.acc),this.destination.complete()),this.unsubscribe()},t.prototype.notifyNext=function(e,t,n,r,i){var o=this.destination;this.acc=t,this.hasValue=!0,o.next(t)},t.prototype.notifyComplete=function(e){var t=this.buffer;this.destination.remove(e),this.active--,t.length>0?this._next(t.shift()):0===this.active&&this.hasCompleted&&(!1===this.hasValue&&this.destination.next(this.acc),this.destination.complete())},t}(s.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/min.js":function(e,t,n){"use strict";n.r(t),n.d(t,"min",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/reduce.js");function i(e){var t="function"==typeof e?function(t,n){return e(t,n)<0?t:n}:function(e,t){return e<t?e:t};return Object(r.reduce)(t)}},"./node_modules/rxjs/_esm5/internal/operators/multicast.js":function(e,t,n){"use strict";n.r(t),n.d(t,"multicast",function(){return i}),n.d(t,"MulticastOperator",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/observable/ConnectableObservable.js");function i(e,t){return function(n){var i;if(i="function"==typeof e?e:function(){return e},"function"==typeof t)return n.lift(new o(i,t));var a=Object.create(n,r.connectableObservableDescriptor);return a.source=n,a.subjectFactory=i,a}}var o=function(){function e(e,t){this.subjectFactory=e,this.selector=t}return e.prototype.call=function(e,t){var n=this.selector,r=this.subjectFactory(),i=n(r).subscribe(e);return i.add(t.subscribe(r)),i},e}()},"./node_modules/rxjs/_esm5/internal/operators/observeOn.js":function(e,t,n){"use strict";n.r(t),n.d(t,"observeOn",function(){return a}),n.d(t,"ObserveOnOperator",function(){return s}),n.d(t,"ObserveOnSubscriber",function(){return u}),n.d(t,"ObserveOnMessage",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/Notification.js");function a(e,t){return void 0===t&&(t=0),function(n){return n.lift(new s(e,t))}}var s=function(){function e(e,t){void 0===t&&(t=0),this.scheduler=e,this.delay=t}return e.prototype.call=function(e,t){return t.subscribe(new u(e,this.scheduler,this.delay))},e}(),u=function(e){function t(t,n,r){void 0===r&&(r=0);var i=e.call(this,t)||this;return i.scheduler=n,i.delay=r,i}return r.__extends(t,e),t.dispatch=function(e){var t=e.notification,n=e.destination;t.observe(n),this.unsubscribe()},t.prototype.scheduleMessage=function(e){this.destination.add(this.scheduler.schedule(t.dispatch,this.delay,new c(e,this.destination)))},t.prototype._next=function(e){this.scheduleMessage(o.Notification.createNext(e))},t.prototype._error=function(e){this.scheduleMessage(o.Notification.createError(e)),this.unsubscribe()},t.prototype._complete=function(){this.scheduleMessage(o.Notification.createComplete()),this.unsubscribe()},t}(i.Subscriber),c=function(){return function(e,t){this.notification=e,this.destination=t}}()},"./node_modules/rxjs/_esm5/internal/operators/onErrorResumeNext.js":function(e,t,n){"use strict";n.r(t),n.d(t,"onErrorResumeNext",function(){return c}),n.d(t,"onErrorResumeNextStatic",function(){return l});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/from.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),a=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/InnerSubscriber.js"),u=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function c(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];return 1===e.length&&Object(o.isArray)(e[0])&&(e=e[0]),function(t){return t.lift(new d(e))}}function l(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];var n;return 1===e.length&&Object(o.isArray)(e[0])&&(e=e[0]),n=e.shift(),Object(i.from)(n,null).lift(new d(e))}var d=function(){function e(e){this.nextSources=e}return e.prototype.call=function(e,t){return t.subscribe(new p(e,this.nextSources))},e}(),p=function(e){function t(t,n){var r=e.call(this,t)||this;return r.destination=t,r.nextSources=n,r}return r.__extends(t,e),t.prototype.notifyError=function(e,t){this.subscribeToNextSource()},t.prototype.notifyComplete=function(e){this.subscribeToNextSource()},t.prototype._error=function(e){this.subscribeToNextSource(),this.unsubscribe()},t.prototype._complete=function(){this.subscribeToNextSource(),this.unsubscribe()},t.prototype.subscribeToNextSource=function(){var e=this.nextSources.shift();if(e){var t=new s.InnerSubscriber(this,void 0,void 0);this.destination.add(t),Object(u.subscribeToResult)(this,e,void 0,void 0,t)}else this.destination.complete()},t}(a.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/pairwise.js":function(e,t,n){"use strict";n.r(t),n.d(t,"pairwise",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(){return function(e){return e.lift(new a)}}var a=function(){function e(){}return e.prototype.call=function(e,t){return t.subscribe(new s(e))},e}(),s=function(e){function t(t){var n=e.call(this,t)||this;return n.hasPrev=!1,n}return r.__extends(t,e),t.prototype._next=function(e){this.hasPrev?this.destination.next([this.prev,e]):this.hasPrev=!0,this.prev=e},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/partition.js":function(e,t,n){"use strict";n.r(t),n.d(t,"partition",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/util/not.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/filter.js");function o(e,t){return function(n){return[Object(i.filter)(e,t)(n),Object(i.filter)(Object(r.not)(e,t))(n)]}}},"./node_modules/rxjs/_esm5/internal/operators/pluck.js":function(e,t,n){"use strict";n.r(t),n.d(t,"pluck",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/map.js");function i(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];var n=e.length;if(0===n)throw new Error("list of properties cannot be empty.");return function(t){return Object(r.map)((i=e,o=n,function(e){for(var t=e,n=0;n<o;n++){var r=t[i[n]];if(void 0===r)return;t=r}return t}))(t);var i,o}}},"./node_modules/rxjs/_esm5/internal/operators/publish.js":function(e,t,n){"use strict";n.r(t),n.d(t,"publish",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/Subject.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/multicast.js");function o(e){return e?Object(i.multicast)(function(){return new r.Subject},e):Object(i.multicast)(new r.Subject)}},"./node_modules/rxjs/_esm5/internal/operators/publishBehavior.js":function(e,t,n){"use strict";n.r(t),n.d(t,"publishBehavior",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/BehaviorSubject.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/multicast.js");function o(e){return function(t){return Object(i.multicast)(new r.BehaviorSubject(e))(t)}}},"./node_modules/rxjs/_esm5/internal/operators/publishLast.js":function(e,t,n){"use strict";n.r(t),n.d(t,"publishLast",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/AsyncSubject.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/multicast.js");function o(){return function(e){return Object(i.multicast)(new r.AsyncSubject)(e)}}},"./node_modules/rxjs/_esm5/internal/operators/publishReplay.js":function(e,t,n){"use strict";n.r(t),n.d(t,"publishReplay",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/ReplaySubject.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/multicast.js");function o(e,t,n,o){n&&"function"!=typeof n&&(o=n);var a="function"==typeof n?n:void 0,s=new r.ReplaySubject(e,t,o);return function(e){return Object(i.multicast)(function(){return s},a)(e)}}},"./node_modules/rxjs/_esm5/internal/operators/race.js":function(e,t,n){"use strict";n.r(t),n.d(t,"race",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/util/isArray.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/race.js");function o(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];return function(t){return 1===e.length&&Object(r.isArray)(e[0])&&(e=e[0]),t.lift.call(i.race.apply(void 0,[t].concat(e)))}}},"./node_modules/rxjs/_esm5/internal/operators/reduce.js":function(e,t,n){"use strict";n.r(t),n.d(t,"reduce",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/operators/scan.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/takeLast.js"),o=n("./node_modules/rxjs/_esm5/internal/operators/defaultIfEmpty.js"),a=n("./node_modules/rxjs/_esm5/internal/util/pipe.js");function s(e,t){return arguments.length>=2?function(n){return Object(a.pipe)(Object(r.scan)(e,t),Object(i.takeLast)(1),Object(o.defaultIfEmpty)(t))(n)}:function(t){return Object(a.pipe)(Object(r.scan)(function(t,n,r){return e(t,n,r+1)}),Object(i.takeLast)(1))(t)}}},"./node_modules/rxjs/_esm5/internal/operators/refCount.js":function(e,t,n){"use strict";n.r(t),n.d(t,"refCount",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(){return function(e){return e.lift(new a(e))}}var a=function(){function e(e){this.connectable=e}return e.prototype.call=function(e,t){var n=this.connectable;n._refCount++;var r=new s(e,n),i=t.subscribe(r);return r.closed||(r.connection=n.connect()),i},e}(),s=function(e){function t(t,n){var r=e.call(this,t)||this;return r.connectable=n,r}return r.__extends(t,e),t.prototype._unsubscribe=function(){var e=this.connectable;if(e){this.connectable=null;var t=e._refCount;if(t<=0)this.connection=null;else if(e._refCount=t-1,t>1)this.connection=null;else{var n=this.connection,r=e._connection;this.connection=null,!r||n&&r!==n||r.unsubscribe()}}else this.connection=null},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/repeat.js":function(e,t,n){"use strict";n.r(t),n.d(t,"repeat",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/empty.js");function a(e){return void 0===e&&(e=-1),function(t){return 0===e?Object(o.empty)():e<0?t.lift(new s(-1,t)):t.lift(new s(e-1,t))}}var s=function(){function e(e,t){this.count=e,this.source=t}return e.prototype.call=function(e,t){return t.subscribe(new u(e,this.count,this.source))},e}(),u=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.count=n,i.source=r,i}return r.__extends(t,e),t.prototype.complete=function(){if(!this.isStopped){var t=this.source,n=this.count;if(0===n)return e.prototype.complete.call(this);n>-1&&(this.count=n-1),t.subscribe(this._unsubscribeAndRecycle())}},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/repeatWhen.js":function(e,t,n){"use strict";n.r(t),n.d(t,"repeatWhen",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),a=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),s=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),u=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function c(e){return function(t){return t.lift(new l(e))}}var l=function(){function e(e){this.notifier=e}return e.prototype.call=function(e,t){return t.subscribe(new d(e,this.notifier,t))},e}(),d=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.notifier=n,i.source=r,i.sourceIsBeingSubscribedTo=!0,i}return r.__extends(t,e),t.prototype.notifyNext=function(e,t,n,r,i){this.sourceIsBeingSubscribedTo=!0,this.source.subscribe(this)},t.prototype.notifyComplete=function(t){if(!1===this.sourceIsBeingSubscribedTo)return e.prototype.complete.call(this)},t.prototype.complete=function(){if(this.sourceIsBeingSubscribedTo=!1,!this.isStopped){if(this.retries||this.subscribeToRetries(),!this.retriesSubscription||this.retriesSubscription.closed)return e.prototype.complete.call(this);this._unsubscribeAndRecycle(),this.notifications.next()}},t.prototype._unsubscribe=function(){var e=this.notifications,t=this.retriesSubscription;e&&(e.unsubscribe(),this.notifications=null),t&&(t.unsubscribe(),this.retriesSubscription=null),this.retries=null},t.prototype._unsubscribeAndRecycle=function(){var t=this._unsubscribe;return this._unsubscribe=null,e.prototype._unsubscribeAndRecycle.call(this),this._unsubscribe=t,this},t.prototype.subscribeToRetries=function(){this.notifications=new i.Subject;var t=Object(o.tryCatch)(this.notifier)(this.notifications);if(t===a.errorObject)return e.prototype.complete.call(this);this.retries=t,this.retriesSubscription=Object(u.subscribeToResult)(this,t)},t}(s.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/retry.js":function(e,t,n){"use strict";n.r(t),n.d(t,"retry",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(e){return void 0===e&&(e=-1),function(t){return t.lift(new a(e,t))}}var a=function(){function e(e,t){this.count=e,this.source=t}return e.prototype.call=function(e,t){return t.subscribe(new s(e,this.count,this.source))},e}(),s=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.count=n,i.source=r,i}return r.__extends(t,e),t.prototype.error=function(t){if(!this.isStopped){var n=this.source,r=this.count;if(0===r)return e.prototype.error.call(this,t);r>-1&&(this.count=r-1),n.subscribe(this._unsubscribeAndRecycle())}},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/retryWhen.js":function(e,t,n){"use strict";n.r(t),n.d(t,"retryWhen",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),a=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),s=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),u=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function c(e){return function(t){return t.lift(new l(e,t))}}var l=function(){function e(e,t){this.notifier=e,this.source=t}return e.prototype.call=function(e,t){return t.subscribe(new d(e,this.notifier,this.source))},e}(),d=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.notifier=n,i.source=r,i}return r.__extends(t,e),t.prototype.error=function(t){if(!this.isStopped){var n=this.errors,r=this.retries,s=this.retriesSubscription;if(r)this.errors=null,this.retriesSubscription=null;else{if(n=new i.Subject,(r=Object(o.tryCatch)(this.notifier)(n))===a.errorObject)return e.prototype.error.call(this,a.errorObject.e);s=Object(u.subscribeToResult)(this,r)}this._unsubscribeAndRecycle(),this.errors=n,this.retries=r,this.retriesSubscription=s,n.next(t)}},t.prototype._unsubscribe=function(){var e=this.errors,t=this.retriesSubscription;e&&(e.unsubscribe(),this.errors=null),t&&(t.unsubscribe(),this.retriesSubscription=null),this.retries=null},t.prototype.notifyNext=function(e,t,n,r,i){var o=this._unsubscribe;this._unsubscribe=null,this._unsubscribeAndRecycle(),this._unsubscribe=o,this.source.subscribe(this)},t}(s.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/sample.js":function(e,t,n){"use strict";n.r(t),n.d(t,"sample",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function a(e){return function(t){return t.lift(new s(e))}}var s=function(){function e(e){this.notifier=e}return e.prototype.call=function(e,t){var n=new u(e),r=t.subscribe(n);return r.add(Object(o.subscribeToResult)(n,this.notifier)),r},e}(),u=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t.hasValue=!1,t}return r.__extends(t,e),t.prototype._next=function(e){this.value=e,this.hasValue=!0},t.prototype.notifyNext=function(e,t,n,r,i){this.emitValue()},t.prototype.notifyComplete=function(){this.emitValue()},t.prototype.emitValue=function(){this.hasValue&&(this.hasValue=!1,this.destination.next(this.value))},t}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/sampleTime.js":function(e,t,n){"use strict";n.r(t),n.d(t,"sampleTime",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js");function a(e,t){return void 0===t&&(t=o.async),function(n){return n.lift(new s(e,t))}}var s=function(){function e(e,t){this.period=e,this.scheduler=t}return e.prototype.call=function(e,t){return t.subscribe(new u(e,this.period,this.scheduler))},e}(),u=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.period=n,i.scheduler=r,i.hasValue=!1,i.add(r.schedule(c,n,{subscriber:i,period:n})),i}return r.__extends(t,e),t.prototype._next=function(e){this.lastValue=e,this.hasValue=!0},t.prototype.notifyNext=function(){this.hasValue&&(this.hasValue=!1,this.destination.next(this.lastValue))},t}(i.Subscriber);function c(e){var t=e.subscriber,n=e.period;t.notifyNext(),this.schedule(e,n)}},"./node_modules/rxjs/_esm5/internal/operators/scan.js":function(e,t,n){"use strict";n.r(t),n.d(t,"scan",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(e,t){var n=!1;return arguments.length>=2&&(n=!0),function(r){return r.lift(new a(e,t,n))}}var a=function(){function e(e,t,n){void 0===n&&(n=!1),this.accumulator=e,this.seed=t,this.hasSeed=n}return e.prototype.call=function(e,t){return t.subscribe(new s(e,this.accumulator,this.seed,this.hasSeed))},e}(),s=function(e){function t(t,n,r,i){var o=e.call(this,t)||this;return o.accumulator=n,o._seed=r,o.hasSeed=i,o.index=0,o}return r.__extends(t,e),Object.defineProperty(t.prototype,"seed",{get:function(){return this._seed},set:function(e){this.hasSeed=!0,this._seed=e},enumerable:!0,configurable:!0}),t.prototype._next=function(e){if(this.hasSeed)return this._tryNext(e);this.seed=e,this.destination.next(e)},t.prototype._tryNext=function(e){var t,n=this.index++;try{t=this.accumulator(this.seed,e,n)}catch(e){this.destination.error(e)}this.seed=t,this.destination.next(t)},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/sequenceEqual.js":function(e,t,n){"use strict";n.r(t),n.d(t,"sequenceEqual",function(){return s}),n.d(t,"SequenceEqualOperator",function(){return u}),n.d(t,"SequenceEqualSubscriber",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),a=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js");function s(e,t){return function(n){return n.lift(new u(e,t))}}var u=function(){function e(e,t){this.compareTo=e,this.comparor=t}return e.prototype.call=function(e,t){return t.subscribe(new c(e,this.compareTo,this.comparor))},e}(),c=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.compareTo=n,i.comparor=r,i._a=[],i._b=[],i._oneComplete=!1,i.destination.add(n.subscribe(new l(t,i))),i}return r.__extends(t,e),t.prototype._next=function(e){this._oneComplete&&0===this._b.length?this.emit(!1):(this._a.push(e),this.checkValues())},t.prototype._complete=function(){this._oneComplete?this.emit(0===this._a.length&&0===this._b.length):this._oneComplete=!0,this.unsubscribe()},t.prototype.checkValues=function(){for(var e=this._a,t=this._b,n=this.comparor;e.length>0&&t.length>0;){var r=e.shift(),i=t.shift(),s=!1;n?(s=Object(o.tryCatch)(n)(r,i))===a.errorObject&&this.destination.error(a.errorObject.e):s=r===i,s||this.emit(!1)}},t.prototype.emit=function(e){var t=this.destination;t.next(e),t.complete()},t.prototype.nextB=function(e){this._oneComplete&&0===this._a.length?this.emit(!1):(this._b.push(e),this.checkValues())},t.prototype.completeB=function(){this._oneComplete?this.emit(0===this._a.length&&0===this._b.length):this._oneComplete=!0},t}(i.Subscriber),l=function(e){function t(t,n){var r=e.call(this,t)||this;return r.parent=n,r}return r.__extends(t,e),t.prototype._next=function(e){this.parent.nextB(e)},t.prototype._error=function(e){this.parent.error(e),this.unsubscribe()},t.prototype._complete=function(){this.parent.completeB(),this.unsubscribe()},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/share.js":function(e,t,n){"use strict";n.r(t),n.d(t,"share",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/operators/multicast.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/refCount.js"),o=n("./node_modules/rxjs/_esm5/internal/Subject.js");function a(){return new o.Subject}function s(){return function(e){return Object(i.refCount)()(Object(r.multicast)(a)(e))}}},"./node_modules/rxjs/_esm5/internal/operators/shareReplay.js":function(e,t,n){"use strict";n.r(t),n.d(t,"shareReplay",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/ReplaySubject.js");function i(e,t,n){return void 0===e&&(e=Number.POSITIVE_INFINITY),void 0===t&&(t=Number.POSITIVE_INFINITY),function(i){return i.lift((o=e,a=t,s=n,l=0,d=!1,p=!1,function(e){l++,u&&!d||(d=!1,u=new r.ReplaySubject(o,a,s),c=e.subscribe({next:function(e){u.next(e)},error:function(e){d=!0,u.error(e)},complete:function(){p=!0,u.complete()}}));var t=u.subscribe(this);return function(){l--,t.unsubscribe(),c&&0===l&&p&&c.unsubscribe()}}));var o,a,s,u,c,l,d,p}}},"./node_modules/rxjs/_esm5/internal/operators/single.js":function(e,t,n){"use strict";n.r(t),n.d(t,"single",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/EmptyError.js");function a(e){return function(t){return t.lift(new s(e,t))}}var s=function(){function e(e,t){this.predicate=e,this.source=t}return e.prototype.call=function(e,t){return t.subscribe(new u(e,this.predicate,this.source))},e}(),u=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.predicate=n,i.source=r,i.seenValue=!1,i.index=0,i}return r.__extends(t,e),t.prototype.applySingleValue=function(e){this.seenValue?this.destination.error("Sequence contains more than one element"):(this.seenValue=!0,this.singleValue=e)},t.prototype._next=function(e){var t=this.index++;this.predicate?this.tryNext(e,t):this.applySingleValue(e)},t.prototype.tryNext=function(e,t){try{this.predicate(e,t,this.source)&&this.applySingleValue(e)}catch(e){this.destination.error(e)}},t.prototype._complete=function(){var e=this.destination;this.index>0?(e.next(this.seenValue?this.singleValue:void 0),e.complete()):e.error(new o.EmptyError)},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/skip.js":function(e,t,n){"use strict";n.r(t),n.d(t,"skip",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(e){return function(t){return t.lift(new a(e))}}var a=function(){function e(e){this.total=e}return e.prototype.call=function(e,t){return t.subscribe(new s(e,this.total))},e}(),s=function(e){function t(t,n){var r=e.call(this,t)||this;return r.total=n,r.count=0,r}return r.__extends(t,e),t.prototype._next=function(e){++this.count>this.total&&this.destination.next(e)},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/skipLast.js":function(e,t,n){"use strict";n.r(t),n.d(t,"skipLast",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/ArgumentOutOfRangeError.js");function a(e){return function(t){return t.lift(new s(e))}}var s=function(){function e(e){if(this._skipCount=e,this._skipCount<0)throw new o.ArgumentOutOfRangeError}return e.prototype.call=function(e,t){return 0===this._skipCount?t.subscribe(new i.Subscriber(e)):t.subscribe(new u(e,this._skipCount))},e}(),u=function(e){function t(t,n){var r=e.call(this,t)||this;return r._skipCount=n,r._count=0,r._ring=new Array(n),r}return r.__extends(t,e),t.prototype._next=function(e){var t=this._skipCount,n=this._count++;if(n<t)this._ring[n]=e;else{var r=n%t,i=this._ring,o=i[r];i[r]=e,this.destination.next(o)}},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/skipUntil.js":function(e,t,n){"use strict";n.r(t),n.d(t,"skipUntil",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/InnerSubscriber.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function s(e){return function(t){return t.lift(new u(e))}}var u=function(){function e(e){this.notifier=e}return e.prototype.call=function(e,t){return t.subscribe(new c(e,this.notifier))},e}(),c=function(e){function t(t,n){var r=e.call(this,t)||this;r.hasValue=!1;var i=new o.InnerSubscriber(r,void 0,void 0);return r.add(i),r.innerSubscription=i,Object(a.subscribeToResult)(r,n,void 0,void 0,i),r}return r.__extends(t,e),t.prototype._next=function(t){this.hasValue&&e.prototype._next.call(this,t)},t.prototype.notifyNext=function(e,t,n,r,i){this.hasValue=!0,this.innerSubscription&&this.innerSubscription.unsubscribe()},t.prototype.notifyComplete=function(){},t}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/skipWhile.js":function(e,t,n){"use strict";n.r(t),n.d(t,"skipWhile",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(e){return function(t){return t.lift(new a(e))}}var a=function(){function e(e){this.predicate=e}return e.prototype.call=function(e,t){return t.subscribe(new s(e,this.predicate))},e}(),s=function(e){function t(t,n){var r=e.call(this,t)||this;return r.predicate=n,r.skipping=!0,r.index=0,r}return r.__extends(t,e),t.prototype._next=function(e){var t=this.destination;this.skipping&&this.tryCallPredicate(e),this.skipping||t.next(e)},t.prototype.tryCallPredicate=function(e){try{var t=this.predicate(e,this.index++);this.skipping=Boolean(t)}catch(e){this.destination.error(e)}},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/startWith.js":function(e,t,n){"use strict";n.r(t),n.d(t,"startWith",function(){return u});var r=n("./node_modules/rxjs/_esm5/internal/observable/fromArray.js"),i=n("./node_modules/rxjs/_esm5/internal/observable/scalar.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/empty.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/concat.js"),s=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js");function u(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];return function(t){var n=e[e.length-1];Object(s.isScheduler)(n)?e.pop():n=null;var u=e.length;return 1!==u||n?u>0?Object(a.concat)(Object(r.fromArray)(e,n),t):Object(a.concat)(Object(o.empty)(n),t):Object(a.concat)(Object(i.scalar)(e[0]),t)}}},"./node_modules/rxjs/_esm5/internal/operators/subscribeOn.js":function(e,t,n){"use strict";n.r(t),n.d(t,"subscribeOn",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/observable/SubscribeOnObservable.js");function i(e,t){return void 0===t&&(t=0),function(n){return n.lift(new o(e,t))}}var o=function(){function e(e,t){this.scheduler=e,this.delay=t}return e.prototype.call=function(e,t){return new r.SubscribeOnObservable(t,this.delay,this.scheduler).subscribe(e)},e}()},"./node_modules/rxjs/_esm5/internal/operators/switchAll.js":function(e,t,n){"use strict";n.r(t),n.d(t,"switchAll",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/operators/switchMap.js"),i=n("./node_modules/rxjs/_esm5/internal/util/identity.js");function o(){return Object(r.switchMap)(i.identity)}},"./node_modules/rxjs/_esm5/internal/operators/switchMap.js":function(e,t,n){"use strict";n.r(t),n.d(t,"switchMap",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/InnerSubscriber.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),s=n("./node_modules/rxjs/_esm5/internal/operators/map.js"),u=n("./node_modules/rxjs/_esm5/internal/observable/from.js");function c(e,t){return"function"==typeof t?function(n){return n.pipe(c(function(n,r){return Object(u.from)(e(n,r)).pipe(Object(s.map)(function(e,i){return t(n,e,r,i)}))}))}:function(t){return t.lift(new l(e))}}var l=function(){function e(e){this.project=e}return e.prototype.call=function(e,t){return t.subscribe(new d(e,this.project))},e}(),d=function(e){function t(t,n){var r=e.call(this,t)||this;return r.project=n,r.index=0,r}return r.__extends(t,e),t.prototype._next=function(e){var t,n=this.index++;try{t=this.project(e,n)}catch(e){return void this.destination.error(e)}this._innerSub(t,e,n)},t.prototype._innerSub=function(e,t,n){var r=this.innerSubscription;r&&r.unsubscribe();var i=new o.InnerSubscriber(this,void 0,void 0);this.destination.add(i),this.innerSubscription=Object(a.subscribeToResult)(this,e,t,n,i)},t.prototype._complete=function(){var t=this.innerSubscription;t&&!t.closed||e.prototype._complete.call(this),this.unsubscribe()},t.prototype._unsubscribe=function(){this.innerSubscription=null},t.prototype.notifyComplete=function(t){this.destination.remove(t),this.innerSubscription=null,this.isStopped&&e.prototype._complete.call(this)},t.prototype.notifyNext=function(e,t,n,r,i){this.destination.next(t)},t}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/switchMapTo.js":function(e,t,n){"use strict";n.r(t),n.d(t,"switchMapTo",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/operators/switchMap.js");function i(e,t){return t?Object(r.switchMap)(function(){return e},t):Object(r.switchMap)(function(){return e})}},"./node_modules/rxjs/_esm5/internal/operators/take.js":function(e,t,n){"use strict";n.r(t),n.d(t,"take",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/ArgumentOutOfRangeError.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/empty.js");function s(e){return function(t){return 0===e?Object(a.empty)():t.lift(new u(e))}}var u=function(){function e(e){if(this.total=e,this.total<0)throw new o.ArgumentOutOfRangeError}return e.prototype.call=function(e,t){return t.subscribe(new c(e,this.total))},e}(),c=function(e){function t(t,n){var r=e.call(this,t)||this;return r.total=n,r.count=0,r}return r.__extends(t,e),t.prototype._next=function(e){var t=this.total,n=++this.count;n<=t&&(this.destination.next(e),n===t&&(this.destination.complete(),this.unsubscribe()))},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/takeLast.js":function(e,t,n){"use strict";n.r(t),n.d(t,"takeLast",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/ArgumentOutOfRangeError.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/empty.js");function s(e){return function(t){return 0===e?Object(a.empty)():t.lift(new u(e))}}var u=function(){function e(e){if(this.total=e,this.total<0)throw new o.ArgumentOutOfRangeError}return e.prototype.call=function(e,t){return t.subscribe(new c(e,this.total))},e}(),c=function(e){function t(t,n){var r=e.call(this,t)||this;return r.total=n,r.ring=new Array,r.count=0,r}return r.__extends(t,e),t.prototype._next=function(e){var t=this.ring,n=this.total,r=this.count++;t.length<n?t.push(e):t[r%n]=e},t.prototype._complete=function(){var e=this.destination,t=this.count;if(t>0)for(var n=this.count>=this.total?this.total:this.count,r=this.ring,i=0;i<n;i++){var o=t++%n;e.next(r[o])}e.complete()},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/takeUntil.js":function(e,t,n){"use strict";n.r(t),n.d(t,"takeUntil",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function a(e){return function(t){return t.lift(new s(e))}}var s=function(){function e(e){this.notifier=e}return e.prototype.call=function(e,t){var n=new u(e),r=Object(o.subscribeToResult)(n,this.notifier);return r&&!n.seenValue?(n.add(r),t.subscribe(n)):n},e}(),u=function(e){function t(t){var n=e.call(this,t)||this;return n.seenValue=!1,n}return r.__extends(t,e),t.prototype.notifyNext=function(e,t,n,r,i){this.seenValue=!0,this.complete()},t.prototype.notifyComplete=function(){},t}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/takeWhile.js":function(e,t,n){"use strict";n.r(t),n.d(t,"takeWhile",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function o(e){return function(t){return t.lift(new a(e))}}var a=function(){function e(e){this.predicate=e}return e.prototype.call=function(e,t){return t.subscribe(new s(e,this.predicate))},e}(),s=function(e){function t(t,n){var r=e.call(this,t)||this;return r.predicate=n,r.index=0,r}return r.__extends(t,e),t.prototype._next=function(e){var t,n=this.destination;try{t=this.predicate(e,this.index++)}catch(e){return void n.error(e)}this.nextOrComplete(e,t)},t.prototype.nextOrComplete=function(e,t){var n=this.destination;Boolean(t)?n.next(e):n.complete()},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/tap.js":function(e,t,n){"use strict";n.r(t),n.d(t,"tap",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/noop.js"),a=n("./node_modules/rxjs/_esm5/internal/util/isFunction.js");function s(e,t,n){return function(r){return r.lift(new u(e,t,n))}}var u=function(){function e(e,t,n){this.nextOrObserver=e,this.error=t,this.complete=n}return e.prototype.call=function(e,t){return t.subscribe(new c(e,this.nextOrObserver,this.error,this.complete))},e}(),c=function(e){function t(t,n,r,i){var s=e.call(this,t)||this;return s._tapNext=o.noop,s._tapError=o.noop,s._tapComplete=o.noop,s._tapError=r||o.noop,s._tapComplete=i||o.noop,Object(a.isFunction)(n)?(s._context=s,s._tapNext=n):n&&(s._context=n,s._tapNext=n.next||o.noop,s._tapError=n.error||o.noop,s._tapComplete=n.complete||o.noop),s}return r.__extends(t,e),t.prototype._next=function(e){try{this._tapNext.call(this._context,e)}catch(e){return void this.destination.error(e)}this.destination.next(e)},t.prototype._error=function(e){try{this._tapError.call(this._context,e)}catch(e){return void this.destination.error(e)}this.destination.error(e)},t.prototype._complete=function(){try{this._tapComplete.call(this._context)}catch(e){return void this.destination.error(e)}return this.destination.complete()},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/throttle.js":function(e,t,n){"use strict";n.r(t),n.d(t,"defaultThrottleConfig",function(){return a}),n.d(t,"throttle",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js"),a={leading:!0,trailing:!1};function s(e,t){return void 0===t&&(t=a),function(n){return n.lift(new u(e,t.leading,t.trailing))}}var u=function(){function e(e,t,n){this.durationSelector=e,this.leading=t,this.trailing=n}return e.prototype.call=function(e,t){return t.subscribe(new c(e,this.durationSelector,this.leading,this.trailing))},e}(),c=function(e){function t(t,n,r,i){var o=e.call(this,t)||this;return o.destination=t,o.durationSelector=n,o._leading=r,o._trailing=i,o._hasValue=!1,o}return r.__extends(t,e),t.prototype._next=function(e){this._hasValue=!0,this._sendValue=e,this._throttled||(this._leading?this.send():this.throttle(e))},t.prototype.send=function(){var e=this._hasValue,t=this._sendValue;e&&(this.destination.next(t),this.throttle(t)),this._hasValue=!1,this._sendValue=null},t.prototype.throttle=function(e){var t=this.tryDurationSelector(e);t&&this.add(this._throttled=Object(o.subscribeToResult)(this,t))},t.prototype.tryDurationSelector=function(e){try{return this.durationSelector(e)}catch(e){return this.destination.error(e),null}},t.prototype.throttlingDone=function(){var e=this._throttled,t=this._trailing;e&&e.unsubscribe(),this._throttled=null,t&&this.send()},t.prototype.notifyNext=function(e,t,n,r,i){this.throttlingDone()},t.prototype.notifyComplete=function(){this.throttlingDone()},t}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/throttleTime.js":function(e,t,n){"use strict";n.r(t),n.d(t,"throttleTime",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),a=n("./node_modules/rxjs/_esm5/internal/operators/throttle.js");function s(e,t,n){return void 0===t&&(t=o.async),void 0===n&&(n=a.defaultThrottleConfig),function(r){return r.lift(new u(e,t,n.leading,n.trailing))}}var u=function(){function e(e,t,n,r){this.duration=e,this.scheduler=t,this.leading=n,this.trailing=r}return e.prototype.call=function(e,t){return t.subscribe(new c(e,this.duration,this.scheduler,this.leading,this.trailing))},e}(),c=function(e){function t(t,n,r,i,o){var a=e.call(this,t)||this;return a.duration=n,a.scheduler=r,a.leading=i,a.trailing=o,a._hasTrailingValue=!1,a._trailingValue=null,a}return r.__extends(t,e),t.prototype._next=function(e){this.throttled?this.trailing&&(this._trailingValue=e,this._hasTrailingValue=!0):(this.add(this.throttled=this.scheduler.schedule(l,this.duration,{subscriber:this})),this.leading&&this.destination.next(e))},t.prototype._complete=function(){this._hasTrailingValue?(this.destination.next(this._trailingValue),this.destination.complete()):this.destination.complete()},t.prototype.clearThrottle=function(){var e=this.throttled;e&&(this.trailing&&this._hasTrailingValue&&(this.destination.next(this._trailingValue),this._trailingValue=null,this._hasTrailingValue=!1),e.unsubscribe(),this.remove(e),this.throttled=null)},t}(i.Subscriber);function l(e){e.subscriber.clearThrottle()}},"./node_modules/rxjs/_esm5/internal/operators/throwIfEmpty.js":function(e,t,n){"use strict";n.r(t),n.d(t,"throwIfEmpty",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/operators/tap.js"),i=n("./node_modules/rxjs/_esm5/internal/util/EmptyError.js"),o=function(e){return void 0===e&&(e=a),Object(r.tap)({hasValue:!1,next:function(){this.hasValue=!0},complete:function(){if(!this.hasValue)throw e()}})};function a(){return new i.EmptyError}},"./node_modules/rxjs/_esm5/internal/operators/timeInterval.js":function(e,t,n){"use strict";n.r(t),n.d(t,"timeInterval",function(){return s}),n.d(t,"TimeInterval",function(){return u});var r=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/scan.js"),o=n("./node_modules/rxjs/_esm5/internal/observable/defer.js"),a=n("./node_modules/rxjs/_esm5/internal/operators/map.js");function s(e){return void 0===e&&(e=r.async),function(t){return Object(o.defer)(function(){return t.pipe(Object(i.scan)(function(t,n){var r=t.current;return{value:n,current:e.now(),last:r}},{current:e.now(),value:void 0,last:void 0}),Object(a.map)(function(e){var t=e.current,n=e.last,r=e.value;return new u(r,t-n)}))})}}var u=function(){return function(e,t){this.value=e,this.interval=t}}()},"./node_modules/rxjs/_esm5/internal/operators/timeout.js":function(e,t,n){"use strict";n.r(t),n.d(t,"timeout",function(){return s});var r=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),i=n("./node_modules/rxjs/_esm5/internal/util/TimeoutError.js"),o=n("./node_modules/rxjs/_esm5/internal/operators/timeoutWith.js"),a=n("./node_modules/rxjs/_esm5/internal/observable/throwError.js");function s(e,t){return void 0===t&&(t=r.async),Object(o.timeoutWith)(e,Object(a.throwError)(new i.TimeoutError),t)}},"./node_modules/rxjs/_esm5/internal/operators/timeoutWith.js":function(e,t,n){"use strict";n.r(t),n.d(t,"timeoutWith",function(){return u});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),o=n("./node_modules/rxjs/_esm5/internal/util/isDate.js"),a=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function u(e,t,n){return void 0===n&&(n=i.async),function(r){var i=Object(o.isDate)(e),a=i?+e-n.now():Math.abs(e);return r.lift(new c(a,i,t,n))}}var c=function(){function e(e,t,n,r){this.waitFor=e,this.absoluteTimeout=t,this.withObservable=n,this.scheduler=r}return e.prototype.call=function(e,t){return t.subscribe(new l(e,this.absoluteTimeout,this.waitFor,this.withObservable,this.scheduler))},e}(),l=function(e){function t(t,n,r,i,o){var a=e.call(this,t)||this;return a.absoluteTimeout=n,a.waitFor=r,a.withObservable=i,a.scheduler=o,a.action=null,a.scheduleTimeout(),a}return r.__extends(t,e),t.dispatchTimeout=function(e){var t=e.withObservable;e._unsubscribeAndRecycle(),e.add(Object(s.subscribeToResult)(e,t))},t.prototype.scheduleTimeout=function(){var e=this.action;e?this.action=e.schedule(this,this.waitFor):this.add(this.action=this.scheduler.schedule(t.dispatchTimeout,this.waitFor,this))},t.prototype._next=function(t){this.absoluteTimeout||this.scheduleTimeout(),e.prototype._next.call(this,t)},t.prototype._unsubscribe=function(){this.action=null,this.scheduler=null,this.withObservable=null},t}(a.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/timestamp.js":function(e,t,n){"use strict";n.r(t),n.d(t,"timestamp",function(){return o}),n.d(t,"Timestamp",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),i=n("./node_modules/rxjs/_esm5/internal/operators/map.js");function o(e){return void 0===e&&(e=r.async),Object(i.map)(function(t){return new a(t,e.now())})}var a=function(){return function(e,t){this.value=e,this.timestamp=t}}()},"./node_modules/rxjs/_esm5/internal/operators/toArray.js":function(e,t,n){"use strict";n.r(t),n.d(t,"toArray",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/operators/reduce.js");function i(e,t,n){return 0===n?[t]:(e.push(t),e)}function o(){return Object(r.reduce)(i,[])}},"./node_modules/rxjs/_esm5/internal/operators/window.js":function(e,t,n){"use strict";n.r(t),n.d(t,"window",function(){return s});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function s(e){return function(t){return t.lift(new u(e))}}var u=function(){function e(e){this.windowBoundaries=e}return e.prototype.call=function(e,t){var n=new c(e),r=t.subscribe(n);return r.closed||n.add(Object(a.subscribeToResult)(n,this.windowBoundaries)),r},e}(),c=function(e){function t(t){var n=e.call(this,t)||this;return n.window=new i.Subject,t.next(n.window),n}return r.__extends(t,e),t.prototype.notifyNext=function(e,t,n,r,i){this.openWindow()},t.prototype.notifyError=function(e,t){this._error(e)},t.prototype.notifyComplete=function(e){this._complete()},t.prototype._next=function(e){this.window.next(e)},t.prototype._error=function(e){this.window.error(e),this.destination.error(e)},t.prototype._complete=function(){this.window.complete(),this.destination.complete()},t.prototype._unsubscribe=function(){this.window=null},t.prototype.openWindow=function(){var e=this.window;e&&e.complete();var t=this.destination,n=this.window=new i.Subject;t.next(n)},t}(o.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/windowCount.js":function(e,t,n){"use strict";n.r(t),n.d(t,"windowCount",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/Subject.js");function a(e,t){return void 0===t&&(t=0),function(n){return n.lift(new s(e,t))}}var s=function(){function e(e,t){this.windowSize=e,this.startWindowEvery=t}return e.prototype.call=function(e,t){return t.subscribe(new u(e,this.windowSize,this.startWindowEvery))},e}(),u=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.destination=t,i.windowSize=n,i.startWindowEvery=r,i.windows=[new o.Subject],i.count=0,t.next(i.windows[0]),i}return r.__extends(t,e),t.prototype._next=function(e){for(var t=this.startWindowEvery>0?this.startWindowEvery:this.windowSize,n=this.destination,r=this.windowSize,i=this.windows,a=i.length,s=0;s<a&&!this.closed;s++)i[s].next(e);var u=this.count-r+1;if(u>=0&&u%t==0&&!this.closed&&i.shift().complete(),++this.count%t==0&&!this.closed){var c=new o.Subject;i.push(c),n.next(c)}},t.prototype._error=function(e){var t=this.windows;if(t)for(;t.length>0&&!this.closed;)t.shift().error(e);this.destination.error(e)},t.prototype._complete=function(){var e=this.windows;if(e)for(;e.length>0&&!this.closed;)e.shift().complete();this.destination.complete()},t.prototype._unsubscribe=function(){this.count=0,this.windows=null},t}(i.Subscriber)},"./node_modules/rxjs/_esm5/internal/operators/windowTime.js":function(e,t,n){"use strict";n.r(t),n.d(t,"windowTime",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/scheduler/async.js"),a=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),s=n("./node_modules/rxjs/_esm5/internal/util/isNumeric.js"),u=n("./node_modules/rxjs/_esm5/internal/util/isScheduler.js");function c(e){var t=o.async,n=null,r=Number.POSITIVE_INFINITY;return Object(u.isScheduler)(arguments[3])&&(t=arguments[3]),Object(u.isScheduler)(arguments[2])?t=arguments[2]:Object(s.isNumeric)(arguments[2])&&(r=arguments[2]),Object(u.isScheduler)(arguments[1])?t=arguments[1]:Object(s.isNumeric)(arguments[1])&&(n=arguments[1]),function(i){return i.lift(new l(e,n,r,t))}}var l=function(){function e(e,t,n,r){this.windowTimeSpan=e,this.windowCreationInterval=t,this.maxWindowSize=n,this.scheduler=r}return e.prototype.call=function(e,t){return t.subscribe(new p(e,this.windowTimeSpan,this.windowCreationInterval,this.maxWindowSize,this.scheduler))},e}(),d=function(e){function t(){var t=null!==e&&e.apply(this,arguments)||this;return t._numberOfNextedValues=0,t}return r.__extends(t,e),t.prototype.next=function(t){this._numberOfNextedValues++,e.prototype.next.call(this,t)},Object.defineProperty(t.prototype,"numberOfNextedValues",{get:function(){return this._numberOfNextedValues},enumerable:!0,configurable:!0}),t}(i.Subject),p=function(e){function t(t,n,r,i,o){var a=e.call(this,t)||this;a.destination=t,a.windowTimeSpan=n,a.windowCreationInterval=r,a.maxWindowSize=i,a.scheduler=o,a.windows=[];var s=a.openWindow();if(null!==r&&r>=0){var u={subscriber:a,window:s,context:null},c={windowTimeSpan:n,windowCreationInterval:r,subscriber:a,scheduler:o};a.add(o.schedule(m,n,u)),a.add(o.schedule(h,r,c))}else{var l={subscriber:a,window:s,windowTimeSpan:n};a.add(o.schedule(f,n,l))}return a}return r.__extends(t,e),t.prototype._next=function(e){for(var t=this.windows,n=t.length,r=0;r<n;r++){var i=t[r];i.closed||(i.next(e),i.numberOfNextedValues>=this.maxWindowSize&&this.closeWindow(i))}},t.prototype._error=function(e){for(var t=this.windows;t.length>0;)t.shift().error(e);this.destination.error(e)},t.prototype._complete=function(){for(var e=this.windows;e.length>0;){var t=e.shift();t.closed||t.complete()}this.destination.complete()},t.prototype.openWindow=function(){var e=new d;return this.windows.push(e),this.destination.next(e),e},t.prototype.closeWindow=function(e){e.complete();var t=this.windows;t.splice(t.indexOf(e),1)},t}(a.Subscriber);function f(e){var t=e.subscriber,n=e.windowTimeSpan,r=e.window;r&&t.closeWindow(r),e.window=t.openWindow(),this.schedule(e,n)}function h(e){var t=e.windowTimeSpan,n=e.subscriber,r=e.scheduler,i=e.windowCreationInterval,o={action:this,subscription:null},a={subscriber:n,window:n.openWindow(),context:o};o.subscription=r.schedule(m,t,a),this.add(o.subscription),this.schedule(e,i)}function m(e){var t=e.subscriber,n=e.window,r=e.context;r&&r.action&&r.subscription&&r.action.remove(r.subscription),t.closeWindow(n)}},"./node_modules/rxjs/_esm5/internal/operators/windowToggle.js":function(e,t,n){"use strict";n.r(t),n.d(t,"windowToggle",function(){return l});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/Subscription.js"),a=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),s=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),u=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),c=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function l(e,t){return function(n){return n.lift(new d(e,t))}}var d=function(){function e(e,t){this.openings=e,this.closingSelector=t}return e.prototype.call=function(e,t){return t.subscribe(new p(e,this.openings,this.closingSelector))},e}(),p=function(e){function t(t,n,r){var i=e.call(this,t)||this;return i.openings=n,i.closingSelector=r,i.contexts=[],i.add(i.openSubscription=Object(c.subscribeToResult)(i,n,n)),i}return r.__extends(t,e),t.prototype._next=function(e){var t=this.contexts;if(t)for(var n=t.length,r=0;r<n;r++)t[r].window.next(e)},t.prototype._error=function(t){var n=this.contexts;if(this.contexts=null,n)for(var r=n.length,i=-1;++i<r;){var o=n[i];o.window.error(t),o.subscription.unsubscribe()}e.prototype._error.call(this,t)},t.prototype._complete=function(){var t=this.contexts;if(this.contexts=null,t)for(var n=t.length,r=-1;++r<n;){var i=t[r];i.window.complete(),i.subscription.unsubscribe()}e.prototype._complete.call(this)},t.prototype._unsubscribe=function(){var e=this.contexts;if(this.contexts=null,e)for(var t=e.length,n=-1;++n<t;){var r=e[n];r.window.unsubscribe(),r.subscription.unsubscribe()}},t.prototype.notifyNext=function(e,t,n,r,u){if(e===this.openings){var l=this.closingSelector,d=Object(a.tryCatch)(l)(t);if(d===s.errorObject)return this.error(s.errorObject.e);var p=new i.Subject,f=new o.Subscription,h={window:p,subscription:f};this.contexts.push(h);var m=Object(c.subscribeToResult)(this,d,h);m.closed?this.closeWindow(this.contexts.length-1):(m.context=h,f.add(m)),this.destination.next(p)}else this.closeWindow(this.contexts.indexOf(e))},t.prototype.notifyError=function(e){this.error(e)},t.prototype.notifyComplete=function(e){e!==this.openSubscription&&this.closeWindow(this.contexts.indexOf(e.context))},t.prototype.closeWindow=function(e){if(-1!==e){var t=this.contexts,n=t[e],r=n.window,i=n.subscription;t.splice(e,1),r.complete(),i.unsubscribe()}},t}(u.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/windowWhen.js":function(e,t,n){"use strict";n.r(t),n.d(t,"windowWhen",function(){return c});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Subject.js"),o=n("./node_modules/rxjs/_esm5/internal/util/tryCatch.js"),a=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js"),s=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),u=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function c(e){return function(t){return t.lift(new l(e))}}var l=function(){function e(e){this.closingSelector=e}return e.prototype.call=function(e,t){return t.subscribe(new d(e,this.closingSelector))},e}(),d=function(e){function t(t,n){var r=e.call(this,t)||this;return r.destination=t,r.closingSelector=n,r.openWindow(),r}return r.__extends(t,e),t.prototype.notifyNext=function(e,t,n,r,i){this.openWindow(i)},t.prototype.notifyError=function(e,t){this._error(e)},t.prototype.notifyComplete=function(e){this.openWindow(e)},t.prototype._next=function(e){this.window.next(e)},t.prototype._error=function(e){this.window.error(e),this.destination.error(e),this.unsubscribeClosingNotification()},t.prototype._complete=function(){this.window.complete(),this.destination.complete(),this.unsubscribeClosingNotification()},t.prototype.unsubscribeClosingNotification=function(){this.closingNotification&&this.closingNotification.unsubscribe()},t.prototype.openWindow=function(e){void 0===e&&(e=null),e&&(this.remove(e),e.unsubscribe());var t=this.window;t&&t.complete();var n=this.window=new i.Subject;this.destination.next(n);var r=Object(o.tryCatch)(this.closingSelector)();if(r===a.errorObject){var s=a.errorObject.e;this.destination.error(s),this.window.error(s)}else this.add(this.closingNotification=Object(u.subscribeToResult)(this,r))},t}(s.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/withLatestFrom.js":function(e,t,n){"use strict";n.r(t),n.d(t,"withLatestFrom",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/OuterSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js");function a(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];return function(t){var n;"function"==typeof e[e.length-1]&&(n=e.pop());var r=e;return t.lift(new s(r,n))}}var s=function(){function e(e,t){this.observables=e,this.project=t}return e.prototype.call=function(e,t){return t.subscribe(new u(e,this.observables,this.project))},e}(),u=function(e){function t(t,n,r){var i=e.call(this,t)||this;i.observables=n,i.project=r,i.toRespond=[];var a=n.length;i.values=new Array(a);for(var s=0;s<a;s++)i.toRespond.push(s);for(s=0;s<a;s++){var u=n[s];i.add(Object(o.subscribeToResult)(i,u,u,s))}return i}return r.__extends(t,e),t.prototype.notifyNext=function(e,t,n,r,i){this.values[n]=t;var o=this.toRespond;if(o.length>0){var a=o.indexOf(n);-1!==a&&o.splice(a,1)}},t.prototype.notifyComplete=function(){},t.prototype._next=function(e){if(0===this.toRespond.length){var t=[e].concat(this.values);this.project?this._tryProject(t):this.destination.next(t)}},t.prototype._tryProject=function(e){var t;try{t=this.project.apply(this,e)}catch(e){return void this.destination.error(e)}this.destination.next(t)},t}(i.OuterSubscriber)},"./node_modules/rxjs/_esm5/internal/operators/zip.js":function(e,t,n){"use strict";n.r(t),n.d(t,"zip",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/observable/zip.js");function i(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];return function(t){return t.lift.call(r.zip.apply(void 0,[t].concat(e)))}}},"./node_modules/rxjs/_esm5/internal/operators/zipAll.js":function(e,t,n){"use strict";n.r(t),n.d(t,"zipAll",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/observable/zip.js");function i(e){return function(t){return t.lift(new r.ZipOperator(e))}}},"./node_modules/rxjs/_esm5/internal/scheduler/Action.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Action",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(e){function t(t,n){return e.call(this)||this}return r.__extends(t,e),t.prototype.schedule=function(e,t){return void 0===t&&(t=0),this},t}(n("./node_modules/rxjs/_esm5/internal/Subscription.js").Subscription)},"./node_modules/rxjs/_esm5/internal/scheduler/AnimationFrameAction.js":function(e,t,n){"use strict";n.r(t),n.d(t,"AnimationFrameAction",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(e){function t(t,n){var r=e.call(this,t,n)||this;return r.scheduler=t,r.work=n,r}return r.__extends(t,e),t.prototype.requestAsyncId=function(t,n,r){return void 0===r&&(r=0),null!==r&&r>0?e.prototype.requestAsyncId.call(this,t,n,r):(t.actions.push(this),t.scheduled||(t.scheduled=requestAnimationFrame(function(){return t.flush(null)})))},t.prototype.recycleAsyncId=function(t,n,r){if(void 0===r&&(r=0),null!==r&&r>0||null===r&&this.delay>0)return e.prototype.recycleAsyncId.call(this,t,n,r);0===t.actions.length&&(cancelAnimationFrame(n),t.scheduled=void 0)},t}(n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncAction.js").AsyncAction)},"./node_modules/rxjs/_esm5/internal/scheduler/AnimationFrameScheduler.js":function(e,t,n){"use strict";n.r(t),n.d(t,"AnimationFrameScheduler",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return r.__extends(t,e),t.prototype.flush=function(e){this.active=!0,this.scheduled=void 0;var t,n=this.actions,r=-1,i=n.length;e=e||n.shift();do{if(t=e.execute(e.state,e.delay))break}while(++r<i&&(e=n.shift()));if(this.active=!1,t){for(;++r<i&&(e=n.shift());)e.unsubscribe();throw t}},t}(n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncScheduler.js").AsyncScheduler)},"./node_modules/rxjs/_esm5/internal/scheduler/AsapAction.js":function(e,t,n){"use strict";n.r(t),n.d(t,"AsapAction",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/util/Immediate.js"),o=function(e){function t(t,n){var r=e.call(this,t,n)||this;return r.scheduler=t,r.work=n,r}return r.__extends(t,e),t.prototype.requestAsyncId=function(t,n,r){return void 0===r&&(r=0),null!==r&&r>0?e.prototype.requestAsyncId.call(this,t,n,r):(t.actions.push(this),t.scheduled||(t.scheduled=i.Immediate.setImmediate(t.flush.bind(t,null))))},t.prototype.recycleAsyncId=function(t,n,r){if(void 0===r&&(r=0),null!==r&&r>0||null===r&&this.delay>0)return e.prototype.recycleAsyncId.call(this,t,n,r);0===t.actions.length&&(i.Immediate.clearImmediate(n),t.scheduled=void 0)},t}(n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncAction.js").AsyncAction)},"./node_modules/rxjs/_esm5/internal/scheduler/AsapScheduler.js":function(e,t,n){"use strict";n.r(t),n.d(t,"AsapScheduler",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return r.__extends(t,e),t.prototype.flush=function(e){this.active=!0,this.scheduled=void 0;var t,n=this.actions,r=-1,i=n.length;e=e||n.shift();do{if(t=e.execute(e.state,e.delay))break}while(++r<i&&(e=n.shift()));if(this.active=!1,t){for(;++r<i&&(e=n.shift());)e.unsubscribe();throw t}},t}(n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncScheduler.js").AsyncScheduler)},"./node_modules/rxjs/_esm5/internal/scheduler/AsyncAction.js":function(e,t,n){"use strict";n.r(t),n.d(t,"AsyncAction",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(e){function t(t,n){var r=e.call(this,t,n)||this;return r.scheduler=t,r.work=n,r.pending=!1,r}return r.__extends(t,e),t.prototype.schedule=function(e,t){if(void 0===t&&(t=0),this.closed)return this;this.state=e;var n=this.id,r=this.scheduler;return null!=n&&(this.id=this.recycleAsyncId(r,n,t)),this.pending=!0,this.delay=t,this.id=this.id||this.requestAsyncId(r,this.id,t),this},t.prototype.requestAsyncId=function(e,t,n){return void 0===n&&(n=0),setInterval(e.flush.bind(e,this),n)},t.prototype.recycleAsyncId=function(e,t,n){if(void 0===n&&(n=0),null!==n&&this.delay===n&&!1===this.pending)return t;clearInterval(t)},t.prototype.execute=function(e,t){if(this.closed)return new Error("executing a cancelled action");this.pending=!1;var n=this._execute(e,t);if(n)return n;!1===this.pending&&null!=this.id&&(this.id=this.recycleAsyncId(this.scheduler,this.id,null))},t.prototype._execute=function(e,t){var n=!1,r=void 0;try{this.work(e)}catch(e){n=!0,r=!!e&&e||new Error(e)}if(n)return this.unsubscribe(),r},t.prototype._unsubscribe=function(){var e=this.id,t=this.scheduler,n=t.actions,r=n.indexOf(this);this.work=null,this.state=null,this.pending=!1,this.scheduler=null,-1!==r&&n.splice(r,1),null!=e&&(this.id=this.recycleAsyncId(t,e,null)),this.delay=null},t}(n("./node_modules/rxjs/_esm5/internal/scheduler/Action.js").Action)},"./node_modules/rxjs/_esm5/internal/scheduler/AsyncScheduler.js":function(e,t,n){"use strict";n.r(t),n.d(t,"AsyncScheduler",function(){return o});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/Scheduler.js"),o=function(e){function t(n,r){void 0===r&&(r=i.Scheduler.now);var o=e.call(this,n,function(){return t.delegate&&t.delegate!==o?t.delegate.now():r()})||this;return o.actions=[],o.active=!1,o.scheduled=void 0,o}return r.__extends(t,e),t.prototype.schedule=function(n,r,i){return void 0===r&&(r=0),t.delegate&&t.delegate!==this?t.delegate.schedule(n,r,i):e.prototype.schedule.call(this,n,r,i)},t.prototype.flush=function(e){var t=this.actions;if(this.active)t.push(e);else{var n;this.active=!0;do{if(n=e.execute(e.state,e.delay))break}while(e=t.shift());if(this.active=!1,n){for(;e=t.shift();)e.unsubscribe();throw n}}},t}(i.Scheduler)},"./node_modules/rxjs/_esm5/internal/scheduler/QueueAction.js":function(e,t,n){"use strict";n.r(t),n.d(t,"QueueAction",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(e){function t(t,n){var r=e.call(this,t,n)||this;return r.scheduler=t,r.work=n,r}return r.__extends(t,e),t.prototype.schedule=function(t,n){return void 0===n&&(n=0),n>0?e.prototype.schedule.call(this,t,n):(this.delay=n,this.state=t,this.scheduler.flush(this),this)},t.prototype.execute=function(t,n){return n>0||this.closed?e.prototype.execute.call(this,t,n):this._execute(t,n)},t.prototype.requestAsyncId=function(t,n,r){return void 0===r&&(r=0),null!==r&&r>0||null===r&&this.delay>0?e.prototype.requestAsyncId.call(this,t,n,r):t.flush(this)},t}(n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncAction.js").AsyncAction)},"./node_modules/rxjs/_esm5/internal/scheduler/QueueScheduler.js":function(e,t,n){"use strict";n.r(t),n.d(t,"QueueScheduler",function(){return i});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=function(e){function t(){return null!==e&&e.apply(this,arguments)||this}return r.__extends(t,e),t}(n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncScheduler.js").AsyncScheduler)},"./node_modules/rxjs/_esm5/internal/scheduler/VirtualTimeScheduler.js":function(e,t,n){"use strict";n.r(t),n.d(t,"VirtualTimeScheduler",function(){return o}),n.d(t,"VirtualAction",function(){return a});var r=n("./node_modules/rxjs/node_modules/tslib/tslib.es6.js"),i=n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncAction.js"),o=function(e){function t(t,n){void 0===t&&(t=a),void 0===n&&(n=Number.POSITIVE_INFINITY);var r=e.call(this,t,function(){return r.frame})||this;return r.maxFrames=n,r.frame=0,r.index=-1,r}return r.__extends(t,e),t.prototype.flush=function(){for(var e,t,n=this.actions,r=this.maxFrames;(t=n.shift())&&(this.frame=t.delay)<=r&&!(e=t.execute(t.state,t.delay)););if(e){for(;t=n.shift();)t.unsubscribe();throw e}},t.frameTimeFactor=10,t}(n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncScheduler.js").AsyncScheduler),a=function(e){function t(t,n,r){void 0===r&&(r=t.index+=1);var i=e.call(this,t,n)||this;return i.scheduler=t,i.work=n,i.index=r,i.active=!0,i.index=t.index=r,i}return r.__extends(t,e),t.prototype.schedule=function(n,r){if(void 0===r&&(r=0),!this.id)return e.prototype.schedule.call(this,n,r);this.active=!1;var i=new t(this.scheduler,this.work);return this.add(i),i.schedule(n,r)},t.prototype.requestAsyncId=function(e,n,r){void 0===r&&(r=0),this.delay=e.frame+r;var i=e.actions;return i.push(this),i.sort(t.sortActions),!0},t.prototype.recycleAsyncId=function(e,t,n){void 0===n&&(n=0)},t.prototype._execute=function(t,n){if(!0===this.active)return e.prototype._execute.call(this,t,n)},t.sortActions=function(e,t){return e.delay===t.delay?e.index===t.index?0:e.index>t.index?1:-1:e.delay>t.delay?1:-1},t}(i.AsyncAction)},"./node_modules/rxjs/_esm5/internal/scheduler/animationFrame.js":function(e,t,n){"use strict";n.r(t),n.d(t,"animationFrame",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/scheduler/AnimationFrameAction.js"),i=new(n("./node_modules/rxjs/_esm5/internal/scheduler/AnimationFrameScheduler.js").AnimationFrameScheduler)(r.AnimationFrameAction)},"./node_modules/rxjs/_esm5/internal/scheduler/asap.js":function(e,t,n){"use strict";n.r(t),n.d(t,"asap",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/scheduler/AsapAction.js"),i=new(n("./node_modules/rxjs/_esm5/internal/scheduler/AsapScheduler.js").AsapScheduler)(r.AsapAction)},"./node_modules/rxjs/_esm5/internal/scheduler/async.js":function(e,t,n){"use strict";n.r(t),n.d(t,"async",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncAction.js"),i=new(n("./node_modules/rxjs/_esm5/internal/scheduler/AsyncScheduler.js").AsyncScheduler)(r.AsyncAction)},"./node_modules/rxjs/_esm5/internal/scheduler/queue.js":function(e,t,n){"use strict";n.r(t),n.d(t,"queue",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/scheduler/QueueAction.js"),i=new(n("./node_modules/rxjs/_esm5/internal/scheduler/QueueScheduler.js").QueueScheduler)(r.QueueAction)},"./node_modules/rxjs/_esm5/internal/symbol/iterator.js":function(e,t,n){"use strict";function r(){return"function"==typeof Symbol&&Symbol.iterator?Symbol.iterator:"@@iterator"}n.r(t),n.d(t,"getSymbolIterator",function(){return r}),n.d(t,"iterator",function(){return i}),n.d(t,"$$iterator",function(){return o});var i=r(),o=i},"./node_modules/rxjs/_esm5/internal/symbol/observable.js":function(e,t,n){"use strict";n.r(t),n.d(t,"observable",function(){return r});var r="function"==typeof Symbol&&Symbol.observable||"@@observable"},"./node_modules/rxjs/_esm5/internal/symbol/rxSubscriber.js":function(e,t,n){"use strict";n.r(t),n.d(t,"rxSubscriber",function(){return r}),n.d(t,"$$rxSubscriber",function(){return i});var r="function"==typeof Symbol?Symbol("rxSubscriber"):"@@rxSubscriber_"+Math.random(),i=r},"./node_modules/rxjs/_esm5/internal/util/ArgumentOutOfRangeError.js":function(e,t,n){"use strict";function r(){return Error.call(this),this.message="argument out of range",this.name="ArgumentOutOfRangeError",this}n.r(t),n.d(t,"ArgumentOutOfRangeError",function(){return i}),r.prototype=Object.create(Error.prototype);var i=r},"./node_modules/rxjs/_esm5/internal/util/EmptyError.js":function(e,t,n){"use strict";function r(){return Error.call(this),this.message="no elements in sequence",this.name="EmptyError",this}n.r(t),n.d(t,"EmptyError",function(){return i}),r.prototype=Object.create(Error.prototype);var i=r},"./node_modules/rxjs/_esm5/internal/util/Immediate.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Immediate",function(){return o});var r=1,i={};var o={setImmediate:function(e){var t=r++;return i[t]=e,Promise.resolve().then(function(){var e;(e=i[t])&&e()}),t},clearImmediate:function(e){delete i[e]}}},"./node_modules/rxjs/_esm5/internal/util/ObjectUnsubscribedError.js":function(e,t,n){"use strict";function r(){return Error.call(this),this.message="object unsubscribed",this.name="ObjectUnsubscribedError",this}n.r(t),n.d(t,"ObjectUnsubscribedError",function(){return i}),r.prototype=Object.create(Error.prototype);var i=r},"./node_modules/rxjs/_esm5/internal/util/TimeoutError.js":function(e,t,n){"use strict";function r(){return Error.call(this),this.message="Timeout has occurred",this.name="TimeoutError",this}n.r(t),n.d(t,"TimeoutError",function(){return i}),r.prototype=Object.create(Error.prototype);var i=r},"./node_modules/rxjs/_esm5/internal/util/UnsubscriptionError.js":function(e,t,n){"use strict";function r(e){return Error.call(this),this.message=e?e.length+" errors occurred during unsubscription:\n"+e.map(function(e,t){return t+1+") "+e.toString()}).join("\n  "):"",this.name="UnsubscriptionError",this.errors=e,this}n.r(t),n.d(t,"UnsubscriptionError",function(){return i}),r.prototype=Object.create(Error.prototype);var i=r},"./node_modules/rxjs/_esm5/internal/util/canReportError.js":function(e,t,n){"use strict";n.r(t),n.d(t,"canReportError",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/Subscriber.js");function i(e){for(;e;){var t=e,n=t.closed,i=t.destination,o=t.isStopped;if(n||o)return!1;e=i&&i instanceof r.Subscriber?i:null}return!0}},"./node_modules/rxjs/_esm5/internal/util/errorObject.js":function(e,t,n){"use strict";n.r(t),n.d(t,"errorObject",function(){return r});var r={e:{}}},"./node_modules/rxjs/_esm5/internal/util/hostReportError.js":function(e,t,n){"use strict";function r(e){setTimeout(function(){throw e})}n.r(t),n.d(t,"hostReportError",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/identity.js":function(e,t,n){"use strict";function r(e){return e}n.r(t),n.d(t,"identity",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/isArray.js":function(e,t,n){"use strict";n.r(t),n.d(t,"isArray",function(){return r});var r=Array.isArray||function(e){return e&&"number"==typeof e.length}},"./node_modules/rxjs/_esm5/internal/util/isArrayLike.js":function(e,t,n){"use strict";n.r(t),n.d(t,"isArrayLike",function(){return r});var r=function(e){return e&&"number"==typeof e.length&&"function"!=typeof e}},"./node_modules/rxjs/_esm5/internal/util/isDate.js":function(e,t,n){"use strict";function r(e){return e instanceof Date&&!isNaN(+e)}n.r(t),n.d(t,"isDate",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/isFunction.js":function(e,t,n){"use strict";function r(e){return"function"==typeof e}n.r(t),n.d(t,"isFunction",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/isInteropObservable.js":function(e,t,n){"use strict";n.r(t),n.d(t,"isInteropObservable",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/symbol/observable.js");function i(e){return e&&"function"==typeof e[r.observable]}},"./node_modules/rxjs/_esm5/internal/util/isIterable.js":function(e,t,n){"use strict";n.r(t),n.d(t,"isIterable",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/symbol/iterator.js");function i(e){return e&&"function"==typeof e[r.iterator]}},"./node_modules/rxjs/_esm5/internal/util/isNumeric.js":function(e,t,n){"use strict";n.r(t),n.d(t,"isNumeric",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/util/isArray.js");function i(e){return!Object(r.isArray)(e)&&e-parseFloat(e)+1>=0}},"./node_modules/rxjs/_esm5/internal/util/isObject.js":function(e,t,n){"use strict";function r(e){return null!=e&&"object"==typeof e}n.r(t),n.d(t,"isObject",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/isObservable.js":function(e,t,n){"use strict";n.r(t),n.d(t,"isObservable",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js");function i(e){return!!e&&(e instanceof r.Observable||"function"==typeof e.lift&&"function"==typeof e.subscribe)}},"./node_modules/rxjs/_esm5/internal/util/isPromise.js":function(e,t,n){"use strict";function r(e){return e&&"function"!=typeof e.subscribe&&"function"==typeof e.then}n.r(t),n.d(t,"isPromise",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/isScheduler.js":function(e,t,n){"use strict";function r(e){return e&&"function"==typeof e.schedule}n.r(t),n.d(t,"isScheduler",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/noop.js":function(e,t,n){"use strict";function r(){}n.r(t),n.d(t,"noop",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/not.js":function(e,t,n){"use strict";function r(e,t){function n(){return!n.pred.apply(n.thisArg,arguments)}return n.pred=e,n.thisArg=t,n}n.r(t),n.d(t,"not",function(){return r})},"./node_modules/rxjs/_esm5/internal/util/pipe.js":function(e,t,n){"use strict";n.r(t),n.d(t,"pipe",function(){return i}),n.d(t,"pipeFromArray",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/util/noop.js");function i(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];return o(e)}function o(e){return e?1===e.length?e[0]:function(t){return e.reduce(function(e,t){return t(e)},t)}:r.noop}},"./node_modules/rxjs/_esm5/internal/util/subscribeTo.js":function(e,t,n){"use strict";n.r(t),n.d(t,"subscribeTo",function(){return f});var r=n("./node_modules/rxjs/_esm5/internal/Observable.js"),i=n("./node_modules/rxjs/_esm5/internal/util/subscribeToArray.js"),o=n("./node_modules/rxjs/_esm5/internal/util/subscribeToPromise.js"),a=n("./node_modules/rxjs/_esm5/internal/util/subscribeToIterable.js"),s=n("./node_modules/rxjs/_esm5/internal/util/subscribeToObservable.js"),u=n("./node_modules/rxjs/_esm5/internal/util/isArrayLike.js"),c=n("./node_modules/rxjs/_esm5/internal/util/isPromise.js"),l=n("./node_modules/rxjs/_esm5/internal/util/isObject.js"),d=n("./node_modules/rxjs/_esm5/internal/symbol/iterator.js"),p=n("./node_modules/rxjs/_esm5/internal/symbol/observable.js"),f=function(e){if(e instanceof r.Observable)return function(t){return e._isScalar?(t.next(e.value),void t.complete()):e.subscribe(t)};if(e&&"function"==typeof e[p.observable])return Object(s.subscribeToObservable)(e);if(Object(u.isArrayLike)(e))return Object(i.subscribeToArray)(e);if(Object(c.isPromise)(e))return Object(o.subscribeToPromise)(e);if(e&&"function"==typeof e[d.iterator])return Object(a.subscribeToIterable)(e);var t=Object(l.isObject)(e)?"an invalid object":"'"+e+"'";throw new TypeError("You provided "+t+" where a stream was expected. You can provide an Observable, Promise, Array, or Iterable.")}},"./node_modules/rxjs/_esm5/internal/util/subscribeToArray.js":function(e,t,n){"use strict";n.r(t),n.d(t,"subscribeToArray",function(){return r});var r=function(e){return function(t){for(var n=0,r=e.length;n<r&&!t.closed;n++)t.next(e[n]);t.closed||t.complete()}}},"./node_modules/rxjs/_esm5/internal/util/subscribeToIterable.js":function(e,t,n){"use strict";n.r(t),n.d(t,"subscribeToIterable",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/symbol/iterator.js"),i=function(e){return function(t){for(var n=e[r.iterator]();;){var i=n.next();if(i.done){t.complete();break}if(t.next(i.value),t.closed)break}return"function"==typeof n.return&&t.add(function(){n.return&&n.return()}),t}}},"./node_modules/rxjs/_esm5/internal/util/subscribeToObservable.js":function(e,t,n){"use strict";n.r(t),n.d(t,"subscribeToObservable",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/symbol/observable.js"),i=function(e){return function(t){var n=e[r.observable]();if("function"!=typeof n.subscribe)throw new TypeError("Provided object does not correctly implement Symbol.observable");return n.subscribe(t)}}},"./node_modules/rxjs/_esm5/internal/util/subscribeToPromise.js":function(e,t,n){"use strict";n.r(t),n.d(t,"subscribeToPromise",function(){return i});var r=n("./node_modules/rxjs/_esm5/internal/util/hostReportError.js"),i=function(e){return function(t){return e.then(function(e){t.closed||(t.next(e),t.complete())},function(e){return t.error(e)}).then(null,r.hostReportError),t}}},"./node_modules/rxjs/_esm5/internal/util/subscribeToResult.js":function(e,t,n){"use strict";n.r(t),n.d(t,"subscribeToResult",function(){return o});var r=n("./node_modules/rxjs/_esm5/internal/InnerSubscriber.js"),i=n("./node_modules/rxjs/_esm5/internal/util/subscribeTo.js");function o(e,t,n,o,a){if(void 0===a&&(a=new r.InnerSubscriber(e,n,o)),!a.closed)return Object(i.subscribeTo)(t)(a)}},"./node_modules/rxjs/_esm5/internal/util/toSubscriber.js":function(e,t,n){"use strict";n.r(t),n.d(t,"toSubscriber",function(){return a});var r=n("./node_modules/rxjs/_esm5/internal/Subscriber.js"),i=n("./node_modules/rxjs/_esm5/internal/symbol/rxSubscriber.js"),o=n("./node_modules/rxjs/_esm5/internal/Observer.js");function a(e,t,n){if(e){if(e instanceof r.Subscriber)return e;if(e[i.rxSubscriber])return e[i.rxSubscriber]()}return e||t||n?new r.Subscriber(e,t,n):new r.Subscriber(o.empty)}},"./node_modules/rxjs/_esm5/internal/util/tryCatch.js":function(e,t,n){"use strict";n.r(t),n.d(t,"tryCatch",function(){return a});var r,i=n("./node_modules/rxjs/_esm5/internal/util/errorObject.js");function o(){try{return r.apply(this,arguments)}catch(e){return i.errorObject.e=e,i.errorObject}}function a(e){return r=e,o}},"./node_modules/rxjs/_esm5/operators/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/rxjs/_esm5/internal/operators/audit.js");n.d(t,"audit",function(){return r.audit});var i=n("./node_modules/rxjs/_esm5/internal/operators/auditTime.js");n.d(t,"auditTime",function(){return i.auditTime});var o=n("./node_modules/rxjs/_esm5/internal/operators/buffer.js");n.d(t,"buffer",function(){return o.buffer});var a=n("./node_modules/rxjs/_esm5/internal/operators/bufferCount.js");n.d(t,"bufferCount",function(){return a.bufferCount});var s=n("./node_modules/rxjs/_esm5/internal/operators/bufferTime.js");n.d(t,"bufferTime",function(){return s.bufferTime});var u=n("./node_modules/rxjs/_esm5/internal/operators/bufferToggle.js");n.d(t,"bufferToggle",function(){return u.bufferToggle});var c=n("./node_modules/rxjs/_esm5/internal/operators/bufferWhen.js");n.d(t,"bufferWhen",function(){return c.bufferWhen});var l=n("./node_modules/rxjs/_esm5/internal/operators/catchError.js");n.d(t,"catchError",function(){return l.catchError});var d=n("./node_modules/rxjs/_esm5/internal/operators/combineAll.js");n.d(t,"combineAll",function(){return d.combineAll});var p=n("./node_modules/rxjs/_esm5/internal/operators/combineLatest.js");n.d(t,"combineLatest",function(){return p.combineLatest});var f=n("./node_modules/rxjs/_esm5/internal/operators/concat.js");n.d(t,"concat",function(){return f.concat});var h=n("./node_modules/rxjs/_esm5/internal/operators/concatAll.js");n.d(t,"concatAll",function(){return h.concatAll});var m=n("./node_modules/rxjs/_esm5/internal/operators/concatMap.js");n.d(t,"concatMap",function(){return m.concatMap});var g=n("./node_modules/rxjs/_esm5/internal/operators/concatMapTo.js");n.d(t,"concatMapTo",function(){return g.concatMapTo});var v=n("./node_modules/rxjs/_esm5/internal/operators/count.js");n.d(t,"count",function(){return v.count});var _=n("./node_modules/rxjs/_esm5/internal/operators/debounce.js");n.d(t,"debounce",function(){return _.debounce});var y=n("./node_modules/rxjs/_esm5/internal/operators/debounceTime.js");n.d(t,"debounceTime",function(){return y.debounceTime});var b=n("./node_modules/rxjs/_esm5/internal/operators/defaultIfEmpty.js");n.d(t,"defaultIfEmpty",function(){return b.defaultIfEmpty});var j=n("./node_modules/rxjs/_esm5/internal/operators/delay.js");n.d(t,"delay",function(){return j.delay});var x=n("./node_modules/rxjs/_esm5/internal/operators/delayWhen.js");n.d(t,"delayWhen",function(){return x.delayWhen});var O=n("./node_modules/rxjs/_esm5/internal/operators/dematerialize.js");n.d(t,"dematerialize",function(){return O.dematerialize});var w=n("./node_modules/rxjs/_esm5/internal/operators/distinct.js");n.d(t,"distinct",function(){return w.distinct});var C=n("./node_modules/rxjs/_esm5/internal/operators/distinctUntilChanged.js");n.d(t,"distinctUntilChanged",function(){return C.distinctUntilChanged});var E=n("./node_modules/rxjs/_esm5/internal/operators/distinctUntilKeyChanged.js");n.d(t,"distinctUntilKeyChanged",function(){return E.distinctUntilKeyChanged});var S=n("./node_modules/rxjs/_esm5/internal/operators/elementAt.js");n.d(t,"elementAt",function(){return S.elementAt});var k=n("./node_modules/rxjs/_esm5/internal/operators/endWith.js");n.d(t,"endWith",function(){return k.endWith});var M=n("./node_modules/rxjs/_esm5/internal/operators/every.js");n.d(t,"every",function(){return M.every});var A=n("./node_modules/rxjs/_esm5/internal/operators/exhaust.js");n.d(t,"exhaust",function(){return A.exhaust});var T=n("./node_modules/rxjs/_esm5/internal/operators/exhaustMap.js");n.d(t,"exhaustMap",function(){return T.exhaustMap});var D=n("./node_modules/rxjs/_esm5/internal/operators/expand.js");n.d(t,"expand",function(){return D.expand});var I=n("./node_modules/rxjs/_esm5/internal/operators/filter.js");n.d(t,"filter",function(){return I.filter});var P=n("./node_modules/rxjs/_esm5/internal/operators/finalize.js");n.d(t,"finalize",function(){return P.finalize});var R=n("./node_modules/rxjs/_esm5/internal/operators/find.js");n.d(t,"find",function(){return R.find});var N=n("./node_modules/rxjs/_esm5/internal/operators/findIndex.js");n.d(t,"findIndex",function(){return N.findIndex});var F=n("./node_modules/rxjs/_esm5/internal/operators/first.js");n.d(t,"first",function(){return F.first});var L=n("./node_modules/rxjs/_esm5/internal/operators/groupBy.js");n.d(t,"groupBy",function(){return L.groupBy});var V=n("./node_modules/rxjs/_esm5/internal/operators/ignoreElements.js");n.d(t,"ignoreElements",function(){return V.ignoreElements});var B=n("./node_modules/rxjs/_esm5/internal/operators/isEmpty.js");n.d(t,"isEmpty",function(){return B.isEmpty});var z=n("./node_modules/rxjs/_esm5/internal/operators/last.js");n.d(t,"last",function(){return z.last});var U=n("./node_modules/rxjs/_esm5/internal/operators/map.js");n.d(t,"map",function(){return U.map});var H=n("./node_modules/rxjs/_esm5/internal/operators/mapTo.js");n.d(t,"mapTo",function(){return H.mapTo});var q=n("./node_modules/rxjs/_esm5/internal/operators/materialize.js");n.d(t,"materialize",function(){return q.materialize});var G=n("./node_modules/rxjs/_esm5/internal/operators/max.js");n.d(t,"max",function(){return G.max});var W=n("./node_modules/rxjs/_esm5/internal/operators/merge.js");n.d(t,"merge",function(){return W.merge});var Y=n("./node_modules/rxjs/_esm5/internal/operators/mergeAll.js");n.d(t,"mergeAll",function(){return Y.mergeAll});var K=n("./node_modules/rxjs/_esm5/internal/operators/mergeMap.js");n.d(t,"mergeMap",function(){return K.mergeMap}),n.d(t,"flatMap",function(){return K.mergeMap});var Q=n("./node_modules/rxjs/_esm5/internal/operators/mergeMapTo.js");n.d(t,"mergeMapTo",function(){return Q.mergeMapTo});var X=n("./node_modules/rxjs/_esm5/internal/operators/mergeScan.js");n.d(t,"mergeScan",function(){return X.mergeScan});var Z=n("./node_modules/rxjs/_esm5/internal/operators/min.js");n.d(t,"min",function(){return Z.min});var $=n("./node_modules/rxjs/_esm5/internal/operators/multicast.js");n.d(t,"multicast",function(){return $.multicast});var J=n("./node_modules/rxjs/_esm5/internal/operators/observeOn.js");n.d(t,"observeOn",function(){return J.observeOn});var ee=n("./node_modules/rxjs/_esm5/internal/operators/onErrorResumeNext.js");n.d(t,"onErrorResumeNext",function(){return ee.onErrorResumeNext});var te=n("./node_modules/rxjs/_esm5/internal/operators/pairwise.js");n.d(t,"pairwise",function(){return te.pairwise});var ne=n("./node_modules/rxjs/_esm5/internal/operators/partition.js");n.d(t,"partition",function(){return ne.partition});var re=n("./node_modules/rxjs/_esm5/internal/operators/pluck.js");n.d(t,"pluck",function(){return re.pluck});var ie=n("./node_modules/rxjs/_esm5/internal/operators/publish.js");n.d(t,"publish",function(){return ie.publish});var oe=n("./node_modules/rxjs/_esm5/internal/operators/publishBehavior.js");n.d(t,"publishBehavior",function(){return oe.publishBehavior});var ae=n("./node_modules/rxjs/_esm5/internal/operators/publishLast.js");n.d(t,"publishLast",function(){return ae.publishLast});var se=n("./node_modules/rxjs/_esm5/internal/operators/publishReplay.js");n.d(t,"publishReplay",function(){return se.publishReplay});var ue=n("./node_modules/rxjs/_esm5/internal/operators/race.js");n.d(t,"race",function(){return ue.race});var ce=n("./node_modules/rxjs/_esm5/internal/operators/reduce.js");n.d(t,"reduce",function(){return ce.reduce});var le=n("./node_modules/rxjs/_esm5/internal/operators/repeat.js");n.d(t,"repeat",function(){return le.repeat});var de=n("./node_modules/rxjs/_esm5/internal/operators/repeatWhen.js");n.d(t,"repeatWhen",function(){return de.repeatWhen});var pe=n("./node_modules/rxjs/_esm5/internal/operators/retry.js");n.d(t,"retry",function(){return pe.retry});var fe=n("./node_modules/rxjs/_esm5/internal/operators/retryWhen.js");n.d(t,"retryWhen",function(){return fe.retryWhen});var he=n("./node_modules/rxjs/_esm5/internal/operators/refCount.js");n.d(t,"refCount",function(){return he.refCount});var me=n("./node_modules/rxjs/_esm5/internal/operators/sample.js");n.d(t,"sample",function(){return me.sample});var ge=n("./node_modules/rxjs/_esm5/internal/operators/sampleTime.js");n.d(t,"sampleTime",function(){return ge.sampleTime});var ve=n("./node_modules/rxjs/_esm5/internal/operators/scan.js");n.d(t,"scan",function(){return ve.scan});var _e=n("./node_modules/rxjs/_esm5/internal/operators/sequenceEqual.js");n.d(t,"sequenceEqual",function(){return _e.sequenceEqual});var ye=n("./node_modules/rxjs/_esm5/internal/operators/share.js");n.d(t,"share",function(){return ye.share});var be=n("./node_modules/rxjs/_esm5/internal/operators/shareReplay.js");n.d(t,"shareReplay",function(){return be.shareReplay});var je=n("./node_modules/rxjs/_esm5/internal/operators/single.js");n.d(t,"single",function(){return je.single});var xe=n("./node_modules/rxjs/_esm5/internal/operators/skip.js");n.d(t,"skip",function(){return xe.skip});var Oe=n("./node_modules/rxjs/_esm5/internal/operators/skipLast.js");n.d(t,"skipLast",function(){return Oe.skipLast});var we=n("./node_modules/rxjs/_esm5/internal/operators/skipUntil.js");n.d(t,"skipUntil",function(){return we.skipUntil});var Ce=n("./node_modules/rxjs/_esm5/internal/operators/skipWhile.js");n.d(t,"skipWhile",function(){return Ce.skipWhile});var Ee=n("./node_modules/rxjs/_esm5/internal/operators/startWith.js");n.d(t,"startWith",function(){return Ee.startWith});var Se=n("./node_modules/rxjs/_esm5/internal/operators/subscribeOn.js");n.d(t,"subscribeOn",function(){return Se.subscribeOn});var ke=n("./node_modules/rxjs/_esm5/internal/operators/switchAll.js");n.d(t,"switchAll",function(){return ke.switchAll});var Me=n("./node_modules/rxjs/_esm5/internal/operators/switchMap.js");n.d(t,"switchMap",function(){return Me.switchMap});var Ae=n("./node_modules/rxjs/_esm5/internal/operators/switchMapTo.js");n.d(t,"switchMapTo",function(){return Ae.switchMapTo});var Te=n("./node_modules/rxjs/_esm5/internal/operators/take.js");n.d(t,"take",function(){return Te.take});var De=n("./node_modules/rxjs/_esm5/internal/operators/takeLast.js");n.d(t,"takeLast",function(){return De.takeLast});var Ie=n("./node_modules/rxjs/_esm5/internal/operators/takeUntil.js");n.d(t,"takeUntil",function(){return Ie.takeUntil});var Pe=n("./node_modules/rxjs/_esm5/internal/operators/takeWhile.js");n.d(t,"takeWhile",function(){return Pe.takeWhile});var Re=n("./node_modules/rxjs/_esm5/internal/operators/tap.js");n.d(t,"tap",function(){return Re.tap});var Ne=n("./node_modules/rxjs/_esm5/internal/operators/throttle.js");n.d(t,"throttle",function(){return Ne.throttle});var Fe=n("./node_modules/rxjs/_esm5/internal/operators/throttleTime.js");n.d(t,"throttleTime",function(){return Fe.throttleTime});var Le=n("./node_modules/rxjs/_esm5/internal/operators/throwIfEmpty.js");n.d(t,"throwIfEmpty",function(){return Le.throwIfEmpty});var Ve=n("./node_modules/rxjs/_esm5/internal/operators/timeInterval.js");n.d(t,"timeInterval",function(){return Ve.timeInterval});var Be=n("./node_modules/rxjs/_esm5/internal/operators/timeout.js");n.d(t,"timeout",function(){return Be.timeout});var ze=n("./node_modules/rxjs/_esm5/internal/operators/timeoutWith.js");n.d(t,"timeoutWith",function(){return ze.timeoutWith});var Ue=n("./node_modules/rxjs/_esm5/internal/operators/timestamp.js");n.d(t,"timestamp",function(){return Ue.timestamp});var He=n("./node_modules/rxjs/_esm5/internal/operators/toArray.js");n.d(t,"toArray",function(){return He.toArray});var qe=n("./node_modules/rxjs/_esm5/internal/operators/window.js");n.d(t,"window",function(){return qe.window});var Ge=n("./node_modules/rxjs/_esm5/internal/operators/windowCount.js");n.d(t,"windowCount",function(){return Ge.windowCount});var We=n("./node_modules/rxjs/_esm5/internal/operators/windowTime.js");n.d(t,"windowTime",function(){return We.windowTime});var Ye=n("./node_modules/rxjs/_esm5/internal/operators/windowToggle.js");n.d(t,"windowToggle",function(){return Ye.windowToggle});var Ke=n("./node_modules/rxjs/_esm5/internal/operators/windowWhen.js");n.d(t,"windowWhen",function(){return Ke.windowWhen});var Qe=n("./node_modules/rxjs/_esm5/internal/operators/withLatestFrom.js");n.d(t,"withLatestFrom",function(){return Qe.withLatestFrom});var Xe=n("./node_modules/rxjs/_esm5/internal/operators/zip.js");n.d(t,"zip",function(){return Xe.zip});var Ze=n("./node_modules/rxjs/_esm5/internal/operators/zipAll.js");n.d(t,"zipAll",function(){return Ze.zipAll})},"./node_modules/rxjs/node_modules/tslib/tslib.es6.js":function(e,t,n){"use strict";n.r(t),n.d(t,"__extends",function(){return i}),n.d(t,"__assign",function(){return o}),n.d(t,"__rest",function(){return a}),n.d(t,"__decorate",function(){return s}),n.d(t,"__param",function(){return u}),n.d(t,"__metadata",function(){return c}),n.d(t,"__awaiter",function(){return l}),n.d(t,"__generator",function(){return d}),n.d(t,"__exportStar",function(){return p}),n.d(t,"__values",function(){return f}),n.d(t,"__read",function(){return h}),n.d(t,"__spread",function(){return m}),n.d(t,"__await",function(){return g}),n.d(t,"__asyncGenerator",function(){return v}),n.d(t,"__asyncDelegator",function(){return _}),n.d(t,"__asyncValues",function(){return y}),n.d(t,"__makeTemplateObject",function(){return b}),n.d(t,"__importStar",function(){return j}),n.d(t,"__importDefault",function(){return x});var r=function(e,t){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,t){e.__proto__=t}||function(e,t){for(var n in t)t.hasOwnProperty(n)&&(e[n]=t[n])})(e,t)};function i(e,t){function n(){this.constructor=e}r(e,t),e.prototype=null===t?Object.create(t):(n.prototype=t.prototype,new n)}var o=function(){return(o=Object.assign||function(e){for(var t,n=1,r=arguments.length;n<r;n++){t=arguments[n];for(var i in t)Object.prototype.hasOwnProperty.call(t,i)&&(e[i]=t[i])}return e}).apply(this,arguments)};function a(e,t){var n={};for(var r in e)Object.prototype.hasOwnProperty.call(e,r)&&t.indexOf(r)<0&&(n[r]=e[r]);if(null!=e&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(e);i<r.length;i++)t.indexOf(r[i])<0&&(n[r[i]]=e[r[i]])}return n}function s(e,t,n,r){var i,o=arguments.length,a=o<3?t:null===r?r=Object.getOwnPropertyDescriptor(t,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(e,t,n,r);else for(var s=e.length-1;s>=0;s--)(i=e[s])&&(a=(o<3?i(a):o>3?i(t,n,a):i(t,n))||a);return o>3&&a&&Object.defineProperty(t,n,a),a}function u(e,t){return function(n,r){t(n,r,e)}}function c(e,t){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(e,t)}function l(e,t,n,r){return new(n||(n=Promise))(function(i,o){function a(e){try{u(r.next(e))}catch(e){o(e)}}function s(e){try{u(r.throw(e))}catch(e){o(e)}}function u(e){e.done?i(e.value):new n(function(t){t(e.value)}).then(a,s)}u((r=r.apply(e,t||[])).next())})}function d(e,t){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=t.call(e,a)}catch(e){o=[6,e],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(e,t){for(var n in e)t.hasOwnProperty(n)||(t[n]=e[n])}function f(e){var t="function"==typeof Symbol&&e[Symbol.iterator],n=0;return t?t.call(e):{next:function(){return e&&n>=e.length&&(e=void 0),{value:e&&e[n++],done:!e}}}}function h(e,t){var n="function"==typeof Symbol&&e[Symbol.iterator];if(!n)return e;var r,i,o=n.call(e),a=[];try{for(;(void 0===t||t-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(e){i={error:e}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var e=[],t=0;t<arguments.length;t++)e=e.concat(h(arguments[t]));return e}function g(e){return this instanceof g?(this.v=e,this):new g(e)}function v(e,t,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(e,t||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(e){i[e]&&(r[e]=function(t){return new Promise(function(n,r){o.push([e,t,n,r])>1||s(e,t)})})}function s(e,t){try{(n=i[e](t)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(e){l(o[0][3],e)}var n}function u(e){s("next",e)}function c(e){s("throw",e)}function l(e,t){e(t),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(e){var t,n;return t={},r("next"),r("throw",function(e){throw e}),r("return"),t[Symbol.iterator]=function(){return this},t;function r(r,i){t[r]=e[r]?function(t){return(n=!n)?{value:g(e[r](t)),done:"return"===r}:i?i(t):t}:i}}function y(e){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var t,n=e[Symbol.asyncIterator];return n?n.call(e):(e=f(e),t={},r("next"),r("throw"),r("return"),t[Symbol.asyncIterator]=function(){return this},t);function r(n){t[n]=e[n]&&function(t){return new Promise(function(r,i){var o,a,s,u;t=e[n](t),o=r,a=i,s=t.done,u=t.value,Promise.resolve(u).then(function(e){o({value:e,done:s})},a)})}}}function b(e,t){return Object.defineProperty?Object.defineProperty(e,"raw",{value:t}):e.raw=t,e}function j(e){if(e&&e.__esModule)return e;var t={};if(null!=e)for(var n in e)Object.hasOwnProperty.call(e,n)&&(t[n]=e[n]);return t.default=e,t}function x(e){return e&&e.__esModule?e:{default:e}}},"./node_modules/tslib/tslib.es6.js":function(e,t,n){"use strict";n.r(t),n.d(t,"__extends",function(){return i}),n.d(t,"__assign",function(){return o}),n.d(t,"__rest",function(){return a}),n.d(t,"__decorate",function(){return s}),n.d(t,"__param",function(){return u}),n.d(t,"__metadata",function(){return c}),n.d(t,"__awaiter",function(){return l}),n.d(t,"__generator",function(){return d}),n.d(t,"__exportStar",function(){return p}),n.d(t,"__values",function(){return f}),n.d(t,"__read",function(){return h}),n.d(t,"__spread",function(){return m}),n.d(t,"__await",function(){return g}),n.d(t,"__asyncGenerator",function(){return v}),n.d(t,"__asyncDelegator",function(){return _}),n.d(t,"__asyncValues",function(){return y}),n.d(t,"__makeTemplateObject",function(){return b}),n.d(t,"__importStar",function(){return j}),n.d(t,"__importDefault",function(){return x});var r=function(e,t){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,t){e.__proto__=t}||function(e,t){for(var n in t)t.hasOwnProperty(n)&&(e[n]=t[n])})(e,t)};function i(e,t){function n(){this.constructor=e}r(e,t),e.prototype=null===t?Object.create(t):(n.prototype=t.prototype,new n)}var o=function(){return(o=Object.assign||function(e){for(var t,n=1,r=arguments.length;n<r;n++){t=arguments[n];for(var i in t)Object.prototype.hasOwnProperty.call(t,i)&&(e[i]=t[i])}return e}).apply(this,arguments)};function a(e,t){var n={};for(var r in e)Object.prototype.hasOwnProperty.call(e,r)&&t.indexOf(r)<0&&(n[r]=e[r]);if(null!=e&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(e);i<r.length;i++)t.indexOf(r[i])<0&&(n[r[i]]=e[r[i]])}return n}function s(e,t,n,r){var i,o=arguments.length,a=o<3?t:null===r?r=Object.getOwnPropertyDescriptor(t,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)a=Reflect.decorate(e,t,n,r);else for(var s=e.length-1;s>=0;s--)(i=e[s])&&(a=(o<3?i(a):o>3?i(t,n,a):i(t,n))||a);return o>3&&a&&Object.defineProperty(t,n,a),a}function u(e,t){return function(n,r){t(n,r,e)}}function c(e,t){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(e,t)}function l(e,t,n,r){return new(n||(n=Promise))(function(i,o){function a(e){try{u(r.next(e))}catch(e){o(e)}}function s(e){try{u(r.throw(e))}catch(e){o(e)}}function u(e){e.done?i(e.value):new n(function(t){t(e.value)}).then(a,s)}u((r=r.apply(e,t||[])).next())})}function d(e,t){var n,r,i,o,a={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function s(o){return function(s){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;a;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return a.label++,{value:o[1],done:!1};case 5:a.label++,r=o[1],o=[0];continue;case 7:o=a.ops.pop(),a.trys.pop();continue;default:if(!(i=(i=a.trys).length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){a=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){a.label=o[1];break}if(6===o[0]&&a.label<i[1]){a.label=i[1],i=o;break}if(i&&a.label<i[2]){a.label=i[2],a.ops.push(o);break}i[2]&&a.ops.pop(),a.trys.pop();continue}o=t.call(e,a)}catch(e){o=[6,e],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,s])}}}function p(e,t){for(var n in e)t.hasOwnProperty(n)||(t[n]=e[n])}function f(e){var t="function"==typeof Symbol&&e[Symbol.iterator],n=0;return t?t.call(e):{next:function(){return e&&n>=e.length&&(e=void 0),{value:e&&e[n++],done:!e}}}}function h(e,t){var n="function"==typeof Symbol&&e[Symbol.iterator];if(!n)return e;var r,i,o=n.call(e),a=[];try{for(;(void 0===t||t-- >0)&&!(r=o.next()).done;)a.push(r.value)}catch(e){i={error:e}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return a}function m(){for(var e=[],t=0;t<arguments.length;t++)e=e.concat(h(arguments[t]));return e}function g(e){return this instanceof g?(this.v=e,this):new g(e)}function v(e,t,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(e,t||[]),o=[];return r={},a("next"),a("throw"),a("return"),r[Symbol.asyncIterator]=function(){return this},r;function a(e){i[e]&&(r[e]=function(t){return new Promise(function(n,r){o.push([e,t,n,r])>1||s(e,t)})})}function s(e,t){try{(n=i[e](t)).value instanceof g?Promise.resolve(n.value.v).then(u,c):l(o[0][2],n)}catch(e){l(o[0][3],e)}var n}function u(e){s("next",e)}function c(e){s("throw",e)}function l(e,t){e(t),o.shift(),o.length&&s(o[0][0],o[0][1])}}function _(e){var t,n;return t={},r("next"),r("throw",function(e){throw e}),r("return"),t[Symbol.iterator]=function(){return this},t;function r(r,i){t[r]=e[r]?function(t){return(n=!n)?{value:g(e[r](t)),done:"return"===r}:i?i(t):t}:i}}function y(e){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var t,n=e[Symbol.asyncIterator];return n?n.call(e):(e=f(e),t={},r("next"),r("throw"),r("return"),t[Symbol.asyncIterator]=function(){return this},t);function r(n){t[n]=e[n]&&function(t){return new Promise(function(r,i){var o,a,s,u;t=e[n](t),o=r,a=i,s=t.done,u=t.value,Promise.resolve(u).then(function(e){o({value:e,done:s})},a)})}}}function b(e,t){return Object.defineProperty?Object.defineProperty(e,"raw",{value:t}):e.raw=t,e}function j(e){if(e&&e.__esModule)return e;var t={};if(null!=e)for(var n in e)Object.hasOwnProperty.call(e,n)&&(t[n]=e[n]);return t.default=e,t}function x(e){return e&&e.__esModule?e:{default:e}}},"./node_modules/vega-canvas/index.js":function(e,t,n){"use strict";n.r(t),n.d(t,"canvas",function(){return o}),n.d(t,"image",function(){return a});var r=n("./node_modules/vega-canvas/src/domCanvas.js"),i=n("./node_modules/vega-canvas/src/nodeCanvas.js");function o(e,t){return Object(r.domCanvas)(e,t)||Object(i.nodeCanvas)(e,t)||null}function a(){return Object(r.domImage)()||Object(i.nodeImage)()||null}n.d(t,"domCanvas",function(){return r.domCanvas}),n.d(t,"nodeCanvas",function(){return i.nodeCanvas})},"./node_modules/vega-canvas/src/domCanvas.js":function(e,t,n){"use strict";function r(e,t){if("undefined"!=typeof document&&document.createElement){var n=document.createElement("canvas");if(n&&n.getContext)return n.width=e,n.height=t,n}return null}function i(){return"undefined"!=typeof Image?Image:null}n.r(t),n.d(t,"domCanvas",function(){return r}),n.d(t,"domImage",function(){return i})},"./node_modules/vega-canvas/src/nodeCanvas.js":function(e,t,n){"use strict";var r;n.r(t),n.d(t,"nodeCanvas",function(){return i}),n.d(t,"nodeImage",function(){return o});try{if(!(r=n(2)))throw 1}catch(e){try{r=n(3)}catch(e){r=null}}function i(e,t){if(r)try{return new r(e,t)}catch(e){}return null}function o(){return r&&r.Image||null}},"./node_modules/vega-crossfilter/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-crossfilter/src/CrossFilter.js");n.d(t,"crossfilter",function(){return r.default});var i=n("./node_modules/vega-crossfilter/src/ResolveFilter.js");n.d(t,"resolvefilter",function(){return i.default})},"./node_modules/vega-crossfilter/src/Bitmaps.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return i});var r=n("./node_modules/vega-crossfilter/src/arrays.js");function i(){var e=8,t=[],n=Object(r.array32)(0),i=o(0,e),a=o(0,e);return{data:function(){return t},seen:function(){return e=n,r=t.length,n=e.length>=r?e:((i=i||new e.constructor(r)).set(e),i);var e,r,i},add:function(e){for(var n,r=0,i=t.length,o=e.length;r<o;++r)(n=e[r])._index=i++,t.push(n)},remove:function(e,n){var r,o,s,u=t.length,c=Array(u-e),l=t;for(o=0;!n[o]&&o<u;++o)c[o]=t[o],l[o]=o;for(s=o;o<u;++o)r=t[o],n[o]?l[o]=-1:(l[o]=s,i[s]=i[o],a[s]=a[o],c[s]=r,r._index=s++),i[o]=0;return t=c,l},size:function(){return t.length},curr:function(){return i},prev:function(){return a},reset:function(e){a[e]=i[e]},all:function(){return e<257?255:e<65537?65535:4294967295},set:function(e,t){i[e]|=t},clear:function(e,t){i[e]&=~t},resize:function(t,n){(t>i.length||n>e)&&(e=Math.max(n,e),i=o(t,e,i),a=o(t,e))}}}function o(e,t,n){var i=(t<257?r.array8:t<65537?r.array16:r.array32)(e);return n&&i.set(n),i}},"./node_modules/vega-crossfilter/src/CrossFilter.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return u});var r=n("./node_modules/vega-crossfilter/src/Bitmaps.js"),i=n("./node_modules/vega-crossfilter/src/Dimension.js"),o=n("./node_modules/vega-crossfilter/src/SortedIndex.js"),a=n("./node_modules/vega-dataflow/index.js"),s=n("./node_modules/vega-util/index.js");function u(e){a.Transform.call(this,Object(r.default)(),e),this._indices=null,this._dims=null}u.Definition={type:"CrossFilter",metadata:{},params:[{name:"fields",type:"field",array:!0,required:!0},{name:"query",type:"array",array:!0,required:!0,content:{type:"number",array:!0,length:2}}]};var c=Object(s.inherits)(u,a.Transform);c.transform=function(e,t){return this._dims?e.modified("fields")||e.fields.some(function(e){return t.modified(e.fields)})?this.reinit(e,t):this.eval(e,t):this.init(e,t)},c.init=function(e,t){for(var n,r,a=e.fields,s=e.query,u=this._indices={},c=this._dims=[],l=s.length,d=0;d<l;++d)r=u[n=a[d].fname]||(u[n]=Object(o.default)()),c.push(Object(i.default)(r,d,s[d]));return this.eval(e,t)},c.reinit=function(e,t){var n,r,a,s,u,c,l,d,p,f=t.materialize().fork(),h=e.fields,m=e.query,g=this._indices,v=this._dims,_=this.value,y=_.curr(),b=_.prev(),j=_.all(),x=f.rem=f.add,O=f.mod,w=m.length,C={};if(b.set(y),t.rem.length&&(u=this.remove(e,t,f)),t.add.length&&_.add(t.add),t.mod.length)for(c={},l=0,d=(s=t.mod).length;l<d;++l)c[s[l]._index]=1;for(l=0;l<w;++l)p=h[l],(!v[l]||e.modified("fields",l)||t.modified(p.fields))&&((n=C[a=p.fname])||(g[a]=r=Object(o.default)(),C[a]=n=r.insert(p,t.source,0)),v[l]=Object(i.default)(r,l,m[l]).onAdd(n,y));for(l=0,d=_.data().length;l<d;++l)u[l]||(b[l]!==y[l]?x.push(l):c[l]&&y[l]!==j&&O.push(l));return _.mask=(1<<w)-1,f},c.eval=function(e,t){var n=t.materialize().fork(),r=this._dims.length,i=0;return t.rem.length&&(this.remove(e,t,n),i|=(1<<r)-1),e.modified("query")&&!e.modified("fields")&&(i|=this.update(e,t,n)),t.add.length&&(this.insert(e,t,n),i|=(1<<r)-1),t.mod.length&&(this.modify(t,n),i|=(1<<r)-1),this.value.mask=i,n},c.insert=function(e,t,n){var r,i,o,a=t.add,s=this.value,u=this._dims,c=this._indices,l=e.fields,d={},p=n.add,f=s.size(),h=f+a.length,m=u.length;s.resize(h,m),s.add(a);var g=s.curr(),v=s.prev(),_=s.all();for(r=0;r<m;++r)o=d[i=l[r].fname]||(d[i]=c[i].insert(l[r],a,f)),u[r].onAdd(o,g);for(;f<h;++f)v[f]=_,g[f]!==_&&p.push(f)},c.modify=function(e,t){var n,r,i,o=t.mod,a=this.value,s=a.curr(),u=a.all(),c=e.mod;for(n=0,r=c.length;n<r;++n)s[i=c[n]._index]!==u&&o.push(i)},c.remove=function(e,t,n){var r,i,o,a,s=this._indices,u=this.value,c=u.curr(),l=u.prev(),d=u.all(),p={},f=n.rem,h=t.rem;for(r=0,i=h.length;r<i;++r)p[o=h[r]._index]=1,l[o]=a=c[o],c[o]=d,a!==d&&f.push(o);for(o in s)s[o].remove(i,p);return this.reindex(t,i,p),p},c.reindex=function(e,t,n){var r=this._indices,i=this.value;e.runAfter(function(){var e=i.remove(t,n);for(var o in r)r[o].reindex(e)})},c.update=function(e,t,n){var r,i,o=this._dims,a=e.query,s=t.stamp,u=o.length,c=0;for(n.filters=0,i=0;i<u;++i)e.modified("query",i)&&(r=i,++c);if(1===c)c=o[r].one,this.incrementOne(o[r],a[r],n.add,n.rem);else for(i=0,c=0;i<u;++i)e.modified("query",i)&&(c|=o[i].one,this.incrementAll(o[i],a[i],s,n.add),n.rem=n.add);return c},c.incrementAll=function(e,t,n,r){var i,o,a,s=this.value,u=s.seen(),c=s.curr(),l=s.prev(),d=e.index(),p=e.bisect(e.range),f=e.bisect(t),h=f[0],m=f[1],g=p[0],v=p[1],_=e.one;if(h<g)for(i=h,o=Math.min(g,m);i<o;++i)u[a=d[i]]!==n&&(l[a]=c[a],u[a]=n,r.push(a)),c[a]^=_;else if(h>g)for(i=g,o=Math.min(h,v);i<o;++i)u[a=d[i]]!==n&&(l[a]=c[a],u[a]=n,r.push(a)),c[a]^=_;if(m>v)for(i=Math.max(h,v),o=m;i<o;++i)u[a=d[i]]!==n&&(l[a]=c[a],u[a]=n,r.push(a)),c[a]^=_;else if(m<v)for(i=Math.max(g,m),o=v;i<o;++i)u[a=d[i]]!==n&&(l[a]=c[a],u[a]=n,r.push(a)),c[a]^=_;e.range=t.slice()},c.incrementOne=function(e,t,n,r){var i,o,a,s=this.value.curr(),u=e.index(),c=e.bisect(e.range),l=e.bisect(t),d=l[0],p=l[1],f=c[0],h=c[1],m=e.one;if(d<f)for(i=d,o=Math.min(f,p);i<o;++i)s[a=u[i]]^=m,n.push(a);else if(d>f)for(i=f,o=Math.min(d,h);i<o;++i)s[a=u[i]]^=m,r.push(a);if(p>h)for(i=Math.max(d,h),o=p;i<o;++i)s[a=u[i]]^=m,n.push(a);else if(p<h)for(i=Math.max(f,p),o=h;i<o;++i)s[a=u[i]]^=m,r.push(a);e.range=t.slice()}},"./node_modules/vega-crossfilter/src/Dimension.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n){var r=1<<t;return{one:r,zero:~r,range:n.slice(),bisect:e.bisect,index:e.index,size:e.size,onAdd:function(e,t){var n,i=this.bisect(this.range,e.value),o=e.index,a=i[0],s=i[1],u=o.length;for(n=0;n<a;++n)t[o[n]]|=r;for(n=s;n<u;++n)t[o[n]]|=r;return this}}}},"./node_modules/vega-crossfilter/src/ResolveFilter.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,null,e)}o.Definition={type:"ResolveFilter",metadata:{},params:[{name:"ignore",type:"number",required:!0,description:"A bit mask indicating which filters to ignore."},{name:"filter",type:"object",required:!0,description:"Per-tuple filter bitmaps from a CrossFilter transform."}]},Object(i.inherits)(o,r.Transform).transform=function(e,t){var n=~(e.ignore||0),r=e.filter,i=r.mask;if(0==(i&n))return t.StopPropagation;var o=t.fork(t.ALL),a=r.data(),s=r.curr(),u=r.prev(),c=function(e){return s[e]&n?null:a[e]};return o.filter(o.MOD,c),i&i-1?(o.filter(o.ADD,function(e){var t=s[e]&n;return!t&&t^u[e]&n?a[e]:null}),o.filter(o.REM,function(e){var t=s[e]&n;return t&&!(t^t^u[e]&n)?a[e]:null})):(o.filter(o.ADD,c),o.filter(o.REM,function(e){return(s[e]&n)===i?a[e]:null})),o.filter(o.SOURCE,function(e){return c(e._index)})}},"./node_modules/vega-crossfilter/src/SortedIndex.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-crossfilter/src/arrays.js"),i=n("./node_modules/d3-array/index.js");function o(){var e=Object(r.array32)(0),t=[],n=0;return{insert:function(o,a,s){if(!a.length)return[];var u,c,l,d,p,f=n,h=a.length,m=Array(h),g=Object(r.array32)(h);for(l=0;l<h;++l)m[l]=o(a[l]),g[l]=l;if(p=g,(d=m).sort.call(p,function(e,t){var n=d[e],r=d[t];return n<r?-1:n>r?1:0}),m=Object(i.permute)(d,p),f)u=t,c=e,t=Array(f+h),e=Object(r.array32)(f+h),function(e,t,n,r,i,o,a,s,u){var c,l=0,d=0;for(c=0;l<r&&d<a;++c)t[l]<i[d]?(s[c]=t[l],u[c]=n[l++]):(s[c]=i[d],u[c]=o[d++]+e);for(;l<r;++l,++c)s[c]=t[l],u[c]=n[l];for(;d<a;++d,++c)s[c]=i[d],u[c]=o[d]+e}(s,u,c,f,m,g,h,t,e);else{if(s>0)for(l=0;l<h;++l)g[l]+=s;t=m,e=g}return n=f+h,{index:g,value:m}},remove:function(r,i){var o,a,s,u=n;for(a=0;!i[e[a]]&&a<u;++a);for(s=a;a<u;++a)i[o=e[a]]||(e[s]=o,t[s]=t[a],++s);n=u-r},bisect:function(e,r){var o;return r?o=r.length:(r=t,o=n),[Object(i.bisectLeft)(r,e[0],0,o),Object(i.bisectRight)(r,e[1],0,o)]},reindex:function(t){for(var r=0,i=n;r<i;++r)e[r]=t[e[r]]},index:function(){return e},size:function(){return n}}}},"./node_modules/vega-crossfilter/src/arrays.js":function(e,t,n){"use strict";function r(e){return new Uint8Array(e)}function i(e){return new Uint16Array(e)}function o(e){return new Uint32Array(e)}n.r(t),n.d(t,"array8",function(){return r}),n.d(t,"array16",function(){return i}),n.d(t,"array32",function(){return o})},"./node_modules/vega-dataflow/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-dataflow/src/util/UniqueList.js");n.d(t,"UniqueList",function(){return r.default});var i=n("./node_modules/vega-dataflow/src/ChangeSet.js");n.d(t,"changeset",function(){return i.default}),n.d(t,"isChangeSet",function(){return i.isChangeSet});var o=n("./node_modules/vega-dataflow/src/dataflow/Dataflow.js");n.d(t,"Dataflow",function(){return o.default});var a=n("./node_modules/vega-dataflow/src/EventStream.js");n.d(t,"EventStream",function(){return a.default});var s=n("./node_modules/vega-dataflow/src/Parameters.js");n.d(t,"Parameters",function(){return s.default});var u=n("./node_modules/vega-dataflow/src/Pulse.js");n.d(t,"Pulse",function(){return u.default});var c=n("./node_modules/vega-dataflow/src/MultiPulse.js");n.d(t,"MultiPulse",function(){return c.default});var l=n("./node_modules/vega-dataflow/src/Operator.js");n.d(t,"Operator",function(){return l.default});var d=n("./node_modules/vega-dataflow/src/Transform.js");n.d(t,"Transform",function(){return d.default});var p=n("./node_modules/vega-dataflow/src/Tuple.js");n.d(t,"derive",function(){return p.derive}),n.d(t,"rederive",function(){return p.rederive}),n.d(t,"ingest",function(){return p.ingest}),n.d(t,"isTuple",function(){return p.isTuple}),n.d(t,"replace",function(){return p.replace}),n.d(t,"tupleid",function(){return p.tupleid});var f=n("./node_modules/vega-dataflow/src/register.js");n.d(t,"definition",function(){return f.definition}),n.d(t,"transform",function(){return f.transform}),n.d(t,"transforms",function(){return f.transforms})},"./node_modules/vega-dataflow/src/ChangeSet.js":function(e,t,n){"use strict";n.r(t),n.d(t,"isChangeSet",function(){return o}),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-dataflow/src/Tuple.js"),i=n("./node_modules/vega-util/index.js");function o(e){return e&&e.constructor===a}function a(){var e=[],t=[],n=[],o=[],s=[],u=!1;return{constructor:a,insert:function(t){for(var n=Object(i.array)(t),r=0,o=n.length;r<o;++r)e.push(n[r]);return this},remove:function(e){for(var n=Object(i.isFunction)(e)?o:t,r=Object(i.array)(e),a=0,s=r.length;a<s;++a)n.push(r[a]);return this},modify:function(e,t,r){var o={field:t,value:Object(i.constant)(r)};return Object(i.isFunction)(e)?(o.filter=e,s.push(o)):(o.tuple=e,n.push(o)),this},encode:function(e,t){return Object(i.isFunction)(e)?s.push({filter:e,field:t}):n.push({tuple:e,field:t}),this},reflow:function(){return u=!0,this},pulse:function(i,a){var c,l,d,p,f,h,m;for(l=0,d=e.length;l<d;++l)i.add.push(Object(r.ingest)(e[l]));for(c={},l=0,d=t.length;l<d;++l)h=t[l],c[Object(r.tupleid)(h)]=h;for(l=0,d=o.length;l<d;++l)f=o[l],a.forEach(function(e){f(e)&&(c[Object(r.tupleid)(e)]=e)});for(m in c)i.rem.push(c[m]);function g(e,t,n){n?e[t]=n(e):i.encode=t,u||(c[Object(r.tupleid)(e)]=e)}for(c={},l=0,d=n.length;l<d;++l)g((p=n[l]).tuple,p.field,p.value),i.modifies(p.field);for(l=0,d=s.length;l<d;++l)p=s[l],f=p.filter,a.forEach(function(e){f(e)&&g(e,p.field,p.value)}),i.modifies(p.field);if(u)i.mod=t.length||o.length?a.filter(function(e){return c.hasOwnProperty(Object(r.tupleid)(e))}):a.slice();else for(m in c)i.mod.push(c[m]);return i}}}},"./node_modules/vega-dataflow/src/EventStream.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a}),n.d(t,"stream",function(){return s});var r=n("./node_modules/vega-dataflow/src/util/UniqueList.js"),i=n("./node_modules/vega-util/index.js"),o=0;function a(e,t,n){this.id=++o,this.value=null,n&&(this.receive=n),e&&(this._filter=e),t&&(this._apply=t)}function s(e,t,n){return new a(e,t,n)}var u=a.prototype;u._filter=i.truthy,u._apply=i.identity,u.targets=function(){return this._targets||(this._targets=Object(r.default)(i.id))},u.consume=function(e){return arguments.length?(this._consume=!!e,this):!!this._consume},u.receive=function(e){if(this._filter(e)){for(var t=this.value=this._apply(e),n=this._targets,r=n?n.length:0,i=0;i<r;++i)n[i].receive(t);this._consume&&(e.preventDefault(),e.stopPropagation())}},u.filter=function(e){var t=s(e);return this.targets().add(t),t},u.apply=function(e){var t=s(null,e);return this.targets().add(t),t},u.merge=function(){var e=s();this.targets().add(e);for(var t=0,n=arguments.length;t<n;++t)arguments[t].targets().add(e);return e},u.throttle=function(e){var t=-1;return this.filter(function(){var n=Date.now();return n-t>e?(t=n,1):0})},u.debounce=function(e){var t=s();return this.targets().add(s(null,null,Object(i.debounce)(e,function(e){var n=e.dataflow;t.receive(e),n&&n.run&&n.run()}))),t},u.between=function(e,t){var n=!1;return e.targets().add(s(null,null,function(){n=!0})),t.targets().add(s(null,null,function(){n=!1})),this.filter(function(){return n})}},"./node_modules/vega-dataflow/src/MultiPulse.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/src/Pulse.js"),i=n("./node_modules/vega-util/index.js");function o(e,t,n,r){var i,o,a,s,u,c=0;for(this.dataflow=e,this.stamp=t,this.fields=null,this.encode=r||null,this.pulses=n,a=0,s=n.length;a<s;++a)if((i=n[a]).stamp===t){if(i.fields){o=this.fields||(this.fields={});for(u in i.fields)o[u]=1}i.changed(this.ADD)&&(c|=this.ADD),i.changed(this.REM)&&(c|=this.REM),i.changed(this.MOD)&&(c|=this.MOD)}this.changes=c}var a=Object(i.inherits)(o,r.default);a.fork=function(e){var t=new r.default(this.dataflow).init(this,e&this.NO_FIELDS);return void 0!==e&&(e&t.ADD&&this.visit(t.ADD,function(e){return t.add.push(e)}),e&t.REM&&this.visit(t.REM,function(e){return t.rem.push(e)}),e&t.MOD&&this.visit(t.MOD,function(e){return t.mod.push(e)})),t},a.changed=function(e){return this.changes&e},a.modified=function(e){var t=this.fields;return t&&this.changes&this.MOD?Object(i.isArray)(e)?e.some(function(e){return t[e]}):t[e]:0},a.filter=function(){Object(i.error)("MultiPulse does not support filtering.")},a.materialize=function(){Object(i.error)("MultiPulse does not support materialization.")},a.visit=function(e,t){var n=this.pulses,r=n.length,i=0;if(e&this.SOURCE)for(;i<r;++i)n[i].visit(e,t);else for(;i<r;++i)n[i].stamp===this.stamp&&n[i].visit(e,t);return this}},"./node_modules/vega-dataflow/src/Operator.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return u});var r=n("./node_modules/vega-dataflow/src/Parameters.js"),i=n("./node_modules/vega-dataflow/src/util/UniqueList.js"),o=n("./node_modules/vega-util/index.js"),a=0,s=new r.default;function u(e,t,n,r){this.id=++a,this.value=e,this.stamp=-1,this.rank=-1,this.qrank=-1,this.flags=0,t&&(this._update=t),n&&this.parameters(n,r)}var c=u.prototype;function l(e){return function(t){var n=this.flags;return 0===arguments.length?!!(n&e):(this.flags=t?n|e:n&~e,this)}}c.targets=function(){return this._targets||(this._targets=Object(i.default)(o.id))},c.set=function(e){return this.value!==e?(this.value=e,1):0},c.skip=l(1),c.modified=l(2),c.parameters=function(e,t){t=!1!==t;var n,i,a,s,c=this,l=c._argval=c._argval||new r.default,d=c._argops=c._argops||[],p=[];function f(e,n,r){r instanceof u?(r!==c&&(t&&r.targets().add(c),p.push(r)),d.push({op:r,name:e,index:n})):l.set(e,n,r)}for(n in e)if(i=e[n],"pulse"===n)Object(o.array)(i).forEach(function(e){e instanceof u?e!==c&&(e.targets().add(c),p.push(e)):Object(o.error)("Pulse parameters must be operator instances.")}),c.source=i;else if(Object(o.isArray)(i))for(l.set(n,-1,Array(a=i.length)),s=0;s<a;++s)f(n,s,i[s]);else f(n,-1,i);return this.marshall().clear(),p},c.marshall=function(e){var t,n,r,i,o,a=this._argval||s,u=this._argops;if(u&&(r=u.length))for(n=0;n<r;++n)o=(i=(t=u[n]).op).modified()&&i.stamp===e,a.set(t.name,t.index,i.value,o);return a},c.evaluate=function(e){if(this._update){var t=this.marshall(e.stamp),n=this._update(t,e);if(t.clear(),n!==this.value)this.value=n;else if(!this.modified())return e.StopPropagation}},c.run=function(e){return e.stamp<=this.stamp?e.StopPropagation:(this.skip()?(this.skip(!1),t=0):t=this.evaluate(e),this.stamp=e.stamp,this.pulse=t,t||e);var t}},"./node_modules/vega-dataflow/src/Parameters.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-util/index.js"),i="_:mod:_";function o(){Object.defineProperty(this,i,{writable:!0,value:{}})}var a=o.prototype;a.set=function(e,t,n,o){var a=this[e],s=this[i];return null!=t&&t>=0?(a[t]!==n||o)&&(a[t]=n,s[t+":"+e]=-1,s[e]=-1):(a!==n||o)&&(this[e]=n,s[e]=Object(r.isArray)(n)?1+n.length:-1),this},a.modified=function(e,t){var n,o=this[i];if(!arguments.length){for(n in o)if(o[n])return!0;return!1}if(Object(r.isArray)(e)){for(n=0;n<e.length;++n)if(o[e[n]])return!0;return!1}return null!=t&&t>=0?t+1<o[e]||!!o[t+":"+e]:!!o[e]},a.clear=function(){return this[i]={},this}},"./node_modules/vega-dataflow/src/Pulse.js":function(e,t,n){"use strict";n.r(t),n.d(t,"StopPropagation",function(){return o}),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-dataflow/src/Tuple.js"),i=n("./node_modules/vega-util/index.js"),o={};function a(e,t,n){this.dataflow=e,this.stamp=null==t?-1:t,this.add=[],this.rem=[],this.mod=[],this.fields=null,this.encode=n||null}var s=a.prototype;function u(e,t){return e?function(n,r){return e(n,r)&&t(n,r)}:t}function c(e,t){var n=[];return Object(i.visitArray)(e,t,function(e){n.push(e)}),n}function l(e,t){var n={};return e.visit(t,function(e){n[Object(r.tupleid)(e)]=1}),function(e){return n[Object(r.tupleid)(e)]?null:e}}s.StopPropagation=o,s.ADD=1,s.REM=2,s.MOD=4,s.ADD_REM=3,s.ADD_MOD=5,s.ALL=7,s.REFLOW=8,s.SOURCE=16,s.NO_SOURCE=32,s.NO_FIELDS=64,s.fork=function(e){return new a(this.dataflow).init(this,e)},s.addAll=function(){var e=this;return this.source&&this.source.length!==this.add.length?((e=new a(this.dataflow).init(this)).add=e.source,e):e},s.init=function(e,t){return this.stamp=e.stamp,this.encode=e.encode,!e.fields||64&t||(this.fields=e.fields),1&t?(this.addF=e.addF,this.add=e.add):(this.addF=null,this.add=[]),2&t?(this.remF=e.remF,this.rem=e.rem):(this.remF=null,this.rem=[]),4&t?(this.modF=e.modF,this.mod=e.mod):(this.modF=null,this.mod=[]),32&t?(this.srcF=null,this.source=null):(this.srcF=e.srcF,this.source=e.source),this},s.runAfter=function(e){this.dataflow.runAfter(e)},s.changed=function(e){var t=e||7;return 1&t&&this.add.length||2&t&&this.rem.length||4&t&&this.mod.length},s.reflow=function(e){if(e)return this.fork(7).reflow();var t=this.add.length,n=this.source&&this.source.length;return n&&n!==t&&(this.mod=this.source,t&&this.filter(4,l(this,1))),this},s.modifies=function(e){var t=Object(i.array)(e),n=this.fields||(this.fields={});return t.forEach(function(e){n[e]=!0}),this},s.modified=function(e){var t=this.fields;return!(!this.mod.length||!t)&&(arguments.length?Object(i.isArray)(e)?e.some(function(e){return t[e]}):t[e]:!!t)},s.filter=function(e,t){return 1&e&&(this.addF=u(this.addF,t)),2&e&&(this.remF=u(this.remF,t)),4&e&&(this.modF=u(this.modF,t)),16&e&&(this.srcF=u(this.srcF,t)),this},s.materialize=function(e){return 1&(e=e||7)&&this.addF&&(this.add=c(this.add,this.addF),this.addF=null),2&e&&this.remF&&(this.rem=c(this.rem,this.remF),this.remF=null),4&e&&this.modF&&(this.mod=c(this.mod,this.modF),this.modF=null),16&e&&this.srcF&&(this.source=this.source.filter(this.srcF),this.srcF=null),this},s.visit=function(e,t){var n,r,o=t;return 16&e?(Object(i.visitArray)(this.source,this.srcF,o),this):(1&e&&Object(i.visitArray)(this.add,this.addF,o),2&e&&Object(i.visitArray)(this.rem,this.remF,o),4&e&&Object(i.visitArray)(this.mod,this.modF,o),8&e&&(n=this.source)&&((r=this.add.length+this.mod.length)===n.length||(r?Object(i.visitArray)(n,l(this,5),o):Object(i.visitArray)(n,this.srcF,o))),this)}},"./node_modules/vega-dataflow/src/Transform.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/src/Operator.js"),i=n("./node_modules/vega-util/index.js");function o(e,t){r.default.call(this,e,null,t)}var a=Object(i.inherits)(o,r.default);a.run=function(e){return e.stamp<=this.stamp?e.StopPropagation:(this.skip()?this.skip(!1):t=this.evaluate(e),(t=t||e)!==e.StopPropagation&&(this.pulse=t),this.stamp=e.stamp,t);var t},a.evaluate=function(e){var t=this.marshall(e.stamp),n=this.transform(t,e);return t.clear(),n},a.transform=function(){}},"./node_modules/vega-dataflow/src/Tuple.js":function(e,t,n){"use strict";n.r(t),n.d(t,"reset",function(){return o}),n.d(t,"isTuple",function(){return a}),n.d(t,"tupleid",function(){return s}),n.d(t,"ingest",function(){return c}),n.d(t,"derive",function(){return l}),n.d(t,"rederive",function(){return d}),n.d(t,"replace",function(){return p});var r=Symbol("vega_id"),i=1;function o(){i=1}function a(e){return!(!e||!s(e))}function s(e){return e[r]}function u(e,t){return e[r]=t,e}function c(e){var t=e===Object(e)?e:{data:e};return s(t)?t:u(t,i++)}function l(e){return d(e,c({}))}function d(e,t){for(var n in e)t[n]=e[n];return t}function p(e,t){return u(t,s(e))}},"./node_modules/vega-dataflow/src/dataflow/Dataflow.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return g});var r=n("./node_modules/vega-dataflow/src/dataflow/add.js"),i=n("./node_modules/vega-dataflow/src/dataflow/connect.js"),o=n("./node_modules/vega-dataflow/src/dataflow/events.js"),a=n("./node_modules/vega-dataflow/src/dataflow/load.js"),s=n("./node_modules/vega-dataflow/src/dataflow/on.js"),u=n("./node_modules/vega-dataflow/src/dataflow/rank.js"),c=n("./node_modules/vega-dataflow/src/dataflow/run.js"),l=n("./node_modules/vega-dataflow/src/dataflow/update.js"),d=n("./node_modules/vega-dataflow/src/ChangeSet.js"),p=n("./node_modules/vega-dataflow/src/util/Heap.js"),f=n("./node_modules/vega-dataflow/src/util/UniqueList.js"),h=n("./node_modules/vega-loader/index.js"),m=n("./node_modules/vega-util/index.js");function g(){this._log=Object(m.logger)(),this.logLevel(m.Error),this._clock=0,this._rank=0;try{this._loader=Object(h.loader)()}catch(e){}this._touched=Object(f.default)(m.id),this._pulses={},this._pulse=null,this._heap=new p.default(function(e,t){return e.qrank-t.qrank}),this._postrun=[]}var v=g.prototype;function _(e){return function(){return this._log[e].apply(this,arguments)}}v.stamp=function(){return this._clock},v.loader=function(e){return arguments.length?(this._loader=e,this):this._loader},v.cleanThreshold=1e4,v.add=r.default,v.connect=i.default,v.rank=u.rank,v.rerank=u.rerank,v.pulse=l.pulse,v.touch=l.touch,v.update=l.update,v.changeset=d.default,v.ingest=a.ingest,v.request=a.request,v.events=o.default,v.on=s.default,v.run=c.run,v.runAsync=c.runAsync,v.runAfter=c.runAfter,v._enqueue=c.enqueue,v._getPulse=c.getPulse,v.error=_("error"),v.warn=_("warn"),v.info=_("info"),v.debug=_("debug"),v.logLevel=_("level")},"./node_modules/vega-dataflow/src/dataflow/add.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-dataflow/src/Operator.js"),i=n("./node_modules/vega-util/index.js");t.default=function(e,t,n,o){var a,s=1;return e instanceof r.default?a=e:e&&e.prototype instanceof r.default?a=new e:Object(i.isFunction)(e)?a=new r.default(null,e):(s=0,a=new r.default(e,t)),this.rank(a),s&&(o=n,n=t),n&&this.connect(a,a.parameters(n,o)),this.touch(a),a}},"./node_modules/vega-dataflow/src/dataflow/connect.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){var n,r,i=e.rank;for(n=0,r=t.length;n<r;++n)if(i<t[n].rank)return void this.rerank(e)}},"./node_modules/vega-dataflow/src/dataflow/events.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-dataflow/src/EventStream.js"),i=n("./node_modules/vega-util/index.js");t.default=function(e,t,n,o){for(var a,s=this,u=Object(r.stream)(n,o),c=function(e){e.dataflow=s;try{u.receive(e)}catch(e){s.error(e)}finally{s.run()}},l=0,d=(a="string"==typeof e&&"undefined"!=typeof document?document.querySelectorAll(e):Object(i.array)(e)).length;l<d;++l)a[l].addEventListener(t,c);return u}},"./node_modules/vega-dataflow/src/dataflow/load.js":function(e,t,n){"use strict";n.r(t),n.d(t,"ingest",function(){return i}),n.d(t,"request",function(){return o});var r=n("./node_modules/vega-loader/index.js");function i(e,t,n){return this.pulse(e,this.changeset().insert(Object(r.read)(t,n)))}function o(e,t,n){var r,i,o,a,s=this,u=s._pending||(r=s,(a=new Promise(function(e,t){i=e,o=t})).requests=0,a.done=function(){0==--a.requests&&r.runAfter(function(){r._pending=null;try{r.run(),i(r)}catch(e){o(e)}})},r._pending=a);u.requests+=1,s.loader().load(t,{context:"dataflow"}).then(function(t){s.ingest(e,t,n)},function(e){s.error("Loading failed",t,e)}).catch(function(e){s.error("Data ingestion failed",t,e)}).then(u.done,u.done)}},"./node_modules/vega-dataflow/src/dataflow/on.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-dataflow/src/Operator.js"),i=n("./node_modules/vega-dataflow/src/ChangeSet.js"),o=n("./node_modules/vega-util/index.js"),a={skip:!0};function s(e,t,n,s,u,c){var l,d,p=Object(o.extend)({},c,a);Object(o.isFunction)(n)||(n=Object(o.constant)(n)),void 0===s?l=function(t){e.touch(n(t))}:Object(o.isFunction)(s)?(d=new r.default(null,s,u,!1),l=function(t){var r,o=n(t);d.evaluate(t),Object(i.isChangeSet)(r=d.value)?e.pulse(o,r,c):e.update(o,r,p)}):l=function(t){e.update(n(t),s,p)},t.apply(l)}function u(e,t,n,i,a,s){var u,c;void 0===i?c=n:(u=Object(o.isFunction)(i)?i:Object(o.constant)(i),i=n?function(e,t){var r=u(e,t);return n.skip()?r:n.skip(!0).value=r}:u,(c=new r.default(null,i,a,!1)).modified(s&&s.force),c.rank=0,n&&(c.skip(!0),c.value=n.value,c.targets().add(n))),t.targets().add(c)}t.default=function(e,t,n,i,o){return(e instanceof r.default?u:s)(this,e,t,n,i,o),this}},"./node_modules/vega-dataflow/src/dataflow/rank.js":function(e,t,n){"use strict";n.r(t),n.d(t,"rank",function(){return i}),n.d(t,"rerank",function(){return o});var r=n("./node_modules/vega-util/index.js");function i(e){e.rank=++this._rank}function o(e){for(var t,n,i,o=[e];o.length;)if(this.rank(t=o.pop()),n=t._targets)for(i=n.length;--i>=0;)o.push(t=n[i]),t===e&&Object(r.error)("Cycle detected in dataflow graph.")}},"./node_modules/vega-dataflow/src/dataflow/run.js":function(e,t,n){"use strict";n.r(t),n.d(t,"run",function(){return s}),n.d(t,"runAsync",function(){return c}),n.d(t,"runAfter",function(){return l}),n.d(t,"enqueue",function(){return d}),n.d(t,"getPulse",function(){return p});var r=n("./node_modules/vega-dataflow/src/Pulse.js"),i=n("./node_modules/vega-dataflow/src/MultiPulse.js"),o=n("./node_modules/vega-dataflow/src/util/UniqueList.js"),a=n("./node_modules/vega-util/index.js");function s(e){var t,n,i,s,c=this,l=0,d=c.logLevel();if(c._pending)return c.info("Awaiting requests, delaying dataflow run."),0;if(c._pulse)return c.error("Dataflow invoked recursively. Use the runAfter method to queue invocation."),0;if(!c._touched.length)return c.info("Dataflow invoked, but nothing to do."),0;c._pulse=new r.default(c,++c._clock,e),d>=a.Info&&(i=Date.now(),c.debug("-- START PROPAGATION ("+c._clock+") -----")),c._touched.forEach(function(e){c._enqueue(e,!0)}),c._touched=Object(o.default)(a.id);try{for(;c._heap.size()>0;)(t=c._heap.pop()).rank===t.qrank?(n=t.run(c._getPulse(t,e)),d>=a.Debug&&c.debug(t.id,n===r.StopPropagation?"STOP":n,t),n!==r.StopPropagation&&(c._pulse=n,t._targets&&t._targets.forEach(function(e){c._enqueue(e)})),++l):c._enqueue(t,!0)}catch(e){s=e}if(c._pulses={},c._pulse=null,d>=a.Info&&(i=Date.now()-i,c.info("> Pulse "+c._clock+": "+l+" operators; "+i+"ms")),s&&(c._postrun=[],c.error(s)),c._onrun)try{c._onrun(c,l,s)}catch(e){c.error(e)}if(c._postrun.length){var p=c._postrun;c._postrun=[],p.sort(function(e,t){return t.priority-e.priority}).forEach(function(e){u(c,e.callback)})}return l}function u(e,t){try{t(e)}catch(t){e.error(t)}}function c(){return this._pending||Promise.resolve(this.run())}function l(e,t,n){this._pulse||t?this._postrun.push({priority:n||0,callback:e}):u(this,e)}function d(e,t){var n=!this._pulses[e.id];n&&(this._pulses[e.id]=this._pulse),(n||t)&&(e.qrank=e.rank,this._heap.push(e))}function p(e,t){var n,o=e.source,s=this._clock;return o&&Object(a.isArray)(o)?(n=o.map(function(e){return e.pulse}),new i.default(this,s,n,t)):(n=this._pulses[e.id],o&&((o=o.pulse)&&o!==r.StopPropagation?o.stamp===s&&n.target!==e?n=o:n.source=o.source:n.source=[]),n)}},"./node_modules/vega-dataflow/src/dataflow/update.js":function(e,t,n){"use strict";n.r(t),n.d(t,"touch",function(){return o}),n.d(t,"update",function(){return a}),n.d(t,"pulse",function(){return s});var r=n("./node_modules/vega-dataflow/src/Pulse.js"),i={skip:!1,force:!1};function o(e,t){var n=t||i;return this._pulse?this._enqueue(e):this._touched.add(e),n.skip&&e.skip(!0),this}function a(e,t,n){var r=n||i;return(e.set(t)||r.force)&&this.touch(e,r),this}function s(e,t,n){this.touch(e,n||i);var o=new r.default(this,this._clock+(this._pulse?0:1)),a=e.pulse&&e.pulse.source||[];return o.target=e,this._pulses[e.id]=t.pulse(o,a),this}},"./node_modules/vega-dataflow/src/register.js":function(e,t,n){"use strict";n.r(t),n.d(t,"transforms",function(){return r}),n.d(t,"definition",function(){return i}),n.d(t,"transform",function(){return o});var r={};function i(e){var t=o(e);return t&&t.Definition||null}function o(e){return e=e&&e.toLowerCase(),r.hasOwnProperty(e)?r[e]:null}},"./node_modules/vega-dataflow/src/util/Heap.js":function(e,t,n){"use strict";function r(e){this.cmp=e,this.nodes=[]}n.r(t),n.d(t,"default",function(){return r});var i=r.prototype;function o(e,t,n,r){var i,o,a;for(i=e[n];n>t&&r(i,o=e[a=n-1>>1])<0;)e[n]=o,n=a;return e[n]=i}function a(e,t,n){for(var r,i=t,a=e.length,s=e[t],u=2*t+1;u<a;)(r=u+1)<a&&n(e[u],e[r])>=0&&(u=r),e[t]=e[u],u=2*(t=u)+1;return e[t]=s,o(e,i,t,n)}i.size=function(){return this.nodes.length},i.clear=function(){return this.nodes=[],this},i.peek=function(){return this.nodes[0]},i.push=function(e){var t=this.nodes;return t.push(e),o(t,0,t.length-1,this.cmp)},i.pop=function(){var e,t=this.nodes,n=t.pop();return t.length?(e=t[0],t[0]=n,a(t,0,this.cmp)):e=n,e},i.replace=function(e){var t=this.nodes,n=t[0];return t[0]=e,a(t,0,this.cmp),n},i.pushpop=function(e){var t=this.nodes,n=t[0];return t.length&&this.cmp(n,e)<0&&(t[0]=e,e=n,a(t,0,this.cmp)),e}},"./node_modules/vega-dataflow/src/util/UniqueList.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return i});var r=n("./node_modules/vega-util/index.js");function i(e){var t=e||r.identity,n=[],i={};return n.add=function(e){var r=t(e);return i[r]||(i[r]=1,n.push(e)),n},n.remove=function(e){var r,o=t(e);return i[o]&&(i[o]=0,(r=n.indexOf(e))>=0&&n.splice(r,1)),n},n}},"./node_modules/vega-encode/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-encode/src/AxisTicks.js");n.d(t,"axisticks",function(){return r.default});var i=n("./node_modules/vega-encode/src/DataJoin.js");n.d(t,"datajoin",function(){return i.default});var o=n("./node_modules/vega-encode/src/Encode.js");n.d(t,"encode",function(){return o.default});var a=n("./node_modules/vega-encode/src/LegendEntries.js");n.d(t,"legendentries",function(){return a.default});var s=n("./node_modules/vega-encode/src/LinkPath.js");n.d(t,"linkpath",function(){return s.default});var u=n("./node_modules/vega-encode/src/Pie.js");n.d(t,"pie",function(){return u.default});var c=n("./node_modules/vega-encode/src/Scale.js");n.d(t,"scale",function(){return c.default});var l=n("./node_modules/vega-encode/src/SortItems.js");n.d(t,"sortitems",function(){return l.default});var d=n("./node_modules/vega-encode/src/Stack.js");n.d(t,"stack",function(){return d.default});var p=n("./node_modules/vega-encode/src/ticks.js");n.d(t,"validTicks",function(){return p.validTicks})},"./node_modules/vega-encode/src/AxisTicks.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-encode/src/ticks.js"),i=n("./node_modules/vega-dataflow/index.js"),o=n("./node_modules/vega-util/index.js");function a(e){i.Transform.call(this,null,e)}Object(o.inherits)(a,i.Transform).transform=function(e,t){if(this.value&&!e.modified())return t.StopPropagation;var n=t.fork(t.NO_SOURCE|t.NO_FIELDS),o=this.value,a=e.scale,s=null==e.count?e.values?e.values.length:10:Object(r.tickCount)(a,e.count),u=e.format||Object(r.tickFormat)(a,s,e.formatSpecifier),c=e.values?Object(r.validTicks)(a,e.values,s):Object(r.tickValues)(a,s);return o&&(n.rem=o),o=c.map(function(e,t){return Object(i.ingest)({index:t/(c.length-1),value:e,label:u(e)})}),e.extra&&o.push(Object(i.ingest)({index:-1,extra:{value:o[0].value},label:""})),n.source=o,n.add=o,this.value=o,n}},"./node_modules/vega-encode/src/DataJoin.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,null,e)}function a(){return Object(r.ingest)({})}function s(e){return e.exit}Object(i.inherits)(o,r.Transform).transform=function(e,t){var n=t.dataflow,o=t.fork(t.NO_SOURCE|t.NO_FIELDS),u=e.item||a,c=e.key||r.tupleid,l=this.value;return Object(i.isArray)(o.encode)&&(o.encode=null),l&&(e.modified("key")||t.modified(c))&&Object(i.error)("DataJoin does not support modified key function or fields."),l||(t=t.addAll(),this.value=l=Object(i.fastmap)().test(s),l.lookup=function(e){return l.get(c(e))}),t.visit(t.ADD,function(e){var t=c(e),n=l.get(t);n?n.exit?(l.empty--,o.add.push(n)):o.mod.push(n):(l.set(t,n=u(e)),o.add.push(n)),n.datum=e,n.exit=!1}),t.visit(t.MOD,function(e){var t=c(e),n=l.get(t);n&&(n.datum=e,o.mod.push(n))}),t.visit(t.REM,function(e){var t=c(e),n=l.get(t);e!==n.datum||n.exit||(o.rem.push(n),n.exit=!0,++l.empty)}),t.changed(t.ADD_MOD)&&o.modifies("datum"),e.clean&&l.empty>n.cleanThreshold&&n.runAfter(l.clean),o}},"./node_modules/vega-encode/src/Encode.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,null,e)}Object(i.inherits)(o,r.Transform).transform=function(e,t){var n=t.fork(t.ADD_REM),r=e.encoders,o=t.encode;if(Object(i.isArray)(o)){if(!n.changed()&&!o.every(function(e){return r[e]}))return t.StopPropagation;o=o[0]}var a="enter"===o,s=r.update||i.falsy,u=r.enter||i.falsy,c=r.exit||i.falsy,l=(o&&!a?r[o]:s)||i.falsy;if(t.changed(t.ADD)&&(t.visit(t.ADD,function(t){u(t,e),s(t,e),l!==i.falsy&&l!==s&&l(t,e)}),n.modifies(u.output),n.modifies(s.output),l!==i.falsy&&l!==s&&n.modifies(l.output)),t.changed(t.REM)&&c!==i.falsy&&(t.visit(t.REM,function(t){c(t,e)}),n.modifies(c.output)),a||l!==i.falsy){var d=t.MOD|(e.modified()?t.REFLOW:0);a?(t.visit(d,function(t){var r=u(t,e);(l(t,e)||r)&&n.mod.push(t)}),n.mod.length&&n.modifies(u.output)):t.visit(d,function(t){l(t,e)&&n.mod.push(t)}),n.mod.length&&n.modifies(l.output)}return n.changed()?n:t.StopPropagation}},"./node_modules/vega-encode/src/LegendEntries.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return u});var r=n("./node_modules/vega-encode/src/labels.js"),i=n("./node_modules/vega-encode/src/ticks.js"),o=n("./node_modules/vega-dataflow/index.js"),a=n("./node_modules/vega-scale/index.js"),s=n("./node_modules/vega-util/index.js");function u(e){o.Transform.call(this,[],e)}Object(s.inherits)(u,o.Transform).transform=function(e,t){if(null!=this.value&&!e.modified())return t.StopPropagation;var n=t.fork(t.NO_SOURCE|t.NO_FIELDS),u=0,c=this.value,l="gradient"===e.type,d=e.scale,p=null==e.count?5:Object(i.tickCount)(d,e.count),f=e.format||Object(i.tickFormat)(d,p,e.formatSpecifier),h=e.values||Object(r.labelValues)(d,p,l);if(f=Object(r.labelFormat)(d,f),c&&(n.rem=c),l)var m=e.values?d.domain():h,g=Object(a.scaleFraction)(d,m[0],Object(s.peek)(m));else{var v,_=e.size;Object(s.isFunction)(_)?(e.values||0!==d(h[0])||(h=h.slice(1)),v=h.reduce(function(t,n){return Math.max(t,_(n,e))},0)):_=Object(s.constant)(v=_||8)}return c=h.map(function(t,n){var r=Object(o.ingest)({index:n,label:f(t,n,h),value:t});return l?r.perc=g(t):(r.offset=v,r.size=_(t,e),r.total=Math.round(u),u+=r.size),r}),n.source=c,n.add=c,this.value=c,n}},"./node_modules/vega-encode/src/LinkPath.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return l});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js"),o=Object(i.fastmap)({line:d,"line-radial":function(e,t,n,r){return d(t*Math.cos(e),t*Math.sin(e),r*Math.cos(n),r*Math.sin(n))},arc:p,"arc-radial":function(e,t,n,r){return p(t*Math.cos(e),t*Math.sin(e),r*Math.cos(n),r*Math.sin(n))},curve:f,"curve-radial":function(e,t,n,r){return f(t*Math.cos(e),t*Math.sin(e),r*Math.cos(n),r*Math.sin(n))},"orthogonal-horizontal":function(e,t,n,r){return"M"+e+","+t+"V"+r+"H"+n},"orthogonal-vertical":function(e,t,n,r){return"M"+e+","+t+"H"+n+"V"+r},"orthogonal-radial":function(e,t,n,r){var i=Math.cos(e),o=Math.sin(e),a=Math.cos(n),s=Math.sin(n),u=Math.abs(n-e)>Math.PI?n<=e:n>e;return"M"+t*i+","+t*o+"A"+t+","+t+" 0 0,"+(u?1:0)+" "+t*a+","+t*s+"L"+r*a+","+r*s},"diagonal-horizontal":function(e,t,n,r){var i=(e+n)/2;return"M"+e+","+t+"C"+i+","+t+" "+i+","+r+" "+n+","+r},"diagonal-vertical":function(e,t,n,r){var i=(t+r)/2;return"M"+e+","+t+"C"+e+","+i+" "+n+","+i+" "+n+","+r},"diagonal-radial":function(e,t,n,r){var i=Math.cos(e),o=Math.sin(e),a=Math.cos(n),s=Math.sin(n),u=(t+r)/2;return"M"+t*i+","+t*o+"C"+u*i+","+u*o+" "+u*a+","+u*s+" "+r*a+","+r*s}});function a(e){return e.source.x}function s(e){return e.source.y}function u(e){return e.target.x}function c(e){return e.target.y}function l(e){r.Transform.call(this,{},e)}function d(e,t,n,r){return"M"+e+","+t+"L"+n+","+r}function p(e,t,n,r){var i=n-e,o=r-t,a=Math.sqrt(i*i+o*o)/2;return"M"+e+","+t+"A"+a+","+a+" "+180*Math.atan2(o,i)/Math.PI+" 0 1 "+n+","+r}function f(e,t,n,r){var i=n-e,o=r-t,a=.2*(i+o),s=.2*(o-i);return"M"+e+","+t+"C"+(e+a)+","+(t+s)+" "+(n+s)+","+(r-a)+" "+n+","+r}l.Definition={type:"LinkPath",metadata:{modifies:!0},params:[{name:"sourceX",type:"field",default:"source.x"},{name:"sourceY",type:"field",default:"source.y"},{name:"targetX",type:"field",default:"target.x"},{name:"targetY",type:"field",default:"target.y"},{name:"orient",type:"enum",default:"vertical",values:["horizontal","vertical","radial"]},{name:"shape",type:"enum",default:"line",values:["line","arc","curve","diagonal","orthogonal"]},{name:"as",type:"string",default:"path"}]},Object(i.inherits)(l,r.Transform).transform=function(e,t){var n=e.sourceX||a,r=e.sourceY||s,l=e.targetX||u,d=e.targetY||c,p=e.as||"path",f=e.orient||"vertical",h=e.shape||"line",m=o.get(h+"-"+f)||o.get(h);return m||Object(i.error)("LinkPath unsupported type: "+e.shape+(e.orient?"-"+e.orient:"")),t.visit(t.SOURCE,function(e){e[p]=m(n(e),r(e),l(e),d(e))}),t.reflow(e.modified()).modifies(p)}},"./node_modules/vega-encode/src/Pie.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js"),o=n("./node_modules/d3-array/index.js");function a(e){r.Transform.call(this,null,e)}a.Definition={type:"Pie",metadata:{modifies:!0},params:[{name:"field",type:"field"},{name:"startAngle",type:"number",default:0},{name:"endAngle",type:"number",default:6.283185307179586},{name:"sort",type:"boolean",default:!1},{name:"as",type:"string",array:!0,length:2,default:["startAngle","endAngle"]}]},Object(i.inherits)(a,r.Transform).transform=function(e,t){var n,r,a,s=e.as||["startAngle","endAngle"],u=s[0],c=s[1],l=e.field||i.one,d=e.startAngle||0,p=null!=e.endAngle?e.endAngle:2*Math.PI,f=t.source,h=f.map(l),m=h.length,g=d,v=(p-d)/Object(o.sum)(h),_=Object(o.range)(m);for(e.sort&&_.sort(function(e,t){return h[e]-h[t]}),n=0;n<m;++n)a=h[_[n]],(r=f[_[n]])[u]=g,r[c]=g+=a*v;return this.value=h,t.reflow(e.modified()).modifies(s)}},"./node_modules/vega-encode/src/Scale.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return f});var r=n("./node_modules/vega-encode/src/ticks.js"),i=n("./node_modules/vega-dataflow/index.js"),o=n("./node_modules/vega-util/index.js"),a=n("./node_modules/vega-encode/src/scale-types.js"),s=n("./node_modules/vega-scale/index.js"),u=n("./node_modules/d3-interpolate/index.js"),c=5,l=Object(o.toSet)([a.Linear,a.Pow,a.Sqrt]),d=Object(o.toSet)([a.Linear,a.Log,a.Pow,a.Sqrt,a.Time,a.Utc]),p=Object(o.toSet)(["set","modified","clear","type","scheme","schemeExtent","schemeCount","domain","domainMin","domainMid","domainMax","domainRaw","nice","zero","range","rangeStep","round","reverse","interpolate","interpolateGamma"]);function f(e){i.Transform.call(this,null,e),this.modified(!0)}function h(e,t,n){return Object(o.isFunction)(e)&&(t||n)?Object(s.interpolateRange)(e,m(t||[0,1],n)):e}function m(e,t){return t?e.slice().reverse():e}Object(o.inherits)(f,i.Transform).transform=function(e,t){var n,i=t.dataflow,f=this.value;f&&!e.modified("type")||(this.value=f=Object(s.scale)((e.type||a.Linear).toLowerCase())());for(n in e)if(!p[n]){if("padding"===n&&d[f.type])continue;Object(o.isFunction)(f[n])?f[n](e[n]):i.warn("Unsupported scale property: "+n)}return function(e,t,n){var r=t.round||!1,i=t.range;if(null!=t.rangeStep)i=function(e,t,n){e!==a.Band&&e!==a.Point&&Object(o.error)("Only band and point scales support rangeStep.");var r=(null!=t.paddingOuter?t.paddingOuter:t.padding)||0,i=e===a.Point?1:(null!=t.paddingInner?t.paddingInner:t.padding)||0;return[0,t.rangeStep*Object(s.bandSpace)(n,i,r)]}(e.type,t,n);else if(t.scheme){if(i=function(e,t,n){var r,i=t.scheme.toLowerCase(),u=Object(s.scheme)(i),l=t.schemeExtent;u||Object(o.error)("Unrecognized scheme name: "+t.scheme);return n=e===a.Threshold?n+1:e===a.BinOrdinal?n-1:e===a.Quantile||e===a.Quantize?+t.schemeCount||c:n,e===a.Sequential?h(u,l,t.reverse):!l&&(r=Object(s.scheme)(i+"-"+n))?r:Object(o.isFunction)(u)?function(e,t){for(var n=new Array(t),r=t-1||1,i=0;i<t;++i)n[i]=e(i/r);return n}(h(u,l),n):e===a.Ordinal?u:u.slice(0,n)}(e.type,t,n),Object(o.isFunction)(i))return e.interpolator(i)}else if(i&&e.type===a.Sequential)return e.interpolator(Object(u.interpolateRgbBasis)(m(i,t.reverse)));i&&t.interpolate&&e.interpolate?e.interpolate(Object(s.interpolate)(t.interpolate,t.interpolateGamma)):Object(o.isFunction)(e.round)?e.round(r):Object(o.isFunction)(e.rangeRound)&&e.interpolate(r?u.interpolateRound:u.interpolate);i&&e.range(m(i,t.reverse))}(f,e,function(e,t,n){var i=(s=e,u=t.domainRaw,u?(s.domain(u),u.length):-1);var s,u;if(i>-1)return i;var c,p,f=t.domain,h=e.type,m=t.zero||void 0===t.zero&&l[h];if(!f)return 0;d[h]&&t.padding&&f[0]!==Object(o.peek)(f)&&(g=h,v=f,_=t.range,y=t.padding,b=t.exponent,j=Math.abs(Object(o.peek)(_)-_[0]),x=j/(j-2*y),O=g===a.Log?Object(o.zoomLog)(v,null,x):g===a.Sqrt?Object(o.zoomPow)(v,null,x,.5):g===a.Pow?Object(o.zoomPow)(v,null,x,b):Object(o.zoomLinear)(v,null,x),(v=v.slice())[0]=O[0],v[v.length-1]=O[1],f=v);var g,v,_,y,b,j,x,O;(m||null!=t.domainMin||null!=t.domainMax||null!=t.domainMid)&&(c=(f=f.slice()).length-1||1,m&&(f[0]>0&&(f[0]=0),f[c]<0&&(f[c]=0)),null!=t.domainMin&&(f[0]=t.domainMin),null!=t.domainMax&&(f[c]=t.domainMax),null!=t.domainMid&&(((p=t.domainMid)<f[0]||p>f[c])&&n.warn("Scale domainMid exceeds domain min or max.",p),f.splice(c,0,p)));e.domain(f),t.nice&&e.nice&&e.nice(!0!==t.nice&&Object(r.tickCount)(e,t.nice)||null);return f.length}(f,e,i)),t.fork(t.NO_SOURCE|t.NO_FIELDS)}},"./node_modules/vega-encode/src/SortItems.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,null,e)}Object(i.inherits)(o,r.Transform).transform=function(e,t){var n=e.modified("sort")||t.changed(t.ADD)||t.modified(e.sort.fields)||t.modified("datum");return n&&t.source.sort(e.sort),this.modified(n),t}},"./node_modules/vega-encode/src/Stack.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,null,e)}function a(e,t,n,r,i){for(var o,a=(t-e.sum)/2,s=e.length,u=0;u<s;++u)(o=e[u])[r]=a,o[i]=a+=Math.abs(n(o))}function s(e,t,n,r,i){for(var o,a=1/e.sum,s=0,u=e.length,c=0,l=0;c<u;++c)(o=e[c])[r]=s,o[i]=s=a*(l+=Math.abs(n(o)))}function u(e,t,n,r,i){for(var o,a,s=0,u=0,c=e.length,l=0;l<c;++l)(o=n(a=e[l]))<0?(a[r]=u,a[i]=u+=o):(a[r]=s,a[i]=s+=o)}o.Definition={type:"Stack",metadata:{modifies:!0},params:[{name:"field",type:"field"},{name:"groupby",type:"field",array:!0},{name:"sort",type:"compare"},{name:"offset",type:"enum",default:"zero",values:["zero","center","normalize"]},{name:"as",type:"string",array:!0,length:2,default:["y0","y1"]}]},Object(i.inherits)(o,r.Transform).transform=function(e,t){var n,r,o,c,l=e.as||["y0","y1"],d=l[0],p=l[1],f=e.field||i.one,h="center"===e.offset?a:"normalize"===e.offset?s:u;for(r=0,o=(n=function(e,t,n,r){var i,o,a,s,u,c,l,d,p,f=[],h=function(e){return e(u)};if(null==t)f.push(e.slice());else for(i={},o=0,a=e.length;o<a;++o)u=e[o],c=t.map(h),(l=i[c])||(i[c]=l=[],f.push(l)),l.push(u);for(c=0,p=0,s=f.length;c<s;++c){for(l=f[c],o=0,d=0,a=l.length;o<a;++o)d+=Math.abs(r(l[o]));l.sum=d,d>p&&(p=d),n&&l.sort(n)}return f.max=p,f}(t.source,e.groupby,e.sort,f)).length,c=n.max;r<o;++r)h(n[r],c,f,d,p);return t.reflow(e.modified()).modifies(l)}},"./node_modules/vega-encode/src/labels.js":function(e,t,n){"use strict";n.r(t),n.d(t,"labelValues",function(){return s}),n.d(t,"labelFormat",function(){return c});var r=n("./node_modules/vega-encode/src/scale-types.js"),i=n("./node_modules/vega-encode/src/ticks.js"),o=n("./node_modules/vega-util/index.js"),a={};function s(e,t,n){if(n)return e.domain();var r=a[e.type];return r?r(e):Object(i.tickValues)(e,t)}function u(e){var t=e.domain();return t.max=t.pop(),t}function c(e,t){return a[e.type]?(r=t,function(e,t,n){var i=n[t+1]||n.max||1/0,o=l(e,r),a=l(i,r);return o&&a?o+"–"+a:a?"< "+a:"≥ "+o}):(n=t,function(e){return n(e)});var n,r}function l(e,t){return isFinite(e)?t(e):null}a[r.Quantile]=function(e){var t=[-1/0].concat(e.quantiles());return t.max=1/0,t},a[r.Quantize]=function(e){var t=e.domain(),n=t[0],r=Object(o.peek)(t),i=e.range().length,a=new Array(i),s=0;a[0]=-1/0;for(;++s<i;)a[s]=(s*r-(s-i)*n)/i;return a.max=1/0,a},a[r.Threshold]=function(e){var t=[-1/0].concat(e.domain());return t.max=1/0,t},a[r.BinLinear]=u,a[r.BinOrdinal]=u},"./node_modules/vega-encode/src/scale-types.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Log",function(){return r}),n.d(t,"Pow",function(){return i}),n.d(t,"Utc",function(){return o}),n.d(t,"Sqrt",function(){return a}),n.d(t,"Band",function(){return s}),n.d(t,"Time",function(){return u}),n.d(t,"Point",function(){return c}),n.d(t,"Linear",function(){return l}),n.d(t,"Ordinal",function(){return d}),n.d(t,"Quantile",function(){return p}),n.d(t,"Quantize",function(){return f}),n.d(t,"Threshold",function(){return h}),n.d(t,"BinLinear",function(){return m}),n.d(t,"BinOrdinal",function(){return g}),n.d(t,"Sequential",function(){return v});var r="log",i="pow",o="utc",a="sqrt",s="band",u="time",c="point",l="linear",d="ordinal",p="quantile",f="quantize",h="threshold",m="bin-linear",g="bin-ordinal",v="sequential"},"./node_modules/vega-encode/src/ticks.js":function(e,t,n){"use strict";n.r(t),n.d(t,"tickCount",function(){return s}),n.d(t,"validTicks",function(){return u}),n.d(t,"tickValues",function(){return c}),n.d(t,"tickFormat",function(){return l});var r=n("./node_modules/vega-encode/src/scale-types.js"),i=n("./node_modules/vega-scale/index.js"),o=n("./node_modules/vega-util/index.js"),a=n("./node_modules/d3-format/index.js");function s(e,t){var n;return Object(o.isObject)(t)&&(n=t.step,t=t.interval),Object(o.isString)(t)&&(t="time"===e.type?Object(i.timeInterval)(t):"utc"===e.type?Object(i.utcInterval)(t):Object(o.error)("Only time and utc scales accept interval strings."),n&&(t=t.every(n))),t}function u(e,t,n){var r=e.range(),i=r[0],a=Object(o.peek)(r);if(i>a&&(r=a,a=i,i=r),t=t.filter(function(t){return!((t=e(t))<i||t>a)}),n>0&&t.length>1){for(var s=[t[0],Object(o.peek)(t)];t.length>n&&t.length>=3;)t=t.filter(function(e,t){return!(t%2)});t.length<3&&(t=s)}return t}function c(e,t){return e.ticks?e.ticks(t):e.domain()}function l(e,t,n){var i,o,s=e.tickFormat?e.tickFormat(t,n):String;return e.type===r.Log?(i=s,o=function(e){var t=Object(a.formatSpecifier)(e||",");if(null==t.precision){switch(t.precision=12,t.type){case"%":t.precision-=2;break;case"e":t.precision-=1}return n=Object(a.format)(t),r=Object(a.format)(".1f")(1)[1],function(e){var t,i,o=n(e),a=o.indexOf(r);if(a<0)return o;for(t=function(e,t){var n,r=e.lastIndexOf("e");if(r>0)return r;for(r=e.length;--r>t;)if((n=e.charCodeAt(r))>=48&&n<=57)return r+1}(o,a),i=t<o.length?o.slice(t):"";--t>a;)if("0"!==o[t]){++t;break}return o.slice(0,t)+i}}return Object(a.format)(t);var n,r}(n),function(e){return i(e)?o(e):""}):s}},"./node_modules/vega-event-selector/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-event-selector/src/event-selector.js");n.d(t,"selector",function(){return r.default})},"./node_modules/vega-event-selector/src/event-selector.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n){return r=t||o,i=n||m,v(e.trim()).map(_)};var r,i,o="view",a="[",s="]",u="{",c="}",l=":",d=",",p="@",f=">",h=/[[\]{}]/,m={"*":1,arc:1,area:1,group:1,image:1,line:1,path:1,rect:1,rule:1,shape:1,symbol:1,text:1,trail:1};function g(e,t,n,r,i){for(var o,a=0,s=e.length;t<s;++t){if(o=e[t],!a&&o===n)return t;i&&i.indexOf(o)>=0?--a:r&&r.indexOf(o)>=0&&++a}return t}function v(e){for(var t=[],n=0,r=e.length,i=0;i<r;)i=g(e,i,d,a+u,s+c),t.push(e.substring(n,i).trim()),n=++i;if(0===t.length)throw"Empty event selector: "+e;return t}function _(e){return"["===e[0]?function(e){var t,n,r=e.length,i=1;if((i=g(e,i,s,a,s))===r)throw"Empty between selector: "+e;if(2!==(t=v(e.substring(1,i))).length)throw"Between selector must have two elements: "+e;if((e=e.slice(i+1).trim())[0]!==f)throw"Expected '>' after between selector: "+e;if(t=t.map(_),(n=_(e.slice(1).trim())).between)return{between:t,stream:n};n.between=t;return n}(e):function(e){var t,n,o={source:r},f=[],m=[0,0],v=0,_=0,y=e.length,b=0;if(e[y-1]===c){if(!((b=e.lastIndexOf(u))>=0))throw"Unmatched right brace: "+e;try{m=function(e){var t=e.split(d);if(!e.length||t.length>2)throw e;return t.map(function(t){var n=+t;if(n!=n)throw e;return n})}(e.substring(b+1,y-1))}catch(t){throw"Invalid throttle specification: "+e}e=e.slice(0,b).trim(),y=e.length,b=0}if(!y)throw e;e[0]===p&&(v=++b);(t=g(e,b,l))<y&&(f.push(e.substring(_,t).trim()),_=b=++t);if((b=g(e,b,a))===y)f.push(e.substring(_,y).trim());else if(f.push(e.substring(_,b).trim()),n=[],(_=++b)===y)throw"Unmatched left bracket: "+e;for(;b<y;){if((b=g(e,b,s))===y)throw"Unmatched left bracket: "+e;if(n.push(e.substring(_,b).trim()),b<y-1&&e[++b]!==a)throw"Expected left bracket: "+e;_=++b}if(!(y=f.length)||h.test(f[y-1]))throw"Invalid event selector: "+e;y>1?(o.type=f[1],v?o.markname=f[0].slice(1):(j=f[0],i.hasOwnProperty(j)?o.marktype=f[0]:o.source=f[0])):o.type=f[0];var j;"!"===o.type.slice(-1)&&(o.consume=!0,o.type=o.type.slice(0,-1));null!=n&&(o.filter=n);m[0]&&(o.throttle=m[0]);m[1]&&(o.debounce=m[1]);return o}(e)}},"./node_modules/vega-expression/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-expression/src/ast.js");n.d(t,"ASTNode",function(){return r.default});var i=n("./node_modules/vega-expression/src/parser.js");n.d(t,"parse",function(){return i.default});var o=n("./node_modules/vega-expression/src/codegen.js");n.d(t,"codegen",function(){return o.default});var a=n("./node_modules/vega-expression/src/functions.js");n.d(t,"functions",function(){return a.default});var s=n("./node_modules/vega-expression/src/constants.js");n.d(t,"constants",function(){return s.default})},"./node_modules/vega-expression/src/ast.js":function(e,t,n){"use strict";function r(e){this.type=e}n.r(t),n.d(t,"default",function(){return r}),r.prototype.visit=function(e){var t,n,r;if(e(this))return 1;for(n=0,r=(t=function(e){switch(e.type){case"ArrayExpression":return e.elements;case"BinaryExpression":case"LogicalExpression":return[e.left,e.right];case"CallExpression":var t=e.arguments.slice();return t.unshift(e.callee),t;case"ConditionalExpression":return[e.test,e.consequent,e.alternate];case"MemberExpression":return[e.object,e.property];case"ObjectExpression":return e.properties;case"Property":return[e.key,e.value];case"UnaryExpression":return[e.argument];case"Identifier":case"Literal":case"RawCode":default:return[]}}(this)).length;n<r;++n)if(t[n].visit(e))return 1}},"./node_modules/vega-expression/src/codegen.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-expression/src/constants.js"),i=n("./node_modules/vega-expression/src/functions.js"),o=n("./node_modules/vega-util/index.js");t.default=function(e){var t=(e=e||{}).whitelist?Object(o.toSet)(e.whitelist):{},n=e.blacklist?Object(o.toSet)(e.blacklist):{},a=e.constants||r.default,s=(e.functions||i.default)(h),u=e.globalvar,c=e.fieldvar,l={},d={},p=0,f=Object(o.isFunction)(u)?u:function(e){return u+'["'+e+'"]'};function h(e){if(Object(o.isString)(e))return e;var t=m[e.type];return null==t&&Object(o.error)("Unsupported type: "+e.type),t(e)}var m={Literal:function(e){return e.raw},Identifier:function(e){var r=e.name;return p>0?r:n.hasOwnProperty(r)?Object(o.error)("Illegal identifier: "+r):a.hasOwnProperty(r)?a[r]:t.hasOwnProperty(r)?r:(l[r]=1,f(r))},MemberExpression:function(e){var t=!e.computed,n=h(e.object);t&&(p+=1);var r=h(e.property);return n===c&&(d[r]=1),t&&(p-=1),n+(t?"."+r:"["+r+"]")},CallExpression:function(e){"Identifier"!==e.callee.type&&Object(o.error)("Illegal callee type: "+e.callee.type);var t=e.callee.name,n=e.arguments,r=s.hasOwnProperty(t)&&s[t];return r||Object(o.error)("Unrecognized function: "+t),Object(o.isFunction)(r)?r(n):r+"("+n.map(h).join(",")+")"},ArrayExpression:function(e){return"["+e.elements.map(h).join(",")+"]"},BinaryExpression:function(e){return"("+h(e.left)+e.operator+h(e.right)+")"},UnaryExpression:function(e){return"("+e.operator+h(e.argument)+")"},ConditionalExpression:function(e){return"("+h(e.test)+"?"+h(e.consequent)+":"+h(e.alternate)+")"},LogicalExpression:function(e){return"("+h(e.left)+e.operator+h(e.right)+")"},ObjectExpression:function(e){return"{"+e.properties.map(h).join(",")+"}"},Property:function(e){p+=1;var t=h(e.key);return p-=1,t+":"+h(e.value)}};function g(e){var t={code:h(e),globals:Object.keys(l),fields:Object.keys(d)};return l={},d={},t}return g.functions=s,g.constants=a,g}},"./node_modules/vega-expression/src/constants.js":function(e,t,n){"use strict";n.r(t),t.default={NaN:"NaN",E:"Math.E",LN2:"Math.LN2",LN10:"Math.LN10",LOG2E:"Math.LOG2E",LOG10E:"Math.LOG10E",PI:"Math.PI",SQRT1_2:"Math.SQRT1_2",SQRT2:"Math.SQRT2",MIN_VALUE:"Number.MIN_VALUE",MAX_VALUE:"Number.MAX_VALUE"}},"./node_modules/vega-expression/src/functions.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/index.js");t.default=function(e){function t(t,n,r){return function(i){return o=t,s=n,u=r,c=e((a=i)[0]),s&&(c=s+"("+c+")",0===s.lastIndexOf("new ",0)&&(c="("+c+")")),c+"."+o+(u<0?"":0===u?"()":"("+a.slice(1).map(e).join(",")+")");var o,a,s,u,c}}var n="new Date";return{isNaN:"isNaN",isFinite:"isFinite",abs:"Math.abs",acos:"Math.acos",asin:"Math.asin",atan:"Math.atan",atan2:"Math.atan2",ceil:"Math.ceil",cos:"Math.cos",exp:"Math.exp",floor:"Math.floor",log:"Math.log",max:"Math.max",min:"Math.min",pow:"Math.pow",random:"Math.random",round:"Math.round",sin:"Math.sin",sqrt:"Math.sqrt",tan:"Math.tan",clamp:function(t){t.length<3&&Object(r.error)("Missing arguments to clamp function."),t.length>3&&Object(r.error)("Too many arguments to clamp function.");var n=t.map(e);return"Math.max("+n[1]+", Math.min("+n[2]+","+n[0]+"))"},now:"Date.now",utc:"Date.UTC",datetime:n,date:t("getDate",n,0),day:t("getDay",n,0),year:t("getFullYear",n,0),month:t("getMonth",n,0),hours:t("getHours",n,0),minutes:t("getMinutes",n,0),seconds:t("getSeconds",n,0),milliseconds:t("getMilliseconds",n,0),time:t("getTime",n,0),timezoneoffset:t("getTimezoneOffset",n,0),utcdate:t("getUTCDate",n,0),utcday:t("getUTCDay",n,0),utcyear:t("getUTCFullYear",n,0),utcmonth:t("getUTCMonth",n,0),utchours:t("getUTCHours",n,0),utcminutes:t("getUTCMinutes",n,0),utcseconds:t("getUTCSeconds",n,0),utcmilliseconds:t("getUTCMilliseconds",n,0),length:t("length",null,-1),indexof:t("indexOf",null),lastindexof:t("lastIndexOf",null),slice:t("slice",null),parseFloat:"parseFloat",parseInt:"parseInt",upper:t("toUpperCase","String",0),lower:t("toLowerCase","String",0),substring:t("substring","String"),replace:t("replace","String"),regexp:"RegExp",test:t("test","RegExp"),if:function(t){t.length<3&&Object(r.error)("Missing arguments to if function."),t.length>3&&Object(r.error)("Too many arguments to if function.");var n=t.map(e);return"("+n[0]+"?"+n[1]+":"+n[2]+")"}}}},"./node_modules/vega-expression/src/parser.js":function(e,t,n){"use strict";n.r(t);var r,i,o,a,s,u=n("./node_modules/vega-expression/src/ast.js"),c=1,l=2,d=3,p=4,f=5,h=6,m=7,g=8;(r={})[c]="Boolean",r[l]="<end>",r[d]="Identifier",r[p]="Keyword",r[f]="Null",r[h]="Numeric",r[m]="Punctuator",r[g]="String",r[9]="RegularExpression";var v="ArrayExpression",_="BinaryExpression",y="CallExpression",b="ConditionalExpression",j="Identifier",x="Literal",O="LogicalExpression",w="MemberExpression",C="ObjectExpression",E="Property",S="UnaryExpression",k="Unexpected token %0",M="Unexpected number",A="Unexpected string",T="Unexpected identifier",D="Unexpected reserved word",I="Unexpected end of input",P="Invalid regular expression",R="Invalid regular expression: missing /",N="Octal literals are not allowed in strict mode.",F="Duplicate data property in object literal not allowed in strict mode",L="ILLEGAL",V="Disabled.",B=new RegExp("[ªµºÀ-ÖØ-öø-ˁˆ-ˑˠ-ˤˬˮͰ-ʹͶͷͺ-ͽͿΆΈ-ΊΌΎ-ΡΣ-ϵϷ-ҁҊ-ԯԱ-Ֆՙա-ևא-תװ-ײؠ-يٮٯٱ-ۓەۥۦۮۯۺ-ۼۿܐܒ-ܯݍ-ޥޱߊ-ߪߴߵߺࠀ-ࠕࠚࠤࠨࡀ-ࡘࢠ-ࢲऄ-हऽॐक़-ॡॱ-ঀঅ-ঌএঐও-নপ-রলশ-হঽৎড়ঢ়য়-ৡৰৱਅ-ਊਏਐਓ-ਨਪ-ਰਲਲ਼ਵਸ਼ਸਹਖ਼-ੜਫ਼ੲ-ੴઅ-ઍએ-ઑઓ-નપ-રલળવ-હઽૐૠૡଅ-ଌଏଐଓ-ନପ-ରଲଳଵ-ହଽଡ଼ଢ଼ୟ-ୡୱஃஅ-ஊஎ-ஐஒ-கஙசஜஞடணதந-பம-ஹௐఅ-ఌఎ-ఐఒ-నప-హఽౘౙౠౡಅ-ಌಎ-ಐಒ-ನಪ-ಳವ-ಹಽೞೠೡೱೲഅ-ഌഎ-ഐഒ-ഺഽൎൠൡൺ-ൿඅ-ඖක-නඳ-රලව-ෆก-ะาำเ-ๆກຂຄງຈຊຍດ-ທນ-ຟມ-ຣລວສຫອ-ະາຳຽເ-ໄໆໜ-ໟༀཀ-ཇཉ-ཬྈ-ྌက-ဪဿၐ-ၕၚ-ၝၡၥၦၮ-ၰၵ-ႁႎႠ-ჅჇჍა-ჺჼ-ቈቊ-ቍቐ-ቖቘቚ-ቝበ-ኈኊ-ኍነ-ኰኲ-ኵኸ-ኾዀዂ-ዅወ-ዖዘ-ጐጒ-ጕጘ-ፚᎀ-ᎏᎠ-Ᏼᐁ-ᙬᙯ-ᙿᚁ-ᚚᚠ-ᛪᛮ-ᛸᜀ-ᜌᜎ-ᜑᜠ-ᜱᝀ-ᝑᝠ-ᝬᝮ-ᝰក-ឳៗៜᠠ-ᡷᢀ-ᢨᢪᢰ-ᣵᤀ-ᤞᥐ-ᥭᥰ-ᥴᦀ-ᦫᧁ-ᧇᨀ-ᨖᨠ-ᩔᪧᬅ-ᬳᭅ-ᭋᮃ-ᮠᮮᮯᮺ-ᯥᰀ-ᰣᱍ-ᱏᱚ-ᱽᳩ-ᳬᳮ-ᳱᳵᳶᴀ-ᶿḀ-ἕἘ-Ἕἠ-ὅὈ-Ὅὐ-ὗὙὛὝὟ-ώᾀ-ᾴᾶ-ᾼιῂ-ῄῆ-ῌῐ-ΐῖ-Ίῠ-Ῥῲ-ῴῶ-ῼⁱⁿₐ-ₜℂℇℊ-ℓℕℙ-ℝℤΩℨK-ℭℯ-ℹℼ-ℿⅅ-ⅉⅎⅠ-ↈⰀ-Ⱞⰰ-ⱞⱠ-ⳤⳫ-ⳮⳲⳳⴀ-ⴥⴧⴭⴰ-ⵧⵯⶀ-ⶖⶠ-ⶦⶨ-ⶮⶰ-ⶶⶸ-ⶾⷀ-ⷆⷈ-ⷎⷐ-ⷖⷘ-ⷞⸯ々-〇〡-〩〱-〵〸-〼ぁ-ゖゝ-ゟァ-ヺー-ヿㄅ-ㄭㄱ-ㆎㆠ-ㆺㇰ-ㇿ㐀-䶵一-鿌ꀀ-ꒌꓐ-ꓽꔀ-ꘌꘐ-ꘟꘪꘫꙀ-ꙮꙿ-ꚝꚠ-ꛯꜗ-ꜟꜢ-ꞈꞋ-ꞎꞐ-ꞭꞰꞱꟷ-ꠁꠃ-ꠅꠇ-ꠊꠌ-ꠢꡀ-ꡳꢂ-ꢳꣲ-ꣷꣻꤊ-ꤥꤰ-ꥆꥠ-ꥼꦄ-ꦲꧏꧠ-ꧤꧦ-ꧯꧺ-ꧾꨀ-ꨨꩀ-ꩂꩄ-ꩋꩠ-ꩶꩺꩾ-ꪯꪱꪵꪶꪹ-ꪽꫀꫂꫛ-ꫝꫠ-ꫪꫲ-ꫴꬁ-ꬆꬉ-ꬎꬑ-ꬖꬠ-ꬦꬨ-ꬮꬰ-ꭚꭜ-ꭟꭤꭥꯀ-ꯢ가-힣ힰ-ퟆퟋ-ퟻ豈-舘並-龎ﬀ-ﬆﬓ-ﬗיִײַ-ﬨשׁ-זּטּ-לּמּנּסּףּפּצּ-ﮱﯓ-ﴽﵐ-ﶏﶒ-ﷇﷰ-ﷻﹰ-ﹴﹶ-ﻼＡ-Ｚａ-ｚｦ-ﾾￂ-ￇￊ-ￏￒ-ￗￚ-ￜ]"),z=new RegExp("[ªµºÀ-ÖØ-öø-ˁˆ-ˑˠ-ˤˬˮ̀-ʹͶͷͺ-ͽͿΆΈ-ΊΌΎ-ΡΣ-ϵϷ-ҁ҃-҇Ҋ-ԯԱ-Ֆՙա-և֑-ׇֽֿׁׂׅׄא-תװ-ײؐ-ؚؠ-٩ٮ-ۓە-ۜ۟-۪ۨ-ۼۿܐ-݊ݍ-ޱ߀-ߵߺࠀ-࠭ࡀ-࡛ࢠ-ࢲࣤ-ॣ०-९ॱ-ঃঅ-ঌএঐও-নপ-রলশ-হ়-ৄেৈো-ৎৗড়ঢ়য়-ৣ০-ৱਁ-ਃਅ-ਊਏਐਓ-ਨਪ-ਰਲਲ਼ਵਸ਼ਸਹ਼ਾ-ੂੇੈੋ-੍ੑਖ਼-ੜਫ਼੦-ੵઁ-ઃઅ-ઍએ-ઑઓ-નપ-રલળવ-હ઼-ૅે-ૉો-્ૐૠ-ૣ૦-૯ଁ-ଃଅ-ଌଏଐଓ-ନପ-ରଲଳଵ-ହ଼-ୄେୈୋ-୍ୖୗଡ଼ଢ଼ୟ-ୣ୦-୯ୱஂஃஅ-ஊஎ-ஐஒ-கஙசஜஞடணதந-பம-ஹா-ூெ-ைொ-்ௐௗ௦-௯ఀ-ఃఅ-ఌఎ-ఐఒ-నప-హఽ-ౄె-ైొ-్ౕౖౘౙౠ-ౣ౦-౯ಁ-ಃಅ-ಌಎ-ಐಒ-ನಪ-ಳವ-ಹ಼-ೄೆ-ೈೊ-್ೕೖೞೠ-ೣ೦-೯ೱೲഁ-ഃഅ-ഌഎ-ഐഒ-ഺഽ-ൄെ-ൈൊ-ൎൗൠ-ൣ൦-൯ൺ-ൿංඃඅ-ඖක-නඳ-රලව-ෆ්ා-ුූෘ-ෟ෦-෯ෲෳก-ฺเ-๎๐-๙ກຂຄງຈຊຍດ-ທນ-ຟມ-ຣລວສຫອ-ູົ-ຽເ-ໄໆ່-ໍ໐-໙ໜ-ໟༀ༘༙༠-༩༹༵༷༾-ཇཉ-ཬཱ-྄྆-ྗྙ-ྼ࿆က-၉ၐ-ႝႠ-ჅჇჍა-ჺჼ-ቈቊ-ቍቐ-ቖቘቚ-ቝበ-ኈኊ-ኍነ-ኰኲ-ኵኸ-ኾዀዂ-ዅወ-ዖዘ-ጐጒ-ጕጘ-ፚ፝-፟ᎀ-ᎏᎠ-Ᏼᐁ-ᙬᙯ-ᙿᚁ-ᚚᚠ-ᛪᛮ-ᛸᜀ-ᜌᜎ-᜔ᜠ-᜴ᝀ-ᝓᝠ-ᝬᝮ-ᝰᝲᝳក-៓ៗៜ៝០-៩᠋-᠍᠐-᠙ᠠ-ᡷᢀ-ᢪᢰ-ᣵᤀ-ᤞᤠ-ᤫᤰ-᤻᥆-ᥭᥰ-ᥴᦀ-ᦫᦰ-ᧉ᧐-᧙ᨀ-ᨛᨠ-ᩞ᩠-᩿᩼-᪉᪐-᪙ᪧ᪰-᪽ᬀ-ᭋ᭐-᭙᭫-᭳ᮀ-᯳ᰀ-᰷᱀-᱉ᱍ-ᱽ᳐-᳔᳒-ᳶ᳸᳹ᴀ-᷵᷼-ἕἘ-Ἕἠ-ὅὈ-Ὅὐ-ὗὙὛὝὟ-ώᾀ-ᾴᾶ-ᾼιῂ-ῄῆ-ῌῐ-ΐῖ-Ίῠ-Ῥῲ-ῴῶ-ῼ‌‍‿⁀⁔ⁱⁿₐ-ₜ⃐-⃥⃜⃡-⃰ℂℇℊ-ℓℕℙ-ℝℤΩℨK-ℭℯ-ℹℼ-ℿⅅ-ⅉⅎⅠ-ↈⰀ-Ⱞⰰ-ⱞⱠ-ⳤⳫ-ⳳⴀ-ⴥⴧⴭⴰ-ⵧⵯ⵿-ⶖⶠ-ⶦⶨ-ⶮⶰ-ⶶⶸ-ⶾⷀ-ⷆⷈ-ⷎⷐ-ⷖⷘ-ⷞⷠ-ⷿⸯ々-〇〡-〯〱-〵〸-〼ぁ-ゖ゙゚ゝ-ゟァ-ヺー-ヿㄅ-ㄭㄱ-ㆎㆠ-ㆺㇰ-ㇿ㐀-䶵一-鿌ꀀ-ꒌꓐ-ꓽꔀ-ꘌꘐ-ꘫꙀ-꙯ꙴ-꙽ꙿ-ꚝꚟ-꛱ꜗ-ꜟꜢ-ꞈꞋ-ꞎꞐ-ꞭꞰꞱꟷ-ꠧꡀ-ꡳꢀ-꣄꣐-꣙꣠-ꣷꣻ꤀-꤭ꤰ-꥓ꥠ-ꥼꦀ-꧀ꧏ-꧙ꧠ-ꧾꨀ-ꨶꩀ-ꩍ꩐-꩙ꩠ-ꩶꩺ-ꫂꫛ-ꫝꫠ-ꫯꫲ-꫶ꬁ-ꬆꬉ-ꬎꬑ-ꬖꬠ-ꬦꬨ-ꬮꬰ-ꭚꭜ-ꭟꭤꭥꯀ-ꯪ꯬꯭꯰-꯹가-힣ힰ-ퟆퟋ-ퟻ豈-舘並-龎ﬀ-ﬆﬓ-ﬗיִ-ﬨשׁ-זּטּ-לּמּנּסּףּפּצּ-ﮱﯓ-ﴽﵐ-ﶏﶒ-ﷇﷰ-ﷻ︀-️︠-︭︳︴﹍-﹏ﹰ-ﹴﹶ-ﻼ０-９Ａ-Ｚ＿ａ-ｚｦ-ﾾￂ-ￇￊ-ￏￒ-ￗￚ-ￜ]");function U(e,t){if(!e)throw new Error("ASSERT: "+t)}function H(e){return e>=48&&e<=57}function q(e){return"0123456789abcdefABCDEF".indexOf(e)>=0}function G(e){return"01234567".indexOf(e)>=0}function W(e){return 10===e||13===e||8232===e||8233===e}function Y(e){return 36===e||95===e||e>=65&&e<=90||e>=97&&e<=122||92===e||e>=128&&B.test(String.fromCharCode(e))}function K(e){return 36===e||95===e||e>=65&&e<=90||e>=97&&e<=122||e>=48&&e<=57||92===e||e>=128&&z.test(String.fromCharCode(e))}var Q={if:1,in:1,do:1,var:1,for:1,new:1,try:1,let:1,this:1,else:1,case:1,void:1,with:1,enum:1,while:1,break:1,catch:1,throw:1,const:1,yield:1,class:1,super:1,return:1,typeof:1,delete:1,switch:1,export:1,import:1,public:1,static:1,default:1,finally:1,extends:1,package:1,private:1,function:1,continue:1,debugger:1,interface:1,protected:1,instanceof:1,implements:1};function X(){for(var e,t;o<a&&(e=i.charCodeAt(o),32===(t=e)||9===t||11===t||12===t||160===t||t>=5760&&[5760,6158,8192,8193,8194,8195,8196,8197,8198,8199,8200,8201,8202,8239,8287,12288,65279].indexOf(t)>=0||W(e));)++o}function Z(e){var t,n,r,s=0;for(n="u"===e?4:2,t=0;t<n;++t)o<a&&q(i[o])?(r=i[o++],s=16*s+"0123456789abcdef".indexOf(r.toLowerCase())):pe({},k,L);return String.fromCharCode(s)}function $(){var e,t,n,r;for(t=0,"}"===(e=i[o])&&pe({},k,L);o<a&&q(e=i[o++]);)t=16*t+"0123456789abcdef".indexOf(e.toLowerCase());return(t>1114111||"}"!==e)&&pe({},k,L),t<=65535?String.fromCharCode(t):(n=55296+(t-65536>>10),r=56320+(t-65536&1023),String.fromCharCode(n,r))}function J(){var e,t;for(e=i.charCodeAt(o++),t=String.fromCharCode(e),92===e&&(117!==i.charCodeAt(o)&&pe({},k,L),++o,(e=Z("u"))&&"\\"!==e&&Y(e.charCodeAt(0))||pe({},k,L),t=e);o<a&&K(e=i.charCodeAt(o));)++o,t+=String.fromCharCode(e),92===e&&(t=t.substr(0,t.length-1),117!==i.charCodeAt(o)&&pe({},k,L),++o,(e=Z("u"))&&"\\"!==e&&K(e.charCodeAt(0))||pe({},k,L),t+=e);return t}function ee(){var e,t;return e=o,{type:1===(t=92===i.charCodeAt(o)?J():function(){var e,t;for(e=o++;o<a;){if(92===(t=i.charCodeAt(o)))return o=e,J();if(!K(t))break;++o}return i.slice(e,o)}()).length?d:Q.hasOwnProperty(t)?p:"null"===t?f:"true"===t||"false"===t?c:d,value:t,start:e,end:o}}function te(){var e,t,n,r,a=o,s=i.charCodeAt(o),u=i[o];switch(s){case 46:case 40:case 41:case 59:case 44:case 123:case 125:case 91:case 93:case 58:case 63:case 126:return++o,{type:m,value:String.fromCharCode(s),start:a,end:o};default:if(61===(e=i.charCodeAt(o+1)))switch(s){case 43:case 45:case 47:case 60:case 62:case 94:case 124:case 37:case 38:case 42:return o+=2,{type:m,value:String.fromCharCode(s)+String.fromCharCode(e),start:a,end:o};case 33:case 61:return o+=2,61===i.charCodeAt(o)&&++o,{type:m,value:i.slice(a,o),start:a,end:o}}}return">>>="===(r=i.substr(o,4))?{type:m,value:r,start:a,end:o+=4}:">>>"===(n=r.substr(0,3))||"<<="===n||">>="===n?{type:m,value:n,start:a,end:o+=3}:u===(t=n.substr(0,2))[1]&&"+-<>&|".indexOf(u)>=0||"=>"===t?{type:m,value:t,start:a,end:o+=2}:"<>=!+-*%&|^/".indexOf(u)>=0?{type:m,value:u,start:a,end:++o}:void pe({},k,L)}function ne(){var e,t,n;if(U(H((n=i[o]).charCodeAt(0))||"."===n,"Numeric literal must start with a decimal digit or a decimal point"),t=o,e="","."!==n){if(e=i[o++],n=i[o],"0"===e){if("x"===n||"X"===n)return++o,function(e){for(var t="";o<a&&q(i[o]);)t+=i[o++];return 0===t.length&&pe({},k,L),Y(i.charCodeAt(o))&&pe({},k,L),{type:h,value:parseInt("0x"+t,16),start:e,end:o}}(t);if(G(n))return function(e){for(var t="0"+i[o++];o<a&&G(i[o]);)t+=i[o++];return(Y(i.charCodeAt(o))||H(i.charCodeAt(o)))&&pe({},k,L),{type:h,value:parseInt(t,8),octal:!0,start:e,end:o}}(t);n&&H(n.charCodeAt(0))&&pe({},k,L)}for(;H(i.charCodeAt(o));)e+=i[o++];n=i[o]}if("."===n){for(e+=i[o++];H(i.charCodeAt(o));)e+=i[o++];n=i[o]}if("e"===n||"E"===n)if(e+=i[o++],"+"!==(n=i[o])&&"-"!==n||(e+=i[o++]),H(i.charCodeAt(o)))for(;H(i.charCodeAt(o));)e+=i[o++];else pe({},k,L);return Y(i.charCodeAt(o))&&pe({},k,L),{type:h,value:parseFloat(e),start:t,end:o}}function re(){var e,t,n,r;return s=null,X(),e=o,t=function(){var e,t,n,r;for(U("/"===(e=i[o]),"Regular expression literal must start with a slash"),t=i[o++],n=!1,r=!1;o<a;)if(t+=e=i[o++],"\\"===e)W((e=i[o++]).charCodeAt(0))&&pe({},R),t+=e;else if(W(e.charCodeAt(0)))pe({},R);else if(n)"]"===e&&(n=!1);else{if("/"===e){r=!0;break}"["===e&&(n=!0)}return r||pe({},R),{value:t.substr(1,t.length-2),literal:t}}(),n=function(){var e,t,n;for(t="",n="";o<a&&K((e=i[o]).charCodeAt(0));)++o,"\\"===e&&o<a?pe({},k,L):(n+=e,t+=e);return n.search(/[^gimuy]/g)>=0&&pe({},P,n),{value:n,literal:t}}(),r=function(e,t){var n=e;t.indexOf("u")>=0&&(n=n.replace(/\\u\{([0-9a-fA-F]+)\}/g,function(e,t){if(parseInt(t,16)<=1114111)return"x";pe({},P)}).replace(/[\uD800-\uDBFF][\uDC00-\uDFFF]/g,"x"));try{new RegExp(n)}catch(e){pe({},P)}try{return new RegExp(e,t)}catch(e){return null}}(t.value,n.value),{literal:t.literal+n.literal,value:r,regex:{pattern:t.value,flags:n.value},start:e,end:o}}function ie(){var e;return X(),o>=a?{type:l,start:o,end:o}:Y(e=i.charCodeAt(o))?ee():40===e||41===e||59===e?te():39===e||34===e?function(){var e,t,n,r,s="",u=!1;for(U("'"===(e=i[o])||'"'===e,"String literal must starts with a quote"),t=o,++o;o<a;){if((n=i[o++])===e){e="";break}if("\\"===n)if((n=i[o++])&&W(n.charCodeAt(0)))"\r"===n&&"\n"===i[o]&&++o;else switch(n){case"u":case"x":"{"===i[o]?(++o,s+=$()):s+=Z(n);break;case"n":s+="\n";break;case"r":s+="\r";break;case"t":s+="\t";break;case"b":s+="\b";break;case"f":s+="\f";break;case"v":s+="\v";break;default:G(n)?(0!==(r="01234567".indexOf(n))&&(u=!0),o<a&&G(i[o])&&(u=!0,r=8*r+"01234567".indexOf(i[o++]),"0123".indexOf(n)>=0&&o<a&&G(i[o])&&(r=8*r+"01234567".indexOf(i[o++]))),s+=String.fromCharCode(r)):s+=n}else{if(W(n.charCodeAt(0)))break;s+=n}}return""!==e&&pe({},k,L),{type:g,value:s,octal:u,start:t,end:o}}():46===e?H(i.charCodeAt(o+1))?ne():te():H(e)?ne():te()}function oe(){var e;return o=(e=s).end,s=ie(),o=e.end,e}function ae(){var e;e=o,s=ie(),o=e}function se(e,t,n){var r=new u.default("||"===e||"&&"===e?O:_);return r.operator=e,r.left=t,r.right=n,r}function ue(e){var t=new u.default(j);return t.name=e,t}function ce(e){var t=new u.default(x);return t.value=e.value,t.raw=i.slice(e.start,e.end),e.regex&&("//"===t.raw&&(t.raw="/(?:)/"),t.regex=e.regex),t}function le(e,t,n){var r=new u.default(w);return r.computed="["===e,r.object=t,r.property=n,r.computed||(n.member=!0),r}function de(e,t,n){var r=new u.default(E);return r.key=t,r.value=n,r.kind=e,r}function pe(e,t){var n,r=Array.prototype.slice.call(arguments,2),i=t.replace(/%(\d)/g,function(e,t){return U(t<r.length,"Message reference must be in range"),r[t]});throw(n=new Error(i)).index=o,n.description=i,n}function fe(e){e.type===l&&pe(e,I),e.type===h&&pe(e,M),e.type===g&&pe(e,A),e.type===d&&pe(e,T),e.type===p&&pe(e,D),pe(e,k,e.value)}function he(e){var t=oe();t.type===m&&t.value===e||fe(t)}function me(e){return s.type===m&&s.value===e}function ge(e){return s.type===p&&s.value===e}function ve(){var e,t,n=[];for(o=s.start,he("[");!me("]");)me(",")?(oe(),n.push(null)):(n.push(Se()),me("]")||he(","));return oe(),e=n,(t=new u.default(v)).elements=e,t}function _e(){var e;return o=s.start,(e=oe()).type===g||e.type===h?(e.octal&&pe(e,N),ce(e)):ue(e.value)}function ye(){var e,t,n,r,i,a,c,p=[],f={},h=String;for(o=s.start,he("{");!me("}");)n=void 0,r=void 0,i=void 0,o=s.start,t="$"+((e=(n=s).type===d?(i=_e(),he(":"),de("init",i,Se())):n.type!==l&&n.type!==m?(r=_e(),he(":"),de("init",r,Se())):void fe(n)).key.type===j?e.key.name:h(e.key.value)),Object.prototype.hasOwnProperty.call(f,t)?pe({},F):f[t]=!0,p.push(e),me("}")||he(",");return he("}"),a=p,(c=new u.default(C)).properties=a,c}var be={if:1,this:1};function je(){var e,t,n,r;if(me("("))return he("("),r=ke(),he(")"),r;if(me("["))return ve();if(me("{"))return ye();if(e=s.type,o=s.start,e===d||be[s.value])n=ue(oe().value);else if(e===g||e===h)s.octal&&pe(s,N),n=ce(oe());else{if(e===p)throw new Error(V);e===c?((t=oe()).value="true"===t.value,n=ce(t)):e===f?((t=oe()).value=null,n=ce(t)):me("/")||me("/=")?(n=ce(re()),ae()):fe(oe())}return n}function xe(){var e=[];if(he("("),!me(")"))for(;o<a&&(e.push(Se()),!me(")"));)he(",");return he(")"),e}function Oe(){var e,t;return o=s.start,e=oe(),(t=e).type!==d&&t.type!==p&&t.type!==c&&t.type!==f&&fe(e),ue(e.value)}function we(){var e=function(){var e,t,n,r,i,o;for(e=je();;)if(me("."))he("."),e=le(".",e,Oe());else if(me("("))t=xe(),r=e,i=t,o=void 0,(o=new u.default(y)).callee=r,o.arguments=i,e=o;else{if(!me("["))break;n=void 0,he("["),n=ke(),he("]"),e=le("[",e,n)}return e}();if(s.type===m&&(me("++")||me("--")))throw new Error(V);return e}function Ce(){var e,t,n,r,i;if(s.type!==m&&s.type!==p)t=we();else{if(me("++")||me("--"))throw new Error(V);if(me("+")||me("-")||me("~")||me("!"))e=oe(),t=Ce(),n=e.value,r=t,(i=new u.default(S)).operator=n,i.argument=r,i.prefix=!0,t=i;else{if(ge("delete")||ge("void")||ge("typeof"))throw new Error(V);t=we()}}return t}function Ee(e){var t=0;if(e.type!==m&&e.type!==p)return 0;switch(e.value){case"||":t=1;break;case"&&":t=2;break;case"|":t=3;break;case"^":t=4;break;case"&":t=5;break;case"==":case"!=":case"===":case"!==":t=6;break;case"<":case">":case"<=":case">=":case"instanceof":case"in":t=7;break;case"<<":case">>":case">>>":t=8;break;case"+":case"-":t=9;break;case"*":case"/":case"%":t=11}return t}function Se(){var e,t,n,r,i,o,a;return e=function(){var e,t,n,r,i,o,a,u,c,l;if(e=s,c=Ce(),0===(i=Ee(r=s)))return c;for(r.prec=i,oe(),t=[e,s],o=[c,r,a=Ce()];(i=Ee(s))>0;){for(;o.length>2&&i<=o[o.length-2].prec;)a=o.pop(),u=o.pop().value,c=o.pop(),t.pop(),n=se(u,c,a),o.push(n);(r=oe()).prec=i,o.push(r),t.push(s),n=Ce(),o.push(n)}for(n=o[l=o.length-1],t.pop();l>1;)t.pop(),n=se(o[l-1].value,o[l-2],n),l-=2;return n}(),me("?")&&(oe(),t=Se(),he(":"),n=Se(),r=e,i=t,o=n,(a=new u.default(b)).test=r,a.consequent=i,a.alternate=o,e=a),e}function ke(){var e=Se();if(me(","))throw new Error(V);return e}t.default=function(e){o=0,a=(i=e).length,s=null,ae();var t=ke();if(s.type!==l)throw new Error("Unexpect token after expression.");return t}},"./node_modules/vega-force/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-force/src/Force.js");n.d(t,"force",function(){return r.default})},"./node_modules/vega-force/src/Force.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return d});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js"),o=n("./node_modules/d3-force/index.js"),a={center:o.forceCenter,collide:o.forceCollide,nbody:o.forceManyBody,link:o.forceLink,x:o.forceX,y:o.forceY},s="forces",u=["alpha","alphaMin","alphaTarget","velocityDecay","forces"],c=["static","iterations"],l=["x","y","vx","vy"];function d(e){r.Transform.call(this,null,e)}d.Definition={type:"Force",metadata:{modifies:!0},params:[{name:"static",type:"boolean",default:!1},{name:"restart",type:"boolean",default:!1},{name:"iterations",type:"number",default:300},{name:"alpha",type:"number",default:1},{name:"alphaMin",type:"number",default:.001},{name:"alphaTarget",type:"number",default:0},{name:"velocityDecay",type:"number",default:.4},{name:"forces",type:"param",array:!0,params:[{key:{force:"center"},params:[{name:"x",type:"number",default:0},{name:"y",type:"number",default:0}]},{key:{force:"collide"},params:[{name:"radius",type:"number",expr:!0},{name:"strength",type:"number",default:.7},{name:"iterations",type:"number",default:1}]},{key:{force:"nbody"},params:[{name:"strength",type:"number",default:-30},{name:"theta",type:"number",default:.9},{name:"distanceMin",type:"number",default:1},{name:"distanceMax",type:"number"}]},{key:{force:"link"},params:[{name:"links",type:"data"},{name:"id",type:"field"},{name:"distance",type:"number",default:30,expr:!0},{name:"strength",type:"number",expr:!0},{name:"iterations",type:"number",default:1}]},{key:{force:"x"},params:[{name:"strength",type:"number",default:.1},{name:"x",type:"field"}]},{key:{force:"y"},params:[{name:"strength",type:"number",default:.1},{name:"y",type:"field"}]}]},{name:"as",type:"string",array:!0,modify:!1,default:l}]};var p=Object(i.inherits)(d,r.Transform);function f(e,t,n,r){var o,a,c,l,d=Object(i.array)(t.forces);for(o=0,a=u.length;o<a;++o)(c=u[o])!==s&&t.modified(c)&&e[c](t[c]);for(o=0,a=d.length;o<a;++o)l=s+o,(c=n||t.modified(s,o)?m(d[o]):r&&h(d[o],r)?e.force(l):null)&&e.force(l,c);for(a=e.numForces||0;o<a;++o)e.force(s+o,null);return e.numForces=d.length,e}function h(e,t){var n,r;for(n in e)if(Object(i.isFunction)(r=e[n])&&t.modified(Object(i.accessorFields)(r)))return 1;return 0}function m(e){var t,n;a.hasOwnProperty(e.force)||Object(i.error)("Unrecognized force: "+e.force),t=a[e.force]();for(n in e)Object(i.isFunction)(t[n])&&g(t[n],e[n],e);return t}function g(e,t,n){e(Object(i.isFunction)(t)?function(e){return t(e,n)}:t)}p.transform=function(e,t){var n,r,i,a,s,l,d,p,h=this.value,m=t.changed(t.ADD_REM),g=e.modified(u),v=e.iterations||300;if(h?(m&&(t.modifies("index"),h.nodes(t.source)),(g||t.changed(t.MOD))&&f(h,e,0,t)):(this.value=(i=t.source,a=e,s=Object(o.forceSimulation)(i),l=!1,d=s.stop,p=s.restart,s.stopped=function(){return l},s.restart=function(){return l=!1,p()},s.stop=function(){return l=!0,d()},h=f(s,a,!0).on("end",function(){l=!0})),h.on("tick",(n=t.dataflow,r=this,function(){n.touch(r).run()})),e.static||(m=!0,h.tick()),t.modifies("index")),g||m||e.modified(c)||t.changed()&&e.restart)if(h.alpha(Math.max(h.alpha(),e.alpha||1)).alphaDecay(1-Math.pow(h.alphaMin(),1/v)),e.static)for(h.stop();--v>=0;)h.tick();else if(h.stopped()&&h.restart(),!m)return t.StopPropagation;return this.finish(e,t)},p.finish=function(e,t){for(var n,r=t.dataflow,i=this._argops,o=0,a=i.length;o<a;++o)if((n=i[o]).name===s&&"link"===n.op._argval.force)for(var u,c=n.op._argops,d=0,p=c.length;d<p;++d)if("links"===c[d].name&&(u=c[d].op.source)){r.pulse(u,r.changeset().reflow());break}return t.reflow(e.modified()).modifies(l)}},"./node_modules/vega-geo/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-geo/src/Contour.js");n.d(t,"contour",function(){return r.default});var i=n("./node_modules/vega-geo/src/GeoJSON.js");n.d(t,"geojson",function(){return i.default});var o=n("./node_modules/vega-geo/src/GeoPath.js");n.d(t,"geopath",function(){return o.default});var a=n("./node_modules/vega-geo/src/GeoPoint.js");n.d(t,"geopoint",function(){return a.default});var s=n("./node_modules/vega-geo/src/GeoShape.js");n.d(t,"geoshape",function(){return s.default});var u=n("./node_modules/vega-geo/src/Graticule.js");n.d(t,"graticule",function(){return u.default});var c=n("./node_modules/vega-geo/src/Projection.js");n.d(t,"projection",function(){return c.default})},"./node_modules/vega-geo/src/Contour.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return c});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js"),o=n("./node_modules/d3-array/index.js"),a=n("./node_modules/d3-contour/index.js"),s=["values","size"],u=["x","y","size","cellSize","bandwidth"];function c(e){r.Transform.call(this,null,e)}c.Definition={type:"Contour",metadata:{generates:!0},params:[{name:"size",type:"number",array:!0,length:2,required:!0},{name:"values",type:"number",array:!0},{name:"x",type:"field"},{name:"y",type:"field"},{name:"cellSize",type:"number"},{name:"bandwidth",type:"number"},{name:"count",type:"number"},{name:"nice",type:"number",default:!1},{name:"thresholds",type:"number",array:!0}]},Object(i.inherits)(c,r.Transform).transform=function(e,t){if(this.value&&!t.changed()&&!e.modified())return t.StopPropagation;var n,i,c,l,d=t.fork(t.NO_SOURCE|t.NO_FIELDS),p=e.count||10;return e.values?(n=Object(a.contours)(),i=s,c=e.values):(n=Object(a.contourDensity)(),i=u,c=t.materialize(t.SOURCE).source),n.thresholds(e.thresholds||(e.nice?p:(l=p,function(e){for(var t=Object(o.extent)(e),n=t[0],r=t[1]-n,i=[],a=1;a<=l;++a)i.push(n+r*a/(l+1));return i}))),i.forEach(function(t){null!=e[t]&&n[t](e[t])}),this.value&&(d.rem=this.value),c=c&&c.length?n(c).map(r.ingest):[],this.value=d.source=d.add=c,d}},"./node_modules/vega-geo/src/GeoJSON.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-geo/src/constants.js"),i=n("./node_modules/vega-dataflow/index.js"),o=n("./node_modules/vega-util/index.js");function a(e){i.Transform.call(this,null,e)}a.Definition={type:"GeoJSON",metadata:{},params:[{name:"fields",type:"field",array:!0,length:2},{name:"geojson",type:"field"}]},Object(o.inherits)(a,i.Transform).transform=function(e,t){var n,i=this._features,a=this._points,s=e.fields,u=s&&s[0],c=s&&s[1],l=e.geojson,d=t.ADD;n=e.modified()||t.changed(t.REM)||t.modified(Object(o.accessorFields)(l))||u&&t.modified(Object(o.accessorFields)(u))||c&&t.modified(Object(o.accessorFields)(c)),this.value&&!n||(d=t.SOURCE,this._features=i=[],this._points=a=[]),l&&t.visit(d,function(e){i.push(l(e))}),u&&c&&(t.visit(d,function(e){var t=u(e),n=c(e);null!=t&&null!=n&&(t=+t)===t&&(n=+n)===n&&a.push([t,n])}),i=i.concat({type:r.Feature,geometry:{type:r.MultiPoint,coordinates:a}})),this.value={type:r.FeatureCollection,features:i}}},"./node_modules/vega-geo/src/GeoPath.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-projection/index.js"),o=n("./node_modules/vega-util/index.js");function a(e){r.Transform.call(this,null,e)}a.Definition={type:"GeoPath",metadata:{modifies:!0},params:[{name:"projection",type:"projection"},{name:"field",type:"field"},{name:"pointRadius",type:"number",expr:!0},{name:"as",type:"string",default:"path"}]},Object(o.inherits)(a,r.Transform).transform=function(e,t){var n=t.fork(t.ALL),r=this.value,a=e.field||o.identity,s=e.as||"path",u=n.SOURCE;!r||e.modified()?(this.value=r=Object(i.getProjectionPath)(e.projection),n.materialize().reflow()):u=a===o.identity||t.modified(a.fields)?n.ADD_MOD:n.ADD;var c=function(e,t){var n=e.pointRadius();e.context(null),null!=t&&e.pointRadius(t);return n}(r,e.pointRadius);return n.visit(u,function(e){e[s]=r(a(e))}),r.pointRadius(c),n.modifies(s)}},"./node_modules/vega-geo/src/GeoPoint.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,null,e)}o.Definition={type:"GeoPoint",metadata:{modifies:!0},params:[{name:"projection",type:"projection",required:!0},{name:"fields",type:"field",array:!0,required:!0,length:2},{name:"as",type:"string",array:!0,length:2,default:["x","y"]}]},Object(i.inherits)(o,r.Transform).transform=function(e,t){var n,r=e.projection,i=e.fields[0],o=e.fields[1],a=e.as||["x","y"],s=a[0],u=a[1];function c(e){var t=r([i(e),o(e)]);t?(e[s]=t[0],e[u]=t[1]):(e[s]=void 0,e[u]=void 0)}return e.modified()?t=t.materialize().reflow(!0).visit(t.SOURCE,c):(n=t.modified(i.fields)||t.modified(o.fields),t.visit(n?t.ADD_MOD:t.ADD,c)),t.modifies(a)}},"./node_modules/vega-geo/src/GeoShape.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-projection/index.js"),o=n("./node_modules/vega-util/index.js");function a(e){r.Transform.call(this,null,e)}a.Definition={type:"GeoShape",metadata:{modifies:!0},params:[{name:"projection",type:"projection"},{name:"field",type:"field",default:"datum"},{name:"pointRadius",type:"number",expr:!0},{name:"as",type:"string",default:"shape"}]},Object(o.inherits)(a,r.Transform).transform=function(e,t){var n,r,a,s,u=t.fork(t.ALL),c=this.value,l=e.field||Object(o.field)("datum"),d=e.as||"shape",p=u.ADD_MOD;return c&&!e.modified()||(this.value=(n=Object(i.getProjectionPath)(e.projection),r=l,a=e.pointRadius,(s=null==a?function(e){return n(r(e))}:function(e){var t=n.pointRadius(),i=n.pointRadius(a)(r(e));return n.pointRadius(t),i}).context=function(e){return n.context(e),s},c=s),u.materialize().reflow(),p=u.SOURCE),u.visit(p,function(e){e[d]=c}),u.modifies(d)}},"./node_modules/vega-geo/src/Graticule.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js"),o=n("./node_modules/d3-geo/index.js");function a(e){r.Transform.call(this,[],e),this.generator=Object(o.geoGraticule)()}a.Definition={type:"Graticule",metadata:{source:!0,generates:!0,changes:!0},params:[{name:"extent",type:"array",array:!0,length:2,content:{type:"number",array:!0,length:2}},{name:"extentMajor",type:"array",array:!0,length:2,content:{type:"number",array:!0,length:2}},{name:"extentMinor",type:"array",array:!0,length:2,content:{type:"number",array:!0,length:2}},{name:"step",type:"number",array:!0,length:2},{name:"stepMajor",type:"number",array:!0,length:2,default:[90,360]},{name:"stepMinor",type:"number",array:!0,length:2,default:[10,10]},{name:"precision",type:"number",default:2.5}]},Object(i.inherits)(a,r.Transform).transform=function(e,t){var n,o=t.fork(),a=this.value,s=this.generator;if(!a.length||e.modified())for(var u in e)Object(i.isFunction)(s[u])&&s[u](e[u]);return n=s(),a.length?o.mod.push(Object(r.replace)(a[0],n)):o.add.push(Object(r.ingest)(n)),a[0]=n,o.source=a,o}},"./node_modules/vega-geo/src/Projection.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return s}),n.d(t,"collectGeoJSON",function(){return c});var r=n("./node_modules/vega-geo/src/constants.js"),i=n("./node_modules/vega-dataflow/index.js"),o=n("./node_modules/vega-projection/index.js"),a=n("./node_modules/vega-util/index.js");function s(e){i.Transform.call(this,null,e),this.modified(!0)}function u(e,t,n){Object(a.isFunction)(e[t])&&e[t](n)}function c(e){return 1===(e=Object(a.array)(e)).length?e[0]:{type:r.FeatureCollection,features:e.reduce(function(e,t){return t&&t.type===r.FeatureCollection?e.push.apply(e,t.features):Object(a.isArray)(t)?e.push.apply(e,t):e.push(t),e},[])}}Object(a.inherits)(s,i.Transform).transform=function(e,t){var n,r,i,s=this.value;return!s||e.modified("type")?(this.value=s=function(e){var t=Object(o.projection)((e||"mercator").toLowerCase());t||Object(a.error)("Unrecognized projection type: "+e);return t()}(e.type),o.projectionProperties.forEach(function(t){null!=e[t]&&u(s,t,e[t])})):o.projectionProperties.forEach(function(t){e.modified(t)&&u(s,t,e[t])}),null!=e.pointRadius&&s.path.pointRadius(e.pointRadius),e.fit&&(n=s,i=c((r=e).fit),r.extent?n.fitExtent(r.extent,i):r.size&&n.fitSize(r.size,i)),t.fork(t.NO_SOURCE|t.NO_FIELDS)}},"./node_modules/vega-geo/src/constants.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Feature",function(){return r}),n.d(t,"FeatureCollection",function(){return i}),n.d(t,"MultiPoint",function(){return o});var r="Feature",i="FeatureCollection",o="MultiPoint"},"./node_modules/vega-hierarchy/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-hierarchy/src/Nest.js");n.d(t,"nest",function(){return r.default});var i=n("./node_modules/vega-hierarchy/src/Pack.js");n.d(t,"pack",function(){return i.default});var o=n("./node_modules/vega-hierarchy/src/Partition.js");n.d(t,"partition",function(){return o.default});var a=n("./node_modules/vega-hierarchy/src/Stratify.js");n.d(t,"stratify",function(){return a.default});var s=n("./node_modules/vega-hierarchy/src/Tree.js");n.d(t,"tree",function(){return s.default});var u=n("./node_modules/vega-hierarchy/src/TreeLinks.js");n.d(t,"treelinks",function(){return u.default});var c=n("./node_modules/vega-hierarchy/src/Treemap.js");n.d(t,"treemap",function(){return c.default})},"./node_modules/vega-hierarchy/src/HierarchyLayout.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,null,e)}Object(i.inherits)(o,r.Transform).transform=function(e,t){t.source&&t.source.root||Object(i.error)(this.constructor.name+" transform requires a backing tree data source.");var n=this.layout(e.method),r=this.fields,o=t.source.root,a=e.as||r;e.field&&o.sum(e.field),e.sort&&o.sort(e.sort),function(e,t,n){for(var r,i=0,o=t.length;i<o;++i)(r=t[i])in n&&e[r](n[r])}(n,this.params,e);try{this.value=n(o)}catch(e){Object(i.error)(e)}return o.each(function(e){!function(e,t,n){for(var r=e.data,i=0,o=t.length-1;i<o;++i)r[n[i]]=e[t[i]];r[n[o]]=e.children?e.children.length:0}(e,r,a)}),t.reflow(e.modified()).modifies(a).modifies("leaf")}},"./node_modules/vega-hierarchy/src/Nest.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return s});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js"),o=n("./node_modules/d3-collection/index.js"),a=n("./node_modules/d3-hierarchy/index.js");function s(e){r.Transform.call(this,null,e)}function u(e){return e.values}s.Definition={type:"Nest",metadata:{treesource:!0,generates:!0},params:[{name:"keys",type:"field",array:!0},{name:"key",type:"field"},{name:"generate",type:"boolean"}]},Object(i.inherits)(s,r.Transform).transform=function(e,t){t.source||Object(i.error)("Nest transform requires an upstream data source.");var n,s,c,l=e.key||r.tupleid,d=e.generate,p=e.modified(),f=d||p?t.fork(t.ALL):t;return(!this.value||p||t.changed())&&(d&&this.value&&(f.materialize(f.REM),this.value.each(function(e){e.children&&f.rem.push(e)})),n=Object(i.array)(e.keys).reduce(function(e,t){return e.key(t),e},Object(o.nest)()).entries(f.materialize(f.SOURCE).source),this.value=s=Object(a.hierarchy)({values:n},u),d&&(f.materialize(f.ADD),f.source=f.source.slice(),s.each(function(e){e.children&&(e=Object(r.ingest)(e.data),f.add.push(e),f.source.push(e))})),c=s.lookup={},s.each(function(e){null!=Object(r.tupleid)(e.data)&&(c[l(e.data)]=e)})),f.source.root=this.value,f}},"./node_modules/vega-hierarchy/src/Pack.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return s});var r=n("./node_modules/vega-hierarchy/src/HierarchyLayout.js"),i=n("./node_modules/vega-util/index.js"),o=n("./node_modules/d3-hierarchy/index.js"),a=["x","y","r","depth","children"];function s(e){r.default.call(this,e)}s.Definition={type:"Pack",metadata:{tree:!0,modifies:!0},params:[{name:"field",type:"field"},{name:"sort",type:"compare"},{name:"padding",type:"number",default:0},{name:"radius",type:"field",default:null},{name:"size",type:"number",array:!0,length:2},{name:"as",type:"string",array:!0,length:3,default:a}]};var u=Object(i.inherits)(s,r.default);u.layout=o.pack,u.params=["size","padding"],u.fields=a},"./node_modules/vega-hierarchy/src/Partition.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return s});var r=n("./node_modules/vega-hierarchy/src/HierarchyLayout.js"),i=n("./node_modules/vega-util/index.js"),o=n("./node_modules/d3-hierarchy/index.js"),a=["x0","y0","x1","y1","depth","children"];function s(e){r.default.call(this,e)}s.Definition={type:"Partition",metadata:{tree:!0,modifies:!0},params:[{name:"field",type:"field"},{name:"sort",type:"compare"},{name:"padding",type:"number",default:0},{name:"round",type:"boolean",default:!1},{name:"size",type:"number",array:!0,length:2},{name:"as",type:"string",array:!0,length:4,default:a}]};var u=Object(i.inherits)(s,r.default);u.layout=o.partition,u.params=["size","round","padding"],u.fields=a},"./node_modules/vega-hierarchy/src/Stratify.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js"),o=n("./node_modules/d3-hierarchy/index.js");function a(e){r.Transform.call(this,null,e)}a.Definition={type:"Stratify",metadata:{treesource:!0},params:[{name:"key",type:"field",required:!0},{name:"parentKey",type:"field",required:!0}]},Object(i.inherits)(a,r.Transform).transform=function(e,t){t.source||Object(i.error)("Stratify transform requires an upstream data source.");var n,r,a=e.modified();return(!this.value||a||t.changed(t.ADD_REM)||t.modified(e.key.fields)||t.modified(e.parentKey.fields))&&(n=Object(o.stratify)().id(e.key).parentId(e.parentKey)(t.source),r=n.lookup={},n.each(function(t){r[e.key(t.data)]=t}),this.value=n),t.source.root=this.value,a?t.fork(t.ALL):t}},"./node_modules/vega-hierarchy/src/Tree.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return u});var r=n("./node_modules/vega-hierarchy/src/HierarchyLayout.js"),i=n("./node_modules/vega-util/index.js"),o=n("./node_modules/d3-hierarchy/index.js"),a={tidy:o.tree,cluster:o.cluster},s=["x","y","depth","children"];function u(e){r.default.call(this,e)}u.Definition={type:"Tree",metadata:{tree:!0,modifies:!0},params:[{name:"field",type:"field"},{name:"sort",type:"compare"},{name:"method",type:"enum",default:"tidy",values:["tidy","cluster"]},{name:"size",type:"number",array:!0,length:2},{name:"nodeSize",type:"number",array:!0,length:2},{name:"as",type:"string",array:!0,length:4,default:s}]};var c=Object(i.inherits)(u,r.default);c.layout=function(e){var t=e||"tidy";if(a.hasOwnProperty(t))return a[t]();Object(i.error)("Unrecognized Tree layout method: "+t)},c.params=["size","nodeSize","separation"],c.fields=s},"./node_modules/vega-hierarchy/src/TreeLinks.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,{},e)}o.Definition={type:"TreeLinks",metadata:{tree:!0,generates:!0,changes:!0},params:[{name:"key",type:"field"}]},Object(i.inherits)(o,r.Transform).transform=function(e,t){t.source&&t.source.root||Object(i.error)("TreeLinks transform requires a backing tree data source.");var n=t.source.root.lookup,o=this.value,a=e.key||r.tupleid,s={},u=t.fork();function c(e){var t=o[e];t&&(s[e]=1,u.mod.push(t))}return t.visit(t.REM,function(e){var t=a(e),n=o[t];n&&(delete o[t],u.rem.push(n))}),t.visit(t.ADD,function(e){var t,i,c,l=a(e);(t=(i=n[l]).parent&&(c=i.parent.data)&&null!=Object(r.tupleid)(c)&&c)&&(u.add.push(o[l]=Object(r.ingest)({source:t,target:e})),s[l]=1)}),t.visit(t.MOD,function(e){var t=a(e),r=n[t].children;if(c(t),r)for(var i=0,o=r.length;i<o;++i)s[t=a(r[i].data)]||c(t)}),u}},"./node_modules/vega-hierarchy/src/Treemap.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return u});var r=n("./node_modules/vega-hierarchy/src/HierarchyLayout.js"),i=n("./node_modules/vega-util/index.js"),o=n("./node_modules/d3-hierarchy/index.js"),a={binary:o.treemapBinary,dice:o.treemapDice,slice:o.treemapSlice,slicedice:o.treemapSliceDice,squarify:o.treemapSquarify,resquarify:o.treemapResquarify},s=["x0","y0","x1","y1","depth","children"];function u(e){r.default.call(this,e)}u.Definition={type:"Treemap",metadata:{tree:!0,modifies:!0},params:[{name:"field",type:"field"},{name:"sort",type:"compare"},{name:"method",type:"enum",default:"squarify",values:["squarify","resquarify","binary","dice","slice","slicedice"]},{name:"padding",type:"number",default:0},{name:"paddingInner",type:"number",default:0},{name:"paddingOuter",type:"number",default:0},{name:"paddingTop",type:"number",default:0},{name:"paddingRight",type:"number",default:0},{name:"paddingBottom",type:"number",default:0},{name:"paddingLeft",type:"number",default:0},{name:"ratio",type:"number",default:1.618033988749895},{name:"round",type:"boolean",default:!1},{name:"size",type:"number",array:!0,length:2},{name:"as",type:"string",array:!0,length:4,default:s}]};var c=Object(i.inherits)(u,r.default);c.layout=function(){var e=Object(o.treemap)();return e.ratio=function(t){var n=e.tile();n.ratio&&e.tile(n.ratio(t))},e.method=function(t){a.hasOwnProperty(t)?e.tile(a[t]):Object(i.error)("Unrecognized Treemap layout method: "+t)},e},c.params=["method","ratio","size","round","padding","paddingInner","paddingOuter","paddingTop","paddingRight","paddingBottom","paddingLeft"],c.fields=s},"./node_modules/vega-lib/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/index.js"),i=n("./node_modules/vega-dataflow/index.js"),o=n("./node_modules/vega-transforms/index.js"),a=n("./node_modules/vega-view-transforms/index.js"),s=n("./node_modules/vega-encode/index.js"),u=n("./node_modules/vega-geo/index.js"),c=n("./node_modules/vega-force/index.js"),l=n("./node_modules/vega-hierarchy/index.js"),d=n("./node_modules/vega-voronoi/index.js"),p=n("./node_modules/vega-wordcloud/index.js"),f=n("./node_modules/vega-crossfilter/index.js"),h=n("./node_modules/vega-lib/package.json");n.d(t,"version",function(){return h.version});var m=n("./node_modules/vega-statistics/index.js");n.d(t,"bin",function(){return m.bin}),n.d(t,"bootstrapCI",function(){return m.bootstrapCI}),n.d(t,"quartiles",function(){return m.quartiles}),n.d(t,"random",function(){return m.random}),n.d(t,"setRandom",function(){return m.setRandom}),n.d(t,"randomInteger",function(){return m.randomInteger}),n.d(t,"randomKDE",function(){return m.randomKDE}),n.d(t,"randomMixture",function(){return m.randomMixture}),n.d(t,"randomNormal",function(){return m.randomNormal}),n.d(t,"randomUniform",function(){return m.randomUniform}),n.d(t,"accessor",function(){return r.accessor}),n.d(t,"accessorName",function(){return r.accessorName}),n.d(t,"accessorFields",function(){return r.accessorFields}),n.d(t,"id",function(){return r.id}),n.d(t,"identity",function(){return r.identity}),n.d(t,"zero",function(){return r.zero}),n.d(t,"one",function(){return r.one}),n.d(t,"truthy",function(){return r.truthy}),n.d(t,"falsy",function(){return r.falsy}),n.d(t,"logger",function(){return r.logger}),n.d(t,"None",function(){return r.None}),n.d(t,"Error",function(){return r.Error}),n.d(t,"Warn",function(){return r.Warn}),n.d(t,"Info",function(){return r.Info}),n.d(t,"Debug",function(){return r.Debug}),n.d(t,"panLinear",function(){return r.panLinear}),n.d(t,"panLog",function(){return r.panLog}),n.d(t,"panPow",function(){return r.panPow}),n.d(t,"zoomLinear",function(){return r.zoomLinear}),n.d(t,"zoomLog",function(){return r.zoomLog}),n.d(t,"zoomPow",function(){return r.zoomPow}),n.d(t,"array",function(){return r.array}),n.d(t,"compare",function(){return r.compare}),n.d(t,"constant",function(){return r.constant}),n.d(t,"debounce",function(){return r.debounce}),n.d(t,"error",function(){return r.error}),n.d(t,"extend",function(){return r.extend}),n.d(t,"extentIndex",function(){return r.extentIndex}),n.d(t,"fastmap",function(){return r.fastmap}),n.d(t,"field",function(){return r.field}),n.d(t,"inherits",function(){return r.inherits}),n.d(t,"isArray",function(){return r.isArray}),n.d(t,"isBoolean",function(){return r.isBoolean}),n.d(t,"isDate",function(){return r.isDate}),n.d(t,"isFunction",function(){return r.isFunction}),n.d(t,"isNumber",function(){return r.isNumber}),n.d(t,"isObject",function(){return r.isObject}),n.d(t,"isRegExp",function(){return r.isRegExp}),n.d(t,"isString",function(){return r.isString}),n.d(t,"key",function(){return r.key}),n.d(t,"merge",function(){return r.merge}),n.d(t,"pad",function(){return r.pad}),n.d(t,"peek",function(){return r.peek}),n.d(t,"repeat",function(){return r.repeat}),n.d(t,"splitAccessPath",function(){return r.splitAccessPath}),n.d(t,"stringValue",function(){return r.stringValue}),n.d(t,"toBoolean",function(){return r.toBoolean}),n.d(t,"toDate",function(){return r.toDate}),n.d(t,"toNumber",function(){return r.toNumber}),n.d(t,"toString",function(){return r.toString}),n.d(t,"toSet",function(){return r.toSet}),n.d(t,"truncate",function(){return r.truncate}),n.d(t,"visitArray",function(){return r.visitArray});var g=n("./node_modules/vega-loader/index.js");n.d(t,"loader",function(){return g.loader}),n.d(t,"read",function(){return g.read}),n.d(t,"inferType",function(){return g.inferType}),n.d(t,"inferTypes",function(){return g.inferTypes}),n.d(t,"typeParsers",function(){return g.typeParsers}),n.d(t,"formats",function(){return g.formats});var v=n("./node_modules/vega-scenegraph/index.js");n.d(t,"Bounds",function(){return v.Bounds}),n.d(t,"Gradient",function(){return v.Gradient}),n.d(t,"GroupItem",function(){return v.GroupItem}),n.d(t,"ResourceLoader",function(){return v.ResourceLoader}),n.d(t,"Item",function(){return v.Item}),n.d(t,"Scenegraph",function(){return v.Scenegraph}),n.d(t,"Handler",function(){return v.Handler}),n.d(t,"Renderer",function(){return v.Renderer}),n.d(t,"CanvasHandler",function(){return v.CanvasHandler}),n.d(t,"CanvasRenderer",function(){return v.CanvasRenderer}),n.d(t,"SVGHandler",function(){return v.SVGHandler}),n.d(t,"SVGRenderer",function(){return v.SVGRenderer}),n.d(t,"SVGStringRenderer",function(){return v.SVGStringRenderer}),n.d(t,"RenderType",function(){return v.RenderType}),n.d(t,"renderModule",function(){return v.renderModule}),n.d(t,"Marks",function(){return v.Marks}),n.d(t,"boundClip",function(){return v.boundClip}),n.d(t,"boundContext",function(){return v.boundContext}),n.d(t,"boundStroke",function(){return v.boundStroke}),n.d(t,"boundItem",function(){return v.boundItem}),n.d(t,"boundMark",function(){return v.boundMark}),n.d(t,"pathCurves",function(){return v.pathCurves}),n.d(t,"pathSymbols",function(){return v.pathSymbols}),n.d(t,"pathRectangle",function(){return v.pathRectangle}),n.d(t,"pathTrail",function(){return v.pathTrail}),n.d(t,"pathParse",function(){return v.pathParse}),n.d(t,"pathRender",function(){return v.pathRender}),n.d(t,"point",function(){return v.point}),n.d(t,"domCreate",function(){return v.domCreate}),n.d(t,"domFind",function(){return v.domFind}),n.d(t,"domChild",function(){return v.domChild}),n.d(t,"domClear",function(){return v.domClear}),n.d(t,"openTag",function(){return v.openTag}),n.d(t,"closeTag",function(){return v.closeTag}),n.d(t,"font",function(){return v.font}),n.d(t,"textMetrics",function(){return v.textMetrics}),n.d(t,"resetSVGClipId",function(){return v.resetSVGClipId}),n.d(t,"sceneEqual",function(){return v.sceneEqual}),n.d(t,"pathEqual",function(){return v.pathEqual}),n.d(t,"sceneToJSON",function(){return v.sceneToJSON}),n.d(t,"sceneFromJSON",function(){return v.sceneFromJSON}),n.d(t,"sceneZOrder",function(){return v.sceneZOrder}),n.d(t,"sceneVisit",function(){return v.sceneVisit}),n.d(t,"scenePickVisit",function(){return v.scenePickVisit}),n.d(t,"Dataflow",function(){return i.Dataflow}),n.d(t,"EventStream",function(){return i.EventStream}),n.d(t,"Parameters",function(){return i.Parameters}),n.d(t,"Pulse",function(){return i.Pulse}),n.d(t,"MultiPulse",function(){return i.MultiPulse}),n.d(t,"Operator",function(){return i.Operator}),n.d(t,"Transform",function(){return i.Transform}),n.d(t,"changeset",function(){return i.changeset}),n.d(t,"ingest",function(){return i.ingest}),n.d(t,"isTuple",function(){return i.isTuple}),n.d(t,"definition",function(){return i.definition}),n.d(t,"transform",function(){return i.transform}),n.d(t,"transforms",function(){return i.transforms}),n.d(t,"tupleid",function(){return i.tupleid});var _=n("./node_modules/vega-scale/index.js");n.d(t,"scale",function(){return _.scale}),n.d(t,"scheme",function(){return _.scheme}),n.d(t,"interpolate",function(){return _.interpolate}),n.d(t,"interpolateRange",function(){return _.interpolateRange}),n.d(t,"timeInterval",function(){return _.timeInterval}),n.d(t,"utcInterval",function(){return _.utcInterval});var y=n("./node_modules/vega-projection/index.js");n.d(t,"projection",function(){return y.projection});var b=n("./node_modules/vega-view/index.js");n.d(t,"View",function(){return b.View});var j=n("./node_modules/vega-parser/index.js");n.d(t,"parse",function(){return j.parse}),n.d(t,"expressionFunction",function(){return j.expressionFunction}),n.d(t,"formatLocale",function(){return j.formatLocale}),n.d(t,"timeFormatLocale",function(){return j.timeFormatLocale});var x=n("./node_modules/vega-runtime/index.js");n.d(t,"runtime",function(){return x.parse}),n.d(t,"runtimeContext",function(){return x.context}),Object(r.extend)(i.transforms,o,a,s,u,c,l,d,p,f)},"./node_modules/vega-lib/package.json":function(e){e.exports={_args:[["vega-lib@3.1.0","/home/travis/build/opencobra/memote/memote-report-app"]],_from:"vega-lib@3.1.0",_id:"vega-lib@3.1.0",_inBundle:!1,_integrity:"sha512-76M6DTK6ErAEIWrbah2IPy1CYgTfvYIyqrdGid5cNanUWWXAonP2arcmibWIYxLhbAr7o6QWL+GQM1vkqwMxhg==",_location:"/vega-lib",_phantomChildren:{},_requested:{type:"version",registry:!0,raw:"vega-lib@3.1.0",name:"vega-lib",escapedName:"vega-lib",rawSpec:"3.1.0",saveSpec:null,fetchSpec:"3.1.0"},_requiredBy:["/"],_resolved:"https://registry.npmjs.org/vega-lib/-/vega-lib-3.1.0.tgz",_spec:"3.1.0",_where:"/home/travis/build/opencobra/memote/memote-report-app",author:{name:"UW Interactive Data Lab",url:"http://idl.cs.washington.edu"},browser:{buffer:!1,canvas:!1,"canvas-prebuilt":!1,fs:!1},bugs:{url:"https://github.com/vega/vega-lib/issues"},contributors:[{name:"Jeffrey Heer",url:"http://idl.cs.washington.edu"},{name:"Arvind Satyanarayan",url:"http://arvindsatya.com"},{name:"Ryan Russell",url:"https://github.com/RussellSprouts"},{name:"Emily Gu",url:"https://github.com/emilygu"},{name:"Dominik Moritz",url:"https://www.domoritz.de"},{name:"Yuri Astrakhan",url:"https://github.com/nyurik"}],dependencies:{"vega-crossfilter":"2","vega-dataflow":"3","vega-encode":"2","vega-expression":"^2.3","vega-force":"2","vega-geo":"^2.2","vega-hierarchy":"^2.1","vega-loader":"2","vega-parser":"^2.5","vega-projection":"1","vega-runtime":"2","vega-scale":"^2.1","vega-scenegraph":"^2.3","vega-statistics":"^1.2","vega-transforms":"^1.2","vega-util":"^1.7","vega-view":"^2.2","vega-view-transforms":"^1.2","vega-voronoi":"2","vega-wordcloud":"^2.1"},description:"Include Vega in projeccts using minimal dependencies.",devDependencies:{eslint:"4",rollup:"0.43","rollup-plugin-json":"2","rollup-plugin-node-resolve":"3",tape:"4",tv4:"1","uglify-js":"3"},homepage:"https://github.com/vega/vega-lib#readme","jsnext:main":"index",keywords:["vega","visualization","interaction","dataflow","library","data","d3"],license:"BSD-3-Clause",main:"build/vega.js",module:"index",name:"vega-lib",repository:{type:"git",url:"git+https://github.com/vega/vega-lib.git"},scripts:{__postpublish:"git push && git push --tags",build:"npm run rollup && bin/rollup -e",postbuild:"uglifyjs build/vega.js -c -m -o build/vega.min.js && uglifyjs build/vega-core.js -c -m -o build/vega-core.min.js",prebuild:"rm -rf build && mkdir build",prepublish:"npm run test && npm run build",rollup:"bin/rollup && bin/schema > build/vega-schema.json",test:"eslint index.js"},version:"3.1.0"}},"./node_modules/vega-loader/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-loader/src/loader.js");n.d(t,"loader",function(){return r.default});var i=n("./node_modules/vega-loader/src/read.js");n.d(t,"read",function(){return i.default});var o=n("./node_modules/vega-loader/src/type.js");n.d(t,"inferType",function(){return o.inferType}),n.d(t,"inferTypes",function(){return o.inferTypes}),n.d(t,"typeParsers",function(){return o.typeParsers});var a=n("./node_modules/vega-loader/src/formats/index.js");n.d(t,"formats",function(){return a.default})},"./node_modules/vega-loader/node_modules/topojson-client/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-loader/node_modules/topojson-client/src/bbox.js");n.d(t,"bbox",function(){return r.default});var i=n("./node_modules/vega-loader/node_modules/topojson-client/src/feature.js");n.d(t,"feature",function(){return i.default});var o=n("./node_modules/vega-loader/node_modules/topojson-client/src/mesh.js");n.d(t,"mesh",function(){return o.default}),n.d(t,"meshArcs",function(){return o.meshArcs});var a=n("./node_modules/vega-loader/node_modules/topojson-client/src/merge.js");n.d(t,"merge",function(){return a.default}),n.d(t,"mergeArcs",function(){return a.mergeArcs});var s=n("./node_modules/vega-loader/node_modules/topojson-client/src/neighbors.js");n.d(t,"neighbors",function(){return s.default});var u=n("./node_modules/vega-loader/node_modules/topojson-client/src/quantize.js");n.d(t,"quantize",function(){return u.default});var c=n("./node_modules/vega-loader/node_modules/topojson-client/src/transform.js");n.d(t,"transform",function(){return c.default});var l=n("./node_modules/vega-loader/node_modules/topojson-client/src/untransform.js");n.d(t,"untransform",function(){return l.default})},"./node_modules/vega-loader/node_modules/topojson-client/src/bbox.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-loader/node_modules/topojson-client/src/transform.js");t.default=function(e){var t,n=Object(r.default)(e.transform),i=1/0,o=i,a=-i,s=-i;function u(e){(e=n(e))[0]<i&&(i=e[0]),e[0]>a&&(a=e[0]),e[1]<o&&(o=e[1]),e[1]>s&&(s=e[1])}function c(e){switch(e.type){case"GeometryCollection":e.geometries.forEach(c);break;case"Point":u(e.coordinates);break;case"MultiPoint":e.coordinates.forEach(u)}}e.arcs.forEach(function(e){for(var t,r=-1,u=e.length;++r<u;)(t=n(e[r],r))[0]<i&&(i=t[0]),t[0]>a&&(a=t[0]),t[1]<o&&(o=t[1]),t[1]>s&&(s=t[1])});for(t in e.objects)c(e.objects[t]);return[i,o,a,s]}},"./node_modules/vega-loader/node_modules/topojson-client/src/bisect.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){for(var n=0,r=e.length;n<r;){var i=n+r>>>1;e[i]<t?n=i+1:r=i}return n}},"./node_modules/vega-loader/node_modules/topojson-client/src/feature.js":function(e,t,n){"use strict";n.r(t),n.d(t,"feature",function(){return o}),n.d(t,"object",function(){return a});var r=n("./node_modules/vega-loader/node_modules/topojson-client/src/reverse.js"),i=n("./node_modules/vega-loader/node_modules/topojson-client/src/transform.js");function o(e,t){var n=t.id,r=t.bbox,i=null==t.properties?{}:t.properties,o=a(e,t);return null==n&&null==r?{type:"Feature",properties:i,geometry:o}:null==r?{type:"Feature",id:n,properties:i,geometry:o}:{type:"Feature",id:n,bbox:r,properties:i,geometry:o}}function a(e,t){var n=Object(i.default)(e.transform),o=e.arcs;function a(e,t){t.length&&t.pop();for(var i=o[e<0?~e:e],a=0,s=i.length;a<s;++a)t.push(n(i[a],a));e<0&&Object(r.default)(t,s)}function s(e){return n(e)}function u(e){for(var t=[],n=0,r=e.length;n<r;++n)a(e[n],t);return t.length<2&&t.push(t[0]),t}function c(e){for(var t=u(e);t.length<4;)t.push(t[0]);return t}function l(e){return e.map(c)}return function e(t){var n,r=t.type;switch(r){case"GeometryCollection":return{type:r,geometries:t.geometries.map(e)};case"Point":n=s(t.coordinates);break;case"MultiPoint":n=t.coordinates.map(s);break;case"LineString":n=u(t.arcs);break;case"MultiLineString":n=t.arcs.map(u);break;case"Polygon":n=l(t.arcs);break;case"MultiPolygon":n=t.arcs.map(l);break;default:return null}return{type:r,coordinates:n}}(t)}t.default=function(e,t){return"GeometryCollection"===t.type?{type:"FeatureCollection",features:t.geometries.map(function(t){return o(e,t)})}:o(e,t)}},"./node_modules/vega-loader/node_modules/topojson-client/src/identity.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return e}},"./node_modules/vega-loader/node_modules/topojson-client/src/merge.js":function(e,t,n){"use strict";n.r(t),n.d(t,"mergeArcs",function(){return o});var r=n("./node_modules/vega-loader/node_modules/topojson-client/src/feature.js"),i=n("./node_modules/vega-loader/node_modules/topojson-client/src/stitch.js");function o(e,t){var n={},o=[],a=[];function s(e){e.forEach(function(t){t.forEach(function(t){(n[t=t<0?~t:t]||(n[t]=[])).push(e)})}),o.push(e)}function u(t){return function(e){for(var t,n=-1,r=e.length,i=e[r-1],o=0;++n<r;)t=i,i=e[n],o+=t[0]*i[1]-t[1]*i[0];return Math.abs(o)}(Object(r.object)(e,{type:"Polygon",arcs:[t]}).coordinates[0])}return t.forEach(function e(t){switch(t.type){case"GeometryCollection":t.geometries.forEach(e);break;case"Polygon":s(t.arcs);break;case"MultiPolygon":t.arcs.forEach(s)}}),o.forEach(function(e){if(!e._){var t=[],r=[e];for(e._=1,a.push(t);e=r.pop();)t.push(e),e.forEach(function(e){e.forEach(function(e){n[e<0?~e:e].forEach(function(e){e._||(e._=1,r.push(e))})})})}}),o.forEach(function(e){delete e._}),{type:"MultiPolygon",arcs:a.map(function(t){var r,o=[];if(t.forEach(function(e){e.forEach(function(e){e.forEach(function(e){n[e<0?~e:e].length<2&&o.push(e)})})}),(r=(o=Object(i.default)(e,o)).length)>1)for(var a,s,c=1,l=u(o[0]);c<r;++c)(a=u(o[c]))>l&&(s=o[0],o[0]=o[c],o[c]=s,l=a);return o})}}t.default=function(e){return Object(r.object)(e,o.apply(this,arguments))}},"./node_modules/vega-loader/node_modules/topojson-client/src/mesh.js":function(e,t,n){"use strict";n.r(t),n.d(t,"meshArcs",function(){return o});var r=n("./node_modules/vega-loader/node_modules/topojson-client/src/feature.js"),i=n("./node_modules/vega-loader/node_modules/topojson-client/src/stitch.js");function o(e,t,n){var r,o,a;if(arguments.length>1)r=function(e,t,n){var r,i=[],o=[];function a(e){var t=e<0?~e:e;(o[t]||(o[t]=[])).push({i:e,g:r})}function s(e){e.forEach(a)}function u(e){e.forEach(s)}return function e(t){switch(r=t,t.type){case"GeometryCollection":t.geometries.forEach(e);break;case"LineString":s(t.arcs);break;case"MultiLineString":case"Polygon":u(t.arcs);break;case"MultiPolygon":t.arcs.forEach(u)}}(t),o.forEach(null==n?function(e){i.push(e[0].i)}:function(e){n(e[0].g,e[e.length-1].g)&&i.push(e[0].i)}),i}(0,t,n);else for(o=0,r=new Array(a=e.arcs.length);o<a;++o)r[o]=o;return{type:"MultiLineString",arcs:Object(i.default)(e,r)}}t.default=function(e){return Object(r.object)(e,o.apply(this,arguments))}},"./node_modules/vega-loader/node_modules/topojson-client/src/neighbors.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-loader/node_modules/topojson-client/src/bisect.js");t.default=function(e){var t={},n=e.map(function(){return[]});function i(e,n){e.forEach(function(e){e<0&&(e=~e);var r=t[e];r?r.push(n):t[e]=[n]})}function o(e,t){e.forEach(function(e){i(e,t)})}var a={LineString:i,MultiLineString:o,Polygon:o,MultiPolygon:function(e,t){e.forEach(function(e){o(e,t)})}};e.forEach(function e(t,n){"GeometryCollection"===t.type?t.geometries.forEach(function(t){e(t,n)}):t.type in a&&a[t.type](t.arcs,n)});for(var s in t)for(var u=t[s],c=u.length,l=0;l<c;++l)for(var d=l+1;d<c;++d){var p,f=u[l],h=u[d];(p=n[f])[s=Object(r.default)(p,h)]!==h&&p.splice(s,0,h),(p=n[h])[s=Object(r.default)(p,f)]!==f&&p.splice(s,0,f)}return n}},"./node_modules/vega-loader/node_modules/topojson-client/src/quantize.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-loader/node_modules/topojson-client/src/bbox.js"),i=n("./node_modules/vega-loader/node_modules/topojson-client/src/untransform.js");t.default=function(e,t){if(e.transform)throw new Error("already quantized");if(t&&t.scale)c=e.bbox;else{if(!((n=Math.floor(t))>=2))throw new Error("n must be ≥2");var n,o=(c=e.bbox||Object(r.default)(e))[0],a=c[1],s=c[2],u=c[3];t={scale:[s-o?(s-o)/(n-1):1,u-a?(u-a)/(n-1):1],translate:[o,a]}}var c,l,d=Object(i.default)(t),p=e.objects,f={};function h(e){return d(e)}function m(e){var t;switch(e.type){case"GeometryCollection":t={type:"GeometryCollection",geometries:e.geometries.map(m)};break;case"Point":t={type:"Point",coordinates:h(e.coordinates)};break;case"MultiPoint":t={type:"MultiPoint",coordinates:e.coordinates.map(h)};break;default:return e}return null!=e.id&&(t.id=e.id),null!=e.bbox&&(t.bbox=e.bbox),null!=e.properties&&(t.properties=e.properties),t}for(l in p)f[l]=m(p[l]);return{type:"Topology",bbox:c,transform:t,objects:f,arcs:e.arcs.map(function(e){var t,n=0,r=1,i=e.length,o=new Array(i);for(o[0]=d(e[0],0);++n<i;)((t=d(e[n],n))[0]||t[1])&&(o[r++]=t);return 1===r&&(o[r++]=[0,0]),o.length=r,o})}}},"./node_modules/vega-loader/node_modules/topojson-client/src/reverse.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){for(var n,r=e.length,i=r-t;i<--r;)n=e[i],e[i++]=e[r],e[r]=n}},"./node_modules/vega-loader/node_modules/topojson-client/src/stitch.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){var n={},r={},i={},o=[],a=-1;function s(e,t){for(var r in e){var i=e[r];delete t[i.start],delete i.start,delete i.end,i.forEach(function(e){n[e<0?~e:e]=1}),o.push(i)}}return t.forEach(function(n,r){var i,o=e.arcs[n<0?~n:n];o.length<3&&!o[1][0]&&!o[1][1]&&(i=t[++a],t[a]=n,t[r]=i)}),t.forEach(function(t){var n,o,a=function(t){var n,r=e.arcs[t<0?~t:t],i=r[0];e.transform?(n=[0,0],r.forEach(function(e){n[0]+=e[0],n[1]+=e[1]})):n=r[r.length-1];return t<0?[n,i]:[i,n]}(t),s=a[0],u=a[1];if(n=i[s])if(delete i[n.end],n.push(t),n.end=u,o=r[u]){delete r[o.start];var c=o===n?n:n.concat(o);r[c.start=n.start]=i[c.end=o.end]=c}else r[n.start]=i[n.end]=n;else if(n=r[u])if(delete r[n.start],n.unshift(t),n.start=s,o=i[s]){delete i[o.end];var l=o===n?n:o.concat(n);r[l.start=o.start]=i[l.end=n.end]=l}else r[n.start]=i[n.end]=n;else r[(n=[t]).start=s]=i[n.end=u]=n}),s(i,r),s(r,i),t.forEach(function(e){n[e<0?~e:e]||o.push([e])}),o}},"./node_modules/vega-loader/node_modules/topojson-client/src/transform.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-loader/node_modules/topojson-client/src/identity.js");t.default=function(e){if(null==e)return r.default;var t,n,i=e.scale[0],o=e.scale[1],a=e.translate[0],s=e.translate[1];return function(e,r){r||(t=n=0);var u=2,c=e.length,l=new Array(c);for(l[0]=(t+=e[0])*i+a,l[1]=(n+=e[1])*o+s;u<c;)l[u]=e[u],++u;return l}}},"./node_modules/vega-loader/node_modules/topojson-client/src/untransform.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-loader/node_modules/topojson-client/src/identity.js");t.default=function(e){if(null==e)return r.default;var t,n,i=e.scale[0],o=e.scale[1],a=e.translate[0],s=e.translate[1];return function(e,r){r||(t=n=0);var u=2,c=e.length,l=new Array(c),d=Math.round((e[0]-a)/i),p=Math.round((e[1]-s)/o);for(l[0]=d-t,t=d,l[1]=p-n,n=p;u<c;)l[u]=e[u],++u;return l}}},"./node_modules/vega-loader/src/formats/dsv.js":function(e,t,n){"use strict";n.r(t),n.d(t,"delimitedFormat",function(){return o}),n.d(t,"default",function(){return a});var r=n("./node_modules/d3-dsv/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){return function(t,n){var r={delimiter:e};return a(t,n?Object(i.extend)(n,r):r)}}function a(e,t){return t.header&&(e=t.header.map(i.stringValue).join(t.delimiter)+"\n"+e),Object(r.dsvFormat)(t.delimiter).parse(e+"")}},"./node_modules/vega-loader/src/formats/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-loader/src/formats/dsv.js"),i=n("./node_modules/vega-loader/src/formats/json.js"),o=n("./node_modules/vega-loader/src/formats/topojson.js"),a={dsv:r.default,csv:Object(r.delimitedFormat)(","),tsv:Object(r.delimitedFormat)("\t"),json:i.default,topojson:o.default};t.default=function(e,t){return arguments.length>1?(a[e]=t,this):a.hasOwnProperty(e)?a[e]:null}},"./node_modules/vega-loader/src/formats/json.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/index.js");t.default=function(e,t){var n,i,o,a=t&&t.property?Object(r.field)(t.property):r.identity;return!Object(r.isObject)(e)||(o=e,"function"==typeof Buffer&&Object(r.isFunction)(Buffer.isBuffer)&&Buffer.isBuffer(o))?a(JSON.parse(e)):(n=a(e),i&&i.copy?JSON.parse(JSON.stringify(n)):n)}},"./node_modules/vega-loader/src/formats/topojson.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-loader/src/formats/json.js"),i=n("./node_modules/vega-loader/node_modules/topojson-client/index.js"),o=n("./node_modules/vega-util/index.js");t.default=function(e,t){var n,a,s;return e=Object(r.default)(e,t),n=t&&(s=t.feature)?i.feature:t&&(s=t.mesh)?i.mesh:Object(o.error)("Missing TopoJSON feature or mesh parameter."),(a=(a=e.objects[s])?n(e,a):Object(o.error)("Invalid TopoJSON object: "+s))&&a.features||[a]}},"./node_modules/vega-loader/src/loader.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/index.js"),i=n("./node_modules/d3-request/index.js"),o=/^([A-Za-z]+:)?\/\//,a="file://",s=["mimeType","responseType","user","password"];function u(e,t){return Object(r.extend)({},e.options,t)}function c(e,t){var n=this;return n.sanitize(e,t).then(function(e){var r=e.href;return e.localFile?n.file(r):n.http(r,t)})}function l(e,t){return t=u(this,t),new Promise(function(n,i){var s,u,c,l,d={href:null};null!=e&&"string"==typeof e?(u=o.test(e),(l=t.baseURL)&&!u&&(h(e,"/")||"/"===l[l.length-1]||(e="/"+e),e=l+e),c=(s=h(e,a))||"file"===t.mode||"http"!==t.mode&&!u&&f(),s?e=e.slice(a.length):h(e,"//")&&("file"===t.defaultProtocol?(e=e.slice(2),c=!0):e=(t.defaultProtocol||"http")+":"+e),Object.defineProperty(d,"localFile",{value:!!c}),d.href=e,n(d)):i("Sanitize failure, invalid URI: "+Object(r.stringValue)(e))})}function d(e,t){return t=u(this,t),new Promise(function(n,r){var o,a=Object(i.request)(e);for(o in t.headers)a.header(o,t.headers[o]);s.forEach(function(e){t[e]&&a[e](t[e])}),a.on("error",function(t){r(t||"Error loading URL: "+e)}).on("load",function(e){var t=e&&e.responseText;e&&0!==e.status?n(t):r(t||"Error")}).get()})}function p(e){return new Promise(function(t,n){var r=f();r?r.readFile(e,function(e,r){e?n(e):t(r)}):n("No file system access for "+e)})}function f(){var e=n(1);return e&&Object(r.isFunction)(e.readFile)?e:null}function h(e,t){return null!=e&&0===e.lastIndexOf(t,0)}t.default=function(e){return{options:e||{},sanitize:l,load:c,file:p,http:d}}},"./node_modules/vega-loader/src/read.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-loader/src/type.js"),i=n("./node_modules/vega-loader/src/formats/index.js"),o=n("./node_modules/vega-util/index.js"),a=n("./node_modules/d3-time-format/index.js");t.default=function(e,t,n){t=t||{};var s=Object(i.default)(t.type||"json");return s||Object(o.error)("Unknown data format type: "+t.type),e=s(e,t),t.parse&&function(e,t,n){if(!e.length)return;n=n||a.timeParse;var i,o,s,u,c,l,d,p=e.columns||Object.keys(e[0]);"auto"===t&&(t=Object(r.inferTypes)(e,p));for(p=Object.keys(t),i=p.map(function(e){var i,o,s=t[e];if(s&&(0===s.indexOf("date:")||0===s.indexOf("utc:")))return("'"===(o=(i=s.split(/:(.+)?/,2))[1])[0]&&"'"===o[o.length-1]||'"'===o[0]&&'"'===o[o.length-1])&&(o=o.slice(1,-1)),"utc"===i[0]?Object(a.utcParse)(o):n(o);if(!r.typeParsers[s])throw Error("Illegal format pattern: "+e+":"+s);return r.typeParsers[s]}),u=0,l=e.length,d=p.length;u<l;++u)for(o=e[u],c=0;c<d;++c)s=p[c],o[s]=i[c](o[s])}(e,t.parse,n),e.hasOwnProperty("columns")&&delete e.columns,e}},"./node_modules/vega-loader/src/type.js":function(e,t,n){"use strict";n.r(t),n.d(t,"typeParsers",function(){return i}),n.d(t,"inferType",function(){return s}),n.d(t,"inferTypes",function(){return u});var r=n("./node_modules/vega-util/index.js"),i={boolean:r.toBoolean,integer:r.toNumber,number:r.toNumber,date:r.toDate,string:r.toString,unknown:r.identity},o=[function(e){return"true"===e||"false"===e||!0===e||!1===e},function(e){return c(e)&&(e=+e)==~~e},c,function(e){return!isNaN(Date.parse(e))}],a=["boolean","integer","number","date"];function s(e,t){if(!e||!e.length)return"unknown";var n,r,i,s,u,c=o.slice();for(r=0,i=e.length;r<i;++r){for(n=t?e[r][t]:e[r],s=0;s<c.length;++s)null==(u=n)||u!=u||c[s](n)||(c.splice(s,1),--s);if(0===c.length)return"string"}return a[o.indexOf(c[0])]}function u(e,t){return t.reduce(function(t,n){return t[n]=s(e,n),t},{})}function c(e){return!(isNaN(+e)||e instanceof Date)}},"./node_modules/vega-parser/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parse.js");n.d(t,"parse",function(){return r.default});var i=n("./node_modules/vega-parser/src/config.js");n.d(t,"config",function(){return i.default});var o=n("./node_modules/vega-parser/src/parsers/signal.js");n.d(t,"signal",function(){return o.default});var a=n("./node_modules/vega-parser/src/parsers/signal-updates.js");n.d(t,"signalUpdates",function(){return a.default});var s=n("./node_modules/vega-parser/src/parsers/stream.js");n.d(t,"stream",function(){return s.default});var u=n("./node_modules/vega-parser/src/parsers/expression/codegen.js");n.d(t,"codeGenerator",function(){return u.codeGenerator}),n.d(t,"functionContext",function(){return u.functionContext}),n.d(t,"expressionFunction",function(){return u.expressionFunction});var c=n("./node_modules/vega-parser/src/parsers/marks/roles.js");n.d(t,"MarkRole",function(){return c.MarkRole}),n.d(t,"FrameRole",function(){return c.FrameRole}),n.d(t,"ScopeRole",function(){return c.ScopeRole}),n.d(t,"AxisRole",function(){return c.AxisRole}),n.d(t,"AxisDomainRole",function(){return c.AxisDomainRole}),n.d(t,"AxisGridRole",function(){return c.AxisGridRole}),n.d(t,"AxisLabelRole",function(){return c.AxisLabelRole}),n.d(t,"AxisTickRole",function(){return c.AxisTickRole}),n.d(t,"AxisTitleRole",function(){return c.AxisTitleRole}),n.d(t,"LegendRole",function(){return c.LegendRole}),n.d(t,"LegendEntryRole",function(){return c.LegendEntryRole}),n.d(t,"LegendLabelRole",function(){return c.LegendLabelRole}),n.d(t,"LegendSymbolRole",function(){return c.LegendSymbolRole}),n.d(t,"LegendTitleRole",function(){return c.LegendTitleRole});var l=n("./node_modules/vega-parser/src/Scope.js");n.d(t,"Scope",function(){return l.default});var d=n("./node_modules/vega-parser/src/DataScope.js");n.d(t,"DataScope",function(){return d.default});var p=n("./node_modules/d3-format/index.js");n.d(t,"formatLocale",function(){return p.formatDefaultLocale});var f=n("./node_modules/d3-time-format/index.js");n.d(t,"timeFormatLocale",function(){return f.timeFormatDefaultLocale})},"./node_modules/vega-parser/src/DataScope.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-parser/src/util.js"),i=n("./node_modules/vega-parser/src/transforms.js"),o=n("./node_modules/vega-util/index.js");function a(e,t,n,r,i){this.scope=e,this.input=t,this.output=n,this.values=r,this.aggregate=i,this.index={}}a.fromEntries=function(e,t){var n=t.length,i=1,o=t[0],s=t[n-1],u=t[n-2],c=null;for(e.add(t[0]);i<n;++i)t[i].params.pulse=Object(r.ref)(t[i-1]),e.add(t[i]),"aggregate"===t[i].type&&(c=t[i]);return new a(e,o,u,s,c)};var s=a.prototype;function u(e){return Object(o.isString)(e)?e:null}function c(e,t,n){var i,o=Object(r.aggrField)(n.op,n.field);if(t.ops){for(var a=0,s=t.as.length;a<s;++a)if(t.as[a]===o)return}else t.ops=["count"],t.fields=[null],t.as=["count"];n.op&&(t.ops.push((i=n.op.signal)?e.signalRef(i):n.op),t.fields.push(e.fieldRef(n.field)),t.as.push(o))}function l(e,t,n,i,o,a,s){var c,l,d=t[n]||(t[n]={}),p=Object(r.sortKey)(a),f=u(o);if(null!=f&&(e=t.scope,c=d[f+=p?"|"+p:""]),!c){var h=a?{field:r.keyFieldRef,pulse:t.countsRef(e,o,a)}:{field:e.fieldRef(o),pulse:Object(r.ref)(t.output)};p&&(h.sort=e.sortRef(a)),l=e.add(Object(r.entry)(i,void 0,h)),s&&(t.index[o]=l),c=Object(r.ref)(l),null!=f&&(d[f]=c)}return c}s.countsRef=function(e,t,n){var o,a,s,l=this.counts||(this.counts={}),d=u(t);return null!=d&&(e=this.scope,o=l[d]),o?n&&n.field&&c(e,o.agg.params,n):(s={groupby:e.fieldRef(t,"key"),pulse:Object(r.ref)(this.output)},n&&n.field&&c(e,s,n),a=e.add(Object(i.Aggregate)(s)),o=e.add(Object(i.Collect)({pulse:Object(r.ref)(a)})),o={agg:a,ref:Object(r.ref)(o)},null!=d&&(l[d]=o)),o.ref},s.tuplesRef=function(){return Object(r.ref)(this.values)},s.extentRef=function(e,t){return l(e,this,"extent","extent",t,!1)},s.domainRef=function(e,t){return l(e,this,"domain","values",t,!1)},s.valuesRef=function(e,t,n){return l(e,this,"vals","values",t,n||!0)},s.lookupRef=function(e,t){return l(e,this,"lookup","tupleindex",t,!1)},s.indataRef=function(e,t){return l(e,this,"indata","tupleindex",t,!0,!0)}},"./node_modules/vega-parser/src/Scope.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return u});var r=n("./node_modules/vega-parser/src/DataScope.js"),i=n("./node_modules/vega-parser/src/util.js"),o=n("./node_modules/vega-parser/src/parsers/expression.js"),a=n("./node_modules/vega-parser/src/transforms.js"),s=n("./node_modules/vega-util/index.js");function u(e){this.config=e,this.bindings=[],this.field={},this.signals={},this.lambdas={},this.scales={},this.events={},this.data={},this.streams=[],this.updates=[],this.operators=[],this.background=null,this.eventConfig=null,this._id=0,this._subid=0,this._nextsub=[0],this._parent=[],this._encode=[],this._lookup=[],this._markpath=[]}function c(e){this.config=e.config,this.field=Object.create(e.field),this.signals=Object.create(e.signals),this.lambdas=Object.create(e.lambdas),this.scales=Object.create(e.scales),this.events=Object.create(e.events),this.data=Object.create(e.data),this.streams=[],this.updates=[],this.operators=[],this._id=0,this._subid=++e._nextsub[0],this._nextsub=e._nextsub,this._parent=e._parent.slice(),this._encode=e._encode.slice(),this._lookup=e._lookup.slice(),this._markpath=e._markpath}var l=u.prototype=c.prototype;function d(e){return(Object(s.isArray)(e)?function(e){for(var t,n="[",r=0,i=e.length;r<i;++r)t=e[r],n+=(r>0?",":"")+(Object(s.isObject)(t)?t.signal||d(t):Object(s.stringValue)(t));return n+"]"}:function(e){var t,n,r="{",i=0;for(t in e)n=e[t],r+=(++i>1?",":"")+Object(s.stringValue)(t)+":"+(Object(s.isObject)(n)?n.signal||d(n):Object(s.stringValue)(n));return r+"}"})(e)}l.fork=function(){return new c(this)},l.toRuntime=function(){return this.finish(),{background:this.background,operators:this.operators,streams:this.streams,updates:this.updates,bindings:this.bindings,eventConfig:this.eventConfig}},l.id=function(){return(this._subid?this._subid+":":0)+this._id++},l.add=function(e){return this.operators.push(e),e.id=this.id(),e.refs&&(e.refs.forEach(function(t){t.$ref=e.id}),e.refs=null),e},l.proxy=function(e){var t=e instanceof i.Entry?Object(i.ref)(e):e;return this.add(Object(a.Proxy)({value:t}))},l.addStream=function(e){return this.streams.push(e),e.id=this.id(),e},l.addUpdate=function(e){return this.updates.push(e),e},l.finish=function(){var e,t;this.root&&(this.root.root=!0);for(e in this.signals)this.signals[e].signal=e;for(e in this.scales)this.scales[e].scale=e;function n(e,t,n){var r;e&&((r=e.data||(e.data={}))[t]||(r[t]=[])).push(n)}for(e in this.data){n((t=this.data[e]).input,e,"input"),n(t.output,e,"output"),n(t.values,e,"values");for(var r in t.index)n(t.index[r],e,"index:"+r)}return this},l.pushState=function(e,t,n){this._encode.push(Object(i.ref)(this.add(Object(a.Sieve)({pulse:e})))),this._parent.push(t),this._lookup.push(n?Object(i.ref)(this.proxy(n)):null),this._markpath.push(-1)},l.popState=function(){this._encode.pop(),this._parent.pop(),this._lookup.pop(),this._markpath.pop()},l.parent=function(){return Object(s.peek)(this._parent)},l.encode=function(){return Object(s.peek)(this._encode)},l.lookup=function(){return Object(s.peek)(this._lookup)},l.markpath=function(){var e=this._markpath;return++e[e.length-1]},l.fieldRef=function(e,t){if(Object(s.isString)(e))return Object(i.fieldRef)(e,t);e.signal||Object(s.error)("Unsupported field reference: "+Object(s.stringValue)(e));var n,r=e.signal,o=this.field[r];return o||(n={name:this.signalRef(r)},t&&(n.as=t),this.field[r]=o=Object(i.ref)(this.add(Object(a.Field)(n)))),o},l.compareRef=function(e,t){function n(e){return Object(i.isSignal)(e)?(o=!0,Object(i.ref)(r[e.signal])):e}var r=this.signals,o=!1,u=Object(s.array)(e.field).map(n),c=Object(s.array)(e.order).map(n);return t&&u.push(i.tupleidRef),o?Object(i.ref)(this.add(Object(a.Compare)({fields:u,orders:c}))):Object(i.compareRef)(u,c)},l.keyRef=function(e,t){var n=this.signals,r=!1;return e=Object(s.array)(e).map(function(e){return Object(i.isSignal)(e)?(r=!0,Object(i.ref)(n[e.signal])):e}),r?Object(i.ref)(this.add(Object(a.Key)({fields:e,flat:t}))):Object(i.keyRef)(e,t)},l.sortRef=function(e){if(!e)return e;var t=[Object(i.aggrField)(e.op,e.field),i.tupleidRef],n=e.order||i.Ascending;return n.signal?Object(i.ref)(this.add(Object(a.Compare)({fields:t,orders:[n=this.signalRef(n.signal),n]}))):Object(i.compareRef)(t,[n,n])},l.event=function(e,t){var n=e+":"+t;if(!this.events[n]){var r=this.id();this.streams.push({id:r,source:e,type:t}),this.events[n]=r}return this.events[n]},l.addSignal=function(e,t){this.signals.hasOwnProperty(e)&&Object(s.error)("Duplicate signal name: "+Object(s.stringValue)(e));var n=t instanceof i.Entry?t:this.add(Object(i.operator)(t));return this.signals[e]=n},l.getSignal=function(e){return this.signals[e]||Object(s.error)("Unrecognized signal name: "+Object(s.stringValue)(e)),this.signals[e]},l.signalRef=function(e){return this.signals[e]?Object(i.ref)(this.signals[e]):(this.lambdas.hasOwnProperty(e)||(this.lambdas[e]=this.add(Object(i.operator)(null))),Object(i.ref)(this.lambdas[e]))},l.parseLambdas=function(){for(var e=Object.keys(this.lambdas),t=0,n=e.length;t<n;++t){var r=e[t],i=Object(o.default)(r,this),a=this.lambdas[r];a.params=i.$params,a.update=i.$expr}},l.property=function(e){return e&&e.signal?this.signalRef(e.signal):e},l.objectProperty=function(e){return e&&Object(s.isObject)(e)?this.signalRef(e.signal||d(e)):e},l.addBinding=function(e,t){this.bindings||Object(s.error)("Nested signals do not support binding: "+Object(s.stringValue)(e)),this.bindings.push(Object(s.extend)({signal:e},t))},l.addScaleProj=function(e,t){this.scales.hasOwnProperty(e)&&Object(s.error)("Duplicate scale or projection name: "+Object(s.stringValue)(e)),this.scales[e]=this.add(t)},l.addScale=function(e,t){this.addScaleProj(e,Object(a.Scale)(t))},l.addProjection=function(e,t){this.addScaleProj(e,Object(a.Projection)(t))},l.getScale=function(e){return this.scales[e]||Object(s.error)("Unrecognized scale name: "+Object(s.stringValue)(e)),this.scales[e]},l.projectionRef=l.scaleRef=function(e){return Object(i.ref)(this.getScale(e))},l.projectionType=l.scaleType=function(e){return this.getScale(e).params.type},l.addData=function(e,t){return this.data.hasOwnProperty(e)&&Object(s.error)("Duplicate data set name: "+Object(s.stringValue)(e)),this.data[e]=t},l.getData=function(e){return this.data[e]||Object(s.error)("Undefined data set name: "+Object(s.stringValue)(e)),this.data[e]},l.addDataPipeline=function(e,t){return this.data.hasOwnProperty(e)&&Object(s.error)("Duplicate data set name: "+Object(s.stringValue)(e)),this.addData(e,r.default.fromEntries(this,t))}},"./node_modules/vega-parser/src/config.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/index.js");t.default=function(e){var t={padding:0,autosize:"pad",background:null,events:{defaults:{allow:["wheel"]}},group:null,mark:null,arc:{fill:s},area:{fill:s},image:null,line:{stroke:s,strokeWidth:a},path:{stroke:s},rect:{fill:s},rule:{stroke:u},shape:{stroke:s},symbol:{fill:s,size:64},text:{fill:u,font:i,fontSize:11},style:{"guide-label":{fill:u,font:i,fontSize:10},"guide-title":{fill:u,font:i,fontSize:11,fontWeight:"bold"},"group-title":{fill:u,font:i,fontSize:13,fontWeight:"bold"},point:{size:o,strokeWidth:a,shape:"circle"},circle:{size:o,strokeWidth:a},square:{size:o,strokeWidth:a,shape:"square"},cell:{fill:"transparent",stroke:l}},axis:{minExtent:0,maxExtent:200,bandPosition:.5,domain:!0,domainWidth:1,domainColor:c,grid:!1,gridWidth:1,gridColor:l,gridOpacity:1,labels:!0,labelAngle:0,labelLimit:180,labelPadding:2,ticks:!0,tickColor:c,tickOffset:0,tickRound:!0,tickSize:5,tickWidth:1,titleAlign:"center",titlePadding:4},axisBand:{tickOffset:-1},legend:{orient:"right",offset:18,padding:0,entryPadding:5,titlePadding:5,gradientWidth:100,gradientHeight:20,gradientStrokeColor:l,gradientStrokeWidth:0,gradientLabelBaseline:"top",gradientLabelOffset:2,labelAlign:"left",labelBaseline:"middle",labelOffset:8,labelLimit:160,symbolType:"circle",symbolSize:100,symbolFillColor:"transparent",symbolStrokeColor:c,symbolStrokeWidth:1.5,titleAlign:"left",titleBaseline:"top",titleLimit:180},title:{orient:"top",anchor:"middle",offset:4},range:{category:{scheme:"tableau10"},ordinal:{scheme:"blues",extent:[.2,1]},heatmap:{scheme:"viridis"},ramp:{scheme:"blues",extent:[.2,1]},diverging:{scheme:"blueorange"},symbol:["circle","square","triangle-up","cross","diamond","triangle-right","triangle-down","triangle-left"]}};return(e||[]).forEach(function(e){var n,i,o;if(e)for(n in e)if("style"===n){o=t.style||(t.style={});for(n in e.style)o[n]=Object(r.extend)(o[n]||{},e.style[n])}else i=e[n],t[n]=Object(r.isObject)(i)&&!Object(r.isArray)(i)?Object(r.extend)(Object(r.isObject)(t[n])?t[n]:{},i):i}),t};var i="sans-serif",o=30,a=2,s="#4c78a8",u="#000",c="#888",l="#ddd"},"./node_modules/vega-parser/src/parse.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/view.js"),i=n("./node_modules/vega-parser/src/Scope.js"),o=n("./node_modules/vega-parser/src/config.js"),a=n("./node_modules/vega-util/index.js");t.default=function(e,t){return Object(a.isObject)(e)||Object(a.error)("Input Vega specification must be an object."),Object(r.default)(e,new i.default(Object(o.default)([t,e.config]))).toRuntime()}},"./node_modules/vega-parser/src/parsers/autosize.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/index.js");t.default=function(e,t){return e=e||t.autosize,Object(r.isObject)(e)?e:{type:e=e||"pad"}}},"./node_modules/vega-parser/src/parsers/axis.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/guides/axis-config.js"),i=n("./node_modules/vega-parser/src/parsers/guides/axis-domain.js"),o=n("./node_modules/vega-parser/src/parsers/guides/axis-grid.js"),a=n("./node_modules/vega-parser/src/parsers/guides/axis-ticks.js"),s=n("./node_modules/vega-parser/src/parsers/guides/axis-labels.js"),u=n("./node_modules/vega-parser/src/parsers/guides/axis-title.js"),c=n("./node_modules/vega-parser/src/parsers/guides/guide-group.js"),l=n("./node_modules/vega-parser/src/parsers/marks/roles.js"),d=n("./node_modules/vega-parser/src/parsers/mark.js"),p=n("./node_modules/vega-parser/src/parsers/encode/encode-util.js"),f=n("./node_modules/vega-parser/src/parsers/guides/constants.js"),h=n("./node_modules/vega-parser/src/util.js"),m=n("./node_modules/vega-parser/src/transforms.js");t.default=function(e,t){var n,g,v,_,y,b,j=Object(r.default)(e,t),x=e.encode||{},O=x.axis||{},w=O.name||void 0,C=O.interactive,E=O.style;return n={orient:e.orient,ticks:!!Object(h.value)(e.ticks,j.ticks),labels:!!Object(h.value)(e.labels,j.labels),grid:!!Object(h.value)(e.grid,j.grid),domain:!!Object(h.value)(e.domain,j.domain),title:!!Object(h.value)(e.title,!1)},g=Object(h.ref)(t.add(Object(m.Collect)({},[n]))),O=Object(p.extendEncode)({update:{range:{signal:'abs(span(range("'+e.scale+'")))'},offset:Object(p.encoder)(Object(h.value)(e.offset,0)),position:Object(p.encoder)(Object(h.value)(e.position,0)),titlePadding:Object(p.encoder)(Object(h.value)(e.titlePadding,j.titlePadding)),minExtent:Object(p.encoder)(Object(h.value)(e.minExtent,j.minExtent)),maxExtent:Object(p.encoder)(Object(h.value)(e.maxExtent,j.maxExtent))}},x.axis,f.Skip),v=Object(h.ref)(t.add(Object(m.AxisTicks)({scale:t.scaleRef(e.scale),extra:j.tickExtra,count:t.objectProperty(e.tickCount),values:t.objectProperty(e.values),formatSpecifier:t.property(e.format)}))),b=[],n.grid&&b.push(Object(o.default)(e,j,x.grid,v)),n.ticks&&(_=Object(h.value)(e.tickSize,j.tickSize),b.push(Object(a.default)(e,j,x.ticks,v,_))),n.labels&&(_=n.ticks?_:0,b.push(Object(s.default)(e,j,x.labels,v,_))),n.domain&&b.push(Object(i.default)(e,j,x.domain,g)),n.title&&b.push(Object(u.default)(e,j,x.title,g)),y=Object(c.default)(l.AxisRole,E,w,g,C,O,b),e.zindex&&(y.zindex=e.zindex),Object(d.default)(y,t)}},"./node_modules/vega-parser/src/parsers/data.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return u});var r=n("./node_modules/vega-parser/src/parsers/transform.js"),i=n("./node_modules/vega-parser/src/parsers/trigger.js"),o=n("./node_modules/vega-parser/src/util.js"),a=n("./node_modules/vega-parser/src/transforms.js"),s=n("./node_modules/vega-util/index.js");function u(e,t){var n=[];e.transform&&e.transform.forEach(function(e){n.push(Object(r.default)(e,t))}),e.on&&e.on.forEach(function(n){Object(i.default)(n,t,e.name)}),t.addDataPipeline(e.name,function(e,t,n){var r,i,u,l,d,p=[],f=null,h=!1,m=!1;e.values?p.push(f=c({$ingest:e.values,$format:e.format})):e.url?p.push(f=c({$request:e.url,$format:e.format})):e.source&&(f=r=Object(s.array)(e.source).map(function(e){return Object(o.ref)(t.getData(e).output)}),p.push(null));for(i=0,u=n.length;i<u;++i)l=n[i],d=l.metadata,f||d.source||p.push(f=c()),p.push(l),d.generates&&(m=!0),d.modifies&&!m&&(h=!0),d.source?f=l:d.changes&&(f=null);r&&(u=r.length-1,p[0]=Object(a.Relay)({derive:h,pulse:u?r:r[0]}),(h||u)&&p.splice(1,0,c()));f||p.push(c());return p.push(Object(a.Sieve)({})),p}(e,t,n))}function c(e){var t=Object(a.Collect)({},e);return t.metadata={source:!0},t}},"./node_modules/vega-parser/src/parsers/encode.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return u});var r=n("./node_modules/vega-parser/src/parsers/encode/adjust-spatial.js"),i=n("./node_modules/vega-parser/src/parsers/encode/entry.js"),o=n("./node_modules/vega-parser/src/parsers/encode/rule.js"),a=n("./node_modules/vega-parser/src/parsers/encode/set.js"),s=n("./node_modules/vega-util/index.js");function u(e,t,n,u){var c,l,d,p={},f="var o=item,datum=o.datum,$;";for(c in e)l=e[c],Object(s.isArray)(l)?f+=Object(o.default)(c,l,u,n,p):(d=Object(i.default)(c,l,u,n,p),f+=Object(a.default)("o",c,d));return f+=Object(r.default)(e,t),{$expr:f+="return 1;",$fields:Object.keys(p),$output:Object.keys(e)}}},"./node_modules/vega-parser/src/parsers/encode/adjust-spatial.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/index.js"),i=Object(r.toSet)(["rule"]),o=Object(r.toSet)(["group","image","rect"]);t.default=function(e,t){var n="";return i[t]?n:(e.x2&&(e.x?(o[t]&&(n+="if(o.x>o.x2)$=o.x,o.x=o.x2,o.x2=$;"),n+="o.width=o.x2-o.x;"):n+="o.x=o.x2-(o.width||0);"),e.xc&&(n+="o.x=o.xc-(o.width||0)/2;"),e.y2&&(e.y?(o[t]&&(n+="if(o.y>o.y2)$=o.y,o.y=o.y2,o.y2=$;"),n+="o.height=o.y2-o.y;"):n+="o.y=o.y2-(o.height||0);"),e.yc&&(n+="o.y=o.yc-(o.height||0)/2;"),n)}},"./node_modules/vega-parser/src/parsers/encode/color.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/encode/entry.js");t.default=function(e,t,n,i){function o(e,o,a,s){return"this."+e+"("+[Object(r.default)(null,o,t,n,i),Object(r.default)(null,a,t,n,i),Object(r.default)(null,s,t,n,i)].join(",")+").toString()"}return e.c?o("hcl",e.h,e.c,e.l):e.h||e.s?o("hsl",e.h,e.s,e.l):e.l||e.a?o("lab",e.l,e.a,e.b):e.r||e.g||e.b?o("rgb",e.r,e.g,e.b):null}},"./node_modules/vega-parser/src/parsers/encode/encode-util.js":function(e,t,n){"use strict";n.r(t),n.d(t,"encoder",function(){return a}),n.d(t,"addEncode",function(){return s}),n.d(t,"extendEncode",function(){return u}),n.d(t,"encoders",function(){return c}),n.d(t,"has",function(){return l});var r=n("./node_modules/vega-parser/src/parsers/encode.js"),i=n("./node_modules/vega-parser/src/parsers/marks/roles.js"),o=n("./node_modules/vega-util/index.js");function a(e){return Object(o.isObject)(e)?e:{value:e}}function s(e,t,n){return null!=n?(e[t]=Object(o.isObject)(n)&&!Object(o.isArray)(n)?n:{value:n},1):0}function u(e,t,n){for(var r in t)n&&n.hasOwnProperty(r)||(e[r]=Object(o.extend)(e[r]||{},t[r]));return e}function c(e,t,n,a,s,u){var c,d;(u=u||{}).encoders={$encode:c={}},e=function(e,t,n,r,a){var s,u,c={};"legend"!=n&&0!==String(n).indexOf("axis")||(n=null);u=n===i.FrameRole?a.group:n===i.MarkRole?Object(o.extend)({},a.mark,a[t]):null;for(s in u)l(s,e)||("fill"===s||"stroke"===s)&&(l("fill",e)||l("stroke",e))||(c[s]={value:u[s]});return Object(o.array)(r).forEach(function(t){var n=a.style&&a.style[t];for(var r in n)l(r,e)||(c[r]={value:n[r]})}),(e=Object(o.extend)({},e)).enter=Object(o.extend)(c,e.enter),e}(e,t,n,a,s.config);for(d in e)c[d]=Object(r.default)(e[d],t,u,s);return u}function l(e,t){return t&&(t.enter&&t.enter[e]||t.update&&t.update[e])}},"./node_modules/vega-parser/src/parsers/encode/entry.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/encode/color.js"),i=n("./node_modules/vega-parser/src/parsers/encode/field.js"),o=n("./node_modules/vega-parser/src/parsers/encode/scale.js"),a=n("./node_modules/vega-parser/src/parsers/encode/gradient.js"),s=n("./node_modules/vega-parser/src/parsers/encode/property.js"),u=n("./node_modules/vega-parser/src/parsers/encode/expression.js"),c=n("./node_modules/vega-util/index.js");t.default=function(e,t,n,l,d){if(null!=t.gradient)return Object(a.default)(t,n,l,d);var p=t.signal?Object(u.default)(t.signal,n,l,d):t.color?Object(r.default)(t.color,n,l,d):null!=t.field?Object(i.default)(t.field,n,l,d):void 0!==t.value?Object(c.stringValue)(t.value):void 0;return null!=t.scale&&(p=Object(o.default)(t,p,n,l,d)),void 0===p&&(p=null),null!=t.exponent&&(p="Math.pow("+p+","+Object(s.default)(t.exponent,n,l,d)+")"),null!=t.mult&&(p+="*"+Object(s.default)(t.mult,n,l,d)),null!=t.offset&&(p+="+"+Object(s.default)(t.offset,n,l,d)),t.round&&(p="Math.round("+p+")"),p}},"./node_modules/vega-parser/src/parsers/encode/expression.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/expression.js"),i=n("./node_modules/vega-util/index.js");t.default=function(e,t,n,o){var a=Object(r.default)(e,t);return a.$fields.forEach(function(e){o[e]=1}),Object(i.extend)(n,a.$params),a.$expr}},"./node_modules/vega-parser/src/parsers/encode/field.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/encode/expression.js"),i=n("./node_modules/vega-util/index.js");t.default=function(e,t,n,o){return function e(t,n,o,a){var s,u,c;if(t.signal)s="datum",c=Object(r.default)(t.signal,n,o,a);else if(t.group||t.parent){for(u=Math.max(1,t.level||1),s="item";u-- >0;)s+=".mark.group";t.parent?(c=t.parent,s+=".datum"):c=t.group}else t.datum?(s="datum",c=t.datum):Object(i.error)("Invalid field reference: "+Object(i.stringValue)(t));t.signal||(Object(i.isString)(c)?(a[c]=1,c=Object(i.splitAccessPath)(c).map(i.stringValue).join("][")):c=e(c,n,o,a));return s+"["+c+"]"}(Object(i.isObject)(e)?e:{datum:e},t,n,o)}},"./node_modules/vega-parser/src/parsers/encode/gradient.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/encode/scale.js"),i=n("./node_modules/vega-util/index.js");t.default=function(e,t,n,o){return"this.gradient("+Object(r.getScale)(e.gradient,t,n,o)+","+Object(i.stringValue)(e.start)+","+Object(i.stringValue)(e.stop)+","+Object(i.stringValue)(e.count)+")"}},"./node_modules/vega-parser/src/parsers/encode/property.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/encode/entry.js"),i=n("./node_modules/vega-util/index.js");t.default=function(e,t,n,o){return Object(i.isObject)(e)?"("+Object(r.default)(null,e,t,n,o)+")":e}},"./node_modules/vega-parser/src/parsers/encode/rule.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/encode/entry.js"),i=n("./node_modules/vega-parser/src/parsers/encode/set.js"),o=n("./node_modules/vega-parser/src/parsers/encode/expression.js");t.default=function(e,t,n,a,s){var u="";return t.forEach(function(t){var i=Object(r.default)(e,t,n,a,s);u+=t.test?Object(o.default)(t.test,n,a,s)+"?"+i+":":i}),Object(i.default)("o",e,u)}},"./node_modules/vega-parser/src/parsers/encode/scale.js":function(e,t,n){"use strict";n.r(t),n.d(t,"getScale",function(){return s});var r=n("./node_modules/vega-parser/src/parsers/encode/expression.js"),i=n("./node_modules/vega-parser/src/parsers/encode/field.js"),o=n("./node_modules/vega-parser/src/parsers/expression/prefixes.js"),a=n("./node_modules/vega-util/index.js");function s(e,t,n,s){var u;if(Object(a.isString)(e))u=o.scalePrefix+e,n.hasOwnProperty(u)||(n[u]=t.scaleRef(e)),u=Object(a.stringValue)(u);else{for(u in t.scales)n[o.scalePrefix+u]=t.scaleRef(u);u=Object(a.stringValue)(o.scalePrefix)+"+"+(e.signal?"("+Object(r.default)(e.signal,t,n,s)+")":Object(i.default)(e,t,n,s))}return"_["+u+"]"}t.default=function(e,t,n,r,i){var o,u,c,l=s(e.scale,n,r,i);return null!=e.range?(u=l+".range()",t=0===(o=+e.range)?u+"[0]":"($="+u+","+(1===o?"$[$.length-1]":"$[0]+"+o+"*($[$.length-1]-$[0])")+")"):(void 0!==t&&(t=l+"("+t+")"),e.band&&(c=function(e,t){if(!Object(a.isString)(e))return-1;var n=t.scaleType(e);return"band"===n||"point"===n?1:0}(e.scale,n))&&(o=(u=l+".bandwidth")+"()"+(1===(o=+e.band)?"":"*"+o),c<0&&(o="("+u+"?"+o+":0)"),t=(t?t+"+":"")+o,e.extra&&(t="(datum.extra?"+l+"(datum.extra.value):"+t+")")),null==t&&(t="0")),t}},"./node_modules/vega-parser/src/parsers/encode/set.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/index.js");t.default=function(e,t,n){return e+"["+Object(r.stringValue)(t)+"]="+n+";"}},"./node_modules/vega-parser/src/parsers/expression.js":function(e,t,n){"use strict";n.r(t),n.d(t,"signalPrefix",function(){return a});var r=n("./node_modules/vega-parser/src/parsers/expression/codegen.js"),i=n("./node_modules/vega-expression/index.js"),o=n("./node_modules/vega-util/index.js"),a="$";t.default=function(e,t,n){var s,u,c={};try{s=Object(i.parse)(e)}catch(t){Object(o.error)("Expression parse error: "+Object(o.stringValue)(e))}return s.visit(function(e){if("CallExpression"===e.type){var n=e.callee.name,i=r.codegenParams.visitors[n];i&&i(n,e.arguments,t,c)}}),(u=Object(r.codeGenerator)(s)).globals.forEach(function(e){var n=a+e;!c.hasOwnProperty(n)&&t.getSignal(e)&&(c[n]=t.signalRef(e))}),{$expr:n?n+"return("+u.code+");":u.code,$fields:u.fields,$params:c}}},"./node_modules/vega-parser/src/parsers/expression/ast.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Literal",function(){return r}),n.d(t,"Identifier",function(){return i});var r="Literal",i="Identifier"},"./node_modules/vega-parser/src/parsers/expression/clamp-range.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n){var r,i=e[0],o=e[1];return o<i&&(r=o,o=i,i=r),(r=o-i)>=n-t?[t,n]:[Math.min(Math.max(i,t),n-r),Math.min(Math.max(o,r),n)]}},"./node_modules/vega-parser/src/parsers/expression/codegen.js":function(e,t,n){"use strict";n.r(t),n.d(t,"functionContext",function(){return k}),n.d(t,"expressionFunction",function(){return I}),n.d(t,"codegenParams",function(){return P}),n.d(t,"codeGenerator",function(){return R});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-expression/index.js"),o=n("./node_modules/vega-statistics/index.js"),a=n("./node_modules/vega-util/index.js"),s=n("./node_modules/d3-color/index.js"),u=n("./node_modules/d3-array/index.js"),c=n("./node_modules/vega-parser/src/parsers/expression/format.js"),l=n("./node_modules/vega-parser/src/parsers/expression/quarter.js"),d=n("./node_modules/vega-parser/src/parsers/expression/log.js"),p=n("./node_modules/vega-parser/src/parsers/expression/inscope.js"),f=n("./node_modules/vega-parser/src/parsers/expression/clamp-range.js"),h=n("./node_modules/vega-parser/src/parsers/expression/pinch.js"),m=n("./node_modules/vega-parser/src/parsers/expression/window.js"),g=n("./node_modules/vega-parser/src/parsers/expression/flush.js"),v=n("./node_modules/vega-parser/src/parsers/expression/span.js"),_=n("./node_modules/vega-parser/src/parsers/expression/scale.js"),y=n("./node_modules/vega-parser/src/parsers/expression/scale-gradient.js"),b=n("./node_modules/vega-parser/src/parsers/expression/geo.js"),j=n("./node_modules/vega-parser/src/parsers/expression/shape.js"),x=n("./node_modules/vega-parser/src/parsers/expression/data.js"),O=n("./node_modules/vega-parser/src/parsers/expression/tree.js"),w=n("./node_modules/vega-parser/src/parsers/expression/inrange.js"),C=n("./node_modules/vega-parser/src/parsers/expression/encode.js"),E=n("./node_modules/vega-parser/src/parsers/expression/modify.js"),S=n("./node_modules/vega-parser/src/parsers/expression/selection.js"),k={random:function(){return Object(o.random)()},isArray:a.isArray,isBoolean:a.isBoolean,isDate:a.isDate,isNumber:a.isNumber,isObject:a.isObject,isRegExp:a.isRegExp,isString:a.isString,isTuple:r.isTuple,toBoolean:a.toBoolean,toDate:a.toDate,toNumber:a.toNumber,toString:a.toString,pad:a.pad,peek:a.peek,truncate:a.truncate,rgb:s.rgb,lab:s.lab,hcl:s.hcl,hsl:s.hsl,sequence:u.range,format:c.format,utcFormat:c.utcFormat,utcParse:c.utcParse,timeFormat:c.timeFormat,timeParse:c.timeParse,monthFormat:c.monthFormat,monthAbbrevFormat:c.monthAbbrevFormat,dayFormat:c.dayFormat,dayAbbrevFormat:c.dayAbbrevFormat,quarter:l.quarter,utcquarter:l.utcquarter,warn:d.warn,info:d.info,debug:d.debug,inScope:p.default,clampRange:f.default,pinchDistance:h.pinchDistance,pinchAngle:h.pinchAngle,screen:m.screen,containerSize:m.containerSize,windowSize:m.windowSize,span:v.default,flush:g.default,bandspace:_.bandspace,inrange:w.default,setdata:x.setdata,pathShape:j.pathShape,panLinear:a.panLinear,panLog:a.panLog,panPow:a.panPow,zoomLinear:a.zoomLinear,zoomLog:a.zoomLog,zoomPow:a.zoomPow,encode:C.default,modify:E.default},M=["view","item","group","xy","x","y"],A="event.vega.",T="this.",D={};function I(e,t,n){return 1===arguments.length?k[e]:(k[e]=t,n&&(D[e]=n),R&&(R.functions[e]=T+e),this)}I("bandwidth",_.bandwidth,_.scaleVisitor),I("copy",_.copy,_.scaleVisitor),I("domain",_.domain,_.scaleVisitor),I("range",_.range,_.scaleVisitor),I("invert",_.invert,_.scaleVisitor),I("scale",_.scale,_.scaleVisitor),I("gradient",y.default,_.scaleVisitor),I("geoArea",b.geoArea,_.scaleVisitor),I("geoBounds",b.geoBounds,_.scaleVisitor),I("geoCentroid",b.geoCentroid,_.scaleVisitor),I("geoShape",j.geoShape,_.scaleVisitor),I("indata",x.indata,x.indataVisitor),I("data",x.data,x.dataVisitor),I("vlSingle",S.vlPoint,x.dataVisitor),I("vlSingleDomain",S.vlPointDomain,x.dataVisitor),I("vlMulti",S.vlPoint,S.vlMultiVisitor),I("vlMultiDomain",S.vlPointDomain,S.vlMultiVisitor),I("vlInterval",S.vlInterval,x.dataVisitor),I("vlIntervalDomain",S.vlIntervalDomain,x.dataVisitor),I("treePath",O.treePath,x.dataVisitor),I("treeAncestors",O.treeAncestors,x.dataVisitor);var P={blacklist:["_"],whitelist:["datum","event","item"],fieldvar:"datum",globalvar:function(e){return"_["+Object(a.stringValue)("$"+e)+"]"},functions:function(e){var t=Object(i.functions)(e);M.forEach(function(e){t[e]=A+e});for(var n in k)t[n]=T+n;return t},constants:i.constants,visitors:D},R=Object(i.codegen)(P)},"./node_modules/vega-parser/src/parsers/expression/data.js":function(e,t,n){"use strict";n.r(t),n.d(t,"data",function(){return a}),n.d(t,"dataVisitor",function(){return s}),n.d(t,"indata",function(){return u}),n.d(t,"indataVisitor",function(){return c}),n.d(t,"setdata",function(){return l});var r=n("./node_modules/vega-parser/src/parsers/expression/ast.js"),i=n("./node_modules/vega-parser/src/parsers/expression/prefixes.js"),o=n("./node_modules/vega-util/index.js");function a(e){var t=this.context.data[e];return t?t.values.value:[]}function s(e,t,n,a){t[0].type!==r.Literal&&Object(o.error)("First argument to data functions must be a string literal.");var s=t[0].value,u=i.dataPrefix+s;a.hasOwnProperty(u)||(a[u]=n.getData(s).tuplesRef())}function u(e,t,n){var r=this.context.data[e]["index:"+t],i=r?r.value.get(n):void 0;return i?i.count:i}function c(e,t,n,a){t[0].type!==r.Literal&&Object(o.error)("First argument to indata must be a string literal."),t[1].type!==r.Literal&&Object(o.error)("Second argument to indata must be a string literal.");var s=t[0].value,u=t[1].value,c=i.indexPrefix+u;a.hasOwnProperty(c)||(a[c]=n.getData(s).indataRef(n,u))}function l(e,t){var n=this.context.dataflow,r=this.context.data[e].input;return n.pulse(r,n.changeset().remove(o.truthy).insert(t)),1}},"./node_modules/vega-parser/src/parsers/expression/encode.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n){if(e){var r=this.context.dataflow,i=e.mark.source;r.pulse(i,r.changeset().encode(e,t))}return void 0!==n?n:e}},"./node_modules/vega-parser/src/parsers/expression/flush.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/index.js");t.default=function(e,t,n,i,o,a){var s=Math.abs(t-e[0]),u=Math.abs(Object(r.peek)(e)-t);return s<u&&s<=n?i:u<=n?o:a}},"./node_modules/vega-parser/src/parsers/expression/format.js":function(e,t,n){"use strict";n.r(t),n.d(t,"format",function(){return s}),n.d(t,"timeFormat",function(){return u}),n.d(t,"utcFormat",function(){return c}),n.d(t,"timeParse",function(){return l}),n.d(t,"utcParse",function(){return d}),n.d(t,"monthFormat",function(){return h}),n.d(t,"monthAbbrevFormat",function(){return m}),n.d(t,"dayFormat",function(){return g}),n.d(t,"dayAbbrevFormat",function(){return v});var r=n("./node_modules/d3-format/index.js"),i=n("./node_modules/d3-time-format/index.js"),o={};function a(e,t,n){var r=e+":"+n,i=o[r];return i&&i[0]===t||(o[r]=i=[t,t(n)]),i[1]}function s(e,t){return a("format",r.format,t)(e)}function u(e,t){return a("timeFormat",i.timeFormat,t)(e)}function c(e,t){return a("utcFormat",i.utcFormat,t)(e)}function l(e,t){return a("timeParse",i.timeParse,t)(e)}function d(e,t){return a("utcParse",i.utcParse,t)(e)}var p=new Date(2e3,0,1);function f(e,t,n){return p.setMonth(e),p.setDate(t),u(p,n)}function h(e){return f(e,1,"%B")}function m(e){return f(e,1,"%b")}function g(e){return f(0,2+e,"%A")}function v(e){return f(0,2+e,"%a")}},"./node_modules/vega-parser/src/parsers/expression/geo.js":function(e,t,n){"use strict";n.r(t),n.d(t,"geoArea",function(){return a}),n.d(t,"geoBounds",function(){return s}),n.d(t,"geoCentroid",function(){return u});var r=n("./node_modules/d3-geo/index.js"),i=n("./node_modules/vega-parser/src/parsers/expression/scale.js");function o(e,t){return function(n,r,o){if(n){var a=Object(i.getScale)(n,(o||this).context);return a&&a.path[e](r)}return t(r)}}var a=o("area",r.geoArea),s=o("bounds",r.geoBounds),u=o("centroid",r.geoCentroid)},"./node_modules/vega-parser/src/parsers/expression/inrange.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n,r){var i,o=t[0],a=t[t.length-1];return o>a&&(i=o,o=a,a=i),n=void 0===n||n,r=void 0===r||r,(n?o<=e:o<e)&&(r?e<=a:e<a)}},"./node_modules/vega-parser/src/parsers/expression/inscope.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){var t=this.context.group,n=!1;if(t)for(;e;){if(e===t){n=!0;break}e=e.mark.group}return n}},"./node_modules/vega-parser/src/parsers/expression/log.js":function(e,t,n){"use strict";function r(e,t,n){try{e[t].apply(e,["EXPRESSION"].concat([].slice.call(n)))}catch(t){e.warn(t)}return n[n.length-1]}function i(){return r(this.context.dataflow,"warn",arguments)}function o(){return r(this.context.dataflow,"info",arguments)}function a(){return r(this.context.dataflow,"debug",arguments)}n.r(t),n.d(t,"warn",function(){return i}),n.d(t,"info",function(){return o}),n.d(t,"debug",function(){return a})},"./node_modules/vega-parser/src/parsers/expression/modify.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e,t){return e===t||e!=e&&t!=t||!(!Object(i.isArray)(e)||!Object(i.isArray)(t)||e.length!==t.length)&&function(e,t){for(var n=0,r=e.length;n<r;++n)if(!o(e[n],t[n]))return!1;return!0}(e,t)}function a(e){return function(t){for(var n in e)if(!o(t[n],e[n]))return!1;return!0}}t.default=function(e,t,n,o,s,u){var c,l,d=this.context.dataflow,p=this.context.data[e],f=p.input,h=p.changes,m=d.stamp();if(!1===d._trigger||!(f.value.length||t||o))return 0;if((!h||h.stamp<m)&&(p.changes=h=d.changeset(),h.stamp=m,d.runAfter(function(){p.modified=!0,d.pulse(f,h).run()},!0,1)),n&&(c=!0===n?i.truthy:Object(i.isArray)(n)||Object(r.isTuple)(n)?n:a(n),h.remove(c)),t&&h.insert(t),o&&(c=a(o),f.value.some(c)?h.remove(c):h.insert(o)),s)for(l in u)h.modify(s,l,u[l]);return 1}},"./node_modules/vega-parser/src/parsers/expression/pinch.js":function(e,t,n){"use strict";function r(e){var t=e.touches,n=t[0].clientX-t[1].clientX,r=t[0].clientY-t[1].clientY;return Math.sqrt(n*n+r*r)}function i(e){var t=e.touches;return Math.atan2(t[0].clientY-t[1].clientY,t[0].clientX-t[1].clientX)}n.r(t),n.d(t,"pinchDistance",function(){return r}),n.d(t,"pinchAngle",function(){return i})},"./node_modules/vega-parser/src/parsers/expression/prefixes.js":function(e,t,n){"use strict";n.r(t),n.d(t,"indexPrefix",function(){return r}),n.d(t,"scalePrefix",function(){return i}),n.d(t,"dataPrefix",function(){return o});var r="@",i="%",o=":"},"./node_modules/vega-parser/src/parsers/expression/quarter.js":function(e,t,n){"use strict";function r(e){return 1+~~(new Date(e).getMonth()/3)}function i(e){return 1+~~(new Date(e).getUTCMonth()/3)}n.r(t),n.d(t,"quarter",function(){return r}),n.d(t,"utcquarter",function(){return i})},"./node_modules/vega-parser/src/parsers/expression/scale-gradient.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/expression/scale.js"),i=n("./node_modules/vega-scenegraph/index.js"),o=n("./node_modules/vega-scale/index.js");t.default=function(e,t,n,a,s){e=Object(r.getScale)(e,(s||this).context);var u=Object(i.Gradient)(t,n),c=e.domain(),l=c[0],d=c[c.length-1],p=Object(o.scaleFraction)(e,l,d);e.ticks&&(l!==(c=e.ticks(+a||15))[0]&&c.unshift(l),d!==c[c.length-1]&&c.push(d));for(var f=0,h=c.length;f<h;++f)u.stop(p(c[f]),e(c[f]));return u}},"./node_modules/vega-parser/src/parsers/expression/scale.js":function(e,t,n){"use strict";n.r(t),n.d(t,"getScale",function(){return s}),n.d(t,"scaleVisitor",function(){return c}),n.d(t,"range",function(){return l}),n.d(t,"domain",function(){return d}),n.d(t,"bandwidth",function(){return p}),n.d(t,"bandspace",function(){return f}),n.d(t,"copy",function(){return h}),n.d(t,"scale",function(){return m}),n.d(t,"invert",function(){return g});var r=n("./node_modules/vega-parser/src/parsers/expression/ast.js"),i=n("./node_modules/vega-parser/src/parsers/expression/prefixes.js"),o=n("./node_modules/vega-scale/index.js"),a=n("./node_modules/vega-util/index.js");function s(e,t){var n;return Object(a.isFunction)(e)?e:Object(a.isString)(e)?(n=t.scales[e])&&n.value:void 0}function u(e,t,n){var r=i.scalePrefix+n;if(!t.hasOwnProperty(r))try{t[r]=e.scaleRef(n)}catch(e){}}function c(e,t,n,i){if(t[0].type===r.Literal)u(n,i,t[0].value);else if(t[0].type===r.Identifier)for(e in n.scales)u(n,i,e)}function l(e,t){var n=s(e,(t||this).context);return n&&n.range?n.range():[]}function d(e,t){var n=s(e,(t||this).context);return n?n.domain():[]}function p(e,t){var n=s(e,(t||this).context);return n&&n.bandwidth?n.bandwidth():0}function f(e,t,n){return Object(o.bandSpace)(e||0,t||0,n||0)}function h(e,t){var n=s(e,(t||this).context);return n?n.copy():void 0}function m(e,t,n){var r=s(e,(n||this).context);return r?r(t):void 0}function g(e,t,n){var r=s(e,(n||this).context);return r?Object(a.isArray)(t)?(r.invertRange||r.invert)(t):(r.invert||r.invertExtent)(t):void 0}},"./node_modules/vega-parser/src/parsers/expression/selection.js":function(e,t,n){"use strict";n.r(t),n.d(t,"vlPoint",function(){return m}),n.d(t,"vlInterval",function(){return g}),n.d(t,"vlMultiVisitor",function(){return v}),n.d(t,"vlPointDomain",function(){return _}),n.d(t,"vlIntervalDomain",function(){return y});var r=n("./node_modules/vega-parser/src/parsers/expression/ast.js"),i=n("./node_modules/vega-parser/src/parsers/expression/data.js"),o=n("./node_modules/vega-parser/src/parsers/expression/inrange.js"),a=n("./node_modules/vega-parser/src/parsers/expression/prefixes.js"),s=n("./node_modules/vega-util/index.js"),u="bin_",c="intersect",l="union",d="index:unit";function p(e,t){for(var n,r=t.fields,i=t.values,a=t.getter||(t.getter=[]),c=r.length,l=0;l<c;++l)if(a[l]=a[l]||Object(s.field)(r[l]),n=a[l](e),Object(s.isDate)(n)&&(n=Object(s.toNumber)(n)),Object(s.isDate)(i[l])&&(i[l]=Object(s.toNumber)(i[l])),t[u+r[l]]){if(Object(s.isDate)(i[l][0])&&(i[l]=i[l].map(s.toNumber)),!Object(o.default)(n,i[l],!0,!1))return!1}else if(n!==i[l])return!1;return!0}function f(e,t){for(var n,r,i=t.intervals,a=i.length,u=0;u<a;++u){if(n=i[u].extent,r=(i[u].getter||(i[u].getter=Object(s.field)(i[u].field)))(e),!n||n[0]===n[1])return!1;if(Object(s.isDate)(r)&&(r=Object(s.toNumber)(r)),Object(s.isDate)(n[0])&&(n=i[u].extent=n.map(s.toNumber)),Object(s.isNumber)(n[0])&&!Object(o.default)(r,n))return!1;if(Object(s.isString)(n[0])&&n.indexOf(r)<0)return!1}return!0}function h(e,t,n,r){for(var i,o,a,s,u,l=this.context.data[e],p=l?l.values.value:[],f=l?l[d]&&l[d].value:void 0,h=n===c,m=p.length,g=0;g<m;++g)if(i=p[g],f&&h){if(-1===(a=(o=o||{})[s=i.unit]||0))continue;if(u=r(t,i),o[s]=u?-1:++a,u&&1===f.size)return!0;if(!u&&a===f.get(s).count)return!1}else if(h^(u=r(t,i)))return u;return m&&h}function m(e,t,n){return h.call(this,e,t,n,p)}function g(e,t,n){return h.call(this,e,t,n,f)}function v(e,t,n,o){t[0].type!==r.Literal&&Object(s.error)("First argument to indata must be a string literal.");var u=t[0].value,l=t.length>=2&&t[t.length-1].value,d=a.indexPrefix+"unit";l!==c||o.hasOwnProperty(d)||(o[d]=n.getData(u).indataRef(n,"unit")),Object(i.dataVisitor)(e,t,n,o)}function _(e,t,n,r){var i,o,a,s,p,f=this.context.data[e],h=f?f.values.value:[],m=f?f[d]&&f[d].value:void 0,g=h[0],v=0;if(g){for(i=t?g.encodings.length:g.fields.length;v<i;++v)if(t&&g.encodings[v]===t||n&&g.fields[v]===n){o=v,s=g[u+g.fields[v]];break}return m&&1===m.size&&(r=l),m&&r===c?(p=h.reduce(function(e,t){return(e[t.unit]||(e[t.unit]=[])).push({unit:t.unit,value:t.values[o]}),e},{}),a=Object.keys(p).map(function(e){return{unit:e,value:s?j(p[e],l):b(p[e],l)}})):a=h.map(function(e){return{unit:e.unit,value:e.values[o]}}),s?j(a,r):b(a,r)}}function y(e,t,n,r){var i,o,a,s,u,c=this.context.data[e],l=c?c.values.value:[],d=l[0],p=0;if(d){for(i=d.intervals.length;p<i;++p)if(o=d.intervals[p],t&&o.encoding===t||n&&o.field===n){if(!o.extent)return;a=p,u=o.extent.length>2;break}return s=l.reduce(function(e,t){var n=t.intervals[a].extent,r=u?n.map(function(e){return{unit:t.unit,value:e}}):{unit:t.unit,value:n};return u?e.push.apply(e,r):e.push(r),e},[]),u?b(s,r):j(s,r)}}function b(e,t){for(var n,r,i,o,a={},s=0,u={},l=[],d=0,p=e.length;d<p;++d)r=(n=e[d]).unit,o=n.value,a[r]||(a[r]=++s),(i=u[o])||(u[o]=i={value:o,units:{},count:0}),i.units[r]||(i.units[r]=++i.count);for(o in u)i=u[o],t===c&&i.count!==s||l.push(i.value);return l.length?l:void 0}function j(e,t){for(var n,r,i,o,a=t===c?O:x,u=0,l=e.length;u<l;++u)n=e[u].value,Object(s.isDate)(n[0])&&(n=n.map(s.toNumber)),(i=n[0])>(o=n[1])&&(o=n[0],i=n[1]),r=r?a(r,i,o):[i,o];return r&&r.length&&+r[0]!=+r[1]?r:void 0}function x(e,t,n){return e[0]>t&&(e[0]=t),e[1]<n&&(e[1]=n),e}function O(e,t,n){return n<e[0]||e[1]<t?[]:(e[0]<t&&(e[0]=t),e[1]>n&&(e[1]=n),e)}},"./node_modules/vega-parser/src/parsers/expression/shape.js":function(e,t,n){"use strict";n.r(t),n.d(t,"geoShape",function(){return o}),n.d(t,"pathShape",function(){return a});var r=n("./node_modules/vega-parser/src/parsers/expression/scale.js"),i=n("./node_modules/vega-scenegraph/index.js");function o(e,t,n){var i=Object(r.getScale)(e,(n||this).context);return function(e){return i?i.path.context(e)(t):""}}function a(e){var t=null;return function(n){return n?Object(i.pathRender)(n,t=t||Object(i.pathParse)(e)):e}}},"./node_modules/vega-parser/src/parsers/expression/span.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return e[e.length-1]-e[0]||0}},"./node_modules/vega-parser/src/parsers/expression/tree.js":function(e,t,n){"use strict";n.r(t),n.d(t,"treePath",function(){return s}),n.d(t,"treeAncestors",function(){return u});var r=n("./node_modules/vega-parser/src/parsers/expression/data.js"),i={};function o(e){return e.data}function a(e,t){var n=r.data.call(t,e);return n.root&&n.root.lookup||i}function s(e,t,n){var r=a(e,this),i=r[t],s=r[n];return i&&s?i.path(s).map(o):void 0}function u(e,t){var n=a(e,this)[t];return n?n.ancestors().map(o):void 0}},"./node_modules/vega-parser/src/parsers/expression/window.js":function(e,t,n){"use strict";n.r(t),n.d(t,"screen",function(){return i}),n.d(t,"windowSize",function(){return o}),n.d(t,"containerSize",function(){return a});var r="undefined"!=typeof window&&window||null;function i(){return r?r.screen:{}}function o(){return r?[r.innerWidth,r.innerHeight]:[void 0,void 0]}function a(){var e=this.context.dataflow,t=e.container&&e.container();return t?[t.clientWidth,t.clientHeight]:[void 0,void 0]}},"./node_modules/vega-parser/src/parsers/guides/axis-config.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/guides/constants.js"),i=n("./node_modules/vega-util/index.js");t.default=function(e,t){var n=t.config,o=e.orient,a=o===r.Top||o===r.Bottom?n.axisX:n.axisY,s=n["axis"+o[0].toUpperCase()+o.slice(1)],u="band"===t.scaleType(e.scale)&&n.axisBand;return a||s||u?Object(i.extend)({},n.axis,a,s,u):n.axis}},"./node_modules/vega-parser/src/parsers/guides/axis-domain.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/guides/constants.js"),i=n("./node_modules/vega-parser/src/parsers/guides/guide-mark.js"),o=n("./node_modules/vega-parser/src/parsers/marks/marktypes.js"),a=n("./node_modules/vega-parser/src/parsers/marks/roles.js"),s=n("./node_modules/vega-parser/src/parsers/encode/encode-util.js");function u(e,t){return{scale:e.scale,range:t}}t.default=function(e,t,n,c){var l,d,p,f,h,m=e.orient,g={value:0},v={};return v.enter=l={opacity:g},Object(s.addEncode)(l,"stroke",t.domainColor),Object(s.addEncode)(l,"strokeWidth",t.domainWidth),v.exit={opacity:g},v.update=d={opacity:{value:1}},m===r.Top||m===r.Bottom?(p="x",h="y"):(p="y",h="x"),f=p+"2",l[h]=g,d[p]=l[p]=u(e,0),d[f]=l[f]=u(e,1),Object(i.default)(o.RuleMark,a.AxisDomainRole,null,null,c,v,n)}},"./node_modules/vega-parser/src/parsers/guides/axis-grid.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/guides/constants.js"),i=n("./node_modules/vega-parser/src/parsers/guides/guide-mark.js"),o=n("./node_modules/vega-parser/src/parsers/marks/marktypes.js"),a=n("./node_modules/vega-parser/src/parsers/marks/roles.js"),s=n("./node_modules/vega-parser/src/parsers/encode/encode-util.js");t.default=function(e,t,n,u){var c,l,d,p,f,h,m,g,v=e.orient,_=e.gridScale,y=v===r.Left||v===r.Top?1:-1,b=y*e.offset||0,j={value:0},x={};return x.enter=c={opacity:j},Object(s.addEncode)(c,"stroke",t.gridColor),Object(s.addEncode)(c,"strokeWidth",t.gridWidth),Object(s.addEncode)(c,"strokeDash",t.gridDash),x.exit=l={opacity:j},x.update=d={},Object(s.addEncode)(d,"opacity",t.gridOpacity),p={scale:e.scale,field:r.Value,band:t.bandPosition,round:t.tickRound,extra:t.tickExtra,offset:t.tickOffset},v===r.Top||v===r.Bottom?(f="x",h="y",g="height"):(f="y",h="x",g="width"),m=h+"2",d[f]=c[f]=l[f]=p,_?(c[h]={scale:_,range:0,mult:y,offset:b},d[m]=c[m]={scale:_,range:1,mult:y,offset:b}):(c[h]={value:b},d[m]=c[m]={signal:g,mult:y,offset:b}),Object(i.default)(o.RuleMark,a.AxisGridRole,null,r.Value,u,x,n)}},"./node_modules/vega-parser/src/parsers/guides/axis-labels.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/guides/constants.js"),i=n("./node_modules/vega-parser/src/parsers/guides/guide-mark.js"),o=n("./node_modules/vega-parser/src/parsers/marks/marktypes.js"),a=n("./node_modules/vega-parser/src/parsers/marks/roles.js"),s=n("./node_modules/vega-parser/src/parsers/encode/encode-util.js"),u=n("./node_modules/vega-parser/src/util.js");function c(e,t,n,r,i){return{signal:'flush(range("'+e+'"), scale("'+e+'", datum.value), '+t+","+n+","+r+","+i+")"}}t.default=function(e,t,n,l,d){var p,f,h,m,g,v=e.orient,_=v===r.Left||v===r.Top?-1:1,y=e.scale,b=Object(u.value)(e.labelPadding,t.labelPadding),j=Object(u.value)(e.labelBound,t.labelBound),x=Object(u.value)(e.labelFlush,t.labelFlush),O=null!=x&&!1!==x&&(x=+x)===x,w=+Object(u.value)(e.labelFlushOffset,t.labelFlushOffset),C=Object(u.value)(e.labelOverlap,t.labelOverlap),E={value:0},S={};return S.enter=p={opacity:E},Object(s.addEncode)(p,"angle",t.labelAngle),Object(s.addEncode)(p,"fill",t.labelColor),Object(s.addEncode)(p,"font",t.labelFont),Object(s.addEncode)(p,"fontSize",t.labelFontSize),Object(s.addEncode)(p,"fontWeight",t.labelFontWeight),Object(s.addEncode)(p,"limit",t.labelLimit),S.exit=f={opacity:E},S.update=h={opacity:{value:1},text:{field:r.Label}},(m=Object(s.encoder)(d)).mult=_,m.offset=Object(s.encoder)(b),m.offset.mult=_,g={scale:y,field:r.Value,band:.5,offset:t.tickOffset},v===r.Top||v===r.Bottom?(h.y=p.y=m,h.x=p.x=f.x=g,Object(s.addEncode)(h,"align",O?c(y,x,'"left"','"right"','"center"'):"center"),O&&w&&Object(s.addEncode)(h,"dx",c(y,x,-w,w,0)),Object(s.addEncode)(h,"baseline",v===r.Top?"bottom":"top")):(h.x=p.x=m,h.y=p.y=f.y=g,Object(s.addEncode)(h,"align",v===r.Right?"left":"right"),Object(s.addEncode)(h,"baseline",O?c(y,x,'"bottom"','"top"','"middle"'):"middle"),O&&w&&Object(s.addEncode)(h,"dy",c(y,x,w,-w,0))),e=Object(i.default)(o.TextMark,a.AxisLabelRole,r.GuideLabelStyle,r.Value,l,S,n),(C||j)&&(e.overlap={method:C,order:"datum.index",bound:j?{scale:y,orient:v,tolerance:+j}:null}),e}},"./node_modules/vega-parser/src/parsers/guides/axis-ticks.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/guides/constants.js"),i=n("./node_modules/vega-parser/src/parsers/guides/guide-mark.js"),o=n("./node_modules/vega-parser/src/parsers/marks/marktypes.js"),a=n("./node_modules/vega-parser/src/parsers/marks/roles.js"),s=n("./node_modules/vega-parser/src/parsers/encode/encode-util.js");t.default=function(e,t,n,u,c){var l,d,p,f,h,m=e.orient,g=m===r.Left||m===r.Top?-1:1,v={value:0},_={};return _.enter=l={opacity:v},Object(s.addEncode)(l,"stroke",t.tickColor),Object(s.addEncode)(l,"strokeWidth",t.tickWidth),_.exit=d={opacity:v},_.update=p={opacity:{value:1}},(f=Object(s.encoder)(c)).mult=g,h={scale:e.scale,field:r.Value,band:t.bandPosition,round:t.tickRound,extra:t.tickExtra,offset:t.tickOffset},m===r.Top||m===r.Bottom?(p.y=l.y=v,p.y2=l.y2=f,p.x=l.x=d.x=h):(p.x=l.x=v,p.x2=l.x2=f,p.y=l.y=d.y=h),Object(i.default)(o.RuleMark,a.AxisTickRole,null,r.Value,u,_,n)}},"./node_modules/vega-parser/src/parsers/guides/axis-title.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/guides/constants.js"),i=n("./node_modules/vega-parser/src/parsers/guides/guide-mark.js"),o=n("./node_modules/vega-parser/src/parsers/encode/encode-util.js"),a=n("./node_modules/vega-parser/src/parsers/marks/marktypes.js"),s=n("./node_modules/vega-parser/src/parsers/marks/roles.js");t.default=function(e,t,n,u){var c,l,d,p=e.orient,f=e.title,h=p===r.Left||p===r.Top?-1:1,m=p===r.Top||p===r.Bottom,g={};return g.enter=c={opacity:{value:0}},Object(o.addEncode)(c,"align",t.titleAlign),Object(o.addEncode)(c,"fill",t.titleColor),Object(o.addEncode)(c,"font",t.titleFont),Object(o.addEncode)(c,"fontSize",t.titleFontSize),Object(o.addEncode)(c,"fontWeight",t.titleFontWeight),Object(o.addEncode)(c,"limit",t.titleLimit),g.exit={opacity:{value:0}},g.update=l={opacity:{value:1},text:f&&f.signal?{signal:f.signal}:{value:f+""}},d={scale:e.scale,range:.5},m?(l.x=d,l.angle={value:0},l.baseline={value:p===r.Top?"bottom":"top"}):(l.y=d,l.angle={value:90*h},l.baseline={value:"bottom"}),Object(o.addEncode)(l,"angle",t.titleAngle),Object(o.addEncode)(l,"baseline",t.titleBaseline),!Object(o.addEncode)(l,"x",t.titleX)&&m&&!Object(o.has)("x",n)&&(g.enter.auto={value:!0}),!Object(o.addEncode)(l,"y",t.titleY)&&!m&&!Object(o.has)("y",n)&&(g.enter.auto={value:!0}),Object(i.default)(a.TextMark,s.AxisTitleRole,r.GuideTitleStyle,null,u,g,n)}},"./node_modules/vega-parser/src/parsers/guides/constants.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Top",function(){return r}),n.d(t,"Left",function(){return i}),n.d(t,"Right",function(){return o}),n.d(t,"Bottom",function(){return a}),n.d(t,"Index",function(){return s}),n.d(t,"Label",function(){return u}),n.d(t,"Offset",function(){return c}),n.d(t,"Perc",function(){return l}),n.d(t,"Size",function(){return d}),n.d(t,"Total",function(){return p}),n.d(t,"Value",function(){return f}),n.d(t,"GuideLabelStyle",function(){return h}),n.d(t,"GuideTitleStyle",function(){return m}),n.d(t,"GroupTitleStyle",function(){return g}),n.d(t,"LegendScales",function(){return v}),n.d(t,"Skip",function(){return _});var r="top",i="left",o="right",a="bottom",s="index",u="label",c="offset",l="perc",d="size",p="total",f="value",h="guide-label",m="guide-title",g="group-title",v=["shape","size","fill","stroke","strokeDash","opacity"],_={name:1,interactive:1}},"./node_modules/vega-parser/src/parsers/guides/guide-group.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/marks/marktypes.js");t.default=function(e,t,n,i,o,a,s){return{type:r.GroupMark,name:n,role:e,style:t,from:i,interactive:o||!1,encode:a,marks:s}}},"./node_modules/vega-parser/src/parsers/guides/guide-mark.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/guides/constants.js"),i=n("./node_modules/vega-parser/src/parsers/encode/encode-util.js");t.default=function(e,t,n,o,a,s,u){return{type:e,name:u?u.name:void 0,role:t,style:u&&u.style||n,key:o,from:a,interactive:!(!u||!u.interactive),encode:Object(i.extendEncode)(s,u,r.Skip)}}},"./node_modules/vega-parser/src/parsers/guides/legend-gradient-labels.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/guides/constants.js"),i=n("./node_modules/vega-parser/src/parsers/guides/guide-mark.js"),o=n("./node_modules/vega-parser/src/parsers/marks/marktypes.js"),a=n("./node_modules/vega-parser/src/parsers/marks/roles.js"),s=n("./node_modules/vega-parser/src/parsers/encode/encode-util.js"),u="datum."+r.Perc+'<=0?"left":datum.'+r.Perc+'>=1?"right":"center"';t.default=function(e,t,n,c){var l,d,p={value:0},f={};return f.enter=l={opacity:p},Object(s.addEncode)(l,"fill",t.labelColor),Object(s.addEncode)(l,"font",t.labelFont),Object(s.addEncode)(l,"fontSize",t.labelFontSize),Object(s.addEncode)(l,"fontWeight",t.labelFontWeight),Object(s.addEncode)(l,"baseline",t.gradientLabelBaseline),Object(s.addEncode)(l,"limit",t.gradientLabelLimit),f.exit={opacity:p},f.update=d={opacity:{value:1},text:{field:r.Label}},l.x=d.x={field:r.Perc,mult:t.gradientWidth},l.y=d.y={value:t.gradientHeight,offset:t.gradientLabelOffset},l.align=d.align={signal:u},Object(i.default)(o.TextMark,a.LegendLabelRole,r.GuideLabelStyle,r.Perc,c,f,n)}},"./node_modules/vega-parser/src/parsers/guides/legend-gradient.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/guides/guide-mark.js"),i=n("./node_modules/vega-parser/src/parsers/marks/marktypes.js"),o=n("./node_modules/vega-parser/src/parsers/marks/roles.js"),a=n("./node_modules/vega-parser/src/parsers/encode/encode-util.js");t.default=function(e,t,n,s){var u,c,l={value:0},d={};return d.enter=u={opacity:l,x:l,y:l},Object(a.addEncode)(u,"width",n.gradientWidth),Object(a.addEncode)(u,"height",n.gradientHeight),Object(a.addEncode)(u,"stroke",n.gradientStrokeColor),Object(a.addEncode)(u,"strokeWidth",n.gradientStrokeWidth),d.exit={opacity:l},d.update=c={x:l,y:l,fill:{gradient:t},opacity:{value:1}},Object(a.addEncode)(c,"width",n.gradientWidth),Object(a.addEncode)(c,"height",n.gradientHeight),Object(r.default)(i.RectMark,o.LegendGradientRole,null,void 0,void 0,d,s)}},"./node_modules/vega-parser/src/parsers/guides/legend-labels.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/guides/constants.js"),i=n("./node_modules/vega-parser/src/parsers/guides/guide-mark.js"),o=n("./node_modules/vega-parser/src/parsers/marks/marktypes.js"),a=n("./node_modules/vega-parser/src/parsers/marks/roles.js"),s=n("./node_modules/vega-parser/src/parsers/encode/encode-util.js");t.default=function(e,t,n,u){var c,l,d={value:0},p={};return p.enter=c={opacity:d},Object(s.addEncode)(c,"align",t.labelAlign),Object(s.addEncode)(c,"baseline",t.labelBaseline),Object(s.addEncode)(c,"fill",t.labelColor),Object(s.addEncode)(c,"font",t.labelFont),Object(s.addEncode)(c,"fontSize",t.labelFontSize),Object(s.addEncode)(c,"fontWeight",t.labelFontWeight),Object(s.addEncode)(c,"limit",t.labelLimit),p.exit={opacity:d},p.update=l={opacity:{value:1},text:{field:r.Label}},c.x=l.x={field:r.Offset,offset:t.labelOffset},c.y=l.y={field:r.Size,mult:.5,offset:{field:r.Total,offset:{field:{group:"entryPadding"},mult:{field:r.Index}}}},Object(i.default)(o.TextMark,a.LegendLabelRole,r.GuideLabelStyle,r.Value,u,p,n)}},"./node_modules/vega-parser/src/parsers/guides/legend-symbols.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/guides/constants.js"),i=n("./node_modules/vega-parser/src/parsers/guides/guide-mark.js"),o=n("./node_modules/vega-parser/src/parsers/marks/marktypes.js"),a=n("./node_modules/vega-parser/src/parsers/marks/roles.js"),s=n("./node_modules/vega-parser/src/parsers/encode/encode-util.js");t.default=function(e,t,n,u){var c,l,d={value:0},p={};return p.enter=c={opacity:d},Object(s.addEncode)(c,"shape",t.symbolType),Object(s.addEncode)(c,"size",t.symbolSize),Object(s.addEncode)(c,"strokeWidth",t.symbolStrokeWidth),e.fill||(Object(s.addEncode)(c,"fill",t.symbolFillColor),Object(s.addEncode)(c,"stroke",t.symbolStrokeColor)),p.exit={opacity:d},p.update=l={opacity:{value:1}},c.x=l.x={field:r.Offset,mult:.5},c.y=l.y={field:r.Size,mult:.5,offset:{field:r.Total,offset:{field:{group:"entryPadding"},mult:{field:r.Index}}}},r.LegendScales.forEach(function(t){e[t]&&(l[t]=c[t]={scale:e[t],field:r.Value})}),Object(i.default)(o.SymbolMark,a.LegendSymbolRole,null,r.Value,u,p,n)}},"./node_modules/vega-parser/src/parsers/guides/legend-title.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/guides/constants.js"),i=n("./node_modules/vega-parser/src/parsers/guides/guide-mark.js"),o=n("./node_modules/vega-parser/src/parsers/marks/marktypes.js"),a=n("./node_modules/vega-parser/src/parsers/marks/roles.js"),s=n("./node_modules/vega-parser/src/parsers/encode/encode-util.js");t.default=function(e,t,n,u){var c,l={value:0},d=e.title,p={};return p.enter=c={x:{field:{group:"padding"}},y:{field:{group:"padding"}},opacity:l},Object(s.addEncode)(c,"align",t.titleAlign),Object(s.addEncode)(c,"baseline",t.titleBaseline),Object(s.addEncode)(c,"fill",t.titleColor),Object(s.addEncode)(c,"font",t.titleFont),Object(s.addEncode)(c,"fontSize",t.titleFontSize),Object(s.addEncode)(c,"fontWeight",t.titleFontWeight),Object(s.addEncode)(c,"limit",t.titleLimit),p.exit={opacity:l},p.update={opacity:{value:1},text:d&&d.signal?{signal:d.signal}:{value:d+""}},Object(i.default)(o.TextMark,a.LegendTitleRole,r.GuideTitleStyle,null,u,p,n)}},"./node_modules/vega-parser/src/parsers/legend.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/guides/legend-gradient.js"),i=n("./node_modules/vega-parser/src/parsers/guides/legend-gradient-labels.js"),o=n("./node_modules/vega-parser/src/parsers/guides/legend-labels.js"),a=n("./node_modules/vega-parser/src/parsers/guides/legend-symbols.js"),s=n("./node_modules/vega-parser/src/parsers/guides/legend-title.js"),u=n("./node_modules/vega-parser/src/parsers/guides/guide-group.js"),c=n("./node_modules/vega-parser/src/parsers/mark.js"),l=n("./node_modules/vega-parser/src/parsers/marks/roles.js"),d=n("./node_modules/vega-parser/src/parsers/encode/encode-util.js"),p=n("./node_modules/vega-parser/src/parsers/guides/constants.js"),f=n("./node_modules/vega-parser/src/util.js"),h=n("./node_modules/vega-parser/src/transforms.js"),m=n("./node_modules/vega-util/index.js");function g(e,t,n,r){var i=t&&(t.update&&t.update[n]||t.enter&&t.enter[n]);return+(i?i.value:r&&(i=e.config.style[r])&&i[n])}t.default=function(e,t){var n,v,_,y,b,j,x,O,w,C,E=e.type||"symbol",S=t.config.legend,k=e.encode||{},M=k.legend||{},A=M.name||void 0,T=M.interactive,D=M.style,I=e.size||e.shape||e.fill||e.stroke||e.strokeDash||e.opacity;return I||Object(m.error)("Missing valid scale for legend."),n={orient:Object(f.value)(e.orient,S.orient),title:null!=e.title},v=Object(f.ref)(t.add(Object(h.Collect)(null,[n]))),M=Object(d.extendEncode)({enter:(w=S,C={},Object(d.addEncode)(C,"fill",w.fillColor)+Object(d.addEncode)(C,"stroke",w.strokeColor)+Object(d.addEncode)(C,"strokeWidth",w.strokeWidth)+Object(d.addEncode)(C,"strokeDash",w.strokeDash)+Object(d.addEncode)(C,"cornerRadius",w.cornerRadius)?C:void 0),update:{offset:Object(d.encoder)(Object(f.value)(e.offset,S.offset)),padding:Object(d.encoder)(Object(f.value)(e.padding,S.padding)),titlePadding:Object(d.encoder)(Object(f.value)(e.titlePadding,S.titlePadding))}},M,p.Skip),j={update:{x:{field:{group:"padding"}},y:{field:{group:"padding"}},entryPadding:Object(d.encoder)(Object(f.value)(e.entryPadding,S.entryPadding))}},"gradient"===E?(_=Object(f.ref)(t.add(Object(h.LegendEntries)({type:"gradient",scale:t.scaleRef(I),count:t.objectProperty(e.tickCount),values:t.objectProperty(e.values),formatSpecifier:t.property(e.format)}))),O=[Object(r.default)(e,I,S,k.gradient),Object(i.default)(e,S,k.labels,_)]):(_=Object(f.ref)(t.add(Object(h.LegendEntries)(x={scale:t.scaleRef(I),count:t.objectProperty(e.tickCount),values:t.objectProperty(e.values),formatSpecifier:t.property(e.format)}))),O=[Object(a.default)(e,S,k.symbols,_),Object(o.default)(e,S,k.labels,_)],x.size=function(e,t,n){var r=g(t,n[1].encode,"fontSize",p.GuideLabelStyle);if(e.size)return{$expr:"Math.max(Math.ceil(Math.sqrt(_.scale(datum))),"+r+")"};var i=g(t,n[0].encode,"size");return Math.max(Math.ceil(Math.sqrt(i)),r)}(e,t,O)),O=[Object(u.default)(l.LegendEntryRole,null,null,v,T,j,O)],n.title&&(b=Object(s.default)(e,S,k.title,v),j.update.y.offset={field:{group:"titlePadding"},offset:g(t,b.encode,"fontSize",p.GuideTitleStyle)},O.push(b)),y=Object(u.default)(l.LegendRole,D,A,v,T,M,O),e.zindex&&(y.zindex=e.zindex),Object(c.default)(y,t)}},"./node_modules/vega-parser/src/parsers/mark.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/marks/clip.js"),i=n("./node_modules/vega-parser/src/parsers/marks/definition.js"),o=n("./node_modules/vega-parser/src/parsers/marks/interactive.js"),a=n("./node_modules/vega-parser/src/parsers/marks/data.js"),s=n("./node_modules/vega-parser/src/parsers/marks/facet.js"),u=n("./node_modules/vega-parser/src/parsers/marks/subflow.js"),c=n("./node_modules/vega-parser/src/parsers/marks/role.js"),l=n("./node_modules/vega-parser/src/parsers/marks/marktypes.js"),d=n("./node_modules/vega-parser/src/parsers/marks/roles.js"),p=n("./node_modules/vega-parser/src/parsers/encode/encode-util.js"),f=n("./node_modules/vega-parser/src/parsers/transform.js"),h=n("./node_modules/vega-parser/src/parsers/trigger.js"),m=n("./node_modules/vega-parser/src/parsers/spec.js"),g=n("./node_modules/vega-parser/src/DataScope.js"),v=n("./node_modules/vega-parser/src/util.js"),_=n("./node_modules/vega-util/index.js"),y=n("./node_modules/vega-parser/src/transforms.js");t.default=function(e,t){var n,b,j,x,O,w,C,E,S,k,M,A,T,D=Object(c.default)(e),I=e.type===l.GroupMark,P=e.from&&e.from.facet,R=e.layout||D===d.ScopeRole||D===d.FrameRole,N=D===d.MarkRole||R||P,F=e.overlap;j=Object(a.default)(e.from,I,t),b=t.add(Object(y.DataJoin)({key:j.key||(e.key?Object(v.fieldRef)(e.key):void 0),pulse:j.pulse,clean:!I})),S=Object(v.ref)(b),b=x=t.add(Object(y.Collect)({pulse:S})),b=t.add(Object(y.Mark)({markdef:Object(i.default)(e),interactive:Object(o.default)(e.interactive,t),clip:Object(r.default)(e.clip,t),context:{$context:!0},groups:t.lookup(),parent:t.signals.parent?t.signalRef("parent"):null,index:t.markpath(),pulse:Object(v.ref)(b)})),k=Object(v.ref)(b),(b=t.add(Object(y.Encode)(Object(p.encoders)(e.encode,e.type,D,e.style,t,{pulse:k})))).params.parent=t.encode(),e.transform&&e.transform.forEach(function(e){var n=Object(f.default)(e,t);(n.metadata.generates||n.metadata.changes)&&Object(_.error)("Mark transforms should not generate new data."),n.params.pulse=Object(v.ref)(b),t.add(b=n)}),e.sort&&(b=t.add(Object(y.SortItems)({sort:t.compareRef(e.sort,!0),pulse:Object(v.ref)(b)}))),M=Object(v.ref)(b),(P||R)&&(R=t.add(Object(y.ViewLayout)({layout:t.objectProperty(e.layout),legendMargin:t.config.legendMargin,mark:k,pulse:M})),A=Object(v.ref)(R)),O=t.add(Object(y.Bound)({mark:k,pulse:A||M})),T=Object(v.ref)(O),I&&(N&&((n=t.operators).pop(),R&&n.pop()),t.pushState(M,A||T,S),P?Object(s.default)(e,t,j):N?Object(u.default)(e,t,j):Object(m.default)(e,t),t.popState(),N&&(R&&n.push(R),n.push(O))),F&&(b={method:!0===F.method?"parity":F.method,pulse:T},F.order&&(b.sort=t.compareRef({field:F.order})),F.bound&&(b.boundScale=t.scaleRef(F.bound.scale),b.boundOrient=F.bound.orient,b.boundTolerance=F.bound.tolerance),T=Object(v.ref)(t.add(Object(y.Overlap)(b)))),w=t.add(Object(y.Render)({pulse:T})),C=t.add(Object(y.Sieve)({pulse:Object(v.ref)(w)},void 0,t.parent())),null!=e.name&&(E=e.name,t.addData(E,new g.default(t,x,w,C)),e.on&&e.on.forEach(function(e){(e.insert||e.remove||e.toggle)&&Object(_.error)("Marks only support modify triggers."),Object(h.default)(e,t,E)}))}},"./node_modules/vega-parser/src/parsers/marks/clip.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/index.js");function i(e){return Object(r.isObject)(e)&&e.signal?e.signal:Object(r.stringValue)(e)}t.default=function(e,t){var n;return Object(r.isObject)(e)&&(e.signal?n=e.signal:e.path?n="pathShape("+i(e.path)+")":e.sphere&&(n="geoShape("+i(e.sphere)+', {type: "Sphere"})')),n?t.signalRef(n):!!e}},"./node_modules/vega-parser/src/parsers/marks/data.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/transform.js"),i=n("./node_modules/vega-parser/src/util.js"),o=n("./node_modules/vega-parser/src/transforms.js"),a=n("./node_modules/vega-util/index.js");t.default=function(e,t,n){var s,u,c,l,d;return e?(s=e.facet)&&(t||Object(a.error)("Only group marks can be faceted."),null!=s.field?l=d=Object(i.ref)(n.getData(s.data).output):(e.data?d=Object(i.ref)(n.getData(e.data).aggregate):((c=Object(r.default)(Object(a.extend)({type:"aggregate",groupby:Object(a.array)(s.groupby)},s.aggregate),n)).params.key=n.keyRef(s.groupby),c.params.pulse=Object(i.ref)(n.getData(s.data).output),l=d=Object(i.ref)(n.add(c))),u=n.keyRef(s.groupby,!0))):l=Object(i.ref)(n.add(Object(o.Collect)(null,[{}]))),l||(l=e.$ref?e:Object(i.ref)(n.getData(e.data).output)),{key:u,pulse:l,parent:d}}},"./node_modules/vega-parser/src/parsers/marks/definition.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/marks/role.js");t.default=function(e){return{marktype:e.type,name:e.name||void 0,role:e.role||Object(r.default)(e),zindex:+e.zindex||void 0}}},"./node_modules/vega-parser/src/parsers/marks/facet.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/spec.js"),i=n("./node_modules/vega-parser/src/DataScope.js"),o=n("./node_modules/vega-parser/src/util.js"),a=n("./node_modules/vega-parser/src/transforms.js"),s=n("./node_modules/vega-util/index.js");t.default=function(e,t,n){var u,c,l,d,p=e.from.facet,f=p.name,h=Object(o.ref)(t.getData(p.data).output);p.name||Object(s.error)("Facet must have a name: "+Object(s.stringValue)(p)),p.data||Object(s.error)("Facet must reference a data set: "+Object(s.stringValue)(p)),p.field?d=t.add(Object(a.PreFacet)({field:t.fieldRef(p.field),pulse:h})):p.groupby?d=t.add(Object(a.Facet)({key:t.keyRef(p.groupby),group:Object(o.ref)(t.proxy(n.parent)),pulse:h})):Object(s.error)("Facet must specify groupby or field: "+Object(s.stringValue)(p)),c=(u=t.fork()).add(Object(a.Collect)()),l=u.add(Object(a.Sieve)({pulse:Object(o.ref)(c)})),u.addData(f,new i.default(u,c,c,l)),u.addSignal("parent",null),d.params.subflow={$subflow:Object(r.default)(e,u).toRuntime()}}},"./node_modules/vega-parser/src/parsers/marks/interactive.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){return e&&e.signal?t.signalRef(e.signal):!1!==e}},"./node_modules/vega-parser/src/parsers/marks/marktypes.js":function(e,t,n){"use strict";n.r(t),n.d(t,"GroupMark",function(){return r}),n.d(t,"RectMark",function(){return i}),n.d(t,"RuleMark",function(){return o}),n.d(t,"SymbolMark",function(){return a}),n.d(t,"TextMark",function(){return s});var r="group",i="rect",o="rule",a="symbol",s="text"},"./node_modules/vega-parser/src/parsers/marks/role.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/marks/marktypes.js"),i=n("./node_modules/vega-parser/src/parsers/marks/roles.js");t.default=function(e){var t=e.role||"";return t.indexOf("axis")&&t.indexOf("legend")?e.type===r.GroupMark?i.ScopeRole:t||i.MarkRole:t}},"./node_modules/vega-parser/src/parsers/marks/roles.js":function(e,t,n){"use strict";n.r(t),n.d(t,"MarkRole",function(){return r}),n.d(t,"FrameRole",function(){return i}),n.d(t,"ScopeRole",function(){return o}),n.d(t,"AxisRole",function(){return a}),n.d(t,"AxisDomainRole",function(){return s}),n.d(t,"AxisGridRole",function(){return u}),n.d(t,"AxisLabelRole",function(){return c}),n.d(t,"AxisTickRole",function(){return l}),n.d(t,"AxisTitleRole",function(){return d}),n.d(t,"LegendRole",function(){return p}),n.d(t,"LegendEntryRole",function(){return f}),n.d(t,"LegendGradientRole",function(){return h}),n.d(t,"LegendLabelRole",function(){return m}),n.d(t,"LegendSymbolRole",function(){return g}),n.d(t,"LegendTitleRole",function(){return v}),n.d(t,"TitleRole",function(){return _});var r="mark",i="frame",o="scope",a="axis",s="axis-domain",u="axis-grid",c="axis-label",l="axis-tick",d="axis-title",p="legend",f="legend-entry",h="legend-gradient",m="legend-label",g="legend-symbol",v="legend-title",_="title"},"./node_modules/vega-parser/src/parsers/marks/subflow.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/spec.js"),i=n("./node_modules/vega-parser/src/transforms.js");t.default=function(e,t,n){var o=t.add(Object(i.PreFacet)({pulse:n.pulse})),a=t.fork();a.add(Object(i.Sieve)()),a.addSignal("parent",null),o.params.subflow={$subflow:Object(r.default)(e,a).toRuntime()}}},"./node_modules/vega-parser/src/parsers/padding.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/index.js");function i(e){return+e||0}t.default=function(e,t){return e=e||t.padding,Object(r.isObject)(e)?{top:i(e.top),bottom:i(e.bottom),left:i(e.left),right:i(e.right)}:{top:n=i(e),bottom:n,left:n,right:n};var n}},"./node_modules/vega-parser/src/parsers/projection.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/index.js");function i(e,t){return Object(r.isArray)(e)?e.map(function(e){return i(e,t)}):Object(r.isObject)(e)?e.signal?t.signalRef(e.signal):Object(r.error)("Unsupported parameter object: "+Object(r.stringValue)(e)):e}t.default=function(e,t){var n={};for(var r in e)"name"!==r&&(n[r]=i(e[r],t));t.addProjection(e.name,n)}},"./node_modules/vega-parser/src/parsers/scale.js":function(e,t,n){"use strict";n.r(t),n.d(t,"isOrdinal",function(){return l}),n.d(t,"isQuantile",function(){return d}),n.d(t,"initScale",function(){return p}),n.d(t,"parseScale",function(){return f});var r=n("./node_modules/vega-parser/src/util.js"),i=n("./node_modules/vega-parser/src/transforms.js"),o=n("./node_modules/vega-util/index.js"),a=0,s=["identity","ordinal","band","point","bin-linear","bin-ordinal","linear","pow","sqrt","log","sequential","time","utc","quantize","quantile","threshold"],u=Object(o.toSet)(s),c=Object(o.toSet)(s.slice(1,6));function l(e){return c.hasOwnProperty(e)}function d(e){return"quantile"===e}function p(e,t){var n=e.type||"linear";u.hasOwnProperty(n)||Object(o.error)("Unrecognized scale type: "+Object(o.stringValue)(n)),t.addScale(e.name,{type:n,domain:void 0})}function f(e,t){var n,r,i,a,s=t.getScale(e.name).params;s.domain=g(e.domain,e,t),null!=e.range&&(s.range=function e(t,n,r){var i=t.range,a=n.config.range;if(i.signal)return n.signalRef(i.signal);if(Object(o.isString)(i)){if(a&&a.hasOwnProperty(i))return t=Object(o.extend)({},t,{range:a[i]}),e(t,n,r);"width"===i?i=[0,{signal:"width"}]:"height"===i?i=l(t.type)?[0,{signal:"height"}]:[{signal:"height"},0]:Object(o.error)("Unrecognized scale range value: "+Object(o.stringValue)(i))}else{if(i.scheme)return r.scheme=h(i.scheme,n),i.extent&&(r.schemeExtent=(s=i.extent,u=n,s.signal?u.signalRef(s.signal):s.map(function(e){return h(e,u)}))),void(i.count&&(r.schemeCount=h(i.count,n)));if(i.step)return void(r.rangeStep=h(i.step,n));if(l(t.type)&&!Object(o.isArray)(i))return g(i,t,n);Object(o.isArray)(i)||Object(o.error)("Unsupported range type: "+Object(o.stringValue)(i))}var s,u;return i.map(function(e){return h(e,n)})}(e,t,s)),null!=e.interpolate&&(r=e.interpolate,(i=s).interpolate=h(r.type||r),null!=r.gamma&&(i.interpolateGamma=h(r.gamma))),null!=e.nice&&(a=e.nice,s.nice=Object(o.isObject)(a)?{interval:h(a.interval),step:h(a.step)}:h(a));for(n in e)s.hasOwnProperty(n)||"name"===n||(s[n]=h(e[n],t))}function h(e,t){return Object(o.isObject)(e)?e.signal?t.signalRef(e.signal):Object(o.error)("Unsupported object: "+Object(o.stringValue)(e)):e}function m(e){Object(o.error)("Can not find data set: "+Object(o.stringValue)(e))}function g(e,t,n){if(e)return e.signal?n.signalRef(e.signal):(Object(o.isArray)(e)?function(e,t,n){return e.map(function(e){return h(e,n)})}:e.fields?function(e,t,n){var s=e.data,u=e.fields.reduce(function(e,t){return t=Object(o.isString)(t)?{data:s,field:t}:Object(o.isArray)(t)||t.signal?function(e,t){var n="_:vega:_"+a++,r=Object(i.Collect)({});if(Object(o.isArray)(e))r.value={$ingest:e};else if(e.signal){var s="setdata("+Object(o.stringValue)(n)+","+e.signal+")";r.params.input=t.signalRef(s)}return t.addDataPipeline(n,[r,Object(i.Sieve)({})]),{data:n,field:"data"}}(t,n):t,e.push(t),e},[]);return(l(t.type)?function(e,t,n){var o,a,s,u;return o=n.map(function(e){var n=t.getData(e.data);return n||m(e.data),n.countsRef(t,e.field)}),a=t.add(Object(i.Aggregate)({groupby:r.keyFieldRef,ops:["sum"],fields:[t.fieldRef("count")],as:["count"],pulse:o})),s=t.add(Object(i.Collect)({pulse:Object(r.ref)(a)})),u=t.add(Object(i.Values)({field:r.keyFieldRef,sort:t.sortRef(v(e.sort,!0)),pulse:Object(r.ref)(s)})),Object(r.ref)(u)}:d(t.type)?function(e,t,n){var o=n.map(function(e){var n=t.getData(e.data);return n||m(e.data),n.domainRef(t,e.field)});return Object(r.ref)(t.add(Object(i.MultiValues)({values:o})))}:function(e,t,n){var o=n.map(function(e){var n=t.getData(e.data);return n||m(e.data),n.extentRef(t,e.field)});return Object(r.ref)(t.add(Object(i.MultiExtent)({extents:o})))})(e,n,u)}:function(e,t,n){var r=n.getData(e.data);r||m(e.data);return l(t.type)?r.valuesRef(n,e.field,v(e.sort,!1)):d(t.type)?r.domainRef(n,e.field):r.extentRef(n,e.field)})(e,t,n);null==t.domainMin&&null==t.domainMax||Object(o.error)("No scale domain defined for domainMin/domainMax to override.")}function v(e,t){return e&&(e.field||e.op?e.field||"count"===e.op?t&&e.field?Object(o.error)("Multiple domain scales can not sort by field."):t&&e.op&&"count"!==e.op&&Object(o.error)("Multiple domain scales support op count only."):Object(o.error)("No field provided for sort aggregate op: "+e.op):Object(o.isObject)(e)?e.field="key":e={field:"key"}),e}},"./node_modules/vega-parser/src/parsers/signal-updates.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/expression.js"),i=n("./node_modules/vega-parser/src/parsers/update.js");t.default=function(e,t){var n=t.getSignal(e.name);if(e.update){var o=Object(r.default)(e.update,t);n.update=o.$expr,n.params=o.$params}e.on&&e.on.forEach(function(e){Object(i.default)(e,t,n.id)})}},"./node_modules/vega-parser/src/parsers/signal.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/index.js"),i=["value","update","react","bind"];function o(e,t){Object(r.error)(e+' for "outer" push: '+Object(r.stringValue)(t))}t.default=function(e,t){var n=e.name;if("outer"===e.push)t.signals[n]||o("No prior signal definition",n),i.forEach(function(t){void 0!==e[t]&&o("Invalid property ",t)});else{var r=t.addSignal(n,e.value);!1===e.react&&(r.react=!1),e.bind&&t.addBinding(n,e.bind)}}},"./node_modules/vega-parser/src/parsers/spec.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/signal-updates.js"),i=n("./node_modules/vega-parser/src/parsers/scale.js"),o=n("./node_modules/vega-parser/src/parsers/projection.js"),a=n("./node_modules/vega-parser/src/parsers/legend.js"),s=n("./node_modules/vega-parser/src/parsers/signal.js"),u=n("./node_modules/vega-parser/src/parsers/title.js"),c=n("./node_modules/vega-parser/src/parsers/data.js"),l=n("./node_modules/vega-parser/src/parsers/mark.js"),d=n("./node_modules/vega-parser/src/parsers/axis.js"),p=n("./node_modules/vega-util/index.js");t.default=function(e,t,n){var f=Object(p.array)(e.signals),h=Object(p.array)(e.scales);return n||f.forEach(function(e){Object(s.default)(e,t)}),Object(p.array)(e.projections).forEach(function(e){Object(o.default)(e,t)}),h.forEach(function(e){Object(i.initScale)(e,t)}),Object(p.array)(e.data).forEach(function(e){Object(c.default)(e,t)}),h.forEach(function(e){Object(i.parseScale)(e,t)}),f.forEach(function(e){Object(r.default)(e,t)}),Object(p.array)(e.axes).forEach(function(e){Object(d.default)(e,t)}),Object(p.array)(e.marks).forEach(function(e){Object(l.default)(e,t)}),Object(p.array)(e.legends).forEach(function(e){Object(a.default)(e,t)}),e.title&&Object(u.default)(e.title,t),t.parseLambdas(),t}},"./node_modules/vega-parser/src/parsers/stream.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/expression.js"),i=n("./node_modules/vega-util/index.js"),o="view",a="scope";function s(e,t){return(e.merge?u:e.stream?c:e.type?l:Object(i.error)("Invalid stream specification: "+Object(i.stringValue)(e)))(e,t)}function u(e,t){var n=d({merge:e.merge.map(function(e){return s(e,t)})},e,t);return t.addStream(n).id}function c(e,t){var n=d({stream:s(e.stream,t)},e,t);return t.addStream(n).id}function l(e,t){var n,r=t.event((n=e.source)===a?o:n||o,e.type),i=d({stream:r},e,t);return 1===Object.keys(i).length?r:t.addStream(i).id}function d(e,t,n){var o,u,c,l=t.between;return l&&(2!==l.length&&Object(i.error)('Stream "between" parameter must have 2 entries: '+Object(i.stringValue)(t)),e.between=[s(l[0],n),s(l[1],n)]),l=t.filter?Object(i.array)(t.filter):[],(t.marktype||t.markname||t.markrole)&&l.push((o=t.marktype,u=t.markname,c=t.markrole,"event.item"+(o&&"*"!==o?"&&event.item.mark.marktype==='"+o+"'":"")+(c?"&&event.item.mark.role==='"+c+"'":"")+(u?"&&event.item.mark.name==='"+u+"'":""))),t.source===a&&l.push("inScope(event.item)"),l.length&&(e.filter=Object(r.default)("("+l.join(")&&(")+")").$expr),null!=(l=t.throttle)&&(e.throttle=+l),null!=(l=t.debounce)&&(e.debounce=+l),t.consume&&(e.consume=!0),e}t.default=function(e,t){return e.signal?t.getSignal(e.signal).id:e.scale?t.getScale(e.scale).id:s(e,t)}},"./node_modules/vega-parser/src/parsers/title.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/guides/constants.js"),i=n("./node_modules/vega-parser/src/parsers/guides/guide-mark.js"),o=n("./node_modules/vega-parser/src/parsers/mark.js"),a=n("./node_modules/vega-parser/src/parsers/marks/marktypes.js"),s=n("./node_modules/vega-parser/src/parsers/marks/roles.js"),u=n("./node_modules/vega-parser/src/parsers/encode/encode-util.js"),c=n("./node_modules/vega-parser/src/util.js"),l=n("./node_modules/vega-parser/src/transforms.js"),d=n("./node_modules/vega-util/index.js");t.default=function(e,t){e=Object(d.isString)(e)?{text:e}:e;var n,p,f,h=t.config.title,m=Object(d.extend)({},e.encode);return n={orient:null!=e.orient?e.orient:h.orient},p=Object(c.ref)(t.add(Object(l.Collect)(null,[n]))),m.name=e.name,m.interactive=e.interactive,f=function(e,t,n,o){var c,l,p,f,h,m,g=e.text,v=e.orient||t.orient,_=e.anchor||t.anchor,y=v===r.Left||v===r.Top?-1:1,b=v===r.Top||v===r.Bottom,j={group:b?"width":"height"},x={};x.enter=c={opacity:{value:0}},Object(u.addEncode)(c,"fill",t.color),Object(u.addEncode)(c,"font",t.font),Object(u.addEncode)(c,"fontSize",t.fontSize),Object(u.addEncode)(c,"fontWeight",t.fontWeight),x.exit={opacity:{value:0}},x.update=l={opacity:{value:1},text:Object(d.isObject)(g)?g:{value:g+""},offset:Object(u.encoder)((null!=e.offset?e.offset:t.offset)||0)},"start"===_?(h=0,m="left"):"end"===_?(h=1,m="right"):(h=.5,m="center");p={field:j,mult:h},f=y<0?{value:0}:b?{field:{group:"height"}}:{field:{group:"width"}},b?(l.x=p,l.y=f,l.angle={value:0},l.baseline={value:v===r.Top?"bottom":"top"}):(l.x=f,l.y=p,l.angle={value:90*y},l.baseline={value:"bottom"});return l.align={value:m},l.limit={field:j},Object(u.addEncode)(l,"angle",t.angle),Object(u.addEncode)(l,"baseline",t.baseline),Object(u.addEncode)(l,"limit",t.limit),Object(i.default)(a.TextMark,s.TitleRole,e.style||r.GroupTitleStyle,null,o,x,n)}(e,h,m,p),e.zindex&&(f.zindex=e.zindex),Object(o.default)(f,t)}},"./node_modules/vega-parser/src/parsers/transform.js":function(e,t,n){"use strict";n.r(t),n.d(t,"outerExpr",function(){return p}),n.d(t,"outerField",function(){return f}),n.d(t,"isData",function(){return h}),n.d(t,"isExpr",function(){return m}),n.d(t,"isField",function(){return g}),n.d(t,"isCompare",function(){return v});var r=n("./node_modules/vega-parser/src/parsers/expression.js"),i=n("./node_modules/vega-parser/src/util.js"),o=n("./node_modules/vega-parser/src/transforms.js"),a=n("./node_modules/vega-dataflow/index.js"),s=n("./node_modules/vega-util/index.js");function u(e,t,n){var r,i,o,a={};for(i=0,o=e.params.length;i<o;++i)a[(r=e.params[i]).name]=c(r,t,n);return a}function c(e,t,n){var r,o,a,u=e.type,c=t[e.name];if("index"===u)return function(e,t,n){Object(s.isString)(t.from)||Object(s.error)('Lookup "from" parameter must be a string literal.');return n.getData(t.from).lookupRef(n,t.key)}(0,t,n);{if(void 0!==c)return"param"===u?(o=n,a=t[(r=e).name],r.array?(Object(s.isArray)(a)||Object(s.error)("Expected an array of sub-parameters. Instead: "+Object(s.stringValue)(a)),a.map(function(e){return d(r,e,o)})):d(r,a,o)):"projection"===u?n.projectionRef(t[e.name]):e.array&&!Object(i.isSignal)(c)?c.map(function(t){return l(e,t,n)}):l(e,c,n);e.required&&Object(s.error)("Missing required "+Object(s.stringValue)(t.type)+" parameter: "+Object(s.stringValue)(e.name))}}function l(e,t,n){var o=e.type;if(Object(i.isSignal)(t))return m(o)?Object(s.error)("Expression references can not be signals."):g(o)?n.fieldRef(t):v(o)?n.compareRef(t):n.signalRef(t.signal);var a=e.expr||g(o);return a&&p(t)?Object(r.default)(t.expr,n):a&&f(t)?Object(i.fieldRef)(t.field):m(o)?Object(r.default)(t,n):h(o)?Object(i.ref)(n.getData(t).values):g(o)?Object(i.fieldRef)(t):v(o)?n.compareRef(t):t}function d(e,t,n){var r,a,c,l,d;for(l=0,d=e.params.length;l<d;++l){a=e.params[l];for(c in a.key)if(a.key[c]!==t[c]){a=null;break}if(a)break}return a||Object(s.error)("Unsupported parameter: "+Object(s.stringValue)(t)),r=Object(s.extend)(u(a,t,n),a.key),Object(i.ref)(n.add(Object(o.Params)(r)))}function p(e){return e&&e.expr}function f(e){return e&&e.field}function h(e){return"data"===e}function m(e){return"expr"===e}function g(e){return"field"===e}function v(e){return"compare"===e}t.default=function(e,t){var n=Object(a.definition)(e.type);n||Object(s.error)("Unrecognized transform type: "+Object(s.stringValue)(e.type));var r=Object(i.entry)(n.type.toLowerCase(),null,u(n,e,t));return e.signal&&t.addSignal(e.signal,t.proxy(r)),r.metadata=n.metadata||{},r}},"./node_modules/vega-parser/src/parsers/trigger.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/expression.js"),i=n("./node_modules/vega-parser/src/util.js");t.default=function(e,t,n){var o,a,s=e.remove,u=e.insert,c=e.toggle,l=e.modify,d=e.values,p=t.add(Object(i.operator)());o="if("+e.trigger+',modify("'+n+'",'+[u,s,c,l,d].map(function(e){return null==e?"null":e}).join(",")+"),0)",a=Object(r.default)(o,t),p.update=a.$expr,p.params=a.$params}},"./node_modules/vega-parser/src/parsers/update.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-parser/src/parsers/expression.js"),i=n("./node_modules/vega-parser/src/parsers/stream.js"),o=n("./node_modules/vega-event-selector/index.js"),a=n("./node_modules/vega-util/index.js"),s="var datum=event.item&&event.item.datum;";t.default=function(e,t,n){var u,c,l=e.events,d=e.update,p=e.encode,f=[];l||Object(a.error)("Signal update missing events specification."),Object(a.isString)(l)&&(l=Object(o.selector)(l)),(l=Object(a.array)(l).filter(function(e){return e.signal||e.scale?(f.push(e),0):1})).length&&f.push(l.length>1?{merge:l}:l[0]),null!=p&&(d&&Object(a.error)("Signal encode and update are mutually exclusive."),d="encode(item(),"+Object(a.stringValue)(p)+")"),u=Object(a.isString)(d)?Object(r.default)(d,t,s):null!=d.expr?Object(r.default)(d.expr,t,s):null!=d.value?d.value:null!=d.signal?{$expr:"_.value",$params:{value:t.signalRef(d.signal)}}:Object(a.error)("Invalid signal update specification."),c={target:n,update:u},e.force&&(c.options={force:!0}),f.forEach(function(e){e={source:Object(i.default)(e,t)},t.addUpdate(Object(a.extend)(e,c))})}},"./node_modules/vega-parser/src/parsers/view.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return m});var r=n("./node_modules/vega-parser/src/parsers/autosize.js"),i=n("./node_modules/vega-parser/src/parsers/padding.js"),o=n("./node_modules/vega-parser/src/parsers/signal.js"),a=n("./node_modules/vega-parser/src/parsers/spec.js"),s=n("./node_modules/vega-parser/src/parsers/encode/encode-util.js"),u=n("./node_modules/vega-parser/src/parsers/marks/marktypes.js"),c=n("./node_modules/vega-parser/src/parsers/marks/roles.js"),l=n("./node_modules/vega-parser/src/util.js"),d=n("./node_modules/vega-parser/src/DataScope.js"),p=n("./node_modules/vega-parser/src/transforms.js"),f=n("./node_modules/vega-util/index.js"),h=Object(f.toSet)(["width","height","padding","autosize"]);function m(e,t){var n,m,g,v,_,y=t.config;return t.background=e.background||y.background,t.eventConfig=y.events,_=Object(l.ref)(t.root=t.add(Object(l.operator)())),t.addSignal("width",e.width||0),t.addSignal("height",e.height||0),t.addSignal("padding",Object(i.default)(e.padding,y)),t.addSignal("autosize",Object(r.default)(e.autosize,y)),Object(f.array)(e.signals).forEach(function(e){h[e.name]||Object(o.default)(e,t)}),m=t.add(Object(p.Collect)()),g=Object(s.extendEncode)({enter:{x:{value:0},y:{value:0}},update:{width:{signal:"width"},height:{signal:"height"}}},e.encode),g=t.add(Object(p.Encode)(Object(s.encoders)(g,u.GroupMark,c.FrameRole,e.style,t,{pulse:Object(l.ref)(m)}))),v=t.add(Object(p.ViewLayout)({layout:t.objectProperty(e.layout),legendMargin:y.legendMargin,autosize:t.signalRef("autosize"),mark:_,pulse:Object(l.ref)(g)})),t.operators.pop(),t.pushState(Object(l.ref)(g),Object(l.ref)(v),null),Object(a.default)(e,t,!0),t.operators.push(v),n=t.add(Object(p.Bound)({mark:_,pulse:Object(l.ref)(v)})),n=t.add(Object(p.Render)({pulse:Object(l.ref)(n)})),n=t.add(Object(p.Sieve)({pulse:Object(l.ref)(n)})),t.addData("root",new d.default(t,m,m,n)),t}},"./node_modules/vega-parser/src/transforms.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Aggregate",function(){return o}),n.d(t,"AxisTicks",function(){return a}),n.d(t,"Bound",function(){return s}),n.d(t,"Collect",function(){return u}),n.d(t,"Compare",function(){return c}),n.d(t,"DataJoin",function(){return l}),n.d(t,"Encode",function(){return d}),n.d(t,"Extent",function(){return p}),n.d(t,"Facet",function(){return f}),n.d(t,"Field",function(){return h}),n.d(t,"Key",function(){return m}),n.d(t,"LegendEntries",function(){return g}),n.d(t,"Mark",function(){return v}),n.d(t,"MultiExtent",function(){return _}),n.d(t,"MultiValues",function(){return y}),n.d(t,"Overlap",function(){return b}),n.d(t,"Params",function(){return j}),n.d(t,"PreFacet",function(){return x}),n.d(t,"Projection",function(){return O}),n.d(t,"Proxy",function(){return w}),n.d(t,"Relay",function(){return C}),n.d(t,"Render",function(){return E}),n.d(t,"Scale",function(){return S}),n.d(t,"Sieve",function(){return k}),n.d(t,"SortItems",function(){return M}),n.d(t,"ViewLayout",function(){return A}),n.d(t,"Values",function(){return T});var r=n("./node_modules/vega-parser/src/util.js");function i(e){return function(t,n,i){return Object(r.entry)(e,n,t||void 0,i)}}var o=i("aggregate"),a=i("axisticks"),s=i("bound"),u=i("collect"),c=i("compare"),l=i("datajoin"),d=i("encode"),p=i("extent"),f=i("facet"),h=i("field"),m=i("key"),g=i("legendentries"),v=i("mark"),_=i("multiextent"),y=i("multivalues"),b=i("overlap"),j=i("params"),x=i("prefacet"),O=i("projection"),w=i("proxy"),C=i("relay"),E=i("render"),S=i("scale"),k=i("sieve"),M=i("sortitems"),A=i("viewlayout"),T=i("values")},"./node_modules/vega-parser/src/util.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Entry",function(){return i}),n.d(t,"entry",function(){return o}),n.d(t,"operator",function(){return a}),n.d(t,"ref",function(){return s}),n.d(t,"tupleidRef",function(){return u}),n.d(t,"fieldRef",function(){return c}),n.d(t,"keyFieldRef",function(){return l}),n.d(t,"compareRef",function(){return d}),n.d(t,"keyRef",function(){return p}),n.d(t,"Ascending",function(){return f}),n.d(t,"Descending",function(){return h}),n.d(t,"sortKey",function(){return m}),n.d(t,"aggrField",function(){return g}),n.d(t,"isSignal",function(){return v}),n.d(t,"value",function(){return _});var r=n("./node_modules/vega-util/index.js");function i(e,t,n,r){this.id=-1,this.type=e,this.value=t,this.params=n,r&&(this.parent=r)}function o(e,t,n,r){return new i(e,t,n,r)}function a(e,t){return o("operator",e,t)}function s(e){var t={$ref:e.id};return e.id<0&&(e.refs=e.refs||[]).push(t),t}var u={$tupleid:1,toString:function(){return":_tupleid_:"}};function c(e,t){return t?{$field:e,$name:t}:{$field:e}}var l=c("key");function d(e,t){return{$compare:e,$order:t}}function p(e,t){var n={$key:e};return t&&(n.$flat=!0),n}var f="ascending",h="descending";function m(e){return Object(r.isObject)(e)?(e.order===h?"-":"+")+g(e.op,e.field):""}function g(e,t){return(e&&e.signal?"$"+e.signal:e||"")+(e&&t?"_":"")+(t&&t.signal?"$"+t.signal:t||"")}function v(e){return e&&e.signal}function _(e,t){return null!=e?e:t}},"./node_modules/vega-projection/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-projection/src/projections.js");n.d(t,"projection",function(){return r.projection}),n.d(t,"projectionProperties",function(){return r.projectionProperties}),n.d(t,"getProjectionPath",function(){return r.getProjectionPath})},"./node_modules/vega-projection/src/projections.js":function(e,t,n){"use strict";n.r(t),n.d(t,"projectionProperties",function(){return o}),n.d(t,"projection",function(){return a}),n.d(t,"getProjectionPath",function(){return s});var r=n("./node_modules/d3-geo/index.js"),i=Object(r.geoPath)(),o=["clipAngle","clipExtent","scale","translate","center","rotate","parallels","precision","coefficient","distance","fraction","lobes","parallel","radius","ratio","spacing","tilt"];function a(e,t){if(!e||"string"!=typeof e)throw new Error("Projection type must be a name string.");return e=e.toLowerCase(),arguments.length>1?(u[e]=(n=e,i=t,function e(){var t=i();return t.type=n,t.path=Object(r.geoPath)().projection(t),t.copy=t.copy||function(){var n=e();return o.forEach(function(e){t.hasOwnProperty(e)&&n[e](t[e]())}),n.path.pointRadius(t.path.pointRadius()),n},t}),this):u.hasOwnProperty(e)?u[e]:null;var n,i}function s(e){return e&&e.path||i}var u={albers:r.geoAlbers,albersusa:r.geoAlbersUsa,azimuthalequalarea:r.geoAzimuthalEqualArea,azimuthalequidistant:r.geoAzimuthalEquidistant,conicconformal:r.geoConicConformal,conicequalarea:r.geoConicEqualArea,conicequidistant:r.geoConicEquidistant,equirectangular:r.geoEquirectangular,gnomonic:r.geoGnomonic,mercator:r.geoMercator,orthographic:r.geoOrthographic,stereographic:r.geoStereographic,transversemercator:r.geoTransverseMercator};for(var c in u)a(c,u[c])},"./node_modules/vega-runtime/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-runtime/src/dataflow.js");n.d(t,"parse",function(){return r.default});var i=n("./node_modules/vega-runtime/src/context.js");n.d(t,"context",function(){return i.default});var o=n("./node_modules/vega-runtime/src/expression.js");n.d(t,"expression",function(){return o.default})},"./node_modules/vega-runtime/src/context.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-runtime/src/state.js"),i=n("./node_modules/vega-runtime/src/util.js");function o(e,t,n){this.dataflow=e,this.transforms=t,this.events=e.events.bind(e),this.signals={},this.scales={},this.nodes={},this.data={},this.fn={},n&&(this.functions=Object.create(n),this.functions.context=this)}function a(e){this.dataflow=e.dataflow,this.transforms=e.transforms,this.functions=e.functions,this.events=e.events,this.signals=Object.create(e.signals),this.scales=Object.create(e.scales),this.nodes=Object.create(e.nodes),this.data=Object.create(e.data),this.fn=Object.create(e.fn),e.functions&&(this.functions=Object.create(e.functions),this.functions.context=this)}t.default=function(e,t,n){return new o(e,t,n)},o.prototype=a.prototype={fork:function(){var e=new a(this);return(this.subcontext||(this.subcontext=[])).push(e),e},get:function(e){return this.nodes[e]},set:function(e,t){return this.nodes[e]=t},add:function(e,t){var n,r=this,o=r.dataflow;if(r.set(e.id,t),Object(i.isCollect)(e.type)&&(n=e.value)&&(n.$ingest?o.ingest(t,n.$ingest,n.$format):n.$request?o.request(t,n.$request,n.$format):o.pulse(t,o.changeset().insert(n))),e.root&&(r.root=t),e.parent){var a=r.get(e.parent.$ref);a?(o.connect(a,[t]),t.targets().add(a)):(r.unresolved=r.unresolved||[]).push(function(){a=r.get(e.parent.$ref),o.connect(a,[t]),t.targets().add(a)})}if(e.signal&&(r.signals[e.signal]=t),e.scale&&(r.scales[e.scale]=t),e.data)for(var s in e.data)n=r.data[s]||(r.data[s]={}),e.data[s].forEach(function(e){n[e]=t})},resolve:function(){return(this.unresolved||[]).forEach(function(e){e()}),delete this.unresolved,this},operator:function(e,t,n){this.add(e,this.dataflow.add(e.value,t,n,e.react))},transform:function(e,t,n){this.add(e,this.dataflow.add(this.transforms[Object(i.canonicalType)(t)],n))},stream:function(e,t){this.set(e.id,t)},update:function(e,t,n,r,i){this.dataflow.on(t,n,r,i,e.options)},getState:r.getState,setState:r.setState}},"./node_modules/vega-runtime/src/dataflow.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-runtime/src/operator.js"),i=n("./node_modules/vega-runtime/src/stream.js"),o=n("./node_modules/vega-runtime/src/update.js");t.default=function(e,t){var n=e.operators||[];return e.background&&(t.background=e.background),e.eventConfig&&(t.eventConfig=e.eventConfig),n.forEach(function(e){Object(r.default)(e,t)}),n.forEach(function(e){Object(r.parseOperatorParameters)(e,t)}),(e.streams||[]).forEach(function(e){Object(i.default)(e,t)}),(e.updates||[]).forEach(function(e){Object(o.default)(e,t)}),t.resolve()}},"./node_modules/vega-runtime/src/expression.js":function(e,t,n){"use strict";function r(e,t,n){";"!==t[t.length-1]&&(t="return("+t+");");var r=Function.apply(null,e.concat(t));return n&&n.functions?r.bind(n.functions):r}function i(e,t){return r(["_"],e,t)}function o(e,t){return r(["datum","_"],e,t)}function a(e,t){return r(["event"],e,t)}function s(e,t){return r(["_","event"],e,t)}function u(e,t){return r(["item","_"],e,t)}n.r(t),n.d(t,"default",function(){return r}),n.d(t,"operatorExpression",function(){return i}),n.d(t,"parameterExpression",function(){return o}),n.d(t,"eventExpression",function(){return a}),n.d(t,"handlerExpression",function(){return s}),n.d(t,"encodeExpression",function(){return u})},"./node_modules/vega-runtime/src/operator.js":function(e,t,n){"use strict";n.r(t),n.d(t,"parseOperatorParameters",function(){return s});var r=n("./node_modules/vega-runtime/src/expression.js"),i=n("./node_modules/vega-runtime/src/parameters.js"),o=n("./node_modules/vega-runtime/src/util.js"),a=n("./node_modules/vega-util/index.js");function s(e,t){var n,r;e.params&&((n=t.get(e.id))||Object(a.error)("Invalid operator id: "+e.id),r=Object(i.default)(e.params,t),t.dataflow.connect(n,n.parameters(r)))}t.default=function(e,t){Object(o.isOperator)(e.type)||!e.type?t.operator(e,e.update?Object(r.operatorExpression)(e.update,t):null):t.transform(e,e.type)}},"./node_modules/vega-runtime/src/parameters.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return s});var r=n("./node_modules/vega-runtime/src/dataflow.js"),i=n("./node_modules/vega-runtime/src/expression.js"),o=n("./node_modules/vega-dataflow/index.js"),a=n("./node_modules/vega-util/index.js");function s(e,t,n){var r,i;n=n||{};for(r in e)(i=e[r])&&i.$expr&&i.$params&&s(i.$params,t,n),n[r]=Object(a.isArray)(i)?i.map(function(e){return u(e,t)}):u(i,t);return n}function u(e,t){if(!e||!Object(a.isObject)(e))return e;for(var n,r=0,i=c.length;r<i;++r)if(n=c[r],e.hasOwnProperty(n.key))return n.parse(e,t);return e}var c=[{key:"$ref",parse:function(e,t){return t.get(e.$ref)||Object(a.error)("Operator not defined: "+e.$ref)}},{key:"$key",parse:function(e,t){var n="k:"+e.$key+"_"+!!e.$flat;return t.fn[n]||(t.fn[n]=Object(a.key)(e.$key,e.$flat))}},{key:"$expr",parse:function(e,t){var n="e:"+e.$expr;return t.fn[n]||(t.fn[n]=Object(a.accessor)(Object(i.parameterExpression)(e.$expr,t),e.$fields,e.$name))}},{key:"$field",parse:function(e,t){if(!e.$field)return null;var n="f:"+e.$field+"_"+e.$name;return t.fn[n]||(t.fn[n]=Object(a.field)(e.$field,e.$name))}},{key:"$encode",parse:function(e,t){var n,r,o=e.$encode,s={};for(n in o)r=o[n],s[n]=Object(a.accessor)(Object(i.encodeExpression)(r.$expr,t),r.$fields),s[n].output=r.$output;return s}},{key:"$compare",parse:function(e,t){var n="c:"+e.$compare+"_"+e.$order,r=Object(a.array)(e.$compare).map(function(e){return e&&e.$tupleid?o.tupleid:e});return t.fn[n]||(t.fn[n]=Object(a.compare)(r,e.$order))}},{key:"$context",parse:function(e,t){return t}},{key:"$subflow",parse:function(e,t){var n=e.$subflow;return function(e,i,o){var a=Object(r.default)(n,t.fork()),s=a.get(n.operators[0].id),u=a.signals.parent;return u&&u.set(o),s}}},{key:"$tupleid",parse:function(){return o.tupleid}}]},"./node_modules/vega-runtime/src/state.js":function(e,t,n){"use strict";n.r(t),n.d(t,"getState",function(){return o}),n.d(t,"setState",function(){return a});var r=n("./node_modules/vega-util/index.js"),i={skip:!0};function o(e){var t=this,n={};if(e.signals){var r=n.signals={};Object.keys(t.signals).forEach(function(n){var i=t.signals[n];e.signals(n,i)&&(r[n]=i.value)})}if(e.data){var i=n.data={};Object.keys(t.data).forEach(function(n){var r=t.data[n];e.data(n,r)&&(i[n]=r.input.value)})}return t.subcontext&&!1!==e.recurse&&(n.subcontext=t.subcontext.map(function(t){return t.getState(e)})),n}function a(e){var t=this,n=t.dataflow,o=e.data,a=e.signals;Object.keys(a||{}).forEach(function(e){n.update(t.signals[e],a[e],i)}),Object.keys(o||{}).forEach(function(e){n.pulse(t.data[e].input,n.changeset().remove(r.truthy).insert(o[e]))}),(e.subcontext||[]).forEach(function(e,n){var r=t.subcontext[n];r&&r.setState(e)})}},"./node_modules/vega-runtime/src/stream.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-runtime/src/expression.js"),i=n("./node_modules/vega-util/index.js");t.default=function(e,t){var n,o=null!=e.filter?Object(r.eventExpression)(e.filter,t):void 0,a=null!=e.stream?t.get(e.stream):void 0;e.source?a=t.events(e.source,e.type,o):e.merge&&(a=(n=e.merge.map(t.get.bind(t)))[0].merge.apply(n[0],n.slice(1))),e.between&&(n=e.between.map(t.get.bind(t)),a=a.between(n[0],n[1])),e.filter&&(a=a.filter(o)),null!=e.throttle&&(a=a.throttle(+e.throttle)),null!=e.debounce&&(a=a.debounce(+e.debounce)),null==a&&Object(i.error)("Invalid stream definition: "+JSON.stringify(e)),e.consume&&a.consume(!0),t.stream(e,a)}},"./node_modules/vega-runtime/src/update.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-runtime/src/expression.js"),i=n("./node_modules/vega-runtime/src/parameters.js"),o=n("./node_modules/vega-util/index.js");t.default=function(e,t){var n=t.get(e.source),a=null,s=e.update,u=void 0;n||Object(o.error)("Source not defined: "+e.source),a=e.target&&e.target.$expr?Object(r.eventExpression)(e.target.$expr,t):t.get(e.target),s&&s.$expr&&(s.$params&&(u=Object(i.default)(s.$params,t)),s=Object(r.handlerExpression)(s.$expr,t)),t.update(e,n,a,s,u)}},"./node_modules/vega-runtime/src/util.js":function(e,t,n){"use strict";function r(e){return(e+"").toLowerCase()}function i(e){return"operator"===r(e)}function o(e){return"collect"===r(e)}n.r(t),n.d(t,"canonicalType",function(){return r}),n.d(t,"isOperator",function(){return i}),n.d(t,"isCollect",function(){return o})},"./node_modules/vega-scale/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scale/src/scales.js");n.d(t,"scale",function(){return r.default});var i=n("./node_modules/vega-scale/src/schemes.js");n.d(t,"scheme",function(){return i.default});var o=n("./node_modules/vega-scale/src/bandSpace.js");n.d(t,"bandSpace",function(){return o.default});var a=n("./node_modules/vega-scale/src/interpolate.js");n.d(t,"interpolate",function(){return a.interpolate}),n.d(t,"interpolateRange",function(){return a.interpolateRange}),n.d(t,"scaleFraction",function(){return a.scaleFraction});var s=n("./node_modules/vega-scale/src/timeInterval.js");n.d(t,"timeInterval",function(){return s.timeInterval}),n.d(t,"utcInterval",function(){return s.utcInterval})},"./node_modules/vega-scale/src/bandSpace.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n){var r=e-t+2*n;return e?r>0?r:1:0}},"./node_modules/vega-scale/src/interpolate.js":function(e,t,n){"use strict";n.r(t),n.d(t,"interpolateRange",function(){return o}),n.d(t,"scaleFraction",function(){return a}),n.d(t,"interpolate",function(){return u});var r=n("./node_modules/vega-util/index.js"),i=n("./node_modules/d3-interpolate/index.js");function o(e,t){var n=t[0],i=Object(r.peek)(t)-n;return function(t){return e(n+t*i)}}function a(e,t,n){var i=n-t;return i?"linear"===e.type||"sequential"===e.type?function(e){return(e-t)/i}:e.copy().domain([t,n]).range([0,1]).interpolate(s):Object(r.constant)(0)}function s(e,t){var n=t-e;return function(t){return e+t*n}}function u(e,t){var n,r=i[(n=e,"interpolate"+n.toLowerCase().split("-").map(function(e){return e[0].toUpperCase()+e.slice(1)}).join(""))];return null!=t&&r&&r.gamma?r.gamma(t):r}},"./node_modules/vega-scale/src/invertRange.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return function(t){var n,r=t[0],i=t[1];return i<r&&(n=r,r=i,i=n),[e.invert(r),e.invert(i)]}}},"./node_modules/vega-scale/src/invertRangeExtent.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return function(t){var n,r,i,o,a=e.range(),s=t[0],u=t[1],c=-1;for(u<s&&(r=s,s=u,u=r),i=0,o=a.length;i<o;++i)a[i]>=s&&a[i]<=u&&(c<0&&(c=i),n=i);if(!(c<0))return s=e.invertExtent(a[c]),u=e.invertExtent(a[n]),[void 0===s[0]?s[1]:s[0],void 0===u[1]?u[0]:u[1]]}}},"./node_modules/vega-scale/src/palettes.js":function(e,t,n){"use strict";function r(e){for(var t=e.length/6|0,n=new Array(t),r=0;r<t;)n[r]="#"+e.slice(6*r,6*++r);return n}n.r(t),n.d(t,"category20",function(){return i}),n.d(t,"category20b",function(){return o}),n.d(t,"category20c",function(){return a}),n.d(t,"tableau10",function(){return s}),n.d(t,"tableau20",function(){return u}),n.d(t,"blueOrange",function(){return c});var i=r("1f77b4aec7e8ff7f0effbb782ca02c98df8ad62728ff98969467bdc5b0d58c564bc49c94e377c2f7b6d27f7f7fc7c7c7bcbd22dbdb8d17becf9edae5"),o=r("393b795254a36b6ecf9c9ede6379398ca252b5cf6bcedb9c8c6d31bd9e39e7ba52e7cb94843c39ad494ad6616be7969c7b4173a55194ce6dbdde9ed6"),a=r("3182bd6baed69ecae1c6dbefe6550dfd8d3cfdae6bfdd0a231a35474c476a1d99bc7e9c0756bb19e9ac8bcbddcdadaeb636363969696bdbdbdd9d9d9"),s=r("4c78a8f58518e4575672b7b254a24beeca3bb279a2ff9da69d755dbab0ac"),u=r("4c78a89ecae9f58518ffbf7954a24b88d27ab79a20f2cf5b43989483bcb6e45756ff9d9879706ebab0acd67195fcbfd2b279a2d6a5c99e765fd8b5a5"),c=new Array(3).concat("67a9cff7f7f7f1a340","0571b092c5defdb863e66101","0571b092c5def7f7f7fdb863e66101","2166ac67a9cfd1e5f0fee0b6f1a340b35806","2166ac67a9cfd1e5f0f7f7f7fee0b6f1a340b35806","2166ac4393c392c5ded1e5f0fee0b6fdb863e08214b35806","2166ac4393c392c5ded1e5f0f7f7f7fee0b6fdb863e08214b35806","0530612166ac4393c392c5ded1e5f0fee0b6fdb863e08214b358067f3b08","0530612166ac4393c392c5ded1e5f0f7f7f7fee0b6fdb863e08214b358067f3b08").map(r)},"./node_modules/vega-scale/src/scaleBand.js":function(e,t,n){"use strict";n.r(t),n.d(t,"band",function(){return a}),n.d(t,"point",function(){return s});var r=n("./node_modules/vega-scale/src/bandSpace.js"),i=n("./node_modules/d3-array/index.js"),o=n("./node_modules/d3-scale/index.js");function a(){var e,t,n=Object(o.scaleOrdinal)().unknown(void 0),s=n.domain,u=n.range,c=[0,1],l=!1,d=0,p=0,f=.5;function h(){var n=s().length,o=c[1]<c[0],a=c[o-0],h=c[1-o],m=Object(r.default)(n,d,p);e=(h-a)/(m||1),l&&(e=Math.floor(e)),a+=(h-a-e*(n-d))*f,t=e*(1-d),l&&(a=Math.round(a),t=Math.round(t));var g=Object(i.range)(n).map(function(t){return a+e*t});return u(o?g.reverse():g)}return delete n.unknown,n.domain=function(e){return arguments.length?(s(e),h()):s()},n.range=function(e){return arguments.length?(c=[+e[0],+e[1]],h()):c.slice()},n.rangeRound=function(e){return c=[+e[0],+e[1]],l=!0,h()},n.bandwidth=function(){return t},n.step=function(){return e},n.round=function(e){return arguments.length?(l=!!e,h()):l},n.padding=function(e){return arguments.length?(p=Math.max(0,Math.min(1,e)),d=p,h()):d},n.paddingInner=function(e){return arguments.length?(d=Math.max(0,Math.min(1,e)),h()):d},n.paddingOuter=function(e){return arguments.length?(p=Math.max(0,Math.min(1,e)),h()):p},n.align=function(e){return arguments.length?(f=Math.max(0,Math.min(1,e)),h()):f},n.invertRange=function(e){if(null!=e[0]&&null!=e[1]){var n,r,o,a=+e[0],l=+e[1],d=c[1]<c[0],p=d?u().reverse():u(),f=p.length-1;if(a==a&&l==l&&(l<a&&(o=a,a=l,l=o),!(l<p[0]||a>c[1-d])))return n=Math.max(0,Object(i.bisectRight)(p,a)-1),r=a===l?n:Object(i.bisectRight)(p,l)-1,a-p[n]>t+1e-10&&++n,d&&(o=n,n=f-r,r=f-o),n>r?void 0:s().slice(n,r+1)}},n.invert=function(e){var t=n.invertRange([e,e]);return t?t[0]:t},n.copy=function(){return a().domain(s()).range(c).round(l).paddingInner(d).paddingOuter(p).align(f)},h()}function s(){return function e(t){var n=t.copy;return t.padding=t.paddingOuter,delete t.paddingInner,t.copy=function(){return e(n())},t}(a().paddingInner(1))}},"./node_modules/vega-scale/src/scaleBin.js":function(e,t,n){"use strict";n.r(t),n.d(t,"binLinear",function(){return c}),n.d(t,"binOrdinal",function(){return l});var r=n("./node_modules/d3-array/index.js"),i=n("./node_modules/d3-scale/index.js"),o=n("./node_modules/vega-util/index.js"),a=Array.prototype.map,s=Array.prototype.slice;function u(e){return a.call(e,function(e){return+e})}function c(){var e=Object(i.scaleLinear)(),t=[];function n(t){return e(t)}return n.domain=function(r){return arguments.length?(t=u(r),e.domain([t[0],Object(o.peek)(t)]),n):t.slice()},n.range=function(t){return arguments.length?(e.range(t),n):e.range()},n.rangeRound=function(t){return arguments.length?(e.rangeRound(t),n):e.rangeRound()},n.interpolate=function(t){return arguments.length?(e.interpolate(t),n):e.interpolate()},n.invert=function(t){return e.invert(t)},n.ticks=function(e){var r=t.length,i=~~(r/(e||r));return i<2?n.domain():t.filter(function(e,t){return!(t%i)})},n.tickFormat=function(){return e.tickFormat.apply(e,arguments)},n.copy=function(){return c().domain(n.domain()).range(n.range())},n}function l(){var e=[],t=[];function n(n){return null==n||n!=n?void 0:t[(Object(r.bisect)(e,n)-1)%t.length]}return n.domain=function(t){return arguments.length?(e=u(t),n):e.slice()},n.range=function(e){return arguments.length?(t=s.call(e),n):t.slice()},n.copy=function(){return l().domain(n.domain()).range(n.range())},n}},"./node_modules/vega-scale/src/scaleSequential.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-util/index.js"),i=n("./node_modules/d3-scale/index.js");function o(e){var t=Object(i.scaleLinear)(),n=0,a=1,s=!1;function u(){var e=t.domain();n=e[0],a=Object(r.peek)(e)-n}function c(t){var r=(t-n)/a;return e(s?Math.max(0,Math.min(1,r)):r)}return c.clamp=function(e){return arguments.length?(s=!!e,c):s},c.domain=function(e){return arguments.length?(t.domain(e),u(),c):t.domain()},c.interpolator=function(t){return arguments.length?(e=t,c):e},c.copy=function(){return o().domain(t.domain()).clamp(s).interpolator(e)},c.ticks=function(e){return t.ticks(e)},c.tickFormat=function(e,n){return t.tickFormat(e,n)},c.nice=function(e){return t.nice(e),u(),c},c}},"./node_modules/vega-scale/src/scales.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return c});var r=n("./node_modules/vega-scale/src/invertRange.js"),i=n("./node_modules/vega-scale/src/invertRangeExtent.js"),o=n("./node_modules/vega-scale/src/scaleBand.js"),a=n("./node_modules/vega-scale/src/scaleBin.js"),s=n("./node_modules/vega-scale/src/scaleSequential.js"),u=n("./node_modules/d3-scale/index.js");function c(e,t){return arguments.length>1?(l[e]=(n=e,o=t,function(){var e=o();return e.invertRange||(e.invertRange=e.invert?Object(r.default)(e):e.invertExtent?Object(i.default)(e):void 0),e.type=n,e}),this):l.hasOwnProperty(e)?l[e]:void 0;var n,o}var l={identity:u.scaleIdentity,linear:u.scaleLinear,log:u.scaleLog,ordinal:u.scaleOrdinal,pow:u.scalePow,sqrt:u.scaleSqrt,quantile:u.scaleQuantile,quantize:u.scaleQuantize,threshold:u.scaleThreshold,time:u.scaleTime,utc:u.scaleUtc,band:o.band,point:o.point,sequential:s.default,"bin-linear":a.binLinear,"bin-ordinal":a.binOrdinal};for(var d in l)c(d,l[d])},"./node_modules/vega-scale/src/schemes.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scale/src/palettes.js"),i=n("./node_modules/d3-scale-chromatic/index.js"),o=n("./node_modules/d3-interpolate/index.js"),a=n("./node_modules/vega-util/index.js"),s={blueorange:r.blueOrange},u={category10:i.schemeCategory10,accent:i.schemeAccent,dark2:i.schemeDark2,paired:i.schemePaired,pastel1:i.schemePastel1,pastel2:i.schemePastel2,set1:i.schemeSet1,set2:i.schemeSet2,set3:i.schemeSet3,category20:r.category20,category20b:r.category20b,category20c:r.category20c,tableau10:r.tableau10,tableau20:r.tableau20,viridis:i.interpolateViridis,magma:i.interpolateMagma,inferno:i.interpolateInferno,plasma:i.interpolatePlasma,blueorange:Object(o.interpolateRgbBasis)(Object(a.peek)(r.blueOrange))};function c(e,t){u[e]=i["interpolate"+t],s[e]=i["scheme"+t]}c("blues","Blues"),c("greens","Greens"),c("greys","Greys"),c("purples","Purples"),c("reds","Reds"),c("oranges","Oranges"),c("brownbluegreen","BrBG"),c("purplegreen","PRGn"),c("pinkyellowgreen","PiYG"),c("purpleorange","PuOr"),c("redblue","RdBu"),c("redgrey","RdGy"),c("redyellowblue","RdYlBu"),c("redyellowgreen","RdYlGn"),c("spectral","Spectral"),c("bluegreen","BuGn"),c("bluepurple","BuPu"),c("greenblue","GnBu"),c("orangered","OrRd"),c("purplebluegreen","PuBuGn"),c("purpleblue","PuBu"),c("purplered","PuRd"),c("redpurple","RdPu"),c("yellowgreenblue","YlGnBu"),c("yellowgreen","YlGn"),c("yelloworangebrown","YlOrBr"),c("yelloworangered","YlOrRd"),t.default=function(e,t){if(arguments.length>1)return u[e]=t,this;var n=e.split("-");return e=n[0],(n=+n[1]+1)&&s.hasOwnProperty(e)?s[e][n-1]:!n&&u.hasOwnProperty(e)?u[e]:void 0}},"./node_modules/vega-scale/src/timeInterval.js":function(e,t,n){"use strict";n.r(t),n.d(t,"timeInterval",function(){return a}),n.d(t,"utcInterval",function(){return s});var r=n("./node_modules/d3-time/index.js"),i={millisecond:r.timeMillisecond,second:r.timeSecond,minute:r.timeMinute,hour:r.timeHour,day:r.timeDay,week:r.timeWeek,month:r.timeMonth,year:r.timeYear},o={millisecond:r.utcMillisecond,second:r.utcSecond,minute:r.utcMinute,hour:r.utcHour,day:r.utcDay,week:r.utcWeek,month:r.utcMonth,year:r.utcYear};function a(e){return i.hasOwnProperty(e)&&i[e]}function s(e){return o.hasOwnProperty(e)&&o[e]}},"./node_modules/vega-scenegraph/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/Bounds.js");n.d(t,"Bounds",function(){return r.default});var i=n("./node_modules/vega-scenegraph/src/Gradient.js");n.d(t,"Gradient",function(){return i.default});var o=n("./node_modules/vega-scenegraph/src/GroupItem.js");n.d(t,"GroupItem",function(){return o.default});var a=n("./node_modules/vega-scenegraph/src/ResourceLoader.js");n.d(t,"ResourceLoader",function(){return a.default});var s=n("./node_modules/vega-scenegraph/src/Item.js");n.d(t,"Item",function(){return s.default});var u=n("./node_modules/vega-scenegraph/src/Scenegraph.js");n.d(t,"Scenegraph",function(){return u.default});var c=n("./node_modules/vega-scenegraph/src/Handler.js");n.d(t,"Handler",function(){return c.default});var l=n("./node_modules/vega-scenegraph/src/Renderer.js");n.d(t,"Renderer",function(){return l.default});var d=n("./node_modules/vega-scenegraph/src/CanvasHandler.js");n.d(t,"CanvasHandler",function(){return d.default});var p=n("./node_modules/vega-scenegraph/src/CanvasRenderer.js");n.d(t,"CanvasRenderer",function(){return p.default});var f=n("./node_modules/vega-scenegraph/src/SVGHandler.js");n.d(t,"SVGHandler",function(){return f.default});var h=n("./node_modules/vega-scenegraph/src/SVGRenderer.js");n.d(t,"SVGRenderer",function(){return h.default});var m=n("./node_modules/vega-scenegraph/src/SVGStringRenderer.js");n.d(t,"SVGStringRenderer",function(){return m.default});var g=n("./node_modules/vega-scenegraph/src/modules.js");n.d(t,"RenderType",function(){return g.RenderType}),n.d(t,"renderModule",function(){return g.renderModule});var v=n("./node_modules/vega-scenegraph/src/marks/index.js");n.d(t,"Marks",function(){return v.default});var _=n("./node_modules/vega-scenegraph/src/bound/boundClip.js");n.d(t,"boundClip",function(){return _.default});var y=n("./node_modules/vega-scenegraph/src/bound/boundContext.js");n.d(t,"boundContext",function(){return y.default});var b=n("./node_modules/vega-scenegraph/src/bound/boundStroke.js");n.d(t,"boundStroke",function(){return b.default});var j=n("./node_modules/vega-scenegraph/src/bound/boundItem.js");n.d(t,"boundItem",function(){return j.default});var x=n("./node_modules/vega-scenegraph/src/bound/boundMark.js");n.d(t,"boundMark",function(){return x.default});var O=n("./node_modules/vega-scenegraph/src/path/curves.js");n.d(t,"pathCurves",function(){return O.default});var w=n("./node_modules/vega-scenegraph/src/path/symbols.js");n.d(t,"pathSymbols",function(){return w.default});var C=n("./node_modules/vega-scenegraph/src/path/rectangle.js");n.d(t,"pathRectangle",function(){return C.default});var E=n("./node_modules/vega-scenegraph/src/path/trail.js");n.d(t,"pathTrail",function(){return E.default});var S=n("./node_modules/vega-scenegraph/src/path/parse.js");n.d(t,"pathParse",function(){return S.default});var k=n("./node_modules/vega-scenegraph/src/path/render.js");n.d(t,"pathRender",function(){return k.default});var M=n("./node_modules/vega-scenegraph/src/util/point.js");n.d(t,"point",function(){return M.default});var A=n("./node_modules/vega-scenegraph/src/util/dom.js");n.d(t,"domCreate",function(){return A.domCreate}),n.d(t,"domFind",function(){return A.domFind}),n.d(t,"domChild",function(){return A.domChild}),n.d(t,"domClear",function(){return A.domClear});var T=n("./node_modules/vega-scenegraph/src/util/tags.js");n.d(t,"openTag",function(){return T.openTag}),n.d(t,"closeTag",function(){return T.closeTag});var D=n("./node_modules/vega-scenegraph/src/util/text.js");n.d(t,"font",function(){return D.font}),n.d(t,"textMetrics",function(){return D.textMetrics});var I=n("./node_modules/vega-scenegraph/src/util/svg/clip.js");n.d(t,"resetSVGClipId",function(){return I.resetSVGClipId});var P=n("./node_modules/vega-scenegraph/src/util/equal.js");n.d(t,"sceneEqual",function(){return P.sceneEqual}),n.d(t,"pathEqual",function(){return P.pathEqual});var R=n("./node_modules/vega-scenegraph/src/util/serialize.js");n.d(t,"sceneToJSON",function(){return R.sceneToJSON}),n.d(t,"sceneFromJSON",function(){return R.sceneFromJSON});var N=n("./node_modules/vega-scenegraph/src/util/visit.js");n.d(t,"sceneZOrder",function(){return N.zorder}),n.d(t,"sceneVisit",function(){return N.visit}),n.d(t,"scenePickVisit",function(){return N.pickVisit})},"./node_modules/vega-scenegraph/src/Bounds.js":function(e,t,n){"use strict";function r(e){this.clear(),e&&this.union(e)}n.r(t),n.d(t,"default",function(){return r});var i=r.prototype;i.clone=function(){return new r(this)},i.clear=function(){return this.x1=+Number.MAX_VALUE,this.y1=+Number.MAX_VALUE,this.x2=-Number.MAX_VALUE,this.y2=-Number.MAX_VALUE,this},i.empty=function(){return this.x1===+Number.MAX_VALUE&&this.y1===+Number.MAX_VALUE&&this.x2===-Number.MAX_VALUE&&this.y2===-Number.MAX_VALUE},i.set=function(e,t,n,r){return n<e?(this.x2=e,this.x1=n):(this.x1=e,this.x2=n),r<t?(this.y2=t,this.y1=r):(this.y1=t,this.y2=r),this},i.add=function(e,t){return e<this.x1&&(this.x1=e),t<this.y1&&(this.y1=t),e>this.x2&&(this.x2=e),t>this.y2&&(this.y2=t),this},i.expand=function(e){return this.x1-=e,this.y1-=e,this.x2+=e,this.y2+=e,this},i.round=function(){return this.x1=Math.floor(this.x1),this.y1=Math.floor(this.y1),this.x2=Math.ceil(this.x2),this.y2=Math.ceil(this.y2),this},i.translate=function(e,t){return this.x1+=e,this.x2+=e,this.y1+=t,this.y2+=t,this},i.rotate=function(e,t,n){var r=Math.cos(e),i=Math.sin(e),o=t-t*r+n*i,a=n-t*i-n*r,s=this.x1,u=this.x2,c=this.y1,l=this.y2;return this.clear().add(r*s-i*c+o,i*s+r*c+a).add(r*s-i*l+o,i*s+r*l+a).add(r*u-i*c+o,i*u+r*c+a).add(r*u-i*l+o,i*u+r*l+a)},i.union=function(e){return e.x1<this.x1&&(this.x1=e.x1),e.y1<this.y1&&(this.y1=e.y1),e.x2>this.x2&&(this.x2=e.x2),e.y2>this.y2&&(this.y2=e.y2),this},i.intersect=function(e){return e.x1>this.x1&&(this.x1=e.x1),e.y1>this.y1&&(this.y1=e.y1),e.x2<this.x2&&(this.x2=e.x2),e.y2<this.y2&&(this.y2=e.y2),this},i.encloses=function(e){return e&&this.x1<=e.x1&&this.x2>=e.x2&&this.y1<=e.y1&&this.y2>=e.y2},i.alignsWith=function(e){return e&&(this.x1==e.x1||this.x2==e.x2||this.y1==e.y1||this.y2==e.y2)},i.intersects=function(e){return e&&!(this.x2<e.x1||this.x1>e.x2||this.y2<e.y1||this.y1>e.y2)},i.contains=function(e,t){return!(e<this.x1||e>this.x2||t<this.y1||t>this.y2)},i.width=function(){return this.x2-this.x1},i.height=function(){return this.y2-this.y1}},"./node_modules/vega-scenegraph/src/CanvasHandler.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return u});var r=n("./node_modules/vega-scenegraph/src/Handler.js"),i=n("./node_modules/vega-scenegraph/src/marks/index.js"),o=n("./node_modules/vega-scenegraph/src/util/point.js"),a=n("./node_modules/vega-scenegraph/src/util/dom.js"),s=n("./node_modules/vega-util/index.js");function u(e){r.default.call(this,e),this._down=null,this._touch=null,this._first=!0}var c=Object(s.inherits)(u,r.default);function l(e,t,n){return function(r){var i=this._active,o=this.pickEvent(r);o===i?this.fire(e,r):(i&&i.exit||this.fire(n,r),this._active=o,this.fire(t,r),this.fire(e,r))}}function d(e){return function(t){this.fire(e,t),this._active=null}}c.initialize=function(e,t,n){var i=this._canvas=e&&Object(a.domFind)(e,"canvas");if(i){var o=this;this.events.forEach(function(e){i.addEventListener(e,function(t){c[e]?c[e].call(o,t):o.fire(e,t)})})}return r.default.prototype.initialize.call(this,e,t,n)},c.canvas=function(){return this._canvas},c.context=function(){return this._canvas.getContext("2d")},c.events=["keydown","keypress","keyup","dragenter","dragleave","dragover","mousedown","mouseup","mousemove","mouseout","mouseover","click","dblclick","wheel","mousewheel","touchstart","touchmove","touchend"],c.DOMMouseScroll=function(e){this.fire("mousewheel",e)},c.mousemove=l("mousemove","mouseover","mouseout"),c.dragover=l("dragover","dragenter","dragleave"),c.mouseout=d("mouseout"),c.dragleave=d("dragleave"),c.mousedown=function(e){this._down=this._active,this.fire("mousedown",e)},c.click=function(e){this._down===this._active&&(this.fire("click",e),this._down=null)},c.touchstart=function(e){this._touch=this.pickEvent(e.changedTouches[0]),this._first&&(this._active=this._touch,this._first=!1),this.fire("touchstart",e,!0)},c.touchmove=function(e){this.fire("touchmove",e,!0)},c.touchend=function(e){this.fire("touchend",e,!0),this._touch=null},c.fire=function(e,t,n){var r,i,o=n?this._touch:this._active,a=this._handlers[e];if("click"===e&&o&&o.href?this.handleHref(t,o,o.href):("mouseover"===e||"mouseout"===e)&&o&&o.tooltip&&this.handleTooltip(t,o,"mouseover"===e?o.tooltip:null),a)for(t.vegaType=e,r=0,i=a.length;r<i;++r)a[r].handler.call(this._obj,t,o)},c.on=function(e,t){var n=this.eventName(e),r=this._handlers;return this._handlerIndex(r[n],e,t)<0&&(r[n]||(r[n]=[])).push({type:e,handler:t}),this},c.off=function(e,t){var n=this.eventName(e),r=this._handlers[n],i=this._handlerIndex(r,e,t);return i>=0&&r.splice(i,1),this},c.pickEvent=function(e){var t=Object(o.default)(e,this._canvas),n=this._origin;return this.pick(this._scene,t[0],t[1],t[0]-n[0],t[1]-n[1])},c.pick=function(e,t,n,r,o){var a=this.context();return i.default[e.marktype].pick.call(this,a,e,t,n,r,o)}},"./node_modules/vega-scenegraph/src/CanvasRenderer.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return d});var r=n("./node_modules/vega-scenegraph/src/Renderer.js"),i=n("./node_modules/vega-scenegraph/src/Bounds.js"),o=n("./node_modules/vega-scenegraph/src/marks/index.js"),a=n("./node_modules/vega-scenegraph/src/util/dom.js"),s=n("./node_modules/vega-scenegraph/src/util/canvas/clip.js"),u=n("./node_modules/vega-scenegraph/src/util/canvas/resize.js"),c=n("./node_modules/vega-canvas/index.js"),l=n("./node_modules/vega-util/index.js");function d(e){r.default.call(this,e),this._redraw=!1,this._dirty=new i.default}var p=Object(l.inherits)(d,r.default),f=r.default.prototype,h=new i.default;p.initialize=function(e,t,n,r,i){return this._canvas=Object(c.canvas)(1,1),e&&(Object(a.domClear)(e,0).appendChild(this._canvas),this._canvas.setAttribute("class","marks")),f.initialize.call(this,e,t,n,r,i)},p.resize=function(e,t,n,r){return f.resize.call(this,e,t,n,r),Object(u.default)(this._canvas,this._width,this._height,this._origin,this._scale),this._redraw=!0,this},p.canvas=function(){return this._canvas},p.context=function(){return this._canvas?this._canvas.getContext("2d"):null},p.dirty=function(e){var t=function(e,t){if(null==t)return e;for(var n=h.clear().union(e);null!=t;t=t.mark.group)n.translate(t.x||0,t.y||0);return n}(e.bounds,e.mark.group);this._dirty.union(t)},p._render=function(e){var t,n,r,i=this.context(),o=this._origin,a=this._width,s=this._height,u=this._dirty;return i.save(),this._redraw||u.empty()?(this._redraw=!1,u=null):(t=i,r=o,(n=u).expand(1).round(),n.translate(-r[0]%1,-r[1]%1),t.beginPath(),t.rect(n.x1,n.y1,n.width(),n.height()),t.clip(),u=n),this.clear(-o[0],-o[1],a,s),this.draw(i,e,u),i.restore(),this._dirty.clear(),this},p.draw=function(e,t,n){var r=o.default[t.marktype];t.clip&&Object(s.default)(e,t),r.draw.call(this,e,t,n),t.clip&&e.restore()},p.clear=function(e,t,n,r){var i=this.context();i.clearRect(e,t,n,r),null!=this._bgcolor&&(i.fillStyle=this._bgcolor,i.fillRect(e,t,n,r))}},"./node_modules/vega-scenegraph/src/Gradient.js":function(e,t,n){"use strict";n.r(t);var r=0;t.default=function(e,t){var n,i=[];return n={id:"gradient_"+r++,x1:e?e[0]:0,y1:e?e[1]:0,x2:t?t[0]:1,y2:t?t[1]:0,stops:i,stop:function(e,t){return i.push({offset:e,color:t}),n}}}},"./node_modules/vega-scenegraph/src/GroupItem.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-scenegraph/src/Item.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.default.call(this,e),this.items=this.items||[]}Object(i.inherits)(o,r.default)},"./node_modules/vega-scenegraph/src/Handler.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-scenegraph/src/util/dom.js"),i=n("./node_modules/vega-loader/index.js");function o(e){this._active=null,this._handlers={},this._loader=e||Object(i.loader)()}var a=o.prototype;a.initialize=function(e,t,n){return this._el=e,this._obj=n||null,this.origin(t)},a.element=function(){return this._el},a.origin=function(e){return this._origin=e||[0,0],this},a.scene=function(e){return arguments.length?(this._scene=e,this):this._scene},a.on=function(){},a.off=function(){},a._handlerIndex=function(e,t,n){for(var r=e?e.length:0;--r>=0;)if(e[r].type===t&&!n||e[r].handler===n)return r;return-1},a.handlers=function(){var e,t=this._handlers,n=[];for(e in t)n.push.apply(n,t[e]);return n},a.eventName=function(e){var t=e.indexOf(".");return t<0?e:e.slice(0,t)},a.handleHref=function(e,t,n){this._loader.sanitize(n,{context:"href"}).then(function(t){var n=new MouseEvent(e.type,e),i=Object(r.domCreate)(null,"a");for(var o in t)i.setAttribute(o,t[o]);i.dispatchEvent(n)}).catch(function(){})},a.handleTooltip=function(e,t,n){this._el.setAttribute("title",n||"")}},"./node_modules/vega-scenegraph/src/Item.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return i});var r=n("./node_modules/vega-scenegraph/src/Bounds.js");function i(e){this.mark=e,this.bounds=this.bounds||new r.default}},"./node_modules/vega-scenegraph/src/Renderer.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return i});var r=n("./node_modules/vega-scenegraph/src/ResourceLoader.js");function i(e){this._el=null,this._bgcolor=null,this._loader=new r.default(e)}var o=i.prototype;o.initialize=function(e,t,n,r,i){return this._el=e,this.resize(t,n,r,i)},o.element=function(){return this._el},o.scene=function(){return this._el&&this._el.firstChild},o.background=function(e){return 0===arguments.length?this._bgcolor:(this._bgcolor=e,this)},o.resize=function(e,t,n,r){return this._width=e,this._height=t,this._origin=n||[0,0],this._scale=r||1,this},o.dirty=function(){},o.render=function(e){var t=this;return t._call=function(){t._render(e)},t._call(),t._call=null,t},o._render=function(){},o.renderAsync=function(e){var t=this.render(e);return this._ready?this._ready.then(function(){return t}):Promise.resolve(t)},o._load=function(e,t){var n=this,r=n._loader[e](t);if(!n._ready){var i=n._call;n._ready=n._loader.ready().then(function(e){e&&i(),n._ready=null})}return r},o.sanitizeURL=function(e){return this._load("sanitizeURL",e)},o.loadImage=function(e){return this._load("loadImage",e)}},"./node_modules/vega-scenegraph/src/ResourceLoader.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-canvas/index.js"),i=n("./node_modules/vega-loader/index.js");function o(e){this._pending=0,this._loader=e||Object(i.loader)()}var a=o.prototype;function s(e){e._pending+=1}function u(e){e._pending-=1}a.pending=function(){return this._pending},a.sanitizeURL=function(e){var t=this;return s(t),t._loader.sanitize(e,{context:"href"}).then(function(e){return u(t),e}).catch(function(){return u(t),null})},a.loadImage=function(e){var t=this,n=Object(r.image)();return s(t),t._loader.sanitize(e,{context:"image"}).then(function(e){var r=e.href;if(!r||!n)throw{url:r};var i=new n;return i.onload=function(){u(t),i.loaded=!0},i.onerror=function(){u(t),i.loaded=!1},i.src=r,i}).catch(function(e){return u(t),{loaded:!1,width:0,height:0,src:e&&e.url||""}})},a.ready=function(){var e=this;return new Promise(function(t){!function n(r){e.pending()?setTimeout(function(){n(!0)},10):t(r)}(!1)})}},"./node_modules/vega-scenegraph/src/SVGHandler.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-scenegraph/src/Handler.js"),i=n("./node_modules/vega-scenegraph/src/util/dom.js"),o=n("./node_modules/vega-util/index.js");function a(e){r.default.call(this,e);var t=this;t._hrefHandler=u(t,function(e,n){n&&n.href&&t.handleHref(e,n,n.href)}),t._tooltipHandler=u(t,function(e,n){n&&n.tooltip&&t.handleTooltip(e,n,"mouseover"===e.type?n.tooltip:null)})}var s=Object(o.inherits)(a,r.default);function u(e,t){return function(n){var r=n.target.__data__;n.vegaType=n.type,r=Array.isArray(r)?r[0]:r,t.call(e._obj,n,r)}}s.initialize=function(e,t,n){var o=this._svg;return o&&(o.removeEventListener("click",this._hrefHandler),o.removeEventListener("mouseover",this._tooltipHandler),o.removeEventListener("mouseout",this._tooltipHandler)),this._svg=o=e&&Object(i.domFind)(e,"svg"),o&&(o.addEventListener("click",this._hrefHandler),o.addEventListener("mouseover",this._tooltipHandler),o.addEventListener("mouseout",this._tooltipHandler)),r.default.prototype.initialize.call(this,e,t,n)},s.svg=function(){return this._svg},s.on=function(e,t){var n=this.eventName(e),r=this._handlers;if(this._handlerIndex(r[n],e,t)<0){var i={type:e,handler:t,listener:u(this,t)};(r[n]||(r[n]=[])).push(i),this._svg&&this._svg.addEventListener(n,i.listener)}return this},s.off=function(e,t){var n=this.eventName(e),r=this._handlers[n],i=this._handlerIndex(r,e,t);return i>=0&&(this._svg&&this._svg.removeEventListener(n,r[i].listener),r.splice(i,1)),this}},"./node_modules/vega-scenegraph/src/SVGRenderer.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return h});var r=n("./node_modules/vega-scenegraph/src/Renderer.js"),i=n("./node_modules/vega-scenegraph/src/marks/index.js"),o=n("./node_modules/vega-scenegraph/src/util/dom.js"),a=n("./node_modules/vega-scenegraph/src/util/tags.js"),s=n("./node_modules/vega-scenegraph/src/util/text.js"),u=n("./node_modules/vega-scenegraph/src/util/visit.js"),c=n("./node_modules/vega-scenegraph/src/util/svg/clip.js"),l=n("./node_modules/vega-scenegraph/src/util/svg/metadata.js"),d=n("./node_modules/vega-scenegraph/src/util/svg/styles.js"),p=n("./node_modules/vega-util/index.js"),f=l.default.xmlns;function h(e){r.default.call(this,e),this._dirtyID=1,this._dirty=[],this._svg=null,this._root=null,this._defs=null}var m=Object(p.inherits)(h,r.default),g=r.default.prototype;function v(e,t,n){var r,i,a;for((e=Object(o.domChild)(e,n,"linearGradient",f)).setAttribute("id",t.id),e.setAttribute("x1",t.x1),e.setAttribute("x2",t.x2),e.setAttribute("y1",t.y1),e.setAttribute("y2",t.y2),r=0,i=t.stops.length;r<i;++r)(a=Object(o.domChild)(e,r,"stop",f)).setAttribute("offset",t.stops[r].offset),a.setAttribute("stop-color",t.stops[r].color);Object(o.domClear)(e,r)}function _(e,t,n){var r;(e=Object(o.domChild)(e,n,"clipPath",f)).setAttribute("id",t.id),t.path?(r=Object(o.domChild)(e,0,"path",f)).setAttribute("d",t.path):((r=Object(o.domChild)(e,0,"rect",f)).setAttribute("x",0),r.setAttribute("y",0),r.setAttribute("width",t.width),r.setAttribute("height",t.height))}function y(e,t){for(;e&&e.dirty!==t;e=e.mark.group){if(e.dirty=t,!e.mark||e.mark.dirty===t)return;e.mark.dirty=t}}function b(e,t,n,r){var i,a=e._svg;if(!a&&(i=t.ownerDocument,a=Object(o.domCreate)(i,r,f),e._svg=a,e.mark&&(a.__data__=e,a.__values__={fill:"default"},"g"===r))){var s=Object(o.domCreate)(i,"path",f);s.setAttribute("class","background"),a.appendChild(s),s.__data__=e;var u=Object(o.domCreate)(i,"g",f);a.appendChild(u),u.__data__=e}return!i&&a.previousSibling===n&&n||t.insertBefore(a,n?n.nextSibling:t.firstChild),a}m.initialize=function(e,t,n,r){return e&&(this._svg=Object(o.domChild)(e,0,"svg",f),this._svg.setAttribute("class","marks"),Object(o.domClear)(e,1),this._root=Object(o.domChild)(this._svg,0,"g",f),Object(o.domClear)(this._svg,1)),this._defs={gradient:{},clipping:{}},this.background(this._bgcolor),g.initialize.call(this,e,t,n,r)},m.background=function(e){return arguments.length&&this._svg&&this._svg.style.setProperty("background-color",e),g.background.apply(this,arguments)},m.resize=function(e,t,n,r){return g.resize.call(this,e,t,n,r),this._svg&&(this._svg.setAttribute("width",this._width*this._scale),this._svg.setAttribute("height",this._height*this._scale),this._svg.setAttribute("viewBox","0 0 "+this._width+" "+this._height),this._root.setAttribute("transform","translate("+this._origin+")")),this._dirty=[],this},m.svg=function(){if(!this._svg)return null;var e={class:"marks",width:this._width*this._scale,height:this._height*this._scale,viewBox:"0 0 "+this._width+" "+this._height};for(var t in l.default)e[t]=l.default[t];var n=this._bgcolor?Object(a.openTag)("rect",{width:this._width,height:this._height,style:"fill: "+this._bgcolor+";"})+Object(a.closeTag)("rect"):"";return Object(a.openTag)("svg",e)+n+this._svg.innerHTML+Object(a.closeTag)("svg")},m._render=function(e){return this._dirtyCheck()&&(this._dirtyAll&&this._resetDefs(),this.draw(this._root,e),Object(o.domClear)(this._root,1)),this.updateDefs(),this._dirty=[],++this._dirtyID,this},m.updateDefs=function(){var e,t=this._svg,n=this._defs,r=n.el,i=0;for(e in n.gradient)r||(n.el=r=Object(o.domChild)(t,0,"defs",f)),v(r,n.gradient[e],i++);for(e in n.clipping)r||(n.el=r=Object(o.domChild)(t,0,"defs",f)),_(r,n.clipping[e],i++);r&&(0===i?(t.removeChild(r),n.el=null):Object(o.domClear)(r,i))},m._resetDefs=function(){var e=this._defs;e.gradient={},e.clipping={}},m.dirty=function(e){e.dirty!==this._dirtyID&&(e.dirty=this._dirtyID,this._dirty.push(e))},m.isDirty=function(e){return this._dirtyAll||!e._svg||e.dirty===this._dirtyID},m._dirtyCheck=function(){this._dirtyAll=!0;var e=this._dirty;if(!e.length)return!0;var t,n,r,o,a,s,u,c=++this._dirtyID;for(a=0,s=e.length;a<s;++a)(n=(t=e[a]).mark).marktype!==r&&(r=n.marktype,o=i.default[r]),n.zdirty&&n.dirty!==c&&(this._dirtyAll=!1,n.dirty=c,y(n.group,c)),t.exit?(o.nested&&n.items.length?(u=n.items[0])._svg&&this._update(o,u._svg,u):t._svg&&(u=t._svg.parentNode)&&u.removeChild(t._svg),t._svg=null):(t=o.nested?n.items[0]:t)._update!==c&&(t._svg&&t._svg.ownerSVGElement?this._update(o,t._svg,t):(this._dirtyAll=!1,y(t,c)),t._update=c);return!this._dirtyAll},m.draw=function(e,t,n){if(!this.isDirty(t))return t._svg;var r,a=this,s=i.default[t.marktype],l=!1===t.interactive?"none":null,d="g"===s.tag,p=null,f=0;function h(e){var t=a.isDirty(e),n=b(e,r,p,s.tag);t&&(a._update(s,n,e),d&&function(e,t,n){t=t.lastChild;var r,i=0;Object(u.visit)(n,function(n){r=e.draw(t,n,r),++i}),Object(o.domClear)(t,1+i)}(a,n,e)),p=n,++f}return(r=b(t,e,n,"g")).setAttribute("class",Object(o.cssClass)(t)),d||r.style.setProperty("pointer-events",l),t.clip?r.setAttribute("clip-path",Object(c.default)(a,t,t.group)):r.removeAttribute("clip-path"),s.nested?t.items.length&&h(t.items[0]):Object(u.visit)(t,h),Object(o.domClear)(r,f),r};var j=null,x=null,O={group:function(e,t,n){x=t.__values__,j=t.childNodes[1],e.foreground(w,n,this),j=t.childNodes[0],e.background(w,n,this);var r=!1===n.mark.interactive?"none":null;r!==x.events&&(j.style.setProperty("pointer-events",r),x.events=r)},text:function(e,t,n){var r=Object(s.textValue)(n);r!==x.text&&(t.textContent=r,x.text=r),(r=Object(s.font)(n))!==x.font&&(t.style.setProperty("font",r),x.font=r)}};function w(e,t,n){t!==x[e]&&(null!=t?n?j.setAttributeNS(n,e,t):j.setAttribute(e,t):n?j.removeAttributeNS(n,e):j.removeAttribute(e),x[e]=t)}m._update=function(e,t,n){j=t,x=t.__values__,e.attr(w,n,this);var r=O[e.type];r&&r.call(this,e,t,n),this.style(j,n)},m.style=function(e,t){var n,r,i,o,a,s;if(null!=t)for(n=0,r=d.styleProperties.length;n<r;++n)(a=t[i=d.styleProperties[n]])!==x[i]&&(o=d.styles[i],null==a?"fill"===o?e.style.setProperty(o,"none"):e.style.removeProperty(o):(a.id&&(this._defs.gradient[a.id]=a,a="url("+(s=void 0,"undefined"==typeof window?"":(s=window.location).hash?s.href.slice(0,-s.hash.length):s.href)+"#"+a.id+")"),e.style.setProperty(o,a+"")),x[i]=a)}},"./node_modules/vega-scenegraph/src/SVGStringRenderer.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return f});var r=n("./node_modules/vega-scenegraph/src/Renderer.js"),i=n("./node_modules/vega-scenegraph/src/marks/index.js"),o=n("./node_modules/vega-scenegraph/src/util/dom.js"),a=n("./node_modules/vega-scenegraph/src/util/tags.js"),s=n("./node_modules/vega-scenegraph/src/util/text.js"),u=n("./node_modules/vega-scenegraph/src/util/visit.js"),c=n("./node_modules/vega-scenegraph/src/util/svg/clip.js"),l=n("./node_modules/vega-scenegraph/src/util/svg/metadata.js"),d=n("./node_modules/vega-scenegraph/src/util/svg/styles.js"),p=n("./node_modules/vega-util/index.js");function f(e){r.default.call(this,e),this._text={head:"",bg:"",root:"",foot:"",defs:"",body:""},this._defs={gradient:{},clipping:{}}}var h,m=Object(p.inherits)(f,r.default),g=r.default.prototype;function v(e,t,n,r){h[r||e]=t}function _(e,t,n,r){if(null==e)return"";var i,o,a,u,c,l="";for("bgrect"===n&&!1===t.interactive&&(l+="pointer-events: none; "),"text"===n&&(l+="font: "+Object(s.font)(e)+"; "),i=0,o=d.styleProperties.length;i<o;++i)a=d.styleProperties[i],u=d.styles[a],null==(c=e[a])?"fill"===u&&(l+="fill: none; "):"transparent"!==c||"fill"!==u&&"stroke"!==u?(c.id&&(r.gradient[c.id]=c,c="url(#"+c.id+")"),l+=u+": "+c+"; "):l+=u+": none; ";return l?'style="'+l.trim()+'"':null}m.resize=function(e,t,n,r){g.resize.call(this,e,t,n,r);var i=this._origin,o=this._text,s={class:"marks",width:this._width*this._scale,height:this._height*this._scale,viewBox:"0 0 "+this._width+" "+this._height};for(var u in l.default)s[u]=l.default[u];o.head=Object(a.openTag)("svg",s);var c=this._bgcolor;return"transparent"!==c&&"none"!==c||(c=null),o.bg=c?Object(a.openTag)("rect",{width:this._width,height:this._height,style:"fill: "+c+";"})+Object(a.closeTag)("rect"):"",o.root=Object(a.openTag)("g",{transform:"translate("+i+")"}),o.foot=Object(a.closeTag)("g")+Object(a.closeTag)("svg"),this},m.background=function(){var e=g.background.apply(this,arguments);return arguments.length&&this._text.head&&this.resize(this._width,this._height,this._origin,this._scale),e},m.svg=function(){var e=this._text;return e.head+e.bg+e.defs+e.root+e.body+e.foot},m._render=function(e){return this._text.body=this.mark(e),this._text.defs=this.buildDefs(),this},m.buildDefs=function(){var e,t,n,r,i=this._defs,o="";for(t in i.gradient){for(r=(n=i.gradient[t]).stops,o+=Object(a.openTag)("linearGradient",{id:t,x1:n.x1,x2:n.x2,y1:n.y1,y2:n.y2}),e=0;e<r.length;++e)o+=Object(a.openTag)("stop",{offset:r[e].offset,"stop-color":r[e].color})+Object(a.closeTag)("stop");o+=Object(a.closeTag)("linearGradient")}for(t in i.clipping)n=i.clipping[t],o+=Object(a.openTag)("clipPath",{id:t}),n.path?o+=Object(a.openTag)("path",{d:n.path})+Object(a.closeTag)("path"):o+=Object(a.openTag)("rect",{x:0,y:0,width:n.width,height:n.height})+Object(a.closeTag)("rect"),o+=Object(a.closeTag)("clipPath");return o.length>0?Object(a.openTag)("defs")+o+Object(a.closeTag)("defs"):""},m.attributes=function(e,t){return h={},e(v,t,this),h},m.href=function(e){var t,n=this,r=e.href;if(r){if(t=n._hrefs&&n._hrefs[r])return t;n.sanitizeURL(r).then(function(e){e["xlink:href"]=e.href,e.href=null,(n._hrefs||(n._hrefs={}))[r]=e})}return null},m.mark=function(e){var t,n=this,r=i.default[e.marktype],l=r.tag,d=this._defs,p="";function f(i){var o=n.href(i);o&&(p+=Object(a.openTag)("a",o)),t="g"!==l?_(i,e,l,d):null,p+=Object(a.openTag)(l,n.attributes(r.attr,i),t),"text"===l?p+=Object(s.textValue)(i).replace(/&/g,"&amp;").replace(/</g,"&lt;").replace(/>/g,"&gt;"):"g"===l&&(p+=Object(a.openTag)("path",n.attributes(r.background,i),_(i,e,"bgrect",d))+Object(a.closeTag)("path"),p+=Object(a.openTag)("g",n.attributes(r.foreground,i))+n.markGroup(i)+Object(a.closeTag)("g")),p+=Object(a.closeTag)(l),o&&(p+=Object(a.closeTag)("a"))}return"g"!==l&&!1===e.interactive&&(t='style="pointer-events: none;"'),p+=Object(a.openTag)("g",{class:Object(o.cssClass)(e),"clip-path":e.clip?Object(c.default)(n,e,e.group):null},t),r.nested?e.items&&e.items.length&&f(e.items[0]):Object(u.visit)(e,f),p+Object(a.closeTag)("g")},m.markGroup=function(e){var t=this,n="";return Object(u.visit)(e,function(e){n+=t.mark(e)}),n}},"./node_modules/vega-scenegraph/src/Scenegraph.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-scenegraph/src/Bounds.js"),i=n("./node_modules/vega-scenegraph/src/GroupItem.js"),o=n("./node_modules/vega-scenegraph/src/util/serialize.js");function a(e){arguments.length?this.root=Object(o.sceneFromJSON)(e):(this.root=u({marktype:"group",name:"root",role:"frame"}),this.root.items=[new i.default(this.root)])}var s=a.prototype;function u(e,t){return{bounds:new r.default,clip:!!e.clip,group:t,interactive:!1!==e.interactive,items:[],marktype:e.marktype,name:e.name||void 0,role:e.role||void 0,zindex:e.zindex||0}}s.toJSON=function(e){return Object(o.sceneToJSON)(this.root,e||0)},s.mark=function(e,t,n){var r=u(e,t=t||this.root.items[0]);return t.items[n]=r,r.zindex&&(r.group.zdirty=!0),r}},"./node_modules/vega-scenegraph/src/bound/boundClip.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/Bounds.js"),i=n("./node_modules/vega-scenegraph/src/bound/boundContext.js"),o=n("./node_modules/vega-util/index.js"),a=new r.default;t.default=function(e){var t=e.clip;if(Object(o.isFunction)(t))t(Object(i.default)(a.clear()));else{if(!t)return;a.set(0,0,e.group.width,e.group.height)}e.bounds.intersect(a)}},"./node_modules/vega-scenegraph/src/bound/boundContext.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return s});var r,i=2*Math.PI,o=i/4,a=i-1e-8;function s(e){return r=e,s}function u(){}function c(e,t){r.add(e,t)}s.beginPath=u,s.closePath=u,s.moveTo=c,s.lineTo=c,s.rect=function(e,t,n,r){c(e,t),c(e+n,t+r)},s.quadraticCurveTo=function(e,t,n,r){c(e,t),c(n,r)},s.bezierCurveTo=function(e,t,n,r,i,o){c(e,t),c(n,r),c(i,o)},s.arc=function(e,t,n,r,s,u){if(Math.abs(s-r)>a)return c(e-n,t-n),void c(e+n,t+n);var l,d,p,f,h=1/0,m=-1/0,g=1/0,v=-1/0;function _(e){p=n*Math.cos(e),f=n*Math.sin(e),p<h&&(h=p),p>m&&(m=p),f<g&&(g=f),f>v&&(v=f)}if(_(r),_(s),s!==r)if((r%=i)<0&&(r+=i),(s%=i)<0&&(s+=i),s<r&&(u=!u,l=r,r=s,s=l),u)for(s-=i,l=r-r%o,d=0;d<4&&l>s;++d,l-=o)_(l);else for(l=r-r%o+o,d=0;d<4&&l<s;++d,l+=o)_(l);c(e+h,t+g),c(e+m,t+v)}},"./node_modules/vega-scenegraph/src/bound/boundItem.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/Bounds.js"),i=n("./node_modules/vega-scenegraph/src/marks/index.js");t.default=function(e,t,n){var o=i.default[e.mark.marktype],a=t||o.bound;return o.nested&&(e=e.mark),a(e.bounds||(e.bounds=new r.default),e,n)}},"./node_modules/vega-scenegraph/src/bound/boundMark.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/Bounds.js"),i=n("./node_modules/vega-scenegraph/src/bound/boundItem.js"),o=n("./node_modules/vega-scenegraph/src/marks/index.js"),a={mark:null};t.default=function(e,t,n){var s,u,c,l,d=o.default[e.marktype],p=d.bound,f=e.items,h=f&&f.length;if(d.nested)return h?c=f[0]:(a.mark=e,c=a),l=Object(i.default)(c,p,n),t=t&&t.union(l)||l;if(t=t||e.bounds&&e.bounds.clear()||new r.default,h)for(s=0,u=f.length;s<u;++s)t.union(Object(i.default)(f[s],p,n));return e.bounds=t}},"./node_modules/vega-scenegraph/src/bound/boundStroke.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){return t.stroke&&0!==t.opacity&&0!==t.strokeOpacity&&e.expand(null!=t.strokeWidth?+t.strokeWidth:1),e}},"./node_modules/vega-scenegraph/src/marks/arc.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/path/shapes.js"),i=n("./node_modules/vega-scenegraph/src/marks/markItemPath.js");t.default=Object(i.default)("arc",r.arc)},"./node_modules/vega-scenegraph/src/marks/area.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/path/shapes.js"),i=n("./node_modules/vega-scenegraph/src/marks/markMultiItemPath.js");t.default=Object(i.default)("area",r.area)},"./node_modules/vega-scenegraph/src/marks/group.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/path/shapes.js"),i=n("./node_modules/vega-scenegraph/src/bound/boundStroke.js"),o=n("./node_modules/vega-scenegraph/src/util/visit.js"),a=n("./node_modules/vega-scenegraph/src/util/canvas/stroke.js"),s=n("./node_modules/vega-scenegraph/src/util/canvas/fill.js"),u=n("./node_modules/vega-scenegraph/src/util/svg/clip.js"),c=n("./node_modules/vega-scenegraph/src/util/svg/translateItem.js"),l=.5;t.default={type:"group",tag:"g",nested:!1,attr:function(e,t){e("transform",Object(c.default)(t))},bound:function(e,t){if(!t.clip&&t.items)for(var n=t.items,r=0,o=n.length;r<o;++r)e.union(n[r].bounds);return(t.clip||t.width||t.height)&&Object(i.default)(e.add(0,0).add(t.width||0,t.height||0),t),e.translate(t.x||0,t.y||0)},draw:function(e,t,n){var i=this;Object(o.visit)(t,function(t){var u,c,d=t.x||0,p=t.y||0,f=t.width||0,h=t.height||0;e.save(),e.translate(d,p),(t.stroke||t.fill)&&(c=null==t.opacity?1:t.opacity)>0&&(e.beginPath(),u=t.stroke?l:0,Object(r.rectangle)(e,t,u,u),t.fill&&Object(s.default)(e,t,c)&&e.fill(),t.stroke&&Object(a.default)(e,t,c)&&e.stroke()),t.clip&&(e.beginPath(),e.rect(0,0,f,h),e.clip()),n&&n.translate(-d,-p),Object(o.visit)(t,function(t){i.draw(e,t,n)}),n&&n.translate(d,p),e.restore()})},pick:function(e,t,n,r,i,a){if(t.bounds&&!t.bounds.contains(i,a)||!t.items)return null;var s=this;return Object(o.pickVisit)(t,function(u){var c,l,d,p;if(!(p=u.bounds)||p.contains(i,a))return l=u.x||0,d=u.y||0,e.save(),e.translate(l,d),l=i-l,d=a-d,c=Object(o.pickVisit)(u,function(e){return i=l,o=d,(!1!==(t=e).interactive||"group"===t.marktype)&&t.bounds&&t.bounds.contains(i,o)?s.pick(e,n,r,l,d):null;var t,i,o}),e.restore(),c||((c=!1!==t.interactive&&(u.fill||u.stroke)&&l>=0&&l<=u.width&&d>=0&&d<=u.height)?u:null)})},background:function(e,t){var n=t.stroke?l:0;e("class","background"),e("d",Object(r.rectangle)(null,t,n,n))},foreground:function(e,t,n){e("clip-path",t.clip?Object(u.default)(n,t,t):null)}}},"./node_modules/vega-scenegraph/src/marks/image.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/util/visit.js"),i=n("./node_modules/vega-scenegraph/src/util/canvas/pick.js"),o=n("./node_modules/vega-scenegraph/src/util/svg/translate.js");function a(e,t){var n=e.image;return n&&n.url===e.url||(n={loaded:!1,width:0,height:0},t.loadImage(e.url).then(function(t){e.image=t,e.image.url=e.url})),n}function s(e,t){return"center"===e?t/2:"right"===e?t:0}function u(e,t){return"middle"===e?t/2:"bottom"===e?t:0}t.default={type:"image",tag:"image",nested:!1,attr:function(e,t,n){var r=a(t,n),i=t.x||0,c=t.y||0,l=(null!=t.width?t.width:r.width)||0,d=(null!=t.height?t.height:r.height)||0,p=!1===t.aspect?"none":"xMidYMid";i-=s(t.align,l),c-=u(t.baseline,d),e("href",r.src||"","http://www.w3.org/1999/xlink","xlink:href"),e("transform",Object(o.default)(i,c)),e("width",l),e("height",d),e("preserveAspectRatio",p)},bound:function(e,t){var n=t.image,r=t.x||0,i=t.y||0,o=(null!=t.width?t.width:n&&n.width)||0,a=(null!=t.height?t.height:n&&n.height)||0;return r-=s(t.align,o),i-=u(t.baseline,a),e.set(r,i,r+o,i+a)},draw:function(e,t,n){var i=this;Object(r.visit)(t,function(t){if(!n||n.intersects(t.bounds)){var r,o,c,l,d=a(t,i),p=t.x||0,f=t.y||0,h=(null!=t.width?t.width:d.width)||0,m=(null!=t.height?t.height:d.height)||0;p-=s(t.align,h),f-=u(t.baseline,m),!1!==t.aspect&&(o=d.width/d.height,c=t.width/t.height,o==o&&c==c&&o!==c&&(c<o?(f+=(m-(l=h/o))/2,m=l):(p+=(h-(l=m*o))/2,h=l))),d.loaded&&(e.globalAlpha=null!=(r=t.opacity)?r:1,e.drawImage(d,p,f,h,m))}})},pick:Object(i.pick)(),get:a,xOffset:s,yOffset:u}},"./node_modules/vega-scenegraph/src/marks/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/marks/arc.js"),i=n("./node_modules/vega-scenegraph/src/marks/area.js"),o=n("./node_modules/vega-scenegraph/src/marks/group.js"),a=n("./node_modules/vega-scenegraph/src/marks/image.js"),s=n("./node_modules/vega-scenegraph/src/marks/line.js"),u=n("./node_modules/vega-scenegraph/src/marks/path.js"),c=n("./node_modules/vega-scenegraph/src/marks/rect.js"),l=n("./node_modules/vega-scenegraph/src/marks/rule.js"),d=n("./node_modules/vega-scenegraph/src/marks/shape.js"),p=n("./node_modules/vega-scenegraph/src/marks/symbol.js"),f=n("./node_modules/vega-scenegraph/src/marks/text.js"),h=n("./node_modules/vega-scenegraph/src/marks/trail.js");t.default={arc:r.default,area:i.default,group:o.default,image:a.default,line:s.default,path:u.default,rect:c.default,rule:l.default,shape:d.default,symbol:p.default,text:f.default,trail:h.default}},"./node_modules/vega-scenegraph/src/marks/line.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/path/shapes.js"),i=n("./node_modules/vega-scenegraph/src/marks/markMultiItemPath.js");t.default=Object(i.default)("line",r.line)},"./node_modules/vega-scenegraph/src/marks/markItemPath.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/bound/boundStroke.js"),i=n("./node_modules/vega-scenegraph/src/bound/boundContext.js"),o=n("./node_modules/vega-scenegraph/src/util/canvas/draw.js"),a=n("./node_modules/vega-scenegraph/src/util/canvas/pick.js"),s=n("./node_modules/vega-scenegraph/src/util/svg/translateItem.js");t.default=function(e,t){function n(e,n){var r=n.x||0,i=n.y||0;e.translate(r,i),e.beginPath(),t(e,n),e.translate(-r,-i)}return{type:e,tag:"path",nested:!1,attr:function(e,n){e("transform",Object(s.default)(n)),e("d",t(null,n))},bound:function(e,n){return t(Object(i.default)(e),n),Object(r.default)(e,n).translate(n.x||0,n.y||0)},draw:Object(o.drawAll)(n),pick:Object(a.pickPath)(n)}}},"./node_modules/vega-scenegraph/src/marks/markMultiItemPath.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/bound/boundStroke.js"),i=n("./node_modules/vega-scenegraph/src/bound/boundContext.js"),o=n("./node_modules/vega-scenegraph/src/util/canvas/draw.js"),a=n("./node_modules/vega-scenegraph/src/util/canvas/pick.js");t.default=function(e,t){function n(e,n){e.beginPath(),t(e,n)}var s=Object(a.hitPath)(n);return{type:e,tag:"path",nested:!0,attr:function(e,n){var r=n.mark.items;r.length&&e("d",t(null,r))},bound:function(e,n){var o=n.items;return 0===o.length?e:(t(Object(i.default)(e),o),Object(r.default)(e,o[0]))},draw:Object(o.drawOne)(n),pick:function(e,t,n,r,i,o){var a=t.items,u=t.bounds;return!a||!a.length||u&&!u.contains(i,o)?null:(e.pixelRatio>1&&(n*=e.pixelRatio,r*=e.pixelRatio),s(e,a,n,r)?a[0]:null)}}}},"./node_modules/vega-scenegraph/src/marks/path.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/bound/boundStroke.js"),i=n("./node_modules/vega-scenegraph/src/bound/boundContext.js"),o=n("./node_modules/vega-scenegraph/src/path/parse.js"),a=n("./node_modules/vega-scenegraph/src/path/render.js"),s=n("./node_modules/vega-scenegraph/src/util/canvas/draw.js"),u=n("./node_modules/vega-scenegraph/src/util/canvas/pick.js"),c=n("./node_modules/vega-scenegraph/src/util/svg/translateItem.js");function l(e,t){var n=t.path;if(null==n)return!0;var r=t.pathCache;r&&r.path===n||((t.pathCache=r=Object(o.default)(n)).path=n),Object(a.default)(e,r,t.x,t.y)}t.default={type:"path",tag:"path",nested:!1,attr:function(e,t){e("transform",Object(c.default)(t)),e("d",t.path)},bound:function(e,t){return l(Object(i.default)(e),t)?e.set(0,0,0,0):Object(r.default)(e,t)},draw:Object(s.drawAll)(l),pick:Object(u.pickPath)(l)}},"./node_modules/vega-scenegraph/src/marks/rect.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/bound/boundStroke.js"),i=n("./node_modules/vega-scenegraph/src/path/shapes.js"),o=n("./node_modules/vega-scenegraph/src/util/canvas/draw.js"),a=n("./node_modules/vega-scenegraph/src/util/canvas/pick.js");function s(e,t){e.beginPath(),Object(i.rectangle)(e,t)}t.default={type:"rect",tag:"path",nested:!1,attr:function(e,t){e("d",Object(i.rectangle)(null,t))},bound:function(e,t){var n,i;return Object(r.default)(e.set(n=t.x||0,i=t.y||0,n+t.width||0,i+t.height||0),t)},draw:Object(o.drawAll)(s),pick:Object(a.pickPath)(s)}},"./node_modules/vega-scenegraph/src/marks/rule.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/bound/boundStroke.js"),i=n("./node_modules/vega-scenegraph/src/util/visit.js"),o=n("./node_modules/vega-scenegraph/src/util/canvas/pick.js"),a=n("./node_modules/vega-scenegraph/src/util/canvas/stroke.js"),s=n("./node_modules/vega-scenegraph/src/util/svg/translateItem.js");function u(e,t,n){var r,i,o,s;return!(!t.stroke||!Object(a.default)(e,t,n))&&(r=t.x||0,i=t.y||0,o=null!=t.x2?t.x2:r,s=null!=t.y2?t.y2:i,e.beginPath(),e.moveTo(r,i),e.lineTo(o,s),!0)}t.default={type:"rule",tag:"line",nested:!1,attr:function(e,t){e("transform",Object(s.default)(t)),e("x2",null!=t.x2?t.x2-(t.x||0):0),e("y2",null!=t.y2?t.y2-(t.y||0):0)},bound:function(e,t){var n,i;return Object(r.default)(e.set(n=t.x||0,i=t.y||0,null!=t.x2?t.x2:n,null!=t.y2?t.y2:i),t)},draw:function(e,t,n){Object(i.visit)(t,function(t){if(!n||n.intersects(t.bounds)){var r=null==t.opacity?1:t.opacity;r&&u(e,t,r)&&e.stroke()}})},pick:Object(o.pick)(function(e,t,n,r){return!!e.isPointInStroke&&u(e,t,1)&&e.isPointInStroke(n,r)})}},"./node_modules/vega-scenegraph/src/marks/shape.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/path/shapes.js"),i=n("./node_modules/vega-scenegraph/src/marks/markItemPath.js");t.default=Object(i.default)("shape",r.shape)},"./node_modules/vega-scenegraph/src/marks/symbol.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/path/shapes.js"),i=n("./node_modules/vega-scenegraph/src/marks/markItemPath.js");t.default=Object(i.default)("symbol",r.symbol)},"./node_modules/vega-scenegraph/src/marks/text.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/Bounds.js"),i=n("./node_modules/vega-scenegraph/src/util/text.js"),o=n("./node_modules/vega-scenegraph/src/util/visit.js"),a=n("./node_modules/vega-scenegraph/src/util/canvas/fill.js"),s=n("./node_modules/vega-scenegraph/src/util/canvas/pick.js"),u=n("./node_modules/vega-scenegraph/src/util/canvas/stroke.js"),c=n("./node_modules/vega-scenegraph/src/util/svg/translate.js"),l={left:"start",center:"middle",right:"end"},d=new r.default;function p(e,t,n){var r,o,a=i.textMetrics.height(t),s=t.align,u=t.radius||0,c=t.x||0,l=t.y||0,d=t.dx||0,p=(t.dy||0)+Object(i.offset)(t)-Math.round(.8*a);return u&&(o=(t.theta||0)-Math.PI/2,c+=u*Math.cos(o),l+=u*Math.sin(o)),r=i.textMetrics.width(t),"center"===s?d-=r/2:"right"===s&&(d-=r),e.set(d+=c,p+=l,d+r,p+a),t.angle&&!n&&e.rotate(t.angle*Math.PI/180,c,l),e.expand(n||!r?0:1)}t.default={type:"text",tag:"text",nested:!1,attr:function(e,t){var n,r=t.dx||0,o=(t.dy||0)+Object(i.offset)(t),a=t.x||0,s=t.y||0,u=t.angle||0,d=t.radius||0;d&&(n=(t.theta||0)-Math.PI/2,a+=d*Math.cos(n),s+=d*Math.sin(n)),e("text-anchor",l[t.align]||"start"),u?(n=Object(c.default)(a,s)+" rotate("+u+")",(r||o)&&(n+=" "+Object(c.default)(r,o))):n=Object(c.default)(a+r,s+o),e("transform",n)},bound:p,draw:function(e,t,n){Object(o.visit)(t,function(t){var r,o,s,c,l,d;n&&!n.intersects(t.bounds)||(d=Object(i.textValue)(t))&&0!==(r=null==t.opacity?1:t.opacity)&&(e.font=Object(i.font)(t),e.textAlign=t.align||"left",o=t.x||0,s=t.y||0,(c=t.radius)&&(l=(t.theta||0)-Math.PI/2,o+=c*Math.cos(l),s+=c*Math.sin(l)),t.angle&&(e.save(),e.translate(o,s),e.rotate(t.angle*Math.PI/180),o=s=0),o+=t.dx||0,s+=(t.dy||0)+Object(i.offset)(t),t.fill&&Object(a.default)(e,t,r)&&e.fillText(d,o,s),t.stroke&&Object(u.default)(e,t,r)&&e.strokeText(d,o,s),t.angle&&e.restore())})},pick:Object(s.pick)(function(e,t,n,r,i,o){if(t.fontSize<=0)return!1;if(!t.angle)return!0;var a=p(d,t,!0),s=-t.angle*Math.PI/180,u=Math.cos(s),c=Math.sin(s),l=t.x,f=t.y,h=u*i-c*o+(l-l*u+f*c),m=c*i+u*o+(f-l*c-f*u);return a.contains(h,m)})}},"./node_modules/vega-scenegraph/src/marks/trail.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/path/shapes.js"),i=n("./node_modules/vega-scenegraph/src/marks/markMultiItemPath.js");t.default=Object(i.default)("trail",r.trail)},"./node_modules/vega-scenegraph/src/modules.js":function(e,t,n){"use strict";n.r(t),n.d(t,"RenderType",function(){return u}),n.d(t,"renderModule",function(){return l});var r=n("./node_modules/vega-scenegraph/src/CanvasHandler.js"),i=n("./node_modules/vega-scenegraph/src/CanvasRenderer.js"),o=n("./node_modules/vega-scenegraph/src/SVGHandler.js"),a=n("./node_modules/vega-scenegraph/src/SVGRenderer.js"),s=n("./node_modules/vega-scenegraph/src/SVGStringRenderer.js"),u={Canvas:"canvas",PNG:"png",SVG:"svg",None:"none"},c={};function l(e,t){return e=String(e||"").toLowerCase(),arguments.length>1?(c[e]=t,this):c[e]}c.canvas=c.png={renderer:i.default,headless:i.default,handler:r.default},c.svg={renderer:a.default,headless:s.default,handler:o.default},c.none={}},"./node_modules/vega-scenegraph/src/path/arc.js":function(e,t,n){"use strict";n.r(t),n.d(t,"segmentCache",function(){return r}),n.d(t,"bezierCache",function(){return i}),n.d(t,"segments",function(){return a}),n.d(t,"bezier",function(){return s});var r={},i={},o=[].join;function a(e,t,n,i,a,s,u,c,l){var d=o.call(arguments);if(r[d])return r[d];var p=u*(Math.PI/180),f=Math.sin(p),h=Math.cos(p),m=h*(c-e)*.5+f*(l-t)*.5,g=h*(l-t)*.5-f*(c-e)*.5,v=m*m/((n=Math.abs(n))*n)+g*g/((i=Math.abs(i))*i);v>1&&(n*=v=Math.sqrt(v),i*=v);var _=h/n,y=f/n,b=-f/i,j=h/i,x=_*c+y*l,O=b*c+j*l,w=_*e+y*t,C=b*e+j*t,E=1/((w-x)*(w-x)+(C-O)*(C-O))-.25;E<0&&(E=0);var S=Math.sqrt(E);s==a&&(S=-S);var k=.5*(x+w)-S*(C-O),M=.5*(O+C)+S*(w-x),A=Math.atan2(O-M,x-k),T=Math.atan2(C-M,w-k)-A;T<0&&1===s?T+=2*Math.PI:T>0&&0===s&&(T-=2*Math.PI);for(var D=Math.ceil(Math.abs(T/(.5*Math.PI+.001))),I=[],P=0;P<D;++P){var R=A+P*T/D,N=A+(P+1)*T/D;I[P]=[k,M,R,N,n,i,f,h]}return r[d]=I}function s(e){var t=o.call(e);if(i[t])return i[t];var n=e[0],r=e[1],a=e[2],s=e[3],u=e[4],c=e[5],l=e[6],d=e[7],p=d*u,f=-l*c,h=l*u,m=d*c,g=Math.cos(a),v=Math.sin(a),_=Math.cos(s),y=Math.sin(s),b=.5*(s-a),j=Math.sin(.5*b),x=8/3*j*j/Math.sin(b),O=n+g-x*v,w=r+v+x*g,C=n+_,E=r+y,S=C+x*y,k=E-x*_;return i[t]=[p*O+f*w,h*O+m*w,p*S+f*k,h*S+m*k,p*C+f*E,h*C+m*E]}},"./node_modules/vega-scenegraph/src/path/curves.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/d3-shape/index.js"),i={basis:{curve:r.curveBasis},"basis-closed":{curve:r.curveBasisClosed},"basis-open":{curve:r.curveBasisOpen},bundle:{curve:r.curveBundle,tension:"beta",value:.85},cardinal:{curve:r.curveCardinal,tension:"tension",value:0},"cardinal-open":{curve:r.curveCardinalOpen,tension:"tension",value:0},"cardinal-closed":{curve:r.curveCardinalClosed,tension:"tension",value:0},"catmull-rom":{curve:r.curveCatmullRom,tension:"alpha",value:.5},"catmull-rom-closed":{curve:r.curveCatmullRomClosed,tension:"alpha",value:.5},"catmull-rom-open":{curve:r.curveCatmullRomOpen,tension:"alpha",value:.5},linear:{curve:r.curveLinear},"linear-closed":{curve:r.curveLinearClosed},monotone:{horizontal:r.curveMonotoneY,vertical:r.curveMonotoneX},natural:{curve:r.curveNatural},step:{curve:r.curveStep},"step-after":{curve:r.curveStepAfter},"step-before":{curve:r.curveStepBefore}};function o(e,t,n){var r=i.hasOwnProperty(e)&&i[e],o=null;return r&&(o=r.curve||r[t||"vertical"],r.tension&&null!=n&&(o=o[r.tension](n))),o}},"./node_modules/vega-scenegraph/src/path/parse.js":function(e,t,n){"use strict";n.r(t);var r={m:2,l:2,h:1,v:1,c:6,s:4,q:4,t:2,a:7},i=[/([MLHVCSQTAZmlhvcsqtaz])/g,/###/,/(\d)([-+])/g,/\s|,|###/];t.default=function(e){var t,n,o,a,s,u,c,l,d,p,f,h=[];for(l=0,p=(t=e.slice().replace(i[0],"###$1").split(i[1]).slice(1)).length;l<p;++l){for(o=(n=t[l]).slice(1).trim().replace(i[2],"$1###$2").split(i[3]),a=[u=n.charAt(0)],d=0,f=o.length;d<f;++d)(s=+o[d])===s&&a.push(s);if(c=r[u.toLowerCase()],a.length-1>c)for(d=1,f=a.length;d<f;d+=c)h.push([u].concat(a.slice(d,d+c)));else h.push(a)}return h}},"./node_modules/vega-scenegraph/src/path/rectangle.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-path/index.js");function i(e){return e.x}function o(e){return e.y}function a(e){return e.width}function s(e){return e.height}function u(e){return function(){return e}}t.default=function(){var e=i,t=o,n=a,c=s,l=u(0),d=null;function p(i,o,a){var s,u=null!=o?o:+e.call(this,i),p=null!=a?a:+t.call(this,i),f=+n.call(this,i),h=+c.call(this,i),m=+l.call(this,i);if(d||(d=s=Object(r.path)()),m<=0)d.rect(u,p,f,h);else{var g=u+f,v=p+h;d.moveTo(u+m,p),d.lineTo(g-m,p),d.quadraticCurveTo(g,p,g,p+m),d.lineTo(g,v-m),d.quadraticCurveTo(g,v,g-m,v),d.lineTo(u+m,v),d.quadraticCurveTo(u,v,u,v-m),d.lineTo(u,p+m),d.quadraticCurveTo(u,p,u+m,p),d.closePath()}if(s)return d=null,s+""||null}return p.x=function(t){return arguments.length?(e="function"==typeof t?t:u(+t),p):e},p.y=function(e){return arguments.length?(t="function"==typeof e?e:u(+e),p):t},p.width=function(e){return arguments.length?(n="function"==typeof e?e:u(+e),p):n},p.height=function(e){return arguments.length?(c="function"==typeof e?e:u(+e),p):c},p.cornerRadius=function(e){return arguments.length?(l="function"==typeof e?e:u(+e),p):l},p.context=function(e){return arguments.length?(d=null==e?null:e,p):d},p}},"./node_modules/vega-scenegraph/src/path/render.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/path/arc.js"),i=["l",0,0,0,0,0,0,0];function o(e,t){var n=i[0]=e[0];if("a"===n||"A"===n)i[1]=t*e[1],i[2]=t*e[2],i[6]=t*e[6],i[7]=t*e[7];else for(var r=1,o=e.length;r<o;++r)i[r]=t*e[r];return i}function a(e,t,n,i){for(var o=Object(r.segments)(i[5],i[6],i[0],i[1],i[3],i[4],i[2],t,n),a=0;a<o.length;++a){var s=Object(r.bezier)(o[a]);e.bezierCurveTo(s[0],s[1],s[2],s[3],s[4],s[5])}}t.default=function(e,t,n,r,i){var s,u,c,l,d,p=null,f=0,h=0,m=0,g=0;null==n&&(n=0),null==r&&(r=0),null==i&&(i=1),e.beginPath&&e.beginPath();for(var v=0,_=t.length;v<_;++v){switch(s=t[v],1!==i&&(s=o(s,i)),s[0]){case"l":f+=s[1],h+=s[2],e.lineTo(f+n,h+r);break;case"L":f=s[1],h=s[2],e.lineTo(f+n,h+r);break;case"h":f+=s[1],e.lineTo(f+n,h+r);break;case"H":f=s[1],e.lineTo(f+n,h+r);break;case"v":h+=s[1],e.lineTo(f+n,h+r);break;case"V":h=s[1],e.lineTo(f+n,h+r);break;case"m":f+=s[1],h+=s[2],e.moveTo(f+n,h+r);break;case"M":f=s[1],h=s[2],e.moveTo(f+n,h+r);break;case"c":u=f+s[5],c=h+s[6],m=f+s[3],g=h+s[4],e.bezierCurveTo(f+s[1]+n,h+s[2]+r,m+n,g+r,u+n,c+r),f=u,h=c;break;case"C":f=s[5],h=s[6],m=s[3],g=s[4],e.bezierCurveTo(s[1]+n,s[2]+r,m+n,g+r,f+n,h+r);break;case"s":u=f+s[3],c=h+s[4],m=2*f-m,g=2*h-g,e.bezierCurveTo(m+n,g+r,f+s[1]+n,h+s[2]+r,u+n,c+r),m=f+s[1],g=h+s[2],f=u,h=c;break;case"S":u=s[3],c=s[4],m=2*f-m,g=2*h-g,e.bezierCurveTo(m+n,g+r,s[1]+n,s[2]+r,u+n,c+r),f=u,h=c,m=s[1],g=s[2];break;case"q":u=f+s[3],c=h+s[4],m=f+s[1],g=h+s[2],e.quadraticCurveTo(m+n,g+r,u+n,c+r),f=u,h=c;break;case"Q":u=s[3],c=s[4],e.quadraticCurveTo(s[1]+n,s[2]+r,u+n,c+r),f=u,h=c,m=s[1],g=s[2];break;case"t":u=f+s[1],c=h+s[2],null===p[0].match(/[QqTt]/)?(m=f,g=h):"t"===p[0]?(m=2*f-l,g=2*h-d):"q"===p[0]&&(m=2*f-m,g=2*h-g),l=m,d=g,e.quadraticCurveTo(m+n,g+r,u+n,c+r),h=c,m=(f=u)+s[1],g=h+s[2];break;case"T":u=s[1],c=s[2],m=2*f-m,g=2*h-g,e.quadraticCurveTo(m+n,g+r,u+n,c+r),f=u,h=c;break;case"a":a(e,f+n,h+r,[s[1],s[2],s[3],s[4],s[5],s[6]+f+n,s[7]+h+r]),f+=s[6],h+=s[7];break;case"A":a(e,f+n,h+r,[s[1],s[2],s[3],s[4],s[5],s[6]+n,s[7]+r]),f=s[6],h=s[7];break;case"z":case"Z":e.closePath()}p=s}}},"./node_modules/vega-scenegraph/src/path/shapes.js":function(e,t,n){"use strict";n.r(t),n.d(t,"arc",function(){return y}),n.d(t,"area",function(){return b}),n.d(t,"line",function(){return j}),n.d(t,"rectangle",function(){return x}),n.d(t,"shape",function(){return O}),n.d(t,"symbol",function(){return w}),n.d(t,"trail",function(){return C});var r=n("./node_modules/vega-scenegraph/src/path/curves.js"),i=n("./node_modules/vega-scenegraph/src/path/symbols.js"),o=n("./node_modules/vega-scenegraph/src/path/rectangle.js"),a=n("./node_modules/vega-scenegraph/src/path/trail.js"),s=n("./node_modules/d3-shape/index.js");function u(e){return e.x||0}function c(e){return e.y||0}function l(e){return e.cornerRadius||0}function d(e){return!(!1===e.defined)}var p=Object(s.arc)().startAngle(function(e){return e.startAngle||0}).endAngle(function(e){return e.endAngle||0}).padAngle(function(e){return e.padAngle||0}).innerRadius(function(e){return e.innerRadius||0}).outerRadius(function(e){return e.outerRadius||0}).cornerRadius(l),f=Object(s.area)().x(u).y1(c).y0(function(e){return(e.y||0)+(e.height||0)}).defined(d),h=Object(s.area)().y(c).x1(u).x0(function(e){return(e.x||0)+(e.width||0)}).defined(d),m=Object(s.line)().x(u).y(c).defined(d),g=Object(o.default)().x(u).y(c).width(function(e){return e.width||0}).height(function(e){return e.height||0}).cornerRadius(l),v=Object(s.symbol)().type(function(e){return Object(i.default)(e.shape||"circle")}).size(function(e){return null==e.size?64:e.size}),_=Object(a.default)().x(u).y(c).defined(d).size(function(e){return e.size||1});function y(e,t){return p.context(e)(t)}function b(e,t){var n=t[0],i=n.interpolate||"linear";return("horizontal"===n.orient?h:f).curve(Object(r.default)(i,n.orient,n.tension)).context(e)(t)}function j(e,t){var n=t[0],i=n.interpolate||"linear";return m.curve(Object(r.default)(i,n.orient,n.tension)).context(e)(t)}function x(e,t,n,r){return g.context(e)(t,n,r)}function O(e,t){return(t.mark.shape||t.shape).context(e)(t)}function w(e,t){return v.context(e)(t)}function C(e,t){return _.context(e)(t)}},"./node_modules/vega-scenegraph/src/path/symbols.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return u});var r=n("./node_modules/vega-scenegraph/src/path/parse.js"),i=n("./node_modules/vega-scenegraph/src/path/render.js"),o=2*Math.PI,a=Math.sqrt(3)/2,s={circle:{draw:function(e,t){var n=Math.sqrt(t)/2;e.moveTo(n,0),e.arc(0,0,n,0,o)}},cross:{draw:function(e,t){var n=Math.sqrt(t)/2,r=n/2.5;e.moveTo(-n,-r),e.lineTo(-n,r),e.lineTo(-r,r),e.lineTo(-r,n),e.lineTo(r,n),e.lineTo(r,r),e.lineTo(n,r),e.lineTo(n,-r),e.lineTo(r,-r),e.lineTo(r,-n),e.lineTo(-r,-n),e.lineTo(-r,-r),e.closePath()}},diamond:{draw:function(e,t){var n=Math.sqrt(t)/2;e.moveTo(-n,0),e.lineTo(0,-n),e.lineTo(n,0),e.lineTo(0,n),e.closePath()}},square:{draw:function(e,t){var n=Math.sqrt(t),r=-n/2;e.rect(r,r,n,n)}},"triangle-up":{draw:function(e,t){var n=Math.sqrt(t)/2,r=a*n;e.moveTo(0,-r),e.lineTo(-n,r),e.lineTo(n,r),e.closePath()}},"triangle-down":{draw:function(e,t){var n=Math.sqrt(t)/2,r=a*n;e.moveTo(0,r),e.lineTo(-n,-r),e.lineTo(n,-r),e.closePath()}},"triangle-right":{draw:function(e,t){var n=Math.sqrt(t)/2,r=a*n;e.moveTo(r,0),e.lineTo(-r,-n),e.lineTo(-r,n),e.closePath()}},"triangle-left":{draw:function(e,t){var n=Math.sqrt(t)/2,r=a*n;e.moveTo(-r,0),e.lineTo(r,-n),e.lineTo(r,n),e.closePath()}}};function u(e){return s.hasOwnProperty(e)?s[e]:function(e){if(!c.hasOwnProperty(e)){var t=Object(r.default)(e);c[e]={draw:function(e,n){Object(i.default)(e,t,0,0,Math.sqrt(n)/2)}}}return c[e]}(e)}var c={}},"./node_modules/vega-scenegraph/src/path/trail.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/d3-path/index.js"),i=Math.PI;t.default=function(){var e,t,n,o,a,s,u,c,l=null;function d(e,t,n){var r=n/2;if(a){var o=u-t,d=e-s;if(o||d){var p=Math.sqrt(o*o+d*d),f=(o/=p)*c,h=(d/=p)*c,m=Math.atan2(d,o);l.moveTo(s-f,u-h),l.lineTo(e-o*r,t-d*r),l.arc(e,t,r,m-i,m),l.lineTo(s+f,u+h),l.arc(s,u,c,m,m+i)}else l.arc(e,t,r,0,2*i);l.closePath()}else a=1;s=e,u=t,c=r}function p(i){var s,u,c,p=i.length,f=!1;for(null==l&&(l=c=Object(r.path)()),s=0;s<=p;++s)!(s<p&&o(u=i[s],s,i))===f&&(f=!f)&&(a=0),f&&d(+e(u,s,i),+t(u,s,i),+n(u,s,i));if(c)return l=null,c+""||null}return p.x=function(t){return arguments.length?(e=t,p):e},p.y=function(e){return arguments.length?(t=e,p):t},p.size=function(e){return arguments.length?(n=e,p):n},p.defined=function(e){return arguments.length?(o=e,p):o},p.context=function(e){return arguments.length?(l=null==e?null:e,p):l},p}},"./node_modules/vega-scenegraph/src/util/canvas/clip.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/index.js");t.default=function(e,t){var n=t.clip;if(e.save(),e.beginPath(),Object(r.isFunction)(n))n(e);else{var i=t.group;e.rect(0,0,i.width||0,i.height||0)}e.clip()}},"./node_modules/vega-scenegraph/src/util/canvas/color.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/util/canvas/gradient.js");t.default=function(e,t,n){return n.id?Object(r.default)(e,n,t.bounds):n}},"./node_modules/vega-scenegraph/src/util/canvas/draw.js":function(e,t,n){"use strict";n.r(t),n.d(t,"drawAll",function(){return a}),n.d(t,"drawOne",function(){return s});var r=n("./node_modules/vega-scenegraph/src/util/canvas/fill.js"),i=n("./node_modules/vega-scenegraph/src/util/canvas/stroke.js"),o=n("./node_modules/vega-scenegraph/src/util/visit.js");function a(e){return function(t,n,r){Object(o.visit)(n,function(n){r&&!r.intersects(n.bounds)||u(e,t,n,n)})}}function s(e){return function(t,n,r){!n.items.length||r&&!r.intersects(n.bounds)||u(e,t,n.items[0],n.items)}}function u(e,t,n,o){var a=null==n.opacity?1:n.opacity;0!==a&&(e(t,o)||(n.fill&&Object(r.default)(t,n,a)&&t.fill(),n.stroke&&Object(i.default)(t,n,a)&&t.stroke()))}},"./node_modules/vega-scenegraph/src/util/canvas/fill.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/util/canvas/color.js");t.default=function(e,t,n){return(n*=null==t.fillOpacity?1:t.fillOpacity)>0&&(e.globalAlpha=n,e.fillStyle=Object(r.default)(e,t,t.fill),!0)}},"./node_modules/vega-scenegraph/src/util/canvas/gradient.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n){for(var r=n.width(),i=n.height(),o=n.x1+t.x1*r,a=n.y1+t.y1*i,s=n.x1+t.x2*r,u=n.y1+t.y2*i,c=t.stops,l=0,d=c.length,p=e.createLinearGradient(o,a,s,u);l<d;++l)p.addColorStop(c[l].offset,c[l].color);return p}},"./node_modules/vega-scenegraph/src/util/canvas/pick.js":function(e,t,n){"use strict";n.r(t),n.d(t,"pick",function(){return o}),n.d(t,"hitPath",function(){return a}),n.d(t,"pickPath",function(){return s});var r=n("./node_modules/vega-scenegraph/src/util/visit.js"),i=function(){return!0};function o(e){return e||(e=i),function(t,n,i,o,a,s){return t.pixelRatio>1&&(i*=t.pixelRatio,o*=t.pixelRatio),Object(r.pickVisit)(n,function(n){var r=n.bounds;if((!r||r.contains(a,s))&&r)return e(t,n,i,o,a,s)?n:void 0})}}function a(e,t){return function(n,r,i,o){var a,s,u=Array.isArray(r)?r[0]:r,c=null==t?u.fill:t,l=u.stroke&&n.isPointInStroke;return l&&(a=u.strokeWidth,s=u.strokeCap,n.lineWidth=null!=a?a:1,n.lineCap=null!=s?s:"butt"),!e(n,r)&&(c&&n.isPointInPath(i,o)||l&&n.isPointInStroke(i,o))}}function s(e){return o(a(e))}},"./node_modules/vega-scenegraph/src/util/canvas/resize.js":function(e,t,n){"use strict";n.r(t),n.d(t,"devicePixelRatio",function(){return r});var r="undefined"!=typeof window?window.devicePixelRatio||1:1;t.default=function(e,t,n,i,o){var a="undefined"!=typeof HTMLElement&&e instanceof HTMLElement&&null!=e.parentNode,s=e.getContext("2d"),u=a?r:o;return e.width=t*u,e.height=n*u,a&&1!==u&&(e.style.width=t+"px",e.style.height=n+"px"),s.pixelRatio=u,s.setTransform(u,0,0,u,u*i[0],u*i[1]),e}},"./node_modules/vega-scenegraph/src/util/canvas/stroke.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/util/canvas/color.js"),i=[];t.default=function(e,t,n){var o=null!=(o=t.strokeWidth)?o:1;return!(o<=0)&&((n*=null==t.strokeOpacity?1:t.strokeOpacity)>0&&(e.globalAlpha=n,e.strokeStyle=Object(r.default)(e,t,t.stroke),e.lineWidth=o,e.lineCap=t.strokeCap||"butt",e.lineJoin=t.strokeJoin||"miter",e.miterLimit=t.strokeMiterLimit||10,e.setLineDash&&(e.setLineDash(t.strokeDash||i),e.lineDashOffset=t.strokeDashOffset||0),!0))}},"./node_modules/vega-scenegraph/src/util/dom.js":function(e,t,n){"use strict";function r(e,t,n){return!e&&"undefined"!=typeof document&&document.createElement&&(e=document),e?n?e.createElementNS(n,t):e.createElement(t):null}function i(e,t){t=t.toLowerCase();for(var n=e.childNodes,r=0,i=n.length;r<i;++r)if(n[r].tagName.toLowerCase()===t)return n[r]}function o(e,t,n,i){var o,a=e.childNodes[t];return a&&a.tagName.toLowerCase()===n.toLowerCase()||(o=a||null,a=r(e.ownerDocument,n,i),e.insertBefore(a,o)),a}function a(e,t){for(var n=e.childNodes,r=n.length;r>t;)e.removeChild(n[--r]);return e}function s(e){return"mark-"+e.marktype+(e.role?" role-"+e.role:"")+(e.name?" "+e.name:"")}n.r(t),n.d(t,"domCreate",function(){return r}),n.d(t,"domFind",function(){return i}),n.d(t,"domChild",function(){return o}),n.d(t,"domClear",function(){return a}),n.d(t,"cssClass",function(){return s})},"./node_modules/vega-scenegraph/src/util/equal.js":function(e,t,n){"use strict";n.r(t),n.d(t,"sceneEqual",function(){return a}),n.d(t,"pathEqual",function(){return s});var r=n("./node_modules/vega-scenegraph/src/path/parse.js"),i=n("./node_modules/vega-util/index.js"),o=1e-9;function a(e,t,n){return e===t||("path"===n?s(e,t):e instanceof Date&&t instanceof Date?+e==+t:Object(i.isNumber)(e)&&Object(i.isNumber)(t)?Math.abs(e-t)<=o:e&&t&&(Object(i.isObject)(e)||Object(i.isObject)(t))?null!=e&&null!=t&&function(e,t){var n,r,i=Object.keys(e),o=Object.keys(t);if(i.length!==o.length)return!1;for(i.sort(),o.sort(),r=i.length-1;r>=0;r--)if(i[r]!=o[r])return!1;for(r=i.length-1;r>=0;r--)if(n=i[r],!a(e[n],t[n],n))return!1;return typeof e==typeof t}(e,t):e==t)}function s(e,t){return a(Object(r.default)(e),Object(r.default)(t))}},"./node_modules/vega-scenegraph/src/util/point.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){var n=t.getBoundingClientRect();return[e.clientX-n.left-(t.clientLeft||0),e.clientY-n.top-(t.clientTop||0)]}},"./node_modules/vega-scenegraph/src/util/serialize.js":function(e,t,n){"use strict";n.r(t),n.d(t,"sceneToJSON",function(){return o}),n.d(t,"sceneFromJSON",function(){return a});var r=n("./node_modules/vega-scenegraph/src/bound/boundMark.js"),i=["marktype","name","role","interactive","clip","items","zindex","x","y","width","height","align","baseline","fill","fillOpacity","opacity","stroke","strokeOpacity","strokeWidth","strokeCap","strokeDash","strokeDashOffset","startAngle","endAngle","innerRadius","outerRadius","cornerRadius","padAngle","interpolate","tension","orient","defined","url","path","x2","y2","size","shape","text","angle","theta","radius","dx","dy","font","fontSize","fontWeight","fontStyle","fontVariant"];function o(e,t){return JSON.stringify(e,i,t)}function a(e){return function e(t){var n,i,o,a=t.marktype,s=t.items;if(s)for(i=0,o=s.length;i<o;++i)n=a?"mark":"group",s[i][n]=t,s[i].zindex&&(s[i][n].zdirty=!0),"group"===(a||n)&&e(s[i]);a&&Object(r.default)(t);return t}("string"==typeof e?JSON.parse(e):e)}},"./node_modules/vega-scenegraph/src/util/svg/clip.js":function(e,t,n){"use strict";n.r(t),n.d(t,"resetSVGClipId",function(){return o});var r=n("./node_modules/vega-util/index.js"),i=1;function o(){i=1}t.default=function(e,t,n){var o=t.clip,a=e._defs,s=t.clip_id||(t.clip_id="clip"+i++),u=a.clipping[s]||(a.clipping[s]={id:s});return Object(r.isFunction)(o)?u.path=o(null):(u.width=n.width||0,u.height=n.height||0),"url(#"+s+")"}},"./node_modules/vega-scenegraph/src/util/svg/metadata.js":function(e,t,n){"use strict";n.r(t),t.default={version:"1.1",xmlns:"http://www.w3.org/2000/svg","xmlns:xlink":"http://www.w3.org/1999/xlink"}},"./node_modules/vega-scenegraph/src/util/svg/styles.js":function(e,t,n){"use strict";n.r(t),n.d(t,"styles",function(){return r}),n.d(t,"styleProperties",function(){return i});var r={fill:"fill",fillOpacity:"fill-opacity",stroke:"stroke",strokeOpacity:"stroke-opacity",strokeWidth:"stroke-width",strokeCap:"stroke-linecap",strokeJoin:"stroke-linejoin",strokeDash:"stroke-dasharray",strokeDashOffset:"stroke-dashoffset",strokeMiterLimit:"stroke-miterlimit",opacity:"opacity"},i=Object.keys(r)},"./node_modules/vega-scenegraph/src/util/svg/translate.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){return"translate("+e+","+t+")"}},"./node_modules/vega-scenegraph/src/util/svg/translateItem.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-scenegraph/src/util/svg/translate.js");t.default=function(e){return Object(r.default)(e.x||0,e.y||0)}},"./node_modules/vega-scenegraph/src/util/tags.js":function(e,t,n){"use strict";function r(e,t,n){var r,i,o="<"+e;if(t)for(r in t)null!=(i=t[r])&&(o+=" "+r+'="'+i+'"');return n&&(o+=" "+n),o+">"}function i(e){return"</"+e+">"}n.r(t),n.d(t,"openTag",function(){return r}),n.d(t,"closeTag",function(){return i})},"./node_modules/vega-scenegraph/src/util/text.js":function(e,t,n){"use strict";n.r(t),n.d(t,"textMetrics",function(){return a}),n.d(t,"textValue",function(){return f}),n.d(t,"truncate",function(){return h}),n.d(t,"font",function(){return m}),n.d(t,"offset",function(){return g});var r,i,o=n("./node_modules/vega-canvas/index.js"),a={height:d,measureWidth:c,estimateWidth:s,width:s,canvas:p};function s(e){return i=d(e),u(f(e))}function u(e){return~~(.8*e.length*i)}function c(e){return r.font=m(e),l(f(e))}function l(e){return r.measureText(e).width}function d(e){return null!=e.fontSize?e.fontSize:11}function p(e){r=e&&(r=Object(o.canvas)(1,1))?r.getContext("2d"):null,a.width=r?c:s}function f(e){var t=e.text;return null==t?"":e.limit>0?h(e):t+""}function h(e){var t,n=+e.limit,o=e.text+"";if(r?(r.font=m(e),t=l):(i=d(e),t=u),t(o)<n)return o;var a,s=e.ellipsis||"…",c="rtl"===e.dir,p=0,f=o.length;if(n-=t(s),c){for(;p<f;)a=p+f>>>1,t(o.slice(a))>n?p=a+1:f=a;return s+o.slice(p)}for(;p<f;)a=1+(p+f>>>1),t(o.slice(0,a))<n?p=a:f=a-1;return o.slice(0,p)+s}function m(e,t){var n=e.font;return t&&n&&(n=String(n).replace(/"/g,"'")),(e.fontStyle?e.fontStyle+" ":"")+(e.fontVariant?e.fontVariant+" ":"")+(e.fontWeight?e.fontWeight+" ":"")+d(e)+"px "+(n||"sans-serif")}function g(e){var t=e.baseline,n=d(e);return Math.round("top"===t?.79*n:"middle"===t?.3*n:"bottom"===t?-.21*n:0)}p(!0)},"./node_modules/vega-scenegraph/src/util/visit.js":function(e,t,n){"use strict";function r(e,t){return e.zindex-t.zindex||e.index-t.index}function i(e){if(!e.zdirty)return e.zitems;var t,n,i,o=e.items,a=[];for(n=0,i=o.length;n<i;++n)(t=o[n]).index=n,t.zindex&&a.push(t);return e.zdirty=!1,e.zitems=a.sort(r)}function o(e,t){var n,r,o=e.items;if(o&&o.length){var a=i(e);if(a&&a.length){for(n=0,r=o.length;n<r;++n)o[n].zindex||t(o[n]);o=a}for(n=0,r=o.length;n<r;++n)t(o[n])}}function a(e,t){var n,r,o=e.items;if(!o||!o.length)return null;var a=i(e);for(a&&a.length&&(o=a),r=o.length;--r>=0;)if(n=t(o[r]))return n;if(o===a)for(r=(o=e.items).length;--r>=0;)if(!o[r].zindex&&(n=t(o[r])))return n;return null}n.r(t),n.d(t,"zorder",function(){return i}),n.d(t,"visit",function(){return o}),n.d(t,"pickVisit",function(){return a})},"./node_modules/vega-statistics/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-statistics/src/bin.js");n.d(t,"bin",function(){return r.default});var i=n("./node_modules/vega-statistics/src/bootstrapCI.js");n.d(t,"bootstrapCI",function(){return i.default});var o=n("./node_modules/vega-statistics/src/quartiles.js");n.d(t,"quartiles",function(){return o.default});var a=n("./node_modules/vega-statistics/src/random.js");n.d(t,"random",function(){return a.random}),n.d(t,"setRandom",function(){return a.setRandom});var s=n("./node_modules/vega-statistics/src/integer.js");n.d(t,"randomInteger",function(){return s.default});var u=n("./node_modules/vega-statistics/src/kde.js");n.d(t,"randomKDE",function(){return u.default});var c=n("./node_modules/vega-statistics/src/mixture.js");n.d(t,"randomMixture",function(){return c.default});var l=n("./node_modules/vega-statistics/src/normal.js");n.d(t,"randomNormal",function(){return l.default});var d=n("./node_modules/vega-statistics/src/uniform.js");n.d(t,"randomUniform",function(){return d.default})},"./node_modules/vega-statistics/src/bin.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){var t,n,r,i,o,a,s,u,c=e.maxbins||20,l=e.base||10,d=Math.log(l),p=e.divide||[5,2],f=e.extent[0],h=e.extent[1],m=h-f;if(e.step)t=e.step;else if(e.steps){for(o=m/c,a=0,s=e.steps.length;a<s&&e.steps[a]<o;++a);t=e.steps[Math.max(0,a-1)]}else{for(n=Math.ceil(Math.log(c)/d),r=e.minstep||0,t=Math.max(r,Math.pow(l,Math.round(Math.log(m)/d)-n));Math.ceil(m/t)>c;)t*=l;for(a=0,s=p.length;a<s;++a)(o=t/p[a])>=r&&m/o<=c&&(t=o)}return i=(o=Math.log(t))>=0?0:1+~~(-o/d),u=Math.pow(l,-i-1),(e.nice||void 0===e.nice)&&(f=f<(o=Math.floor(f/t+u)*t)?o-t:o,h=Math.ceil(h/t)*t),{start:f,stop:h,step:t}}},"./node_modules/vega-statistics/src/bootstrapCI.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-statistics/src/numbers.js"),i=n("./node_modules/vega-statistics/src/random.js"),o=n("./node_modules/d3-array/index.js");t.default=function(e,t,n,a){var s,u,c,l,d=Object(r.default)(e,a),p=d.length,f=t;for(c=0,l=Array(f);c<f;++c){for(s=0,u=0;u<p;++u)s+=d[~~(Object(i.random)()*p)];l[c]=s/p}return[Object(o.quantile)(l.sort(o.ascending),n/2),Object(o.quantile)(l,1-n/2)]}},"./node_modules/vega-statistics/src/integer.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-statistics/src/random.js");t.default=function(e,t){null==t&&(t=e,e=0);var n,i,o,a={};return a.min=function(e){return arguments.length?(o=i-(n=e||0),a):n},a.max=function(e){return arguments.length?(o=(i=e||0)-n,a):i},a.sample=function(){return n+Math.floor(o*Object(r.random)())},a.pdf=function(e){return e===Math.floor(e)&&e>=n&&e<i?1/o:0},a.cdf=function(e){var t=Math.floor(e);return t<n?0:t>=i?1:(t-n+1)/o},a.icdf=function(e){return e>=0&&e<=1?n-1+Math.floor(e*o):NaN},a.min(e).max(t)}},"./node_modules/vega-statistics/src/kde.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-statistics/src/normal.js"),i=n("./node_modules/vega-statistics/src/quartiles.js"),o=n("./node_modules/vega-statistics/src/random.js"),a=n("./node_modules/d3-array/index.js");t.default=function(e,t){var n=Object(r.default)(),s={},u=0;return s.data=function(n){return arguments.length?(e=n,u=n?n.length:0,s.bandwidth(t)):e},s.bandwidth=function(n){return arguments.length?(!(t=n)&&e&&(o=(r=e).length,u=Object(i.default)(r),c=(u[2]-u[0])/1.34,t=1.06*Math.min(Math.sqrt(Object(a.variance)(r)),c)*Math.pow(o,-.2)),s):t;var r,o,u,c},s.sample=function(){return e[~~(Object(o.random)()*u)]+t*n.sample()},s.pdf=function(r){for(var i=0,o=0;o<u;++o)i+=n.pdf((r-e[o])/t);return i/t/u},s.cdf=function(r){for(var i=0,o=0;o<u;++o)i+=n.cdf((r-e[o])/t);return i/u},s.icdf=function(){throw Error("KDE icdf not supported.")},s.data(e)}},"./node_modules/vega-statistics/src/mixture.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-statistics/src/random.js");t.default=function(e,t){var n,i={},o=0;return i.weights=function(e){return arguments.length?(n=function(e){var t,n=[],r=0;for(t=0;t<o;++t)r+=n[t]=null==e[t]?1:+e[t];for(t=0;t<o;++t)n[t]/=r;return n}(t=e||[]),i):t},i.distributions=function(n){return arguments.length?(n?(o=n.length,e=n):(o=0,e=[]),i.weights(t)):e},i.sample=function(){for(var t=Object(r.random)(),i=e[o-1],a=n[0],s=0;s<o-1;a+=n[++s])if(t<a){i=e[s];break}return i.sample()},i.pdf=function(t){for(var r=0,i=0;i<o;++i)r+=n[i]*e[i].pdf(t);return r},i.cdf=function(t){for(var r=0,i=0;i<o;++i)r+=n[i]*e[i].cdf(t);return r},i.icdf=function(){throw Error("Mixture icdf not supported.")},i.distributions(e).weights(t)}},"./node_modules/vega-statistics/src/normal.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-statistics/src/random.js");t.default=function(e,t){var n,i,o=NaN,a={};return a.mean=function(e){return arguments.length?(n=e||0,o=NaN,a):n},a.stdev=function(e){return arguments.length?(i=null==e?1:e,o=NaN,a):i},a.sample=function(){var e,t,a=0,s=0;if(o==o)return a=o,o=NaN,a;do{e=(a=2*Object(r.random)()-1)*a+(s=2*Object(r.random)()-1)*s}while(0===e||e>1);return t=Math.sqrt(-2*Math.log(e)/e),o=n+s*t*i,n+a*t*i},a.pdf=function(e){var t=Math.exp(Math.pow(e-n,2)/(-2*Math.pow(i,2)));return 1/(i*Math.sqrt(2*Math.PI))*t},a.cdf=function(e){var t,r=(e-n)/i,o=Math.abs(r);if(o>37)t=0;else{var a=Math.exp(-o*o/2);o<7.07106781186547?(t=a*((((((.0352624965998911*o+.700383064443688)*o+6.37396220353165)*o+33.912866078383)*o+112.079291497871)*o+221.213596169931)*o+220.206867912376),t/=((((((.0883883476483184*o+1.75566716318264)*o+16.064177579207)*o+86.7807322029461)*o+296.564248779674)*o+637.333633378831)*o+793.826512519948)*o+440.413735824752):t=a/(o+1/(o+2/(o+3/(o+4/(o+.65)))))/2.506628274631}return r>0?1-t:t},a.icdf=function(e){if(e<=0||e>=1)return NaN;var t=2*e-1,r=8*(Math.PI-3)/(3*Math.PI*(4-Math.PI)),o=2/(Math.PI*r)+Math.log(1-Math.pow(t,2))/2,a=Math.log(1-t*t)/r,s=(t>0?1:-1)*Math.sqrt(Math.sqrt(o*o-a)-o);return n+i*Math.SQRT2*s},a.mean(e).stdev(t)}},"./node_modules/vega-statistics/src/numbers.js":function(e,t,n){"use strict";function r(e){return null===e?NaN:+e}n.r(t),t.default=function(e,t){var n,i=[],o=e.length,a=-1;if(null==t)for(;++a<o;)isNaN(n=r(e[a]))||i.push(n);else for(;++a<o;)isNaN(n=r(t(e[a],a,e)))||i.push(n);return i}},"./node_modules/vega-statistics/src/quartiles.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-statistics/src/numbers.js"),i=n("./node_modules/d3-array/index.js");t.default=function(e,t){var n=Object(r.default)(e,t);return[Object(i.quantile)(n.sort(i.ascending),.25),Object(i.quantile)(n,.5),Object(i.quantile)(n,.75)]}},"./node_modules/vega-statistics/src/random.js":function(e,t,n){"use strict";n.r(t),n.d(t,"random",function(){return r}),n.d(t,"setRandom",function(){return i});var r=Math.random;function i(e){r=e}},"./node_modules/vega-statistics/src/uniform.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-statistics/src/random.js");t.default=function(e,t){null==t&&(t=null==e?1:e,e=0);var n,i,o,a={};return a.min=function(e){return arguments.length?(o=i-(n=e||0),a):n},a.max=function(e){return arguments.length?(o=(i=e||0)-n,a):i},a.sample=function(){return n+o*Object(r.random)()},a.pdf=function(e){return e>=n&&e<=i?1/o:0},a.cdf=function(e){return e<n?0:e>i?1:(e-n)/o},a.icdf=function(e){return e>=0&&e<=1?n+e*o:NaN},a.min(e).max(t)}},"./node_modules/vega-transforms/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-transforms/src/Aggregate.js");n.d(t,"aggregate",function(){return r.default});var i=n("./node_modules/vega-transforms/src/Bin.js");n.d(t,"bin",function(){return i.default});var o=n("./node_modules/vega-transforms/src/Collect.js");n.d(t,"collect",function(){return o.default});var a=n("./node_modules/vega-transforms/src/Compare.js");n.d(t,"compare",function(){return a.default});var s=n("./node_modules/vega-transforms/src/CountPattern.js");n.d(t,"countpattern",function(){return s.default});var u=n("./node_modules/vega-transforms/src/Cross.js");n.d(t,"cross",function(){return u.default});var c=n("./node_modules/vega-transforms/src/Density.js");n.d(t,"density",function(){return c.default});var l=n("./node_modules/vega-transforms/src/Extent.js");n.d(t,"extent",function(){return l.default});var d=n("./node_modules/vega-transforms/src/Facet.js");n.d(t,"facet",function(){return d.default});var p=n("./node_modules/vega-transforms/src/Field.js");n.d(t,"field",function(){return p.default});var f=n("./node_modules/vega-transforms/src/Filter.js");n.d(t,"filter",function(){return f.default});var h=n("./node_modules/vega-transforms/src/Flatten.js");n.d(t,"flatten",function(){return h.default});var m=n("./node_modules/vega-transforms/src/Fold.js");n.d(t,"fold",function(){return m.default});var g=n("./node_modules/vega-transforms/src/Formula.js");n.d(t,"formula",function(){return g.default});var v=n("./node_modules/vega-transforms/src/Generate.js");n.d(t,"generate",function(){return v.default});var _=n("./node_modules/vega-transforms/src/Impute.js");n.d(t,"impute",function(){return _.default});var y=n("./node_modules/vega-transforms/src/JoinAggregate.js");n.d(t,"joinaggregate",function(){return y.default});var b=n("./node_modules/vega-transforms/src/Key.js");n.d(t,"key",function(){return b.default});var j=n("./node_modules/vega-transforms/src/Lookup.js");n.d(t,"lookup",function(){return j.default});var x=n("./node_modules/vega-transforms/src/MultiExtent.js");n.d(t,"multiextent",function(){return x.default});var O=n("./node_modules/vega-transforms/src/MultiValues.js");n.d(t,"multivalues",function(){return O.default});var w=n("./node_modules/vega-transforms/src/Params.js");n.d(t,"params",function(){return w.default});var C=n("./node_modules/vega-transforms/src/PreFacet.js");n.d(t,"prefacet",function(){return C.default});var E=n("./node_modules/vega-transforms/src/Project.js");n.d(t,"project",function(){return E.default});var S=n("./node_modules/vega-transforms/src/Proxy.js");n.d(t,"proxy",function(){return S.default});var k=n("./node_modules/vega-transforms/src/Relay.js");n.d(t,"relay",function(){return k.default});var M=n("./node_modules/vega-transforms/src/Sample.js");n.d(t,"sample",function(){return M.default});var A=n("./node_modules/vega-transforms/src/Sequence.js");n.d(t,"sequence",function(){return A.default});var T=n("./node_modules/vega-transforms/src/Sieve.js");n.d(t,"sieve",function(){return T.default});var D=n("./node_modules/vega-transforms/src/Subflow.js");n.d(t,"subflow",function(){return D.default});var I=n("./node_modules/vega-transforms/src/TupleIndex.js");n.d(t,"tupleindex",function(){return I.default});var P=n("./node_modules/vega-transforms/src/Values.js");n.d(t,"values",function(){return P.default});var R=n("./node_modules/vega-transforms/src/Window.js");n.d(t,"window",function(){return R.default})},"./node_modules/vega-transforms/src/Aggregate.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return u});var r=n("./node_modules/vega-transforms/src/util/AggregateKeys.js"),i=n("./node_modules/vega-transforms/src/util/AggregateOps.js"),o=n("./node_modules/vega-transforms/src/util/TupleStore.js"),a=n("./node_modules/vega-dataflow/index.js"),s=n("./node_modules/vega-util/index.js");function u(e){a.Transform.call(this,null,e),this._adds=[],this._mods=[],this._alen=0,this._mlen=0,this._drop=!0,this._cross=!1,this._dims=[],this._dnames=[],this._measures=[],this._countOnly=!1,this._counts=null,this._prev=null,this._inputs=null,this._outputs=null}u.Definition={type:"Aggregate",metadata:{generates:!0,changes:!0},params:[{name:"groupby",type:"field",array:!0},{name:"ops",type:"enum",array:!0,values:i.ValidAggregateOps},{name:"fields",type:"field",null:!0,array:!0},{name:"as",type:"string",null:!0,array:!0},{name:"drop",type:"boolean",default:!0},{name:"cross",type:"boolean",default:!1},{name:"key",type:"field"}]};var c=Object(s.inherits)(u,a.Transform);c.transform=function(e,t){var n,r=this,i=t.fork(t.NO_SOURCE|t.NO_FIELDS);return this.stamp=i.stamp,this.value&&((n=e.modified())||t.modified(this._inputs))?(this._prev=this.value,this.value=n?this.init(e):{},t.visit(t.SOURCE,function(e){r.add(e)})):(this.value=this.value||this.init(e),t.visit(t.REM,function(e){r.rem(e)}),t.visit(t.ADD,function(e){r.add(e)})),i.modifies(this._outputs),r._drop=!1!==e.drop,e.cross&&r._dims.length>1&&(r._drop=!1,this.cross()),r.changes(i)},c.cross=function(){var e=this,t=e.value,n=e._dnames,r=n.map(function(){return{}}),i=n.length;function o(e){var t,o,a,s;for(t in e)for(a=e[t].tuple,o=0;o<i;++o)r[o][s=a[n[o]]]=s}o(e._prev),o(t),function o(a,s,u){var c,l,d=n[u],p=r[u++];for(c in p)s[d]=p[c],l=a?a+"|"+c:c,u<i?o(l,s,u):t[l]||e.cell(l,s)}("",{},0)},c.init=function(e){var t=this._inputs=[],n=this._outputs=[],o={};function a(e){for(var n,r=Object(s.array)(Object(s.accessorFields)(e)),i=0,a=r.length;i<a;++i)o[n=r[i]]||(o[n]=1,t.push(n))}this._dims=Object(s.array)(e.groupby),this._dnames=this._dims.map(function(e){var t=Object(s.accessorName)(e);return a(e),n.push(t),t}),this.cellkey=e.key?e.key:Object(r.groupkey)(this._dims),this._countOnly=!0,this._counts=[],this._measures=[];var u,c,l,d,p,f,h=e.fields||[null],m=e.ops||["count"],g=e.as||[],v=h.length,_={};for(v!==m.length&&Object(s.error)("Unmatched number of fields and aggregate ops."),f=0;f<v;++f)u=h[f],c=m[f],null==u&&"count"!==c&&Object(s.error)("Null aggregate field specified."),d=Object(s.accessorName)(u),p=Object(i.measureName)(c,d,g[f]),n.push(p),"count"!==c?((l=_[d])||(a(u),(l=_[d]=[]).field=u,this._measures.push(l)),"count"!==c&&(this._countOnly=!1),l.push(Object(i.createMeasure)(c,p))):this._counts.push(p);return this._measures=this._measures.map(function(e){return Object(i.compileMeasures)(e,e.field)}),{}},c.cellkey=Object(r.groupkey)(),c.cell=function(e,t){var n=this.value[e];return n?0===n.num&&this._drop&&n.stamp<this.stamp?(n.stamp=this.stamp,this._adds[this._alen++]=n):n.stamp<this.stamp&&(n.stamp=this.stamp,this._mods[this._mlen++]=n):(n=this.value[e]=this.newcell(e,t),this._adds[this._alen++]=n),n},c.newcell=function(e,t){var n={key:e,num:0,agg:null,tuple:this.newtuple(t,this._prev&&this._prev[e]),stamp:this.stamp,store:!1};if(!this._countOnly){var r,i=this._measures,a=i.length;for(n.agg=Array(a),r=0;r<a;++r)n.agg[r]=new i[r](n)}return n.store&&(n.data=new o.default),n},c.newtuple=function(e,t){var n,r,i=this._dnames,o=this._dims,s={};for(n=0,r=o.length;n<r;++n)s[i[n]]=o[n](e);return t?Object(a.replace)(t.tuple,s):Object(a.ingest)(s)},c.add=function(e){var t,n,r,i=this.cellkey(e),o=this.cell(i,e);if(o.num+=1,!this._countOnly)for(o.store&&o.data.add(e),n=0,r=(t=o.agg).length;n<r;++n)t[n].add(t[n].get(e),e)},c.rem=function(e){var t,n,r,i=this.cellkey(e),o=this.cell(i,e);if(o.num-=1,!this._countOnly)for(o.store&&o.data.rem(e),n=0,r=(t=o.agg).length;n<r;++n)t[n].rem(t[n].get(e),e)},c.celltuple=function(e){var t,n,r,i=e.tuple,o=this._counts;for(e.store&&e.data.values(),n=0,r=o.length;n<r;++n)i[o[n]]=e.num;if(!this._countOnly)for(n=0,r=(t=e.agg).length;n<r;++n)t[n].set(i);return i},c.changes=function(e){var t,n,r,i,o=this._adds,a=this._mods,s=this._prev,u=this._drop,c=e.add,l=e.rem,d=e.mod;if(s)for(n in s)t=s[n],u&&!t.num||l.push(t.tuple);for(r=0,i=this._alen;r<i;++r)c.push(this.celltuple(o[r])),o[r]=null;for(r=0,i=this._mlen;r<i;++r)(0===(t=a[r]).num&&u?l:d).push(this.celltuple(t)),a[r]=null;return this._alen=this._mlen=0,this._prev=null,e}},"./node_modules/vega-transforms/src/Bin.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-statistics/index.js"),o=n("./node_modules/vega-util/index.js");function a(e){r.Transform.call(this,null,e)}a.Definition={type:"Bin",metadata:{modifies:!0},params:[{name:"field",type:"field",required:!0},{name:"anchor",type:"number"},{name:"maxbins",type:"number",default:20},{name:"base",type:"number",default:10},{name:"divide",type:"number",array:!0,default:[5,2]},{name:"extent",type:"number",array:!0,length:2,required:!0},{name:"step",type:"number"},{name:"steps",type:"number",array:!0},{name:"minstep",type:"number",default:0},{name:"nice",type:"boolean",default:!0},{name:"name",type:"string"},{name:"as",type:"string",array:!0,length:2,default:["bin0","bin1"]}]};var s=Object(o.inherits)(a,r.Transform);s.transform=function(e,t){var n,r=this._bins(e),i=r.start,a=r.step,s=e.as||["bin0","bin1"],u=s[0],c=s[1];return n=e.modified()?(t=t.reflow(!0)).SOURCE:t.modified(Object(o.accessorFields)(e.field))?t.ADD_MOD:t.ADD,t.visit(n,function(e){var t=r(e);e[u]=t,e[c]=null==t?null:i+a*(1+(t-i)/a)}),t.modifies(s)},s._bins=function(e){if(this.value&&!e.modified())return this.value;var t,n,r=e.field,a=Object(i.bin)(e),s=a.start,u=a.stop,c=a.step;null!=(t=e.anchor)&&(n=t-(s+c*Math.floor((t-s)/c)),s+=n,u+=n);var l=function(e){var t=r(e);return null==t?null:(t=Math.max(s,Math.min(+t,u-c)),s+c*Math.floor((t-s)/c))};return l.start=s,l.stop=u,l.step=c,this.value=Object(o.accessor)(l,Object(o.accessorFields)(r),e.name||"bin_"+Object(o.accessorName)(r))}},"./node_modules/vega-transforms/src/Collect.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-transforms/src/util/SortedList.js"),i=n("./node_modules/vega-dataflow/index.js"),o=n("./node_modules/vega-util/index.js");function a(e){i.Transform.call(this,[],e)}a.Definition={type:"Collect",metadata:{source:!0},params:[{name:"sort",type:"compare"}]},Object(o.inherits)(a,i.Transform).transform=function(e,t){var n=t.fork(t.ALL),o=Object(r.default)(i.tupleid,this.value,n.materialize(n.ADD).add),a=e.sort,s=t.changed()||a&&(e.modified("sort")||t.modified(a.fields));return n.visit(n.REM,o.remove),this.modified(s),this.value=n.source=o.data(a,s),t.source&&t.source.root&&(this.value.root=t.source.root),n}},"./node_modules/vega-transforms/src/Compare.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Operator.call(this,null,a,e)}function a(e){return this.value&&!e.modified()?this.value:Object(i.compare)(e.fields,e.orders)}Object(i.inherits)(o,r.Operator)},"./node_modules/vega-transforms/src/CountPattern.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,null,e)}o.Definition={type:"CountPattern",metadata:{generates:!0,changes:!0},params:[{name:"field",type:"field",required:!0},{name:"case",type:"enum",values:["upper","lower","mixed"],default:"mixed"},{name:"pattern",type:"string",default:'[\\w"]+'},{name:"stopwords",type:"string",default:""},{name:"as",type:"string",array:!0,length:2,default:["text","count"]}]};var a=Object(i.inherits)(o,r.Transform);a.transform=function(e,t){function n(t){return function(n){for(var r,i=function(e,t,n){switch(t){case"upper":e=e.toUpperCase();break;case"lower":e=e.toLowerCase()}return e.match(n)}(s(n),e.case,o)||[],u=0,c=i.length;u<c;++u)a.test(r=i[u])||t(r)}}var r=this._parameterCheck(e,t),i=this._counts,o=this._match,a=this._stop,s=e.field,u=e.as||["text","count"],c=n(function(e){i[e]=1+(i[e]||0)}),l=n(function(e){i[e]-=1});return r?t.visit(t.SOURCE,c):(t.visit(t.ADD,c),t.visit(t.REM,l)),this._finish(t,u)},a._parameterCheck=function(e,t){var n=!1;return!e.modified("stopwords")&&this._stop||(this._stop=new RegExp("^"+(e.stopwords||"")+"$","i"),n=!0),!e.modified("pattern")&&this._match||(this._match=new RegExp(e.pattern||"[\\w']+","g"),n=!0),(e.modified("field")||t.modified(e.field.fields))&&(n=!0),n&&(this._counts={}),n},a._finish=function(e,t){var n,i,o,a=this._counts,s=this._tuples||(this._tuples={}),u=t[0],c=t[1],l=e.fork(e.NO_SOURCE|e.NO_FIELDS);for(n in a)i=s[n],o=a[n]||0,!i&&o?(s[n]=i=Object(r.ingest)({}),i[u]=n,i[c]=o,l.add.push(i)):0===o?(i&&l.rem.push(i),a[n]=null,s[n]=null):i[c]!==o&&(i[c]=o,l.mod.push(i));return l.modifies(t)}},"./node_modules/vega-transforms/src/Cross.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,null,e)}o.Definition={type:"Cross",metadata:{generates:!0},params:[{name:"filter",type:"expr"},{name:"as",type:"string",array:!0,length:2,default:["a","b"]}]},Object(i.inherits)(o,r.Transform).transform=function(e,t){var n=t.fork(t.NO_SOURCE),o=this.value,a=e.as||["a","b"],s=a[0],u=a[1];return!o||t.changed(t.ADD_REM)||e.modified("as")||e.modified("filter")?(o&&(n.rem=o),o=t.materialize(t.SOURCE).source,n.add=this.value=function(e,t,n,i){for(var o,a,s=[],u={},c=e.length,l=0;l<c;++l)for(u[t]=a=e[l],o=0;o<c;++o)u[n]=e[o],i(u)&&(s.push(Object(r.ingest)(u)),(u={})[t]=a);return s}(o,s,u,e.filter||i.truthy)):n.mod=o,n.source=this.value,n.modifies(a)}},"./node_modules/vega-transforms/src/Density.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return s});var r=n("./node_modules/vega-transforms/src/util/Distributions.js"),i=n("./node_modules/vega-dataflow/index.js"),o=n("./node_modules/vega-util/index.js"),a=n("./node_modules/d3-array/index.js");function s(e){i.Transform.call(this,null,e)}var u=[{key:{function:"normal"},params:[{name:"mean",type:"number",default:0},{name:"stdev",type:"number",default:1}]},{key:{function:"uniform"},params:[{name:"min",type:"number",default:0},{name:"max",type:"number",default:1}]},{key:{function:"kde"},params:[{name:"field",type:"field",required:!0},{name:"from",type:"data"},{name:"bandwidth",type:"number",default:0}]}],c={key:{function:"mixture"},params:[{name:"distributions",type:"param",array:!0,params:u},{name:"weights",type:"number",array:!0}]};s.Definition={type:"Density",metadata:{generates:!0},params:[{name:"extent",type:"number",array:!0,length:2},{name:"steps",type:"number",default:100},{name:"method",type:"string",default:"pdf",values:["pdf","cdf"]},{name:"distribution",type:"param",params:u.concat(c)},{name:"as",type:"string",array:!0,default:["value","density"]}]},Object(o.inherits)(s,i.Transform).transform=function(e,t){var n,s=t.fork(t.NO_SOURCE|t.NO_FIELDS);if(!this.value||t.changed()||e.modified()){var u=Object(r.default)(e.distribution,(n=t,function(){return n.materialize(n.SOURCE).source})),c=e.method||"pdf";"pdf"!==c&&"cdf"!==c&&Object(o.error)("Invalid density method: "+c),e.extent||u.data||Object(o.error)("Missing density extent parameter."),c=u[c];var l=e.as||["value","density"],d=e.extent||Object(a.extent)(u.data()),p=(d[1]-d[0])/(e.steps||100),f=Object(a.range)(d[0],d[1]+p/2,p).map(function(e){var t={};return t[l[0]]=e,t[l[1]]=c(e),Object(i.ingest)(t)});this.value&&(s.rem=this.value),this.value=s.add=s.source=f}return s}},"./node_modules/vega-transforms/src/Extent.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,[1/0,-1/0],e)}o.Definition={type:"Extent",metadata:{},params:[{name:"field",type:"field",required:!0}]},Object(i.inherits)(o,r.Transform).transform=function(e,t){var n=this.value,r=e.field,i=n[0],o=n[1],a=t.ADD;(t.changed()||t.modified(r.fields)||e.modified("field"))&&(a=t.SOURCE,i=1/0,o=-1/0),t.visit(a,function(e){var t=r(e);null!=t&&((t=+t)<i&&(i=t),t>o&&(o=t))}),this.value=[i,o]}},"./node_modules/vega-transforms/src/Facet.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-transforms/src/Subflow.js"),i=n("./node_modules/vega-dataflow/index.js"),o=n("./node_modules/vega-util/index.js");function a(e){i.Transform.call(this,{},e),this._keys=Object(o.fastmap)();var t=this._targets=[];t.active=0,t.forEach=function(e){for(var n=0,r=t.active;n<r;++n)e(t[n],n,t)}}var s=Object(o.inherits)(a,i.Transform);s.activate=function(e){this._targets[this._targets.active++]=e},s.subflow=function(e,t,n,i){var o,a,s=this.value,u=s.hasOwnProperty(e)&&s[e];return u?u.value.stamp<n.stamp&&(u.init(n),this.activate(u)):(a=i||(a=this._group[e])&&a.tuple,u=(o=n.dataflow).add(new r.default(n.fork(n.NO_SOURCE),this)).connect(t(o,e,a)),s[e]=u,this.activate(u)),u},s.transform=function(e,t){var n=t.dataflow,r=this,o=e.key,a=e.subflow,s=this._keys,u=e.modified("key");function c(e){return r.subflow(e,a,t)}return this._group=e.group||{},this._targets.active=0,t.visit(t.REM,function(e){var t=Object(i.tupleid)(e),n=s.get(t);void 0!==n&&(s.delete(t),c(n).rem(e))}),t.visit(t.ADD,function(e){var t=o(e);s.set(Object(i.tupleid)(e),t),c(t).add(e)}),u||t.modified(o.fields)?t.visit(t.MOD,function(e){var t=Object(i.tupleid)(e),n=s.get(t),r=o(e);n===r?c(r).mod(e):(s.set(t,r),c(n).rem(e),c(r).add(e))}):t.changed(t.MOD)&&t.visit(t.MOD,function(e){c(s.get(Object(i.tupleid)(e))).mod(e)}),u&&t.visit(t.REFLOW,function(e){var t=Object(i.tupleid)(e),n=s.get(t),r=o(e);n!==r&&(s.set(t,r),c(n).rem(e),c(r).add(e))}),s.empty>n.cleanThreshold&&n.runAfter(s.clean),t}},"./node_modules/vega-transforms/src/Field.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Operator.call(this,null,a,e)}function a(e){return this.value&&!e.modified()?this.value:Object(i.isArray)(e.name)?Object(i.array)(e.name).map(function(e){return Object(i.field)(e)}):Object(i.field)(e.name,e.as)}Object(i.inherits)(o,r.Operator)},"./node_modules/vega-transforms/src/Filter.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,Object(i.fastmap)(),e)}o.Definition={type:"Filter",metadata:{changes:!0},params:[{name:"expr",type:"expr",required:!0}]},Object(i.inherits)(o,r.Transform).transform=function(e,t){var n=t.dataflow,i=this.value,o=t.fork(),a=o.add,s=o.rem,u=o.mod,c=e.expr,l=!0;function d(t){var n=Object(r.tupleid)(t),o=c(t,e),d=i.get(n);o&&d?(i.delete(n),a.push(t)):o||d?l&&o&&!d&&u.push(t):(i.set(n,1),s.push(t))}return t.visit(t.REM,function(e){var t=Object(r.tupleid)(e);i.has(t)?i.delete(t):s.push(e)}),t.visit(t.ADD,function(t){c(t,e)?a.push(t):i.set(Object(r.tupleid)(t),1)}),t.visit(t.MOD,d),e.modified()&&(l=!1,t.visit(t.REFLOW,d)),i.empty>n.cleanThreshold&&n.runAfter(i.clean),o}},"./node_modules/vega-transforms/src/Flatten.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-transforms/src/util/util.js"),i=n("./node_modules/vega-dataflow/index.js"),o=n("./node_modules/vega-util/index.js");function a(e){i.Transform.call(this,[],e)}a.Definition={type:"Flatten",metadata:{generates:!0,source:!0},params:[{name:"fields",type:"field",array:!0,required:!0},{name:"as",type:"string",array:!0}]},Object(o.inherits)(a,i.Transform).transform=function(e,t){var n=t.fork(t.NO_SOURCE),o=e.fields,a=Object(r.fieldNames)(o,e.as||[]),s=a.length;return n.rem=this.value,t.visit(t.SOURCE,function(e){for(var t,r,u,c=o.map(function(t){return t(e)}),l=c.reduce(function(e,t){return Math.max(e,t.length)},0),d=0;d<l;++d){for(r=Object(i.derive)(e),t=0;t<s;++t)r[a[t]]=null==(u=c[t][d])?null:u;n.add.push(r)}}),this.value=n.source=n.add,n.modifies(a)}},"./node_modules/vega-transforms/src/Fold.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,[],e)}o.Definition={type:"Fold",metadata:{generates:!0,source:!0},params:[{name:"fields",type:"field",array:!0,required:!0},{name:"as",type:"string",array:!0,length:2,default:["key","value"]}]},Object(i.inherits)(o,r.Transform).transform=function(e,t){var n=t.fork(t.NO_SOURCE),o=e.fields,a=o.map(i.accessorName),s=e.as||["key","value"],u=s[0],c=s[1],l=o.length;return n.rem=this.value,t.visit(t.SOURCE,function(e){for(var t,i=0;i<l;++i)(t=Object(r.derive)(e))[u]=a[i],t[c]=o[i](e),n.add.push(t)}),this.value=n.source=n.add,n.modifies(s)}},"./node_modules/vega-transforms/src/Formula.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,null,e)}o.Definition={type:"Formula",metadata:{modifies:!0},params:[{name:"expr",type:"expr",required:!0},{name:"as",type:"string",required:!0},{name:"initonly",type:"boolean"}]},Object(i.inherits)(o,r.Transform).transform=function(e,t){var n=e.expr,r=e.as,i=e.modified(),o=e.initonly?t.ADD:i?t.SOURCE:t.modified(n.fields)?t.ADD_MOD:t.ADD;return i&&(t=t.materialize().reflow(!0)),e.initonly||t.modifies(r),t.visit(o,function(t){t[r]=n(t,e)})}},"./node_modules/vega-transforms/src/Generate.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,[],e)}Object(i.inherits)(o,r.Transform).transform=function(e,t){var n,i,o,a=this.value,s=t.fork(t.ALL),u=e.size-a.length,c=e.generator;if(u>0){for(n=[];--u>=0;)n.push(o=Object(r.ingest)(c(e))),a.push(o);s.add=s.add.length?s.materialize(s.ADD).add.concat(n):n}else i=a.slice(0,-u),s.rem=s.rem.length?s.materialize(s.REM).rem.concat(i):i,a=a.slice(-u);return s.source=this.value=a,s}},"./node_modules/vega-transforms/src/Impute.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return u});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js"),o=n("./node_modules/d3-array/index.js"),a={value:"value",median:o.median,mean:o.mean,min:o.min,max:o.max},s=[];function u(e){r.Transform.call(this,[],e)}u.Definition={type:"Impute",metadata:{generates:!0,changes:!0},params:[{name:"field",type:"field",required:!0},{name:"key",type:"field",required:!0},{name:"keyvals",array:!0},{name:"groupby",type:"field",array:!0},{name:"method",type:"enum",default:"value",values:["value","mean","median","max","min"]},{name:"value",default:0}]},Object(i.inherits)(u,r.Transform).transform=function(e,t){var n,o,u,c,l,d,p,f,h,m,g,v=t.fork(t.ALL),_=function(e){var t,n=e.method||a.value;if(null!=a[n])return n===a.value?(t=void 0!==e.value?e.value:0,function(){return t}):a[n];Object(i.error)("Unrecognized imputation method: "+n)}(e),y=(g=e.field,function(e){return e?g(e):NaN}),b=Object(i.accessorName)(e.field),j=Object(i.accessorName)(e.key),x=(e.groupby||[]).map(i.accessorName),O=function(e,t,n,r){var i,o,a,u,c,l,d,p,f=function(e){return e(p)},h=[],m=r?r.slice():[],g={},v={};for(m.forEach(function(e,t){g[e]=t+1}),u=0,d=e.length;u<d;++u)p=e[u],l=n(p),c=g[l]||(g[l]=m.push(l)),o=(i=t?t.map(f):s)+"",(a=v[o])||(a=v[o]=[],h.push(a),a.values=i),a[c-1]=p;return h.domain=m,h}(t.source,e.groupby,e.key,e.keyvals),w=[],C=this.value,E=O.domain.length;for(l=0,f=O.length;l<f;++l)for(u=(n=O[l]).values,o=NaN,p=0;p<E;++p)if(null==n[p]){for(c=O.domain[p],m={_impute:!0},d=0,h=u.length;d<h;++d)m[x[d]]=u[d];m[j]=c,m[b]=isNaN(o)?o=_(n,y):o,w.push(Object(r.ingest)(m))}return w.length&&(v.add=v.materialize(v.ADD).add.concat(w)),C.length&&(v.rem=v.materialize(v.REM).rem.concat(C)),this.value=w,v}},"./node_modules/vega-transforms/src/JoinAggregate.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-transforms/src/Aggregate.js"),i=n("./node_modules/vega-transforms/src/util/AggregateOps.js"),o=n("./node_modules/vega-util/index.js");function a(e){r.default.call(this,e)}a.Definition={type:"JoinAggregate",metadata:{modifies:!0},params:[{name:"groupby",type:"field",array:!0},{name:"fields",type:"field",null:!0,array:!0},{name:"ops",type:"enum",array:!0,values:i.ValidAggregateOps},{name:"as",type:"string",null:!0,array:!0},{name:"key",type:"field"}]};var s=Object(o.inherits)(a,r.default);s.transform=function(e,t){var n,r=this,i=e.modified();return r.value&&(i||t.modified(r._inputs))?(n=r.value=i?r.init(e):{},t.visit(t.SOURCE,function(e){r.add(e)})):(n=r.value=r.value||this.init(e),t.visit(t.REM,function(e){r.rem(e)}),t.visit(t.ADD,function(e){r.add(e)})),r.changes(),t.visit(t.SOURCE,function(e){Object(o.extend)(e,n[r.cellkey(e)].tuple)}),t.reflow(i).modifies(this._outputs)},s.changes=function(){var e,t,n=this._adds,r=this._mods;for(e=0,t=this._alen;e<t;++e)this.celltuple(n[e]),n[e]=null;for(e=0,t=this._mlen;e<t;++e)this.celltuple(r[e]),r[e]=null;this._alen=this._mlen=0}},"./node_modules/vega-transforms/src/Key.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Operator.call(this,null,a,e)}function a(e){return this.value&&!e.modified()?this.value:Object(i.key)(e.fields,e.flat)}Object(i.inherits)(o,r.Operator)},"./node_modules/vega-transforms/src/Lookup.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,{},e)}o.Definition={type:"Lookup",metadata:{modifies:!0},params:[{name:"index",type:"index",params:[{name:"from",type:"data",required:!0},{name:"key",type:"field",required:!0}]},{name:"values",type:"field",array:!0},{name:"fields",type:"field",array:!0,required:!0},{name:"as",type:"string",array:!0},{name:"default",default:null}]},Object(i.inherits)(o,r.Transform).transform=function(e,t){var n,r,o=t,a=e.as,s=e.fields,u=e.index,c=e.values,l=null==e.default?null:e.default,d=e.modified(),p=d?t.SOURCE:t.ADD,f=s.length;return c?(r=c.length,f>1&&!a&&Object(i.error)('Multi-field lookup requires explicit "as" parameter.'),a&&a.length!==f*r&&Object(i.error)('The "as" parameter has too few output field names.'),a=a||c.map(i.accessorName),n=function(e){for(var t,n,i=0,o=0;i<f;++i)if(null==(n=u.get(s[i](e))))for(t=0;t<r;++t,++o)e[a[o]]=l;else for(t=0;t<r;++t,++o)e[a[o]]=c[t](n)}):(a||Object(i.error)("Missing output field names."),n=function(e){for(var t,n=0;n<f;++n)t=u.get(s[n](e)),e[a[n]]=null==t?l:t}),d?o=t.reflow(!0):p|=s.some(function(e){return t.modified(e.fields)})?t.MOD:0,t.visit(p,n),o.modifies(a)}},"./node_modules/vega-transforms/src/MultiExtent.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Operator.call(this,null,a,e)}function a(e){if(this.value&&!e.modified())return this.value;var t,n,r,i=1/0,o=-1/0,a=e.extents;for(t=0,n=a.length;t<n;++t)(r=a[t])[0]<i&&(i=r[0]),r[1]>o&&(o=r[1]);return[i,o]}Object(i.inherits)(o,r.Operator)},"./node_modules/vega-transforms/src/MultiValues.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Operator.call(this,null,a,e)}function a(e){return this.value&&!e.modified()?this.value:e.values.reduce(function(e,t){return e.concat(t)},[])}Object(i.inherits)(o,r.Operator)},"./node_modules/vega-transforms/src/Params.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,null,e)}Object(i.inherits)(o,r.Transform),o.prototype.transform=function(e,t){return this.modified(e.modified()),this.value=e,t.fork(t.NO_SOURCE|t.NO_FIELDS)}},"./node_modules/vega-transforms/src/PreFacet.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-transforms/src/Facet.js"),i=n("./node_modules/vega-dataflow/index.js"),o=n("./node_modules/vega-util/index.js");function a(e){r.default.call(this,e)}Object(o.inherits)(a,r.default).transform=function(e,t){var n=this,r=e.subflow,a=e.field;return(e.modified("field")||a&&t.modified(Object(o.accessorFields)(a)))&&Object(o.error)("PreFacet does not support field modification."),this._targets.active=0,t.visit(t.MOD,function(e){var o=n.subflow(Object(i.tupleid)(e),r,t,e);a?a(e).forEach(function(e){o.mod(e)}):o.mod(e)}),t.visit(t.ADD,function(e){var o=n.subflow(Object(i.tupleid)(e),r,t,e);a?a(e).forEach(function(e){o.add(Object(i.ingest)(e))}):o.add(e)}),t.visit(t.REM,function(e){var o=n.subflow(Object(i.tupleid)(e),r,t,e);a?a(e).forEach(function(e){o.rem(e)}):o.rem(e)}),t}},"./node_modules/vega-transforms/src/Project.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-transforms/src/util/util.js"),i=n("./node_modules/vega-dataflow/index.js"),o=n("./node_modules/vega-util/index.js");function a(e){i.Transform.call(this,null,e)}a.Definition={type:"Project",metadata:{generates:!0,changes:!0,modifies:!0},params:[{name:"fields",type:"field",array:!0},{name:"as",type:"string",null:!0,array:!0}]},Object(o.inherits)(a,i.Transform).transform=function(e,t){var n,o,a=e.fields,s=Object(r.fieldNames)(e.fields,e.as||[]),u=a?function(e,t){return function(e,t,n,r){for(var i=0,o=n.length;i<o;++i)t[r[i]]=n[i](e);return t}(e,t,a,s)}:i.rederive;return this.value?o=this.value:(t=t.addAll(),o=this.value={}),n=t.fork(t.NO_SOURCE),t.visit(t.REM,function(e){var t=Object(i.tupleid)(e);n.rem.push(o[t]),o[t]=null}),t.visit(t.ADD,function(e){var t=u(e,Object(i.ingest)({}));o[Object(i.tupleid)(e)]=t,n.add.push(t)}),t.visit(t.MOD,function(e){n.mod.push(u(e,o[Object(i.tupleid)(e)]))}),n}},"./node_modules/vega-transforms/src/Proxy.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,null,e)}Object(i.inherits)(o,r.Transform).transform=function(e,t){return this.value=e.value,e.modified("value")?t.fork(t.NO_SOURCE|t.NO_FIELDS):t.StopPropagation}},"./node_modules/vega-transforms/src/Relay.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,null,e)}Object(i.inherits)(o,r.Transform).transform=function(e,t){var n,i;return this.value?i=this.value:(n=t=t.addAll(),i=this.value={}),e.derive&&(n=t.fork(t.NO_SOURCE),t.visit(t.REM,function(e){var t=Object(r.tupleid)(e);n.rem.push(i[t]),i[t]=null}),t.visit(t.ADD,function(e){var t=Object(r.derive)(e);i[Object(r.tupleid)(e)]=t,n.add.push(t)}),t.visit(t.MOD,function(e){n.mod.push(Object(r.rederive)(e,i[Object(r.tupleid)(e)]))})),n}},"./node_modules/vega-transforms/src/Sample.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-statistics/index.js"),o=n("./node_modules/vega-util/index.js");function a(e){r.Transform.call(this,[],e),this.count=0}a.Definition={type:"Sample",metadata:{},params:[{name:"size",type:"number",default:1e3}]},Object(o.inherits)(a,r.Transform).transform=function(e,t){var n=t.fork(t.NO_SOURCE),o=e.modified("size"),a=e.size,s=this.value,u=this.count,c=0,l=s.reduce(function(e,t){return e[Object(r.tupleid)(t)]=1,e},{});function d(e){var t,o;s.length<a?s.push(e):(o=~~((u+1)*Object(i.random)()))<s.length&&o>=c&&(t=s[o],l[Object(r.tupleid)(t)]&&n.rem.push(t),s[o]=e),++u}if(t.rem.length&&(t.visit(t.REM,function(e){var t=Object(r.tupleid)(e);l[t]&&(l[t]=-1,n.rem.push(e)),--u}),s=s.filter(function(e){return-1!==l[Object(r.tupleid)(e)]})),(t.rem.length||o)&&s.length<a&&t.source&&(c=u=s.length,t.visit(t.SOURCE,function(e){l[Object(r.tupleid)(e)]||d(e)}),c=-1),o&&s.length>a){for(var p=0,f=s.length-a;p<f;++p)l[Object(r.tupleid)(s[p])]=-1,n.rem.push(s[p]);s=s.slice(f)}return t.mod.length&&t.visit(t.MOD,function(e){l[Object(r.tupleid)(e)]&&n.mod.push(e)}),t.add.length&&t.visit(t.ADD,d),(t.add.length||c<0)&&(n.add=s.filter(function(e){return!l[Object(r.tupleid)(e)]})),this.count=u,this.value=n.source=s,n}},"./node_modules/vega-transforms/src/Sequence.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js"),o=n("./node_modules/d3-array/index.js");function a(e){r.Transform.call(this,null,e)}a.Definition={type:"Sequence",metadata:{generates:!0,changes:!0},params:[{name:"start",type:"number",required:!0},{name:"stop",type:"number",required:!0},{name:"step",type:"number",default:1}],output:["value"]},Object(i.inherits)(a,r.Transform).transform=function(e,t){if(!this.value||e.modified()){var n=t.materialize().fork(t.MOD);return n.rem=this.value?t.rem.concat(this.value):t.rem,this.value=Object(o.range)(e.start,e.stop,e.step||1).map(r.ingest),n.add=t.add.concat(this.value),n}}},"./node_modules/vega-transforms/src/Sieve.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,null,e),this.modified(!0)}Object(i.inherits)(o,r.Transform).transform=function(e,t){return this.value=t.source,t.changed()?t.fork(t.NO_SOURCE|t.NO_FIELDS):t.StopPropagation}},"./node_modules/vega-transforms/src/Subflow.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e,t){r.Operator.call(this,e),this.parent=t}var a=Object(i.inherits)(o,r.Operator);a.connect=function(e){return this.targets().add(e),e.source=this},a.add=function(e){this.value.add.push(e)},a.rem=function(e){this.value.rem.push(e)},a.mod=function(e){this.value.mod.push(e)},a.init=function(e){this.value.init(e,e.NO_SOURCE)},a.evaluate=function(){return this.value}},"./node_modules/vega-transforms/src/TupleIndex.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,Object(i.fastmap)(),e)}Object(i.inherits)(o,r.Transform).transform=function(e,t){var n=t.dataflow,r=e.field,i=this.value,o=!0;function a(e){i.set(r(e),e)}return e.modified("field")||t.modified(r.fields)?(i.clear(),t.visit(t.SOURCE,a)):t.changed()?(t.visit(t.REM,function(e){i.delete(r(e))}),t.visit(t.ADD,a)):o=!1,this.modified(o),i.empty>n.cleanThreshold&&n.runAfter(i.clean),t.fork()}},"./node_modules/vega-transforms/src/Values.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,null,e)}Object(i.inherits)(o,r.Transform).transform=function(e,t){(!this.value||e.modified("field")||e.modified("sort")||t.changed()||e.sort&&t.modified(e.sort.fields))&&(this.value=(e.sort?t.source.slice().sort(e.sort):t.source).map(e.field))}},"./node_modules/vega-transforms/src/Window.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return d});var r=n("./node_modules/vega-transforms/src/util/AggregateKeys.js"),i=n("./node_modules/vega-transforms/src/util/AggregateOps.js"),o=n("./node_modules/vega-transforms/src/util/SortedList.js"),a=n("./node_modules/vega-transforms/src/util/WindowOps.js"),s=n("./node_modules/vega-transforms/src/util/WindowState.js"),u=n("./node_modules/vega-dataflow/index.js"),c=n("./node_modules/vega-util/index.js"),l=n("./node_modules/d3-array/index.js");function d(e){u.Transform.call(this,{},e),this._mlen=0,this._mods=[]}d.Definition={type:"Window",metadata:{modifies:!0},params:[{name:"sort",type:"compare"},{name:"groupby",type:"field",array:!0},{name:"ops",type:"enum",array:!0,values:a.ValidWindowOps.concat(i.ValidAggregateOps)},{name:"params",type:"number",null:!0,array:!0},{name:"fields",type:"field",null:!0,array:!0},{name:"as",type:"string",null:!0,array:!0},{name:"frame",type:"number",null:!0,array:!0,length:2,default:[null,0]},{name:"ignorePeers",type:"boolean",default:!1}]};var p=Object(c.inherits)(d,u.Transform);function f(e,t,n){var r=n.sort,i=r&&!n.ignorePeers,o=n.frame||[null,0],a=e.data(r),s=a.length,u=0,d=i?Object(l.bisector)(r):null,p={i0:0,i1:0,p0:0,p1:0,index:0,data:a,compare:r||Object(c.constant)(-1)};for(t.init();u<s;++u)h(p,o,u,s),i&&m(p,d),t.update(p,a[u])}function h(e,t,n,r){e.p0=e.i0,e.p1=e.i1,e.i0=null==t[0]?0:Math.max(0,n-Math.abs(t[0])),e.i1=null==t[1]?r:Math.min(r,n+Math.abs(t[1])+1),e.index=n}function m(e,t){var n=e.i0,r=e.i1-1,i=e.compare,o=e.data,a=o.length-1;n>0&&!i(o[n],o[n-1])&&(e.i0=t.left(o,o[n])),r<a&&!i(o[r],o[r+1])&&(e.i1=t.right(o,o[r]))}p.transform=function(e,t){var n,i,o=this,a=o.state,u=e.modified();this.stamp=t.stamp,a&&!u||(a=o.state=new s.default(e));var c=Object(r.groupkey)(e.groupby);function l(e){return o.group(c(e))}for(u||t.modified(a.inputs)?(o.value={},t.visit(t.SOURCE,function(e){l(e).add(e)})):(t.visit(t.REM,function(e){l(e).remove(e)}),t.visit(t.ADD,function(e){l(e).add(e)})),n=0,i=o._mlen;n<i;++n)f(o._mods[n],a,e);return o._mlen=0,o._mods=[],t.reflow(u).modifies(a.outputs)},p.group=function(e){var t=this.value[e];return t||((t=this.value[e]=Object(o.default)(u.tupleid)).stamp=-1),t.stamp<this.stamp&&(t.stamp=this.stamp,this._mods[this._mlen++]=t),t}},"./node_modules/vega-transforms/src/util/AggregateKeys.js":function(e,t,n){"use strict";function r(e){return function(t){for(var n=e.length,r=1,i=String(e[0](t));r<n;++r)i+="|"+e[r](t);return i}}function i(e){return e&&e.length?1===e.length?e[0]:r(e):function(){return""}}n.r(t),n.d(t,"multikey",function(){return r}),n.d(t,"groupkey",function(){return i})},"./node_modules/vega-transforms/src/util/AggregateOps.js":function(e,t,n){"use strict";n.r(t),n.d(t,"measureName",function(){return i}),n.d(t,"AggregateOps",function(){return o}),n.d(t,"ValidAggregateOps",function(){return a}),n.d(t,"createMeasure",function(){return s}),n.d(t,"compileMeasures",function(){return l});var r=n("./node_modules/vega-util/index.js");function i(e,t,n){return n||e+(t?"_"+t:"")}var o={values:u({name:"values",init:"cell.store = true;",set:"cell.data.values()",idx:-1}),count:u({name:"count",set:"cell.num"}),missing:u({name:"missing",set:"this.missing"}),valid:u({name:"valid",set:"this.valid"}),sum:u({name:"sum",init:"this.sum = 0;",add:"this.sum += v;",rem:"this.sum -= v;",set:"this.sum"}),mean:u({name:"mean",init:"this.mean = 0;",add:"var d = v - this.mean; this.mean += d / this.valid;",rem:"var d = v - this.mean; this.mean -= this.valid ? d / this.valid : this.mean;",set:"this.mean"}),average:u({name:"average",set:"this.mean",req:["mean"],idx:1}),variance:u({name:"variance",init:"this.dev = 0;",add:"this.dev += d * (v - this.mean);",rem:"this.dev -= d * (v - this.mean);",set:"this.valid > 1 ? this.dev / (this.valid-1) : 0",req:["mean"],idx:1}),variancep:u({name:"variancep",set:"this.valid > 1 ? this.dev / this.valid : 0",req:["variance"],idx:2}),stdev:u({name:"stdev",set:"this.valid > 1 ? Math.sqrt(this.dev / (this.valid-1)) : 0",req:["variance"],idx:2}),stdevp:u({name:"stdevp",set:"this.valid > 1 ? Math.sqrt(this.dev / this.valid) : 0",req:["variance"],idx:2}),stderr:u({name:"stderr",set:"this.valid > 1 ? Math.sqrt(this.dev / (this.valid * (this.valid-1))) : 0",req:["variance"],idx:2}),distinct:u({name:"distinct",set:"cell.data.distinct(this.get)",req:["values"],idx:3}),ci0:u({name:"ci0",set:"cell.data.ci0(this.get)",req:["values"],idx:3}),ci1:u({name:"ci1",set:"cell.data.ci1(this.get)",req:["values"],idx:3}),median:u({name:"median",set:"cell.data.q2(this.get)",req:["values"],idx:3}),q1:u({name:"q1",set:"cell.data.q1(this.get)",req:["values"],idx:3}),q3:u({name:"q3",set:"cell.data.q3(this.get)",req:["values"],idx:3}),argmin:u({name:"argmin",init:"this.argmin = null;",add:"if (v < this.min) this.argmin = t;",rem:"if (v <= this.min) this.argmin = null;",set:"this.argmin || cell.data.argmin(this.get)",req:["min"],str:["values"],idx:3}),argmax:u({name:"argmax",init:"this.argmax = null;",add:"if (v > this.max) this.argmax = t;",rem:"if (v >= this.max) this.argmax = null;",set:"this.argmax || cell.data.argmax(this.get)",req:["max"],str:["values"],idx:3}),min:u({name:"min",init:"this.min = null;",add:"if (v < this.min || this.min === null) this.min = v;",rem:"if (v <= this.min) this.min = NaN;",set:"this.min = (isNaN(this.min) ? cell.data.min(this.get) : this.min)",str:["values"],idx:4}),max:u({name:"max",init:"this.max = null;",add:"if (v > this.max || this.max === null) this.max = v;",rem:"if (v >= this.max) this.max = NaN;",set:"this.max = (isNaN(this.max) ? cell.data.max(this.get) : this.max)",str:["values"],idx:4})},a=Object.keys(o);function s(e,t){return o[e](t)}function u(e){return function(t){var n=Object(r.extend)({init:"",add:"",rem:"",idx:0},e);return n.out=t||e.name,n}}function c(e,t){return e.idx-t.idx}function l(e,t){var n=t||r.identity,i="var cell = this.cell; this.valid = 0; this.missing = 0;",a="this.cell = cell; this.init();",s="if(v==null){++this.missing; return;} if(v!==v) return; ++this.valid;",u="if(v==null){--this.missing; return;} if(v!==v) return; --this.valid;",l="var cell = this.cell;";return function(e,t){var n,r=e.reduce(function e(n,r){function i(t){n[t]||e(n,n[t]=o[t]())}return r.req&&r.req.forEach(i),t&&r.str&&r.str.forEach(i),n},e.reduce(function(e,t){return e[t.name]=t,e},{})),i=[];for(n in r)i.push(r[n]);return i.sort(c)}(e,!0).forEach(function(e){i+=e.init,s+=e.add,u+=e.rem}),e.slice().sort(c).forEach(function(e){l+="t['"+e.out+"']="+e.set+";"}),l+="return t;",(a=Function("cell",a)).prototype.init=Function(i),a.prototype.add=Function("v","t",s),a.prototype.rem=Function("v","t",u),a.prototype.set=Function("t",l),a.prototype.get=n,a.fields=e.map(function(e){return e.out}),a}},"./node_modules/vega-transforms/src/util/Distributions.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return c});var r=n("./node_modules/vega-statistics/index.js"),i=n("./node_modules/vega-util/index.js"),o={kde:r.randomKDE,mixture:r.randomMixture,normal:r.randomNormal,uniform:r.randomUniform},a="distributions",s="function",u="field";function c(e,t){var n=e[s];o.hasOwnProperty(n)||Object(i.error)("Unknown distribution function: "+n);var r=o[n]();for(var l in e)l===u?r.data((e.from||t()).map(e[l])):l===a?r[l](e[l].map(function(e){return c(e,t)})):typeof r[l]===s&&r[l](e[l]);return r}},"./node_modules/vega-transforms/src/util/SortedList.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/index.js");t.default=function(e,t,n){var i=e,o=t||[],a=n||[],s={},u=0;return{add:function(e){a.push(e)},remove:function(e){s[i(e)]=++u},size:function(){return o.length},data:function(e,t){return u&&(o=o.filter(function(e){return!s[i(e)]}),s={},u=0),t&&e&&o.sort(e),a.length&&(o=e?Object(r.merge)(e,o,a.sort(e)):o.concat(a),a=[]),o}}}},"./node_modules/vega-transforms/src/util/TupleStore.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-statistics/index.js"),o=n("./node_modules/vega-util/index.js");function a(e){this._key=e?Object(o.field)(e):r.tupleid,this.reset()}var s=a.prototype;s.reset=function(){this._add=[],this._rem=[],this._ext=null,this._get=null,this._q=null},s.add=function(e){this._add.push(e)},s.rem=function(e){this._rem.push(e)},s.values=function(){if(this._get=null,0===this._rem.length)return this._add;var e,t,n,r=this._add,i=this._rem,o=this._key,a=r.length,s=i.length,u=Array(a-s),c={};for(e=0;e<s;++e)c[o(i[e])]=1;for(e=0,t=0;e<a;++e)c[o(n=r[e])]?c[o(n)]=0:u[t++]=n;return this._rem=[],this._add=u},s.distinct=function(e){for(var t,n=this.values(),r=n.length,i={},o=0;--r>=0;)t=e(n[r])+"",i.hasOwnProperty(t)||(i[t]=1,++o);return o},s.extent=function(e){if(this._get!==e||!this._ext){var t=this.values(),n=Object(o.extentIndex)(t,e);this._ext=[t[n[0]],t[n[1]]],this._get=e}return this._ext},s.argmin=function(e){return this.extent(e)[0]||{}},s.argmax=function(e){return this.extent(e)[1]||{}},s.min=function(e){var t=this.extent(e)[0];return null!=t?e(t):1/0},s.max=function(e){var t=this.extent(e)[1];return null!=t?e(t):-1/0},s.quartile=function(e){return this._get===e&&this._q||(this._q=Object(i.quartiles)(this.values(),e),this._get=e),this._q},s.q1=function(e){return this.quartile(e)[0]},s.q2=function(e){return this.quartile(e)[1]},s.q3=function(e){return this.quartile(e)[2]},s.ci=function(e){return this._get===e&&this._ci||(this._ci=Object(i.bootstrapCI)(this.values(),1e3,.05,e),this._get=e),this._ci},s.ci0=function(e){return this.ci(e)[0]},s.ci1=function(e){return this.ci(e)[1]}},"./node_modules/vega-transforms/src/util/WindowOps.js":function(e,t,n){"use strict";n.r(t),n.d(t,"WindowOp",function(){return i}),n.d(t,"WindowOps",function(){return o}),n.d(t,"ValidWindowOps",function(){return a});var r=n("./node_modules/vega-util/index.js");function i(e,t,n,i){var a=o[e](t,n);return{init:a.init||r.zero,update:function(e,t){t[i]=a.next(e)}}}var o={row_number:function(){return{next:function(e){return e.index+1}}},rank:function(){var e;return{init:function(){e=1},next:function(t){var n=t.index,r=t.data;return n&&t.compare(r[n-1],r[n])?e=n+1:e}}},dense_rank:function(){var e;return{init:function(){e=1},next:function(t){var n=t.index,r=t.data;return n&&t.compare(r[n-1],r[n])?++e:e}}},percent_rank:function(){var e=o.rank(),t=e.next;return{init:e.init,next:function(e){return(t(e)-1)/(e.data.length-1)}}},cume_dist:function(){var e;return{init:function(){e=0},next:function(t){var n=t.index,r=t.data,i=t.compare;if(e<n){for(;n+1<r.length&&!i(r[n],r[n+1]);)++n;e=n}return(1+e)/r.length}}},ntile:function(e,t){(t=+t)>0||Object(r.error)("ntile num must be greater than zero.");var n=o.cume_dist(),i=n.next;return{init:n.init,next:function(e){return Math.ceil(t*i(e))}}},lag:function(e,t){return t=+t||1,{next:function(n){var r=n.index-t;return r>=0?e(n.data[r]):null}}},lead:function(e,t){return t=+t||1,{next:function(n){var r=n.index+t,i=n.data;return r<i.length?e(i[r]):null}}},first_value:function(e){return{next:function(t){return e(t.data[t.i0])}}},last_value:function(e){return{next:function(t){return e(t.data[t.i1-1])}}},nth_value:function(e,t){return(t=+t)>0||Object(r.error)("nth_value nth must be greater than zero."),{next:function(n){var r=n.i0+(t-1);return r<n.i1?e(n.data[r]):null}}}},a=Object.keys(o)},"./node_modules/vega-transforms/src/util/WindowState.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return s});var r=n("./node_modules/vega-transforms/src/util/AggregateOps.js"),i=n("./node_modules/vega-transforms/src/util/TupleStore.js"),o=n("./node_modules/vega-transforms/src/util/WindowOps.js"),a=n("./node_modules/vega-util/index.js");function s(e){var t=Object(a.array)(e.ops),n=Object(a.array)(e.fields),s=Object(a.array)(e.params),u=Object(a.array)(e.as),c=this.outputs=[],l=this.windows=[],d={},p={},f=!0,h=[],m=[];function g(e){Object(a.array)(Object(a.accessorFields)(e)).forEach(function(e){d[e]=1})}g(e.sort),t.forEach(function(e,t){var i=n[t],d=Object(a.accessorName)(i),v=Object(r.measureName)(e,d,u[t]);if(g(i),c.push(v),o.WindowOps.hasOwnProperty(e))l.push(Object(o.WindowOp)(e,n[t],s[t],v));else{if(null==i&&"count"!==e&&Object(a.error)("Null aggregate field specified."),"count"===e)return void h.push(v);f=!1;var _=p[d];_||((_=p[d]=[]).field=i,m.push(_)),_.push(Object(r.createMeasure)(e,v))}}),(h.length||m.length)&&(this.cell=function(e,t,n){e=e.map(function(e){return Object(r.compileMeasures)(e,e.field)});var o={num:0,agg:null,store:!1,count:t};if(!n)for(var a=e.length,s=o.agg=Array(a),u=0;u<a;++u)s[u]=new e[u](o);if(o.store)var c=o.data=new i.default;return o.add=function(e){if(o.num+=1,!n){c&&c.add(e);for(var t=0;t<a;++t)s[t].add(s[t].get(e),e)}},o.rem=function(e){if(o.num-=1,!n){c&&c.rem(e);for(var t=0;t<a;++t)s[t].rem(s[t].get(e),e)}},o.set=function(e){var r,i;for(c&&c.values(),r=0,i=t.length;r<i;++r)e[t[r]]=o.num;if(!n)for(r=0,i=s.length;r<i;++r)s[r].set(e)},o.init=function(){o.num=0,c&&c.reset();for(var e=0;e<a;++e)s[e].init()},o}(m,h,f)),this.inputs=Object.keys(d)}var u=s.prototype;u.init=function(){this.windows.forEach(function(e){e.init()}),this.cell&&this.cell.init()},u.update=function(e,t){var n,r=this.cell,i=this.windows,o=e.data,a=i&&i.length;if(r){for(n=e.p0;n<e.i0;++n)r.rem(o[n]);for(n=e.p1;n<e.i1;++n)r.add(o[n]);r.set(t)}for(n=0;n<a;++n)i[n].update(e,t)}},"./node_modules/vega-transforms/src/util/util.js":function(e,t,n){"use strict";n.r(t),n.d(t,"fieldNames",function(){return i});var r=n("./node_modules/vega-util/index.js");function i(e,t){return e?e.map(function(e,n){return t[n]||Object(r.accessorName)(e)}):null}},"./node_modules/vega-util/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/src/accessor.js");n.d(t,"accessor",function(){return r.default}),n.d(t,"accessorName",function(){return r.accessorName}),n.d(t,"accessorFields",function(){return r.accessorFields});var i=n("./node_modules/vega-util/src/accessors.js");n.d(t,"id",function(){return i.id}),n.d(t,"identity",function(){return i.identity}),n.d(t,"zero",function(){return i.zero}),n.d(t,"one",function(){return i.one}),n.d(t,"truthy",function(){return i.truthy}),n.d(t,"falsy",function(){return i.falsy});var o=n("./node_modules/vega-util/src/logger.js");n.d(t,"logger",function(){return o.default}),n.d(t,"None",function(){return o.None}),n.d(t,"Error",function(){return o.Error}),n.d(t,"Warn",function(){return o.Warn}),n.d(t,"Info",function(){return o.Info}),n.d(t,"Debug",function(){return o.Debug});var a=n("./node_modules/vega-util/src/transform.js");n.d(t,"panLinear",function(){return a.panLinear}),n.d(t,"panLog",function(){return a.panLog}),n.d(t,"panPow",function(){return a.panPow}),n.d(t,"zoomLinear",function(){return a.zoomLinear}),n.d(t,"zoomLog",function(){return a.zoomLog}),n.d(t,"zoomPow",function(){return a.zoomPow});var s=n("./node_modules/vega-util/src/array.js");n.d(t,"array",function(){return s.default});var u=n("./node_modules/vega-util/src/compare.js");n.d(t,"compare",function(){return u.default});var c=n("./node_modules/vega-util/src/constant.js");n.d(t,"constant",function(){return c.default});var l=n("./node_modules/vega-util/src/debounce.js");n.d(t,"debounce",function(){return l.default});var d=n("./node_modules/vega-util/src/error.js");n.d(t,"error",function(){return d.default});var p=n("./node_modules/vega-util/src/extend.js");n.d(t,"extend",function(){return p.default});var f=n("./node_modules/vega-util/src/extentIndex.js");n.d(t,"extentIndex",function(){return f.default});var h=n("./node_modules/vega-util/src/fastmap.js");n.d(t,"fastmap",function(){return h.default});var m=n("./node_modules/vega-util/src/field.js");n.d(t,"field",function(){return m.default});var g=n("./node_modules/vega-util/src/inherits.js");n.d(t,"inherits",function(){return g.default});var v=n("./node_modules/vega-util/src/isArray.js");n.d(t,"isArray",function(){return v.default});var _=n("./node_modules/vega-util/src/isBoolean.js");n.d(t,"isBoolean",function(){return _.default});var y=n("./node_modules/vega-util/src/isDate.js");n.d(t,"isDate",function(){return y.default});var b=n("./node_modules/vega-util/src/isFunction.js");n.d(t,"isFunction",function(){return b.default});var j=n("./node_modules/vega-util/src/isNumber.js");n.d(t,"isNumber",function(){return j.default});var x=n("./node_modules/vega-util/src/isObject.js");n.d(t,"isObject",function(){return x.default});var O=n("./node_modules/vega-util/src/isRegExp.js");n.d(t,"isRegExp",function(){return O.default});var w=n("./node_modules/vega-util/src/isString.js");n.d(t,"isString",function(){return w.default});var C=n("./node_modules/vega-util/src/key.js");n.d(t,"key",function(){return C.default});var E=n("./node_modules/vega-util/src/merge.js");n.d(t,"merge",function(){return E.default});var S=n("./node_modules/vega-util/src/pad.js");n.d(t,"pad",function(){return S.default});var k=n("./node_modules/vega-util/src/peek.js");n.d(t,"peek",function(){return k.default});var M=n("./node_modules/vega-util/src/repeat.js");n.d(t,"repeat",function(){return M.default});var A=n("./node_modules/vega-util/src/splitAccessPath.js");n.d(t,"splitAccessPath",function(){return A.default});var T=n("./node_modules/vega-util/src/stringValue.js");n.d(t,"stringValue",function(){return T.default});var D=n("./node_modules/vega-util/src/toBoolean.js");n.d(t,"toBoolean",function(){return D.default});var I=n("./node_modules/vega-util/src/toDate.js");n.d(t,"toDate",function(){return I.default});var P=n("./node_modules/vega-util/src/toNumber.js");n.d(t,"toNumber",function(){return P.default});var R=n("./node_modules/vega-util/src/toString.js");n.d(t,"toString",function(){return R.default});var N=n("./node_modules/vega-util/src/toSet.js");n.d(t,"toSet",function(){return N.default});var F=n("./node_modules/vega-util/src/truncate.js");n.d(t,"truncate",function(){return F.default});var L=n("./node_modules/vega-util/src/visitArray.js");n.d(t,"visitArray",function(){return L.default})},"./node_modules/vega-util/src/accessor.js":function(e,t,n){"use strict";function r(e){return null==e?null:e.fname}function i(e){return null==e?null:e.fields}n.r(t),n.d(t,"accessorName",function(){return r}),n.d(t,"accessorFields",function(){return i}),t.default=function(e,t,n){return e.fields=t||[],e.fname=n,e}},"./node_modules/vega-util/src/accessors.js":function(e,t,n){"use strict";n.r(t),n.d(t,"id",function(){return a}),n.d(t,"identity",function(){return s}),n.d(t,"zero",function(){return u}),n.d(t,"one",function(){return c}),n.d(t,"truthy",function(){return l}),n.d(t,"falsy",function(){return d});var r=n("./node_modules/vega-util/src/accessor.js"),i=n("./node_modules/vega-util/src/field.js"),o=[],a=Object(i.default)("id"),s=Object(r.default)(function(e){return e},o,"identity"),u=Object(r.default)(function(){return 0},o,"zero"),c=Object(r.default)(function(){return 1},o,"one"),l=Object(r.default)(function(){return!0},o,"true"),d=Object(r.default)(function(){return!1},o,"false")},"./node_modules/vega-util/src/array.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/src/isArray.js");t.default=function(e){return null!=e?Object(r.default)(e)?e:[e]:[]}},"./node_modules/vega-util/src/compare.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/src/accessor.js"),i=n("./node_modules/vega-util/src/array.js"),o=n("./node_modules/vega-util/src/isFunction.js"),a=n("./node_modules/vega-util/src/splitAccessPath.js"),s=n("./node_modules/vega-util/src/stringValue.js");t.default=function(e,t){var n,u,c,l,d,p,f,h,m,g=[],v=(e=Object(i.default)(e)).map(function(e,t){return null==e?null:(g.push(t),Object(o.default)(e)?e:Object(a.default)(e).map(s.default).join("]["))}),_=g.length-1,y=Object(i.default)(t),b="var u,v;return ";if(_<0)return null;for(u=0;u<=_;++u)c=v[n=g[u]],Object(o.default)(c)?(l="(u=this."+(p="f"+n)+"(a))",d="(v=this."+p+"(b))",(f=f||{})[p]=c):(l="(u=a["+c+"])",d="(v=b["+c+"])"),p="((v=v instanceof Date?+v:v),(u=u instanceof Date?+u:u))","descending"!==y[n]?(m=1,h=-1):(m=-1,h=1),b+="("+l+"<"+d+"||u==null)&&v!=null?"+h+":(u>v||v==null)&&u!=null?"+m+":"+p+"!==u&&v===v?"+h+":v!==v&&u===u?"+m+(n<_?":":":0");return c=Function("a","b",b+";"),f&&(c=c.bind(f)),e=e.reduce(function(e,t){return Object(o.default)(t)?(Object(r.accessorFields)(t)||[]).forEach(function(t){e[t]=1}):null!=t&&(e[t+""]=1),e},{}),Object(r.default)(c,Object.keys(e))}},"./node_modules/vega-util/src/constant.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/src/isFunction.js");t.default=function(e){return Object(r.default)(e)?e:function(){return e}}},"./node_modules/vega-util/src/debounce.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){var n,r;function i(){t(r),n=r=null}return function(t){r=t,n&&clearTimeout(n),n=setTimeout(i,e)}}},"./node_modules/vega-util/src/error.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){throw Error(e)}},"./node_modules/vega-util/src/extend.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){for(var t,n,r=1,i=arguments.length;r<i;++r){t=arguments[r];for(n in t)e[n]=t[n]}return e}},"./node_modules/vega-util/src/extentIndex.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){var n,r,i,o,a,s=-1,u=e.length;if(null==t){for(;++s<u;)if(null!=(r=e[s])&&r>=r){n=i=r;break}for(o=a=s;++s<u;)null!=(r=e[s])&&(n>r&&(n=r,o=s),i<r&&(i=r,a=s))}else{for(;++s<u;)if(null!=(r=t(e[s],s,e))&&r>=r){n=i=r;break}for(o=a=s;++s<u;)null!=(r=t(e[s],s,e))&&(n>r&&(n=r,o=s),i<r&&(i=r,a=s))}return[o,a]}},"./node_modules/vega-util/src/fastmap.js":function(e,t,n){"use strict";n.r(t);var r={};t.default=function(e){var t,n,i={};function o(e){return i.hasOwnProperty(e)&&i[e]!==r}return t={size:0,empty:0,object:i,has:o,get:function(e){return o(e)?i[e]:void 0},set:function(e,n){return o(e)||(++t.size,i[e]===r&&--t.empty),i[e]=n,this},delete:function(e){return o(e)&&(--t.size,++t.empty,i[e]=r),this},clear:function(){t.size=t.empty=0,t.object=i={}},test:function(e){return arguments.length?(n=e,t):n},clean:function(){var e,o,a={},s=0;for(e in i)(o=i[e])===r||n&&n(o)||(a[e]=o,++s);t.size=s,t.empty=0,t.object=i=a}},e&&Object.keys(e).forEach(function(n){t.set(n,e[n])}),t}},"./node_modules/vega-util/src/field.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/src/accessor.js"),i=n("./node_modules/vega-util/src/splitAccessPath.js"),o=n("./node_modules/vega-util/src/stringValue.js");t.default=function(e,t){var n=Object(i.default)(e),a="return _["+n.map(o.default).join("][")+"];";return Object(r.default)(Function("_",a),[e=1===n.length?n[0]:e],t||e)}},"./node_modules/vega-util/src/inherits.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){var n=e.prototype=Object.create(t.prototype);return n.constructor=e,n}},"./node_modules/vega-util/src/isArray.js":function(e,t,n){"use strict";n.r(t),t.default=Array.isArray},"./node_modules/vega-util/src/isBoolean.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return"boolean"==typeof e}},"./node_modules/vega-util/src/isDate.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return"[object Date]"===Object.prototype.toString.call(e)}},"./node_modules/vega-util/src/isFunction.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return"function"==typeof e}},"./node_modules/vega-util/src/isNumber.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return"number"==typeof e}},"./node_modules/vega-util/src/isObject.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return e===Object(e)}},"./node_modules/vega-util/src/isRegExp.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return"[object RegExp]"===Object.prototype.toString.call(e)}},"./node_modules/vega-util/src/isString.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return"string"==typeof e}},"./node_modules/vega-util/src/key.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/src/accessor.js"),i=n("./node_modules/vega-util/src/array.js"),o=n("./node_modules/vega-util/src/splitAccessPath.js"),a=n("./node_modules/vega-util/src/stringValue.js");t.default=function(e,t){e&&(e=t?Object(i.default)(e).map(function(e){return e.replace(/\\(.)/g,"$1")}):Object(i.default)(e));var n=e&&e.length?Function("_","return ''+"+e.map(function(e){return"_["+(t?Object(a.default)(e):Object(o.default)(e).map(a.default).join("]["))+"]"}).join("+'|'+")+";"):function(){return""};return Object(r.default)(n,e,"key")}},"./node_modules/vega-util/src/logger.js":function(e,t,n){"use strict";function r(e,t,n){var r=[t].concat([].slice.call(n));console[e].apply(console,r)}n.r(t),n.d(t,"None",function(){return i}),n.d(t,"Error",function(){return o}),n.d(t,"Warn",function(){return a}),n.d(t,"Info",function(){return s}),n.d(t,"Debug",function(){return u});var i=0,o=1,a=2,s=3,u=4;t.default=function(e){var t=e||i;return{level:function(e){return arguments.length?(t=+e,this):t},error:function(){return t>=o&&r("error","ERROR",arguments),this},warn:function(){return t>=a&&r("warn","WARN",arguments),this},info:function(){return t>=s&&r("log","INFO",arguments),this},debug:function(){return t>=u&&r("log","DEBUG",arguments),this}}}},"./node_modules/vega-util/src/merge.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n,r){var i=t.length,o=n.length;if(!o)return t;if(!i)return n;for(var a=r||new t.constructor(i+o),s=0,u=0,c=0;s<i&&u<o;++c)a[c]=e(t[s],n[u])>0?n[u++]:t[s++];for(;s<i;++s,++c)a[c]=t[s];for(;u<o;++u,++c)a[c]=n[u];return a}},"./node_modules/vega-util/src/pad.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/src/repeat.js");t.default=function(e,t,n,i){var o=n||" ",a=e+"",s=t-a.length;return s<=0?a:"left"===i?Object(r.default)(o,s)+a:"center"===i?Object(r.default)(o,~~(s/2))+a+Object(r.default)(o,Math.ceil(s/2)):a+Object(r.default)(o,s)}},"./node_modules/vega-util/src/peek.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return e[e.length-1]}},"./node_modules/vega-util/src/repeat.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t){for(var n="";--t>=0;)n+=e;return n}},"./node_modules/vega-util/src/splitAccessPath.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/src/error.js");t.default=function(e){var t,n,i,o=[],a=null,s=0,u=e.length,c="";function l(){o.push(c+e.substring(t,n)),c="",t=n+1}for(e+="",t=n=0;n<u;++n)if("\\"===(i=e[n]))c+=e.substring(t,n),t=++n;else if(i===a)l(),a=null,s=-1;else{if(a)continue;t===s&&'"'===i?(t=n+1,a=i):t===s&&"'"===i?(t=n+1,a=i):"."!==i||s?"["===i?(n>t&&l(),s=t=n+1):"]"===i&&(s||Object(r.default)("Access path missing open bracket: "+e),s>0&&l(),s=0,t=n+1):n>t?l():t=n+1}return s&&Object(r.default)("Access path missing closing bracket: "+e),a&&Object(r.default)("Access path missing closing quote: "+e),n>t&&(n++,l()),o}},"./node_modules/vega-util/src/stringValue.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-util/src/isArray.js"),i=n("./node_modules/vega-util/src/isObject.js"),o=n("./node_modules/vega-util/src/isString.js");function a(e){return Object(r.default)(e)?"["+e.map(a)+"]":Object(i.default)(e)||Object(o.default)(e)?JSON.stringify(e).replace("\u2028","\\u2028").replace("\u2029","\\u2029"):e}},"./node_modules/vega-util/src/toBoolean.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return null==e||""===e?null:!(!e||"false"===e||"0"===e)&&!!e}},"./node_modules/vega-util/src/toDate.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/src/isDate.js"),i=n("./node_modules/vega-util/src/isNumber.js");function o(e){return Object(i.default)(e)?e:Object(r.default)(e)?e:Date.parse(e)}t.default=function(e,t){return t=t||o,null==e||""===e?null:t(e)}},"./node_modules/vega-util/src/toNumber.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return null==e||""===e?null:+e}},"./node_modules/vega-util/src/toSet.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){for(var t={},n=0,r=e.length;n<r;++n)t[e[n]]=!0;return t}},"./node_modules/vega-util/src/toString.js":function(e,t,n){"use strict";n.r(t),t.default=function(e){return null==e||""===e?null:e+""}},"./node_modules/vega-util/src/transform.js":function(e,t,n){"use strict";n.r(t),n.d(t,"panLinear",function(){return l}),n.d(t,"panLog",function(){return d}),n.d(t,"panPow",function(){return p}),n.d(t,"zoomLinear",function(){return h}),n.d(t,"zoomLog",function(){return m}),n.d(t,"zoomPow",function(){return g});var r=n("./node_modules/vega-util/src/accessors.js"),i=n("./node_modules/vega-util/src/peek.js"),o=n("./node_modules/vega-util/src/toNumber.js");function a(e){return function(t){return e*Math.exp(t)}}function s(e){return function(t){return Math.log(e*t)}}function u(e){return function(t){return t<0?-Math.pow(-t,e):Math.pow(t,e)}}function c(e,t,n,r){var o=n(e[0]),a=n(Object(i.default)(e)),s=(a-o)*t;return[r(o-s),r(a-s)]}function l(e,t){return c(e,t,o.default,r.identity)}function d(e,t){var n=Math.sign(e[0]);return c(e,t,s(n),a(n))}function p(e,t,n){return c(e,t,u(n),u(1/n))}function f(e,t,n,r,o){var a=r(e[0]),s=r(Object(i.default)(e)),u=null!=t?r(t):(a+s)/2;return[o(u+(a-u)*n),o(u+(s-u)*n)]}function h(e,t,n){return f(e,t,n,o.default,r.identity)}function m(e,t,n){var r=Math.sign(e[0]);return f(e,t,n,s(r),a(r))}function g(e,t,n,r){return f(e,t,n,u(r),u(1/r))}},"./node_modules/vega-util/src/truncate.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n,r){var i=null!=r?r:"…",o=e+"",a=o.length,s=Math.max(0,t-i.length);return a<=t?o:"left"===n?i+o.slice(a-s):"center"===n?o.slice(0,Math.ceil(s/2))+i+o.slice(a-~~(s/2)):o.slice(0,s)+i}},"./node_modules/vega-util/src/visitArray.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n){if(e){var r,i=0,o=e.length;if(t)for(;i<o;++i)(r=t(e[i]))&&n(r,i,e);else e.forEach(n)}}},"./node_modules/vega-view-transforms/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-view-transforms/src/Bound.js");n.d(t,"bound",function(){return r.default});var i=n("./node_modules/vega-view-transforms/src/Identifier.js");n.d(t,"identifier",function(){return i.default});var o=n("./node_modules/vega-view-transforms/src/Mark.js");n.d(t,"mark",function(){return o.default});var a=n("./node_modules/vega-view-transforms/src/Overlap.js");n.d(t,"overlap",function(){return a.default});var s=n("./node_modules/vega-view-transforms/src/Render.js");n.d(t,"render",function(){return s.default});var u=n("./node_modules/vega-view-transforms/src/ViewLayout.js");n.d(t,"viewlayout",function(){return u.default})},"./node_modules/vega-view-transforms/src/Bound.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-scenegraph/index.js"),o=n("./node_modules/vega-util/index.js");function a(e){r.Transform.call(this,null,e)}function s(e,t,n){return t(e.bounds.clear(),e,n)}Object(o.inherits)(a,r.Transform).transform=function(e,t){var n,r=t.dataflow,o=e.mark,a=o.marktype,u=i.Marks[a],c=u.bound,l=o.bounds;return u.nested?(o.items.length&&r.dirty(o.items[0]),l=s(o,c),o.items.forEach(function(e){e.bounds.clear().union(l)})):"group"===a||e.modified()?(t.visit(t.MOD,function(e){r.dirty(e)}),l.clear(),o.items.forEach(function(e){l.union(s(e,c))})):(n=t.changed(t.REM),t.visit(t.ADD,function(e){l.union(s(e,c))}),t.visit(t.MOD,function(e){n=n||l.alignsWith(e.bounds),r.dirty(e),l.union(s(e,c))}),n&&(l.clear(),o.items.forEach(function(e){l.union(e.bounds)}))),Object(i.boundClip)(o),t.modifies("bounds")}},"./node_modules/vega-view-transforms/src/Identifier.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js"),o=":vega_identifier:";function a(e){r.Transform.call(this,0,e)}a.Definition={type:"Identifier",metadata:{modifies:!0},params:[{name:"as",type:"string",required:!0}]},Object(i.inherits)(a,r.Transform).transform=function(e,t){var n=function(e){var t=e._signals[o];t||(e._signals[o]=t=e.add(0));return t}(t.dataflow),r=n.value,i=e.as;return t.visit(t.ADD,function(e){e[i]||(e[i]=++r)}),n.set(this.value=r),t}},"./node_modules/vega-view-transforms/src/Mark.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-scenegraph/index.js"),o=n("./node_modules/vega-util/index.js");function a(e){r.Transform.call(this,null,e)}Object(o.inherits)(a,r.Transform).transform=function(e,t){var n,r,o,a=this.value;a||((a=t.dataflow.scenegraph().mark(e.markdef,(r=(n=e).groups,o=n.parent,r&&1===r.size?r.get(Object.keys(r.object)[0]):r&&o?r.lookup(o):null),e.index)).group.context=e.context,e.context.group||(e.context.group=a.group),a.source=this,a.clip=e.clip,a.interactive=e.interactive,this.value=a);var s="group"===a.marktype?i.GroupItem:i.Item;return t.visit(t.ADD,function(e){s.call(e,a)}),(e.modified("clip")||e.modified("interactive"))&&(a.clip=e.clip,a.interactive=!!e.interactive,a.zdirty=!0,t.reflow()),a.items=t.source,t}},"./node_modules/vega-view-transforms/src/Overlap.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return s});var r=n("./node_modules/vega-view-transforms/src/orient.js"),i=n("./node_modules/vega-dataflow/index.js"),o=n("./node_modules/vega-scenegraph/index.js"),a=n("./node_modules/vega-util/index.js");function s(e){i.Transform.call(this,null,e)}var u={parity:function(e){return e.filter(function(e,t){return t%2?e.opacity=0:1})},greedy:function(e){var t;return e.filter(function(e,n){return n&&c(t.bounds,e.bounds)?e.opacity=0:(t=e,1)})}};function c(e,t){return!(e.x2-1<t.x1||e.x1+1>t.x2||e.y2-1<t.y1||e.y1+1>t.y2)}function l(e){for(var t,n=1,r=e.length,i=e[0].bounds;n<r;i=t,++n)if(c(i,t=e[n].bounds))return!0}function d(e){var t=e.bounds;return t.width()>1&&t.height()>1}Object(a.inherits)(s,i.Transform).transform=function(e,t){var n=u[e.method]||u.parity,i=t.materialize(t.SOURCE).source;if(i){e.sort&&(i=i.slice().sort(e.sort)),"greedy"===e.method&&(i=i.filter(d)),i.forEach(function(e){e.opacity=1});var s,c,p,f,h,m=i;if(m.length>=3&&l(m)){t=t.reflow(e.modified()).modifies("opacity");do{m=n(m)}while(m.length>=3&&l(m));m.length<3&&!Object(a.peek)(i).opacity&&(m.length>1&&(Object(a.peek)(m).opacity=0),Object(a.peek)(i).opacity=1)}if(e.boundScale){var g=(s=e.boundScale,c=e.boundOrient,p=e.boundTolerance,f=s.range(),h=new o.Bounds,c===r.Top||c===r.Bottom?h.set(f[0],-1/0,f[1],1/0):h.set(-1/0,f[0],1/0,f[1]),h.expand(p||1),function(e){return h.encloses(e.bounds)});i.forEach(function(e){g(e)||(e.opacity=0)})}return t}}},"./node_modules/vega-view-transforms/src/Render.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return o});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){r.Transform.call(this,null,e)}Object(i.inherits)(o,r.Transform).transform=function(e,t){var n=t.dataflow;if(t.visit(t.ALL,function(e){n.dirty(e)}),t.fields&&t.fields.zindex){var r=t.source&&t.source[0];r&&(r.mark.zdirty=!0)}}},"./node_modules/vega-view-transforms/src/ViewLayout.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return C});var r=n("./node_modules/vega-view-transforms/src/gridLayout.js"),i=n("./node_modules/vega-view-transforms/src/orient.js"),o=n("./node_modules/vega-dataflow/index.js"),a=n("./node_modules/vega-scenegraph/index.js"),s=n("./node_modules/vega-util/index.js"),u="fit",c="fit-x",l="fit-y",d="pad",p="none",f="padding",h="axis",m="title",g="frame",v="legend",_="scope",y="row-header",b="row-footer",j="column-header",x="column-footer",O=.5,w=new a.Bounds;function C(e){o.Transform.call(this,null,e)}function E(e,t,n){return e[t]===n?0:(e[t]=n,1)}function S(e){var t=e.items[0].datum.orient;return t===i.Left||t===i.Right}function k(e,t,n,r){var o,s,u,c,l=t.items[0],d=l.datum,p=d.orient,f=(c=+(u=d).grid,[u.ticks?c++:-1,u.labels?c++:-1,c+ +u.domain]),h=l.range,m=l.offset,g=l.position,v=l.minExtent,_=l.maxExtent,y=d.title&&l.items[f[2]].items[0],b=l.titlePadding,j=l.bounds,x=0,C=0;switch(w.clear().union(j),j.clear(),(o=f[0])>-1&&j.union(l.items[o].bounds),(o=f[1])>-1&&j.union(l.items[o].bounds),p){case i.Top:x=g||0,C=-m,s=Math.max(v,Math.min(_,-j.y1)),y&&(y.auto?(s+=b,y.y=-s,s+=y.bounds.height(),j.add(y.bounds.x1,0).add(y.bounds.x2,0)):j.union(y.bounds)),j.add(0,-s).add(h,0);break;case i.Left:x=-m,C=g||0,s=Math.max(v,Math.min(_,-j.x1)),y&&(y.auto?(s+=b,y.x=-s,s+=y.bounds.width(),j.add(0,y.bounds.y1).add(0,y.bounds.y2)):j.union(y.bounds)),j.add(-s,0).add(0,h);break;case i.Right:x=n+m,C=g||0,s=Math.max(v,Math.min(_,j.x2)),y&&(y.auto?(s+=b,y.x=s,s+=y.bounds.width(),j.add(0,y.bounds.y1).add(0,y.bounds.y2)):j.union(y.bounds)),j.add(0,0).add(s,h);break;case i.Bottom:x=g||0,C=r+m,s=Math.max(v,Math.min(_,j.y2)),y&&(y.auto?(s+=b,y.y=s,s+=y.bounds.height(),j.add(y.bounds.x1,0).add(y.bounds.x2,0)):j.union(y.bounds)),j.add(0,0).add(h,s);break;default:x=l.x,C=l.y}return Object(a.boundStroke)(j.translate(x,C),l),E(l,"x",x+O)|E(l,"y",C+O)&&(l.bounds=w,e.dirty(l),l.bounds=j,e.dirty(l)),l.mark.bounds.clear().union(j)}function M(e,t,n,r,o,s,u){var c,l,d,p=t.items[0],f=p.datum.orient,h=p.offset,m=p.bounds,g=0,v=0;switch(f===i.Top||f===i.Bottom?(d=o,g=n[f]):f!==i.Left&&f!==i.Right||(d=r,v=n[f]),w.clear().union(m),m.clear(),p.items.forEach(function(e){m.union(e.bounds)}),c=Math.round(m.width())+2*p.padding-1,l=Math.round(m.height())+2*p.padding-1,f){case i.Left:g-=c+h-Math.floor(d.x1),n.left+=l+n.margin;break;case i.Right:g+=h+Math.ceil(d.x2),n.right+=l+n.margin;break;case i.Top:v-=l+h-Math.floor(d.y1),n.top+=c+n.margin;break;case i.Bottom:v+=h+Math.ceil(d.y2),n.bottom+=c+n.margin;break;case"top-left":g+=h,v+=h;break;case"top-right":g+=s-c-h,v+=h;break;case"bottom-left":g+=h,v+=u-l-h;break;case"bottom-right":g+=s-c-h,v+=u-l-h;break;default:g=p.x,v=p.y}return Object(a.boundStroke)(m.set(g,v,g+c,v+l),p),E(p,"x",g)|E(p,"width",c)|E(p,"y",v)|E(p,"height",l)&&(p.bounds=w,e.dirty(p),p.bounds=m,e.dirty(p)),p.mark.bounds.clear().union(m)}Object(s.inherits)(C,o.Transform).transform=function(e,t){var n=t.dataflow;return e.mark.items.forEach(function(t){e.layout&&Object(r.gridLayout)(n,t,e.layout),function(e,t,n){var r,o,s,O,C,A,T=t.items,D=Math.max(0,t.width||0),I=Math.max(0,t.height||0),P=(new a.Bounds).set(0,0,D,I),R=P.clone(),N=P.clone(),F=P.clone(),L=[];for(C=0,A=T.length;C<A;++C)switch((o=T[C]).role){case h:R.union(O=k(e,o,D,I)),(S(o)?N:F).union(O);break;case m:r=o;break;case v:L.push(o);break;case g:case _:case y:case b:case j:case x:N.union(o.bounds),F.union(o.bounds);break;default:P.union(o.bounds)}r&&(R.union(O=function(e,t,n){var r=t.items[0],o=r.datum.orient,a=r.offset,s=r.bounds,u=0,c=0;switch(w.clear().union(s),o){case i.Top:u=r.x,c=n.y1-a;break;case i.Left:u=n.x1-a,c=r.y;break;case i.Right:u=n.x2+a,c=r.y;break;case i.Bottom:u=r.x,c=n.y2+a;break;default:u=r.x,c=r.y}s.translate(u-r.x,c-r.y),E(r,"x",u)|E(r,"y",c)&&(r.bounds=w,e.dirty(r),r.bounds=s,e.dirty(r));return t.bounds.clear().union(s)}(e,r,R)),(S(r)?N:F).union(O));if(L.length)for(s={left:0,right:0,top:0,bottom:0,margin:n.legendMargin||8},C=0,A=L.length;C<A;++C)if(O=M(e,L[C],s,N,F,D,I),n.autosize&&n.autosize.type===u){var V=L[C].items[0].datum.orient;V===i.Left||V===i.Right?P.add(O.x1,0).add(O.x2,0):V!==i.Top&&V!==i.Bottom||P.add(0,O.y1).add(0,O.y2)}else P.union(O);P.union(N).union(F).union(R),function(e,t,n,r){var i=r.autosize||{},o=i.type,a=e._width,s=e._height,h=e.padding();if(e._autosize<1||!o)return;var m=Math.max(0,t.width||0),g=Math.max(0,Math.ceil(-n.x1)),v=Math.max(0,Math.ceil(n.x2-m)),_=Math.max(0,t.height||0),y=Math.max(0,Math.ceil(-n.y1)),b=Math.max(0,Math.ceil(n.y2-_));i.contains===f&&(a-=h.left+h.right,s-=h.top+h.bottom);o===p?(g=0,y=0,m=a,_=s):o===u?(m=Math.max(0,a-g-v),_=Math.max(0,s-y-b)):o===c?(m=Math.max(0,a-g-v),s=_+y+b):o===l?(a=m+g+v,_=Math.max(0,s-y-b)):o===d&&(a=m+g+v,s=_+y+b);e._resizeView(a,s,m,_,[g,y],i.resize)}(e,t,P,n)}(n,t,e)}),t}},"./node_modules/vega-view-transforms/src/gridLayout.js":function(e,t,n){"use strict";n.r(t),n.d(t,"gridLayout",function(){return b});var r=n("./node_modules/vega-util/index.js"),i=n("./node_modules/vega-scenegraph/index.js"),o="axis",a="legend",s="row-header",u="row-footer",c="row-title",l="column-header",d="column-footer",p="column-title";function f(e,t){for(var n=0,r=e.length;n<r;++n)t.push(e[n])}function h(e){return{x1:0,y1:0,x2:e.width||0,y2:e.height||0}}function m(e){var t=e.bounds.clone();return t.empty()?t.set(0,0,0,0):t.translate(-(e.x||0),-(e.y||0))}function g(e,t){return"x1"===t?e.x||0:"y1"===t?e.y||0:"x2"===t?(e.x||0)+(e.width||0):"y2"===t?(e.y||0)+(e.height||0):void 0}function v(e,t){return e.bounds[t]}function _(e,t,n){var i=Object(r.isObject)(e)?e[t]:e;return null!=i?i:void 0!==n?n:0}function y(e){return e<0?Math.ceil(-e):0}function b(e,t,n){var r,b,O,w,C,E,S,k,M,A,T,D,I,P=function(e){for(var t,n,r=e.items,i=r.length,h=0,m={marks:[],rowheaders:[],rowfooters:[],colheaders:[],colfooters:[],rowtitle:null,coltitle:null};h<i;++h)if(n=(t=r[h]).items,"group"===t.marktype)switch(t.role){case o:case a:break;case s:f(n,m.rowheaders);break;case u:f(n,m.rowfooters);break;case l:f(n,m.colheaders);break;case d:f(n,m.colfooters);break;case c:m.rowtitle=n[0];break;case p:m.coltitle=n[0];break;default:f(n,m.marks)}return m}(t),R=P.marks,N="flush"===n.bounds,F=N?h:m,L=new i.Bounds(0,0,0,0),V=_(n.align,"column"),B=_(n.align,"row"),z=_(n.padding,"column"),U=_(n.padding,"row"),H=n.offset,q=t.columns||n.columns||R.length,G=q<0?1:Math.ceil(R.length/q),W=G*q,Y=[],K=[],Q=[],X=[],Z=R.length;for(b=0;b<q;++b)K[b]=0;for(b=0;b<G;++b)X[b]=0;for(b=0;b<Z;++b)C=F(R[b]),w=~~(b/q),S=(O=b%q)?Math.ceil(F(R[b-1]).x2):0,k=w?Math.ceil(F(R[b-q]).y2):0,K[O]=Math.max(K[O],S),X[w]=Math.max(X[w],k),Y.push(z+y(C.x1)),Q.push(U+y(C.y1)),e.dirty(R[b]);for(b=0;b<Z;++b)b%q==0&&(Y[b]=0),b<q&&(Q[b]=0);if("each"===V)for(O=1;O<q;++O){for(I=0,b=O;b<Z;b+=q)I<Y[b]&&(I=Y[b]);for(b=O;b<Z;b+=q)Y[b]=I+K[O]}else if("all"===V){for(D=0,O=1;O<q;++O)D<K[O]&&(D=K[O]);for(I=0,b=0;b<Z;++b)b%q&&I<Y[b]&&(I=Y[b]);for(b=0;b<Z;++b)b%q&&(Y[b]=I+D)}else for(O=1;O<q;++O)for(b=O;b<Z;b+=q)Y[b]+=K[O];if("each"===B)for(w=1;w<G;++w){for(I=0,r=(b=w*q)+q;b<r;++b)I<Q[b]&&(I=Q[b]);for(b=w*q;b<r;++b)Q[b]=I+X[w]}else if("all"===B){for(D=0,w=1;w<G;++w)D<X[w]&&(D=X[w]);for(I=0,b=q;b<Z;++b)I<Q[b]&&(I=Q[b]);for(b=q;b<Z;++b)Q[b]=I+D}else for(w=1;w<G;++w)for(r=(b=w*q)+q;b<r;++b)Q[b]+=X[w];for(M=0,b=0;b<Z;++b)S=(E=R[b]).x||0,E.x=M=Y[b]+(b%q?M:0),E.bounds.translate(M-S,0);for(O=0;O<q;++O)for(A=0,b=O;b<Z;b+=q)k=(E=R[b]).y||0,E.y=A+=Q[b],E.bounds.translate(0,A-k);for(b=0;b<Z;++b)R[b].mark.bounds.clear();for(b=0;b<Z;++b)E=R[b],e.dirty(E),L.union(E.mark.bounds.union(E.bounds));function $(e,t){return Math.floor(Math.min(e,t))}function J(e,t){return Math.ceil(Math.max(e,t))}F=N?g:v,T=_(n.headerBand,"row",null),M=j(e,P.rowheaders,R,q,G,-_(H,"rowHeader"),$,0,F,"x1",0,q,1,T),T=_(n.headerBand,"column",null),A=j(e,P.colheaders,R,q,q,-_(H,"columnHeader"),$,1,F,"y1",0,1,q,T),T=_(n.footerBand,"row",null),j(e,P.rowfooters,R,q,G,_(H,"rowFooter"),J,0,F,"x2",q-1,q,1,T),T=_(n.footerBand,"column",null),j(e,P.colfooters,R,q,q,_(H,"columnFooter"),J,1,F,"y2",W-q,1,q,T),P.rowtitle&&(I=M-_(H,"rowTitle"),T=_(n.titleBand,"row",.5),x(e,P.rowtitle,I,0,L,T)),P.coltitle&&(I=A-_(H,"columnTitle"),T=_(n.titleBand,"column",.5),x(e,P.coltitle,I,1,L,T))}function j(e,t,n,r,i,o,a,s,u,c,l,d,p,f){var h,m,g,v,_,y,b,j,x,O=n.length,w=0,C=0;if(!O)return w;for(h=l;h<O;h+=d)n[h]&&(w=a(w,u(n[h],c)));if(!t.length)return w;for(t.length>i&&(e.warn("Grid headers exceed limit: "+i),t=t.slice(0,i)),w+=o,m=0,v=t.length;m<v;++m)e.dirty(t[m]),t[m].mark.bounds.clear();for(h=l,m=0,v=t.length;m<v;++m,h+=d){for(_=(y=t[m]).mark.bounds,g=h;g>=0&&null==(b=n[g]);g-=p);s?(j=null==f?b.x:Math.round(b.bounds.x1+f*b.bounds.width()),x=w):(j=w,x=null==f?b.y:Math.round(b.bounds.y1+f*b.bounds.height())),_.union(y.bounds.translate(j-(y.x||0),x-(y.y||0))),y.x=j,y.y=x,e.dirty(y),C=a(C,_[c])}return C}function x(e,t,n,r,i,o){if(t){e.dirty(t);var a=n,s=n;r?a=Math.round(i.x1+o*i.width()):s=Math.round(i.y1+o*i.height()),t.bounds.translate(a-(t.x||0),s-(t.y||0)),t.mark.bounds.clear().union(t.bounds),t.x=a,t.y=s,e.dirty(t)}}},"./node_modules/vega-view-transforms/src/orient.js":function(e,t,n){"use strict";n.r(t),n.d(t,"Top",function(){return r}),n.d(t,"Left",function(){return i}),n.d(t,"Right",function(){return o}),n.d(t,"Bottom",function(){return a});var r="top",i="left",o="right",a="bottom"},"./node_modules/vega-view/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-view/src/View.js");n.d(t,"View",function(){return r.default})},"./node_modules/vega-view/src/View.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return y});var r=n("./node_modules/vega-view/src/cursor.js"),i=n("./node_modules/vega-view/src/data.js"),o=n("./node_modules/vega-view/src/events.js"),a=n("./node_modules/vega-view/src/hover.js"),s=n("./node_modules/vega-view/src/finalize.js"),u=n("./node_modules/vega-view/src/initialize.js"),c=n("./node_modules/vega-view/src/render-to-image-url.js"),l=n("./node_modules/vega-view/src/render-to-canvas.js"),d=n("./node_modules/vega-view/src/render-to-svg.js"),p=n("./node_modules/vega-view/src/render-size.js"),f=n("./node_modules/vega-view/src/runtime.js"),h=n("./node_modules/vega-view/src/size.js"),m=n("./node_modules/vega-view/src/state.js"),g=n("./node_modules/vega-dataflow/index.js"),v=n("./node_modules/vega-util/index.js"),_=n("./node_modules/vega-scenegraph/index.js");function y(e,t){t=t||{},g.Dataflow.call(this),this.loader(t.loader||this._loader),this.logLevel(t.logLevel||0),this._el=null,this._renderType=t.renderer||_.RenderType.Canvas,this._scenegraph=new _.Scenegraph;var n=this._scenegraph.root;this._renderer=null,this._redraw=!0,this._handler=(new _.CanvasHandler).scene(n),this._preventDefault=!1,this._eventListeners=[],this._resizeListeners=[];var i=Object(f.default)(this,e,t.functions);this._runtime=i,this._signals=i.signals,this._bind=(e.bindings||[]).map(function(e){return{state:null,param:Object(v.extend)({},e)}}),i.root&&i.root.set(n),n.source=i.data.root.input,this.pulse(i.data.root.input,this.changeset().insert(n.items)),this._background=i.background||null,this._eventConfig=Object(o.initializeEventConfig)(i.eventConfig),this._width=this.width(),this._height=this.height(),this._viewWidth=Object(h.viewWidth)(this,this._width),this._viewHeight=Object(h.viewHeight)(this,this._height),this._origin=[0,0],this._resize=0,this._autosize=1,Object(h.initializeResize)(this),Object(r.default)(this)}var b=Object(v.inherits)(y,g.Dataflow);function j(e,t){return e._signals.hasOwnProperty(t)?e._signals[t]:Object(v.error)("Unrecognized signal name: "+Object(v.stringValue)(t))}function x(e,t){var n=(e._targets||[]).filter(function(e){var n=e._update;return n&&n.handler===t});return n.length?n[0]:null}b.run=function(e){if(g.Dataflow.prototype.run.call(this,e),this._redraw||this._resize)try{this.render()}catch(e){this.error(e)}return this},b.render=function(){return this._renderer&&(this._resize&&(this._resize=0,Object(p.resizeRenderer)(this)),this._renderer.render(this._scenegraph.root)),this._redraw=!1,this},b.dirty=function(e){this._redraw=!0,this._renderer&&this._renderer.dirty(e)},b.container=function(){return this._el},b.scenegraph=function(){return this._scenegraph},b.signal=function(e,t,n){var r=j(this,e);return 1===arguments.length?r.value:this.update(r,t,n)},b.background=function(e){return arguments.length?(this._background=e,this._resize=1,this):this._background},b.width=function(e){return arguments.length?this.signal("width",e):this.signal("width")},b.height=function(e){return arguments.length?this.signal("height",e):this.signal("height")},b.padding=function(e){return arguments.length?this.signal("padding",e):this.signal("padding")},b.autosize=function(e){return arguments.length?this.signal("autosize",e):this.signal("autosize")},b.renderer=function(e){return arguments.length?(Object(_.renderModule)(e)||Object(v.error)("Unrecognized renderer type: "+e),e!==this._renderType&&(this._renderType=e,this._renderer&&(this._renderer=null,this.initialize(this._el))),this):this._renderType},b.loader=function(e){return arguments.length?(e!==this._loader&&(g.Dataflow.prototype.loader.call(this,e),this._renderer&&(this._renderer=null,this.initialize(this._el))),this):this._loader},b.resize=function(){return this._autosize=1,this},b._resizeView=h.resizeView,b.addEventListener=function(e,t){return this._handler.on(e,t),this},b.removeEventListener=function(e,t){return this._handler.off(e,t),this},b.addResizeListener=function(e){var t=this._resizeListeners;return t.indexOf(e)<0&&t.push(e),this},b.removeResizeListener=function(e){var t=this._resizeListeners,n=t.indexOf(e);return n>=0&&t.splice(n,1),this},b.addSignalListener=function(e,t){var n=j(this,e),r=x(n,t);return r||((r=function(){t(e,n.value)}).handler=t,this.on(n,null,r)),this},b.removeSignalListener=function(e,t){var n=j(this,e),r=x(n,t);return r&&n._targets.remove(r),this},b.preventDefault=function(e){return arguments.length?(this._preventDefault=e,this):this._preventDefault},b.tooltipHandler=function(e){var t=this._handler;return arguments.length?(t.handleTooltip=e||_.Handler.prototype.handleTooltip,this):t.handleTooltip},b.events=o.events,b.finalize=s.default,b.hover=a.default,b.data=i.data,b.change=i.change,b.insert=i.insert,b.remove=i.remove,b.initialize=u.default,b.toImageURL=c.default,b.toCanvas=l.default,b.toSVG=d.default,b.getState=m.getState,b.setState=m.setState},"./node_modules/vega-view/src/bind.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-view/src/element.js"),i=n("./node_modules/vega-util/index.js"),o=n("./node_modules/d3-array/index.js"),a="vega-bind",s="vega-bind-name",u="vega-bind-radio",c="vega-option-";function l(e,t,n,i){var o=Object(r.default)("input");for(var a in n)"signal"!==a&&"element"!==a&&o.setAttribute("input"===a?"type":a,n[a]);o.setAttribute("name",n.signal),o.value=i,t.appendChild(o),o.addEventListener("input",function(){e.update(o.value)}),e.elements=[o],e.set=function(e){o.value=e}}function d(e,t,n,i){var o={type:"checkbox",name:n.signal};i&&(o.checked=!0);var a=Object(r.default)("input",o);t.appendChild(a),a.addEventListener("change",function(){e.update(a.checked)}),e.elements=[a],e.set=function(e){a.checked=!!e||null}}function p(e,t,n,i){var o=Object(r.default)("select",{name:n.signal});n.options.forEach(function(e){var t={value:e};m(e,i)&&(t.selected=!0),o.appendChild(Object(r.default)("option",t,e+""))}),t.appendChild(o),o.addEventListener("change",function(){e.update(n.options[o.selectedIndex])}),e.elements=[o],e.set=function(e){for(var t=0,r=n.options.length;t<r;++t)if(m(n.options[t],e))return void(o.selectedIndex=t)}}function f(e,t,n,i){var o=Object(r.default)("span",{class:u});t.appendChild(o),e.elements=n.options.map(function(t){var a=c+n.signal+"-"+t,s={id:a,type:"radio",name:n.signal,value:t};m(t,i)&&(s.checked=!0);var u=Object(r.default)("input",s);return u.addEventListener("change",function(){e.update(t)}),o.appendChild(u),o.appendChild(Object(r.default)("label",{for:a},t+"")),u}),e.set=function(t){for(var n=e.elements,r=0,i=n.length;r<i;++r)m(n[r].value,t)&&(n[r].checked=!0)}}function h(e,t,n,i){i=void 0!==i?i:(+n.max+ +n.min)/2;var a=n.min||Math.min(0,+i)||0,s=n.max||Math.max(100,+i)||100,u=n.step||Object(o.tickStep)(a,s,100),c=Object(r.default)("input",{type:"range",name:n.signal,min:a,max:s,step:u});c.value=i;var l=Object(r.default)("label",{},+i);function d(){l.textContent=c.value,e.update(+c.value)}t.appendChild(c),t.appendChild(l),c.addEventListener("input",d),c.addEventListener("change",d),e.elements=[c],e.set=function(e){c.value=e,l.textContent=e}}function m(e,t){return e===t||e+""==t+""}t.default=function(e,t,n){if(t){var o=n.param,u=n.state;return u||(u=n.state={elements:null,active:!1,set:null,update:function(t){u.source=!0,e.signal(o.signal,t).run()}},o.debounce&&(u.update=Object(i.debounce)(o.debounce,u.update))),function(e,t,n,i){var o=Object(r.default)("div",{class:a});o.appendChild(Object(r.default)("span",{class:s},n.name||n.signal)),t.appendChild(o);var u=l;switch(n.input){case"checkbox":u=d;break;case"select":u=p;break;case"radio":u=f;break;case"range":u=h}u(e,o,n,i)}(u,t,o,e.signal(o.signal)),u.active||(e.on(e._signals[o.signal],null,function(){u.source?u.source=!1:u.set(e.signal(o.signal))}),u.active=!0),u}}},"./node_modules/vega-view/src/cursor.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-util/index.js");t.default=function(e){var t=e._signals.cursor;t||(e._signals.cursor=t=e.add({user:"default",item:null})),e.on(e.events("view","mousemove"),t,function(e,n){var i=t.value,o=i?Object(r.isString)(i)?i:i.user:"default",a=n.item&&n.item.cursor||null;return i&&o===i.user&&a==i.item?i:{user:o,item:a}}),e.add(null,function(e){var t,n=e.cursor,i=this.value;return Object(r.isString)(n)||(i=n.item,n=n.user),t=n&&"default"!==n?n:i||n,"undefined"!=typeof document&&document.body&&(document.body.style.cursor=t),i},{cursor:t})}},"./node_modules/vega-view/src/data.js":function(e,t,n){"use strict";n.r(t),n.d(t,"data",function(){return a}),n.d(t,"change",function(){return s}),n.d(t,"insert",function(){return u}),n.d(t,"remove",function(){return c});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e,t){var n=e._runtime.data;return n.hasOwnProperty(t)||Object(i.error)("Unrecognized data set: "+t),n[t]}function a(e){return o(this,e).values.value}function s(e,t){Object(r.isChangeSet)(t)||Object(i.error)("Second argument to changes must be a changeset.");var n=o(this,e);return n.modified=!0,this.pulse(n.input,t)}function u(e,t){return s.call(this,e,Object(r.changeset)().insert(t))}function c(e,t){return s.call(this,e,Object(r.changeset)().remove(t))}},"./node_modules/vega-view/src/element.js":function(e,t,n){"use strict";n.r(t),t.default=function(e,t,n){var r=document.createElement(e);for(var i in t)r.setAttribute(i,t[i]);return null!=n&&(r.textContent=n),r}},"./node_modules/vega-view/src/events-extend.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-view/src/render-size.js"),i=n("./node_modules/vega-util/index.js"),o=n("./node_modules/vega-scenegraph/index.js");t.default=function(e,t,n){var a,s,u,c=e._renderer.scene();return c&&(u=Object(r.offset)(e),s=t.changedTouches?t.changedTouches[0]:t,(a=Object(o.point)(s,c))[0]-=u[0],a[1]-=u[1]),t.dataflow=e,t.vega=function(e,t,n){var r=t?"group"===t.mark.marktype?t:t.mark.group:null;function o(e){var n,i=r;if(e)for(n=t;n;n=n.mark.group)if(n.mark.name===e){i=n;break}return i&&i.mark&&i.mark.interactive?i:{}}function a(e){if(!e)return n;Object(i.isString)(e)&&(e=o(e));for(var t=n.slice();e;)t[0]-=e.x||0,t[1]-=e.y||0,e=e.mark&&e.mark.group;return t}return{view:Object(i.constant)(e),item:Object(i.constant)(t||{}),group:o,xy:a,x:function(e){return a(e)[0]},y:function(e){return a(e)[1]}}}(e,n,a),t.item=n,t}},"./node_modules/vega-view/src/events.js":function(e,t,n){"use strict";n.r(t),n.d(t,"initializeEventConfig",function(){return u}),n.d(t,"events",function(){return c});var r=n("./node_modules/vega-view/src/events-extend.js"),i=n("./node_modules/vega-dataflow/index.js"),o=n("./node_modules/vega-util/index.js"),a="view",s="window";function u(e){var t=(e=Object(o.extend)({},e)).defaults;return t&&(Object(o.isArray)(t.prevent)&&(t.prevent=Object(o.toSet)(t.prevent)),Object(o.isArray)(t.allow)&&(t.allow=Object(o.toSet)(t.allow))),e}function c(e,t,n){var o,u=this,c=new i.EventStream(n),l=function(n,i){var o,s,l,d,p;e!==a||(s=t,l=(o=u)._eventConfig.defaults,d=l&&l.prevent,p=l&&l.allow,!1===d||!0===p||!0!==d&&!1!==p&&(d?!d[s]:p?p[s]:!o.preventDefault()))||n.preventDefault();try{c.receive(Object(r.default)(u,n,i))}catch(e){u.error(e)}finally{u.run()}};if(e===a)return u.addEventListener(t,l),c;if(e===s?"undefined"!=typeof window&&(o=[window]):"undefined"!=typeof document&&(o=document.querySelectorAll(e)),!o)return u.warn("Can not resolve event source: "+e),c;for(var d=0,p=o.length;d<p;++d)o[d].addEventListener(t,l);return u._eventListeners.push({type:t,sources:o,handler:l}),c}},"./node_modules/vega-view/src/finalize.js":function(e,t,n){"use strict";n.r(t),t.default=function(){for(var e,t,n=this._eventListeners,r=n.length;--r>=0;)for(e=(t=n[r]).sources.length;--e>=0;)t.sources[e].removeEventListener(t.type,t.handler)}},"./node_modules/vega-view/src/hover.js":function(e,t,n){"use strict";function r(e){return e.item}function i(e){var t=e.item.mark.source;return t.source||t}function o(e){return function(t,n){return n.vega.view().changeset().encode(n.item,e)}}n.r(t),t.default=function(e,t){return e=[e||"hover"],t=[t||"update",e],this.on(this.events("view","mouseover",r),i,o(e)),this.on(this.events("view","mouseout",r),i,o(t)),this}},"./node_modules/vega-view/src/initialize-handler.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-view/src/render-size.js");t.default=function(e,t,n,i){var o=(new i).scene(e.scenegraph().root).initialize(n,Object(r.offset)(e),e);return t&&(o.handleTooltip=t.handleTooltip,t.handlers().forEach(function(e){o.on(e.type,e.handler)})),o}},"./node_modules/vega-view/src/initialize-renderer.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-view/src/render-size.js");t.default=function(e,t,n,i,o){return(t=t||new i(e.loader())).initialize(n,Object(r.width)(e),Object(r.height)(e),Object(r.offset)(e),o).background(e._background)}},"./node_modules/vega-view/src/initialize.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-view/src/bind.js"),i=n("./node_modules/vega-view/src/element.js"),o=n("./node_modules/vega-view/src/initialize-renderer.js"),a=n("./node_modules/vega-view/src/initialize-handler.js"),s=n("./node_modules/vega-scenegraph/index.js");function u(e,t){if("string"==typeof t){if("undefined"==typeof document)return e.error("DOM document instance not found."),null;if(!(t=document.querySelector(t)))return e.error("Signal bind element not found: "+t),null}if(t)try{t.innerHTML=""}catch(n){t=null,e.error(n)}return t}t.default=function(e,t){var n,c,l=this,d=l._renderType,p=Object(s.renderModule)(d);return e=l._el=e?u(l,e):null,p||l.error("Unrecognized renderer type: "+d),n=p.handler||s.CanvasHandler,c=e?p.renderer:p.headless,l._renderer=c?Object(o.default)(l,l._renderer,e,c):null,l._handler=Object(a.default)(l,l._handler,e,n),l._redraw=!0,e&&(t=t?u(l,t):e.appendChild(Object(i.default)("div",{class:"vega-bindings"})),l._bind.forEach(function(e){e.param.element&&(e.element=u(l,e.param.element))}),l._bind.forEach(function(e){Object(r.default)(l,e.element||t,e)})),l}},"./node_modules/vega-view/src/render-headless.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-view/src/initialize-renderer.js"),i=n("./node_modules/vega-scenegraph/index.js");t.default=function(e,t,n){var o=Object(i.renderModule)(t),a=o&&o.headless;return a?e.runAsync().then(function(){return Object(r.default)(e,null,null,a,n).renderAsync(e._scenegraph.root)}):Promise.reject("Unrecognized renderer type: "+t)}},"./node_modules/vega-view/src/render-size.js":function(e,t,n){"use strict";function r(e){var t=e.padding();return Math.max(0,e._viewWidth+t.left+t.right)}function i(e){var t=e.padding();return Math.max(0,e._viewHeight+t.top+t.bottom)}function o(e){var t=e.padding(),n=e._origin;return[t.left+n[0],t.top+n[1]]}function a(e){var t=o(e),n=r(e),a=i(e);e._renderer.background(e._background),e._renderer.resize(n,a,t),e._handler.origin(t),e._resizeListeners.forEach(function(e){e(n,a)})}n.r(t),n.d(t,"width",function(){return r}),n.d(t,"height",function(){return i}),n.d(t,"offset",function(){return o}),n.d(t,"resizeRenderer",function(){return a})},"./node_modules/vega-view/src/render-to-canvas.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-view/src/render-headless.js"),i=n("./node_modules/vega-scenegraph/index.js");t.default=function(e){return Object(r.default)(this,i.RenderType.Canvas,e).then(function(e){return e.canvas()})}},"./node_modules/vega-view/src/render-to-image-url.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-view/src/render-headless.js"),i=n("./node_modules/vega-scenegraph/index.js");t.default=function(e,t){return e!==i.RenderType.Canvas&&e!==i.RenderType.SVG&&e!==i.RenderType.PNG?Promise.reject("Unrecognized image type: "+e):Object(r.default)(this,e,t).then(function(t){return e===i.RenderType.SVG?(n=t.svg(),r=new Blob([n],{type:"image/svg+xml"}),window.URL.createObjectURL(r)):t.canvas().toDataURL("image/png");var n,r})}},"./node_modules/vega-view/src/render-to-svg.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-view/src/render-headless.js"),i=n("./node_modules/vega-scenegraph/index.js");t.default=function(e){return Object(r.default)(this,i.RenderType.SVG,e).then(function(e){return e.svg()})}},"./node_modules/vega-view/src/runtime.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-parser/index.js"),o=n("./node_modules/vega-runtime/index.js");t.default=function(e,t,n){var a=n||i.functionContext;return Object(o.parse)(t,Object(o.context)(e,r.transforms,a))}},"./node_modules/vega-view/src/size.js":function(e,t,n){"use strict";n.r(t),n.d(t,"viewWidth",function(){return i}),n.d(t,"viewHeight",function(){return o}),n.d(t,"initializeResize",function(){return a}),n.d(t,"resizeView",function(){return s});var r="padding";function i(e,t){var n=e.autosize(),i=e.padding();return t-(n&&n.contains===r?i.left+i.right:0)}function o(e,t){var n=e.autosize(),i=e.padding();return t-(n&&n.contains===r?i.top+i.bottom:0)}function a(e){var t=e._signals,n=t.width,r=t.height,a=t.padding;function s(){e._autosize=e._resize=1}e._resizeWidth=e.add(null,function(t){e._width=t.size,e._viewWidth=i(e,t.size),s()},{size:n}),e._resizeHeight=e.add(null,function(t){e._height=t.size,e._viewHeight=o(e,t.size),s()},{size:r});var u=e.add(null,s,{pad:a});e._resizeWidth.rank=n.rank+1,e._resizeHeight.rank=r.rank+1,u.rank=a.rank+1}function s(e,t,n,r,i,o){this.runAfter(function(a){var s=0;a._autosize=0,a.width()!==n&&(s=1,a.width(n),a._resizeWidth.skip(!0)),a.height()!==r&&(s=1,a.height(r),a._resizeHeight.skip(!0)),a._viewWidth!==e&&(a._resize=1,a._viewWidth=e),a._viewHeight!==t&&(a._resize=1,a._viewHeight=t),a._origin[0]===i[0]&&a._origin[1]===i[1]||(a._resize=1,a._origin=i),s&&a.run("enter"),o&&a.runAfter(function(){a.resize()})},!1,1)}},"./node_modules/vega-view/src/state.js":function(e,t,n){"use strict";n.r(t),n.d(t,"getState",function(){return o}),n.d(t,"setState",function(){return u});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js");function o(e){return this._runtime.getState(e||{data:a,signals:s,recurse:!0})}function a(e,t){return t.modified&&Object(i.isArray)(t.input.value)&&e.indexOf("_:vega:_")}function s(e,t){return!("parent"===e||t instanceof r.transforms.proxy)}function u(e){var t=this;return t.runAfter(function(){t._trigger=!1,t._runtime.setState(e),t.run().runAfter(function(){t._trigger=!0})}),this}},"./node_modules/vega-voronoi/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-voronoi/src/Voronoi.js");n.d(t,"voronoi",function(){return r.default})},"./node_modules/vega-voronoi/src/Voronoi.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return a});var r=n("./node_modules/vega-dataflow/index.js"),i=n("./node_modules/vega-util/index.js"),o=n("./node_modules/d3-voronoi/index.js");function a(e){r.Transform.call(this,null,e)}a.Definition={type:"Voronoi",metadata:{modifies:!0},params:[{name:"x",type:"field",required:!0},{name:"y",type:"field",required:!0},{name:"size",type:"number",array:!0,length:2},{name:"extent",type:"array",array:!0,length:2,default:[[-1e5,-1e5],[1e5,1e5]],content:{type:"number",array:!0,length:2}},{name:"as",type:"string",default:"path"}]};var s=[[-1e5,-1e5],[1e5,1e5]];Object(i.inherits)(a,r.Transform).transform=function(e,t){var n,r,i,a,u=e.as||"path",c=t.source;for(n=Object(o.voronoi)().x(e.x).y(e.y),e.size?n.size(e.size):n.extent(e.extent||s),this.value=n=n(c),r=n.polygons(),i=0,a=c.length;i<a;++i)c[i][u]=r[i]?"M"+r[i].join("L")+"Z":null;return t.reflow(e.modified()).modifies(u)}},"./node_modules/vega-wordcloud/index.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-wordcloud/src/Wordcloud.js");n.d(t,"wordcloud",function(){return r.default})},"./node_modules/vega-wordcloud/src/CloudLayout.js":function(e,t,n){"use strict";n.r(t);var r=n("./node_modules/vega-canvas/index.js"),i=Math.PI/180,o=64,a=2048;function s(e,t,n,r){if(!t.sprite){var s=e.context,u=e.ratio;s.clearRect(0,0,(o<<5)/u,a/u);var c,l,d,p,f,h=0,m=0,g=0,v=n.length;for(--r;++r<v;){if(t=n[r],s.save(),s.font=t.style+" "+t.weight+" "+~~((t.size+1)/u)+"px "+t.font,c=s.measureText(t.text+"m").width*u,d=t.size<<1,t.rotate){var _=Math.sin(t.rotate*i),y=Math.cos(t.rotate*i),b=c*y,j=c*_,x=d*y,O=d*_;c=Math.max(Math.abs(b+O),Math.abs(b-O))+31>>5<<5,d=~~Math.max(Math.abs(j+x),Math.abs(j-x))}else c=c+31>>5<<5;if(d>g&&(g=d),h+c>=o<<5&&(h=0,m+=g,g=0),m+d>=a)break;s.translate((h+(c>>1))/u,(m+(d>>1))/u),t.rotate&&s.rotate(t.rotate*i),s.fillText(t.text,0,0),t.padding&&(s.lineWidth=2*t.padding,s.strokeText(t.text,0,0)),s.restore(),t.width=c,t.height=d,t.xoff=h,t.yoff=m,t.x1=c>>1,t.y1=d>>1,t.x0=-t.x1,t.y0=-t.y1,t.hasText=!0,h+=c}for(var w=s.getImageData(0,0,(o<<5)/u,a/u).data,C=[];--r>=0;)if((t=n[r]).hasText){for(l=(c=t.width)>>5,d=t.y1-t.y0,p=0;p<d*l;p++)C[p]=0;if(null==(h=t.xoff))return;m=t.yoff;var E=0,S=-1;for(f=0;f<d;f++){for(p=0;p<c;p++){var k=l*f+(p>>5),M=w[(m+f)*(o<<5)+(h+p)<<2]?1<<31-p%32:0;C[k]|=M,E|=M}E?S=f:(t.y0++,d--,f--,m++)}t.y1=t.y0+S,t.sprite=C.slice(0,(t.y1-t.y0)*l)}}}function u(e,t,n){n>>=5;for(var r,i=e.sprite,o=e.width>>5,a=e.x-(o<<4),s=127&a,u=32-s,c=e.y1-e.y0,l=(e.y+e.y0)*n+(a>>5),d=0;d<c;d++){r=0;for(var p=0;p<=o;p++)if((r<<u|(p<o?(r=i[d*o+p])>>>s:0))&t[l+p])return!0;l+=n}return!1}function c(e,t){var n=e[0],r=e[1];t.x+t.x0<n.x&&(n.x=t.x+t.x0),t.y+t.y0<n.y&&(n.y=t.y+t.y0),t.x+t.x1>r.x&&(r.x=t.x+t.x1),t.y+t.y1>r.y&&(r.y=t.y+t.y1)}function l(e){var t=e[0]/e[1];return function(e){return[t*(e*=.1)*Math.cos(e),e*Math.sin(e)]}}function d(e){return"function"==typeof e?e:function(){return e}}t.default=function(){var e,t,n,i,f,h,m,g=[256,256],v=l,_=[],y=Math.random,b={};function j(e,t,n){for(var r,i,o,a,s,c=t.x,l=t.y,d=Math.sqrt(g[0]*g[0]+g[1]*g[1]),p=v(g),f=y()<.5?1:-1,h=-f;(r=p(h+=f))&&(i=~~r[0],o=~~r[1],!(Math.min(Math.abs(i),Math.abs(o))>=d));)if(t.x=c+i,t.y=l+o,!(t.x+t.x0<0||t.y+t.y0<0||t.x+t.x1>g[0]||t.y+t.y1>g[1]||n&&u(t,e,g[0])||n&&(s=n,!((a=t).x+a.x1>s[0].x&&a.x+a.x0<s[1].x&&a.y+a.y1>s[0].y&&a.y+a.y0<s[1].y)))){for(var m,_=t.sprite,b=t.width>>5,j=g[0]>>5,x=t.x-(b<<4),O=127&x,w=32-O,C=t.y1-t.y0,E=(t.y+t.y0)*j+(x>>5),S=0;S<C;S++){m=0;for(var k=0;k<=b;k++)e[E+k]|=m<<w|(k<b?(m=_[S*b+k])>>>O:0);E+=j}return t.sprite=null,!0}return!1}return b.layout=function(){for(var u=function(e){e.width=e.height=1;var t=Math.sqrt(e.getContext("2d").getImageData(0,0,1,1).data.length>>2);e.width=(o<<5)/t,e.height=a/t;var n=e.getContext("2d");return n.fillStyle=n.strokeStyle="red",n.textAlign="center",{context:n,ratio:t}}(Object(r.canvas)()),l=function(e){var t=[],n=-1;for(;++n<e;)t[n]=0;return t}((g[0]>>5)*g[1]),d=null,p=_.length,v=-1,b=[],x=_.map(function(r){return{text:e(r),font:t(r),style:i(r),weight:f(r),rotate:h(r),size:~~n(r),padding:m(r),xoff:0,yoff:0,x1:0,y1:0,x0:0,y0:0,hasText:!1,sprite:null,datum:r}}).sort(function(e,t){return t.size-e.size});++v<p;){var O=x[v];O.x=g[0]*(y()+.5)>>1,O.y=g[1]*(y()+.5)>>1,s(u,O,x,v),O.hasText&&j(l,O,d)&&(b.push(O),d?c(d,O):d=[{x:O.x+O.x0,y:O.y+O.y0},{x:O.x+O.x1,y:O.y+O.y1}],O.x-=g[0]>>1,O.y-=g[1]>>1)}return b},b.words=function(e){return arguments.length?(_=e,b):_},b.size=function(e){return arguments.length?(g=[+e[0],+e[1]],b):g},b.font=function(e){return arguments.length?(t=d(e),b):t},b.fontStyle=function(e){return arguments.length?(i=d(e),b):i},b.fontWeight=function(e){return arguments.length?(f=d(e),b):f},b.rotate=function(e){return arguments.length?(h=d(e),b):h},b.text=function(t){return arguments.length?(e=d(t),b):e},b.spiral=function(e){return arguments.length?(v=p[e]||e,b):v},b.fontSize=function(e){return arguments.length?(n=d(e),b):n},b.padding=function(e){return arguments.length?(m=d(e),b):m},b.random=function(e){return arguments.length?(y=e,b):y},b};var p={archimedean:l,rectangular:function(e){var t=4*e[0]/e[1],n=0,r=0;return function(e){var i=e<0?-1:1;switch(Math.sqrt(1+4*i*e)-i&3){case 0:n+=t;break;case 1:r+=4;break;case 2:n-=t;break;default:r-=4}return[n,r]}}}},"./node_modules/vega-wordcloud/src/Wordcloud.js":function(e,t,n){"use strict";n.r(t),n.d(t,"default",function(){return l});var r=n("./node_modules/vega-wordcloud/src/CloudLayout.js"),i=n("./node_modules/vega-dataflow/index.js"),o=n("./node_modules/vega-util/index.js"),a=n("./node_modules/vega-scale/index.js"),s=n("./node_modules/vega-statistics/index.js"),u=["x","y","font","fontSize","fontStyle","fontWeight","angle"],c=["text","font","rotate","fontSize","fontStyle","fontWeight"];function l(e){i.Transform.call(this,Object(r.default)(),e)}l.Definition={type:"Wordcloud",metadata:{modifies:!0},params:[{name:"size",type:"number",array:!0,length:2},{name:"font",type:"string",expr:!0,default:"sans-serif"},{name:"fontStyle",type:"string",expr:!0,default:"normal"},{name:"fontWeight",type:"string",expr:!0,default:"normal"},{name:"fontSize",type:"number",expr:!0,default:14},{name:"fontSizeRange",type:"number",array:"nullable",default:[10,50]},{name:"rotate",type:"number",expr:!0,default:0},{name:"text",type:"field"},{name:"spiral",type:"string",values:["archimedean","rectangular"]},{name:"padding",type:"number",expr:!0},{name:"as",type:"string",array:!0,length:7,default:u}]},Object(o.inherits)(l,i.Transform).transform=function(e,t){var n=e.modified();if(n||t.changed(t.ADD_REM)||c.some(function(n){var r=e[n];return Object(o.isFunction)(r)&&t.modified(r.fields)})){var r,i=t.materialize(t.SOURCE).source,l=this.value,d=e.as||u,p=e.fontSize||14;if(Object(o.isFunction)(p)?r=e.fontSizeRange:p=Object(o.constant)(p),r){var f=p,h=Object(a.scale)("sqrt")().domain(function(e,t){for(var n,r=1/0,i=-1/0,o=0,a=t.length;o<a;++o)(n=e(t[o]))<r&&(r=n),n>i&&(i=n);return[r,i]}(f,i)).range(r);p=function(e){return h(f(e))}}i.forEach(function(e){e[d[0]]=NaN,e[d[1]]=NaN,e[d[3]]=0});for(var m,g,v=l.words(i).text(e.text).size(e.size||[500,500]).padding(e.padding||1).spiral(e.spiral||"archimedean").rotate(e.rotate||0).font(e.font||"sans-serif").fontStyle(e.fontStyle||"normal").fontWeight(e.fontWeight||"normal").fontSize(p).random(s.random).layout(),_=l.size(),y=_[0]>>1,b=_[1]>>1,j=0,x=v.length;j<x;++j)(g=(m=v[j]).datum)[d[0]]=m.x+y,g[d[1]]=m.y+b,g[d[2]]=m.font,g[d[3]]=m.size,g[d[4]]=m.style,g[d[5]]=m.weight,g[d[6]]=m.rotate;return t.reflow(n).modifies(d)}}}}]);
</script><script type="text/javascript">
(window.webpackJsonp=window.webpackJsonp||[]).push([["main"],{"./src/$$_lazy_route_resource lazy recursive":function(t,e){function a(t){return Promise.resolve().then(function(){var e=new Error("Cannot find module '"+t+"'");throw e.code="MODULE_NOT_FOUND",e})}a.keys=function(){return[]},a.resolve=a,t.exports=a,a.id="./src/$$_lazy_route_resource lazy recursive"},"./src/app/accordion/accordion.component.html":function(t,e){t.exports='<div *ngIf="data.byReg(case).length > 1">\n  <mat-expansion-panel\n  [expanded]="data.allExpandState"\n  #expPanel\n  class="mat-elevation-z0">\n     <mat-expansion-panel-header\n     class="mat-expansion-panel-header"\n     collapsedHeight="auto"\n     expandedHeight="auto"\n     style="padding:0.05vh">\n       <mat-panel-title fxLayoutAlign=\'space-between center\'>\n         <td class="test-title">\n           {{ data.byID(data.byReg(case)[0])?.title }}\n         </td>\n         <td>\n           <button mat-button style="color:black"> Info </button>\n         </td>\n       </mat-panel-title>\n     </mat-expansion-panel-header>\n       <div>\n         <p> {{ data.byID(data.byReg(case)[0])?.summary }} </p>\n       </div>\n </mat-expansion-panel>\n</div>\n  <mat-accordion\n  multi="true"\n  displayMode="flat">\n       <mat-expansion-panel\n       *ngFor="let testId of data.byReg(case); let i = index"\n       #expPanel\n       [expanded]="data.allExpandState"\n       class="mat-elevation-z0">\n          <mat-expansion-panel-header\n          class="mat-expansion-panel-header"\n          collapsedHeight="auto"\n          expandedHeight="auto"\n          style="padding:0.05vh">\n            <mat-panel-title\n            *ngIf="data.byReg(case).length === 1; else noTitle"\n            fxLayoutAlign=\'space-between center\'>\n              <td class="test-title">\n                {{ data.byID(testId)?.title }}\n              </td>\n                <div [ngSwitch]="data.reportType">\n                <td *ngSwitchCase="\'snapshot\'">\n                  <app-result-button [testId]="testId"></app-result-button>\n                </td>\n                  <td *ngSwitchCase="\'diff\'">\n                    <app-diff-button [testId]="testId" ></app-diff-button>\n                  </td>\n              </div>\n            </mat-panel-title>\n            <ng-template #noTitle>\n              <mat-panel-title fxLayoutAlign=\'space-between center\'>\n                <td style="padding-left:1vh"> {{ data.getParam(testId, 1) }} </td>\n                <div [ngSwitch]="data.reportType">\n                <td *ngSwitchCase="\'snapshot\'">\n                  <app-result-button [testId]="testId"></app-result-button>\n                </td>\n                <td *ngSwitchCase="\'diff\'">\n                    <app-diff-button [testId]="testId"></app-diff-button>\n                  </td>\n                </div>\n              </mat-panel-title>\n            </ng-template>\n          <div *ngIf=\'isWeighted(testId) ; else whiteBadge\'> \n            <div class=\'badge\'> <span class="vert-align"> X{{ weight }} </span> </div> \n          </div>\n          <ng-template #whiteBadge>\n              <div class=\'badge\' style="background-color: white"></div> \n          </ng-template>\n          </mat-expansion-panel-header>\n          <div fxLayout=\'column\' flexLayoutAlign=\'space-between center\'>\n            <div *ngIf="data.byReg(case).length === 1; else justMessages">\n              <p>{{ data.byID(testId)?.summary }}</p>\n              <p>{{ data.byID(testId)?.message }}</p>\n            </div>\n            <ng-template #justMessages>\n              <div>\n                <tr> {{ data.byID(testId)?.message }} </tr>\n              </div>\n            </ng-template>\n            <div [ngSwitch]="data.reportType">\n              <div *ngSwitchCase="\'snapshot\'">\n              <br>\n              <textarea rows="4" style="overflow-wrap: break-word; width:100%">\n                {{ data.getString(data.byID(testId)?.data) }}\n              </textarea>\n              </div>\n              <div *ngSwitchCase="\'diff\'">\n                  <div *ngFor=\'let results of data.byID(testId)?.diff; let i = index\'>\n                    <p> For model "{{ results.model }}" : {{ results.message }} </p>\n                  </div>\n                  <br>\n                  <textarea rows="4" style="overflow-wrap: break-word; width:100%">\n                      {{ data.extractData(testId) }}\n                  </textarea>\n              </div>\n              <div *ngSwitchCase="\'history\'">\n                  <br>\n                    <div>\n                      <app-vega-plot [testObject]="data.byID(testId)"></app-vega-plot>\n                    </div>\n              </div>\n             </div>\n          </div>\n      </mat-expansion-panel>\n  </mat-accordion>\n'},"./src/app/accordion/accordion.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL2FjY29yZGlvbi9hY2NvcmRpb24uY29tcG9uZW50LnNjc3MifQ== */"},"./src/app/accordion/accordion.component.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"AccordionComponent",function(){return i});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=a("./src/app/report-data.service.ts"),i=function(){function t(t){this.data=t,this.weight=1}return t.prototype.ngOnInit=function(){},t.prototype.isWeighted=function(t){return t in this.data.testWeightsExpanded&&1!==this.data.testWeightsExpanded[t]&&null!==this.data.testWeightsExpanded[t]&&(this.weight=this.data.testWeightsExpanded[t],!0)},n.__decorate([Object(s.Input)(),n.__metadata("design:type",String)],t.prototype,"case",void 0),t=n.__decorate([Object(s.Component)({selector:"app-accordion",template:a("./src/app/accordion/accordion.component.html"),encapsulation:s.ViewEncapsulation.None,styles:[a("./src/app/accordion/accordion.component.scss")]}),n.__metadata("design:paramtypes",[o.ReportDataService])],t)}()},"./src/app/app.component.html":function(t,e){t.exports="<app-header></app-header>\n<app-dashboard></app-dashboard>\n"},"./src/app/app.component.scss":function(t,e){t.exports=".example-container {\n  display: flex;\n  flex-direction: column;\n  min-width: 300px; }\n\n.mat-table {\n  overflow: auto;\n  max-height: 500px; }\n\n.mat-header-cell.mat-sort-header-sorted {\n  color: black; }\n\n.mat-expansion-panel-content:not(.mat-expanded) {\n  height: 0px;\n  visibility: hidden; }\n\n.mat-expansion-panel-header {\n  height: 48px; }\n\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbIi9ob21lL3RyYXZpcy9idWlsZC9vcGVuY29icmEvbWVtb3RlL21lbW90ZS1yZXBvcnQtYXBwL3NyYy9hcHAvYXBwLmNvbXBvbmVudC5zY3NzIl0sIm5hbWVzIjpbXSwibWFwcGluZ3MiOiJBQUFBO0VBQ0UsY0FBYTtFQUNiLHVCQUFzQjtFQUN0QixpQkFBZ0IsRUFDakI7O0FBRUQ7RUFDRSxlQUFjO0VBQ2Qsa0JBQWlCLEVBQ2xCOztBQUVEO0VBQ0UsYUFBWSxFQUNiOztBQUVEO0VBQ0ksWUFBVztFQUNYLG1CQUFrQixFQUNyQjs7QUFFRDtFQUNJLGFBQVksRUFDZiIsImZpbGUiOiJzcmMvYXBwL2FwcC5jb21wb25lbnQuc2NzcyIsInNvdXJjZXNDb250ZW50IjpbIi5leGFtcGxlLWNvbnRhaW5lciB7XG4gIGRpc3BsYXk6IGZsZXg7XG4gIGZsZXgtZGlyZWN0aW9uOiBjb2x1bW47XG4gIG1pbi13aWR0aDogMzAwcHg7XG59XG5cbi5tYXQtdGFibGUge1xuICBvdmVyZmxvdzogYXV0bztcbiAgbWF4LWhlaWdodDogNTAwcHg7XG59XG5cbi5tYXQtaGVhZGVyLWNlbGwubWF0LXNvcnQtaGVhZGVyLXNvcnRlZCB7XG4gIGNvbG9yOiBibGFjaztcbn1cblxuLm1hdC1leHBhbnNpb24tcGFuZWwtY29udGVudDpub3QoLm1hdC1leHBhbmRlZCkge1xuICAgIGhlaWdodDogMHB4O1xuICAgIHZpc2liaWxpdHk6IGhpZGRlbjtcbn1cblxuLm1hdC1leHBhbnNpb24tcGFuZWwtaGVhZGVyICB7XG4gICAgaGVpZ2h0OiA0OHB4O1xufVxuIl19 */"},"./src/app/app.component.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"AppComponent",function(){return o});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=function(){function t(){this.title="app"}return t=n.__decorate([Object(s.Component)({selector:"app-root",template:a("./src/app/app.component.html"),styles:[a("./src/app/app.component.scss")]})],t)}()},"./src/app/app.material.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"AppMaterialModule",function(){return r});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=a("./node_modules/@angular/platform-browser/fesm5/animations.js"),i=a("./node_modules/@angular/material/esm5/material.es5.js"),r=function(){function t(){}return t=n.__decorate([Object(s.NgModule)({imports:[o.BrowserAnimationsModule,i.MatBadgeModule,i.MatButtonModule,i.MatCheckboxModule,i.MatListModule,i.MatToolbarModule,i.MatIconModule,i.MatFormFieldModule,i.MatInputModule,i.MatCardModule,i.MatSidenavModule,i.MatMenuModule,i.MatAutocompleteModule,i.MatDialogModule,i.MatTableModule,i.MatSortModule,i.MatPaginatorModule,i.MatTooltipModule,i.MatProgressBarModule,i.MatGridListModule,i.MatSelectModule,i.MatExpansionModule],exports:[o.BrowserAnimationsModule,i.MatBadgeModule,i.MatButtonModule,i.MatCheckboxModule,i.MatListModule,i.MatToolbarModule,i.MatIconModule,i.MatFormFieldModule,i.MatInputModule,i.MatCardModule,i.MatSidenavModule,i.MatMenuModule,i.MatAutocompleteModule,i.MatDialogModule,i.MatTableModule,i.MatSortModule,i.MatPaginatorModule,i.MatTooltipModule,i.MatProgressBarModule,i.MatGridListModule,i.MatSelectModule,i.MatExpansionModule]})],t)}()},"./src/app/app.module.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"AppModule",function(){return T});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/platform-browser/fesm5/platform-browser.js"),o=a("./node_modules/@angular/core/fesm5/core.js"),i=a("./src/app/app.material.ts"),r=a("./node_modules/@angular/flex-layout/esm5/flex-layout.es5.js"),c=a("./node_modules/@angular/common/fesm5/http.js"),l=a("./src/app/app.component.ts"),d=a("./src/app/header/header.component.ts"),p=a("./src/app/dashboard/dashboard.component.ts"),u=a("./src/app/dashboard/universal-core/universal-core.component.ts"),m=a("./src/app/dashboard/statistics/statistics.component.ts"),M=a("./src/app/report-data.service.ts"),h=a("./src/app/keys.pipe.ts"),y=a("./src/app/dashboard/system-information/system-information.component.ts"),f=a("./src/app/accordion/accordion.component.ts"),g=a("./src/app/result-button/result-button.component.ts"),b=a("./src/app/diff-button/diff-button.component.ts"),x=a("./src/app/vega-plot/vega-plot.component.ts"),L=a("./src/app/vega-bar-chart/vega-bar-chart.component.ts"),v=a("./src/app/vega-total-score-bar-chart/vega-total-score-bar-chart.component.ts"),j=a("./src/app/vega-grouped-bar-chart/vega-grouped-bar-chart.component.ts"),w=a("./src/app/help-dialog/help-dialog.component.ts"),N=a("./src/app/score-formula/score-formula.component.ts"),T=function(){function t(t){this.reportDataService=t,this.reportDataService.loadResults()}return t.prototype.ngOnInit=function(){},t=n.__decorate([Object(o.NgModule)({declarations:[l.AppComponent,d.HeaderComponent,p.DashboardComponent,u.UniversalCoreComponent,m.StatisticsComponent,h.KeysPipe,y.SystemInformationComponent,f.AccordionComponent,g.ResultButtonComponent,b.DiffButtonComponent,x.VegaPlotComponent,L.VegaBarChartComponent,j.VegaGroupedBarChartComponent,v.VegaTotalScoreBarChartComponent,w.HelpDialogComponent,w.HelpDialogTextComponent,N.ScoreFormulaComponent],imports:[s.BrowserModule,i.AppMaterialModule,r.FlexLayoutModule,c.HttpClientModule],exports:[s.BrowserModule,i.AppMaterialModule],entryComponents:[w.HelpDialogComponent,w.HelpDialogTextComponent],providers:[M.ReportDataService],bootstrap:[l.AppComponent]}),n.__metadata("design:paramtypes",[M.ReportDataService])],t)}()},"./src/app/dashboard/dashboard.component.html":function(t,e){t.exports="<div\n  fxLayout='row wrap'\n  fxLayoutAlign='center stretch'\n  fxLayout.xs='column'>\n  <div\n    fxLayout='column'\n    fxFlex=\"45%\"\n    fxFlexAlign='stretch'\n    style=\"margin:0.5vmax\">\n    <mat-card\n      class=\"mat-elevation-z2\"\n      style=\"background-color:whitesmoke; height: 5vmax\"> \n      <mat-card-title> Independent Section </mat-card-title>\n      <mat-card-subtitle> Contains tests that are independent of the class of modeled organism, a model's complexity or types of identifiers\n          that are used to describe its components. Parameterization or initialization of the network is not required. See readme for more details. </mat-card-subtitle>\n      </mat-card>\n      <mat-card fxFlex *ngIf='data.scoredCard'>\n        <app-universal-core [scoreData]=\"data.scoredCard\"></app-universal-core>\n      </mat-card>\n  </div>\n  <div\n    fxLayout='column'\n    fxFlex=\"45%\"\n    fxFlexAlign='stretch'\n    fxLayoutAlign='start stretch'\n    style=\"margin:0.5vmax\"\n    *ngIf='data'>\n      <mat-card\n      class=\"mat-elevation-z2\"\n      style=\"background-color:whitesmoke; height: 5vmax\"> \n      <mat-card-title> Specific Section </mat-card-title>\n      <mat-card-subtitle> Covers general statistics and specific aspects of a metabolic network that are not universally applicable. See readme for more details. </mat-card-subtitle>\n      </mat-card>\n      <div\n        *ngFor=\"let card of data.statisticsCards\">\n          <app-statistics [currentCard]=\"card\"></app-statistics>\n      </div>\n      <div\n        *ngIf='data.metaData'>\n          <app-system-information></app-system-information>\n      </div>\n   </div>\n</div>\n"},"./src/app/dashboard/dashboard.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL2Rhc2hib2FyZC9kYXNoYm9hcmQuY29tcG9uZW50LnNjc3MifQ== */"},"./src/app/dashboard/dashboard.component.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"DashboardComponent",function(){return i});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=a("./src/app/report-data.service.ts"),i=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},t=n.__decorate([Object(s.Component)({selector:"app-dashboard",template:a("./src/app/dashboard/dashboard.component.html"),encapsulation:s.ViewEncapsulation.None,styles:[a("./src/app/dashboard/dashboard.component.scss")]}),n.__metadata("design:paramtypes",[o.ReportDataService])],t)}()},"./src/app/dashboard/statistics/statistics.component.html":function(t,e){t.exports='<mat-card fxFlex>\n  <mat-card-title>{{ currentCard.cardName }}</mat-card-title>\n    <mat-card-content>\n      <div *ngFor="let case of currentCard.associatedTests; let first = first">\n            <mat-expansion-panel\n            *ngIf="first && data.reportType === \'diff\'"\n            disabled\n            #expPanel\n            class="mat-elevation-z0">\n               <mat-expansion-panel-header\n               class="mat-expansion-panel-header"\n               collapsedHeight="auto"\n               style="padding:0.05vh">\n                 <mat-panel-title fxLayoutAlign=\'space-between center\'>\n                    <td class="test-title">\n                        Compared Models\n                    </td>\n                    <div style="margin-right: 10px">\n                    <td *ngFor="let results of data.byID(case)?.diff">\n                      <div class="MatButtonEmulation" style="background: white; color:black"> {{ results.model }} </div>\n                    </td>\n                    </div>\n                 </mat-panel-title>\n               </mat-expansion-panel-header>\n              </mat-expansion-panel>\n         <app-accordion [case]="case"></app-accordion>\n      </div>\n    </mat-card-content>\n  </mat-card>\n'},"./src/app/dashboard/statistics/statistics.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL2Rhc2hib2FyZC9zdGF0aXN0aWNzL3N0YXRpc3RpY3MuY29tcG9uZW50LnNjc3MifQ== */"},"./src/app/dashboard/statistics/statistics.component.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"StatisticsComponent",function(){return r});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=a("./src/app/resultcard.model.ts"),i=a("./src/app/report-data.service.ts"),r=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},n.__decorate([Object(s.Input)(),n.__metadata("design:type",o.ResultCard)],t.prototype,"currentCard",void 0),t=n.__decorate([Object(s.Component)({selector:"app-statistics",template:a("./src/app/dashboard/statistics/statistics.component.html"),encapsulation:s.ViewEncapsulation.None,styles:[a("./src/app/dashboard/statistics/statistics.component.scss")]}),n.__metadata("design:paramtypes",[i.ReportDataService])],t)}()},"./src/app/dashboard/system-information/system-information.component.html":function(t,e){t.exports='  <mat-card fxFlex>\n    <mat-card-title>Environment</mat-card-title>\n    <mat-card-content>\n      <mat-accordion\n      multi="true"\n      displayMode="flat" >\n         <mat-expansion-panel\n         class="mat-elevation-z0"\n         hideToggle="true">\n          <mat-expansion-panel-header\n          class="mat-expansion-panel-header"\n          collapsedHeight="auto"\n          expandedHeight="auto"\n          style="padding:0.05vh">\n            <mat-panel-description fxLayoutAlign="space-between start">\n              <span> Python Version </span>\n              <span>{{data.metaData.python}}</span>\n            </mat-panel-description>\n          </mat-expansion-panel-header>\n        </mat-expansion-panel>\n        <mat-expansion-panel\n        class="mat-elevation-z0"\n        hideToggle="true">\n         <mat-expansion-panel-header\n         class="mat-expansion-panel-header"\n         collapsedHeight="auto"\n         expandedHeight="auto"\n         style="padding:0.05vh">\n           <mat-panel-description fxLayoutAlign="space-between start">\n             <span> Platform </span>\n             <span>{{data.metaData.platform}}</span>\n           </mat-panel-description>\n         </mat-expansion-panel-header>\n       </mat-expansion-panel>\n       <mat-expansion-panel\n       class="mat-elevation-z0"\n       *ngIf=\'data.metaData.packages\'\n       hideToggle="true">\n        <mat-expansion-panel-header\n        class="mat-expansion-panel-header"\n        collapsedHeight="auto"\n        expandedHeight="auto"\n        style="padding:0.05vh">\n          <mat-panel-description fxLayoutAlign="space-between start">\n            <span> Memote Version </span>\n            <span>{{data.metaData.packages.memote}}</span>\n          </mat-panel-description>\n        </mat-expansion-panel-header>\n        <div fxLayout=\'column\' flexLayoutAlign=\'space-between center\'>\n        <div> Package Versions </div>\n        <div>\n          <p></p>\n          <textarea rows="4" style="overflow-wrap: break-word; width:100%">\n          {{data.getString(data.metaData.packages)}}\n          </textarea>\n        </div>\n        </div>\n      </mat-expansion-panel>\n      </mat-accordion>\n    </mat-card-content>\n  </mat-card>\n'},"./src/app/dashboard/system-information/system-information.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL2Rhc2hib2FyZC9zeXN0ZW0taW5mb3JtYXRpb24vc3lzdGVtLWluZm9ybWF0aW9uLmNvbXBvbmVudC5zY3NzIn0= */"},"./src/app/dashboard/system-information/system-information.component.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"SystemInformationComponent",function(){return i});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=a("./src/app/report-data.service.ts"),i=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},t=n.__decorate([Object(s.Component)({selector:"app-system-information",template:a("./src/app/dashboard/system-information/system-information.component.html"),encapsulation:s.ViewEncapsulation.None,styles:[a("./src/app/dashboard/system-information/system-information.component.scss")]}),n.__metadata("design:paramtypes",[o.ReportDataService])],t)}()},"./src/app/dashboard/universal-core/universal-core.component.html":function(t,e){t.exports='<mat-card-title></mat-card-title>\n    <mat-card-content>\n      <div>\n      <div fxFlex>\n        \x3c!-- <h2>Scored Tests</h2> --\x3e\n        <div *ngFor="let section of scoreData.sections | keys">\n          <h2> {{ scoreData.sections[section].title }}</h2>\n          <div *ngFor="let case of scoreData.sections[section].cases; let first = first">\n              <mat-expansion-panel\n              *ngIf="first && data.reportType === \'diff\'"\n              disabled\n              #expPanel\n              class="mat-elevation-z0">\n                 <mat-expansion-panel-header\n                 class="mat-expansion-panel-header"\n                 collapsedHeight="auto"\n                 style="padding:0.05vh">\n                 <mat-panel-title fxLayoutAlign=\'space-between center\'>\n                    <td class="test-title">\n                        Compared Models\n                    </td>\n                    <div style="margin-right: 10px">\n                    <td *ngFor="let results of data.byID(case)?.diff">\n                      <div class="MatButtonEmulation" style="background: white; color:black"> {{ results.model }} </div>\n                    </td>\n                    </div>\n                   </mat-panel-title>\n                 </mat-expansion-panel-header>\n                </mat-expansion-panel>\n            <app-accordion [case]="case"></app-accordion>\n          </div>\n          <div *ngIf="data.reportType != \'history\'">\n          <hr>\n          <app-score-formula [position]="section"></app-score-formula>\n          </div>\n        </div>\n      </div>\n      </div>\n      <div *ngIf="data.reportType != \'history\'">\n          <hr>\n          <hr>\n          <app-score-formula [position]="\'Total Score\'"></app-score-formula>\n          <hr>\n          </div>\n      <div\n        fxLayout=\'row\'\n        fxLayout.lt-lg=\'column\'\n        fxLayoutAlign=\'space-between start\'\n        *ngIf="data.reportType != \'history\'; else vegaHistoryScore">\n        <div fxFlex> Total Score\n          <div *ngIf="data.reportType === \'snapshot\'; else diffScore">\n            <div style="font-size:10vw">{{ data.score.total_score | percent:\'.\'  }}</div>\n          </div>\n          <ng-template #diffScore>\n              <div> \n                <app-vega-total-score-bar-chart> </app-vega-total-score-bar-chart>\n              </div>\n          </ng-template>\n        </div>\n        <div>\n        <div \n          fxFlex \n          class="svg-container"\n          *ngIf="data.reportType === \'snapshot\'; else diffSectionScore"> Score per Category\n          <br>\n              <app-vega-bar-chart ></app-vega-bar-chart>\n        </div>\n        <ng-template #diffSectionScore>\n            <div fxFlex class="svg-container"> Score per Category\n            <br>\n                <app-vega-grouped-bar-chart ></app-vega-grouped-bar-chart>\n          </div>\n        </ng-template>\n      </div>\n      </div>\n      <ng-template #vegaHistoryScore>\n          <div>\n            Total Score\n            <br>\n            <app-vega-plot [testObject]="data.score.total_score"></app-vega-plot>\n          </div>\n      </ng-template>\n</mat-card-content>\n'},"./src/app/dashboard/universal-core/universal-core.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL2Rhc2hib2FyZC91bml2ZXJzYWwtY29yZS91bml2ZXJzYWwtY29yZS5jb21wb25lbnQuc2NzcyJ9 */"},"./src/app/dashboard/universal-core/universal-core.component.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"UniversalCoreComponent",function(){return i});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=a("./src/app/report-data.service.ts"),i=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},n.__decorate([Object(s.Input)(),n.__metadata("design:type",Object)],t.prototype,"scoreData",void 0),t=n.__decorate([Object(s.Component)({selector:"app-universal-core",template:a("./src/app/dashboard/universal-core/universal-core.component.html"),encapsulation:s.ViewEncapsulation.None,styles:[a("./src/app/dashboard/universal-core/universal-core.component.scss")]}),n.__metadata("design:paramtypes",[o.ReportDataService])],t)}()},"./src/app/diff-button/diff-button.component.html":function(t,e){t.exports='<div *ngIf="data.isScored(data.getParam(testId, 0)); else notScored">\n    \x3c!-- If they are from scored tests use a percentage formatting for the results --\x3e\n    <div fxLayoutAlign=\'center stretch\' fxLayout=\'row wrap\'>\n    <div *ngFor="let results of data.byID(testId)?.diff">\n        <div\n          *ngIf="results.result === \'passed\' || results.result === \'failed\'"\n          class="MatButtonEmulation"\n          [ngStyle]="{\'background-color\':highlightColor}">\n            {{ 1 - results.metric | percent : \'1.1\' }}\n        </div>\n          \x3c!-- Skipped tests are marked with a gray button with beige text --\x3e\n        <div\n          *ngIf="results.result === \'skipped\'"\n          class="MatButtonEmulation"\n          style="background:gray; color:beige">\n            Skipped\n        </div>\n          \x3c!-- Tests with errors are marked with a black button with red text --\x3e\n        <div\n          class="MatButtonEmulation"\n          style="color:red;background-color:black"\n          *ngIf="results.result === \'error\'">\n            Errored\n        </div>\n    </div>\n  </div>\n</div>\n<ng-template #notScored>\n  \x3c!-- If they are from unscored tests the user has some control over how --\x3e\n  \x3c!-- they are displayed using the \'format_type\' attribute of each test. --\x3e\n  <div  fxLayoutAlign=\'center stretch\' fxLayout=\'row wrap\'>\n      <div *ngFor="let results of data.byID(testId)?.diff">\n        \x3c!-- Skipped tests are marked with a gray button with beige text --\x3e\n        <div\n        *ngIf="results.result === \'skipped\'"\n        class="MatButtonEmulation"\n        style="background:gray; color:beige">\n          Skipped\n        </div>\n          \x3c!-- Tests with errors are marked with a black button with red text --\x3e\n        <div\n          class="MatButtonEmulation"\n          style="color:red;background-color:black"\n          *ngIf="results.result === \'error\'">\n            Errored\n        </div>\n    <div\n    *ngIf="results.result === \'passed\' || results.result === \'failed\'"\n    class="MatButtonEmulation"\n    [ngStyle]="{\'background-color\':highlightColor}">\n      \x3c!-- Hence for each test, determine now how the result ought to be formatted --\x3e\n      <div *ngIf="data.byID(testId)?.format_type === \'count\'">\n        {{ results.data?.length | number : \'1.\'}}\n      </div>\n      <div *ngIf="data.byID(testId)?.format_type === \'number\'">\n        {{ results.data | number : \'1.2-2\'}}\n      </div>\n      <div *ngIf="data.byID(testId)?.format_type === \'percent\'">\n        {{ results.metric | number : \'1.2-2\'}}\n      </div>\n      <div *ngIf="data.byID(testId)?.format_type === \'raw\'">\n        {{ results.data }}\n      </div>\n    </div>\n    </div>\n  </div>\n</ng-template>\n'},"./src/app/diff-button/diff-button.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL2RpZmYtYnV0dG9uL2RpZmYtYnV0dG9uLmNvbXBvbmVudC5zY3NzIn0= */"},"./src/app/diff-button/diff-button.component.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"DiffButtonComponent",function(){return i});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=a("./src/app/report-data.service.ts"),i=function(){function t(t){this.data=t,this.metricList=[]}return t.prototype.ngOnInit=function(){this.getDiffColor()},t.prototype.getDiffColor=function(){if(this.data.isScored(this.data.getParam(this.testId,0))){for(var t=0,e=this.data.byID(this.testId).diff;t<e.length;t++){var a=e[t];this.metricList.push(a.metric)}this.spectrum(this.ratioOrVariance(this.metricList))}else if("raw"!==this.data.byID(this.testId).format_type){switch(this.data.byID(this.testId).format_type){case"count":for(var n=0,s=this.data.byID(this.testId).diff;n<s.length;n++){a=s[n];this.metricList.push(a.data.length)}break;case"number":for(var o=0,i=this.data.byID(this.testId).diff;o<i.length;o++){a=i[o];this.metricList.push(a.data)}break;case"percent":for(var r=0,c=this.data.byID(this.testId).diff;r<c.length;r++){a=c[r];this.metricList.push(a.metric)}}this.spectrum(this.ratioOrVariance(this.metricList))}else{for(var l=0,d=this.data.byID(this.testId).diff;l<d.length;l++){a=d[l];this.metricList.push(a.data)}this.spectrum(this.determineStringEquality(this.metricList))}},t.prototype.ratioOrVariance=function(t){var e=Math.min.apply(Math,t),a=Math.max.apply(Math,t);return 2===t.length&&t[0]===t[1]?0:0===e?100*(1-(e+1)/(a+1)):100*(1-e/a)},t.prototype.determineStringEquality=function(t){return!!t.length&&!!t.reduce(function(t,e){return t===e?t:NaN})?0:100},t.prototype.spectrum=function(t){var e=(1.19*t+42).toString(10),a=(123-1.05*t).toString(10),n=(184-1.66*t).toString(10);this.highlightColor=["rgb(",e,",",a,",",n,")"].join("")},n.__decorate([Object(s.Input)(),n.__metadata("design:type",String)],t.prototype,"testId",void 0),t=n.__decorate([Object(s.Component)({selector:"app-diff-button",template:a("./src/app/diff-button/diff-button.component.html"),encapsulation:s.ViewEncapsulation.None,styles:[a("./src/app/diff-button/diff-button.component.scss")]}),n.__metadata("design:paramtypes",[o.ReportDataService])],t)}()},"./src/app/header/header.component.html":function(t,e){t.exports='<mat-toolbar\n  class="mat-elevation-z9 fixed-header"\n  color=\'primary\'>\n      <img aria-label="logo" style="width: 5vw; padding-right:1vw"\n       width="48px"\n       height="48px"\n       src="data:image/svg+xml;base64, "/>\n      <span style="text-align:left; padding-right:1vw" *ngIf="data.byID(\'test_model_id_presence\')"> {{ data.byID("test_model_id_presence").data }} </span>\n      <button mat-button (click)="data.togglePanel()"> {{ !data.allExpandState ? "Expand" : "Collapse" }} All</button>\n      <app-help-dialog></app-help-dialog>\n      <span style="flex:1 1 auto"></span>\n      <span style="text-align:right; padding-right:1vw" *ngIf="data.metaData">{{ data.metaData.timestamp | date:\'yyyy-MM-dd HH:mm\' }}</span>\n      <a\n        aria-label="GitHub Repository"\n        href="https://github.com/opencobra/memote"\n        mat-button>\n        <span class="mat-button-wrapper">\n          <img\n          width="40px"\n          height="40px"\n          src="data:image/svg+xml;base64,PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0idXRmLTgiPz48IURPQ1RZUEUgc3ZnIFBVQkxJQyAiLS8vVzNDLy9EVEQgU1ZHIDEuMS8vRU4iICJodHRwOi8vd3d3LnczLm9yZy9HcmFwaGljcy9TVkcvMS4xL0RURC9zdmcxMS5kdGQiPjxzdmcgdmVyc2lvbj0iMS4xIiBpZD0iTGF5ZXJfMSIgeG1sbnM9Imh0dHA6Ly93d3cudzMub3JnLzIwMDAvc3ZnIiB4bWxuczp4bGluaz0iaHR0cDovL3d3dy53My5vcmcvMTk5OS94bGluayIgeD0iMHB4IiB5PSIwcHgiIHdpZHRoPSIxNnB4IiBoZWlnaHQ9IjE2cHgiIHZpZXdCb3g9IjAgMCAxNiAxNiIgZW5hYmxlLWJhY2tncm91bmQ9Im5ldyAwIDAgMTYgMTYiIHhtbDpzcGFjZT0icHJlc2VydmUiPjx0aXRsZT5tYXJrLWdpdGh1YjwvdGl0bGU+PGRlc2M+Q3JlYXRlZCB3aXRoIFNrZXRjaC48L2Rlc2M+PGcgaWQ9Ik9jdGljb25zIj48ZyBpZD0ibWFyay1naXRodWIiPjxwYXRoIGlkPSJTaGFwZSIgZmlsbD0iI0ZGRkZGRiIgZD0iTTgsMEMzLjU4LDAsMCwzLjU4LDAsOGMwLDMuNTQsMi4yOSw2LjUzLDUuNDcsNy41OWMwLjQsMC4wNywwLjU1LTAuMTcsMC41NS0wLjM4YzAtMC4xOS0wLjAxLTAuODE5LTAuMDEtMS40OUM0LDE0LjA5LDMuNDgsMTMuMjMsMy4zMiwxMi43OGMtMC4wOS0wLjIzLTAuNDgtMC45NC0wLjgyLTEuMTNjLTAuMjgtMC4xNS0wLjY4LTAuNTIxLTAuMDEtMC41M2MwLjYzLTAuMDExLDEuMDgsMC41OCwxLjIzLDAuODE5YzAuNzIsMS4yMTEsMS44NywwLjg3MSwyLjMzLDAuNjZjMC4wNy0wLjUyLDAuMjgtMC44NjksMC41MS0xLjA2OWMtMS43OC0wLjItMy42NC0wLjg5LTMuNjQtMy45NWMwLTAuODcsMC4zMS0xLjU5LDAuODItMi4xNWMtMC4wOC0wLjItMC4zNi0xLjAyLDAuMDgtMi4xMmMwLDAsMC42Ny0wLjIxLDIuMiwwLjgyYzAuNjQtMC4xOCwxLjMyLTAuMjcsMi0wLjI3czEuMzYsMC4wOSwyLDAuMjdjMS41My0xLjA0LDIuMjAxLTAuODIsMi4yMDEtMC44MmMwLjQzOSwxLjEsMC4xNTksMS45MiwwLjA3OSwyLjEyYzAuNTExLDAuNTYsMC44MiwxLjI3LDAuODIsMi4xNWMwLDMuMDctMS44NywzLjc1LTMuNjQ5LDMuOTVjMC4yODksMC4yNSwwLjUzOSwwLjcyOSwwLjUzOSwxLjQ3OWMwLDEuMDctMC4wMSwxLjkzLTAuMDEsMi4yYzAsMC4yMSwwLjE1LDAuNDYsMC41NSwwLjM4QzEzLjcxLDE0LjUzLDE2LDExLjUzLDE2LDhDMTYsMy41OCwxMi40MiwwLDgsMEw4LDB6Ii8+PC9nPjwvZz48L3N2Zz4="/>\n        </span>\n        </a>\n</mat-toolbar>'},"./src/app/header/header.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL2hlYWRlci9oZWFkZXIuY29tcG9uZW50LnNjc3MifQ== */"},"./src/app/header/header.component.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"HeaderComponent",function(){return i});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=a("./src/app/report-data.service.ts"),i=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},t=n.__decorate([Object(s.Component)({selector:"app-header",template:a("./src/app/header/header.component.html"),encapsulation:s.ViewEncapsulation.None,styles:[a("./src/app/header/header.component.scss")]}),n.__metadata("design:paramtypes",[o.ReportDataService])],t)}()},"./src/app/help-dialog/help-dialog-text.component.html":function(t,e){t.exports='<h2 mat-dialog-title>Readme</h2>\n<mat-dialog-content>\n    <mat-card>\n    <mat-card-title>What do the colours mean?</mat-card-title>\n    <mat-card-content>\n    <h4>Snapshot Report</h4>\n    Results without highlights are kept in the main\n        <span style="color: rgb(42, 123, 184)"> <b> blue </b> </span> \n        colour of the memote colour scheme. Scored results will be marked with a gradient ranging from \n        <span style="color: rgb(161, 18, 18)"> <b> red </b> </span> to \n        <span style="color: rgb(18, 161, 46)"> <b> green </b> </span> denoting a low or a high score respectively:\n        <div fxLayoutAlign=\'center center\'>\n            <span> <b>0%</b></span>\n            <div style="display: inline-block; vertical-align: middle">\n                <svg width="120" height="60">\n                    <defs>\n                        <linearGradient id="grad1">\n                        <stop offset="0%" style="stop-color:rgb(161, 18, 18);stop-opacity:1" />\n                                <stop offset="100%" style="stop-color:rgb(18, 161, 46);stop-opacity:1" />\n                        </linearGradient>\n                    </defs>\n                    <rect fill="url(#grad1)"\n                        x="10" y="10" width="100" height="100"/>\n                </svg>\n            </div>     \n        <span> <b>100% </b></span>\n        </div>\n        <h4>Diff Report</h4>\n        The colour in the Diff Report depends on the ratio of the sample minimum to the sample maxium. \n        Result sets where the sample minimum and the sample maximum are identical will be coloured in the main\n        <span style="color: rgb(42, 123, 184)"> <b> blue </b> </span> \n        colour of the memote colour scheme. Result sets where the sample minimum is very small relative to the sample maximum will appear\n        <span style="color: rgb(161, 18, 18)"> <b> red </b> </span>. \n        This ratio is calculated with the following formula:\n        <br>\n        <b> 1 - (Min / Max)) * 100 </b>\n        This is then mapped to the following gradient:\n        <div fxLayoutAlign=\'center center\'>\n            <span> <b>Identical</b></span>\n            <div style="display: inline-block; vertical-align: middle">\n                <svg width="120" height="60">\n                    <defs>\n                        <linearGradient id="grad2">\n                        <stop offset="0%" style="stop-color:rgb(42, 123, 184);stop-opacity:1" />\n                        <stop offset="100%" style="stop-color:rgb(161, 18, 18);stop-opacity:1" />\n                        </linearGradient>\n                    </defs>\n                    <rect fill="url(#grad2)"\n                        x="10" y="10" width="100" height="100"/>\n                </svg>\n            </div>     \n        <span> <b>Different</b></span>\n        </div>\n    </mat-card-content>\n    </mat-card>\n    <mat-card>\n        <mat-card-title>Why are the tests divided into two principal sections?</mat-card-title>\n        <mat-card-content>\n    <p>The tests in the left section are independent of the class of modeled organism, the complexity of a genome-scale metabolic model (GSM) or the types of identifiers\n         that are used to describe its components. Because of this, it is possible to calculate a score for these tests which further allows for the quick comparison of GSMs. The unscored section on ther right \n         provides general statistics and covers specific aspects of a metabolic network that are not universally applicable. For instance,\n           dedicated quality control of the biomass equation only applies to GSM which are used to investigate cell growth.</p>\n        </mat-card-content>\n    </mat-card>\n    <mat-card>\n            <mat-card-title>"Reconstructions" versus "Models"</mat-card-title>\n            <mat-card-content>\n        <p>The variety of constraints-based modeling approaches and the fundamental differences between various organisms compound the assessment of GSMs. For\n                instance, some authors may publish metabolic networks which are parameterized, ready to run flux balance analysis (FBA), these are referred to simply as \'models\'. Alternatively,\n                 others may publish unconstrained metabolic databases (referred to as \'reconstructions\'), from which several models can be derived by applying differential constraints. Both can be encoded in SBML. \n                 With having an independent scored test section, we attempt to make both \'models\' and \'reconstructions\' comparable, although a user should be aware that this difference exists and is subject to <a href="https://github.com/opencobra/memote/issues/228"> some discussion </a>.\n                Please note, that some tests in the specific section may error for a reconstruction as they require initialization.</p>\n            </mat-card-content>\n        </mat-card>\n    <mat-card>\n    <mat-card-title>How is the score calculated?</mat-card-title>\n    <mat-card-content>\n    <p>Each test in the independent section provides a relative measure of completeness with regard to the tested property. The final score is the weighted sum of all individual test\n         results normalized by the maximally achievable score i.e. all individual results at 100%. Individual tests can be weighted, but it is also possible to apply weighting to entire test \n         categories. Hence the final score is calculated:</p>\n         <br>\n         \x3c!-- Generated with this latex formula \\[ TotalScore = \\frac{\\sum_{Sections} weight_{section} \\times ( \\sum_{Tests} weight_{test} \\times score_{test})}{MaxScore}\\] at http://www.tlhiv.org/ltxpreview/ --\x3e\n         <svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" viewBox="0 0 310 23">\n            <defs>\n              <symbol id="a" overflow="visible">\n                <path d="M4.25-6.05c.08-.28.1-.34.23-.37.1-.02.43-.02.63-.02 1.01 0 1.45.03 1.45.82 0 .15-.03.54-.08.8l-.03.18c0 .06.03.14.13.14s.12-.08.15-.23L7-6.47c.02-.05.02-.14.02-.17 0-.11-.1-.11-.27-.11H1.22c-.24 0-.25.02-.33.2L.3-4.8c0 .02-.06.16-.06.2 0 .04.07.1.13.1.1 0 .1-.06.17-.22.53-1.55.8-1.72 2.27-1.72h.39c.28 0 .28.03.28.13 0 .04-.03.17-.05.2L2.1-.78C2-.42 1.97-.31.91-.31c-.36 0-.43 0-.43.18C.48 0 .6 0 .66 0l.8-.02.87-.01.83.01.86.02c.09 0 .21 0 .21-.2 0-.11-.07-.11-.34-.11-.23 0-.37 0-.62-.02-.3-.03-.38-.06-.38-.22 0-.01 0-.06.05-.2zm0 0"/>\n              </symbol>\n              <symbol id="b" overflow="visible">\n                <path d="M4.67-2.72c0-1.05-.69-1.69-1.6-1.69C1.76-4.4.42-2.98.42-1.58.4-.59 1.08.11 2 .11c1.33 0 2.67-1.38 2.67-2.83zM2.02-.1c-.44 0-.88-.31-.88-1.1 0-.48.27-1.56.58-2.06.5-.76 1.08-.92 1.36-.92.58 0 .87.49.87 1.08 0 .4-.2 1.44-.58 2.08-.35.58-.9.92-1.35.92zm0 0"/>\n              </symbol>\n              <symbol id="c" overflow="visible">\n                <path d="M2.05-3.98h.93c.2 0 .32 0 .32-.2 0-.12-.11-.12-.28-.12h-.88c.36-1.42.4-1.6.4-1.67 0-.17-.12-.26-.29-.26-.03 0-.31 0-.4.36L1.48-4.3H.53c-.2 0-.3 0-.3.2 0 .12.08.12.29.12h.87C.67-1.16.63-.98.63-.81c0 .54.37.92.92.92 1.01 0 1.57-1.45 1.57-1.53 0-.11-.07-.11-.1-.11-.1 0-.11.03-.16.14C2.44-.34 1.9-.11 1.56-.11c-.2 0-.31-.12-.31-.45 0-.25.03-.31.06-.49zm0 0"/>\n              </symbol>\n              <symbol id="d" overflow="visible">\n                <path d="M3.72-3.77c-.19-.37-.47-.64-.92-.64C1.64-4.4.4-2.94.4-1.48.4-.55.96.1 1.73.1c.2 0 .7-.05 1.3-.75.07.42.43.75.9.75.36 0 .58-.23.75-.55.16-.36.3-.97.3-.98 0-.11-.1-.11-.13-.11-.09 0-.1.05-.14.19-.17.64-.34 1.23-.75 1.23-.28 0-.3-.27-.3-.45 0-.22.02-.31.13-.75.11-.4.13-.52.22-.9l.36-1.38c.06-.29.06-.3.06-.35 0-.17-.1-.26-.28-.26-.25 0-.39.22-.42.43zm-.64 2.58c-.06.19-.06.2-.2.38-.44.54-.85.7-1.13.7-.5 0-.64-.55-.64-.94 0-.5.31-1.72.55-2.18.3-.58.75-.96 1.15-.96.64 0 .78.82.78.88 0 .06-.01.12-.03.17zm0 0"/>\n              </symbol>\n              <symbol id="e" overflow="visible">\n                <path d="M2.58-6.81s0-.11-.14-.11c-.22 0-.96.08-1.22.1-.08 0-.19.02-.19.21 0 .11.11.11.25.11.49 0 .5.1.5.17l-.03.2L.48-1.13a.97.97 0 0 0-.04.33c0 .58.43.92.9.92.33 0 .58-.2.75-.56.18-.38.3-.96.3-.97 0-.11-.1-.11-.12-.11-.1 0-.11.05-.13.19-.17.64-.36 1.23-.77 1.23-.3 0-.3-.31-.3-.45 0-.25.02-.3.07-.49zm0 0"/>\n              </symbol>\n              <symbol id="f" overflow="visible">\n                <path d="M6.44-6.92c0-.03-.03-.11-.13-.11-.04 0-.04.01-.17.16l-.48.56c-.25-.47-.77-.72-1.43-.72-1.26 0-2.45 1.16-2.45 2.36 0 .8.52 1.26 1.03 1.4l1.06.29c.38.09.93.25.93 1.06C4.8-1.03 3.98-.1 3-.1c-.64 0-1.75-.22-1.75-1.46 0-.23.06-.48.06-.53l.02-.08c0-.09-.06-.1-.11-.1a.14.14 0 0 0-.11.04C1.08-2.19.5.1.5.13c0 .04.05.09.11.09C.68.22.7.2.82.06L1.3-.5c.42.58 1.1.72 1.68.72 1.36 0 2.54-1.33 2.54-2.56 0-.7-.35-1.04-.49-1.18-.23-.21-.39-.26-1.26-.48l-.68-.19a.93.93 0 0 1-.6-.9c0-.82.8-1.66 1.74-1.66.82 0 1.43.44 1.43 1.55 0 .3-.04.5-.04.56 0 0 0 .1.11.1s.13-.04.16-.21zm0 0"/>\n              </symbol>\n              <symbol id="g" overflow="visible">\n                <path d="M3.95-3.78c-.17 0-.3 0-.43.12a.53.53 0 0 0-.2.4c0 .24.2.35.38.35.28 0 .55-.25.55-.64 0-.48-.47-.86-1.17-.86C1.73-4.4.4-2.98.4-1.58.4-.68.98.11 2.03.11 3.45.1 4.28-.95 4.28-1.06c0-.06-.05-.14-.1-.14-.07 0-.09.03-.15.1-.78 1-1.87 1-1.98 1-.63 0-.9-.5-.9-1.1 0-.4.2-1.38.54-1.99.31-.58.86-1 1.4-1 .33 0 .72.13.86.4zm0 0"/>\n              </symbol>\n              <symbol id="h" overflow="visible">\n                <path d="M.88-.6l-.1.44c0 .18.14.27.3.27.12 0 .3-.08.37-.28.02-.03.35-1.4.4-1.58.07-.33.26-1.02.3-1.3.05-.12.33-.6.57-.8.08-.07.37-.34.8-.34.26 0 .4.13.42.13-.3.04-.52.28-.52.54 0 .16.11.35.38.35.26 0 .54-.24.54-.6 0-.34-.3-.64-.82-.64-.65 0-1.08.49-1.27.77a.92.92 0 0 0-.92-.77c-.46 0-.64.4-.74.57-.17.34-.3.93-.3.96 0 .11.1.11.12.11.1 0 .1-.01.17-.23.17-.7.37-1.19.73-1.19.16 0 .3.08.3.46 0 .21-.03.32-.16.84zm0 0"/>\n              </symbol>\n              <symbol id="i" overflow="visible">\n                <path d="M1.86-2.3c.3 0 1.03-.03 1.53-.23.7-.3.75-.9.75-1.03 0-.44-.37-.85-1.06-.85-1.11 0-2.63.97-2.63 2.72 0 1.02.6 1.8 1.58 1.8C3.45.1 4.28-.95 4.28-1.06c0-.06-.05-.14-.1-.14-.07 0-.09.03-.15.1-.78 1-1.87 1-1.98 1-.78 0-.88-.85-.88-1.17 0-.12.02-.42.16-1.03zm-.47-.22c.4-1.51 1.42-1.67 1.69-1.67.45 0 .73.3.73.63 0 1.04-1.6 1.04-2.01 1.04zm0 0"/>\n              </symbol>\n              <symbol id="t" overflow="visible">\n                <path d="M4.6-3.38c.06-.21.15-.59.15-.65 0-.17-.14-.27-.28-.27-.13 0-.3.08-.38.28-.03.07-.5 1.97-.56 2.24a2.34 2.34 0 0 0-.08.83c-.23.53-.53.84-.92.84-.8 0-.8-.73-.8-.9 0-.32.05-.7.52-1.94.1-.3.17-.44.17-.64a.8.8 0 0 0-.81-.82C.66-4.4.3-2.95.3-2.87c0 .1.1.1.1.1.12 0 .12-.03.16-.18.28-.92.66-1.24 1.02-1.24.1 0 .25.02.25.33 0 .25-.11.53-.19.7a6.08 6.08 0 0 0-.55 2.02C1.1-.24 1.75.1 2.5.1c.17 0 .64 0 1.03-.7.27.64.95.7 1.25.7.75 0 1.19-.63 1.45-1.22.33-.78.66-2.12.66-2.6 0-.54-.26-.7-.44-.7-.25 0-.5.27-.5.49 0 .12.07.19.14.26.11.11.36.36.36.85a5.9 5.9 0 0 1-.54 1.83c-.25.53-.61.87-1.1.87-.47 0-.73-.3-.73-.87 0-.27.06-.58.1-.72zm0 0"/>\n              </symbol>\n              <symbol id="u" overflow="visible">\n                <path d="M2.83-6.23c0-.2-.14-.36-.36-.36-.28 0-.55.26-.55.53 0 .18.14.36.38.36.23 0 .53-.24.53-.53zm-.75 3.75c.1-.29.1-.32.22-.58.08-.2.12-.35.12-.53a.79.79 0 0 0-.81-.82C.67-4.4.3-2.95.3-2.87c0 .1.1.1.1.1.12 0 .12-.03.16-.18.28-.94.67-1.24 1.02-1.24.08 0 .25 0 .25.32 0 .21-.08.42-.11.53-.08.25-.53 1.4-.69 1.84-.1.25-.23.58-.23.8 0 .47.34.8.8.8.95 0 1.32-1.43 1.32-1.52 0-.11-.1-.11-.12-.11-.1 0-.1.03-.14.19-.2.62-.52 1.23-1.02 1.23-.17 0-.25-.1-.25-.33 0-.25.06-.39.3-1zm0 0"/>\n              </symbol>\n              <symbol id="v" overflow="visible">\n                <path d="M4.69-3.77c.01-.04.03-.1.03-.17 0-.17-.11-.26-.28-.26-.1 0-.38.06-.4.42-.2-.36-.54-.63-.95-.63C1.97-4.4.73-3 .73-1.58.73-.59 1.33 0 2.05 0c.6 0 1.06-.47 1.15-.58l.02.02C3.02.3 2.89.74 2.89.75c-.05.1-.37 1.08-1.44 1.08-.18 0-.51-.02-.8-.11.3-.08.41-.34.41-.52 0-.15-.1-.34-.37-.34a.55.55 0 0 0-.53.58c0 .4.36.6 1.3.6 1.26 0 1.98-.77 2.13-1.37zM3.4-1.28c-.07.26-.3.51-.52.7-.2.17-.52.36-.81.36-.5 0-.64-.51-.64-.92 0-.47.28-1.66.56-2.16.27-.48.69-.89 1.1-.89.67 0 .8.82.8.86 0 .05 0 .11-.02.14zm0 0"/>\n              </symbol>\n              <symbol id="w" overflow="visible">\n                <path d="M2.86-6.81s0-.11-.13-.11c-.23 0-.95.08-1.21.1-.08 0-.2.02-.2.2 0 .12.1.12.24.12.49 0 .5.06.5.17l-.03.2L.6-.38c-.04.14-.04.16-.04.22 0 .23.2.28.3.28a.4.4 0 0 0 .35-.27l.2-.75.2-.89.18-.67c.02-.06.11-.39.11-.45.03-.1.34-.64.69-.92.22-.16.51-.35.95-.35.42 0 .53.35.53.7 0 .54-.37 1.63-.6 2.24-.08.22-.15.34-.15.55 0 .47.36.8.83.8.94 0 1.3-1.44 1.3-1.52 0-.11-.08-.11-.11-.11-.11 0-.11.03-.16.19-.14.53-.47 1.23-1.01 1.23-.18 0-.24-.1-.24-.33 0-.25.08-.48.17-.7.16-.44.61-1.63.61-2.2 0-.64-.39-1.07-1.14-1.07-.62 0-1.1.32-1.48.77zm0 0"/>\n              </symbol>\n              <symbol id="A" overflow="visible">\n                <path d="M3.9-3.73c-.27.01-.48.23-.48.45 0 .14.1.3.31.3.22 0 .46-.18.46-.57C4.19-4 3.77-4.4 3-4.4c-1.31 0-1.69 1.02-1.69 1.46 0 .78.74.92 1.03.98.52.11 1.04.22 1.04.77 0 .25-.22 1.1-1.43 1.1-.14 0-.9 0-1.14-.54.4.05.64-.25.64-.53 0-.22-.17-.35-.37-.35-.27 0-.56.2-.56.66 0 .56.57.97 1.42.97 1.62 0 2.01-1.2 2.01-1.66 0-.36-.18-.6-.3-.72-.27-.28-.57-.34-1-.42-.37-.08-.76-.15-.76-.6 0-.29.24-.9 1.11-.9.25 0 .75.08.9.46zm0 0"/>\n              </symbol>\n              <symbol id="C" overflow="visible">\n                <path d="M9.23-6.05c.1-.36.11-.45.86-.45.22 0 .32 0 .32-.2 0-.11-.1-.11-.27-.11H8.83c-.27 0-.28 0-.4.18L4.8-.92l-.78-5.65c-.04-.23-.05-.23-.32-.23H2.34c-.18 0-.3 0-.3.18 0 .13.1.13.29.13l.44.02c.15.03.21.04.21.17l-.04.18-1.27 5.07c-.1.4-.26.72-1.08.75-.04 0-.17.01-.17.18 0 .1.06.13.14.13.31 0 .66-.03 1-.03L2.58 0c.04 0 .19 0 .19-.2 0-.11-.11-.11-.2-.11-.55 0-.66-.2-.66-.44 0-.06 0-.13.03-.23L3.3-6.41h.01l.86 6.18c.02.12.03.23.14.23.11 0 .17-.1.22-.17l4.03-6.31h.02L7.14-.78c-.1.39-.1.47-.9.47-.16 0-.27 0-.27.18 0 .13.1.13.14.13l1.23-.03L8.6 0c.07 0 .2 0 .2-.2 0-.11-.1-.11-.29-.11-.36 0-.64 0-.64-.17 0-.05 0-.07.05-.25zm0 0"/>\n              </symbol>\n              <symbol id="D" overflow="visible">\n                <path d="M3.33-3.02c.06-.25.3-1.17.98-1.17a1 1 0 0 1 .5.13.58.58 0 0 0-.47.54c0 .16.11.35.38.35.22 0 .53-.17.53-.58 0-.52-.58-.66-.92-.66-.58 0-.92.54-1.05.75A1.1 1.1 0 0 0 2.2-4.4C1.17-4.4.6-3.12.6-2.87c0 .1.1.1.12.1.08 0 .1-.03.12-.1.35-1.07 1-1.32 1.35-1.32.19 0 .53.1.53.67 0 .32-.17.97-.53 2.38-.16.6-.52 1.03-.96 1.03-.06 0-.28 0-.5-.12.25-.07.47-.27.47-.55 0-.27-.22-.34-.36-.34-.3 0-.54.25-.54.57 0 .46.48.66.92.66.67 0 1.03-.7 1.05-.75.12.36.48.75 1.07.75 1.04 0 1.6-1.28 1.6-1.53 0-.11-.08-.11-.11-.11-.1 0-.11.05-.14.1C4.36-.33 3.69-.1 3.37-.1c-.39 0-.54-.31-.54-.66 0-.21.04-.43.15-.87zm0 0"/>\n              </symbol>\n              <symbol id="j" overflow="visible">\n                <path d="M6.84-3.27c.16 0 .35 0 .35-.18 0-.2-.19-.2-.33-.2H.89c-.14 0-.33 0-.33.2 0 .18.19.18.33.18zm.02 1.94c.14 0 .33 0 .33-.2 0-.19-.19-.19-.35-.19H.9c-.14 0-.33 0-.33.19 0 .2.19.2.33.2zm0 0"/>\n              </symbol>\n              <symbol id="y" overflow="visible">\n                <path d="M3.3 2.4c0-.04 0-.06-.17-.23C1.89.92 1.55-.97 1.55-2.5c0-1.73.38-3.47 1.61-4.7.13-.13.13-.14.13-.17 0-.08-.03-.11-.1-.11-.1 0-1 .68-1.6 1.95-.5 1.1-.62 2.2-.62 3.03 0 .78.11 1.98.66 3.13A4.37 4.37 0 0 0 3.2 2.5c.07 0 .1-.03.1-.1zm0 0"/>\n              </symbol>\n              <symbol id="B" overflow="visible">\n                <path d="M2.88-2.5c0-.77-.11-1.97-.66-3.1A4.39 4.39 0 0 0 .67-7.49a.1.1 0 0 0-.1.1c0 .04 0 .05.18.24.98.98 1.55 2.56 1.55 4.64C2.3-.78 1.94.97.7 2.22c-.14.12-.14.14-.14.17 0 .06.05.11.11.11.1 0 1-.69 1.58-1.95.52-1.1.63-2.2.63-3.05zm0 0"/>\n              </symbol>\n              <symbol id="k" overflow="visible">\n                <path d="M4.2 5.33L.66 9.7c-.08.1-.1.11-.1.16 0 .1.1.1.28.1H9.1l.86-2.48H9.7a2.7 2.7 0 0 1-1.76 1.64c-.16.07-.85.3-2.32.3H1.4l3.47-4.28c.06-.1.08-.1.08-.16 0-.04 0-.04-.07-.14L1.65.41h3.94C6.72.4 9.02.47 9.7 2.33h.25L9.1 0H.84C.56 0 .56.02.56.31zm0 0"/>\n              </symbol>\n              <symbol id="l" overflow="visible">\n                <path d="M4.95-4.81c0-.03-.03-.1-.1-.1-.04 0-.04.02-.13.11l-.34.4c-.27-.38-.7-.5-1.13-.5-.98 0-1.86.8-1.86 1.6 0 .1.03.38.24.64.23.27.5.33.96.46.14.03.49.1.6.14.22.04.64.2.64.72 0 .56-.6 1.23-1.38 1.23-.62 0-1.36-.22-1.36-.98 0-.08.02-.24.05-.36v-.03c0-.1-.08-.1-.1-.1-.1 0-.12.02-.13.14L.55-.04C.53-.03.52.01.52.05c0 .03.03.08.1.08.04 0 .05-.02.15-.1l.32-.38c.38.4.91.48 1.33.48 1.06 0 1.9-.9 1.9-1.73 0-.3-.12-.58-.27-.74-.24-.25-.35-.28-1.25-.5l-.44-.1c-.17-.07-.47-.24-.47-.63 0-.56.63-1.11 1.34-1.11.77 0 1.13.42 1.13 1.08l-.03.3c0 .12.08.12.12.12.1 0 .11-.05.13-.16zm0 0"/>\n              </symbol>\n              <symbol id="m" overflow="visible">\n                <path d="M1.56-1.6c.19 0 .75 0 1.13-.13.51-.2.6-.52.6-.72 0-.4-.38-.63-.85-.63-.85 0-1.97.64-1.97 1.85 0 .7.44 1.3 1.25 1.3 1.19 0 1.73-.7 1.73-.79 0-.05-.06-.12-.12-.12-.03 0-.05.01-.11.07-.55.65-1.36.65-1.49.65-.42 0-.7-.29-.7-.85 0-.1 0-.23.1-.62zm-.39-.18c.3-1.02 1.06-1.1 1.27-1.1.3 0 .56.16.56.43 0 .67-1.19.67-1.48.67zm0 0"/>\n              </symbol>\n              <symbol id="n" overflow="visible">\n                <path d="M3.05-2.67c-.25.04-.35.23-.35.39 0 .19.14.26.27.26.15 0 .39-.1.39-.45 0-.47-.53-.6-.9-.6-1.05 0-2.02.96-2.02 1.93 0 .6.4 1.2 1.28 1.2 1.19 0 1.73-.69 1.73-.78 0-.05-.06-.12-.12-.12-.03 0-.05.01-.11.07-.55.65-1.36.65-1.49.65-.5 0-.71-.35-.71-.79 0-.18.09-.95.45-1.43.26-.35.62-.54.98-.54.1 0 .42.02.6.2zm0 0"/>\n              </symbol>\n              <symbol id="o" overflow="visible">\n                <path d="M1.72-2.75h.7c.14 0 .22 0 .22-.16 0-.09-.08-.09-.2-.09h-.66l.25-1.03a.32.32 0 0 0 .03-.1c0-.14-.1-.23-.25-.23-.17 0-.26.13-.33.3-.04.18.05-.16-.26 1.06h-.7C.39-3 .3-3 .3-2.84c0 .09.08.09.2.09h.66L.75-1.11c-.05.17-.1.42-.1.52 0 .4.35.65.74.65C2.17.06 2.61-.9 2.61-1c0-.1-.1-.1-.11-.1-.1 0-.1.02-.16.15-.18.43-.54.82-.92.82-.15 0-.25-.09-.25-.34 0-.06.03-.22.05-.28zm0 0"/>\n              </symbol>\n              <symbol id="p" overflow="visible">\n                <path d="M2.27-4.36c0-.1-.1-.26-.29-.26-.18 0-.39.18-.39.39 0 .1.08.26.28.26s.4-.2.4-.39zM.84-.8c-.03.1-.06.17-.06.3 0 .32.27.57.66.57.69 0 1-.95 1-1.06 0-.1-.1-.1-.11-.1-.1 0-.11.05-.14.13-.16.56-.46.84-.74.84-.14 0-.17-.09-.17-.24 0-.16.05-.29.11-.44l.22-.56c.06-.18.33-.8.34-.9A.5.5 0 0 0 2-2.48c0-.33-.28-.6-.66-.6C.64-3.08.33-2.12.33-2c0 .08.1.08.12.08.1 0 .1-.03.13-.11.17-.6.48-.85.73-.85.11 0 .17.05.17.24 0 .17-.03.27-.2.7zm0 0"/>\n              </symbol>\n              <symbol id="q" overflow="visible">\n                <path d="M3.7-1.86c0-.76-.58-1.22-1.26-1.22-1.03 0-2 .97-2 1.92C.44-.46.94.06 1.7.06c1 0 2-.9 2-1.92zm-2 1.74c-.34 0-.68-.22-.68-.79 0-.28.12-.96.4-1.36.3-.43.7-.6 1.02-.6.37 0 .69.25.69.76 0 .17-.08.86-.4 1.34-.26.43-.67.65-1.03.65zm0 0"/>\n              </symbol>\n              <symbol id="r" overflow="visible">\n                <path d="M.84-.44c-.01.1-.06.27-.06.28 0 .16.13.22.24.22.12 0 .23-.08.28-.14.03-.06.07-.3.12-.44l.14-.62c.05-.16.1-.31.13-.47.08-.28.1-.34.3-.62.18-.29.51-.65 1.04-.65.4 0 .4.36.4.49 0 .42-.29 1.19-.4 1.48-.08.2-.1.27-.1.38 0 .37.29.6.65.6.7 0 1-.96 1-1.07 0-.1-.08-.1-.11-.1-.1 0-.1.05-.13.13-.15.56-.46.84-.73.84-.16 0-.19-.09-.19-.24 0-.16.05-.26.17-.57.08-.22.36-.95.36-1.34 0-.67-.53-.8-.9-.8-.58 0-.97.36-1.17.64-.05-.48-.46-.64-.75-.64-.3 0-.46.22-.55.38-.16.26-.25.65-.25.7 0 .08.1.08.12.08.1 0 .1-.02.14-.2.11-.41.25-.75.52-.75.19 0 .23.15.23.34 0 .12-.06.39-.12.58l-.14.62zm0 0"/>\n              </symbol>\n              <symbol id="s" overflow="visible">\n                <path d="M3-2.63c-.17.05-.3.2-.3.33 0 .17.14.24.24.24.08 0 .34-.05.34-.4 0-.46-.5-.62-.94-.62-1.07 0-1.28.81-1.28 1.02 0 .26.16.44.25.51.17.14.3.18.78.25.16.03.6.11.6.46 0 .12-.08.39-.36.56-.28.16-.63.16-.7.16-.29 0-.68-.07-.85-.3.24-.03.4-.2.4-.4 0-.16-.13-.26-.27-.26C.7-1.08.5-.9.5-.6c0 .42.44.67 1.1.67 1.3 0 1.54-.87 1.54-1.14 0-.64-.7-.76-.97-.81l-.28-.06c-.25-.05-.37-.2-.37-.35 0-.15.12-.34.28-.45.18-.1.42-.13.54-.13.14 0 .5.02.66.25zm0 0"/>\n              </symbol>\n              <symbol id="z" overflow="visible">\n                <path d="M3.4-4.22c.04-.2.05-.22.2-.23h.42c.43 0 .6 0 .78.04.3.1.33.3.33.55 0 .11 0 .2-.05.56l-.02.08c0 .08.05.11.13.11.1 0 .1-.06.12-.17l.19-1.33c0-.1-.08-.1-.2-.1H1.02c-.18 0-.2 0-.24.15L.33-3.33.3-3.2c0 .03.01.1.12.1.1 0 .1-.04.14-.18.4-1.11.64-1.17 1.7-1.17h.29c.22 0 .22 0 .22.06 0 0 0 .05-.04.14L1.81-.58c-.06.25-.08.33-.81.33-.25 0-.31 0-.31.16C.69-.08.7 0 .8 0l.6-.02L2-.03l.64.01.58.02c.06 0 .15 0 .15-.16 0-.09-.06-.09-.28-.09l-.4-.02c-.24-.01-.25-.04-.25-.12 0-.06 0-.06.03-.17zm0 0"/>\n              </symbol>\n              <symbol id="x" overflow="visible">\n                <path d="M3.88-2.77L1.89-4.75c-.12-.13-.14-.14-.22-.14a.2.2 0 0 0-.2.2c0 .07.01.08.12.19l2 2-2 2.02c-.1.1-.12.12-.12.18 0 .13.1.2.2.2.08 0 .1 0 .22-.13l1.99-1.99L5.94-.16c.01.02.08.07.14.07.12 0 .2-.08.2-.2 0-.02 0-.05-.03-.12-.02-.01-1.6-1.57-2.1-2.09l1.83-1.81c.05-.07.2-.19.25-.25 0-.02.05-.07.05-.13 0-.12-.08-.2-.2-.2-.08 0-.11.03-.22.14zm0 0"/>\n              </symbol>\n            </defs>\n            <use x=".11" y="15.07" xlink:href="#a"/>\n            <use x="7.32" y="15.07" xlink:href="#b"/>\n            <use x="12.15" y="15.07" xlink:href="#c"/>\n            <use x="15.75" y="15.07" xlink:href="#d"/>\n            <use x="21.02" y="15.07" xlink:href="#e"/>\n            <use x="24.19" y="15.07" xlink:href="#f"/>\n            <use x="30.87" y="15.07" xlink:href="#g"/>\n            <use x="35.18" y="15.07" xlink:href="#b"/>\n            <use x="40.02" y="15.07" xlink:href="#h"/>\n            <use x="44.79" y="15.07" xlink:href="#i"/>\n            <use x="52.18" y="15.07" xlink:href="#j"/>\n            <use x="63.89" y=".73" xlink:href="#k"/>\n            <use x="74.41" y="11.19" xlink:href="#l"/>\n            <use x="79.68" y="11.19" xlink:href="#m"/>\n            <use x="83.46" y="11.19" xlink:href="#n"/>\n            <use x="87.02" y="11.19" xlink:href="#o"/>\n            <use x="90.03" y="11.19" xlink:href="#p"/>\n            <use x="92.85" y="11.19" xlink:href="#q"/>\n            <use x="96.78" y="11.19" xlink:href="#r"/>\n            <use x="101.7" y="11.19" xlink:href="#s"/>\n            <use x="107.62" y="8.2" xlink:href="#t"/>\n            <use x="115.02" y="8.2" xlink:href="#i"/>\n            <use x="119.66" y="8.2" xlink:href="#u"/>\n            <use x="123.09" y="8.2" xlink:href="#v"/>\n            <use x="128.2" y="8.2" xlink:href="#w"/>\n            <use x="133.94" y="8.2" xlink:href="#c"/>\n            <use x="137.54" y="9.69" xlink:href="#s"/>\n            <use x="141.3" y="9.69" xlink:href="#m"/>\n            <use x="145.08" y="9.69" xlink:href="#n"/>\n            <use x="148.65" y="9.69" xlink:href="#o"/>\n            <use x="151.65" y="9.69" xlink:href="#p"/>\n            <use x="154.47" y="9.69" xlink:href="#q"/>\n            <use x="158.4" y="9.69" xlink:href="#r"/>\n            <use x="166.04" y="8.2" xlink:href="#x"/>\n            <use x="176" y="8.2" xlink:href="#y"/>\n            <use x="179.87" y=".73" xlink:href="#k"/>\n            <use x="190.39" y="11.19" xlink:href="#z"/>\n            <use x="196.17" y="11.19" xlink:href="#m"/>\n            <use x="199.95" y="11.19" xlink:href="#s"/>\n            <use x="203.71" y="11.19" xlink:href="#o"/>\n            <use x="206.71" y="11.19" xlink:href="#s"/>\n            <use x="212.64" y="8.2" xlink:href="#t"/>\n            <use x="220.03" y="8.2" xlink:href="#i"/>\n            <use x="224.67" y="8.2" xlink:href="#u"/>\n            <use x="228.1" y="8.2" xlink:href="#v"/>\n            <use x="233.21" y="8.2" xlink:href="#w"/>\n            <use x="238.95" y="8.2" xlink:href="#c"/>\n            <use x="242.55" y="9.69" xlink:href="#o"/>\n            <use x="245.56" y="9.69" xlink:href="#m"/>\n            <use x="249.34" y="9.69" xlink:href="#s"/>\n            <use x="253.1" y="9.69" xlink:href="#o"/>\n            <use x="258.82" y="8.2" xlink:href="#x"/>\n            <use x="268.79" y="8.2" xlink:href="#A"/>\n            <use x="273.46" y="8.2" xlink:href="#g"/>\n            <use x="277.77" y="8.2" xlink:href="#b"/>\n            <use x="282.6" y="8.2" xlink:href="#h"/>\n            <use x="287.37" y="8.2" xlink:href="#i"/>\n            <use x="292.01" y="9.69" xlink:href="#o"/>\n            <use x="295.01" y="9.69" xlink:href="#m"/>\n            <use x="298.79" y="9.69" xlink:href="#s"/>\n            <g>\n              <use x="302.56" y="9.69" xlink:href="#o"/>\n            </g>\n            <g>\n              <use x="306.07" y="8.2" xlink:href="#B"/>\n            </g>\n            <path fill="none" stroke="#000" stroke-miterlimit="10" stroke-width=".405" d="M63.89 12.58h246.06"/>\n            <g>\n              <use x="163.44" y="21.91" xlink:href="#C"/>\n            </g>\n            <g>\n              <use x="174.2" y="21.91" xlink:href="#d"/>\n            </g>\n            <g>\n              <use x="179.46" y="21.91" xlink:href="#D"/>\n            </g>\n            <g>\n              <use x="185.16" y="21.91" xlink:href="#f"/>\n            </g>\n            <g>\n              <use x="191.85" y="21.91" xlink:href="#g"/>\n            </g>\n            <g>\n              <use x="196.16" y="21.91" xlink:href="#b"/>\n            </g>\n            <g>\n              <use x="200.99" y="21.91" xlink:href="#h"/>\n            </g>\n            <g>\n              <use x="205.76" y="21.91" xlink:href="#i"/>\n            </g>\n         </svg>          \n         <br>\n    <p> Weights for sections and individual tests are indicated by a white number inside a margenta badge. Like so: </p>\n        <mat-expansion-panel>\n            <mat-expansion-panel-header>\n                <mat-panel-title fxLayoutAlign=\'space-between center\'>\n                    <td class="test-title"> Test Title </td>\n                    <td class="MatButtonEmulation"> 100% </td>\n                </mat-panel-title>\n                <div class=\'badge\'> <span class=\'vert-align\'> X3 </span> </div>\n            </mat-expansion-panel-header>\n          </mat-expansion-panel>\n    <p> No badge means that the weight is either equal to 1 or undefined.</p>\n    </mat-card-content>\n    </mat-card>\n</mat-dialog-content>\n<mat-dialog-actions>\n        <button mat-button mat-dialog-close>Close</button>\n</mat-dialog-actions>'},"./src/app/help-dialog/help-dialog.component.html":function(t,e){t.exports='<button mat-button (click)="openDialog()">Readme</button>'},"./src/app/help-dialog/help-dialog.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL2hlbHAtZGlhbG9nL2hlbHAtZGlhbG9nLmNvbXBvbmVudC5zY3NzIn0= */"},"./src/app/help-dialog/help-dialog.component.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"HelpDialogComponent",function(){return i}),a.d(e,"HelpDialogTextComponent",function(){return r});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=a("./node_modules/@angular/material/esm5/material.es5.js"),i=function(){function t(t){this.dialog=t}return t.prototype.openDialog=function(){this.dialog.open(r,{height:"90%",width:"70%"}).afterClosed().subscribe(function(t){console.log("Dialog result: "+t)})},t=n.__decorate([Object(s.Component)({selector:"app-help-dialog",template:a("./src/app/help-dialog/help-dialog.component.html"),styles:[a("./src/app/help-dialog/help-dialog.component.scss")]}),n.__metadata("design:paramtypes",[o.MatDialog])],t)}(),r=function(){function t(){}return t=n.__decorate([Object(s.Component)({selector:"app-help-dialog-text",template:a("./src/app/help-dialog/help-dialog-text.component.html")})],t)}()},"./src/app/keys.pipe.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"KeysPipe",function(){return o});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=function(){function t(){}return t.prototype.transform=function(t,e){return void 0===e&&(e=null),Object.keys(t)},t=n.__decorate([Object(s.Pipe)({name:"keys",pure:!1})],t)}()},"./src/app/report-data.service.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"ReportDataService",function(){return d});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=a("./node_modules/@angular/common/fesm5/http.js"),i=a("./src/app/resultcard.model.ts"),r=a("./src/app/test-result.model.ts"),c=a("./src/app/test-history.model.ts"),l=a("./src/app/test-diff.model.ts"),d=function(){function t(t){this.http=t,this.allTests=[],this.statisticsCards=[],this.scoredTests=[],this.allExpandState=!1,this.scorePerSection={},this.testWeights={},this.testWeightsExpanded={}}return t.prototype.loadResults=function(){var t=this;switch(this.reportType=window.reportType,this.reportType){case"history":this.convertHistoryResults(window.data);break;case"snapshot":this.convertResults(window.data);break;case"diff":this.convertDiffResults(window.data);break;default:this.http.get("/data/testData.json").subscribe(function(e){t.convertResults(e)}),this.reportType="snapshot"}},t.prototype.togglePanel=function(){this.allExpandState=!this.allExpandState},t.prototype.byID=function(t){return this.allTests.find(function(e){return e.id===t})},t.prototype.byReg=function(t){for(var e=new RegExp(t),a=[],n=0,s=this.allTests;n<s.length;n++){var o=s[n];e.test(o.id)&&a.push(o.id)}return a},t.prototype.isScored=function(t){return this.scoredTests.includes(t)},t.prototype.getParam=function(t,e){return t.split(":")[e]},t.prototype.getString=function(t){return JSON.stringify(t)},t.prototype.convertResults=function(t){for(var e=0,a=Object.keys(t.tests);e<a.length;e++){var n=a[e];if(t.tests[n].message instanceof Object)for(var s=0,o=Object.keys(t.tests[n].result);s<o.length;s++){var i=o[s],c=n+":"+i;this.allTests.push(new r.TestResult(c,{data:t.tests[n].data[i],duration:t.tests[n].duration[i],message:t.tests[n].message[i],metric:t.tests[n].metric[i],result:t.tests[n].result[i],summary:t.tests[n].summary,title:t.tests[n].title,format_type:t.tests[n].format_type}))}else this.allTests.push(new r.TestResult(n,t.tests[n]))}this.extractMetadata(t),this.extractScoring(t),this.extractTestWeights(t),this.distributeCardsToSections(t),this.determineScoredTests(),this.determineScorePerSection()},t.prototype.convertHistoryResults=function(t){for(var e=0,a=Object.keys(t.tests);e<a.length;e++){var n=a[e];if(t.tests[n].history instanceof Array)this.allTests.push(new c.TestHistory(n,t.tests[n]));else for(var s=0,o=Object.keys(t.tests[n].history);s<o.length;s++){var i=o[s],r=n+":"+i;this.allTests.push(new c.TestHistory(r,{history:t.tests[n].history[i],summary:t.tests[n].summary,title:t.tests[n].title,format_type:t.tests[n].format_type}))}}this.distributeCardsToSections(t),this.determineScoredTests(),this.extractScoring(t)},t.prototype.convertDiffResults=function(t){for(var e=0,a=Object.keys(t.tests);e<a.length;e++){var n=a[e];if(t.tests[n].diff instanceof Array)this.allTests.push(new l.TestDiff(n,t.tests[n]));else for(var s=0,o=Object.keys(t.tests[n].diff);s<o.length;s++){var i=o[s],r=n+":"+i;this.allTests.push(new l.TestDiff(r,{diff:t.tests[n].diff[i],summary:t.tests[n].summary,title:t.tests[n].title,format_type:t.tests[n].format_type}))}}this.distributeCardsToSections(t),this.determineScoredTests(),this.extractScoring(t),this.extractTestWeights(t),this.determineScorePerSection()},t.prototype.extractMetadata=function(t){this.metaData=t.meta},t.prototype.extractScoring=function(t){this.score=t.score},t.prototype.extractTestWeights=function(t){this.testWeights=t.weights;for(var e=0,a=Object.keys(this.testWeights);e<a.length;e++)for(var n=a[e],s=0,o=this.byReg(n);s<o.length;s++){var i=o[s];this.testWeightsExpanded[i]=this.testWeights[n]}},t.prototype.distributeCardsToSections=function(t){for(var e=0,a=Object.keys(t.cards);e<a.length;e++){var n=a[e];"scored"===n?this.scoredCard=t.cards.scored:this.statisticsCards.push(new i.ResultCard(n,t.cards[n]))}},t.prototype.determineScoredTests=function(){for(var t=0,e=Object.keys(this.scoredCard.sections);t<e.length;t++){var a=e[t];if(this.scoredCard.sections[a].cases instanceof Array)for(var n=0,s=this.scoredCard.sections[a].cases;n<s.length;n++){var o=s[n];this.scoredTests.push(o)}}},t.prototype.extractData=function(t){for(var e={},a=0,n=this.byID(t).diff;a<n.length;a++){var s=n[a];e[s.model]=s.data}return this.getString(e)},t.prototype.determineScorePerSection=function(){var t=[],e=[];if("snapshot"===this.reportType)e=this.score.sections,t=[this.score.total_score];else{e=this.score.sections.diff;for(var a=0,n=this.score.total_score.diff;a<n.length;a++){var s=n[a];t.push(s.total_score)}}for(var o=[],i=[],r=0,c=e;r<c.length;r++){var l=c[r],d=l.section;this.scorePerSection.hasOwnProperty(d)?(this.scorePerSection[d].weights.push(this.scoredCard.sections[d].weight),this.scorePerSection[d].scores.push(l.score)):(o.push(this.scoredCard.sections[d].weight),i.push(l.score),this.scorePerSection[d]={weights:o,scores:i}),o=[],i=[]}this.scorePerSection["Total Score"]={scores:t}},t=n.__decorate([Object(s.Injectable)(),n.__metadata("design:paramtypes",[o.HttpClient])],t)}()},"./src/app/result-button/result-button.component.html":function(t,e){t.exports="\x3c!-- First distinguish wether the tests have passed/failed or skipped or errored --\x3e\n<div *ngIf=\"data.byID(testId)?.result === 'passed' || data.byID(testId)?.result === 'failed'\">\n  \x3c!-- Then determine if the test results are from scored or from unscored tests --\x3e\n  <div *ngIf=\"data.isScored(data.getParam(testId, 0)); else notScored\">\n    \x3c!-- If they are from scored tests colorize the result button --\x3e\n    <div\n    class=\"MatButtonEmulation\"\n    [ngStyle]=\"{'background-color':getColor(1 - data.byID(testId)?.metric | number : '1.2-2')}\">\n      {{ 1 - data.byID(testId)?.metric | percent : '1.1' }}\n      \x3c!-- {{ 1 - data.byID(testId)?.metric | percent : '1.2-2' }} --\x3e\n  </div>\n  </div>\n  <ng-template #notScored>\n    \x3c!-- If they are from unscored tests the user has some control over how --\x3e\n    \x3c!-- they are displayed using the 'format_type' attribute of each test. --\x3e\n    <div\n    class=\"MatButtonEmulation\">\n      \x3c!-- Hence for each test, determine now how the result ought to be formatted --\x3e\n      <div *ngIf=\"data.byID(testId)?.format_type === 'count'\">\n        {{ data.byID(testId)?.data.length | number : '1.'}}\n      </div>\n      <div *ngIf=\"data.byID(testId)?.format_type === 'number'\">\n        {{ data.byID(testId)?.data | number : '1.2-2'}}\n      </div>\n      <div *ngIf=\"data.byID(testId)?.format_type === 'percent'\">\n        {{ data.byID(testId)?.metric | number : '1.2-2'}}\n      </div>\n      <div *ngIf=\"data.byID(testId)?.format_type === 'raw'\">\n        {{ data.byID(testId)?.data }}\n      </div>\n    </div>\n  </ng-template>\n</div>\n<div *ngIf=\"data.byID(testId)?.result === 'skipped'\">\n\x3c!-- Skipped tests are marked with a gray button with beige text --\x3e\n    <div\n     class=\"MatButtonEmulation\"\n     style=\"background:gray; color:beige\">\n       Skipped\n     </div>\n</div>\n<div *ngIf=\"data.byID(testId)?.result === 'error'\">\n\x3c!-- Tests with errors are marked with a black button with red text --\x3e\n    <div\n     class=\"MatButtonEmulation\"\n     style=\"color:red;background-color:black\">\n       Errored\n    </div>\n</div>\n"},"./src/app/result-button/result-button.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL3Jlc3VsdC1idXR0b24vcmVzdWx0LWJ1dHRvbi5jb21wb25lbnQuc2NzcyJ9 */"},"./src/app/result-button/result-button.component.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"ResultButtonComponent",function(){return i});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=a("./src/app/report-data.service.ts"),i=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},t.prototype.getColor=function(t){return["hsl(",(132*t).toString(10),",80%,35%)"].join("")},n.__decorate([Object(s.Input)(),n.__metadata("design:type",String)],t.prototype,"testId",void 0),t=n.__decorate([Object(s.Component)({selector:"app-result-button",template:a("./src/app/result-button/result-button.component.html"),encapsulation:s.ViewEncapsulation.None,styles:[a("./src/app/result-button/result-button.component.scss")]}),n.__metadata("design:paramtypes",[o.ReportDataService])],t)}()},"./src/app/resultcard.model.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"ResultCard",function(){return n});var n=function(){return function(t,e){var a=e.title,n=e.cases;this.cardID=t,this.cardName=a,this.associatedTests=n}}()},"./src/app/score-formula/score-formula.component.html":function(t,e){t.exports='<div>\n    <mat-expansion-panel\n    #expPanel\n    class="mat-elevation-z0">\n       <mat-expansion-panel-header\n       class="mat-expansion-panel-header"\n       collapsedHeight="auto"\n       style="padding:0.05vh">\n       <mat-panel-title fxLayoutAlign=\'space-between center\'>\n          <td class="test-title">\n              {{ scoreType }}\n          </td>\n          <div style="margin-right: 10px">\n          <td *ngFor=\'let score of modelScore\'>\n            <div class="MatButtonEmulation" style="background: white; color:black"> {{ score | percent : \'1.0\' }} </div>\n          </td>\n          </div>\n         </mat-panel-title>\n         <div *ngIf=\'getSectionWeight() ; else whiteBadge\'> \n            <div class=\'badge\'> <span class="vert-align"> X{{ sectionWeight }} </span> </div> \n          </div>\n          <ng-template #whiteBadge>\n              <div class=\'badge\' style="background-color: white"></div> \n          </ng-template>\n       </mat-expansion-panel-header>\n       <div>\n         <p>The {{ scoreType}} is the result of the following calculation. For more information please click on "Readme" in the top left of the report.</p>\n       </div>\n       <div *ngFor=\'let formula of formulae\'>\n          <p style="font-size: 36px; font-size: 0.6vmax"> {{ formula }} </p>\n        </div>\n      </mat-expansion-panel>  \n</div>\n'},"./src/app/score-formula/score-formula.component.scss":function(t,e){t.exports="\n/*# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbXSwibmFtZXMiOltdLCJtYXBwaW5ncyI6IiIsImZpbGUiOiJzcmMvYXBwL3Njb3JlLWZvcm11bGEvc2NvcmUtZm9ybXVsYS5jb21wb25lbnQuc2NzcyJ9 */"},"./src/app/score-formula/score-formula.component.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"ScoreFormulaComponent",function(){return i});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=a("./src/app/report-data.service.ts"),i=function(){function t(t){this.data=t,this.scoreType="Total Score",this.modelScore=[],this.formulae=[],this.sections=[],this.sectionWeight=1}return t.prototype.ngOnInit=function(){this.getScore(),this.constructFormulae()},t.prototype.getSectionWeight=function(){if("Total Score"!==this.position){var t=this.data.scorePerSection[this.position].weights[0];return 1!==t&&void 0!==t&&(this.sectionWeight=t,!0)}},t.prototype.getScore=function(){this.modelScore=this.data.scorePerSection[this.position].scores,"Total Score"!==this.position&&(this.scoreType="Sub Total")},t.prototype.constructFormulae=function(){var t;for(t=0;t<this.data.scorePerSection["Total Score"].scores.length;t++)if("Total Score"===this.scoreType){var e=this.constructTotalFormula(t);this.formulae.push(e)}else{e=this.constructSubTotalFormula(t);this.formulae.push(e)}},t.prototype.constructTotalFormula=function(t){for(var e="",a="",n=!0,s="",o=0,i=Object.keys(this.data.scorePerSection);o<i.length;o++){var r=i[o];if("Total Score"!==r){var c=this.data.scorePerSection[r].weights[t];e+=s+"("+c+" * "+(100*this.data.scorePerSection[r].scores[t]).toFixed(2)+")",a+=s+"("+c+" * 100)"}n&&(s=" + ",n=!1)}return"`("+e+")/("+a+") = "+(100*this.data.scorePerSection["Total Score"].scores[t]).toFixed(2)+"`"},t.prototype.constructSubTotalFormula=function(t){for(var e="",a="",n=0,s=0,o=0,i=0,r=0,c=this.data.scoredCard.sections[this.position].cases;r<c.length;r++){var l=c[r];if(l in Object.keys(this.data.testWeights))this.data.testWeights[l];for(var d=0,p=this.data.byReg(l);d<p.length;d++){var u=p[d];0!==(i="snapshot"===this.data.reportType?100*(1-this.data.byID(u).metric):100*(1-this.data.byID(u).diff[t].metric))?100!==i?(e+=i.toFixed(2)+" + ",n+=1):(s+=1,n+=1):(n+=1,o+=1)}}return s>0&&(e+="("+s+" * 100) + "),o>0&&(e+="("+o+" * 0)"),"`("+e+")/("+(a+="("+n+" * 100)")+") = "+(100*this.data.scorePerSection[this.position].scores[t]).toFixed(2)+"`"},n.__decorate([Object(s.Input)(),n.__metadata("design:type",String)],t.prototype,"position",void 0),t=n.__decorate([Object(s.Component)({selector:"app-score-formula",template:a("./src/app/score-formula/score-formula.component.html"),encapsulation:s.ViewEncapsulation.None,styles:[a("./src/app/score-formula/score-formula.component.scss")]}),n.__metadata("design:paramtypes",[o.ReportDataService])],t)}()},"./src/app/test-diff.model.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"TestDiff",function(){return n});var n=function(){return function(t,e){var a=e.diff,n=e.summary,s=e.title,o=e.format_type;this.id=t,this.diff=a,this.summary=n,this.title=s,this.format_type=o}}()},"./src/app/test-history.model.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"TestHistory",function(){return n});var n=function(){return function(t,e){var a=e.history,n=e.summary,s=e.title,o=e.format_type;this.id=t,this.history=a,this.summary=n,this.title=s,this.format_type=o}}()},"./src/app/test-result.model.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"TestResult",function(){return n});var n=function(){function t(t,e){var a=e.data,n=e.duration,s=e.message,o=e.metric,i=e.result,r=e.summary,c=e.title,l=e.format_type;this.id=t,this.data=a,this.duration=n,this.message=s,this.metric=o,this.errorFailsafe(i),this.summary=r,this.title=c,this.format_type=l}return t.prototype.errorFailsafe=function(t){"skipped"===t||null!==this.data&&void 0!==this.data?this.result=t:this.result="error"},t}()},"./src/app/vega-bar-chart/vega-bar-chart.bar-chart-spec.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"specBarChart",function(){return n});var n={$schema:"https://vega.github.io/schema/vega/v4.json",width:300,height:200,autosize:"fit",padding:"strict",data:[{name:"table",values:[],transform:[{type:"formula",as:"percent",expr:"round(datum.score * 100)"}]}],signals:[],scales:[{name:"xscale",type:"band",domain:{data:"table",field:"section"},range:"width",padding:.05,round:!0},{name:"yscale",domain:{data:"table",field:"percent"},nice:!0,range:"height"}],axes:[{orient:"bottom",scale:"xscale"},{orient:"left",scale:"yscale"}],marks:[{type:"rect",from:{data:"table"},encode:{enter:{x:{scale:"xscale",field:"section"},width:{scale:"xscale",band:1},y:{scale:"yscale",field:"percent"},y2:{scale:"yscale",value:0}}}},{type:"text",from:{data:"table"},encode:{enter:{text:{field:"percent"},x:{scale:"xscale",field:"section",band:.5},y:{scale:"yscale",field:"percent",offset:-2},fill:{value:"black"},align:{value:"center"},baseline:{value:"bottom"},fontSize:{value:18}}}}]}},"./src/app/vega-bar-chart/vega-bar-chart.component.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"VegaBarChartComponent",function(){return c});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=a("./node_modules/vega-lib/index.js"),i=a("./src/app/report-data.service.ts"),r=a("./src/app/vega-bar-chart/vega-bar-chart.bar-chart-spec.ts"),c=function(){function t(t,e){this.data=t,this.elementRef=e,this.nativeElement=e.nativeElement}return t.prototype.initialize=function(){r.specBarChart.data[0].values=this.data.score.sections;new o.View(Object(o.parse)(r.specBarChart)).renderer("svg").initialize(this.nativeElement).hover().logLevel(o.Warn).run()},t.prototype.ngOnInit=function(){this.initialize()},n.__decorate([Object(s.Input)(),n.__metadata("design:type",String)],t.prototype,"testId",void 0),t=n.__decorate([Object(s.Component)({selector:"app-vega-bar-chart",template:"",encapsulation:s.ViewEncapsulation.None}),n.__metadata("design:paramtypes",[i.ReportDataService,s.ElementRef])],t)}()},"./src/app/vega-grouped-bar-chart/vega-grouped-bar-chart.bar-chart-spec.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"specGroupedBarChart",function(){return n});var n={$schema:"https://vega.github.io/schema/vega/v4.json",width:300,height:240,padding:5,data:[{name:"diff",values:[],transform:[{type:"formula",as:"percent",expr:"round(datum.score * 100)"}]}],legends:[{columns:0,orient:"bottom",fill:"color",encode:{title:{update:{fontSize:{value:14}}},labels:{interactive:!1}}}],scales:[{name:"yscale",type:"band",domain:{data:"diff",field:"model"},range:"height",padding:.2},{name:"xscale",type:"linear",domain:{data:"diff",field:"percent"},range:"width",round:!0,zero:!0,nice:!0},{name:"color",type:"ordinal",domain:{data:"diff",field:"section"},range:{scheme:"category20"}}],axes:[{orient:"left",scale:"yscale",tickSize:0,labelPadding:4,zindex:1},{orient:"bottom",scale:"xscale"}],marks:[{type:"group",from:{facet:{data:"diff",name:"facet",groupby:"model"}},encode:{enter:{y:{scale:"yscale",field:"model"}}},signals:[{name:"height",update:'bandwidth("yscale")'}],scales:[{name:"pos",type:"band",range:"height",domain:{data:"facet",field:"section"}}],marks:[{name:"bars",from:{data:"facet"},type:"rect",encode:{enter:{y:{scale:"pos",field:"section"},height:{scale:"pos",band:1},x:{scale:"xscale",field:"percent"},x2:{scale:"xscale",percent:0},fill:{scale:"color",field:"section"}}}},{type:"text",from:{data:"bars"},encode:{enter:{x:{field:"x2",band:1,offset:4},y:{field:"y",offset:{field:"height",mult:.5}},fill:{value:"black"},align:{value:"left"},baseline:{value:"middle"},text:{field:"datum.percent"},fontSize:{value:18}}}}]}]}},"./src/app/vega-grouped-bar-chart/vega-grouped-bar-chart.component.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"VegaGroupedBarChartComponent",function(){return c});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=a("./node_modules/vega-lib/index.js"),i=a("./src/app/report-data.service.ts"),r=a("./src/app/vega-grouped-bar-chart/vega-grouped-bar-chart.bar-chart-spec.ts"),c=function(){function t(t,e){this.data=t,this.elementRef=e,this.nativeElement=e.nativeElement}return t.prototype.initialize=function(){r.specGroupedBarChart.data[0].values=this.data.score.sections.diff;new o.View(Object(o.parse)(r.specGroupedBarChart)).renderer("svg").initialize(this.nativeElement).hover().logLevel(o.Warn).run()},t.prototype.ngOnInit=function(){this.initialize()},n.__decorate([Object(s.Input)(),n.__metadata("design:type",Object)],t.prototype,"nativeElement",void 0),t=n.__decorate([Object(s.Component)({selector:"app-vega-grouped-bar-chart",template:"",encapsulation:s.ViewEncapsulation.None}),n.__metadata("design:paramtypes",[i.ReportDataService,s.ElementRef])],t)}()},"./src/app/vega-plot/vega-plot.component.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"VegaPlotComponent",function(){return c});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=a("./node_modules/vega-lib/index.js"),i=a("./src/app/report-data.service.ts"),r=a("./src/app/vega-plot/vega-plot.metric-spec.ts"),c=function(){function t(t,e){this.data=t,this.elementRef=e,this.nativeElement=e.nativeElement}return t.prototype.initialize=function(){this.format_type=this.testObject.format_type,["count","raw","number"].includes(this.format_type)?this.format_type="data":this.format_type="metric",r.specMetric.data[0].values=this.testObject.history,new o.View(Object(o.parse)(r.specMetric)).renderer("svg").signal("type",this.format_type).initialize(this.nativeElement).hover().run().addEventListener("click",function(t,e){e&&e.datum&&e.datum.commit&&alert("Navigating to snapshort report for commit"+e.datum.commit)})},t.prototype.ngOnInit=function(){this.initialize()},n.__decorate([Object(s.Input)(),n.__metadata("design:type",Object)],t.prototype,"testObject",void 0),t=n.__decorate([Object(s.Component)({selector:"app-vega-plot",template:"",encapsulation:s.ViewEncapsulation.None}),n.__metadata("design:paramtypes",[i.ReportDataService,s.ElementRef])],t)}()},"./src/app/vega-plot/vega-plot.metric-spec.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"specMetric",function(){return n});var n={$schema:"https://vega.github.io/schema/vega/v3.0.json",width:350,height:200,padding:0,signals:[{name:"shift",value:!1,on:[{events:"@legendSymbol:click, @legendLabel:click",update:"event.shiftKey",force:!0}]},{name:"clicked",value:null,on:[{events:"@legendSymbol:click, @legendLabel:click",update:"{value: datum.value}",force:!0}]},{name:"type",value:"data",description:"This keyword describes which data is displayed on the y-axis."}],data:[{name:"scores",values:[]},{name:"selected",on:[{trigger:"clicked",toggle:"clicked"}]}],scales:[{name:"x",type:"point",range:"width",domain:{data:"scores",field:"commit"}},{name:"y",type:"linear",range:"height",nice:!0,zero:!0,domain:{data:"scores",field:{signal:"type"}}},{name:"color",type:"ordinal",range:["#000000","#E69F00","#56B4E9","#009E73","#F0E442","#0072B2","#D55E00","#CC79A7"],domain:{data:"scores",field:"branch"}}],axes:[{orient:"bottom",scale:"x",title:"History",encode:{labels:{interactive:!0,update:{angle:{value:50},limit:{value:50},baseline:{value:"top"},dx:{value:3},align:{value:"left"}}}}},{orient:"left",scale:"y",title:{signal:"type === 'metric' ? 'Metric': 'Data'"}}],marks:[{type:"group",from:{facet:{name:"series",data:"scores",groupby:"branch"}},legends:[{stroke:"color",title:"Branches",padding:4,orient:"left",offset:30,encode:{symbols:{name:"legendSymbol",interactive:!0,enter:{cursor:{value:"pointer"}},update:{fill:{value:"transparent"},strokeWidth:{value:2},opacity:[{test:"!length(data('selected')) || indata('selected', 'value', datum.value)",value:.7},{value:.15}],size:{value:64}}},labels:{name:"legendLabel",interactive:!0,enter:{cursor:{value:"pointer"}},update:{opacity:[{test:"!length(data('selected')) || indata('selected', 'value', datum.value)",value:1},{value:.25}]}}}}],marks:[{type:"line",from:{data:"series"},encode:{enter:{x:{scale:"x",field:"commit"},y:{scale:"y",field:{signal:"type"}},stroke:{scale:"color",field:"branch"},strokeWidth:{value:2}},update:{interpolate:{value:"linear"},opacity:[{test:"!length(data('selected')) || indata('selected', 'value', datum.branch)",value:1},{value:0}]}}},{type:"symbol",from:{data:"series"},encode:{enter:{x:{scale:"x",field:"commit"},y:{scale:"y",field:{signal:"type"}},fill:{scale:"color",field:"branch"},size:{value:90},tooltip:{scale:"x",field:"commit"}},update:{opacity:[{test:"!length(data('selected')) || indata('selected', 'value', datum.branch)",value:1},{value:0}]},hover:{opacity:[{test:"!length(data('selected')) || indata('selected', 'value', datum.branch)",value:.5},{value:0}],cursor:{value:"pointer"}}}}]}]}},"./src/app/vega-total-score-bar-chart/vega-total-score-bar-chart.component.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"VegaTotalScoreBarChartComponent",function(){return c});var n=a("./node_modules/tslib/tslib.es6.js"),s=a("./node_modules/@angular/core/fesm5/core.js"),o=a("./node_modules/vega-lib/index.js"),i=a("./src/app/report-data.service.ts"),r=a("./src/app/vega-total-score-bar-chart/vega-total-score-bar-chart.total-score-bar-chart-spec.ts"),c=function(){function t(t,e){this.data=t,this.elementRef=e,this.nativeElement=e.nativeElement}return t.prototype.initialize=function(){r.specTotalScoreBarChart.data[0].values=this.data.score.total_score.diff;new o.View(Object(o.parse)(r.specTotalScoreBarChart)).renderer("svg").initialize(this.nativeElement).hover().logLevel(o.Warn).run()},t.prototype.ngOnInit=function(){this.initialize()},n.__decorate([Object(s.Input)(),n.__metadata("design:type",String)],t.prototype,"testId",void 0),t=n.__decorate([Object(s.Component)({selector:"app-vega-total-score-bar-chart",template:"",encapsulation:s.ViewEncapsulation.None}),n.__metadata("design:paramtypes",[i.ReportDataService,s.ElementRef])],t)}()},"./src/app/vega-total-score-bar-chart/vega-total-score-bar-chart.total-score-bar-chart-spec.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"specTotalScoreBarChart",function(){return n});var n={$schema:"https://vega.github.io/schema/vega/v4.json",width:300,height:240,autosize:"fit",padding:"strict",data:[{name:"table",values:[],transform:[{type:"formula",as:"percent",expr:"round(datum.total_score * 100)"}]}],scales:[{name:"xscale",type:"band",domain:{data:"table",field:"model"},range:"width",padding:.05,round:!0},{name:"yscale",domain:{data:"table",field:"percent"},nice:!0,range:"height"}],axes:[{orient:"bottom",scale:"xscale"},{orient:"left",scale:"yscale"}],marks:[{type:"rect",from:{data:"table"},encode:{enter:{x:{scale:"xscale",field:"model"},width:{scale:"xscale",band:1},y:{scale:"yscale",field:"percent"},y2:{scale:"yscale",value:0}}}},{type:"text",from:{data:"table"},encode:{enter:{text:{field:"percent"},x:{scale:"xscale",field:"model",band:.5},y:{scale:"yscale",field:"percent",offset:-2},fill:{value:"black"},align:{value:"center"},baseline:{value:"bottom"},fontSize:{value:18}}}}]}},"./src/environments/environment.ts":function(t,e,a){"use strict";a.r(e),a.d(e,"environment",function(){return n});var n={production:!1}},"./src/main.ts":function(t,e,a){"use strict";a.r(e);var n=a("./node_modules/@angular/core/fesm5/core.js"),s=a("./node_modules/@angular/platform-browser-dynamic/fesm5/platform-browser-dynamic.js"),o=a("./src/app/app.module.ts");a("./src/environments/environment.ts").environment.production&&Object(n.enableProdMode)(),Object(s.platformBrowserDynamic)().bootstrapModule(o.AppModule).catch(function(t){return console.log(t)})},0:function(t,e,a){t.exports=a("./src/main.ts")},1:function(t,e){},2:function(t,e){},3:function(t,e){}},[[0,"runtime","vendor"]]]);
</script></body>
</html>
